/* Generated by Cython 3.0.4 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "extra_compile_args": [
            "-g",
            "-O2"
        ],
        "extra_link_args": [
            "-g"
        ],
        "name": "cavro",
        "sources": [
            "cavro.pyx"
        ]
    },
    "module_name": "cavro"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif
    

    #if (PY_VERSION_HEX < 0x030700b1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM < 0x07030600)) && !defined(PyContextVar_Get)
    #define PyContextVar_Get(var, d, v)         ((d) ?             ((void)(var), Py_INCREF(d), (v)[0] = (d), 0) :             ((v)[0] = NULL, 0)         )
    #endif
    
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_4" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x030004F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info; // borrowed
        #endif
        PyObject *py_minor_version = NULL;
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11; // we don't yet need to distinguish between versions > 11
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        Py_XDECREF(py_minor_version);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);  // we don't have access to __pyx_empty_bytes here
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE(obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__cavro
#define __PYX_HAVE_API__cavro
/* Early includes */
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include "pythread.h"

    #if CYTHON_COMPILING_IN_PYPY
    #ifdef _MSC_VER
    #pragma message ("This module uses CPython specific internals of 'array.array', which are not available in PyPy.")
    #else
    #warning This module uses CPython specific internals of 'array.array', which are not available in PyPy.
    #endif
    #endif
    
#include <float.h>
#include <stdlib.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const wchar_t *u)
{
    const wchar_t *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#endif
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else  // Py < 3.12
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "src/error.pxi",
  "<stringsource>",
  "src/type.pxi",
  "src/codec.pxi",
  "src/option.pxi",
  "src/io.pxi",
  "src/tests/test_zigzag.pxi",
  "src/tests/test_perf.pxi",
  "src/enum.pxi",
  "src/map.pxi",
  "src/numeric_types.pxi",
  "src/record.pxi",
  "src/string_types.pxi",
  "cavro.pyx",
  "contextvars.pxd",
  "array.pxd",
  "src/buffer.pxi",
  "src/zigzag.pxi",
  "src/rabin.pxi",
  "src/tests/test_buffer.pxi",
  "src/array.pxi",
  "src/null.pxi",
  "src/union.pxi",
  "src/schema.pxi",
  "src/logical.pxi",
  "src/container.pxi",
  "src/promotions.pxi",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* BufferFormatStructs.proto */
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_incr_aligned(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_decr_aligned(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_incr_aligned(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_decr_aligned(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_incr_aligned(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #pragma intrinsic (_InterlockedExchangeAdd)
    #define __pyx_atomic_incr_aligned(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_aligned(value) _InterlockedExchangeAdd(value, -1)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview))
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview))
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* #### Code section: numeric_typedefs ### */

/* "cavro.pyx":36
 * from libc.string cimport memcpy
 * 
 * ctypedef bint bool             # <<<<<<<<<<<<<<
 * 
 * if cython.sizeof(Py_ssize_t) == 8:
 */
typedef int __pyx_t_5cavro_bool;
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
#ifndef _ARRAYARRAY_H
struct arrayobject;
typedef struct arrayobject arrayobject;
#endif
struct __pyx_obj_5cavro_Options;
struct __pyx_obj_5cavro__Writer;
struct __pyx_obj_5cavro__Reader;
struct __pyx_obj_5cavro_FileReader;
struct __pyx_obj_5cavro_FileWriter;
struct __pyx_obj_5cavro_MemoryWriter;
struct __pyx_obj_5cavro_MemoryReader;
struct __pyx_obj_5cavro_Rabin;
struct __pyx_obj_5cavro_Record;
struct __pyx_obj_5cavro_RecordField;
struct __pyx_obj_5cavro__FieldAccessor;
struct __pyx_obj_5cavro__Sentinel;
struct __pyx_obj_5cavro__CanonicalForm;
struct __pyx_obj_5cavro_AvroType;
struct __pyx_obj_5cavro_ArrayType;
struct __pyx_obj_5cavro_MapType;
struct __pyx_obj_5cavro_NullType;
struct __pyx_obj_5cavro_BoolType;
struct __pyx_obj_5cavro_IntType;
struct __pyx_obj_5cavro_LongType;
struct __pyx_obj_5cavro_FloatType;
struct __pyx_obj_5cavro_DoubleType;
struct __pyx_obj_5cavro__PlaceholderType;
struct __pyx_obj_5cavro_BytesType;
struct __pyx_obj_5cavro_StringType;
struct __pyx_obj_5cavro_UnionType;
struct __pyx_obj_5cavro__NamedType;
struct __pyx_obj_5cavro_EnumType;
struct __pyx_obj_5cavro_PromotingEnumType;
struct __pyx_obj_5cavro_RecordType;
struct __pyx_obj_5cavro_PromotingRecordType;
struct __pyx_obj_5cavro_FixedType;
struct __pyx_obj_5cavro_Schema;
struct __pyx_obj_5cavro_ResolvedSchema;
struct __pyx_obj_5cavro_ValueAdapter;
struct __pyx_obj_5cavro_LogicalType;
struct __pyx_obj_5cavro_CustomLogicalType;
struct __pyx_obj_5cavro_DecimalType;
struct __pyx_obj_5cavro_UUIDBase;
struct __pyx_obj_5cavro_UUIDStringType;
struct __pyx_obj_5cavro_UUIDFixedType;
struct __pyx_obj_5cavro_Date;
struct __pyx_obj_5cavro_TimeMillis;
struct __pyx_obj_5cavro_TimeMicros;
struct __pyx_obj_5cavro_TimestampMillis;
struct __pyx_obj_5cavro_TimestampMicros;
struct __pyx_obj_5cavro_Codec;
struct __pyx_obj_5cavro__SnappyCodec;
struct __pyx_obj_5cavro__NullCodec;
struct __pyx_obj_5cavro__DeflateCodec;
struct __pyx_obj_5cavro___Bzip2Codec;
struct __pyx_obj_5cavro__LzmaCodec;
struct __pyx_obj_5cavro__ZStandardCodec;
struct __pyx_obj_5cavro__Lz4Codec;
struct __pyx_obj_5cavro_ContainerReader;
struct __pyx_obj_5cavro_ContainerWriter;
struct __pyx_obj_5cavro_PromoteToFloat;
struct __pyx_obj_5cavro_PromoteBytesToString;
struct __pyx_obj_5cavro_PromoteStringToBytes;
struct __pyx_obj_5cavro_CannotPromote;
struct __pyx_obj_5cavro___pyx_scope_struct__genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types;
struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer;
struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types;
struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_8___init__;
struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types;
struct __pyx_obj_5cavro___pyx_scope_struct_11___init__;
struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types;
struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr;
struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types;
struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases;
struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;
struct __pyx_opt_args_7cpython_11contextvars_get_value;
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default;

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value {
  int __pyx_n;
  PyObject *default_value;
};

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default {
  int __pyx_n;
  PyObject *default_value;
};
struct __pyx_t_5cavro_random_t;
struct __pyx_opt_args_5cavro_8AvroType_clone_base;
struct __pyx_opt_args_5cavro_8AvroType_convert_value;
struct __pyx_opt_args_5cavro_8AvroType_for_writer;
struct __pyx_opt_args_5cavro_8AvroType_get_schema;
struct __pyx_opt_args_5cavro_9ArrayType_convert_value;
struct __pyx_opt_args_5cavro_7MapType_convert_value;
struct __pyx_opt_args_5cavro_9UnionType_get_schema;
struct __pyx_opt_args_5cavro_9UnionType_convert_value;
struct __pyx_opt_args_5cavro_10_NamedType_clone_base;
struct __pyx_opt_args_5cavro_6Schema_find_type;
struct __pyx_opt_args_5cavro_15ContainerWriter__flush_block;

/* "src/tests/test_zigzag.pxi":8
 *     return " ".join(parts)
 * 
 * cdef struct random_t:             # <<<<<<<<<<<<<<
 *     uint64_t state
 *     uint64_t inc
 */
struct __pyx_t_5cavro_random_t {
  uint64_t state;
  uint64_t inc;
};

/* "src/type.pxi":99
 *             self.value_adapters = value_adapters
 * 
 *     cdef AvroType clone_base(self, cls=None):             # <<<<<<<<<<<<<<
 *         if cls is None:
 *             cls = self.__class__
 */
struct __pyx_opt_args_5cavro_8AvroType_clone_base {
  int __pyx_n;
  PyObject *cls;
};

/* "src/type.pxi":134
 *             f"{type(self).__name__} does not implement _convert_value")
 * 
 *     cpdef object convert_value(self, object value, check_value=False):             # <<<<<<<<<<<<<<
 *         if check_value:
 *             self.assert_value(value)
 */
struct __pyx_opt_args_5cavro_8AvroType_convert_value {
  int __pyx_n;
  PyObject *check_value;
};

/* "src/type.pxi":236
 *             f"{type(self).__name__} does not implement _get_schema_extra")
 * 
 *     cdef AvroType for_writer(self, AvroType writer, bint _allow_deferrals=True):             # <<<<<<<<<<<<<<
 *         cdef UnionType writer_union
 *         cdef AvroType promoted
 */
struct __pyx_opt_args_5cavro_8AvroType_for_writer {
  int __pyx_n;
  int _allow_deferrals;
};

/* "src/type.pxi":292
 *         yield self
 * 
 *     cpdef get_schema(self, created=None):             # <<<<<<<<<<<<<<
 *         if created is None:
 *             created = set()
 */
struct __pyx_opt_args_5cavro_8AvroType_get_schema {
  int __pyx_n;
  PyObject *created;
};

/* "src/array.pxi":92
 *         return out
 * 
 *     cpdef object convert_value(self, object value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef int item_fitness
 *         cdef AvroType item_type = self.item_type
 */
struct __pyx_opt_args_5cavro_9ArrayType_convert_value {
  int __pyx_n;
  PyObject *check_value;
};

/* "src/map.pxi":122
 *         return out
 * 
 *     cpdef object convert_value(self, object orig_value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef int key_fitness
 *         cdef int value_fitness
 */
struct __pyx_opt_args_5cavro_7MapType_convert_value {
  int __pyx_n;
  PyObject *check_value;
};

/* "src/union.pxi":97
 *             yield from t.walk_types(visited)
 * 
 *     cpdef get_schema(self, created=None):             # <<<<<<<<<<<<<<
 *         if created is None:
 *             created = set()
 */
struct __pyx_opt_args_5cavro_9UnionType_get_schema {
  int __pyx_n;
  PyObject *created;
};

/* "src/union.pxi":170
 *         return item_type.json_decode(item_value)
 * 
 *     cpdef object convert_value(self, object value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t index = self.resolve_from_value(value)
 *         return self.union_types[index].convert_value(value)
 */
struct __pyx_opt_args_5cavro_9UnionType_convert_value {
  int __pyx_n;
  PyObject *check_value;
};

/* "src/type.pxi":409
 *         return schema
 * 
 *     cdef AvroType clone_base(self, cls=None):             # <<<<<<<<<<<<<<
 *         cdef _NamedType inst = AvroType.clone_base(self, cls)
 *         inst.name = self.name
 */
struct __pyx_opt_args_5cavro_10_NamedType_clone_base {
  int __pyx_n;
  PyObject *cls;
};

/* "src/schema.pxi":138
 *         return hasher
 * 
 *     cpdef AvroType find_type(self, str namespace, str name, bint _raise=True):             # <<<<<<<<<<<<<<
 *         """
 *         Given a namespace and name (namespace may be None), find and return the `AvroType` instance matching this name.
 */
struct __pyx_opt_args_5cavro_6Schema_find_type {
  int __pyx_n;
  int _raise;
};

/* "src/container.pxi":246
 *         self.next_item = None
 * 
 *     cdef int _flush_block(self, int force=False) except -1:             # <<<<<<<<<<<<<<
 *         cdef size_t pending_len = self.pending_block.len
 *         if force or self.num_pending > 0:
 */
struct __pyx_opt_args_5cavro_15ContainerWriter__flush_block {
  int __pyx_n;
  int force;
};

/* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_Options {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cavro_Options *__pyx_vtab;
  int fingerprint_returns_digest;
  int canonical_form_repeat_fixed;
  int canonical_form_repeat_enum;
  int record_can_encode_dict;
  int record_values_type_hint;
  int record_decodes_to_dict;
  int record_allow_extra_fields;
  int record_encode_use_defaults;
  int allow_tuple_notation;
  PyObject *union_decodes_to;
  int union_json_encodes_type_name;
  int allow_primitive_name_collision;
  int allow_primitive_names_in_namespaces;
  int named_type_names_must_be_unique;
  int enum_symbols_must_be_unique;
  int enforce_enum_symbol_name_rules;
  int enforce_type_name_rules;
  int enforce_namespace_name_rules;
  int record_fields_must_be_unique;
  int ascii_name_rules;
  int missing_values_can_be_null;
  int missing_values_can_be_empty_container;
  int allow_false_values_for_null;
  int allow_empty_unions;
  int allow_nested_unions;
  int allow_duplicate_union_types;
  int allow_union_default_any_member;
  int allow_aliases_to_be_string;
  int coerce_values_to_int;
  int coerce_values_to_float;
  int coerce_int_to_float;
  int coerce_values_to_boolean;
  int coerce_values_to_str;
  PyObject *bytes_codec;
  PyObject *fixed_codec;
  int null_pad_fixed;
  int truncate_fixed;
  int clamp_int_overflow;
  int clamp_float_overflow;
  int float_out_of_range_inf;
  int bytes_default_value_utf8;
  int string_types_default_unchanged;
  int decimal_check_exp_overflow;
  int types_str_to_schema;
  PyObject *logical_types;
  int adapt_record_types;
  int return_uuid_object;
  int allow_error_type;
  int allow_leading_dot_in_names;
  int naive_dt_assume_utc;
  int alternate_timestamp_millis_encoding;
  int date_type_accepts_string;
  int raise_on_invalid_logical;
  int inline_namespaces;
  int expand_types_in_schema;
  PyObject *unicode_errors;
  int container_fill_blocks;
  int defer_schema_promotion_errors;
  int invalid_value_includes_record_name;
  int invalid_value_include_array_index;
  int invalid_value_include_map_key;
  int allow_invalid_default_values;
  PyObject *externally_defined_types;
};


/* "src/io.pxi":4
 * 
 * 
 * cdef class _Writer:             # <<<<<<<<<<<<<<
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:
 */
struct __pyx_obj_5cavro__Writer {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cavro__Writer *__pyx_vtab;
};


/* "src/io.pxi":18
 * 
 * 
 * cdef class _Reader:             # <<<<<<<<<<<<<<
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:
 */
struct __pyx_obj_5cavro__Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cavro__Reader *__pyx_vtab;
};


/* "src/io.pxi":33
 * 
 * 
 * cdef class FileReader(_Reader):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_FileReader {
  struct __pyx_obj_5cavro__Reader __pyx_base;
  PyObject *file_obj;
};


/* "src/io.pxi":68
 * 
 * 
 * cdef class FileWriter(_Writer):             # <<<<<<<<<<<<<<
 *     """
 *     A cavro wrapper for writing data to a file-like object (Object that implements `.write(data)` and `.flush()`).
 */
struct __pyx_obj_5cavro_FileWriter {
  struct __pyx_obj_5cavro__Writer __pyx_base;
  PyObject *file_obj;
};


/* "src/buffer.pxi":7
 * @cython.no_gc_clear
 * @cython.final
 * cdef class MemoryWriter(_Writer):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_MemoryWriter {
  struct __pyx_obj_5cavro__Writer __pyx_base;
  arrayobject *buffer;
  size_t len;
};


/* "src/buffer.pxi":51
 * 
 * 
 * cdef class MemoryReader(_Reader):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_MemoryReader {
  struct __pyx_obj_5cavro__Reader __pyx_base;
  __Pyx_memviewslice data;
  uint8_t const *ptr;
  uint8_t const *end_ptr;
};


/* "src/rabin.pxi":25
 * 
 * @cython.final
 * cdef class Rabin:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_Rabin {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cavro_Rabin *__pyx_vtab;
  int64_t value;
};


/* "src/record.pxi":88
 * 
 * @cython.freelist(8)
 * cdef class Record:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_Record {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cavro_Record *__pyx_vtab;
  PyObject *data;
};


/* "src/record.pxi":188
 * 
 * @cython.final
 * cdef class RecordField:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_RecordField {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cavro_RecordField *__pyx_vtab;
  PyObject *name;
  PyObject *writer_name;
  PyObject *doc;
  struct __pyx_obj_5cavro_AvroType *type;
  PyObject *default_value;
  PyObject *order;
  PyObject *aliases;
};


/* "src/record.pxi":267
 * 
 * @cython.final
 * cdef class _FieldAccessor:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t index
 * 
 */
struct __pyx_obj_5cavro__FieldAccessor {
  PyObject_HEAD
  Py_ssize_t index;
};


/* "src/type.pxi":10
 * CANONICAL_FORM_KEYS = ('name', 'type', 'fields', 'symbols', 'items', 'values', 'size')
 * 
 * cdef class _Sentinel:             # <<<<<<<<<<<<<<
 *     cdef readonly str name
 * 
 */
struct __pyx_obj_5cavro__Sentinel {
  PyObject_HEAD
  PyObject *name;
};


/* "src/type.pxi":21
 * MISSING_VALUE = _Sentinel('MISSING_VALUE')
 * 
 * cdef class _CanonicalForm(str):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5cavro__CanonicalForm {
  PyUnicodeObject __pyx_base;
};


/* "src/type.pxi":45
 * 
 * 
 * cdef class AvroType:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_AvroType {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cavro_AvroType *__pyx_vtab;
  struct __pyx_obj_5cavro_Options *options;
  PyObject *metadata;
  PyObject *value_adapters;
};


/* "src/array.pxi":5
 * 
 * @cython.final
 * cdef class ArrayType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro array type."""
 *     type_name = "array"
 */
struct __pyx_obj_5cavro_ArrayType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
  struct __pyx_obj_5cavro_AvroType *item_type;
};


/* "src/map.pxi":2
 * 
 * cdef class MapType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro map type."""
 *     type_name = "map"
 */
struct __pyx_obj_5cavro_MapType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
  struct __pyx_obj_5cavro_StringType *key_type;
  struct __pyx_obj_5cavro_AvroType *value_type;
};


/* "src/null.pxi":2
 * 
 * cdef class NullType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro null type."""
 *     type_name = "null"
 */
struct __pyx_obj_5cavro_NullType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
};


/* "src/numeric_types.pxi":11
 * 
 * @cython.final
 * cdef class BoolType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro boolean type."""
 *     type_name = "boolean"
 */
struct __pyx_obj_5cavro_BoolType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
};


/* "src/numeric_types.pxi":61
 * 
 * @cython.final
 * cdef class IntType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro int type."""
 *     type_name = "int"
 */
struct __pyx_obj_5cavro_IntType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
};


/* "src/numeric_types.pxi":133
 * 
 * @cython.final
 * cdef class LongType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro long type."""
 *     type_name = "long"
 */
struct __pyx_obj_5cavro_LongType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
};


/* "src/numeric_types.pxi":209
 * 
 * @cython.final
 * cdef class FloatType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro float type."""
 *     type_name = "float"
 */
struct __pyx_obj_5cavro_FloatType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
};


/* "src/numeric_types.pxi":324
 * 
 * @cython.final
 * cdef class DoubleType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro double type."""
 *     type_name = "double"
 */
struct __pyx_obj_5cavro_DoubleType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
};


/* "src/record.pxi":15
 * 
 * 
 * cdef class _PlaceholderType(AvroType):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro__PlaceholderType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
  PyObject *default_value;
};


/* "src/string_types.pxi":3
 * 
 * @cython.final
 * cdef class BytesType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro bytes type"""
 *     type_name = "bytes"
 */
struct __pyx_obj_5cavro_BytesType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
};


/* "src/string_types.pxi":93
 * 
 * @cython.final
 * cdef class StringType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro string type"""
 *     type_name = "string"
 */
struct __pyx_obj_5cavro_StringType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
};


/* "src/union.pxi":25
 * 
 * @cython.final
 * cdef class UnionType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro union type"""
 *     type_name = "union"
 */
struct __pyx_obj_5cavro_UnionType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
  PyObject *union_types;
  PyObject *by_name_map;
  PyObject *return_type_tuple;
};


/* "src/type.pxi":320
 * 
 * 
 * cdef class _NamedType(AvroType):             # <<<<<<<<<<<<<<
 *     """Base class for all named types"""
 * 
 */
struct __pyx_obj_5cavro__NamedType {
  struct __pyx_obj_5cavro_AvroType __pyx_base;
  PyObject *name;
  PyObject *namespace;
  PyObject *effective_namespace;
  PyObject *aliases;
};


/* "src/enum.pxi":2
 * 
 * cdef class EnumType(_NamedType):             # <<<<<<<<<<<<<<
 *     """The avro enum type"""
 *     type_name = 'enum'
 */
struct __pyx_obj_5cavro_EnumType {
  struct __pyx_obj_5cavro__NamedType __pyx_base;
  PyObject *symbols;
  PyObject *symbol_indexes;
  PyObject *default_value;
  PyObject *doc;
};


/* "src/enum.pxi":136
 * 
 * 
 * cdef class PromotingEnumType(EnumType):             # <<<<<<<<<<<<<<
 *     cdef readonly set unknown_symbols
 *     cdef readonly EnumType reader_type
 */
struct __pyx_obj_5cavro_PromotingEnumType {
  struct __pyx_obj_5cavro_EnumType __pyx_base;
  PyObject *unknown_symbols;
  struct __pyx_obj_5cavro_EnumType *reader_type;
  struct __pyx_obj_5cavro_EnumType *writer_type;
};


/* "src/record.pxi":301
 * 
 * 
 * cdef class RecordType(_NamedType):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_RecordType {
  struct __pyx_obj_5cavro__NamedType __pyx_base;
  PyObject *doc;
  PyObject *fields;
  PyObject *field_dict;
  PyTypeObject *record;
  int _setting_up;
};


/* "src/record.pxi":610
 * 
 * 
 * cdef class PromotingRecordType(RecordType):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_PromotingRecordType {
  struct __pyx_obj_5cavro_RecordType __pyx_base;
  __Pyx_memviewslice decode_indexes;
};


/* "src/string_types.pxi":151
 * 
 * @cython.final
 * cdef class FixedType(_NamedType):             # <<<<<<<<<<<<<<
 *     """The avro fixed type"""
 *     type_name = "fixed"
 */
struct __pyx_obj_5cavro_FixedType {
  struct __pyx_obj_5cavro__NamedType __pyx_base;
  Py_ssize_t size;
};


/* "src/schema.pxi":18
 * 
 * 
 * cdef class Schema:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_Schema {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cavro_Schema *__pyx_vtab;
  PyObject *named_types;
  PyObject *source;
  struct __pyx_obj_5cavro_Options *options;
  struct __pyx_obj_5cavro_AvroType *type;
  PyObject *_canonical_form;
  PyObject *logical_types;
};


/* "src/schema.pxi":226
 * 
 * 
 * cdef class ResolvedSchema(Schema):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_ResolvedSchema {
  struct __pyx_obj_5cavro_Schema __pyx_base;
};


/* "src/logical.pxi":1
 * cdef class ValueAdapter:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_ValueAdapter {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cavro_ValueAdapter *__pyx_vtab;
};


/* "src/logical.pxi":23
 * 
 * 
 * cdef class LogicalType(ValueAdapter):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_LogicalType {
  struct __pyx_obj_5cavro_ValueAdapter __pyx_base;
};


/* "src/logical.pxi":47
 * 
 * 
 * cdef class CustomLogicalType(LogicalType):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_CustomLogicalType {
  struct __pyx_obj_5cavro_LogicalType __pyx_base;
};


/* "src/logical.pxi":72
 * 
 * 
 * cdef class DecimalType(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for decimal values."""
 * 
 */
struct __pyx_obj_5cavro_DecimalType {
  struct __pyx_obj_5cavro_LogicalType __pyx_base;
  int precision;
  int scale;
  PyObject *scale_val;
  PyObject *context;
  PyObject *size;
  PyObject *underlying_name;
  int check_exp_overflow;
};


/* "src/logical.pxi":167
 * 
 * 
 * cdef class UUIDBase(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for UUID values"""
 *     logical_name = 'uuid'
 */
struct __pyx_obj_5cavro_UUIDBase {
  struct __pyx_obj_5cavro_LogicalType __pyx_base;
  int return_uuid_object;
};


/* "src/logical.pxi":180
 * 
 * 
 * cdef class UUIDStringType(UUIDBase):             # <<<<<<<<<<<<<<
 *     underlying_types = (StringType, )
 * 
 */
struct __pyx_obj_5cavro_UUIDStringType {
  struct __pyx_obj_5cavro_UUIDBase __pyx_base;
};


/* "src/logical.pxi":209
 * 
 * 
 * cdef class UUIDFixedType(UUIDBase):             # <<<<<<<<<<<<<<
 *     underlying_types = (FixedType, )
 * 
 */
struct __pyx_obj_5cavro_UUIDFixedType {
  struct __pyx_obj_5cavro_UUIDBase __pyx_base;
};


/* "src/logical.pxi":237
 * EPOCH_DT_NO_TZ = datetime.datetime(1970, 1, 1)
 * 
 * cdef class Date(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for Date values"""
 *     logical_name = 'date'
 */
struct __pyx_obj_5cavro_Date {
  struct __pyx_obj_5cavro_LogicalType __pyx_base;
  int accepts_string;
};


/* "src/logical.pxi":292
 * 
 * 
 * cdef class TimeMillis(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for time-millis values"""
 *     logical_name = 'time-millis'
 */
struct __pyx_obj_5cavro_TimeMillis {
  struct __pyx_obj_5cavro_LogicalType __pyx_base;
};


/* "src/logical.pxi":316
 * 
 * 
 * cdef class TimeMicros(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for time-micros values"""
 *     logical_name = 'time-micros'
 */
struct __pyx_obj_5cavro_TimeMicros {
  struct __pyx_obj_5cavro_LogicalType __pyx_base;
};


/* "src/logical.pxi":340
 * 
 * 
 * cdef class TimestampMillis(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for timestamp-micros values"""
 *     logical_name = 'timestamp-millis'
 */
struct __pyx_obj_5cavro_TimestampMillis {
  struct __pyx_obj_5cavro_LogicalType __pyx_base;
  int alternate_timestamp_encoding;
};


/* "src/logical.pxi":375
 * 
 * 
 * cdef class TimestampMicros(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for timestamp-micros values"""
 *     logical_name = 'timestamp-micros'
 */
struct __pyx_obj_5cavro_TimestampMicros {
  struct __pyx_obj_5cavro_LogicalType __pyx_base;
};


/* "src/codec.pxi":50
 * 
 * 
 * cdef class Codec:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_Codec {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cavro_Codec *__pyx_vtab;
};


/* "src/codec.pxi":70
 * 
 * @cython.final
 * cdef class _SnappyCodec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implemented the snappy codec if python-snappy is installed.
 */
struct __pyx_obj_5cavro__SnappyCodec {
  struct __pyx_obj_5cavro_Codec __pyx_base;
};


/* "src/codec.pxi":93
 * 
 * 
 * cdef class _NullCodec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implements the null codec (no compression)
 */
struct __pyx_obj_5cavro__NullCodec {
  struct __pyx_obj_5cavro_Codec __pyx_base;
};


/* "src/codec.pxi":107
 * 
 * 
 * cdef class _DeflateCodec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implements the deflate codec
 */
struct __pyx_obj_5cavro__DeflateCodec {
  struct __pyx_obj_5cavro_Codec __pyx_base;
};


/* "src/codec.pxi":122
 * 
 * 
 * cdef class __Bzip2Codec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implements the bzip2 codec
 */
struct __pyx_obj_5cavro___Bzip2Codec {
  struct __pyx_obj_5cavro_Codec __pyx_base;
};


/* "src/codec.pxi":137
 * 
 * 
 * cdef class _LzmaCodec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implements the lzma codec
 */
struct __pyx_obj_5cavro__LzmaCodec {
  struct __pyx_obj_5cavro_Codec __pyx_base;
};


/* "src/codec.pxi":152
 * 
 * 
 * cdef class _ZStandardCodec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implements the zstandard codec if zstandard is installed.
 */
struct __pyx_obj_5cavro__ZStandardCodec {
  struct __pyx_obj_5cavro_Codec __pyx_base;
  PyObject *compressor;
  PyObject *decompressor;
};


/* "src/codec.pxi":175
 * 
 * 
 * cdef class _Lz4Codec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implements the lz4 codec if lz4 is installed.
 */
struct __pyx_obj_5cavro__Lz4Codec {
  struct __pyx_obj_5cavro_Codec __pyx_base;
};


/* "src/container.pxi":41
 * 
 * 
 * cdef class ContainerReader:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_ContainerReader {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cavro_ContainerReader *__pyx_vtab;
  PyObject *metadata;
  __Pyx_memviewslice marker;
  struct __pyx_obj_5cavro_Schema *writer_schema;
  struct __pyx_obj_5cavro_Schema *reader_schema;
  struct __pyx_obj_5cavro_Schema *schema;
  PyObject *codec_name;
  size_t objects_left_in_block;
  struct __pyx_obj_5cavro_MemoryReader *current_block;
  struct __pyx_obj_5cavro_Codec *codec;
  struct __pyx_obj_5cavro__Reader *reader;
};


/* "src/container.pxi":131
 * 
 * @cython.no_gc_clear
 * cdef class ContainerWriter:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_ContainerWriter {
  PyObject_HEAD
  struct __pyx_vtabstruct_5cavro_ContainerWriter *__pyx_vtab;
  struct __pyx_obj_5cavro__Writer *writer;
  struct __pyx_obj_5cavro_MemoryWriter *pending_block;
  struct __pyx_obj_5cavro_MemoryWriter *next_block;
  struct __pyx_obj_5cavro_MemoryWriter *next_item;
  struct __pyx_obj_5cavro_Schema *schema;
  struct __pyx_obj_5cavro_Codec *codec;
  __Pyx_memviewslice marker;
  size_t max_blocksize;
  struct __pyx_obj_5cavro_Options *options;
  int should_write_header;
  size_t num_pending;
  int blocks_written;
  PyObject *metadata;
};


/* "src/promotions.pxi":3
 * 
 * 
 * cdef class PromoteToFloat(ValueAdapter):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_PromoteToFloat {
  struct __pyx_obj_5cavro_ValueAdapter __pyx_base;
};


/* "src/promotions.pxi":16
 * 
 * 
 * cdef class PromoteBytesToString(ValueAdapter):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_PromoteBytesToString {
  struct __pyx_obj_5cavro_ValueAdapter __pyx_base;
};


/* "src/promotions.pxi":29
 * 
 * 
 * cdef class PromoteStringToBytes(ValueAdapter):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_PromoteStringToBytes {
  struct __pyx_obj_5cavro_ValueAdapter __pyx_base;
};


/* "src/promotions.pxi":42
 * 
 * 
 * cdef class CannotPromote(ValueAdapter):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
struct __pyx_obj_5cavro_CannotPromote {
  struct __pyx_obj_5cavro_ValueAdapter __pyx_base;
  struct __pyx_obj_5cavro_AvroType *reader_type;
  struct __pyx_obj_5cavro_AvroType *writer_type;
  PyObject *extra;
};


/* "cavro.pyx":52
 *         name = test.__name__
 *         if args:
 *             name += '-' + ('-'.join(str(a) for a in args))             # <<<<<<<<<<<<<<
 *         __tests[name] = (test, ) + args
 *     fn(add)
 */
struct __pyx_obj_5cavro___pyx_scope_struct__genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_a;
};


/* "cavro.pyx":62
 *         name = test.__name__
 *         if args:
 *             name += '-' + ('-'.join(str(a) for a in args))             # <<<<<<<<<<<<<<
 *         __perf_tests[name] = (test, ) + args
 *     fn(add)
 */
struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_a;
};


/* "src/error.pxi":90
 *     def __str__(self):
 *         if self.schema_path:
 *             path_str = ".".join(str(p) for p in self.schema_path)             # <<<<<<<<<<<<<<
 *             return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name} at {path_str}"
 *         return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name}"
 */
struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_p;
};


/* "src/array.pxi":23
 *         return _strip_keys(dict(source), {'type', 'items'})
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types {
  PyObject_HEAD
  struct __pyx_obj_5cavro_ArrayType *__pyx_v_self;
  PyObject *__pyx_v_visited;
};


/* "src/enum.pxi":110
 *         })
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         if not isinstance(writer, EnumType):
 *             return
 */
struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer {
  PyObject_HEAD
  PyObject *__pyx_v_reader_symbols;
  struct __pyx_obj_5cavro_EnumType *__pyx_v_self;
};


/* "src/enum.pxi":125
 *         new_type.reader_type = self
 *         new_type.writer_type = writer_enum
 *         new_type.symbols = tuple(s if s in reader_symbols else self.default_value for s in writer.symbols)             # <<<<<<<<<<<<<<
 *         new_type.unknown_symbols = writer_extra_symbols
 *         new_type.symbol_indexes = writer_enum.symbol_indexes
 */
struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_s;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "src/map.pxi":23
 *         return _strip_keys(dict(source), {'type', 'values'})
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types {
  PyObject_HEAD
  struct __pyx_obj_5cavro_MapType *__pyx_v_self;
  PyObject *__pyx_v_visited;
};


/* "src/record.pxi":142
 *             return "{...}"
 *         more = False
 *         key_len = sum(len(f.name) + 1 for f in self.Type.fields)             # <<<<<<<<<<<<<<
 *         child_remain = remain - key_len
 *         vals = []
 */
struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_f;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "src/record.pxi":321
 *     cdef bint _setting_up
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         cdef Schema schema_ = schema
 *         self._setting_up = True
 */
struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ {
  PyObject_HEAD
  struct __pyx_obj_5cavro_Schema *__pyx_v_schema_;
  struct __pyx_obj_5cavro_RecordType *__pyx_v_self;
};


/* "src/record.pxi":327
 *         _NamedType.__init__(self, schema, source, namespace)
 *         self.fields = tuple(
 *             RecordField(schema_, f, self.effective_namespace) for f in source['fields']             # <<<<<<<<<<<<<<
 *         )
 *         self.field_dict = {}
 */
struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_f;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "src/record.pxi":366
 *         })
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types {
  PyObject_HEAD
  PyObject *__pyx_v_field;
  struct __pyx_obj_5cavro_RecordType *__pyx_v_self;
  PyObject *__pyx_v_visited;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "src/union.pxi":33
 *     cdef readonly tuple return_type_tuple
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         super().__init__(schema, source, namespace)
 *         self.union_types = tuple(AvroType.for_source(schema, s, namespace) for s in source)
 */
struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_is_ambiguous;
  int __pyx_v_is_amibiguous;
  PyObject *__pyx_v_namespace;
  PyObject *__pyx_v_schema;
};


/* "src/union.pxi":35
 *     def __init__(self, schema, source, namespace):
 *         super().__init__(schema, source, namespace)
 *         self.union_types = tuple(AvroType.for_source(schema, s, namespace) for s in source)             # <<<<<<<<<<<<<<
 *         if len(self.union_types) == 0 and not self.options.allow_empty_unions:
 *             raise ValueError("Unions must contain at least one member type")
 */
struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_s;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "src/union.pxi":44
 *         decode_option = self.options.union_decodes_to
 *         if decode_option == UnionDecodeOption.RAW_VALUES:
 *             self.return_type_tuple = tuple(False for _ in self.union_types)             # <<<<<<<<<<<<<<
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_AMBIGUOUS:
 *             is_amibiguous = are_union_values_ambiguous(self.union_types)
 */
struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v__;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "src/union.pxi":47
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_AMBIGUOUS:
 *             is_amibiguous = are_union_values_ambiguous(self.union_types)
 *             self.return_type_tuple = tuple(is_amibiguous for _ in self.union_types)             # <<<<<<<<<<<<<<
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD_AMBIGUOUS:
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1
 */
struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v__;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "src/union.pxi":49
 *             self.return_type_tuple = tuple(is_amibiguous for _ in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD_AMBIGUOUS:
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1             # <<<<<<<<<<<<<<
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:
 */
struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_t;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "src/union.pxi":50
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD_AMBIGUOUS:
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)             # <<<<<<<<<<<<<<
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)
 */
struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_ty;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "src/union.pxi":52
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)             # <<<<<<<<<<<<<<
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_ALWAYS:
 *             self.return_type_tuple = tuple(True for _ in self.union_types)
 */
struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_ty;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "src/union.pxi":54
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_ALWAYS:
 *             self.return_type_tuple = tuple(True for _ in self.union_types)             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError(f"Invalid union_decodes_to option: {decode_option}")
 */
struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v__;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "src/union.pxi":90
 *         return dict()
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types {
  PyObject_HEAD
  struct __pyx_obj_5cavro_UnionType *__pyx_v_self;
  PyObject *__pyx_v_t;
  PyObject *__pyx_v_visited;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "src/type.pxi":30
 *             if key in data:
 *                 pairs.append((f'"{key}"', dict_to_canonical(data[key])))
 *         pair_str = ','.join(':'.join(p) for p in pairs)             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('{' + pair_str + '}')
 *     elif isinstance(data, (tuple, list)):
 */
struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_p;
};


/* "src/type.pxi":286
 *             raise TypeError(f"Default value {schema_default!r} is not valid for field: {field}") from e
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types {
  PyObject_HEAD
  struct __pyx_obj_5cavro_AvroType *__pyx_v_self;
  PyObject *__pyx_v_visited;
};


/* "src/type.pxi":390
 *         return False
 * 
 *     cdef frozenset get_namespaced_aliases(self):             # <<<<<<<<<<<<<<
 *         cdef frozenset all_names = self.aliases | {self.name}
 *         if not self.effective_namespace:
 */
struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases {
  PyObject_HEAD
  struct __pyx_obj_5cavro__NamedType *__pyx_v_self;
};


/* "src/type.pxi":394
 *         if not self.effective_namespace:
 *             return all_names
 *         return frozenset(resolve_namespaced_name(self.effective_namespace, alias) for alias in all_names)             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_alias;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};


/* "View.MemoryView":114
 * @cython.collection_type("sequence")
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_array_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_array *__pyx_vtab;
  char *data;
  Py_ssize_t len;
  char *format;
  int ndim;
  Py_ssize_t *_shape;
  Py_ssize_t *_strides;
  Py_ssize_t itemsize;
  PyObject *mode;
  PyObject *_format;
  void (*callback_free_data)(void *);
  int free_data;
  int dtype_is_object;
};


/* "View.MemoryView":302
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */
struct __pyx_MemviewEnum_obj {
  PyObject_HEAD
  PyObject *name;
};


/* "View.MemoryView":337
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview:             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */
struct __pyx_memoryview_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;
  PyObject *obj;
  PyObject *_size;
  PyObject *_array_interface;
  PyThread_type_lock lock;
  __pyx_atomic_int_type acquisition_count;
  Py_buffer view;
  int flags;
  int dtype_is_object;
  __Pyx_TypeInfo *typeinfo;
};


/* "View.MemoryView":952
 * @cython.collection_type("sequence")
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */
struct __pyx_memoryviewslice_obj {
  struct __pyx_memoryview_obj __pyx_base;
  __Pyx_memviewslice from_slice;
  PyObject *from_object;
  PyObject *(*to_object_func)(char *);
  int (*to_dtype_func)(char *, PyObject *);
};



/* "View.MemoryView":114
 * @cython.collection_type("sequence")
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_array {
  PyObject *(*get_memview)(struct __pyx_array_obj *);
};
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;


/* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_Options {
  int (*can_have_missing_values)(struct __pyx_obj_5cavro_Options *);
};
static struct __pyx_vtabstruct_5cavro_Options *__pyx_vtabptr_5cavro_Options;


/* "src/io.pxi":4
 * 
 * 
 * cdef class _Writer:             # <<<<<<<<<<<<<<
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:
 */

struct __pyx_vtabstruct_5cavro__Writer {
  int (*write_u8)(struct __pyx_obj_5cavro__Writer *, uint8_t);
  int (*write_n)(struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice);
  int (*flush)(struct __pyx_obj_5cavro__Writer *);
};
static struct __pyx_vtabstruct_5cavro__Writer *__pyx_vtabptr_5cavro__Writer;


/* "src/io.pxi":18
 * 
 * 
 * cdef class _Reader:             # <<<<<<<<<<<<<<
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:
 */

struct __pyx_vtabstruct_5cavro__Reader {
  uint8_t (*read_u8)(struct __pyx_obj_5cavro__Reader *);
  PyObject *(*read_bytes)(struct __pyx_obj_5cavro__Reader *, Py_ssize_t);
  __Pyx_memviewslice (*read_n)(struct __pyx_obj_5cavro__Reader *, Py_ssize_t);
};
static struct __pyx_vtabstruct_5cavro__Reader *__pyx_vtabptr_5cavro__Reader;


/* "src/io.pxi":33
 * 
 * 
 * cdef class FileReader(_Reader):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_FileReader {
  struct __pyx_vtabstruct_5cavro__Reader __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_FileReader *__pyx_vtabptr_5cavro_FileReader;


/* "src/io.pxi":68
 * 
 * 
 * cdef class FileWriter(_Writer):             # <<<<<<<<<<<<<<
 *     """
 *     A cavro wrapper for writing data to a file-like object (Object that implements `.write(data)` and `.flush()`).
 */

struct __pyx_vtabstruct_5cavro_FileWriter {
  struct __pyx_vtabstruct_5cavro__Writer __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_FileWriter *__pyx_vtabptr_5cavro_FileWriter;


/* "src/buffer.pxi":7
 * @cython.no_gc_clear
 * @cython.final
 * cdef class MemoryWriter(_Writer):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_MemoryWriter {
  struct __pyx_vtabstruct_5cavro__Writer __pyx_base;
  PyObject *(*bytes)(struct __pyx_obj_5cavro_MemoryWriter *);
  __Pyx_memviewslice (*view)(struct __pyx_obj_5cavro_MemoryWriter *);
  PyObject *(*reset)(struct __pyx_obj_5cavro_MemoryWriter *);
};
static struct __pyx_vtabstruct_5cavro_MemoryWriter *__pyx_vtabptr_5cavro_MemoryWriter;
static PyObject *__pyx_f_5cavro_12MemoryWriter_bytes(struct __pyx_obj_5cavro_MemoryWriter *);
static __Pyx_memviewslice __pyx_f_5cavro_12MemoryWriter_view(struct __pyx_obj_5cavro_MemoryWriter *);
static PyObject *__pyx_f_5cavro_12MemoryWriter_reset(struct __pyx_obj_5cavro_MemoryWriter *);


/* "src/buffer.pxi":51
 * 
 * 
 * cdef class MemoryReader(_Reader):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_MemoryReader {
  struct __pyx_vtabstruct_5cavro__Reader __pyx_base;
  void (*_reset_to)(struct __pyx_obj_5cavro_MemoryReader *, __Pyx_memviewslice);
  uint8_t const *(*advance)(struct __pyx_obj_5cavro_MemoryReader *, size_t);
  int (*ensure)(struct __pyx_obj_5cavro_MemoryReader *, size_t);
};
static struct __pyx_vtabstruct_5cavro_MemoryReader *__pyx_vtabptr_5cavro_MemoryReader;
static CYTHON_INLINE int __pyx_f_5cavro_12MemoryReader_ensure(struct __pyx_obj_5cavro_MemoryReader *, size_t);


/* "src/rabin.pxi":25
 * 
 * @cython.final
 * cdef class Rabin:             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_Rabin {
  PyObject *(*update)(struct __pyx_obj_5cavro_Rabin *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5cavro_Rabin *__pyx_vtabptr_5cavro_Rabin;
static PyObject *__pyx_f_5cavro_5Rabin_update(struct __pyx_obj_5cavro_Rabin *, PyObject *, int __pyx_skip_dispatch);


/* "src/record.pxi":88
 * 
 * @cython.freelist(8)
 * cdef class Record:             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_Record {
  PyObject *(*_repr_children)(struct __pyx_obj_5cavro_Record *, PyObject *);
};
static struct __pyx_vtabstruct_5cavro_Record *__pyx_vtabptr_5cavro_Record;


/* "src/record.pxi":188
 * 
 * @cython.final
 * cdef class RecordField:             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_RecordField {
  PyObject *(*for_reader)(struct __pyx_obj_5cavro_RecordField *, PyObject *, struct __pyx_obj_5cavro_AvroType *);
  PyObject *(*placeholder)(struct __pyx_obj_5cavro_RecordField *);
  struct __pyx_obj_5cavro__CanonicalForm *(*canonical_form)(struct __pyx_obj_5cavro_RecordField *, PyObject *);
};
static struct __pyx_vtabstruct_5cavro_RecordField *__pyx_vtabptr_5cavro_RecordField;
static PyObject *__pyx_f_5cavro_11RecordField_for_reader(struct __pyx_obj_5cavro_RecordField *, PyObject *, struct __pyx_obj_5cavro_AvroType *);
static PyObject *__pyx_f_5cavro_11RecordField_placeholder(struct __pyx_obj_5cavro_RecordField *);
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_11RecordField_canonical_form(struct __pyx_obj_5cavro_RecordField *, PyObject *);


/* "src/type.pxi":45
 * 
 * 
 * cdef class AvroType:             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_AvroType {
  struct __pyx_obj_5cavro_AvroType *(*clone_base)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_opt_args_5cavro_8AvroType_clone_base *__pyx_optional_args);
  struct __pyx_obj_5cavro_AvroType *(*copy)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch);
  PyObject *(*_make_logical)(struct __pyx_obj_5cavro_AvroType *, PyObject *, PyObject *);
  PyObject *(*_extract_metadata)(struct __pyx_obj_5cavro_AvroType *, PyObject *);
  PyObject *(*_convert_value)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*convert_value)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_8AvroType_convert_value *__pyx_optional_args);
  int (*binary_buffer_encode)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *);
  int (*_binary_buffer_encode)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *);
  PyObject *(*binary_buffer_decode)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *);
  PyObject *(*_binary_buffer_decode)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *);
  int (*get_value_fitness)(struct __pyx_obj_5cavro_AvroType *, PyObject *);
  int (*_get_value_fitness)(struct __pyx_obj_5cavro_AvroType *, PyObject *);
  int (*assert_value)(struct __pyx_obj_5cavro_AvroType *, PyObject *);
  PyObject *(*json_format)(struct __pyx_obj_5cavro_AvroType *, PyObject *);
  PyObject *(*_json_format)(struct __pyx_obj_5cavro_AvroType *, PyObject *);
  PyObject *(*json_decode)(struct __pyx_obj_5cavro_AvroType *, PyObject *);
  PyObject *(*_json_decode)(struct __pyx_obj_5cavro_AvroType *, PyObject *);
  struct __pyx_obj_5cavro__CanonicalForm *(*canonical_form)(struct __pyx_obj_5cavro_AvroType *, PyObject *);
  PyObject *(*_get_schema_extra)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_5cavro_AvroType *(*for_writer)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *, struct __pyx_opt_args_5cavro_8AvroType_for_writer *__pyx_optional_args);
  struct __pyx_obj_5cavro_AvroType *(*_for_writer)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *);
  int (*accepts_missing_value)(struct __pyx_obj_5cavro_AvroType *);
  PyObject *(*resolve_default_value)(struct __pyx_obj_5cavro_AvroType *, PyObject *, PyObject *);
  PyObject *(*get_schema)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_8AvroType_get_schema *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5cavro_AvroType *__pyx_vtabptr_5cavro_AvroType;


/* "src/union.pxi":25
 * 
 * @cython.final
 * cdef class UnionType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro union type"""
 *     type_name = "union"
 */

struct __pyx_vtabstruct_5cavro_UnionType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
  PyObject *(*_populate_name_map)(struct __pyx_obj_5cavro_UnionType *);
  Py_ssize_t (*resolve_from_value)(struct __pyx_obj_5cavro_UnionType *, PyObject *);
  struct __pyx_obj_5cavro_AvroType *(*for_reader)(struct __pyx_obj_5cavro_UnionType *, struct __pyx_obj_5cavro_AvroType *);
};
static struct __pyx_vtabstruct_5cavro_UnionType *__pyx_vtabptr_5cavro_UnionType;
static PyObject *__pyx_f_5cavro_9UnionType__populate_name_map(struct __pyx_obj_5cavro_UnionType *);
static Py_ssize_t __pyx_f_5cavro_9UnionType_resolve_from_value(struct __pyx_obj_5cavro_UnionType *, PyObject *);
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9UnionType_for_reader(struct __pyx_obj_5cavro_UnionType *, struct __pyx_obj_5cavro_AvroType *);


/* "src/string_types.pxi":93
 * 
 * @cython.final
 * cdef class StringType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro string type"""
 *     type_name = "string"
 */

struct __pyx_vtabstruct_5cavro_StringType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_StringType *__pyx_vtabptr_5cavro_StringType;


/* "src/string_types.pxi":3
 * 
 * @cython.final
 * cdef class BytesType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro bytes type"""
 *     type_name = "bytes"
 */

struct __pyx_vtabstruct_5cavro_BytesType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_BytesType *__pyx_vtabptr_5cavro_BytesType;


/* "src/record.pxi":15
 * 
 * 
 * cdef class _PlaceholderType(AvroType):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro__PlaceholderType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro__PlaceholderType *__pyx_vtabptr_5cavro__PlaceholderType;


/* "src/numeric_types.pxi":324
 * 
 * @cython.final
 * cdef class DoubleType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro double type."""
 *     type_name = "double"
 */

struct __pyx_vtabstruct_5cavro_DoubleType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_DoubleType *__pyx_vtabptr_5cavro_DoubleType;


/* "src/numeric_types.pxi":209
 * 
 * @cython.final
 * cdef class FloatType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro float type."""
 *     type_name = "float"
 */

struct __pyx_vtabstruct_5cavro_FloatType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_FloatType *__pyx_vtabptr_5cavro_FloatType;


/* "src/numeric_types.pxi":133
 * 
 * @cython.final
 * cdef class LongType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro long type."""
 *     type_name = "long"
 */

struct __pyx_vtabstruct_5cavro_LongType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_LongType *__pyx_vtabptr_5cavro_LongType;


/* "src/numeric_types.pxi":61
 * 
 * @cython.final
 * cdef class IntType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro int type."""
 *     type_name = "int"
 */

struct __pyx_vtabstruct_5cavro_IntType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_IntType *__pyx_vtabptr_5cavro_IntType;


/* "src/numeric_types.pxi":11
 * 
 * @cython.final
 * cdef class BoolType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro boolean type."""
 *     type_name = "boolean"
 */

struct __pyx_vtabstruct_5cavro_BoolType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_BoolType *__pyx_vtabptr_5cavro_BoolType;


/* "src/null.pxi":2
 * 
 * cdef class NullType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro null type."""
 *     type_name = "null"
 */

struct __pyx_vtabstruct_5cavro_NullType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_NullType *__pyx_vtabptr_5cavro_NullType;


/* "src/map.pxi":2
 * 
 * cdef class MapType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro map type."""
 *     type_name = "map"
 */

struct __pyx_vtabstruct_5cavro_MapType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
  PyObject *(*_make_converted_map)(struct __pyx_obj_5cavro_MapType *, PyObject *);
};
static struct __pyx_vtabstruct_5cavro_MapType *__pyx_vtabptr_5cavro_MapType;


/* "src/array.pxi":5
 * 
 * @cython.final
 * cdef class ArrayType(AvroType):             # <<<<<<<<<<<<<<
 *     """The avro array type."""
 *     type_name = "array"
 */

struct __pyx_vtabstruct_5cavro_ArrayType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
  PyObject *(*_make_converted_list)(struct __pyx_obj_5cavro_ArrayType *, PyObject *);
};
static struct __pyx_vtabstruct_5cavro_ArrayType *__pyx_vtabptr_5cavro_ArrayType;
static PyObject *__pyx_f_5cavro_9ArrayType__make_converted_list(struct __pyx_obj_5cavro_ArrayType *, PyObject *);


/* "src/type.pxi":320
 * 
 * 
 * cdef class _NamedType(AvroType):             # <<<<<<<<<<<<<<
 *     """Base class for all named types"""
 * 
 */

struct __pyx_vtabstruct_5cavro__NamedType {
  struct __pyx_vtabstruct_5cavro_AvroType __pyx_base;
  int (*name_matches)(struct __pyx_obj_5cavro__NamedType *, struct __pyx_obj_5cavro__NamedType *);
  PyObject *(*get_namespaced_aliases)(struct __pyx_obj_5cavro__NamedType *);
};
static struct __pyx_vtabstruct_5cavro__NamedType *__pyx_vtabptr_5cavro__NamedType;


/* "src/string_types.pxi":151
 * 
 * @cython.final
 * cdef class FixedType(_NamedType):             # <<<<<<<<<<<<<<
 *     """The avro fixed type"""
 *     type_name = "fixed"
 */

struct __pyx_vtabstruct_5cavro_FixedType {
  struct __pyx_vtabstruct_5cavro__NamedType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_FixedType *__pyx_vtabptr_5cavro_FixedType;


/* "src/record.pxi":301
 * 
 * 
 * cdef class RecordType(_NamedType):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_RecordType {
  struct __pyx_vtabstruct_5cavro__NamedType __pyx_base;
  PyObject *(*_binary_buffer_decode_record)(struct __pyx_obj_5cavro_RecordType *, struct __pyx_obj_5cavro__Reader *);
  PyObject *(*_binary_buffer_decode_dict)(struct __pyx_obj_5cavro_RecordType *, struct __pyx_obj_5cavro__Reader *);
  PyObject *(*json_decode_record)(struct __pyx_obj_5cavro_RecordType *, PyObject *);
  PyObject *(*json_decode_dict)(struct __pyx_obj_5cavro_RecordType *, PyObject *);
};
static struct __pyx_vtabstruct_5cavro_RecordType *__pyx_vtabptr_5cavro_RecordType;


/* "src/record.pxi":610
 * 
 * 
 * cdef class PromotingRecordType(RecordType):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_PromotingRecordType {
  struct __pyx_vtabstruct_5cavro_RecordType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_PromotingRecordType *__pyx_vtabptr_5cavro_PromotingRecordType;


/* "src/enum.pxi":2
 * 
 * cdef class EnumType(_NamedType):             # <<<<<<<<<<<<<<
 *     """The avro enum type"""
 *     type_name = 'enum'
 */

struct __pyx_vtabstruct_5cavro_EnumType {
  struct __pyx_vtabstruct_5cavro__NamedType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_EnumType *__pyx_vtabptr_5cavro_EnumType;


/* "src/enum.pxi":136
 * 
 * 
 * cdef class PromotingEnumType(EnumType):             # <<<<<<<<<<<<<<
 *     cdef readonly set unknown_symbols
 *     cdef readonly EnumType reader_type
 */

struct __pyx_vtabstruct_5cavro_PromotingEnumType {
  struct __pyx_vtabstruct_5cavro_EnumType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_PromotingEnumType *__pyx_vtabptr_5cavro_PromotingEnumType;


/* "src/schema.pxi":18
 * 
 * 
 * cdef class Schema:             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_Schema {
  void (*register_type)(struct __pyx_obj_5cavro_Schema *, PyObject *, PyObject *, struct __pyx_obj_5cavro_AvroType *);
  struct __pyx_obj_5cavro_AvroType *(*find_type)(struct __pyx_obj_5cavro_Schema *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_6Schema_find_type *__pyx_optional_args);
  PyObject *(*binary_read)(struct __pyx_obj_5cavro_Schema *, struct __pyx_obj_5cavro__Reader *, int __pyx_skip_dispatch);
  PyObject *(*binary_write)(struct __pyx_obj_5cavro_Schema *, struct __pyx_obj_5cavro__Writer *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_5cavro_Schema *(*reader_for_writer)(struct __pyx_obj_5cavro_Schema *, struct __pyx_obj_5cavro_Schema *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5cavro_Schema *__pyx_vtabptr_5cavro_Schema;


/* "src/schema.pxi":226
 * 
 * 
 * cdef class ResolvedSchema(Schema):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_ResolvedSchema {
  struct __pyx_vtabstruct_5cavro_Schema __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_ResolvedSchema *__pyx_vtabptr_5cavro_ResolvedSchema;


/* "src/logical.pxi":1
 * cdef class ValueAdapter:             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_ValueAdapter {
  PyObject *(*encode_value)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *);
  PyObject *(*decode_value)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *);
};
static struct __pyx_vtabstruct_5cavro_ValueAdapter *__pyx_vtabptr_5cavro_ValueAdapter;


/* "src/logical.pxi":23
 * 
 * 
 * cdef class LogicalType(ValueAdapter):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_LogicalType {
  struct __pyx_vtabstruct_5cavro_ValueAdapter __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_LogicalType *__pyx_vtabptr_5cavro_LogicalType;


/* "src/logical.pxi":47
 * 
 * 
 * cdef class CustomLogicalType(LogicalType):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_CustomLogicalType {
  struct __pyx_vtabstruct_5cavro_LogicalType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_CustomLogicalType *__pyx_vtabptr_5cavro_CustomLogicalType;


/* "src/logical.pxi":72
 * 
 * 
 * cdef class DecimalType(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for decimal values."""
 * 
 */

struct __pyx_vtabstruct_5cavro_DecimalType {
  struct __pyx_vtabstruct_5cavro_LogicalType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_DecimalType *__pyx_vtabptr_5cavro_DecimalType;


/* "src/logical.pxi":167
 * 
 * 
 * cdef class UUIDBase(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for UUID values"""
 *     logical_name = 'uuid'
 */

struct __pyx_vtabstruct_5cavro_UUIDBase {
  struct __pyx_vtabstruct_5cavro_LogicalType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_UUIDBase *__pyx_vtabptr_5cavro_UUIDBase;


/* "src/logical.pxi":180
 * 
 * 
 * cdef class UUIDStringType(UUIDBase):             # <<<<<<<<<<<<<<
 *     underlying_types = (StringType, )
 * 
 */

struct __pyx_vtabstruct_5cavro_UUIDStringType {
  struct __pyx_vtabstruct_5cavro_UUIDBase __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_UUIDStringType *__pyx_vtabptr_5cavro_UUIDStringType;


/* "src/logical.pxi":209
 * 
 * 
 * cdef class UUIDFixedType(UUIDBase):             # <<<<<<<<<<<<<<
 *     underlying_types = (FixedType, )
 * 
 */

struct __pyx_vtabstruct_5cavro_UUIDFixedType {
  struct __pyx_vtabstruct_5cavro_UUIDBase __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_UUIDFixedType *__pyx_vtabptr_5cavro_UUIDFixedType;


/* "src/logical.pxi":237
 * EPOCH_DT_NO_TZ = datetime.datetime(1970, 1, 1)
 * 
 * cdef class Date(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for Date values"""
 *     logical_name = 'date'
 */

struct __pyx_vtabstruct_5cavro_Date {
  struct __pyx_vtabstruct_5cavro_LogicalType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_Date *__pyx_vtabptr_5cavro_Date;


/* "src/logical.pxi":292
 * 
 * 
 * cdef class TimeMillis(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for time-millis values"""
 *     logical_name = 'time-millis'
 */

struct __pyx_vtabstruct_5cavro_TimeMillis {
  struct __pyx_vtabstruct_5cavro_LogicalType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_TimeMillis *__pyx_vtabptr_5cavro_TimeMillis;


/* "src/logical.pxi":316
 * 
 * 
 * cdef class TimeMicros(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for time-micros values"""
 *     logical_name = 'time-micros'
 */

struct __pyx_vtabstruct_5cavro_TimeMicros {
  struct __pyx_vtabstruct_5cavro_LogicalType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_TimeMicros *__pyx_vtabptr_5cavro_TimeMicros;


/* "src/logical.pxi":340
 * 
 * 
 * cdef class TimestampMillis(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for timestamp-micros values"""
 *     logical_name = 'timestamp-millis'
 */

struct __pyx_vtabstruct_5cavro_TimestampMillis {
  struct __pyx_vtabstruct_5cavro_LogicalType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_TimestampMillis *__pyx_vtabptr_5cavro_TimestampMillis;


/* "src/logical.pxi":375
 * 
 * 
 * cdef class TimestampMicros(LogicalType):             # <<<<<<<<<<<<<<
 *     """Logical type for timestamp-micros values"""
 *     logical_name = 'timestamp-micros'
 */

struct __pyx_vtabstruct_5cavro_TimestampMicros {
  struct __pyx_vtabstruct_5cavro_LogicalType __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_TimestampMicros *__pyx_vtabptr_5cavro_TimestampMicros;


/* "src/codec.pxi":50
 * 
 * 
 * cdef class Codec:             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_Codec {
  __Pyx_memviewslice (*read_block)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Reader *, size_t);
  Py_ssize_t (*write_block)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice);
};
static struct __pyx_vtabstruct_5cavro_Codec *__pyx_vtabptr_5cavro_Codec;


/* "src/codec.pxi":70
 * 
 * @cython.final
 * cdef class _SnappyCodec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implemented the snappy codec if python-snappy is installed.
 */

struct __pyx_vtabstruct_5cavro__SnappyCodec {
  struct __pyx_vtabstruct_5cavro_Codec __pyx_base;
};
static struct __pyx_vtabstruct_5cavro__SnappyCodec *__pyx_vtabptr_5cavro__SnappyCodec;


/* "src/codec.pxi":93
 * 
 * 
 * cdef class _NullCodec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implements the null codec (no compression)
 */

struct __pyx_vtabstruct_5cavro__NullCodec {
  struct __pyx_vtabstruct_5cavro_Codec __pyx_base;
};
static struct __pyx_vtabstruct_5cavro__NullCodec *__pyx_vtabptr_5cavro__NullCodec;


/* "src/codec.pxi":107
 * 
 * 
 * cdef class _DeflateCodec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implements the deflate codec
 */

struct __pyx_vtabstruct_5cavro__DeflateCodec {
  struct __pyx_vtabstruct_5cavro_Codec __pyx_base;
};
static struct __pyx_vtabstruct_5cavro__DeflateCodec *__pyx_vtabptr_5cavro__DeflateCodec;


/* "src/codec.pxi":122
 * 
 * 
 * cdef class __Bzip2Codec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implements the bzip2 codec
 */

struct __pyx_vtabstruct_5cavro___Bzip2Codec {
  struct __pyx_vtabstruct_5cavro_Codec __pyx_base;
};
static struct __pyx_vtabstruct_5cavro___Bzip2Codec *__pyx_vtabptr_5cavro___Bzip2Codec;


/* "src/codec.pxi":137
 * 
 * 
 * cdef class _LzmaCodec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implements the lzma codec
 */

struct __pyx_vtabstruct_5cavro__LzmaCodec {
  struct __pyx_vtabstruct_5cavro_Codec __pyx_base;
};
static struct __pyx_vtabstruct_5cavro__LzmaCodec *__pyx_vtabptr_5cavro__LzmaCodec;


/* "src/codec.pxi":152
 * 
 * 
 * cdef class _ZStandardCodec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implements the zstandard codec if zstandard is installed.
 */

struct __pyx_vtabstruct_5cavro__ZStandardCodec {
  struct __pyx_vtabstruct_5cavro_Codec __pyx_base;
};
static struct __pyx_vtabstruct_5cavro__ZStandardCodec *__pyx_vtabptr_5cavro__ZStandardCodec;


/* "src/codec.pxi":175
 * 
 * 
 * cdef class _Lz4Codec(Codec):             # <<<<<<<<<<<<<<
 *     """
 *     Implements the lz4 codec if lz4 is installed.
 */

struct __pyx_vtabstruct_5cavro__Lz4Codec {
  struct __pyx_vtabstruct_5cavro_Codec __pyx_base;
};
static struct __pyx_vtabstruct_5cavro__Lz4Codec *__pyx_vtabptr_5cavro__Lz4Codec;


/* "src/container.pxi":41
 * 
 * 
 * cdef class ContainerReader:             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_ContainerReader {
  PyObject *(*_read_marker)(struct __pyx_obj_5cavro_ContainerReader *, int __pyx_skip_dispatch);
  int (*next_block)(struct __pyx_obj_5cavro_ContainerReader *);
  PyObject *(*next_object)(struct __pyx_obj_5cavro_ContainerReader *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5cavro_ContainerReader *__pyx_vtabptr_5cavro_ContainerReader;


/* "src/container.pxi":131
 * 
 * @cython.no_gc_clear
 * cdef class ContainerWriter:             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_ContainerWriter {
  int (*_write_header)(struct __pyx_obj_5cavro_ContainerWriter *);
  int (*_flush_block)(struct __pyx_obj_5cavro_ContainerWriter *, struct __pyx_opt_args_5cavro_15ContainerWriter__flush_block *__pyx_optional_args);
  int (*write_one)(struct __pyx_obj_5cavro_ContainerWriter *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5cavro_ContainerWriter *__pyx_vtabptr_5cavro_ContainerWriter;


/* "src/promotions.pxi":3
 * 
 * 
 * cdef class PromoteToFloat(ValueAdapter):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_PromoteToFloat {
  struct __pyx_vtabstruct_5cavro_ValueAdapter __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_PromoteToFloat *__pyx_vtabptr_5cavro_PromoteToFloat;


/* "src/promotions.pxi":16
 * 
 * 
 * cdef class PromoteBytesToString(ValueAdapter):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_PromoteBytesToString {
  struct __pyx_vtabstruct_5cavro_ValueAdapter __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_PromoteBytesToString *__pyx_vtabptr_5cavro_PromoteBytesToString;


/* "src/promotions.pxi":29
 * 
 * 
 * cdef class PromoteStringToBytes(ValueAdapter):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_PromoteStringToBytes {
  struct __pyx_vtabstruct_5cavro_ValueAdapter __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_PromoteStringToBytes *__pyx_vtabptr_5cavro_PromoteStringToBytes;


/* "src/promotions.pxi":42
 * 
 * 
 * cdef class CannotPromote(ValueAdapter):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

struct __pyx_vtabstruct_5cavro_CannotPromote {
  struct __pyx_vtabstruct_5cavro_ValueAdapter __pyx_base;
};
static struct __pyx_vtabstruct_5cavro_CannotPromote *__pyx_vtabptr_5cavro_CannotPromote;


/* "View.MemoryView":337
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview:             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

struct __pyx_vtabstruct_memoryview {
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);
  PyObject *(*_get_base)(struct __pyx_memoryview_obj *);
};
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;


/* "View.MemoryView":952
 * @cython.collection_type("sequence")
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

struct __pyx_vtabstruct__memoryviewslice {
  struct __pyx_vtabstruct_memoryview __pyx_base;
};
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg // no-op
    #define __Pyx_Arg_XDECREF_VARARGS(arg) // no-op - arg is borrowed
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg // no-op, __Pyx_Arg_FASTCALL is direct and this needs
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)  // no-op - arg was returned from array
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

CYTHON_UNUSED static int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* DivInt[Py_ssize_t].proto */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t, Py_ssize_t);

/* UnaryNegOverflows.proto */
#define __Pyx_UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (1)
#elif CYTHON_COMPILING_IN_LIMITED_API  ||  (CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030C0000)
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

CYTHON_UNUSED static int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PySequenceMultiply.proto */
#define __Pyx_PySequence_Multiply_Left(mul, seq)  __Pyx_PySequence_Multiply(seq, mul)
static CYTHON_INLINE PyObject* __Pyx_PySequence_Multiply(PyObject *seq, Py_ssize_t mul);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* RaiseUnboundLocalError.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* DivInt[long].proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyObject_Str.proto */
#define __Pyx_PyObject_Str(obj)\
    (likely(PyString_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* CBIntToPyUnicode.proto */
#define __Pyx_PyUnicode_FromBInt_bint(value)\
    ((value) ? __Pyx_NewRef(__pyx_n_u_True) : __Pyx_NewRef(__pyx_n_u_False))

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* UnicodeConcatInPlace.proto */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
    #if CYTHON_REFNANNY
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, __pyx_refnanny)
    #else
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right)
    #endif
    static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    );
#else
#define __Pyx_PyUnicode_ConcatInPlace __Pyx_PyUnicode_Concat
#endif
#define __Pyx_PyUnicode_ConcatInPlaceSafe(left, right) ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_ConcatInPlace(left, right))

/* UnicodeAsUCS4.proto */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject*);

/* object_ord.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Ord(c)\
    (likely(PyUnicode_Check(c)) ? (long)__Pyx_PyUnicode_AsPy_UCS4(c) : __Pyx__PyObject_Ord(c))
#else
#define __Pyx_PyObject_Ord(c) __Pyx__PyObject_Ord(c)
#endif
static long __Pyx__PyObject_Ord(PyObject* c);

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AndObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAnd(op1, op2) : PyNumber_And(op1, op2))
#endif

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_MultiplyObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceMultiply(op1, op2) : PyNumber_Multiply(op1, op2))
#endif

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);

/* GeneratorYieldFrom.proto */
static CYTHON_INLINE PyObject* __Pyx_Generator_Yield_From(__pyx_CoroutineObject *gen, PyObject *source);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* RaiseClosureNameError.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* IterNext.proto */
#define __Pyx_PyIter_Next(obj) __Pyx_PyIter_Next2(obj, NULL)
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject *, PyObject *);

/* pybytes_as_double.proto */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj);
static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length);
static CYTHON_INLINE double __Pyx_PyBytes_AsDouble(PyObject *obj) {
    return __Pyx__PyBytes_AsDouble(obj, PyBytes_AS_STRING(obj), PyBytes_GET_SIZE(obj));
}
static CYTHON_INLINE double __Pyx_PyByteArray_AsDouble(PyObject *obj) {
    return __Pyx__PyBytes_AsDouble(obj, PyByteArray_AS_STRING(obj), PyByteArray_GET_SIZE(obj));
}

/* pyunicode_as_double.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static const char* __Pyx__PyUnicode_AsDouble_Copy(const void* data, const int kind, char* buffer, Py_ssize_t start, Py_ssize_t end) {
    int last_was_punctuation;
    Py_ssize_t i;
    last_was_punctuation = 1;
    for (i=start; i <= end; i++) {
        Py_UCS4 chr = PyUnicode_READ(kind, data, i);
        int is_punctuation = (chr == '_') | (chr == '.');
        *buffer = (char)chr;
        buffer += (chr != '_');
        if (unlikely(chr > 127)) goto parse_failure;
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyUnicode_AsDouble_inf_nan(const void* data, int kind, Py_ssize_t start, Py_ssize_t length) {
    int matches = 1;
    Py_UCS4 chr;
    Py_UCS4 sign = PyUnicode_READ(kind, data, start);
    int is_signed = (sign == '-') | (sign == '+');
    start += is_signed;
    length -= is_signed;
    switch (PyUnicode_READ(kind, data, start)) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'a') | (chr == 'A');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'n') | (chr == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'f') | (chr == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+3);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+4);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+5);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+6);
            matches &= (chr == 't') | (chr == 'T');
            chr = PyUnicode_READ(kind, data, start+7);
            matches &= (chr == 'y') | (chr == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static double __Pyx_PyUnicode_AsDouble_WithSpaces(PyObject *obj) {
    double value;
    const char *last;
    char *end;
    Py_ssize_t start, length = PyUnicode_GET_LENGTH(obj);
    const int kind = PyUnicode_KIND(obj);
    const void* data = PyUnicode_DATA(obj);
    start = 0;
    while (Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, start)))
        start++;
    while (start < length - 1 && Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, length - 1)))
        length--;
    length -= start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyUnicode_AsDouble_inf_nan(data, kind, start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    if (length < 40) {
        char number[40];
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((length + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}
#endif
static CYTHON_INLINE double __Pyx_PyUnicode_AsDouble(PyObject *obj) {
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
    if (unlikely(__Pyx_PyUnicode_READY(obj) == -1))
        return (double)-1;
    if (likely(PyUnicode_IS_ASCII(obj))) {
        const char *s;
        Py_ssize_t length;
        s = PyUnicode_AsUTF8AndSize(obj, &length);
        return __Pyx__PyBytes_AsDouble(obj, s, length);
    }
    return __Pyx_PyUnicode_AsDouble_WithSpaces(obj);
#else
    return __Pyx_SlowPyString_AsDouble(obj);
#endif
}

/* pynumber_float.proto */
static CYTHON_INLINE PyObject* __Pyx__PyNumber_Float(PyObject* obj);
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : __Pyx__PyNumber_Float(x))

/* py_dict_keys.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d);

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* py_set_discard.proto */
static CYTHON_INLINE int __Pyx_PySet_Discard(PyObject *set, PyObject *key);

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* py_dict_items.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* SliceTupleAndList.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
#else
#define __Pyx_PyList_GetSlice(seq, start, stop)   PySequence_GetSlice(seq, start, stop)
#define __Pyx_PyTuple_GetSlice(seq, start, stop)  PySequence_GetSlice(seq, start, stop)
#endif

/* PyUnicodeContains.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_ContainsTF(PyObject* substring, PyObject* text, int eq) {
    int result = PyUnicode_Contains(text, substring);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* dict_setdefault.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value, int is_safe_type);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_MultiplyCObj(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceMultiply(op1, op2) : PyNumber_Multiply(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_FloorDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_FloorDivideObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceFloorDivide(op1, op2) : PyNumber_FloorDivide(op1, op2))
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE int __Pyx_PyInt_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_RemainderObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_RemainderObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceRemainder(op1, op2) : PyNumber_Remainder(op1, op2))
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear);

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* ValidateExternBase.proto */
static int __Pyx_validate_extern_base(PyTypeObject *base);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_4
#define __PYX_HAVE_RT_ImportType_proto_3_0_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_4 {
   __Pyx_ImportType_CheckSize_Error_3_0_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_4 check_size);
#endif

/* PatchInspect.proto */
static PyObject* __Pyx_patch_inspect(PyObject* module);

/* Py3UpdateBases.proto */
static PyObject* __Pyx_PEP560_update_bases(PyObject *bases);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* CyFunctionClassCell.proto */
static int __Pyx_CyFunction_InitClassCell(PyObject *cyfunctions, PyObject *classobj);

/* DataclassesCallHelper.proto */
static PyObject* __Pyx_DataclassesCallHelper(PyObject *callable, PyObject *kwds);

/* ModuleLoader.proto */
static PyObject* __Pyx_LoadInternalModule(const char* name, const char* fallback_code);

/* SpecificModuleLoader.proto */
static PyObject* __Pyx_Load_dataclasses_Module(void);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* ClassMethod.proto */
#include "descrobject.h"
CYTHON_UNUSED static PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* ArrayAPI.proto */
#ifndef _ARRAYARRAY_H
#define _ARRAYARRAY_H
typedef struct arraydescr {
    int typecode;
    int itemsize;
    PyObject * (*getitem)(struct arrayobject *, Py_ssize_t);
    int (*setitem)(struct arrayobject *, Py_ssize_t, PyObject *);
#if PY_MAJOR_VERSION >= 3
    char *formats;
#endif
} arraydescr;
struct arrayobject {
    PyObject_HEAD
    Py_ssize_t ob_size;
    union {
        char *ob_item;
        float *as_floats;
        double *as_doubles;
        int *as_ints;
        unsigned int *as_uints;
        unsigned char *as_uchars;
        signed char *as_schars;
        char *as_chars;
        unsigned long *as_ulongs;
        long *as_longs;
#if PY_MAJOR_VERSION >= 3
        unsigned long long *as_ulonglongs;
        long long *as_longlongs;
#endif
        short *as_shorts;
        unsigned short *as_ushorts;
        Py_UNICODE *as_pyunicodes;
        void *as_voidptr;
    } data;
    Py_ssize_t allocated;
    struct arraydescr *ob_descr;
    PyObject *weakreflist;
#if PY_MAJOR_VERSION >= 3
        int ob_exports;
#endif
};
#ifndef NO_NEWARRAY_INLINE
static CYTHON_INLINE PyObject * newarrayobject(PyTypeObject *type, Py_ssize_t size,
    struct arraydescr *descr) {
    arrayobject *op;
    size_t nbytes;
    if (size < 0) {
        PyErr_BadInternalCall();
        return NULL;
    }
    nbytes = size * descr->itemsize;
    if (nbytes / descr->itemsize != (size_t)size) {
        return PyErr_NoMemory();
    }
    op = (arrayobject *) type->tp_alloc(type, 0);
    if (op == NULL) {
        return NULL;
    }
    op->ob_descr = descr;
    op->allocated = size;
    op->weakreflist = NULL;
    __Pyx_SET_SIZE(op, size);
    if (size <= 0) {
        op->data.ob_item = NULL;
    }
    else {
        op->data.ob_item = PyMem_NEW(char, nbytes);
        if (op->data.ob_item == NULL) {
            Py_DECREF(op);
            return PyErr_NoMemory();
        }
    }
    return (PyObject *) op;
}
#else
PyObject* newarrayobject(PyTypeObject *type, Py_ssize_t size,
    struct arraydescr *descr);
#endif
static CYTHON_INLINE int resize(arrayobject *self, Py_ssize_t n) {
    void *items = (void*) self->data.ob_item;
    PyMem_Resize(items, char, (size_t)(n * self->ob_descr->itemsize));
    if (items == NULL) {
        PyErr_NoMemory();
        return -1;
    }
    self->data.ob_item = (char*) items;
    __Pyx_SET_SIZE(self, n);
    self->allocated = n;
    return 0;
}
static CYTHON_INLINE int resize_smart(arrayobject *self, Py_ssize_t n) {
    void *items = (void*) self->data.ob_item;
    Py_ssize_t newsize;
    if (n < self->allocated && n*4 > self->allocated) {
        __Pyx_SET_SIZE(self, n);
        return 0;
    }
    newsize = n + (n / 2) + 1;
    if (newsize <= n) {
        PyErr_NoMemory();
        return -1;
    }
    PyMem_Resize(items, char, (size_t)(newsize * self->ob_descr->itemsize));
    if (items == NULL) {
        PyErr_NoMemory();
        return -1;
    }
    self->data.ob_item = (char*) items;
    __Pyx_SET_SIZE(self, n);
    self->allocated = newsize;
    return 0;
}
#endif

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn_uint8_t__const__(const char *itemp);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* TypeInfoCompare.proto */
static int __pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b);

/* MemviewSliceValidateAndInit.proto */
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(PyObject *, int writable_flag);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_Py_ssize_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_Py_ssize_t(const char *itemp, PyObject *obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_Py_ssize_t(PyObject *, int writable_flag);

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int_type *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int_type *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (&memview->acquisition_count)
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XCLEAR_MEMVIEW(slice, have_gil) __Pyx_XCLEAR_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XCLEAR_MEMVIEW(__Pyx_memviewslice *, int, int);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int32_t __Pyx_PyInt_As_int32_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview__get_base(struct __pyx_memoryview_obj *__pyx_v_self); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice__get_base(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self); /* proto*/
static int __pyx_f_5cavro_7Options_can_have_missing_values(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto*/
static int __pyx_f_5cavro_7_Writer_write_u8(struct __pyx_obj_5cavro__Writer *__pyx_v_self, CYTHON_UNUSED uint8_t __pyx_v_val); /* proto*/
static int __pyx_f_5cavro_7_Writer_write_n(struct __pyx_obj_5cavro__Writer *__pyx_v_self, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_data); /* proto*/
static int __pyx_f_5cavro_7_Writer_flush(CYTHON_UNUSED struct __pyx_obj_5cavro__Writer *__pyx_v_self); /* proto*/
static uint8_t __pyx_f_5cavro_7_Reader_read_u8(struct __pyx_obj_5cavro__Reader *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5cavro_7_Reader_read_bytes(struct __pyx_obj_5cavro__Reader *__pyx_v_self, CYTHON_UNUSED Py_ssize_t __pyx_v_n); /* proto*/
static __Pyx_memviewslice __pyx_f_5cavro_7_Reader_read_n(struct __pyx_obj_5cavro__Reader *__pyx_v_self, CYTHON_UNUSED Py_ssize_t __pyx_v_n); /* proto*/
static uint8_t __pyx_f_5cavro_10FileReader_read_u8(struct __pyx_obj_5cavro_FileReader *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5cavro_10FileReader_read_bytes(struct __pyx_obj_5cavro_FileReader *__pyx_v_self, Py_ssize_t __pyx_v_n); /* proto*/
static __Pyx_memviewslice __pyx_f_5cavro_10FileReader_read_n(struct __pyx_obj_5cavro_FileReader *__pyx_v_self, Py_ssize_t __pyx_v_n); /* proto*/
static int __pyx_f_5cavro_10FileWriter_write_u8(struct __pyx_obj_5cavro_FileWriter *__pyx_v_self, uint8_t __pyx_v_val); /* proto*/
static int __pyx_f_5cavro_10FileWriter_write_n(struct __pyx_obj_5cavro_FileWriter *__pyx_v_self, __Pyx_memviewslice __pyx_v_data); /* proto*/
static int __pyx_f_5cavro_10FileWriter_flush(struct __pyx_obj_5cavro_FileWriter *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5cavro_12MemoryWriter_bytes(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self); /* proto*/
static __Pyx_memviewslice __pyx_f_5cavro_12MemoryWriter_view(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5cavro_12MemoryWriter_reset(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self); /* proto*/
static int __pyx_f_5cavro_12MemoryWriter_write_u8(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self, uint8_t __pyx_v_val); /* proto*/
static int __pyx_f_5cavro_12MemoryWriter_write_n(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self, __Pyx_memviewslice __pyx_v_data); /* proto*/
static void __pyx_f_5cavro_12MemoryReader__reset_to(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, __Pyx_memviewslice __pyx_v_data); /* proto*/
static uint8_t const *__pyx_f_5cavro_12MemoryReader_advance(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, size_t __pyx_v_num); /* proto*/
static CYTHON_INLINE int __pyx_f_5cavro_12MemoryReader_ensure(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, size_t __pyx_v_num); /* proto*/
static uint8_t __pyx_f_5cavro_12MemoryReader_read_u8(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self); /* proto*/
static __Pyx_memviewslice __pyx_f_5cavro_12MemoryReader_read_n(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, Py_ssize_t __pyx_v_n); /* proto*/
static PyObject *__pyx_f_5cavro_12MemoryReader_read_bytes(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, Py_ssize_t __pyx_v_n); /* proto*/
static PyObject *__pyx_f_5cavro_5Rabin_update(struct __pyx_obj_5cavro_Rabin *__pyx_v_self, PyObject *__pyx_v_data, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9ArrayType_copy(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_9ArrayType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_9ArrayType__get_schema_extra(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_9ArrayType__binary_buffer_encode(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9ArrayType__binary_buffer_decode(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static PyObject *__pyx_f_5cavro_9ArrayType__json_format(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9ArrayType__json_decode(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static int __pyx_f_5cavro_9ArrayType__get_value_fitness(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9ArrayType__make_converted_list(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9ArrayType_convert_value(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_9ArrayType_convert_value *__pyx_optional_args); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_9ArrayType_canonical_form(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9ArrayType__for_writer(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer); /* proto*/
static int __pyx_f_5cavro_9ArrayType_accepts_missing_value(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8EnumType_copy(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_8EnumType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_8EnumType__get_schema_extra(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_created, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_8EnumType__binary_buffer_encode(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8EnumType__binary_buffer_decode(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_8EnumType__get_value_fitness(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8EnumType__convert_value(CYTHON_UNUSED struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_value, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_8EnumType__json_format(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8EnumType__json_decode(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_8EnumType_canonical_form(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8EnumType__for_writer(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer); /* proto*/
static PyObject *__pyx_f_5cavro_8EnumType_resolve_default_value(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_schema_default, PyObject *__pyx_v_field); /* proto*/
static PyObject *__pyx_f_5cavro_17PromotingEnumType__json_decode(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_17PromotingEnumType__binary_buffer_decode(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_7MapType_copy(struct __pyx_obj_5cavro_MapType *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_7MapType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_7MapType__get_schema_extra(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_created, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_7MapType__binary_buffer_encode(struct __pyx_obj_5cavro_MapType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_7MapType__binary_buffer_decode(struct __pyx_obj_5cavro_MapType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static PyObject *__pyx_f_5cavro_7MapType__json_format(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_7MapType__json_decode(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static int __pyx_f_5cavro_7MapType__get_value_fitness(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_7MapType__make_converted_map(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_7MapType_convert_value(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_orig_value, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_7MapType_convert_value *__pyx_optional_args); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_7MapType_canonical_form(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_7MapType__for_writer(struct __pyx_obj_5cavro_MapType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer); /* proto*/
static int __pyx_f_5cavro_7MapType_accepts_missing_value(struct __pyx_obj_5cavro_MapType *__pyx_v_self); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8NullType_copy(struct __pyx_obj_5cavro_NullType *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_8NullType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_8NullType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_8NullType__binary_buffer_encode(struct __pyx_obj_5cavro_NullType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8NullType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_8NullType__get_value_fitness(struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8NullType__json_format(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8NullType__json_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8NullType__convert_value(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_8NullType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created); /* proto*/
static int __pyx_f_5cavro_8NullType_accepts_missing_value(struct __pyx_obj_5cavro_NullType *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5cavro_8NullType_resolve_default_value(struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v_schema_default, PyObject *__pyx_v_field); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8BoolType_copy(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_8BoolType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_8BoolType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_BoolType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_8BoolType__binary_buffer_encode(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8BoolType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_BoolType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_8BoolType__get_value_fitness(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8BoolType__convert_value(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_8BoolType__json_format(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8BoolType__json_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_8BoolType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_BoolType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_7IntType_copy(struct __pyx_obj_5cavro_IntType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_7IntType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_7IntType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_IntType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_7IntType__binary_buffer_encode(struct __pyx_obj_5cavro_IntType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_7IntType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_IntType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_7IntType__get_value_fitness(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_7IntType__convert_value(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_7IntType__json_format(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_7IntType__json_decode(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_7IntType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_IntType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8LongType_copy(struct __pyx_obj_5cavro_LongType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_8LongType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_8LongType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_LongType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_8LongType__binary_buffer_encode(struct __pyx_obj_5cavro_LongType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8LongType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_LongType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_8LongType__get_value_fitness(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8LongType__convert_value(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_8LongType_json_format(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8LongType__json_decode(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_8LongType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_LongType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8LongType__for_writer(CYTHON_UNUSED struct __pyx_obj_5cavro_LongType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9FloatType_copy(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_9FloatType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_9FloatType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_FloatType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_9FloatType__binary_buffer_encode(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9FloatType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_FloatType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_9FloatType__get_value_fitness(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9FloatType__convert_value(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_9FloatType__json_format(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9FloatType__json_decode(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_9FloatType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_FloatType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9FloatType__for_writer(CYTHON_UNUSED struct __pyx_obj_5cavro_FloatType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10DoubleType_copy(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_10DoubleType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_10DoubleType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_10DoubleType__binary_buffer_encode(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10DoubleType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_10DoubleType__get_value_fitness(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10DoubleType__convert_value(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_10DoubleType__json_format(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10DoubleType__json_decode(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_10DoubleType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10DoubleType__for_writer(CYTHON_UNUSED struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer); /* proto*/
static PyObject *__pyx_f_5cavro_16_PlaceholderType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_16_PlaceholderType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_16_PlaceholderType__make_logical(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_schema, CYTHON_UNUSED PyObject *__pyx_v_source); /* proto*/
static int __pyx_f_5cavro_16_PlaceholderType_binary_buffer_encode(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_16_PlaceholderType__binary_buffer_decode(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_16_PlaceholderType__get_value_fitness(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_16_PlaceholderType__json_format(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_16_PlaceholderType__json_decode(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_16_PlaceholderType__convert_value(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_16_PlaceholderType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created); /* proto*/
static PyObject *__pyx_f_5cavro_6Record__repr_children(struct __pyx_obj_5cavro_Record *__pyx_v_self, PyObject *__pyx_v_remain); /* proto*/
static PyObject *__pyx_f_5cavro_11RecordField_for_reader(struct __pyx_obj_5cavro_RecordField *__pyx_v_self, PyObject *__pyx_v_reader_name, struct __pyx_obj_5cavro_AvroType *__pyx_v_reader_type); /* proto*/
static PyObject *__pyx_f_5cavro_11RecordField_placeholder(struct __pyx_obj_5cavro_RecordField *__pyx_v_self); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_11RecordField_canonical_form(struct __pyx_obj_5cavro_RecordField *__pyx_v_self, PyObject *__pyx_v_created); /* proto*/
static PyObject *__pyx_f_5cavro_10RecordType__make_logical(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10RecordType_copy(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_10RecordType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_10RecordType__get_schema_extra(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_created, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_10RecordType__binary_buffer_encode(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10RecordType__binary_buffer_decode_record(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static PyObject *__pyx_f_5cavro_10RecordType__binary_buffer_decode_dict(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static PyObject *__pyx_f_5cavro_10RecordType__binary_buffer_decode(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_10RecordType__get_value_fitness(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10RecordType__json_format(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10RecordType_json_decode_record(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10RecordType_json_decode_dict(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10RecordType__json_decode(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10RecordType__convert_value(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_10RecordType_canonical_form(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10RecordType__for_writer(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer); /* proto*/
static PyObject *__pyx_f_5cavro_19PromotingRecordType__binary_buffer_decode_record(struct __pyx_obj_5cavro_PromotingRecordType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9BytesType_copy(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_9BytesType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_9BytesType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_BytesType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_9BytesType__binary_buffer_encode(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9BytesType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_BytesType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_9BytesType__get_value_fitness(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9BytesType__json_format(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9BytesType__json_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9BytesType__convert_value(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_9BytesType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_BytesType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9BytesType__for_writer(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer); /* proto*/
static PyObject *__pyx_f_5cavro_9BytesType_resolve_default_value(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_schema_default, PyObject *__pyx_v_field); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10StringType_copy(struct __pyx_obj_5cavro_StringType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_10StringType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_10StringType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_StringType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_10StringType__binary_buffer_encode(struct __pyx_obj_5cavro_StringType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10StringType__binary_buffer_decode(struct __pyx_obj_5cavro_StringType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_10StringType__get_value_fitness(struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10StringType__json_format(struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10StringType__json_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10StringType__convert_value(struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_10StringType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_StringType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10StringType__for_writer(struct __pyx_obj_5cavro_StringType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9FixedType_copy(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_9FixedType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_9FixedType__get_schema_extra(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_9FixedType__binary_buffer_encode(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9FixedType__binary_buffer_decode(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_9FixedType__get_value_fitness(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9FixedType__json_format(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9FixedType__json_decode(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9FixedType__convert_value(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_9FixedType_canonical_form(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9FixedType__for_writer(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer); /* proto*/
static PyObject *__pyx_f_5cavro_9FixedType_resolve_default_value(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_schema_default, PyObject *__pyx_v_field); /* proto*/
static PyObject *__pyx_f_5cavro_9UnionType__populate_name_map(struct __pyx_obj_5cavro_UnionType *__pyx_v_self); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9UnionType_copy(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_9UnionType__make_logical(CYTHON_UNUSED struct __pyx_obj_5cavro_UnionType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_schema, CYTHON_UNUSED PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_9UnionType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_UnionType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_9UnionType_get_schema(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_9UnionType_get_schema *__pyx_optional_args); /* proto*/
static Py_ssize_t __pyx_f_5cavro_9UnionType_resolve_from_value(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static int __pyx_f_5cavro_9UnionType__binary_buffer_encode(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9UnionType__binary_buffer_decode(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_9UnionType__get_value_fitness(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9UnionType__json_format(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9UnionType__json_decode(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_9UnionType_convert_value(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_9UnionType_convert_value *__pyx_optional_args); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_9UnionType_canonical_form(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_created); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9UnionType__for_writer(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9UnionType_for_reader(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_reader); /* proto*/
static int __pyx_f_5cavro_9UnionType_accepts_missing_value(struct __pyx_obj_5cavro_UnionType *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5cavro_9UnionType_resolve_default_value(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_schema_default, PyObject *__pyx_v_field); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8AvroType_clone_base(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, struct __pyx_opt_args_5cavro_8AvroType_clone_base *__pyx_optional_args); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8AvroType_copy(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType__make_logical(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType__extract_metadata(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_source); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType__convert_value(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType_convert_value(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_8AvroType_convert_value *__pyx_optional_args); /* proto*/
static int __pyx_f_5cavro_8AvroType_binary_buffer_encode(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value); /* proto*/
static int __pyx_f_5cavro_8AvroType__binary_buffer_encode(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType_binary_buffer_decode(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType__binary_buffer_decode(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Reader *__pyx_v_buffer); /* proto*/
static int __pyx_f_5cavro_8AvroType_get_value_fitness(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static int __pyx_f_5cavro_8AvroType__get_value_fitness(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static int __pyx_f_5cavro_8AvroType_assert_value(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType_json_format(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType__json_format(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType_json_decode(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType__json_decode(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_8AvroType_canonical_form(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType__get_schema_extra(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8AvroType_for_writer(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer, struct __pyx_opt_args_5cavro_8AvroType_for_writer *__pyx_optional_args); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8AvroType__for_writer(CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_writer); /* proto*/
static int __pyx_f_5cavro_8AvroType_accepts_missing_value(CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType_resolve_default_value(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_schema_default, PyObject *__pyx_v_field); /* proto*/
static PyObject *__pyx_f_5cavro_8AvroType_get_schema(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_8AvroType_get_schema *__pyx_optional_args); /* proto*/
static int __pyx_f_5cavro_10_NamedType_name_matches(struct __pyx_obj_5cavro__NamedType *__pyx_v_self, struct __pyx_obj_5cavro__NamedType *__pyx_v_other); /* proto*/
static PyObject *__pyx_f_5cavro_10_NamedType_get_namespaced_aliases(struct __pyx_obj_5cavro__NamedType *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5cavro_10_NamedType__get_schema_extra(struct __pyx_obj_5cavro__NamedType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10_NamedType_clone_base(struct __pyx_obj_5cavro__NamedType *__pyx_v_self, struct __pyx_opt_args_5cavro_10_NamedType_clone_base *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_5cavro_10_NamedType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro__NamedType *__pyx_v_self, PyObject *__pyx_v_source); /* proto*/
static void __pyx_f_5cavro_6Schema_register_type(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_name, struct __pyx_obj_5cavro_AvroType *__pyx_v_avro_type); /* proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_6Schema_find_type(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_name, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_6Schema_find_type *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_5cavro_6Schema_binary_read(struct __pyx_obj_5cavro_Schema *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_6Schema_binary_write(struct __pyx_obj_5cavro_Schema *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, PyObject *__pyx_v_value, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_5cavro_Schema *__pyx_f_5cavro_6Schema_reader_for_writer(struct __pyx_obj_5cavro_Schema *__pyx_v_self, struct __pyx_obj_5cavro_Schema *__pyx_v_writer_schema, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_12ValueAdapter_encode_value(struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_12ValueAdapter_decode_value(struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_17CustomLogicalType_encode_value(struct __pyx_obj_5cavro_CustomLogicalType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_17CustomLogicalType_decode_value(struct __pyx_obj_5cavro_CustomLogicalType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_11DecimalType_encode_value(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_11DecimalType_decode_value(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_14UUIDStringType_encode_value(struct __pyx_obj_5cavro_UUIDStringType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_14UUIDStringType_decode_value(struct __pyx_obj_5cavro_UUIDStringType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_13UUIDFixedType_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_UUIDFixedType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_13UUIDFixedType_decode_value(struct __pyx_obj_5cavro_UUIDFixedType *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_4Date_encode_value(struct __pyx_obj_5cavro_Date *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_4Date_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_Date *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10TimeMillis_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimeMillis *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10TimeMillis_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimeMillis *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10TimeMicros_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimeMicros *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_10TimeMicros_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimeMicros *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_15TimestampMillis_encode_value(struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_15TimestampMillis_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_15TimestampMicros_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimestampMicros *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_15TimestampMicros_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimestampMicros *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static __Pyx_memviewslice __pyx_f_5cavro_5Codec_read_block(struct __pyx_obj_5cavro_Codec *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Reader *__pyx_v_reader, CYTHON_UNUSED size_t __pyx_v_length); /* proto*/
static Py_ssize_t __pyx_f_5cavro_5Codec_write_block(struct __pyx_obj_5cavro_Codec *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Writer *__pyx_v_writer, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_data); /* proto*/
static __Pyx_memviewslice __pyx_f_5cavro_12_SnappyCodec_read_block(CYTHON_UNUSED struct __pyx_obj_5cavro__SnappyCodec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length); /* proto*/
static Py_ssize_t __pyx_f_5cavro_12_SnappyCodec_write_block(CYTHON_UNUSED struct __pyx_obj_5cavro__SnappyCodec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data); /* proto*/
static __Pyx_memviewslice __pyx_f_5cavro_10_NullCodec_read_block(CYTHON_UNUSED struct __pyx_obj_5cavro__NullCodec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length); /* proto*/
static Py_ssize_t __pyx_f_5cavro_10_NullCodec_write_block(CYTHON_UNUSED struct __pyx_obj_5cavro__NullCodec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data); /* proto*/
static __Pyx_memviewslice __pyx_f_5cavro_13_DeflateCodec_read_block(CYTHON_UNUSED struct __pyx_obj_5cavro__DeflateCodec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length); /* proto*/
static Py_ssize_t __pyx_f_5cavro_13_DeflateCodec_write_block(CYTHON_UNUSED struct __pyx_obj_5cavro__DeflateCodec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data); /* proto*/
static __Pyx_memviewslice __pyx_f_5cavro_12__Bzip2Codec_read_block(CYTHON_UNUSED struct __pyx_obj_5cavro___Bzip2Codec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length); /* proto*/
static Py_ssize_t __pyx_f_5cavro_12__Bzip2Codec_write_block(CYTHON_UNUSED struct __pyx_obj_5cavro___Bzip2Codec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data); /* proto*/
static __Pyx_memviewslice __pyx_f_5cavro_10_LzmaCodec_read_block(CYTHON_UNUSED struct __pyx_obj_5cavro__LzmaCodec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length); /* proto*/
static Py_ssize_t __pyx_f_5cavro_10_LzmaCodec_write_block(CYTHON_UNUSED struct __pyx_obj_5cavro__LzmaCodec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data); /* proto*/
static __Pyx_memviewslice __pyx_f_5cavro_15_ZStandardCodec_read_block(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length); /* proto*/
static Py_ssize_t __pyx_f_5cavro_15_ZStandardCodec_write_block(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data); /* proto*/
static __Pyx_memviewslice __pyx_f_5cavro_9_Lz4Codec_read_block(CYTHON_UNUSED struct __pyx_obj_5cavro__Lz4Codec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length); /* proto*/
static Py_ssize_t __pyx_f_5cavro_9_Lz4Codec_write_block(CYTHON_UNUSED struct __pyx_obj_5cavro__Lz4Codec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data); /* proto*/
static PyObject *__pyx_f_5cavro_15ContainerReader__read_marker(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_15ContainerReader_next_block(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5cavro_15ContainerReader_next_object(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5cavro_15ContainerWriter__write_header(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto*/
static int __pyx_f_5cavro_15ContainerWriter__flush_block(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, struct __pyx_opt_args_5cavro_15ContainerWriter__flush_block *__pyx_optional_args); /* proto*/
static int __pyx_f_5cavro_15ContainerWriter_write_one(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5cavro_14PromoteToFloat_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_PromoteToFloat *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_14PromoteToFloat_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_PromoteToFloat *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_20PromoteBytesToString_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_PromoteBytesToString *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_20PromoteBytesToString_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_PromoteBytesToString *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_20PromoteStringToBytes_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_PromoteStringToBytes *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_20PromoteStringToBytes_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_PromoteStringToBytes *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_13CannotPromote_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_5cavro_13CannotPromote_decode_value(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value); /* proto*/

/* Module declarations from "cython.view" */

/* Module declarations from "cython.dataclasses" */

/* Module declarations from "cython" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython.version" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.module" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.sequence" */

/* Module declarations from "cpython.mapping" */

/* Module declarations from "cpython.iterator" */

/* Module declarations from "cpython.number" */

/* Module declarations from "cpython.int" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.float" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.complex" */

/* Module declarations from "cpython.string" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.instance" */

/* Module declarations from "cpython.function" */

/* Module declarations from "cpython.method" */

/* Module declarations from "cpython.weakref" */

/* Module declarations from "cpython.getargs" */

/* Module declarations from "cpython.pythread" */

/* Module declarations from "cpython.pystate" */

/* Module declarations from "cpython.cobject" */

/* Module declarations from "cpython.oldbuffer" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "cpython.contextvars" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "libc.stdint" */

/* Module declarations from "array" */

/* Module declarations from "cpython.array" */
static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_clone(arrayobject *, Py_ssize_t, int); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_5array_extend_buffer(arrayobject *, char *, Py_ssize_t); /*proto*/

/* Module declarations from "libc.float" */

/* Module declarations from "cavro" */
static arrayobject *__pyx_v_5cavro_byte_buffer_template = 0;
static PyObject *__pyx_v_5cavro_empty_buffer = 0;
static int64_t __pyx_v_5cavro_RABIN_EMPTY;
static int64_t __pyx_v_5cavro_RABIN_TABLE[0x100];
static int __pyx_v_5cavro_RABIN_TABLE_CONFIGURED;
static float __pyx_v_5cavro_FLOAT_INT_THRESHOLD;
static int __pyx_v_5cavro_FIT_NONE;
static int __pyx_v_5cavro_FIT_POOR;
static int __pyx_v_5cavro_FIT_OK;
static int __pyx_v_5cavro_FIT_EXACT;
static __Pyx_memviewslice __pyx_v_5cavro_OBJ_MAGIC = { 0, 0, { 0 }, { 0 }, { 0 } };
static PyObject *__pyx_dataclass_dflt5cavro_Options5cavro_union_decodes_to;
static PyObject *__pyx_collections_abc_Sequence = 0;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static PyObject *__pyx_f_5cavro__get_unicode_name_re(void); /*proto*/
static uint32_t __pyx_f_5cavro_read_varint(struct __pyx_obj_5cavro__Reader *); /*proto*/
static uint64_t __pyx_f_5cavro_read_varlong(struct __pyx_obj_5cavro__Reader *); /*proto*/
static int32_t __pyx_f_5cavro_zigzag_decode_int(struct __pyx_obj_5cavro__Reader *); /*proto*/
static int64_t __pyx_f_5cavro_zigzag_decode_long(struct __pyx_obj_5cavro__Reader *); /*proto*/
static int __pyx_f_5cavro_zigzag_encode_int(struct __pyx_obj_5cavro__Writer *, int32_t); /*proto*/
static int __pyx_f_5cavro_zigzag_encode_long(struct __pyx_obj_5cavro__Writer *, int64_t); /*proto*/
static PyObject *__pyx_f_5cavro_init_rabin_table(void); /*proto*/
static PyObject *__pyx_f_5cavro__ubin(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static uint32_t __pyx_f_5cavro_rand(struct __pyx_t_5cavro_random_t *); /*proto*/
static PyObject *__pyx_f_5cavro_boring_varint_encoder(uint64_t); /*proto*/
static PyObject *__pyx_f_5cavro__record_data_from_dict(struct __pyx_obj_5cavro_RecordType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro__make_record_class(struct __pyx_obj_5cavro_RecordType *); /*proto*/
static int __pyx_f_5cavro_are_union_values_ambiguous(PyObject *); /*proto*/
static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_dict_to_canonical(PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro__strip_keys(PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro_resolve_namespaced_name(PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro_dt_to_timedelta(PyObject *); /*proto*/
static int __pyx_f_5cavro_viewcmp(__Pyx_memviewslice, __Pyx_memviewslice); /*proto*/
static struct __pyx_obj_5cavro__Reader *__pyx_f_5cavro_make_reader(PyObject *); /*proto*/
static struct __pyx_obj_5cavro__Writer *__pyx_f_5cavro_make_writer(PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_Options__set_state(struct __pyx_obj_5cavro_Options *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__Writer__set_state(struct __pyx_obj_5cavro__Writer *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__Reader__set_state(struct __pyx_obj_5cavro__Reader *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_FileReader__set_state(struct __pyx_obj_5cavro_FileReader *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_FileWriter__set_state(struct __pyx_obj_5cavro_FileWriter *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_MemoryReader__set_state(struct __pyx_obj_5cavro_MemoryReader *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_Rabin__set_state(struct __pyx_obj_5cavro_Rabin *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_ArrayType__set_state(struct __pyx_obj_5cavro_ArrayType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_EnumType__set_state(struct __pyx_obj_5cavro_EnumType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_PromotingEnumType__set_state(struct __pyx_obj_5cavro_PromotingEnumType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_MapType__set_state(struct __pyx_obj_5cavro_MapType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_NullType__set_state(struct __pyx_obj_5cavro_NullType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_BoolType__set_state(struct __pyx_obj_5cavro_BoolType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_IntType__set_state(struct __pyx_obj_5cavro_IntType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_LongType__set_state(struct __pyx_obj_5cavro_LongType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_FloatType__set_state(struct __pyx_obj_5cavro_FloatType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_DoubleType__set_state(struct __pyx_obj_5cavro_DoubleType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__PlaceholderType__set_state(struct __pyx_obj_5cavro__PlaceholderType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_Record__set_state(struct __pyx_obj_5cavro_Record *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_RecordField__set_state(struct __pyx_obj_5cavro_RecordField *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__FieldAccessor__set_state(struct __pyx_obj_5cavro__FieldAccessor *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_RecordType__set_state(struct __pyx_obj_5cavro_RecordType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_PromotingRecordType__set_state(struct __pyx_obj_5cavro_PromotingRecordType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_BytesType__set_state(struct __pyx_obj_5cavro_BytesType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_StringType__set_state(struct __pyx_obj_5cavro_StringType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_FixedType__set_state(struct __pyx_obj_5cavro_FixedType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_UnionType__set_state(struct __pyx_obj_5cavro_UnionType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__Sentinel__set_state(struct __pyx_obj_5cavro__Sentinel *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__CanonicalForm__set_state(struct __pyx_obj_5cavro__CanonicalForm *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_AvroType__set_state(struct __pyx_obj_5cavro_AvroType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__NamedType__set_state(struct __pyx_obj_5cavro__NamedType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_Schema__set_state(struct __pyx_obj_5cavro_Schema *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_ResolvedSchema__set_state(struct __pyx_obj_5cavro_ResolvedSchema *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_ValueAdapter__set_state(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_LogicalType__set_state(struct __pyx_obj_5cavro_LogicalType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_CustomLogicalType__set_state(struct __pyx_obj_5cavro_CustomLogicalType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_DecimalType__set_state(struct __pyx_obj_5cavro_DecimalType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_UUIDBase__set_state(struct __pyx_obj_5cavro_UUIDBase *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_UUIDStringType__set_state(struct __pyx_obj_5cavro_UUIDStringType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_UUIDFixedType__set_state(struct __pyx_obj_5cavro_UUIDFixedType *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_Date__set_state(struct __pyx_obj_5cavro_Date *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_TimeMillis__set_state(struct __pyx_obj_5cavro_TimeMillis *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_TimeMicros__set_state(struct __pyx_obj_5cavro_TimeMicros *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_TimestampMillis__set_state(struct __pyx_obj_5cavro_TimestampMillis *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_TimestampMicros__set_state(struct __pyx_obj_5cavro_TimestampMicros *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_Codec__set_state(struct __pyx_obj_5cavro_Codec *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__SnappyCodec__set_state(struct __pyx_obj_5cavro__SnappyCodec *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__NullCodec__set_state(struct __pyx_obj_5cavro__NullCodec *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__DeflateCodec__set_state(struct __pyx_obj_5cavro__DeflateCodec *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle___Bzip2Codec__set_state(struct __pyx_obj_5cavro___Bzip2Codec *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__LzmaCodec__set_state(struct __pyx_obj_5cavro__LzmaCodec *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__ZStandardCodec__set_state(struct __pyx_obj_5cavro__ZStandardCodec *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle__Lz4Codec__set_state(struct __pyx_obj_5cavro__Lz4Codec *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_ContainerReader__set_state(struct __pyx_obj_5cavro_ContainerReader *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_PromoteToFloat__set_state(struct __pyx_obj_5cavro_PromoteToFloat *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_PromoteBytesToString__set_state(struct __pyx_obj_5cavro_PromoteBytesToString *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_PromoteStringToBytes__set_state(struct __pyx_obj_5cavro_PromoteStringToBytes *, PyObject *); /*proto*/
static PyObject *__pyx_f_5cavro___pyx_unpickle_CannotPromote__set_state(struct __pyx_obj_5cavro_CannotPromote *, PyObject *); /*proto*/
static int __pyx_array_allocate_buffer(struct __pyx_array_obj *); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static int assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, PyObject *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, PyObject *); /*proto*/
static int __pyx_memoryview_err_no_memory(void); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
static __Pyx_TypeInfo __Pyx_TypeInfo_nn_uint8_t__const__ = { "const uint8_t", NULL, sizeof(uint8_t const ), { 0 }, 0, __PYX_IS_UNSIGNED(uint8_t const ) ? 'U' : 'I', __PYX_IS_UNSIGNED(uint8_t const ), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_Py_ssize_t = { "Py_ssize_t", NULL, sizeof(Py_ssize_t), { 0 }, 0, __PYX_IS_UNSIGNED(Py_ssize_t) ? 'U' : 'I', __PYX_IS_UNSIGNED(Py_ssize_t), 0 };
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "cavro"
extern int __pyx_module_is_main_cavro;
int __pyx_module_is_main_cavro = 0;

/* Implementation of "cavro" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin___import__;
static PyObject *__pyx_builtin_NotImplemented;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_chr;
static PyObject *__pyx_builtin_eval;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_EOFError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_bin;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_print;
static PyObject *__pyx_builtin_hex;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_OverflowError;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_UnicodeEncodeError;
static PyObject *__pyx_builtin_reversed;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_Ellipsis;
/* #### Code section: string_decls ### */
static const char __pyx_k_[] = ": ";
static const char __pyx_k_0[] = "0";
static const char __pyx_k_1[] = "1";
static const char __pyx_k_B[] = "B";
static const char __pyx_k_O[] = "O";
static const char __pyx_k_a[] = "a";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_e[] = "e";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_l[] = "l";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_q[] = "q";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_w[] = "_]\\w*";
static const char __pyx_k_Ll[] = "Ll";
static const char __pyx_k_Lu[] = "Lu";
static const char __pyx_k__2[] = ".";
static const char __pyx_k__3[] = "*";
static const char __pyx_k__6[] = "'";
static const char __pyx_k__7[] = ")";
static const char __pyx_k__9[] = "-";
static const char __pyx_k_at[] = " at ";
static const char __pyx_k_eq[] = "eq";
static const char __pyx_k_fn[] = "fn";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_io[] = "io";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_ns[] = " ns";
static const char __pyx_k_ob[] = "ob";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_rb[] = "rb";
static const char __pyx_k_re[] = "re";
static const char __pyx_k_to[] = " to ";
static const char __pyx_k_wb[] = "wb";
static const char __pyx_k_xz[] = "xz";
static const char __pyx_k_ASC[] = "ASC";
static const char __pyx_k_Obj[] = "Obj\001";
static const char __pyx_k__13[] = "";
static const char __pyx_k__14[] = "[";
static const char __pyx_k__18[] = "...";
static const char __pyx_k__24[] = " ";
static const char __pyx_k__25[] = ":\n   ";
static const char __pyx_k__26[] = "\n!= ";
static const char __pyx_k__27[] = "\n   ";
static const char __pyx_k__34[] = "\001\002\003";
static const char __pyx_k__48[] = "\"";
static const char __pyx_k__60[] = "{...}";
static const char __pyx_k__61[] = "{";
static const char __pyx_k__62[] = "}";
static const char __pyx_k__63[] = ", ";
static const char __pyx_k__64[] = ">";
static const char __pyx_k__70[] = " != ";
static const char __pyx_k__72[] = "\000";
static const char __pyx_k__76[] = "')";
static const char __pyx_k__77[] = ",";
static const char __pyx_k__78[] = "]";
static const char __pyx_k__79[] = "<";
static const char __pyx_k__80[] = ":";
static const char __pyx_k_abc[] = "abc";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_and[] = " and ";
static const char __pyx_k_big[] = "big";
static const char __pyx_k_bin[] = "bin";
static const char __pyx_k_bz2[] = "bz2";
static const char __pyx_k_chr[] = "chr";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_dir[] = "__dir__";
static const char __pyx_k_doc[] = "doc";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_got[] = " (got ";
static const char __pyx_k_hex[] = "hex";
static const char __pyx_k_inf[] = "-inf";
static const char __pyx_k_int[] = "\"int\"";
static const char __pyx_k_lz4[] = "lz4";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_msg[] = "msg";
static const char __pyx_k_nan[] = "nan";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_out[] = "out";
static const char __pyx_k_pad[] = "pad";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_src[] = "src";
static const char __pyx_k_str[] = "__str__";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_t_a[] = "t_a";
static const char __pyx_k_utc[] = "utc";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_DESC[] = "DESC";
static const char __pyx_k_Date[] = "Date";
static const char __pyx_k_Enum[] = "Enum";
static const char __pyx_k_Name[] = "Name ";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_Path[] = "Path";
static const char __pyx_k_True[] = "True";
static const char __pyx_k_Type[] = "Type";
static const char __pyx_k_UUID[] = "UUID";
static const char __pyx_k__122[] = "?";
static const char __pyx_k__152[] = "_";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_bint[] = "bint";
static const char __pyx_k_bool[] = "bool";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_date[] = "date";
static const char __pyx_k_days[] = "days";
static const char __pyx_k_dest[] = "dest";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_diff[] = "diff";
static const char __pyx_k_enum[] = "enum";
static const char __pyx_k_eval[] = "eval";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_func[] = "func";
static const char __pyx_k_hash[] = "hash";
static const char __pyx_k_hour[] = "hour";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_inst[] = "inst";
static const char __pyx_k_json[] = "json";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_long[] = "\"long\"";
static const char __pyx_k_lzma[] = "lzma";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mask[] = "mask";
static const char __pyx_k_math[] = "math";
static const char __pyx_k_meta[] = "meta";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_null[] = "\"null\"";
static const char __pyx_k_objs[] = "objs";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_perf[] = "_perf";
static const char __pyx_k_prec[] = "prec";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_repr[] = "repr";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "test";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_ubin[] = "_ubin";
static const char __pyx_k_utf8[] = "utf8";
static const char __pyx_k_uuid[] = "uuid";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_zlib[] = "zlib";
static const char __pyx_k_1_0_0[] = "1.0.0";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_Codec[] = "Codec";
static const char __pyx_k_FIELD[] = "_FIELD";
static const char __pyx_k_False[] = "False";
static const char __pyx_k_Field[] = "Field ";
static const char __pyx_k_Order[] = "Order";
static const char __pyx_k_Rabin[] = "Rabin";
static const char __pyx_k_Union[] = "Union";
static const char __pyx_k_Value[] = "Value ";
static const char __pyx_k_Y_m_d[] = "%Y-%m-%d";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_bytes[] = " bytes";
static const char __pyx_k_bzip2[] = "bzip2";
static const char __pyx_k_cavro[] = "cavro";
static const char __pyx_k_clamp[] = "clamp";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_codec[] = "codec";
static const char __pyx_k_count[] = "count";
static const char __pyx_k_crc32[] = "crc32";
static const char __pyx_k_doc_2[] = "__doc__";
static const char __pyx_k_dumps[] = "dumps";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_extra[] = "extra";
static const char __pyx_k_field[] = "field";
static const char __pyx_k_fixed[] = "fixed";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_float[] = "\"float\"";
static const char __pyx_k_floor[] = "floor";
static const char __pyx_k_flush[] = "flush";
static const char __pyx_k_force[] = "force";
static const char __pyx_k_frame[] = "frame";
static const char __pyx_k_get_2[] = "__get__";
static const char __pyx_k_given[] = "given";
static const char __pyx_k_got_2[] = ", got: ";
static const char __pyx_k_got_3[] = " got ";
static const char __pyx_k_got_4[] = ", got ";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_inf_2[] = "inf";
static const char __pyx_k_int_2[] = "int";
static const char __pyx_k_isinf[] = "isinf";
static const char __pyx_k_isnan[] = "isnan";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_loads[] = "loads";
static const char __pyx_k_local[] = "local";
static const char __pyx_k_log10[] = "log10";
static const char __pyx_k_new_2[] = "new";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_order[] = "order";
static const char __pyx_k_other[] = "other";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_rabin[] = "rabin";
static const char __pyx_k_raise[] = "_raise";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_rjust[] = "rjust";
static const char __pyx_k_scale[] = "scale";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_slots[] = "__slots__";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_str_2[] = "str";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_tests[] = "__tests";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_types[] = "types";
static const char __pyx_k_union[] = "union";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_uuid4[] = "uuid4";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_wbits[] = "wbits";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_CODECS[] = "CODECS";
static const char __pyx_k_IGNORE[] = "IGNORE";
static const char __pyx_k_Reader[] = "_Reader";
static const char __pyx_k_Record[] = "Record ";
static const char __pyx_k_Schema[] = "Schema";
static const char __pyx_k_Writer[] = "_Writer";
static const char __pyx_k_abacus[] = "abacus";
static const char __pyx_k_actual[] = "actual";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_asdict[] = "_asdict";
static const char __pyx_k_bool_2[] = "bool_";
static const char __pyx_k_buffer[] = "buffer";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_dict_2[] = "_dict";
static const char __pyx_k_digest[] = "digest";
static const char __pyx_k_double[] = "\"double\"";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_equals[] = "equals";
static const char __pyx_k_errors[] = "errors";
static const char __pyx_k_escape[] = "escape";
static const char __pyx_k_eval_2[] = "_eval";
static const char __pyx_k_exc_tb[] = "exc_tb";
static const char __pyx_k_fields[] = "fields";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_frozen[] = "frozen";
static const char __pyx_k_hasher[] = "hasher";
static const char __pyx_k_ignore[] = "ignore";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_indent[] = "indent";
static const char __pyx_k_init_2[] = "init";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_latin1[] = "latin1";
static const char __pyx_k_little[] = "little";
static const char __pyx_k_long_2[] = "long";
static const char __pyx_k_marker[] = "marker";
static const char __pyx_k_method[] = "method";
static const char __pyx_k_minute[] = "minute";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_np_f16[] = "np_f16";
static const char __pyx_k_np_f32[] = "np_f32";
static const char __pyx_k_np_f64[] = "np_f64";
static const char __pyx_k_null_2[] = "null";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_pytest[] = "pytest";
static const char __pyx_k_raises[] = "raises";
static const char __pyx_k_reader[] = "reader";
static const char __pyx_k_record[] = "record";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_rsplit[] = "rsplit";
static const char __pyx_k_s_read[] = " s, read: ";
static const char __pyx_k_scaleb[] = "scaleb";
static const char __pyx_k_schema[] = "schema";
static const char __pyx_k_second[] = "second";
static const char __pyx_k_signed[] = "signed";
static const char __pyx_k_snappy[] = "snappy";
static const char __pyx_k_source[] = "source";
static const char __pyx_k_strict[] = "strict";
static const char __pyx_k_string[] = "string";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_test_2[] = "__test__";
static const char __pyx_k_type_2[] = "-type";
static const char __pyx_k_type_3[] = "_type";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_tzinfo[] = "tzinfo";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_writer[] = "writer";
static const char __pyx_k_Context[] = "Context";
static const char __pyx_k_Decimal[] = "Decimal";
static const char __pyx_k_HAVE_XZ[] = "HAVE_XZ";
static const char __pyx_k_IntType[] = "IntType";
static const char __pyx_k_MapType[] = "MapType";
static const char __pyx_k_Options[] = "Options";
static const char __pyx_k_UNKNOWN[] = "UNKNOWN";
static const char __pyx_k_aliases[] = "aliases";
static const char __pyx_k_boolean[] = "\"boolean\"";
static const char __pyx_k_bytes_2[] = "\"bytes\"";
static const char __pyx_k_bytes_3[] = "bytes";
static const char __pyx_k_changes[] = "changes";
static const char __pyx_k_compare[] = "compare";
static const char __pyx_k_compile[] = "compile";
static const char __pyx_k_context[] = "context";
static const char __pyx_k_counter[] = "counter";
static const char __pyx_k_created[] = "created";
static const char __pyx_k_decimal[] = "decimal";
static const char __pyx_k_default[] = "default";
static const char __pyx_k_deflate[] = "deflate";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_exc_val[] = "exc_val";
static const char __pyx_k_fitness[] = "fitness";
static const char __pyx_k_float16[] = "float16";
static const char __pyx_k_float32[] = "float32";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_float_2[] = "float";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_getitem[] = "__getitem__";
static const char __pyx_k_hashlib[] = "hashlib";
static const char __pyx_k_inspect[] = "inspect";
static const char __pyx_k_integer[] = "integer";
static const char __pyx_k_items_2[] = " items";
static const char __pyx_k_kw_only[] = "kw_only";
static const char __pyx_k_latin_1[] = "latin-1";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_missing[] = "<missing>";
static const char __pyx_k_n_bytes[] = "n_bytes";
static const char __pyx_k_ndarray[] = "ndarray";
static const char __pyx_k_new_ver[] = "new_ver";
static const char __pyx_k_ns_copy[] = " ns, copy: ";
static const char __pyx_k_options[] = "options";
static const char __pyx_k_partial[] = "partial";
static const char __pyx_k_pathlib[] = "pathlib";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_replace[] = "replace";
static const char __pyx_k_running[] = "running";
static const char __pyx_k_seconds[] = "seconds";
static const char __pyx_k_slots_2[] = "slots";
static const char __pyx_k_symbols[] = "symbols";
static const char __pyx_k_tests_2[] = "_tests";
static const char __pyx_k_value_2[] = " (value: ";
static const char __pyx_k_version[] = "__version__";
static const char __pyx_k_visited[] = "visited";
static const char __pyx_k_wrapped[] = "__wrapped__";
static const char __pyx_k_write_2[] = ": write: ";
static const char __pyx_k_AvroType[] = "AvroType";
static const char __pyx_k_BoolType[] = "BoolType";
static const char __pyx_k_EMPTY_MP[] = "_EMPTY_MP";
static const char __pyx_k_EOFError[] = "EOFError";
static const char __pyx_k_EPOCH_DT[] = "EPOCH_DT";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_EnumType[] = "EnumType";
static const char __pyx_k_HAVE_LZ4[] = "HAVE_LZ4";
static const char __pyx_k_Iterable[] = "Iterable";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_LongType[] = "LongType";
static const char __pyx_k_Lz4Codec[] = "_Lz4Codec";
static const char __pyx_k_MAX_PREC[] = "MAX_PREC";
static const char __pyx_k_MAX_TIME[] = "MAX_TIME";
static const char __pyx_k_NullType[] = "NullType";
static const char __pyx_k_Record_2[] = "<Record:";
static const char __pyx_k_Record_3[] = "Record";
static const char __pyx_k_Sentinel[] = "_Sentinel";
static const char __pyx_k_Sequence[] = "Sequence";
static const char __pyx_k_UUIDBase[] = "UUIDBase";
static const char __pyx_k_as_tuple[] = "as_tuple";
static const char __pyx_k_category[] = "category";
static const char __pyx_k_compress[] = "compress";
static const char __pyx_k_datetime[] = "datetime";
static const char __pyx_k_double_2[] = "double";
static const char __pyx_k_exc_type[] = "exc_type";
static const char __pyx_k_file_obj[] = "file_obj";
static const char __pyx_k_for_type[] = " for type ";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_int_date[] = "int(date)";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_metadata[] = "metadata";
static const char __pyx_k_new_inst[] = "new_inst";
static const char __pyx_k_path_str[] = "path_str";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_quantize[] = "quantize";
static const char __pyx_k_register[] = "register";
static const char __pyx_k_reversed[] = "reversed";
static const char __pyx_k_self_val[] = "self_val";
static const char __pyx_k_set_name[] = "__set_name__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_string_2[] = "\"string\"";
static const char __pyx_k_strptime[] = "strptime";
static const char __pyx_k_timezone[] = "timezone";
static const char __pyx_k_to_bytes[] = "to_bytes";
static const char __pyx_k_type_str[] = "type_str";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_ArrayType[] = "ArrayType";
static const char __pyx_k_BytesType[] = "BytesType";
static const char __pyx_k_FixedType[] = "FixedType";
static const char __pyx_k_FloatType[] = "FloatType";
static const char __pyx_k_HAVE_ZLIB[] = "HAVE_ZLIB";
static const char __pyx_k_HAVE_ZSTD[] = "HAVE_ZSTD";
static const char __pyx_k_LzmaCodec[] = "_LzmaCodec";
static const char __pyx_k_NamedType[] = "_NamedType";
static const char __pyx_k_Namespace[] = "Namespace '";
static const char __pyx_k_NullCodec[] = "_NullCodec";
static const char __pyx_k_Precision[] = ". Precision: ";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_Type_name[] = "Type name '";
static const char __pyx_k_UnionType[] = "UnionType";
static const char __pyx_k_ascending[] = "ascending";
static const char __pyx_k_avro_type[] = "avro_type";
static const char __pyx_k_boolean_2[] = "boolean";
static const char __pyx_k_byteorder[] = "byteorder";
static const char __pyx_k_cavro_pyx[] = "cavro.pyx";
static const char __pyx_k_copy_time[] = "copy_time";
static const char __pyx_k_decimal_2[] = "(decimal)";
static const char __pyx_k_dest_type[] = "dest_type";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_find_type[] = "find_type";
static const char __pyx_k_for_field[] = " for field ";
static const char __pyx_k_fullmatch[] = "fullmatch";
static const char __pyx_k_functools[] = "functools";
static const char __pyx_k_hexdigest[] = "hexdigest";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_item_type[] = "item_type";
static const char __pyx_k_lz4_frame[] = "lz4.frame";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_missing_2[] = "' missing";
static const char __pyx_k_namespace[] = "namespace";
static const char __pyx_k_new_types[] = "new_types";
static const char __pyx_k_other_val[] = "other_val";
static const char __pyx_k_out_state[] = "out_state";
static const char __pyx_k_precision[] = "precision";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_read_perf[] = "read_perf";
static const char __pyx_k_read_time[] = "read_time";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_schema_ob[] = "schema_ob";
static const char __pyx_k_serialize[] = "serialize";
static const char __pyx_k_signature[] = "__signature__";
static const char __pyx_k_threading[] = "threading";
static const char __pyx_k_timedelta[] = "timedelta";
static const char __pyx_k_timestamp[] = "timestamp";
static const char __pyx_k_type_name[] = "type_name";
static const char __pyx_k_wrap_type[] = "_wrap_type";
static const char __pyx_k_write_one[] = "write_one";
static const char __pyx_k_zstandard[] = "zstandard";
static const char __pyx_k_BATCH_SIZE[] = "BATCH_SIZE";
static const char __pyx_k_Bzip2Codec[] = "__Bzip2Codec";
static const char __pyx_k_DoubleType[] = "DoubleType";
static const char __pyx_k_EPOCH_DATE[] = "EPOCH_DATE";
static const char __pyx_k_FileReader[] = "FileReader";
static const char __pyx_k_FileWriter[] = "FileWriter";
static const char __pyx_k_HAVE_BZIP2[] = "HAVE_BZIP2";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_NO_DEFAULT[] = "NO_DEFAULT";
static const char __pyx_k_RAW_VALUES[] = "RAW_VALUES";
static const char __pyx_k_Rabin_copy[] = "Rabin.copy";
static const char __pyx_k_RecordType[] = "RecordType";
static const char __pyx_k_SLICE_TYPE[] = "_SLICE_TYPE";
static const char __pyx_k_StringType[] = "StringType";
static const char __pyx_k_TimeMicros[] = "TimeMicros";
static const char __pyx_k_TimeMillis[] = "TimeMillis";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_astimezone[] = "astimezone";
static const char __pyx_k_avro_codec[] = "avro.codec";
static const char __pyx_k_bit_length[] = "bit_length";
static const char __pyx_k_block_size[] = "block_size";
static const char __pyx_k_can_encode[] = "can_encode";
static const char __pyx_k_decompress[] = "decompress";
static const char __pyx_k_descending[] = "descending";
static const char __pyx_k_field_type[] = "_field_type";
static const char __pyx_k_fixed_uuid[] = "fixed(uuid)";
static const char __pyx_k_for_schema[] = "for_schema";
static const char __pyx_k_for_source[] = "for_source";
static const char __pyx_k_for_type_2[] = "for_type";
static const char __pyx_k_for_type_3[] = "_for_type";
static const char __pyx_k_from_bytes[] = "from_bytes";
static const char __pyx_k_get_schema[] = "get_schema";
static const char __pyx_k_match_args[] = "match_args";
static const char __pyx_k_maxunicode[] = "maxunicode";
static const char __pyx_k_namespaced[] = "namespaced";
static const char __pyx_k_orig_value[] = "orig_value";
static const char __pyx_k_parse_json[] = "parse_json";
static const char __pyx_k_perf_tests[] = "__perf_tests";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_setdefault[] = "setdefault";
static const char __pyx_k_total_nums[] = "total_nums";
static const char __pyx_k_underlying[] = "underlying";
static const char __pyx_k_value_type[] = "value_type";
static const char __pyx_k_walk_types[] = "walk_types";
static const char __pyx_k_write_many[] = "write_many";
static const char __pyx_k_write_perf[] = "write_perf";
static const char __pyx_k_write_time[] = "write_time";
static const char __pyx_k_DecimalType[] = "DecimalType";
static const char __pyx_k_EMPTY_ARRAY[] = "_EMPTY_ARRAY";
static const char __pyx_k_Enum_symbol[] = "Enum symbol '";
static const char __pyx_k_HAVE_SNAPPY[] = "HAVE_SNAPPY";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_InvalidName[] = "InvalidName";
static const char __pyx_k_LogicalType[] = "LogicalType";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_RecordField[] = "RecordField";
static const char __pyx_k_SnappyCodec[] = "_SnappyCodec";
static const char __pyx_k_UnknownType[] = "UnknownType";
static const char __pyx_k_annotations[] = "annotations";
static const char __pyx_k_avro_schema[] = "avro.schema";
static const char __pyx_k_binary_read[] = "binary_read";
static const char __pyx_k_bytes_codec[] = "bytes_codec";
static const char __pyx_k_check_value[] = "check_value";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_dataclasses[] = "dataclasses";
static const char __pyx_k_deserialize[] = "deserialize";
static const char __pyx_k_digest_size[] = "digest_size";
static const char __pyx_k_fingerprint[] = "fingerprint";
static const char __pyx_k_fixed_codec[] = "fixed_codec";
static const char __pyx_k_found_twice[] = "' found twice";
static const char __pyx_k_json_decode[] = "json_decode";
static const char __pyx_k_json_encode[] = "json_encode";
static const char __pyx_k_logicalType[] = "logicalType";
static const char __pyx_k_microsecond[] = "microsecond";
static const char __pyx_k_mro_entries[] = "__mro_entries__";
static const char __pyx_k_named_types[] = "named_types";
static const char __pyx_k_next_object[] = "next_object";
static const char __pyx_k_placeholder[] = "placeholder";
static const char __pyx_k_precision_2[] = ", precision: ";
static const char __pyx_k_read_marker[] = "_read_marker";
static const char __pyx_k_reader_type[] = "reader_type";
static const char __pyx_k_schema_path[] = "schema_path";
static const char __pyx_k_signature_2[] = "signature";
static const char __pyx_k_src_map_pxi[] = "src/map.pxi";
static const char __pyx_k_string_uuid[] = "string(uuid)";
static const char __pyx_k_time_micros[] = "time-micros";
static const char __pyx_k_time_millis[] = "time-millis";
static const char __pyx_k_unicodedata[] = "unicodedata";
static const char __pyx_k_unsafe_hash[] = "unsafe_hash";
static const char __pyx_k_writer_type[] = "writer_type";
static const char __pyx_k_DeflateCodec[] = "_DeflateCodec";
static const char __pyx_k_IntType_copy[] = "IntType.copy";
static const char __pyx_k_InvalidValue[] = "InvalidValue";
static const char __pyx_k_MapType_copy[] = "MapType.copy";
static const char __pyx_k_MemoryReader[] = "MemoryReader";
static const char __pyx_k_MemoryWriter[] = "MemoryWriter";
static const char __pyx_k_Options_diff[] = "Options.diff";
static const char __pyx_k_Rabin_digest[] = "Rabin.digest";
static const char __pyx_k_Rabin_update[] = "Rabin.update";
static const char __pyx_k_Record___dir[] = "Record.__dir__";
static const char __pyx_k_Sequence_str[] = "Sequence[str]";
static const char __pyx_k_Unknown_type[] = "Unknown type: ";
static const char __pyx_k_ValueAdapter[] = "ValueAdapter";
static const char __pyx_k_binary_write[] = "binary_write";
static const char __pyx_k_defaul_value[] = "defaul_value";
static const char __pyx_k_ensure_ascii[] = "ensure_ascii";
static const char __pyx_k_has_exponent[] = " has exponent ";
static const char __pyx_k_initial_size[] = "initial_size";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_is_not_valid[] = "' is not valid";
static const char __pyx_k_logical_name[] = "logical_name";
static const char __pyx_k_logical_type[] = "logical_type";
static const char __pyx_k_microseconds[] = "microseconds";
static const char __pyx_k_name_pattern[] = "name_pattern";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_src_enum_pxi[] = "src/enum.pxi";
static const char __pyx_k_src_null_pxi[] = "src/null.pxi";
static const char __pyx_k_src_type_pxi[] = "src/type.pxi";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_version_info[] = "version_info";
static const char __pyx_k_weakref_slot[] = "weakref_slot";
static const char __pyx_k_write_header[] = "write_header";
static const char __pyx_k_AvroType_copy[] = "AvroType.copy";
static const char __pyx_k_BoolType_copy[] = "BoolType.copy";
static const char __pyx_k_CannotPromote[] = "CannotPromote";
static const char __pyx_k_CanonicalForm[] = "_CanonicalForm";
static const char __pyx_k_Decimal_value[] = "Decimal value ";
static const char __pyx_k_Default_value[] = "Default value '";
static const char __pyx_k_DuplicateName[] = "DuplicateName";
static const char __pyx_k_EnumType_copy[] = "EnumType.copy";
static const char __pyx_k_FieldAccessor[] = "_FieldAccessor";
static const char __pyx_k_InvalidHasher[] = "InvalidHasher";
static const char __pyx_k_Invalid_value[] = "Invalid value ";
static const char __pyx_k_LOGICAL_TYPES[] = "LOGICAL_TYPES";
static const char __pyx_k_LongType_copy[] = "LongType.copy";
static const char __pyx_k_MISSING_VALUE[] = "MISSING_VALUE";
static const char __pyx_k_NullType_copy[] = "NullType.copy";
static const char __pyx_k_OverflowError[] = "OverflowError";
static const char __pyx_k_StopIteration[] = "StopIteration";
static const char __pyx_k_TYPES_BY_NAME[] = "TYPES_BY_NAME";
static const char __pyx_k_UUIDFixedType[] = "UUIDFixedType";
static const char __pyx_k_annotations_2[] = "__annotations__";
static const char __pyx_k_binary_decode[] = "binary_decode";
static const char __pyx_k_binary_encode[] = "binary_encode";
static const char __pyx_k_bytes_codec_2[] = ", bytes_codec=";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_convert_value[] = "_convert_value";
static const char __pyx_k_decompressobj[] = "decompressobj";
static const char __pyx_k_default_value[] = "default_value";
static const char __pyx_k_fixed_codec_2[] = ", fixed_codec=";
static const char __pyx_k_init_subclass[] = "__init_subclass__";
static const char __pyx_k_logical_types[] = "logical_types";
static const char __pyx_k_max_blocksize[] = "max_blocksize";
static const char __pyx_k_max_precision[] = "max_precision";
static const char __pyx_k_not_valid_for[] = "' not valid for ";
static const char __pyx_k_reader_schema[] = "reader_schema";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_resolved_type[] = "resolved_type";
static const char __pyx_k_src_array_pxi[] = "src/array.pxi";
static const char __pyx_k_src_error_pxi[] = "src/error.pxi";
static const char __pyx_k_src_rabin_pxi[] = "src/rabin.pxi";
static const char __pyx_k_src_union_pxi[] = "src/union.pxi";
static const char __pyx_k_writer_schema[] = "writer_schema";
static const char __pyx_k_ArrayType_copy[] = "ArrayType.copy";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_BytesType_copy[] = "BytesType.copy";
static const char __pyx_k_Cannot_promote[] = "Cannot promote ";
static const char __pyx_k_CavroException[] = "CavroException";
static const char __pyx_k_Date__for_type[] = "Date._for_type";
static const char __pyx_k_EPOCH_DT_NO_TZ[] = "EPOCH_DT_NO_TZ";
static const char __pyx_k_FixedType_copy[] = "FixedType.copy";
static const char __pyx_k_FloatType_copy[] = "FloatType.copy";
static const char __pyx_k_NAME_RE_STRICT[] = "NAME_RE_STRICT";
static const char __pyx_k_NotImplemented[] = "NotImplemented";
static const char __pyx_k_OptionsWrapper[] = "_OptionsWrapper";
static const char __pyx_k_Options_equals[] = "Options.equals";
static const char __pyx_k_PromoteToFloat[] = "PromoteToFloat";
static const char __pyx_k_Record__asdict[] = "Record._asdict";
static const char __pyx_k_ResolvedSchema[] = "ResolvedSchema";
static const char __pyx_k_SSIZE_TYPECODE[] = "SSIZE_TYPECODE";
static const char __pyx_k_UUIDStringType[] = "UUIDStringType";
static const char __pyx_k_UnionType_copy[] = "UnionType.copy";
static const char __pyx_k_ZStandardCodec[] = "_ZStandardCodec";
static const char __pyx_k_ZstdCompressor[] = "ZstdCompressor";
static const char __pyx_k_abacuscounter1[] = "abacuscounter1";
static const char __pyx_k_accepts_string[] = "accepts_string";
static const char __pyx_k_canonical_form[] = "canonical_form";
static const char __pyx_k_create_decimal[] = "create_decimal";
static const char __pyx_k_field_to_index[] = "_field_to_index";
static const char __pyx_k_getfullargspec[] = "getfullargspec";
static const char __pyx_k_make_dataclass[] = "make_dataclass";
static const char __pyx_k_micros_to_time[] = "_micros_to_time";
static const char __pyx_k_null_pad_fixed[] = "null_pad_fixed";
static const char __pyx_k_required_field[] = "required field '";
static const char __pyx_k_src_option_pxi[] = "src/option.pxi";
static const char __pyx_k_src_record_pxi[] = "src/record.pxi";
static const char __pyx_k_src_schema_pxi[] = "src/schema.pxi";
static const char __pyx_k_time_to_micros[] = "_time_to_micros";
static const char __pyx_k_truncate_fixed[] = "truncate_fixed";
static const char __pyx_k_unicode_errors[] = "unicode_errors";
static const char __pyx_k_value_adapters[] = "value_adapters";
static const char __pyx_k_AMBIGUOUS_TYPES[] = "AMBIGUOUS_TYPES";
static const char __pyx_k_Cannot_write_to[] = "Cannot write to '";
static const char __pyx_k_ContainerReader[] = "ContainerReader";
static const char __pyx_k_ContainerWriter[] = "ContainerWriter";
static const char __pyx_k_DEFAULT_OPTIONS[] = "DEFAULT_OPTIONS";
static const char __pyx_k_DataclassParams[] = "_DataclassParams";
static const char __pyx_k_Default_value_2[] = "Default value ";
static const char __pyx_k_DoubleType_copy[] = "DoubleType.copy";
static const char __pyx_k_NAME_RE_UNICODE[] = "NAME_RE_UNICODE";
static const char __pyx_k_OBJ_MAGIC_BYTES[] = "OBJ_MAGIC_BYTES";
static const char __pyx_k_Options_replace[] = "Options.replace";
static const char __pyx_k_PRIMITIVE_TYPES[] = "PRIMITIVE_TYPES";
static const char __pyx_k_PlaceholderType[] = "_PlaceholderType";
static const char __pyx_k_Rabin_hexdigest[] = "Rabin.hexdigest";
static const char __pyx_k_RecordType_copy[] = "RecordType.copy";
static const char __pyx_k_ResourceWarning[] = "ResourceWarning";
static const char __pyx_k_StringType_copy[] = "StringType.copy";
static const char __pyx_k_TimestampMicros[] = "TimestampMicros";
static const char __pyx_k_TimestampMillis[] = "TimestampMillis";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_collections_abc[] = "collections.abc";
static const char __pyx_k_convert_value_2[] = "convert_value";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_int_time_millis[] = "int(time-millis)";
static const char __pyx_k_logical_by_name[] = "logical_by_name";
static const char __pyx_k_logical_types_2[] = ", logical_types=";
static const char __pyx_k_parse_namespace[] = "parse_namespace";
static const char __pyx_k_perf_counter_ns[] = "perf_counter_ns";
static const char __pyx_k_perf_locals_add[] = "_perf.<locals>.add";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_src_logical_pxi[] = "src/logical.pxi";
static const char __pyx_k_test_readvarint[] = "test_readvarint";
static const char __pyx_k_underlying_name[] = "underlying_name";
static const char __pyx_k_underlying_type[] = "underlying_type";
static const char __pyx_k_Cannot_read_from[] = "Cannot read from '";
static const char __pyx_k_CodecUnavailable[] = "CodecUnavailable";
static const char __pyx_k_ExponentTooLarge[] = "ExponentTooLarge";
static const char __pyx_k_MappingProxyType[] = "MappingProxyType";
static const char __pyx_k_Max_precision_is[] = ", Max precision is ";
static const char __pyx_k_Schema_find_type[] = "Schema.find_type";
static const char __pyx_k_SignatureWrapper[] = "_SignatureWrapper";
static const char __pyx_k_Union_str_object[] = "Union[str, object]";
static const char __pyx_k_ZstdDecompressor[] = "ZstdDecompressor";
static const char __pyx_k_allow_error_type[] = "allow_error_type";
static const char __pyx_k_ascii_name_rules[] = "ascii_name_rules";
static const char __pyx_k_dataclass_fields[] = "__dataclass_fields__";
static const char __pyx_k_dataclass_params[] = "__dataclass_params__";
static const char __pyx_k_get_schema_extra[] = "_get_schema_extra";
static const char __pyx_k_invalid_for_enum[] = "' invalid for enum";
static const char __pyx_k_is_not_a_mapping[] = "' is not a mapping";
static const char __pyx_k_long_time_micros[] = "long(time-micros)";
static const char __pyx_k_null_pad_fixed_2[] = ", null_pad_fixed=";
static const char __pyx_k_test_readvarlong[] = "test_readvarlong";
static const char __pyx_k_test_zigzag_long[] = "test_zigzag_long";
static const char __pyx_k_test_zigzag_perf[] = "_test_zigzag_perf";
static const char __pyx_k_tests_locals_add[] = "_tests.<locals>.add";
static const char __pyx_k_timestamp_micros[] = "timestamp-micros";
static const char __pyx_k_timestamp_millis[] = "timestamp-millis";
static const char __pyx_k_truncate_fixed_2[] = ", truncate_fixed=";
static const char __pyx_k_underlying_types[] = "underlying_types";
static const char __pyx_k_unicode_errors_2[] = ", unicode_errors=";
static const char __pyx_k_union_decodes_to[] = "union_decodes_to";
static const char __pyx_k_A_Za_z__A_Za_z0_9[] = "[A-Za-z_][A-Za-z0-9_]*";
static const char __pyx_k_CustomLogicalType[] = "CustomLogicalType";
static const char __pyx_k_Expected_null_got[] = "Expected null, got ";
static const char __pyx_k_OBJ_FILE_METADATA[] = "OBJ_FILE_METADATA";
static const char __pyx_k_PromotingEnumType[] = "PromotingEnumType";
static const char __pyx_k_Schema__wrap_type[] = "Schema._wrap_type";
static const char __pyx_k_Schema_can_encode[] = "Schema.can_encode";
static const char __pyx_k_TYPE_TUPLE_ALWAYS[] = "TYPE_TUPLE_ALWAYS";
static const char __pyx_k_UnionDecodeOption[] = "UnionDecodeOption";
static const char __pyx_k_Unsupported_codec[] = "Unsupported codec: '";
static const char __pyx_k_class_inst_method[] = "_class_inst_method";
static const char __pyx_k_dict_str_AvroType[] = "dict[str, AvroType]";
static const char __pyx_k_inline_namespaces[] = "inline_namespaces";
static const char __pyx_k_pyx_unpickle_Date[] = "__pyx_unpickle_Date";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_reader_for_writer[] = "reader_for_writer";
static const char __pyx_k_src_container_pxi[] = "src/container.pxi";
static const char __pyx_k_tuple_LogicalType[] = "tuple[LogicalType]";
static const char __pyx_k_CannotPromoteError[] = "CannotPromoteError";
static const char __pyx_k_InvalidValue___str[] = "InvalidValue.__str__";
static const char __pyx_k_MapType_walk_types[] = "MapType.walk_types";
static const char __pyx_k_PERMISSIVE_OPTIONS[] = "PERMISSIVE_OPTIONS";
static const char __pyx_k_Schema_binary_read[] = "Schema.binary_read";
static const char __pyx_k_Schema_fingerprint[] = "Schema.fingerprint";
static const char __pyx_k_Schema_is_required[] = "Schema is required";
static const char __pyx_k_Schema_json_decode[] = "Schema.json_decode";
static const char __pyx_k_Schema_json_encode[] = "Schema.json_encode";
static const char __pyx_k_UUIDBase__for_type[] = "UUIDBase._for_type";
static const char __pyx_k_UnicodeEncodeError[] = "UnicodeEncodeError";
static const char __pyx_k_UnknownType___init[] = "UnknownType.__init__";
static const char __pyx_k_adapt_record_types[] = "adapt_record_types";
static const char __pyx_k_allow_empty_unions[] = "allow_empty_unions";
static const char __pyx_k_allow_error_type_2[] = ", allow_error_type=";
static const char __pyx_k_ascii_name_rules_2[] = ", ascii_name_rules=";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_check_exp_overflow[] = "check_exp_overflow";
static const char __pyx_k_clamp_int_overflow[] = "clamp_int_overflow";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_make_logical_types[] = "_make_logical_types";
static const char __pyx_k_not_valid_for_long[] = " not valid for long";
static const char __pyx_k_pyx_unpickle_Codec[] = "__pyx_unpickle_Codec";
static const char __pyx_k_pyx_unpickle_Rabin[] = "__pyx_unpickle_Rabin";
static const char __pyx_k_return_uuid_object[] = "return_uuid_object";
static const char __pyx_k_set_from_dataclass[] = "_set_from_dataclass";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_test_reader_read_n[] = "_test_reader_read_n";
static const char __pyx_k_union_decodes_to_2[] = ", union_decodes_to=";
static const char __pyx_k_with_logical_types[] = "with_logical_types";
static const char __pyx_k_AvroType_for_schema[] = "AvroType.for_schema";
static const char __pyx_k_AvroType_for_source[] = "AvroType.for_source";
static const char __pyx_k_AvroType_get_schema[] = "AvroType.get_schema";
static const char __pyx_k_AvroType_walk_types[] = "AvroType.walk_types";
static const char __pyx_k_CANONICAL_FORM_KEYS[] = "CANONICAL_FORM_KEYS";
static const char __pyx_k_Container_is_closed[] = "Container is closed";
static const char __pyx_k_InvalidValue___init[] = "InvalidValue.__init__";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_PromotingRecordType[] = "PromotingRecordType";
static const char __pyx_k_Schema_binary_write[] = "Schema.binary_write";
static const char __pyx_k_Unknown_hash_method[] = "Unknown hash method: ";
static const char __pyx_k_allow_nested_unions[] = "allow_nested_unions";
static const char __pyx_k_coerce_int_to_float[] = "coerce_int_to_float";
static const char __pyx_k_custom_decode_value[] = "custom_decode_value";
static const char __pyx_k_custom_encode_value[] = "custom_encode_value";
static const char __pyx_k_inline_namespaces_2[] = ", inline_namespaces=";
static const char __pyx_k_is_not_a_valid_name[] = "' is not a valid name";
static const char __pyx_k_naive_dt_assume_utc[] = "naive_dt_assume_utc";
static const char __pyx_k_not_in_enum_symbols[] = "' not in enum symbols";
static const char __pyx_k_pyx_unpickle_Record[] = "__pyx_unpickle_Record";
static const char __pyx_k_pyx_unpickle_Schema[] = "__pyx_unpickle_Schema";
static const char __pyx_k_test_reader_read_u8[] = "_test_reader_read_u8";
static const char __pyx_k_types_str_to_schema[] = "types_str_to_schema";
static const char __pyx_k_with_external_types[] = "with_external_types";
static const char __pyx_k_ArrayType_walk_types[] = "ArrayType.walk_types";
static const char __pyx_k_Date___reduce_cython[] = "Date.__reduce_cython__";
static const char __pyx_k_Duplicate_field_name[] = "Duplicate field name: ";
static const char __pyx_k_LogicalType_for_type[] = "LogicalType.for_type";
static const char __pyx_k_PromoteBytesToString[] = "PromoteBytesToString";
static const char __pyx_k_PromoteStringToBytes[] = "PromoteStringToBytes";
static const char __pyx_k_Schema_binary_decode[] = "Schema.binary_decode";
static const char __pyx_k_Schema_binary_encode[] = "Schema.binary_encode";
static const char __pyx_k_TYPE_TUPLE_IF_RECORD[] = "TYPE_TUPLE_IF_RECORD";
static const char __pyx_k_TimeMicros__for_type[] = "TimeMicros._for_type";
static const char __pyx_k_TimeMillis__for_type[] = "TimeMillis._for_type";
static const char __pyx_k_UnionType_get_schema[] = "UnionType.get_schema";
static const char __pyx_k_UnionType_walk_types[] = "UnionType.walk_types";
static const char __pyx_k_adapt_record_types_2[] = ", adapt_record_types=";
static const char __pyx_k_allow_empty_unions_2[] = ", allow_empty_unions=";
static const char __pyx_k_allow_tuple_notation[] = "allow_tuple_notation";
static const char __pyx_k_cannot_be_adapted_to[] = " cannot be adapted to ";
static const char __pyx_k_clamp_float_overflow[] = "clamp_float_overflow";
static const char __pyx_k_clamp_int_overflow_2[] = ", clamp_int_overflow=";
static const char __pyx_k_coerce_values_to_int[] = "coerce_values_to_int";
static const char __pyx_k_coerce_values_to_str[] = "coerce_values_to_str";
static const char __pyx_k_pyx_unpickle_IntType[] = "__pyx_unpickle_IntType";
static const char __pyx_k_pyx_unpickle_MapType[] = "__pyx_unpickle_MapType";
static const char __pyx_k_pyx_unpickle_Options[] = "__pyx_unpickle_Options";
static const char __pyx_k_pyx_unpickle__Reader[] = "__pyx_unpickle__Reader";
static const char __pyx_k_pyx_unpickle__Writer[] = "__pyx_unpickle__Writer";
static const char __pyx_k_return_uuid_object_2[] = ", return_uuid_object=";
static const char __pyx_k_src_string_types_pxi[] = "src/string_types.pxi";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_Codec___reduce_cython[] = "Codec.__reduce_cython__";
static const char __pyx_k_ContainerWriter_close[] = "ContainerWriter.close";
static const char __pyx_k_ContainerWriter_flush[] = "ContainerWriter.flush";
static const char __pyx_k_DecimalType__for_type[] = "DecimalType._for_type";
static const char __pyx_k_Invalid_shape_in_axis[] = "Invalid shape in axis ";
static const char __pyx_k_MapType_convert_value[] = "MapType.convert_value";
static const char __pyx_k_Rabin___reduce_cython[] = "Rabin.__reduce_cython__";
static const char __pyx_k_RecordType_walk_types[] = "RecordType.walk_types";
static const char __pyx_k_allow_nested_unions_2[] = ", allow_nested_unions=";
static const char __pyx_k_coerce_int_to_float_2[] = ", coerce_int_to_float=";
static const char __pyx_k_container_fill_blocks[] = "container_fill_blocks";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_init___locals_genexpr[] = "__init__.<locals>.genexpr";
static const char __pyx_k_long_timestamp_micros[] = "long(timestamp-micros)";
static const char __pyx_k_long_timestamp_millis[] = "long(timestamp-millis)";
static const char __pyx_k_naive_dt_assume_utc_2[] = ", naive_dt_assume_utc=";
static const char __pyx_k_pyx_unpickle_AvroType[] = "__pyx_unpickle_AvroType";
static const char __pyx_k_pyx_unpickle_BoolType[] = "__pyx_unpickle_BoolType";
static const char __pyx_k_pyx_unpickle_EnumType[] = "__pyx_unpickle_EnumType";
static const char __pyx_k_pyx_unpickle_LongType[] = "__pyx_unpickle_LongType";
static const char __pyx_k_pyx_unpickle_NullType[] = "__pyx_unpickle_NullType";
static const char __pyx_k_pyx_unpickle_UUIDBase[] = "__pyx_unpickle_UUIDBase";
static const char __pyx_k_src_numeric_types_pxi[] = "src/numeric_types.pxi";
static const char __pyx_k_types_str_to_schema_2[] = ", types_str_to_schema=";
static const char __pyx_k_AvroType_convert_value[] = "AvroType.convert_value";
static const char __pyx_k_Cannot_index_with_type[] = "Cannot index with type '";
static const char __pyx_k_ContainerWriter___exit[] = "ContainerWriter.__exit__";
static const char __pyx_k_Date___setstate_cython[] = "Date.__setstate_cython__";
static const char __pyx_k_IntType__convert_value[] = "IntType._convert_value";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_Reader___reduce_cython[] = "_Reader.__reduce_cython__";
static const char __pyx_k_RecordField_get_schema[] = "RecordField.get_schema";
static const char __pyx_k_Record___reduce_cython[] = "Record.__reduce_cython__";
static const char __pyx_k_Schema___reduce_cython[] = "Schema.__reduce_cython__";
static const char __pyx_k_Unknown_value_for_enum[] = "Unknown value for enum";
static const char __pyx_k_Writer___reduce_cython[] = "_Writer.__reduce_cython__";
static const char __pyx_k_allow_tuple_notation_2[] = ", allow_tuple_notation=";
static const char __pyx_k_clamp_float_overflow_2[] = ", clamp_float_overflow=";
static const char __pyx_k_coerce_values_to_float[] = "coerce_values_to_float";
static const char __pyx_k_coerce_values_to_int_2[] = ", coerce_values_to_int=";
static const char __pyx_k_coerce_values_to_str_2[] = ", coerce_values_to_str=";
static const char __pyx_k_expand_types_in_schema[] = "expand_types_in_schema";
static const char __pyx_k_float_out_of_range_inf[] = "float_out_of_range_inf";
static const char __pyx_k_is_not_valid_for_field[] = " is not valid for field: ";
static const char __pyx_k_is_required_for_record[] = " is required for record ";
static const char __pyx_k_pyx_unpickle_ArrayType[] = "__pyx_unpickle_ArrayType";
static const char __pyx_k_pyx_unpickle_BytesType[] = "__pyx_unpickle_BytesType";
static const char __pyx_k_pyx_unpickle_FixedType[] = "__pyx_unpickle_FixedType";
static const char __pyx_k_pyx_unpickle_FloatType[] = "__pyx_unpickle_FloatType";
static const char __pyx_k_pyx_unpickle_UnionType[] = "__pyx_unpickle_UnionType";
static const char __pyx_k_pyx_unpickle__Lz4Codec[] = "__pyx_unpickle__Lz4Codec";
static const char __pyx_k_pyx_unpickle__Sentinel[] = "__pyx_unpickle__Sentinel";
static const char __pyx_k_record_can_encode_dict[] = "record_can_encode_dict";
static const char __pyx_k_record_decodes_to_dict[] = "record_decodes_to_dict";
static const char __pyx_k_ArrayType_convert_value[] = "ArrayType.convert_value";
static const char __pyx_k_AvroType__convert_value[] = "AvroType._convert_value";
static const char __pyx_k_BoolType__convert_value[] = "BoolType._convert_value";
static const char __pyx_k_Codec___setstate_cython[] = "Codec.__setstate_cython__";
static const char __pyx_k_ContainerWriter___enter[] = "ContainerWriter.__enter__";
static const char __pyx_k_EnumType__convert_value[] = "EnumType._convert_value";
static const char __pyx_k_IntType___reduce_cython[] = "IntType.__reduce_cython__";
static const char __pyx_k_LongType__convert_value[] = "LongType._convert_value";
static const char __pyx_k_MapType___reduce_cython[] = "MapType.__reduce_cython__";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_NullType__convert_value[] = "NullType._convert_value";
static const char __pyx_k_Options___reduce_cython[] = "Options.__reduce_cython__";
static const char __pyx_k_Rabin___setstate_cython[] = "Rabin.__setstate_cython__";
static const char __pyx_k_SignatureWrapper___init[] = "_SignatureWrapper.__init__";
static const char __pyx_k_TYPE_TUPLE_IF_AMBIGUOUS[] = "TYPE_TUPLE_IF_AMBIGUOUS";
static const char __pyx_k_UUIDFixedType__for_type[] = "UUIDFixedType._for_type";
static const char __pyx_k_UnionType_convert_value[] = "UnionType.convert_value";
static const char __pyx_k_class_inst_method___get[] = "_class_inst_method.__get__";
static const char __pyx_k_container_fill_blocks_2[] = ", container_fill_blocks=";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_enforce_type_name_rules[] = "enforce_type_name_rules";
static const char __pyx_k_pyx_unpickle_DoubleType[] = "__pyx_unpickle_DoubleType";
static const char __pyx_k_pyx_unpickle_FileReader[] = "__pyx_unpickle_FileReader";
static const char __pyx_k_pyx_unpickle_FileWriter[] = "__pyx_unpickle_FileWriter";
static const char __pyx_k_pyx_unpickle_RecordType[] = "__pyx_unpickle_RecordType";
static const char __pyx_k_pyx_unpickle_StringType[] = "__pyx_unpickle_StringType";
static const char __pyx_k_pyx_unpickle_TimeMicros[] = "__pyx_unpickle_TimeMicros";
static const char __pyx_k_pyx_unpickle_TimeMillis[] = "__pyx_unpickle_TimeMillis";
static const char __pyx_k_pyx_unpickle__LzmaCodec[] = "__pyx_unpickle__LzmaCodec";
static const char __pyx_k_pyx_unpickle__NamedType[] = "__pyx_unpickle__NamedType";
static const char __pyx_k_pyx_unpickle__NullCodec[] = "__pyx_unpickle__NullCodec";
static const char __pyx_k_record_values_type_hint[] = "record_values_type_hint";
static const char __pyx_k_src_tests_test_perf_pxi[] = "src/tests/test_perf.pxi";
static const char __pyx_k_AvroType___reduce_cython[] = "AvroType.__reduce_cython__";
static const char __pyx_k_BoolType___reduce_cython[] = "BoolType.__reduce_cython__";
static const char __pyx_k_BytesType__convert_value[] = "BytesType._convert_value";
static const char __pyx_k_EnumType___reduce_cython[] = "EnumType.__reduce_cython__";
static const char __pyx_k_FixedType__convert_value[] = "FixedType._convert_value";
static const char __pyx_k_FloatType__convert_value[] = "FloatType._convert_value";
static const char __pyx_k_LongType___reduce_cython[] = "LongType.__reduce_cython__";
static const char __pyx_k_Lz4Codec___reduce_cython[] = "_Lz4Codec.__reduce_cython__";
static const char __pyx_k_Marker_must_be_bytes_got[] = "Marker must be bytes, got: ";
static const char __pyx_k_NullType___reduce_cython[] = "NullType.__reduce_cython__";
static const char __pyx_k_Reader___setstate_cython[] = "_Reader.__setstate_cython__";
static const char __pyx_k_Record___setstate_cython[] = "Record.__setstate_cython__";
static const char __pyx_k_Schema___setstate_cython[] = "Schema.__setstate_cython__";
static const char __pyx_k_Schema_reader_for_writer[] = "Schema.reader_for_writer";
static const char __pyx_k_Sentinel___reduce_cython[] = "_Sentinel.__reduce_cython__";
static const char __pyx_k_UUIDBase___reduce_cython[] = "UUIDBase.__reduce_cython__";
static const char __pyx_k_UUIDStringType__for_type[] = "UUIDStringType._for_type";
static const char __pyx_k_Writer___setstate_cython[] = "_Writer.__setstate_cython__";
static const char __pyx_k_bytes_default_value_utf8[] = "bytes_default_value_utf8";
static const char __pyx_k_class_inst_method___init[] = "_class_inst_method.__init__";
static const char __pyx_k_coerce_values_to_boolean[] = "coerce_values_to_boolean";
static const char __pyx_k_coerce_values_to_float_2[] = ", coerce_values_to_float=";
static const char __pyx_k_date_type_accepts_string[] = "date_type_accepts_string";
static const char __pyx_k_does_not_implement_clone[] = " does not implement clone";
static const char __pyx_k_expand_types_in_schema_2[] = ", expand_types_in_schema=";
static const char __pyx_k_externally_defined_types[] = "externally_defined_types";
static const char __pyx_k_float_out_of_range_inf_2[] = ", float_out_of_range_inf=";
static const char __pyx_k_is_not_allowed_as_a_name[] = "' is not allowed as a name";
static const char __pyx_k_pyx_recursive_repr_guard[] = "__pyx_recursive_repr_guard";
static const char __pyx_k_pyx_unpickle_DecimalType[] = "__pyx_unpickle_DecimalType";
static const char __pyx_k_pyx_unpickle_LogicalType[] = "__pyx_unpickle_LogicalType";
static const char __pyx_k_pyx_unpickle_RecordField[] = "__pyx_unpickle_RecordField";
static const char __pyx_k_raise_on_invalid_logical[] = "raise_on_invalid_logical";
static const char __pyx_k_record_can_encode_dict_2[] = ", record_can_encode_dict=";
static const char __pyx_k_record_decodes_to_dict_2[] = ", record_decodes_to_dict=";
static const char __pyx_k_ArrayType___reduce_cython[] = "ArrayType.__reduce_cython__";
static const char __pyx_k_BytesType___reduce_cython[] = "BytesType.__reduce_cython__";
static const char __pyx_k_CannotPromoteError___init[] = "CannotPromoteError.__init__";
static const char __pyx_k_ContainerWriter_write_one[] = "ContainerWriter.write_one";
static const char __pyx_k_Dimension_d_is_not_direct[] = "Dimension %d is not direct";
static const char __pyx_k_DoubleType__convert_value[] = "DoubleType._convert_value";
static const char __pyx_k_FixedType___reduce_cython[] = "FixedType.__reduce_cython__";
static const char __pyx_k_FloatType___reduce_cython[] = "FloatType.__reduce_cython__";
static const char __pyx_k_IntType___setstate_cython[] = "IntType.__setstate_cython__";
static const char __pyx_k_IntType__get_schema_extra[] = "IntType._get_schema_extra";
static const char __pyx_k_LzmaCodec___reduce_cython[] = "_LzmaCodec.__reduce_cython__";
static const char __pyx_k_MapType___setstate_cython[] = "MapType.__setstate_cython__";
static const char __pyx_k_MapType__get_schema_extra[] = "MapType._get_schema_extra";
static const char __pyx_k_NamedType___reduce_cython[] = "_NamedType.__reduce_cython__";
static const char __pyx_k_NullCodec___reduce_cython[] = "_NullCodec.__reduce_cython__";
static const char __pyx_k_Options___setstate_cython[] = "Options.__setstate_cython__";
static const char __pyx_k_RecordType__convert_value[] = "RecordType._convert_value";
static const char __pyx_k_StringType__convert_value[] = "StringType._convert_value";
static const char __pyx_k_TimestampMicros__for_type[] = "TimestampMicros._for_type";
static const char __pyx_k_TimestampMillis__for_type[] = "TimestampMillis._for_type";
static const char __pyx_k_UnionType___reduce_cython[] = "UnionType.__reduce_cython__";
static const char __pyx_k_does_not_implement_read_n[] = " does not implement read_n";
static const char __pyx_k_enforce_type_name_rules_2[] = ", enforce_type_name_rules=";
static const char __pyx_k_pyx_unpickle_MemoryReader[] = "__pyx_unpickle_MemoryReader";
static const char __pyx_k_pyx_unpickle_ValueAdapter[] = "__pyx_unpickle_ValueAdapter";
static const char __pyx_k_pyx_unpickle__SnappyCodec[] = "__pyx_unpickle__SnappyCodec";
static const char __pyx_k_pyx_unpickle___Bzip2Codec[] = "__pyx_unpickle___Bzip2Codec";
static const char __pyx_k_record_allow_extra_fields[] = "record_allow_extra_fields";
static const char __pyx_k_record_values_type_hint_2[] = ", record_values_type_hint=";
static const char __pyx_k_src_tests_test_buffer_pxi[] = "src/tests/test_buffer.pxi";
static const char __pyx_k_src_tests_test_zigzag_pxi[] = "src/tests/test_zigzag.pxi";
static const char __pyx_k_AvroType___setstate_cython[] = "AvroType.__setstate_cython__";
static const char __pyx_k_AvroType__get_schema_extra[] = "AvroType._get_schema_extra";
static const char __pyx_k_BoolType___setstate_cython[] = "BoolType.__setstate_cython__";
static const char __pyx_k_BoolType__get_schema_extra[] = "BoolType._get_schema_extra";
static const char __pyx_k_Bzip2Codec___reduce_cython[] = "__Bzip2Codec.__reduce_cython__";
static const char __pyx_k_ContainerWriter_write_many[] = "ContainerWriter.write_many";
static const char __pyx_k_DoubleType___reduce_cython[] = "DoubleType.__reduce_cython__";
static const char __pyx_k_EnumType___setstate_cython[] = "EnumType.__setstate_cython__";
static const char __pyx_k_EnumType__get_schema_extra[] = "EnumType._get_schema_extra";
static const char __pyx_k_Expected_datetime_date_got[] = "Expected datetime.date, got ";
static const char __pyx_k_FileReader___reduce_cython[] = "FileReader.__reduce_cython__";
static const char __pyx_k_FileWriter___reduce_cython[] = "FileWriter.__reduce_cython__";
static const char __pyx_k_Index_out_of_bounds_axis_d[] = "Index out of bounds (axis %d)";
static const char __pyx_k_LongType___setstate_cython[] = "LongType.__setstate_cython__";
static const char __pyx_k_LongType__get_schema_extra[] = "LongType._get_schema_extra";
static const char __pyx_k_Lz4Codec___setstate_cython[] = "_Lz4Codec.__setstate_cython__";
static const char __pyx_k_NullType___setstate_cython[] = "NullType.__setstate_cython__";
static const char __pyx_k_NullType__get_schema_extra[] = "NullType._get_schema_extra";
static const char __pyx_k_Options_with_logical_types[] = "Options.with_logical_types";
static const char __pyx_k_RecordType___reduce_cython[] = "RecordType.__reduce_cython__";
static const char __pyx_k_Schema__make_logical_types[] = "Schema._make_logical_types";
static const char __pyx_k_Sentinel___setstate_cython[] = "_Sentinel.__setstate_cython__";
static const char __pyx_k_StringType___reduce_cython[] = "StringType.__reduce_cython__";
static const char __pyx_k_TimeMicros___reduce_cython[] = "TimeMicros.__reduce_cython__";
static const char __pyx_k_TimeMillis___reduce_cython[] = "TimeMillis.__reduce_cython__";
static const char __pyx_k_UUIDBase___setstate_cython[] = "UUIDBase.__setstate_cython__";
static const char __pyx_k_allow_aliases_to_be_string[] = "allow_aliases_to_be_string";
static const char __pyx_k_allow_leading_dot_in_names[] = "allow_leading_dot_in_names";
static const char __pyx_k_bytes_default_value_utf8_2[] = ", bytes_default_value_utf8=";
static const char __pyx_k_canonical_form_repeat_enum[] = "canonical_form_repeat_enum";
static const char __pyx_k_coerce_values_to_boolean_2[] = ", coerce_values_to_boolean=";
static const char __pyx_k_date_type_accepts_string_2[] = ", date_type_accepts_string=";
static const char __pyx_k_decimal_check_exp_overflow[] = "decimal_check_exp_overflow";
static const char __pyx_k_does_not_implement_read_u8[] = " does not implement read_u8";
static const char __pyx_k_does_not_implement_write_n[] = " does not implement write_n";
static const char __pyx_k_externally_defined_types_2[] = ", externally_defined_types=";
static const char __pyx_k_fingerprint_returns_digest[] = "fingerprint_returns_digest";
static const char __pyx_k_missing_values_can_be_null[] = "missing_values_can_be_null";
static const char __pyx_k_pyx_unpickle_CannotPromote[] = "__pyx_unpickle_CannotPromote";
static const char __pyx_k_pyx_unpickle_UUIDFixedType[] = "__pyx_unpickle_UUIDFixedType";
static const char __pyx_k_pyx_unpickle__DeflateCodec[] = "__pyx_unpickle__DeflateCodec";
static const char __pyx_k_raise_on_invalid_logical_2[] = ", raise_on_invalid_logical=";
static const char __pyx_k_record_encode_use_defaults[] = "record_encode_use_defaults";
static const char __pyx_k_ArrayType___setstate_cython[] = "ArrayType.__setstate_cython__";
static const char __pyx_k_ArrayType__get_schema_extra[] = "ArrayType._get_schema_extra";
static const char __pyx_k_BytesType___setstate_cython[] = "BytesType.__setstate_cython__";
static const char __pyx_k_BytesType__get_schema_extra[] = "BytesType._get_schema_extra";
static const char __pyx_k_ContainerReader_next_object[] = "ContainerReader.next_object";
static const char __pyx_k_DecimalType___reduce_cython[] = "DecimalType.__reduce_cython__";
static const char __pyx_k_Enum_symbols_must_be_unique[] = "Enum symbols must be unique. '";
static const char __pyx_k_FixedType___setstate_cython[] = "FixedType.__setstate_cython__";
static const char __pyx_k_FixedType__get_schema_extra[] = "FixedType._get_schema_extra";
static const char __pyx_k_FloatType___setstate_cython[] = "FloatType.__setstate_cython__";
static const char __pyx_k_FloatType__get_schema_extra[] = "FloatType._get_schema_extra";
static const char __pyx_k_LogicalType___reduce_cython[] = "LogicalType.__reduce_cython__";
static const char __pyx_k_LzmaCodec___setstate_cython[] = "_LzmaCodec.__setstate_cython__";
static const char __pyx_k_NamedType___setstate_cython[] = "_NamedType.__setstate_cython__";
static const char __pyx_k_NamedType__get_schema_extra[] = "_NamedType._get_schema_extra";
static const char __pyx_k_NullCodec___setstate_cython[] = "_NullCodec.__setstate_cython__";
static const char __pyx_k_Options_with_external_types[] = "Options.with_external_types";
static const char __pyx_k_RecordField___reduce_cython[] = "RecordField.__reduce_cython__";
static const char __pyx_k_SnappyCodec___reduce_cython[] = "_SnappyCodec.__reduce_cython__";
static const char __pyx_k_Step_may_not_be_zero_axis_d[] = "Step may not be zero (axis %d)";
static const char __pyx_k_UnionType___setstate_cython[] = "UnionType.__setstate_cython__";
static const char __pyx_k_allow_duplicate_union_types[] = "allow_duplicate_union_types";
static const char __pyx_k_allow_false_values_for_null[] = "allow_false_values_for_null";
static const char __pyx_k_canonical_form_repeat_fixed[] = "canonical_form_repeat_fixed";
static const char __pyx_k_does_not_implement_write_u8[] = " does not implement write_u8";
static const char __pyx_k_enum_symbols_must_be_unique[] = "enum_symbols_must_be_unique";
static const char __pyx_k_is_not_valid_for_a_union_of[] = " is not valid for a union of ";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_pyx_unpickle_PromoteToFloat[] = "__pyx_unpickle_PromoteToFloat";
static const char __pyx_k_pyx_unpickle_ResolvedSchema[] = "__pyx_unpickle_ResolvedSchema";
static const char __pyx_k_pyx_unpickle_UUIDStringType[] = "__pyx_unpickle_UUIDStringType";
static const char __pyx_k_pyx_unpickle__CanonicalForm[] = "__pyx_unpickle__CanonicalForm";
static const char __pyx_k_pyx_unpickle__FieldAccessor[] = "__pyx_unpickle__FieldAccessor";
static const char __pyx_k_record_allow_extra_fields_2[] = ", record_allow_extra_fields=";
static const char __pyx_k_Bzip2Codec___setstate_cython[] = "__Bzip2Codec.__setstate_cython__";
static const char __pyx_k_ContainerReader__read_marker[] = "ContainerReader._read_marker";
static const char __pyx_k_DeflateCodec___reduce_cython[] = "_DeflateCodec.__reduce_cython__";
static const char __pyx_k_DoubleType___setstate_cython[] = "DoubleType.__setstate_cython__";
static const char __pyx_k_DoubleType__get_schema_extra[] = "DoubleType._get_schema_extra";
static const char __pyx_k_FileReader___setstate_cython[] = "FileReader.__setstate_cython__";
static const char __pyx_k_FileWriter___setstate_cython[] = "FileWriter.__setstate_cython__";
static const char __pyx_k_Invalid_file_header_expected[] = "Invalid file header, expected: ";
static const char __pyx_k_MemoryReader___reduce_cython[] = "MemoryReader.__reduce_cython__";
static const char __pyx_k_MemoryWriter___reduce_cython[] = "MemoryWriter.__reduce_cython__";
static const char __pyx_k_RecordType___setstate_cython[] = "RecordType.__setstate_cython__";
static const char __pyx_k_RecordType__get_schema_extra[] = "RecordType._get_schema_extra";
static const char __pyx_k_StringType___setstate_cython[] = "StringType.__setstate_cython__";
static const char __pyx_k_StringType__get_schema_extra[] = "StringType._get_schema_extra";
static const char __pyx_k_TimeMicros___setstate_cython[] = "TimeMicros.__setstate_cython__";
static const char __pyx_k_TimeMillis___setstate_cython[] = "TimeMillis.__setstate_cython__";
static const char __pyx_k_ValueAdapter___reduce_cython[] = "ValueAdapter.__reduce_cython__";
static const char __pyx_k_allow_aliases_to_be_string_2[] = ", allow_aliases_to_be_string=";
static const char __pyx_k_allow_invalid_default_values[] = "allow_invalid_default_values";
static const char __pyx_k_allow_leading_dot_in_names_2[] = ", allow_leading_dot_in_names=";
static const char __pyx_k_alternate_timestamp_encoding[] = "alternate_timestamp_encoding";
static const char __pyx_k_canonical_form_repeat_enum_2[] = ", canonical_form_repeat_enum=";
static const char __pyx_k_decimal_check_exp_overflow_2[] = ", decimal_check_exp_overflow=";
static const char __pyx_k_dict_str_tuple_object_object[] = "dict[str, tuple[object, object]]";
static const char __pyx_k_enforce_namespace_name_rules[] = "enforce_namespace_name_rules";
static const char __pyx_k_fingerprint_returns_digest_2[] = "(fingerprint_returns_digest=";
static const char __pyx_k_is_not_valid_for_bytes_field[] = " is not valid for bytes field: ";
static const char __pyx_k_is_not_valid_for_fixed_field[] = " is not valid for fixed field: ";
static const char __pyx_k_missing_values_can_be_null_2[] = ", missing_values_can_be_null=";
static const char __pyx_k_not_compatible_with_NullType[] = " not compatible with NullType";
static const char __pyx_k_pyx_unpickle_ContainerReader[] = "__pyx_unpickle_ContainerReader";
static const char __pyx_k_pyx_unpickle_TimestampMicros[] = "__pyx_unpickle_TimestampMicros";
static const char __pyx_k_pyx_unpickle_TimestampMillis[] = "__pyx_unpickle_TimestampMillis";
static const char __pyx_k_pyx_unpickle__ZStandardCodec[] = "__pyx_unpickle__ZStandardCodec";
static const char __pyx_k_record_encode_use_defaults_2[] = ", record_encode_use_defaults=";
static const char __pyx_k_record_fields_must_be_unique[] = "record_fields_must_be_unique";
static const char __pyx_k_tests_locals_test_readvarint[] = "_tests.<locals>.test_readvarint";
static const char __pyx_k_union_json_encodes_type_name[] = "union_json_encodes_type_name";
static const char __pyx_k_CannotPromote___reduce_cython[] = "CannotPromote.__reduce_cython__";
static const char __pyx_k_CanonicalForm___reduce_cython[] = "_CanonicalForm.__reduce_cython__";
static const char __pyx_k_DecimalType___setstate_cython[] = "DecimalType.__setstate_cython__";
static const char __pyx_k_FieldAccessor___reduce_cython[] = "_FieldAccessor.__reduce_cython__";
static const char __pyx_k_LogicalType___setstate_cython[] = "LogicalType.__setstate_cython__";
static const char __pyx_k_RecordField___setstate_cython[] = "RecordField.__setstate_cython__";
static const char __pyx_k_SnappyCodec___setstate_cython[] = "_SnappyCodec.__setstate_cython__";
static const char __pyx_k_UUIDFixedType___reduce_cython[] = "UUIDFixedType.__reduce_cython__";
static const char __pyx_k_allow_duplicate_union_types_2[] = ", allow_duplicate_union_types=";
static const char __pyx_k_allow_false_values_for_null_2[] = ", allow_false_values_for_null=";
static const char __pyx_k_canonical_form_repeat_fixed_2[] = ", canonical_form_repeat_fixed=";
static const char __pyx_k_defer_schema_promotion_errors[] = "defer_schema_promotion_errors";
static const char __pyx_k_does_not_implement_read_block[] = " does not implement read_block";
static const char __pyx_k_does_not_implement_read_bytes[] = " does not implement read_bytes";
static const char __pyx_k_enum_symbols_must_be_unique_2[] = ", enum_symbols_must_be_unique=";
static const char __pyx_k_invalid_value_include_map_key[] = "invalid_value_include_map_key";
static const char __pyx_k_perf_locals__test_zigzag_perf[] = "_perf.<locals>._test_zigzag_perf";
static const char __pyx_k_pyx_unpickle__PlaceholderType[] = "__pyx_unpickle__PlaceholderType";
static const char __pyx_k_tests_locals_test_readvarlong[] = "_tests.<locals>.test_readvarlong";
static const char __pyx_k_tests_locals_test_zigzag_long[] = "_tests.<locals>.test_zigzag_long";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_A_schema_cannot_be_promoted_to[] = "\n    A schema cannot be promoted to another schema. (reader/writer schema promotion)\n\n    Attributes:\n     * `reader_type`: The schema type of the reader\n     * `writer_type`: The schema type of the writer\n     * `extra`: An optional extra message\n    ";
static const char __pyx_k_A_value_is_invalid_for_a_given[] = "\n    A value is invalid for a given avro type.\n        \n    Attributes:\n     * `value`: The value that caused the error\n     * `dest_type`: The schema type that the value was being converted to\n     * `schema_path`: A sequence of identifiers (field names etc) to help locate the value that caused the error\n    ";
static const char __pyx_k_As_of_cython_3_0_2_dataclasses[] = "\n    As of cython 3.0.2, dataclasses do not have a class-level signature\n    (unlike pure python classes/dataclasses)\n\n    This wrapper allows us to implement a signature on a frozen dataclass, meaning\n    that `inspect.signature(MyDataclass)` will work as expected.\n\n    It does this by creating a temporary pure-python version of the dataclass, and then\n    copying the signature from that.\n    ";
static const char __pyx_k_DeflateCodec___setstate_cython[] = "_DeflateCodec.__setstate_cython__";
static const char __pyx_k_Expected_datetime_datetime_got[] = "Expected datetime.datetime, got ";
static const char __pyx_k_Invalid_length_for_fixed_field[] = "Invalid length for fixed field: ";
static const char __pyx_k_MemoryReader___setstate_cython[] = "MemoryReader.__setstate_cython__";
static const char __pyx_k_MemoryWriter___setstate_cython[] = "MemoryWriter.__setstate_cython__";
static const char __pyx_k_PlaceholderType__convert_value[] = "_PlaceholderType._convert_value";
static const char __pyx_k_PromoteToFloat___reduce_cython[] = "PromoteToFloat.__reduce_cython__";
static const char __pyx_k_ResolvedSchema___reduce_cython[] = "ResolvedSchema.__reduce_cython__";
static const char __pyx_k_TYPE_TUPLE_IF_RECORD_AMBIGUOUS[] = "TYPE_TUPLE_IF_RECORD_AMBIGUOUS";
static const char __pyx_k_UUIDStringType___reduce_cython[] = "UUIDStringType.__reduce_cython__";
static const char __pyx_k_ValueAdapter___setstate_cython[] = "ValueAdapter.__setstate_cython__";
static const char __pyx_k_ZStandardCodec___reduce_cython[] = "_ZStandardCodec.__reduce_cython__";
static const char __pyx_k_allow_invalid_default_values_2[] = ", allow_invalid_default_values=";
static const char __pyx_k_allow_primitive_name_collision[] = "allow_primitive_name_collision";
static const char __pyx_k_allow_union_default_any_member[] = "allow_union_default_any_member";
static const char __pyx_k_does_not_implement__get_schema[] = " does not implement _get_schema_extra";
static const char __pyx_k_does_not_implement_write_block[] = " does not implement write_block";
static const char __pyx_k_enforce_enum_symbol_name_rules[] = "enforce_enum_symbol_name_rules";
static const char __pyx_k_enforce_namespace_name_rules_2[] = ", enforce_namespace_name_rules=";
static const char __pyx_k_is_not_a_valid_value_for_array[] = "' is not a valid value for array";
static const char __pyx_k_perf_locals_add_locals_genexpr[] = "_perf.<locals>.add.<locals>.genexpr";
static const char __pyx_k_pyx_unpickle_CustomLogicalType[] = "__pyx_unpickle_CustomLogicalType";
static const char __pyx_k_pyx_unpickle_PromoteBytesToStr[] = "__pyx_unpickle_PromoteBytesToString";
static const char __pyx_k_pyx_unpickle_PromoteStringToBy[] = "__pyx_unpickle_PromoteStringToBytes";
static const char __pyx_k_pyx_unpickle_PromotingEnumType[] = "__pyx_unpickle_PromotingEnumType";
static const char __pyx_k_pyx_unpickle_PromotingRecordTy[] = "__pyx_unpickle_PromotingRecordType";
static const char __pyx_k_record_fields_must_be_unique_2[] = ", record_fields_must_be_unique=";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_string_types_default_unchanged[] = "string_types_default_unchanged";
static const char __pyx_k_tests_locals__test_reader_read[] = "_tests.<locals>._test_reader_read_u8";
static const char __pyx_k_union_json_encodes_type_name_2[] = ", union_json_encodes_type_name=";
static const char __pyx_k_A_record_contains_multiple_fiel[] = "\n    A record contains multiple fields with the same name, a schema contains multiple types of the same name, or an enum has multiple identical symbols.\n    ";
static const char __pyx_k_A_requested_codec_or_codec_in_a[] = "\n    A requested codec (or codec in a container) is not available or is unknown.\n    ";
static const char __pyx_k_An_unknown_hash_method_was_requ[] = "\n    An unknown hash method was requested\n    ";
static const char __pyx_k_Base_class_for_exceptions_raise[] = "\n    Base class for exceptions raised by cavro\n    ";
static const char __pyx_k_CannotPromote___setstate_cython[] = "CannotPromote.__setstate_cython__";
static const char __pyx_k_CanonicalForm___setstate_cython[] = "_CanonicalForm.__setstate_cython__";
static const char __pyx_k_ContainerReader___reduce_cython[] = "ContainerReader.__reduce_cython__";
static const char __pyx_k_ContainerWriter___reduce_cython[] = "ContainerWriter.__reduce_cython__";
static const char __pyx_k_Controls_how_union_values_are_d[] = "\n    Controls how union values are decoded:\n     * RAW_VALUES - The value of the matching union type is returned unmodified\n     * TYPE_TUPLE_IF_AMBIGUOUS - If union contains types that might be ambiguous (Record + Map) or (String + Enum), then the value is returned as a 2-tuple of (<type>, <value>)\n     * TYPE_TUPLE_IF_RECORD_AMBIGUOUS - If union contains multiple record (or map) types, then the value is returned as a 2-tuple of (<type>, <value>)\n     * TYPE_TUPLE_IF_RECORD - values matching any union member that is record are returned as a 2-tuple of (<type>, <value>)\n     * TYPE_TUPLE_ALWAYS - values are always returned as a 2-tuple of (<type>, <value>)\n    ";
static const char __pyx_k_FieldAccessor___setstate_cython[] = "_FieldAccessor.__setstate_cython__";
static const char __pyx_k_Invalid_negative_size_for_fixed[] = "Invalid negative size for fixed: ";
static const char __pyx_k_Invalid_non_str_value_for_fixed[] = "Invalid non-str value for fixed: '";
static const char __pyx_k_Invalid_union_decodes_to_option[] = "Invalid union_decodes_to option: ";
static const char __pyx_k_Marker_must_be_exactly_16_bytes[] = "Marker must be exactly 16 bytes, got: ";
static const char __pyx_k_NamedType_get_namespaced_aliase[] = "_NamedType.get_namespaced_aliases.<locals>.genexpr";
static const char __pyx_k_PlaceholderType___reduce_cython[] = "_PlaceholderType.__reduce_cython__";
static const char __pyx_k_PlaceholderType___setstate_cyth[] = "_PlaceholderType.__setstate_cython__";
static const char __pyx_k_PlaceholderType__get_schema_ext[] = "_PlaceholderType._get_schema_extra";
static const char __pyx_k_PromoteBytesToString___setstate[] = "PromoteBytesToString.__setstate_cython__";
static const char __pyx_k_PromoteStringToBytes___setstate[] = "PromoteStringToBytes.__setstate_cython__";
static const char __pyx_k_Promoting_schemas_reader_writer[] = "Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.";
static const char __pyx_k_SignatureWrapper__set_from_data[] = "_SignatureWrapper._set_from_dataclass";
static const char __pyx_k_The_exponent_of_a_decimal_value[] = "\n    The exponent of a decimal value is too large to be represented in the given type\n    ";
static const char __pyx_k_The_order_of_a_field_in_a_recor[] = "\n    The order of a field in a record.\n    ";
static const char __pyx_k_The_schema_contains_a_type_or_e[] = "\n    The schema contains a type or enum symbol with an invalid name (as per the avro specification)\n    ";
static const char __pyx_k_The_schema_contains_an_unexptec[] = "\n    The schema contains an unexptected type name (either a missing named-type definition, or invalid primitive type)\n    ";
static const char __pyx_k_TimestampMicros___reduce_cython[] = "TimestampMicros.__reduce_cython__";
static const char __pyx_k_TimestampMillis___reduce_cython[] = "TimestampMillis.__reduce_cython__";
static const char __pyx_k_UUIDFixedType___setstate_cython[] = "UUIDFixedType.__setstate_cython__";
static const char __pyx_k_ZStandardCodec___setstate_cytho[] = "_ZStandardCodec.__setstate_cython__";
static const char __pyx_k_allow_primitive_names_in_namesp[] = ", allow_primitive_names_in_namespaces=";
static const char __pyx_k_alternate_timestamp_millis_enco[] = ", alternate_timestamp_millis_encoding=";
static const char __pyx_k_appears_multiple_times_in_schem[] = " appears multiple times in schema";
static const char __pyx_k_cavro_is_a_library_for_encoding[] = "\n`cavro` is a library for encoding and decoding data in the Apache Avro format.\n\nIt's written in Cython, with a focus on performance, correctness and ease-of-use.\n";
static const char __pyx_k_defer_schema_promotion_errors_2[] = ", defer_schema_promotion_errors=";
static const char __pyx_k_does_not_implement__binary_buff[] = " does not implement _binary_buffer_encode";
static const char __pyx_k_does_not_implement__convert_val[] = " does not implement _convert_value";
static const char __pyx_k_does_not_implement__extract_met[] = " does not implement _extract_metadata";
static const char __pyx_k_does_not_implement__get_value_f[] = " does not implement _get_value_fitness";
static const char __pyx_k_does_not_implement__json_decode[] = " does not implement _json_decode";
static const char __pyx_k_does_not_implement__json_format[] = " does not implement _json_format";
static const char __pyx_k_does_not_implement_binary_buffe[] = " does not implement binary_buffer_decode";
static const char __pyx_k_does_not_implement_canonical_fo[] = " does not implement canonical_form";
static const char __pyx_k_does_not_implement_decode_value[] = " does not implement decode_value";
static const char __pyx_k_does_not_implement_encode_value[] = " does not implement encode_value";
static const char __pyx_k_invalid_value_include_array_ind[] = ", invalid_value_include_array_index=";
static const char __pyx_k_invalid_value_include_map_key_2[] = ", invalid_value_include_map_key=";
static const char __pyx_k_invalid_value_includes_record_n[] = ", invalid_value_includes_record_name=";
static const char __pyx_k_is_not_a_valid_key_for_map_type[] = "' is not a valid key for map type";
static const char __pyx_k_is_not_a_valid_union_value_expe[] = " is not a valid union value (expect exactly one item)";
static const char __pyx_k_is_not_a_valid_union_value_unkn[] = " is not a valid union value (unknown type '";
static const char __pyx_k_is_not_a_valid_value_for_map_ty[] = "' is not a valid value for map type: ";
static const char __pyx_k_is_not_valid_for_union_with_fir[] = " is not valid for union with first type: ";
static const char __pyx_k_missing_values_can_be_empty_con[] = ", missing_values_can_be_empty_container=";
static const char __pyx_k_named_type_names_must_be_unique[] = "named_type_names_must_be_unique";
static const char __pyx_k_tests_locals_add_locals_genexpr[] = "_tests.<locals>.add.<locals>.genexpr";
static const char __pyx_k_which_is_greater_than_the_scale[] = " which is greater than the scale ";
static const char __pyx_k_Aliases_must_be_a_list_tuple_set[] = "Aliases must be a list/tuple/set, got: ";
static const char __pyx_k_All_dimensions_preceding_dimensi[] = "All dimensions preceding dimension %d must be indexed and not sliced";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Cannot_transpose_memoryview_with[] = "Cannot transpose memoryview with indirect dimensions";
static const char __pyx_k_ContainerReader___setstate_cytho[] = "ContainerReader.__setstate_cython__";
static const char __pyx_k_ContainerWriter___setstate_cytho[] = "ContainerWriter.__setstate_cython__";
static const char __pyx_k_Could_not_find_key_type_in_schem[] = "Could not find key 'type' in schema definition: ";
static const char __pyx_k_CustomLogicalType___reduce_cytho[] = "CustomLogicalType.__reduce_cython__";
static const char __pyx_k_CustomLogicalType___setstate_cyt[] = "CustomLogicalType.__setstate_cython__";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_End_of_file_found_trying_to_read[] = "End of file found trying to read 1 byte";
static const char __pyx_k_EnumType__for_writer_locals_gene[] = "EnumType._for_writer.<locals>.genexpr";
static const char __pyx_k_Enum_symbols_must_be_a_list_of_s[] = "Enum symbols must be a list of strings";
static const char __pyx_k_Error_closing_container_file_dur[] = "Error closing container file during __del__: ";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_InvalidValue___str___locals_gene[] = "InvalidValue.__str__.<locals>.genexpr";
static const char __pyx_k_Invalid_block_sync_marker_expect[] = "Invalid block sync marker, expected ";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got ";
static const char __pyx_k_Invalid_non_bytes_value_for_fixe[] = "Invalid non-bytes value for fixed: '";
static const char __pyx_k_Not_enough_input_data_to_read_va[] = "Not enough input data to read value";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis ";
static const char __pyx_k_Placeholder_types_cannot_be_enco[] = "Placeholder types cannot be encoded";
static const char __pyx_k_Placeholder_types_have_no_canoni[] = "Placeholder types have no canonical form";
static const char __pyx_k_Precision_is_too_large_for_fixed[] = "Precision is too large for fixed size ";
static const char __pyx_k_Precision_must_be_an_integer_bet[] = "Precision must be an integer between 1 and ";
static const char __pyx_k_Precision_must_be_greater_than_s[] = "Precision must be greater than scale. scale: ";
static const char __pyx_k_PromoteBytesToString___reduce_cy[] = "PromoteBytesToString.__reduce_cython__";
static const char __pyx_k_PromoteStringToBytes___reduce_cy[] = "PromoteStringToBytes.__reduce_cython__";
static const char __pyx_k_PromoteToFloat___setstate_cython[] = "PromoteToFloat.__setstate_cython__";
static const char __pyx_k_PromotingEnumType___reduce_cytho[] = "PromotingEnumType.__reduce_cython__";
static const char __pyx_k_PromotingEnumType___setstate_cyt[] = "PromotingEnumType.__setstate_cython__";
static const char __pyx_k_PromotingRecordType___reduce_cyt[] = "PromotingRecordType.__reduce_cython__";
static const char __pyx_k_PromotingRecordType___setstate_c[] = "PromotingRecordType.__setstate_cython__";
static const char __pyx_k_Record__repr_children_locals_gen[] = "Record._repr_children.<locals>.genexpr";
static const char __pyx_k_Records_may_either_be_instantiat[] = "Records may either be instantiated with a record, or **kwargs, not both";
static const char __pyx_k_ResolvedSchema___setstate_cython[] = "ResolvedSchema.__setstate_cython__";
static const char __pyx_k_Scale_must_be_a_positive_integer[] = "Scale must be a positive integer, got: ";
static const char __pyx_k_The_null_namespace_cannot_be_spe[] = "The null namespace cannot be specified in name fields";
static const char __pyx_k_TimestampMicros___setstate_cytho[] = "TimestampMicros.__setstate_cython__";
static const char __pyx_k_TimestampMillis___setstate_cytho[] = "TimestampMillis.__setstate_cython__";
static const char __pyx_k_Trying_to_close_a_closed_Contain[] = "Trying to close a closed Container";
static const char __pyx_k_Trying_to_write_to_closed_Contai[] = "Trying to write to closed Container";
static const char __pyx_k_UUIDStringType___setstate_cython[] = "UUIDStringType.__setstate_cython__";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_Union_bytes_hashlib__hashlib_HAS[] = "Union[bytes, hashlib._hashlib.HASH]";
static const char __pyx_k_Unions_may_not_have_more_than_on[] = "Unions may not have more than one member of type '";
static const char __pyx_k_Unions_may_not_immediately_conta[] = "Unions may not immediately contain other unions";
static const char __pyx_k_Unions_must_contain_at_least_one[] = "Unions must contain at least one member type";
static const char __pyx_k_allow_primitive_name_collision_2[] = ", allow_primitive_name_collision=";
static const char __pyx_k_allow_primitive_names_in_namespa[] = "allow_primitive_names_in_namespaces";
static const char __pyx_k_allow_union_default_any_member_2[] = ", allow_union_default_any_member=";
static const char __pyx_k_alternate_timestamp_millis_encod[] = "alternate_timestamp_millis_encoding";
static const char __pyx_k_dict_to_canonical_locals_genexpr[] = "dict_to_canonical.<locals>.genexpr";
static const char __pyx_k_enforce_enum_symbol_name_rules_2[] = ", enforce_enum_symbol_name_rules=";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension ";
static const char __pyx_k_invalid_value_include_array_inde[] = "invalid_value_include_array_index";
static const char __pyx_k_invalid_value_includes_record_na[] = "invalid_value_includes_record_name";
static const char __pyx_k_missing_values_can_be_empty_cont[] = "missing_values_can_be_empty_container";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_string_types_default_unchanged_2[] = ", string_types_default_unchanged=";
static const char __pyx_k_tests_locals__test_reader_read_2[] = "_tests.<locals>._test_reader_read_n";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static const char __pyx_k_named_type_names_must_be_unique_2[] = ", named_type_names_must_be_unique=";
static const char __pyx_k_End_of_file_found_trying_to_read_2[] = "End of file found trying to read ";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0xc466972, 0x8dd8fb8, 0xf76cf3c) = (adapt_record_types, allow_aliases_to_be_string, allow_duplicate_union_types, allow_empty_unions, allow_error_type, allow_false_values_for_null, allow_invalid_default_values, allow_leading_dot_in_names, allow_nested_unions, allow_primitive_name_collision, allow_primitive_names_in_namespaces, allow_tuple_notation, allow_union_default_any_member, alternate_timestamp_millis_encoding, ascii_name_rules, bytes_codec, bytes_default_value_utf8, canonical_form_repeat_enum, canonical_form_repeat_fixed, clamp_float_overflow, clamp_int_overflow, coerce_int_to_float, coerce_values_to_boolean, coerce_values_to_float, coerce_values_to_int, coerce_values_to_str, container_fill_blocks, date_type_accepts_string, decimal_check_exp_overflow, defer_schema_promotion_errors, enforce_enum_symbol_name_rules, enforce_namespace_name_rules, enforce_type_name_rules, enum_symbols_must_be_unique, expand_types_in_schema, externally_defined_types, fingerprint_returns_digest, fixed_codec, float_out_of_range_inf, inline_namespaces, invalid_value_include_array_index, invalid_value_include_map_key, invalid_value_includes_record_name, logical_types, missing_values_can_be_empty_container, missing_values_can_be_null, naive_dt_assume_utc, named_type_names_must_be_unique, null_pad_fixed, raise_on_invalid_logical, record_allow_extra_fields, record_can_encode_dict, record_decodes_to_dict, record_encode_use_defaults, record_fields_must_be_unique, record_values_type_hint, return_uuid_object, string_types_default_unchanged, truncate_fixed, types_str_to_schema, unicode_errors, union_decodes_to, union_json_encodes_type_name))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_4[] = "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_5[] = "Incompatible checksums (0x%x vs (0xec0f6ac, 0xbcf8db1, 0xf4af146) = (data, end_ptr, ptr))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_6[] = "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_7[] = "Incompatible checksums (0x%x vs (0xcdf32d8, 0x60f601e, 0xa92a64f) = (item_type, metadata, options, value_adapters))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_8[] = "Incompatible checksums (0x%x vs (0x8f63b31, 0x117bc6b, 0xb063604) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, symbol_indexes, symbols, value_adapters))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_9[] = "Incompatible checksums (0x%x vs (0x8a61ff8, 0x47d8f26, 0xbae6ea0) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, reader_type, symbol_indexes, symbols, unknown_symbols, value_adapters, writer_type))";
static const char __pyx_k_Records_may_either_be_instantiat_2[] = "Records may either be instantiated with a single dict, or **kwargs, not both";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_10[] = "Incompatible checksums (0x%x vs (0xaa8230c, 0x4c3ce07, 0x4f7d725) = (key_type, metadata, options, value_adapters, value_type))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_11[] = "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_12[] = "Incompatible checksums (0x%x vs (0x5c2152b, 0x99117dc, 0xe166855) = (default_value, metadata, options, value_adapters))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_13[] = "Incompatible checksums (0x%x vs (0x3a6eb07, 0xa17c9aa, 0x8d777f3) = (data))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_14[] = "Incompatible checksums (0x%x vs (0x1d9a0f8, 0x4d78739, 0x24a1a05) = (aliases, default_value, doc, name, order, type, writer_name))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_15[] = "Incompatible checksums (0x%x vs (0x1bc04b5, 0xe540cdd, 0x6a992d5) = (index))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_16[] = "Incompatible checksums (0x%x vs (0x182acf2, 0x17ad56f, 0xde7d9e1) = (_setting_up, aliases, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_17[] = "Incompatible checksums (0x%x vs (0x40bba2b, 0x72480de, 0x6ac5437) = (_setting_up, aliases, decode_indexes, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_18[] = "Incompatible checksums (0x%x vs (0xd9f3afe, 0xacb1013, 0x605aa9f) = (aliases, effective_namespace, metadata, name, namespace, options, size, value_adapters))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_19[] = "Incompatible checksums (0x%x vs (0xabef18e, 0x6e50000, 0x97a7fa4) = (by_name_map, metadata, options, return_type_tuple, union_types, value_adapters))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_20[] = "Incompatible checksums (0x%x vs (0x24d7e16, 0xc138f68, 0xdaff9f5) = (aliases, effective_namespace, metadata, name, namespace, options, value_adapters))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_21[] = "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_22[] = "Incompatible checksums (0x%x vs (0xe9db223, 0xc0cb30d, 0x997aa79) = (check_exp_overflow, context, precision, scale, scale_val, size, underlying_name))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_23[] = "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_24[] = "Incompatible checksums (0x%x vs (0xc5331ff, 0xabd30c3, 0x97e5bf4) = (accepts_string))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_25[] = "Incompatible checksums (0x%x vs (0x8d5b257, 0x5b53641, 0x2d8f391) = (alternate_timestamp_encoding))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_26[] = "Incompatible checksums (0x%x vs (0x8d57cb2, 0xfb98f56, 0xb8cd0c2) = (compressor, decompressor))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_27[] = "Incompatible checksums (0x%x vs (0x289807d, 0x75481bf, 0x9c7e164) = (codec, codec_name, current_block, marker, metadata, objects_left_in_block, reader, reader_schema, schema, writer_schema))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_28[] = "Incompatible checksums (0x%x vs (0xcbfa7ba, 0xf8f7277, 0xca3fdae) = (extra, reader_type, writer_type))";
/* #### Code section: decls ### */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7cpython_5array_5array___getbuffer__(arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_7cpython_5array_5array_2__releasebuffer__(CYTHON_UNUSED arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_pf_5cavro_6_tests_3add_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_6_tests_add(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_test, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_5cavro__tests(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_fn); /* proto */
static PyObject *__pyx_pf_5cavro_5_perf_3add_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_5_perf_add(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_test, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_5cavro_2_perf(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_fn); /* proto */
static PyObject *__pyx_pf_5cavro_11UnknownType___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_5cavro_18CannotPromoteError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_reader_type, PyObject *__pyx_v_writer_type, PyObject *__pyx_v_extra); /* proto */
static PyObject *__pyx_pf_5cavro_12InvalidValue___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_dest_type, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_5cavro_12InvalidValue_7__str___genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_12InvalidValue_2__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_4_eval(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_5cavro_17_SignatureWrapper___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_17_SignatureWrapper_2_set_from_dataclass(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_cls); /* proto */
static struct __pyx_obj_5cavro_Options *__pyx_pf_5cavro_7Options_replace(struct __pyx_obj_5cavro_Options *__pyx_v_self, PyObject *__pyx_v_changes); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_2with_logical_types(struct __pyx_obj_5cavro_Options *__pyx_v_self, PyObject *__pyx_v_logical_types); /* proto */
static struct __pyx_obj_5cavro_Options *__pyx_pf_5cavro_7Options_4with_external_types(struct __pyx_obj_5cavro_Options *__pyx_v_self, PyObject *__pyx_v_named_types); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_12name_pattern___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_6equals(struct __pyx_obj_5cavro_Options *__pyx_v_self, struct __pyx_obj_5cavro_Options *__pyx_v_other, PyObject *__pyx_v_ignore); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_8diff(struct __pyx_obj_5cavro_Options *__pyx_v_self, struct __pyx_obj_5cavro_Options *__pyx_v_other, PyObject *__pyx_v_ignore); /* proto */
static int __pyx_pf_5cavro_7Options_10__init__(struct __pyx_obj_5cavro_Options *__pyx_v_self, PyObject *__pyx_v_fingerprint_returns_digest, PyObject *__pyx_v_canonical_form_repeat_fixed, PyObject *__pyx_v_canonical_form_repeat_enum, PyObject *__pyx_v_record_can_encode_dict, PyObject *__pyx_v_record_values_type_hint, PyObject *__pyx_v_record_decodes_to_dict, PyObject *__pyx_v_record_allow_extra_fields, PyObject *__pyx_v_record_encode_use_defaults, PyObject *__pyx_v_allow_tuple_notation, PyObject *__pyx_v_union_decodes_to, PyObject *__pyx_v_union_json_encodes_type_name, PyObject *__pyx_v_allow_primitive_name_collision, PyObject *__pyx_v_allow_primitive_names_in_namespaces, PyObject *__pyx_v_named_type_names_must_be_unique, PyObject *__pyx_v_enum_symbols_must_be_unique, PyObject *__pyx_v_enforce_enum_symbol_name_rules, PyObject *__pyx_v_enforce_type_name_rules, PyObject *__pyx_v_enforce_namespace_name_rules, PyObject *__pyx_v_record_fields_must_be_unique, PyObject *__pyx_v_ascii_name_rules, PyObject *__pyx_v_missing_values_can_be_null, PyObject *__pyx_v_missing_values_can_be_empty_container, PyObject *__pyx_v_allow_false_values_for_null, PyObject *__pyx_v_allow_empty_unions, PyObject *__pyx_v_allow_nested_unions, PyObject *__pyx_v_allow_duplicate_union_types, PyObject *__pyx_v_allow_union_default_any_member, PyObject *__pyx_v_allow_aliases_to_be_string, PyObject *__pyx_v_coerce_values_to_int, PyObject *__pyx_v_coerce_values_to_float, PyObject *__pyx_v_coerce_int_to_float, PyObject *__pyx_v_coerce_values_to_boolean, PyObject *__pyx_v_coerce_values_to_str, PyObject *__pyx_v_bytes_codec, PyObject *__pyx_v_fixed_codec, PyObject *__pyx_v_null_pad_fixed, PyObject *__pyx_v_truncate_fixed, PyObject *__pyx_v_clamp_int_overflow, PyObject *__pyx_v_clamp_float_overflow, PyObject *__pyx_v_float_out_of_range_inf, PyObject *__pyx_v_bytes_default_value_utf8, PyObject *__pyx_v_string_types_default_unchanged, PyObject *__pyx_v_decimal_check_exp_overflow, PyObject *__pyx_v_types_str_to_schema, PyObject *__pyx_v_logical_types, PyObject *__pyx_v_adapt_record_types, PyObject *__pyx_v_return_uuid_object, PyObject *__pyx_v_allow_error_type, PyObject *__pyx_v_allow_leading_dot_in_names, PyObject *__pyx_v_naive_dt_assume_utc, PyObject *__pyx_v_alternate_timestamp_millis_encoding, PyObject *__pyx_v_date_type_accepts_string, PyObject *__pyx_v_raise_on_invalid_logical, PyObject *__pyx_v_inline_namespaces, PyObject *__pyx_v_expand_types_in_schema, PyObject *__pyx_v_unicode_errors, PyObject *__pyx_v_container_fill_blocks, PyObject *__pyx_v_defer_schema_promotion_errors, PyObject *__pyx_v_invalid_value_includes_record_name, PyObject *__pyx_v_invalid_value_include_array_index, PyObject *__pyx_v_invalid_value_include_map_key, PyObject *__pyx_v_allow_invalid_default_values, PyObject *__pyx_v_externally_defined_types); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_12__repr__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_14__eq__(struct __pyx_obj_5cavro_Options *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static Py_hash_t __pyx_pf_5cavro_7Options_16__hash__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_26fingerprint_returns_digest___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_27canonical_form_repeat_fixed___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_26canonical_form_repeat_enum___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_22record_can_encode_dict___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_23record_values_type_hint___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_22record_decodes_to_dict___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_25record_allow_extra_fields___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_26record_encode_use_defaults___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_20allow_tuple_notation___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_16union_decodes_to___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_28union_json_encodes_type_name___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_30allow_primitive_name_collision___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_35allow_primitive_names_in_namespaces___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_31named_type_names_must_be_unique___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_27enum_symbols_must_be_unique___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_30enforce_enum_symbol_name_rules___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_23enforce_type_name_rules___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_28enforce_namespace_name_rules___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_28record_fields_must_be_unique___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_16ascii_name_rules___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_26missing_values_can_be_null___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_37missing_values_can_be_empty_container___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_27allow_false_values_for_null___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_18allow_empty_unions___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_19allow_nested_unions___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_27allow_duplicate_union_types___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_30allow_union_default_any_member___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_26allow_aliases_to_be_string___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_20coerce_values_to_int___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_22coerce_values_to_float___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_19coerce_int_to_float___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_24coerce_values_to_boolean___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_20coerce_values_to_str___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_11bytes_codec___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_11fixed_codec___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_14null_pad_fixed___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_14truncate_fixed___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_18clamp_int_overflow___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_20clamp_float_overflow___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_22float_out_of_range_inf___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_24bytes_default_value_utf8___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_30string_types_default_unchanged___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_26decimal_check_exp_overflow___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_19types_str_to_schema___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_13logical_types___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_18adapt_record_types___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_18return_uuid_object___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_16allow_error_type___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_26allow_leading_dot_in_names___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_19naive_dt_assume_utc___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_35alternate_timestamp_millis_encoding___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_24date_type_accepts_string___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_24raise_on_invalid_logical___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_17inline_namespaces___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_22expand_types_in_schema___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_14unicode_errors___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_21container_fill_blocks___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_29defer_schema_promotion_errors___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_34invalid_value_includes_record_name___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_33invalid_value_include_array_index___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_29invalid_value_include_map_key___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_28allow_invalid_default_values___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_24externally_defined_types___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_18__reduce_cython__(struct __pyx_obj_5cavro_Options *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7Options_20__setstate_cython__(struct __pyx_obj_5cavro_Options *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_7_Writer___reduce_cython__(struct __pyx_obj_5cavro__Writer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7_Writer_2__setstate_cython__(struct __pyx_obj_5cavro__Writer *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_7_Reader___reduce_cython__(struct __pyx_obj_5cavro__Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7_Reader_2__setstate_cython__(struct __pyx_obj_5cavro__Reader *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_10FileReader___init__(struct __pyx_obj_5cavro_FileReader *__pyx_v_self, PyObject *__pyx_v_file_obj); /* proto */
static PyObject *__pyx_pf_5cavro_10FileReader_2__reduce_cython__(struct __pyx_obj_5cavro_FileReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10FileReader_4__setstate_cython__(struct __pyx_obj_5cavro_FileReader *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_10FileWriter___init__(struct __pyx_obj_5cavro_FileWriter *__pyx_v_self, PyObject *__pyx_v_file_obj); /* proto */
static PyObject *__pyx_pf_5cavro_10FileWriter_2__reduce_cython__(struct __pyx_obj_5cavro_FileWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10FileWriter_4__setstate_cython__(struct __pyx_obj_5cavro_FileWriter *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_12MemoryWriter___cinit__(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self, PyObject *__pyx_v_initial_size); /* proto */
static PyObject *__pyx_pf_5cavro_12MemoryWriter_6buffer___get__(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_12MemoryWriter_3len___get__(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_12MemoryWriter_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_12MemoryWriter_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_12MemoryReader___init__(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_5cavro_12MemoryReader_2__reduce_cython__(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_12MemoryReader_4__setstate_cython__(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_5Rabin___init__(struct __pyx_obj_5cavro_Rabin *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_5Rabin_2update(struct __pyx_obj_5cavro_Rabin *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_5cavro_5Rabin_4digest(struct __pyx_obj_5cavro_Rabin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_5Rabin_6hexdigest(struct __pyx_obj_5cavro_Rabin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_5Rabin_8copy(struct __pyx_obj_5cavro_Rabin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_5Rabin_5value___get__(struct __pyx_obj_5cavro_Rabin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_5Rabin_10__reduce_cython__(struct __pyx_obj_5cavro_Rabin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_5Rabin_12__setstate_cython__(struct __pyx_obj_5cavro_Rabin *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_6_ubin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_val, PyObject *__pyx_v_n_bytes); /* proto */
static PyObject *__pyx_pf_5cavro_6_tests_2test_readvarint(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_mask); /* proto */
static PyObject *__pyx_pf_5cavro_6_tests_4test_zigzag_long(CYTHON_UNUSED PyObject *__pyx_self, uint64_t __pyx_v_mask); /* proto */
static PyObject *__pyx_pf_5cavro_6_tests_6test_readvarlong(CYTHON_UNUSED PyObject *__pyx_self, uint64_t __pyx_v_mask); /* proto */
static PyObject *__pyx_pf_5cavro_8_tests(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_add); /* proto */
static PyObject *__pyx_pf_5cavro_6_tests_8_test_reader_read_u8(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5cavro_6_tests_10_test_reader_read_n(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5cavro_10_tests(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_add); /* proto */
static PyObject *__pyx_pf_5cavro_5_perf_2_test_zigzag_perf(CYTHON_UNUSED PyObject *__pyx_self, uint64_t __pyx_v_mask); /* proto */
static PyObject *__pyx_pf_5cavro_12_perf(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_add); /* proto */
static int __pyx_pf_5cavro_9ArrayType___init__(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace); /* proto */
static PyObject *__pyx_pf_5cavro_9ArrayType_2copy(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9ArrayType_4walk_types(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_visited); /* proto */
static PyObject *__pyx_pf_5cavro_9ArrayType_7_get_schema_extra(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_9ArrayType_9convert_value(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_check_value); /* proto */
static PyObject *__pyx_pf_5cavro_9ArrayType_9item_type___get__(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9ArrayType_11__reduce_cython__(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9ArrayType_13__setstate_cython__(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_8EnumType___init__(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace); /* proto */
static PyObject *__pyx_pf_5cavro_8EnumType_2copy(struct __pyx_obj_5cavro_EnumType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8EnumType_4_get_schema_extra(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_8EnumType_6_convert_value(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_8EnumType_11_for_writer_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_8EnumType_7symbols___get__(struct __pyx_obj_5cavro_EnumType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8EnumType_13default_value___get__(struct __pyx_obj_5cavro_EnumType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8EnumType_3doc___get__(struct __pyx_obj_5cavro_EnumType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8EnumType_8__reduce_cython__(struct __pyx_obj_5cavro_EnumType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8EnumType_10__setstate_cython__(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_17PromotingEnumType_15unknown_symbols___get__(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_17PromotingEnumType_11reader_type___get__(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_17PromotingEnumType_11writer_type___get__(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_17PromotingEnumType___reduce_cython__(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_17PromotingEnumType_2__setstate_cython__(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_7MapType___init__(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace); /* proto */
static PyObject *__pyx_pf_5cavro_7MapType_2copy(struct __pyx_obj_5cavro_MapType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7MapType_4walk_types(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_visited); /* proto */
static PyObject *__pyx_pf_5cavro_7MapType_7_get_schema_extra(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_7MapType_9convert_value(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_orig_value, PyObject *__pyx_v_check_value); /* proto */
static PyObject *__pyx_pf_5cavro_7MapType_10value_type___get__(struct __pyx_obj_5cavro_MapType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7MapType_11__reduce_cython__(struct __pyx_obj_5cavro_MapType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7MapType_13__setstate_cython__(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_8NullType_copy(struct __pyx_obj_5cavro_NullType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8NullType_2_get_schema_extra(struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_8NullType_4_convert_value(struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_8NullType_6__reduce_cython__(struct __pyx_obj_5cavro_NullType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8NullType_8__setstate_cython__(struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_8BoolType_copy(struct __pyx_obj_5cavro_BoolType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8BoolType_2_get_schema_extra(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_8BoolType_4_convert_value(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_8BoolType_6__reduce_cython__(struct __pyx_obj_5cavro_BoolType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8BoolType_8__setstate_cython__(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_7IntType_copy(struct __pyx_obj_5cavro_IntType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7IntType_2_get_schema_extra(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_7IntType_4_convert_value(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_7IntType_6__reduce_cython__(struct __pyx_obj_5cavro_IntType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_7IntType_8__setstate_cython__(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_8LongType_copy(struct __pyx_obj_5cavro_LongType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8LongType_2_get_schema_extra(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_8LongType_4_convert_value(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_8LongType_6__reduce_cython__(struct __pyx_obj_5cavro_LongType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8LongType_8__setstate_cython__(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_9FloatType_copy(struct __pyx_obj_5cavro_FloatType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9FloatType_2_get_schema_extra(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_9FloatType_4_convert_value(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_9FloatType_6__reduce_cython__(struct __pyx_obj_5cavro_FloatType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9FloatType_8__setstate_cython__(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_10DoubleType_copy(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10DoubleType_2_get_schema_extra(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_10DoubleType_4_convert_value(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_10DoubleType_6__reduce_cython__(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10DoubleType_8__setstate_cython__(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_16_PlaceholderType___init__(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, PyObject *__pyx_v_options, PyObject *__pyx_v_default_value); /* proto */
static PyObject *__pyx_pf_5cavro_16_PlaceholderType_2_get_schema_extra(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_16_PlaceholderType_4_convert_value(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_16_PlaceholderType_13default_value___get__(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_16_PlaceholderType_6__reduce_cython__(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_16_PlaceholderType_8__setstate_cython__(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_6Record___init__(struct __pyx_obj_5cavro_Record *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_5cavro_6Record_2__dir__(struct __pyx_obj_5cavro_Record *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Record_4__getitem__(struct __pyx_obj_5cavro_Record *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_5cavro_6Record_14_repr_children_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_6Record_6__repr__(struct __pyx_obj_5cavro_Record *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Record_8_asdict(struct __pyx_obj_5cavro_Record *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Record_10__eq__(struct __pyx_obj_5cavro_Record *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_5cavro_6Record_12__reduce_cython__(struct __pyx_obj_5cavro_Record *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Record_14__setstate_cython__(struct __pyx_obj_5cavro_Record *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_11RecordField___init__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace); /* proto */
static PyObject *__pyx_pf_5cavro_11RecordField_2get_schema(struct __pyx_obj_5cavro_RecordField *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_11RecordField_4name___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11RecordField_11writer_name___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11RecordField_3doc___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11RecordField_4type___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11RecordField_13default_value___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11RecordField_5order___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11RecordField_7aliases___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11RecordField_4__reduce_cython__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11RecordField_6__setstate_cython__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_14_FieldAccessor___init__(struct __pyx_obj_5cavro__FieldAccessor *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static PyObject *__pyx_pf_5cavro_14_FieldAccessor_2__get__(struct __pyx_obj_5cavro__FieldAccessor *__pyx_v_self, PyObject *__pyx_v_inst, CYTHON_UNUSED PyObject *__pyx_v_cls); /* proto */
static int __pyx_pf_5cavro_14_FieldAccessor_4__set__(struct __pyx_obj_5cavro__FieldAccessor *__pyx_v_self, PyObject *__pyx_v_inst, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_14_FieldAccessor_6__reduce_cython__(struct __pyx_obj_5cavro__FieldAccessor *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_14_FieldAccessor_8__setstate_cython__(struct __pyx_obj_5cavro__FieldAccessor *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_10RecordType_8__init___genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static int __pyx_pf_5cavro_10RecordType___init__(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace); /* proto */
static PyObject *__pyx_pf_5cavro_10RecordType_2copy(struct __pyx_obj_5cavro_RecordType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10RecordType_4walk_types(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_visited); /* proto */
static PyObject *__pyx_pf_5cavro_10RecordType_7_get_schema_extra(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_10RecordType_9_convert_value(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_10RecordType_3doc___get__(struct __pyx_obj_5cavro_RecordType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10RecordType_6fields___get__(struct __pyx_obj_5cavro_RecordType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10RecordType_6record___get__(struct __pyx_obj_5cavro_RecordType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10RecordType_11__reduce_cython__(struct __pyx_obj_5cavro_RecordType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10RecordType_13__setstate_cython__(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_19PromotingRecordType___reduce_cython__(struct __pyx_obj_5cavro_PromotingRecordType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_19PromotingRecordType_2__setstate_cython__(struct __pyx_obj_5cavro_PromotingRecordType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_9BytesType_copy(struct __pyx_obj_5cavro_BytesType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9BytesType_2_get_schema_extra(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_9BytesType_4_convert_value(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_9BytesType_6__reduce_cython__(struct __pyx_obj_5cavro_BytesType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9BytesType_8__setstate_cython__(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_10StringType_copy(struct __pyx_obj_5cavro_StringType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10StringType_2_get_schema_extra(struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_10StringType_4_convert_value(struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_10StringType_6__reduce_cython__(struct __pyx_obj_5cavro_StringType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10StringType_8__setstate_cython__(struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_9FixedType___init__(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace); /* proto */
static PyObject *__pyx_pf_5cavro_9FixedType_2copy(struct __pyx_obj_5cavro_FixedType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9FixedType_4_get_schema_extra(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_9FixedType_6_convert_value(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_9FixedType_4size___get__(struct __pyx_obj_5cavro_FixedType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9FixedType_8__reduce_cython__(struct __pyx_obj_5cavro_FixedType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9FixedType_10__setstate_cython__(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_8__init___genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_8__init___3genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_8__init___6genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_8__init___9genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_8__init___12genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_8__init___15genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_8__init___18genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static int __pyx_pf_5cavro_9UnionType___init__(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_2copy(struct __pyx_obj_5cavro_UnionType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_4walk_types(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_visited); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_7get_schema(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_9convert_value(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_check_value); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_11union_types___get__(struct __pyx_obj_5cavro_UnionType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_11by_name_map___get__(struct __pyx_obj_5cavro_UnionType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_17return_type_tuple___get__(struct __pyx_obj_5cavro_UnionType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_11__reduce_cython__(struct __pyx_obj_5cavro_UnionType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9UnionType_13__setstate_cython__(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_9_Sentinel___init__(struct __pyx_obj_5cavro__Sentinel *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_5cavro_9_Sentinel_2__repr__(struct __pyx_obj_5cavro__Sentinel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9_Sentinel_4name___get__(struct __pyx_obj_5cavro__Sentinel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9_Sentinel_4__reduce_cython__(struct __pyx_obj_5cavro__Sentinel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9_Sentinel_6__setstate_cython__(struct __pyx_obj_5cavro__Sentinel *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_14_CanonicalForm___reduce_cython__(struct __pyx_obj_5cavro__CanonicalForm *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_14_CanonicalForm_2__setstate_cython__(struct __pyx_obj_5cavro__CanonicalForm *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_17dict_to_canonical_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_for_source(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_2for_schema(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_schema); /* proto */
static int __pyx_pf_5cavro_8AvroType_4__init__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, CYTHON_UNUSED PyObject *__pyx_v_namespace, PyObject *__pyx_v_value_adapters); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_6copy(struct __pyx_obj_5cavro_AvroType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_4type___get__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_8_convert_value(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_10convert_value(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_check_value); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_12_get_schema_extra(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_14walk_types(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_visited); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_17get_schema(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_19__str__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_7options___get__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_8metadata___get__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_14value_adapters___get__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_21__reduce_cython__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8AvroType_23__setstate_cython__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_10_NamedType___init__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_parse_namespace); /* proto */
static PyObject *__pyx_pf_5cavro_10_NamedType_4type___get__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10_NamedType_22get_namespaced_aliases_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_5cavro_10_NamedType_2_get_schema_extra(struct __pyx_obj_5cavro__NamedType *__pyx_v_self, PyObject *__pyx_v_created); /* proto */
static PyObject *__pyx_pf_5cavro_10_NamedType_4name___get__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10_NamedType_9namespace___get__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10_NamedType_19effective_namespace___get__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10_NamedType_7aliases___get__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10_NamedType_4__reduce_cython__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10_NamedType_6__setstate_cython__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_18_class_inst_method___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_5cavro_18_class_inst_method_2__get__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_inst, PyObject *__pyx_v_cls); /* proto */
static int __pyx_pf_5cavro_6Schema___init__(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_source, struct __pyx_obj_5cavro_Options *__pyx_v_options, PyObject *__pyx_v_named_types, __pyx_t_5cavro_bool __pyx_v_parse_json, PyObject *__pyx_v__type, PyObject *__pyx_v_extra_options); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_2_wrap_type(struct __pyx_obj_5cavro_Schema *__pyx_v_inst, PyObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_avro_type, struct __pyx_obj_5cavro_Options *__pyx_v_options); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_4_make_logical_types(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_options); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_14canonical_form___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_5cavro_6Schema_6__hash__(struct __pyx_obj_5cavro_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_6schema___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_10schema_str___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_8fingerprint(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_method, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_10find_type(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_name, int __pyx_v__raise); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_12can_encode(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_14binary_encode(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_16binary_decode(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_18binary_read(struct __pyx_obj_5cavro_Schema *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_20binary_write(struct __pyx_obj_5cavro_Schema *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_22json_encode(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_serialize, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_24json_decode(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_value, __pyx_t_5cavro_bool __pyx_v_deserialize, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_26reader_for_writer(struct __pyx_obj_5cavro_Schema *__pyx_v_self, struct __pyx_obj_5cavro_Schema *__pyx_v_writer_schema); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_11named_types___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_6source___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_7options___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_4type___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_13logical_types___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_28__reduce_cython__(struct __pyx_obj_5cavro_Schema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_6Schema_30__setstate_cython__(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_14ResolvedSchema___init__(struct __pyx_obj_5cavro_ResolvedSchema *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_resolved_type, struct __pyx_obj_5cavro_Options *__pyx_v_options); /* proto */
static PyObject *__pyx_pf_5cavro_14ResolvedSchema_2__reduce_cython__(struct __pyx_obj_5cavro_ResolvedSchema *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_14ResolvedSchema_4__setstate_cython__(struct __pyx_obj_5cavro_ResolvedSchema *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_12ValueAdapter___reduce_cython__(struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_12ValueAdapter_2__setstate_cython__(struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_11LogicalType_for_type(struct __pyx_obj_5cavro_LogicalType *__pyx_v_inst, PyObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying); /* proto */
static PyObject *__pyx_pf_5cavro_11LogicalType_2__reduce_cython__(struct __pyx_obj_5cavro_LogicalType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11LogicalType_4__setstate_cython__(struct __pyx_obj_5cavro_LogicalType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_17CustomLogicalType___reduce_cython__(struct __pyx_obj_5cavro_CustomLogicalType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_17CustomLogicalType_2__setstate_cython__(struct __pyx_obj_5cavro_CustomLogicalType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_11DecimalType___init__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self, PyObject *__pyx_v_precision, PyObject *__pyx_v_scale, PyObject *__pyx_v_size, PyObject *__pyx_v_check_exp_overflow, PyObject *__pyx_v_underlying_name); /* proto */
static PyObject *__pyx_pf_5cavro_11DecimalType_9type_name___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11DecimalType_2_for_type(PyTypeObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying); /* proto */
static PyObject *__pyx_pf_5cavro_11DecimalType_9precision___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11DecimalType_5scale___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11DecimalType_9scale_val___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11DecimalType_7context___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11DecimalType_4size___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11DecimalType_18check_exp_overflow___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11DecimalType_4__reduce_cython__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_11DecimalType_6__setstate_cython__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_8UUIDBase___init__(struct __pyx_obj_5cavro_UUIDBase *__pyx_v_self, PyObject *__pyx_v_return_uuid_object); /* proto */
static PyObject *__pyx_pf_5cavro_8UUIDBase_2_for_type(PyTypeObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying); /* proto */
static PyObject *__pyx_pf_5cavro_8UUIDBase_4__reduce_cython__(struct __pyx_obj_5cavro_UUIDBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_8UUIDBase_6__setstate_cython__(struct __pyx_obj_5cavro_UUIDBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_14UUIDStringType__for_type(PyTypeObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying); /* proto */
static PyObject *__pyx_pf_5cavro_14UUIDStringType_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_UUIDStringType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_14UUIDStringType_2__reduce_cython__(struct __pyx_obj_5cavro_UUIDStringType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_14UUIDStringType_4__setstate_cython__(struct __pyx_obj_5cavro_UUIDStringType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_13UUIDFixedType__for_type(PyTypeObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying); /* proto */
static PyObject *__pyx_pf_5cavro_13UUIDFixedType_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_UUIDFixedType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_13UUIDFixedType_2__reduce_cython__(struct __pyx_obj_5cavro_UUIDFixedType *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_13UUIDFixedType_4__setstate_cython__(struct __pyx_obj_5cavro_UUIDFixedType *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_4Date___init__(struct __pyx_obj_5cavro_Date *__pyx_v_self, PyObject *__pyx_v_accepts_string); /* proto */
static PyObject *__pyx_pf_5cavro_4Date_2_for_type(PyTypeObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying); /* proto */
static PyObject *__pyx_pf_5cavro_4Date_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_Date *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_4Date_14accepts_string___get__(struct __pyx_obj_5cavro_Date *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_4Date_4__reduce_cython__(struct __pyx_obj_5cavro_Date *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_4Date_6__setstate_cython__(struct __pyx_obj_5cavro_Date *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_14_time_to_micros(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_16_micros_to_time(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5cavro_10TimeMillis__for_type(PyTypeObject *__pyx_v_cls, CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying); /* proto */
static PyObject *__pyx_pf_5cavro_10TimeMillis_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_TimeMillis *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10TimeMillis_2__reduce_cython__(struct __pyx_obj_5cavro_TimeMillis *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10TimeMillis_4__setstate_cython__(struct __pyx_obj_5cavro_TimeMillis *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_10TimeMicros__for_type(PyTypeObject *__pyx_v_cls, CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying); /* proto */
static PyObject *__pyx_pf_5cavro_10TimeMicros_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_TimeMicros *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10TimeMicros_2__reduce_cython__(struct __pyx_obj_5cavro_TimeMicros *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10TimeMicros_4__setstate_cython__(struct __pyx_obj_5cavro_TimeMicros *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_15TimestampMillis___init__(struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self, PyObject *__pyx_v_alternate_timestamp_encoding); /* proto */
static PyObject *__pyx_pf_5cavro_15TimestampMillis_2_for_type(PyTypeObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying); /* proto */
static PyObject *__pyx_pf_5cavro_15TimestampMillis_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15TimestampMillis_28alternate_timestamp_encoding___get__(struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15TimestampMillis_4__reduce_cython__(struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15TimestampMillis_6__setstate_cython__(struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_15TimestampMicros__for_type(PyTypeObject *__pyx_v_cls, CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying); /* proto */
static PyObject *__pyx_pf_5cavro_15TimestampMicros_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_TimestampMicros *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15TimestampMicros_2__reduce_cython__(struct __pyx_obj_5cavro_TimestampMicros *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15TimestampMicros_4__setstate_cython__(struct __pyx_obj_5cavro_TimestampMicros *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_5Codec___reduce_cython__(struct __pyx_obj_5cavro_Codec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_5Codec_2__setstate_cython__(struct __pyx_obj_5cavro_Codec *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_12_SnappyCodec___reduce_cython__(struct __pyx_obj_5cavro__SnappyCodec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_12_SnappyCodec_2__setstate_cython__(struct __pyx_obj_5cavro__SnappyCodec *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_10_NullCodec___reduce_cython__(struct __pyx_obj_5cavro__NullCodec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10_NullCodec_2__setstate_cython__(struct __pyx_obj_5cavro__NullCodec *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_13_DeflateCodec___reduce_cython__(struct __pyx_obj_5cavro__DeflateCodec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_13_DeflateCodec_2__setstate_cython__(struct __pyx_obj_5cavro__DeflateCodec *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_12__Bzip2Codec___reduce_cython__(struct __pyx_obj_5cavro___Bzip2Codec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_12__Bzip2Codec_2__setstate_cython__(struct __pyx_obj_5cavro___Bzip2Codec *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_10_LzmaCodec___reduce_cython__(struct __pyx_obj_5cavro__LzmaCodec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_10_LzmaCodec_2__setstate_cython__(struct __pyx_obj_5cavro__LzmaCodec *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_15_ZStandardCodec___init__(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15_ZStandardCodec_10compressor___get__(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15_ZStandardCodec_12decompressor___get__(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15_ZStandardCodec_2__reduce_cython__(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15_ZStandardCodec_4__setstate_cython__(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_9_Lz4Codec___reduce_cython__(struct __pyx_obj_5cavro__Lz4Codec *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_9_Lz4Codec_2__setstate_cython__(struct __pyx_obj_5cavro__Lz4Codec *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_15ContainerReader___init__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self, PyObject *__pyx_v_src, PyObject *__pyx_v_reader_schema, PyObject *__pyx_v_options); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_2_read_marker(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_4next_object(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_6__iter__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_8__next__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_8metadata___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_6marker___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_13writer_schema___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_13reader_schema___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_6schema___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_10codec_name___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_21objects_left_in_block___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_10__reduce_cython__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerReader_12__setstate_cython__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_15ContainerWriter___cinit__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, PyObject *__pyx_v_dest, struct __pyx_obj_5cavro_Schema *__pyx_v_schema, PyObject *__pyx_v_codec, size_t __pyx_v_max_blocksize, PyObject *__pyx_v_write_header, PyObject *__pyx_v_metadata, PyObject *__pyx_v_marker, PyObject *__pyx_v_options); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_2__enter__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_4__exit__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_exc_type, CYTHON_UNUSED PyObject *__pyx_v_exc_val, CYTHON_UNUSED PyObject *__pyx_v_exc_tb); /* proto */
static void __pyx_pf_5cavro_15ContainerWriter_6__del__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_6closed___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_8close(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_10flush(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, PyObject *__pyx_v_force); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_12write_one(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_14write_many(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, PyObject *__pyx_v_objs); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_6schema___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_5codec___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_6marker___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_13max_blocksize___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_7options___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_19should_write_header___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_11num_pending___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_14blocks_written___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_8metadata___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_15ContainerWriter_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_14PromoteToFloat___reduce_cython__(struct __pyx_obj_5cavro_PromoteToFloat *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_14PromoteToFloat_2__setstate_cython__(struct __pyx_obj_5cavro_PromoteToFloat *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_20PromoteBytesToString___reduce_cython__(struct __pyx_obj_5cavro_PromoteBytesToString *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_20PromoteBytesToString_2__setstate_cython__(struct __pyx_obj_5cavro_PromoteBytesToString *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_20PromoteStringToBytes___reduce_cython__(struct __pyx_obj_5cavro_PromoteStringToBytes *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_20PromoteStringToBytes_2__setstate_cython__(struct __pyx_obj_5cavro_PromoteStringToBytes *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5cavro_13CannotPromote___init__(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self, PyObject *__pyx_v_reader_type, PyObject *__pyx_v_writer_type, PyObject *__pyx_v_extra); /* proto */
static PyObject *__pyx_pf_5cavro_13CannotPromote_11reader_type___get__(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_13CannotPromote_11writer_type___get__(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_13CannotPromote_5extra___get__(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_13CannotPromote_2__reduce_cython__(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5cavro_13CannotPromote_4__setstate_cython__(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_18__pyx_unpickle_Options(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_20__pyx_unpickle__Writer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_22__pyx_unpickle__Reader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_24__pyx_unpickle_FileReader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_26__pyx_unpickle_FileWriter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_28__pyx_unpickle_MemoryReader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_30__pyx_unpickle_Rabin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_32__pyx_unpickle_ArrayType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_34__pyx_unpickle_EnumType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_36__pyx_unpickle_PromotingEnumType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_38__pyx_unpickle_MapType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_40__pyx_unpickle_NullType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_42__pyx_unpickle_BoolType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_44__pyx_unpickle_IntType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_46__pyx_unpickle_LongType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_48__pyx_unpickle_FloatType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_50__pyx_unpickle_DoubleType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_52__pyx_unpickle__PlaceholderType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_54__pyx_unpickle_Record(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_56__pyx_unpickle_RecordField(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_58__pyx_unpickle__FieldAccessor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_60__pyx_unpickle_RecordType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_62__pyx_unpickle_PromotingRecordType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_64__pyx_unpickle_BytesType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_66__pyx_unpickle_StringType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_68__pyx_unpickle_FixedType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_70__pyx_unpickle_UnionType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_72__pyx_unpickle__Sentinel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_74__pyx_unpickle__CanonicalForm(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_76__pyx_unpickle_AvroType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_78__pyx_unpickle__NamedType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_80__pyx_unpickle_Schema(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_82__pyx_unpickle_ResolvedSchema(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_84__pyx_unpickle_ValueAdapter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_86__pyx_unpickle_LogicalType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_88__pyx_unpickle_CustomLogicalType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_90__pyx_unpickle_DecimalType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_92__pyx_unpickle_UUIDBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_94__pyx_unpickle_UUIDStringType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_96__pyx_unpickle_UUIDFixedType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_98__pyx_unpickle_Date(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_100__pyx_unpickle_TimeMillis(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_102__pyx_unpickle_TimeMicros(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_104__pyx_unpickle_TimestampMillis(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_106__pyx_unpickle_TimestampMicros(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_108__pyx_unpickle_Codec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_110__pyx_unpickle__SnappyCodec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_112__pyx_unpickle__NullCodec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_114__pyx_unpickle__DeflateCodec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_116__pyx_unpickle___Bzip2Codec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_118__pyx_unpickle__LzmaCodec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_120__pyx_unpickle__ZStandardCodec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_122__pyx_unpickle__Lz4Codec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_124__pyx_unpickle_ContainerReader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_126__pyx_unpickle_PromoteToFloat(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_128__pyx_unpickle_PromoteBytesToString(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_130__pyx_unpickle_PromoteStringToBytes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5cavro_132__pyx_unpickle_CannotPromote(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_5cavro_Options(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro__Writer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro__Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_FileReader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_FileWriter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_MemoryWriter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_MemoryReader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_Rabin(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_Record(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_RecordField(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro__FieldAccessor(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro__Sentinel(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_AvroType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_UnionType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_StringType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_BytesType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro__PlaceholderType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_DoubleType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_FloatType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_LongType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_IntType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_BoolType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_NullType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_MapType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_ArrayType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro__NamedType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_FixedType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_RecordType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_PromotingRecordType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_EnumType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_PromotingEnumType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_Schema(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_ResolvedSchema(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_ValueAdapter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_LogicalType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_CustomLogicalType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_DecimalType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_UUIDBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_UUIDStringType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_UUIDFixedType(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_Date(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_TimeMillis(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_TimeMicros(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_TimestampMillis(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_TimestampMicros(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_Codec(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro__SnappyCodec(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro__NullCodec(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro__DeflateCodec(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___Bzip2Codec(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro__LzmaCodec(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro__ZStandardCodec(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro__Lz4Codec(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_ContainerReader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_ContainerWriter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_PromoteToFloat(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_PromoteBytesToString(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_PromoteStringToBytes(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro_CannotPromote(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct__genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_2_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_3_walk_types(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_5_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_6_walk_types(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_7_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_8___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_9_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_10_walk_types(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_11___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_12_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_13_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_14_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_15_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_16_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_17_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_18_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_19_walk_types(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_20_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_21_walk_types(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_22_get_namespaced_aliases(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_23_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_keys = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_setdefault = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_update = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type_rjust = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type_rsplit = {0, 0, 0, 0, 0};
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_7complex_complex;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_5array_array;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_5cavro_Options;
  PyObject *__pyx_type_5cavro__Writer;
  PyObject *__pyx_type_5cavro__Reader;
  PyObject *__pyx_type_5cavro_FileReader;
  PyObject *__pyx_type_5cavro_FileWriter;
  PyObject *__pyx_type_5cavro_MemoryWriter;
  PyObject *__pyx_type_5cavro_MemoryReader;
  PyObject *__pyx_type_5cavro_Rabin;
  PyObject *__pyx_type_5cavro_Record;
  PyObject *__pyx_type_5cavro_RecordField;
  PyObject *__pyx_type_5cavro__FieldAccessor;
  PyObject *__pyx_type_5cavro__Sentinel;
  PyObject *__pyx_type_5cavro__CanonicalForm;
  PyObject *__pyx_type_5cavro_AvroType;
  PyObject *__pyx_type_5cavro_UnionType;
  PyObject *__pyx_type_5cavro_StringType;
  PyObject *__pyx_type_5cavro_BytesType;
  PyObject *__pyx_type_5cavro__PlaceholderType;
  PyObject *__pyx_type_5cavro_DoubleType;
  PyObject *__pyx_type_5cavro_FloatType;
  PyObject *__pyx_type_5cavro_LongType;
  PyObject *__pyx_type_5cavro_IntType;
  PyObject *__pyx_type_5cavro_BoolType;
  PyObject *__pyx_type_5cavro_NullType;
  PyObject *__pyx_type_5cavro_MapType;
  PyObject *__pyx_type_5cavro_ArrayType;
  PyObject *__pyx_type_5cavro__NamedType;
  PyObject *__pyx_type_5cavro_FixedType;
  PyObject *__pyx_type_5cavro_RecordType;
  PyObject *__pyx_type_5cavro_PromotingRecordType;
  PyObject *__pyx_type_5cavro_EnumType;
  PyObject *__pyx_type_5cavro_PromotingEnumType;
  PyObject *__pyx_type_5cavro_Schema;
  PyObject *__pyx_type_5cavro_ResolvedSchema;
  PyObject *__pyx_type_5cavro_ValueAdapter;
  PyObject *__pyx_type_5cavro_LogicalType;
  PyObject *__pyx_type_5cavro_CustomLogicalType;
  PyObject *__pyx_type_5cavro_DecimalType;
  PyObject *__pyx_type_5cavro_UUIDBase;
  PyObject *__pyx_type_5cavro_UUIDStringType;
  PyObject *__pyx_type_5cavro_UUIDFixedType;
  PyObject *__pyx_type_5cavro_Date;
  PyObject *__pyx_type_5cavro_TimeMillis;
  PyObject *__pyx_type_5cavro_TimeMicros;
  PyObject *__pyx_type_5cavro_TimestampMillis;
  PyObject *__pyx_type_5cavro_TimestampMicros;
  PyObject *__pyx_type_5cavro_Codec;
  PyObject *__pyx_type_5cavro__SnappyCodec;
  PyObject *__pyx_type_5cavro__NullCodec;
  PyObject *__pyx_type_5cavro__DeflateCodec;
  PyObject *__pyx_type_5cavro___Bzip2Codec;
  PyObject *__pyx_type_5cavro__LzmaCodec;
  PyObject *__pyx_type_5cavro__ZStandardCodec;
  PyObject *__pyx_type_5cavro__Lz4Codec;
  PyObject *__pyx_type_5cavro_ContainerReader;
  PyObject *__pyx_type_5cavro_ContainerWriter;
  PyObject *__pyx_type_5cavro_PromoteToFloat;
  PyObject *__pyx_type_5cavro_PromoteBytesToString;
  PyObject *__pyx_type_5cavro_PromoteStringToBytes;
  PyObject *__pyx_type_5cavro_CannotPromote;
  PyObject *__pyx_type_5cavro___pyx_scope_struct__genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_1_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_2_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_3_walk_types;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_5_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_6_walk_types;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_7_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_8___init__;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_9_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_10_walk_types;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_11___init__;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_12_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_13_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_14_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_15_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_16_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_17_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_18_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_19_walk_types;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_20_genexpr;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_21_walk_types;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_22_get_namespaced_aliases;
  PyObject *__pyx_type_5cavro___pyx_scope_struct_23_genexpr;
  PyObject *__pyx_type___pyx_array;
  PyObject *__pyx_type___pyx_MemviewEnum;
  PyObject *__pyx_type___pyx_memoryview;
  PyObject *__pyx_type___pyx_memoryviewslice;
  #endif
  PyTypeObject *__pyx_ptype_5cavro_Options;
  PyTypeObject *__pyx_ptype_5cavro__Writer;
  PyTypeObject *__pyx_ptype_5cavro__Reader;
  PyTypeObject *__pyx_ptype_5cavro_FileReader;
  PyTypeObject *__pyx_ptype_5cavro_FileWriter;
  PyTypeObject *__pyx_ptype_5cavro_MemoryWriter;
  PyTypeObject *__pyx_ptype_5cavro_MemoryReader;
  PyTypeObject *__pyx_ptype_5cavro_Rabin;
  PyTypeObject *__pyx_ptype_5cavro_Record;
  PyTypeObject *__pyx_ptype_5cavro_RecordField;
  PyTypeObject *__pyx_ptype_5cavro__FieldAccessor;
  PyTypeObject *__pyx_ptype_5cavro__Sentinel;
  PyTypeObject *__pyx_ptype_5cavro__CanonicalForm;
  PyTypeObject *__pyx_ptype_5cavro_AvroType;
  PyTypeObject *__pyx_ptype_5cavro_UnionType;
  PyTypeObject *__pyx_ptype_5cavro_StringType;
  PyTypeObject *__pyx_ptype_5cavro_BytesType;
  PyTypeObject *__pyx_ptype_5cavro__PlaceholderType;
  PyTypeObject *__pyx_ptype_5cavro_DoubleType;
  PyTypeObject *__pyx_ptype_5cavro_FloatType;
  PyTypeObject *__pyx_ptype_5cavro_LongType;
  PyTypeObject *__pyx_ptype_5cavro_IntType;
  PyTypeObject *__pyx_ptype_5cavro_BoolType;
  PyTypeObject *__pyx_ptype_5cavro_NullType;
  PyTypeObject *__pyx_ptype_5cavro_MapType;
  PyTypeObject *__pyx_ptype_5cavro_ArrayType;
  PyTypeObject *__pyx_ptype_5cavro__NamedType;
  PyTypeObject *__pyx_ptype_5cavro_FixedType;
  PyTypeObject *__pyx_ptype_5cavro_RecordType;
  PyTypeObject *__pyx_ptype_5cavro_PromotingRecordType;
  PyTypeObject *__pyx_ptype_5cavro_EnumType;
  PyTypeObject *__pyx_ptype_5cavro_PromotingEnumType;
  PyTypeObject *__pyx_ptype_5cavro_Schema;
  PyTypeObject *__pyx_ptype_5cavro_ResolvedSchema;
  PyTypeObject *__pyx_ptype_5cavro_ValueAdapter;
  PyTypeObject *__pyx_ptype_5cavro_LogicalType;
  PyTypeObject *__pyx_ptype_5cavro_CustomLogicalType;
  PyTypeObject *__pyx_ptype_5cavro_DecimalType;
  PyTypeObject *__pyx_ptype_5cavro_UUIDBase;
  PyTypeObject *__pyx_ptype_5cavro_UUIDStringType;
  PyTypeObject *__pyx_ptype_5cavro_UUIDFixedType;
  PyTypeObject *__pyx_ptype_5cavro_Date;
  PyTypeObject *__pyx_ptype_5cavro_TimeMillis;
  PyTypeObject *__pyx_ptype_5cavro_TimeMicros;
  PyTypeObject *__pyx_ptype_5cavro_TimestampMillis;
  PyTypeObject *__pyx_ptype_5cavro_TimestampMicros;
  PyTypeObject *__pyx_ptype_5cavro_Codec;
  PyTypeObject *__pyx_ptype_5cavro__SnappyCodec;
  PyTypeObject *__pyx_ptype_5cavro__NullCodec;
  PyTypeObject *__pyx_ptype_5cavro__DeflateCodec;
  PyTypeObject *__pyx_ptype_5cavro___Bzip2Codec;
  PyTypeObject *__pyx_ptype_5cavro__LzmaCodec;
  PyTypeObject *__pyx_ptype_5cavro__ZStandardCodec;
  PyTypeObject *__pyx_ptype_5cavro__Lz4Codec;
  PyTypeObject *__pyx_ptype_5cavro_ContainerReader;
  PyTypeObject *__pyx_ptype_5cavro_ContainerWriter;
  PyTypeObject *__pyx_ptype_5cavro_PromoteToFloat;
  PyTypeObject *__pyx_ptype_5cavro_PromoteBytesToString;
  PyTypeObject *__pyx_ptype_5cavro_PromoteStringToBytes;
  PyTypeObject *__pyx_ptype_5cavro_CannotPromote;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct__genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_1_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_2_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_3_walk_types;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_5_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_6_walk_types;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_7_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_8___init__;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_9_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_10_walk_types;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_11___init__;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_12_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_13_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_14_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_15_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_16_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_17_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_18_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_19_walk_types;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_20_genexpr;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_21_walk_types;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases;
  PyTypeObject *__pyx_ptype_5cavro___pyx_scope_struct_23_genexpr;
  PyTypeObject *__pyx_array_type;
  PyTypeObject *__pyx_MemviewEnum_type;
  PyTypeObject *__pyx_memoryview_type;
  PyTypeObject *__pyx_memoryviewslice_type;
  PyObject *__pyx_kp_u_;
  PyObject *__pyx_kp_u_0;
  PyObject *__pyx_kp_u_1;
  PyObject *__pyx_kp_u_1_0_0;
  PyObject *__pyx_n_s_AMBIGUOUS_TYPES;
  PyObject *__pyx_n_s_ASC;
  PyObject *__pyx_n_s_ASCII;
  PyObject *__pyx_kp_u_A_Za_z__A_Za_z0_9;
  PyObject *__pyx_kp_s_A_record_contains_multiple_fiel;
  PyObject *__pyx_kp_s_A_requested_codec_or_codec_in_a;
  PyObject *__pyx_kp_s_A_schema_cannot_be_promoted_to;
  PyObject *__pyx_kp_s_A_value_is_invalid_for_a_given;
  PyObject *__pyx_kp_u_Aliases_must_be_a_list_tuple_set;
  PyObject *__pyx_kp_s_All_dimensions_preceding_dimensi;
  PyObject *__pyx_kp_s_An_unknown_hash_method_was_requ;
  PyObject *__pyx_n_s_ArrayType;
  PyObject *__pyx_n_s_ArrayType___reduce_cython;
  PyObject *__pyx_n_s_ArrayType___setstate_cython;
  PyObject *__pyx_n_s_ArrayType__get_schema_extra;
  PyObject *__pyx_n_s_ArrayType_convert_value;
  PyObject *__pyx_n_s_ArrayType_copy;
  PyObject *__pyx_n_s_ArrayType_walk_types;
  PyObject *__pyx_kp_s_As_of_cython_3_0_2_dataclasses;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_s_AttributeError;
  PyObject *__pyx_n_s_AvroType;
  PyObject *__pyx_n_s_AvroType___reduce_cython;
  PyObject *__pyx_n_s_AvroType___setstate_cython;
  PyObject *__pyx_n_s_AvroType__convert_value;
  PyObject *__pyx_n_s_AvroType__get_schema_extra;
  PyObject *__pyx_n_s_AvroType_convert_value;
  PyObject *__pyx_n_s_AvroType_copy;
  PyObject *__pyx_n_s_AvroType_for_schema;
  PyObject *__pyx_n_s_AvroType_for_source;
  PyObject *__pyx_n_s_AvroType_get_schema;
  PyObject *__pyx_n_s_AvroType_walk_types;
  PyObject *__pyx_n_u_B;
  PyObject *__pyx_n_s_BATCH_SIZE;
  PyObject *__pyx_kp_s_Base_class_for_exceptions_raise;
  PyObject *__pyx_n_s_BoolType;
  PyObject *__pyx_n_s_BoolType___reduce_cython;
  PyObject *__pyx_n_s_BoolType___setstate_cython;
  PyObject *__pyx_n_s_BoolType__convert_value;
  PyObject *__pyx_n_s_BoolType__get_schema_extra;
  PyObject *__pyx_n_s_BoolType_copy;
  PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
  PyObject *__pyx_n_s_BytesType;
  PyObject *__pyx_n_s_BytesType___reduce_cython;
  PyObject *__pyx_n_s_BytesType___setstate_cython;
  PyObject *__pyx_n_s_BytesType__convert_value;
  PyObject *__pyx_n_s_BytesType__get_schema_extra;
  PyObject *__pyx_n_s_BytesType_copy;
  PyObject *__pyx_n_s_Bzip2Codec;
  PyObject *__pyx_n_s_Bzip2Codec___reduce_cython;
  PyObject *__pyx_n_s_Bzip2Codec___setstate_cython;
  PyObject *__pyx_n_s_CANONICAL_FORM_KEYS;
  PyObject *__pyx_n_s_CODECS;
  PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
  PyObject *__pyx_n_s_CannotPromote;
  PyObject *__pyx_n_s_CannotPromoteError;
  PyObject *__pyx_n_s_CannotPromoteError___init;
  PyObject *__pyx_n_s_CannotPromote___reduce_cython;
  PyObject *__pyx_n_s_CannotPromote___setstate_cython;
  PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
  PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
  PyObject *__pyx_kp_u_Cannot_index_with_type;
  PyObject *__pyx_kp_u_Cannot_promote;
  PyObject *__pyx_kp_u_Cannot_read_from;
  PyObject *__pyx_kp_s_Cannot_transpose_memoryview_with;
  PyObject *__pyx_kp_u_Cannot_write_to;
  PyObject *__pyx_n_s_CanonicalForm;
  PyObject *__pyx_n_s_CanonicalForm___reduce_cython;
  PyObject *__pyx_n_s_CanonicalForm___setstate_cython;
  PyObject *__pyx_n_s_CavroException;
  PyObject *__pyx_n_s_Codec;
  PyObject *__pyx_n_s_CodecUnavailable;
  PyObject *__pyx_n_s_Codec___reduce_cython;
  PyObject *__pyx_n_s_Codec___setstate_cython;
  PyObject *__pyx_n_s_ContainerReader;
  PyObject *__pyx_n_s_ContainerReader___reduce_cython;
  PyObject *__pyx_n_s_ContainerReader___setstate_cytho;
  PyObject *__pyx_n_s_ContainerReader__read_marker;
  PyObject *__pyx_n_s_ContainerReader_next_object;
  PyObject *__pyx_n_s_ContainerWriter;
  PyObject *__pyx_n_s_ContainerWriter___enter;
  PyObject *__pyx_n_s_ContainerWriter___exit;
  PyObject *__pyx_n_s_ContainerWriter___reduce_cython;
  PyObject *__pyx_n_s_ContainerWriter___setstate_cytho;
  PyObject *__pyx_n_s_ContainerWriter_close;
  PyObject *__pyx_n_s_ContainerWriter_flush;
  PyObject *__pyx_n_s_ContainerWriter_write_many;
  PyObject *__pyx_n_s_ContainerWriter_write_one;
  PyObject *__pyx_kp_u_Container_is_closed;
  PyObject *__pyx_n_s_Context;
  PyObject *__pyx_kp_s_Controls_how_union_values_are_d;
  PyObject *__pyx_kp_u_Could_not_find_key_type_in_schem;
  PyObject *__pyx_n_s_CustomLogicalType;
  PyObject *__pyx_n_s_CustomLogicalType___reduce_cytho;
  PyObject *__pyx_n_s_CustomLogicalType___setstate_cyt;
  PyObject *__pyx_n_s_DEFAULT_OPTIONS;
  PyObject *__pyx_n_s_DESC;
  PyObject *__pyx_n_s_DataclassParams;
  PyObject *__pyx_n_s_Date;
  PyObject *__pyx_n_s_Date___reduce_cython;
  PyObject *__pyx_n_s_Date___setstate_cython;
  PyObject *__pyx_n_s_Date__for_type;
  PyObject *__pyx_n_s_Decimal;
  PyObject *__pyx_n_s_DecimalType;
  PyObject *__pyx_n_s_DecimalType___reduce_cython;
  PyObject *__pyx_n_s_DecimalType___setstate_cython;
  PyObject *__pyx_n_s_DecimalType__for_type;
  PyObject *__pyx_kp_u_Decimal_value;
  PyObject *__pyx_kp_u_Default_value;
  PyObject *__pyx_kp_u_Default_value_2;
  PyObject *__pyx_n_s_DeflateCodec;
  PyObject *__pyx_n_s_DeflateCodec___reduce_cython;
  PyObject *__pyx_n_s_DeflateCodec___setstate_cython;
  PyObject *__pyx_kp_s_Dimension_d_is_not_direct;
  PyObject *__pyx_n_s_DoubleType;
  PyObject *__pyx_n_s_DoubleType___reduce_cython;
  PyObject *__pyx_n_s_DoubleType___setstate_cython;
  PyObject *__pyx_n_s_DoubleType__convert_value;
  PyObject *__pyx_n_s_DoubleType__get_schema_extra;
  PyObject *__pyx_n_s_DoubleType_copy;
  PyObject *__pyx_n_s_DuplicateName;
  PyObject *__pyx_kp_u_Duplicate_field_name;
  PyObject *__pyx_n_s_EMPTY_ARRAY;
  PyObject *__pyx_n_s_EMPTY_MP;
  PyObject *__pyx_n_s_EOFError;
  PyObject *__pyx_n_s_EPOCH_DATE;
  PyObject *__pyx_n_s_EPOCH_DT;
  PyObject *__pyx_n_s_EPOCH_DT_NO_TZ;
  PyObject *__pyx_n_s_Ellipsis;
  PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
  PyObject *__pyx_kp_u_End_of_file_found_trying_to_read;
  PyObject *__pyx_kp_u_End_of_file_found_trying_to_read_2;
  PyObject *__pyx_n_s_Enum;
  PyObject *__pyx_n_s_EnumType;
  PyObject *__pyx_n_s_EnumType___reduce_cython;
  PyObject *__pyx_n_s_EnumType___setstate_cython;
  PyObject *__pyx_n_s_EnumType__convert_value;
  PyObject *__pyx_n_s_EnumType__for_writer_locals_gene;
  PyObject *__pyx_n_s_EnumType__get_schema_extra;
  PyObject *__pyx_n_s_EnumType_copy;
  PyObject *__pyx_kp_u_Enum_symbol;
  PyObject *__pyx_kp_u_Enum_symbols_must_be_a_list_of_s;
  PyObject *__pyx_kp_u_Enum_symbols_must_be_unique;
  PyObject *__pyx_kp_u_Error_closing_container_file_dur;
  PyObject *__pyx_kp_u_Expected_datetime_date_got;
  PyObject *__pyx_kp_u_Expected_datetime_datetime_got;
  PyObject *__pyx_kp_u_Expected_null_got;
  PyObject *__pyx_n_s_ExponentTooLarge;
  PyObject *__pyx_n_s_FIELD;
  PyObject *__pyx_n_u_False;
  PyObject *__pyx_kp_u_Field;
  PyObject *__pyx_n_s_FieldAccessor;
  PyObject *__pyx_n_s_FieldAccessor___reduce_cython;
  PyObject *__pyx_n_s_FieldAccessor___setstate_cython;
  PyObject *__pyx_n_s_FileReader;
  PyObject *__pyx_n_s_FileReader___reduce_cython;
  PyObject *__pyx_n_s_FileReader___setstate_cython;
  PyObject *__pyx_n_s_FileWriter;
  PyObject *__pyx_n_s_FileWriter___reduce_cython;
  PyObject *__pyx_n_s_FileWriter___setstate_cython;
  PyObject *__pyx_n_s_FixedType;
  PyObject *__pyx_n_s_FixedType___reduce_cython;
  PyObject *__pyx_n_s_FixedType___setstate_cython;
  PyObject *__pyx_n_s_FixedType__convert_value;
  PyObject *__pyx_n_s_FixedType__get_schema_extra;
  PyObject *__pyx_n_s_FixedType_copy;
  PyObject *__pyx_n_s_FloatType;
  PyObject *__pyx_n_s_FloatType___reduce_cython;
  PyObject *__pyx_n_s_FloatType___setstate_cython;
  PyObject *__pyx_n_s_FloatType__convert_value;
  PyObject *__pyx_n_s_FloatType__get_schema_extra;
  PyObject *__pyx_n_s_FloatType_copy;
  PyObject *__pyx_n_s_HAVE_BZIP2;
  PyObject *__pyx_n_s_HAVE_LZ4;
  PyObject *__pyx_n_s_HAVE_SNAPPY;
  PyObject *__pyx_n_s_HAVE_XZ;
  PyObject *__pyx_n_s_HAVE_ZLIB;
  PyObject *__pyx_n_s_HAVE_ZSTD;
  PyObject *__pyx_n_s_IGNORE;
  PyObject *__pyx_n_s_ImportError;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_25;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_26;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_27;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_28;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9;
  PyObject *__pyx_n_s_IndexError;
  PyObject *__pyx_kp_s_Index_out_of_bounds_axis_d;
  PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
  PyObject *__pyx_n_s_IntType;
  PyObject *__pyx_n_s_IntType___reduce_cython;
  PyObject *__pyx_n_s_IntType___setstate_cython;
  PyObject *__pyx_n_s_IntType__convert_value;
  PyObject *__pyx_n_s_IntType__get_schema_extra;
  PyObject *__pyx_n_s_IntType_copy;
  PyObject *__pyx_n_s_InvalidHasher;
  PyObject *__pyx_n_s_InvalidName;
  PyObject *__pyx_n_s_InvalidValue;
  PyObject *__pyx_n_s_InvalidValue___init;
  PyObject *__pyx_n_s_InvalidValue___str;
  PyObject *__pyx_n_s_InvalidValue___str___locals_gene;
  PyObject *__pyx_kp_u_Invalid_block_sync_marker_expect;
  PyObject *__pyx_kp_u_Invalid_file_header_expected;
  PyObject *__pyx_kp_u_Invalid_length_for_fixed_field;
  PyObject *__pyx_kp_u_Invalid_mode_expected_c_or_fortr;
  PyObject *__pyx_kp_u_Invalid_negative_size_for_fixed;
  PyObject *__pyx_kp_u_Invalid_non_bytes_value_for_fixe;
  PyObject *__pyx_kp_u_Invalid_non_str_value_for_fixed;
  PyObject *__pyx_kp_u_Invalid_shape_in_axis;
  PyObject *__pyx_kp_u_Invalid_union_decodes_to_option;
  PyObject *__pyx_kp_u_Invalid_value;
  PyObject *__pyx_n_s_Iterable;
  PyObject *__pyx_n_s_KeyError;
  PyObject *__pyx_n_s_LOGICAL_TYPES;
  PyObject *__pyx_n_u_Ll;
  PyObject *__pyx_n_s_LogicalType;
  PyObject *__pyx_n_s_LogicalType___reduce_cython;
  PyObject *__pyx_n_s_LogicalType___setstate_cython;
  PyObject *__pyx_n_s_LogicalType_for_type;
  PyObject *__pyx_n_s_LongType;
  PyObject *__pyx_n_s_LongType___reduce_cython;
  PyObject *__pyx_n_s_LongType___setstate_cython;
  PyObject *__pyx_n_s_LongType__convert_value;
  PyObject *__pyx_n_s_LongType__get_schema_extra;
  PyObject *__pyx_n_s_LongType_copy;
  PyObject *__pyx_n_u_Lu;
  PyObject *__pyx_n_s_Lz4Codec;
  PyObject *__pyx_n_s_Lz4Codec___reduce_cython;
  PyObject *__pyx_n_s_Lz4Codec___setstate_cython;
  PyObject *__pyx_n_s_LzmaCodec;
  PyObject *__pyx_n_s_LzmaCodec___reduce_cython;
  PyObject *__pyx_n_s_LzmaCodec___setstate_cython;
  PyObject *__pyx_n_s_MAX_PREC;
  PyObject *__pyx_n_s_MAX_TIME;
  PyObject *__pyx_n_s_MISSING_VALUE;
  PyObject *__pyx_n_u_MISSING_VALUE;
  PyObject *__pyx_n_s_MapType;
  PyObject *__pyx_n_s_MapType___reduce_cython;
  PyObject *__pyx_n_s_MapType___setstate_cython;
  PyObject *__pyx_n_s_MapType__get_schema_extra;
  PyObject *__pyx_n_s_MapType_convert_value;
  PyObject *__pyx_n_s_MapType_copy;
  PyObject *__pyx_n_s_MapType_walk_types;
  PyObject *__pyx_n_s_MappingProxyType;
  PyObject *__pyx_kp_u_Marker_must_be_bytes_got;
  PyObject *__pyx_kp_u_Marker_must_be_exactly_16_bytes;
  PyObject *__pyx_kp_u_Max_precision_is;
  PyObject *__pyx_n_s_MemoryError;
  PyObject *__pyx_n_s_MemoryReader;
  PyObject *__pyx_n_s_MemoryReader___reduce_cython;
  PyObject *__pyx_n_s_MemoryReader___setstate_cython;
  PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
  PyObject *__pyx_kp_s_MemoryView_of_r_object;
  PyObject *__pyx_n_s_MemoryWriter;
  PyObject *__pyx_n_s_MemoryWriter___reduce_cython;
  PyObject *__pyx_n_s_MemoryWriter___setstate_cython;
  PyObject *__pyx_n_s_NAME_RE_STRICT;
  PyObject *__pyx_n_s_NAME_RE_UNICODE;
  PyObject *__pyx_n_s_NO_DEFAULT;
  PyObject *__pyx_n_u_NO_DEFAULT;
  PyObject *__pyx_kp_u_Name;
  PyObject *__pyx_n_s_NamedType;
  PyObject *__pyx_n_s_NamedType___reduce_cython;
  PyObject *__pyx_n_s_NamedType___setstate_cython;
  PyObject *__pyx_n_s_NamedType__get_schema_extra;
  PyObject *__pyx_n_s_NamedType_get_namespaced_aliase;
  PyObject *__pyx_kp_u_Namespace;
  PyObject *__pyx_kp_u_None;
  PyObject *__pyx_n_s_NotImplemented;
  PyObject *__pyx_n_s_NotImplementedError;
  PyObject *__pyx_kp_u_Not_enough_input_data_to_read_va;
  PyObject *__pyx_n_s_NullCodec;
  PyObject *__pyx_n_s_NullCodec___reduce_cython;
  PyObject *__pyx_n_s_NullCodec___setstate_cython;
  PyObject *__pyx_n_s_NullType;
  PyObject *__pyx_n_s_NullType___reduce_cython;
  PyObject *__pyx_n_s_NullType___setstate_cython;
  PyObject *__pyx_n_s_NullType__convert_value;
  PyObject *__pyx_n_s_NullType__get_schema_extra;
  PyObject *__pyx_n_s_NullType_copy;
  PyObject *__pyx_n_b_O;
  PyObject *__pyx_n_s_OBJ_FILE_METADATA;
  PyObject *__pyx_n_s_OBJ_MAGIC_BYTES;
  PyObject *__pyx_kp_b_Obj;
  PyObject *__pyx_n_s_Options;
  PyObject *__pyx_n_s_OptionsWrapper;
  PyObject *__pyx_n_s_Options___reduce_cython;
  PyObject *__pyx_n_s_Options___setstate_cython;
  PyObject *__pyx_n_s_Options_diff;
  PyObject *__pyx_n_s_Options_equals;
  PyObject *__pyx_n_s_Options_replace;
  PyObject *__pyx_n_s_Options_with_external_types;
  PyObject *__pyx_n_s_Options_with_logical_types;
  PyObject *__pyx_n_s_Order;
  PyObject *__pyx_kp_u_Out_of_bounds_on_buffer_access_a;
  PyObject *__pyx_n_s_OverflowError;
  PyObject *__pyx_n_s_PERMISSIVE_OPTIONS;
  PyObject *__pyx_n_s_PRIMITIVE_TYPES;
  PyObject *__pyx_n_s_Path;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_n_s_PlaceholderType;
  PyObject *__pyx_n_s_PlaceholderType___reduce_cython;
  PyObject *__pyx_n_s_PlaceholderType___setstate_cyth;
  PyObject *__pyx_n_s_PlaceholderType__convert_value;
  PyObject *__pyx_n_s_PlaceholderType__get_schema_ext;
  PyObject *__pyx_kp_u_Placeholder_types_cannot_be_enco;
  PyObject *__pyx_kp_u_Placeholder_types_have_no_canoni;
  PyObject *__pyx_kp_u_Precision;
  PyObject *__pyx_kp_u_Precision_is_too_large_for_fixed;
  PyObject *__pyx_kp_u_Precision_must_be_an_integer_bet;
  PyObject *__pyx_kp_u_Precision_must_be_greater_than_s;
  PyObject *__pyx_n_s_PromoteBytesToString;
  PyObject *__pyx_n_s_PromoteBytesToString___reduce_cy;
  PyObject *__pyx_n_s_PromoteBytesToString___setstate;
  PyObject *__pyx_n_s_PromoteStringToBytes;
  PyObject *__pyx_n_s_PromoteStringToBytes___reduce_cy;
  PyObject *__pyx_n_s_PromoteStringToBytes___setstate;
  PyObject *__pyx_n_s_PromoteToFloat;
  PyObject *__pyx_n_s_PromoteToFloat___reduce_cython;
  PyObject *__pyx_n_s_PromoteToFloat___setstate_cython;
  PyObject *__pyx_n_s_PromotingEnumType;
  PyObject *__pyx_n_s_PromotingEnumType___reduce_cytho;
  PyObject *__pyx_n_s_PromotingEnumType___setstate_cyt;
  PyObject *__pyx_n_s_PromotingRecordType;
  PyObject *__pyx_n_s_PromotingRecordType___reduce_cyt;
  PyObject *__pyx_n_s_PromotingRecordType___setstate_c;
  PyObject *__pyx_kp_u_Promoting_schemas_reader_writer;
  PyObject *__pyx_n_s_RAW_VALUES;
  PyObject *__pyx_n_s_Rabin;
  PyObject *__pyx_n_s_Rabin___reduce_cython;
  PyObject *__pyx_n_s_Rabin___setstate_cython;
  PyObject *__pyx_n_s_Rabin_copy;
  PyObject *__pyx_n_s_Rabin_digest;
  PyObject *__pyx_n_s_Rabin_hexdigest;
  PyObject *__pyx_n_s_Rabin_update;
  PyObject *__pyx_n_s_Reader;
  PyObject *__pyx_n_s_Reader___reduce_cython;
  PyObject *__pyx_n_s_Reader___setstate_cython;
  PyObject *__pyx_kp_u_Record;
  PyObject *__pyx_n_s_RecordField;
  PyObject *__pyx_n_s_RecordField___reduce_cython;
  PyObject *__pyx_n_s_RecordField___setstate_cython;
  PyObject *__pyx_n_s_RecordField_get_schema;
  PyObject *__pyx_n_s_RecordType;
  PyObject *__pyx_n_s_RecordType___reduce_cython;
  PyObject *__pyx_n_s_RecordType___setstate_cython;
  PyObject *__pyx_n_s_RecordType__convert_value;
  PyObject *__pyx_n_s_RecordType__get_schema_extra;
  PyObject *__pyx_n_s_RecordType_copy;
  PyObject *__pyx_n_s_RecordType_walk_types;
  PyObject *__pyx_kp_u_Record_2;
  PyObject *__pyx_n_s_Record_3;
  PyObject *__pyx_n_s_Record___dir;
  PyObject *__pyx_n_s_Record___reduce_cython;
  PyObject *__pyx_n_s_Record___setstate_cython;
  PyObject *__pyx_n_s_Record__asdict;
  PyObject *__pyx_n_s_Record__repr_children_locals_gen;
  PyObject *__pyx_kp_u_Records_may_either_be_instantiat;
  PyObject *__pyx_kp_u_Records_may_either_be_instantiat_2;
  PyObject *__pyx_n_s_ResolvedSchema;
  PyObject *__pyx_n_s_ResolvedSchema___reduce_cython;
  PyObject *__pyx_n_s_ResolvedSchema___setstate_cython;
  PyObject *__pyx_n_s_ResourceWarning;
  PyObject *__pyx_n_s_SLICE_TYPE;
  PyObject *__pyx_n_s_SSIZE_TYPECODE;
  PyObject *__pyx_kp_u_Scale_must_be_a_positive_integer;
  PyObject *__pyx_n_s_Schema;
  PyObject *__pyx_n_s_Schema___reduce_cython;
  PyObject *__pyx_n_s_Schema___setstate_cython;
  PyObject *__pyx_n_s_Schema__make_logical_types;
  PyObject *__pyx_n_s_Schema__wrap_type;
  PyObject *__pyx_n_s_Schema_binary_decode;
  PyObject *__pyx_n_s_Schema_binary_encode;
  PyObject *__pyx_n_s_Schema_binary_read;
  PyObject *__pyx_n_s_Schema_binary_write;
  PyObject *__pyx_n_s_Schema_can_encode;
  PyObject *__pyx_n_s_Schema_find_type;
  PyObject *__pyx_n_s_Schema_fingerprint;
  PyObject *__pyx_kp_u_Schema_is_required;
  PyObject *__pyx_n_s_Schema_json_decode;
  PyObject *__pyx_n_s_Schema_json_encode;
  PyObject *__pyx_n_s_Schema_reader_for_writer;
  PyObject *__pyx_n_s_Sentinel;
  PyObject *__pyx_n_s_Sentinel___reduce_cython;
  PyObject *__pyx_n_s_Sentinel___setstate_cython;
  PyObject *__pyx_n_s_Sequence;
  PyObject *__pyx_kp_s_Sequence_str;
  PyObject *__pyx_n_s_SignatureWrapper;
  PyObject *__pyx_n_s_SignatureWrapper___init;
  PyObject *__pyx_n_s_SignatureWrapper__set_from_data;
  PyObject *__pyx_n_s_SnappyCodec;
  PyObject *__pyx_n_s_SnappyCodec___reduce_cython;
  PyObject *__pyx_n_s_SnappyCodec___setstate_cython;
  PyObject *__pyx_kp_s_Step_may_not_be_zero_axis_d;
  PyObject *__pyx_n_s_StopIteration;
  PyObject *__pyx_n_s_StringType;
  PyObject *__pyx_n_s_StringType___reduce_cython;
  PyObject *__pyx_n_s_StringType___setstate_cython;
  PyObject *__pyx_n_s_StringType__convert_value;
  PyObject *__pyx_n_s_StringType__get_schema_extra;
  PyObject *__pyx_n_s_StringType_copy;
  PyObject *__pyx_n_s_TYPES_BY_NAME;
  PyObject *__pyx_n_s_TYPE_TUPLE_ALWAYS;
  PyObject *__pyx_n_s_TYPE_TUPLE_IF_AMBIGUOUS;
  PyObject *__pyx_n_s_TYPE_TUPLE_IF_RECORD;
  PyObject *__pyx_n_s_TYPE_TUPLE_IF_RECORD_AMBIGUOUS;
  PyObject *__pyx_kp_s_The_exponent_of_a_decimal_value;
  PyObject *__pyx_kp_u_The_null_namespace_cannot_be_spe;
  PyObject *__pyx_kp_s_The_order_of_a_field_in_a_recor;
  PyObject *__pyx_kp_s_The_schema_contains_a_type_or_e;
  PyObject *__pyx_kp_s_The_schema_contains_an_unexptec;
  PyObject *__pyx_n_s_TimeMicros;
  PyObject *__pyx_n_s_TimeMicros___reduce_cython;
  PyObject *__pyx_n_s_TimeMicros___setstate_cython;
  PyObject *__pyx_n_s_TimeMicros__for_type;
  PyObject *__pyx_n_s_TimeMillis;
  PyObject *__pyx_n_s_TimeMillis___reduce_cython;
  PyObject *__pyx_n_s_TimeMillis___setstate_cython;
  PyObject *__pyx_n_s_TimeMillis__for_type;
  PyObject *__pyx_n_s_TimestampMicros;
  PyObject *__pyx_n_s_TimestampMicros___reduce_cython;
  PyObject *__pyx_n_s_TimestampMicros___setstate_cytho;
  PyObject *__pyx_n_s_TimestampMicros__for_type;
  PyObject *__pyx_n_s_TimestampMillis;
  PyObject *__pyx_n_s_TimestampMillis___reduce_cython;
  PyObject *__pyx_n_s_TimestampMillis___setstate_cytho;
  PyObject *__pyx_n_s_TimestampMillis__for_type;
  PyObject *__pyx_n_u_True;
  PyObject *__pyx_kp_u_Trying_to_close_a_closed_Contain;
  PyObject *__pyx_kp_u_Trying_to_write_to_closed_Contai;
  PyObject *__pyx_n_s_Type;
  PyObject *__pyx_n_u_Type;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_kp_u_Type_name;
  PyObject *__pyx_n_u_UNKNOWN;
  PyObject *__pyx_n_s_UUID;
  PyObject *__pyx_n_s_UUIDBase;
  PyObject *__pyx_n_s_UUIDBase___reduce_cython;
  PyObject *__pyx_n_s_UUIDBase___setstate_cython;
  PyObject *__pyx_n_s_UUIDBase__for_type;
  PyObject *__pyx_n_s_UUIDFixedType;
  PyObject *__pyx_n_s_UUIDFixedType___reduce_cython;
  PyObject *__pyx_n_s_UUIDFixedType___setstate_cython;
  PyObject *__pyx_n_s_UUIDFixedType__for_type;
  PyObject *__pyx_n_s_UUIDStringType;
  PyObject *__pyx_n_s_UUIDStringType___reduce_cython;
  PyObject *__pyx_n_s_UUIDStringType___setstate_cython;
  PyObject *__pyx_n_s_UUIDStringType__for_type;
  PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
  PyObject *__pyx_n_s_UnicodeEncodeError;
  PyObject *__pyx_n_s_Union;
  PyObject *__pyx_n_s_UnionDecodeOption;
  PyObject *__pyx_n_s_UnionType;
  PyObject *__pyx_n_s_UnionType___reduce_cython;
  PyObject *__pyx_n_s_UnionType___setstate_cython;
  PyObject *__pyx_n_s_UnionType_convert_value;
  PyObject *__pyx_n_s_UnionType_copy;
  PyObject *__pyx_n_s_UnionType_get_schema;
  PyObject *__pyx_n_s_UnionType_walk_types;
  PyObject *__pyx_kp_s_Union_bytes_hashlib__hashlib_HAS;
  PyObject *__pyx_kp_s_Union_str_object;
  PyObject *__pyx_kp_u_Unions_may_not_have_more_than_on;
  PyObject *__pyx_kp_u_Unions_may_not_immediately_conta;
  PyObject *__pyx_kp_u_Unions_must_contain_at_least_one;
  PyObject *__pyx_n_s_UnknownType;
  PyObject *__pyx_n_s_UnknownType___init;
  PyObject *__pyx_kp_u_Unknown_hash_method;
  PyObject *__pyx_kp_u_Unknown_type;
  PyObject *__pyx_kp_u_Unknown_value_for_enum;
  PyObject *__pyx_kp_u_Unsupported_codec;
  PyObject *__pyx_kp_u_Value;
  PyObject *__pyx_n_s_ValueAdapter;
  PyObject *__pyx_n_s_ValueAdapter___reduce_cython;
  PyObject *__pyx_n_s_ValueAdapter___setstate_cython;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_n_s_View_MemoryView;
  PyObject *__pyx_n_s_Writer;
  PyObject *__pyx_n_s_Writer___reduce_cython;
  PyObject *__pyx_n_s_Writer___setstate_cython;
  PyObject *__pyx_kp_u_Y_m_d;
  PyObject *__pyx_n_s_ZStandardCodec;
  PyObject *__pyx_n_s_ZStandardCodec___reduce_cython;
  PyObject *__pyx_n_s_ZStandardCodec___setstate_cytho;
  PyObject *__pyx_n_s_ZstdCompressor;
  PyObject *__pyx_n_s_ZstdDecompressor;
  PyObject *__pyx_n_s__122;
  PyObject *__pyx_kp_b__13;
  PyObject *__pyx_kp_u__13;
  PyObject *__pyx_kp_u__14;
  PyObject *__pyx_n_s__152;
  PyObject *__pyx_kp_s__18;
  PyObject *__pyx_kp_u__18;
  PyObject *__pyx_kp_u__2;
  PyObject *__pyx_kp_u__24;
  PyObject *__pyx_kp_u__25;
  PyObject *__pyx_kp_u__26;
  PyObject *__pyx_kp_u__27;
  PyObject *__pyx_n_s__3;
  PyObject *__pyx_kp_b__34;
  PyObject *__pyx_kp_u__48;
  PyObject *__pyx_kp_u__6;
  PyObject *__pyx_kp_u__60;
  PyObject *__pyx_kp_u__61;
  PyObject *__pyx_kp_u__62;
  PyObject *__pyx_kp_u__63;
  PyObject *__pyx_kp_u__64;
  PyObject *__pyx_kp_u__7;
  PyObject *__pyx_kp_u__70;
  PyObject *__pyx_kp_b__72;
  PyObject *__pyx_kp_u__76;
  PyObject *__pyx_kp_u__77;
  PyObject *__pyx_kp_u__78;
  PyObject *__pyx_kp_u__79;
  PyObject *__pyx_kp_u__80;
  PyObject *__pyx_kp_u__9;
  PyObject *__pyx_n_s_a;
  PyObject *__pyx_n_b_abacus;
  PyObject *__pyx_n_b_abacuscounter1;
  PyObject *__pyx_n_s_abc;
  PyObject *__pyx_n_s_accepts_string;
  PyObject *__pyx_n_s_actual;
  PyObject *__pyx_n_s_adapt_record_types;
  PyObject *__pyx_kp_u_adapt_record_types_2;
  PyObject *__pyx_n_s_add;
  PyObject *__pyx_n_u_aliases;
  PyObject *__pyx_n_s_allocate_buffer;
  PyObject *__pyx_n_s_allow_aliases_to_be_string;
  PyObject *__pyx_kp_u_allow_aliases_to_be_string_2;
  PyObject *__pyx_n_s_allow_duplicate_union_types;
  PyObject *__pyx_kp_u_allow_duplicate_union_types_2;
  PyObject *__pyx_n_s_allow_empty_unions;
  PyObject *__pyx_kp_u_allow_empty_unions_2;
  PyObject *__pyx_n_s_allow_error_type;
  PyObject *__pyx_kp_u_allow_error_type_2;
  PyObject *__pyx_n_s_allow_false_values_for_null;
  PyObject *__pyx_kp_u_allow_false_values_for_null_2;
  PyObject *__pyx_n_s_allow_invalid_default_values;
  PyObject *__pyx_kp_u_allow_invalid_default_values_2;
  PyObject *__pyx_n_s_allow_leading_dot_in_names;
  PyObject *__pyx_kp_u_allow_leading_dot_in_names_2;
  PyObject *__pyx_n_s_allow_nested_unions;
  PyObject *__pyx_kp_u_allow_nested_unions_2;
  PyObject *__pyx_n_s_allow_primitive_name_collision;
  PyObject *__pyx_kp_u_allow_primitive_name_collision_2;
  PyObject *__pyx_kp_u_allow_primitive_names_in_namesp;
  PyObject *__pyx_n_s_allow_primitive_names_in_namespa;
  PyObject *__pyx_n_s_allow_tuple_notation;
  PyObject *__pyx_kp_u_allow_tuple_notation_2;
  PyObject *__pyx_n_s_allow_union_default_any_member;
  PyObject *__pyx_kp_u_allow_union_default_any_member_2;
  PyObject *__pyx_n_s_alternate_timestamp_encoding;
  PyObject *__pyx_kp_u_alternate_timestamp_millis_enco;
  PyObject *__pyx_n_s_alternate_timestamp_millis_encod;
  PyObject *__pyx_kp_u_and;
  PyObject *__pyx_n_s_annotations;
  PyObject *__pyx_n_s_annotations_2;
  PyObject *__pyx_kp_u_appears_multiple_times_in_schem;
  PyObject *__pyx_n_s_append;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_array;
  PyObject *__pyx_n_u_array;
  PyObject *__pyx_n_s_as_tuple;
  PyObject *__pyx_n_u_ascending;
  PyObject *__pyx_n_s_ascii_name_rules;
  PyObject *__pyx_kp_u_ascii_name_rules_2;
  PyObject *__pyx_n_s_asdict;
  PyObject *__pyx_n_u_asdict;
  PyObject *__pyx_n_s_astimezone;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_kp_u_at;
  PyObject *__pyx_kp_u_avro_codec;
  PyObject *__pyx_kp_u_avro_schema;
  PyObject *__pyx_n_s_avro_type;
  PyObject *__pyx_n_s_b;
  PyObject *__pyx_n_s_base;
  PyObject *__pyx_n_u_big;
  PyObject *__pyx_n_s_bin;
  PyObject *__pyx_n_s_binary_decode;
  PyObject *__pyx_n_s_binary_encode;
  PyObject *__pyx_n_s_binary_read;
  PyObject *__pyx_n_s_binary_write;
  PyObject *__pyx_n_s_bint;
  PyObject *__pyx_n_u_bint;
  PyObject *__pyx_n_s_bit_length;
  PyObject *__pyx_n_s_block_size;
  PyObject *__pyx_n_s_bool;
  PyObject *__pyx_n_u_bool;
  PyObject *__pyx_n_s_bool_2;
  PyObject *__pyx_kp_u_boolean;
  PyObject *__pyx_n_u_boolean_2;
  PyObject *__pyx_n_s_buffer;
  PyObject *__pyx_n_s_byteorder;
  PyObject *__pyx_kp_u_bytes;
  PyObject *__pyx_kp_u_bytes_2;
  PyObject *__pyx_n_s_bytes_3;
  PyObject *__pyx_n_u_bytes_3;
  PyObject *__pyx_n_s_bytes_codec;
  PyObject *__pyx_kp_u_bytes_codec_2;
  PyObject *__pyx_n_s_bytes_default_value_utf8;
  PyObject *__pyx_kp_u_bytes_default_value_utf8_2;
  PyObject *__pyx_n_s_bz2;
  PyObject *__pyx_n_b_bzip2;
  PyObject *__pyx_n_u_bzip2;
  PyObject *__pyx_n_s_c;
  PyObject *__pyx_n_u_c;
  PyObject *__pyx_n_s_can_encode;
  PyObject *__pyx_kp_u_cannot_be_adapted_to;
  PyObject *__pyx_n_s_canonical_form;
  PyObject *__pyx_n_s_canonical_form_repeat_enum;
  PyObject *__pyx_kp_u_canonical_form_repeat_enum_2;
  PyObject *__pyx_n_s_canonical_form_repeat_fixed;
  PyObject *__pyx_kp_u_canonical_form_repeat_fixed_2;
  PyObject *__pyx_n_s_category;
  PyObject *__pyx_n_s_cavro;
  PyObject *__pyx_kp_s_cavro_pyx;
  PyObject *__pyx_n_s_changes;
  PyObject *__pyx_n_s_check_exp_overflow;
  PyObject *__pyx_n_s_check_value;
  PyObject *__pyx_n_s_chr;
  PyObject *__pyx_n_s_clamp;
  PyObject *__pyx_n_s_clamp_float_overflow;
  PyObject *__pyx_kp_u_clamp_float_overflow_2;
  PyObject *__pyx_n_s_clamp_int_overflow;
  PyObject *__pyx_kp_u_clamp_int_overflow_2;
  PyObject *__pyx_n_s_class;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_class_inst_method;
  PyObject *__pyx_n_s_class_inst_method___get;
  PyObject *__pyx_n_s_class_inst_method___init;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_s_cls;
  PyObject *__pyx_n_s_codec;
  PyObject *__pyx_n_s_coerce_int_to_float;
  PyObject *__pyx_kp_u_coerce_int_to_float_2;
  PyObject *__pyx_n_s_coerce_values_to_boolean;
  PyObject *__pyx_kp_u_coerce_values_to_boolean_2;
  PyObject *__pyx_n_s_coerce_values_to_float;
  PyObject *__pyx_kp_u_coerce_values_to_float_2;
  PyObject *__pyx_n_s_coerce_values_to_int;
  PyObject *__pyx_kp_u_coerce_values_to_int_2;
  PyObject *__pyx_n_s_coerce_values_to_str;
  PyObject *__pyx_kp_u_coerce_values_to_str_2;
  PyObject *__pyx_n_s_collections;
  PyObject *__pyx_kp_s_collections_abc;
  PyObject *__pyx_n_s_compare;
  PyObject *__pyx_n_s_compile;
  PyObject *__pyx_n_s_compress;
  PyObject *__pyx_n_s_container_fill_blocks;
  PyObject *__pyx_kp_u_container_fill_blocks_2;
  PyObject *__pyx_n_s_context;
  PyObject *__pyx_kp_s_contiguous_and_direct;
  PyObject *__pyx_kp_s_contiguous_and_indirect;
  PyObject *__pyx_n_s_convert_value;
  PyObject *__pyx_n_s_convert_value_2;
  PyObject *__pyx_n_s_copy;
  PyObject *__pyx_n_s_copy_time;
  PyObject *__pyx_n_s_count;
  PyObject *__pyx_n_b_counter;
  PyObject *__pyx_n_s_crc32;
  PyObject *__pyx_n_s_create_decimal;
  PyObject *__pyx_n_s_created;
  PyObject *__pyx_n_s_custom_decode_value;
  PyObject *__pyx_n_s_custom_encode_value;
  PyObject *__pyx_n_s_data;
  PyObject *__pyx_n_s_dataclass_fields;
  PyObject *__pyx_n_s_dataclass_params;
  PyObject *__pyx_n_s_dataclasses;
  PyObject *__pyx_n_s_date;
  PyObject *__pyx_n_u_date;
  PyObject *__pyx_n_s_date_type_accepts_string;
  PyObject *__pyx_kp_u_date_type_accepts_string_2;
  PyObject *__pyx_n_s_datetime;
  PyObject *__pyx_n_s_days;
  PyObject *__pyx_n_s_decimal;
  PyObject *__pyx_n_u_decimal;
  PyObject *__pyx_kp_u_decimal_2;
  PyObject *__pyx_n_s_decimal_check_exp_overflow;
  PyObject *__pyx_kp_u_decimal_check_exp_overflow_2;
  PyObject *__pyx_n_s_decode;
  PyObject *__pyx_n_s_decompress;
  PyObject *__pyx_n_s_decompressobj;
  PyObject *__pyx_n_s_defaul_value;
  PyObject *__pyx_n_s_default;
  PyObject *__pyx_n_u_default;
  PyObject *__pyx_n_s_default_value;
  PyObject *__pyx_n_s_defer_schema_promotion_errors;
  PyObject *__pyx_kp_u_defer_schema_promotion_errors_2;
  PyObject *__pyx_n_b_deflate;
  PyObject *__pyx_n_u_deflate;
  PyObject *__pyx_n_u_descending;
  PyObject *__pyx_n_s_deserialize;
  PyObject *__pyx_n_s_dest;
  PyObject *__pyx_n_s_dest_type;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dict_2;
  PyObject *__pyx_kp_s_dict_str_AvroType;
  PyObject *__pyx_kp_s_dict_str_tuple_object_object;
  PyObject *__pyx_n_s_dict_to_canonical_locals_genexpr;
  PyObject *__pyx_n_s_diff;
  PyObject *__pyx_n_s_digest;
  PyObject *__pyx_n_s_digest_size;
  PyObject *__pyx_n_s_dir;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_u_doc;
  PyObject *__pyx_n_s_doc_2;
  PyObject *__pyx_kp_u_does_not_implement__binary_buff;
  PyObject *__pyx_kp_u_does_not_implement__convert_val;
  PyObject *__pyx_kp_u_does_not_implement__extract_met;
  PyObject *__pyx_kp_u_does_not_implement__get_schema;
  PyObject *__pyx_kp_u_does_not_implement__get_value_f;
  PyObject *__pyx_kp_u_does_not_implement__json_decode;
  PyObject *__pyx_kp_u_does_not_implement__json_format;
  PyObject *__pyx_kp_u_does_not_implement_binary_buffe;
  PyObject *__pyx_kp_u_does_not_implement_canonical_fo;
  PyObject *__pyx_kp_u_does_not_implement_clone;
  PyObject *__pyx_kp_u_does_not_implement_decode_value;
  PyObject *__pyx_kp_u_does_not_implement_encode_value;
  PyObject *__pyx_kp_u_does_not_implement_read_block;
  PyObject *__pyx_kp_u_does_not_implement_read_bytes;
  PyObject *__pyx_kp_u_does_not_implement_read_n;
  PyObject *__pyx_kp_u_does_not_implement_read_u8;
  PyObject *__pyx_kp_u_does_not_implement_write_block;
  PyObject *__pyx_kp_u_does_not_implement_write_n;
  PyObject *__pyx_kp_u_does_not_implement_write_u8;
  PyObject *__pyx_kp_u_double;
  PyObject *__pyx_n_u_double_2;
  PyObject *__pyx_n_s_dtype_is_object;
  PyObject *__pyx_n_s_dumps;
  PyObject *__pyx_n_s_e;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_encode;
  PyObject *__pyx_n_s_enforce_enum_symbol_name_rules;
  PyObject *__pyx_kp_u_enforce_enum_symbol_name_rules_2;
  PyObject *__pyx_n_s_enforce_namespace_name_rules;
  PyObject *__pyx_kp_u_enforce_namespace_name_rules_2;
  PyObject *__pyx_n_s_enforce_type_name_rules;
  PyObject *__pyx_kp_u_enforce_type_name_rules_2;
  PyObject *__pyx_n_s_ensure_ascii;
  PyObject *__pyx_n_s_enter;
  PyObject *__pyx_n_s_enum;
  PyObject *__pyx_n_u_enum;
  PyObject *__pyx_n_s_enum_symbols_must_be_unique;
  PyObject *__pyx_kp_u_enum_symbols_must_be_unique_2;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_eq;
  PyObject *__pyx_n_s_equals;
  PyObject *__pyx_n_s_error;
  PyObject *__pyx_n_u_error;
  PyObject *__pyx_n_s_errors;
  PyObject *__pyx_n_s_escape;
  PyObject *__pyx_n_s_eval;
  PyObject *__pyx_n_s_eval_2;
  PyObject *__pyx_n_s_exc_tb;
  PyObject *__pyx_n_s_exc_type;
  PyObject *__pyx_n_s_exc_val;
  PyObject *__pyx_n_s_exit;
  PyObject *__pyx_n_s_expand_types_in_schema;
  PyObject *__pyx_kp_u_expand_types_in_schema_2;
  PyObject *__pyx_n_s_externally_defined_types;
  PyObject *__pyx_kp_u_externally_defined_types_2;
  PyObject *__pyx_n_s_extra;
  PyObject *__pyx_n_s_f;
  PyObject *__pyx_n_s_field;
  PyObject *__pyx_n_s_field_to_index;
  PyObject *__pyx_n_u_field_to_index;
  PyObject *__pyx_n_s_field_type;
  PyObject *__pyx_n_s_fields;
  PyObject *__pyx_n_u_fields;
  PyObject *__pyx_n_s_file_obj;
  PyObject *__pyx_n_s_find_type;
  PyObject *__pyx_n_s_fingerprint;
  PyObject *__pyx_n_s_fingerprint_returns_digest;
  PyObject *__pyx_kp_u_fingerprint_returns_digest_2;
  PyObject *__pyx_n_s_fitness;
  PyObject *__pyx_n_s_fixed;
  PyObject *__pyx_n_u_fixed;
  PyObject *__pyx_n_s_fixed_codec;
  PyObject *__pyx_kp_u_fixed_codec_2;
  PyObject *__pyx_kp_u_fixed_uuid;
  PyObject *__pyx_n_s_flags;
  PyObject *__pyx_kp_u_float;
  PyObject *__pyx_n_s_float16;
  PyObject *__pyx_n_s_float32;
  PyObject *__pyx_n_s_float64;
  PyObject *__pyx_n_u_float_2;
  PyObject *__pyx_n_s_float_out_of_range_inf;
  PyObject *__pyx_kp_u_float_out_of_range_inf_2;
  PyObject *__pyx_n_s_floor;
  PyObject *__pyx_n_s_flush;
  PyObject *__pyx_n_s_fn;
  PyObject *__pyx_kp_u_for_field;
  PyObject *__pyx_n_s_for_schema;
  PyObject *__pyx_n_s_for_source;
  PyObject *__pyx_kp_u_for_type;
  PyObject *__pyx_n_s_for_type_2;
  PyObject *__pyx_n_s_for_type_3;
  PyObject *__pyx_n_s_force;
  PyObject *__pyx_n_s_format;
  PyObject *__pyx_n_s_fortran;
  PyObject *__pyx_n_u_fortran;
  PyObject *__pyx_kp_u_found_twice;
  PyObject *__pyx_n_s_frame;
  PyObject *__pyx_n_s_from_bytes;
  PyObject *__pyx_n_s_frozen;
  PyObject *__pyx_n_s_fullmatch;
  PyObject *__pyx_n_s_func;
  PyObject *__pyx_n_s_functools;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_genexpr;
  PyObject *__pyx_n_s_get;
  PyObject *__pyx_n_s_get_2;
  PyObject *__pyx_n_s_get_schema;
  PyObject *__pyx_n_s_get_schema_extra;
  PyObject *__pyx_kp_u_getfullargspec;
  PyObject *__pyx_n_u_getitem;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_s_given;
  PyObject *__pyx_kp_u_got;
  PyObject *__pyx_kp_u_got_2;
  PyObject *__pyx_kp_u_got_3;
  PyObject *__pyx_kp_u_got_4;
  PyObject *__pyx_kp_u_got_differing_extents_in_dimensi;
  PyObject *__pyx_kp_u_has_exponent;
  PyObject *__pyx_n_s_hash;
  PyObject *__pyx_n_s_hasher;
  PyObject *__pyx_n_s_hashlib;
  PyObject *__pyx_n_s_hex;
  PyObject *__pyx_n_s_hexdigest;
  PyObject *__pyx_n_s_hour;
  PyObject *__pyx_n_s_i;
  PyObject *__pyx_n_s_id;
  PyObject *__pyx_n_s_ignore;
  PyObject *__pyx_n_u_ignore;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_indent;
  PyObject *__pyx_n_s_index;
  PyObject *__pyx_kp_u_inf;
  PyObject *__pyx_n_u_inf_2;
  PyObject *__pyx_n_s_init;
  PyObject *__pyx_n_s_init_2;
  PyObject *__pyx_n_s_init___locals_genexpr;
  PyObject *__pyx_n_s_init_subclass;
  PyObject *__pyx_n_s_initial_size;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_inline_namespaces;
  PyObject *__pyx_kp_u_inline_namespaces_2;
  PyObject *__pyx_n_s_inspect;
  PyObject *__pyx_n_s_inst;
  PyObject *__pyx_kp_u_int;
  PyObject *__pyx_n_u_int_2;
  PyObject *__pyx_kp_u_int_date;
  PyObject *__pyx_kp_u_int_time_millis;
  PyObject *__pyx_n_s_integer;
  PyObject *__pyx_kp_u_invalid_for_enum;
  PyObject *__pyx_kp_u_invalid_value_include_array_ind;
  PyObject *__pyx_n_s_invalid_value_include_array_inde;
  PyObject *__pyx_n_s_invalid_value_include_map_key;
  PyObject *__pyx_kp_u_invalid_value_include_map_key_2;
  PyObject *__pyx_kp_u_invalid_value_includes_record_n;
  PyObject *__pyx_n_s_invalid_value_includes_record_na;
  PyObject *__pyx_n_s_io;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_kp_u_is_not_a_mapping;
  PyObject *__pyx_kp_u_is_not_a_valid_key_for_map_type;
  PyObject *__pyx_kp_u_is_not_a_valid_name;
  PyObject *__pyx_kp_u_is_not_a_valid_union_value_expe;
  PyObject *__pyx_kp_u_is_not_a_valid_union_value_unkn;
  PyObject *__pyx_kp_u_is_not_a_valid_value_for_array;
  PyObject *__pyx_kp_u_is_not_a_valid_value_for_map_ty;
  PyObject *__pyx_kp_u_is_not_allowed_as_a_name;
  PyObject *__pyx_kp_u_is_not_valid;
  PyObject *__pyx_kp_u_is_not_valid_for_a_union_of;
  PyObject *__pyx_kp_u_is_not_valid_for_bytes_field;
  PyObject *__pyx_kp_u_is_not_valid_for_field;
  PyObject *__pyx_kp_u_is_not_valid_for_fixed_field;
  PyObject *__pyx_kp_u_is_not_valid_for_union_with_fir;
  PyObject *__pyx_kp_u_is_required_for_record;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_isinf;
  PyObject *__pyx_n_s_isnan;
  PyObject *__pyx_n_s_item_type;
  PyObject *__pyx_n_s_items;
  PyObject *__pyx_n_u_items;
  PyObject *__pyx_kp_u_items_2;
  PyObject *__pyx_n_s_itemsize;
  PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
  PyObject *__pyx_n_s_json;
  PyObject *__pyx_n_s_json_decode;
  PyObject *__pyx_n_s_json_encode;
  PyObject *__pyx_n_s_keys;
  PyObject *__pyx_n_s_kw_only;
  PyObject *__pyx_n_s_kwargs;
  PyObject *__pyx_n_b_l;
  PyObject *__pyx_n_u_latin1;
  PyObject *__pyx_kp_u_latin_1;
  PyObject *__pyx_n_u_little;
  PyObject *__pyx_n_s_loads;
  PyObject *__pyx_n_s_local;
  PyObject *__pyx_n_s_log10;
  PyObject *__pyx_n_u_logicalType;
  PyObject *__pyx_n_s_logical_by_name;
  PyObject *__pyx_n_s_logical_name;
  PyObject *__pyx_n_s_logical_type;
  PyObject *__pyx_n_s_logical_types;
  PyObject *__pyx_kp_u_logical_types_2;
  PyObject *__pyx_kp_u_long;
  PyObject *__pyx_n_u_long_2;
  PyObject *__pyx_kp_u_long_time_micros;
  PyObject *__pyx_kp_u_long_timestamp_micros;
  PyObject *__pyx_kp_u_long_timestamp_millis;
  PyObject *__pyx_n_b_lz4;
  PyObject *__pyx_n_s_lz4;
  PyObject *__pyx_n_u_lz4;
  PyObject *__pyx_n_s_lz4_frame;
  PyObject *__pyx_n_s_lzma;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_make_dataclass;
  PyObject *__pyx_n_s_make_logical_types;
  PyObject *__pyx_n_s_map;
  PyObject *__pyx_n_u_map;
  PyObject *__pyx_n_s_marker;
  PyObject *__pyx_n_s_mask;
  PyObject *__pyx_n_s_match_args;
  PyObject *__pyx_n_s_math;
  PyObject *__pyx_n_s_max;
  PyObject *__pyx_n_s_max_blocksize;
  PyObject *__pyx_n_s_max_precision;
  PyObject *__pyx_n_s_maxunicode;
  PyObject *__pyx_n_s_memview;
  PyObject *__pyx_n_s_meta;
  PyObject *__pyx_n_s_metaclass;
  PyObject *__pyx_n_s_metadata;
  PyObject *__pyx_n_s_method;
  PyObject *__pyx_n_s_micros_to_time;
  PyObject *__pyx_n_s_microsecond;
  PyObject *__pyx_n_s_microseconds;
  PyObject *__pyx_n_s_minute;
  PyObject *__pyx_kp_u_missing;
  PyObject *__pyx_kp_u_missing_2;
  PyObject *__pyx_kp_u_missing_values_can_be_empty_con;
  PyObject *__pyx_n_s_missing_values_can_be_empty_cont;
  PyObject *__pyx_n_s_missing_values_can_be_null;
  PyObject *__pyx_kp_u_missing_values_can_be_null_2;
  PyObject *__pyx_n_s_mode;
  PyObject *__pyx_n_s_module;
  PyObject *__pyx_n_s_mro_entries;
  PyObject *__pyx_n_s_msg;
  PyObject *__pyx_n_s_n;
  PyObject *__pyx_n_s_n_bytes;
  PyObject *__pyx_n_s_naive_dt_assume_utc;
  PyObject *__pyx_kp_u_naive_dt_assume_utc_2;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_u_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_s_name_pattern;
  PyObject *__pyx_n_s_named_type_names_must_be_unique;
  PyObject *__pyx_kp_u_named_type_names_must_be_unique_2;
  PyObject *__pyx_n_s_named_types;
  PyObject *__pyx_n_s_namespace;
  PyObject *__pyx_n_u_namespace;
  PyObject *__pyx_n_s_namespaced;
  PyObject *__pyx_n_u_nan;
  PyObject *__pyx_n_s_ndarray;
  PyObject *__pyx_n_s_ndim;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_n_s_new_2;
  PyObject *__pyx_n_s_new_inst;
  PyObject *__pyx_n_s_new_types;
  PyObject *__pyx_n_s_new_ver;
  PyObject *__pyx_n_s_next_object;
  PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
  PyObject *__pyx_kp_u_not_compatible_with_NullType;
  PyObject *__pyx_kp_u_not_in_enum_symbols;
  PyObject *__pyx_kp_u_not_valid_for;
  PyObject *__pyx_kp_u_not_valid_for_long;
  PyObject *__pyx_n_s_np;
  PyObject *__pyx_n_s_np_f16;
  PyObject *__pyx_n_s_np_f32;
  PyObject *__pyx_n_s_np_f64;
  PyObject *__pyx_kp_u_ns;
  PyObject *__pyx_kp_u_ns_copy;
  PyObject *__pyx_kp_u_null;
  PyObject *__pyx_n_b_null_2;
  PyObject *__pyx_n_u_null_2;
  PyObject *__pyx_n_s_null_pad_fixed;
  PyObject *__pyx_kp_u_null_pad_fixed_2;
  PyObject *__pyx_n_s_numpy;
  PyObject *__pyx_n_s_ob;
  PyObject *__pyx_n_s_obj;
  PyObject *__pyx_n_s_object;
  PyObject *__pyx_n_s_objs;
  PyObject *__pyx_n_s_open;
  PyObject *__pyx_n_s_options;
  PyObject *__pyx_n_s_order;
  PyObject *__pyx_n_u_order;
  PyObject *__pyx_n_s_orig_value;
  PyObject *__pyx_n_s_os;
  PyObject *__pyx_n_s_other;
  PyObject *__pyx_n_s_other_val;
  PyObject *__pyx_n_s_out;
  PyObject *__pyx_n_s_out_state;
  PyObject *__pyx_n_s_pack;
  PyObject *__pyx_n_s_pad;
  PyObject *__pyx_n_s_parse_json;
  PyObject *__pyx_n_s_parse_namespace;
  PyObject *__pyx_n_s_partial;
  PyObject *__pyx_n_s_path;
  PyObject *__pyx_n_s_path_str;
  PyObject *__pyx_n_s_pathlib;
  PyObject *__pyx_n_s_perf;
  PyObject *__pyx_n_s_perf_counter_ns;
  PyObject *__pyx_n_s_perf_locals__test_zigzag_perf;
  PyObject *__pyx_n_s_perf_locals_add;
  PyObject *__pyx_n_s_perf_locals_add_locals_genexpr;
  PyObject *__pyx_n_s_perf_tests;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_u_placeholder;
  PyObject *__pyx_n_s_pop;
  PyObject *__pyx_n_s_prec;
  PyObject *__pyx_n_s_precision;
  PyObject *__pyx_n_u_precision;
  PyObject *__pyx_kp_u_precision_2;
  PyObject *__pyx_n_s_prepare;
  PyObject *__pyx_n_s_print;
  PyObject *__pyx_n_s_pytest;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_recursive_repr_guard;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle_ArrayType;
  PyObject *__pyx_n_s_pyx_unpickle_AvroType;
  PyObject *__pyx_n_s_pyx_unpickle_BoolType;
  PyObject *__pyx_n_s_pyx_unpickle_BytesType;
  PyObject *__pyx_n_s_pyx_unpickle_CannotPromote;
  PyObject *__pyx_n_s_pyx_unpickle_Codec;
  PyObject *__pyx_n_s_pyx_unpickle_ContainerReader;
  PyObject *__pyx_n_s_pyx_unpickle_CustomLogicalType;
  PyObject *__pyx_n_s_pyx_unpickle_Date;
  PyObject *__pyx_n_s_pyx_unpickle_DecimalType;
  PyObject *__pyx_n_s_pyx_unpickle_DoubleType;
  PyObject *__pyx_n_s_pyx_unpickle_Enum;
  PyObject *__pyx_n_s_pyx_unpickle_EnumType;
  PyObject *__pyx_n_s_pyx_unpickle_FileReader;
  PyObject *__pyx_n_s_pyx_unpickle_FileWriter;
  PyObject *__pyx_n_s_pyx_unpickle_FixedType;
  PyObject *__pyx_n_s_pyx_unpickle_FloatType;
  PyObject *__pyx_n_s_pyx_unpickle_IntType;
  PyObject *__pyx_n_s_pyx_unpickle_LogicalType;
  PyObject *__pyx_n_s_pyx_unpickle_LongType;
  PyObject *__pyx_n_s_pyx_unpickle_MapType;
  PyObject *__pyx_n_s_pyx_unpickle_MemoryReader;
  PyObject *__pyx_n_s_pyx_unpickle_NullType;
  PyObject *__pyx_n_s_pyx_unpickle_Options;
  PyObject *__pyx_n_s_pyx_unpickle_PromoteBytesToStr;
  PyObject *__pyx_n_s_pyx_unpickle_PromoteStringToBy;
  PyObject *__pyx_n_s_pyx_unpickle_PromoteToFloat;
  PyObject *__pyx_n_s_pyx_unpickle_PromotingEnumType;
  PyObject *__pyx_n_s_pyx_unpickle_PromotingRecordTy;
  PyObject *__pyx_n_s_pyx_unpickle_Rabin;
  PyObject *__pyx_n_s_pyx_unpickle_Record;
  PyObject *__pyx_n_s_pyx_unpickle_RecordField;
  PyObject *__pyx_n_s_pyx_unpickle_RecordType;
  PyObject *__pyx_n_s_pyx_unpickle_ResolvedSchema;
  PyObject *__pyx_n_s_pyx_unpickle_Schema;
  PyObject *__pyx_n_s_pyx_unpickle_StringType;
  PyObject *__pyx_n_s_pyx_unpickle_TimeMicros;
  PyObject *__pyx_n_s_pyx_unpickle_TimeMillis;
  PyObject *__pyx_n_s_pyx_unpickle_TimestampMicros;
  PyObject *__pyx_n_s_pyx_unpickle_TimestampMillis;
  PyObject *__pyx_n_s_pyx_unpickle_UUIDBase;
  PyObject *__pyx_n_s_pyx_unpickle_UUIDFixedType;
  PyObject *__pyx_n_s_pyx_unpickle_UUIDStringType;
  PyObject *__pyx_n_s_pyx_unpickle_UnionType;
  PyObject *__pyx_n_s_pyx_unpickle_ValueAdapter;
  PyObject *__pyx_n_s_pyx_unpickle__CanonicalForm;
  PyObject *__pyx_n_s_pyx_unpickle__DeflateCodec;
  PyObject *__pyx_n_s_pyx_unpickle__FieldAccessor;
  PyObject *__pyx_n_s_pyx_unpickle__Lz4Codec;
  PyObject *__pyx_n_s_pyx_unpickle__LzmaCodec;
  PyObject *__pyx_n_s_pyx_unpickle__NamedType;
  PyObject *__pyx_n_s_pyx_unpickle__NullCodec;
  PyObject *__pyx_n_s_pyx_unpickle__PlaceholderType;
  PyObject *__pyx_n_s_pyx_unpickle__Reader;
  PyObject *__pyx_n_s_pyx_unpickle__Sentinel;
  PyObject *__pyx_n_s_pyx_unpickle__SnappyCodec;
  PyObject *__pyx_n_s_pyx_unpickle__Writer;
  PyObject *__pyx_n_s_pyx_unpickle__ZStandardCodec;
  PyObject *__pyx_n_s_pyx_unpickle___Bzip2Codec;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_b_q;
  PyObject *__pyx_n_s_qualname;
  PyObject *__pyx_n_s_quantize;
  PyObject *__pyx_n_u_rabin;
  PyObject *__pyx_n_s_raise;
  PyObject *__pyx_n_s_raise_on_invalid_logical;
  PyObject *__pyx_kp_u_raise_on_invalid_logical_2;
  PyObject *__pyx_n_s_raises;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_u_rb;
  PyObject *__pyx_n_s_re;
  PyObject *__pyx_n_s_read;
  PyObject *__pyx_n_u_read;
  PyObject *__pyx_n_s_read_marker;
  PyObject *__pyx_n_s_read_perf;
  PyObject *__pyx_n_s_read_time;
  PyObject *__pyx_n_s_reader;
  PyObject *__pyx_n_s_reader_for_writer;
  PyObject *__pyx_n_s_reader_schema;
  PyObject *__pyx_n_s_reader_type;
  PyObject *__pyx_n_s_record;
  PyObject *__pyx_n_u_record;
  PyObject *__pyx_n_s_record_allow_extra_fields;
  PyObject *__pyx_kp_u_record_allow_extra_fields_2;
  PyObject *__pyx_n_s_record_can_encode_dict;
  PyObject *__pyx_kp_u_record_can_encode_dict_2;
  PyObject *__pyx_n_s_record_decodes_to_dict;
  PyObject *__pyx_kp_u_record_decodes_to_dict_2;
  PyObject *__pyx_n_s_record_encode_use_defaults;
  PyObject *__pyx_kp_u_record_encode_use_defaults_2;
  PyObject *__pyx_n_s_record_fields_must_be_unique;
  PyObject *__pyx_kp_u_record_fields_must_be_unique_2;
  PyObject *__pyx_n_s_record_values_type_hint;
  PyObject *__pyx_kp_u_record_values_type_hint_2;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_register;
  PyObject *__pyx_n_s_remove;
  PyObject *__pyx_n_s_replace;
  PyObject *__pyx_n_u_replace;
  PyObject *__pyx_n_s_repr;
  PyObject *__pyx_kp_u_required_field;
  PyObject *__pyx_n_s_resolved_type;
  PyObject *__pyx_n_s_return;
  PyObject *__pyx_n_s_return_uuid_object;
  PyObject *__pyx_kp_u_return_uuid_object_2;
  PyObject *__pyx_n_s_reversed;
  PyObject *__pyx_n_s_rjust;
  PyObject *__pyx_n_s_rsplit;
  PyObject *__pyx_n_s_running;
  PyObject *__pyx_kp_u_s_read;
  PyObject *__pyx_n_s_scale;
  PyObject *__pyx_n_u_scale;
  PyObject *__pyx_n_s_scaleb;
  PyObject *__pyx_n_s_schema;
  PyObject *__pyx_n_s_schema_ob;
  PyObject *__pyx_n_s_schema_path;
  PyObject *__pyx_n_s_second;
  PyObject *__pyx_n_s_seconds;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_n_s_self_val;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_n_s_serialize;
  PyObject *__pyx_n_s_set_from_dataclass;
  PyObject *__pyx_n_s_set_name;
  PyObject *__pyx_n_s_setdefault;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_shape;
  PyObject *__pyx_n_s_signature;
  PyObject *__pyx_n_s_signature_2;
  PyObject *__pyx_n_s_signed;
  PyObject *__pyx_n_s_size;
  PyObject *__pyx_n_u_size;
  PyObject *__pyx_n_u_slots;
  PyObject *__pyx_n_s_slots_2;
  PyObject *__pyx_n_b_snappy;
  PyObject *__pyx_n_s_snappy;
  PyObject *__pyx_n_u_snappy;
  PyObject *__pyx_n_s_source;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_s_src;
  PyObject *__pyx_kp_s_src_array_pxi;
  PyObject *__pyx_kp_s_src_container_pxi;
  PyObject *__pyx_kp_s_src_enum_pxi;
  PyObject *__pyx_kp_s_src_error_pxi;
  PyObject *__pyx_kp_s_src_logical_pxi;
  PyObject *__pyx_kp_s_src_map_pxi;
  PyObject *__pyx_kp_s_src_null_pxi;
  PyObject *__pyx_kp_s_src_numeric_types_pxi;
  PyObject *__pyx_kp_s_src_option_pxi;
  PyObject *__pyx_kp_s_src_rabin_pxi;
  PyObject *__pyx_kp_s_src_record_pxi;
  PyObject *__pyx_kp_s_src_schema_pxi;
  PyObject *__pyx_kp_s_src_string_types_pxi;
  PyObject *__pyx_kp_s_src_tests_test_buffer_pxi;
  PyObject *__pyx_kp_s_src_tests_test_perf_pxi;
  PyObject *__pyx_kp_s_src_tests_test_zigzag_pxi;
  PyObject *__pyx_kp_s_src_type_pxi;
  PyObject *__pyx_kp_s_src_union_pxi;
  PyObject *__pyx_n_s_start;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_s_step;
  PyObject *__pyx_n_s_stop;
  PyObject *__pyx_n_s_str;
  PyObject *__pyx_n_s_str_2;
  PyObject *__pyx_n_u_strict;
  PyObject *__pyx_kp_s_strided_and_direct;
  PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
  PyObject *__pyx_kp_s_strided_and_indirect;
  PyObject *__pyx_n_u_string;
  PyObject *__pyx_kp_u_string_2;
  PyObject *__pyx_n_s_string_types_default_unchanged;
  PyObject *__pyx_kp_u_string_types_default_unchanged_2;
  PyObject *__pyx_kp_u_string_uuid;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_strptime;
  PyObject *__pyx_n_s_struct;
  PyObject *__pyx_n_s_sum;
  PyObject *__pyx_n_s_super;
  PyObject *__pyx_n_s_symbols;
  PyObject *__pyx_n_u_symbols;
  PyObject *__pyx_n_s_sys;
  PyObject *__pyx_n_s_t;
  PyObject *__pyx_n_s_t_a;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_test_2;
  PyObject *__pyx_n_s_test_reader_read_n;
  PyObject *__pyx_n_s_test_reader_read_u8;
  PyObject *__pyx_n_s_test_readvarint;
  PyObject *__pyx_n_s_test_readvarlong;
  PyObject *__pyx_n_s_test_zigzag_long;
  PyObject *__pyx_n_s_test_zigzag_perf;
  PyObject *__pyx_n_s_tests;
  PyObject *__pyx_n_s_tests_2;
  PyObject *__pyx_n_s_tests_locals__test_reader_read;
  PyObject *__pyx_n_s_tests_locals__test_reader_read_2;
  PyObject *__pyx_n_s_tests_locals_add;
  PyObject *__pyx_n_s_tests_locals_add_locals_genexpr;
  PyObject *__pyx_n_s_tests_locals_test_readvarint;
  PyObject *__pyx_n_s_tests_locals_test_readvarlong;
  PyObject *__pyx_n_s_tests_locals_test_zigzag_long;
  PyObject *__pyx_n_s_threading;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_time;
  PyObject *__pyx_kp_u_time_micros;
  PyObject *__pyx_kp_u_time_millis;
  PyObject *__pyx_n_s_time_to_micros;
  PyObject *__pyx_n_s_timedelta;
  PyObject *__pyx_n_s_timestamp;
  PyObject *__pyx_kp_u_timestamp_micros;
  PyObject *__pyx_kp_u_timestamp_millis;
  PyObject *__pyx_n_s_timezone;
  PyObject *__pyx_kp_u_to;
  PyObject *__pyx_n_s_to_bytes;
  PyObject *__pyx_n_s_total_nums;
  PyObject *__pyx_n_s_truncate_fixed;
  PyObject *__pyx_kp_u_truncate_fixed_2;
  PyObject *__pyx_kp_s_tuple_LogicalType;
  PyObject *__pyx_n_s_type;
  PyObject *__pyx_n_u_type;
  PyObject *__pyx_kp_u_type_2;
  PyObject *__pyx_n_s_type_3;
  PyObject *__pyx_n_s_type_name;
  PyObject *__pyx_n_s_type_str;
  PyObject *__pyx_n_s_types;
  PyObject *__pyx_n_s_types_str_to_schema;
  PyObject *__pyx_kp_u_types_str_to_schema_2;
  PyObject *__pyx_n_s_typing;
  PyObject *__pyx_n_s_tzinfo;
  PyObject *__pyx_n_s_ubin;
  PyObject *__pyx_kp_s_unable_to_allocate_array_data;
  PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
  PyObject *__pyx_n_s_underlying;
  PyObject *__pyx_n_s_underlying_name;
  PyObject *__pyx_n_s_underlying_type;
  PyObject *__pyx_n_s_underlying_types;
  PyObject *__pyx_n_s_unicode_errors;
  PyObject *__pyx_kp_u_unicode_errors_2;
  PyObject *__pyx_n_s_unicodedata;
  PyObject *__pyx_n_u_union;
  PyObject *__pyx_n_s_union_decodes_to;
  PyObject *__pyx_kp_u_union_decodes_to_2;
  PyObject *__pyx_n_s_union_json_encodes_type_name;
  PyObject *__pyx_kp_u_union_json_encodes_type_name_2;
  PyObject *__pyx_n_s_unpack;
  PyObject *__pyx_n_s_unsafe_hash;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_use_setstate;
  PyObject *__pyx_n_s_utc;
  PyObject *__pyx_n_u_utf8;
  PyObject *__pyx_kp_u_utf_8;
  PyObject *__pyx_n_s_uuid;
  PyObject *__pyx_n_u_uuid;
  PyObject *__pyx_n_s_uuid4;
  PyObject *__pyx_n_s_val;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_kp_u_value_2;
  PyObject *__pyx_n_s_value_adapters;
  PyObject *__pyx_n_s_value_type;
  PyObject *__pyx_n_u_values;
  PyObject *__pyx_n_s_version;
  PyObject *__pyx_n_s_version_info;
  PyObject *__pyx_n_s_visited;
  PyObject *__pyx_kp_u_w;
  PyObject *__pyx_n_s_walk_types;
  PyObject *__pyx_n_s_warn;
  PyObject *__pyx_n_s_warnings;
  PyObject *__pyx_n_u_wb;
  PyObject *__pyx_n_s_wbits;
  PyObject *__pyx_n_s_weakref_slot;
  PyObject *__pyx_kp_u_which_is_greater_than_the_scale;
  PyObject *__pyx_n_s_with_external_types;
  PyObject *__pyx_n_s_with_logical_types;
  PyObject *__pyx_n_s_wrap_type;
  PyObject *__pyx_n_s_wrapped;
  PyObject *__pyx_n_s_write;
  PyObject *__pyx_n_u_write;
  PyObject *__pyx_kp_u_write_2;
  PyObject *__pyx_n_s_write_header;
  PyObject *__pyx_n_s_write_many;
  PyObject *__pyx_n_s_write_one;
  PyObject *__pyx_n_s_write_perf;
  PyObject *__pyx_n_s_write_time;
  PyObject *__pyx_n_s_writer;
  PyObject *__pyx_n_s_writer_schema;
  PyObject *__pyx_n_s_writer_type;
  PyObject *__pyx_n_b_xz;
  PyObject *__pyx_n_u_xz;
  PyObject *__pyx_n_s_zip;
  PyObject *__pyx_n_s_zlib;
  PyObject *__pyx_n_b_zstandard;
  PyObject *__pyx_n_s_zstandard;
  PyObject *__pyx_n_u_zstandard;
  PyObject *__pyx_float_0_0;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_2;
  PyObject *__pyx_int_3;
  PyObject *__pyx_int_4;
  PyObject *__pyx_int_8;
  PyObject *__pyx_int_10;
  PyObject *__pyx_int_16;
  PyObject *__pyx_int_24;
  PyObject *__pyx_int_60;
  PyObject *__pyx_int_70;
  PyObject *__pyx_int_127;
  PyObject *__pyx_int_255;
  PyObject *__pyx_int_1000;
  PyObject *__pyx_int_1970;
  PyObject *__pyx_int_3600;
  PyObject *__pyx_int_4096;
  PyObject *__pyx_int_32767;
  PyObject *__pyx_int_65536;
  PyObject *__pyx_int_1000000;
  PyObject *__pyx_int_8388607;
  PyObject *__pyx_int_18332779;
  PyObject *__pyx_int_24827247;
  PyObject *__pyx_int_25341170;
  PyObject *__pyx_int_29099189;
  PyObject *__pyx_int_31039736;
  PyObject *__pyx_int_33963030;
  PyObject *__pyx_int_38410757;
  PyObject *__pyx_int_38632982;
  PyObject *__pyx_int_39872412;
  PyObject *__pyx_int_42565757;
  PyObject *__pyx_int_47772561;
  PyObject *__pyx_int_60000000;
  PyObject *__pyx_int_61270791;
  PyObject *__pyx_int_61440409;
  PyObject *__pyx_int_61765448;
  PyObject *__pyx_int_67877419;
  PyObject *__pyx_int_75337510;
  PyObject *__pyx_int_79389776;
  PyObject *__pyx_int_79941127;
  PyObject *__pyx_int_81233721;
  PyObject *__pyx_int_83351333;
  PyObject *__pyx_int_95761985;
  PyObject *__pyx_int_96605483;
  PyObject *__pyx_int_101034655;
  PyObject *__pyx_int_101670942;
  PyObject *__pyx_int_111776469;
  PyObject *__pyx_int_111957047;
  PyObject *__pyx_int_112105877;
  PyObject *__pyx_int_115671040;
  PyObject *__pyx_int_119832798;
  PyObject *__pyx_int_122446337;
  PyObject *__pyx_int_122978751;
  PyObject *__pyx_int_136983863;
  PyObject *__pyx_int_139219010;
  PyObject *__pyx_int_145104888;
  PyObject *__pyx_int_148208818;
  PyObject *__pyx_int_148222551;
  PyObject *__pyx_int_148338675;
  PyObject *__pyx_int_148737976;
  PyObject *__pyx_int_150354737;
  PyObject *__pyx_int_159023012;
  PyObject *__pyx_int_159276020;
  PyObject *__pyx_int_159891379;
  PyObject *__pyx_int_160503772;
  PyObject *__pyx_int_160934521;
  PyObject *__pyx_int_164094308;
  PyObject *__pyx_int_169331114;
  PyObject *__pyx_int_175892883;
  PyObject *__pyx_int_176081420;
  PyObject *__pyx_int_176267327;
  PyObject *__pyx_int_177382991;
  PyObject *__pyx_int_178791180;
  PyObject *__pyx_int_180170947;
  PyObject *__pyx_int_180285838;
  PyObject *__pyx_int_181080083;
  PyObject *__pyx_int_184956420;
  PyObject *__pyx_int_184977713;
  PyObject *__pyx_int_193777858;
  PyObject *__pyx_int_195980960;
  PyObject *__pyx_int_198151601;
  PyObject *__pyx_int_202158861;
  PyObject *__pyx_int_202608488;
  PyObject *__pyx_int_203789025;
  PyObject *__pyx_int_205941106;
  PyObject *__pyx_int_206778879;
  PyObject *__pyx_int_212073902;
  PyObject *__pyx_int_213886906;
  PyObject *__pyx_int_215229444;
  PyObject *__pyx_int_215954136;
  PyObject *__pyx_int_222419149;
  PyObject *__pyx_int_228539134;
  PyObject *__pyx_int_228825662;
  PyObject *__pyx_int_229636597;
  PyObject *__pyx_int_233298401;
  PyObject *__pyx_int_236180632;
  PyObject *__pyx_int_236349525;
  PyObject *__pyx_int_238750788;
  PyObject *__pyx_int_240389341;
  PyObject *__pyx_int_245215779;
  PyObject *__pyx_int_247527084;
  PyObject *__pyx_int_254981756;
  PyObject *__pyx_int_256569670;
  PyObject *__pyx_int_259444540;
  PyObject *__pyx_int_261059191;
  PyObject *__pyx_int_263819094;
  PyObject *__pyx_int_2147483647;
  PyObject *__pyx_int_3442987655;
  PyObject *__pyx_int_3600000000;
  PyObject *__pyx_int_neg_1;
  PyObject *__pyx_int_neg_15;
  PyObject *__pyx_int_549755813887L;
  PyObject *__pyx_int_0x7fffffffffffL;
  PyObject *__pyx_int_0x7fffffffffffffL;
  PyObject *__pyx_int_0x7fffffffffffffffL;
  PyObject *__pyx_k__15;
  PyObject *__pyx_k__16;
  PyObject *__pyx_k__17;
  PyObject *__pyx_k__21;
  PyObject *__pyx_k__81;
  struct __pyx_obj_5cavro_Options *__pyx_k__83;
  struct __pyx_obj_5cavro_Options *__pyx_k__84;
  PyObject *__pyx_k__87;
  PyObject *__pyx_k__89;
  PyObject *__pyx_slice__5;
  PyObject *__pyx_tuple__4;
  PyObject *__pyx_tuple__8;
  PyObject *__pyx_slice__23;
  PyObject *__pyx_slice__86;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__19;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__28;
  PyObject *__pyx_tuple__30;
  PyObject *__pyx_tuple__32;
  PyObject *__pyx_tuple__35;
  PyObject *__pyx_tuple__36;
  PyObject *__pyx_tuple__37;
  PyObject *__pyx_tuple__38;
  PyObject *__pyx_tuple__40;
  PyObject *__pyx_tuple__42;
  PyObject *__pyx_tuple__43;
  PyObject *__pyx_tuple__46;
  PyObject *__pyx_tuple__47;
  PyObject *__pyx_tuple__50;
  PyObject *__pyx_tuple__51;
  PyObject *__pyx_tuple__52;
  PyObject *__pyx_tuple__53;
  PyObject *__pyx_tuple__54;
  PyObject *__pyx_tuple__55;
  PyObject *__pyx_tuple__56;
  PyObject *__pyx_tuple__57;
  PyObject *__pyx_tuple__58;
  PyObject *__pyx_tuple__59;
  PyObject *__pyx_tuple__65;
  PyObject *__pyx_tuple__67;
  PyObject *__pyx_tuple__68;
  PyObject *__pyx_tuple__69;
  PyObject *__pyx_tuple__71;
  PyObject *__pyx_tuple__73;
  PyObject *__pyx_tuple__74;
  PyObject *__pyx_tuple__85;
  PyObject *__pyx_tuple__88;
  PyObject *__pyx_tuple__90;
  PyObject *__pyx_tuple__91;
  PyObject *__pyx_tuple__92;
  PyObject *__pyx_tuple__93;
  PyObject *__pyx_tuple__94;
  PyObject *__pyx_tuple__95;
  PyObject *__pyx_tuple__96;
  PyObject *__pyx_tuple__97;
  PyObject *__pyx_tuple__98;
  PyObject *__pyx_tuple__99;
  PyObject *__pyx_tuple__100;
  PyObject *__pyx_tuple__101;
  PyObject *__pyx_tuple__102;
  PyObject *__pyx_tuple__103;
  PyObject *__pyx_tuple__104;
  PyObject *__pyx_tuple__105;
  PyObject *__pyx_tuple__106;
  PyObject *__pyx_tuple__107;
  PyObject *__pyx_tuple__108;
  PyObject *__pyx_tuple__109;
  PyObject *__pyx_tuple__110;
  PyObject *__pyx_tuple__111;
  PyObject *__pyx_tuple__112;
  PyObject *__pyx_tuple__113;
  PyObject *__pyx_tuple__114;
  PyObject *__pyx_tuple__115;
  PyObject *__pyx_tuple__116;
  PyObject *__pyx_tuple__117;
  PyObject *__pyx_tuple__118;
  PyObject *__pyx_tuple__119;
  PyObject *__pyx_tuple__120;
  PyObject *__pyx_tuple__121;
  PyObject *__pyx_tuple__123;
  PyObject *__pyx_tuple__124;
  PyObject *__pyx_tuple__125;
  PyObject *__pyx_tuple__126;
  PyObject *__pyx_tuple__127;
  PyObject *__pyx_tuple__128;
  PyObject *__pyx_tuple__129;
  PyObject *__pyx_tuple__130;
  PyObject *__pyx_tuple__131;
  PyObject *__pyx_tuple__132;
  PyObject *__pyx_tuple__134;
  PyObject *__pyx_tuple__137;
  PyObject *__pyx_tuple__139;
  PyObject *__pyx_tuple__141;
  PyObject *__pyx_tuple__142;
  PyObject *__pyx_tuple__144;
  PyObject *__pyx_tuple__145;
  PyObject *__pyx_tuple__147;
  PyObject *__pyx_tuple__148;
  PyObject *__pyx_tuple__150;
  PyObject *__pyx_tuple__153;
  PyObject *__pyx_tuple__155;
  PyObject *__pyx_tuple__157;
  PyObject *__pyx_tuple__159;
  PyObject *__pyx_tuple__161;
  PyObject *__pyx_tuple__163;
  PyObject *__pyx_tuple__164;
  PyObject *__pyx_tuple__166;
  PyObject *__pyx_tuple__167;
  PyObject *__pyx_tuple__169;
  PyObject *__pyx_tuple__183;
  PyObject *__pyx_tuple__185;
  PyObject *__pyx_tuple__188;
  PyObject *__pyx_tuple__192;
  PyObject *__pyx_tuple__194;
  PyObject *__pyx_tuple__196;
  PyObject *__pyx_tuple__198;
  PyObject *__pyx_tuple__201;
  PyObject *__pyx_tuple__202;
  PyObject *__pyx_tuple__204;
  PyObject *__pyx_tuple__206;
  PyObject *__pyx_tuple__218;
  PyObject *__pyx_tuple__220;
  PyObject *__pyx_tuple__253;
  PyObject *__pyx_tuple__258;
  PyObject *__pyx_tuple__260;
  PyObject *__pyx_tuple__264;
  PyObject *__pyx_tuple__271;
  PyObject *__pyx_tuple__294;
  PyObject *__pyx_tuple__297;
  PyObject *__pyx_tuple__300;
  PyObject *__pyx_tuple__303;
  PyObject *__pyx_tuple__306;
  PyObject *__pyx_tuple__308;
  PyObject *__pyx_tuple__313;
  PyObject *__pyx_tuple__321;
  PyObject *__pyx_tuple__323;
  PyObject *__pyx_tuple__325;
  PyObject *__pyx_tuple__327;
  PyObject *__pyx_tuple__329;
  PyObject *__pyx_tuple__331;
  PyObject *__pyx_tuple__332;
  PyObject *__pyx_tuple__334;
  PyObject *__pyx_tuple__335;
  PyObject *__pyx_tuple__337;
  PyObject *__pyx_tuple__340;
  PyObject *__pyx_tuple__342;
  PyObject *__pyx_tuple__344;
  PyObject *__pyx_tuple__346;
  PyObject *__pyx_tuple__347;
  PyObject *__pyx_tuple__349;
  PyObject *__pyx_tuple__357;
  PyObject *__pyx_tuple__363;
  PyObject *__pyx_tuple__367;
  PyObject *__pyx_tuple__377;
  PyObject *__pyx_tuple__381;
  PyObject *__pyx_tuple__396;
  PyObject *__pyx_tuple__418;
  PyObject *__pyx_tuple__420;
  PyObject *__pyx_tuple__422;
  PyObject *__pyx_tuple__424;
  PyObject *__pyx_tuple__425;
  PyObject *__pyx_tuple__427;
  PyObject *__pyx_codeobj__11;
  PyObject *__pyx_codeobj__12;
  PyObject *__pyx_codeobj__29;
  PyObject *__pyx_codeobj__31;
  PyObject *__pyx_codeobj__33;
  PyObject *__pyx_codeobj__39;
  PyObject *__pyx_codeobj__41;
  PyObject *__pyx_codeobj__44;
  PyObject *__pyx_codeobj__45;
  PyObject *__pyx_codeobj__49;
  PyObject *__pyx_codeobj__66;
  PyObject *__pyx_codeobj__75;
  PyObject *__pyx_codeobj__82;
  PyObject *__pyx_codeobj__133;
  PyObject *__pyx_codeobj__135;
  PyObject *__pyx_codeobj__136;
  PyObject *__pyx_codeobj__138;
  PyObject *__pyx_codeobj__140;
  PyObject *__pyx_codeobj__143;
  PyObject *__pyx_codeobj__146;
  PyObject *__pyx_codeobj__149;
  PyObject *__pyx_codeobj__151;
  PyObject *__pyx_codeobj__154;
  PyObject *__pyx_codeobj__156;
  PyObject *__pyx_codeobj__158;
  PyObject *__pyx_codeobj__160;
  PyObject *__pyx_codeobj__162;
  PyObject *__pyx_codeobj__165;
  PyObject *__pyx_codeobj__168;
  PyObject *__pyx_codeobj__170;
  PyObject *__pyx_codeobj__171;
  PyObject *__pyx_codeobj__172;
  PyObject *__pyx_codeobj__173;
  PyObject *__pyx_codeobj__174;
  PyObject *__pyx_codeobj__175;
  PyObject *__pyx_codeobj__176;
  PyObject *__pyx_codeobj__177;
  PyObject *__pyx_codeobj__178;
  PyObject *__pyx_codeobj__179;
  PyObject *__pyx_codeobj__180;
  PyObject *__pyx_codeobj__181;
  PyObject *__pyx_codeobj__182;
  PyObject *__pyx_codeobj__184;
  PyObject *__pyx_codeobj__186;
  PyObject *__pyx_codeobj__187;
  PyObject *__pyx_codeobj__189;
  PyObject *__pyx_codeobj__190;
  PyObject *__pyx_codeobj__191;
  PyObject *__pyx_codeobj__193;
  PyObject *__pyx_codeobj__195;
  PyObject *__pyx_codeobj__197;
  PyObject *__pyx_codeobj__199;
  PyObject *__pyx_codeobj__200;
  PyObject *__pyx_codeobj__203;
  PyObject *__pyx_codeobj__205;
  PyObject *__pyx_codeobj__207;
  PyObject *__pyx_codeobj__208;
  PyObject *__pyx_codeobj__209;
  PyObject *__pyx_codeobj__210;
  PyObject *__pyx_codeobj__211;
  PyObject *__pyx_codeobj__212;
  PyObject *__pyx_codeobj__213;
  PyObject *__pyx_codeobj__214;
  PyObject *__pyx_codeobj__215;
  PyObject *__pyx_codeobj__216;
  PyObject *__pyx_codeobj__217;
  PyObject *__pyx_codeobj__219;
  PyObject *__pyx_codeobj__221;
  PyObject *__pyx_codeobj__222;
  PyObject *__pyx_codeobj__223;
  PyObject *__pyx_codeobj__224;
  PyObject *__pyx_codeobj__225;
  PyObject *__pyx_codeobj__226;
  PyObject *__pyx_codeobj__227;
  PyObject *__pyx_codeobj__228;
  PyObject *__pyx_codeobj__229;
  PyObject *__pyx_codeobj__230;
  PyObject *__pyx_codeobj__231;
  PyObject *__pyx_codeobj__232;
  PyObject *__pyx_codeobj__233;
  PyObject *__pyx_codeobj__234;
  PyObject *__pyx_codeobj__235;
  PyObject *__pyx_codeobj__236;
  PyObject *__pyx_codeobj__237;
  PyObject *__pyx_codeobj__238;
  PyObject *__pyx_codeobj__239;
  PyObject *__pyx_codeobj__240;
  PyObject *__pyx_codeobj__241;
  PyObject *__pyx_codeobj__242;
  PyObject *__pyx_codeobj__243;
  PyObject *__pyx_codeobj__244;
  PyObject *__pyx_codeobj__245;
  PyObject *__pyx_codeobj__246;
  PyObject *__pyx_codeobj__247;
  PyObject *__pyx_codeobj__248;
  PyObject *__pyx_codeobj__249;
  PyObject *__pyx_codeobj__250;
  PyObject *__pyx_codeobj__251;
  PyObject *__pyx_codeobj__252;
  PyObject *__pyx_codeobj__254;
  PyObject *__pyx_codeobj__255;
  PyObject *__pyx_codeobj__256;
  PyObject *__pyx_codeobj__257;
  PyObject *__pyx_codeobj__259;
  PyObject *__pyx_codeobj__261;
  PyObject *__pyx_codeobj__262;
  PyObject *__pyx_codeobj__263;
  PyObject *__pyx_codeobj__265;
  PyObject *__pyx_codeobj__266;
  PyObject *__pyx_codeobj__267;
  PyObject *__pyx_codeobj__268;
  PyObject *__pyx_codeobj__269;
  PyObject *__pyx_codeobj__270;
  PyObject *__pyx_codeobj__272;
  PyObject *__pyx_codeobj__273;
  PyObject *__pyx_codeobj__274;
  PyObject *__pyx_codeobj__275;
  PyObject *__pyx_codeobj__276;
  PyObject *__pyx_codeobj__277;
  PyObject *__pyx_codeobj__278;
  PyObject *__pyx_codeobj__279;
  PyObject *__pyx_codeobj__280;
  PyObject *__pyx_codeobj__281;
  PyObject *__pyx_codeobj__282;
  PyObject *__pyx_codeobj__283;
  PyObject *__pyx_codeobj__284;
  PyObject *__pyx_codeobj__285;
  PyObject *__pyx_codeobj__286;
  PyObject *__pyx_codeobj__287;
  PyObject *__pyx_codeobj__288;
  PyObject *__pyx_codeobj__289;
  PyObject *__pyx_codeobj__290;
  PyObject *__pyx_codeobj__291;
  PyObject *__pyx_codeobj__292;
  PyObject *__pyx_codeobj__293;
  PyObject *__pyx_codeobj__295;
  PyObject *__pyx_codeobj__296;
  PyObject *__pyx_codeobj__298;
  PyObject *__pyx_codeobj__299;
  PyObject *__pyx_codeobj__301;
  PyObject *__pyx_codeobj__302;
  PyObject *__pyx_codeobj__304;
  PyObject *__pyx_codeobj__305;
  PyObject *__pyx_codeobj__307;
  PyObject *__pyx_codeobj__309;
  PyObject *__pyx_codeobj__310;
  PyObject *__pyx_codeobj__311;
  PyObject *__pyx_codeobj__312;
  PyObject *__pyx_codeobj__314;
  PyObject *__pyx_codeobj__315;
  PyObject *__pyx_codeobj__316;
  PyObject *__pyx_codeobj__317;
  PyObject *__pyx_codeobj__318;
  PyObject *__pyx_codeobj__319;
  PyObject *__pyx_codeobj__320;
  PyObject *__pyx_codeobj__322;
  PyObject *__pyx_codeobj__324;
  PyObject *__pyx_codeobj__326;
  PyObject *__pyx_codeobj__328;
  PyObject *__pyx_codeobj__330;
  PyObject *__pyx_codeobj__333;
  PyObject *__pyx_codeobj__336;
  PyObject *__pyx_codeobj__338;
  PyObject *__pyx_codeobj__339;
  PyObject *__pyx_codeobj__341;
  PyObject *__pyx_codeobj__343;
  PyObject *__pyx_codeobj__345;
  PyObject *__pyx_codeobj__348;
  PyObject *__pyx_codeobj__350;
  PyObject *__pyx_codeobj__351;
  PyObject *__pyx_codeobj__352;
  PyObject *__pyx_codeobj__353;
  PyObject *__pyx_codeobj__354;
  PyObject *__pyx_codeobj__355;
  PyObject *__pyx_codeobj__356;
  PyObject *__pyx_codeobj__358;
  PyObject *__pyx_codeobj__359;
  PyObject *__pyx_codeobj__360;
  PyObject *__pyx_codeobj__361;
  PyObject *__pyx_codeobj__362;
  PyObject *__pyx_codeobj__364;
  PyObject *__pyx_codeobj__365;
  PyObject *__pyx_codeobj__366;
  PyObject *__pyx_codeobj__368;
  PyObject *__pyx_codeobj__369;
  PyObject *__pyx_codeobj__370;
  PyObject *__pyx_codeobj__371;
  PyObject *__pyx_codeobj__372;
  PyObject *__pyx_codeobj__373;
  PyObject *__pyx_codeobj__374;
  PyObject *__pyx_codeobj__375;
  PyObject *__pyx_codeobj__376;
  PyObject *__pyx_codeobj__378;
  PyObject *__pyx_codeobj__379;
  PyObject *__pyx_codeobj__380;
  PyObject *__pyx_codeobj__382;
  PyObject *__pyx_codeobj__383;
  PyObject *__pyx_codeobj__384;
  PyObject *__pyx_codeobj__385;
  PyObject *__pyx_codeobj__386;
  PyObject *__pyx_codeobj__387;
  PyObject *__pyx_codeobj__388;
  PyObject *__pyx_codeobj__389;
  PyObject *__pyx_codeobj__390;
  PyObject *__pyx_codeobj__391;
  PyObject *__pyx_codeobj__392;
  PyObject *__pyx_codeobj__393;
  PyObject *__pyx_codeobj__394;
  PyObject *__pyx_codeobj__395;
  PyObject *__pyx_codeobj__397;
  PyObject *__pyx_codeobj__398;
  PyObject *__pyx_codeobj__399;
  PyObject *__pyx_codeobj__400;
  PyObject *__pyx_codeobj__401;
  PyObject *__pyx_codeobj__402;
  PyObject *__pyx_codeobj__403;
  PyObject *__pyx_codeobj__404;
  PyObject *__pyx_codeobj__405;
  PyObject *__pyx_codeobj__406;
  PyObject *__pyx_codeobj__407;
  PyObject *__pyx_codeobj__408;
  PyObject *__pyx_codeobj__409;
  PyObject *__pyx_codeobj__410;
  PyObject *__pyx_codeobj__411;
  PyObject *__pyx_codeobj__412;
  PyObject *__pyx_codeobj__413;
  PyObject *__pyx_codeobj__414;
  PyObject *__pyx_codeobj__415;
  PyObject *__pyx_codeobj__416;
  PyObject *__pyx_codeobj__417;
  PyObject *__pyx_codeobj__419;
  PyObject *__pyx_codeobj__421;
  PyObject *__pyx_codeobj__423;
  PyObject *__pyx_codeobj__426;
  PyObject *__pyx_codeobj__428;
  PyObject *__pyx_codeobj__429;
  PyObject *__pyx_codeobj__430;
  PyObject *__pyx_codeobj__431;
  PyObject *__pyx_codeobj__432;
  PyObject *__pyx_codeobj__433;
  PyObject *__pyx_codeobj__434;
  PyObject *__pyx_codeobj__435;
  PyObject *__pyx_codeobj__436;
  PyObject *__pyx_codeobj__437;
  PyObject *__pyx_codeobj__438;
  PyObject *__pyx_codeobj__439;
  PyObject *__pyx_codeobj__440;
  PyObject *__pyx_codeobj__441;
  PyObject *__pyx_codeobj__442;
  PyObject *__pyx_codeobj__443;
  PyObject *__pyx_codeobj__444;
  PyObject *__pyx_codeobj__445;
  PyObject *__pyx_codeobj__446;
  PyObject *__pyx_codeobj__447;
  PyObject *__pyx_codeobj__448;
  PyObject *__pyx_codeobj__449;
  PyObject *__pyx_codeobj__450;
  PyObject *__pyx_codeobj__451;
  PyObject *__pyx_codeobj__452;
  PyObject *__pyx_codeobj__453;
  PyObject *__pyx_codeobj__454;
  PyObject *__pyx_codeobj__455;
  PyObject *__pyx_codeobj__456;
  PyObject *__pyx_codeobj__457;
  PyObject *__pyx_codeobj__458;
  PyObject *__pyx_codeobj__459;
  PyObject *__pyx_codeobj__460;
  PyObject *__pyx_codeobj__461;
  PyObject *__pyx_codeobj__462;
  PyObject *__pyx_codeobj__463;
  PyObject *__pyx_codeobj__464;
  PyObject *__pyx_codeobj__465;
  PyObject *__pyx_codeobj__466;
  PyObject *__pyx_codeobj__467;
  PyObject *__pyx_codeobj__468;
  PyObject *__pyx_codeobj__469;
  PyObject *__pyx_codeobj__470;
  PyObject *__pyx_codeobj__471;
  PyObject *__pyx_codeobj__472;
  PyObject *__pyx_codeobj__473;
  PyObject *__pyx_codeobj__474;
  PyObject *__pyx_codeobj__475;
  PyObject *__pyx_codeobj__476;
  PyObject *__pyx_codeobj__477;
  PyObject *__pyx_codeobj__478;
  PyObject *__pyx_codeobj__479;
  PyObject *__pyx_codeobj__480;
  PyObject *__pyx_codeobj__481;
  PyObject *__pyx_codeobj__482;
  PyObject *__pyx_codeobj__483;
  PyObject *__pyx_codeobj__484;
  PyObject *__pyx_codeobj__485;
  PyObject *__pyx_codeobj__486;
  PyObject *__pyx_codeobj__487;
  PyObject *__pyx_codeobj__488;
  PyObject *__pyx_codeobj__489;
  PyObject *__pyx_codeobj__490;
  PyObject *__pyx_codeobj__491;
  PyObject *__pyx_codeobj__492;
  PyObject *__pyx_codeobj__493;
  PyObject *__pyx_codeobj__494;
  PyObject *__pyx_codeobj__495;
  PyObject *__pyx_codeobj__496;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_5array_array);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_Options);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_Options);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__Writer);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__Writer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__Reader);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__Reader);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_FileReader);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_FileReader);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_FileWriter);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_FileWriter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_MemoryWriter);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_MemoryWriter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_MemoryReader);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_MemoryReader);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_Rabin);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_Rabin);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_Record);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_Record);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_RecordField);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_RecordField);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__FieldAccessor);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__FieldAccessor);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__Sentinel);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__Sentinel);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__CanonicalForm);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__CanonicalForm);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_AvroType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_AvroType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_UnionType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_UnionType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_StringType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_StringType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_BytesType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_BytesType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__PlaceholderType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__PlaceholderType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_DoubleType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_DoubleType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_FloatType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_FloatType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_LongType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_LongType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_IntType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_IntType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_BoolType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_BoolType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_NullType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_NullType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_MapType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_MapType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_ArrayType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_ArrayType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__NamedType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__NamedType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_FixedType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_FixedType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_RecordType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_RecordType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_PromotingRecordType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_PromotingRecordType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_EnumType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_EnumType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_PromotingEnumType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_PromotingEnumType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_Schema);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_Schema);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_ResolvedSchema);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_ResolvedSchema);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_ValueAdapter);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_ValueAdapter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_LogicalType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_LogicalType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_CustomLogicalType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_CustomLogicalType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_DecimalType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_DecimalType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_UUIDBase);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_UUIDBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_UUIDStringType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_UUIDStringType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_UUIDFixedType);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_UUIDFixedType);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_Date);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_Date);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_TimeMillis);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_TimeMillis);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_TimeMicros);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_TimeMicros);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_TimestampMillis);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_TimestampMillis);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_TimestampMicros);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_TimestampMicros);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_Codec);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_Codec);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__SnappyCodec);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__SnappyCodec);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__NullCodec);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__NullCodec);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__DeflateCodec);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__DeflateCodec);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___Bzip2Codec);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___Bzip2Codec);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__LzmaCodec);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__LzmaCodec);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__ZStandardCodec);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__ZStandardCodec);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro__Lz4Codec);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro__Lz4Codec);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_ContainerReader);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_ContainerReader);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_ContainerWriter);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_ContainerWriter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_PromoteToFloat);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_PromoteToFloat);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_PromoteBytesToString);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_PromoteBytesToString);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_PromoteStringToBytes);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_PromoteStringToBytes);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro_CannotPromote);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro_CannotPromote);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct__genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct__genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_1_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_1_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_2_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_2_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_3_walk_types);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_3_walk_types);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_5_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_5_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_6_walk_types);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_6_walk_types);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_7_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_7_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_8___init__);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_8___init__);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_9_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_9_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_10_walk_types);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_10_walk_types);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_11___init__);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_11___init__);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_12_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_12_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_13_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_13_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_14_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_14_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_15_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_15_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_16_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_16_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_17_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_17_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_18_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_18_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_19_walk_types);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_19_walk_types);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_20_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_20_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_21_walk_types);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_21_walk_types);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_22_get_namespaced_aliases);
  Py_CLEAR(clear_module_state->__pyx_ptype_5cavro___pyx_scope_struct_23_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_5cavro___pyx_scope_struct_23_genexpr);
  Py_CLEAR(clear_module_state->__pyx_array_type);
  Py_CLEAR(clear_module_state->__pyx_type___pyx_array);
  Py_CLEAR(clear_module_state->__pyx_MemviewEnum_type);
  Py_CLEAR(clear_module_state->__pyx_type___pyx_MemviewEnum);
  Py_CLEAR(clear_module_state->__pyx_memoryview_type);
  Py_CLEAR(clear_module_state->__pyx_type___pyx_memoryview);
  Py_CLEAR(clear_module_state->__pyx_memoryviewslice_type);
  Py_CLEAR(clear_module_state->__pyx_type___pyx_memoryviewslice);
  Py_CLEAR(clear_module_state->__pyx_kp_u_);
  Py_CLEAR(clear_module_state->__pyx_kp_u_0);
  Py_CLEAR(clear_module_state->__pyx_kp_u_1);
  Py_CLEAR(clear_module_state->__pyx_kp_u_1_0_0);
  Py_CLEAR(clear_module_state->__pyx_n_s_AMBIGUOUS_TYPES);
  Py_CLEAR(clear_module_state->__pyx_n_s_ASC);
  Py_CLEAR(clear_module_state->__pyx_n_s_ASCII);
  Py_CLEAR(clear_module_state->__pyx_kp_u_A_Za_z__A_Za_z0_9);
  Py_CLEAR(clear_module_state->__pyx_kp_s_A_record_contains_multiple_fiel);
  Py_CLEAR(clear_module_state->__pyx_kp_s_A_requested_codec_or_codec_in_a);
  Py_CLEAR(clear_module_state->__pyx_kp_s_A_schema_cannot_be_promoted_to);
  Py_CLEAR(clear_module_state->__pyx_kp_s_A_value_is_invalid_for_a_given);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Aliases_must_be_a_list_tuple_set);
  Py_CLEAR(clear_module_state->__pyx_kp_s_All_dimensions_preceding_dimensi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_An_unknown_hash_method_was_requ);
  Py_CLEAR(clear_module_state->__pyx_n_s_ArrayType);
  Py_CLEAR(clear_module_state->__pyx_n_s_ArrayType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ArrayType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ArrayType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_ArrayType_convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_ArrayType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_ArrayType_walk_types);
  Py_CLEAR(clear_module_state->__pyx_kp_s_As_of_cython_3_0_2_dataclasses);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_AttributeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvroType);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvroType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvroType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvroType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvroType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvroType_convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvroType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvroType_for_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvroType_for_source);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvroType_get_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvroType_walk_types);
  Py_CLEAR(clear_module_state->__pyx_n_u_B);
  Py_CLEAR(clear_module_state->__pyx_n_s_BATCH_SIZE);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Base_class_for_exceptions_raise);
  Py_CLEAR(clear_module_state->__pyx_n_s_BoolType);
  Py_CLEAR(clear_module_state->__pyx_n_s_BoolType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BoolType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BoolType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_BoolType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_BoolType_copy);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Buffer_view_does_not_expose_stri);
  Py_CLEAR(clear_module_state->__pyx_n_s_BytesType);
  Py_CLEAR(clear_module_state->__pyx_n_s_BytesType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BytesType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BytesType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_BytesType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_BytesType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_Bzip2Codec);
  Py_CLEAR(clear_module_state->__pyx_n_s_Bzip2Codec___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Bzip2Codec___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_CANONICAL_FORM_KEYS);
  Py_CLEAR(clear_module_state->__pyx_n_s_CODECS);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Can_only_create_a_buffer_that_is);
  Py_CLEAR(clear_module_state->__pyx_n_s_CannotPromote);
  Py_CLEAR(clear_module_state->__pyx_n_s_CannotPromoteError);
  Py_CLEAR(clear_module_state->__pyx_n_s_CannotPromoteError___init);
  Py_CLEAR(clear_module_state->__pyx_n_s_CannotPromote___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_CannotPromote___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_assign_to_read_only_memor);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_create_writable_memory_vi);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_index_with_type);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_promote);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_read_from);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_transpose_memoryview_with);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_write_to);
  Py_CLEAR(clear_module_state->__pyx_n_s_CanonicalForm);
  Py_CLEAR(clear_module_state->__pyx_n_s_CanonicalForm___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_CanonicalForm___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_CavroException);
  Py_CLEAR(clear_module_state->__pyx_n_s_Codec);
  Py_CLEAR(clear_module_state->__pyx_n_s_CodecUnavailable);
  Py_CLEAR(clear_module_state->__pyx_n_s_Codec___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Codec___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerReader);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerReader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerReader___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerReader__read_marker);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerReader_next_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerWriter);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerWriter___enter);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerWriter___exit);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerWriter___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerWriter___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerWriter_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerWriter_flush);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerWriter_write_many);
  Py_CLEAR(clear_module_state->__pyx_n_s_ContainerWriter_write_one);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Container_is_closed);
  Py_CLEAR(clear_module_state->__pyx_n_s_Context);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Controls_how_union_values_are_d);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Could_not_find_key_type_in_schem);
  Py_CLEAR(clear_module_state->__pyx_n_s_CustomLogicalType);
  Py_CLEAR(clear_module_state->__pyx_n_s_CustomLogicalType___reduce_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_CustomLogicalType___setstate_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_DEFAULT_OPTIONS);
  Py_CLEAR(clear_module_state->__pyx_n_s_DESC);
  Py_CLEAR(clear_module_state->__pyx_n_s_DataclassParams);
  Py_CLEAR(clear_module_state->__pyx_n_s_Date);
  Py_CLEAR(clear_module_state->__pyx_n_s_Date___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Date___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Date__for_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_Decimal);
  Py_CLEAR(clear_module_state->__pyx_n_s_DecimalType);
  Py_CLEAR(clear_module_state->__pyx_n_s_DecimalType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_DecimalType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_DecimalType__for_type);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Decimal_value);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Default_value);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Default_value_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_DeflateCodec);
  Py_CLEAR(clear_module_state->__pyx_n_s_DeflateCodec___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_DeflateCodec___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Dimension_d_is_not_direct);
  Py_CLEAR(clear_module_state->__pyx_n_s_DoubleType);
  Py_CLEAR(clear_module_state->__pyx_n_s_DoubleType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_DoubleType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_DoubleType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_DoubleType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_DoubleType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_DuplicateName);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Duplicate_field_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_EMPTY_ARRAY);
  Py_CLEAR(clear_module_state->__pyx_n_s_EMPTY_MP);
  Py_CLEAR(clear_module_state->__pyx_n_s_EOFError);
  Py_CLEAR(clear_module_state->__pyx_n_s_EPOCH_DATE);
  Py_CLEAR(clear_module_state->__pyx_n_s_EPOCH_DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_EPOCH_DT_NO_TZ);
  Py_CLEAR(clear_module_state->__pyx_n_s_Ellipsis);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Empty_shape_tuple_for_cython_arr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_End_of_file_found_trying_to_read);
  Py_CLEAR(clear_module_state->__pyx_kp_u_End_of_file_found_trying_to_read_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Enum);
  Py_CLEAR(clear_module_state->__pyx_n_s_EnumType);
  Py_CLEAR(clear_module_state->__pyx_n_s_EnumType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_EnumType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_EnumType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_EnumType__for_writer_locals_gene);
  Py_CLEAR(clear_module_state->__pyx_n_s_EnumType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_EnumType_copy);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Enum_symbol);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Enum_symbols_must_be_a_list_of_s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Enum_symbols_must_be_unique);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Error_closing_container_file_dur);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Expected_datetime_date_got);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Expected_datetime_datetime_got);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Expected_null_got);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExponentTooLarge);
  Py_CLEAR(clear_module_state->__pyx_n_s_FIELD);
  Py_CLEAR(clear_module_state->__pyx_n_u_False);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Field);
  Py_CLEAR(clear_module_state->__pyx_n_s_FieldAccessor);
  Py_CLEAR(clear_module_state->__pyx_n_s_FieldAccessor___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FieldAccessor___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FileReader);
  Py_CLEAR(clear_module_state->__pyx_n_s_FileReader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FileReader___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FileWriter);
  Py_CLEAR(clear_module_state->__pyx_n_s_FileWriter___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FileWriter___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FixedType);
  Py_CLEAR(clear_module_state->__pyx_n_s_FixedType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FixedType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FixedType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_FixedType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_FixedType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloatType);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloatType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloatType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloatType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloatType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_FloatType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_HAVE_BZIP2);
  Py_CLEAR(clear_module_state->__pyx_n_s_HAVE_LZ4);
  Py_CLEAR(clear_module_state->__pyx_n_s_HAVE_SNAPPY);
  Py_CLEAR(clear_module_state->__pyx_n_s_HAVE_XZ);
  Py_CLEAR(clear_module_state->__pyx_n_s_HAVE_ZLIB);
  Py_CLEAR(clear_module_state->__pyx_n_s_HAVE_ZSTD);
  Py_CLEAR(clear_module_state->__pyx_n_s_IGNORE);
  Py_CLEAR(clear_module_state->__pyx_n_s_ImportError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_25);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_26);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_27);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_28);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Index_out_of_bounds_axis_d);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Indirect_dimensions_not_supporte);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntType);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_InvalidHasher);
  Py_CLEAR(clear_module_state->__pyx_n_s_InvalidName);
  Py_CLEAR(clear_module_state->__pyx_n_s_InvalidValue);
  Py_CLEAR(clear_module_state->__pyx_n_s_InvalidValue___init);
  Py_CLEAR(clear_module_state->__pyx_n_s_InvalidValue___str);
  Py_CLEAR(clear_module_state->__pyx_n_s_InvalidValue___str___locals_gene);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_block_sync_marker_expect);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_file_header_expected);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_length_for_fixed_field);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_mode_expected_c_or_fortr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_negative_size_for_fixed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_non_bytes_value_for_fixe);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_non_str_value_for_fixed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_shape_in_axis);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_union_decodes_to_option);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_Iterable);
  Py_CLEAR(clear_module_state->__pyx_n_s_KeyError);
  Py_CLEAR(clear_module_state->__pyx_n_s_LOGICAL_TYPES);
  Py_CLEAR(clear_module_state->__pyx_n_u_Ll);
  Py_CLEAR(clear_module_state->__pyx_n_s_LogicalType);
  Py_CLEAR(clear_module_state->__pyx_n_s_LogicalType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_LogicalType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_LogicalType_for_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_LongType);
  Py_CLEAR(clear_module_state->__pyx_n_s_LongType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_LongType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_LongType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_LongType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_LongType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_u_Lu);
  Py_CLEAR(clear_module_state->__pyx_n_s_Lz4Codec);
  Py_CLEAR(clear_module_state->__pyx_n_s_Lz4Codec___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Lz4Codec___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_LzmaCodec);
  Py_CLEAR(clear_module_state->__pyx_n_s_LzmaCodec___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_LzmaCodec___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_MAX_PREC);
  Py_CLEAR(clear_module_state->__pyx_n_s_MAX_TIME);
  Py_CLEAR(clear_module_state->__pyx_n_s_MISSING_VALUE);
  Py_CLEAR(clear_module_state->__pyx_n_u_MISSING_VALUE);
  Py_CLEAR(clear_module_state->__pyx_n_s_MapType);
  Py_CLEAR(clear_module_state->__pyx_n_s_MapType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_MapType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_MapType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_MapType_convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_MapType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_MapType_walk_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_MappingProxyType);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Marker_must_be_bytes_got);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Marker_must_be_exactly_16_bytes);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Max_precision_is);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryError);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryReader);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryReader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryReader___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_s_MemoryView_of_r_at_0x_x);
  Py_CLEAR(clear_module_state->__pyx_kp_s_MemoryView_of_r_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryWriter);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryWriter___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryWriter___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_NAME_RE_STRICT);
  Py_CLEAR(clear_module_state->__pyx_n_s_NAME_RE_UNICODE);
  Py_CLEAR(clear_module_state->__pyx_n_s_NO_DEFAULT);
  Py_CLEAR(clear_module_state->__pyx_n_u_NO_DEFAULT);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Name);
  Py_CLEAR(clear_module_state->__pyx_n_s_NamedType);
  Py_CLEAR(clear_module_state->__pyx_n_s_NamedType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_NamedType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_NamedType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_NamedType_get_namespaced_aliase);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Namespace);
  Py_CLEAR(clear_module_state->__pyx_kp_u_None);
  Py_CLEAR(clear_module_state->__pyx_n_s_NotImplemented);
  Py_CLEAR(clear_module_state->__pyx_n_s_NotImplementedError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Not_enough_input_data_to_read_va);
  Py_CLEAR(clear_module_state->__pyx_n_s_NullCodec);
  Py_CLEAR(clear_module_state->__pyx_n_s_NullCodec___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_NullCodec___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_NullType);
  Py_CLEAR(clear_module_state->__pyx_n_s_NullType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_NullType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_NullType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_NullType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_NullType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_b_O);
  Py_CLEAR(clear_module_state->__pyx_n_s_OBJ_FILE_METADATA);
  Py_CLEAR(clear_module_state->__pyx_n_s_OBJ_MAGIC_BYTES);
  Py_CLEAR(clear_module_state->__pyx_kp_b_Obj);
  Py_CLEAR(clear_module_state->__pyx_n_s_Options);
  Py_CLEAR(clear_module_state->__pyx_n_s_OptionsWrapper);
  Py_CLEAR(clear_module_state->__pyx_n_s_Options___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Options___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Options_diff);
  Py_CLEAR(clear_module_state->__pyx_n_s_Options_equals);
  Py_CLEAR(clear_module_state->__pyx_n_s_Options_replace);
  Py_CLEAR(clear_module_state->__pyx_n_s_Options_with_external_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_Options_with_logical_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_Order);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
  Py_CLEAR(clear_module_state->__pyx_n_s_OverflowError);
  Py_CLEAR(clear_module_state->__pyx_n_s_PERMISSIVE_OPTIONS);
  Py_CLEAR(clear_module_state->__pyx_n_s_PRIMITIVE_TYPES);
  Py_CLEAR(clear_module_state->__pyx_n_s_Path);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_PlaceholderType);
  Py_CLEAR(clear_module_state->__pyx_n_s_PlaceholderType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PlaceholderType___setstate_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_PlaceholderType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_PlaceholderType__get_schema_ext);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Placeholder_types_cannot_be_enco);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Placeholder_types_have_no_canoni);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Precision);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Precision_is_too_large_for_fixed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Precision_must_be_an_integer_bet);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Precision_must_be_greater_than_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromoteBytesToString);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromoteBytesToString___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromoteBytesToString___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromoteStringToBytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromoteStringToBytes___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromoteStringToBytes___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromoteToFloat);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromoteToFloat___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromoteToFloat___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromotingEnumType);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromotingEnumType___reduce_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromotingEnumType___setstate_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromotingRecordType);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromotingRecordType___reduce_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_PromotingRecordType___setstate_c);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Promoting_schemas_reader_writer);
  Py_CLEAR(clear_module_state->__pyx_n_s_RAW_VALUES);
  Py_CLEAR(clear_module_state->__pyx_n_s_Rabin);
  Py_CLEAR(clear_module_state->__pyx_n_s_Rabin___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Rabin___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Rabin_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_Rabin_digest);
  Py_CLEAR(clear_module_state->__pyx_n_s_Rabin_hexdigest);
  Py_CLEAR(clear_module_state->__pyx_n_s_Rabin_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Reader___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Reader___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Record);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecordField);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecordField___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecordField___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecordField_get_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecordType);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecordType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecordType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecordType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecordType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecordType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_RecordType_walk_types);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Record_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Record_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_Record___dir);
  Py_CLEAR(clear_module_state->__pyx_n_s_Record___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Record___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Record__asdict);
  Py_CLEAR(clear_module_state->__pyx_n_s_Record__repr_children_locals_gen);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Records_may_either_be_instantiat);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Records_may_either_be_instantiat_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_ResolvedSchema);
  Py_CLEAR(clear_module_state->__pyx_n_s_ResolvedSchema___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ResolvedSchema___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ResourceWarning);
  Py_CLEAR(clear_module_state->__pyx_n_s_SLICE_TYPE);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSIZE_TYPECODE);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Scale_must_be_a_positive_integer);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema__make_logical_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema__wrap_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema_binary_decode);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema_binary_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema_binary_read);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema_binary_write);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema_can_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema_find_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema_fingerprint);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Schema_is_required);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema_json_decode);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema_json_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_Schema_reader_for_writer);
  Py_CLEAR(clear_module_state->__pyx_n_s_Sentinel);
  Py_CLEAR(clear_module_state->__pyx_n_s_Sentinel___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Sentinel___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Sequence);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Sequence_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_SignatureWrapper);
  Py_CLEAR(clear_module_state->__pyx_n_s_SignatureWrapper___init);
  Py_CLEAR(clear_module_state->__pyx_n_s_SignatureWrapper__set_from_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_SnappyCodec);
  Py_CLEAR(clear_module_state->__pyx_n_s_SnappyCodec___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SnappyCodec___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Step_may_not_be_zero_axis_d);
  Py_CLEAR(clear_module_state->__pyx_n_s_StopIteration);
  Py_CLEAR(clear_module_state->__pyx_n_s_StringType);
  Py_CLEAR(clear_module_state->__pyx_n_s_StringType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_StringType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_StringType__convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_StringType__get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_StringType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_TYPES_BY_NAME);
  Py_CLEAR(clear_module_state->__pyx_n_s_TYPE_TUPLE_ALWAYS);
  Py_CLEAR(clear_module_state->__pyx_n_s_TYPE_TUPLE_IF_AMBIGUOUS);
  Py_CLEAR(clear_module_state->__pyx_n_s_TYPE_TUPLE_IF_RECORD);
  Py_CLEAR(clear_module_state->__pyx_n_s_TYPE_TUPLE_IF_RECORD_AMBIGUOUS);
  Py_CLEAR(clear_module_state->__pyx_kp_s_The_exponent_of_a_decimal_value);
  Py_CLEAR(clear_module_state->__pyx_kp_u_The_null_namespace_cannot_be_spe);
  Py_CLEAR(clear_module_state->__pyx_kp_s_The_order_of_a_field_in_a_recor);
  Py_CLEAR(clear_module_state->__pyx_kp_s_The_schema_contains_a_type_or_e);
  Py_CLEAR(clear_module_state->__pyx_kp_s_The_schema_contains_an_unexptec);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimeMicros);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimeMicros___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimeMicros___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimeMicros__for_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimeMillis);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimeMillis___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimeMillis___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimeMillis__for_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimestampMicros);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimestampMicros___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimestampMicros___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimestampMicros__for_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimestampMillis);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimestampMillis___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimestampMillis___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimestampMillis__for_type);
  Py_CLEAR(clear_module_state->__pyx_n_u_True);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Trying_to_close_a_closed_Contain);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Trying_to_write_to_closed_Contai);
  Py_CLEAR(clear_module_state->__pyx_n_s_Type);
  Py_CLEAR(clear_module_state->__pyx_n_u_Type);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Type_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_UNKNOWN);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUID);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUIDBase);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUIDBase___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUIDBase___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUIDBase__for_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUIDFixedType);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUIDFixedType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUIDFixedType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUIDFixedType__for_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUIDStringType);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUIDStringType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUIDStringType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UUIDStringType__for_type);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Unable_to_convert_item_to_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnicodeEncodeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Union);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnionDecodeOption);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnionType);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnionType___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnionType___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnionType_convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnionType_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnionType_get_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnionType_walk_types);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Union_bytes_hashlib__hashlib_HAS);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Union_str_object);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unions_may_not_have_more_than_on);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unions_may_not_immediately_conta);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unions_must_contain_at_least_one);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnknownType);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnknownType___init);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unknown_hash_method);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unknown_type);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unknown_value_for_enum);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unsupported_codec);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Value);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueAdapter);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueAdapter___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueAdapter___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_n_s_View_MemoryView);
  Py_CLEAR(clear_module_state->__pyx_n_s_Writer);
  Py_CLEAR(clear_module_state->__pyx_n_s_Writer___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Writer___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Y_m_d);
  Py_CLEAR(clear_module_state->__pyx_n_s_ZStandardCodec);
  Py_CLEAR(clear_module_state->__pyx_n_s_ZStandardCodec___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ZStandardCodec___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_ZstdCompressor);
  Py_CLEAR(clear_module_state->__pyx_n_s_ZstdDecompressor);
  Py_CLEAR(clear_module_state->__pyx_n_s__122);
  Py_CLEAR(clear_module_state->__pyx_kp_b__13);
  Py_CLEAR(clear_module_state->__pyx_kp_u__13);
  Py_CLEAR(clear_module_state->__pyx_kp_u__14);
  Py_CLEAR(clear_module_state->__pyx_n_s__152);
  Py_CLEAR(clear_module_state->__pyx_kp_s__18);
  Py_CLEAR(clear_module_state->__pyx_kp_u__18);
  Py_CLEAR(clear_module_state->__pyx_kp_u__2);
  Py_CLEAR(clear_module_state->__pyx_kp_u__24);
  Py_CLEAR(clear_module_state->__pyx_kp_u__25);
  Py_CLEAR(clear_module_state->__pyx_kp_u__26);
  Py_CLEAR(clear_module_state->__pyx_kp_u__27);
  Py_CLEAR(clear_module_state->__pyx_n_s__3);
  Py_CLEAR(clear_module_state->__pyx_kp_b__34);
  Py_CLEAR(clear_module_state->__pyx_kp_u__48);
  Py_CLEAR(clear_module_state->__pyx_kp_u__6);
  Py_CLEAR(clear_module_state->__pyx_kp_u__60);
  Py_CLEAR(clear_module_state->__pyx_kp_u__61);
  Py_CLEAR(clear_module_state->__pyx_kp_u__62);
  Py_CLEAR(clear_module_state->__pyx_kp_u__63);
  Py_CLEAR(clear_module_state->__pyx_kp_u__64);
  Py_CLEAR(clear_module_state->__pyx_kp_u__7);
  Py_CLEAR(clear_module_state->__pyx_kp_u__70);
  Py_CLEAR(clear_module_state->__pyx_kp_b__72);
  Py_CLEAR(clear_module_state->__pyx_kp_u__76);
  Py_CLEAR(clear_module_state->__pyx_kp_u__77);
  Py_CLEAR(clear_module_state->__pyx_kp_u__78);
  Py_CLEAR(clear_module_state->__pyx_kp_u__79);
  Py_CLEAR(clear_module_state->__pyx_kp_u__80);
  Py_CLEAR(clear_module_state->__pyx_kp_u__9);
  Py_CLEAR(clear_module_state->__pyx_n_s_a);
  Py_CLEAR(clear_module_state->__pyx_n_b_abacus);
  Py_CLEAR(clear_module_state->__pyx_n_b_abacuscounter1);
  Py_CLEAR(clear_module_state->__pyx_n_s_abc);
  Py_CLEAR(clear_module_state->__pyx_n_s_accepts_string);
  Py_CLEAR(clear_module_state->__pyx_n_s_actual);
  Py_CLEAR(clear_module_state->__pyx_n_s_adapt_record_types);
  Py_CLEAR(clear_module_state->__pyx_kp_u_adapt_record_types_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_add);
  Py_CLEAR(clear_module_state->__pyx_n_u_aliases);
  Py_CLEAR(clear_module_state->__pyx_n_s_allocate_buffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_aliases_to_be_string);
  Py_CLEAR(clear_module_state->__pyx_kp_u_allow_aliases_to_be_string_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_duplicate_union_types);
  Py_CLEAR(clear_module_state->__pyx_kp_u_allow_duplicate_union_types_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_empty_unions);
  Py_CLEAR(clear_module_state->__pyx_kp_u_allow_empty_unions_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_error_type);
  Py_CLEAR(clear_module_state->__pyx_kp_u_allow_error_type_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_false_values_for_null);
  Py_CLEAR(clear_module_state->__pyx_kp_u_allow_false_values_for_null_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_invalid_default_values);
  Py_CLEAR(clear_module_state->__pyx_kp_u_allow_invalid_default_values_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_leading_dot_in_names);
  Py_CLEAR(clear_module_state->__pyx_kp_u_allow_leading_dot_in_names_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_nested_unions);
  Py_CLEAR(clear_module_state->__pyx_kp_u_allow_nested_unions_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_primitive_name_collision);
  Py_CLEAR(clear_module_state->__pyx_kp_u_allow_primitive_name_collision_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_allow_primitive_names_in_namesp);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_primitive_names_in_namespa);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_tuple_notation);
  Py_CLEAR(clear_module_state->__pyx_kp_u_allow_tuple_notation_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_union_default_any_member);
  Py_CLEAR(clear_module_state->__pyx_kp_u_allow_union_default_any_member_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_alternate_timestamp_encoding);
  Py_CLEAR(clear_module_state->__pyx_kp_u_alternate_timestamp_millis_enco);
  Py_CLEAR(clear_module_state->__pyx_n_s_alternate_timestamp_millis_encod);
  Py_CLEAR(clear_module_state->__pyx_kp_u_and);
  Py_CLEAR(clear_module_state->__pyx_n_s_annotations);
  Py_CLEAR(clear_module_state->__pyx_n_s_annotations_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_appears_multiple_times_in_schem);
  Py_CLEAR(clear_module_state->__pyx_n_s_append);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_array);
  Py_CLEAR(clear_module_state->__pyx_n_u_array);
  Py_CLEAR(clear_module_state->__pyx_n_s_as_tuple);
  Py_CLEAR(clear_module_state->__pyx_n_u_ascending);
  Py_CLEAR(clear_module_state->__pyx_n_s_ascii_name_rules);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ascii_name_rules_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_asdict);
  Py_CLEAR(clear_module_state->__pyx_n_u_asdict);
  Py_CLEAR(clear_module_state->__pyx_n_s_astimezone);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_kp_u_at);
  Py_CLEAR(clear_module_state->__pyx_kp_u_avro_codec);
  Py_CLEAR(clear_module_state->__pyx_kp_u_avro_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_avro_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_b);
  Py_CLEAR(clear_module_state->__pyx_n_s_base);
  Py_CLEAR(clear_module_state->__pyx_n_u_big);
  Py_CLEAR(clear_module_state->__pyx_n_s_bin);
  Py_CLEAR(clear_module_state->__pyx_n_s_binary_decode);
  Py_CLEAR(clear_module_state->__pyx_n_s_binary_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_binary_read);
  Py_CLEAR(clear_module_state->__pyx_n_s_binary_write);
  Py_CLEAR(clear_module_state->__pyx_n_s_bint);
  Py_CLEAR(clear_module_state->__pyx_n_u_bint);
  Py_CLEAR(clear_module_state->__pyx_n_s_bit_length);
  Py_CLEAR(clear_module_state->__pyx_n_s_block_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_bool);
  Py_CLEAR(clear_module_state->__pyx_n_u_bool);
  Py_CLEAR(clear_module_state->__pyx_n_s_bool_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_boolean);
  Py_CLEAR(clear_module_state->__pyx_n_u_boolean_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_buffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_byteorder);
  Py_CLEAR(clear_module_state->__pyx_kp_u_bytes);
  Py_CLEAR(clear_module_state->__pyx_kp_u_bytes_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_bytes_3);
  Py_CLEAR(clear_module_state->__pyx_n_u_bytes_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_bytes_codec);
  Py_CLEAR(clear_module_state->__pyx_kp_u_bytes_codec_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_bytes_default_value_utf8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_bytes_default_value_utf8_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_bz2);
  Py_CLEAR(clear_module_state->__pyx_n_b_bzip2);
  Py_CLEAR(clear_module_state->__pyx_n_u_bzip2);
  Py_CLEAR(clear_module_state->__pyx_n_s_c);
  Py_CLEAR(clear_module_state->__pyx_n_u_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_can_encode);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cannot_be_adapted_to);
  Py_CLEAR(clear_module_state->__pyx_n_s_canonical_form);
  Py_CLEAR(clear_module_state->__pyx_n_s_canonical_form_repeat_enum);
  Py_CLEAR(clear_module_state->__pyx_kp_u_canonical_form_repeat_enum_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_canonical_form_repeat_fixed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_canonical_form_repeat_fixed_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_category);
  Py_CLEAR(clear_module_state->__pyx_n_s_cavro);
  Py_CLEAR(clear_module_state->__pyx_kp_s_cavro_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_changes);
  Py_CLEAR(clear_module_state->__pyx_n_s_check_exp_overflow);
  Py_CLEAR(clear_module_state->__pyx_n_s_check_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_chr);
  Py_CLEAR(clear_module_state->__pyx_n_s_clamp);
  Py_CLEAR(clear_module_state->__pyx_n_s_clamp_float_overflow);
  Py_CLEAR(clear_module_state->__pyx_kp_u_clamp_float_overflow_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_clamp_int_overflow);
  Py_CLEAR(clear_module_state->__pyx_kp_u_clamp_int_overflow_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_inst_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_inst_method___get);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_inst_method___init);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_cls);
  Py_CLEAR(clear_module_state->__pyx_n_s_codec);
  Py_CLEAR(clear_module_state->__pyx_n_s_coerce_int_to_float);
  Py_CLEAR(clear_module_state->__pyx_kp_u_coerce_int_to_float_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_coerce_values_to_boolean);
  Py_CLEAR(clear_module_state->__pyx_kp_u_coerce_values_to_boolean_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_coerce_values_to_float);
  Py_CLEAR(clear_module_state->__pyx_kp_u_coerce_values_to_float_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_coerce_values_to_int);
  Py_CLEAR(clear_module_state->__pyx_kp_u_coerce_values_to_int_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_coerce_values_to_str);
  Py_CLEAR(clear_module_state->__pyx_kp_u_coerce_values_to_str_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections);
  Py_CLEAR(clear_module_state->__pyx_kp_s_collections_abc);
  Py_CLEAR(clear_module_state->__pyx_n_s_compare);
  Py_CLEAR(clear_module_state->__pyx_n_s_compile);
  Py_CLEAR(clear_module_state->__pyx_n_s_compress);
  Py_CLEAR(clear_module_state->__pyx_n_s_container_fill_blocks);
  Py_CLEAR(clear_module_state->__pyx_kp_u_container_fill_blocks_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_context);
  Py_CLEAR(clear_module_state->__pyx_kp_s_contiguous_and_direct);
  Py_CLEAR(clear_module_state->__pyx_kp_s_contiguous_and_indirect);
  Py_CLEAR(clear_module_state->__pyx_n_s_convert_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_convert_value_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_count);
  Py_CLEAR(clear_module_state->__pyx_n_b_counter);
  Py_CLEAR(clear_module_state->__pyx_n_s_crc32);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_decimal);
  Py_CLEAR(clear_module_state->__pyx_n_s_created);
  Py_CLEAR(clear_module_state->__pyx_n_s_custom_decode_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_custom_encode_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_dataclass_fields);
  Py_CLEAR(clear_module_state->__pyx_n_s_dataclass_params);
  Py_CLEAR(clear_module_state->__pyx_n_s_dataclasses);
  Py_CLEAR(clear_module_state->__pyx_n_s_date);
  Py_CLEAR(clear_module_state->__pyx_n_u_date);
  Py_CLEAR(clear_module_state->__pyx_n_s_date_type_accepts_string);
  Py_CLEAR(clear_module_state->__pyx_kp_u_date_type_accepts_string_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_datetime);
  Py_CLEAR(clear_module_state->__pyx_n_s_days);
  Py_CLEAR(clear_module_state->__pyx_n_s_decimal);
  Py_CLEAR(clear_module_state->__pyx_n_u_decimal);
  Py_CLEAR(clear_module_state->__pyx_kp_u_decimal_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_decimal_check_exp_overflow);
  Py_CLEAR(clear_module_state->__pyx_kp_u_decimal_check_exp_overflow_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_decode);
  Py_CLEAR(clear_module_state->__pyx_n_s_decompress);
  Py_CLEAR(clear_module_state->__pyx_n_s_decompressobj);
  Py_CLEAR(clear_module_state->__pyx_n_s_defaul_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_default);
  Py_CLEAR(clear_module_state->__pyx_n_u_default);
  Py_CLEAR(clear_module_state->__pyx_n_s_default_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_defer_schema_promotion_errors);
  Py_CLEAR(clear_module_state->__pyx_kp_u_defer_schema_promotion_errors_2);
  Py_CLEAR(clear_module_state->__pyx_n_b_deflate);
  Py_CLEAR(clear_module_state->__pyx_n_u_deflate);
  Py_CLEAR(clear_module_state->__pyx_n_u_descending);
  Py_CLEAR(clear_module_state->__pyx_n_s_deserialize);
  Py_CLEAR(clear_module_state->__pyx_n_s_dest);
  Py_CLEAR(clear_module_state->__pyx_n_s_dest_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_dict_str_AvroType);
  Py_CLEAR(clear_module_state->__pyx_kp_s_dict_str_tuple_object_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_to_canonical_locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_diff);
  Py_CLEAR(clear_module_state->__pyx_n_s_digest);
  Py_CLEAR(clear_module_state->__pyx_n_s_digest_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_dir);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_u_doc);
  Py_CLEAR(clear_module_state->__pyx_n_s_doc_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement__binary_buff);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement__convert_val);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement__extract_met);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement__get_schema);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement__get_value_f);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement__json_decode);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement__json_format);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement_binary_buffe);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement_canonical_fo);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement_clone);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement_decode_value);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement_encode_value);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement_read_block);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement_read_bytes);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement_read_n);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement_read_u8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement_write_block);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement_write_n);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_implement_write_u8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_double);
  Py_CLEAR(clear_module_state->__pyx_n_u_double_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_dtype_is_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_dumps);
  Py_CLEAR(clear_module_state->__pyx_n_s_e);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_enforce_enum_symbol_name_rules);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enforce_enum_symbol_name_rules_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_enforce_namespace_name_rules);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enforce_namespace_name_rules_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_enforce_type_name_rules);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enforce_type_name_rules_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_ensure_ascii);
  Py_CLEAR(clear_module_state->__pyx_n_s_enter);
  Py_CLEAR(clear_module_state->__pyx_n_s_enum);
  Py_CLEAR(clear_module_state->__pyx_n_u_enum);
  Py_CLEAR(clear_module_state->__pyx_n_s_enum_symbols_must_be_unique);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enum_symbols_must_be_unique_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_eq);
  Py_CLEAR(clear_module_state->__pyx_n_s_equals);
  Py_CLEAR(clear_module_state->__pyx_n_s_error);
  Py_CLEAR(clear_module_state->__pyx_n_u_error);
  Py_CLEAR(clear_module_state->__pyx_n_s_errors);
  Py_CLEAR(clear_module_state->__pyx_n_s_escape);
  Py_CLEAR(clear_module_state->__pyx_n_s_eval);
  Py_CLEAR(clear_module_state->__pyx_n_s_eval_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_exc_tb);
  Py_CLEAR(clear_module_state->__pyx_n_s_exc_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_exc_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_exit);
  Py_CLEAR(clear_module_state->__pyx_n_s_expand_types_in_schema);
  Py_CLEAR(clear_module_state->__pyx_kp_u_expand_types_in_schema_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_externally_defined_types);
  Py_CLEAR(clear_module_state->__pyx_kp_u_externally_defined_types_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_f);
  Py_CLEAR(clear_module_state->__pyx_n_s_field);
  Py_CLEAR(clear_module_state->__pyx_n_s_field_to_index);
  Py_CLEAR(clear_module_state->__pyx_n_u_field_to_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_field_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_fields);
  Py_CLEAR(clear_module_state->__pyx_n_u_fields);
  Py_CLEAR(clear_module_state->__pyx_n_s_file_obj);
  Py_CLEAR(clear_module_state->__pyx_n_s_find_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_fingerprint);
  Py_CLEAR(clear_module_state->__pyx_n_s_fingerprint_returns_digest);
  Py_CLEAR(clear_module_state->__pyx_kp_u_fingerprint_returns_digest_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_fitness);
  Py_CLEAR(clear_module_state->__pyx_n_s_fixed);
  Py_CLEAR(clear_module_state->__pyx_n_u_fixed);
  Py_CLEAR(clear_module_state->__pyx_n_s_fixed_codec);
  Py_CLEAR(clear_module_state->__pyx_kp_u_fixed_codec_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_fixed_uuid);
  Py_CLEAR(clear_module_state->__pyx_n_s_flags);
  Py_CLEAR(clear_module_state->__pyx_kp_u_float);
  Py_CLEAR(clear_module_state->__pyx_n_s_float16);
  Py_CLEAR(clear_module_state->__pyx_n_s_float32);
  Py_CLEAR(clear_module_state->__pyx_n_s_float64);
  Py_CLEAR(clear_module_state->__pyx_n_u_float_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_float_out_of_range_inf);
  Py_CLEAR(clear_module_state->__pyx_kp_u_float_out_of_range_inf_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_floor);
  Py_CLEAR(clear_module_state->__pyx_n_s_flush);
  Py_CLEAR(clear_module_state->__pyx_n_s_fn);
  Py_CLEAR(clear_module_state->__pyx_kp_u_for_field);
  Py_CLEAR(clear_module_state->__pyx_n_s_for_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_for_source);
  Py_CLEAR(clear_module_state->__pyx_kp_u_for_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_for_type_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_for_type_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_force);
  Py_CLEAR(clear_module_state->__pyx_n_s_format);
  Py_CLEAR(clear_module_state->__pyx_n_s_fortran);
  Py_CLEAR(clear_module_state->__pyx_n_u_fortran);
  Py_CLEAR(clear_module_state->__pyx_kp_u_found_twice);
  Py_CLEAR(clear_module_state->__pyx_n_s_frame);
  Py_CLEAR(clear_module_state->__pyx_n_s_from_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_frozen);
  Py_CLEAR(clear_module_state->__pyx_n_s_fullmatch);
  Py_CLEAR(clear_module_state->__pyx_n_s_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_functools);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_schema_extra);
  Py_CLEAR(clear_module_state->__pyx_kp_u_getfullargspec);
  Py_CLEAR(clear_module_state->__pyx_n_u_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_given);
  Py_CLEAR(clear_module_state->__pyx_kp_u_got);
  Py_CLEAR(clear_module_state->__pyx_kp_u_got_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_got_3);
  Py_CLEAR(clear_module_state->__pyx_kp_u_got_4);
  Py_CLEAR(clear_module_state->__pyx_kp_u_got_differing_extents_in_dimensi);
  Py_CLEAR(clear_module_state->__pyx_kp_u_has_exponent);
  Py_CLEAR(clear_module_state->__pyx_n_s_hash);
  Py_CLEAR(clear_module_state->__pyx_n_s_hasher);
  Py_CLEAR(clear_module_state->__pyx_n_s_hashlib);
  Py_CLEAR(clear_module_state->__pyx_n_s_hex);
  Py_CLEAR(clear_module_state->__pyx_n_s_hexdigest);
  Py_CLEAR(clear_module_state->__pyx_n_s_hour);
  Py_CLEAR(clear_module_state->__pyx_n_s_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_ignore);
  Py_CLEAR(clear_module_state->__pyx_n_u_ignore);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_indent);
  Py_CLEAR(clear_module_state->__pyx_n_s_index);
  Py_CLEAR(clear_module_state->__pyx_kp_u_inf);
  Py_CLEAR(clear_module_state->__pyx_n_u_inf_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_init);
  Py_CLEAR(clear_module_state->__pyx_n_s_init_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_init___locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_init_subclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_initial_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_inline_namespaces);
  Py_CLEAR(clear_module_state->__pyx_kp_u_inline_namespaces_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_inspect);
  Py_CLEAR(clear_module_state->__pyx_n_s_inst);
  Py_CLEAR(clear_module_state->__pyx_kp_u_int);
  Py_CLEAR(clear_module_state->__pyx_n_u_int_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_int_date);
  Py_CLEAR(clear_module_state->__pyx_kp_u_int_time_millis);
  Py_CLEAR(clear_module_state->__pyx_n_s_integer);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_for_enum);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_value_include_array_ind);
  Py_CLEAR(clear_module_state->__pyx_n_s_invalid_value_include_array_inde);
  Py_CLEAR(clear_module_state->__pyx_n_s_invalid_value_include_map_key);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_value_include_map_key_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_value_includes_record_n);
  Py_CLEAR(clear_module_state->__pyx_n_s_invalid_value_includes_record_na);
  Py_CLEAR(clear_module_state->__pyx_n_s_io);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_a_mapping);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_a_valid_key_for_map_type);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_a_valid_name);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_a_valid_union_value_expe);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_a_valid_union_value_unkn);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_a_valid_value_for_array);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_a_valid_value_for_map_ty);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_allowed_as_a_name);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_valid);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_valid_for_a_union_of);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_valid_for_bytes_field);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_valid_for_field);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_valid_for_fixed_field);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_valid_for_union_with_fir);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_required_for_record);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_isinf);
  Py_CLEAR(clear_module_state->__pyx_n_s_isnan);
  Py_CLEAR(clear_module_state->__pyx_n_s_item_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_items);
  Py_CLEAR(clear_module_state->__pyx_n_u_items);
  Py_CLEAR(clear_module_state->__pyx_kp_u_items_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_itemsize);
  Py_CLEAR(clear_module_state->__pyx_kp_s_itemsize_0_for_cython_array);
  Py_CLEAR(clear_module_state->__pyx_n_s_json);
  Py_CLEAR(clear_module_state->__pyx_n_s_json_decode);
  Py_CLEAR(clear_module_state->__pyx_n_s_json_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_kw_only);
  Py_CLEAR(clear_module_state->__pyx_n_s_kwargs);
  Py_CLEAR(clear_module_state->__pyx_n_b_l);
  Py_CLEAR(clear_module_state->__pyx_n_u_latin1);
  Py_CLEAR(clear_module_state->__pyx_kp_u_latin_1);
  Py_CLEAR(clear_module_state->__pyx_n_u_little);
  Py_CLEAR(clear_module_state->__pyx_n_s_loads);
  Py_CLEAR(clear_module_state->__pyx_n_s_local);
  Py_CLEAR(clear_module_state->__pyx_n_s_log10);
  Py_CLEAR(clear_module_state->__pyx_n_u_logicalType);
  Py_CLEAR(clear_module_state->__pyx_n_s_logical_by_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_logical_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_logical_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_logical_types);
  Py_CLEAR(clear_module_state->__pyx_kp_u_logical_types_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_long);
  Py_CLEAR(clear_module_state->__pyx_n_u_long_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_long_time_micros);
  Py_CLEAR(clear_module_state->__pyx_kp_u_long_timestamp_micros);
  Py_CLEAR(clear_module_state->__pyx_kp_u_long_timestamp_millis);
  Py_CLEAR(clear_module_state->__pyx_n_b_lz4);
  Py_CLEAR(clear_module_state->__pyx_n_s_lz4);
  Py_CLEAR(clear_module_state->__pyx_n_u_lz4);
  Py_CLEAR(clear_module_state->__pyx_n_s_lz4_frame);
  Py_CLEAR(clear_module_state->__pyx_n_s_lzma);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_make_dataclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_make_logical_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_map);
  Py_CLEAR(clear_module_state->__pyx_n_u_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_marker);
  Py_CLEAR(clear_module_state->__pyx_n_s_mask);
  Py_CLEAR(clear_module_state->__pyx_n_s_match_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_math);
  Py_CLEAR(clear_module_state->__pyx_n_s_max);
  Py_CLEAR(clear_module_state->__pyx_n_s_max_blocksize);
  Py_CLEAR(clear_module_state->__pyx_n_s_max_precision);
  Py_CLEAR(clear_module_state->__pyx_n_s_maxunicode);
  Py_CLEAR(clear_module_state->__pyx_n_s_memview);
  Py_CLEAR(clear_module_state->__pyx_n_s_meta);
  Py_CLEAR(clear_module_state->__pyx_n_s_metaclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_metadata);
  Py_CLEAR(clear_module_state->__pyx_n_s_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_micros_to_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_microsecond);
  Py_CLEAR(clear_module_state->__pyx_n_s_microseconds);
  Py_CLEAR(clear_module_state->__pyx_n_s_minute);
  Py_CLEAR(clear_module_state->__pyx_kp_u_missing);
  Py_CLEAR(clear_module_state->__pyx_kp_u_missing_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_missing_values_can_be_empty_con);
  Py_CLEAR(clear_module_state->__pyx_n_s_missing_values_can_be_empty_cont);
  Py_CLEAR(clear_module_state->__pyx_n_s_missing_values_can_be_null);
  Py_CLEAR(clear_module_state->__pyx_kp_u_missing_values_can_be_null_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_module);
  Py_CLEAR(clear_module_state->__pyx_n_s_mro_entries);
  Py_CLEAR(clear_module_state->__pyx_n_s_msg);
  Py_CLEAR(clear_module_state->__pyx_n_s_n);
  Py_CLEAR(clear_module_state->__pyx_n_s_n_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_naive_dt_assume_utc);
  Py_CLEAR(clear_module_state->__pyx_kp_u_naive_dt_assume_utc_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_pattern);
  Py_CLEAR(clear_module_state->__pyx_n_s_named_type_names_must_be_unique);
  Py_CLEAR(clear_module_state->__pyx_kp_u_named_type_names_must_be_unique_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_named_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_namespace);
  Py_CLEAR(clear_module_state->__pyx_n_u_namespace);
  Py_CLEAR(clear_module_state->__pyx_n_s_namespaced);
  Py_CLEAR(clear_module_state->__pyx_n_u_nan);
  Py_CLEAR(clear_module_state->__pyx_n_s_ndarray);
  Py_CLEAR(clear_module_state->__pyx_n_s_ndim);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_n_s_new_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_new_inst);
  Py_CLEAR(clear_module_state->__pyx_n_s_new_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_new_ver);
  Py_CLEAR(clear_module_state->__pyx_n_s_next_object);
  Py_CLEAR(clear_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_CLEAR(clear_module_state->__pyx_kp_u_not_compatible_with_NullType);
  Py_CLEAR(clear_module_state->__pyx_kp_u_not_in_enum_symbols);
  Py_CLEAR(clear_module_state->__pyx_kp_u_not_valid_for);
  Py_CLEAR(clear_module_state->__pyx_kp_u_not_valid_for_long);
  Py_CLEAR(clear_module_state->__pyx_n_s_np);
  Py_CLEAR(clear_module_state->__pyx_n_s_np_f16);
  Py_CLEAR(clear_module_state->__pyx_n_s_np_f32);
  Py_CLEAR(clear_module_state->__pyx_n_s_np_f64);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ns);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ns_copy);
  Py_CLEAR(clear_module_state->__pyx_kp_u_null);
  Py_CLEAR(clear_module_state->__pyx_n_b_null_2);
  Py_CLEAR(clear_module_state->__pyx_n_u_null_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_null_pad_fixed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_null_pad_fixed_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_numpy);
  Py_CLEAR(clear_module_state->__pyx_n_s_ob);
  Py_CLEAR(clear_module_state->__pyx_n_s_obj);
  Py_CLEAR(clear_module_state->__pyx_n_s_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_objs);
  Py_CLEAR(clear_module_state->__pyx_n_s_open);
  Py_CLEAR(clear_module_state->__pyx_n_s_options);
  Py_CLEAR(clear_module_state->__pyx_n_s_order);
  Py_CLEAR(clear_module_state->__pyx_n_u_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_orig_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_os);
  Py_CLEAR(clear_module_state->__pyx_n_s_other);
  Py_CLEAR(clear_module_state->__pyx_n_s_other_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_out);
  Py_CLEAR(clear_module_state->__pyx_n_s_out_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pack);
  Py_CLEAR(clear_module_state->__pyx_n_s_pad);
  Py_CLEAR(clear_module_state->__pyx_n_s_parse_json);
  Py_CLEAR(clear_module_state->__pyx_n_s_parse_namespace);
  Py_CLEAR(clear_module_state->__pyx_n_s_partial);
  Py_CLEAR(clear_module_state->__pyx_n_s_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_path_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_pathlib);
  Py_CLEAR(clear_module_state->__pyx_n_s_perf);
  Py_CLEAR(clear_module_state->__pyx_n_s_perf_counter_ns);
  Py_CLEAR(clear_module_state->__pyx_n_s_perf_locals__test_zigzag_perf);
  Py_CLEAR(clear_module_state->__pyx_n_s_perf_locals_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_perf_locals_add_locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_perf_tests);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_u_placeholder);
  Py_CLEAR(clear_module_state->__pyx_n_s_pop);
  Py_CLEAR(clear_module_state->__pyx_n_s_prec);
  Py_CLEAR(clear_module_state->__pyx_n_s_precision);
  Py_CLEAR(clear_module_state->__pyx_n_u_precision);
  Py_CLEAR(clear_module_state->__pyx_kp_u_precision_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_prepare);
  Py_CLEAR(clear_module_state->__pyx_n_s_print);
  Py_CLEAR(clear_module_state->__pyx_n_s_pytest);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_recursive_repr_guard);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_ArrayType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_AvroType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_BoolType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_BytesType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_CannotPromote);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Codec);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_ContainerReader);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_CustomLogicalType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Date);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_DecimalType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_DoubleType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Enum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_EnumType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_FileReader);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_FileWriter);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_FixedType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_FloatType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_IntType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_LogicalType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_LongType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_MapType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_MemoryReader);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_NullType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Options);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PromoteBytesToStr);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PromoteStringToBy);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PromoteToFloat);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PromotingEnumType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PromotingRecordTy);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Rabin);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Record);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_RecordField);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_RecordType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_ResolvedSchema);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_StringType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_TimeMicros);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_TimeMillis);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_TimestampMicros);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_TimestampMillis);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_UUIDBase);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_UUIDFixedType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_UUIDStringType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_UnionType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_ValueAdapter);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__CanonicalForm);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__DeflateCodec);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__FieldAccessor);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__Lz4Codec);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__LzmaCodec);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__NamedType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__NullCodec);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__PlaceholderType);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__Reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__Sentinel);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__SnappyCodec);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__Writer);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle__ZStandardCodec);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle___Bzip2Codec);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_b_q);
  Py_CLEAR(clear_module_state->__pyx_n_s_qualname);
  Py_CLEAR(clear_module_state->__pyx_n_s_quantize);
  Py_CLEAR(clear_module_state->__pyx_n_u_rabin);
  Py_CLEAR(clear_module_state->__pyx_n_s_raise);
  Py_CLEAR(clear_module_state->__pyx_n_s_raise_on_invalid_logical);
  Py_CLEAR(clear_module_state->__pyx_kp_u_raise_on_invalid_logical_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_raises);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_u_rb);
  Py_CLEAR(clear_module_state->__pyx_n_s_re);
  Py_CLEAR(clear_module_state->__pyx_n_s_read);
  Py_CLEAR(clear_module_state->__pyx_n_u_read);
  Py_CLEAR(clear_module_state->__pyx_n_s_read_marker);
  Py_CLEAR(clear_module_state->__pyx_n_s_read_perf);
  Py_CLEAR(clear_module_state->__pyx_n_s_read_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_reader_for_writer);
  Py_CLEAR(clear_module_state->__pyx_n_s_reader_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_reader_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_record);
  Py_CLEAR(clear_module_state->__pyx_n_u_record);
  Py_CLEAR(clear_module_state->__pyx_n_s_record_allow_extra_fields);
  Py_CLEAR(clear_module_state->__pyx_kp_u_record_allow_extra_fields_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_record_can_encode_dict);
  Py_CLEAR(clear_module_state->__pyx_kp_u_record_can_encode_dict_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_record_decodes_to_dict);
  Py_CLEAR(clear_module_state->__pyx_kp_u_record_decodes_to_dict_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_record_encode_use_defaults);
  Py_CLEAR(clear_module_state->__pyx_kp_u_record_encode_use_defaults_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_record_fields_must_be_unique);
  Py_CLEAR(clear_module_state->__pyx_kp_u_record_fields_must_be_unique_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_record_values_type_hint);
  Py_CLEAR(clear_module_state->__pyx_kp_u_record_values_type_hint_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_register);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove);
  Py_CLEAR(clear_module_state->__pyx_n_s_replace);
  Py_CLEAR(clear_module_state->__pyx_n_u_replace);
  Py_CLEAR(clear_module_state->__pyx_n_s_repr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_required_field);
  Py_CLEAR(clear_module_state->__pyx_n_s_resolved_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_return);
  Py_CLEAR(clear_module_state->__pyx_n_s_return_uuid_object);
  Py_CLEAR(clear_module_state->__pyx_kp_u_return_uuid_object_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_reversed);
  Py_CLEAR(clear_module_state->__pyx_n_s_rjust);
  Py_CLEAR(clear_module_state->__pyx_n_s_rsplit);
  Py_CLEAR(clear_module_state->__pyx_n_s_running);
  Py_CLEAR(clear_module_state->__pyx_kp_u_s_read);
  Py_CLEAR(clear_module_state->__pyx_n_s_scale);
  Py_CLEAR(clear_module_state->__pyx_n_u_scale);
  Py_CLEAR(clear_module_state->__pyx_n_s_scaleb);
  Py_CLEAR(clear_module_state->__pyx_n_s_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_schema_ob);
  Py_CLEAR(clear_module_state->__pyx_n_s_schema_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_second);
  Py_CLEAR(clear_module_state->__pyx_n_s_seconds);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_self_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_serialize);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_from_dataclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_setdefault);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_shape);
  Py_CLEAR(clear_module_state->__pyx_n_s_signature);
  Py_CLEAR(clear_module_state->__pyx_n_s_signature_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_signed);
  Py_CLEAR(clear_module_state->__pyx_n_s_size);
  Py_CLEAR(clear_module_state->__pyx_n_u_size);
  Py_CLEAR(clear_module_state->__pyx_n_u_slots);
  Py_CLEAR(clear_module_state->__pyx_n_s_slots_2);
  Py_CLEAR(clear_module_state->__pyx_n_b_snappy);
  Py_CLEAR(clear_module_state->__pyx_n_s_snappy);
  Py_CLEAR(clear_module_state->__pyx_n_u_snappy);
  Py_CLEAR(clear_module_state->__pyx_n_s_source);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_s_src);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_array_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_container_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_enum_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_error_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_logical_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_map_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_null_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_numeric_types_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_option_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_rabin_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_record_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_schema_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_string_types_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_tests_test_buffer_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_tests_test_perf_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_tests_test_zigzag_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_type_pxi);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_union_pxi);
  Py_CLEAR(clear_module_state->__pyx_n_s_start);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_step);
  Py_CLEAR(clear_module_state->__pyx_n_s_stop);
  Py_CLEAR(clear_module_state->__pyx_n_s_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_str_2);
  Py_CLEAR(clear_module_state->__pyx_n_u_strict);
  Py_CLEAR(clear_module_state->__pyx_kp_s_strided_and_direct);
  Py_CLEAR(clear_module_state->__pyx_kp_s_strided_and_direct_or_indirect);
  Py_CLEAR(clear_module_state->__pyx_kp_s_strided_and_indirect);
  Py_CLEAR(clear_module_state->__pyx_n_u_string);
  Py_CLEAR(clear_module_state->__pyx_kp_u_string_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_string_types_default_unchanged);
  Py_CLEAR(clear_module_state->__pyx_kp_u_string_types_default_unchanged_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_string_uuid);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_strptime);
  Py_CLEAR(clear_module_state->__pyx_n_s_struct);
  Py_CLEAR(clear_module_state->__pyx_n_s_sum);
  Py_CLEAR(clear_module_state->__pyx_n_s_super);
  Py_CLEAR(clear_module_state->__pyx_n_s_symbols);
  Py_CLEAR(clear_module_state->__pyx_n_u_symbols);
  Py_CLEAR(clear_module_state->__pyx_n_s_sys);
  Py_CLEAR(clear_module_state->__pyx_n_s_t);
  Py_CLEAR(clear_module_state->__pyx_n_s_t_a);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_reader_read_n);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_reader_read_u8);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_readvarint);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_readvarlong);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_zigzag_long);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_zigzag_perf);
  Py_CLEAR(clear_module_state->__pyx_n_s_tests);
  Py_CLEAR(clear_module_state->__pyx_n_s_tests_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_tests_locals__test_reader_read);
  Py_CLEAR(clear_module_state->__pyx_n_s_tests_locals__test_reader_read_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_tests_locals_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_tests_locals_add_locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_tests_locals_test_readvarint);
  Py_CLEAR(clear_module_state->__pyx_n_s_tests_locals_test_readvarlong);
  Py_CLEAR(clear_module_state->__pyx_n_s_tests_locals_test_zigzag_long);
  Py_CLEAR(clear_module_state->__pyx_n_s_threading);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_time);
  Py_CLEAR(clear_module_state->__pyx_kp_u_time_micros);
  Py_CLEAR(clear_module_state->__pyx_kp_u_time_millis);
  Py_CLEAR(clear_module_state->__pyx_n_s_time_to_micros);
  Py_CLEAR(clear_module_state->__pyx_n_s_timedelta);
  Py_CLEAR(clear_module_state->__pyx_n_s_timestamp);
  Py_CLEAR(clear_module_state->__pyx_kp_u_timestamp_micros);
  Py_CLEAR(clear_module_state->__pyx_kp_u_timestamp_millis);
  Py_CLEAR(clear_module_state->__pyx_n_s_timezone);
  Py_CLEAR(clear_module_state->__pyx_kp_u_to);
  Py_CLEAR(clear_module_state->__pyx_n_s_to_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_total_nums);
  Py_CLEAR(clear_module_state->__pyx_n_s_truncate_fixed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_truncate_fixed_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_tuple_LogicalType);
  Py_CLEAR(clear_module_state->__pyx_n_s_type);
  Py_CLEAR(clear_module_state->__pyx_n_u_type);
  Py_CLEAR(clear_module_state->__pyx_kp_u_type_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_type_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_type_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_type_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_types_str_to_schema);
  Py_CLEAR(clear_module_state->__pyx_kp_u_types_str_to_schema_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_typing);
  Py_CLEAR(clear_module_state->__pyx_n_s_tzinfo);
  Py_CLEAR(clear_module_state->__pyx_n_s_ubin);
  Py_CLEAR(clear_module_state->__pyx_kp_s_unable_to_allocate_array_data);
  Py_CLEAR(clear_module_state->__pyx_kp_s_unable_to_allocate_shape_and_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_underlying);
  Py_CLEAR(clear_module_state->__pyx_n_s_underlying_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_underlying_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_underlying_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_unicode_errors);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unicode_errors_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_unicodedata);
  Py_CLEAR(clear_module_state->__pyx_n_u_union);
  Py_CLEAR(clear_module_state->__pyx_n_s_union_decodes_to);
  Py_CLEAR(clear_module_state->__pyx_kp_u_union_decodes_to_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_union_json_encodes_type_name);
  Py_CLEAR(clear_module_state->__pyx_kp_u_union_json_encodes_type_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_unpack);
  Py_CLEAR(clear_module_state->__pyx_n_s_unsafe_hash);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_utc);
  Py_CLEAR(clear_module_state->__pyx_n_u_utf8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_utf_8);
  Py_CLEAR(clear_module_state->__pyx_n_s_uuid);
  Py_CLEAR(clear_module_state->__pyx_n_u_uuid);
  Py_CLEAR(clear_module_state->__pyx_n_s_uuid4);
  Py_CLEAR(clear_module_state->__pyx_n_s_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_kp_u_value_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_value_adapters);
  Py_CLEAR(clear_module_state->__pyx_n_s_value_type);
  Py_CLEAR(clear_module_state->__pyx_n_u_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_version);
  Py_CLEAR(clear_module_state->__pyx_n_s_version_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_visited);
  Py_CLEAR(clear_module_state->__pyx_kp_u_w);
  Py_CLEAR(clear_module_state->__pyx_n_s_walk_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_warn);
  Py_CLEAR(clear_module_state->__pyx_n_s_warnings);
  Py_CLEAR(clear_module_state->__pyx_n_u_wb);
  Py_CLEAR(clear_module_state->__pyx_n_s_wbits);
  Py_CLEAR(clear_module_state->__pyx_n_s_weakref_slot);
  Py_CLEAR(clear_module_state->__pyx_kp_u_which_is_greater_than_the_scale);
  Py_CLEAR(clear_module_state->__pyx_n_s_with_external_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_with_logical_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_wrap_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_wrapped);
  Py_CLEAR(clear_module_state->__pyx_n_s_write);
  Py_CLEAR(clear_module_state->__pyx_n_u_write);
  Py_CLEAR(clear_module_state->__pyx_kp_u_write_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_write_header);
  Py_CLEAR(clear_module_state->__pyx_n_s_write_many);
  Py_CLEAR(clear_module_state->__pyx_n_s_write_one);
  Py_CLEAR(clear_module_state->__pyx_n_s_write_perf);
  Py_CLEAR(clear_module_state->__pyx_n_s_write_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_writer);
  Py_CLEAR(clear_module_state->__pyx_n_s_writer_schema);
  Py_CLEAR(clear_module_state->__pyx_n_s_writer_type);
  Py_CLEAR(clear_module_state->__pyx_n_b_xz);
  Py_CLEAR(clear_module_state->__pyx_n_u_xz);
  Py_CLEAR(clear_module_state->__pyx_n_s_zip);
  Py_CLEAR(clear_module_state->__pyx_n_s_zlib);
  Py_CLEAR(clear_module_state->__pyx_n_b_zstandard);
  Py_CLEAR(clear_module_state->__pyx_n_s_zstandard);
  Py_CLEAR(clear_module_state->__pyx_n_u_zstandard);
  Py_CLEAR(clear_module_state->__pyx_float_0_0);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_2);
  Py_CLEAR(clear_module_state->__pyx_int_3);
  Py_CLEAR(clear_module_state->__pyx_int_4);
  Py_CLEAR(clear_module_state->__pyx_int_8);
  Py_CLEAR(clear_module_state->__pyx_int_10);
  Py_CLEAR(clear_module_state->__pyx_int_16);
  Py_CLEAR(clear_module_state->__pyx_int_24);
  Py_CLEAR(clear_module_state->__pyx_int_60);
  Py_CLEAR(clear_module_state->__pyx_int_70);
  Py_CLEAR(clear_module_state->__pyx_int_127);
  Py_CLEAR(clear_module_state->__pyx_int_255);
  Py_CLEAR(clear_module_state->__pyx_int_1000);
  Py_CLEAR(clear_module_state->__pyx_int_1970);
  Py_CLEAR(clear_module_state->__pyx_int_3600);
  Py_CLEAR(clear_module_state->__pyx_int_4096);
  Py_CLEAR(clear_module_state->__pyx_int_32767);
  Py_CLEAR(clear_module_state->__pyx_int_65536);
  Py_CLEAR(clear_module_state->__pyx_int_1000000);
  Py_CLEAR(clear_module_state->__pyx_int_8388607);
  Py_CLEAR(clear_module_state->__pyx_int_18332779);
  Py_CLEAR(clear_module_state->__pyx_int_24827247);
  Py_CLEAR(clear_module_state->__pyx_int_25341170);
  Py_CLEAR(clear_module_state->__pyx_int_29099189);
  Py_CLEAR(clear_module_state->__pyx_int_31039736);
  Py_CLEAR(clear_module_state->__pyx_int_33963030);
  Py_CLEAR(clear_module_state->__pyx_int_38410757);
  Py_CLEAR(clear_module_state->__pyx_int_38632982);
  Py_CLEAR(clear_module_state->__pyx_int_39872412);
  Py_CLEAR(clear_module_state->__pyx_int_42565757);
  Py_CLEAR(clear_module_state->__pyx_int_47772561);
  Py_CLEAR(clear_module_state->__pyx_int_60000000);
  Py_CLEAR(clear_module_state->__pyx_int_61270791);
  Py_CLEAR(clear_module_state->__pyx_int_61440409);
  Py_CLEAR(clear_module_state->__pyx_int_61765448);
  Py_CLEAR(clear_module_state->__pyx_int_67877419);
  Py_CLEAR(clear_module_state->__pyx_int_75337510);
  Py_CLEAR(clear_module_state->__pyx_int_79389776);
  Py_CLEAR(clear_module_state->__pyx_int_79941127);
  Py_CLEAR(clear_module_state->__pyx_int_81233721);
  Py_CLEAR(clear_module_state->__pyx_int_83351333);
  Py_CLEAR(clear_module_state->__pyx_int_95761985);
  Py_CLEAR(clear_module_state->__pyx_int_96605483);
  Py_CLEAR(clear_module_state->__pyx_int_101034655);
  Py_CLEAR(clear_module_state->__pyx_int_101670942);
  Py_CLEAR(clear_module_state->__pyx_int_111776469);
  Py_CLEAR(clear_module_state->__pyx_int_111957047);
  Py_CLEAR(clear_module_state->__pyx_int_112105877);
  Py_CLEAR(clear_module_state->__pyx_int_115671040);
  Py_CLEAR(clear_module_state->__pyx_int_119832798);
  Py_CLEAR(clear_module_state->__pyx_int_122446337);
  Py_CLEAR(clear_module_state->__pyx_int_122978751);
  Py_CLEAR(clear_module_state->__pyx_int_136983863);
  Py_CLEAR(clear_module_state->__pyx_int_139219010);
  Py_CLEAR(clear_module_state->__pyx_int_145104888);
  Py_CLEAR(clear_module_state->__pyx_int_148208818);
  Py_CLEAR(clear_module_state->__pyx_int_148222551);
  Py_CLEAR(clear_module_state->__pyx_int_148338675);
  Py_CLEAR(clear_module_state->__pyx_int_148737976);
  Py_CLEAR(clear_module_state->__pyx_int_150354737);
  Py_CLEAR(clear_module_state->__pyx_int_159023012);
  Py_CLEAR(clear_module_state->__pyx_int_159276020);
  Py_CLEAR(clear_module_state->__pyx_int_159891379);
  Py_CLEAR(clear_module_state->__pyx_int_160503772);
  Py_CLEAR(clear_module_state->__pyx_int_160934521);
  Py_CLEAR(clear_module_state->__pyx_int_164094308);
  Py_CLEAR(clear_module_state->__pyx_int_169331114);
  Py_CLEAR(clear_module_state->__pyx_int_175892883);
  Py_CLEAR(clear_module_state->__pyx_int_176081420);
  Py_CLEAR(clear_module_state->__pyx_int_176267327);
  Py_CLEAR(clear_module_state->__pyx_int_177382991);
  Py_CLEAR(clear_module_state->__pyx_int_178791180);
  Py_CLEAR(clear_module_state->__pyx_int_180170947);
  Py_CLEAR(clear_module_state->__pyx_int_180285838);
  Py_CLEAR(clear_module_state->__pyx_int_181080083);
  Py_CLEAR(clear_module_state->__pyx_int_184956420);
  Py_CLEAR(clear_module_state->__pyx_int_184977713);
  Py_CLEAR(clear_module_state->__pyx_int_193777858);
  Py_CLEAR(clear_module_state->__pyx_int_195980960);
  Py_CLEAR(clear_module_state->__pyx_int_198151601);
  Py_CLEAR(clear_module_state->__pyx_int_202158861);
  Py_CLEAR(clear_module_state->__pyx_int_202608488);
  Py_CLEAR(clear_module_state->__pyx_int_203789025);
  Py_CLEAR(clear_module_state->__pyx_int_205941106);
  Py_CLEAR(clear_module_state->__pyx_int_206778879);
  Py_CLEAR(clear_module_state->__pyx_int_212073902);
  Py_CLEAR(clear_module_state->__pyx_int_213886906);
  Py_CLEAR(clear_module_state->__pyx_int_215229444);
  Py_CLEAR(clear_module_state->__pyx_int_215954136);
  Py_CLEAR(clear_module_state->__pyx_int_222419149);
  Py_CLEAR(clear_module_state->__pyx_int_228539134);
  Py_CLEAR(clear_module_state->__pyx_int_228825662);
  Py_CLEAR(clear_module_state->__pyx_int_229636597);
  Py_CLEAR(clear_module_state->__pyx_int_233298401);
  Py_CLEAR(clear_module_state->__pyx_int_236180632);
  Py_CLEAR(clear_module_state->__pyx_int_236349525);
  Py_CLEAR(clear_module_state->__pyx_int_238750788);
  Py_CLEAR(clear_module_state->__pyx_int_240389341);
  Py_CLEAR(clear_module_state->__pyx_int_245215779);
  Py_CLEAR(clear_module_state->__pyx_int_247527084);
  Py_CLEAR(clear_module_state->__pyx_int_254981756);
  Py_CLEAR(clear_module_state->__pyx_int_256569670);
  Py_CLEAR(clear_module_state->__pyx_int_259444540);
  Py_CLEAR(clear_module_state->__pyx_int_261059191);
  Py_CLEAR(clear_module_state->__pyx_int_263819094);
  Py_CLEAR(clear_module_state->__pyx_int_2147483647);
  Py_CLEAR(clear_module_state->__pyx_int_3442987655);
  Py_CLEAR(clear_module_state->__pyx_int_3600000000);
  Py_CLEAR(clear_module_state->__pyx_int_neg_1);
  Py_CLEAR(clear_module_state->__pyx_int_neg_15);
  Py_CLEAR(clear_module_state->__pyx_int_549755813887L);
  Py_CLEAR(clear_module_state->__pyx_int_0x7fffffffffffL);
  Py_CLEAR(clear_module_state->__pyx_int_0x7fffffffffffffL);
  Py_CLEAR(clear_module_state->__pyx_int_0x7fffffffffffffffL);
  Py_CLEAR(clear_module_state->__pyx_k__15);
  Py_CLEAR(clear_module_state->__pyx_k__16);
  Py_CLEAR(clear_module_state->__pyx_k__17);
  Py_CLEAR(clear_module_state->__pyx_k__21);
  Py_CLEAR(clear_module_state->__pyx_k__81);
  Py_CLEAR(clear_module_state->__pyx_k__83);
  Py_CLEAR(clear_module_state->__pyx_k__84);
  Py_CLEAR(clear_module_state->__pyx_k__87);
  Py_CLEAR(clear_module_state->__pyx_k__89);
  Py_CLEAR(clear_module_state->__pyx_slice__5);
  Py_CLEAR(clear_module_state->__pyx_tuple__4);
  Py_CLEAR(clear_module_state->__pyx_tuple__8);
  Py_CLEAR(clear_module_state->__pyx_slice__23);
  Py_CLEAR(clear_module_state->__pyx_slice__86);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__19);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__28);
  Py_CLEAR(clear_module_state->__pyx_tuple__30);
  Py_CLEAR(clear_module_state->__pyx_tuple__32);
  Py_CLEAR(clear_module_state->__pyx_tuple__35);
  Py_CLEAR(clear_module_state->__pyx_tuple__36);
  Py_CLEAR(clear_module_state->__pyx_tuple__37);
  Py_CLEAR(clear_module_state->__pyx_tuple__38);
  Py_CLEAR(clear_module_state->__pyx_tuple__40);
  Py_CLEAR(clear_module_state->__pyx_tuple__42);
  Py_CLEAR(clear_module_state->__pyx_tuple__43);
  Py_CLEAR(clear_module_state->__pyx_tuple__46);
  Py_CLEAR(clear_module_state->__pyx_tuple__47);
  Py_CLEAR(clear_module_state->__pyx_tuple__50);
  Py_CLEAR(clear_module_state->__pyx_tuple__51);
  Py_CLEAR(clear_module_state->__pyx_tuple__52);
  Py_CLEAR(clear_module_state->__pyx_tuple__53);
  Py_CLEAR(clear_module_state->__pyx_tuple__54);
  Py_CLEAR(clear_module_state->__pyx_tuple__55);
  Py_CLEAR(clear_module_state->__pyx_tuple__56);
  Py_CLEAR(clear_module_state->__pyx_tuple__57);
  Py_CLEAR(clear_module_state->__pyx_tuple__58);
  Py_CLEAR(clear_module_state->__pyx_tuple__59);
  Py_CLEAR(clear_module_state->__pyx_tuple__65);
  Py_CLEAR(clear_module_state->__pyx_tuple__67);
  Py_CLEAR(clear_module_state->__pyx_tuple__68);
  Py_CLEAR(clear_module_state->__pyx_tuple__69);
  Py_CLEAR(clear_module_state->__pyx_tuple__71);
  Py_CLEAR(clear_module_state->__pyx_tuple__73);
  Py_CLEAR(clear_module_state->__pyx_tuple__74);
  Py_CLEAR(clear_module_state->__pyx_tuple__85);
  Py_CLEAR(clear_module_state->__pyx_tuple__88);
  Py_CLEAR(clear_module_state->__pyx_tuple__90);
  Py_CLEAR(clear_module_state->__pyx_tuple__91);
  Py_CLEAR(clear_module_state->__pyx_tuple__92);
  Py_CLEAR(clear_module_state->__pyx_tuple__93);
  Py_CLEAR(clear_module_state->__pyx_tuple__94);
  Py_CLEAR(clear_module_state->__pyx_tuple__95);
  Py_CLEAR(clear_module_state->__pyx_tuple__96);
  Py_CLEAR(clear_module_state->__pyx_tuple__97);
  Py_CLEAR(clear_module_state->__pyx_tuple__98);
  Py_CLEAR(clear_module_state->__pyx_tuple__99);
  Py_CLEAR(clear_module_state->__pyx_tuple__100);
  Py_CLEAR(clear_module_state->__pyx_tuple__101);
  Py_CLEAR(clear_module_state->__pyx_tuple__102);
  Py_CLEAR(clear_module_state->__pyx_tuple__103);
  Py_CLEAR(clear_module_state->__pyx_tuple__104);
  Py_CLEAR(clear_module_state->__pyx_tuple__105);
  Py_CLEAR(clear_module_state->__pyx_tuple__106);
  Py_CLEAR(clear_module_state->__pyx_tuple__107);
  Py_CLEAR(clear_module_state->__pyx_tuple__108);
  Py_CLEAR(clear_module_state->__pyx_tuple__109);
  Py_CLEAR(clear_module_state->__pyx_tuple__110);
  Py_CLEAR(clear_module_state->__pyx_tuple__111);
  Py_CLEAR(clear_module_state->__pyx_tuple__112);
  Py_CLEAR(clear_module_state->__pyx_tuple__113);
  Py_CLEAR(clear_module_state->__pyx_tuple__114);
  Py_CLEAR(clear_module_state->__pyx_tuple__115);
  Py_CLEAR(clear_module_state->__pyx_tuple__116);
  Py_CLEAR(clear_module_state->__pyx_tuple__117);
  Py_CLEAR(clear_module_state->__pyx_tuple__118);
  Py_CLEAR(clear_module_state->__pyx_tuple__119);
  Py_CLEAR(clear_module_state->__pyx_tuple__120);
  Py_CLEAR(clear_module_state->__pyx_tuple__121);
  Py_CLEAR(clear_module_state->__pyx_tuple__123);
  Py_CLEAR(clear_module_state->__pyx_tuple__124);
  Py_CLEAR(clear_module_state->__pyx_tuple__125);
  Py_CLEAR(clear_module_state->__pyx_tuple__126);
  Py_CLEAR(clear_module_state->__pyx_tuple__127);
  Py_CLEAR(clear_module_state->__pyx_tuple__128);
  Py_CLEAR(clear_module_state->__pyx_tuple__129);
  Py_CLEAR(clear_module_state->__pyx_tuple__130);
  Py_CLEAR(clear_module_state->__pyx_tuple__131);
  Py_CLEAR(clear_module_state->__pyx_tuple__132);
  Py_CLEAR(clear_module_state->__pyx_tuple__134);
  Py_CLEAR(clear_module_state->__pyx_tuple__137);
  Py_CLEAR(clear_module_state->__pyx_tuple__139);
  Py_CLEAR(clear_module_state->__pyx_tuple__141);
  Py_CLEAR(clear_module_state->__pyx_tuple__142);
  Py_CLEAR(clear_module_state->__pyx_tuple__144);
  Py_CLEAR(clear_module_state->__pyx_tuple__145);
  Py_CLEAR(clear_module_state->__pyx_tuple__147);
  Py_CLEAR(clear_module_state->__pyx_tuple__148);
  Py_CLEAR(clear_module_state->__pyx_tuple__150);
  Py_CLEAR(clear_module_state->__pyx_tuple__153);
  Py_CLEAR(clear_module_state->__pyx_tuple__155);
  Py_CLEAR(clear_module_state->__pyx_tuple__157);
  Py_CLEAR(clear_module_state->__pyx_tuple__159);
  Py_CLEAR(clear_module_state->__pyx_tuple__161);
  Py_CLEAR(clear_module_state->__pyx_tuple__163);
  Py_CLEAR(clear_module_state->__pyx_tuple__164);
  Py_CLEAR(clear_module_state->__pyx_tuple__166);
  Py_CLEAR(clear_module_state->__pyx_tuple__167);
  Py_CLEAR(clear_module_state->__pyx_tuple__169);
  Py_CLEAR(clear_module_state->__pyx_tuple__183);
  Py_CLEAR(clear_module_state->__pyx_tuple__185);
  Py_CLEAR(clear_module_state->__pyx_tuple__188);
  Py_CLEAR(clear_module_state->__pyx_tuple__192);
  Py_CLEAR(clear_module_state->__pyx_tuple__194);
  Py_CLEAR(clear_module_state->__pyx_tuple__196);
  Py_CLEAR(clear_module_state->__pyx_tuple__198);
  Py_CLEAR(clear_module_state->__pyx_tuple__201);
  Py_CLEAR(clear_module_state->__pyx_tuple__202);
  Py_CLEAR(clear_module_state->__pyx_tuple__204);
  Py_CLEAR(clear_module_state->__pyx_tuple__206);
  Py_CLEAR(clear_module_state->__pyx_tuple__218);
  Py_CLEAR(clear_module_state->__pyx_tuple__220);
  Py_CLEAR(clear_module_state->__pyx_tuple__253);
  Py_CLEAR(clear_module_state->__pyx_tuple__258);
  Py_CLEAR(clear_module_state->__pyx_tuple__260);
  Py_CLEAR(clear_module_state->__pyx_tuple__264);
  Py_CLEAR(clear_module_state->__pyx_tuple__271);
  Py_CLEAR(clear_module_state->__pyx_tuple__294);
  Py_CLEAR(clear_module_state->__pyx_tuple__297);
  Py_CLEAR(clear_module_state->__pyx_tuple__300);
  Py_CLEAR(clear_module_state->__pyx_tuple__303);
  Py_CLEAR(clear_module_state->__pyx_tuple__306);
  Py_CLEAR(clear_module_state->__pyx_tuple__308);
  Py_CLEAR(clear_module_state->__pyx_tuple__313);
  Py_CLEAR(clear_module_state->__pyx_tuple__321);
  Py_CLEAR(clear_module_state->__pyx_tuple__323);
  Py_CLEAR(clear_module_state->__pyx_tuple__325);
  Py_CLEAR(clear_module_state->__pyx_tuple__327);
  Py_CLEAR(clear_module_state->__pyx_tuple__329);
  Py_CLEAR(clear_module_state->__pyx_tuple__331);
  Py_CLEAR(clear_module_state->__pyx_tuple__332);
  Py_CLEAR(clear_module_state->__pyx_tuple__334);
  Py_CLEAR(clear_module_state->__pyx_tuple__335);
  Py_CLEAR(clear_module_state->__pyx_tuple__337);
  Py_CLEAR(clear_module_state->__pyx_tuple__340);
  Py_CLEAR(clear_module_state->__pyx_tuple__342);
  Py_CLEAR(clear_module_state->__pyx_tuple__344);
  Py_CLEAR(clear_module_state->__pyx_tuple__346);
  Py_CLEAR(clear_module_state->__pyx_tuple__347);
  Py_CLEAR(clear_module_state->__pyx_tuple__349);
  Py_CLEAR(clear_module_state->__pyx_tuple__357);
  Py_CLEAR(clear_module_state->__pyx_tuple__363);
  Py_CLEAR(clear_module_state->__pyx_tuple__367);
  Py_CLEAR(clear_module_state->__pyx_tuple__377);
  Py_CLEAR(clear_module_state->__pyx_tuple__381);
  Py_CLEAR(clear_module_state->__pyx_tuple__396);
  Py_CLEAR(clear_module_state->__pyx_tuple__418);
  Py_CLEAR(clear_module_state->__pyx_tuple__420);
  Py_CLEAR(clear_module_state->__pyx_tuple__422);
  Py_CLEAR(clear_module_state->__pyx_tuple__424);
  Py_CLEAR(clear_module_state->__pyx_tuple__425);
  Py_CLEAR(clear_module_state->__pyx_tuple__427);
  Py_CLEAR(clear_module_state->__pyx_codeobj__11);
  Py_CLEAR(clear_module_state->__pyx_codeobj__12);
  Py_CLEAR(clear_module_state->__pyx_codeobj__29);
  Py_CLEAR(clear_module_state->__pyx_codeobj__31);
  Py_CLEAR(clear_module_state->__pyx_codeobj__33);
  Py_CLEAR(clear_module_state->__pyx_codeobj__39);
  Py_CLEAR(clear_module_state->__pyx_codeobj__41);
  Py_CLEAR(clear_module_state->__pyx_codeobj__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__45);
  Py_CLEAR(clear_module_state->__pyx_codeobj__49);
  Py_CLEAR(clear_module_state->__pyx_codeobj__66);
  Py_CLEAR(clear_module_state->__pyx_codeobj__75);
  Py_CLEAR(clear_module_state->__pyx_codeobj__82);
  Py_CLEAR(clear_module_state->__pyx_codeobj__133);
  Py_CLEAR(clear_module_state->__pyx_codeobj__135);
  Py_CLEAR(clear_module_state->__pyx_codeobj__136);
  Py_CLEAR(clear_module_state->__pyx_codeobj__138);
  Py_CLEAR(clear_module_state->__pyx_codeobj__140);
  Py_CLEAR(clear_module_state->__pyx_codeobj__143);
  Py_CLEAR(clear_module_state->__pyx_codeobj__146);
  Py_CLEAR(clear_module_state->__pyx_codeobj__149);
  Py_CLEAR(clear_module_state->__pyx_codeobj__151);
  Py_CLEAR(clear_module_state->__pyx_codeobj__154);
  Py_CLEAR(clear_module_state->__pyx_codeobj__156);
  Py_CLEAR(clear_module_state->__pyx_codeobj__158);
  Py_CLEAR(clear_module_state->__pyx_codeobj__160);
  Py_CLEAR(clear_module_state->__pyx_codeobj__162);
  Py_CLEAR(clear_module_state->__pyx_codeobj__165);
  Py_CLEAR(clear_module_state->__pyx_codeobj__168);
  Py_CLEAR(clear_module_state->__pyx_codeobj__170);
  Py_CLEAR(clear_module_state->__pyx_codeobj__171);
  Py_CLEAR(clear_module_state->__pyx_codeobj__172);
  Py_CLEAR(clear_module_state->__pyx_codeobj__173);
  Py_CLEAR(clear_module_state->__pyx_codeobj__174);
  Py_CLEAR(clear_module_state->__pyx_codeobj__175);
  Py_CLEAR(clear_module_state->__pyx_codeobj__176);
  Py_CLEAR(clear_module_state->__pyx_codeobj__177);
  Py_CLEAR(clear_module_state->__pyx_codeobj__178);
  Py_CLEAR(clear_module_state->__pyx_codeobj__179);
  Py_CLEAR(clear_module_state->__pyx_codeobj__180);
  Py_CLEAR(clear_module_state->__pyx_codeobj__181);
  Py_CLEAR(clear_module_state->__pyx_codeobj__182);
  Py_CLEAR(clear_module_state->__pyx_codeobj__184);
  Py_CLEAR(clear_module_state->__pyx_codeobj__186);
  Py_CLEAR(clear_module_state->__pyx_codeobj__187);
  Py_CLEAR(clear_module_state->__pyx_codeobj__189);
  Py_CLEAR(clear_module_state->__pyx_codeobj__190);
  Py_CLEAR(clear_module_state->__pyx_codeobj__191);
  Py_CLEAR(clear_module_state->__pyx_codeobj__193);
  Py_CLEAR(clear_module_state->__pyx_codeobj__195);
  Py_CLEAR(clear_module_state->__pyx_codeobj__197);
  Py_CLEAR(clear_module_state->__pyx_codeobj__199);
  Py_CLEAR(clear_module_state->__pyx_codeobj__200);
  Py_CLEAR(clear_module_state->__pyx_codeobj__203);
  Py_CLEAR(clear_module_state->__pyx_codeobj__205);
  Py_CLEAR(clear_module_state->__pyx_codeobj__207);
  Py_CLEAR(clear_module_state->__pyx_codeobj__208);
  Py_CLEAR(clear_module_state->__pyx_codeobj__209);
  Py_CLEAR(clear_module_state->__pyx_codeobj__210);
  Py_CLEAR(clear_module_state->__pyx_codeobj__211);
  Py_CLEAR(clear_module_state->__pyx_codeobj__212);
  Py_CLEAR(clear_module_state->__pyx_codeobj__213);
  Py_CLEAR(clear_module_state->__pyx_codeobj__214);
  Py_CLEAR(clear_module_state->__pyx_codeobj__215);
  Py_CLEAR(clear_module_state->__pyx_codeobj__216);
  Py_CLEAR(clear_module_state->__pyx_codeobj__217);
  Py_CLEAR(clear_module_state->__pyx_codeobj__219);
  Py_CLEAR(clear_module_state->__pyx_codeobj__221);
  Py_CLEAR(clear_module_state->__pyx_codeobj__222);
  Py_CLEAR(clear_module_state->__pyx_codeobj__223);
  Py_CLEAR(clear_module_state->__pyx_codeobj__224);
  Py_CLEAR(clear_module_state->__pyx_codeobj__225);
  Py_CLEAR(clear_module_state->__pyx_codeobj__226);
  Py_CLEAR(clear_module_state->__pyx_codeobj__227);
  Py_CLEAR(clear_module_state->__pyx_codeobj__228);
  Py_CLEAR(clear_module_state->__pyx_codeobj__229);
  Py_CLEAR(clear_module_state->__pyx_codeobj__230);
  Py_CLEAR(clear_module_state->__pyx_codeobj__231);
  Py_CLEAR(clear_module_state->__pyx_codeobj__232);
  Py_CLEAR(clear_module_state->__pyx_codeobj__233);
  Py_CLEAR(clear_module_state->__pyx_codeobj__234);
  Py_CLEAR(clear_module_state->__pyx_codeobj__235);
  Py_CLEAR(clear_module_state->__pyx_codeobj__236);
  Py_CLEAR(clear_module_state->__pyx_codeobj__237);
  Py_CLEAR(clear_module_state->__pyx_codeobj__238);
  Py_CLEAR(clear_module_state->__pyx_codeobj__239);
  Py_CLEAR(clear_module_state->__pyx_codeobj__240);
  Py_CLEAR(clear_module_state->__pyx_codeobj__241);
  Py_CLEAR(clear_module_state->__pyx_codeobj__242);
  Py_CLEAR(clear_module_state->__pyx_codeobj__243);
  Py_CLEAR(clear_module_state->__pyx_codeobj__244);
  Py_CLEAR(clear_module_state->__pyx_codeobj__245);
  Py_CLEAR(clear_module_state->__pyx_codeobj__246);
  Py_CLEAR(clear_module_state->__pyx_codeobj__247);
  Py_CLEAR(clear_module_state->__pyx_codeobj__248);
  Py_CLEAR(clear_module_state->__pyx_codeobj__249);
  Py_CLEAR(clear_module_state->__pyx_codeobj__250);
  Py_CLEAR(clear_module_state->__pyx_codeobj__251);
  Py_CLEAR(clear_module_state->__pyx_codeobj__252);
  Py_CLEAR(clear_module_state->__pyx_codeobj__254);
  Py_CLEAR(clear_module_state->__pyx_codeobj__255);
  Py_CLEAR(clear_module_state->__pyx_codeobj__256);
  Py_CLEAR(clear_module_state->__pyx_codeobj__257);
  Py_CLEAR(clear_module_state->__pyx_codeobj__259);
  Py_CLEAR(clear_module_state->__pyx_codeobj__261);
  Py_CLEAR(clear_module_state->__pyx_codeobj__262);
  Py_CLEAR(clear_module_state->__pyx_codeobj__263);
  Py_CLEAR(clear_module_state->__pyx_codeobj__265);
  Py_CLEAR(clear_module_state->__pyx_codeobj__266);
  Py_CLEAR(clear_module_state->__pyx_codeobj__267);
  Py_CLEAR(clear_module_state->__pyx_codeobj__268);
  Py_CLEAR(clear_module_state->__pyx_codeobj__269);
  Py_CLEAR(clear_module_state->__pyx_codeobj__270);
  Py_CLEAR(clear_module_state->__pyx_codeobj__272);
  Py_CLEAR(clear_module_state->__pyx_codeobj__273);
  Py_CLEAR(clear_module_state->__pyx_codeobj__274);
  Py_CLEAR(clear_module_state->__pyx_codeobj__275);
  Py_CLEAR(clear_module_state->__pyx_codeobj__276);
  Py_CLEAR(clear_module_state->__pyx_codeobj__277);
  Py_CLEAR(clear_module_state->__pyx_codeobj__278);
  Py_CLEAR(clear_module_state->__pyx_codeobj__279);
  Py_CLEAR(clear_module_state->__pyx_codeobj__280);
  Py_CLEAR(clear_module_state->__pyx_codeobj__281);
  Py_CLEAR(clear_module_state->__pyx_codeobj__282);
  Py_CLEAR(clear_module_state->__pyx_codeobj__283);
  Py_CLEAR(clear_module_state->__pyx_codeobj__284);
  Py_CLEAR(clear_module_state->__pyx_codeobj__285);
  Py_CLEAR(clear_module_state->__pyx_codeobj__286);
  Py_CLEAR(clear_module_state->__pyx_codeobj__287);
  Py_CLEAR(clear_module_state->__pyx_codeobj__288);
  Py_CLEAR(clear_module_state->__pyx_codeobj__289);
  Py_CLEAR(clear_module_state->__pyx_codeobj__290);
  Py_CLEAR(clear_module_state->__pyx_codeobj__291);
  Py_CLEAR(clear_module_state->__pyx_codeobj__292);
  Py_CLEAR(clear_module_state->__pyx_codeobj__293);
  Py_CLEAR(clear_module_state->__pyx_codeobj__295);
  Py_CLEAR(clear_module_state->__pyx_codeobj__296);
  Py_CLEAR(clear_module_state->__pyx_codeobj__298);
  Py_CLEAR(clear_module_state->__pyx_codeobj__299);
  Py_CLEAR(clear_module_state->__pyx_codeobj__301);
  Py_CLEAR(clear_module_state->__pyx_codeobj__302);
  Py_CLEAR(clear_module_state->__pyx_codeobj__304);
  Py_CLEAR(clear_module_state->__pyx_codeobj__305);
  Py_CLEAR(clear_module_state->__pyx_codeobj__307);
  Py_CLEAR(clear_module_state->__pyx_codeobj__309);
  Py_CLEAR(clear_module_state->__pyx_codeobj__310);
  Py_CLEAR(clear_module_state->__pyx_codeobj__311);
  Py_CLEAR(clear_module_state->__pyx_codeobj__312);
  Py_CLEAR(clear_module_state->__pyx_codeobj__314);
  Py_CLEAR(clear_module_state->__pyx_codeobj__315);
  Py_CLEAR(clear_module_state->__pyx_codeobj__316);
  Py_CLEAR(clear_module_state->__pyx_codeobj__317);
  Py_CLEAR(clear_module_state->__pyx_codeobj__318);
  Py_CLEAR(clear_module_state->__pyx_codeobj__319);
  Py_CLEAR(clear_module_state->__pyx_codeobj__320);
  Py_CLEAR(clear_module_state->__pyx_codeobj__322);
  Py_CLEAR(clear_module_state->__pyx_codeobj__324);
  Py_CLEAR(clear_module_state->__pyx_codeobj__326);
  Py_CLEAR(clear_module_state->__pyx_codeobj__328);
  Py_CLEAR(clear_module_state->__pyx_codeobj__330);
  Py_CLEAR(clear_module_state->__pyx_codeobj__333);
  Py_CLEAR(clear_module_state->__pyx_codeobj__336);
  Py_CLEAR(clear_module_state->__pyx_codeobj__338);
  Py_CLEAR(clear_module_state->__pyx_codeobj__339);
  Py_CLEAR(clear_module_state->__pyx_codeobj__341);
  Py_CLEAR(clear_module_state->__pyx_codeobj__343);
  Py_CLEAR(clear_module_state->__pyx_codeobj__345);
  Py_CLEAR(clear_module_state->__pyx_codeobj__348);
  Py_CLEAR(clear_module_state->__pyx_codeobj__350);
  Py_CLEAR(clear_module_state->__pyx_codeobj__351);
  Py_CLEAR(clear_module_state->__pyx_codeobj__352);
  Py_CLEAR(clear_module_state->__pyx_codeobj__353);
  Py_CLEAR(clear_module_state->__pyx_codeobj__354);
  Py_CLEAR(clear_module_state->__pyx_codeobj__355);
  Py_CLEAR(clear_module_state->__pyx_codeobj__356);
  Py_CLEAR(clear_module_state->__pyx_codeobj__358);
  Py_CLEAR(clear_module_state->__pyx_codeobj__359);
  Py_CLEAR(clear_module_state->__pyx_codeobj__360);
  Py_CLEAR(clear_module_state->__pyx_codeobj__361);
  Py_CLEAR(clear_module_state->__pyx_codeobj__362);
  Py_CLEAR(clear_module_state->__pyx_codeobj__364);
  Py_CLEAR(clear_module_state->__pyx_codeobj__365);
  Py_CLEAR(clear_module_state->__pyx_codeobj__366);
  Py_CLEAR(clear_module_state->__pyx_codeobj__368);
  Py_CLEAR(clear_module_state->__pyx_codeobj__369);
  Py_CLEAR(clear_module_state->__pyx_codeobj__370);
  Py_CLEAR(clear_module_state->__pyx_codeobj__371);
  Py_CLEAR(clear_module_state->__pyx_codeobj__372);
  Py_CLEAR(clear_module_state->__pyx_codeobj__373);
  Py_CLEAR(clear_module_state->__pyx_codeobj__374);
  Py_CLEAR(clear_module_state->__pyx_codeobj__375);
  Py_CLEAR(clear_module_state->__pyx_codeobj__376);
  Py_CLEAR(clear_module_state->__pyx_codeobj__378);
  Py_CLEAR(clear_module_state->__pyx_codeobj__379);
  Py_CLEAR(clear_module_state->__pyx_codeobj__380);
  Py_CLEAR(clear_module_state->__pyx_codeobj__382);
  Py_CLEAR(clear_module_state->__pyx_codeobj__383);
  Py_CLEAR(clear_module_state->__pyx_codeobj__384);
  Py_CLEAR(clear_module_state->__pyx_codeobj__385);
  Py_CLEAR(clear_module_state->__pyx_codeobj__386);
  Py_CLEAR(clear_module_state->__pyx_codeobj__387);
  Py_CLEAR(clear_module_state->__pyx_codeobj__388);
  Py_CLEAR(clear_module_state->__pyx_codeobj__389);
  Py_CLEAR(clear_module_state->__pyx_codeobj__390);
  Py_CLEAR(clear_module_state->__pyx_codeobj__391);
  Py_CLEAR(clear_module_state->__pyx_codeobj__392);
  Py_CLEAR(clear_module_state->__pyx_codeobj__393);
  Py_CLEAR(clear_module_state->__pyx_codeobj__394);
  Py_CLEAR(clear_module_state->__pyx_codeobj__395);
  Py_CLEAR(clear_module_state->__pyx_codeobj__397);
  Py_CLEAR(clear_module_state->__pyx_codeobj__398);
  Py_CLEAR(clear_module_state->__pyx_codeobj__399);
  Py_CLEAR(clear_module_state->__pyx_codeobj__400);
  Py_CLEAR(clear_module_state->__pyx_codeobj__401);
  Py_CLEAR(clear_module_state->__pyx_codeobj__402);
  Py_CLEAR(clear_module_state->__pyx_codeobj__403);
  Py_CLEAR(clear_module_state->__pyx_codeobj__404);
  Py_CLEAR(clear_module_state->__pyx_codeobj__405);
  Py_CLEAR(clear_module_state->__pyx_codeobj__406);
  Py_CLEAR(clear_module_state->__pyx_codeobj__407);
  Py_CLEAR(clear_module_state->__pyx_codeobj__408);
  Py_CLEAR(clear_module_state->__pyx_codeobj__409);
  Py_CLEAR(clear_module_state->__pyx_codeobj__410);
  Py_CLEAR(clear_module_state->__pyx_codeobj__411);
  Py_CLEAR(clear_module_state->__pyx_codeobj__412);
  Py_CLEAR(clear_module_state->__pyx_codeobj__413);
  Py_CLEAR(clear_module_state->__pyx_codeobj__414);
  Py_CLEAR(clear_module_state->__pyx_codeobj__415);
  Py_CLEAR(clear_module_state->__pyx_codeobj__416);
  Py_CLEAR(clear_module_state->__pyx_codeobj__417);
  Py_CLEAR(clear_module_state->__pyx_codeobj__419);
  Py_CLEAR(clear_module_state->__pyx_codeobj__421);
  Py_CLEAR(clear_module_state->__pyx_codeobj__423);
  Py_CLEAR(clear_module_state->__pyx_codeobj__426);
  Py_CLEAR(clear_module_state->__pyx_codeobj__428);
  Py_CLEAR(clear_module_state->__pyx_codeobj__429);
  Py_CLEAR(clear_module_state->__pyx_codeobj__430);
  Py_CLEAR(clear_module_state->__pyx_codeobj__431);
  Py_CLEAR(clear_module_state->__pyx_codeobj__432);
  Py_CLEAR(clear_module_state->__pyx_codeobj__433);
  Py_CLEAR(clear_module_state->__pyx_codeobj__434);
  Py_CLEAR(clear_module_state->__pyx_codeobj__435);
  Py_CLEAR(clear_module_state->__pyx_codeobj__436);
  Py_CLEAR(clear_module_state->__pyx_codeobj__437);
  Py_CLEAR(clear_module_state->__pyx_codeobj__438);
  Py_CLEAR(clear_module_state->__pyx_codeobj__439);
  Py_CLEAR(clear_module_state->__pyx_codeobj__440);
  Py_CLEAR(clear_module_state->__pyx_codeobj__441);
  Py_CLEAR(clear_module_state->__pyx_codeobj__442);
  Py_CLEAR(clear_module_state->__pyx_codeobj__443);
  Py_CLEAR(clear_module_state->__pyx_codeobj__444);
  Py_CLEAR(clear_module_state->__pyx_codeobj__445);
  Py_CLEAR(clear_module_state->__pyx_codeobj__446);
  Py_CLEAR(clear_module_state->__pyx_codeobj__447);
  Py_CLEAR(clear_module_state->__pyx_codeobj__448);
  Py_CLEAR(clear_module_state->__pyx_codeobj__449);
  Py_CLEAR(clear_module_state->__pyx_codeobj__450);
  Py_CLEAR(clear_module_state->__pyx_codeobj__451);
  Py_CLEAR(clear_module_state->__pyx_codeobj__452);
  Py_CLEAR(clear_module_state->__pyx_codeobj__453);
  Py_CLEAR(clear_module_state->__pyx_codeobj__454);
  Py_CLEAR(clear_module_state->__pyx_codeobj__455);
  Py_CLEAR(clear_module_state->__pyx_codeobj__456);
  Py_CLEAR(clear_module_state->__pyx_codeobj__457);
  Py_CLEAR(clear_module_state->__pyx_codeobj__458);
  Py_CLEAR(clear_module_state->__pyx_codeobj__459);
  Py_CLEAR(clear_module_state->__pyx_codeobj__460);
  Py_CLEAR(clear_module_state->__pyx_codeobj__461);
  Py_CLEAR(clear_module_state->__pyx_codeobj__462);
  Py_CLEAR(clear_module_state->__pyx_codeobj__463);
  Py_CLEAR(clear_module_state->__pyx_codeobj__464);
  Py_CLEAR(clear_module_state->__pyx_codeobj__465);
  Py_CLEAR(clear_module_state->__pyx_codeobj__466);
  Py_CLEAR(clear_module_state->__pyx_codeobj__467);
  Py_CLEAR(clear_module_state->__pyx_codeobj__468);
  Py_CLEAR(clear_module_state->__pyx_codeobj__469);
  Py_CLEAR(clear_module_state->__pyx_codeobj__470);
  Py_CLEAR(clear_module_state->__pyx_codeobj__471);
  Py_CLEAR(clear_module_state->__pyx_codeobj__472);
  Py_CLEAR(clear_module_state->__pyx_codeobj__473);
  Py_CLEAR(clear_module_state->__pyx_codeobj__474);
  Py_CLEAR(clear_module_state->__pyx_codeobj__475);
  Py_CLEAR(clear_module_state->__pyx_codeobj__476);
  Py_CLEAR(clear_module_state->__pyx_codeobj__477);
  Py_CLEAR(clear_module_state->__pyx_codeobj__478);
  Py_CLEAR(clear_module_state->__pyx_codeobj__479);
  Py_CLEAR(clear_module_state->__pyx_codeobj__480);
  Py_CLEAR(clear_module_state->__pyx_codeobj__481);
  Py_CLEAR(clear_module_state->__pyx_codeobj__482);
  Py_CLEAR(clear_module_state->__pyx_codeobj__483);
  Py_CLEAR(clear_module_state->__pyx_codeobj__484);
  Py_CLEAR(clear_module_state->__pyx_codeobj__485);
  Py_CLEAR(clear_module_state->__pyx_codeobj__486);
  Py_CLEAR(clear_module_state->__pyx_codeobj__487);
  Py_CLEAR(clear_module_state->__pyx_codeobj__488);
  Py_CLEAR(clear_module_state->__pyx_codeobj__489);
  Py_CLEAR(clear_module_state->__pyx_codeobj__490);
  Py_CLEAR(clear_module_state->__pyx_codeobj__491);
  Py_CLEAR(clear_module_state->__pyx_codeobj__492);
  Py_CLEAR(clear_module_state->__pyx_codeobj__493);
  Py_CLEAR(clear_module_state->__pyx_codeobj__494);
  Py_CLEAR(clear_module_state->__pyx_codeobj__495);
  Py_CLEAR(clear_module_state->__pyx_codeobj__496);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_5array_array);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_Options);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_Options);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__Writer);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__Writer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__Reader);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__Reader);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_FileReader);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_FileReader);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_FileWriter);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_FileWriter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_MemoryWriter);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_MemoryWriter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_MemoryReader);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_MemoryReader);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_Rabin);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_Rabin);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_Record);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_Record);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_RecordField);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_RecordField);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__FieldAccessor);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__FieldAccessor);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__Sentinel);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__Sentinel);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__CanonicalForm);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__CanonicalForm);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_AvroType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_AvroType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_UnionType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_UnionType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_StringType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_StringType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_BytesType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_BytesType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__PlaceholderType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__PlaceholderType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_DoubleType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_DoubleType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_FloatType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_FloatType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_LongType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_LongType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_IntType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_IntType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_BoolType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_BoolType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_NullType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_NullType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_MapType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_MapType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_ArrayType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_ArrayType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__NamedType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__NamedType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_FixedType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_FixedType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_RecordType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_RecordType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_PromotingRecordType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_PromotingRecordType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_EnumType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_EnumType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_PromotingEnumType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_PromotingEnumType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_Schema);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_Schema);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_ResolvedSchema);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_ResolvedSchema);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_ValueAdapter);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_ValueAdapter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_LogicalType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_LogicalType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_CustomLogicalType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_CustomLogicalType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_DecimalType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_DecimalType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_UUIDBase);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_UUIDBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_UUIDStringType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_UUIDStringType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_UUIDFixedType);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_UUIDFixedType);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_Date);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_Date);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_TimeMillis);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_TimeMillis);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_TimeMicros);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_TimeMicros);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_TimestampMillis);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_TimestampMillis);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_TimestampMicros);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_TimestampMicros);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_Codec);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_Codec);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__SnappyCodec);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__SnappyCodec);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__NullCodec);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__NullCodec);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__DeflateCodec);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__DeflateCodec);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___Bzip2Codec);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___Bzip2Codec);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__LzmaCodec);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__LzmaCodec);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__ZStandardCodec);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__ZStandardCodec);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro__Lz4Codec);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro__Lz4Codec);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_ContainerReader);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_ContainerReader);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_ContainerWriter);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_ContainerWriter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_PromoteToFloat);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_PromoteToFloat);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_PromoteBytesToString);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_PromoteBytesToString);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_PromoteStringToBytes);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_PromoteStringToBytes);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro_CannotPromote);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro_CannotPromote);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct__genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct__genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_1_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_1_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_2_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_2_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_3_walk_types);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_3_walk_types);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_5_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_5_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_6_walk_types);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_6_walk_types);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_7_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_7_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_8___init__);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_8___init__);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_9_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_9_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_10_walk_types);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_10_walk_types);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_11___init__);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_11___init__);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_12_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_12_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_13_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_13_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_14_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_14_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_15_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_15_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_16_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_16_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_17_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_17_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_18_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_18_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_19_walk_types);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_19_walk_types);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_20_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_20_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_21_walk_types);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_21_walk_types);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_22_get_namespaced_aliases);
  Py_VISIT(traverse_module_state->__pyx_ptype_5cavro___pyx_scope_struct_23_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_5cavro___pyx_scope_struct_23_genexpr);
  Py_VISIT(traverse_module_state->__pyx_array_type);
  Py_VISIT(traverse_module_state->__pyx_type___pyx_array);
  Py_VISIT(traverse_module_state->__pyx_MemviewEnum_type);
  Py_VISIT(traverse_module_state->__pyx_type___pyx_MemviewEnum);
  Py_VISIT(traverse_module_state->__pyx_memoryview_type);
  Py_VISIT(traverse_module_state->__pyx_type___pyx_memoryview);
  Py_VISIT(traverse_module_state->__pyx_memoryviewslice_type);
  Py_VISIT(traverse_module_state->__pyx_type___pyx_memoryviewslice);
  Py_VISIT(traverse_module_state->__pyx_kp_u_);
  Py_VISIT(traverse_module_state->__pyx_kp_u_0);
  Py_VISIT(traverse_module_state->__pyx_kp_u_1);
  Py_VISIT(traverse_module_state->__pyx_kp_u_1_0_0);
  Py_VISIT(traverse_module_state->__pyx_n_s_AMBIGUOUS_TYPES);
  Py_VISIT(traverse_module_state->__pyx_n_s_ASC);
  Py_VISIT(traverse_module_state->__pyx_n_s_ASCII);
  Py_VISIT(traverse_module_state->__pyx_kp_u_A_Za_z__A_Za_z0_9);
  Py_VISIT(traverse_module_state->__pyx_kp_s_A_record_contains_multiple_fiel);
  Py_VISIT(traverse_module_state->__pyx_kp_s_A_requested_codec_or_codec_in_a);
  Py_VISIT(traverse_module_state->__pyx_kp_s_A_schema_cannot_be_promoted_to);
  Py_VISIT(traverse_module_state->__pyx_kp_s_A_value_is_invalid_for_a_given);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Aliases_must_be_a_list_tuple_set);
  Py_VISIT(traverse_module_state->__pyx_kp_s_All_dimensions_preceding_dimensi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_An_unknown_hash_method_was_requ);
  Py_VISIT(traverse_module_state->__pyx_n_s_ArrayType);
  Py_VISIT(traverse_module_state->__pyx_n_s_ArrayType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ArrayType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ArrayType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_ArrayType_convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_ArrayType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_ArrayType_walk_types);
  Py_VISIT(traverse_module_state->__pyx_kp_s_As_of_cython_3_0_2_dataclasses);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_AttributeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvroType);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvroType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvroType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvroType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvroType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvroType_convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvroType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvroType_for_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvroType_for_source);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvroType_get_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvroType_walk_types);
  Py_VISIT(traverse_module_state->__pyx_n_u_B);
  Py_VISIT(traverse_module_state->__pyx_n_s_BATCH_SIZE);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Base_class_for_exceptions_raise);
  Py_VISIT(traverse_module_state->__pyx_n_s_BoolType);
  Py_VISIT(traverse_module_state->__pyx_n_s_BoolType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BoolType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BoolType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_BoolType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_BoolType_copy);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Buffer_view_does_not_expose_stri);
  Py_VISIT(traverse_module_state->__pyx_n_s_BytesType);
  Py_VISIT(traverse_module_state->__pyx_n_s_BytesType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BytesType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BytesType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_BytesType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_BytesType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_Bzip2Codec);
  Py_VISIT(traverse_module_state->__pyx_n_s_Bzip2Codec___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Bzip2Codec___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_CANONICAL_FORM_KEYS);
  Py_VISIT(traverse_module_state->__pyx_n_s_CODECS);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Can_only_create_a_buffer_that_is);
  Py_VISIT(traverse_module_state->__pyx_n_s_CannotPromote);
  Py_VISIT(traverse_module_state->__pyx_n_s_CannotPromoteError);
  Py_VISIT(traverse_module_state->__pyx_n_s_CannotPromoteError___init);
  Py_VISIT(traverse_module_state->__pyx_n_s_CannotPromote___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_CannotPromote___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_assign_to_read_only_memor);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_create_writable_memory_vi);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_index_with_type);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_promote);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_read_from);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_transpose_memoryview_with);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_write_to);
  Py_VISIT(traverse_module_state->__pyx_n_s_CanonicalForm);
  Py_VISIT(traverse_module_state->__pyx_n_s_CanonicalForm___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_CanonicalForm___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_CavroException);
  Py_VISIT(traverse_module_state->__pyx_n_s_Codec);
  Py_VISIT(traverse_module_state->__pyx_n_s_CodecUnavailable);
  Py_VISIT(traverse_module_state->__pyx_n_s_Codec___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Codec___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerReader);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerReader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerReader___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerReader__read_marker);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerReader_next_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerWriter);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerWriter___enter);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerWriter___exit);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerWriter___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerWriter___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerWriter_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerWriter_flush);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerWriter_write_many);
  Py_VISIT(traverse_module_state->__pyx_n_s_ContainerWriter_write_one);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Container_is_closed);
  Py_VISIT(traverse_module_state->__pyx_n_s_Context);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Controls_how_union_values_are_d);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Could_not_find_key_type_in_schem);
  Py_VISIT(traverse_module_state->__pyx_n_s_CustomLogicalType);
  Py_VISIT(traverse_module_state->__pyx_n_s_CustomLogicalType___reduce_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_CustomLogicalType___setstate_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_DEFAULT_OPTIONS);
  Py_VISIT(traverse_module_state->__pyx_n_s_DESC);
  Py_VISIT(traverse_module_state->__pyx_n_s_DataclassParams);
  Py_VISIT(traverse_module_state->__pyx_n_s_Date);
  Py_VISIT(traverse_module_state->__pyx_n_s_Date___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Date___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Date__for_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_Decimal);
  Py_VISIT(traverse_module_state->__pyx_n_s_DecimalType);
  Py_VISIT(traverse_module_state->__pyx_n_s_DecimalType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_DecimalType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_DecimalType__for_type);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Decimal_value);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Default_value);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Default_value_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_DeflateCodec);
  Py_VISIT(traverse_module_state->__pyx_n_s_DeflateCodec___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_DeflateCodec___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Dimension_d_is_not_direct);
  Py_VISIT(traverse_module_state->__pyx_n_s_DoubleType);
  Py_VISIT(traverse_module_state->__pyx_n_s_DoubleType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_DoubleType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_DoubleType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_DoubleType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_DoubleType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_DuplicateName);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Duplicate_field_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_EMPTY_ARRAY);
  Py_VISIT(traverse_module_state->__pyx_n_s_EMPTY_MP);
  Py_VISIT(traverse_module_state->__pyx_n_s_EOFError);
  Py_VISIT(traverse_module_state->__pyx_n_s_EPOCH_DATE);
  Py_VISIT(traverse_module_state->__pyx_n_s_EPOCH_DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_EPOCH_DT_NO_TZ);
  Py_VISIT(traverse_module_state->__pyx_n_s_Ellipsis);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Empty_shape_tuple_for_cython_arr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_End_of_file_found_trying_to_read);
  Py_VISIT(traverse_module_state->__pyx_kp_u_End_of_file_found_trying_to_read_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Enum);
  Py_VISIT(traverse_module_state->__pyx_n_s_EnumType);
  Py_VISIT(traverse_module_state->__pyx_n_s_EnumType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_EnumType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_EnumType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_EnumType__for_writer_locals_gene);
  Py_VISIT(traverse_module_state->__pyx_n_s_EnumType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_EnumType_copy);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Enum_symbol);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Enum_symbols_must_be_a_list_of_s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Enum_symbols_must_be_unique);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Error_closing_container_file_dur);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Expected_datetime_date_got);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Expected_datetime_datetime_got);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Expected_null_got);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExponentTooLarge);
  Py_VISIT(traverse_module_state->__pyx_n_s_FIELD);
  Py_VISIT(traverse_module_state->__pyx_n_u_False);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Field);
  Py_VISIT(traverse_module_state->__pyx_n_s_FieldAccessor);
  Py_VISIT(traverse_module_state->__pyx_n_s_FieldAccessor___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FieldAccessor___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FileReader);
  Py_VISIT(traverse_module_state->__pyx_n_s_FileReader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FileReader___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FileWriter);
  Py_VISIT(traverse_module_state->__pyx_n_s_FileWriter___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FileWriter___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FixedType);
  Py_VISIT(traverse_module_state->__pyx_n_s_FixedType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FixedType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FixedType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_FixedType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_FixedType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloatType);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloatType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloatType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloatType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloatType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_FloatType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_HAVE_BZIP2);
  Py_VISIT(traverse_module_state->__pyx_n_s_HAVE_LZ4);
  Py_VISIT(traverse_module_state->__pyx_n_s_HAVE_SNAPPY);
  Py_VISIT(traverse_module_state->__pyx_n_s_HAVE_XZ);
  Py_VISIT(traverse_module_state->__pyx_n_s_HAVE_ZLIB);
  Py_VISIT(traverse_module_state->__pyx_n_s_HAVE_ZSTD);
  Py_VISIT(traverse_module_state->__pyx_n_s_IGNORE);
  Py_VISIT(traverse_module_state->__pyx_n_s_ImportError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_25);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_26);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_27);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_28);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Index_out_of_bounds_axis_d);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Indirect_dimensions_not_supporte);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntType);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_InvalidHasher);
  Py_VISIT(traverse_module_state->__pyx_n_s_InvalidName);
  Py_VISIT(traverse_module_state->__pyx_n_s_InvalidValue);
  Py_VISIT(traverse_module_state->__pyx_n_s_InvalidValue___init);
  Py_VISIT(traverse_module_state->__pyx_n_s_InvalidValue___str);
  Py_VISIT(traverse_module_state->__pyx_n_s_InvalidValue___str___locals_gene);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_block_sync_marker_expect);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_file_header_expected);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_length_for_fixed_field);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_mode_expected_c_or_fortr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_negative_size_for_fixed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_non_bytes_value_for_fixe);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_non_str_value_for_fixed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_shape_in_axis);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_union_decodes_to_option);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_Iterable);
  Py_VISIT(traverse_module_state->__pyx_n_s_KeyError);
  Py_VISIT(traverse_module_state->__pyx_n_s_LOGICAL_TYPES);
  Py_VISIT(traverse_module_state->__pyx_n_u_Ll);
  Py_VISIT(traverse_module_state->__pyx_n_s_LogicalType);
  Py_VISIT(traverse_module_state->__pyx_n_s_LogicalType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_LogicalType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_LogicalType_for_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_LongType);
  Py_VISIT(traverse_module_state->__pyx_n_s_LongType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_LongType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_LongType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_LongType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_LongType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_u_Lu);
  Py_VISIT(traverse_module_state->__pyx_n_s_Lz4Codec);
  Py_VISIT(traverse_module_state->__pyx_n_s_Lz4Codec___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Lz4Codec___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_LzmaCodec);
  Py_VISIT(traverse_module_state->__pyx_n_s_LzmaCodec___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_LzmaCodec___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_MAX_PREC);
  Py_VISIT(traverse_module_state->__pyx_n_s_MAX_TIME);
  Py_VISIT(traverse_module_state->__pyx_n_s_MISSING_VALUE);
  Py_VISIT(traverse_module_state->__pyx_n_u_MISSING_VALUE);
  Py_VISIT(traverse_module_state->__pyx_n_s_MapType);
  Py_VISIT(traverse_module_state->__pyx_n_s_MapType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_MapType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_MapType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_MapType_convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_MapType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_MapType_walk_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_MappingProxyType);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Marker_must_be_bytes_got);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Marker_must_be_exactly_16_bytes);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Max_precision_is);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryError);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryReader);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryReader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryReader___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_s_MemoryView_of_r_at_0x_x);
  Py_VISIT(traverse_module_state->__pyx_kp_s_MemoryView_of_r_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryWriter);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryWriter___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryWriter___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_NAME_RE_STRICT);
  Py_VISIT(traverse_module_state->__pyx_n_s_NAME_RE_UNICODE);
  Py_VISIT(traverse_module_state->__pyx_n_s_NO_DEFAULT);
  Py_VISIT(traverse_module_state->__pyx_n_u_NO_DEFAULT);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Name);
  Py_VISIT(traverse_module_state->__pyx_n_s_NamedType);
  Py_VISIT(traverse_module_state->__pyx_n_s_NamedType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_NamedType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_NamedType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_NamedType_get_namespaced_aliase);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Namespace);
  Py_VISIT(traverse_module_state->__pyx_kp_u_None);
  Py_VISIT(traverse_module_state->__pyx_n_s_NotImplemented);
  Py_VISIT(traverse_module_state->__pyx_n_s_NotImplementedError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Not_enough_input_data_to_read_va);
  Py_VISIT(traverse_module_state->__pyx_n_s_NullCodec);
  Py_VISIT(traverse_module_state->__pyx_n_s_NullCodec___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_NullCodec___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_NullType);
  Py_VISIT(traverse_module_state->__pyx_n_s_NullType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_NullType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_NullType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_NullType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_NullType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_b_O);
  Py_VISIT(traverse_module_state->__pyx_n_s_OBJ_FILE_METADATA);
  Py_VISIT(traverse_module_state->__pyx_n_s_OBJ_MAGIC_BYTES);
  Py_VISIT(traverse_module_state->__pyx_kp_b_Obj);
  Py_VISIT(traverse_module_state->__pyx_n_s_Options);
  Py_VISIT(traverse_module_state->__pyx_n_s_OptionsWrapper);
  Py_VISIT(traverse_module_state->__pyx_n_s_Options___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Options___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Options_diff);
  Py_VISIT(traverse_module_state->__pyx_n_s_Options_equals);
  Py_VISIT(traverse_module_state->__pyx_n_s_Options_replace);
  Py_VISIT(traverse_module_state->__pyx_n_s_Options_with_external_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_Options_with_logical_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_Order);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
  Py_VISIT(traverse_module_state->__pyx_n_s_OverflowError);
  Py_VISIT(traverse_module_state->__pyx_n_s_PERMISSIVE_OPTIONS);
  Py_VISIT(traverse_module_state->__pyx_n_s_PRIMITIVE_TYPES);
  Py_VISIT(traverse_module_state->__pyx_n_s_Path);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_PlaceholderType);
  Py_VISIT(traverse_module_state->__pyx_n_s_PlaceholderType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PlaceholderType___setstate_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_PlaceholderType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_PlaceholderType__get_schema_ext);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Placeholder_types_cannot_be_enco);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Placeholder_types_have_no_canoni);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Precision);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Precision_is_too_large_for_fixed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Precision_must_be_an_integer_bet);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Precision_must_be_greater_than_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromoteBytesToString);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromoteBytesToString___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromoteBytesToString___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromoteStringToBytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromoteStringToBytes___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromoteStringToBytes___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromoteToFloat);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromoteToFloat___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromoteToFloat___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromotingEnumType);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromotingEnumType___reduce_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromotingEnumType___setstate_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromotingRecordType);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromotingRecordType___reduce_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_PromotingRecordType___setstate_c);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Promoting_schemas_reader_writer);
  Py_VISIT(traverse_module_state->__pyx_n_s_RAW_VALUES);
  Py_VISIT(traverse_module_state->__pyx_n_s_Rabin);
  Py_VISIT(traverse_module_state->__pyx_n_s_Rabin___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Rabin___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Rabin_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_Rabin_digest);
  Py_VISIT(traverse_module_state->__pyx_n_s_Rabin_hexdigest);
  Py_VISIT(traverse_module_state->__pyx_n_s_Rabin_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Reader___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Reader___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Record);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecordField);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecordField___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecordField___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecordField_get_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecordType);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecordType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecordType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecordType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecordType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecordType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_RecordType_walk_types);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Record_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Record_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_Record___dir);
  Py_VISIT(traverse_module_state->__pyx_n_s_Record___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Record___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Record__asdict);
  Py_VISIT(traverse_module_state->__pyx_n_s_Record__repr_children_locals_gen);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Records_may_either_be_instantiat);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Records_may_either_be_instantiat_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_ResolvedSchema);
  Py_VISIT(traverse_module_state->__pyx_n_s_ResolvedSchema___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ResolvedSchema___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ResourceWarning);
  Py_VISIT(traverse_module_state->__pyx_n_s_SLICE_TYPE);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSIZE_TYPECODE);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Scale_must_be_a_positive_integer);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema__make_logical_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema__wrap_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema_binary_decode);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema_binary_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema_binary_read);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema_binary_write);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema_can_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema_find_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema_fingerprint);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Schema_is_required);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema_json_decode);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema_json_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_Schema_reader_for_writer);
  Py_VISIT(traverse_module_state->__pyx_n_s_Sentinel);
  Py_VISIT(traverse_module_state->__pyx_n_s_Sentinel___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Sentinel___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Sequence);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Sequence_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_SignatureWrapper);
  Py_VISIT(traverse_module_state->__pyx_n_s_SignatureWrapper___init);
  Py_VISIT(traverse_module_state->__pyx_n_s_SignatureWrapper__set_from_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_SnappyCodec);
  Py_VISIT(traverse_module_state->__pyx_n_s_SnappyCodec___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SnappyCodec___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Step_may_not_be_zero_axis_d);
  Py_VISIT(traverse_module_state->__pyx_n_s_StopIteration);
  Py_VISIT(traverse_module_state->__pyx_n_s_StringType);
  Py_VISIT(traverse_module_state->__pyx_n_s_StringType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_StringType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_StringType__convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_StringType__get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_StringType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_TYPES_BY_NAME);
  Py_VISIT(traverse_module_state->__pyx_n_s_TYPE_TUPLE_ALWAYS);
  Py_VISIT(traverse_module_state->__pyx_n_s_TYPE_TUPLE_IF_AMBIGUOUS);
  Py_VISIT(traverse_module_state->__pyx_n_s_TYPE_TUPLE_IF_RECORD);
  Py_VISIT(traverse_module_state->__pyx_n_s_TYPE_TUPLE_IF_RECORD_AMBIGUOUS);
  Py_VISIT(traverse_module_state->__pyx_kp_s_The_exponent_of_a_decimal_value);
  Py_VISIT(traverse_module_state->__pyx_kp_u_The_null_namespace_cannot_be_spe);
  Py_VISIT(traverse_module_state->__pyx_kp_s_The_order_of_a_field_in_a_recor);
  Py_VISIT(traverse_module_state->__pyx_kp_s_The_schema_contains_a_type_or_e);
  Py_VISIT(traverse_module_state->__pyx_kp_s_The_schema_contains_an_unexptec);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimeMicros);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimeMicros___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimeMicros___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimeMicros__for_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimeMillis);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimeMillis___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimeMillis___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimeMillis__for_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimestampMicros);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimestampMicros___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimestampMicros___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimestampMicros__for_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimestampMillis);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimestampMillis___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimestampMillis___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimestampMillis__for_type);
  Py_VISIT(traverse_module_state->__pyx_n_u_True);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Trying_to_close_a_closed_Contain);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Trying_to_write_to_closed_Contai);
  Py_VISIT(traverse_module_state->__pyx_n_s_Type);
  Py_VISIT(traverse_module_state->__pyx_n_u_Type);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Type_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_UNKNOWN);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUID);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUIDBase);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUIDBase___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUIDBase___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUIDBase__for_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUIDFixedType);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUIDFixedType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUIDFixedType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUIDFixedType__for_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUIDStringType);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUIDStringType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUIDStringType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UUIDStringType__for_type);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Unable_to_convert_item_to_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnicodeEncodeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Union);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnionDecodeOption);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnionType);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnionType___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnionType___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnionType_convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnionType_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnionType_get_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnionType_walk_types);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Union_bytes_hashlib__hashlib_HAS);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Union_str_object);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unions_may_not_have_more_than_on);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unions_may_not_immediately_conta);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unions_must_contain_at_least_one);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnknownType);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnknownType___init);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unknown_hash_method);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unknown_type);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unknown_value_for_enum);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unsupported_codec);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Value);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueAdapter);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueAdapter___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueAdapter___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_n_s_View_MemoryView);
  Py_VISIT(traverse_module_state->__pyx_n_s_Writer);
  Py_VISIT(traverse_module_state->__pyx_n_s_Writer___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Writer___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Y_m_d);
  Py_VISIT(traverse_module_state->__pyx_n_s_ZStandardCodec);
  Py_VISIT(traverse_module_state->__pyx_n_s_ZStandardCodec___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ZStandardCodec___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_ZstdCompressor);
  Py_VISIT(traverse_module_state->__pyx_n_s_ZstdDecompressor);
  Py_VISIT(traverse_module_state->__pyx_n_s__122);
  Py_VISIT(traverse_module_state->__pyx_kp_b__13);
  Py_VISIT(traverse_module_state->__pyx_kp_u__13);
  Py_VISIT(traverse_module_state->__pyx_kp_u__14);
  Py_VISIT(traverse_module_state->__pyx_n_s__152);
  Py_VISIT(traverse_module_state->__pyx_kp_s__18);
  Py_VISIT(traverse_module_state->__pyx_kp_u__18);
  Py_VISIT(traverse_module_state->__pyx_kp_u__2);
  Py_VISIT(traverse_module_state->__pyx_kp_u__24);
  Py_VISIT(traverse_module_state->__pyx_kp_u__25);
  Py_VISIT(traverse_module_state->__pyx_kp_u__26);
  Py_VISIT(traverse_module_state->__pyx_kp_u__27);
  Py_VISIT(traverse_module_state->__pyx_n_s__3);
  Py_VISIT(traverse_module_state->__pyx_kp_b__34);
  Py_VISIT(traverse_module_state->__pyx_kp_u__48);
  Py_VISIT(traverse_module_state->__pyx_kp_u__6);
  Py_VISIT(traverse_module_state->__pyx_kp_u__60);
  Py_VISIT(traverse_module_state->__pyx_kp_u__61);
  Py_VISIT(traverse_module_state->__pyx_kp_u__62);
  Py_VISIT(traverse_module_state->__pyx_kp_u__63);
  Py_VISIT(traverse_module_state->__pyx_kp_u__64);
  Py_VISIT(traverse_module_state->__pyx_kp_u__7);
  Py_VISIT(traverse_module_state->__pyx_kp_u__70);
  Py_VISIT(traverse_module_state->__pyx_kp_b__72);
  Py_VISIT(traverse_module_state->__pyx_kp_u__76);
  Py_VISIT(traverse_module_state->__pyx_kp_u__77);
  Py_VISIT(traverse_module_state->__pyx_kp_u__78);
  Py_VISIT(traverse_module_state->__pyx_kp_u__79);
  Py_VISIT(traverse_module_state->__pyx_kp_u__80);
  Py_VISIT(traverse_module_state->__pyx_kp_u__9);
  Py_VISIT(traverse_module_state->__pyx_n_s_a);
  Py_VISIT(traverse_module_state->__pyx_n_b_abacus);
  Py_VISIT(traverse_module_state->__pyx_n_b_abacuscounter1);
  Py_VISIT(traverse_module_state->__pyx_n_s_abc);
  Py_VISIT(traverse_module_state->__pyx_n_s_accepts_string);
  Py_VISIT(traverse_module_state->__pyx_n_s_actual);
  Py_VISIT(traverse_module_state->__pyx_n_s_adapt_record_types);
  Py_VISIT(traverse_module_state->__pyx_kp_u_adapt_record_types_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_add);
  Py_VISIT(traverse_module_state->__pyx_n_u_aliases);
  Py_VISIT(traverse_module_state->__pyx_n_s_allocate_buffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_aliases_to_be_string);
  Py_VISIT(traverse_module_state->__pyx_kp_u_allow_aliases_to_be_string_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_duplicate_union_types);
  Py_VISIT(traverse_module_state->__pyx_kp_u_allow_duplicate_union_types_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_empty_unions);
  Py_VISIT(traverse_module_state->__pyx_kp_u_allow_empty_unions_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_error_type);
  Py_VISIT(traverse_module_state->__pyx_kp_u_allow_error_type_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_false_values_for_null);
  Py_VISIT(traverse_module_state->__pyx_kp_u_allow_false_values_for_null_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_invalid_default_values);
  Py_VISIT(traverse_module_state->__pyx_kp_u_allow_invalid_default_values_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_leading_dot_in_names);
  Py_VISIT(traverse_module_state->__pyx_kp_u_allow_leading_dot_in_names_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_nested_unions);
  Py_VISIT(traverse_module_state->__pyx_kp_u_allow_nested_unions_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_primitive_name_collision);
  Py_VISIT(traverse_module_state->__pyx_kp_u_allow_primitive_name_collision_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_allow_primitive_names_in_namesp);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_primitive_names_in_namespa);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_tuple_notation);
  Py_VISIT(traverse_module_state->__pyx_kp_u_allow_tuple_notation_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_union_default_any_member);
  Py_VISIT(traverse_module_state->__pyx_kp_u_allow_union_default_any_member_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_alternate_timestamp_encoding);
  Py_VISIT(traverse_module_state->__pyx_kp_u_alternate_timestamp_millis_enco);
  Py_VISIT(traverse_module_state->__pyx_n_s_alternate_timestamp_millis_encod);
  Py_VISIT(traverse_module_state->__pyx_kp_u_and);
  Py_VISIT(traverse_module_state->__pyx_n_s_annotations);
  Py_VISIT(traverse_module_state->__pyx_n_s_annotations_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_appears_multiple_times_in_schem);
  Py_VISIT(traverse_module_state->__pyx_n_s_append);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_array);
  Py_VISIT(traverse_module_state->__pyx_n_u_array);
  Py_VISIT(traverse_module_state->__pyx_n_s_as_tuple);
  Py_VISIT(traverse_module_state->__pyx_n_u_ascending);
  Py_VISIT(traverse_module_state->__pyx_n_s_ascii_name_rules);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ascii_name_rules_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_asdict);
  Py_VISIT(traverse_module_state->__pyx_n_u_asdict);
  Py_VISIT(traverse_module_state->__pyx_n_s_astimezone);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_kp_u_at);
  Py_VISIT(traverse_module_state->__pyx_kp_u_avro_codec);
  Py_VISIT(traverse_module_state->__pyx_kp_u_avro_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_avro_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_b);
  Py_VISIT(traverse_module_state->__pyx_n_s_base);
  Py_VISIT(traverse_module_state->__pyx_n_u_big);
  Py_VISIT(traverse_module_state->__pyx_n_s_bin);
  Py_VISIT(traverse_module_state->__pyx_n_s_binary_decode);
  Py_VISIT(traverse_module_state->__pyx_n_s_binary_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_binary_read);
  Py_VISIT(traverse_module_state->__pyx_n_s_binary_write);
  Py_VISIT(traverse_module_state->__pyx_n_s_bint);
  Py_VISIT(traverse_module_state->__pyx_n_u_bint);
  Py_VISIT(traverse_module_state->__pyx_n_s_bit_length);
  Py_VISIT(traverse_module_state->__pyx_n_s_block_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_bool);
  Py_VISIT(traverse_module_state->__pyx_n_u_bool);
  Py_VISIT(traverse_module_state->__pyx_n_s_bool_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_boolean);
  Py_VISIT(traverse_module_state->__pyx_n_u_boolean_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_buffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_byteorder);
  Py_VISIT(traverse_module_state->__pyx_kp_u_bytes);
  Py_VISIT(traverse_module_state->__pyx_kp_u_bytes_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_bytes_3);
  Py_VISIT(traverse_module_state->__pyx_n_u_bytes_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_bytes_codec);
  Py_VISIT(traverse_module_state->__pyx_kp_u_bytes_codec_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_bytes_default_value_utf8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_bytes_default_value_utf8_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_bz2);
  Py_VISIT(traverse_module_state->__pyx_n_b_bzip2);
  Py_VISIT(traverse_module_state->__pyx_n_u_bzip2);
  Py_VISIT(traverse_module_state->__pyx_n_s_c);
  Py_VISIT(traverse_module_state->__pyx_n_u_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_can_encode);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cannot_be_adapted_to);
  Py_VISIT(traverse_module_state->__pyx_n_s_canonical_form);
  Py_VISIT(traverse_module_state->__pyx_n_s_canonical_form_repeat_enum);
  Py_VISIT(traverse_module_state->__pyx_kp_u_canonical_form_repeat_enum_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_canonical_form_repeat_fixed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_canonical_form_repeat_fixed_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_category);
  Py_VISIT(traverse_module_state->__pyx_n_s_cavro);
  Py_VISIT(traverse_module_state->__pyx_kp_s_cavro_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_changes);
  Py_VISIT(traverse_module_state->__pyx_n_s_check_exp_overflow);
  Py_VISIT(traverse_module_state->__pyx_n_s_check_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_chr);
  Py_VISIT(traverse_module_state->__pyx_n_s_clamp);
  Py_VISIT(traverse_module_state->__pyx_n_s_clamp_float_overflow);
  Py_VISIT(traverse_module_state->__pyx_kp_u_clamp_float_overflow_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_clamp_int_overflow);
  Py_VISIT(traverse_module_state->__pyx_kp_u_clamp_int_overflow_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_inst_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_inst_method___get);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_inst_method___init);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_cls);
  Py_VISIT(traverse_module_state->__pyx_n_s_codec);
  Py_VISIT(traverse_module_state->__pyx_n_s_coerce_int_to_float);
  Py_VISIT(traverse_module_state->__pyx_kp_u_coerce_int_to_float_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_coerce_values_to_boolean);
  Py_VISIT(traverse_module_state->__pyx_kp_u_coerce_values_to_boolean_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_coerce_values_to_float);
  Py_VISIT(traverse_module_state->__pyx_kp_u_coerce_values_to_float_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_coerce_values_to_int);
  Py_VISIT(traverse_module_state->__pyx_kp_u_coerce_values_to_int_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_coerce_values_to_str);
  Py_VISIT(traverse_module_state->__pyx_kp_u_coerce_values_to_str_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections);
  Py_VISIT(traverse_module_state->__pyx_kp_s_collections_abc);
  Py_VISIT(traverse_module_state->__pyx_n_s_compare);
  Py_VISIT(traverse_module_state->__pyx_n_s_compile);
  Py_VISIT(traverse_module_state->__pyx_n_s_compress);
  Py_VISIT(traverse_module_state->__pyx_n_s_container_fill_blocks);
  Py_VISIT(traverse_module_state->__pyx_kp_u_container_fill_blocks_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_context);
  Py_VISIT(traverse_module_state->__pyx_kp_s_contiguous_and_direct);
  Py_VISIT(traverse_module_state->__pyx_kp_s_contiguous_and_indirect);
  Py_VISIT(traverse_module_state->__pyx_n_s_convert_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_convert_value_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_count);
  Py_VISIT(traverse_module_state->__pyx_n_b_counter);
  Py_VISIT(traverse_module_state->__pyx_n_s_crc32);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_decimal);
  Py_VISIT(traverse_module_state->__pyx_n_s_created);
  Py_VISIT(traverse_module_state->__pyx_n_s_custom_decode_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_custom_encode_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_dataclass_fields);
  Py_VISIT(traverse_module_state->__pyx_n_s_dataclass_params);
  Py_VISIT(traverse_module_state->__pyx_n_s_dataclasses);
  Py_VISIT(traverse_module_state->__pyx_n_s_date);
  Py_VISIT(traverse_module_state->__pyx_n_u_date);
  Py_VISIT(traverse_module_state->__pyx_n_s_date_type_accepts_string);
  Py_VISIT(traverse_module_state->__pyx_kp_u_date_type_accepts_string_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_datetime);
  Py_VISIT(traverse_module_state->__pyx_n_s_days);
  Py_VISIT(traverse_module_state->__pyx_n_s_decimal);
  Py_VISIT(traverse_module_state->__pyx_n_u_decimal);
  Py_VISIT(traverse_module_state->__pyx_kp_u_decimal_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_decimal_check_exp_overflow);
  Py_VISIT(traverse_module_state->__pyx_kp_u_decimal_check_exp_overflow_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_decode);
  Py_VISIT(traverse_module_state->__pyx_n_s_decompress);
  Py_VISIT(traverse_module_state->__pyx_n_s_decompressobj);
  Py_VISIT(traverse_module_state->__pyx_n_s_defaul_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_default);
  Py_VISIT(traverse_module_state->__pyx_n_u_default);
  Py_VISIT(traverse_module_state->__pyx_n_s_default_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_defer_schema_promotion_errors);
  Py_VISIT(traverse_module_state->__pyx_kp_u_defer_schema_promotion_errors_2);
  Py_VISIT(traverse_module_state->__pyx_n_b_deflate);
  Py_VISIT(traverse_module_state->__pyx_n_u_deflate);
  Py_VISIT(traverse_module_state->__pyx_n_u_descending);
  Py_VISIT(traverse_module_state->__pyx_n_s_deserialize);
  Py_VISIT(traverse_module_state->__pyx_n_s_dest);
  Py_VISIT(traverse_module_state->__pyx_n_s_dest_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_dict_str_AvroType);
  Py_VISIT(traverse_module_state->__pyx_kp_s_dict_str_tuple_object_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_to_canonical_locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_diff);
  Py_VISIT(traverse_module_state->__pyx_n_s_digest);
  Py_VISIT(traverse_module_state->__pyx_n_s_digest_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_dir);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_u_doc);
  Py_VISIT(traverse_module_state->__pyx_n_s_doc_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement__binary_buff);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement__convert_val);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement__extract_met);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement__get_schema);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement__get_value_f);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement__json_decode);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement__json_format);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement_binary_buffe);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement_canonical_fo);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement_clone);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement_decode_value);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement_encode_value);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement_read_block);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement_read_bytes);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement_read_n);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement_read_u8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement_write_block);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement_write_n);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_implement_write_u8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_double);
  Py_VISIT(traverse_module_state->__pyx_n_u_double_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_dtype_is_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_dumps);
  Py_VISIT(traverse_module_state->__pyx_n_s_e);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_enforce_enum_symbol_name_rules);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enforce_enum_symbol_name_rules_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_enforce_namespace_name_rules);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enforce_namespace_name_rules_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_enforce_type_name_rules);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enforce_type_name_rules_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_ensure_ascii);
  Py_VISIT(traverse_module_state->__pyx_n_s_enter);
  Py_VISIT(traverse_module_state->__pyx_n_s_enum);
  Py_VISIT(traverse_module_state->__pyx_n_u_enum);
  Py_VISIT(traverse_module_state->__pyx_n_s_enum_symbols_must_be_unique);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enum_symbols_must_be_unique_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_eq);
  Py_VISIT(traverse_module_state->__pyx_n_s_equals);
  Py_VISIT(traverse_module_state->__pyx_n_s_error);
  Py_VISIT(traverse_module_state->__pyx_n_u_error);
  Py_VISIT(traverse_module_state->__pyx_n_s_errors);
  Py_VISIT(traverse_module_state->__pyx_n_s_escape);
  Py_VISIT(traverse_module_state->__pyx_n_s_eval);
  Py_VISIT(traverse_module_state->__pyx_n_s_eval_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_exc_tb);
  Py_VISIT(traverse_module_state->__pyx_n_s_exc_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_exc_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_exit);
  Py_VISIT(traverse_module_state->__pyx_n_s_expand_types_in_schema);
  Py_VISIT(traverse_module_state->__pyx_kp_u_expand_types_in_schema_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_externally_defined_types);
  Py_VISIT(traverse_module_state->__pyx_kp_u_externally_defined_types_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_f);
  Py_VISIT(traverse_module_state->__pyx_n_s_field);
  Py_VISIT(traverse_module_state->__pyx_n_s_field_to_index);
  Py_VISIT(traverse_module_state->__pyx_n_u_field_to_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_field_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_fields);
  Py_VISIT(traverse_module_state->__pyx_n_u_fields);
  Py_VISIT(traverse_module_state->__pyx_n_s_file_obj);
  Py_VISIT(traverse_module_state->__pyx_n_s_find_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_fingerprint);
  Py_VISIT(traverse_module_state->__pyx_n_s_fingerprint_returns_digest);
  Py_VISIT(traverse_module_state->__pyx_kp_u_fingerprint_returns_digest_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_fitness);
  Py_VISIT(traverse_module_state->__pyx_n_s_fixed);
  Py_VISIT(traverse_module_state->__pyx_n_u_fixed);
  Py_VISIT(traverse_module_state->__pyx_n_s_fixed_codec);
  Py_VISIT(traverse_module_state->__pyx_kp_u_fixed_codec_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_fixed_uuid);
  Py_VISIT(traverse_module_state->__pyx_n_s_flags);
  Py_VISIT(traverse_module_state->__pyx_kp_u_float);
  Py_VISIT(traverse_module_state->__pyx_n_s_float16);
  Py_VISIT(traverse_module_state->__pyx_n_s_float32);
  Py_VISIT(traverse_module_state->__pyx_n_s_float64);
  Py_VISIT(traverse_module_state->__pyx_n_u_float_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_float_out_of_range_inf);
  Py_VISIT(traverse_module_state->__pyx_kp_u_float_out_of_range_inf_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_floor);
  Py_VISIT(traverse_module_state->__pyx_n_s_flush);
  Py_VISIT(traverse_module_state->__pyx_n_s_fn);
  Py_VISIT(traverse_module_state->__pyx_kp_u_for_field);
  Py_VISIT(traverse_module_state->__pyx_n_s_for_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_for_source);
  Py_VISIT(traverse_module_state->__pyx_kp_u_for_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_for_type_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_for_type_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_force);
  Py_VISIT(traverse_module_state->__pyx_n_s_format);
  Py_VISIT(traverse_module_state->__pyx_n_s_fortran);
  Py_VISIT(traverse_module_state->__pyx_n_u_fortran);
  Py_VISIT(traverse_module_state->__pyx_kp_u_found_twice);
  Py_VISIT(traverse_module_state->__pyx_n_s_frame);
  Py_VISIT(traverse_module_state->__pyx_n_s_from_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_frozen);
  Py_VISIT(traverse_module_state->__pyx_n_s_fullmatch);
  Py_VISIT(traverse_module_state->__pyx_n_s_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_functools);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_schema_extra);
  Py_VISIT(traverse_module_state->__pyx_kp_u_getfullargspec);
  Py_VISIT(traverse_module_state->__pyx_n_u_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_given);
  Py_VISIT(traverse_module_state->__pyx_kp_u_got);
  Py_VISIT(traverse_module_state->__pyx_kp_u_got_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_got_3);
  Py_VISIT(traverse_module_state->__pyx_kp_u_got_4);
  Py_VISIT(traverse_module_state->__pyx_kp_u_got_differing_extents_in_dimensi);
  Py_VISIT(traverse_module_state->__pyx_kp_u_has_exponent);
  Py_VISIT(traverse_module_state->__pyx_n_s_hash);
  Py_VISIT(traverse_module_state->__pyx_n_s_hasher);
  Py_VISIT(traverse_module_state->__pyx_n_s_hashlib);
  Py_VISIT(traverse_module_state->__pyx_n_s_hex);
  Py_VISIT(traverse_module_state->__pyx_n_s_hexdigest);
  Py_VISIT(traverse_module_state->__pyx_n_s_hour);
  Py_VISIT(traverse_module_state->__pyx_n_s_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_ignore);
  Py_VISIT(traverse_module_state->__pyx_n_u_ignore);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_indent);
  Py_VISIT(traverse_module_state->__pyx_n_s_index);
  Py_VISIT(traverse_module_state->__pyx_kp_u_inf);
  Py_VISIT(traverse_module_state->__pyx_n_u_inf_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_init);
  Py_VISIT(traverse_module_state->__pyx_n_s_init_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_init___locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_init_subclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_initial_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_inline_namespaces);
  Py_VISIT(traverse_module_state->__pyx_kp_u_inline_namespaces_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_inspect);
  Py_VISIT(traverse_module_state->__pyx_n_s_inst);
  Py_VISIT(traverse_module_state->__pyx_kp_u_int);
  Py_VISIT(traverse_module_state->__pyx_n_u_int_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_int_date);
  Py_VISIT(traverse_module_state->__pyx_kp_u_int_time_millis);
  Py_VISIT(traverse_module_state->__pyx_n_s_integer);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_for_enum);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_value_include_array_ind);
  Py_VISIT(traverse_module_state->__pyx_n_s_invalid_value_include_array_inde);
  Py_VISIT(traverse_module_state->__pyx_n_s_invalid_value_include_map_key);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_value_include_map_key_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_value_includes_record_n);
  Py_VISIT(traverse_module_state->__pyx_n_s_invalid_value_includes_record_na);
  Py_VISIT(traverse_module_state->__pyx_n_s_io);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_a_mapping);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_a_valid_key_for_map_type);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_a_valid_name);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_a_valid_union_value_expe);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_a_valid_union_value_unkn);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_a_valid_value_for_array);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_a_valid_value_for_map_ty);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_allowed_as_a_name);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_valid);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_valid_for_a_union_of);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_valid_for_bytes_field);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_valid_for_field);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_valid_for_fixed_field);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_valid_for_union_with_fir);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_required_for_record);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_isinf);
  Py_VISIT(traverse_module_state->__pyx_n_s_isnan);
  Py_VISIT(traverse_module_state->__pyx_n_s_item_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_items);
  Py_VISIT(traverse_module_state->__pyx_n_u_items);
  Py_VISIT(traverse_module_state->__pyx_kp_u_items_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_itemsize);
  Py_VISIT(traverse_module_state->__pyx_kp_s_itemsize_0_for_cython_array);
  Py_VISIT(traverse_module_state->__pyx_n_s_json);
  Py_VISIT(traverse_module_state->__pyx_n_s_json_decode);
  Py_VISIT(traverse_module_state->__pyx_n_s_json_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_kw_only);
  Py_VISIT(traverse_module_state->__pyx_n_s_kwargs);
  Py_VISIT(traverse_module_state->__pyx_n_b_l);
  Py_VISIT(traverse_module_state->__pyx_n_u_latin1);
  Py_VISIT(traverse_module_state->__pyx_kp_u_latin_1);
  Py_VISIT(traverse_module_state->__pyx_n_u_little);
  Py_VISIT(traverse_module_state->__pyx_n_s_loads);
  Py_VISIT(traverse_module_state->__pyx_n_s_local);
  Py_VISIT(traverse_module_state->__pyx_n_s_log10);
  Py_VISIT(traverse_module_state->__pyx_n_u_logicalType);
  Py_VISIT(traverse_module_state->__pyx_n_s_logical_by_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_logical_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_logical_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_logical_types);
  Py_VISIT(traverse_module_state->__pyx_kp_u_logical_types_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_long);
  Py_VISIT(traverse_module_state->__pyx_n_u_long_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_long_time_micros);
  Py_VISIT(traverse_module_state->__pyx_kp_u_long_timestamp_micros);
  Py_VISIT(traverse_module_state->__pyx_kp_u_long_timestamp_millis);
  Py_VISIT(traverse_module_state->__pyx_n_b_lz4);
  Py_VISIT(traverse_module_state->__pyx_n_s_lz4);
  Py_VISIT(traverse_module_state->__pyx_n_u_lz4);
  Py_VISIT(traverse_module_state->__pyx_n_s_lz4_frame);
  Py_VISIT(traverse_module_state->__pyx_n_s_lzma);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_make_dataclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_make_logical_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_map);
  Py_VISIT(traverse_module_state->__pyx_n_u_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_marker);
  Py_VISIT(traverse_module_state->__pyx_n_s_mask);
  Py_VISIT(traverse_module_state->__pyx_n_s_match_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_math);
  Py_VISIT(traverse_module_state->__pyx_n_s_max);
  Py_VISIT(traverse_module_state->__pyx_n_s_max_blocksize);
  Py_VISIT(traverse_module_state->__pyx_n_s_max_precision);
  Py_VISIT(traverse_module_state->__pyx_n_s_maxunicode);
  Py_VISIT(traverse_module_state->__pyx_n_s_memview);
  Py_VISIT(traverse_module_state->__pyx_n_s_meta);
  Py_VISIT(traverse_module_state->__pyx_n_s_metaclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_metadata);
  Py_VISIT(traverse_module_state->__pyx_n_s_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_micros_to_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_microsecond);
  Py_VISIT(traverse_module_state->__pyx_n_s_microseconds);
  Py_VISIT(traverse_module_state->__pyx_n_s_minute);
  Py_VISIT(traverse_module_state->__pyx_kp_u_missing);
  Py_VISIT(traverse_module_state->__pyx_kp_u_missing_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_missing_values_can_be_empty_con);
  Py_VISIT(traverse_module_state->__pyx_n_s_missing_values_can_be_empty_cont);
  Py_VISIT(traverse_module_state->__pyx_n_s_missing_values_can_be_null);
  Py_VISIT(traverse_module_state->__pyx_kp_u_missing_values_can_be_null_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_module);
  Py_VISIT(traverse_module_state->__pyx_n_s_mro_entries);
  Py_VISIT(traverse_module_state->__pyx_n_s_msg);
  Py_VISIT(traverse_module_state->__pyx_n_s_n);
  Py_VISIT(traverse_module_state->__pyx_n_s_n_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_naive_dt_assume_utc);
  Py_VISIT(traverse_module_state->__pyx_kp_u_naive_dt_assume_utc_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_pattern);
  Py_VISIT(traverse_module_state->__pyx_n_s_named_type_names_must_be_unique);
  Py_VISIT(traverse_module_state->__pyx_kp_u_named_type_names_must_be_unique_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_named_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_namespace);
  Py_VISIT(traverse_module_state->__pyx_n_u_namespace);
  Py_VISIT(traverse_module_state->__pyx_n_s_namespaced);
  Py_VISIT(traverse_module_state->__pyx_n_u_nan);
  Py_VISIT(traverse_module_state->__pyx_n_s_ndarray);
  Py_VISIT(traverse_module_state->__pyx_n_s_ndim);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_n_s_new_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_new_inst);
  Py_VISIT(traverse_module_state->__pyx_n_s_new_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_new_ver);
  Py_VISIT(traverse_module_state->__pyx_n_s_next_object);
  Py_VISIT(traverse_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_VISIT(traverse_module_state->__pyx_kp_u_not_compatible_with_NullType);
  Py_VISIT(traverse_module_state->__pyx_kp_u_not_in_enum_symbols);
  Py_VISIT(traverse_module_state->__pyx_kp_u_not_valid_for);
  Py_VISIT(traverse_module_state->__pyx_kp_u_not_valid_for_long);
  Py_VISIT(traverse_module_state->__pyx_n_s_np);
  Py_VISIT(traverse_module_state->__pyx_n_s_np_f16);
  Py_VISIT(traverse_module_state->__pyx_n_s_np_f32);
  Py_VISIT(traverse_module_state->__pyx_n_s_np_f64);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ns);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ns_copy);
  Py_VISIT(traverse_module_state->__pyx_kp_u_null);
  Py_VISIT(traverse_module_state->__pyx_n_b_null_2);
  Py_VISIT(traverse_module_state->__pyx_n_u_null_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_null_pad_fixed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_null_pad_fixed_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_numpy);
  Py_VISIT(traverse_module_state->__pyx_n_s_ob);
  Py_VISIT(traverse_module_state->__pyx_n_s_obj);
  Py_VISIT(traverse_module_state->__pyx_n_s_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_objs);
  Py_VISIT(traverse_module_state->__pyx_n_s_open);
  Py_VISIT(traverse_module_state->__pyx_n_s_options);
  Py_VISIT(traverse_module_state->__pyx_n_s_order);
  Py_VISIT(traverse_module_state->__pyx_n_u_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_orig_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_os);
  Py_VISIT(traverse_module_state->__pyx_n_s_other);
  Py_VISIT(traverse_module_state->__pyx_n_s_other_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_out);
  Py_VISIT(traverse_module_state->__pyx_n_s_out_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pack);
  Py_VISIT(traverse_module_state->__pyx_n_s_pad);
  Py_VISIT(traverse_module_state->__pyx_n_s_parse_json);
  Py_VISIT(traverse_module_state->__pyx_n_s_parse_namespace);
  Py_VISIT(traverse_module_state->__pyx_n_s_partial);
  Py_VISIT(traverse_module_state->__pyx_n_s_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_path_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_pathlib);
  Py_VISIT(traverse_module_state->__pyx_n_s_perf);
  Py_VISIT(traverse_module_state->__pyx_n_s_perf_counter_ns);
  Py_VISIT(traverse_module_state->__pyx_n_s_perf_locals__test_zigzag_perf);
  Py_VISIT(traverse_module_state->__pyx_n_s_perf_locals_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_perf_locals_add_locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_perf_tests);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_u_placeholder);
  Py_VISIT(traverse_module_state->__pyx_n_s_pop);
  Py_VISIT(traverse_module_state->__pyx_n_s_prec);
  Py_VISIT(traverse_module_state->__pyx_n_s_precision);
  Py_VISIT(traverse_module_state->__pyx_n_u_precision);
  Py_VISIT(traverse_module_state->__pyx_kp_u_precision_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_prepare);
  Py_VISIT(traverse_module_state->__pyx_n_s_print);
  Py_VISIT(traverse_module_state->__pyx_n_s_pytest);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_recursive_repr_guard);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_ArrayType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_AvroType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_BoolType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_BytesType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_CannotPromote);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Codec);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_ContainerReader);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_CustomLogicalType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Date);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_DecimalType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_DoubleType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Enum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_EnumType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_FileReader);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_FileWriter);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_FixedType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_FloatType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_IntType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_LogicalType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_LongType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_MapType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_MemoryReader);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_NullType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Options);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PromoteBytesToStr);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PromoteStringToBy);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PromoteToFloat);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PromotingEnumType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PromotingRecordTy);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Rabin);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Record);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_RecordField);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_RecordType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_ResolvedSchema);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_StringType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_TimeMicros);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_TimeMillis);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_TimestampMicros);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_TimestampMillis);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_UUIDBase);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_UUIDFixedType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_UUIDStringType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_UnionType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_ValueAdapter);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__CanonicalForm);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__DeflateCodec);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__FieldAccessor);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__Lz4Codec);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__LzmaCodec);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__NamedType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__NullCodec);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__PlaceholderType);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__Reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__Sentinel);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__SnappyCodec);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__Writer);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle__ZStandardCodec);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle___Bzip2Codec);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_b_q);
  Py_VISIT(traverse_module_state->__pyx_n_s_qualname);
  Py_VISIT(traverse_module_state->__pyx_n_s_quantize);
  Py_VISIT(traverse_module_state->__pyx_n_u_rabin);
  Py_VISIT(traverse_module_state->__pyx_n_s_raise);
  Py_VISIT(traverse_module_state->__pyx_n_s_raise_on_invalid_logical);
  Py_VISIT(traverse_module_state->__pyx_kp_u_raise_on_invalid_logical_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_raises);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_u_rb);
  Py_VISIT(traverse_module_state->__pyx_n_s_re);
  Py_VISIT(traverse_module_state->__pyx_n_s_read);
  Py_VISIT(traverse_module_state->__pyx_n_u_read);
  Py_VISIT(traverse_module_state->__pyx_n_s_read_marker);
  Py_VISIT(traverse_module_state->__pyx_n_s_read_perf);
  Py_VISIT(traverse_module_state->__pyx_n_s_read_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_reader_for_writer);
  Py_VISIT(traverse_module_state->__pyx_n_s_reader_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_reader_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_record);
  Py_VISIT(traverse_module_state->__pyx_n_u_record);
  Py_VISIT(traverse_module_state->__pyx_n_s_record_allow_extra_fields);
  Py_VISIT(traverse_module_state->__pyx_kp_u_record_allow_extra_fields_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_record_can_encode_dict);
  Py_VISIT(traverse_module_state->__pyx_kp_u_record_can_encode_dict_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_record_decodes_to_dict);
  Py_VISIT(traverse_module_state->__pyx_kp_u_record_decodes_to_dict_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_record_encode_use_defaults);
  Py_VISIT(traverse_module_state->__pyx_kp_u_record_encode_use_defaults_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_record_fields_must_be_unique);
  Py_VISIT(traverse_module_state->__pyx_kp_u_record_fields_must_be_unique_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_record_values_type_hint);
  Py_VISIT(traverse_module_state->__pyx_kp_u_record_values_type_hint_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_register);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove);
  Py_VISIT(traverse_module_state->__pyx_n_s_replace);
  Py_VISIT(traverse_module_state->__pyx_n_u_replace);
  Py_VISIT(traverse_module_state->__pyx_n_s_repr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_required_field);
  Py_VISIT(traverse_module_state->__pyx_n_s_resolved_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_return);
  Py_VISIT(traverse_module_state->__pyx_n_s_return_uuid_object);
  Py_VISIT(traverse_module_state->__pyx_kp_u_return_uuid_object_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_reversed);
  Py_VISIT(traverse_module_state->__pyx_n_s_rjust);
  Py_VISIT(traverse_module_state->__pyx_n_s_rsplit);
  Py_VISIT(traverse_module_state->__pyx_n_s_running);
  Py_VISIT(traverse_module_state->__pyx_kp_u_s_read);
  Py_VISIT(traverse_module_state->__pyx_n_s_scale);
  Py_VISIT(traverse_module_state->__pyx_n_u_scale);
  Py_VISIT(traverse_module_state->__pyx_n_s_scaleb);
  Py_VISIT(traverse_module_state->__pyx_n_s_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_schema_ob);
  Py_VISIT(traverse_module_state->__pyx_n_s_schema_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_second);
  Py_VISIT(traverse_module_state->__pyx_n_s_seconds);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_self_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_serialize);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_from_dataclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_setdefault);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_shape);
  Py_VISIT(traverse_module_state->__pyx_n_s_signature);
  Py_VISIT(traverse_module_state->__pyx_n_s_signature_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_signed);
  Py_VISIT(traverse_module_state->__pyx_n_s_size);
  Py_VISIT(traverse_module_state->__pyx_n_u_size);
  Py_VISIT(traverse_module_state->__pyx_n_u_slots);
  Py_VISIT(traverse_module_state->__pyx_n_s_slots_2);
  Py_VISIT(traverse_module_state->__pyx_n_b_snappy);
  Py_VISIT(traverse_module_state->__pyx_n_s_snappy);
  Py_VISIT(traverse_module_state->__pyx_n_u_snappy);
  Py_VISIT(traverse_module_state->__pyx_n_s_source);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_s_src);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_array_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_container_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_enum_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_error_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_logical_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_map_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_null_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_numeric_types_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_option_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_rabin_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_record_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_schema_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_string_types_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_tests_test_buffer_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_tests_test_perf_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_tests_test_zigzag_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_type_pxi);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_union_pxi);
  Py_VISIT(traverse_module_state->__pyx_n_s_start);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_step);
  Py_VISIT(traverse_module_state->__pyx_n_s_stop);
  Py_VISIT(traverse_module_state->__pyx_n_s_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_str_2);
  Py_VISIT(traverse_module_state->__pyx_n_u_strict);
  Py_VISIT(traverse_module_state->__pyx_kp_s_strided_and_direct);
  Py_VISIT(traverse_module_state->__pyx_kp_s_strided_and_direct_or_indirect);
  Py_VISIT(traverse_module_state->__pyx_kp_s_strided_and_indirect);
  Py_VISIT(traverse_module_state->__pyx_n_u_string);
  Py_VISIT(traverse_module_state->__pyx_kp_u_string_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_string_types_default_unchanged);
  Py_VISIT(traverse_module_state->__pyx_kp_u_string_types_default_unchanged_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_string_uuid);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_strptime);
  Py_VISIT(traverse_module_state->__pyx_n_s_struct);
  Py_VISIT(traverse_module_state->__pyx_n_s_sum);
  Py_VISIT(traverse_module_state->__pyx_n_s_super);
  Py_VISIT(traverse_module_state->__pyx_n_s_symbols);
  Py_VISIT(traverse_module_state->__pyx_n_u_symbols);
  Py_VISIT(traverse_module_state->__pyx_n_s_sys);
  Py_VISIT(traverse_module_state->__pyx_n_s_t);
  Py_VISIT(traverse_module_state->__pyx_n_s_t_a);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_reader_read_n);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_reader_read_u8);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_readvarint);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_readvarlong);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_zigzag_long);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_zigzag_perf);
  Py_VISIT(traverse_module_state->__pyx_n_s_tests);
  Py_VISIT(traverse_module_state->__pyx_n_s_tests_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_tests_locals__test_reader_read);
  Py_VISIT(traverse_module_state->__pyx_n_s_tests_locals__test_reader_read_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_tests_locals_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_tests_locals_add_locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_tests_locals_test_readvarint);
  Py_VISIT(traverse_module_state->__pyx_n_s_tests_locals_test_readvarlong);
  Py_VISIT(traverse_module_state->__pyx_n_s_tests_locals_test_zigzag_long);
  Py_VISIT(traverse_module_state->__pyx_n_s_threading);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_time);
  Py_VISIT(traverse_module_state->__pyx_kp_u_time_micros);
  Py_VISIT(traverse_module_state->__pyx_kp_u_time_millis);
  Py_VISIT(traverse_module_state->__pyx_n_s_time_to_micros);
  Py_VISIT(traverse_module_state->__pyx_n_s_timedelta);
  Py_VISIT(traverse_module_state->__pyx_n_s_timestamp);
  Py_VISIT(traverse_module_state->__pyx_kp_u_timestamp_micros);
  Py_VISIT(traverse_module_state->__pyx_kp_u_timestamp_millis);
  Py_VISIT(traverse_module_state->__pyx_n_s_timezone);
  Py_VISIT(traverse_module_state->__pyx_kp_u_to);
  Py_VISIT(traverse_module_state->__pyx_n_s_to_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_total_nums);
  Py_VISIT(traverse_module_state->__pyx_n_s_truncate_fixed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_truncate_fixed_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_tuple_LogicalType);
  Py_VISIT(traverse_module_state->__pyx_n_s_type);
  Py_VISIT(traverse_module_state->__pyx_n_u_type);
  Py_VISIT(traverse_module_state->__pyx_kp_u_type_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_type_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_type_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_type_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_types_str_to_schema);
  Py_VISIT(traverse_module_state->__pyx_kp_u_types_str_to_schema_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_typing);
  Py_VISIT(traverse_module_state->__pyx_n_s_tzinfo);
  Py_VISIT(traverse_module_state->__pyx_n_s_ubin);
  Py_VISIT(traverse_module_state->__pyx_kp_s_unable_to_allocate_array_data);
  Py_VISIT(traverse_module_state->__pyx_kp_s_unable_to_allocate_shape_and_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_underlying);
  Py_VISIT(traverse_module_state->__pyx_n_s_underlying_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_underlying_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_underlying_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_unicode_errors);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unicode_errors_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_unicodedata);
  Py_VISIT(traverse_module_state->__pyx_n_u_union);
  Py_VISIT(traverse_module_state->__pyx_n_s_union_decodes_to);
  Py_VISIT(traverse_module_state->__pyx_kp_u_union_decodes_to_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_union_json_encodes_type_name);
  Py_VISIT(traverse_module_state->__pyx_kp_u_union_json_encodes_type_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_unpack);
  Py_VISIT(traverse_module_state->__pyx_n_s_unsafe_hash);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_utc);
  Py_VISIT(traverse_module_state->__pyx_n_u_utf8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_utf_8);
  Py_VISIT(traverse_module_state->__pyx_n_s_uuid);
  Py_VISIT(traverse_module_state->__pyx_n_u_uuid);
  Py_VISIT(traverse_module_state->__pyx_n_s_uuid4);
  Py_VISIT(traverse_module_state->__pyx_n_s_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_kp_u_value_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_value_adapters);
  Py_VISIT(traverse_module_state->__pyx_n_s_value_type);
  Py_VISIT(traverse_module_state->__pyx_n_u_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_version);
  Py_VISIT(traverse_module_state->__pyx_n_s_version_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_visited);
  Py_VISIT(traverse_module_state->__pyx_kp_u_w);
  Py_VISIT(traverse_module_state->__pyx_n_s_walk_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_warn);
  Py_VISIT(traverse_module_state->__pyx_n_s_warnings);
  Py_VISIT(traverse_module_state->__pyx_n_u_wb);
  Py_VISIT(traverse_module_state->__pyx_n_s_wbits);
  Py_VISIT(traverse_module_state->__pyx_n_s_weakref_slot);
  Py_VISIT(traverse_module_state->__pyx_kp_u_which_is_greater_than_the_scale);
  Py_VISIT(traverse_module_state->__pyx_n_s_with_external_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_with_logical_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_wrap_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_wrapped);
  Py_VISIT(traverse_module_state->__pyx_n_s_write);
  Py_VISIT(traverse_module_state->__pyx_n_u_write);
  Py_VISIT(traverse_module_state->__pyx_kp_u_write_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_write_header);
  Py_VISIT(traverse_module_state->__pyx_n_s_write_many);
  Py_VISIT(traverse_module_state->__pyx_n_s_write_one);
  Py_VISIT(traverse_module_state->__pyx_n_s_write_perf);
  Py_VISIT(traverse_module_state->__pyx_n_s_write_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_writer);
  Py_VISIT(traverse_module_state->__pyx_n_s_writer_schema);
  Py_VISIT(traverse_module_state->__pyx_n_s_writer_type);
  Py_VISIT(traverse_module_state->__pyx_n_b_xz);
  Py_VISIT(traverse_module_state->__pyx_n_u_xz);
  Py_VISIT(traverse_module_state->__pyx_n_s_zip);
  Py_VISIT(traverse_module_state->__pyx_n_s_zlib);
  Py_VISIT(traverse_module_state->__pyx_n_b_zstandard);
  Py_VISIT(traverse_module_state->__pyx_n_s_zstandard);
  Py_VISIT(traverse_module_state->__pyx_n_u_zstandard);
  Py_VISIT(traverse_module_state->__pyx_float_0_0);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_2);
  Py_VISIT(traverse_module_state->__pyx_int_3);
  Py_VISIT(traverse_module_state->__pyx_int_4);
  Py_VISIT(traverse_module_state->__pyx_int_8);
  Py_VISIT(traverse_module_state->__pyx_int_10);
  Py_VISIT(traverse_module_state->__pyx_int_16);
  Py_VISIT(traverse_module_state->__pyx_int_24);
  Py_VISIT(traverse_module_state->__pyx_int_60);
  Py_VISIT(traverse_module_state->__pyx_int_70);
  Py_VISIT(traverse_module_state->__pyx_int_127);
  Py_VISIT(traverse_module_state->__pyx_int_255);
  Py_VISIT(traverse_module_state->__pyx_int_1000);
  Py_VISIT(traverse_module_state->__pyx_int_1970);
  Py_VISIT(traverse_module_state->__pyx_int_3600);
  Py_VISIT(traverse_module_state->__pyx_int_4096);
  Py_VISIT(traverse_module_state->__pyx_int_32767);
  Py_VISIT(traverse_module_state->__pyx_int_65536);
  Py_VISIT(traverse_module_state->__pyx_int_1000000);
  Py_VISIT(traverse_module_state->__pyx_int_8388607);
  Py_VISIT(traverse_module_state->__pyx_int_18332779);
  Py_VISIT(traverse_module_state->__pyx_int_24827247);
  Py_VISIT(traverse_module_state->__pyx_int_25341170);
  Py_VISIT(traverse_module_state->__pyx_int_29099189);
  Py_VISIT(traverse_module_state->__pyx_int_31039736);
  Py_VISIT(traverse_module_state->__pyx_int_33963030);
  Py_VISIT(traverse_module_state->__pyx_int_38410757);
  Py_VISIT(traverse_module_state->__pyx_int_38632982);
  Py_VISIT(traverse_module_state->__pyx_int_39872412);
  Py_VISIT(traverse_module_state->__pyx_int_42565757);
  Py_VISIT(traverse_module_state->__pyx_int_47772561);
  Py_VISIT(traverse_module_state->__pyx_int_60000000);
  Py_VISIT(traverse_module_state->__pyx_int_61270791);
  Py_VISIT(traverse_module_state->__pyx_int_61440409);
  Py_VISIT(traverse_module_state->__pyx_int_61765448);
  Py_VISIT(traverse_module_state->__pyx_int_67877419);
  Py_VISIT(traverse_module_state->__pyx_int_75337510);
  Py_VISIT(traverse_module_state->__pyx_int_79389776);
  Py_VISIT(traverse_module_state->__pyx_int_79941127);
  Py_VISIT(traverse_module_state->__pyx_int_81233721);
  Py_VISIT(traverse_module_state->__pyx_int_83351333);
  Py_VISIT(traverse_module_state->__pyx_int_95761985);
  Py_VISIT(traverse_module_state->__pyx_int_96605483);
  Py_VISIT(traverse_module_state->__pyx_int_101034655);
  Py_VISIT(traverse_module_state->__pyx_int_101670942);
  Py_VISIT(traverse_module_state->__pyx_int_111776469);
  Py_VISIT(traverse_module_state->__pyx_int_111957047);
  Py_VISIT(traverse_module_state->__pyx_int_112105877);
  Py_VISIT(traverse_module_state->__pyx_int_115671040);
  Py_VISIT(traverse_module_state->__pyx_int_119832798);
  Py_VISIT(traverse_module_state->__pyx_int_122446337);
  Py_VISIT(traverse_module_state->__pyx_int_122978751);
  Py_VISIT(traverse_module_state->__pyx_int_136983863);
  Py_VISIT(traverse_module_state->__pyx_int_139219010);
  Py_VISIT(traverse_module_state->__pyx_int_145104888);
  Py_VISIT(traverse_module_state->__pyx_int_148208818);
  Py_VISIT(traverse_module_state->__pyx_int_148222551);
  Py_VISIT(traverse_module_state->__pyx_int_148338675);
  Py_VISIT(traverse_module_state->__pyx_int_148737976);
  Py_VISIT(traverse_module_state->__pyx_int_150354737);
  Py_VISIT(traverse_module_state->__pyx_int_159023012);
  Py_VISIT(traverse_module_state->__pyx_int_159276020);
  Py_VISIT(traverse_module_state->__pyx_int_159891379);
  Py_VISIT(traverse_module_state->__pyx_int_160503772);
  Py_VISIT(traverse_module_state->__pyx_int_160934521);
  Py_VISIT(traverse_module_state->__pyx_int_164094308);
  Py_VISIT(traverse_module_state->__pyx_int_169331114);
  Py_VISIT(traverse_module_state->__pyx_int_175892883);
  Py_VISIT(traverse_module_state->__pyx_int_176081420);
  Py_VISIT(traverse_module_state->__pyx_int_176267327);
  Py_VISIT(traverse_module_state->__pyx_int_177382991);
  Py_VISIT(traverse_module_state->__pyx_int_178791180);
  Py_VISIT(traverse_module_state->__pyx_int_180170947);
  Py_VISIT(traverse_module_state->__pyx_int_180285838);
  Py_VISIT(traverse_module_state->__pyx_int_181080083);
  Py_VISIT(traverse_module_state->__pyx_int_184956420);
  Py_VISIT(traverse_module_state->__pyx_int_184977713);
  Py_VISIT(traverse_module_state->__pyx_int_193777858);
  Py_VISIT(traverse_module_state->__pyx_int_195980960);
  Py_VISIT(traverse_module_state->__pyx_int_198151601);
  Py_VISIT(traverse_module_state->__pyx_int_202158861);
  Py_VISIT(traverse_module_state->__pyx_int_202608488);
  Py_VISIT(traverse_module_state->__pyx_int_203789025);
  Py_VISIT(traverse_module_state->__pyx_int_205941106);
  Py_VISIT(traverse_module_state->__pyx_int_206778879);
  Py_VISIT(traverse_module_state->__pyx_int_212073902);
  Py_VISIT(traverse_module_state->__pyx_int_213886906);
  Py_VISIT(traverse_module_state->__pyx_int_215229444);
  Py_VISIT(traverse_module_state->__pyx_int_215954136);
  Py_VISIT(traverse_module_state->__pyx_int_222419149);
  Py_VISIT(traverse_module_state->__pyx_int_228539134);
  Py_VISIT(traverse_module_state->__pyx_int_228825662);
  Py_VISIT(traverse_module_state->__pyx_int_229636597);
  Py_VISIT(traverse_module_state->__pyx_int_233298401);
  Py_VISIT(traverse_module_state->__pyx_int_236180632);
  Py_VISIT(traverse_module_state->__pyx_int_236349525);
  Py_VISIT(traverse_module_state->__pyx_int_238750788);
  Py_VISIT(traverse_module_state->__pyx_int_240389341);
  Py_VISIT(traverse_module_state->__pyx_int_245215779);
  Py_VISIT(traverse_module_state->__pyx_int_247527084);
  Py_VISIT(traverse_module_state->__pyx_int_254981756);
  Py_VISIT(traverse_module_state->__pyx_int_256569670);
  Py_VISIT(traverse_module_state->__pyx_int_259444540);
  Py_VISIT(traverse_module_state->__pyx_int_261059191);
  Py_VISIT(traverse_module_state->__pyx_int_263819094);
  Py_VISIT(traverse_module_state->__pyx_int_2147483647);
  Py_VISIT(traverse_module_state->__pyx_int_3442987655);
  Py_VISIT(traverse_module_state->__pyx_int_3600000000);
  Py_VISIT(traverse_module_state->__pyx_int_neg_1);
  Py_VISIT(traverse_module_state->__pyx_int_neg_15);
  Py_VISIT(traverse_module_state->__pyx_int_549755813887L);
  Py_VISIT(traverse_module_state->__pyx_int_0x7fffffffffffL);
  Py_VISIT(traverse_module_state->__pyx_int_0x7fffffffffffffL);
  Py_VISIT(traverse_module_state->__pyx_int_0x7fffffffffffffffL);
  Py_VISIT(traverse_module_state->__pyx_k__15);
  Py_VISIT(traverse_module_state->__pyx_k__16);
  Py_VISIT(traverse_module_state->__pyx_k__17);
  Py_VISIT(traverse_module_state->__pyx_k__21);
  Py_VISIT(traverse_module_state->__pyx_k__81);
  Py_VISIT(traverse_module_state->__pyx_k__83);
  Py_VISIT(traverse_module_state->__pyx_k__84);
  Py_VISIT(traverse_module_state->__pyx_k__87);
  Py_VISIT(traverse_module_state->__pyx_k__89);
  Py_VISIT(traverse_module_state->__pyx_slice__5);
  Py_VISIT(traverse_module_state->__pyx_tuple__4);
  Py_VISIT(traverse_module_state->__pyx_tuple__8);
  Py_VISIT(traverse_module_state->__pyx_slice__23);
  Py_VISIT(traverse_module_state->__pyx_slice__86);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__19);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__28);
  Py_VISIT(traverse_module_state->__pyx_tuple__30);
  Py_VISIT(traverse_module_state->__pyx_tuple__32);
  Py_VISIT(traverse_module_state->__pyx_tuple__35);
  Py_VISIT(traverse_module_state->__pyx_tuple__36);
  Py_VISIT(traverse_module_state->__pyx_tuple__37);
  Py_VISIT(traverse_module_state->__pyx_tuple__38);
  Py_VISIT(traverse_module_state->__pyx_tuple__40);
  Py_VISIT(traverse_module_state->__pyx_tuple__42);
  Py_VISIT(traverse_module_state->__pyx_tuple__43);
  Py_VISIT(traverse_module_state->__pyx_tuple__46);
  Py_VISIT(traverse_module_state->__pyx_tuple__47);
  Py_VISIT(traverse_module_state->__pyx_tuple__50);
  Py_VISIT(traverse_module_state->__pyx_tuple__51);
  Py_VISIT(traverse_module_state->__pyx_tuple__52);
  Py_VISIT(traverse_module_state->__pyx_tuple__53);
  Py_VISIT(traverse_module_state->__pyx_tuple__54);
  Py_VISIT(traverse_module_state->__pyx_tuple__55);
  Py_VISIT(traverse_module_state->__pyx_tuple__56);
  Py_VISIT(traverse_module_state->__pyx_tuple__57);
  Py_VISIT(traverse_module_state->__pyx_tuple__58);
  Py_VISIT(traverse_module_state->__pyx_tuple__59);
  Py_VISIT(traverse_module_state->__pyx_tuple__65);
  Py_VISIT(traverse_module_state->__pyx_tuple__67);
  Py_VISIT(traverse_module_state->__pyx_tuple__68);
  Py_VISIT(traverse_module_state->__pyx_tuple__69);
  Py_VISIT(traverse_module_state->__pyx_tuple__71);
  Py_VISIT(traverse_module_state->__pyx_tuple__73);
  Py_VISIT(traverse_module_state->__pyx_tuple__74);
  Py_VISIT(traverse_module_state->__pyx_tuple__85);
  Py_VISIT(traverse_module_state->__pyx_tuple__88);
  Py_VISIT(traverse_module_state->__pyx_tuple__90);
  Py_VISIT(traverse_module_state->__pyx_tuple__91);
  Py_VISIT(traverse_module_state->__pyx_tuple__92);
  Py_VISIT(traverse_module_state->__pyx_tuple__93);
  Py_VISIT(traverse_module_state->__pyx_tuple__94);
  Py_VISIT(traverse_module_state->__pyx_tuple__95);
  Py_VISIT(traverse_module_state->__pyx_tuple__96);
  Py_VISIT(traverse_module_state->__pyx_tuple__97);
  Py_VISIT(traverse_module_state->__pyx_tuple__98);
  Py_VISIT(traverse_module_state->__pyx_tuple__99);
  Py_VISIT(traverse_module_state->__pyx_tuple__100);
  Py_VISIT(traverse_module_state->__pyx_tuple__101);
  Py_VISIT(traverse_module_state->__pyx_tuple__102);
  Py_VISIT(traverse_module_state->__pyx_tuple__103);
  Py_VISIT(traverse_module_state->__pyx_tuple__104);
  Py_VISIT(traverse_module_state->__pyx_tuple__105);
  Py_VISIT(traverse_module_state->__pyx_tuple__106);
  Py_VISIT(traverse_module_state->__pyx_tuple__107);
  Py_VISIT(traverse_module_state->__pyx_tuple__108);
  Py_VISIT(traverse_module_state->__pyx_tuple__109);
  Py_VISIT(traverse_module_state->__pyx_tuple__110);
  Py_VISIT(traverse_module_state->__pyx_tuple__111);
  Py_VISIT(traverse_module_state->__pyx_tuple__112);
  Py_VISIT(traverse_module_state->__pyx_tuple__113);
  Py_VISIT(traverse_module_state->__pyx_tuple__114);
  Py_VISIT(traverse_module_state->__pyx_tuple__115);
  Py_VISIT(traverse_module_state->__pyx_tuple__116);
  Py_VISIT(traverse_module_state->__pyx_tuple__117);
  Py_VISIT(traverse_module_state->__pyx_tuple__118);
  Py_VISIT(traverse_module_state->__pyx_tuple__119);
  Py_VISIT(traverse_module_state->__pyx_tuple__120);
  Py_VISIT(traverse_module_state->__pyx_tuple__121);
  Py_VISIT(traverse_module_state->__pyx_tuple__123);
  Py_VISIT(traverse_module_state->__pyx_tuple__124);
  Py_VISIT(traverse_module_state->__pyx_tuple__125);
  Py_VISIT(traverse_module_state->__pyx_tuple__126);
  Py_VISIT(traverse_module_state->__pyx_tuple__127);
  Py_VISIT(traverse_module_state->__pyx_tuple__128);
  Py_VISIT(traverse_module_state->__pyx_tuple__129);
  Py_VISIT(traverse_module_state->__pyx_tuple__130);
  Py_VISIT(traverse_module_state->__pyx_tuple__131);
  Py_VISIT(traverse_module_state->__pyx_tuple__132);
  Py_VISIT(traverse_module_state->__pyx_tuple__134);
  Py_VISIT(traverse_module_state->__pyx_tuple__137);
  Py_VISIT(traverse_module_state->__pyx_tuple__139);
  Py_VISIT(traverse_module_state->__pyx_tuple__141);
  Py_VISIT(traverse_module_state->__pyx_tuple__142);
  Py_VISIT(traverse_module_state->__pyx_tuple__144);
  Py_VISIT(traverse_module_state->__pyx_tuple__145);
  Py_VISIT(traverse_module_state->__pyx_tuple__147);
  Py_VISIT(traverse_module_state->__pyx_tuple__148);
  Py_VISIT(traverse_module_state->__pyx_tuple__150);
  Py_VISIT(traverse_module_state->__pyx_tuple__153);
  Py_VISIT(traverse_module_state->__pyx_tuple__155);
  Py_VISIT(traverse_module_state->__pyx_tuple__157);
  Py_VISIT(traverse_module_state->__pyx_tuple__159);
  Py_VISIT(traverse_module_state->__pyx_tuple__161);
  Py_VISIT(traverse_module_state->__pyx_tuple__163);
  Py_VISIT(traverse_module_state->__pyx_tuple__164);
  Py_VISIT(traverse_module_state->__pyx_tuple__166);
  Py_VISIT(traverse_module_state->__pyx_tuple__167);
  Py_VISIT(traverse_module_state->__pyx_tuple__169);
  Py_VISIT(traverse_module_state->__pyx_tuple__183);
  Py_VISIT(traverse_module_state->__pyx_tuple__185);
  Py_VISIT(traverse_module_state->__pyx_tuple__188);
  Py_VISIT(traverse_module_state->__pyx_tuple__192);
  Py_VISIT(traverse_module_state->__pyx_tuple__194);
  Py_VISIT(traverse_module_state->__pyx_tuple__196);
  Py_VISIT(traverse_module_state->__pyx_tuple__198);
  Py_VISIT(traverse_module_state->__pyx_tuple__201);
  Py_VISIT(traverse_module_state->__pyx_tuple__202);
  Py_VISIT(traverse_module_state->__pyx_tuple__204);
  Py_VISIT(traverse_module_state->__pyx_tuple__206);
  Py_VISIT(traverse_module_state->__pyx_tuple__218);
  Py_VISIT(traverse_module_state->__pyx_tuple__220);
  Py_VISIT(traverse_module_state->__pyx_tuple__253);
  Py_VISIT(traverse_module_state->__pyx_tuple__258);
  Py_VISIT(traverse_module_state->__pyx_tuple__260);
  Py_VISIT(traverse_module_state->__pyx_tuple__264);
  Py_VISIT(traverse_module_state->__pyx_tuple__271);
  Py_VISIT(traverse_module_state->__pyx_tuple__294);
  Py_VISIT(traverse_module_state->__pyx_tuple__297);
  Py_VISIT(traverse_module_state->__pyx_tuple__300);
  Py_VISIT(traverse_module_state->__pyx_tuple__303);
  Py_VISIT(traverse_module_state->__pyx_tuple__306);
  Py_VISIT(traverse_module_state->__pyx_tuple__308);
  Py_VISIT(traverse_module_state->__pyx_tuple__313);
  Py_VISIT(traverse_module_state->__pyx_tuple__321);
  Py_VISIT(traverse_module_state->__pyx_tuple__323);
  Py_VISIT(traverse_module_state->__pyx_tuple__325);
  Py_VISIT(traverse_module_state->__pyx_tuple__327);
  Py_VISIT(traverse_module_state->__pyx_tuple__329);
  Py_VISIT(traverse_module_state->__pyx_tuple__331);
  Py_VISIT(traverse_module_state->__pyx_tuple__332);
  Py_VISIT(traverse_module_state->__pyx_tuple__334);
  Py_VISIT(traverse_module_state->__pyx_tuple__335);
  Py_VISIT(traverse_module_state->__pyx_tuple__337);
  Py_VISIT(traverse_module_state->__pyx_tuple__340);
  Py_VISIT(traverse_module_state->__pyx_tuple__342);
  Py_VISIT(traverse_module_state->__pyx_tuple__344);
  Py_VISIT(traverse_module_state->__pyx_tuple__346);
  Py_VISIT(traverse_module_state->__pyx_tuple__347);
  Py_VISIT(traverse_module_state->__pyx_tuple__349);
  Py_VISIT(traverse_module_state->__pyx_tuple__357);
  Py_VISIT(traverse_module_state->__pyx_tuple__363);
  Py_VISIT(traverse_module_state->__pyx_tuple__367);
  Py_VISIT(traverse_module_state->__pyx_tuple__377);
  Py_VISIT(traverse_module_state->__pyx_tuple__381);
  Py_VISIT(traverse_module_state->__pyx_tuple__396);
  Py_VISIT(traverse_module_state->__pyx_tuple__418);
  Py_VISIT(traverse_module_state->__pyx_tuple__420);
  Py_VISIT(traverse_module_state->__pyx_tuple__422);
  Py_VISIT(traverse_module_state->__pyx_tuple__424);
  Py_VISIT(traverse_module_state->__pyx_tuple__425);
  Py_VISIT(traverse_module_state->__pyx_tuple__427);
  Py_VISIT(traverse_module_state->__pyx_codeobj__11);
  Py_VISIT(traverse_module_state->__pyx_codeobj__12);
  Py_VISIT(traverse_module_state->__pyx_codeobj__29);
  Py_VISIT(traverse_module_state->__pyx_codeobj__31);
  Py_VISIT(traverse_module_state->__pyx_codeobj__33);
  Py_VISIT(traverse_module_state->__pyx_codeobj__39);
  Py_VISIT(traverse_module_state->__pyx_codeobj__41);
  Py_VISIT(traverse_module_state->__pyx_codeobj__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__45);
  Py_VISIT(traverse_module_state->__pyx_codeobj__49);
  Py_VISIT(traverse_module_state->__pyx_codeobj__66);
  Py_VISIT(traverse_module_state->__pyx_codeobj__75);
  Py_VISIT(traverse_module_state->__pyx_codeobj__82);
  Py_VISIT(traverse_module_state->__pyx_codeobj__133);
  Py_VISIT(traverse_module_state->__pyx_codeobj__135);
  Py_VISIT(traverse_module_state->__pyx_codeobj__136);
  Py_VISIT(traverse_module_state->__pyx_codeobj__138);
  Py_VISIT(traverse_module_state->__pyx_codeobj__140);
  Py_VISIT(traverse_module_state->__pyx_codeobj__143);
  Py_VISIT(traverse_module_state->__pyx_codeobj__146);
  Py_VISIT(traverse_module_state->__pyx_codeobj__149);
  Py_VISIT(traverse_module_state->__pyx_codeobj__151);
  Py_VISIT(traverse_module_state->__pyx_codeobj__154);
  Py_VISIT(traverse_module_state->__pyx_codeobj__156);
  Py_VISIT(traverse_module_state->__pyx_codeobj__158);
  Py_VISIT(traverse_module_state->__pyx_codeobj__160);
  Py_VISIT(traverse_module_state->__pyx_codeobj__162);
  Py_VISIT(traverse_module_state->__pyx_codeobj__165);
  Py_VISIT(traverse_module_state->__pyx_codeobj__168);
  Py_VISIT(traverse_module_state->__pyx_codeobj__170);
  Py_VISIT(traverse_module_state->__pyx_codeobj__171);
  Py_VISIT(traverse_module_state->__pyx_codeobj__172);
  Py_VISIT(traverse_module_state->__pyx_codeobj__173);
  Py_VISIT(traverse_module_state->__pyx_codeobj__174);
  Py_VISIT(traverse_module_state->__pyx_codeobj__175);
  Py_VISIT(traverse_module_state->__pyx_codeobj__176);
  Py_VISIT(traverse_module_state->__pyx_codeobj__177);
  Py_VISIT(traverse_module_state->__pyx_codeobj__178);
  Py_VISIT(traverse_module_state->__pyx_codeobj__179);
  Py_VISIT(traverse_module_state->__pyx_codeobj__180);
  Py_VISIT(traverse_module_state->__pyx_codeobj__181);
  Py_VISIT(traverse_module_state->__pyx_codeobj__182);
  Py_VISIT(traverse_module_state->__pyx_codeobj__184);
  Py_VISIT(traverse_module_state->__pyx_codeobj__186);
  Py_VISIT(traverse_module_state->__pyx_codeobj__187);
  Py_VISIT(traverse_module_state->__pyx_codeobj__189);
  Py_VISIT(traverse_module_state->__pyx_codeobj__190);
  Py_VISIT(traverse_module_state->__pyx_codeobj__191);
  Py_VISIT(traverse_module_state->__pyx_codeobj__193);
  Py_VISIT(traverse_module_state->__pyx_codeobj__195);
  Py_VISIT(traverse_module_state->__pyx_codeobj__197);
  Py_VISIT(traverse_module_state->__pyx_codeobj__199);
  Py_VISIT(traverse_module_state->__pyx_codeobj__200);
  Py_VISIT(traverse_module_state->__pyx_codeobj__203);
  Py_VISIT(traverse_module_state->__pyx_codeobj__205);
  Py_VISIT(traverse_module_state->__pyx_codeobj__207);
  Py_VISIT(traverse_module_state->__pyx_codeobj__208);
  Py_VISIT(traverse_module_state->__pyx_codeobj__209);
  Py_VISIT(traverse_module_state->__pyx_codeobj__210);
  Py_VISIT(traverse_module_state->__pyx_codeobj__211);
  Py_VISIT(traverse_module_state->__pyx_codeobj__212);
  Py_VISIT(traverse_module_state->__pyx_codeobj__213);
  Py_VISIT(traverse_module_state->__pyx_codeobj__214);
  Py_VISIT(traverse_module_state->__pyx_codeobj__215);
  Py_VISIT(traverse_module_state->__pyx_codeobj__216);
  Py_VISIT(traverse_module_state->__pyx_codeobj__217);
  Py_VISIT(traverse_module_state->__pyx_codeobj__219);
  Py_VISIT(traverse_module_state->__pyx_codeobj__221);
  Py_VISIT(traverse_module_state->__pyx_codeobj__222);
  Py_VISIT(traverse_module_state->__pyx_codeobj__223);
  Py_VISIT(traverse_module_state->__pyx_codeobj__224);
  Py_VISIT(traverse_module_state->__pyx_codeobj__225);
  Py_VISIT(traverse_module_state->__pyx_codeobj__226);
  Py_VISIT(traverse_module_state->__pyx_codeobj__227);
  Py_VISIT(traverse_module_state->__pyx_codeobj__228);
  Py_VISIT(traverse_module_state->__pyx_codeobj__229);
  Py_VISIT(traverse_module_state->__pyx_codeobj__230);
  Py_VISIT(traverse_module_state->__pyx_codeobj__231);
  Py_VISIT(traverse_module_state->__pyx_codeobj__232);
  Py_VISIT(traverse_module_state->__pyx_codeobj__233);
  Py_VISIT(traverse_module_state->__pyx_codeobj__234);
  Py_VISIT(traverse_module_state->__pyx_codeobj__235);
  Py_VISIT(traverse_module_state->__pyx_codeobj__236);
  Py_VISIT(traverse_module_state->__pyx_codeobj__237);
  Py_VISIT(traverse_module_state->__pyx_codeobj__238);
  Py_VISIT(traverse_module_state->__pyx_codeobj__239);
  Py_VISIT(traverse_module_state->__pyx_codeobj__240);
  Py_VISIT(traverse_module_state->__pyx_codeobj__241);
  Py_VISIT(traverse_module_state->__pyx_codeobj__242);
  Py_VISIT(traverse_module_state->__pyx_codeobj__243);
  Py_VISIT(traverse_module_state->__pyx_codeobj__244);
  Py_VISIT(traverse_module_state->__pyx_codeobj__245);
  Py_VISIT(traverse_module_state->__pyx_codeobj__246);
  Py_VISIT(traverse_module_state->__pyx_codeobj__247);
  Py_VISIT(traverse_module_state->__pyx_codeobj__248);
  Py_VISIT(traverse_module_state->__pyx_codeobj__249);
  Py_VISIT(traverse_module_state->__pyx_codeobj__250);
  Py_VISIT(traverse_module_state->__pyx_codeobj__251);
  Py_VISIT(traverse_module_state->__pyx_codeobj__252);
  Py_VISIT(traverse_module_state->__pyx_codeobj__254);
  Py_VISIT(traverse_module_state->__pyx_codeobj__255);
  Py_VISIT(traverse_module_state->__pyx_codeobj__256);
  Py_VISIT(traverse_module_state->__pyx_codeobj__257);
  Py_VISIT(traverse_module_state->__pyx_codeobj__259);
  Py_VISIT(traverse_module_state->__pyx_codeobj__261);
  Py_VISIT(traverse_module_state->__pyx_codeobj__262);
  Py_VISIT(traverse_module_state->__pyx_codeobj__263);
  Py_VISIT(traverse_module_state->__pyx_codeobj__265);
  Py_VISIT(traverse_module_state->__pyx_codeobj__266);
  Py_VISIT(traverse_module_state->__pyx_codeobj__267);
  Py_VISIT(traverse_module_state->__pyx_codeobj__268);
  Py_VISIT(traverse_module_state->__pyx_codeobj__269);
  Py_VISIT(traverse_module_state->__pyx_codeobj__270);
  Py_VISIT(traverse_module_state->__pyx_codeobj__272);
  Py_VISIT(traverse_module_state->__pyx_codeobj__273);
  Py_VISIT(traverse_module_state->__pyx_codeobj__274);
  Py_VISIT(traverse_module_state->__pyx_codeobj__275);
  Py_VISIT(traverse_module_state->__pyx_codeobj__276);
  Py_VISIT(traverse_module_state->__pyx_codeobj__277);
  Py_VISIT(traverse_module_state->__pyx_codeobj__278);
  Py_VISIT(traverse_module_state->__pyx_codeobj__279);
  Py_VISIT(traverse_module_state->__pyx_codeobj__280);
  Py_VISIT(traverse_module_state->__pyx_codeobj__281);
  Py_VISIT(traverse_module_state->__pyx_codeobj__282);
  Py_VISIT(traverse_module_state->__pyx_codeobj__283);
  Py_VISIT(traverse_module_state->__pyx_codeobj__284);
  Py_VISIT(traverse_module_state->__pyx_codeobj__285);
  Py_VISIT(traverse_module_state->__pyx_codeobj__286);
  Py_VISIT(traverse_module_state->__pyx_codeobj__287);
  Py_VISIT(traverse_module_state->__pyx_codeobj__288);
  Py_VISIT(traverse_module_state->__pyx_codeobj__289);
  Py_VISIT(traverse_module_state->__pyx_codeobj__290);
  Py_VISIT(traverse_module_state->__pyx_codeobj__291);
  Py_VISIT(traverse_module_state->__pyx_codeobj__292);
  Py_VISIT(traverse_module_state->__pyx_codeobj__293);
  Py_VISIT(traverse_module_state->__pyx_codeobj__295);
  Py_VISIT(traverse_module_state->__pyx_codeobj__296);
  Py_VISIT(traverse_module_state->__pyx_codeobj__298);
  Py_VISIT(traverse_module_state->__pyx_codeobj__299);
  Py_VISIT(traverse_module_state->__pyx_codeobj__301);
  Py_VISIT(traverse_module_state->__pyx_codeobj__302);
  Py_VISIT(traverse_module_state->__pyx_codeobj__304);
  Py_VISIT(traverse_module_state->__pyx_codeobj__305);
  Py_VISIT(traverse_module_state->__pyx_codeobj__307);
  Py_VISIT(traverse_module_state->__pyx_codeobj__309);
  Py_VISIT(traverse_module_state->__pyx_codeobj__310);
  Py_VISIT(traverse_module_state->__pyx_codeobj__311);
  Py_VISIT(traverse_module_state->__pyx_codeobj__312);
  Py_VISIT(traverse_module_state->__pyx_codeobj__314);
  Py_VISIT(traverse_module_state->__pyx_codeobj__315);
  Py_VISIT(traverse_module_state->__pyx_codeobj__316);
  Py_VISIT(traverse_module_state->__pyx_codeobj__317);
  Py_VISIT(traverse_module_state->__pyx_codeobj__318);
  Py_VISIT(traverse_module_state->__pyx_codeobj__319);
  Py_VISIT(traverse_module_state->__pyx_codeobj__320);
  Py_VISIT(traverse_module_state->__pyx_codeobj__322);
  Py_VISIT(traverse_module_state->__pyx_codeobj__324);
  Py_VISIT(traverse_module_state->__pyx_codeobj__326);
  Py_VISIT(traverse_module_state->__pyx_codeobj__328);
  Py_VISIT(traverse_module_state->__pyx_codeobj__330);
  Py_VISIT(traverse_module_state->__pyx_codeobj__333);
  Py_VISIT(traverse_module_state->__pyx_codeobj__336);
  Py_VISIT(traverse_module_state->__pyx_codeobj__338);
  Py_VISIT(traverse_module_state->__pyx_codeobj__339);
  Py_VISIT(traverse_module_state->__pyx_codeobj__341);
  Py_VISIT(traverse_module_state->__pyx_codeobj__343);
  Py_VISIT(traverse_module_state->__pyx_codeobj__345);
  Py_VISIT(traverse_module_state->__pyx_codeobj__348);
  Py_VISIT(traverse_module_state->__pyx_codeobj__350);
  Py_VISIT(traverse_module_state->__pyx_codeobj__351);
  Py_VISIT(traverse_module_state->__pyx_codeobj__352);
  Py_VISIT(traverse_module_state->__pyx_codeobj__353);
  Py_VISIT(traverse_module_state->__pyx_codeobj__354);
  Py_VISIT(traverse_module_state->__pyx_codeobj__355);
  Py_VISIT(traverse_module_state->__pyx_codeobj__356);
  Py_VISIT(traverse_module_state->__pyx_codeobj__358);
  Py_VISIT(traverse_module_state->__pyx_codeobj__359);
  Py_VISIT(traverse_module_state->__pyx_codeobj__360);
  Py_VISIT(traverse_module_state->__pyx_codeobj__361);
  Py_VISIT(traverse_module_state->__pyx_codeobj__362);
  Py_VISIT(traverse_module_state->__pyx_codeobj__364);
  Py_VISIT(traverse_module_state->__pyx_codeobj__365);
  Py_VISIT(traverse_module_state->__pyx_codeobj__366);
  Py_VISIT(traverse_module_state->__pyx_codeobj__368);
  Py_VISIT(traverse_module_state->__pyx_codeobj__369);
  Py_VISIT(traverse_module_state->__pyx_codeobj__370);
  Py_VISIT(traverse_module_state->__pyx_codeobj__371);
  Py_VISIT(traverse_module_state->__pyx_codeobj__372);
  Py_VISIT(traverse_module_state->__pyx_codeobj__373);
  Py_VISIT(traverse_module_state->__pyx_codeobj__374);
  Py_VISIT(traverse_module_state->__pyx_codeobj__375);
  Py_VISIT(traverse_module_state->__pyx_codeobj__376);
  Py_VISIT(traverse_module_state->__pyx_codeobj__378);
  Py_VISIT(traverse_module_state->__pyx_codeobj__379);
  Py_VISIT(traverse_module_state->__pyx_codeobj__380);
  Py_VISIT(traverse_module_state->__pyx_codeobj__382);
  Py_VISIT(traverse_module_state->__pyx_codeobj__383);
  Py_VISIT(traverse_module_state->__pyx_codeobj__384);
  Py_VISIT(traverse_module_state->__pyx_codeobj__385);
  Py_VISIT(traverse_module_state->__pyx_codeobj__386);
  Py_VISIT(traverse_module_state->__pyx_codeobj__387);
  Py_VISIT(traverse_module_state->__pyx_codeobj__388);
  Py_VISIT(traverse_module_state->__pyx_codeobj__389);
  Py_VISIT(traverse_module_state->__pyx_codeobj__390);
  Py_VISIT(traverse_module_state->__pyx_codeobj__391);
  Py_VISIT(traverse_module_state->__pyx_codeobj__392);
  Py_VISIT(traverse_module_state->__pyx_codeobj__393);
  Py_VISIT(traverse_module_state->__pyx_codeobj__394);
  Py_VISIT(traverse_module_state->__pyx_codeobj__395);
  Py_VISIT(traverse_module_state->__pyx_codeobj__397);
  Py_VISIT(traverse_module_state->__pyx_codeobj__398);
  Py_VISIT(traverse_module_state->__pyx_codeobj__399);
  Py_VISIT(traverse_module_state->__pyx_codeobj__400);
  Py_VISIT(traverse_module_state->__pyx_codeobj__401);
  Py_VISIT(traverse_module_state->__pyx_codeobj__402);
  Py_VISIT(traverse_module_state->__pyx_codeobj__403);
  Py_VISIT(traverse_module_state->__pyx_codeobj__404);
  Py_VISIT(traverse_module_state->__pyx_codeobj__405);
  Py_VISIT(traverse_module_state->__pyx_codeobj__406);
  Py_VISIT(traverse_module_state->__pyx_codeobj__407);
  Py_VISIT(traverse_module_state->__pyx_codeobj__408);
  Py_VISIT(traverse_module_state->__pyx_codeobj__409);
  Py_VISIT(traverse_module_state->__pyx_codeobj__410);
  Py_VISIT(traverse_module_state->__pyx_codeobj__411);
  Py_VISIT(traverse_module_state->__pyx_codeobj__412);
  Py_VISIT(traverse_module_state->__pyx_codeobj__413);
  Py_VISIT(traverse_module_state->__pyx_codeobj__414);
  Py_VISIT(traverse_module_state->__pyx_codeobj__415);
  Py_VISIT(traverse_module_state->__pyx_codeobj__416);
  Py_VISIT(traverse_module_state->__pyx_codeobj__417);
  Py_VISIT(traverse_module_state->__pyx_codeobj__419);
  Py_VISIT(traverse_module_state->__pyx_codeobj__421);
  Py_VISIT(traverse_module_state->__pyx_codeobj__423);
  Py_VISIT(traverse_module_state->__pyx_codeobj__426);
  Py_VISIT(traverse_module_state->__pyx_codeobj__428);
  Py_VISIT(traverse_module_state->__pyx_codeobj__429);
  Py_VISIT(traverse_module_state->__pyx_codeobj__430);
  Py_VISIT(traverse_module_state->__pyx_codeobj__431);
  Py_VISIT(traverse_module_state->__pyx_codeobj__432);
  Py_VISIT(traverse_module_state->__pyx_codeobj__433);
  Py_VISIT(traverse_module_state->__pyx_codeobj__434);
  Py_VISIT(traverse_module_state->__pyx_codeobj__435);
  Py_VISIT(traverse_module_state->__pyx_codeobj__436);
  Py_VISIT(traverse_module_state->__pyx_codeobj__437);
  Py_VISIT(traverse_module_state->__pyx_codeobj__438);
  Py_VISIT(traverse_module_state->__pyx_codeobj__439);
  Py_VISIT(traverse_module_state->__pyx_codeobj__440);
  Py_VISIT(traverse_module_state->__pyx_codeobj__441);
  Py_VISIT(traverse_module_state->__pyx_codeobj__442);
  Py_VISIT(traverse_module_state->__pyx_codeobj__443);
  Py_VISIT(traverse_module_state->__pyx_codeobj__444);
  Py_VISIT(traverse_module_state->__pyx_codeobj__445);
  Py_VISIT(traverse_module_state->__pyx_codeobj__446);
  Py_VISIT(traverse_module_state->__pyx_codeobj__447);
  Py_VISIT(traverse_module_state->__pyx_codeobj__448);
  Py_VISIT(traverse_module_state->__pyx_codeobj__449);
  Py_VISIT(traverse_module_state->__pyx_codeobj__450);
  Py_VISIT(traverse_module_state->__pyx_codeobj__451);
  Py_VISIT(traverse_module_state->__pyx_codeobj__452);
  Py_VISIT(traverse_module_state->__pyx_codeobj__453);
  Py_VISIT(traverse_module_state->__pyx_codeobj__454);
  Py_VISIT(traverse_module_state->__pyx_codeobj__455);
  Py_VISIT(traverse_module_state->__pyx_codeobj__456);
  Py_VISIT(traverse_module_state->__pyx_codeobj__457);
  Py_VISIT(traverse_module_state->__pyx_codeobj__458);
  Py_VISIT(traverse_module_state->__pyx_codeobj__459);
  Py_VISIT(traverse_module_state->__pyx_codeobj__460);
  Py_VISIT(traverse_module_state->__pyx_codeobj__461);
  Py_VISIT(traverse_module_state->__pyx_codeobj__462);
  Py_VISIT(traverse_module_state->__pyx_codeobj__463);
  Py_VISIT(traverse_module_state->__pyx_codeobj__464);
  Py_VISIT(traverse_module_state->__pyx_codeobj__465);
  Py_VISIT(traverse_module_state->__pyx_codeobj__466);
  Py_VISIT(traverse_module_state->__pyx_codeobj__467);
  Py_VISIT(traverse_module_state->__pyx_codeobj__468);
  Py_VISIT(traverse_module_state->__pyx_codeobj__469);
  Py_VISIT(traverse_module_state->__pyx_codeobj__470);
  Py_VISIT(traverse_module_state->__pyx_codeobj__471);
  Py_VISIT(traverse_module_state->__pyx_codeobj__472);
  Py_VISIT(traverse_module_state->__pyx_codeobj__473);
  Py_VISIT(traverse_module_state->__pyx_codeobj__474);
  Py_VISIT(traverse_module_state->__pyx_codeobj__475);
  Py_VISIT(traverse_module_state->__pyx_codeobj__476);
  Py_VISIT(traverse_module_state->__pyx_codeobj__477);
  Py_VISIT(traverse_module_state->__pyx_codeobj__478);
  Py_VISIT(traverse_module_state->__pyx_codeobj__479);
  Py_VISIT(traverse_module_state->__pyx_codeobj__480);
  Py_VISIT(traverse_module_state->__pyx_codeobj__481);
  Py_VISIT(traverse_module_state->__pyx_codeobj__482);
  Py_VISIT(traverse_module_state->__pyx_codeobj__483);
  Py_VISIT(traverse_module_state->__pyx_codeobj__484);
  Py_VISIT(traverse_module_state->__pyx_codeobj__485);
  Py_VISIT(traverse_module_state->__pyx_codeobj__486);
  Py_VISIT(traverse_module_state->__pyx_codeobj__487);
  Py_VISIT(traverse_module_state->__pyx_codeobj__488);
  Py_VISIT(traverse_module_state->__pyx_codeobj__489);
  Py_VISIT(traverse_module_state->__pyx_codeobj__490);
  Py_VISIT(traverse_module_state->__pyx_codeobj__491);
  Py_VISIT(traverse_module_state->__pyx_codeobj__492);
  Py_VISIT(traverse_module_state->__pyx_codeobj__493);
  Py_VISIT(traverse_module_state->__pyx_codeobj__494);
  Py_VISIT(traverse_module_state->__pyx_codeobj__495);
  Py_VISIT(traverse_module_state->__pyx_codeobj__496);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4bool_bool __pyx_mstate_global->__pyx_ptype_7cpython_4bool_bool
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_7complex_complex __pyx_mstate_global->__pyx_ptype_7cpython_7complex_complex
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_5array_array __pyx_mstate_global->__pyx_ptype_7cpython_5array_array
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_5cavro_Options __pyx_mstate_global->__pyx_type_5cavro_Options
#define __pyx_type_5cavro__Writer __pyx_mstate_global->__pyx_type_5cavro__Writer
#define __pyx_type_5cavro__Reader __pyx_mstate_global->__pyx_type_5cavro__Reader
#define __pyx_type_5cavro_FileReader __pyx_mstate_global->__pyx_type_5cavro_FileReader
#define __pyx_type_5cavro_FileWriter __pyx_mstate_global->__pyx_type_5cavro_FileWriter
#define __pyx_type_5cavro_MemoryWriter __pyx_mstate_global->__pyx_type_5cavro_MemoryWriter
#define __pyx_type_5cavro_MemoryReader __pyx_mstate_global->__pyx_type_5cavro_MemoryReader
#define __pyx_type_5cavro_Rabin __pyx_mstate_global->__pyx_type_5cavro_Rabin
#define __pyx_type_5cavro_Record __pyx_mstate_global->__pyx_type_5cavro_Record
#define __pyx_type_5cavro_RecordField __pyx_mstate_global->__pyx_type_5cavro_RecordField
#define __pyx_type_5cavro__FieldAccessor __pyx_mstate_global->__pyx_type_5cavro__FieldAccessor
#define __pyx_type_5cavro__Sentinel __pyx_mstate_global->__pyx_type_5cavro__Sentinel
#define __pyx_type_5cavro__CanonicalForm __pyx_mstate_global->__pyx_type_5cavro__CanonicalForm
#define __pyx_type_5cavro_AvroType __pyx_mstate_global->__pyx_type_5cavro_AvroType
#define __pyx_type_5cavro_UnionType __pyx_mstate_global->__pyx_type_5cavro_UnionType
#define __pyx_type_5cavro_StringType __pyx_mstate_global->__pyx_type_5cavro_StringType
#define __pyx_type_5cavro_BytesType __pyx_mstate_global->__pyx_type_5cavro_BytesType
#define __pyx_type_5cavro__PlaceholderType __pyx_mstate_global->__pyx_type_5cavro__PlaceholderType
#define __pyx_type_5cavro_DoubleType __pyx_mstate_global->__pyx_type_5cavro_DoubleType
#define __pyx_type_5cavro_FloatType __pyx_mstate_global->__pyx_type_5cavro_FloatType
#define __pyx_type_5cavro_LongType __pyx_mstate_global->__pyx_type_5cavro_LongType
#define __pyx_type_5cavro_IntType __pyx_mstate_global->__pyx_type_5cavro_IntType
#define __pyx_type_5cavro_BoolType __pyx_mstate_global->__pyx_type_5cavro_BoolType
#define __pyx_type_5cavro_NullType __pyx_mstate_global->__pyx_type_5cavro_NullType
#define __pyx_type_5cavro_MapType __pyx_mstate_global->__pyx_type_5cavro_MapType
#define __pyx_type_5cavro_ArrayType __pyx_mstate_global->__pyx_type_5cavro_ArrayType
#define __pyx_type_5cavro__NamedType __pyx_mstate_global->__pyx_type_5cavro__NamedType
#define __pyx_type_5cavro_FixedType __pyx_mstate_global->__pyx_type_5cavro_FixedType
#define __pyx_type_5cavro_RecordType __pyx_mstate_global->__pyx_type_5cavro_RecordType
#define __pyx_type_5cavro_PromotingRecordType __pyx_mstate_global->__pyx_type_5cavro_PromotingRecordType
#define __pyx_type_5cavro_EnumType __pyx_mstate_global->__pyx_type_5cavro_EnumType
#define __pyx_type_5cavro_PromotingEnumType __pyx_mstate_global->__pyx_type_5cavro_PromotingEnumType
#define __pyx_type_5cavro_Schema __pyx_mstate_global->__pyx_type_5cavro_Schema
#define __pyx_type_5cavro_ResolvedSchema __pyx_mstate_global->__pyx_type_5cavro_ResolvedSchema
#define __pyx_type_5cavro_ValueAdapter __pyx_mstate_global->__pyx_type_5cavro_ValueAdapter
#define __pyx_type_5cavro_LogicalType __pyx_mstate_global->__pyx_type_5cavro_LogicalType
#define __pyx_type_5cavro_CustomLogicalType __pyx_mstate_global->__pyx_type_5cavro_CustomLogicalType
#define __pyx_type_5cavro_DecimalType __pyx_mstate_global->__pyx_type_5cavro_DecimalType
#define __pyx_type_5cavro_UUIDBase __pyx_mstate_global->__pyx_type_5cavro_UUIDBase
#define __pyx_type_5cavro_UUIDStringType __pyx_mstate_global->__pyx_type_5cavro_UUIDStringType
#define __pyx_type_5cavro_UUIDFixedType __pyx_mstate_global->__pyx_type_5cavro_UUIDFixedType
#define __pyx_type_5cavro_Date __pyx_mstate_global->__pyx_type_5cavro_Date
#define __pyx_type_5cavro_TimeMillis __pyx_mstate_global->__pyx_type_5cavro_TimeMillis
#define __pyx_type_5cavro_TimeMicros __pyx_mstate_global->__pyx_type_5cavro_TimeMicros
#define __pyx_type_5cavro_TimestampMillis __pyx_mstate_global->__pyx_type_5cavro_TimestampMillis
#define __pyx_type_5cavro_TimestampMicros __pyx_mstate_global->__pyx_type_5cavro_TimestampMicros
#define __pyx_type_5cavro_Codec __pyx_mstate_global->__pyx_type_5cavro_Codec
#define __pyx_type_5cavro__SnappyCodec __pyx_mstate_global->__pyx_type_5cavro__SnappyCodec
#define __pyx_type_5cavro__NullCodec __pyx_mstate_global->__pyx_type_5cavro__NullCodec
#define __pyx_type_5cavro__DeflateCodec __pyx_mstate_global->__pyx_type_5cavro__DeflateCodec
#define __pyx_type_5cavro___Bzip2Codec __pyx_mstate_global->__pyx_type_5cavro___Bzip2Codec
#define __pyx_type_5cavro__LzmaCodec __pyx_mstate_global->__pyx_type_5cavro__LzmaCodec
#define __pyx_type_5cavro__ZStandardCodec __pyx_mstate_global->__pyx_type_5cavro__ZStandardCodec
#define __pyx_type_5cavro__Lz4Codec __pyx_mstate_global->__pyx_type_5cavro__Lz4Codec
#define __pyx_type_5cavro_ContainerReader __pyx_mstate_global->__pyx_type_5cavro_ContainerReader
#define __pyx_type_5cavro_ContainerWriter __pyx_mstate_global->__pyx_type_5cavro_ContainerWriter
#define __pyx_type_5cavro_PromoteToFloat __pyx_mstate_global->__pyx_type_5cavro_PromoteToFloat
#define __pyx_type_5cavro_PromoteBytesToString __pyx_mstate_global->__pyx_type_5cavro_PromoteBytesToString
#define __pyx_type_5cavro_PromoteStringToBytes __pyx_mstate_global->__pyx_type_5cavro_PromoteStringToBytes
#define __pyx_type_5cavro_CannotPromote __pyx_mstate_global->__pyx_type_5cavro_CannotPromote
#define __pyx_type_5cavro___pyx_scope_struct__genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct__genexpr
#define __pyx_type_5cavro___pyx_scope_struct_1_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_1_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_2_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_2_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_3_walk_types __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_3_walk_types
#define __pyx_type_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer
#define __pyx_type_5cavro___pyx_scope_struct_5_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_5_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_6_walk_types __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_6_walk_types
#define __pyx_type_5cavro___pyx_scope_struct_7_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_7_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_8___init__ __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_8___init__
#define __pyx_type_5cavro___pyx_scope_struct_9_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_9_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_10_walk_types __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_10_walk_types
#define __pyx_type_5cavro___pyx_scope_struct_11___init__ __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_11___init__
#define __pyx_type_5cavro___pyx_scope_struct_12_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_12_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_13_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_13_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_14_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_14_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_15_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_15_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_16_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_16_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_17_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_17_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_18_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_18_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_19_walk_types __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_19_walk_types
#define __pyx_type_5cavro___pyx_scope_struct_20_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_20_genexpr
#define __pyx_type_5cavro___pyx_scope_struct_21_walk_types __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_21_walk_types
#define __pyx_type_5cavro___pyx_scope_struct_22_get_namespaced_aliases __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_22_get_namespaced_aliases
#define __pyx_type_5cavro___pyx_scope_struct_23_genexpr __pyx_mstate_global->__pyx_type_5cavro___pyx_scope_struct_23_genexpr
#define __pyx_type___pyx_array __pyx_mstate_global->__pyx_type___pyx_array
#define __pyx_type___pyx_MemviewEnum __pyx_mstate_global->__pyx_type___pyx_MemviewEnum
#define __pyx_type___pyx_memoryview __pyx_mstate_global->__pyx_type___pyx_memoryview
#define __pyx_type___pyx_memoryviewslice __pyx_mstate_global->__pyx_type___pyx_memoryviewslice
#endif
#define __pyx_ptype_5cavro_Options __pyx_mstate_global->__pyx_ptype_5cavro_Options
#define __pyx_ptype_5cavro__Writer __pyx_mstate_global->__pyx_ptype_5cavro__Writer
#define __pyx_ptype_5cavro__Reader __pyx_mstate_global->__pyx_ptype_5cavro__Reader
#define __pyx_ptype_5cavro_FileReader __pyx_mstate_global->__pyx_ptype_5cavro_FileReader
#define __pyx_ptype_5cavro_FileWriter __pyx_mstate_global->__pyx_ptype_5cavro_FileWriter
#define __pyx_ptype_5cavro_MemoryWriter __pyx_mstate_global->__pyx_ptype_5cavro_MemoryWriter
#define __pyx_ptype_5cavro_MemoryReader __pyx_mstate_global->__pyx_ptype_5cavro_MemoryReader
#define __pyx_ptype_5cavro_Rabin __pyx_mstate_global->__pyx_ptype_5cavro_Rabin
#define __pyx_ptype_5cavro_Record __pyx_mstate_global->__pyx_ptype_5cavro_Record
#define __pyx_ptype_5cavro_RecordField __pyx_mstate_global->__pyx_ptype_5cavro_RecordField
#define __pyx_ptype_5cavro__FieldAccessor __pyx_mstate_global->__pyx_ptype_5cavro__FieldAccessor
#define __pyx_ptype_5cavro__Sentinel __pyx_mstate_global->__pyx_ptype_5cavro__Sentinel
#define __pyx_ptype_5cavro__CanonicalForm __pyx_mstate_global->__pyx_ptype_5cavro__CanonicalForm
#define __pyx_ptype_5cavro_AvroType __pyx_mstate_global->__pyx_ptype_5cavro_AvroType
#define __pyx_ptype_5cavro_UnionType __pyx_mstate_global->__pyx_ptype_5cavro_UnionType
#define __pyx_ptype_5cavro_StringType __pyx_mstate_global->__pyx_ptype_5cavro_StringType
#define __pyx_ptype_5cavro_BytesType __pyx_mstate_global->__pyx_ptype_5cavro_BytesType
#define __pyx_ptype_5cavro__PlaceholderType __pyx_mstate_global->__pyx_ptype_5cavro__PlaceholderType
#define __pyx_ptype_5cavro_DoubleType __pyx_mstate_global->__pyx_ptype_5cavro_DoubleType
#define __pyx_ptype_5cavro_FloatType __pyx_mstate_global->__pyx_ptype_5cavro_FloatType
#define __pyx_ptype_5cavro_LongType __pyx_mstate_global->__pyx_ptype_5cavro_LongType
#define __pyx_ptype_5cavro_IntType __pyx_mstate_global->__pyx_ptype_5cavro_IntType
#define __pyx_ptype_5cavro_BoolType __pyx_mstate_global->__pyx_ptype_5cavro_BoolType
#define __pyx_ptype_5cavro_NullType __pyx_mstate_global->__pyx_ptype_5cavro_NullType
#define __pyx_ptype_5cavro_MapType __pyx_mstate_global->__pyx_ptype_5cavro_MapType
#define __pyx_ptype_5cavro_ArrayType __pyx_mstate_global->__pyx_ptype_5cavro_ArrayType
#define __pyx_ptype_5cavro__NamedType __pyx_mstate_global->__pyx_ptype_5cavro__NamedType
#define __pyx_ptype_5cavro_FixedType __pyx_mstate_global->__pyx_ptype_5cavro_FixedType
#define __pyx_ptype_5cavro_RecordType __pyx_mstate_global->__pyx_ptype_5cavro_RecordType
#define __pyx_ptype_5cavro_PromotingRecordType __pyx_mstate_global->__pyx_ptype_5cavro_PromotingRecordType
#define __pyx_ptype_5cavro_EnumType __pyx_mstate_global->__pyx_ptype_5cavro_EnumType
#define __pyx_ptype_5cavro_PromotingEnumType __pyx_mstate_global->__pyx_ptype_5cavro_PromotingEnumType
#define __pyx_ptype_5cavro_Schema __pyx_mstate_global->__pyx_ptype_5cavro_Schema
#define __pyx_ptype_5cavro_ResolvedSchema __pyx_mstate_global->__pyx_ptype_5cavro_ResolvedSchema
#define __pyx_ptype_5cavro_ValueAdapter __pyx_mstate_global->__pyx_ptype_5cavro_ValueAdapter
#define __pyx_ptype_5cavro_LogicalType __pyx_mstate_global->__pyx_ptype_5cavro_LogicalType
#define __pyx_ptype_5cavro_CustomLogicalType __pyx_mstate_global->__pyx_ptype_5cavro_CustomLogicalType
#define __pyx_ptype_5cavro_DecimalType __pyx_mstate_global->__pyx_ptype_5cavro_DecimalType
#define __pyx_ptype_5cavro_UUIDBase __pyx_mstate_global->__pyx_ptype_5cavro_UUIDBase
#define __pyx_ptype_5cavro_UUIDStringType __pyx_mstate_global->__pyx_ptype_5cavro_UUIDStringType
#define __pyx_ptype_5cavro_UUIDFixedType __pyx_mstate_global->__pyx_ptype_5cavro_UUIDFixedType
#define __pyx_ptype_5cavro_Date __pyx_mstate_global->__pyx_ptype_5cavro_Date
#define __pyx_ptype_5cavro_TimeMillis __pyx_mstate_global->__pyx_ptype_5cavro_TimeMillis
#define __pyx_ptype_5cavro_TimeMicros __pyx_mstate_global->__pyx_ptype_5cavro_TimeMicros
#define __pyx_ptype_5cavro_TimestampMillis __pyx_mstate_global->__pyx_ptype_5cavro_TimestampMillis
#define __pyx_ptype_5cavro_TimestampMicros __pyx_mstate_global->__pyx_ptype_5cavro_TimestampMicros
#define __pyx_ptype_5cavro_Codec __pyx_mstate_global->__pyx_ptype_5cavro_Codec
#define __pyx_ptype_5cavro__SnappyCodec __pyx_mstate_global->__pyx_ptype_5cavro__SnappyCodec
#define __pyx_ptype_5cavro__NullCodec __pyx_mstate_global->__pyx_ptype_5cavro__NullCodec
#define __pyx_ptype_5cavro__DeflateCodec __pyx_mstate_global->__pyx_ptype_5cavro__DeflateCodec
#define __pyx_ptype_5cavro___Bzip2Codec __pyx_mstate_global->__pyx_ptype_5cavro___Bzip2Codec
#define __pyx_ptype_5cavro__LzmaCodec __pyx_mstate_global->__pyx_ptype_5cavro__LzmaCodec
#define __pyx_ptype_5cavro__ZStandardCodec __pyx_mstate_global->__pyx_ptype_5cavro__ZStandardCodec
#define __pyx_ptype_5cavro__Lz4Codec __pyx_mstate_global->__pyx_ptype_5cavro__Lz4Codec
#define __pyx_ptype_5cavro_ContainerReader __pyx_mstate_global->__pyx_ptype_5cavro_ContainerReader
#define __pyx_ptype_5cavro_ContainerWriter __pyx_mstate_global->__pyx_ptype_5cavro_ContainerWriter
#define __pyx_ptype_5cavro_PromoteToFloat __pyx_mstate_global->__pyx_ptype_5cavro_PromoteToFloat
#define __pyx_ptype_5cavro_PromoteBytesToString __pyx_mstate_global->__pyx_ptype_5cavro_PromoteBytesToString
#define __pyx_ptype_5cavro_PromoteStringToBytes __pyx_mstate_global->__pyx_ptype_5cavro_PromoteStringToBytes
#define __pyx_ptype_5cavro_CannotPromote __pyx_mstate_global->__pyx_ptype_5cavro_CannotPromote
#define __pyx_ptype_5cavro___pyx_scope_struct__genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct__genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_1_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_1_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_2_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_2_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_3_walk_types __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_3_walk_types
#define __pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer
#define __pyx_ptype_5cavro___pyx_scope_struct_5_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_5_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_6_walk_types __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_6_walk_types
#define __pyx_ptype_5cavro___pyx_scope_struct_7_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_7_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_8___init__ __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_8___init__
#define __pyx_ptype_5cavro___pyx_scope_struct_9_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_9_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_10_walk_types __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_10_walk_types
#define __pyx_ptype_5cavro___pyx_scope_struct_11___init__ __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_11___init__
#define __pyx_ptype_5cavro___pyx_scope_struct_12_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_12_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_13_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_13_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_14_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_14_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_15_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_15_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_16_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_16_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_17_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_17_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_18_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_18_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_19_walk_types __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_19_walk_types
#define __pyx_ptype_5cavro___pyx_scope_struct_20_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_20_genexpr
#define __pyx_ptype_5cavro___pyx_scope_struct_21_walk_types __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_21_walk_types
#define __pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases
#define __pyx_ptype_5cavro___pyx_scope_struct_23_genexpr __pyx_mstate_global->__pyx_ptype_5cavro___pyx_scope_struct_23_genexpr
#define __pyx_array_type __pyx_mstate_global->__pyx_array_type
#define __pyx_MemviewEnum_type __pyx_mstate_global->__pyx_MemviewEnum_type
#define __pyx_memoryview_type __pyx_mstate_global->__pyx_memoryview_type
#define __pyx_memoryviewslice_type __pyx_mstate_global->__pyx_memoryviewslice_type
#define __pyx_kp_u_ __pyx_mstate_global->__pyx_kp_u_
#define __pyx_kp_u_0 __pyx_mstate_global->__pyx_kp_u_0
#define __pyx_kp_u_1 __pyx_mstate_global->__pyx_kp_u_1
#define __pyx_kp_u_1_0_0 __pyx_mstate_global->__pyx_kp_u_1_0_0
#define __pyx_n_s_AMBIGUOUS_TYPES __pyx_mstate_global->__pyx_n_s_AMBIGUOUS_TYPES
#define __pyx_n_s_ASC __pyx_mstate_global->__pyx_n_s_ASC
#define __pyx_n_s_ASCII __pyx_mstate_global->__pyx_n_s_ASCII
#define __pyx_kp_u_A_Za_z__A_Za_z0_9 __pyx_mstate_global->__pyx_kp_u_A_Za_z__A_Za_z0_9
#define __pyx_kp_s_A_record_contains_multiple_fiel __pyx_mstate_global->__pyx_kp_s_A_record_contains_multiple_fiel
#define __pyx_kp_s_A_requested_codec_or_codec_in_a __pyx_mstate_global->__pyx_kp_s_A_requested_codec_or_codec_in_a
#define __pyx_kp_s_A_schema_cannot_be_promoted_to __pyx_mstate_global->__pyx_kp_s_A_schema_cannot_be_promoted_to
#define __pyx_kp_s_A_value_is_invalid_for_a_given __pyx_mstate_global->__pyx_kp_s_A_value_is_invalid_for_a_given
#define __pyx_kp_u_Aliases_must_be_a_list_tuple_set __pyx_mstate_global->__pyx_kp_u_Aliases_must_be_a_list_tuple_set
#define __pyx_kp_s_All_dimensions_preceding_dimensi __pyx_mstate_global->__pyx_kp_s_All_dimensions_preceding_dimensi
#define __pyx_kp_s_An_unknown_hash_method_was_requ __pyx_mstate_global->__pyx_kp_s_An_unknown_hash_method_was_requ
#define __pyx_n_s_ArrayType __pyx_mstate_global->__pyx_n_s_ArrayType
#define __pyx_n_s_ArrayType___reduce_cython __pyx_mstate_global->__pyx_n_s_ArrayType___reduce_cython
#define __pyx_n_s_ArrayType___setstate_cython __pyx_mstate_global->__pyx_n_s_ArrayType___setstate_cython
#define __pyx_n_s_ArrayType__get_schema_extra __pyx_mstate_global->__pyx_n_s_ArrayType__get_schema_extra
#define __pyx_n_s_ArrayType_convert_value __pyx_mstate_global->__pyx_n_s_ArrayType_convert_value
#define __pyx_n_s_ArrayType_copy __pyx_mstate_global->__pyx_n_s_ArrayType_copy
#define __pyx_n_s_ArrayType_walk_types __pyx_mstate_global->__pyx_n_s_ArrayType_walk_types
#define __pyx_kp_s_As_of_cython_3_0_2_dataclasses __pyx_mstate_global->__pyx_kp_s_As_of_cython_3_0_2_dataclasses
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_s_AttributeError __pyx_mstate_global->__pyx_n_s_AttributeError
#define __pyx_n_s_AvroType __pyx_mstate_global->__pyx_n_s_AvroType
#define __pyx_n_s_AvroType___reduce_cython __pyx_mstate_global->__pyx_n_s_AvroType___reduce_cython
#define __pyx_n_s_AvroType___setstate_cython __pyx_mstate_global->__pyx_n_s_AvroType___setstate_cython
#define __pyx_n_s_AvroType__convert_value __pyx_mstate_global->__pyx_n_s_AvroType__convert_value
#define __pyx_n_s_AvroType__get_schema_extra __pyx_mstate_global->__pyx_n_s_AvroType__get_schema_extra
#define __pyx_n_s_AvroType_convert_value __pyx_mstate_global->__pyx_n_s_AvroType_convert_value
#define __pyx_n_s_AvroType_copy __pyx_mstate_global->__pyx_n_s_AvroType_copy
#define __pyx_n_s_AvroType_for_schema __pyx_mstate_global->__pyx_n_s_AvroType_for_schema
#define __pyx_n_s_AvroType_for_source __pyx_mstate_global->__pyx_n_s_AvroType_for_source
#define __pyx_n_s_AvroType_get_schema __pyx_mstate_global->__pyx_n_s_AvroType_get_schema
#define __pyx_n_s_AvroType_walk_types __pyx_mstate_global->__pyx_n_s_AvroType_walk_types
#define __pyx_n_u_B __pyx_mstate_global->__pyx_n_u_B
#define __pyx_n_s_BATCH_SIZE __pyx_mstate_global->__pyx_n_s_BATCH_SIZE
#define __pyx_kp_s_Base_class_for_exceptions_raise __pyx_mstate_global->__pyx_kp_s_Base_class_for_exceptions_raise
#define __pyx_n_s_BoolType __pyx_mstate_global->__pyx_n_s_BoolType
#define __pyx_n_s_BoolType___reduce_cython __pyx_mstate_global->__pyx_n_s_BoolType___reduce_cython
#define __pyx_n_s_BoolType___setstate_cython __pyx_mstate_global->__pyx_n_s_BoolType___setstate_cython
#define __pyx_n_s_BoolType__convert_value __pyx_mstate_global->__pyx_n_s_BoolType__convert_value
#define __pyx_n_s_BoolType__get_schema_extra __pyx_mstate_global->__pyx_n_s_BoolType__get_schema_extra
#define __pyx_n_s_BoolType_copy __pyx_mstate_global->__pyx_n_s_BoolType_copy
#define __pyx_kp_s_Buffer_view_does_not_expose_stri __pyx_mstate_global->__pyx_kp_s_Buffer_view_does_not_expose_stri
#define __pyx_n_s_BytesType __pyx_mstate_global->__pyx_n_s_BytesType
#define __pyx_n_s_BytesType___reduce_cython __pyx_mstate_global->__pyx_n_s_BytesType___reduce_cython
#define __pyx_n_s_BytesType___setstate_cython __pyx_mstate_global->__pyx_n_s_BytesType___setstate_cython
#define __pyx_n_s_BytesType__convert_value __pyx_mstate_global->__pyx_n_s_BytesType__convert_value
#define __pyx_n_s_BytesType__get_schema_extra __pyx_mstate_global->__pyx_n_s_BytesType__get_schema_extra
#define __pyx_n_s_BytesType_copy __pyx_mstate_global->__pyx_n_s_BytesType_copy
#define __pyx_n_s_Bzip2Codec __pyx_mstate_global->__pyx_n_s_Bzip2Codec
#define __pyx_n_s_Bzip2Codec___reduce_cython __pyx_mstate_global->__pyx_n_s_Bzip2Codec___reduce_cython
#define __pyx_n_s_Bzip2Codec___setstate_cython __pyx_mstate_global->__pyx_n_s_Bzip2Codec___setstate_cython
#define __pyx_n_s_CANONICAL_FORM_KEYS __pyx_mstate_global->__pyx_n_s_CANONICAL_FORM_KEYS
#define __pyx_n_s_CODECS __pyx_mstate_global->__pyx_n_s_CODECS
#define __pyx_kp_s_Can_only_create_a_buffer_that_is __pyx_mstate_global->__pyx_kp_s_Can_only_create_a_buffer_that_is
#define __pyx_n_s_CannotPromote __pyx_mstate_global->__pyx_n_s_CannotPromote
#define __pyx_n_s_CannotPromoteError __pyx_mstate_global->__pyx_n_s_CannotPromoteError
#define __pyx_n_s_CannotPromoteError___init __pyx_mstate_global->__pyx_n_s_CannotPromoteError___init
#define __pyx_n_s_CannotPromote___reduce_cython __pyx_mstate_global->__pyx_n_s_CannotPromote___reduce_cython
#define __pyx_n_s_CannotPromote___setstate_cython __pyx_mstate_global->__pyx_n_s_CannotPromote___setstate_cython
#define __pyx_kp_s_Cannot_assign_to_read_only_memor __pyx_mstate_global->__pyx_kp_s_Cannot_assign_to_read_only_memor
#define __pyx_kp_s_Cannot_create_writable_memory_vi __pyx_mstate_global->__pyx_kp_s_Cannot_create_writable_memory_vi
#define __pyx_kp_u_Cannot_index_with_type __pyx_mstate_global->__pyx_kp_u_Cannot_index_with_type
#define __pyx_kp_u_Cannot_promote __pyx_mstate_global->__pyx_kp_u_Cannot_promote
#define __pyx_kp_u_Cannot_read_from __pyx_mstate_global->__pyx_kp_u_Cannot_read_from
#define __pyx_kp_s_Cannot_transpose_memoryview_with __pyx_mstate_global->__pyx_kp_s_Cannot_transpose_memoryview_with
#define __pyx_kp_u_Cannot_write_to __pyx_mstate_global->__pyx_kp_u_Cannot_write_to
#define __pyx_n_s_CanonicalForm __pyx_mstate_global->__pyx_n_s_CanonicalForm
#define __pyx_n_s_CanonicalForm___reduce_cython __pyx_mstate_global->__pyx_n_s_CanonicalForm___reduce_cython
#define __pyx_n_s_CanonicalForm___setstate_cython __pyx_mstate_global->__pyx_n_s_CanonicalForm___setstate_cython
#define __pyx_n_s_CavroException __pyx_mstate_global->__pyx_n_s_CavroException
#define __pyx_n_s_Codec __pyx_mstate_global->__pyx_n_s_Codec
#define __pyx_n_s_CodecUnavailable __pyx_mstate_global->__pyx_n_s_CodecUnavailable
#define __pyx_n_s_Codec___reduce_cython __pyx_mstate_global->__pyx_n_s_Codec___reduce_cython
#define __pyx_n_s_Codec___setstate_cython __pyx_mstate_global->__pyx_n_s_Codec___setstate_cython
#define __pyx_n_s_ContainerReader __pyx_mstate_global->__pyx_n_s_ContainerReader
#define __pyx_n_s_ContainerReader___reduce_cython __pyx_mstate_global->__pyx_n_s_ContainerReader___reduce_cython
#define __pyx_n_s_ContainerReader___setstate_cytho __pyx_mstate_global->__pyx_n_s_ContainerReader___setstate_cytho
#define __pyx_n_s_ContainerReader__read_marker __pyx_mstate_global->__pyx_n_s_ContainerReader__read_marker
#define __pyx_n_s_ContainerReader_next_object __pyx_mstate_global->__pyx_n_s_ContainerReader_next_object
#define __pyx_n_s_ContainerWriter __pyx_mstate_global->__pyx_n_s_ContainerWriter
#define __pyx_n_s_ContainerWriter___enter __pyx_mstate_global->__pyx_n_s_ContainerWriter___enter
#define __pyx_n_s_ContainerWriter___exit __pyx_mstate_global->__pyx_n_s_ContainerWriter___exit
#define __pyx_n_s_ContainerWriter___reduce_cython __pyx_mstate_global->__pyx_n_s_ContainerWriter___reduce_cython
#define __pyx_n_s_ContainerWriter___setstate_cytho __pyx_mstate_global->__pyx_n_s_ContainerWriter___setstate_cytho
#define __pyx_n_s_ContainerWriter_close __pyx_mstate_global->__pyx_n_s_ContainerWriter_close
#define __pyx_n_s_ContainerWriter_flush __pyx_mstate_global->__pyx_n_s_ContainerWriter_flush
#define __pyx_n_s_ContainerWriter_write_many __pyx_mstate_global->__pyx_n_s_ContainerWriter_write_many
#define __pyx_n_s_ContainerWriter_write_one __pyx_mstate_global->__pyx_n_s_ContainerWriter_write_one
#define __pyx_kp_u_Container_is_closed __pyx_mstate_global->__pyx_kp_u_Container_is_closed
#define __pyx_n_s_Context __pyx_mstate_global->__pyx_n_s_Context
#define __pyx_kp_s_Controls_how_union_values_are_d __pyx_mstate_global->__pyx_kp_s_Controls_how_union_values_are_d
#define __pyx_kp_u_Could_not_find_key_type_in_schem __pyx_mstate_global->__pyx_kp_u_Could_not_find_key_type_in_schem
#define __pyx_n_s_CustomLogicalType __pyx_mstate_global->__pyx_n_s_CustomLogicalType
#define __pyx_n_s_CustomLogicalType___reduce_cytho __pyx_mstate_global->__pyx_n_s_CustomLogicalType___reduce_cytho
#define __pyx_n_s_CustomLogicalType___setstate_cyt __pyx_mstate_global->__pyx_n_s_CustomLogicalType___setstate_cyt
#define __pyx_n_s_DEFAULT_OPTIONS __pyx_mstate_global->__pyx_n_s_DEFAULT_OPTIONS
#define __pyx_n_s_DESC __pyx_mstate_global->__pyx_n_s_DESC
#define __pyx_n_s_DataclassParams __pyx_mstate_global->__pyx_n_s_DataclassParams
#define __pyx_n_s_Date __pyx_mstate_global->__pyx_n_s_Date
#define __pyx_n_s_Date___reduce_cython __pyx_mstate_global->__pyx_n_s_Date___reduce_cython
#define __pyx_n_s_Date___setstate_cython __pyx_mstate_global->__pyx_n_s_Date___setstate_cython
#define __pyx_n_s_Date__for_type __pyx_mstate_global->__pyx_n_s_Date__for_type
#define __pyx_n_s_Decimal __pyx_mstate_global->__pyx_n_s_Decimal
#define __pyx_n_s_DecimalType __pyx_mstate_global->__pyx_n_s_DecimalType
#define __pyx_n_s_DecimalType___reduce_cython __pyx_mstate_global->__pyx_n_s_DecimalType___reduce_cython
#define __pyx_n_s_DecimalType___setstate_cython __pyx_mstate_global->__pyx_n_s_DecimalType___setstate_cython
#define __pyx_n_s_DecimalType__for_type __pyx_mstate_global->__pyx_n_s_DecimalType__for_type
#define __pyx_kp_u_Decimal_value __pyx_mstate_global->__pyx_kp_u_Decimal_value
#define __pyx_kp_u_Default_value __pyx_mstate_global->__pyx_kp_u_Default_value
#define __pyx_kp_u_Default_value_2 __pyx_mstate_global->__pyx_kp_u_Default_value_2
#define __pyx_n_s_DeflateCodec __pyx_mstate_global->__pyx_n_s_DeflateCodec
#define __pyx_n_s_DeflateCodec___reduce_cython __pyx_mstate_global->__pyx_n_s_DeflateCodec___reduce_cython
#define __pyx_n_s_DeflateCodec___setstate_cython __pyx_mstate_global->__pyx_n_s_DeflateCodec___setstate_cython
#define __pyx_kp_s_Dimension_d_is_not_direct __pyx_mstate_global->__pyx_kp_s_Dimension_d_is_not_direct
#define __pyx_n_s_DoubleType __pyx_mstate_global->__pyx_n_s_DoubleType
#define __pyx_n_s_DoubleType___reduce_cython __pyx_mstate_global->__pyx_n_s_DoubleType___reduce_cython
#define __pyx_n_s_DoubleType___setstate_cython __pyx_mstate_global->__pyx_n_s_DoubleType___setstate_cython
#define __pyx_n_s_DoubleType__convert_value __pyx_mstate_global->__pyx_n_s_DoubleType__convert_value
#define __pyx_n_s_DoubleType__get_schema_extra __pyx_mstate_global->__pyx_n_s_DoubleType__get_schema_extra
#define __pyx_n_s_DoubleType_copy __pyx_mstate_global->__pyx_n_s_DoubleType_copy
#define __pyx_n_s_DuplicateName __pyx_mstate_global->__pyx_n_s_DuplicateName
#define __pyx_kp_u_Duplicate_field_name __pyx_mstate_global->__pyx_kp_u_Duplicate_field_name
#define __pyx_n_s_EMPTY_ARRAY __pyx_mstate_global->__pyx_n_s_EMPTY_ARRAY
#define __pyx_n_s_EMPTY_MP __pyx_mstate_global->__pyx_n_s_EMPTY_MP
#define __pyx_n_s_EOFError __pyx_mstate_global->__pyx_n_s_EOFError
#define __pyx_n_s_EPOCH_DATE __pyx_mstate_global->__pyx_n_s_EPOCH_DATE
#define __pyx_n_s_EPOCH_DT __pyx_mstate_global->__pyx_n_s_EPOCH_DT
#define __pyx_n_s_EPOCH_DT_NO_TZ __pyx_mstate_global->__pyx_n_s_EPOCH_DT_NO_TZ
#define __pyx_n_s_Ellipsis __pyx_mstate_global->__pyx_n_s_Ellipsis
#define __pyx_kp_s_Empty_shape_tuple_for_cython_arr __pyx_mstate_global->__pyx_kp_s_Empty_shape_tuple_for_cython_arr
#define __pyx_kp_u_End_of_file_found_trying_to_read __pyx_mstate_global->__pyx_kp_u_End_of_file_found_trying_to_read
#define __pyx_kp_u_End_of_file_found_trying_to_read_2 __pyx_mstate_global->__pyx_kp_u_End_of_file_found_trying_to_read_2
#define __pyx_n_s_Enum __pyx_mstate_global->__pyx_n_s_Enum
#define __pyx_n_s_EnumType __pyx_mstate_global->__pyx_n_s_EnumType
#define __pyx_n_s_EnumType___reduce_cython __pyx_mstate_global->__pyx_n_s_EnumType___reduce_cython
#define __pyx_n_s_EnumType___setstate_cython __pyx_mstate_global->__pyx_n_s_EnumType___setstate_cython
#define __pyx_n_s_EnumType__convert_value __pyx_mstate_global->__pyx_n_s_EnumType__convert_value
#define __pyx_n_s_EnumType__for_writer_locals_gene __pyx_mstate_global->__pyx_n_s_EnumType__for_writer_locals_gene
#define __pyx_n_s_EnumType__get_schema_extra __pyx_mstate_global->__pyx_n_s_EnumType__get_schema_extra
#define __pyx_n_s_EnumType_copy __pyx_mstate_global->__pyx_n_s_EnumType_copy
#define __pyx_kp_u_Enum_symbol __pyx_mstate_global->__pyx_kp_u_Enum_symbol
#define __pyx_kp_u_Enum_symbols_must_be_a_list_of_s __pyx_mstate_global->__pyx_kp_u_Enum_symbols_must_be_a_list_of_s
#define __pyx_kp_u_Enum_symbols_must_be_unique __pyx_mstate_global->__pyx_kp_u_Enum_symbols_must_be_unique
#define __pyx_kp_u_Error_closing_container_file_dur __pyx_mstate_global->__pyx_kp_u_Error_closing_container_file_dur
#define __pyx_kp_u_Expected_datetime_date_got __pyx_mstate_global->__pyx_kp_u_Expected_datetime_date_got
#define __pyx_kp_u_Expected_datetime_datetime_got __pyx_mstate_global->__pyx_kp_u_Expected_datetime_datetime_got
#define __pyx_kp_u_Expected_null_got __pyx_mstate_global->__pyx_kp_u_Expected_null_got
#define __pyx_n_s_ExponentTooLarge __pyx_mstate_global->__pyx_n_s_ExponentTooLarge
#define __pyx_n_s_FIELD __pyx_mstate_global->__pyx_n_s_FIELD
#define __pyx_n_u_False __pyx_mstate_global->__pyx_n_u_False
#define __pyx_kp_u_Field __pyx_mstate_global->__pyx_kp_u_Field
#define __pyx_n_s_FieldAccessor __pyx_mstate_global->__pyx_n_s_FieldAccessor
#define __pyx_n_s_FieldAccessor___reduce_cython __pyx_mstate_global->__pyx_n_s_FieldAccessor___reduce_cython
#define __pyx_n_s_FieldAccessor___setstate_cython __pyx_mstate_global->__pyx_n_s_FieldAccessor___setstate_cython
#define __pyx_n_s_FileReader __pyx_mstate_global->__pyx_n_s_FileReader
#define __pyx_n_s_FileReader___reduce_cython __pyx_mstate_global->__pyx_n_s_FileReader___reduce_cython
#define __pyx_n_s_FileReader___setstate_cython __pyx_mstate_global->__pyx_n_s_FileReader___setstate_cython
#define __pyx_n_s_FileWriter __pyx_mstate_global->__pyx_n_s_FileWriter
#define __pyx_n_s_FileWriter___reduce_cython __pyx_mstate_global->__pyx_n_s_FileWriter___reduce_cython
#define __pyx_n_s_FileWriter___setstate_cython __pyx_mstate_global->__pyx_n_s_FileWriter___setstate_cython
#define __pyx_n_s_FixedType __pyx_mstate_global->__pyx_n_s_FixedType
#define __pyx_n_s_FixedType___reduce_cython __pyx_mstate_global->__pyx_n_s_FixedType___reduce_cython
#define __pyx_n_s_FixedType___setstate_cython __pyx_mstate_global->__pyx_n_s_FixedType___setstate_cython
#define __pyx_n_s_FixedType__convert_value __pyx_mstate_global->__pyx_n_s_FixedType__convert_value
#define __pyx_n_s_FixedType__get_schema_extra __pyx_mstate_global->__pyx_n_s_FixedType__get_schema_extra
#define __pyx_n_s_FixedType_copy __pyx_mstate_global->__pyx_n_s_FixedType_copy
#define __pyx_n_s_FloatType __pyx_mstate_global->__pyx_n_s_FloatType
#define __pyx_n_s_FloatType___reduce_cython __pyx_mstate_global->__pyx_n_s_FloatType___reduce_cython
#define __pyx_n_s_FloatType___setstate_cython __pyx_mstate_global->__pyx_n_s_FloatType___setstate_cython
#define __pyx_n_s_FloatType__convert_value __pyx_mstate_global->__pyx_n_s_FloatType__convert_value
#define __pyx_n_s_FloatType__get_schema_extra __pyx_mstate_global->__pyx_n_s_FloatType__get_schema_extra
#define __pyx_n_s_FloatType_copy __pyx_mstate_global->__pyx_n_s_FloatType_copy
#define __pyx_n_s_HAVE_BZIP2 __pyx_mstate_global->__pyx_n_s_HAVE_BZIP2
#define __pyx_n_s_HAVE_LZ4 __pyx_mstate_global->__pyx_n_s_HAVE_LZ4
#define __pyx_n_s_HAVE_SNAPPY __pyx_mstate_global->__pyx_n_s_HAVE_SNAPPY
#define __pyx_n_s_HAVE_XZ __pyx_mstate_global->__pyx_n_s_HAVE_XZ
#define __pyx_n_s_HAVE_ZLIB __pyx_mstate_global->__pyx_n_s_HAVE_ZLIB
#define __pyx_n_s_HAVE_ZSTD __pyx_mstate_global->__pyx_n_s_HAVE_ZSTD
#define __pyx_n_s_IGNORE __pyx_mstate_global->__pyx_n_s_IGNORE
#define __pyx_n_s_ImportError __pyx_mstate_global->__pyx_n_s_ImportError
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_25 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_25
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_26 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_26
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_27 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_27
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_28 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_28
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9
#define __pyx_n_s_IndexError __pyx_mstate_global->__pyx_n_s_IndexError
#define __pyx_kp_s_Index_out_of_bounds_axis_d __pyx_mstate_global->__pyx_kp_s_Index_out_of_bounds_axis_d
#define __pyx_kp_s_Indirect_dimensions_not_supporte __pyx_mstate_global->__pyx_kp_s_Indirect_dimensions_not_supporte
#define __pyx_n_s_IntType __pyx_mstate_global->__pyx_n_s_IntType
#define __pyx_n_s_IntType___reduce_cython __pyx_mstate_global->__pyx_n_s_IntType___reduce_cython
#define __pyx_n_s_IntType___setstate_cython __pyx_mstate_global->__pyx_n_s_IntType___setstate_cython
#define __pyx_n_s_IntType__convert_value __pyx_mstate_global->__pyx_n_s_IntType__convert_value
#define __pyx_n_s_IntType__get_schema_extra __pyx_mstate_global->__pyx_n_s_IntType__get_schema_extra
#define __pyx_n_s_IntType_copy __pyx_mstate_global->__pyx_n_s_IntType_copy
#define __pyx_n_s_InvalidHasher __pyx_mstate_global->__pyx_n_s_InvalidHasher
#define __pyx_n_s_InvalidName __pyx_mstate_global->__pyx_n_s_InvalidName
#define __pyx_n_s_InvalidValue __pyx_mstate_global->__pyx_n_s_InvalidValue
#define __pyx_n_s_InvalidValue___init __pyx_mstate_global->__pyx_n_s_InvalidValue___init
#define __pyx_n_s_InvalidValue___str __pyx_mstate_global->__pyx_n_s_InvalidValue___str
#define __pyx_n_s_InvalidValue___str___locals_gene __pyx_mstate_global->__pyx_n_s_InvalidValue___str___locals_gene
#define __pyx_kp_u_Invalid_block_sync_marker_expect __pyx_mstate_global->__pyx_kp_u_Invalid_block_sync_marker_expect
#define __pyx_kp_u_Invalid_file_header_expected __pyx_mstate_global->__pyx_kp_u_Invalid_file_header_expected
#define __pyx_kp_u_Invalid_length_for_fixed_field __pyx_mstate_global->__pyx_kp_u_Invalid_length_for_fixed_field
#define __pyx_kp_u_Invalid_mode_expected_c_or_fortr __pyx_mstate_global->__pyx_kp_u_Invalid_mode_expected_c_or_fortr
#define __pyx_kp_u_Invalid_negative_size_for_fixed __pyx_mstate_global->__pyx_kp_u_Invalid_negative_size_for_fixed
#define __pyx_kp_u_Invalid_non_bytes_value_for_fixe __pyx_mstate_global->__pyx_kp_u_Invalid_non_bytes_value_for_fixe
#define __pyx_kp_u_Invalid_non_str_value_for_fixed __pyx_mstate_global->__pyx_kp_u_Invalid_non_str_value_for_fixed
#define __pyx_kp_u_Invalid_shape_in_axis __pyx_mstate_global->__pyx_kp_u_Invalid_shape_in_axis
#define __pyx_kp_u_Invalid_union_decodes_to_option __pyx_mstate_global->__pyx_kp_u_Invalid_union_decodes_to_option
#define __pyx_kp_u_Invalid_value __pyx_mstate_global->__pyx_kp_u_Invalid_value
#define __pyx_n_s_Iterable __pyx_mstate_global->__pyx_n_s_Iterable
#define __pyx_n_s_KeyError __pyx_mstate_global->__pyx_n_s_KeyError
#define __pyx_n_s_LOGICAL_TYPES __pyx_mstate_global->__pyx_n_s_LOGICAL_TYPES
#define __pyx_n_u_Ll __pyx_mstate_global->__pyx_n_u_Ll
#define __pyx_n_s_LogicalType __pyx_mstate_global->__pyx_n_s_LogicalType
#define __pyx_n_s_LogicalType___reduce_cython __pyx_mstate_global->__pyx_n_s_LogicalType___reduce_cython
#define __pyx_n_s_LogicalType___setstate_cython __pyx_mstate_global->__pyx_n_s_LogicalType___setstate_cython
#define __pyx_n_s_LogicalType_for_type __pyx_mstate_global->__pyx_n_s_LogicalType_for_type
#define __pyx_n_s_LongType __pyx_mstate_global->__pyx_n_s_LongType
#define __pyx_n_s_LongType___reduce_cython __pyx_mstate_global->__pyx_n_s_LongType___reduce_cython
#define __pyx_n_s_LongType___setstate_cython __pyx_mstate_global->__pyx_n_s_LongType___setstate_cython
#define __pyx_n_s_LongType__convert_value __pyx_mstate_global->__pyx_n_s_LongType__convert_value
#define __pyx_n_s_LongType__get_schema_extra __pyx_mstate_global->__pyx_n_s_LongType__get_schema_extra
#define __pyx_n_s_LongType_copy __pyx_mstate_global->__pyx_n_s_LongType_copy
#define __pyx_n_u_Lu __pyx_mstate_global->__pyx_n_u_Lu
#define __pyx_n_s_Lz4Codec __pyx_mstate_global->__pyx_n_s_Lz4Codec
#define __pyx_n_s_Lz4Codec___reduce_cython __pyx_mstate_global->__pyx_n_s_Lz4Codec___reduce_cython
#define __pyx_n_s_Lz4Codec___setstate_cython __pyx_mstate_global->__pyx_n_s_Lz4Codec___setstate_cython
#define __pyx_n_s_LzmaCodec __pyx_mstate_global->__pyx_n_s_LzmaCodec
#define __pyx_n_s_LzmaCodec___reduce_cython __pyx_mstate_global->__pyx_n_s_LzmaCodec___reduce_cython
#define __pyx_n_s_LzmaCodec___setstate_cython __pyx_mstate_global->__pyx_n_s_LzmaCodec___setstate_cython
#define __pyx_n_s_MAX_PREC __pyx_mstate_global->__pyx_n_s_MAX_PREC
#define __pyx_n_s_MAX_TIME __pyx_mstate_global->__pyx_n_s_MAX_TIME
#define __pyx_n_s_MISSING_VALUE __pyx_mstate_global->__pyx_n_s_MISSING_VALUE
#define __pyx_n_u_MISSING_VALUE __pyx_mstate_global->__pyx_n_u_MISSING_VALUE
#define __pyx_n_s_MapType __pyx_mstate_global->__pyx_n_s_MapType
#define __pyx_n_s_MapType___reduce_cython __pyx_mstate_global->__pyx_n_s_MapType___reduce_cython
#define __pyx_n_s_MapType___setstate_cython __pyx_mstate_global->__pyx_n_s_MapType___setstate_cython
#define __pyx_n_s_MapType__get_schema_extra __pyx_mstate_global->__pyx_n_s_MapType__get_schema_extra
#define __pyx_n_s_MapType_convert_value __pyx_mstate_global->__pyx_n_s_MapType_convert_value
#define __pyx_n_s_MapType_copy __pyx_mstate_global->__pyx_n_s_MapType_copy
#define __pyx_n_s_MapType_walk_types __pyx_mstate_global->__pyx_n_s_MapType_walk_types
#define __pyx_n_s_MappingProxyType __pyx_mstate_global->__pyx_n_s_MappingProxyType
#define __pyx_kp_u_Marker_must_be_bytes_got __pyx_mstate_global->__pyx_kp_u_Marker_must_be_bytes_got
#define __pyx_kp_u_Marker_must_be_exactly_16_bytes __pyx_mstate_global->__pyx_kp_u_Marker_must_be_exactly_16_bytes
#define __pyx_kp_u_Max_precision_is __pyx_mstate_global->__pyx_kp_u_Max_precision_is
#define __pyx_n_s_MemoryError __pyx_mstate_global->__pyx_n_s_MemoryError
#define __pyx_n_s_MemoryReader __pyx_mstate_global->__pyx_n_s_MemoryReader
#define __pyx_n_s_MemoryReader___reduce_cython __pyx_mstate_global->__pyx_n_s_MemoryReader___reduce_cython
#define __pyx_n_s_MemoryReader___setstate_cython __pyx_mstate_global->__pyx_n_s_MemoryReader___setstate_cython
#define __pyx_kp_s_MemoryView_of_r_at_0x_x __pyx_mstate_global->__pyx_kp_s_MemoryView_of_r_at_0x_x
#define __pyx_kp_s_MemoryView_of_r_object __pyx_mstate_global->__pyx_kp_s_MemoryView_of_r_object
#define __pyx_n_s_MemoryWriter __pyx_mstate_global->__pyx_n_s_MemoryWriter
#define __pyx_n_s_MemoryWriter___reduce_cython __pyx_mstate_global->__pyx_n_s_MemoryWriter___reduce_cython
#define __pyx_n_s_MemoryWriter___setstate_cython __pyx_mstate_global->__pyx_n_s_MemoryWriter___setstate_cython
#define __pyx_n_s_NAME_RE_STRICT __pyx_mstate_global->__pyx_n_s_NAME_RE_STRICT
#define __pyx_n_s_NAME_RE_UNICODE __pyx_mstate_global->__pyx_n_s_NAME_RE_UNICODE
#define __pyx_n_s_NO_DEFAULT __pyx_mstate_global->__pyx_n_s_NO_DEFAULT
#define __pyx_n_u_NO_DEFAULT __pyx_mstate_global->__pyx_n_u_NO_DEFAULT
#define __pyx_kp_u_Name __pyx_mstate_global->__pyx_kp_u_Name
#define __pyx_n_s_NamedType __pyx_mstate_global->__pyx_n_s_NamedType
#define __pyx_n_s_NamedType___reduce_cython __pyx_mstate_global->__pyx_n_s_NamedType___reduce_cython
#define __pyx_n_s_NamedType___setstate_cython __pyx_mstate_global->__pyx_n_s_NamedType___setstate_cython
#define __pyx_n_s_NamedType__get_schema_extra __pyx_mstate_global->__pyx_n_s_NamedType__get_schema_extra
#define __pyx_n_s_NamedType_get_namespaced_aliase __pyx_mstate_global->__pyx_n_s_NamedType_get_namespaced_aliase
#define __pyx_kp_u_Namespace __pyx_mstate_global->__pyx_kp_u_Namespace
#define __pyx_kp_u_None __pyx_mstate_global->__pyx_kp_u_None
#define __pyx_n_s_NotImplemented __pyx_mstate_global->__pyx_n_s_NotImplemented
#define __pyx_n_s_NotImplementedError __pyx_mstate_global->__pyx_n_s_NotImplementedError
#define __pyx_kp_u_Not_enough_input_data_to_read_va __pyx_mstate_global->__pyx_kp_u_Not_enough_input_data_to_read_va
#define __pyx_n_s_NullCodec __pyx_mstate_global->__pyx_n_s_NullCodec
#define __pyx_n_s_NullCodec___reduce_cython __pyx_mstate_global->__pyx_n_s_NullCodec___reduce_cython
#define __pyx_n_s_NullCodec___setstate_cython __pyx_mstate_global->__pyx_n_s_NullCodec___setstate_cython
#define __pyx_n_s_NullType __pyx_mstate_global->__pyx_n_s_NullType
#define __pyx_n_s_NullType___reduce_cython __pyx_mstate_global->__pyx_n_s_NullType___reduce_cython
#define __pyx_n_s_NullType___setstate_cython __pyx_mstate_global->__pyx_n_s_NullType___setstate_cython
#define __pyx_n_s_NullType__convert_value __pyx_mstate_global->__pyx_n_s_NullType__convert_value
#define __pyx_n_s_NullType__get_schema_extra __pyx_mstate_global->__pyx_n_s_NullType__get_schema_extra
#define __pyx_n_s_NullType_copy __pyx_mstate_global->__pyx_n_s_NullType_copy
#define __pyx_n_b_O __pyx_mstate_global->__pyx_n_b_O
#define __pyx_n_s_OBJ_FILE_METADATA __pyx_mstate_global->__pyx_n_s_OBJ_FILE_METADATA
#define __pyx_n_s_OBJ_MAGIC_BYTES __pyx_mstate_global->__pyx_n_s_OBJ_MAGIC_BYTES
#define __pyx_kp_b_Obj __pyx_mstate_global->__pyx_kp_b_Obj
#define __pyx_n_s_Options __pyx_mstate_global->__pyx_n_s_Options
#define __pyx_n_s_OptionsWrapper __pyx_mstate_global->__pyx_n_s_OptionsWrapper
#define __pyx_n_s_Options___reduce_cython __pyx_mstate_global->__pyx_n_s_Options___reduce_cython
#define __pyx_n_s_Options___setstate_cython __pyx_mstate_global->__pyx_n_s_Options___setstate_cython
#define __pyx_n_s_Options_diff __pyx_mstate_global->__pyx_n_s_Options_diff
#define __pyx_n_s_Options_equals __pyx_mstate_global->__pyx_n_s_Options_equals
#define __pyx_n_s_Options_replace __pyx_mstate_global->__pyx_n_s_Options_replace
#define __pyx_n_s_Options_with_external_types __pyx_mstate_global->__pyx_n_s_Options_with_external_types
#define __pyx_n_s_Options_with_logical_types __pyx_mstate_global->__pyx_n_s_Options_with_logical_types
#define __pyx_n_s_Order __pyx_mstate_global->__pyx_n_s_Order
#define __pyx_kp_u_Out_of_bounds_on_buffer_access_a __pyx_mstate_global->__pyx_kp_u_Out_of_bounds_on_buffer_access_a
#define __pyx_n_s_OverflowError __pyx_mstate_global->__pyx_n_s_OverflowError
#define __pyx_n_s_PERMISSIVE_OPTIONS __pyx_mstate_global->__pyx_n_s_PERMISSIVE_OPTIONS
#define __pyx_n_s_PRIMITIVE_TYPES __pyx_mstate_global->__pyx_n_s_PRIMITIVE_TYPES
#define __pyx_n_s_Path __pyx_mstate_global->__pyx_n_s_Path
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_n_s_PlaceholderType __pyx_mstate_global->__pyx_n_s_PlaceholderType
#define __pyx_n_s_PlaceholderType___reduce_cython __pyx_mstate_global->__pyx_n_s_PlaceholderType___reduce_cython
#define __pyx_n_s_PlaceholderType___setstate_cyth __pyx_mstate_global->__pyx_n_s_PlaceholderType___setstate_cyth
#define __pyx_n_s_PlaceholderType__convert_value __pyx_mstate_global->__pyx_n_s_PlaceholderType__convert_value
#define __pyx_n_s_PlaceholderType__get_schema_ext __pyx_mstate_global->__pyx_n_s_PlaceholderType__get_schema_ext
#define __pyx_kp_u_Placeholder_types_cannot_be_enco __pyx_mstate_global->__pyx_kp_u_Placeholder_types_cannot_be_enco
#define __pyx_kp_u_Placeholder_types_have_no_canoni __pyx_mstate_global->__pyx_kp_u_Placeholder_types_have_no_canoni
#define __pyx_kp_u_Precision __pyx_mstate_global->__pyx_kp_u_Precision
#define __pyx_kp_u_Precision_is_too_large_for_fixed __pyx_mstate_global->__pyx_kp_u_Precision_is_too_large_for_fixed
#define __pyx_kp_u_Precision_must_be_an_integer_bet __pyx_mstate_global->__pyx_kp_u_Precision_must_be_an_integer_bet
#define __pyx_kp_u_Precision_must_be_greater_than_s __pyx_mstate_global->__pyx_kp_u_Precision_must_be_greater_than_s
#define __pyx_n_s_PromoteBytesToString __pyx_mstate_global->__pyx_n_s_PromoteBytesToString
#define __pyx_n_s_PromoteBytesToString___reduce_cy __pyx_mstate_global->__pyx_n_s_PromoteBytesToString___reduce_cy
#define __pyx_n_s_PromoteBytesToString___setstate __pyx_mstate_global->__pyx_n_s_PromoteBytesToString___setstate
#define __pyx_n_s_PromoteStringToBytes __pyx_mstate_global->__pyx_n_s_PromoteStringToBytes
#define __pyx_n_s_PromoteStringToBytes___reduce_cy __pyx_mstate_global->__pyx_n_s_PromoteStringToBytes___reduce_cy
#define __pyx_n_s_PromoteStringToBytes___setstate __pyx_mstate_global->__pyx_n_s_PromoteStringToBytes___setstate
#define __pyx_n_s_PromoteToFloat __pyx_mstate_global->__pyx_n_s_PromoteToFloat
#define __pyx_n_s_PromoteToFloat___reduce_cython __pyx_mstate_global->__pyx_n_s_PromoteToFloat___reduce_cython
#define __pyx_n_s_PromoteToFloat___setstate_cython __pyx_mstate_global->__pyx_n_s_PromoteToFloat___setstate_cython
#define __pyx_n_s_PromotingEnumType __pyx_mstate_global->__pyx_n_s_PromotingEnumType
#define __pyx_n_s_PromotingEnumType___reduce_cytho __pyx_mstate_global->__pyx_n_s_PromotingEnumType___reduce_cytho
#define __pyx_n_s_PromotingEnumType___setstate_cyt __pyx_mstate_global->__pyx_n_s_PromotingEnumType___setstate_cyt
#define __pyx_n_s_PromotingRecordType __pyx_mstate_global->__pyx_n_s_PromotingRecordType
#define __pyx_n_s_PromotingRecordType___reduce_cyt __pyx_mstate_global->__pyx_n_s_PromotingRecordType___reduce_cyt
#define __pyx_n_s_PromotingRecordType___setstate_c __pyx_mstate_global->__pyx_n_s_PromotingRecordType___setstate_c
#define __pyx_kp_u_Promoting_schemas_reader_writer __pyx_mstate_global->__pyx_kp_u_Promoting_schemas_reader_writer
#define __pyx_n_s_RAW_VALUES __pyx_mstate_global->__pyx_n_s_RAW_VALUES
#define __pyx_n_s_Rabin __pyx_mstate_global->__pyx_n_s_Rabin
#define __pyx_n_s_Rabin___reduce_cython __pyx_mstate_global->__pyx_n_s_Rabin___reduce_cython
#define __pyx_n_s_Rabin___setstate_cython __pyx_mstate_global->__pyx_n_s_Rabin___setstate_cython
#define __pyx_n_s_Rabin_copy __pyx_mstate_global->__pyx_n_s_Rabin_copy
#define __pyx_n_s_Rabin_digest __pyx_mstate_global->__pyx_n_s_Rabin_digest
#define __pyx_n_s_Rabin_hexdigest __pyx_mstate_global->__pyx_n_s_Rabin_hexdigest
#define __pyx_n_s_Rabin_update __pyx_mstate_global->__pyx_n_s_Rabin_update
#define __pyx_n_s_Reader __pyx_mstate_global->__pyx_n_s_Reader
#define __pyx_n_s_Reader___reduce_cython __pyx_mstate_global->__pyx_n_s_Reader___reduce_cython
#define __pyx_n_s_Reader___setstate_cython __pyx_mstate_global->__pyx_n_s_Reader___setstate_cython
#define __pyx_kp_u_Record __pyx_mstate_global->__pyx_kp_u_Record
#define __pyx_n_s_RecordField __pyx_mstate_global->__pyx_n_s_RecordField
#define __pyx_n_s_RecordField___reduce_cython __pyx_mstate_global->__pyx_n_s_RecordField___reduce_cython
#define __pyx_n_s_RecordField___setstate_cython __pyx_mstate_global->__pyx_n_s_RecordField___setstate_cython
#define __pyx_n_s_RecordField_get_schema __pyx_mstate_global->__pyx_n_s_RecordField_get_schema
#define __pyx_n_s_RecordType __pyx_mstate_global->__pyx_n_s_RecordType
#define __pyx_n_s_RecordType___reduce_cython __pyx_mstate_global->__pyx_n_s_RecordType___reduce_cython
#define __pyx_n_s_RecordType___setstate_cython __pyx_mstate_global->__pyx_n_s_RecordType___setstate_cython
#define __pyx_n_s_RecordType__convert_value __pyx_mstate_global->__pyx_n_s_RecordType__convert_value
#define __pyx_n_s_RecordType__get_schema_extra __pyx_mstate_global->__pyx_n_s_RecordType__get_schema_extra
#define __pyx_n_s_RecordType_copy __pyx_mstate_global->__pyx_n_s_RecordType_copy
#define __pyx_n_s_RecordType_walk_types __pyx_mstate_global->__pyx_n_s_RecordType_walk_types
#define __pyx_kp_u_Record_2 __pyx_mstate_global->__pyx_kp_u_Record_2
#define __pyx_n_s_Record_3 __pyx_mstate_global->__pyx_n_s_Record_3
#define __pyx_n_s_Record___dir __pyx_mstate_global->__pyx_n_s_Record___dir
#define __pyx_n_s_Record___reduce_cython __pyx_mstate_global->__pyx_n_s_Record___reduce_cython
#define __pyx_n_s_Record___setstate_cython __pyx_mstate_global->__pyx_n_s_Record___setstate_cython
#define __pyx_n_s_Record__asdict __pyx_mstate_global->__pyx_n_s_Record__asdict
#define __pyx_n_s_Record__repr_children_locals_gen __pyx_mstate_global->__pyx_n_s_Record__repr_children_locals_gen
#define __pyx_kp_u_Records_may_either_be_instantiat __pyx_mstate_global->__pyx_kp_u_Records_may_either_be_instantiat
#define __pyx_kp_u_Records_may_either_be_instantiat_2 __pyx_mstate_global->__pyx_kp_u_Records_may_either_be_instantiat_2
#define __pyx_n_s_ResolvedSchema __pyx_mstate_global->__pyx_n_s_ResolvedSchema
#define __pyx_n_s_ResolvedSchema___reduce_cython __pyx_mstate_global->__pyx_n_s_ResolvedSchema___reduce_cython
#define __pyx_n_s_ResolvedSchema___setstate_cython __pyx_mstate_global->__pyx_n_s_ResolvedSchema___setstate_cython
#define __pyx_n_s_ResourceWarning __pyx_mstate_global->__pyx_n_s_ResourceWarning
#define __pyx_n_s_SLICE_TYPE __pyx_mstate_global->__pyx_n_s_SLICE_TYPE
#define __pyx_n_s_SSIZE_TYPECODE __pyx_mstate_global->__pyx_n_s_SSIZE_TYPECODE
#define __pyx_kp_u_Scale_must_be_a_positive_integer __pyx_mstate_global->__pyx_kp_u_Scale_must_be_a_positive_integer
#define __pyx_n_s_Schema __pyx_mstate_global->__pyx_n_s_Schema
#define __pyx_n_s_Schema___reduce_cython __pyx_mstate_global->__pyx_n_s_Schema___reduce_cython
#define __pyx_n_s_Schema___setstate_cython __pyx_mstate_global->__pyx_n_s_Schema___setstate_cython
#define __pyx_n_s_Schema__make_logical_types __pyx_mstate_global->__pyx_n_s_Schema__make_logical_types
#define __pyx_n_s_Schema__wrap_type __pyx_mstate_global->__pyx_n_s_Schema__wrap_type
#define __pyx_n_s_Schema_binary_decode __pyx_mstate_global->__pyx_n_s_Schema_binary_decode
#define __pyx_n_s_Schema_binary_encode __pyx_mstate_global->__pyx_n_s_Schema_binary_encode
#define __pyx_n_s_Schema_binary_read __pyx_mstate_global->__pyx_n_s_Schema_binary_read
#define __pyx_n_s_Schema_binary_write __pyx_mstate_global->__pyx_n_s_Schema_binary_write
#define __pyx_n_s_Schema_can_encode __pyx_mstate_global->__pyx_n_s_Schema_can_encode
#define __pyx_n_s_Schema_find_type __pyx_mstate_global->__pyx_n_s_Schema_find_type
#define __pyx_n_s_Schema_fingerprint __pyx_mstate_global->__pyx_n_s_Schema_fingerprint
#define __pyx_kp_u_Schema_is_required __pyx_mstate_global->__pyx_kp_u_Schema_is_required
#define __pyx_n_s_Schema_json_decode __pyx_mstate_global->__pyx_n_s_Schema_json_decode
#define __pyx_n_s_Schema_json_encode __pyx_mstate_global->__pyx_n_s_Schema_json_encode
#define __pyx_n_s_Schema_reader_for_writer __pyx_mstate_global->__pyx_n_s_Schema_reader_for_writer
#define __pyx_n_s_Sentinel __pyx_mstate_global->__pyx_n_s_Sentinel
#define __pyx_n_s_Sentinel___reduce_cython __pyx_mstate_global->__pyx_n_s_Sentinel___reduce_cython
#define __pyx_n_s_Sentinel___setstate_cython __pyx_mstate_global->__pyx_n_s_Sentinel___setstate_cython
#define __pyx_n_s_Sequence __pyx_mstate_global->__pyx_n_s_Sequence
#define __pyx_kp_s_Sequence_str __pyx_mstate_global->__pyx_kp_s_Sequence_str
#define __pyx_n_s_SignatureWrapper __pyx_mstate_global->__pyx_n_s_SignatureWrapper
#define __pyx_n_s_SignatureWrapper___init __pyx_mstate_global->__pyx_n_s_SignatureWrapper___init
#define __pyx_n_s_SignatureWrapper__set_from_data __pyx_mstate_global->__pyx_n_s_SignatureWrapper__set_from_data
#define __pyx_n_s_SnappyCodec __pyx_mstate_global->__pyx_n_s_SnappyCodec
#define __pyx_n_s_SnappyCodec___reduce_cython __pyx_mstate_global->__pyx_n_s_SnappyCodec___reduce_cython
#define __pyx_n_s_SnappyCodec___setstate_cython __pyx_mstate_global->__pyx_n_s_SnappyCodec___setstate_cython
#define __pyx_kp_s_Step_may_not_be_zero_axis_d __pyx_mstate_global->__pyx_kp_s_Step_may_not_be_zero_axis_d
#define __pyx_n_s_StopIteration __pyx_mstate_global->__pyx_n_s_StopIteration
#define __pyx_n_s_StringType __pyx_mstate_global->__pyx_n_s_StringType
#define __pyx_n_s_StringType___reduce_cython __pyx_mstate_global->__pyx_n_s_StringType___reduce_cython
#define __pyx_n_s_StringType___setstate_cython __pyx_mstate_global->__pyx_n_s_StringType___setstate_cython
#define __pyx_n_s_StringType__convert_value __pyx_mstate_global->__pyx_n_s_StringType__convert_value
#define __pyx_n_s_StringType__get_schema_extra __pyx_mstate_global->__pyx_n_s_StringType__get_schema_extra
#define __pyx_n_s_StringType_copy __pyx_mstate_global->__pyx_n_s_StringType_copy
#define __pyx_n_s_TYPES_BY_NAME __pyx_mstate_global->__pyx_n_s_TYPES_BY_NAME
#define __pyx_n_s_TYPE_TUPLE_ALWAYS __pyx_mstate_global->__pyx_n_s_TYPE_TUPLE_ALWAYS
#define __pyx_n_s_TYPE_TUPLE_IF_AMBIGUOUS __pyx_mstate_global->__pyx_n_s_TYPE_TUPLE_IF_AMBIGUOUS
#define __pyx_n_s_TYPE_TUPLE_IF_RECORD __pyx_mstate_global->__pyx_n_s_TYPE_TUPLE_IF_RECORD
#define __pyx_n_s_TYPE_TUPLE_IF_RECORD_AMBIGUOUS __pyx_mstate_global->__pyx_n_s_TYPE_TUPLE_IF_RECORD_AMBIGUOUS
#define __pyx_kp_s_The_exponent_of_a_decimal_value __pyx_mstate_global->__pyx_kp_s_The_exponent_of_a_decimal_value
#define __pyx_kp_u_The_null_namespace_cannot_be_spe __pyx_mstate_global->__pyx_kp_u_The_null_namespace_cannot_be_spe
#define __pyx_kp_s_The_order_of_a_field_in_a_recor __pyx_mstate_global->__pyx_kp_s_The_order_of_a_field_in_a_recor
#define __pyx_kp_s_The_schema_contains_a_type_or_e __pyx_mstate_global->__pyx_kp_s_The_schema_contains_a_type_or_e
#define __pyx_kp_s_The_schema_contains_an_unexptec __pyx_mstate_global->__pyx_kp_s_The_schema_contains_an_unexptec
#define __pyx_n_s_TimeMicros __pyx_mstate_global->__pyx_n_s_TimeMicros
#define __pyx_n_s_TimeMicros___reduce_cython __pyx_mstate_global->__pyx_n_s_TimeMicros___reduce_cython
#define __pyx_n_s_TimeMicros___setstate_cython __pyx_mstate_global->__pyx_n_s_TimeMicros___setstate_cython
#define __pyx_n_s_TimeMicros__for_type __pyx_mstate_global->__pyx_n_s_TimeMicros__for_type
#define __pyx_n_s_TimeMillis __pyx_mstate_global->__pyx_n_s_TimeMillis
#define __pyx_n_s_TimeMillis___reduce_cython __pyx_mstate_global->__pyx_n_s_TimeMillis___reduce_cython
#define __pyx_n_s_TimeMillis___setstate_cython __pyx_mstate_global->__pyx_n_s_TimeMillis___setstate_cython
#define __pyx_n_s_TimeMillis__for_type __pyx_mstate_global->__pyx_n_s_TimeMillis__for_type
#define __pyx_n_s_TimestampMicros __pyx_mstate_global->__pyx_n_s_TimestampMicros
#define __pyx_n_s_TimestampMicros___reduce_cython __pyx_mstate_global->__pyx_n_s_TimestampMicros___reduce_cython
#define __pyx_n_s_TimestampMicros___setstate_cytho __pyx_mstate_global->__pyx_n_s_TimestampMicros___setstate_cytho
#define __pyx_n_s_TimestampMicros__for_type __pyx_mstate_global->__pyx_n_s_TimestampMicros__for_type
#define __pyx_n_s_TimestampMillis __pyx_mstate_global->__pyx_n_s_TimestampMillis
#define __pyx_n_s_TimestampMillis___reduce_cython __pyx_mstate_global->__pyx_n_s_TimestampMillis___reduce_cython
#define __pyx_n_s_TimestampMillis___setstate_cytho __pyx_mstate_global->__pyx_n_s_TimestampMillis___setstate_cytho
#define __pyx_n_s_TimestampMillis__for_type __pyx_mstate_global->__pyx_n_s_TimestampMillis__for_type
#define __pyx_n_u_True __pyx_mstate_global->__pyx_n_u_True
#define __pyx_kp_u_Trying_to_close_a_closed_Contain __pyx_mstate_global->__pyx_kp_u_Trying_to_close_a_closed_Contain
#define __pyx_kp_u_Trying_to_write_to_closed_Contai __pyx_mstate_global->__pyx_kp_u_Trying_to_write_to_closed_Contai
#define __pyx_n_s_Type __pyx_mstate_global->__pyx_n_s_Type
#define __pyx_n_u_Type __pyx_mstate_global->__pyx_n_u_Type
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_kp_u_Type_name __pyx_mstate_global->__pyx_kp_u_Type_name
#define __pyx_n_u_UNKNOWN __pyx_mstate_global->__pyx_n_u_UNKNOWN
#define __pyx_n_s_UUID __pyx_mstate_global->__pyx_n_s_UUID
#define __pyx_n_s_UUIDBase __pyx_mstate_global->__pyx_n_s_UUIDBase
#define __pyx_n_s_UUIDBase___reduce_cython __pyx_mstate_global->__pyx_n_s_UUIDBase___reduce_cython
#define __pyx_n_s_UUIDBase___setstate_cython __pyx_mstate_global->__pyx_n_s_UUIDBase___setstate_cython
#define __pyx_n_s_UUIDBase__for_type __pyx_mstate_global->__pyx_n_s_UUIDBase__for_type
#define __pyx_n_s_UUIDFixedType __pyx_mstate_global->__pyx_n_s_UUIDFixedType
#define __pyx_n_s_UUIDFixedType___reduce_cython __pyx_mstate_global->__pyx_n_s_UUIDFixedType___reduce_cython
#define __pyx_n_s_UUIDFixedType___setstate_cython __pyx_mstate_global->__pyx_n_s_UUIDFixedType___setstate_cython
#define __pyx_n_s_UUIDFixedType__for_type __pyx_mstate_global->__pyx_n_s_UUIDFixedType__for_type
#define __pyx_n_s_UUIDStringType __pyx_mstate_global->__pyx_n_s_UUIDStringType
#define __pyx_n_s_UUIDStringType___reduce_cython __pyx_mstate_global->__pyx_n_s_UUIDStringType___reduce_cython
#define __pyx_n_s_UUIDStringType___setstate_cython __pyx_mstate_global->__pyx_n_s_UUIDStringType___setstate_cython
#define __pyx_n_s_UUIDStringType__for_type __pyx_mstate_global->__pyx_n_s_UUIDStringType__for_type
#define __pyx_kp_s_Unable_to_convert_item_to_object __pyx_mstate_global->__pyx_kp_s_Unable_to_convert_item_to_object
#define __pyx_n_s_UnicodeEncodeError __pyx_mstate_global->__pyx_n_s_UnicodeEncodeError
#define __pyx_n_s_Union __pyx_mstate_global->__pyx_n_s_Union
#define __pyx_n_s_UnionDecodeOption __pyx_mstate_global->__pyx_n_s_UnionDecodeOption
#define __pyx_n_s_UnionType __pyx_mstate_global->__pyx_n_s_UnionType
#define __pyx_n_s_UnionType___reduce_cython __pyx_mstate_global->__pyx_n_s_UnionType___reduce_cython
#define __pyx_n_s_UnionType___setstate_cython __pyx_mstate_global->__pyx_n_s_UnionType___setstate_cython
#define __pyx_n_s_UnionType_convert_value __pyx_mstate_global->__pyx_n_s_UnionType_convert_value
#define __pyx_n_s_UnionType_copy __pyx_mstate_global->__pyx_n_s_UnionType_copy
#define __pyx_n_s_UnionType_get_schema __pyx_mstate_global->__pyx_n_s_UnionType_get_schema
#define __pyx_n_s_UnionType_walk_types __pyx_mstate_global->__pyx_n_s_UnionType_walk_types
#define __pyx_kp_s_Union_bytes_hashlib__hashlib_HAS __pyx_mstate_global->__pyx_kp_s_Union_bytes_hashlib__hashlib_HAS
#define __pyx_kp_s_Union_str_object __pyx_mstate_global->__pyx_kp_s_Union_str_object
#define __pyx_kp_u_Unions_may_not_have_more_than_on __pyx_mstate_global->__pyx_kp_u_Unions_may_not_have_more_than_on
#define __pyx_kp_u_Unions_may_not_immediately_conta __pyx_mstate_global->__pyx_kp_u_Unions_may_not_immediately_conta
#define __pyx_kp_u_Unions_must_contain_at_least_one __pyx_mstate_global->__pyx_kp_u_Unions_must_contain_at_least_one
#define __pyx_n_s_UnknownType __pyx_mstate_global->__pyx_n_s_UnknownType
#define __pyx_n_s_UnknownType___init __pyx_mstate_global->__pyx_n_s_UnknownType___init
#define __pyx_kp_u_Unknown_hash_method __pyx_mstate_global->__pyx_kp_u_Unknown_hash_method
#define __pyx_kp_u_Unknown_type __pyx_mstate_global->__pyx_kp_u_Unknown_type
#define __pyx_kp_u_Unknown_value_for_enum __pyx_mstate_global->__pyx_kp_u_Unknown_value_for_enum
#define __pyx_kp_u_Unsupported_codec __pyx_mstate_global->__pyx_kp_u_Unsupported_codec
#define __pyx_kp_u_Value __pyx_mstate_global->__pyx_kp_u_Value
#define __pyx_n_s_ValueAdapter __pyx_mstate_global->__pyx_n_s_ValueAdapter
#define __pyx_n_s_ValueAdapter___reduce_cython __pyx_mstate_global->__pyx_n_s_ValueAdapter___reduce_cython
#define __pyx_n_s_ValueAdapter___setstate_cython __pyx_mstate_global->__pyx_n_s_ValueAdapter___setstate_cython
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_n_s_View_MemoryView __pyx_mstate_global->__pyx_n_s_View_MemoryView
#define __pyx_n_s_Writer __pyx_mstate_global->__pyx_n_s_Writer
#define __pyx_n_s_Writer___reduce_cython __pyx_mstate_global->__pyx_n_s_Writer___reduce_cython
#define __pyx_n_s_Writer___setstate_cython __pyx_mstate_global->__pyx_n_s_Writer___setstate_cython
#define __pyx_kp_u_Y_m_d __pyx_mstate_global->__pyx_kp_u_Y_m_d
#define __pyx_n_s_ZStandardCodec __pyx_mstate_global->__pyx_n_s_ZStandardCodec
#define __pyx_n_s_ZStandardCodec___reduce_cython __pyx_mstate_global->__pyx_n_s_ZStandardCodec___reduce_cython
#define __pyx_n_s_ZStandardCodec___setstate_cytho __pyx_mstate_global->__pyx_n_s_ZStandardCodec___setstate_cytho
#define __pyx_n_s_ZstdCompressor __pyx_mstate_global->__pyx_n_s_ZstdCompressor
#define __pyx_n_s_ZstdDecompressor __pyx_mstate_global->__pyx_n_s_ZstdDecompressor
#define __pyx_n_s__122 __pyx_mstate_global->__pyx_n_s__122
#define __pyx_kp_b__13 __pyx_mstate_global->__pyx_kp_b__13
#define __pyx_kp_u__13 __pyx_mstate_global->__pyx_kp_u__13
#define __pyx_kp_u__14 __pyx_mstate_global->__pyx_kp_u__14
#define __pyx_n_s__152 __pyx_mstate_global->__pyx_n_s__152
#define __pyx_kp_s__18 __pyx_mstate_global->__pyx_kp_s__18
#define __pyx_kp_u__18 __pyx_mstate_global->__pyx_kp_u__18
#define __pyx_kp_u__2 __pyx_mstate_global->__pyx_kp_u__2
#define __pyx_kp_u__24 __pyx_mstate_global->__pyx_kp_u__24
#define __pyx_kp_u__25 __pyx_mstate_global->__pyx_kp_u__25
#define __pyx_kp_u__26 __pyx_mstate_global->__pyx_kp_u__26
#define __pyx_kp_u__27 __pyx_mstate_global->__pyx_kp_u__27
#define __pyx_n_s__3 __pyx_mstate_global->__pyx_n_s__3
#define __pyx_kp_b__34 __pyx_mstate_global->__pyx_kp_b__34
#define __pyx_kp_u__48 __pyx_mstate_global->__pyx_kp_u__48
#define __pyx_kp_u__6 __pyx_mstate_global->__pyx_kp_u__6
#define __pyx_kp_u__60 __pyx_mstate_global->__pyx_kp_u__60
#define __pyx_kp_u__61 __pyx_mstate_global->__pyx_kp_u__61
#define __pyx_kp_u__62 __pyx_mstate_global->__pyx_kp_u__62
#define __pyx_kp_u__63 __pyx_mstate_global->__pyx_kp_u__63
#define __pyx_kp_u__64 __pyx_mstate_global->__pyx_kp_u__64
#define __pyx_kp_u__7 __pyx_mstate_global->__pyx_kp_u__7
#define __pyx_kp_u__70 __pyx_mstate_global->__pyx_kp_u__70
#define __pyx_kp_b__72 __pyx_mstate_global->__pyx_kp_b__72
#define __pyx_kp_u__76 __pyx_mstate_global->__pyx_kp_u__76
#define __pyx_kp_u__77 __pyx_mstate_global->__pyx_kp_u__77
#define __pyx_kp_u__78 __pyx_mstate_global->__pyx_kp_u__78
#define __pyx_kp_u__79 __pyx_mstate_global->__pyx_kp_u__79
#define __pyx_kp_u__80 __pyx_mstate_global->__pyx_kp_u__80
#define __pyx_kp_u__9 __pyx_mstate_global->__pyx_kp_u__9
#define __pyx_n_s_a __pyx_mstate_global->__pyx_n_s_a
#define __pyx_n_b_abacus __pyx_mstate_global->__pyx_n_b_abacus
#define __pyx_n_b_abacuscounter1 __pyx_mstate_global->__pyx_n_b_abacuscounter1
#define __pyx_n_s_abc __pyx_mstate_global->__pyx_n_s_abc
#define __pyx_n_s_accepts_string __pyx_mstate_global->__pyx_n_s_accepts_string
#define __pyx_n_s_actual __pyx_mstate_global->__pyx_n_s_actual
#define __pyx_n_s_adapt_record_types __pyx_mstate_global->__pyx_n_s_adapt_record_types
#define __pyx_kp_u_adapt_record_types_2 __pyx_mstate_global->__pyx_kp_u_adapt_record_types_2
#define __pyx_n_s_add __pyx_mstate_global->__pyx_n_s_add
#define __pyx_n_u_aliases __pyx_mstate_global->__pyx_n_u_aliases
#define __pyx_n_s_allocate_buffer __pyx_mstate_global->__pyx_n_s_allocate_buffer
#define __pyx_n_s_allow_aliases_to_be_string __pyx_mstate_global->__pyx_n_s_allow_aliases_to_be_string
#define __pyx_kp_u_allow_aliases_to_be_string_2 __pyx_mstate_global->__pyx_kp_u_allow_aliases_to_be_string_2
#define __pyx_n_s_allow_duplicate_union_types __pyx_mstate_global->__pyx_n_s_allow_duplicate_union_types
#define __pyx_kp_u_allow_duplicate_union_types_2 __pyx_mstate_global->__pyx_kp_u_allow_duplicate_union_types_2
#define __pyx_n_s_allow_empty_unions __pyx_mstate_global->__pyx_n_s_allow_empty_unions
#define __pyx_kp_u_allow_empty_unions_2 __pyx_mstate_global->__pyx_kp_u_allow_empty_unions_2
#define __pyx_n_s_allow_error_type __pyx_mstate_global->__pyx_n_s_allow_error_type
#define __pyx_kp_u_allow_error_type_2 __pyx_mstate_global->__pyx_kp_u_allow_error_type_2
#define __pyx_n_s_allow_false_values_for_null __pyx_mstate_global->__pyx_n_s_allow_false_values_for_null
#define __pyx_kp_u_allow_false_values_for_null_2 __pyx_mstate_global->__pyx_kp_u_allow_false_values_for_null_2
#define __pyx_n_s_allow_invalid_default_values __pyx_mstate_global->__pyx_n_s_allow_invalid_default_values
#define __pyx_kp_u_allow_invalid_default_values_2 __pyx_mstate_global->__pyx_kp_u_allow_invalid_default_values_2
#define __pyx_n_s_allow_leading_dot_in_names __pyx_mstate_global->__pyx_n_s_allow_leading_dot_in_names
#define __pyx_kp_u_allow_leading_dot_in_names_2 __pyx_mstate_global->__pyx_kp_u_allow_leading_dot_in_names_2
#define __pyx_n_s_allow_nested_unions __pyx_mstate_global->__pyx_n_s_allow_nested_unions
#define __pyx_kp_u_allow_nested_unions_2 __pyx_mstate_global->__pyx_kp_u_allow_nested_unions_2
#define __pyx_n_s_allow_primitive_name_collision __pyx_mstate_global->__pyx_n_s_allow_primitive_name_collision
#define __pyx_kp_u_allow_primitive_name_collision_2 __pyx_mstate_global->__pyx_kp_u_allow_primitive_name_collision_2
#define __pyx_kp_u_allow_primitive_names_in_namesp __pyx_mstate_global->__pyx_kp_u_allow_primitive_names_in_namesp
#define __pyx_n_s_allow_primitive_names_in_namespa __pyx_mstate_global->__pyx_n_s_allow_primitive_names_in_namespa
#define __pyx_n_s_allow_tuple_notation __pyx_mstate_global->__pyx_n_s_allow_tuple_notation
#define __pyx_kp_u_allow_tuple_notation_2 __pyx_mstate_global->__pyx_kp_u_allow_tuple_notation_2
#define __pyx_n_s_allow_union_default_any_member __pyx_mstate_global->__pyx_n_s_allow_union_default_any_member
#define __pyx_kp_u_allow_union_default_any_member_2 __pyx_mstate_global->__pyx_kp_u_allow_union_default_any_member_2
#define __pyx_n_s_alternate_timestamp_encoding __pyx_mstate_global->__pyx_n_s_alternate_timestamp_encoding
#define __pyx_kp_u_alternate_timestamp_millis_enco __pyx_mstate_global->__pyx_kp_u_alternate_timestamp_millis_enco
#define __pyx_n_s_alternate_timestamp_millis_encod __pyx_mstate_global->__pyx_n_s_alternate_timestamp_millis_encod
#define __pyx_kp_u_and __pyx_mstate_global->__pyx_kp_u_and
#define __pyx_n_s_annotations __pyx_mstate_global->__pyx_n_s_annotations
#define __pyx_n_s_annotations_2 __pyx_mstate_global->__pyx_n_s_annotations_2
#define __pyx_kp_u_appears_multiple_times_in_schem __pyx_mstate_global->__pyx_kp_u_appears_multiple_times_in_schem
#define __pyx_n_s_append __pyx_mstate_global->__pyx_n_s_append
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_array __pyx_mstate_global->__pyx_n_s_array
#define __pyx_n_u_array __pyx_mstate_global->__pyx_n_u_array
#define __pyx_n_s_as_tuple __pyx_mstate_global->__pyx_n_s_as_tuple
#define __pyx_n_u_ascending __pyx_mstate_global->__pyx_n_u_ascending
#define __pyx_n_s_ascii_name_rules __pyx_mstate_global->__pyx_n_s_ascii_name_rules
#define __pyx_kp_u_ascii_name_rules_2 __pyx_mstate_global->__pyx_kp_u_ascii_name_rules_2
#define __pyx_n_s_asdict __pyx_mstate_global->__pyx_n_s_asdict
#define __pyx_n_u_asdict __pyx_mstate_global->__pyx_n_u_asdict
#define __pyx_n_s_astimezone __pyx_mstate_global->__pyx_n_s_astimezone
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_kp_u_at __pyx_mstate_global->__pyx_kp_u_at
#define __pyx_kp_u_avro_codec __pyx_mstate_global->__pyx_kp_u_avro_codec
#define __pyx_kp_u_avro_schema __pyx_mstate_global->__pyx_kp_u_avro_schema
#define __pyx_n_s_avro_type __pyx_mstate_global->__pyx_n_s_avro_type
#define __pyx_n_s_b __pyx_mstate_global->__pyx_n_s_b
#define __pyx_n_s_base __pyx_mstate_global->__pyx_n_s_base
#define __pyx_n_u_big __pyx_mstate_global->__pyx_n_u_big
#define __pyx_n_s_bin __pyx_mstate_global->__pyx_n_s_bin
#define __pyx_n_s_binary_decode __pyx_mstate_global->__pyx_n_s_binary_decode
#define __pyx_n_s_binary_encode __pyx_mstate_global->__pyx_n_s_binary_encode
#define __pyx_n_s_binary_read __pyx_mstate_global->__pyx_n_s_binary_read
#define __pyx_n_s_binary_write __pyx_mstate_global->__pyx_n_s_binary_write
#define __pyx_n_s_bint __pyx_mstate_global->__pyx_n_s_bint
#define __pyx_n_u_bint __pyx_mstate_global->__pyx_n_u_bint
#define __pyx_n_s_bit_length __pyx_mstate_global->__pyx_n_s_bit_length
#define __pyx_n_s_block_size __pyx_mstate_global->__pyx_n_s_block_size
#define __pyx_n_s_bool __pyx_mstate_global->__pyx_n_s_bool
#define __pyx_n_u_bool __pyx_mstate_global->__pyx_n_u_bool
#define __pyx_n_s_bool_2 __pyx_mstate_global->__pyx_n_s_bool_2
#define __pyx_kp_u_boolean __pyx_mstate_global->__pyx_kp_u_boolean
#define __pyx_n_u_boolean_2 __pyx_mstate_global->__pyx_n_u_boolean_2
#define __pyx_n_s_buffer __pyx_mstate_global->__pyx_n_s_buffer
#define __pyx_n_s_byteorder __pyx_mstate_global->__pyx_n_s_byteorder
#define __pyx_kp_u_bytes __pyx_mstate_global->__pyx_kp_u_bytes
#define __pyx_kp_u_bytes_2 __pyx_mstate_global->__pyx_kp_u_bytes_2
#define __pyx_n_s_bytes_3 __pyx_mstate_global->__pyx_n_s_bytes_3
#define __pyx_n_u_bytes_3 __pyx_mstate_global->__pyx_n_u_bytes_3
#define __pyx_n_s_bytes_codec __pyx_mstate_global->__pyx_n_s_bytes_codec
#define __pyx_kp_u_bytes_codec_2 __pyx_mstate_global->__pyx_kp_u_bytes_codec_2
#define __pyx_n_s_bytes_default_value_utf8 __pyx_mstate_global->__pyx_n_s_bytes_default_value_utf8
#define __pyx_kp_u_bytes_default_value_utf8_2 __pyx_mstate_global->__pyx_kp_u_bytes_default_value_utf8_2
#define __pyx_n_s_bz2 __pyx_mstate_global->__pyx_n_s_bz2
#define __pyx_n_b_bzip2 __pyx_mstate_global->__pyx_n_b_bzip2
#define __pyx_n_u_bzip2 __pyx_mstate_global->__pyx_n_u_bzip2
#define __pyx_n_s_c __pyx_mstate_global->__pyx_n_s_c
#define __pyx_n_u_c __pyx_mstate_global->__pyx_n_u_c
#define __pyx_n_s_can_encode __pyx_mstate_global->__pyx_n_s_can_encode
#define __pyx_kp_u_cannot_be_adapted_to __pyx_mstate_global->__pyx_kp_u_cannot_be_adapted_to
#define __pyx_n_s_canonical_form __pyx_mstate_global->__pyx_n_s_canonical_form
#define __pyx_n_s_canonical_form_repeat_enum __pyx_mstate_global->__pyx_n_s_canonical_form_repeat_enum
#define __pyx_kp_u_canonical_form_repeat_enum_2 __pyx_mstate_global->__pyx_kp_u_canonical_form_repeat_enum_2
#define __pyx_n_s_canonical_form_repeat_fixed __pyx_mstate_global->__pyx_n_s_canonical_form_repeat_fixed
#define __pyx_kp_u_canonical_form_repeat_fixed_2 __pyx_mstate_global->__pyx_kp_u_canonical_form_repeat_fixed_2
#define __pyx_n_s_category __pyx_mstate_global->__pyx_n_s_category
#define __pyx_n_s_cavro __pyx_mstate_global->__pyx_n_s_cavro
#define __pyx_kp_s_cavro_pyx __pyx_mstate_global->__pyx_kp_s_cavro_pyx
#define __pyx_n_s_changes __pyx_mstate_global->__pyx_n_s_changes
#define __pyx_n_s_check_exp_overflow __pyx_mstate_global->__pyx_n_s_check_exp_overflow
#define __pyx_n_s_check_value __pyx_mstate_global->__pyx_n_s_check_value
#define __pyx_n_s_chr __pyx_mstate_global->__pyx_n_s_chr
#define __pyx_n_s_clamp __pyx_mstate_global->__pyx_n_s_clamp
#define __pyx_n_s_clamp_float_overflow __pyx_mstate_global->__pyx_n_s_clamp_float_overflow
#define __pyx_kp_u_clamp_float_overflow_2 __pyx_mstate_global->__pyx_kp_u_clamp_float_overflow_2
#define __pyx_n_s_clamp_int_overflow __pyx_mstate_global->__pyx_n_s_clamp_int_overflow
#define __pyx_kp_u_clamp_int_overflow_2 __pyx_mstate_global->__pyx_kp_u_clamp_int_overflow_2
#define __pyx_n_s_class __pyx_mstate_global->__pyx_n_s_class
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_class_inst_method __pyx_mstate_global->__pyx_n_s_class_inst_method
#define __pyx_n_s_class_inst_method___get __pyx_mstate_global->__pyx_n_s_class_inst_method___get
#define __pyx_n_s_class_inst_method___init __pyx_mstate_global->__pyx_n_s_class_inst_method___init
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_s_cls __pyx_mstate_global->__pyx_n_s_cls
#define __pyx_n_s_codec __pyx_mstate_global->__pyx_n_s_codec
#define __pyx_n_s_coerce_int_to_float __pyx_mstate_global->__pyx_n_s_coerce_int_to_float
#define __pyx_kp_u_coerce_int_to_float_2 __pyx_mstate_global->__pyx_kp_u_coerce_int_to_float_2
#define __pyx_n_s_coerce_values_to_boolean __pyx_mstate_global->__pyx_n_s_coerce_values_to_boolean
#define __pyx_kp_u_coerce_values_to_boolean_2 __pyx_mstate_global->__pyx_kp_u_coerce_values_to_boolean_2
#define __pyx_n_s_coerce_values_to_float __pyx_mstate_global->__pyx_n_s_coerce_values_to_float
#define __pyx_kp_u_coerce_values_to_float_2 __pyx_mstate_global->__pyx_kp_u_coerce_values_to_float_2
#define __pyx_n_s_coerce_values_to_int __pyx_mstate_global->__pyx_n_s_coerce_values_to_int
#define __pyx_kp_u_coerce_values_to_int_2 __pyx_mstate_global->__pyx_kp_u_coerce_values_to_int_2
#define __pyx_n_s_coerce_values_to_str __pyx_mstate_global->__pyx_n_s_coerce_values_to_str
#define __pyx_kp_u_coerce_values_to_str_2 __pyx_mstate_global->__pyx_kp_u_coerce_values_to_str_2
#define __pyx_n_s_collections __pyx_mstate_global->__pyx_n_s_collections
#define __pyx_kp_s_collections_abc __pyx_mstate_global->__pyx_kp_s_collections_abc
#define __pyx_n_s_compare __pyx_mstate_global->__pyx_n_s_compare
#define __pyx_n_s_compile __pyx_mstate_global->__pyx_n_s_compile
#define __pyx_n_s_compress __pyx_mstate_global->__pyx_n_s_compress
#define __pyx_n_s_container_fill_blocks __pyx_mstate_global->__pyx_n_s_container_fill_blocks
#define __pyx_kp_u_container_fill_blocks_2 __pyx_mstate_global->__pyx_kp_u_container_fill_blocks_2
#define __pyx_n_s_context __pyx_mstate_global->__pyx_n_s_context
#define __pyx_kp_s_contiguous_and_direct __pyx_mstate_global->__pyx_kp_s_contiguous_and_direct
#define __pyx_kp_s_contiguous_and_indirect __pyx_mstate_global->__pyx_kp_s_contiguous_and_indirect
#define __pyx_n_s_convert_value __pyx_mstate_global->__pyx_n_s_convert_value
#define __pyx_n_s_convert_value_2 __pyx_mstate_global->__pyx_n_s_convert_value_2
#define __pyx_n_s_copy __pyx_mstate_global->__pyx_n_s_copy
#define __pyx_n_s_copy_time __pyx_mstate_global->__pyx_n_s_copy_time
#define __pyx_n_s_count __pyx_mstate_global->__pyx_n_s_count
#define __pyx_n_b_counter __pyx_mstate_global->__pyx_n_b_counter
#define __pyx_n_s_crc32 __pyx_mstate_global->__pyx_n_s_crc32
#define __pyx_n_s_create_decimal __pyx_mstate_global->__pyx_n_s_create_decimal
#define __pyx_n_s_created __pyx_mstate_global->__pyx_n_s_created
#define __pyx_n_s_custom_decode_value __pyx_mstate_global->__pyx_n_s_custom_decode_value
#define __pyx_n_s_custom_encode_value __pyx_mstate_global->__pyx_n_s_custom_encode_value
#define __pyx_n_s_data __pyx_mstate_global->__pyx_n_s_data
#define __pyx_n_s_dataclass_fields __pyx_mstate_global->__pyx_n_s_dataclass_fields
#define __pyx_n_s_dataclass_params __pyx_mstate_global->__pyx_n_s_dataclass_params
#define __pyx_n_s_dataclasses __pyx_mstate_global->__pyx_n_s_dataclasses
#define __pyx_n_s_date __pyx_mstate_global->__pyx_n_s_date
#define __pyx_n_u_date __pyx_mstate_global->__pyx_n_u_date
#define __pyx_n_s_date_type_accepts_string __pyx_mstate_global->__pyx_n_s_date_type_accepts_string
#define __pyx_kp_u_date_type_accepts_string_2 __pyx_mstate_global->__pyx_kp_u_date_type_accepts_string_2
#define __pyx_n_s_datetime __pyx_mstate_global->__pyx_n_s_datetime
#define __pyx_n_s_days __pyx_mstate_global->__pyx_n_s_days
#define __pyx_n_s_decimal __pyx_mstate_global->__pyx_n_s_decimal
#define __pyx_n_u_decimal __pyx_mstate_global->__pyx_n_u_decimal
#define __pyx_kp_u_decimal_2 __pyx_mstate_global->__pyx_kp_u_decimal_2
#define __pyx_n_s_decimal_check_exp_overflow __pyx_mstate_global->__pyx_n_s_decimal_check_exp_overflow
#define __pyx_kp_u_decimal_check_exp_overflow_2 __pyx_mstate_global->__pyx_kp_u_decimal_check_exp_overflow_2
#define __pyx_n_s_decode __pyx_mstate_global->__pyx_n_s_decode
#define __pyx_n_s_decompress __pyx_mstate_global->__pyx_n_s_decompress
#define __pyx_n_s_decompressobj __pyx_mstate_global->__pyx_n_s_decompressobj
#define __pyx_n_s_defaul_value __pyx_mstate_global->__pyx_n_s_defaul_value
#define __pyx_n_s_default __pyx_mstate_global->__pyx_n_s_default
#define __pyx_n_u_default __pyx_mstate_global->__pyx_n_u_default
#define __pyx_n_s_default_value __pyx_mstate_global->__pyx_n_s_default_value
#define __pyx_n_s_defer_schema_promotion_errors __pyx_mstate_global->__pyx_n_s_defer_schema_promotion_errors
#define __pyx_kp_u_defer_schema_promotion_errors_2 __pyx_mstate_global->__pyx_kp_u_defer_schema_promotion_errors_2
#define __pyx_n_b_deflate __pyx_mstate_global->__pyx_n_b_deflate
#define __pyx_n_u_deflate __pyx_mstate_global->__pyx_n_u_deflate
#define __pyx_n_u_descending __pyx_mstate_global->__pyx_n_u_descending
#define __pyx_n_s_deserialize __pyx_mstate_global->__pyx_n_s_deserialize
#define __pyx_n_s_dest __pyx_mstate_global->__pyx_n_s_dest
#define __pyx_n_s_dest_type __pyx_mstate_global->__pyx_n_s_dest_type
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dict_2 __pyx_mstate_global->__pyx_n_s_dict_2
#define __pyx_kp_s_dict_str_AvroType __pyx_mstate_global->__pyx_kp_s_dict_str_AvroType
#define __pyx_kp_s_dict_str_tuple_object_object __pyx_mstate_global->__pyx_kp_s_dict_str_tuple_object_object
#define __pyx_n_s_dict_to_canonical_locals_genexpr __pyx_mstate_global->__pyx_n_s_dict_to_canonical_locals_genexpr
#define __pyx_n_s_diff __pyx_mstate_global->__pyx_n_s_diff
#define __pyx_n_s_digest __pyx_mstate_global->__pyx_n_s_digest
#define __pyx_n_s_digest_size __pyx_mstate_global->__pyx_n_s_digest_size
#define __pyx_n_s_dir __pyx_mstate_global->__pyx_n_s_dir
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_u_doc __pyx_mstate_global->__pyx_n_u_doc
#define __pyx_n_s_doc_2 __pyx_mstate_global->__pyx_n_s_doc_2
#define __pyx_kp_u_does_not_implement__binary_buff __pyx_mstate_global->__pyx_kp_u_does_not_implement__binary_buff
#define __pyx_kp_u_does_not_implement__convert_val __pyx_mstate_global->__pyx_kp_u_does_not_implement__convert_val
#define __pyx_kp_u_does_not_implement__extract_met __pyx_mstate_global->__pyx_kp_u_does_not_implement__extract_met
#define __pyx_kp_u_does_not_implement__get_schema __pyx_mstate_global->__pyx_kp_u_does_not_implement__get_schema
#define __pyx_kp_u_does_not_implement__get_value_f __pyx_mstate_global->__pyx_kp_u_does_not_implement__get_value_f
#define __pyx_kp_u_does_not_implement__json_decode __pyx_mstate_global->__pyx_kp_u_does_not_implement__json_decode
#define __pyx_kp_u_does_not_implement__json_format __pyx_mstate_global->__pyx_kp_u_does_not_implement__json_format
#define __pyx_kp_u_does_not_implement_binary_buffe __pyx_mstate_global->__pyx_kp_u_does_not_implement_binary_buffe
#define __pyx_kp_u_does_not_implement_canonical_fo __pyx_mstate_global->__pyx_kp_u_does_not_implement_canonical_fo
#define __pyx_kp_u_does_not_implement_clone __pyx_mstate_global->__pyx_kp_u_does_not_implement_clone
#define __pyx_kp_u_does_not_implement_decode_value __pyx_mstate_global->__pyx_kp_u_does_not_implement_decode_value
#define __pyx_kp_u_does_not_implement_encode_value __pyx_mstate_global->__pyx_kp_u_does_not_implement_encode_value
#define __pyx_kp_u_does_not_implement_read_block __pyx_mstate_global->__pyx_kp_u_does_not_implement_read_block
#define __pyx_kp_u_does_not_implement_read_bytes __pyx_mstate_global->__pyx_kp_u_does_not_implement_read_bytes
#define __pyx_kp_u_does_not_implement_read_n __pyx_mstate_global->__pyx_kp_u_does_not_implement_read_n
#define __pyx_kp_u_does_not_implement_read_u8 __pyx_mstate_global->__pyx_kp_u_does_not_implement_read_u8
#define __pyx_kp_u_does_not_implement_write_block __pyx_mstate_global->__pyx_kp_u_does_not_implement_write_block
#define __pyx_kp_u_does_not_implement_write_n __pyx_mstate_global->__pyx_kp_u_does_not_implement_write_n
#define __pyx_kp_u_does_not_implement_write_u8 __pyx_mstate_global->__pyx_kp_u_does_not_implement_write_u8
#define __pyx_kp_u_double __pyx_mstate_global->__pyx_kp_u_double
#define __pyx_n_u_double_2 __pyx_mstate_global->__pyx_n_u_double_2
#define __pyx_n_s_dtype_is_object __pyx_mstate_global->__pyx_n_s_dtype_is_object
#define __pyx_n_s_dumps __pyx_mstate_global->__pyx_n_s_dumps
#define __pyx_n_s_e __pyx_mstate_global->__pyx_n_s_e
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_encode __pyx_mstate_global->__pyx_n_s_encode
#define __pyx_n_s_enforce_enum_symbol_name_rules __pyx_mstate_global->__pyx_n_s_enforce_enum_symbol_name_rules
#define __pyx_kp_u_enforce_enum_symbol_name_rules_2 __pyx_mstate_global->__pyx_kp_u_enforce_enum_symbol_name_rules_2
#define __pyx_n_s_enforce_namespace_name_rules __pyx_mstate_global->__pyx_n_s_enforce_namespace_name_rules
#define __pyx_kp_u_enforce_namespace_name_rules_2 __pyx_mstate_global->__pyx_kp_u_enforce_namespace_name_rules_2
#define __pyx_n_s_enforce_type_name_rules __pyx_mstate_global->__pyx_n_s_enforce_type_name_rules
#define __pyx_kp_u_enforce_type_name_rules_2 __pyx_mstate_global->__pyx_kp_u_enforce_type_name_rules_2
#define __pyx_n_s_ensure_ascii __pyx_mstate_global->__pyx_n_s_ensure_ascii
#define __pyx_n_s_enter __pyx_mstate_global->__pyx_n_s_enter
#define __pyx_n_s_enum __pyx_mstate_global->__pyx_n_s_enum
#define __pyx_n_u_enum __pyx_mstate_global->__pyx_n_u_enum
#define __pyx_n_s_enum_symbols_must_be_unique __pyx_mstate_global->__pyx_n_s_enum_symbols_must_be_unique
#define __pyx_kp_u_enum_symbols_must_be_unique_2 __pyx_mstate_global->__pyx_kp_u_enum_symbols_must_be_unique_2
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_eq __pyx_mstate_global->__pyx_n_s_eq
#define __pyx_n_s_equals __pyx_mstate_global->__pyx_n_s_equals
#define __pyx_n_s_error __pyx_mstate_global->__pyx_n_s_error
#define __pyx_n_u_error __pyx_mstate_global->__pyx_n_u_error
#define __pyx_n_s_errors __pyx_mstate_global->__pyx_n_s_errors
#define __pyx_n_s_escape __pyx_mstate_global->__pyx_n_s_escape
#define __pyx_n_s_eval __pyx_mstate_global->__pyx_n_s_eval
#define __pyx_n_s_eval_2 __pyx_mstate_global->__pyx_n_s_eval_2
#define __pyx_n_s_exc_tb __pyx_mstate_global->__pyx_n_s_exc_tb
#define __pyx_n_s_exc_type __pyx_mstate_global->__pyx_n_s_exc_type
#define __pyx_n_s_exc_val __pyx_mstate_global->__pyx_n_s_exc_val
#define __pyx_n_s_exit __pyx_mstate_global->__pyx_n_s_exit
#define __pyx_n_s_expand_types_in_schema __pyx_mstate_global->__pyx_n_s_expand_types_in_schema
#define __pyx_kp_u_expand_types_in_schema_2 __pyx_mstate_global->__pyx_kp_u_expand_types_in_schema_2
#define __pyx_n_s_externally_defined_types __pyx_mstate_global->__pyx_n_s_externally_defined_types
#define __pyx_kp_u_externally_defined_types_2 __pyx_mstate_global->__pyx_kp_u_externally_defined_types_2
#define __pyx_n_s_extra __pyx_mstate_global->__pyx_n_s_extra
#define __pyx_n_s_f __pyx_mstate_global->__pyx_n_s_f
#define __pyx_n_s_field __pyx_mstate_global->__pyx_n_s_field
#define __pyx_n_s_field_to_index __pyx_mstate_global->__pyx_n_s_field_to_index
#define __pyx_n_u_field_to_index __pyx_mstate_global->__pyx_n_u_field_to_index
#define __pyx_n_s_field_type __pyx_mstate_global->__pyx_n_s_field_type
#define __pyx_n_s_fields __pyx_mstate_global->__pyx_n_s_fields
#define __pyx_n_u_fields __pyx_mstate_global->__pyx_n_u_fields
#define __pyx_n_s_file_obj __pyx_mstate_global->__pyx_n_s_file_obj
#define __pyx_n_s_find_type __pyx_mstate_global->__pyx_n_s_find_type
#define __pyx_n_s_fingerprint __pyx_mstate_global->__pyx_n_s_fingerprint
#define __pyx_n_s_fingerprint_returns_digest __pyx_mstate_global->__pyx_n_s_fingerprint_returns_digest
#define __pyx_kp_u_fingerprint_returns_digest_2 __pyx_mstate_global->__pyx_kp_u_fingerprint_returns_digest_2
#define __pyx_n_s_fitness __pyx_mstate_global->__pyx_n_s_fitness
#define __pyx_n_s_fixed __pyx_mstate_global->__pyx_n_s_fixed
#define __pyx_n_u_fixed __pyx_mstate_global->__pyx_n_u_fixed
#define __pyx_n_s_fixed_codec __pyx_mstate_global->__pyx_n_s_fixed_codec
#define __pyx_kp_u_fixed_codec_2 __pyx_mstate_global->__pyx_kp_u_fixed_codec_2
#define __pyx_kp_u_fixed_uuid __pyx_mstate_global->__pyx_kp_u_fixed_uuid
#define __pyx_n_s_flags __pyx_mstate_global->__pyx_n_s_flags
#define __pyx_kp_u_float __pyx_mstate_global->__pyx_kp_u_float
#define __pyx_n_s_float16 __pyx_mstate_global->__pyx_n_s_float16
#define __pyx_n_s_float32 __pyx_mstate_global->__pyx_n_s_float32
#define __pyx_n_s_float64 __pyx_mstate_global->__pyx_n_s_float64
#define __pyx_n_u_float_2 __pyx_mstate_global->__pyx_n_u_float_2
#define __pyx_n_s_float_out_of_range_inf __pyx_mstate_global->__pyx_n_s_float_out_of_range_inf
#define __pyx_kp_u_float_out_of_range_inf_2 __pyx_mstate_global->__pyx_kp_u_float_out_of_range_inf_2
#define __pyx_n_s_floor __pyx_mstate_global->__pyx_n_s_floor
#define __pyx_n_s_flush __pyx_mstate_global->__pyx_n_s_flush
#define __pyx_n_s_fn __pyx_mstate_global->__pyx_n_s_fn
#define __pyx_kp_u_for_field __pyx_mstate_global->__pyx_kp_u_for_field
#define __pyx_n_s_for_schema __pyx_mstate_global->__pyx_n_s_for_schema
#define __pyx_n_s_for_source __pyx_mstate_global->__pyx_n_s_for_source
#define __pyx_kp_u_for_type __pyx_mstate_global->__pyx_kp_u_for_type
#define __pyx_n_s_for_type_2 __pyx_mstate_global->__pyx_n_s_for_type_2
#define __pyx_n_s_for_type_3 __pyx_mstate_global->__pyx_n_s_for_type_3
#define __pyx_n_s_force __pyx_mstate_global->__pyx_n_s_force
#define __pyx_n_s_format __pyx_mstate_global->__pyx_n_s_format
#define __pyx_n_s_fortran __pyx_mstate_global->__pyx_n_s_fortran
#define __pyx_n_u_fortran __pyx_mstate_global->__pyx_n_u_fortran
#define __pyx_kp_u_found_twice __pyx_mstate_global->__pyx_kp_u_found_twice
#define __pyx_n_s_frame __pyx_mstate_global->__pyx_n_s_frame
#define __pyx_n_s_from_bytes __pyx_mstate_global->__pyx_n_s_from_bytes
#define __pyx_n_s_frozen __pyx_mstate_global->__pyx_n_s_frozen
#define __pyx_n_s_fullmatch __pyx_mstate_global->__pyx_n_s_fullmatch
#define __pyx_n_s_func __pyx_mstate_global->__pyx_n_s_func
#define __pyx_n_s_functools __pyx_mstate_global->__pyx_n_s_functools
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_genexpr __pyx_mstate_global->__pyx_n_s_genexpr
#define __pyx_n_s_get __pyx_mstate_global->__pyx_n_s_get
#define __pyx_n_s_get_2 __pyx_mstate_global->__pyx_n_s_get_2
#define __pyx_n_s_get_schema __pyx_mstate_global->__pyx_n_s_get_schema
#define __pyx_n_s_get_schema_extra __pyx_mstate_global->__pyx_n_s_get_schema_extra
#define __pyx_kp_u_getfullargspec __pyx_mstate_global->__pyx_kp_u_getfullargspec
#define __pyx_n_u_getitem __pyx_mstate_global->__pyx_n_u_getitem
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_s_given __pyx_mstate_global->__pyx_n_s_given
#define __pyx_kp_u_got __pyx_mstate_global->__pyx_kp_u_got
#define __pyx_kp_u_got_2 __pyx_mstate_global->__pyx_kp_u_got_2
#define __pyx_kp_u_got_3 __pyx_mstate_global->__pyx_kp_u_got_3
#define __pyx_kp_u_got_4 __pyx_mstate_global->__pyx_kp_u_got_4
#define __pyx_kp_u_got_differing_extents_in_dimensi __pyx_mstate_global->__pyx_kp_u_got_differing_extents_in_dimensi
#define __pyx_kp_u_has_exponent __pyx_mstate_global->__pyx_kp_u_has_exponent
#define __pyx_n_s_hash __pyx_mstate_global->__pyx_n_s_hash
#define __pyx_n_s_hasher __pyx_mstate_global->__pyx_n_s_hasher
#define __pyx_n_s_hashlib __pyx_mstate_global->__pyx_n_s_hashlib
#define __pyx_n_s_hex __pyx_mstate_global->__pyx_n_s_hex
#define __pyx_n_s_hexdigest __pyx_mstate_global->__pyx_n_s_hexdigest
#define __pyx_n_s_hour __pyx_mstate_global->__pyx_n_s_hour
#define __pyx_n_s_i __pyx_mstate_global->__pyx_n_s_i
#define __pyx_n_s_id __pyx_mstate_global->__pyx_n_s_id
#define __pyx_n_s_ignore __pyx_mstate_global->__pyx_n_s_ignore
#define __pyx_n_u_ignore __pyx_mstate_global->__pyx_n_u_ignore
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_indent __pyx_mstate_global->__pyx_n_s_indent
#define __pyx_n_s_index __pyx_mstate_global->__pyx_n_s_index
#define __pyx_kp_u_inf __pyx_mstate_global->__pyx_kp_u_inf
#define __pyx_n_u_inf_2 __pyx_mstate_global->__pyx_n_u_inf_2
#define __pyx_n_s_init __pyx_mstate_global->__pyx_n_s_init
#define __pyx_n_s_init_2 __pyx_mstate_global->__pyx_n_s_init_2
#define __pyx_n_s_init___locals_genexpr __pyx_mstate_global->__pyx_n_s_init___locals_genexpr
#define __pyx_n_s_init_subclass __pyx_mstate_global->__pyx_n_s_init_subclass
#define __pyx_n_s_initial_size __pyx_mstate_global->__pyx_n_s_initial_size
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_inline_namespaces __pyx_mstate_global->__pyx_n_s_inline_namespaces
#define __pyx_kp_u_inline_namespaces_2 __pyx_mstate_global->__pyx_kp_u_inline_namespaces_2
#define __pyx_n_s_inspect __pyx_mstate_global->__pyx_n_s_inspect
#define __pyx_n_s_inst __pyx_mstate_global->__pyx_n_s_inst
#define __pyx_kp_u_int __pyx_mstate_global->__pyx_kp_u_int
#define __pyx_n_u_int_2 __pyx_mstate_global->__pyx_n_u_int_2
#define __pyx_kp_u_int_date __pyx_mstate_global->__pyx_kp_u_int_date
#define __pyx_kp_u_int_time_millis __pyx_mstate_global->__pyx_kp_u_int_time_millis
#define __pyx_n_s_integer __pyx_mstate_global->__pyx_n_s_integer
#define __pyx_kp_u_invalid_for_enum __pyx_mstate_global->__pyx_kp_u_invalid_for_enum
#define __pyx_kp_u_invalid_value_include_array_ind __pyx_mstate_global->__pyx_kp_u_invalid_value_include_array_ind
#define __pyx_n_s_invalid_value_include_array_inde __pyx_mstate_global->__pyx_n_s_invalid_value_include_array_inde
#define __pyx_n_s_invalid_value_include_map_key __pyx_mstate_global->__pyx_n_s_invalid_value_include_map_key
#define __pyx_kp_u_invalid_value_include_map_key_2 __pyx_mstate_global->__pyx_kp_u_invalid_value_include_map_key_2
#define __pyx_kp_u_invalid_value_includes_record_n __pyx_mstate_global->__pyx_kp_u_invalid_value_includes_record_n
#define __pyx_n_s_invalid_value_includes_record_na __pyx_mstate_global->__pyx_n_s_invalid_value_includes_record_na
#define __pyx_n_s_io __pyx_mstate_global->__pyx_n_s_io
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_kp_u_is_not_a_mapping __pyx_mstate_global->__pyx_kp_u_is_not_a_mapping
#define __pyx_kp_u_is_not_a_valid_key_for_map_type __pyx_mstate_global->__pyx_kp_u_is_not_a_valid_key_for_map_type
#define __pyx_kp_u_is_not_a_valid_name __pyx_mstate_global->__pyx_kp_u_is_not_a_valid_name
#define __pyx_kp_u_is_not_a_valid_union_value_expe __pyx_mstate_global->__pyx_kp_u_is_not_a_valid_union_value_expe
#define __pyx_kp_u_is_not_a_valid_union_value_unkn __pyx_mstate_global->__pyx_kp_u_is_not_a_valid_union_value_unkn
#define __pyx_kp_u_is_not_a_valid_value_for_array __pyx_mstate_global->__pyx_kp_u_is_not_a_valid_value_for_array
#define __pyx_kp_u_is_not_a_valid_value_for_map_ty __pyx_mstate_global->__pyx_kp_u_is_not_a_valid_value_for_map_ty
#define __pyx_kp_u_is_not_allowed_as_a_name __pyx_mstate_global->__pyx_kp_u_is_not_allowed_as_a_name
#define __pyx_kp_u_is_not_valid __pyx_mstate_global->__pyx_kp_u_is_not_valid
#define __pyx_kp_u_is_not_valid_for_a_union_of __pyx_mstate_global->__pyx_kp_u_is_not_valid_for_a_union_of
#define __pyx_kp_u_is_not_valid_for_bytes_field __pyx_mstate_global->__pyx_kp_u_is_not_valid_for_bytes_field
#define __pyx_kp_u_is_not_valid_for_field __pyx_mstate_global->__pyx_kp_u_is_not_valid_for_field
#define __pyx_kp_u_is_not_valid_for_fixed_field __pyx_mstate_global->__pyx_kp_u_is_not_valid_for_fixed_field
#define __pyx_kp_u_is_not_valid_for_union_with_fir __pyx_mstate_global->__pyx_kp_u_is_not_valid_for_union_with_fir
#define __pyx_kp_u_is_required_for_record __pyx_mstate_global->__pyx_kp_u_is_required_for_record
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_isinf __pyx_mstate_global->__pyx_n_s_isinf
#define __pyx_n_s_isnan __pyx_mstate_global->__pyx_n_s_isnan
#define __pyx_n_s_item_type __pyx_mstate_global->__pyx_n_s_item_type
#define __pyx_n_s_items __pyx_mstate_global->__pyx_n_s_items
#define __pyx_n_u_items __pyx_mstate_global->__pyx_n_u_items
#define __pyx_kp_u_items_2 __pyx_mstate_global->__pyx_kp_u_items_2
#define __pyx_n_s_itemsize __pyx_mstate_global->__pyx_n_s_itemsize
#define __pyx_kp_s_itemsize_0_for_cython_array __pyx_mstate_global->__pyx_kp_s_itemsize_0_for_cython_array
#define __pyx_n_s_json __pyx_mstate_global->__pyx_n_s_json
#define __pyx_n_s_json_decode __pyx_mstate_global->__pyx_n_s_json_decode
#define __pyx_n_s_json_encode __pyx_mstate_global->__pyx_n_s_json_encode
#define __pyx_n_s_keys __pyx_mstate_global->__pyx_n_s_keys
#define __pyx_n_s_kw_only __pyx_mstate_global->__pyx_n_s_kw_only
#define __pyx_n_s_kwargs __pyx_mstate_global->__pyx_n_s_kwargs
#define __pyx_n_b_l __pyx_mstate_global->__pyx_n_b_l
#define __pyx_n_u_latin1 __pyx_mstate_global->__pyx_n_u_latin1
#define __pyx_kp_u_latin_1 __pyx_mstate_global->__pyx_kp_u_latin_1
#define __pyx_n_u_little __pyx_mstate_global->__pyx_n_u_little
#define __pyx_n_s_loads __pyx_mstate_global->__pyx_n_s_loads
#define __pyx_n_s_local __pyx_mstate_global->__pyx_n_s_local
#define __pyx_n_s_log10 __pyx_mstate_global->__pyx_n_s_log10
#define __pyx_n_u_logicalType __pyx_mstate_global->__pyx_n_u_logicalType
#define __pyx_n_s_logical_by_name __pyx_mstate_global->__pyx_n_s_logical_by_name
#define __pyx_n_s_logical_name __pyx_mstate_global->__pyx_n_s_logical_name
#define __pyx_n_s_logical_type __pyx_mstate_global->__pyx_n_s_logical_type
#define __pyx_n_s_logical_types __pyx_mstate_global->__pyx_n_s_logical_types
#define __pyx_kp_u_logical_types_2 __pyx_mstate_global->__pyx_kp_u_logical_types_2
#define __pyx_kp_u_long __pyx_mstate_global->__pyx_kp_u_long
#define __pyx_n_u_long_2 __pyx_mstate_global->__pyx_n_u_long_2
#define __pyx_kp_u_long_time_micros __pyx_mstate_global->__pyx_kp_u_long_time_micros
#define __pyx_kp_u_long_timestamp_micros __pyx_mstate_global->__pyx_kp_u_long_timestamp_micros
#define __pyx_kp_u_long_timestamp_millis __pyx_mstate_global->__pyx_kp_u_long_timestamp_millis
#define __pyx_n_b_lz4 __pyx_mstate_global->__pyx_n_b_lz4
#define __pyx_n_s_lz4 __pyx_mstate_global->__pyx_n_s_lz4
#define __pyx_n_u_lz4 __pyx_mstate_global->__pyx_n_u_lz4
#define __pyx_n_s_lz4_frame __pyx_mstate_global->__pyx_n_s_lz4_frame
#define __pyx_n_s_lzma __pyx_mstate_global->__pyx_n_s_lzma
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_make_dataclass __pyx_mstate_global->__pyx_n_s_make_dataclass
#define __pyx_n_s_make_logical_types __pyx_mstate_global->__pyx_n_s_make_logical_types
#define __pyx_n_s_map __pyx_mstate_global->__pyx_n_s_map
#define __pyx_n_u_map __pyx_mstate_global->__pyx_n_u_map
#define __pyx_n_s_marker __pyx_mstate_global->__pyx_n_s_marker
#define __pyx_n_s_mask __pyx_mstate_global->__pyx_n_s_mask
#define __pyx_n_s_match_args __pyx_mstate_global->__pyx_n_s_match_args
#define __pyx_n_s_math __pyx_mstate_global->__pyx_n_s_math
#define __pyx_n_s_max __pyx_mstate_global->__pyx_n_s_max
#define __pyx_n_s_max_blocksize __pyx_mstate_global->__pyx_n_s_max_blocksize
#define __pyx_n_s_max_precision __pyx_mstate_global->__pyx_n_s_max_precision
#define __pyx_n_s_maxunicode __pyx_mstate_global->__pyx_n_s_maxunicode
#define __pyx_n_s_memview __pyx_mstate_global->__pyx_n_s_memview
#define __pyx_n_s_meta __pyx_mstate_global->__pyx_n_s_meta
#define __pyx_n_s_metaclass __pyx_mstate_global->__pyx_n_s_metaclass
#define __pyx_n_s_metadata __pyx_mstate_global->__pyx_n_s_metadata
#define __pyx_n_s_method __pyx_mstate_global->__pyx_n_s_method
#define __pyx_n_s_micros_to_time __pyx_mstate_global->__pyx_n_s_micros_to_time
#define __pyx_n_s_microsecond __pyx_mstate_global->__pyx_n_s_microsecond
#define __pyx_n_s_microseconds __pyx_mstate_global->__pyx_n_s_microseconds
#define __pyx_n_s_minute __pyx_mstate_global->__pyx_n_s_minute
#define __pyx_kp_u_missing __pyx_mstate_global->__pyx_kp_u_missing
#define __pyx_kp_u_missing_2 __pyx_mstate_global->__pyx_kp_u_missing_2
#define __pyx_kp_u_missing_values_can_be_empty_con __pyx_mstate_global->__pyx_kp_u_missing_values_can_be_empty_con
#define __pyx_n_s_missing_values_can_be_empty_cont __pyx_mstate_global->__pyx_n_s_missing_values_can_be_empty_cont
#define __pyx_n_s_missing_values_can_be_null __pyx_mstate_global->__pyx_n_s_missing_values_can_be_null
#define __pyx_kp_u_missing_values_can_be_null_2 __pyx_mstate_global->__pyx_kp_u_missing_values_can_be_null_2
#define __pyx_n_s_mode __pyx_mstate_global->__pyx_n_s_mode
#define __pyx_n_s_module __pyx_mstate_global->__pyx_n_s_module
#define __pyx_n_s_mro_entries __pyx_mstate_global->__pyx_n_s_mro_entries
#define __pyx_n_s_msg __pyx_mstate_global->__pyx_n_s_msg
#define __pyx_n_s_n __pyx_mstate_global->__pyx_n_s_n
#define __pyx_n_s_n_bytes __pyx_mstate_global->__pyx_n_s_n_bytes
#define __pyx_n_s_naive_dt_assume_utc __pyx_mstate_global->__pyx_n_s_naive_dt_assume_utc
#define __pyx_kp_u_naive_dt_assume_utc_2 __pyx_mstate_global->__pyx_kp_u_naive_dt_assume_utc_2
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_u_name __pyx_mstate_global->__pyx_n_u_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_s_name_pattern __pyx_mstate_global->__pyx_n_s_name_pattern
#define __pyx_n_s_named_type_names_must_be_unique __pyx_mstate_global->__pyx_n_s_named_type_names_must_be_unique
#define __pyx_kp_u_named_type_names_must_be_unique_2 __pyx_mstate_global->__pyx_kp_u_named_type_names_must_be_unique_2
#define __pyx_n_s_named_types __pyx_mstate_global->__pyx_n_s_named_types
#define __pyx_n_s_namespace __pyx_mstate_global->__pyx_n_s_namespace
#define __pyx_n_u_namespace __pyx_mstate_global->__pyx_n_u_namespace
#define __pyx_n_s_namespaced __pyx_mstate_global->__pyx_n_s_namespaced
#define __pyx_n_u_nan __pyx_mstate_global->__pyx_n_u_nan
#define __pyx_n_s_ndarray __pyx_mstate_global->__pyx_n_s_ndarray
#define __pyx_n_s_ndim __pyx_mstate_global->__pyx_n_s_ndim
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_n_s_new_2 __pyx_mstate_global->__pyx_n_s_new_2
#define __pyx_n_s_new_inst __pyx_mstate_global->__pyx_n_s_new_inst
#define __pyx_n_s_new_types __pyx_mstate_global->__pyx_n_s_new_types
#define __pyx_n_s_new_ver __pyx_mstate_global->__pyx_n_s_new_ver
#define __pyx_n_s_next_object __pyx_mstate_global->__pyx_n_s_next_object
#define __pyx_kp_s_no_default___reduce___due_to_non __pyx_mstate_global->__pyx_kp_s_no_default___reduce___due_to_non
#define __pyx_kp_u_not_compatible_with_NullType __pyx_mstate_global->__pyx_kp_u_not_compatible_with_NullType
#define __pyx_kp_u_not_in_enum_symbols __pyx_mstate_global->__pyx_kp_u_not_in_enum_symbols
#define __pyx_kp_u_not_valid_for __pyx_mstate_global->__pyx_kp_u_not_valid_for
#define __pyx_kp_u_not_valid_for_long __pyx_mstate_global->__pyx_kp_u_not_valid_for_long
#define __pyx_n_s_np __pyx_mstate_global->__pyx_n_s_np
#define __pyx_n_s_np_f16 __pyx_mstate_global->__pyx_n_s_np_f16
#define __pyx_n_s_np_f32 __pyx_mstate_global->__pyx_n_s_np_f32
#define __pyx_n_s_np_f64 __pyx_mstate_global->__pyx_n_s_np_f64
#define __pyx_kp_u_ns __pyx_mstate_global->__pyx_kp_u_ns
#define __pyx_kp_u_ns_copy __pyx_mstate_global->__pyx_kp_u_ns_copy
#define __pyx_kp_u_null __pyx_mstate_global->__pyx_kp_u_null
#define __pyx_n_b_null_2 __pyx_mstate_global->__pyx_n_b_null_2
#define __pyx_n_u_null_2 __pyx_mstate_global->__pyx_n_u_null_2
#define __pyx_n_s_null_pad_fixed __pyx_mstate_global->__pyx_n_s_null_pad_fixed
#define __pyx_kp_u_null_pad_fixed_2 __pyx_mstate_global->__pyx_kp_u_null_pad_fixed_2
#define __pyx_n_s_numpy __pyx_mstate_global->__pyx_n_s_numpy
#define __pyx_n_s_ob __pyx_mstate_global->__pyx_n_s_ob
#define __pyx_n_s_obj __pyx_mstate_global->__pyx_n_s_obj
#define __pyx_n_s_object __pyx_mstate_global->__pyx_n_s_object
#define __pyx_n_s_objs __pyx_mstate_global->__pyx_n_s_objs
#define __pyx_n_s_open __pyx_mstate_global->__pyx_n_s_open
#define __pyx_n_s_options __pyx_mstate_global->__pyx_n_s_options
#define __pyx_n_s_order __pyx_mstate_global->__pyx_n_s_order
#define __pyx_n_u_order __pyx_mstate_global->__pyx_n_u_order
#define __pyx_n_s_orig_value __pyx_mstate_global->__pyx_n_s_orig_value
#define __pyx_n_s_os __pyx_mstate_global->__pyx_n_s_os
#define __pyx_n_s_other __pyx_mstate_global->__pyx_n_s_other
#define __pyx_n_s_other_val __pyx_mstate_global->__pyx_n_s_other_val
#define __pyx_n_s_out __pyx_mstate_global->__pyx_n_s_out
#define __pyx_n_s_out_state __pyx_mstate_global->__pyx_n_s_out_state
#define __pyx_n_s_pack __pyx_mstate_global->__pyx_n_s_pack
#define __pyx_n_s_pad __pyx_mstate_global->__pyx_n_s_pad
#define __pyx_n_s_parse_json __pyx_mstate_global->__pyx_n_s_parse_json
#define __pyx_n_s_parse_namespace __pyx_mstate_global->__pyx_n_s_parse_namespace
#define __pyx_n_s_partial __pyx_mstate_global->__pyx_n_s_partial
#define __pyx_n_s_path __pyx_mstate_global->__pyx_n_s_path
#define __pyx_n_s_path_str __pyx_mstate_global->__pyx_n_s_path_str
#define __pyx_n_s_pathlib __pyx_mstate_global->__pyx_n_s_pathlib
#define __pyx_n_s_perf __pyx_mstate_global->__pyx_n_s_perf
#define __pyx_n_s_perf_counter_ns __pyx_mstate_global->__pyx_n_s_perf_counter_ns
#define __pyx_n_s_perf_locals__test_zigzag_perf __pyx_mstate_global->__pyx_n_s_perf_locals__test_zigzag_perf
#define __pyx_n_s_perf_locals_add __pyx_mstate_global->__pyx_n_s_perf_locals_add
#define __pyx_n_s_perf_locals_add_locals_genexpr __pyx_mstate_global->__pyx_n_s_perf_locals_add_locals_genexpr
#define __pyx_n_s_perf_tests __pyx_mstate_global->__pyx_n_s_perf_tests
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_u_placeholder __pyx_mstate_global->__pyx_n_u_placeholder
#define __pyx_n_s_pop __pyx_mstate_global->__pyx_n_s_pop
#define __pyx_n_s_prec __pyx_mstate_global->__pyx_n_s_prec
#define __pyx_n_s_precision __pyx_mstate_global->__pyx_n_s_precision
#define __pyx_n_u_precision __pyx_mstate_global->__pyx_n_u_precision
#define __pyx_kp_u_precision_2 __pyx_mstate_global->__pyx_kp_u_precision_2
#define __pyx_n_s_prepare __pyx_mstate_global->__pyx_n_s_prepare
#define __pyx_n_s_print __pyx_mstate_global->__pyx_n_s_print
#define __pyx_n_s_pytest __pyx_mstate_global->__pyx_n_s_pytest
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_recursive_repr_guard __pyx_mstate_global->__pyx_n_s_pyx_recursive_repr_guard
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle_ArrayType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_ArrayType
#define __pyx_n_s_pyx_unpickle_AvroType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_AvroType
#define __pyx_n_s_pyx_unpickle_BoolType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_BoolType
#define __pyx_n_s_pyx_unpickle_BytesType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_BytesType
#define __pyx_n_s_pyx_unpickle_CannotPromote __pyx_mstate_global->__pyx_n_s_pyx_unpickle_CannotPromote
#define __pyx_n_s_pyx_unpickle_Codec __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Codec
#define __pyx_n_s_pyx_unpickle_ContainerReader __pyx_mstate_global->__pyx_n_s_pyx_unpickle_ContainerReader
#define __pyx_n_s_pyx_unpickle_CustomLogicalType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_CustomLogicalType
#define __pyx_n_s_pyx_unpickle_Date __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Date
#define __pyx_n_s_pyx_unpickle_DecimalType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_DecimalType
#define __pyx_n_s_pyx_unpickle_DoubleType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_DoubleType
#define __pyx_n_s_pyx_unpickle_Enum __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Enum
#define __pyx_n_s_pyx_unpickle_EnumType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_EnumType
#define __pyx_n_s_pyx_unpickle_FileReader __pyx_mstate_global->__pyx_n_s_pyx_unpickle_FileReader
#define __pyx_n_s_pyx_unpickle_FileWriter __pyx_mstate_global->__pyx_n_s_pyx_unpickle_FileWriter
#define __pyx_n_s_pyx_unpickle_FixedType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_FixedType
#define __pyx_n_s_pyx_unpickle_FloatType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_FloatType
#define __pyx_n_s_pyx_unpickle_IntType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_IntType
#define __pyx_n_s_pyx_unpickle_LogicalType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_LogicalType
#define __pyx_n_s_pyx_unpickle_LongType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_LongType
#define __pyx_n_s_pyx_unpickle_MapType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_MapType
#define __pyx_n_s_pyx_unpickle_MemoryReader __pyx_mstate_global->__pyx_n_s_pyx_unpickle_MemoryReader
#define __pyx_n_s_pyx_unpickle_NullType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_NullType
#define __pyx_n_s_pyx_unpickle_Options __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Options
#define __pyx_n_s_pyx_unpickle_PromoteBytesToStr __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PromoteBytesToStr
#define __pyx_n_s_pyx_unpickle_PromoteStringToBy __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PromoteStringToBy
#define __pyx_n_s_pyx_unpickle_PromoteToFloat __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PromoteToFloat
#define __pyx_n_s_pyx_unpickle_PromotingEnumType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PromotingEnumType
#define __pyx_n_s_pyx_unpickle_PromotingRecordTy __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PromotingRecordTy
#define __pyx_n_s_pyx_unpickle_Rabin __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Rabin
#define __pyx_n_s_pyx_unpickle_Record __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Record
#define __pyx_n_s_pyx_unpickle_RecordField __pyx_mstate_global->__pyx_n_s_pyx_unpickle_RecordField
#define __pyx_n_s_pyx_unpickle_RecordType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_RecordType
#define __pyx_n_s_pyx_unpickle_ResolvedSchema __pyx_mstate_global->__pyx_n_s_pyx_unpickle_ResolvedSchema
#define __pyx_n_s_pyx_unpickle_Schema __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Schema
#define __pyx_n_s_pyx_unpickle_StringType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_StringType
#define __pyx_n_s_pyx_unpickle_TimeMicros __pyx_mstate_global->__pyx_n_s_pyx_unpickle_TimeMicros
#define __pyx_n_s_pyx_unpickle_TimeMillis __pyx_mstate_global->__pyx_n_s_pyx_unpickle_TimeMillis
#define __pyx_n_s_pyx_unpickle_TimestampMicros __pyx_mstate_global->__pyx_n_s_pyx_unpickle_TimestampMicros
#define __pyx_n_s_pyx_unpickle_TimestampMillis __pyx_mstate_global->__pyx_n_s_pyx_unpickle_TimestampMillis
#define __pyx_n_s_pyx_unpickle_UUIDBase __pyx_mstate_global->__pyx_n_s_pyx_unpickle_UUIDBase
#define __pyx_n_s_pyx_unpickle_UUIDFixedType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_UUIDFixedType
#define __pyx_n_s_pyx_unpickle_UUIDStringType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_UUIDStringType
#define __pyx_n_s_pyx_unpickle_UnionType __pyx_mstate_global->__pyx_n_s_pyx_unpickle_UnionType
#define __pyx_n_s_pyx_unpickle_ValueAdapter __pyx_mstate_global->__pyx_n_s_pyx_unpickle_ValueAdapter
#define __pyx_n_s_pyx_unpickle__CanonicalForm __pyx_mstate_global->__pyx_n_s_pyx_unpickle__CanonicalForm
#define __pyx_n_s_pyx_unpickle__DeflateCodec __pyx_mstate_global->__pyx_n_s_pyx_unpickle__DeflateCodec
#define __pyx_n_s_pyx_unpickle__FieldAccessor __pyx_mstate_global->__pyx_n_s_pyx_unpickle__FieldAccessor
#define __pyx_n_s_pyx_unpickle__Lz4Codec __pyx_mstate_global->__pyx_n_s_pyx_unpickle__Lz4Codec
#define __pyx_n_s_pyx_unpickle__LzmaCodec __pyx_mstate_global->__pyx_n_s_pyx_unpickle__LzmaCodec
#define __pyx_n_s_pyx_unpickle__NamedType __pyx_mstate_global->__pyx_n_s_pyx_unpickle__NamedType
#define __pyx_n_s_pyx_unpickle__NullCodec __pyx_mstate_global->__pyx_n_s_pyx_unpickle__NullCodec
#define __pyx_n_s_pyx_unpickle__PlaceholderType __pyx_mstate_global->__pyx_n_s_pyx_unpickle__PlaceholderType
#define __pyx_n_s_pyx_unpickle__Reader __pyx_mstate_global->__pyx_n_s_pyx_unpickle__Reader
#define __pyx_n_s_pyx_unpickle__Sentinel __pyx_mstate_global->__pyx_n_s_pyx_unpickle__Sentinel
#define __pyx_n_s_pyx_unpickle__SnappyCodec __pyx_mstate_global->__pyx_n_s_pyx_unpickle__SnappyCodec
#define __pyx_n_s_pyx_unpickle__Writer __pyx_mstate_global->__pyx_n_s_pyx_unpickle__Writer
#define __pyx_n_s_pyx_unpickle__ZStandardCodec __pyx_mstate_global->__pyx_n_s_pyx_unpickle__ZStandardCodec
#define __pyx_n_s_pyx_unpickle___Bzip2Codec __pyx_mstate_global->__pyx_n_s_pyx_unpickle___Bzip2Codec
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_b_q __pyx_mstate_global->__pyx_n_b_q
#define __pyx_n_s_qualname __pyx_mstate_global->__pyx_n_s_qualname
#define __pyx_n_s_quantize __pyx_mstate_global->__pyx_n_s_quantize
#define __pyx_n_u_rabin __pyx_mstate_global->__pyx_n_u_rabin
#define __pyx_n_s_raise __pyx_mstate_global->__pyx_n_s_raise
#define __pyx_n_s_raise_on_invalid_logical __pyx_mstate_global->__pyx_n_s_raise_on_invalid_logical
#define __pyx_kp_u_raise_on_invalid_logical_2 __pyx_mstate_global->__pyx_kp_u_raise_on_invalid_logical_2
#define __pyx_n_s_raises __pyx_mstate_global->__pyx_n_s_raises
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_u_rb __pyx_mstate_global->__pyx_n_u_rb
#define __pyx_n_s_re __pyx_mstate_global->__pyx_n_s_re
#define __pyx_n_s_read __pyx_mstate_global->__pyx_n_s_read
#define __pyx_n_u_read __pyx_mstate_global->__pyx_n_u_read
#define __pyx_n_s_read_marker __pyx_mstate_global->__pyx_n_s_read_marker
#define __pyx_n_s_read_perf __pyx_mstate_global->__pyx_n_s_read_perf
#define __pyx_n_s_read_time __pyx_mstate_global->__pyx_n_s_read_time
#define __pyx_n_s_reader __pyx_mstate_global->__pyx_n_s_reader
#define __pyx_n_s_reader_for_writer __pyx_mstate_global->__pyx_n_s_reader_for_writer
#define __pyx_n_s_reader_schema __pyx_mstate_global->__pyx_n_s_reader_schema
#define __pyx_n_s_reader_type __pyx_mstate_global->__pyx_n_s_reader_type
#define __pyx_n_s_record __pyx_mstate_global->__pyx_n_s_record
#define __pyx_n_u_record __pyx_mstate_global->__pyx_n_u_record
#define __pyx_n_s_record_allow_extra_fields __pyx_mstate_global->__pyx_n_s_record_allow_extra_fields
#define __pyx_kp_u_record_allow_extra_fields_2 __pyx_mstate_global->__pyx_kp_u_record_allow_extra_fields_2
#define __pyx_n_s_record_can_encode_dict __pyx_mstate_global->__pyx_n_s_record_can_encode_dict
#define __pyx_kp_u_record_can_encode_dict_2 __pyx_mstate_global->__pyx_kp_u_record_can_encode_dict_2
#define __pyx_n_s_record_decodes_to_dict __pyx_mstate_global->__pyx_n_s_record_decodes_to_dict
#define __pyx_kp_u_record_decodes_to_dict_2 __pyx_mstate_global->__pyx_kp_u_record_decodes_to_dict_2
#define __pyx_n_s_record_encode_use_defaults __pyx_mstate_global->__pyx_n_s_record_encode_use_defaults
#define __pyx_kp_u_record_encode_use_defaults_2 __pyx_mstate_global->__pyx_kp_u_record_encode_use_defaults_2
#define __pyx_n_s_record_fields_must_be_unique __pyx_mstate_global->__pyx_n_s_record_fields_must_be_unique
#define __pyx_kp_u_record_fields_must_be_unique_2 __pyx_mstate_global->__pyx_kp_u_record_fields_must_be_unique_2
#define __pyx_n_s_record_values_type_hint __pyx_mstate_global->__pyx_n_s_record_values_type_hint
#define __pyx_kp_u_record_values_type_hint_2 __pyx_mstate_global->__pyx_kp_u_record_values_type_hint_2
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_register __pyx_mstate_global->__pyx_n_s_register
#define __pyx_n_s_remove __pyx_mstate_global->__pyx_n_s_remove
#define __pyx_n_s_replace __pyx_mstate_global->__pyx_n_s_replace
#define __pyx_n_u_replace __pyx_mstate_global->__pyx_n_u_replace
#define __pyx_n_s_repr __pyx_mstate_global->__pyx_n_s_repr
#define __pyx_kp_u_required_field __pyx_mstate_global->__pyx_kp_u_required_field
#define __pyx_n_s_resolved_type __pyx_mstate_global->__pyx_n_s_resolved_type
#define __pyx_n_s_return __pyx_mstate_global->__pyx_n_s_return
#define __pyx_n_s_return_uuid_object __pyx_mstate_global->__pyx_n_s_return_uuid_object
#define __pyx_kp_u_return_uuid_object_2 __pyx_mstate_global->__pyx_kp_u_return_uuid_object_2
#define __pyx_n_s_reversed __pyx_mstate_global->__pyx_n_s_reversed
#define __pyx_n_s_rjust __pyx_mstate_global->__pyx_n_s_rjust
#define __pyx_n_s_rsplit __pyx_mstate_global->__pyx_n_s_rsplit
#define __pyx_n_s_running __pyx_mstate_global->__pyx_n_s_running
#define __pyx_kp_u_s_read __pyx_mstate_global->__pyx_kp_u_s_read
#define __pyx_n_s_scale __pyx_mstate_global->__pyx_n_s_scale
#define __pyx_n_u_scale __pyx_mstate_global->__pyx_n_u_scale
#define __pyx_n_s_scaleb __pyx_mstate_global->__pyx_n_s_scaleb
#define __pyx_n_s_schema __pyx_mstate_global->__pyx_n_s_schema
#define __pyx_n_s_schema_ob __pyx_mstate_global->__pyx_n_s_schema_ob
#define __pyx_n_s_schema_path __pyx_mstate_global->__pyx_n_s_schema_path
#define __pyx_n_s_second __pyx_mstate_global->__pyx_n_s_second
#define __pyx_n_s_seconds __pyx_mstate_global->__pyx_n_s_seconds
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_n_s_self_val __pyx_mstate_global->__pyx_n_s_self_val
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_n_s_serialize __pyx_mstate_global->__pyx_n_s_serialize
#define __pyx_n_s_set_from_dataclass __pyx_mstate_global->__pyx_n_s_set_from_dataclass
#define __pyx_n_s_set_name __pyx_mstate_global->__pyx_n_s_set_name
#define __pyx_n_s_setdefault __pyx_mstate_global->__pyx_n_s_setdefault
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_shape __pyx_mstate_global->__pyx_n_s_shape
#define __pyx_n_s_signature __pyx_mstate_global->__pyx_n_s_signature
#define __pyx_n_s_signature_2 __pyx_mstate_global->__pyx_n_s_signature_2
#define __pyx_n_s_signed __pyx_mstate_global->__pyx_n_s_signed
#define __pyx_n_s_size __pyx_mstate_global->__pyx_n_s_size
#define __pyx_n_u_size __pyx_mstate_global->__pyx_n_u_size
#define __pyx_n_u_slots __pyx_mstate_global->__pyx_n_u_slots
#define __pyx_n_s_slots_2 __pyx_mstate_global->__pyx_n_s_slots_2
#define __pyx_n_b_snappy __pyx_mstate_global->__pyx_n_b_snappy
#define __pyx_n_s_snappy __pyx_mstate_global->__pyx_n_s_snappy
#define __pyx_n_u_snappy __pyx_mstate_global->__pyx_n_u_snappy
#define __pyx_n_s_source __pyx_mstate_global->__pyx_n_s_source
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_s_src __pyx_mstate_global->__pyx_n_s_src
#define __pyx_kp_s_src_array_pxi __pyx_mstate_global->__pyx_kp_s_src_array_pxi
#define __pyx_kp_s_src_container_pxi __pyx_mstate_global->__pyx_kp_s_src_container_pxi
#define __pyx_kp_s_src_enum_pxi __pyx_mstate_global->__pyx_kp_s_src_enum_pxi
#define __pyx_kp_s_src_error_pxi __pyx_mstate_global->__pyx_kp_s_src_error_pxi
#define __pyx_kp_s_src_logical_pxi __pyx_mstate_global->__pyx_kp_s_src_logical_pxi
#define __pyx_kp_s_src_map_pxi __pyx_mstate_global->__pyx_kp_s_src_map_pxi
#define __pyx_kp_s_src_null_pxi __pyx_mstate_global->__pyx_kp_s_src_null_pxi
#define __pyx_kp_s_src_numeric_types_pxi __pyx_mstate_global->__pyx_kp_s_src_numeric_types_pxi
#define __pyx_kp_s_src_option_pxi __pyx_mstate_global->__pyx_kp_s_src_option_pxi
#define __pyx_kp_s_src_rabin_pxi __pyx_mstate_global->__pyx_kp_s_src_rabin_pxi
#define __pyx_kp_s_src_record_pxi __pyx_mstate_global->__pyx_kp_s_src_record_pxi
#define __pyx_kp_s_src_schema_pxi __pyx_mstate_global->__pyx_kp_s_src_schema_pxi
#define __pyx_kp_s_src_string_types_pxi __pyx_mstate_global->__pyx_kp_s_src_string_types_pxi
#define __pyx_kp_s_src_tests_test_buffer_pxi __pyx_mstate_global->__pyx_kp_s_src_tests_test_buffer_pxi
#define __pyx_kp_s_src_tests_test_perf_pxi __pyx_mstate_global->__pyx_kp_s_src_tests_test_perf_pxi
#define __pyx_kp_s_src_tests_test_zigzag_pxi __pyx_mstate_global->__pyx_kp_s_src_tests_test_zigzag_pxi
#define __pyx_kp_s_src_type_pxi __pyx_mstate_global->__pyx_kp_s_src_type_pxi
#define __pyx_kp_s_src_union_pxi __pyx_mstate_global->__pyx_kp_s_src_union_pxi
#define __pyx_n_s_start __pyx_mstate_global->__pyx_n_s_start
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_s_step __pyx_mstate_global->__pyx_n_s_step
#define __pyx_n_s_stop __pyx_mstate_global->__pyx_n_s_stop
#define __pyx_n_s_str __pyx_mstate_global->__pyx_n_s_str
#define __pyx_n_s_str_2 __pyx_mstate_global->__pyx_n_s_str_2
#define __pyx_n_u_strict __pyx_mstate_global->__pyx_n_u_strict
#define __pyx_kp_s_strided_and_direct __pyx_mstate_global->__pyx_kp_s_strided_and_direct
#define __pyx_kp_s_strided_and_direct_or_indirect __pyx_mstate_global->__pyx_kp_s_strided_and_direct_or_indirect
#define __pyx_kp_s_strided_and_indirect __pyx_mstate_global->__pyx_kp_s_strided_and_indirect
#define __pyx_n_u_string __pyx_mstate_global->__pyx_n_u_string
#define __pyx_kp_u_string_2 __pyx_mstate_global->__pyx_kp_u_string_2
#define __pyx_n_s_string_types_default_unchanged __pyx_mstate_global->__pyx_n_s_string_types_default_unchanged
#define __pyx_kp_u_string_types_default_unchanged_2 __pyx_mstate_global->__pyx_kp_u_string_types_default_unchanged_2
#define __pyx_kp_u_string_uuid __pyx_mstate_global->__pyx_kp_u_string_uuid
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_strptime __pyx_mstate_global->__pyx_n_s_strptime
#define __pyx_n_s_struct __pyx_mstate_global->__pyx_n_s_struct
#define __pyx_n_s_sum __pyx_mstate_global->__pyx_n_s_sum
#define __pyx_n_s_super __pyx_mstate_global->__pyx_n_s_super
#define __pyx_n_s_symbols __pyx_mstate_global->__pyx_n_s_symbols
#define __pyx_n_u_symbols __pyx_mstate_global->__pyx_n_u_symbols
#define __pyx_n_s_sys __pyx_mstate_global->__pyx_n_s_sys
#define __pyx_n_s_t __pyx_mstate_global->__pyx_n_s_t
#define __pyx_n_s_t_a __pyx_mstate_global->__pyx_n_s_t_a
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_test_2 __pyx_mstate_global->__pyx_n_s_test_2
#define __pyx_n_s_test_reader_read_n __pyx_mstate_global->__pyx_n_s_test_reader_read_n
#define __pyx_n_s_test_reader_read_u8 __pyx_mstate_global->__pyx_n_s_test_reader_read_u8
#define __pyx_n_s_test_readvarint __pyx_mstate_global->__pyx_n_s_test_readvarint
#define __pyx_n_s_test_readvarlong __pyx_mstate_global->__pyx_n_s_test_readvarlong
#define __pyx_n_s_test_zigzag_long __pyx_mstate_global->__pyx_n_s_test_zigzag_long
#define __pyx_n_s_test_zigzag_perf __pyx_mstate_global->__pyx_n_s_test_zigzag_perf
#define __pyx_n_s_tests __pyx_mstate_global->__pyx_n_s_tests
#define __pyx_n_s_tests_2 __pyx_mstate_global->__pyx_n_s_tests_2
#define __pyx_n_s_tests_locals__test_reader_read __pyx_mstate_global->__pyx_n_s_tests_locals__test_reader_read
#define __pyx_n_s_tests_locals__test_reader_read_2 __pyx_mstate_global->__pyx_n_s_tests_locals__test_reader_read_2
#define __pyx_n_s_tests_locals_add __pyx_mstate_global->__pyx_n_s_tests_locals_add
#define __pyx_n_s_tests_locals_add_locals_genexpr __pyx_mstate_global->__pyx_n_s_tests_locals_add_locals_genexpr
#define __pyx_n_s_tests_locals_test_readvarint __pyx_mstate_global->__pyx_n_s_tests_locals_test_readvarint
#define __pyx_n_s_tests_locals_test_readvarlong __pyx_mstate_global->__pyx_n_s_tests_locals_test_readvarlong
#define __pyx_n_s_tests_locals_test_zigzag_long __pyx_mstate_global->__pyx_n_s_tests_locals_test_zigzag_long
#define __pyx_n_s_threading __pyx_mstate_global->__pyx_n_s_threading
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_time __pyx_mstate_global->__pyx_n_s_time
#define __pyx_kp_u_time_micros __pyx_mstate_global->__pyx_kp_u_time_micros
#define __pyx_kp_u_time_millis __pyx_mstate_global->__pyx_kp_u_time_millis
#define __pyx_n_s_time_to_micros __pyx_mstate_global->__pyx_n_s_time_to_micros
#define __pyx_n_s_timedelta __pyx_mstate_global->__pyx_n_s_timedelta
#define __pyx_n_s_timestamp __pyx_mstate_global->__pyx_n_s_timestamp
#define __pyx_kp_u_timestamp_micros __pyx_mstate_global->__pyx_kp_u_timestamp_micros
#define __pyx_kp_u_timestamp_millis __pyx_mstate_global->__pyx_kp_u_timestamp_millis
#define __pyx_n_s_timezone __pyx_mstate_global->__pyx_n_s_timezone
#define __pyx_kp_u_to __pyx_mstate_global->__pyx_kp_u_to
#define __pyx_n_s_to_bytes __pyx_mstate_global->__pyx_n_s_to_bytes
#define __pyx_n_s_total_nums __pyx_mstate_global->__pyx_n_s_total_nums
#define __pyx_n_s_truncate_fixed __pyx_mstate_global->__pyx_n_s_truncate_fixed
#define __pyx_kp_u_truncate_fixed_2 __pyx_mstate_global->__pyx_kp_u_truncate_fixed_2
#define __pyx_kp_s_tuple_LogicalType __pyx_mstate_global->__pyx_kp_s_tuple_LogicalType
#define __pyx_n_s_type __pyx_mstate_global->__pyx_n_s_type
#define __pyx_n_u_type __pyx_mstate_global->__pyx_n_u_type
#define __pyx_kp_u_type_2 __pyx_mstate_global->__pyx_kp_u_type_2
#define __pyx_n_s_type_3 __pyx_mstate_global->__pyx_n_s_type_3
#define __pyx_n_s_type_name __pyx_mstate_global->__pyx_n_s_type_name
#define __pyx_n_s_type_str __pyx_mstate_global->__pyx_n_s_type_str
#define __pyx_n_s_types __pyx_mstate_global->__pyx_n_s_types
#define __pyx_n_s_types_str_to_schema __pyx_mstate_global->__pyx_n_s_types_str_to_schema
#define __pyx_kp_u_types_str_to_schema_2 __pyx_mstate_global->__pyx_kp_u_types_str_to_schema_2
#define __pyx_n_s_typing __pyx_mstate_global->__pyx_n_s_typing
#define __pyx_n_s_tzinfo __pyx_mstate_global->__pyx_n_s_tzinfo
#define __pyx_n_s_ubin __pyx_mstate_global->__pyx_n_s_ubin
#define __pyx_kp_s_unable_to_allocate_array_data __pyx_mstate_global->__pyx_kp_s_unable_to_allocate_array_data
#define __pyx_kp_s_unable_to_allocate_shape_and_str __pyx_mstate_global->__pyx_kp_s_unable_to_allocate_shape_and_str
#define __pyx_n_s_underlying __pyx_mstate_global->__pyx_n_s_underlying
#define __pyx_n_s_underlying_name __pyx_mstate_global->__pyx_n_s_underlying_name
#define __pyx_n_s_underlying_type __pyx_mstate_global->__pyx_n_s_underlying_type
#define __pyx_n_s_underlying_types __pyx_mstate_global->__pyx_n_s_underlying_types
#define __pyx_n_s_unicode_errors __pyx_mstate_global->__pyx_n_s_unicode_errors
#define __pyx_kp_u_unicode_errors_2 __pyx_mstate_global->__pyx_kp_u_unicode_errors_2
#define __pyx_n_s_unicodedata __pyx_mstate_global->__pyx_n_s_unicodedata
#define __pyx_n_u_union __pyx_mstate_global->__pyx_n_u_union
#define __pyx_n_s_union_decodes_to __pyx_mstate_global->__pyx_n_s_union_decodes_to
#define __pyx_kp_u_union_decodes_to_2 __pyx_mstate_global->__pyx_kp_u_union_decodes_to_2
#define __pyx_n_s_union_json_encodes_type_name __pyx_mstate_global->__pyx_n_s_union_json_encodes_type_name
#define __pyx_kp_u_union_json_encodes_type_name_2 __pyx_mstate_global->__pyx_kp_u_union_json_encodes_type_name_2
#define __pyx_n_s_unpack __pyx_mstate_global->__pyx_n_s_unpack
#define __pyx_n_s_unsafe_hash __pyx_mstate_global->__pyx_n_s_unsafe_hash
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_use_setstate __pyx_mstate_global->__pyx_n_s_use_setstate
#define __pyx_n_s_utc __pyx_mstate_global->__pyx_n_s_utc
#define __pyx_n_u_utf8 __pyx_mstate_global->__pyx_n_u_utf8
#define __pyx_kp_u_utf_8 __pyx_mstate_global->__pyx_kp_u_utf_8
#define __pyx_n_s_uuid __pyx_mstate_global->__pyx_n_s_uuid
#define __pyx_n_u_uuid __pyx_mstate_global->__pyx_n_u_uuid
#define __pyx_n_s_uuid4 __pyx_mstate_global->__pyx_n_s_uuid4
#define __pyx_n_s_val __pyx_mstate_global->__pyx_n_s_val
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_kp_u_value_2 __pyx_mstate_global->__pyx_kp_u_value_2
#define __pyx_n_s_value_adapters __pyx_mstate_global->__pyx_n_s_value_adapters
#define __pyx_n_s_value_type __pyx_mstate_global->__pyx_n_s_value_type
#define __pyx_n_u_values __pyx_mstate_global->__pyx_n_u_values
#define __pyx_n_s_version __pyx_mstate_global->__pyx_n_s_version
#define __pyx_n_s_version_info __pyx_mstate_global->__pyx_n_s_version_info
#define __pyx_n_s_visited __pyx_mstate_global->__pyx_n_s_visited
#define __pyx_kp_u_w __pyx_mstate_global->__pyx_kp_u_w
#define __pyx_n_s_walk_types __pyx_mstate_global->__pyx_n_s_walk_types
#define __pyx_n_s_warn __pyx_mstate_global->__pyx_n_s_warn
#define __pyx_n_s_warnings __pyx_mstate_global->__pyx_n_s_warnings
#define __pyx_n_u_wb __pyx_mstate_global->__pyx_n_u_wb
#define __pyx_n_s_wbits __pyx_mstate_global->__pyx_n_s_wbits
#define __pyx_n_s_weakref_slot __pyx_mstate_global->__pyx_n_s_weakref_slot
#define __pyx_kp_u_which_is_greater_than_the_scale __pyx_mstate_global->__pyx_kp_u_which_is_greater_than_the_scale
#define __pyx_n_s_with_external_types __pyx_mstate_global->__pyx_n_s_with_external_types
#define __pyx_n_s_with_logical_types __pyx_mstate_global->__pyx_n_s_with_logical_types
#define __pyx_n_s_wrap_type __pyx_mstate_global->__pyx_n_s_wrap_type
#define __pyx_n_s_wrapped __pyx_mstate_global->__pyx_n_s_wrapped
#define __pyx_n_s_write __pyx_mstate_global->__pyx_n_s_write
#define __pyx_n_u_write __pyx_mstate_global->__pyx_n_u_write
#define __pyx_kp_u_write_2 __pyx_mstate_global->__pyx_kp_u_write_2
#define __pyx_n_s_write_header __pyx_mstate_global->__pyx_n_s_write_header
#define __pyx_n_s_write_many __pyx_mstate_global->__pyx_n_s_write_many
#define __pyx_n_s_write_one __pyx_mstate_global->__pyx_n_s_write_one
#define __pyx_n_s_write_perf __pyx_mstate_global->__pyx_n_s_write_perf
#define __pyx_n_s_write_time __pyx_mstate_global->__pyx_n_s_write_time
#define __pyx_n_s_writer __pyx_mstate_global->__pyx_n_s_writer
#define __pyx_n_s_writer_schema __pyx_mstate_global->__pyx_n_s_writer_schema
#define __pyx_n_s_writer_type __pyx_mstate_global->__pyx_n_s_writer_type
#define __pyx_n_b_xz __pyx_mstate_global->__pyx_n_b_xz
#define __pyx_n_u_xz __pyx_mstate_global->__pyx_n_u_xz
#define __pyx_n_s_zip __pyx_mstate_global->__pyx_n_s_zip
#define __pyx_n_s_zlib __pyx_mstate_global->__pyx_n_s_zlib
#define __pyx_n_b_zstandard __pyx_mstate_global->__pyx_n_b_zstandard
#define __pyx_n_s_zstandard __pyx_mstate_global->__pyx_n_s_zstandard
#define __pyx_n_u_zstandard __pyx_mstate_global->__pyx_n_u_zstandard
#define __pyx_float_0_0 __pyx_mstate_global->__pyx_float_0_0
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_2 __pyx_mstate_global->__pyx_int_2
#define __pyx_int_3 __pyx_mstate_global->__pyx_int_3
#define __pyx_int_4 __pyx_mstate_global->__pyx_int_4
#define __pyx_int_8 __pyx_mstate_global->__pyx_int_8
#define __pyx_int_10 __pyx_mstate_global->__pyx_int_10
#define __pyx_int_16 __pyx_mstate_global->__pyx_int_16
#define __pyx_int_24 __pyx_mstate_global->__pyx_int_24
#define __pyx_int_60 __pyx_mstate_global->__pyx_int_60
#define __pyx_int_70 __pyx_mstate_global->__pyx_int_70
#define __pyx_int_127 __pyx_mstate_global->__pyx_int_127
#define __pyx_int_255 __pyx_mstate_global->__pyx_int_255
#define __pyx_int_1000 __pyx_mstate_global->__pyx_int_1000
#define __pyx_int_1970 __pyx_mstate_global->__pyx_int_1970
#define __pyx_int_3600 __pyx_mstate_global->__pyx_int_3600
#define __pyx_int_4096 __pyx_mstate_global->__pyx_int_4096
#define __pyx_int_32767 __pyx_mstate_global->__pyx_int_32767
#define __pyx_int_65536 __pyx_mstate_global->__pyx_int_65536
#define __pyx_int_1000000 __pyx_mstate_global->__pyx_int_1000000
#define __pyx_int_8388607 __pyx_mstate_global->__pyx_int_8388607
#define __pyx_int_18332779 __pyx_mstate_global->__pyx_int_18332779
#define __pyx_int_24827247 __pyx_mstate_global->__pyx_int_24827247
#define __pyx_int_25341170 __pyx_mstate_global->__pyx_int_25341170
#define __pyx_int_29099189 __pyx_mstate_global->__pyx_int_29099189
#define __pyx_int_31039736 __pyx_mstate_global->__pyx_int_31039736
#define __pyx_int_33963030 __pyx_mstate_global->__pyx_int_33963030
#define __pyx_int_38410757 __pyx_mstate_global->__pyx_int_38410757
#define __pyx_int_38632982 __pyx_mstate_global->__pyx_int_38632982
#define __pyx_int_39872412 __pyx_mstate_global->__pyx_int_39872412
#define __pyx_int_42565757 __pyx_mstate_global->__pyx_int_42565757
#define __pyx_int_47772561 __pyx_mstate_global->__pyx_int_47772561
#define __pyx_int_60000000 __pyx_mstate_global->__pyx_int_60000000
#define __pyx_int_61270791 __pyx_mstate_global->__pyx_int_61270791
#define __pyx_int_61440409 __pyx_mstate_global->__pyx_int_61440409
#define __pyx_int_61765448 __pyx_mstate_global->__pyx_int_61765448
#define __pyx_int_67877419 __pyx_mstate_global->__pyx_int_67877419
#define __pyx_int_75337510 __pyx_mstate_global->__pyx_int_75337510
#define __pyx_int_79389776 __pyx_mstate_global->__pyx_int_79389776
#define __pyx_int_79941127 __pyx_mstate_global->__pyx_int_79941127
#define __pyx_int_81233721 __pyx_mstate_global->__pyx_int_81233721
#define __pyx_int_83351333 __pyx_mstate_global->__pyx_int_83351333
#define __pyx_int_95761985 __pyx_mstate_global->__pyx_int_95761985
#define __pyx_int_96605483 __pyx_mstate_global->__pyx_int_96605483
#define __pyx_int_101034655 __pyx_mstate_global->__pyx_int_101034655
#define __pyx_int_101670942 __pyx_mstate_global->__pyx_int_101670942
#define __pyx_int_111776469 __pyx_mstate_global->__pyx_int_111776469
#define __pyx_int_111957047 __pyx_mstate_global->__pyx_int_111957047
#define __pyx_int_112105877 __pyx_mstate_global->__pyx_int_112105877
#define __pyx_int_115671040 __pyx_mstate_global->__pyx_int_115671040
#define __pyx_int_119832798 __pyx_mstate_global->__pyx_int_119832798
#define __pyx_int_122446337 __pyx_mstate_global->__pyx_int_122446337
#define __pyx_int_122978751 __pyx_mstate_global->__pyx_int_122978751
#define __pyx_int_136983863 __pyx_mstate_global->__pyx_int_136983863
#define __pyx_int_139219010 __pyx_mstate_global->__pyx_int_139219010
#define __pyx_int_145104888 __pyx_mstate_global->__pyx_int_145104888
#define __pyx_int_148208818 __pyx_mstate_global->__pyx_int_148208818
#define __pyx_int_148222551 __pyx_mstate_global->__pyx_int_148222551
#define __pyx_int_148338675 __pyx_mstate_global->__pyx_int_148338675
#define __pyx_int_148737976 __pyx_mstate_global->__pyx_int_148737976
#define __pyx_int_150354737 __pyx_mstate_global->__pyx_int_150354737
#define __pyx_int_159023012 __pyx_mstate_global->__pyx_int_159023012
#define __pyx_int_159276020 __pyx_mstate_global->__pyx_int_159276020
#define __pyx_int_159891379 __pyx_mstate_global->__pyx_int_159891379
#define __pyx_int_160503772 __pyx_mstate_global->__pyx_int_160503772
#define __pyx_int_160934521 __pyx_mstate_global->__pyx_int_160934521
#define __pyx_int_164094308 __pyx_mstate_global->__pyx_int_164094308
#define __pyx_int_169331114 __pyx_mstate_global->__pyx_int_169331114
#define __pyx_int_175892883 __pyx_mstate_global->__pyx_int_175892883
#define __pyx_int_176081420 __pyx_mstate_global->__pyx_int_176081420
#define __pyx_int_176267327 __pyx_mstate_global->__pyx_int_176267327
#define __pyx_int_177382991 __pyx_mstate_global->__pyx_int_177382991
#define __pyx_int_178791180 __pyx_mstate_global->__pyx_int_178791180
#define __pyx_int_180170947 __pyx_mstate_global->__pyx_int_180170947
#define __pyx_int_180285838 __pyx_mstate_global->__pyx_int_180285838
#define __pyx_int_181080083 __pyx_mstate_global->__pyx_int_181080083
#define __pyx_int_184956420 __pyx_mstate_global->__pyx_int_184956420
#define __pyx_int_184977713 __pyx_mstate_global->__pyx_int_184977713
#define __pyx_int_193777858 __pyx_mstate_global->__pyx_int_193777858
#define __pyx_int_195980960 __pyx_mstate_global->__pyx_int_195980960
#define __pyx_int_198151601 __pyx_mstate_global->__pyx_int_198151601
#define __pyx_int_202158861 __pyx_mstate_global->__pyx_int_202158861
#define __pyx_int_202608488 __pyx_mstate_global->__pyx_int_202608488
#define __pyx_int_203789025 __pyx_mstate_global->__pyx_int_203789025
#define __pyx_int_205941106 __pyx_mstate_global->__pyx_int_205941106
#define __pyx_int_206778879 __pyx_mstate_global->__pyx_int_206778879
#define __pyx_int_212073902 __pyx_mstate_global->__pyx_int_212073902
#define __pyx_int_213886906 __pyx_mstate_global->__pyx_int_213886906
#define __pyx_int_215229444 __pyx_mstate_global->__pyx_int_215229444
#define __pyx_int_215954136 __pyx_mstate_global->__pyx_int_215954136
#define __pyx_int_222419149 __pyx_mstate_global->__pyx_int_222419149
#define __pyx_int_228539134 __pyx_mstate_global->__pyx_int_228539134
#define __pyx_int_228825662 __pyx_mstate_global->__pyx_int_228825662
#define __pyx_int_229636597 __pyx_mstate_global->__pyx_int_229636597
#define __pyx_int_233298401 __pyx_mstate_global->__pyx_int_233298401
#define __pyx_int_236180632 __pyx_mstate_global->__pyx_int_236180632
#define __pyx_int_236349525 __pyx_mstate_global->__pyx_int_236349525
#define __pyx_int_238750788 __pyx_mstate_global->__pyx_int_238750788
#define __pyx_int_240389341 __pyx_mstate_global->__pyx_int_240389341
#define __pyx_int_245215779 __pyx_mstate_global->__pyx_int_245215779
#define __pyx_int_247527084 __pyx_mstate_global->__pyx_int_247527084
#define __pyx_int_254981756 __pyx_mstate_global->__pyx_int_254981756
#define __pyx_int_256569670 __pyx_mstate_global->__pyx_int_256569670
#define __pyx_int_259444540 __pyx_mstate_global->__pyx_int_259444540
#define __pyx_int_261059191 __pyx_mstate_global->__pyx_int_261059191
#define __pyx_int_263819094 __pyx_mstate_global->__pyx_int_263819094
#define __pyx_int_2147483647 __pyx_mstate_global->__pyx_int_2147483647
#define __pyx_int_3442987655 __pyx_mstate_global->__pyx_int_3442987655
#define __pyx_int_3600000000 __pyx_mstate_global->__pyx_int_3600000000
#define __pyx_int_neg_1 __pyx_mstate_global->__pyx_int_neg_1
#define __pyx_int_neg_15 __pyx_mstate_global->__pyx_int_neg_15
#define __pyx_int_549755813887L __pyx_mstate_global->__pyx_int_549755813887L
#define __pyx_int_0x7fffffffffffL __pyx_mstate_global->__pyx_int_0x7fffffffffffL
#define __pyx_int_0x7fffffffffffffL __pyx_mstate_global->__pyx_int_0x7fffffffffffffL
#define __pyx_int_0x7fffffffffffffffL __pyx_mstate_global->__pyx_int_0x7fffffffffffffffL
#define __pyx_k__15 __pyx_mstate_global->__pyx_k__15
#define __pyx_k__16 __pyx_mstate_global->__pyx_k__16
#define __pyx_k__17 __pyx_mstate_global->__pyx_k__17
#define __pyx_k__21 __pyx_mstate_global->__pyx_k__21
#define __pyx_k__81 __pyx_mstate_global->__pyx_k__81
#define __pyx_k__83 __pyx_mstate_global->__pyx_k__83
#define __pyx_k__84 __pyx_mstate_global->__pyx_k__84
#define __pyx_k__87 __pyx_mstate_global->__pyx_k__87
#define __pyx_k__89 __pyx_mstate_global->__pyx_k__89
#define __pyx_slice__5 __pyx_mstate_global->__pyx_slice__5
#define __pyx_tuple__4 __pyx_mstate_global->__pyx_tuple__4
#define __pyx_tuple__8 __pyx_mstate_global->__pyx_tuple__8
#define __pyx_slice__23 __pyx_mstate_global->__pyx_slice__23
#define __pyx_slice__86 __pyx_mstate_global->__pyx_slice__86
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__19 __pyx_mstate_global->__pyx_tuple__19
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__28 __pyx_mstate_global->__pyx_tuple__28
#define __pyx_tuple__30 __pyx_mstate_global->__pyx_tuple__30
#define __pyx_tuple__32 __pyx_mstate_global->__pyx_tuple__32
#define __pyx_tuple__35 __pyx_mstate_global->__pyx_tuple__35
#define __pyx_tuple__36 __pyx_mstate_global->__pyx_tuple__36
#define __pyx_tuple__37 __pyx_mstate_global->__pyx_tuple__37
#define __pyx_tuple__38 __pyx_mstate_global->__pyx_tuple__38
#define __pyx_tuple__40 __pyx_mstate_global->__pyx_tuple__40
#define __pyx_tuple__42 __pyx_mstate_global->__pyx_tuple__42
#define __pyx_tuple__43 __pyx_mstate_global->__pyx_tuple__43
#define __pyx_tuple__46 __pyx_mstate_global->__pyx_tuple__46
#define __pyx_tuple__47 __pyx_mstate_global->__pyx_tuple__47
#define __pyx_tuple__50 __pyx_mstate_global->__pyx_tuple__50
#define __pyx_tuple__51 __pyx_mstate_global->__pyx_tuple__51
#define __pyx_tuple__52 __pyx_mstate_global->__pyx_tuple__52
#define __pyx_tuple__53 __pyx_mstate_global->__pyx_tuple__53
#define __pyx_tuple__54 __pyx_mstate_global->__pyx_tuple__54
#define __pyx_tuple__55 __pyx_mstate_global->__pyx_tuple__55
#define __pyx_tuple__56 __pyx_mstate_global->__pyx_tuple__56
#define __pyx_tuple__57 __pyx_mstate_global->__pyx_tuple__57
#define __pyx_tuple__58 __pyx_mstate_global->__pyx_tuple__58
#define __pyx_tuple__59 __pyx_mstate_global->__pyx_tuple__59
#define __pyx_tuple__65 __pyx_mstate_global->__pyx_tuple__65
#define __pyx_tuple__67 __pyx_mstate_global->__pyx_tuple__67
#define __pyx_tuple__68 __pyx_mstate_global->__pyx_tuple__68
#define __pyx_tuple__69 __pyx_mstate_global->__pyx_tuple__69
#define __pyx_tuple__71 __pyx_mstate_global->__pyx_tuple__71
#define __pyx_tuple__73 __pyx_mstate_global->__pyx_tuple__73
#define __pyx_tuple__74 __pyx_mstate_global->__pyx_tuple__74
#define __pyx_tuple__85 __pyx_mstate_global->__pyx_tuple__85
#define __pyx_tuple__88 __pyx_mstate_global->__pyx_tuple__88
#define __pyx_tuple__90 __pyx_mstate_global->__pyx_tuple__90
#define __pyx_tuple__91 __pyx_mstate_global->__pyx_tuple__91
#define __pyx_tuple__92 __pyx_mstate_global->__pyx_tuple__92
#define __pyx_tuple__93 __pyx_mstate_global->__pyx_tuple__93
#define __pyx_tuple__94 __pyx_mstate_global->__pyx_tuple__94
#define __pyx_tuple__95 __pyx_mstate_global->__pyx_tuple__95
#define __pyx_tuple__96 __pyx_mstate_global->__pyx_tuple__96
#define __pyx_tuple__97 __pyx_mstate_global->__pyx_tuple__97
#define __pyx_tuple__98 __pyx_mstate_global->__pyx_tuple__98
#define __pyx_tuple__99 __pyx_mstate_global->__pyx_tuple__99
#define __pyx_tuple__100 __pyx_mstate_global->__pyx_tuple__100
#define __pyx_tuple__101 __pyx_mstate_global->__pyx_tuple__101
#define __pyx_tuple__102 __pyx_mstate_global->__pyx_tuple__102
#define __pyx_tuple__103 __pyx_mstate_global->__pyx_tuple__103
#define __pyx_tuple__104 __pyx_mstate_global->__pyx_tuple__104
#define __pyx_tuple__105 __pyx_mstate_global->__pyx_tuple__105
#define __pyx_tuple__106 __pyx_mstate_global->__pyx_tuple__106
#define __pyx_tuple__107 __pyx_mstate_global->__pyx_tuple__107
#define __pyx_tuple__108 __pyx_mstate_global->__pyx_tuple__108
#define __pyx_tuple__109 __pyx_mstate_global->__pyx_tuple__109
#define __pyx_tuple__110 __pyx_mstate_global->__pyx_tuple__110
#define __pyx_tuple__111 __pyx_mstate_global->__pyx_tuple__111
#define __pyx_tuple__112 __pyx_mstate_global->__pyx_tuple__112
#define __pyx_tuple__113 __pyx_mstate_global->__pyx_tuple__113
#define __pyx_tuple__114 __pyx_mstate_global->__pyx_tuple__114
#define __pyx_tuple__115 __pyx_mstate_global->__pyx_tuple__115
#define __pyx_tuple__116 __pyx_mstate_global->__pyx_tuple__116
#define __pyx_tuple__117 __pyx_mstate_global->__pyx_tuple__117
#define __pyx_tuple__118 __pyx_mstate_global->__pyx_tuple__118
#define __pyx_tuple__119 __pyx_mstate_global->__pyx_tuple__119
#define __pyx_tuple__120 __pyx_mstate_global->__pyx_tuple__120
#define __pyx_tuple__121 __pyx_mstate_global->__pyx_tuple__121
#define __pyx_tuple__123 __pyx_mstate_global->__pyx_tuple__123
#define __pyx_tuple__124 __pyx_mstate_global->__pyx_tuple__124
#define __pyx_tuple__125 __pyx_mstate_global->__pyx_tuple__125
#define __pyx_tuple__126 __pyx_mstate_global->__pyx_tuple__126
#define __pyx_tuple__127 __pyx_mstate_global->__pyx_tuple__127
#define __pyx_tuple__128 __pyx_mstate_global->__pyx_tuple__128
#define __pyx_tuple__129 __pyx_mstate_global->__pyx_tuple__129
#define __pyx_tuple__130 __pyx_mstate_global->__pyx_tuple__130
#define __pyx_tuple__131 __pyx_mstate_global->__pyx_tuple__131
#define __pyx_tuple__132 __pyx_mstate_global->__pyx_tuple__132
#define __pyx_tuple__134 __pyx_mstate_global->__pyx_tuple__134
#define __pyx_tuple__137 __pyx_mstate_global->__pyx_tuple__137
#define __pyx_tuple__139 __pyx_mstate_global->__pyx_tuple__139
#define __pyx_tuple__141 __pyx_mstate_global->__pyx_tuple__141
#define __pyx_tuple__142 __pyx_mstate_global->__pyx_tuple__142
#define __pyx_tuple__144 __pyx_mstate_global->__pyx_tuple__144
#define __pyx_tuple__145 __pyx_mstate_global->__pyx_tuple__145
#define __pyx_tuple__147 __pyx_mstate_global->__pyx_tuple__147
#define __pyx_tuple__148 __pyx_mstate_global->__pyx_tuple__148
#define __pyx_tuple__150 __pyx_mstate_global->__pyx_tuple__150
#define __pyx_tuple__153 __pyx_mstate_global->__pyx_tuple__153
#define __pyx_tuple__155 __pyx_mstate_global->__pyx_tuple__155
#define __pyx_tuple__157 __pyx_mstate_global->__pyx_tuple__157
#define __pyx_tuple__159 __pyx_mstate_global->__pyx_tuple__159
#define __pyx_tuple__161 __pyx_mstate_global->__pyx_tuple__161
#define __pyx_tuple__163 __pyx_mstate_global->__pyx_tuple__163
#define __pyx_tuple__164 __pyx_mstate_global->__pyx_tuple__164
#define __pyx_tuple__166 __pyx_mstate_global->__pyx_tuple__166
#define __pyx_tuple__167 __pyx_mstate_global->__pyx_tuple__167
#define __pyx_tuple__169 __pyx_mstate_global->__pyx_tuple__169
#define __pyx_tuple__183 __pyx_mstate_global->__pyx_tuple__183
#define __pyx_tuple__185 __pyx_mstate_global->__pyx_tuple__185
#define __pyx_tuple__188 __pyx_mstate_global->__pyx_tuple__188
#define __pyx_tuple__192 __pyx_mstate_global->__pyx_tuple__192
#define __pyx_tuple__194 __pyx_mstate_global->__pyx_tuple__194
#define __pyx_tuple__196 __pyx_mstate_global->__pyx_tuple__196
#define __pyx_tuple__198 __pyx_mstate_global->__pyx_tuple__198
#define __pyx_tuple__201 __pyx_mstate_global->__pyx_tuple__201
#define __pyx_tuple__202 __pyx_mstate_global->__pyx_tuple__202
#define __pyx_tuple__204 __pyx_mstate_global->__pyx_tuple__204
#define __pyx_tuple__206 __pyx_mstate_global->__pyx_tuple__206
#define __pyx_tuple__218 __pyx_mstate_global->__pyx_tuple__218
#define __pyx_tuple__220 __pyx_mstate_global->__pyx_tuple__220
#define __pyx_tuple__253 __pyx_mstate_global->__pyx_tuple__253
#define __pyx_tuple__258 __pyx_mstate_global->__pyx_tuple__258
#define __pyx_tuple__260 __pyx_mstate_global->__pyx_tuple__260
#define __pyx_tuple__264 __pyx_mstate_global->__pyx_tuple__264
#define __pyx_tuple__271 __pyx_mstate_global->__pyx_tuple__271
#define __pyx_tuple__294 __pyx_mstate_global->__pyx_tuple__294
#define __pyx_tuple__297 __pyx_mstate_global->__pyx_tuple__297
#define __pyx_tuple__300 __pyx_mstate_global->__pyx_tuple__300
#define __pyx_tuple__303 __pyx_mstate_global->__pyx_tuple__303
#define __pyx_tuple__306 __pyx_mstate_global->__pyx_tuple__306
#define __pyx_tuple__308 __pyx_mstate_global->__pyx_tuple__308
#define __pyx_tuple__313 __pyx_mstate_global->__pyx_tuple__313
#define __pyx_tuple__321 __pyx_mstate_global->__pyx_tuple__321
#define __pyx_tuple__323 __pyx_mstate_global->__pyx_tuple__323
#define __pyx_tuple__325 __pyx_mstate_global->__pyx_tuple__325
#define __pyx_tuple__327 __pyx_mstate_global->__pyx_tuple__327
#define __pyx_tuple__329 __pyx_mstate_global->__pyx_tuple__329
#define __pyx_tuple__331 __pyx_mstate_global->__pyx_tuple__331
#define __pyx_tuple__332 __pyx_mstate_global->__pyx_tuple__332
#define __pyx_tuple__334 __pyx_mstate_global->__pyx_tuple__334
#define __pyx_tuple__335 __pyx_mstate_global->__pyx_tuple__335
#define __pyx_tuple__337 __pyx_mstate_global->__pyx_tuple__337
#define __pyx_tuple__340 __pyx_mstate_global->__pyx_tuple__340
#define __pyx_tuple__342 __pyx_mstate_global->__pyx_tuple__342
#define __pyx_tuple__344 __pyx_mstate_global->__pyx_tuple__344
#define __pyx_tuple__346 __pyx_mstate_global->__pyx_tuple__346
#define __pyx_tuple__347 __pyx_mstate_global->__pyx_tuple__347
#define __pyx_tuple__349 __pyx_mstate_global->__pyx_tuple__349
#define __pyx_tuple__357 __pyx_mstate_global->__pyx_tuple__357
#define __pyx_tuple__363 __pyx_mstate_global->__pyx_tuple__363
#define __pyx_tuple__367 __pyx_mstate_global->__pyx_tuple__367
#define __pyx_tuple__377 __pyx_mstate_global->__pyx_tuple__377
#define __pyx_tuple__381 __pyx_mstate_global->__pyx_tuple__381
#define __pyx_tuple__396 __pyx_mstate_global->__pyx_tuple__396
#define __pyx_tuple__418 __pyx_mstate_global->__pyx_tuple__418
#define __pyx_tuple__420 __pyx_mstate_global->__pyx_tuple__420
#define __pyx_tuple__422 __pyx_mstate_global->__pyx_tuple__422
#define __pyx_tuple__424 __pyx_mstate_global->__pyx_tuple__424
#define __pyx_tuple__425 __pyx_mstate_global->__pyx_tuple__425
#define __pyx_tuple__427 __pyx_mstate_global->__pyx_tuple__427
#define __pyx_codeobj__11 __pyx_mstate_global->__pyx_codeobj__11
#define __pyx_codeobj__12 __pyx_mstate_global->__pyx_codeobj__12
#define __pyx_codeobj__29 __pyx_mstate_global->__pyx_codeobj__29
#define __pyx_codeobj__31 __pyx_mstate_global->__pyx_codeobj__31
#define __pyx_codeobj__33 __pyx_mstate_global->__pyx_codeobj__33
#define __pyx_codeobj__39 __pyx_mstate_global->__pyx_codeobj__39
#define __pyx_codeobj__41 __pyx_mstate_global->__pyx_codeobj__41
#define __pyx_codeobj__44 __pyx_mstate_global->__pyx_codeobj__44
#define __pyx_codeobj__45 __pyx_mstate_global->__pyx_codeobj__45
#define __pyx_codeobj__49 __pyx_mstate_global->__pyx_codeobj__49
#define __pyx_codeobj__66 __pyx_mstate_global->__pyx_codeobj__66
#define __pyx_codeobj__75 __pyx_mstate_global->__pyx_codeobj__75
#define __pyx_codeobj__82 __pyx_mstate_global->__pyx_codeobj__82
#define __pyx_codeobj__133 __pyx_mstate_global->__pyx_codeobj__133
#define __pyx_codeobj__135 __pyx_mstate_global->__pyx_codeobj__135
#define __pyx_codeobj__136 __pyx_mstate_global->__pyx_codeobj__136
#define __pyx_codeobj__138 __pyx_mstate_global->__pyx_codeobj__138
#define __pyx_codeobj__140 __pyx_mstate_global->__pyx_codeobj__140
#define __pyx_codeobj__143 __pyx_mstate_global->__pyx_codeobj__143
#define __pyx_codeobj__146 __pyx_mstate_global->__pyx_codeobj__146
#define __pyx_codeobj__149 __pyx_mstate_global->__pyx_codeobj__149
#define __pyx_codeobj__151 __pyx_mstate_global->__pyx_codeobj__151
#define __pyx_codeobj__154 __pyx_mstate_global->__pyx_codeobj__154
#define __pyx_codeobj__156 __pyx_mstate_global->__pyx_codeobj__156
#define __pyx_codeobj__158 __pyx_mstate_global->__pyx_codeobj__158
#define __pyx_codeobj__160 __pyx_mstate_global->__pyx_codeobj__160
#define __pyx_codeobj__162 __pyx_mstate_global->__pyx_codeobj__162
#define __pyx_codeobj__165 __pyx_mstate_global->__pyx_codeobj__165
#define __pyx_codeobj__168 __pyx_mstate_global->__pyx_codeobj__168
#define __pyx_codeobj__170 __pyx_mstate_global->__pyx_codeobj__170
#define __pyx_codeobj__171 __pyx_mstate_global->__pyx_codeobj__171
#define __pyx_codeobj__172 __pyx_mstate_global->__pyx_codeobj__172
#define __pyx_codeobj__173 __pyx_mstate_global->__pyx_codeobj__173
#define __pyx_codeobj__174 __pyx_mstate_global->__pyx_codeobj__174
#define __pyx_codeobj__175 __pyx_mstate_global->__pyx_codeobj__175
#define __pyx_codeobj__176 __pyx_mstate_global->__pyx_codeobj__176
#define __pyx_codeobj__177 __pyx_mstate_global->__pyx_codeobj__177
#define __pyx_codeobj__178 __pyx_mstate_global->__pyx_codeobj__178
#define __pyx_codeobj__179 __pyx_mstate_global->__pyx_codeobj__179
#define __pyx_codeobj__180 __pyx_mstate_global->__pyx_codeobj__180
#define __pyx_codeobj__181 __pyx_mstate_global->__pyx_codeobj__181
#define __pyx_codeobj__182 __pyx_mstate_global->__pyx_codeobj__182
#define __pyx_codeobj__184 __pyx_mstate_global->__pyx_codeobj__184
#define __pyx_codeobj__186 __pyx_mstate_global->__pyx_codeobj__186
#define __pyx_codeobj__187 __pyx_mstate_global->__pyx_codeobj__187
#define __pyx_codeobj__189 __pyx_mstate_global->__pyx_codeobj__189
#define __pyx_codeobj__190 __pyx_mstate_global->__pyx_codeobj__190
#define __pyx_codeobj__191 __pyx_mstate_global->__pyx_codeobj__191
#define __pyx_codeobj__193 __pyx_mstate_global->__pyx_codeobj__193
#define __pyx_codeobj__195 __pyx_mstate_global->__pyx_codeobj__195
#define __pyx_codeobj__197 __pyx_mstate_global->__pyx_codeobj__197
#define __pyx_codeobj__199 __pyx_mstate_global->__pyx_codeobj__199
#define __pyx_codeobj__200 __pyx_mstate_global->__pyx_codeobj__200
#define __pyx_codeobj__203 __pyx_mstate_global->__pyx_codeobj__203
#define __pyx_codeobj__205 __pyx_mstate_global->__pyx_codeobj__205
#define __pyx_codeobj__207 __pyx_mstate_global->__pyx_codeobj__207
#define __pyx_codeobj__208 __pyx_mstate_global->__pyx_codeobj__208
#define __pyx_codeobj__209 __pyx_mstate_global->__pyx_codeobj__209
#define __pyx_codeobj__210 __pyx_mstate_global->__pyx_codeobj__210
#define __pyx_codeobj__211 __pyx_mstate_global->__pyx_codeobj__211
#define __pyx_codeobj__212 __pyx_mstate_global->__pyx_codeobj__212
#define __pyx_codeobj__213 __pyx_mstate_global->__pyx_codeobj__213
#define __pyx_codeobj__214 __pyx_mstate_global->__pyx_codeobj__214
#define __pyx_codeobj__215 __pyx_mstate_global->__pyx_codeobj__215
#define __pyx_codeobj__216 __pyx_mstate_global->__pyx_codeobj__216
#define __pyx_codeobj__217 __pyx_mstate_global->__pyx_codeobj__217
#define __pyx_codeobj__219 __pyx_mstate_global->__pyx_codeobj__219
#define __pyx_codeobj__221 __pyx_mstate_global->__pyx_codeobj__221
#define __pyx_codeobj__222 __pyx_mstate_global->__pyx_codeobj__222
#define __pyx_codeobj__223 __pyx_mstate_global->__pyx_codeobj__223
#define __pyx_codeobj__224 __pyx_mstate_global->__pyx_codeobj__224
#define __pyx_codeobj__225 __pyx_mstate_global->__pyx_codeobj__225
#define __pyx_codeobj__226 __pyx_mstate_global->__pyx_codeobj__226
#define __pyx_codeobj__227 __pyx_mstate_global->__pyx_codeobj__227
#define __pyx_codeobj__228 __pyx_mstate_global->__pyx_codeobj__228
#define __pyx_codeobj__229 __pyx_mstate_global->__pyx_codeobj__229
#define __pyx_codeobj__230 __pyx_mstate_global->__pyx_codeobj__230
#define __pyx_codeobj__231 __pyx_mstate_global->__pyx_codeobj__231
#define __pyx_codeobj__232 __pyx_mstate_global->__pyx_codeobj__232
#define __pyx_codeobj__233 __pyx_mstate_global->__pyx_codeobj__233
#define __pyx_codeobj__234 __pyx_mstate_global->__pyx_codeobj__234
#define __pyx_codeobj__235 __pyx_mstate_global->__pyx_codeobj__235
#define __pyx_codeobj__236 __pyx_mstate_global->__pyx_codeobj__236
#define __pyx_codeobj__237 __pyx_mstate_global->__pyx_codeobj__237
#define __pyx_codeobj__238 __pyx_mstate_global->__pyx_codeobj__238
#define __pyx_codeobj__239 __pyx_mstate_global->__pyx_codeobj__239
#define __pyx_codeobj__240 __pyx_mstate_global->__pyx_codeobj__240
#define __pyx_codeobj__241 __pyx_mstate_global->__pyx_codeobj__241
#define __pyx_codeobj__242 __pyx_mstate_global->__pyx_codeobj__242
#define __pyx_codeobj__243 __pyx_mstate_global->__pyx_codeobj__243
#define __pyx_codeobj__244 __pyx_mstate_global->__pyx_codeobj__244
#define __pyx_codeobj__245 __pyx_mstate_global->__pyx_codeobj__245
#define __pyx_codeobj__246 __pyx_mstate_global->__pyx_codeobj__246
#define __pyx_codeobj__247 __pyx_mstate_global->__pyx_codeobj__247
#define __pyx_codeobj__248 __pyx_mstate_global->__pyx_codeobj__248
#define __pyx_codeobj__249 __pyx_mstate_global->__pyx_codeobj__249
#define __pyx_codeobj__250 __pyx_mstate_global->__pyx_codeobj__250
#define __pyx_codeobj__251 __pyx_mstate_global->__pyx_codeobj__251
#define __pyx_codeobj__252 __pyx_mstate_global->__pyx_codeobj__252
#define __pyx_codeobj__254 __pyx_mstate_global->__pyx_codeobj__254
#define __pyx_codeobj__255 __pyx_mstate_global->__pyx_codeobj__255
#define __pyx_codeobj__256 __pyx_mstate_global->__pyx_codeobj__256
#define __pyx_codeobj__257 __pyx_mstate_global->__pyx_codeobj__257
#define __pyx_codeobj__259 __pyx_mstate_global->__pyx_codeobj__259
#define __pyx_codeobj__261 __pyx_mstate_global->__pyx_codeobj__261
#define __pyx_codeobj__262 __pyx_mstate_global->__pyx_codeobj__262
#define __pyx_codeobj__263 __pyx_mstate_global->__pyx_codeobj__263
#define __pyx_codeobj__265 __pyx_mstate_global->__pyx_codeobj__265
#define __pyx_codeobj__266 __pyx_mstate_global->__pyx_codeobj__266
#define __pyx_codeobj__267 __pyx_mstate_global->__pyx_codeobj__267
#define __pyx_codeobj__268 __pyx_mstate_global->__pyx_codeobj__268
#define __pyx_codeobj__269 __pyx_mstate_global->__pyx_codeobj__269
#define __pyx_codeobj__270 __pyx_mstate_global->__pyx_codeobj__270
#define __pyx_codeobj__272 __pyx_mstate_global->__pyx_codeobj__272
#define __pyx_codeobj__273 __pyx_mstate_global->__pyx_codeobj__273
#define __pyx_codeobj__274 __pyx_mstate_global->__pyx_codeobj__274
#define __pyx_codeobj__275 __pyx_mstate_global->__pyx_codeobj__275
#define __pyx_codeobj__276 __pyx_mstate_global->__pyx_codeobj__276
#define __pyx_codeobj__277 __pyx_mstate_global->__pyx_codeobj__277
#define __pyx_codeobj__278 __pyx_mstate_global->__pyx_codeobj__278
#define __pyx_codeobj__279 __pyx_mstate_global->__pyx_codeobj__279
#define __pyx_codeobj__280 __pyx_mstate_global->__pyx_codeobj__280
#define __pyx_codeobj__281 __pyx_mstate_global->__pyx_codeobj__281
#define __pyx_codeobj__282 __pyx_mstate_global->__pyx_codeobj__282
#define __pyx_codeobj__283 __pyx_mstate_global->__pyx_codeobj__283
#define __pyx_codeobj__284 __pyx_mstate_global->__pyx_codeobj__284
#define __pyx_codeobj__285 __pyx_mstate_global->__pyx_codeobj__285
#define __pyx_codeobj__286 __pyx_mstate_global->__pyx_codeobj__286
#define __pyx_codeobj__287 __pyx_mstate_global->__pyx_codeobj__287
#define __pyx_codeobj__288 __pyx_mstate_global->__pyx_codeobj__288
#define __pyx_codeobj__289 __pyx_mstate_global->__pyx_codeobj__289
#define __pyx_codeobj__290 __pyx_mstate_global->__pyx_codeobj__290
#define __pyx_codeobj__291 __pyx_mstate_global->__pyx_codeobj__291
#define __pyx_codeobj__292 __pyx_mstate_global->__pyx_codeobj__292
#define __pyx_codeobj__293 __pyx_mstate_global->__pyx_codeobj__293
#define __pyx_codeobj__295 __pyx_mstate_global->__pyx_codeobj__295
#define __pyx_codeobj__296 __pyx_mstate_global->__pyx_codeobj__296
#define __pyx_codeobj__298 __pyx_mstate_global->__pyx_codeobj__298
#define __pyx_codeobj__299 __pyx_mstate_global->__pyx_codeobj__299
#define __pyx_codeobj__301 __pyx_mstate_global->__pyx_codeobj__301
#define __pyx_codeobj__302 __pyx_mstate_global->__pyx_codeobj__302
#define __pyx_codeobj__304 __pyx_mstate_global->__pyx_codeobj__304
#define __pyx_codeobj__305 __pyx_mstate_global->__pyx_codeobj__305
#define __pyx_codeobj__307 __pyx_mstate_global->__pyx_codeobj__307
#define __pyx_codeobj__309 __pyx_mstate_global->__pyx_codeobj__309
#define __pyx_codeobj__310 __pyx_mstate_global->__pyx_codeobj__310
#define __pyx_codeobj__311 __pyx_mstate_global->__pyx_codeobj__311
#define __pyx_codeobj__312 __pyx_mstate_global->__pyx_codeobj__312
#define __pyx_codeobj__314 __pyx_mstate_global->__pyx_codeobj__314
#define __pyx_codeobj__315 __pyx_mstate_global->__pyx_codeobj__315
#define __pyx_codeobj__316 __pyx_mstate_global->__pyx_codeobj__316
#define __pyx_codeobj__317 __pyx_mstate_global->__pyx_codeobj__317
#define __pyx_codeobj__318 __pyx_mstate_global->__pyx_codeobj__318
#define __pyx_codeobj__319 __pyx_mstate_global->__pyx_codeobj__319
#define __pyx_codeobj__320 __pyx_mstate_global->__pyx_codeobj__320
#define __pyx_codeobj__322 __pyx_mstate_global->__pyx_codeobj__322
#define __pyx_codeobj__324 __pyx_mstate_global->__pyx_codeobj__324
#define __pyx_codeobj__326 __pyx_mstate_global->__pyx_codeobj__326
#define __pyx_codeobj__328 __pyx_mstate_global->__pyx_codeobj__328
#define __pyx_codeobj__330 __pyx_mstate_global->__pyx_codeobj__330
#define __pyx_codeobj__333 __pyx_mstate_global->__pyx_codeobj__333
#define __pyx_codeobj__336 __pyx_mstate_global->__pyx_codeobj__336
#define __pyx_codeobj__338 __pyx_mstate_global->__pyx_codeobj__338
#define __pyx_codeobj__339 __pyx_mstate_global->__pyx_codeobj__339
#define __pyx_codeobj__341 __pyx_mstate_global->__pyx_codeobj__341
#define __pyx_codeobj__343 __pyx_mstate_global->__pyx_codeobj__343
#define __pyx_codeobj__345 __pyx_mstate_global->__pyx_codeobj__345
#define __pyx_codeobj__348 __pyx_mstate_global->__pyx_codeobj__348
#define __pyx_codeobj__350 __pyx_mstate_global->__pyx_codeobj__350
#define __pyx_codeobj__351 __pyx_mstate_global->__pyx_codeobj__351
#define __pyx_codeobj__352 __pyx_mstate_global->__pyx_codeobj__352
#define __pyx_codeobj__353 __pyx_mstate_global->__pyx_codeobj__353
#define __pyx_codeobj__354 __pyx_mstate_global->__pyx_codeobj__354
#define __pyx_codeobj__355 __pyx_mstate_global->__pyx_codeobj__355
#define __pyx_codeobj__356 __pyx_mstate_global->__pyx_codeobj__356
#define __pyx_codeobj__358 __pyx_mstate_global->__pyx_codeobj__358
#define __pyx_codeobj__359 __pyx_mstate_global->__pyx_codeobj__359
#define __pyx_codeobj__360 __pyx_mstate_global->__pyx_codeobj__360
#define __pyx_codeobj__361 __pyx_mstate_global->__pyx_codeobj__361
#define __pyx_codeobj__362 __pyx_mstate_global->__pyx_codeobj__362
#define __pyx_codeobj__364 __pyx_mstate_global->__pyx_codeobj__364
#define __pyx_codeobj__365 __pyx_mstate_global->__pyx_codeobj__365
#define __pyx_codeobj__366 __pyx_mstate_global->__pyx_codeobj__366
#define __pyx_codeobj__368 __pyx_mstate_global->__pyx_codeobj__368
#define __pyx_codeobj__369 __pyx_mstate_global->__pyx_codeobj__369
#define __pyx_codeobj__370 __pyx_mstate_global->__pyx_codeobj__370
#define __pyx_codeobj__371 __pyx_mstate_global->__pyx_codeobj__371
#define __pyx_codeobj__372 __pyx_mstate_global->__pyx_codeobj__372
#define __pyx_codeobj__373 __pyx_mstate_global->__pyx_codeobj__373
#define __pyx_codeobj__374 __pyx_mstate_global->__pyx_codeobj__374
#define __pyx_codeobj__375 __pyx_mstate_global->__pyx_codeobj__375
#define __pyx_codeobj__376 __pyx_mstate_global->__pyx_codeobj__376
#define __pyx_codeobj__378 __pyx_mstate_global->__pyx_codeobj__378
#define __pyx_codeobj__379 __pyx_mstate_global->__pyx_codeobj__379
#define __pyx_codeobj__380 __pyx_mstate_global->__pyx_codeobj__380
#define __pyx_codeobj__382 __pyx_mstate_global->__pyx_codeobj__382
#define __pyx_codeobj__383 __pyx_mstate_global->__pyx_codeobj__383
#define __pyx_codeobj__384 __pyx_mstate_global->__pyx_codeobj__384
#define __pyx_codeobj__385 __pyx_mstate_global->__pyx_codeobj__385
#define __pyx_codeobj__386 __pyx_mstate_global->__pyx_codeobj__386
#define __pyx_codeobj__387 __pyx_mstate_global->__pyx_codeobj__387
#define __pyx_codeobj__388 __pyx_mstate_global->__pyx_codeobj__388
#define __pyx_codeobj__389 __pyx_mstate_global->__pyx_codeobj__389
#define __pyx_codeobj__390 __pyx_mstate_global->__pyx_codeobj__390
#define __pyx_codeobj__391 __pyx_mstate_global->__pyx_codeobj__391
#define __pyx_codeobj__392 __pyx_mstate_global->__pyx_codeobj__392
#define __pyx_codeobj__393 __pyx_mstate_global->__pyx_codeobj__393
#define __pyx_codeobj__394 __pyx_mstate_global->__pyx_codeobj__394
#define __pyx_codeobj__395 __pyx_mstate_global->__pyx_codeobj__395
#define __pyx_codeobj__397 __pyx_mstate_global->__pyx_codeobj__397
#define __pyx_codeobj__398 __pyx_mstate_global->__pyx_codeobj__398
#define __pyx_codeobj__399 __pyx_mstate_global->__pyx_codeobj__399
#define __pyx_codeobj__400 __pyx_mstate_global->__pyx_codeobj__400
#define __pyx_codeobj__401 __pyx_mstate_global->__pyx_codeobj__401
#define __pyx_codeobj__402 __pyx_mstate_global->__pyx_codeobj__402
#define __pyx_codeobj__403 __pyx_mstate_global->__pyx_codeobj__403
#define __pyx_codeobj__404 __pyx_mstate_global->__pyx_codeobj__404
#define __pyx_codeobj__405 __pyx_mstate_global->__pyx_codeobj__405
#define __pyx_codeobj__406 __pyx_mstate_global->__pyx_codeobj__406
#define __pyx_codeobj__407 __pyx_mstate_global->__pyx_codeobj__407
#define __pyx_codeobj__408 __pyx_mstate_global->__pyx_codeobj__408
#define __pyx_codeobj__409 __pyx_mstate_global->__pyx_codeobj__409
#define __pyx_codeobj__410 __pyx_mstate_global->__pyx_codeobj__410
#define __pyx_codeobj__411 __pyx_mstate_global->__pyx_codeobj__411
#define __pyx_codeobj__412 __pyx_mstate_global->__pyx_codeobj__412
#define __pyx_codeobj__413 __pyx_mstate_global->__pyx_codeobj__413
#define __pyx_codeobj__414 __pyx_mstate_global->__pyx_codeobj__414
#define __pyx_codeobj__415 __pyx_mstate_global->__pyx_codeobj__415
#define __pyx_codeobj__416 __pyx_mstate_global->__pyx_codeobj__416
#define __pyx_codeobj__417 __pyx_mstate_global->__pyx_codeobj__417
#define __pyx_codeobj__419 __pyx_mstate_global->__pyx_codeobj__419
#define __pyx_codeobj__421 __pyx_mstate_global->__pyx_codeobj__421
#define __pyx_codeobj__423 __pyx_mstate_global->__pyx_codeobj__423
#define __pyx_codeobj__426 __pyx_mstate_global->__pyx_codeobj__426
#define __pyx_codeobj__428 __pyx_mstate_global->__pyx_codeobj__428
#define __pyx_codeobj__429 __pyx_mstate_global->__pyx_codeobj__429
#define __pyx_codeobj__430 __pyx_mstate_global->__pyx_codeobj__430
#define __pyx_codeobj__431 __pyx_mstate_global->__pyx_codeobj__431
#define __pyx_codeobj__432 __pyx_mstate_global->__pyx_codeobj__432
#define __pyx_codeobj__433 __pyx_mstate_global->__pyx_codeobj__433
#define __pyx_codeobj__434 __pyx_mstate_global->__pyx_codeobj__434
#define __pyx_codeobj__435 __pyx_mstate_global->__pyx_codeobj__435
#define __pyx_codeobj__436 __pyx_mstate_global->__pyx_codeobj__436
#define __pyx_codeobj__437 __pyx_mstate_global->__pyx_codeobj__437
#define __pyx_codeobj__438 __pyx_mstate_global->__pyx_codeobj__438
#define __pyx_codeobj__439 __pyx_mstate_global->__pyx_codeobj__439
#define __pyx_codeobj__440 __pyx_mstate_global->__pyx_codeobj__440
#define __pyx_codeobj__441 __pyx_mstate_global->__pyx_codeobj__441
#define __pyx_codeobj__442 __pyx_mstate_global->__pyx_codeobj__442
#define __pyx_codeobj__443 __pyx_mstate_global->__pyx_codeobj__443
#define __pyx_codeobj__444 __pyx_mstate_global->__pyx_codeobj__444
#define __pyx_codeobj__445 __pyx_mstate_global->__pyx_codeobj__445
#define __pyx_codeobj__446 __pyx_mstate_global->__pyx_codeobj__446
#define __pyx_codeobj__447 __pyx_mstate_global->__pyx_codeobj__447
#define __pyx_codeobj__448 __pyx_mstate_global->__pyx_codeobj__448
#define __pyx_codeobj__449 __pyx_mstate_global->__pyx_codeobj__449
#define __pyx_codeobj__450 __pyx_mstate_global->__pyx_codeobj__450
#define __pyx_codeobj__451 __pyx_mstate_global->__pyx_codeobj__451
#define __pyx_codeobj__452 __pyx_mstate_global->__pyx_codeobj__452
#define __pyx_codeobj__453 __pyx_mstate_global->__pyx_codeobj__453
#define __pyx_codeobj__454 __pyx_mstate_global->__pyx_codeobj__454
#define __pyx_codeobj__455 __pyx_mstate_global->__pyx_codeobj__455
#define __pyx_codeobj__456 __pyx_mstate_global->__pyx_codeobj__456
#define __pyx_codeobj__457 __pyx_mstate_global->__pyx_codeobj__457
#define __pyx_codeobj__458 __pyx_mstate_global->__pyx_codeobj__458
#define __pyx_codeobj__459 __pyx_mstate_global->__pyx_codeobj__459
#define __pyx_codeobj__460 __pyx_mstate_global->__pyx_codeobj__460
#define __pyx_codeobj__461 __pyx_mstate_global->__pyx_codeobj__461
#define __pyx_codeobj__462 __pyx_mstate_global->__pyx_codeobj__462
#define __pyx_codeobj__463 __pyx_mstate_global->__pyx_codeobj__463
#define __pyx_codeobj__464 __pyx_mstate_global->__pyx_codeobj__464
#define __pyx_codeobj__465 __pyx_mstate_global->__pyx_codeobj__465
#define __pyx_codeobj__466 __pyx_mstate_global->__pyx_codeobj__466
#define __pyx_codeobj__467 __pyx_mstate_global->__pyx_codeobj__467
#define __pyx_codeobj__468 __pyx_mstate_global->__pyx_codeobj__468
#define __pyx_codeobj__469 __pyx_mstate_global->__pyx_codeobj__469
#define __pyx_codeobj__470 __pyx_mstate_global->__pyx_codeobj__470
#define __pyx_codeobj__471 __pyx_mstate_global->__pyx_codeobj__471
#define __pyx_codeobj__472 __pyx_mstate_global->__pyx_codeobj__472
#define __pyx_codeobj__473 __pyx_mstate_global->__pyx_codeobj__473
#define __pyx_codeobj__474 __pyx_mstate_global->__pyx_codeobj__474
#define __pyx_codeobj__475 __pyx_mstate_global->__pyx_codeobj__475
#define __pyx_codeobj__476 __pyx_mstate_global->__pyx_codeobj__476
#define __pyx_codeobj__477 __pyx_mstate_global->__pyx_codeobj__477
#define __pyx_codeobj__478 __pyx_mstate_global->__pyx_codeobj__478
#define __pyx_codeobj__479 __pyx_mstate_global->__pyx_codeobj__479
#define __pyx_codeobj__480 __pyx_mstate_global->__pyx_codeobj__480
#define __pyx_codeobj__481 __pyx_mstate_global->__pyx_codeobj__481
#define __pyx_codeobj__482 __pyx_mstate_global->__pyx_codeobj__482
#define __pyx_codeobj__483 __pyx_mstate_global->__pyx_codeobj__483
#define __pyx_codeobj__484 __pyx_mstate_global->__pyx_codeobj__484
#define __pyx_codeobj__485 __pyx_mstate_global->__pyx_codeobj__485
#define __pyx_codeobj__486 __pyx_mstate_global->__pyx_codeobj__486
#define __pyx_codeobj__487 __pyx_mstate_global->__pyx_codeobj__487
#define __pyx_codeobj__488 __pyx_mstate_global->__pyx_codeobj__488
#define __pyx_codeobj__489 __pyx_mstate_global->__pyx_codeobj__489
#define __pyx_codeobj__490 __pyx_mstate_global->__pyx_codeobj__490
#define __pyx_codeobj__491 __pyx_mstate_global->__pyx_codeobj__491
#define __pyx_codeobj__492 __pyx_mstate_global->__pyx_codeobj__492
#define __pyx_codeobj__493 __pyx_mstate_global->__pyx_codeobj__493
#define __pyx_codeobj__494 __pyx_mstate_global->__pyx_codeobj__494
#define __pyx_codeobj__495 __pyx_mstate_global->__pyx_codeobj__495
#define __pyx_codeobj__496 __pyx_mstate_global->__pyx_codeobj__496
/* #### Code section: module_code ### */

/* "View.MemoryView":131
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

/* Python wrapper */
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_format = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_v_allocate_buffer;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    values[3] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_n_s_c));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_shape)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_itemsize)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); __PYX_ERR(1, 131, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_format)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); __PYX_ERR(1, 131, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mode);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_allocate_buffer);
          if (value) { values[4] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 131, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = ((PyObject*)values[0]);
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L3_error)
    __pyx_v_format = values[2];
    __pyx_v_mode = values[3];
    if (values[4]) {
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 132, __pyx_L3_error)
    } else {

      /* "View.MemoryView":132
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */
      __pyx_v_allocate_buffer = ((int)1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, __pyx_nargs); __PYX_ERR(1, 131, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(1, 131, __pyx_L1_error)
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(1, 131, __pyx_L1_error)
  }
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":131
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_v_dim;
  char __pyx_v_order;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  char *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":137
 *         cdef Py_ssize_t dim
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */
  if (unlikely(__pyx_v_shape == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 137, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 137, __pyx_L1_error)
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":138
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":140
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Empty shape tuple for cython.array"
 * 
 */
  __pyx_t_2 = (!(__pyx_v_self->ndim != 0));
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":141
 * 
 *         if not self.ndim:
 *             raise ValueError, "Empty shape tuple for cython.array"             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Empty_shape_tuple_for_cython_arr, 0, 0);
    __PYX_ERR(1, 141, __pyx_L1_error)

    /* "View.MemoryView":140
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Empty shape tuple for cython.array"
 * 
 */
  }

  /* "View.MemoryView":143
 *             raise ValueError, "Empty shape tuple for cython.array"
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError, "itemsize <= 0 for cython.array"
 * 
 */
  __pyx_t_2 = (__pyx_v_itemsize <= 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":144
 * 
 *         if itemsize <= 0:
 *             raise ValueError, "itemsize <= 0 for cython.array"             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_itemsize_0_for_cython_array, 0, 0);
    __PYX_ERR(1, 144, __pyx_L1_error)

    /* "View.MemoryView":143
 *             raise ValueError, "Empty shape tuple for cython.array"
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError, "itemsize <= 0 for cython.array"
 * 
 */
  }

  /* "View.MemoryView":146
 *             raise ValueError, "itemsize <= 0 for cython.array"
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "View.MemoryView":147
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_n_s_ASCII};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 147, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":146
 *             raise ValueError, "itemsize <= 0 for cython.array"
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  }

  /* "View.MemoryView":148
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_v_format))) __PYX_ERR(1, 148, __pyx_L1_error)
  __pyx_t_4 = __pyx_v_format;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_format);
  __Pyx_DECREF(__pyx_v_self->_format);
  __pyx_v_self->_format = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":149
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_self->_format == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(1, 149, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(1, 149, __pyx_L1_error)
  __pyx_v_self->format = __pyx_t_8;

  /* "View.MemoryView":152
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":153
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":155
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError, "unable to allocate shape and strides."
 * 
 */
  __pyx_t_3 = (!(__pyx_v_self->_shape != 0));
  if (unlikely(__pyx_t_3)) {

    /* "View.MemoryView":156
 * 
 *         if not self._shape:
 *             raise MemoryError, "unable to allocate shape and strides."             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_Raise(__pyx_builtin_MemoryError, __pyx_kp_s_unable_to_allocate_shape_and_str, 0, 0);
    __PYX_ERR(1, 156, __pyx_L1_error)

    /* "View.MemoryView":155
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError, "unable to allocate shape and strides."
 * 
 */
  }

  /* "View.MemoryView":159
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 */
  __pyx_t_7 = 0;
  __pyx_t_4 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_4);
  __pyx_t_1 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 159, __pyx_L1_error)
      #endif
      if (__pyx_t_1 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely((0 < 0))) __PYX_ERR(1, 159, __pyx_L1_error)
    #else
    __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 159, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_9;
    __pyx_v_idx = __pyx_t_7;
    __pyx_t_7 = (__pyx_t_7 + 1);

    /* "View.MemoryView":160
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 *             self._shape[idx] = dim
 */
    __pyx_t_3 = (__pyx_v_dim <= 0);
    if (unlikely(__pyx_t_3)) {

      /* "View.MemoryView":161
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */
      __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;
      __Pyx_INCREF(__pyx_kp_u_Invalid_shape_in_axis);
      __pyx_t_9 += 22;
      __Pyx_GIVEREF(__pyx_kp_u_Invalid_shape_in_axis);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Invalid_shape_in_axis);
      __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_idx, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_kp_u_);
      __pyx_t_9 += 2;
      __Pyx_GIVEREF(__pyx_kp_u_);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u_);
      __pyx_t_6 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_dim, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_kp_u__2);
      __pyx_t_9 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__2);
      PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_kp_u__2);
      __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_5, 5, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_builtin_ValueError, __pyx_t_6, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(1, 161, __pyx_L1_error)

      /* "View.MemoryView":160
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 *             self._shape[idx] = dim
 */
    }

    /* "View.MemoryView":162
 *             if dim <= 0:
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":159
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "View.MemoryView":165
 * 
 *         cdef char order
 *         if mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(1, 165, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "View.MemoryView":166
 *         cdef char order
 *         if mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         elif mode == 'fortran':
 */
    __pyx_v_order = 'C';

    /* "View.MemoryView":167
 *         if mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         elif mode == 'fortran':
 *             order = b'F'
 */
    __Pyx_INCREF(__pyx_n_u_c);
    __Pyx_GIVEREF(__pyx_n_u_c);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":165
 * 
 *         cdef char order
 *         if mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
    goto __pyx_L11;
  }

  /* "View.MemoryView":168
 *             order = b'C'
 *             self.mode = u'c'
 *         elif mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(1, 168, __pyx_L1_error)
  if (likely(__pyx_t_3)) {

    /* "View.MemoryView":169
 *             self.mode = u'c'
 *         elif mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         else:
 */
    __pyx_v_order = 'F';

    /* "View.MemoryView":170
 *         elif mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError, f"Invalid mode, expected 'c' or 'fortran', got {mode}"
 */
    __Pyx_INCREF(__pyx_n_u_fortran);
    __Pyx_GIVEREF(__pyx_n_u_fortran);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":168
 *             order = b'C'
 *             self.mode = u'c'
 *         elif mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
    goto __pyx_L11;
  }

  /* "View.MemoryView":172
 *             self.mode = u'fortran'
 *         else:
 *             raise ValueError, f"Invalid mode, expected 'c' or 'fortran', got {mode}"             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides, itemsize, self.ndim, order)
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_mode, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Invalid_mode_expected_c_or_fortr, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_t_6, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 172, __pyx_L1_error)
  }
  __pyx_L11:;

  /* "View.MemoryView":174
 *             raise ValueError, f"Invalid mode, expected 'c' or 'fortran', got {mode}"
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides, itemsize, self.ndim, order)             # <<<<<<<<<<<<<<
 * 
 *         self.free_data = allocate_buffer
 */
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":176
 *         self.len = fill_contig_strides_array(self._shape, self._strides, itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 * 
 */
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":177
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 * 
 *         if allocate_buffer:
 */
  __pyx_t_6 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_self->dtype_is_object = __pyx_t_3;

  /* "View.MemoryView":179
 *         self.dtype_is_object = format == b'O'
 * 
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 *             _allocate_buffer(self)
 * 
 */
  if (__pyx_v_allocate_buffer) {

    /* "View.MemoryView":180
 * 
 *         if allocate_buffer:
 *             _allocate_buffer(self)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_t_7 = __pyx_array_allocate_buffer(__pyx_v_self); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 180, __pyx_L1_error)

    /* "View.MemoryView":179
 *         self.dtype_is_object = format == b'O'
 * 
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 *             _allocate_buffer(self)
 * 
 */
  }

  /* "View.MemoryView":131
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":182
 *             _allocate_buffer(self)
 * 
 *     @cname('getbuffer')             # <<<<<<<<<<<<<<
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 */

/* Python wrapper */
CYTHON_UNUSED static int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
CYTHON_UNUSED static int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  char *__pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (unlikely(__pyx_v_info == NULL)) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":184
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if flags & (PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS):
 *             if self.mode == u"c":
 */
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":185
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if flags & (PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS):             # <<<<<<<<<<<<<<
 *             if self.mode == u"c":
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  __pyx_t_1 = ((__pyx_v_flags & ((PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS) | PyBUF_ANY_CONTIGUOUS)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":186
 *         cdef int bufmode = -1
 *         if flags & (PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS):
 *             if self.mode == u"c":             # <<<<<<<<<<<<<<
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             elif self.mode == u"fortran":
 */
    __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 186, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":187
 *         if flags & (PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS):
 *             if self.mode == u"c":
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *             elif self.mode == u"fortran":
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
      __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

      /* "View.MemoryView":186
 *         cdef int bufmode = -1
 *         if flags & (PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS):
 *             if self.mode == u"c":             # <<<<<<<<<<<<<<
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             elif self.mode == u"fortran":
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":188
 *             if self.mode == u"c":
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             if not (flags & bufmode):
 */
    __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 188, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":189
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             elif self.mode == u"fortran":
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *             if not (flags & bufmode):
 *                 raise ValueError, "Can only create a buffer that is contiguous in memory."
 */
      __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

      /* "View.MemoryView":188
 *             if self.mode == u"c":
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             if not (flags & bufmode):
 */
    }
    __pyx_L4:;

    /* "View.MemoryView":190
 *             elif self.mode == u"fortran":
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *                 raise ValueError, "Can only create a buffer that is contiguous in memory."
 *         info.buf = self.data
 */
    __pyx_t_1 = (!((__pyx_v_flags & __pyx_v_bufmode) != 0));
    if (unlikely(__pyx_t_1)) {

      /* "View.MemoryView":191
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             if not (flags & bufmode):
 *                 raise ValueError, "Can only create a buffer that is contiguous in memory."             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
      __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Can_only_create_a_buffer_that_is, 0, 0);
      __PYX_ERR(1, 191, __pyx_L1_error)

      /* "View.MemoryView":190
 *             elif self.mode == u"fortran":
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *                 raise ValueError, "Can only create a buffer that is contiguous in memory."
 *         info.buf = self.data
 */
    }

    /* "View.MemoryView":185
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if flags & (PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS):             # <<<<<<<<<<<<<<
 *             if self.mode == u"c":
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  }

  /* "View.MemoryView":192
 *             if not (flags & bufmode):
 *                 raise ValueError, "Can only create a buffer that is contiguous in memory."
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 * 
 */
  __pyx_t_2 = __pyx_v_self->data;
  __pyx_v_info->buf = __pyx_t_2;

  /* "View.MemoryView":193
 *                 raise ValueError, "Can only create a buffer that is contiguous in memory."
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  __pyx_t_3 = __pyx_v_self->len;
  __pyx_v_info->len = __pyx_t_3;

  /* "View.MemoryView":195
 *         info.len = self.len
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.ndim = self.ndim
 *             info.shape = self._shape
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":196
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *             info.shape = self._shape
 *             info.strides = self._strides
 */
    __pyx_t_4 = __pyx_v_self->ndim;
    __pyx_v_info->ndim = __pyx_t_4;

    /* "View.MemoryView":197
 *         if flags & PyBUF_STRIDES:
 *             info.ndim = self.ndim
 *             info.shape = self._shape             # <<<<<<<<<<<<<<
 *             info.strides = self._strides
 *         else:
 */
    __pyx_t_5 = __pyx_v_self->_shape;
    __pyx_v_info->shape = __pyx_t_5;

    /* "View.MemoryView":198
 *             info.ndim = self.ndim
 *             info.shape = self._shape
 *             info.strides = self._strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.ndim = 1
 */
    __pyx_t_5 = __pyx_v_self->_strides;
    __pyx_v_info->strides = __pyx_t_5;

    /* "View.MemoryView":195
 *         info.len = self.len
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.ndim = self.ndim
 *             info.shape = self._shape
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":200
 *             info.strides = self._strides
 *         else:
 *             info.ndim = 1             # <<<<<<<<<<<<<<
 *             info.shape = &self.len if flags & PyBUF_ND else NULL
 *             info.strides = NULL
 */
  /*else*/ {
    __pyx_v_info->ndim = 1;

    /* "View.MemoryView":201
 *         else:
 *             info.ndim = 1
 *             info.shape = &self.len if flags & PyBUF_ND else NULL             # <<<<<<<<<<<<<<
 *             info.strides = NULL
 * 
 */
    __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
    if (__pyx_t_1) {
      __pyx_t_5 = (&__pyx_v_self->len);
    } else {
      __pyx_t_5 = NULL;
    }
    __pyx_v_info->shape = __pyx_t_5;

    /* "View.MemoryView":202
 *             info.ndim = 1
 *             info.shape = &self.len if flags & PyBUF_ND else NULL
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.suboffsets = NULL
 */
    __pyx_v_info->strides = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":204
 *             info.strides = NULL
 * 
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":205
 * 
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 *         info.format = self.format if flags & PyBUF_FORMAT else NULL
 */
  __pyx_t_3 = __pyx_v_self->itemsize;
  __pyx_v_info->itemsize = __pyx_t_3;

  /* "View.MemoryView":206
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 *         info.format = self.format if flags & PyBUF_FORMAT else NULL
 *         info.obj = self
 */
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":207
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 *         info.format = self.format if flags & PyBUF_FORMAT else NULL             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {
    __pyx_t_2 = __pyx_v_self->format;
  } else {
    __pyx_t_2 = NULL;
  }
  __pyx_v_info->format = __pyx_t_2;

  /* "View.MemoryView":208
 *         info.readonly = 0
 *         info.format = self.format if flags & PyBUF_FORMAT else NULL
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":182
 *             _allocate_buffer(self)
 * 
 *     @cname('getbuffer')             # <<<<<<<<<<<<<<
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":210
 *         info.obj = self
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

/* Python wrapper */
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_t_1;
  int __pyx_t_2;

  /* "View.MemoryView":211
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:
 */
  __pyx_t_1 = (__pyx_v_self->callback_free_data != NULL);
  if (__pyx_t_1) {

    /* "View.MemoryView":212
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data and self.data is not NULL:
 *             if self.dtype_is_object:
 */
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":211
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":213
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 */
  if (__pyx_v_self->free_data) {
  } else {
    __pyx_t_1 = __pyx_v_self->free_data;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->data != NULL);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":214
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 *             free(self.data)
 */
    if (__pyx_v_self->dtype_is_object) {

      /* "View.MemoryView":215
 *         elif self.free_data and self.data is not NULL:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)             # <<<<<<<<<<<<<<
 *             free(self.data)
 *         PyObject_Free(self._shape)
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":214
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 *             free(self.data)
 */
    }

    /* "View.MemoryView":216
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */
    free(__pyx_v_self->data);

    /* "View.MemoryView":213
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":217
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":210
 *         info.obj = self
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

  /* function exit code */
}

/* "View.MemoryView":219
 *         PyObject_Free(self._shape)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def memview(self):
 *         return self.get_memview()
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":221
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":219
 *         PyObject_Free(self._shape)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def memview(self):
 *         return self.get_memview()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":224
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_memview", 1);

  /* "View.MemoryView":225
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":226
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self))) __PYX_ERR(1, 226, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 226, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2)) __PYX_ERR(1, 226, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":224
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":228
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;

  /* "View.MemoryView":229
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */
  __pyx_r = (__pyx_v_self->_shape[0]);
  goto __pyx_L0;

  /* "View.MemoryView":228
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":231
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattr__", 1);

  /* "View.MemoryView":232
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":231
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":234
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 1);

  /* "View.MemoryView":235
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":234
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":237
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

/* Python wrapper */
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 1);

  /* "View.MemoryView":238
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely((PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0))) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":237
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":248
 * 
 * @cname("__pyx_array_allocate_buffer")
 * cdef int _allocate_buffer(array self) except -1:             # <<<<<<<<<<<<<<
 * 
 * 
 */

static int __pyx_array_allocate_buffer(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_i;
  PyObject **__pyx_v_p;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":254
 *     cdef PyObject **p
 * 
 *     self.free_data = True             # <<<<<<<<<<<<<<
 *     self.data = <char *>malloc(self.len)
 *     if not self.data:
 */
  __pyx_v_self->free_data = 1;

  /* "View.MemoryView":255
 * 
 *     self.free_data = True
 *     self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *     if not self.data:
 *         raise MemoryError, "unable to allocate array data."
 */
  __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

  /* "View.MemoryView":256
 *     self.free_data = True
 *     self.data = <char *>malloc(self.len)
 *     if not self.data:             # <<<<<<<<<<<<<<
 *         raise MemoryError, "unable to allocate array data."
 * 
 */
  __pyx_t_1 = (!(__pyx_v_self->data != 0));
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":257
 *     self.data = <char *>malloc(self.len)
 *     if not self.data:
 *         raise MemoryError, "unable to allocate array data."             # <<<<<<<<<<<<<<
 * 
 *     if self.dtype_is_object:
 */
    __Pyx_Raise(__pyx_builtin_MemoryError, __pyx_kp_s_unable_to_allocate_array_data, 0, 0);
    __PYX_ERR(1, 257, __pyx_L1_error)

    /* "View.MemoryView":256
 *     self.free_data = True
 *     self.data = <char *>malloc(self.len)
 *     if not self.data:             # <<<<<<<<<<<<<<
 *         raise MemoryError, "unable to allocate array data."
 * 
 */
  }

  /* "View.MemoryView":259
 *         raise MemoryError, "unable to allocate array data."
 * 
 *     if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *         p = <PyObject **> self.data
 *         for i in range(self.len // self.itemsize):
 */
  if (__pyx_v_self->dtype_is_object) {

    /* "View.MemoryView":260
 * 
 *     if self.dtype_is_object:
 *         p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *         for i in range(self.len // self.itemsize):
 *             p[i] = Py_None
 */
    __pyx_v_p = ((PyObject **)__pyx_v_self->data);

    /* "View.MemoryView":261
 *     if self.dtype_is_object:
 *         p = <PyObject **> self.data
 *         for i in range(self.len // self.itemsize):             # <<<<<<<<<<<<<<
 *             p[i] = Py_None
 *             Py_INCREF(Py_None)
 */
    if (unlikely(__pyx_v_self->itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(1, 261, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_self->itemsize == (Py_ssize_t)-1)  && unlikely(__Pyx_UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(1, 261, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_div_Py_ssize_t(__pyx_v_self->len, __pyx_v_self->itemsize);
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":262
 *         p = <PyObject **> self.data
 *         for i in range(self.len // self.itemsize):
 *             p[i] = Py_None             # <<<<<<<<<<<<<<
 *             Py_INCREF(Py_None)
 *     return 0
 */
      (__pyx_v_p[__pyx_v_i]) = Py_None;

      /* "View.MemoryView":263
 *         for i in range(self.len // self.itemsize):
 *             p[i] = Py_None
 *             Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
      Py_INCREF(Py_None);
    }

    /* "View.MemoryView":259
 *         raise MemoryError, "unable to allocate array data."
 * 
 *     if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *         p = <PyObject **> self.data
 *         for i in range(self.len // self.itemsize):
 */
  }

  /* "View.MemoryView":264
 *             p[i] = Py_None
 *             Py_INCREF(Py_None)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":248
 * 
 * @cname("__pyx_array_allocate_buffer")
 * cdef int _allocate_buffer(array self) except -1:             # <<<<<<<<<<<<<<
 * 
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._allocate_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":268
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format, char *c_mode, char *buf):             # <<<<<<<<<<<<<<
 *     cdef array result
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.
 */

static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_c_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0;
  PyObject *__pyx_v_mode = 0;
  struct __pyx_array_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("array_cwrapper", 1);

  /* "View.MemoryView":270
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format, char *c_mode, char *buf):
 *     cdef array result
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.             # <<<<<<<<<<<<<<
 * 
 *     if buf is NULL:
 */
  __pyx_t_2 = ((__pyx_v_c_mode[0]) == 'f');
  if (__pyx_t_2) {
    __Pyx_INCREF(__pyx_n_s_fortran);
    __pyx_t_1 = __pyx_n_s_fortran;
  } else {
    __Pyx_INCREF(__pyx_n_s_c);
    __pyx_t_1 = __pyx_n_s_c;
  }
  __pyx_v_mode = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":272
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.
 * 
 *     if buf is NULL:             # <<<<<<<<<<<<<<
 *         result = array.__new__(array, shape, itemsize, format, mode)
 *     else:
 */
  __pyx_t_2 = (__pyx_v_buf == NULL);
  if (__pyx_t_2) {

    /* "View.MemoryView":273
 * 
 *     if buf is NULL:
 *         result = array.__new__(array, shape, itemsize, format, mode)             # <<<<<<<<<<<<<<
 *     else:
 *         result = array.__new__(array, shape, itemsize, format, mode, allocate_buffer=False)
 */
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_shape)) __PYX_ERR(1, 273, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 273, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3)) __PYX_ERR(1, 273, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_mode);
    __Pyx_GIVEREF(__pyx_v_mode);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_mode)) __PYX_ERR(1, 273, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = ((PyObject *)__pyx_tp_new_array(((PyTypeObject *)__pyx_array_type), __pyx_t_4, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 273, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":272
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.
 * 
 *     if buf is NULL:             # <<<<<<<<<<<<<<
 *         result = array.__new__(array, shape, itemsize, format, mode)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":275
 *         result = array.__new__(array, shape, itemsize, format, mode)
 *     else:
 *         result = array.__new__(array, shape, itemsize, format, mode, allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */
  /*else*/ {
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_shape)) __PYX_ERR(1, 275, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3)) __PYX_ERR(1, 275, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4)) __PYX_ERR(1, 275, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_mode);
    __Pyx_GIVEREF(__pyx_v_mode);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_mode)) __PYX_ERR(1, 275, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(1, 275, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_tp_new_array(((PyTypeObject *)__pyx_array_type), __pyx_t_1, __pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 275, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":276
 *     else:
 *         result = array.__new__(array, shape, itemsize, format, mode, allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_v_result->data = __pyx_v_buf;
  }
  __pyx_L3:;

  /* "View.MemoryView":278
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":268
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format, char *c_mode, char *buf):             # <<<<<<<<<<<<<<
 *     cdef array result
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_mode);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":304
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

/* Python wrapper */
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 304, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 304, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 304, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "View.MemoryView":305
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":304
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":306
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "View.MemoryView":307
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "View.MemoryView":306
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_136983863);
    __Pyx_GIVEREF(__pyx_int_136983863);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_136983863)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_136983863);
    __Pyx_GIVEREF(__pyx_int_136983863);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_136983863)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":349
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

/* Python wrapper */
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_v_flags;
  int __pyx_v_dtype_is_object;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_obj)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 349, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_flags)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 349, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(1, 349, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dtype_is_object);
          if (value) { values[2] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 349, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 349, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 349, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 349, __pyx_L3_error)
    } else {
      __pyx_v_dtype_is_object = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, __pyx_nargs); __PYX_ERR(1, 349, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_intptr_t __pyx_t_4;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "View.MemoryView":350
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":351
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":352
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_obj != Py_None);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":353
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */
    __pyx_t_3 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 353, __pyx_L1_error)

    /* "View.MemoryView":354
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->view.obj) == NULL);
    if (__pyx_t_1) {

      /* "View.MemoryView":355
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":356
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 */
      Py_INCREF(Py_None);

      /* "View.MemoryView":354
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    }

    /* "View.MemoryView":352
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  }

  /* "View.MemoryView":358
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < 8:
 */
  __pyx_t_1 = (!__PYX_CYTHON_ATOMICS_ENABLED());
  if (__pyx_t_1) {

    /* "View.MemoryView":360
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < 8:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */
    __pyx_t_1 = (__pyx_memoryview_thread_locks_used < 8);
    if (__pyx_t_1) {

      /* "View.MemoryView":361
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < 8:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 */
      __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

      /* "View.MemoryView":362
 *             if __pyx_memoryview_thread_locks_used < 8:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 */
      __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

      /* "View.MemoryView":360
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < 8:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */
    }

    /* "View.MemoryView":363
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */
    __pyx_t_1 = (__pyx_v_self->lock == NULL);
    if (__pyx_t_1) {

      /* "View.MemoryView":364
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *                 if self.lock is NULL:
 *                     raise MemoryError
 */
      __pyx_v_self->lock = PyThread_allocate_lock();

      /* "View.MemoryView":365
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */
      __pyx_t_1 = (__pyx_v_self->lock == NULL);
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":366
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 *                     raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
        PyErr_NoMemory(); __PYX_ERR(1, 366, __pyx_L1_error)

        /* "View.MemoryView":365
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */
      }

      /* "View.MemoryView":363
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */
    }

    /* "View.MemoryView":358
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < 8:
 */
  }

  /* "View.MemoryView":368
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":369
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */
    __pyx_t_2 = ((__pyx_v_self->view.format[0]) == 'O');
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_self->view.format[1]) == '\x00');
    __pyx_t_1 = __pyx_t_2;
    __pyx_L12_bool_binop_done:;
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":368
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
    goto __pyx_L11;
  }

  /* "View.MemoryView":371
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         assert <Py_intptr_t><void*>(&self.acquisition_count) % sizeof(__pyx_atomic_int_type) == 0
 */
  /*else*/ {
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;
  }
  __pyx_L11:;

  /* "View.MemoryView":373
 *             self.dtype_is_object = dtype_is_object
 * 
 *         assert <Py_intptr_t><void*>(&self.acquisition_count) % sizeof(__pyx_atomic_int_type) == 0             # <<<<<<<<<<<<<<
 *         self.typeinfo = NULL
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_4 = ((Py_intptr_t)((void *)(&__pyx_v_self->acquisition_count)));
    __pyx_t_5 = (sizeof(__pyx_atomic_int_type));
    if (unlikely(__pyx_t_5 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(1, 373, __pyx_L1_error)
    }
    __pyx_t_1 = ((__pyx_t_4 % __pyx_t_5) == 0);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(1, 373, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(1, 373, __pyx_L1_error)
  #endif

  /* "View.MemoryView":374
 * 
 *         assert <Py_intptr_t><void*>(&self.acquisition_count) % sizeof(__pyx_atomic_int_type) == 0
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":349
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":376
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

/* Python wrapper */
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyThread_type_lock __pyx_t_5;
  PyThread_type_lock __pyx_t_6;

  /* "View.MemoryView":377
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);
  if (__pyx_t_1) {

    /* "View.MemoryView":378
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":377
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":379
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  __pyx_t_1 = (((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None);
  if (__pyx_t_1) {

    /* "View.MemoryView":381
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":382
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
    Py_DECREF(Py_None);

    /* "View.MemoryView":379
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":386
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  __pyx_t_1 = (__pyx_v_self->lock != NULL);
  if (__pyx_t_1) {

    /* "View.MemoryView":387
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */
    __pyx_t_2 = __pyx_memoryview_thread_locks_used;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":388
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      __pyx_t_1 = ((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock);
      if (__pyx_t_1) {

        /* "View.MemoryView":389
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":390
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        __pyx_t_1 = (__pyx_v_i != __pyx_memoryview_thread_locks_used);
        if (__pyx_t_1) {

          /* "View.MemoryView":392
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
          __pyx_t_5 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":391
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_5;
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_6;

          /* "View.MemoryView":390
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        }

        /* "View.MemoryView":393
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */
        goto __pyx_L6_break;

        /* "View.MemoryView":388
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      }
    }
    /*else*/ {

      /* "View.MemoryView":395
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */
      PyThread_free_lock(__pyx_v_self->lock);
    }
    __pyx_L6_break:;

    /* "View.MemoryView":386
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  }

  /* "View.MemoryView":376
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

  /* function exit code */
}

/* "View.MemoryView":397
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim;
  char *__pyx_v_itemp;
  PyObject *__pyx_v_idx = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  char *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_item_pointer", 1);

  /* "View.MemoryView":399
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":401
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 401, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 401, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(1, 401, __pyx_L1_error)
        #else
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 401, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(1, 401, __pyx_L1_error)
        #else
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 401, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":402
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 402, __pyx_L1_error)
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(1, 402, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":401
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":404
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_itemp;
  goto __pyx_L0;

  /* "View.MemoryView":397
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":407
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_indices = NULL;
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  char *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 1);

  /* "View.MemoryView":408
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);
  if (__pyx_t_1) {

    /* "View.MemoryView":409
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "View.MemoryView":408
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "View.MemoryView":411
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 411, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 411, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_indices = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "View.MemoryView":414
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 414, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "View.MemoryView":415
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":414
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  }

  /* "View.MemoryView":417
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */
  /*else*/ {
    __pyx_t_5 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_5 == ((char *)NULL))) __PYX_ERR(1, 417, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_5;

    /* "View.MemoryView":418
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":407
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":420
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError, "Cannot assign to read-only memoryview"
 */

/* Python wrapper */
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_obj = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":421
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError, "Cannot assign to read-only memoryview"
 * 
 */
  if (unlikely(__pyx_v_self->view.readonly)) {

    /* "View.MemoryView":422
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError, "Cannot assign to read-only memoryview"             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
    __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_Cannot_assign_to_read_only_memor, 0, 0);
    __PYX_ERR(1, 422, __pyx_L1_error)

    /* "View.MemoryView":421
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError, "Cannot assign to read-only memoryview"
 * 
 */
  }

  /* "View.MemoryView":424
 *             raise TypeError, "Cannot assign to read-only memoryview"
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */
  __pyx_t_1 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(__pyx_t_1 != Py_None)) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 424, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 424, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_2;
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":426
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(1, 426, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "View.MemoryView":427
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_obj = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "View.MemoryView":428
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(1, 428, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "View.MemoryView":429
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */
      __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_1, __pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":428
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":431
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 431, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(1, 431, __pyx_L1_error)
      __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_3), __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 431, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L5:;

    /* "View.MemoryView":426
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":433
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */
  /*else*/ {
    __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L4:;

  /* "View.MemoryView":420
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError, "Cannot assign to read-only memoryview"
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":435
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_slice", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":436
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "View.MemoryView":437
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "View.MemoryView":438
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 438, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":439
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 439, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":438
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 438, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj)) __PYX_ERR(1, 438, __pyx_L4_error);
        __Pyx_GIVEREF(__pyx_t_6);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6)) __PYX_ERR(1, 438, __pyx_L4_error);
        __Pyx_GIVEREF(__pyx_t_7);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7)) __PYX_ERR(1, 438, __pyx_L4_error);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 438, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "View.MemoryView":437
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":440
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(1, 440, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);

        /* "View.MemoryView":441
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;

      /* "View.MemoryView":437
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L9_try_end:;
    }

    /* "View.MemoryView":436
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  }

  /* "View.MemoryView":443
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "View.MemoryView":435
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":445
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_src_slice;
  __Pyx_memviewslice __pyx_v_msrc;
  __Pyx_memviewslice __pyx_v_mdst;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 1);

  /* "View.MemoryView":448
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 *         cdef __Pyx_memviewslice msrc = get_slice_from_memview(src, &src_slice)[0]             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mdst = get_slice_from_memview(dst, &dst_slice)[0]
 * 
 */
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(1, 448, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 448, __pyx_L1_error)
  __pyx_v_msrc = (__pyx_t_1[0]);

  /* "View.MemoryView":449
 *         cdef __Pyx_memviewslice src_slice
 *         cdef __Pyx_memviewslice msrc = get_slice_from_memview(src, &src_slice)[0]
 *         cdef __Pyx_memviewslice mdst = get_slice_from_memview(dst, &dst_slice)[0]             # <<<<<<<<<<<<<<
 * 
 *         memoryview_copy_contents(msrc, mdst, src.ndim, dst.ndim, self.dtype_is_object)
 */
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(1, 449, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 449, __pyx_L1_error)
  __pyx_v_mdst = (__pyx_t_1[0]);

  /* "View.MemoryView":451
 *         cdef __Pyx_memviewslice mdst = get_slice_from_memview(dst, &dst_slice)[0]
 * 
 *         memoryview_copy_contents(msrc, mdst, src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 451, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 451, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __pyx_memoryview_copy_contents(__pyx_v_msrc, __pyx_v_mdst, __pyx_t_3, __pyx_t_4, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 451, __pyx_L1_error)

  /* "View.MemoryView":445
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":453
 *         memoryview_copy_contents(msrc, mdst, src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80];
  void *__pyx_v_tmp;
  void *__pyx_v_item;
  __Pyx_memviewslice *__pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_tmp_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 1);

  /* "View.MemoryView":455
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":460
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 460, __pyx_L1_error)
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":462
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
  __pyx_t_2 = (((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array)));
  if (__pyx_t_2) {

    /* "View.MemoryView":463
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":464
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    __pyx_t_2 = (__pyx_v_tmp == NULL);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":465
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */
      PyErr_NoMemory(); __PYX_ERR(1, 465, __pyx_L1_error)

      /* "View.MemoryView":464
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    }

    /* "View.MemoryView":466
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":462
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":468
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  /*else*/ {
    __pyx_v_item = ((void *)__pyx_v_array);
  }
  __pyx_L3:;

  /* "View.MemoryView":470
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */
  /*try:*/ {

    /* "View.MemoryView":471
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
    if (__pyx_v_self->dtype_is_object) {

      /* "View.MemoryView":472
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":471
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":474
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 474, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L8:;

    /* "View.MemoryView":478
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    __pyx_t_2 = (__pyx_v_self->view.suboffsets != NULL);
    if (__pyx_t_2) {

      /* "View.MemoryView":479
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */
      __pyx_t_4 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 479, __pyx_L6_error)

      /* "View.MemoryView":478
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    }

    /* "View.MemoryView":480
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);
  }

  /* "View.MemoryView":483
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyMem_Free(__pyx_v_tmp);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_tmp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "View.MemoryView":453
 *         memoryview_copy_contents(msrc, mdst, src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":485
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_indexed", 1);

  /* "View.MemoryView":486
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(1, 486, __pyx_L1_error)
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":487
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":485
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":489
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_bytesitem = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 1);

  /* "View.MemoryView":492
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":495
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":496
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "View.MemoryView":497
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError, "Unable to convert item to object"
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 497, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 497, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 497, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_v_result = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "View.MemoryView":496
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    }

    /* "View.MemoryView":501
 *             raise ValueError, "Unable to convert item to object"
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
    /*else:*/ {
      __pyx_t_9 = __Pyx_ssize_strlen(__pyx_v_self->view.format); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(1, 501, __pyx_L5_except_error)
      __pyx_t_10 = (__pyx_t_9 == 1);
      if (__pyx_t_10) {

        /* "View.MemoryView":502
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 502, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L6_except_return;

        /* "View.MemoryView":501
 *             raise ValueError, "Unable to convert item to object"
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
      }

      /* "View.MemoryView":503
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "View.MemoryView":498
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Unable to convert item to object"
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 498, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_6);
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(1, 498, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_1);

      /* "View.MemoryView":499
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError, "Unable to convert item to object"             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
      __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Unable_to_convert_item_to_object, 0, 0);
      __PYX_ERR(1, 499, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "View.MemoryView":496
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "View.MemoryView":489
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesitem);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":505
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL;
  char __pyx_v_c;
  PyObject *__pyx_v_bytesvalue = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  char *__pyx_t_10;
  char *__pyx_t_11;
  char *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 1);

  /* "View.MemoryView":508
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":513
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  if (__pyx_t_2) {

    /* "View.MemoryView":514
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 514, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Add(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(1, 514, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":513
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":516
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */
  /*else*/ {
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_1, __pyx_v_value};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 516, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(1, 516, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":518
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
  __pyx_t_7 = 0;
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(1, 518, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_bytesvalue);
  __pyx_t_8 = __pyx_v_bytesvalue;
  __pyx_t_10 = PyBytes_AS_STRING(__pyx_t_8);
  __pyx_t_11 = (__pyx_t_10 + PyBytes_GET_SIZE(__pyx_t_8));
  for (__pyx_t_12 = __pyx_t_10; __pyx_t_12 < __pyx_t_11; __pyx_t_12++) {
    __pyx_t_9 = __pyx_t_12;
    __pyx_v_c = (__pyx_t_9[0]);

    /* "View.MemoryView":519
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_v_i = __pyx_t_7;

    /* "View.MemoryView":518
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
    __pyx_t_7 = (__pyx_t_7 + 1);

    /* "View.MemoryView":519
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "View.MemoryView":505
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":521
 *             itemp[i] = c
 * 
 *     @cname('getbuffer')             # <<<<<<<<<<<<<<
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 */

/* Python wrapper */
CYTHON_UNUSED static int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
CYTHON_UNUSED static int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  char *__pyx_t_4;
  void *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (unlikely(__pyx_v_info == NULL)) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":523
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"
 * 
 */
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":524
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Cannot_create_writable_memory_vi, 0, 0);
    __PYX_ERR(1, 524, __pyx_L1_error)

    /* "View.MemoryView":523
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"
 * 
 */
  }

  /* "View.MemoryView":526
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":527
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */
    __pyx_t_3 = __pyx_v_self->view.shape;
    __pyx_v_info->shape = __pyx_t_3;

    /* "View.MemoryView":526
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":529
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  /*else*/ {
    __pyx_v_info->shape = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":531
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":532
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */
    __pyx_t_3 = __pyx_v_self->view.strides;
    __pyx_v_info->strides = __pyx_t_3;

    /* "View.MemoryView":531
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
    goto __pyx_L7;
  }

  /* "View.MemoryView":534
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */
  /*else*/ {
    __pyx_v_info->strides = NULL;
  }
  __pyx_L7:;

  /* "View.MemoryView":536
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":537
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */
    __pyx_t_3 = __pyx_v_self->view.suboffsets;
    __pyx_v_info->suboffsets = __pyx_t_3;

    /* "View.MemoryView":536
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
    goto __pyx_L8;
  }

  /* "View.MemoryView":539
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  /*else*/ {
    __pyx_v_info->suboffsets = NULL;
  }
  __pyx_L8:;

  /* "View.MemoryView":541
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":542
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.format;
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":541
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":544
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L9:;

  /* "View.MemoryView":546
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */
  __pyx_t_5 = __pyx_v_self->view.buf;
  __pyx_v_info->buf = __pyx_t_5;

  /* "View.MemoryView":547
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */
  __pyx_t_6 = __pyx_v_self->view.ndim;
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":548
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */
  __pyx_t_7 = __pyx_v_self->view.itemsize;
  __pyx_v_info->itemsize = __pyx_t_7;

  /* "View.MemoryView":549
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */
  __pyx_t_7 = __pyx_v_self->view.len;
  __pyx_v_info->len = __pyx_t_7;

  /* "View.MemoryView":550
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":551
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":521
 *             itemp[i] = c
 * 
 *     @cname('getbuffer')             # <<<<<<<<<<<<<<
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":554
 * 
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":556
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(1, 556, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":557
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(1, 557, __pyx_L1_error)

  /* "View.MemoryView":558
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":554
 * 
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":560
 *         return result
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def base(self):
 *         return self._get_base()
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":562
 *     @property
 *     def base(self):
 *         return self._get_base()             # <<<<<<<<<<<<<<
 * 
 *     cdef _get_base(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->_get_base(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":560
 *         return result
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def base(self):
 *         return self._get_base()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.base.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":564
 *         return self._get_base()
 * 
 *     cdef _get_base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

static PyObject *__pyx_memoryview__get_base(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_base", 1);

  /* "View.MemoryView":565
 * 
 *     cdef _get_base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* "View.MemoryView":564
 *         return self._get_base()
 * 
 *     cdef _get_base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":567
 *         return self.obj
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_7genexpr__pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":569
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
      __pyx_t_2 = __pyx_t_4;
      __pyx_7genexpr__pyx_v_length = (__pyx_t_2[0]);
      __pyx_t_5 = PyInt_FromSsize_t(__pyx_7genexpr__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 569, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":567
 *         return self.obj
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":571
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def strides(self):
 *         if self.view.strides == NULL:
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_8genexpr1__pyx_v_stride;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":573
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError, "Buffer view does not expose strides"
 */
  __pyx_t_1 = (__pyx_v_self->view.strides == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":575
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError, "Buffer view does not expose strides"             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Buffer_view_does_not_expose_stri, 0, 0);
    __PYX_ERR(1, 575, __pyx_L1_error)

    /* "View.MemoryView":573
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError, "Buffer view does not expose strides"
 */
  }

  /* "View.MemoryView":577
 *             raise ValueError, "Buffer view does not expose strides"
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_8genexpr1__pyx_v_stride = (__pyx_t_3[0]);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_8genexpr1__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(1, 577, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  } /* exit inner scope */
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":571
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def strides(self):
 *         if self.view.strides == NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":579
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_8genexpr2__pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":581
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.suboffsets == NULL);
  if (__pyx_t_1) {

    /* "View.MemoryView":582
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PySequence_Multiply(__pyx_tuple__4, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 582, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":581
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  }

  /* "View.MemoryView":584
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.suboffsets; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_8genexpr2__pyx_v_suboffset = (__pyx_t_3[0]);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_8genexpr2__pyx_v_suboffset); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(1, 584, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  } /* exit inner scope */
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":579
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":586
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ndim(self):
 *         return self.view.ndim
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":588
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":586
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ndim(self):
 *         return self.view.ndim
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":590
 *         return self.view.ndim
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def itemsize(self):
 *         return self.view.itemsize
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":592
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":590
 *         return self.view.ndim
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def itemsize(self):
 *         return self.view.itemsize
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":594
 *         return self.view.itemsize
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def nbytes(self):
 *         return self.size * self.view.itemsize
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":596
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":594
 *         return self.view.itemsize
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def nbytes(self):
 *         return self.size * self.view.itemsize
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":598
 *         return self.size * self.view.itemsize
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def size(self):
 *         if self._size is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":600
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);
  if (__pyx_t_1) {

    /* "View.MemoryView":601
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */
    __Pyx_INCREF(__pyx_int_1);
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":603
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */
    __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
      __pyx_t_2 = __pyx_t_4;
      __pyx_t_5 = PyInt_FromSsize_t((__pyx_t_2[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "View.MemoryView":604
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */
      __pyx_t_5 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_5);
      __pyx_t_5 = 0;
    }

    /* "View.MemoryView":606
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_size);
    __Pyx_DECREF(__pyx_v_self->_size);
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":600
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  }

  /* "View.MemoryView":608
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_size);
  __pyx_r = __pyx_v_self->_size;
  goto __pyx_L0;

  /* "View.MemoryView":598
 *         return self.size * self.view.itemsize
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def size(self):
 *         if self._size is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":610
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

/* Python wrapper */
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":611
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.ndim >= 1);
  if (__pyx_t_1) {

    /* "View.MemoryView":612
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_r = (__pyx_v_self->view.shape[0]);
    goto __pyx_L0;

    /* "View.MemoryView":611
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  }

  /* "View.MemoryView":614
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":610
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":616
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "View.MemoryView":617
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":618
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":617
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(1, 617, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(1, 617, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":616
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":620
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "View.MemoryView":621
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 621, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":620
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":624
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_c_contig", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_c_contig", 0))) return NULL;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_c_contig", 1);

  /* "View.MemoryView":627
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 627, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":628
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":624
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":630
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_f_contig", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_f_contig", 0))) return NULL;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_f_contig", 1);

  /* "View.MemoryView":633
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 633, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":634
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":630
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":636
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "View.MemoryView":638
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":640
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":641
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 641, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":646
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":636
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":648
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy_fortran", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy_fortran", 0))) return NULL;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy_fortran", 1);

  /* "View.MemoryView":650
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":652
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":653
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 653, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":658
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":648
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":662
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 1);

  /* "View.MemoryView":663
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o)) __PYX_ERR(1, 663, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 663, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2)) __PYX_ERR(1, 663, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":664
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":665
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":662
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":668
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {
  int __pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":669
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o) noexcept:
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":668
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":671
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_idx;
  PyObject *__pyx_v_tup = NULL;
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_have_slices;
  int __pyx_v_seen_ellipsis;
  PyObject *__pyx_v_item = NULL;
  Py_ssize_t __pyx_v_nslices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unellipsify", 1);

  /* "View.MemoryView":677
 *     """
 *     cdef Py_ssize_t idx
 *     tup = <tuple>index if isinstance(index, tuple) else (index,)             # <<<<<<<<<<<<<<
 * 
 *     result = [slice(None)] * ndim
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_index); 
  if (__pyx_t_2) {
    __Pyx_INCREF(((PyObject*)__pyx_v_index));
    __pyx_t_1 = __pyx_v_index;
  } else {
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 677, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index)) __PYX_ERR(1, 677, __pyx_L1_error);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_v_tup = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":679
 *     tup = <tuple>index if isinstance(index, tuple) else (index,)
 * 
 *     result = [slice(None)] * ndim             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_t_1 = PyList_New(1 * ((__pyx_v_ndim<0) ? 0:__pyx_v_ndim)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_v_ndim; __pyx_temp++) {
      __Pyx_INCREF(__pyx_slice__5);
      __Pyx_GIVEREF(__pyx_slice__5);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_1, __pyx_temp, __pyx_slice__5)) __PYX_ERR(1, 679, __pyx_L1_error);
    }
  }
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":680
 * 
 *     result = [slice(None)] * ndim
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     idx = 0
 */
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":681
 *     result = [slice(None)] * ndim
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     idx = 0
 *     for item in tup:
 */
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":682
 *     have_slices = False
 *     seen_ellipsis = False
 *     idx = 0             # <<<<<<<<<<<<<<
 *     for item in tup:
 *         if item is Ellipsis:
 */
  __pyx_v_idx = 0;

  /* "View.MemoryView":683
 *     seen_ellipsis = False
 *     idx = 0
 *     for item in tup:             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  if (unlikely(__pyx_v_tup == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 683, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_4 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 683, __pyx_L1_error)
      #endif
      if (__pyx_t_4 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(1, 683, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":684
 *     idx = 0
 *     for item in tup:
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 idx += ndim - len(tup)
 */
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);
    if (__pyx_t_2) {

      /* "View.MemoryView":685
 *     for item in tup:
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 idx += ndim - len(tup)
 *                 seen_ellipsis = True
 */
      __pyx_t_2 = (!__pyx_v_seen_ellipsis);
      if (__pyx_t_2) {

        /* "View.MemoryView":686
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 idx += ndim - len(tup)             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             have_slices = True
 */
        if (unlikely(__pyx_v_tup == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
          __PYX_ERR(1, 686, __pyx_L1_error)
        }
        __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v_tup); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 686, __pyx_L1_error)
        __pyx_v_idx = (__pyx_v_idx + (__pyx_v_ndim - __pyx_t_5));

        /* "View.MemoryView":687
 *             if not seen_ellipsis:
 *                 idx += ndim - len(tup)
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":685
 *     for item in tup:
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 idx += ndim - len(tup)
 *                 seen_ellipsis = True
 */
      }

      /* "View.MemoryView":688
 *                 idx += ndim - len(tup)
 *                 seen_ellipsis = True
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if isinstance(item, slice):
 */
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":684
 *     idx = 0
 *     for item in tup:
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 idx += ndim - len(tup)
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if isinstance(item, slice):             # <<<<<<<<<<<<<<
 *                 have_slices = True
 *             elif not PyIndex_Check(item):
 */
    /*else*/ {
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 
      if (__pyx_t_2) {

        /* "View.MemoryView":691
 *         else:
 *             if isinstance(item, slice):
 *                 have_slices = True             # <<<<<<<<<<<<<<
 *             elif not PyIndex_Check(item):
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 */
        __pyx_v_have_slices = 1;

        /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if isinstance(item, slice):             # <<<<<<<<<<<<<<
 *                 have_slices = True
 *             elif not PyIndex_Check(item):
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":692
 *             if isinstance(item, slice):
 *                 have_slices = True
 *             elif not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 *             result[idx] = item
 */
      __pyx_t_2 = (!(PyIndex_Check(__pyx_v_item) != 0));
      if (unlikely(__pyx_t_2)) {

        /* "View.MemoryView":693
 *                 have_slices = True
 *             elif not PyIndex_Check(item):
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"             # <<<<<<<<<<<<<<
 *             result[idx] = item
 *         idx += 1
 */
        __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 693, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = 0;
        __pyx_t_6 = 127;
        __Pyx_INCREF(__pyx_kp_u_Cannot_index_with_type);
        __pyx_t_5 += 24;
        __Pyx_GIVEREF(__pyx_kp_u_Cannot_index_with_type);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Cannot_index_with_type);
        __pyx_t_7 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_item)), __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 693, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
        __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_7);
        __pyx_t_7 = 0;
        __Pyx_INCREF(__pyx_kp_u__6);
        __pyx_t_5 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__6);
        PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__6);
        __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 693, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_builtin_TypeError, __pyx_t_7, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(1, 693, __pyx_L1_error)

        /* "View.MemoryView":692
 *             if isinstance(item, slice):
 *                 have_slices = True
 *             elif not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 *             result[idx] = item
 */
      }
      __pyx_L7:;

      /* "View.MemoryView":694
 *             elif not PyIndex_Check(item):
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 *             result[idx] = item             # <<<<<<<<<<<<<<
 *         idx += 1
 * 
 */
      if (unlikely((__Pyx_SetItemInt(__pyx_v_result, __pyx_v_idx, __pyx_v_item, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1) < 0))) __PYX_ERR(1, 694, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "View.MemoryView":695
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 *             result[idx] = item
 *         idx += 1             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - idx
 */
    __pyx_v_idx = (__pyx_v_idx + 1);

    /* "View.MemoryView":683
 *     seen_ellipsis = False
 *     idx = 0
 *     for item in tup:             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":697
 *         idx += 1
 * 
 *     nslices = ndim - idx             # <<<<<<<<<<<<<<
 *     return have_slices or nslices, tuple(result)
 * 
 */
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_v_idx);

  /* "View.MemoryView":698
 * 
 *     nslices = ndim - idx
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef int assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  if (!__pyx_v_have_slices) {
  } else {
    __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_7 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __pyx_t_7;
  __pyx_t_7 = 0;
  __pyx_L9_bool_binop_done:;
  __pyx_t_7 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(1, 698, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_7)) __PYX_ERR(1, 698, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_7 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":671
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tup);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef int assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim) except -1:             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

static int assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset;
  int __pyx_r;
  Py_ssize_t *__pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":701
 * 
 * cdef int assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim) except -1:
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError, "Indirect dimensions not supported"
 */
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {
    __pyx_t_1 = __pyx_t_3;
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":702
 * cdef int assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim) except -1:
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Indirect dimensions not supported"
 *     return 0  # return type just used as an error flag
 */
    __pyx_t_4 = (__pyx_v_suboffset >= 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError, "Indirect dimensions not supported"             # <<<<<<<<<<<<<<
 *     return 0  # return type just used as an error flag
 * 
 */
      __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Indirect_dimensions_not_supporte, 0, 0);
      __PYX_ERR(1, 703, __pyx_L1_error)

      /* "View.MemoryView":702
 * cdef int assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim) except -1:
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Indirect dimensions not supported"
 *     return 0  # return type just used as an error flag
 */
    }
  }

  /* "View.MemoryView":704
 *         if suboffset >= 0:
 *             raise ValueError, "Indirect dimensions not supported"
 *     return 0  # return type just used as an error flag             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef int assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim) except -1:             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":711
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim;
  int __pyx_v_suboffset_dim;
  int __pyx_v_dim;
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  __Pyx_memviewslice *__pyx_v_p_src;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0;
  __Pyx_memviewslice *__pyx_v_p_dst;
  int *__pyx_v_p_suboffset_dim;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  Py_ssize_t __pyx_v_cindex;
  int __pyx_v_have_start;
  int __pyx_v_have_stop;
  int __pyx_v_have_step;
  PyObject *__pyx_v_index = NULL;
  struct __pyx_memoryview_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_memoryview_obj *__pyx_t_3;
  char *__pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memview_slice", 1);

  /* "View.MemoryView":712
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */
  __pyx_v_new_ndim = 0;
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":719
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":723
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (__pyx_v_memview->view.ndim > 0);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(1, 723, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(1, 723, __pyx_L1_error)
  #endif

  /* "View.MemoryView":725
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  if (__pyx_t_1) {

    /* "View.MemoryView":726
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(1, 726, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":727
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":725
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":729
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":730
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_p_src = (&__pyx_v_src);
  }
  __pyx_L3:;

  /* "View.MemoryView":736
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */
  __pyx_t_3 = __pyx_v_p_src->memview;
  __pyx_v_dst.memview = __pyx_t_3;

  /* "View.MemoryView":737
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __pyx_v_p_src->data;
  __pyx_v_dst.data = __pyx_t_4;

  /* "View.MemoryView":742
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step, cindex
 */
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":743
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step, cindex
 *     cdef bint have_start, have_stop, have_step
 */
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":747
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             cindex = index
 */
  __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {
    __pyx_t_2 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 747, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 747, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(1, 747, __pyx_L1_error)
        #else
        __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 747, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 747, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(1, 747, __pyx_L1_error)
        #else
        __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 747, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      }
    } else {
      __pyx_t_8 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_8)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 747, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_v_dim = __pyx_t_5;
    __pyx_t_5 = (__pyx_t_5 + 1);

    /* "View.MemoryView":748
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             cindex = index
 *             slice_memviewslice(
 */
    __pyx_t_1 = (PyIndex_Check(__pyx_v_index) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":749
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             cindex = index             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 749, __pyx_L1_error)
      __pyx_v_cindex = __pyx_t_9;

      /* "View.MemoryView":750
 *         if PyIndex_Check(index):
 *             cindex = index
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_10 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_cindex, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(1, 750, __pyx_L1_error)

      /* "View.MemoryView":748
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             cindex = index
 *             slice_memviewslice(
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
    __pyx_t_1 = (__pyx_v_index == Py_None);
    if (__pyx_t_1) {

      /* "View.MemoryView":757
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":758
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":759
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":760
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":762
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */
    /*else*/ {
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 762, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else {
        __pyx_t_11 = __Pyx_PyIndex_AsSsize_t(__pyx_t_8); if (unlikely((__pyx_t_11 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 762, __pyx_L1_error)
        __pyx_t_9 = __pyx_t_11;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_9 = 0;
      __pyx_L7_bool_binop_done:;
      __pyx_v_start = __pyx_t_9;

      /* "View.MemoryView":763
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 763, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 763, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else {
        __pyx_t_11 = __Pyx_PyIndex_AsSsize_t(__pyx_t_8); if (unlikely((__pyx_t_11 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 763, __pyx_L1_error)
        __pyx_t_9 = __pyx_t_11;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_9 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_v_stop = __pyx_t_9;

      /* "View.MemoryView":764
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 764, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 764, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else {
        __pyx_t_11 = __Pyx_PyIndex_AsSsize_t(__pyx_t_8); if (unlikely((__pyx_t_11 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 764, __pyx_L1_error)
        __pyx_t_9 = __pyx_t_11;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_9 = 0;
      __pyx_L11_bool_binop_done:;
      __pyx_v_step = __pyx_t_9;

      /* "View.MemoryView":766
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = (__pyx_t_8 != Py_None);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":767
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 767, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = (__pyx_t_8 != Py_None);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 768, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = (__pyx_t_8 != Py_None);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":770
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_10 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(1, 770, __pyx_L1_error)

      /* "View.MemoryView":776
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);
    }
    __pyx_L6:;

    /* "View.MemoryView":747
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             cindex = index
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  if (__pyx_t_1) {

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);

    /* "View.MemoryView":780
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(1, 780, __pyx_L1_error) }

    /* "View.MemoryView":781
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(1, 781, __pyx_L1_error) }

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 779, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_memoryview_type))))) __PYX_ERR(1, 779, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  }

  /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF((PyObject *)__pyx_r);

    /* "View.MemoryView":785
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 784, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_memoryview_type))))) __PYX_ERR(1, 784, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":711
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":793
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape;
  int __pyx_v_negative_step;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "View.MemoryView":813
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
  __pyx_t_1 = (!__pyx_v_is_slice);
  if (__pyx_t_1) {

    /* "View.MemoryView":815
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    __pyx_t_1 = (__pyx_v_start < 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":816
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(PyExc_IndexError, "Index out of bounds (axis %d)", dim)
 */
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":815
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    }

    /* "View.MemoryView":817
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(PyExc_IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    __pyx_t_1 = (0 <= __pyx_v_start);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);
    }
    __pyx_t_2 = (!__pyx_t_1);
    if (__pyx_t_2) {

      /* "View.MemoryView":818
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(PyExc_IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(PyExc_IndexError, __pyx_kp_s_Index_out_of_bounds_axis_d, __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 818, __pyx_L1_error)

      /* "View.MemoryView":817
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(PyExc_IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    }

    /* "View.MemoryView":813
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":821
 *     else:
 * 
 *         if have_step:             # <<<<<<<<<<<<<<
 *             negative_step = step < 0
 *             if step == 0:
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_have_step != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":822
 * 
 *         if have_step:
 *             negative_step = step < 0             # <<<<<<<<<<<<<<
 *             if step == 0:
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)
 */
      __pyx_v_negative_step = (__pyx_v_step < 0);

      /* "View.MemoryView":823
 *         if have_step:
 *             negative_step = step < 0
 *             if step == 0:             # <<<<<<<<<<<<<<
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)
 *         else:
 */
      __pyx_t_2 = (__pyx_v_step == 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":824
 *             negative_step = step < 0
 *             if step == 0:
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             negative_step = False
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(PyExc_ValueError, __pyx_kp_s_Step_may_not_be_zero_axis_d, __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 824, __pyx_L1_error)

        /* "View.MemoryView":823
 *         if have_step:
 *             negative_step = step < 0
 *             if step == 0:             # <<<<<<<<<<<<<<
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)
 *         else:
 */
      }

      /* "View.MemoryView":821
 *     else:
 * 
 *         if have_step:             # <<<<<<<<<<<<<<
 *             negative_step = step < 0
 *             if step == 0:
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":826
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)
 *         else:
 *             negative_step = False             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    /*else*/ {
      __pyx_v_negative_step = 0;

      /* "View.MemoryView":827
 *         else:
 *             negative_step = False
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_step = 1;
    }
    __pyx_L6:;

    /* "View.MemoryView":830
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
    __pyx_t_2 = (__pyx_v_have_start != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":831
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
      __pyx_t_2 = (__pyx_v_start < 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":832
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":833
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        __pyx_t_2 = (__pyx_v_start < 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":834
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */
          __pyx_v_start = 0;

          /* "View.MemoryView":833
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        }

        /* "View.MemoryView":831
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
        goto __pyx_L9;
      }

      /* "View.MemoryView":835
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      __pyx_t_2 = (__pyx_v_start >= __pyx_v_shape);
      if (__pyx_t_2) {

        /* "View.MemoryView":836
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
        if (__pyx_v_negative_step) {

          /* "View.MemoryView":837
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":836
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
          goto __pyx_L11;
        }

        /* "View.MemoryView":839
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        /*else*/ {
          __pyx_v_start = __pyx_v_shape;
        }
        __pyx_L11:;

        /* "View.MemoryView":835
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      }
      __pyx_L9:;

      /* "View.MemoryView":830
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":841
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
    /*else*/ {
      if (__pyx_v_negative_step) {

        /* "View.MemoryView":842
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":841
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
        goto __pyx_L12;
      }

      /* "View.MemoryView":844
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */
      /*else*/ {
        __pyx_v_start = 0;
      }
      __pyx_L12:;
    }
    __pyx_L8:;

    /* "View.MemoryView":846
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
    __pyx_t_2 = (__pyx_v_have_stop != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":847
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
      __pyx_t_2 = (__pyx_v_stop < 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":848
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":849
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        __pyx_t_2 = (__pyx_v_stop < 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":850
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */
          __pyx_v_stop = 0;

          /* "View.MemoryView":849
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        }

        /* "View.MemoryView":847
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
        goto __pyx_L14;
      }

      /* "View.MemoryView":851
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      __pyx_t_2 = (__pyx_v_stop > __pyx_v_shape);
      if (__pyx_t_2) {

        /* "View.MemoryView":852
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":851
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      }
      __pyx_L14:;

      /* "View.MemoryView":846
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
      goto __pyx_L13;
    }

    /* "View.MemoryView":854
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
    /*else*/ {
      if (__pyx_v_negative_step) {

        /* "View.MemoryView":855
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */
        __pyx_v_stop = -1L;

        /* "View.MemoryView":854
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
        goto __pyx_L16;
      }

      /* "View.MemoryView":857
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 * 
 */
      /*else*/ {
        __pyx_v_stop = __pyx_v_shape;
      }
      __pyx_L16:;
    }
    __pyx_L13:;

    /* "View.MemoryView":861
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":863
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":864
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":863
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    }

    /* "View.MemoryView":866
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    __pyx_t_2 = (__pyx_v_new_shape < 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":867
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":866
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    }

    /* "View.MemoryView":870
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":871
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":872
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;
  }
  __pyx_L3:;

  /* "View.MemoryView":875
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
  __pyx_t_2 = ((__pyx_v_suboffset_dim[0]) < 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":876
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":875
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
    goto __pyx_L19;
  }

  /* "View.MemoryView":878
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));
  }
  __pyx_L19:;

  /* "View.MemoryView":880
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  __pyx_t_2 = (__pyx_v_suboffset >= 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":881
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
    __pyx_t_2 = (!__pyx_v_is_slice);
    if (__pyx_t_2) {

      /* "View.MemoryView":882
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
      __pyx_t_2 = (__pyx_v_new_ndim == 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":883
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(PyExc_IndexError, "All dimensions preceding dimension %d "
 */
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":882
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
        goto __pyx_L22;
      }

      /* "View.MemoryView":885
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(PyExc_IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */
      /*else*/ {

        /* "View.MemoryView":886
 *             else:
 *                 _err_dim(PyExc_IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(PyExc_IndexError, __pyx_kp_s_All_dimensions_preceding_dimensi, __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 885, __pyx_L1_error)
      }
      __pyx_L22:;

      /* "View.MemoryView":881
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
      goto __pyx_L21;
    }

    /* "View.MemoryView":888
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    /*else*/ {
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;
    }
    __pyx_L21:;

    /* "View.MemoryView":880
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  }

  /* "View.MemoryView":890
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":793
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":896
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_suboffset;
  Py_ssize_t __pyx_v_itemsize;
  char *__pyx_v_resultp;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pybuffer_index", 1);

  /* "View.MemoryView":898
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":899
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */
  __pyx_t_1 = __pyx_v_view->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":902
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len // itemsize
 *         stride = itemsize
 */
  __pyx_t_2 = (__pyx_v_view->ndim == 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":903
 * 
 *     if view.ndim == 0:
 *         shape = view.len // itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */
    if (unlikely(__pyx_v_itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(1, 903, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(__Pyx_UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(1, 903, __pyx_L1_error)
    }
    __pyx_v_shape = __Pyx_div_Py_ssize_t(__pyx_v_view->len, __pyx_v_itemsize);

    /* "View.MemoryView":904
 *     if view.ndim == 0:
 *         shape = view.len // itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":902
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len // itemsize
 *         stride = itemsize
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":906
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */
  /*else*/ {
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":907
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":908
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    __pyx_t_2 = (__pyx_v_view->suboffsets != NULL);
    if (__pyx_t_2) {

      /* "View.MemoryView":909
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":908
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":911
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  __pyx_t_2 = (__pyx_v_index < 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":912
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 */
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":913
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 */
    __pyx_t_2 = (__pyx_v_index < 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":914
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = 0;
      __pyx_t_4 = 127;
      __Pyx_INCREF(__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
      __pyx_t_1 += 37;
      __Pyx_GIVEREF(__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Out_of_bounds_on_buffer_access_a);
      __pyx_t_5 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_dim, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u__7);
      __pyx_t_1 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__7);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__7);
      __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_builtin_IndexError, __pyx_t_5, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(1, 914, __pyx_L1_error)

      /* "View.MemoryView":913
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 */
    }

    /* "View.MemoryView":911
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  }

  /* "View.MemoryView":916
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_shape);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":917
 * 
 *     if index >= shape:
 *         raise IndexError, f"Out of bounds on buffer access (axis {dim})"             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
    __pyx_t_1 += 37;
    __Pyx_GIVEREF(__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Out_of_bounds_on_buffer_access_a);
    __pyx_t_3 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_dim, 0, ' ', 'd'); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_kp_u__7);
    __pyx_t_1 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__7);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u__7);
    __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_builtin_IndexError, __pyx_t_3, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 917, __pyx_L1_error)

    /* "View.MemoryView":916
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 */
  }

  /* "View.MemoryView":919
 *         raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":920
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  __pyx_t_2 = (__pyx_v_suboffset >= 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":921
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":920
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  }

  /* "View.MemoryView":923
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_resultp;
  goto __pyx_L0;

  /* "View.MemoryView":896
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":929
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) except -1 nogil:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "View.MemoryView":930
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) except -1 nogil:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":932
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->shape;
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":933
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->strides;
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":937
 * 
 *     cdef int i, j
 *     for i in range(ndim // 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */
  __pyx_t_3 = __Pyx_div_long(__pyx_v_ndim, 2);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":938
 *     cdef int i, j
 *     for i in range(ndim // 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":939
 *     for i in range(ndim // 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":940
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":942
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(PyExc_ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    __pyx_t_8 = ((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = ((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_7) {

      /* "View.MemoryView":943
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(PyExc_ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
      __pyx_t_9 = __pyx_memoryview_err(PyExc_ValueError, __pyx_kp_s_Cannot_transpose_memoryview_with); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 943, __pyx_L1_error)

      /* "View.MemoryView":942
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(PyExc_ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    }
  }

  /* "View.MemoryView":945
 *             _err(PyExc_ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":929
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) except -1 nogil:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":963
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XCLEAR_MEMVIEW(&self.from_slice, 1)
 * 
 */

/* Python wrapper */
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {

  /* "View.MemoryView":964
 * 
 *     def __dealloc__(self):
 *         __PYX_XCLEAR_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __PYX_XCLEAR_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":963
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XCLEAR_MEMVIEW(&self.from_slice, 1)
 * 
 */

  /* function exit code */
}

/* "View.MemoryView":966
 *         __PYX_XCLEAR_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 1);

  /* "View.MemoryView":967
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->to_object_func != NULL);
  if (__pyx_t_1) {

    /* "View.MemoryView":968
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":967
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  }

  /* "View.MemoryView":970
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 970, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":966
 *         __PYX_XCLEAR_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":972
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 1);

  /* "View.MemoryView":973
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->to_dtype_func != NULL);
  if (__pyx_t_1) {

    /* "View.MemoryView":974
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(1, 974, __pyx_L1_error)

    /* "View.MemoryView":973
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":976
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _get_base(self):
 */
  /*else*/ {
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 976, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":972
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":978
 *             memoryview.assign_item_from_object(self, itemp, value)
 * 
 *     cdef _get_base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

static PyObject *__pyx_memoryviewslice__get_base(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_base", 1);

  /* "View.MemoryView":979
 * 
 *     cdef _get_base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->from_object);
  __pyx_r = __pyx_v_self->from_object;
  goto __pyx_L0;

  /* "View.MemoryView":978
 *             memoryview.assign_item_from_object(self, itemp, value)
 * 
 *     cdef _get_base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_TypeInfo *__pyx_t_4;
  Py_buffer __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  Py_ssize_t *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_fromslice", 1);

  /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_memviewslice.memview) == Py_None);
  if (__pyx_t_1) {

    /* "View.MemoryView":1008
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "View.MemoryView":1013
 * 
 * 
 *     result = _memoryviewslice.__new__(_memoryviewslice, None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None)) __PYX_ERR(1, 1013, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0)) __PYX_ERR(1, 1013, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2)) __PYX_ERR(1, 1013, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = ((PyObject *)__pyx_tp_new__memoryviewslice(((PyTypeObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":1015
 *     result = _memoryviewslice.__new__(_memoryviewslice, None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1016
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview)._get_base()
 */
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1018
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview)._get_base()             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->__pyx_vtab)->_get_base(((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_result->from_object);
  __Pyx_DECREF(__pyx_v_result->from_object);
  __pyx_v_result->from_object = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "View.MemoryView":1019
 * 
 *     result.from_object = (<memoryview> memviewslice.memview)._get_base()
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1021
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1022
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1023
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1024
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1025
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */
  Py_INCREF(Py_None);

  /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1028
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":1030
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */
  /*else*/ {
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;
  }
  __pyx_L4:;

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1033
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1036
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1037
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    __pyx_t_1 = (__pyx_v_suboffset >= 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":1039
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1040
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */
      goto __pyx_L6_break;

      /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    }
  }
  __pyx_L6_break:;

  /* "View.MemoryView":1042
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1043
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1044
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;
  }

  /* "View.MemoryView":1046
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1047
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1049
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0;
  __Pyx_memviewslice *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slice_from_memview", 1);

  /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  if (__pyx_t_1) {

    /* "View.MemoryView":1056
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(1, 1056, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1057
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */
    __pyx_r = (&__pyx_v_obj->from_slice);
    goto __pyx_L0;

    /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  }

  /* "View.MemoryView":1059
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1060
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */
    __pyx_r = __pyx_v_mslice;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst) noexcept:             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  Py_ssize_t *__pyx_v_suboffsets;
  Py_ssize_t *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;

  /* "View.MemoryView":1067
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */
  __pyx_t_1 = __pyx_v_memview->view.shape;
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1068
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */
  __pyx_t_1 = __pyx_v_memview->view.strides;
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1069
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1071
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1072
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1074
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */
  __pyx_t_2 = __pyx_v_memview->view.ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1075
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1076
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1077
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */
    __pyx_t_6 = (__pyx_v_suboffsets != 0);
    if (__pyx_t_6) {
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);
    } else {
      __pyx_t_5 = -1L;
    }
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;
  }

  /* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst) noexcept:             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

  /* function exit code */
}

/* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy", 1);

  /* "View.MemoryView":1083
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1084
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *);
  int (*__pyx_v_to_dtype_func)(char *, PyObject *);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *(*__pyx_t_2)(char *);
  int (*__pyx_t_3)(char *, PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 1);

  /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  if (__pyx_t_1) {

    /* "View.MemoryView":1095
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */
    __pyx_t_2 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;
    __pyx_v_to_object_func = __pyx_t_2;

    /* "View.MemoryView":1096
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;
    __pyx_v_to_dtype_func = __pyx_t_3;

    /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1098
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */
  /*else*/ {
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1099
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */
    __pyx_v_to_dtype_func = NULL;
  }
  __pyx_L3:;

  /* "View.MemoryView":1101
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1103
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) noexcept nogil:             # <<<<<<<<<<<<<<
 *     return -arg if arg < 0 else arg
 * 
 */

static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;

  /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) noexcept nogil:
 *     return -arg if arg < 0 else arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */
  __pyx_t_2 = (__pyx_v_arg < 0);
  if (__pyx_t_2) {
    __pyx_t_1 = (-__pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) noexcept nogil:             # <<<<<<<<<<<<<<
 *     return -arg if arg < 0 else arg
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1113
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_c_stride;
  Py_ssize_t __pyx_v_f_stride;
  char __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1118
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1119
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1121
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1122
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = ((__pyx_v_mslice->shape[__pyx_v_i]) > 1);
    if (__pyx_t_2) {

      /* "View.MemoryView":1123
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1124
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
      goto __pyx_L4_break;

      /* "View.MemoryView":1122
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L4_break:;

  /* "View.MemoryView":1126
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */
  __pyx_t_1 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1127
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = ((__pyx_v_mslice->shape[__pyx_v_i]) > 1);
    if (__pyx_t_2) {

      /* "View.MemoryView":1128
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1129
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */
      goto __pyx_L7_break;

      /* "View.MemoryView":1127
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L7_break:;

  /* "View.MemoryView":1131
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  __pyx_t_2 = (abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride));
  if (__pyx_t_2) {

    /* "View.MemoryView":1132
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */
    __pyx_r = 'C';
    goto __pyx_L0;

    /* "View.MemoryView":1131
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  }

  /* "View.MemoryView":1134
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  /*else*/ {
    __pyx_r = 'F';
    goto __pyx_L0;
  }

  /* "View.MemoryView":1113
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1137
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent;
  Py_ssize_t __pyx_v_dst_extent;
  Py_ssize_t __pyx_v_src_stride;
  Py_ssize_t __pyx_v_dst_stride;
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;

  /* "View.MemoryView":1144
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1145
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1146
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1147
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1149
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         if (src_stride > 0 and dst_stride > 0 and
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
  __pyx_t_1 = (__pyx_v_ndim == 1);
  if (__pyx_t_1) {

    /* "View.MemoryView":1150
 * 
 *     if ndim == 1:
 *         if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):
 *             memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    __pyx_t_2 = (__pyx_v_src_stride > 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_dst_stride > 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }

    /* "View.MemoryView":1151
 *     if ndim == 1:
 *         if (src_stride > 0 and dst_stride > 0 and
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *             memcpy(dst_data, src_data, itemsize * dst_extent)
 *         else:
 */
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));
    }
    __pyx_t_1 = __pyx_t_2;
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1150
 * 
 *     if ndim == 1:
 *         if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):
 *             memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    if (__pyx_t_1) {

      /* "View.MemoryView":1152
 *         if (src_stride > 0 and dst_stride > 0 and
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):
 *             memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *         else:
 *             for i in range(dst_extent):
 */
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1150
 * 
 *     if ndim == 1:
 *         if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):
 *             memcpy(dst_data, src_data, itemsize * dst_extent)
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":1154
 *             memcpy(dst_data, src_data, itemsize * dst_extent)
 *         else:
 *             for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                 memcpy(dst_data, src_data, itemsize)
 *                 src_data += src_stride
 */
    /*else*/ {
      __pyx_t_3 = __pyx_v_dst_extent;
      __pyx_t_4 = __pyx_t_3;
      for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
        __pyx_v_i = __pyx_t_5;

        /* "View.MemoryView":1155
 *         else:
 *             for i in range(dst_extent):
 *                 memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                 src_data += src_stride
 *                 dst_data += dst_stride
 */
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1156
 *             for i in range(dst_extent):
 *                 memcpy(dst_data, src_data, itemsize)
 *                 src_data += src_stride             # <<<<<<<<<<<<<<
 *                 dst_data += dst_stride
 *     else:
 */
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1157
 *                 memcpy(dst_data, src_data, itemsize)
 *                 src_data += src_stride
 *                 dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
      }
    }
    __pyx_L4:;

    /* "View.MemoryView":1149
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         if (src_stride > 0 and dst_stride > 0 and
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1159
 *                 dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */
  /*else*/ {
    __pyx_t_3 = __pyx_v_dst_extent;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":1160
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1164
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1165
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1137
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

  /* function exit code */
}

/* "View.MemoryView":1167
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) noexcept nogil:
 */

static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1170
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) noexcept nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1167
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) noexcept nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1174
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) noexcept nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;

  /* "View.MemoryView":1176
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) noexcept nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1178
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1179
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);
  }

  /* "View.MemoryView":1181
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "View.MemoryView":1174
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) noexcept nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1184
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) noexcept nogil:
 */

static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1193
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
  __pyx_t_1 = (__pyx_v_order == 'F');
  if (__pyx_t_1) {

    /* "View.MemoryView":1194
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
    __pyx_t_2 = __pyx_v_ndim;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1195
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1196
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }

    /* "View.MemoryView":1193
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1198
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
  /*else*/ {
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1199
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1200
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1202
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */
  __pyx_r = __pyx_v_stride;
  goto __pyx_L0;

  /* "View.MemoryView":1184
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) noexcept nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1205
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i;
  void *__pyx_v_result;
  size_t __pyx_v_itemsize;
  size_t __pyx_v_size;
  void *__pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_memoryview_obj *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "View.MemoryView":1216
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1217
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1219
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err_no_memory()
 */
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1220
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err_no_memory()
 * 
 */
  __pyx_t_2 = (!(__pyx_v_result != 0));
  if (__pyx_t_2) {

    /* "View.MemoryView":1221
 *     result = malloc(size)
 *     if not result:
 *         _err_no_memory()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_err_no_memory(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 1221, __pyx_L1_error)

    /* "View.MemoryView":1220
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err_no_memory()
 * 
 */
  }

  /* "View.MemoryView":1224
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1225
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */
  __pyx_t_4 = __pyx_v_src->memview;
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1226
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1227
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1228
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize, ndim, order)
 */
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1230
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize, ndim, order)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1233
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1234
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    __pyx_t_2 = ((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1);
    if (__pyx_t_2) {

      /* "View.MemoryView":1235
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1234
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    }
  }

  /* "View.MemoryView":1237
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
  __pyx_t_2 = __pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim);
  if (__pyx_t_2) {

    /* "View.MemoryView":1238
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1237
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":1240
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  /*else*/ {
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);
  }
  __pyx_L9:;

  /* "View.MemoryView":1242
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":1205
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1247
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError, f"got differing extents in dimension {i} (got {extent1} and {extent2})"
 */

static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1249
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError, f"got differing extents in dimension {i} (got {extent1} and {extent2})"             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_got_differing_extents_in_dimensi);
  __pyx_t_2 += 35;
  __Pyx_GIVEREF(__pyx_kp_u_got_differing_extents_in_dimensi);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_got_differing_extents_in_dimensi);
  __pyx_t_4 = __Pyx_PyUnicode_From_int(__pyx_v_i, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_got);
  __pyx_t_2 += 6;
  __Pyx_GIVEREF(__pyx_kp_u_got);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_got);
  __pyx_t_4 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_extent1, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_and);
  __pyx_t_2 += 5;
  __Pyx_GIVEREF(__pyx_kp_u_and);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_and);
  __pyx_t_4 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_extent2, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__7);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__7);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__7);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_builtin_ValueError, __pyx_t_4, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(1, 1249, __pyx_L1_error)

  /* "View.MemoryView":1247
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError, f"got differing extents in dimension {i} (got {extent1} and {extent2})"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1252
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(PyObject *error, str msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise <object>error, msg % dim
 * 
 */

static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, PyObject *__pyx_v_msg, int __pyx_v_dim) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_dim", 0);
  __Pyx_INCREF(__pyx_v_msg);

  /* "View.MemoryView":1253
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(PyObject *error, str msg, int dim) except -1 with gil:
 *     raise <object>error, msg % dim             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_v_msg, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(((PyObject *)__pyx_v_error), __pyx_t_2, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(1, 1253, __pyx_L1_error)

  /* "View.MemoryView":1252
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(PyObject *error, str msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise <object>error, msg % dim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1256
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(PyObject *error, str msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise <object>error, msg
 * 
 */

static int __pyx_memoryview_err(PyObject *__pyx_v_error, PyObject *__pyx_v_msg) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err", 0);
  __Pyx_INCREF(__pyx_v_msg);

  /* "View.MemoryView":1257
 * @cname('__pyx_memoryview_err')
 * cdef int _err(PyObject *error, str msg) except -1 with gil:
 *     raise <object>error, msg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_no_memory')
 */
  __Pyx_Raise(((PyObject *)__pyx_v_error), __pyx_v_msg, 0, 0);
  __PYX_ERR(1, 1257, __pyx_L1_error)

  /* "View.MemoryView":1256
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(PyObject *error, str msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise <object>error, msg
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1260
 * 
 * @cname('__pyx_memoryview_err_no_memory')
 * cdef int _err_no_memory() except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise MemoryError
 * 
 */

static int __pyx_memoryview_err_no_memory(void) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif

  /* "View.MemoryView":1261
 * @cname('__pyx_memoryview_err_no_memory')
 * cdef int _err_no_memory() except -1 with gil:
 *     raise MemoryError             # <<<<<<<<<<<<<<
 * 
 * 
 */
  PyErr_NoMemory(); __PYX_ERR(1, 1261, __pyx_L1_error)

  /* "View.MemoryView":1260
 * 
 * @cname('__pyx_memoryview_err_no_memory')
 * cdef int _err_no_memory() except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise MemoryError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._err_no_memory", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1265
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata;
  size_t __pyx_v_itemsize;
  int __pyx_v_i;
  char __pyx_v_order;
  int __pyx_v_broadcasting;
  int __pyx_v_direct_copy;
  __Pyx_memviewslice __pyx_v_tmp;
  int __pyx_v_ndim;
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "View.MemoryView":1273
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1274
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1276
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1277
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1278
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1281
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
  __pyx_t_2 = (__pyx_v_src_ndim < __pyx_v_dst_ndim);
  if (__pyx_t_2) {

    /* "View.MemoryView":1282
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1281
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1283
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  __pyx_t_2 = (__pyx_v_dst_ndim < __pyx_v_src_ndim);
  if (__pyx_t_2) {

    /* "View.MemoryView":1284
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1283
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":1286
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_3 = __pyx_v_dst_ndim;
  __pyx_t_4 = __pyx_v_src_ndim;
  __pyx_t_2 = (__pyx_t_3 > __pyx_t_4);
  if (__pyx_t_2) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1288
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */
  __pyx_t_5 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1289
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    __pyx_t_2 = ((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i]));
    if (__pyx_t_2) {

      /* "View.MemoryView":1290
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
      __pyx_t_2 = ((__pyx_v_src.shape[__pyx_v_i]) == 1);
      if (__pyx_t_2) {

        /* "View.MemoryView":1291
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1292
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1290
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":1294
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */
      /*else*/ {
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1294, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":1289
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    }

    /* "View.MemoryView":1296
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)
 * 
 */
    __pyx_t_2 = ((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1297
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */
      __pyx_t_6 = __pyx_memoryview_err_dim(PyExc_ValueError, __pyx_kp_s_Dimension_d_is_not_direct, __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1297, __pyx_L1_error)

      /* "View.MemoryView":1296
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)
 * 
 */
    }
  }

  /* "View.MemoryView":1299
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  __pyx_t_2 = __pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);
  if (__pyx_t_2) {

    /* "View.MemoryView":1301
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    __pyx_t_2 = (!__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim));
    if (__pyx_t_2) {

      /* "View.MemoryView":1302
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1301
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    }

    /* "View.MemoryView":1304
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(1, 1304, __pyx_L1_error)
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1305
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1299
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  }

  /* "View.MemoryView":1307
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (!__pyx_v_broadcasting);
  if (__pyx_t_2) {

    /* "View.MemoryView":1310
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
    __pyx_t_2 = __pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim);
    if (__pyx_t_2) {

      /* "View.MemoryView":1311
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1310
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
      goto __pyx_L12;
    }

    /* "View.MemoryView":1312
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    __pyx_t_2 = __pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim);
    if (__pyx_t_2) {

      /* "View.MemoryView":1313
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1312
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    }
    __pyx_L12:;

    /* "View.MemoryView":1315
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 */
    if (__pyx_v_direct_copy) {

      /* "View.MemoryView":1317
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1318
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 *             free(tmpdata)
 */
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1319
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1320
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1321
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "View.MemoryView":1315
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 */
    }

    /* "View.MemoryView":1307
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1323
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (__pyx_v_order == 'F');
  if (__pyx_t_2) {
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));
  }
  if (__pyx_t_2) {

    /* "View.MemoryView":1326
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 1326, __pyx_L1_error)

    /* "View.MemoryView":1327
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 1327, __pyx_L1_error)

    /* "View.MemoryView":1323
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1329
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1330
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 * 
 */
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1331
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1333
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1334
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1265
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1337
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) noexcept nogil:
 */

static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":1341
 *                             int ndim_other) noexcept nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1343
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1344
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1345
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1346
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);
  }

  /* "View.MemoryView":1348
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */
  __pyx_t_1 = __pyx_v_offset;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1349
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1350
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1351
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1337
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) noexcept nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1359
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object, int ndim, bint inc) noexcept nogil:             # <<<<<<<<<<<<<<
 * 
 *     if dtype_is_object:
 */

static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {

  /* "View.MemoryView":1361
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object, int ndim, bint inc) noexcept nogil:
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape, dst.strides, ndim, inc)
 * 
 */
  if (__pyx_v_dtype_is_object) {

    /* "View.MemoryView":1362
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape, dst.strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 */
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1361
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object, int ndim, bint inc) noexcept nogil:
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape, dst.strides, ndim, inc)
 * 
 */
  }

  /* "View.MemoryView":1359
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object, int ndim, bint inc) noexcept nogil:             # <<<<<<<<<<<<<<
 * 
 *     if dtype_is_object:
 */

  /* function exit code */
}

/* "View.MemoryView":1365
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) noexcept with gil:
 */

static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif

  /* "View.MemoryView":1368
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) noexcept with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1365
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) noexcept with gil:
 */

  /* function exit code */
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc) noexcept:
 *     cdef Py_ssize_t i
 */

static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1374
 *                                     Py_ssize_t *strides, int ndim, bint inc) noexcept:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(shape[0]):
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1376
 *     cdef Py_ssize_t stride = strides[0]
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */
  __pyx_t_1 = (__pyx_v_shape[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1377
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
    __pyx_t_4 = (__pyx_v_ndim == 1);
    if (__pyx_t_4) {

      /* "View.MemoryView":1378
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
      if (__pyx_v_inc) {

        /* "View.MemoryView":1379
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1378
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
        goto __pyx_L6;
      }

      /* "View.MemoryView":1381
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1, ndim - 1, inc)
 */
      /*else*/ {
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));
      }
      __pyx_L6:;

      /* "View.MemoryView":1377
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":1383
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1, ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += stride
 */
    /*else*/ {
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);
    }
    __pyx_L5:;

    /* "View.MemoryView":1385
 *             refcount_objects_in_slice(data, shape + 1, strides + 1, ndim - 1, inc)
 * 
 *         data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
  }

  /* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc) noexcept:
 *     cdef Py_ssize_t i
 */

  /* function exit code */
}

/* "View.MemoryView":1391
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) noexcept nogil:
 */

static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1394
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) noexcept nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, inc=False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim, itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, inc=True)
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1395
 *                               bint dtype_is_object) noexcept nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, inc=False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim, itemsize, item)             # <<<<<<<<<<<<<<
 *     refcount_copying(dst, dtype_is_object, ndim, inc=True)
 * 
 */
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1396
 *     refcount_copying(dst, dtype_is_object, ndim, inc=False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim, itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, inc=True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1391
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) noexcept nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1400
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) noexcept nogil:
 */

static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_extent;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1404
 *                               size_t itemsize, void *item) noexcept nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1405
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1407
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
  __pyx_t_1 = (__pyx_v_ndim == 1);
  if (__pyx_t_1) {

    /* "View.MemoryView":1408
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1409
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1410
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }

    /* "View.MemoryView":1407
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1412
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1, ndim - 1, itemsize, item)
 *             data += stride
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1413
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1, ndim - 1, itemsize, item)             # <<<<<<<<<<<<<<
 *             data += stride
 * 
 */
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1414
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1, ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1400
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) noexcept nogil:
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = {"__pyx_unpickle_Enum", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__8, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = Enum.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self):             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":20
 *         @property
 *         cdef inline double real(self):
 *             return self.cval.real             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = __pyx_v_self->cval.real;
  goto __pyx_L0;

  /* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self):             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self):             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":24
 *         @property
 *         cdef inline double imag(self):
 *             return self.cval.imag             # <<<<<<<<<<<<<<
 * 
 *     # PyTypeObject PyComplex_Type
 */
  __pyx_r = __pyx_v_self->cval.imag;
  goto __pyx_L0;

  /* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self):             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":117
 *     or None if no such value or default was found.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":118
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)             # <<<<<<<<<<<<<<
 *     if value is NULL:
 *         # context variable does not have a default
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, NULL, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(14, 118, __pyx_L1_error)

  /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
  __pyx_t_2 = (__pyx_v_value == NULL);
  if (__pyx_t_2) {

    /* "cpython/contextvars.pxd":121
 *     if value is NULL:
 *         # context variable does not have a default
 *         pyvalue = default_value             # <<<<<<<<<<<<<<
 *     else:
 *         # value or default value of context variable
 */
    __Pyx_INCREF(__pyx_v_default_value);
    __pyx_v_pyvalue = __pyx_v_default_value;

    /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
    goto __pyx_L3;
  }

  /* "cpython/contextvars.pxd":124
 *     else:
 *         # value or default value of context variable
 *         pyvalue = <object>value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_pyvalue = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cpython/contextvars.pxd":125
 *         # value or default value of context variable
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 * 
 */
    Py_XDECREF(__pyx_v_value);
  }
  __pyx_L3:;

  /* "cpython/contextvars.pxd":126
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.contextvars.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value_no_default(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value_no_default", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":135
 *     Ignores the default value of the context variable, if any.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":136
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)             # <<<<<<<<<<<<<<
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, ((PyObject *)__pyx_v_default_value), (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(14, 136, __pyx_L1_error)

  /* "cpython/contextvars.pxd":138
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value             # <<<<<<<<<<<<<<
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_value);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_pyvalue = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cpython/contextvars.pxd":139
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 */
  Py_XDECREF(__pyx_v_value);

  /* "cpython/contextvars.pxd":140
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.contextvars.get_value_no_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":104
 *             __data_union data
 * 
 *         def __getbuffer__(self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
CYTHON_UNUSED static int __pyx_pw_7cpython_5array_5array_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
CYTHON_UNUSED static int __pyx_pw_7cpython_5array_5array_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7cpython_5array_5array___getbuffer__(((arrayobject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7cpython_5array_5array___getbuffer__(arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info, CYTHON_UNUSED int __pyx_v_flags) {
  PyObject *__pyx_v_item_count = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  char __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (unlikely(__pyx_v_info == NULL)) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "array.pxd":109
 *             # In particular strided access is always provided regardless
 *             # of flags
 *             item_count = Py_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *             info.suboffsets = NULL
 */
  __pyx_t_1 = PyInt_FromSsize_t(Py_SIZE(((PyObject *)__pyx_v_self))); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_item_count = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "array.pxd":111
 *             item_count = Py_SIZE(self)
 * 
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.buf = self.data.as_chars
 *             info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "array.pxd":112
 * 
 *             info.suboffsets = NULL
 *             info.buf = self.data.as_chars             # <<<<<<<<<<<<<<
 *             info.readonly = 0
 *             info.ndim = 1
 */
  __pyx_t_2 = __pyx_v_self->data.as_chars;
  __pyx_v_info->buf = __pyx_t_2;

  /* "array.pxd":113
 *             info.suboffsets = NULL
 *             info.buf = self.data.as_chars
 *             info.readonly = 0             # <<<<<<<<<<<<<<
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 */
  __pyx_v_info->readonly = 0;

  /* "array.pxd":114
 *             info.buf = self.data.as_chars
 *             info.readonly = 0
 *             info.ndim = 1             # <<<<<<<<<<<<<<
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 *             info.len = info.itemsize * item_count
 */
  __pyx_v_info->ndim = 1;

  /* "array.pxd":115
 *             info.readonly = 0
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)             # <<<<<<<<<<<<<<
 *             info.len = info.itemsize * item_count
 * 
 */
  __pyx_t_3 = __pyx_v_self->ob_descr->itemsize;
  __pyx_v_info->itemsize = __pyx_t_3;

  /* "array.pxd":116
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 *             info.len = info.itemsize * item_count             # <<<<<<<<<<<<<<
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 */
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_info->itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_1, __pyx_v_item_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(15, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(15, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_info->len = __pyx_t_5;

  /* "array.pxd":118
 *             info.len = info.itemsize * item_count
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)             # <<<<<<<<<<<<<<
 *             if not info.shape:
 *                 raise MemoryError()
 */
  __pyx_v_info->shape = ((Py_ssize_t *)PyObject_Malloc(((sizeof(Py_ssize_t)) + 2)));

  /* "array.pxd":119
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 */
  __pyx_t_6 = (!(__pyx_v_info->shape != 0));
  if (unlikely(__pyx_t_6)) {

    /* "array.pxd":120
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 *             info.shape[0] = item_count      # constant regardless of resizing
 *             info.strides = &info.itemsize
 */
    PyErr_NoMemory(); __PYX_ERR(15, 120, __pyx_L1_error)

    /* "array.pxd":119
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 */
  }

  /* "array.pxd":121
 *             if not info.shape:
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing             # <<<<<<<<<<<<<<
 *             info.strides = &info.itemsize
 * 
 */
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_v_item_count); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(15, 121, __pyx_L1_error)
  (__pyx_v_info->shape[0]) = __pyx_t_5;

  /* "array.pxd":122
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 *             info.strides = &info.itemsize             # <<<<<<<<<<<<<<
 * 
 *             info.format = <char*> (info.shape + 1)
 */
  __pyx_v_info->strides = (&__pyx_v_info->itemsize);

  /* "array.pxd":124
 *             info.strides = &info.itemsize
 * 
 *             info.format = <char*> (info.shape + 1)             # <<<<<<<<<<<<<<
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0
 */
  __pyx_v_info->format = ((char *)(__pyx_v_info->shape + 1));

  /* "array.pxd":125
 * 
 *             info.format = <char*> (info.shape + 1)
 *             info.format[0] = self.ob_descr.typecode             # <<<<<<<<<<<<<<
 *             info.format[1] = 0
 *             info.obj = self
 */
  __pyx_t_7 = __pyx_v_self->ob_descr->typecode;
  (__pyx_v_info->format[0]) = __pyx_t_7;

  /* "array.pxd":126
 *             info.format = <char*> (info.shape + 1)
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0             # <<<<<<<<<<<<<<
 *             info.obj = self
 * 
 */
  (__pyx_v_info->format[1]) = 0;

  /* "array.pxd":127
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(self, Py_buffer* info):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "array.pxd":104
 *             __data_union data
 * 
 *         def __getbuffer__(self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cpython.array.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_item_count);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":129
 *             info.obj = self
 * 
 *         def __releasebuffer__(self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             PyObject_Free(info.shape)
 * 
 */

/* Python wrapper */
CYTHON_UNUSED static void __pyx_pw_7cpython_5array_5array_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
CYTHON_UNUSED static void __pyx_pw_7cpython_5array_5array_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_7cpython_5array_5array_2__releasebuffer__(((arrayobject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7cpython_5array_5array_2__releasebuffer__(CYTHON_UNUSED arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info) {

  /* "array.pxd":130
 * 
 *         def __releasebuffer__(self, Py_buffer* info):
 *             PyObject_Free(info.shape)             # <<<<<<<<<<<<<<
 * 
 *     array newarrayobject(PyTypeObject* type, Py_ssize_t size, arraydescr *descr)
 */
  PyObject_Free(__pyx_v_info->shape);

  /* "array.pxd":129
 *             info.obj = self
 * 
 *         def __releasebuffer__(self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             PyObject_Free(info.shape)
 * 
 */

  /* function exit code */
}

/* "array.pxd":141
 * 
 * 
 * cdef inline array clone(array template, Py_ssize_t length, bint zero):             # <<<<<<<<<<<<<<
 *     """ fast creation of a new array, given a template array.
 *     type will be same as template.
 */

static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_clone(arrayobject *__pyx_v_template, Py_ssize_t __pyx_v_length, int __pyx_v_zero) {
  arrayobject *__pyx_v_op = 0;
  arrayobject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 1);

  /* "array.pxd":145
 *     type will be same as template.
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)             # <<<<<<<<<<<<<<
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 */
  __pyx_t_1 = ((PyObject *)newarrayobject(Py_TYPE(((PyObject *)__pyx_v_template)), __pyx_v_length, __pyx_v_template->ob_descr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "array.pxd":146
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:             # <<<<<<<<<<<<<<
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op
 */
  if (__pyx_v_zero) {
  } else {
    __pyx_t_2 = __pyx_v_zero;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (((PyObject *)__pyx_v_op) != Py_None);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "array.pxd":147
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 *     return op
 * 
 */
    (void)(memset(__pyx_v_op->data.as_chars, 0, (__pyx_v_length * __pyx_v_op->ob_descr->itemsize)));

    /* "array.pxd":146
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:             # <<<<<<<<<<<<<<
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op
 */
  }

  /* "array.pxd":148
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op             # <<<<<<<<<<<<<<
 * 
 * cdef inline array copy(array self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_op);
  __pyx_r = __pyx_v_op;
  goto __pyx_L0;

  /* "array.pxd":141
 * 
 * 
 * cdef inline array clone(array template, Py_ssize_t length, bint zero):             # <<<<<<<<<<<<<<
 *     """ fast creation of a new array, given a template array.
 *     type will be same as template.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.array.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":150
 *     return op
 * 
 * cdef inline array copy(array self):             # <<<<<<<<<<<<<<
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 */

static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_copy(arrayobject *__pyx_v_self) {
  arrayobject *__pyx_v_op = 0;
  arrayobject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "array.pxd":152
 * cdef inline array copy(array self):
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)             # <<<<<<<<<<<<<<
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)
 *     return op
 */
  __pyx_t_1 = ((PyObject *)newarrayobject(Py_TYPE(((PyObject *)__pyx_v_self)), Py_SIZE(((PyObject *)__pyx_v_self)), __pyx_v_self->ob_descr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "array.pxd":153
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 *     return op
 * 
 */
  (void)(memcpy(__pyx_v_op->data.as_chars, __pyx_v_self->data.as_chars, (Py_SIZE(((PyObject *)__pyx_v_op)) * __pyx_v_op->ob_descr->itemsize)));

  /* "array.pxd":154
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)
 *     return op             # <<<<<<<<<<<<<<
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_op);
  __pyx_r = __pyx_v_op;
  goto __pyx_L0;

  /* "array.pxd":150
 *     return op
 * 
 * cdef inline array copy(array self):             # <<<<<<<<<<<<<<
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.array.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "array.pxd":156
 *     return op
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:             # <<<<<<<<<<<<<<
 *     """ efficient appending of new stuff of same type
 *     (e.g. of same array type)
 */

static CYTHON_INLINE int __pyx_f_7cpython_5array_extend_buffer(arrayobject *__pyx_v_self, char *__pyx_v_stuff, Py_ssize_t __pyx_v_n) {
  Py_ssize_t __pyx_v_itemsize;
  Py_ssize_t __pyx_v_origsize;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "array.pxd":160
 *     (e.g. of same array type)
 *     n: number of elements (not number of bytes!) """
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)
 */
  __pyx_t_1 = __pyx_v_self->ob_descr->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "array.pxd":161
 *     n: number of elements (not number of bytes!) """
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize
 *     cdef Py_ssize_t origsize = Py_SIZE(self)             # <<<<<<<<<<<<<<
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 */
  __pyx_v_origsize = Py_SIZE(((PyObject *)__pyx_v_self));

  /* "array.pxd":162
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)             # <<<<<<<<<<<<<<
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 *     return 0
 */
  __pyx_t_1 = resize_smart(__pyx_v_self, (__pyx_v_origsize + __pyx_v_n)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(15, 162, __pyx_L1_error)

  /* "array.pxd":163
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  (void)(memcpy((__pyx_v_self->data.as_chars + (__pyx_v_origsize * __pyx_v_itemsize)), __pyx_v_stuff, (__pyx_v_n * __pyx_v_itemsize)));

  /* "array.pxd":164
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef inline int extend(array self, array other) except -1:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "array.pxd":156
 *     return op
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:             # <<<<<<<<<<<<<<
 *     """ efficient appending of new stuff of same type
 *     (e.g. of same array type)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cpython.array.extend_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "array.pxd":166
 *     return 0
 * 
 * cdef inline int extend(array self, array other) except -1:             # <<<<<<<<<<<<<<
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 */

static CYTHON_INLINE int __pyx_f_7cpython_5array_extend(arrayobject *__pyx_v_self, arrayobject *__pyx_v_other) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "array.pxd":168
 * cdef inline int extend(array self, array other) except -1:
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:             # <<<<<<<<<<<<<<
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 */
  __pyx_t_1 = (__pyx_v_self->ob_descr->typecode != __pyx_v_other->ob_descr->typecode);
  if (__pyx_t_1) {

    /* "array.pxd":169
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 *         PyErr_BadArgument()             # <<<<<<<<<<<<<<
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 */
    __pyx_t_2 = PyErr_BadArgument(); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(15, 169, __pyx_L1_error)

    /* "array.pxd":168
 * cdef inline int extend(array self, array other) except -1:
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:             # <<<<<<<<<<<<<<
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 */
  }

  /* "array.pxd":170
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))             # <<<<<<<<<<<<<<
 * 
 * cdef inline void zero(array self):
 */
  __pyx_t_2 = __pyx_f_7cpython_5array_extend_buffer(__pyx_v_self, __pyx_v_other->data.as_chars, Py_SIZE(((PyObject *)__pyx_v_other))); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(15, 170, __pyx_L1_error)
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "array.pxd":166
 *     return 0
 * 
 * cdef inline int extend(array self, array other) except -1:             # <<<<<<<<<<<<<<
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cpython.array.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "array.pxd":172
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

static CYTHON_INLINE void __pyx_f_7cpython_5array_zero(arrayobject *__pyx_v_self) {

  /* "array.pxd":174
 * cdef inline void zero(array self):
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 */
  (void)(memset(__pyx_v_self->data.as_chars, 0, (Py_SIZE(((PyObject *)__pyx_v_self)) * __pyx_v_self->ob_descr->itemsize)));

  /* "array.pxd":172
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

  /* function exit code */
}

/* "cavro.pyx":48
 * 
 * __tests = {}
 * def _tests(fn):             # <<<<<<<<<<<<<<
 *     def add(test, *args):
 *         name = test.__name__
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_1_tests(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_1_tests = {"_tests", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_1_tests, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_1_tests(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fn = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_tests (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fn,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fn)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(13, 48, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_tests") < 0)) __PYX_ERR(13, 48, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_fn = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_tests", 1, 1, 1, __pyx_nargs); __PYX_ERR(13, 48, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._tests", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro__tests(__pyx_self, __pyx_v_fn);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cavro.pyx":49
 * __tests = {}
 * def _tests(fn):
 *     def add(test, *args):             # <<<<<<<<<<<<<<
 *         name = test.__name__
 *         if args:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6_tests_1add(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_6_tests_1add = {"add", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6_tests_1add, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_6_tests_1add(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_test = 0;
  PyObject *__pyx_v_args = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 1, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_test,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_test)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(13, 49, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 1) ? kwd_pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, used_pos_args, "add") < 0)) __PYX_ERR(13, 49, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_test = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add", 0, 1, 1, __pyx_nargs); __PYX_ERR(13, 49, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("cavro._tests.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6_tests_add(__pyx_self, __pyx_v_test, __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_6_tests_3add_2generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "cavro.pyx":52
 *         name = test.__name__
 *         if args:
 *             name += '-' + ('-'.join(str(a) for a in args))             # <<<<<<<<<<<<<<
 *         __tests[name] = (test, ) + args
 *     fn(add)
 */

static PyObject *__pyx_pf_5cavro_6_tests_3add_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct__genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct__genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct__genexpr(__pyx_ptype_5cavro___pyx_scope_struct__genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct__genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(13, 52, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_6_tests_3add_2generator5, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_tests_locals_add_locals_genexpr, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(13, 52, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro._tests.add.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_6_tests_3add_2generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct__genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct__genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(13, 52, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(13, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(13, 52, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(13, 52, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(13, 52, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_a);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_a, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Str(__pyx_cur_scope->__pyx_v_a); if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_3))) __PYX_ERR(13, 52, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cavro.pyx":49
 * __tests = {}
 * def _tests(fn):
 *     def add(test, *args):             # <<<<<<<<<<<<<<
 *         name = test.__name__
 *         if args:
 */

static PyObject *__pyx_pf_5cavro_6_tests_add(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_test, PyObject *__pyx_v_args) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_gb_5cavro_6_tests_3add_2generator5 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add", 1);

  /* "cavro.pyx":50
 * def _tests(fn):
 *     def add(test, *args):
 *         name = test.__name__             # <<<<<<<<<<<<<<
 *         if args:
 *             name += '-' + ('-'.join(str(a) for a in args))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_test, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cavro.pyx":51
 *     def add(test, *args):
 *         name = test.__name__
 *         if args:             # <<<<<<<<<<<<<<
 *             name += '-' + ('-'.join(str(a) for a in args))
 *         __tests[name] = (test, ) + args
 */
  __pyx_t_2 = (PyTuple_GET_SIZE(__pyx_v_args) != 0);
  if (__pyx_t_2) {

    /* "cavro.pyx":52
 *         name = test.__name__
 *         if args:
 *             name += '-' + ('-'.join(str(a) for a in args))             # <<<<<<<<<<<<<<
 *         __tests[name] = (test, ) + args
 *     fn(add)
 */
    __pyx_t_1 = __pyx_pf_5cavro_6_tests_3add_genexpr(NULL, __pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__9, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u__9, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_name, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_name, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "cavro.pyx":51
 *     def add(test, *args):
 *         name = test.__name__
 *         if args:             # <<<<<<<<<<<<<<
 *             name += '-' + ('-'.join(str(a) for a in args))
 *         __tests[name] = (test, ) + args
 */
  }

  /* "cavro.pyx":53
 *         if args:
 *             name += '-' + ('-'.join(str(a) for a in args))
 *         __tests[name] = (test, ) + args             # <<<<<<<<<<<<<<
 *     fn(add)
 *     return _tests
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_test);
  __Pyx_GIVEREF(__pyx_v_test);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_test)) __PYX_ERR(13, 53, __pyx_L1_error);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_tests); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely((PyObject_SetItem(__pyx_t_1, __pyx_v_name, __pyx_t_3) < 0))) __PYX_ERR(13, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cavro.pyx":49
 * __tests = {}
 * def _tests(fn):
 *     def add(test, *args):             # <<<<<<<<<<<<<<
 *         name = test.__name__
 *         if args:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro._tests.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_gb_5cavro_6_tests_3add_2generator5);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cavro.pyx":48
 * 
 * __tests = {}
 * def _tests(fn):             # <<<<<<<<<<<<<<
 *     def add(test, *args):
 *         name = test.__name__
 */

static PyObject *__pyx_pf_5cavro__tests(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_fn) {
  PyObject *__pyx_v_add = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_tests", 1);

  /* "cavro.pyx":49
 * __tests = {}
 * def _tests(fn):
 *     def add(test, *args):             # <<<<<<<<<<<<<<
 *         name = test.__name__
 *         if args:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6_tests_1add, 0, __pyx_n_s_tests_locals_add, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__11)); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_add = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cavro.pyx":54
 *             name += '-' + ('-'.join(str(a) for a in args))
 *         __tests[name] = (test, ) + args
 *     fn(add)             # <<<<<<<<<<<<<<
 *     return _tests
 * 
 */
  __Pyx_INCREF(__pyx_v_fn);
  __pyx_t_2 = __pyx_v_fn; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_add};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cavro.pyx":55
 *         __tests[name] = (test, ) + args
 *     fn(add)
 *     return _tests             # <<<<<<<<<<<<<<
 * 
 * __perf_tests = {}
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_tests_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cavro.pyx":48
 * 
 * __tests = {}
 * def _tests(fn):             # <<<<<<<<<<<<<<
 *     def add(test, *args):
 *         name = test.__name__
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro._tests", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_add);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cavro.pyx":58
 * 
 * __perf_tests = {}
 * def _perf(fn):             # <<<<<<<<<<<<<<
 *     def add(test, *args):
 *         name = test.__name__
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_3_perf(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_3_perf = {"_perf", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_3_perf, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_3_perf(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fn = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_perf (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fn,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fn)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(13, 58, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_perf") < 0)) __PYX_ERR(13, 58, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_fn = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_perf", 1, 1, 1, __pyx_nargs); __PYX_ERR(13, 58, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._perf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_2_perf(__pyx_self, __pyx_v_fn);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cavro.pyx":59
 * __perf_tests = {}
 * def _perf(fn):
 *     def add(test, *args):             # <<<<<<<<<<<<<<
 *         name = test.__name__
 *         if args:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_5_perf_1add(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_5_perf_1add = {"add", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5_perf_1add, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_5_perf_1add(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_test = 0;
  PyObject *__pyx_v_args = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 1, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_test,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_test)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(13, 59, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 1) ? kwd_pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, used_pos_args, "add") < 0)) __PYX_ERR(13, 59, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_test = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add", 0, 1, 1, __pyx_nargs); __PYX_ERR(13, 59, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("cavro._perf.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_5_perf_add(__pyx_self, __pyx_v_test, __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_5_perf_3add_2generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "cavro.pyx":62
 *         name = test.__name__
 *         if args:
 *             name += '-' + ('-'.join(str(a) for a in args))             # <<<<<<<<<<<<<<
 *         __perf_tests[name] = (test, ) + args
 *     fn(add)
 */

static PyObject *__pyx_pf_5cavro_5_perf_3add_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_1_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(13, 62, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_5_perf_3add_2generator6, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_perf_locals_add_locals_genexpr, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(13, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro._perf.add.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_5_perf_3add_2generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(13, 62, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(13, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(13, 62, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(13, 62, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(13, 62, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_a);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_a, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Str(__pyx_cur_scope->__pyx_v_a); if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_3))) __PYX_ERR(13, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cavro.pyx":59
 * __perf_tests = {}
 * def _perf(fn):
 *     def add(test, *args):             # <<<<<<<<<<<<<<
 *         name = test.__name__
 *         if args:
 */

static PyObject *__pyx_pf_5cavro_5_perf_add(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_test, PyObject *__pyx_v_args) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_gb_5cavro_5_perf_3add_2generator6 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add", 1);

  /* "cavro.pyx":60
 * def _perf(fn):
 *     def add(test, *args):
 *         name = test.__name__             # <<<<<<<<<<<<<<
 *         if args:
 *             name += '-' + ('-'.join(str(a) for a in args))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_test, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cavro.pyx":61
 *     def add(test, *args):
 *         name = test.__name__
 *         if args:             # <<<<<<<<<<<<<<
 *             name += '-' + ('-'.join(str(a) for a in args))
 *         __perf_tests[name] = (test, ) + args
 */
  __pyx_t_2 = (PyTuple_GET_SIZE(__pyx_v_args) != 0);
  if (__pyx_t_2) {

    /* "cavro.pyx":62
 *         name = test.__name__
 *         if args:
 *             name += '-' + ('-'.join(str(a) for a in args))             # <<<<<<<<<<<<<<
 *         __perf_tests[name] = (test, ) + args
 *     fn(add)
 */
    __pyx_t_1 = __pyx_pf_5cavro_5_perf_3add_genexpr(NULL, __pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__9, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u__9, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_name, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_name, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "cavro.pyx":61
 *     def add(test, *args):
 *         name = test.__name__
 *         if args:             # <<<<<<<<<<<<<<
 *             name += '-' + ('-'.join(str(a) for a in args))
 *         __perf_tests[name] = (test, ) + args
 */
  }

  /* "cavro.pyx":63
 *         if args:
 *             name += '-' + ('-'.join(str(a) for a in args))
 *         __perf_tests[name] = (test, ) + args             # <<<<<<<<<<<<<<
 *     fn(add)
 *     return _perf
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_test);
  __Pyx_GIVEREF(__pyx_v_test);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_test)) __PYX_ERR(13, 63, __pyx_L1_error);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(13, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_perf_tests); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely((PyObject_SetItem(__pyx_t_1, __pyx_v_name, __pyx_t_3) < 0))) __PYX_ERR(13, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cavro.pyx":59
 * __perf_tests = {}
 * def _perf(fn):
 *     def add(test, *args):             # <<<<<<<<<<<<<<
 *         name = test.__name__
 *         if args:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro._perf.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_gb_5cavro_5_perf_3add_2generator6);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cavro.pyx":58
 * 
 * __perf_tests = {}
 * def _perf(fn):             # <<<<<<<<<<<<<<
 *     def add(test, *args):
 *         name = test.__name__
 */

static PyObject *__pyx_pf_5cavro_2_perf(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_fn) {
  PyObject *__pyx_v_add = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_perf", 1);

  /* "cavro.pyx":59
 * __perf_tests = {}
 * def _perf(fn):
 *     def add(test, *args):             # <<<<<<<<<<<<<<
 *         name = test.__name__
 *         if args:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_5_perf_1add, 0, __pyx_n_s_perf_locals_add, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__12)); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_add = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cavro.pyx":64
 *             name += '-' + ('-'.join(str(a) for a in args))
 *         __perf_tests[name] = (test, ) + args
 *     fn(add)             # <<<<<<<<<<<<<<
 *     return _perf
 * 
 */
  __Pyx_INCREF(__pyx_v_fn);
  __pyx_t_2 = __pyx_v_fn; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_add};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cavro.pyx":65
 *         __perf_tests[name] = (test, ) + args
 *     fn(add)
 *     return _perf             # <<<<<<<<<<<<<<
 * 
 * include "src/error.pxi"
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_perf); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cavro.pyx":58
 * 
 * __perf_tests = {}
 * def _perf(fn):             # <<<<<<<<<<<<<<
 *     def add(test, *args):
 *         name = test.__name__
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro._perf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_add);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/error.pxi":20
 *     """
 * 
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *         super().__init__(f"Unknown type: {name}")
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11UnknownType_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_11UnknownType_1__init__ = {"__init__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11UnknownType_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_11UnknownType_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 20, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 20, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 20, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 20, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 20, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UnknownType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_11UnknownType___init__(__pyx_self, __pyx_v_self, __pyx_v_name);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11UnknownType___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/error.pxi":21
 * 
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *         super().__init__(f"Unknown type: {name}")
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(0, 21, __pyx_L1_error)

  /* "src/error.pxi":22
 *     def __init__(self, name):
 *         self.name = name
 *         super().__init__(f"Unknown type: {name}")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 22, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self)) __PYX_ERR(0, 22, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_name, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Unknown_type, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/error.pxi":20
 *     """
 * 
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *         super().__init__(f"Unknown type: {name}")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.UnknownType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/error.pxi":61
 *     """
 * 
 *     def __init__(self, reader_type, writer_type, extra=None):             # <<<<<<<<<<<<<<
 *         self.reader_type = reader_type
 *         self.writer_type = writer_type
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_18CannotPromoteError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_18CannotPromoteError_1__init__ = {"__init__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_18CannotPromoteError_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_18CannotPromoteError_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_reader_type = 0;
  PyObject *__pyx_v_writer_type = 0;
  PyObject *__pyx_v_extra = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_reader_type,&__pyx_n_s_writer_type,&__pyx_n_s_extra,0};
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_reader_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 1); __PYX_ERR(0, 61, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_writer_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 2); __PYX_ERR(0, 61, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_extra);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 61, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_reader_type = values[1];
    __pyx_v_writer_type = values[2];
    __pyx_v_extra = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 61, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.CannotPromoteError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_18CannotPromoteError___init__(__pyx_self, __pyx_v_self, __pyx_v_reader_type, __pyx_v_writer_type, __pyx_v_extra);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_18CannotPromoteError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_reader_type, PyObject *__pyx_v_writer_type, PyObject *__pyx_v_extra) {
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/error.pxi":62
 * 
 *     def __init__(self, reader_type, writer_type, extra=None):
 *         self.reader_type = reader_type             # <<<<<<<<<<<<<<
 *         self.writer_type = writer_type
 *         self.extra = extra
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_reader_type, __pyx_v_reader_type) < 0) __PYX_ERR(0, 62, __pyx_L1_error)

  /* "src/error.pxi":63
 *     def __init__(self, reader_type, writer_type, extra=None):
 *         self.reader_type = reader_type
 *         self.writer_type = writer_type             # <<<<<<<<<<<<<<
 *         self.extra = extra
 *         msg = f"Cannot promote {reader_type.get_schema()} to {writer_type.get_schema()}"
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_writer_type, __pyx_v_writer_type) < 0) __PYX_ERR(0, 63, __pyx_L1_error)

  /* "src/error.pxi":64
 *         self.reader_type = reader_type
 *         self.writer_type = writer_type
 *         self.extra = extra             # <<<<<<<<<<<<<<
 *         msg = f"Cannot promote {reader_type.get_schema()} to {writer_type.get_schema()}"
 *         if extra:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_extra, __pyx_v_extra) < 0) __PYX_ERR(0, 64, __pyx_L1_error)

  /* "src/error.pxi":65
 *         self.writer_type = writer_type
 *         self.extra = extra
 *         msg = f"Cannot promote {reader_type.get_schema()} to {writer_type.get_schema()}"             # <<<<<<<<<<<<<<
 *         if extra:
 *             msg += ': ' + extra
 */
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Cannot_promote);
  __pyx_t_2 += 15;
  __Pyx_GIVEREF(__pyx_kp_u_Cannot_promote);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Cannot_promote);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_reader_type, __pyx_n_s_get_schema); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u_to);
  __pyx_t_2 += 4;
  __Pyx_GIVEREF(__pyx_kp_u_to);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_to);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer_type, __pyx_n_s_get_schema); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_msg = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "src/error.pxi":66
 *         self.extra = extra
 *         msg = f"Cannot promote {reader_type.get_schema()} to {writer_type.get_schema()}"
 *         if extra:             # <<<<<<<<<<<<<<
 *             msg += ': ' + extra
 *         super().__init__(msg)
 */
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_extra); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 66, __pyx_L1_error)
  if (__pyx_t_8) {

    /* "src/error.pxi":67
 *         msg = f"Cannot promote {reader_type.get_schema()} to {writer_type.get_schema()}"
 *         if extra:
 *             msg += ': ' + extra             # <<<<<<<<<<<<<<
 *         super().__init__(msg)
 * 
 */
    __pyx_t_4 = PyNumber_Add(__pyx_kp_u_, __pyx_v_extra); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_msg, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_msg, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/error.pxi":66
 *         self.extra = extra
 *         msg = f"Cannot promote {reader_type.get_schema()} to {writer_type.get_schema()}"
 *         if extra:             # <<<<<<<<<<<<<<
 *             msg += ': ' + extra
 *         super().__init__(msg)
 */
  }

  /* "src/error.pxi":68
 *         if extra:
 *             msg += ': ' + extra
 *         super().__init__(msg)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_4) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 68, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4)) __PYX_ERR(0, 68, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self)) __PYX_ERR(0, 68, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_init); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_msg};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/error.pxi":61
 *     """
 * 
 *     def __init__(self, reader_type, writer_type, extra=None):             # <<<<<<<<<<<<<<
 *         self.reader_type = reader_type
 *         self.writer_type = writer_type
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.CannotPromoteError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/error.pxi":82
 *     """
 * 
 *     def __init__(self, value, dest_type, path=()):             # <<<<<<<<<<<<<<
 *         self.value = value
 *         self.dest_type = dest_type
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12InvalidValue_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_12InvalidValue_1__init__ = {"__init__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12InvalidValue_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_12InvalidValue_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_dest_type = 0;
  PyObject *__pyx_v_path = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_value,&__pyx_n_s_dest_type,&__pyx_n_s_path,0};
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)((PyObject*)__pyx_empty_tuple)));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 82, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 82, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 1); __PYX_ERR(0, 82, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dest_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 82, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 2); __PYX_ERR(0, 82, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_path);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 82, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 82, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_value = values[1];
    __pyx_v_dest_type = values[2];
    __pyx_v_path = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, __pyx_nargs); __PYX_ERR(0, 82, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.InvalidValue.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_12InvalidValue___init__(__pyx_self, __pyx_v_self, __pyx_v_value, __pyx_v_dest_type, __pyx_v_path);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12InvalidValue___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_dest_type, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/error.pxi":83
 * 
 *     def __init__(self, value, dest_type, path=()):
 *         self.value = value             # <<<<<<<<<<<<<<
 *         self.dest_type = dest_type
 *         self.schema_path = path
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_value, __pyx_v_value) < 0) __PYX_ERR(0, 83, __pyx_L1_error)

  /* "src/error.pxi":84
 *     def __init__(self, value, dest_type, path=()):
 *         self.value = value
 *         self.dest_type = dest_type             # <<<<<<<<<<<<<<
 *         self.schema_path = path
 *         super().__init__()
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_dest_type, __pyx_v_dest_type) < 0) __PYX_ERR(0, 84, __pyx_L1_error)

  /* "src/error.pxi":85
 *         self.value = value
 *         self.dest_type = dest_type
 *         self.schema_path = path             # <<<<<<<<<<<<<<
 *         super().__init__()
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_schema_path, __pyx_v_path) < 0) __PYX_ERR(0, 85, __pyx_L1_error)

  /* "src/error.pxi":86
 *         self.dest_type = dest_type
 *         self.schema_path = path
 *         super().__init__()             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 86, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self)) __PYX_ERR(0, 86, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/error.pxi":82
 *     """
 * 
 *     def __init__(self, value, dest_type, path=()):             # <<<<<<<<<<<<<<
 *         self.value = value
 *         self.dest_type = dest_type
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.InvalidValue.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/error.pxi":88
 *         super().__init__()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.schema_path:
 *             path_str = ".".join(str(p) for p in self.schema_path)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12InvalidValue_3__str__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_12InvalidValue_3__str__ = {"__str__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12InvalidValue_3__str__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_12InvalidValue_3__str__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 88, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__str__") < 0)) __PYX_ERR(0, 88, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__str__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 88, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.InvalidValue.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_12InvalidValue_2__str__(__pyx_self, __pyx_v_self);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_12InvalidValue_7__str___2generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/error.pxi":90
 *     def __str__(self):
 *         if self.schema_path:
 *             path_str = ".".join(str(p) for p in self.schema_path)             # <<<<<<<<<<<<<<
 *             return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name} at {path_str}"
 *         return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name}"
 */

static PyObject *__pyx_pf_5cavro_12InvalidValue_7__str___genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_2_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_2_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 90, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_12InvalidValue_7__str___2generator7, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_InvalidValue___str___locals_gene, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.InvalidValue.__str__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_12InvalidValue_7__str___2generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 90, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 90, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 90, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 90, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 90, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 90, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 90, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 90, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 90, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 90, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_p);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_p, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Str(__pyx_cur_scope->__pyx_v_p); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/error.pxi":88
 *         super().__init__()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.schema_path:
 *             path_str = ".".join(str(p) for p in self.schema_path)
 */

static PyObject *__pyx_pf_5cavro_12InvalidValue_2__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_path_str = NULL;
  PyObject *__pyx_gb_5cavro_12InvalidValue_7__str___2generator7 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "src/error.pxi":89
 * 
 *     def __str__(self):
 *         if self.schema_path:             # <<<<<<<<<<<<<<
 *             path_str = ".".join(str(p) for p in self.schema_path)
 *             return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name} at {path_str}"
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_schema_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/error.pxi":90
 *     def __str__(self):
 *         if self.schema_path:
 *             path_str = ".".join(str(p) for p in self.schema_path)             # <<<<<<<<<<<<<<
 *             return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name} at {path_str}"
 *         return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name}"
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_schema_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_pf_5cavro_12InvalidValue_7__str___genexpr(NULL, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_Generator_Next(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyUnicode_Join(__pyx_kp_u__2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_path_str = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/error.pxi":91
 *         if self.schema_path:
 *             path_str = ".".join(str(p) for p in self.schema_path)
 *             return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name} at {path_str}"             # <<<<<<<<<<<<<<
 *         return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name}"
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Invalid_value);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_Invalid_value);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Invalid_value);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_6, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_for_type);
    __pyx_t_4 += 10;
    __Pyx_GIVEREF(__pyx_kp_u_for_type);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_for_type);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_dest_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_type_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_6, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_at);
    __pyx_t_4 += 4;
    __Pyx_GIVEREF(__pyx_kp_u_at);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_at);
    __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_v_path_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_3, 6, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/error.pxi":89
 * 
 *     def __str__(self):
 *         if self.schema_path:             # <<<<<<<<<<<<<<
 *             path_str = ".".join(str(p) for p in self.schema_path)
 *             return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name} at {path_str}"
 */
  }

  /* "src/error.pxi":92
 *             path_str = ".".join(str(p) for p in self.schema_path)
 *             return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name} at {path_str}"
 *         return f"Invalid value {repr(self.value)} for type {self.dest_type.type_name}"             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = 0;
  __pyx_t_5 = 127;
  __Pyx_INCREF(__pyx_kp_u_Invalid_value);
  __pyx_t_4 += 14;
  __Pyx_GIVEREF(__pyx_kp_u_Invalid_value);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Invalid_value);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = PyObject_Repr(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_6, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_5;
  __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u_for_type);
  __pyx_t_4 += 10;
  __Pyx_GIVEREF(__pyx_kp_u_for_type);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_for_type);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_dest_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_type_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_6, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_5;
  __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/error.pxi":88
 *         super().__init__()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.schema_path:
 *             path_str = ".".join(str(p) for p in self.schema_path)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.InvalidValue.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_path_str);
  __Pyx_XDECREF(__pyx_gb_5cavro_12InvalidValue_7__str___2generator7);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":6
 * NAME_RE_UNICODE = None
 * 
 * cdef _get_unicode_name_re():             # <<<<<<<<<<<<<<
 *     global NAME_RE_UNICODE
 *     if NAME_RE_UNICODE is None:
 */

static PyObject *__pyx_f_5cavro__get_unicode_name_re(void) {
  PyObject *__pyx_v_unicodedata = NULL;
  PyObject *__pyx_v_sys = NULL;
  PyObject *__pyx_v_letters = NULL;
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_v_letter_pattern = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_UCS4 __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_unicode_name_re", 1);

  /* "src/option.pxi":8
 * cdef _get_unicode_name_re():
 *     global NAME_RE_UNICODE
 *     if NAME_RE_UNICODE is None:             # <<<<<<<<<<<<<<
 *         import unicodedata, sys
 *         letters = set()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NAME_RE_UNICODE); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/option.pxi":9
 *     global NAME_RE_UNICODE
 *     if NAME_RE_UNICODE is None:
 *         import unicodedata, sys             # <<<<<<<<<<<<<<
 *         letters = set()
 *         for c in range(sys.maxunicode + 1):
 */
    __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_unicodedata, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_unicodedata = __pyx_t_1;
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_sys, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_sys = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "src/option.pxi":10
 *     if NAME_RE_UNICODE is None:
 *         import unicodedata, sys
 *         letters = set()             # <<<<<<<<<<<<<<
 *         for c in range(sys.maxunicode + 1):
 *             if unicodedata.category(chr(c)) in ('Ll', 'Lu'):
 */
    __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_letters = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/option.pxi":11
 *         import unicodedata, sys
 *         letters = set()
 *         for c in range(sys.maxunicode + 1):             # <<<<<<<<<<<<<<
 *             if unicodedata.category(chr(c)) in ('Ll', 'Lu'):
 *                 letters.add(chr(c))
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_maxunicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 11, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 11, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 11, __pyx_L1_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(4, 11, __pyx_L1_error)
          #else
          __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 11, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 11, __pyx_L1_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(4, 11, __pyx_L1_error)
          #else
          __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 11, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(4, 11, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "src/option.pxi":12
 *         letters = set()
 *         for c in range(sys.maxunicode + 1):
 *             if unicodedata.category(chr(c)) in ('Ll', 'Lu'):             # <<<<<<<<<<<<<<
 *                 letters.add(chr(c))
 *         letter_pattern = re.escape(''.join(letters))
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_unicodedata, __pyx_n_s_category); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 12, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_chr, __pyx_v_c); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 12, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_9 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_7};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 12, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_Ll, Py_EQ)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(4, 12, __pyx_L1_error)
      if (!__pyx_t_10) {
      } else {
        __pyx_t_2 = __pyx_t_10;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_Lu, Py_EQ)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(4, 12, __pyx_L1_error)
      __pyx_t_2 = __pyx_t_10;
      __pyx_L7_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_10 = __pyx_t_2;
      if (__pyx_t_10) {

        /* "src/option.pxi":13
 *         for c in range(sys.maxunicode + 1):
 *             if unicodedata.category(chr(c)) in ('Ll', 'Lu'):
 *                 letters.add(chr(c))             # <<<<<<<<<<<<<<
 *         letter_pattern = re.escape(''.join(letters))
 *         NAME_RE_UNICODE = re.compile(fr'[{letter_pattern}_]\w*')
 */
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_chr, __pyx_v_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 13, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PySet_Add(__pyx_v_letters, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(4, 13, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "src/option.pxi":12
 *         letters = set()
 *         for c in range(sys.maxunicode + 1):
 *             if unicodedata.category(chr(c)) in ('Ll', 'Lu'):             # <<<<<<<<<<<<<<
 *                 letters.add(chr(c))
 *         letter_pattern = re.escape(''.join(letters))
 */
      }

      /* "src/option.pxi":11
 *         import unicodedata, sys
 *         letters = set()
 *         for c in range(sys.maxunicode + 1):             # <<<<<<<<<<<<<<
 *             if unicodedata.category(chr(c)) in ('Ll', 'Lu'):
 *                 letters.add(chr(c))
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "src/option.pxi":14
 *             if unicodedata.category(chr(c)) in ('Ll', 'Lu'):
 *                 letters.add(chr(c))
 *         letter_pattern = re.escape(''.join(letters))             # <<<<<<<<<<<<<<
 *         NAME_RE_UNICODE = re.compile(fr'[{letter_pattern}_]\w*')
 *     return NAME_RE_UNICODE
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_re); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_escape); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__13, __pyx_v_letters); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_1};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_v_letter_pattern = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "src/option.pxi":15
 *                 letters.add(chr(c))
 *         letter_pattern = re.escape(''.join(letters))
 *         NAME_RE_UNICODE = re.compile(fr'[{letter_pattern}_]\w*')             # <<<<<<<<<<<<<<
 *     return NAME_RE_UNICODE
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_re); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_compile); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = 0;
    __pyx_t_12 = 127;
    __Pyx_INCREF(__pyx_kp_u__14);
    __pyx_t_4 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__14);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_u__14);
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_v_letter_pattern, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_12;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_w);
    __pyx_t_4 += 5;
    __Pyx_GIVEREF(__pyx_kp_u_w);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_kp_u_w);
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_6, 3, __pyx_t_4, __pyx_t_12); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_7};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 15, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_NAME_RE_UNICODE, __pyx_t_3) < 0) __PYX_ERR(4, 15, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "src/option.pxi":8
 * cdef _get_unicode_name_re():
 *     global NAME_RE_UNICODE
 *     if NAME_RE_UNICODE is None:             # <<<<<<<<<<<<<<
 *         import unicodedata, sys
 *         letters = set()
 */
  }

  /* "src/option.pxi":16
 *         letter_pattern = re.escape(''.join(letters))
 *         NAME_RE_UNICODE = re.compile(fr'[{letter_pattern}_]\w*')
 *     return NAME_RE_UNICODE             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_NAME_RE_UNICODE); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/option.pxi":6
 * NAME_RE_UNICODE = None
 * 
 * cdef _get_unicode_name_re():             # <<<<<<<<<<<<<<
 *     global NAME_RE_UNICODE
 *     if NAME_RE_UNICODE is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cavro._get_unicode_name_re", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_unicodedata);
  __Pyx_XDECREF(__pyx_v_sys);
  __Pyx_XDECREF(__pyx_v_letters);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XDECREF(__pyx_v_letter_pattern);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":49
 * _EMPTY_MP = MappingProxyType({})
 * 
 * def _eval(val):             # <<<<<<<<<<<<<<
 *     if isinstance(val, str):
 *         if val == 'bint':
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_5_eval(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_5_eval = {"_eval", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5_eval, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_5_eval(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_val = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_eval (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_val,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_val)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 49, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_eval") < 0)) __PYX_ERR(4, 49, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_val = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_eval", 1, 1, 1, __pyx_nargs); __PYX_ERR(4, 49, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._eval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_4_eval(__pyx_self, __pyx_v_val);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_4_eval(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_val) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_eval", 0);
  __Pyx_INCREF(__pyx_v_val);

  /* "src/option.pxi":50
 * 
 * def _eval(val):
 *     if isinstance(val, str):             # <<<<<<<<<<<<<<
 *         if val == 'bint':
 *             val = 'bool'
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_val); 
  if (__pyx_t_1) {

    /* "src/option.pxi":51
 * def _eval(val):
 *     if isinstance(val, str):
 *         if val == 'bint':             # <<<<<<<<<<<<<<
 *             val = 'bool'
 *         return eval(val, globals(), {})
 */
    __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_val, __pyx_n_u_bint, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(4, 51, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "src/option.pxi":52
 *     if isinstance(val, str):
 *         if val == 'bint':
 *             val = 'bool'             # <<<<<<<<<<<<<<
 *         return eval(val, globals(), {})
 *     return val
 */
      __Pyx_INCREF(__pyx_n_u_bool);
      __Pyx_DECREF_SET(__pyx_v_val, __pyx_n_u_bool);

      /* "src/option.pxi":51
 * def _eval(val):
 *     if isinstance(val, str):
 *         if val == 'bint':             # <<<<<<<<<<<<<<
 *             val = 'bool'
 *         return eval(val, globals(), {})
 */
    }

    /* "src/option.pxi":53
 *         if val == 'bint':
 *             val = 'bool'
 *         return eval(val, globals(), {})             # <<<<<<<<<<<<<<
 *     return val
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_Globals(); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_val);
    __Pyx_GIVEREF(__pyx_v_val);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_val)) __PYX_ERR(4, 53, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(4, 53, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3)) __PYX_ERR(4, 53, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_eval, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "src/option.pxi":50
 * 
 * def _eval(val):
 *     if isinstance(val, str):             # <<<<<<<<<<<<<<
 *         if val == 'bint':
 *             val = 'bool'
 */
  }

  /* "src/option.pxi":54
 *             val = 'bool'
 *         return eval(val, globals(), {})
 *     return val             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_val);
  __pyx_r = __pyx_v_val;
  goto __pyx_L0;

  /* "src/option.pxi":49
 * _EMPTY_MP = MappingProxyType({})
 * 
 * def _eval(val):             # <<<<<<<<<<<<<<
 *     if isinstance(val, str):
 *         if val == 'bint':
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._eval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":70
 *     '''
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.__signature__ = None
 *         self.annotations = {}
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_17_SignatureWrapper_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_17_SignatureWrapper_1__init__ = {"__init__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_17_SignatureWrapper_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_17_SignatureWrapper_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 70, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(4, 70, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(4, 70, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._SignatureWrapper.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_17_SignatureWrapper___init__(__pyx_self, __pyx_v_self);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_17_SignatureWrapper___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/option.pxi":71
 * 
 *     def __init__(self):
 *         self.__signature__ = None             # <<<<<<<<<<<<<<
 *         self.annotations = {}
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_signature, Py_None) < 0) __PYX_ERR(4, 71, __pyx_L1_error)

  /* "src/option.pxi":72
 *     def __init__(self):
 *         self.__signature__ = None
 *         self.annotations = {}             # <<<<<<<<<<<<<<
 * 
 *     def _set_from_dataclass(self, cls):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_annotations, __pyx_t_1) < 0) __PYX_ERR(4, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/option.pxi":70
 *     '''
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.__signature__ = None
 *         self.annotations = {}
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._SignatureWrapper.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":74
 *         self.annotations = {}
 * 
 *     def _set_from_dataclass(self, cls):             # <<<<<<<<<<<<<<
 *         try:
 *             fields = [(f.name, f.type, f) for f in dataclasses.fields(cls)]
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_17_SignatureWrapper_3_set_from_dataclass(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_17_SignatureWrapper_3_set_from_dataclass = {"_set_from_dataclass", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_17_SignatureWrapper_3_set_from_dataclass, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_17_SignatureWrapper_3_set_from_dataclass(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_cls = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_from_dataclass (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_cls,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 74, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_cls)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 74, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_set_from_dataclass", 1, 2, 2, 1); __PYX_ERR(4, 74, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_set_from_dataclass") < 0)) __PYX_ERR(4, 74, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_cls = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_set_from_dataclass", 1, 2, 2, __pyx_nargs); __PYX_ERR(4, 74, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._SignatureWrapper._set_from_dataclass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_17_SignatureWrapper_2_set_from_dataclass(__pyx_self, __pyx_v_self, __pyx_v_cls);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_17_SignatureWrapper_2_set_from_dataclass(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_cls) {
  PyObject *__pyx_v_fields = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_type_str = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_inst = NULL;
  PyObject *__pyx_8genexpr3__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_from_dataclass", 1);

  /* "src/option.pxi":75
 * 
 *     def _set_from_dataclass(self, cls):
 *         try:             # <<<<<<<<<<<<<<
 *             fields = [(f.name, f.type, f) for f in dataclasses.fields(cls)]
 *             for name, type_str, _ in fields:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "src/option.pxi":76
 *     def _set_from_dataclass(self, cls):
 *         try:
 *             fields = [(f.name, f.type, f) for f in dataclasses.fields(cls)]             # <<<<<<<<<<<<<<
 *             for name, type_str, _ in fields:
 *                 self.annotations[name] = _eval(type_str)
 */
      { /* enter inner scope */
        __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 76, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_dataclasses); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 76, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 76, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_cls};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 76, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
          __pyx_t_7 = __pyx_t_5; __Pyx_INCREF(__pyx_t_7);
          __pyx_t_9 = 0;
          __pyx_t_10 = NULL;
        } else {
          __pyx_t_9 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 76, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 76, __pyx_L11_error)
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        for (;;) {
          if (likely(!__pyx_t_10)) {
            if (likely(PyList_CheckExact(__pyx_t_7))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_7);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 76, __pyx_L11_error)
                #endif
                if (__pyx_t_9 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_9); __Pyx_INCREF(__pyx_t_5); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(4, 76, __pyx_L11_error)
              #else
              __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 76, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_7);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 76, __pyx_L11_error)
                #endif
                if (__pyx_t_9 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_9); __Pyx_INCREF(__pyx_t_5); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(4, 76, __pyx_L11_error)
              #else
              __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 76, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            }
          } else {
            __pyx_t_5 = __pyx_t_10(__pyx_t_7);
            if (unlikely(!__pyx_t_5)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(4, 76, __pyx_L11_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_5);
          }
          __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_f, __pyx_t_5);
          __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr3__pyx_v_f, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 76, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr3__pyx_v_f, __pyx_n_s_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 76, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 76, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_GIVEREF(__pyx_t_5);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_5)) __PYX_ERR(4, 76, __pyx_L11_error);
          __Pyx_GIVEREF(__pyx_t_6);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_6)) __PYX_ERR(4, 76, __pyx_L11_error);
          __Pyx_INCREF(__pyx_8genexpr3__pyx_v_f);
          __Pyx_GIVEREF(__pyx_8genexpr3__pyx_v_f);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_8genexpr3__pyx_v_f)) __PYX_ERR(4, 76, __pyx_L11_error);
          __pyx_t_5 = 0;
          __pyx_t_6 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_11))) __PYX_ERR(4, 76, __pyx_L11_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_f); __pyx_8genexpr3__pyx_v_f = 0;
        goto __pyx_L15_exit_scope;
        __pyx_L11_error:;
        __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_f); __pyx_8genexpr3__pyx_v_f = 0;
        goto __pyx_L3_error;
        __pyx_L15_exit_scope:;
      } /* exit inner scope */
      __pyx_v_fields = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "src/option.pxi":77
 *         try:
 *             fields = [(f.name, f.type, f) for f in dataclasses.fields(cls)]
 *             for name, type_str, _ in fields:             # <<<<<<<<<<<<<<
 *                 self.annotations[name] = _eval(type_str)
 *             inst = dataclasses.make_dataclass(cls.__name__, fields)
 */
      __pyx_t_4 = __pyx_v_fields; __Pyx_INCREF(__pyx_t_4);
      __pyx_t_9 = 0;
      for (;;) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 77, __pyx_L3_error)
          #endif
          if (__pyx_t_9 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_7); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(4, 77, __pyx_L3_error)
        #else
        __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 77, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
        if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
          PyObject* sequence = __pyx_t_7;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 3)) {
            if (size > 3) __Pyx_RaiseTooManyValuesError(3);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(4, 77, __pyx_L3_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_11 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
            __pyx_t_5 = PyTuple_GET_ITEM(sequence, 2); 
          } else {
            __pyx_t_11 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
            __pyx_t_5 = PyList_GET_ITEM(sequence, 2); 
          }
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          #else
          __pyx_t_11 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 77, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 77, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 77, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_12 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 77, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_13 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_12);
          index = 0; __pyx_t_11 = __pyx_t_13(__pyx_t_12); if (unlikely(!__pyx_t_11)) goto __pyx_L18_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_11);
          index = 1; __pyx_t_6 = __pyx_t_13(__pyx_t_12); if (unlikely(!__pyx_t_6)) goto __pyx_L18_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_6);
          index = 2; __pyx_t_5 = __pyx_t_13(__pyx_t_12); if (unlikely(!__pyx_t_5)) goto __pyx_L18_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_5);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_12), 3) < 0) __PYX_ERR(4, 77, __pyx_L3_error)
          __pyx_t_13 = NULL;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          goto __pyx_L19_unpacking_done;
          __pyx_L18_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_13 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(4, 77, __pyx_L3_error)
          __pyx_L19_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_11);
        __pyx_t_11 = 0;
        __Pyx_XDECREF_SET(__pyx_v_type_str, __pyx_t_6);
        __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "src/option.pxi":78
 *             fields = [(f.name, f.type, f) for f in dataclasses.fields(cls)]
 *             for name, type_str, _ in fields:
 *                 self.annotations[name] = _eval(type_str)             # <<<<<<<<<<<<<<
 *             inst = dataclasses.make_dataclass(cls.__name__, fields)
 *             self.__signature__ = inspect.signature(inst)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_eval_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 78, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_type_str};
          __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 78, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_annotations); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 78, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely((PyObject_SetItem(__pyx_t_5, __pyx_v_name, __pyx_t_7) < 0))) __PYX_ERR(4, 78, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "src/option.pxi":77
 *         try:
 *             fields = [(f.name, f.type, f) for f in dataclasses.fields(cls)]
 *             for name, type_str, _ in fields:             # <<<<<<<<<<<<<<
 *                 self.annotations[name] = _eval(type_str)
 *             inst = dataclasses.make_dataclass(cls.__name__, fields)
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "src/option.pxi":79
 *             for name, type_str, _ in fields:
 *                 self.annotations[name] = _eval(type_str)
 *             inst = dataclasses.make_dataclass(cls.__name__, fields)             # <<<<<<<<<<<<<<
 *             self.__signature__ = inspect.signature(inst)
 *         except Exception:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_dataclasses); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 79, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_make_dataclass); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 79, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_name_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 79, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_t_7, __pyx_v_fields};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 79, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_v_inst = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "src/option.pxi":80
 *                 self.annotations[name] = _eval(type_str)
 *             inst = dataclasses.make_dataclass(cls.__name__, fields)
 *             self.__signature__ = inspect.signature(inst)             # <<<<<<<<<<<<<<
 *         except Exception:
 *             pass # 3.11 doesn't allow mappingproxy in frozen dataclasses, but 3.12 makes mappingproxy hashable, which should solve this
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_inspect); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 80, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_signature_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 80, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_inst};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 80, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_signature, __pyx_t_4) < 0) __PYX_ERR(4, 80, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "src/option.pxi":75
 * 
 *     def _set_from_dataclass(self, cls):
 *         try:             # <<<<<<<<<<<<<<
 *             fields = [(f.name, f.type, f) for f in dataclasses.fields(cls)]
 *             for name, type_str, _ in fields:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "src/option.pxi":81
 *             inst = dataclasses.make_dataclass(cls.__name__, fields)
 *             self.__signature__ = inspect.signature(inst)
 *         except Exception:             # <<<<<<<<<<<<<<
 *             pass # 3.11 doesn't allow mappingproxy in frozen dataclasses, but 3.12 makes mappingproxy hashable, which should solve this
 *             # We only really use this for documentation, so avoiding the issue should be easy
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_8) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "src/option.pxi":75
 * 
 *     def _set_from_dataclass(self, cls):
 *         try:             # <<<<<<<<<<<<<<
 *             fields = [(f.name, f.type, f) for f in dataclasses.fields(cls)]
 *             for name, type_str, _ in fields:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "src/option.pxi":74
 *         self.annotations = {}
 * 
 *     def _set_from_dataclass(self, cls):             # <<<<<<<<<<<<<<
 *         try:
 *             fields = [(f.name, f.type, f) for f in dataclasses.fields(cls)]
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("cavro._SignatureWrapper._set_from_dataclass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_type_str);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_inst);
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":285
 *     externally_defined_types: MappingProxyType = _EMPTY_MP
 * 
 *     def replace(self, **changes) -> Options:             # <<<<<<<<<<<<<<
 *         '''
 *         Return a copy of the options with the specified fields replaced.
 */

/* Python wrapper */
static struct __pyx_obj_5cavro_Options *__pyx_pw_5cavro_7Options_1replace(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_7Options_replace, "\n        Return a copy of the options with the specified fields replaced.\n        This is a simple wrapper around `dataclasses.replace`\n        ");
static PyMethodDef __pyx_mdef_5cavro_7Options_1replace = {"replace", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7Options_1replace, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_7Options_replace};
static struct __pyx_obj_5cavro_Options *__pyx_pw_5cavro_7Options_1replace(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_changes = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  struct __pyx_obj_5cavro_Options *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("replace (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("replace", 1, 0, 0, __pyx_nargs); return NULL;}
  if (__pyx_kwds && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "replace", 1))) return NULL;
  if (__pyx_kwds) {
    __pyx_v_changes = __Pyx_KwargsAsDict_FASTCALL(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_changes)) return NULL;
    __Pyx_GOTREF(__pyx_v_changes);
  } else {
    __pyx_v_changes = PyDict_New();
    if (unlikely(!__pyx_v_changes)) return NULL;
    __Pyx_GOTREF(__pyx_v_changes);
  }
  __pyx_r = __pyx_pf_5cavro_7Options_replace(((struct __pyx_obj_5cavro_Options *)__pyx_v_self), __pyx_v_changes);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_changes);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_5cavro_Options *__pyx_pf_5cavro_7Options_replace(struct __pyx_obj_5cavro_Options *__pyx_v_self, PyObject *__pyx_v_changes) {
  struct __pyx_obj_5cavro_Options *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("replace", 1);

  /* "src/option.pxi":290
 *         This is a simple wrapper around `dataclasses.replace`
 *         '''
 *         return dataclasses.replace(self, **changes)             # <<<<<<<<<<<<<<
 * 
 *     def with_logical_types(self, *logical_types):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_dataclasses); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_replace); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self))) __PYX_ERR(4, 290, __pyx_L1_error);
  __pyx_t_3 = PyDict_Copy(__pyx_v_changes); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5cavro_Options))))) __PYX_ERR(4, 290, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_5cavro_Options *)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "src/option.pxi":285
 *     externally_defined_types: MappingProxyType = _EMPTY_MP
 * 
 *     def replace(self, **changes) -> Options:             # <<<<<<<<<<<<<<
 *         '''
 *         Return a copy of the options with the specified fields replaced.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Options.replace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":292
 *         return dataclasses.replace(self, **changes)
 * 
 *     def with_logical_types(self, *logical_types):             # <<<<<<<<<<<<<<
 *         '''
 *         Return a copy of the options with additional logical types added.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_3with_logical_types(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_7Options_2with_logical_types, "\n        Return a copy of the options with additional logical types added.\n        ");
static PyMethodDef __pyx_mdef_5cavro_7Options_3with_logical_types = {"with_logical_types", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5cavro_7Options_3with_logical_types, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5cavro_7Options_2with_logical_types};
static PyObject *__pyx_pw_5cavro_7Options_3with_logical_types(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_logical_types = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("with_logical_types (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "with_logical_types", 0))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_logical_types = __pyx_args;
  __pyx_r = __pyx_pf_5cavro_7Options_2with_logical_types(((struct __pyx_obj_5cavro_Options *)__pyx_v_self), __pyx_v_logical_types);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_logical_types);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_2with_logical_types(struct __pyx_obj_5cavro_Options *__pyx_v_self, PyObject *__pyx_v_logical_types) {
  PyObject *__pyx_v_new_types = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("with_logical_types", 1);

  /* "src/option.pxi":296
 *         Return a copy of the options with additional logical types added.
 *         '''
 *         new_types = logical_types + self.logical_types             # <<<<<<<<<<<<<<
 *         return self.replace(logical_types=new_types)
 * 
 */
  __pyx_t_1 = PyNumber_Add(__pyx_v_logical_types, __pyx_v_self->logical_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_new_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/option.pxi":297
 *         '''
 *         new_types = logical_types + self.logical_types
 *         return self.replace(logical_types=new_types)             # <<<<<<<<<<<<<<
 * 
 *     def with_external_types(self, named_types: dict[str, AvroType]) -> Options:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_replace); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_logical_types, __pyx_v_new_types) < 0) __PYX_ERR(4, 297, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/option.pxi":292
 *         return dataclasses.replace(self, **changes)
 * 
 *     def with_logical_types(self, *logical_types):             # <<<<<<<<<<<<<<
 *         '''
 *         Return a copy of the options with additional logical types added.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.Options.with_logical_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_new_types);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":299
 *         return self.replace(logical_types=new_types)
 * 
 *     def with_external_types(self, named_types: dict[str, AvroType]) -> Options:             # <<<<<<<<<<<<<<
 *         '''
 *         Return a copy of the options with pre-parsed external named types added
 */

/* Python wrapper */
static struct __pyx_obj_5cavro_Options *__pyx_pw_5cavro_7Options_5with_external_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_7Options_4with_external_types, "\n        Return a copy of the options with pre-parsed external named types added\n        (Allows for references to types in the schema that have not been defined in the schema)\n        ");
static PyMethodDef __pyx_mdef_5cavro_7Options_5with_external_types = {"with_external_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7Options_5with_external_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_7Options_4with_external_types};
static struct __pyx_obj_5cavro_Options *__pyx_pw_5cavro_7Options_5with_external_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_named_types = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_5cavro_Options *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("with_external_types (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_named_types,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_named_types)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 299, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "with_external_types") < 0)) __PYX_ERR(4, 299, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_named_types = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("with_external_types", 1, 1, 1, __pyx_nargs); __PYX_ERR(4, 299, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Options.with_external_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_named_types), (&PyDict_Type), 0, "named_types", 1))) __PYX_ERR(4, 299, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_7Options_4with_external_types(((struct __pyx_obj_5cavro_Options *)__pyx_v_self), __pyx_v_named_types);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_5cavro_Options *__pyx_pf_5cavro_7Options_4with_external_types(struct __pyx_obj_5cavro_Options *__pyx_v_self, PyObject *__pyx_v_named_types) {
  PyObject *__pyx_v_new_types = NULL;
  struct __pyx_obj_5cavro_Options *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("with_external_types", 1);

  /* "src/option.pxi":304
 *         (Allows for references to types in the schema that have not been defined in the schema)
 *         '''
 *         new_types = dict(self.externally_defined_types)             # <<<<<<<<<<<<<<
 *         new_types.update(named_types)
 *         return self.replace(externally_defined_types=new_types)
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_self->externally_defined_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_new_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/option.pxi":305
 *         '''
 *         new_types = dict(self.externally_defined_types)
 *         new_types.update(named_types)             # <<<<<<<<<<<<<<
 *         return self.replace(externally_defined_types=new_types)
 * 
 */
  __pyx_t_1 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_update, __pyx_v_new_types, __pyx_v_named_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/option.pxi":306
 *         new_types = dict(self.externally_defined_types)
 *         new_types.update(named_types)
 *         return self.replace(externally_defined_types=new_types)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_replace); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_externally_defined_types, __pyx_v_new_types) < 0) __PYX_ERR(4, 306, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_Options))))) __PYX_ERR(4, 306, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_5cavro_Options *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/option.pxi":299
 *         return self.replace(logical_types=new_types)
 * 
 *     def with_external_types(self, named_types: dict[str, AvroType]) -> Options:             # <<<<<<<<<<<<<<
 *         '''
 *         Return a copy of the options with pre-parsed external named types added
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.Options.with_external_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_new_types);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":308
 *         return self.replace(externally_defined_types=new_types)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def name_pattern(self):
 *         '''
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_12name_pattern_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_12name_pattern_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_12name_pattern___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_12name_pattern___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/option.pxi":313
 *         Returns the relevant regular expression for validating names based on the current options.
 *         '''
 *         if self.ascii_name_rules:             # <<<<<<<<<<<<<<
 *             return NAME_RE_STRICT
 *         return _get_unicode_name_re()
 */
  if (__pyx_v_self->ascii_name_rules) {

    /* "src/option.pxi":314
 *         '''
 *         if self.ascii_name_rules:
 *             return NAME_RE_STRICT             # <<<<<<<<<<<<<<
 *         return _get_unicode_name_re()
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NAME_RE_STRICT); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/option.pxi":313
 *         Returns the relevant regular expression for validating names based on the current options.
 *         '''
 *         if self.ascii_name_rules:             # <<<<<<<<<<<<<<
 *             return NAME_RE_STRICT
 *         return _get_unicode_name_re()
 */
  }

  /* "src/option.pxi":315
 *         if self.ascii_name_rules:
 *             return NAME_RE_STRICT
 *         return _get_unicode_name_re()             # <<<<<<<<<<<<<<
 * 
 *     cdef bint can_have_missing_values(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro__get_unicode_name_re(); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/option.pxi":308
 *         return self.replace(externally_defined_types=new_types)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def name_pattern(self):
 *         '''
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.name_pattern.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":317
 *         return _get_unicode_name_re()
 * 
 *     cdef bint can_have_missing_values(self):             # <<<<<<<<<<<<<<
 *         '''
 *         If we are expecting a value, and there's no default in the schema, is it ever possible for that value to be absent without error?
 */

static int __pyx_f_5cavro_7Options_can_have_missing_values(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  int __pyx_r;
  int __pyx_t_1;

  /* "src/option.pxi":321
 *         If we are expecting a value, and there's no default in the schema, is it ever possible for that value to be absent without error?
 *         '''
 *         return self.missing_values_can_be_empty_container or self.missing_values_can_be_null             # <<<<<<<<<<<<<<
 * 
 *     def equals(self, Options other: Options, ignore: Sequence[str]=()) -> bool:
 */
  if (!__pyx_v_self->missing_values_can_be_empty_container) {
  } else {
    __pyx_t_1 = __pyx_v_self->missing_values_can_be_empty_container;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_1 = __pyx_v_self->missing_values_can_be_null;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "src/option.pxi":317
 *         return _get_unicode_name_re()
 * 
 *     cdef bint can_have_missing_values(self):             # <<<<<<<<<<<<<<
 *         '''
 *         If we are expecting a value, and there's no default in the schema, is it ever possible for that value to be absent without error?
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "src/option.pxi":323
 *         return self.missing_values_can_be_empty_container or self.missing_values_can_be_null
 * 
 *     def equals(self, Options other: Options, ignore: Sequence[str]=()) -> bool:             # <<<<<<<<<<<<<<
 *         '''
 *         Compares two Options objects, this is equivalent to
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_7equals(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_7Options_6equals, "\n        Compares two Options objects, this is equivalent to \n        `self == other` except you can pass in a list of field names to ignore\n        ");
static PyMethodDef __pyx_mdef_5cavro_7Options_7equals = {"equals", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7Options_7equals, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_7Options_6equals};
static PyObject *__pyx_pw_5cavro_7Options_7equals(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_5cavro_Options *__pyx_v_other = 0;
  PyObject *__pyx_v_ignore = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("equals (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,&__pyx_n_s_ignore,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_empty_tuple));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 323, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ignore);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 323, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "equals") < 0)) __PYX_ERR(4, 323, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_other = ((struct __pyx_obj_5cavro_Options *)values[0]);
    __pyx_v_ignore = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("equals", 0, 1, 2, __pyx_nargs); __PYX_ERR(4, 323, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Options.equals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_5cavro_Options, 0, "other", 0))) __PYX_ERR(4, 323, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_7Options_6equals(((struct __pyx_obj_5cavro_Options *)__pyx_v_self), __pyx_v_other, __pyx_v_ignore);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_6equals(struct __pyx_obj_5cavro_Options *__pyx_v_self, struct __pyx_obj_5cavro_Options *__pyx_v_other, PyObject *__pyx_v_ignore) {
  PyObject *__pyx_v_field = NULL;
  PyObject *__pyx_v_self_val = NULL;
  PyObject *__pyx_v_other_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("equals", 1);

  /* "src/option.pxi":328
 *         `self == other` except you can pass in a list of field names to ignore
 *         '''
 *         if not ignore:             # <<<<<<<<<<<<<<
 *             return self == other
 *         for field in dataclasses.fields(Options):
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_ignore); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(4, 328, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "src/option.pxi":329
 *         '''
 *         if not ignore:
 *             return self == other             # <<<<<<<<<<<<<<
 *         for field in dataclasses.fields(Options):
 *             if field.name in ignore:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyObject_RichCompare(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other), Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 329, __pyx_L1_error)
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "src/option.pxi":328
 *         `self == other` except you can pass in a list of field names to ignore
 *         '''
 *         if not ignore:             # <<<<<<<<<<<<<<
 *             return self == other
 *         for field in dataclasses.fields(Options):
 */
  }

  /* "src/option.pxi":330
 *         if not ignore:
 *             return self == other
 *         for field in dataclasses.fields(Options):             # <<<<<<<<<<<<<<
 *             if field.name in ignore:
 *                 continue
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_dataclasses); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_ptype_5cavro_Options)};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_5 = __pyx_t_3; __Pyx_INCREF(__pyx_t_5);
    __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 330, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 330, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(4, 330, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 330, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 330, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(4, 330, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 330, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_5);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(4, 330, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/option.pxi":331
 *             return self == other
 *         for field in dataclasses.fields(Options):
 *             if field.name in ignore:             # <<<<<<<<<<<<<<
 *                 continue
 *             self_val = getattr(self, field.name)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_v_ignore, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 331, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "src/option.pxi":332
 *         for field in dataclasses.fields(Options):
 *             if field.name in ignore:
 *                 continue             # <<<<<<<<<<<<<<
 *             self_val = getattr(self, field.name)
 *             other_val = getattr(other, field.name)
 */
      goto __pyx_L4_continue;

      /* "src/option.pxi":331
 *             return self == other
 *         for field in dataclasses.fields(Options):
 *             if field.name in ignore:             # <<<<<<<<<<<<<<
 *                 continue
 *             self_val = getattr(self, field.name)
 */
    }

    /* "src/option.pxi":333
 *             if field.name in ignore:
 *                 continue
 *             self_val = getattr(self, field.name)             # <<<<<<<<<<<<<<
 *             other_val = getattr(other, field.name)
 *             if self_val != other_val:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_GetAttr(((PyObject *)__pyx_v_self), __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_self_val, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/option.pxi":334
 *                 continue
 *             self_val = getattr(self, field.name)
 *             other_val = getattr(other, field.name)             # <<<<<<<<<<<<<<
 *             if self_val != other_val:
 *                 return False
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_GetAttr(((PyObject *)__pyx_v_other), __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_other_val, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/option.pxi":335
 *             self_val = getattr(self, field.name)
 *             other_val = getattr(other, field.name)
 *             if self_val != other_val:             # <<<<<<<<<<<<<<
 *                 return False
 *         return True
 */
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_self_val, __pyx_v_other_val, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 335, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 335, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "src/option.pxi":336
 *             other_val = getattr(other, field.name)
 *             if self_val != other_val:
 *                 return False             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "src/option.pxi":335
 *             self_val = getattr(self, field.name)
 *             other_val = getattr(other, field.name)
 *             if self_val != other_val:             # <<<<<<<<<<<<<<
 *                 return False
 *         return True
 */
    }

    /* "src/option.pxi":330
 *         if not ignore:
 *             return self == other
 *         for field in dataclasses.fields(Options):             # <<<<<<<<<<<<<<
 *             if field.name in ignore:
 *                 continue
 */
    __pyx_L4_continue:;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/option.pxi":337
 *             if self_val != other_val:
 *                 return False
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     def diff(self, Options other: Options, ignore: Sequence[str]=()) -> dict[str, tuple[object, object]]:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "src/option.pxi":323
 *         return self.missing_values_can_be_empty_container or self.missing_values_can_be_null
 * 
 *     def equals(self, Options other: Options, ignore: Sequence[str]=()) -> bool:             # <<<<<<<<<<<<<<
 *         '''
 *         Compares two Options objects, this is equivalent to
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.Options.equals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_self_val);
  __Pyx_XDECREF(__pyx_v_other_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":339
 *         return True
 * 
 *     def diff(self, Options other: Options, ignore: Sequence[str]=()) -> dict[str, tuple[object, object]]:             # <<<<<<<<<<<<<<
 *         '''
 *         Returns a dictionary containing just the fields that are different between two Options objects.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_9diff(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_7Options_8diff, "\n        Returns a dictionary containing just the fields that are different between two Options objects.\n        Any fields whose names are in `ignore` are not included.\n        The return dictionary is of the form: `{field_name: (self_value, other_value)}`\n        ");
static PyMethodDef __pyx_mdef_5cavro_7Options_9diff = {"diff", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7Options_9diff, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_7Options_8diff};
static PyObject *__pyx_pw_5cavro_7Options_9diff(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_5cavro_Options *__pyx_v_other = 0;
  PyObject *__pyx_v_ignore = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("diff (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,&__pyx_n_s_ignore,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_empty_tuple));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 339, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ignore);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 339, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "diff") < 0)) __PYX_ERR(4, 339, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_other = ((struct __pyx_obj_5cavro_Options *)values[0]);
    __pyx_v_ignore = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("diff", 0, 1, 2, __pyx_nargs); __PYX_ERR(4, 339, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Options.diff", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_5cavro_Options, 0, "other", 0))) __PYX_ERR(4, 339, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_7Options_8diff(((struct __pyx_obj_5cavro_Options *)__pyx_v_self), __pyx_v_other, __pyx_v_ignore);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_8diff(struct __pyx_obj_5cavro_Options *__pyx_v_self, struct __pyx_obj_5cavro_Options *__pyx_v_other, PyObject *__pyx_v_ignore) {
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_field = NULL;
  PyObject *__pyx_v_self_val = NULL;
  PyObject *__pyx_v_other_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("diff", 1);

  /* "src/option.pxi":345
 *         The return dictionary is of the form: `{field_name: (self_value, other_value)}`
 *         '''
 *         out = {}             # <<<<<<<<<<<<<<
 *         for field in dataclasses.fields(Options):
 *             if field.name in ignore:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/option.pxi":346
 *         '''
 *         out = {}
 *         for field in dataclasses.fields(Options):             # <<<<<<<<<<<<<<
 *             if field.name in ignore:
 *                 continue
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_dataclasses); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_fields); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, ((PyObject *)__pyx_ptype_5cavro_Options)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3);
    __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 346, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 346, __pyx_L1_error)
          #endif
          if (__pyx_t_5 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(4, 346, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 346, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 346, __pyx_L1_error)
          #endif
          if (__pyx_t_5 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(4, 346, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 346, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(4, 346, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/option.pxi":347
 *         out = {}
 *         for field in dataclasses.fields(Options):
 *             if field.name in ignore:             # <<<<<<<<<<<<<<
 *                 continue
 *             self_val = getattr(self, field.name)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_v_ignore, Py_EQ)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(4, 347, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_7) {

      /* "src/option.pxi":348
 *         for field in dataclasses.fields(Options):
 *             if field.name in ignore:
 *                 continue             # <<<<<<<<<<<<<<
 *             self_val = getattr(self, field.name)
 *             other_val = getattr(other, field.name)
 */
      goto __pyx_L3_continue;

      /* "src/option.pxi":347
 *         out = {}
 *         for field in dataclasses.fields(Options):
 *             if field.name in ignore:             # <<<<<<<<<<<<<<
 *                 continue
 *             self_val = getattr(self, field.name)
 */
    }

    /* "src/option.pxi":349
 *             if field.name in ignore:
 *                 continue
 *             self_val = getattr(self, field.name)             # <<<<<<<<<<<<<<
 *             other_val = getattr(other, field.name)
 *             if self_val != other_val:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_GetAttr(((PyObject *)__pyx_v_self), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_self_val, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/option.pxi":350
 *                 continue
 *             self_val = getattr(self, field.name)
 *             other_val = getattr(other, field.name)             # <<<<<<<<<<<<<<
 *             if self_val != other_val:
 *                 out[field.name] = (self_val, other_val)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_GetAttr(((PyObject *)__pyx_v_other), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_other_val, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/option.pxi":351
 *             self_val = getattr(self, field.name)
 *             other_val = getattr(other, field.name)
 *             if self_val != other_val:             # <<<<<<<<<<<<<<
 *                 out[field.name] = (self_val, other_val)
 *         return out
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_val, __pyx_v_other_val, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 351, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(4, 351, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_7) {

      /* "src/option.pxi":352
 *             other_val = getattr(other, field.name)
 *             if self_val != other_val:
 *                 out[field.name] = (self_val, other_val)             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_self_val);
      __Pyx_GIVEREF(__pyx_v_self_val);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self_val)) __PYX_ERR(4, 352, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_other_val);
      __Pyx_GIVEREF(__pyx_v_other_val);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_other_val)) __PYX_ERR(4, 352, __pyx_L1_error);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely((PyDict_SetItem(__pyx_v_out, __pyx_t_2, __pyx_t_1) < 0))) __PYX_ERR(4, 352, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "src/option.pxi":351
 *             self_val = getattr(self, field.name)
 *             other_val = getattr(other, field.name)
 *             if self_val != other_val:             # <<<<<<<<<<<<<<
 *                 out[field.name] = (self_val, other_val)
 *         return out
 */
    }

    /* "src/option.pxi":346
 *         '''
 *         out = {}
 *         for field in dataclasses.fields(Options):             # <<<<<<<<<<<<<<
 *             if field.name in ignore:
 *                 continue
 */
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "src/option.pxi":353
 *             if self_val != other_val:
 *                 out[field.name] = (self_val, other_val)
 *         return out             # <<<<<<<<<<<<<<
 * 
 * _OptionsWrapper._set_from_dataclass(Options)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "src/option.pxi":339
 *         return True
 * 
 *     def diff(self, Options other: Options, ignore: Sequence[str]=()) -> dict[str, tuple[object, object]]:             # <<<<<<<<<<<<<<
 *         '''
 *         Returns a dictionary containing just the fields that are different between two Options objects.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.Options.diff", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_self_val);
  __Pyx_XDECREF(__pyx_v_other_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __init__(self, *, fingerprint_returns_digest: bint = DATACLASS_PLACEHOLDER_1, canonical_form_repeat_fixed: bint = DATACLASS_PLACEHOLDER_2, canonical_form_repeat_enum: bint = DATACLASS_PLACEHOLDER_3, record_can_encode_dict: bint = DATACLASS_PLACEHOLDER_4, record_values_type_hint: bint = DATACLASS_PLACEHOLDER_5, record_decodes_to_dict: bint = DATACLASS_PLACEHOLDER_6, record_allow_extra_fields: bint = DATACLASS_PLACEHOLDER_7, record_encode_use_defaults: bint = DATACLASS_PLACEHOLDER_8, allow_tuple_notation: bint = DATACLASS_PLACEHOLDER_9, union_decodes_to: UnionDecodeOption = DATACLASS_PLACEHOLDER_10, union_json_encodes_type_name: bint = DATACLASS_PLACEHOLDER_11, allow_primitive_name_collision: bint = DATACLASS_PLACEHOLDER_12, allow_primitive_names_in_namespaces: bint = DATACLASS_PLACEHOLDER_13, named_type_names_must_be_unique: bint = DATACLASS_PLACEHOLDER_14, enum_symbols_must_be_unique: bint = DATACLASS_PLACEHOLDER_15, enforce_enum_symbol_name_rules: bint = DATACLASS_PLACEHOLDER_16, enforce_type_name_rules: bint = DATACLASS_PLACEHOLDER_17, enforce_namespace_name_rules: bint = DATACLASS_PLACEHOLDER_18, record_fields_must_be_unique: bint = DATACLASS_PLACEHOLDER_19, ascii_name_rules: bint = DATACLASS_PLACEHOLDER_20, missing_values_can_be_null: bint = DATACLASS_PLACEHOLDER_21, missing_values_can_be_empty_container: bint = DATACLASS_PLACEHOLDER_22, allow_false_values_for_null: bint = DATACLASS_PLACEHOLDER_23, allow_empty_unions: bint = DATACLASS_PLACEHOLDER_24, allow_nested_unions: bint = DATACLASS_PLACEHOLDER_25, allow_duplicate_union_types: bint = DATACLASS_PLACEHOLDER_26, allow_union_default_any_member: bint = DATACLASS_PLACEHOLDER_27, allow_aliases_to_be_string: bint = DATACLASS_PLACEHOLDER_28, coerce_values_to_int: bint = DATACLASS_PLACEHOLDER_29, coerce_values_to_float: bint = DATACLASS_PLACEHOLDER_30, coerce_int_to_float: bint = DATACLASS_PLACEHOLDER_31, coerce_values_to_boolean: bint = DATACLASS_PLACEHOLDER_32, coerce_values_to_str: bint = DATACLASS_PLACEHOLDER_33, bytes_codec: str = DATACLASS_PLACEHOLDER_34, fixed_codec: str = DATACLASS_PLACEHOLDER_35, null_pad_fixed: bint = DATACLASS_PLACEHOLDER_36, truncate_fixed: bint = DATACLASS_PLACEHOLDER_37, clamp_int_overflow: bint = DATACLASS_PLACEHOLDER_38, clamp_float_overflow: bint = DATACLASS_PLACEHOLDER_39, float_out_of_range_inf: bint = DATACLASS_PLACEHOLDER_40, bytes_default_value_utf8: bint = DATACLASS_PLACEHOLDER_41, string_types_default_unchanged: bint = DATACLASS_PLACEHOLDER_42, decimal_check_exp_overflow: bint = DATACLASS_PLACEHOLDER_43, types_str_to_schema: bint = DATACLASS_PLACEHOLDER_44, logical_types: tuple[LogicalType] = DATACLASS_PLACEHOLDER_45, adapt_record_types: bint = DATACLASS_PLACEHOLDER_46, return_uuid_object: bint = DATACLASS_PLACEHOLDER_47, allow_error_type: bint = DATACLASS_PLACEHOLDER_48, allow_leading_dot_in_names: bint = DATACLASS_PLACEHOLDER_49, naive_dt_assume_utc: bint = DATACLASS_PLACEHOLDER_50, alternate_timestamp_millis_encoding: bint = DATACLASS_PLACEHOLDER_51, date_type_accepts_string: bint = DATACLASS_PLACEHOLDER_52, raise_on_invalid_logical: bint = DATACLASS_PLACEHOLDER_53, inline_namespaces: bint = DATACLASS_PLACEHOLDER_54, expand_types_in_schema: bint = DATACLASS_PLACEHOLDER_55, unicode_errors: str = DATACLASS_PLACEHOLDER_56, container_fill_blocks: bint = DATACLASS_PLACEHOLDER_57, defer_schema_promotion_errors: bint = DATACLASS_PLACEHOLDER_58, invalid_value_includes_record_name: bint = DATACLASS_PLACEHOLDER_59, invalid_value_include_array_index: bint = DATACLASS_PLACEHOLDER_60, invalid_value_include_map_key: bint = DATACLASS_PLACEHOLDER_61, allow_invalid_default_values: bint = DATACLASS_PLACEHOLDER_62, externally_defined_types: MappingProxyType = DATACLASS_PLACEHOLDER_63):             # <<<<<<<<<<<<<<
 *     self.fingerprint_returns_digest = fingerprint_returns_digest
 *     self.canonical_form_repeat_fixed = canonical_form_repeat_fixed
 */

/* Python wrapper */
static int __pyx_pw_5cavro_7Options_11__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_7Options_11__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fingerprint_returns_digest = 0;
  PyObject *__pyx_v_canonical_form_repeat_fixed = 0;
  PyObject *__pyx_v_canonical_form_repeat_enum = 0;
  PyObject *__pyx_v_record_can_encode_dict = 0;
  PyObject *__pyx_v_record_values_type_hint = 0;
  PyObject *__pyx_v_record_decodes_to_dict = 0;
  PyObject *__pyx_v_record_allow_extra_fields = 0;
  PyObject *__pyx_v_record_encode_use_defaults = 0;
  PyObject *__pyx_v_allow_tuple_notation = 0;
  PyObject *__pyx_v_union_decodes_to = 0;
  PyObject *__pyx_v_union_json_encodes_type_name = 0;
  PyObject *__pyx_v_allow_primitive_name_collision = 0;
  PyObject *__pyx_v_allow_primitive_names_in_namespaces = 0;
  PyObject *__pyx_v_named_type_names_must_be_unique = 0;
  PyObject *__pyx_v_enum_symbols_must_be_unique = 0;
  PyObject *__pyx_v_enforce_enum_symbol_name_rules = 0;
  PyObject *__pyx_v_enforce_type_name_rules = 0;
  PyObject *__pyx_v_enforce_namespace_name_rules = 0;
  PyObject *__pyx_v_record_fields_must_be_unique = 0;
  PyObject *__pyx_v_ascii_name_rules = 0;
  PyObject *__pyx_v_missing_values_can_be_null = 0;
  PyObject *__pyx_v_missing_values_can_be_empty_container = 0;
  PyObject *__pyx_v_allow_false_values_for_null = 0;
  PyObject *__pyx_v_allow_empty_unions = 0;
  PyObject *__pyx_v_allow_nested_unions = 0;
  PyObject *__pyx_v_allow_duplicate_union_types = 0;
  PyObject *__pyx_v_allow_union_default_any_member = 0;
  PyObject *__pyx_v_allow_aliases_to_be_string = 0;
  PyObject *__pyx_v_coerce_values_to_int = 0;
  PyObject *__pyx_v_coerce_values_to_float = 0;
  PyObject *__pyx_v_coerce_int_to_float = 0;
  PyObject *__pyx_v_coerce_values_to_boolean = 0;
  PyObject *__pyx_v_coerce_values_to_str = 0;
  PyObject *__pyx_v_bytes_codec = 0;
  PyObject *__pyx_v_fixed_codec = 0;
  PyObject *__pyx_v_null_pad_fixed = 0;
  PyObject *__pyx_v_truncate_fixed = 0;
  PyObject *__pyx_v_clamp_int_overflow = 0;
  PyObject *__pyx_v_clamp_float_overflow = 0;
  PyObject *__pyx_v_float_out_of_range_inf = 0;
  PyObject *__pyx_v_bytes_default_value_utf8 = 0;
  PyObject *__pyx_v_string_types_default_unchanged = 0;
  PyObject *__pyx_v_decimal_check_exp_overflow = 0;
  PyObject *__pyx_v_types_str_to_schema = 0;
  PyObject *__pyx_v_logical_types = 0;
  PyObject *__pyx_v_adapt_record_types = 0;
  PyObject *__pyx_v_return_uuid_object = 0;
  PyObject *__pyx_v_allow_error_type = 0;
  PyObject *__pyx_v_allow_leading_dot_in_names = 0;
  PyObject *__pyx_v_naive_dt_assume_utc = 0;
  PyObject *__pyx_v_alternate_timestamp_millis_encoding = 0;
  PyObject *__pyx_v_date_type_accepts_string = 0;
  PyObject *__pyx_v_raise_on_invalid_logical = 0;
  PyObject *__pyx_v_inline_namespaces = 0;
  PyObject *__pyx_v_expand_types_in_schema = 0;
  PyObject *__pyx_v_unicode_errors = 0;
  PyObject *__pyx_v_container_fill_blocks = 0;
  PyObject *__pyx_v_defer_schema_promotion_errors = 0;
  PyObject *__pyx_v_invalid_value_includes_record_name = 0;
  PyObject *__pyx_v_invalid_value_include_array_index = 0;
  PyObject *__pyx_v_invalid_value_include_map_key = 0;
  PyObject *__pyx_v_allow_invalid_default_values = 0;
  PyObject *__pyx_v_externally_defined_types = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[63] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fingerprint_returns_digest,&__pyx_n_s_canonical_form_repeat_fixed,&__pyx_n_s_canonical_form_repeat_enum,&__pyx_n_s_record_can_encode_dict,&__pyx_n_s_record_values_type_hint,&__pyx_n_s_record_decodes_to_dict,&__pyx_n_s_record_allow_extra_fields,&__pyx_n_s_record_encode_use_defaults,&__pyx_n_s_allow_tuple_notation,&__pyx_n_s_union_decodes_to,&__pyx_n_s_union_json_encodes_type_name,&__pyx_n_s_allow_primitive_name_collision,&__pyx_n_s_allow_primitive_names_in_namespa,&__pyx_n_s_named_type_names_must_be_unique,&__pyx_n_s_enum_symbols_must_be_unique,&__pyx_n_s_enforce_enum_symbol_name_rules,&__pyx_n_s_enforce_type_name_rules,&__pyx_n_s_enforce_namespace_name_rules,&__pyx_n_s_record_fields_must_be_unique,&__pyx_n_s_ascii_name_rules,&__pyx_n_s_missing_values_can_be_null,&__pyx_n_s_missing_values_can_be_empty_cont,&__pyx_n_s_allow_false_values_for_null,&__pyx_n_s_allow_empty_unions,&__pyx_n_s_allow_nested_unions,&__pyx_n_s_allow_duplicate_union_types,&__pyx_n_s_allow_union_default_any_member,&__pyx_n_s_allow_aliases_to_be_string,&__pyx_n_s_coerce_values_to_int,&__pyx_n_s_coerce_values_to_float,&__pyx_n_s_coerce_int_to_float,&__pyx_n_s_coerce_values_to_boolean,&__pyx_n_s_coerce_values_to_str,&__pyx_n_s_bytes_codec,&__pyx_n_s_fixed_codec,&__pyx_n_s_null_pad_fixed,&__pyx_n_s_truncate_fixed,&__pyx_n_s_clamp_int_overflow,&__pyx_n_s_clamp_float_overflow,&__pyx_n_s_float_out_of_range_inf,&__pyx_n_s_bytes_default_value_utf8,&__pyx_n_s_string_types_default_unchanged,&__pyx_n_s_decimal_check_exp_overflow,&__pyx_n_s_types_str_to_schema,&__pyx_n_s_logical_types,&__pyx_n_s_adapt_record_types,&__pyx_n_s_return_uuid_object,&__pyx_n_s_allow_error_type,&__pyx_n_s_allow_leading_dot_in_names,&__pyx_n_s_naive_dt_assume_utc,&__pyx_n_s_alternate_timestamp_millis_encod,&__pyx_n_s_date_type_accepts_string,&__pyx_n_s_raise_on_invalid_logical,&__pyx_n_s_inline_namespaces,&__pyx_n_s_expand_types_in_schema,&__pyx_n_s_unicode_errors,&__pyx_n_s_container_fill_blocks,&__pyx_n_s_defer_schema_promotion_errors,&__pyx_n_s_invalid_value_includes_record_na,&__pyx_n_s_invalid_value_include_array_inde,&__pyx_n_s_invalid_value_include_map_key,&__pyx_n_s_allow_invalid_default_values,&__pyx_n_s_externally_defined_types,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[1] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[2] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[3] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[4] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[5] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[6] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[7] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[8] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[9] = __Pyx_Arg_NewRef_VARARGS(__pyx_k__15);
    values[10] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[11] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[12] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[13] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[14] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[15] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[16] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[17] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[18] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[19] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[20] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[21] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[22] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[23] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[24] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[25] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[26] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[27] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[28] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[29] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[30] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[31] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[32] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[33] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    values[34] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    values[35] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[36] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[37] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[38] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[39] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[40] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[41] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[42] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[43] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[44] = __Pyx_Arg_NewRef_VARARGS(__pyx_k__16);
    values[45] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[46] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[47] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[48] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[49] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[50] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[51] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[52] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[53] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[54] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[55] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_n_u_strict));
    values[56] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[57] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[58] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[59] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[60] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[61] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[62] = __Pyx_Arg_NewRef_VARARGS(__pyx_k__17);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      if (kw_args > 0 && likely(kw_args <= 63)) {
        Py_ssize_t index;
        for (index = 0; index < 63 && kw_args > 0; index++) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
          if (value) { values[index] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, 0, "__init__") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 0)) {
      goto __pyx_L5_argtuple_error;
    } else {
    }
    __pyx_v_fingerprint_returns_digest = values[0];
    __pyx_v_canonical_form_repeat_fixed = values[1];
    __pyx_v_canonical_form_repeat_enum = values[2];
    __pyx_v_record_can_encode_dict = values[3];
    __pyx_v_record_values_type_hint = values[4];
    __pyx_v_record_decodes_to_dict = values[5];
    __pyx_v_record_allow_extra_fields = values[6];
    __pyx_v_record_encode_use_defaults = values[7];
    __pyx_v_allow_tuple_notation = values[8];
    __pyx_v_union_decodes_to = values[9];
    __pyx_v_union_json_encodes_type_name = values[10];
    __pyx_v_allow_primitive_name_collision = values[11];
    __pyx_v_allow_primitive_names_in_namespaces = values[12];
    __pyx_v_named_type_names_must_be_unique = values[13];
    __pyx_v_enum_symbols_must_be_unique = values[14];
    __pyx_v_enforce_enum_symbol_name_rules = values[15];
    __pyx_v_enforce_type_name_rules = values[16];
    __pyx_v_enforce_namespace_name_rules = values[17];
    __pyx_v_record_fields_must_be_unique = values[18];
    __pyx_v_ascii_name_rules = values[19];
    __pyx_v_missing_values_can_be_null = values[20];
    __pyx_v_missing_values_can_be_empty_container = values[21];
    __pyx_v_allow_false_values_for_null = values[22];
    __pyx_v_allow_empty_unions = values[23];
    __pyx_v_allow_nested_unions = values[24];
    __pyx_v_allow_duplicate_union_types = values[25];
    __pyx_v_allow_union_default_any_member = values[26];
    __pyx_v_allow_aliases_to_be_string = values[27];
    __pyx_v_coerce_values_to_int = values[28];
    __pyx_v_coerce_values_to_float = values[29];
    __pyx_v_coerce_int_to_float = values[30];
    __pyx_v_coerce_values_to_boolean = values[31];
    __pyx_v_coerce_values_to_str = values[32];
    __pyx_v_bytes_codec = values[33];
    __pyx_v_fixed_codec = values[34];
    __pyx_v_null_pad_fixed = values[35];
    __pyx_v_truncate_fixed = values[36];
    __pyx_v_clamp_int_overflow = values[37];
    __pyx_v_clamp_float_overflow = values[38];
    __pyx_v_float_out_of_range_inf = values[39];
    __pyx_v_bytes_default_value_utf8 = values[40];
    __pyx_v_string_types_default_unchanged = values[41];
    __pyx_v_decimal_check_exp_overflow = values[42];
    __pyx_v_types_str_to_schema = values[43];
    __pyx_v_logical_types = values[44];
    __pyx_v_adapt_record_types = values[45];
    __pyx_v_return_uuid_object = values[46];
    __pyx_v_allow_error_type = values[47];
    __pyx_v_allow_leading_dot_in_names = values[48];
    __pyx_v_naive_dt_assume_utc = values[49];
    __pyx_v_alternate_timestamp_millis_encoding = values[50];
    __pyx_v_date_type_accepts_string = values[51];
    __pyx_v_raise_on_invalid_logical = values[52];
    __pyx_v_inline_namespaces = values[53];
    __pyx_v_expand_types_in_schema = values[54];
    __pyx_v_unicode_errors = values[55];
    __pyx_v_container_fill_blocks = values[56];
    __pyx_v_defer_schema_promotion_errors = values[57];
    __pyx_v_invalid_value_includes_record_name = values[58];
    __pyx_v_invalid_value_include_array_index = values[59];
    __pyx_v_invalid_value_include_map_key = values[60];
    __pyx_v_allow_invalid_default_values = values[61];
    __pyx_v_externally_defined_types = values[62];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Options.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_7Options_10__init__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self), __pyx_v_fingerprint_returns_digest, __pyx_v_canonical_form_repeat_fixed, __pyx_v_canonical_form_repeat_enum, __pyx_v_record_can_encode_dict, __pyx_v_record_values_type_hint, __pyx_v_record_decodes_to_dict, __pyx_v_record_allow_extra_fields, __pyx_v_record_encode_use_defaults, __pyx_v_allow_tuple_notation, __pyx_v_union_decodes_to, __pyx_v_union_json_encodes_type_name, __pyx_v_allow_primitive_name_collision, __pyx_v_allow_primitive_names_in_namespaces, __pyx_v_named_type_names_must_be_unique, __pyx_v_enum_symbols_must_be_unique, __pyx_v_enforce_enum_symbol_name_rules, __pyx_v_enforce_type_name_rules, __pyx_v_enforce_namespace_name_rules, __pyx_v_record_fields_must_be_unique, __pyx_v_ascii_name_rules, __pyx_v_missing_values_can_be_null, __pyx_v_missing_values_can_be_empty_container, __pyx_v_allow_false_values_for_null, __pyx_v_allow_empty_unions, __pyx_v_allow_nested_unions, __pyx_v_allow_duplicate_union_types, __pyx_v_allow_union_default_any_member, __pyx_v_allow_aliases_to_be_string, __pyx_v_coerce_values_to_int, __pyx_v_coerce_values_to_float, __pyx_v_coerce_int_to_float, __pyx_v_coerce_values_to_boolean, __pyx_v_coerce_values_to_str, __pyx_v_bytes_codec, __pyx_v_fixed_codec, __pyx_v_null_pad_fixed, __pyx_v_truncate_fixed, __pyx_v_clamp_int_overflow, __pyx_v_clamp_float_overflow, __pyx_v_float_out_of_range_inf, __pyx_v_bytes_default_value_utf8, __pyx_v_string_types_default_unchanged, __pyx_v_decimal_check_exp_overflow, __pyx_v_types_str_to_schema, __pyx_v_logical_types, __pyx_v_adapt_record_types, __pyx_v_return_uuid_object, __pyx_v_allow_error_type, __pyx_v_allow_leading_dot_in_names, __pyx_v_naive_dt_assume_utc, __pyx_v_alternate_timestamp_millis_encoding, __pyx_v_date_type_accepts_string, __pyx_v_raise_on_invalid_logical, __pyx_v_inline_namespaces, __pyx_v_expand_types_in_schema, __pyx_v_unicode_errors, __pyx_v_container_fill_blocks, __pyx_v_defer_schema_promotion_errors, __pyx_v_invalid_value_includes_record_name, __pyx_v_invalid_value_include_array_index, __pyx_v_invalid_value_include_map_key, __pyx_v_allow_invalid_default_values, __pyx_v_externally_defined_types);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_7Options_10__init__(struct __pyx_obj_5cavro_Options *__pyx_v_self, PyObject *__pyx_v_fingerprint_returns_digest, PyObject *__pyx_v_canonical_form_repeat_fixed, PyObject *__pyx_v_canonical_form_repeat_enum, PyObject *__pyx_v_record_can_encode_dict, PyObject *__pyx_v_record_values_type_hint, PyObject *__pyx_v_record_decodes_to_dict, PyObject *__pyx_v_record_allow_extra_fields, PyObject *__pyx_v_record_encode_use_defaults, PyObject *__pyx_v_allow_tuple_notation, PyObject *__pyx_v_union_decodes_to, PyObject *__pyx_v_union_json_encodes_type_name, PyObject *__pyx_v_allow_primitive_name_collision, PyObject *__pyx_v_allow_primitive_names_in_namespaces, PyObject *__pyx_v_named_type_names_must_be_unique, PyObject *__pyx_v_enum_symbols_must_be_unique, PyObject *__pyx_v_enforce_enum_symbol_name_rules, PyObject *__pyx_v_enforce_type_name_rules, PyObject *__pyx_v_enforce_namespace_name_rules, PyObject *__pyx_v_record_fields_must_be_unique, PyObject *__pyx_v_ascii_name_rules, PyObject *__pyx_v_missing_values_can_be_null, PyObject *__pyx_v_missing_values_can_be_empty_container, PyObject *__pyx_v_allow_false_values_for_null, PyObject *__pyx_v_allow_empty_unions, PyObject *__pyx_v_allow_nested_unions, PyObject *__pyx_v_allow_duplicate_union_types, PyObject *__pyx_v_allow_union_default_any_member, PyObject *__pyx_v_allow_aliases_to_be_string, PyObject *__pyx_v_coerce_values_to_int, PyObject *__pyx_v_coerce_values_to_float, PyObject *__pyx_v_coerce_int_to_float, PyObject *__pyx_v_coerce_values_to_boolean, PyObject *__pyx_v_coerce_values_to_str, PyObject *__pyx_v_bytes_codec, PyObject *__pyx_v_fixed_codec, PyObject *__pyx_v_null_pad_fixed, PyObject *__pyx_v_truncate_fixed, PyObject *__pyx_v_clamp_int_overflow, PyObject *__pyx_v_clamp_float_overflow, PyObject *__pyx_v_float_out_of_range_inf, PyObject *__pyx_v_bytes_default_value_utf8, PyObject *__pyx_v_string_types_default_unchanged, PyObject *__pyx_v_decimal_check_exp_overflow, PyObject *__pyx_v_types_str_to_schema, PyObject *__pyx_v_logical_types, PyObject *__pyx_v_adapt_record_types, PyObject *__pyx_v_return_uuid_object, PyObject *__pyx_v_allow_error_type, PyObject *__pyx_v_allow_leading_dot_in_names, PyObject *__pyx_v_naive_dt_assume_utc, PyObject *__pyx_v_alternate_timestamp_millis_encoding, PyObject *__pyx_v_date_type_accepts_string, PyObject *__pyx_v_raise_on_invalid_logical, PyObject *__pyx_v_inline_namespaces, PyObject *__pyx_v_expand_types_in_schema, PyObject *__pyx_v_unicode_errors, PyObject *__pyx_v_container_fill_blocks, PyObject *__pyx_v_defer_schema_promotion_errors, PyObject *__pyx_v_invalid_value_includes_record_name, PyObject *__pyx_v_invalid_value_include_array_index, PyObject *__pyx_v_invalid_value_include_map_key, PyObject *__pyx_v_allow_invalid_default_values, PyObject *__pyx_v_externally_defined_types) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "(tree fragment)":2
 * def __init__(self, *, fingerprint_returns_digest: bint = DATACLASS_PLACEHOLDER_1, canonical_form_repeat_fixed: bint = DATACLASS_PLACEHOLDER_2, canonical_form_repeat_enum: bint = DATACLASS_PLACEHOLDER_3, record_can_encode_dict: bint = DATACLASS_PLACEHOLDER_4, record_values_type_hint: bint = DATACLASS_PLACEHOLDER_5, record_decodes_to_dict: bint = DATACLASS_PLACEHOLDER_6, record_allow_extra_fields: bint = DATACLASS_PLACEHOLDER_7, record_encode_use_defaults: bint = DATACLASS_PLACEHOLDER_8, allow_tuple_notation: bint = DATACLASS_PLACEHOLDER_9, union_decodes_to: UnionDecodeOption = DATACLASS_PLACEHOLDER_10, union_json_encodes_type_name: bint = DATACLASS_PLACEHOLDER_11, allow_primitive_name_collision: bint = DATACLASS_PLACEHOLDER_12, allow_primitive_names_in_namespaces: bint = DATACLASS_PLACEHOLDER_13, named_type_names_must_be_unique: bint = DATACLASS_PLACEHOLDER_14, enum_symbols_must_be_unique: bint = DATACLASS_PLACEHOLDER_15, enforce_enum_symbol_name_rules: bint = DATACLASS_PLACEHOLDER_16, enforce_type_name_rules: bint = DATACLASS_PLACEHOLDER_17, enforce_namespace_name_rules: bint = DATACLASS_PLACEHOLDER_18, record_fields_must_be_unique: bint = DATACLASS_PLACEHOLDER_19, ascii_name_rules: bint = DATACLASS_PLACEHOLDER_20, missing_values_can_be_null: bint = DATACLASS_PLACEHOLDER_21, missing_values_can_be_empty_container: bint = DATACLASS_PLACEHOLDER_22, allow_false_values_for_null: bint = DATACLASS_PLACEHOLDER_23, allow_empty_unions: bint = DATACLASS_PLACEHOLDER_24, allow_nested_unions: bint = DATACLASS_PLACEHOLDER_25, allow_duplicate_union_types: bint = DATACLASS_PLACEHOLDER_26, allow_union_default_any_member: bint = DATACLASS_PLACEHOLDER_27, allow_aliases_to_be_string: bint = DATACLASS_PLACEHOLDER_28, coerce_values_to_int: bint = DATACLASS_PLACEHOLDER_29, coerce_values_to_float: bint = DATACLASS_PLACEHOLDER_30, coerce_int_to_float: bint = DATACLASS_PLACEHOLDER_31, coerce_values_to_boolean: bint = DATACLASS_PLACEHOLDER_32, coerce_values_to_str: bint = DATACLASS_PLACEHOLDER_33, bytes_codec: str = DATACLASS_PLACEHOLDER_34, fixed_codec: str = DATACLASS_PLACEHOLDER_35, null_pad_fixed: bint = DATACLASS_PLACEHOLDER_36, truncate_fixed: bint = DATACLASS_PLACEHOLDER_37, clamp_int_overflow: bint = DATACLASS_PLACEHOLDER_38, clamp_float_overflow: bint = DATACLASS_PLACEHOLDER_39, float_out_of_range_inf: bint = DATACLASS_PLACEHOLDER_40, bytes_default_value_utf8: bint = DATACLASS_PLACEHOLDER_41, string_types_default_unchanged: bint = DATACLASS_PLACEHOLDER_42, decimal_check_exp_overflow: bint = DATACLASS_PLACEHOLDER_43, types_str_to_schema: bint = DATACLASS_PLACEHOLDER_44, logical_types: tuple[LogicalType] = DATACLASS_PLACEHOLDER_45, adapt_record_types: bint = DATACLASS_PLACEHOLDER_46, return_uuid_object: bint = DATACLASS_PLACEHOLDER_47, allow_error_type: bint = DATACLASS_PLACEHOLDER_48, allow_leading_dot_in_names: bint = DATACLASS_PLACEHOLDER_49, naive_dt_assume_utc: bint = DATACLASS_PLACEHOLDER_50, alternate_timestamp_millis_encoding: bint = DATACLASS_PLACEHOLDER_51, date_type_accepts_string: bint = DATACLASS_PLACEHOLDER_52, raise_on_invalid_logical: bint = DATACLASS_PLACEHOLDER_53, inline_namespaces: bint = DATACLASS_PLACEHOLDER_54, expand_types_in_schema: bint = DATACLASS_PLACEHOLDER_55, unicode_errors: str = DATACLASS_PLACEHOLDER_56, container_fill_blocks: bint = DATACLASS_PLACEHOLDER_57, defer_schema_promotion_errors: bint = DATACLASS_PLACEHOLDER_58, invalid_value_includes_record_name: bint = DATACLASS_PLACEHOLDER_59, invalid_value_include_array_index: bint = DATACLASS_PLACEHOLDER_60, invalid_value_include_map_key: bint = DATACLASS_PLACEHOLDER_61, allow_invalid_default_values: bint = DATACLASS_PLACEHOLDER_62, externally_defined_types: MappingProxyType = DATACLASS_PLACEHOLDER_63):
 *     self.fingerprint_returns_digest = fingerprint_returns_digest             # <<<<<<<<<<<<<<
 *     self.canonical_form_repeat_fixed = canonical_form_repeat_fixed
 *     self.canonical_form_repeat_enum = canonical_form_repeat_enum
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fingerprint_returns_digest); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_v_self->fingerprint_returns_digest = __pyx_t_1;

  /* "(tree fragment)":3
 * def __init__(self, *, fingerprint_returns_digest: bint = DATACLASS_PLACEHOLDER_1, canonical_form_repeat_fixed: bint = DATACLASS_PLACEHOLDER_2, canonical_form_repeat_enum: bint = DATACLASS_PLACEHOLDER_3, record_can_encode_dict: bint = DATACLASS_PLACEHOLDER_4, record_values_type_hint: bint = DATACLASS_PLACEHOLDER_5, record_decodes_to_dict: bint = DATACLASS_PLACEHOLDER_6, record_allow_extra_fields: bint = DATACLASS_PLACEHOLDER_7, record_encode_use_defaults: bint = DATACLASS_PLACEHOLDER_8, allow_tuple_notation: bint = DATACLASS_PLACEHOLDER_9, union_decodes_to: UnionDecodeOption = DATACLASS_PLACEHOLDER_10, union_json_encodes_type_name: bint = DATACLASS_PLACEHOLDER_11, allow_primitive_name_collision: bint = DATACLASS_PLACEHOLDER_12, allow_primitive_names_in_namespaces: bint = DATACLASS_PLACEHOLDER_13, named_type_names_must_be_unique: bint = DATACLASS_PLACEHOLDER_14, enum_symbols_must_be_unique: bint = DATACLASS_PLACEHOLDER_15, enforce_enum_symbol_name_rules: bint = DATACLASS_PLACEHOLDER_16, enforce_type_name_rules: bint = DATACLASS_PLACEHOLDER_17, enforce_namespace_name_rules: bint = DATACLASS_PLACEHOLDER_18, record_fields_must_be_unique: bint = DATACLASS_PLACEHOLDER_19, ascii_name_rules: bint = DATACLASS_PLACEHOLDER_20, missing_values_can_be_null: bint = DATACLASS_PLACEHOLDER_21, missing_values_can_be_empty_container: bint = DATACLASS_PLACEHOLDER_22, allow_false_values_for_null: bint = DATACLASS_PLACEHOLDER_23, allow_empty_unions: bint = DATACLASS_PLACEHOLDER_24, allow_nested_unions: bint = DATACLASS_PLACEHOLDER_25, allow_duplicate_union_types: bint = DATACLASS_PLACEHOLDER_26, allow_union_default_any_member: bint = DATACLASS_PLACEHOLDER_27, allow_aliases_to_be_string: bint = DATACLASS_PLACEHOLDER_28, coerce_values_to_int: bint = DATACLASS_PLACEHOLDER_29, coerce_values_to_float: bint = DATACLASS_PLACEHOLDER_30, coerce_int_to_float: bint = DATACLASS_PLACEHOLDER_31, coerce_values_to_boolean: bint = DATACLASS_PLACEHOLDER_32, coerce_values_to_str: bint = DATACLASS_PLACEHOLDER_33, bytes_codec: str = DATACLASS_PLACEHOLDER_34, fixed_codec: str = DATACLASS_PLACEHOLDER_35, null_pad_fixed: bint = DATACLASS_PLACEHOLDER_36, truncate_fixed: bint = DATACLASS_PLACEHOLDER_37, clamp_int_overflow: bint = DATACLASS_PLACEHOLDER_38, clamp_float_overflow: bint = DATACLASS_PLACEHOLDER_39, float_out_of_range_inf: bint = DATACLASS_PLACEHOLDER_40, bytes_default_value_utf8: bint = DATACLASS_PLACEHOLDER_41, string_types_default_unchanged: bint = DATACLASS_PLACEHOLDER_42, decimal_check_exp_overflow: bint = DATACLASS_PLACEHOLDER_43, types_str_to_schema: bint = DATACLASS_PLACEHOLDER_44, logical_types: tuple[LogicalType] = DATACLASS_PLACEHOLDER_45, adapt_record_types: bint = DATACLASS_PLACEHOLDER_46, return_uuid_object: bint = DATACLASS_PLACEHOLDER_47, allow_error_type: bint = DATACLASS_PLACEHOLDER_48, allow_leading_dot_in_names: bint = DATACLASS_PLACEHOLDER_49, naive_dt_assume_utc: bint = DATACLASS_PLACEHOLDER_50, alternate_timestamp_millis_encoding: bint = DATACLASS_PLACEHOLDER_51, date_type_accepts_string: bint = DATACLASS_PLACEHOLDER_52, raise_on_invalid_logical: bint = DATACLASS_PLACEHOLDER_53, inline_namespaces: bint = DATACLASS_PLACEHOLDER_54, expand_types_in_schema: bint = DATACLASS_PLACEHOLDER_55, unicode_errors: str = DATACLASS_PLACEHOLDER_56, container_fill_blocks: bint = DATACLASS_PLACEHOLDER_57, defer_schema_promotion_errors: bint = DATACLASS_PLACEHOLDER_58, invalid_value_includes_record_name: bint = DATACLASS_PLACEHOLDER_59, invalid_value_include_array_index: bint = DATACLASS_PLACEHOLDER_60, invalid_value_include_map_key: bint = DATACLASS_PLACEHOLDER_61, allow_invalid_default_values: bint = DATACLASS_PLACEHOLDER_62, externally_defined_types: MappingProxyType = DATACLASS_PLACEHOLDER_63):
 *     self.fingerprint_returns_digest = fingerprint_returns_digest
 *     self.canonical_form_repeat_fixed = canonical_form_repeat_fixed             # <<<<<<<<<<<<<<
 *     self.canonical_form_repeat_enum = canonical_form_repeat_enum
 *     self.record_can_encode_dict = record_can_encode_dict
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_canonical_form_repeat_fixed); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L1_error)
  __pyx_v_self->canonical_form_repeat_fixed = __pyx_t_1;

  /* "(tree fragment)":4
 *     self.fingerprint_returns_digest = fingerprint_returns_digest
 *     self.canonical_form_repeat_fixed = canonical_form_repeat_fixed
 *     self.canonical_form_repeat_enum = canonical_form_repeat_enum             # <<<<<<<<<<<<<<
 *     self.record_can_encode_dict = record_can_encode_dict
 *     self.record_values_type_hint = record_values_type_hint
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_canonical_form_repeat_enum); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 4, __pyx_L1_error)
  __pyx_v_self->canonical_form_repeat_enum = __pyx_t_1;

  /* "(tree fragment)":5
 *     self.canonical_form_repeat_fixed = canonical_form_repeat_fixed
 *     self.canonical_form_repeat_enum = canonical_form_repeat_enum
 *     self.record_can_encode_dict = record_can_encode_dict             # <<<<<<<<<<<<<<
 *     self.record_values_type_hint = record_values_type_hint
 *     self.record_decodes_to_dict = record_decodes_to_dict
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_record_can_encode_dict); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 5, __pyx_L1_error)
  __pyx_v_self->record_can_encode_dict = __pyx_t_1;

  /* "(tree fragment)":6
 *     self.canonical_form_repeat_enum = canonical_form_repeat_enum
 *     self.record_can_encode_dict = record_can_encode_dict
 *     self.record_values_type_hint = record_values_type_hint             # <<<<<<<<<<<<<<
 *     self.record_decodes_to_dict = record_decodes_to_dict
 *     self.record_allow_extra_fields = record_allow_extra_fields
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_record_values_type_hint); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 6, __pyx_L1_error)
  __pyx_v_self->record_values_type_hint = __pyx_t_1;

  /* "(tree fragment)":7
 *     self.record_can_encode_dict = record_can_encode_dict
 *     self.record_values_type_hint = record_values_type_hint
 *     self.record_decodes_to_dict = record_decodes_to_dict             # <<<<<<<<<<<<<<
 *     self.record_allow_extra_fields = record_allow_extra_fields
 *     self.record_encode_use_defaults = record_encode_use_defaults
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_record_decodes_to_dict); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 7, __pyx_L1_error)
  __pyx_v_self->record_decodes_to_dict = __pyx_t_1;

  /* "(tree fragment)":8
 *     self.record_values_type_hint = record_values_type_hint
 *     self.record_decodes_to_dict = record_decodes_to_dict
 *     self.record_allow_extra_fields = record_allow_extra_fields             # <<<<<<<<<<<<<<
 *     self.record_encode_use_defaults = record_encode_use_defaults
 *     self.allow_tuple_notation = allow_tuple_notation
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_record_allow_extra_fields); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 8, __pyx_L1_error)
  __pyx_v_self->record_allow_extra_fields = __pyx_t_1;

  /* "(tree fragment)":9
 *     self.record_decodes_to_dict = record_decodes_to_dict
 *     self.record_allow_extra_fields = record_allow_extra_fields
 *     self.record_encode_use_defaults = record_encode_use_defaults             # <<<<<<<<<<<<<<
 *     self.allow_tuple_notation = allow_tuple_notation
 *     self.union_decodes_to = union_decodes_to
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_record_encode_use_defaults); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 9, __pyx_L1_error)
  __pyx_v_self->record_encode_use_defaults = __pyx_t_1;

  /* "(tree fragment)":10
 *     self.record_allow_extra_fields = record_allow_extra_fields
 *     self.record_encode_use_defaults = record_encode_use_defaults
 *     self.allow_tuple_notation = allow_tuple_notation             # <<<<<<<<<<<<<<
 *     self.union_decodes_to = union_decodes_to
 *     self.union_json_encodes_type_name = union_json_encodes_type_name
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_allow_tuple_notation); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 10, __pyx_L1_error)
  __pyx_v_self->allow_tuple_notation = __pyx_t_1;

  /* "(tree fragment)":11
 *     self.record_encode_use_defaults = record_encode_use_defaults
 *     self.allow_tuple_notation = allow_tuple_notation
 *     self.union_decodes_to = union_decodes_to             # <<<<<<<<<<<<<<
 *     self.union_json_encodes_type_name = union_json_encodes_type_name
 *     self.allow_primitive_name_collision = allow_primitive_name_collision
 */
  __Pyx_INCREF(__pyx_v_union_decodes_to);
  __Pyx_GIVEREF(__pyx_v_union_decodes_to);
  __Pyx_GOTREF(__pyx_v_self->union_decodes_to);
  __Pyx_DECREF(__pyx_v_self->union_decodes_to);
  __pyx_v_self->union_decodes_to = __pyx_v_union_decodes_to;

  /* "(tree fragment)":12
 *     self.allow_tuple_notation = allow_tuple_notation
 *     self.union_decodes_to = union_decodes_to
 *     self.union_json_encodes_type_name = union_json_encodes_type_name             # <<<<<<<<<<<<<<
 *     self.allow_primitive_name_collision = allow_primitive_name_collision
 *     self.allow_primitive_names_in_namespaces = allow_primitive_names_in_namespaces
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_union_json_encodes_type_name); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v_self->union_json_encodes_type_name = __pyx_t_1;

  /* "(tree fragment)":13
 *     self.union_decodes_to = union_decodes_to
 *     self.union_json_encodes_type_name = union_json_encodes_type_name
 *     self.allow_primitive_name_collision = allow_primitive_name_collision             # <<<<<<<<<<<<<<
 *     self.allow_primitive_names_in_namespaces = allow_primitive_names_in_namespaces
 *     self.named_type_names_must_be_unique = named_type_names_must_be_unique
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_allow_primitive_name_collision); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_v_self->allow_primitive_name_collision = __pyx_t_1;

  /* "(tree fragment)":14
 *     self.union_json_encodes_type_name = union_json_encodes_type_name
 *     self.allow_primitive_name_collision = allow_primitive_name_collision
 *     self.allow_primitive_names_in_namespaces = allow_primitive_names_in_namespaces             # <<<<<<<<<<<<<<
 *     self.named_type_names_must_be_unique = named_type_names_must_be_unique
 *     self.enum_symbols_must_be_unique = enum_symbols_must_be_unique
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_allow_primitive_names_in_namespaces); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 14, __pyx_L1_error)
  __pyx_v_self->allow_primitive_names_in_namespaces = __pyx_t_1;

  /* "(tree fragment)":15
 *     self.allow_primitive_name_collision = allow_primitive_name_collision
 *     self.allow_primitive_names_in_namespaces = allow_primitive_names_in_namespaces
 *     self.named_type_names_must_be_unique = named_type_names_must_be_unique             # <<<<<<<<<<<<<<
 *     self.enum_symbols_must_be_unique = enum_symbols_must_be_unique
 *     self.enforce_enum_symbol_name_rules = enforce_enum_symbol_name_rules
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_named_type_names_must_be_unique); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 15, __pyx_L1_error)
  __pyx_v_self->named_type_names_must_be_unique = __pyx_t_1;

  /* "(tree fragment)":16
 *     self.allow_primitive_names_in_namespaces = allow_primitive_names_in_namespaces
 *     self.named_type_names_must_be_unique = named_type_names_must_be_unique
 *     self.enum_symbols_must_be_unique = enum_symbols_must_be_unique             # <<<<<<<<<<<<<<
 *     self.enforce_enum_symbol_name_rules = enforce_enum_symbol_name_rules
 *     self.enforce_type_name_rules = enforce_type_name_rules
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_enum_symbols_must_be_unique); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L1_error)
  __pyx_v_self->enum_symbols_must_be_unique = __pyx_t_1;

  /* "(tree fragment)":17
 *     self.named_type_names_must_be_unique = named_type_names_must_be_unique
 *     self.enum_symbols_must_be_unique = enum_symbols_must_be_unique
 *     self.enforce_enum_symbol_name_rules = enforce_enum_symbol_name_rules             # <<<<<<<<<<<<<<
 *     self.enforce_type_name_rules = enforce_type_name_rules
 *     self.enforce_namespace_name_rules = enforce_namespace_name_rules
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_enforce_enum_symbol_name_rules); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_v_self->enforce_enum_symbol_name_rules = __pyx_t_1;

  /* "(tree fragment)":18
 *     self.enum_symbols_must_be_unique = enum_symbols_must_be_unique
 *     self.enforce_enum_symbol_name_rules = enforce_enum_symbol_name_rules
 *     self.enforce_type_name_rules = enforce_type_name_rules             # <<<<<<<<<<<<<<
 *     self.enforce_namespace_name_rules = enforce_namespace_name_rules
 *     self.record_fields_must_be_unique = record_fields_must_be_unique
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_enforce_type_name_rules); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 18, __pyx_L1_error)
  __pyx_v_self->enforce_type_name_rules = __pyx_t_1;

  /* "(tree fragment)":19
 *     self.enforce_enum_symbol_name_rules = enforce_enum_symbol_name_rules
 *     self.enforce_type_name_rules = enforce_type_name_rules
 *     self.enforce_namespace_name_rules = enforce_namespace_name_rules             # <<<<<<<<<<<<<<
 *     self.record_fields_must_be_unique = record_fields_must_be_unique
 *     self.ascii_name_rules = ascii_name_rules
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_enforce_namespace_name_rules); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 19, __pyx_L1_error)
  __pyx_v_self->enforce_namespace_name_rules = __pyx_t_1;

  /* "(tree fragment)":20
 *     self.enforce_type_name_rules = enforce_type_name_rules
 *     self.enforce_namespace_name_rules = enforce_namespace_name_rules
 *     self.record_fields_must_be_unique = record_fields_must_be_unique             # <<<<<<<<<<<<<<
 *     self.ascii_name_rules = ascii_name_rules
 *     self.missing_values_can_be_null = missing_values_can_be_null
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_record_fields_must_be_unique); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 20, __pyx_L1_error)
  __pyx_v_self->record_fields_must_be_unique = __pyx_t_1;

  /* "(tree fragment)":21
 *     self.enforce_namespace_name_rules = enforce_namespace_name_rules
 *     self.record_fields_must_be_unique = record_fields_must_be_unique
 *     self.ascii_name_rules = ascii_name_rules             # <<<<<<<<<<<<<<
 *     self.missing_values_can_be_null = missing_values_can_be_null
 *     self.missing_values_can_be_empty_container = missing_values_can_be_empty_container
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_ascii_name_rules); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 21, __pyx_L1_error)
  __pyx_v_self->ascii_name_rules = __pyx_t_1;

  /* "(tree fragment)":22
 *     self.record_fields_must_be_unique = record_fields_must_be_unique
 *     self.ascii_name_rules = ascii_name_rules
 *     self.missing_values_can_be_null = missing_values_can_be_null             # <<<<<<<<<<<<<<
 *     self.missing_values_can_be_empty_container = missing_values_can_be_empty_container
 *     self.allow_false_values_for_null = allow_false_values_for_null
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_missing_values_can_be_null); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 22, __pyx_L1_error)
  __pyx_v_self->missing_values_can_be_null = __pyx_t_1;

  /* "(tree fragment)":23
 *     self.ascii_name_rules = ascii_name_rules
 *     self.missing_values_can_be_null = missing_values_can_be_null
 *     self.missing_values_can_be_empty_container = missing_values_can_be_empty_container             # <<<<<<<<<<<<<<
 *     self.allow_false_values_for_null = allow_false_values_for_null
 *     self.allow_empty_unions = allow_empty_unions
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_missing_values_can_be_empty_container); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 23, __pyx_L1_error)
  __pyx_v_self->missing_values_can_be_empty_container = __pyx_t_1;

  /* "(tree fragment)":24
 *     self.missing_values_can_be_null = missing_values_can_be_null
 *     self.missing_values_can_be_empty_container = missing_values_can_be_empty_container
 *     self.allow_false_values_for_null = allow_false_values_for_null             # <<<<<<<<<<<<<<
 *     self.allow_empty_unions = allow_empty_unions
 *     self.allow_nested_unions = allow_nested_unions
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_allow_false_values_for_null); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 24, __pyx_L1_error)
  __pyx_v_self->allow_false_values_for_null = __pyx_t_1;

  /* "(tree fragment)":25
 *     self.missing_values_can_be_empty_container = missing_values_can_be_empty_container
 *     self.allow_false_values_for_null = allow_false_values_for_null
 *     self.allow_empty_unions = allow_empty_unions             # <<<<<<<<<<<<<<
 *     self.allow_nested_unions = allow_nested_unions
 *     self.allow_duplicate_union_types = allow_duplicate_union_types
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_allow_empty_unions); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 25, __pyx_L1_error)
  __pyx_v_self->allow_empty_unions = __pyx_t_1;

  /* "(tree fragment)":26
 *     self.allow_false_values_for_null = allow_false_values_for_null
 *     self.allow_empty_unions = allow_empty_unions
 *     self.allow_nested_unions = allow_nested_unions             # <<<<<<<<<<<<<<
 *     self.allow_duplicate_union_types = allow_duplicate_union_types
 *     self.allow_union_default_any_member = allow_union_default_any_member
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_allow_nested_unions); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 26, __pyx_L1_error)
  __pyx_v_self->allow_nested_unions = __pyx_t_1;

  /* "(tree fragment)":27
 *     self.allow_empty_unions = allow_empty_unions
 *     self.allow_nested_unions = allow_nested_unions
 *     self.allow_duplicate_union_types = allow_duplicate_union_types             # <<<<<<<<<<<<<<
 *     self.allow_union_default_any_member = allow_union_default_any_member
 *     self.allow_aliases_to_be_string = allow_aliases_to_be_string
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_allow_duplicate_union_types); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 27, __pyx_L1_error)
  __pyx_v_self->allow_duplicate_union_types = __pyx_t_1;

  /* "(tree fragment)":28
 *     self.allow_nested_unions = allow_nested_unions
 *     self.allow_duplicate_union_types = allow_duplicate_union_types
 *     self.allow_union_default_any_member = allow_union_default_any_member             # <<<<<<<<<<<<<<
 *     self.allow_aliases_to_be_string = allow_aliases_to_be_string
 *     self.coerce_values_to_int = coerce_values_to_int
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_allow_union_default_any_member); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 28, __pyx_L1_error)
  __pyx_v_self->allow_union_default_any_member = __pyx_t_1;

  /* "(tree fragment)":29
 *     self.allow_duplicate_union_types = allow_duplicate_union_types
 *     self.allow_union_default_any_member = allow_union_default_any_member
 *     self.allow_aliases_to_be_string = allow_aliases_to_be_string             # <<<<<<<<<<<<<<
 *     self.coerce_values_to_int = coerce_values_to_int
 *     self.coerce_values_to_float = coerce_values_to_float
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_allow_aliases_to_be_string); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 29, __pyx_L1_error)
  __pyx_v_self->allow_aliases_to_be_string = __pyx_t_1;

  /* "(tree fragment)":30
 *     self.allow_union_default_any_member = allow_union_default_any_member
 *     self.allow_aliases_to_be_string = allow_aliases_to_be_string
 *     self.coerce_values_to_int = coerce_values_to_int             # <<<<<<<<<<<<<<
 *     self.coerce_values_to_float = coerce_values_to_float
 *     self.coerce_int_to_float = coerce_int_to_float
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_coerce_values_to_int); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 30, __pyx_L1_error)
  __pyx_v_self->coerce_values_to_int = __pyx_t_1;

  /* "(tree fragment)":31
 *     self.allow_aliases_to_be_string = allow_aliases_to_be_string
 *     self.coerce_values_to_int = coerce_values_to_int
 *     self.coerce_values_to_float = coerce_values_to_float             # <<<<<<<<<<<<<<
 *     self.coerce_int_to_float = coerce_int_to_float
 *     self.coerce_values_to_boolean = coerce_values_to_boolean
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_coerce_values_to_float); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 31, __pyx_L1_error)
  __pyx_v_self->coerce_values_to_float = __pyx_t_1;

  /* "(tree fragment)":32
 *     self.coerce_values_to_int = coerce_values_to_int
 *     self.coerce_values_to_float = coerce_values_to_float
 *     self.coerce_int_to_float = coerce_int_to_float             # <<<<<<<<<<<<<<
 *     self.coerce_values_to_boolean = coerce_values_to_boolean
 *     self.coerce_values_to_str = coerce_values_to_str
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_coerce_int_to_float); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 32, __pyx_L1_error)
  __pyx_v_self->coerce_int_to_float = __pyx_t_1;

  /* "(tree fragment)":33
 *     self.coerce_values_to_float = coerce_values_to_float
 *     self.coerce_int_to_float = coerce_int_to_float
 *     self.coerce_values_to_boolean = coerce_values_to_boolean             # <<<<<<<<<<<<<<
 *     self.coerce_values_to_str = coerce_values_to_str
 *     self.bytes_codec = bytes_codec
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_coerce_values_to_boolean); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 33, __pyx_L1_error)
  __pyx_v_self->coerce_values_to_boolean = __pyx_t_1;

  /* "(tree fragment)":34
 *     self.coerce_int_to_float = coerce_int_to_float
 *     self.coerce_values_to_boolean = coerce_values_to_boolean
 *     self.coerce_values_to_str = coerce_values_to_str             # <<<<<<<<<<<<<<
 *     self.bytes_codec = bytes_codec
 *     self.fixed_codec = fixed_codec
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_coerce_values_to_str); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 34, __pyx_L1_error)
  __pyx_v_self->coerce_values_to_str = __pyx_t_1;

  /* "(tree fragment)":35
 *     self.coerce_values_to_boolean = coerce_values_to_boolean
 *     self.coerce_values_to_str = coerce_values_to_str
 *     self.bytes_codec = bytes_codec             # <<<<<<<<<<<<<<
 *     self.fixed_codec = fixed_codec
 *     self.null_pad_fixed = null_pad_fixed
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_bytes_codec))||((__pyx_v_bytes_codec) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_bytes_codec))) __PYX_ERR(1, 35, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_bytes_codec;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->bytes_codec);
  __Pyx_DECREF(__pyx_v_self->bytes_codec);
  __pyx_v_self->bytes_codec = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":36
 *     self.coerce_values_to_str = coerce_values_to_str
 *     self.bytes_codec = bytes_codec
 *     self.fixed_codec = fixed_codec             # <<<<<<<<<<<<<<
 *     self.null_pad_fixed = null_pad_fixed
 *     self.truncate_fixed = truncate_fixed
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_fixed_codec))||((__pyx_v_fixed_codec) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_fixed_codec))) __PYX_ERR(1, 36, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_fixed_codec;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->fixed_codec);
  __Pyx_DECREF(__pyx_v_self->fixed_codec);
  __pyx_v_self->fixed_codec = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":37
 *     self.bytes_codec = bytes_codec
 *     self.fixed_codec = fixed_codec
 *     self.null_pad_fixed = null_pad_fixed             # <<<<<<<<<<<<<<
 *     self.truncate_fixed = truncate_fixed
 *     self.clamp_int_overflow = clamp_int_overflow
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_null_pad_fixed); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 37, __pyx_L1_error)
  __pyx_v_self->null_pad_fixed = __pyx_t_1;

  /* "(tree fragment)":38
 *     self.fixed_codec = fixed_codec
 *     self.null_pad_fixed = null_pad_fixed
 *     self.truncate_fixed = truncate_fixed             # <<<<<<<<<<<<<<
 *     self.clamp_int_overflow = clamp_int_overflow
 *     self.clamp_float_overflow = clamp_float_overflow
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_truncate_fixed); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 38, __pyx_L1_error)
  __pyx_v_self->truncate_fixed = __pyx_t_1;

  /* "(tree fragment)":39
 *     self.null_pad_fixed = null_pad_fixed
 *     self.truncate_fixed = truncate_fixed
 *     self.clamp_int_overflow = clamp_int_overflow             # <<<<<<<<<<<<<<
 *     self.clamp_float_overflow = clamp_float_overflow
 *     self.float_out_of_range_inf = float_out_of_range_inf
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_clamp_int_overflow); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L1_error)
  __pyx_v_self->clamp_int_overflow = __pyx_t_1;

  /* "(tree fragment)":40
 *     self.truncate_fixed = truncate_fixed
 *     self.clamp_int_overflow = clamp_int_overflow
 *     self.clamp_float_overflow = clamp_float_overflow             # <<<<<<<<<<<<<<
 *     self.float_out_of_range_inf = float_out_of_range_inf
 *     self.bytes_default_value_utf8 = bytes_default_value_utf8
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_clamp_float_overflow); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 40, __pyx_L1_error)
  __pyx_v_self->clamp_float_overflow = __pyx_t_1;

  /* "(tree fragment)":41
 *     self.clamp_int_overflow = clamp_int_overflow
 *     self.clamp_float_overflow = clamp_float_overflow
 *     self.float_out_of_range_inf = float_out_of_range_inf             # <<<<<<<<<<<<<<
 *     self.bytes_default_value_utf8 = bytes_default_value_utf8
 *     self.string_types_default_unchanged = string_types_default_unchanged
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_float_out_of_range_inf); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 41, __pyx_L1_error)
  __pyx_v_self->float_out_of_range_inf = __pyx_t_1;

  /* "(tree fragment)":42
 *     self.clamp_float_overflow = clamp_float_overflow
 *     self.float_out_of_range_inf = float_out_of_range_inf
 *     self.bytes_default_value_utf8 = bytes_default_value_utf8             # <<<<<<<<<<<<<<
 *     self.string_types_default_unchanged = string_types_default_unchanged
 *     self.decimal_check_exp_overflow = decimal_check_exp_overflow
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_bytes_default_value_utf8); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 42, __pyx_L1_error)
  __pyx_v_self->bytes_default_value_utf8 = __pyx_t_1;

  /* "(tree fragment)":43
 *     self.float_out_of_range_inf = float_out_of_range_inf
 *     self.bytes_default_value_utf8 = bytes_default_value_utf8
 *     self.string_types_default_unchanged = string_types_default_unchanged             # <<<<<<<<<<<<<<
 *     self.decimal_check_exp_overflow = decimal_check_exp_overflow
 *     self.types_str_to_schema = types_str_to_schema
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_string_types_default_unchanged); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 43, __pyx_L1_error)
  __pyx_v_self->string_types_default_unchanged = __pyx_t_1;

  /* "(tree fragment)":44
 *     self.bytes_default_value_utf8 = bytes_default_value_utf8
 *     self.string_types_default_unchanged = string_types_default_unchanged
 *     self.decimal_check_exp_overflow = decimal_check_exp_overflow             # <<<<<<<<<<<<<<
 *     self.types_str_to_schema = types_str_to_schema
 *     self.logical_types = logical_types
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_decimal_check_exp_overflow); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 44, __pyx_L1_error)
  __pyx_v_self->decimal_check_exp_overflow = __pyx_t_1;

  /* "(tree fragment)":45
 *     self.string_types_default_unchanged = string_types_default_unchanged
 *     self.decimal_check_exp_overflow = decimal_check_exp_overflow
 *     self.types_str_to_schema = types_str_to_schema             # <<<<<<<<<<<<<<
 *     self.logical_types = logical_types
 *     self.adapt_record_types = adapt_record_types
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_types_str_to_schema); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 45, __pyx_L1_error)
  __pyx_v_self->types_str_to_schema = __pyx_t_1;

  /* "(tree fragment)":46
 *     self.decimal_check_exp_overflow = decimal_check_exp_overflow
 *     self.types_str_to_schema = types_str_to_schema
 *     self.logical_types = logical_types             # <<<<<<<<<<<<<<
 *     self.adapt_record_types = adapt_record_types
 *     self.return_uuid_object = return_uuid_object
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v_logical_types))||((__pyx_v_logical_types) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v_logical_types))) __PYX_ERR(1, 46, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_logical_types;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->logical_types);
  __Pyx_DECREF(__pyx_v_self->logical_types);
  __pyx_v_self->logical_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":47
 *     self.types_str_to_schema = types_str_to_schema
 *     self.logical_types = logical_types
 *     self.adapt_record_types = adapt_record_types             # <<<<<<<<<<<<<<
 *     self.return_uuid_object = return_uuid_object
 *     self.allow_error_type = allow_error_type
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_adapt_record_types); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 47, __pyx_L1_error)
  __pyx_v_self->adapt_record_types = __pyx_t_1;

  /* "(tree fragment)":48
 *     self.logical_types = logical_types
 *     self.adapt_record_types = adapt_record_types
 *     self.return_uuid_object = return_uuid_object             # <<<<<<<<<<<<<<
 *     self.allow_error_type = allow_error_type
 *     self.allow_leading_dot_in_names = allow_leading_dot_in_names
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_uuid_object); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 48, __pyx_L1_error)
  __pyx_v_self->return_uuid_object = __pyx_t_1;

  /* "(tree fragment)":49
 *     self.adapt_record_types = adapt_record_types
 *     self.return_uuid_object = return_uuid_object
 *     self.allow_error_type = allow_error_type             # <<<<<<<<<<<<<<
 *     self.allow_leading_dot_in_names = allow_leading_dot_in_names
 *     self.naive_dt_assume_utc = naive_dt_assume_utc
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_allow_error_type); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 49, __pyx_L1_error)
  __pyx_v_self->allow_error_type = __pyx_t_1;

  /* "(tree fragment)":50
 *     self.return_uuid_object = return_uuid_object
 *     self.allow_error_type = allow_error_type
 *     self.allow_leading_dot_in_names = allow_leading_dot_in_names             # <<<<<<<<<<<<<<
 *     self.naive_dt_assume_utc = naive_dt_assume_utc
 *     self.alternate_timestamp_millis_encoding = alternate_timestamp_millis_encoding
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_allow_leading_dot_in_names); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 50, __pyx_L1_error)
  __pyx_v_self->allow_leading_dot_in_names = __pyx_t_1;

  /* "(tree fragment)":51
 *     self.allow_error_type = allow_error_type
 *     self.allow_leading_dot_in_names = allow_leading_dot_in_names
 *     self.naive_dt_assume_utc = naive_dt_assume_utc             # <<<<<<<<<<<<<<
 *     self.alternate_timestamp_millis_encoding = alternate_timestamp_millis_encoding
 *     self.date_type_accepts_string = date_type_accepts_string
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_naive_dt_assume_utc); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 51, __pyx_L1_error)
  __pyx_v_self->naive_dt_assume_utc = __pyx_t_1;

  /* "(tree fragment)":52
 *     self.allow_leading_dot_in_names = allow_leading_dot_in_names
 *     self.naive_dt_assume_utc = naive_dt_assume_utc
 *     self.alternate_timestamp_millis_encoding = alternate_timestamp_millis_encoding             # <<<<<<<<<<<<<<
 *     self.date_type_accepts_string = date_type_accepts_string
 *     self.raise_on_invalid_logical = raise_on_invalid_logical
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_alternate_timestamp_millis_encoding); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 52, __pyx_L1_error)
  __pyx_v_self->alternate_timestamp_millis_encoding = __pyx_t_1;

  /* "(tree fragment)":53
 *     self.naive_dt_assume_utc = naive_dt_assume_utc
 *     self.alternate_timestamp_millis_encoding = alternate_timestamp_millis_encoding
 *     self.date_type_accepts_string = date_type_accepts_string             # <<<<<<<<<<<<<<
 *     self.raise_on_invalid_logical = raise_on_invalid_logical
 *     self.inline_namespaces = inline_namespaces
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_date_type_accepts_string); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 53, __pyx_L1_error)
  __pyx_v_self->date_type_accepts_string = __pyx_t_1;

  /* "(tree fragment)":54
 *     self.alternate_timestamp_millis_encoding = alternate_timestamp_millis_encoding
 *     self.date_type_accepts_string = date_type_accepts_string
 *     self.raise_on_invalid_logical = raise_on_invalid_logical             # <<<<<<<<<<<<<<
 *     self.inline_namespaces = inline_namespaces
 *     self.expand_types_in_schema = expand_types_in_schema
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_raise_on_invalid_logical); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 54, __pyx_L1_error)
  __pyx_v_self->raise_on_invalid_logical = __pyx_t_1;

  /* "(tree fragment)":55
 *     self.date_type_accepts_string = date_type_accepts_string
 *     self.raise_on_invalid_logical = raise_on_invalid_logical
 *     self.inline_namespaces = inline_namespaces             # <<<<<<<<<<<<<<
 *     self.expand_types_in_schema = expand_types_in_schema
 *     self.unicode_errors = unicode_errors
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_inline_namespaces); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 55, __pyx_L1_error)
  __pyx_v_self->inline_namespaces = __pyx_t_1;

  /* "(tree fragment)":56
 *     self.raise_on_invalid_logical = raise_on_invalid_logical
 *     self.inline_namespaces = inline_namespaces
 *     self.expand_types_in_schema = expand_types_in_schema             # <<<<<<<<<<<<<<
 *     self.unicode_errors = unicode_errors
 *     self.container_fill_blocks = container_fill_blocks
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_expand_types_in_schema); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 56, __pyx_L1_error)
  __pyx_v_self->expand_types_in_schema = __pyx_t_1;

  /* "(tree fragment)":57
 *     self.inline_namespaces = inline_namespaces
 *     self.expand_types_in_schema = expand_types_in_schema
 *     self.unicode_errors = unicode_errors             # <<<<<<<<<<<<<<
 *     self.container_fill_blocks = container_fill_blocks
 *     self.defer_schema_promotion_errors = defer_schema_promotion_errors
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_unicode_errors))||((__pyx_v_unicode_errors) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_unicode_errors))) __PYX_ERR(1, 57, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_unicode_errors;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->unicode_errors);
  __Pyx_DECREF(__pyx_v_self->unicode_errors);
  __pyx_v_self->unicode_errors = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":58
 *     self.expand_types_in_schema = expand_types_in_schema
 *     self.unicode_errors = unicode_errors
 *     self.container_fill_blocks = container_fill_blocks             # <<<<<<<<<<<<<<
 *     self.defer_schema_promotion_errors = defer_schema_promotion_errors
 *     self.invalid_value_includes_record_name = invalid_value_includes_record_name
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_container_fill_blocks); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 58, __pyx_L1_error)
  __pyx_v_self->container_fill_blocks = __pyx_t_1;

  /* "(tree fragment)":59
 *     self.unicode_errors = unicode_errors
 *     self.container_fill_blocks = container_fill_blocks
 *     self.defer_schema_promotion_errors = defer_schema_promotion_errors             # <<<<<<<<<<<<<<
 *     self.invalid_value_includes_record_name = invalid_value_includes_record_name
 *     self.invalid_value_include_array_index = invalid_value_include_array_index
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_defer_schema_promotion_errors); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 59, __pyx_L1_error)
  __pyx_v_self->defer_schema_promotion_errors = __pyx_t_1;

  /* "(tree fragment)":60
 *     self.container_fill_blocks = container_fill_blocks
 *     self.defer_schema_promotion_errors = defer_schema_promotion_errors
 *     self.invalid_value_includes_record_name = invalid_value_includes_record_name             # <<<<<<<<<<<<<<
 *     self.invalid_value_include_array_index = invalid_value_include_array_index
 *     self.invalid_value_include_map_key = invalid_value_include_map_key
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_invalid_value_includes_record_name); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 60, __pyx_L1_error)
  __pyx_v_self->invalid_value_includes_record_name = __pyx_t_1;

  /* "(tree fragment)":61
 *     self.defer_schema_promotion_errors = defer_schema_promotion_errors
 *     self.invalid_value_includes_record_name = invalid_value_includes_record_name
 *     self.invalid_value_include_array_index = invalid_value_include_array_index             # <<<<<<<<<<<<<<
 *     self.invalid_value_include_map_key = invalid_value_include_map_key
 *     self.allow_invalid_default_values = allow_invalid_default_values
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_invalid_value_include_array_index); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 61, __pyx_L1_error)
  __pyx_v_self->invalid_value_include_array_index = __pyx_t_1;

  /* "(tree fragment)":62
 *     self.invalid_value_includes_record_name = invalid_value_includes_record_name
 *     self.invalid_value_include_array_index = invalid_value_include_array_index
 *     self.invalid_value_include_map_key = invalid_value_include_map_key             # <<<<<<<<<<<<<<
 *     self.allow_invalid_default_values = allow_invalid_default_values
 *     self.externally_defined_types = externally_defined_types
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_invalid_value_include_map_key); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 62, __pyx_L1_error)
  __pyx_v_self->invalid_value_include_map_key = __pyx_t_1;

  /* "(tree fragment)":63
 *     self.invalid_value_include_array_index = invalid_value_include_array_index
 *     self.invalid_value_include_map_key = invalid_value_include_map_key
 *     self.allow_invalid_default_values = allow_invalid_default_values             # <<<<<<<<<<<<<<
 *     self.externally_defined_types = externally_defined_types
 * __pyx_recursive_repr_guard = __import__('threading').local()
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_allow_invalid_default_values); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 63, __pyx_L1_error)
  __pyx_v_self->allow_invalid_default_values = __pyx_t_1;

  /* "(tree fragment)":64
 *     self.invalid_value_include_map_key = invalid_value_include_map_key
 *     self.allow_invalid_default_values = allow_invalid_default_values
 *     self.externally_defined_types = externally_defined_types             # <<<<<<<<<<<<<<
 * __pyx_recursive_repr_guard = __import__('threading').local()
 * __pyx_recursive_repr_guard.running = set()
 */
  __Pyx_INCREF(__pyx_v_externally_defined_types);
  __Pyx_GIVEREF(__pyx_v_externally_defined_types);
  __Pyx_GOTREF(__pyx_v_self->externally_defined_types);
  __Pyx_DECREF(__pyx_v_self->externally_defined_types);
  __pyx_v_self->externally_defined_types = __pyx_v_externally_defined_types;

  /* "(tree fragment)":1
 * def __init__(self, *, fingerprint_returns_digest: bint = DATACLASS_PLACEHOLDER_1, canonical_form_repeat_fixed: bint = DATACLASS_PLACEHOLDER_2, canonical_form_repeat_enum: bint = DATACLASS_PLACEHOLDER_3, record_can_encode_dict: bint = DATACLASS_PLACEHOLDER_4, record_values_type_hint: bint = DATACLASS_PLACEHOLDER_5, record_decodes_to_dict: bint = DATACLASS_PLACEHOLDER_6, record_allow_extra_fields: bint = DATACLASS_PLACEHOLDER_7, record_encode_use_defaults: bint = DATACLASS_PLACEHOLDER_8, allow_tuple_notation: bint = DATACLASS_PLACEHOLDER_9, union_decodes_to: UnionDecodeOption = DATACLASS_PLACEHOLDER_10, union_json_encodes_type_name: bint = DATACLASS_PLACEHOLDER_11, allow_primitive_name_collision: bint = DATACLASS_PLACEHOLDER_12, allow_primitive_names_in_namespaces: bint = DATACLASS_PLACEHOLDER_13, named_type_names_must_be_unique: bint = DATACLASS_PLACEHOLDER_14, enum_symbols_must_be_unique: bint = DATACLASS_PLACEHOLDER_15, enforce_enum_symbol_name_rules: bint = DATACLASS_PLACEHOLDER_16, enforce_type_name_rules: bint = DATACLASS_PLACEHOLDER_17, enforce_namespace_name_rules: bint = DATACLASS_PLACEHOLDER_18, record_fields_must_be_unique: bint = DATACLASS_PLACEHOLDER_19, ascii_name_rules: bint = DATACLASS_PLACEHOLDER_20, missing_values_can_be_null: bint = DATACLASS_PLACEHOLDER_21, missing_values_can_be_empty_container: bint = DATACLASS_PLACEHOLDER_22, allow_false_values_for_null: bint = DATACLASS_PLACEHOLDER_23, allow_empty_unions: bint = DATACLASS_PLACEHOLDER_24, allow_nested_unions: bint = DATACLASS_PLACEHOLDER_25, allow_duplicate_union_types: bint = DATACLASS_PLACEHOLDER_26, allow_union_default_any_member: bint = DATACLASS_PLACEHOLDER_27, allow_aliases_to_be_string: bint = DATACLASS_PLACEHOLDER_28, coerce_values_to_int: bint = DATACLASS_PLACEHOLDER_29, coerce_values_to_float: bint = DATACLASS_PLACEHOLDER_30, coerce_int_to_float: bint = DATACLASS_PLACEHOLDER_31, coerce_values_to_boolean: bint = DATACLASS_PLACEHOLDER_32, coerce_values_to_str: bint = DATACLASS_PLACEHOLDER_33, bytes_codec: str = DATACLASS_PLACEHOLDER_34, fixed_codec: str = DATACLASS_PLACEHOLDER_35, null_pad_fixed: bint = DATACLASS_PLACEHOLDER_36, truncate_fixed: bint = DATACLASS_PLACEHOLDER_37, clamp_int_overflow: bint = DATACLASS_PLACEHOLDER_38, clamp_float_overflow: bint = DATACLASS_PLACEHOLDER_39, float_out_of_range_inf: bint = DATACLASS_PLACEHOLDER_40, bytes_default_value_utf8: bint = DATACLASS_PLACEHOLDER_41, string_types_default_unchanged: bint = DATACLASS_PLACEHOLDER_42, decimal_check_exp_overflow: bint = DATACLASS_PLACEHOLDER_43, types_str_to_schema: bint = DATACLASS_PLACEHOLDER_44, logical_types: tuple[LogicalType] = DATACLASS_PLACEHOLDER_45, adapt_record_types: bint = DATACLASS_PLACEHOLDER_46, return_uuid_object: bint = DATACLASS_PLACEHOLDER_47, allow_error_type: bint = DATACLASS_PLACEHOLDER_48, allow_leading_dot_in_names: bint = DATACLASS_PLACEHOLDER_49, naive_dt_assume_utc: bint = DATACLASS_PLACEHOLDER_50, alternate_timestamp_millis_encoding: bint = DATACLASS_PLACEHOLDER_51, date_type_accepts_string: bint = DATACLASS_PLACEHOLDER_52, raise_on_invalid_logical: bint = DATACLASS_PLACEHOLDER_53, inline_namespaces: bint = DATACLASS_PLACEHOLDER_54, expand_types_in_schema: bint = DATACLASS_PLACEHOLDER_55, unicode_errors: str = DATACLASS_PLACEHOLDER_56, container_fill_blocks: bint = DATACLASS_PLACEHOLDER_57, defer_schema_promotion_errors: bint = DATACLASS_PLACEHOLDER_58, invalid_value_includes_record_name: bint = DATACLASS_PLACEHOLDER_59, invalid_value_include_array_index: bint = DATACLASS_PLACEHOLDER_60, invalid_value_include_map_key: bint = DATACLASS_PLACEHOLDER_61, allow_invalid_default_values: bint = DATACLASS_PLACEHOLDER_62, externally_defined_types: MappingProxyType = DATACLASS_PLACEHOLDER_63):             # <<<<<<<<<<<<<<
 *     self.fingerprint_returns_digest = fingerprint_returns_digest
 *     self.canonical_form_repeat_fixed = canonical_form_repeat_fixed
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.Options.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":67
 * __pyx_recursive_repr_guard = __import__('threading').local()
 * __pyx_recursive_repr_guard.running = set()
 * def __repr__(self):             # <<<<<<<<<<<<<<
 *     key = id(self)
 *     guard_set = self.__pyx_recursive_repr_guard.running
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_13__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_13__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_12__repr__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_12__repr__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_guard_set = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  int __pyx_t_8;
  char const *__pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "(tree fragment)":68
 * __pyx_recursive_repr_guard.running = set()
 * def __repr__(self):
 *     key = id(self)             # <<<<<<<<<<<<<<
 *     guard_set = self.__pyx_recursive_repr_guard.running
 *     if key in guard_set: return '...'
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_key = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":69
 * def __repr__(self):
 *     key = id(self)
 *     guard_set = self.__pyx_recursive_repr_guard.running             # <<<<<<<<<<<<<<
 *     if key in guard_set: return '...'
 *     guard_set.add(key)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_recursive_repr_guard); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_running); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_guard_set = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":70
 *     key = id(self)
 *     guard_set = self.__pyx_recursive_repr_guard.running
 *     if key in guard_set: return '...'             # <<<<<<<<<<<<<<
 *     guard_set.add(key)
 *     try:
 */
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_key, __pyx_v_guard_set, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(1, 70, __pyx_L1_error)
  if (__pyx_t_3) {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_kp_s__18);
    __pyx_r = __pyx_kp_s__18;
    goto __pyx_L0;
  }

  /* "(tree fragment)":71
 *     guard_set = self.__pyx_recursive_repr_guard.running
 *     if key in guard_set: return '...'
 *     guard_set.add(key)             # <<<<<<<<<<<<<<
 *     try:
 *         name = getattr(type(self), "__qualname__", type(self).__name__)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_guard_set, __pyx_n_s_add); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_key};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":72
 *     if key in guard_set: return '...'
 *     guard_set.add(key)
 *     try:             # <<<<<<<<<<<<<<
 *         name = getattr(type(self), "__qualname__", type(self).__name__)
 *         return f'{name}(fingerprint_returns_digest={self.fingerprint_returns_digest!r}, canonical_form_repeat_fixed={self.canonical_form_repeat_fixed!r}, canonical_form_repeat_enum={self.canonical_form_repeat_enum!r}, record_can_encode_dict={self.record_can_encode_dict!r}, record_values_type_hint={self.record_values_type_hint!r}, record_decodes_to_dict={self.record_decodes_to_dict!r}, record_allow_extra_fields={self.record_allow_extra_fields!r}, record_encode_use_defaults={self.record_encode_use_defaults!r}, allow_tuple_notation={self.allow_tuple_notation!r}, union_decodes_to={self.union_decodes_to!r}, union_json_encodes_type_name={self.union_json_encodes_type_name!r}, allow_primitive_name_collision={self.allow_primitive_name_collision!r}, allow_primitive_names_in_namespaces={self.allow_primitive_names_in_namespaces!r}, named_type_names_must_be_unique={self.named_type_names_must_be_unique!r}, enum_symbols_must_be_unique={self.enum_symbols_must_be_unique!r}, enforce_enum_symbol_name_rules={self.enforce_enum_symbol_name_rules!r}, enforce_type_name_rules={self.enforce_type_name_rules!r}, enforce_namespace_name_rules={self.enforce_namespace_name_rules!r}, record_fields_must_be_unique={self.record_fields_must_be_unique!r}, ascii_name_rules={self.ascii_name_rules!r}, missing_values_can_be_null={self.missing_values_can_be_null!r}, missing_values_can_be_empty_container={self.missing_values_can_be_empty_container!r}, allow_false_values_for_null={self.allow_false_values_for_null!r}, allow_empty_unions={self.allow_empty_unions!r}, allow_nested_unions={self.allow_nested_unions!r}, allow_duplicate_union_types={self.allow_duplicate_union_types!r}, allow_union_default_any_member={self.allow_union_default_any_member!r}, allow_aliases_to_be_string={self.allow_aliases_to_be_string!r}, coerce_values_to_int={self.coerce_values_to_int!r}, coerce_values_to_float={self.coerce_values_to_float!r}, coerce_int_to_float={self.coerce_int_to_float!r}, coerce_values_to_boolean={self.coerce_values_to_boolean!r}, coerce_values_to_str={self.coerce_values_to_str!r}, bytes_codec={self.bytes_codec!r}, fixed_codec={self.fixed_codec!r}, null_pad_fixed={self.null_pad_fixed!r}, truncate_fixed={self.truncate_fixed!r}, clamp_int_overflow={self.clamp_int_overflow!r}, clamp_float_overflow={self.clamp_float_overflow!r}, float_out_of_range_inf={self.float_out_of_range_inf!r}, bytes_default_value_utf8={self.bytes_default_value_utf8!r}, string_types_default_unchanged={self.string_types_default_unchanged!r}, decimal_check_exp_overflow={self.decimal_check_exp_overflow!r}, types_str_to_schema={self.types_str_to_schema!r}, logical_types={self.logical_types!r}, adapt_record_types={self.adapt_record_types!r}, return_uuid_object={self.return_uuid_object!r}, allow_error_type={self.allow_error_type!r}, allow_leading_dot_in_names={self.allow_leading_dot_in_names!r}, naive_dt_assume_utc={self.naive_dt_assume_utc!r}, alternate_timestamp_millis_encoding={self.alternate_timestamp_millis_encoding!r}, date_type_accepts_string={self.date_type_accepts_string!r}, raise_on_invalid_logical={self.raise_on_invalid_logical!r}, inline_namespaces={self.inline_namespaces!r}, expand_types_in_schema={self.expand_types_in_schema!r}, unicode_errors={self.unicode_errors!r}, container_fill_blocks={self.container_fill_blocks!r}, defer_schema_promotion_errors={self.defer_schema_promotion_errors!r}, invalid_value_includes_record_name={self.invalid_value_includes_record_name!r}, invalid_value_include_array_index={self.invalid_value_include_array_index!r}, invalid_value_include_map_key={self.invalid_value_include_map_key!r}, allow_invalid_default_values={self.allow_invalid_default_values!r}, externally_defined_types={self.externally_defined_types!r})'
 */
  /*try:*/ {

    /* "(tree fragment)":73
 *     guard_set.add(key)
 *     try:
 *         name = getattr(type(self), "__qualname__", type(self).__name__)             # <<<<<<<<<<<<<<
 *         return f'{name}(fingerprint_returns_digest={self.fingerprint_returns_digest!r}, canonical_form_repeat_fixed={self.canonical_form_repeat_fixed!r}, canonical_form_repeat_enum={self.canonical_form_repeat_enum!r}, record_can_encode_dict={self.record_can_encode_dict!r}, record_values_type_hint={self.record_values_type_hint!r}, record_decodes_to_dict={self.record_decodes_to_dict!r}, record_allow_extra_fields={self.record_allow_extra_fields!r}, record_encode_use_defaults={self.record_encode_use_defaults!r}, allow_tuple_notation={self.allow_tuple_notation!r}, union_decodes_to={self.union_decodes_to!r}, union_json_encodes_type_name={self.union_json_encodes_type_name!r}, allow_primitive_name_collision={self.allow_primitive_name_collision!r}, allow_primitive_names_in_namespaces={self.allow_primitive_names_in_namespaces!r}, named_type_names_must_be_unique={self.named_type_names_must_be_unique!r}, enum_symbols_must_be_unique={self.enum_symbols_must_be_unique!r}, enforce_enum_symbol_name_rules={self.enforce_enum_symbol_name_rules!r}, enforce_type_name_rules={self.enforce_type_name_rules!r}, enforce_namespace_name_rules={self.enforce_namespace_name_rules!r}, record_fields_must_be_unique={self.record_fields_must_be_unique!r}, ascii_name_rules={self.ascii_name_rules!r}, missing_values_can_be_null={self.missing_values_can_be_null!r}, missing_values_can_be_empty_container={self.missing_values_can_be_empty_container!r}, allow_false_values_for_null={self.allow_false_values_for_null!r}, allow_empty_unions={self.allow_empty_unions!r}, allow_nested_unions={self.allow_nested_unions!r}, allow_duplicate_union_types={self.allow_duplicate_union_types!r}, allow_union_default_any_member={self.allow_union_default_any_member!r}, allow_aliases_to_be_string={self.allow_aliases_to_be_string!r}, coerce_values_to_int={self.coerce_values_to_int!r}, coerce_values_to_float={self.coerce_values_to_float!r}, coerce_int_to_float={self.coerce_int_to_float!r}, coerce_values_to_boolean={self.coerce_values_to_boolean!r}, coerce_values_to_str={self.coerce_values_to_str!r}, bytes_codec={self.bytes_codec!r}, fixed_codec={self.fixed_codec!r}, null_pad_fixed={self.null_pad_fixed!r}, truncate_fixed={self.truncate_fixed!r}, clamp_int_overflow={self.clamp_int_overflow!r}, clamp_float_overflow={self.clamp_float_overflow!r}, float_out_of_range_inf={self.float_out_of_range_inf!r}, bytes_default_value_utf8={self.bytes_default_value_utf8!r}, string_types_default_unchanged={self.string_types_default_unchanged!r}, decimal_check_exp_overflow={self.decimal_check_exp_overflow!r}, types_str_to_schema={self.types_str_to_schema!r}, logical_types={self.logical_types!r}, adapt_record_types={self.adapt_record_types!r}, return_uuid_object={self.return_uuid_object!r}, allow_error_type={self.allow_error_type!r}, allow_leading_dot_in_names={self.allow_leading_dot_in_names!r}, naive_dt_assume_utc={self.naive_dt_assume_utc!r}, alternate_timestamp_millis_encoding={self.alternate_timestamp_millis_encoding!r}, date_type_accepts_string={self.date_type_accepts_string!r}, raise_on_invalid_logical={self.raise_on_invalid_logical!r}, inline_namespaces={self.inline_namespaces!r}, expand_types_in_schema={self.expand_types_in_schema!r}, unicode_errors={self.unicode_errors!r}, container_fill_blocks={self.container_fill_blocks!r}, defer_schema_promotion_errors={self.defer_schema_promotion_errors!r}, invalid_value_includes_record_name={self.invalid_value_includes_record_name!r}, invalid_value_include_array_index={self.invalid_value_include_array_index!r}, invalid_value_include_map_key={self.invalid_value_include_map_key!r}, allow_invalid_default_values={self.allow_invalid_default_values!r}, externally_defined_types={self.externally_defined_types!r})'
 *     finally:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 73, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_qualname, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 73, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_name = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "(tree fragment)":74
 *     try:
 *         name = getattr(type(self), "__qualname__", type(self).__name__)
 *         return f'{name}(fingerprint_returns_digest={self.fingerprint_returns_digest!r}, canonical_form_repeat_fixed={self.canonical_form_repeat_fixed!r}, canonical_form_repeat_enum={self.canonical_form_repeat_enum!r}, record_can_encode_dict={self.record_can_encode_dict!r}, record_values_type_hint={self.record_values_type_hint!r}, record_decodes_to_dict={self.record_decodes_to_dict!r}, record_allow_extra_fields={self.record_allow_extra_fields!r}, record_encode_use_defaults={self.record_encode_use_defaults!r}, allow_tuple_notation={self.allow_tuple_notation!r}, union_decodes_to={self.union_decodes_to!r}, union_json_encodes_type_name={self.union_json_encodes_type_name!r}, allow_primitive_name_collision={self.allow_primitive_name_collision!r}, allow_primitive_names_in_namespaces={self.allow_primitive_names_in_namespaces!r}, named_type_names_must_be_unique={self.named_type_names_must_be_unique!r}, enum_symbols_must_be_unique={self.enum_symbols_must_be_unique!r}, enforce_enum_symbol_name_rules={self.enforce_enum_symbol_name_rules!r}, enforce_type_name_rules={self.enforce_type_name_rules!r}, enforce_namespace_name_rules={self.enforce_namespace_name_rules!r}, record_fields_must_be_unique={self.record_fields_must_be_unique!r}, ascii_name_rules={self.ascii_name_rules!r}, missing_values_can_be_null={self.missing_values_can_be_null!r}, missing_values_can_be_empty_container={self.missing_values_can_be_empty_container!r}, allow_false_values_for_null={self.allow_false_values_for_null!r}, allow_empty_unions={self.allow_empty_unions!r}, allow_nested_unions={self.allow_nested_unions!r}, allow_duplicate_union_types={self.allow_duplicate_union_types!r}, allow_union_default_any_member={self.allow_union_default_any_member!r}, allow_aliases_to_be_string={self.allow_aliases_to_be_string!r}, coerce_values_to_int={self.coerce_values_to_int!r}, coerce_values_to_float={self.coerce_values_to_float!r}, coerce_int_to_float={self.coerce_int_to_float!r}, coerce_values_to_boolean={self.coerce_values_to_boolean!r}, coerce_values_to_str={self.coerce_values_to_str!r}, bytes_codec={self.bytes_codec!r}, fixed_codec={self.fixed_codec!r}, null_pad_fixed={self.null_pad_fixed!r}, truncate_fixed={self.truncate_fixed!r}, clamp_int_overflow={self.clamp_int_overflow!r}, clamp_float_overflow={self.clamp_float_overflow!r}, float_out_of_range_inf={self.float_out_of_range_inf!r}, bytes_default_value_utf8={self.bytes_default_value_utf8!r}, string_types_default_unchanged={self.string_types_default_unchanged!r}, decimal_check_exp_overflow={self.decimal_check_exp_overflow!r}, types_str_to_schema={self.types_str_to_schema!r}, logical_types={self.logical_types!r}, adapt_record_types={self.adapt_record_types!r}, return_uuid_object={self.return_uuid_object!r}, allow_error_type={self.allow_error_type!r}, allow_leading_dot_in_names={self.allow_leading_dot_in_names!r}, naive_dt_assume_utc={self.naive_dt_assume_utc!r}, alternate_timestamp_millis_encoding={self.alternate_timestamp_millis_encoding!r}, date_type_accepts_string={self.date_type_accepts_string!r}, raise_on_invalid_logical={self.raise_on_invalid_logical!r}, inline_namespaces={self.inline_namespaces!r}, expand_types_in_schema={self.expand_types_in_schema!r}, unicode_errors={self.unicode_errors!r}, container_fill_blocks={self.container_fill_blocks!r}, defer_schema_promotion_errors={self.defer_schema_promotion_errors!r}, invalid_value_includes_record_name={self.invalid_value_includes_record_name!r}, invalid_value_include_array_index={self.invalid_value_include_array_index!r}, invalid_value_include_map_key={self.invalid_value_include_map_key!r}, allow_invalid_default_values={self.allow_invalid_default_values!r}, externally_defined_types={self.externally_defined_types!r})'             # <<<<<<<<<<<<<<
 *     finally:
 *         guard_set.remove(key)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyTuple_New(128); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_7 = 127;
    __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_name, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_fingerprint_returns_digest_2);
    __pyx_t_6 += 28;
    __Pyx_GIVEREF(__pyx_kp_u_fingerprint_returns_digest_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_fingerprint_returns_digest_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->fingerprint_returns_digest); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_canonical_form_repeat_fixed_2);
    __pyx_t_6 += 30;
    __Pyx_GIVEREF(__pyx_kp_u_canonical_form_repeat_fixed_2);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u_canonical_form_repeat_fixed_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->canonical_form_repeat_fixed); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_canonical_form_repeat_enum_2);
    __pyx_t_6 += 29;
    __Pyx_GIVEREF(__pyx_kp_u_canonical_form_repeat_enum_2);
    PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_kp_u_canonical_form_repeat_enum_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->canonical_form_repeat_enum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_record_can_encode_dict_2);
    __pyx_t_6 += 25;
    __Pyx_GIVEREF(__pyx_kp_u_record_can_encode_dict_2);
    PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_kp_u_record_can_encode_dict_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->record_can_encode_dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_record_values_type_hint_2);
    __pyx_t_6 += 26;
    __Pyx_GIVEREF(__pyx_kp_u_record_values_type_hint_2);
    PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_kp_u_record_values_type_hint_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->record_values_type_hint); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_record_decodes_to_dict_2);
    __pyx_t_6 += 25;
    __Pyx_GIVEREF(__pyx_kp_u_record_decodes_to_dict_2);
    PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_kp_u_record_decodes_to_dict_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->record_decodes_to_dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_record_allow_extra_fields_2);
    __pyx_t_6 += 28;
    __Pyx_GIVEREF(__pyx_kp_u_record_allow_extra_fields_2);
    PyTuple_SET_ITEM(__pyx_t_1, 13, __pyx_kp_u_record_allow_extra_fields_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->record_allow_extra_fields); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 14, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_record_encode_use_defaults_2);
    __pyx_t_6 += 29;
    __Pyx_GIVEREF(__pyx_kp_u_record_encode_use_defaults_2);
    PyTuple_SET_ITEM(__pyx_t_1, 15, __pyx_kp_u_record_encode_use_defaults_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->record_encode_use_defaults); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 16, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_allow_tuple_notation_2);
    __pyx_t_6 += 23;
    __Pyx_GIVEREF(__pyx_kp_u_allow_tuple_notation_2);
    PyTuple_SET_ITEM(__pyx_t_1, 17, __pyx_kp_u_allow_tuple_notation_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->allow_tuple_notation); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 18, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_union_decodes_to_2);
    __pyx_t_6 += 19;
    __Pyx_GIVEREF(__pyx_kp_u_union_decodes_to_2);
    PyTuple_SET_ITEM(__pyx_t_1, 19, __pyx_kp_u_union_decodes_to_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->union_decodes_to), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 20, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_union_json_encodes_type_name_2);
    __pyx_t_6 += 31;
    __Pyx_GIVEREF(__pyx_kp_u_union_json_encodes_type_name_2);
    PyTuple_SET_ITEM(__pyx_t_1, 21, __pyx_kp_u_union_json_encodes_type_name_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->union_json_encodes_type_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 22, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_allow_primitive_name_collision_2);
    __pyx_t_6 += 33;
    __Pyx_GIVEREF(__pyx_kp_u_allow_primitive_name_collision_2);
    PyTuple_SET_ITEM(__pyx_t_1, 23, __pyx_kp_u_allow_primitive_name_collision_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->allow_primitive_name_collision); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 24, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_allow_primitive_names_in_namesp);
    __pyx_t_6 += 38;
    __Pyx_GIVEREF(__pyx_kp_u_allow_primitive_names_in_namesp);
    PyTuple_SET_ITEM(__pyx_t_1, 25, __pyx_kp_u_allow_primitive_names_in_namesp);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->allow_primitive_names_in_namespaces); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 26, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_named_type_names_must_be_unique_2);
    __pyx_t_6 += 34;
    __Pyx_GIVEREF(__pyx_kp_u_named_type_names_must_be_unique_2);
    PyTuple_SET_ITEM(__pyx_t_1, 27, __pyx_kp_u_named_type_names_must_be_unique_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->named_type_names_must_be_unique); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 28, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_enum_symbols_must_be_unique_2);
    __pyx_t_6 += 30;
    __Pyx_GIVEREF(__pyx_kp_u_enum_symbols_must_be_unique_2);
    PyTuple_SET_ITEM(__pyx_t_1, 29, __pyx_kp_u_enum_symbols_must_be_unique_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->enum_symbols_must_be_unique); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 30, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_enforce_enum_symbol_name_rules_2);
    __pyx_t_6 += 33;
    __Pyx_GIVEREF(__pyx_kp_u_enforce_enum_symbol_name_rules_2);
    PyTuple_SET_ITEM(__pyx_t_1, 31, __pyx_kp_u_enforce_enum_symbol_name_rules_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->enforce_enum_symbol_name_rules); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 32, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_enforce_type_name_rules_2);
    __pyx_t_6 += 26;
    __Pyx_GIVEREF(__pyx_kp_u_enforce_type_name_rules_2);
    PyTuple_SET_ITEM(__pyx_t_1, 33, __pyx_kp_u_enforce_type_name_rules_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->enforce_type_name_rules); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 34, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_enforce_namespace_name_rules_2);
    __pyx_t_6 += 31;
    __Pyx_GIVEREF(__pyx_kp_u_enforce_namespace_name_rules_2);
    PyTuple_SET_ITEM(__pyx_t_1, 35, __pyx_kp_u_enforce_namespace_name_rules_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->enforce_namespace_name_rules); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 36, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_record_fields_must_be_unique_2);
    __pyx_t_6 += 31;
    __Pyx_GIVEREF(__pyx_kp_u_record_fields_must_be_unique_2);
    PyTuple_SET_ITEM(__pyx_t_1, 37, __pyx_kp_u_record_fields_must_be_unique_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->record_fields_must_be_unique); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 38, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_ascii_name_rules_2);
    __pyx_t_6 += 19;
    __Pyx_GIVEREF(__pyx_kp_u_ascii_name_rules_2);
    PyTuple_SET_ITEM(__pyx_t_1, 39, __pyx_kp_u_ascii_name_rules_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->ascii_name_rules); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 40, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_missing_values_can_be_null_2);
    __pyx_t_6 += 29;
    __Pyx_GIVEREF(__pyx_kp_u_missing_values_can_be_null_2);
    PyTuple_SET_ITEM(__pyx_t_1, 41, __pyx_kp_u_missing_values_can_be_null_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->missing_values_can_be_null); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 42, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_missing_values_can_be_empty_con);
    __pyx_t_6 += 40;
    __Pyx_GIVEREF(__pyx_kp_u_missing_values_can_be_empty_con);
    PyTuple_SET_ITEM(__pyx_t_1, 43, __pyx_kp_u_missing_values_can_be_empty_con);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->missing_values_can_be_empty_container); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 44, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_allow_false_values_for_null_2);
    __pyx_t_6 += 30;
    __Pyx_GIVEREF(__pyx_kp_u_allow_false_values_for_null_2);
    PyTuple_SET_ITEM(__pyx_t_1, 45, __pyx_kp_u_allow_false_values_for_null_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->allow_false_values_for_null); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 46, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_allow_empty_unions_2);
    __pyx_t_6 += 21;
    __Pyx_GIVEREF(__pyx_kp_u_allow_empty_unions_2);
    PyTuple_SET_ITEM(__pyx_t_1, 47, __pyx_kp_u_allow_empty_unions_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->allow_empty_unions); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 48, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_allow_nested_unions_2);
    __pyx_t_6 += 22;
    __Pyx_GIVEREF(__pyx_kp_u_allow_nested_unions_2);
    PyTuple_SET_ITEM(__pyx_t_1, 49, __pyx_kp_u_allow_nested_unions_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->allow_nested_unions); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 50, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_allow_duplicate_union_types_2);
    __pyx_t_6 += 30;
    __Pyx_GIVEREF(__pyx_kp_u_allow_duplicate_union_types_2);
    PyTuple_SET_ITEM(__pyx_t_1, 51, __pyx_kp_u_allow_duplicate_union_types_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->allow_duplicate_union_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 52, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_allow_union_default_any_member_2);
    __pyx_t_6 += 33;
    __Pyx_GIVEREF(__pyx_kp_u_allow_union_default_any_member_2);
    PyTuple_SET_ITEM(__pyx_t_1, 53, __pyx_kp_u_allow_union_default_any_member_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->allow_union_default_any_member); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 54, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_allow_aliases_to_be_string_2);
    __pyx_t_6 += 29;
    __Pyx_GIVEREF(__pyx_kp_u_allow_aliases_to_be_string_2);
    PyTuple_SET_ITEM(__pyx_t_1, 55, __pyx_kp_u_allow_aliases_to_be_string_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->allow_aliases_to_be_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 56, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_coerce_values_to_int_2);
    __pyx_t_6 += 23;
    __Pyx_GIVEREF(__pyx_kp_u_coerce_values_to_int_2);
    PyTuple_SET_ITEM(__pyx_t_1, 57, __pyx_kp_u_coerce_values_to_int_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->coerce_values_to_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 58, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_coerce_values_to_float_2);
    __pyx_t_6 += 25;
    __Pyx_GIVEREF(__pyx_kp_u_coerce_values_to_float_2);
    PyTuple_SET_ITEM(__pyx_t_1, 59, __pyx_kp_u_coerce_values_to_float_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->coerce_values_to_float); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 60, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_coerce_int_to_float_2);
    __pyx_t_6 += 22;
    __Pyx_GIVEREF(__pyx_kp_u_coerce_int_to_float_2);
    PyTuple_SET_ITEM(__pyx_t_1, 61, __pyx_kp_u_coerce_int_to_float_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->coerce_int_to_float); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 62, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_coerce_values_to_boolean_2);
    __pyx_t_6 += 27;
    __Pyx_GIVEREF(__pyx_kp_u_coerce_values_to_boolean_2);
    PyTuple_SET_ITEM(__pyx_t_1, 63, __pyx_kp_u_coerce_values_to_boolean_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->coerce_values_to_boolean); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 64, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_coerce_values_to_str_2);
    __pyx_t_6 += 23;
    __Pyx_GIVEREF(__pyx_kp_u_coerce_values_to_str_2);
    PyTuple_SET_ITEM(__pyx_t_1, 65, __pyx_kp_u_coerce_values_to_str_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->coerce_values_to_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 66, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_bytes_codec_2);
    __pyx_t_6 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_bytes_codec_2);
    PyTuple_SET_ITEM(__pyx_t_1, 67, __pyx_kp_u_bytes_codec_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->bytes_codec), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 68, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_fixed_codec_2);
    __pyx_t_6 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_fixed_codec_2);
    PyTuple_SET_ITEM(__pyx_t_1, 69, __pyx_kp_u_fixed_codec_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->fixed_codec), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 70, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_null_pad_fixed_2);
    __pyx_t_6 += 17;
    __Pyx_GIVEREF(__pyx_kp_u_null_pad_fixed_2);
    PyTuple_SET_ITEM(__pyx_t_1, 71, __pyx_kp_u_null_pad_fixed_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->null_pad_fixed); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 72, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_truncate_fixed_2);
    __pyx_t_6 += 17;
    __Pyx_GIVEREF(__pyx_kp_u_truncate_fixed_2);
    PyTuple_SET_ITEM(__pyx_t_1, 73, __pyx_kp_u_truncate_fixed_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->truncate_fixed); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 74, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_clamp_int_overflow_2);
    __pyx_t_6 += 21;
    __Pyx_GIVEREF(__pyx_kp_u_clamp_int_overflow_2);
    PyTuple_SET_ITEM(__pyx_t_1, 75, __pyx_kp_u_clamp_int_overflow_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->clamp_int_overflow); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 76, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_clamp_float_overflow_2);
    __pyx_t_6 += 23;
    __Pyx_GIVEREF(__pyx_kp_u_clamp_float_overflow_2);
    PyTuple_SET_ITEM(__pyx_t_1, 77, __pyx_kp_u_clamp_float_overflow_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->clamp_float_overflow); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 78, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_float_out_of_range_inf_2);
    __pyx_t_6 += 25;
    __Pyx_GIVEREF(__pyx_kp_u_float_out_of_range_inf_2);
    PyTuple_SET_ITEM(__pyx_t_1, 79, __pyx_kp_u_float_out_of_range_inf_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->float_out_of_range_inf); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 80, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_bytes_default_value_utf8_2);
    __pyx_t_6 += 27;
    __Pyx_GIVEREF(__pyx_kp_u_bytes_default_value_utf8_2);
    PyTuple_SET_ITEM(__pyx_t_1, 81, __pyx_kp_u_bytes_default_value_utf8_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->bytes_default_value_utf8); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 82, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_string_types_default_unchanged_2);
    __pyx_t_6 += 33;
    __Pyx_GIVEREF(__pyx_kp_u_string_types_default_unchanged_2);
    PyTuple_SET_ITEM(__pyx_t_1, 83, __pyx_kp_u_string_types_default_unchanged_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->string_types_default_unchanged); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 84, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_decimal_check_exp_overflow_2);
    __pyx_t_6 += 29;
    __Pyx_GIVEREF(__pyx_kp_u_decimal_check_exp_overflow_2);
    PyTuple_SET_ITEM(__pyx_t_1, 85, __pyx_kp_u_decimal_check_exp_overflow_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->decimal_check_exp_overflow); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 86, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_types_str_to_schema_2);
    __pyx_t_6 += 22;
    __Pyx_GIVEREF(__pyx_kp_u_types_str_to_schema_2);
    PyTuple_SET_ITEM(__pyx_t_1, 87, __pyx_kp_u_types_str_to_schema_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->types_str_to_schema); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 88, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_logical_types_2);
    __pyx_t_6 += 16;
    __Pyx_GIVEREF(__pyx_kp_u_logical_types_2);
    PyTuple_SET_ITEM(__pyx_t_1, 89, __pyx_kp_u_logical_types_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->logical_types), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 90, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_adapt_record_types_2);
    __pyx_t_6 += 21;
    __Pyx_GIVEREF(__pyx_kp_u_adapt_record_types_2);
    PyTuple_SET_ITEM(__pyx_t_1, 91, __pyx_kp_u_adapt_record_types_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->adapt_record_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 92, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_return_uuid_object_2);
    __pyx_t_6 += 21;
    __Pyx_GIVEREF(__pyx_kp_u_return_uuid_object_2);
    PyTuple_SET_ITEM(__pyx_t_1, 93, __pyx_kp_u_return_uuid_object_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->return_uuid_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 94, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_allow_error_type_2);
    __pyx_t_6 += 19;
    __Pyx_GIVEREF(__pyx_kp_u_allow_error_type_2);
    PyTuple_SET_ITEM(__pyx_t_1, 95, __pyx_kp_u_allow_error_type_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->allow_error_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 96, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_allow_leading_dot_in_names_2);
    __pyx_t_6 += 29;
    __Pyx_GIVEREF(__pyx_kp_u_allow_leading_dot_in_names_2);
    PyTuple_SET_ITEM(__pyx_t_1, 97, __pyx_kp_u_allow_leading_dot_in_names_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->allow_leading_dot_in_names); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 98, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_naive_dt_assume_utc_2);
    __pyx_t_6 += 22;
    __Pyx_GIVEREF(__pyx_kp_u_naive_dt_assume_utc_2);
    PyTuple_SET_ITEM(__pyx_t_1, 99, __pyx_kp_u_naive_dt_assume_utc_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->naive_dt_assume_utc); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 100, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_alternate_timestamp_millis_enco);
    __pyx_t_6 += 38;
    __Pyx_GIVEREF(__pyx_kp_u_alternate_timestamp_millis_enco);
    PyTuple_SET_ITEM(__pyx_t_1, 101, __pyx_kp_u_alternate_timestamp_millis_enco);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->alternate_timestamp_millis_encoding); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 102, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_date_type_accepts_string_2);
    __pyx_t_6 += 27;
    __Pyx_GIVEREF(__pyx_kp_u_date_type_accepts_string_2);
    PyTuple_SET_ITEM(__pyx_t_1, 103, __pyx_kp_u_date_type_accepts_string_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->date_type_accepts_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 104, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_raise_on_invalid_logical_2);
    __pyx_t_6 += 27;
    __Pyx_GIVEREF(__pyx_kp_u_raise_on_invalid_logical_2);
    PyTuple_SET_ITEM(__pyx_t_1, 105, __pyx_kp_u_raise_on_invalid_logical_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->raise_on_invalid_logical); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 106, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_inline_namespaces_2);
    __pyx_t_6 += 20;
    __Pyx_GIVEREF(__pyx_kp_u_inline_namespaces_2);
    PyTuple_SET_ITEM(__pyx_t_1, 107, __pyx_kp_u_inline_namespaces_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->inline_namespaces); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 108, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_expand_types_in_schema_2);
    __pyx_t_6 += 25;
    __Pyx_GIVEREF(__pyx_kp_u_expand_types_in_schema_2);
    PyTuple_SET_ITEM(__pyx_t_1, 109, __pyx_kp_u_expand_types_in_schema_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->expand_types_in_schema); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 110, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_unicode_errors_2);
    __pyx_t_6 += 17;
    __Pyx_GIVEREF(__pyx_kp_u_unicode_errors_2);
    PyTuple_SET_ITEM(__pyx_t_1, 111, __pyx_kp_u_unicode_errors_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->unicode_errors), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 112, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_container_fill_blocks_2);
    __pyx_t_6 += 24;
    __Pyx_GIVEREF(__pyx_kp_u_container_fill_blocks_2);
    PyTuple_SET_ITEM(__pyx_t_1, 113, __pyx_kp_u_container_fill_blocks_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->container_fill_blocks); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 114, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_defer_schema_promotion_errors_2);
    __pyx_t_6 += 32;
    __Pyx_GIVEREF(__pyx_kp_u_defer_schema_promotion_errors_2);
    PyTuple_SET_ITEM(__pyx_t_1, 115, __pyx_kp_u_defer_schema_promotion_errors_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->defer_schema_promotion_errors); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 116, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_invalid_value_includes_record_n);
    __pyx_t_6 += 37;
    __Pyx_GIVEREF(__pyx_kp_u_invalid_value_includes_record_n);
    PyTuple_SET_ITEM(__pyx_t_1, 117, __pyx_kp_u_invalid_value_includes_record_n);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->invalid_value_includes_record_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 118, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_invalid_value_include_array_ind);
    __pyx_t_6 += 36;
    __Pyx_GIVEREF(__pyx_kp_u_invalid_value_include_array_ind);
    PyTuple_SET_ITEM(__pyx_t_1, 119, __pyx_kp_u_invalid_value_include_array_ind);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->invalid_value_include_array_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 120, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_invalid_value_include_map_key_2);
    __pyx_t_6 += 32;
    __Pyx_GIVEREF(__pyx_kp_u_invalid_value_include_map_key_2);
    PyTuple_SET_ITEM(__pyx_t_1, 121, __pyx_kp_u_invalid_value_include_map_key_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->invalid_value_include_map_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 122, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_allow_invalid_default_values_2);
    __pyx_t_6 += 31;
    __Pyx_GIVEREF(__pyx_kp_u_allow_invalid_default_values_2);
    PyTuple_SET_ITEM(__pyx_t_1, 123, __pyx_kp_u_allow_invalid_default_values_2);
    __pyx_t_2 = __Pyx_PyUnicode_FromBInt_bint(__pyx_v_self->allow_invalid_default_values); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 124, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_externally_defined_types_2);
    __pyx_t_6 += 27;
    __Pyx_GIVEREF(__pyx_kp_u_externally_defined_types_2);
    PyTuple_SET_ITEM(__pyx_t_1, 125, __pyx_kp_u_externally_defined_types_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->externally_defined_types), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 126, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u__7);
    __pyx_t_6 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__7);
    PyTuple_SET_ITEM(__pyx_t_1, 127, __pyx_kp_u__7);
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 128, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 74, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L4_return;
  }

  /* "(tree fragment)":76
 *         return f'{name}(fingerprint_returns_digest={self.fingerprint_returns_digest!r}, canonical_form_repeat_fixed={self.canonical_form_repeat_fixed!r}, canonical_form_repeat_enum={self.canonical_form_repeat_enum!r}, record_can_encode_dict={self.record_can_encode_dict!r}, record_values_type_hint={self.record_values_type_hint!r}, record_decodes_to_dict={self.record_decodes_to_dict!r}, record_allow_extra_fields={self.record_allow_extra_fields!r}, record_encode_use_defaults={self.record_encode_use_defaults!r}, allow_tuple_notation={self.allow_tuple_notation!r}, union_decodes_to={self.union_decodes_to!r}, union_json_encodes_type_name={self.union_json_encodes_type_name!r}, allow_primitive_name_collision={self.allow_primitive_name_collision!r}, allow_primitive_names_in_namespaces={self.allow_primitive_names_in_namespaces!r}, named_type_names_must_be_unique={self.named_type_names_must_be_unique!r}, enum_symbols_must_be_unique={self.enum_symbols_must_be_unique!r}, enforce_enum_symbol_name_rules={self.enforce_enum_symbol_name_rules!r}, enforce_type_name_rules={self.enforce_type_name_rules!r}, enforce_namespace_name_rules={self.enforce_namespace_name_rules!r}, record_fields_must_be_unique={self.record_fields_must_be_unique!r}, ascii_name_rules={self.ascii_name_rules!r}, missing_values_can_be_null={self.missing_values_can_be_null!r}, missing_values_can_be_empty_container={self.missing_values_can_be_empty_container!r}, allow_false_values_for_null={self.allow_false_values_for_null!r}, allow_empty_unions={self.allow_empty_unions!r}, allow_nested_unions={self.allow_nested_unions!r}, allow_duplicate_union_types={self.allow_duplicate_union_types!r}, allow_union_default_any_member={self.allow_union_default_any_member!r}, allow_aliases_to_be_string={self.allow_aliases_to_be_string!r}, coerce_values_to_int={self.coerce_values_to_int!r}, coerce_values_to_float={self.coerce_values_to_float!r}, coerce_int_to_float={self.coerce_int_to_float!r}, coerce_values_to_boolean={self.coerce_values_to_boolean!r}, coerce_values_to_str={self.coerce_values_to_str!r}, bytes_codec={self.bytes_codec!r}, fixed_codec={self.fixed_codec!r}, null_pad_fixed={self.null_pad_fixed!r}, truncate_fixed={self.truncate_fixed!r}, clamp_int_overflow={self.clamp_int_overflow!r}, clamp_float_overflow={self.clamp_float_overflow!r}, float_out_of_range_inf={self.float_out_of_range_inf!r}, bytes_default_value_utf8={self.bytes_default_value_utf8!r}, string_types_default_unchanged={self.string_types_default_unchanged!r}, decimal_check_exp_overflow={self.decimal_check_exp_overflow!r}, types_str_to_schema={self.types_str_to_schema!r}, logical_types={self.logical_types!r}, adapt_record_types={self.adapt_record_types!r}, return_uuid_object={self.return_uuid_object!r}, allow_error_type={self.allow_error_type!r}, allow_leading_dot_in_names={self.allow_leading_dot_in_names!r}, naive_dt_assume_utc={self.naive_dt_assume_utc!r}, alternate_timestamp_millis_encoding={self.alternate_timestamp_millis_encoding!r}, date_type_accepts_string={self.date_type_accepts_string!r}, raise_on_invalid_logical={self.raise_on_invalid_logical!r}, inline_namespaces={self.inline_namespaces!r}, expand_types_in_schema={self.expand_types_in_schema!r}, unicode_errors={self.unicode_errors!r}, container_fill_blocks={self.container_fill_blocks!r}, defer_schema_promotion_errors={self.defer_schema_promotion_errors!r}, invalid_value_includes_record_name={self.invalid_value_includes_record_name!r}, invalid_value_include_array_index={self.invalid_value_include_array_index!r}, invalid_value_include_map_key={self.invalid_value_include_map_key!r}, allow_invalid_default_values={self.allow_invalid_default_values!r}, externally_defined_types={self.externally_defined_types!r})'
 *     finally:
 *         guard_set.remove(key)             # <<<<<<<<<<<<<<
 * def __eq__(self, other):
 *     if not isinstance(other, Options):
 */
  /*finally:*/ {
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __pyx_t_5 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_9 = __pyx_filename;
      {
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_guard_set, __pyx_n_s_remove); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 76, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = NULL;
        __pyx_t_16 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_16 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_key};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_16, 1+__pyx_t_16);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 76, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_9;
      goto __pyx_L1_error;
      __pyx_L8_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L4_return: {
      __pyx_t_15 = __pyx_r;
      __pyx_r = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_guard_set, __pyx_n_s_remove); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_key};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 76, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_15;
      __pyx_t_15 = 0;
      goto __pyx_L0;
    }
  }

  /* "(tree fragment)":67
 * __pyx_recursive_repr_guard = __import__('threading').local()
 * __pyx_recursive_repr_guard.running = set()
 * def __repr__(self):             # <<<<<<<<<<<<<<
 *     key = id(self)
 *     guard_set = self.__pyx_recursive_repr_guard.running
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Options.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_guard_set);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":77
 *     finally:
 *         guard_set.remove(key)
 * def __eq__(self, other):             # <<<<<<<<<<<<<<
 *     if not isinstance(other, Options):
 *         return NotImplemented
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_15__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_15__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_14__eq__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_14__eq__(struct __pyx_obj_5cavro_Options *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_5cavro_Options *__pyx_v_other_cast = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 1);

  /* "(tree fragment)":78
 *         guard_set.remove(key)
 * def __eq__(self, other):
 *     if not isinstance(other, Options):             # <<<<<<<<<<<<<<
 *         return NotImplemented
 *     cdef Options other_cast
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_5cavro_Options); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "(tree fragment)":79
 * def __eq__(self, other):
 *     if not isinstance(other, Options):
 *         return NotImplemented             # <<<<<<<<<<<<<<
 *     cdef Options other_cast
 *     other_cast = <Options>other
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;

    /* "(tree fragment)":78
 *         guard_set.remove(key)
 * def __eq__(self, other):
 *     if not isinstance(other, Options):             # <<<<<<<<<<<<<<
 *         return NotImplemented
 *     cdef Options other_cast
 */
  }

  /* "(tree fragment)":81
 *         return NotImplemented
 *     cdef Options other_cast
 *     other_cast = <Options>other             # <<<<<<<<<<<<<<
 *     return (self.fingerprint_returns_digest == other_cast.fingerprint_returns_digest) and (self.canonical_form_repeat_fixed == other_cast.canonical_form_repeat_fixed) and (self.canonical_form_repeat_enum == other_cast.canonical_form_repeat_enum) and (self.record_can_encode_dict == other_cast.record_can_encode_dict) and (self.record_values_type_hint == other_cast.record_values_type_hint) and (self.record_decodes_to_dict == other_cast.record_decodes_to_dict) and (self.record_allow_extra_fields == other_cast.record_allow_extra_fields) and (self.record_encode_use_defaults == other_cast.record_encode_use_defaults) and (self.allow_tuple_notation == other_cast.allow_tuple_notation) and (self.union_decodes_to == other_cast.union_decodes_to) and (self.union_json_encodes_type_name == other_cast.union_json_encodes_type_name) and (self.allow_primitive_name_collision == other_cast.allow_primitive_name_collision) and (self.allow_primitive_names_in_namespaces == other_cast.allow_primitive_names_in_namespaces) and (self.named_type_names_must_be_unique == other_cast.named_type_names_must_be_unique) and (self.enum_symbols_must_be_unique == other_cast.enum_symbols_must_be_unique) and (self.enforce_enum_symbol_name_rules == other_cast.enforce_enum_symbol_name_rules) and (self.enforce_type_name_rules == other_cast.enforce_type_name_rules) and (self.enforce_namespace_name_rules == other_cast.enforce_namespace_name_rules) and (self.record_fields_must_be_unique == other_cast.record_fields_must_be_unique) and (self.ascii_name_rules == other_cast.ascii_name_rules) and (self.missing_values_can_be_null == other_cast.missing_values_can_be_null) and (self.missing_values_can_be_empty_container == other_cast.missing_values_can_be_empty_container) and (self.allow_false_values_for_null == other_cast.allow_false_values_for_null) and (self.allow_empty_unions == other_cast.allow_empty_unions) and (self.allow_nested_unions == other_cast.allow_nested_unions) and (self.allow_duplicate_union_types == other_cast.allow_duplicate_union_types) and (self.allow_union_default_any_member == other_cast.allow_union_default_any_member) and (self.allow_aliases_to_be_string == other_cast.allow_aliases_to_be_string) and (self.coerce_values_to_int == other_cast.coerce_values_to_int) and (self.coerce_values_to_float == other_cast.coerce_values_to_float) and (self.coerce_int_to_float == other_cast.coerce_int_to_float) and (self.coerce_values_to_boolean == other_cast.coerce_values_to_boolean) and (self.coerce_values_to_str == other_cast.coerce_values_to_str) and (self.bytes_codec == other_cast.bytes_codec) and (self.fixed_codec == other_cast.fixed_codec) and (self.null_pad_fixed == other_cast.null_pad_fixed) and (self.truncate_fixed == other_cast.truncate_fixed) and (self.clamp_int_overflow == other_cast.clamp_int_overflow) and (self.clamp_float_overflow == other_cast.clamp_float_overflow) and (self.float_out_of_range_inf == other_cast.float_out_of_range_inf) and (self.bytes_default_value_utf8 == other_cast.bytes_default_value_utf8) and (self.string_types_default_unchanged == other_cast.string_types_default_unchanged) and (self.decimal_check_exp_overflow == other_cast.decimal_check_exp_overflow) and (self.types_str_to_schema == other_cast.types_str_to_schema) and (self.logical_types == other_cast.logical_types) and (self.adapt_record_types == other_cast.adapt_record_types) and (self.return_uuid_object == other_cast.return_uuid_object) and (self.allow_error_type == other_cast.allow_error_type) and (self.allow_leading_dot_in_names == other_cast.allow_leading_dot_in_names) and (self.naive_dt_assume_utc == other_cast.naive_dt_assume_utc) and (self.alternate_timestamp_millis_encoding == other_cast.alternate_timestamp_millis_encoding) and (self.date_type_accepts_string == other_cast.date_type_accepts_string) and (self.raise_on_invalid_logical == other_cast.raise_on_invalid_logical) and (self.inline_namespaces == other_cast.inline_namespaces) and (self.expand_types_in_schema == other_cast.expand_types_in_schema) and (self.unicode_errors == other_cast.unicode_errors) and (self.container_fill_blocks == other_cast.container_fill_blocks) and (self.defer_schema_promotion_errors == other_cast.defer_schema_promotion_errors) and (self.invalid_value_includes_record_name == other_cast.invalid_value_includes_record_name) and (self.invalid_value_include_array_index == other_cast.invalid_value_include_array_index) and (self.invalid_value_include_map_key == other_cast.invalid_value_include_map_key) and (self.allow_invalid_default_values == other_cast.allow_invalid_default_values) and (self.externally_defined_types == other_cast.externally_defined_types)
 * def __hash__(self):
 */
  __pyx_t_3 = __pyx_v_other;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_other_cast = ((struct __pyx_obj_5cavro_Options *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":82
 *     cdef Options other_cast
 *     other_cast = <Options>other
 *     return (self.fingerprint_returns_digest == other_cast.fingerprint_returns_digest) and (self.canonical_form_repeat_fixed == other_cast.canonical_form_repeat_fixed) and (self.canonical_form_repeat_enum == other_cast.canonical_form_repeat_enum) and (self.record_can_encode_dict == other_cast.record_can_encode_dict) and (self.record_values_type_hint == other_cast.record_values_type_hint) and (self.record_decodes_to_dict == other_cast.record_decodes_to_dict) and (self.record_allow_extra_fields == other_cast.record_allow_extra_fields) and (self.record_encode_use_defaults == other_cast.record_encode_use_defaults) and (self.allow_tuple_notation == other_cast.allow_tuple_notation) and (self.union_decodes_to == other_cast.union_decodes_to) and (self.union_json_encodes_type_name == other_cast.union_json_encodes_type_name) and (self.allow_primitive_name_collision == other_cast.allow_primitive_name_collision) and (self.allow_primitive_names_in_namespaces == other_cast.allow_primitive_names_in_namespaces) and (self.named_type_names_must_be_unique == other_cast.named_type_names_must_be_unique) and (self.enum_symbols_must_be_unique == other_cast.enum_symbols_must_be_unique) and (self.enforce_enum_symbol_name_rules == other_cast.enforce_enum_symbol_name_rules) and (self.enforce_type_name_rules == other_cast.enforce_type_name_rules) and (self.enforce_namespace_name_rules == other_cast.enforce_namespace_name_rules) and (self.record_fields_must_be_unique == other_cast.record_fields_must_be_unique) and (self.ascii_name_rules == other_cast.ascii_name_rules) and (self.missing_values_can_be_null == other_cast.missing_values_can_be_null) and (self.missing_values_can_be_empty_container == other_cast.missing_values_can_be_empty_container) and (self.allow_false_values_for_null == other_cast.allow_false_values_for_null) and (self.allow_empty_unions == other_cast.allow_empty_unions) and (self.allow_nested_unions == other_cast.allow_nested_unions) and (self.allow_duplicate_union_types == other_cast.allow_duplicate_union_types) and (self.allow_union_default_any_member == other_cast.allow_union_default_any_member) and (self.allow_aliases_to_be_string == other_cast.allow_aliases_to_be_string) and (self.coerce_values_to_int == other_cast.coerce_values_to_int) and (self.coerce_values_to_float == other_cast.coerce_values_to_float) and (self.coerce_int_to_float == other_cast.coerce_int_to_float) and (self.coerce_values_to_boolean == other_cast.coerce_values_to_boolean) and (self.coerce_values_to_str == other_cast.coerce_values_to_str) and (self.bytes_codec == other_cast.bytes_codec) and (self.fixed_codec == other_cast.fixed_codec) and (self.null_pad_fixed == other_cast.null_pad_fixed) and (self.truncate_fixed == other_cast.truncate_fixed) and (self.clamp_int_overflow == other_cast.clamp_int_overflow) and (self.clamp_float_overflow == other_cast.clamp_float_overflow) and (self.float_out_of_range_inf == other_cast.float_out_of_range_inf) and (self.bytes_default_value_utf8 == other_cast.bytes_default_value_utf8) and (self.string_types_default_unchanged == other_cast.string_types_default_unchanged) and (self.decimal_check_exp_overflow == other_cast.decimal_check_exp_overflow) and (self.types_str_to_schema == other_cast.types_str_to_schema) and (self.logical_types == other_cast.logical_types) and (self.adapt_record_types == other_cast.adapt_record_types) and (self.return_uuid_object == other_cast.return_uuid_object) and (self.allow_error_type == other_cast.allow_error_type) and (self.allow_leading_dot_in_names == other_cast.allow_leading_dot_in_names) and (self.naive_dt_assume_utc == other_cast.naive_dt_assume_utc) and (self.alternate_timestamp_millis_encoding == other_cast.alternate_timestamp_millis_encoding) and (self.date_type_accepts_string == other_cast.date_type_accepts_string) and (self.raise_on_invalid_logical == other_cast.raise_on_invalid_logical) and (self.inline_namespaces == other_cast.inline_namespaces) and (self.expand_types_in_schema == other_cast.expand_types_in_schema) and (self.unicode_errors == other_cast.unicode_errors) and (self.container_fill_blocks == other_cast.container_fill_blocks) and (self.defer_schema_promotion_errors == other_cast.defer_schema_promotion_errors) and (self.invalid_value_includes_record_name == other_cast.invalid_value_includes_record_name) and (self.invalid_value_include_array_index == other_cast.invalid_value_include_array_index) and (self.invalid_value_include_map_key == other_cast.invalid_value_include_map_key) and (self.allow_invalid_default_values == other_cast.allow_invalid_default_values) and (self.externally_defined_types == other_cast.externally_defined_types)             # <<<<<<<<<<<<<<
 * def __hash__(self):
 *     return hash((self.fingerprint_returns_digest, self.canonical_form_repeat_fixed, self.canonical_form_repeat_enum, self.record_can_encode_dict, self.record_values_type_hint, self.record_decodes_to_dict, self.record_allow_extra_fields, self.record_encode_use_defaults, self.allow_tuple_notation, self.union_decodes_to, self.union_json_encodes_type_name, self.allow_primitive_name_collision, self.allow_primitive_names_in_namespaces, self.named_type_names_must_be_unique, self.enum_symbols_must_be_unique, self.enforce_enum_symbol_name_rules, self.enforce_type_name_rules, self.enforce_namespace_name_rules, self.record_fields_must_be_unique, self.ascii_name_rules, self.missing_values_can_be_null, self.missing_values_can_be_empty_container, self.allow_false_values_for_null, self.allow_empty_unions, self.allow_nested_unions, self.allow_duplicate_union_types, self.allow_union_default_any_member, self.allow_aliases_to_be_string, self.coerce_values_to_int, self.coerce_values_to_float, self.coerce_int_to_float, self.coerce_values_to_boolean, self.coerce_values_to_str, self.bytes_codec, self.fixed_codec, self.null_pad_fixed, self.truncate_fixed, self.clamp_int_overflow, self.clamp_float_overflow, self.float_out_of_range_inf, self.bytes_default_value_utf8, self.string_types_default_unchanged, self.decimal_check_exp_overflow, self.types_str_to_schema, self.logical_types, self.adapt_record_types, self.return_uuid_object, self.allow_error_type, self.allow_leading_dot_in_names, self.naive_dt_assume_utc, self.alternate_timestamp_millis_encoding, self.date_type_accepts_string, self.raise_on_invalid_logical, self.inline_namespaces, self.expand_types_in_schema, self.unicode_errors, self.container_fill_blocks, self.defer_schema_promotion_errors, self.invalid_value_includes_record_name, self.invalid_value_include_array_index, self.invalid_value_include_map_key, self.allow_invalid_default_values, self.externally_defined_types,))
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = (__pyx_v_self->fingerprint_returns_digest == __pyx_v_other_cast->fingerprint_returns_digest);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->canonical_form_repeat_fixed == __pyx_v_other_cast->canonical_form_repeat_fixed);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->canonical_form_repeat_enum == __pyx_v_other_cast->canonical_form_repeat_enum);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->record_can_encode_dict == __pyx_v_other_cast->record_can_encode_dict);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->record_values_type_hint == __pyx_v_other_cast->record_values_type_hint);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->record_decodes_to_dict == __pyx_v_other_cast->record_decodes_to_dict);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->record_allow_extra_fields == __pyx_v_other_cast->record_allow_extra_fields);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->record_encode_use_defaults == __pyx_v_other_cast->record_encode_use_defaults);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->allow_tuple_notation == __pyx_v_other_cast->allow_tuple_notation);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_self->union_decodes_to, __pyx_v_other_cast->union_decodes_to, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 82, __pyx_L1_error)
  if (__pyx_t_2) {
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->union_json_encodes_type_name == __pyx_v_other_cast->union_json_encodes_type_name);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->allow_primitive_name_collision == __pyx_v_other_cast->allow_primitive_name_collision);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->allow_primitive_names_in_namespaces == __pyx_v_other_cast->allow_primitive_names_in_namespaces);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->named_type_names_must_be_unique == __pyx_v_other_cast->named_type_names_must_be_unique);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->enum_symbols_must_be_unique == __pyx_v_other_cast->enum_symbols_must_be_unique);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->enforce_enum_symbol_name_rules == __pyx_v_other_cast->enforce_enum_symbol_name_rules);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->enforce_type_name_rules == __pyx_v_other_cast->enforce_type_name_rules);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->enforce_namespace_name_rules == __pyx_v_other_cast->enforce_namespace_name_rules);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->record_fields_must_be_unique == __pyx_v_other_cast->record_fields_must_be_unique);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->ascii_name_rules == __pyx_v_other_cast->ascii_name_rules);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->missing_values_can_be_null == __pyx_v_other_cast->missing_values_can_be_null);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->missing_values_can_be_empty_container == __pyx_v_other_cast->missing_values_can_be_empty_container);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->allow_false_values_for_null == __pyx_v_other_cast->allow_false_values_for_null);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->allow_empty_unions == __pyx_v_other_cast->allow_empty_unions);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->allow_nested_unions == __pyx_v_other_cast->allow_nested_unions);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->allow_duplicate_union_types == __pyx_v_other_cast->allow_duplicate_union_types);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->allow_union_default_any_member == __pyx_v_other_cast->allow_union_default_any_member);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->allow_aliases_to_be_string == __pyx_v_other_cast->allow_aliases_to_be_string);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->coerce_values_to_int == __pyx_v_other_cast->coerce_values_to_int);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->coerce_values_to_float == __pyx_v_other_cast->coerce_values_to_float);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->coerce_int_to_float == __pyx_v_other_cast->coerce_int_to_float);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->coerce_values_to_boolean == __pyx_v_other_cast->coerce_values_to_boolean);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->coerce_values_to_str == __pyx_v_other_cast->coerce_values_to_str);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->bytes_codec, __pyx_v_other_cast->bytes_codec, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 82, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->fixed_codec, __pyx_v_other_cast->fixed_codec, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 82, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->null_pad_fixed == __pyx_v_other_cast->null_pad_fixed);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->truncate_fixed == __pyx_v_other_cast->truncate_fixed);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->clamp_int_overflow == __pyx_v_other_cast->clamp_int_overflow);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->clamp_float_overflow == __pyx_v_other_cast->clamp_float_overflow);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->float_out_of_range_inf == __pyx_v_other_cast->float_out_of_range_inf);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->bytes_default_value_utf8 == __pyx_v_other_cast->bytes_default_value_utf8);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->string_types_default_unchanged == __pyx_v_other_cast->string_types_default_unchanged);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->decimal_check_exp_overflow == __pyx_v_other_cast->decimal_check_exp_overflow);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->types_str_to_schema == __pyx_v_other_cast->types_str_to_schema);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_self->logical_types, __pyx_v_other_cast->logical_types, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 82, __pyx_L1_error)
  if (__pyx_t_2) {
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->adapt_record_types == __pyx_v_other_cast->adapt_record_types);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->return_uuid_object == __pyx_v_other_cast->return_uuid_object);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->allow_error_type == __pyx_v_other_cast->allow_error_type);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->allow_leading_dot_in_names == __pyx_v_other_cast->allow_leading_dot_in_names);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->naive_dt_assume_utc == __pyx_v_other_cast->naive_dt_assume_utc);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->alternate_timestamp_millis_encoding == __pyx_v_other_cast->alternate_timestamp_millis_encoding);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->date_type_accepts_string == __pyx_v_other_cast->date_type_accepts_string);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->raise_on_invalid_logical == __pyx_v_other_cast->raise_on_invalid_logical);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->inline_namespaces == __pyx_v_other_cast->inline_namespaces);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->expand_types_in_schema == __pyx_v_other_cast->expand_types_in_schema);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->unicode_errors, __pyx_v_other_cast->unicode_errors, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 82, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->container_fill_blocks == __pyx_v_other_cast->container_fill_blocks);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->defer_schema_promotion_errors == __pyx_v_other_cast->defer_schema_promotion_errors);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->invalid_value_includes_record_name == __pyx_v_other_cast->invalid_value_includes_record_name);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->invalid_value_include_array_index == __pyx_v_other_cast->invalid_value_include_array_index);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->invalid_value_include_map_key == __pyx_v_other_cast->invalid_value_include_map_key);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->allow_invalid_default_values == __pyx_v_other_cast->allow_invalid_default_values);
  if (__pyx_t_2) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_self->externally_defined_types, __pyx_v_other_cast->externally_defined_types, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 82, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_L4_bool_binop_done:;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "(tree fragment)":77
 *     finally:
 *         guard_set.remove(key)
 * def __eq__(self, other):             # <<<<<<<<<<<<<<
 *     if not isinstance(other, Options):
 *         return NotImplemented
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Options.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_other_cast);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":83
 *     other_cast = <Options>other
 *     return (self.fingerprint_returns_digest == other_cast.fingerprint_returns_digest) and (self.canonical_form_repeat_fixed == other_cast.canonical_form_repeat_fixed) and (self.canonical_form_repeat_enum == other_cast.canonical_form_repeat_enum) and (self.record_can_encode_dict == other_cast.record_can_encode_dict) and (self.record_values_type_hint == other_cast.record_values_type_hint) and (self.record_decodes_to_dict == other_cast.record_decodes_to_dict) and (self.record_allow_extra_fields == other_cast.record_allow_extra_fields) and (self.record_encode_use_defaults == other_cast.record_encode_use_defaults) and (self.allow_tuple_notation == other_cast.allow_tuple_notation) and (self.union_decodes_to == other_cast.union_decodes_to) and (self.union_json_encodes_type_name == other_cast.union_json_encodes_type_name) and (self.allow_primitive_name_collision == other_cast.allow_primitive_name_collision) and (self.allow_primitive_names_in_namespaces == other_cast.allow_primitive_names_in_namespaces) and (self.named_type_names_must_be_unique == other_cast.named_type_names_must_be_unique) and (self.enum_symbols_must_be_unique == other_cast.enum_symbols_must_be_unique) and (self.enforce_enum_symbol_name_rules == other_cast.enforce_enum_symbol_name_rules) and (self.enforce_type_name_rules == other_cast.enforce_type_name_rules) and (self.enforce_namespace_name_rules == other_cast.enforce_namespace_name_rules) and (self.record_fields_must_be_unique == other_cast.record_fields_must_be_unique) and (self.ascii_name_rules == other_cast.ascii_name_rules) and (self.missing_values_can_be_null == other_cast.missing_values_can_be_null) and (self.missing_values_can_be_empty_container == other_cast.missing_values_can_be_empty_container) and (self.allow_false_values_for_null == other_cast.allow_false_values_for_null) and (self.allow_empty_unions == other_cast.allow_empty_unions) and (self.allow_nested_unions == other_cast.allow_nested_unions) and (self.allow_duplicate_union_types == other_cast.allow_duplicate_union_types) and (self.allow_union_default_any_member == other_cast.allow_union_default_any_member) and (self.allow_aliases_to_be_string == other_cast.allow_aliases_to_be_string) and (self.coerce_values_to_int == other_cast.coerce_values_to_int) and (self.coerce_values_to_float == other_cast.coerce_values_to_float) and (self.coerce_int_to_float == other_cast.coerce_int_to_float) and (self.coerce_values_to_boolean == other_cast.coerce_values_to_boolean) and (self.coerce_values_to_str == other_cast.coerce_values_to_str) and (self.bytes_codec == other_cast.bytes_codec) and (self.fixed_codec == other_cast.fixed_codec) and (self.null_pad_fixed == other_cast.null_pad_fixed) and (self.truncate_fixed == other_cast.truncate_fixed) and (self.clamp_int_overflow == other_cast.clamp_int_overflow) and (self.clamp_float_overflow == other_cast.clamp_float_overflow) and (self.float_out_of_range_inf == other_cast.float_out_of_range_inf) and (self.bytes_default_value_utf8 == other_cast.bytes_default_value_utf8) and (self.string_types_default_unchanged == other_cast.string_types_default_unchanged) and (self.decimal_check_exp_overflow == other_cast.decimal_check_exp_overflow) and (self.types_str_to_schema == other_cast.types_str_to_schema) and (self.logical_types == other_cast.logical_types) and (self.adapt_record_types == other_cast.adapt_record_types) and (self.return_uuid_object == other_cast.return_uuid_object) and (self.allow_error_type == other_cast.allow_error_type) and (self.allow_leading_dot_in_names == other_cast.allow_leading_dot_in_names) and (self.naive_dt_assume_utc == other_cast.naive_dt_assume_utc) and (self.alternate_timestamp_millis_encoding == other_cast.alternate_timestamp_millis_encoding) and (self.date_type_accepts_string == other_cast.date_type_accepts_string) and (self.raise_on_invalid_logical == other_cast.raise_on_invalid_logical) and (self.inline_namespaces == other_cast.inline_namespaces) and (self.expand_types_in_schema == other_cast.expand_types_in_schema) and (self.unicode_errors == other_cast.unicode_errors) and (self.container_fill_blocks == other_cast.container_fill_blocks) and (self.defer_schema_promotion_errors == other_cast.defer_schema_promotion_errors) and (self.invalid_value_includes_record_name == other_cast.invalid_value_includes_record_name) and (self.invalid_value_include_array_index == other_cast.invalid_value_include_array_index) and (self.invalid_value_include_map_key == other_cast.invalid_value_include_map_key) and (self.allow_invalid_default_values == other_cast.allow_invalid_default_values) and (self.externally_defined_types == other_cast.externally_defined_types)
 * def __hash__(self):             # <<<<<<<<<<<<<<
 *     return hash((self.fingerprint_returns_digest, self.canonical_form_repeat_fixed, self.canonical_form_repeat_enum, self.record_can_encode_dict, self.record_values_type_hint, self.record_decodes_to_dict, self.record_allow_extra_fields, self.record_encode_use_defaults, self.allow_tuple_notation, self.union_decodes_to, self.union_json_encodes_type_name, self.allow_primitive_name_collision, self.allow_primitive_names_in_namespaces, self.named_type_names_must_be_unique, self.enum_symbols_must_be_unique, self.enforce_enum_symbol_name_rules, self.enforce_type_name_rules, self.enforce_namespace_name_rules, self.record_fields_must_be_unique, self.ascii_name_rules, self.missing_values_can_be_null, self.missing_values_can_be_empty_container, self.allow_false_values_for_null, self.allow_empty_unions, self.allow_nested_unions, self.allow_duplicate_union_types, self.allow_union_default_any_member, self.allow_aliases_to_be_string, self.coerce_values_to_int, self.coerce_values_to_float, self.coerce_int_to_float, self.coerce_values_to_boolean, self.coerce_values_to_str, self.bytes_codec, self.fixed_codec, self.null_pad_fixed, self.truncate_fixed, self.clamp_int_overflow, self.clamp_float_overflow, self.float_out_of_range_inf, self.bytes_default_value_utf8, self.string_types_default_unchanged, self.decimal_check_exp_overflow, self.types_str_to_schema, self.logical_types, self.adapt_record_types, self.return_uuid_object, self.allow_error_type, self.allow_leading_dot_in_names, self.naive_dt_assume_utc, self.alternate_timestamp_millis_encoding, self.date_type_accepts_string, self.raise_on_invalid_logical, self.inline_namespaces, self.expand_types_in_schema, self.unicode_errors, self.container_fill_blocks, self.defer_schema_promotion_errors, self.invalid_value_includes_record_name, self.invalid_value_include_array_index, self.invalid_value_include_map_key, self.allow_invalid_default_values, self.externally_defined_types,))
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_5cavro_7Options_17__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_5cavro_7Options_17__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_16__hash__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_5cavro_7Options_16__hash__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  PyObject *__pyx_t_36 = NULL;
  PyObject *__pyx_t_37 = NULL;
  PyObject *__pyx_t_38 = NULL;
  PyObject *__pyx_t_39 = NULL;
  PyObject *__pyx_t_40 = NULL;
  PyObject *__pyx_t_41 = NULL;
  PyObject *__pyx_t_42 = NULL;
  PyObject *__pyx_t_43 = NULL;
  PyObject *__pyx_t_44 = NULL;
  PyObject *__pyx_t_45 = NULL;
  PyObject *__pyx_t_46 = NULL;
  PyObject *__pyx_t_47 = NULL;
  PyObject *__pyx_t_48 = NULL;
  PyObject *__pyx_t_49 = NULL;
  PyObject *__pyx_t_50 = NULL;
  PyObject *__pyx_t_51 = NULL;
  PyObject *__pyx_t_52 = NULL;
  PyObject *__pyx_t_53 = NULL;
  PyObject *__pyx_t_54 = NULL;
  PyObject *__pyx_t_55 = NULL;
  PyObject *__pyx_t_56 = NULL;
  PyObject *__pyx_t_57 = NULL;
  PyObject *__pyx_t_58 = NULL;
  Py_hash_t __pyx_t_59;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 1);

  /* "(tree fragment)":84
 *     return (self.fingerprint_returns_digest == other_cast.fingerprint_returns_digest) and (self.canonical_form_repeat_fixed == other_cast.canonical_form_repeat_fixed) and (self.canonical_form_repeat_enum == other_cast.canonical_form_repeat_enum) and (self.record_can_encode_dict == other_cast.record_can_encode_dict) and (self.record_values_type_hint == other_cast.record_values_type_hint) and (self.record_decodes_to_dict == other_cast.record_decodes_to_dict) and (self.record_allow_extra_fields == other_cast.record_allow_extra_fields) and (self.record_encode_use_defaults == other_cast.record_encode_use_defaults) and (self.allow_tuple_notation == other_cast.allow_tuple_notation) and (self.union_decodes_to == other_cast.union_decodes_to) and (self.union_json_encodes_type_name == other_cast.union_json_encodes_type_name) and (self.allow_primitive_name_collision == other_cast.allow_primitive_name_collision) and (self.allow_primitive_names_in_namespaces == other_cast.allow_primitive_names_in_namespaces) and (self.named_type_names_must_be_unique == other_cast.named_type_names_must_be_unique) and (self.enum_symbols_must_be_unique == other_cast.enum_symbols_must_be_unique) and (self.enforce_enum_symbol_name_rules == other_cast.enforce_enum_symbol_name_rules) and (self.enforce_type_name_rules == other_cast.enforce_type_name_rules) and (self.enforce_namespace_name_rules == other_cast.enforce_namespace_name_rules) and (self.record_fields_must_be_unique == other_cast.record_fields_must_be_unique) and (self.ascii_name_rules == other_cast.ascii_name_rules) and (self.missing_values_can_be_null == other_cast.missing_values_can_be_null) and (self.missing_values_can_be_empty_container == other_cast.missing_values_can_be_empty_container) and (self.allow_false_values_for_null == other_cast.allow_false_values_for_null) and (self.allow_empty_unions == other_cast.allow_empty_unions) and (self.allow_nested_unions == other_cast.allow_nested_unions) and (self.allow_duplicate_union_types == other_cast.allow_duplicate_union_types) and (self.allow_union_default_any_member == other_cast.allow_union_default_any_member) and (self.allow_aliases_to_be_string == other_cast.allow_aliases_to_be_string) and (self.coerce_values_to_int == other_cast.coerce_values_to_int) and (self.coerce_values_to_float == other_cast.coerce_values_to_float) and (self.coerce_int_to_float == other_cast.coerce_int_to_float) and (self.coerce_values_to_boolean == other_cast.coerce_values_to_boolean) and (self.coerce_values_to_str == other_cast.coerce_values_to_str) and (self.bytes_codec == other_cast.bytes_codec) and (self.fixed_codec == other_cast.fixed_codec) and (self.null_pad_fixed == other_cast.null_pad_fixed) and (self.truncate_fixed == other_cast.truncate_fixed) and (self.clamp_int_overflow == other_cast.clamp_int_overflow) and (self.clamp_float_overflow == other_cast.clamp_float_overflow) and (self.float_out_of_range_inf == other_cast.float_out_of_range_inf) and (self.bytes_default_value_utf8 == other_cast.bytes_default_value_utf8) and (self.string_types_default_unchanged == other_cast.string_types_default_unchanged) and (self.decimal_check_exp_overflow == other_cast.decimal_check_exp_overflow) and (self.types_str_to_schema == other_cast.types_str_to_schema) and (self.logical_types == other_cast.logical_types) and (self.adapt_record_types == other_cast.adapt_record_types) and (self.return_uuid_object == other_cast.return_uuid_object) and (self.allow_error_type == other_cast.allow_error_type) and (self.allow_leading_dot_in_names == other_cast.allow_leading_dot_in_names) and (self.naive_dt_assume_utc == other_cast.naive_dt_assume_utc) and (self.alternate_timestamp_millis_encoding == other_cast.alternate_timestamp_millis_encoding) and (self.date_type_accepts_string == other_cast.date_type_accepts_string) and (self.raise_on_invalid_logical == other_cast.raise_on_invalid_logical) and (self.inline_namespaces == other_cast.inline_namespaces) and (self.expand_types_in_schema == other_cast.expand_types_in_schema) and (self.unicode_errors == other_cast.unicode_errors) and (self.container_fill_blocks == other_cast.container_fill_blocks) and (self.defer_schema_promotion_errors == other_cast.defer_schema_promotion_errors) and (self.invalid_value_includes_record_name == other_cast.invalid_value_includes_record_name) and (self.invalid_value_include_array_index == other_cast.invalid_value_include_array_index) and (self.invalid_value_include_map_key == other_cast.invalid_value_include_map_key) and (self.allow_invalid_default_values == other_cast.allow_invalid_default_values) and (self.externally_defined_types == other_cast.externally_defined_types)
 * def __hash__(self):
 *     return hash((self.fingerprint_returns_digest, self.canonical_form_repeat_fixed, self.canonical_form_repeat_enum, self.record_can_encode_dict, self.record_values_type_hint, self.record_decodes_to_dict, self.record_allow_extra_fields, self.record_encode_use_defaults, self.allow_tuple_notation, self.union_decodes_to, self.union_json_encodes_type_name, self.allow_primitive_name_collision, self.allow_primitive_names_in_namespaces, self.named_type_names_must_be_unique, self.enum_symbols_must_be_unique, self.enforce_enum_symbol_name_rules, self.enforce_type_name_rules, self.enforce_namespace_name_rules, self.record_fields_must_be_unique, self.ascii_name_rules, self.missing_values_can_be_null, self.missing_values_can_be_empty_container, self.allow_false_values_for_null, self.allow_empty_unions, self.allow_nested_unions, self.allow_duplicate_union_types, self.allow_union_default_any_member, self.allow_aliases_to_be_string, self.coerce_values_to_int, self.coerce_values_to_float, self.coerce_int_to_float, self.coerce_values_to_boolean, self.coerce_values_to_str, self.bytes_codec, self.fixed_codec, self.null_pad_fixed, self.truncate_fixed, self.clamp_int_overflow, self.clamp_float_overflow, self.float_out_of_range_inf, self.bytes_default_value_utf8, self.string_types_default_unchanged, self.decimal_check_exp_overflow, self.types_str_to_schema, self.logical_types, self.adapt_record_types, self.return_uuid_object, self.allow_error_type, self.allow_leading_dot_in_names, self.naive_dt_assume_utc, self.alternate_timestamp_millis_encoding, self.date_type_accepts_string, self.raise_on_invalid_logical, self.inline_namespaces, self.expand_types_in_schema, self.unicode_errors, self.container_fill_blocks, self.defer_schema_promotion_errors, self.invalid_value_includes_record_name, self.invalid_value_include_array_index, self.invalid_value_include_map_key, self.allow_invalid_default_values, self.externally_defined_types,))             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->fingerprint_returns_digest); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->canonical_form_repeat_fixed); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->canonical_form_repeat_enum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->record_can_encode_dict); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->record_values_type_hint); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->record_decodes_to_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->record_allow_extra_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyBool_FromLong(__pyx_v_self->record_encode_use_defaults); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_tuple_notation); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyBool_FromLong(__pyx_v_self->union_json_encodes_type_name); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_primitive_name_collision); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_primitive_names_in_namespaces); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyBool_FromLong(__pyx_v_self->named_type_names_must_be_unique); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_14 = __Pyx_PyBool_FromLong(__pyx_v_self->enum_symbols_must_be_unique); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __Pyx_PyBool_FromLong(__pyx_v_self->enforce_enum_symbol_name_rules); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = __Pyx_PyBool_FromLong(__pyx_v_self->enforce_type_name_rules); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __Pyx_PyBool_FromLong(__pyx_v_self->enforce_namespace_name_rules); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_18 = __Pyx_PyBool_FromLong(__pyx_v_self->record_fields_must_be_unique); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_19 = __Pyx_PyBool_FromLong(__pyx_v_self->ascii_name_rules); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_20 = __Pyx_PyBool_FromLong(__pyx_v_self->missing_values_can_be_null); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = __Pyx_PyBool_FromLong(__pyx_v_self->missing_values_can_be_empty_container); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_22 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_false_values_for_null); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_23 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_empty_unions); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_24 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_nested_unions); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_25 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_duplicate_union_types); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_26 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_union_default_any_member); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_27 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_aliases_to_be_string); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_28 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_values_to_int); if (unlikely(!__pyx_t_28)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __pyx_t_29 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_values_to_float); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_30 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_int_to_float); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_31 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_values_to_boolean); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_32 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_values_to_str); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_33 = __Pyx_PyBool_FromLong(__pyx_v_self->null_pad_fixed); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = __Pyx_PyBool_FromLong(__pyx_v_self->truncate_fixed); if (unlikely(!__pyx_t_34)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_35 = __Pyx_PyBool_FromLong(__pyx_v_self->clamp_int_overflow); if (unlikely(!__pyx_t_35)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __pyx_t_36 = __Pyx_PyBool_FromLong(__pyx_v_self->clamp_float_overflow); if (unlikely(!__pyx_t_36)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __pyx_t_37 = __Pyx_PyBool_FromLong(__pyx_v_self->float_out_of_range_inf); if (unlikely(!__pyx_t_37)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_38 = __Pyx_PyBool_FromLong(__pyx_v_self->bytes_default_value_utf8); if (unlikely(!__pyx_t_38)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __pyx_t_39 = __Pyx_PyBool_FromLong(__pyx_v_self->string_types_default_unchanged); if (unlikely(!__pyx_t_39)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __pyx_t_40 = __Pyx_PyBool_FromLong(__pyx_v_self->decimal_check_exp_overflow); if (unlikely(!__pyx_t_40)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __pyx_t_41 = __Pyx_PyBool_FromLong(__pyx_v_self->types_str_to_schema); if (unlikely(!__pyx_t_41)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __pyx_t_42 = __Pyx_PyBool_FromLong(__pyx_v_self->adapt_record_types); if (unlikely(!__pyx_t_42)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __pyx_t_43 = __Pyx_PyBool_FromLong(__pyx_v_self->return_uuid_object); if (unlikely(!__pyx_t_43)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __pyx_t_44 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_error_type); if (unlikely(!__pyx_t_44)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __pyx_t_45 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_leading_dot_in_names); if (unlikely(!__pyx_t_45)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __pyx_t_46 = __Pyx_PyBool_FromLong(__pyx_v_self->naive_dt_assume_utc); if (unlikely(!__pyx_t_46)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __pyx_t_47 = __Pyx_PyBool_FromLong(__pyx_v_self->alternate_timestamp_millis_encoding); if (unlikely(!__pyx_t_47)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __pyx_t_48 = __Pyx_PyBool_FromLong(__pyx_v_self->date_type_accepts_string); if (unlikely(!__pyx_t_48)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __pyx_t_49 = __Pyx_PyBool_FromLong(__pyx_v_self->raise_on_invalid_logical); if (unlikely(!__pyx_t_49)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __pyx_t_50 = __Pyx_PyBool_FromLong(__pyx_v_self->inline_namespaces); if (unlikely(!__pyx_t_50)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);
  __pyx_t_51 = __Pyx_PyBool_FromLong(__pyx_v_self->expand_types_in_schema); if (unlikely(!__pyx_t_51)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  __pyx_t_52 = __Pyx_PyBool_FromLong(__pyx_v_self->container_fill_blocks); if (unlikely(!__pyx_t_52)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  __pyx_t_53 = __Pyx_PyBool_FromLong(__pyx_v_self->defer_schema_promotion_errors); if (unlikely(!__pyx_t_53)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  __pyx_t_54 = __Pyx_PyBool_FromLong(__pyx_v_self->invalid_value_includes_record_name); if (unlikely(!__pyx_t_54)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __pyx_t_55 = __Pyx_PyBool_FromLong(__pyx_v_self->invalid_value_include_array_index); if (unlikely(!__pyx_t_55)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  __pyx_t_56 = __Pyx_PyBool_FromLong(__pyx_v_self->invalid_value_include_map_key); if (unlikely(!__pyx_t_56)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  __pyx_t_57 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_invalid_default_values); if (unlikely(!__pyx_t_57)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  __pyx_t_58 = PyTuple_New(63); if (unlikely(!__pyx_t_58)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 0, __pyx_t_1)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 1, __pyx_t_2)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 2, __pyx_t_3)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 3, __pyx_t_4)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 4, __pyx_t_5)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 5, __pyx_t_6)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 6, __pyx_t_7)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 7, __pyx_t_8)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 8, __pyx_t_9)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->union_decodes_to);
  __Pyx_GIVEREF(__pyx_v_self->union_decodes_to);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 9, __pyx_v_self->union_decodes_to)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 10, __pyx_t_10)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_11);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 11, __pyx_t_11)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 12, __pyx_t_12)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_13);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 13, __pyx_t_13)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_14);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 14, __pyx_t_14)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_15);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 15, __pyx_t_15)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_16);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 16, __pyx_t_16)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_17);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 17, __pyx_t_17)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_18);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 18, __pyx_t_18)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_19);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 19, __pyx_t_19)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_20);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 20, __pyx_t_20)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_21);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 21, __pyx_t_21)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_22);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 22, __pyx_t_22)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_23);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 23, __pyx_t_23)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_24);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 24, __pyx_t_24)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_25);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 25, __pyx_t_25)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_26);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 26, __pyx_t_26)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_27);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 27, __pyx_t_27)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_28);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 28, __pyx_t_28)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_29);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 29, __pyx_t_29)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_30);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 30, __pyx_t_30)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_31);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 31, __pyx_t_31)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_32);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 32, __pyx_t_32)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->bytes_codec);
  __Pyx_GIVEREF(__pyx_v_self->bytes_codec);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 33, __pyx_v_self->bytes_codec)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->fixed_codec);
  __Pyx_GIVEREF(__pyx_v_self->fixed_codec);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 34, __pyx_v_self->fixed_codec)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_33);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 35, __pyx_t_33)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_34);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 36, __pyx_t_34)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_35);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 37, __pyx_t_35)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_36);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 38, __pyx_t_36)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_37);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 39, __pyx_t_37)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_38);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 40, __pyx_t_38)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_39);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 41, __pyx_t_39)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 42, __pyx_t_40)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_41);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 43, __pyx_t_41)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->logical_types);
  __Pyx_GIVEREF(__pyx_v_self->logical_types);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 44, __pyx_v_self->logical_types)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_42);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 45, __pyx_t_42)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_43);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 46, __pyx_t_43)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_44);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 47, __pyx_t_44)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_45);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 48, __pyx_t_45)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_46);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 49, __pyx_t_46)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_47);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 50, __pyx_t_47)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_48);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 51, __pyx_t_48)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_49);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 52, __pyx_t_49)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_50);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 53, __pyx_t_50)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_51);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 54, __pyx_t_51)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->unicode_errors);
  __Pyx_GIVEREF(__pyx_v_self->unicode_errors);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 55, __pyx_v_self->unicode_errors)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_52);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 56, __pyx_t_52)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_53);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 57, __pyx_t_53)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_54);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 58, __pyx_t_54)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_55);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 59, __pyx_t_55)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_56);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 60, __pyx_t_56)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_57);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 61, __pyx_t_57)) __PYX_ERR(1, 84, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->externally_defined_types);
  __Pyx_GIVEREF(__pyx_v_self->externally_defined_types);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 62, __pyx_v_self->externally_defined_types)) __PYX_ERR(1, 84, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_15 = 0;
  __pyx_t_16 = 0;
  __pyx_t_17 = 0;
  __pyx_t_18 = 0;
  __pyx_t_19 = 0;
  __pyx_t_20 = 0;
  __pyx_t_21 = 0;
  __pyx_t_22 = 0;
  __pyx_t_23 = 0;
  __pyx_t_24 = 0;
  __pyx_t_25 = 0;
  __pyx_t_26 = 0;
  __pyx_t_27 = 0;
  __pyx_t_28 = 0;
  __pyx_t_29 = 0;
  __pyx_t_30 = 0;
  __pyx_t_31 = 0;
  __pyx_t_32 = 0;
  __pyx_t_33 = 0;
  __pyx_t_34 = 0;
  __pyx_t_35 = 0;
  __pyx_t_36 = 0;
  __pyx_t_37 = 0;
  __pyx_t_38 = 0;
  __pyx_t_39 = 0;
  __pyx_t_40 = 0;
  __pyx_t_41 = 0;
  __pyx_t_42 = 0;
  __pyx_t_43 = 0;
  __pyx_t_44 = 0;
  __pyx_t_45 = 0;
  __pyx_t_46 = 0;
  __pyx_t_47 = 0;
  __pyx_t_48 = 0;
  __pyx_t_49 = 0;
  __pyx_t_50 = 0;
  __pyx_t_51 = 0;
  __pyx_t_52 = 0;
  __pyx_t_53 = 0;
  __pyx_t_54 = 0;
  __pyx_t_55 = 0;
  __pyx_t_56 = 0;
  __pyx_t_57 = 0;
  __pyx_t_59 = PyObject_Hash(__pyx_t_58); if (unlikely(__pyx_t_59 == ((Py_hash_t)-1))) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
  __pyx_r = __pyx_t_59;
  goto __pyx_L0;

  /* "(tree fragment)":83
 *     other_cast = <Options>other
 *     return (self.fingerprint_returns_digest == other_cast.fingerprint_returns_digest) and (self.canonical_form_repeat_fixed == other_cast.canonical_form_repeat_fixed) and (self.canonical_form_repeat_enum == other_cast.canonical_form_repeat_enum) and (self.record_can_encode_dict == other_cast.record_can_encode_dict) and (self.record_values_type_hint == other_cast.record_values_type_hint) and (self.record_decodes_to_dict == other_cast.record_decodes_to_dict) and (self.record_allow_extra_fields == other_cast.record_allow_extra_fields) and (self.record_encode_use_defaults == other_cast.record_encode_use_defaults) and (self.allow_tuple_notation == other_cast.allow_tuple_notation) and (self.union_decodes_to == other_cast.union_decodes_to) and (self.union_json_encodes_type_name == other_cast.union_json_encodes_type_name) and (self.allow_primitive_name_collision == other_cast.allow_primitive_name_collision) and (self.allow_primitive_names_in_namespaces == other_cast.allow_primitive_names_in_namespaces) and (self.named_type_names_must_be_unique == other_cast.named_type_names_must_be_unique) and (self.enum_symbols_must_be_unique == other_cast.enum_symbols_must_be_unique) and (self.enforce_enum_symbol_name_rules == other_cast.enforce_enum_symbol_name_rules) and (self.enforce_type_name_rules == other_cast.enforce_type_name_rules) and (self.enforce_namespace_name_rules == other_cast.enforce_namespace_name_rules) and (self.record_fields_must_be_unique == other_cast.record_fields_must_be_unique) and (self.ascii_name_rules == other_cast.ascii_name_rules) and (self.missing_values_can_be_null == other_cast.missing_values_can_be_null) and (self.missing_values_can_be_empty_container == other_cast.missing_values_can_be_empty_container) and (self.allow_false_values_for_null == other_cast.allow_false_values_for_null) and (self.allow_empty_unions == other_cast.allow_empty_unions) and (self.allow_nested_unions == other_cast.allow_nested_unions) and (self.allow_duplicate_union_types == other_cast.allow_duplicate_union_types) and (self.allow_union_default_any_member == other_cast.allow_union_default_any_member) and (self.allow_aliases_to_be_string == other_cast.allow_aliases_to_be_string) and (self.coerce_values_to_int == other_cast.coerce_values_to_int) and (self.coerce_values_to_float == other_cast.coerce_values_to_float) and (self.coerce_int_to_float == other_cast.coerce_int_to_float) and (self.coerce_values_to_boolean == other_cast.coerce_values_to_boolean) and (self.coerce_values_to_str == other_cast.coerce_values_to_str) and (self.bytes_codec == other_cast.bytes_codec) and (self.fixed_codec == other_cast.fixed_codec) and (self.null_pad_fixed == other_cast.null_pad_fixed) and (self.truncate_fixed == other_cast.truncate_fixed) and (self.clamp_int_overflow == other_cast.clamp_int_overflow) and (self.clamp_float_overflow == other_cast.clamp_float_overflow) and (self.float_out_of_range_inf == other_cast.float_out_of_range_inf) and (self.bytes_default_value_utf8 == other_cast.bytes_default_value_utf8) and (self.string_types_default_unchanged == other_cast.string_types_default_unchanged) and (self.decimal_check_exp_overflow == other_cast.decimal_check_exp_overflow) and (self.types_str_to_schema == other_cast.types_str_to_schema) and (self.logical_types == other_cast.logical_types) and (self.adapt_record_types == other_cast.adapt_record_types) and (self.return_uuid_object == other_cast.return_uuid_object) and (self.allow_error_type == other_cast.allow_error_type) and (self.allow_leading_dot_in_names == other_cast.allow_leading_dot_in_names) and (self.naive_dt_assume_utc == other_cast.naive_dt_assume_utc) and (self.alternate_timestamp_millis_encoding == other_cast.alternate_timestamp_millis_encoding) and (self.date_type_accepts_string == other_cast.date_type_accepts_string) and (self.raise_on_invalid_logical == other_cast.raise_on_invalid_logical) and (self.inline_namespaces == other_cast.inline_namespaces) and (self.expand_types_in_schema == other_cast.expand_types_in_schema) and (self.unicode_errors == other_cast.unicode_errors) and (self.container_fill_blocks == other_cast.container_fill_blocks) and (self.defer_schema_promotion_errors == other_cast.defer_schema_promotion_errors) and (self.invalid_value_includes_record_name == other_cast.invalid_value_includes_record_name) and (self.invalid_value_include_array_index == other_cast.invalid_value_include_array_index) and (self.invalid_value_include_map_key == other_cast.invalid_value_include_map_key) and (self.allow_invalid_default_values == other_cast.allow_invalid_default_values) and (self.externally_defined_types == other_cast.externally_defined_types)
 * def __hash__(self):             # <<<<<<<<<<<<<<
 *     return hash((self.fingerprint_returns_digest, self.canonical_form_repeat_fixed, self.canonical_form_repeat_enum, self.record_can_encode_dict, self.record_values_type_hint, self.record_decodes_to_dict, self.record_allow_extra_fields, self.record_encode_use_defaults, self.allow_tuple_notation, self.union_decodes_to, self.union_json_encodes_type_name, self.allow_primitive_name_collision, self.allow_primitive_names_in_namespaces, self.named_type_names_must_be_unique, self.enum_symbols_must_be_unique, self.enforce_enum_symbol_name_rules, self.enforce_type_name_rules, self.enforce_namespace_name_rules, self.record_fields_must_be_unique, self.ascii_name_rules, self.missing_values_can_be_null, self.missing_values_can_be_empty_container, self.allow_false_values_for_null, self.allow_empty_unions, self.allow_nested_unions, self.allow_duplicate_union_types, self.allow_union_default_any_member, self.allow_aliases_to_be_string, self.coerce_values_to_int, self.coerce_values_to_float, self.coerce_int_to_float, self.coerce_values_to_boolean, self.coerce_values_to_str, self.bytes_codec, self.fixed_codec, self.null_pad_fixed, self.truncate_fixed, self.clamp_int_overflow, self.clamp_float_overflow, self.float_out_of_range_inf, self.bytes_default_value_utf8, self.string_types_default_unchanged, self.decimal_check_exp_overflow, self.types_str_to_schema, self.logical_types, self.adapt_record_types, self.return_uuid_object, self.allow_error_type, self.allow_leading_dot_in_names, self.naive_dt_assume_utc, self.alternate_timestamp_millis_encoding, self.date_type_accepts_string, self.raise_on_invalid_logical, self.inline_namespaces, self.expand_types_in_schema, self.unicode_errors, self.container_fill_blocks, self.defer_schema_promotion_errors, self.invalid_value_includes_record_name, self.invalid_value_include_array_index, self.invalid_value_include_map_key, self.allow_invalid_default_values, self.externally_defined_types,))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __Pyx_XDECREF(__pyx_t_34);
  __Pyx_XDECREF(__pyx_t_35);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_37);
  __Pyx_XDECREF(__pyx_t_38);
  __Pyx_XDECREF(__pyx_t_39);
  __Pyx_XDECREF(__pyx_t_40);
  __Pyx_XDECREF(__pyx_t_41);
  __Pyx_XDECREF(__pyx_t_42);
  __Pyx_XDECREF(__pyx_t_43);
  __Pyx_XDECREF(__pyx_t_44);
  __Pyx_XDECREF(__pyx_t_45);
  __Pyx_XDECREF(__pyx_t_46);
  __Pyx_XDECREF(__pyx_t_47);
  __Pyx_XDECREF(__pyx_t_48);
  __Pyx_XDECREF(__pyx_t_49);
  __Pyx_XDECREF(__pyx_t_50);
  __Pyx_XDECREF(__pyx_t_51);
  __Pyx_XDECREF(__pyx_t_52);
  __Pyx_XDECREF(__pyx_t_53);
  __Pyx_XDECREF(__pyx_t_54);
  __Pyx_XDECREF(__pyx_t_55);
  __Pyx_XDECREF(__pyx_t_56);
  __Pyx_XDECREF(__pyx_t_57);
  __Pyx_XDECREF(__pyx_t_58);
  __Pyx_AddTraceback("cavro.Options.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":205
 *     __annotations__ = _OptionsWrapper.annotations
 * 
 *     fingerprint_returns_digest: bint = False             # <<<<<<<<<<<<<<
 *     canonical_form_repeat_fixed: bint = False
 *     canonical_form_repeat_enum: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_26fingerprint_returns_digest_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_26fingerprint_returns_digest_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_26fingerprint_returns_digest___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_26fingerprint_returns_digest___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->fingerprint_returns_digest); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.fingerprint_returns_digest.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":206
 * 
 *     fingerprint_returns_digest: bint = False
 *     canonical_form_repeat_fixed: bint = False             # <<<<<<<<<<<<<<
 *     canonical_form_repeat_enum: bint = False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_27canonical_form_repeat_fixed_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_27canonical_form_repeat_fixed_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_27canonical_form_repeat_fixed___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_27canonical_form_repeat_fixed___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->canonical_form_repeat_fixed); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.canonical_form_repeat_fixed.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":207
 *     fingerprint_returns_digest: bint = False
 *     canonical_form_repeat_fixed: bint = False
 *     canonical_form_repeat_enum: bint = False             # <<<<<<<<<<<<<<
 * 
 *     record_can_encode_dict: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_26canonical_form_repeat_enum_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_26canonical_form_repeat_enum_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_26canonical_form_repeat_enum___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_26canonical_form_repeat_enum___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->canonical_form_repeat_enum); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.canonical_form_repeat_enum.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":209
 *     canonical_form_repeat_enum: bint = False
 * 
 *     record_can_encode_dict: bint = True             # <<<<<<<<<<<<<<
 *     record_values_type_hint: bint = False
 *     record_decodes_to_dict: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_22record_can_encode_dict_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_22record_can_encode_dict_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_22record_can_encode_dict___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_22record_can_encode_dict___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->record_can_encode_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.record_can_encode_dict.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":210
 * 
 *     record_can_encode_dict: bint = True
 *     record_values_type_hint: bint = False             # <<<<<<<<<<<<<<
 *     record_decodes_to_dict: bint = False
 *     record_allow_extra_fields: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_23record_values_type_hint_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_23record_values_type_hint_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_23record_values_type_hint___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_23record_values_type_hint___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->record_values_type_hint); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.record_values_type_hint.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":211
 *     record_can_encode_dict: bint = True
 *     record_values_type_hint: bint = False
 *     record_decodes_to_dict: bint = False             # <<<<<<<<<<<<<<
 *     record_allow_extra_fields: bint = True
 *     record_encode_use_defaults: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_22record_decodes_to_dict_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_22record_decodes_to_dict_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_22record_decodes_to_dict___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_22record_decodes_to_dict___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->record_decodes_to_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.record_decodes_to_dict.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":212
 *     record_values_type_hint: bint = False
 *     record_decodes_to_dict: bint = False
 *     record_allow_extra_fields: bint = True             # <<<<<<<<<<<<<<
 *     record_encode_use_defaults: bint = True
 *     allow_tuple_notation: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_25record_allow_extra_fields_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_25record_allow_extra_fields_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_25record_allow_extra_fields___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_25record_allow_extra_fields___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->record_allow_extra_fields); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.record_allow_extra_fields.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":213
 *     record_decodes_to_dict: bint = False
 *     record_allow_extra_fields: bint = True
 *     record_encode_use_defaults: bint = True             # <<<<<<<<<<<<<<
 *     allow_tuple_notation: bint = False
 *     union_decodes_to: UnionDecodeOption = UnionDecodeOption.RAW_VALUES
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_26record_encode_use_defaults_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_26record_encode_use_defaults_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_26record_encode_use_defaults___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_26record_encode_use_defaults___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->record_encode_use_defaults); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.record_encode_use_defaults.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":214
 *     record_allow_extra_fields: bint = True
 *     record_encode_use_defaults: bint = True
 *     allow_tuple_notation: bint = False             # <<<<<<<<<<<<<<
 *     union_decodes_to: UnionDecodeOption = UnionDecodeOption.RAW_VALUES
 *     union_json_encodes_type_name: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_20allow_tuple_notation_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_20allow_tuple_notation_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_20allow_tuple_notation___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_20allow_tuple_notation___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_tuple_notation); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.allow_tuple_notation.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":215
 *     record_encode_use_defaults: bint = True
 *     allow_tuple_notation: bint = False
 *     union_decodes_to: UnionDecodeOption = UnionDecodeOption.RAW_VALUES             # <<<<<<<<<<<<<<
 *     union_json_encodes_type_name: bint = True
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_16union_decodes_to_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_16union_decodes_to_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_16union_decodes_to___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_16union_decodes_to___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->union_decodes_to);
  __pyx_r = __pyx_v_self->union_decodes_to;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":216
 *     allow_tuple_notation: bint = False
 *     union_decodes_to: UnionDecodeOption = UnionDecodeOption.RAW_VALUES
 *     union_json_encodes_type_name: bint = True             # <<<<<<<<<<<<<<
 * 
 *     allow_primitive_name_collision: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_28union_json_encodes_type_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_28union_json_encodes_type_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_28union_json_encodes_type_name___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_28union_json_encodes_type_name___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->union_json_encodes_type_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.union_json_encodes_type_name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":218
 *     union_json_encodes_type_name: bint = True
 * 
 *     allow_primitive_name_collision: bint = False             # <<<<<<<<<<<<<<
 *     allow_primitive_names_in_namespaces: bint = False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_30allow_primitive_name_collision_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_30allow_primitive_name_collision_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_30allow_primitive_name_collision___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_30allow_primitive_name_collision___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_primitive_name_collision); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.allow_primitive_name_collision.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":219
 * 
 *     allow_primitive_name_collision: bint = False
 *     allow_primitive_names_in_namespaces: bint = False             # <<<<<<<<<<<<<<
 * 
 *     named_type_names_must_be_unique: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_35allow_primitive_names_in_namespaces_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_35allow_primitive_names_in_namespaces_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_35allow_primitive_names_in_namespaces___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_35allow_primitive_names_in_namespaces___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_primitive_names_in_namespaces); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.allow_primitive_names_in_namespaces.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":221
 *     allow_primitive_names_in_namespaces: bint = False
 * 
 *     named_type_names_must_be_unique: bint = True             # <<<<<<<<<<<<<<
 *     enum_symbols_must_be_unique: bint = True
 *     enforce_enum_symbol_name_rules: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_31named_type_names_must_be_unique_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_31named_type_names_must_be_unique_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_31named_type_names_must_be_unique___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_31named_type_names_must_be_unique___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->named_type_names_must_be_unique); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.named_type_names_must_be_unique.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":222
 * 
 *     named_type_names_must_be_unique: bint = True
 *     enum_symbols_must_be_unique: bint = True             # <<<<<<<<<<<<<<
 *     enforce_enum_symbol_name_rules: bint = True
 *     enforce_type_name_rules: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_27enum_symbols_must_be_unique_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_27enum_symbols_must_be_unique_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_27enum_symbols_must_be_unique___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_27enum_symbols_must_be_unique___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->enum_symbols_must_be_unique); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.enum_symbols_must_be_unique.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":223
 *     named_type_names_must_be_unique: bint = True
 *     enum_symbols_must_be_unique: bint = True
 *     enforce_enum_symbol_name_rules: bint = True             # <<<<<<<<<<<<<<
 *     enforce_type_name_rules: bint = True
 *     enforce_namespace_name_rules: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_30enforce_enum_symbol_name_rules_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_30enforce_enum_symbol_name_rules_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_30enforce_enum_symbol_name_rules___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_30enforce_enum_symbol_name_rules___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->enforce_enum_symbol_name_rules); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.enforce_enum_symbol_name_rules.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":224
 *     enum_symbols_must_be_unique: bint = True
 *     enforce_enum_symbol_name_rules: bint = True
 *     enforce_type_name_rules: bint = True             # <<<<<<<<<<<<<<
 *     enforce_namespace_name_rules: bint = True
 *     record_fields_must_be_unique: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_23enforce_type_name_rules_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_23enforce_type_name_rules_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_23enforce_type_name_rules___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_23enforce_type_name_rules___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->enforce_type_name_rules); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.enforce_type_name_rules.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":225
 *     enforce_enum_symbol_name_rules: bint = True
 *     enforce_type_name_rules: bint = True
 *     enforce_namespace_name_rules: bint = True             # <<<<<<<<<<<<<<
 *     record_fields_must_be_unique: bint = True
 *     ascii_name_rules: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_28enforce_namespace_name_rules_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_28enforce_namespace_name_rules_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_28enforce_namespace_name_rules___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_28enforce_namespace_name_rules___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->enforce_namespace_name_rules); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.enforce_namespace_name_rules.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":226
 *     enforce_type_name_rules: bint = True
 *     enforce_namespace_name_rules: bint = True
 *     record_fields_must_be_unique: bint = True             # <<<<<<<<<<<<<<
 *     ascii_name_rules: bint = True
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_28record_fields_must_be_unique_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_28record_fields_must_be_unique_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_28record_fields_must_be_unique___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_28record_fields_must_be_unique___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->record_fields_must_be_unique); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.record_fields_must_be_unique.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":227
 *     enforce_namespace_name_rules: bint = True
 *     record_fields_must_be_unique: bint = True
 *     ascii_name_rules: bint = True             # <<<<<<<<<<<<<<
 * 
 *     missing_values_can_be_null: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_16ascii_name_rules_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_16ascii_name_rules_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_16ascii_name_rules___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_16ascii_name_rules___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->ascii_name_rules); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.ascii_name_rules.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":229
 *     ascii_name_rules: bint = True
 * 
 *     missing_values_can_be_null: bint = False             # <<<<<<<<<<<<<<
 *     missing_values_can_be_empty_container: bint = False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_26missing_values_can_be_null_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_26missing_values_can_be_null_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_26missing_values_can_be_null___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_26missing_values_can_be_null___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->missing_values_can_be_null); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.missing_values_can_be_null.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":230
 * 
 *     missing_values_can_be_null: bint = False
 *     missing_values_can_be_empty_container: bint = False             # <<<<<<<<<<<<<<
 * 
 *     allow_false_values_for_null: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_37missing_values_can_be_empty_container_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_37missing_values_can_be_empty_container_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_37missing_values_can_be_empty_container___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_37missing_values_can_be_empty_container___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->missing_values_can_be_empty_container); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.missing_values_can_be_empty_container.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":232
 *     missing_values_can_be_empty_container: bint = False
 * 
 *     allow_false_values_for_null: bint = False             # <<<<<<<<<<<<<<
 *     allow_empty_unions: bint = False
 *     allow_nested_unions: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_27allow_false_values_for_null_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_27allow_false_values_for_null_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_27allow_false_values_for_null___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_27allow_false_values_for_null___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_false_values_for_null); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.allow_false_values_for_null.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":233
 * 
 *     allow_false_values_for_null: bint = False
 *     allow_empty_unions: bint = False             # <<<<<<<<<<<<<<
 *     allow_nested_unions: bint = False
 *     allow_duplicate_union_types: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_18allow_empty_unions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_18allow_empty_unions_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_18allow_empty_unions___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_18allow_empty_unions___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_empty_unions); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.allow_empty_unions.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":234
 *     allow_false_values_for_null: bint = False
 *     allow_empty_unions: bint = False
 *     allow_nested_unions: bint = False             # <<<<<<<<<<<<<<
 *     allow_duplicate_union_types: bint = False
 *     allow_union_default_any_member: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_19allow_nested_unions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_19allow_nested_unions_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_19allow_nested_unions___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_19allow_nested_unions___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_nested_unions); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.allow_nested_unions.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":235
 *     allow_empty_unions: bint = False
 *     allow_nested_unions: bint = False
 *     allow_duplicate_union_types: bint = False             # <<<<<<<<<<<<<<
 *     allow_union_default_any_member: bint = False
 *     allow_aliases_to_be_string: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_27allow_duplicate_union_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_27allow_duplicate_union_types_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_27allow_duplicate_union_types___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_27allow_duplicate_union_types___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_duplicate_union_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.allow_duplicate_union_types.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":236
 *     allow_nested_unions: bint = False
 *     allow_duplicate_union_types: bint = False
 *     allow_union_default_any_member: bint = False             # <<<<<<<<<<<<<<
 *     allow_aliases_to_be_string: bint = False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_30allow_union_default_any_member_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_30allow_union_default_any_member_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_30allow_union_default_any_member___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_30allow_union_default_any_member___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_union_default_any_member); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.allow_union_default_any_member.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":237
 *     allow_duplicate_union_types: bint = False
 *     allow_union_default_any_member: bint = False
 *     allow_aliases_to_be_string: bint = False             # <<<<<<<<<<<<<<
 * 
 *     coerce_values_to_int: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_26allow_aliases_to_be_string_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_26allow_aliases_to_be_string_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_26allow_aliases_to_be_string___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_26allow_aliases_to_be_string___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_aliases_to_be_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.allow_aliases_to_be_string.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":239
 *     allow_aliases_to_be_string: bint = False
 * 
 *     coerce_values_to_int: bint = False             # <<<<<<<<<<<<<<
 *     coerce_values_to_float: bint = False
 *     coerce_int_to_float: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_20coerce_values_to_int_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_20coerce_values_to_int_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_20coerce_values_to_int___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_20coerce_values_to_int___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_values_to_int); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.coerce_values_to_int.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":240
 * 
 *     coerce_values_to_int: bint = False
 *     coerce_values_to_float: bint = False             # <<<<<<<<<<<<<<
 *     coerce_int_to_float: bint = True
 *     coerce_values_to_boolean: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_22coerce_values_to_float_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_22coerce_values_to_float_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_22coerce_values_to_float___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_22coerce_values_to_float___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_values_to_float); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.coerce_values_to_float.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":241
 *     coerce_values_to_int: bint = False
 *     coerce_values_to_float: bint = False
 *     coerce_int_to_float: bint = True             # <<<<<<<<<<<<<<
 *     coerce_values_to_boolean: bint = False
 *     coerce_values_to_str: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_19coerce_int_to_float_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_19coerce_int_to_float_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_19coerce_int_to_float___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_19coerce_int_to_float___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_int_to_float); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.coerce_int_to_float.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":242
 *     coerce_values_to_float: bint = False
 *     coerce_int_to_float: bint = True
 *     coerce_values_to_boolean: bint = False             # <<<<<<<<<<<<<<
 *     coerce_values_to_str: bint = False
 *     bytes_codec: str = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_24coerce_values_to_boolean_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_24coerce_values_to_boolean_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_24coerce_values_to_boolean___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_24coerce_values_to_boolean___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_values_to_boolean); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.coerce_values_to_boolean.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":243
 *     coerce_int_to_float: bint = True
 *     coerce_values_to_boolean: bint = False
 *     coerce_values_to_str: bint = False             # <<<<<<<<<<<<<<
 *     bytes_codec: str = None
 *     fixed_codec: str = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_20coerce_values_to_str_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_20coerce_values_to_str_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_20coerce_values_to_str___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_20coerce_values_to_str___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_values_to_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.coerce_values_to_str.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":244
 *     coerce_values_to_boolean: bint = False
 *     coerce_values_to_str: bint = False
 *     bytes_codec: str = None             # <<<<<<<<<<<<<<
 *     fixed_codec: str = None
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_11bytes_codec_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_11bytes_codec_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_11bytes_codec___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_11bytes_codec___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->bytes_codec);
  __pyx_r = __pyx_v_self->bytes_codec;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":245
 *     coerce_values_to_str: bint = False
 *     bytes_codec: str = None
 *     fixed_codec: str = None             # <<<<<<<<<<<<<<
 * 
 *     null_pad_fixed: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_11fixed_codec_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_11fixed_codec_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_11fixed_codec___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_11fixed_codec___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->fixed_codec);
  __pyx_r = __pyx_v_self->fixed_codec;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":247
 *     fixed_codec: str = None
 * 
 *     null_pad_fixed: bint = False             # <<<<<<<<<<<<<<
 *     truncate_fixed: bint = False
 *     clamp_int_overflow: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_14null_pad_fixed_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_14null_pad_fixed_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_14null_pad_fixed___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_14null_pad_fixed___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->null_pad_fixed); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.null_pad_fixed.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":248
 * 
 *     null_pad_fixed: bint = False
 *     truncate_fixed: bint = False             # <<<<<<<<<<<<<<
 *     clamp_int_overflow: bint = False
 *     clamp_float_overflow: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_14truncate_fixed_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_14truncate_fixed_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_14truncate_fixed___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_14truncate_fixed___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->truncate_fixed); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.truncate_fixed.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":249
 *     null_pad_fixed: bint = False
 *     truncate_fixed: bint = False
 *     clamp_int_overflow: bint = False             # <<<<<<<<<<<<<<
 *     clamp_float_overflow: bint = False
 *     float_out_of_range_inf: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_18clamp_int_overflow_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_18clamp_int_overflow_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_18clamp_int_overflow___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_18clamp_int_overflow___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->clamp_int_overflow); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.clamp_int_overflow.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":250
 *     truncate_fixed: bint = False
 *     clamp_int_overflow: bint = False
 *     clamp_float_overflow: bint = False             # <<<<<<<<<<<<<<
 *     float_out_of_range_inf: bint = False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_20clamp_float_overflow_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_20clamp_float_overflow_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_20clamp_float_overflow___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_20clamp_float_overflow___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->clamp_float_overflow); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.clamp_float_overflow.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":251
 *     clamp_int_overflow: bint = False
 *     clamp_float_overflow: bint = False
 *     float_out_of_range_inf: bint = False             # <<<<<<<<<<<<<<
 * 
 *     bytes_default_value_utf8: bint = False  # Avro 1.11/12 utf8 encode the default value for bytes (rather than latin1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_22float_out_of_range_inf_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_22float_out_of_range_inf_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_22float_out_of_range_inf___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_22float_out_of_range_inf___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->float_out_of_range_inf); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.float_out_of_range_inf.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":253
 *     float_out_of_range_inf: bint = False
 * 
 *     bytes_default_value_utf8: bint = False  # Avro 1.11/12 utf8 encode the default value for bytes (rather than latin1)             # <<<<<<<<<<<<<<
 *     string_types_default_unchanged: bint = False # Avro < 1.11 pass default value back unmodified for bytes/fixed etc..
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_24bytes_default_value_utf8_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_24bytes_default_value_utf8_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_24bytes_default_value_utf8___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_24bytes_default_value_utf8___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->bytes_default_value_utf8); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.bytes_default_value_utf8.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":254
 * 
 *     bytes_default_value_utf8: bint = False  # Avro 1.11/12 utf8 encode the default value for bytes (rather than latin1)
 *     string_types_default_unchanged: bint = False # Avro < 1.11 pass default value back unmodified for bytes/fixed etc..             # <<<<<<<<<<<<<<
 * 
 *     decimal_check_exp_overflow: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_30string_types_default_unchanged_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_30string_types_default_unchanged_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_30string_types_default_unchanged___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_30string_types_default_unchanged___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->string_types_default_unchanged); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.string_types_default_unchanged.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":256
 *     string_types_default_unchanged: bint = False # Avro < 1.11 pass default value back unmodified for bytes/fixed etc..
 * 
 *     decimal_check_exp_overflow: bint = True             # <<<<<<<<<<<<<<
 * 
 *     types_str_to_schema: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_26decimal_check_exp_overflow_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_26decimal_check_exp_overflow_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_26decimal_check_exp_overflow___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_26decimal_check_exp_overflow___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->decimal_check_exp_overflow); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.decimal_check_exp_overflow.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":258
 *     decimal_check_exp_overflow: bint = True
 * 
 *     types_str_to_schema: bint = False             # <<<<<<<<<<<<<<
 *     logical_types: tuple[LogicalType] = LOGICAL_TYPES
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_19types_str_to_schema_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_19types_str_to_schema_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_19types_str_to_schema___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_19types_str_to_schema___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->types_str_to_schema); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.types_str_to_schema.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":259
 * 
 *     types_str_to_schema: bint = False
 *     logical_types: tuple[LogicalType] = LOGICAL_TYPES             # <<<<<<<<<<<<<<
 * 
 *     adapt_record_types: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_13logical_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_13logical_types_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_13logical_types___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_13logical_types___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->logical_types);
  __pyx_r = __pyx_v_self->logical_types;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":261
 *     logical_types: tuple[LogicalType] = LOGICAL_TYPES
 * 
 *     adapt_record_types: bint = True             # <<<<<<<<<<<<<<
 *     return_uuid_object: bint = True
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_18adapt_record_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_18adapt_record_types_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_18adapt_record_types___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_18adapt_record_types___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->adapt_record_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.adapt_record_types.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":262
 * 
 *     adapt_record_types: bint = True
 *     return_uuid_object: bint = True             # <<<<<<<<<<<<<<
 * 
 *     allow_error_type: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_18return_uuid_object_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_18return_uuid_object_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_18return_uuid_object___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_18return_uuid_object___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->return_uuid_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.return_uuid_object.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":264
 *     return_uuid_object: bint = True
 * 
 *     allow_error_type: bint = False             # <<<<<<<<<<<<<<
 *     allow_leading_dot_in_names: bint = True
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_16allow_error_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_16allow_error_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_16allow_error_type___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_16allow_error_type___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_error_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.allow_error_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":265
 * 
 *     allow_error_type: bint = False
 *     allow_leading_dot_in_names: bint = True             # <<<<<<<<<<<<<<
 * 
 *     naive_dt_assume_utc: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_26allow_leading_dot_in_names_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_26allow_leading_dot_in_names_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_26allow_leading_dot_in_names___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_26allow_leading_dot_in_names___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_leading_dot_in_names); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.allow_leading_dot_in_names.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":267
 *     allow_leading_dot_in_names: bint = True
 * 
 *     naive_dt_assume_utc: bint = False             # <<<<<<<<<<<<<<
 *     alternate_timestamp_millis_encoding: bint = False
 *     date_type_accepts_string: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_19naive_dt_assume_utc_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_19naive_dt_assume_utc_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_19naive_dt_assume_utc___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_19naive_dt_assume_utc___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->naive_dt_assume_utc); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.naive_dt_assume_utc.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":268
 * 
 *     naive_dt_assume_utc: bint = False
 *     alternate_timestamp_millis_encoding: bint = False             # <<<<<<<<<<<<<<
 *     date_type_accepts_string: bint = False
 *     raise_on_invalid_logical: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_35alternate_timestamp_millis_encoding_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_35alternate_timestamp_millis_encoding_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_35alternate_timestamp_millis_encoding___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_35alternate_timestamp_millis_encoding___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->alternate_timestamp_millis_encoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.alternate_timestamp_millis_encoding.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":269
 *     naive_dt_assume_utc: bint = False
 *     alternate_timestamp_millis_encoding: bint = False
 *     date_type_accepts_string: bint = False             # <<<<<<<<<<<<<<
 *     raise_on_invalid_logical: bint = False
 *     inline_namespaces: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_24date_type_accepts_string_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_24date_type_accepts_string_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_24date_type_accepts_string___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_24date_type_accepts_string___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->date_type_accepts_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.date_type_accepts_string.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":270
 *     alternate_timestamp_millis_encoding: bint = False
 *     date_type_accepts_string: bint = False
 *     raise_on_invalid_logical: bint = False             # <<<<<<<<<<<<<<
 *     inline_namespaces: bint = False
 *     expand_types_in_schema: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_24raise_on_invalid_logical_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_24raise_on_invalid_logical_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_24raise_on_invalid_logical___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_24raise_on_invalid_logical___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->raise_on_invalid_logical); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.raise_on_invalid_logical.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":271
 *     date_type_accepts_string: bint = False
 *     raise_on_invalid_logical: bint = False
 *     inline_namespaces: bint = False             # <<<<<<<<<<<<<<
 *     expand_types_in_schema: bint = False
 *     unicode_errors: str = 'strict'
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_17inline_namespaces_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_17inline_namespaces_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_17inline_namespaces___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_17inline_namespaces___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->inline_namespaces); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.inline_namespaces.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":272
 *     raise_on_invalid_logical: bint = False
 *     inline_namespaces: bint = False
 *     expand_types_in_schema: bint = False             # <<<<<<<<<<<<<<
 *     unicode_errors: str = 'strict'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_22expand_types_in_schema_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_22expand_types_in_schema_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_22expand_types_in_schema___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_22expand_types_in_schema___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->expand_types_in_schema); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.expand_types_in_schema.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":273
 *     inline_namespaces: bint = False
 *     expand_types_in_schema: bint = False
 *     unicode_errors: str = 'strict'             # <<<<<<<<<<<<<<
 * 
 *     container_fill_blocks: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_14unicode_errors_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_14unicode_errors_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_14unicode_errors___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_14unicode_errors___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->unicode_errors);
  __pyx_r = __pyx_v_self->unicode_errors;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":275
 *     unicode_errors: str = 'strict'
 * 
 *     container_fill_blocks: bint = False             # <<<<<<<<<<<<<<
 *     defer_schema_promotion_errors: bint = False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_21container_fill_blocks_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_21container_fill_blocks_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_21container_fill_blocks___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_21container_fill_blocks___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->container_fill_blocks); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.container_fill_blocks.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":276
 * 
 *     container_fill_blocks: bint = False
 *     defer_schema_promotion_errors: bint = False             # <<<<<<<<<<<<<<
 * 
 *     invalid_value_includes_record_name: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_29defer_schema_promotion_errors_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_29defer_schema_promotion_errors_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_29defer_schema_promotion_errors___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_29defer_schema_promotion_errors___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->defer_schema_promotion_errors); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.defer_schema_promotion_errors.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":278
 *     defer_schema_promotion_errors: bint = False
 * 
 *     invalid_value_includes_record_name: bint = False             # <<<<<<<<<<<<<<
 *     invalid_value_include_array_index: bint = True
 *     invalid_value_include_map_key: bint = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_34invalid_value_includes_record_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_34invalid_value_includes_record_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_34invalid_value_includes_record_name___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_34invalid_value_includes_record_name___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->invalid_value_includes_record_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.invalid_value_includes_record_name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":279
 * 
 *     invalid_value_includes_record_name: bint = False
 *     invalid_value_include_array_index: bint = True             # <<<<<<<<<<<<<<
 *     invalid_value_include_map_key: bint = True
 *     allow_invalid_default_values: bint = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_33invalid_value_include_array_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_33invalid_value_include_array_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_33invalid_value_include_array_index___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_33invalid_value_include_array_index___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->invalid_value_include_array_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.invalid_value_include_array_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":280
 *     invalid_value_includes_record_name: bint = False
 *     invalid_value_include_array_index: bint = True
 *     invalid_value_include_map_key: bint = True             # <<<<<<<<<<<<<<
 *     allow_invalid_default_values: bint = False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_29invalid_value_include_map_key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_29invalid_value_include_map_key_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_29invalid_value_include_map_key___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_29invalid_value_include_map_key___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->invalid_value_include_map_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.invalid_value_include_map_key.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":281
 *     invalid_value_include_array_index: bint = True
 *     invalid_value_include_map_key: bint = True
 *     allow_invalid_default_values: bint = False             # <<<<<<<<<<<<<<
 * 
 *     externally_defined_types: MappingProxyType = _EMPTY_MP
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_28allow_invalid_default_values_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_28allow_invalid_default_values_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_28allow_invalid_default_values___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_28allow_invalid_default_values___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_invalid_default_values); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.allow_invalid_default_values.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/option.pxi":283
 *     allow_invalid_default_values: bint = False
 * 
 *     externally_defined_types: MappingProxyType = _EMPTY_MP             # <<<<<<<<<<<<<<
 * 
 *     def replace(self, **changes) -> Options:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_24externally_defined_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7Options_24externally_defined_types_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7Options_24externally_defined_types___get__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_24externally_defined_types___get__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->externally_defined_types);
  __pyx_r = __pyx_v_self->externally_defined_types;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_19__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7Options_19__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7Options_19__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7Options_19__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_7Options_18__reduce_cython__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_18__reduce_cython__(struct __pyx_obj_5cavro_Options *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  PyObject *__pyx_t_36 = NULL;
  PyObject *__pyx_t_37 = NULL;
  PyObject *__pyx_t_38 = NULL;
  PyObject *__pyx_t_39 = NULL;
  PyObject *__pyx_t_40 = NULL;
  PyObject *__pyx_t_41 = NULL;
  PyObject *__pyx_t_42 = NULL;
  PyObject *__pyx_t_43 = NULL;
  PyObject *__pyx_t_44 = NULL;
  PyObject *__pyx_t_45 = NULL;
  PyObject *__pyx_t_46 = NULL;
  PyObject *__pyx_t_47 = NULL;
  PyObject *__pyx_t_48 = NULL;
  PyObject *__pyx_t_49 = NULL;
  PyObject *__pyx_t_50 = NULL;
  PyObject *__pyx_t_51 = NULL;
  PyObject *__pyx_t_52 = NULL;
  PyObject *__pyx_t_53 = NULL;
  PyObject *__pyx_t_54 = NULL;
  PyObject *__pyx_t_55 = NULL;
  PyObject *__pyx_t_56 = NULL;
  PyObject *__pyx_t_57 = NULL;
  PyObject *__pyx_t_58 = NULL;
  int __pyx_t_59;
  int __pyx_t_60;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.adapt_record_types, self.allow_aliases_to_be_string, self.allow_duplicate_union_types, self.allow_empty_unions, self.allow_error_type, self.allow_false_values_for_null, self.allow_invalid_default_values, self.allow_leading_dot_in_names, self.allow_nested_unions, self.allow_primitive_name_collision, self.allow_primitive_names_in_namespaces, self.allow_tuple_notation, self.allow_union_default_any_member, self.alternate_timestamp_millis_encoding, self.ascii_name_rules, self.bytes_codec, self.bytes_default_value_utf8, self.canonical_form_repeat_enum, self.canonical_form_repeat_fixed, self.clamp_float_overflow, self.clamp_int_overflow, self.coerce_int_to_float, self.coerce_values_to_boolean, self.coerce_values_to_float, self.coerce_values_to_int, self.coerce_values_to_str, self.container_fill_blocks, self.date_type_accepts_string, self.decimal_check_exp_overflow, self.defer_schema_promotion_errors, self.enforce_enum_symbol_name_rules, self.enforce_namespace_name_rules, self.enforce_type_name_rules, self.enum_symbols_must_be_unique, self.expand_types_in_schema, self.externally_defined_types, self.fingerprint_returns_digest, self.fixed_codec, self.float_out_of_range_inf, self.inline_namespaces, self.invalid_value_include_array_index, self.invalid_value_include_map_key, self.invalid_value_includes_record_name, self.logical_types, self.missing_values_can_be_empty_container, self.missing_values_can_be_null, self.naive_dt_assume_utc, self.named_type_names_must_be_unique, self.null_pad_fixed, self.raise_on_invalid_logical, self.record_allow_extra_fields, self.record_can_encode_dict, self.record_decodes_to_dict, self.record_encode_use_defaults, self.record_fields_must_be_unique, self.record_values_type_hint, self.return_uuid_object, self.string_types_default_unchanged, self.truncate_fixed, self.types_str_to_schema, self.unicode_errors, self.union_decodes_to, self.union_json_encodes_type_name)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->adapt_record_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_aliases_to_be_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_duplicate_union_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_empty_unions); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_error_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_false_values_for_null); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_invalid_default_values); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_leading_dot_in_names); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_nested_unions); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_primitive_name_collision); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_primitive_names_in_namespaces); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_tuple_notation); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyBool_FromLong(__pyx_v_self->allow_union_default_any_member); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_14 = __Pyx_PyBool_FromLong(__pyx_v_self->alternate_timestamp_millis_encoding); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __Pyx_PyBool_FromLong(__pyx_v_self->ascii_name_rules); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = __Pyx_PyBool_FromLong(__pyx_v_self->bytes_default_value_utf8); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __Pyx_PyBool_FromLong(__pyx_v_self->canonical_form_repeat_enum); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_18 = __Pyx_PyBool_FromLong(__pyx_v_self->canonical_form_repeat_fixed); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_19 = __Pyx_PyBool_FromLong(__pyx_v_self->clamp_float_overflow); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_20 = __Pyx_PyBool_FromLong(__pyx_v_self->clamp_int_overflow); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_int_to_float); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_22 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_values_to_boolean); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_23 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_values_to_float); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_24 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_values_to_int); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_25 = __Pyx_PyBool_FromLong(__pyx_v_self->coerce_values_to_str); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_26 = __Pyx_PyBool_FromLong(__pyx_v_self->container_fill_blocks); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_27 = __Pyx_PyBool_FromLong(__pyx_v_self->date_type_accepts_string); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_28 = __Pyx_PyBool_FromLong(__pyx_v_self->decimal_check_exp_overflow); if (unlikely(!__pyx_t_28)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __pyx_t_29 = __Pyx_PyBool_FromLong(__pyx_v_self->defer_schema_promotion_errors); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_30 = __Pyx_PyBool_FromLong(__pyx_v_self->enforce_enum_symbol_name_rules); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_31 = __Pyx_PyBool_FromLong(__pyx_v_self->enforce_namespace_name_rules); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_32 = __Pyx_PyBool_FromLong(__pyx_v_self->enforce_type_name_rules); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_33 = __Pyx_PyBool_FromLong(__pyx_v_self->enum_symbols_must_be_unique); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = __Pyx_PyBool_FromLong(__pyx_v_self->expand_types_in_schema); if (unlikely(!__pyx_t_34)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_35 = __Pyx_PyBool_FromLong(__pyx_v_self->fingerprint_returns_digest); if (unlikely(!__pyx_t_35)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __pyx_t_36 = __Pyx_PyBool_FromLong(__pyx_v_self->float_out_of_range_inf); if (unlikely(!__pyx_t_36)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __pyx_t_37 = __Pyx_PyBool_FromLong(__pyx_v_self->inline_namespaces); if (unlikely(!__pyx_t_37)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_38 = __Pyx_PyBool_FromLong(__pyx_v_self->invalid_value_include_array_index); if (unlikely(!__pyx_t_38)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __pyx_t_39 = __Pyx_PyBool_FromLong(__pyx_v_self->invalid_value_include_map_key); if (unlikely(!__pyx_t_39)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __pyx_t_40 = __Pyx_PyBool_FromLong(__pyx_v_self->invalid_value_includes_record_name); if (unlikely(!__pyx_t_40)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __pyx_t_41 = __Pyx_PyBool_FromLong(__pyx_v_self->missing_values_can_be_empty_container); if (unlikely(!__pyx_t_41)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __pyx_t_42 = __Pyx_PyBool_FromLong(__pyx_v_self->missing_values_can_be_null); if (unlikely(!__pyx_t_42)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __pyx_t_43 = __Pyx_PyBool_FromLong(__pyx_v_self->naive_dt_assume_utc); if (unlikely(!__pyx_t_43)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __pyx_t_44 = __Pyx_PyBool_FromLong(__pyx_v_self->named_type_names_must_be_unique); if (unlikely(!__pyx_t_44)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __pyx_t_45 = __Pyx_PyBool_FromLong(__pyx_v_self->null_pad_fixed); if (unlikely(!__pyx_t_45)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __pyx_t_46 = __Pyx_PyBool_FromLong(__pyx_v_self->raise_on_invalid_logical); if (unlikely(!__pyx_t_46)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __pyx_t_47 = __Pyx_PyBool_FromLong(__pyx_v_self->record_allow_extra_fields); if (unlikely(!__pyx_t_47)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __pyx_t_48 = __Pyx_PyBool_FromLong(__pyx_v_self->record_can_encode_dict); if (unlikely(!__pyx_t_48)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __pyx_t_49 = __Pyx_PyBool_FromLong(__pyx_v_self->record_decodes_to_dict); if (unlikely(!__pyx_t_49)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_49);
  __pyx_t_50 = __Pyx_PyBool_FromLong(__pyx_v_self->record_encode_use_defaults); if (unlikely(!__pyx_t_50)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_50);
  __pyx_t_51 = __Pyx_PyBool_FromLong(__pyx_v_self->record_fields_must_be_unique); if (unlikely(!__pyx_t_51)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_51);
  __pyx_t_52 = __Pyx_PyBool_FromLong(__pyx_v_self->record_values_type_hint); if (unlikely(!__pyx_t_52)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_52);
  __pyx_t_53 = __Pyx_PyBool_FromLong(__pyx_v_self->return_uuid_object); if (unlikely(!__pyx_t_53)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_53);
  __pyx_t_54 = __Pyx_PyBool_FromLong(__pyx_v_self->string_types_default_unchanged); if (unlikely(!__pyx_t_54)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_54);
  __pyx_t_55 = __Pyx_PyBool_FromLong(__pyx_v_self->truncate_fixed); if (unlikely(!__pyx_t_55)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_55);
  __pyx_t_56 = __Pyx_PyBool_FromLong(__pyx_v_self->types_str_to_schema); if (unlikely(!__pyx_t_56)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_56);
  __pyx_t_57 = __Pyx_PyBool_FromLong(__pyx_v_self->union_json_encodes_type_name); if (unlikely(!__pyx_t_57)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_57);
  __pyx_t_58 = PyTuple_New(63); if (unlikely(!__pyx_t_58)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 1, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 2, __pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 3, __pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 4, __pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 5, __pyx_t_6)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 6, __pyx_t_7)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 7, __pyx_t_8)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 8, __pyx_t_9)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 9, __pyx_t_10)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_11);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 10, __pyx_t_11)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 11, __pyx_t_12)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_13);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 12, __pyx_t_13)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_14);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 13, __pyx_t_14)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_15);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 14, __pyx_t_15)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->bytes_codec);
  __Pyx_GIVEREF(__pyx_v_self->bytes_codec);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 15, __pyx_v_self->bytes_codec)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_16);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 16, __pyx_t_16)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_17);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 17, __pyx_t_17)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_18);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 18, __pyx_t_18)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_19);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 19, __pyx_t_19)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_20);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 20, __pyx_t_20)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_21);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 21, __pyx_t_21)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_22);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 22, __pyx_t_22)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_23);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 23, __pyx_t_23)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_24);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 24, __pyx_t_24)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_25);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 25, __pyx_t_25)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_26);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 26, __pyx_t_26)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_27);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 27, __pyx_t_27)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_28);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 28, __pyx_t_28)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_29);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 29, __pyx_t_29)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_30);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 30, __pyx_t_30)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_31);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 31, __pyx_t_31)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_32);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 32, __pyx_t_32)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_33);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 33, __pyx_t_33)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_34);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 34, __pyx_t_34)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->externally_defined_types);
  __Pyx_GIVEREF(__pyx_v_self->externally_defined_types);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 35, __pyx_v_self->externally_defined_types)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_35);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 36, __pyx_t_35)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->fixed_codec);
  __Pyx_GIVEREF(__pyx_v_self->fixed_codec);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 37, __pyx_v_self->fixed_codec)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_36);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 38, __pyx_t_36)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_37);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 39, __pyx_t_37)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_38);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 40, __pyx_t_38)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_39);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 41, __pyx_t_39)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_40);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 42, __pyx_t_40)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->logical_types);
  __Pyx_GIVEREF(__pyx_v_self->logical_types);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 43, __pyx_v_self->logical_types)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_41);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 44, __pyx_t_41)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_42);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 45, __pyx_t_42)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_43);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 46, __pyx_t_43)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_44);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 47, __pyx_t_44)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_45);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 48, __pyx_t_45)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_46);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 49, __pyx_t_46)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_47);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 50, __pyx_t_47)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_48);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 51, __pyx_t_48)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_49);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 52, __pyx_t_49)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_50);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 53, __pyx_t_50)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_51);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 54, __pyx_t_51)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_52);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 55, __pyx_t_52)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_53);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 56, __pyx_t_53)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_54);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 57, __pyx_t_54)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_55);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 58, __pyx_t_55)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_56);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 59, __pyx_t_56)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->unicode_errors);
  __Pyx_GIVEREF(__pyx_v_self->unicode_errors);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 60, __pyx_v_self->unicode_errors)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->union_decodes_to);
  __Pyx_GIVEREF(__pyx_v_self->union_decodes_to);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 61, __pyx_v_self->union_decodes_to)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_57);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 62, __pyx_t_57)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_15 = 0;
  __pyx_t_16 = 0;
  __pyx_t_17 = 0;
  __pyx_t_18 = 0;
  __pyx_t_19 = 0;
  __pyx_t_20 = 0;
  __pyx_t_21 = 0;
  __pyx_t_22 = 0;
  __pyx_t_23 = 0;
  __pyx_t_24 = 0;
  __pyx_t_25 = 0;
  __pyx_t_26 = 0;
  __pyx_t_27 = 0;
  __pyx_t_28 = 0;
  __pyx_t_29 = 0;
  __pyx_t_30 = 0;
  __pyx_t_31 = 0;
  __pyx_t_32 = 0;
  __pyx_t_33 = 0;
  __pyx_t_34 = 0;
  __pyx_t_35 = 0;
  __pyx_t_36 = 0;
  __pyx_t_37 = 0;
  __pyx_t_38 = 0;
  __pyx_t_39 = 0;
  __pyx_t_40 = 0;
  __pyx_t_41 = 0;
  __pyx_t_42 = 0;
  __pyx_t_43 = 0;
  __pyx_t_44 = 0;
  __pyx_t_45 = 0;
  __pyx_t_46 = 0;
  __pyx_t_47 = 0;
  __pyx_t_48 = 0;
  __pyx_t_49 = 0;
  __pyx_t_50 = 0;
  __pyx_t_51 = 0;
  __pyx_t_52 = 0;
  __pyx_t_53 = 0;
  __pyx_t_54 = 0;
  __pyx_t_55 = 0;
  __pyx_t_56 = 0;
  __pyx_t_57 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_58);
  __pyx_t_58 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.adapt_record_types, self.allow_aliases_to_be_string, self.allow_duplicate_union_types, self.allow_empty_unions, self.allow_error_type, self.allow_false_values_for_null, self.allow_invalid_default_values, self.allow_leading_dot_in_names, self.allow_nested_unions, self.allow_primitive_name_collision, self.allow_primitive_names_in_namespaces, self.allow_tuple_notation, self.allow_union_default_any_member, self.alternate_timestamp_millis_encoding, self.ascii_name_rules, self.bytes_codec, self.bytes_default_value_utf8, self.canonical_form_repeat_enum, self.canonical_form_repeat_fixed, self.clamp_float_overflow, self.clamp_int_overflow, self.coerce_int_to_float, self.coerce_values_to_boolean, self.coerce_values_to_float, self.coerce_values_to_int, self.coerce_values_to_str, self.container_fill_blocks, self.date_type_accepts_string, self.decimal_check_exp_overflow, self.defer_schema_promotion_errors, self.enforce_enum_symbol_name_rules, self.enforce_namespace_name_rules, self.enforce_type_name_rules, self.enum_symbols_must_be_unique, self.expand_types_in_schema, self.externally_defined_types, self.fingerprint_returns_digest, self.fixed_codec, self.float_out_of_range_inf, self.inline_namespaces, self.invalid_value_include_array_index, self.invalid_value_include_map_key, self.invalid_value_includes_record_name, self.logical_types, self.missing_values_can_be_empty_container, self.missing_values_can_be_null, self.naive_dt_assume_utc, self.named_type_names_must_be_unique, self.null_pad_fixed, self.raise_on_invalid_logical, self.record_allow_extra_fields, self.record_can_encode_dict, self.record_decodes_to_dict, self.record_encode_use_defaults, self.record_fields_must_be_unique, self.record_values_type_hint, self.return_uuid_object, self.string_types_default_unchanged, self.truncate_fixed, self.types_str_to_schema, self.unicode_errors, self.union_decodes_to, self.union_json_encodes_type_name)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_58 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_58)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_58);
  __pyx_v__dict = __pyx_t_58;
  __pyx_t_58 = 0;

  /* "(tree fragment)":7
 *     state = (self.adapt_record_types, self.allow_aliases_to_be_string, self.allow_duplicate_union_types, self.allow_empty_unions, self.allow_error_type, self.allow_false_values_for_null, self.allow_invalid_default_values, self.allow_leading_dot_in_names, self.allow_nested_unions, self.allow_primitive_name_collision, self.allow_primitive_names_in_namespaces, self.allow_tuple_notation, self.allow_union_default_any_member, self.alternate_timestamp_millis_encoding, self.ascii_name_rules, self.bytes_codec, self.bytes_default_value_utf8, self.canonical_form_repeat_enum, self.canonical_form_repeat_fixed, self.clamp_float_overflow, self.clamp_int_overflow, self.coerce_int_to_float, self.coerce_values_to_boolean, self.coerce_values_to_float, self.coerce_values_to_int, self.coerce_values_to_str, self.container_fill_blocks, self.date_type_accepts_string, self.decimal_check_exp_overflow, self.defer_schema_promotion_errors, self.enforce_enum_symbol_name_rules, self.enforce_namespace_name_rules, self.enforce_type_name_rules, self.enum_symbols_must_be_unique, self.expand_types_in_schema, self.externally_defined_types, self.fingerprint_returns_digest, self.fixed_codec, self.float_out_of_range_inf, self.inline_namespaces, self.invalid_value_include_array_index, self.invalid_value_include_map_key, self.invalid_value_includes_record_name, self.logical_types, self.missing_values_can_be_empty_container, self.missing_values_can_be_null, self.naive_dt_assume_utc, self.named_type_names_must_be_unique, self.null_pad_fixed, self.raise_on_invalid_logical, self.record_allow_extra_fields, self.record_can_encode_dict, self.record_decodes_to_dict, self.record_encode_use_defaults, self.record_fields_must_be_unique, self.record_values_type_hint, self.return_uuid_object, self.string_types_default_unchanged, self.truncate_fixed, self.types_str_to_schema, self.unicode_errors, self.union_decodes_to, self.union_json_encodes_type_name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_59 = (__pyx_v__dict != Py_None);
  if (__pyx_t_59) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_58 = PyTuple_New(1); if (unlikely(!__pyx_t_58)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_58);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_57 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_58); if (unlikely(!__pyx_t_57)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_57);
    __Pyx_DECREF(__pyx_t_58); __pyx_t_58 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_57));
    __pyx_t_57 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.bytes_codec is not None or self.externally_defined_types is not None or self.fixed_codec is not None or self.logical_types is not None or self.unicode_errors is not None or self.union_decodes_to is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.adapt_record_types, self.allow_aliases_to_be_string, self.allow_duplicate_union_types, self.allow_empty_unions, self.allow_error_type, self.allow_false_values_for_null, self.allow_invalid_default_values, self.allow_leading_dot_in_names, self.allow_nested_unions, self.allow_primitive_name_collision, self.allow_primitive_names_in_namespaces, self.allow_tuple_notation, self.allow_union_default_any_member, self.alternate_timestamp_millis_encoding, self.ascii_name_rules, self.bytes_codec, self.bytes_default_value_utf8, self.canonical_form_repeat_enum, self.canonical_form_repeat_fixed, self.clamp_float_overflow, self.clamp_int_overflow, self.coerce_int_to_float, self.coerce_values_to_boolean, self.coerce_values_to_float, self.coerce_values_to_int, self.coerce_values_to_str, self.container_fill_blocks, self.date_type_accepts_string, self.decimal_check_exp_overflow, self.defer_schema_promotion_errors, self.enforce_enum_symbol_name_rules, self.enforce_namespace_name_rules, self.enforce_type_name_rules, self.enum_symbols_must_be_unique, self.expand_types_in_schema, self.externally_defined_types, self.fingerprint_returns_digest, self.fixed_codec, self.float_out_of_range_inf, self.inline_namespaces, self.invalid_value_include_array_index, self.invalid_value_include_map_key, self.invalid_value_includes_record_name, self.logical_types, self.missing_values_can_be_empty_container, self.missing_values_can_be_null, self.naive_dt_assume_utc, self.named_type_names_must_be_unique, self.null_pad_fixed, self.raise_on_invalid_logical, self.record_allow_extra_fields, self.record_can_encode_dict, self.record_decodes_to_dict, self.record_encode_use_defaults, self.record_fields_must_be_unique, self.record_values_type_hint, self.return_uuid_object, self.string_types_default_unchanged, self.truncate_fixed, self.types_str_to_schema, self.unicode_errors, self.union_decodes_to, self.union_json_encodes_type_name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.bytes_codec is not None or self.externally_defined_types is not None or self.fixed_codec is not None or self.logical_types is not None or self.unicode_errors is not None or self.union_decodes_to is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Options, (type(self), 0xc466972, None), state
 */
  /*else*/ {
    __pyx_t_60 = (__pyx_v_self->bytes_codec != ((PyObject*)Py_None));
    if (!__pyx_t_60) {
    } else {
      __pyx_t_59 = __pyx_t_60;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_60 = (__pyx_v_self->externally_defined_types != Py_None);
    if (!__pyx_t_60) {
    } else {
      __pyx_t_59 = __pyx_t_60;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_60 = (__pyx_v_self->fixed_codec != ((PyObject*)Py_None));
    if (!__pyx_t_60) {
    } else {
      __pyx_t_59 = __pyx_t_60;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_60 = (__pyx_v_self->logical_types != ((PyObject*)Py_None));
    if (!__pyx_t_60) {
    } else {
      __pyx_t_59 = __pyx_t_60;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_60 = (__pyx_v_self->unicode_errors != ((PyObject*)Py_None));
    if (!__pyx_t_60) {
    } else {
      __pyx_t_59 = __pyx_t_60;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_60 = (__pyx_v_self->union_decodes_to != Py_None);
    __pyx_t_59 = __pyx_t_60;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_59;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.bytes_codec is not None or self.externally_defined_types is not None or self.fixed_codec is not None or self.logical_types is not None or self.unicode_errors is not None or self.union_decodes_to is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Options, (type(self), 0xc466972, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.bytes_codec is not None or self.externally_defined_types is not None or self.fixed_codec is not None or self.logical_types is not None or self.unicode_errors is not None or self.union_decodes_to is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Options, (type(self), 0xc466972, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Options, (type(self), 0xc466972, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_57, __pyx_n_s_pyx_unpickle_Options); if (unlikely(!__pyx_t_57)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_57);
    __pyx_t_58 = PyTuple_New(3); if (unlikely(!__pyx_t_58)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_58);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_205941106);
    __Pyx_GIVEREF(__pyx_int_205941106);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 1, __pyx_int_205941106)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_56 = PyTuple_New(3); if (unlikely(!__pyx_t_56)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_56);
    __Pyx_GIVEREF(__pyx_t_57);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_56, 0, __pyx_t_57)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_58);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_56, 1, __pyx_t_58)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_56, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_57 = 0;
    __pyx_t_58 = 0;
    __pyx_r = __pyx_t_56;
    __pyx_t_56 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.bytes_codec is not None or self.externally_defined_types is not None or self.fixed_codec is not None or self.logical_types is not None or self.unicode_errors is not None or self.union_decodes_to is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Options, (type(self), 0xc466972, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Options, (type(self), 0xc466972, None), state
 *     else:
 *         return __pyx_unpickle_Options, (type(self), 0xc466972, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Options__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_56, __pyx_n_s_pyx_unpickle_Options); if (unlikely(!__pyx_t_56)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_56);
    __pyx_t_58 = PyTuple_New(3); if (unlikely(!__pyx_t_58)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_58);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_205941106);
    __Pyx_GIVEREF(__pyx_int_205941106);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 1, __pyx_int_205941106)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_58, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_57 = PyTuple_New(2); if (unlikely(!__pyx_t_57)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_57);
    __Pyx_GIVEREF(__pyx_t_56);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_57, 0, __pyx_t_56)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_58);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_57, 1, __pyx_t_58)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_56 = 0;
    __pyx_t_58 = 0;
    __pyx_r = __pyx_t_57;
    __pyx_t_57 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __Pyx_XDECREF(__pyx_t_34);
  __Pyx_XDECREF(__pyx_t_35);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_37);
  __Pyx_XDECREF(__pyx_t_38);
  __Pyx_XDECREF(__pyx_t_39);
  __Pyx_XDECREF(__pyx_t_40);
  __Pyx_XDECREF(__pyx_t_41);
  __Pyx_XDECREF(__pyx_t_42);
  __Pyx_XDECREF(__pyx_t_43);
  __Pyx_XDECREF(__pyx_t_44);
  __Pyx_XDECREF(__pyx_t_45);
  __Pyx_XDECREF(__pyx_t_46);
  __Pyx_XDECREF(__pyx_t_47);
  __Pyx_XDECREF(__pyx_t_48);
  __Pyx_XDECREF(__pyx_t_49);
  __Pyx_XDECREF(__pyx_t_50);
  __Pyx_XDECREF(__pyx_t_51);
  __Pyx_XDECREF(__pyx_t_52);
  __Pyx_XDECREF(__pyx_t_53);
  __Pyx_XDECREF(__pyx_t_54);
  __Pyx_XDECREF(__pyx_t_55);
  __Pyx_XDECREF(__pyx_t_56);
  __Pyx_XDECREF(__pyx_t_57);
  __Pyx_XDECREF(__pyx_t_58);
  __Pyx_AddTraceback("cavro.Options.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Options, (type(self), 0xc466972, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Options__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7Options_21__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7Options_21__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7Options_21__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7Options_21__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Options.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_7Options_20__setstate_cython__(((struct __pyx_obj_5cavro_Options *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7Options_20__setstate_cython__(struct __pyx_obj_5cavro_Options *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Options, (type(self), 0xc466972, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Options__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_Options__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Options, (type(self), 0xc466972, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Options__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Options.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":6
 * cdef class _Writer:
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement write_u8")
 */

static int __pyx_f_5cavro_7_Writer_write_u8(struct __pyx_obj_5cavro__Writer *__pyx_v_self, CYTHON_UNUSED uint8_t __pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_u8", 1);

  /* "src/io.pxi":8
 *     cdef int write_u8(self, uint8_t val) except -1:
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement write_u8")             # <<<<<<<<<<<<<<
 * 
 *     cdef int write_n(self, const uint8_t[:] data) except -1:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement_write_u8); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/io.pxi":7
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement write_u8")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(5, 7, __pyx_L1_error)

  /* "src/io.pxi":6
 * cdef class _Writer:
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement write_u8")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro._Writer.write_u8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":10
 *             f"{type(self).__name__} does not implement write_u8")
 * 
 *     cdef int write_n(self, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement write_n")
 */

static int __pyx_f_5cavro_7_Writer_write_n(struct __pyx_obj_5cavro__Writer *__pyx_v_self, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_data) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_n", 1);

  /* "src/io.pxi":12
 *     cdef int write_n(self, const uint8_t[:] data) except -1:
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement write_n")             # <<<<<<<<<<<<<<
 * 
 *     cdef int flush(self) except -1:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement_write_n); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/io.pxi":11
 * 
 *     cdef int write_n(self, const uint8_t[:] data) except -1:
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement write_n")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(5, 11, __pyx_L1_error)

  /* "src/io.pxi":10
 *             f"{type(self).__name__} does not implement write_u8")
 * 
 *     cdef int write_n(self, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement write_n")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro._Writer.write_n", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":14
 *             f"{type(self).__name__} does not implement write_n")
 * 
 *     cdef int flush(self) except -1:             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static int __pyx_f_5cavro_7_Writer_flush(CYTHON_UNUSED struct __pyx_obj_5cavro__Writer *__pyx_v_self) {
  int __pyx_r;

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7_Writer_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7_Writer_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7_Writer_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7_Writer_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_7_Writer___reduce_cython__(((struct __pyx_obj_5cavro__Writer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7_Writer___reduce_cython__(struct __pyx_obj_5cavro__Writer *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__Writer, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__Writer, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle__Writer, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__Writer, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle__Writer); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__Writer, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__Writer, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle__Writer, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__Writer__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle__Writer); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._Writer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Writer, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Writer__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7_Writer_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7_Writer_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7_Writer_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7_Writer_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._Writer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_7_Writer_2__setstate_cython__(((struct __pyx_obj_5cavro__Writer *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7_Writer_2__setstate_cython__(struct __pyx_obj_5cavro__Writer *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__Writer, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__Writer__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__Writer__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Writer, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Writer__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._Writer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":20
 * cdef class _Reader:
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement read_u8")
 */

static uint8_t __pyx_f_5cavro_7_Reader_read_u8(struct __pyx_obj_5cavro__Reader *__pyx_v_self) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_u8", 1);

  /* "src/io.pxi":22
 *     cdef uint8_t read_u8(self) except? 0xba:
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement read_u8")             # <<<<<<<<<<<<<<
 * 
 *     cdef bytes read_bytes(self, Py_ssize_t n):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement_read_u8); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/io.pxi":21
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement read_u8")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(5, 21, __pyx_L1_error)

  /* "src/io.pxi":20
 * cdef class _Reader:
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement read_u8")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro._Reader.read_u8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0xba;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":24
 *             f"{type(self).__name__} does not implement read_u8")
 * 
 *     cdef bytes read_bytes(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement read_bytes")
 */

static PyObject *__pyx_f_5cavro_7_Reader_read_bytes(struct __pyx_obj_5cavro__Reader *__pyx_v_self, CYTHON_UNUSED Py_ssize_t __pyx_v_n) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_bytes", 1);

  /* "src/io.pxi":26
 *     cdef bytes read_bytes(self, Py_ssize_t n):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement read_bytes")             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement_read_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/io.pxi":25
 * 
 *     cdef bytes read_bytes(self, Py_ssize_t n):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement read_bytes")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(5, 25, __pyx_L1_error)

  /* "src/io.pxi":24
 *             f"{type(self).__name__} does not implement read_u8")
 * 
 *     cdef bytes read_bytes(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement read_bytes")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro._Reader.read_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":28
 *             f"{type(self).__name__} does not implement read_bytes")
 * 
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement read_n")
 */

static __Pyx_memviewslice __pyx_f_5cavro_7_Reader_read_n(struct __pyx_obj_5cavro__Reader *__pyx_v_self, CYTHON_UNUSED Py_ssize_t __pyx_v_n) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_n", 1);

  /* "src/io.pxi":30
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement read_n")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement_read_n); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/io.pxi":29
 * 
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement read_n")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(5, 29, __pyx_L1_error)

  /* "src/io.pxi":28
 *             f"{type(self).__name__} does not implement read_bytes")
 * 
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement read_n")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("cavro._Reader.read_n", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7_Reader_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7_Reader_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7_Reader_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7_Reader_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_7_Reader___reduce_cython__(((struct __pyx_obj_5cavro__Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7_Reader___reduce_cython__(struct __pyx_obj_5cavro__Reader *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__Reader, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__Reader, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle__Reader, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__Reader, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle__Reader); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__Reader, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__Reader, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle__Reader, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__Reader__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle__Reader); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Reader, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Reader__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7_Reader_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7_Reader_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7_Reader_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7_Reader_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_7_Reader_2__setstate_cython__(((struct __pyx_obj_5cavro__Reader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7_Reader_2__setstate_cython__(struct __pyx_obj_5cavro__Reader *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__Reader, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__Reader__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__Reader__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Reader, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Reader__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":46
 *     cdef object file_obj
 * 
 *     def __init__(self, file_obj):             # <<<<<<<<<<<<<<
 *         self.file_obj = file_obj
 * 
 */

/* Python wrapper */
static int __pyx_pw_5cavro_10FileReader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_10FileReader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_file_obj = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_file_obj,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_file_obj)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 46, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(5, 46, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_file_obj = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(5, 46, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.FileReader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10FileReader___init__(((struct __pyx_obj_5cavro_FileReader *)__pyx_v_self), __pyx_v_file_obj);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_10FileReader___init__(struct __pyx_obj_5cavro_FileReader *__pyx_v_self, PyObject *__pyx_v_file_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/io.pxi":47
 * 
 *     def __init__(self, file_obj):
 *         self.file_obj = file_obj             # <<<<<<<<<<<<<<
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:
 */
  __Pyx_INCREF(__pyx_v_file_obj);
  __Pyx_GIVEREF(__pyx_v_file_obj);
  __Pyx_GOTREF(__pyx_v_self->file_obj);
  __Pyx_DECREF(__pyx_v_self->file_obj);
  __pyx_v_self->file_obj = __pyx_v_file_obj;

  /* "src/io.pxi":46
 *     cdef object file_obj
 * 
 *     def __init__(self, file_obj):             # <<<<<<<<<<<<<<
 *         self.file_obj = file_obj
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":49
 *         self.file_obj = file_obj
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:             # <<<<<<<<<<<<<<
 *         cdef bytes data = self.file_obj.read(1)
 *         if len(data) != 1:
 */

static uint8_t __pyx_f_5cavro_10FileReader_read_u8(struct __pyx_obj_5cavro_FileReader *__pyx_v_self) {
  PyObject *__pyx_v_data = 0;
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  long __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_u8", 1);

  /* "src/io.pxi":50
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:
 *         cdef bytes data = self.file_obj.read(1)             # <<<<<<<<<<<<<<
 *         if len(data) != 1:
 *             raise EOFError(f"End of file found trying to read 1 byte")
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->file_obj, __pyx_n_s_read); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_int_1};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(5, 50, __pyx_L1_error)
  __pyx_v_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/io.pxi":51
 *     cdef uint8_t read_u8(self) except? 0xba:
 *         cdef bytes data = self.file_obj.read(1)
 *         if len(data) != 1:             # <<<<<<<<<<<<<<
 *             raise EOFError(f"End of file found trying to read 1 byte")
 *         return ord(data)
 */
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(5, 51, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyBytes_GET_SIZE(__pyx_v_data); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(5, 51, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 1);
  if (unlikely(__pyx_t_6)) {

    /* "src/io.pxi":52
 *         cdef bytes data = self.file_obj.read(1)
 *         if len(data) != 1:
 *             raise EOFError(f"End of file found trying to read 1 byte")             # <<<<<<<<<<<<<<
 *         return ord(data)
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_EOFError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(5, 52, __pyx_L1_error)

    /* "src/io.pxi":51
 *     cdef uint8_t read_u8(self) except? 0xba:
 *         cdef bytes data = self.file_obj.read(1)
 *         if len(data) != 1:             # <<<<<<<<<<<<<<
 *             raise EOFError(f"End of file found trying to read 1 byte")
 *         return ord(data)
 */
  }

  /* "src/io.pxi":53
 *         if len(data) != 1:
 *             raise EOFError(f"End of file found trying to read 1 byte")
 *         return ord(data)             # <<<<<<<<<<<<<<
 * 
 *     cdef bytes read_bytes(self, Py_ssize_t n):
 */
  __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_v_data); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(5, 53, __pyx_L1_error)
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "src/io.pxi":49
 *         self.file_obj = file_obj
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:             # <<<<<<<<<<<<<<
 *         cdef bytes data = self.file_obj.read(1)
 *         if len(data) != 1:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.FileReader.read_u8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0xba;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":55
 *         return ord(data)
 * 
 *     cdef bytes read_bytes(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         cdef bytes result = self.file_obj.read(n)
 *         if len(result) != n:
 */

static PyObject *__pyx_f_5cavro_10FileReader_read_bytes(struct __pyx_obj_5cavro_FileReader *__pyx_v_self, Py_ssize_t __pyx_v_n) {
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_bytes", 1);

  /* "src/io.pxi":56
 * 
 *     cdef bytes read_bytes(self, Py_ssize_t n):
 *         cdef bytes result = self.file_obj.read(n)             # <<<<<<<<<<<<<<
 *         if len(result) != n:
 *             raise EOFError(f"End of file found trying to read {n} bytes")
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->file_obj, __pyx_n_s_read); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(5, 56, __pyx_L1_error)
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/io.pxi":57
 *     cdef bytes read_bytes(self, Py_ssize_t n):
 *         cdef bytes result = self.file_obj.read(n)
 *         if len(result) != n:             # <<<<<<<<<<<<<<
 *             raise EOFError(f"End of file found trying to read {n} bytes")
 *         return result
 */
  if (unlikely(__pyx_v_result == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(5, 57, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyBytes_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(5, 57, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != __pyx_v_n);
  if (unlikely(__pyx_t_7)) {

    /* "src/io.pxi":58
 *         cdef bytes result = self.file_obj.read(n)
 *         if len(result) != n:
 *             raise EOFError(f"End of file found trying to read {n} bytes")             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_8 = 127;
    __Pyx_INCREF(__pyx_kp_u_End_of_file_found_trying_to_read_2);
    __pyx_t_6 += 33;
    __Pyx_GIVEREF(__pyx_kp_u_End_of_file_found_trying_to_read_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_End_of_file_found_trying_to_read_2);
    __pyx_t_2 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_n, 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_bytes);
    __pyx_t_6 += 6;
    __Pyx_GIVEREF(__pyx_kp_u_bytes);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_bytes);
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_6, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_EOFError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(5, 58, __pyx_L1_error)

    /* "src/io.pxi":57
 *     cdef bytes read_bytes(self, Py_ssize_t n):
 *         cdef bytes result = self.file_obj.read(n)
 *         if len(result) != n:             # <<<<<<<<<<<<<<
 *             raise EOFError(f"End of file found trying to read {n} bytes")
 *         return result
 */
  }

  /* "src/io.pxi":59
 *         if len(result) != n:
 *             raise EOFError(f"End of file found trying to read {n} bytes")
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "src/io.pxi":55
 *         return ord(data)
 * 
 *     cdef bytes read_bytes(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         cdef bytes result = self.file_obj.read(n)
 *         if len(result) != n:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.FileReader.read_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":61
 *         return result
 * 
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         cdef bytes result = self.file_obj.read(n)
 *         if len(result) != n:
 */

static __Pyx_memviewslice __pyx_f_5cavro_10FileReader_read_n(struct __pyx_obj_5cavro_FileReader *__pyx_v_self, Py_ssize_t __pyx_v_n) {
  PyObject *__pyx_v_result = 0;
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_n", 1);

  /* "src/io.pxi":62
 * 
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):
 *         cdef bytes result = self.file_obj.read(n)             # <<<<<<<<<<<<<<
 *         if len(result) != n:
 *             raise EOFError(f"End of file found trying to read {n} bytes")
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->file_obj, __pyx_n_s_read); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(5, 62, __pyx_L1_error)
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/io.pxi":63
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):
 *         cdef bytes result = self.file_obj.read(n)
 *         if len(result) != n:             # <<<<<<<<<<<<<<
 *             raise EOFError(f"End of file found trying to read {n} bytes")
 *         return result
 */
  if (unlikely(__pyx_v_result == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(5, 63, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyBytes_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(5, 63, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != __pyx_v_n);
  if (unlikely(__pyx_t_7)) {

    /* "src/io.pxi":64
 *         cdef bytes result = self.file_obj.read(n)
 *         if len(result) != n:
 *             raise EOFError(f"End of file found trying to read {n} bytes")             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_8 = 127;
    __Pyx_INCREF(__pyx_kp_u_End_of_file_found_trying_to_read_2);
    __pyx_t_6 += 33;
    __Pyx_GIVEREF(__pyx_kp_u_End_of_file_found_trying_to_read_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_End_of_file_found_trying_to_read_2);
    __pyx_t_2 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_n, 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_bytes);
    __pyx_t_6 += 6;
    __Pyx_GIVEREF(__pyx_kp_u_bytes);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_bytes);
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_6, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_EOFError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(5, 64, __pyx_L1_error)

    /* "src/io.pxi":63
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):
 *         cdef bytes result = self.file_obj.read(n)
 *         if len(result) != n:             # <<<<<<<<<<<<<<
 *             raise EOFError(f"End of file found trying to read {n} bytes")
 *         return result
 */
  }

  /* "src/io.pxi":65
 *         if len(result) != n:
 *             raise EOFError(f"End of file found trying to read {n} bytes")
 *         return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_result, 0); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(5, 65, __pyx_L1_error)
  __pyx_r = __pyx_t_9;
  __pyx_t_9.memview = NULL;
  __pyx_t_9.data = NULL;
  goto __pyx_L0;

  /* "src/io.pxi":61
 *         return result
 * 
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         cdef bytes result = self.file_obj.read(n)
 *         if len(result) != n:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_9, 1);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("cavro.FileReader.read_n", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10FileReader_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10FileReader_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10FileReader_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10FileReader_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10FileReader_2__reduce_cython__(((struct __pyx_obj_5cavro_FileReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10FileReader_2__reduce_cython__(struct __pyx_obj_5cavro_FileReader *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.file_obj,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->file_obj);
  __Pyx_GIVEREF(__pyx_v_self->file_obj);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->file_obj)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.file_obj,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.file_obj,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.file_obj is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.file_obj,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.file_obj is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_FileReader, (type(self), 0x74c6201, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->file_obj != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.file_obj is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FileReader, (type(self), 0x74c6201, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.file_obj is not None
 *     if use_setstate:
 *         return __pyx_unpickle_FileReader, (type(self), 0x74c6201, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_FileReader, (type(self), 0x74c6201, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_FileReader); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_122446337);
    __Pyx_GIVEREF(__pyx_int_122446337);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_122446337)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.file_obj is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FileReader, (type(self), 0x74c6201, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_FileReader, (type(self), 0x74c6201, None), state
 *     else:
 *         return __pyx_unpickle_FileReader, (type(self), 0x74c6201, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FileReader__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_FileReader); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_122446337);
    __Pyx_GIVEREF(__pyx_int_122446337);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_122446337)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.FileReader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FileReader, (type(self), 0x74c6201, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FileReader__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10FileReader_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10FileReader_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10FileReader_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10FileReader_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.FileReader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10FileReader_4__setstate_cython__(((struct __pyx_obj_5cavro_FileReader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10FileReader_4__setstate_cython__(struct __pyx_obj_5cavro_FileReader *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_FileReader, (type(self), 0x74c6201, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FileReader__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_FileReader__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FileReader, (type(self), 0x74c6201, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FileReader__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FileReader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":77
 *     cdef object file_obj
 * 
 *     def __init__(self, file_obj):             # <<<<<<<<<<<<<<
 *         self.file_obj = file_obj
 * 
 */

/* Python wrapper */
static int __pyx_pw_5cavro_10FileWriter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_10FileWriter_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_file_obj = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_file_obj,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_file_obj)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 77, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(5, 77, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_file_obj = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(5, 77, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.FileWriter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10FileWriter___init__(((struct __pyx_obj_5cavro_FileWriter *)__pyx_v_self), __pyx_v_file_obj);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_10FileWriter___init__(struct __pyx_obj_5cavro_FileWriter *__pyx_v_self, PyObject *__pyx_v_file_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/io.pxi":78
 * 
 *     def __init__(self, file_obj):
 *         self.file_obj = file_obj             # <<<<<<<<<<<<<<
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:
 */
  __Pyx_INCREF(__pyx_v_file_obj);
  __Pyx_GIVEREF(__pyx_v_file_obj);
  __Pyx_GOTREF(__pyx_v_self->file_obj);
  __Pyx_DECREF(__pyx_v_self->file_obj);
  __pyx_v_self->file_obj = __pyx_v_file_obj;

  /* "src/io.pxi":77
 *     cdef object file_obj
 * 
 *     def __init__(self, file_obj):             # <<<<<<<<<<<<<<
 *         self.file_obj = file_obj
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":80
 *         self.file_obj = file_obj
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:             # <<<<<<<<<<<<<<
 *         self.file_obj.write(bytes([val]))
 * 
 */

static int __pyx_f_5cavro_10FileWriter_write_u8(struct __pyx_obj_5cavro_FileWriter *__pyx_v_self, uint8_t __pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_u8", 1);

  /* "src/io.pxi":81
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:
 *         self.file_obj.write(bytes([val]))             # <<<<<<<<<<<<<<
 * 
 *     cdef int write_n(self, const uint8_t[:] data) except -1:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->file_obj, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_uint8_t(__pyx_v_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(5, 81, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/io.pxi":80
 *         self.file_obj = file_obj
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:             # <<<<<<<<<<<<<<
 *         self.file_obj.write(bytes([val]))
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.FileWriter.write_u8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":83
 *         self.file_obj.write(bytes([val]))
 * 
 *     cdef int write_n(self, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         self.file_obj.write(data)
 * 
 */

static int __pyx_f_5cavro_10FileWriter_write_n(struct __pyx_obj_5cavro_FileWriter *__pyx_v_self, __Pyx_memviewslice __pyx_v_data) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_n", 1);

  /* "src/io.pxi":84
 * 
 *     cdef int write_n(self, const uint8_t[:] data) except -1:
 *         self.file_obj.write(data)             # <<<<<<<<<<<<<<
 * 
 *     cdef int flush(self) except -1:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->file_obj, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/io.pxi":83
 *         self.file_obj.write(bytes([val]))
 * 
 *     cdef int write_n(self, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         self.file_obj.write(data)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.FileWriter.write_n", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/io.pxi":86
 *         self.file_obj.write(data)
 * 
 *     cdef int flush(self) except -1:             # <<<<<<<<<<<<<<
 *         self.file_obj.flush()
 */

static int __pyx_f_5cavro_10FileWriter_flush(struct __pyx_obj_5cavro_FileWriter *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("flush", 1);

  /* "src/io.pxi":87
 * 
 *     cdef int flush(self) except -1:
 *         self.file_obj.flush()             # <<<<<<<<<<<<<<
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->file_obj, __pyx_n_s_flush); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/io.pxi":86
 *         self.file_obj.write(data)
 * 
 *     cdef int flush(self) except -1:             # <<<<<<<<<<<<<<
 *         self.file_obj.flush()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.FileWriter.flush", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10FileWriter_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10FileWriter_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10FileWriter_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10FileWriter_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10FileWriter_2__reduce_cython__(((struct __pyx_obj_5cavro_FileWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10FileWriter_2__reduce_cython__(struct __pyx_obj_5cavro_FileWriter *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.file_obj,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->file_obj);
  __Pyx_GIVEREF(__pyx_v_self->file_obj);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->file_obj)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.file_obj,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.file_obj,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.file_obj is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.file_obj,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.file_obj is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_FileWriter, (type(self), 0x74c6201, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->file_obj != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.file_obj is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FileWriter, (type(self), 0x74c6201, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.file_obj is not None
 *     if use_setstate:
 *         return __pyx_unpickle_FileWriter, (type(self), 0x74c6201, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_FileWriter, (type(self), 0x74c6201, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_FileWriter); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_122446337);
    __Pyx_GIVEREF(__pyx_int_122446337);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_122446337)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.file_obj is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FileWriter, (type(self), 0x74c6201, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_FileWriter, (type(self), 0x74c6201, None), state
 *     else:
 *         return __pyx_unpickle_FileWriter, (type(self), 0x74c6201, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FileWriter__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_FileWriter); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_122446337);
    __Pyx_GIVEREF(__pyx_int_122446337);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_122446337)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.FileWriter.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FileWriter, (type(self), 0x74c6201, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FileWriter__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10FileWriter_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10FileWriter_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10FileWriter_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10FileWriter_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.FileWriter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10FileWriter_4__setstate_cython__(((struct __pyx_obj_5cavro_FileWriter *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10FileWriter_4__setstate_cython__(struct __pyx_obj_5cavro_FileWriter *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_FileWriter, (type(self), 0x74c6201, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FileWriter__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_FileWriter__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FileWriter, (type(self), 0x74c6201, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FileWriter__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FileWriter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/buffer.pxi":18
 *     cdef readonly size_t len
 * 
 *     def __cinit__(self, initial_size=4096):             # <<<<<<<<<<<<<<
 *         self.buffer = array.clone(byte_buffer_template, initial_size, zero=True)
 *         self.len = 0
 */

/* Python wrapper */
static int __pyx_pw_5cavro_12MemoryWriter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_12MemoryWriter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_initial_size = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_initial_size,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_int_4096));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_initial_size);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 18, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(16, 18, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_initial_size = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 1, __pyx_nargs); __PYX_ERR(16, 18, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.MemoryWriter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_12MemoryWriter___cinit__(((struct __pyx_obj_5cavro_MemoryWriter *)__pyx_v_self), __pyx_v_initial_size);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_12MemoryWriter___cinit__(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self, PyObject *__pyx_v_initial_size) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "src/buffer.pxi":19
 * 
 *     def __cinit__(self, initial_size=4096):
 *         self.buffer = array.clone(byte_buffer_template, initial_size, zero=True)             # <<<<<<<<<<<<<<
 *         self.len = 0
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_5cavro_byte_buffer_template);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyIndex_AsSsize_t(__pyx_v_initial_size); if (unlikely((__pyx_t_2 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(16, 19, __pyx_L1_error)
  __pyx_t_3 = ((PyObject *)__pyx_f_7cpython_5array_clone(((arrayobject *)__pyx_t_1), __pyx_t_2, 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->buffer);
  __Pyx_DECREF((PyObject *)__pyx_v_self->buffer);
  __pyx_v_self->buffer = ((arrayobject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/buffer.pxi":20
 *     def __cinit__(self, initial_size=4096):
 *         self.buffer = array.clone(byte_buffer_template, initial_size, zero=True)
 *         self.len = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef bytes bytes(self):
 */
  __pyx_v_self->len = 0;

  /* "src/buffer.pxi":18
 *     cdef readonly size_t len
 * 
 *     def __cinit__(self, initial_size=4096):             # <<<<<<<<<<<<<<
 *         self.buffer = array.clone(byte_buffer_template, initial_size, zero=True)
 *         self.len = 0
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.MemoryWriter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/buffer.pxi":22
 *         self.len = 0
 * 
 *     cdef bytes bytes(self):             # <<<<<<<<<<<<<<
 *         return self.buffer.data.as_chars[:self.len]
 * 
 */

static PyObject *__pyx_f_5cavro_12MemoryWriter_bytes(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bytes", 1);

  /* "src/buffer.pxi":23
 * 
 *     cdef bytes bytes(self):
 *         return self.buffer.data.as_chars[:self.len]             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t[:] view(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_self->buffer->data.as_chars + 0, __pyx_v_self->len - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/buffer.pxi":22
 *         self.len = 0
 * 
 *     cdef bytes bytes(self):             # <<<<<<<<<<<<<<
 *         return self.buffer.data.as_chars[:self.len]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.MemoryWriter.bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/buffer.pxi":25
 *         return self.buffer.data.as_chars[:self.len]
 * 
 *     cdef const uint8_t[:] view(self):             # <<<<<<<<<<<<<<
 *         return self.buffer.data.as_uchars[:self.len]
 * 
 */

static __Pyx_memviewslice __pyx_f_5cavro_12MemoryWriter_view(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("view", 1);

  /* "src/buffer.pxi":26
 * 
 *     cdef const uint8_t[:] view(self):
 *         return self.buffer.data.as_uchars[:self.len]             # <<<<<<<<<<<<<<
 * 
 *     cdef reset(self):
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_self->buffer->data.as_uchars) + 0, __pyx_v_self->len - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_1, 0); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(16, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  goto __pyx_L0;

  /* "src/buffer.pxi":25
 *         return self.buffer.data.as_chars[:self.len]
 * 
 *     cdef const uint8_t[:] view(self):             # <<<<<<<<<<<<<<
 *         return self.buffer.data.as_uchars[:self.len]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_2, 1);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("cavro.MemoryWriter.view", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/buffer.pxi":28
 *         return self.buffer.data.as_uchars[:self.len]
 * 
 *     cdef reset(self):             # <<<<<<<<<<<<<<
 *         self.len = 0
 * 
 */

static PyObject *__pyx_f_5cavro_12MemoryWriter_reset(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset", 1);

  /* "src/buffer.pxi":29
 * 
 *     cdef reset(self):
 *         self.len = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:
 */
  __pyx_v_self->len = 0;

  /* "src/buffer.pxi":28
 *         return self.buffer.data.as_uchars[:self.len]
 * 
 *     cdef reset(self):             # <<<<<<<<<<<<<<
 *         self.len = 0
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/buffer.pxi":31
 *         self.len = 0
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:             # <<<<<<<<<<<<<<
 *         if self.buffer.ob_size - self.len < 1:
 *             array.resize_smart(self.buffer, self.len + 1)
 */

static int __pyx_f_5cavro_12MemoryWriter_write_u8(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self, uint8_t __pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_u8", 1);

  /* "src/buffer.pxi":32
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:
 *         if self.buffer.ob_size - self.len < 1:             # <<<<<<<<<<<<<<
 *             array.resize_smart(self.buffer, self.len + 1)
 *         self.buffer.data.as_uchars[self.len] = val
 */
  __pyx_t_1 = ((__pyx_v_self->buffer->ob_size - __pyx_v_self->len) < 1);
  if (__pyx_t_1) {

    /* "src/buffer.pxi":33
 *     cdef int write_u8(self, uint8_t val) except -1:
 *         if self.buffer.ob_size - self.len < 1:
 *             array.resize_smart(self.buffer, self.len + 1)             # <<<<<<<<<<<<<<
 *         self.buffer.data.as_uchars[self.len] = val
 *         self.len += 1
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_self->buffer);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = resize_smart(((arrayobject *)__pyx_t_2), (__pyx_v_self->len + 1)); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(16, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/buffer.pxi":32
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:
 *         if self.buffer.ob_size - self.len < 1:             # <<<<<<<<<<<<<<
 *             array.resize_smart(self.buffer, self.len + 1)
 *         self.buffer.data.as_uchars[self.len] = val
 */
  }

  /* "src/buffer.pxi":34
 *         if self.buffer.ob_size - self.len < 1:
 *             array.resize_smart(self.buffer, self.len + 1)
 *         self.buffer.data.as_uchars[self.len] = val             # <<<<<<<<<<<<<<
 *         self.len += 1
 * 
 */
  (__pyx_v_self->buffer->data.as_uchars[__pyx_v_self->len]) = __pyx_v_val;

  /* "src/buffer.pxi":35
 *             array.resize_smart(self.buffer, self.len + 1)
 *         self.buffer.data.as_uchars[self.len] = val
 *         self.len += 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int write_n(self, const uint8_t[:] data) except -1:
 */
  __pyx_v_self->len = (__pyx_v_self->len + 1);

  /* "src/buffer.pxi":31
 *         self.len = 0
 * 
 *     cdef int write_u8(self, uint8_t val) except -1:             # <<<<<<<<<<<<<<
 *         if self.buffer.ob_size - self.len < 1:
 *             array.resize_smart(self.buffer, self.len + 1)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.MemoryWriter.write_u8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/buffer.pxi":37
 *         self.len += 1
 * 
 *     cdef int write_n(self, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef size_t num = data.shape[0]
 *         if num == 0:
 */

static int __pyx_f_5cavro_12MemoryWriter_write_n(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self, __Pyx_memviewslice __pyx_v_data) {
  size_t __pyx_v_num;
  uint8_t const *__pyx_v_data_ptr;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_n", 1);

  /* "src/buffer.pxi":38
 * 
 *     cdef int write_n(self, const uint8_t[:] data) except -1:
 *         cdef size_t num = data.shape[0]             # <<<<<<<<<<<<<<
 *         if num == 0:
 *             return 0
 */
  __pyx_v_num = (__pyx_v_data.shape[0]);

  /* "src/buffer.pxi":39
 *     cdef int write_n(self, const uint8_t[:] data) except -1:
 *         cdef size_t num = data.shape[0]
 *         if num == 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         if self.buffer.ob_size - self.len < num:
 */
  __pyx_t_1 = (__pyx_v_num == 0);
  if (__pyx_t_1) {

    /* "src/buffer.pxi":40
 *         cdef size_t num = data.shape[0]
 *         if num == 0:
 *             return 0             # <<<<<<<<<<<<<<
 *         if self.buffer.ob_size - self.len < num:
 *             array.resize_smart(self.buffer, self.len + num)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "src/buffer.pxi":39
 *     cdef int write_n(self, const uint8_t[:] data) except -1:
 *         cdef size_t num = data.shape[0]
 *         if num == 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         if self.buffer.ob_size - self.len < num:
 */
  }

  /* "src/buffer.pxi":41
 *         if num == 0:
 *             return 0
 *         if self.buffer.ob_size - self.len < num:             # <<<<<<<<<<<<<<
 *             array.resize_smart(self.buffer, self.len + num)
 *         cdef const uint8_t* data_ptr = &data[0]
 */
  __pyx_t_1 = ((__pyx_v_self->buffer->ob_size - __pyx_v_self->len) < __pyx_v_num);
  if (__pyx_t_1) {

    /* "src/buffer.pxi":42
 *             return 0
 *         if self.buffer.ob_size - self.len < num:
 *             array.resize_smart(self.buffer, self.len + num)             # <<<<<<<<<<<<<<
 *         cdef const uint8_t* data_ptr = &data[0]
 *         memcpy(self.buffer.data.as_chars + self.len, data_ptr, num)
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_self->buffer);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = resize_smart(((arrayobject *)__pyx_t_2), (__pyx_v_self->len + __pyx_v_num)); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(16, 42, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/buffer.pxi":41
 *         if num == 0:
 *             return 0
 *         if self.buffer.ob_size - self.len < num:             # <<<<<<<<<<<<<<
 *             array.resize_smart(self.buffer, self.len + num)
 *         cdef const uint8_t* data_ptr = &data[0]
 */
  }

  /* "src/buffer.pxi":43
 *         if self.buffer.ob_size - self.len < num:
 *             array.resize_smart(self.buffer, self.len + num)
 *         cdef const uint8_t* data_ptr = &data[0]             # <<<<<<<<<<<<<<
 *         memcpy(self.buffer.data.as_chars + self.len, data_ptr, num)
 *         self.len += num
 */
  __pyx_t_4 = 0;
  __pyx_t_3 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_v_data.shape[0];
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_3 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_v_data.shape[0])) __pyx_t_3 = 0;
  if (unlikely(__pyx_t_3 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_3);
    __PYX_ERR(16, 43, __pyx_L1_error)
  }
  __pyx_v_data_ptr = (&(*((uint8_t const  *) ( /* dim=0 */ (__pyx_v_data.data + __pyx_t_4 * __pyx_v_data.strides[0]) ))));

  /* "src/buffer.pxi":44
 *             array.resize_smart(self.buffer, self.len + num)
 *         cdef const uint8_t* data_ptr = &data[0]
 *         memcpy(self.buffer.data.as_chars + self.len, data_ptr, num)             # <<<<<<<<<<<<<<
 *         self.len += num
 * 
 */
  (void)(memcpy((__pyx_v_self->buffer->data.as_chars + __pyx_v_self->len), __pyx_v_data_ptr, __pyx_v_num));

  /* "src/buffer.pxi":45
 *         cdef const uint8_t* data_ptr = &data[0]
 *         memcpy(self.buffer.data.as_chars + self.len, data_ptr, num)
 *         self.len += num             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->len = (__pyx_v_self->len + __pyx_v_num);

  /* "src/buffer.pxi":37
 *         self.len += 1
 * 
 *     cdef int write_n(self, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef size_t num = data.shape[0]
 *         if num == 0:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.MemoryWriter.write_n", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/buffer.pxi":15
 *     """
 * 
 *     cdef readonly array.array buffer             # <<<<<<<<<<<<<<
 *     cdef readonly size_t len
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12MemoryWriter_6buffer_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_12MemoryWriter_6buffer_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_12MemoryWriter_6buffer___get__(((struct __pyx_obj_5cavro_MemoryWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12MemoryWriter_6buffer___get__(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->buffer);
  __pyx_r = ((PyObject *)__pyx_v_self->buffer);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/buffer.pxi":16
 * 
 *     cdef readonly array.array buffer
 *     cdef readonly size_t len             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, initial_size=4096):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12MemoryWriter_3len_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_12MemoryWriter_3len_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_12MemoryWriter_3len___get__(((struct __pyx_obj_5cavro_MemoryWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12MemoryWriter_3len___get__(struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->len); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.MemoryWriter.len.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12MemoryWriter_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_12MemoryWriter_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12MemoryWriter_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_12MemoryWriter_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_12MemoryWriter_2__reduce_cython__(((struct __pyx_obj_5cavro_MemoryWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12MemoryWriter_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.MemoryWriter.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12MemoryWriter_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_12MemoryWriter_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12MemoryWriter_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_12MemoryWriter_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.MemoryWriter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_12MemoryWriter_4__setstate_cython__(((struct __pyx_obj_5cavro_MemoryWriter *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12MemoryWriter_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.MemoryWriter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/buffer.pxi":64
 *     cdef const uint8_t *end_ptr
 * 
 *     def __init__(self, data):             # <<<<<<<<<<<<<<
 *         self._reset_to(data)
 * 
 */

/* Python wrapper */
static int __pyx_pw_5cavro_12MemoryReader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_12MemoryReader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 64, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(16, 64, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_data = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(16, 64, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.MemoryReader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_12MemoryReader___init__(((struct __pyx_obj_5cavro_MemoryReader *)__pyx_v_self), __pyx_v_data);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_12MemoryReader___init__(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, PyObject *__pyx_v_data) {
  int __pyx_r;
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/buffer.pxi":65
 * 
 *     def __init__(self, data):
 *         self._reset_to(data)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _reset_to(self, const uint8_t[:] data):
 */
  __pyx_t_1 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_data, 0); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(16, 65, __pyx_L1_error)
  ((struct __pyx_vtabstruct_5cavro_MemoryReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_reset_to(__pyx_v_self, __pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 65, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_1, 1);
  __pyx_t_1.memview = NULL; __pyx_t_1.data = NULL;

  /* "src/buffer.pxi":64
 *     cdef const uint8_t *end_ptr
 * 
 *     def __init__(self, data):             # <<<<<<<<<<<<<<
 *         self._reset_to(data)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_AddTraceback("cavro.MemoryReader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/buffer.pxi":67
 *         self._reset_to(data)
 * 
 *     cdef void _reset_to(self, const uint8_t[:] data):             # <<<<<<<<<<<<<<
 *         if len(data):
 *             self.data = data
 */

static void __pyx_f_5cavro_12MemoryReader__reset_to(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, __Pyx_memviewslice __pyx_v_data) {
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/buffer.pxi":68
 * 
 *     cdef void _reset_to(self, const uint8_t[:] data):
 *         if len(data):             # <<<<<<<<<<<<<<
 *             self.data = data
 *         else:
 */
  __pyx_t_1 = __Pyx_MemoryView_Len(__pyx_v_data); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "src/buffer.pxi":69
 *     cdef void _reset_to(self, const uint8_t[:] data):
 *         if len(data):
 *             self.data = data             # <<<<<<<<<<<<<<
 *         else:
 *             self.data = empty_buffer
 */
    __PYX_XCLEAR_MEMVIEW(&__pyx_v_self->data, 0);
    __PYX_INC_MEMVIEW(&__pyx_v_data, 1);
    __pyx_v_self->data = __pyx_v_data;

    /* "src/buffer.pxi":68
 * 
 *     cdef void _reset_to(self, const uint8_t[:] data):
 *         if len(data):             # <<<<<<<<<<<<<<
 *             self.data = data
 *         else:
 */
    goto __pyx_L3;
  }

  /* "src/buffer.pxi":71
 *             self.data = data
 *         else:
 *             self.data = empty_buffer             # <<<<<<<<<<<<<<
 * 
 *         self.ptr = &self.data[0]
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_5cavro_empty_buffer, 0); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(16, 71, __pyx_L1_error)
    __PYX_XCLEAR_MEMVIEW(&__pyx_v_self->data, 0);
    __pyx_v_self->data = __pyx_t_3;
    __pyx_t_3.memview = NULL;
    __pyx_t_3.data = NULL;
  }
  __pyx_L3:;

  /* "src/buffer.pxi":73
 *             self.data = empty_buffer
 * 
 *         self.ptr = &self.data[0]             # <<<<<<<<<<<<<<
 *         self.end_ptr = self.ptr + len(data)
 * 
 */
  if (unlikely(!__pyx_v_self->data.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(16, 73, __pyx_L1_error)}
  __pyx_t_4 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_v_self->data.shape[0];
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_v_self->data.shape[0])) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(16, 73, __pyx_L1_error)
  }
  __pyx_v_self->ptr = (&(*((uint8_t const  *) ( /* dim=0 */ (__pyx_v_self->data.data + __pyx_t_4 * __pyx_v_self->data.strides[0]) ))));

  /* "src/buffer.pxi":74
 * 
 *         self.ptr = &self.data[0]
 *         self.end_ptr = self.ptr + len(data)             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t *advance(self, size_t num) except <uint8_t*>0:
 */
  __pyx_t_1 = __Pyx_MemoryView_Len(__pyx_v_data); 
  __pyx_v_self->end_ptr = (__pyx_v_self->ptr + __pyx_t_1);

  /* "src/buffer.pxi":67
 *         self._reset_to(data)
 * 
 *     cdef void _reset_to(self, const uint8_t[:] data):             # <<<<<<<<<<<<<<
 *         if len(data):
 *             self.data = data
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_AddTraceback("cavro.MemoryReader._reset_to", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* "src/buffer.pxi":76
 *         self.end_ptr = self.ptr + len(data)
 * 
 *     cdef const uint8_t *advance(self, size_t num) except <uint8_t*>0:             # <<<<<<<<<<<<<<
 *         cdef const uint8_t *ptr = self.ptr
 *         if <size_t>(self.end_ptr - self.ptr) < num:
 */

static uint8_t const *__pyx_f_5cavro_12MemoryReader_advance(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, size_t __pyx_v_num) {
  uint8_t const *__pyx_v_ptr;
  uint8_t const *__pyx_r;
  __Pyx_RefNannyDeclarations
  uint8_t const *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("advance", 1);

  /* "src/buffer.pxi":77
 * 
 *     cdef const uint8_t *advance(self, size_t num) except <uint8_t*>0:
 *         cdef const uint8_t *ptr = self.ptr             # <<<<<<<<<<<<<<
 *         if <size_t>(self.end_ptr - self.ptr) < num:
 *             raise EOFError("Not enough input data to read value")
 */
  __pyx_t_1 = __pyx_v_self->ptr;
  __pyx_v_ptr = __pyx_t_1;

  /* "src/buffer.pxi":78
 *     cdef const uint8_t *advance(self, size_t num) except <uint8_t*>0:
 *         cdef const uint8_t *ptr = self.ptr
 *         if <size_t>(self.end_ptr - self.ptr) < num:             # <<<<<<<<<<<<<<
 *             raise EOFError("Not enough input data to read value")
 *         self.ptr += num
 */
  __pyx_t_2 = (((size_t)(__pyx_v_self->end_ptr - __pyx_v_self->ptr)) < __pyx_v_num);
  if (unlikely(__pyx_t_2)) {

    /* "src/buffer.pxi":79
 *         cdef const uint8_t *ptr = self.ptr
 *         if <size_t>(self.end_ptr - self.ptr) < num:
 *             raise EOFError("Not enough input data to read value")             # <<<<<<<<<<<<<<
 *         self.ptr += num
 *         return ptr
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_EOFError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(16, 79, __pyx_L1_error)

    /* "src/buffer.pxi":78
 *     cdef const uint8_t *advance(self, size_t num) except <uint8_t*>0:
 *         cdef const uint8_t *ptr = self.ptr
 *         if <size_t>(self.end_ptr - self.ptr) < num:             # <<<<<<<<<<<<<<
 *             raise EOFError("Not enough input data to read value")
 *         self.ptr += num
 */
  }

  /* "src/buffer.pxi":80
 *         if <size_t>(self.end_ptr - self.ptr) < num:
 *             raise EOFError("Not enough input data to read value")
 *         self.ptr += num             # <<<<<<<<<<<<<<
 *         return ptr
 * 
 */
  __pyx_v_self->ptr = (__pyx_v_self->ptr + __pyx_v_num);

  /* "src/buffer.pxi":81
 *             raise EOFError("Not enough input data to read value")
 *         self.ptr += num
 *         return ptr             # <<<<<<<<<<<<<<
 * 
 *     cdef inline int ensure(self, size_t num) except -1:
 */
  __pyx_r = __pyx_v_ptr;
  goto __pyx_L0;

  /* "src/buffer.pxi":76
 *         self.end_ptr = self.ptr + len(data)
 * 
 *     cdef const uint8_t *advance(self, size_t num) except <uint8_t*>0:             # <<<<<<<<<<<<<<
 *         cdef const uint8_t *ptr = self.ptr
 *         if <size_t>(self.end_ptr - self.ptr) < num:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.MemoryReader.advance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = ((uint8_t *)0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/buffer.pxi":83
 *         return ptr
 * 
 *     cdef inline int ensure(self, size_t num) except -1:             # <<<<<<<<<<<<<<
 *         if <size_t>(self.end_ptr - self.ptr) < num:
 *             raise EOFError("Not enough input data to read value")
 */

static CYTHON_INLINE int __pyx_f_5cavro_12MemoryReader_ensure(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, size_t __pyx_v_num) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ensure", 1);

  /* "src/buffer.pxi":84
 * 
 *     cdef inline int ensure(self, size_t num) except -1:
 *         if <size_t>(self.end_ptr - self.ptr) < num:             # <<<<<<<<<<<<<<
 *             raise EOFError("Not enough input data to read value")
 * 
 */
  __pyx_t_1 = (((size_t)(__pyx_v_self->end_ptr - __pyx_v_self->ptr)) < __pyx_v_num);
  if (unlikely(__pyx_t_1)) {

    /* "src/buffer.pxi":85
 *     cdef inline int ensure(self, size_t num) except -1:
 *         if <size_t>(self.end_ptr - self.ptr) < num:
 *             raise EOFError("Not enough input data to read value")             # <<<<<<<<<<<<<<
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_EOFError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(16, 85, __pyx_L1_error)

    /* "src/buffer.pxi":84
 * 
 *     cdef inline int ensure(self, size_t num) except -1:
 *         if <size_t>(self.end_ptr - self.ptr) < num:             # <<<<<<<<<<<<<<
 *             raise EOFError("Not enough input data to read value")
 * 
 */
  }

  /* "src/buffer.pxi":83
 *         return ptr
 * 
 *     cdef inline int ensure(self, size_t num) except -1:             # <<<<<<<<<<<<<<
 *         if <size_t>(self.end_ptr - self.ptr) < num:
 *             raise EOFError("Not enough input data to read value")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.MemoryReader.ensure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/buffer.pxi":87
 *             raise EOFError("Not enough input data to read value")
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:             # <<<<<<<<<<<<<<
 *         return self.advance(1)[0]
 * 
 */

static uint8_t __pyx_f_5cavro_12MemoryReader_read_u8(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self) {
  uint8_t __pyx_r;
  uint8_t const *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/buffer.pxi":88
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:
 *         return self.advance(1)[0]             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_MemoryReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->advance(__pyx_v_self, 1); if (unlikely(__pyx_t_1 == ((uint8_t const *)((uint8_t *)0)))) __PYX_ERR(16, 88, __pyx_L1_error)
  __pyx_r = (__pyx_t_1[0]);
  goto __pyx_L0;

  /* "src/buffer.pxi":87
 *             raise EOFError("Not enough input data to read value")
 * 
 *     cdef uint8_t read_u8(self) except? 0xba:             # <<<<<<<<<<<<<<
 *         return self.advance(1)[0]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.MemoryReader.read_u8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0xba;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/buffer.pxi":90
 *         return self.advance(1)[0]
 * 
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         cdef const uint8_t* ptr = self.ptr
 *         self.advance(n)
 */

static __Pyx_memviewslice __pyx_f_5cavro_12MemoryReader_read_n(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, Py_ssize_t __pyx_v_n) {
  uint8_t const *__pyx_v_ptr;
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  uint8_t const *__pyx_t_1;
  uint8_t const *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_n", 1);

  /* "src/buffer.pxi":91
 * 
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):
 *         cdef const uint8_t* ptr = self.ptr             # <<<<<<<<<<<<<<
 *         self.advance(n)
 *         return ptr[:n]
 */
  __pyx_t_1 = __pyx_v_self->ptr;
  __pyx_v_ptr = __pyx_t_1;

  /* "src/buffer.pxi":92
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):
 *         cdef const uint8_t* ptr = self.ptr
 *         self.advance(n)             # <<<<<<<<<<<<<<
 *         return ptr[:n]
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_MemoryReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->advance(__pyx_v_self, __pyx_v_n); if (unlikely(__pyx_t_2 == ((uint8_t const *)((uint8_t *)0)))) __PYX_ERR(16, 92, __pyx_L1_error)

  /* "src/buffer.pxi":93
 *         cdef const uint8_t* ptr = self.ptr
 *         self.advance(n)
 *         return ptr[:n]             # <<<<<<<<<<<<<<
 * 
 *     cdef bytes read_bytes(self, Py_ssize_t n):
 */
  __pyx_t_3 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_ptr) + 0, __pyx_v_n - 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_3, 0); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(16, 93, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;
  goto __pyx_L0;

  /* "src/buffer.pxi":90
 *         return self.advance(1)[0]
 * 
 *     cdef const uint8_t[:] read_n(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         cdef const uint8_t* ptr = self.ptr
 *         self.advance(n)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_4, 1);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("cavro.MemoryReader.read_n", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/buffer.pxi":95
 *         return ptr[:n]
 * 
 *     cdef bytes read_bytes(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         cdef const uint8_t* ptr = self.ptr
 *         self.advance(n)
 */

static PyObject *__pyx_f_5cavro_12MemoryReader_read_bytes(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, Py_ssize_t __pyx_v_n) {
  uint8_t const *__pyx_v_ptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint8_t const *__pyx_t_1;
  uint8_t const *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_bytes", 1);

  /* "src/buffer.pxi":96
 * 
 *     cdef bytes read_bytes(self, Py_ssize_t n):
 *         cdef const uint8_t* ptr = self.ptr             # <<<<<<<<<<<<<<
 *         self.advance(n)
 *         return ptr[:n]
 */
  __pyx_t_1 = __pyx_v_self->ptr;
  __pyx_v_ptr = __pyx_t_1;

  /* "src/buffer.pxi":97
 *     cdef bytes read_bytes(self, Py_ssize_t n):
 *         cdef const uint8_t* ptr = self.ptr
 *         self.advance(n)             # <<<<<<<<<<<<<<
 *         return ptr[:n]
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_MemoryReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->advance(__pyx_v_self, __pyx_v_n); if (unlikely(__pyx_t_2 == ((uint8_t const *)((uint8_t *)0)))) __PYX_ERR(16, 97, __pyx_L1_error)

  /* "src/buffer.pxi":98
 *         cdef const uint8_t* ptr = self.ptr
 *         self.advance(n)
 *         return ptr[:n]             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_ptr) + 0, __pyx_v_n - 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/buffer.pxi":95
 *         return ptr[:n]
 * 
 *     cdef bytes read_bytes(self, Py_ssize_t n):             # <<<<<<<<<<<<<<
 *         cdef const uint8_t* ptr = self.ptr
 *         self.advance(n)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.MemoryReader.read_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12MemoryReader_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_12MemoryReader_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12MemoryReader_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_12MemoryReader_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_12MemoryReader_2__reduce_cython__(((struct __pyx_obj_5cavro_MemoryReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12MemoryReader_2__reduce_cython__(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.data, self.end_ptr, self.ptr)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->data.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBytes_FromCString(__pyx_v_self->end_ptr); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBytes_FromCString(__pyx_v_self->ptr); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.data, self.end_ptr, self.ptr)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.data, self.end_ptr, self.ptr)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.data, self.end_ptr, self.ptr)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_MemoryReader, (type(self), 0xec0f6ac, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MemoryReader, (type(self), 0xec0f6ac, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_MemoryReader, (type(self), 0xec0f6ac, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_MemoryReader, (type(self), 0xec0f6ac, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_MemoryReader); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_247527084);
    __Pyx_GIVEREF(__pyx_int_247527084);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_247527084)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MemoryReader, (type(self), 0xec0f6ac, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_MemoryReader, (type(self), 0xec0f6ac, None), state
 *     else:
 *         return __pyx_unpickle_MemoryReader, (type(self), 0xec0f6ac, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MemoryReader__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_MemoryReader); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_247527084);
    __Pyx_GIVEREF(__pyx_int_247527084);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_247527084)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.MemoryReader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MemoryReader, (type(self), 0xec0f6ac, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MemoryReader__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12MemoryReader_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_12MemoryReader_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12MemoryReader_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_12MemoryReader_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.MemoryReader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_12MemoryReader_4__setstate_cython__(((struct __pyx_obj_5cavro_MemoryReader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12MemoryReader_4__setstate_cython__(struct __pyx_obj_5cavro_MemoryReader *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_MemoryReader, (type(self), 0xec0f6ac, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MemoryReader__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_MemoryReader__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MemoryReader, (type(self), 0xec0f6ac, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MemoryReader__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.MemoryReader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/zigzag.pxi":3
 * from libc.stdint cimport *
 * 
 * cdef uint32_t read_varint(_Reader buf) except? 0xfffffbad:             # <<<<<<<<<<<<<<
 *     cdef uint32_t cur
 *     cdef uint32_t shift = 0
 */

static uint32_t __pyx_f_5cavro_read_varint(struct __pyx_obj_5cavro__Reader *__pyx_v_buf) {
  uint32_t __pyx_v_cur;
  uint32_t __pyx_v_shift;
  uint32_t __pyx_v_value;
  uint32_t __pyx_r;
  uint8_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/zigzag.pxi":5
 * cdef uint32_t read_varint(_Reader buf) except? 0xfffffbad:
 *     cdef uint32_t cur
 *     cdef uint32_t shift = 0             # <<<<<<<<<<<<<<
 *     cdef uint32_t value = 0
 *     while True:
 */
  __pyx_v_shift = 0;

  /* "src/zigzag.pxi":6
 *     cdef uint32_t cur
 *     cdef uint32_t shift = 0
 *     cdef uint32_t value = 0             # <<<<<<<<<<<<<<
 *     while True:
 *         cur = buf.read_u8()
 */
  __pyx_v_value = 0;

  /* "src/zigzag.pxi":7
 *     cdef uint32_t shift = 0
 *     cdef uint32_t value = 0
 *     while True:             # <<<<<<<<<<<<<<
 *         cur = buf.read_u8()
 *         value |= (cur & 0b01111111) << shift
 */
  while (1) {

    /* "src/zigzag.pxi":8
 *     cdef uint32_t value = 0
 *     while True:
 *         cur = buf.read_u8()             # <<<<<<<<<<<<<<
 *         value |= (cur & 0b01111111) << shift
 *         shift += 7
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_buf->__pyx_vtab)->read_u8(__pyx_v_buf); if (unlikely(__pyx_t_1 == ((uint8_t)0xba) && PyErr_Occurred())) __PYX_ERR(17, 8, __pyx_L1_error)
    __pyx_v_cur = __pyx_t_1;

    /* "src/zigzag.pxi":9
 *     while True:
 *         cur = buf.read_u8()
 *         value |= (cur & 0b01111111) << shift             # <<<<<<<<<<<<<<
 *         shift += 7
 *         if not cur & 0b10000000:
 */
    __pyx_v_value = (__pyx_v_value | ((__pyx_v_cur & 127) << __pyx_v_shift));

    /* "src/zigzag.pxi":10
 *         cur = buf.read_u8()
 *         value |= (cur & 0b01111111) << shift
 *         shift += 7             # <<<<<<<<<<<<<<
 *         if not cur & 0b10000000:
 *             return value
 */
    __pyx_v_shift = (__pyx_v_shift + 7);

    /* "src/zigzag.pxi":11
 *         value |= (cur & 0b01111111) << shift
 *         shift += 7
 *         if not cur & 0b10000000:             # <<<<<<<<<<<<<<
 *             return value
 * 
 */
    __pyx_t_2 = (!((__pyx_v_cur & 128) != 0));
    if (__pyx_t_2) {

      /* "src/zigzag.pxi":12
 *         shift += 7
 *         if not cur & 0b10000000:
 *             return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_r = __pyx_v_value;
      goto __pyx_L0;

      /* "src/zigzag.pxi":11
 *         value |= (cur & 0b01111111) << shift
 *         shift += 7
 *         if not cur & 0b10000000:             # <<<<<<<<<<<<<<
 *             return value
 * 
 */
    }
  }

  /* "src/zigzag.pxi":3
 * from libc.stdint cimport *
 * 
 * cdef uint32_t read_varint(_Reader buf) except? 0xfffffbad:             # <<<<<<<<<<<<<<
 *     cdef uint32_t cur
 *     cdef uint32_t shift = 0
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.read_varint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0xfffffbad;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/zigzag.pxi":15
 * 
 * 
 * cdef uint64_t read_varlong(_Reader buf) except? 0xfffffffffffffbadull:             # <<<<<<<<<<<<<<
 *     cdef uint64_t cur
 *     cdef uint64_t shift = 0
 */

static uint64_t __pyx_f_5cavro_read_varlong(struct __pyx_obj_5cavro__Reader *__pyx_v_buf) {
  uint64_t __pyx_v_cur;
  uint64_t __pyx_v_shift;
  uint64_t __pyx_v_value;
  uint64_t __pyx_r;
  uint8_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/zigzag.pxi":17
 * cdef uint64_t read_varlong(_Reader buf) except? 0xfffffffffffffbadull:
 *     cdef uint64_t cur
 *     cdef uint64_t shift = 0             # <<<<<<<<<<<<<<
 *     cdef uint64_t value = 0
 *     while True:
 */
  __pyx_v_shift = 0;

  /* "src/zigzag.pxi":18
 *     cdef uint64_t cur
 *     cdef uint64_t shift = 0
 *     cdef uint64_t value = 0             # <<<<<<<<<<<<<<
 *     while True:
 *         cur = buf.read_u8()
 */
  __pyx_v_value = 0;

  /* "src/zigzag.pxi":19
 *     cdef uint64_t shift = 0
 *     cdef uint64_t value = 0
 *     while True:             # <<<<<<<<<<<<<<
 *         cur = buf.read_u8()
 *         value |= (cur & 0b01111111) << shift
 */
  while (1) {

    /* "src/zigzag.pxi":20
 *     cdef uint64_t value = 0
 *     while True:
 *         cur = buf.read_u8()             # <<<<<<<<<<<<<<
 *         value |= (cur & 0b01111111) << shift
 *         shift += 7
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_buf->__pyx_vtab)->read_u8(__pyx_v_buf); if (unlikely(__pyx_t_1 == ((uint8_t)0xba) && PyErr_Occurred())) __PYX_ERR(17, 20, __pyx_L1_error)
    __pyx_v_cur = __pyx_t_1;

    /* "src/zigzag.pxi":21
 *     while True:
 *         cur = buf.read_u8()
 *         value |= (cur & 0b01111111) << shift             # <<<<<<<<<<<<<<
 *         shift += 7
 *         if not cur & 0b10000000:
 */
    __pyx_v_value = (__pyx_v_value | ((__pyx_v_cur & 127) << __pyx_v_shift));

    /* "src/zigzag.pxi":22
 *         cur = buf.read_u8()
 *         value |= (cur & 0b01111111) << shift
 *         shift += 7             # <<<<<<<<<<<<<<
 *         if not cur & 0b10000000:
 *             return value
 */
    __pyx_v_shift = (__pyx_v_shift + 7);

    /* "src/zigzag.pxi":23
 *         value |= (cur & 0b01111111) << shift
 *         shift += 7
 *         if not cur & 0b10000000:             # <<<<<<<<<<<<<<
 *             return value
 * 
 */
    __pyx_t_2 = (!((__pyx_v_cur & 128) != 0));
    if (__pyx_t_2) {

      /* "src/zigzag.pxi":24
 *         shift += 7
 *         if not cur & 0b10000000:
 *             return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_r = __pyx_v_value;
      goto __pyx_L0;

      /* "src/zigzag.pxi":23
 *         value |= (cur & 0b01111111) << shift
 *         shift += 7
 *         if not cur & 0b10000000:             # <<<<<<<<<<<<<<
 *             return value
 * 
 */
    }
  }

  /* "src/zigzag.pxi":15
 * 
 * 
 * cdef uint64_t read_varlong(_Reader buf) except? 0xfffffffffffffbadull:             # <<<<<<<<<<<<<<
 *     cdef uint64_t cur
 *     cdef uint64_t shift = 0
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.read_varlong", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0xfffffffffffffbadULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/zigzag.pxi":27
 * 
 * 
 * cdef int32_t zigzag_decode_int(_Reader buf) except? 0x7ffffbadu:             # <<<<<<<<<<<<<<
 *     cdef uint32_t value = read_varint(buf)
 *     return (value >> 1) ^ (-(value & 1))
 */

static int32_t __pyx_f_5cavro_zigzag_decode_int(struct __pyx_obj_5cavro__Reader *__pyx_v_buf) {
  uint32_t __pyx_v_value;
  int32_t __pyx_r;
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/zigzag.pxi":28
 * 
 * cdef int32_t zigzag_decode_int(_Reader buf) except? 0x7ffffbadu:
 *     cdef uint32_t value = read_varint(buf)             # <<<<<<<<<<<<<<
 *     return (value >> 1) ^ (-(value & 1))
 * 
 */
  __pyx_t_1 = __pyx_f_5cavro_read_varint(__pyx_v_buf); if (unlikely(__pyx_t_1 == ((uint32_t)0xfffffbad) && PyErr_Occurred())) __PYX_ERR(17, 28, __pyx_L1_error)
  __pyx_v_value = __pyx_t_1;

  /* "src/zigzag.pxi":29
 * cdef int32_t zigzag_decode_int(_Reader buf) except? 0x7ffffbadu:
 *     cdef uint32_t value = read_varint(buf)
 *     return (value >> 1) ^ (-(value & 1))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((__pyx_v_value >> 1) ^ (-(__pyx_v_value & 1)));
  goto __pyx_L0;

  /* "src/zigzag.pxi":27
 * 
 * 
 * cdef int32_t zigzag_decode_int(_Reader buf) except? 0x7ffffbadu:             # <<<<<<<<<<<<<<
 *     cdef uint32_t value = read_varint(buf)
 *     return (value >> 1) ^ (-(value & 1))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.zigzag_decode_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0x7ffffbadU;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/zigzag.pxi":32
 * 
 * 
 * cdef int64_t zigzag_decode_long(_Reader buf) except? 0x7ffffffffffffbadull:             # <<<<<<<<<<<<<<
 *     cdef uint64_t value = read_varlong(buf)
 *     return (value >> 1) ^ (-(value & 1ull))
 */

static int64_t __pyx_f_5cavro_zigzag_decode_long(struct __pyx_obj_5cavro__Reader *__pyx_v_buf) {
  uint64_t __pyx_v_value;
  int64_t __pyx_r;
  uint64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/zigzag.pxi":33
 * 
 * cdef int64_t zigzag_decode_long(_Reader buf) except? 0x7ffffffffffffbadull:
 *     cdef uint64_t value = read_varlong(buf)             # <<<<<<<<<<<<<<
 *     return (value >> 1) ^ (-(value & 1ull))
 * 
 */
  __pyx_t_1 = __pyx_f_5cavro_read_varlong(__pyx_v_buf); if (unlikely(__pyx_t_1 == ((uint64_t)0xfffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(17, 33, __pyx_L1_error)
  __pyx_v_value = __pyx_t_1;

  /* "src/zigzag.pxi":34
 * cdef int64_t zigzag_decode_long(_Reader buf) except? 0x7ffffffffffffbadull:
 *     cdef uint64_t value = read_varlong(buf)
 *     return (value >> 1) ^ (-(value & 1ull))             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = ((__pyx_v_value >> 1) ^ (-(__pyx_v_value & 1ULL)));
  goto __pyx_L0;

  /* "src/zigzag.pxi":32
 * 
 * 
 * cdef int64_t zigzag_decode_long(_Reader buf) except? 0x7ffffffffffffbadull:             # <<<<<<<<<<<<<<
 *     cdef uint64_t value = read_varlong(buf)
 *     return (value >> 1) ^ (-(value & 1ull))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.zigzag_decode_long", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0x7ffffffffffffbadULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/zigzag.pxi":37
 * 
 * @cython.cdivision(True)
 * cdef int zigzag_encode_int(_Writer buf, int32_t value) except -1:             # <<<<<<<<<<<<<<
 *     cdef uint32_t zz = (value << 1) ^ (value >> 31)
 *     cdef uint8_t cur
 */

static int __pyx_f_5cavro_zigzag_encode_int(struct __pyx_obj_5cavro__Writer *__pyx_v_buf, int32_t __pyx_v_value) {
  uint32_t __pyx_v_zz;
  uint8_t __pyx_v_cur;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/zigzag.pxi":38
 * @cython.cdivision(True)
 * cdef int zigzag_encode_int(_Writer buf, int32_t value) except -1:
 *     cdef uint32_t zz = (value << 1) ^ (value >> 31)             # <<<<<<<<<<<<<<
 *     cdef uint8_t cur
 *     if zz == 0:
 */
  __pyx_v_zz = ((__pyx_v_value << 1) ^ (__pyx_v_value >> 31));

  /* "src/zigzag.pxi":40
 *     cdef uint32_t zz = (value << 1) ^ (value >> 31)
 *     cdef uint8_t cur
 *     if zz == 0:             # <<<<<<<<<<<<<<
 *         buf.write_u8(0)
 *     while zz:
 */
  __pyx_t_1 = (__pyx_v_zz == 0);
  if (__pyx_t_1) {

    /* "src/zigzag.pxi":41
 *     cdef uint8_t cur
 *     if zz == 0:
 *         buf.write_u8(0)             # <<<<<<<<<<<<<<
 *     while zz:
 *         cur = zz & 0x7f
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_buf->__pyx_vtab)->write_u8(__pyx_v_buf, 0); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(17, 41, __pyx_L1_error)

    /* "src/zigzag.pxi":40
 *     cdef uint32_t zz = (value << 1) ^ (value >> 31)
 *     cdef uint8_t cur
 *     if zz == 0:             # <<<<<<<<<<<<<<
 *         buf.write_u8(0)
 *     while zz:
 */
  }

  /* "src/zigzag.pxi":42
 *     if zz == 0:
 *         buf.write_u8(0)
 *     while zz:             # <<<<<<<<<<<<<<
 *         cur = zz & 0x7f
 *         zz >>= 7
 */
  while (1) {
    __pyx_t_1 = (__pyx_v_zz != 0);
    if (!__pyx_t_1) break;

    /* "src/zigzag.pxi":43
 *         buf.write_u8(0)
 *     while zz:
 *         cur = zz & 0x7f             # <<<<<<<<<<<<<<
 *         zz >>= 7
 *         if zz:
 */
    __pyx_v_cur = (__pyx_v_zz & 0x7f);

    /* "src/zigzag.pxi":44
 *     while zz:
 *         cur = zz & 0x7f
 *         zz >>= 7             # <<<<<<<<<<<<<<
 *         if zz:
 *             cur |= 0x80
 */
    __pyx_v_zz = (__pyx_v_zz >> 7);

    /* "src/zigzag.pxi":45
 *         cur = zz & 0x7f
 *         zz >>= 7
 *         if zz:             # <<<<<<<<<<<<<<
 *             cur |= 0x80
 *         buf.write_u8(cur)
 */
    __pyx_t_1 = (__pyx_v_zz != 0);
    if (__pyx_t_1) {

      /* "src/zigzag.pxi":46
 *         zz >>= 7
 *         if zz:
 *             cur |= 0x80             # <<<<<<<<<<<<<<
 *         buf.write_u8(cur)
 * 
 */
      __pyx_v_cur = (__pyx_v_cur | 0x80);

      /* "src/zigzag.pxi":45
 *         cur = zz & 0x7f
 *         zz >>= 7
 *         if zz:             # <<<<<<<<<<<<<<
 *             cur |= 0x80
 *         buf.write_u8(cur)
 */
    }

    /* "src/zigzag.pxi":47
 *         if zz:
 *             cur |= 0x80
 *         buf.write_u8(cur)             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_buf->__pyx_vtab)->write_u8(__pyx_v_buf, __pyx_v_cur); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(17, 47, __pyx_L1_error)
  }

  /* "src/zigzag.pxi":37
 * 
 * @cython.cdivision(True)
 * cdef int zigzag_encode_int(_Writer buf, int32_t value) except -1:             # <<<<<<<<<<<<<<
 *     cdef uint32_t zz = (value << 1) ^ (value >> 31)
 *     cdef uint8_t cur
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.zigzag_encode_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/zigzag.pxi":50
 * 
 * @cython.cdivision(True)
 * cdef int zigzag_encode_long(_Writer buf, int64_t value) except -1:             # <<<<<<<<<<<<<<
 *     cdef uint64_t zz = (value << 1) ^ (value >> 63)
 *     cdef uint8_t cur
 */

static int __pyx_f_5cavro_zigzag_encode_long(struct __pyx_obj_5cavro__Writer *__pyx_v_buf, int64_t __pyx_v_value) {
  uint64_t __pyx_v_zz;
  uint8_t __pyx_v_cur;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/zigzag.pxi":51
 * @cython.cdivision(True)
 * cdef int zigzag_encode_long(_Writer buf, int64_t value) except -1:
 *     cdef uint64_t zz = (value << 1) ^ (value >> 63)             # <<<<<<<<<<<<<<
 *     cdef uint8_t cur
 *     if zz == 0:
 */
  __pyx_v_zz = ((__pyx_v_value << 1) ^ (__pyx_v_value >> 63));

  /* "src/zigzag.pxi":53
 *     cdef uint64_t zz = (value << 1) ^ (value >> 63)
 *     cdef uint8_t cur
 *     if zz == 0:             # <<<<<<<<<<<<<<
 *         buf.write_u8(0)
 *     while zz:
 */
  __pyx_t_1 = (__pyx_v_zz == 0);
  if (__pyx_t_1) {

    /* "src/zigzag.pxi":54
 *     cdef uint8_t cur
 *     if zz == 0:
 *         buf.write_u8(0)             # <<<<<<<<<<<<<<
 *     while zz:
 *         cur = zz & 0b01111111
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_buf->__pyx_vtab)->write_u8(__pyx_v_buf, 0); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(17, 54, __pyx_L1_error)

    /* "src/zigzag.pxi":53
 *     cdef uint64_t zz = (value << 1) ^ (value >> 63)
 *     cdef uint8_t cur
 *     if zz == 0:             # <<<<<<<<<<<<<<
 *         buf.write_u8(0)
 *     while zz:
 */
  }

  /* "src/zigzag.pxi":55
 *     if zz == 0:
 *         buf.write_u8(0)
 *     while zz:             # <<<<<<<<<<<<<<
 *         cur = zz & 0b01111111
 *         zz >>= 7
 */
  while (1) {
    __pyx_t_1 = (__pyx_v_zz != 0);
    if (!__pyx_t_1) break;

    /* "src/zigzag.pxi":56
 *         buf.write_u8(0)
 *     while zz:
 *         cur = zz & 0b01111111             # <<<<<<<<<<<<<<
 *         zz >>= 7
 *         if zz:
 */
    __pyx_v_cur = (__pyx_v_zz & 127);

    /* "src/zigzag.pxi":57
 *     while zz:
 *         cur = zz & 0b01111111
 *         zz >>= 7             # <<<<<<<<<<<<<<
 *         if zz:
 *             cur |= 0b10000000
 */
    __pyx_v_zz = (__pyx_v_zz >> 7);

    /* "src/zigzag.pxi":58
 *         cur = zz & 0b01111111
 *         zz >>= 7
 *         if zz:             # <<<<<<<<<<<<<<
 *             cur |= 0b10000000
 *         buf.write_u8(cur)
 */
    __pyx_t_1 = (__pyx_v_zz != 0);
    if (__pyx_t_1) {

      /* "src/zigzag.pxi":59
 *         zz >>= 7
 *         if zz:
 *             cur |= 0b10000000             # <<<<<<<<<<<<<<
 *         buf.write_u8(cur)
 */
      __pyx_v_cur = (__pyx_v_cur | 128);

      /* "src/zigzag.pxi":58
 *         cur = zz & 0b01111111
 *         zz >>= 7
 *         if zz:             # <<<<<<<<<<<<<<
 *             cur |= 0b10000000
 *         buf.write_u8(cur)
 */
    }

    /* "src/zigzag.pxi":60
 *         if zz:
 *             cur |= 0b10000000
 *         buf.write_u8(cur)             # <<<<<<<<<<<<<<
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_buf->__pyx_vtab)->write_u8(__pyx_v_buf, __pyx_v_cur); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(17, 60, __pyx_L1_error)
  }

  /* "src/zigzag.pxi":50
 * 
 * @cython.cdivision(True)
 * cdef int zigzag_encode_long(_Writer buf, int64_t value) except -1:             # <<<<<<<<<<<<<<
 *     cdef uint64_t zz = (value << 1) ^ (value >> 63)
 *     cdef uint8_t cur
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.zigzag_encode_long", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/rabin.pxi":10
 * 
 * 
 * cdef init_rabin_table():             # <<<<<<<<<<<<<<
 *     global RABIN_TABLE_CONFIGURED
 *     cdef size_t i, j
 */

static PyObject *__pyx_f_5cavro_init_rabin_table(void) {
  size_t __pyx_v_i;
  CYTHON_UNUSED size_t __pyx_v_j;
  int64_t __pyx_v_mask;
  int64_t __pyx_v_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  __Pyx_RefNannySetupContext("init_rabin_table", 1);

  /* "src/rabin.pxi":15
 *     cdef int64_t mask;
 *     cdef int64_t value
 *     for i in range(256):             # <<<<<<<<<<<<<<
 *         value = i
 *         for j in range(8):
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 0x100; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "src/rabin.pxi":16
 *     cdef int64_t value
 *     for i in range(256):
 *         value = i             # <<<<<<<<<<<<<<
 *         for j in range(8):
 *             mask = -(value & 0x1);
 */
    __pyx_v_value = __pyx_v_i;

    /* "src/rabin.pxi":17
 *     for i in range(256):
 *         value = i
 *         for j in range(8):             # <<<<<<<<<<<<<<
 *             mask = -(value & 0x1);
 *             value = <int64_t>((<uint64_t>value >> 1) ^ (RABIN_EMPTY & mask));
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 8; __pyx_t_2+=1) {
      __pyx_v_j = __pyx_t_2;

      /* "src/rabin.pxi":18
 *         value = i
 *         for j in range(8):
 *             mask = -(value & 0x1);             # <<<<<<<<<<<<<<
 *             value = <int64_t>((<uint64_t>value >> 1) ^ (RABIN_EMPTY & mask));
 *         RABIN_TABLE[i] = value
 */
      __pyx_v_mask = (-(__pyx_v_value & 0x1));

      /* "src/rabin.pxi":19
 *         for j in range(8):
 *             mask = -(value & 0x1);
 *             value = <int64_t>((<uint64_t>value >> 1) ^ (RABIN_EMPTY & mask));             # <<<<<<<<<<<<<<
 *         RABIN_TABLE[i] = value
 *     RABIN_TABLE_CONFIGURED = 1
 */
      __pyx_v_value = ((int64_t)((((uint64_t)__pyx_v_value) >> 1) ^ (__pyx_v_5cavro_RABIN_EMPTY & __pyx_v_mask)));
    }

    /* "src/rabin.pxi":20
 *             mask = -(value & 0x1);
 *             value = <int64_t>((<uint64_t>value >> 1) ^ (RABIN_EMPTY & mask));
 *         RABIN_TABLE[i] = value             # <<<<<<<<<<<<<<
 *     RABIN_TABLE_CONFIGURED = 1
 * 
 */
    (__pyx_v_5cavro_RABIN_TABLE[__pyx_v_i]) = __pyx_v_value;
  }

  /* "src/rabin.pxi":21
 *             value = <int64_t>((<uint64_t>value >> 1) ^ (RABIN_EMPTY & mask));
 *         RABIN_TABLE[i] = value
 *     RABIN_TABLE_CONFIGURED = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_5cavro_RABIN_TABLE_CONFIGURED = 1;

  /* "src/rabin.pxi":10
 * 
 * 
 * cdef init_rabin_table():             # <<<<<<<<<<<<<<
 *     global RABIN_TABLE_CONFIGURED
 *     cdef size_t i, j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/rabin.pxi":39
 *     cdef readonly int64_t value
 * 
 *     def __init__(self, value=RABIN_EMPTY):             # <<<<<<<<<<<<<<
 *         if not RABIN_TABLE_CONFIGURED:
 *             init_rabin_table()
 */

/* Python wrapper */
static int __pyx_pw_5cavro_5Rabin_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_5Rabin_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_value = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(__pyx_k__21);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 39, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(18, 39, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, __pyx_nargs); __PYX_ERR(18, 39, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Rabin.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_5Rabin___init__(((struct __pyx_obj_5cavro_Rabin *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_5Rabin___init__(struct __pyx_obj_5cavro_Rabin *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int64_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/rabin.pxi":40
 * 
 *     def __init__(self, value=RABIN_EMPTY):
 *         if not RABIN_TABLE_CONFIGURED:             # <<<<<<<<<<<<<<
 *             init_rabin_table()
 *         self.value = value
 */
  __pyx_t_1 = (!__pyx_v_5cavro_RABIN_TABLE_CONFIGURED);
  if (__pyx_t_1) {

    /* "src/rabin.pxi":41
 *     def __init__(self, value=RABIN_EMPTY):
 *         if not RABIN_TABLE_CONFIGURED:
 *             init_rabin_table()             # <<<<<<<<<<<<<<
 *         self.value = value
 * 
 */
    __pyx_t_2 = __pyx_f_5cavro_init_rabin_table(); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/rabin.pxi":40
 * 
 *     def __init__(self, value=RABIN_EMPTY):
 *         if not RABIN_TABLE_CONFIGURED:             # <<<<<<<<<<<<<<
 *             init_rabin_table()
 *         self.value = value
 */
  }

  /* "src/rabin.pxi":42
 *         if not RABIN_TABLE_CONFIGURED:
 *             init_rabin_table()
 *         self.value = value             # <<<<<<<<<<<<<<
 * 
 *     cpdef update(self, bytes data):
 */
  __pyx_t_3 = __Pyx_PyInt_As_int64_t(__pyx_v_value); if (unlikely((__pyx_t_3 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(18, 42, __pyx_L1_error)
  __pyx_v_self->value = __pyx_t_3;

  /* "src/rabin.pxi":39
 *     cdef readonly int64_t value
 * 
 *     def __init__(self, value=RABIN_EMPTY):             # <<<<<<<<<<<<<<
 *         if not RABIN_TABLE_CONFIGURED:
 *             init_rabin_table()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.Rabin.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/rabin.pxi":44
 *         self.value = value
 * 
 *     cpdef update(self, bytes data):             # <<<<<<<<<<<<<<
 *         cdef int64_t swizzle
 *         cdef int64_t shifted
 */

static PyObject *__pyx_pw_5cavro_5Rabin_3update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_5Rabin_update(struct __pyx_obj_5cavro_Rabin *__pyx_v_self, PyObject *__pyx_v_data, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int64_t __pyx_v_swizzle;
  int64_t __pyx_v_shifted;
  PyObject *__pyx_v_char = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *(*__pyx_t_2)(PyObject *);
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 1);

  /* "src/rabin.pxi":47
 *         cdef int64_t swizzle
 *         cdef int64_t shifted
 *         for char in data:             # <<<<<<<<<<<<<<
 *             swizzle = RABIN_TABLE[(self.value ^ char) & 0xff]
 *             shifted = <int64_t>(<uint64_t>self.value >> 8)
 */
  __pyx_t_1 = PyObject_GetIter(__pyx_v_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 47, __pyx_L1_error)
  for (;;) {
    {
      __pyx_t_3 = __pyx_t_2(__pyx_t_1);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(18, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_char, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/rabin.pxi":48
 *         cdef int64_t shifted
 *         for char in data:
 *             swizzle = RABIN_TABLE[(self.value ^ char) & 0xff]             # <<<<<<<<<<<<<<
 *             shifted = <int64_t>(<uint64_t>self.value >> 8)
 *             self.value = shifted ^ swizzle
 */
    __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_self->value); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyNumber_Xor(__pyx_t_3, __pyx_v_char); if (unlikely(!__pyx_t_4)) __PYX_ERR(18, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_AndObjC(__pyx_t_4, __pyx_int_255, 0xff, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(18, 48, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_swizzle = (__pyx_v_5cavro_RABIN_TABLE[__pyx_t_5]);

    /* "src/rabin.pxi":49
 *         for char in data:
 *             swizzle = RABIN_TABLE[(self.value ^ char) & 0xff]
 *             shifted = <int64_t>(<uint64_t>self.value >> 8)             # <<<<<<<<<<<<<<
 *             self.value = shifted ^ swizzle
 * 
 */
    __pyx_v_shifted = ((int64_t)(((uint64_t)__pyx_v_self->value) >> 8));

    /* "src/rabin.pxi":50
 *             swizzle = RABIN_TABLE[(self.value ^ char) & 0xff]
 *             shifted = <int64_t>(<uint64_t>self.value >> 8)
 *             self.value = shifted ^ swizzle             # <<<<<<<<<<<<<<
 * 
 *     def digest(self):
 */
    __pyx_v_self->value = (__pyx_v_shifted ^ __pyx_v_swizzle);

    /* "src/rabin.pxi":47
 *         cdef int64_t swizzle
 *         cdef int64_t shifted
 *         for char in data:             # <<<<<<<<<<<<<<
 *             swizzle = RABIN_TABLE[(self.value ^ char) & 0xff]
 *             shifted = <int64_t>(<uint64_t>self.value >> 8)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/rabin.pxi":44
 *         self.value = value
 * 
 *     cpdef update(self, bytes data):             # <<<<<<<<<<<<<<
 *         cdef int64_t swizzle
 *         cdef int64_t shifted
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Rabin.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_char);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_5Rabin_3update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_5Rabin_3update = {"update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Rabin_3update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_5Rabin_3update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 44, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "update") < 0)) __PYX_ERR(18, 44, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_data = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("update", 1, 1, 1, __pyx_nargs); __PYX_ERR(18, 44, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Rabin.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data), (&PyBytes_Type), 1, "data", 1))) __PYX_ERR(18, 44, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_5Rabin_2update(((struct __pyx_obj_5cavro_Rabin *)__pyx_v_self), __pyx_v_data);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_5Rabin_2update(struct __pyx_obj_5cavro_Rabin *__pyx_v_self, PyObject *__pyx_v_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_5Rabin_update(__pyx_v_self, __pyx_v_data, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Rabin.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/rabin.pxi":52
 *             self.value = shifted ^ swizzle
 * 
 *     def digest(self):             # <<<<<<<<<<<<<<
 *         cdef uint64_t value = self.value
 *         return int(value).to_bytes(8, 'little', signed=False)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_5Rabin_5digest(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_5Rabin_5digest = {"digest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Rabin_5digest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_5Rabin_5digest(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("digest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("digest", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "digest", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_5Rabin_4digest(((struct __pyx_obj_5cavro_Rabin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_5Rabin_4digest(struct __pyx_obj_5cavro_Rabin *__pyx_v_self) {
  uint64_t __pyx_v_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("digest", 1);

  /* "src/rabin.pxi":53
 * 
 *     def digest(self):
 *         cdef uint64_t value = self.value             # <<<<<<<<<<<<<<
 *         return int(value).to_bytes(8, 'little', signed=False)
 * 
 */
  __pyx_t_1 = __pyx_v_self->value;
  __pyx_v_value = __pyx_t_1;

  /* "src/rabin.pxi":54
 *     def digest(self):
 *         cdef uint64_t value = self.value
 *         return int(value).to_bytes(8, 'little', signed=False)             # <<<<<<<<<<<<<<
 * 
 *     def hexdigest(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_uint64_t(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyInt_Type)), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_to_bytes); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_signed, Py_False) < 0) __PYX_ERR(18, 54, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__22, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(18, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "src/rabin.pxi":52
 *             self.value = shifted ^ swizzle
 * 
 *     def digest(self):             # <<<<<<<<<<<<<<
 *         cdef uint64_t value = self.value
 *         return int(value).to_bytes(8, 'little', signed=False)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Rabin.digest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/rabin.pxi":56
 *         return int(value).to_bytes(8, 'little', signed=False)
 * 
 *     def hexdigest(self):             # <<<<<<<<<<<<<<
 *         return self.digest().hex()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_5Rabin_7hexdigest(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_5Rabin_7hexdigest = {"hexdigest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Rabin_7hexdigest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_5Rabin_7hexdigest(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hexdigest (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("hexdigest", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "hexdigest", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_5Rabin_6hexdigest(((struct __pyx_obj_5cavro_Rabin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_5Rabin_6hexdigest(struct __pyx_obj_5cavro_Rabin *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hexdigest", 1);

  /* "src/rabin.pxi":57
 * 
 *     def hexdigest(self):
 *         return self.digest().hex()             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_digest); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_hex); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/rabin.pxi":56
 *         return int(value).to_bytes(8, 'little', signed=False)
 * 
 *     def hexdigest(self):             # <<<<<<<<<<<<<<
 *         return self.digest().hex()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Rabin.hexdigest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/rabin.pxi":59
 *         return self.digest().hex()
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         new_ver = Rabin()
 *         new_ver.value = self.value
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_5Rabin_9copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_5Rabin_9copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Rabin_9copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_5Rabin_9copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_5Rabin_8copy(((struct __pyx_obj_5cavro_Rabin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_5Rabin_8copy(struct __pyx_obj_5cavro_Rabin *__pyx_v_self) {
  struct __pyx_obj_5cavro_Rabin *__pyx_v_new_ver = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "src/rabin.pxi":60
 * 
 *     def copy(self):
 *         new_ver = Rabin()             # <<<<<<<<<<<<<<
 *         new_ver.value = self.value
 *         return new_ver
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro_Rabin)); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_new_ver = ((struct __pyx_obj_5cavro_Rabin *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/rabin.pxi":61
 *     def copy(self):
 *         new_ver = Rabin()
 *         new_ver.value = self.value             # <<<<<<<<<<<<<<
 *         return new_ver
 */
  __pyx_t_2 = __pyx_v_self->value;
  __pyx_v_new_ver->value = __pyx_t_2;

  /* "src/rabin.pxi":62
 *         new_ver = Rabin()
 *         new_ver.value = self.value
 *         return new_ver             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_new_ver);
  __pyx_r = ((PyObject *)__pyx_v_new_ver);
  goto __pyx_L0;

  /* "src/rabin.pxi":59
 *         return self.digest().hex()
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         new_ver = Rabin()
 *         new_ver.value = self.value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Rabin.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_new_ver);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/rabin.pxi":37
 *     block_size = 8
 * 
 *     cdef readonly int64_t value             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, value=RABIN_EMPTY):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_5Rabin_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_5Rabin_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_5Rabin_5value___get__(((struct __pyx_obj_5cavro_Rabin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_5Rabin_5value___get__(struct __pyx_obj_5cavro_Rabin *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->value); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Rabin.value.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_5Rabin_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_5Rabin_11__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Rabin_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_5Rabin_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_5Rabin_10__reduce_cython__(((struct __pyx_obj_5cavro_Rabin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_5Rabin_10__reduce_cython__(struct __pyx_obj_5cavro_Rabin *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.value,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->value); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.value,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.value,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.value,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Rabin, (type(self), 0xcd42404, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Rabin, (type(self), 0xcd42404, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_Rabin, (type(self), 0xcd42404, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Rabin, (type(self), 0xcd42404, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Rabin); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_215229444);
    __Pyx_GIVEREF(__pyx_int_215229444);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_215229444)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Rabin, (type(self), 0xcd42404, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Rabin, (type(self), 0xcd42404, None), state
 *     else:
 *         return __pyx_unpickle_Rabin, (type(self), 0xcd42404, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Rabin__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Rabin); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_215229444);
    __Pyx_GIVEREF(__pyx_int_215229444);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_215229444)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Rabin.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Rabin, (type(self), 0xcd42404, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Rabin__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_5Rabin_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_5Rabin_13__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Rabin_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_5Rabin_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Rabin.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_5Rabin_12__setstate_cython__(((struct __pyx_obj_5cavro_Rabin *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_5Rabin_12__setstate_cython__(struct __pyx_obj_5cavro_Rabin *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Rabin, (type(self), 0xcd42404, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Rabin__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_Rabin__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Rabin, (type(self), 0xcd42404, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Rabin__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Rabin.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_zigzag.pxi":2
 * 
 * cpdef str _ubin(val, n_bytes):             # <<<<<<<<<<<<<<
 *     cdef str raw = bin(val)[2:]
 *     padded = raw.rjust(n_bytes * 8, '0')
 */

static PyObject *__pyx_pw_5cavro_7_ubin(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro__ubin(PyObject *__pyx_v_val, PyObject *__pyx_v_n_bytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_raw = 0;
  PyObject *__pyx_v_padded = NULL;
  PyObject *__pyx_v_parts = NULL;
  Py_ssize_t __pyx_8genexpr4__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ubin", 1);

  /* "src/tests/test_zigzag.pxi":3
 * 
 * cpdef str _ubin(val, n_bytes):
 *     cdef str raw = bin(val)[2:]             # <<<<<<<<<<<<<<
 *     padded = raw.rjust(n_bytes * 8, '0')
 *     parts = [padded[i:i+8] for i in range(0, len(padded), 8)]
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_bin, __pyx_v_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_t_1, 2, 0, NULL, NULL, &__pyx_slice__23, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(6, 3, __pyx_L1_error)
  __pyx_v_raw = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/tests/test_zigzag.pxi":4
 * cpdef str _ubin(val, n_bytes):
 *     cdef str raw = bin(val)[2:]
 *     padded = raw.rjust(n_bytes * 8, '0')             # <<<<<<<<<<<<<<
 *     parts = [padded[i:i+8] for i in range(0, len(padded), 8)]
 *     return " ".join(parts)
 */
  __pyx_t_2 = __Pyx_PyInt_MultiplyObjC(__pyx_v_n_bytes, __pyx_int_8, 8, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyUnicode_Type_rjust, __pyx_v_raw, __pyx_t_2, __pyx_kp_u_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_padded = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/tests/test_zigzag.pxi":5
 *     cdef str raw = bin(val)[2:]
 *     padded = raw.rjust(n_bytes * 8, '0')
 *     parts = [padded[i:i+8] for i in range(0, len(padded), 8)]             # <<<<<<<<<<<<<<
 *     return " ".join(parts)
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_Length(__pyx_v_padded); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(6, 5, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=8) {
      __pyx_8genexpr4__pyx_v_i = __pyx_t_5;
      __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_padded, __pyx_8genexpr4__pyx_v_i, (__pyx_8genexpr4__pyx_v_i + 8), NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(6, 5, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  } /* exit inner scope */
  __pyx_v_parts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/tests/test_zigzag.pxi":6
 *     padded = raw.rjust(n_bytes * 8, '0')
 *     parts = [padded[i:i+8] for i in range(0, len(padded), 8)]
 *     return " ".join(parts)             # <<<<<<<<<<<<<<
 * 
 * cdef struct random_t:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__24, __pyx_v_parts); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/tests/test_zigzag.pxi":2
 * 
 * cpdef str _ubin(val, n_bytes):             # <<<<<<<<<<<<<<
 *     cdef str raw = bin(val)[2:]
 *     padded = raw.rjust(n_bytes * 8, '0')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro._ubin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_raw);
  __Pyx_XDECREF(__pyx_v_padded);
  __Pyx_XDECREF(__pyx_v_parts);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7_ubin(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7_ubin = {"_ubin", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7_ubin, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7_ubin(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_val = 0;
  PyObject *__pyx_v_n_bytes = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_ubin (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_val,&__pyx_n_s_n_bytes,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_val)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 2, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_n_bytes)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 2, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_ubin", 1, 2, 2, 1); __PYX_ERR(6, 2, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_ubin") < 0)) __PYX_ERR(6, 2, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_val = values[0];
    __pyx_v_n_bytes = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_ubin", 1, 2, 2, __pyx_nargs); __PYX_ERR(6, 2, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._ubin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6_ubin(__pyx_self, __pyx_v_val, __pyx_v_n_bytes);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6_ubin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_val, PyObject *__pyx_v_n_bytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ubin", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro__ubin(__pyx_v_val, __pyx_v_n_bytes, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._ubin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_zigzag.pxi":12
 *     uint64_t inc
 * 
 * cdef uint32_t rand(random_t *rng):             # <<<<<<<<<<<<<<
 *     cdef uint64_t oldstate = rng.state;
 *     rng.state = oldstate * 6364136223846793005ULL + (rng.inc|1);
 */

static uint32_t __pyx_f_5cavro_rand(struct __pyx_t_5cavro_random_t *__pyx_v_rng) {
  uint64_t __pyx_v_oldstate;
  uint32_t __pyx_v_xorshifted;
  uint32_t __pyx_v_rot;
  uint32_t __pyx_r;
  uint64_t __pyx_t_1;

  /* "src/tests/test_zigzag.pxi":13
 * 
 * cdef uint32_t rand(random_t *rng):
 *     cdef uint64_t oldstate = rng.state;             # <<<<<<<<<<<<<<
 *     rng.state = oldstate * 6364136223846793005ULL + (rng.inc|1);
 *     cdef uint32_t xorshifted = ((oldstate >> 18u) ^ oldstate) >> 27u;
 */
  __pyx_t_1 = __pyx_v_rng->state;
  __pyx_v_oldstate = __pyx_t_1;

  /* "src/tests/test_zigzag.pxi":14
 * cdef uint32_t rand(random_t *rng):
 *     cdef uint64_t oldstate = rng.state;
 *     rng.state = oldstate * 6364136223846793005ULL + (rng.inc|1);             # <<<<<<<<<<<<<<
 *     cdef uint32_t xorshifted = ((oldstate >> 18u) ^ oldstate) >> 27u;
 *     cdef uint32_t rot = oldstate >> 59u;
 */
  __pyx_v_rng->state = ((__pyx_v_oldstate * 6364136223846793005ULL) + (__pyx_v_rng->inc | 1));

  /* "src/tests/test_zigzag.pxi":15
 *     cdef uint64_t oldstate = rng.state;
 *     rng.state = oldstate * 6364136223846793005ULL + (rng.inc|1);
 *     cdef uint32_t xorshifted = ((oldstate >> 18u) ^ oldstate) >> 27u;             # <<<<<<<<<<<<<<
 *     cdef uint32_t rot = oldstate >> 59u;
 *     return (xorshifted >> rot) | (xorshifted << ((-rot) & 31));
 */
  __pyx_v_xorshifted = (((__pyx_v_oldstate >> 18U) ^ __pyx_v_oldstate) >> 27U);

  /* "src/tests/test_zigzag.pxi":16
 *     rng.state = oldstate * 6364136223846793005ULL + (rng.inc|1);
 *     cdef uint32_t xorshifted = ((oldstate >> 18u) ^ oldstate) >> 27u;
 *     cdef uint32_t rot = oldstate >> 59u;             # <<<<<<<<<<<<<<
 *     return (xorshifted >> rot) | (xorshifted << ((-rot) & 31));
 * 
 */
  __pyx_v_rot = (__pyx_v_oldstate >> 59U);

  /* "src/tests/test_zigzag.pxi":17
 *     cdef uint32_t xorshifted = ((oldstate >> 18u) ^ oldstate) >> 27u;
 *     cdef uint32_t rot = oldstate >> 59u;
 *     return (xorshifted >> rot) | (xorshifted << ((-rot) & 31));             # <<<<<<<<<<<<<<
 * 
 * cdef bytes boring_varint_encoder(uint64_t val):
 */
  __pyx_r = ((__pyx_v_xorshifted >> __pyx_v_rot) | (__pyx_v_xorshifted << ((-__pyx_v_rot) & 31)));
  goto __pyx_L0;

  /* "src/tests/test_zigzag.pxi":12
 *     uint64_t inc
 * 
 * cdef uint32_t rand(random_t *rng):             # <<<<<<<<<<<<<<
 *     cdef uint64_t oldstate = rng.state;
 *     rng.state = oldstate * 6364136223846793005ULL + (rng.inc|1);
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "src/tests/test_zigzag.pxi":19
 *     return (xorshifted >> rot) | (xorshifted << ((-rot) & 31));
 * 
 * cdef bytes boring_varint_encoder(uint64_t val):             # <<<<<<<<<<<<<<
 *     cdef uint8_t cur = 0
 *     out = []
 */

static PyObject *__pyx_f_5cavro_boring_varint_encoder(uint64_t __pyx_v_val) {
  uint8_t __pyx_v_cur;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("boring_varint_encoder", 1);

  /* "src/tests/test_zigzag.pxi":20
 * 
 * cdef bytes boring_varint_encoder(uint64_t val):
 *     cdef uint8_t cur = 0             # <<<<<<<<<<<<<<
 *     out = []
 *     while val:
 */
  __pyx_v_cur = 0;

  /* "src/tests/test_zigzag.pxi":21
 * cdef bytes boring_varint_encoder(uint64_t val):
 *     cdef uint8_t cur = 0
 *     out = []             # <<<<<<<<<<<<<<
 *     while val:
 *         cur = val & 0x7f
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/tests/test_zigzag.pxi":22
 *     cdef uint8_t cur = 0
 *     out = []
 *     while val:             # <<<<<<<<<<<<<<
 *         cur = val & 0x7f
 *         val >>= 7
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_val != 0);
    if (!__pyx_t_2) break;

    /* "src/tests/test_zigzag.pxi":23
 *     out = []
 *     while val:
 *         cur = val & 0x7f             # <<<<<<<<<<<<<<
 *         val >>= 7
 *         if val:
 */
    __pyx_v_cur = (__pyx_v_val & 0x7f);

    /* "src/tests/test_zigzag.pxi":24
 *     while val:
 *         cur = val & 0x7f
 *         val >>= 7             # <<<<<<<<<<<<<<
 *         if val:
 *             out.append(cur | 0x80)
 */
    __pyx_v_val = (__pyx_v_val >> 7);

    /* "src/tests/test_zigzag.pxi":25
 *         cur = val & 0x7f
 *         val >>= 7
 *         if val:             # <<<<<<<<<<<<<<
 *             out.append(cur | 0x80)
 *     out.append(cur)
 */
    __pyx_t_2 = (__pyx_v_val != 0);
    if (__pyx_t_2) {

      /* "src/tests/test_zigzag.pxi":26
 *         val >>= 7
 *         if val:
 *             out.append(cur | 0x80)             # <<<<<<<<<<<<<<
 *     out.append(cur)
 *     return bytes(out)
 */
      __pyx_t_1 = __Pyx_PyInt_From_long((__pyx_v_cur | 0x80)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 26, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(6, 26, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "src/tests/test_zigzag.pxi":25
 *         cur = val & 0x7f
 *         val >>= 7
 *         if val:             # <<<<<<<<<<<<<<
 *             out.append(cur | 0x80)
 *     out.append(cur)
 */
    }
  }

  /* "src/tests/test_zigzag.pxi":27
 *         if val:
 *             out.append(cur | 0x80)
 *     out.append(cur)             # <<<<<<<<<<<<<<
 *     return bytes(out)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_v_cur); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(6, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/tests/test_zigzag.pxi":28
 *             out.append(cur | 0x80)
 *     out.append(cur)
 *     return bytes(out)             # <<<<<<<<<<<<<<
 * 
 * @_tests
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_v_out); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/tests/test_zigzag.pxi":19
 *     return (xorshifted >> rot) | (xorshifted << ((-rot) & 31));
 * 
 * cdef bytes boring_varint_encoder(uint64_t val):             # <<<<<<<<<<<<<<
 *     cdef uint8_t cur = 0
 *     out = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.boring_varint_encoder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_zigzag.pxi":30
 *     return bytes(out)
 * 
 * @_tests             # <<<<<<<<<<<<<<
 * def _tests(add):
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9_tests(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9_tests = {"_tests", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9_tests, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9_tests(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_add = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_tests (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_add,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_add)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 30, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_tests") < 0)) __PYX_ERR(6, 30, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_add = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_tests", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 30, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._tests", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8_tests(__pyx_self, __pyx_v_add);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_zigzag.pxi":33
 * def _tests(add):
 * 
 *     def test_readvarint(uint32_t mask):             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6_tests_3test_readvarint(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_6_tests_3test_readvarint = {"test_readvarint", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6_tests_3test_readvarint, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_6_tests_3test_readvarint(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  uint32_t __pyx_v_mask;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("test_readvarint (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mask,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mask)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 33, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "test_readvarint") < 0)) __PYX_ERR(6, 33, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_mask = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_mask == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(6, 33, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("test_readvarint", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 33, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._tests.test_readvarint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6_tests_2test_readvarint(__pyx_self, __pyx_v_mask);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6_tests_2test_readvarint(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_mask) {
  struct __pyx_t_5cavro_random_t __pyx_v_state;
  uint32_t __pyx_v_actual;
  uint32_t __pyx_v_given;
  PyObject *__pyx_v_src = 0;
  struct __pyx_obj_5cavro_MemoryReader *__pyx_v_reader = 0;
  CYTHON_UNUSED uint32_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  uint32_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("test_readvarint", 1);

  /* "src/tests/test_zigzag.pxi":43
 *         cdef uint32_t i
 * 
 *         for i in range(50_000):             # <<<<<<<<<<<<<<
 *             given = rand(&state) & mask
 *             src = boring_varint_encoder(given)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 0xC350; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "src/tests/test_zigzag.pxi":44
 * 
 *         for i in range(50_000):
 *             given = rand(&state) & mask             # <<<<<<<<<<<<<<
 *             src = boring_varint_encoder(given)
 *             reader = MemoryReader(src)
 */
    __pyx_t_2 = __pyx_f_5cavro_rand((&__pyx_v_state)); if (unlikely(__pyx_t_2 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(6, 44, __pyx_L1_error)
    __pyx_v_given = (__pyx_t_2 & __pyx_v_mask);

    /* "src/tests/test_zigzag.pxi":45
 *         for i in range(50_000):
 *             given = rand(&state) & mask
 *             src = boring_varint_encoder(given)             # <<<<<<<<<<<<<<
 *             reader = MemoryReader(src)
 *             actual = read_varint(reader)
 */
    __pyx_t_3 = __pyx_f_5cavro_boring_varint_encoder(__pyx_v_given); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_src, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/tests/test_zigzag.pxi":46
 *             given = rand(&state) & mask
 *             src = boring_varint_encoder(given)
 *             reader = MemoryReader(src)             # <<<<<<<<<<<<<<
 *             actual = read_varint(reader)
 *             if actual != given:
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro_MemoryReader), __pyx_v_src); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_reader, ((struct __pyx_obj_5cavro_MemoryReader *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/tests/test_zigzag.pxi":47
 *             src = boring_varint_encoder(given)
 *             reader = MemoryReader(src)
 *             actual = read_varint(reader)             # <<<<<<<<<<<<<<
 *             if actual != given:
 *                 raise AssertionError(f"{src}:\n   {_ubin(actual, 4)}\n!= {_ubin(given, 4)}")
 */
    __pyx_t_2 = __pyx_f_5cavro_read_varint(((struct __pyx_obj_5cavro__Reader *)__pyx_v_reader)); if (unlikely(__pyx_t_2 == ((uint32_t)0xfffffbad) && PyErr_Occurred())) __PYX_ERR(6, 47, __pyx_L1_error)
    __pyx_v_actual = __pyx_t_2;

    /* "src/tests/test_zigzag.pxi":48
 *             reader = MemoryReader(src)
 *             actual = read_varint(reader)
 *             if actual != given:             # <<<<<<<<<<<<<<
 *                 raise AssertionError(f"{src}:\n   {_ubin(actual, 4)}\n!= {_ubin(given, 4)}")
 *     for mask in [0x7f, 0x7fff, 0x7fffff, 0x7fffffff, UINT32_MAX]:
 */
    __pyx_t_4 = (__pyx_v_actual != __pyx_v_given);
    if (unlikely(__pyx_t_4)) {

      /* "src/tests/test_zigzag.pxi":49
 *             actual = read_varint(reader)
 *             if actual != given:
 *                 raise AssertionError(f"{src}:\n   {_ubin(actual, 4)}\n!= {_ubin(given, 4)}")             # <<<<<<<<<<<<<<
 *     for mask in [0x7f, 0x7fff, 0x7fffff, 0x7fffffff, UINT32_MAX]:
 *         add(test_readvarint, mask)
 */
      __pyx_t_3 = PyTuple_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = 0;
      __pyx_t_6 = 127;
      __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_v_src, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_INCREF(__pyx_kp_u__25);
      __pyx_t_5 += 5;
      __Pyx_GIVEREF(__pyx_kp_u__25);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_kp_u__25);
      __pyx_t_7 = __Pyx_PyInt_From_uint32_t(__pyx_v_actual); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __pyx_f_5cavro__ubin(__pyx_t_7, __pyx_int_4, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(6, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyUnicode_Unicode(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_INCREF(__pyx_kp_u__26);
      __pyx_t_5 += 4;
      __Pyx_GIVEREF(__pyx_kp_u__26);
      PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_kp_u__26);
      __pyx_t_7 = __Pyx_PyInt_From_uint32_t(__pyx_v_given); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __pyx_f_5cavro__ubin(__pyx_t_7, __pyx_int_4, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(6, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyUnicode_Unicode(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_3, 5, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(6, 49, __pyx_L1_error)

      /* "src/tests/test_zigzag.pxi":48
 *             reader = MemoryReader(src)
 *             actual = read_varint(reader)
 *             if actual != given:             # <<<<<<<<<<<<<<
 *                 raise AssertionError(f"{src}:\n   {_ubin(actual, 4)}\n!= {_ubin(given, 4)}")
 *     for mask in [0x7f, 0x7fff, 0x7fffff, 0x7fffffff, UINT32_MAX]:
 */
    }
  }

  /* "src/tests/test_zigzag.pxi":33
 * def _tests(add):
 * 
 *     def test_readvarint(uint32_t mask):             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cavro._tests.test_readvarint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_src);
  __Pyx_XDECREF((PyObject *)__pyx_v_reader);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_zigzag.pxi":53
 *         add(test_readvarint, mask)
 * 
 *     def test_zigzag_long(uint64_t mask):             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6_tests_5test_zigzag_long(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_6_tests_5test_zigzag_long = {"test_zigzag_long", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6_tests_5test_zigzag_long, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_6_tests_5test_zigzag_long(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  uint64_t __pyx_v_mask;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("test_zigzag_long (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mask,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mask)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 53, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "test_zigzag_long") < 0)) __PYX_ERR(6, 53, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_mask = __Pyx_PyInt_As_uint64_t(values[0]); if (unlikely((__pyx_v_mask == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(6, 53, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("test_zigzag_long", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 53, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._tests.test_zigzag_long", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6_tests_4test_zigzag_long(__pyx_self, __pyx_v_mask);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6_tests_4test_zigzag_long(CYTHON_UNUSED PyObject *__pyx_self, uint64_t __pyx_v_mask) {
  struct __pyx_t_5cavro_random_t __pyx_v_state;
  int64_t __pyx_v_actual;
  int64_t __pyx_v_given;
  struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_writer = 0;
  struct __pyx_obj_5cavro_MemoryReader *__pyx_v_reader = 0;
  CYTHON_UNUSED uint64_t __pyx_v_i;
  CYTHON_UNUSED uint64_t __pyx_v_pad;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint32_t __pyx_t_2;
  uint32_t __pyx_t_3;
  uint64_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int64_t __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("test_zigzag_long", 1);

  /* "src/tests/test_zigzag.pxi":58
 *         cdef int64_t actual
 *         cdef int64_t given
 *         cdef MemoryWriter writer = MemoryWriter()             # <<<<<<<<<<<<<<
 *         cdef MemoryReader reader
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro_MemoryWriter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_writer = ((struct __pyx_obj_5cavro_MemoryWriter *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/tests/test_zigzag.pxi":64
 *         cdef uint64_t pad
 * 
 *         pad = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32))             # <<<<<<<<<<<<<<
 *         for i in range(50_000):
 *             writer.reset()
 */
  __pyx_t_2 = __pyx_f_5cavro_rand((&__pyx_v_state)); if (unlikely(__pyx_t_2 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(6, 64, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro_rand((&__pyx_v_state)); if (unlikely(__pyx_t_3 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(6, 64, __pyx_L1_error)
  __pyx_v_pad = (((uint64_t)__pyx_t_2) | (((uint64_t)__pyx_t_3) << 32));

  /* "src/tests/test_zigzag.pxi":65
 * 
 *         pad = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32))
 *         for i in range(50_000):             # <<<<<<<<<<<<<<
 *             writer.reset()
 *             given = <int64_t>((<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask)
 */
  for (__pyx_t_4 = 0; __pyx_t_4 < 0xC350; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "src/tests/test_zigzag.pxi":66
 *         pad = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32))
 *         for i in range(50_000):
 *             writer.reset()             # <<<<<<<<<<<<<<
 *             given = <int64_t>((<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask)
 *             zigzag_encode_long(writer, given)
 */
    __pyx_t_1 = __pyx_f_5cavro_12MemoryWriter_reset(__pyx_v_writer); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/tests/test_zigzag.pxi":67
 *         for i in range(50_000):
 *             writer.reset()
 *             given = <int64_t>((<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask)             # <<<<<<<<<<<<<<
 *             zigzag_encode_long(writer, given)
 *             reader = MemoryReader(writer.bytes())
 */
    __pyx_t_3 = __pyx_f_5cavro_rand((&__pyx_v_state)); if (unlikely(__pyx_t_3 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(6, 67, __pyx_L1_error)
    __pyx_t_2 = __pyx_f_5cavro_rand((&__pyx_v_state)); if (unlikely(__pyx_t_2 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(6, 67, __pyx_L1_error)
    __pyx_v_given = ((int64_t)((((uint64_t)__pyx_t_3) | (((uint64_t)__pyx_t_2) << 32)) & __pyx_v_mask));

    /* "src/tests/test_zigzag.pxi":68
 *             writer.reset()
 *             given = <int64_t>((<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask)
 *             zigzag_encode_long(writer, given)             # <<<<<<<<<<<<<<
 *             reader = MemoryReader(writer.bytes())
 *             actual = zigzag_decode_long(reader)
 */
    __pyx_t_5 = __pyx_f_5cavro_zigzag_encode_long(((struct __pyx_obj_5cavro__Writer *)__pyx_v_writer), __pyx_v_given); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(6, 68, __pyx_L1_error)

    /* "src/tests/test_zigzag.pxi":69
 *             given = <int64_t>((<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask)
 *             zigzag_encode_long(writer, given)
 *             reader = MemoryReader(writer.bytes())             # <<<<<<<<<<<<<<
 *             actual = zigzag_decode_long(reader)
 *             if actual != given:
 */
    __pyx_t_1 = __pyx_f_5cavro_12MemoryWriter_bytes(__pyx_v_writer); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro_MemoryReader), __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_reader, ((struct __pyx_obj_5cavro_MemoryReader *)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "src/tests/test_zigzag.pxi":70
 *             zigzag_encode_long(writer, given)
 *             reader = MemoryReader(writer.bytes())
 *             actual = zigzag_decode_long(reader)             # <<<<<<<<<<<<<<
 *             if actual != given:
 *                 raise AssertionError(f"{writer.bytes()}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")
 */
    __pyx_t_7 = __pyx_f_5cavro_zigzag_decode_long(((struct __pyx_obj_5cavro__Reader *)__pyx_v_reader)); if (unlikely(__pyx_t_7 == ((int64_t)0x7ffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(6, 70, __pyx_L1_error)
    __pyx_v_actual = __pyx_t_7;

    /* "src/tests/test_zigzag.pxi":71
 *             reader = MemoryReader(writer.bytes())
 *             actual = zigzag_decode_long(reader)
 *             if actual != given:             # <<<<<<<<<<<<<<
 *                 raise AssertionError(f"{writer.bytes()}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")
 * 
 */
    __pyx_t_8 = (__pyx_v_actual != __pyx_v_given);
    if (unlikely(__pyx_t_8)) {

      /* "src/tests/test_zigzag.pxi":72
 *             actual = zigzag_decode_long(reader)
 *             if actual != given:
 *                 raise AssertionError(f"{writer.bytes()}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")             # <<<<<<<<<<<<<<
 * 
 *     def test_readvarlong(uint64_t mask):
 */
      __pyx_t_6 = PyTuple_New(7); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;
      __pyx_t_1 = __pyx_f_5cavro_12MemoryWriter_bytes(__pyx_v_writer); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_11);
      __pyx_t_11 = 0;
      __Pyx_INCREF(__pyx_kp_u_);
      __pyx_t_9 += 2;
      __Pyx_GIVEREF(__pyx_kp_u_);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_kp_u_);
      __pyx_t_11 = __Pyx_PyInt_From_int64_t(__pyx_v_given); if (unlikely(!__pyx_t_11)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_11, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_INCREF(__pyx_kp_u__27);
      __pyx_t_9 += 4;
      __Pyx_GIVEREF(__pyx_kp_u__27);
      PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_kp_u__27);
      __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_actual); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __pyx_f_5cavro__ubin(__pyx_t_1, __pyx_int_8, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_INCREF(__pyx_kp_u__26);
      __pyx_t_9 += 4;
      __Pyx_GIVEREF(__pyx_kp_u__26);
      PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_kp_u__26);
      __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_given); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __pyx_f_5cavro__ubin(__pyx_t_1, __pyx_int_8, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 6, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_6, 7, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(6, 72, __pyx_L1_error)

      /* "src/tests/test_zigzag.pxi":71
 *             reader = MemoryReader(writer.bytes())
 *             actual = zigzag_decode_long(reader)
 *             if actual != given:             # <<<<<<<<<<<<<<
 *                 raise AssertionError(f"{writer.bytes()}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")
 * 
 */
    }
  }

  /* "src/tests/test_zigzag.pxi":53
 *         add(test_readvarint, mask)
 * 
 *     def test_zigzag_long(uint64_t mask):             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("cavro._tests.test_zigzag_long", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_writer);
  __Pyx_XDECREF((PyObject *)__pyx_v_reader);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_zigzag.pxi":74
 *                 raise AssertionError(f"{writer.bytes()}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")
 * 
 *     def test_readvarlong(uint64_t mask):             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6_tests_7test_readvarlong(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_6_tests_7test_readvarlong = {"test_readvarlong", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6_tests_7test_readvarlong, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_6_tests_7test_readvarlong(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  uint64_t __pyx_v_mask;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("test_readvarlong (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mask,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mask)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 74, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "test_readvarlong") < 0)) __PYX_ERR(6, 74, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_mask = __Pyx_PyInt_As_uint64_t(values[0]); if (unlikely((__pyx_v_mask == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(6, 74, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("test_readvarlong", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 74, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._tests.test_readvarlong", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6_tests_6test_readvarlong(__pyx_self, __pyx_v_mask);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6_tests_6test_readvarlong(CYTHON_UNUSED PyObject *__pyx_self, uint64_t __pyx_v_mask) {
  struct __pyx_t_5cavro_random_t __pyx_v_state;
  uint64_t __pyx_v_actual;
  uint64_t __pyx_v_given;
  PyObject *__pyx_v_src = 0;
  struct __pyx_obj_5cavro_MemoryReader *__pyx_v_reader = 0;
  CYTHON_UNUSED uint64_t __pyx_v_i;
  uint64_t __pyx_v_pad;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  uint32_t __pyx_t_2;
  uint64_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  uint64_t __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("test_readvarlong", 1);

  /* "src/tests/test_zigzag.pxi":85
 *         cdef uint64_t pad
 * 
 *         pad = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32))             # <<<<<<<<<<<<<<
 *         for i in range(50_000):
 *             given = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask
 */
  __pyx_t_1 = __pyx_f_5cavro_rand((&__pyx_v_state)); if (unlikely(__pyx_t_1 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(6, 85, __pyx_L1_error)
  __pyx_t_2 = __pyx_f_5cavro_rand((&__pyx_v_state)); if (unlikely(__pyx_t_2 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(6, 85, __pyx_L1_error)
  __pyx_v_pad = (((uint64_t)__pyx_t_1) | (((uint64_t)__pyx_t_2) << 32));

  /* "src/tests/test_zigzag.pxi":86
 * 
 *         pad = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32))
 *         for i in range(50_000):             # <<<<<<<<<<<<<<
 *             given = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask
 *             src = boring_varint_encoder(given)
 */
  for (__pyx_t_3 = 0; __pyx_t_3 < 0xC350; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "src/tests/test_zigzag.pxi":87
 *         pad = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32))
 *         for i in range(50_000):
 *             given = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask             # <<<<<<<<<<<<<<
 *             src = boring_varint_encoder(given)
 *             src += (<char*>&pad)[:8]
 */
    __pyx_t_2 = __pyx_f_5cavro_rand((&__pyx_v_state)); if (unlikely(__pyx_t_2 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(6, 87, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro_rand((&__pyx_v_state)); if (unlikely(__pyx_t_1 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(6, 87, __pyx_L1_error)
    __pyx_v_given = ((((uint64_t)__pyx_t_2) | (((uint64_t)__pyx_t_1) << 32)) & __pyx_v_mask);

    /* "src/tests/test_zigzag.pxi":88
 *         for i in range(50_000):
 *             given = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask
 *             src = boring_varint_encoder(given)             # <<<<<<<<<<<<<<
 *             src += (<char*>&pad)[:8]
 *             reader = MemoryReader(src)
 */
    __pyx_t_4 = __pyx_f_5cavro_boring_varint_encoder(__pyx_v_given); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_src, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "src/tests/test_zigzag.pxi":89
 *             given = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask
 *             src = boring_varint_encoder(given)
 *             src += (<char*>&pad)[:8]             # <<<<<<<<<<<<<<
 *             reader = MemoryReader(src)
 *             actual = read_varlong(reader)
 */
    __pyx_t_4 = __Pyx_PyBytes_FromStringAndSize(((char *)(&__pyx_v_pad)) + 0, 8 - 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_src, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_src, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "src/tests/test_zigzag.pxi":90
 *             src = boring_varint_encoder(given)
 *             src += (<char*>&pad)[:8]
 *             reader = MemoryReader(src)             # <<<<<<<<<<<<<<
 *             actual = read_varlong(reader)
 *             if actual != given:
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro_MemoryReader), __pyx_v_src); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_reader, ((struct __pyx_obj_5cavro_MemoryReader *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "src/tests/test_zigzag.pxi":91
 *             src += (<char*>&pad)[:8]
 *             reader = MemoryReader(src)
 *             actual = read_varlong(reader)             # <<<<<<<<<<<<<<
 *             if actual != given:
 *                 raise AssertionError(f"{src}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")
 */
    __pyx_t_6 = __pyx_f_5cavro_read_varlong(((struct __pyx_obj_5cavro__Reader *)__pyx_v_reader)); if (unlikely(__pyx_t_6 == ((uint64_t)0xfffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(6, 91, __pyx_L1_error)
    __pyx_v_actual = __pyx_t_6;

    /* "src/tests/test_zigzag.pxi":92
 *             reader = MemoryReader(src)
 *             actual = read_varlong(reader)
 *             if actual != given:             # <<<<<<<<<<<<<<
 *                 raise AssertionError(f"{src}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")
 * 
 */
    __pyx_t_7 = (__pyx_v_actual != __pyx_v_given);
    if (unlikely(__pyx_t_7)) {

      /* "src/tests/test_zigzag.pxi":93
 *             actual = read_varlong(reader)
 *             if actual != given:
 *                 raise AssertionError(f"{src}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")             # <<<<<<<<<<<<<<
 * 
 *     for mask in [
 */
      __pyx_t_5 = PyTuple_New(7); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = 0;
      __pyx_t_9 = 127;
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_src, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_9;
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_INCREF(__pyx_kp_u_);
      __pyx_t_8 += 2;
      __Pyx_GIVEREF(__pyx_kp_u_);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_kp_u_);
      __pyx_t_4 = __Pyx_PyInt_From_uint64_t(__pyx_v_given); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_10 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_9;
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_INCREF(__pyx_kp_u__27);
      __pyx_t_8 += 4;
      __Pyx_GIVEREF(__pyx_kp_u__27);
      PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_kp_u__27);
      __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_actual); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_4 = __pyx_f_5cavro__ubin(__pyx_t_10, __pyx_int_8, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyUnicode_Unicode(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_9;
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_INCREF(__pyx_kp_u__26);
      __pyx_t_8 += 4;
      __Pyx_GIVEREF(__pyx_kp_u__26);
      PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_kp_u__26);
      __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_given); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_4 = __pyx_f_5cavro__ubin(__pyx_t_10, __pyx_int_8, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyUnicode_Unicode(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_9;
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_5, 6, __pyx_t_10);
      __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyUnicode_Join(__pyx_t_5, 7, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(6, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(6, 93, __pyx_L1_error)

      /* "src/tests/test_zigzag.pxi":92
 *             reader = MemoryReader(src)
 *             actual = read_varlong(reader)
 *             if actual != given:             # <<<<<<<<<<<<<<
 *                 raise AssertionError(f"{src}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")
 * 
 */
    }
  }

  /* "src/tests/test_zigzag.pxi":74
 *                 raise AssertionError(f"{writer.bytes()}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")
 * 
 *     def test_readvarlong(uint64_t mask):             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("cavro._tests.test_readvarlong", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_src);
  __Pyx_XDECREF((PyObject *)__pyx_v_reader);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_zigzag.pxi":30
 *     return bytes(out)
 * 
 * @_tests             # <<<<<<<<<<<<<<
 * def _tests(add):
 * 
 */

static PyObject *__pyx_pf_5cavro_8_tests(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_add) {
  PyObject *__pyx_v_test_readvarint = 0;
  PyObject *__pyx_v_mask = NULL;
  PyObject *__pyx_v_test_zigzag_long = 0;
  PyObject *__pyx_v_test_readvarlong = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_tests", 1);

  /* "src/tests/test_zigzag.pxi":33
 * def _tests(add):
 * 
 *     def test_readvarint(uint32_t mask):             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6_tests_3test_readvarint, 0, __pyx_n_s_tests_locals_test_readvarint, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_test_readvarint = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/tests/test_zigzag.pxi":50
 *             if actual != given:
 *                 raise AssertionError(f"{src}:\n   {_ubin(actual, 4)}\n!= {_ubin(given, 4)}")
 *     for mask in [0x7f, 0x7fff, 0x7fffff, 0x7fffffff, UINT32_MAX]:             # <<<<<<<<<<<<<<
 *         add(test_readvarint, mask)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(UINT32_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_127);
  __Pyx_GIVEREF(__pyx_int_127);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_127)) __PYX_ERR(6, 50, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_32767);
  __Pyx_GIVEREF(__pyx_int_32767);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_32767)) __PYX_ERR(6, 50, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_8388607);
  __Pyx_GIVEREF(__pyx_int_8388607);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_int_8388607)) __PYX_ERR(6, 50, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2147483647);
  __Pyx_GIVEREF(__pyx_int_2147483647);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_int_2147483647)) __PYX_ERR(6, 50, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_1)) __PYX_ERR(6, 50, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_3 >= 5) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(6, 50, __pyx_L1_error)
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_mask, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/tests/test_zigzag.pxi":51
 *                 raise AssertionError(f"{src}:\n   {_ubin(actual, 4)}\n!= {_ubin(given, 4)}")
 *     for mask in [0x7f, 0x7fff, 0x7fffff, 0x7fffffff, UINT32_MAX]:
 *         add(test_readvarint, mask)             # <<<<<<<<<<<<<<
 * 
 *     def test_zigzag_long(uint64_t mask):
 */
    __Pyx_INCREF(__pyx_v_add);
    __pyx_t_4 = __pyx_v_add; __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_test_readvarint, __pyx_v_mask};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 51, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/tests/test_zigzag.pxi":50
 *             if actual != given:
 *                 raise AssertionError(f"{src}:\n   {_ubin(actual, 4)}\n!= {_ubin(given, 4)}")
 *     for mask in [0x7f, 0x7fff, 0x7fffff, 0x7fffffff, UINT32_MAX]:             # <<<<<<<<<<<<<<
 *         add(test_readvarint, mask)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/tests/test_zigzag.pxi":53
 *         add(test_readvarint, mask)
 * 
 *     def test_zigzag_long(uint64_t mask):             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6_tests_5test_zigzag_long, 0, __pyx_n_s_tests_locals_test_zigzag_long, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_test_zigzag_long = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/tests/test_zigzag.pxi":74
 *                 raise AssertionError(f"{writer.bytes()}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")
 * 
 *     def test_readvarlong(uint64_t mask):             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6_tests_7test_readvarlong, 0, __pyx_n_s_tests_locals_test_readvarlong, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_test_readvarlong = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/tests/test_zigzag.pxi":98
 *         0x7f, 0x7fff, 0x7fffff, 0x7fffffff, 0x7fffffffffull,
 *          0x7fffffffffffull, 0x7fffffffffffffull, 0x7fffffffffffffffull,
 *         UINT64_MAX]:             # <<<<<<<<<<<<<<
 *         add(test_readvarlong, mask)
 *         add(test_zigzag_long, mask)
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(UINT64_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/tests/test_zigzag.pxi":95
 *                 raise AssertionError(f"{src}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")
 * 
 *     for mask in [             # <<<<<<<<<<<<<<
 *         0x7f, 0x7fff, 0x7fffff, 0x7fffffff, 0x7fffffffffull,
 *          0x7fffffffffffull, 0x7fffffffffffffull, 0x7fffffffffffffffull,
 */
  __pyx_t_2 = PyTuple_New(9); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_127);
  __Pyx_GIVEREF(__pyx_int_127);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_127)) __PYX_ERR(6, 95, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_32767);
  __Pyx_GIVEREF(__pyx_int_32767);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_32767)) __PYX_ERR(6, 95, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_8388607);
  __Pyx_GIVEREF(__pyx_int_8388607);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_int_8388607)) __PYX_ERR(6, 95, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2147483647);
  __Pyx_GIVEREF(__pyx_int_2147483647);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_int_2147483647)) __PYX_ERR(6, 95, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_549755813887L);
  __Pyx_GIVEREF(__pyx_int_549755813887L);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_int_549755813887L)) __PYX_ERR(6, 95, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0x7fffffffffffL);
  __Pyx_GIVEREF(__pyx_int_0x7fffffffffffL);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_int_0x7fffffffffffL)) __PYX_ERR(6, 95, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0x7fffffffffffffL);
  __Pyx_GIVEREF(__pyx_int_0x7fffffffffffffL);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_int_0x7fffffffffffffL)) __PYX_ERR(6, 95, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0x7fffffffffffffffL);
  __Pyx_GIVEREF(__pyx_int_0x7fffffffffffffffL);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_int_0x7fffffffffffffffL)) __PYX_ERR(6, 95, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_t_1)) __PYX_ERR(6, 95, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_3 >= 9) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(6, 95, __pyx_L1_error)
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_mask, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/tests/test_zigzag.pxi":99
 *          0x7fffffffffffull, 0x7fffffffffffffull, 0x7fffffffffffffffull,
 *         UINT64_MAX]:
 *         add(test_readvarlong, mask)             # <<<<<<<<<<<<<<
 *         add(test_zigzag_long, mask)
 */
    __Pyx_INCREF(__pyx_v_add);
    __pyx_t_4 = __pyx_v_add; __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_test_readvarlong, __pyx_v_mask};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/tests/test_zigzag.pxi":100
 *         UINT64_MAX]:
 *         add(test_readvarlong, mask)
 *         add(test_zigzag_long, mask)             # <<<<<<<<<<<<<<
 */
    __Pyx_INCREF(__pyx_v_add);
    __pyx_t_4 = __pyx_v_add; __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_test_zigzag_long, __pyx_v_mask};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/tests/test_zigzag.pxi":95
 *                 raise AssertionError(f"{src}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")
 * 
 *     for mask in [             # <<<<<<<<<<<<<<
 *         0x7f, 0x7fff, 0x7fffff, 0x7fffffff, 0x7fffffffffull,
 *          0x7fffffffffffull, 0x7fffffffffffffull, 0x7fffffffffffffffull,
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/tests/test_zigzag.pxi":30
 *     return bytes(out)
 * 
 * @_tests             # <<<<<<<<<<<<<<
 * def _tests(add):
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro._tests", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_test_readvarint);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XDECREF(__pyx_v_test_zigzag_long);
  __Pyx_XDECREF(__pyx_v_test_readvarlong);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_buffer.pxi":2
 * 
 * @_tests             # <<<<<<<<<<<<<<
 * def _tests(add):
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11_tests(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_11_tests = {"_tests", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11_tests, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_11_tests(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_add = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_tests (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_add,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_add)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(19, 2, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_tests") < 0)) __PYX_ERR(19, 2, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_add = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_tests", 1, 1, 1, __pyx_nargs); __PYX_ERR(19, 2, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._tests", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10_tests(__pyx_self, __pyx_v_add);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_buffer.pxi":5
 * def _tests(add):
 * 
 *     @add             # <<<<<<<<<<<<<<
 *     def _test_reader_read_u8():
 *         buffer = MemoryReader(b'\x01\x02\x03')
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6_tests_9_test_reader_read_u8(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_6_tests_9_test_reader_read_u8 = {"_test_reader_read_u8", (PyCFunction)__pyx_pw_5cavro_6_tests_9_test_reader_read_u8, METH_NOARGS, 0};
static PyObject *__pyx_pw_5cavro_6_tests_9_test_reader_read_u8(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_reader_read_u8 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6_tests_8_test_reader_read_u8(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6_tests_8_test_reader_read_u8(CYTHON_UNUSED PyObject *__pyx_self) {
  struct __pyx_obj_5cavro_MemoryReader *__pyx_v_buffer = NULL;
  PyObject *__pyx_v_pytest = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint8_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_reader_read_u8", 1);

  /* "src/tests/test_buffer.pxi":7
 *     @add
 *     def _test_reader_read_u8():
 *         buffer = MemoryReader(b'\x01\x02\x03')             # <<<<<<<<<<<<<<
 *         assert buffer.read_u8() == 1
 *         assert buffer.read_u8() == 2
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_MemoryReader), __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buffer = ((struct __pyx_obj_5cavro_MemoryReader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/tests/test_buffer.pxi":8
 *     def _test_reader_read_u8():
 *         buffer = MemoryReader(b'\x01\x02\x03')
 *         assert buffer.read_u8() == 1             # <<<<<<<<<<<<<<
 *         assert buffer.read_u8() == 2
 *         assert buffer.read_u8() == 3
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_MemoryReader *)__pyx_v_buffer->__pyx_base.__pyx_vtab)->__pyx_base.read_u8(((struct __pyx_obj_5cavro__Reader *)__pyx_v_buffer)); if (unlikely(__pyx_t_2 == ((uint8_t)0xba) && PyErr_Occurred())) __PYX_ERR(19, 8, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_2 == 1);
    if (unlikely(!__pyx_t_3)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(19, 8, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(19, 8, __pyx_L1_error)
  #endif

  /* "src/tests/test_buffer.pxi":9
 *         buffer = MemoryReader(b'\x01\x02\x03')
 *         assert buffer.read_u8() == 1
 *         assert buffer.read_u8() == 2             # <<<<<<<<<<<<<<
 *         assert buffer.read_u8() == 3
 *         import pytest
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_MemoryReader *)__pyx_v_buffer->__pyx_base.__pyx_vtab)->__pyx_base.read_u8(((struct __pyx_obj_5cavro__Reader *)__pyx_v_buffer)); if (unlikely(__pyx_t_2 == ((uint8_t)0xba) && PyErr_Occurred())) __PYX_ERR(19, 9, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_2 == 2);
    if (unlikely(!__pyx_t_3)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(19, 9, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(19, 9, __pyx_L1_error)
  #endif

  /* "src/tests/test_buffer.pxi":10
 *         assert buffer.read_u8() == 1
 *         assert buffer.read_u8() == 2
 *         assert buffer.read_u8() == 3             # <<<<<<<<<<<<<<
 *         import pytest
 *         with pytest.raises(EOFError):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_MemoryReader *)__pyx_v_buffer->__pyx_base.__pyx_vtab)->__pyx_base.read_u8(((struct __pyx_obj_5cavro__Reader *)__pyx_v_buffer)); if (unlikely(__pyx_t_2 == ((uint8_t)0xba) && PyErr_Occurred())) __PYX_ERR(19, 10, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_2 == 3);
    if (unlikely(!__pyx_t_3)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(19, 10, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(19, 10, __pyx_L1_error)
  #endif

  /* "src/tests/test_buffer.pxi":11
 *         assert buffer.read_u8() == 2
 *         assert buffer.read_u8() == 3
 *         import pytest             # <<<<<<<<<<<<<<
 *         with pytest.raises(EOFError):
 *             buffer.read_u8()
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_pytest, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pytest = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/tests/test_buffer.pxi":12
 *         assert buffer.read_u8() == 3
 *         import pytest
 *         with pytest.raises(EOFError):             # <<<<<<<<<<<<<<
 *             buffer.read_u8()
 * 
 */
  /*with:*/ {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_pytest, __pyx_n_s_raises); if (unlikely(!__pyx_t_4)) __PYX_ERR(19, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_builtin_EOFError};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 12, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_7 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_7)) __PYX_ERR(19, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(19, 12, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(19, 12, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        /*try:*/ {

          /* "src/tests/test_buffer.pxi":13
 *         import pytest
 *         with pytest.raises(EOFError):
 *             buffer.read_u8()             # <<<<<<<<<<<<<<
 * 
 *     @add
 */
          __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_MemoryReader *)__pyx_v_buffer->__pyx_base.__pyx_vtab)->__pyx_base.read_u8(((struct __pyx_obj_5cavro__Reader *)__pyx_v_buffer)); if (unlikely(__pyx_t_2 == ((uint8_t)0xba) && PyErr_Occurred())) __PYX_ERR(19, 13, __pyx_L7_error)

          /* "src/tests/test_buffer.pxi":12
 *         assert buffer.read_u8() == 3
 *         import pytest
 *         with pytest.raises(EOFError):             # <<<<<<<<<<<<<<
 *             buffer.read_u8()
 * 
 */
        }
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("cavro._tests._test_reader_read_u8", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(19, 12, __pyx_L9_except_error)
          __Pyx_XGOTREF(__pyx_t_1);
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_t_5);
          __pyx_t_8 = PyTuple_Pack(3, __pyx_t_1, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(19, 12, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_8, NULL);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(19, 12, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (__pyx_t_3 < 0) __PYX_ERR(19, 12, __pyx_L9_except_error)
          __pyx_t_13 = (!__pyx_t_3);
          if (unlikely(__pyx_t_13)) {
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_XGIVEREF(__pyx_t_5);
            __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_4, __pyx_t_5);
            __pyx_t_1 = 0; __pyx_t_4 = 0; __pyx_t_5 = 0; 
            __PYX_ERR(19, 12, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_7) {
          __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__36, NULL);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(19, 12, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L16;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L1_error;
    __pyx_L16:;
  }

  /* "src/tests/test_buffer.pxi":5
 * def _tests(add):
 * 
 *     @add             # <<<<<<<<<<<<<<
 *     def _test_reader_read_u8():
 *         buffer = MemoryReader(b'\x01\x02\x03')
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cavro._tests._test_reader_read_u8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_buffer);
  __Pyx_XDECREF(__pyx_v_pytest);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_buffer.pxi":15
 *             buffer.read_u8()
 * 
 *     @add             # <<<<<<<<<<<<<<
 *     def _test_reader_read_n():
 *         buffer = MemoryReader(b'abacuscounter1')
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6_tests_11_test_reader_read_n(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_6_tests_11_test_reader_read_n = {"_test_reader_read_n", (PyCFunction)__pyx_pw_5cavro_6_tests_11_test_reader_read_n, METH_NOARGS, 0};
static PyObject *__pyx_pw_5cavro_6_tests_11_test_reader_read_n(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_reader_read_n (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6_tests_10_test_reader_read_n(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6_tests_10_test_reader_read_n(CYTHON_UNUSED PyObject *__pyx_self) {
  struct __pyx_obj_5cavro_MemoryReader *__pyx_v_buffer = NULL;
  PyObject *__pyx_v_a = NULL;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_v_pytest = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_reader_read_n", 1);

  /* "src/tests/test_buffer.pxi":17
 *     @add
 *     def _test_reader_read_n():
 *         buffer = MemoryReader(b'abacuscounter1')             # <<<<<<<<<<<<<<
 *         a = bytes(buffer.read_n(6)[:6])
 *         assert a == b'abacus', a
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_MemoryReader), __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buffer = ((struct __pyx_obj_5cavro_MemoryReader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/tests/test_buffer.pxi":18
 *     def _test_reader_read_n():
 *         buffer = MemoryReader(b'abacuscounter1')
 *         a = bytes(buffer.read_n(6)[:6])             # <<<<<<<<<<<<<<
 *         assert a == b'abacus', a
 *         b = bytes(buffer.read_n(7)[:7])
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_MemoryReader *)__pyx_v_buffer->__pyx_base.__pyx_vtab)->__pyx_base.read_n(((struct __pyx_obj_5cavro__Reader *)__pyx_v_buffer), 6); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(19, 18, __pyx_L1_error)
  __pyx_t_3.data = __pyx_t_2.data;
  __pyx_t_3.memview = __pyx_t_2.memview;
  __PYX_INC_MEMVIEW(&__pyx_t_3, 1);
  __pyx_t_4 = -1;
  if (unlikely(__pyx_memoryview_slice_memviewslice(
    &__pyx_t_3,
    __pyx_t_2.shape[0], __pyx_t_2.strides[0], __pyx_t_2.suboffsets[0],
    0,
    0,
    &__pyx_t_4,
    0,
    6,
    0,
    0,
    1,
    0,
    1) < 0))
{
    __PYX_ERR(19, 18, __pyx_L1_error)
}

__PYX_XCLEAR_MEMVIEW(&__pyx_t_2, 1);
  __pyx_t_2.memview = NULL; __pyx_t_2.data = NULL;
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_t_3, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_3, 1);
  __pyx_t_3.memview = NULL; __pyx_t_3.data = NULL;
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(19, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_a = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "src/tests/test_buffer.pxi":19
 *         buffer = MemoryReader(b'abacuscounter1')
 *         a = bytes(buffer.read_n(6)[:6])
 *         assert a == b'abacus', a             # <<<<<<<<<<<<<<
 *         b = bytes(buffer.read_n(7)[:7])
 *         assert b == b'counter', b
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_6 = (__Pyx_PyBytes_Equals(__pyx_v_a, __pyx_n_b_abacus, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(19, 19, __pyx_L1_error)
    if (unlikely(!__pyx_t_6)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_v_a, 0, 0);
      __PYX_ERR(19, 19, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(19, 19, __pyx_L1_error)
  #endif

  /* "src/tests/test_buffer.pxi":20
 *         a = bytes(buffer.read_n(6)[:6])
 *         assert a == b'abacus', a
 *         b = bytes(buffer.read_n(7)[:7])             # <<<<<<<<<<<<<<
 *         assert b == b'counter', b
 *         import pytest
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_MemoryReader *)__pyx_v_buffer->__pyx_base.__pyx_vtab)->__pyx_base.read_n(((struct __pyx_obj_5cavro__Reader *)__pyx_v_buffer), 7); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(19, 20, __pyx_L1_error)
  __pyx_t_2.data = __pyx_t_3.data;
  __pyx_t_2.memview = __pyx_t_3.memview;
  __PYX_INC_MEMVIEW(&__pyx_t_2, 1);
  __pyx_t_4 = -1;
  if (unlikely(__pyx_memoryview_slice_memviewslice(
    &__pyx_t_2,
    __pyx_t_3.shape[0], __pyx_t_3.strides[0], __pyx_t_3.suboffsets[0],
    0,
    0,
    &__pyx_t_4,
    0,
    7,
    0,
    0,
    1,
    0,
    1) < 0))
{
    __PYX_ERR(19, 20, __pyx_L1_error)
}

__PYX_XCLEAR_MEMVIEW(&__pyx_t_3, 1);
  __pyx_t_3.memview = NULL; __pyx_t_3.data = NULL;
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_t_2, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(19, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_2, 1);
  __pyx_t_2.memview = NULL; __pyx_t_2.data = NULL;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_b = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/tests/test_buffer.pxi":21
 *         assert a == b'abacus', a
 *         b = bytes(buffer.read_n(7)[:7])
 *         assert b == b'counter', b             # <<<<<<<<<<<<<<
 *         import pytest
 *         with pytest.raises(EOFError):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_6 = (__Pyx_PyBytes_Equals(__pyx_v_b, __pyx_n_b_counter, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(19, 21, __pyx_L1_error)
    if (unlikely(!__pyx_t_6)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_v_b, 0, 0);
      __PYX_ERR(19, 21, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(19, 21, __pyx_L1_error)
  #endif

  /* "src/tests/test_buffer.pxi":22
 *         b = bytes(buffer.read_n(7)[:7])
 *         assert b == b'counter', b
 *         import pytest             # <<<<<<<<<<<<<<
 *         with pytest.raises(EOFError):
 *             buffer.read_n(2)
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_pytest, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pytest = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/tests/test_buffer.pxi":23
 *         assert b == b'counter', b
 *         import pytest
 *         with pytest.raises(EOFError):             # <<<<<<<<<<<<<<
 *             buffer.read_n(2)
 */
  /*with:*/ {
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_pytest, __pyx_n_s_raises); if (unlikely(!__pyx_t_5)) __PYX_ERR(19, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_builtin_EOFError};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 23, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_8 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_8)) __PYX_ERR(19, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_7)) __PYX_ERR(19, 23, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(19, 23, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        /*try:*/ {

          /* "src/tests/test_buffer.pxi":24
 *         import pytest
 *         with pytest.raises(EOFError):
 *             buffer.read_n(2)             # <<<<<<<<<<<<<<
 */
          __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_MemoryReader *)__pyx_v_buffer->__pyx_base.__pyx_vtab)->__pyx_base.read_n(((struct __pyx_obj_5cavro__Reader *)__pyx_v_buffer), 2); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(19, 24, __pyx_L7_error)
          __PYX_XCLEAR_MEMVIEW(&__pyx_t_2, 1);
          __pyx_t_2.memview = NULL; __pyx_t_2.data = NULL;

          /* "src/tests/test_buffer.pxi":23
 *         assert b == b'counter', b
 *         import pytest
 *         with pytest.raises(EOFError):             # <<<<<<<<<<<<<<
 *             buffer.read_n(2)
 */
        }
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_XCLEAR_MEMVIEW(&__pyx_t_2, 1);
        __pyx_t_2.memview = NULL; __pyx_t_2.data = NULL;
        __PYX_XCLEAR_MEMVIEW(&__pyx_t_3, 1);
        __pyx_t_3.memview = NULL; __pyx_t_3.data = NULL;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("cavro._tests._test_reader_read_n", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_5, &__pyx_t_7) < 0) __PYX_ERR(19, 23, __pyx_L9_except_error)
          __Pyx_XGOTREF(__pyx_t_1);
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_7);
          __pyx_t_9 = PyTuple_Pack(3, __pyx_t_1, __pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(19, 23, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, NULL);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(19, 23, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (__pyx_t_6 < 0) __PYX_ERR(19, 23, __pyx_L9_except_error)
          __pyx_t_14 = (!__pyx_t_6);
          if (unlikely(__pyx_t_14)) {
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_5);
            __Pyx_XGIVEREF(__pyx_t_7);
            __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_5, __pyx_t_7);
            __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_7 = 0; 
            __PYX_ERR(19, 23, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_8) {
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__36, NULL);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(19, 23, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L16;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L1_error;
    __pyx_L16:;
  }

  /* "src/tests/test_buffer.pxi":15
 *             buffer.read_u8()
 * 
 *     @add             # <<<<<<<<<<<<<<
 *     def _test_reader_read_n():
 *         buffer = MemoryReader(b'abacuscounter1')
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cavro._tests._test_reader_read_n", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_buffer);
  __Pyx_XDECREF(__pyx_v_a);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XDECREF(__pyx_v_pytest);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_buffer.pxi":2
 * 
 * @_tests             # <<<<<<<<<<<<<<
 * def _tests(add):
 * 
 */

static PyObject *__pyx_pf_5cavro_10_tests(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_add) {
  CYTHON_UNUSED PyObject *__pyx_v__test_reader_read_u8 = 0;
  CYTHON_UNUSED PyObject *__pyx_v__test_reader_read_n = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_tests", 1);

  /* "src/tests/test_buffer.pxi":5
 * def _tests(add):
 * 
 *     @add             # <<<<<<<<<<<<<<
 *     def _test_reader_read_u8():
 *         buffer = MemoryReader(b'\x01\x02\x03')
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6_tests_9_test_reader_read_u8, 0, __pyx_n_s_tests_locals__test_reader_read, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_2)) __PYX_ERR(19, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_add);
  __pyx_t_3 = __pyx_v_add; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v__test_reader_read_u8 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/tests/test_buffer.pxi":15
 *             buffer.read_u8()
 * 
 *     @add             # <<<<<<<<<<<<<<
 *     def _test_reader_read_n():
 *         buffer = MemoryReader(b'abacuscounter1')
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6_tests_11_test_reader_read_n, 0, __pyx_n_s_tests_locals__test_reader_read_2, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_3)) __PYX_ERR(19, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_add);
  __pyx_t_2 = __pyx_v_add; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v__test_reader_read_n = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/tests/test_buffer.pxi":2
 * 
 * @_tests             # <<<<<<<<<<<<<<
 * def _tests(add):
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._tests", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__test_reader_read_u8);
  __Pyx_XDECREF(__pyx_v__test_reader_read_n);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_perf.pxi":2
 * 
 * @_perf             # <<<<<<<<<<<<<<
 * def _perf(add):
 *     def _test_zigzag_perf(const uint64_t mask):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_13_perf(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_13_perf = {"_perf", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13_perf, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_13_perf(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_add = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_perf (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_add,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_add)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(7, 2, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_perf") < 0)) __PYX_ERR(7, 2, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_add = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_perf", 1, 1, 1, __pyx_nargs); __PYX_ERR(7, 2, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._perf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_12_perf(__pyx_self, __pyx_v_add);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_perf.pxi":4
 * @_perf
 * def _perf(add):
 *     def _test_zigzag_perf(const uint64_t mask):             # <<<<<<<<<<<<<<
 *         from time import perf_counter_ns
 *         import os
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_5_perf_3_test_zigzag_perf(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_5_perf_3_test_zigzag_perf = {"_test_zigzag_perf", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5_perf_3_test_zigzag_perf, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_5_perf_3_test_zigzag_perf(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  uint64_t __pyx_v_mask;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_zigzag_perf (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mask,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mask)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(7, 4, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_test_zigzag_perf") < 0)) __PYX_ERR(7, 4, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_mask = __Pyx_PyInt_As_uint64_t(values[0]); if (unlikely((__pyx_v_mask == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(7, 4, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_test_zigzag_perf", 1, 1, 1, __pyx_nargs); __PYX_ERR(7, 4, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._perf._test_zigzag_perf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_5_perf_2_test_zigzag_perf(__pyx_self, __pyx_v_mask);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_5_perf_2_test_zigzag_perf(CYTHON_UNUSED PyObject *__pyx_self, uint64_t __pyx_v_mask) {
  PyObject *__pyx_v_perf_counter_ns = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_os = NULL;
  struct __pyx_t_5cavro_random_t __pyx_v_state;
  uint64_t __pyx_v_BATCH_SIZE;
  uint64_t __pyx_v_given;
  struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_writer = 0;
  struct __pyx_obj_5cavro_MemoryReader *__pyx_v_reader = 0;
  size_t __pyx_v_total_nums;
  CYTHON_UNUSED size_t __pyx_v_n;
  CYTHON_UNUSED uint64_t __pyx_v_i;
  uint64_t __pyx_v_out_state;
  PyObject *__pyx_v_t_a = NULL;
  PyObject *__pyx_v_write_time = NULL;
  PyObject *__pyx_v_copy_time = NULL;
  PyObject *__pyx_v_read_time = NULL;
  PyObject *__pyx_v_write_perf = NULL;
  PyObject *__pyx_v_read_perf = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  uint64_t __pyx_t_5;
  uint64_t __pyx_t_6;
  uint64_t __pyx_t_7;
  uint32_t __pyx_t_8;
  uint32_t __pyx_t_9;
  size_t __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  int64_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_UCS4 __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_zigzag_perf", 1);

  /* "src/tests/test_perf.pxi":5
 * def _perf(add):
 *     def _test_zigzag_perf(const uint64_t mask):
 *         from time import perf_counter_ns             # <<<<<<<<<<<<<<
 *         import os
 *         cdef random_t state
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_perf_counter_ns);
  __Pyx_GIVEREF(__pyx_n_s_perf_counter_ns);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_perf_counter_ns)) __PYX_ERR(7, 5, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_time, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_perf_counter_ns); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_perf_counter_ns = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/tests/test_perf.pxi":6
 *     def _test_zigzag_perf(const uint64_t mask):
 *         from time import perf_counter_ns
 *         import os             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_os, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_os = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "src/tests/test_perf.pxi":9
 *         cdef random_t state
 * 
 *         cdef uint64_t BATCH_SIZE = 5_000_000             # <<<<<<<<<<<<<<
 * 
 *         cdef uint64_t actual
 */
  __pyx_v_BATCH_SIZE = 0x4C4B40;

  /* "src/tests/test_perf.pxi":14
 *         cdef uint64_t given
 *         cdef bytes src
 *         cdef MemoryWriter writer = MemoryWriter(3442987655)             # <<<<<<<<<<<<<<
 *         cdef MemoryReader reader
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_MemoryWriter), __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_writer = ((struct __pyx_obj_5cavro_MemoryWriter *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/tests/test_perf.pxi":17
 *         cdef MemoryReader reader
 * 
 *         cdef size_t total_nums = 0             # <<<<<<<<<<<<<<
 *         cdef size_t n
 *         cdef uint64_t i
 */
  __pyx_v_total_nums = 0;

  /* "src/tests/test_perf.pxi":20
 *         cdef size_t n
 *         cdef uint64_t i
 *         cdef uint64_t out_state = 0             # <<<<<<<<<<<<<<
 *         t_a = perf_counter_ns()
 *         for i in range(BATCH_SIZE):
 */
  __pyx_v_out_state = 0;

  /* "src/tests/test_perf.pxi":21
 *         cdef uint64_t i
 *         cdef uint64_t out_state = 0
 *         t_a = perf_counter_ns()             # <<<<<<<<<<<<<<
 *         for i in range(BATCH_SIZE):
 *             given = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask
 */
  __Pyx_INCREF(__pyx_v_perf_counter_ns);
  __pyx_t_1 = __pyx_v_perf_counter_ns; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 21, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_v_t_a = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "src/tests/test_perf.pxi":22
 *         cdef uint64_t out_state = 0
 *         t_a = perf_counter_ns()
 *         for i in range(BATCH_SIZE):             # <<<<<<<<<<<<<<
 *             given = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask
 *             zigzag_encode_long(writer, given)
 */
  __pyx_t_5 = __pyx_v_BATCH_SIZE;
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "src/tests/test_perf.pxi":23
 *         t_a = perf_counter_ns()
 *         for i in range(BATCH_SIZE):
 *             given = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask             # <<<<<<<<<<<<<<
 *             zigzag_encode_long(writer, given)
 *             total_nums += 1
 */
    __pyx_t_8 = __pyx_f_5cavro_rand((&__pyx_v_state)); if (unlikely(__pyx_t_8 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(7, 23, __pyx_L1_error)
    __pyx_t_9 = __pyx_f_5cavro_rand((&__pyx_v_state)); if (unlikely(__pyx_t_9 == ((uint32_t)-1) && PyErr_Occurred())) __PYX_ERR(7, 23, __pyx_L1_error)
    __pyx_v_given = ((((uint64_t)__pyx_t_8) | (((uint64_t)__pyx_t_9) << 32)) & __pyx_v_mask);

    /* "src/tests/test_perf.pxi":24
 *         for i in range(BATCH_SIZE):
 *             given = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask
 *             zigzag_encode_long(writer, given)             # <<<<<<<<<<<<<<
 *             total_nums += 1
 *         write_time = perf_counter_ns() - t_a
 */
    __pyx_t_4 = __pyx_f_5cavro_zigzag_encode_long(((struct __pyx_obj_5cavro__Writer *)__pyx_v_writer), __pyx_v_given); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(7, 24, __pyx_L1_error)

    /* "src/tests/test_perf.pxi":25
 *             given = (<uint64_t>rand(&state) | (<uint64_t>rand(&state) << 32)) & mask
 *             zigzag_encode_long(writer, given)
 *             total_nums += 1             # <<<<<<<<<<<<<<
 *         write_time = perf_counter_ns() - t_a
 *         t_a = perf_counter_ns()
 */
    __pyx_v_total_nums = (__pyx_v_total_nums + 1);
  }

  /* "src/tests/test_perf.pxi":26
 *             zigzag_encode_long(writer, given)
 *             total_nums += 1
 *         write_time = perf_counter_ns() - t_a             # <<<<<<<<<<<<<<
 *         t_a = perf_counter_ns()
 *         reader = MemoryReader(writer.bytes())
 */
  __Pyx_INCREF(__pyx_v_perf_counter_ns);
  __pyx_t_1 = __pyx_v_perf_counter_ns; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 26, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_1 = PyNumber_Subtract(__pyx_t_2, __pyx_v_t_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_write_time = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/tests/test_perf.pxi":27
 *             total_nums += 1
 *         write_time = perf_counter_ns() - t_a
 *         t_a = perf_counter_ns()             # <<<<<<<<<<<<<<
 *         reader = MemoryReader(writer.bytes())
 *         copy_time = perf_counter_ns() - t_a
 */
  __Pyx_INCREF(__pyx_v_perf_counter_ns);
  __pyx_t_2 = __pyx_v_perf_counter_ns; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 27, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF_SET(__pyx_v_t_a, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/tests/test_perf.pxi":28
 *         write_time = perf_counter_ns() - t_a
 *         t_a = perf_counter_ns()
 *         reader = MemoryReader(writer.bytes())             # <<<<<<<<<<<<<<
 *         copy_time = perf_counter_ns() - t_a
 *         t_a = perf_counter_ns()
 */
  __pyx_t_1 = __pyx_f_5cavro_12MemoryWriter_bytes(__pyx_v_writer); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro_MemoryReader), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_reader = ((struct __pyx_obj_5cavro_MemoryReader *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/tests/test_perf.pxi":29
 *         t_a = perf_counter_ns()
 *         reader = MemoryReader(writer.bytes())
 *         copy_time = perf_counter_ns() - t_a             # <<<<<<<<<<<<<<
 *         t_a = perf_counter_ns()
 *         for n in range(total_nums):
 */
  __Pyx_INCREF(__pyx_v_perf_counter_ns);
  __pyx_t_1 = __pyx_v_perf_counter_ns; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_1 = PyNumber_Subtract(__pyx_t_2, __pyx_v_t_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_copy_time = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/tests/test_perf.pxi":30
 *         reader = MemoryReader(writer.bytes())
 *         copy_time = perf_counter_ns() - t_a
 *         t_a = perf_counter_ns()             # <<<<<<<<<<<<<<
 *         for n in range(total_nums):
 *             out_state ^= zigzag_decode_long(reader)
 */
  __Pyx_INCREF(__pyx_v_perf_counter_ns);
  __pyx_t_2 = __pyx_v_perf_counter_ns; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF_SET(__pyx_v_t_a, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/tests/test_perf.pxi":31
 *         copy_time = perf_counter_ns() - t_a
 *         t_a = perf_counter_ns()
 *         for n in range(total_nums):             # <<<<<<<<<<<<<<
 *             out_state ^= zigzag_decode_long(reader)
 *         read_time = perf_counter_ns() - t_a
 */
  __pyx_t_10 = __pyx_v_total_nums;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_n = __pyx_t_12;

    /* "src/tests/test_perf.pxi":32
 *         t_a = perf_counter_ns()
 *         for n in range(total_nums):
 *             out_state ^= zigzag_decode_long(reader)             # <<<<<<<<<<<<<<
 *         read_time = perf_counter_ns() - t_a
 * 
 */
    __pyx_t_13 = __pyx_f_5cavro_zigzag_decode_long(((struct __pyx_obj_5cavro__Reader *)__pyx_v_reader)); if (unlikely(__pyx_t_13 == ((int64_t)0x7ffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(7, 32, __pyx_L1_error)
    __pyx_v_out_state = (__pyx_v_out_state ^ __pyx_t_13);
  }

  /* "src/tests/test_perf.pxi":33
 *         for n in range(total_nums):
 *             out_state ^= zigzag_decode_long(reader)
 *         read_time = perf_counter_ns() - t_a             # <<<<<<<<<<<<<<
 * 
 *         write_perf = (write_time) / BATCH_SIZE
 */
  __Pyx_INCREF(__pyx_v_perf_counter_ns);
  __pyx_t_2 = __pyx_v_perf_counter_ns; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyNumber_Subtract(__pyx_t_1, __pyx_v_t_a); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_read_time = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "src/tests/test_perf.pxi":35
 *         read_time = perf_counter_ns() - t_a
 * 
 *         write_perf = (write_time) / BATCH_SIZE             # <<<<<<<<<<<<<<
 *         read_perf = (read_time) / BATCH_SIZE
 *         print(f"{hex(mask)}: {out_state}: write: {write_perf} ns, copy: {copy_time} s, read: {read_perf} ns")
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint64_t(__pyx_v_BATCH_SIZE); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_v_write_time, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_write_perf = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/tests/test_perf.pxi":36
 * 
 *         write_perf = (write_time) / BATCH_SIZE
 *         read_perf = (read_time) / BATCH_SIZE             # <<<<<<<<<<<<<<
 *         print(f"{hex(mask)}: {out_state}: write: {write_perf} ns, copy: {copy_time} s, read: {read_perf} ns")
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_BATCH_SIZE); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_v_read_time, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_read_perf = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "src/tests/test_perf.pxi":37
 *         write_perf = (write_time) / BATCH_SIZE
 *         read_perf = (read_time) / BATCH_SIZE
 *         print(f"{hex(mask)}: {out_state}: write: {write_perf} ns, copy: {copy_time} s, read: {read_perf} ns")             # <<<<<<<<<<<<<<
 * 
 *     for mask in [
 */
  __pyx_t_2 = PyTuple_New(10); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_14 = 0;
  __pyx_t_15 = 127;
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_hex, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_15 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_15) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_15;
  __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_14 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_kp_u_);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_out_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_15 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_15) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_15;
  __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u_write_2);
  __pyx_t_14 += 9;
  __Pyx_GIVEREF(__pyx_kp_u_write_2);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_kp_u_write_2);
  __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_write_perf, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_15 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_15) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_15;
  __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u_ns_copy);
  __pyx_t_14 += 11;
  __Pyx_GIVEREF(__pyx_kp_u_ns_copy);
  PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_kp_u_ns_copy);
  __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_copy_time, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_15 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_15) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_15;
  __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u_s_read);
  __pyx_t_14 += 10;
  __Pyx_GIVEREF(__pyx_kp_u_s_read);
  PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_kp_u_s_read);
  __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_read_perf, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_15 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_15) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_15;
  __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u_ns);
  __pyx_t_14 += 3;
  __Pyx_GIVEREF(__pyx_kp_u_ns);
  PyTuple_SET_ITEM(__pyx_t_2, 9, __pyx_kp_u_ns);
  __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_2, 10, __pyx_t_14, __pyx_t_15); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/tests/test_perf.pxi":4
 * @_perf
 * def _perf(add):
 *     def _test_zigzag_perf(const uint64_t mask):             # <<<<<<<<<<<<<<
 *         from time import perf_counter_ns
 *         import os
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro._perf._test_zigzag_perf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_perf_counter_ns);
  __Pyx_XDECREF(__pyx_v_os);
  __Pyx_XDECREF((PyObject *)__pyx_v_writer);
  __Pyx_XDECREF((PyObject *)__pyx_v_reader);
  __Pyx_XDECREF(__pyx_v_t_a);
  __Pyx_XDECREF(__pyx_v_write_time);
  __Pyx_XDECREF(__pyx_v_copy_time);
  __Pyx_XDECREF(__pyx_v_read_time);
  __Pyx_XDECREF(__pyx_v_write_perf);
  __Pyx_XDECREF(__pyx_v_read_perf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/tests/test_perf.pxi":2
 * 
 * @_perf             # <<<<<<<<<<<<<<
 * def _perf(add):
 *     def _test_zigzag_perf(const uint64_t mask):
 */

static PyObject *__pyx_pf_5cavro_12_perf(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_add) {
  PyObject *__pyx_v__test_zigzag_perf = 0;
  PyObject *__pyx_v_mask = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_perf", 1);

  /* "src/tests/test_perf.pxi":4
 * @_perf
 * def _perf(add):
 *     def _test_zigzag_perf(const uint64_t mask):             # <<<<<<<<<<<<<<
 *         from time import perf_counter_ns
 *         import os
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_5_perf_3_test_zigzag_perf, 0, __pyx_n_s_perf_locals__test_zigzag_perf, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__test_zigzag_perf = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/tests/test_perf.pxi":40
 * 
 *     for mask in [
 *         0x7f, 0x7fff, 0x7fffff, 0x7fffffff, UINT32_MAX,             # <<<<<<<<<<<<<<
 *         0x7ffffffffful, 0x7ffffffffffful, 0x7ffffffffffffful,
 *         0x7ffffffffffffffful, UINT64_MAX
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(UINT32_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/tests/test_perf.pxi":42
 *         0x7f, 0x7fff, 0x7fffff, 0x7fffffff, UINT32_MAX,
 *         0x7ffffffffful, 0x7ffffffffffful, 0x7ffffffffffffful,
 *         0x7ffffffffffffffful, UINT64_MAX             # <<<<<<<<<<<<<<
 *         ]:
 *         add(_test_zigzag_perf, mask)
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint64_t(UINT64_MAX); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "src/tests/test_perf.pxi":39
 *         print(f"{hex(mask)}: {out_state}: write: {write_perf} ns, copy: {copy_time} s, read: {read_perf} ns")
 * 
 *     for mask in [             # <<<<<<<<<<<<<<
 *         0x7f, 0x7fff, 0x7fffff, 0x7fffffff, UINT32_MAX,
 *         0x7ffffffffful, 0x7ffffffffffful, 0x7ffffffffffffful,
 */
  __pyx_t_3 = PyTuple_New(10); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_int_127);
  __Pyx_GIVEREF(__pyx_int_127);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_int_127)) __PYX_ERR(7, 39, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_32767);
  __Pyx_GIVEREF(__pyx_int_32767);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_32767)) __PYX_ERR(7, 39, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_8388607);
  __Pyx_GIVEREF(__pyx_int_8388607);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_int_8388607)) __PYX_ERR(7, 39, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2147483647);
  __Pyx_GIVEREF(__pyx_int_2147483647);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_int_2147483647)) __PYX_ERR(7, 39, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_t_1)) __PYX_ERR(7, 39, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_549755813887L);
  __Pyx_GIVEREF(__pyx_int_549755813887L);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_int_549755813887L)) __PYX_ERR(7, 39, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0x7fffffffffffL);
  __Pyx_GIVEREF(__pyx_int_0x7fffffffffffL);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_int_0x7fffffffffffL)) __PYX_ERR(7, 39, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0x7fffffffffffffL);
  __Pyx_GIVEREF(__pyx_int_0x7fffffffffffffL);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_int_0x7fffffffffffffL)) __PYX_ERR(7, 39, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0x7fffffffffffffffL);
  __Pyx_GIVEREF(__pyx_int_0x7fffffffffffffffL);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_int_0x7fffffffffffffffL)) __PYX_ERR(7, 39, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2)) __PYX_ERR(7, 39, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (__pyx_t_4 >= 10) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(7, 39, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_mask, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/tests/test_perf.pxi":44
 *         0x7ffffffffffffffful, UINT64_MAX
 *         ]:
 *         add(_test_zigzag_perf, mask)             # <<<<<<<<<<<<<<
 */
    __Pyx_INCREF(__pyx_v_add);
    __pyx_t_1 = __pyx_v_add; __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v__test_zigzag_perf, __pyx_v_mask};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 44, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "src/tests/test_perf.pxi":39
 *         print(f"{hex(mask)}: {out_state}: write: {write_perf} ns, copy: {copy_time} s, read: {read_perf} ns")
 * 
 *     for mask in [             # <<<<<<<<<<<<<<
 *         0x7f, 0x7fff, 0x7fffff, 0x7fffffff, UINT32_MAX,
 *         0x7ffffffffful, 0x7ffffffffffful, 0x7ffffffffffffful,
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/tests/test_perf.pxi":2
 * 
 * @_perf             # <<<<<<<<<<<<<<
 * def _perf(add):
 *     def _test_zigzag_perf(const uint64_t mask):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro._perf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__test_zigzag_perf);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":11
 *     cdef readonly AvroType item_type
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         super().__init__(schema, source, namespace)
 *         self.item_type = AvroType.for_source(schema, source['items'], namespace)
 */

/* Python wrapper */
static int __pyx_pw_5cavro_9ArrayType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_9ArrayType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_namespace = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_schema,&__pyx_n_s_source,&__pyx_n_s_namespace,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 11, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 11, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(20, 11, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_namespace)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 11, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(20, 11, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(20, 11, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_schema = values[0];
    __pyx_v_source = values[1];
    __pyx_v_namespace = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(20, 11, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ArrayType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9ArrayType___init__(((struct __pyx_obj_5cavro_ArrayType *)__pyx_v_self), __pyx_v_schema, __pyx_v_source, __pyx_v_namespace);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_9ArrayType___init__(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/array.pxi":12
 * 
 *     def __init__(self, schema, source, namespace):
 *         super().__init__(schema, source, namespace)             # <<<<<<<<<<<<<<
 *         self.item_type = AvroType.for_source(schema, source['items'], namespace)
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_ArrayType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_ArrayType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_5cavro_ArrayType))) __PYX_ERR(20, 12, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(20, 12, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_schema, __pyx_v_source, __pyx_v_namespace};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/array.pxi":13
 *     def __init__(self, schema, source, namespace):
 *         super().__init__(schema, source, namespace)
 *         self.item_type = AvroType.for_source(schema, source['items'], namespace)             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_AvroType), __pyx_n_s_for_source); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_source, __pyx_n_u_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_v_schema, __pyx_t_3, __pyx_v_namespace};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(20, 13, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->item_type);
  __Pyx_DECREF((PyObject *)__pyx_v_self->item_type);
  __pyx_v_self->item_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/array.pxi":11
 *     cdef readonly AvroType item_type
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         super().__init__(schema, source, namespace)
 *         self.item_type = AvroType.for_source(schema, source['items'], namespace)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.ArrayType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":15
 *         self.item_type = AvroType.for_source(schema, source['items'], namespace)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef ArrayType new_inst = self.clone_base()
 *         new_inst.item_type = self.item_type
 */

static PyObject *__pyx_pw_5cavro_9ArrayType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9ArrayType_copy(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_ArrayType *__pyx_v_new_inst = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "src/array.pxi":16
 * 
 *     cpdef AvroType copy(self):
 *         cdef ArrayType new_inst = self.clone_base()             # <<<<<<<<<<<<<<
 *         new_inst.item_type = self.item_type
 *         return new_inst
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_ArrayType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_ArrayType))))) __PYX_ERR(20, 16, __pyx_L1_error)
  __pyx_v_new_inst = ((struct __pyx_obj_5cavro_ArrayType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/array.pxi":17
 *     cpdef AvroType copy(self):
 *         cdef ArrayType new_inst = self.clone_base()
 *         new_inst.item_type = self.item_type             # <<<<<<<<<<<<<<
 *         return new_inst
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->item_type);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_new_inst->item_type);
  __Pyx_DECREF((PyObject *)__pyx_v_new_inst->item_type);
  __pyx_v_new_inst->item_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/array.pxi":18
 *         cdef ArrayType new_inst = self.clone_base()
 *         new_inst.item_type = self.item_type
 *         return new_inst             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_new_inst);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_new_inst);
  goto __pyx_L0;

  /* "src/array.pxi":15
 *         self.item_type = AvroType.for_source(schema, source['items'], namespace)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef ArrayType new_inst = self.clone_base()
 *         new_inst.item_type = self.item_type
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ArrayType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_new_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9ArrayType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9ArrayType_3copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9ArrayType_3copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9ArrayType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_9ArrayType_2copy(((struct __pyx_obj_5cavro_ArrayType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9ArrayType_2copy(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_9ArrayType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ArrayType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":20
 *         return new_inst
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type', 'items'})
 * 
 */

static PyObject *__pyx_f_5cavro_9ArrayType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/array.pxi":21
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {'type', 'items'})             # <<<<<<<<<<<<<<
 * 
 *     def walk_types(self, visited):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(20, 21, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_items) < 0) __PYX_ERR(20, 21, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/array.pxi":20
 *         return new_inst
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type', 'items'})
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.ArrayType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_9ArrayType_6generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/array.pxi":23
 *         return _strip_keys(dict(source), {'type', 'items'})
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9ArrayType_5walk_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9ArrayType_5walk_types = {"walk_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9ArrayType_5walk_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9ArrayType_5walk_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_visited = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("walk_types (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_visited,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_visited)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 23, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "walk_types") < 0)) __PYX_ERR(20, 23, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_visited = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("walk_types", 1, 1, 1, __pyx_nargs); __PYX_ERR(20, 23, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ArrayType.walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9ArrayType_4walk_types(((struct __pyx_obj_5cavro_ArrayType *)__pyx_v_self), __pyx_v_visited);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9ArrayType_4walk_types(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_visited) {
  struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("walk_types", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types *)__pyx_tp_new_5cavro___pyx_scope_struct_3_walk_types(__pyx_ptype_5cavro___pyx_scope_struct_3_walk_types, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(20, 23, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_visited = __pyx_v_visited;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_visited);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_visited);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_9ArrayType_6generator1, __pyx_codeobj__45, (PyObject *) __pyx_cur_scope, __pyx_n_s_walk_types, __pyx_n_s_ArrayType_walk_types, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(20, 23, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.ArrayType.walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_9ArrayType_6generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("walk_types", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L5_resume_from_yield_from;
    case 2: goto __pyx_L6_resume_from_yield_from;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(20, 23, __pyx_L1_error)

  /* "src/array.pxi":24
 * 
 *     def walk_types(self, visited):
 *         if self in visited:             # <<<<<<<<<<<<<<
 *             return
 *         yield from super().walk_types(visited)
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_visited, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(20, 24, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/array.pxi":25
 *     def walk_types(self, visited):
 *         if self in visited:
 *             return             # <<<<<<<<<<<<<<
 *         yield from super().walk_types(visited)
 *         yield from self.item_type.walk_types(visited)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "src/array.pxi":24
 * 
 *     def walk_types(self, visited):
 *         if self in visited:             # <<<<<<<<<<<<<<
 *             return
 *         yield from super().walk_types(visited)
 */
  }

  /* "src/array.pxi":26
 *         if self in visited:
 *             return
 *         yield from super().walk_types(visited)             # <<<<<<<<<<<<<<
 *         yield from self.item_type.walk_types(visited)
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_ArrayType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_ArrayType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_ptype_5cavro_ArrayType))) __PYX_ERR(20, 26, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_cur_scope->__pyx_v_self))) __PYX_ERR(20, 26, __pyx_L1_error);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_walk_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_visited};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 26, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __Pyx_Generator_Yield_From(__pyx_generator, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L5_resume_from_yield_from:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(20, 26, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(20, 26, __pyx_L1_error)
    }
  }

  /* "src/array.pxi":27
 *             return
 *         yield from super().walk_types(visited)
 *         yield from self.item_type.walk_types(visited)             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->item_type), __pyx_n_s_walk_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_visited};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 27, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __Pyx_Generator_Yield_From(__pyx_generator, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 2;
    return __pyx_r;
    __pyx_L6_resume_from_yield_from:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(20, 27, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(20, 27, __pyx_L1_error)
    }
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "src/array.pxi":23
 *         return _strip_keys(dict(source), {'type', 'items'})
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":29
 *         yield from self.item_type.walk_types(visited)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {'items': self.item_type.get_schema(created)}
 * 
 */

static PyObject *__pyx_pw_5cavro_9ArrayType_8_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_9ArrayType__get_schema_extra(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_get_schema __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);

  /* "src/array.pxi":30
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         return {'items': self.item_type.get_schema(created)}             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3.__pyx_n = 1;
  __pyx_t_3.created = __pyx_v_created;
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->item_type->__pyx_vtab)->get_schema(__pyx_v_self->item_type, 0, &__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_items, __pyx_t_2) < 0) __PYX_ERR(20, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/array.pxi":29
 *         yield from self.item_type.walk_types(visited)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {'items': self.item_type.get_schema(created)}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.ArrayType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9ArrayType_8_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9ArrayType_8_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9ArrayType_8_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9ArrayType_8_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 29, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(20, 29, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(20, 29, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ArrayType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(20, 29, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_9ArrayType_7_get_schema_extra(((struct __pyx_obj_5cavro_ArrayType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9ArrayType_7_get_schema_extra(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_9ArrayType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ArrayType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":32
 *         return {'items': self.item_type.get_schema(created)}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         if len(value):
 */

static int __pyx_f_5cavro_9ArrayType__binary_buffer_encode(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  size_t __pyx_v_idx;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_v_e = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 1);

  /* "src/array.pxi":33
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef size_t idx = 0             # <<<<<<<<<<<<<<
 *         if len(value):
 *             zigzag_encode_long(buffer, len(value))
 */
  __pyx_v_idx = 0;

  /* "src/array.pxi":34
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef size_t idx = 0
 *         if len(value):             # <<<<<<<<<<<<<<
 *             zigzag_encode_long(buffer, len(value))
 *             for item in value:
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(20, 34, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "src/array.pxi":35
 *         cdef size_t idx = 0
 *         if len(value):
 *             zigzag_encode_long(buffer, len(value))             # <<<<<<<<<<<<<<
 *             for item in value:
 *                 try:
 */
    __pyx_t_1 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(20, 35, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_5cavro_zigzag_encode_long(__pyx_v_buffer, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(20, 35, __pyx_L1_error)

    /* "src/array.pxi":36
 *         if len(value):
 *             zigzag_encode_long(buffer, len(value))
 *             for item in value:             # <<<<<<<<<<<<<<
 *                 try:
 *                     self.item_type.binary_buffer_encode(buffer, item)
 */
    if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
      __pyx_t_4 = __pyx_v_value; __Pyx_INCREF(__pyx_t_4);
      __pyx_t_1 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_1 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 36, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 36, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 36, __pyx_L1_error)
            #endif
            if (__pyx_t_1 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_1); __Pyx_INCREF(__pyx_t_6); __pyx_t_1++; if (unlikely((0 < 0))) __PYX_ERR(20, 36, __pyx_L1_error)
          #else
          __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_6)) __PYX_ERR(20, 36, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 36, __pyx_L1_error)
            #endif
            if (__pyx_t_1 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_1); __Pyx_INCREF(__pyx_t_6); __pyx_t_1++; if (unlikely((0 < 0))) __PYX_ERR(20, 36, __pyx_L1_error)
          #else
          __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_6)) __PYX_ERR(20, 36, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_5(__pyx_t_4);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(20, 36, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "src/array.pxi":37
 *             zigzag_encode_long(buffer, len(value))
 *             for item in value:
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.item_type.binary_buffer_encode(buffer, item)
 *                 except InvalidValue as e:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "src/array.pxi":38
 *             for item in value:
 *                 try:
 *                     self.item_type.binary_buffer_encode(buffer, item)             # <<<<<<<<<<<<<<
 *                 except InvalidValue as e:
 *                     if self.options.invalid_value_include_array_index:
 */
          __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->item_type->__pyx_vtab)->binary_buffer_encode(__pyx_v_self->item_type, __pyx_v_buffer, __pyx_v_item); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(20, 38, __pyx_L6_error)

          /* "src/array.pxi":37
 *             zigzag_encode_long(buffer, len(value))
 *             for item in value:
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.item_type.binary_buffer_encode(buffer, item)
 *                 except InvalidValue as e:
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L13_try_end;
        __pyx_L6_error:;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "src/array.pxi":39
 *                 try:
 *                     self.item_type.binary_buffer_encode(buffer, item)
 *                 except InvalidValue as e:             # <<<<<<<<<<<<<<
 *                     if self.options.invalid_value_include_array_index:
 *                         e.schema_path = (idx, ) + e.schema_path
 */
        __Pyx_ErrFetch(&__pyx_t_6, &__pyx_t_10, &__pyx_t_11);
        __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_12)) __PYX_ERR(20, 39, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_3 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_6, __pyx_t_12);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_ErrRestore(__pyx_t_6, __pyx_t_10, __pyx_t_11);
        __pyx_t_6 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __Pyx_AddTraceback("cavro.ArrayType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_11, &__pyx_t_10, &__pyx_t_6) < 0) __PYX_ERR(20, 39, __pyx_L8_except_error)
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_10);
          __pyx_v_e = __pyx_t_10;
          /*try:*/ {

            /* "src/array.pxi":40
 *                     self.item_type.binary_buffer_encode(buffer, item)
 *                 except InvalidValue as e:
 *                     if self.options.invalid_value_include_array_index:             # <<<<<<<<<<<<<<
 *                         e.schema_path = (idx, ) + e.schema_path
 *                     raise
 */
            if (__pyx_v_self->__pyx_base.options->invalid_value_include_array_index) {

              /* "src/array.pxi":41
 *                 except InvalidValue as e:
 *                     if self.options.invalid_value_include_array_index:
 *                         e.schema_path = (idx, ) + e.schema_path             # <<<<<<<<<<<<<<
 *                     raise
 *                 idx += 1
 */
              __pyx_t_12 = __Pyx_PyInt_FromSize_t(__pyx_v_idx); if (unlikely(!__pyx_t_12)) __PYX_ERR(20, 41, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_13 = PyTuple_New(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(20, 41, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_13);
              __Pyx_GIVEREF(__pyx_t_12);
              if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_12)) __PYX_ERR(20, 41, __pyx_L19_error);
              __pyx_t_12 = 0;
              __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_schema_path); if (unlikely(!__pyx_t_12)) __PYX_ERR(20, 41, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_14 = PyNumber_Add(__pyx_t_13, __pyx_t_12); if (unlikely(!__pyx_t_14)) __PYX_ERR(20, 41, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_e, __pyx_n_s_schema_path, __pyx_t_14) < 0) __PYX_ERR(20, 41, __pyx_L19_error)
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

              /* "src/array.pxi":40
 *                     self.item_type.binary_buffer_encode(buffer, item)
 *                 except InvalidValue as e:
 *                     if self.options.invalid_value_include_array_index:             # <<<<<<<<<<<<<<
 *                         e.schema_path = (idx, ) + e.schema_path
 *                     raise
 */
            }

            /* "src/array.pxi":42
 *                     if self.options.invalid_value_include_array_index:
 *                         e.schema_path = (idx, ) + e.schema_path
 *                     raise             # <<<<<<<<<<<<<<
 *                 idx += 1
 *         zigzag_encode_long(buffer, 0)
 */
            __Pyx_GIVEREF(__pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_6);
            __Pyx_ErrRestoreWithState(__pyx_t_11, __pyx_t_10, __pyx_t_6);
            __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_6 = 0; 
            __PYX_ERR(20, 42, __pyx_L19_error)
          }

          /* "src/array.pxi":39
 *                 try:
 *                     self.item_type.binary_buffer_encode(buffer, item)
 *                 except InvalidValue as e:             # <<<<<<<<<<<<<<
 *                     if self.options.invalid_value_include_array_index:
 *                         e.schema_path = (idx, ) + e.schema_path
 */
          /*finally:*/ {
            __pyx_L19_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_22);
              __pyx_t_3 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_XGIVEREF(__pyx_t_22);
                __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
              }
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
              __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
              __pyx_lineno = __pyx_t_3; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
              goto __pyx_L8_except_error;
            }
          }
        }
        goto __pyx_L8_except_error;

        /* "src/array.pxi":37
 *             zigzag_encode_long(buffer, len(value))
 *             for item in value:
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.item_type.binary_buffer_encode(buffer, item)
 *                 except InvalidValue as e:
 */
        __pyx_L8_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L13_try_end:;
      }

      /* "src/array.pxi":43
 *                         e.schema_path = (idx, ) + e.schema_path
 *                     raise
 *                 idx += 1             # <<<<<<<<<<<<<<
 *         zigzag_encode_long(buffer, 0)
 * 
 */
      __pyx_v_idx = (__pyx_v_idx + 1);

      /* "src/array.pxi":36
 *         if len(value):
 *             zigzag_encode_long(buffer, len(value))
 *             for item in value:             # <<<<<<<<<<<<<<
 *                 try:
 *                     self.item_type.binary_buffer_encode(buffer, item)
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "src/array.pxi":34
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef size_t idx = 0
 *         if len(value):             # <<<<<<<<<<<<<<
 *             zigzag_encode_long(buffer, len(value))
 *             for item in value:
 */
  }

  /* "src/array.pxi":44
 *                     raise
 *                 idx += 1
 *         zigzag_encode_long(buffer, 0)             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_t_15 = __pyx_f_5cavro_zigzag_encode_long(__pyx_v_buffer, 0); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(20, 44, __pyx_L1_error)

  /* "src/array.pxi":32
 *         return {'items': self.item_type.get_schema(created)}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         if len(value):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("cavro.ArrayType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":46
 *         zigzag_encode_long(buffer, 0)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef list out = []
 *         cdef size_t length
 */

static PyObject *__pyx_f_5cavro_9ArrayType__binary_buffer_decode(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  PyObject *__pyx_v_out = 0;
  size_t __pyx_v_length;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/array.pxi":47
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef list out = []             # <<<<<<<<<<<<<<
 *         cdef size_t length
 *         cdef str key
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/array.pxi":50
 *         cdef size_t length
 *         cdef str key
 *         while True:             # <<<<<<<<<<<<<<
 *             length = zigzag_decode_long(buffer)
 *             if length == 0:
 */
  while (1) {

    /* "src/array.pxi":51
 *         cdef str key
 *         while True:
 *             length = zigzag_decode_long(buffer)             # <<<<<<<<<<<<<<
 *             if length == 0:
 *                 return out
 */
    __pyx_t_2 = __pyx_f_5cavro_zigzag_decode_long(__pyx_v_buffer); if (unlikely(__pyx_t_2 == ((int64_t)0x7ffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(20, 51, __pyx_L1_error)
    __pyx_v_length = __pyx_t_2;

    /* "src/array.pxi":52
 *         while True:
 *             length = zigzag_decode_long(buffer)
 *             if length == 0:             # <<<<<<<<<<<<<<
 *                 return out
 *             while length:
 */
    __pyx_t_3 = (__pyx_v_length == 0);
    if (__pyx_t_3) {

      /* "src/array.pxi":53
 *             length = zigzag_decode_long(buffer)
 *             if length == 0:
 *                 return out             # <<<<<<<<<<<<<<
 *             while length:
 *                 value = self.item_type.binary_buffer_decode(buffer)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_out);
      __pyx_r = __pyx_v_out;
      goto __pyx_L0;

      /* "src/array.pxi":52
 *         while True:
 *             length = zigzag_decode_long(buffer)
 *             if length == 0:             # <<<<<<<<<<<<<<
 *                 return out
 *             while length:
 */
    }

    /* "src/array.pxi":54
 *             if length == 0:
 *                 return out
 *             while length:             # <<<<<<<<<<<<<<
 *                 value = self.item_type.binary_buffer_decode(buffer)
 *                 out.append(value)
 */
    while (1) {
      __pyx_t_3 = (__pyx_v_length != 0);
      if (!__pyx_t_3) break;

      /* "src/array.pxi":55
 *                 return out
 *             while length:
 *                 value = self.item_type.binary_buffer_decode(buffer)             # <<<<<<<<<<<<<<
 *                 out.append(value)
 *                 length -= 1
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->item_type->__pyx_vtab)->binary_buffer_decode(__pyx_v_self->item_type, __pyx_v_buffer); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 55, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "src/array.pxi":56
 *             while length:
 *                 value = self.item_type.binary_buffer_decode(buffer)
 *                 out.append(value)             # <<<<<<<<<<<<<<
 *                 length -= 1
 * 
 */
      __pyx_t_4 = __Pyx_PyList_Append(__pyx_v_out, __pyx_v_value); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(20, 56, __pyx_L1_error)

      /* "src/array.pxi":57
 *                 value = self.item_type.binary_buffer_decode(buffer)
 *                 out.append(value)
 *                 length -= 1             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
      __pyx_v_length = (__pyx_v_length - 1);
    }
  }

  /* "src/array.pxi":46
 *         zigzag_encode_long(buffer, 0)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef list out = []
 *         cdef size_t length
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ArrayType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":59
 *                 length -= 1
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return [self.item_type.json_format(item) for item in value]
 * 
 */

static PyObject *__pyx_f_5cavro_9ArrayType__json_format(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_8genexpr5__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 1);

  /* "src/array.pxi":60
 * 
 *     cdef _json_format(self, value):
 *         return [self.item_type.json_format(item) for item in value]             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 60, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
      __pyx_t_2 = __pyx_v_value; __Pyx_INCREF(__pyx_t_2);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 60, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 60, __pyx_L5_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 60, __pyx_L5_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(20, 60, __pyx_L5_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 60, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 60, __pyx_L5_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(20, 60, __pyx_L5_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 60, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_4(__pyx_t_2);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(20, 60, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_item, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->item_type->__pyx_vtab)->json_format(__pyx_v_self->item_type, __pyx_8genexpr5__pyx_v_item); if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 60, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(20, 60, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_item); __pyx_8genexpr5__pyx_v_item = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_item); __pyx_8genexpr5__pyx_v_item = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/array.pxi":59
 *                 length -= 1
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return [self.item_type.json_format(item) for item in value]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.ArrayType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":62
 *         return [self.item_type.json_format(item) for item in value]
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         return [self.item_type.json_decode(item) for item in value]
 * 
 */

static PyObject *__pyx_f_5cavro_9ArrayType__json_decode(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_8genexpr6__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/array.pxi":63
 * 
 *     cdef _json_decode(self, value):
 *         return [self.item_type.json_decode(item) for item in value]             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 63, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
      __pyx_t_2 = __pyx_v_value; __Pyx_INCREF(__pyx_t_2);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 63, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 63, __pyx_L5_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 63, __pyx_L5_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(20, 63, __pyx_L5_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 63, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 63, __pyx_L5_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(20, 63, __pyx_L5_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 63, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_4(__pyx_t_2);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(20, 63, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_item, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->item_type->__pyx_vtab)->json_decode(__pyx_v_self->item_type, __pyx_8genexpr6__pyx_v_item); if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 63, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(20, 63, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_item); __pyx_8genexpr6__pyx_v_item = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_item); __pyx_8genexpr6__pyx_v_item = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/array.pxi":62
 *         return [self.item_type.json_format(item) for item in value]
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         return [self.item_type.json_decode(item) for item in value]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.ArrayType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":65
 *         return [self.item_type.json_decode(item) for item in value]
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef int level = FIT_OK
 *         if isinstance(value, (list, tuple, numpy.ndarray)):
 */

static int __pyx_f_5cavro_9ArrayType__get_value_fitness(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_v_level;
  PyObject *__pyx_v_it = NULL;
  PyObject *__pyx_v_item = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 1);

  /* "src/array.pxi":66
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         cdef int level = FIT_OK             # <<<<<<<<<<<<<<
 *         if isinstance(value, (list, tuple, numpy.ndarray)):
 *             level = FIT_EXACT
 */
  __pyx_v_level = __pyx_v_5cavro_FIT_OK;

  /* "src/array.pxi":67
 *     cdef int _get_value_fitness(self, value) except -1:
 *         cdef int level = FIT_OK
 *         if isinstance(value, (list, tuple, numpy.ndarray)):             # <<<<<<<<<<<<<<
 *             level = FIT_EXACT
 *         elif isinstance(value, dict):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ndarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = PyList_Check(__pyx_v_value); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyTuple_Check(__pyx_v_value); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); 
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "src/array.pxi":68
 *         cdef int level = FIT_OK
 *         if isinstance(value, (list, tuple, numpy.ndarray)):
 *             level = FIT_EXACT             # <<<<<<<<<<<<<<
 *         elif isinstance(value, dict):
 *             level = FIT_POOR
 */
    __pyx_v_level = __pyx_v_5cavro_FIT_EXACT;

    /* "src/array.pxi":67
 *     cdef int _get_value_fitness(self, value) except -1:
 *         cdef int level = FIT_OK
 *         if isinstance(value, (list, tuple, numpy.ndarray)):             # <<<<<<<<<<<<<<
 *             level = FIT_EXACT
 *         elif isinstance(value, dict):
 */
    goto __pyx_L3;
  }

  /* "src/array.pxi":69
 *         if isinstance(value, (list, tuple, numpy.ndarray)):
 *             level = FIT_EXACT
 *         elif isinstance(value, dict):             # <<<<<<<<<<<<<<
 *             level = FIT_POOR
 *         elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
 */
  __pyx_t_3 = PyDict_Check(__pyx_v_value); 
  if (__pyx_t_3) {

    /* "src/array.pxi":70
 *             level = FIT_EXACT
 *         elif isinstance(value, dict):
 *             level = FIT_POOR             # <<<<<<<<<<<<<<
 *         elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
 *             return FIT_POOR
 */
    __pyx_v_level = __pyx_v_5cavro_FIT_POOR;

    /* "src/array.pxi":69
 *         if isinstance(value, (list, tuple, numpy.ndarray)):
 *             level = FIT_EXACT
 *         elif isinstance(value, dict):             # <<<<<<<<<<<<<<
 *             level = FIT_POOR
 *         elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
 */
    goto __pyx_L3;
  }

  /* "src/array.pxi":71
 *         elif isinstance(value, dict):
 *             level = FIT_POOR
 *         elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *             return FIT_POOR
 *         elif isinstance(value, (bytes, str)) or not isinstance(value, collections.abc.Iterable):
 */
  if (__pyx_v_self->__pyx_base.options->missing_values_can_be_empty_container) {
  } else {
    __pyx_t_3 = __pyx_v_self->__pyx_base.options->missing_values_can_be_empty_container;
    goto __pyx_L7_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_v_value == __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __pyx_t_4;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_3) {

    /* "src/array.pxi":72
 *             level = FIT_POOR
 *         elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
 *             return FIT_POOR             # <<<<<<<<<<<<<<
 *         elif isinstance(value, (bytes, str)) or not isinstance(value, collections.abc.Iterable):
 *             return FIT_NONE
 */
    __pyx_r = __pyx_v_5cavro_FIT_POOR;
    goto __pyx_L0;

    /* "src/array.pxi":71
 *         elif isinstance(value, dict):
 *             level = FIT_POOR
 *         elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *             return FIT_POOR
 *         elif isinstance(value, (bytes, str)) or not isinstance(value, collections.abc.Iterable):
 */
  }

  /* "src/array.pxi":73
 *         elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
 *             return FIT_POOR
 *         elif isinstance(value, (bytes, str)) or not isinstance(value, collections.abc.Iterable):             # <<<<<<<<<<<<<<
 *             return FIT_NONE
 *         try:
 */
  __pyx_t_5 = PyBytes_Check(__pyx_v_value); 
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_5 = PyUnicode_Check(__pyx_v_value); 
  __pyx_t_4 = __pyx_t_5;
  __pyx_L11_bool_binop_done:;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L9_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_collections); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_abc); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Iterable); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(20, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (!__pyx_t_4);
  __pyx_t_3 = __pyx_t_5;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_3) {

    /* "src/array.pxi":74
 *             return FIT_POOR
 *         elif isinstance(value, (bytes, str)) or not isinstance(value, collections.abc.Iterable):
 *             return FIT_NONE             # <<<<<<<<<<<<<<
 *         try:
 *             it = iter(value)
 */
    __pyx_r = __pyx_v_5cavro_FIT_NONE;
    goto __pyx_L0;

    /* "src/array.pxi":73
 *         elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
 *             return FIT_POOR
 *         elif isinstance(value, (bytes, str)) or not isinstance(value, collections.abc.Iterable):             # <<<<<<<<<<<<<<
 *             return FIT_NONE
 *         try:
 */
  }
  __pyx_L3:;

  /* "src/array.pxi":75
 *         elif isinstance(value, (bytes, str)) or not isinstance(value, collections.abc.Iterable):
 *             return FIT_NONE
 *         try:             # <<<<<<<<<<<<<<
 *             it = iter(value)
 *         except TypeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "src/array.pxi":76
 *             return FIT_NONE
 *         try:
 *             it = iter(value)             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             return FIT_NONE
 */
      __pyx_t_2 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 76, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_it = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "src/array.pxi":75
 *         elif isinstance(value, (bytes, str)) or not isinstance(value, collections.abc.Iterable):
 *             return FIT_NONE
 *         try:             # <<<<<<<<<<<<<<
 *             it = iter(value)
 *         except TypeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L18_try_end;
    __pyx_L13_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/array.pxi":77
 *         try:
 *             it = iter(value)
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             return FIT_NONE
 *         for item in it:
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("cavro.ArrayType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_1, &__pyx_t_10) < 0) __PYX_ERR(20, 77, __pyx_L15_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_10);

      /* "src/array.pxi":78
 *             it = iter(value)
 *         except TypeError:
 *             return FIT_NONE             # <<<<<<<<<<<<<<
 *         for item in it:
 *             level = min(self.item_type.get_value_fitness(item), level)
 */
      __pyx_r = __pyx_v_5cavro_FIT_NONE;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L16_except_return;
    }
    goto __pyx_L15_except_error;

    /* "src/array.pxi":75
 *         elif isinstance(value, (bytes, str)) or not isinstance(value, collections.abc.Iterable):
 *             return FIT_NONE
 *         try:             # <<<<<<<<<<<<<<
 *             it = iter(value)
 *         except TypeError:
 */
    __pyx_L15_except_error:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L16_except_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L18_try_end:;
  }

  /* "src/array.pxi":79
 *         except TypeError:
 *             return FIT_NONE
 *         for item in it:             # <<<<<<<<<<<<<<
 *             level = min(self.item_type.get_value_fitness(item), level)
 *             if level <= FIT_NONE:
 */
  if (likely(PyList_CheckExact(__pyx_v_it)) || PyTuple_CheckExact(__pyx_v_it)) {
    __pyx_t_10 = __pyx_v_it; __Pyx_INCREF(__pyx_t_10);
    __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_11 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_v_it); if (unlikely(!__pyx_t_10)) __PYX_ERR(20, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(20, 79, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_10))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_10);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 79, __pyx_L1_error)
          #endif
          if (__pyx_t_11 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_11); __Pyx_INCREF(__pyx_t_1); __pyx_t_11++; if (unlikely((0 < 0))) __PYX_ERR(20, 79, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_10, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 79, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_10);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 79, __pyx_L1_error)
          #endif
          if (__pyx_t_11 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_11); __Pyx_INCREF(__pyx_t_1); __pyx_t_11++; if (unlikely((0 < 0))) __PYX_ERR(20, 79, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_10, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 79, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_12(__pyx_t_10);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(20, 79, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/array.pxi":80
 *             return FIT_NONE
 *         for item in it:
 *             level = min(self.item_type.get_value_fitness(item), level)             # <<<<<<<<<<<<<<
 *             if level <= FIT_NONE:
 *                 break
 */
    __pyx_t_9 = __pyx_v_level;
    __pyx_t_13 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->item_type->__pyx_vtab)->get_value_fitness(__pyx_v_self->item_type, __pyx_v_item); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(20, 80, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_9 < __pyx_t_13);
    if (__pyx_t_3) {
      __pyx_t_14 = __pyx_t_9;
    } else {
      __pyx_t_14 = __pyx_t_13;
    }
    __pyx_v_level = __pyx_t_14;

    /* "src/array.pxi":81
 *         for item in it:
 *             level = min(self.item_type.get_value_fitness(item), level)
 *             if level <= FIT_NONE:             # <<<<<<<<<<<<<<
 *                 break
 *         return level
 */
    __pyx_t_3 = (__pyx_v_level <= __pyx_v_5cavro_FIT_NONE);
    if (__pyx_t_3) {

      /* "src/array.pxi":82
 *             level = min(self.item_type.get_value_fitness(item), level)
 *             if level <= FIT_NONE:
 *                 break             # <<<<<<<<<<<<<<
 *         return level
 * 
 */
      goto __pyx_L22_break;

      /* "src/array.pxi":81
 *         for item in it:
 *             level = min(self.item_type.get_value_fitness(item), level)
 *             if level <= FIT_NONE:             # <<<<<<<<<<<<<<
 *                 break
 *         return level
 */
    }

    /* "src/array.pxi":79
 *         except TypeError:
 *             return FIT_NONE
 *         for item in it:             # <<<<<<<<<<<<<<
 *             level = min(self.item_type.get_value_fitness(item), level)
 *             if level <= FIT_NONE:
 */
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  goto __pyx_L24_for_end;
  __pyx_L22_break:;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  goto __pyx_L24_for_end;
  __pyx_L24_for_end:;

  /* "src/array.pxi":83
 *             if level <= FIT_NONE:
 *                 break
 *         return level             # <<<<<<<<<<<<<<
 * 
 *     cdef list _make_converted_list(self, value):
 */
  __pyx_r = __pyx_v_level;
  goto __pyx_L0;

  /* "src/array.pxi":65
 *         return [self.item_type.json_decode(item) for item in value]
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef int level = FIT_OK
 *         if isinstance(value, (list, tuple, numpy.ndarray)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("cavro.ArrayType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_it);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":85
 *         return level
 * 
 *     cdef list _make_converted_list(self, value):             # <<<<<<<<<<<<<<
 *         cdef AvroType item_type = self.item_type
 *         cdef list out = []
 */

static PyObject *__pyx_f_5cavro_9ArrayType__make_converted_list(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_item_type = 0;
  PyObject *__pyx_v_out = 0;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_converted_list", 1);

  /* "src/array.pxi":86
 * 
 *     cdef list _make_converted_list(self, value):
 *         cdef AvroType item_type = self.item_type             # <<<<<<<<<<<<<<
 *         cdef list out = []
 *         for item in value:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->item_type);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_item_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/array.pxi":87
 *     cdef list _make_converted_list(self, value):
 *         cdef AvroType item_type = self.item_type
 *         cdef list out = []             # <<<<<<<<<<<<<<
 *         for item in value:
 *             out.append(item_type.convert_value(item))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/array.pxi":88
 *         cdef AvroType item_type = self.item_type
 *         cdef list out = []
 *         for item in value:             # <<<<<<<<<<<<<<
 *             out.append(item_type.convert_value(item))
 *         return out
 */
  if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
    __pyx_t_1 = __pyx_v_value; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 88, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 88, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(20, 88, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 88, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 88, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(20, 88, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 88, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(20, 88, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/array.pxi":89
 *         cdef list out = []
 *         for item in value:
 *             out.append(item_type.convert_value(item))             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_item_type->__pyx_vtab)->convert_value(__pyx_v_item_type, __pyx_v_item, 0, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(20, 89, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "src/array.pxi":88
 *         cdef AvroType item_type = self.item_type
 *         cdef list out = []
 *         for item in value:             # <<<<<<<<<<<<<<
 *             out.append(item_type.convert_value(item))
 *         return out
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/array.pxi":90
 *         for item in value:
 *             out.append(item_type.convert_value(item))
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cpdef object convert_value(self, object value, check_value=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "src/array.pxi":85
 *         return level
 * 
 *     cdef list _make_converted_list(self, value):             # <<<<<<<<<<<<<<
 *         cdef AvroType item_type = self.item_type
 *         cdef list out = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.ArrayType._make_converted_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_item_type);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":92
 *         return out
 * 
 *     cpdef object convert_value(self, object value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef int item_fitness
 *         cdef AvroType item_type = self.item_type
 */

static PyObject *__pyx_pw_5cavro_9ArrayType_10convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_9ArrayType_convert_value(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_9ArrayType_convert_value *__pyx_optional_args) {
  int __pyx_v_item_fitness;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_item_type = 0;
  PyObject *__pyx_v_it = NULL;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_value", 1);
  if (__pyx_optional_args) {
  }

  /* "src/array.pxi":94
 *     cpdef object convert_value(self, object value, check_value=True):
 *         cdef int item_fitness
 *         cdef AvroType item_type = self.item_type             # <<<<<<<<<<<<<<
 * 
 *         if self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->item_type);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_item_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/array.pxi":96
 *         cdef AvroType item_type = self.item_type
 * 
 *         if self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *             return []
 * 
 */
  if (__pyx_v_self->__pyx_base.options->missing_values_can_be_empty_container) {
  } else {
    __pyx_t_2 = __pyx_v_self->__pyx_base.options->missing_values_can_be_empty_container;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_value == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "src/array.pxi":97
 * 
 *         if self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
 *             return []             # <<<<<<<<<<<<<<
 * 
 *         it = iter(value)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/array.pxi":96
 *         cdef AvroType item_type = self.item_type
 * 
 *         if self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *             return []
 * 
 */
  }

  /* "src/array.pxi":99
 *             return []
 * 
 *         it = iter(value)             # <<<<<<<<<<<<<<
 *         for item in it:
 *             item_fitness = item_type.get_value_fitness(item)
 */
  __pyx_t_1 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_it = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/array.pxi":100
 * 
 *         it = iter(value)
 *         for item in it:             # <<<<<<<<<<<<<<
 *             item_fitness = item_type.get_value_fitness(item)
 *             if item_fitness == FIT_NONE:
 */
  if (likely(PyList_CheckExact(__pyx_v_it)) || PyTuple_CheckExact(__pyx_v_it)) {
    __pyx_t_1 = __pyx_v_it; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_it); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 100, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 100, __pyx_L1_error)
          #endif
          if (__pyx_t_4 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(20, 100, __pyx_L1_error)
        #else
        __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(20, 100, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 100, __pyx_L1_error)
          #endif
          if (__pyx_t_4 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(20, 100, __pyx_L1_error)
        #else
        __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(20, 100, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_5(__pyx_t_1);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(20, 100, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "src/array.pxi":101
 *         it = iter(value)
 *         for item in it:
 *             item_fitness = item_type.get_value_fitness(item)             # <<<<<<<<<<<<<<
 *             if item_fitness == FIT_NONE:
 *                 raise ValueError(f"'{item}' is not a valid value for array")
 */
    __pyx_t_7 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_item_type->__pyx_vtab)->get_value_fitness(__pyx_v_item_type, __pyx_v_item); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(20, 101, __pyx_L1_error)
    __pyx_v_item_fitness = __pyx_t_7;

    /* "src/array.pxi":102
 *         for item in it:
 *             item_fitness = item_type.get_value_fitness(item)
 *             if item_fitness == FIT_NONE:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"'{item}' is not a valid value for array")
 *             elif item_fitness < FIT_EXACT:
 */
    __pyx_t_2 = (__pyx_v_item_fitness == __pyx_v_5cavro_FIT_NONE);
    if (unlikely(__pyx_t_2)) {

      /* "src/array.pxi":103
 *             item_fitness = item_type.get_value_fitness(item)
 *             if item_fitness == FIT_NONE:
 *                 raise ValueError(f"'{item}' is not a valid value for array")             # <<<<<<<<<<<<<<
 *             elif item_fitness < FIT_EXACT:
 *                 return self._make_converted_list(iter(value))
 */
      __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(20, 103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = 0;
      __pyx_t_9 = 127;
      __Pyx_INCREF(__pyx_kp_u__6);
      __pyx_t_8 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__6);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_u__6);
      __pyx_t_10 = __Pyx_PyObject_FormatSimple(__pyx_v_item, __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(20, 103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_9;
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_INCREF(__pyx_kp_u_is_not_a_valid_value_for_array);
      __pyx_t_8 += 32;
      __Pyx_GIVEREF(__pyx_kp_u_is_not_a_valid_value_for_array);
      PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_kp_u_is_not_a_valid_value_for_array);
      __pyx_t_10 = __Pyx_PyUnicode_Join(__pyx_t_6, 3, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(20, 103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(20, 103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(20, 103, __pyx_L1_error)

      /* "src/array.pxi":102
 *         for item in it:
 *             item_fitness = item_type.get_value_fitness(item)
 *             if item_fitness == FIT_NONE:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"'{item}' is not a valid value for array")
 *             elif item_fitness < FIT_EXACT:
 */
    }

    /* "src/array.pxi":104
 *             if item_fitness == FIT_NONE:
 *                 raise ValueError(f"'{item}' is not a valid value for array")
 *             elif item_fitness < FIT_EXACT:             # <<<<<<<<<<<<<<
 *                 return self._make_converted_list(iter(value))
 *         return value
 */
    __pyx_t_2 = (__pyx_v_item_fitness < __pyx_v_5cavro_FIT_EXACT);
    if (__pyx_t_2) {

      /* "src/array.pxi":105
 *                 raise ValueError(f"'{item}' is not a valid value for array")
 *             elif item_fitness < FIT_EXACT:
 *                 return self._make_converted_list(iter(value))             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_6 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(20, 105, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = __pyx_f_5cavro_9ArrayType__make_converted_list(__pyx_v_self, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(20, 105, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = __pyx_t_10;
      __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "src/array.pxi":104
 *             if item_fitness == FIT_NONE:
 *                 raise ValueError(f"'{item}' is not a valid value for array")
 *             elif item_fitness < FIT_EXACT:             # <<<<<<<<<<<<<<
 *                 return self._make_converted_list(iter(value))
 *         return value
 */
    }

    /* "src/array.pxi":100
 * 
 *         it = iter(value)
 *         for item in it:             # <<<<<<<<<<<<<<
 *             item_fitness = item_type.get_value_fitness(item)
 *             if item_fitness == FIT_NONE:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/array.pxi":106
 *             elif item_fitness < FIT_EXACT:
 *                 return self._make_converted_list(iter(value))
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/array.pxi":92
 *         return out
 * 
 *     cpdef object convert_value(self, object value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef int item_fitness
 *         cdef AvroType item_type = self.item_type
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("cavro.ArrayType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_item_type);
  __Pyx_XDECREF(__pyx_v_it);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9ArrayType_10convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9ArrayType_10convert_value = {"convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9ArrayType_10convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9ArrayType_10convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_check_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,&__pyx_n_s_check_value,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_True));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 92, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_check_value);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 92, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "convert_value") < 0)) __PYX_ERR(20, 92, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_value = values[0];
    __pyx_v_check_value = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convert_value", 0, 1, 2, __pyx_nargs); __PYX_ERR(20, 92, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ArrayType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9ArrayType_9convert_value(((struct __pyx_obj_5cavro_ArrayType *)__pyx_v_self), __pyx_v_value, __pyx_v_check_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9ArrayType_9convert_value(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_check_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_convert_value __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.check_value = __pyx_v_check_value;
  __pyx_t_1 = __pyx_vtabptr_5cavro_ArrayType->__pyx_base.convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ArrayType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":108
 *         return value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return dict_to_canonical({
 *             'type': 'array',
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_9ArrayType_canonical_form(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/array.pxi":109
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return dict_to_canonical({             # <<<<<<<<<<<<<<
 *             'type': 'array',
 *             'items': self.item_type.canonical_form(created)
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);

  /* "src/array.pxi":110
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return dict_to_canonical({
 *             'type': 'array',             # <<<<<<<<<<<<<<
 *             'items': self.item_type.canonical_form(created)
 *         })
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_type, __pyx_n_u_array) < 0) __PYX_ERR(20, 110, __pyx_L1_error)

  /* "src/array.pxi":111
 *         return dict_to_canonical({
 *             'type': 'array',
 *             'items': self.item_type.canonical_form(created)             # <<<<<<<<<<<<<<
 *         })
 * 
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->item_type->__pyx_vtab)->canonical_form(__pyx_v_self->item_type, __pyx_v_created)); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_items, __pyx_t_2) < 0) __PYX_ERR(20, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/array.pxi":109
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return dict_to_canonical({             # <<<<<<<<<<<<<<
 *             'type': 'array',
 *             'items': self.item_type.canonical_form(created)
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_5cavro_dict_to_canonical(__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/array.pxi":108
 *         return value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return dict_to_canonical({
 *             'type': 'array',
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.ArrayType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":114
 *         })
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         cdef ArrayType cloned
 *         if isinstance(writer, ArrayType):
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9ArrayType__for_writer(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer) {
  struct __pyx_obj_5cavro_ArrayType *__pyx_v_cloned = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_promoted_item = NULL;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_writer", 1);

  /* "src/array.pxi":116
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         cdef ArrayType cloned
 *         if isinstance(writer, ArrayType):             # <<<<<<<<<<<<<<
 *             promoted_item = self.item_type._for_writer(writer.item_type)
 *             if promoted_item is not None:
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_ArrayType); 
  if (__pyx_t_1) {

    /* "src/array.pxi":117
 *         cdef ArrayType cloned
 *         if isinstance(writer, ArrayType):
 *             promoted_item = self.item_type._for_writer(writer.item_type)             # <<<<<<<<<<<<<<
 *             if promoted_item is not None:
 *                 cloned = self.clone_base()
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_writer), __pyx_n_s_item_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(20, 117, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->item_type->__pyx_vtab)->_for_writer(__pyx_v_self->item_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2))); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_promoted_item = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/array.pxi":118
 *         if isinstance(writer, ArrayType):
 *             promoted_item = self.item_type._for_writer(writer.item_type)
 *             if promoted_item is not None:             # <<<<<<<<<<<<<<
 *                 cloned = self.clone_base()
 *                 cloned.item_type = promoted_item
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_promoted_item) != Py_None);
    if (__pyx_t_1) {

      /* "src/array.pxi":119
 *             promoted_item = self.item_type._for_writer(writer.item_type)
 *             if promoted_item is not None:
 *                 cloned = self.clone_base()             # <<<<<<<<<<<<<<
 *                 cloned.item_type = promoted_item
 *                 return cloned
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_ArrayType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_ArrayType))))) __PYX_ERR(20, 119, __pyx_L1_error)
      __pyx_v_cloned = ((struct __pyx_obj_5cavro_ArrayType *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "src/array.pxi":120
 *             if promoted_item is not None:
 *                 cloned = self.clone_base()
 *                 cloned.item_type = promoted_item             # <<<<<<<<<<<<<<
 *                 return cloned
 * 
 */
      __Pyx_INCREF((PyObject *)__pyx_v_promoted_item);
      __Pyx_GIVEREF((PyObject *)__pyx_v_promoted_item);
      __Pyx_GOTREF((PyObject *)__pyx_v_cloned->item_type);
      __Pyx_DECREF((PyObject *)__pyx_v_cloned->item_type);
      __pyx_v_cloned->item_type = __pyx_v_promoted_item;

      /* "src/array.pxi":121
 *                 cloned = self.clone_base()
 *                 cloned.item_type = promoted_item
 *                 return cloned             # <<<<<<<<<<<<<<
 * 
 *     cdef bint accepts_missing_value(self):
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __Pyx_INCREF((PyObject *)__pyx_v_cloned);
      __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_cloned);
      goto __pyx_L0;

      /* "src/array.pxi":118
 *         if isinstance(writer, ArrayType):
 *             promoted_item = self.item_type._for_writer(writer.item_type)
 *             if promoted_item is not None:             # <<<<<<<<<<<<<<
 *                 cloned = self.clone_base()
 *                 cloned.item_type = promoted_item
 */
    }

    /* "src/array.pxi":116
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         cdef ArrayType cloned
 *         if isinstance(writer, ArrayType):             # <<<<<<<<<<<<<<
 *             promoted_item = self.item_type._for_writer(writer.item_type)
 *             if promoted_item is not None:
 */
  }

  /* "src/array.pxi":114
 *         })
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         cdef ArrayType cloned
 *         if isinstance(writer, ArrayType):
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.ArrayType._for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cloned);
  __Pyx_XDECREF((PyObject *)__pyx_v_promoted_item);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/array.pxi":123
 *                 return cloned
 * 
 *     cdef bint accepts_missing_value(self):             # <<<<<<<<<<<<<<
 *         if self.options.missing_values_can_be_empty_container:
 *             return True
 */

static int __pyx_f_5cavro_9ArrayType_accepts_missing_value(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self) {
  int __pyx_r;

  /* "src/array.pxi":124
 * 
 *     cdef bint accepts_missing_value(self):
 *         if self.options.missing_values_can_be_empty_container:             # <<<<<<<<<<<<<<
 *             return True
 */
  if (__pyx_v_self->__pyx_base.options->missing_values_can_be_empty_container) {

    /* "src/array.pxi":125
 *     cdef bint accepts_missing_value(self):
 *         if self.options.missing_values_can_be_empty_container:
 *             return True             # <<<<<<<<<<<<<<
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "src/array.pxi":124
 * 
 *     cdef bint accepts_missing_value(self):
 *         if self.options.missing_values_can_be_empty_container:             # <<<<<<<<<<<<<<
 *             return True
 */
  }

  /* "src/array.pxi":123
 *                 return cloned
 * 
 *     cdef bint accepts_missing_value(self):             # <<<<<<<<<<<<<<
 *         if self.options.missing_values_can_be_empty_container:
 *             return True
 */

  /* function exit code */
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/array.pxi":9
 *     type_name = "array"
 * 
 *     cdef readonly AvroType item_type             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, schema, source, namespace):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9ArrayType_9item_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_9ArrayType_9item_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_9ArrayType_9item_type___get__(((struct __pyx_obj_5cavro_ArrayType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9ArrayType_9item_type___get__(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->item_type);
  __pyx_r = ((PyObject *)__pyx_v_self->item_type);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9ArrayType_12__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9ArrayType_12__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9ArrayType_12__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9ArrayType_12__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_9ArrayType_11__reduce_cython__(((struct __pyx_obj_5cavro_ArrayType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9ArrayType_11__reduce_cython__(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.item_type, self.metadata, self.options, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->item_type);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->item_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->item_type))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.item_type, self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.item_type, self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.item_type is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.item_type, self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.item_type is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ArrayType, (type(self), 0xcdf32d8, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->item_type) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.item_type is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ArrayType, (type(self), 0xcdf32d8, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.item_type is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_ArrayType, (type(self), 0xcdf32d8, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ArrayType, (type(self), 0xcdf32d8, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_ArrayType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_215954136);
    __Pyx_GIVEREF(__pyx_int_215954136);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_215954136)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.item_type is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ArrayType, (type(self), 0xcdf32d8, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ArrayType, (type(self), 0xcdf32d8, None), state
 *     else:
 *         return __pyx_unpickle_ArrayType, (type(self), 0xcdf32d8, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ArrayType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_ArrayType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_215954136);
    __Pyx_GIVEREF(__pyx_int_215954136);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_215954136)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.ArrayType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ArrayType, (type(self), 0xcdf32d8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ArrayType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9ArrayType_14__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9ArrayType_14__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9ArrayType_14__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9ArrayType_14__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ArrayType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9ArrayType_13__setstate_cython__(((struct __pyx_obj_5cavro_ArrayType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9ArrayType_13__setstate_cython__(struct __pyx_obj_5cavro_ArrayType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ArrayType, (type(self), 0xcdf32d8, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ArrayType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_ArrayType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ArrayType, (type(self), 0xcdf32d8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ArrayType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ArrayType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":11
 *     cdef readonly str doc
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         _NamedType.__init__(self, schema, source, namespace)
 *         cdef Schema schema_ = schema
 */

/* Python wrapper */
static int __pyx_pw_5cavro_8EnumType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_8EnumType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_namespace = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_schema,&__pyx_n_s_source,&__pyx_n_s_namespace,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(8, 11, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(8, 11, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(8, 11, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_namespace)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(8, 11, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(8, 11, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(8, 11, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_schema = values[0];
    __pyx_v_source = values[1];
    __pyx_v_namespace = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(8, 11, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.EnumType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8EnumType___init__(((struct __pyx_obj_5cavro_EnumType *)__pyx_v_self), __pyx_v_schema, __pyx_v_source, __pyx_v_namespace);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_8EnumType___init__(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace) {
  struct __pyx_obj_5cavro_Schema *__pyx_v_schema_ = 0;
  PyObject *__pyx_v_raw_symbols = NULL;
  PyObject *__pyx_v_seen_symbols = NULL;
  PyObject *__pyx_v_symbol = NULL;
  PyObject *__pyx_v_name_pattern = NULL;
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  Py_UCS4 __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/enum.pxi":12
 * 
 *     def __init__(self, schema, source, namespace):
 *         _NamedType.__init__(self, schema, source, namespace)             # <<<<<<<<<<<<<<
 *         cdef Schema schema_ = schema
 *         raw_symbols = source['symbols']
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__NamedType), __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_schema, __pyx_v_source, __pyx_v_namespace};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 4+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/enum.pxi":13
 *     def __init__(self, schema, source, namespace):
 *         _NamedType.__init__(self, schema, source, namespace)
 *         cdef Schema schema_ = schema             # <<<<<<<<<<<<<<
 *         raw_symbols = source['symbols']
 *         if not isinstance(raw_symbols, (list, tuple)):
 */
  if (!(likely(((__pyx_v_schema) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_schema, __pyx_ptype_5cavro_Schema))))) __PYX_ERR(8, 13, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_schema;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_schema_ = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/enum.pxi":14
 *         _NamedType.__init__(self, schema, source, namespace)
 *         cdef Schema schema_ = schema
 *         raw_symbols = source['symbols']             # <<<<<<<<<<<<<<
 *         if not isinstance(raw_symbols, (list, tuple)):
 *             raise ValueError("Enum symbols must be a list of strings")
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_source, __pyx_n_u_symbols); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_raw_symbols = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/enum.pxi":15
 *         cdef Schema schema_ = schema
 *         raw_symbols = source['symbols']
 *         if not isinstance(raw_symbols, (list, tuple)):             # <<<<<<<<<<<<<<
 *             raise ValueError("Enum symbols must be a list of strings")
 * 
 */
  __pyx_t_6 = PyList_Check(__pyx_v_raw_symbols); 
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = PyTuple_Check(__pyx_v_raw_symbols); 
  __pyx_t_5 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  __pyx_t_6 = (!__pyx_t_5);
  if (unlikely(__pyx_t_6)) {

    /* "src/enum.pxi":16
 *         raw_symbols = source['symbols']
 *         if not isinstance(raw_symbols, (list, tuple)):
 *             raise ValueError("Enum symbols must be a list of strings")             # <<<<<<<<<<<<<<
 * 
 *         if schema_.options.enum_symbols_must_be_unique:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 16, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(8, 16, __pyx_L1_error)

    /* "src/enum.pxi":15
 *         cdef Schema schema_ = schema
 *         raw_symbols = source['symbols']
 *         if not isinstance(raw_symbols, (list, tuple)):             # <<<<<<<<<<<<<<
 *             raise ValueError("Enum symbols must be a list of strings")
 * 
 */
  }

  /* "src/enum.pxi":18
 *             raise ValueError("Enum symbols must be a list of strings")
 * 
 *         if schema_.options.enum_symbols_must_be_unique:             # <<<<<<<<<<<<<<
 *             seen_symbols = set()
 *             for symbol in raw_symbols:
 */
  if (__pyx_v_schema_->options->enum_symbols_must_be_unique) {

    /* "src/enum.pxi":19
 * 
 *         if schema_.options.enum_symbols_must_be_unique:
 *             seen_symbols = set()             # <<<<<<<<<<<<<<
 *             for symbol in raw_symbols:
 *                 if symbol in seen_symbols:
 */
    __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 19, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_seen_symbols = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "src/enum.pxi":20
 *         if schema_.options.enum_symbols_must_be_unique:
 *             seen_symbols = set()
 *             for symbol in raw_symbols:             # <<<<<<<<<<<<<<
 *                 if symbol in seen_symbols:
 *                     raise DuplicateName(f"Enum symbols must be unique. '{symbol}' found twice")
 */
    if (likely(PyList_CheckExact(__pyx_v_raw_symbols)) || PyTuple_CheckExact(__pyx_v_raw_symbols)) {
      __pyx_t_1 = __pyx_v_raw_symbols; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_raw_symbols); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 20, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(8, 20, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(8, 20, __pyx_L1_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(8, 20, __pyx_L1_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 20, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(8, 20, __pyx_L1_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(8, 20, __pyx_L1_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 20, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_8(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(8, 20, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_symbol, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "src/enum.pxi":21
 *             seen_symbols = set()
 *             for symbol in raw_symbols:
 *                 if symbol in seen_symbols:             # <<<<<<<<<<<<<<
 *                     raise DuplicateName(f"Enum symbols must be unique. '{symbol}' found twice")
 *                 seen_symbols.add(symbol)
 */
      __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_v_symbol, __pyx_v_seen_symbols, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(8, 21, __pyx_L1_error)
      if (unlikely(__pyx_t_6)) {

        /* "src/enum.pxi":22
 *             for symbol in raw_symbols:
 *                 if symbol in seen_symbols:
 *                     raise DuplicateName(f"Enum symbols must be unique. '{symbol}' found twice")             # <<<<<<<<<<<<<<
 *                 seen_symbols.add(symbol)
 *         else:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_DuplicateName); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 22, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(8, 22, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = 0;
        __pyx_t_11 = 127;
        __Pyx_INCREF(__pyx_kp_u_Enum_symbols_must_be_unique);
        __pyx_t_10 += 30;
        __Pyx_GIVEREF(__pyx_kp_u_Enum_symbols_must_be_unique);
        PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_kp_u_Enum_symbols_must_be_unique);
        __pyx_t_12 = __Pyx_PyObject_FormatSimple(__pyx_v_symbol, __pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(8, 22, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
        __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_12);
        __pyx_t_12 = 0;
        __Pyx_INCREF(__pyx_kp_u_found_twice);
        __pyx_t_10 += 13;
        __Pyx_GIVEREF(__pyx_kp_u_found_twice);
        PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_kp_u_found_twice);
        __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_9, 3, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(8, 22, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = NULL;
        __pyx_t_4 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_4 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_12};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 22, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(8, 22, __pyx_L1_error)

        /* "src/enum.pxi":21
 *             seen_symbols = set()
 *             for symbol in raw_symbols:
 *                 if symbol in seen_symbols:             # <<<<<<<<<<<<<<
 *                     raise DuplicateName(f"Enum symbols must be unique. '{symbol}' found twice")
 *                 seen_symbols.add(symbol)
 */
      }

      /* "src/enum.pxi":23
 *                 if symbol in seen_symbols:
 *                     raise DuplicateName(f"Enum symbols must be unique. '{symbol}' found twice")
 *                 seen_symbols.add(symbol)             # <<<<<<<<<<<<<<
 *         else:
 *             seen_symbols = raw_symbols
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_seen_symbols, __pyx_n_s_add); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 23, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_12 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_symbol};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 23, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "src/enum.pxi":20
 *         if schema_.options.enum_symbols_must_be_unique:
 *             seen_symbols = set()
 *             for symbol in raw_symbols:             # <<<<<<<<<<<<<<
 *                 if symbol in seen_symbols:
 *                     raise DuplicateName(f"Enum symbols must be unique. '{symbol}' found twice")
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/enum.pxi":18
 *             raise ValueError("Enum symbols must be a list of strings")
 * 
 *         if schema_.options.enum_symbols_must_be_unique:             # <<<<<<<<<<<<<<
 *             seen_symbols = set()
 *             for symbol in raw_symbols:
 */
    goto __pyx_L6;
  }

  /* "src/enum.pxi":25
 *                 seen_symbols.add(symbol)
 *         else:
 *             seen_symbols = raw_symbols             # <<<<<<<<<<<<<<
 * 
 *         name_pattern = schema_.options.name_pattern
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_raw_symbols);
    __pyx_v_seen_symbols = __pyx_v_raw_symbols;
  }
  __pyx_L6:;

  /* "src/enum.pxi":27
 *             seen_symbols = raw_symbols
 * 
 *         name_pattern = schema_.options.name_pattern             # <<<<<<<<<<<<<<
 *         if schema_.options.enforce_enum_symbol_name_rules:
 *             for symbol in seen_symbols:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_schema_->options), __pyx_n_s_name_pattern); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_name_pattern = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/enum.pxi":28
 * 
 *         name_pattern = schema_.options.name_pattern
 *         if schema_.options.enforce_enum_symbol_name_rules:             # <<<<<<<<<<<<<<
 *             for symbol in seen_symbols:
 *                 if not isinstance(symbol, str):
 */
  if (__pyx_v_schema_->options->enforce_enum_symbol_name_rules) {

    /* "src/enum.pxi":29
 *         name_pattern = schema_.options.name_pattern
 *         if schema_.options.enforce_enum_symbol_name_rules:
 *             for symbol in seen_symbols:             # <<<<<<<<<<<<<<
 *                 if not isinstance(symbol, str):
 *                     raise InvalidName('Enum symbols must be a list of strings')
 */
    if (likely(PyList_CheckExact(__pyx_v_seen_symbols)) || PyTuple_CheckExact(__pyx_v_seen_symbols)) {
      __pyx_t_1 = __pyx_v_seen_symbols; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_seen_symbols); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 29, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(8, 29, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(8, 29, __pyx_L1_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(8, 29, __pyx_L1_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 29, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(8, 29, __pyx_L1_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(8, 29, __pyx_L1_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 29, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_8(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(8, 29, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_symbol, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "src/enum.pxi":30
 *         if schema_.options.enforce_enum_symbol_name_rules:
 *             for symbol in seen_symbols:
 *                 if not isinstance(symbol, str):             # <<<<<<<<<<<<<<
 *                     raise InvalidName('Enum symbols must be a list of strings')
 *                 if schema_.options.enforce_enum_symbol_name_rules:
 */
      __pyx_t_6 = PyUnicode_Check(__pyx_v_symbol); 
      __pyx_t_5 = (!__pyx_t_6);
      if (unlikely(__pyx_t_5)) {

        /* "src/enum.pxi":31
 *             for symbol in seen_symbols:
 *                 if not isinstance(symbol, str):
 *                     raise InvalidName('Enum symbols must be a list of strings')             # <<<<<<<<<<<<<<
 *                 if schema_.options.enforce_enum_symbol_name_rules:
 *                     if not name_pattern.fullmatch(symbol):
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_InvalidName); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 31, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_12 = NULL;
        __pyx_t_4 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_4 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_kp_u_Enum_symbols_must_be_a_list_of_s};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 31, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(8, 31, __pyx_L1_error)

        /* "src/enum.pxi":30
 *         if schema_.options.enforce_enum_symbol_name_rules:
 *             for symbol in seen_symbols:
 *                 if not isinstance(symbol, str):             # <<<<<<<<<<<<<<
 *                     raise InvalidName('Enum symbols must be a list of strings')
 *                 if schema_.options.enforce_enum_symbol_name_rules:
 */
      }

      /* "src/enum.pxi":32
 *                 if not isinstance(symbol, str):
 *                     raise InvalidName('Enum symbols must be a list of strings')
 *                 if schema_.options.enforce_enum_symbol_name_rules:             # <<<<<<<<<<<<<<
 *                     if not name_pattern.fullmatch(symbol):
 *                         raise InvalidName(f"Enum symbol '{symbol}' is not a valid name")
 */
      if (__pyx_v_schema_->options->enforce_enum_symbol_name_rules) {

        /* "src/enum.pxi":33
 *                     raise InvalidName('Enum symbols must be a list of strings')
 *                 if schema_.options.enforce_enum_symbol_name_rules:
 *                     if not name_pattern.fullmatch(symbol):             # <<<<<<<<<<<<<<
 *                         raise InvalidName(f"Enum symbol '{symbol}' is not a valid name")
 * 
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_name_pattern, __pyx_n_s_fullmatch); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 33, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_12 = NULL;
        __pyx_t_4 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_4 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_symbol};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 33, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(8, 33, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_6 = (!__pyx_t_5);
        if (unlikely(__pyx_t_6)) {

          /* "src/enum.pxi":34
 *                 if schema_.options.enforce_enum_symbol_name_rules:
 *                     if not name_pattern.fullmatch(symbol):
 *                         raise InvalidName(f"Enum symbol '{symbol}' is not a valid name")             # <<<<<<<<<<<<<<
 * 
 *         self.symbols = tuple(raw_symbols)
 */
          __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_InvalidName); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 34, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(8, 34, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_10 = 0;
          __pyx_t_11 = 127;
          __Pyx_INCREF(__pyx_kp_u_Enum_symbol);
          __pyx_t_10 += 13;
          __Pyx_GIVEREF(__pyx_kp_u_Enum_symbol);
          PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_kp_u_Enum_symbol);
          __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_v_symbol, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(8, 34, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_11;
          __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_9);
          __pyx_t_9 = 0;
          __Pyx_INCREF(__pyx_kp_u_is_not_a_valid_name);
          __pyx_t_10 += 21;
          __Pyx_GIVEREF(__pyx_kp_u_is_not_a_valid_name);
          PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_kp_u_is_not_a_valid_name);
          __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_12, 3, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(8, 34, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_12 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_9};
            __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 34, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_Raise(__pyx_t_2, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __PYX_ERR(8, 34, __pyx_L1_error)

          /* "src/enum.pxi":33
 *                     raise InvalidName('Enum symbols must be a list of strings')
 *                 if schema_.options.enforce_enum_symbol_name_rules:
 *                     if not name_pattern.fullmatch(symbol):             # <<<<<<<<<<<<<<
 *                         raise InvalidName(f"Enum symbol '{symbol}' is not a valid name")
 * 
 */
        }

        /* "src/enum.pxi":32
 *                 if not isinstance(symbol, str):
 *                     raise InvalidName('Enum symbols must be a list of strings')
 *                 if schema_.options.enforce_enum_symbol_name_rules:             # <<<<<<<<<<<<<<
 *                     if not name_pattern.fullmatch(symbol):
 *                         raise InvalidName(f"Enum symbol '{symbol}' is not a valid name")
 */
      }

      /* "src/enum.pxi":29
 *         name_pattern = schema_.options.name_pattern
 *         if schema_.options.enforce_enum_symbol_name_rules:
 *             for symbol in seen_symbols:             # <<<<<<<<<<<<<<
 *                 if not isinstance(symbol, str):
 *                     raise InvalidName('Enum symbols must be a list of strings')
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/enum.pxi":28
 * 
 *         name_pattern = schema_.options.name_pattern
 *         if schema_.options.enforce_enum_symbol_name_rules:             # <<<<<<<<<<<<<<
 *             for symbol in seen_symbols:
 *                 if not isinstance(symbol, str):
 */
  }

  /* "src/enum.pxi":36
 *                         raise InvalidName(f"Enum symbol '{symbol}' is not a valid name")
 * 
 *         self.symbols = tuple(raw_symbols)             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         self.symbol_indexes = {}
 */
  __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_v_raw_symbols); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->symbols);
  __Pyx_DECREF(__pyx_v_self->symbols);
  __pyx_v_self->symbols = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/enum.pxi":38
 *         self.symbols = tuple(raw_symbols)
 *         cdef size_t i
 *         self.symbol_indexes = {}             # <<<<<<<<<<<<<<
 *         for i, symbol in enumerate(self.symbols):
 *             self.symbol_indexes[symbol] = i
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->symbol_indexes);
  __Pyx_DECREF(__pyx_v_self->symbol_indexes);
  __pyx_v_self->symbol_indexes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/enum.pxi":39
 *         cdef size_t i
 *         self.symbol_indexes = {}
 *         for i, symbol in enumerate(self.symbols):             # <<<<<<<<<<<<<<
 *             self.symbol_indexes[symbol] = i
 *         self.default_value = source.get('default', NO_DEFAULT)
 */
  __pyx_t_13 = 0;
  __pyx_t_1 = __pyx_v_self->symbols; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_7 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(8, 39, __pyx_L1_error)
      #endif
      if (__pyx_t_7 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(8, 39, __pyx_L1_error)
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_symbol, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_v_i = __pyx_t_13;
    __pyx_t_13 = (__pyx_t_13 + 1);

    /* "src/enum.pxi":40
 *         self.symbol_indexes = {}
 *         for i, symbol in enumerate(self.symbols):
 *             self.symbol_indexes[symbol] = i             # <<<<<<<<<<<<<<
 *         self.default_value = source.get('default', NO_DEFAULT)
 *         if self.default_value is not NO_DEFAULT:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_self->symbol_indexes == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(8, 40, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__pyx_v_self->symbol_indexes, __pyx_v_symbol, __pyx_t_2) < 0))) __PYX_ERR(8, 40, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/enum.pxi":39
 *         cdef size_t i
 *         self.symbol_indexes = {}
 *         for i, symbol in enumerate(self.symbols):             # <<<<<<<<<<<<<<
 *             self.symbol_indexes[symbol] = i
 *         self.default_value = source.get('default', NO_DEFAULT)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/enum.pxi":41
 *         for i, symbol in enumerate(self.symbols):
 *             self.symbol_indexes[symbol] = i
 *         self.default_value = source.get('default', NO_DEFAULT)             # <<<<<<<<<<<<<<
 *         if self.default_value is not NO_DEFAULT:
 *             if self.default_value not in self.symbol_indexes:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_9, __pyx_n_u_default, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->default_value);
  __Pyx_DECREF(__pyx_v_self->default_value);
  __pyx_v_self->default_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/enum.pxi":42
 *             self.symbol_indexes[symbol] = i
 *         self.default_value = source.get('default', NO_DEFAULT)
 *         if self.default_value is not NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             if self.default_value not in self.symbol_indexes:
 *                 raise ValueError(f"Default value '{self.default_value}' not in enum symbols")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = (__pyx_v_self->default_value != __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "src/enum.pxi":43
 *         self.default_value = source.get('default', NO_DEFAULT)
 *         if self.default_value is not NO_DEFAULT:
 *             if self.default_value not in self.symbol_indexes:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Default value '{self.default_value}' not in enum symbols")
 *         self.doc = source.get('doc', '')
 */
    if (unlikely(__pyx_v_self->symbol_indexes == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(8, 43, __pyx_L1_error)
    }
    __pyx_t_6 = (__Pyx_PyDict_ContainsTF(__pyx_v_self->default_value, __pyx_v_self->symbol_indexes, Py_NE)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(8, 43, __pyx_L1_error)
    if (unlikely(__pyx_t_6)) {

      /* "src/enum.pxi":44
 *         if self.default_value is not NO_DEFAULT:
 *             if self.default_value not in self.symbol_indexes:
 *                 raise ValueError(f"Default value '{self.default_value}' not in enum symbols")             # <<<<<<<<<<<<<<
 *         self.doc = source.get('doc', '')
 * 
 */
      __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 44, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = 0;
      __pyx_t_11 = 127;
      __Pyx_INCREF(__pyx_kp_u_Default_value);
      __pyx_t_7 += 15;
      __Pyx_GIVEREF(__pyx_kp_u_Default_value);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Default_value);
      __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_self->default_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 44, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_11;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_INCREF(__pyx_kp_u_not_in_enum_symbols);
      __pyx_t_7 += 21;
      __Pyx_GIVEREF(__pyx_kp_u_not_in_enum_symbols);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_not_in_enum_symbols);
      __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 44, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 44, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(8, 44, __pyx_L1_error)

      /* "src/enum.pxi":43
 *         self.default_value = source.get('default', NO_DEFAULT)
 *         if self.default_value is not NO_DEFAULT:
 *             if self.default_value not in self.symbol_indexes:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Default value '{self.default_value}' not in enum symbols")
 *         self.doc = source.get('doc', '')
 */
    }

    /* "src/enum.pxi":42
 *             self.symbol_indexes[symbol] = i
 *         self.default_value = source.get('default', NO_DEFAULT)
 *         if self.default_value is not NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             if self.default_value not in self.symbol_indexes:
 *                 raise ValueError(f"Default value '{self.default_value}' not in enum symbols")
 */
  }

  /* "src/enum.pxi":45
 *             if self.default_value not in self.symbol_indexes:
 *                 raise ValueError(f"Default value '{self.default_value}' not in enum symbols")
 *         self.doc = source.get('doc', '')             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(8, 45, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->doc);
  __Pyx_DECREF(__pyx_v_self->doc);
  __pyx_v_self->doc = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/enum.pxi":11
 *     cdef readonly str doc
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         _NamedType.__init__(self, schema, source, namespace)
 *         cdef Schema schema_ = schema
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("cavro.EnumType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_schema_);
  __Pyx_XDECREF(__pyx_v_raw_symbols);
  __Pyx_XDECREF(__pyx_v_seen_symbols);
  __Pyx_XDECREF(__pyx_v_symbol);
  __Pyx_XDECREF(__pyx_v_name_pattern);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":47
 *         self.doc = source.get('doc', '')
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef EnumType new_inst = self.clone_base()
 *         new_inst.symbols = self.symbols
 */

static PyObject *__pyx_pw_5cavro_8EnumType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8EnumType_copy(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_EnumType *__pyx_v_new_inst = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 47, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_8EnumType_3copy)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 47, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(8, 47, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/enum.pxi":48
 * 
 *     cpdef AvroType copy(self):
 *         cdef EnumType new_inst = self.clone_base()             # <<<<<<<<<<<<<<
 *         new_inst.symbols = self.symbols
 *         new_inst.symbol_indexes = self.symbol_indexes
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_EnumType *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_EnumType))))) __PYX_ERR(8, 48, __pyx_L1_error)
  __pyx_v_new_inst = ((struct __pyx_obj_5cavro_EnumType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/enum.pxi":49
 *     cpdef AvroType copy(self):
 *         cdef EnumType new_inst = self.clone_base()
 *         new_inst.symbols = self.symbols             # <<<<<<<<<<<<<<
 *         new_inst.symbol_indexes = self.symbol_indexes
 *         new_inst.default_value = self.default_value
 */
  __pyx_t_1 = __pyx_v_self->symbols;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_new_inst->symbols);
  __Pyx_DECREF(__pyx_v_new_inst->symbols);
  __pyx_v_new_inst->symbols = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/enum.pxi":50
 *         cdef EnumType new_inst = self.clone_base()
 *         new_inst.symbols = self.symbols
 *         new_inst.symbol_indexes = self.symbol_indexes             # <<<<<<<<<<<<<<
 *         new_inst.default_value = self.default_value
 *         new_inst.doc = self.doc
 */
  __pyx_t_1 = __pyx_v_self->symbol_indexes;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_new_inst->symbol_indexes);
  __Pyx_DECREF(__pyx_v_new_inst->symbol_indexes);
  __pyx_v_new_inst->symbol_indexes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/enum.pxi":51
 *         new_inst.symbols = self.symbols
 *         new_inst.symbol_indexes = self.symbol_indexes
 *         new_inst.default_value = self.default_value             # <<<<<<<<<<<<<<
 *         new_inst.doc = self.doc
 *         return new_inst
 */
  __pyx_t_1 = __pyx_v_self->default_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_new_inst->default_value);
  __Pyx_DECREF(__pyx_v_new_inst->default_value);
  __pyx_v_new_inst->default_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/enum.pxi":52
 *         new_inst.symbol_indexes = self.symbol_indexes
 *         new_inst.default_value = self.default_value
 *         new_inst.doc = self.doc             # <<<<<<<<<<<<<<
 *         return new_inst
 * 
 */
  __pyx_t_1 = __pyx_v_self->doc;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_new_inst->doc);
  __Pyx_DECREF(__pyx_v_new_inst->doc);
  __pyx_v_new_inst->doc = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/enum.pxi":53
 *         new_inst.default_value = self.default_value
 *         new_inst.doc = self.doc
 *         return new_inst             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_new_inst);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_new_inst);
  goto __pyx_L0;

  /* "src/enum.pxi":47
 *         self.doc = source.get('doc', '')
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef EnumType new_inst = self.clone_base()
 *         new_inst.symbols = self.symbols
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.EnumType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_new_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8EnumType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8EnumType_3copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8EnumType_3copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8EnumType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_8EnumType_2copy(((struct __pyx_obj_5cavro_EnumType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8EnumType_2copy(struct __pyx_obj_5cavro_EnumType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_8EnumType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.EnumType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":55
 *         return new_inst
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {
 *             'type',
 */

static PyObject *__pyx_f_5cavro_8EnumType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/enum.pxi":56
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {             # <<<<<<<<<<<<<<
 *             'type',
 *             'name',
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(8, 57, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_name) < 0) __PYX_ERR(8, 57, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_namespace) < 0) __PYX_ERR(8, 57, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_aliases) < 0) __PYX_ERR(8, 57, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_doc) < 0) __PYX_ERR(8, 57, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_symbols) < 0) __PYX_ERR(8, 57, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/enum.pxi":55
 *         return new_inst
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {
 *             'type',
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.EnumType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":65
 *         })
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         extra = _NamedType._get_schema_extra(self, created)
 *         if self.default_value is not NO_DEFAULT:
 */

static PyObject *__pyx_pw_5cavro_8EnumType_5_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_8EnumType__get_schema_extra(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_created, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_extra = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_schema_extra); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 65, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_8EnumType_5_get_schema_extra)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_created};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 65, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_2))) __PYX_ERR(8, 65, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/enum.pxi":66
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         extra = _NamedType._get_schema_extra(self, created)             # <<<<<<<<<<<<<<
 *         if self.default_value is not NO_DEFAULT:
 *             extra['default'] = self.default_value
 */
  __pyx_t_1 = __pyx_f_5cavro_10_NamedType__get_schema_extra(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self), __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_extra = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/enum.pxi":67
 *     cpdef dict _get_schema_extra(self, set created):
 *         extra = _NamedType._get_schema_extra(self, created)
 *         if self.default_value is not NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             extra['default'] = self.default_value
 *         extra['symbols'] = list(self.symbols)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = (__pyx_v_self->default_value != __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "src/enum.pxi":68
 *         extra = _NamedType._get_schema_extra(self, created)
 *         if self.default_value is not NO_DEFAULT:
 *             extra['default'] = self.default_value             # <<<<<<<<<<<<<<
 *         extra['symbols'] = list(self.symbols)
 *         return extra
 */
    __pyx_t_1 = __pyx_v_self->default_value;
    __Pyx_INCREF(__pyx_t_1);
    if (unlikely(__pyx_v_extra == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(8, 68, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__pyx_v_extra, __pyx_n_u_default, __pyx_t_1) < 0))) __PYX_ERR(8, 68, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/enum.pxi":67
 *     cpdef dict _get_schema_extra(self, set created):
 *         extra = _NamedType._get_schema_extra(self, created)
 *         if self.default_value is not NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             extra['default'] = self.default_value
 *         extra['symbols'] = list(self.symbols)
 */
  }

  /* "src/enum.pxi":69
 *         if self.default_value is not NO_DEFAULT:
 *             extra['default'] = self.default_value
 *         extra['symbols'] = list(self.symbols)             # <<<<<<<<<<<<<<
 *         return extra
 * 
 */
  __pyx_t_1 = PySequence_List(__pyx_v_self->symbols); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_v_extra == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(8, 69, __pyx_L1_error)
  }
  if (unlikely((PyDict_SetItem(__pyx_v_extra, __pyx_n_u_symbols, __pyx_t_1) < 0))) __PYX_ERR(8, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/enum.pxi":70
 *             extra['default'] = self.default_value
 *         extra['symbols'] = list(self.symbols)
 *         return extra             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_extra);
  __pyx_r = __pyx_v_extra;
  goto __pyx_L0;

  /* "src/enum.pxi":65
 *         })
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         extra = _NamedType._get_schema_extra(self, created)
 *         if self.default_value is not NO_DEFAULT:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.EnumType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_extra);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8EnumType_5_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8EnumType_5_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8EnumType_5_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8EnumType_5_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(8, 65, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(8, 65, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(8, 65, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.EnumType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(8, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_8EnumType_4_get_schema_extra(((struct __pyx_obj_5cavro_EnumType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8EnumType_4_get_schema_extra(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8EnumType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.EnumType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":72
 *         return extra
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef size_t index = self.symbol_indexes[value]
 *         zigzag_encode_long(buffer, index)
 */

static int __pyx_f_5cavro_8EnumType__binary_buffer_encode(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  size_t __pyx_v_index;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 1);

  /* "src/enum.pxi":73
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef size_t index = self.symbol_indexes[value]             # <<<<<<<<<<<<<<
 *         zigzag_encode_long(buffer, index)
 * 
 */
  if (unlikely(__pyx_v_self->symbol_indexes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(8, 73, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->symbol_indexes, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(8, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_index = __pyx_t_2;

  /* "src/enum.pxi":74
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef size_t index = self.symbol_indexes[value]
 *         zigzag_encode_long(buffer, index)             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_t_3 = __pyx_f_5cavro_zigzag_encode_long(__pyx_v_buffer, __pyx_v_index); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(8, 74, __pyx_L1_error)

  /* "src/enum.pxi":72
 *         return extra
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef size_t index = self.symbol_indexes[value]
 *         zigzag_encode_long(buffer, index)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.EnumType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":76
 *         zigzag_encode_long(buffer, index)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return self.symbols[zigzag_decode_long(buffer)]
 * 
 */

static PyObject *__pyx_f_5cavro_8EnumType__binary_buffer_decode(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/enum.pxi":77
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         return self.symbols[zigzag_decode_long(buffer)]             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->symbols == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(8, 77, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_f_5cavro_zigzag_decode_long(__pyx_v_buffer); if (unlikely(__pyx_t_1 == ((int64_t)0x7ffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(8, 77, __pyx_L1_error)
  __pyx_t_2 = __Pyx_GetItemInt_Tuple(__pyx_v_self->symbols, __pyx_t_1, int64_t, 1, __Pyx_PyInt_From_int64_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/enum.pxi":76
 *         zigzag_encode_long(buffer, index)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return self.symbols[zigzag_decode_long(buffer)]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.EnumType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":79
 *         return self.symbols[zigzag_decode_long(buffer)]
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         try:
 *             if value in self.symbol_indexes:
 */

static int __pyx_f_5cavro_8EnumType__get_value_fitness(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 1);

  /* "src/enum.pxi":80
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         try:             # <<<<<<<<<<<<<<
 *             if value in self.symbol_indexes:
 *                 return FIT_EXACT
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "src/enum.pxi":81
 *     cdef int _get_value_fitness(self, value) except -1:
 *         try:
 *             if value in self.symbol_indexes:             # <<<<<<<<<<<<<<
 *                 return FIT_EXACT
 *         except TypeError:
 */
      if (unlikely(__pyx_v_self->symbol_indexes == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(8, 81, __pyx_L3_error)
      }
      __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_v_value, __pyx_v_self->symbol_indexes, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(8, 81, __pyx_L3_error)
      if (__pyx_t_4) {

        /* "src/enum.pxi":82
 *         try:
 *             if value in self.symbol_indexes:
 *                 return FIT_EXACT             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             return FIT_NONE
 */
        __pyx_r = __pyx_v_5cavro_FIT_EXACT;
        goto __pyx_L7_try_return;

        /* "src/enum.pxi":81
 *     cdef int _get_value_fitness(self, value) except -1:
 *         try:
 *             if value in self.symbol_indexes:             # <<<<<<<<<<<<<<
 *                 return FIT_EXACT
 *         except TypeError:
 */
      }

      /* "src/enum.pxi":80
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         try:             # <<<<<<<<<<<<<<
 *             if value in self.symbol_indexes:
 *                 return FIT_EXACT
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "src/enum.pxi":83
 *             if value in self.symbol_indexes:
 *                 return FIT_EXACT
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             return FIT_NONE
 * 
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_5) {
      __Pyx_AddTraceback("cavro.EnumType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(8, 83, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "src/enum.pxi":84
 *                 return FIT_EXACT
 *         except TypeError:
 *             return FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
      __pyx_r = __pyx_v_5cavro_FIT_NONE;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "src/enum.pxi":80
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         try:             # <<<<<<<<<<<<<<
 *             if value in self.symbol_indexes:
 *                 return FIT_EXACT
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "src/enum.pxi":79
 *         return self.symbols[zigzag_decode_long(buffer)]
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         try:
 *             if value in self.symbol_indexes:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cavro.EnumType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":86
 *             return FIT_NONE
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         # if value is valid for type, then it needs no conversion
 *         return value
 */

static PyObject *__pyx_pw_5cavro_8EnumType_7_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_8EnumType__convert_value(CYTHON_UNUSED struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_value, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_convert_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 86, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_8EnumType_7_convert_value)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_value};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 86, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/enum.pxi":88
 *     cpdef object _convert_value(self, object value):
 *         # if value is valid for type, then it needs no conversion
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/enum.pxi":86
 *             return FIT_NONE
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         # if value is valid for type, then it needs no conversion
 *         return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.EnumType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8EnumType_7_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8EnumType_7_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8EnumType_7_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8EnumType_7_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(8, 86, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(8, 86, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(8, 86, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.EnumType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8EnumType_6_convert_value(((struct __pyx_obj_5cavro_EnumType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8EnumType_6_convert_value(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8EnumType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.EnumType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":90
 *         return value
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         if value not in self.symbol_indexes:
 *             raise KeyError(f"'{value}' invalid for enum")
 */

static PyObject *__pyx_f_5cavro_8EnumType__json_format(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 1);

  /* "src/enum.pxi":91
 * 
 *     cdef _json_format(self, value):
 *         if value not in self.symbol_indexes:             # <<<<<<<<<<<<<<
 *             raise KeyError(f"'{value}' invalid for enum")
 *         return value
 */
  if (unlikely(__pyx_v_self->symbol_indexes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(8, 91, __pyx_L1_error)
  }
  __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_value, __pyx_v_self->symbol_indexes, Py_NE)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(8, 91, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "src/enum.pxi":92
 *     cdef _json_format(self, value):
 *         if value not in self.symbol_indexes:
 *             raise KeyError(f"'{value}' invalid for enum")             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u__6);
    __pyx_t_3 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__6);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u__6);
    __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_invalid_for_enum);
    __pyx_t_3 += 18;
    __Pyx_GIVEREF(__pyx_kp_u_invalid_for_enum);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_invalid_for_enum);
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(8, 92, __pyx_L1_error)

    /* "src/enum.pxi":91
 * 
 *     cdef _json_format(self, value):
 *         if value not in self.symbol_indexes:             # <<<<<<<<<<<<<<
 *             raise KeyError(f"'{value}' invalid for enum")
 *         return value
 */
  }

  /* "src/enum.pxi":93
 *         if value not in self.symbol_indexes:
 *             raise KeyError(f"'{value}' invalid for enum")
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/enum.pxi":90
 *         return value
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         if value not in self.symbol_indexes:
 *             raise KeyError(f"'{value}' invalid for enum")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.EnumType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":95
 *         return value
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if value not in self.symbol_indexes:
 *             raise ValueError(f"'{value}' invalid for enum")
 */

static PyObject *__pyx_f_5cavro_8EnumType__json_decode(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/enum.pxi":96
 * 
 *     cdef _json_decode(self, value):
 *         if value not in self.symbol_indexes:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"'{value}' invalid for enum")
 *         return value
 */
  if (unlikely(__pyx_v_self->symbol_indexes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(8, 96, __pyx_L1_error)
  }
  __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_value, __pyx_v_self->symbol_indexes, Py_NE)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(8, 96, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "src/enum.pxi":97
 *     cdef _json_decode(self, value):
 *         if value not in self.symbol_indexes:
 *             raise ValueError(f"'{value}' invalid for enum")             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u__6);
    __pyx_t_3 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__6);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u__6);
    __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_invalid_for_enum);
    __pyx_t_3 += 18;
    __Pyx_GIVEREF(__pyx_kp_u_invalid_for_enum);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_invalid_for_enum);
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(8, 97, __pyx_L1_error)

    /* "src/enum.pxi":96
 * 
 *     cdef _json_decode(self, value):
 *         if value not in self.symbol_indexes:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"'{value}' invalid for enum")
 *         return value
 */
  }

  /* "src/enum.pxi":98
 *         if value not in self.symbol_indexes:
 *             raise ValueError(f"'{value}' invalid for enum")
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/enum.pxi":95
 *         return value
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if value not in self.symbol_indexes:
 *             raise ValueError(f"'{value}' invalid for enum")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.EnumType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":100
 *         return value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         if self in created and not self.options.canonical_form_repeat_enum:
 *             return _CanonicalForm('"' + self.type + '"')
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_8EnumType_canonical_form(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/enum.pxi":101
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         if self in created and not self.options.canonical_form_repeat_enum:             # <<<<<<<<<<<<<<
 *             return _CanonicalForm('"' + self.type + '"')
 *         created.add(self)
 */
  if (unlikely(__pyx_v_created == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(8, 101, __pyx_L1_error)
  }
  __pyx_t_2 = (__Pyx_PySet_ContainsTF(((PyObject *)__pyx_v_self), __pyx_v_created, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 101, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (!__pyx_v_self->__pyx_base.__pyx_base.options->canonical_form_repeat_enum);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/enum.pxi":102
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         if self in created and not self.options.canonical_form_repeat_enum:
 *             return _CanonicalForm('"' + self.type + '"')             # <<<<<<<<<<<<<<
 *         created.add(self)
 *         return dict_to_canonical({
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyNumber_Add(__pyx_kp_u__48, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_kp_u__48); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_4);
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "src/enum.pxi":101
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         if self in created and not self.options.canonical_form_repeat_enum:             # <<<<<<<<<<<<<<
 *             return _CanonicalForm('"' + self.type + '"')
 *         created.add(self)
 */
  }

  /* "src/enum.pxi":103
 *         if self in created and not self.options.canonical_form_repeat_enum:
 *             return _CanonicalForm('"' + self.type + '"')
 *         created.add(self)             # <<<<<<<<<<<<<<
 *         return dict_to_canonical({
 *             'type': 'enum',
 */
  if (unlikely(__pyx_v_created == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
    __PYX_ERR(8, 103, __pyx_L1_error)
  }
  __pyx_t_5 = PySet_Add(__pyx_v_created, ((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(8, 103, __pyx_L1_error)

  /* "src/enum.pxi":104
 *             return _CanonicalForm('"' + self.type + '"')
 *         created.add(self)
 *         return dict_to_canonical({             # <<<<<<<<<<<<<<
 *             'type': 'enum',
 *             'name': self.type,
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);

  /* "src/enum.pxi":105
 *         created.add(self)
 *         return dict_to_canonical({
 *             'type': 'enum',             # <<<<<<<<<<<<<<
 *             'name': self.type,
 *             'symbols': self.symbols
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_type, __pyx_n_u_enum) < 0) __PYX_ERR(8, 105, __pyx_L1_error)

  /* "src/enum.pxi":106
 *         return dict_to_canonical({
 *             'type': 'enum',
 *             'name': self.type,             # <<<<<<<<<<<<<<
 *             'symbols': self.symbols
 *         })
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_name, __pyx_t_3) < 0) __PYX_ERR(8, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "src/enum.pxi":107
 *             'type': 'enum',
 *             'name': self.type,
 *             'symbols': self.symbols             # <<<<<<<<<<<<<<
 *         })
 * 
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_symbols, __pyx_v_self->symbols) < 0) __PYX_ERR(8, 105, __pyx_L1_error)

  /* "src/enum.pxi":104
 *             return _CanonicalForm('"' + self.type + '"')
 *         created.add(self)
 *         return dict_to_canonical({             # <<<<<<<<<<<<<<
 *             'type': 'enum',
 *             'name': self.type,
 */
  __pyx_t_3 = ((PyObject *)__pyx_f_5cavro_dict_to_canonical(__pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/enum.pxi":100
 *         return value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         if self in created and not self.options.canonical_form_repeat_enum:
 *             return _CanonicalForm('"' + self.type + '"')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.EnumType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_8EnumType_11_for_writer_2generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/enum.pxi":125
 *         new_type.reader_type = self
 *         new_type.writer_type = writer_enum
 *         new_type.symbols = tuple(s if s in reader_symbols else self.default_value for s in writer.symbols)             # <<<<<<<<<<<<<<
 *         new_type.unknown_symbols = writer_extra_symbols
 *         new_type.symbol_indexes = writer_enum.symbol_indexes
 */

static PyObject *__pyx_pf_5cavro_8EnumType_11_for_writer_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_5_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_5_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(8, 125, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_8EnumType_11_for_writer_2generator8, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_EnumType__for_writer_locals_gene, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(8, 125, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.EnumType._for_writer.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_8EnumType_11_for_writer_2generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(8, 125, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(8, 125, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 125, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(8, 125, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(8, 125, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(8, 125, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(8, 125, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(8, 125, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_reader_symbols)) { __Pyx_RaiseClosureNameError("reader_symbols"); __PYX_ERR(8, 125, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_reader_symbols == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(8, 125, __pyx_L1_error)
    }
    __pyx_t_5 = (__Pyx_PySet_ContainsTF(__pyx_cur_scope->__pyx_v_s, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_reader_symbols, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(8, 125, __pyx_L1_error)
    if (__pyx_t_5) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_s);
      __pyx_t_4 = __pyx_cur_scope->__pyx_v_s;
    } else {
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(8, 125, __pyx_L1_error) }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->default_value);
      __pyx_t_4 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->default_value;
    }
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(8, 125, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":110
 *         })
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         if not isinstance(writer, EnumType):
 *             return
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8EnumType__for_writer(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer) {
  struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *__pyx_cur_scope;
  struct __pyx_obj_5cavro_EnumType *__pyx_v_writer_enum = 0;
  PyObject *__pyx_v_writer_symbols = NULL;
  PyObject *__pyx_v_writer_extra_symbols = NULL;
  struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_new_type = 0;
  PyObject *__pyx_gb_5cavro_8EnumType_11_for_writer_2generator8 = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_clone_base __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_writer", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *)__pyx_tp_new_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer(__pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(8, 110, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "src/enum.pxi":111
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if not isinstance(writer, EnumType):             # <<<<<<<<<<<<<<
 *             return
 *         cdef EnumType writer_enum = writer
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_EnumType); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "src/enum.pxi":112
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if not isinstance(writer, EnumType):
 *             return             # <<<<<<<<<<<<<<
 *         cdef EnumType writer_enum = writer
 *         if not self.name_matches(writer_enum):
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/enum.pxi":111
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if not isinstance(writer, EnumType):             # <<<<<<<<<<<<<<
 *             return
 *         cdef EnumType writer_enum = writer
 */
  }

  /* "src/enum.pxi":113
 *         if not isinstance(writer, EnumType):
 *             return
 *         cdef EnumType writer_enum = writer             # <<<<<<<<<<<<<<
 *         if not self.name_matches(writer_enum):
 *             return
 */
  if (!(likely(((((PyObject *)__pyx_v_writer)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_EnumType))))) __PYX_ERR(8, 113, __pyx_L1_error)
  __pyx_t_3 = ((PyObject *)__pyx_v_writer);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_writer_enum = ((struct __pyx_obj_5cavro_EnumType *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/enum.pxi":114
 *             return
 *         cdef EnumType writer_enum = writer
 *         if not self.name_matches(writer_enum):             # <<<<<<<<<<<<<<
 *             return
 *         reader_symbols = set(self.symbols)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_EnumType *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.name_matches(((struct __pyx_obj_5cavro__NamedType *)__pyx_cur_scope->__pyx_v_self), ((struct __pyx_obj_5cavro__NamedType *)__pyx_v_writer_enum)); if (unlikely(PyErr_Occurred())) __PYX_ERR(8, 114, __pyx_L1_error)
  __pyx_t_1 = (!__pyx_t_2);
  if (__pyx_t_1) {

    /* "src/enum.pxi":115
 *         cdef EnumType writer_enum = writer
 *         if not self.name_matches(writer_enum):
 *             return             # <<<<<<<<<<<<<<
 *         reader_symbols = set(self.symbols)
 *         writer_symbols = set(writer_enum.symbols)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/enum.pxi":114
 *             return
 *         cdef EnumType writer_enum = writer
 *         if not self.name_matches(writer_enum):             # <<<<<<<<<<<<<<
 *             return
 *         reader_symbols = set(self.symbols)
 */
  }

  /* "src/enum.pxi":116
 *         if not self.name_matches(writer_enum):
 *             return
 *         reader_symbols = set(self.symbols)             # <<<<<<<<<<<<<<
 *         writer_symbols = set(writer_enum.symbols)
 *         writer_extra_symbols = writer_symbols - reader_symbols
 */
  __pyx_t_3 = PySet_New(__pyx_cur_scope->__pyx_v_self->symbols); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_reader_symbols = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/enum.pxi":117
 *             return
 *         reader_symbols = set(self.symbols)
 *         writer_symbols = set(writer_enum.symbols)             # <<<<<<<<<<<<<<
 *         writer_extra_symbols = writer_symbols - reader_symbols
 *         if not writer_extra_symbols:  # reader knows about all possible symbols
 */
  __pyx_t_3 = PySet_New(__pyx_v_writer_enum->symbols); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_writer_symbols = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/enum.pxi":118
 *         reader_symbols = set(self.symbols)
 *         writer_symbols = set(writer_enum.symbols)
 *         writer_extra_symbols = writer_symbols - reader_symbols             # <<<<<<<<<<<<<<
 *         if not writer_extra_symbols:  # reader knows about all possible symbols
 *             return self
 */
  __pyx_t_3 = PyNumber_Subtract(__pyx_v_writer_symbols, __pyx_cur_scope->__pyx_v_reader_symbols); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_writer_extra_symbols = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "src/enum.pxi":119
 *         writer_symbols = set(writer_enum.symbols)
 *         writer_extra_symbols = writer_symbols - reader_symbols
 *         if not writer_extra_symbols:  # reader knows about all possible symbols             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_writer_extra_symbols); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(8, 119, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "src/enum.pxi":120
 *         writer_extra_symbols = writer_symbols - reader_symbols
 *         if not writer_extra_symbols:  # reader knows about all possible symbols
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         cdef PromotingEnumType new_type = writer.clone_base(PromotingEnumType)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_cur_scope->__pyx_v_self);
    goto __pyx_L0;

    /* "src/enum.pxi":119
 *         writer_symbols = set(writer_enum.symbols)
 *         writer_extra_symbols = writer_symbols - reader_symbols
 *         if not writer_extra_symbols:  # reader knows about all possible symbols             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "src/enum.pxi":122
 *             return self
 * 
 *         cdef PromotingEnumType new_type = writer.clone_base(PromotingEnumType)             # <<<<<<<<<<<<<<
 *         new_type.reader_type = self
 *         new_type.writer_type = writer_enum
 */
  __pyx_t_4.__pyx_n = 1;
  __pyx_t_4.cls = ((PyObject *)__pyx_ptype_5cavro_PromotingEnumType);
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_writer->__pyx_vtab)->clone_base(__pyx_v_writer, &__pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_PromotingEnumType))))) __PYX_ERR(8, 122, __pyx_L1_error)
  __pyx_v_new_type = ((struct __pyx_obj_5cavro_PromotingEnumType *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/enum.pxi":123
 * 
 *         cdef PromotingEnumType new_type = writer.clone_base(PromotingEnumType)
 *         new_type.reader_type = self             # <<<<<<<<<<<<<<
 *         new_type.writer_type = writer_enum
 *         new_type.symbols = tuple(s if s in reader_symbols else self.default_value for s in writer.symbols)
 */
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GOTREF((PyObject *)__pyx_v_new_type->reader_type);
  __Pyx_DECREF((PyObject *)__pyx_v_new_type->reader_type);
  __pyx_v_new_type->reader_type = __pyx_cur_scope->__pyx_v_self;

  /* "src/enum.pxi":124
 *         cdef PromotingEnumType new_type = writer.clone_base(PromotingEnumType)
 *         new_type.reader_type = self
 *         new_type.writer_type = writer_enum             # <<<<<<<<<<<<<<
 *         new_type.symbols = tuple(s if s in reader_symbols else self.default_value for s in writer.symbols)
 *         new_type.unknown_symbols = writer_extra_symbols
 */
  __Pyx_INCREF((PyObject *)__pyx_v_writer_enum);
  __Pyx_GIVEREF((PyObject *)__pyx_v_writer_enum);
  __Pyx_GOTREF((PyObject *)__pyx_v_new_type->writer_type);
  __Pyx_DECREF((PyObject *)__pyx_v_new_type->writer_type);
  __pyx_v_new_type->writer_type = __pyx_v_writer_enum;

  /* "src/enum.pxi":125
 *         new_type.reader_type = self
 *         new_type.writer_type = writer_enum
 *         new_type.symbols = tuple(s if s in reader_symbols else self.default_value for s in writer.symbols)             # <<<<<<<<<<<<<<
 *         new_type.unknown_symbols = writer_extra_symbols
 *         new_type.symbol_indexes = writer_enum.symbol_indexes
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_writer), __pyx_n_s_symbols); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __pyx_pf_5cavro_8EnumType_11_for_writer_genexpr(((PyObject*)__pyx_cur_scope), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_new_type->__pyx_base.symbols);
  __Pyx_DECREF(__pyx_v_new_type->__pyx_base.symbols);
  __pyx_v_new_type->__pyx_base.symbols = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/enum.pxi":126
 *         new_type.writer_type = writer_enum
 *         new_type.symbols = tuple(s if s in reader_symbols else self.default_value for s in writer.symbols)
 *         new_type.unknown_symbols = writer_extra_symbols             # <<<<<<<<<<<<<<
 *         new_type.symbol_indexes = writer_enum.symbol_indexes
 *         return new_type
 */
  if (!(likely(PySet_CheckExact(__pyx_v_writer_extra_symbols)) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_v_writer_extra_symbols))) __PYX_ERR(8, 126, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_writer_extra_symbols;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_new_type->unknown_symbols);
  __Pyx_DECREF(__pyx_v_new_type->unknown_symbols);
  __pyx_v_new_type->unknown_symbols = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/enum.pxi":127
 *         new_type.symbols = tuple(s if s in reader_symbols else self.default_value for s in writer.symbols)
 *         new_type.unknown_symbols = writer_extra_symbols
 *         new_type.symbol_indexes = writer_enum.symbol_indexes             # <<<<<<<<<<<<<<
 *         return new_type
 * 
 */
  __pyx_t_3 = __pyx_v_writer_enum->symbol_indexes;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_new_type->__pyx_base.symbol_indexes);
  __Pyx_DECREF(__pyx_v_new_type->__pyx_base.symbol_indexes);
  __pyx_v_new_type->__pyx_base.symbol_indexes = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/enum.pxi":128
 *         new_type.unknown_symbols = writer_extra_symbols
 *         new_type.symbol_indexes = writer_enum.symbol_indexes
 *         return new_type             # <<<<<<<<<<<<<<
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_new_type);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_new_type);
  goto __pyx_L0;

  /* "src/enum.pxi":110
 *         })
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         if not isinstance(writer, EnumType):
 *             return
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.EnumType._for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_writer_enum);
  __Pyx_XDECREF(__pyx_v_writer_symbols);
  __Pyx_XDECREF(__pyx_v_writer_extra_symbols);
  __Pyx_XDECREF((PyObject *)__pyx_v_new_type);
  __Pyx_XDECREF(__pyx_gb_5cavro_8EnumType_11_for_writer_2generator8);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":130
 *         return new_type
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):             # <<<<<<<<<<<<<<
 *         if self.options.string_types_default_unchanged:
 *             return schema_default
 */

static PyObject *__pyx_f_5cavro_8EnumType_resolve_default_value(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v_schema_default, PyObject *__pyx_v_field) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resolve_default_value", 1);

  /* "src/enum.pxi":131
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if self.options.string_types_default_unchanged:             # <<<<<<<<<<<<<<
 *             return schema_default
 *         return AvroType.resolve_default_value(self, schema_default, field)
 */
  if (__pyx_v_self->__pyx_base.__pyx_base.options->string_types_default_unchanged) {

    /* "src/enum.pxi":132
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if self.options.string_types_default_unchanged:
 *             return schema_default             # <<<<<<<<<<<<<<
 *         return AvroType.resolve_default_value(self, schema_default, field)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_schema_default);
    __pyx_r = __pyx_v_schema_default;
    goto __pyx_L0;

    /* "src/enum.pxi":131
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if self.options.string_types_default_unchanged:             # <<<<<<<<<<<<<<
 *             return schema_default
 *         return AvroType.resolve_default_value(self, schema_default, field)
 */
  }

  /* "src/enum.pxi":133
 *         if self.options.string_types_default_unchanged:
 *             return schema_default
 *         return AvroType.resolve_default_value(self, schema_default, field)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8AvroType_resolve_default_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_schema_default, __pyx_v_field); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/enum.pxi":130
 *         return new_type
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):             # <<<<<<<<<<<<<<
 *         if self.options.string_types_default_unchanged:
 *             return schema_default
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.EnumType.resolve_default_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":6
 *     type_name = 'enum'
 * 
 *     cdef readonly tuple symbols             # <<<<<<<<<<<<<<
 *     cdef dict symbol_indexes
 *     cdef readonly object default_value
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8EnumType_7symbols_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_8EnumType_7symbols_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_8EnumType_7symbols___get__(((struct __pyx_obj_5cavro_EnumType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8EnumType_7symbols___get__(struct __pyx_obj_5cavro_EnumType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->symbols);
  __pyx_r = __pyx_v_self->symbols;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":8
 *     cdef readonly tuple symbols
 *     cdef dict symbol_indexes
 *     cdef readonly object default_value             # <<<<<<<<<<<<<<
 *     cdef readonly str doc
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8EnumType_13default_value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_8EnumType_13default_value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_8EnumType_13default_value___get__(((struct __pyx_obj_5cavro_EnumType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8EnumType_13default_value___get__(struct __pyx_obj_5cavro_EnumType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->default_value);
  __pyx_r = __pyx_v_self->default_value;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":9
 *     cdef dict symbol_indexes
 *     cdef readonly object default_value
 *     cdef readonly str doc             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, schema, source, namespace):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8EnumType_3doc_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_8EnumType_3doc_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_8EnumType_3doc___get__(((struct __pyx_obj_5cavro_EnumType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8EnumType_3doc___get__(struct __pyx_obj_5cavro_EnumType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->doc);
  __pyx_r = __pyx_v_self->doc;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8EnumType_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8EnumType_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8EnumType_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8EnumType_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_8EnumType_8__reduce_cython__(((struct __pyx_obj_5cavro_EnumType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8EnumType_8__reduce_cython__(struct __pyx_obj_5cavro_EnumType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.aliases, self.default_value, self.doc, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.symbol_indexes, self.symbols, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(11); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.aliases);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.aliases);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.aliases)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->default_value);
  __Pyx_GIVEREF(__pyx_v_self->default_value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->default_value)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->doc);
  __Pyx_GIVEREF(__pyx_v_self->doc);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->doc)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.effective_namespace);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.effective_namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->__pyx_base.effective_namespace)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_self->__pyx_base.__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.name);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_v_self->__pyx_base.name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.namespace);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_v_self->__pyx_base.namespace)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 7, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->symbol_indexes);
  __Pyx_GIVEREF(__pyx_v_self->symbol_indexes);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_v_self->symbol_indexes)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->symbols);
  __Pyx_GIVEREF(__pyx_v_self->symbols);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_v_self->symbols)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_v_self->__pyx_base.__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.aliases, self.default_value, self.doc, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.symbol_indexes, self.symbols, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.aliases, self.default_value, self.doc, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.symbol_indexes, self.symbols, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.symbol_indexes is not None or self.symbols is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.aliases, self.default_value, self.doc, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.symbol_indexes, self.symbols, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.symbol_indexes is not None or self.symbols is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_EnumType, (type(self), 0x8f63b31, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base.aliases != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->default_value != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->doc != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.effective_namespace != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.name != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.namespace != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->symbol_indexes != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->symbols != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.symbol_indexes is not None or self.symbols is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_EnumType, (type(self), 0x8f63b31, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.symbol_indexes is not None or self.symbols is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_EnumType, (type(self), 0x8f63b31, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_EnumType, (type(self), 0x8f63b31, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_EnumType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_150354737);
    __Pyx_GIVEREF(__pyx_int_150354737);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_150354737)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.symbol_indexes is not None or self.symbols is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_EnumType, (type(self), 0x8f63b31, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_EnumType, (type(self), 0x8f63b31, None), state
 *     else:
 *         return __pyx_unpickle_EnumType, (type(self), 0x8f63b31, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_EnumType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_EnumType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_150354737);
    __Pyx_GIVEREF(__pyx_int_150354737);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_150354737)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.EnumType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EnumType, (type(self), 0x8f63b31, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EnumType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8EnumType_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8EnumType_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8EnumType_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8EnumType_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.EnumType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8EnumType_10__setstate_cython__(((struct __pyx_obj_5cavro_EnumType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8EnumType_10__setstate_cython__(struct __pyx_obj_5cavro_EnumType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_EnumType, (type(self), 0x8f63b31, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_EnumType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_EnumType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EnumType, (type(self), 0x8f63b31, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EnumType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.EnumType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":141
 *     cdef readonly EnumType writer_type
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         value = EnumType._json_decode(self, value)
 *         if value in self.unknown_symbols:
 */

static PyObject *__pyx_f_5cavro_17PromotingEnumType__json_decode(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/enum.pxi":142
 * 
 *     cdef _json_decode(self, value):
 *         value = EnumType._json_decode(self, value)             # <<<<<<<<<<<<<<
 *         if value in self.unknown_symbols:
 *             if self.defaul_value is NO_DEFAULT:
 */
  __pyx_t_1 = __pyx_f_5cavro_8EnumType__json_decode(((struct __pyx_obj_5cavro_EnumType *)__pyx_v_self), __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/enum.pxi":143
 *     cdef _json_decode(self, value):
 *         value = EnumType._json_decode(self, value)
 *         if value in self.unknown_symbols:             # <<<<<<<<<<<<<<
 *             if self.defaul_value is NO_DEFAULT:
 *                 raise CannotPromote(self.reader_type, self.writer_type, f"'{value}' invalid for enum")
 */
  if (unlikely(__pyx_v_self->unknown_symbols == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(8, 143, __pyx_L1_error)
  }
  __pyx_t_2 = (__Pyx_PySet_ContainsTF(__pyx_v_value, __pyx_v_self->unknown_symbols, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(8, 143, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "src/enum.pxi":144
 *         value = EnumType._json_decode(self, value)
 *         if value in self.unknown_symbols:
 *             if self.defaul_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                 raise CannotPromote(self.reader_type, self.writer_type, f"'{value}' invalid for enum")
 *             return self.default_value
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_defaul_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = (__pyx_t_1 == __pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__pyx_t_2)) {

      /* "src/enum.pxi":145
 *         if value in self.unknown_symbols:
 *             if self.defaul_value is NO_DEFAULT:
 *                 raise CannotPromote(self.reader_type, self.writer_type, f"'{value}' invalid for enum")             # <<<<<<<<<<<<<<
 *             return self.default_value
 *         return value
 */
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = 127;
      __Pyx_INCREF(__pyx_kp_u__6);
      __pyx_t_4 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__6);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u__6);
      __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_5;
      __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_INCREF(__pyx_kp_u_invalid_for_enum);
      __pyx_t_4 += 18;
      __Pyx_GIVEREF(__pyx_kp_u_invalid_for_enum);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_invalid_for_enum);
      __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF((PyObject *)__pyx_v_self->reader_type);
      __Pyx_GIVEREF((PyObject *)__pyx_v_self->reader_type);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self->reader_type))) __PYX_ERR(8, 145, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_self->writer_type);
      __Pyx_GIVEREF((PyObject *)__pyx_v_self->writer_type);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self->writer_type))) __PYX_ERR(8, 145, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_1)) __PYX_ERR(8, 145, __pyx_L1_error);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_CannotPromote), __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(8, 145, __pyx_L1_error)

      /* "src/enum.pxi":144
 *         value = EnumType._json_decode(self, value)
 *         if value in self.unknown_symbols:
 *             if self.defaul_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                 raise CannotPromote(self.reader_type, self.writer_type, f"'{value}' invalid for enum")
 *             return self.default_value
 */
    }

    /* "src/enum.pxi":146
 *             if self.defaul_value is NO_DEFAULT:
 *                 raise CannotPromote(self.reader_type, self.writer_type, f"'{value}' invalid for enum")
 *             return self.default_value             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->__pyx_base.default_value);
    __pyx_r = __pyx_v_self->__pyx_base.default_value;
    goto __pyx_L0;

    /* "src/enum.pxi":143
 *     cdef _json_decode(self, value):
 *         value = EnumType._json_decode(self, value)
 *         if value in self.unknown_symbols:             # <<<<<<<<<<<<<<
 *             if self.defaul_value is NO_DEFAULT:
 *                 raise CannotPromote(self.reader_type, self.writer_type, f"'{value}' invalid for enum")
 */
  }

  /* "src/enum.pxi":147
 *                 raise CannotPromote(self.reader_type, self.writer_type, f"'{value}' invalid for enum")
 *             return self.default_value
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/enum.pxi":141
 *     cdef readonly EnumType writer_type
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         value = EnumType._json_decode(self, value)
 *         if value in self.unknown_symbols:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.PromotingEnumType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":149
 *         return value
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         value = EnumType._binary_buffer_decode(self, buffer)
 *         if value is NO_DEFAULT:
 */

static PyObject *__pyx_f_5cavro_17PromotingEnumType__binary_buffer_decode(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/enum.pxi":150
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         value = EnumType._binary_buffer_decode(self, buffer)             # <<<<<<<<<<<<<<
 *         if value is NO_DEFAULT:
 *             raise CannotPromoteError(self.reader_type, self.writer_type, f"Unknown value for enum")
 */
  __pyx_t_1 = __pyx_f_5cavro_8EnumType__binary_buffer_decode(((struct __pyx_obj_5cavro_EnumType *)__pyx_v_self), __pyx_v_buffer); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/enum.pxi":151
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         value = EnumType._binary_buffer_decode(self, buffer)
 *         if value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             raise CannotPromoteError(self.reader_type, self.writer_type, f"Unknown value for enum")
 *         return value
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_v_value == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "src/enum.pxi":152
 *         value = EnumType._binary_buffer_decode(self, buffer)
 *         if value is NO_DEFAULT:
 *             raise CannotPromoteError(self.reader_type, self.writer_type, f"Unknown value for enum")             # <<<<<<<<<<<<<<
 *         return value
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CannotPromoteError); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_4, ((PyObject *)__pyx_v_self->reader_type), ((PyObject *)__pyx_v_self->writer_type), __pyx_kp_u_Unknown_value_for_enum};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(8, 152, __pyx_L1_error)

    /* "src/enum.pxi":151
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         value = EnumType._binary_buffer_decode(self, buffer)
 *         if value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             raise CannotPromoteError(self.reader_type, self.writer_type, f"Unknown value for enum")
 *         return value
 */
  }

  /* "src/enum.pxi":153
 *         if value is NO_DEFAULT:
 *             raise CannotPromoteError(self.reader_type, self.writer_type, f"Unknown value for enum")
 *         return value             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/enum.pxi":149
 *         return value
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         value = EnumType._binary_buffer_decode(self, buffer)
 *         if value is NO_DEFAULT:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.PromotingEnumType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":137
 * 
 * cdef class PromotingEnumType(EnumType):
 *     cdef readonly set unknown_symbols             # <<<<<<<<<<<<<<
 *     cdef readonly EnumType reader_type
 *     cdef readonly EnumType writer_type
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_17PromotingEnumType_15unknown_symbols_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_17PromotingEnumType_15unknown_symbols_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_17PromotingEnumType_15unknown_symbols___get__(((struct __pyx_obj_5cavro_PromotingEnumType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_17PromotingEnumType_15unknown_symbols___get__(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->unknown_symbols);
  __pyx_r = __pyx_v_self->unknown_symbols;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":138
 * cdef class PromotingEnumType(EnumType):
 *     cdef readonly set unknown_symbols
 *     cdef readonly EnumType reader_type             # <<<<<<<<<<<<<<
 *     cdef readonly EnumType writer_type
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_17PromotingEnumType_11reader_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_17PromotingEnumType_11reader_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_17PromotingEnumType_11reader_type___get__(((struct __pyx_obj_5cavro_PromotingEnumType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_17PromotingEnumType_11reader_type___get__(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->reader_type);
  __pyx_r = ((PyObject *)__pyx_v_self->reader_type);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/enum.pxi":139
 *     cdef readonly set unknown_symbols
 *     cdef readonly EnumType reader_type
 *     cdef readonly EnumType writer_type             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_17PromotingEnumType_11writer_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_17PromotingEnumType_11writer_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_17PromotingEnumType_11writer_type___get__(((struct __pyx_obj_5cavro_PromotingEnumType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_17PromotingEnumType_11writer_type___get__(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->writer_type);
  __pyx_r = ((PyObject *)__pyx_v_self->writer_type);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_17PromotingEnumType_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_17PromotingEnumType_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_17PromotingEnumType_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_17PromotingEnumType_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_17PromotingEnumType___reduce_cython__(((struct __pyx_obj_5cavro_PromotingEnumType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_17PromotingEnumType___reduce_cython__(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.aliases, self.default_value, self.doc, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.reader_type, self.symbol_indexes, self.symbols, self.unknown_symbols, self.value_adapters, self.writer_type)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(14); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.aliases);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.aliases);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.__pyx_base.aliases)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.default_value);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.default_value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->__pyx_base.default_value)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.doc);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.doc);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->__pyx_base.doc)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.effective_namespace);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.effective_namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->__pyx_base.__pyx_base.effective_namespace)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.name);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_v_self->__pyx_base.__pyx_base.name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.namespace);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_v_self->__pyx_base.__pyx_base.namespace)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 7, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->reader_type);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->reader_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 8, ((PyObject *)__pyx_v_self->reader_type))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.symbol_indexes);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.symbol_indexes);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_v_self->__pyx_base.symbol_indexes)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.symbols);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.symbols);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_v_self->__pyx_base.symbols)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->unknown_symbols);
  __Pyx_GIVEREF(__pyx_v_self->unknown_symbols);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_v_self->unknown_symbols)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->writer_type);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->writer_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 13, ((PyObject *)__pyx_v_self->writer_type))) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.aliases, self.default_value, self.doc, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.reader_type, self.symbol_indexes, self.symbols, self.unknown_symbols, self.value_adapters, self.writer_type)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.aliases, self.default_value, self.doc, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.reader_type, self.symbol_indexes, self.symbols, self.unknown_symbols, self.value_adapters, self.writer_type)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.reader_type is not None or self.symbol_indexes is not None or self.symbols is not None or self.unknown_symbols is not None or self.value_adapters is not None or self.writer_type is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.aliases, self.default_value, self.doc, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.reader_type, self.symbol_indexes, self.symbols, self.unknown_symbols, self.value_adapters, self.writer_type)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.reader_type is not None or self.symbol_indexes is not None or self.symbols is not None or self.unknown_symbols is not None or self.value_adapters is not None or self.writer_type is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PromotingEnumType, (type(self), 0x8a61ff8, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base.__pyx_base.aliases != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.default_value != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.doc != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.__pyx_base.effective_namespace != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.__pyx_base.name != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.__pyx_base.namespace != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->reader_type) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.symbol_indexes != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.symbols != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->unknown_symbols != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.value_adapters != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->writer_type) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.reader_type is not None or self.symbol_indexes is not None or self.symbols is not None or self.unknown_symbols is not None or self.value_adapters is not None or self.writer_type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PromotingEnumType, (type(self), 0x8a61ff8, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.reader_type is not None or self.symbol_indexes is not None or self.symbols is not None or self.unknown_symbols is not None or self.value_adapters is not None or self.writer_type is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PromotingEnumType, (type(self), 0x8a61ff8, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PromotingEnumType, (type(self), 0x8a61ff8, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_PromotingEnumType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_145104888);
    __Pyx_GIVEREF(__pyx_int_145104888);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_145104888)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.reader_type is not None or self.symbol_indexes is not None or self.symbols is not None or self.unknown_symbols is not None or self.value_adapters is not None or self.writer_type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PromotingEnumType, (type(self), 0x8a61ff8, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PromotingEnumType, (type(self), 0x8a61ff8, None), state
 *     else:
 *         return __pyx_unpickle_PromotingEnumType, (type(self), 0x8a61ff8, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PromotingEnumType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_PromotingEnumType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_145104888);
    __Pyx_GIVEREF(__pyx_int_145104888);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_145104888)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.PromotingEnumType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromotingEnumType, (type(self), 0x8a61ff8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromotingEnumType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_17PromotingEnumType_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_17PromotingEnumType_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_17PromotingEnumType_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_17PromotingEnumType_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.PromotingEnumType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_17PromotingEnumType_2__setstate_cython__(((struct __pyx_obj_5cavro_PromotingEnumType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_17PromotingEnumType_2__setstate_cython__(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PromotingEnumType, (type(self), 0x8a61ff8, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PromotingEnumType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_PromotingEnumType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromotingEnumType, (type(self), 0x8a61ff8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromotingEnumType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.PromotingEnumType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":9
 *     cdef readonly AvroType value_type
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         super().__init__(schema, source, namespace)
 *         self.key_type = StringType(schema, {'type': 'string'}, namespace) # This way we pick up options
 */

/* Python wrapper */
static int __pyx_pw_5cavro_7MapType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_7MapType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_namespace = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_schema,&__pyx_n_s_source,&__pyx_n_s_namespace,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(9, 9, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(9, 9, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(9, 9, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_namespace)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(9, 9, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(9, 9, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(9, 9, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_schema = values[0];
    __pyx_v_source = values[1];
    __pyx_v_namespace = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(9, 9, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.MapType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_7MapType___init__(((struct __pyx_obj_5cavro_MapType *)__pyx_v_self), __pyx_v_schema, __pyx_v_source, __pyx_v_namespace);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_7MapType___init__(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/map.pxi":10
 * 
 *     def __init__(self, schema, source, namespace):
 *         super().__init__(schema, source, namespace)             # <<<<<<<<<<<<<<
 *         self.key_type = StringType(schema, {'type': 'string'}, namespace) # This way we pick up options
 *         self.value_type = AvroType.for_source(schema, source['values'], namespace)
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_MapType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_MapType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_5cavro_MapType))) __PYX_ERR(9, 10, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(9, 10, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_schema, __pyx_v_source, __pyx_v_namespace};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/map.pxi":11
 *     def __init__(self, schema, source, namespace):
 *         super().__init__(schema, source, namespace)
 *         self.key_type = StringType(schema, {'type': 'string'}, namespace) # This way we pick up options             # <<<<<<<<<<<<<<
 *         self.value_type = AvroType.for_source(schema, source['values'], namespace)
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_type, __pyx_n_u_string) < 0) __PYX_ERR(9, 11, __pyx_L1_error)
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_schema);
  __Pyx_GIVEREF(__pyx_v_schema);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_schema)) __PYX_ERR(9, 11, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1)) __PYX_ERR(9, 11, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_namespace);
  __Pyx_GIVEREF(__pyx_v_namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_namespace)) __PYX_ERR(9, 11, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_StringType), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->key_type);
  __Pyx_DECREF((PyObject *)__pyx_v_self->key_type);
  __pyx_v_self->key_type = ((struct __pyx_obj_5cavro_StringType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":12
 *         super().__init__(schema, source, namespace)
 *         self.key_type = StringType(schema, {'type': 'string'}, namespace) # This way we pick up options
 *         self.value_type = AvroType.for_source(schema, source['values'], namespace)             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_AvroType), __pyx_n_s_for_source); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_source, __pyx_n_u_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_v_schema, __pyx_t_3, __pyx_v_namespace};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(9, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->value_type);
  __Pyx_DECREF((PyObject *)__pyx_v_self->value_type);
  __pyx_v_self->value_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":9
 *     cdef readonly AvroType value_type
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         super().__init__(schema, source, namespace)
 *         self.key_type = StringType(schema, {'type': 'string'}, namespace) # This way we pick up options
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.MapType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":14
 *         self.value_type = AvroType.for_source(schema, source['values'], namespace)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef MapType new_inst = self.clone_base()
 *         new_inst.key_type = self.key_type
 */

static PyObject *__pyx_pw_5cavro_7MapType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_7MapType_copy(struct __pyx_obj_5cavro_MapType *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_MapType *__pyx_v_new_inst = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_7MapType_3copy)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 14, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(9, 14, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/map.pxi":15
 * 
 *     cpdef AvroType copy(self):
 *         cdef MapType new_inst = self.clone_base()             # <<<<<<<<<<<<<<
 *         new_inst.key_type = self.key_type
 *         new_inst.value_type = self.value_type
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_MapType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_MapType))))) __PYX_ERR(9, 15, __pyx_L1_error)
  __pyx_v_new_inst = ((struct __pyx_obj_5cavro_MapType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":16
 *     cpdef AvroType copy(self):
 *         cdef MapType new_inst = self.clone_base()
 *         new_inst.key_type = self.key_type             # <<<<<<<<<<<<<<
 *         new_inst.value_type = self.value_type
 *         return new_inst
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->key_type);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_new_inst->key_type);
  __Pyx_DECREF((PyObject *)__pyx_v_new_inst->key_type);
  __pyx_v_new_inst->key_type = ((struct __pyx_obj_5cavro_StringType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":17
 *         cdef MapType new_inst = self.clone_base()
 *         new_inst.key_type = self.key_type
 *         new_inst.value_type = self.value_type             # <<<<<<<<<<<<<<
 *         return new_inst
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->value_type);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_new_inst->value_type);
  __Pyx_DECREF((PyObject *)__pyx_v_new_inst->value_type);
  __pyx_v_new_inst->value_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":18
 *         new_inst.key_type = self.key_type
 *         new_inst.value_type = self.value_type
 *         return new_inst             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_new_inst);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_new_inst);
  goto __pyx_L0;

  /* "src/map.pxi":14
 *         self.value_type = AvroType.for_source(schema, source['values'], namespace)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef MapType new_inst = self.clone_base()
 *         new_inst.key_type = self.key_type
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.MapType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_new_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7MapType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7MapType_3copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7MapType_3copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7MapType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_7MapType_2copy(((struct __pyx_obj_5cavro_MapType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7MapType_2copy(struct __pyx_obj_5cavro_MapType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_7MapType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.MapType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":20
 *         return new_inst
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type', 'values'})
 * 
 */

static PyObject *__pyx_f_5cavro_7MapType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/map.pxi":21
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {'type', 'values'})             # <<<<<<<<<<<<<<
 * 
 *     def walk_types(self, visited):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(9, 21, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_values) < 0) __PYX_ERR(9, 21, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/map.pxi":20
 *         return new_inst
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type', 'values'})
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.MapType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_7MapType_6generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/map.pxi":23
 *         return _strip_keys(dict(source), {'type', 'values'})
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7MapType_5walk_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7MapType_5walk_types = {"walk_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7MapType_5walk_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7MapType_5walk_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_visited = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("walk_types (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_visited,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_visited)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(9, 23, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "walk_types") < 0)) __PYX_ERR(9, 23, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_visited = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("walk_types", 1, 1, 1, __pyx_nargs); __PYX_ERR(9, 23, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.MapType.walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_7MapType_4walk_types(((struct __pyx_obj_5cavro_MapType *)__pyx_v_self), __pyx_v_visited);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7MapType_4walk_types(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_visited) {
  struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("walk_types", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types *)__pyx_tp_new_5cavro___pyx_scope_struct_6_walk_types(__pyx_ptype_5cavro___pyx_scope_struct_6_walk_types, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(9, 23, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_visited = __pyx_v_visited;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_visited);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_visited);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_7MapType_6generator2, __pyx_codeobj__49, (PyObject *) __pyx_cur_scope, __pyx_n_s_walk_types, __pyx_n_s_MapType_walk_types, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(9, 23, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.MapType.walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_7MapType_6generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("walk_types", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L5_resume_from_yield_from;
    case 2: goto __pyx_L6_resume_from_yield_from;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(9, 23, __pyx_L1_error)

  /* "src/map.pxi":24
 * 
 *     def walk_types(self, visited):
 *         if self in visited:             # <<<<<<<<<<<<<<
 *             return
 *         yield from super().walk_types(visited)
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_visited, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(9, 24, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/map.pxi":25
 *     def walk_types(self, visited):
 *         if self in visited:
 *             return             # <<<<<<<<<<<<<<
 *         yield from super().walk_types(visited)
 *         yield from self.value_type.walk_types(visited)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "src/map.pxi":24
 * 
 *     def walk_types(self, visited):
 *         if self in visited:             # <<<<<<<<<<<<<<
 *             return
 *         yield from super().walk_types(visited)
 */
  }

  /* "src/map.pxi":26
 *         if self in visited:
 *             return
 *         yield from super().walk_types(visited)             # <<<<<<<<<<<<<<
 *         yield from self.value_type.walk_types(visited)
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_MapType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_MapType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_ptype_5cavro_MapType))) __PYX_ERR(9, 26, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_cur_scope->__pyx_v_self))) __PYX_ERR(9, 26, __pyx_L1_error);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_walk_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_visited};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 26, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __Pyx_Generator_Yield_From(__pyx_generator, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L5_resume_from_yield_from:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(9, 26, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(9, 26, __pyx_L1_error)
    }
  }

  /* "src/map.pxi":27
 *             return
 *         yield from super().walk_types(visited)
 *         yield from self.value_type.walk_types(visited)             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->value_type), __pyx_n_s_walk_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_visited};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 27, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __Pyx_Generator_Yield_From(__pyx_generator, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 2;
    return __pyx_r;
    __pyx_L6_resume_from_yield_from:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(9, 27, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(9, 27, __pyx_L1_error)
    }
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "src/map.pxi":23
 *         return _strip_keys(dict(source), {'type', 'values'})
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":29
 *         yield from self.value_type.walk_types(visited)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {'values': self.value_type.get_schema(created)}
 * 
 */

static PyObject *__pyx_pw_5cavro_7MapType_8_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_7MapType__get_schema_extra(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_created, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  struct __pyx_opt_args_5cavro_8AvroType_get_schema __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_schema_extra); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 29, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_7MapType_8_get_schema_extra)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_created};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 29, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_2))) __PYX_ERR(9, 29, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/map.pxi":30
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         return {'values': self.value_type.get_schema(created)}             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6.__pyx_n = 1;
  __pyx_t_6.created = __pyx_v_created;
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->value_type->__pyx_vtab)->get_schema(__pyx_v_self->value_type, 0, &__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_values, __pyx_t_2) < 0) __PYX_ERR(9, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/map.pxi":29
 *         yield from self.value_type.walk_types(visited)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {'values': self.value_type.get_schema(created)}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.MapType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7MapType_8_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7MapType_8_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7MapType_8_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7MapType_8_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(9, 29, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(9, 29, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(9, 29, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.MapType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(9, 29, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_7MapType_7_get_schema_extra(((struct __pyx_obj_5cavro_MapType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7MapType_7_get_schema_extra(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_7MapType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.MapType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":32
 *         return {'values': self.value_type.get_schema(created)}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         if hasattr(value, 'items'):
 *             value = value.items()
 */

static int __pyx_f_5cavro_7MapType__binary_buffer_encode(struct __pyx_obj_5cavro_MapType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_item_value = NULL;
  PyObject *__pyx_v_e = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  char const *__pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/map.pxi":33
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         if hasattr(value, 'items'):             # <<<<<<<<<<<<<<
 *             value = value.items()
 *         zigzag_encode_long(buffer, len(value))
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_value, __pyx_n_u_items); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(9, 33, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/map.pxi":34
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         if hasattr(value, 'items'):
 *             value = value.items()             # <<<<<<<<<<<<<<
 *         zigzag_encode_long(buffer, len(value))
 *         if value:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 34, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/map.pxi":33
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         if hasattr(value, 'items'):             # <<<<<<<<<<<<<<
 *             value = value.items()
 *         zigzag_encode_long(buffer, len(value))
 */
  }

  /* "src/map.pxi":35
 *         if hasattr(value, 'items'):
 *             value = value.items()
 *         zigzag_encode_long(buffer, len(value))             # <<<<<<<<<<<<<<
 *         if value:
 *             for key, item_value in value:
 */
  __pyx_t_6 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(9, 35, __pyx_L1_error)
  __pyx_t_5 = __pyx_f_5cavro_zigzag_encode_long(__pyx_v_buffer, __pyx_t_6); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(9, 35, __pyx_L1_error)

  /* "src/map.pxi":36
 *             value = value.items()
 *         zigzag_encode_long(buffer, len(value))
 *         if value:             # <<<<<<<<<<<<<<
 *             for key, item_value in value:
 *                 self.key_type.binary_buffer_encode(buffer, key)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(9, 36, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/map.pxi":37
 *         zigzag_encode_long(buffer, len(value))
 *         if value:
 *             for key, item_value in value:             # <<<<<<<<<<<<<<
 *                 self.key_type.binary_buffer_encode(buffer, key)
 *                 try:
 */
    if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
      __pyx_t_2 = __pyx_v_value; __Pyx_INCREF(__pyx_t_2);
      __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 37, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(9, 37, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(9, 37, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(9, 37, __pyx_L1_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 37, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(9, 37, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(9, 37, __pyx_L1_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 37, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(9, 37, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
        PyObject* sequence = __pyx_t_3;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(9, 37, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 37, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(9, 37, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(9, 37, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_9);
        index = 0; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(9, 37, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L8_unpacking_done;
        __pyx_L7_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(9, 37, __pyx_L1_error)
        __pyx_L8_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_item_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "src/map.pxi":38
 *         if value:
 *             for key, item_value in value:
 *                 self.key_type.binary_buffer_encode(buffer, key)             # <<<<<<<<<<<<<<
 *                 try:
 *                     self.value_type.binary_buffer_encode(buffer, item_value)
 */
      __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_StringType *)__pyx_v_self->key_type->__pyx_base.__pyx_vtab)->__pyx_base.binary_buffer_encode(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self->key_type), __pyx_v_buffer, __pyx_v_key); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(9, 38, __pyx_L1_error)

      /* "src/map.pxi":39
 *             for key, item_value in value:
 *                 self.key_type.binary_buffer_encode(buffer, key)
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.value_type.binary_buffer_encode(buffer, item_value)
 *                 except InvalidValue as e:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_13);
        /*try:*/ {

          /* "src/map.pxi":40
 *                 self.key_type.binary_buffer_encode(buffer, key)
 *                 try:
 *                     self.value_type.binary_buffer_encode(buffer, item_value)             # <<<<<<<<<<<<<<
 *                 except InvalidValue as e:
 *                     if self.options.invalid_value_include_map_key:
 */
          __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->value_type->__pyx_vtab)->binary_buffer_encode(__pyx_v_self->value_type, __pyx_v_buffer, __pyx_v_item_value); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(9, 40, __pyx_L9_error)

          /* "src/map.pxi":39
 *             for key, item_value in value:
 *                 self.key_type.binary_buffer_encode(buffer, key)
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.value_type.binary_buffer_encode(buffer, item_value)
 *                 except InvalidValue as e:
 */
        }
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        goto __pyx_L16_try_end;
        __pyx_L9_error:;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "src/map.pxi":41
 *                 try:
 *                     self.value_type.binary_buffer_encode(buffer, item_value)
 *                 except InvalidValue as e:             # <<<<<<<<<<<<<<
 *                     if self.options.invalid_value_include_map_key:
 *                         e.schema_path = (key, ) + e.schema_path
 */
        __Pyx_ErrFetch(&__pyx_t_3, &__pyx_t_8, &__pyx_t_4);
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_9)) __PYX_ERR(9, 41, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_5 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_t_9);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_ErrRestore(__pyx_t_3, __pyx_t_8, __pyx_t_4);
        __pyx_t_3 = 0; __pyx_t_8 = 0; __pyx_t_4 = 0;
        if (__pyx_t_5) {
          __Pyx_AddTraceback("cavro.MapType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_8, &__pyx_t_3) < 0) __PYX_ERR(9, 41, __pyx_L11_except_error)
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_8);
          __pyx_v_e = __pyx_t_8;
          /*try:*/ {

            /* "src/map.pxi":42
 *                     self.value_type.binary_buffer_encode(buffer, item_value)
 *                 except InvalidValue as e:
 *                     if self.options.invalid_value_include_map_key:             # <<<<<<<<<<<<<<
 *                         e.schema_path = (key, ) + e.schema_path
 *                     raise
 */
            if (__pyx_v_self->__pyx_base.options->invalid_value_include_map_key) {

              /* "src/map.pxi":43
 *                 except InvalidValue as e:
 *                     if self.options.invalid_value_include_map_key:
 *                         e.schema_path = (key, ) + e.schema_path             # <<<<<<<<<<<<<<
 *                     raise
 *             zigzag_encode_long(buffer, 0)
 */
              __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(9, 43, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_9);
              __Pyx_INCREF(__pyx_v_key);
              __Pyx_GIVEREF(__pyx_v_key);
              if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_key)) __PYX_ERR(9, 43, __pyx_L22_error);
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_schema_path); if (unlikely(!__pyx_t_14)) __PYX_ERR(9, 43, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_15 = PyNumber_Add(__pyx_t_9, __pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(9, 43, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_e, __pyx_n_s_schema_path, __pyx_t_15) < 0) __PYX_ERR(9, 43, __pyx_L22_error)
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

              /* "src/map.pxi":42
 *                     self.value_type.binary_buffer_encode(buffer, item_value)
 *                 except InvalidValue as e:
 *                     if self.options.invalid_value_include_map_key:             # <<<<<<<<<<<<<<
 *                         e.schema_path = (key, ) + e.schema_path
 *                     raise
 */
            }

            /* "src/map.pxi":44
 *                     if self.options.invalid_value_include_map_key:
 *                         e.schema_path = (key, ) + e.schema_path
 *                     raise             # <<<<<<<<<<<<<<
 *             zigzag_encode_long(buffer, 0)
 *         return 0
 */
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_8);
            __Pyx_XGIVEREF(__pyx_t_3);
            __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_8, __pyx_t_3);
            __pyx_t_4 = 0; __pyx_t_8 = 0; __pyx_t_3 = 0; 
            __PYX_ERR(9, 44, __pyx_L22_error)
          }

          /* "src/map.pxi":41
 *                 try:
 *                     self.value_type.binary_buffer_encode(buffer, item_value)
 *                 except InvalidValue as e:             # <<<<<<<<<<<<<<
 *                     if self.options.invalid_value_include_map_key:
 *                         e.schema_path = (key, ) + e.schema_path
 */
          /*finally:*/ {
            __pyx_L22_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_22);
              __Pyx_XGOTREF(__pyx_t_23);
              __pyx_t_5 = __pyx_lineno; __pyx_t_16 = __pyx_clineno; __pyx_t_17 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_XGIVEREF(__pyx_t_22);
                __Pyx_XGIVEREF(__pyx_t_23);
                __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
              }
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ErrRestore(__pyx_t_18, __pyx_t_19, __pyx_t_20);
              __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
              __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_16; __pyx_filename = __pyx_t_17;
              goto __pyx_L11_except_error;
            }
          }
        }
        goto __pyx_L11_except_error;

        /* "src/map.pxi":39
 *             for key, item_value in value:
 *                 self.key_type.binary_buffer_encode(buffer, key)
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.value_type.binary_buffer_encode(buffer, item_value)
 *                 except InvalidValue as e:
 */
        __pyx_L11_except_error:;
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
        goto __pyx_L1_error;
        __pyx_L16_try_end:;
      }

      /* "src/map.pxi":37
 *         zigzag_encode_long(buffer, len(value))
 *         if value:
 *             for key, item_value in value:             # <<<<<<<<<<<<<<
 *                 self.key_type.binary_buffer_encode(buffer, key)
 *                 try:
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/map.pxi":45
 *                         e.schema_path = (key, ) + e.schema_path
 *                     raise
 *             zigzag_encode_long(buffer, 0)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_16 = __pyx_f_5cavro_zigzag_encode_long(__pyx_v_buffer, 0); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(9, 45, __pyx_L1_error)

    /* "src/map.pxi":36
 *             value = value.items()
 *         zigzag_encode_long(buffer, len(value))
 *         if value:             # <<<<<<<<<<<<<<
 *             for key, item_value in value:
 *                 self.key_type.binary_buffer_encode(buffer, key)
 */
  }

  /* "src/map.pxi":46
 *                     raise
 *             zigzag_encode_long(buffer, 0)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "src/map.pxi":32
 *         return {'values': self.value_type.get_schema(created)}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         if hasattr(value, 'items'):
 *             value = value.items()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("cavro.MapType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_item_value);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":48
 *         return 0
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef dict out = {}
 *         cdef size_t length
 */

static PyObject *__pyx_f_5cavro_7MapType__binary_buffer_decode(struct __pyx_obj_5cavro_MapType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  PyObject *__pyx_v_out = 0;
  size_t __pyx_v_length;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/map.pxi":49
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef dict out = {}             # <<<<<<<<<<<<<<
 *         cdef size_t length
 *         cdef str key
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":52
 *         cdef size_t length
 *         cdef str key
 *         while True:             # <<<<<<<<<<<<<<
 *             length = zigzag_decode_long(buffer)
 *             if length == 0:
 */
  while (1) {

    /* "src/map.pxi":53
 *         cdef str key
 *         while True:
 *             length = zigzag_decode_long(buffer)             # <<<<<<<<<<<<<<
 *             if length == 0:
 *                 return out
 */
    __pyx_t_2 = __pyx_f_5cavro_zigzag_decode_long(__pyx_v_buffer); if (unlikely(__pyx_t_2 == ((int64_t)0x7ffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(9, 53, __pyx_L1_error)
    __pyx_v_length = __pyx_t_2;

    /* "src/map.pxi":54
 *         while True:
 *             length = zigzag_decode_long(buffer)
 *             if length == 0:             # <<<<<<<<<<<<<<
 *                 return out
 *             while length:
 */
    __pyx_t_3 = (__pyx_v_length == 0);
    if (__pyx_t_3) {

      /* "src/map.pxi":55
 *             length = zigzag_decode_long(buffer)
 *             if length == 0:
 *                 return out             # <<<<<<<<<<<<<<
 *             while length:
 *                 key = self.key_type.binary_buffer_decode(buffer)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_out);
      __pyx_r = __pyx_v_out;
      goto __pyx_L0;

      /* "src/map.pxi":54
 *         while True:
 *             length = zigzag_decode_long(buffer)
 *             if length == 0:             # <<<<<<<<<<<<<<
 *                 return out
 *             while length:
 */
    }

    /* "src/map.pxi":56
 *             if length == 0:
 *                 return out
 *             while length:             # <<<<<<<<<<<<<<
 *                 key = self.key_type.binary_buffer_decode(buffer)
 *                 value = self.value_type.binary_buffer_decode(buffer)
 */
    while (1) {
      __pyx_t_3 = (__pyx_v_length != 0);
      if (!__pyx_t_3) break;

      /* "src/map.pxi":57
 *                 return out
 *             while length:
 *                 key = self.key_type.binary_buffer_decode(buffer)             # <<<<<<<<<<<<<<
 *                 value = self.value_type.binary_buffer_decode(buffer)
 *                 out[key] = value
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_StringType *)__pyx_v_self->key_type->__pyx_base.__pyx_vtab)->__pyx_base.binary_buffer_decode(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self->key_type), __pyx_v_buffer); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 57, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(9, 57, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_key, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "src/map.pxi":58
 *             while length:
 *                 key = self.key_type.binary_buffer_decode(buffer)
 *                 value = self.value_type.binary_buffer_decode(buffer)             # <<<<<<<<<<<<<<
 *                 out[key] = value
 *                 length -= 1
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->value_type->__pyx_vtab)->binary_buffer_decode(__pyx_v_self->value_type, __pyx_v_buffer); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 58, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "src/map.pxi":59
 *                 key = self.key_type.binary_buffer_decode(buffer)
 *                 value = self.value_type.binary_buffer_decode(buffer)
 *                 out[key] = value             # <<<<<<<<<<<<<<
 *                 length -= 1
 * 
 */
      if (unlikely((PyDict_SetItem(__pyx_v_out, __pyx_v_key, __pyx_v_value) < 0))) __PYX_ERR(9, 59, __pyx_L1_error)

      /* "src/map.pxi":60
 *                 value = self.value_type.binary_buffer_decode(buffer)
 *                 out[key] = value
 *                 length -= 1             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
      __pyx_v_length = (__pyx_v_length - 1);
    }
  }

  /* "src/map.pxi":48
 *         return 0
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef dict out = {}
 *         cdef size_t length
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.MapType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":62
 *                 length -= 1
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         cdef str key
 *         cdef dict out = {}
 */

static PyObject *__pyx_f_5cavro_7MapType__json_format(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_out = 0;
  PyObject *__pyx_v_item_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/map.pxi":64
 *     cdef _json_format(self, value):
 *         cdef str key
 *         cdef dict out = {}             # <<<<<<<<<<<<<<
 *         if hasattr(value, 'items'):
 *             value = value.items()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":65
 *         cdef str key
 *         cdef dict out = {}
 *         if hasattr(value, 'items'):             # <<<<<<<<<<<<<<
 *             value = value.items()
 *         for key, item_value in value:
 */
  __pyx_t_2 = __Pyx_HasAttr(__pyx_v_value, __pyx_n_u_items); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(9, 65, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "src/map.pxi":66
 *         cdef dict out = {}
 *         if hasattr(value, 'items'):
 *             value = value.items()             # <<<<<<<<<<<<<<
 *         for key, item_value in value:
 *             out[key] = self.value_type.json_format(item_value)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 66, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/map.pxi":65
 *         cdef str key
 *         cdef dict out = {}
 *         if hasattr(value, 'items'):             # <<<<<<<<<<<<<<
 *             value = value.items()
 *         for key, item_value in value:
 */
  }

  /* "src/map.pxi":67
 *         if hasattr(value, 'items'):
 *             value = value.items()
 *         for key, item_value in value:             # <<<<<<<<<<<<<<
 *             out[key] = self.value_type.json_format(item_value)
 *         return out
 */
  if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
    __pyx_t_1 = __pyx_v_value; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(9, 67, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(9, 67, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(9, 67, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 67, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(9, 67, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(9, 67, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 67, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(9, 67, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(9, 67, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 67, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(9, 67, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(9, 67, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_9);
      index = 0; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(9, 67, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(9, 67, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_4))) __PYX_ERR(9, 67, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_key, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_item_value, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "src/map.pxi":68
 *             value = value.items()
 *         for key, item_value in value:
 *             out[key] = self.value_type.json_format(item_value)             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->value_type->__pyx_vtab)->json_format(__pyx_v_self->value_type, __pyx_v_item_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely((PyDict_SetItem(__pyx_v_out, __pyx_v_key, __pyx_t_3) < 0))) __PYX_ERR(9, 68, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "src/map.pxi":67
 *         if hasattr(value, 'items'):
 *             value = value.items()
 *         for key, item_value in value:             # <<<<<<<<<<<<<<
 *             out[key] = self.value_type.json_format(item_value)
 *         return out
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/map.pxi":69
 *         for key, item_value in value:
 *             out[key] = self.value_type.json_format(item_value)
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "src/map.pxi":62
 *                 length -= 1
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         cdef str key
 *         cdef dict out = {}
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cavro.MapType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_item_value);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":71
 *         return out
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         cdef dict inp = value
 *         cdef dict out = {}
 */

static PyObject *__pyx_f_5cavro_7MapType__json_decode(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_inp = 0;
  PyObject *__pyx_v_out = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/map.pxi":72
 * 
 *     cdef _json_decode(self, value):
 *         cdef dict inp = value             # <<<<<<<<<<<<<<
 *         cdef dict out = {}
 *         cdef str key
 */
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(9, 72, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_inp = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":73
 *     cdef _json_decode(self, value):
 *         cdef dict inp = value
 *         cdef dict out = {}             # <<<<<<<<<<<<<<
 *         cdef str key
 *         for key, val in inp.items():
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":75
 *         cdef dict out = {}
 *         cdef str key
 *         for key, val in inp.items():             # <<<<<<<<<<<<<<
 *             out[key] = self.value_type.json_decode(val)
 *         return out
 */
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v_inp == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(9, 75, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_inp, 1, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(9, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_5))) __PYX_ERR(9, 75, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_key, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "src/map.pxi":76
 *         cdef str key
 *         for key, val in inp.items():
 *             out[key] = self.value_type.json_decode(val)             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->value_type->__pyx_vtab)->json_decode(__pyx_v_self->value_type, __pyx_v_val); if (unlikely(!__pyx_t_6)) __PYX_ERR(9, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely((PyDict_SetItem(__pyx_v_out, __pyx_v_key, __pyx_t_6) < 0))) __PYX_ERR(9, 76, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/map.pxi":77
 *         for key, val in inp.items():
 *             out[key] = self.value_type.json_decode(val)
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "src/map.pxi":71
 *         return out
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         cdef dict inp = value
 *         cdef dict out = {}
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.MapType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_inp);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":79
 *         return out
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef int level = FIT_OK
 *         if isinstance(value, dict):
 */

static int __pyx_f_5cavro_7MapType__get_value_fitness(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_v_level;
  PyObject *__pyx_v_it = NULL;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_item_value = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/map.pxi":80
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         cdef int level = FIT_OK             # <<<<<<<<<<<<<<
 *         if isinstance(value, dict):
 *             level = FIT_EXACT
 */
  __pyx_v_level = __pyx_v_5cavro_FIT_OK;

  /* "src/map.pxi":81
 *     cdef int _get_value_fitness(self, value) except -1:
 *         cdef int level = FIT_OK
 *         if isinstance(value, dict):             # <<<<<<<<<<<<<<
 *             level = FIT_EXACT
 * 
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_value); 
  if (__pyx_t_1) {

    /* "src/map.pxi":82
 *         cdef int level = FIT_OK
 *         if isinstance(value, dict):
 *             level = FIT_EXACT             # <<<<<<<<<<<<<<
 * 
 *         elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
 */
    __pyx_v_level = __pyx_v_5cavro_FIT_EXACT;

    /* "src/map.pxi":81
 *     cdef int _get_value_fitness(self, value) except -1:
 *         cdef int level = FIT_OK
 *         if isinstance(value, dict):             # <<<<<<<<<<<<<<
 *             level = FIT_EXACT
 * 
 */
    goto __pyx_L3;
  }

  /* "src/map.pxi":84
 *             level = FIT_EXACT
 * 
 *         elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *             return FIT_POOR
 * 
 */
  if (__pyx_v_self->__pyx_base.options->missing_values_can_be_empty_container) {
  } else {
    __pyx_t_1 = __pyx_v_self->__pyx_base.options->missing_values_can_be_empty_container;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__pyx_v_value == __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/map.pxi":85
 * 
 *         elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
 *             return FIT_POOR             # <<<<<<<<<<<<<<
 * 
 *         if hasattr(value, 'items'):
 */
    __pyx_r = __pyx_v_5cavro_FIT_POOR;
    goto __pyx_L0;

    /* "src/map.pxi":84
 *             level = FIT_EXACT
 * 
 *         elif self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *             return FIT_POOR
 * 
 */
  }
  __pyx_L3:;

  /* "src/map.pxi":87
 *             return FIT_POOR
 * 
 *         if hasattr(value, 'items'):             # <<<<<<<<<<<<<<
 *             value = value.items()
 *         try:
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_value, __pyx_n_u_items); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(9, 87, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/map.pxi":88
 * 
 *         if hasattr(value, 'items'):
 *             value = value.items()             # <<<<<<<<<<<<<<
 *         try:
 *             it = iter(value)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 88, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/map.pxi":87
 *             return FIT_POOR
 * 
 *         if hasattr(value, 'items'):             # <<<<<<<<<<<<<<
 *             value = value.items()
 *         try:
 */
  }

  /* "src/map.pxi":89
 *         if hasattr(value, 'items'):
 *             value = value.items()
 *         try:             # <<<<<<<<<<<<<<
 *             it = iter(value)
 *         except (TypeError, KeyError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "src/map.pxi":90
 *             value = value.items()
 *         try:
 *             it = iter(value)             # <<<<<<<<<<<<<<
 *         except (TypeError, KeyError):
 *             return False
 */
      __pyx_t_2 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 90, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_it = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "src/map.pxi":89
 *         if hasattr(value, 'items'):
 *             value = value.items()
 *         try:             # <<<<<<<<<<<<<<
 *             it = iter(value)
 *         except (TypeError, KeyError):
 */
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L12_try_end;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "src/map.pxi":91
 *         try:
 *             it = iter(value)
 *         except (TypeError, KeyError):             # <<<<<<<<<<<<<<
 *             return False
 *         while True:
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_TypeError, __pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("cavro.MapType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(9, 91, __pyx_L9_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);

      /* "src/map.pxi":92
 *             it = iter(value)
 *         except (TypeError, KeyError):
 *             return False             # <<<<<<<<<<<<<<
 *         while True:
 *             try:
 */
      __pyx_r = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_except_return;
    }
    goto __pyx_L9_except_error;

    /* "src/map.pxi":89
 *         if hasattr(value, 'items'):
 *             value = value.items()
 *         try:             # <<<<<<<<<<<<<<
 *             it = iter(value)
 *         except (TypeError, KeyError):
 */
    __pyx_L9_except_error:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L10_except_return:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L0;
    __pyx_L12_try_end:;
  }

  /* "src/map.pxi":93
 *         except (TypeError, KeyError):
 *             return False
 *         while True:             # <<<<<<<<<<<<<<
 *             try:
 *                 item = next(it)
 */
  while (1) {

    /* "src/map.pxi":94
 *             return False
 *         while True:
 *             try:             # <<<<<<<<<<<<<<
 *                 item = next(it)
 *             except StopIteration:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "src/map.pxi":95
 *         while True:
 *             try:
 *                 item = next(it)             # <<<<<<<<<<<<<<
 *             except StopIteration:
 *                 break
 */
        __pyx_t_5 = __Pyx_PyIter_Next(__pyx_v_it); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 95, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "src/map.pxi":94
 *             return False
 *         while True:
 *             try:             # <<<<<<<<<<<<<<
 *                 item = next(it)
 *             except StopIteration:
 */
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L24_try_end;
      __pyx_L17_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "src/map.pxi":96
 *             try:
 *                 item = next(it)
 *             except StopIteration:             # <<<<<<<<<<<<<<
 *                 break
 *             except:
 */
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_StopIteration);
      if (__pyx_t_6) {
        __Pyx_AddTraceback("cavro.MapType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_2) < 0) __PYX_ERR(9, 96, __pyx_L19_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_2);

        /* "src/map.pxi":97
 *                 item = next(it)
 *             except StopIteration:
 *                 break             # <<<<<<<<<<<<<<
 *             except:
 *                 return FIT_NONE
 */
        goto __pyx_L26_except_break;
        __pyx_L26_except_break:;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L22_try_break;
      }

      /* "src/map.pxi":98
 *             except StopIteration:
 *                 break
 *             except:             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *             try:
 */
      /*except:*/ {
        __Pyx_AddTraceback("cavro.MapType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(9, 98, __pyx_L19_except_error)
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_5);

        /* "src/map.pxi":99
 *                 break
 *             except:
 *                 return FIT_NONE             # <<<<<<<<<<<<<<
 *             try:
 *                 key, item_value = item
 */
        __pyx_r = __pyx_v_5cavro_FIT_NONE;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L20_except_return;
      }

      /* "src/map.pxi":94
 *             return False
 *         while True:
 *             try:             # <<<<<<<<<<<<<<
 *                 item = next(it)
 *             except StopIteration:
 */
      __pyx_L19_except_error:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L22_try_break:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      goto __pyx_L16_break;
      __pyx_L20_except_return:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      goto __pyx_L0;
      __pyx_L24_try_end:;
    }

    /* "src/map.pxi":100
 *             except:
 *                 return FIT_NONE
 *             try:             # <<<<<<<<<<<<<<
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError, KeyError):
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "src/map.pxi":101
 *                 return FIT_NONE
 *             try:
 *                 key, item_value = item             # <<<<<<<<<<<<<<
 *             except (TypeError, ValueError, IndexError, KeyError):
 *                 return FIT_NONE
 */
        if ((likely(PyTuple_CheckExact(__pyx_v_item))) || (PyList_CheckExact(__pyx_v_item))) {
          PyObject* sequence = __pyx_v_item;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(9, 101, __pyx_L29_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          #else
          __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 101, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 101, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          Py_ssize_t index = -1;
          __pyx_t_2 = PyObject_GetIter(__pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 101, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2);
          index = 0; __pyx_t_5 = __pyx_t_10(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L37_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_5);
          index = 1; __pyx_t_4 = __pyx_t_10(__pyx_t_2); if (unlikely(!__pyx_t_4)) goto __pyx_L37_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_4);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_2), 2) < 0) __PYX_ERR(9, 101, __pyx_L29_error)
          __pyx_t_10 = NULL;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L38_unpacking_done;
          __pyx_L37_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_10 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(9, 101, __pyx_L29_error)
          __pyx_L38_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_5);
        __pyx_t_5 = 0;
        __Pyx_XDECREF_SET(__pyx_v_item_value, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "src/map.pxi":100
 *             except:
 *                 return FIT_NONE
 *             try:             # <<<<<<<<<<<<<<
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError, KeyError):
 */
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L36_try_end;
      __pyx_L29_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "src/map.pxi":102
 *             try:
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError, KeyError):             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *             level = min(self.key_type.get_value_fitness(key), level)
 */
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_6) {
        __Pyx_AddTraceback("cavro.MapType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_2) < 0) __PYX_ERR(9, 102, __pyx_L31_except_error)
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_2);

        /* "src/map.pxi":103
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError, KeyError):
 *                 return FIT_NONE             # <<<<<<<<<<<<<<
 *             level = min(self.key_type.get_value_fitness(key), level)
 *             level = min(self.value_type.get_value_fitness(item_value), level)
 */
        __pyx_r = __pyx_v_5cavro_FIT_NONE;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L32_except_return;
      }
      goto __pyx_L31_except_error;

      /* "src/map.pxi":100
 *             except:
 *                 return FIT_NONE
 *             try:             # <<<<<<<<<<<<<<
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError, KeyError):
 */
      __pyx_L31_except_error:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L32_except_return:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L0;
      __pyx_L36_try_end:;
    }

    /* "src/map.pxi":104
 *             except (TypeError, ValueError, IndexError, KeyError):
 *                 return FIT_NONE
 *             level = min(self.key_type.get_value_fitness(key), level)             # <<<<<<<<<<<<<<
 *             level = min(self.value_type.get_value_fitness(item_value), level)
 *             if level <= FIT_NONE:
 */
    __pyx_t_6 = __pyx_v_level;
    __pyx_t_11 = ((struct __pyx_vtabstruct_5cavro_StringType *)__pyx_v_self->key_type->__pyx_base.__pyx_vtab)->__pyx_base.get_value_fitness(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self->key_type), __pyx_v_key); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(9, 104, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_6 < __pyx_t_11);
    if (__pyx_t_1) {
      __pyx_t_12 = __pyx_t_6;
    } else {
      __pyx_t_12 = __pyx_t_11;
    }
    __pyx_v_level = __pyx_t_12;

    /* "src/map.pxi":105
 *                 return FIT_NONE
 *             level = min(self.key_type.get_value_fitness(key), level)
 *             level = min(self.value_type.get_value_fitness(item_value), level)             # <<<<<<<<<<<<<<
 *             if level <= FIT_NONE:
 *                 break
 */
    __pyx_t_12 = __pyx_v_level;
    __pyx_t_6 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->value_type->__pyx_vtab)->get_value_fitness(__pyx_v_self->value_type, __pyx_v_item_value); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(9, 105, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_12 < __pyx_t_6);
    if (__pyx_t_1) {
      __pyx_t_11 = __pyx_t_12;
    } else {
      __pyx_t_11 = __pyx_t_6;
    }
    __pyx_v_level = __pyx_t_11;

    /* "src/map.pxi":106
 *             level = min(self.key_type.get_value_fitness(key), level)
 *             level = min(self.value_type.get_value_fitness(item_value), level)
 *             if level <= FIT_NONE:             # <<<<<<<<<<<<<<
 *                 break
 *         return level
 */
    __pyx_t_1 = (__pyx_v_level <= __pyx_v_5cavro_FIT_NONE);
    if (__pyx_t_1) {

      /* "src/map.pxi":107
 *             level = min(self.value_type.get_value_fitness(item_value), level)
 *             if level <= FIT_NONE:
 *                 break             # <<<<<<<<<<<<<<
 *         return level
 * 
 */
      goto __pyx_L16_break;

      /* "src/map.pxi":106
 *             level = min(self.key_type.get_value_fitness(key), level)
 *             level = min(self.value_type.get_value_fitness(item_value), level)
 *             if level <= FIT_NONE:             # <<<<<<<<<<<<<<
 *                 break
 *         return level
 */
    }
  }
  __pyx_L16_break:;

  /* "src/map.pxi":108
 *             if level <= FIT_NONE:
 *                 break
 *         return level             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _make_converted_map(self, value):
 */
  __pyx_r = __pyx_v_level;
  goto __pyx_L0;

  /* "src/map.pxi":79
 *         return out
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef int level = FIT_OK
 *         if isinstance(value, dict):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.MapType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_it);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_item_value);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":110
 *         return level
 * 
 *     cdef dict _make_converted_map(self, value):             # <<<<<<<<<<<<<<
 *         cdef AvroType key_type = self.key_type
 *         cdef AvroType value_type = self.value_type
 */

static PyObject *__pyx_f_5cavro_7MapType__make_converted_map(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_key_type = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_value_type = 0;
  PyObject *__pyx_v_out = 0;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_item_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_UCS4 __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_converted_map", 1);

  /* "src/map.pxi":111
 * 
 *     cdef dict _make_converted_map(self, value):
 *         cdef AvroType key_type = self.key_type             # <<<<<<<<<<<<<<
 *         cdef AvroType value_type = self.value_type
 *         cdef dict out = {}
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->key_type);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_key_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":112
 *     cdef dict _make_converted_map(self, value):
 *         cdef AvroType key_type = self.key_type
 *         cdef AvroType value_type = self.value_type             # <<<<<<<<<<<<<<
 *         cdef dict out = {}
 *         for item in value:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->value_type);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_value_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":113
 *         cdef AvroType key_type = self.key_type
 *         cdef AvroType value_type = self.value_type
 *         cdef dict out = {}             # <<<<<<<<<<<<<<
 *         for item in value:
 *             try:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":114
 *         cdef AvroType value_type = self.value_type
 *         cdef dict out = {}
 *         for item in value:             # <<<<<<<<<<<<<<
 *             try:
 *                 key, item_value = item
 */
  if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
    __pyx_t_1 = __pyx_v_value; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 114, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(9, 114, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(9, 114, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 114, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(9, 114, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(9, 114, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 114, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(9, 114, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/map.pxi":115
 *         cdef dict out = {}
 *         for item in value:
 *             try:             # <<<<<<<<<<<<<<
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError):
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "src/map.pxi":116
 *         for item in value:
 *             try:
 *                 key, item_value = item             # <<<<<<<<<<<<<<
 *             except (TypeError, ValueError, IndexError):
 *                 raise ValueError(f"'{value}' is not a mapping")
 */
        if ((likely(PyTuple_CheckExact(__pyx_v_item))) || (PyList_CheckExact(__pyx_v_item))) {
          PyObject* sequence = __pyx_v_item;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(9, 116, __pyx_L5_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_8);
          #else
          __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 116, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(9, 116, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        } else {
          Py_ssize_t index = -1;
          __pyx_t_9 = PyObject_GetIter(__pyx_v_item); if (unlikely(!__pyx_t_9)) __PYX_ERR(9, 116, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_9);
          index = 0; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L13_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_4);
          index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L13_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_8);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(9, 116, __pyx_L5_error)
          __pyx_t_10 = NULL;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L14_unpacking_done;
          __pyx_L13_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_10 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(9, 116, __pyx_L5_error)
          __pyx_L14_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_4);
        __pyx_t_4 = 0;
        __Pyx_XDECREF_SET(__pyx_v_item_value, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "src/map.pxi":115
 *         cdef dict out = {}
 *         for item in value:
 *             try:             # <<<<<<<<<<<<<<
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError):
 */
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "src/map.pxi":117
 *             try:
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError):             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"'{value}' is not a mapping")
 *             out[key_type._convert_value(key)] = value_type._convert_value(item_value)
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
      if (__pyx_t_11) {
        __Pyx_AddTraceback("cavro.MapType._make_converted_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_4, &__pyx_t_9) < 0) __PYX_ERR(9, 117, __pyx_L7_except_error)
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_9);

        /* "src/map.pxi":118
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError):
 *                 raise ValueError(f"'{value}' is not a mapping")             # <<<<<<<<<<<<<<
 *             out[key_type._convert_value(key)] = value_type._convert_value(item_value)
 *         return out
 */
        __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(9, 118, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = 0;
        __pyx_t_14 = 127;
        __Pyx_INCREF(__pyx_kp_u__6);
        __pyx_t_13 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__6);
        PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_kp_u__6);
        __pyx_t_15 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_15)) __PYX_ERR(9, 118, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_14 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) > __pyx_t_14) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) : __pyx_t_14;
        __pyx_t_13 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15);
        __Pyx_GIVEREF(__pyx_t_15);
        PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_15);
        __pyx_t_15 = 0;
        __Pyx_INCREF(__pyx_kp_u_is_not_a_mapping);
        __pyx_t_13 += 18;
        __Pyx_GIVEREF(__pyx_kp_u_is_not_a_mapping);
        PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_kp_u_is_not_a_mapping);
        __pyx_t_15 = __Pyx_PyUnicode_Join(__pyx_t_12, 3, __pyx_t_13, __pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(9, 118, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_12 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_15); if (unlikely(!__pyx_t_12)) __PYX_ERR(9, 118, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_Raise(__pyx_t_12, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __PYX_ERR(9, 118, __pyx_L7_except_error)
      }
      goto __pyx_L7_except_error;

      /* "src/map.pxi":115
 *         cdef dict out = {}
 *         for item in value:
 *             try:             # <<<<<<<<<<<<<<
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError):
 */
      __pyx_L7_except_error:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L12_try_end:;
    }

    /* "src/map.pxi":119
 *             except (TypeError, ValueError, IndexError):
 *                 raise ValueError(f"'{value}' is not a mapping")
 *             out[key_type._convert_value(key)] = value_type._convert_value(item_value)             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
    __pyx_t_9 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_value_type->__pyx_vtab)->_convert_value(__pyx_v_value_type, __pyx_v_item_value, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(9, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_key_type->__pyx_vtab)->_convert_value(__pyx_v_key_type, __pyx_v_key, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely((PyDict_SetItem(__pyx_v_out, __pyx_t_4, __pyx_t_9) < 0))) __PYX_ERR(9, 119, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "src/map.pxi":114
 *         cdef AvroType value_type = self.value_type
 *         cdef dict out = {}
 *         for item in value:             # <<<<<<<<<<<<<<
 *             try:
 *                 key, item_value = item
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/map.pxi":120
 *                 raise ValueError(f"'{value}' is not a mapping")
 *             out[key_type._convert_value(key)] = value_type._convert_value(item_value)
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cpdef object convert_value(self, object orig_value, check_value=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "src/map.pxi":110
 *         return level
 * 
 *     cdef dict _make_converted_map(self, value):             # <<<<<<<<<<<<<<
 *         cdef AvroType key_type = self.key_type
 *         cdef AvroType value_type = self.value_type
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("cavro.MapType._make_converted_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_key_type);
  __Pyx_XDECREF((PyObject *)__pyx_v_value_type);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_item_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":122
 *         return out
 * 
 *     cpdef object convert_value(self, object orig_value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef int key_fitness
 *         cdef int value_fitness
 */

static PyObject *__pyx_pw_5cavro_7MapType_10convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_7MapType_convert_value(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_orig_value, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_7MapType_convert_value *__pyx_optional_args) {
  PyObject *__pyx_v_check_value = ((PyObject *)Py_True);
  int __pyx_v_key_fitness;
  int __pyx_v_value_fitness;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_key_type = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_value_type = 0;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_it = NULL;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_item_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  Py_UCS4 __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_value", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_check_value = __pyx_optional_args->check_value;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_convert_value_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_7MapType_10convert_value)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_orig_value, __pyx_v_check_value};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 122, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/map.pxi":125
 *         cdef int key_fitness
 *         cdef int value_fitness
 *         cdef AvroType key_type = self.key_type             # <<<<<<<<<<<<<<
 *         cdef AvroType value_type = self.value_type
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->key_type);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_key_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":126
 *         cdef int value_fitness
 *         cdef AvroType key_type = self.key_type
 *         cdef AvroType value_type = self.value_type             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(orig_value, dict):
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->value_type);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_value_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/map.pxi":128
 *         cdef AvroType value_type = self.value_type
 * 
 *         if isinstance(orig_value, dict):             # <<<<<<<<<<<<<<
 *             value = orig_value.items()
 *         else:
 */
  __pyx_t_6 = PyDict_Check(__pyx_v_orig_value); 
  if (__pyx_t_6) {

    /* "src/map.pxi":129
 * 
 *         if isinstance(orig_value, dict):
 *             value = orig_value.items()             # <<<<<<<<<<<<<<
 *         else:
 *             value = orig_value
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_orig_value, __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_v_value = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "src/map.pxi":128
 *         cdef AvroType value_type = self.value_type
 * 
 *         if isinstance(orig_value, dict):             # <<<<<<<<<<<<<<
 *             value = orig_value.items()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "src/map.pxi":131
 *             value = orig_value.items()
 *         else:
 *             value = orig_value             # <<<<<<<<<<<<<<
 * 
 *         if self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_orig_value);
    __pyx_v_value = __pyx_v_orig_value;
  }
  __pyx_L3:;

  /* "src/map.pxi":133
 *             value = orig_value
 * 
 *         if self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *             return {}
 * 
 */
  if (__pyx_v_self->__pyx_base.options->missing_values_can_be_empty_container) {
  } else {
    __pyx_t_6 = __pyx_v_self->__pyx_base.options->missing_values_can_be_empty_container;
    goto __pyx_L5_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = (__pyx_v_value == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_6) {

    /* "src/map.pxi":134
 * 
 *         if self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:
 *             return {}             # <<<<<<<<<<<<<<
 * 
 *         it = iter(value)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/map.pxi":133
 *             value = orig_value
 * 
 *         if self.options.missing_values_can_be_empty_container and value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *             return {}
 * 
 */
  }

  /* "src/map.pxi":136
 *             return {}
 * 
 *         it = iter(value)             # <<<<<<<<<<<<<<
 *         for item in it:
 *             try:
 */
  __pyx_t_1 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_it = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/map.pxi":137
 * 
 *         it = iter(value)
 *         for item in it:             # <<<<<<<<<<<<<<
 *             try:
 *                 key, item_value = item
 */
  if (likely(PyList_CheckExact(__pyx_v_it)) || PyTuple_CheckExact(__pyx_v_it)) {
    __pyx_t_1 = __pyx_v_it; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_it); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(9, 137, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(9, 137, __pyx_L1_error)
          #endif
          if (__pyx_t_8 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(9, 137, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 137, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(9, 137, __pyx_L1_error)
          #endif
          if (__pyx_t_8 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(9, 137, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 137, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_9(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(9, 137, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/map.pxi":138
 *         it = iter(value)
 *         for item in it:
 *             try:             # <<<<<<<<<<<<<<
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError):
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      /*try:*/ {

        /* "src/map.pxi":139
 *         for item in it:
 *             try:
 *                 key, item_value = item             # <<<<<<<<<<<<<<
 *             except (TypeError, ValueError, IndexError):
 *                 raise ValueError(f"'{value}' is not a mapping")
 */
        if ((likely(PyTuple_CheckExact(__pyx_v_item))) || (PyList_CheckExact(__pyx_v_item))) {
          PyObject* sequence = __pyx_v_item;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(9, 139, __pyx_L9_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          #else
          __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 139, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 139, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          Py_ssize_t index = -1;
          __pyx_t_4 = PyObject_GetIter(__pyx_v_item); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 139, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_13 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
          index = 0; __pyx_t_2 = __pyx_t_13(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L17_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          index = 1; __pyx_t_3 = __pyx_t_13(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L17_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_3);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_4), 2) < 0) __PYX_ERR(9, 139, __pyx_L9_error)
          __pyx_t_13 = NULL;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L18_unpacking_done;
          __pyx_L17_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_13 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(9, 139, __pyx_L9_error)
          __pyx_L18_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_item_value, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "src/map.pxi":138
 *         it = iter(value)
 *         for item in it:
 *             try:             # <<<<<<<<<<<<<<
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError):
 */
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L16_try_end;
      __pyx_L9_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "src/map.pxi":140
 *             try:
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError):             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"'{value}' is not a mapping")
 *             key_fitness = key_type.get_value_fitness(key)
 */
      __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
      if (__pyx_t_5) {
        __Pyx_AddTraceback("cavro.MapType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_4) < 0) __PYX_ERR(9, 140, __pyx_L11_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_4);

        /* "src/map.pxi":141
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError):
 *                 raise ValueError(f"'{value}' is not a mapping")             # <<<<<<<<<<<<<<
 *             key_fitness = key_type.get_value_fitness(key)
 *             if key_fitness == FIT_NONE:
 */
        __pyx_t_14 = PyTuple_New(3); if (unlikely(!__pyx_t_14)) __PYX_ERR(9, 141, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = 0;
        __pyx_t_16 = 127;
        __Pyx_INCREF(__pyx_kp_u__6);
        __pyx_t_15 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__6);
        PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_kp_u__6);
        __pyx_t_17 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_17)) __PYX_ERR(9, 141, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) : __pyx_t_16;
        __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_17);
        __Pyx_GIVEREF(__pyx_t_17);
        PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_17);
        __pyx_t_17 = 0;
        __Pyx_INCREF(__pyx_kp_u_is_not_a_mapping);
        __pyx_t_15 += 18;
        __Pyx_GIVEREF(__pyx_kp_u_is_not_a_mapping);
        PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_kp_u_is_not_a_mapping);
        __pyx_t_17 = __Pyx_PyUnicode_Join(__pyx_t_14, 3, __pyx_t_15, __pyx_t_16); if (unlikely(!__pyx_t_17)) __PYX_ERR(9, 141, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_17); if (unlikely(!__pyx_t_14)) __PYX_ERR(9, 141, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_Raise(__pyx_t_14, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __PYX_ERR(9, 141, __pyx_L11_except_error)
      }
      goto __pyx_L11_except_error;

      /* "src/map.pxi":138
 *         it = iter(value)
 *         for item in it:
 *             try:             # <<<<<<<<<<<<<<
 *                 key, item_value = item
 *             except (TypeError, ValueError, IndexError):
 */
      __pyx_L11_except_error:;
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      goto __pyx_L1_error;
      __pyx_L16_try_end:;
    }

    /* "src/map.pxi":142
 *             except (TypeError, ValueError, IndexError):
 *                 raise ValueError(f"'{value}' is not a mapping")
 *             key_fitness = key_type.get_value_fitness(key)             # <<<<<<<<<<<<<<
 *             if key_fitness == FIT_NONE:
 *                 raise ValueError(f"'{key}' is not a valid key for map type")
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_key_type->__pyx_vtab)->get_value_fitness(__pyx_v_key_type, __pyx_v_key); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(9, 142, __pyx_L1_error)
    __pyx_v_key_fitness = __pyx_t_5;

    /* "src/map.pxi":143
 *                 raise ValueError(f"'{value}' is not a mapping")
 *             key_fitness = key_type.get_value_fitness(key)
 *             if key_fitness == FIT_NONE:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"'{key}' is not a valid key for map type")
 *             elif key_fitness < FIT_EXACT:
 */
    __pyx_t_6 = (__pyx_v_key_fitness == __pyx_v_5cavro_FIT_NONE);
    if (unlikely(__pyx_t_6)) {

      /* "src/map.pxi":144
 *             key_fitness = key_type.get_value_fitness(key)
 *             if key_fitness == FIT_NONE:
 *                 raise ValueError(f"'{key}' is not a valid key for map type")             # <<<<<<<<<<<<<<
 *             elif key_fitness < FIT_EXACT:
 *                 return self._make_converted_map(iter(value))
 */
      __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_15 = 0;
      __pyx_t_16 = 127;
      __Pyx_INCREF(__pyx_kp_u__6);
      __pyx_t_15 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__6);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u__6);
      __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_key, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_16;
      __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_INCREF(__pyx_kp_u_is_not_a_valid_key_for_map_type);
      __pyx_t_15 += 33;
      __Pyx_GIVEREF(__pyx_kp_u_is_not_a_valid_key_for_map_type);
      PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_is_not_a_valid_key_for_map_type);
      __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_4, 3, __pyx_t_15, __pyx_t_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(9, 144, __pyx_L1_error)

      /* "src/map.pxi":143
 *                 raise ValueError(f"'{value}' is not a mapping")
 *             key_fitness = key_type.get_value_fitness(key)
 *             if key_fitness == FIT_NONE:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"'{key}' is not a valid key for map type")
 *             elif key_fitness < FIT_EXACT:
 */
    }

    /* "src/map.pxi":145
 *             if key_fitness == FIT_NONE:
 *                 raise ValueError(f"'{key}' is not a valid key for map type")
 *             elif key_fitness < FIT_EXACT:             # <<<<<<<<<<<<<<
 *                 return self._make_converted_map(iter(value))
 * 
 */
    __pyx_t_6 = (__pyx_v_key_fitness < __pyx_v_5cavro_FIT_EXACT);
    if (__pyx_t_6) {

      /* "src/map.pxi":146
 *                 raise ValueError(f"'{key}' is not a valid key for map type")
 *             elif key_fitness < FIT_EXACT:
 *                 return self._make_converted_map(iter(value))             # <<<<<<<<<<<<<<
 * 
 *             value_fitness = value_type.get_value_fitness(item_value)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_MapType *)__pyx_v_self->__pyx_base.__pyx_vtab)->_make_converted_map(__pyx_v_self, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "src/map.pxi":145
 *             if key_fitness == FIT_NONE:
 *                 raise ValueError(f"'{key}' is not a valid key for map type")
 *             elif key_fitness < FIT_EXACT:             # <<<<<<<<<<<<<<
 *                 return self._make_converted_map(iter(value))
 * 
 */
    }

    /* "src/map.pxi":148
 *                 return self._make_converted_map(iter(value))
 * 
 *             value_fitness = value_type.get_value_fitness(item_value)             # <<<<<<<<<<<<<<
 *             if value_fitness == FIT_NONE:
 *                 raise ValueError(f"'{item_value}' is not a valid value for map type: {type(self.value_type).__name__}")
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_value_type->__pyx_vtab)->get_value_fitness(__pyx_v_value_type, __pyx_v_item_value); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(9, 148, __pyx_L1_error)
    __pyx_v_value_fitness = __pyx_t_5;

    /* "src/map.pxi":149
 * 
 *             value_fitness = value_type.get_value_fitness(item_value)
 *             if value_fitness == FIT_NONE:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"'{item_value}' is not a valid value for map type: {type(self.value_type).__name__}")
 *             elif value_fitness < FIT_EXACT:
 */
    __pyx_t_6 = (__pyx_v_value_fitness == __pyx_v_5cavro_FIT_NONE);
    if (unlikely(__pyx_t_6)) {

      /* "src/map.pxi":150
 *             value_fitness = value_type.get_value_fitness(item_value)
 *             if value_fitness == FIT_NONE:
 *                 raise ValueError(f"'{item_value}' is not a valid value for map type: {type(self.value_type).__name__}")             # <<<<<<<<<<<<<<
 *             elif value_fitness < FIT_EXACT:
 *                 return self._make_converted_map(iter(value))
 */
      __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_15 = 0;
      __pyx_t_16 = 127;
      __Pyx_INCREF(__pyx_kp_u__6);
      __pyx_t_15 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__6);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u__6);
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_item_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_16;
      __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_INCREF(__pyx_kp_u_is_not_a_valid_value_for_map_ty);
      __pyx_t_15 += 37;
      __Pyx_GIVEREF(__pyx_kp_u_is_not_a_valid_value_for_map_ty);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_is_not_a_valid_value_for_map_ty);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self->value_type))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_16;
      __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_2, 4, __pyx_t_15, __pyx_t_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(9, 150, __pyx_L1_error)

      /* "src/map.pxi":149
 * 
 *             value_fitness = value_type.get_value_fitness(item_value)
 *             if value_fitness == FIT_NONE:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"'{item_value}' is not a valid value for map type: {type(self.value_type).__name__}")
 *             elif value_fitness < FIT_EXACT:
 */
    }

    /* "src/map.pxi":151
 *             if value_fitness == FIT_NONE:
 *                 raise ValueError(f"'{item_value}' is not a valid value for map type: {type(self.value_type).__name__}")
 *             elif value_fitness < FIT_EXACT:             # <<<<<<<<<<<<<<
 *                 return self._make_converted_map(iter(value))
 *         return orig_value
 */
    __pyx_t_6 = (__pyx_v_value_fitness < __pyx_v_5cavro_FIT_EXACT);
    if (__pyx_t_6) {

      /* "src/map.pxi":152
 *                 raise ValueError(f"'{item_value}' is not a valid value for map type: {type(self.value_type).__name__}")
 *             elif value_fitness < FIT_EXACT:
 *                 return self._make_converted_map(iter(value))             # <<<<<<<<<<<<<<
 *         return orig_value
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_MapType *)__pyx_v_self->__pyx_base.__pyx_vtab)->_make_converted_map(__pyx_v_self, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "src/map.pxi":151
 *             if value_fitness == FIT_NONE:
 *                 raise ValueError(f"'{item_value}' is not a valid value for map type: {type(self.value_type).__name__}")
 *             elif value_fitness < FIT_EXACT:             # <<<<<<<<<<<<<<
 *                 return self._make_converted_map(iter(value))
 *         return orig_value
 */
    }

    /* "src/map.pxi":137
 * 
 *         it = iter(value)
 *         for item in it:             # <<<<<<<<<<<<<<
 *             try:
 *                 key, item_value = item
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/map.pxi":153
 *             elif value_fitness < FIT_EXACT:
 *                 return self._make_converted_map(iter(value))
 *         return orig_value             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_orig_value);
  __pyx_r = __pyx_v_orig_value;
  goto __pyx_L0;

  /* "src/map.pxi":122
 *         return out
 * 
 *     cpdef object convert_value(self, object orig_value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef int key_fitness
 *         cdef int value_fitness
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("cavro.MapType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_key_type);
  __Pyx_XDECREF((PyObject *)__pyx_v_value_type);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_it);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_item_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7MapType_10convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7MapType_10convert_value = {"convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7MapType_10convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7MapType_10convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_orig_value = 0;
  PyObject *__pyx_v_check_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_orig_value,&__pyx_n_s_check_value,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_True));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_orig_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(9, 122, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_check_value);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(9, 122, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "convert_value") < 0)) __PYX_ERR(9, 122, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_orig_value = values[0];
    __pyx_v_check_value = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convert_value", 0, 1, 2, __pyx_nargs); __PYX_ERR(9, 122, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.MapType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_7MapType_9convert_value(((struct __pyx_obj_5cavro_MapType *)__pyx_v_self), __pyx_v_orig_value, __pyx_v_check_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7MapType_9convert_value(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_orig_value, PyObject *__pyx_v_check_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_convert_value __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.check_value = __pyx_v_check_value;
  __pyx_t_1 = __pyx_vtabptr_5cavro_MapType->__pyx_base.convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_orig_value, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.MapType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":155
 *         return orig_value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return dict_to_canonical({
 *             'type': 'map',
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_7MapType_canonical_form(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/map.pxi":156
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return dict_to_canonical({             # <<<<<<<<<<<<<<
 *             'type': 'map',
 *             'values': self.value_type.canonical_form(created)
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);

  /* "src/map.pxi":157
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return dict_to_canonical({
 *             'type': 'map',             # <<<<<<<<<<<<<<
 *             'values': self.value_type.canonical_form(created)
 *         })
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_type, __pyx_n_u_map) < 0) __PYX_ERR(9, 157, __pyx_L1_error)

  /* "src/map.pxi":158
 *         return dict_to_canonical({
 *             'type': 'map',
 *             'values': self.value_type.canonical_form(created)             # <<<<<<<<<<<<<<
 *         })
 * 
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->value_type->__pyx_vtab)->canonical_form(__pyx_v_self->value_type, __pyx_v_created)); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_values, __pyx_t_2) < 0) __PYX_ERR(9, 157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/map.pxi":156
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return dict_to_canonical({             # <<<<<<<<<<<<<<
 *             'type': 'map',
 *             'values': self.value_type.canonical_form(created)
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_5cavro_dict_to_canonical(__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/map.pxi":155
 *         return orig_value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return dict_to_canonical({
 *             'type': 'map',
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.MapType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":161
 *         })
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         cdef MapType cloned
 *         cdef MapType writer_map
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_7MapType__for_writer(struct __pyx_obj_5cavro_MapType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer) {
  struct __pyx_obj_5cavro_MapType *__pyx_v_cloned = 0;
  struct __pyx_obj_5cavro_MapType *__pyx_v_writer_map = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_promoted_value = NULL;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_writer", 1);

  /* "src/map.pxi":164
 *         cdef MapType cloned
 *         cdef MapType writer_map
 *         if isinstance(writer, MapType):             # <<<<<<<<<<<<<<
 *             writer_map = writer
 *             promoted_value = self.value_type._for_writer(writer.value_type)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_MapType); 
  if (__pyx_t_1) {

    /* "src/map.pxi":165
 *         cdef MapType writer_map
 *         if isinstance(writer, MapType):
 *             writer_map = writer             # <<<<<<<<<<<<<<
 *             promoted_value = self.value_type._for_writer(writer.value_type)
 *             if promoted_value is not None:
 */
    if (!(likely(((((PyObject *)__pyx_v_writer)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_MapType))))) __PYX_ERR(9, 165, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)__pyx_v_writer);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_writer_map = ((struct __pyx_obj_5cavro_MapType *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/map.pxi":166
 *         if isinstance(writer, MapType):
 *             writer_map = writer
 *             promoted_value = self.value_type._for_writer(writer.value_type)             # <<<<<<<<<<<<<<
 *             if promoted_value is not None:
 *                 cloned = self.clone_base()
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_writer), __pyx_n_s_value_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(9, 166, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->value_type->__pyx_vtab)->_for_writer(__pyx_v_self->value_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2))); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_promoted_value = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/map.pxi":167
 *             writer_map = writer
 *             promoted_value = self.value_type._for_writer(writer.value_type)
 *             if promoted_value is not None:             # <<<<<<<<<<<<<<
 *                 cloned = self.clone_base()
 *                 cloned.key_type = writer_map.key_type
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_promoted_value) != Py_None);
    if (__pyx_t_1) {

      /* "src/map.pxi":168
 *             promoted_value = self.value_type._for_writer(writer.value_type)
 *             if promoted_value is not None:
 *                 cloned = self.clone_base()             # <<<<<<<<<<<<<<
 *                 cloned.key_type = writer_map.key_type
 *                 cloned.value_type = promoted_value
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_MapType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_MapType))))) __PYX_ERR(9, 168, __pyx_L1_error)
      __pyx_v_cloned = ((struct __pyx_obj_5cavro_MapType *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "src/map.pxi":169
 *             if promoted_value is not None:
 *                 cloned = self.clone_base()
 *                 cloned.key_type = writer_map.key_type             # <<<<<<<<<<<<<<
 *                 cloned.value_type = promoted_value
 *                 return cloned
 */
      __pyx_t_3 = ((PyObject *)__pyx_v_writer_map->key_type);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_GOTREF((PyObject *)__pyx_v_cloned->key_type);
      __Pyx_DECREF((PyObject *)__pyx_v_cloned->key_type);
      __pyx_v_cloned->key_type = ((struct __pyx_obj_5cavro_StringType *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "src/map.pxi":170
 *                 cloned = self.clone_base()
 *                 cloned.key_type = writer_map.key_type
 *                 cloned.value_type = promoted_value             # <<<<<<<<<<<<<<
 *                 return cloned
 * 
 */
      __Pyx_INCREF((PyObject *)__pyx_v_promoted_value);
      __Pyx_GIVEREF((PyObject *)__pyx_v_promoted_value);
      __Pyx_GOTREF((PyObject *)__pyx_v_cloned->value_type);
      __Pyx_DECREF((PyObject *)__pyx_v_cloned->value_type);
      __pyx_v_cloned->value_type = __pyx_v_promoted_value;

      /* "src/map.pxi":171
 *                 cloned.key_type = writer_map.key_type
 *                 cloned.value_type = promoted_value
 *                 return cloned             # <<<<<<<<<<<<<<
 * 
 *     cdef bint accepts_missing_value(self):
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __Pyx_INCREF((PyObject *)__pyx_v_cloned);
      __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_cloned);
      goto __pyx_L0;

      /* "src/map.pxi":167
 *             writer_map = writer
 *             promoted_value = self.value_type._for_writer(writer.value_type)
 *             if promoted_value is not None:             # <<<<<<<<<<<<<<
 *                 cloned = self.clone_base()
 *                 cloned.key_type = writer_map.key_type
 */
    }

    /* "src/map.pxi":164
 *         cdef MapType cloned
 *         cdef MapType writer_map
 *         if isinstance(writer, MapType):             # <<<<<<<<<<<<<<
 *             writer_map = writer
 *             promoted_value = self.value_type._for_writer(writer.value_type)
 */
  }

  /* "src/map.pxi":161
 *         })
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         cdef MapType cloned
 *         cdef MapType writer_map
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.MapType._for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cloned);
  __Pyx_XDECREF((PyObject *)__pyx_v_writer_map);
  __Pyx_XDECREF((PyObject *)__pyx_v_promoted_value);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/map.pxi":173
 *                 return cloned
 * 
 *     cdef bint accepts_missing_value(self):             # <<<<<<<<<<<<<<
 *         if self.options .missing_values_can_be_empty_container:
 *             return True
 */

static int __pyx_f_5cavro_7MapType_accepts_missing_value(struct __pyx_obj_5cavro_MapType *__pyx_v_self) {
  int __pyx_r;

  /* "src/map.pxi":174
 * 
 *     cdef bint accepts_missing_value(self):
 *         if self.options .missing_values_can_be_empty_container:             # <<<<<<<<<<<<<<
 *             return True
 */
  if (__pyx_v_self->__pyx_base.options->missing_values_can_be_empty_container) {

    /* "src/map.pxi":175
 *     cdef bint accepts_missing_value(self):
 *         if self.options .missing_values_can_be_empty_container:
 *             return True             # <<<<<<<<<<<<<<
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "src/map.pxi":174
 * 
 *     cdef bint accepts_missing_value(self):
 *         if self.options .missing_values_can_be_empty_container:             # <<<<<<<<<<<<<<
 *             return True
 */
  }

  /* "src/map.pxi":173
 *                 return cloned
 * 
 *     cdef bint accepts_missing_value(self):             # <<<<<<<<<<<<<<
 *         if self.options .missing_values_can_be_empty_container:
 *             return True
 */

  /* function exit code */
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/map.pxi":7
 * 
 *     cdef StringType key_type
 *     cdef readonly AvroType value_type             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, schema, source, namespace):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7MapType_10value_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_7MapType_10value_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_7MapType_10value_type___get__(((struct __pyx_obj_5cavro_MapType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7MapType_10value_type___get__(struct __pyx_obj_5cavro_MapType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->value_type);
  __pyx_r = ((PyObject *)__pyx_v_self->value_type);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7MapType_12__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7MapType_12__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7MapType_12__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7MapType_12__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_7MapType_11__reduce_cython__(((struct __pyx_obj_5cavro_MapType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7MapType_11__reduce_cython__(struct __pyx_obj_5cavro_MapType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.key_type, self.metadata, self.options, self.value_adapters, self.value_type)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->key_type);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->key_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->key_type))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->value_type);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->value_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, ((PyObject *)__pyx_v_self->value_type))) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.key_type, self.metadata, self.options, self.value_adapters, self.value_type)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.key_type, self.metadata, self.options, self.value_adapters, self.value_type)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.key_type is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None or self.value_type is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.key_type, self.metadata, self.options, self.value_adapters, self.value_type)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.key_type is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None or self.value_type is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_MapType, (type(self), 0xaa8230c, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->key_type) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->value_type) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.key_type is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None or self.value_type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MapType, (type(self), 0xaa8230c, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.key_type is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None or self.value_type is not None
 *     if use_setstate:
 *         return __pyx_unpickle_MapType, (type(self), 0xaa8230c, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_MapType, (type(self), 0xaa8230c, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_MapType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_178791180);
    __Pyx_GIVEREF(__pyx_int_178791180);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_178791180)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.key_type is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None or self.value_type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MapType, (type(self), 0xaa8230c, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_MapType, (type(self), 0xaa8230c, None), state
 *     else:
 *         return __pyx_unpickle_MapType, (type(self), 0xaa8230c, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MapType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_MapType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_178791180);
    __Pyx_GIVEREF(__pyx_int_178791180);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_178791180)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.MapType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MapType, (type(self), 0xaa8230c, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MapType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7MapType_14__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7MapType_14__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7MapType_14__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7MapType_14__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.MapType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_7MapType_13__setstate_cython__(((struct __pyx_obj_5cavro_MapType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7MapType_13__setstate_cython__(struct __pyx_obj_5cavro_MapType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_MapType, (type(self), 0xaa8230c, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MapType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_MapType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MapType, (type(self), 0xaa8230c, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MapType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.MapType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/null.pxi":6
 *     type_name = "null"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

static PyObject *__pyx_pw_5cavro_8NullType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8NullType_copy(struct __pyx_obj_5cavro_NullType *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 6, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_8NullType_1copy)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 6, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(21, 6, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/null.pxi":7
 * 
 *     cpdef AvroType copy(self):
 *         return self.clone_base()             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_NullType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/null.pxi":6
 *     type_name = "null"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.NullType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8NullType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8NullType_1copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8NullType_1copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8NullType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_8NullType_copy(((struct __pyx_obj_5cavro_NullType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8NullType_copy(struct __pyx_obj_5cavro_NullType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_8NullType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.NullType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/null.pxi":9
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

static PyObject *__pyx_f_5cavro_8NullType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/null.pxi":10
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {'type'})             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(21, 10, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/null.pxi":9
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.NullType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/null.pxi":12
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_pw_5cavro_8NullType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_8NullType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_schema_extra); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 12, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_8NullType_3_get_schema_extra)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_created};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 12, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_2))) __PYX_ERR(21, 12, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/null.pxi":13
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/null.pxi":12
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.NullType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8NullType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8NullType_3_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8NullType_3_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8NullType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(21, 12, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(21, 12, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(21, 12, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.NullType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(21, 12, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_8NullType_2_get_schema_extra(((struct __pyx_obj_5cavro_NullType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8NullType_2_get_schema_extra(struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8NullType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.NullType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/null.pxi":15
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         if value is None or value is MISSING_VALUE or (self.options.allow_false_values_for_null and not value):
 *             return 0
 */

static int __pyx_f_5cavro_8NullType__binary_buffer_encode(struct __pyx_obj_5cavro_NullType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 1);

  /* "src/null.pxi":16
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         if value is None or value is MISSING_VALUE or (self.options.allow_false_values_for_null and not value):             # <<<<<<<<<<<<<<
 *             return 0
 *         raise ValueError(f'{repr(value)} not compatible with NullType')
 */
  __pyx_t_2 = (__pyx_v_value == Py_None);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = (__pyx_v_value == __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  if (__pyx_v_self->__pyx_base.options->allow_false_values_for_null) {
  } else {
    __pyx_t_1 = __pyx_v_self->__pyx_base.options->allow_false_values_for_null;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(21, 16, __pyx_L1_error)
  __pyx_t_4 = (!__pyx_t_2);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/null.pxi":17
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         if value is None or value is MISSING_VALUE or (self.options.allow_false_values_for_null and not value):
 *             return 0             # <<<<<<<<<<<<<<
 *         raise ValueError(f'{repr(value)} not compatible with NullType')
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "src/null.pxi":16
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         if value is None or value is MISSING_VALUE or (self.options.allow_false_values_for_null and not value):             # <<<<<<<<<<<<<<
 *             return 0
 *         raise ValueError(f'{repr(value)} not compatible with NullType')
 */
  }

  /* "src/null.pxi":18
 *         if value is None or value is MISSING_VALUE or (self.options.allow_false_values_for_null and not value):
 *             return 0
 *         raise ValueError(f'{repr(value)} not compatible with NullType')             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_t_3 = PyObject_Repr(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(21, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_5, __pyx_kp_u_not_compatible_with_NullType); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(21, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_5, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __PYX_ERR(21, 18, __pyx_L1_error)

  /* "src/null.pxi":15
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         if value is None or value is MISSING_VALUE or (self.options.allow_false_values_for_null and not value):
 *             return 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.NullType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/null.pxi":20
 *         raise ValueError(f'{repr(value)} not compatible with NullType')
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

static PyObject *__pyx_f_5cavro_8NullType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/null.pxi":21
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "src/null.pxi":20
 *         raise ValueError(f'{repr(value)} not compatible with NullType')
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/null.pxi":23
 *         return None
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         if value is None:
 *             return FIT_EXACT
 */

static int __pyx_f_5cavro_8NullType__get_value_fitness(struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 1);

  /* "src/null.pxi":24
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         if value is None:             # <<<<<<<<<<<<<<
 *             return FIT_EXACT
 *         if self.options.allow_false_values_for_null and not value:
 */
  __pyx_t_1 = (__pyx_v_value == Py_None);
  if (__pyx_t_1) {

    /* "src/null.pxi":25
 *     cdef int _get_value_fitness(self, value) except -1:
 *         if value is None:
 *             return FIT_EXACT             # <<<<<<<<<<<<<<
 *         if self.options.allow_false_values_for_null and not value:
 *             return FIT_POOR
 */
    __pyx_r = __pyx_v_5cavro_FIT_EXACT;
    goto __pyx_L0;

    /* "src/null.pxi":24
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         if value is None:             # <<<<<<<<<<<<<<
 *             return FIT_EXACT
 *         if self.options.allow_false_values_for_null and not value:
 */
  }

  /* "src/null.pxi":26
 *         if value is None:
 *             return FIT_EXACT
 *         if self.options.allow_false_values_for_null and not value:             # <<<<<<<<<<<<<<
 *             return FIT_POOR
 *         if self.options.missing_values_can_be_null and value is MISSING_VALUE:
 */
  if (__pyx_v_self->__pyx_base.options->allow_false_values_for_null) {
  } else {
    __pyx_t_1 = __pyx_v_self->__pyx_base.options->allow_false_values_for_null;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(21, 26, __pyx_L1_error)
  __pyx_t_3 = (!__pyx_t_2);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/null.pxi":27
 *             return FIT_EXACT
 *         if self.options.allow_false_values_for_null and not value:
 *             return FIT_POOR             # <<<<<<<<<<<<<<
 *         if self.options.missing_values_can_be_null and value is MISSING_VALUE:
 *             return FIT_OK
 */
    __pyx_r = __pyx_v_5cavro_FIT_POOR;
    goto __pyx_L0;

    /* "src/null.pxi":26
 *         if value is None:
 *             return FIT_EXACT
 *         if self.options.allow_false_values_for_null and not value:             # <<<<<<<<<<<<<<
 *             return FIT_POOR
 *         if self.options.missing_values_can_be_null and value is MISSING_VALUE:
 */
  }

  /* "src/null.pxi":28
 *         if self.options.allow_false_values_for_null and not value:
 *             return FIT_POOR
 *         if self.options.missing_values_can_be_null and value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *             return FIT_OK
 *         return FIT_NONE
 */
  if (__pyx_v_self->__pyx_base.options->missing_values_can_be_null) {
  } else {
    __pyx_t_1 = __pyx_v_self->__pyx_base.options->missing_values_can_be_null;
    goto __pyx_L8_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_4)) __PYX_ERR(21, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__pyx_v_value == __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/null.pxi":29
 *             return FIT_POOR
 *         if self.options.missing_values_can_be_null and value is MISSING_VALUE:
 *             return FIT_OK             # <<<<<<<<<<<<<<
 *         return FIT_NONE
 * 
 */
    __pyx_r = __pyx_v_5cavro_FIT_OK;
    goto __pyx_L0;

    /* "src/null.pxi":28
 *         if self.options.allow_false_values_for_null and not value:
 *             return FIT_POOR
 *         if self.options.missing_values_can_be_null and value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *             return FIT_OK
 *         return FIT_NONE
 */
  }

  /* "src/null.pxi":30
 *         if self.options.missing_values_can_be_null and value is MISSING_VALUE:
 *             return FIT_OK
 *         return FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
  __pyx_r = __pyx_v_5cavro_FIT_NONE;
  goto __pyx_L0;

  /* "src/null.pxi":23
 *         return None
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         if value is None:
 *             return FIT_EXACT
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.NullType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/null.pxi":32
 *         return FIT_NONE
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

static PyObject *__pyx_f_5cavro_8NullType__json_format(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_json_format", 1);

  /* "src/null.pxi":33
 * 
 *     cdef _json_format(self, value):
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "src/null.pxi":32
 *         return FIT_NONE
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/null.pxi":35
 *         return None
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if value is not None:
 *             raise ValueError(f'Expected null, got {repr(value)}')
 */

static PyObject *__pyx_f_5cavro_8NullType__json_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/null.pxi":36
 * 
 *     cdef _json_decode(self, value):
 *         if value is not None:             # <<<<<<<<<<<<<<
 *             raise ValueError(f'Expected null, got {repr(value)}')
 *         return None
 */
  __pyx_t_1 = (__pyx_v_value != Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "src/null.pxi":37
 *     cdef _json_decode(self, value):
 *         if value is not None:
 *             raise ValueError(f'Expected null, got {repr(value)}')             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
    __pyx_t_2 = PyObject_Repr(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Expected_null_got, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(21, 37, __pyx_L1_error)

    /* "src/null.pxi":36
 * 
 *     cdef _json_decode(self, value):
 *         if value is not None:             # <<<<<<<<<<<<<<
 *             raise ValueError(f'Expected null, got {repr(value)}')
 *         return None
 */
  }

  /* "src/null.pxi":38
 *         if value is not None:
 *             raise ValueError(f'Expected null, got {repr(value)}')
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "src/null.pxi":35
 *         return None
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if value is not None:
 *             raise ValueError(f'Expected null, got {repr(value)}')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.NullType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/null.pxi":40
 *         return None
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

static PyObject *__pyx_pw_5cavro_8NullType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_8NullType__convert_value(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_convert_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 40, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_8NullType_5_convert_value)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_value};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 40, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/null.pxi":41
 * 
 *     cpdef object _convert_value(self, object value):
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "src/null.pxi":40
 *         return None
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.NullType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8NullType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8NullType_5_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8NullType_5_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8NullType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(21, 40, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(21, 40, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(21, 40, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.NullType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8NullType_4_convert_value(((struct __pyx_obj_5cavro_NullType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8NullType_4_convert_value(struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8NullType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.NullType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/null.pxi":43
 *         return None
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"null"')
 * 
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_8NullType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_NullType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/null.pxi":44
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"null"')             # <<<<<<<<<<<<<<
 * 
 *     cdef bint accepts_missing_value(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_tuple__50, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/null.pxi":43
 *         return None
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"null"')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.NullType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/null.pxi":46
 *         return _CanonicalForm('"null"')
 * 
 *     cdef bint accepts_missing_value(self):             # <<<<<<<<<<<<<<
 *         if self.options.missing_values_can_be_null:
 *             return True
 */

static int __pyx_f_5cavro_8NullType_accepts_missing_value(struct __pyx_obj_5cavro_NullType *__pyx_v_self) {
  int __pyx_r;

  /* "src/null.pxi":47
 * 
 *     cdef bint accepts_missing_value(self):
 *         if self.options.missing_values_can_be_null:             # <<<<<<<<<<<<<<
 *             return True
 * 
 */
  if (__pyx_v_self->__pyx_base.options->missing_values_can_be_null) {

    /* "src/null.pxi":48
 *     cdef bint accepts_missing_value(self):
 *         if self.options.missing_values_can_be_null:
 *             return True             # <<<<<<<<<<<<<<
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "src/null.pxi":47
 * 
 *     cdef bint accepts_missing_value(self):
 *         if self.options.missing_values_can_be_null:             # <<<<<<<<<<<<<<
 *             return True
 * 
 */
  }

  /* "src/null.pxi":46
 *         return _CanonicalForm('"null"')
 * 
 *     cdef bint accepts_missing_value(self):             # <<<<<<<<<<<<<<
 *         if self.options.missing_values_can_be_null:
 *             return True
 */

  /* function exit code */
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/null.pxi":50
 *             return True
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):             # <<<<<<<<<<<<<<
 *         if self.options.missing_values_can_be_null and schema_default is NO_DEFAULT:
 *             return MISSING_VALUE
 */

static PyObject *__pyx_f_5cavro_8NullType_resolve_default_value(struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v_schema_default, PyObject *__pyx_v_field) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resolve_default_value", 1);

  /* "src/null.pxi":51
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if self.options.missing_values_can_be_null and schema_default is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             return MISSING_VALUE
 *         return AvroType.resolve_default_value(self, schema_default, field)
 */
  if (__pyx_v_self->__pyx_base.options->missing_values_can_be_null) {
  } else {
    __pyx_t_1 = __pyx_v_self->__pyx_base.options->missing_values_can_be_null;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__pyx_v_schema_default == __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/null.pxi":52
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if self.options.missing_values_can_be_null and schema_default is NO_DEFAULT:
 *             return MISSING_VALUE             # <<<<<<<<<<<<<<
 *         return AvroType.resolve_default_value(self, schema_default, field)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "src/null.pxi":51
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if self.options.missing_values_can_be_null and schema_default is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             return MISSING_VALUE
 *         return AvroType.resolve_default_value(self, schema_default, field)
 */
  }

  /* "src/null.pxi":53
 *         if self.options.missing_values_can_be_null and schema_default is NO_DEFAULT:
 *             return MISSING_VALUE
 *         return AvroType.resolve_default_value(self, schema_default, field)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_f_5cavro_8AvroType_resolve_default_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_schema_default, __pyx_v_field); if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/null.pxi":50
 *             return True
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):             # <<<<<<<<<<<<<<
 *         if self.options.missing_values_can_be_null and schema_default is NO_DEFAULT:
 *             return MISSING_VALUE
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.NullType.resolve_default_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8NullType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8NullType_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8NullType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8NullType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_8NullType_6__reduce_cython__(((struct __pyx_obj_5cavro_NullType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8NullType_6__reduce_cython__(struct __pyx_obj_5cavro_NullType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_NullType, (type(self), 0x84c5042, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NullType, (type(self), 0x84c5042, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_NullType, (type(self), 0x84c5042, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_NullType, (type(self), 0x84c5042, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_NullType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NullType, (type(self), 0x84c5042, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_NullType, (type(self), 0x84c5042, None), state
 *     else:
 *         return __pyx_unpickle_NullType, (type(self), 0x84c5042, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NullType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_NullType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.NullType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NullType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NullType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8NullType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8NullType_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8NullType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8NullType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.NullType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8NullType_8__setstate_cython__(((struct __pyx_obj_5cavro_NullType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8NullType_8__setstate_cython__(struct __pyx_obj_5cavro_NullType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_NullType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NullType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_NullType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NullType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NullType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.NullType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":15
 *     type_name = "boolean"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

static PyObject *__pyx_pw_5cavro_8BoolType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8BoolType_copy(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "src/numeric_types.pxi":16
 * 
 *     cpdef AvroType copy(self):
 *         return self.clone_base()             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_BoolType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":15
 *     type_name = "boolean"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BoolType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8BoolType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8BoolType_1copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8BoolType_1copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8BoolType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_8BoolType_copy(((struct __pyx_obj_5cavro_BoolType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8BoolType_copy(struct __pyx_obj_5cavro_BoolType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_8BoolType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BoolType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":18
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

static PyObject *__pyx_f_5cavro_8BoolType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/numeric_types.pxi":19
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {'type'})             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(10, 19, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":18
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.BoolType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":21
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_pw_5cavro_8BoolType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_8BoolType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_BoolType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);

  /* "src/numeric_types.pxi":22
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":21
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BoolType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8BoolType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8BoolType_3_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8BoolType_3_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8BoolType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 21, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(10, 21, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(10, 21, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.BoolType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(10, 21, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_8BoolType_2_get_schema_extra(((struct __pyx_obj_5cavro_BoolType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8BoolType_2_get_schema_extra(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8BoolType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BoolType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":24
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef bint bool_val = self._convert_value(value)
 *         buffer.write_u8(bool_val)
 */

static int __pyx_f_5cavro_8BoolType__binary_buffer_encode(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  int __pyx_v_bool_val;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 1);

  /* "src/numeric_types.pxi":25
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef bint bool_val = self._convert_value(value)             # <<<<<<<<<<<<<<
 *         buffer.write_u8(bool_val)
 * 
 */
  __pyx_t_1 = __pyx_f_5cavro_8BoolType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(10, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_bool_val = __pyx_t_2;

  /* "src/numeric_types.pxi":26
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef bint bool_val = self._convert_value(value)
 *         buffer.write_u8(bool_val)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_buffer->__pyx_vtab)->write_u8(__pyx_v_buffer, __pyx_v_bool_val); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(10, 26, __pyx_L1_error)

  /* "src/numeric_types.pxi":24
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef bint bool_val = self._convert_value(value)
 *         buffer.write_u8(bool_val)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BoolType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":28
 *         buffer.write_u8(bool_val)
 * 
 *     cdef object _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return py_bool(buffer.read_u8())
 * 
 */

static PyObject *__pyx_f_5cavro_8BoolType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_BoolType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint8_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/numeric_types.pxi":29
 * 
 *     cdef object _binary_buffer_decode(self, _Reader buffer):
 *         return py_bool(buffer.read_u8())             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_buffer->__pyx_vtab)->read_u8(__pyx_v_buffer); if (unlikely(__pyx_t_1 == ((uint8_t)0xba) && PyErr_Occurred())) __PYX_ERR(10, 29, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_uint8_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7cpython_4bool_bool), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":28
 *         buffer.write_u8(bool_val)
 * 
 *     cdef object _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return py_bool(buffer.read_u8())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.BoolType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":31
 *         return py_bool(buffer.read_u8())
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         if isinstance(value, (py_bool, bool_)):
 *             return FIT_EXACT
 */

static int __pyx_f_5cavro_8BoolType__get_value_fitness(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 1);

  /* "src/numeric_types.pxi":32
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         if isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_boolean:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/numeric_types.pxi":33
 *     cdef int _get_value_fitness(self, value) except -1:
 *         if isinstance(value, (py_bool, bool_)):
 *             return FIT_EXACT             # <<<<<<<<<<<<<<
 *         if self.options.coerce_values_to_boolean:
 *             try:
 */
    __pyx_r = __pyx_v_5cavro_FIT_EXACT;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":32
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         if isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_boolean:
 */
  }

  /* "src/numeric_types.pxi":34
 *         if isinstance(value, (py_bool, bool_)):
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_boolean:             # <<<<<<<<<<<<<<
 *             try:
 *                 py_bool(value)
 */
  if (__pyx_v_self->__pyx_base.options->coerce_values_to_boolean) {

    /* "src/numeric_types.pxi":35
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_boolean:
 *             try:             # <<<<<<<<<<<<<<
 *                 py_bool(value)
 *             except (ValueError, TypeError):
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "src/numeric_types.pxi":36
 *         if self.options.coerce_values_to_boolean:
 *             try:
 *                 py_bool(value)             # <<<<<<<<<<<<<<
 *             except (ValueError, TypeError):
 *                 return FIT_NONE
 */
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7cpython_4bool_bool), __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 36, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "src/numeric_types.pxi":35
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_boolean:
 *             try:             # <<<<<<<<<<<<<<
 *                 py_bool(value)
 *             except (ValueError, TypeError):
 */
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L12_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "src/numeric_types.pxi":37
 *             try:
 *                 py_bool(value)
 *             except (ValueError, TypeError):             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *             return FIT_POOR
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_ValueError, __pyx_builtin_TypeError);
      if (__pyx_t_7) {
        __Pyx_AddTraceback("cavro.BoolType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(10, 37, __pyx_L9_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);

        /* "src/numeric_types.pxi":38
 *                 py_bool(value)
 *             except (ValueError, TypeError):
 *                 return FIT_NONE             # <<<<<<<<<<<<<<
 *             return FIT_POOR
 *         return FIT_NONE
 */
        __pyx_r = __pyx_v_5cavro_FIT_NONE;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L10_except_return;
      }
      goto __pyx_L9_except_error;

      /* "src/numeric_types.pxi":35
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_boolean:
 *             try:             # <<<<<<<<<<<<<<
 *                 py_bool(value)
 *             except (ValueError, TypeError):
 */
      __pyx_L9_except_error:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L10_except_return:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L0;
      __pyx_L12_try_end:;
    }

    /* "src/numeric_types.pxi":39
 *             except (ValueError, TypeError):
 *                 return FIT_NONE
 *             return FIT_POOR             # <<<<<<<<<<<<<<
 *         return FIT_NONE
 * 
 */
    __pyx_r = __pyx_v_5cavro_FIT_POOR;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":34
 *         if isinstance(value, (py_bool, bool_)):
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_boolean:             # <<<<<<<<<<<<<<
 *             try:
 *                 py_bool(value)
 */
  }

  /* "src/numeric_types.pxi":40
 *                 return FIT_NONE
 *             return FIT_POOR
 *         return FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
  __pyx_r = __pyx_v_5cavro_FIT_NONE;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":31
 *         return py_bool(buffer.read_u8())
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         if isinstance(value, (py_bool, bool_)):
 *             return FIT_EXACT
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cavro.BoolType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":42
 *         return FIT_NONE
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (py_bool, bool_)):
 *             return value
 */

static PyObject *__pyx_pw_5cavro_8BoolType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_8BoolType__convert_value(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);

  /* "src/numeric_types.pxi":43
 * 
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *             return value
 *         if self.options.coerce_values_to_boolean:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/numeric_types.pxi":44
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, (py_bool, bool_)):
 *             return value             # <<<<<<<<<<<<<<
 *         if self.options.coerce_values_to_boolean:
 *             return py_bool(value)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":43
 * 
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *             return value
 *         if self.options.coerce_values_to_boolean:
 */
  }

  /* "src/numeric_types.pxi":45
 *         if isinstance(value, (py_bool, bool_)):
 *             return value
 *         if self.options.coerce_values_to_boolean:             # <<<<<<<<<<<<<<
 *             return py_bool(value)
 *         raise InvalidValue(value, self)
 */
  if (__pyx_v_self->__pyx_base.options->coerce_values_to_boolean) {

    /* "src/numeric_types.pxi":46
 *             return value
 *         if self.options.coerce_values_to_boolean:
 *             return py_bool(value)             # <<<<<<<<<<<<<<
 *         raise InvalidValue(value, self)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7cpython_4bool_bool), __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":45
 *         if isinstance(value, (py_bool, bool_)):
 *             return value
 *         if self.options.coerce_values_to_boolean:             # <<<<<<<<<<<<<<
 *             return py_bool(value)
 *         raise InvalidValue(value, self)
 */
  }

  /* "src/numeric_types.pxi":47
 *         if self.options.coerce_values_to_boolean:
 *             return py_bool(value)
 *         raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_value, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(10, 47, __pyx_L1_error)

  /* "src/numeric_types.pxi":42
 *         return FIT_NONE
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (py_bool, bool_)):
 *             return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.BoolType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8BoolType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8BoolType_5_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8BoolType_5_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8BoolType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 42, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(10, 42, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(10, 42, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.BoolType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8BoolType_4_convert_value(((struct __pyx_obj_5cavro_BoolType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8BoolType_4_convert_value(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8BoolType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BoolType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":49
 *         raise InvalidValue(value, self)
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */

static PyObject *__pyx_f_5cavro_8BoolType__json_format(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 1);

  /* "src/numeric_types.pxi":50
 * 
 *     cdef _json_format(self, value):
 *         return self._convert_value(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8BoolType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":49
 *         raise InvalidValue(value, self)
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BoolType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":52
 *         return self._convert_value(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         cdef py_bool decoded = value
 *         return decoded
 */

static PyObject *__pyx_f_5cavro_8BoolType__json_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyBoolObject *__pyx_v_decoded = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/numeric_types.pxi":53
 * 
 *     cdef _json_decode(self, value):
 *         cdef py_bool decoded = value             # <<<<<<<<<<<<<<
 *         return decoded
 * 
 */
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool))))) __PYX_ERR(10, 53, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_decoded = ((PyBoolObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/numeric_types.pxi":54
 *     cdef _json_decode(self, value):
 *         cdef py_bool decoded = value
 *         return decoded             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_decoded);
  __pyx_r = ((PyObject *)__pyx_v_decoded);
  goto __pyx_L0;

  /* "src/numeric_types.pxi":52
 *         return self._convert_value(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         cdef py_bool decoded = value
 *         return decoded
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BoolType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_decoded);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":56
 *         return decoded
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"boolean"')
 * 
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_8BoolType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_BoolType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/numeric_types.pxi":57
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"boolean"')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_tuple__51, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":56
 *         return decoded
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"boolean"')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BoolType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8BoolType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8BoolType_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8BoolType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8BoolType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_8BoolType_6__reduce_cython__(((struct __pyx_obj_5cavro_BoolType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8BoolType_6__reduce_cython__(struct __pyx_obj_5cavro_BoolType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_BoolType, (type(self), 0x84c5042, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BoolType, (type(self), 0x84c5042, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_BoolType, (type(self), 0x84c5042, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_BoolType, (type(self), 0x84c5042, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_BoolType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BoolType, (type(self), 0x84c5042, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_BoolType, (type(self), 0x84c5042, None), state
 *     else:
 *         return __pyx_unpickle_BoolType, (type(self), 0x84c5042, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BoolType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_BoolType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.BoolType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BoolType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BoolType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8BoolType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8BoolType_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8BoolType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8BoolType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.BoolType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8BoolType_8__setstate_cython__(((struct __pyx_obj_5cavro_BoolType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8BoolType_8__setstate_cython__(struct __pyx_obj_5cavro_BoolType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_BoolType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BoolType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_BoolType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BoolType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BoolType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BoolType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":65
 *     type_name = "int"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

static PyObject *__pyx_pw_5cavro_7IntType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_7IntType_copy(struct __pyx_obj_5cavro_IntType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "src/numeric_types.pxi":66
 * 
 *     cpdef AvroType copy(self):
 *         return self.clone_base()             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_IntType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":65
 *     type_name = "int"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.IntType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7IntType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7IntType_1copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7IntType_1copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7IntType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_7IntType_copy(((struct __pyx_obj_5cavro_IntType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7IntType_copy(struct __pyx_obj_5cavro_IntType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_7IntType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.IntType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":68
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

static PyObject *__pyx_f_5cavro_7IntType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/numeric_types.pxi":69
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {'type'})             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(10, 69, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":68
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.IntType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":71
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_pw_5cavro_7IntType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_7IntType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_IntType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);

  /* "src/numeric_types.pxi":72
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":71
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.IntType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7IntType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7IntType_3_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7IntType_3_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7IntType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 71, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(10, 71, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(10, 71, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.IntType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(10, 71, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_7IntType_2_get_schema_extra(((struct __pyx_obj_5cavro_IntType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7IntType_2_get_schema_extra(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_7IntType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.IntType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":74
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         value = self._convert_value(value)
 *         zigzag_encode_int(buffer, value)
 */

static int __pyx_f_5cavro_7IntType__binary_buffer_encode(struct __pyx_obj_5cavro_IntType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int32_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/numeric_types.pxi":75
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         value = self._convert_value(value)             # <<<<<<<<<<<<<<
 *         zigzag_encode_int(buffer, value)
 * 
 */
  __pyx_t_1 = __pyx_f_5cavro_7IntType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/numeric_types.pxi":76
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         value = self._convert_value(value)
 *         zigzag_encode_int(buffer, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_t_2 = __Pyx_PyInt_As_int32_t(__pyx_v_value); if (unlikely((__pyx_t_2 == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(10, 76, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro_zigzag_encode_int(__pyx_v_buffer, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(10, 76, __pyx_L1_error)

  /* "src/numeric_types.pxi":74
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         value = self._convert_value(value)
 *         zigzag_encode_int(buffer, value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.IntType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":78
 *         zigzag_encode_int(buffer, value)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return zigzag_decode_int(buffer)
 * 
 */

static PyObject *__pyx_f_5cavro_7IntType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_IntType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/numeric_types.pxi":79
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         return zigzag_decode_int(buffer)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_zigzag_decode_int(__pyx_v_buffer); if (unlikely(__pyx_t_1 == ((int32_t)0x7ffffbadU) && PyErr_Occurred())) __PYX_ERR(10, 79, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":78
 *         zigzag_encode_int(buffer, value)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return zigzag_decode_int(buffer)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.IntType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":81
 *         return zigzag_decode_int(buffer)
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         max_fit = FIT_EXACT
 *         if isinstance(value, (bool_, py_bool)):
 */

static int __pyx_f_5cavro_7IntType__get_value_fitness(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_v_max_fit;
  PyObject *__pyx_v_new_value = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/numeric_types.pxi":82
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         max_fit = FIT_EXACT             # <<<<<<<<<<<<<<
 *         if isinstance(value, (bool_, py_bool)):
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 */
  __pyx_v_max_fit = __pyx_v_5cavro_FIT_EXACT;

  /* "src/numeric_types.pxi":83
 *     cdef int _get_value_fitness(self, value) except -1:
 *         max_fit = FIT_EXACT
 *         if isinstance(value, (bool_, py_bool)):             # <<<<<<<<<<<<<<
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 *         if not isinstance(value, (int, integer)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/numeric_types.pxi":84
 *         max_fit = FIT_EXACT
 *         if isinstance(value, (bool_, py_bool)):
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (int, integer)):
 *             if not self.options.coerce_values_to_int:
 */
    if (__pyx_v_self->__pyx_base.options->coerce_values_to_int) {
      __pyx_t_4 = __pyx_v_5cavro_FIT_POOR;
    } else {
      __pyx_t_4 = __pyx_v_5cavro_FIT_NONE;
    }
    __pyx_r = __pyx_t_4;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":83
 *     cdef int _get_value_fitness(self, value) except -1:
 *         max_fit = FIT_EXACT
 *         if isinstance(value, (bool_, py_bool)):             # <<<<<<<<<<<<<<
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 *         if not isinstance(value, (int, integer)):
 */
  }

  /* "src/numeric_types.pxi":85
 *         if isinstance(value, (bool_, py_bool)):
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 *         if not isinstance(value, (int, integer)):             # <<<<<<<<<<<<<<
 *             if not self.options.coerce_values_to_int:
 *                 return FIT_NONE
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_integer); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_Check(__pyx_v_value); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "src/numeric_types.pxi":86
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 *         if not isinstance(value, (int, integer)):
 *             if not self.options.coerce_values_to_int:             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *             try:
 */
    __pyx_t_3 = (!__pyx_v_self->__pyx_base.options->coerce_values_to_int);
    if (__pyx_t_3) {

      /* "src/numeric_types.pxi":87
 *         if not isinstance(value, (int, integer)):
 *             if not self.options.coerce_values_to_int:
 *                 return FIT_NONE             # <<<<<<<<<<<<<<
 *             try:
 *                 new_value = int(value)
 */
      __pyx_r = __pyx_v_5cavro_FIT_NONE;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":86
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 *         if not isinstance(value, (int, integer)):
 *             if not self.options.coerce_values_to_int:             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *             try:
 */
    }

    /* "src/numeric_types.pxi":88
 *             if not self.options.coerce_values_to_int:
 *                 return FIT_NONE
 *             try:             # <<<<<<<<<<<<<<
 *                 new_value = int(value)
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "src/numeric_types.pxi":89
 *                 return FIT_NONE
 *             try:
 *                 new_value = int(value)             # <<<<<<<<<<<<<<
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR
 *             except (ValueError, TypeError):
 */
        __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 89, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_v_new_value = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "src/numeric_types.pxi":90
 *             try:
 *                 new_value = int(value)
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR             # <<<<<<<<<<<<<<
 *             except (ValueError, TypeError):
 *                 return FIT_NONE
 */
        __pyx_t_1 = PyObject_RichCompare(__pyx_v_new_value, __pyx_v_value, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 90, __pyx_L10_error)
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(10, 90, __pyx_L10_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_3) {
          __pyx_t_4 = __pyx_v_5cavro_FIT_OK;
        } else {
          __pyx_t_4 = __pyx_v_5cavro_FIT_POOR;
        }
        __pyx_v_max_fit = __pyx_t_4;

        /* "src/numeric_types.pxi":88
 *             if not self.options.coerce_values_to_int:
 *                 return FIT_NONE
 *             try:             # <<<<<<<<<<<<<<
 *                 new_value = int(value)
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR
 */
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L15_try_end;
      __pyx_L10_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "src/numeric_types.pxi":91
 *                 new_value = int(value)
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR
 *             except (ValueError, TypeError):             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *             value = new_value
 */
      __pyx_t_4 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_ValueError, __pyx_builtin_TypeError);
      if (__pyx_t_4) {
        __Pyx_AddTraceback("cavro.IntType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(10, 91, __pyx_L12_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);

        /* "src/numeric_types.pxi":92
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR
 *             except (ValueError, TypeError):
 *                 return FIT_NONE             # <<<<<<<<<<<<<<
 *             value = new_value
 * 
 */
        __pyx_r = __pyx_v_5cavro_FIT_NONE;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L13_except_return;
      }
      goto __pyx_L12_except_error;

      /* "src/numeric_types.pxi":88
 *             if not self.options.coerce_values_to_int:
 *                 return FIT_NONE
 *             try:             # <<<<<<<<<<<<<<
 *                 new_value = int(value)
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR
 */
      __pyx_L12_except_error:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L13_except_return:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L0;
      __pyx_L15_try_end:;
    }

    /* "src/numeric_types.pxi":93
 *             except (ValueError, TypeError):
 *                 return FIT_NONE
 *             value = new_value             # <<<<<<<<<<<<<<
 * 
 *         if value > INT32_MAX or value < INT32_MIN:
 */
    __Pyx_INCREF(__pyx_v_new_value);
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_v_new_value);

    /* "src/numeric_types.pxi":85
 *         if isinstance(value, (bool_, py_bool)):
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 *         if not isinstance(value, (int, integer)):             # <<<<<<<<<<<<<<
 *             if not self.options.coerce_values_to_int:
 *                 return FIT_NONE
 */
  }

  /* "src/numeric_types.pxi":95
 *             value = new_value
 * 
 *         if value > INT32_MAX or value < INT32_MIN:             # <<<<<<<<<<<<<<
 *             return FIT_POOR if self.options.clamp_int_overflow else FIT_NONE
 * 
 */
  __pyx_t_9 = __Pyx_PyInt_From_int32_t(INT32_MAX); if (unlikely(!__pyx_t_9)) __PYX_ERR(10, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = PyObject_RichCompare(__pyx_v_value, __pyx_t_9, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(10, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(10, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (!__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L19_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_PyInt_From_int32_t(INT32_MIN); if (unlikely(!__pyx_t_8)) __PYX_ERR(10, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyObject_RichCompare(__pyx_v_value, __pyx_t_8, Py_LT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(10, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(10, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_3 = __pyx_t_2;
  __pyx_L19_bool_binop_done:;
  if (__pyx_t_3) {

    /* "src/numeric_types.pxi":96
 * 
 *         if value > INT32_MAX or value < INT32_MIN:
 *             return FIT_POOR if self.options.clamp_int_overflow else FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *         return max_fit
 */
    if (__pyx_v_self->__pyx_base.options->clamp_int_overflow) {
      __pyx_t_4 = __pyx_v_5cavro_FIT_POOR;
    } else {
      __pyx_t_4 = __pyx_v_5cavro_FIT_NONE;
    }
    __pyx_r = __pyx_t_4;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":95
 *             value = new_value
 * 
 *         if value > INT32_MAX or value < INT32_MIN:             # <<<<<<<<<<<<<<
 *             return FIT_POOR if self.options.clamp_int_overflow else FIT_NONE
 * 
 */
  }

  /* "src/numeric_types.pxi":98
 *             return FIT_POOR if self.options.clamp_int_overflow else FIT_NONE
 * 
 *         return max_fit             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
  __pyx_r = __pyx_v_max_fit;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":81
 *         return zigzag_decode_int(buffer)
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         max_fit = FIT_EXACT
 *         if isinstance(value, (bool_, py_bool)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cavro.IntType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_new_value);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":100
 *         return max_fit
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')
 */

static PyObject *__pyx_pw_5cavro_7IntType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_7IntType__convert_value(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/numeric_types.pxi":101
 * 
 *     cpdef object _convert_value(self, object value):
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):             # <<<<<<<<<<<<<<
 *             raise ValueError(f'{value} not valid for long')
 *         if not isinstance(value, int):
 */
  __pyx_t_2 = (!__pyx_v_self->__pyx_base.options->coerce_values_to_int);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
  __pyx_t_2 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "src/numeric_types.pxi":102
 *     cpdef object _convert_value(self, object value):
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')             # <<<<<<<<<<<<<<
 *         if not isinstance(value, int):
 *             if self.options.coerce_values_to_int or isinstance(value, integer):
 */
    __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_3, __pyx_kp_u_not_valid_for_long); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(10, 102, __pyx_L1_error)

    /* "src/numeric_types.pxi":101
 * 
 *     cpdef object _convert_value(self, object value):
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):             # <<<<<<<<<<<<<<
 *             raise ValueError(f'{value} not valid for long')
 *         if not isinstance(value, int):
 */
  }

  /* "src/numeric_types.pxi":103
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')
 *         if not isinstance(value, int):             # <<<<<<<<<<<<<<
 *             if self.options.coerce_values_to_int or isinstance(value, integer):
 *                 value = int(value)
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_value); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "src/numeric_types.pxi":104
 *             raise ValueError(f'{value} not valid for long')
 *         if not isinstance(value, int):
 *             if self.options.coerce_values_to_int or isinstance(value, integer):             # <<<<<<<<<<<<<<
 *                 value = int(value)
 *             else:
 */
    if (!__pyx_v_self->__pyx_base.options->coerce_values_to_int) {
    } else {
      __pyx_t_2 = __pyx_v_self->__pyx_base.options->coerce_values_to_int;
      goto __pyx_L10_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_integer); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(10, 104, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __pyx_t_1;
    __pyx_L10_bool_binop_done:;
    if (likely(__pyx_t_2)) {

      /* "src/numeric_types.pxi":105
 *         if not isinstance(value, int):
 *             if self.options.coerce_values_to_int or isinstance(value, integer):
 *                 value = int(value)             # <<<<<<<<<<<<<<
 *             else:
 *                 raise InvalidValue(value, self)
 */
      __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 105, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "src/numeric_types.pxi":104
 *             raise ValueError(f'{value} not valid for long')
 *         if not isinstance(value, int):
 *             if self.options.coerce_values_to_int or isinstance(value, integer):             # <<<<<<<<<<<<<<
 *                 value = int(value)
 *             else:
 */
      goto __pyx_L9;
    }

    /* "src/numeric_types.pxi":107
 *                 value = int(value)
 *             else:
 *                 raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         if value > INT32_MAX:
 *             if self.options.clamp_int_overflow:
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_value, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 107, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(10, 107, __pyx_L1_error)
    }
    __pyx_L9:;

    /* "src/numeric_types.pxi":103
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')
 *         if not isinstance(value, int):             # <<<<<<<<<<<<<<
 *             if self.options.coerce_values_to_int or isinstance(value, integer):
 *                 value = int(value)
 */
  }

  /* "src/numeric_types.pxi":108
 *             else:
 *                 raise InvalidValue(value, self)
 *         if value > INT32_MAX:             # <<<<<<<<<<<<<<
 *             if self.options.clamp_int_overflow:
 *                 return INT32_MAX
 */
  __pyx_t_3 = __Pyx_PyInt_From_int32_t(INT32_MAX); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_value, __pyx_t_3, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(10, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "src/numeric_types.pxi":109
 *                 raise InvalidValue(value, self)
 *         if value > INT32_MAX:
 *             if self.options.clamp_int_overflow:             # <<<<<<<<<<<<<<
 *                 return INT32_MAX
 *         elif value < INT32_MIN:
 */
    if (__pyx_v_self->__pyx_base.options->clamp_int_overflow) {

      /* "src/numeric_types.pxi":110
 *         if value > INT32_MAX:
 *             if self.options.clamp_int_overflow:
 *                 return INT32_MAX             # <<<<<<<<<<<<<<
 *         elif value < INT32_MIN:
 *             if self.options.clamp_int_overflow:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = __Pyx_PyInt_From_int32_t(INT32_MAX); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":109
 *                 raise InvalidValue(value, self)
 *         if value > INT32_MAX:
 *             if self.options.clamp_int_overflow:             # <<<<<<<<<<<<<<
 *                 return INT32_MAX
 *         elif value < INT32_MIN:
 */
    }

    /* "src/numeric_types.pxi":108
 *             else:
 *                 raise InvalidValue(value, self)
 *         if value > INT32_MAX:             # <<<<<<<<<<<<<<
 *             if self.options.clamp_int_overflow:
 *                 return INT32_MAX
 */
    goto __pyx_L12;
  }

  /* "src/numeric_types.pxi":111
 *             if self.options.clamp_int_overflow:
 *                 return INT32_MAX
 *         elif value < INT32_MIN:             # <<<<<<<<<<<<<<
 *             if self.options.clamp_int_overflow:
 *                 return INT32_MIN
 */
  __pyx_t_5 = __Pyx_PyInt_From_int32_t(INT32_MIN); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_value, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(10, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "src/numeric_types.pxi":112
 *                 return INT32_MAX
 *         elif value < INT32_MIN:
 *             if self.options.clamp_int_overflow:             # <<<<<<<<<<<<<<
 *                 return INT32_MIN
 *         else:
 */
    if (__pyx_v_self->__pyx_base.options->clamp_int_overflow) {

      /* "src/numeric_types.pxi":113
 *         elif value < INT32_MIN:
 *             if self.options.clamp_int_overflow:
 *                 return INT32_MIN             # <<<<<<<<<<<<<<
 *         else:
 *             return value
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int32_t(INT32_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":112
 *                 return INT32_MAX
 *         elif value < INT32_MIN:
 *             if self.options.clamp_int_overflow:             # <<<<<<<<<<<<<<
 *                 return INT32_MIN
 *         else:
 */
    }

    /* "src/numeric_types.pxi":111
 *             if self.options.clamp_int_overflow:
 *                 return INT32_MAX
 *         elif value < INT32_MIN:             # <<<<<<<<<<<<<<
 *             if self.options.clamp_int_overflow:
 *                 return INT32_MIN
 */
    goto __pyx_L12;
  }

  /* "src/numeric_types.pxi":115
 *                 return INT32_MIN
 *         else:
 *             return value             # <<<<<<<<<<<<<<
 *         raise InvalidValue(value, self)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;
  }
  __pyx_L12:;

  /* "src/numeric_types.pxi":116
 *         else:
 *             return value
 *         raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_value, ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_Raise(__pyx_t_3, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __PYX_ERR(10, 116, __pyx_L1_error)

  /* "src/numeric_types.pxi":100
 *         return max_fit
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.IntType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7IntType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7IntType_5_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7IntType_5_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7IntType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 100, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(10, 100, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(10, 100, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.IntType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_7IntType_4_convert_value(((struct __pyx_obj_5cavro_IntType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7IntType_4_convert_value(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_7IntType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.IntType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":118
 *         raise InvalidValue(value, self)
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */

static PyObject *__pyx_f_5cavro_7IntType__json_format(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 1);

  /* "src/numeric_types.pxi":119
 * 
 *     cdef _json_format(self, value):
 *         return self._convert_value(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_7IntType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":118
 *         raise InvalidValue(value, self)
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.IntType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":121
 *         return self._convert_value(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):
 *             raise InvalidValue(value, self)
 */

static PyObject *__pyx_f_5cavro_7IntType__json_decode(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/numeric_types.pxi":122
 * 
 *     cdef _json_decode(self, value):
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if value < INT32_MIN or value > INT32_MAX:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyFloat_Check(__pyx_v_value); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); 
  __pyx_t_3 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyInt_Check(__pyx_v_value); 
  __pyx_t_4 = (!__pyx_t_3);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "src/numeric_types.pxi":123
 *     cdef _json_decode(self, value):
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         if value < INT32_MIN or value > INT32_MAX:
 *             raise InvalidValue(value, self)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(10, 123, __pyx_L1_error)

    /* "src/numeric_types.pxi":122
 * 
 *     cdef _json_decode(self, value):
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if value < INT32_MIN or value > INT32_MAX:
 */
  }

  /* "src/numeric_types.pxi":124
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):
 *             raise InvalidValue(value, self)
 *         if value < INT32_MIN or value > INT32_MAX:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         return value
 */
  __pyx_t_2 = __Pyx_PyInt_From_int32_t(INT32_MIN); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_value, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(10, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyInt_From_int32_t(INT32_MAX); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_value, __pyx_t_5, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(10, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L10_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "src/numeric_types.pxi":125
 *             raise InvalidValue(value, self)
 *         if value < INT32_MIN or value > INT32_MAX:
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 125, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(10, 125, __pyx_L1_error)

    /* "src/numeric_types.pxi":124
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):
 *             raise InvalidValue(value, self)
 *         if value < INT32_MIN or value > INT32_MAX:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         return value
 */
  }

  /* "src/numeric_types.pxi":126
 *         if value < INT32_MIN or value > INT32_MAX:
 *             raise InvalidValue(value, self)
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":121
 *         return self._convert_value(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):
 *             raise InvalidValue(value, self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.IntType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":128
 *         return value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"int"')
 * 
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_7IntType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_IntType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/numeric_types.pxi":129
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"int"')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_tuple__52, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":128
 *         return value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"int"')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.IntType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7IntType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7IntType_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7IntType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7IntType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_7IntType_6__reduce_cython__(((struct __pyx_obj_5cavro_IntType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7IntType_6__reduce_cython__(struct __pyx_obj_5cavro_IntType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IntType, (type(self), 0x84c5042, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IntType, (type(self), 0x84c5042, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_IntType, (type(self), 0x84c5042, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IntType, (type(self), 0x84c5042, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_IntType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IntType, (type(self), 0x84c5042, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IntType, (type(self), 0x84c5042, None), state
 *     else:
 *         return __pyx_unpickle_IntType, (type(self), 0x84c5042, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IntType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_IntType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.IntType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_7IntType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_7IntType_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7IntType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_7IntType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.IntType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_7IntType_8__setstate_cython__(((struct __pyx_obj_5cavro_IntType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_7IntType_8__setstate_cython__(struct __pyx_obj_5cavro_IntType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IntType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IntType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_IntType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.IntType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":137
 *     type_name = "long"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

static PyObject *__pyx_pw_5cavro_8LongType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8LongType_copy(struct __pyx_obj_5cavro_LongType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "src/numeric_types.pxi":138
 * 
 *     cpdef AvroType copy(self):
 *         return self.clone_base()             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_LongType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":137
 *     type_name = "long"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.LongType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8LongType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8LongType_1copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8LongType_1copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8LongType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_8LongType_copy(((struct __pyx_obj_5cavro_LongType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8LongType_copy(struct __pyx_obj_5cavro_LongType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_8LongType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.LongType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":140
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

static PyObject *__pyx_f_5cavro_8LongType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/numeric_types.pxi":141
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {'type'})             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(10, 141, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":140
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.LongType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":143
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_pw_5cavro_8LongType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_8LongType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_LongType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);

  /* "src/numeric_types.pxi":144
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":143
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.LongType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8LongType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8LongType_3_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8LongType_3_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8LongType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 143, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(10, 143, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(10, 143, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.LongType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(10, 143, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_8LongType_2_get_schema_extra(((struct __pyx_obj_5cavro_LongType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8LongType_2_get_schema_extra(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8LongType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.LongType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":146
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         value = self._convert_value(value)
 *         zigzag_encode_long(buffer, value)
 */

static int __pyx_f_5cavro_8LongType__binary_buffer_encode(struct __pyx_obj_5cavro_LongType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/numeric_types.pxi":147
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         value = self._convert_value(value)             # <<<<<<<<<<<<<<
 *         zigzag_encode_long(buffer, value)
 * 
 */
  __pyx_t_1 = __pyx_f_5cavro_8LongType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/numeric_types.pxi":148
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         value = self._convert_value(value)
 *         zigzag_encode_long(buffer, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(__pyx_v_value); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(10, 148, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro_zigzag_encode_long(__pyx_v_buffer, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(10, 148, __pyx_L1_error)

  /* "src/numeric_types.pxi":146
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         value = self._convert_value(value)
 *         zigzag_encode_long(buffer, value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.LongType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":150
 *         zigzag_encode_long(buffer, value)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return zigzag_decode_long(buffer)
 * 
 */

static PyObject *__pyx_f_5cavro_8LongType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_LongType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/numeric_types.pxi":151
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         return zigzag_decode_long(buffer)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_zigzag_decode_long(__pyx_v_buffer); if (unlikely(__pyx_t_1 == ((int64_t)0x7ffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(10, 151, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":150
 *         zigzag_encode_long(buffer, value)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return zigzag_decode_long(buffer)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.LongType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":153
 *         return zigzag_decode_long(buffer)
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         max_fit = FIT_EXACT
 *         if isinstance(value, (bool_, py_bool)):
 */

static int __pyx_f_5cavro_8LongType__get_value_fitness(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_v_max_fit;
  PyObject *__pyx_v_new_value = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/numeric_types.pxi":154
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         max_fit = FIT_EXACT             # <<<<<<<<<<<<<<
 *         if isinstance(value, (bool_, py_bool)):
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 */
  __pyx_v_max_fit = __pyx_v_5cavro_FIT_EXACT;

  /* "src/numeric_types.pxi":155
 *     cdef int _get_value_fitness(self, value) except -1:
 *         max_fit = FIT_EXACT
 *         if isinstance(value, (bool_, py_bool)):             # <<<<<<<<<<<<<<
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 *         if not isinstance(value, (int, integer)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/numeric_types.pxi":156
 *         max_fit = FIT_EXACT
 *         if isinstance(value, (bool_, py_bool)):
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (int, integer)):
 *             if not self.options.coerce_values_to_int:
 */
    if (__pyx_v_self->__pyx_base.options->coerce_values_to_int) {
      __pyx_t_4 = __pyx_v_5cavro_FIT_POOR;
    } else {
      __pyx_t_4 = __pyx_v_5cavro_FIT_NONE;
    }
    __pyx_r = __pyx_t_4;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":155
 *     cdef int _get_value_fitness(self, value) except -1:
 *         max_fit = FIT_EXACT
 *         if isinstance(value, (bool_, py_bool)):             # <<<<<<<<<<<<<<
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 *         if not isinstance(value, (int, integer)):
 */
  }

  /* "src/numeric_types.pxi":157
 *         if isinstance(value, (bool_, py_bool)):
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 *         if not isinstance(value, (int, integer)):             # <<<<<<<<<<<<<<
 *             if not self.options.coerce_values_to_int:
 *                 return FIT_NONE
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_integer); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_Check(__pyx_v_value); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "src/numeric_types.pxi":158
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 *         if not isinstance(value, (int, integer)):
 *             if not self.options.coerce_values_to_int:             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *             try:
 */
    __pyx_t_3 = (!__pyx_v_self->__pyx_base.options->coerce_values_to_int);
    if (__pyx_t_3) {

      /* "src/numeric_types.pxi":159
 *         if not isinstance(value, (int, integer)):
 *             if not self.options.coerce_values_to_int:
 *                 return FIT_NONE             # <<<<<<<<<<<<<<
 *             try:
 *                 new_value = int(value)
 */
      __pyx_r = __pyx_v_5cavro_FIT_NONE;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":158
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 *         if not isinstance(value, (int, integer)):
 *             if not self.options.coerce_values_to_int:             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *             try:
 */
    }

    /* "src/numeric_types.pxi":160
 *             if not self.options.coerce_values_to_int:
 *                 return FIT_NONE
 *             try:             # <<<<<<<<<<<<<<
 *                 new_value = int(value)
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "src/numeric_types.pxi":161
 *                 return FIT_NONE
 *             try:
 *                 new_value = int(value)             # <<<<<<<<<<<<<<
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR
 *             except (ValueError, TypeError):
 */
        __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 161, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_v_new_value = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "src/numeric_types.pxi":162
 *             try:
 *                 new_value = int(value)
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR             # <<<<<<<<<<<<<<
 *             except (ValueError, TypeError):
 *                 return FIT_NONE
 */
        __pyx_t_1 = PyObject_RichCompare(__pyx_v_new_value, __pyx_v_value, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 162, __pyx_L10_error)
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(10, 162, __pyx_L10_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_3) {
          __pyx_t_4 = __pyx_v_5cavro_FIT_OK;
        } else {
          __pyx_t_4 = __pyx_v_5cavro_FIT_POOR;
        }
        __pyx_v_max_fit = __pyx_t_4;

        /* "src/numeric_types.pxi":160
 *             if not self.options.coerce_values_to_int:
 *                 return FIT_NONE
 *             try:             # <<<<<<<<<<<<<<
 *                 new_value = int(value)
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR
 */
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L15_try_end;
      __pyx_L10_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "src/numeric_types.pxi":163
 *                 new_value = int(value)
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR
 *             except (ValueError, TypeError):             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *             value = new_value
 */
      __pyx_t_4 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_ValueError, __pyx_builtin_TypeError);
      if (__pyx_t_4) {
        __Pyx_AddTraceback("cavro.LongType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(10, 163, __pyx_L12_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);

        /* "src/numeric_types.pxi":164
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR
 *             except (ValueError, TypeError):
 *                 return FIT_NONE             # <<<<<<<<<<<<<<
 *             value = new_value
 * 
 */
        __pyx_r = __pyx_v_5cavro_FIT_NONE;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L13_except_return;
      }
      goto __pyx_L12_except_error;

      /* "src/numeric_types.pxi":160
 *             if not self.options.coerce_values_to_int:
 *                 return FIT_NONE
 *             try:             # <<<<<<<<<<<<<<
 *                 new_value = int(value)
 *                 max_fit = FIT_OK if new_value == value else FIT_POOR
 */
      __pyx_L12_except_error:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L13_except_return:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L0;
      __pyx_L15_try_end:;
    }

    /* "src/numeric_types.pxi":165
 *             except (ValueError, TypeError):
 *                 return FIT_NONE
 *             value = new_value             # <<<<<<<<<<<<<<
 * 
 *         if value > INT64_MAX or value < INT64_MIN:
 */
    __Pyx_INCREF(__pyx_v_new_value);
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_v_new_value);

    /* "src/numeric_types.pxi":157
 *         if isinstance(value, (bool_, py_bool)):
 *             return FIT_POOR if self.options.coerce_values_to_int else FIT_NONE
 *         if not isinstance(value, (int, integer)):             # <<<<<<<<<<<<<<
 *             if not self.options.coerce_values_to_int:
 *                 return FIT_NONE
 */
  }

  /* "src/numeric_types.pxi":167
 *             value = new_value
 * 
 *         if value > INT64_MAX or value < INT64_MIN:             # <<<<<<<<<<<<<<
 *             return FIT_POOR if self.options.clamp_int_overflow else FIT_NONE
 * 
 */
  __pyx_t_9 = __Pyx_PyInt_From_int64_t(INT64_MAX); if (unlikely(!__pyx_t_9)) __PYX_ERR(10, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = PyObject_RichCompare(__pyx_v_value, __pyx_t_9, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(10, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(10, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (!__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L19_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_PyInt_From_int64_t(INT64_MIN); if (unlikely(!__pyx_t_8)) __PYX_ERR(10, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyObject_RichCompare(__pyx_v_value, __pyx_t_8, Py_LT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(10, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(10, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_3 = __pyx_t_2;
  __pyx_L19_bool_binop_done:;
  if (__pyx_t_3) {

    /* "src/numeric_types.pxi":168
 * 
 *         if value > INT64_MAX or value < INT64_MIN:
 *             return FIT_POOR if self.options.clamp_int_overflow else FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *         return max_fit
 */
    if (__pyx_v_self->__pyx_base.options->clamp_int_overflow) {
      __pyx_t_4 = __pyx_v_5cavro_FIT_POOR;
    } else {
      __pyx_t_4 = __pyx_v_5cavro_FIT_NONE;
    }
    __pyx_r = __pyx_t_4;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":167
 *             value = new_value
 * 
 *         if value > INT64_MAX or value < INT64_MIN:             # <<<<<<<<<<<<<<
 *             return FIT_POOR if self.options.clamp_int_overflow else FIT_NONE
 * 
 */
  }

  /* "src/numeric_types.pxi":170
 *             return FIT_POOR if self.options.clamp_int_overflow else FIT_NONE
 * 
 *         return max_fit             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
  __pyx_r = __pyx_v_max_fit;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":153
 *         return zigzag_decode_long(buffer)
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         max_fit = FIT_EXACT
 *         if isinstance(value, (bool_, py_bool)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cavro.LongType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_new_value);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":172
 *         return max_fit
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')
 */

static PyObject *__pyx_pw_5cavro_8LongType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_8LongType__convert_value(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/numeric_types.pxi":173
 * 
 *     cpdef object _convert_value(self, object value):
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):             # <<<<<<<<<<<<<<
 *             raise ValueError(f'{value} not valid for long')
 *         if not isinstance(value, int):
 */
  __pyx_t_2 = (!__pyx_v_self->__pyx_base.options->coerce_values_to_int);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
  __pyx_t_2 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "src/numeric_types.pxi":174
 *     cpdef object _convert_value(self, object value):
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')             # <<<<<<<<<<<<<<
 *         if not isinstance(value, int):
 *             if self.options.coerce_values_to_int or isinstance(value, integer):
 */
    __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_3, __pyx_kp_u_not_valid_for_long); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(10, 174, __pyx_L1_error)

    /* "src/numeric_types.pxi":173
 * 
 *     cpdef object _convert_value(self, object value):
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):             # <<<<<<<<<<<<<<
 *             raise ValueError(f'{value} not valid for long')
 *         if not isinstance(value, int):
 */
  }

  /* "src/numeric_types.pxi":175
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')
 *         if not isinstance(value, int):             # <<<<<<<<<<<<<<
 *             if self.options.coerce_values_to_int or isinstance(value, integer):
 *                 value = int(value)
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_value); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "src/numeric_types.pxi":176
 *             raise ValueError(f'{value} not valid for long')
 *         if not isinstance(value, int):
 *             if self.options.coerce_values_to_int or isinstance(value, integer):             # <<<<<<<<<<<<<<
 *                 value = int(value)
 *             else:
 */
    if (!__pyx_v_self->__pyx_base.options->coerce_values_to_int) {
    } else {
      __pyx_t_2 = __pyx_v_self->__pyx_base.options->coerce_values_to_int;
      goto __pyx_L10_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_integer); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(10, 176, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __pyx_t_1;
    __pyx_L10_bool_binop_done:;
    if (likely(__pyx_t_2)) {

      /* "src/numeric_types.pxi":177
 *         if not isinstance(value, int):
 *             if self.options.coerce_values_to_int or isinstance(value, integer):
 *                 value = int(value)             # <<<<<<<<<<<<<<
 *             else:
 *                 raise InvalidValue(value, self)
 */
      __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "src/numeric_types.pxi":176
 *             raise ValueError(f'{value} not valid for long')
 *         if not isinstance(value, int):
 *             if self.options.coerce_values_to_int or isinstance(value, integer):             # <<<<<<<<<<<<<<
 *                 value = int(value)
 *             else:
 */
      goto __pyx_L9;
    }

    /* "src/numeric_types.pxi":179
 *                 value = int(value)
 *             else:
 *                 raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         if value > INT64_MAX:
 *             if self.options.clamp_int_overflow:
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 179, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_value, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 179, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(10, 179, __pyx_L1_error)
    }
    __pyx_L9:;

    /* "src/numeric_types.pxi":175
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')
 *         if not isinstance(value, int):             # <<<<<<<<<<<<<<
 *             if self.options.coerce_values_to_int or isinstance(value, integer):
 *                 value = int(value)
 */
  }

  /* "src/numeric_types.pxi":180
 *             else:
 *                 raise InvalidValue(value, self)
 *         if value > INT64_MAX:             # <<<<<<<<<<<<<<
 *             if self.options.clamp_int_overflow:
 *                 return INT64_MAX
 */
  __pyx_t_3 = __Pyx_PyInt_From_int64_t(INT64_MAX); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_value, __pyx_t_3, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(10, 180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "src/numeric_types.pxi":181
 *                 raise InvalidValue(value, self)
 *         if value > INT64_MAX:
 *             if self.options.clamp_int_overflow:             # <<<<<<<<<<<<<<
 *                 return INT64_MAX
 *         elif value < INT64_MIN:
 */
    if (__pyx_v_self->__pyx_base.options->clamp_int_overflow) {

      /* "src/numeric_types.pxi":182
 *         if value > INT64_MAX:
 *             if self.options.clamp_int_overflow:
 *                 return INT64_MAX             # <<<<<<<<<<<<<<
 *         elif value < INT64_MIN:
 *             if self.options.clamp_int_overflow:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = __Pyx_PyInt_From_int64_t(INT64_MAX); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 182, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":181
 *                 raise InvalidValue(value, self)
 *         if value > INT64_MAX:
 *             if self.options.clamp_int_overflow:             # <<<<<<<<<<<<<<
 *                 return INT64_MAX
 *         elif value < INT64_MIN:
 */
    }

    /* "src/numeric_types.pxi":180
 *             else:
 *                 raise InvalidValue(value, self)
 *         if value > INT64_MAX:             # <<<<<<<<<<<<<<
 *             if self.options.clamp_int_overflow:
 *                 return INT64_MAX
 */
    goto __pyx_L12;
  }

  /* "src/numeric_types.pxi":183
 *             if self.options.clamp_int_overflow:
 *                 return INT64_MAX
 *         elif value < INT64_MIN:             # <<<<<<<<<<<<<<
 *             if self.options.clamp_int_overflow:
 *                 return INT64_MIN
 */
  __pyx_t_5 = __Pyx_PyInt_From_int64_t(INT64_MIN); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_value, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(10, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "src/numeric_types.pxi":184
 *                 return INT64_MAX
 *         elif value < INT64_MIN:
 *             if self.options.clamp_int_overflow:             # <<<<<<<<<<<<<<
 *                 return INT64_MIN
 *         else:
 */
    if (__pyx_v_self->__pyx_base.options->clamp_int_overflow) {

      /* "src/numeric_types.pxi":185
 *         elif value < INT64_MIN:
 *             if self.options.clamp_int_overflow:
 *                 return INT64_MIN             # <<<<<<<<<<<<<<
 *         else:
 *             return value
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int64_t(INT64_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":184
 *                 return INT64_MAX
 *         elif value < INT64_MIN:
 *             if self.options.clamp_int_overflow:             # <<<<<<<<<<<<<<
 *                 return INT64_MIN
 *         else:
 */
    }

    /* "src/numeric_types.pxi":183
 *             if self.options.clamp_int_overflow:
 *                 return INT64_MAX
 *         elif value < INT64_MIN:             # <<<<<<<<<<<<<<
 *             if self.options.clamp_int_overflow:
 *                 return INT64_MIN
 */
    goto __pyx_L12;
  }

  /* "src/numeric_types.pxi":187
 *                 return INT64_MIN
 *         else:
 *             return value             # <<<<<<<<<<<<<<
 *         raise InvalidValue(value, self)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;
  }
  __pyx_L12:;

  /* "src/numeric_types.pxi":188
 *         else:
 *             return value
 *         raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 * 
 *     cdef json_format(self, object value):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_value, ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_Raise(__pyx_t_3, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __PYX_ERR(10, 188, __pyx_L1_error)

  /* "src/numeric_types.pxi":172
 *         return max_fit
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.LongType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8LongType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8LongType_5_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8LongType_5_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8LongType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 172, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(10, 172, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(10, 172, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.LongType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8LongType_4_convert_value(((struct __pyx_obj_5cavro_LongType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8LongType_4_convert_value(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8LongType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.LongType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":190
 *         raise InvalidValue(value, self)
 * 
 *     cdef json_format(self, object value):             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */

static PyObject *__pyx_f_5cavro_8LongType_json_format(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("json_format", 1);

  /* "src/numeric_types.pxi":191
 * 
 *     cdef json_format(self, object value):
 *         return self._convert_value(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8LongType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":190
 *         raise InvalidValue(value, self)
 * 
 *     cdef json_format(self, object value):             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.LongType.json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":193
 *         return self._convert_value(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):
 *             raise InvalidValue(value, self)
 */

static PyObject *__pyx_f_5cavro_8LongType__json_decode(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/numeric_types.pxi":194
 * 
 *     cdef _json_decode(self, value):
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if value < INT64_MIN or value > INT64_MAX:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyFloat_Check(__pyx_v_value); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); 
  __pyx_t_3 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyInt_Check(__pyx_v_value); 
  __pyx_t_4 = (!__pyx_t_3);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "src/numeric_types.pxi":195
 *     cdef _json_decode(self, value):
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         if value < INT64_MIN or value > INT64_MAX:
 *             raise InvalidValue(value, self)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 195, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(10, 195, __pyx_L1_error)

    /* "src/numeric_types.pxi":194
 * 
 *     cdef _json_decode(self, value):
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if value < INT64_MIN or value > INT64_MAX:
 */
  }

  /* "src/numeric_types.pxi":196
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):
 *             raise InvalidValue(value, self)
 *         if value < INT64_MIN or value > INT64_MAX:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         return value
 */
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(INT64_MIN); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_value, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(10, 196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyInt_From_int64_t(INT64_MAX); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_value, __pyx_t_5, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(10, 196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L10_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "src/numeric_types.pxi":197
 *             raise InvalidValue(value, self)
 *         if value < INT64_MIN or value > INT64_MAX:
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 197, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(10, 197, __pyx_L1_error)

    /* "src/numeric_types.pxi":196
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):
 *             raise InvalidValue(value, self)
 *         if value < INT64_MIN or value > INT64_MAX:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         return value
 */
  }

  /* "src/numeric_types.pxi":198
 *         if value < INT64_MIN or value > INT64_MAX:
 *             raise InvalidValue(value, self)
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":193
 *         return self._convert_value(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (float, py_bool, bool_)) or not isinstance(value, int):
 *             raise InvalidValue(value, self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.LongType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":200
 *         return value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"long"')
 * 
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_8LongType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_LongType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/numeric_types.pxi":201
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"long"')             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_tuple__53, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":200
 *         return value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"long"')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.LongType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":203
 *         return _CanonicalForm('"long"')
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         if isinstance(writer, IntType):
 *             return writer # Int is a subset of Long
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8LongType__for_writer(CYTHON_UNUSED struct __pyx_obj_5cavro_LongType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_for_writer", 1);

  /* "src/numeric_types.pxi":204
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if isinstance(writer, IntType):             # <<<<<<<<<<<<<<
 *             return writer # Int is a subset of Long
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_IntType); 
  if (__pyx_t_1) {

    /* "src/numeric_types.pxi":205
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if isinstance(writer, IntType):
 *             return writer # Int is a subset of Long             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_writer);
    __pyx_r = __pyx_v_writer;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":204
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if isinstance(writer, IntType):             # <<<<<<<<<<<<<<
 *             return writer # Int is a subset of Long
 * 
 */
  }

  /* "src/numeric_types.pxi":203
 *         return _CanonicalForm('"long"')
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         if isinstance(writer, IntType):
 *             return writer # Int is a subset of Long
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8LongType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8LongType_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8LongType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8LongType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_8LongType_6__reduce_cython__(((struct __pyx_obj_5cavro_LongType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8LongType_6__reduce_cython__(struct __pyx_obj_5cavro_LongType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_LongType, (type(self), 0x84c5042, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LongType, (type(self), 0x84c5042, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_LongType, (type(self), 0x84c5042, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_LongType, (type(self), 0x84c5042, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_LongType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LongType, (type(self), 0x84c5042, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_LongType, (type(self), 0x84c5042, None), state
 *     else:
 *         return __pyx_unpickle_LongType, (type(self), 0x84c5042, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LongType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_LongType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.LongType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LongType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LongType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8LongType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8LongType_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8LongType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8LongType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.LongType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8LongType_8__setstate_cython__(((struct __pyx_obj_5cavro_LongType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8LongType_8__setstate_cython__(struct __pyx_obj_5cavro_LongType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_LongType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LongType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_LongType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LongType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LongType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.LongType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":213
 *     type_name = "float"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

static PyObject *__pyx_pw_5cavro_9FloatType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9FloatType_copy(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "src/numeric_types.pxi":214
 * 
 *     cpdef AvroType copy(self):
 *         return self.clone_base()             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_FloatType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":213
 *     type_name = "float"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FloatType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9FloatType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9FloatType_1copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FloatType_1copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9FloatType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_9FloatType_copy(((struct __pyx_obj_5cavro_FloatType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9FloatType_copy(struct __pyx_obj_5cavro_FloatType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_9FloatType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FloatType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":216
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

static PyObject *__pyx_f_5cavro_9FloatType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/numeric_types.pxi":217
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {'type'})             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(10, 217, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":216
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.FloatType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":219
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_pw_5cavro_9FloatType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_9FloatType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_FloatType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);

  /* "src/numeric_types.pxi":220
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":219
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FloatType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9FloatType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9FloatType_3_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FloatType_3_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9FloatType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 219, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(10, 219, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(10, 219, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.FloatType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(10, 219, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_9FloatType_2_get_schema_extra(((struct __pyx_obj_5cavro_FloatType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9FloatType_2_get_schema_extra(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_9FloatType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FloatType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":222
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef float float_val = self._convert_value(value)
 *         cdef uint8_t *int_val = <uint8_t*>&float_val
 */

static int __pyx_f_5cavro_9FloatType__binary_buffer_encode(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  float __pyx_v_float_val;
  uint8_t *__pyx_v_int_val;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  float __pyx_t_2;
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 1);

  /* "src/numeric_types.pxi":223
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef float float_val = self._convert_value(value)             # <<<<<<<<<<<<<<
 *         cdef uint8_t *int_val = <uint8_t*>&float_val
 *         buffer.write_n(int_val[:4])
 */
  __pyx_t_1 = __pyx_f_5cavro_9FloatType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_2 == (float)-1) && PyErr_Occurred())) __PYX_ERR(10, 223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_float_val = __pyx_t_2;

  /* "src/numeric_types.pxi":224
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef float float_val = self._convert_value(value)
 *         cdef uint8_t *int_val = <uint8_t*>&float_val             # <<<<<<<<<<<<<<
 *         buffer.write_n(int_val[:4])
 * 
 */
  __pyx_v_int_val = ((uint8_t *)(&__pyx_v_float_val));

  /* "src/numeric_types.pxi":225
 *         cdef float float_val = self._convert_value(value)
 *         cdef uint8_t *int_val = <uint8_t*>&float_val
 *         buffer.write_n(int_val[:4])             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_int_val) + 0, 4 - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_1, 0); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(10, 225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_buffer->__pyx_vtab)->write_n(__pyx_v_buffer, __pyx_t_3); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(10, 225, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_3, 1);
  __pyx_t_3.memview = NULL; __pyx_t_3.data = NULL;

  /* "src/numeric_types.pxi":222
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef float float_val = self._convert_value(value)
 *         cdef uint8_t *int_val = <uint8_t*>&float_val
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_AddTraceback("cavro.FloatType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":227
 *         buffer.write_n(int_val[:4])
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef const uint8_t[:] val = buffer.read_n(4)
 *         return (<float*>(&val[0]))[0]
 */

static PyObject *__pyx_f_5cavro_9FloatType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_FloatType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  __Pyx_memviewslice __pyx_v_val = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/numeric_types.pxi":228
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef const uint8_t[:] val = buffer.read_n(4)             # <<<<<<<<<<<<<<
 *         return (<float*>(&val[0]))[0]
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_buffer->__pyx_vtab)->read_n(__pyx_v_buffer, 4); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(10, 228, __pyx_L1_error)
  __pyx_v_val = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;

  /* "src/numeric_types.pxi":229
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef const uint8_t[:] val = buffer.read_n(4)
 *         return (<float*>(&val[0]))[0]             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = 0;
  __pyx_t_3 = -1;
  if (__pyx_t_2 < 0) {
    __pyx_t_2 += __pyx_v_val.shape[0];
    if (unlikely(__pyx_t_2 < 0)) __pyx_t_3 = 0;
  } else if (unlikely(__pyx_t_2 >= __pyx_v_val.shape[0])) __pyx_t_3 = 0;
  if (unlikely(__pyx_t_3 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_3);
    __PYX_ERR(10, 229, __pyx_L1_error)
  }
  __pyx_t_4 = PyFloat_FromDouble((((float *)(&(*((uint8_t const  *) ( /* dim=0 */ (__pyx_v_val.data + __pyx_t_2 * __pyx_v_val.strides[0]) )))))[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":227
 *         buffer.write_n(int_val[:4])
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef const uint8_t[:] val = buffer.read_n(4)
 *         return (<float*>(&val[0]))[0]
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.FloatType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XCLEAR_MEMVIEW(&__pyx_v_val, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":231
 *         return (<float*>(&val[0]))[0]
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         max_fit = FIT_EXACT
 *         if not isinstance(value, (float, np_f16, np_f32)):
 */

static int __pyx_f_5cavro_9FloatType__get_value_fitness(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_v_max_fit;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/numeric_types.pxi":232
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         max_fit = FIT_EXACT             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (float, np_f16, np_f32)):
 *             max_fit = FIT_OK
 */
  __pyx_v_max_fit = __pyx_v_5cavro_FIT_EXACT;

  /* "src/numeric_types.pxi":233
 *     cdef int _get_value_fitness(self, value) except -1:
 *         max_fit = FIT_EXACT
 *         if not isinstance(value, (float, np_f16, np_f32)):             # <<<<<<<<<<<<<<
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np_f16); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np_f32); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyFloat_Check(__pyx_v_value); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); 
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (!__pyx_t_3);
  if (__pyx_t_4) {

    /* "src/numeric_types.pxi":234
 *         max_fit = FIT_EXACT
 *         if not isinstance(value, (float, np_f16, np_f32)):
 *             max_fit = FIT_OK             # <<<<<<<<<<<<<<
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 *                 try:
 */
    __pyx_v_max_fit = __pyx_v_5cavro_FIT_OK;

    /* "src/numeric_types.pxi":235
 *         if not isinstance(value, (float, np_f16, np_f32)):
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *                 try:
 *                     value = float(value)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_integer); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyInt_Check(__pyx_v_value); 
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_5 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
    __pyx_t_3 = __pyx_t_5;
    __pyx_L10_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {
    } else {
      __pyx_t_4 = __pyx_t_3;
      goto __pyx_L8_bool_binop_done;
    }
    if (__pyx_v_self->__pyx_base.options->coerce_int_to_float) {
    } else {
      __pyx_t_4 = __pyx_v_self->__pyx_base.options->coerce_int_to_float;
      goto __pyx_L8_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_5 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
    __pyx_t_3 = __pyx_t_5;
    __pyx_L13_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = (!__pyx_t_3);
    __pyx_t_4 = __pyx_t_5;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_4) {

      /* "src/numeric_types.pxi":236
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = float(value)
 *                 except OverflowError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {

          /* "src/numeric_types.pxi":237
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 *                 try:
 *                     value = float(value)             # <<<<<<<<<<<<<<
 *                 except OverflowError:
 *                     return FIT_POOR if self.options.clamp_float_overflow else FIT_NONE
 */
          __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 237, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "src/numeric_types.pxi":236
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = float(value)
 *                 except OverflowError:
 */
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L20_try_end;
        __pyx_L15_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "src/numeric_types.pxi":238
 *                 try:
 *                     value = float(value)
 *                 except OverflowError:             # <<<<<<<<<<<<<<
 *                     return FIT_POOR if self.options.clamp_float_overflow else FIT_NONE
 *             else:
 */
        __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OverflowError);
        if (__pyx_t_9) {
          __Pyx_AddTraceback("cavro.FloatType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_10) < 0) __PYX_ERR(10, 238, __pyx_L17_except_error)
          __Pyx_XGOTREF(__pyx_t_1);
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_10);

          /* "src/numeric_types.pxi":239
 *                     value = float(value)
 *                 except OverflowError:
 *                     return FIT_POOR if self.options.clamp_float_overflow else FIT_NONE             # <<<<<<<<<<<<<<
 *             else:
 *                 max_fit = FIT_POOR
 */
          if (__pyx_v_self->__pyx_base.options->clamp_float_overflow) {
            __pyx_t_9 = __pyx_v_5cavro_FIT_POOR;
          } else {
            __pyx_t_9 = __pyx_v_5cavro_FIT_NONE;
          }
          __pyx_r = __pyx_t_9;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L18_except_return;
        }
        goto __pyx_L17_except_error;

        /* "src/numeric_types.pxi":236
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = float(value)
 *                 except OverflowError:
 */
        __pyx_L17_except_error:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L18_except_return:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L0;
        __pyx_L20_try_end:;
      }

      /* "src/numeric_types.pxi":235
 *         if not isinstance(value, (float, np_f16, np_f32)):
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *                 try:
 *                     value = float(value)
 */
      goto __pyx_L7;
    }

    /* "src/numeric_types.pxi":241
 *                     return FIT_POOR if self.options.clamp_float_overflow else FIT_NONE
 *             else:
 *                 max_fit = FIT_POOR             # <<<<<<<<<<<<<<
 *                 if not self.options.coerce_values_to_float:
 *                     return FIT_NONE
 */
    /*else*/ {
      __pyx_v_max_fit = __pyx_v_5cavro_FIT_POOR;

      /* "src/numeric_types.pxi":242
 *             else:
 *                 max_fit = FIT_POOR
 *                 if not self.options.coerce_values_to_float:             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 *                 try:
 */
      __pyx_t_4 = (!__pyx_v_self->__pyx_base.options->coerce_values_to_float);
      if (__pyx_t_4) {

        /* "src/numeric_types.pxi":243
 *                 max_fit = FIT_POOR
 *                 if not self.options.coerce_values_to_float:
 *                     return FIT_NONE             # <<<<<<<<<<<<<<
 *                 try:
 *                     value = float(value)
 */
        __pyx_r = __pyx_v_5cavro_FIT_NONE;
        goto __pyx_L0;

        /* "src/numeric_types.pxi":242
 *             else:
 *                 max_fit = FIT_POOR
 *                 if not self.options.coerce_values_to_float:             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 *                 try:
 */
      }

      /* "src/numeric_types.pxi":244
 *                 if not self.options.coerce_values_to_float:
 *                     return FIT_NONE
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = float(value)
 *                 except (ValueError, TypeError, OverflowError):
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);
        /*try:*/ {

          /* "src/numeric_types.pxi":245
 *                     return FIT_NONE
 *                 try:
 *                     value = float(value)             # <<<<<<<<<<<<<<
 *                 except (ValueError, TypeError, OverflowError):
 *                     return FIT_NONE
 */
          __pyx_t_10 = __Pyx_PyNumber_Float(__pyx_v_value); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 245, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_10);
          __pyx_t_10 = 0;

          /* "src/numeric_types.pxi":244
 *                 if not self.options.coerce_values_to_float:
 *                     return FIT_NONE
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = float(value)
 *                 except (ValueError, TypeError, OverflowError):
 */
        }
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L29_try_end;
        __pyx_L24_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "src/numeric_types.pxi":246
 *                 try:
 *                     value = float(value)
 *                 except (ValueError, TypeError, OverflowError):             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 * 
 */
        __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OverflowError);
        if (__pyx_t_9) {
          __Pyx_AddTraceback("cavro.FloatType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_10, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(10, 246, __pyx_L26_except_error)
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_1);

          /* "src/numeric_types.pxi":247
 *                     value = float(value)
 *                 except (ValueError, TypeError, OverflowError):
 *                     return FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *         if self.options.clamp_float_overflow or self.options.float_out_of_range_inf:
 */
          __pyx_r = __pyx_v_5cavro_FIT_NONE;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L27_except_return;
        }
        goto __pyx_L26_except_error;

        /* "src/numeric_types.pxi":244
 *                 if not self.options.coerce_values_to_float:
 *                     return FIT_NONE
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = float(value)
 *                 except (ValueError, TypeError, OverflowError):
 */
        __pyx_L26_except_error:;
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
        goto __pyx_L1_error;
        __pyx_L27_except_return:;
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
        goto __pyx_L0;
        __pyx_L29_try_end:;
      }
    }
    __pyx_L7:;

    /* "src/numeric_types.pxi":233
 *     cdef int _get_value_fitness(self, value) except -1:
 *         max_fit = FIT_EXACT
 *         if not isinstance(value, (float, np_f16, np_f32)):             # <<<<<<<<<<<<<<
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 */
  }

  /* "src/numeric_types.pxi":249
 *                     return FIT_NONE
 * 
 *         if self.options.clamp_float_overflow or self.options.float_out_of_range_inf:             # <<<<<<<<<<<<<<
 *             return max_fit
 * 
 */
  if (!__pyx_v_self->__pyx_base.options->clamp_float_overflow) {
  } else {
    __pyx_t_4 = __pyx_v_self->__pyx_base.options->clamp_float_overflow;
    goto __pyx_L33_bool_binop_done;
  }
  __pyx_t_4 = __pyx_v_self->__pyx_base.options->float_out_of_range_inf;
  __pyx_L33_bool_binop_done:;
  if (__pyx_t_4) {

    /* "src/numeric_types.pxi":250
 * 
 *         if self.options.clamp_float_overflow or self.options.float_out_of_range_inf:
 *             return max_fit             # <<<<<<<<<<<<<<
 * 
 *         if (value >= -FLT_MAX and value <= FLT_MAX) or isnan(value) or isinf(value):
 */
    __pyx_r = __pyx_v_max_fit;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":249
 *                     return FIT_NONE
 * 
 *         if self.options.clamp_float_overflow or self.options.float_out_of_range_inf:             # <<<<<<<<<<<<<<
 *             return max_fit
 * 
 */
  }

  /* "src/numeric_types.pxi":252
 *             return max_fit
 * 
 *         if (value >= -FLT_MAX and value <= FLT_MAX) or isnan(value) or isinf(value):             # <<<<<<<<<<<<<<
 *             return max_fit
 *         return FIT_NONE
 */
  __pyx_t_1 = PyFloat_FromDouble((-FLT_MAX)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_value, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(10, 252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_5) {
    goto __pyx_L37_next_or;
  } else {
  }
  __pyx_t_2 = PyFloat_FromDouble(FLT_MAX); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_value, __pyx_t_2, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(10, 252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L36_bool_binop_done;
  }
  __pyx_L37_next_or:;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isnan); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(10, 252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L36_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_isinf); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(10, 252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_t_5;
  __pyx_L36_bool_binop_done:;
  if (__pyx_t_4) {

    /* "src/numeric_types.pxi":253
 * 
 *         if (value >= -FLT_MAX and value <= FLT_MAX) or isnan(value) or isinf(value):
 *             return max_fit             # <<<<<<<<<<<<<<
 *         return FIT_NONE
 * 
 */
    __pyx_r = __pyx_v_max_fit;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":252
 *             return max_fit
 * 
 *         if (value >= -FLT_MAX and value <= FLT_MAX) or isnan(value) or isinf(value):             # <<<<<<<<<<<<<<
 *             return max_fit
 *         return FIT_NONE
 */
  }

  /* "src/numeric_types.pxi":254
 *         if (value >= -FLT_MAX and value <= FLT_MAX) or isnan(value) or isinf(value):
 *             return max_fit
 *         return FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
  __pyx_r = __pyx_v_5cavro_FIT_NONE;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":231
 *         return (<float*>(&val[0]))[0]
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         max_fit = FIT_EXACT
 *         if not isinstance(value, (float, np_f16, np_f32)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("cavro.FloatType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":256
 *         return FIT_NONE
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, float):
 *             pass
 */

static PyObject *__pyx_pw_5cavro_9FloatType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_9FloatType__convert_value(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  double __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/numeric_types.pxi":257
 * 
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, float):             # <<<<<<<<<<<<<<
 *             pass
 *         elif isinstance(value, (np_f16, np_f32)):
 */
  __pyx_t_1 = PyFloat_Check(__pyx_v_value); 
  if (__pyx_t_1) {
    goto __pyx_L3;
  }

  /* "src/numeric_types.pxi":259
 *         if isinstance(value, float):
 *             pass
 *         elif isinstance(value, (np_f16, np_f32)):             # <<<<<<<<<<<<<<
 *             value = float(value)
 *         else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np_f16); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np_f32); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); 
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); 
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "src/numeric_types.pxi":260
 *             pass
 *         elif isinstance(value, (np_f16, np_f32)):
 *             value = float(value)             # <<<<<<<<<<<<<<
 *         else:
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 */
    __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/numeric_types.pxi":259
 *         if isinstance(value, float):
 *             pass
 *         elif isinstance(value, (np_f16, np_f32)):             # <<<<<<<<<<<<<<
 *             value = float(value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "src/numeric_types.pxi":262
 *             value = float(value)
 *         else:
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *                 value = float(value)
 *             else:
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_integer); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyInt_Check(__pyx_v_value); 
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_5 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); 
    __pyx_t_4 = __pyx_t_5;
    __pyx_L9_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_4) {
    } else {
      __pyx_t_1 = __pyx_t_4;
      goto __pyx_L7_bool_binop_done;
    }
    if (__pyx_v_self->__pyx_base.options->coerce_int_to_float) {
    } else {
      __pyx_t_1 = __pyx_v_self->__pyx_base.options->coerce_int_to_float;
      goto __pyx_L7_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_5 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); 
    __pyx_t_4 = __pyx_t_5;
    __pyx_L12_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (!__pyx_t_4);
    __pyx_t_1 = __pyx_t_5;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_1) {

      /* "src/numeric_types.pxi":263
 *         else:
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 *                 value = float(value)             # <<<<<<<<<<<<<<
 *             else:
 *                 if self.options.coerce_values_to_float:
 */
      __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 263, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "src/numeric_types.pxi":262
 *             value = float(value)
 *         else:
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *                 value = float(value)
 *             else:
 */
      goto __pyx_L6;
    }

    /* "src/numeric_types.pxi":265
 *                 value = float(value)
 *             else:
 *                 if self.options.coerce_values_to_float:             # <<<<<<<<<<<<<<
 *                     try:
 *                         value = float(value)
 */
    /*else*/ {
      if (likely(__pyx_v_self->__pyx_base.options->coerce_values_to_float)) {

        /* "src/numeric_types.pxi":266
 *             else:
 *                 if self.options.coerce_values_to_float:
 *                     try:             # <<<<<<<<<<<<<<
 *                         value = float(value)
 *                     except OverflowError:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_8);
          /*try:*/ {

            /* "src/numeric_types.pxi":267
 *                 if self.options.coerce_values_to_float:
 *                     try:
 *                         value = float(value)             # <<<<<<<<<<<<<<
 *                     except OverflowError:
 *                         if self.options.clamp_float_overflow:
 */
            __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 267, __pyx_L15_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
            __pyx_t_2 = 0;

            /* "src/numeric_types.pxi":266
 *             else:
 *                 if self.options.coerce_values_to_float:
 *                     try:             # <<<<<<<<<<<<<<
 *                         value = float(value)
 *                     except OverflowError:
 */
          }
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          goto __pyx_L20_try_end;
          __pyx_L15_error:;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "src/numeric_types.pxi":268
 *                     try:
 *                         value = float(value)
 *                     except OverflowError:             # <<<<<<<<<<<<<<
 *                         if self.options.clamp_float_overflow:
 *                             return FLT_MAX
 */
          __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OverflowError);
          if (__pyx_t_9) {
            __Pyx_AddTraceback("cavro.FloatType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_10) < 0) __PYX_ERR(10, 268, __pyx_L17_except_error)
            __Pyx_XGOTREF(__pyx_t_2);
            __Pyx_XGOTREF(__pyx_t_3);
            __Pyx_XGOTREF(__pyx_t_10);

            /* "src/numeric_types.pxi":269
 *                         value = float(value)
 *                     except OverflowError:
 *                         if self.options.clamp_float_overflow:             # <<<<<<<<<<<<<<
 *                             return FLT_MAX
 *                         else:
 */
            if (likely(__pyx_v_self->__pyx_base.options->clamp_float_overflow)) {

              /* "src/numeric_types.pxi":270
 *                     except OverflowError:
 *                         if self.options.clamp_float_overflow:
 *                             return FLT_MAX             # <<<<<<<<<<<<<<
 *                         else:
 *                             raise
 */
              __Pyx_XDECREF(__pyx_r);
              __pyx_t_11 = PyFloat_FromDouble(FLT_MAX); if (unlikely(!__pyx_t_11)) __PYX_ERR(10, 270, __pyx_L17_except_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_r = __pyx_t_11;
              __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              goto __pyx_L18_except_return;

              /* "src/numeric_types.pxi":269
 *                         value = float(value)
 *                     except OverflowError:
 *                         if self.options.clamp_float_overflow:             # <<<<<<<<<<<<<<
 *                             return FLT_MAX
 *                         else:
 */
            }

            /* "src/numeric_types.pxi":272
 *                             return FLT_MAX
 *                         else:
 *                             raise             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise InvalidValue(value, self)
 */
            /*else*/ {
              __Pyx_GIVEREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3);
              __Pyx_XGIVEREF(__pyx_t_10);
              __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_3, __pyx_t_10);
              __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_10 = 0; 
              __PYX_ERR(10, 272, __pyx_L17_except_error)
            }
          }
          goto __pyx_L17_except_error;

          /* "src/numeric_types.pxi":266
 *             else:
 *                 if self.options.coerce_values_to_float:
 *                     try:             # <<<<<<<<<<<<<<
 *                         value = float(value)
 *                     except OverflowError:
 */
          __pyx_L17_except_error:;
          __Pyx_XGIVEREF(__pyx_t_6);
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
          goto __pyx_L1_error;
          __pyx_L18_except_return:;
          __Pyx_XGIVEREF(__pyx_t_6);
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
          goto __pyx_L0;
          __pyx_L20_try_end:;
        }

        /* "src/numeric_types.pxi":265
 *                 value = float(value)
 *             else:
 *                 if self.options.coerce_values_to_float:             # <<<<<<<<<<<<<<
 *                     try:
 *                         value = float(value)
 */
        goto __pyx_L14;
      }

      /* "src/numeric_types.pxi":274
 *                             raise
 *                 else:
 *                     raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 * 
 *         if isnan(value):
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 274, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_9 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_9 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_value, ((PyObject *)__pyx_v_self)};
          __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 2+__pyx_t_9);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 274, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_Raise(__pyx_t_10, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __PYX_ERR(10, 274, __pyx_L1_error)
      }
      __pyx_L14:;
    }
    __pyx_L6:;
  }
  __pyx_L3:;

  /* "src/numeric_types.pxi":276
 *                     raise InvalidValue(value, self)
 * 
 *         if isnan(value):             # <<<<<<<<<<<<<<
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:
 *                 return 0.0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isnan); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_value};
    __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(10, 276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (__pyx_t_1) {

    /* "src/numeric_types.pxi":277
 * 
 *         if isnan(value):
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:             # <<<<<<<<<<<<<<
 *                 return 0.0
 *             return value
 */
    if (__pyx_v_self->__pyx_base.options->clamp_float_overflow) {
    } else {
      __pyx_t_1 = __pyx_v_self->__pyx_base.options->clamp_float_overflow;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_5 = (!__pyx_v_self->__pyx_base.options->float_out_of_range_inf);
    __pyx_t_1 = __pyx_t_5;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_1) {

      /* "src/numeric_types.pxi":278
 *         if isnan(value):
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:
 *                 return 0.0             # <<<<<<<<<<<<<<
 *             return value
 *         if isinf(value):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_float_0_0);
      __pyx_r = __pyx_float_0_0;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":277
 * 
 *         if isnan(value):
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:             # <<<<<<<<<<<<<<
 *                 return 0.0
 *             return value
 */
    }

    /* "src/numeric_types.pxi":279
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:
 *                 return 0.0
 *             return value             # <<<<<<<<<<<<<<
 *         if isinf(value):
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":276
 *                     raise InvalidValue(value, self)
 * 
 *         if isnan(value):             # <<<<<<<<<<<<<<
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:
 *                 return 0.0
 */
  }

  /* "src/numeric_types.pxi":280
 *                 return 0.0
 *             return value
 *         if isinf(value):             # <<<<<<<<<<<<<<
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:
 *                 return FLT_MAX if value > 0 else -FLT_MAX
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isinf); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_value};
    __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(10, 280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (__pyx_t_1) {

    /* "src/numeric_types.pxi":281
 *             return value
 *         if isinf(value):
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:             # <<<<<<<<<<<<<<
 *                 return FLT_MAX if value > 0 else -FLT_MAX
 *             return value
 */
    if (__pyx_v_self->__pyx_base.options->clamp_float_overflow) {
    } else {
      __pyx_t_1 = __pyx_v_self->__pyx_base.options->clamp_float_overflow;
      goto __pyx_L30_bool_binop_done;
    }
    __pyx_t_5 = (!__pyx_v_self->__pyx_base.options->float_out_of_range_inf);
    __pyx_t_1 = __pyx_t_5;
    __pyx_L30_bool_binop_done:;
    if (__pyx_t_1) {

      /* "src/numeric_types.pxi":282
 *         if isinf(value):
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:
 *                 return FLT_MAX if value > 0 else -FLT_MAX             # <<<<<<<<<<<<<<
 *             return value
 *         if value < -FLT_MAX:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_value, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 282, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(10, 282, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_1) {
        __pyx_t_3 = PyFloat_FromDouble(FLT_MAX); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 282, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_10 = __pyx_t_3;
        __pyx_t_3 = 0;
      } else {
        __pyx_t_3 = PyFloat_FromDouble((-FLT_MAX)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 282, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_10 = __pyx_t_3;
        __pyx_t_3 = 0;
      }
      __pyx_r = __pyx_t_10;
      __pyx_t_10 = 0;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":281
 *             return value
 *         if isinf(value):
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:             # <<<<<<<<<<<<<<
 *                 return FLT_MAX if value > 0 else -FLT_MAX
 *             return value
 */
    }

    /* "src/numeric_types.pxi":283
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:
 *                 return FLT_MAX if value > 0 else -FLT_MAX
 *             return value             # <<<<<<<<<<<<<<
 *         if value < -FLT_MAX:
 *             if self.options.clamp_float_overflow:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":280
 *                 return 0.0
 *             return value
 *         if isinf(value):             # <<<<<<<<<<<<<<
 *             if self.options.clamp_float_overflow and not self.options.float_out_of_range_inf:
 *                 return FLT_MAX if value > 0 else -FLT_MAX
 */
  }

  /* "src/numeric_types.pxi":284
 *                 return FLT_MAX if value > 0 else -FLT_MAX
 *             return value
 *         if value < -FLT_MAX:             # <<<<<<<<<<<<<<
 *             if self.options.clamp_float_overflow:
 *                 return -FLT_MAX
 */
  __pyx_t_10 = PyFloat_FromDouble((-FLT_MAX)); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_value, __pyx_t_10, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(10, 284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "src/numeric_types.pxi":285
 *             return value
 *         if value < -FLT_MAX:
 *             if self.options.clamp_float_overflow:             # <<<<<<<<<<<<<<
 *                 return -FLT_MAX
 *             if self.options.float_out_of_range_inf:
 */
    if (__pyx_v_self->__pyx_base.options->clamp_float_overflow) {

      /* "src/numeric_types.pxi":286
 *         if value < -FLT_MAX:
 *             if self.options.clamp_float_overflow:
 *                 return -FLT_MAX             # <<<<<<<<<<<<<<
 *             if self.options.float_out_of_range_inf:
 *                 return float('-inf')
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = PyFloat_FromDouble((-FLT_MAX)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":285
 *             return value
 *         if value < -FLT_MAX:
 *             if self.options.clamp_float_overflow:             # <<<<<<<<<<<<<<
 *                 return -FLT_MAX
 *             if self.options.float_out_of_range_inf:
 */
    }

    /* "src/numeric_types.pxi":287
 *             if self.options.clamp_float_overflow:
 *                 return -FLT_MAX
 *             if self.options.float_out_of_range_inf:             # <<<<<<<<<<<<<<
 *                 return float('-inf')
 *             raise InvalidValue(value, self)
 */
    if (__pyx_v_self->__pyx_base.options->float_out_of_range_inf) {

      /* "src/numeric_types.pxi":288
 *                 return -FLT_MAX
 *             if self.options.float_out_of_range_inf:
 *                 return float('-inf')             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         elif value > FLT_MAX:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_12 = __Pyx_PyUnicode_AsDouble(__pyx_kp_u_inf); if (unlikely(__pyx_t_12 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(10, 288, __pyx_L1_error)
      __pyx_t_3 = PyFloat_FromDouble(__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":287
 *             if self.options.clamp_float_overflow:
 *                 return -FLT_MAX
 *             if self.options.float_out_of_range_inf:             # <<<<<<<<<<<<<<
 *                 return float('-inf')
 *             raise InvalidValue(value, self)
 */
    }

    /* "src/numeric_types.pxi":289
 *             if self.options.float_out_of_range_inf:
 *                 return float('-inf')
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         elif value > FLT_MAX:
 *             if self.options.clamp_float_overflow:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_2 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_9, 2+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(10, 289, __pyx_L1_error)

    /* "src/numeric_types.pxi":284
 *                 return FLT_MAX if value > 0 else -FLT_MAX
 *             return value
 *         if value < -FLT_MAX:             # <<<<<<<<<<<<<<
 *             if self.options.clamp_float_overflow:
 *                 return -FLT_MAX
 */
  }

  /* "src/numeric_types.pxi":290
 *                 return float('-inf')
 *             raise InvalidValue(value, self)
 *         elif value > FLT_MAX:             # <<<<<<<<<<<<<<
 *             if self.options.clamp_float_overflow:
 *                 return FLT_MAX
 */
  __pyx_t_3 = PyFloat_FromDouble(FLT_MAX); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_value, __pyx_t_3, Py_GT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 290, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(10, 290, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (__pyx_t_1) {

    /* "src/numeric_types.pxi":291
 *             raise InvalidValue(value, self)
 *         elif value > FLT_MAX:
 *             if self.options.clamp_float_overflow:             # <<<<<<<<<<<<<<
 *                 return FLT_MAX
 *             if self.options.float_out_of_range_inf:
 */
    if (__pyx_v_self->__pyx_base.options->clamp_float_overflow) {

      /* "src/numeric_types.pxi":292
 *         elif value > FLT_MAX:
 *             if self.options.clamp_float_overflow:
 *                 return FLT_MAX             # <<<<<<<<<<<<<<
 *             if self.options.float_out_of_range_inf:
 *                 return float('inf')
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_10 = PyFloat_FromDouble(FLT_MAX); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_r = __pyx_t_10;
      __pyx_t_10 = 0;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":291
 *             raise InvalidValue(value, self)
 *         elif value > FLT_MAX:
 *             if self.options.clamp_float_overflow:             # <<<<<<<<<<<<<<
 *                 return FLT_MAX
 *             if self.options.float_out_of_range_inf:
 */
    }

    /* "src/numeric_types.pxi":293
 *             if self.options.clamp_float_overflow:
 *                 return FLT_MAX
 *             if self.options.float_out_of_range_inf:             # <<<<<<<<<<<<<<
 *                 return float('inf')
 *             raise InvalidValue(value, self)
 */
    if (__pyx_v_self->__pyx_base.options->float_out_of_range_inf) {

      /* "src/numeric_types.pxi":294
 *                 return FLT_MAX
 *             if self.options.float_out_of_range_inf:
 *                 return float('inf')             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         else:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_12 = __Pyx_PyUnicode_AsDouble(__pyx_n_u_inf_2); if (unlikely(__pyx_t_12 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(10, 294, __pyx_L1_error)
      __pyx_t_10 = PyFloat_FromDouble(__pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 294, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_r = __pyx_t_10;
      __pyx_t_10 = 0;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":293
 *             if self.options.clamp_float_overflow:
 *                 return FLT_MAX
 *             if self.options.float_out_of_range_inf:             # <<<<<<<<<<<<<<
 *                 return float('inf')
 *             raise InvalidValue(value, self)
 */
    }

    /* "src/numeric_types.pxi":295
 *             if self.options.float_out_of_range_inf:
 *                 return float('inf')
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         else:
 *             return value
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 2+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 295, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_ERR(10, 295, __pyx_L1_error)

    /* "src/numeric_types.pxi":290
 *                 return float('-inf')
 *             raise InvalidValue(value, self)
 *         elif value > FLT_MAX:             # <<<<<<<<<<<<<<
 *             if self.options.clamp_float_overflow:
 *                 return FLT_MAX
 */
  }

  /* "src/numeric_types.pxi":297
 *             raise InvalidValue(value, self)
 *         else:
 *             return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;
  }

  /* "src/numeric_types.pxi":256
 *         return FIT_NONE
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, float):
 *             pass
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("cavro.FloatType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9FloatType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9FloatType_5_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FloatType_5_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9FloatType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 256, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(10, 256, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(10, 256, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.FloatType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9FloatType_4_convert_value(((struct __pyx_obj_5cavro_FloatType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9FloatType_4_convert_value(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_9FloatType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FloatType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":299
 *             return value
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */

static PyObject *__pyx_f_5cavro_9FloatType__json_format(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 1);

  /* "src/numeric_types.pxi":300
 * 
 *     cdef _json_format(self, value):
 *         return self._convert_value(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_9FloatType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":299
 *             return value
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FloatType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":302
 *         return self._convert_value(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (py_bool, bool_)):
 *             raise InvalidValue(value, self)
 */

static PyObject *__pyx_f_5cavro_9FloatType__json_decode(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/numeric_types.pxi":303
 * 
 *     cdef _json_decode(self, value):
 *         if isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if value in ('nan', 'inf', '-inf'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "src/numeric_types.pxi":304
 *     cdef _json_decode(self, value):
 *         if isinstance(value, (py_bool, bool_)):
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         if value in ('nan', 'inf', '-inf'):
 *             return float(value)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(10, 304, __pyx_L1_error)

    /* "src/numeric_types.pxi":303
 * 
 *     cdef _json_decode(self, value):
 *         if isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if value in ('nan', 'inf', '-inf'):
 */
  }

  /* "src/numeric_types.pxi":305
 *         if isinstance(value, (py_bool, bool_)):
 *             raise InvalidValue(value, self)
 *         if value in ('nan', 'inf', '-inf'):             # <<<<<<<<<<<<<<
 *             return float(value)
 *         if not isinstance(value, (float, int)):
 */
  __Pyx_INCREF(__pyx_v_value);
  __pyx_t_1 = __pyx_v_value;
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_nan, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(10, 305, __pyx_L1_error)
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_inf_2, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(10, 305, __pyx_L1_error)
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_kp_u_inf, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(10, 305, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_t_2;
  if (__pyx_t_3) {

    /* "src/numeric_types.pxi":306
 *             raise InvalidValue(value, self)
 *         if value in ('nan', 'inf', '-inf'):
 *             return float(value)             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (float, int)):
 *             raise InvalidValue(value, self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":305
 *         if isinstance(value, (py_bool, bool_)):
 *             raise InvalidValue(value, self)
 *         if value in ('nan', 'inf', '-inf'):             # <<<<<<<<<<<<<<
 *             return float(value)
 *         if not isinstance(value, (float, int)):
 */
  }

  /* "src/numeric_types.pxi":307
 *         if value in ('nan', 'inf', '-inf'):
 *             return float(value)
 *         if not isinstance(value, (float, int)):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if (value < -FLT_MAX or value > FLT_MAX) and not isnan(value) and not isinf(value):
 */
  __pyx_t_2 = PyFloat_Check(__pyx_v_value); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_2 = PyInt_Check(__pyx_v_value); 
  __pyx_t_3 = __pyx_t_2;
  __pyx_L11_bool_binop_done:;
  __pyx_t_2 = (!__pyx_t_3);
  if (unlikely(__pyx_t_2)) {

    /* "src/numeric_types.pxi":308
 *             return float(value)
 *         if not isinstance(value, (float, int)):
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         if (value < -FLT_MAX or value > FLT_MAX) and not isnan(value) and not isinf(value):
 *             raise InvalidValue(value, self)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(10, 308, __pyx_L1_error)

    /* "src/numeric_types.pxi":307
 *         if value in ('nan', 'inf', '-inf'):
 *             return float(value)
 *         if not isinstance(value, (float, int)):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if (value < -FLT_MAX or value > FLT_MAX) and not isnan(value) and not isinf(value):
 */
  }

  /* "src/numeric_types.pxi":309
 *         if not isinstance(value, (float, int)):
 *             raise InvalidValue(value, self)
 *         if (value < -FLT_MAX or value > FLT_MAX) and not isnan(value) and not isinf(value):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         return value
 */
  __pyx_t_1 = PyFloat_FromDouble((-FLT_MAX)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_value, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(10, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!__pyx_t_3) {
  } else {
    goto __pyx_L15_next_and;
  }
  __pyx_t_4 = PyFloat_FromDouble(FLT_MAX); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_value, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(10, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_L15_next_and:;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_isnan); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(10, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (!__pyx_t_3);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L14_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_isinf); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(10, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_7);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L14_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "src/numeric_types.pxi":310
 *             raise InvalidValue(value, self)
 *         if (value < -FLT_MAX or value > FLT_MAX) and not isnan(value) and not isinf(value):
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(10, 310, __pyx_L1_error)

    /* "src/numeric_types.pxi":309
 *         if not isinstance(value, (float, int)):
 *             raise InvalidValue(value, self)
 *         if (value < -FLT_MAX or value > FLT_MAX) and not isnan(value) and not isinf(value):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         return value
 */
  }

  /* "src/numeric_types.pxi":311
 *         if (value < -FLT_MAX or value > FLT_MAX) and not isnan(value) and not isinf(value):
 *             raise InvalidValue(value, self)
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":302
 *         return self._convert_value(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (py_bool, bool_)):
 *             raise InvalidValue(value, self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.FloatType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":313
 *         return value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"float"')
 * 
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_9FloatType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_FloatType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/numeric_types.pxi":314
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"float"')             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_tuple__54, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":313
 *         return value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"float"')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FloatType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":316
 *         return _CanonicalForm('"float"')
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         if isinstance(writer, (IntType, LongType)):
 *             promoted = writer.clone_base()
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9FloatType__for_writer(CYTHON_UNUSED struct __pyx_obj_5cavro_FloatType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_promoted = NULL;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_writer", 1);

  /* "src/numeric_types.pxi":317
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if isinstance(writer, (IntType, LongType)):             # <<<<<<<<<<<<<<
 *             promoted = writer.clone_base()
 *             promoted.value_adapters = promoted.value_adapters + (PromoteToFloat(),)
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_IntType); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_LongType); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/numeric_types.pxi":318
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if isinstance(writer, (IntType, LongType)):
 *             promoted = writer.clone_base()             # <<<<<<<<<<<<<<
 *             promoted.value_adapters = promoted.value_adapters + (PromoteToFloat(),)
 *             return promoted
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_writer->__pyx_vtab)->clone_base(__pyx_v_writer, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_promoted = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/numeric_types.pxi":319
 *         if isinstance(writer, (IntType, LongType)):
 *             promoted = writer.clone_base()
 *             promoted.value_adapters = promoted.value_adapters + (PromoteToFloat(),)             # <<<<<<<<<<<<<<
 *             return promoted
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro_PromoteToFloat)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(10, 319, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_v_promoted->value_adapters, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_promoted->value_adapters);
    __Pyx_DECREF(__pyx_v_promoted->value_adapters);
    __pyx_v_promoted->value_adapters = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/numeric_types.pxi":320
 *             promoted = writer.clone_base()
 *             promoted.value_adapters = promoted.value_adapters + (PromoteToFloat(),)
 *             return promoted             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_promoted);
    __pyx_r = __pyx_v_promoted;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":317
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if isinstance(writer, (IntType, LongType)):             # <<<<<<<<<<<<<<
 *             promoted = writer.clone_base()
 *             promoted.value_adapters = promoted.value_adapters + (PromoteToFloat(),)
 */
  }

  /* "src/numeric_types.pxi":316
 *         return _CanonicalForm('"float"')
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         if isinstance(writer, (IntType, LongType)):
 *             promoted = writer.clone_base()
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.FloatType._for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_promoted);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9FloatType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9FloatType_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FloatType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9FloatType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_9FloatType_6__reduce_cython__(((struct __pyx_obj_5cavro_FloatType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9FloatType_6__reduce_cython__(struct __pyx_obj_5cavro_FloatType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_FloatType, (type(self), 0x84c5042, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FloatType, (type(self), 0x84c5042, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_FloatType, (type(self), 0x84c5042, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_FloatType, (type(self), 0x84c5042, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_FloatType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FloatType, (type(self), 0x84c5042, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_FloatType, (type(self), 0x84c5042, None), state
 *     else:
 *         return __pyx_unpickle_FloatType, (type(self), 0x84c5042, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FloatType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_FloatType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.FloatType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FloatType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FloatType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9FloatType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9FloatType_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FloatType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9FloatType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.FloatType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9FloatType_8__setstate_cython__(((struct __pyx_obj_5cavro_FloatType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9FloatType_8__setstate_cython__(struct __pyx_obj_5cavro_FloatType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_FloatType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FloatType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_FloatType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FloatType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FloatType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FloatType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":328
 *     type_name = "double"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

static PyObject *__pyx_pw_5cavro_10DoubleType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10DoubleType_copy(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "src/numeric_types.pxi":329
 * 
 *     cpdef AvroType copy(self):
 *         return self.clone_base()             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_DoubleType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":328
 *     type_name = "double"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.DoubleType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10DoubleType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10DoubleType_1copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10DoubleType_1copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10DoubleType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10DoubleType_copy(((struct __pyx_obj_5cavro_DoubleType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10DoubleType_copy(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_10DoubleType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.DoubleType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":331
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

static PyObject *__pyx_f_5cavro_10DoubleType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/numeric_types.pxi":332
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {'type'})             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(10, 332, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":331
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.DoubleType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":334
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_pw_5cavro_10DoubleType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_10DoubleType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);

  /* "src/numeric_types.pxi":335
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":334
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.DoubleType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10DoubleType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10DoubleType_3_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10DoubleType_3_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10DoubleType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 334, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(10, 334, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(10, 334, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.DoubleType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(10, 334, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_10DoubleType_2_get_schema_extra(((struct __pyx_obj_5cavro_DoubleType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10DoubleType_2_get_schema_extra(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_10DoubleType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.DoubleType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":337
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef double float_val = self._convert_value(value)
 *         cdef uint8_t *int_val = <uint8_t*>&float_val
 */

static int __pyx_f_5cavro_10DoubleType__binary_buffer_encode(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  double __pyx_v_float_val;
  uint8_t *__pyx_v_int_val;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 1);

  /* "src/numeric_types.pxi":338
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef double float_val = self._convert_value(value)             # <<<<<<<<<<<<<<
 *         cdef uint8_t *int_val = <uint8_t*>&float_val
 *         buffer.write_n(int_val[:8])
 */
  __pyx_t_1 = __pyx_f_5cavro_10DoubleType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(10, 338, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_float_val = __pyx_t_2;

  /* "src/numeric_types.pxi":339
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef double float_val = self._convert_value(value)
 *         cdef uint8_t *int_val = <uint8_t*>&float_val             # <<<<<<<<<<<<<<
 *         buffer.write_n(int_val[:8])
 * 
 */
  __pyx_v_int_val = ((uint8_t *)(&__pyx_v_float_val));

  /* "src/numeric_types.pxi":340
 *         cdef double float_val = self._convert_value(value)
 *         cdef uint8_t *int_val = <uint8_t*>&float_val
 *         buffer.write_n(int_val[:8])             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_int_val) + 0, 8 - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_1, 0); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(10, 340, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_buffer->__pyx_vtab)->write_n(__pyx_v_buffer, __pyx_t_3); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(10, 340, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_3, 1);
  __pyx_t_3.memview = NULL; __pyx_t_3.data = NULL;

  /* "src/numeric_types.pxi":337
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef double float_val = self._convert_value(value)
 *         cdef uint8_t *int_val = <uint8_t*>&float_val
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_AddTraceback("cavro.DoubleType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":342
 *         buffer.write_n(int_val[:8])
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef const uint8_t[:] val = buffer.read_n(8)
 *         return (<double*>(&val[0]))[0]
 */

static PyObject *__pyx_f_5cavro_10DoubleType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  __Pyx_memviewslice __pyx_v_val = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/numeric_types.pxi":343
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef const uint8_t[:] val = buffer.read_n(8)             # <<<<<<<<<<<<<<
 *         return (<double*>(&val[0]))[0]
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_buffer->__pyx_vtab)->read_n(__pyx_v_buffer, 8); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(10, 343, __pyx_L1_error)
  __pyx_v_val = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;

  /* "src/numeric_types.pxi":344
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef const uint8_t[:] val = buffer.read_n(8)
 *         return (<double*>(&val[0]))[0]             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = 0;
  __pyx_t_3 = -1;
  if (__pyx_t_2 < 0) {
    __pyx_t_2 += __pyx_v_val.shape[0];
    if (unlikely(__pyx_t_2 < 0)) __pyx_t_3 = 0;
  } else if (unlikely(__pyx_t_2 >= __pyx_v_val.shape[0])) __pyx_t_3 = 0;
  if (unlikely(__pyx_t_3 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_3);
    __PYX_ERR(10, 344, __pyx_L1_error)
  }
  __pyx_t_4 = PyFloat_FromDouble((((double *)(&(*((uint8_t const  *) ( /* dim=0 */ (__pyx_v_val.data + __pyx_t_2 * __pyx_v_val.strides[0]) )))))[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":342
 *         buffer.write_n(int_val[:8])
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef const uint8_t[:] val = buffer.read_n(8)
 *         return (<double*>(&val[0]))[0]
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.DoubleType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XCLEAR_MEMVIEW(&__pyx_v_val, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":346
 *         return (<double*>(&val[0]))[0]
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         max_fit = FIT_EXACT
 *         if not isinstance(value, (float, np_f16, np_f32, np_f64)):
 */

static int __pyx_f_5cavro_10DoubleType__get_value_fitness(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_v_max_fit;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/numeric_types.pxi":347
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         max_fit = FIT_EXACT             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (float, np_f16, np_f32, np_f64)):
 *             max_fit = FIT_OK
 */
  __pyx_v_max_fit = __pyx_v_5cavro_FIT_EXACT;

  /* "src/numeric_types.pxi":348
 *     cdef int _get_value_fitness(self, value) except -1:
 *         max_fit = FIT_EXACT
 *         if not isinstance(value, (float, np_f16, np_f32, np_f64)):             # <<<<<<<<<<<<<<
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np_f16); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np_f32); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np_f64); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyFloat_Check(__pyx_v_value); 
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); 
  if (!__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); 
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (!__pyx_t_4);
  if (__pyx_t_5) {

    /* "src/numeric_types.pxi":349
 *         max_fit = FIT_EXACT
 *         if not isinstance(value, (float, np_f16, np_f32, np_f64)):
 *             max_fit = FIT_OK             # <<<<<<<<<<<<<<
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 *                 try:
 */
    __pyx_v_max_fit = __pyx_v_5cavro_FIT_OK;

    /* "src/numeric_types.pxi":350
 *         if not isinstance(value, (float, np_f16, np_f32, np_f64)):
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *                 try:
 *                     value = float(value)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_integer); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyInt_Check(__pyx_v_value); 
    if (!__pyx_t_6) {
    } else {
      __pyx_t_4 = __pyx_t_6;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_6 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
    __pyx_t_4 = __pyx_t_6;
    __pyx_L11_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {
    } else {
      __pyx_t_5 = __pyx_t_4;
      goto __pyx_L9_bool_binop_done;
    }
    if (__pyx_v_self->__pyx_base.options->coerce_int_to_float) {
    } else {
      __pyx_t_5 = __pyx_v_self->__pyx_base.options->coerce_int_to_float;
      goto __pyx_L9_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
    if (!__pyx_t_6) {
    } else {
      __pyx_t_4 = __pyx_t_6;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_6 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
    __pyx_t_4 = __pyx_t_6;
    __pyx_L14_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (!__pyx_t_4);
    __pyx_t_5 = __pyx_t_6;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_5) {

      /* "src/numeric_types.pxi":351
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = float(value)
 *                 except OverflowError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "src/numeric_types.pxi":352
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 *                 try:
 *                     value = float(value)             # <<<<<<<<<<<<<<
 *                 except OverflowError:
 *                     return FIT_POOR if self.options.clamp_float_overflow else FIT_NONE
 */
          __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 352, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "src/numeric_types.pxi":351
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = float(value)
 *                 except OverflowError:
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L21_try_end;
        __pyx_L16_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "src/numeric_types.pxi":353
 *                 try:
 *                     value = float(value)
 *                 except OverflowError:             # <<<<<<<<<<<<<<
 *                     return FIT_POOR if self.options.clamp_float_overflow else FIT_NONE
 *             else:
 */
        __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OverflowError);
        if (__pyx_t_10) {
          __Pyx_AddTraceback("cavro.DoubleType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(10, 353, __pyx_L18_except_error)
          __Pyx_XGOTREF(__pyx_t_1);
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_3);

          /* "src/numeric_types.pxi":354
 *                     value = float(value)
 *                 except OverflowError:
 *                     return FIT_POOR if self.options.clamp_float_overflow else FIT_NONE             # <<<<<<<<<<<<<<
 *             else:
 *                 max_fit = FIT_POOR
 */
          if (__pyx_v_self->__pyx_base.options->clamp_float_overflow) {
            __pyx_t_10 = __pyx_v_5cavro_FIT_POOR;
          } else {
            __pyx_t_10 = __pyx_v_5cavro_FIT_NONE;
          }
          __pyx_r = __pyx_t_10;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L19_except_return;
        }
        goto __pyx_L18_except_error;

        /* "src/numeric_types.pxi":351
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = float(value)
 *                 except OverflowError:
 */
        __pyx_L18_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L19_except_return:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L0;
        __pyx_L21_try_end:;
      }

      /* "src/numeric_types.pxi":350
 *         if not isinstance(value, (float, np_f16, np_f32, np_f64)):
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *                 try:
 *                     value = float(value)
 */
      goto __pyx_L8;
    }

    /* "src/numeric_types.pxi":356
 *                     return FIT_POOR if self.options.clamp_float_overflow else FIT_NONE
 *             else:
 *                 max_fit = FIT_POOR             # <<<<<<<<<<<<<<
 *                 if not self.options.coerce_values_to_float:
 *                     return FIT_NONE
 */
    /*else*/ {
      __pyx_v_max_fit = __pyx_v_5cavro_FIT_POOR;

      /* "src/numeric_types.pxi":357
 *             else:
 *                 max_fit = FIT_POOR
 *                 if not self.options.coerce_values_to_float:             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 *                 try:
 */
      __pyx_t_5 = (!__pyx_v_self->__pyx_base.options->coerce_values_to_float);
      if (__pyx_t_5) {

        /* "src/numeric_types.pxi":358
 *                 max_fit = FIT_POOR
 *                 if not self.options.coerce_values_to_float:
 *                     return FIT_NONE             # <<<<<<<<<<<<<<
 *                 try:
 *                     value = float(value)
 */
        __pyx_r = __pyx_v_5cavro_FIT_NONE;
        goto __pyx_L0;

        /* "src/numeric_types.pxi":357
 *             else:
 *                 max_fit = FIT_POOR
 *                 if not self.options.coerce_values_to_float:             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 *                 try:
 */
      }

      /* "src/numeric_types.pxi":359
 *                 if not self.options.coerce_values_to_float:
 *                     return FIT_NONE
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = float(value)
 *                 except (ValueError, TypeError, OverflowError):
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);
        /*try:*/ {

          /* "src/numeric_types.pxi":360
 *                     return FIT_NONE
 *                 try:
 *                     value = float(value)             # <<<<<<<<<<<<<<
 *                 except (ValueError, TypeError, OverflowError):
 *                     return FIT_NONE
 */
          __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 360, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "src/numeric_types.pxi":359
 *                 if not self.options.coerce_values_to_float:
 *                     return FIT_NONE
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = float(value)
 *                 except (ValueError, TypeError, OverflowError):
 */
        }
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L30_try_end;
        __pyx_L25_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "src/numeric_types.pxi":361
 *                 try:
 *                     value = float(value)
 *                 except (ValueError, TypeError, OverflowError):             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 * 
 */
        __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OverflowError);
        if (__pyx_t_10) {
          __Pyx_AddTraceback("cavro.DoubleType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(10, 361, __pyx_L27_except_error)
          __Pyx_XGOTREF(__pyx_t_3);
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_1);

          /* "src/numeric_types.pxi":362
 *                     value = float(value)
 *                 except (ValueError, TypeError, OverflowError):
 *                     return FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *         return max_fit
 */
          __pyx_r = __pyx_v_5cavro_FIT_NONE;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L28_except_return;
        }
        goto __pyx_L27_except_error;

        /* "src/numeric_types.pxi":359
 *                 if not self.options.coerce_values_to_float:
 *                     return FIT_NONE
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = float(value)
 *                 except (ValueError, TypeError, OverflowError):
 */
        __pyx_L27_except_error:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
        goto __pyx_L1_error;
        __pyx_L28_except_return:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
        goto __pyx_L0;
        __pyx_L30_try_end:;
      }
    }
    __pyx_L8:;

    /* "src/numeric_types.pxi":348
 *     cdef int _get_value_fitness(self, value) except -1:
 *         max_fit = FIT_EXACT
 *         if not isinstance(value, (float, np_f16, np_f32, np_f64)):             # <<<<<<<<<<<<<<
 *             max_fit = FIT_OK
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 */
  }

  /* "src/numeric_types.pxi":364
 *                     return FIT_NONE
 * 
 *         return max_fit             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
  __pyx_r = __pyx_v_max_fit;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":346
 *         return (<double*>(&val[0]))[0]
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         max_fit = FIT_EXACT
 *         if not isinstance(value, (float, np_f16, np_f32, np_f64)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.DoubleType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":366
 *         return max_fit
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, float):
 *             pass
 */

static PyObject *__pyx_pw_5cavro_10DoubleType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_10DoubleType__convert_value(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/numeric_types.pxi":367
 * 
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, float):             # <<<<<<<<<<<<<<
 *             pass
 *         elif isinstance(value, (np_f16, np_f32, np_f64)):
 */
  __pyx_t_1 = PyFloat_Check(__pyx_v_value); 
  if (__pyx_t_1) {
    goto __pyx_L3;
  }

  /* "src/numeric_types.pxi":369
 *         if isinstance(value, float):
 *             pass
 *         elif isinstance(value, (np_f16, np_f32, np_f64)):             # <<<<<<<<<<<<<<
 *             value = float(value)
 *         else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np_f16); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np_f32); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np_f64); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); 
  if (!__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); 
  if (!__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_value, __pyx_t_4); 
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "src/numeric_types.pxi":370
 *             pass
 *         elif isinstance(value, (np_f16, np_f32, np_f64)):
 *             value = float(value)             # <<<<<<<<<<<<<<
 *         else:
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 */
    __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/numeric_types.pxi":369
 *         if isinstance(value, float):
 *             pass
 *         elif isinstance(value, (np_f16, np_f32, np_f64)):             # <<<<<<<<<<<<<<
 *             value = float(value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "src/numeric_types.pxi":372
 *             value = float(value)
 *         else:
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *                 value = float(value)
 *             else:
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_integer); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyInt_Check(__pyx_v_value); 
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_6 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); 
    __pyx_t_5 = __pyx_t_6;
    __pyx_L10_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_1 = __pyx_t_5;
      goto __pyx_L8_bool_binop_done;
    }
    if (__pyx_v_self->__pyx_base.options->coerce_int_to_float) {
    } else {
      __pyx_t_1 = __pyx_v_self->__pyx_base.options->coerce_int_to_float;
      goto __pyx_L8_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_6 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); 
    __pyx_t_5 = __pyx_t_6;
    __pyx_L13_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = (!__pyx_t_5);
    __pyx_t_1 = __pyx_t_6;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_1) {

      /* "src/numeric_types.pxi":373
 *         else:
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):
 *                 value = float(value)             # <<<<<<<<<<<<<<
 *             else:
 *                 if self.options.coerce_values_to_float:
 */
      __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 373, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "src/numeric_types.pxi":372
 *             value = float(value)
 *         else:
 *             if isinstance(value, (int, integer)) and self.options.coerce_int_to_float and not isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *                 value = float(value)
 *             else:
 */
      goto __pyx_L7;
    }

    /* "src/numeric_types.pxi":375
 *                 value = float(value)
 *             else:
 *                 if self.options.coerce_values_to_float:             # <<<<<<<<<<<<<<
 *                     try:
 *                         value = float(value)
 */
    /*else*/ {
      if (likely(__pyx_v_self->__pyx_base.options->coerce_values_to_float)) {

        /* "src/numeric_types.pxi":376
 *             else:
 *                 if self.options.coerce_values_to_float:
 *                     try:             # <<<<<<<<<<<<<<
 *                         value = float(value)
 *                     except OverflowError:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_9);
          /*try:*/ {

            /* "src/numeric_types.pxi":377
 *                 if self.options.coerce_values_to_float:
 *                     try:
 *                         value = float(value)             # <<<<<<<<<<<<<<
 *                     except OverflowError:
 *                         if self.options.clamp_float_overflow:
 */
            __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 377, __pyx_L16_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
            __pyx_t_2 = 0;

            /* "src/numeric_types.pxi":376
 *             else:
 *                 if self.options.coerce_values_to_float:
 *                     try:             # <<<<<<<<<<<<<<
 *                         value = float(value)
 *                     except OverflowError:
 */
          }
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L21_try_end;
          __pyx_L16_error:;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "src/numeric_types.pxi":378
 *                     try:
 *                         value = float(value)
 *                     except OverflowError:             # <<<<<<<<<<<<<<
 *                         if self.options.clamp_float_overflow:
 *                             return DBL_MAX
 */
          __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OverflowError);
          if (__pyx_t_10) {
            __Pyx_AddTraceback("cavro.DoubleType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(10, 378, __pyx_L18_except_error)
            __Pyx_XGOTREF(__pyx_t_2);
            __Pyx_XGOTREF(__pyx_t_3);
            __Pyx_XGOTREF(__pyx_t_4);

            /* "src/numeric_types.pxi":379
 *                         value = float(value)
 *                     except OverflowError:
 *                         if self.options.clamp_float_overflow:             # <<<<<<<<<<<<<<
 *                             return DBL_MAX
 *                         else:
 */
            if (likely(__pyx_v_self->__pyx_base.options->clamp_float_overflow)) {

              /* "src/numeric_types.pxi":380
 *                     except OverflowError:
 *                         if self.options.clamp_float_overflow:
 *                             return DBL_MAX             # <<<<<<<<<<<<<<
 *                         else:
 *                             raise
 */
              __Pyx_XDECREF(__pyx_r);
              __pyx_t_11 = PyFloat_FromDouble(DBL_MAX); if (unlikely(!__pyx_t_11)) __PYX_ERR(10, 380, __pyx_L18_except_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_r = __pyx_t_11;
              __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              goto __pyx_L19_except_return;

              /* "src/numeric_types.pxi":379
 *                         value = float(value)
 *                     except OverflowError:
 *                         if self.options.clamp_float_overflow:             # <<<<<<<<<<<<<<
 *                             return DBL_MAX
 *                         else:
 */
            }

            /* "src/numeric_types.pxi":382
 *                             return DBL_MAX
 *                         else:
 *                             raise             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise InvalidValue(value, self)
 */
            /*else*/ {
              __Pyx_GIVEREF(__pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_3);
              __Pyx_XGIVEREF(__pyx_t_4);
              __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_3, __pyx_t_4);
              __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_4 = 0; 
              __PYX_ERR(10, 382, __pyx_L18_except_error)
            }
          }
          goto __pyx_L18_except_error;

          /* "src/numeric_types.pxi":376
 *             else:
 *                 if self.options.coerce_values_to_float:
 *                     try:             # <<<<<<<<<<<<<<
 *                         value = float(value)
 *                     except OverflowError:
 */
          __pyx_L18_except_error:;
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
          goto __pyx_L1_error;
          __pyx_L19_except_return:;
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
          goto __pyx_L0;
          __pyx_L21_try_end:;
        }

        /* "src/numeric_types.pxi":375
 *                 value = float(value)
 *             else:
 *                 if self.options.coerce_values_to_float:             # <<<<<<<<<<<<<<
 *                     try:
 *                         value = float(value)
 */
        goto __pyx_L15;
      }

      /* "src/numeric_types.pxi":384
 *                             raise
 *                 else:
 *                     raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         if isnan(value):
 *             if self.options.clamp_float_overflow:
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 384, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        __pyx_t_10 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_10 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_value, ((PyObject *)__pyx_v_self)};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_10, 2+__pyx_t_10);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 384, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(10, 384, __pyx_L1_error)
      }
      __pyx_L15:;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "src/numeric_types.pxi":385
 *                 else:
 *                     raise InvalidValue(value, self)
 *         if isnan(value):             # <<<<<<<<<<<<<<
 *             if self.options.clamp_float_overflow:
 *                 return 0.0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_isnan); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_10 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_10 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_value};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(10, 385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_1) {

    /* "src/numeric_types.pxi":386
 *                     raise InvalidValue(value, self)
 *         if isnan(value):
 *             if self.options.clamp_float_overflow:             # <<<<<<<<<<<<<<
 *                 return 0.0
 *         elif  (value):
 */
    if (__pyx_v_self->__pyx_base.options->clamp_float_overflow) {

      /* "src/numeric_types.pxi":387
 *         if isnan(value):
 *             if self.options.clamp_float_overflow:
 *                 return 0.0             # <<<<<<<<<<<<<<
 *         elif  (value):
 *             if self.options.clamp_float_overflow:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_float_0_0);
      __pyx_r = __pyx_float_0_0;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":386
 *                     raise InvalidValue(value, self)
 *         if isnan(value):
 *             if self.options.clamp_float_overflow:             # <<<<<<<<<<<<<<
 *                 return 0.0
 *         elif  (value):
 */
    }

    /* "src/numeric_types.pxi":385
 *                 else:
 *                     raise InvalidValue(value, self)
 *         if isnan(value):             # <<<<<<<<<<<<<<
 *             if self.options.clamp_float_overflow:
 *                 return 0.0
 */
    goto __pyx_L25;
  }

  /* "src/numeric_types.pxi":388
 *             if self.options.clamp_float_overflow:
 *                 return 0.0
 *         elif  (value):             # <<<<<<<<<<<<<<
 *             if self.options.clamp_float_overflow:
 *                 return DBL_MAX
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(10, 388, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/numeric_types.pxi":389
 *                 return 0.0
 *         elif  (value):
 *             if self.options.clamp_float_overflow:             # <<<<<<<<<<<<<<
 *                 return DBL_MAX
 * 
 */
    if (__pyx_v_self->__pyx_base.options->clamp_float_overflow) {

      /* "src/numeric_types.pxi":390
 *         elif  (value):
 *             if self.options.clamp_float_overflow:
 *                 return DBL_MAX             # <<<<<<<<<<<<<<
 * 
 *         return value
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = PyFloat_FromDouble(DBL_MAX); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 390, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "src/numeric_types.pxi":389
 *                 return 0.0
 *         elif  (value):
 *             if self.options.clamp_float_overflow:             # <<<<<<<<<<<<<<
 *                 return DBL_MAX
 * 
 */
    }

    /* "src/numeric_types.pxi":388
 *             if self.options.clamp_float_overflow:
 *                 return 0.0
 *         elif  (value):             # <<<<<<<<<<<<<<
 *             if self.options.clamp_float_overflow:
 *                 return DBL_MAX
 */
  }
  __pyx_L25:;

  /* "src/numeric_types.pxi":392
 *                 return DBL_MAX
 * 
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":366
 *         return max_fit
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, float):
 *             pass
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("cavro.DoubleType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10DoubleType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10DoubleType_5_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10DoubleType_5_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10DoubleType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 366, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(10, 366, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(10, 366, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.DoubleType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10DoubleType_4_convert_value(((struct __pyx_obj_5cavro_DoubleType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10DoubleType_4_convert_value(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_10DoubleType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.DoubleType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":394
 *         return value
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */

static PyObject *__pyx_f_5cavro_10DoubleType__json_format(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 1);

  /* "src/numeric_types.pxi":395
 * 
 *     cdef _json_format(self, value):
 *         return self._convert_value(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_10DoubleType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":394
 *         return value
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.DoubleType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":397
 *         return self._convert_value(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (py_bool, bool_)):
 *             raise InvalidValue(value, self)
 */

static PyObject *__pyx_f_5cavro_10DoubleType__json_decode(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/numeric_types.pxi":398
 * 
 *     cdef _json_decode(self, value):
 *         if isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if not isinstance(value, (float, int)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_7cpython_4bool_bool); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_1); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "src/numeric_types.pxi":399
 *     cdef _json_decode(self, value):
 *         if isinstance(value, (py_bool, bool_)):
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (float, int)):
 *             raise InvalidValue(value, self)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 399, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(10, 399, __pyx_L1_error)

    /* "src/numeric_types.pxi":398
 * 
 *     cdef _json_decode(self, value):
 *         if isinstance(value, (py_bool, bool_)):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if not isinstance(value, (float, int)):
 */
  }

  /* "src/numeric_types.pxi":400
 *         if isinstance(value, (py_bool, bool_)):
 *             raise InvalidValue(value, self)
 *         if not isinstance(value, (float, int)):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         return value
 */
  __pyx_t_3 = PyFloat_Check(__pyx_v_value); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = PyInt_Check(__pyx_v_value); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  __pyx_t_3 = (!__pyx_t_2);
  if (unlikely(__pyx_t_3)) {

    /* "src/numeric_types.pxi":401
 *             raise InvalidValue(value, self)
 *         if not isinstance(value, (float, int)):
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 401, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(10, 401, __pyx_L1_error)

    /* "src/numeric_types.pxi":400
 *         if isinstance(value, (py_bool, bool_)):
 *             raise InvalidValue(value, self)
 *         if not isinstance(value, (float, int)):             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         return value
 */
  }

  /* "src/numeric_types.pxi":402
 *         if not isinstance(value, (float, int)):
 *             raise InvalidValue(value, self)
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":397
 *         return self._convert_value(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (py_bool, bool_)):
 *             raise InvalidValue(value, self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.DoubleType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":404
 *         return value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"double"')
 * 
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_10DoubleType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/numeric_types.pxi":405
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"double"')             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_tuple__55, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/numeric_types.pxi":404
 *         return value
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"double"')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.DoubleType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/numeric_types.pxi":407
 *         return _CanonicalForm('"double"')
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         cdef AvroType promoted
 *         if isinstance(writer, FloatType):
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10DoubleType__for_writer(CYTHON_UNUSED struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_promoted = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_writer", 1);

  /* "src/numeric_types.pxi":409
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         cdef AvroType promoted
 *         if isinstance(writer, FloatType):             # <<<<<<<<<<<<<<
 *             return writer
 *         if isinstance(writer, (IntType, LongType)):
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_FloatType); 
  if (__pyx_t_1) {

    /* "src/numeric_types.pxi":410
 *         cdef AvroType promoted
 *         if isinstance(writer, FloatType):
 *             return writer             # <<<<<<<<<<<<<<
 *         if isinstance(writer, (IntType, LongType)):
 *             promoted = writer.clone_base()
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_writer);
    __pyx_r = __pyx_v_writer;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":409
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         cdef AvroType promoted
 *         if isinstance(writer, FloatType):             # <<<<<<<<<<<<<<
 *             return writer
 *         if isinstance(writer, (IntType, LongType)):
 */
  }

  /* "src/numeric_types.pxi":411
 *         if isinstance(writer, FloatType):
 *             return writer
 *         if isinstance(writer, (IntType, LongType)):             # <<<<<<<<<<<<<<
 *             promoted = writer.clone_base()
 *             promoted.value_adapters = promoted.value_adapters + (PromoteToFloat(),)
 */
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_IntType); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_LongType); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/numeric_types.pxi":412
 *             return writer
 *         if isinstance(writer, (IntType, LongType)):
 *             promoted = writer.clone_base()             # <<<<<<<<<<<<<<
 *             promoted.value_adapters = promoted.value_adapters + (PromoteToFloat(),)
 *             return promoted
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_writer->__pyx_vtab)->clone_base(__pyx_v_writer, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_promoted = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/numeric_types.pxi":413
 *         if isinstance(writer, (IntType, LongType)):
 *             promoted = writer.clone_base()
 *             promoted.value_adapters = promoted.value_adapters + (PromoteToFloat(),)             # <<<<<<<<<<<<<<
 *             return promoted
 */
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro_PromoteToFloat)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(10, 413, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_v_promoted->value_adapters, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_promoted->value_adapters);
    __Pyx_DECREF(__pyx_v_promoted->value_adapters);
    __pyx_v_promoted->value_adapters = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/numeric_types.pxi":414
 *             promoted = writer.clone_base()
 *             promoted.value_adapters = promoted.value_adapters + (PromoteToFloat(),)
 *             return promoted             # <<<<<<<<<<<<<<
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_promoted);
    __pyx_r = __pyx_v_promoted;
    goto __pyx_L0;

    /* "src/numeric_types.pxi":411
 *         if isinstance(writer, FloatType):
 *             return writer
 *         if isinstance(writer, (IntType, LongType)):             # <<<<<<<<<<<<<<
 *             promoted = writer.clone_base()
 *             promoted.value_adapters = promoted.value_adapters + (PromoteToFloat(),)
 */
  }

  /* "src/numeric_types.pxi":407
 *         return _CanonicalForm('"double"')
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         cdef AvroType promoted
 *         if isinstance(writer, FloatType):
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.DoubleType._for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_promoted);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10DoubleType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10DoubleType_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10DoubleType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10DoubleType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10DoubleType_6__reduce_cython__(((struct __pyx_obj_5cavro_DoubleType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10DoubleType_6__reduce_cython__(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_DoubleType, (type(self), 0x84c5042, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DoubleType, (type(self), 0x84c5042, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_DoubleType, (type(self), 0x84c5042, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_DoubleType, (type(self), 0x84c5042, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_DoubleType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DoubleType, (type(self), 0x84c5042, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_DoubleType, (type(self), 0x84c5042, None), state
 *     else:
 *         return __pyx_unpickle_DoubleType, (type(self), 0x84c5042, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DoubleType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_DoubleType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.DoubleType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DoubleType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DoubleType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10DoubleType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10DoubleType_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10DoubleType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10DoubleType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.DoubleType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10DoubleType_8__setstate_cython__(((struct __pyx_obj_5cavro_DoubleType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10DoubleType_8__setstate_cython__(struct __pyx_obj_5cavro_DoubleType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_DoubleType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DoubleType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_DoubleType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DoubleType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DoubleType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.DoubleType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":26
 *     cdef readonly object default_value
 * 
 *     def __init__(self, options, default_value):             # <<<<<<<<<<<<<<
 *         schema = Schema('"null"', options)
 *         AvroType.__init__(self, schema, None, None)
 */

/* Python wrapper */
static int __pyx_pw_5cavro_16_PlaceholderType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_16_PlaceholderType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_options = 0;
  PyObject *__pyx_v_default_value = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_options,&__pyx_n_s_default_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_options)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 26, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_default_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 26, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(11, 26, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(11, 26, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_options = values[0];
    __pyx_v_default_value = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(11, 26, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._PlaceholderType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_16_PlaceholderType___init__(((struct __pyx_obj_5cavro__PlaceholderType *)__pyx_v_self), __pyx_v_options, __pyx_v_default_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_16_PlaceholderType___init__(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, PyObject *__pyx_v_options, PyObject *__pyx_v_default_value) {
  struct __pyx_obj_5cavro_Schema *__pyx_v_schema = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/record.pxi":27
 * 
 *     def __init__(self, options, default_value):
 *         schema = Schema('"null"', options)             # <<<<<<<<<<<<<<
 *         AvroType.__init__(self, schema, None, None)
 *         self.default_value = None if default_value is MISSING_VALUE else default_value
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_kp_u_null);
  __Pyx_GIVEREF(__pyx_kp_u_null);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_null)) __PYX_ERR(11, 27, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_options);
  __Pyx_GIVEREF(__pyx_v_options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_options)) __PYX_ERR(11, 27, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_Schema), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_schema = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/record.pxi":28
 *     def __init__(self, options, default_value):
 *         schema = Schema('"null"', options)
 *         AvroType.__init__(self, schema, None, None)             # <<<<<<<<<<<<<<
 *         self.default_value = None if default_value is MISSING_VALUE else default_value
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_AvroType), __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_3, ((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_schema), Py_None, Py_None};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 4+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 28, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/record.pxi":29
 *         schema = Schema('"null"', options)
 *         AvroType.__init__(self, schema, None, None)
 *         self.default_value = None if default_value is MISSING_VALUE else default_value             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = (__pyx_v_default_value == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {
    __Pyx_INCREF(Py_None);
    __pyx_t_2 = Py_None;
  } else {
    __Pyx_INCREF(__pyx_v_default_value);
    __pyx_t_2 = __pyx_v_default_value;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->default_value);
  __Pyx_DECREF(__pyx_v_self->default_value);
  __pyx_v_self->default_value = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "src/record.pxi":26
 *     cdef readonly object default_value
 * 
 *     def __init__(self, options, default_value):             # <<<<<<<<<<<<<<
 *         schema = Schema('"null"', options)
 *         AvroType.__init__(self, schema, None, None)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro._PlaceholderType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_schema);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":31
 *         self.default_value = None if default_value is MISSING_VALUE else default_value
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_f_5cavro_16_PlaceholderType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/record.pxi":32
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":31
 *         self.default_value = None if default_value is MISSING_VALUE else default_value
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._PlaceholderType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":34
 *         return {}
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_pw_5cavro_16_PlaceholderType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_16_PlaceholderType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_schema_extra); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 34, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_16_PlaceholderType_3_get_schema_extra)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_created};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 34, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_2))) __PYX_ERR(11, 34, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/record.pxi":35
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     cdef _make_logical(self, schema, source):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":34
 *         return {}
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._PlaceholderType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_16_PlaceholderType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_16_PlaceholderType_3_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_16_PlaceholderType_3_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_16_PlaceholderType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 34, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(11, 34, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(11, 34, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._PlaceholderType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(11, 34, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_16_PlaceholderType_2_get_schema_extra(((struct __pyx_obj_5cavro__PlaceholderType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_16_PlaceholderType_2_get_schema_extra(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_16_PlaceholderType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._PlaceholderType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":37
 *         return {}
 * 
 *     cdef _make_logical(self, schema, source):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static PyObject *__pyx_f_5cavro_16_PlaceholderType__make_logical(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_schema, CYTHON_UNUSED PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_make_logical", 1);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":40
 *         pass
 * 
 *     cdef int binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("Placeholder types cannot be encoded")
 * 
 */

static int __pyx_f_5cavro_16_PlaceholderType_binary_buffer_encode(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, CYTHON_UNUSED PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("binary_buffer_encode", 1);

  /* "src/record.pxi":41
 * 
 *     cdef int binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         raise NotImplementedError("Placeholder types cannot be encoded")             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(11, 41, __pyx_L1_error)

  /* "src/record.pxi":40
 *         pass
 * 
 *     cdef int binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("Placeholder types cannot be encoded")
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._PlaceholderType.binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":43
 *         raise NotImplementedError("Placeholder types cannot be encoded")
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return self.default_value
 * 
 */

static PyObject *__pyx_f_5cavro_16_PlaceholderType__binary_buffer_decode(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/record.pxi":44
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         return self.default_value             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->default_value);
  __pyx_r = __pyx_v_self->default_value;
  goto __pyx_L0;

  /* "src/record.pxi":43
 *         raise NotImplementedError("Placeholder types cannot be encoded")
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return self.default_value
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":46
 *         return self.default_value
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         return FIT_NONE
 * 
 */

static int __pyx_f_5cavro_16_PlaceholderType__get_value_fitness(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  int __pyx_r;

  /* "src/record.pxi":47
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         return FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
  __pyx_r = __pyx_v_5cavro_FIT_NONE;
  goto __pyx_L0;

  /* "src/record.pxi":46
 *         return self.default_value
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         return FIT_NONE
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "src/record.pxi":49
 *         return FIT_NONE
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("Placeholder types cannot be encoded")
 * 
 */

static PyObject *__pyx_f_5cavro_16_PlaceholderType__json_format(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 1);

  /* "src/record.pxi":50
 * 
 *     cdef _json_format(self, value):
 *         raise NotImplementedError("Placeholder types cannot be encoded")             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(11, 50, __pyx_L1_error)

  /* "src/record.pxi":49
 *         return FIT_NONE
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("Placeholder types cannot be encoded")
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._PlaceholderType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":52
 *         raise NotImplementedError("Placeholder types cannot be encoded")
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         return self.default_value
 * 
 */

static PyObject *__pyx_f_5cavro_16_PlaceholderType__json_decode(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/record.pxi":53
 * 
 *     cdef _json_decode(self, value):
 *         return self.default_value             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->default_value);
  __pyx_r = __pyx_v_self->default_value;
  goto __pyx_L0;

  /* "src/record.pxi":52
 *         raise NotImplementedError("Placeholder types cannot be encoded")
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         return self.default_value
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":55
 *         return self.default_value
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

static PyObject *__pyx_pw_5cavro_16_PlaceholderType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_16_PlaceholderType__convert_value(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_convert_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 55, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_16_PlaceholderType_5_convert_value)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_value};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 55, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/record.pxi":56
 * 
 *     cpdef object _convert_value(self, object value):
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "src/record.pxi":55
 *         return self.default_value
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._PlaceholderType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_16_PlaceholderType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_16_PlaceholderType_5_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_16_PlaceholderType_5_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_16_PlaceholderType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 55, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(11, 55, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(11, 55, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._PlaceholderType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_16_PlaceholderType_4_convert_value(((struct __pyx_obj_5cavro__PlaceholderType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_16_PlaceholderType_4_convert_value(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_16_PlaceholderType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._PlaceholderType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":58
 *         return None
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("Placeholder types have no canonical form")
 * 
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_16_PlaceholderType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/record.pxi":59
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         raise NotImplementedError("Placeholder types have no canonical form")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(11, 59, __pyx_L1_error)

  /* "src/record.pxi":58
 *         return None
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError("Placeholder types have no canonical form")
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._PlaceholderType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":24
 *     type_name = "placeholder"
 * 
 *     cdef readonly object default_value             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, options, default_value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_16_PlaceholderType_13default_value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_16_PlaceholderType_13default_value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_16_PlaceholderType_13default_value___get__(((struct __pyx_obj_5cavro__PlaceholderType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_16_PlaceholderType_13default_value___get__(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->default_value);
  __pyx_r = __pyx_v_self->default_value;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_16_PlaceholderType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_16_PlaceholderType_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_16_PlaceholderType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_16_PlaceholderType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_16_PlaceholderType_6__reduce_cython__(((struct __pyx_obj_5cavro__PlaceholderType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_16_PlaceholderType_6__reduce_cython__(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.default_value, self.metadata, self.options, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->default_value);
  __Pyx_GIVEREF(__pyx_v_self->default_value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->default_value)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.default_value, self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.default_value, self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.default_value is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.default_value, self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.default_value is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__PlaceholderType, (type(self), 0x5c2152b, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->default_value != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.default_value is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__PlaceholderType, (type(self), 0x5c2152b, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.default_value is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle__PlaceholderType, (type(self), 0x5c2152b, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__PlaceholderType, (type(self), 0x5c2152b, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle__PlaceholderType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_96605483);
    __Pyx_GIVEREF(__pyx_int_96605483);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_96605483)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.default_value is not None or self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__PlaceholderType, (type(self), 0x5c2152b, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__PlaceholderType, (type(self), 0x5c2152b, None), state
 *     else:
 *         return __pyx_unpickle__PlaceholderType, (type(self), 0x5c2152b, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__PlaceholderType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle__PlaceholderType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_96605483);
    __Pyx_GIVEREF(__pyx_int_96605483);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_96605483)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro._PlaceholderType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__PlaceholderType, (type(self), 0x5c2152b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__PlaceholderType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_16_PlaceholderType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_16_PlaceholderType_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_16_PlaceholderType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_16_PlaceholderType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._PlaceholderType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_16_PlaceholderType_8__setstate_cython__(((struct __pyx_obj_5cavro__PlaceholderType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_16_PlaceholderType_8__setstate_cython__(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__PlaceholderType, (type(self), 0x5c2152b, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__PlaceholderType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__PlaceholderType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__PlaceholderType, (type(self), 0x5c2152b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__PlaceholderType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._PlaceholderType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":62
 * 
 * 
 * cdef list _record_data_from_dict(RecordType record, dict data):             # <<<<<<<<<<<<<<
 *     cdef list field_data = [None] * len(record.fields)
 *     cdef Py_ssize_t index = 0
 */

static PyObject *__pyx_f_5cavro__record_data_from_dict(struct __pyx_obj_5cavro_RecordType *__pyx_v_record, PyObject *__pyx_v_data) {
  PyObject *__pyx_v_field_data = 0;
  Py_ssize_t __pyx_v_index;
  struct __pyx_obj_5cavro_Options *__pyx_v_options = 0;
  PyObject *__pyx_v_extra_fields = NULL;
  PyObject *__pyx_v_extra_field = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_missing_fields = NULL;
  PyObject *__pyx_v_missing_field = NULL;
  PyObject *__pyx_v_field = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_record_data_from_dict", 1);

  /* "src/record.pxi":63
 * 
 * cdef list _record_data_from_dict(RecordType record, dict data):
 *     cdef list field_data = [None] * len(record.fields)             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t index = 0
 *     cdef Options options = record.options
 */
  __pyx_t_1 = __pyx_v_record->fields;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(11, 63, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(11, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(1 * ((__pyx_t_2<0) ? 0:__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_t_2; __pyx_temp++) {
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_1, __pyx_temp, Py_None)) __PYX_ERR(11, 63, __pyx_L1_error);
    }
  }
  __pyx_v_field_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":64
 * cdef list _record_data_from_dict(RecordType record, dict data):
 *     cdef list field_data = [None] * len(record.fields)
 *     cdef Py_ssize_t index = 0             # <<<<<<<<<<<<<<
 *     cdef Options options = record.options
 * 
 */
  __pyx_v_index = 0;

  /* "src/record.pxi":65
 *     cdef list field_data = [None] * len(record.fields)
 *     cdef Py_ssize_t index = 0
 *     cdef Options options = record.options             # <<<<<<<<<<<<<<
 * 
 *     if not options.record_allow_extra_fields:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_record->__pyx_base.__pyx_base.options);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":67
 *     cdef Options options = record.options
 * 
 *     if not options.record_allow_extra_fields:             # <<<<<<<<<<<<<<
 *         extra_fields = data.keys() - record.field_dict.keys()
 *         if extra_fields:
 */
  __pyx_t_3 = (!__pyx_v_options->record_allow_extra_fields);
  if (__pyx_t_3) {

    /* "src/record.pxi":68
 * 
 *     if not options.record_allow_extra_fields:
 *         extra_fields = data.keys() - record.field_dict.keys()             # <<<<<<<<<<<<<<
 *         if extra_fields:
 *             extra_field, *_ = extra_fields
 */
    if (unlikely(__pyx_v_data == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
      __PYX_ERR(11, 68, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_Keys(__pyx_v_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_record->field_dict == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
      __PYX_ERR(11, 68, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyDict_Keys(__pyx_v_record->field_dict); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyNumber_Subtract(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_extra_fields = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "src/record.pxi":69
 *     if not options.record_allow_extra_fields:
 *         extra_fields = data.keys() - record.field_dict.keys()
 *         if extra_fields:             # <<<<<<<<<<<<<<
 *             extra_field, *_ = extra_fields
 *             raise InvalidValue('...', record, (extra_field, ))
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_extra_fields); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(11, 69, __pyx_L1_error)
    if (unlikely(__pyx_t_3)) {

      /* "src/record.pxi":70
 *         extra_fields = data.keys() - record.field_dict.keys()
 *         if extra_fields:
 *             extra_field, *_ = extra_fields             # <<<<<<<<<<<<<<
 *             raise InvalidValue('...', record, (extra_field, ))
 *     if not options.record_encode_use_defaults:
 */
      {
        Py_ssize_t index = -1;
        PyObject** temps[2] = {&__pyx_t_5};
        __pyx_t_4 = PyObject_GetIter(__pyx_v_extra_fields); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 70, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
        for (index=0; index < 1; index++) {
          PyObject* item = __pyx_t_6(__pyx_t_4); if (unlikely(!item)) goto __pyx_L5_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        goto __pyx_L6_unpacking_done;
        __pyx_L5_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_6 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(11, 70, __pyx_L1_error)
        __pyx_L6_unpacking_done:;
      }
      __pyx_t_1 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 70, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_extra_field = __pyx_t_5;
      __pyx_t_5 = 0;
      __pyx_v__ = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "src/record.pxi":71
 *         if extra_fields:
 *             extra_field, *_ = extra_fields
 *             raise InvalidValue('...', record, (extra_field, ))             # <<<<<<<<<<<<<<
 *     if not options.record_encode_use_defaults:
 *         missing_fields = record.field_dict.keys() - data.keys()
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 71, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 71, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_extra_field);
      __Pyx_GIVEREF(__pyx_v_extra_field);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_extra_field)) __PYX_ERR(11, 71, __pyx_L1_error);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_7, __pyx_kp_u__18, ((PyObject *)__pyx_v_record), __pyx_t_4};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 3+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 71, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(11, 71, __pyx_L1_error)

      /* "src/record.pxi":69
 *     if not options.record_allow_extra_fields:
 *         extra_fields = data.keys() - record.field_dict.keys()
 *         if extra_fields:             # <<<<<<<<<<<<<<
 *             extra_field, *_ = extra_fields
 *             raise InvalidValue('...', record, (extra_field, ))
 */
    }

    /* "src/record.pxi":67
 *     cdef Options options = record.options
 * 
 *     if not options.record_allow_extra_fields:             # <<<<<<<<<<<<<<
 *         extra_fields = data.keys() - record.field_dict.keys()
 *         if extra_fields:
 */
  }

  /* "src/record.pxi":72
 *             extra_field, *_ = extra_fields
 *             raise InvalidValue('...', record, (extra_field, ))
 *     if not options.record_encode_use_defaults:             # <<<<<<<<<<<<<<
 *         missing_fields = record.field_dict.keys() - data.keys()
 *         if missing_fields:
 */
  __pyx_t_3 = (!__pyx_v_options->record_encode_use_defaults);
  if (__pyx_t_3) {

    /* "src/record.pxi":73
 *             raise InvalidValue('...', record, (extra_field, ))
 *     if not options.record_encode_use_defaults:
 *         missing_fields = record.field_dict.keys() - data.keys()             # <<<<<<<<<<<<<<
 *         if missing_fields:
 *             missing_field, *_ = missing_fields
 */
    if (unlikely(__pyx_v_record->field_dict == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
      __PYX_ERR(11, 73, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_Keys(__pyx_v_record->field_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_data == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
      __PYX_ERR(11, 73, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyDict_Keys(__pyx_v_data); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyNumber_Subtract(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_missing_fields = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "src/record.pxi":74
 *     if not options.record_encode_use_defaults:
 *         missing_fields = record.field_dict.keys() - data.keys()
 *         if missing_fields:             # <<<<<<<<<<<<<<
 *             missing_field, *_ = missing_fields
 *             raise InvalidValue('<missing>', record, (missing_field, ))
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_missing_fields); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(11, 74, __pyx_L1_error)
    if (unlikely(__pyx_t_3)) {

      /* "src/record.pxi":75
 *         missing_fields = record.field_dict.keys() - data.keys()
 *         if missing_fields:
 *             missing_field, *_ = missing_fields             # <<<<<<<<<<<<<<
 *             raise InvalidValue('<missing>', record, (missing_field, ))
 * 
 */
      {
        Py_ssize_t index = -1;
        PyObject** temps[2] = {&__pyx_t_4};
        __pyx_t_5 = PyObject_GetIter(__pyx_v_missing_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 75, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5);
        for (index=0; index < 1; index++) {
          PyObject* item = __pyx_t_6(__pyx_t_5); if (unlikely(!item)) goto __pyx_L9_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        goto __pyx_L10_unpacking_done;
        __pyx_L9_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_6 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(11, 75, __pyx_L1_error)
        __pyx_L10_unpacking_done:;
      }
      __pyx_t_1 = PySequence_List(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 75, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_missing_field = __pyx_t_4;
      __pyx_t_4 = 0;
      __pyx_v__ = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "src/record.pxi":76
 *         if missing_fields:
 *             missing_field, *_ = missing_fields
 *             raise InvalidValue('<missing>', record, (missing_field, ))             # <<<<<<<<<<<<<<
 * 
 *     for field in record.fields:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_missing_field);
      __Pyx_GIVEREF(__pyx_v_missing_field);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_missing_field)) __PYX_ERR(11, 76, __pyx_L1_error);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_7, __pyx_kp_u_missing, ((PyObject *)__pyx_v_record), __pyx_t_5};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 3+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 76, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(11, 76, __pyx_L1_error)

      /* "src/record.pxi":74
 *     if not options.record_encode_use_defaults:
 *         missing_fields = record.field_dict.keys() - data.keys()
 *         if missing_fields:             # <<<<<<<<<<<<<<
 *             missing_field, *_ = missing_fields
 *             raise InvalidValue('<missing>', record, (missing_field, ))
 */
    }

    /* "src/record.pxi":72
 *             extra_field, *_ = extra_fields
 *             raise InvalidValue('...', record, (extra_field, ))
 *     if not options.record_encode_use_defaults:             # <<<<<<<<<<<<<<
 *         missing_fields = record.field_dict.keys() - data.keys()
 *         if missing_fields:
 */
  }

  /* "src/record.pxi":78
 *             raise InvalidValue('<missing>', record, (missing_field, ))
 * 
 *     for field in record.fields:             # <<<<<<<<<<<<<<
 *         value = data.get(field.name, field.default_value)
 *         if value is NO_DEFAULT:
 */
  if (unlikely(__pyx_v_record->fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(11, 78, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_record->fields; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 78, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(11, 78, __pyx_L1_error)
    #else
    __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/record.pxi":79
 * 
 *     for field in record.fields:
 *         value = data.get(field.name, field.default_value)             # <<<<<<<<<<<<<<
 *         if value is NO_DEFAULT:
 *             raise ValueError(f"Field {field.name} is required for record {record.type}")
 */
    if (unlikely(__pyx_v_data == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
      __PYX_ERR(11, 79, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_default_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PyDict_GetItemDefault(__pyx_v_data, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "src/record.pxi":80
 *     for field in record.fields:
 *         value = data.get(field.name, field.default_value)
 *         if value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Field {field.name} is required for record {record.type}")
 *         field_data[index] = field.type.convert_value(value)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = (__pyx_v_value == __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_t_3)) {

      /* "src/record.pxi":81
 *         value = data.get(field.name, field.default_value)
 *         if value is NO_DEFAULT:
 *             raise ValueError(f"Field {field.name} is required for record {record.type}")             # <<<<<<<<<<<<<<
 *         field_data[index] = field.type.convert_value(value)
 *         index += 1
 */
      __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;
      __Pyx_INCREF(__pyx_kp_u_Field);
      __pyx_t_9 += 6;
      __Pyx_GIVEREF(__pyx_kp_u_Field);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_kp_u_Field);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_INCREF(__pyx_kp_u_is_required_for_record);
      __pyx_t_9 += 24;
      __Pyx_GIVEREF(__pyx_kp_u_is_required_for_record);
      PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_kp_u_is_required_for_record);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_record), __pyx_n_s_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_7, 4, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(11, 81, __pyx_L1_error)

      /* "src/record.pxi":80
 *     for field in record.fields:
 *         value = data.get(field.name, field.default_value)
 *         if value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Field {field.name} is required for record {record.type}")
 *         field_data[index] = field.type.convert_value(value)
 */
    }

    /* "src/record.pxi":82
 *         if value is NO_DEFAULT:
 *             raise ValueError(f"Field {field.name} is required for record {record.type}")
 *         field_data[index] = field.type.convert_value(value)             # <<<<<<<<<<<<<<
 *         index += 1
 *     return field_data
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_convert_value_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_value};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 82, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    if (unlikely((__Pyx_SetItemInt(__pyx_v_field_data, __pyx_v_index, __pyx_t_7, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1) < 0))) __PYX_ERR(11, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "src/record.pxi":83
 *             raise ValueError(f"Field {field.name} is required for record {record.type}")
 *         field_data[index] = field.type.convert_value(value)
 *         index += 1             # <<<<<<<<<<<<<<
 *     return field_data
 * 
 */
    __pyx_v_index = (__pyx_v_index + 1);

    /* "src/record.pxi":78
 *             raise InvalidValue('<missing>', record, (missing_field, ))
 * 
 *     for field in record.fields:             # <<<<<<<<<<<<<<
 *         value = data.get(field.name, field.default_value)
 *         if value is NO_DEFAULT:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/record.pxi":84
 *         field_data[index] = field.type.convert_value(value)
 *         index += 1
 *     return field_data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_field_data);
  __pyx_r = __pyx_v_field_data;
  goto __pyx_L0;

  /* "src/record.pxi":62
 * 
 * 
 * cdef list _record_data_from_dict(RecordType record, dict data):             # <<<<<<<<<<<<<<
 *     cdef list field_data = [None] * len(record.fields)
 *     cdef Py_ssize_t index = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro._record_data_from_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_field_data);
  __Pyx_XDECREF((PyObject *)__pyx_v_options);
  __Pyx_XDECREF(__pyx_v_extra_fields);
  __Pyx_XDECREF(__pyx_v_extra_field);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_missing_fields);
  __Pyx_XDECREF(__pyx_v_missing_field);
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":107
 *     cdef list data
 * 
 *     def __init__(self, data=None, **kwargs):             # <<<<<<<<<<<<<<
 *         cdef dict data_dict
 *         cdef Record rec
 */

/* Python wrapper */
static int __pyx_pw_5cavro_6Record_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_6Record_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 107, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(11, 107, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_data = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, __pyx_nargs); __PYX_ERR(11, 107, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("cavro.Record.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6Record___init__(((struct __pyx_obj_5cavro_Record *)__pyx_v_self), __pyx_v_data, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_6Record___init__(struct __pyx_obj_5cavro_Record *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_data_dict = 0;
  struct __pyx_obj_5cavro_Record *__pyx_v_rec = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/record.pxi":110
 *         cdef dict data_dict
 *         cdef Record rec
 *         if isinstance(data, type(self)):             # <<<<<<<<<<<<<<
 *             rec = data
 *             self.data = rec.data
 */
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_data, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(11, 110, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/record.pxi":111
 *         cdef Record rec
 *         if isinstance(data, type(self)):
 *             rec = data             # <<<<<<<<<<<<<<
 *             self.data = rec.data
 *             return
 */
    if (!(likely(((__pyx_v_data) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_data, __pyx_ptype_5cavro_Record))))) __PYX_ERR(11, 111, __pyx_L1_error)
    __pyx_t_2 = __pyx_v_data;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_rec = ((struct __pyx_obj_5cavro_Record *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/record.pxi":112
 *         if isinstance(data, type(self)):
 *             rec = data
 *             self.data = rec.data             # <<<<<<<<<<<<<<
 *             return
 *         if isinstance(data, Record):
 */
    __pyx_t_2 = __pyx_v_rec->data;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->data);
    __Pyx_DECREF(__pyx_v_self->data);
    __pyx_v_self->data = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/record.pxi":113
 *             rec = data
 *             self.data = rec.data
 *             return             # <<<<<<<<<<<<<<
 *         if isinstance(data, Record):
 *             if data.Type.name == self.Type.name and self.Type.options.adapt_record_types:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "src/record.pxi":110
 *         cdef dict data_dict
 *         cdef Record rec
 *         if isinstance(data, type(self)):             # <<<<<<<<<<<<<<
 *             rec = data
 *             self.data = rec.data
 */
  }

  /* "src/record.pxi":114
 *             self.data = rec.data
 *             return
 *         if isinstance(data, Record):             # <<<<<<<<<<<<<<
 *             if data.Type.name == self.Type.name and self.Type.options.adapt_record_types:
 *                 data_dict = data._asdict()
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_data, __pyx_ptype_5cavro_Record); 
  if (__pyx_t_1) {

    /* "src/record.pxi":115
 *             return
 *         if isinstance(data, Record):
 *             if data.Type.name == self.Type.name and self.Type.options.adapt_record_types:             # <<<<<<<<<<<<<<
 *                 data_dict = data._asdict()
 *             else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_Type); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Type); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 115, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(11, 115, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_1 = __pyx_t_5;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Type); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_options); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_adapt_record_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(11, 115, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __pyx_t_5;
    __pyx_L6_bool_binop_done:;
    if (likely(__pyx_t_1)) {

      /* "src/record.pxi":116
 *         if isinstance(data, Record):
 *             if data.Type.name == self.Type.name and self.Type.options.adapt_record_types:
 *                 data_dict = data._asdict()             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(f"Record {data} cannot be adapted to {self.Type}")
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_asdict); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 116, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 116, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_2))) __PYX_ERR(11, 116, __pyx_L1_error)
      __pyx_v_data_dict = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "src/record.pxi":115
 *             return
 *         if isinstance(data, Record):
 *             if data.Type.name == self.Type.name and self.Type.options.adapt_record_types:             # <<<<<<<<<<<<<<
 *                 data_dict = data._asdict()
 *             else:
 */
      goto __pyx_L5;
    }

    /* "src/record.pxi":118
 *                 data_dict = data._asdict()
 *             else:
 *                 raise ValueError(f"Record {data} cannot be adapted to {self.Type}")             # <<<<<<<<<<<<<<
 *             if kwargs:
 *                 raise ValueError(f"Records may either be instantiated with a record, or **kwargs, not both")
 */
    /*else*/ {
      __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = 0;
      __pyx_t_8 = 127;
      __Pyx_INCREF(__pyx_kp_u_Record);
      __pyx_t_7 += 7;
      __Pyx_GIVEREF(__pyx_kp_u_Record);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Record);
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_data, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_8;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_INCREF(__pyx_kp_u_cannot_be_adapted_to);
      __pyx_t_7 += 22;
      __Pyx_GIVEREF(__pyx_kp_u_cannot_be_adapted_to);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_cannot_be_adapted_to);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Type); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_8;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_2, 4, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(11, 118, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "src/record.pxi":119
 *             else:
 *                 raise ValueError(f"Record {data} cannot be adapted to {self.Type}")
 *             if kwargs:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Records may either be instantiated with a record, or **kwargs, not both")
 *         elif data:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(11, 119, __pyx_L1_error)
    if (unlikely(__pyx_t_1)) {

      /* "src/record.pxi":120
 *                 raise ValueError(f"Record {data} cannot be adapted to {self.Type}")
 *             if kwargs:
 *                 raise ValueError(f"Records may either be instantiated with a record, or **kwargs, not both")             # <<<<<<<<<<<<<<
 *         elif data:
 *             if kwargs:
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(11, 120, __pyx_L1_error)

      /* "src/record.pxi":119
 *             else:
 *                 raise ValueError(f"Record {data} cannot be adapted to {self.Type}")
 *             if kwargs:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Records may either be instantiated with a record, or **kwargs, not both")
 *         elif data:
 */
    }

    /* "src/record.pxi":114
 *             self.data = rec.data
 *             return
 *         if isinstance(data, Record):             # <<<<<<<<<<<<<<
 *             if data.Type.name == self.Type.name and self.Type.options.adapt_record_types:
 *                 data_dict = data._asdict()
 */
    goto __pyx_L4;
  }

  /* "src/record.pxi":121
 *             if kwargs:
 *                 raise ValueError(f"Records may either be instantiated with a record, or **kwargs, not both")
 *         elif data:             # <<<<<<<<<<<<<<
 *             if kwargs:
 *                 raise ValueError(f"Records may either be instantiated with a single dict, or **kwargs, not both")
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_data); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(11, 121, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/record.pxi":122
 *                 raise ValueError(f"Records may either be instantiated with a record, or **kwargs, not both")
 *         elif data:
 *             if kwargs:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Records may either be instantiated with a single dict, or **kwargs, not both")
 *             data_dict = data
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(11, 122, __pyx_L1_error)
    if (unlikely(__pyx_t_1)) {

      /* "src/record.pxi":123
 *         elif data:
 *             if kwargs:
 *                 raise ValueError(f"Records may either be instantiated with a single dict, or **kwargs, not both")             # <<<<<<<<<<<<<<
 *             data_dict = data
 *         else:
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__59, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(11, 123, __pyx_L1_error)

      /* "src/record.pxi":122
 *                 raise ValueError(f"Records may either be instantiated with a record, or **kwargs, not both")
 *         elif data:
 *             if kwargs:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Records may either be instantiated with a single dict, or **kwargs, not both")
 *             data_dict = data
 */
    }

    /* "src/record.pxi":124
 *             if kwargs:
 *                 raise ValueError(f"Records may either be instantiated with a single dict, or **kwargs, not both")
 *             data_dict = data             # <<<<<<<<<<<<<<
 *         else:
 *             data_dict = kwargs
 */
    if (!(likely(PyDict_CheckExact(__pyx_v_data))||((__pyx_v_data) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_data))) __PYX_ERR(11, 124, __pyx_L1_error)
    __pyx_t_2 = __pyx_v_data;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_data_dict = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/record.pxi":121
 *             if kwargs:
 *                 raise ValueError(f"Records may either be instantiated with a record, or **kwargs, not both")
 *         elif data:             # <<<<<<<<<<<<<<
 *             if kwargs:
 *                 raise ValueError(f"Records may either be instantiated with a single dict, or **kwargs, not both")
 */
    goto __pyx_L4;
  }

  /* "src/record.pxi":126
 *             data_dict = data
 *         else:
 *             data_dict = kwargs             # <<<<<<<<<<<<<<
 *         self.data = _record_data_from_dict(self.Type, data_dict)
 * 
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_kwargs);
    __pyx_v_data_dict = __pyx_v_kwargs;
  }
  __pyx_L4:;

  /* "src/record.pxi":127
 *         else:
 *             data_dict = kwargs
 *         self.data = _record_data_from_dict(self.Type, data_dict)             # <<<<<<<<<<<<<<
 * 
 *     def __dir__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Type); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_RecordType))))) __PYX_ERR(11, 127, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__record_data_from_dict(((struct __pyx_obj_5cavro_RecordType *)__pyx_t_2), __pyx_v_data_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->data);
  __Pyx_DECREF(__pyx_v_self->data);
  __pyx_v_self->data = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/record.pxi":107
 *     cdef list data
 * 
 *     def __init__(self, data=None, **kwargs):             # <<<<<<<<<<<<<<
 *         cdef dict data_dict
 *         cdef Record rec
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Record.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data_dict);
  __Pyx_XDECREF((PyObject *)__pyx_v_rec);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":129
 *         self.data = _record_data_from_dict(self.Type, data_dict)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return ['Type'] + [f.name for f in self.Type.fields] + ['_asdict', '__getitem__']
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Record_3__dir__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_6Record_3__dir__ = {"__dir__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Record_3__dir__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_6Record_3__dir__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dir__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__dir__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__dir__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_6Record_2__dir__(((struct __pyx_obj_5cavro_Record *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Record_2__dir__(struct __pyx_obj_5cavro_Record *__pyx_v_self) {
  PyObject *__pyx_8genexpr8__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dir__", 1);

  /* "src/record.pxi":130
 * 
 *     def __dir__(self):
 *         return ['Type'] + [f.name for f in self.Type.fields] + ['_asdict', '__getitem__']             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, name):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Type);
  __Pyx_GIVEREF(__pyx_n_u_Type);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Type)) __PYX_ERR(11, 130, __pyx_L1_error);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 130, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Type); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 130, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_fields); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 130, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_3 = __pyx_t_4; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 130, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(11, 130, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 130, __pyx_L5_error)
            #endif
            if (__pyx_t_5 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(11, 130, __pyx_L5_error)
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 130, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 130, __pyx_L5_error)
            #endif
            if (__pyx_t_5 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(11, 130, __pyx_L5_error)
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 130, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_6(__pyx_t_3);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(11, 130, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr8__pyx_v_f, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr8__pyx_v_f, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 130, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_4))) __PYX_ERR(11, 130, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_f); __pyx_8genexpr8__pyx_v_f = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_f); __pyx_8genexpr8__pyx_v_f = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_u_asdict);
  __Pyx_GIVEREF(__pyx_n_u_asdict);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_u_asdict)) __PYX_ERR(11, 130, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_getitem);
  __Pyx_GIVEREF(__pyx_n_u_getitem);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_u_getitem)) __PYX_ERR(11, 130, __pyx_L1_error);
  __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":129
 *         self.data = _record_data_from_dict(self.Type, data_dict)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return ['Type'] + [f.name for f in self.Type.fields] + ['_asdict', '__getitem__']
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Record.__dir__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr8__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":132
 *         return ['Type'] + [f.name for f in self.Type.fields] + ['_asdict', '__getitem__']
 * 
 *     def __getitem__(self, name):             # <<<<<<<<<<<<<<
 *         cdef dict indexes = self._field_to_index
 *         cdef Py_ssize_t field_index = indexes[name]
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Record_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_name); /*proto*/
static PyObject *__pyx_pw_5cavro_6Record_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_name) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6Record_4__getitem__(((struct __pyx_obj_5cavro_Record *)__pyx_v_self), ((PyObject *)__pyx_v_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Record_4__getitem__(struct __pyx_obj_5cavro_Record *__pyx_v_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_v_indexes = 0;
  Py_ssize_t __pyx_v_field_index;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 1);

  /* "src/record.pxi":133
 * 
 *     def __getitem__(self, name):
 *         cdef dict indexes = self._field_to_index             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t field_index = indexes[name]
 *         return self.data[field_index]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_field_to_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(11, 133, __pyx_L1_error)
  __pyx_v_indexes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":134
 *     def __getitem__(self, name):
 *         cdef dict indexes = self._field_to_index
 *         cdef Py_ssize_t field_index = indexes[name]             # <<<<<<<<<<<<<<
 *         return self.data[field_index]
 * 
 */
  if (unlikely(__pyx_v_indexes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(11, 134, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_indexes, __pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(11, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_field_index = __pyx_t_2;

  /* "src/record.pxi":135
 *         cdef dict indexes = self._field_to_index
 *         cdef Py_ssize_t field_index = indexes[name]
 *         return self.data[field_index]             # <<<<<<<<<<<<<<
 * 
 *     cdef _repr_children(self, remain):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(11, 135, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_self->data, __pyx_v_field_index, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":132
 *         return ['Type'] + [f.name for f in self.Type.fields] + ['_asdict', '__getitem__']
 * 
 *     def __getitem__(self, name):             # <<<<<<<<<<<<<<
 *         cdef dict indexes = self._field_to_index
 *         cdef Py_ssize_t field_index = indexes[name]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Record.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_indexes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_6Record_14_repr_children_2generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/record.pxi":142
 *             return "{...}"
 *         more = False
 *         key_len = sum(len(f.name) + 1 for f in self.Type.fields)             # <<<<<<<<<<<<<<
 *         child_remain = remain - key_len
 *         vals = []
 */

static PyObject *__pyx_pf_5cavro_6Record_14_repr_children_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_7_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_7_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(11, 142, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_6Record_14_repr_children_2generator9, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Record__repr_children_locals_gen, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(11, 142, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.Record._repr_children.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_6Record_14_repr_children_2generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(11, 142, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(11, 142, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 142, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 142, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(11, 142, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 142, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 142, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(11, 142, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 142, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(11, 142, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_f);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_f, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_f, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(11, 142, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyInt_FromSsize_t((__pyx_t_5 + 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(11, 142, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":137
 *         return self.data[field_index]
 * 
 *     cdef _repr_children(self, remain):             # <<<<<<<<<<<<<<
 *         cdef Record rec
 *         if remain < 4:
 */

static PyObject *__pyx_f_5cavro_6Record__repr_children(struct __pyx_obj_5cavro_Record *__pyx_v_self, PyObject *__pyx_v_remain) {
  struct __pyx_obj_5cavro_Record *__pyx_v_rec = 0;
  int __pyx_v_more;
  PyObject *__pyx_v_key_len = NULL;
  PyObject *__pyx_v_child_remain = NULL;
  PyObject *__pyx_v_vals = NULL;
  PyObject *__pyx_v_field = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_repr_val = NULL;
  PyObject *__pyx_gb_5cavro_6Record_14_repr_children_2generator9 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_repr_children", 0);
  __Pyx_INCREF(__pyx_v_remain);

  /* "src/record.pxi":139
 *     cdef _repr_children(self, remain):
 *         cdef Record rec
 *         if remain < 4:             # <<<<<<<<<<<<<<
 *             return "{...}"
 *         more = False
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_remain, __pyx_int_4, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 139, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(11, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/record.pxi":140
 *         cdef Record rec
 *         if remain < 4:
 *             return "{...}"             # <<<<<<<<<<<<<<
 *         more = False
 *         key_len = sum(len(f.name) + 1 for f in self.Type.fields)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_kp_u__60);
    __pyx_r = __pyx_kp_u__60;
    goto __pyx_L0;

    /* "src/record.pxi":139
 *     cdef _repr_children(self, remain):
 *         cdef Record rec
 *         if remain < 4:             # <<<<<<<<<<<<<<
 *             return "{...}"
 *         more = False
 */
  }

  /* "src/record.pxi":141
 *         if remain < 4:
 *             return "{...}"
 *         more = False             # <<<<<<<<<<<<<<
 *         key_len = sum(len(f.name) + 1 for f in self.Type.fields)
 *         child_remain = remain - key_len
 */
  __pyx_v_more = 0;

  /* "src/record.pxi":142
 *             return "{...}"
 *         more = False
 *         key_len = sum(len(f.name) + 1 for f in self.Type.fields)             # <<<<<<<<<<<<<<
 *         child_remain = remain - key_len
 *         vals = []
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Type); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_fields); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_pf_5cavro_6Record_14_repr_children_genexpr(NULL, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_key_len = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "src/record.pxi":143
 *         more = False
 *         key_len = sum(len(f.name) + 1 for f in self.Type.fields)
 *         child_remain = remain - key_len             # <<<<<<<<<<<<<<
 *         vals = []
 *         for field in self.Type.fields:
 */
  __pyx_t_3 = PyNumber_Subtract(__pyx_v_remain, __pyx_v_key_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_child_remain = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "src/record.pxi":144
 *         key_len = sum(len(f.name) + 1 for f in self.Type.fields)
 *         child_remain = remain - key_len
 *         vals = []             # <<<<<<<<<<<<<<
 *         for field in self.Type.fields:
 *             value = self[field.name]
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_vals = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/record.pxi":145
 *         child_remain = remain - key_len
 *         vals = []
 *         for field in self.Type.fields:             # <<<<<<<<<<<<<<
 *             value = self[field.name]
 *             if isinstance(value, Record):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Type); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_fields); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 145, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 145, __pyx_L1_error)
          #endif
          if (__pyx_t_4 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(11, 145, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 145, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 145, __pyx_L1_error)
          #endif
          if (__pyx_t_4 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(11, 145, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 145, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(11, 145, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/record.pxi":146
 *         vals = []
 *         for field in self.Type.fields:
 *             value = self[field.name]             # <<<<<<<<<<<<<<
 *             if isinstance(value, Record):
 *                 rec = value
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(11, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "src/record.pxi":147
 *         for field in self.Type.fields:
 *             value = self[field.name]
 *             if isinstance(value, Record):             # <<<<<<<<<<<<<<
 *                 rec = value
 *                 value = rec._repr_children(child_remain)
 */
    __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_5cavro_Record); 
    if (__pyx_t_2) {

      /* "src/record.pxi":148
 *             value = self[field.name]
 *             if isinstance(value, Record):
 *                 rec = value             # <<<<<<<<<<<<<<
 *                 value = rec._repr_children(child_remain)
 *             else:
 */
      if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5cavro_Record))))) __PYX_ERR(11, 148, __pyx_L1_error)
      __pyx_t_6 = __pyx_v_value;
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_rec, ((struct __pyx_obj_5cavro_Record *)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "src/record.pxi":149
 *             if isinstance(value, Record):
 *                 rec = value
 *                 value = rec._repr_children(child_remain)             # <<<<<<<<<<<<<<
 *             else:
 *                 value = repr(value)
 */
      __pyx_t_6 = ((struct __pyx_vtabstruct_5cavro_Record *)__pyx_v_rec->__pyx_vtab)->_repr_children(__pyx_v_rec, __pyx_v_child_remain); if (unlikely(!__pyx_t_6)) __PYX_ERR(11, 149, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "src/record.pxi":147
 *         for field in self.Type.fields:
 *             value = self[field.name]
 *             if isinstance(value, Record):             # <<<<<<<<<<<<<<
 *                 rec = value
 *                 value = rec._repr_children(child_remain)
 */
      goto __pyx_L6;
    }

    /* "src/record.pxi":151
 *                 value = rec._repr_children(child_remain)
 *             else:
 *                 value = repr(value)             # <<<<<<<<<<<<<<
 *             repr_val = f'{field.name}: {value}'
 *             vals.append(repr_val)
 */
    /*else*/ {
      __pyx_t_6 = PyObject_Repr(__pyx_v_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(11, 151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;
    }
    __pyx_L6:;

    /* "src/record.pxi":152
 *             else:
 *                 value = repr(value)
 *             repr_val = f'{field.name}: {value}'             # <<<<<<<<<<<<<<
 *             vals.append(repr_val)
 *             remain -= len(repr_val) + 1
 */
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(11, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = 0;
    __pyx_t_8 = 127;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_8;
    __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_9);
    __pyx_t_9 = 0;
    __Pyx_INCREF(__pyx_kp_u_);
    __pyx_t_7 += 2;
    __Pyx_GIVEREF(__pyx_kp_u_);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_kp_u_);
    __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_8;
    __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_6, 3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_repr_val, ((PyObject*)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "src/record.pxi":153
 *                 value = repr(value)
 *             repr_val = f'{field.name}: {value}'
 *             vals.append(repr_val)             # <<<<<<<<<<<<<<
 *             remain -= len(repr_val) + 1
 *             if remain < 4:
 */
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_vals, __pyx_v_repr_val); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(11, 153, __pyx_L1_error)

    /* "src/record.pxi":154
 *             repr_val = f'{field.name}: {value}'
 *             vals.append(repr_val)
 *             remain -= len(repr_val) + 1             # <<<<<<<<<<<<<<
 *             if remain < 4:
 *                 more = True
 */
    __pyx_t_7 = __Pyx_PyUnicode_GET_LENGTH(__pyx_v_repr_val); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(11, 154, __pyx_L1_error)
    __pyx_t_9 = PyInt_FromSsize_t((__pyx_t_7 + 1)); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_6 = PyNumber_InPlaceSubtract(__pyx_v_remain, __pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(11, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_remain, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "src/record.pxi":155
 *             vals.append(repr_val)
 *             remain -= len(repr_val) + 1
 *             if remain < 4:             # <<<<<<<<<<<<<<
 *                 more = True
 *                 break
 */
    __pyx_t_6 = PyObject_RichCompare(__pyx_v_remain, __pyx_int_4, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(11, 155, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(11, 155, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_2) {

      /* "src/record.pxi":156
 *             remain -= len(repr_val) + 1
 *             if remain < 4:
 *                 more = True             # <<<<<<<<<<<<<<
 *                 break
 *         return '{' + (" ".join(vals)) + ("..." if more else "") + "}"
 */
      __pyx_v_more = 1;

      /* "src/record.pxi":157
 *             if remain < 4:
 *                 more = True
 *                 break             # <<<<<<<<<<<<<<
 *         return '{' + (" ".join(vals)) + ("..." if more else "") + "}"
 * 
 */
      goto __pyx_L5_break;

      /* "src/record.pxi":155
 *             vals.append(repr_val)
 *             remain -= len(repr_val) + 1
 *             if remain < 4:             # <<<<<<<<<<<<<<
 *                 more = True
 *                 break
 */
    }

    /* "src/record.pxi":145
 *         child_remain = remain - key_len
 *         vals = []
 *         for field in self.Type.fields:             # <<<<<<<<<<<<<<
 *             value = self[field.name]
 *             if isinstance(value, Record):
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  goto __pyx_L8_for_end;
  __pyx_L5_break:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  goto __pyx_L8_for_end;
  __pyx_L8_for_end:;

  /* "src/record.pxi":158
 *                 more = True
 *                 break
 *         return '{' + (" ".join(vals)) + ("..." if more else "") + "}"             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyUnicode_Join(__pyx_kp_u__24, __pyx_v_vals); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyUnicode_Concat(__pyx_kp_u__61, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(11, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_v_more) {
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_3 = __pyx_kp_u__18;
  } else {
    __Pyx_INCREF(__pyx_kp_u__13);
    __pyx_t_3 = __pyx_kp_u__13;
  }
  __pyx_t_9 = __Pyx_PyUnicode_ConcatInPlaceSafe(__pyx_t_6, __pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_9, __pyx_kp_u__62); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":137
 *         return self.data[field_index]
 * 
 *     cdef _repr_children(self, remain):             # <<<<<<<<<<<<<<
 *         cdef Record rec
 *         if remain < 4:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cavro.Record._repr_children", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_rec);
  __Pyx_XDECREF(__pyx_v_key_len);
  __Pyx_XDECREF(__pyx_v_child_remain);
  __Pyx_XDECREF(__pyx_v_vals);
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_repr_val);
  __Pyx_XDECREF(__pyx_gb_5cavro_6Record_14_repr_children_2generator9);
  __Pyx_XDECREF(__pyx_v_remain);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":160
 *         return '{' + (" ".join(vals)) + ("..." if more else "") + "}"
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         child_reprs = {f.name: self[f.name] for f in self.Type.fields}
 *         child_desc = ', '.join([f''])
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Record_7__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_6Record_7__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6Record_6__repr__(((struct __pyx_obj_5cavro_Record *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Record_6__repr__(struct __pyx_obj_5cavro_Record *__pyx_v_self) {
  CYTHON_UNUSED PyObject *__pyx_v_child_reprs = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_child_desc = NULL;
  PyObject *__pyx_9genexpr10__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_UCS4 __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "src/record.pxi":161
 * 
 *     def __repr__(self):
 *         child_reprs = {f.name: self[f.name] for f in self.Type.fields}             # <<<<<<<<<<<<<<
 *         child_desc = ', '.join([f''])
 *         return f'<Record:{self.Type.type} {self._repr_children(70)}>'
 */
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 161, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Type); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 161, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_fields); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 161, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 161, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 161, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 161, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(11, 161, __pyx_L5_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 161, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 161, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(11, 161, __pyx_L5_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 161, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_5(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(11, 161, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr10__pyx_v_f, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr10__pyx_v_f, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 161, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr10__pyx_v_f, __pyx_n_s_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(11, 161, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 161, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_t_3, (PyObject*)__pyx_t_7))) __PYX_ERR(11, 161, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_f); __pyx_9genexpr10__pyx_v_f = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_f); __pyx_9genexpr10__pyx_v_f = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_v_child_reprs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":162
 *     def __repr__(self):
 *         child_reprs = {f.name: self[f.name] for f in self.Type.fields}
 *         child_desc = ', '.join([f''])             # <<<<<<<<<<<<<<
 *         return f'<Record:{self.Type.type} {self._repr_children(70)}>'
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_kp_u__13);
  __Pyx_GIVEREF(__pyx_kp_u__13);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u__13)) __PYX_ERR(11, 162, __pyx_L1_error);
  __pyx_t_2 = PyUnicode_Join(__pyx_kp_u__63, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_child_desc = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/record.pxi":163
 *         child_reprs = {f.name: self[f.name] for f in self.Type.fields}
 *         child_desc = ', '.join([f''])
 *         return f'<Record:{self.Type.type} {self._repr_children(70)}>'             # <<<<<<<<<<<<<<
 * 
 *     def _asdict(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = 0;
  __pyx_t_8 = 127;
  __Pyx_INCREF(__pyx_kp_u_Record_2);
  __pyx_t_4 += 8;
  __Pyx_GIVEREF(__pyx_kp_u_Record_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Record_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Type); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_7, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_8;
  __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_INCREF(__pyx_kp_u__24);
  __pyx_t_4 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__24);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u__24);
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_Record *)__pyx_v_self->__pyx_vtab)->_repr_children(__pyx_v_self, __pyx_int_70); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_8;
  __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_7);
  __pyx_t_7 = 0;
  __Pyx_INCREF(__pyx_kp_u__64);
  __pyx_t_4 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__64);
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_kp_u__64);
  __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_2, 5, __pyx_t_4, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":160
 *         return '{' + (" ".join(vals)) + ("..." if more else "") + "}"
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         child_reprs = {f.name: self[f.name] for f in self.Type.fields}
 *         child_desc = ', '.join([f''])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.Record.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_child_reprs);
  __Pyx_XDECREF(__pyx_v_child_desc);
  __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":165
 *         return f'<Record:{self.Type.type} {self._repr_children(70)}>'
 * 
 *     def _asdict(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the record as a dict, with field names as keys.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Record_9_asdict(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_6Record_8_asdict, "\n        Returns the record as a dict, with field names as keys.\n        ");
static PyMethodDef __pyx_mdef_5cavro_6Record_9_asdict = {"_asdict", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Record_9_asdict, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Record_8_asdict};
static PyObject *__pyx_pw_5cavro_6Record_9_asdict(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_asdict (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("_asdict", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "_asdict", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_6Record_8_asdict(((struct __pyx_obj_5cavro_Record *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Record_8_asdict(struct __pyx_obj_5cavro_Record *__pyx_v_self) {
  PyObject *__pyx_v_items = 0;
  struct __pyx_obj_5cavro_RecordField *__pyx_v_field = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_asdict", 1);

  /* "src/record.pxi":169
 *         Returns the record as a dict, with field names as keys.
 *         """
 *         cdef dict items = {}             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         cdef Py_ssize_t i = 0
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_items = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":171
 *         cdef dict items = {}
 *         cdef RecordField field
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         for field in self.Type.fields:
 *             data = self.data[i]
 */
  __pyx_v_i = 0;

  /* "src/record.pxi":172
 *         cdef RecordField field
 *         cdef Py_ssize_t i = 0
 *         for field in self.Type.fields:             # <<<<<<<<<<<<<<
 *             data = self.data[i]
 *             if isinstance(data, Record):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Type); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_fields); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 172, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 172, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(11, 172, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 172, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 172, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(11, 172, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 172, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(11, 172, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 172, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "src/record.pxi":173
 *         cdef Py_ssize_t i = 0
 *         for field in self.Type.fields:
 *             data = self.data[i]             # <<<<<<<<<<<<<<
 *             if isinstance(data, Record):
 *                 data = data._asdict()
 */
    if (unlikely(__pyx_v_self->data == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(11, 173, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_self->data, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_data, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/record.pxi":174
 *         for field in self.Type.fields:
 *             data = self.data[i]
 *             if isinstance(data, Record):             # <<<<<<<<<<<<<<
 *                 data = data._asdict()
 *             items[field.name] = data
 */
    __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_data, __pyx_ptype_5cavro_Record); 
    if (__pyx_t_5) {

      /* "src/record.pxi":175
 *             data = self.data[i]
 *             if isinstance(data, Record):
 *                 data = data._asdict()             # <<<<<<<<<<<<<<
 *             items[field.name] = data
 *             i += 1
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_asdict); if (unlikely(!__pyx_t_6)) __PYX_ERR(11, 175, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "src/record.pxi":174
 *         for field in self.Type.fields:
 *             data = self.data[i]
 *             if isinstance(data, Record):             # <<<<<<<<<<<<<<
 *                 data = data._asdict()
 *             items[field.name] = data
 */
    }

    /* "src/record.pxi":176
 *             if isinstance(data, Record):
 *                 data = data._asdict()
 *             items[field.name] = data             # <<<<<<<<<<<<<<
 *             i += 1
 *         return items
 */
    if (unlikely((PyDict_SetItem(__pyx_v_items, __pyx_v_field->name, __pyx_v_data) < 0))) __PYX_ERR(11, 176, __pyx_L1_error)

    /* "src/record.pxi":177
 *                 data = data._asdict()
 *             items[field.name] = data
 *             i += 1             # <<<<<<<<<<<<<<
 *         return items
 * 
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "src/record.pxi":172
 *         cdef RecordField field
 *         cdef Py_ssize_t i = 0
 *         for field in self.Type.fields:             # <<<<<<<<<<<<<<
 *             data = self.data[i]
 *             if isinstance(data, Record):
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/record.pxi":178
 *             items[field.name] = data
 *             i += 1
 *         return items             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_items);
  __pyx_r = __pyx_v_items;
  goto __pyx_L0;

  /* "src/record.pxi":165
 *         return f'<Record:{self.Type.type} {self._repr_children(70)}>'
 * 
 *     def _asdict(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the record as a dict, with field names as keys.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.Record._asdict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":180
 *         return items
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if not isinstance(other, self.Type.record):
 *             return False
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Record_11__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_5cavro_6Record_11__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6Record_10__eq__(((struct __pyx_obj_5cavro_Record *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Record_10__eq__(struct __pyx_obj_5cavro_Record *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_5cavro_Record *__pyx_v_other_rec = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 1);

  /* "src/record.pxi":181
 * 
 *     def __eq__(self, other):
 *         if not isinstance(other, self.Type.record):             # <<<<<<<<<<<<<<
 *             return False
 *         cdef Record other_rec = other
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Type); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_record); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_other, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(11, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (!__pyx_t_3);
  if (__pyx_t_4) {

    /* "src/record.pxi":182
 *     def __eq__(self, other):
 *         if not isinstance(other, self.Type.record):
 *             return False             # <<<<<<<<<<<<<<
 *         cdef Record other_rec = other
 *         return self.data == other_rec.data
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "src/record.pxi":181
 * 
 *     def __eq__(self, other):
 *         if not isinstance(other, self.Type.record):             # <<<<<<<<<<<<<<
 *             return False
 *         cdef Record other_rec = other
 */
  }

  /* "src/record.pxi":183
 *         if not isinstance(other, self.Type.record):
 *             return False
 *         cdef Record other_rec = other             # <<<<<<<<<<<<<<
 *         return self.data == other_rec.data
 * 
 */
  if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_5cavro_Record))))) __PYX_ERR(11, 183, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_other;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_other_rec = ((struct __pyx_obj_5cavro_Record *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/record.pxi":184
 *             return False
 *         cdef Record other_rec = other
 *         return self.data == other_rec.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_self->data, __pyx_v_other_rec->data, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 184, __pyx_L1_error)
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":180
 *         return items
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if not isinstance(other, self.Type.record):
 *             return False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.Record.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_other_rec);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Record_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_6Record_13__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Record_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_6Record_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_6Record_12__reduce_cython__(((struct __pyx_obj_5cavro_Record *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Record_12__reduce_cython__(struct __pyx_obj_5cavro_Record *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.data,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->data);
  __Pyx_GIVEREF(__pyx_v_self->data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->data)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.data,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.data is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.data,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.data is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Record, (type(self), 0x3a6eb07, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->data != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Record, (type(self), 0x3a6eb07, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.data is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Record, (type(self), 0x3a6eb07, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Record, (type(self), 0x3a6eb07, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Record); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_61270791);
    __Pyx_GIVEREF(__pyx_int_61270791);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_61270791)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.data is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Record, (type(self), 0x3a6eb07, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Record, (type(self), 0x3a6eb07, None), state
 *     else:
 *         return __pyx_unpickle_Record, (type(self), 0x3a6eb07, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Record__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Record); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_61270791);
    __Pyx_GIVEREF(__pyx_int_61270791);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_61270791)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Record.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Record, (type(self), 0x3a6eb07, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Record__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Record_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_6Record_15__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Record_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_6Record_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Record.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6Record_14__setstate_cython__(((struct __pyx_obj_5cavro_Record *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Record_14__setstate_cython__(struct __pyx_obj_5cavro_Record *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Record, (type(self), 0x3a6eb07, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Record__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_Record__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Record, (type(self), 0x3a6eb07, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Record__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Record.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":203
 *     cdef readonly frozenset aliases
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         cdef AvroType sub_type
 *         self.name = source['name']
 */

/* Python wrapper */
static int __pyx_pw_5cavro_11RecordField_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_11RecordField_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_namespace = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_schema,&__pyx_n_s_source,&__pyx_n_s_namespace,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 203, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 203, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(11, 203, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_namespace)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 203, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(11, 203, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(11, 203, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_schema = values[0];
    __pyx_v_source = values[1];
    __pyx_v_namespace = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(11, 203, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.RecordField.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_11RecordField___init__(((struct __pyx_obj_5cavro_RecordField *)__pyx_v_self), __pyx_v_schema, __pyx_v_source, __pyx_v_namespace);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_11RecordField___init__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace) {
  PyObject *__pyx_v_alias_val = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/record.pxi":205
 *     def __init__(self, schema, source, namespace):
 *         cdef AvroType sub_type
 *         self.name = source['name']             # <<<<<<<<<<<<<<
 *         self.writer_name = self.name
 *         self.doc = source.get('doc', '')
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_source, __pyx_n_u_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(11, 205, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":206
 *         cdef AvroType sub_type
 *         self.name = source['name']
 *         self.writer_name = self.name             # <<<<<<<<<<<<<<
 *         self.doc = source.get('doc', '')
 *         self.type = AvroType.for_source(schema, source['type'], namespace)
 */
  __pyx_t_1 = __pyx_v_self->name;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->writer_name);
  __Pyx_DECREF(__pyx_v_self->writer_name);
  __pyx_v_self->writer_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":207
 *         self.name = source['name']
 *         self.writer_name = self.name
 *         self.doc = source.get('doc', '')             # <<<<<<<<<<<<<<
 *         self.type = AvroType.for_source(schema, source['type'], namespace)
 *         self.default_value = self.type.resolve_default_value(source.get('default', NO_DEFAULT), self.name)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(11, 207, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->doc);
  __Pyx_DECREF(__pyx_v_self->doc);
  __pyx_v_self->doc = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/record.pxi":208
 *         self.writer_name = self.name
 *         self.doc = source.get('doc', '')
 *         self.type = AvroType.for_source(schema, source['type'], namespace)             # <<<<<<<<<<<<<<
 *         self.default_value = self.type.resolve_default_value(source.get('default', NO_DEFAULT), self.name)
 *         self.order = Order(source.get('order', 'ascending'))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_AvroType), __pyx_n_s_for_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_source, __pyx_n_u_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_v_schema, __pyx_t_3, __pyx_v_namespace};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(11, 208, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->type);
  __Pyx_DECREF((PyObject *)__pyx_v_self->type);
  __pyx_v_self->type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/record.pxi":209
 *         self.doc = source.get('doc', '')
 *         self.type = AvroType.for_source(schema, source['type'], namespace)
 *         self.default_value = self.type.resolve_default_value(source.get('default', NO_DEFAULT), self.name)             # <<<<<<<<<<<<<<
 *         self.order = Order(source.get('order', 'ascending'))
 *         alias_val = source.get('aliases', [])
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_n_u_default, __pyx_t_3};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_1 = __pyx_v_self->name;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->resolve_default_value(__pyx_v_self->type, __pyx_t_2, ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->default_value);
  __Pyx_DECREF(__pyx_v_self->default_value);
  __pyx_v_self->default_value = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "src/record.pxi":210
 *         self.type = AvroType.for_source(schema, source['type'], namespace)
 *         self.default_value = self.type.resolve_default_value(source.get('default', NO_DEFAULT), self.name)
 *         self.order = Order(source.get('order', 'ascending'))             # <<<<<<<<<<<<<<
 *         alias_val = source.get('aliases', [])
 *         if not isinstance(alias_val, (list, tuple, set)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Order); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__65, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->order);
  __Pyx_DECREF(__pyx_v_self->order);
  __pyx_v_self->order = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "src/record.pxi":211
 *         self.default_value = self.type.resolve_default_value(source.get('default', NO_DEFAULT), self.name)
 *         self.order = Order(source.get('order', 'ascending'))
 *         alias_val = source.get('aliases', [])             # <<<<<<<<<<<<<<
 *         if not isinstance(alias_val, (list, tuple, set)):
 *             raise ValueError(f"Aliases must be a list/tuple/set, got: {repr(alias_val)}")
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_n_u_aliases, __pyx_t_4};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_v_alias_val = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "src/record.pxi":212
 *         self.order = Order(source.get('order', 'ascending'))
 *         alias_val = source.get('aliases', [])
 *         if not isinstance(alias_val, (list, tuple, set)):             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Aliases must be a list/tuple/set, got: {repr(alias_val)}")
 *         self.aliases = frozenset(alias_val)
 */
  __pyx_t_7 = PyList_Check(__pyx_v_alias_val); 
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = PyTuple_Check(__pyx_v_alias_val); 
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = PySet_Check(__pyx_v_alias_val); 
  __pyx_t_6 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  __pyx_t_7 = (!__pyx_t_6);
  if (unlikely(__pyx_t_7)) {

    /* "src/record.pxi":213
 *         alias_val = source.get('aliases', [])
 *         if not isinstance(alias_val, (list, tuple, set)):
 *             raise ValueError(f"Aliases must be a list/tuple/set, got: {repr(alias_val)}")             # <<<<<<<<<<<<<<
 *         self.aliases = frozenset(alias_val)
 * 
 */
    __pyx_t_3 = PyObject_Repr(__pyx_v_alias_val); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Aliases_must_be_a_list_tuple_set, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(11, 213, __pyx_L1_error)

    /* "src/record.pxi":212
 *         self.order = Order(source.get('order', 'ascending'))
 *         alias_val = source.get('aliases', [])
 *         if not isinstance(alias_val, (list, tuple, set)):             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Aliases must be a list/tuple/set, got: {repr(alias_val)}")
 *         self.aliases = frozenset(alias_val)
 */
  }

  /* "src/record.pxi":214
 *         if not isinstance(alias_val, (list, tuple, set)):
 *             raise ValueError(f"Aliases must be a list/tuple/set, got: {repr(alias_val)}")
 *         self.aliases = frozenset(alias_val)             # <<<<<<<<<<<<<<
 * 
 *     cdef for_reader(self, str reader_name, AvroType reader_type):
 */
  __pyx_t_1 = __Pyx_PyFrozenSet_New(__pyx_v_alias_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->aliases);
  __Pyx_DECREF(__pyx_v_self->aliases);
  __pyx_v_self->aliases = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":203
 *     cdef readonly frozenset aliases
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         cdef AvroType sub_type
 *         self.name = source['name']
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.RecordField.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_alias_val);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":216
 *         self.aliases = frozenset(alias_val)
 * 
 *     cdef for_reader(self, str reader_name, AvroType reader_type):             # <<<<<<<<<<<<<<
 *         cdef RecordField cloned
 *         cloned = RecordField.__new__(RecordField)
 */

static PyObject *__pyx_f_5cavro_11RecordField_for_reader(struct __pyx_obj_5cavro_RecordField *__pyx_v_self, PyObject *__pyx_v_reader_name, struct __pyx_obj_5cavro_AvroType *__pyx_v_reader_type) {
  struct __pyx_obj_5cavro_RecordField *__pyx_v_cloned = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("for_reader", 1);

  /* "src/record.pxi":218
 *     cdef for_reader(self, str reader_name, AvroType reader_type):
 *         cdef RecordField cloned
 *         cloned = RecordField.__new__(RecordField)             # <<<<<<<<<<<<<<
 *         cloned.name = reader_name
 *         cloned.writer_name = self.writer_name
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_5cavro_RecordField(((PyTypeObject *)__pyx_ptype_5cavro_RecordField), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 218, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_cloned = ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":219
 *         cdef RecordField cloned
 *         cloned = RecordField.__new__(RecordField)
 *         cloned.name = reader_name             # <<<<<<<<<<<<<<
 *         cloned.writer_name = self.writer_name
 *         if reader_type is not None:
 */
  __Pyx_INCREF(__pyx_v_reader_name);
  __Pyx_GIVEREF(__pyx_v_reader_name);
  __Pyx_GOTREF(__pyx_v_cloned->name);
  __Pyx_DECREF(__pyx_v_cloned->name);
  __pyx_v_cloned->name = __pyx_v_reader_name;

  /* "src/record.pxi":220
 *         cloned = RecordField.__new__(RecordField)
 *         cloned.name = reader_name
 *         cloned.writer_name = self.writer_name             # <<<<<<<<<<<<<<
 *         if reader_type is not None:
 *             cloned.type = reader_type.for_writer(self.type)
 */
  __pyx_t_1 = __pyx_v_self->writer_name;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_cloned->writer_name);
  __Pyx_DECREF(__pyx_v_cloned->writer_name);
  __pyx_v_cloned->writer_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":221
 *         cloned.name = reader_name
 *         cloned.writer_name = self.writer_name
 *         if reader_type is not None:             # <<<<<<<<<<<<<<
 *             cloned.type = reader_type.for_writer(self.type)
 *         else:
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_reader_type) != Py_None);
  if (__pyx_t_2) {

    /* "src/record.pxi":222
 *         cloned.writer_name = self.writer_name
 *         if reader_type is not None:
 *             cloned.type = reader_type.for_writer(self.type)             # <<<<<<<<<<<<<<
 *         else:
 *             cloned.type = self.type
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_self->type);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_reader_type->__pyx_vtab)->for_writer(__pyx_v_reader_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1), NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF((PyObject *)__pyx_v_cloned->type);
    __Pyx_DECREF((PyObject *)__pyx_v_cloned->type);
    __pyx_v_cloned->type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/record.pxi":221
 *         cloned.name = reader_name
 *         cloned.writer_name = self.writer_name
 *         if reader_type is not None:             # <<<<<<<<<<<<<<
 *             cloned.type = reader_type.for_writer(self.type)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "src/record.pxi":224
 *             cloned.type = reader_type.for_writer(self.type)
 *         else:
 *             cloned.type = self.type             # <<<<<<<<<<<<<<
 *         cloned.default_value = NO_DEFAULT # If we're reading, then no default
 *         return cloned
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_self->type);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF((PyObject *)__pyx_v_cloned->type);
    __Pyx_DECREF((PyObject *)__pyx_v_cloned->type);
    __pyx_v_cloned->type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "src/record.pxi":225
 *         else:
 *             cloned.type = self.type
 *         cloned.default_value = NO_DEFAULT # If we're reading, then no default             # <<<<<<<<<<<<<<
 *         return cloned
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_cloned->default_value);
  __Pyx_DECREF(__pyx_v_cloned->default_value);
  __pyx_v_cloned->default_value = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "src/record.pxi":226
 *             cloned.type = self.type
 *         cloned.default_value = NO_DEFAULT # If we're reading, then no default
 *         return cloned             # <<<<<<<<<<<<<<
 * 
 *     cdef placeholder(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_cloned);
  __pyx_r = ((PyObject *)__pyx_v_cloned);
  goto __pyx_L0;

  /* "src/record.pxi":216
 *         self.aliases = frozenset(alias_val)
 * 
 *     cdef for_reader(self, str reader_name, AvroType reader_type):             # <<<<<<<<<<<<<<
 *         cdef RecordField cloned
 *         cloned = RecordField.__new__(RecordField)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.RecordField.for_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cloned);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":228
 *         return cloned
 * 
 *     cdef placeholder(self):             # <<<<<<<<<<<<<<
 *         cdef RecordField rec = RecordField.__new__(RecordField)
 *         rec.name = self.name
 */

static PyObject *__pyx_f_5cavro_11RecordField_placeholder(struct __pyx_obj_5cavro_RecordField *__pyx_v_self) {
  struct __pyx_obj_5cavro_RecordField *__pyx_v_rec = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("placeholder", 1);

  /* "src/record.pxi":229
 * 
 *     cdef placeholder(self):
 *         cdef RecordField rec = RecordField.__new__(RecordField)             # <<<<<<<<<<<<<<
 *         rec.name = self.name
 *         rec.doc = ''
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_5cavro_RecordField(((PyTypeObject *)__pyx_ptype_5cavro_RecordField), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 229, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_rec = ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":230
 *     cdef placeholder(self):
 *         cdef RecordField rec = RecordField.__new__(RecordField)
 *         rec.name = self.name             # <<<<<<<<<<<<<<
 *         rec.doc = ''
 *         rec.type = _PlaceholderType(self.type.options, self.default_value)
 */
  __pyx_t_1 = __pyx_v_self->name;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_rec->name);
  __Pyx_DECREF(__pyx_v_rec->name);
  __pyx_v_rec->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":231
 *         cdef RecordField rec = RecordField.__new__(RecordField)
 *         rec.name = self.name
 *         rec.doc = ''             # <<<<<<<<<<<<<<
 *         rec.type = _PlaceholderType(self.type.options, self.default_value)
 *         rec.default_value = self.default_value
 */
  __Pyx_INCREF(__pyx_kp_u__13);
  __Pyx_GIVEREF(__pyx_kp_u__13);
  __Pyx_GOTREF(__pyx_v_rec->doc);
  __Pyx_DECREF(__pyx_v_rec->doc);
  __pyx_v_rec->doc = __pyx_kp_u__13;

  /* "src/record.pxi":232
 *         rec.name = self.name
 *         rec.doc = ''
 *         rec.type = _PlaceholderType(self.type.options, self.default_value)             # <<<<<<<<<<<<<<
 *         rec.default_value = self.default_value
 *         rec.order = Order.ASC
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->type->options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->type->options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->type->options))) __PYX_ERR(11, 232, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->default_value);
  __Pyx_GIVEREF(__pyx_v_self->default_value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->default_value)) __PYX_ERR(11, 232, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro__PlaceholderType), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_rec->type);
  __Pyx_DECREF((PyObject *)__pyx_v_rec->type);
  __pyx_v_rec->type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/record.pxi":233
 *         rec.doc = ''
 *         rec.type = _PlaceholderType(self.type.options, self.default_value)
 *         rec.default_value = self.default_value             # <<<<<<<<<<<<<<
 *         rec.order = Order.ASC
 *         rec.aliases = frozenset()
 */
  __pyx_t_2 = __pyx_v_self->default_value;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_rec->default_value);
  __Pyx_DECREF(__pyx_v_rec->default_value);
  __pyx_v_rec->default_value = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "src/record.pxi":234
 *         rec.type = _PlaceholderType(self.type.options, self.default_value)
 *         rec.default_value = self.default_value
 *         rec.order = Order.ASC             # <<<<<<<<<<<<<<
 *         rec.aliases = frozenset()
 *         return rec
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Order); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ASC); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_rec->order);
  __Pyx_DECREF(__pyx_v_rec->order);
  __pyx_v_rec->order = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/record.pxi":235
 *         rec.default_value = self.default_value
 *         rec.order = Order.ASC
 *         rec.aliases = frozenset()             # <<<<<<<<<<<<<<
 *         return rec
 * 
 */
  __pyx_t_1 = __Pyx_PyFrozenSet_New(((PyObject *)NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_rec->aliases);
  __Pyx_DECREF(__pyx_v_rec->aliases);
  __pyx_v_rec->aliases = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":236
 *         rec.order = Order.ASC
 *         rec.aliases = frozenset()
 *         return rec             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_rec);
  __pyx_r = ((PyObject *)__pyx_v_rec);
  goto __pyx_L0;

  /* "src/record.pxi":228
 *         return cloned
 * 
 *     cdef placeholder(self):             # <<<<<<<<<<<<<<
 *         cdef RecordField rec = RecordField.__new__(RecordField)
 *         rec.name = self.name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.RecordField.placeholder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_rec);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":238
 *         return rec
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return dict_to_canonical({
 *             'name': self.name,
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_11RecordField_canonical_form(struct __pyx_obj_5cavro_RecordField *__pyx_v_self, PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/record.pxi":239
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return dict_to_canonical({             # <<<<<<<<<<<<<<
 *             'name': self.name,
 *             'type': self.type.canonical_form(created),
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);

  /* "src/record.pxi":240
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return dict_to_canonical({
 *             'name': self.name,             # <<<<<<<<<<<<<<
 *             'type': self.type.canonical_form(created),
 *         })
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_name, __pyx_v_self->name) < 0) __PYX_ERR(11, 240, __pyx_L1_error)

  /* "src/record.pxi":241
 *         return dict_to_canonical({
 *             'name': self.name,
 *             'type': self.type.canonical_form(created),             # <<<<<<<<<<<<<<
 *         })
 * 
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->canonical_form(__pyx_v_self->type, __pyx_v_created)); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_type, __pyx_t_2) < 0) __PYX_ERR(11, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/record.pxi":239
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return dict_to_canonical({             # <<<<<<<<<<<<<<
 *             'name': self.name,
 *             'type': self.type.canonical_form(created),
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_5cavro_dict_to_canonical(__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":238
 *         return rec
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return dict_to_canonical({
 *             'name': self.name,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.RecordField.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":244
 *         })
 * 
 *     def get_schema(self, created):             # <<<<<<<<<<<<<<
 *         schema = {
 *             'name': self.name,
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11RecordField_3get_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_11RecordField_3get_schema = {"get_schema", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11RecordField_3get_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_11RecordField_3get_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_schema (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 244, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_schema") < 0)) __PYX_ERR(11, 244, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_schema", 1, 1, 1, __pyx_nargs); __PYX_ERR(11, 244, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.RecordField.get_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_11RecordField_2get_schema(((struct __pyx_obj_5cavro_RecordField *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11RecordField_2get_schema(struct __pyx_obj_5cavro_RecordField *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_v_schema = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_get_schema __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_schema", 1);

  /* "src/record.pxi":246
 *     def get_schema(self, created):
 *         schema = {
 *             'name': self.name,             # <<<<<<<<<<<<<<
 *             'type': self.type.get_schema(created),
 *         }
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_name, __pyx_v_self->name) < 0) __PYX_ERR(11, 246, __pyx_L1_error)

  /* "src/record.pxi":247
 *         schema = {
 *             'name': self.name,
 *             'type': self.type.get_schema(created),             # <<<<<<<<<<<<<<
 *         }
 *         if self.doc:
 */
  __pyx_t_3.__pyx_n = 1;
  __pyx_t_3.created = __pyx_v_created;
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->get_schema(__pyx_v_self->type, 0, &__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_type, __pyx_t_2) < 0) __PYX_ERR(11, 246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_schema = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":249
 *             'type': self.type.get_schema(created),
 *         }
 *         if self.doc:             # <<<<<<<<<<<<<<
 *             schema['doc'] = self.doc
 *         if self.aliases:
 */
  __pyx_t_4 = (__pyx_v_self->doc != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_v_self->doc) != 0);
  if (__pyx_t_4) {

    /* "src/record.pxi":250
 *         }
 *         if self.doc:
 *             schema['doc'] = self.doc             # <<<<<<<<<<<<<<
 *         if self.aliases:
 *             schema['aliases'] = list(self.aliases)
 */
    __pyx_t_1 = __pyx_v_self->doc;
    __Pyx_INCREF(__pyx_t_1);
    if (unlikely((PyDict_SetItem(__pyx_v_schema, __pyx_n_u_doc, __pyx_t_1) < 0))) __PYX_ERR(11, 250, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/record.pxi":249
 *             'type': self.type.get_schema(created),
 *         }
 *         if self.doc:             # <<<<<<<<<<<<<<
 *             schema['doc'] = self.doc
 *         if self.aliases:
 */
  }

  /* "src/record.pxi":251
 *         if self.doc:
 *             schema['doc'] = self.doc
 *         if self.aliases:             # <<<<<<<<<<<<<<
 *             schema['aliases'] = list(self.aliases)
 *         if self.order != Order.ASC:
 */
  __pyx_t_4 = (__pyx_v_self->aliases != Py_None)&&(PySet_GET_SIZE(__pyx_v_self->aliases) != 0);
  if (__pyx_t_4) {

    /* "src/record.pxi":252
 *             schema['doc'] = self.doc
 *         if self.aliases:
 *             schema['aliases'] = list(self.aliases)             # <<<<<<<<<<<<<<
 *         if self.order != Order.ASC:
 *             schema['order'] = self.order.value
 */
    __pyx_t_1 = PySequence_List(__pyx_v_self->aliases); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely((PyDict_SetItem(__pyx_v_schema, __pyx_n_u_aliases, __pyx_t_1) < 0))) __PYX_ERR(11, 252, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/record.pxi":251
 *         if self.doc:
 *             schema['doc'] = self.doc
 *         if self.aliases:             # <<<<<<<<<<<<<<
 *             schema['aliases'] = list(self.aliases)
 *         if self.order != Order.ASC:
 */
  }

  /* "src/record.pxi":253
 *         if self.aliases:
 *             schema['aliases'] = list(self.aliases)
 *         if self.order != Order.ASC:             # <<<<<<<<<<<<<<
 *             schema['order'] = self.order.value
 *         if self.default_value not in {NO_DEFAULT, MISSING_VALUE}:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Order); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ASC); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_self->order, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(11, 253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "src/record.pxi":254
 *             schema['aliases'] = list(self.aliases)
 *         if self.order != Order.ASC:
 *             schema['order'] = self.order.value             # <<<<<<<<<<<<<<
 *         if self.default_value not in {NO_DEFAULT, MISSING_VALUE}:
 *             try:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->order, __pyx_n_s_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely((PyDict_SetItem(__pyx_v_schema, __pyx_n_u_order, __pyx_t_1) < 0))) __PYX_ERR(11, 254, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/record.pxi":253
 *         if self.aliases:
 *             schema['aliases'] = list(self.aliases)
 *         if self.order != Order.ASC:             # <<<<<<<<<<<<<<
 *             schema['order'] = self.order.value
 *         if self.default_value not in {NO_DEFAULT, MISSING_VALUE}:
 */
  }

  /* "src/record.pxi":255
 *         if self.order != Order.ASC:
 *             schema['order'] = self.order.value
 *         if self.default_value not in {NO_DEFAULT, MISSING_VALUE}:             # <<<<<<<<<<<<<<
 *             try:
 *                 schema['default'] = self.type.json_format(self.default_value)
 */
  __Pyx_INCREF(__pyx_v_self->default_value);
  __pyx_t_1 = __pyx_v_self->default_value;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(11, 255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L7_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(11, 255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_t_6;
  __pyx_L7_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __pyx_t_4;
  if (__pyx_t_6) {

    /* "src/record.pxi":256
 *             schema['order'] = self.order.value
 *         if self.default_value not in {NO_DEFAULT, MISSING_VALUE}:
 *             try:             # <<<<<<<<<<<<<<
 *                 schema['default'] = self.type.json_format(self.default_value)
 *             except InvalidValue:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "src/record.pxi":257
 *         if self.default_value not in {NO_DEFAULT, MISSING_VALUE}:
 *             try:
 *                 schema['default'] = self.type.json_format(self.default_value)             # <<<<<<<<<<<<<<
 *             except InvalidValue:
 *                 if self.type.options.allow_invalid_default_values:
 */
        __pyx_t_1 = __pyx_v_self->default_value;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->json_format(__pyx_v_self->type, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 257, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely((PyDict_SetItem(__pyx_v_schema, __pyx_n_u_default, __pyx_t_2) < 0))) __PYX_ERR(11, 257, __pyx_L9_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "src/record.pxi":256
 *             schema['order'] = self.order.value
 *         if self.default_value not in {NO_DEFAULT, MISSING_VALUE}:
 *             try:             # <<<<<<<<<<<<<<
 *                 schema['default'] = self.type.json_format(self.default_value)
 *             except InvalidValue:
 */
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L14_try_end;
      __pyx_L9_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "src/record.pxi":258
 *             try:
 *                 schema['default'] = self.type.json_format(self.default_value)
 *             except InvalidValue:             # <<<<<<<<<<<<<<
 *                 if self.type.options.allow_invalid_default_values:
 *                     schema['default'] = self.default_value
 */
      __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_1, &__pyx_t_5);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_10)) __PYX_ERR(11, 258, __pyx_L11_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_10);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_ErrRestore(__pyx_t_2, __pyx_t_1, __pyx_t_5);
      __pyx_t_2 = 0; __pyx_t_1 = 0; __pyx_t_5 = 0;
      if (__pyx_t_11) {
        __Pyx_AddTraceback("cavro.RecordField.get_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(11, 258, __pyx_L11_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_2);

        /* "src/record.pxi":259
 *                 schema['default'] = self.type.json_format(self.default_value)
 *             except InvalidValue:
 *                 if self.type.options.allow_invalid_default_values:             # <<<<<<<<<<<<<<
 *                     schema['default'] = self.default_value
 *                 else:
 */
        if (likely(__pyx_v_self->type->options->allow_invalid_default_values)) {

          /* "src/record.pxi":260
 *             except InvalidValue:
 *                 if self.type.options.allow_invalid_default_values:
 *                     schema['default'] = self.default_value             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise
 */
          __pyx_t_10 = __pyx_v_self->default_value;
          __Pyx_INCREF(__pyx_t_10);
          if (unlikely((PyDict_SetItem(__pyx_v_schema, __pyx_n_u_default, __pyx_t_10) < 0))) __PYX_ERR(11, 260, __pyx_L11_except_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "src/record.pxi":259
 *                 schema['default'] = self.type.json_format(self.default_value)
 *             except InvalidValue:
 *                 if self.type.options.allow_invalid_default_values:             # <<<<<<<<<<<<<<
 *                     schema['default'] = self.default_value
 *                 else:
 */
          goto __pyx_L17;
        }

        /* "src/record.pxi":262
 *                     schema['default'] = self.default_value
 *                 else:
 *                     raise             # <<<<<<<<<<<<<<
 *         return schema
 * 
 */
        /*else*/ {
          __Pyx_GIVEREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_1);
          __Pyx_XGIVEREF(__pyx_t_2);
          __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_1, __pyx_t_2);
          __pyx_t_5 = 0; __pyx_t_1 = 0; __pyx_t_2 = 0; 
          __PYX_ERR(11, 262, __pyx_L11_except_error)
        }
        __pyx_L17:;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L10_exception_handled;
      }
      goto __pyx_L11_except_error;

      /* "src/record.pxi":256
 *             schema['order'] = self.order.value
 *         if self.default_value not in {NO_DEFAULT, MISSING_VALUE}:
 *             try:             # <<<<<<<<<<<<<<
 *                 schema['default'] = self.type.json_format(self.default_value)
 *             except InvalidValue:
 */
      __pyx_L11_except_error:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L10_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_L14_try_end:;
    }

    /* "src/record.pxi":255
 *         if self.order != Order.ASC:
 *             schema['order'] = self.order.value
 *         if self.default_value not in {NO_DEFAULT, MISSING_VALUE}:             # <<<<<<<<<<<<<<
 *             try:
 *                 schema['default'] = self.type.json_format(self.default_value)
 */
  }

  /* "src/record.pxi":263
 *                 else:
 *                     raise
 *         return schema             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_schema);
  __pyx_r = __pyx_v_schema;
  goto __pyx_L0;

  /* "src/record.pxi":244
 *         })
 * 
 *     def get_schema(self, created):             # <<<<<<<<<<<<<<
 *         schema = {
 *             'name': self.name,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("cavro.RecordField.get_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_schema);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":195
 *     """
 * 
 *     cdef readonly str name             # <<<<<<<<<<<<<<
 *     cdef readonly str writer_name
 *     cdef readonly str doc
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11RecordField_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11RecordField_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11RecordField_4name___get__(((struct __pyx_obj_5cavro_RecordField *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11RecordField_4name___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":196
 * 
 *     cdef readonly str name
 *     cdef readonly str writer_name             # <<<<<<<<<<<<<<
 *     cdef readonly str doc
 *     cdef readonly AvroType type
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11RecordField_11writer_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11RecordField_11writer_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11RecordField_11writer_name___get__(((struct __pyx_obj_5cavro_RecordField *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11RecordField_11writer_name___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->writer_name);
  __pyx_r = __pyx_v_self->writer_name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":197
 *     cdef readonly str name
 *     cdef readonly str writer_name
 *     cdef readonly str doc             # <<<<<<<<<<<<<<
 *     cdef readonly AvroType type
 *     cdef readonly object default_value
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11RecordField_3doc_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11RecordField_3doc_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11RecordField_3doc___get__(((struct __pyx_obj_5cavro_RecordField *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11RecordField_3doc___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->doc);
  __pyx_r = __pyx_v_self->doc;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":198
 *     cdef readonly str writer_name
 *     cdef readonly str doc
 *     cdef readonly AvroType type             # <<<<<<<<<<<<<<
 *     cdef readonly object default_value
 *     cdef readonly object order
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11RecordField_4type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11RecordField_4type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11RecordField_4type___get__(((struct __pyx_obj_5cavro_RecordField *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11RecordField_4type___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->type);
  __pyx_r = ((PyObject *)__pyx_v_self->type);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":199
 *     cdef readonly str doc
 *     cdef readonly AvroType type
 *     cdef readonly object default_value             # <<<<<<<<<<<<<<
 *     cdef readonly object order
 *     cdef readonly frozenset aliases
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11RecordField_13default_value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11RecordField_13default_value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11RecordField_13default_value___get__(((struct __pyx_obj_5cavro_RecordField *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11RecordField_13default_value___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->default_value);
  __pyx_r = __pyx_v_self->default_value;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":200
 *     cdef readonly AvroType type
 *     cdef readonly object default_value
 *     cdef readonly object order             # <<<<<<<<<<<<<<
 *     cdef readonly frozenset aliases
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11RecordField_5order_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11RecordField_5order_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11RecordField_5order___get__(((struct __pyx_obj_5cavro_RecordField *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11RecordField_5order___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->order);
  __pyx_r = __pyx_v_self->order;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":201
 *     cdef readonly object default_value
 *     cdef readonly object order
 *     cdef readonly frozenset aliases             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, schema, source, namespace):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11RecordField_7aliases_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11RecordField_7aliases_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11RecordField_7aliases___get__(((struct __pyx_obj_5cavro_RecordField *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11RecordField_7aliases___get__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->aliases);
  __pyx_r = __pyx_v_self->aliases;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11RecordField_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_11RecordField_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11RecordField_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_11RecordField_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_11RecordField_4__reduce_cython__(((struct __pyx_obj_5cavro_RecordField *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11RecordField_4__reduce_cython__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.aliases, self.default_value, self.doc, self.name, self.order, self.type, self.writer_name)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->aliases);
  __Pyx_GIVEREF(__pyx_v_self->aliases);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->aliases)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->default_value);
  __Pyx_GIVEREF(__pyx_v_self->default_value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->default_value)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->doc);
  __Pyx_GIVEREF(__pyx_v_self->doc);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->doc)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->order);
  __Pyx_GIVEREF(__pyx_v_self->order);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_self->order)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->type);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, ((PyObject *)__pyx_v_self->type))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->writer_name);
  __Pyx_GIVEREF(__pyx_v_self->writer_name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_v_self->writer_name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.aliases, self.default_value, self.doc, self.name, self.order, self.type, self.writer_name)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.aliases, self.default_value, self.doc, self.name, self.order, self.type, self.writer_name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.name is not None or self.order is not None or self.type is not None or self.writer_name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.aliases, self.default_value, self.doc, self.name, self.order, self.type, self.writer_name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.name is not None or self.order is not None or self.type is not None or self.writer_name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_RecordField, (type(self), 0x1d9a0f8, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->aliases != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->default_value != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->doc != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->name != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->order != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->type) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->writer_name != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.name is not None or self.order is not None or self.type is not None or self.writer_name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_RecordField, (type(self), 0x1d9a0f8, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.name is not None or self.order is not None or self.type is not None or self.writer_name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_RecordField, (type(self), 0x1d9a0f8, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_RecordField, (type(self), 0x1d9a0f8, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_RecordField); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_31039736);
    __Pyx_GIVEREF(__pyx_int_31039736);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_31039736)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.default_value is not None or self.doc is not None or self.name is not None or self.order is not None or self.type is not None or self.writer_name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_RecordField, (type(self), 0x1d9a0f8, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_RecordField, (type(self), 0x1d9a0f8, None), state
 *     else:
 *         return __pyx_unpickle_RecordField, (type(self), 0x1d9a0f8, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_RecordField__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_RecordField); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_31039736);
    __Pyx_GIVEREF(__pyx_int_31039736);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_31039736)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.RecordField.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RecordField, (type(self), 0x1d9a0f8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RecordField__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11RecordField_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_11RecordField_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11RecordField_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_11RecordField_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.RecordField.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_11RecordField_6__setstate_cython__(((struct __pyx_obj_5cavro_RecordField *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11RecordField_6__setstate_cython__(struct __pyx_obj_5cavro_RecordField *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_RecordField, (type(self), 0x1d9a0f8, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_RecordField__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_RecordField__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RecordField, (type(self), 0x1d9a0f8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RecordField__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.RecordField.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":270
 *     cdef Py_ssize_t index
 * 
 *     def __init__(self, index):             # <<<<<<<<<<<<<<
 *         self.index = index
 * 
 */

/* Python wrapper */
static int __pyx_pw_5cavro_14_FieldAccessor_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_14_FieldAccessor_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_index = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_index,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_index)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 270, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(11, 270, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_index = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(11, 270, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._FieldAccessor.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_14_FieldAccessor___init__(((struct __pyx_obj_5cavro__FieldAccessor *)__pyx_v_self), __pyx_v_index);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_14_FieldAccessor___init__(struct __pyx_obj_5cavro__FieldAccessor *__pyx_v_self, PyObject *__pyx_v_index) {
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/record.pxi":271
 * 
 *     def __init__(self, index):
 *         self.index = index             # <<<<<<<<<<<<<<
 * 
 *     def __get__(self, inst, cls):
 */
  __pyx_t_1 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(11, 271, __pyx_L1_error)
  __pyx_v_self->index = __pyx_t_1;

  /* "src/record.pxi":270
 *     cdef Py_ssize_t index
 * 
 *     def __init__(self, index):             # <<<<<<<<<<<<<<
 *         self.index = index
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro._FieldAccessor.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/record.pxi":273
 *         self.index = index
 * 
 *     def __get__(self, inst, cls):             # <<<<<<<<<<<<<<
 *         cdef Record record = inst
 *         return record.data[self.index]
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14_FieldAccessor_3__get__(PyObject *__pyx_v_self, PyObject *__pyx_v_inst, PyObject *__pyx_v_cls); /*proto*/
static PyObject *__pyx_pw_5cavro_14_FieldAccessor_3__get__(PyObject *__pyx_v_self, PyObject *__pyx_v_inst, PyObject *__pyx_v_cls) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_14_FieldAccessor_2__get__(((struct __pyx_obj_5cavro__FieldAccessor *)__pyx_v_self), ((PyObject *)__pyx_v_inst), ((PyObject *)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14_FieldAccessor_2__get__(struct __pyx_obj_5cavro__FieldAccessor *__pyx_v_self, PyObject *__pyx_v_inst, CYTHON_UNUSED PyObject *__pyx_v_cls) {
  struct __pyx_obj_5cavro_Record *__pyx_v_record = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/record.pxi":274
 * 
 *     def __get__(self, inst, cls):
 *         cdef Record record = inst             # <<<<<<<<<<<<<<
 *         return record.data[self.index]
 * 
 */
  if (!(likely(((__pyx_v_inst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_inst, __pyx_ptype_5cavro_Record))))) __PYX_ERR(11, 274, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_inst;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_record = ((struct __pyx_obj_5cavro_Record *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":275
 *     def __get__(self, inst, cls):
 *         cdef Record record = inst
 *         return record.data[self.index]             # <<<<<<<<<<<<<<
 * 
 *     def __set__(self, inst, value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_record->data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(11, 275, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_record->data, __pyx_v_self->index, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":273
 *         self.index = index
 * 
 *     def __get__(self, inst, cls):             # <<<<<<<<<<<<<<
 *         cdef Record record = inst
 *         return record.data[self.index]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._FieldAccessor.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_record);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":277
 *         return record.data[self.index]
 * 
 *     def __set__(self, inst, value):             # <<<<<<<<<<<<<<
 *         cdef Record record = inst
 *         record.data[self.index] = value
 */

/* Python wrapper */
static int __pyx_pw_5cavro_14_FieldAccessor_5__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_inst, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5cavro_14_FieldAccessor_5__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_inst, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_14_FieldAccessor_4__set__(((struct __pyx_obj_5cavro__FieldAccessor *)__pyx_v_self), ((PyObject *)__pyx_v_inst), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_14_FieldAccessor_4__set__(struct __pyx_obj_5cavro__FieldAccessor *__pyx_v_self, PyObject *__pyx_v_inst, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_Record *__pyx_v_record = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "src/record.pxi":278
 * 
 *     def __set__(self, inst, value):
 *         cdef Record record = inst             # <<<<<<<<<<<<<<
 *         record.data[self.index] = value
 * 
 */
  if (!(likely(((__pyx_v_inst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_inst, __pyx_ptype_5cavro_Record))))) __PYX_ERR(11, 278, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_inst;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_record = ((struct __pyx_obj_5cavro_Record *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":279
 *     def __set__(self, inst, value):
 *         cdef Record record = inst
 *         record.data[self.index] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_record->data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(11, 279, __pyx_L1_error)
  }
  if (unlikely((__Pyx_SetItemInt(__pyx_v_record->data, __pyx_v_self->index, __pyx_v_value, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1) < 0))) __PYX_ERR(11, 279, __pyx_L1_error)

  /* "src/record.pxi":277
 *         return record.data[self.index]
 * 
 *     def __set__(self, inst, value):             # <<<<<<<<<<<<<<
 *         cdef Record record = inst
 *         record.data[self.index] = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._FieldAccessor.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_record);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14_FieldAccessor_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_14_FieldAccessor_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14_FieldAccessor_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_14_FieldAccessor_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_14_FieldAccessor_6__reduce_cython__(((struct __pyx_obj_5cavro__FieldAccessor *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14_FieldAccessor_6__reduce_cython__(struct __pyx_obj_5cavro__FieldAccessor *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.index,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.index,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.index,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.index,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__FieldAccessor, (type(self), 0x1bc04b5, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__FieldAccessor, (type(self), 0x1bc04b5, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle__FieldAccessor, (type(self), 0x1bc04b5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__FieldAccessor, (type(self), 0x1bc04b5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle__FieldAccessor); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_29099189);
    __Pyx_GIVEREF(__pyx_int_29099189);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_29099189)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__FieldAccessor, (type(self), 0x1bc04b5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__FieldAccessor, (type(self), 0x1bc04b5, None), state
 *     else:
 *         return __pyx_unpickle__FieldAccessor, (type(self), 0x1bc04b5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__FieldAccessor__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle__FieldAccessor); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_29099189);
    __Pyx_GIVEREF(__pyx_int_29099189);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_29099189)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._FieldAccessor.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__FieldAccessor, (type(self), 0x1bc04b5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__FieldAccessor__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14_FieldAccessor_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_14_FieldAccessor_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14_FieldAccessor_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_14_FieldAccessor_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._FieldAccessor.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_14_FieldAccessor_8__setstate_cython__(((struct __pyx_obj_5cavro__FieldAccessor *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14_FieldAccessor_8__setstate_cython__(struct __pyx_obj_5cavro__FieldAccessor *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__FieldAccessor, (type(self), 0x1bc04b5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__FieldAccessor__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__FieldAccessor__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__FieldAccessor, (type(self), 0x1bc04b5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__FieldAccessor__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._FieldAccessor.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":282
 * 
 * 
 * cdef object _make_record_class(RecordType record_type):             # <<<<<<<<<<<<<<
 *     attrs = {}
 *     field_to_index = {}
 */

static PyObject *__pyx_f_5cavro__make_record_class(struct __pyx_obj_5cavro_RecordType *__pyx_v_record_type) {
  PyObject *__pyx_v_attrs = NULL;
  PyObject *__pyx_v_field_to_index = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_field = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_record_class", 1);

  /* "src/record.pxi":283
 * 
 * cdef object _make_record_class(RecordType record_type):
 *     attrs = {}             # <<<<<<<<<<<<<<
 *     field_to_index = {}
 *     for i, field in enumerate(record_type.fields):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_attrs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":284
 * cdef object _make_record_class(RecordType record_type):
 *     attrs = {}
 *     field_to_index = {}             # <<<<<<<<<<<<<<
 *     for i, field in enumerate(record_type.fields):
 *         attrs[field.name] = _FieldAccessor(i)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_field_to_index = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":285
 *     attrs = {}
 *     field_to_index = {}
 *     for i, field in enumerate(record_type.fields):             # <<<<<<<<<<<<<<
 *         attrs[field.name] = _FieldAccessor(i)
 *         field_to_index[field.name] = i
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_1 = __pyx_int_0;
  __pyx_t_2 = __pyx_v_record_type->fields; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 285, __pyx_L1_error)
      #endif
      if (__pyx_t_3 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(11, 285, __pyx_L1_error)
    #else
    __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "src/record.pxi":286
 *     field_to_index = {}
 *     for i, field in enumerate(record_type.fields):
 *         attrs[field.name] = _FieldAccessor(i)             # <<<<<<<<<<<<<<
 *         field_to_index[field.name] = i
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro__FieldAccessor), __pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely((PyDict_SetItem(__pyx_v_attrs, __pyx_t_5, __pyx_t_4) < 0))) __PYX_ERR(11, 286, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "src/record.pxi":287
 *     for i, field in enumerate(record_type.fields):
 *         attrs[field.name] = _FieldAccessor(i)
 *         field_to_index[field.name] = i             # <<<<<<<<<<<<<<
 * 
 *     attrs['Type'] = record_type
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely((PyDict_SetItem(__pyx_v_field_to_index, __pyx_t_4, __pyx_v_i) < 0))) __PYX_ERR(11, 287, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "src/record.pxi":285
 *     attrs = {}
 *     field_to_index = {}
 *     for i, field in enumerate(record_type.fields):             # <<<<<<<<<<<<<<
 *         attrs[field.name] = _FieldAccessor(i)
 *         field_to_index[field.name] = i
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/record.pxi":289
 *         field_to_index[field.name] = i
 * 
 *     attrs['Type'] = record_type             # <<<<<<<<<<<<<<
 *     attrs['__slots__'] = ()
 *     attrs['_field_to_index'] = field_to_index
 */
  if (unlikely((PyDict_SetItem(__pyx_v_attrs, __pyx_n_u_Type, ((PyObject *)__pyx_v_record_type)) < 0))) __PYX_ERR(11, 289, __pyx_L1_error)

  /* "src/record.pxi":290
 * 
 *     attrs['Type'] = record_type
 *     attrs['__slots__'] = ()             # <<<<<<<<<<<<<<
 *     attrs['_field_to_index'] = field_to_index
 *     return type(
 */
  if (unlikely((PyDict_SetItem(__pyx_v_attrs, __pyx_n_u_slots, __pyx_empty_tuple) < 0))) __PYX_ERR(11, 290, __pyx_L1_error)

  /* "src/record.pxi":291
 *     attrs['Type'] = record_type
 *     attrs['__slots__'] = ()
 *     attrs['_field_to_index'] = field_to_index             # <<<<<<<<<<<<<<
 *     return type(
 *         record_type.name,
 */
  if (unlikely((PyDict_SetItem(__pyx_v_attrs, __pyx_n_u_field_to_index, __pyx_v_field_to_index) < 0))) __PYX_ERR(11, 291, __pyx_L1_error)

  /* "src/record.pxi":292
 *     attrs['__slots__'] = ()
 *     attrs['_field_to_index'] = field_to_index
 *     return type(             # <<<<<<<<<<<<<<
 *         record_type.name,
 *         (Record, ),
 */
  __Pyx_XDECREF(__pyx_r);

  /* "src/record.pxi":294
 *     return type(
 *         record_type.name,
 *         (Record, ),             # <<<<<<<<<<<<<<
 *         attrs
 *     )
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_Record);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_Record);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_ptype_5cavro_Record))) __PYX_ERR(11, 294, __pyx_L1_error);

  /* "src/record.pxi":292
 *     attrs['__slots__'] = ()
 *     attrs['_field_to_index'] = field_to_index
 *     return type(             # <<<<<<<<<<<<<<
 *         record_type.name,
 *         (Record, ),
 */
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_record_type->__pyx_base.name);
  __Pyx_GIVEREF(__pyx_v_record_type->__pyx_base.name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_record_type->__pyx_base.name)) __PYX_ERR(11, 292, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1)) __PYX_ERR(11, 292, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_attrs);
  __Pyx_GIVEREF(__pyx_v_attrs);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_attrs)) __PYX_ERR(11, 292, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)(&PyType_Type)), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":282
 * 
 * 
 * cdef object _make_record_class(RecordType record_type):             # <<<<<<<<<<<<<<
 *     attrs = {}
 *     field_to_index = {}
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro._make_record_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_attrs);
  __Pyx_XDECREF(__pyx_v_field_to_index);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":321
 *     cdef bint _setting_up
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         cdef Schema schema_ = schema
 *         self._setting_up = True
 */

/* Python wrapper */
static int __pyx_pw_5cavro_10RecordType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_10RecordType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_namespace = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_schema,&__pyx_n_s_source,&__pyx_n_s_namespace,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 321, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 321, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(11, 321, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_namespace)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 321, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(11, 321, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(11, 321, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_schema = values[0];
    __pyx_v_source = values[1];
    __pyx_v_namespace = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(11, 321, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.RecordType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10RecordType___init__(((struct __pyx_obj_5cavro_RecordType *)__pyx_v_self), __pyx_v_schema, __pyx_v_source, __pyx_v_namespace);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_10RecordType_8__init___2generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/record.pxi":327
 *         _NamedType.__init__(self, schema, source, namespace)
 *         self.fields = tuple(
 *             RecordField(schema_, f, self.effective_namespace) for f in source['fields']             # <<<<<<<<<<<<<<
 *         )
 *         self.field_dict = {}
 */

static PyObject *__pyx_pf_5cavro_10RecordType_8__init___genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_9_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_9_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(11, 327, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_10RecordType_8__init___2generator10, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_init___locals_genexpr, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(11, 327, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.RecordType.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_10RecordType_8__init___2generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(11, 327, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(11, 327, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 327, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 327, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(11, 327, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 327, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 327, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(11, 327, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 327, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(11, 327, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_f);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_f, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_schema_)) { __Pyx_RaiseClosureNameError("schema_"); __PYX_ERR(11, 327, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(11, 327, __pyx_L1_error) }
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_schema_);
    __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_schema_);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_schema_))) __PYX_ERR(11, 327, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_f);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_f);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_cur_scope->__pyx_v_f)) __PYX_ERR(11, 327, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->__pyx_base.effective_namespace);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->__pyx_base.effective_namespace);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->__pyx_base.effective_namespace)) __PYX_ERR(11, 327, __pyx_L1_error);
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_RecordField), __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(11, 327, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":321
 *     cdef bint _setting_up
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         cdef Schema schema_ = schema
 *         self._setting_up = True
 */

static int __pyx_pf_5cavro_10RecordType___init__(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace) {
  struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *__pyx_cur_scope;
  PyObject *__pyx_v_field = NULL;
  CYTHON_UNUSED Py_ssize_t __pyx_v_n_fields;
  PyObject *__pyx_v_logical = NULL;
  PyObject *__pyx_gb_5cavro_10RecordType_8__init___2generator10 = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *)__pyx_tp_new_5cavro___pyx_scope_struct_8___init__(__pyx_ptype_5cavro___pyx_scope_struct_8___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(11, 321, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "src/record.pxi":322
 * 
 *     def __init__(self, schema, source, namespace):
 *         cdef Schema schema_ = schema             # <<<<<<<<<<<<<<
 *         self._setting_up = True
 *         self.doc = source.get('doc', '')
 */
  if (!(likely(((__pyx_v_schema) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_schema, __pyx_ptype_5cavro_Schema))))) __PYX_ERR(11, 322, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_schema;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_schema_ = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":323
 *     def __init__(self, schema, source, namespace):
 *         cdef Schema schema_ = schema
 *         self._setting_up = True             # <<<<<<<<<<<<<<
 *         self.doc = source.get('doc', '')
 *         _NamedType.__init__(self, schema, source, namespace)
 */
  __pyx_cur_scope->__pyx_v_self->_setting_up = 1;

  /* "src/record.pxi":324
 *         cdef Schema schema_ = schema
 *         self._setting_up = True
 *         self.doc = source.get('doc', '')             # <<<<<<<<<<<<<<
 *         _NamedType.__init__(self, schema, source, namespace)
 *         self.fields = tuple(
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(11, 324, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->doc);
  __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->doc);
  __pyx_cur_scope->__pyx_v_self->doc = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/record.pxi":325
 *         self._setting_up = True
 *         self.doc = source.get('doc', '')
 *         _NamedType.__init__(self, schema, source, namespace)             # <<<<<<<<<<<<<<
 *         self.fields = tuple(
 *             RecordField(schema_, f, self.effective_namespace) for f in source['fields']
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__NamedType), __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_3, ((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_v_schema, __pyx_v_source, __pyx_v_namespace};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 4+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/record.pxi":327
 *         _NamedType.__init__(self, schema, source, namespace)
 *         self.fields = tuple(
 *             RecordField(schema_, f, self.effective_namespace) for f in source['fields']             # <<<<<<<<<<<<<<
 *         )
 *         self.field_dict = {}
 */
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_source, __pyx_n_u_fields); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_pf_5cavro_10RecordType_8__init___genexpr(((PyObject*)__pyx_cur_scope), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/record.pxi":326
 *         self.doc = source.get('doc', '')
 *         _NamedType.__init__(self, schema, source, namespace)
 *         self.fields = tuple(             # <<<<<<<<<<<<<<
 *             RecordField(schema_, f, self.effective_namespace) for f in source['fields']
 *         )
 */
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->fields);
  __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->fields);
  __pyx_cur_scope->__pyx_v_self->fields = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/record.pxi":329
 *             RecordField(schema_, f, self.effective_namespace) for f in source['fields']
 *         )
 *         self.field_dict = {}             # <<<<<<<<<<<<<<
 *         for field in self.fields:
 *             if schema_.options.record_fields_must_be_unique and field.name in self.field_dict:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->field_dict);
  __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->field_dict);
  __pyx_cur_scope->__pyx_v_self->field_dict = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/record.pxi":330
 *         )
 *         self.field_dict = {}
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             if schema_.options.record_fields_must_be_unique and field.name in self.field_dict:
 *                 raise InvalidName(f'Duplicate field name: {field.name}')
 */
  if (unlikely(__pyx_cur_scope->__pyx_v_self->fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(11, 330, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_self->fields; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 330, __pyx_L1_error)
      #endif
      if (__pyx_t_5 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(11, 330, __pyx_L1_error)
    #else
    __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/record.pxi":331
 *         self.field_dict = {}
 *         for field in self.fields:
 *             if schema_.options.record_fields_must_be_unique and field.name in self.field_dict:             # <<<<<<<<<<<<<<
 *                 raise InvalidName(f'Duplicate field name: {field.name}')
 *             self.field_dict[field.name] = field
 */
    if (__pyx_cur_scope->__pyx_v_schema_->options->record_fields_must_be_unique) {
    } else {
      __pyx_t_6 = __pyx_cur_scope->__pyx_v_schema_->options->record_fields_must_be_unique;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_cur_scope->__pyx_v_self->field_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(11, 331, __pyx_L1_error)
    }
    __pyx_t_7 = (__Pyx_PyDict_ContainsTF(__pyx_t_1, __pyx_cur_scope->__pyx_v_self->field_dict, Py_EQ)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(11, 331, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __pyx_t_7;
    __pyx_L6_bool_binop_done:;
    if (unlikely(__pyx_t_6)) {

      /* "src/record.pxi":332
 *         for field in self.fields:
 *             if schema_.options.record_fields_must_be_unique and field.name in self.field_dict:
 *                 raise InvalidName(f'Duplicate field name: {field.name}')             # <<<<<<<<<<<<<<
 *             self.field_dict[field.name] = field
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_InvalidName); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_8)) __PYX_ERR(11, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Duplicate_field_name, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(11, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_8};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 332, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(11, 332, __pyx_L1_error)

      /* "src/record.pxi":331
 *         self.field_dict = {}
 *         for field in self.fields:
 *             if schema_.options.record_fields_must_be_unique and field.name in self.field_dict:             # <<<<<<<<<<<<<<
 *                 raise InvalidName(f'Duplicate field name: {field.name}')
 *             self.field_dict[field.name] = field
 */
    }

    /* "src/record.pxi":333
 *             if schema_.options.record_fields_must_be_unique and field.name in self.field_dict:
 *                 raise InvalidName(f'Duplicate field name: {field.name}')
 *             self.field_dict[field.name] = field             # <<<<<<<<<<<<<<
 * 
 *         n_fields = len(self.fields)
 */
    if (unlikely(__pyx_cur_scope->__pyx_v_self->field_dict == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(11, 333, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely((PyDict_SetItem(__pyx_cur_scope->__pyx_v_self->field_dict, __pyx_t_1, __pyx_v_field) < 0))) __PYX_ERR(11, 333, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/record.pxi":330
 *         )
 *         self.field_dict = {}
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             if schema_.options.record_fields_must_be_unique and field.name in self.field_dict:
 *                 raise InvalidName(f'Duplicate field name: {field.name}')
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/record.pxi":335
 *             self.field_dict[field.name] = field
 * 
 *         n_fields = len(self.fields)             # <<<<<<<<<<<<<<
 *         self.record = _make_record_class(self)
 * 
 */
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_self->fields;
  __Pyx_INCREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(11, 335, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_t_2); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(11, 335, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_n_fields = __pyx_t_5;

  /* "src/record.pxi":336
 * 
 *         n_fields = len(self.fields)
 *         self.record = _make_record_class(self)             # <<<<<<<<<<<<<<
 * 
 *         self._setting_up = False
 */
  __pyx_t_2 = ((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_f_5cavro__make_record_class(((struct __pyx_obj_5cavro_RecordType *)__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyType_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("type", __pyx_t_1))) __PYX_ERR(11, 336, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_self->record);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope->__pyx_v_self->record);
  __pyx_cur_scope->__pyx_v_self->record = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":338
 *         self.record = _make_record_class(self)
 * 
 *         self._setting_up = False             # <<<<<<<<<<<<<<
 *         logical = self._make_logical(schema, source)
 *         if logical is not None:
 */
  __pyx_cur_scope->__pyx_v_self->_setting_up = 0;

  /* "src/record.pxi":339
 * 
 *         self._setting_up = False
 *         logical = self._make_logical(schema, source)             # <<<<<<<<<<<<<<
 *         if logical is not None:
 *             self.value_adapters = self.value_adapters + (logical,)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_RecordType *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._make_logical(((struct __pyx_obj_5cavro_AvroType *)__pyx_cur_scope->__pyx_v_self), __pyx_v_schema, __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_logical = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/record.pxi":340
 *         self._setting_up = False
 *         logical = self._make_logical(schema, source)
 *         if logical is not None:             # <<<<<<<<<<<<<<
 *             self.value_adapters = self.value_adapters + (logical,)
 * 
 */
  __pyx_t_6 = (__pyx_v_logical != Py_None);
  if (__pyx_t_6) {

    /* "src/record.pxi":341
 *         logical = self._make_logical(schema, source)
 *         if logical is not None:
 *             self.value_adapters = self.value_adapters + (logical,)             # <<<<<<<<<<<<<<
 * 
 *     cdef _make_logical(self, schema, source):
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_logical);
    __Pyx_GIVEREF(__pyx_v_logical);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_logical)) __PYX_ERR(11, 341, __pyx_L1_error);
    __pyx_t_2 = PyNumber_Add(__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.value_adapters, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.value_adapters);
    __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.value_adapters);
    __pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.value_adapters = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/record.pxi":340
 *         self._setting_up = False
 *         logical = self._make_logical(schema, source)
 *         if logical is not None:             # <<<<<<<<<<<<<<
 *             self.value_adapters = self.value_adapters + (logical,)
 * 
 */
  }

  /* "src/record.pxi":321
 *     cdef bint _setting_up
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         cdef Schema schema_ = schema
 *         self._setting_up = True
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cavro.RecordType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_logical);
  __Pyx_XDECREF(__pyx_gb_5cavro_10RecordType_8__init___2generator10);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":343
 *             self.value_adapters = self.value_adapters + (logical,)
 * 
 *     cdef _make_logical(self, schema, source):             # <<<<<<<<<<<<<<
 *         if self._setting_up:
 *             return
 */

static PyObject *__pyx_f_5cavro_10RecordType__make_logical(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_logical", 1);

  /* "src/record.pxi":344
 * 
 *     cdef _make_logical(self, schema, source):
 *         if self._setting_up:             # <<<<<<<<<<<<<<
 *             return
 *         return AvroType._make_logical(self, schema, source)
 */
  if (__pyx_v_self->_setting_up) {

    /* "src/record.pxi":345
 *     cdef _make_logical(self, schema, source):
 *         if self._setting_up:
 *             return             # <<<<<<<<<<<<<<
 *         return AvroType._make_logical(self, schema, source)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/record.pxi":344
 * 
 *     cdef _make_logical(self, schema, source):
 *         if self._setting_up:             # <<<<<<<<<<<<<<
 *             return
 *         return AvroType._make_logical(self, schema, source)
 */
  }

  /* "src/record.pxi":346
 *         if self._setting_up:
 *             return
 *         return AvroType._make_logical(self, schema, source)             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8AvroType__make_logical(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_schema, __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":343
 *             self.value_adapters = self.value_adapters + (logical,)
 * 
 *     cdef _make_logical(self, schema, source):             # <<<<<<<<<<<<<<
 *         if self._setting_up:
 *             return
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.RecordType._make_logical", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":348
 *         return AvroType._make_logical(self, schema, source)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef RecordType new_inst = self.clone_base()
 *         new_inst.doc = self.doc
 */

static PyObject *__pyx_pw_5cavro_10RecordType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10RecordType_copy(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_RecordType *__pyx_v_new_inst = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_10RecordType_3copy)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 348, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(11, 348, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/record.pxi":349
 * 
 *     cpdef AvroType copy(self):
 *         cdef RecordType new_inst = self.clone_base()             # <<<<<<<<<<<<<<
 *         new_inst.doc = self.doc
 *         new_inst.fields = self.fields
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_RecordType *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_RecordType))))) __PYX_ERR(11, 349, __pyx_L1_error)
  __pyx_v_new_inst = ((struct __pyx_obj_5cavro_RecordType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":350
 *     cpdef AvroType copy(self):
 *         cdef RecordType new_inst = self.clone_base()
 *         new_inst.doc = self.doc             # <<<<<<<<<<<<<<
 *         new_inst.fields = self.fields
 *         new_inst.field_dict = self.field_dict
 */
  __pyx_t_1 = __pyx_v_self->doc;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_new_inst->doc);
  __Pyx_DECREF(__pyx_v_new_inst->doc);
  __pyx_v_new_inst->doc = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":351
 *         cdef RecordType new_inst = self.clone_base()
 *         new_inst.doc = self.doc
 *         new_inst.fields = self.fields             # <<<<<<<<<<<<<<
 *         new_inst.field_dict = self.field_dict
 *         new_inst.record = self.record
 */
  __pyx_t_1 = __pyx_v_self->fields;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_new_inst->fields);
  __Pyx_DECREF(__pyx_v_new_inst->fields);
  __pyx_v_new_inst->fields = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":352
 *         new_inst.doc = self.doc
 *         new_inst.fields = self.fields
 *         new_inst.field_dict = self.field_dict             # <<<<<<<<<<<<<<
 *         new_inst.record = self.record
 *         return new_inst
 */
  __pyx_t_1 = __pyx_v_self->field_dict;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_new_inst->field_dict);
  __Pyx_DECREF(__pyx_v_new_inst->field_dict);
  __pyx_v_new_inst->field_dict = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":353
 *         new_inst.fields = self.fields
 *         new_inst.field_dict = self.field_dict
 *         new_inst.record = self.record             # <<<<<<<<<<<<<<
 *         return new_inst
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->record);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_new_inst->record);
  __Pyx_DECREF((PyObject *)__pyx_v_new_inst->record);
  __pyx_v_new_inst->record = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":354
 *         new_inst.field_dict = self.field_dict
 *         new_inst.record = self.record
 *         return new_inst             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_new_inst);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_new_inst);
  goto __pyx_L0;

  /* "src/record.pxi":348
 *         return AvroType._make_logical(self, schema, source)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef RecordType new_inst = self.clone_base()
 *         new_inst.doc = self.doc
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.RecordType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_new_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10RecordType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10RecordType_3copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10RecordType_3copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10RecordType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10RecordType_2copy(((struct __pyx_obj_5cavro_RecordType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10RecordType_2copy(struct __pyx_obj_5cavro_RecordType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_10RecordType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.RecordType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":356
 *         return new_inst
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {
 *             'type',
 */

static PyObject *__pyx_f_5cavro_10RecordType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/record.pxi":357
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {             # <<<<<<<<<<<<<<
 *             'type',
 *             'name',
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(11, 358, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_name) < 0) __PYX_ERR(11, 358, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_namespace) < 0) __PYX_ERR(11, 358, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_doc) < 0) __PYX_ERR(11, 358, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_aliases) < 0) __PYX_ERR(11, 358, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_fields) < 0) __PYX_ERR(11, 358, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":356
 *         return new_inst
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {
 *             'type',
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.RecordType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_10RecordType_6generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/record.pxi":366
 *         })
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10RecordType_5walk_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10RecordType_5walk_types = {"walk_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10RecordType_5walk_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10RecordType_5walk_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_visited = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("walk_types (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_visited,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_visited)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 366, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "walk_types") < 0)) __PYX_ERR(11, 366, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_visited = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("walk_types", 1, 1, 1, __pyx_nargs); __PYX_ERR(11, 366, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.RecordType.walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10RecordType_4walk_types(((struct __pyx_obj_5cavro_RecordType *)__pyx_v_self), __pyx_v_visited);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10RecordType_4walk_types(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_visited) {
  struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("walk_types", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types *)__pyx_tp_new_5cavro___pyx_scope_struct_10_walk_types(__pyx_ptype_5cavro___pyx_scope_struct_10_walk_types, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(11, 366, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_visited = __pyx_v_visited;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_visited);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_visited);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_10RecordType_6generator4, __pyx_codeobj__66, (PyObject *) __pyx_cur_scope, __pyx_n_s_walk_types, __pyx_n_s_RecordType_walk_types, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(11, 366, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.RecordType.walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_10RecordType_6generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("walk_types", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L5_resume_from_yield_from;
    case 2: goto __pyx_L8_resume_from_yield_from;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(11, 366, __pyx_L1_error)

  /* "src/record.pxi":367
 * 
 *     def walk_types(self, visited):
 *         if self in visited:             # <<<<<<<<<<<<<<
 *             return
 *         yield from super().walk_types(visited)
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_visited, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(11, 367, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/record.pxi":368
 *     def walk_types(self, visited):
 *         if self in visited:
 *             return             # <<<<<<<<<<<<<<
 *         yield from super().walk_types(visited)
 *         for field in self.fields:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "src/record.pxi":367
 * 
 *     def walk_types(self, visited):
 *         if self in visited:             # <<<<<<<<<<<<<<
 *             return
 *         yield from super().walk_types(visited)
 */
  }

  /* "src/record.pxi":369
 *         if self in visited:
 *             return
 *         yield from super().walk_types(visited)             # <<<<<<<<<<<<<<
 *         for field in self.fields:
 *             yield from field.type.walk_types(visited)
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_RecordType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_RecordType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_ptype_5cavro_RecordType))) __PYX_ERR(11, 369, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_cur_scope->__pyx_v_self))) __PYX_ERR(11, 369, __pyx_L1_error);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_walk_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_visited};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __Pyx_Generator_Yield_From(__pyx_generator, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L5_resume_from_yield_from:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(11, 369, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(11, 369, __pyx_L1_error)
    }
  }

  /* "src/record.pxi":370
 *             return
 *         yield from super().walk_types(visited)
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             yield from field.type.walk_types(visited)
 * 
 */
  if (unlikely(__pyx_cur_scope->__pyx_v_self->fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(11, 370, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_self->fields; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_6 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 370, __pyx_L1_error)
      #endif
      if (__pyx_t_6 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(11, 370, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_field);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_field, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/record.pxi":371
 *         yield from super().walk_types(visited)
 *         for field in self.fields:
 *             yield from field.type.walk_types(visited)             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_field, __pyx_n_s_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_walk_types); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_visited};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __pyx_r = __Pyx_Generator_Yield_From(__pyx_generator, __pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XGOTREF(__pyx_r);
    if (likely(__pyx_r)) {
      __Pyx_XGIVEREF(__pyx_t_2);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_6;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 2;
      return __pyx_r;
      __pyx_L8_resume_from_yield_from:;
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_2);
      __pyx_t_6 = __pyx_cur_scope->__pyx_t_1;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(11, 371, __pyx_L1_error)
    } else {
      PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
      if (exc_type) {
        if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
        else __PYX_ERR(11, 371, __pyx_L1_error)
      }
    }

    /* "src/record.pxi":370
 *             return
 *         yield from super().walk_types(visited)
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             yield from field.type.walk_types(visited)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "src/record.pxi":366
 *         })
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":373
 *             yield from field.type.walk_types(visited)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         extra = super(RecordType, self)._get_schema_extra(created)
 *         return dict(extra, fields=[f.get_schema(created) for f in self.fields])
 */

static PyObject *__pyx_pw_5cavro_10RecordType_8_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_10RecordType__get_schema_extra(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_created, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_extra = NULL;
  PyObject *__pyx_9genexpr12__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_schema_extra); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 373, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_10RecordType_8_get_schema_extra)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_created};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 373, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_2))) __PYX_ERR(11, 373, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/record.pxi":374
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         extra = super(RecordType, self)._get_schema_extra(created)             # <<<<<<<<<<<<<<
 *         return dict(extra, fields=[f.get_schema(created) for f in self.fields])
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_RecordType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_RecordType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_5cavro_RecordType))) __PYX_ERR(11, 374, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(11, 374, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get_schema_extra); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_created};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_extra = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/record.pxi":375
 *     cpdef dict _get_schema_extra(self, set created):
 *         extra = super(RecordType, self)._get_schema_extra(created)
 *         return dict(extra, fields=[f.get_schema(created) for f in self.fields])             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_extra);
  __Pyx_GIVEREF(__pyx_v_extra);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_extra)) __PYX_ERR(11, 375, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 375, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_self->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(11, 375, __pyx_L5_error)
    }
    __pyx_t_4 = __pyx_v_self->fields; __Pyx_INCREF(__pyx_t_4);
    __pyx_t_6 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 375, __pyx_L5_error)
        #endif
        if (__pyx_t_6 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_7); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(11, 375, __pyx_L5_error)
      #else
      __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 375, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_f, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr12__pyx_v_f, __pyx_n_s_get_schema); if (unlikely(!__pyx_t_8)) __PYX_ERR(11, 375, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_created};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 375, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_7))) __PYX_ERR(11, 375, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_f); __pyx_9genexpr12__pyx_v_f = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_f); __pyx_9genexpr12__pyx_v_f = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fields, __pyx_t_3) < 0) __PYX_ERR(11, 375, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)(&PyDict_Type)), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":373
 *             yield from field.type.walk_types(visited)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         extra = super(RecordType, self)._get_schema_extra(created)
 *         return dict(extra, fields=[f.get_schema(created) for f in self.fields])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cavro.RecordType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_extra);
  __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10RecordType_8_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10RecordType_8_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10RecordType_8_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10RecordType_8_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 373, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(11, 373, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(11, 373, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.RecordType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(11, 373, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_10RecordType_7_get_schema_extra(((struct __pyx_obj_5cavro_RecordType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10RecordType_7_get_schema_extra(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_10RecordType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.RecordType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":377
 *         return dict(extra, fields=[f.get_schema(created) for f in self.fields])
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         cdef list rec_data
 */

static int __pyx_f_5cavro_10RecordType__binary_buffer_encode(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_RecordField *__pyx_v_field = 0;
  PyObject *__pyx_v_rec_data = 0;
  struct __pyx_obj_5cavro_Record *__pyx_v_rec = 0;
  Py_ssize_t __pyx_v_index;
  PyObject *__pyx_v_field_value = NULL;
  PyObject *__pyx_v_extra_fields = NULL;
  PyObject *__pyx_v_extra_field = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_missing_fields = NULL;
  PyObject *__pyx_v_missing_field = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_prefix = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_t_18;
  char const *__pyx_t_19;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  char const *__pyx_t_26;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 1);

  /* "src/record.pxi":381
 *         cdef list rec_data
 *         cdef Record rec
 *         cdef Py_ssize_t index = 0             # <<<<<<<<<<<<<<
 *         if isinstance(value, Record):
 *             if isinstance(value, self.record):
 */
  __pyx_v_index = 0;

  /* "src/record.pxi":382
 *         cdef Record rec
 *         cdef Py_ssize_t index = 0
 *         if isinstance(value, Record):             # <<<<<<<<<<<<<<
 *             if isinstance(value, self.record):
 *                 rec = value
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_5cavro_Record); 
  if (__pyx_t_1) {

    /* "src/record.pxi":383
 *         cdef Py_ssize_t index = 0
 *         if isinstance(value, Record):
 *             if isinstance(value, self.record):             # <<<<<<<<<<<<<<
 *                 rec = value
 *             else:
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_self->record);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_value, __pyx_t_2); 
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_1) {

      /* "src/record.pxi":384
 *         if isinstance(value, Record):
 *             if isinstance(value, self.record):
 *                 rec = value             # <<<<<<<<<<<<<<
 *             else:
 *                 rec = self.record(value)
 */
      if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5cavro_Record))))) __PYX_ERR(11, 384, __pyx_L1_error)
      __pyx_t_2 = __pyx_v_value;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_v_rec = ((struct __pyx_obj_5cavro_Record *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "src/record.pxi":383
 *         cdef Py_ssize_t index = 0
 *         if isinstance(value, Record):
 *             if isinstance(value, self.record):             # <<<<<<<<<<<<<<
 *                 rec = value
 *             else:
 */
      goto __pyx_L4;
    }

    /* "src/record.pxi":386
 *                 rec = value
 *             else:
 *                 rec = self.record(value)             # <<<<<<<<<<<<<<
 *             rec_data = rec.data
 *             for field in self.fields:
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_v_self->record), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 386, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_Record))))) __PYX_ERR(11, 386, __pyx_L1_error)
      __pyx_v_rec = ((struct __pyx_obj_5cavro_Record *)__pyx_t_2);
      __pyx_t_2 = 0;
    }
    __pyx_L4:;

    /* "src/record.pxi":387
 *             else:
 *                 rec = self.record(value)
 *             rec_data = rec.data             # <<<<<<<<<<<<<<
 *             for field in self.fields:
 *                 field_value = rec_data[index]
 */
    __pyx_t_2 = __pyx_v_rec->data;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_rec_data = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/record.pxi":388
 *                 rec = self.record(value)
 *             rec_data = rec.data
 *             for field in self.fields:             # <<<<<<<<<<<<<<
 *                 field_value = rec_data[index]
 *                 if field_value is NO_DEFAULT:
 */
    if (unlikely(__pyx_v_self->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(11, 388, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_v_self->fields; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 388, __pyx_L1_error)
        #endif
        if (__pyx_t_3 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(11, 388, __pyx_L1_error)
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 388, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "src/record.pxi":389
 *             rec_data = rec.data
 *             for field in self.fields:
 *                 field_value = rec_data[index]             # <<<<<<<<<<<<<<
 *                 if field_value is NO_DEFAULT:
 *                     raise ValueError(f"required field '{field.name}' missing")
 */
      if (unlikely(__pyx_v_rec_data == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(11, 389, __pyx_L1_error)
      }
      __pyx_t_4 = __Pyx_GetItemInt_List(__pyx_v_rec_data, __pyx_v_index, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_field_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "src/record.pxi":390
 *             for field in self.fields:
 *                 field_value = rec_data[index]
 *                 if field_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                     raise ValueError(f"required field '{field.name}' missing")
 *                 field.type.binary_buffer_encode(buffer, field_value)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 390, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = (__pyx_v_field_value == __pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(__pyx_t_1)) {

        /* "src/record.pxi":391
 *                 field_value = rec_data[index]
 *                 if field_value is NO_DEFAULT:
 *                     raise ValueError(f"required field '{field.name}' missing")             # <<<<<<<<<<<<<<
 *                 field.type.binary_buffer_encode(buffer, field_value)
 *                 index = index + 1
 */
        __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 391, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = 0;
        __pyx_t_6 = 127;
        __Pyx_INCREF(__pyx_kp_u_required_field);
        __pyx_t_5 += 16;
        __Pyx_GIVEREF(__pyx_kp_u_required_field);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_required_field);
        __pyx_t_7 = __Pyx_PyUnicode_Unicode(__pyx_v_field->name); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 391, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
        __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_7);
        __pyx_t_7 = 0;
        __Pyx_INCREF(__pyx_kp_u_missing_2);
        __pyx_t_5 += 9;
        __Pyx_GIVEREF(__pyx_kp_u_missing_2);
        PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_missing_2);
        __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_4, 3, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 391, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 391, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(11, 391, __pyx_L1_error)

        /* "src/record.pxi":390
 *             for field in self.fields:
 *                 field_value = rec_data[index]
 *                 if field_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                     raise ValueError(f"required field '{field.name}' missing")
 *                 field.type.binary_buffer_encode(buffer, field_value)
 */
      }

      /* "src/record.pxi":392
 *                 if field_value is NO_DEFAULT:
 *                     raise ValueError(f"required field '{field.name}' missing")
 *                 field.type.binary_buffer_encode(buffer, field_value)             # <<<<<<<<<<<<<<
 *                 index = index + 1
 *             return 0
 */
      __pyx_t_8 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_field->type->__pyx_vtab)->binary_buffer_encode(__pyx_v_field->type, __pyx_v_buffer, __pyx_v_field_value); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(11, 392, __pyx_L1_error)

      /* "src/record.pxi":393
 *                     raise ValueError(f"required field '{field.name}' missing")
 *                 field.type.binary_buffer_encode(buffer, field_value)
 *                 index = index + 1             # <<<<<<<<<<<<<<
 *             return 0
 *         if not self.options.record_can_encode_dict:
 */
      __pyx_v_index = (__pyx_v_index + 1);

      /* "src/record.pxi":388
 *                 rec = self.record(value)
 *             rec_data = rec.data
 *             for field in self.fields:             # <<<<<<<<<<<<<<
 *                 field_value = rec_data[index]
 *                 if field_value is NO_DEFAULT:
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/record.pxi":394
 *                 field.type.binary_buffer_encode(buffer, field_value)
 *                 index = index + 1
 *             return 0             # <<<<<<<<<<<<<<
 *         if not self.options.record_can_encode_dict:
 *             raise InvalidValue(value, self, (self.name, ))
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "src/record.pxi":382
 *         cdef Record rec
 *         cdef Py_ssize_t index = 0
 *         if isinstance(value, Record):             # <<<<<<<<<<<<<<
 *             if isinstance(value, self.record):
 *                 rec = value
 */
  }

  /* "src/record.pxi":395
 *                 index = index + 1
 *             return 0
 *         if not self.options.record_can_encode_dict:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self, (self.name, ))
 *         if not self.options.record_allow_extra_fields:
 */
  __pyx_t_1 = (!__pyx_v_self->__pyx_base.__pyx_base.options->record_can_encode_dict);
  if (unlikely(__pyx_t_1)) {

    /* "src/record.pxi":396
 *             return 0
 *         if not self.options.record_can_encode_dict:
 *             raise InvalidValue(value, self, (self.name, ))             # <<<<<<<<<<<<<<
 *         if not self.options.record_allow_extra_fields:
 *             extra_fields = value.keys() - self.field_dict.keys()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_self->__pyx_base.name);
    __Pyx_GIVEREF(__pyx_v_self->__pyx_base.name);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_self->__pyx_base.name)) __PYX_ERR(11, 396, __pyx_L1_error);
    __pyx_t_9 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_9, __pyx_v_value, ((PyObject *)__pyx_v_self), __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 3+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(11, 396, __pyx_L1_error)

    /* "src/record.pxi":395
 *                 index = index + 1
 *             return 0
 *         if not self.options.record_can_encode_dict:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self, (self.name, ))
 *         if not self.options.record_allow_extra_fields:
 */
  }

  /* "src/record.pxi":397
 *         if not self.options.record_can_encode_dict:
 *             raise InvalidValue(value, self, (self.name, ))
 *         if not self.options.record_allow_extra_fields:             # <<<<<<<<<<<<<<
 *             extra_fields = value.keys() - self.field_dict.keys()
 *             if extra_fields:
 */
  __pyx_t_1 = (!__pyx_v_self->__pyx_base.__pyx_base.options->record_allow_extra_fields);
  if (__pyx_t_1) {

    /* "src/record.pxi":398
 *             raise InvalidValue(value, self, (self.name, ))
 *         if not self.options.record_allow_extra_fields:
 *             extra_fields = value.keys() - self.field_dict.keys()             # <<<<<<<<<<<<<<
 *             if extra_fields:
 *                 extra_field, *_ = extra_fields
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_keys); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    if (unlikely(__pyx_v_self->field_dict == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
      __PYX_ERR(11, 398, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyDict_Keys(__pyx_v_self->field_dict); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = PyNumber_Subtract(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_extra_fields = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "src/record.pxi":399
 *         if not self.options.record_allow_extra_fields:
 *             extra_fields = value.keys() - self.field_dict.keys()
 *             if extra_fields:             # <<<<<<<<<<<<<<
 *                 extra_field, *_ = extra_fields
 *                 raise InvalidValue('...', self, (extra_field, ))
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_extra_fields); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(11, 399, __pyx_L1_error)
    if (unlikely(__pyx_t_1)) {

      /* "src/record.pxi":400
 *             extra_fields = value.keys() - self.field_dict.keys()
 *             if extra_fields:
 *                 extra_field, *_ = extra_fields             # <<<<<<<<<<<<<<
 *                 raise InvalidValue('...', self, (extra_field, ))
 *         if not self.options.record_encode_use_defaults:
 */
      {
        Py_ssize_t index = -1;
        PyObject** temps[2] = {&__pyx_t_7};
        __pyx_t_4 = PyObject_GetIter(__pyx_v_extra_fields); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 400, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
        for (index=0; index < 1; index++) {
          PyObject* item = __pyx_t_10(__pyx_t_4); if (unlikely(!item)) goto __pyx_L12_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        goto __pyx_L13_unpacking_done;
        __pyx_L12_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(11, 400, __pyx_L1_error)
        __pyx_L13_unpacking_done:;
      }
      __pyx_t_2 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 400, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_extra_field = __pyx_t_7;
      __pyx_t_7 = 0;
      __pyx_v__ = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "src/record.pxi":401
 *             if extra_fields:
 *                 extra_field, *_ = extra_fields
 *                 raise InvalidValue('...', self, (extra_field, ))             # <<<<<<<<<<<<<<
 *         if not self.options.record_encode_use_defaults:
 *             missing_fields = self.field_dict.keys() - value.keys()
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 401, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 401, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_extra_field);
      __Pyx_GIVEREF(__pyx_v_extra_field);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_extra_field)) __PYX_ERR(11, 401, __pyx_L1_error);
      __pyx_t_9 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_9, __pyx_kp_u__18, ((PyObject *)__pyx_v_self), __pyx_t_4};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 3+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(11, 401, __pyx_L1_error)

      /* "src/record.pxi":399
 *         if not self.options.record_allow_extra_fields:
 *             extra_fields = value.keys() - self.field_dict.keys()
 *             if extra_fields:             # <<<<<<<<<<<<<<
 *                 extra_field, *_ = extra_fields
 *                 raise InvalidValue('...', self, (extra_field, ))
 */
    }

    /* "src/record.pxi":397
 *         if not self.options.record_can_encode_dict:
 *             raise InvalidValue(value, self, (self.name, ))
 *         if not self.options.record_allow_extra_fields:             # <<<<<<<<<<<<<<
 *             extra_fields = value.keys() - self.field_dict.keys()
 *             if extra_fields:
 */
  }

  /* "src/record.pxi":402
 *                 extra_field, *_ = extra_fields
 *                 raise InvalidValue('...', self, (extra_field, ))
 *         if not self.options.record_encode_use_defaults:             # <<<<<<<<<<<<<<
 *             missing_fields = self.field_dict.keys() - value.keys()
 *             if missing_fields:
 */
  __pyx_t_1 = (!__pyx_v_self->__pyx_base.__pyx_base.options->record_encode_use_defaults);
  if (__pyx_t_1) {

    /* "src/record.pxi":403
 *                 raise InvalidValue('...', self, (extra_field, ))
 *         if not self.options.record_encode_use_defaults:
 *             missing_fields = self.field_dict.keys() - value.keys()             # <<<<<<<<<<<<<<
 *             if missing_fields:
 *                 missing_field, *_ = missing_fields
 */
    if (unlikely(__pyx_v_self->field_dict == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
      __PYX_ERR(11, 403, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyDict_Keys(__pyx_v_self->field_dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_keys); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 403, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = PyNumber_Subtract(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_missing_fields = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "src/record.pxi":404
 *         if not self.options.record_encode_use_defaults:
 *             missing_fields = self.field_dict.keys() - value.keys()
 *             if missing_fields:             # <<<<<<<<<<<<<<
 *                 missing_field, *_ = missing_fields
 *                 raise InvalidValue('<missing>', self, (missing_field, ))
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_missing_fields); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(11, 404, __pyx_L1_error)
    if (unlikely(__pyx_t_1)) {

      /* "src/record.pxi":405
 *             missing_fields = self.field_dict.keys() - value.keys()
 *             if missing_fields:
 *                 missing_field, *_ = missing_fields             # <<<<<<<<<<<<<<
 *                 raise InvalidValue('<missing>', self, (missing_field, ))
 *         for field in self.fields:
 */
      {
        Py_ssize_t index = -1;
        PyObject** temps[2] = {&__pyx_t_4};
        __pyx_t_7 = PyObject_GetIter(__pyx_v_missing_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 405, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
        for (index=0; index < 1; index++) {
          PyObject* item = __pyx_t_10(__pyx_t_7); if (unlikely(!item)) goto __pyx_L16_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        goto __pyx_L17_unpacking_done;
        __pyx_L16_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(11, 405, __pyx_L1_error)
        __pyx_L17_unpacking_done:;
      }
      __pyx_t_2 = PySequence_List(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 405, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_missing_field = __pyx_t_4;
      __pyx_t_4 = 0;
      __pyx_v__ = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "src/record.pxi":406
 *             if missing_fields:
 *                 missing_field, *_ = missing_fields
 *                 raise InvalidValue('<missing>', self, (missing_field, ))             # <<<<<<<<<<<<<<
 *         for field in self.fields:
 *             try:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_v_missing_field);
      __Pyx_GIVEREF(__pyx_v_missing_field);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_missing_field)) __PYX_ERR(11, 406, __pyx_L1_error);
      __pyx_t_9 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_9, __pyx_kp_u_missing, ((PyObject *)__pyx_v_self), __pyx_t_7};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 3+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 406, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(11, 406, __pyx_L1_error)

      /* "src/record.pxi":404
 *         if not self.options.record_encode_use_defaults:
 *             missing_fields = self.field_dict.keys() - value.keys()
 *             if missing_fields:             # <<<<<<<<<<<<<<
 *                 missing_field, *_ = missing_fields
 *                 raise InvalidValue('<missing>', self, (missing_field, ))
 */
    }

    /* "src/record.pxi":402
 *                 extra_field, *_ = extra_fields
 *                 raise InvalidValue('...', self, (extra_field, ))
 *         if not self.options.record_encode_use_defaults:             # <<<<<<<<<<<<<<
 *             missing_fields = self.field_dict.keys() - value.keys()
 *             if missing_fields:
 */
  }

  /* "src/record.pxi":407
 *                 missing_field, *_ = missing_fields
 *                 raise InvalidValue('<missing>', self, (missing_field, ))
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             try:
 *                 field_value = value.get(field.name, field.default_value)
 */
  if (unlikely(__pyx_v_self->fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(11, 407, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_self->fields; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 407, __pyx_L1_error)
      #endif
      if (__pyx_t_3 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(11, 407, __pyx_L1_error)
    #else
    __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 407, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "src/record.pxi":408
 *                 raise InvalidValue('<missing>', self, (missing_field, ))
 *         for field in self.fields:
 *             try:             # <<<<<<<<<<<<<<
 *                 field_value = value.get(field.name, field.default_value)
 *             except AttributeError as e:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      /*try:*/ {

        /* "src/record.pxi":409
 *         for field in self.fields:
 *             try:
 *                 field_value = value.get(field.name, field.default_value)             # <<<<<<<<<<<<<<
 *             except AttributeError as e:
 *                 raise InvalidValue(value, self, (self.name, )) from e
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_get); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 409, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_9 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_9, __pyx_v_field->name, __pyx_v_field->default_value};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 409, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_XDECREF_SET(__pyx_v_field_value, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "src/record.pxi":408
 *                 raise InvalidValue('<missing>', self, (missing_field, ))
 *         for field in self.fields:
 *             try:             # <<<<<<<<<<<<<<
 *                 field_value = value.get(field.name, field.default_value)
 *             except AttributeError as e:
 */
      }
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      goto __pyx_L27_try_end;
      __pyx_L20_error:;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "src/record.pxi":410
 *             try:
 *                 field_value = value.get(field.name, field.default_value)
 *             except AttributeError as e:             # <<<<<<<<<<<<<<
 *                 raise InvalidValue(value, self, (self.name, )) from e
 *             if field_value is NO_DEFAULT:
 */
      __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
      if (__pyx_t_8) {
        __Pyx_AddTraceback("cavro.RecordType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_7, &__pyx_t_9) < 0) __PYX_ERR(11, 410, __pyx_L22_except_error)
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_e = __pyx_t_7;
        /*try:*/ {

          /* "src/record.pxi":411
 *                 field_value = value.get(field.name, field.default_value)
 *             except AttributeError as e:
 *                 raise InvalidValue(value, self, (self.name, )) from e             # <<<<<<<<<<<<<<
 *             if field_value is NO_DEFAULT:
 *                 raise ValueError(f"required field '{field.name}' missing")
 */
          __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_15)) __PYX_ERR(11, 411, __pyx_L33_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_16 = PyTuple_New(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(11, 411, __pyx_L33_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_INCREF(__pyx_v_self->__pyx_base.name);
          __Pyx_GIVEREF(__pyx_v_self->__pyx_base.name);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_self->__pyx_base.name)) __PYX_ERR(11, 411, __pyx_L33_error);
          __pyx_t_17 = NULL;
          __pyx_t_8 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_15))) {
            __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_15);
            if (likely(__pyx_t_17)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
              __Pyx_INCREF(__pyx_t_17);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_15, function);
              __pyx_t_8 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[4] = {__pyx_t_17, __pyx_v_value, ((PyObject *)__pyx_v_self), __pyx_t_16};
            __pyx_t_14 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_8, 3+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (unlikely(!__pyx_t_14)) __PYX_ERR(11, 411, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          __Pyx_Raise(__pyx_t_14, 0, 0, __pyx_v_e);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __PYX_ERR(11, 411, __pyx_L33_error)
        }

        /* "src/record.pxi":410
 *             try:
 *                 field_value = value.get(field.name, field.default_value)
 *             except AttributeError as e:             # <<<<<<<<<<<<<<
 *                 raise InvalidValue(value, self, (self.name, )) from e
 *             if field_value is NO_DEFAULT:
 */
        /*finally:*/ {
          __pyx_L33_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_23, &__pyx_t_24, &__pyx_t_25);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22) < 0)) __Pyx_ErrFetch(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_23);
            __Pyx_XGOTREF(__pyx_t_24);
            __Pyx_XGOTREF(__pyx_t_25);
            __pyx_t_8 = __pyx_lineno; __pyx_t_18 = __pyx_clineno; __pyx_t_19 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_23);
              __Pyx_XGIVEREF(__pyx_t_24);
              __Pyx_XGIVEREF(__pyx_t_25);
              __Pyx_ExceptionReset(__pyx_t_23, __pyx_t_24, __pyx_t_25);
            }
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_ErrRestore(__pyx_t_20, __pyx_t_21, __pyx_t_22);
            __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0;
            __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_18; __pyx_filename = __pyx_t_19;
            goto __pyx_L22_except_error;
          }
        }
      }
      goto __pyx_L22_except_error;

      /* "src/record.pxi":408
 *                 raise InvalidValue('<missing>', self, (missing_field, ))
 *         for field in self.fields:
 *             try:             # <<<<<<<<<<<<<<
 *                 field_value = value.get(field.name, field.default_value)
 *             except AttributeError as e:
 */
      __pyx_L22_except_error:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      goto __pyx_L1_error;
      __pyx_L27_try_end:;
    }

    /* "src/record.pxi":412
 *             except AttributeError as e:
 *                 raise InvalidValue(value, self, (self.name, )) from e
 *             if field_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"required field '{field.name}' missing")
 *             try:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = (__pyx_v_field_value == __pyx_t_9);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "src/record.pxi":413
 *                 raise InvalidValue(value, self, (self.name, )) from e
 *             if field_value is NO_DEFAULT:
 *                 raise ValueError(f"required field '{field.name}' missing")             # <<<<<<<<<<<<<<
 *             try:
 *                 field.type.binary_buffer_encode(buffer, field_value)
 */
      __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_5 = 0;
      __pyx_t_6 = 127;
      __Pyx_INCREF(__pyx_kp_u_required_field);
      __pyx_t_5 += 16;
      __Pyx_GIVEREF(__pyx_kp_u_required_field);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_kp_u_required_field);
      __pyx_t_7 = __Pyx_PyUnicode_Unicode(__pyx_v_field->name); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_INCREF(__pyx_kp_u_missing_2);
      __pyx_t_5 += 9;
      __Pyx_GIVEREF(__pyx_kp_u_missing_2);
      PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_kp_u_missing_2);
      __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_9, 3, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(11, 413, __pyx_L1_error)

      /* "src/record.pxi":412
 *             except AttributeError as e:
 *                 raise InvalidValue(value, self, (self.name, )) from e
 *             if field_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"required field '{field.name}' missing")
 *             try:
 */
    }

    /* "src/record.pxi":414
 *             if field_value is NO_DEFAULT:
 *                 raise ValueError(f"required field '{field.name}' missing")
 *             try:             # <<<<<<<<<<<<<<
 *                 field.type.binary_buffer_encode(buffer, field_value)
 *             except InvalidValue as e:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_12, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_11);
      /*try:*/ {

        /* "src/record.pxi":415
 *                 raise ValueError(f"required field '{field.name}' missing")
 *             try:
 *                 field.type.binary_buffer_encode(buffer, field_value)             # <<<<<<<<<<<<<<
 *             except InvalidValue as e:
 *                 prefix = (field.name, )
 */
        __pyx_t_18 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_field->type->__pyx_vtab)->binary_buffer_encode(__pyx_v_field->type, __pyx_v_buffer, __pyx_v_field_value); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(11, 415, __pyx_L40_error)

        /* "src/record.pxi":414
 *             if field_value is NO_DEFAULT:
 *                 raise ValueError(f"required field '{field.name}' missing")
 *             try:             # <<<<<<<<<<<<<<
 *                 field.type.binary_buffer_encode(buffer, field_value)
 *             except InvalidValue as e:
 */
      }
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L47_try_end;
      __pyx_L40_error:;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "src/record.pxi":416
 *             try:
 *                 field.type.binary_buffer_encode(buffer, field_value)
 *             except InvalidValue as e:             # <<<<<<<<<<<<<<
 *                 prefix = (field.name, )
 *                 if self.options.invalid_value_includes_record_name:
 */
      __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_7, &__pyx_t_4);
      __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_14)) __PYX_ERR(11, 416, __pyx_L42_except_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_18 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_9, __pyx_t_14);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_7, __pyx_t_4);
      __pyx_t_9 = 0; __pyx_t_7 = 0; __pyx_t_4 = 0;
      if (__pyx_t_18) {
        __Pyx_AddTraceback("cavro.RecordType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_7, &__pyx_t_9) < 0) __PYX_ERR(11, 416, __pyx_L42_except_error)
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_e = __pyx_t_7;
        /*try:*/ {

          /* "src/record.pxi":417
 *                 field.type.binary_buffer_encode(buffer, field_value)
 *             except InvalidValue as e:
 *                 prefix = (field.name, )             # <<<<<<<<<<<<<<
 *                 if self.options.invalid_value_includes_record_name:
 *                     prefix = (self.type, ) + prefix
 */
          __pyx_t_14 = PyTuple_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(11, 417, __pyx_L53_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_INCREF(__pyx_v_field->name);
          __Pyx_GIVEREF(__pyx_v_field->name);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_v_field->name)) __PYX_ERR(11, 417, __pyx_L53_error);
          __pyx_v_prefix = ((PyObject*)__pyx_t_14);
          __pyx_t_14 = 0;

          /* "src/record.pxi":418
 *             except InvalidValue as e:
 *                 prefix = (field.name, )
 *                 if self.options.invalid_value_includes_record_name:             # <<<<<<<<<<<<<<
 *                     prefix = (self.type, ) + prefix
 *                 e.schema_path = prefix + e.schema_path
 */
          if (__pyx_v_self->__pyx_base.__pyx_base.options->invalid_value_includes_record_name) {

            /* "src/record.pxi":419
 *                 prefix = (field.name, )
 *                 if self.options.invalid_value_includes_record_name:
 *                     prefix = (self.type, ) + prefix             # <<<<<<<<<<<<<<
 *                 e.schema_path = prefix + e.schema_path
 *                 raise
 */
            __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_14)) __PYX_ERR(11, 419, __pyx_L53_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(11, 419, __pyx_L53_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_14);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_14)) __PYX_ERR(11, 419, __pyx_L53_error);
            __pyx_t_14 = 0;
            __pyx_t_14 = PyNumber_Add(__pyx_t_15, __pyx_v_prefix); if (unlikely(!__pyx_t_14)) __PYX_ERR(11, 419, __pyx_L53_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF_SET(__pyx_v_prefix, ((PyObject*)__pyx_t_14));
            __pyx_t_14 = 0;

            /* "src/record.pxi":418
 *             except InvalidValue as e:
 *                 prefix = (field.name, )
 *                 if self.options.invalid_value_includes_record_name:             # <<<<<<<<<<<<<<
 *                     prefix = (self.type, ) + prefix
 *                 e.schema_path = prefix + e.schema_path
 */
          }

          /* "src/record.pxi":420
 *                 if self.options.invalid_value_includes_record_name:
 *                     prefix = (self.type, ) + prefix
 *                 e.schema_path = prefix + e.schema_path             # <<<<<<<<<<<<<<
 *                 raise
 * 
 */
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_schema_path); if (unlikely(!__pyx_t_14)) __PYX_ERR(11, 420, __pyx_L53_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_15 = PyNumber_Add(__pyx_v_prefix, __pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(11, 420, __pyx_L53_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_e, __pyx_n_s_schema_path, __pyx_t_15) < 0) __PYX_ERR(11, 420, __pyx_L53_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "src/record.pxi":421
 *                     prefix = (self.type, ) + prefix
 *                 e.schema_path = prefix + e.schema_path
 *                 raise             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode_record(self, _Reader buffer):
 */
          __Pyx_GIVEREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_7, __pyx_t_9);
          __pyx_t_4 = 0; __pyx_t_7 = 0; __pyx_t_9 = 0; 
          __PYX_ERR(11, 421, __pyx_L53_error)
        }

        /* "src/record.pxi":416
 *             try:
 *                 field.type.binary_buffer_encode(buffer, field_value)
 *             except InvalidValue as e:             # <<<<<<<<<<<<<<
 *                 prefix = (field.name, )
 *                 if self.options.invalid_value_includes_record_name:
 */
        /*finally:*/ {
          __pyx_L53_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_25 = 0; __pyx_t_24 = 0; __pyx_t_23 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_21, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_25, &__pyx_t_24, &__pyx_t_23) < 0)) __Pyx_ErrFetch(&__pyx_t_25, &__pyx_t_24, &__pyx_t_23);
            __Pyx_XGOTREF(__pyx_t_25);
            __Pyx_XGOTREF(__pyx_t_24);
            __Pyx_XGOTREF(__pyx_t_23);
            __Pyx_XGOTREF(__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_18 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_26 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_22);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_21, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_25);
            __Pyx_XGIVEREF(__pyx_t_24);
            __Pyx_XGIVEREF(__pyx_t_23);
            __Pyx_ErrRestore(__pyx_t_25, __pyx_t_24, __pyx_t_23);
            __pyx_t_25 = 0; __pyx_t_24 = 0; __pyx_t_23 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_18; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_26;
            goto __pyx_L42_except_error;
          }
        }
      }
      goto __pyx_L42_except_error;

      /* "src/record.pxi":414
 *             if field_value is NO_DEFAULT:
 *                 raise ValueError(f"required field '{field.name}' missing")
 *             try:             # <<<<<<<<<<<<<<
 *                 field.type.binary_buffer_encode(buffer, field_value)
 *             except InvalidValue as e:
 */
      __pyx_L42_except_error:;
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_12, __pyx_t_11);
      goto __pyx_L1_error;
      __pyx_L47_try_end:;
    }

    /* "src/record.pxi":407
 *                 missing_field, *_ = missing_fields
 *                 raise InvalidValue('<missing>', self, (missing_field, ))
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             try:
 *                 field_value = value.get(field.name, field.default_value)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/record.pxi":377
 *         return dict(extra, fields=[f.get_schema(created) for f in self.fields])
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         cdef list rec_data
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("cavro.RecordType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_rec_data);
  __Pyx_XDECREF((PyObject *)__pyx_v_rec);
  __Pyx_XDECREF(__pyx_v_field_value);
  __Pyx_XDECREF(__pyx_v_extra_fields);
  __Pyx_XDECREF(__pyx_v_extra_field);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_missing_fields);
  __Pyx_XDECREF(__pyx_v_missing_field);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_prefix);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":423
 *                 raise
 * 
 *     cdef _binary_buffer_decode_record(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         cdef list data = [None] * len(self.fields)
 */

static PyObject *__pyx_f_5cavro_10RecordType__binary_buffer_decode_record(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  struct __pyx_obj_5cavro_RecordField *__pyx_v_field = 0;
  PyObject *__pyx_v_data = 0;
  struct __pyx_obj_5cavro_Record *__pyx_v_rec = 0;
  Py_ssize_t __pyx_v_index;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode_record", 1);

  /* "src/record.pxi":425
 *     cdef _binary_buffer_decode_record(self, _Reader buffer):
 *         cdef RecordField field
 *         cdef list data = [None] * len(self.fields)             # <<<<<<<<<<<<<<
 *         cdef Record rec
 *         cdef Py_ssize_t index = 0
 */
  __pyx_t_1 = __pyx_v_self->fields;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(11, 425, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(11, 425, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(1 * ((__pyx_t_2<0) ? 0:__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_t_2; __pyx_temp++) {
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_1, __pyx_temp, Py_None)) __PYX_ERR(11, 425, __pyx_L1_error);
    }
  }
  __pyx_v_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":427
 *         cdef list data = [None] * len(self.fields)
 *         cdef Record rec
 *         cdef Py_ssize_t index = 0             # <<<<<<<<<<<<<<
 *         for field in self.fields:
 *             data[index] = field.type.binary_buffer_decode(buffer)
 */
  __pyx_v_index = 0;

  /* "src/record.pxi":428
 *         cdef Record rec
 *         cdef Py_ssize_t index = 0
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             data[index] = field.type.binary_buffer_decode(buffer)
 *             index += 1
 */
  if (unlikely(__pyx_v_self->fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(11, 428, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->fields; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 428, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(11, 428, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 428, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/record.pxi":429
 *         cdef Py_ssize_t index = 0
 *         for field in self.fields:
 *             data[index] = field.type.binary_buffer_decode(buffer)             # <<<<<<<<<<<<<<
 *             index += 1
 *         rec = Record.__new__(self.record)
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_field->type->__pyx_vtab)->binary_buffer_decode(__pyx_v_field->type, __pyx_v_buffer); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely((__Pyx_SetItemInt(__pyx_v_data, __pyx_v_index, __pyx_t_3, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1) < 0))) __PYX_ERR(11, 429, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "src/record.pxi":430
 *         for field in self.fields:
 *             data[index] = field.type.binary_buffer_decode(buffer)
 *             index += 1             # <<<<<<<<<<<<<<
 *         rec = Record.__new__(self.record)
 *         rec.data = data
 */
    __pyx_v_index = (__pyx_v_index + 1);

    /* "src/record.pxi":428
 *         cdef Record rec
 *         cdef Py_ssize_t index = 0
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             data[index] = field.type.binary_buffer_decode(buffer)
 *             index += 1
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/record.pxi":431
 *             data[index] = field.type.binary_buffer_decode(buffer)
 *             index += 1
 *         rec = Record.__new__(self.record)             # <<<<<<<<<<<<<<
 *         rec.data = data
 *         return rec
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_Record), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_self->record)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Record))))) __PYX_ERR(11, 431, __pyx_L1_error)
  __pyx_v_rec = ((struct __pyx_obj_5cavro_Record *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":432
 *             index += 1
 *         rec = Record.__new__(self.record)
 *         rec.data = data             # <<<<<<<<<<<<<<
 *         return rec
 * 
 */
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  __Pyx_GOTREF(__pyx_v_rec->data);
  __Pyx_DECREF(__pyx_v_rec->data);
  __pyx_v_rec->data = __pyx_v_data;

  /* "src/record.pxi":433
 *         rec = Record.__new__(self.record)
 *         rec.data = data
 *         return rec             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode_dict(self, _Reader buffer):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_rec);
  __pyx_r = ((PyObject *)__pyx_v_rec);
  goto __pyx_L0;

  /* "src/record.pxi":423
 *                 raise
 * 
 *     cdef _binary_buffer_decode_record(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         cdef list data = [None] * len(self.fields)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.RecordType._binary_buffer_decode_record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF((PyObject *)__pyx_v_rec);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":435
 *         return rec
 * 
 *     cdef _binary_buffer_decode_dict(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         cdef dict data = {}
 */

static PyObject *__pyx_f_5cavro_10RecordType__binary_buffer_decode_dict(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  struct __pyx_obj_5cavro_RecordField *__pyx_v_field = 0;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode_dict", 1);

  /* "src/record.pxi":437
 *     cdef _binary_buffer_decode_dict(self, _Reader buffer):
 *         cdef RecordField field
 *         cdef dict data = {}             # <<<<<<<<<<<<<<
 *         for field in self.fields:
 *             value = field.type.binary_buffer_decode(buffer)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":438
 *         cdef RecordField field
 *         cdef dict data = {}
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             value = field.type.binary_buffer_decode(buffer)
 *             if field.name is not None:
 */
  if (unlikely(__pyx_v_self->fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(11, 438, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->fields; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 438, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(11, 438, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 438, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/record.pxi":439
 *         cdef dict data = {}
 *         for field in self.fields:
 *             value = field.type.binary_buffer_decode(buffer)             # <<<<<<<<<<<<<<
 *             if field.name is not None:
 *                 data[field.name] = value
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_field->type->__pyx_vtab)->binary_buffer_decode(__pyx_v_field->type, __pyx_v_buffer); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/record.pxi":440
 *         for field in self.fields:
 *             value = field.type.binary_buffer_decode(buffer)
 *             if field.name is not None:             # <<<<<<<<<<<<<<
 *                 data[field.name] = value
 *         return data
 */
    __pyx_t_4 = (__pyx_v_field->name != ((PyObject*)Py_None));
    if (__pyx_t_4) {

      /* "src/record.pxi":441
 *             value = field.type.binary_buffer_decode(buffer)
 *             if field.name is not None:
 *                 data[field.name] = value             # <<<<<<<<<<<<<<
 *         return data
 * 
 */
      if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_v_field->name, __pyx_v_value) < 0))) __PYX_ERR(11, 441, __pyx_L1_error)

      /* "src/record.pxi":440
 *         for field in self.fields:
 *             value = field.type.binary_buffer_decode(buffer)
 *             if field.name is not None:             # <<<<<<<<<<<<<<
 *                 data[field.name] = value
 *         return data
 */
    }

    /* "src/record.pxi":438
 *         cdef RecordField field
 *         cdef dict data = {}
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             value = field.type.binary_buffer_decode(buffer)
 *             if field.name is not None:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/record.pxi":442
 *             if field.name is not None:
 *                 data[field.name] = value
 *         return data             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_data);
  __pyx_r = __pyx_v_data;
  goto __pyx_L0;

  /* "src/record.pxi":435
 *         return rec
 * 
 *     cdef _binary_buffer_decode_dict(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         cdef dict data = {}
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.RecordType._binary_buffer_decode_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":444
 *         return data
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         if self.options.record_decodes_to_dict:
 *             return self._binary_buffer_decode_dict(buffer)
 */

static PyObject *__pyx_f_5cavro_10RecordType__binary_buffer_decode(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/record.pxi":445
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         if self.options.record_decodes_to_dict:             # <<<<<<<<<<<<<<
 *             return self._binary_buffer_decode_dict(buffer)
 *         return self._binary_buffer_decode_record(buffer)
 */
  if (__pyx_v_self->__pyx_base.__pyx_base.options->record_decodes_to_dict) {

    /* "src/record.pxi":446
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         if self.options.record_decodes_to_dict:
 *             return self._binary_buffer_decode_dict(buffer)             # <<<<<<<<<<<<<<
 *         return self._binary_buffer_decode_record(buffer)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_RecordType *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_binary_buffer_decode_dict(__pyx_v_self, __pyx_v_buffer); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/record.pxi":445
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         if self.options.record_decodes_to_dict:             # <<<<<<<<<<<<<<
 *             return self._binary_buffer_decode_dict(buffer)
 *         return self._binary_buffer_decode_record(buffer)
 */
  }

  /* "src/record.pxi":447
 *         if self.options.record_decodes_to_dict:
 *             return self._binary_buffer_decode_dict(buffer)
 *         return self._binary_buffer_decode_record(buffer)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_RecordType *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_binary_buffer_decode_record(__pyx_v_self, __pyx_v_buffer); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":444
 *         return data
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         if self.options.record_decodes_to_dict:
 *             return self._binary_buffer_decode_dict(buffer)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.RecordType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":449
 *         return self._binary_buffer_decode_record(buffer)
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef int level = FIT_OK
 *         cdef RecordField field
 */

static int __pyx_f_5cavro_10RecordType__get_value_fitness(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_v_level;
  struct __pyx_obj_5cavro_RecordField *__pyx_v_field = 0;
  PyObject *__pyx_v_type_val = NULL;
  PyObject *__pyx_v_remaining_keys = NULL;
  PyObject *__pyx_v_field_value = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/record.pxi":450
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         cdef int level = FIT_OK             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         if isinstance(value, self.record):
 */
  __pyx_v_level = __pyx_v_5cavro_FIT_OK;

  /* "src/record.pxi":452
 *         cdef int level = FIT_OK
 *         cdef RecordField field
 *         if isinstance(value, self.record):             # <<<<<<<<<<<<<<
 *             return FIT_EXACT
 *         if isinstance(value, Record):
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->record);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_value, __pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/record.pxi":453
 *         cdef RecordField field
 *         if isinstance(value, self.record):
 *             return FIT_EXACT             # <<<<<<<<<<<<<<
 *         if isinstance(value, Record):
 *             if value.Type.name != self.type:
 */
    __pyx_r = __pyx_v_5cavro_FIT_EXACT;
    goto __pyx_L0;

    /* "src/record.pxi":452
 *         cdef int level = FIT_OK
 *         cdef RecordField field
 *         if isinstance(value, self.record):             # <<<<<<<<<<<<<<
 *             return FIT_EXACT
 *         if isinstance(value, Record):
 */
  }

  /* "src/record.pxi":454
 *         if isinstance(value, self.record):
 *             return FIT_EXACT
 *         if isinstance(value, Record):             # <<<<<<<<<<<<<<
 *             if value.Type.name != self.type:
 *                 return FIT_NONE
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_5cavro_Record); 
  if (__pyx_t_2) {

    /* "src/record.pxi":455
 *             return FIT_EXACT
 *         if isinstance(value, Record):
 *             if value.Type.name != self.type:             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *             value = value._asdict()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_Type); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 455, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(11, 455, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {

      /* "src/record.pxi":456
 *         if isinstance(value, Record):
 *             if value.Type.name != self.type:
 *                 return FIT_NONE             # <<<<<<<<<<<<<<
 *             value = value._asdict()
 *         if isinstance(value, dict):
 */
      __pyx_r = __pyx_v_5cavro_FIT_NONE;
      goto __pyx_L0;

      /* "src/record.pxi":455
 *             return FIT_EXACT
 *         if isinstance(value, Record):
 *             if value.Type.name != self.type:             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *             value = value._asdict()
 */
    }

    /* "src/record.pxi":457
 *             if value.Type.name != self.type:
 *                 return FIT_NONE
 *             value = value._asdict()             # <<<<<<<<<<<<<<
 *         if isinstance(value, dict):
 *             if self.options.record_values_type_hint and '-type' in value:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_asdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/record.pxi":454
 *         if isinstance(value, self.record):
 *             return FIT_EXACT
 *         if isinstance(value, Record):             # <<<<<<<<<<<<<<
 *             if value.Type.name != self.type:
 *                 return FIT_NONE
 */
  }

  /* "src/record.pxi":458
 *                 return FIT_NONE
 *             value = value._asdict()
 *         if isinstance(value, dict):             # <<<<<<<<<<<<<<
 *             if self.options.record_values_type_hint and '-type' in value:
 *                 value = value.copy()
 */
  __pyx_t_2 = PyDict_Check(__pyx_v_value); 
  if (__pyx_t_2) {

    /* "src/record.pxi":459
 *             value = value._asdict()
 *         if isinstance(value, dict):
 *             if self.options.record_values_type_hint and '-type' in value:             # <<<<<<<<<<<<<<
 *                 value = value.copy()
 *                 type_val = value.pop('-type')
 */
    if (__pyx_v_self->__pyx_base.__pyx_base.options->record_values_type_hint) {
    } else {
      __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.options->record_values_type_hint;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_kp_u_type_2, __pyx_v_value, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(11, 459, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_6;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_2) {

      /* "src/record.pxi":460
 *         if isinstance(value, dict):
 *             if self.options.record_values_type_hint and '-type' in value:
 *                 value = value.copy()             # <<<<<<<<<<<<<<
 *                 type_val = value.pop('-type')
 *                 if type_val != self.type:
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 460, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 460, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "src/record.pxi":461
 *             if self.options.record_values_type_hint and '-type' in value:
 *                 value = value.copy()
 *                 type_val = value.pop('-type')             # <<<<<<<<<<<<<<
 *                 if type_val != self.type:
 *                     return FIT_NONE
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 461, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_type_2};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 461, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_v_type_val = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "src/record.pxi":462
 *                 value = value.copy()
 *                 type_val = value.pop('-type')
 *                 if type_val != self.type:             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 *             remaining_keys = set(value.keys())
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_type_val, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 462, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(11, 462, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {

        /* "src/record.pxi":463
 *                 type_val = value.pop('-type')
 *                 if type_val != self.type:
 *                     return FIT_NONE             # <<<<<<<<<<<<<<
 *             remaining_keys = set(value.keys())
 *             for field in self.fields:
 */
        __pyx_r = __pyx_v_5cavro_FIT_NONE;
        goto __pyx_L0;

        /* "src/record.pxi":462
 *                 value = value.copy()
 *                 type_val = value.pop('-type')
 *                 if type_val != self.type:             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 *             remaining_keys = set(value.keys())
 */
      }

      /* "src/record.pxi":459
 *             value = value._asdict()
 *         if isinstance(value, dict):
 *             if self.options.record_values_type_hint and '-type' in value:             # <<<<<<<<<<<<<<
 *                 value = value.copy()
 *                 type_val = value.pop('-type')
 */
    }

    /* "src/record.pxi":464
 *                 if type_val != self.type:
 *                     return FIT_NONE
 *             remaining_keys = set(value.keys())             # <<<<<<<<<<<<<<
 *             for field in self.fields:
 *                 field_value = value.get(field.name, field.default_value)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_keys); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 464, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_remaining_keys = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/record.pxi":465
 *                     return FIT_NONE
 *             remaining_keys = set(value.keys())
 *             for field in self.fields:             # <<<<<<<<<<<<<<
 *                 field_value = value.get(field.name, field.default_value)
 *                 if field_value is NO_DEFAULT:
 */
    if (unlikely(__pyx_v_self->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(11, 465, __pyx_L1_error)
    }
    __pyx_t_4 = __pyx_v_self->fields; __Pyx_INCREF(__pyx_t_4);
    __pyx_t_7 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 465, __pyx_L1_error)
        #endif
        if (__pyx_t_7 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(11, 465, __pyx_L1_error)
      #else
      __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 465, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 465, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "src/record.pxi":466
 *             remaining_keys = set(value.keys())
 *             for field in self.fields:
 *                 field_value = value.get(field.name, field.default_value)             # <<<<<<<<<<<<<<
 *                 if field_value is NO_DEFAULT:
 *                     return FIT_NONE
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 466, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_v_field->name, __pyx_v_field->default_value};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 466, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_XDECREF_SET(__pyx_v_field_value, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "src/record.pxi":467
 *             for field in self.fields:
 *                 field_value = value.get(field.name, field.default_value)
 *                 if field_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 *                 level = min(level, field.type.get_value_fitness(field_value))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 467, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = (__pyx_v_field_value == __pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {

        /* "src/record.pxi":468
 *                 field_value = value.get(field.name, field.default_value)
 *                 if field_value is NO_DEFAULT:
 *                     return FIT_NONE             # <<<<<<<<<<<<<<
 *                 level = min(level, field.type.get_value_fitness(field_value))
 *                 if level <= FIT_NONE:
 */
        __pyx_r = __pyx_v_5cavro_FIT_NONE;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L0;

        /* "src/record.pxi":467
 *             for field in self.fields:
 *                 field_value = value.get(field.name, field.default_value)
 *                 if field_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 *                 level = min(level, field.type.get_value_fitness(field_value))
 */
      }

      /* "src/record.pxi":469
 *                 if field_value is NO_DEFAULT:
 *                     return FIT_NONE
 *                 level = min(level, field.type.get_value_fitness(field_value))             # <<<<<<<<<<<<<<
 *                 if level <= FIT_NONE:
 *                     return level
 */
      __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_field->type->__pyx_vtab)->get_value_fitness(__pyx_v_field->type, __pyx_v_field_value); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(11, 469, __pyx_L1_error)
      __pyx_t_9 = __pyx_v_level;
      __pyx_t_2 = (__pyx_t_5 < __pyx_t_9);
      if (__pyx_t_2) {
        __pyx_t_10 = __pyx_t_5;
      } else {
        __pyx_t_10 = __pyx_t_9;
      }
      __pyx_v_level = __pyx_t_10;

      /* "src/record.pxi":470
 *                     return FIT_NONE
 *                 level = min(level, field.type.get_value_fitness(field_value))
 *                 if level <= FIT_NONE:             # <<<<<<<<<<<<<<
 *                     return level
 *                 remaining_keys.discard(field.name)
 */
      __pyx_t_2 = (__pyx_v_level <= __pyx_v_5cavro_FIT_NONE);
      if (__pyx_t_2) {

        /* "src/record.pxi":471
 *                 level = min(level, field.type.get_value_fitness(field_value))
 *                 if level <= FIT_NONE:
 *                     return level             # <<<<<<<<<<<<<<
 *                 remaining_keys.discard(field.name)
 *             if remaining_keys:
 */
        __pyx_r = __pyx_v_level;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L0;

        /* "src/record.pxi":470
 *                     return FIT_NONE
 *                 level = min(level, field.type.get_value_fitness(field_value))
 *                 if level <= FIT_NONE:             # <<<<<<<<<<<<<<
 *                     return level
 *                 remaining_keys.discard(field.name)
 */
      }

      /* "src/record.pxi":472
 *                 if level <= FIT_NONE:
 *                     return level
 *                 remaining_keys.discard(field.name)             # <<<<<<<<<<<<<<
 *             if remaining_keys:
 *                 if self.options.record_allow_extra_fields:
 */
      __pyx_t_1 = __pyx_v_field->name;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_11 = __Pyx_PySet_Discard(__pyx_v_remaining_keys, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(11, 472, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "src/record.pxi":465
 *                     return FIT_NONE
 *             remaining_keys = set(value.keys())
 *             for field in self.fields:             # <<<<<<<<<<<<<<
 *                 field_value = value.get(field.name, field.default_value)
 *                 if field_value is NO_DEFAULT:
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "src/record.pxi":473
 *                     return level
 *                 remaining_keys.discard(field.name)
 *             if remaining_keys:             # <<<<<<<<<<<<<<
 *                 if self.options.record_allow_extra_fields:
 *                     return FIT_POOR
 */
    __pyx_t_2 = (PySet_GET_SIZE(__pyx_v_remaining_keys) != 0);
    if (__pyx_t_2) {

      /* "src/record.pxi":474
 *                 remaining_keys.discard(field.name)
 *             if remaining_keys:
 *                 if self.options.record_allow_extra_fields:             # <<<<<<<<<<<<<<
 *                     return FIT_POOR
 *                 return FIT_NONE
 */
      if (__pyx_v_self->__pyx_base.__pyx_base.options->record_allow_extra_fields) {

        /* "src/record.pxi":475
 *             if remaining_keys:
 *                 if self.options.record_allow_extra_fields:
 *                     return FIT_POOR             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *             return level
 */
        __pyx_r = __pyx_v_5cavro_FIT_POOR;
        goto __pyx_L0;

        /* "src/record.pxi":474
 *                 remaining_keys.discard(field.name)
 *             if remaining_keys:
 *                 if self.options.record_allow_extra_fields:             # <<<<<<<<<<<<<<
 *                     return FIT_POOR
 *                 return FIT_NONE
 */
      }

      /* "src/record.pxi":476
 *                 if self.options.record_allow_extra_fields:
 *                     return FIT_POOR
 *                 return FIT_NONE             # <<<<<<<<<<<<<<
 *             return level
 * 
 */
      __pyx_r = __pyx_v_5cavro_FIT_NONE;
      goto __pyx_L0;

      /* "src/record.pxi":473
 *                     return level
 *                 remaining_keys.discard(field.name)
 *             if remaining_keys:             # <<<<<<<<<<<<<<
 *                 if self.options.record_allow_extra_fields:
 *                     return FIT_POOR
 */
    }

    /* "src/record.pxi":477
 *                     return FIT_POOR
 *                 return FIT_NONE
 *             return level             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
    __pyx_r = __pyx_v_level;
    goto __pyx_L0;

    /* "src/record.pxi":458
 *                 return FIT_NONE
 *             value = value._asdict()
 *         if isinstance(value, dict):             # <<<<<<<<<<<<<<
 *             if self.options.record_values_type_hint and '-type' in value:
 *                 value = value.copy()
 */
  }

  /* "src/record.pxi":449
 *         return self._binary_buffer_decode_record(buffer)
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef int level = FIT_OK
 *         cdef RecordField field
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cavro.RecordType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_type_val);
  __Pyx_XDECREF(__pyx_v_remaining_keys);
  __Pyx_XDECREF(__pyx_v_field_value);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":479
 *             return level
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         cdef Record record = self._convert_value(value)
 *         cdef AvroType field_type
 */

static PyObject *__pyx_f_5cavro_10RecordType__json_format(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_Record *__pyx_v_record = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_field_type = 0;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_field = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/record.pxi":480
 * 
 *     cdef _json_format(self, value):
 *         cdef Record record = self._convert_value(value)             # <<<<<<<<<<<<<<
 *         cdef AvroType field_type
 *         out = {}
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_RecordType *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Record))))) __PYX_ERR(11, 480, __pyx_L1_error)
  __pyx_v_record = ((struct __pyx_obj_5cavro_Record *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":482
 *         cdef Record record = self._convert_value(value)
 *         cdef AvroType field_type
 *         out = {}             # <<<<<<<<<<<<<<
 *         for field, value in zip(self.fields, record.data):
 *             field_type = field.type
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":483
 *         cdef AvroType field_type
 *         out = {}
 *         for field, value in zip(self.fields, record.data):             # <<<<<<<<<<<<<<
 *             field_type = field.type
 *             out[field.name] = field_type.json_format(value)
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->fields);
  __Pyx_GIVEREF(__pyx_v_self->fields);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->fields)) __PYX_ERR(11, 483, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_record->data);
  __Pyx_GIVEREF(__pyx_v_record->data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_record->data)) __PYX_ERR(11, 483, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 483, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 483, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(11, 483, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 483, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 483, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(11, 483, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 483, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(11, 483, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(11, 483, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(11, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(11, 483, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(11, 483, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_field, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "src/record.pxi":484
 *         out = {}
 *         for field, value in zip(self.fields, record.data):
 *             field_type = field.type             # <<<<<<<<<<<<<<
 *             out[field.name] = field_type.json_format(value)
 *         return out
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(11, 484, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_field_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "src/record.pxi":485
 *         for field, value in zip(self.fields, record.data):
 *             field_type = field.type
 *             out[field.name] = field_type.json_format(value)             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_field_type->__pyx_vtab)->json_format(__pyx_v_field_type, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_field, __pyx_n_s_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(11, 485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely((PyDict_SetItem(__pyx_v_out, __pyx_t_6, __pyx_t_2) < 0))) __PYX_ERR(11, 485, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/record.pxi":483
 *         cdef AvroType field_type
 *         out = {}
 *         for field, value in zip(self.fields, record.data):             # <<<<<<<<<<<<<<
 *             field_type = field.type
 *             out[field.name] = field_type.json_format(value)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/record.pxi":486
 *             field_type = field.type
 *             out[field.name] = field_type.json_format(value)
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     cdef json_decode_record(self, dict value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "src/record.pxi":479
 *             return level
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         cdef Record record = self._convert_value(value)
 *         cdef AvroType field_type
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.RecordType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_record);
  __Pyx_XDECREF((PyObject *)__pyx_v_field_type);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":488
 *         return out
 * 
 *     cdef json_decode_record(self, dict value):             # <<<<<<<<<<<<<<
 *         cdef list data = [None] * len(self.fields)
 *         cdef RecordField field
 */

static PyObject *__pyx_f_5cavro_10RecordType_json_decode_record(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_data = 0;
  struct __pyx_obj_5cavro_RecordField *__pyx_v_field = 0;
  struct __pyx_obj_5cavro_Record *__pyx_v_rec = 0;
  Py_ssize_t __pyx_v_index;
  PyObject *__pyx_v_field_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("json_decode_record", 1);

  /* "src/record.pxi":489
 * 
 *     cdef json_decode_record(self, dict value):
 *         cdef list data = [None] * len(self.fields)             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         cdef Record rec
 */
  __pyx_t_1 = __pyx_v_self->fields;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(11, 489, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(11, 489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(1 * ((__pyx_t_2<0) ? 0:__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_t_2; __pyx_temp++) {
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_1, __pyx_temp, Py_None)) __PYX_ERR(11, 489, __pyx_L1_error);
    }
  }
  __pyx_v_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":492
 *         cdef RecordField field
 *         cdef Record rec
 *         cdef Py_ssize_t index = 0             # <<<<<<<<<<<<<<
 * 
 *         for field in self.fields:
 */
  __pyx_v_index = 0;

  /* "src/record.pxi":494
 *         cdef Py_ssize_t index = 0
 * 
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             field_value = value.get(field.writer_name, MISSING_VALUE)
 *             if field_value is MISSING_VALUE:
 */
  if (unlikely(__pyx_v_self->fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(11, 494, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->fields; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 494, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(11, 494, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 494, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/record.pxi":495
 * 
 *         for field in self.fields:
 *             field_value = value.get(field.writer_name, MISSING_VALUE)             # <<<<<<<<<<<<<<
 *             if field_value is MISSING_VALUE:
 *                 if field.default_value is NO_DEFAULT:
 */
    if (unlikely(__pyx_v_value == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
      __PYX_ERR(11, 495, __pyx_L1_error)
    }
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyDict_GetItemDefault(__pyx_v_value, __pyx_v_field->writer_name, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_field_value, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/record.pxi":496
 *         for field in self.fields:
 *             field_value = value.get(field.writer_name, MISSING_VALUE)
 *             if field_value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *                 if field.default_value is NO_DEFAULT:
 *                     raise ValueError(f"required field '{field.name}' missing")
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = (__pyx_v_field_value == __pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "src/record.pxi":497
 *             field_value = value.get(field.writer_name, MISSING_VALUE)
 *             if field_value is MISSING_VALUE:
 *                 if field.default_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                     raise ValueError(f"required field '{field.name}' missing")
 *                 field_value = field.default_value
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 497, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = (__pyx_v_field->default_value == __pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(__pyx_t_5)) {

        /* "src/record.pxi":498
 *             if field_value is MISSING_VALUE:
 *                 if field.default_value is NO_DEFAULT:
 *                     raise ValueError(f"required field '{field.name}' missing")             # <<<<<<<<<<<<<<
 *                 field_value = field.default_value
 *             else:
 */
        __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = 0;
        __pyx_t_7 = 127;
        __Pyx_INCREF(__pyx_kp_u_required_field);
        __pyx_t_6 += 16;
        __Pyx_GIVEREF(__pyx_kp_u_required_field);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_required_field);
        __pyx_t_3 = __Pyx_PyUnicode_Unicode(__pyx_v_field->name); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_7;
        __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u_missing_2);
        __pyx_t_6 += 9;
        __Pyx_GIVEREF(__pyx_kp_u_missing_2);
        PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_missing_2);
        __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_4, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(11, 498, __pyx_L1_error)

        /* "src/record.pxi":497
 *             field_value = value.get(field.writer_name, MISSING_VALUE)
 *             if field_value is MISSING_VALUE:
 *                 if field.default_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                     raise ValueError(f"required field '{field.name}' missing")
 *                 field_value = field.default_value
 */
      }

      /* "src/record.pxi":499
 *                 if field.default_value is NO_DEFAULT:
 *                     raise ValueError(f"required field '{field.name}' missing")
 *                 field_value = field.default_value             # <<<<<<<<<<<<<<
 *             else:
 *                 field_value = field.type.json_decode(field_value)
 */
      __pyx_t_4 = __pyx_v_field->default_value;
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_field_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "src/record.pxi":496
 *         for field in self.fields:
 *             field_value = value.get(field.writer_name, MISSING_VALUE)
 *             if field_value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *                 if field.default_value is NO_DEFAULT:
 *                     raise ValueError(f"required field '{field.name}' missing")
 */
      goto __pyx_L5;
    }

    /* "src/record.pxi":501
 *                 field_value = field.default_value
 *             else:
 *                 field_value = field.type.json_decode(field_value)             # <<<<<<<<<<<<<<
 *             data[index] = field_value
 *             index += 1
 */
    /*else*/ {
      __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_field->type->__pyx_vtab)->json_decode(__pyx_v_field->type, __pyx_v_field_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_field_value, __pyx_t_4);
      __pyx_t_4 = 0;
    }
    __pyx_L5:;

    /* "src/record.pxi":502
 *             else:
 *                 field_value = field.type.json_decode(field_value)
 *             data[index] = field_value             # <<<<<<<<<<<<<<
 *             index += 1
 *         rec = Record.__new__(self.record)
 */
    if (unlikely((__Pyx_SetItemInt(__pyx_v_data, __pyx_v_index, __pyx_v_field_value, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1) < 0))) __PYX_ERR(11, 502, __pyx_L1_error)

    /* "src/record.pxi":503
 *                 field_value = field.type.json_decode(field_value)
 *             data[index] = field_value
 *             index += 1             # <<<<<<<<<<<<<<
 *         rec = Record.__new__(self.record)
 *         rec.data = data
 */
    __pyx_v_index = (__pyx_v_index + 1);

    /* "src/record.pxi":494
 *         cdef Py_ssize_t index = 0
 * 
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             field_value = value.get(field.writer_name, MISSING_VALUE)
 *             if field_value is MISSING_VALUE:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/record.pxi":504
 *             data[index] = field_value
 *             index += 1
 *         rec = Record.__new__(self.record)             # <<<<<<<<<<<<<<
 *         rec.data = data
 *         return rec
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_Record), __pyx_n_s_new); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_self->record)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Record))))) __PYX_ERR(11, 504, __pyx_L1_error)
  __pyx_v_rec = ((struct __pyx_obj_5cavro_Record *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":505
 *             index += 1
 *         rec = Record.__new__(self.record)
 *         rec.data = data             # <<<<<<<<<<<<<<
 *         return rec
 * 
 */
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  __Pyx_GOTREF(__pyx_v_rec->data);
  __Pyx_DECREF(__pyx_v_rec->data);
  __pyx_v_rec->data = __pyx_v_data;

  /* "src/record.pxi":506
 *         rec = Record.__new__(self.record)
 *         rec.data = data
 *         return rec             # <<<<<<<<<<<<<<
 * 
 *     cdef json_decode_dict(self, dict value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_rec);
  __pyx_r = ((PyObject *)__pyx_v_rec);
  goto __pyx_L0;

  /* "src/record.pxi":488
 *         return out
 * 
 *     cdef json_decode_record(self, dict value):             # <<<<<<<<<<<<<<
 *         cdef list data = [None] * len(self.fields)
 *         cdef RecordField field
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.RecordType.json_decode_record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XDECREF((PyObject *)__pyx_v_rec);
  __Pyx_XDECREF(__pyx_v_field_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":508
 *         return rec
 * 
 *     cdef json_decode_dict(self, dict value):             # <<<<<<<<<<<<<<
 *         cdef dict data = {}
 *         cdef RecordField field
 */

static PyObject *__pyx_f_5cavro_10RecordType_json_decode_dict(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_data = 0;
  struct __pyx_obj_5cavro_RecordField *__pyx_v_field = 0;
  PyObject *__pyx_v_field_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("json_decode_dict", 1);

  /* "src/record.pxi":509
 * 
 *     cdef json_decode_dict(self, dict value):
 *         cdef dict data = {}             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":512
 *         cdef RecordField field
 * 
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             field_value = value.get(field.writer_name, MISSING_VALUE)
 *             if field_value is MISSING_VALUE:
 */
  if (unlikely(__pyx_v_self->fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(11, 512, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->fields; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 512, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(11, 512, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 512, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/record.pxi":513
 * 
 *         for field in self.fields:
 *             field_value = value.get(field.writer_name, MISSING_VALUE)             # <<<<<<<<<<<<<<
 *             if field_value is MISSING_VALUE:
 *                 if field.default_value is NO_DEFAULT:
 */
    if (unlikely(__pyx_v_value == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
      __PYX_ERR(11, 513, __pyx_L1_error)
    }
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyDict_GetItemDefault(__pyx_v_value, __pyx_v_field->writer_name, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_field_value, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/record.pxi":514
 *         for field in self.fields:
 *             field_value = value.get(field.writer_name, MISSING_VALUE)
 *             if field_value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *                 if field.default_value is NO_DEFAULT:
 *                     raise ValueError(f"required field '{field.name}' missing")
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = (__pyx_v_field_value == __pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "src/record.pxi":515
 *             field_value = value.get(field.writer_name, MISSING_VALUE)
 *             if field_value is MISSING_VALUE:
 *                 if field.default_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                     raise ValueError(f"required field '{field.name}' missing")
 *                 field_value = field.default_value
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 515, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = (__pyx_v_field->default_value == __pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(__pyx_t_5)) {

        /* "src/record.pxi":516
 *             if field_value is MISSING_VALUE:
 *                 if field.default_value is NO_DEFAULT:
 *                     raise ValueError(f"required field '{field.name}' missing")             # <<<<<<<<<<<<<<
 *                 field_value = field.default_value
 *             else:
 */
        __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 516, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = 0;
        __pyx_t_7 = 127;
        __Pyx_INCREF(__pyx_kp_u_required_field);
        __pyx_t_6 += 16;
        __Pyx_GIVEREF(__pyx_kp_u_required_field);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_required_field);
        __pyx_t_3 = __Pyx_PyUnicode_Unicode(__pyx_v_field->name); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 516, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_7;
        __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_kp_u_missing_2);
        __pyx_t_6 += 9;
        __Pyx_GIVEREF(__pyx_kp_u_missing_2);
        PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_missing_2);
        __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_4, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 516, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 516, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(11, 516, __pyx_L1_error)

        /* "src/record.pxi":515
 *             field_value = value.get(field.writer_name, MISSING_VALUE)
 *             if field_value is MISSING_VALUE:
 *                 if field.default_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                     raise ValueError(f"required field '{field.name}' missing")
 *                 field_value = field.default_value
 */
      }

      /* "src/record.pxi":517
 *                 if field.default_value is NO_DEFAULT:
 *                     raise ValueError(f"required field '{field.name}' missing")
 *                 field_value = field.default_value             # <<<<<<<<<<<<<<
 *             else:
 *                 field_value = field.type.json_decode(field_value)
 */
      __pyx_t_4 = __pyx_v_field->default_value;
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_field_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "src/record.pxi":514
 *         for field in self.fields:
 *             field_value = value.get(field.writer_name, MISSING_VALUE)
 *             if field_value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *                 if field.default_value is NO_DEFAULT:
 *                     raise ValueError(f"required field '{field.name}' missing")
 */
      goto __pyx_L5;
    }

    /* "src/record.pxi":519
 *                 field_value = field.default_value
 *             else:
 *                 field_value = field.type.json_decode(field_value)             # <<<<<<<<<<<<<<
 *             data[field.name] = field_value
 *         return data
 */
    /*else*/ {
      __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_field->type->__pyx_vtab)->json_decode(__pyx_v_field->type, __pyx_v_field_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 519, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_field_value, __pyx_t_4);
      __pyx_t_4 = 0;
    }
    __pyx_L5:;

    /* "src/record.pxi":520
 *             else:
 *                 field_value = field.type.json_decode(field_value)
 *             data[field.name] = field_value             # <<<<<<<<<<<<<<
 *         return data
 * 
 */
    if (unlikely((PyDict_SetItem(__pyx_v_data, __pyx_v_field->name, __pyx_v_field_value) < 0))) __PYX_ERR(11, 520, __pyx_L1_error)

    /* "src/record.pxi":512
 *         cdef RecordField field
 * 
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             field_value = value.get(field.writer_name, MISSING_VALUE)
 *             if field_value is MISSING_VALUE:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/record.pxi":521
 *                 field_value = field.type.json_decode(field_value)
 *             data[field.name] = field_value
 *         return data             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_data);
  __pyx_r = __pyx_v_data;
  goto __pyx_L0;

  /* "src/record.pxi":508
 *         return rec
 * 
 *     cdef json_decode_dict(self, dict value):             # <<<<<<<<<<<<<<
 *         cdef dict data = {}
 *         cdef RecordField field
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.RecordType.json_decode_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_field_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":523
 *         return data
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if self.options.record_decodes_to_dict:
 *             return self.json_decode_dict(value)
 */

static PyObject *__pyx_f_5cavro_10RecordType__json_decode(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/record.pxi":524
 * 
 *     cdef _json_decode(self, value):
 *         if self.options.record_decodes_to_dict:             # <<<<<<<<<<<<<<
 *             return self.json_decode_dict(value)
 *         return self.json_decode_record(value)
 */
  if (__pyx_v_self->__pyx_base.__pyx_base.options->record_decodes_to_dict) {

    /* "src/record.pxi":525
 *     cdef _json_decode(self, value):
 *         if self.options.record_decodes_to_dict:
 *             return self.json_decode_dict(value)             # <<<<<<<<<<<<<<
 *         return self.json_decode_record(value)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(11, 525, __pyx_L1_error)
    __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_RecordType *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->json_decode_dict(__pyx_v_self, ((PyObject*)__pyx_v_value)); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/record.pxi":524
 * 
 *     cdef _json_decode(self, value):
 *         if self.options.record_decodes_to_dict:             # <<<<<<<<<<<<<<
 *             return self.json_decode_dict(value)
 *         return self.json_decode_record(value)
 */
  }

  /* "src/record.pxi":526
 *         if self.options.record_decodes_to_dict:
 *             return self.json_decode_dict(value)
 *         return self.json_decode_record(value)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(11, 526, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_RecordType *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->json_decode_record(__pyx_v_self, ((PyObject*)__pyx_v_value)); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":523
 *         return data
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if self.options.record_decodes_to_dict:
 *             return self.json_decode_dict(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.RecordType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":528
 *         return self.json_decode_record(value)
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, self.record):
 *             return value
 */

static PyObject *__pyx_pw_5cavro_10RecordType_10_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_10RecordType__convert_value(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_convert_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 528, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_10RecordType_10_convert_value)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_value};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 528, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/record.pxi":529
 * 
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, self.record):             # <<<<<<<<<<<<<<
 *             return value
 *         return self.record(value)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->record);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_TypeCheck(__pyx_v_value, __pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "src/record.pxi":530
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, self.record):
 *             return value             # <<<<<<<<<<<<<<
 *         return self.record(value)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/record.pxi":529
 * 
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, self.record):             # <<<<<<<<<<<<<<
 *             return value
 *         return self.record(value)
 */
  }

  /* "src/record.pxi":531
 *         if isinstance(value, self.record):
 *             return value
 *         return self.record(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_v_self->record), __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":528
 *         return self.json_decode_record(value)
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, self.record):
 *             return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.RecordType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10RecordType_10_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10RecordType_10_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10RecordType_10_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10RecordType_10_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 528, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(11, 528, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(11, 528, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.RecordType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10RecordType_9_convert_value(((struct __pyx_obj_5cavro_RecordType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10RecordType_9_convert_value(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_10RecordType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.RecordType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":533
 *         return self.record(value)
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         if self in created:
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_10RecordType_canonical_form(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro_RecordField *__pyx_9genexpr13__pyx_v_field = NULL;
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/record.pxi":535
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         cdef RecordField field
 *         if self in created:             # <<<<<<<<<<<<<<
 *             return _CanonicalForm('"' + self.type + '"')
 *         created.add(self)
 */
  if (unlikely(__pyx_v_created == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(11, 535, __pyx_L1_error)
  }
  __pyx_t_1 = (__Pyx_PySet_ContainsTF(((PyObject *)__pyx_v_self), __pyx_v_created, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(11, 535, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/record.pxi":536
 *         cdef RecordField field
 *         if self in created:
 *             return _CanonicalForm('"' + self.type + '"')             # <<<<<<<<<<<<<<
 *         created.add(self)
 *         return dict_to_canonical({
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Add(__pyx_kp_u__48, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Add(__pyx_t_3, __pyx_kp_u__48); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "src/record.pxi":535
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         cdef RecordField field
 *         if self in created:             # <<<<<<<<<<<<<<
 *             return _CanonicalForm('"' + self.type + '"')
 *         created.add(self)
 */
  }

  /* "src/record.pxi":537
 *         if self in created:
 *             return _CanonicalForm('"' + self.type + '"')
 *         created.add(self)             # <<<<<<<<<<<<<<
 *         return dict_to_canonical({
 *             'type': 'record',
 */
  if (unlikely(__pyx_v_created == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
    __PYX_ERR(11, 537, __pyx_L1_error)
  }
  __pyx_t_4 = PySet_Add(__pyx_v_created, ((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(11, 537, __pyx_L1_error)

  /* "src/record.pxi":538
 *             return _CanonicalForm('"' + self.type + '"')
 *         created.add(self)
 *         return dict_to_canonical({             # <<<<<<<<<<<<<<
 *             'type': 'record',
 *             'name': self.type,
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);

  /* "src/record.pxi":539
 *         created.add(self)
 *         return dict_to_canonical({
 *             'type': 'record',             # <<<<<<<<<<<<<<
 *             'name': self.type,
 *             'fields': [field.canonical_form(created) for field in self.fields]
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_type, __pyx_n_u_record) < 0) __PYX_ERR(11, 539, __pyx_L1_error)

  /* "src/record.pxi":540
 *         return dict_to_canonical({
 *             'type': 'record',
 *             'name': self.type,             # <<<<<<<<<<<<<<
 *             'fields': [field.canonical_form(created) for field in self.fields]
 *         })
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_name, __pyx_t_2) < 0) __PYX_ERR(11, 539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  { /* enter inner scope */

    /* "src/record.pxi":541
 *             'type': 'record',
 *             'name': self.type,
 *             'fields': [field.canonical_form(created) for field in self.fields]             # <<<<<<<<<<<<<<
 *         })
 * 
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 541, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_self->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(11, 541, __pyx_L6_error)
    }
    __pyx_t_5 = __pyx_v_self->fields; __Pyx_INCREF(__pyx_t_5);
    __pyx_t_6 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 541, __pyx_L6_error)
        #endif
        if (__pyx_t_6 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_7); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(11, 541, __pyx_L6_error)
      #else
      __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 541, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 541, __pyx_L6_error)
      __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_7));
      __pyx_t_7 = 0;
      __pyx_t_7 = ((PyObject *)__pyx_f_5cavro_11RecordField_canonical_form(__pyx_9genexpr13__pyx_v_field, __pyx_v_created)); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 541, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_7))) __PYX_ERR(11, 541, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF((PyObject *)__pyx_9genexpr13__pyx_v_field); __pyx_9genexpr13__pyx_v_field = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF((PyObject *)__pyx_9genexpr13__pyx_v_field); __pyx_9genexpr13__pyx_v_field = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_fields, __pyx_t_2) < 0) __PYX_ERR(11, 539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/record.pxi":538
 *             return _CanonicalForm('"' + self.type + '"')
 *         created.add(self)
 *         return dict_to_canonical({             # <<<<<<<<<<<<<<
 *             'type': 'record',
 *             'name': self.type,
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_5cavro_dict_to_canonical(__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/record.pxi":533
 *         return self.record(value)
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         if self in created:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.RecordType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_9genexpr13__pyx_v_field);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":544
 *         })
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         cdef PromotingRecordType cloned
 *         cdef RecordType writer_rec
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10RecordType__for_writer(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer) {
  struct __pyx_obj_5cavro_PromotingRecordType *__pyx_v_cloned = 0;
  struct __pyx_obj_5cavro_RecordType *__pyx_v_writer_rec = 0;
  struct __pyx_obj_5cavro_RecordField *__pyx_v_field = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_v_reader_field_idx = NULL;
  PyObject *__pyx_v_reader_fields = NULL;
  PyObject *__pyx_v_aliases = NULL;
  PyObject *__pyx_v_alias = NULL;
  PyObject *__pyx_v_out_decode_indexes = NULL;
  PyObject *__pyx_v_out_fields = NULL;
  PyObject *__pyx_v_decoded_reader_fields = NULL;
  PyObject *__pyx_v_reader_index = NULL;
  PyObject *__pyx_v_reader_field = NULL;
  PyObject *__pyx_v_extra_reader_fields = NULL;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_9genexpr14__pyx_v_f = NULL;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  Py_UCS4 __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_clone_base __pyx_t_16;
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_writer", 1);

  /* "src/record.pxi":549
 *         cdef RecordField field
 *         cdef Py_ssize_t i
 *         if not isinstance(writer, RecordType):             # <<<<<<<<<<<<<<
 *             return
 *         writer_rec = writer
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_RecordType); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "src/record.pxi":550
 *         cdef Py_ssize_t i
 *         if not isinstance(writer, RecordType):
 *             return             # <<<<<<<<<<<<<<
 *         writer_rec = writer
 *         if not self.name_matches(writer_rec):
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/record.pxi":549
 *         cdef RecordField field
 *         cdef Py_ssize_t i
 *         if not isinstance(writer, RecordType):             # <<<<<<<<<<<<<<
 *             return
 *         writer_rec = writer
 */
  }

  /* "src/record.pxi":551
 *         if not isinstance(writer, RecordType):
 *             return
 *         writer_rec = writer             # <<<<<<<<<<<<<<
 *         if not self.name_matches(writer_rec):
 *             return
 */
  if (!(likely(((((PyObject *)__pyx_v_writer)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_RecordType))))) __PYX_ERR(11, 551, __pyx_L1_error)
  __pyx_t_3 = ((PyObject *)__pyx_v_writer);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_writer_rec = ((struct __pyx_obj_5cavro_RecordType *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/record.pxi":552
 *             return
 *         writer_rec = writer
 *         if not self.name_matches(writer_rec):             # <<<<<<<<<<<<<<
 *             return
 *         # This allows us to go from the field name to the index in the reader record
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_RecordType *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.name_matches(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self), ((struct __pyx_obj_5cavro__NamedType *)__pyx_v_writer_rec)); if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 552, __pyx_L1_error)
  __pyx_t_1 = (!__pyx_t_2);
  if (__pyx_t_1) {

    /* "src/record.pxi":553
 *         writer_rec = writer
 *         if not self.name_matches(writer_rec):
 *             return             # <<<<<<<<<<<<<<
 *         # This allows us to go from the field name to the index in the reader record
 *         reader_field_idx = {}
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/record.pxi":552
 *             return
 *         writer_rec = writer
 *         if not self.name_matches(writer_rec):             # <<<<<<<<<<<<<<
 *             return
 *         # This allows us to go from the field name to the index in the reader record
 */
  }

  /* "src/record.pxi":555
 *             return
 *         # This allows us to go from the field name to the index in the reader record
 *         reader_field_idx = {}             # <<<<<<<<<<<<<<
 *         reader_fields = {}
 *         aliases = {}
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_reader_field_idx = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/record.pxi":556
 *         # This allows us to go from the field name to the index in the reader record
 *         reader_field_idx = {}
 *         reader_fields = {}             # <<<<<<<<<<<<<<
 *         aliases = {}
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_reader_fields = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/record.pxi":557
 *         reader_field_idx = {}
 *         reader_fields = {}
 *         aliases = {}             # <<<<<<<<<<<<<<
 * 
 *         i = 0
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_aliases = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/record.pxi":559
 *         aliases = {}
 * 
 *         i = 0             # <<<<<<<<<<<<<<
 *         for field in self.fields:
 *             reader_field_idx[field.name] = i
 */
  __pyx_v_i = 0;

  /* "src/record.pxi":560
 * 
 *         i = 0
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             reader_field_idx[field.name] = i
 *             reader_fields[i] = field
 */
  if (unlikely(__pyx_v_self->fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(11, 560, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_self->fields; __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 560, __pyx_L1_error)
      #endif
      if (__pyx_t_4 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(11, 560, __pyx_L1_error)
    #else
    __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 560, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "src/record.pxi":561
 *         i = 0
 *         for field in self.fields:
 *             reader_field_idx[field.name] = i             # <<<<<<<<<<<<<<
 *             reader_fields[i] = field
 *             for alias in field.aliases:
 */
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 561, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely((PyDict_SetItem(__pyx_v_reader_field_idx, __pyx_v_field->name, __pyx_t_5) < 0))) __PYX_ERR(11, 561, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "src/record.pxi":562
 *         for field in self.fields:
 *             reader_field_idx[field.name] = i
 *             reader_fields[i] = field             # <<<<<<<<<<<<<<
 *             for alias in field.aliases:
 *                 aliases[alias] = i
 */
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely((PyDict_SetItem(__pyx_v_reader_fields, __pyx_t_5, ((PyObject *)__pyx_v_field)) < 0))) __PYX_ERR(11, 562, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "src/record.pxi":563
 *             reader_field_idx[field.name] = i
 *             reader_fields[i] = field
 *             for alias in field.aliases:             # <<<<<<<<<<<<<<
 *                 aliases[alias] = i
 *             i += 1
 */
    __pyx_t_6 = 0;
    __pyx_t_9 = __Pyx_set_iterator(__pyx_v_field->aliases, 0, (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_5);
    __pyx_t_5 = __pyx_t_9;
    __pyx_t_9 = 0;
    while (1) {
      __pyx_t_10 = __Pyx_set_iter_next(__pyx_t_5, __pyx_t_7, &__pyx_t_6, &__pyx_t_9, __pyx_t_8);
      if (unlikely(__pyx_t_10 == 0)) break;
      if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(11, 563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_v_alias, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "src/record.pxi":564
 *             reader_fields[i] = field
 *             for alias in field.aliases:
 *                 aliases[alias] = i             # <<<<<<<<<<<<<<
 *             i += 1
 * 
 */
      __pyx_t_9 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (unlikely((PyDict_SetItem(__pyx_v_aliases, __pyx_v_alias, __pyx_t_9) < 0))) __PYX_ERR(11, 564, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "src/record.pxi":565
 *             for alias in field.aliases:
 *                 aliases[alias] = i
 *             i += 1             # <<<<<<<<<<<<<<
 * 
 *         # We're going to have to read each writer field in order anyway, but if it's a reader field,
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "src/record.pxi":560
 * 
 *         i = 0
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             reader_field_idx[field.name] = i
 *             reader_fields[i] = field
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "src/record.pxi":569
 *         # We're going to have to read each writer field in order anyway, but if it's a reader field,
 *         # Then it ends up in the corresponding decode_index slot, but should also be promoted etc..
 *         out_decode_indexes = []             # <<<<<<<<<<<<<<
 *         out_fields = []
 *         decoded_reader_fields = set()
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_out_decode_indexes = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/record.pxi":570
 *         # Then it ends up in the corresponding decode_index slot, but should also be promoted etc..
 *         out_decode_indexes = []
 *         out_fields = []             # <<<<<<<<<<<<<<
 *         decoded_reader_fields = set()
 * 
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_out_fields = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/record.pxi":571
 *         out_decode_indexes = []
 *         out_fields = []
 *         decoded_reader_fields = set()             # <<<<<<<<<<<<<<
 * 
 *         # Todo, only use promoting type if the field order/naming is different
 */
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_decoded_reader_fields = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/record.pxi":574
 * 
 *         # Todo, only use promoting type if the field order/naming is different
 *         for field in  writer_rec.fields:             # <<<<<<<<<<<<<<
 *             if field.name in reader_field_idx:
 *                 reader_index = reader_field_idx[field.name]
 */
  if (unlikely(__pyx_v_writer_rec->fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(11, 574, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_writer_rec->fields; __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 574, __pyx_L1_error)
      #endif
      if (__pyx_t_4 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(11, 574, __pyx_L1_error)
    #else
    __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 574, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "src/record.pxi":575
 *         # Todo, only use promoting type if the field order/naming is different
 *         for field in  writer_rec.fields:
 *             if field.name in reader_field_idx:             # <<<<<<<<<<<<<<
 *                 reader_index = reader_field_idx[field.name]
 *                 reader_field = reader_fields[reader_index]
 */
    __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_field->name, __pyx_v_reader_field_idx, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(11, 575, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "src/record.pxi":576
 *         for field in  writer_rec.fields:
 *             if field.name in reader_field_idx:
 *                 reader_index = reader_field_idx[field.name]             # <<<<<<<<<<<<<<
 *                 reader_field = reader_fields[reader_index]
 *             elif field.name in aliases:
 */
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_reader_field_idx, __pyx_v_field->name); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_reader_index, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "src/record.pxi":577
 *             if field.name in reader_field_idx:
 *                 reader_index = reader_field_idx[field.name]
 *                 reader_field = reader_fields[reader_index]             # <<<<<<<<<<<<<<
 *             elif field.name in aliases:
 *                 reader_index = aliases[field.name]
 */
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_reader_fields, __pyx_v_reader_index); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_reader_field, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "src/record.pxi":575
 *         # Todo, only use promoting type if the field order/naming is different
 *         for field in  writer_rec.fields:
 *             if field.name in reader_field_idx:             # <<<<<<<<<<<<<<
 *                 reader_index = reader_field_idx[field.name]
 *                 reader_field = reader_fields[reader_index]
 */
      goto __pyx_L12;
    }

    /* "src/record.pxi":578
 *                 reader_index = reader_field_idx[field.name]
 *                 reader_field = reader_fields[reader_index]
 *             elif field.name in aliases:             # <<<<<<<<<<<<<<
 *                 reader_index = aliases[field.name]
 *                 reader_field = reader_fields[reader_index]
 */
    __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_field->name, __pyx_v_aliases, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(11, 578, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "src/record.pxi":579
 *                 reader_field = reader_fields[reader_index]
 *             elif field.name in aliases:
 *                 reader_index = aliases[field.name]             # <<<<<<<<<<<<<<
 *                 reader_field = reader_fields[reader_index]
 *             else:
 */
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_aliases, __pyx_v_field->name); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 579, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_reader_index, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "src/record.pxi":580
 *             elif field.name in aliases:
 *                 reader_index = aliases[field.name]
 *                 reader_field = reader_fields[reader_index]             # <<<<<<<<<<<<<<
 *             else:
 *                 # This field is not in the reader, so we read it and discard (index = -1)
 */
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_reader_fields, __pyx_v_reader_index); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 580, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_reader_field, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "src/record.pxi":578
 *                 reader_index = reader_field_idx[field.name]
 *                 reader_field = reader_fields[reader_index]
 *             elif field.name in aliases:             # <<<<<<<<<<<<<<
 *                 reader_index = aliases[field.name]
 *                 reader_field = reader_fields[reader_index]
 */
      goto __pyx_L12;
    }

    /* "src/record.pxi":583
 *             else:
 *                 # This field is not in the reader, so we read it and discard (index = -1)
 *                 out_decode_indexes.append(-1)             # <<<<<<<<<<<<<<
 *                 out_fields.append(field.for_reader(None, None))
 *                 continue
 */
    /*else*/ {
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_out_decode_indexes, __pyx_int_neg_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(11, 583, __pyx_L1_error)

      /* "src/record.pxi":584
 *                 # This field is not in the reader, so we read it and discard (index = -1)
 *                 out_decode_indexes.append(-1)
 *                 out_fields.append(field.for_reader(None, None))             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
      __pyx_t_5 = __pyx_f_5cavro_11RecordField_for_reader(__pyx_v_field, ((PyObject*)Py_None), ((struct __pyx_obj_5cavro_AvroType *)Py_None)); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_out_fields, __pyx_t_5); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(11, 584, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "src/record.pxi":585
 *                 out_decode_indexes.append(-1)
 *                 out_fields.append(field.for_reader(None, None))
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             out_decode_indexes.append(reader_index)
 */
      goto __pyx_L10_continue;
    }
    __pyx_L12:;

    /* "src/record.pxi":587
 *                 continue
 * 
 *             out_decode_indexes.append(reader_index)             # <<<<<<<<<<<<<<
 *             out_fields.append(field.for_reader(reader_field.name, reader_field.type))
 *             decoded_reader_fields.add(reader_index)
 */
    __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_out_decode_indexes, __pyx_v_reader_index); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(11, 587, __pyx_L1_error)

    /* "src/record.pxi":588
 * 
 *             out_decode_indexes.append(reader_index)
 *             out_fields.append(field.for_reader(reader_field.name, reader_field.type))             # <<<<<<<<<<<<<<
 *             decoded_reader_fields.add(reader_index)
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_reader_field, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_5))) __PYX_ERR(11, 588, __pyx_L1_error)
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_reader_field, __pyx_n_s_type); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(11, 588, __pyx_L1_error)
    __pyx_t_12 = __pyx_f_5cavro_11RecordField_for_reader(__pyx_v_field, ((PyObject*)__pyx_t_5), ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_9)); if (unlikely(!__pyx_t_12)) __PYX_ERR(11, 588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_out_fields, __pyx_t_12); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(11, 588, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "src/record.pxi":589
 *             out_decode_indexes.append(reader_index)
 *             out_fields.append(field.for_reader(reader_field.name, reader_field.type))
 *             decoded_reader_fields.add(reader_index)             # <<<<<<<<<<<<<<
 * 
 *         extra_reader_fields = reader_fields.keys() - decoded_reader_fields
 */
    __pyx_t_11 = PySet_Add(__pyx_v_decoded_reader_fields, __pyx_v_reader_index); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(11, 589, __pyx_L1_error)

    /* "src/record.pxi":574
 * 
 *         # Todo, only use promoting type if the field order/naming is different
 *         for field in  writer_rec.fields:             # <<<<<<<<<<<<<<
 *             if field.name in reader_field_idx:
 *                 reader_index = reader_field_idx[field.name]
 */
    __pyx_L10_continue:;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "src/record.pxi":591
 *             decoded_reader_fields.add(reader_index)
 * 
 *         extra_reader_fields = reader_fields.keys() - decoded_reader_fields             # <<<<<<<<<<<<<<
 *         for reader_index in extra_reader_fields:
 *             field = reader_fields[reader_index]
 */
  __pyx_t_3 = __Pyx_PyDict_Keys(__pyx_v_reader_fields); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_12 = PyNumber_Subtract(__pyx_t_3, __pyx_v_decoded_reader_fields); if (unlikely(!__pyx_t_12)) __PYX_ERR(11, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_extra_reader_fields = __pyx_t_12;
  __pyx_t_12 = 0;

  /* "src/record.pxi":592
 * 
 *         extra_reader_fields = reader_fields.keys() - decoded_reader_fields
 *         for reader_index in extra_reader_fields:             # <<<<<<<<<<<<<<
 *             field = reader_fields[reader_index]
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_extra_reader_fields)) || PyTuple_CheckExact(__pyx_v_extra_reader_fields)) {
    __pyx_t_12 = __pyx_v_extra_reader_fields; __Pyx_INCREF(__pyx_t_12);
    __pyx_t_4 = 0;
    __pyx_t_13 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_12 = PyObject_GetIter(__pyx_v_extra_reader_fields); if (unlikely(!__pyx_t_12)) __PYX_ERR(11, 592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(11, 592, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_13)) {
      if (likely(PyList_CheckExact(__pyx_t_12))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_12);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 592, __pyx_L1_error)
          #endif
          if (__pyx_t_4 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_12, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(11, 592, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_12, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 592, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_12);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 592, __pyx_L1_error)
          #endif
          if (__pyx_t_4 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_12, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(11, 592, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_12, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 592, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_13(__pyx_t_12);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(11, 592, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_reader_index, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/record.pxi":593
 *         extra_reader_fields = reader_fields.keys() - decoded_reader_fields
 *         for reader_index in extra_reader_fields:
 *             field = reader_fields[reader_index]             # <<<<<<<<<<<<<<
 * 
 *             if field.default_value is NO_DEFAULT:
 */
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_reader_fields, __pyx_v_reader_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 593, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 593, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/record.pxi":595
 *             field = reader_fields[reader_index]
 * 
 *             if field.default_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                 raise CannotPromoteError(self, writer, f"required field '{field.name}' missing")
 *             out_decode_indexes.append(reader_index)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 595, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = (__pyx_v_field->default_value == __pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "src/record.pxi":596
 * 
 *             if field.default_value is NO_DEFAULT:
 *                 raise CannotPromoteError(self, writer, f"required field '{field.name}' missing")             # <<<<<<<<<<<<<<
 *             out_decode_indexes.append(reader_index)
 *             out_fields.append(field.placeholder())
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_CannotPromoteError); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = 0;
      __pyx_t_14 = 127;
      __Pyx_INCREF(__pyx_kp_u_required_field);
      __pyx_t_7 += 16;
      __Pyx_GIVEREF(__pyx_kp_u_required_field);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_required_field);
      __pyx_t_15 = __Pyx_PyUnicode_Unicode(__pyx_v_field->name); if (unlikely(!__pyx_t_15)) __PYX_ERR(11, 596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_14 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) > __pyx_t_14) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) : __pyx_t_14;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15);
      __Pyx_GIVEREF(__pyx_t_15);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_15);
      __pyx_t_15 = 0;
      __Pyx_INCREF(__pyx_kp_u_missing_2);
      __pyx_t_7 += 9;
      __Pyx_GIVEREF(__pyx_kp_u_missing_2);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u_missing_2);
      __pyx_t_15 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, __pyx_t_7, __pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(11, 596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_5, ((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_writer), __pyx_t_15};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_8, 3+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 596, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(11, 596, __pyx_L1_error)

      /* "src/record.pxi":595
 *             field = reader_fields[reader_index]
 * 
 *             if field.default_value is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *                 raise CannotPromoteError(self, writer, f"required field '{field.name}' missing")
 *             out_decode_indexes.append(reader_index)
 */
    }

    /* "src/record.pxi":597
 *             if field.default_value is NO_DEFAULT:
 *                 raise CannotPromoteError(self, writer, f"required field '{field.name}' missing")
 *             out_decode_indexes.append(reader_index)             # <<<<<<<<<<<<<<
 *             out_fields.append(field.placeholder())
 * 
 */
    __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_out_decode_indexes, __pyx_v_reader_index); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(11, 597, __pyx_L1_error)

    /* "src/record.pxi":598
 *                 raise CannotPromoteError(self, writer, f"required field '{field.name}' missing")
 *             out_decode_indexes.append(reader_index)
 *             out_fields.append(field.placeholder())             # <<<<<<<<<<<<<<
 * 
 *         cloned = self.clone_base(PromotingRecordType)
 */
    __pyx_t_3 = __pyx_f_5cavro_11RecordField_placeholder(__pyx_v_field); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_out_fields, __pyx_t_3); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(11, 598, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "src/record.pxi":592
 * 
 *         extra_reader_fields = reader_fields.keys() - decoded_reader_fields
 *         for reader_index in extra_reader_fields:             # <<<<<<<<<<<<<<
 *             field = reader_fields[reader_index]
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "src/record.pxi":600
 *             out_fields.append(field.placeholder())
 * 
 *         cloned = self.clone_base(PromotingRecordType)             # <<<<<<<<<<<<<<
 *         cloned.decode_indexes = cvarray((len(out_decode_indexes), ), sizeof(Py_ssize_t), SSIZE_TYPECODE)
 *         for i, idx in enumerate(out_decode_indexes):
 */
  __pyx_t_16.__pyx_n = 1;
  __pyx_t_16.cls = ((PyObject *)__pyx_ptype_5cavro_PromotingRecordType);
  __pyx_t_12 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_RecordType *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), &__pyx_t_16)); if (unlikely(!__pyx_t_12)) __PYX_ERR(11, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (!(likely(((__pyx_t_12) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_12, __pyx_ptype_5cavro_PromotingRecordType))))) __PYX_ERR(11, 600, __pyx_L1_error)
  __pyx_v_cloned = ((struct __pyx_obj_5cavro_PromotingRecordType *)__pyx_t_12);
  __pyx_t_12 = 0;

  /* "src/record.pxi":601
 * 
 *         cloned = self.clone_base(PromotingRecordType)
 *         cloned.decode_indexes = cvarray((len(out_decode_indexes), ), sizeof(Py_ssize_t), SSIZE_TYPECODE)             # <<<<<<<<<<<<<<
 *         for i, idx in enumerate(out_decode_indexes):
 *             cloned.decode_indexes[i] = idx
 */
  __pyx_t_4 = __Pyx_PyList_GET_SIZE(__pyx_v_out_decode_indexes); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(11, 601, __pyx_L1_error)
  __pyx_t_12 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_12)) __PYX_ERR(11, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_12)) __PYX_ERR(11, 601, __pyx_L1_error);
  __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyInt_FromSize_t((sizeof(Py_ssize_t))); if (unlikely(!__pyx_t_12)) __PYX_ERR(11, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_SSIZE_TYPECODE); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(11, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_3)) __PYX_ERR(11, 601, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_12)) __PYX_ERR(11, 601, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_9)) __PYX_ERR(11, 601, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_12 = 0;
  __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_15, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_17 = __Pyx_PyObject_to_MemoryviewSlice_ds_Py_ssize_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_17.memview)) __PYX_ERR(11, 601, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __PYX_XCLEAR_MEMVIEW(&__pyx_v_cloned->decode_indexes, 0);
  __pyx_v_cloned->decode_indexes = __pyx_t_17;
  __pyx_t_17.memview = NULL;
  __pyx_t_17.data = NULL;

  /* "src/record.pxi":602
 *         cloned = self.clone_base(PromotingRecordType)
 *         cloned.decode_indexes = cvarray((len(out_decode_indexes), ), sizeof(Py_ssize_t), SSIZE_TYPECODE)
 *         for i, idx in enumerate(out_decode_indexes):             # <<<<<<<<<<<<<<
 *             cloned.decode_indexes[i] = idx
 *         cloned.fields = tuple(out_fields)
 */
  __pyx_t_4 = 0;
  __pyx_t_9 = __pyx_v_out_decode_indexes; __Pyx_INCREF(__pyx_t_9);
  __pyx_t_7 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 602, __pyx_L1_error)
      #endif
      if (__pyx_t_7 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_15 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_7); __Pyx_INCREF(__pyx_t_15); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(11, 602, __pyx_L1_error)
    #else
    __pyx_t_15 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_15)) __PYX_ERR(11, 602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_15);
    __pyx_t_15 = 0;
    __pyx_v_i = __pyx_t_4;
    __pyx_t_4 = (__pyx_t_4 + 1);

    /* "src/record.pxi":603
 *         cloned.decode_indexes = cvarray((len(out_decode_indexes), ), sizeof(Py_ssize_t), SSIZE_TYPECODE)
 *         for i, idx in enumerate(out_decode_indexes):
 *             cloned.decode_indexes[i] = idx             # <<<<<<<<<<<<<<
 *         cloned.fields = tuple(out_fields)
 *         cloned.field_dict = {f.name: f for f in cloned.fields}
 */
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(11, 603, __pyx_L1_error)
    if (unlikely(!__pyx_v_cloned->decode_indexes.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(11, 603, __pyx_L1_error)}
    __pyx_t_18 = __pyx_v_i;
    __pyx_t_8 = -1;
    if (__pyx_t_18 < 0) {
      __pyx_t_18 += __pyx_v_cloned->decode_indexes.shape[0];
      if (unlikely(__pyx_t_18 < 0)) __pyx_t_8 = 0;
    } else if (unlikely(__pyx_t_18 >= __pyx_v_cloned->decode_indexes.shape[0])) __pyx_t_8 = 0;
    if (unlikely(__pyx_t_8 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_8);
      __PYX_ERR(11, 603, __pyx_L1_error)
    }
    *((Py_ssize_t *) ( /* dim=0 */ (__pyx_v_cloned->decode_indexes.data + __pyx_t_18 * __pyx_v_cloned->decode_indexes.strides[0]) )) = __pyx_t_6;

    /* "src/record.pxi":602
 *         cloned = self.clone_base(PromotingRecordType)
 *         cloned.decode_indexes = cvarray((len(out_decode_indexes), ), sizeof(Py_ssize_t), SSIZE_TYPECODE)
 *         for i, idx in enumerate(out_decode_indexes):             # <<<<<<<<<<<<<<
 *             cloned.decode_indexes[i] = idx
 *         cloned.fields = tuple(out_fields)
 */
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "src/record.pxi":604
 *         for i, idx in enumerate(out_decode_indexes):
 *             cloned.decode_indexes[i] = idx
 *         cloned.fields = tuple(out_fields)             # <<<<<<<<<<<<<<
 *         cloned.field_dict = {f.name: f for f in cloned.fields}
 *         cloned.record = self.record
 */
  __pyx_t_9 = PyList_AsTuple(__pyx_v_out_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_9);
  __Pyx_GOTREF(__pyx_v_cloned->__pyx_base.fields);
  __Pyx_DECREF(__pyx_v_cloned->__pyx_base.fields);
  __pyx_v_cloned->__pyx_base.fields = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "src/record.pxi":605
 *             cloned.decode_indexes[i] = idx
 *         cloned.fields = tuple(out_fields)
 *         cloned.field_dict = {f.name: f for f in cloned.fields}             # <<<<<<<<<<<<<<
 *         cloned.record = self.record
 *         return cloned
 */
  { /* enter inner scope */
    __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 605, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (unlikely(__pyx_v_cloned->__pyx_base.fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(11, 605, __pyx_L23_error)
    }
    __pyx_t_15 = __pyx_v_cloned->__pyx_base.fields; __Pyx_INCREF(__pyx_t_15);
    __pyx_t_4 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_15);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 605, __pyx_L23_error)
        #endif
        if (__pyx_t_4 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_12 = PyTuple_GET_ITEM(__pyx_t_15, __pyx_t_4); __Pyx_INCREF(__pyx_t_12); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(11, 605, __pyx_L23_error)
      #else
      __pyx_t_12 = __Pyx_PySequence_ITEM(__pyx_t_15, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_12)) __PYX_ERR(11, 605, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_12);
      #endif
      __Pyx_XDECREF_SET(__pyx_9genexpr14__pyx_v_f, __pyx_t_12);
      __pyx_t_12 = 0;
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_f, __pyx_n_s_name); if (unlikely(!__pyx_t_12)) __PYX_ERR(11, 605, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_12);
      if (unlikely(PyDict_SetItem(__pyx_t_9, (PyObject*)__pyx_t_12, (PyObject*)__pyx_9genexpr14__pyx_v_f))) __PYX_ERR(11, 605, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    }
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_f); __pyx_9genexpr14__pyx_v_f = 0;
    goto __pyx_L27_exit_scope;
    __pyx_L23_error:;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_f); __pyx_9genexpr14__pyx_v_f = 0;
    goto __pyx_L1_error;
    __pyx_L27_exit_scope:;
  } /* exit inner scope */
  __Pyx_GIVEREF(__pyx_t_9);
  __Pyx_GOTREF(__pyx_v_cloned->__pyx_base.field_dict);
  __Pyx_DECREF(__pyx_v_cloned->__pyx_base.field_dict);
  __pyx_v_cloned->__pyx_base.field_dict = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "src/record.pxi":606
 *         cloned.fields = tuple(out_fields)
 *         cloned.field_dict = {f.name: f for f in cloned.fields}
 *         cloned.record = self.record             # <<<<<<<<<<<<<<
 *         return cloned
 * 
 */
  __pyx_t_9 = ((PyObject *)__pyx_v_self->record);
  __Pyx_INCREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_9);
  __Pyx_GOTREF((PyObject *)__pyx_v_cloned->__pyx_base.record);
  __Pyx_DECREF((PyObject *)__pyx_v_cloned->__pyx_base.record);
  __pyx_v_cloned->__pyx_base.record = ((PyTypeObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "src/record.pxi":607
 *         cloned.field_dict = {f.name: f for f in cloned.fields}
 *         cloned.record = self.record
 *         return cloned             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_cloned);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_cloned);
  goto __pyx_L0;

  /* "src/record.pxi":544
 *         })
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         cdef PromotingRecordType cloned
 *         cdef RecordType writer_rec
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_15);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("cavro.RecordType._for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cloned);
  __Pyx_XDECREF((PyObject *)__pyx_v_writer_rec);
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_reader_field_idx);
  __Pyx_XDECREF(__pyx_v_reader_fields);
  __Pyx_XDECREF(__pyx_v_aliases);
  __Pyx_XDECREF(__pyx_v_alias);
  __Pyx_XDECREF(__pyx_v_out_decode_indexes);
  __Pyx_XDECREF(__pyx_v_out_fields);
  __Pyx_XDECREF(__pyx_v_decoded_reader_fields);
  __Pyx_XDECREF(__pyx_v_reader_index);
  __Pyx_XDECREF(__pyx_v_reader_field);
  __Pyx_XDECREF(__pyx_v_extra_reader_fields);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_f);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":314
 *     type_name = 'record'
 * 
 *     cdef readonly str doc             # <<<<<<<<<<<<<<
 *     cdef readonly tuple fields
 *     cdef dict field_dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10RecordType_3doc_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_10RecordType_3doc_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_10RecordType_3doc___get__(((struct __pyx_obj_5cavro_RecordType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10RecordType_3doc___get__(struct __pyx_obj_5cavro_RecordType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->doc);
  __pyx_r = __pyx_v_self->doc;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":315
 * 
 *     cdef readonly str doc
 *     cdef readonly tuple fields             # <<<<<<<<<<<<<<
 *     cdef dict field_dict
 *     cdef readonly type record
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10RecordType_6fields_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_10RecordType_6fields_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_10RecordType_6fields___get__(((struct __pyx_obj_5cavro_RecordType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10RecordType_6fields___get__(struct __pyx_obj_5cavro_RecordType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->fields);
  __pyx_r = __pyx_v_self->fields;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":317
 *     cdef readonly tuple fields
 *     cdef dict field_dict
 *     cdef readonly type record             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _setting_up
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10RecordType_6record_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_10RecordType_6record_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_10RecordType_6record___get__(((struct __pyx_obj_5cavro_RecordType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10RecordType_6record___get__(struct __pyx_obj_5cavro_RecordType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->record);
  __pyx_r = ((PyObject *)__pyx_v_self->record);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10RecordType_12__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10RecordType_12__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10RecordType_12__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10RecordType_12__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10RecordType_11__reduce_cython__(((struct __pyx_obj_5cavro_RecordType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10RecordType_11__reduce_cython__(struct __pyx_obj_5cavro_RecordType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._setting_up, self.aliases, self.doc, self.effective_namespace, self.field_dict, self.fields, self.metadata, self.name, self.namespace, self.options, self.record, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_setting_up); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(12); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.aliases);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.aliases);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->__pyx_base.aliases)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->doc);
  __Pyx_GIVEREF(__pyx_v_self->doc);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->doc)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.effective_namespace);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.effective_namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_v_self->__pyx_base.effective_namespace)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->field_dict);
  __Pyx_GIVEREF(__pyx_v_self->field_dict);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_v_self->field_dict)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->fields);
  __Pyx_GIVEREF(__pyx_v_self->fields);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_v_self->fields)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_v_self->__pyx_base.__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.name);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_v_self->__pyx_base.name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.namespace);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_v_self->__pyx_base.namespace)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 9, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->record);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->record);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 10, ((PyObject *)__pyx_v_self->record))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 11, __pyx_v_self->__pyx_base.__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._setting_up, self.aliases, self.doc, self.effective_namespace, self.field_dict, self.fields, self.metadata, self.name, self.namespace, self.options, self.record, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self._setting_up, self.aliases, self.doc, self.effective_namespace, self.field_dict, self.fields, self.metadata, self.name, self.namespace, self.options, self.record, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.aliases is not None or self.doc is not None or self.effective_namespace is not None or self.field_dict is not None or self.fields is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.record is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._setting_up, self.aliases, self.doc, self.effective_namespace, self.field_dict, self.fields, self.metadata, self.name, self.namespace, self.options, self.record, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.aliases is not None or self.doc is not None or self.effective_namespace is not None or self.field_dict is not None or self.fields is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.record is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_RecordType, (type(self), 0x182acf2, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base.aliases != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->doc != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.effective_namespace != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->field_dict != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->fields != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.name != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.namespace != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->record != ((PyTypeObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_3 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.doc is not None or self.effective_namespace is not None or self.field_dict is not None or self.fields is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.record is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_RecordType, (type(self), 0x182acf2, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.aliases is not None or self.doc is not None or self.effective_namespace is not None or self.field_dict is not None or self.fields is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.record is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_RecordType, (type(self), 0x182acf2, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_RecordType, (type(self), 0x182acf2, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_RecordType); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_25341170);
    __Pyx_GIVEREF(__pyx_int_25341170);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_25341170)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.doc is not None or self.effective_namespace is not None or self.field_dict is not None or self.fields is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.record is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_RecordType, (type(self), 0x182acf2, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_RecordType, (type(self), 0x182acf2, None), state
 *     else:
 *         return __pyx_unpickle_RecordType, (type(self), 0x182acf2, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_RecordType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_RecordType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_25341170);
    __Pyx_GIVEREF(__pyx_int_25341170);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_25341170)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.RecordType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RecordType, (type(self), 0x182acf2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RecordType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10RecordType_14__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10RecordType_14__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10RecordType_14__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10RecordType_14__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.RecordType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10RecordType_13__setstate_cython__(((struct __pyx_obj_5cavro_RecordType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10RecordType_13__setstate_cython__(struct __pyx_obj_5cavro_RecordType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_RecordType, (type(self), 0x182acf2, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_RecordType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_RecordType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RecordType, (type(self), 0x182acf2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RecordType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.RecordType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/record.pxi":618
 *     cdef Py_ssize_t [:] decode_indexes
 * 
 *     cdef _binary_buffer_decode_record(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         cdef list data = [None] * len(self.fields)
 */

static PyObject *__pyx_f_5cavro_19PromotingRecordType__binary_buffer_decode_record(struct __pyx_obj_5cavro_PromotingRecordType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  struct __pyx_obj_5cavro_RecordField *__pyx_v_field = 0;
  PyObject *__pyx_v_data = 0;
  struct __pyx_obj_5cavro_Record *__pyx_v_rec = 0;
  Py_ssize_t __pyx_v_index;
  Py_ssize_t __pyx_v_field_index;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode_record", 1);

  /* "src/record.pxi":620
 *     cdef _binary_buffer_decode_record(self, _Reader buffer):
 *         cdef RecordField field
 *         cdef list data = [None] * len(self.fields)             # <<<<<<<<<<<<<<
 *         cdef Record rec
 *         cdef Py_ssize_t index = 0
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.fields;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(11, 620, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(11, 620, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(1 * ((__pyx_t_2<0) ? 0:__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_t_2; __pyx_temp++) {
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_1, __pyx_temp, Py_None)) __PYX_ERR(11, 620, __pyx_L1_error);
    }
  }
  __pyx_v_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":622
 *         cdef list data = [None] * len(self.fields)
 *         cdef Record rec
 *         cdef Py_ssize_t index = 0             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t field_index
 *         for field in self.fields:
 */
  __pyx_v_index = 0;

  /* "src/record.pxi":624
 *         cdef Py_ssize_t index = 0
 *         cdef Py_ssize_t field_index
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             field_index = self.decode_indexes[index]
 *             value = field.type.binary_buffer_decode(buffer)
 */
  if (unlikely(__pyx_v_self->__pyx_base.fields == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(11, 624, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->__pyx_base.fields; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(11, 624, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(11, 624, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 624, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_RecordField))))) __PYX_ERR(11, 624, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_field, ((struct __pyx_obj_5cavro_RecordField *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/record.pxi":625
 *         cdef Py_ssize_t field_index
 *         for field in self.fields:
 *             field_index = self.decode_indexes[index]             # <<<<<<<<<<<<<<
 *             value = field.type.binary_buffer_decode(buffer)
 *             if field_index >= 0:
 */
    if (unlikely(!__pyx_v_self->decode_indexes.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(11, 625, __pyx_L1_error)}
    __pyx_t_4 = __pyx_v_index;
    __pyx_t_5 = -1;
    if (__pyx_t_4 < 0) {
      __pyx_t_4 += __pyx_v_self->decode_indexes.shape[0];
      if (unlikely(__pyx_t_4 < 0)) __pyx_t_5 = 0;
    } else if (unlikely(__pyx_t_4 >= __pyx_v_self->decode_indexes.shape[0])) __pyx_t_5 = 0;
    if (unlikely(__pyx_t_5 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_5);
      __PYX_ERR(11, 625, __pyx_L1_error)
    }
    __pyx_v_field_index = (*((Py_ssize_t *) ( /* dim=0 */ (__pyx_v_self->decode_indexes.data + __pyx_t_4 * __pyx_v_self->decode_indexes.strides[0]) )));

    /* "src/record.pxi":626
 *         for field in self.fields:
 *             field_index = self.decode_indexes[index]
 *             value = field.type.binary_buffer_decode(buffer)             # <<<<<<<<<<<<<<
 *             if field_index >= 0:
 *                 data[field_index] = value
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_field->type->__pyx_vtab)->binary_buffer_decode(__pyx_v_field->type, __pyx_v_buffer); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/record.pxi":627
 *             field_index = self.decode_indexes[index]
 *             value = field.type.binary_buffer_decode(buffer)
 *             if field_index >= 0:             # <<<<<<<<<<<<<<
 *                 data[field_index] = value
 *             index += 1
 */
    __pyx_t_6 = (__pyx_v_field_index >= 0);
    if (__pyx_t_6) {

      /* "src/record.pxi":628
 *             value = field.type.binary_buffer_decode(buffer)
 *             if field_index >= 0:
 *                 data[field_index] = value             # <<<<<<<<<<<<<<
 *             index += 1
 *         rec = Record.__new__(self.record)
 */
      if (unlikely((__Pyx_SetItemInt(__pyx_v_data, __pyx_v_field_index, __pyx_v_value, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1) < 0))) __PYX_ERR(11, 628, __pyx_L1_error)

      /* "src/record.pxi":627
 *             field_index = self.decode_indexes[index]
 *             value = field.type.binary_buffer_decode(buffer)
 *             if field_index >= 0:             # <<<<<<<<<<<<<<
 *                 data[field_index] = value
 *             index += 1
 */
    }

    /* "src/record.pxi":629
 *             if field_index >= 0:
 *                 data[field_index] = value
 *             index += 1             # <<<<<<<<<<<<<<
 *         rec = Record.__new__(self.record)
 *         rec.data = data
 */
    __pyx_v_index = (__pyx_v_index + 1);

    /* "src/record.pxi":624
 *         cdef Py_ssize_t index = 0
 *         cdef Py_ssize_t field_index
 *         for field in self.fields:             # <<<<<<<<<<<<<<
 *             field_index = self.decode_indexes[index]
 *             value = field.type.binary_buffer_decode(buffer)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/record.pxi":630
 *                 data[field_index] = value
 *             index += 1
 *         rec = Record.__new__(self.record)             # <<<<<<<<<<<<<<
 *         rec.data = data
 *         return rec
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_Record), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, ((PyObject *)__pyx_v_self->__pyx_base.record)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Record))))) __PYX_ERR(11, 630, __pyx_L1_error)
  __pyx_v_rec = ((struct __pyx_obj_5cavro_Record *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/record.pxi":631
 *             index += 1
 *         rec = Record.__new__(self.record)
 *         rec.data = data             # <<<<<<<<<<<<<<
 *         return rec
 */
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  __Pyx_GOTREF(__pyx_v_rec->data);
  __Pyx_DECREF(__pyx_v_rec->data);
  __pyx_v_rec->data = __pyx_v_data;

  /* "src/record.pxi":632
 *         rec = Record.__new__(self.record)
 *         rec.data = data
 *         return rec             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_rec);
  __pyx_r = ((PyObject *)__pyx_v_rec);
  goto __pyx_L0;

  /* "src/record.pxi":618
 *     cdef Py_ssize_t [:] decode_indexes
 * 
 *     cdef _binary_buffer_decode_record(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef RecordField field
 *         cdef list data = [None] * len(self.fields)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.PromotingRecordType._binary_buffer_decode_record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF((PyObject *)__pyx_v_rec);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_19PromotingRecordType_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_19PromotingRecordType_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_19PromotingRecordType_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_19PromotingRecordType_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_19PromotingRecordType___reduce_cython__(((struct __pyx_obj_5cavro_PromotingRecordType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_19PromotingRecordType___reduce_cython__(struct __pyx_obj_5cavro_PromotingRecordType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._setting_up, self.aliases, self.decode_indexes, self.doc, self.effective_namespace, self.field_dict, self.fields, self.metadata, self.name, self.namespace, self.options, self.record, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base._setting_up); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->decode_indexes.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->decode_indexes, 1, (PyObject *(*)(char *)) __pyx_memview_get_Py_ssize_t, (int (*)(char *, PyObject *)) __pyx_memview_set_Py_ssize_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(13); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.aliases);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.aliases);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->__pyx_base.__pyx_base.aliases)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.doc);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.doc);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_self->__pyx_base.doc)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.effective_namespace);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.effective_namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_self->__pyx_base.__pyx_base.effective_namespace)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.field_dict);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.field_dict);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_v_self->__pyx_base.field_dict)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.fields);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.fields);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_v_self->__pyx_base.fields)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.name);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_v_self->__pyx_base.__pyx_base.name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.namespace);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_v_self->__pyx_base.__pyx_base.namespace)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 10, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.record);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.record);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 11, ((PyObject *)__pyx_v_self->__pyx_base.record))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 12, __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._setting_up, self.aliases, self.decode_indexes, self.doc, self.effective_namespace, self.field_dict, self.fields, self.metadata, self.name, self.namespace, self.options, self.record, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self._setting_up, self.aliases, self.decode_indexes, self.doc, self.effective_namespace, self.field_dict, self.fields, self.metadata, self.name, self.namespace, self.options, self.record, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.aliases is not None or self.doc is not None or self.effective_namespace is not None or self.field_dict is not None or self.fields is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.record is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._setting_up, self.aliases, self.decode_indexes, self.doc, self.effective_namespace, self.field_dict, self.fields, self.metadata, self.name, self.namespace, self.options, self.record, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.aliases is not None or self.doc is not None or self.effective_namespace is not None or self.field_dict is not None or self.fields is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.record is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PromotingRecordType, (type(self), 0x40bba2b, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.aliases != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.doc != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.effective_namespace != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.field_dict != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.fields != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.metadata != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.name != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.namespace != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.options) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.record != ((PyTypeObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.doc is not None or self.effective_namespace is not None or self.field_dict is not None or self.fields is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.record is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PromotingRecordType, (type(self), 0x40bba2b, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.aliases is not None or self.doc is not None or self.effective_namespace is not None or self.field_dict is not None or self.fields is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.record is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PromotingRecordType, (type(self), 0x40bba2b, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PromotingRecordType, (type(self), 0x40bba2b, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_PromotingRecordTy); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_67877419);
    __Pyx_GIVEREF(__pyx_int_67877419);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_67877419)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.doc is not None or self.effective_namespace is not None or self.field_dict is not None or self.fields is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.record is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PromotingRecordType, (type(self), 0x40bba2b, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PromotingRecordType, (type(self), 0x40bba2b, None), state
 *     else:
 *         return __pyx_unpickle_PromotingRecordType, (type(self), 0x40bba2b, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PromotingRecordType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_PromotingRecordTy); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_67877419);
    __Pyx_GIVEREF(__pyx_int_67877419);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_67877419)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.PromotingRecordType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromotingRecordType, (type(self), 0x40bba2b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromotingRecordType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_19PromotingRecordType_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_19PromotingRecordType_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_19PromotingRecordType_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_19PromotingRecordType_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.PromotingRecordType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_19PromotingRecordType_2__setstate_cython__(((struct __pyx_obj_5cavro_PromotingRecordType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_19PromotingRecordType_2__setstate_cython__(struct __pyx_obj_5cavro_PromotingRecordType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PromotingRecordType, (type(self), 0x40bba2b, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PromotingRecordType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_PromotingRecordType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromotingRecordType, (type(self), 0x40bba2b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromotingRecordType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.PromotingRecordType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":7
 *     type_name = "bytes"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

static PyObject *__pyx_pw_5cavro_9BytesType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9BytesType_copy(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "src/string_types.pxi":8
 * 
 *     cpdef AvroType copy(self):
 *         return self.clone_base()             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_BytesType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":7
 *     type_name = "bytes"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BytesType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9BytesType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9BytesType_1copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9BytesType_1copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9BytesType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_9BytesType_copy(((struct __pyx_obj_5cavro_BytesType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9BytesType_copy(struct __pyx_obj_5cavro_BytesType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_9BytesType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BytesType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":10
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

static PyObject *__pyx_f_5cavro_9BytesType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/string_types.pxi":11
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {'type'})             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(12, 11, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":10
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.BytesType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":13
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_pw_5cavro_9BytesType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_9BytesType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_BytesType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);

  /* "src/string_types.pxi":14
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":13
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BytesType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9BytesType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9BytesType_3_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9BytesType_3_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9BytesType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 13, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(12, 13, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(12, 13, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.BytesType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(12, 13, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_9BytesType_2_get_schema_extra(((struct __pyx_obj_5cavro_BytesType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9BytesType_2_get_schema_extra(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_9BytesType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BytesType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":16
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         value = self.convert_value(value)
 *         cdef Py_ssize_t length = len(value)
 */

static int __pyx_f_5cavro_9BytesType__binary_buffer_encode(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  Py_ssize_t __pyx_v_length;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/string_types.pxi":17
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         value = self.convert_value(value)             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t length = len(value)
 *         zigzag_encode_long(buffer, length)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_BytesType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/string_types.pxi":18
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         value = self.convert_value(value)
 *         cdef Py_ssize_t length = len(value)             # <<<<<<<<<<<<<<
 *         zigzag_encode_long(buffer, length)
 *         if length:
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(12, 18, __pyx_L1_error)
  __pyx_v_length = __pyx_t_2;

  /* "src/string_types.pxi":19
 *         value = self.convert_value(value)
 *         cdef Py_ssize_t length = len(value)
 *         zigzag_encode_long(buffer, length)             # <<<<<<<<<<<<<<
 *         if length:
 *             buffer.write_n(value)
 */
  __pyx_t_3 = __pyx_f_5cavro_zigzag_encode_long(__pyx_v_buffer, __pyx_v_length); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(12, 19, __pyx_L1_error)

  /* "src/string_types.pxi":20
 *         cdef Py_ssize_t length = len(value)
 *         zigzag_encode_long(buffer, length)
 *         if length:             # <<<<<<<<<<<<<<
 *             buffer.write_n(value)
 * 
 */
  __pyx_t_4 = (__pyx_v_length != 0);
  if (__pyx_t_4) {

    /* "src/string_types.pxi":21
 *         zigzag_encode_long(buffer, length)
 *         if length:
 *             buffer.write_n(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
    __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_value, 0); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(12, 21, __pyx_L1_error)
    __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_buffer->__pyx_vtab)->write_n(__pyx_v_buffer, __pyx_t_5); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(12, 21, __pyx_L1_error)
    __PYX_XCLEAR_MEMVIEW(&__pyx_t_5, 1);
    __pyx_t_5.memview = NULL; __pyx_t_5.data = NULL;

    /* "src/string_types.pxi":20
 *         cdef Py_ssize_t length = len(value)
 *         zigzag_encode_long(buffer, length)
 *         if length:             # <<<<<<<<<<<<<<
 *             buffer.write_n(value)
 * 
 */
  }

  /* "src/string_types.pxi":16
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         value = self.convert_value(value)
 *         cdef Py_ssize_t length = len(value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_AddTraceback("cavro.BytesType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":23
 *             buffer.write_n(value)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef uint64_t length = zigzag_decode_long(buffer)
 *         return buffer.read_bytes(length)
 */

static PyObject *__pyx_f_5cavro_9BytesType__binary_buffer_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_BytesType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  uint64_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/string_types.pxi":24
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef uint64_t length = zigzag_decode_long(buffer)             # <<<<<<<<<<<<<<
 *         return buffer.read_bytes(length)
 * 
 */
  __pyx_t_1 = __pyx_f_5cavro_zigzag_decode_long(__pyx_v_buffer); if (unlikely(__pyx_t_1 == ((int64_t)0x7ffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(12, 24, __pyx_L1_error)
  __pyx_v_length = __pyx_t_1;

  /* "src/string_types.pxi":25
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef uint64_t length = zigzag_decode_long(buffer)
 *         return buffer.read_bytes(length)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_buffer->__pyx_vtab)->read_bytes(__pyx_v_buffer, __pyx_v_length); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":23
 *             buffer.write_n(value)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef uint64_t length = zigzag_decode_long(buffer)
 *         return buffer.read_bytes(length)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.BytesType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":27
 *         return buffer.read_bytes(length)
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         if isinstance(value, (bytes, bytearray)):  # If bytes, we're good
 *             return FIT_EXACT
 */

static int __pyx_f_5cavro_9BytesType__get_value_fitness(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_v_MAX_FIT;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/string_types.pxi":28
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         if isinstance(value, (bytes, bytearray)):  # If bytes, we're good             # <<<<<<<<<<<<<<
 *             return FIT_EXACT
 *         # If there's a bytes_codec, then if we get get value to be a str, then we have a way through
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_value); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyByteArray_Check(__pyx_v_value); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/string_types.pxi":29
 *     cdef int _get_value_fitness(self, value) except -1:
 *         if isinstance(value, (bytes, bytearray)):  # If bytes, we're good
 *             return FIT_EXACT             # <<<<<<<<<<<<<<
 *         # If there's a bytes_codec, then if we get get value to be a str, then we have a way through
 *         if not self.options.bytes_codec:
 */
    __pyx_r = __pyx_v_5cavro_FIT_EXACT;
    goto __pyx_L0;

    /* "src/string_types.pxi":28
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         if isinstance(value, (bytes, bytearray)):  # If bytes, we're good             # <<<<<<<<<<<<<<
 *             return FIT_EXACT
 *         # If there's a bytes_codec, then if we get get value to be a str, then we have a way through
 */
  }

  /* "src/string_types.pxi":31
 *             return FIT_EXACT
 *         # If there's a bytes_codec, then if we get get value to be a str, then we have a way through
 *         if not self.options.bytes_codec:             # <<<<<<<<<<<<<<
 *             return FIT_NONE
 *         # We may be allowed to coerce to str
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base.options->bytes_codec != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_v_self->__pyx_base.options->bytes_codec) != 0);
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "src/string_types.pxi":32
 *         # If there's a bytes_codec, then if we get get value to be a str, then we have a way through
 *         if not self.options.bytes_codec:
 *             return FIT_NONE             # <<<<<<<<<<<<<<
 *         # We may be allowed to coerce to str
 *         MAX_FIT = FIT_OK
 */
    __pyx_r = __pyx_v_5cavro_FIT_NONE;
    goto __pyx_L0;

    /* "src/string_types.pxi":31
 *             return FIT_EXACT
 *         # If there's a bytes_codec, then if we get get value to be a str, then we have a way through
 *         if not self.options.bytes_codec:             # <<<<<<<<<<<<<<
 *             return FIT_NONE
 *         # We may be allowed to coerce to str
 */
  }

  /* "src/string_types.pxi":34
 *             return FIT_NONE
 *         # We may be allowed to coerce to str
 *         MAX_FIT = FIT_OK             # <<<<<<<<<<<<<<
 *         if not isinstance(value, str):
 *             MAX_FIT = FIT_POOR
 */
  __pyx_v_MAX_FIT = __pyx_v_5cavro_FIT_OK;

  /* "src/string_types.pxi":35
 *         # We may be allowed to coerce to str
 *         MAX_FIT = FIT_OK
 *         if not isinstance(value, str):             # <<<<<<<<<<<<<<
 *             MAX_FIT = FIT_POOR
 *             if self.options.coerce_values_to_str:
 */
  __pyx_t_2 = PyUnicode_Check(__pyx_v_value); 
  __pyx_t_1 = (!__pyx_t_2);
  if (__pyx_t_1) {

    /* "src/string_types.pxi":36
 *         MAX_FIT = FIT_OK
 *         if not isinstance(value, str):
 *             MAX_FIT = FIT_POOR             # <<<<<<<<<<<<<<
 *             if self.options.coerce_values_to_str:
 *                 value = str(value)
 */
    __pyx_v_MAX_FIT = __pyx_v_5cavro_FIT_POOR;

    /* "src/string_types.pxi":37
 *         if not isinstance(value, str):
 *             MAX_FIT = FIT_POOR
 *             if self.options.coerce_values_to_str:             # <<<<<<<<<<<<<<
 *                 value = str(value)
 *             else:
 */
    if (__pyx_v_self->__pyx_base.options->coerce_values_to_str) {

      /* "src/string_types.pxi":38
 *             MAX_FIT = FIT_POOR
 *             if self.options.coerce_values_to_str:
 *                 value = str(value)             # <<<<<<<<<<<<<<
 *             else:
 *                 return FIT_NONE # We can't get a str
 */
      __pyx_t_3 = __Pyx_PyObject_Str(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 38, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "src/string_types.pxi":37
 *         if not isinstance(value, str):
 *             MAX_FIT = FIT_POOR
 *             if self.options.coerce_values_to_str:             # <<<<<<<<<<<<<<
 *                 value = str(value)
 *             else:
 */
      goto __pyx_L8;
    }

    /* "src/string_types.pxi":40
 *                 value = str(value)
 *             else:
 *                 return FIT_NONE # We can't get a str             # <<<<<<<<<<<<<<
 * 
 *         # We now know that value is str and we have a codec
 */
    /*else*/ {
      __pyx_r = __pyx_v_5cavro_FIT_NONE;
      goto __pyx_L0;
    }
    __pyx_L8:;

    /* "src/string_types.pxi":35
 *         # We may be allowed to coerce to str
 *         MAX_FIT = FIT_OK
 *         if not isinstance(value, str):             # <<<<<<<<<<<<<<
 *             MAX_FIT = FIT_POOR
 *             if self.options.coerce_values_to_str:
 */
  }

  /* "src/string_types.pxi":43
 * 
 *         # We now know that value is str and we have a codec
 *         try:             # <<<<<<<<<<<<<<
 *             value.encode(self.options.bytes_codec)
 *         except (TypeError, ValueError, UnicodeEncodeError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "src/string_types.pxi":44
 *         # We now know that value is str and we have a codec
 *         try:
 *             value.encode(self.options.bytes_codec)             # <<<<<<<<<<<<<<
 *         except (TypeError, ValueError, UnicodeEncodeError):
 *             return FIT_NONE
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_7)) __PYX_ERR(12, 44, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_self->__pyx_base.options->bytes_codec};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 44, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "src/string_types.pxi":43
 * 
 *         # We now know that value is str and we have a codec
 *         try:             # <<<<<<<<<<<<<<
 *             value.encode(self.options.bytes_codec)
 *         except (TypeError, ValueError, UnicodeEncodeError):
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L14_try_end;
    __pyx_L9_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "src/string_types.pxi":45
 *         try:
 *             value.encode(self.options.bytes_codec)
 *         except (TypeError, ValueError, UnicodeEncodeError):             # <<<<<<<<<<<<<<
 *             return FIT_NONE
 *         return MAX_FIT
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeEncodeError);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("cavro.BytesType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(12, 45, __pyx_L11_except_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "src/string_types.pxi":46
 *             value.encode(self.options.bytes_codec)
 *         except (TypeError, ValueError, UnicodeEncodeError):
 *             return FIT_NONE             # <<<<<<<<<<<<<<
 *         return MAX_FIT
 * 
 */
      __pyx_r = __pyx_v_5cavro_FIT_NONE;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L12_except_return;
    }
    goto __pyx_L11_except_error;

    /* "src/string_types.pxi":43
 * 
 *         # We now know that value is str and we have a codec
 *         try:             # <<<<<<<<<<<<<<
 *             value.encode(self.options.bytes_codec)
 *         except (TypeError, ValueError, UnicodeEncodeError):
 */
    __pyx_L11_except_error:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L12_except_return:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L0;
    __pyx_L14_try_end:;
  }

  /* "src/string_types.pxi":47
 *         except (TypeError, ValueError, UnicodeEncodeError):
 *             return FIT_NONE
 *         return MAX_FIT             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
  __pyx_r = __pyx_v_MAX_FIT;
  goto __pyx_L0;

  /* "src/string_types.pxi":27
 *         return buffer.read_bytes(length)
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         if isinstance(value, (bytes, bytearray)):  # If bytes, we're good
 *             return FIT_EXACT
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cavro.BytesType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":49
 *         return MAX_FIT
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         value = self.convert_value(value)
 *         return value.decode('latin-1')
 */

static PyObject *__pyx_f_5cavro_9BytesType__json_format(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/string_types.pxi":50
 * 
 *     cdef _json_format(self, value):
 *         value = self.convert_value(value)             # <<<<<<<<<<<<<<
 *         return value.decode('latin-1')
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_BytesType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/string_types.pxi":51
 *     cdef _json_format(self, value):
 *         value = self.convert_value(value)
 *         return value.decode('latin-1')             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_decode); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_latin_1};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":49
 *         return MAX_FIT
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         value = self.convert_value(value)
 *         return value.decode('latin-1')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.BytesType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":53
 *         return value.decode('latin-1')
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         cdef str sval = value
 *         return sval.encode('latin-1')
 */

static PyObject *__pyx_f_5cavro_9BytesType__json_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_sval = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/string_types.pxi":54
 * 
 *     cdef _json_decode(self, value):
 *         cdef str sval = value             # <<<<<<<<<<<<<<
 *         return sval.encode('latin-1')
 * 
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_value))) __PYX_ERR(12, 54, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_sval = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/string_types.pxi":55
 *     cdef _json_decode(self, value):
 *         cdef str sval = value
 *         return sval.encode('latin-1')             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_sval == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(12, 55, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsLatin1String(__pyx_v_sval); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":53
 *         return value.decode('latin-1')
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         cdef str sval = value
 *         return sval.encode('latin-1')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BytesType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sval);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":57
 *         return sval.encode('latin-1')
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (bytes, bytearray)):
 *             return value
 */

static PyObject *__pyx_pw_5cavro_9BytesType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_9BytesType__convert_value(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_codec = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/string_types.pxi":58
 * 
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, (bytes, bytearray)):             # <<<<<<<<<<<<<<
 *             return value
 *         codec = self.options.bytes_codec
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_value); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyByteArray_Check(__pyx_v_value); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/string_types.pxi":59
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, (bytes, bytearray)):
 *             return value             # <<<<<<<<<<<<<<
 *         codec = self.options.bytes_codec
 *         if not codec:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/string_types.pxi":58
 * 
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, (bytes, bytearray)):             # <<<<<<<<<<<<<<
 *             return value
 *         codec = self.options.bytes_codec
 */
  }

  /* "src/string_types.pxi":60
 *         if isinstance(value, (bytes, bytearray)):
 *             return value
 *         codec = self.options.bytes_codec             # <<<<<<<<<<<<<<
 *         if not codec:
 *             raise InvalidValue(value, self)
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.options->bytes_codec;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_codec = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/string_types.pxi":61
 *             return value
 *         codec = self.options.bytes_codec
 *         if not codec:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if not isinstance(value, str):
 */
  __pyx_t_1 = (__pyx_v_codec != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_v_codec) != 0);
  __pyx_t_2 = (!__pyx_t_1);
  if (unlikely(__pyx_t_2)) {

    /* "src/string_types.pxi":62
 *         codec = self.options.bytes_codec
 *         if not codec:
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         if not isinstance(value, str):
 *             if self.options.coerce_values_to_str:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_4)) __PYX_ERR(12, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 62, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(12, 62, __pyx_L1_error)

    /* "src/string_types.pxi":61
 *             return value
 *         codec = self.options.bytes_codec
 *         if not codec:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if not isinstance(value, str):
 */
  }

  /* "src/string_types.pxi":63
 *         if not codec:
 *             raise InvalidValue(value, self)
 *         if not isinstance(value, str):             # <<<<<<<<<<<<<<
 *             if self.options.coerce_values_to_str:
 *                 value = str(value)
 */
  __pyx_t_2 = PyUnicode_Check(__pyx_v_value); 
  __pyx_t_1 = (!__pyx_t_2);
  if (__pyx_t_1) {

    /* "src/string_types.pxi":64
 *             raise InvalidValue(value, self)
 *         if not isinstance(value, str):
 *             if self.options.coerce_values_to_str:             # <<<<<<<<<<<<<<
 *                 value = str(value)
 *             raise InvalidValue(value, self)
 */
    if (__pyx_v_self->__pyx_base.options->coerce_values_to_str) {

      /* "src/string_types.pxi":65
 *         if not isinstance(value, str):
 *             if self.options.coerce_values_to_str:
 *                 value = str(value)             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         # We have a str, and codec
 */
      __pyx_t_3 = __Pyx_PyObject_Str(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 65, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "src/string_types.pxi":64
 *             raise InvalidValue(value, self)
 *         if not isinstance(value, str):
 *             if self.options.coerce_values_to_str:             # <<<<<<<<<<<<<<
 *                 value = str(value)
 *             raise InvalidValue(value, self)
 */
    }

    /* "src/string_types.pxi":66
 *             if self.options.coerce_values_to_str:
 *                 value = str(value)
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         # We have a str, and codec
 *         return value.encode(codec)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_4)) __PYX_ERR(12, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 66, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(12, 66, __pyx_L1_error)

    /* "src/string_types.pxi":63
 *         if not codec:
 *             raise InvalidValue(value, self)
 *         if not isinstance(value, str):             # <<<<<<<<<<<<<<
 *             if self.options.coerce_values_to_str:
 *                 value = str(value)
 */
  }

  /* "src/string_types.pxi":68
 *             raise InvalidValue(value, self)
 *         # We have a str, and codec
 *         return value.encode(codec)             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(12, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_codec};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":57
 *         return sval.encode('latin-1')
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (bytes, bytearray)):
 *             return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.BytesType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_codec);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9BytesType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9BytesType_5_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9BytesType_5_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9BytesType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 57, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(12, 57, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(12, 57, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.BytesType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9BytesType_4_convert_value(((struct __pyx_obj_5cavro_BytesType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9BytesType_4_convert_value(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_9BytesType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BytesType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":70
 *         return value.encode(codec)
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"bytes"')
 * 
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_9BytesType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_BytesType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/string_types.pxi":71
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"bytes"')             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_tuple__67, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":70
 *         return value.encode(codec)
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"bytes"')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BytesType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":73
 *         return _CanonicalForm('"bytes"')
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         if isinstance(writer, StringType):
 *             return self # Decoding is identical for bytes / str
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9BytesType__for_writer(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_for_writer", 1);

  /* "src/string_types.pxi":74
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if isinstance(writer, StringType):             # <<<<<<<<<<<<<<
 *             return self # Decoding is identical for bytes / str
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_StringType); 
  if (__pyx_t_1) {

    /* "src/string_types.pxi":75
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if isinstance(writer, StringType):
 *             return self # Decoding is identical for bytes / str             # <<<<<<<<<<<<<<
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self);
    goto __pyx_L0;

    /* "src/string_types.pxi":74
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if isinstance(writer, StringType):             # <<<<<<<<<<<<<<
 *             return self # Decoding is identical for bytes / str
 * 
 */
  }

  /* "src/string_types.pxi":73
 *         return _CanonicalForm('"bytes"')
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         if isinstance(writer, StringType):
 *             return self # Decoding is identical for bytes / str
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":77
 *             return self # Decoding is identical for bytes / str
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):             # <<<<<<<<<<<<<<
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 */

static PyObject *__pyx_f_5cavro_9BytesType_resolve_default_value(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v_schema_default, PyObject *__pyx_v_field) {
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  Py_UCS4 __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resolve_default_value", 0);
  __Pyx_INCREF(__pyx_v_schema_default);

  /* "src/string_types.pxi":78
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if schema_default is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             return NO_DEFAULT
 *         if self.options.string_types_default_unchanged:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_v_schema_default == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/string_types.pxi":79
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT             # <<<<<<<<<<<<<<
 *         if self.options.string_types_default_unchanged:
 *             return schema_default
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/string_types.pxi":78
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if schema_default is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             return NO_DEFAULT
 *         if self.options.string_types_default_unchanged:
 */
  }

  /* "src/string_types.pxi":80
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 *         if self.options.string_types_default_unchanged:             # <<<<<<<<<<<<<<
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:
 */
  if (__pyx_v_self->__pyx_base.options->string_types_default_unchanged) {

    /* "src/string_types.pxi":81
 *             return NO_DEFAULT
 *         if self.options.string_types_default_unchanged:
 *             return schema_default             # <<<<<<<<<<<<<<
 *         if self.options.bytes_default_value_utf8:
 *             try:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_schema_default);
    __pyx_r = __pyx_v_schema_default;
    goto __pyx_L0;

    /* "src/string_types.pxi":80
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 *         if self.options.string_types_default_unchanged:             # <<<<<<<<<<<<<<
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:
 */
  }

  /* "src/string_types.pxi":82
 *         if self.options.string_types_default_unchanged:
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:             # <<<<<<<<<<<<<<
 *             try:
 *                 schema_default = schema_default.encode('utf-8')
 */
  if (__pyx_v_self->__pyx_base.options->bytes_default_value_utf8) {

    /* "src/string_types.pxi":83
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:
 *             try:             # <<<<<<<<<<<<<<
 *                 schema_default = schema_default.encode('utf-8')
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "src/string_types.pxi":84
 *         if self.options.bytes_default_value_utf8:
 *             try:
 *                 schema_default = schema_default.encode('utf-8')             # <<<<<<<<<<<<<<
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:
 *                 raise TypeError(f"Default value {schema_default!r} is not valid for bytes field: {field}") from e
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema_default, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 84, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_kp_u_utf_8};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 84, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF_SET(__pyx_v_schema_default, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "src/string_types.pxi":83
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:
 *             try:             # <<<<<<<<<<<<<<
 *                 schema_default = schema_default.encode('utf-8')
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L11_try_end;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "src/string_types.pxi":85
 *             try:
 *                 schema_default = schema_default.encode('utf-8')
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:             # <<<<<<<<<<<<<<
 *                 raise TypeError(f"Default value {schema_default!r} is not valid for bytes field: {field}") from e
 *             return schema_default
 */
      __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeEncodeError);
      if (__pyx_t_8) {
        __Pyx_AddTraceback("cavro.BytesType.resolve_default_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(12, 85, __pyx_L8_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __pyx_v_e = __pyx_t_6;
        /*try:*/ {

          /* "src/string_types.pxi":86
 *                 schema_default = schema_default.encode('utf-8')
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:
 *                 raise TypeError(f"Default value {schema_default!r} is not valid for bytes field: {field}") from e             # <<<<<<<<<<<<<<
 *             return schema_default
 *         return AvroType.resolve_default_value(self, schema_default, field)
 */
          __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 86, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_10 = 0;
          __pyx_t_11 = 127;
          __Pyx_INCREF(__pyx_kp_u_Default_value_2);
          __pyx_t_10 += 14;
          __Pyx_GIVEREF(__pyx_kp_u_Default_value_2);
          PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_kp_u_Default_value_2);
          __pyx_t_12 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_schema_default), __pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(12, 86, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
          __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_12);
          __pyx_t_12 = 0;
          __Pyx_INCREF(__pyx_kp_u_is_not_valid_for_bytes_field);
          __pyx_t_10 += 31;
          __Pyx_GIVEREF(__pyx_kp_u_is_not_valid_for_bytes_field);
          PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_kp_u_is_not_valid_for_bytes_field);
          __pyx_t_12 = __Pyx_PyUnicode_Unicode(__pyx_v_field); if (unlikely(!__pyx_t_12)) __PYX_ERR(12, 86, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
          __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_12);
          __pyx_t_12 = 0;
          __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_9, 4, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(12, 86, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 86, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_Raise(__pyx_t_9, 0, 0, __pyx_v_e);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __PYX_ERR(12, 86, __pyx_L17_error)
        }

        /* "src/string_types.pxi":85
 *             try:
 *                 schema_default = schema_default.encode('utf-8')
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:             # <<<<<<<<<<<<<<
 *                 raise TypeError(f"Default value {schema_default!r} is not valid for bytes field: {field}") from e
 *             return schema_default
 */
        /*finally:*/ {
          __pyx_L17_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_8 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
        }
      }
      goto __pyx_L8_except_error;

      /* "src/string_types.pxi":83
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:
 *             try:             # <<<<<<<<<<<<<<
 *                 schema_default = schema_default.encode('utf-8')
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:
 */
      __pyx_L8_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L11_try_end:;
    }

    /* "src/string_types.pxi":87
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:
 *                 raise TypeError(f"Default value {schema_default!r} is not valid for bytes field: {field}") from e
 *             return schema_default             # <<<<<<<<<<<<<<
 *         return AvroType.resolve_default_value(self, schema_default, field)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_schema_default);
    __pyx_r = __pyx_v_schema_default;
    goto __pyx_L0;

    /* "src/string_types.pxi":82
 *         if self.options.string_types_default_unchanged:
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:             # <<<<<<<<<<<<<<
 *             try:
 *                 schema_default = schema_default.encode('utf-8')
 */
  }

  /* "src/string_types.pxi":88
 *                 raise TypeError(f"Default value {schema_default!r} is not valid for bytes field: {field}") from e
 *             return schema_default
 *         return AvroType.resolve_default_value(self, schema_default, field)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = __pyx_f_5cavro_8AvroType_resolve_default_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_schema_default, __pyx_v_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(12, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":77
 *             return self # Decoding is identical for bytes / str
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):             # <<<<<<<<<<<<<<
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("cavro.BytesType.resolve_default_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_schema_default);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9BytesType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9BytesType_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9BytesType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9BytesType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_9BytesType_6__reduce_cython__(((struct __pyx_obj_5cavro_BytesType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9BytesType_6__reduce_cython__(struct __pyx_obj_5cavro_BytesType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_BytesType, (type(self), 0x84c5042, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BytesType, (type(self), 0x84c5042, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_BytesType, (type(self), 0x84c5042, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_BytesType, (type(self), 0x84c5042, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_BytesType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BytesType, (type(self), 0x84c5042, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_BytesType, (type(self), 0x84c5042, None), state
 *     else:
 *         return __pyx_unpickle_BytesType, (type(self), 0x84c5042, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BytesType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_BytesType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.BytesType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BytesType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BytesType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9BytesType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9BytesType_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9BytesType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9BytesType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.BytesType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9BytesType_8__setstate_cython__(((struct __pyx_obj_5cavro_BytesType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9BytesType_8__setstate_cython__(struct __pyx_obj_5cavro_BytesType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_BytesType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BytesType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_BytesType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BytesType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BytesType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.BytesType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":97
 *     type_name = "string"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

static PyObject *__pyx_pw_5cavro_10StringType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10StringType_copy(struct __pyx_obj_5cavro_StringType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "src/string_types.pxi":98
 * 
 *     cpdef AvroType copy(self):
 *         return self.clone_base()             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_StringType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":97
 *     type_name = "string"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.StringType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10StringType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10StringType_1copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10StringType_1copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10StringType_1copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10StringType_copy(((struct __pyx_obj_5cavro_StringType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10StringType_copy(struct __pyx_obj_5cavro_StringType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_10StringType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.StringType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":100
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

static PyObject *__pyx_f_5cavro_10StringType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/string_types.pxi":101
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {'type'})             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(12, 101, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":100
 *         return self.clone_base()
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type'})
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.StringType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":103
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

static PyObject *__pyx_pw_5cavro_10StringType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_10StringType__get_schema_extra(CYTHON_UNUSED struct __pyx_obj_5cavro_StringType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);

  /* "src/string_types.pxi":104
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":103
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.StringType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10StringType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10StringType_3_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10StringType_3_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10StringType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 103, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(12, 103, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(12, 103, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.StringType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(12, 103, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_10StringType_2_get_schema_extra(((struct __pyx_obj_5cavro_StringType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10StringType_2_get_schema_extra(struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_10StringType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.StringType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":106
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         value = self._convert_value(value).encode('utf8')
 *         cdef size_t length = len(value)
 */

static int __pyx_f_5cavro_10StringType__binary_buffer_encode(struct __pyx_obj_5cavro_StringType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  size_t __pyx_v_length;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/string_types.pxi":107
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         value = self._convert_value(value).encode('utf8')             # <<<<<<<<<<<<<<
 *         cdef size_t length = len(value)
 *         zigzag_encode_long(buffer, length)
 */
  __pyx_t_2 = __pyx_f_5cavro_10StringType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_n_u_utf8};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/string_types.pxi":108
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         value = self._convert_value(value).encode('utf8')
 *         cdef size_t length = len(value)             # <<<<<<<<<<<<<<
 *         zigzag_encode_long(buffer, length)
 *         if length:
 */
  __pyx_t_5 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(12, 108, __pyx_L1_error)
  __pyx_v_length = __pyx_t_5;

  /* "src/string_types.pxi":109
 *         value = self._convert_value(value).encode('utf8')
 *         cdef size_t length = len(value)
 *         zigzag_encode_long(buffer, length)             # <<<<<<<<<<<<<<
 *         if length:
 *             buffer.write_n(value)
 */
  __pyx_t_4 = __pyx_f_5cavro_zigzag_encode_long(__pyx_v_buffer, __pyx_v_length); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(12, 109, __pyx_L1_error)

  /* "src/string_types.pxi":110
 *         cdef size_t length = len(value)
 *         zigzag_encode_long(buffer, length)
 *         if length:             # <<<<<<<<<<<<<<
 *             buffer.write_n(value)
 * 
 */
  __pyx_t_6 = (__pyx_v_length != 0);
  if (__pyx_t_6) {

    /* "src/string_types.pxi":111
 *         zigzag_encode_long(buffer, length)
 *         if length:
 *             buffer.write_n(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_value, 0); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(12, 111, __pyx_L1_error)
    __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_buffer->__pyx_vtab)->write_n(__pyx_v_buffer, __pyx_t_7); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(12, 111, __pyx_L1_error)
    __PYX_XCLEAR_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL; __pyx_t_7.data = NULL;

    /* "src/string_types.pxi":110
 *         cdef size_t length = len(value)
 *         zigzag_encode_long(buffer, length)
 *         if length:             # <<<<<<<<<<<<<<
 *             buffer.write_n(value)
 * 
 */
  }

  /* "src/string_types.pxi":106
 *         return {}
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         value = self._convert_value(value).encode('utf8')
 *         cdef size_t length = len(value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_AddTraceback("cavro.StringType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":113
 *             buffer.write_n(value)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef uint64_t length = zigzag_decode_long(buffer)
 *         return buffer.read_bytes(length).decode('utf-8', errors=self.options.unicode_errors)
 */

static PyObject *__pyx_f_5cavro_10StringType__binary_buffer_decode(struct __pyx_obj_5cavro_StringType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  uint64_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/string_types.pxi":114
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef uint64_t length = zigzag_decode_long(buffer)             # <<<<<<<<<<<<<<
 *         return buffer.read_bytes(length).decode('utf-8', errors=self.options.unicode_errors)
 * 
 */
  __pyx_t_1 = __pyx_f_5cavro_zigzag_decode_long(__pyx_v_buffer); if (unlikely(__pyx_t_1 == ((int64_t)0x7ffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(12, 114, __pyx_L1_error)
  __pyx_v_length = __pyx_t_1;

  /* "src/string_types.pxi":115
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef uint64_t length = zigzag_decode_long(buffer)
 *         return buffer.read_bytes(length).decode('utf-8', errors=self.options.unicode_errors)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_buffer->__pyx_vtab)->read_bytes(__pyx_v_buffer, __pyx_v_length); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_errors, __pyx_v_self->__pyx_base.options->unicode_errors) < 0) __PYX_ERR(12, 115, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__68, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(12, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":113
 *             buffer.write_n(value)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef uint64_t length = zigzag_decode_long(buffer)
 *         return buffer.read_bytes(length).decode('utf-8', errors=self.options.unicode_errors)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.StringType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":117
 *         return buffer.read_bytes(length).decode('utf-8', errors=self.options.unicode_errors)
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         if isinstance(value, str):
 *             return FIT_EXACT
 */

static int __pyx_f_5cavro_10StringType__get_value_fitness(struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  int __pyx_t_1;

  /* "src/string_types.pxi":118
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         if isinstance(value, str):             # <<<<<<<<<<<<<<
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_str:
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_value); 
  if (__pyx_t_1) {

    /* "src/string_types.pxi":119
 *     cdef int _get_value_fitness(self, value) except -1:
 *         if isinstance(value, str):
 *             return FIT_EXACT             # <<<<<<<<<<<<<<
 *         if self.options.coerce_values_to_str:
 *             if isinstance(value, bytes):
 */
    __pyx_r = __pyx_v_5cavro_FIT_EXACT;
    goto __pyx_L0;

    /* "src/string_types.pxi":118
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         if isinstance(value, str):             # <<<<<<<<<<<<<<
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_str:
 */
  }

  /* "src/string_types.pxi":120
 *         if isinstance(value, str):
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_str:             # <<<<<<<<<<<<<<
 *             if isinstance(value, bytes):
 *                 return FIT_OK
 */
  if (__pyx_v_self->__pyx_base.options->coerce_values_to_str) {

    /* "src/string_types.pxi":121
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_str:
 *             if isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 return FIT_OK
 *             return FIT_POOR
 */
    __pyx_t_1 = PyBytes_Check(__pyx_v_value); 
    if (__pyx_t_1) {

      /* "src/string_types.pxi":122
 *         if self.options.coerce_values_to_str:
 *             if isinstance(value, bytes):
 *                 return FIT_OK             # <<<<<<<<<<<<<<
 *             return FIT_POOR
 *         return FIT_NONE
 */
      __pyx_r = __pyx_v_5cavro_FIT_OK;
      goto __pyx_L0;

      /* "src/string_types.pxi":121
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_str:
 *             if isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 return FIT_OK
 *             return FIT_POOR
 */
    }

    /* "src/string_types.pxi":123
 *             if isinstance(value, bytes):
 *                 return FIT_OK
 *             return FIT_POOR             # <<<<<<<<<<<<<<
 *         return FIT_NONE
 * 
 */
    __pyx_r = __pyx_v_5cavro_FIT_POOR;
    goto __pyx_L0;

    /* "src/string_types.pxi":120
 *         if isinstance(value, str):
 *             return FIT_EXACT
 *         if self.options.coerce_values_to_str:             # <<<<<<<<<<<<<<
 *             if isinstance(value, bytes):
 *                 return FIT_OK
 */
  }

  /* "src/string_types.pxi":124
 *                 return FIT_OK
 *             return FIT_POOR
 *         return FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
  __pyx_r = __pyx_v_5cavro_FIT_NONE;
  goto __pyx_L0;

  /* "src/string_types.pxi":117
 *         return buffer.read_bytes(length).decode('utf-8', errors=self.options.unicode_errors)
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         if isinstance(value, str):
 *             return FIT_EXACT
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "src/string_types.pxi":126
 *         return FIT_NONE
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */

static PyObject *__pyx_f_5cavro_10StringType__json_format(struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 1);

  /* "src/string_types.pxi":127
 * 
 *     cdef _json_format(self, value):
 *         return self._convert_value(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_10StringType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":126
 *         return FIT_NONE
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.StringType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":129
 *         return self._convert_value(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         cdef str sval = value
 *         return sval
 */

static PyObject *__pyx_f_5cavro_10StringType__json_decode(CYTHON_UNUSED struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_sval = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/string_types.pxi":130
 * 
 *     cdef _json_decode(self, value):
 *         cdef str sval = value             # <<<<<<<<<<<<<<
 *         return sval
 * 
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_value))) __PYX_ERR(12, 130, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_sval = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/string_types.pxi":131
 *     cdef _json_decode(self, value):
 *         cdef str sval = value
 *         return sval             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_sval);
  __pyx_r = __pyx_v_sval;
  goto __pyx_L0;

  /* "src/string_types.pxi":129
 *         return self._convert_value(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         cdef str sval = value
 *         return sval
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.StringType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sval);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":133
 *         return sval
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, str):
 *             return value
 */

static PyObject *__pyx_pw_5cavro_10StringType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_10StringType__convert_value(struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);

  /* "src/string_types.pxi":134
 * 
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, str):             # <<<<<<<<<<<<<<
 *             return value
 *         if not self.options.coerce_values_to_str:
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_value); 
  if (__pyx_t_1) {

    /* "src/string_types.pxi":135
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, str):
 *             return value             # <<<<<<<<<<<<<<
 *         if not self.options.coerce_values_to_str:
 *             raise InvalidValue(value, self)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/string_types.pxi":134
 * 
 *     cpdef object _convert_value(self, object value):
 *         if isinstance(value, str):             # <<<<<<<<<<<<<<
 *             return value
 *         if not self.options.coerce_values_to_str:
 */
  }

  /* "src/string_types.pxi":136
 *         if isinstance(value, str):
 *             return value
 *         if not self.options.coerce_values_to_str:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if isinstance(value, bytes):
 */
  __pyx_t_1 = (!__pyx_v_self->__pyx_base.options->coerce_values_to_str);
  if (unlikely(__pyx_t_1)) {

    /* "src/string_types.pxi":137
 *             return value
 *         if not self.options.coerce_values_to_str:
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         if isinstance(value, bytes):
 *             return value.decode(errors=self.options.unicode_errors)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(12, 137, __pyx_L1_error)

    /* "src/string_types.pxi":136
 *         if isinstance(value, str):
 *             return value
 *         if not self.options.coerce_values_to_str:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         if isinstance(value, bytes):
 */
  }

  /* "src/string_types.pxi":138
 *         if not self.options.coerce_values_to_str:
 *             raise InvalidValue(value, self)
 *         if isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *             return value.decode(errors=self.options.unicode_errors)
 *         return str(value)
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_value); 
  if (__pyx_t_1) {

    /* "src/string_types.pxi":139
 *             raise InvalidValue(value, self)
 *         if isinstance(value, bytes):
 *             return value.decode(errors=self.options.unicode_errors)             # <<<<<<<<<<<<<<
 *         return str(value)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_decode); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_errors, __pyx_v_self->__pyx_base.options->unicode_errors) < 0) __PYX_ERR(12, 139, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(12, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "src/string_types.pxi":138
 *         if not self.options.coerce_values_to_str:
 *             raise InvalidValue(value, self)
 *         if isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *             return value.decode(errors=self.options.unicode_errors)
 *         return str(value)
 */
  }

  /* "src/string_types.pxi":140
 *         if isinstance(value, bytes):
 *             return value.decode(errors=self.options.unicode_errors)
 *         return str(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_Str(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(12, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":133
 *         return sval
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, str):
 *             return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.StringType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10StringType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10StringType_5_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10StringType_5_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10StringType_5_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 133, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(12, 133, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(12, 133, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.StringType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10StringType_4_convert_value(((struct __pyx_obj_5cavro_StringType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10StringType_4_convert_value(struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_10StringType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.StringType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":142
 *         return str(value)
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"string"')
 * 
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_10StringType_canonical_form(CYTHON_UNUSED struct __pyx_obj_5cavro_StringType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/string_types.pxi":143
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"string"')             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_tuple__69, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":142
 *         return str(value)
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('"string"')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.StringType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":145
 *         return _CanonicalForm('"string"')
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         if isinstance(writer, BytesType):
 *             return self
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10StringType__for_writer(struct __pyx_obj_5cavro_StringType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_for_writer", 1);

  /* "src/string_types.pxi":146
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if isinstance(writer, BytesType):             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_BytesType); 
  if (__pyx_t_1) {

    /* "src/string_types.pxi":147
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if isinstance(writer, BytesType):
 *             return self             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self);
    goto __pyx_L0;

    /* "src/string_types.pxi":146
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         if isinstance(writer, BytesType):             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "src/string_types.pxi":145
 *         return _CanonicalForm('"string"')
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         if isinstance(writer, BytesType):
 *             return self
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10StringType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10StringType_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10StringType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10StringType_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10StringType_6__reduce_cython__(((struct __pyx_obj_5cavro_StringType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10StringType_6__reduce_cython__(struct __pyx_obj_5cavro_StringType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_StringType, (type(self), 0x84c5042, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_StringType, (type(self), 0x84c5042, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_StringType, (type(self), 0x84c5042, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_StringType, (type(self), 0x84c5042, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_StringType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_StringType, (type(self), 0x84c5042, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_StringType, (type(self), 0x84c5042, None), state
 *     else:
 *         return __pyx_unpickle_StringType, (type(self), 0x84c5042, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_StringType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_StringType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.StringType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_StringType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_StringType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10StringType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10StringType_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10StringType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10StringType_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.StringType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10StringType_8__setstate_cython__(((struct __pyx_obj_5cavro_StringType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10StringType_8__setstate_cython__(struct __pyx_obj_5cavro_StringType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_StringType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_StringType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_StringType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_StringType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_StringType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.StringType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":157
 *     cdef readonly Py_ssize_t size
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         self.size = source['size']
 *         if self.size < 0:
 */

/* Python wrapper */
static int __pyx_pw_5cavro_9FixedType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_9FixedType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_namespace = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_schema,&__pyx_n_s_source,&__pyx_n_s_namespace,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 157, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 157, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(12, 157, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_namespace)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 157, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(12, 157, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(12, 157, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_schema = values[0];
    __pyx_v_source = values[1];
    __pyx_v_namespace = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(12, 157, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.FixedType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9FixedType___init__(((struct __pyx_obj_5cavro_FixedType *)__pyx_v_self), __pyx_v_schema, __pyx_v_source, __pyx_v_namespace);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_9FixedType___init__(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/string_types.pxi":158
 * 
 *     def __init__(self, schema, source, namespace):
 *         self.size = source['size']             # <<<<<<<<<<<<<<
 *         if self.size < 0:
 *             raise ValueError(f'Invalid negative size for fixed: {self.size}')
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_source, __pyx_n_u_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(12, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->size = __pyx_t_2;

  /* "src/string_types.pxi":159
 *     def __init__(self, schema, source, namespace):
 *         self.size = source['size']
 *         if self.size < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError(f'Invalid negative size for fixed: {self.size}')
 *         super().__init__(schema, source, namespace)
 */
  __pyx_t_3 = (__pyx_v_self->size < 0);
  if (unlikely(__pyx_t_3)) {

    /* "src/string_types.pxi":160
 *         self.size = source['size']
 *         if self.size < 0:
 *             raise ValueError(f'Invalid negative size for fixed: {self.size}')             # <<<<<<<<<<<<<<
 *         super().__init__(schema, source, namespace)
 * 
 */
    __pyx_t_1 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_self->size, 0, ' ', 'd'); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Invalid_negative_size_for_fixed, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(12, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(12, 160, __pyx_L1_error)

    /* "src/string_types.pxi":159
 *     def __init__(self, schema, source, namespace):
 *         self.size = source['size']
 *         if self.size < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError(f'Invalid negative size for fixed: {self.size}')
 *         super().__init__(schema, source, namespace)
 */
  }

  /* "src/string_types.pxi":161
 *         if self.size < 0:
 *             raise ValueError(f'Invalid negative size for fixed: {self.size}')
 *         super().__init__(schema, source, namespace)             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(12, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_FixedType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_FixedType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_ptype_5cavro_FixedType))) __PYX_ERR(12, 161, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(12, 161, __pyx_L1_error);
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(12, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_init); if (unlikely(!__pyx_t_4)) __PYX_ERR(12, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_v_schema, __pyx_v_source, __pyx_v_namespace};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/string_types.pxi":157
 *     cdef readonly Py_ssize_t size
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         self.size = source['size']
 *         if self.size < 0:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.FixedType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":163
 *         super().__init__(schema, source, namespace)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef FixedType new_inst =  self.clone_base()
 *         new_inst.size = self.size
 */

static PyObject *__pyx_pw_5cavro_9FixedType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9FixedType_copy(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_FixedType *__pyx_v_new_inst = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "src/string_types.pxi":164
 * 
 *     cpdef AvroType copy(self):
 *         cdef FixedType new_inst =  self.clone_base()             # <<<<<<<<<<<<<<
 *         new_inst.size = self.size
 *         return new_inst
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_FixedType *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_FixedType))))) __PYX_ERR(12, 164, __pyx_L1_error)
  __pyx_v_new_inst = ((struct __pyx_obj_5cavro_FixedType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/string_types.pxi":165
 *     cpdef AvroType copy(self):
 *         cdef FixedType new_inst =  self.clone_base()
 *         new_inst.size = self.size             # <<<<<<<<<<<<<<
 *         return new_inst
 * 
 */
  __pyx_t_2 = __pyx_v_self->size;
  __pyx_v_new_inst->size = __pyx_t_2;

  /* "src/string_types.pxi":166
 *         cdef FixedType new_inst =  self.clone_base()
 *         new_inst.size = self.size
 *         return new_inst             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_new_inst);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_new_inst);
  goto __pyx_L0;

  /* "src/string_types.pxi":163
 *         super().__init__(schema, source, namespace)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef FixedType new_inst =  self.clone_base()
 *         new_inst.size = self.size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FixedType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_new_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9FixedType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9FixedType_3copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FixedType_3copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9FixedType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_9FixedType_2copy(((struct __pyx_obj_5cavro_FixedType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9FixedType_2copy(struct __pyx_obj_5cavro_FixedType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_9FixedType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FixedType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":168
 *         return new_inst
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type', 'name', 'namespace', 'aliases', 'size'})
 * 
 */

static PyObject *__pyx_f_5cavro_9FixedType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/string_types.pxi":169
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {'type', 'name', 'namespace', 'aliases', 'size'})             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(12, 169, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_name) < 0) __PYX_ERR(12, 169, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_namespace) < 0) __PYX_ERR(12, 169, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_aliases) < 0) __PYX_ERR(12, 169, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_size) < 0) __PYX_ERR(12, 169, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":168
 *         return new_inst
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {'type', 'name', 'namespace', 'aliases', 'size'})
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.FixedType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":171
 *         return _strip_keys(dict(source), {'type', 'name', 'namespace', 'aliases', 'size'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return dict(_NamedType._get_schema_extra(self, created), size=self.size)
 * 
 */

static PyObject *__pyx_pw_5cavro_9FixedType_5_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_9FixedType__get_schema_extra(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_created, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);

  /* "src/string_types.pxi":172
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         return dict(_NamedType._get_schema_extra(self, created), size=self.size)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_10_NamedType__get_schema_extra(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self), __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(12, 172, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_self->size); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_size, __pyx_t_3) < 0) __PYX_ERR(12, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)(&PyDict_Type)), __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":171
 *         return _strip_keys(dict(source), {'type', 'name', 'namespace', 'aliases', 'size'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return dict(_NamedType._get_schema_extra(self, created), size=self.size)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.FixedType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9FixedType_5_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9FixedType_5_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FixedType_5_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9FixedType_5_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 171, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(12, 171, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(12, 171, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.FixedType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(12, 171, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_9FixedType_4_get_schema_extra(((struct __pyx_obj_5cavro_FixedType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9FixedType_4_get_schema_extra(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_9FixedType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FixedType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":174
 *         return dict(_NamedType._get_schema_extra(self, created), size=self.size)
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         value = self._convert_value(value)
 *         cdef Py_ssize_t length = len(value)
 */

static int __pyx_f_5cavro_9FixedType__binary_buffer_encode(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  Py_ssize_t __pyx_v_length;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/string_types.pxi":175
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         value = self._convert_value(value)             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t length = len(value)
 *         if length != self.size:
 */
  __pyx_t_1 = __pyx_f_5cavro_9FixedType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/string_types.pxi":176
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         value = self._convert_value(value)
 *         cdef Py_ssize_t length = len(value)             # <<<<<<<<<<<<<<
 *         if length != self.size:
 *             raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(12, 176, __pyx_L1_error)
  __pyx_v_length = __pyx_t_2;

  /* "src/string_types.pxi":177
 *         value = self._convert_value(value)
 *         cdef Py_ssize_t length = len(value)
 *         if length != self.size:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")
 *         buffer.write_n(value)
 */
  __pyx_t_3 = (__pyx_v_length != __pyx_v_self->size);
  if (unlikely(__pyx_t_3)) {

    /* "src/string_types.pxi":178
 *         cdef Py_ssize_t length = len(value)
 *         if length != self.size:
 *             raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")             # <<<<<<<<<<<<<<
 *         buffer.write_n(value)
 * 
 */
    __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u_Invalid_length_for_fixed_field);
    __pyx_t_2 += 32;
    __Pyx_GIVEREF(__pyx_kp_u_Invalid_length_for_fixed_field);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Invalid_length_for_fixed_field);
    __pyx_t_5 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_length, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(12, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__70);
    __pyx_t_2 += 4;
    __Pyx_GIVEREF(__pyx_kp_u__70);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__70);
    __pyx_t_5 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_self->size, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(12, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_value_2);
    __pyx_t_2 += 9;
    __Pyx_GIVEREF(__pyx_kp_u_value_2);
    PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_value_2);
    __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(12, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__7);
    __pyx_t_2 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__7);
    PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__7);
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(12, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(12, 178, __pyx_L1_error)

    /* "src/string_types.pxi":177
 *         value = self._convert_value(value)
 *         cdef Py_ssize_t length = len(value)
 *         if length != self.size:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")
 *         buffer.write_n(value)
 */
  }

  /* "src/string_types.pxi":179
 *         if length != self.size:
 *             raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")
 *         buffer.write_n(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_value, 0); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(12, 179, __pyx_L1_error)
  __pyx_t_7 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_buffer->__pyx_vtab)->write_n(__pyx_v_buffer, __pyx_t_6); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(12, 179, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __pyx_t_6.memview = NULL; __pyx_t_6.data = NULL;

  /* "src/string_types.pxi":174
 *         return dict(_NamedType._get_schema_extra(self, created), size=self.size)
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         value = self._convert_value(value)
 *         cdef Py_ssize_t length = len(value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("cavro.FixedType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":181
 *         buffer.write_n(value)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return bytes(buffer.read_n(self.size))
 * 
 */

static PyObject *__pyx_f_5cavro_9FixedType__binary_buffer_decode(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/string_types.pxi":182
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         return bytes(buffer.read_n(self.size))             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_buffer->__pyx_vtab)->read_n(__pyx_v_buffer, __pyx_v_self->size); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(12, 182, __pyx_L1_error)
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_t_1, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_1, 1);
  __pyx_t_1.memview = NULL; __pyx_t_1.data = NULL;
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":181
 *         buffer.write_n(value)
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         return bytes(buffer.read_n(self.size))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.FixedType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":184
 *         return bytes(buffer.read_n(self.size))
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         MAX_FIT = FIT_EXACT
 *         if not isinstance(value, (bytes, bytearray)):
 */

static int __pyx_f_5cavro_9FixedType__get_value_fitness(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED int __pyx_v_MAX_FIT;
  Py_ssize_t __pyx_v_length;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/string_types.pxi":185
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         MAX_FIT = FIT_EXACT             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (bytes, bytearray)):
 *             MAX_FIT = FIT_OK
 */
  __pyx_v_MAX_FIT = __pyx_v_5cavro_FIT_EXACT;

  /* "src/string_types.pxi":186
 *     cdef int _get_value_fitness(self, value) except -1:
 *         MAX_FIT = FIT_EXACT
 *         if not isinstance(value, (bytes, bytearray)):             # <<<<<<<<<<<<<<
 *             MAX_FIT = FIT_OK
 *             # Let's see if we can recover this
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_value); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyByteArray_Check(__pyx_v_value); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "src/string_types.pxi":187
 *         MAX_FIT = FIT_EXACT
 *         if not isinstance(value, (bytes, bytearray)):
 *             MAX_FIT = FIT_OK             # <<<<<<<<<<<<<<
 *             # Let's see if we can recover this
 *             if self.options.fixed_codec is None:
 */
    __pyx_v_MAX_FIT = __pyx_v_5cavro_FIT_OK;

    /* "src/string_types.pxi":189
 *             MAX_FIT = FIT_OK
 *             # Let's see if we can recover this
 *             if self.options.fixed_codec is None:             # <<<<<<<<<<<<<<
 *                 return FIT_NONE # No codec, so can't encode
 * 
 */
    __pyx_t_2 = (__pyx_v_self->__pyx_base.__pyx_base.options->fixed_codec == ((PyObject*)Py_None));
    if (__pyx_t_2) {

      /* "src/string_types.pxi":190
 *             # Let's see if we can recover this
 *             if self.options.fixed_codec is None:
 *                 return FIT_NONE # No codec, so can't encode             # <<<<<<<<<<<<<<
 * 
 *             # It's not even a str, but can we coerce it?
 */
      __pyx_r = __pyx_v_5cavro_FIT_NONE;
      goto __pyx_L0;

      /* "src/string_types.pxi":189
 *             MAX_FIT = FIT_OK
 *             # Let's see if we can recover this
 *             if self.options.fixed_codec is None:             # <<<<<<<<<<<<<<
 *                 return FIT_NONE # No codec, so can't encode
 * 
 */
    }

    /* "src/string_types.pxi":193
 * 
 *             # It's not even a str, but can we coerce it?
 *             if not isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 if not self.options.coerce_values_to_str: # Not allowed
 *                     return FIT_NONE
 */
    __pyx_t_2 = PyUnicode_Check(__pyx_v_value); 
    __pyx_t_1 = (!__pyx_t_2);
    if (__pyx_t_1) {

      /* "src/string_types.pxi":194
 *             # It's not even a str, but can we coerce it?
 *             if not isinstance(value, str):
 *                 if not self.options.coerce_values_to_str: # Not allowed             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 *                 MAX_FIT = FIT_POOR
 */
      __pyx_t_1 = (!__pyx_v_self->__pyx_base.__pyx_base.options->coerce_values_to_str);
      if (__pyx_t_1) {

        /* "src/string_types.pxi":195
 *             if not isinstance(value, str):
 *                 if not self.options.coerce_values_to_str: # Not allowed
 *                     return FIT_NONE             # <<<<<<<<<<<<<<
 *                 MAX_FIT = FIT_POOR
 *                 try:
 */
        __pyx_r = __pyx_v_5cavro_FIT_NONE;
        goto __pyx_L0;

        /* "src/string_types.pxi":194
 *             # It's not even a str, but can we coerce it?
 *             if not isinstance(value, str):
 *                 if not self.options.coerce_values_to_str: # Not allowed             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 *                 MAX_FIT = FIT_POOR
 */
      }

      /* "src/string_types.pxi":196
 *                 if not self.options.coerce_values_to_str: # Not allowed
 *                     return FIT_NONE
 *                 MAX_FIT = FIT_POOR             # <<<<<<<<<<<<<<
 *                 try:
 *                     value = str(value)
 */
      __pyx_v_MAX_FIT = __pyx_v_5cavro_FIT_POOR;

      /* "src/string_types.pxi":197
 *                     return FIT_NONE
 *                 MAX_FIT = FIT_POOR
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = str(value)
 *                 except (TypeError, ValueError, UnicodeEncodeError):
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_5);
        /*try:*/ {

          /* "src/string_types.pxi":198
 *                 MAX_FIT = FIT_POOR
 *                 try:
 *                     value = str(value)             # <<<<<<<<<<<<<<
 *                 except (TypeError, ValueError, UnicodeEncodeError):
 *                     return FIT_NONE
 */
          __pyx_t_6 = __Pyx_PyObject_Str(__pyx_v_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 198, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "src/string_types.pxi":197
 *                     return FIT_NONE
 *                 MAX_FIT = FIT_POOR
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = str(value)
 *                 except (TypeError, ValueError, UnicodeEncodeError):
 */
        }
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L14_try_end;
        __pyx_L9_error:;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "src/string_types.pxi":199
 *                 try:
 *                     value = str(value)
 *                 except (TypeError, ValueError, UnicodeEncodeError):             # <<<<<<<<<<<<<<
 *                     return FIT_NONE
 * 
 */
        __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeEncodeError);
        if (__pyx_t_7) {
          __Pyx_AddTraceback("cavro.FixedType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(12, 199, __pyx_L11_except_error)
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_9);

          /* "src/string_types.pxi":200
 *                     value = str(value)
 *                 except (TypeError, ValueError, UnicodeEncodeError):
 *                     return FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *             # We have a str, and a codec
 */
          __pyx_r = __pyx_v_5cavro_FIT_NONE;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L12_except_return;
        }
        goto __pyx_L11_except_error;

        /* "src/string_types.pxi":197
 *                     return FIT_NONE
 *                 MAX_FIT = FIT_POOR
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = str(value)
 *                 except (TypeError, ValueError, UnicodeEncodeError):
 */
        __pyx_L11_except_error:;
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
        goto __pyx_L1_error;
        __pyx_L12_except_return:;
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
        goto __pyx_L0;
        __pyx_L14_try_end:;
      }

      /* "src/string_types.pxi":193
 * 
 *             # It's not even a str, but can we coerce it?
 *             if not isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 if not self.options.coerce_values_to_str: # Not allowed
 *                     return FIT_NONE
 */
    }

    /* "src/string_types.pxi":203
 * 
 *             # We have a str, and a codec
 *             try:             # <<<<<<<<<<<<<<
 *                 value = value.encode(self.options.fixed_codec)
 *             except (TypeError, ValueError, UnicodeEncodeError):
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      /*try:*/ {

        /* "src/string_types.pxi":204
 *             # We have a str, and a codec
 *             try:
 *                 value = value.encode(self.options.fixed_codec)             # <<<<<<<<<<<<<<
 *             except (TypeError, ValueError, UnicodeEncodeError):
 *                 return FIT_NONE
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(12, 204, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_self->__pyx_base.__pyx_base.options->fixed_codec};
          __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 204, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_9);
        __pyx_t_9 = 0;

        /* "src/string_types.pxi":203
 * 
 *             # We have a str, and a codec
 *             try:             # <<<<<<<<<<<<<<
 *                 value = value.encode(self.options.fixed_codec)
 *             except (TypeError, ValueError, UnicodeEncodeError):
 */
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L22_try_end;
      __pyx_L17_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "src/string_types.pxi":205
 *             try:
 *                 value = value.encode(self.options.fixed_codec)
 *             except (TypeError, ValueError, UnicodeEncodeError):             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *         # We have bytes now, check the length
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeEncodeError);
      if (__pyx_t_7) {
        __Pyx_AddTraceback("cavro.FixedType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(12, 205, __pyx_L19_except_error)
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);

        /* "src/string_types.pxi":206
 *                 value = value.encode(self.options.fixed_codec)
 *             except (TypeError, ValueError, UnicodeEncodeError):
 *                 return FIT_NONE             # <<<<<<<<<<<<<<
 *         # We have bytes now, check the length
 *         cdef Py_ssize_t length = len(value)
 */
        __pyx_r = __pyx_v_5cavro_FIT_NONE;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L20_except_return;
      }
      goto __pyx_L19_except_error;

      /* "src/string_types.pxi":203
 * 
 *             # We have a str, and a codec
 *             try:             # <<<<<<<<<<<<<<
 *                 value = value.encode(self.options.fixed_codec)
 *             except (TypeError, ValueError, UnicodeEncodeError):
 */
      __pyx_L19_except_error:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
      goto __pyx_L1_error;
      __pyx_L20_except_return:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
      goto __pyx_L0;
      __pyx_L22_try_end:;
    }

    /* "src/string_types.pxi":186
 *     cdef int _get_value_fitness(self, value) except -1:
 *         MAX_FIT = FIT_EXACT
 *         if not isinstance(value, (bytes, bytearray)):             # <<<<<<<<<<<<<<
 *             MAX_FIT = FIT_OK
 *             # Let's see if we can recover this
 */
  }

  /* "src/string_types.pxi":208
 *                 return FIT_NONE
 *         # We have bytes now, check the length
 *         cdef Py_ssize_t length = len(value)             # <<<<<<<<<<<<<<
 *         if length == self.size:
 *             return FIT_EXACT
 */
  __pyx_t_10 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(12, 208, __pyx_L1_error)
  __pyx_v_length = __pyx_t_10;

  /* "src/string_types.pxi":209
 *         # We have bytes now, check the length
 *         cdef Py_ssize_t length = len(value)
 *         if length == self.size:             # <<<<<<<<<<<<<<
 *             return FIT_EXACT
 *         if length < self.size and self.options.null_pad_fixed:
 */
  __pyx_t_1 = (__pyx_v_length == __pyx_v_self->size);
  if (__pyx_t_1) {

    /* "src/string_types.pxi":210
 *         cdef Py_ssize_t length = len(value)
 *         if length == self.size:
 *             return FIT_EXACT             # <<<<<<<<<<<<<<
 *         if length < self.size and self.options.null_pad_fixed:
 *             return FIT_OK
 */
    __pyx_r = __pyx_v_5cavro_FIT_EXACT;
    goto __pyx_L0;

    /* "src/string_types.pxi":209
 *         # We have bytes now, check the length
 *         cdef Py_ssize_t length = len(value)
 *         if length == self.size:             # <<<<<<<<<<<<<<
 *             return FIT_EXACT
 *         if length < self.size and self.options.null_pad_fixed:
 */
  }

  /* "src/string_types.pxi":211
 *         if length == self.size:
 *             return FIT_EXACT
 *         if length < self.size and self.options.null_pad_fixed:             # <<<<<<<<<<<<<<
 *             return FIT_OK
 *         if self.options.truncate_fixed:
 */
  __pyx_t_2 = (__pyx_v_length < __pyx_v_self->size);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L27_bool_binop_done;
  }
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.options->null_pad_fixed;
  __pyx_L27_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/string_types.pxi":212
 *             return FIT_EXACT
 *         if length < self.size and self.options.null_pad_fixed:
 *             return FIT_OK             # <<<<<<<<<<<<<<
 *         if self.options.truncate_fixed:
 *             return FIT_OK
 */
    __pyx_r = __pyx_v_5cavro_FIT_OK;
    goto __pyx_L0;

    /* "src/string_types.pxi":211
 *         if length == self.size:
 *             return FIT_EXACT
 *         if length < self.size and self.options.null_pad_fixed:             # <<<<<<<<<<<<<<
 *             return FIT_OK
 *         if self.options.truncate_fixed:
 */
  }

  /* "src/string_types.pxi":213
 *         if length < self.size and self.options.null_pad_fixed:
 *             return FIT_OK
 *         if self.options.truncate_fixed:             # <<<<<<<<<<<<<<
 *             return FIT_OK
 *         return FIT_NONE
 */
  if (__pyx_v_self->__pyx_base.__pyx_base.options->truncate_fixed) {

    /* "src/string_types.pxi":214
 *             return FIT_OK
 *         if self.options.truncate_fixed:
 *             return FIT_OK             # <<<<<<<<<<<<<<
 *         return FIT_NONE
 * 
 */
    __pyx_r = __pyx_v_5cavro_FIT_OK;
    goto __pyx_L0;

    /* "src/string_types.pxi":213
 *         if length < self.size and self.options.null_pad_fixed:
 *             return FIT_OK
 *         if self.options.truncate_fixed:             # <<<<<<<<<<<<<<
 *             return FIT_OK
 *         return FIT_NONE
 */
  }

  /* "src/string_types.pxi":215
 *         if self.options.truncate_fixed:
 *             return FIT_OK
 *         return FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
  __pyx_r = __pyx_v_5cavro_FIT_NONE;
  goto __pyx_L0;

  /* "src/string_types.pxi":184
 *         return bytes(buffer.read_n(self.size))
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         MAX_FIT = FIT_EXACT
 *         if not isinstance(value, (bytes, bytearray)):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cavro.FixedType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":217
 *         return FIT_NONE
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         value = self._convert_value(value)
 *         return value.decode('latin1', errors=self.options.unicode_errors)
 */

static PyObject *__pyx_f_5cavro_9FixedType__json_format(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/string_types.pxi":218
 * 
 *     cdef _json_format(self, value):
 *         value = self._convert_value(value)             # <<<<<<<<<<<<<<
 *         return value.decode('latin1', errors=self.options.unicode_errors)
 * 
 */
  __pyx_t_1 = __pyx_f_5cavro_9FixedType__convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/string_types.pxi":219
 *     cdef _json_format(self, value):
 *         value = self._convert_value(value)
 *         return value.decode('latin1', errors=self.options.unicode_errors)             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_decode); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_errors, __pyx_v_self->__pyx_base.__pyx_base.options->unicode_errors) < 0) __PYX_ERR(12, 219, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__71, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":217
 *         return FIT_NONE
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         value = self._convert_value(value)
 *         return value.decode('latin1', errors=self.options.unicode_errors)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.FixedType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":221
 *         return value.decode('latin1', errors=self.options.unicode_errors)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         cdef str sval = value
 *         cdef bytes encoded = sval.encode('latin1')
 */

static PyObject *__pyx_f_5cavro_9FixedType__json_decode(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_sval = 0;
  PyObject *__pyx_v_encoded = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/string_types.pxi":222
 * 
 *     cdef _json_decode(self, value):
 *         cdef str sval = value             # <<<<<<<<<<<<<<
 *         cdef bytes encoded = sval.encode('latin1')
 *         if len(encoded) != self.size:
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_value))) __PYX_ERR(12, 222, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_sval = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/string_types.pxi":223
 *     cdef _json_decode(self, value):
 *         cdef str sval = value
 *         cdef bytes encoded = sval.encode('latin1')             # <<<<<<<<<<<<<<
 *         if len(encoded) != self.size:
 *             raise ValueError(f"Invalid length for fixed field: {len(encoded)} != {self.size} (value: {value})")
 */
  if (unlikely(__pyx_v_sval == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(12, 223, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsLatin1String(__pyx_v_sval); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_encoded = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/string_types.pxi":224
 *         cdef str sval = value
 *         cdef bytes encoded = sval.encode('latin1')
 *         if len(encoded) != self.size:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Invalid length for fixed field: {len(encoded)} != {self.size} (value: {value})")
 *         return encoded
 */
  if (unlikely(__pyx_v_encoded == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(12, 224, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_GET_SIZE(__pyx_v_encoded); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(12, 224, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 != __pyx_v_self->size);
  if (unlikely(__pyx_t_3)) {

    /* "src/string_types.pxi":225
 *         cdef bytes encoded = sval.encode('latin1')
 *         if len(encoded) != self.size:
 *             raise ValueError(f"Invalid length for fixed field: {len(encoded)} != {self.size} (value: {value})")             # <<<<<<<<<<<<<<
 *         return encoded
 * 
 */
    __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u_Invalid_length_for_fixed_field);
    __pyx_t_2 += 32;
    __Pyx_GIVEREF(__pyx_kp_u_Invalid_length_for_fixed_field);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Invalid_length_for_fixed_field);
    if (unlikely(__pyx_v_encoded == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(12, 225, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyBytes_GET_SIZE(__pyx_v_encoded); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(12, 225, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_5, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__70);
    __pyx_t_2 += 4;
    __Pyx_GIVEREF(__pyx_kp_u__70);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__70);
    __pyx_t_6 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_self->size, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_value_2);
    __pyx_t_2 += 9;
    __Pyx_GIVEREF(__pyx_kp_u_value_2);
    PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_value_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_4;
    __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__7);
    __pyx_t_2 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__7);
    PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__7);
    __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(12, 225, __pyx_L1_error)

    /* "src/string_types.pxi":224
 *         cdef str sval = value
 *         cdef bytes encoded = sval.encode('latin1')
 *         if len(encoded) != self.size:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Invalid length for fixed field: {len(encoded)} != {self.size} (value: {value})")
 *         return encoded
 */
  }

  /* "src/string_types.pxi":226
 *         if len(encoded) != self.size:
 *             raise ValueError(f"Invalid length for fixed field: {len(encoded)} != {self.size} (value: {value})")
 *         return encoded             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_encoded);
  __pyx_r = __pyx_v_encoded;
  goto __pyx_L0;

  /* "src/string_types.pxi":221
 *         return value.decode('latin1', errors=self.options.unicode_errors)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         cdef str sval = value
 *         cdef bytes encoded = sval.encode('latin1')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.FixedType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sval);
  __Pyx_XDECREF(__pyx_v_encoded);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":228
 *         return encoded
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (bytes, bytearray)):
 *             # Let's see if we can recover this
 */

static PyObject *__pyx_pw_5cavro_9FixedType_7_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_9FixedType__convert_value(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/string_types.pxi":229
 * 
 *     cpdef object _convert_value(self, object value):
 *         if not isinstance(value, (bytes, bytearray)):             # <<<<<<<<<<<<<<
 *             # Let's see if we can recover this
 *             if self.options.fixed_codec is None:
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_value); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyByteArray_Check(__pyx_v_value); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "src/string_types.pxi":231
 *         if not isinstance(value, (bytes, bytearray)):
 *             # Let's see if we can recover this
 *             if self.options.fixed_codec is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Invalid non-bytes value for fixed: '{value}'")
 * 
 */
    __pyx_t_2 = (__pyx_v_self->__pyx_base.__pyx_base.options->fixed_codec == ((PyObject*)Py_None));
    if (unlikely(__pyx_t_2)) {

      /* "src/string_types.pxi":232
 *             # Let's see if we can recover this
 *             if self.options.fixed_codec is None:
 *                 raise ValueError(f"Invalid non-bytes value for fixed: '{value}'")             # <<<<<<<<<<<<<<
 * 
 *             # It's not even a str, but can we coerce it?
 */
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = 127;
      __Pyx_INCREF(__pyx_kp_u_Invalid_non_bytes_value_for_fixe);
      __pyx_t_4 += 36;
      __Pyx_GIVEREF(__pyx_kp_u_Invalid_non_bytes_value_for_fixe);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Invalid_non_bytes_value_for_fixe);
      __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
      __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_kp_u__6);
      __pyx_t_4 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__6);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__6);
      __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(12, 232, __pyx_L1_error)

      /* "src/string_types.pxi":231
 *         if not isinstance(value, (bytes, bytearray)):
 *             # Let's see if we can recover this
 *             if self.options.fixed_codec is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Invalid non-bytes value for fixed: '{value}'")
 * 
 */
    }

    /* "src/string_types.pxi":235
 * 
 *             # It's not even a str, but can we coerce it?
 *             if not isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 if not self.options.coerce_values_to_str: # Not allowed
 *                     raise ValueError(f"Invalid non-str value for fixed: '{value}'")
 */
    __pyx_t_2 = PyUnicode_Check(__pyx_v_value); 
    __pyx_t_1 = (!__pyx_t_2);
    if (__pyx_t_1) {

      /* "src/string_types.pxi":236
 *             # It's not even a str, but can we coerce it?
 *             if not isinstance(value, str):
 *                 if not self.options.coerce_values_to_str: # Not allowed             # <<<<<<<<<<<<<<
 *                     raise ValueError(f"Invalid non-str value for fixed: '{value}'")
 *                 value = str(value)
 */
      __pyx_t_1 = (!__pyx_v_self->__pyx_base.__pyx_base.options->coerce_values_to_str);
      if (unlikely(__pyx_t_1)) {

        /* "src/string_types.pxi":237
 *             if not isinstance(value, str):
 *                 if not self.options.coerce_values_to_str: # Not allowed
 *                     raise ValueError(f"Invalid non-str value for fixed: '{value}'")             # <<<<<<<<<<<<<<
 *                 value = str(value)
 * 
 */
        __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = 0;
        __pyx_t_5 = 127;
        __Pyx_INCREF(__pyx_kp_u_Invalid_non_str_value_for_fixed);
        __pyx_t_4 += 34;
        __Pyx_GIVEREF(__pyx_kp_u_Invalid_non_str_value_for_fixed);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Invalid_non_str_value_for_fixed);
        __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
        __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
        __pyx_t_6 = 0;
        __Pyx_INCREF(__pyx_kp_u__6);
        __pyx_t_4 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__6);
        PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__6);
        __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(12, 237, __pyx_L1_error)

        /* "src/string_types.pxi":236
 *             # It's not even a str, but can we coerce it?
 *             if not isinstance(value, str):
 *                 if not self.options.coerce_values_to_str: # Not allowed             # <<<<<<<<<<<<<<
 *                     raise ValueError(f"Invalid non-str value for fixed: '{value}'")
 *                 value = str(value)
 */
      }

      /* "src/string_types.pxi":238
 *                 if not self.options.coerce_values_to_str: # Not allowed
 *                     raise ValueError(f"Invalid non-str value for fixed: '{value}'")
 *                 value = str(value)             # <<<<<<<<<<<<<<
 * 
 *             value = value.encode(self.options.fixed_codec)
 */
      __pyx_t_3 = __Pyx_PyObject_Str(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 238, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "src/string_types.pxi":235
 * 
 *             # It's not even a str, but can we coerce it?
 *             if not isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 if not self.options.coerce_values_to_str: # Not allowed
 *                     raise ValueError(f"Invalid non-str value for fixed: '{value}'")
 */
    }

    /* "src/string_types.pxi":240
 *                 value = str(value)
 * 
 *             value = value.encode(self.options.fixed_codec)             # <<<<<<<<<<<<<<
 *         # We have bytes now, check the length
 *         cdef Py_ssize_t length = len(value)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 240, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_self->__pyx_base.__pyx_base.options->fixed_codec};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/string_types.pxi":229
 * 
 *     cpdef object _convert_value(self, object value):
 *         if not isinstance(value, (bytes, bytearray)):             # <<<<<<<<<<<<<<
 *             # Let's see if we can recover this
 *             if self.options.fixed_codec is None:
 */
  }

  /* "src/string_types.pxi":242
 *             value = value.encode(self.options.fixed_codec)
 *         # We have bytes now, check the length
 *         cdef Py_ssize_t length = len(value)             # <<<<<<<<<<<<<<
 *         if length == self.size:
 *             return value
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(12, 242, __pyx_L1_error)
  __pyx_v_length = __pyx_t_4;

  /* "src/string_types.pxi":243
 *         # We have bytes now, check the length
 *         cdef Py_ssize_t length = len(value)
 *         if length == self.size:             # <<<<<<<<<<<<<<
 *             return value
 *         if length < self.size and self.options.null_pad_fixed:
 */
  __pyx_t_1 = (__pyx_v_length == __pyx_v_self->size);
  if (__pyx_t_1) {

    /* "src/string_types.pxi":244
 *         cdef Py_ssize_t length = len(value)
 *         if length == self.size:
 *             return value             # <<<<<<<<<<<<<<
 *         if length < self.size and self.options.null_pad_fixed:
 *             return value + b'\x00' * (self.size - length)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/string_types.pxi":243
 *         # We have bytes now, check the length
 *         cdef Py_ssize_t length = len(value)
 *         if length == self.size:             # <<<<<<<<<<<<<<
 *             return value
 *         if length < self.size and self.options.null_pad_fixed:
 */
  }

  /* "src/string_types.pxi":245
 *         if length == self.size:
 *             return value
 *         if length < self.size and self.options.null_pad_fixed:             # <<<<<<<<<<<<<<
 *             return value + b'\x00' * (self.size - length)
 *         if self.options.truncate_fixed:
 */
  __pyx_t_2 = (__pyx_v_length < __pyx_v_self->size);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.options->null_pad_fixed;
  __pyx_L11_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/string_types.pxi":246
 *             return value
 *         if length < self.size and self.options.null_pad_fixed:
 *             return value + b'\x00' * (self.size - length)             # <<<<<<<<<<<<<<
 *         if self.options.truncate_fixed:
 *             return value[:self.size]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PySequence_Multiply(__pyx_kp_b__72, (__pyx_v_self->size - __pyx_v_length)); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = PyNumber_Add(__pyx_v_value, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "src/string_types.pxi":245
 *         if length == self.size:
 *             return value
 *         if length < self.size and self.options.null_pad_fixed:             # <<<<<<<<<<<<<<
 *             return value + b'\x00' * (self.size - length)
 *         if self.options.truncate_fixed:
 */
  }

  /* "src/string_types.pxi":247
 *         if length < self.size and self.options.null_pad_fixed:
 *             return value + b'\x00' * (self.size - length)
 *         if self.options.truncate_fixed:             # <<<<<<<<<<<<<<
 *             return value[:self.size]
 *         raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")
 */
  if (__pyx_v_self->__pyx_base.__pyx_base.options->truncate_fixed) {

    /* "src/string_types.pxi":248
 *             return value + b'\x00' * (self.size - length)
 *         if self.options.truncate_fixed:
 *             return value[:self.size]             # <<<<<<<<<<<<<<
 *         raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_6 = __Pyx_PyObject_GetSlice(__pyx_v_value, 0, __pyx_v_self->size, NULL, NULL, NULL, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "src/string_types.pxi":247
 *         if length < self.size and self.options.null_pad_fixed:
 *             return value + b'\x00' * (self.size - length)
 *         if self.options.truncate_fixed:             # <<<<<<<<<<<<<<
 *             return value[:self.size]
 *         raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")
 */
  }

  /* "src/string_types.pxi":249
 *         if self.options.truncate_fixed:
 *             return value[:self.size]
 *         raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __pyx_t_6 = PyTuple_New(7); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = 0;
  __pyx_t_5 = 127;
  __Pyx_INCREF(__pyx_kp_u_Invalid_length_for_fixed_field);
  __pyx_t_4 += 32;
  __Pyx_GIVEREF(__pyx_kp_u_Invalid_length_for_fixed_field);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_u_Invalid_length_for_fixed_field);
  __pyx_t_3 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_length, 0, ' ', 'd'); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u__70);
  __pyx_t_4 += 4;
  __Pyx_GIVEREF(__pyx_kp_u__70);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_kp_u__70);
  __pyx_t_3 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_self->size, 0, ' ', 'd'); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u_value_2);
  __pyx_t_4 += 9;
  __Pyx_GIVEREF(__pyx_kp_u_value_2);
  PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_kp_u_value_2);
  __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_5;
  __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u__7);
  __pyx_t_4 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__7);
  PyTuple_SET_ITEM(__pyx_t_6, 6, __pyx_kp_u__7);
  __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_6, 7, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_6, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __PYX_ERR(12, 249, __pyx_L1_error)

  /* "src/string_types.pxi":228
 *         return encoded
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (bytes, bytearray)):
 *             # Let's see if we can recover this
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.FixedType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9FixedType_7_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9FixedType_7_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FixedType_7_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9FixedType_7_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 228, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(12, 228, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(12, 228, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.FixedType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9FixedType_6_convert_value(((struct __pyx_obj_5cavro_FixedType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9FixedType_6_convert_value(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_9FixedType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FixedType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":251
 *         raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         if self in created and not self.options.canonical_form_repeat_fixed:
 *             return _CanonicalForm(f'"{self.type}"')
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_9FixedType_canonical_form(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/string_types.pxi":252
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         if self in created and not self.options.canonical_form_repeat_fixed:             # <<<<<<<<<<<<<<
 *             return _CanonicalForm(f'"{self.type}"')
 *         created.add(self)
 */
  if (unlikely(__pyx_v_created == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(12, 252, __pyx_L1_error)
  }
  __pyx_t_2 = (__Pyx_PySet_ContainsTF(((PyObject *)__pyx_v_self), __pyx_v_created, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(12, 252, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (!__pyx_v_self->__pyx_base.__pyx_base.options->canonical_form_repeat_fixed);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/string_types.pxi":253
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         if self in created and not self.options.canonical_form_repeat_fixed:
 *             return _CanonicalForm(f'"{self.type}"')             # <<<<<<<<<<<<<<
 *         created.add(self)
 *         return dict_to_canonical({
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u__48);
    __pyx_t_4 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__48);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u__48);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_6, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(12, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u__48);
    __pyx_t_4 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__48);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__48);
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(12, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "src/string_types.pxi":252
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         if self in created and not self.options.canonical_form_repeat_fixed:             # <<<<<<<<<<<<<<
 *             return _CanonicalForm(f'"{self.type}"')
 *         created.add(self)
 */
  }

  /* "src/string_types.pxi":254
 *         if self in created and not self.options.canonical_form_repeat_fixed:
 *             return _CanonicalForm(f'"{self.type}"')
 *         created.add(self)             # <<<<<<<<<<<<<<
 *         return dict_to_canonical({
 *             'type': 'fixed',
 */
  if (unlikely(__pyx_v_created == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
    __PYX_ERR(12, 254, __pyx_L1_error)
  }
  __pyx_t_8 = PySet_Add(__pyx_v_created, ((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(12, 254, __pyx_L1_error)

  /* "src/string_types.pxi":255
 *             return _CanonicalForm(f'"{self.type}"')
 *         created.add(self)
 *         return dict_to_canonical({             # <<<<<<<<<<<<<<
 *             'type': 'fixed',
 *             'name': self.type,
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);

  /* "src/string_types.pxi":256
 *         created.add(self)
 *         return dict_to_canonical({
 *             'type': 'fixed',             # <<<<<<<<<<<<<<
 *             'name': self.type,
 *             'size': self.size
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_type, __pyx_n_u_fixed) < 0) __PYX_ERR(12, 256, __pyx_L1_error)

  /* "src/string_types.pxi":257
 *         return dict_to_canonical({
 *             'type': 'fixed',
 *             'name': self.type,             # <<<<<<<<<<<<<<
 *             'size': self.size
 *         })
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(12, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_name, __pyx_t_7) < 0) __PYX_ERR(12, 256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/string_types.pxi":258
 *             'type': 'fixed',
 *             'name': self.type,
 *             'size': self.size             # <<<<<<<<<<<<<<
 *         })
 * 
 */
  __pyx_t_7 = PyInt_FromSsize_t(__pyx_v_self->size); if (unlikely(!__pyx_t_7)) __PYX_ERR(12, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_size, __pyx_t_7) < 0) __PYX_ERR(12, 256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/string_types.pxi":255
 *             return _CanonicalForm(f'"{self.type}"')
 *         created.add(self)
 *         return dict_to_canonical({             # <<<<<<<<<<<<<<
 *             'type': 'fixed',
 *             'name': self.type,
 */
  __pyx_t_7 = ((PyObject *)__pyx_f_5cavro_dict_to_canonical(__pyx_t_3)); if (unlikely(!__pyx_t_7)) __PYX_ERR(12, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_7);
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":251
 *         raise ValueError(f"Invalid length for fixed field: {length} != {self.size} (value: {value})")
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         if self in created and not self.options.canonical_form_repeat_fixed:
 *             return _CanonicalForm(f'"{self.type}"')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.FixedType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":261
 *         })
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         cdef FixedType writer_fixed
 *         if not isinstance(writer, FixedType):
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9FixedType__for_writer(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer) {
  struct __pyx_obj_5cavro_FixedType *__pyx_v_writer_fixed = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_writer", 1);

  /* "src/string_types.pxi":263
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         cdef FixedType writer_fixed
 *         if not isinstance(writer, FixedType):             # <<<<<<<<<<<<<<
 *             return
 *         writer_fixed = writer
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_FixedType); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "src/string_types.pxi":264
 *         cdef FixedType writer_fixed
 *         if not isinstance(writer, FixedType):
 *             return             # <<<<<<<<<<<<<<
 *         writer_fixed = writer
 *         if not self.name_matches(writer_fixed):
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/string_types.pxi":263
 *     cdef AvroType _for_writer(self, AvroType writer):
 *         cdef FixedType writer_fixed
 *         if not isinstance(writer, FixedType):             # <<<<<<<<<<<<<<
 *             return
 *         writer_fixed = writer
 */
  }

  /* "src/string_types.pxi":265
 *         if not isinstance(writer, FixedType):
 *             return
 *         writer_fixed = writer             # <<<<<<<<<<<<<<
 *         if not self.name_matches(writer_fixed):
 *             return
 */
  if (!(likely(((((PyObject *)__pyx_v_writer)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_FixedType))))) __PYX_ERR(12, 265, __pyx_L1_error)
  __pyx_t_3 = ((PyObject *)__pyx_v_writer);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_writer_fixed = ((struct __pyx_obj_5cavro_FixedType *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/string_types.pxi":266
 *             return
 *         writer_fixed = writer
 *         if not self.name_matches(writer_fixed):             # <<<<<<<<<<<<<<
 *             return
 *         if writer_fixed.size != self.size:
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_FixedType *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.name_matches(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self), ((struct __pyx_obj_5cavro__NamedType *)__pyx_v_writer_fixed)); if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 266, __pyx_L1_error)
  __pyx_t_1 = (!__pyx_t_2);
  if (__pyx_t_1) {

    /* "src/string_types.pxi":267
 *         writer_fixed = writer
 *         if not self.name_matches(writer_fixed):
 *             return             # <<<<<<<<<<<<<<
 *         if writer_fixed.size != self.size:
 *             return
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/string_types.pxi":266
 *             return
 *         writer_fixed = writer
 *         if not self.name_matches(writer_fixed):             # <<<<<<<<<<<<<<
 *             return
 *         if writer_fixed.size != self.size:
 */
  }

  /* "src/string_types.pxi":268
 *         if not self.name_matches(writer_fixed):
 *             return
 *         if writer_fixed.size != self.size:             # <<<<<<<<<<<<<<
 *             return
 *         return self
 */
  __pyx_t_1 = (__pyx_v_writer_fixed->size != __pyx_v_self->size);
  if (__pyx_t_1) {

    /* "src/string_types.pxi":269
 *             return
 *         if writer_fixed.size != self.size:
 *             return             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/string_types.pxi":268
 *         if not self.name_matches(writer_fixed):
 *             return
 *         if writer_fixed.size != self.size:             # <<<<<<<<<<<<<<
 *             return
 *         return self
 */
  }

  /* "src/string_types.pxi":270
 *         if writer_fixed.size != self.size:
 *             return
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self);
  goto __pyx_L0;

  /* "src/string_types.pxi":261
 *         })
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         cdef FixedType writer_fixed
 *         if not isinstance(writer, FixedType):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.FixedType._for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_writer_fixed);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":272
 *         return self
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):             # <<<<<<<<<<<<<<
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 */

static PyObject *__pyx_f_5cavro_9FixedType_resolve_default_value(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v_schema_default, PyObject *__pyx_v_field) {
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  Py_UCS4 __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resolve_default_value", 0);
  __Pyx_INCREF(__pyx_v_schema_default);

  /* "src/string_types.pxi":273
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if schema_default is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             return NO_DEFAULT
 *         if self.options.string_types_default_unchanged:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_v_schema_default == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/string_types.pxi":274
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT             # <<<<<<<<<<<<<<
 *         if self.options.string_types_default_unchanged:
 *             return schema_default
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/string_types.pxi":273
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if schema_default is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             return NO_DEFAULT
 *         if self.options.string_types_default_unchanged:
 */
  }

  /* "src/string_types.pxi":275
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 *         if self.options.string_types_default_unchanged:             # <<<<<<<<<<<<<<
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:
 */
  if (__pyx_v_self->__pyx_base.__pyx_base.options->string_types_default_unchanged) {

    /* "src/string_types.pxi":276
 *             return NO_DEFAULT
 *         if self.options.string_types_default_unchanged:
 *             return schema_default             # <<<<<<<<<<<<<<
 *         if self.options.bytes_default_value_utf8:
 *             try:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_schema_default);
    __pyx_r = __pyx_v_schema_default;
    goto __pyx_L0;

    /* "src/string_types.pxi":275
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 *         if self.options.string_types_default_unchanged:             # <<<<<<<<<<<<<<
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:
 */
  }

  /* "src/string_types.pxi":277
 *         if self.options.string_types_default_unchanged:
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:             # <<<<<<<<<<<<<<
 *             try:
 *                 schema_default = schema_default.encode('utf-8')
 */
  if (__pyx_v_self->__pyx_base.__pyx_base.options->bytes_default_value_utf8) {

    /* "src/string_types.pxi":278
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:
 *             try:             # <<<<<<<<<<<<<<
 *                 schema_default = schema_default.encode('utf-8')
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "src/string_types.pxi":279
 *         if self.options.bytes_default_value_utf8:
 *             try:
 *                 schema_default = schema_default.encode('utf-8')             # <<<<<<<<<<<<<<
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:
 *                 raise TypeError(f"Default value {schema_default!r} is not valid for fixed field: {field}") from e
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema_default, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(12, 279, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_kp_u_utf_8};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 279, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF_SET(__pyx_v_schema_default, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "src/string_types.pxi":278
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:
 *             try:             # <<<<<<<<<<<<<<
 *                 schema_default = schema_default.encode('utf-8')
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L11_try_end;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "src/string_types.pxi":280
 *             try:
 *                 schema_default = schema_default.encode('utf-8')
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:             # <<<<<<<<<<<<<<
 *                 raise TypeError(f"Default value {schema_default!r} is not valid for fixed field: {field}") from e
 *             return schema_default
 */
      __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeEncodeError);
      if (__pyx_t_8) {
        __Pyx_AddTraceback("cavro.FixedType.resolve_default_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(12, 280, __pyx_L8_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __pyx_v_e = __pyx_t_6;
        /*try:*/ {

          /* "src/string_types.pxi":281
 *                 schema_default = schema_default.encode('utf-8')
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:
 *                 raise TypeError(f"Default value {schema_default!r} is not valid for fixed field: {field}") from e             # <<<<<<<<<<<<<<
 *             return schema_default
 *         return AvroType.resolve_default_value(self, schema_default, field)
 */
          __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 281, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_10 = 0;
          __pyx_t_11 = 127;
          __Pyx_INCREF(__pyx_kp_u_Default_value_2);
          __pyx_t_10 += 14;
          __Pyx_GIVEREF(__pyx_kp_u_Default_value_2);
          PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_kp_u_Default_value_2);
          __pyx_t_12 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_schema_default), __pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(12, 281, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
          __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_12);
          __pyx_t_12 = 0;
          __Pyx_INCREF(__pyx_kp_u_is_not_valid_for_fixed_field);
          __pyx_t_10 += 31;
          __Pyx_GIVEREF(__pyx_kp_u_is_not_valid_for_fixed_field);
          PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_kp_u_is_not_valid_for_fixed_field);
          __pyx_t_12 = __Pyx_PyUnicode_Unicode(__pyx_v_field); if (unlikely(!__pyx_t_12)) __PYX_ERR(12, 281, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
          __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_12);
          __pyx_t_12 = 0;
          __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_9, 4, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(12, 281, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 281, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_Raise(__pyx_t_9, 0, 0, __pyx_v_e);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __PYX_ERR(12, 281, __pyx_L17_error)
        }

        /* "src/string_types.pxi":280
 *             try:
 *                 schema_default = schema_default.encode('utf-8')
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:             # <<<<<<<<<<<<<<
 *                 raise TypeError(f"Default value {schema_default!r} is not valid for fixed field: {field}") from e
 *             return schema_default
 */
        /*finally:*/ {
          __pyx_L17_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_8 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L8_except_error;
          }
        }
      }
      goto __pyx_L8_except_error;

      /* "src/string_types.pxi":278
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:
 *             try:             # <<<<<<<<<<<<<<
 *                 schema_default = schema_default.encode('utf-8')
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:
 */
      __pyx_L8_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L11_try_end:;
    }

    /* "src/string_types.pxi":282
 *             except (AttributeError, TypeError, UnicodeEncodeError) as e:
 *                 raise TypeError(f"Default value {schema_default!r} is not valid for fixed field: {field}") from e
 *             return schema_default             # <<<<<<<<<<<<<<
 *         return AvroType.resolve_default_value(self, schema_default, field)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_schema_default);
    __pyx_r = __pyx_v_schema_default;
    goto __pyx_L0;

    /* "src/string_types.pxi":277
 *         if self.options.string_types_default_unchanged:
 *             return schema_default
 *         if self.options.bytes_default_value_utf8:             # <<<<<<<<<<<<<<
 *             try:
 *                 schema_default = schema_default.encode('utf-8')
 */
  }

  /* "src/string_types.pxi":283
 *                 raise TypeError(f"Default value {schema_default!r} is not valid for fixed field: {field}") from e
 *             return schema_default
 *         return AvroType.resolve_default_value(self, schema_default, field)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = __pyx_f_5cavro_8AvroType_resolve_default_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_schema_default, __pyx_v_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(12, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "src/string_types.pxi":272
 *         return self
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):             # <<<<<<<<<<<<<<
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("cavro.FixedType.resolve_default_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_schema_default);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/string_types.pxi":155
 *     type_name = "fixed"
 * 
 *     cdef readonly Py_ssize_t size             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, schema, source, namespace):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9FixedType_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_9FixedType_4size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_9FixedType_4size___get__(((struct __pyx_obj_5cavro_FixedType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9FixedType_4size___get__(struct __pyx_obj_5cavro_FixedType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FixedType.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9FixedType_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9FixedType_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FixedType_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9FixedType_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_9FixedType_8__reduce_cython__(((struct __pyx_obj_5cavro_FixedType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9FixedType_8__reduce_cython__(struct __pyx_obj_5cavro_FixedType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.aliases, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.size, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(8); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.aliases);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.aliases);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->__pyx_base.aliases)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.effective_namespace);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.effective_namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->__pyx_base.effective_namespace)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->__pyx_base.__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.name);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_v_self->__pyx_base.name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.namespace);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_v_self->__pyx_base.namespace)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 5, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_v_self->__pyx_base.__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.aliases, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.size, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.aliases, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.size, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.aliases is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.aliases, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.size, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.aliases is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_FixedType, (type(self), 0xd9f3afe, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base.aliases != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.effective_namespace != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.name != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.namespace != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_3 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FixedType, (type(self), 0xd9f3afe, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.aliases is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_FixedType, (type(self), 0xd9f3afe, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_FixedType, (type(self), 0xd9f3afe, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_FixedType); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_228539134);
    __Pyx_GIVEREF(__pyx_int_228539134);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_228539134)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FixedType, (type(self), 0xd9f3afe, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_FixedType, (type(self), 0xd9f3afe, None), state
 *     else:
 *         return __pyx_unpickle_FixedType, (type(self), 0xd9f3afe, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FixedType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_FixedType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_228539134);
    __Pyx_GIVEREF(__pyx_int_228539134);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_228539134)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.FixedType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FixedType, (type(self), 0xd9f3afe, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FixedType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9FixedType_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9FixedType_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FixedType_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9FixedType_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.FixedType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9FixedType_10__setstate_cython__(((struct __pyx_obj_5cavro_FixedType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9FixedType_10__setstate_cython__(struct __pyx_obj_5cavro_FixedType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_FixedType, (type(self), 0xd9f3afe, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FixedType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_FixedType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FixedType, (type(self), 0xd9f3afe, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FixedType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.FixedType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":8
 * ]
 * 
 * cdef bint are_union_values_ambiguous(sub_types):             # <<<<<<<<<<<<<<
 *     cdef AvroType sub_type
 *     cdef Py_ssize_t i
 */

static int __pyx_f_5cavro_are_union_values_ambiguous(PyObject *__pyx_v_sub_types) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_sub_type = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_v_seen_classes = NULL;
  PyObject *__pyx_v_bucket = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("are_union_values_ambiguous", 1);

  /* "src/union.pxi":11
 *     cdef AvroType sub_type
 *     cdef Py_ssize_t i
 *     seen_classes = set()             # <<<<<<<<<<<<<<
 * 
 *     for sub_type in sub_types:
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_seen_classes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/union.pxi":13
 *     seen_classes = set()
 * 
 *     for sub_type in sub_types:             # <<<<<<<<<<<<<<
 *         i = 0
 *         for bucket in AMBIGUOUS_TYPES:
 */
  if (likely(PyList_CheckExact(__pyx_v_sub_types)) || PyTuple_CheckExact(__pyx_v_sub_types)) {
    __pyx_t_1 = __pyx_v_sub_types; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_sub_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 13, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 13, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 13, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 13, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 13, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 13, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 13, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(22, 13, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 13, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_sub_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "src/union.pxi":14
 * 
 *     for sub_type in sub_types:
 *         i = 0             # <<<<<<<<<<<<<<
 *         for bucket in AMBIGUOUS_TYPES:
 *             if type(sub_type) in bucket:
 */
    __pyx_v_i = 0;

    /* "src/union.pxi":15
 *     for sub_type in sub_types:
 *         i = 0
 *         for bucket in AMBIGUOUS_TYPES:             # <<<<<<<<<<<<<<
 *             if type(sub_type) in bucket:
 *                 if i in seen_classes:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_AMBIGUOUS_TYPES); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5);
      __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(22, 15, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 15, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 15, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(22, 15, __pyx_L1_error)
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 15, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 15, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(22, 15, __pyx_L1_error)
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 15, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_7(__pyx_t_5);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(22, 15, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_v_bucket, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "src/union.pxi":16
 *         i = 0
 *         for bucket in AMBIGUOUS_TYPES:
 *             if type(sub_type) in bucket:             # <<<<<<<<<<<<<<
 *                 if i in seen_classes:
 *                     return True
 */
      __pyx_t_8 = (__Pyx_PySequence_ContainsTF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_sub_type))), __pyx_v_bucket, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(22, 16, __pyx_L1_error)
      if (__pyx_t_8) {

        /* "src/union.pxi":17
 *         for bucket in AMBIGUOUS_TYPES:
 *             if type(sub_type) in bucket:
 *                 if i in seen_classes:             # <<<<<<<<<<<<<<
 *                     return True
 *             seen_classes.add(i)
 */
        __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 17, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = (__Pyx_PySet_ContainsTF(__pyx_t_4, __pyx_v_seen_classes, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(22, 17, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_8) {

          /* "src/union.pxi":18
 *             if type(sub_type) in bucket:
 *                 if i in seen_classes:
 *                     return True             # <<<<<<<<<<<<<<
 *             seen_classes.add(i)
 *             i += 1
 */
          __pyx_r = 1;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          goto __pyx_L0;

          /* "src/union.pxi":17
 *         for bucket in AMBIGUOUS_TYPES:
 *             if type(sub_type) in bucket:
 *                 if i in seen_classes:             # <<<<<<<<<<<<<<
 *                     return True
 *             seen_classes.add(i)
 */
        }

        /* "src/union.pxi":16
 *         i = 0
 *         for bucket in AMBIGUOUS_TYPES:
 *             if type(sub_type) in bucket:             # <<<<<<<<<<<<<<
 *                 if i in seen_classes:
 *                     return True
 */
      }

      /* "src/union.pxi":19
 *                 if i in seen_classes:
 *                     return True
 *             seen_classes.add(i)             # <<<<<<<<<<<<<<
 *             i += 1
 *     return False
 */
      __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 19, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = PySet_Add(__pyx_v_seen_classes, __pyx_t_4); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(22, 19, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "src/union.pxi":20
 *                     return True
 *             seen_classes.add(i)
 *             i += 1             # <<<<<<<<<<<<<<
 *     return False
 * 
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "src/union.pxi":15
 *     for sub_type in sub_types:
 *         i = 0
 *         for bucket in AMBIGUOUS_TYPES:             # <<<<<<<<<<<<<<
 *             if type(sub_type) in bucket:
 *                 if i in seen_classes:
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "src/union.pxi":13
 *     seen_classes = set()
 * 
 *     for sub_type in sub_types:             # <<<<<<<<<<<<<<
 *         i = 0
 *         for bucket in AMBIGUOUS_TYPES:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/union.pxi":21
 *             seen_classes.add(i)
 *             i += 1
 *     return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "src/union.pxi":8
 * ]
 * 
 * cdef bint are_union_values_ambiguous(sub_types):             # <<<<<<<<<<<<<<
 *     cdef AvroType sub_type
 *     cdef Py_ssize_t i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.are_union_values_ambiguous", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sub_type);
  __Pyx_XDECREF(__pyx_v_seen_classes);
  __Pyx_XDECREF(__pyx_v_bucket);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":33
 *     cdef readonly tuple return_type_tuple
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         super().__init__(schema, source, namespace)
 *         self.union_types = tuple(AvroType.for_source(schema, s, namespace) for s in source)
 */

/* Python wrapper */
static int __pyx_pw_5cavro_9UnionType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_9UnionType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_namespace = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_schema,&__pyx_n_s_source,&__pyx_n_s_namespace,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 33, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 33, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(22, 33, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_namespace)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 33, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(22, 33, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(22, 33, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_schema = values[0];
    __pyx_v_source = values[1];
    __pyx_v_namespace = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(22, 33, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UnionType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9UnionType___init__(((struct __pyx_obj_5cavro_UnionType *)__pyx_v_self), __pyx_v_schema, __pyx_v_source, __pyx_v_namespace);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_9UnionType_8__init___2generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/union.pxi":35
 *     def __init__(self, schema, source, namespace):
 *         super().__init__(schema, source, namespace)
 *         self.union_types = tuple(AvroType.for_source(schema, s, namespace) for s in source)             # <<<<<<<<<<<<<<
 *         if len(self.union_types) == 0 and not self.options.allow_empty_unions:
 *             raise ValueError("Unions must contain at least one member type")
 */

static PyObject *__pyx_pf_5cavro_9UnionType_8__init___genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_12_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_12_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(22, 35, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_9UnionType_8__init___2generator11, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_init___locals_genexpr, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(22, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.UnionType.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_9UnionType_8__init___2generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 35, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(22, 35, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 35, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 35, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 35, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 35, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 35, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 35, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 35, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(22, 35, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_AvroType), __pyx_n_s_for_source); if (unlikely(!__pyx_t_5)) __PYX_ERR(22, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_schema)) { __Pyx_RaiseClosureNameError("schema"); __PYX_ERR(22, 35, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_namespace)) { __Pyx_RaiseClosureNameError("namespace"); __PYX_ERR(22, 35, __pyx_L1_error) }
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_6, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_schema, __pyx_cur_scope->__pyx_v_s, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_namespace};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 3+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 35, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 35, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_9UnionType_8__init___5generator12(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/union.pxi":44
 *         decode_option = self.options.union_decodes_to
 *         if decode_option == UnionDecodeOption.RAW_VALUES:
 *             self.return_type_tuple = tuple(False for _ in self.union_types)             # <<<<<<<<<<<<<<
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_AMBIGUOUS:
 *             is_amibiguous = are_union_values_ambiguous(self.union_types)
 */

static PyObject *__pyx_pf_5cavro_9UnionType_8__init___3genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_13_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_13_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(22, 44, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_9UnionType_8__init___5generator12, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_init___locals_genexpr, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(22, 44, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.UnionType.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_9UnionType_8__init___5generator12(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 44, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(22, 44, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_genexpr_arg_0 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 44, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 44, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 44, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v__);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v__, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 44, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_9UnionType_8__init___8generator13(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/union.pxi":47
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_AMBIGUOUS:
 *             is_amibiguous = are_union_values_ambiguous(self.union_types)
 *             self.return_type_tuple = tuple(is_amibiguous for _ in self.union_types)             # <<<<<<<<<<<<<<
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD_AMBIGUOUS:
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1
 */

static PyObject *__pyx_pf_5cavro_9UnionType_8__init___6genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_14_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_14_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(22, 47, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_9UnionType_8__init___8generator13, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_init___locals_genexpr, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(22, 47, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.UnionType.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_9UnionType_8__init___8generator13(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 47, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(22, 47, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_genexpr_arg_0 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 47, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 47, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 47, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v__);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v__, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_is_amibiguous); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 47, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_9UnionType_8__init___11generator14(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/union.pxi":49
 *             self.return_type_tuple = tuple(is_amibiguous for _ in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD_AMBIGUOUS:
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1             # <<<<<<<<<<<<<<
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:
 */

static PyObject *__pyx_pf_5cavro_9UnionType_8__init___9genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_15_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_15_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(22, 49, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_9UnionType_8__init___11generator14, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_init___locals_genexpr, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(22, 49, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.UnionType.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_9UnionType_8__init___11generator14(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L9_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 49, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(22, 49, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_genexpr_arg_0 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 49, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 49, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 49, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_t);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_t, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_TypeCheck(__pyx_cur_scope->__pyx_v_t, __pyx_ptype_5cavro_RecordType); 
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_TypeCheck(__pyx_cur_scope->__pyx_v_t, __pyx_ptype_5cavro_MapType); 
    __pyx_t_4 = __pyx_t_5;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_4) {
      __Pyx_INCREF(__pyx_int_1);
      __pyx_r = __pyx_int_1;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L9_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 49, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_9UnionType_8__init___14generator15(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/union.pxi":50
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD_AMBIGUOUS:
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)             # <<<<<<<<<<<<<<
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)
 */

static PyObject *__pyx_pf_5cavro_9UnionType_8__init___12genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_16_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_16_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(22, 50, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_9UnionType_8__init___14generator15, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_init___locals_genexpr, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(22, 50, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.UnionType.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_9UnionType_8__init___14generator15(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L8_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 50, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(22, 50, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_genexpr_arg_0 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 50, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 50, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 50, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_ty);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_ty, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_is_ambiguous)) { __Pyx_RaiseClosureNameError("is_ambiguous"); __PYX_ERR(22, 50, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_is_ambiguous); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(22, 50, __pyx_L1_error)
    if (__pyx_t_4) {
    } else {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_is_ambiguous);
      __pyx_t_3 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_is_ambiguous;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_cur_scope->__pyx_v_ty, __pyx_ptype_5cavro_RecordType); 
    __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(22, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_L6_bool_binop_done:;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L8_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 50, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_9UnionType_8__init___17generator16(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/union.pxi":52
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)             # <<<<<<<<<<<<<<
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_ALWAYS:
 *             self.return_type_tuple = tuple(True for _ in self.union_types)
 */

static PyObject *__pyx_pf_5cavro_9UnionType_8__init___15genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_17_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_17_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(22, 52, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_9UnionType_8__init___17generator16, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_init___locals_genexpr, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(22, 52, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.UnionType.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_9UnionType_8__init___17generator16(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 52, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(22, 52, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_genexpr_arg_0 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 52, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 52, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 52, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_ty);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_ty, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_cur_scope->__pyx_v_ty, __pyx_ptype_5cavro_RecordType); 
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 52, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_9UnionType_8__init___20generator17(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/union.pxi":54
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_ALWAYS:
 *             self.return_type_tuple = tuple(True for _ in self.union_types)             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError(f"Invalid union_decodes_to option: {decode_option}")
 */

static PyObject *__pyx_pf_5cavro_9UnionType_8__init___18genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_18_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_18_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(22, 54, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_9UnionType_8__init___20generator17, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_init___locals_genexpr, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(22, 54, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.UnionType.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_9UnionType_8__init___20generator17(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 54, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(22, 54, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_genexpr_arg_0 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 54, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 54, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 54, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v__);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v__, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 54, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":33
 *     cdef readonly tuple return_type_tuple
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         super().__init__(schema, source, namespace)
 *         self.union_types = tuple(AvroType.for_source(schema, s, namespace) for s in source)
 */

static int __pyx_pf_5cavro_9UnionType___init__(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace) {
  struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *__pyx_cur_scope;
  PyObject *__pyx_v_decode_option = NULL;
  PyObject *__pyx_gb_5cavro_9UnionType_8__init___2generator11 = 0;
  PyObject *__pyx_gb_5cavro_9UnionType_8__init___5generator12 = 0;
  PyObject *__pyx_gb_5cavro_9UnionType_8__init___8generator13 = 0;
  PyObject *__pyx_gb_5cavro_9UnionType_8__init___11generator14 = 0;
  PyObject *__pyx_gb_5cavro_9UnionType_8__init___14generator15 = 0;
  PyObject *__pyx_gb_5cavro_9UnionType_8__init___17generator16 = 0;
  PyObject *__pyx_gb_5cavro_9UnionType_8__init___20generator17 = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *)__pyx_tp_new_5cavro___pyx_scope_struct_11___init__(__pyx_ptype_5cavro___pyx_scope_struct_11___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(22, 33, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_schema = __pyx_v_schema;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_schema);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_schema);
  __pyx_cur_scope->__pyx_v_namespace = __pyx_v_namespace;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_namespace);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_namespace);

  /* "src/union.pxi":34
 * 
 *     def __init__(self, schema, source, namespace):
 *         super().__init__(schema, source, namespace)             # <<<<<<<<<<<<<<
 *         self.union_types = tuple(AvroType.for_source(schema, s, namespace) for s in source)
 *         if len(self.union_types) == 0 and not self.options.allow_empty_unions:
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_UnionType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_UnionType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_5cavro_UnionType))) __PYX_ERR(22, 34, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(22, 34, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_cur_scope->__pyx_v_schema, __pyx_v_source, __pyx_cur_scope->__pyx_v_namespace};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/union.pxi":35
 *     def __init__(self, schema, source, namespace):
 *         super().__init__(schema, source, namespace)
 *         self.union_types = tuple(AvroType.for_source(schema, s, namespace) for s in source)             # <<<<<<<<<<<<<<
 *         if len(self.union_types) == 0 and not self.options.allow_empty_unions:
 *             raise ValueError("Unions must contain at least one member type")
 */
  __pyx_t_1 = __pyx_pf_5cavro_9UnionType_8__init___genexpr(((PyObject*)__pyx_cur_scope), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->union_types);
  __Pyx_DECREF(__pyx_v_self->union_types);
  __pyx_v_self->union_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/union.pxi":36
 *         super().__init__(schema, source, namespace)
 *         self.union_types = tuple(AvroType.for_source(schema, s, namespace) for s in source)
 *         if len(self.union_types) == 0 and not self.options.allow_empty_unions:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unions must contain at least one member type")
 * 
 */
  __pyx_t_2 = __pyx_v_self->union_types;
  __Pyx_INCREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(22, 36, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_t_2); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(22, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = (__pyx_t_6 == 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_5 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = (!__pyx_v_self->__pyx_base.options->allow_empty_unions);
  __pyx_t_5 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_5)) {

    /* "src/union.pxi":37
 *         self.union_types = tuple(AvroType.for_source(schema, s, namespace) for s in source)
 *         if len(self.union_types) == 0 and not self.options.allow_empty_unions:
 *             raise ValueError("Unions must contain at least one member type")             # <<<<<<<<<<<<<<
 * 
 *         if not (self.options.allow_duplicate_union_types and self.options.allow_nested_unions):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__73, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(22, 37, __pyx_L1_error)

    /* "src/union.pxi":36
 *         super().__init__(schema, source, namespace)
 *         self.union_types = tuple(AvroType.for_source(schema, s, namespace) for s in source)
 *         if len(self.union_types) == 0 and not self.options.allow_empty_unions:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unions must contain at least one member type")
 * 
 */
  }

  /* "src/union.pxi":39
 *             raise ValueError("Unions must contain at least one member type")
 * 
 *         if not (self.options.allow_duplicate_union_types and self.options.allow_nested_unions):             # <<<<<<<<<<<<<<
 *             self._populate_name_map()
 * 
 */
  if (__pyx_v_self->__pyx_base.options->allow_duplicate_union_types) {
  } else {
    __pyx_t_5 = __pyx_v_self->__pyx_base.options->allow_duplicate_union_types;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_5 = __pyx_v_self->__pyx_base.options->allow_nested_unions;
  __pyx_L7_bool_binop_done:;
  __pyx_t_7 = (!__pyx_t_5);
  if (__pyx_t_7) {

    /* "src/union.pxi":40
 * 
 *         if not (self.options.allow_duplicate_union_types and self.options.allow_nested_unions):
 *             self._populate_name_map()             # <<<<<<<<<<<<<<
 * 
 *         decode_option = self.options.union_decodes_to
 */
    __pyx_t_2 = __pyx_f_5cavro_9UnionType__populate_name_map(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/union.pxi":39
 *             raise ValueError("Unions must contain at least one member type")
 * 
 *         if not (self.options.allow_duplicate_union_types and self.options.allow_nested_unions):             # <<<<<<<<<<<<<<
 *             self._populate_name_map()
 * 
 */
  }

  /* "src/union.pxi":42
 *             self._populate_name_map()
 * 
 *         decode_option = self.options.union_decodes_to             # <<<<<<<<<<<<<<
 *         if decode_option == UnionDecodeOption.RAW_VALUES:
 *             self.return_type_tuple = tuple(False for _ in self.union_types)
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.options->union_decodes_to;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_decode_option = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "src/union.pxi":43
 * 
 *         decode_option = self.options.union_decodes_to
 *         if decode_option == UnionDecodeOption.RAW_VALUES:             # <<<<<<<<<<<<<<
 *             self.return_type_tuple = tuple(False for _ in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_AMBIGUOUS:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UnionDecodeOption); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_RAW_VALUES); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_decode_option, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(22, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_7) {

    /* "src/union.pxi":44
 *         decode_option = self.options.union_decodes_to
 *         if decode_option == UnionDecodeOption.RAW_VALUES:
 *             self.return_type_tuple = tuple(False for _ in self.union_types)             # <<<<<<<<<<<<<<
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_AMBIGUOUS:
 *             is_amibiguous = are_union_values_ambiguous(self.union_types)
 */
    __pyx_t_2 = __pyx_pf_5cavro_9UnionType_8__init___3genexpr(NULL, __pyx_v_self->union_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->return_type_tuple);
    __Pyx_DECREF(__pyx_v_self->return_type_tuple);
    __pyx_v_self->return_type_tuple = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/union.pxi":43
 * 
 *         decode_option = self.options.union_decodes_to
 *         if decode_option == UnionDecodeOption.RAW_VALUES:             # <<<<<<<<<<<<<<
 *             self.return_type_tuple = tuple(False for _ in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_AMBIGUOUS:
 */
    goto __pyx_L9;
  }

  /* "src/union.pxi":45
 *         if decode_option == UnionDecodeOption.RAW_VALUES:
 *             self.return_type_tuple = tuple(False for _ in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_AMBIGUOUS:             # <<<<<<<<<<<<<<
 *             is_amibiguous = are_union_values_ambiguous(self.union_types)
 *             self.return_type_tuple = tuple(is_amibiguous for _ in self.union_types)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_UnionDecodeOption); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_TYPE_TUPLE_IF_AMBIGUOUS); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_decode_option, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(22, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "src/union.pxi":46
 *             self.return_type_tuple = tuple(False for _ in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_AMBIGUOUS:
 *             is_amibiguous = are_union_values_ambiguous(self.union_types)             # <<<<<<<<<<<<<<
 *             self.return_type_tuple = tuple(is_amibiguous for _ in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD_AMBIGUOUS:
 */
    __pyx_t_1 = __pyx_v_self->union_types;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7 = __pyx_f_5cavro_are_union_values_ambiguous(__pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(22, 46, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_cur_scope->__pyx_v_is_amibiguous = __pyx_t_7;

    /* "src/union.pxi":47
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_AMBIGUOUS:
 *             is_amibiguous = are_union_values_ambiguous(self.union_types)
 *             self.return_type_tuple = tuple(is_amibiguous for _ in self.union_types)             # <<<<<<<<<<<<<<
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD_AMBIGUOUS:
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1
 */
    __pyx_t_1 = __pyx_pf_5cavro_9UnionType_8__init___6genexpr(((PyObject*)__pyx_cur_scope), __pyx_v_self->union_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->return_type_tuple);
    __Pyx_DECREF(__pyx_v_self->return_type_tuple);
    __pyx_v_self->return_type_tuple = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/union.pxi":45
 *         if decode_option == UnionDecodeOption.RAW_VALUES:
 *             self.return_type_tuple = tuple(False for _ in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_AMBIGUOUS:             # <<<<<<<<<<<<<<
 *             is_amibiguous = are_union_values_ambiguous(self.union_types)
 *             self.return_type_tuple = tuple(is_amibiguous for _ in self.union_types)
 */
    goto __pyx_L9;
  }

  /* "src/union.pxi":48
 *             is_amibiguous = are_union_values_ambiguous(self.union_types)
 *             self.return_type_tuple = tuple(is_amibiguous for _ in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD_AMBIGUOUS:             # <<<<<<<<<<<<<<
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UnionDecodeOption); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_TYPE_TUPLE_IF_RECORD_AMBIGUOUS); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_decode_option, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(22, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_7) {

    /* "src/union.pxi":49
 *             self.return_type_tuple = tuple(is_amibiguous for _ in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD_AMBIGUOUS:
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1             # <<<<<<<<<<<<<<
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:
 */
    __pyx_t_2 = __pyx_pf_5cavro_9UnionType_8__init___9genexpr(NULL, __pyx_v_self->union_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_int_1, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 49, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_v_is_ambiguous = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "src/union.pxi":50
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD_AMBIGUOUS:
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)             # <<<<<<<<<<<<<<
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)
 */
    __pyx_t_2 = __pyx_pf_5cavro_9UnionType_8__init___12genexpr(((PyObject*)__pyx_cur_scope), __pyx_v_self->union_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->return_type_tuple);
    __Pyx_DECREF(__pyx_v_self->return_type_tuple);
    __pyx_v_self->return_type_tuple = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/union.pxi":48
 *             is_amibiguous = are_union_values_ambiguous(self.union_types)
 *             self.return_type_tuple = tuple(is_amibiguous for _ in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD_AMBIGUOUS:             # <<<<<<<<<<<<<<
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)
 */
    goto __pyx_L9;
  }

  /* "src/union.pxi":51
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:             # <<<<<<<<<<<<<<
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_ALWAYS:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_UnionDecodeOption); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_TYPE_TUPLE_IF_RECORD); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_decode_option, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(22, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "src/union.pxi":52
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)             # <<<<<<<<<<<<<<
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_ALWAYS:
 *             self.return_type_tuple = tuple(True for _ in self.union_types)
 */
    __pyx_t_1 = __pyx_pf_5cavro_9UnionType_8__init___15genexpr(NULL, __pyx_v_self->union_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->return_type_tuple);
    __Pyx_DECREF(__pyx_v_self->return_type_tuple);
    __pyx_v_self->return_type_tuple = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/union.pxi":51
 *             is_ambiguous = sum(1 for t in self.union_types if isinstance(t, (RecordType, MapType))) > 1
 *             self.return_type_tuple = tuple(is_ambiguous and isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:             # <<<<<<<<<<<<<<
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_ALWAYS:
 */
    goto __pyx_L9;
  }

  /* "src/union.pxi":53
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_ALWAYS:             # <<<<<<<<<<<<<<
 *             self.return_type_tuple = tuple(True for _ in self.union_types)
 *         else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UnionDecodeOption); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_TYPE_TUPLE_ALWAYS); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_decode_option, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(22, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(__pyx_t_7)) {

    /* "src/union.pxi":54
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_ALWAYS:
 *             self.return_type_tuple = tuple(True for _ in self.union_types)             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError(f"Invalid union_decodes_to option: {decode_option}")
 */
    __pyx_t_2 = __pyx_pf_5cavro_9UnionType_8__init___18genexpr(NULL, __pyx_v_self->union_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->return_type_tuple);
    __Pyx_DECREF(__pyx_v_self->return_type_tuple);
    __pyx_v_self->return_type_tuple = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/union.pxi":53
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_IF_RECORD:
 *             self.return_type_tuple = tuple(isinstance(ty, RecordType) for ty in self.union_types)
 *         elif decode_option == UnionDecodeOption.TYPE_TUPLE_ALWAYS:             # <<<<<<<<<<<<<<
 *             self.return_type_tuple = tuple(True for _ in self.union_types)
 *         else:
 */
    goto __pyx_L9;
  }

  /* "src/union.pxi":56
 *             self.return_type_tuple = tuple(True for _ in self.union_types)
 *         else:
 *             raise ValueError(f"Invalid union_decodes_to option: {decode_option}")             # <<<<<<<<<<<<<<
 * 
 *     cdef _populate_name_map(self):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_decode_option, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Invalid_union_decodes_to_option, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(22, 56, __pyx_L1_error)
  }
  __pyx_L9:;

  /* "src/union.pxi":33
 *     cdef readonly tuple return_type_tuple
 * 
 *     def __init__(self, schema, source, namespace):             # <<<<<<<<<<<<<<
 *         super().__init__(schema, source, namespace)
 *         self.union_types = tuple(AvroType.for_source(schema, s, namespace) for s in source)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.UnionType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_decode_option);
  __Pyx_XDECREF(__pyx_gb_5cavro_9UnionType_8__init___2generator11);
  __Pyx_XDECREF(__pyx_gb_5cavro_9UnionType_8__init___5generator12);
  __Pyx_XDECREF(__pyx_gb_5cavro_9UnionType_8__init___8generator13);
  __Pyx_XDECREF(__pyx_gb_5cavro_9UnionType_8__init___11generator14);
  __Pyx_XDECREF(__pyx_gb_5cavro_9UnionType_8__init___14generator15);
  __Pyx_XDECREF(__pyx_gb_5cavro_9UnionType_8__init___17generator16);
  __Pyx_XDECREF(__pyx_gb_5cavro_9UnionType_8__init___20generator17);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":58
 *             raise ValueError(f"Invalid union_decodes_to option: {decode_option}")
 * 
 *     cdef _populate_name_map(self):             # <<<<<<<<<<<<<<
 *         seen_types = set()
 *         self.by_name_map = {}
 */

static PyObject *__pyx_f_5cavro_9UnionType__populate_name_map(struct __pyx_obj_5cavro_UnionType *__pyx_v_self) {
  PyObject *__pyx_v_seen_types = NULL;
  PyObject *__pyx_v_member = NULL;
  PyTypeObject *__pyx_v_member_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_populate_name_map", 1);

  /* "src/union.pxi":59
 * 
 *     cdef _populate_name_map(self):
 *         seen_types = set()             # <<<<<<<<<<<<<<
 *         self.by_name_map = {}
 *         for member in self.union_types:
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_seen_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/union.pxi":60
 *     cdef _populate_name_map(self):
 *         seen_types = set()
 *         self.by_name_map = {}             # <<<<<<<<<<<<<<
 *         for member in self.union_types:
 *             if isinstance(member, _NamedType):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->by_name_map);
  __Pyx_DECREF(__pyx_v_self->by_name_map);
  __pyx_v_self->by_name_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/union.pxi":61
 *         seen_types = set()
 *         self.by_name_map = {}
 *         for member in self.union_types:             # <<<<<<<<<<<<<<
 *             if isinstance(member, _NamedType):
 *                 self.by_name_map[member.name] = member
 */
  if (unlikely(__pyx_v_self->union_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 61, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->union_types; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 61, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 61, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_member, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/union.pxi":62
 *         self.by_name_map = {}
 *         for member in self.union_types:
 *             if isinstance(member, _NamedType):             # <<<<<<<<<<<<<<
 *                 self.by_name_map[member.name] = member
 *                 self.by_name_map[member.type] = member
 */
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_member, __pyx_ptype_5cavro__NamedType); 
    if (__pyx_t_4) {

      /* "src/union.pxi":63
 *         for member in self.union_types:
 *             if isinstance(member, _NamedType):
 *                 self.by_name_map[member.name] = member             # <<<<<<<<<<<<<<
 *                 self.by_name_map[member.type] = member
 *             else:
 */
      if (unlikely(__pyx_v_self->by_name_map == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(22, 63, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_member, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 63, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely((PyDict_SetItem(__pyx_v_self->by_name_map, __pyx_t_3, __pyx_v_member) < 0))) __PYX_ERR(22, 63, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "src/union.pxi":64
 *             if isinstance(member, _NamedType):
 *                 self.by_name_map[member.name] = member
 *                 self.by_name_map[member.type] = member             # <<<<<<<<<<<<<<
 *             else:
 *                 self.by_name_map[member.type_name] = member
 */
      if (unlikely(__pyx_v_self->by_name_map == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(22, 64, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_member, __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely((PyDict_SetItem(__pyx_v_self->by_name_map, __pyx_t_3, __pyx_v_member) < 0))) __PYX_ERR(22, 64, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "src/union.pxi":62
 *         self.by_name_map = {}
 *         for member in self.union_types:
 *             if isinstance(member, _NamedType):             # <<<<<<<<<<<<<<
 *                 self.by_name_map[member.name] = member
 *                 self.by_name_map[member.type] = member
 */
      goto __pyx_L5;
    }

    /* "src/union.pxi":66
 *                 self.by_name_map[member.type] = member
 *             else:
 *                 self.by_name_map[member.type_name] = member             # <<<<<<<<<<<<<<
 * 
 *             if not self.options.allow_nested_unions and isinstance(member, UnionType):
 */
    /*else*/ {
      if (unlikely(__pyx_v_self->by_name_map == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(22, 66, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_member, __pyx_n_s_type_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 66, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely((PyDict_SetItem(__pyx_v_self->by_name_map, __pyx_t_3, __pyx_v_member) < 0))) __PYX_ERR(22, 66, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L5:;

    /* "src/union.pxi":68
 *                 self.by_name_map[member.type_name] = member
 * 
 *             if not self.options.allow_nested_unions and isinstance(member, UnionType):             # <<<<<<<<<<<<<<
 *                 raise ValueError("Unions may not immediately contain other unions")
 *             if not self.options.allow_duplicate_union_types:
 */
    __pyx_t_5 = (!__pyx_v_self->__pyx_base.options->allow_nested_unions);
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_member, __pyx_ptype_5cavro_UnionType); 
    __pyx_t_4 = __pyx_t_5;
    __pyx_L7_bool_binop_done:;
    if (unlikely(__pyx_t_4)) {

      /* "src/union.pxi":69
 * 
 *             if not self.options.allow_nested_unions and isinstance(member, UnionType):
 *                 raise ValueError("Unions may not immediately contain other unions")             # <<<<<<<<<<<<<<
 *             if not self.options.allow_duplicate_union_types:
 *                 if not isinstance(member, (RecordType, FixedType, EnumType)):
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__74, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 69, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(22, 69, __pyx_L1_error)

      /* "src/union.pxi":68
 *                 self.by_name_map[member.type_name] = member
 * 
 *             if not self.options.allow_nested_unions and isinstance(member, UnionType):             # <<<<<<<<<<<<<<
 *                 raise ValueError("Unions may not immediately contain other unions")
 *             if not self.options.allow_duplicate_union_types:
 */
    }

    /* "src/union.pxi":70
 *             if not self.options.allow_nested_unions and isinstance(member, UnionType):
 *                 raise ValueError("Unions may not immediately contain other unions")
 *             if not self.options.allow_duplicate_union_types:             # <<<<<<<<<<<<<<
 *                 if not isinstance(member, (RecordType, FixedType, EnumType)):
 *                     member_type = type(member)
 */
    __pyx_t_4 = (!__pyx_v_self->__pyx_base.options->allow_duplicate_union_types);
    if (__pyx_t_4) {

      /* "src/union.pxi":71
 *                 raise ValueError("Unions may not immediately contain other unions")
 *             if not self.options.allow_duplicate_union_types:
 *                 if not isinstance(member, (RecordType, FixedType, EnumType)):             # <<<<<<<<<<<<<<
 *                     member_type = type(member)
 *                     if member_type in seen_types:
 */
      __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_member, __pyx_ptype_5cavro_RecordType); 
      if (!__pyx_t_5) {
      } else {
        __pyx_t_4 = __pyx_t_5;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_member, __pyx_ptype_5cavro_FixedType); 
      if (!__pyx_t_5) {
      } else {
        __pyx_t_4 = __pyx_t_5;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_member, __pyx_ptype_5cavro_EnumType); 
      __pyx_t_4 = __pyx_t_5;
      __pyx_L11_bool_binop_done:;
      __pyx_t_5 = (!__pyx_t_4);
      if (__pyx_t_5) {

        /* "src/union.pxi":72
 *             if not self.options.allow_duplicate_union_types:
 *                 if not isinstance(member, (RecordType, FixedType, EnumType)):
 *                     member_type = type(member)             # <<<<<<<<<<<<<<
 *                     if member_type in seen_types:
 *                         raise ValueError(f"Unions may not have more than one member of type '{ member_type.type_name }'")
 */
        __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_member)));
        __Pyx_XDECREF_SET(__pyx_v_member_type, ((PyTypeObject*)((PyObject *)Py_TYPE(__pyx_v_member))));

        /* "src/union.pxi":73
 *                 if not isinstance(member, (RecordType, FixedType, EnumType)):
 *                     member_type = type(member)
 *                     if member_type in seen_types:             # <<<<<<<<<<<<<<
 *                         raise ValueError(f"Unions may not have more than one member of type '{ member_type.type_name }'")
 *                     seen_types.add(member_type)
 */
        __pyx_t_5 = (__Pyx_PySet_ContainsTF(((PyObject *)__pyx_v_member_type), __pyx_v_seen_types, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(22, 73, __pyx_L1_error)
        if (unlikely(__pyx_t_5)) {

          /* "src/union.pxi":74
 *                     member_type = type(member)
 *                     if member_type in seen_types:
 *                         raise ValueError(f"Unions may not have more than one member of type '{ member_type.type_name }'")             # <<<<<<<<<<<<<<
 *                     seen_types.add(member_type)
 * 
 */
          __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 74, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = 0;
          __pyx_t_7 = 127;
          __Pyx_INCREF(__pyx_kp_u_Unions_may_not_have_more_than_on);
          __pyx_t_6 += 50;
          __Pyx_GIVEREF(__pyx_kp_u_Unions_may_not_have_more_than_on);
          PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Unions_may_not_have_more_than_on);
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_member_type), __pyx_n_s_type_name); if (unlikely(!__pyx_t_8)) __PYX_ERR(22, 74, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(22, 74, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_7;
          __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_9);
          __pyx_t_9 = 0;
          __Pyx_INCREF(__pyx_kp_u__6);
          __pyx_t_6 += 1;
          __Pyx_GIVEREF(__pyx_kp_u__6);
          PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__6);
          __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(22, 74, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 74, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(22, 74, __pyx_L1_error)

          /* "src/union.pxi":73
 *                 if not isinstance(member, (RecordType, FixedType, EnumType)):
 *                     member_type = type(member)
 *                     if member_type in seen_types:             # <<<<<<<<<<<<<<
 *                         raise ValueError(f"Unions may not have more than one member of type '{ member_type.type_name }'")
 *                     seen_types.add(member_type)
 */
        }

        /* "src/union.pxi":75
 *                     if member_type in seen_types:
 *                         raise ValueError(f"Unions may not have more than one member of type '{ member_type.type_name }'")
 *                     seen_types.add(member_type)             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
        __pyx_t_10 = PySet_Add(__pyx_v_seen_types, ((PyObject *)__pyx_v_member_type)); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(22, 75, __pyx_L1_error)

        /* "src/union.pxi":71
 *                 raise ValueError("Unions may not immediately contain other unions")
 *             if not self.options.allow_duplicate_union_types:
 *                 if not isinstance(member, (RecordType, FixedType, EnumType)):             # <<<<<<<<<<<<<<
 *                     member_type = type(member)
 *                     if member_type in seen_types:
 */
      }

      /* "src/union.pxi":70
 *             if not self.options.allow_nested_unions and isinstance(member, UnionType):
 *                 raise ValueError("Unions may not immediately contain other unions")
 *             if not self.options.allow_duplicate_union_types:             # <<<<<<<<<<<<<<
 *                 if not isinstance(member, (RecordType, FixedType, EnumType)):
 *                     member_type = type(member)
 */
    }

    /* "src/union.pxi":61
 *         seen_types = set()
 *         self.by_name_map = {}
 *         for member in self.union_types:             # <<<<<<<<<<<<<<
 *             if isinstance(member, _NamedType):
 *                 self.by_name_map[member.name] = member
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/union.pxi":58
 *             raise ValueError(f"Invalid union_decodes_to option: {decode_option}")
 * 
 *     cdef _populate_name_map(self):             # <<<<<<<<<<<<<<
 *         seen_types = set()
 *         self.by_name_map = {}
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cavro.UnionType._populate_name_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_seen_types);
  __Pyx_XDECREF(__pyx_v_member);
  __Pyx_XDECREF((PyObject *)__pyx_v_member_type);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":77
 *                     seen_types.add(member_type)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef UnionType new_inst = self.clone_base()
 *         new_inst.union_types = self.union_types
 */

static PyObject *__pyx_pw_5cavro_9UnionType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9UnionType_copy(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_UnionType *__pyx_v_new_inst = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "src/union.pxi":78
 * 
 *     cpdef AvroType copy(self):
 *         cdef UnionType new_inst = self.clone_base()             # <<<<<<<<<<<<<<
 *         new_inst.union_types = self.union_types
 *         new_inst.by_name_map = self.by_name_map.copy()
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_UnionType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_UnionType))))) __PYX_ERR(22, 78, __pyx_L1_error)
  __pyx_v_new_inst = ((struct __pyx_obj_5cavro_UnionType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/union.pxi":79
 *     cpdef AvroType copy(self):
 *         cdef UnionType new_inst = self.clone_base()
 *         new_inst.union_types = self.union_types             # <<<<<<<<<<<<<<
 *         new_inst.by_name_map = self.by_name_map.copy()
 *         new_inst.return_type_tuple = self.return_type_tuple
 */
  __pyx_t_1 = __pyx_v_self->union_types;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_new_inst->union_types);
  __Pyx_DECREF(__pyx_v_new_inst->union_types);
  __pyx_v_new_inst->union_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/union.pxi":80
 *         cdef UnionType new_inst = self.clone_base()
 *         new_inst.union_types = self.union_types
 *         new_inst.by_name_map = self.by_name_map.copy()             # <<<<<<<<<<<<<<
 *         new_inst.return_type_tuple = self.return_type_tuple
 *         return new_inst
 */
  if (unlikely(__pyx_v_self->by_name_map == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "copy");
    __PYX_ERR(22, 80, __pyx_L1_error)
  }
  __pyx_t_1 = PyDict_Copy(__pyx_v_self->by_name_map); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_new_inst->by_name_map);
  __Pyx_DECREF(__pyx_v_new_inst->by_name_map);
  __pyx_v_new_inst->by_name_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/union.pxi":81
 *         new_inst.union_types = self.union_types
 *         new_inst.by_name_map = self.by_name_map.copy()
 *         new_inst.return_type_tuple = self.return_type_tuple             # <<<<<<<<<<<<<<
 *         return new_inst
 * 
 */
  __pyx_t_1 = __pyx_v_self->return_type_tuple;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_new_inst->return_type_tuple);
  __Pyx_DECREF(__pyx_v_new_inst->return_type_tuple);
  __pyx_v_new_inst->return_type_tuple = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/union.pxi":82
 *         new_inst.by_name_map = self.by_name_map.copy()
 *         new_inst.return_type_tuple = self.return_type_tuple
 *         return new_inst             # <<<<<<<<<<<<<<
 * 
 *     cdef _make_logical(self, schema, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_new_inst);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_new_inst);
  goto __pyx_L0;

  /* "src/union.pxi":77
 *                     seen_types.add(member_type)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef UnionType new_inst = self.clone_base()
 *         new_inst.union_types = self.union_types
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.UnionType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_new_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9UnionType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9UnionType_3copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9UnionType_3copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9UnionType_3copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_9UnionType_2copy(((struct __pyx_obj_5cavro_UnionType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9UnionType_2copy(struct __pyx_obj_5cavro_UnionType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_9UnionType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.UnionType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":84
 *         return new_inst
 * 
 *     cdef _make_logical(self, schema, source):             # <<<<<<<<<<<<<<
 *         return
 * 
 */

static PyObject *__pyx_f_5cavro_9UnionType__make_logical(CYTHON_UNUSED struct __pyx_obj_5cavro_UnionType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_schema, CYTHON_UNUSED PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_make_logical", 1);

  /* "src/union.pxi":85
 * 
 *     cdef _make_logical(self, schema, source):
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "src/union.pxi":84
 *         return new_inst
 * 
 *     cdef _make_logical(self, schema, source):             # <<<<<<<<<<<<<<
 *         return
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":87
 *         return
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return dict()
 * 
 */

static PyObject *__pyx_f_5cavro_9UnionType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro_UnionType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/union.pxi":88
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return dict()             # <<<<<<<<<<<<<<
 * 
 *     def walk_types(self, visited):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/union.pxi":87
 *         return
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return dict()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.UnionType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_9UnionType_6generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/union.pxi":90
 *         return dict()
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9UnionType_5walk_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9UnionType_5walk_types = {"walk_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9UnionType_5walk_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9UnionType_5walk_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_visited = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("walk_types (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_visited,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_visited)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 90, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "walk_types") < 0)) __PYX_ERR(22, 90, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_visited = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("walk_types", 1, 1, 1, __pyx_nargs); __PYX_ERR(22, 90, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UnionType.walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9UnionType_4walk_types(((struct __pyx_obj_5cavro_UnionType *)__pyx_v_self), __pyx_v_visited);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9UnionType_4walk_types(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_visited) {
  struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("walk_types", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types *)__pyx_tp_new_5cavro___pyx_scope_struct_19_walk_types(__pyx_ptype_5cavro___pyx_scope_struct_19_walk_types, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(22, 90, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_visited = __pyx_v_visited;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_visited);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_visited);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_9UnionType_6generator3, __pyx_codeobj__75, (PyObject *) __pyx_cur_scope, __pyx_n_s_walk_types, __pyx_n_s_UnionType_walk_types, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(22, 90, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.UnionType.walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_9UnionType_6generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("walk_types", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L5_resume_from_yield_from;
    case 2: goto __pyx_L8_resume_from_yield_from;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 90, __pyx_L1_error)

  /* "src/union.pxi":91
 * 
 *     def walk_types(self, visited):
 *         if self in visited:             # <<<<<<<<<<<<<<
 *             return
 *         yield from super().walk_types(visited)
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_visited, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(22, 91, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/union.pxi":92
 *     def walk_types(self, visited):
 *         if self in visited:
 *             return             # <<<<<<<<<<<<<<
 *         yield from super().walk_types(visited)
 *         for t in self.union_types:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "src/union.pxi":91
 * 
 *     def walk_types(self, visited):
 *         if self in visited:             # <<<<<<<<<<<<<<
 *             return
 *         yield from super().walk_types(visited)
 */
  }

  /* "src/union.pxi":93
 *         if self in visited:
 *             return
 *         yield from super().walk_types(visited)             # <<<<<<<<<<<<<<
 *         for t in self.union_types:
 *             yield from t.walk_types(visited)
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_UnionType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_UnionType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_ptype_5cavro_UnionType))) __PYX_ERR(22, 93, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_cur_scope->__pyx_v_self))) __PYX_ERR(22, 93, __pyx_L1_error);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_walk_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_visited};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __Pyx_Generator_Yield_From(__pyx_generator, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L5_resume_from_yield_from:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 93, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(22, 93, __pyx_L1_error)
    }
  }

  /* "src/union.pxi":94
 *             return
 *         yield from super().walk_types(visited)
 *         for t in self.union_types:             # <<<<<<<<<<<<<<
 *             yield from t.walk_types(visited)
 * 
 */
  if (unlikely(__pyx_cur_scope->__pyx_v_self->union_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 94, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_self->union_types; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_6 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 94, __pyx_L1_error)
      #endif
      if (__pyx_t_6 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(22, 94, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_t);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_t, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/union.pxi":95
 *         yield from super().walk_types(visited)
 *         for t in self.union_types:
 *             yield from t.walk_types(visited)             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_schema(self, created=None):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_t, __pyx_n_s_walk_types); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_cur_scope->__pyx_v_visited};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_r = __Pyx_Generator_Yield_From(__pyx_generator, __pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XGOTREF(__pyx_r);
    if (likely(__pyx_r)) {
      __Pyx_XGIVEREF(__pyx_t_2);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_6;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 2;
      return __pyx_r;
      __pyx_L8_resume_from_yield_from:;
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_2);
      __pyx_t_6 = __pyx_cur_scope->__pyx_t_1;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(22, 95, __pyx_L1_error)
    } else {
      PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
      if (exc_type) {
        if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
        else __PYX_ERR(22, 95, __pyx_L1_error)
      }
    }

    /* "src/union.pxi":94
 *             return
 *         yield from super().walk_types(visited)
 *         for t in self.union_types:             # <<<<<<<<<<<<<<
 *             yield from t.walk_types(visited)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "src/union.pxi":90
 *         return dict()
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":97
 *             yield from t.walk_types(visited)
 * 
 *     cpdef get_schema(self, created=None):             # <<<<<<<<<<<<<<
 *         if created is None:
 *             created = set()
 */

static PyObject *__pyx_pw_5cavro_9UnionType_8get_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_9UnionType_get_schema(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_9UnionType_get_schema *__pyx_optional_args) {
  PyObject *__pyx_v_created = ((PyObject *)Py_None);
  PyObject *__pyx_9genexpr22__pyx_v_t = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_schema", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_created = __pyx_optional_args->created;
    }
  }
  __Pyx_INCREF(__pyx_v_created);

  /* "src/union.pxi":98
 * 
 *     cpdef get_schema(self, created=None):
 *         if created is None:             # <<<<<<<<<<<<<<
 *             created = set()
 *         return [t.get_schema(created) for t in self.union_types]
 */
  __pyx_t_1 = (__pyx_v_created == Py_None);
  if (__pyx_t_1) {

    /* "src/union.pxi":99
 *     cpdef get_schema(self, created=None):
 *         if created is None:
 *             created = set()             # <<<<<<<<<<<<<<
 *         return [t.get_schema(created) for t in self.union_types]
 * 
 */
    __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_created, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/union.pxi":98
 * 
 *     cpdef get_schema(self, created=None):
 *         if created is None:             # <<<<<<<<<<<<<<
 *             created = set()
 *         return [t.get_schema(created) for t in self.union_types]
 */
  }

  /* "src/union.pxi":100
 *         if created is None:
 *             created = set()
 *         return [t.get_schema(created) for t in self.union_types]             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t resolve_from_value(self, object value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 100, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_self->union_types == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(22, 100, __pyx_L6_error)
    }
    __pyx_t_3 = __pyx_v_self->union_types; __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 100, __pyx_L6_error)
        #endif
        if (__pyx_t_4 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(22, 100, __pyx_L6_error)
      #else
      __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(22, 100, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_XDECREF_SET(__pyx_9genexpr22__pyx_v_t, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr22__pyx_v_t, __pyx_n_s_get_schema); if (unlikely(!__pyx_t_6)) __PYX_ERR(22, 100, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_created};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(22, 100, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_5))) __PYX_ERR(22, 100, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr22__pyx_v_t); __pyx_9genexpr22__pyx_v_t = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_9genexpr22__pyx_v_t); __pyx_9genexpr22__pyx_v_t = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/union.pxi":97
 *             yield from t.walk_types(visited)
 * 
 *     cpdef get_schema(self, created=None):             # <<<<<<<<<<<<<<
 *         if created is None:
 *             created = set()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.UnionType.get_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr22__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_created);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9UnionType_8get_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9UnionType_8get_schema = {"get_schema", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9UnionType_8get_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9UnionType_8get_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_schema (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 97, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_schema") < 0)) __PYX_ERR(22, 97, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_created = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_schema", 0, 0, 1, __pyx_nargs); __PYX_ERR(22, 97, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UnionType.get_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9UnionType_7get_schema(((struct __pyx_obj_5cavro_UnionType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9UnionType_7get_schema(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_get_schema __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_schema", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.created = __pyx_v_created;
  __pyx_t_1 = __pyx_vtabptr_5cavro_UnionType->__pyx_base.get_schema(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.UnionType.get_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":102
 *         return [t.get_schema(created) for t in self.union_types]
 * 
 *     cdef Py_ssize_t resolve_from_value(self, object value) except -1:             # <<<<<<<<<<<<<<
 *         cdef AvroType candidate
 *         cdef int type_fitness
 */

static Py_ssize_t __pyx_f_5cavro_9UnionType_resolve_from_value(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_candidate = 0;
  int __pyx_v_type_fitness;
  int __pyx_v_cur_fit;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_best_index;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resolve_from_value", 1);

  /* "src/union.pxi":105
 *         cdef AvroType candidate
 *         cdef int type_fitness
 *         cdef int cur_fit = FIT_NONE             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t i = 0
 *         cdef Py_ssize_t best_index = -1
 */
  __pyx_v_cur_fit = __pyx_v_5cavro_FIT_NONE;

  /* "src/union.pxi":106
 *         cdef int type_fitness
 *         cdef int cur_fit = FIT_NONE
 *         cdef Py_ssize_t i = 0             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t best_index = -1
 *         for candidate in self.union_types:
 */
  __pyx_v_i = 0;

  /* "src/union.pxi":107
 *         cdef int cur_fit = FIT_NONE
 *         cdef Py_ssize_t i = 0
 *         cdef Py_ssize_t best_index = -1             # <<<<<<<<<<<<<<
 *         for candidate in self.union_types:
 *             type_fitness = candidate.get_value_fitness(value)
 */
  __pyx_v_best_index = -1L;

  /* "src/union.pxi":108
 *         cdef Py_ssize_t i = 0
 *         cdef Py_ssize_t best_index = -1
 *         for candidate in self.union_types:             # <<<<<<<<<<<<<<
 *             type_fitness = candidate.get_value_fitness(value)
 *             if type_fitness > cur_fit:
 */
  if (unlikely(__pyx_v_self->union_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 108, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->union_types; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 108, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 108, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 108, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_candidate, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/union.pxi":109
 *         cdef Py_ssize_t best_index = -1
 *         for candidate in self.union_types:
 *             type_fitness = candidate.get_value_fitness(value)             # <<<<<<<<<<<<<<
 *             if type_fitness > cur_fit:
 *                 best_index = i
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_candidate->__pyx_vtab)->get_value_fitness(__pyx_v_candidate, __pyx_v_value); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(22, 109, __pyx_L1_error)
    __pyx_v_type_fitness = __pyx_t_4;

    /* "src/union.pxi":110
 *         for candidate in self.union_types:
 *             type_fitness = candidate.get_value_fitness(value)
 *             if type_fitness > cur_fit:             # <<<<<<<<<<<<<<
 *                 best_index = i
 *                 if type_fitness == FIT_EXACT:
 */
    __pyx_t_5 = (__pyx_v_type_fitness > __pyx_v_cur_fit);
    if (__pyx_t_5) {

      /* "src/union.pxi":111
 *             type_fitness = candidate.get_value_fitness(value)
 *             if type_fitness > cur_fit:
 *                 best_index = i             # <<<<<<<<<<<<<<
 *                 if type_fitness == FIT_EXACT:
 *                     return i
 */
      __pyx_v_best_index = __pyx_v_i;

      /* "src/union.pxi":112
 *             if type_fitness > cur_fit:
 *                 best_index = i
 *                 if type_fitness == FIT_EXACT:             # <<<<<<<<<<<<<<
 *                     return i
 *                 cur_fit = type_fitness
 */
      __pyx_t_5 = (__pyx_v_type_fitness == __pyx_v_5cavro_FIT_EXACT);
      if (__pyx_t_5) {

        /* "src/union.pxi":113
 *                 best_index = i
 *                 if type_fitness == FIT_EXACT:
 *                     return i             # <<<<<<<<<<<<<<
 *                 cur_fit = type_fitness
 *             i += 1
 */
        __pyx_r = __pyx_v_i;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;

        /* "src/union.pxi":112
 *             if type_fitness > cur_fit:
 *                 best_index = i
 *                 if type_fitness == FIT_EXACT:             # <<<<<<<<<<<<<<
 *                     return i
 *                 cur_fit = type_fitness
 */
      }

      /* "src/union.pxi":114
 *                 if type_fitness == FIT_EXACT:
 *                     return i
 *                 cur_fit = type_fitness             # <<<<<<<<<<<<<<
 *             i += 1
 *         if cur_fit == FIT_NONE or best_index < 0:
 */
      __pyx_v_cur_fit = __pyx_v_type_fitness;

      /* "src/union.pxi":110
 *         for candidate in self.union_types:
 *             type_fitness = candidate.get_value_fitness(value)
 *             if type_fitness > cur_fit:             # <<<<<<<<<<<<<<
 *                 best_index = i
 *                 if type_fitness == FIT_EXACT:
 */
    }

    /* "src/union.pxi":115
 *                     return i
 *                 cur_fit = type_fitness
 *             i += 1             # <<<<<<<<<<<<<<
 *         if cur_fit == FIT_NONE or best_index < 0:
 *             raise InvalidValue(value, self)
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "src/union.pxi":108
 *         cdef Py_ssize_t i = 0
 *         cdef Py_ssize_t best_index = -1
 *         for candidate in self.union_types:             # <<<<<<<<<<<<<<
 *             type_fitness = candidate.get_value_fitness(value)
 *             if type_fitness > cur_fit:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/union.pxi":116
 *                 cur_fit = type_fitness
 *             i += 1
 *         if cur_fit == FIT_NONE or best_index < 0:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         return best_index
 */
  __pyx_t_6 = (__pyx_v_cur_fit == __pyx_v_5cavro_FIT_NONE);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_6 = (__pyx_v_best_index < 0);
  __pyx_t_5 = __pyx_t_6;
  __pyx_L9_bool_binop_done:;
  if (unlikely(__pyx_t_5)) {

    /* "src/union.pxi":117
 *             i += 1
 *         if cur_fit == FIT_NONE or best_index < 0:
 *             raise InvalidValue(value, self)             # <<<<<<<<<<<<<<
 *         return best_index
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_v_value, ((PyObject *)__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(22, 117, __pyx_L1_error)

    /* "src/union.pxi":116
 *                 cur_fit = type_fitness
 *             i += 1
 *         if cur_fit == FIT_NONE or best_index < 0:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self)
 *         return best_index
 */
  }

  /* "src/union.pxi":118
 *         if cur_fit == FIT_NONE or best_index < 0:
 *             raise InvalidValue(value, self)
 *         return best_index             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __pyx_r = __pyx_v_best_index;
  goto __pyx_L0;

  /* "src/union.pxi":102
 *         return [t.get_schema(created) for t in self.union_types]
 * 
 *     cdef Py_ssize_t resolve_from_value(self, object value) except -1:             # <<<<<<<<<<<<<<
 *         cdef AvroType candidate
 *         cdef int type_fitness
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.UnionType.resolve_from_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_candidate);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":120
 *         return best_index
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef size_t type_index = self.resolve_from_value(value)
 *         cdef AvroType encode_type = self.union_types[type_index]
 */

static int __pyx_f_5cavro_9UnionType__binary_buffer_encode(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  size_t __pyx_v_type_index;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_encode_type = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 1);

  /* "src/union.pxi":121
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef size_t type_index = self.resolve_from_value(value)             # <<<<<<<<<<<<<<
 *         cdef AvroType encode_type = self.union_types[type_index]
 *         zigzag_encode_long(buffer, type_index)
 */
  __pyx_t_1 = __pyx_f_5cavro_9UnionType_resolve_from_value(__pyx_v_self, __pyx_v_value); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1L))) __PYX_ERR(22, 121, __pyx_L1_error)
  __pyx_v_type_index = __pyx_t_1;

  /* "src/union.pxi":122
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         cdef size_t type_index = self.resolve_from_value(value)
 *         cdef AvroType encode_type = self.union_types[type_index]             # <<<<<<<<<<<<<<
 *         zigzag_encode_long(buffer, type_index)
 *         encode_type.binary_buffer_encode(buffer, value)
 */
  if (unlikely(__pyx_v_self->union_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(22, 122, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_GetItemInt_Tuple(__pyx_v_self->union_types, __pyx_v_type_index, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 122, __pyx_L1_error)
  __pyx_v_encode_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/union.pxi":123
 *         cdef size_t type_index = self.resolve_from_value(value)
 *         cdef AvroType encode_type = self.union_types[type_index]
 *         zigzag_encode_long(buffer, type_index)             # <<<<<<<<<<<<<<
 *         encode_type.binary_buffer_encode(buffer, value)
 *         return 0
 */
  __pyx_t_3 = __pyx_f_5cavro_zigzag_encode_long(__pyx_v_buffer, __pyx_v_type_index); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(22, 123, __pyx_L1_error)

  /* "src/union.pxi":124
 *         cdef AvroType encode_type = self.union_types[type_index]
 *         zigzag_encode_long(buffer, type_index)
 *         encode_type.binary_buffer_encode(buffer, value)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_encode_type->__pyx_vtab)->binary_buffer_encode(__pyx_v_encode_type, __pyx_v_buffer, __pyx_v_value); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(22, 124, __pyx_L1_error)

  /* "src/union.pxi":125
 *         zigzag_encode_long(buffer, type_index)
 *         encode_type.binary_buffer_encode(buffer, value)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "src/union.pxi":120
 *         return best_index
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef size_t type_index = self.resolve_from_value(value)
 *         cdef AvroType encode_type = self.union_types[type_index]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.UnionType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_encode_type);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":127
 *         return 0
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t index = zigzag_decode_long(buffer)
 *         if index < 0 or index >= len(self.union_types):
 */

static PyObject *__pyx_f_5cavro_9UnionType__binary_buffer_decode(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  Py_ssize_t __pyx_v_index;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_item = 0;
  PyObject *__pyx_v_decoded = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/union.pxi":128
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef Py_ssize_t index = zigzag_decode_long(buffer)             # <<<<<<<<<<<<<<
 *         if index < 0 or index >= len(self.union_types):
 *             raise ValueError(f"Value {index} is not valid for a union of {len(self.union_types)} items")
 */
  __pyx_t_1 = __pyx_f_5cavro_zigzag_decode_long(__pyx_v_buffer); if (unlikely(__pyx_t_1 == ((int64_t)0x7ffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(22, 128, __pyx_L1_error)
  __pyx_v_index = __pyx_t_1;

  /* "src/union.pxi":129
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef Py_ssize_t index = zigzag_decode_long(buffer)
 *         if index < 0 or index >= len(self.union_types):             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Value {index} is not valid for a union of {len(self.union_types)} items")
 *         cdef AvroType item = self.union_types[index]
 */
  __pyx_t_3 = (__pyx_v_index < 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __pyx_v_self->union_types;
  __Pyx_INCREF(__pyx_t_4);
  if (unlikely(__pyx_t_4 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(22, 129, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(22, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__pyx_v_index >= __pyx_t_5);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "src/union.pxi":130
 *         cdef Py_ssize_t index = zigzag_decode_long(buffer)
 *         if index < 0 or index >= len(self.union_types):
 *             raise ValueError(f"Value {index} is not valid for a union of {len(self.union_types)} items")             # <<<<<<<<<<<<<<
 *         cdef AvroType item = self.union_types[index]
 *         decoded = item.binary_buffer_decode(buffer)
 */
    __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_Value);
    __pyx_t_5 += 6;
    __Pyx_GIVEREF(__pyx_kp_u_Value);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Value);
    __pyx_t_7 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_index, 0, ' ', 'd'); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_is_not_valid_for_a_union_of);
    __pyx_t_5 += 29;
    __Pyx_GIVEREF(__pyx_kp_u_is_not_valid_for_a_union_of);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_is_not_valid_for_a_union_of);
    __pyx_t_7 = __pyx_v_self->union_types;
    __Pyx_INCREF(__pyx_t_7);
    if (unlikely(__pyx_t_7 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(22, 130, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyTuple_GET_SIZE(__pyx_t_7); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(22, 130, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_8, 0, ' ', 'd'); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_items_2);
    __pyx_t_5 += 6;
    __Pyx_GIVEREF(__pyx_kp_u_items_2);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u_items_2);
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_4, 5, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(22, 130, __pyx_L1_error)

    /* "src/union.pxi":129
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef Py_ssize_t index = zigzag_decode_long(buffer)
 *         if index < 0 or index >= len(self.union_types):             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Value {index} is not valid for a union of {len(self.union_types)} items")
 *         cdef AvroType item = self.union_types[index]
 */
  }

  /* "src/union.pxi":131
 *         if index < 0 or index >= len(self.union_types):
 *             raise ValueError(f"Value {index} is not valid for a union of {len(self.union_types)} items")
 *         cdef AvroType item = self.union_types[index]             # <<<<<<<<<<<<<<
 *         decoded = item.binary_buffer_decode(buffer)
 *         if self.return_type_tuple[index]:
 */
  if (unlikely(__pyx_v_self->union_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(22, 131, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_GetItemInt_Tuple(__pyx_v_self->union_types, __pyx_v_index, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 131, __pyx_L1_error)
  __pyx_v_item = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "src/union.pxi":132
 *             raise ValueError(f"Value {index} is not valid for a union of {len(self.union_types)} items")
 *         cdef AvroType item = self.union_types[index]
 *         decoded = item.binary_buffer_decode(buffer)             # <<<<<<<<<<<<<<
 *         if self.return_type_tuple[index]:
 *             return (item.type, decoded)
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_item->__pyx_vtab)->binary_buffer_decode(__pyx_v_item, __pyx_v_buffer); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_decoded = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "src/union.pxi":133
 *         cdef AvroType item = self.union_types[index]
 *         decoded = item.binary_buffer_decode(buffer)
 *         if self.return_type_tuple[index]:             # <<<<<<<<<<<<<<
 *             return (item.type, decoded)
 *         return decoded
 */
  if (unlikely(__pyx_v_self->return_type_tuple == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(22, 133, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_GetItemInt_Tuple(__pyx_v_self->return_type_tuple, __pyx_v_index, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(22, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_2) {

    /* "src/union.pxi":134
 *         decoded = item.binary_buffer_decode(buffer)
 *         if self.return_type_tuple[index]:
 *             return (item.type, decoded)             # <<<<<<<<<<<<<<
 *         return decoded
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_item), __pyx_n_s_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4)) __PYX_ERR(22, 134, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_decoded);
    __Pyx_GIVEREF(__pyx_v_decoded);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_decoded)) __PYX_ERR(22, 134, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;

    /* "src/union.pxi":133
 *         cdef AvroType item = self.union_types[index]
 *         decoded = item.binary_buffer_decode(buffer)
 *         if self.return_type_tuple[index]:             # <<<<<<<<<<<<<<
 *             return (item.type, decoded)
 *         return decoded
 */
  }

  /* "src/union.pxi":135
 *         if self.return_type_tuple[index]:
 *             return (item.type, decoded)
 *         return decoded             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_decoded);
  __pyx_r = __pyx_v_decoded;
  goto __pyx_L0;

  /* "src/union.pxi":127
 *         return 0
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t index = zigzag_decode_long(buffer)
 *         if index < 0 or index >= len(self.union_types):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.UnionType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_item);
  __Pyx_XDECREF(__pyx_v_decoded);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":137
 *         return decoded
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef AvroType union_type
 *         cdef int level = FIT_NONE
 */

static int __pyx_f_5cavro_9UnionType__get_value_fitness(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_union_type = 0;
  int __pyx_v_level;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 1);

  /* "src/union.pxi":139
 *     cdef int _get_value_fitness(self, value) except -1:
 *         cdef AvroType union_type
 *         cdef int level = FIT_NONE             # <<<<<<<<<<<<<<
 *         for union_type in self.union_types:
 *             level = max(level, union_type.get_value_fitness(value))
 */
  __pyx_v_level = __pyx_v_5cavro_FIT_NONE;

  /* "src/union.pxi":140
 *         cdef AvroType union_type
 *         cdef int level = FIT_NONE
 *         for union_type in self.union_types:             # <<<<<<<<<<<<<<
 *             level = max(level, union_type.get_value_fitness(value))
 *             if level == FIT_EXACT:
 */
  if (unlikely(__pyx_v_self->union_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 140, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->union_types; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 140, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 140, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 140, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_union_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/union.pxi":141
 *         cdef int level = FIT_NONE
 *         for union_type in self.union_types:
 *             level = max(level, union_type.get_value_fitness(value))             # <<<<<<<<<<<<<<
 *             if level == FIT_EXACT:
 *                 return level
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_union_type->__pyx_vtab)->get_value_fitness(__pyx_v_union_type, __pyx_v_value); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(22, 141, __pyx_L1_error)
    __pyx_t_5 = __pyx_v_level;
    __pyx_t_7 = (__pyx_t_4 > __pyx_t_5);
    if (__pyx_t_7) {
      __pyx_t_6 = __pyx_t_4;
    } else {
      __pyx_t_6 = __pyx_t_5;
    }
    __pyx_v_level = __pyx_t_6;

    /* "src/union.pxi":142
 *         for union_type in self.union_types:
 *             level = max(level, union_type.get_value_fitness(value))
 *             if level == FIT_EXACT:             # <<<<<<<<<<<<<<
 *                 return level
 *         return level
 */
    __pyx_t_7 = (__pyx_v_level == __pyx_v_5cavro_FIT_EXACT);
    if (__pyx_t_7) {

      /* "src/union.pxi":143
 *             level = max(level, union_type.get_value_fitness(value))
 *             if level == FIT_EXACT:
 *                 return level             # <<<<<<<<<<<<<<
 *         return level
 * 
 */
      __pyx_r = __pyx_v_level;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "src/union.pxi":142
 *         for union_type in self.union_types:
 *             level = max(level, union_type.get_value_fitness(value))
 *             if level == FIT_EXACT:             # <<<<<<<<<<<<<<
 *                 return level
 *         return level
 */
    }

    /* "src/union.pxi":140
 *         cdef AvroType union_type
 *         cdef int level = FIT_NONE
 *         for union_type in self.union_types:             # <<<<<<<<<<<<<<
 *             level = max(level, union_type.get_value_fitness(value))
 *             if level == FIT_EXACT:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/union.pxi":144
 *             if level == FIT_EXACT:
 *                 return level
 *         return level             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
  __pyx_r = __pyx_v_level;
  goto __pyx_L0;

  /* "src/union.pxi":137
 *         return decoded
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef AvroType union_type
 *         cdef int level = FIT_NONE
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.UnionType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_union_type);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":146
 *         return level
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         cdef size_t type_index = self.resolve_from_value(value)
 *         cdef AvroType union_type = self.union_types[type_index]
 */

static PyObject *__pyx_f_5cavro_9UnionType__json_format(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_value) {
  size_t __pyx_v_type_index;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_union_type = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 1);

  /* "src/union.pxi":147
 * 
 *     cdef _json_format(self, value):
 *         cdef size_t type_index = self.resolve_from_value(value)             # <<<<<<<<<<<<<<
 *         cdef AvroType union_type = self.union_types[type_index]
 *         if isinstance(union_type, NullType):
 */
  __pyx_t_1 = __pyx_f_5cavro_9UnionType_resolve_from_value(__pyx_v_self, __pyx_v_value); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1L))) __PYX_ERR(22, 147, __pyx_L1_error)
  __pyx_v_type_index = __pyx_t_1;

  /* "src/union.pxi":148
 *     cdef _json_format(self, value):
 *         cdef size_t type_index = self.resolve_from_value(value)
 *         cdef AvroType union_type = self.union_types[type_index]             # <<<<<<<<<<<<<<
 *         if isinstance(union_type, NullType):
 *             return None
 */
  if (unlikely(__pyx_v_self->union_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(22, 148, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_GetItemInt_Tuple(__pyx_v_self->union_types, __pyx_v_type_index, size_t, 0, __Pyx_PyInt_FromSize_t, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 148, __pyx_L1_error)
  __pyx_v_union_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/union.pxi":149
 *         cdef size_t type_index = self.resolve_from_value(value)
 *         cdef AvroType union_type = self.union_types[type_index]
 *         if isinstance(union_type, NullType):             # <<<<<<<<<<<<<<
 *             return None
 *         if self.options.union_json_encodes_type_name:
 */
  __pyx_t_3 = __Pyx_TypeCheck(((PyObject *)__pyx_v_union_type), __pyx_ptype_5cavro_NullType); 
  if (__pyx_t_3) {

    /* "src/union.pxi":150
 *         cdef AvroType union_type = self.union_types[type_index]
 *         if isinstance(union_type, NullType):
 *             return None             # <<<<<<<<<<<<<<
 *         if self.options.union_json_encodes_type_name:
 *             return {union_type.type: union_type.json_format(value)}
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/union.pxi":149
 *         cdef size_t type_index = self.resolve_from_value(value)
 *         cdef AvroType union_type = self.union_types[type_index]
 *         if isinstance(union_type, NullType):             # <<<<<<<<<<<<<<
 *             return None
 *         if self.options.union_json_encodes_type_name:
 */
  }

  /* "src/union.pxi":151
 *         if isinstance(union_type, NullType):
 *             return None
 *         if self.options.union_json_encodes_type_name:             # <<<<<<<<<<<<<<
 *             return {union_type.type: union_type.json_format(value)}
 *         return union_type.json_format(value)
 */
  if (__pyx_v_self->__pyx_base.options->union_json_encodes_type_name) {

    /* "src/union.pxi":152
 *             return None
 *         if self.options.union_json_encodes_type_name:
 *             return {union_type.type: union_type.json_format(value)}             # <<<<<<<<<<<<<<
 *         return union_type.json_format(value)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_union_type), __pyx_n_s_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_union_type->__pyx_vtab)->json_format(__pyx_v_union_type, __pyx_v_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(22, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_2, __pyx_t_4, __pyx_t_5) < 0) __PYX_ERR(22, 152, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "src/union.pxi":151
 *         if isinstance(union_type, NullType):
 *             return None
 *         if self.options.union_json_encodes_type_name:             # <<<<<<<<<<<<<<
 *             return {union_type.type: union_type.json_format(value)}
 *         return union_type.json_format(value)
 */
  }

  /* "src/union.pxi":153
 *         if self.options.union_json_encodes_type_name:
 *             return {union_type.type: union_type.json_format(value)}
 *         return union_type.json_format(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_union_type->__pyx_vtab)->json_format(__pyx_v_union_type, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/union.pxi":146
 *         return level
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         cdef size_t type_index = self.resolve_from_value(value)
 *         cdef AvroType union_type = self.union_types[type_index]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.UnionType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_union_type);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":155
 *         return union_type.json_format(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if value is None or value is MISSING_VALUE:
 *             value = {'null': None}
 */

static PyObject *__pyx_f_5cavro_9UnionType__json_decode(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_value_dict = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_item_type = 0;
  PyObject *__pyx_v_type_name = NULL;
  PyObject *__pyx_v_item_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/union.pxi":156
 * 
 *     cdef _json_decode(self, value):
 *         if value is None or value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *             value = {'null': None}
 * 
 */
  __pyx_t_2 = (__pyx_v_value == Py_None);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = (__pyx_v_value == __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/union.pxi":157
 *     cdef _json_decode(self, value):
 *         if value is None or value is MISSING_VALUE:
 *             value = {'null': None}             # <<<<<<<<<<<<<<
 * 
 *         cdef dict value_dict = value
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_null_2, Py_None) < 0) __PYX_ERR(22, 157, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/union.pxi":156
 * 
 *     cdef _json_decode(self, value):
 *         if value is None or value is MISSING_VALUE:             # <<<<<<<<<<<<<<
 *             value = {'null': None}
 * 
 */
  }

  /* "src/union.pxi":159
 *             value = {'null': None}
 * 
 *         cdef dict value_dict = value             # <<<<<<<<<<<<<<
 *         if len(value) > 1:
 *             raise ValueError(f"Value {value} is not a valid union value (expect exactly one item)")
 */
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_value))) __PYX_ERR(22, 159, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_value_dict = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/union.pxi":160
 * 
 *         cdef dict value_dict = value
 *         if len(value) > 1:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Value {value} is not a valid union value (expect exactly one item)")
 *         cdef AvroType item_type
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(22, 160, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_4 > 1);
  if (unlikely(__pyx_t_1)) {

    /* "src/union.pxi":161
 *         cdef dict value_dict = value
 *         if len(value) > 1:
 *             raise ValueError(f"Value {value} is not a valid union value (expect exactly one item)")             # <<<<<<<<<<<<<<
 *         cdef AvroType item_type
 *         (type_name, item_value), = value_dict.items()
 */
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Value);
    __pyx_t_4 += 6;
    __Pyx_GIVEREF(__pyx_kp_u_Value);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Value);
    __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(22, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_is_not_a_valid_union_value_expe);
    __pyx_t_4 += 53;
    __Pyx_GIVEREF(__pyx_kp_u_is_not_a_valid_union_value_expe);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_is_not_a_valid_union_value_expe);
    __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(22, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(22, 161, __pyx_L1_error)

    /* "src/union.pxi":160
 * 
 *         cdef dict value_dict = value
 *         if len(value) > 1:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Value {value} is not a valid union value (expect exactly one item)")
 *         cdef AvroType item_type
 */
  }

  /* "src/union.pxi":163
 *             raise ValueError(f"Value {value} is not a valid union value (expect exactly one item)")
 *         cdef AvroType item_type
 *         (type_name, item_value), = value_dict.items()             # <<<<<<<<<<<<<<
 *         try:
 *             item_type = self.by_name_map[type_name]
 */
  if (unlikely(__pyx_v_value_dict == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(22, 163, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyDict_Items(__pyx_v_value_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 1)) {
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(22, 163, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
    } else {
      __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
    }
    __Pyx_INCREF(__pyx_t_6);
    #else
    __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(22, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_7 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
    index = 0; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L7_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 1) < 0) __PYX_ERR(22, 163, __pyx_L1_error)
    __pyx_t_8 = NULL;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L8_unpacking_done;
    __pyx_L7_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(22, 163, __pyx_L1_error)
    __pyx_L8_unpacking_done:;
  }
  if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
    PyObject* sequence = __pyx_t_6;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(22, 163, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_9);
    #else
    __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(22, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    #endif
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_10 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(22, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_10);
    index = 0; __pyx_t_7 = __pyx_t_8(__pyx_t_10); if (unlikely(!__pyx_t_7)) goto __pyx_L9_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_7);
    index = 1; __pyx_t_9 = __pyx_t_8(__pyx_t_10); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_9);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_10), 2) < 0) __PYX_ERR(22, 163, __pyx_L1_error)
    __pyx_t_8 = NULL;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    goto __pyx_L10_unpacking_done;
    __pyx_L9_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(22, 163, __pyx_L1_error)
    __pyx_L10_unpacking_done:;
  }
  __pyx_v_type_name = __pyx_t_7;
  __pyx_t_7 = 0;
  __pyx_v_item_value = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "src/union.pxi":164
 *         cdef AvroType item_type
 *         (type_name, item_value), = value_dict.items()
 *         try:             # <<<<<<<<<<<<<<
 *             item_type = self.by_name_map[type_name]
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
    __Pyx_XGOTREF(__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_12);
    __Pyx_XGOTREF(__pyx_t_13);
    /*try:*/ {

      /* "src/union.pxi":165
 *         (type_name, item_value), = value_dict.items()
 *         try:
 *             item_type = self.by_name_map[type_name]             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             raise ValueError(f"Value {value} is not a valid union value (unknown type '{type_name}')")
 */
      if (unlikely(__pyx_v_self->by_name_map == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(22, 165, __pyx_L11_error)
      }
      __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_self->by_name_map, __pyx_v_type_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 165, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 165, __pyx_L11_error)
      __pyx_v_item_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "src/union.pxi":164
 *         cdef AvroType item_type
 *         (type_name, item_value), = value_dict.items()
 *         try:             # <<<<<<<<<<<<<<
 *             item_type = self.by_name_map[type_name]
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    goto __pyx_L16_try_end;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "src/union.pxi":166
 *         try:
 *             item_type = self.by_name_map[type_name]
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Value {value} is not a valid union value (unknown type '{type_name}')")
 *         return item_type.json_decode(item_value)
 */
    __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_14) {
      __Pyx_AddTraceback("cavro.UnionType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_6, &__pyx_t_9) < 0) __PYX_ERR(22, 166, __pyx_L13_except_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_9);

      /* "src/union.pxi":167
 *             item_type = self.by_name_map[type_name]
 *         except KeyError:
 *             raise ValueError(f"Value {value} is not a valid union value (unknown type '{type_name}')")             # <<<<<<<<<<<<<<
 *         return item_type.json_decode(item_value)
 * 
 */
      __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 167, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = 0;
      __pyx_t_5 = 127;
      __Pyx_INCREF(__pyx_kp_u_Value);
      __pyx_t_4 += 6;
      __Pyx_GIVEREF(__pyx_kp_u_Value);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_kp_u_Value);
      __pyx_t_10 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(22, 167, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_5;
      __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_INCREF(__pyx_kp_u_is_not_a_valid_union_value_unkn);
      __pyx_t_4 += 43;
      __Pyx_GIVEREF(__pyx_kp_u_is_not_a_valid_union_value_unkn);
      PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_kp_u_is_not_a_valid_union_value_unkn);
      __pyx_t_10 = __Pyx_PyObject_FormatSimple(__pyx_v_type_name, __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(22, 167, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_5;
      __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_INCREF(__pyx_kp_u__76);
      __pyx_t_4 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__76);
      PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_kp_u__76);
      __pyx_t_10 = __Pyx_PyUnicode_Join(__pyx_t_7, 5, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(22, 167, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 167, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(22, 167, __pyx_L13_except_error)
    }
    goto __pyx_L13_except_error;

    /* "src/union.pxi":164
 *         cdef AvroType item_type
 *         (type_name, item_value), = value_dict.items()
 *         try:             # <<<<<<<<<<<<<<
 *             item_type = self.by_name_map[type_name]
 *         except KeyError:
 */
    __pyx_L13_except_error:;
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_XGIVEREF(__pyx_t_12);
    __Pyx_XGIVEREF(__pyx_t_13);
    __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
    goto __pyx_L1_error;
    __pyx_L16_try_end:;
  }

  /* "src/union.pxi":168
 *         except KeyError:
 *             raise ValueError(f"Value {value} is not a valid union value (unknown type '{type_name}')")
 *         return item_type.json_decode(item_value)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object convert_value(self, object value, check_value=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_9 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_item_type->__pyx_vtab)->json_decode(__pyx_v_item_type, __pyx_v_item_value); if (unlikely(!__pyx_t_9)) __PYX_ERR(22, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_r = __pyx_t_9;
  __pyx_t_9 = 0;
  goto __pyx_L0;

  /* "src/union.pxi":155
 *         return union_type.json_format(value)
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         if value is None or value is MISSING_VALUE:
 *             value = {'null': None}
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("cavro.UnionType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value_dict);
  __Pyx_XDECREF((PyObject *)__pyx_v_item_type);
  __Pyx_XDECREF(__pyx_v_type_name);
  __Pyx_XDECREF(__pyx_v_item_value);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":170
 *         return item_type.json_decode(item_value)
 * 
 *     cpdef object convert_value(self, object value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t index = self.resolve_from_value(value)
 *         return self.union_types[index].convert_value(value)
 */

static PyObject *__pyx_pw_5cavro_9UnionType_10convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_9UnionType_convert_value(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_9UnionType_convert_value *__pyx_optional_args) {
  Py_ssize_t __pyx_v_index;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_value", 1);
  if (__pyx_optional_args) {
  }

  /* "src/union.pxi":171
 * 
 *     cpdef object convert_value(self, object value, check_value=True):
 *         cdef Py_ssize_t index = self.resolve_from_value(value)             # <<<<<<<<<<<<<<
 *         return self.union_types[index].convert_value(value)
 * 
 */
  __pyx_t_1 = __pyx_f_5cavro_9UnionType_resolve_from_value(__pyx_v_self, __pyx_v_value); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1L))) __PYX_ERR(22, 171, __pyx_L1_error)
  __pyx_v_index = __pyx_t_1;

  /* "src/union.pxi":172
 *     cpdef object convert_value(self, object value, check_value=True):
 *         cdef Py_ssize_t index = self.resolve_from_value(value)
 *         return self.union_types[index].convert_value(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->union_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(22, 172, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_GetItemInt_Tuple(__pyx_v_self->union_types, __pyx_v_index, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_convert_value_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_value};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/union.pxi":170
 *         return item_type.json_decode(item_value)
 * 
 *     cpdef object convert_value(self, object value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t index = self.resolve_from_value(value)
 *         return self.union_types[index].convert_value(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.UnionType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9UnionType_10convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9UnionType_10convert_value = {"convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9UnionType_10convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9UnionType_10convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_check_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,&__pyx_n_s_check_value,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_True));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 170, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_check_value);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 170, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "convert_value") < 0)) __PYX_ERR(22, 170, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_value = values[0];
    __pyx_v_check_value = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convert_value", 0, 1, 2, __pyx_nargs); __PYX_ERR(22, 170, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UnionType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9UnionType_9convert_value(((struct __pyx_obj_5cavro_UnionType *)__pyx_v_self), __pyx_v_value, __pyx_v_check_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9UnionType_9convert_value(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_check_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_convert_value __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.check_value = __pyx_v_check_value;
  __pyx_t_1 = __pyx_vtabptr_5cavro_UnionType->__pyx_base.convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.UnionType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":174
 *         return self.union_types[index].convert_value(value)
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         cdef list parts = []
 *         cdef AvroType avro_type
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_9UnionType_canonical_form(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_v_parts = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_avro_type = 0;
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/union.pxi":175
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         cdef list parts = []             # <<<<<<<<<<<<<<
 *         cdef AvroType avro_type
 *         for avro_type in self.union_types:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_parts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/union.pxi":177
 *         cdef list parts = []
 *         cdef AvroType avro_type
 *         for avro_type in self.union_types:             # <<<<<<<<<<<<<<
 *             parts.append(avro_type.canonical_form(created))
 *         return _CanonicalForm('[' + ','.join(parts) + ']')
 */
  if (unlikely(__pyx_v_self->union_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 177, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->union_types; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 177, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 177, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 177, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_avro_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/union.pxi":178
 *         cdef AvroType avro_type
 *         for avro_type in self.union_types:
 *             parts.append(avro_type.canonical_form(created))             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('[' + ','.join(parts) + ']')
 * 
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_avro_type->__pyx_vtab)->canonical_form(__pyx_v_avro_type, __pyx_v_created)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyList_Append(__pyx_v_parts, __pyx_t_3); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(22, 178, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "src/union.pxi":177
 *         cdef list parts = []
 *         cdef AvroType avro_type
 *         for avro_type in self.union_types:             # <<<<<<<<<<<<<<
 *             parts.append(avro_type.canonical_form(created))
 *         return _CanonicalForm('[' + ','.join(parts) + ']')
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/union.pxi":179
 *         for avro_type in self.union_types:
 *             parts.append(avro_type.canonical_form(created))
 *         return _CanonicalForm('[' + ','.join(parts) + ']')             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__77, __pyx_v_parts); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u__14, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_3, __pyx_kp_u__78); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/union.pxi":174
 *         return self.union_types[index].convert_value(value)
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         cdef list parts = []
 *         cdef AvroType avro_type
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.UnionType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_parts);
  __Pyx_XDECREF((PyObject *)__pyx_v_avro_type);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":181
 *         return _CanonicalForm('[' + ','.join(parts) + ']')
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         cdef AvroType sub_type
 *         cdef UnionType writer_union
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9UnionType__for_writer(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_sub_type = 0;
  struct __pyx_obj_5cavro_UnionType *__pyx_v_writer_union = 0;
  CYTHON_UNUSED int __pyx_v_got_one;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_for_writer __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_writer", 1);

  /* "src/union.pxi":184
 *         cdef AvroType sub_type
 *         cdef UnionType writer_union
 *         cdef bint got_one = False             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(writer, UnionType):
 */
  __pyx_v_got_one = 0;

  /* "src/union.pxi":186
 *         cdef bint got_one = False
 * 
 *         if not isinstance(writer, UnionType):             # <<<<<<<<<<<<<<
 *             for sub_type in self.union_types:
 *                 try:
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_UnionType); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "src/union.pxi":187
 * 
 *         if not isinstance(writer, UnionType):
 *             for sub_type in self.union_types:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return sub_type.for_writer(writer, False)
 */
    if (unlikely(__pyx_v_self->union_types == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(22, 187, __pyx_L1_error)
    }
    __pyx_t_3 = __pyx_v_self->union_types; __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 187, __pyx_L1_error)
        #endif
        if (__pyx_t_4 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(22, 187, __pyx_L1_error)
      #else
      __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(22, 187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 187, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_sub_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "src/union.pxi":188
 *         if not isinstance(writer, UnionType):
 *             for sub_type in self.union_types:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return sub_type.for_writer(writer, False)
 *                 except CannotPromoteError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {

          /* "src/union.pxi":189
 *             for sub_type in self.union_types:
 *                 try:
 *                     return sub_type.for_writer(writer, False)             # <<<<<<<<<<<<<<
 *                 except CannotPromoteError:
 *                     pass
 */
          __Pyx_XDECREF((PyObject *)__pyx_r);
          __pyx_t_9.__pyx_n = 1;
          __pyx_t_9._allow_deferrals = 0;
          __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_sub_type->__pyx_vtab)->for_writer(__pyx_v_sub_type, __pyx_v_writer, &__pyx_t_9)); if (unlikely(!__pyx_t_5)) __PYX_ERR(22, 189, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_5);
          __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L10_try_return;

          /* "src/union.pxi":188
 *         if not isinstance(writer, UnionType):
 *             for sub_type in self.union_types:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return sub_type.for_writer(writer, False)
 *                 except CannotPromoteError:
 */
        }
        __pyx_L6_error:;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "src/union.pxi":190
 *                 try:
 *                     return sub_type.for_writer(writer, False)
 *                 except CannotPromoteError:             # <<<<<<<<<<<<<<
 *                     pass
 *             return
 */
        __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_10, &__pyx_t_11);
        __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_CannotPromoteError); if (unlikely(!__pyx_t_12)) __PYX_ERR(22, 190, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_5, __pyx_t_12);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_ErrRestore(__pyx_t_5, __pyx_t_10, __pyx_t_11);
        __pyx_t_5 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
        if (__pyx_t_13) {
          __Pyx_ErrRestore(0,0,0);
          goto __pyx_L7_exception_handled;
        }
        goto __pyx_L8_except_error;

        /* "src/union.pxi":188
 *         if not isinstance(writer, UnionType):
 *             for sub_type in self.union_types:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return sub_type.for_writer(writer, False)
 *                 except CannotPromoteError:
 */
        __pyx_L8_except_error:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L10_try_return:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L0;
        __pyx_L7_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      }

      /* "src/union.pxi":187
 * 
 *         if not isinstance(writer, UnionType):
 *             for sub_type in self.union_types:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return sub_type.for_writer(writer, False)
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "src/union.pxi":192
 *                 except CannotPromoteError:
 *                     pass
 *             return             # <<<<<<<<<<<<<<
 *         writer_union = writer
 *         return writer_union.for_reader(self)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/union.pxi":186
 *         cdef bint got_one = False
 * 
 *         if not isinstance(writer, UnionType):             # <<<<<<<<<<<<<<
 *             for sub_type in self.union_types:
 *                 try:
 */
  }

  /* "src/union.pxi":193
 *                     pass
 *             return
 *         writer_union = writer             # <<<<<<<<<<<<<<
 *         return writer_union.for_reader(self)
 * 
 */
  if (!(likely(((((PyObject *)__pyx_v_writer)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_UnionType))))) __PYX_ERR(22, 193, __pyx_L1_error)
  __pyx_t_3 = ((PyObject *)__pyx_v_writer);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_writer_union = ((struct __pyx_obj_5cavro_UnionType *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/union.pxi":194
 *             return
 *         writer_union = writer
 *         return writer_union.for_reader(self)             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType for_reader(self, AvroType reader):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_3 = ((PyObject *)__pyx_f_5cavro_9UnionType_for_reader(__pyx_v_writer_union, ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self))); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/union.pxi":181
 *         return _CanonicalForm('[' + ','.join(parts) + ']')
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         cdef AvroType sub_type
 *         cdef UnionType writer_union
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("cavro.UnionType._for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sub_type);
  __Pyx_XDECREF((PyObject *)__pyx_v_writer_union);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":196
 *         return writer_union.for_reader(self)
 * 
 *     cdef AvroType for_reader(self, AvroType reader):             # <<<<<<<<<<<<<<
 *         cdef AvroType sub_type
 *         cdef UnionType new_writer
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_9UnionType_for_reader(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_reader) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_sub_type = 0;
  struct __pyx_obj_5cavro_UnionType *__pyx_v_new_writer = 0;
  int __pyx_v_got_one;
  PyObject *__pyx_v_new_sub_types = NULL;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_sub_adapted = NULL;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_for_writer __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("for_reader", 1);

  /* "src/union.pxi":199
 *         cdef AvroType sub_type
 *         cdef UnionType new_writer
 *         cdef bint got_one = False             # <<<<<<<<<<<<<<
 *         new_sub_types = []
 *         for sub_type in self.union_types:
 */
  __pyx_v_got_one = 0;

  /* "src/union.pxi":200
 *         cdef UnionType new_writer
 *         cdef bint got_one = False
 *         new_sub_types = []             # <<<<<<<<<<<<<<
 *         for sub_type in self.union_types:
 *             try:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_new_sub_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/union.pxi":201
 *         cdef bint got_one = False
 *         new_sub_types = []
 *         for sub_type in self.union_types:             # <<<<<<<<<<<<<<
 *             try:
 *                 sub_adapted = reader.for_writer(sub_type, False)
 */
  if (unlikely(__pyx_v_self->union_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 201, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->union_types; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 201, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 201, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 201, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_sub_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/union.pxi":202
 *         new_sub_types = []
 *         for sub_type in self.union_types:
 *             try:             # <<<<<<<<<<<<<<
 *                 sub_adapted = reader.for_writer(sub_type, False)
 *             except CannotPromoteError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "src/union.pxi":203
 *         for sub_type in self.union_types:
 *             try:
 *                 sub_adapted = reader.for_writer(sub_type, False)             # <<<<<<<<<<<<<<
 *             except CannotPromoteError:
 *                 sub_adapted = sub_type.copy()
 */
        __pyx_t_7.__pyx_n = 1;
        __pyx_t_7._allow_deferrals = 0;
        __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_reader->__pyx_vtab)->for_writer(__pyx_v_reader, __pyx_v_sub_type, &__pyx_t_7)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 203, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_v_sub_adapted, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3));
        __pyx_t_3 = 0;

        /* "src/union.pxi":202
 *         new_sub_types = []
 *         for sub_type in self.union_types:
 *             try:             # <<<<<<<<<<<<<<
 *                 sub_adapted = reader.for_writer(sub_type, False)
 *             except CannotPromoteError:
 */
      }

      /* "src/union.pxi":208
 *                 sub_adapted.value_adapters = (CannotPromote(reader, sub_type), ) + sub_adapted.value_adapters
 *             else:
 *                 got_one = True             # <<<<<<<<<<<<<<
 *             new_sub_types.append(sub_adapted)
 *         if not got_one:
 */
      /*else:*/ {
        __pyx_v_got_one = 1;
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "src/union.pxi":204
 *             try:
 *                 sub_adapted = reader.for_writer(sub_type, False)
 *             except CannotPromoteError:             # <<<<<<<<<<<<<<
 *                 sub_adapted = sub_type.copy()
 *                 sub_adapted.value_adapters = (CannotPromote(reader, sub_type), ) + sub_adapted.value_adapters
 */
      __Pyx_ErrFetch(&__pyx_t_3, &__pyx_t_8, &__pyx_t_9);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_CannotPromoteError); if (unlikely(!__pyx_t_10)) __PYX_ERR(22, 204, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_t_10);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_ErrRestore(__pyx_t_3, __pyx_t_8, __pyx_t_9);
      __pyx_t_3 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0;
      if (__pyx_t_11) {
        __Pyx_AddTraceback("cavro.UnionType.for_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_3) < 0) __PYX_ERR(22, 204, __pyx_L7_except_error)
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_3);

        /* "src/union.pxi":205
 *                 sub_adapted = reader.for_writer(sub_type, False)
 *             except CannotPromoteError:
 *                 sub_adapted = sub_type.copy()             # <<<<<<<<<<<<<<
 *                 sub_adapted.value_adapters = (CannotPromote(reader, sub_type), ) + sub_adapted.value_adapters
 *             else:
 */
        __pyx_t_10 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_sub_type->__pyx_vtab)->copy(__pyx_v_sub_type, 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(22, 205, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_XDECREF_SET(__pyx_v_sub_adapted, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_10));
        __pyx_t_10 = 0;

        /* "src/union.pxi":206
 *             except CannotPromoteError:
 *                 sub_adapted = sub_type.copy()
 *                 sub_adapted.value_adapters = (CannotPromote(reader, sub_type), ) + sub_adapted.value_adapters             # <<<<<<<<<<<<<<
 *             else:
 *                 got_one = True
 */
        __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(22, 206, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_INCREF((PyObject *)__pyx_v_reader);
        __Pyx_GIVEREF((PyObject *)__pyx_v_reader);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, ((PyObject *)__pyx_v_reader))) __PYX_ERR(22, 206, __pyx_L7_except_error);
        __Pyx_INCREF((PyObject *)__pyx_v_sub_type);
        __Pyx_GIVEREF((PyObject *)__pyx_v_sub_type);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, ((PyObject *)__pyx_v_sub_type))) __PYX_ERR(22, 206, __pyx_L7_except_error);
        __pyx_t_12 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_CannotPromote), __pyx_t_10, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(22, 206, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(22, 206, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_12);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_12)) __PYX_ERR(22, 206, __pyx_L7_except_error);
        __pyx_t_12 = 0;
        __pyx_t_12 = PyNumber_Add(__pyx_t_10, __pyx_v_sub_adapted->value_adapters); if (unlikely(!__pyx_t_12)) __PYX_ERR(22, 206, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GIVEREF(__pyx_t_12);
        __Pyx_GOTREF(__pyx_v_sub_adapted->value_adapters);
        __Pyx_DECREF(__pyx_v_sub_adapted->value_adapters);
        __pyx_v_sub_adapted->value_adapters = ((PyObject*)__pyx_t_12);
        __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;

      /* "src/union.pxi":202
 *         new_sub_types = []
 *         for sub_type in self.union_types:
 *             try:             # <<<<<<<<<<<<<<
 *                 sub_adapted = reader.for_writer(sub_type, False)
 *             except CannotPromoteError:
 */
      __pyx_L7_except_error:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L6_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      __pyx_L12_try_end:;
    }

    /* "src/union.pxi":209
 *             else:
 *                 got_one = True
 *             new_sub_types.append(sub_adapted)             # <<<<<<<<<<<<<<
 *         if not got_one:
 *             return
 */
    __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_new_sub_types, ((PyObject *)__pyx_v_sub_adapted)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(22, 209, __pyx_L1_error)

    /* "src/union.pxi":201
 *         cdef bint got_one = False
 *         new_sub_types = []
 *         for sub_type in self.union_types:             # <<<<<<<<<<<<<<
 *             try:
 *                 sub_adapted = reader.for_writer(sub_type, False)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/union.pxi":210
 *                 got_one = True
 *             new_sub_types.append(sub_adapted)
 *         if not got_one:             # <<<<<<<<<<<<<<
 *             return
 *         new_writer = self.clone_base()
 */
  __pyx_t_14 = (!__pyx_v_got_one);
  if (__pyx_t_14) {

    /* "src/union.pxi":211
 *             new_sub_types.append(sub_adapted)
 *         if not got_one:
 *             return             # <<<<<<<<<<<<<<
 *         new_writer = self.clone_base()
 *         new_writer.union_types = tuple(new_sub_types)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/union.pxi":210
 *                 got_one = True
 *             new_sub_types.append(sub_adapted)
 *         if not got_one:             # <<<<<<<<<<<<<<
 *             return
 *         new_writer = self.clone_base()
 */
  }

  /* "src/union.pxi":212
 *         if not got_one:
 *             return
 *         new_writer = self.clone_base()             # <<<<<<<<<<<<<<
 *         new_writer.union_types = tuple(new_sub_types)
 *         new_writer.return_type_tuple = self.return_type_tuple
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_UnionType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_UnionType))))) __PYX_ERR(22, 212, __pyx_L1_error)
  __pyx_v_new_writer = ((struct __pyx_obj_5cavro_UnionType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/union.pxi":213
 *             return
 *         new_writer = self.clone_base()
 *         new_writer.union_types = tuple(new_sub_types)             # <<<<<<<<<<<<<<
 *         new_writer.return_type_tuple = self.return_type_tuple
 *         new_writer._populate_name_map()
 */
  __pyx_t_1 = PyList_AsTuple(__pyx_v_new_sub_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_new_writer->union_types);
  __Pyx_DECREF(__pyx_v_new_writer->union_types);
  __pyx_v_new_writer->union_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/union.pxi":214
 *         new_writer = self.clone_base()
 *         new_writer.union_types = tuple(new_sub_types)
 *         new_writer.return_type_tuple = self.return_type_tuple             # <<<<<<<<<<<<<<
 *         new_writer._populate_name_map()
 *         return new_writer
 */
  __pyx_t_1 = __pyx_v_self->return_type_tuple;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_new_writer->return_type_tuple);
  __Pyx_DECREF(__pyx_v_new_writer->return_type_tuple);
  __pyx_v_new_writer->return_type_tuple = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/union.pxi":215
 *         new_writer.union_types = tuple(new_sub_types)
 *         new_writer.return_type_tuple = self.return_type_tuple
 *         new_writer._populate_name_map()             # <<<<<<<<<<<<<<
 *         return new_writer
 * 
 */
  __pyx_t_1 = __pyx_f_5cavro_9UnionType__populate_name_map(__pyx_v_new_writer); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/union.pxi":216
 *         new_writer.return_type_tuple = self.return_type_tuple
 *         new_writer._populate_name_map()
 *         return new_writer             # <<<<<<<<<<<<<<
 * 
 *     cdef bint accepts_missing_value(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_new_writer);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_new_writer);
  goto __pyx_L0;

  /* "src/union.pxi":196
 *         return writer_union.for_reader(self)
 * 
 *     cdef AvroType for_reader(self, AvroType reader):             # <<<<<<<<<<<<<<
 *         cdef AvroType sub_type
 *         cdef UnionType new_writer
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("cavro.UnionType.for_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sub_type);
  __Pyx_XDECREF((PyObject *)__pyx_v_new_writer);
  __Pyx_XDECREF(__pyx_v_new_sub_types);
  __Pyx_XDECREF((PyObject *)__pyx_v_sub_adapted);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":218
 *         return new_writer
 * 
 *     cdef bint accepts_missing_value(self):             # <<<<<<<<<<<<<<
 *         cdef AvroType union_type
 *         for union_type in self.union_types:
 */

static int __pyx_f_5cavro_9UnionType_accepts_missing_value(struct __pyx_obj_5cavro_UnionType *__pyx_v_self) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_union_type = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("accepts_missing_value", 1);

  /* "src/union.pxi":220
 *     cdef bint accepts_missing_value(self):
 *         cdef AvroType union_type
 *         for union_type in self.union_types:             # <<<<<<<<<<<<<<
 *             if union_type.accepts_missing_value():
 *                 return True
 */
  if (unlikely(__pyx_v_self->union_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 220, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->union_types; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 220, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(22, 220, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 220, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_union_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/union.pxi":221
 *         cdef AvroType union_type
 *         for union_type in self.union_types:
 *             if union_type.accepts_missing_value():             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_union_type->__pyx_vtab)->accepts_missing_value(__pyx_v_union_type); if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 221, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "src/union.pxi":222
 *         for union_type in self.union_types:
 *             if union_type.accepts_missing_value():
 *                 return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
      __pyx_r = 1;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "src/union.pxi":221
 *         cdef AvroType union_type
 *         for union_type in self.union_types:
 *             if union_type.accepts_missing_value():             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
 */
    }

    /* "src/union.pxi":220
 *     cdef bint accepts_missing_value(self):
 *         cdef AvroType union_type
 *         for union_type in self.union_types:             # <<<<<<<<<<<<<<
 *             if union_type.accepts_missing_value():
 *                 return True
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/union.pxi":223
 *             if union_type.accepts_missing_value():
 *                 return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "src/union.pxi":218
 *         return new_writer
 * 
 *     cdef bint accepts_missing_value(self):             # <<<<<<<<<<<<<<
 *         cdef AvroType union_type
 *         for union_type in self.union_types:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.UnionType.accepts_missing_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_union_type);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":225
 *         return False
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):             # <<<<<<<<<<<<<<
 *         cdef AvroType sub_type
 *         if schema_default is NO_DEFAULT:
 */

static PyObject *__pyx_f_5cavro_9UnionType_resolve_default_value(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v_schema_default, PyObject *__pyx_v_field) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_sub_type = 0;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_UCS4 __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  char const *__pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resolve_default_value", 1);

  /* "src/union.pxi":227
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         cdef AvroType sub_type
 *         if schema_default is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             for sub_type in self.union_types:
 *                 if sub_type.accepts_missing_value():
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_v_schema_default == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/union.pxi":228
 *         cdef AvroType sub_type
 *         if schema_default is NO_DEFAULT:
 *             for sub_type in self.union_types:             # <<<<<<<<<<<<<<
 *                 if sub_type.accepts_missing_value():
 *                     return MISSING_VALUE
 */
    if (unlikely(__pyx_v_self->union_types == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(22, 228, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_self->union_types; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 228, __pyx_L1_error)
        #endif
        if (__pyx_t_3 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(22, 228, __pyx_L1_error)
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 228, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 228, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_sub_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "src/union.pxi":229
 *         if schema_default is NO_DEFAULT:
 *             for sub_type in self.union_types:
 *                 if sub_type.accepts_missing_value():             # <<<<<<<<<<<<<<
 *                     return MISSING_VALUE
 *         sub_type = self.union_types[0]
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_sub_type->__pyx_vtab)->accepts_missing_value(__pyx_v_sub_type); if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 229, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "src/union.pxi":230
 *             for sub_type in self.union_types:
 *                 if sub_type.accepts_missing_value():
 *                     return MISSING_VALUE             # <<<<<<<<<<<<<<
 *         sub_type = self.union_types[0]
 *         try:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_MISSING_VALUE); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 230, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_r = __pyx_t_4;
        __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;

        /* "src/union.pxi":229
 *         if schema_default is NO_DEFAULT:
 *             for sub_type in self.union_types:
 *                 if sub_type.accepts_missing_value():             # <<<<<<<<<<<<<<
 *                     return MISSING_VALUE
 *         sub_type = self.union_types[0]
 */
      }

      /* "src/union.pxi":228
 *         cdef AvroType sub_type
 *         if schema_default is NO_DEFAULT:
 *             for sub_type in self.union_types:             # <<<<<<<<<<<<<<
 *                 if sub_type.accepts_missing_value():
 *                     return MISSING_VALUE
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/union.pxi":227
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         cdef AvroType sub_type
 *         if schema_default is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             for sub_type in self.union_types:
 *                 if sub_type.accepts_missing_value():
 */
  }

  /* "src/union.pxi":231
 *                 if sub_type.accepts_missing_value():
 *                     return MISSING_VALUE
 *         sub_type = self.union_types[0]             # <<<<<<<<<<<<<<
 *         try:
 *             return sub_type.resolve_default_value(schema_default, field)
 */
  if (unlikely(__pyx_v_self->union_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(22, 231, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_self->union_types, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 231, __pyx_L1_error)
  __Pyx_XDECREF_SET(__pyx_v_sub_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "src/union.pxi":232
 *                     return MISSING_VALUE
 *         sub_type = self.union_types[0]
 *         try:             # <<<<<<<<<<<<<<
 *             return sub_type.resolve_default_value(schema_default, field)
 *         except TypeError as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "src/union.pxi":233
 *         sub_type = self.union_types[0]
 *         try:
 *             return sub_type.resolve_default_value(schema_default, field)             # <<<<<<<<<<<<<<
 *         except TypeError as e:
 *             if self.options.allow_union_default_any_member:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_sub_type->__pyx_vtab)->resolve_default_value(__pyx_v_sub_type, __pyx_v_schema_default, __pyx_v_field); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 233, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L12_try_return;

      /* "src/union.pxi":232
 *                     return MISSING_VALUE
 *         sub_type = self.union_types[0]
 *         try:             # <<<<<<<<<<<<<<
 *             return sub_type.resolve_default_value(schema_default, field)
 *         except TypeError as e:
 */
    }
    __pyx_L8_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "src/union.pxi":234
 *         try:
 *             return sub_type.resolve_default_value(schema_default, field)
 *         except TypeError as e:             # <<<<<<<<<<<<<<
 *             if self.options.allow_union_default_any_member:
 *                 try:
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("cavro.UnionType.resolve_default_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_9) < 0) __PYX_ERR(22, 234, __pyx_L10_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_e = __pyx_t_4;
      /*try:*/ {

        /* "src/union.pxi":235
 *             return sub_type.resolve_default_value(schema_default, field)
 *         except TypeError as e:
 *             if self.options.allow_union_default_any_member:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.json_decode(schema_default)
 */
        if (__pyx_v_self->__pyx_base.options->allow_union_default_any_member) {

          /* "src/union.pxi":236
 *         except TypeError as e:
 *             if self.options.allow_union_default_any_member:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.json_decode(schema_default)
 *                 except:
 */
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_10);
            __Pyx_XGOTREF(__pyx_t_11);
            __Pyx_XGOTREF(__pyx_t_12);
            /*try:*/ {

              /* "src/union.pxi":237
 *             if self.options.allow_union_default_any_member:
 *                 try:
 *                     return self.json_decode(schema_default)             # <<<<<<<<<<<<<<
 *                 except:
 *                     pass
 */
              __Pyx_XDECREF(__pyx_r);
              __pyx_t_13 = ((struct __pyx_vtabstruct_5cavro_UnionType *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.json_decode(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_schema_default); if (unlikely(!__pyx_t_13)) __PYX_ERR(22, 237, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_13);
              __pyx_r = __pyx_t_13;
              __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              goto __pyx_L26_try_return;

              /* "src/union.pxi":236
 *         except TypeError as e:
 *             if self.options.allow_union_default_any_member:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.json_decode(schema_default)
 *                 except:
 */
            }
            __pyx_L22_error:;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;

            /* "src/union.pxi":238
 *                 try:
 *                     return self.json_decode(schema_default)
 *                 except:             # <<<<<<<<<<<<<<
 *                     pass
 *                 for sub_type in self.union_types[1:]:
 */
            /*except:*/ {
              __Pyx_ErrRestore(0,0,0);
              goto __pyx_L23_exception_handled;
            }

            /* "src/union.pxi":236
 *         except TypeError as e:
 *             if self.options.allow_union_default_any_member:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.json_decode(schema_default)
 *                 except:
 */
            __pyx_L26_try_return:;
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
            goto __pyx_L18_return;
            __pyx_L23_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
          }

          /* "src/union.pxi":240
 *                 except:
 *                     pass
 *                 for sub_type in self.union_types[1:]:             # <<<<<<<<<<<<<<
 *                     try:
 *                         return sub_type.resolve_default_value(schema_default, field)
 */
          if (unlikely(__pyx_v_self->union_types == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(22, 240, __pyx_L19_error)
          }
          __pyx_t_13 = __Pyx_PyTuple_GetSlice(__pyx_v_self->union_types, 1, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_13)) __PYX_ERR(22, 240, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_14 = __pyx_t_13; __Pyx_INCREF(__pyx_t_14);
          __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          for (;;) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_14);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(22, 240, __pyx_L19_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_13 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_3); __Pyx_INCREF(__pyx_t_13); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(22, 240, __pyx_L19_error)
            #else
            __pyx_t_13 = __Pyx_PySequence_ITEM(__pyx_t_14, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_13)) __PYX_ERR(22, 240, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_13);
            #endif
            if (!(likely(((__pyx_t_13) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_13, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(22, 240, __pyx_L19_error)
            __Pyx_DECREF_SET(__pyx_v_sub_type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_13));
            __pyx_t_13 = 0;

            /* "src/union.pxi":241
 *                     pass
 *                 for sub_type in self.union_types[1:]:
 *                     try:             # <<<<<<<<<<<<<<
 *                         return sub_type.resolve_default_value(schema_default, field)
 *                     except TypeError:
 */
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10);
              __Pyx_XGOTREF(__pyx_t_12);
              __Pyx_XGOTREF(__pyx_t_11);
              __Pyx_XGOTREF(__pyx_t_10);
              /*try:*/ {

                /* "src/union.pxi":242
 *                 for sub_type in self.union_types[1:]:
 *                     try:
 *                         return sub_type.resolve_default_value(schema_default, field)             # <<<<<<<<<<<<<<
 *                     except TypeError:
 *                         pass
 */
                __Pyx_XDECREF(__pyx_r);
                __pyx_t_13 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_sub_type->__pyx_vtab)->resolve_default_value(__pyx_v_sub_type, __pyx_v_schema_default, __pyx_v_field); if (unlikely(!__pyx_t_13)) __PYX_ERR(22, 242, __pyx_L32_error)
                __Pyx_GOTREF(__pyx_t_13);
                __pyx_r = __pyx_t_13;
                __pyx_t_13 = 0;
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                goto __pyx_L36_try_return;

                /* "src/union.pxi":241
 *                     pass
 *                 for sub_type in self.union_types[1:]:
 *                     try:             # <<<<<<<<<<<<<<
 *                         return sub_type.resolve_default_value(schema_default, field)
 *                     except TypeError:
 */
              }
              __pyx_L32_error:;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;

              /* "src/union.pxi":243
 *                     try:
 *                         return sub_type.resolve_default_value(schema_default, field)
 *                     except TypeError:             # <<<<<<<<<<<<<<
 *                         pass
 *             raise TypeError(f"Default value {schema_default!r} for field {field} is not valid for union with first type: {sub_type.type_name}") from e
 */
              __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
              if (__pyx_t_8) {
                __Pyx_ErrRestore(0,0,0);
                goto __pyx_L33_exception_handled;
              }
              goto __pyx_L34_except_error;

              /* "src/union.pxi":241
 *                     pass
 *                 for sub_type in self.union_types[1:]:
 *                     try:             # <<<<<<<<<<<<<<
 *                         return sub_type.resolve_default_value(schema_default, field)
 *                     except TypeError:
 */
              __pyx_L34_except_error:;
              __Pyx_XGIVEREF(__pyx_t_12);
              __Pyx_XGIVEREF(__pyx_t_11);
              __Pyx_XGIVEREF(__pyx_t_10);
              __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_11, __pyx_t_10);
              goto __pyx_L19_error;
              __pyx_L36_try_return:;
              __Pyx_XGIVEREF(__pyx_t_12);
              __Pyx_XGIVEREF(__pyx_t_11);
              __Pyx_XGIVEREF(__pyx_t_10);
              __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_11, __pyx_t_10);
              goto __pyx_L18_return;
              __pyx_L33_exception_handled:;
              __Pyx_XGIVEREF(__pyx_t_12);
              __Pyx_XGIVEREF(__pyx_t_11);
              __Pyx_XGIVEREF(__pyx_t_10);
              __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_11, __pyx_t_10);
            }

            /* "src/union.pxi":240
 *                 except:
 *                     pass
 *                 for sub_type in self.union_types[1:]:             # <<<<<<<<<<<<<<
 *                     try:
 *                         return sub_type.resolve_default_value(schema_default, field)
 */
          }
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

          /* "src/union.pxi":235
 *             return sub_type.resolve_default_value(schema_default, field)
 *         except TypeError as e:
 *             if self.options.allow_union_default_any_member:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.json_decode(schema_default)
 */
        }

        /* "src/union.pxi":245
 *                     except TypeError:
 *                         pass
 *             raise TypeError(f"Default value {schema_default!r} for field {field} is not valid for union with first type: {sub_type.type_name}") from e             # <<<<<<<<<<<<<<
 */
        __pyx_t_14 = PyTuple_New(6); if (unlikely(!__pyx_t_14)) __PYX_ERR(22, 245, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_3 = 0;
        __pyx_t_15 = 127;
        __Pyx_INCREF(__pyx_kp_u_Default_value_2);
        __pyx_t_3 += 14;
        __Pyx_GIVEREF(__pyx_kp_u_Default_value_2);
        PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_kp_u_Default_value_2);
        __pyx_t_13 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_schema_default), __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(22, 245, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_15 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_15) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_15;
        __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_13);
        __pyx_t_13 = 0;
        __Pyx_INCREF(__pyx_kp_u_for_field);
        __pyx_t_3 += 11;
        __Pyx_GIVEREF(__pyx_kp_u_for_field);
        PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_kp_u_for_field);
        __pyx_t_13 = __Pyx_PyUnicode_Unicode(__pyx_v_field); if (unlikely(!__pyx_t_13)) __PYX_ERR(22, 245, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_15 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_15) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_15;
        __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_14, 3, __pyx_t_13);
        __pyx_t_13 = 0;
        __Pyx_INCREF(__pyx_kp_u_is_not_valid_for_union_with_fir);
        __pyx_t_3 += 41;
        __Pyx_GIVEREF(__pyx_kp_u_is_not_valid_for_union_with_fir);
        PyTuple_SET_ITEM(__pyx_t_14, 4, __pyx_kp_u_is_not_valid_for_union_with_fir);
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_sub_type), __pyx_n_s_type_name); if (unlikely(!__pyx_t_13)) __PYX_ERR(22, 245, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_16 = __Pyx_PyObject_FormatSimple(__pyx_t_13, __pyx_empty_unicode); if (unlikely(!__pyx_t_16)) __PYX_ERR(22, 245, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_15 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16) > __pyx_t_15) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16) : __pyx_t_15;
        __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_16);
        __Pyx_GIVEREF(__pyx_t_16);
        PyTuple_SET_ITEM(__pyx_t_14, 5, __pyx_t_16);
        __pyx_t_16 = 0;
        __pyx_t_16 = __Pyx_PyUnicode_Join(__pyx_t_14, 6, __pyx_t_3, __pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(22, 245, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_16); if (unlikely(!__pyx_t_14)) __PYX_ERR(22, 245, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_Raise(__pyx_t_14, 0, 0, __pyx_v_e);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __PYX_ERR(22, 245, __pyx_L19_error)
      }

      /* "src/union.pxi":234
 *         try:
 *             return sub_type.resolve_default_value(schema_default, field)
 *         except TypeError as e:             # <<<<<<<<<<<<<<
 *             if self.options.allow_union_default_any_member:
 *                 try:
 */
      /*finally:*/ {
        __pyx_L19_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __pyx_t_8 = __pyx_lineno; __pyx_t_17 = __pyx_clineno; __pyx_t_18 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
          }
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_ErrRestore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
          __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_17; __pyx_filename = __pyx_t_18;
          goto __pyx_L10_except_error;
        }
        __pyx_L18_return: {
          __pyx_t_21 = __pyx_r;
          __pyx_r = 0;
          __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          __pyx_r = __pyx_t_21;
          __pyx_t_21 = 0;
          goto __pyx_L11_except_return;
        }
      }
    }
    goto __pyx_L10_except_error;

    /* "src/union.pxi":232
 *                     return MISSING_VALUE
 *         sub_type = self.union_types[0]
 *         try:             # <<<<<<<<<<<<<<
 *             return sub_type.resolve_default_value(schema_default, field)
 *         except TypeError as e:
 */
    __pyx_L10_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L12_try_return:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L0;
    __pyx_L11_except_return:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L0;
  }

  /* "src/union.pxi":225
 *         return False
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):             # <<<<<<<<<<<<<<
 *         cdef AvroType sub_type
 *         if schema_default is NO_DEFAULT:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("cavro.UnionType.resolve_default_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sub_type);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":29
 *     type_name = "union"
 * 
 *     cdef readonly tuple union_types             # <<<<<<<<<<<<<<
 *     cdef readonly dict by_name_map
 *     cdef readonly tuple return_type_tuple
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9UnionType_11union_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_9UnionType_11union_types_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_9UnionType_11union_types___get__(((struct __pyx_obj_5cavro_UnionType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9UnionType_11union_types___get__(struct __pyx_obj_5cavro_UnionType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->union_types);
  __pyx_r = __pyx_v_self->union_types;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":30
 * 
 *     cdef readonly tuple union_types
 *     cdef readonly dict by_name_map             # <<<<<<<<<<<<<<
 *     cdef readonly tuple return_type_tuple
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9UnionType_11by_name_map_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_9UnionType_11by_name_map_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_9UnionType_11by_name_map___get__(((struct __pyx_obj_5cavro_UnionType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9UnionType_11by_name_map___get__(struct __pyx_obj_5cavro_UnionType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->by_name_map);
  __pyx_r = __pyx_v_self->by_name_map;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/union.pxi":31
 *     cdef readonly tuple union_types
 *     cdef readonly dict by_name_map
 *     cdef readonly tuple return_type_tuple             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, schema, source, namespace):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9UnionType_17return_type_tuple_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_9UnionType_17return_type_tuple_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_9UnionType_17return_type_tuple___get__(((struct __pyx_obj_5cavro_UnionType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9UnionType_17return_type_tuple___get__(struct __pyx_obj_5cavro_UnionType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->return_type_tuple);
  __pyx_r = __pyx_v_self->return_type_tuple;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9UnionType_12__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9UnionType_12__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9UnionType_12__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9UnionType_12__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_9UnionType_11__reduce_cython__(((struct __pyx_obj_5cavro_UnionType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9UnionType_11__reduce_cython__(struct __pyx_obj_5cavro_UnionType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.by_name_map, self.metadata, self.options, self.return_type_tuple, self.union_types, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->by_name_map);
  __Pyx_GIVEREF(__pyx_v_self->by_name_map);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->by_name_map)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->return_type_tuple);
  __Pyx_GIVEREF(__pyx_v_self->return_type_tuple);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->return_type_tuple)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->union_types);
  __Pyx_GIVEREF(__pyx_v_self->union_types);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_self->union_types)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.by_name_map, self.metadata, self.options, self.return_type_tuple, self.union_types, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.by_name_map, self.metadata, self.options, self.return_type_tuple, self.union_types, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.by_name_map is not None or self.metadata is not None or self.options is not None or self.return_type_tuple is not None or self.union_types is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.by_name_map, self.metadata, self.options, self.return_type_tuple, self.union_types, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.by_name_map is not None or self.metadata is not None or self.options is not None or self.return_type_tuple is not None or self.union_types is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_UnionType, (type(self), 0xabef18e, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->by_name_map != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->return_type_tuple != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->union_types != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.by_name_map is not None or self.metadata is not None or self.options is not None or self.return_type_tuple is not None or self.union_types is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UnionType, (type(self), 0xabef18e, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.by_name_map is not None or self.metadata is not None or self.options is not None or self.return_type_tuple is not None or self.union_types is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_UnionType, (type(self), 0xabef18e, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_UnionType, (type(self), 0xabef18e, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_UnionType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_180285838);
    __Pyx_GIVEREF(__pyx_int_180285838);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_180285838)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.by_name_map is not None or self.metadata is not None or self.options is not None or self.return_type_tuple is not None or self.union_types is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UnionType, (type(self), 0xabef18e, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_UnionType, (type(self), 0xabef18e, None), state
 *     else:
 *         return __pyx_unpickle_UnionType, (type(self), 0xabef18e, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UnionType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_UnionType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_180285838);
    __Pyx_GIVEREF(__pyx_int_180285838);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_180285838)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.UnionType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UnionType, (type(self), 0xabef18e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UnionType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9UnionType_14__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9UnionType_14__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9UnionType_14__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9UnionType_14__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UnionType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9UnionType_13__setstate_cython__(((struct __pyx_obj_5cavro_UnionType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9UnionType_13__setstate_cython__(struct __pyx_obj_5cavro_UnionType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_UnionType, (type(self), 0xabef18e, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UnionType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_UnionType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UnionType, (type(self), 0xabef18e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UnionType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.UnionType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":13
 *     cdef readonly str name
 * 
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 * 
 */

/* Python wrapper */
static int __pyx_pw_5cavro_9_Sentinel_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_9_Sentinel_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 13, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(2, 13, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 13, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._Sentinel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9_Sentinel___init__(((struct __pyx_obj_5cavro__Sentinel *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_9_Sentinel___init__(struct __pyx_obj_5cavro__Sentinel *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/type.pxi":14
 * 
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_name))||((__pyx_v_name) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_name))) __PYX_ERR(2, 14, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_name;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":13
 *     cdef readonly str name
 * 
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._Sentinel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":16
 *         self.name = name
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'<{self.name}: {id(self)}>'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9_Sentinel_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_9_Sentinel_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_9_Sentinel_2__repr__(((struct __pyx_obj_5cavro__Sentinel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9_Sentinel_2__repr__(struct __pyx_obj_5cavro__Sentinel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "src/type.pxi":17
 * 
 *     def __repr__(self):
 *         return f'<{self.name}: {id(self)}>'             # <<<<<<<<<<<<<<
 * 
 * MISSING_VALUE = _Sentinel('MISSING_VALUE')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u__79);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__79);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u__79);
  __pyx_t_4 = __Pyx_PyUnicode_Unicode(__pyx_v_self->name); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_);
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__64);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__64);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__64);
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "src/type.pxi":16
 *         self.name = name
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return f'<{self.name}: {id(self)}>'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro._Sentinel.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":11
 * 
 * cdef class _Sentinel:
 *     cdef readonly str name             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, name):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9_Sentinel_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_9_Sentinel_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_9_Sentinel_4name___get__(((struct __pyx_obj_5cavro__Sentinel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9_Sentinel_4name___get__(struct __pyx_obj_5cavro__Sentinel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9_Sentinel_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9_Sentinel_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9_Sentinel_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9_Sentinel_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_9_Sentinel_4__reduce_cython__(((struct __pyx_obj_5cavro__Sentinel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9_Sentinel_4__reduce_cython__(struct __pyx_obj_5cavro__Sentinel *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__Sentinel, (type(self), 0x82a3537, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->name != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__Sentinel, (type(self), 0x82a3537, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle__Sentinel, (type(self), 0x82a3537, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__Sentinel, (type(self), 0x82a3537, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle__Sentinel); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_136983863);
    __Pyx_GIVEREF(__pyx_int_136983863);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_136983863)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__Sentinel, (type(self), 0x82a3537, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__Sentinel, (type(self), 0x82a3537, None), state
 *     else:
 *         return __pyx_unpickle__Sentinel, (type(self), 0x82a3537, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__Sentinel__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle__Sentinel); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_136983863);
    __Pyx_GIVEREF(__pyx_int_136983863);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_136983863)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._Sentinel.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Sentinel, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Sentinel__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9_Sentinel_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9_Sentinel_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9_Sentinel_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9_Sentinel_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._Sentinel.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9_Sentinel_6__setstate_cython__(((struct __pyx_obj_5cavro__Sentinel *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9_Sentinel_6__setstate_cython__(struct __pyx_obj_5cavro__Sentinel *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__Sentinel, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__Sentinel__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__Sentinel__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Sentinel, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Sentinel__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._Sentinel.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14_CanonicalForm_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_14_CanonicalForm_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14_CanonicalForm_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_14_CanonicalForm_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_14_CanonicalForm___reduce_cython__(((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14_CanonicalForm___reduce_cython__(struct __pyx_obj_5cavro__CanonicalForm *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__CanonicalForm, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__CanonicalForm, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle__CanonicalForm, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__CanonicalForm, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle__CanonicalForm); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__CanonicalForm, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__CanonicalForm, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle__CanonicalForm, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__CanonicalForm__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle__CanonicalForm); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._CanonicalForm.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__CanonicalForm, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__CanonicalForm__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14_CanonicalForm_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_14_CanonicalForm_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14_CanonicalForm_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_14_CanonicalForm_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._CanonicalForm.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_14_CanonicalForm_2__setstate_cython__(((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14_CanonicalForm_2__setstate_cython__(struct __pyx_obj_5cavro__CanonicalForm *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__CanonicalForm, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__CanonicalForm__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__CanonicalForm__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__CanonicalForm, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__CanonicalForm__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._CanonicalForm.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_17dict_to_canonical_2generator18(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/type.pxi":30
 *             if key in data:
 *                 pairs.append((f'"{key}"', dict_to_canonical(data[key])))
 *         pair_str = ','.join(':'.join(p) for p in pairs)             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('{' + pair_str + '}')
 *     elif isinstance(data, (tuple, list)):
 */

static PyObject *__pyx_pf_5cavro_17dict_to_canonical_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_20_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_20_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(2, 30, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_17dict_to_canonical_2generator18, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_dict_to_canonical_locals_genexpr, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(2, 30, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.dict_to_canonical.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_17dict_to_canonical_2generator18(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(2, 30, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(2, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(2, 30, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 30, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(2, 30, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_p);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_p, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = PyUnicode_Join(__pyx_kp_u__80, __pyx_cur_scope->__pyx_v_p); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_3))) __PYX_ERR(2, 30, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":24
 *     pass
 * 
 * cdef _CanonicalForm dict_to_canonical(data):             # <<<<<<<<<<<<<<
 *     if isinstance(data, dict):
 *         pairs=[]
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_dict_to_canonical(PyObject *__pyx_v_data) {
  PyObject *__pyx_v_pairs = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_pair_str = NULL;
  PyObject *__pyx_v_values = NULL;
  PyObject *__pyx_gb_5cavro_17dict_to_canonical_2generator18 = 0;
  PyObject *__pyx_9genexpr24__pyx_v_v = NULL;
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dict_to_canonical", 1);

  /* "src/type.pxi":25
 * 
 * cdef _CanonicalForm dict_to_canonical(data):
 *     if isinstance(data, dict):             # <<<<<<<<<<<<<<
 *         pairs=[]
 *         for key in CANONICAL_FORM_KEYS:
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_data); 
  if (__pyx_t_1) {

    /* "src/type.pxi":26
 * cdef _CanonicalForm dict_to_canonical(data):
 *     if isinstance(data, dict):
 *         pairs=[]             # <<<<<<<<<<<<<<
 *         for key in CANONICAL_FORM_KEYS:
 *             if key in data:
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 26, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_pairs = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/type.pxi":27
 *     if isinstance(data, dict):
 *         pairs=[]
 *         for key in CANONICAL_FORM_KEYS:             # <<<<<<<<<<<<<<
 *             if key in data:
 *                 pairs.append((f'"{key}"', dict_to_canonical(data[key])))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CANONICAL_FORM_KEYS); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 27, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 27, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 27, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 27, __pyx_L1_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(2, 27, __pyx_L1_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 27, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 27, __pyx_L1_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(2, 27, __pyx_L1_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 27, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(2, 27, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "src/type.pxi":28
 *         pairs=[]
 *         for key in CANONICAL_FORM_KEYS:
 *             if key in data:             # <<<<<<<<<<<<<<
 *                 pairs.append((f'"{key}"', dict_to_canonical(data[key])))
 *         pair_str = ','.join(':'.join(p) for p in pairs)
 */
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_key, __pyx_v_data, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(2, 28, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "src/type.pxi":29
 *         for key in CANONICAL_FORM_KEYS:
 *             if key in data:
 *                 pairs.append((f'"{key}"', dict_to_canonical(data[key])))             # <<<<<<<<<<<<<<
 *         pair_str = ','.join(':'.join(p) for p in pairs)
 *         return _CanonicalForm('{' + pair_str + '}')
 */
        __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 29, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = 0;
        __pyx_t_7 = 127;
        __Pyx_INCREF(__pyx_kp_u__48);
        __pyx_t_6 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__48);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u__48);
        __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_v_key, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 29, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_7;
        __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_8);
        __pyx_t_8 = 0;
        __Pyx_INCREF(__pyx_kp_u__48);
        __pyx_t_6 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__48);
        PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u__48);
        __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 29, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_data, __pyx_v_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 29, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_9 = ((PyObject *)__pyx_f_5cavro_dict_to_canonical(__pyx_t_2)); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 29, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 29, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_8);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_8)) __PYX_ERR(2, 29, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_9);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_9)) __PYX_ERR(2, 29, __pyx_L1_error);
        __pyx_t_8 = 0;
        __pyx_t_9 = 0;
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_pairs, __pyx_t_2); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(2, 29, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "src/type.pxi":28
 *         pairs=[]
 *         for key in CANONICAL_FORM_KEYS:
 *             if key in data:             # <<<<<<<<<<<<<<
 *                 pairs.append((f'"{key}"', dict_to_canonical(data[key])))
 *         pair_str = ','.join(':'.join(p) for p in pairs)
 */
      }

      /* "src/type.pxi":27
 *     if isinstance(data, dict):
 *         pairs=[]
 *         for key in CANONICAL_FORM_KEYS:             # <<<<<<<<<<<<<<
 *             if key in data:
 *                 pairs.append((f'"{key}"', dict_to_canonical(data[key])))
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "src/type.pxi":30
 *             if key in data:
 *                 pairs.append((f'"{key}"', dict_to_canonical(data[key])))
 *         pair_str = ','.join(':'.join(p) for p in pairs)             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('{' + pair_str + '}')
 *     elif isinstance(data, (tuple, list)):
 */
    __pyx_t_3 = __pyx_pf_5cavro_17dict_to_canonical_genexpr(NULL, __pyx_v_pairs); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_Generator_Next(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyUnicode_Join(__pyx_kp_u__77, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_pair_str = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/type.pxi":31
 *                 pairs.append((f'"{key}"', dict_to_canonical(data[key])))
 *         pair_str = ','.join(':'.join(p) for p in pairs)
 *         return _CanonicalForm('{' + pair_str + '}')             # <<<<<<<<<<<<<<
 *     elif isinstance(data, (tuple, list)):
 *         values = ','.join([dict_to_canonical(v) for v in data])
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_kp_u__61, __pyx_v_pair_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 31, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_3, __pyx_kp_u__62); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 31, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 31, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "src/type.pxi":25
 * 
 * cdef _CanonicalForm dict_to_canonical(data):
 *     if isinstance(data, dict):             # <<<<<<<<<<<<<<
 *         pairs=[]
 *         for key in CANONICAL_FORM_KEYS:
 */
  }

  /* "src/type.pxi":32
 *         pair_str = ','.join(':'.join(p) for p in pairs)
 *         return _CanonicalForm('{' + pair_str + '}')
 *     elif isinstance(data, (tuple, list)):             # <<<<<<<<<<<<<<
 *         values = ','.join([dict_to_canonical(v) for v in data])
 *         return _CanonicalForm('[' + values + ']')
 */
  __pyx_t_11 = PyTuple_Check(__pyx_v_data); 
  if (!__pyx_t_11) {
  } else {
    __pyx_t_1 = __pyx_t_11;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_11 = PyList_Check(__pyx_v_data); 
  __pyx_t_1 = __pyx_t_11;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/type.pxi":33
 *         return _CanonicalForm('{' + pair_str + '}')
 *     elif isinstance(data, (tuple, list)):
 *         values = ','.join([dict_to_canonical(v) for v in data])             # <<<<<<<<<<<<<<
 *         return _CanonicalForm('[' + values + ']')
 *     elif isinstance(data, _CanonicalForm):
 */
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 33, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (likely(PyList_CheckExact(__pyx_v_data)) || PyTuple_CheckExact(__pyx_v_data)) {
        __pyx_t_2 = __pyx_v_data; __Pyx_INCREF(__pyx_t_2);
        __pyx_t_4 = 0;
        __pyx_t_5 = NULL;
      } else {
        __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_data); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 33, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 33, __pyx_L12_error)
      }
      for (;;) {
        if (likely(!__pyx_t_5)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 33, __pyx_L12_error)
              #endif
              if (__pyx_t_4 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_9 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_9); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(2, 33, __pyx_L12_error)
            #else
            __pyx_t_9 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 33, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_9);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 33, __pyx_L12_error)
              #endif
              if (__pyx_t_4 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_9); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(2, 33, __pyx_L12_error)
            #else
            __pyx_t_9 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 33, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_9);
            #endif
          }
        } else {
          __pyx_t_9 = __pyx_t_5(__pyx_t_2);
          if (unlikely(!__pyx_t_9)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(2, 33, __pyx_L12_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_9);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr24__pyx_v_v, __pyx_t_9);
        __pyx_t_9 = 0;
        __pyx_t_9 = ((PyObject *)__pyx_f_5cavro_dict_to_canonical(__pyx_9genexpr24__pyx_v_v)); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 33, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_9))) __PYX_ERR(2, 33, __pyx_L12_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_v); __pyx_9genexpr24__pyx_v_v = 0;
      goto __pyx_L16_exit_scope;
      __pyx_L12_error:;
      __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_v); __pyx_9genexpr24__pyx_v_v = 0;
      goto __pyx_L1_error;
      __pyx_L16_exit_scope:;
    } /* exit inner scope */
    __pyx_t_2 = PyUnicode_Join(__pyx_kp_u__77, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_values = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/type.pxi":34
 *     elif isinstance(data, (tuple, list)):
 *         values = ','.join([dict_to_canonical(v) for v in data])
 *         return _CanonicalForm('[' + values + ']')             # <<<<<<<<<<<<<<
 *     elif isinstance(data, _CanonicalForm):
 *         return data
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = __Pyx_PyUnicode_ConcatSafe(__pyx_kp_u__14, __pyx_v_values); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u__78); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "src/type.pxi":32
 *         pair_str = ','.join(':'.join(p) for p in pairs)
 *         return _CanonicalForm('{' + pair_str + '}')
 *     elif isinstance(data, (tuple, list)):             # <<<<<<<<<<<<<<
 *         values = ','.join([dict_to_canonical(v) for v in data])
 *         return _CanonicalForm('[' + values + ']')
 */
  }

  /* "src/type.pxi":35
 *         values = ','.join([dict_to_canonical(v) for v in data])
 *         return _CanonicalForm('[' + values + ']')
 *     elif isinstance(data, _CanonicalForm):             # <<<<<<<<<<<<<<
 *         return data
 *     else:
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_data, __pyx_ptype_5cavro__CanonicalForm); 
  if (__pyx_t_1) {

    /* "src/type.pxi":36
 *         return _CanonicalForm('[' + values + ']')
 *     elif isinstance(data, _CanonicalForm):
 *         return data             # <<<<<<<<<<<<<<
 *     else:
 *         return _CanonicalForm(json.dumps(data, ensure_ascii=False))
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    if (!(likely(((__pyx_v_data) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_data, __pyx_ptype_5cavro__CanonicalForm))))) __PYX_ERR(2, 36, __pyx_L1_error)
    __Pyx_INCREF(__pyx_v_data);
    __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_v_data);
    goto __pyx_L0;

    /* "src/type.pxi":35
 *         values = ','.join([dict_to_canonical(v) for v in data])
 *         return _CanonicalForm('[' + values + ']')
 *     elif isinstance(data, _CanonicalForm):             # <<<<<<<<<<<<<<
 *         return data
 *     else:
 */
  }

  /* "src/type.pxi":38
 *         return data
 *     else:
 *         return _CanonicalForm(json.dumps(data, ensure_ascii=False))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_json); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_dumps); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_data);
    __Pyx_GIVEREF(__pyx_v_data);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_data)) __PYX_ERR(2, 38, __pyx_L1_error);
    __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ensure_ascii, Py_False) < 0) __PYX_ERR(2, 38, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_r = ((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_t_9);
    __pyx_t_9 = 0;
    goto __pyx_L0;
  }

  /* "src/type.pxi":24
 *     pass
 * 
 * cdef _CanonicalForm dict_to_canonical(data):             # <<<<<<<<<<<<<<
 *     if isinstance(data, dict):
 *         pairs=[]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cavro.dict_to_canonical", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pairs);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_pair_str);
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XDECREF(__pyx_gb_5cavro_17dict_to_canonical_2generator18);
  __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_v);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":41
 * 
 * 
 * cdef dict _strip_keys(dict source, set keys):             # <<<<<<<<<<<<<<
 *     return {k: v for k, v in source.items() if k not in keys}
 * 
 */

static PyObject *__pyx_f_5cavro__strip_keys(PyObject *__pyx_v_source, PyObject *__pyx_v_keys) {
  PyObject *__pyx_9genexpr25__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr25__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_strip_keys", 1);

  /* "src/type.pxi":42
 * 
 * cdef dict _strip_keys(dict source, set keys):
 *     return {k: v for k, v in source.items() if k not in keys}             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 42, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_source == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(2, 42, __pyx_L5_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_source, 1, __pyx_n_s_items, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 42, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_8 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, &__pyx_t_7, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_8 == 0)) break;
      if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(2, 42, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_9genexpr25__pyx_v_k, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr25__pyx_v_v, __pyx_t_7);
      __pyx_t_7 = 0;
      if (unlikely(__pyx_v_keys == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(2, 42, __pyx_L5_error)
      }
      __pyx_t_9 = (__Pyx_PySet_ContainsTF(__pyx_9genexpr25__pyx_v_k, __pyx_v_keys, Py_NE)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(2, 42, __pyx_L5_error)
      if (__pyx_t_9) {
        if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_9genexpr25__pyx_v_k, (PyObject*)__pyx_9genexpr25__pyx_v_v))) __PYX_ERR(2, 42, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_k); __pyx_9genexpr25__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_v); __pyx_9genexpr25__pyx_v_v = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_k); __pyx_9genexpr25__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_v); __pyx_9genexpr25__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/type.pxi":41
 * 
 * 
 * cdef dict _strip_keys(dict source, set keys):             # <<<<<<<<<<<<<<
 *     return {k: v for k, v in source.items() if k not in keys}
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro._strip_keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":63
 *     cdef readonly tuple[ValueAdapter] value_adapters
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def for_source(cls, schema, source, namespace=None):
 *         cdef Schema schema_ob = schema
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_1for_source(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8AvroType_1for_source = {"for_source", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_1for_source, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8AvroType_1for_source(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_namespace = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_source (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_schema,&__pyx_n_s_source,&__pyx_n_s_namespace,0};

    /* "src/type.pxi":64
 * 
 *     @classmethod
 *     def for_source(cls, schema, source, namespace=None):             # <<<<<<<<<<<<<<
 *         cdef Schema schema_ob = schema
 *         if isinstance(source, (list, tuple)):
 */
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 63, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 63, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("for_source", 0, 2, 3, 1); __PYX_ERR(2, 63, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_namespace);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 63, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "for_source") < 0)) __PYX_ERR(2, 63, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_schema = values[0];
    __pyx_v_source = values[1];
    __pyx_v_namespace = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("for_source", 0, 2, 3, __pyx_nargs); __PYX_ERR(2, 63, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.AvroType.for_source", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8AvroType_for_source(((PyTypeObject*)__pyx_v_cls), __pyx_v_schema, __pyx_v_source, __pyx_v_namespace);

  /* "src/type.pxi":63
 *     cdef readonly tuple[ValueAdapter] value_adapters
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def for_source(cls, schema, source, namespace=None):
 *         cdef Schema schema_ob = schema
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_for_source(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_namespace) {
  struct __pyx_obj_5cavro_Schema *__pyx_v_schema_ob = 0;
  PyObject *__pyx_v_type_name = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_e = NULL;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_inst = NULL;
  PyObject *__pyx_v_namespaced = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  struct __pyx_opt_args_5cavro_6Schema_find_type __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("for_source", 0);
  __Pyx_INCREF(__pyx_v_source);

  /* "src/type.pxi":65
 *     @classmethod
 *     def for_source(cls, schema, source, namespace=None):
 *         cdef Schema schema_ob = schema             # <<<<<<<<<<<<<<
 *         if isinstance(source, (list, tuple)):
 *             return UnionType(schema, source, namespace)
 */
  if (!(likely(((__pyx_v_schema) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_schema, __pyx_ptype_5cavro_Schema))))) __PYX_ERR(2, 65, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_schema;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_schema_ob = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":66
 *     def for_source(cls, schema, source, namespace=None):
 *         cdef Schema schema_ob = schema
 *         if isinstance(source, (list, tuple)):             # <<<<<<<<<<<<<<
 *             return UnionType(schema, source, namespace)
 *         if isinstance(source, str):
 */
  __pyx_t_3 = PyList_Check(__pyx_v_source); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyTuple_Check(__pyx_v_source); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "src/type.pxi":67
 *         cdef Schema schema_ob = schema
 *         if isinstance(source, (list, tuple)):
 *             return UnionType(schema, source, namespace)             # <<<<<<<<<<<<<<
 *         if isinstance(source, str):
 *             source = {'type': source}
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_schema);
    __Pyx_GIVEREF(__pyx_v_schema);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_schema)) __PYX_ERR(2, 67, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_source);
    __Pyx_GIVEREF(__pyx_v_source);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_source)) __PYX_ERR(2, 67, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_namespace);
    __Pyx_GIVEREF(__pyx_v_namespace);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_namespace)) __PYX_ERR(2, 67, __pyx_L1_error);
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_UnionType), __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "src/type.pxi":66
 *     def for_source(cls, schema, source, namespace=None):
 *         cdef Schema schema_ob = schema
 *         if isinstance(source, (list, tuple)):             # <<<<<<<<<<<<<<
 *             return UnionType(schema, source, namespace)
 *         if isinstance(source, str):
 */
  }

  /* "src/type.pxi":68
 *         if isinstance(source, (list, tuple)):
 *             return UnionType(schema, source, namespace)
 *         if isinstance(source, str):             # <<<<<<<<<<<<<<
 *             source = {'type': source}
 *         try:
 */
  __pyx_t_2 = PyUnicode_Check(__pyx_v_source); 
  if (__pyx_t_2) {

    /* "src/type.pxi":69
 *             return UnionType(schema, source, namespace)
 *         if isinstance(source, str):
 *             source = {'type': source}             # <<<<<<<<<<<<<<
 *         try:
 *             type_name = source['type']
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_type, __pyx_v_source) < 0) __PYX_ERR(2, 69, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_source, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/type.pxi":68
 *         if isinstance(source, (list, tuple)):
 *             return UnionType(schema, source, namespace)
 *         if isinstance(source, str):             # <<<<<<<<<<<<<<
 *             source = {'type': source}
 *         try:
 */
  }

  /* "src/type.pxi":70
 *         if isinstance(source, str):
 *             source = {'type': source}
 *         try:             # <<<<<<<<<<<<<<
 *             type_name = source['type']
 *         except (TypeError, KeyError) as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "src/type.pxi":71
 *             source = {'type': source}
 *         try:
 *             type_name = source['type']             # <<<<<<<<<<<<<<
 *         except (TypeError, KeyError) as e:
 *             raise ValueError(f"Could not find key 'type' in schema definition: {repr(source)}")
 */
      __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_source, __pyx_n_u_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 71, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_type_name = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "src/type.pxi":70
 *         if isinstance(source, str):
 *             source = {'type': source}
 *         try:             # <<<<<<<<<<<<<<
 *             type_name = source['type']
 *         except (TypeError, KeyError) as e:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L12_try_end;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "src/type.pxi":72
 *         try:
 *             type_name = source['type']
 *         except (TypeError, KeyError) as e:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Could not find key 'type' in schema definition: {repr(source)}")
 *         inst = schema_ob.find_type(namespace, type_name, False)
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_TypeError, __pyx_builtin_KeyError);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("cavro.AvroType.for_source", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_9) < 0) __PYX_ERR(2, 72, __pyx_L9_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_v_e = __pyx_t_1;
      /*try:*/ {

        /* "src/type.pxi":73
 *             type_name = source['type']
 *         except (TypeError, KeyError) as e:
 *             raise ValueError(f"Could not find key 'type' in schema definition: {repr(source)}")             # <<<<<<<<<<<<<<
 *         inst = schema_ob.find_type(namespace, type_name, False)
 *         if inst is not None:
 */
        __pyx_t_10 = PyObject_Repr(__pyx_v_source); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 73, __pyx_L18_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_10, __pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 73, __pyx_L18_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Could_not_find_key_type_in_schem, __pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 73, __pyx_L18_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 73, __pyx_L18_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(2, 73, __pyx_L18_error)
      }

      /* "src/type.pxi":72
 *         try:
 *             type_name = source['type']
 *         except (TypeError, KeyError) as e:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Could not find key 'type' in schema definition: {repr(source)}")
 *         inst = schema_ob.find_type(namespace, type_name, False)
 */
      /*finally:*/ {
        __pyx_L18_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_8 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L9_except_error;
        }
      }
    }
    goto __pyx_L9_except_error;

    /* "src/type.pxi":70
 *         if isinstance(source, str):
 *             source = {'type': source}
 *         try:             # <<<<<<<<<<<<<<
 *             type_name = source['type']
 *         except (TypeError, KeyError) as e:
 */
    __pyx_L9_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L12_try_end:;
  }

  /* "src/type.pxi":74
 *         except (TypeError, KeyError) as e:
 *             raise ValueError(f"Could not find key 'type' in schema definition: {repr(source)}")
 *         inst = schema_ob.find_type(namespace, type_name, False)             # <<<<<<<<<<<<<<
 *         if inst is not None:
 *             return inst
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_namespace))||((__pyx_v_namespace) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_namespace))) __PYX_ERR(2, 74, __pyx_L1_error)
  if (!(likely(PyUnicode_CheckExact(__pyx_v_type_name))||((__pyx_v_type_name) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_type_name))) __PYX_ERR(2, 74, __pyx_L1_error)
  __pyx_t_20.__pyx_n = 1;
  __pyx_t_20._raise = 0;
  __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_Schema *)__pyx_v_schema_ob->__pyx_vtab)->find_type(__pyx_v_schema_ob, ((PyObject*)__pyx_v_namespace), ((PyObject*)__pyx_v_type_name), 0, &__pyx_t_20)); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v_inst = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "src/type.pxi":75
 *             raise ValueError(f"Could not find key 'type' in schema definition: {repr(source)}")
 *         inst = schema_ob.find_type(namespace, type_name, False)
 *         if inst is not None:             # <<<<<<<<<<<<<<
 *             return inst
 *         if type_name in TYPES_BY_NAME:
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_inst) != Py_None);
  if (__pyx_t_2) {

    /* "src/type.pxi":76
 *         inst = schema_ob.find_type(namespace, type_name, False)
 *         if inst is not None:
 *             return inst             # <<<<<<<<<<<<<<
 *         if type_name in TYPES_BY_NAME:
 *             return TYPES_BY_NAME[type_name](schema, source, namespace)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_inst);
    __pyx_r = ((PyObject *)__pyx_v_inst);
    goto __pyx_L0;

    /* "src/type.pxi":75
 *             raise ValueError(f"Could not find key 'type' in schema definition: {repr(source)}")
 *         inst = schema_ob.find_type(namespace, type_name, False)
 *         if inst is not None:             # <<<<<<<<<<<<<<
 *             return inst
 *         if type_name in TYPES_BY_NAME:
 */
  }

  /* "src/type.pxi":77
 *         if inst is not None:
 *             return inst
 *         if type_name in TYPES_BY_NAME:             # <<<<<<<<<<<<<<
 *             return TYPES_BY_NAME[type_name](schema, source, namespace)
 *         if type_name == 'error' and schema_ob.options.allow_error_type:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_TYPES_BY_NAME); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_type_name, __pyx_t_9, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__pyx_t_2) {

    /* "src/type.pxi":78
 *             return inst
 *         if type_name in TYPES_BY_NAME:
 *             return TYPES_BY_NAME[type_name](schema, source, namespace)             # <<<<<<<<<<<<<<
 *         if type_name == 'error' and schema_ob.options.allow_error_type:
 *             return RecordType(schema, source, namespace)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TYPES_BY_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_type_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    __pyx_t_12 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_12 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_1, __pyx_v_schema, __pyx_v_source, __pyx_v_namespace};
      __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_12, 3+__pyx_t_12);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 78, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_r = __pyx_t_9;
    __pyx_t_9 = 0;
    goto __pyx_L0;

    /* "src/type.pxi":77
 *         if inst is not None:
 *             return inst
 *         if type_name in TYPES_BY_NAME:             # <<<<<<<<<<<<<<
 *             return TYPES_BY_NAME[type_name](schema, source, namespace)
 *         if type_name == 'error' and schema_ob.options.allow_error_type:
 */
  }

  /* "src/type.pxi":79
 *         if type_name in TYPES_BY_NAME:
 *             return TYPES_BY_NAME[type_name](schema, source, namespace)
 *         if type_name == 'error' and schema_ob.options.allow_error_type:             # <<<<<<<<<<<<<<
 *             return RecordType(schema, source, namespace)
 *         namespaced = resolve_namespaced_name(namespace, type_name)
 */
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_type_name, __pyx_n_u_error, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(2, 79, __pyx_L1_error)
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L27_bool_binop_done;
  }
  __pyx_t_2 = __pyx_v_schema_ob->options->allow_error_type;
  __pyx_L27_bool_binop_done:;
  if (__pyx_t_2) {

    /* "src/type.pxi":80
 *             return TYPES_BY_NAME[type_name](schema, source, namespace)
 *         if type_name == 'error' and schema_ob.options.allow_error_type:
 *             return RecordType(schema, source, namespace)             # <<<<<<<<<<<<<<
 *         namespaced = resolve_namespaced_name(namespace, type_name)
 *         raise UnknownType(namespaced)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_v_schema);
    __Pyx_GIVEREF(__pyx_v_schema);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_schema)) __PYX_ERR(2, 80, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_source);
    __Pyx_GIVEREF(__pyx_v_source);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_source)) __PYX_ERR(2, 80, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_namespace);
    __Pyx_GIVEREF(__pyx_v_namespace);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_namespace)) __PYX_ERR(2, 80, __pyx_L1_error);
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_RecordType), __pyx_t_9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "src/type.pxi":79
 *         if type_name in TYPES_BY_NAME:
 *             return TYPES_BY_NAME[type_name](schema, source, namespace)
 *         if type_name == 'error' and schema_ob.options.allow_error_type:             # <<<<<<<<<<<<<<
 *             return RecordType(schema, source, namespace)
 *         namespaced = resolve_namespaced_name(namespace, type_name)
 */
  }

  /* "src/type.pxi":81
 *         if type_name == 'error' and schema_ob.options.allow_error_type:
 *             return RecordType(schema, source, namespace)
 *         namespaced = resolve_namespaced_name(namespace, type_name)             # <<<<<<<<<<<<<<
 *         raise UnknownType(namespaced)
 * 
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_namespace))||((__pyx_v_namespace) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_namespace))) __PYX_ERR(2, 81, __pyx_L1_error)
  if (!(likely(PyUnicode_CheckExact(__pyx_v_type_name))||((__pyx_v_type_name) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_type_name))) __PYX_ERR(2, 81, __pyx_L1_error)
  __pyx_t_4 = __pyx_f_5cavro_resolve_namespaced_name(((PyObject*)__pyx_v_namespace), ((PyObject*)__pyx_v_type_name)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_namespaced = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "src/type.pxi":82
 *             return RecordType(schema, source, namespace)
 *         namespaced = resolve_namespaced_name(namespace, type_name)
 *         raise UnknownType(namespaced)             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_UnknownType); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = NULL;
  __pyx_t_12 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
      __pyx_t_12 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_namespaced};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(2, 82, __pyx_L1_error)

  /* "src/type.pxi":63
 *     cdef readonly tuple[ValueAdapter] value_adapters
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def for_source(cls, schema, source, namespace=None):
 *         cdef Schema schema_ob = schema
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("cavro.AvroType.for_source", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_schema_ob);
  __Pyx_XDECREF(__pyx_v_type_name);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XDECREF(__pyx_v_namespaced);
  __Pyx_XDECREF(__pyx_v_source);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":84
 *         raise UnknownType(namespaced)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def for_schema(cls, schema):
 *         return AvroType.for_source(schema, schema.source)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_3for_schema(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8AvroType_3for_schema = {"for_schema", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_3for_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8AvroType_3for_schema(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_schema = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_schema (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_schema,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 84, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "for_schema") < 0)) __PYX_ERR(2, 84, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_schema = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("for_schema", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 84, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.AvroType.for_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8AvroType_2for_schema(((PyTypeObject*)__pyx_v_cls), __pyx_v_schema);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_2for_schema(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_schema) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("for_schema", 1);

  /* "src/type.pxi":86
 *     @classmethod
 *     def for_schema(cls, schema):
 *         return AvroType.for_source(schema, schema.source)             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, schema, source, namespace, value_adapters=tuple()):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_AvroType), __pyx_n_s_for_source); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_source); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_schema, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/type.pxi":84
 *         raise UnknownType(namespaced)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def for_schema(cls, schema):
 *         return AvroType.for_source(schema, schema.source)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.AvroType.for_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":88
 *         return AvroType.for_source(schema, schema.source)
 * 
 *     def __init__(self, schema, source, namespace, value_adapters=tuple()):             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter logical
 *         self.options = schema.options
 */

/* Python wrapper */
static int __pyx_pw_5cavro_8AvroType_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_8AvroType_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_v_source = 0;
  CYTHON_UNUSED PyObject *__pyx_v_namespace = 0;
  PyObject *__pyx_v_value_adapters = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_schema,&__pyx_n_s_source,&__pyx_n_s_namespace,&__pyx_n_s_value_adapters,0};
    values[3] = __Pyx_Arg_NewRef_VARARGS(__pyx_k__81);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 88, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 88, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 1); __PYX_ERR(2, 88, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_namespace)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 88, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 2); __PYX_ERR(2, 88, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value_adapters);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 88, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(2, 88, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_schema = values[0];
    __pyx_v_source = values[1];
    __pyx_v_namespace = values[2];
    __pyx_v_value_adapters = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, __pyx_nargs); __PYX_ERR(2, 88, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.AvroType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8AvroType_4__init__(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_schema, __pyx_v_source, __pyx_v_namespace, __pyx_v_value_adapters);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_8AvroType_4__init__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, CYTHON_UNUSED PyObject *__pyx_v_namespace, PyObject *__pyx_v_value_adapters) {
  struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_logical = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/type.pxi":90
 *     def __init__(self, schema, source, namespace, value_adapters=tuple()):
 *         cdef ValueAdapter logical
 *         self.options = schema.options             # <<<<<<<<<<<<<<
 *         self.metadata = PyDictProxy_New(self._extract_metadata(source))
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_options); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(2, 90, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->options);
  __Pyx_DECREF((PyObject *)__pyx_v_self->options);
  __pyx_v_self->options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":91
 *         cdef ValueAdapter logical
 *         self.options = schema.options
 *         self.metadata = PyDictProxy_New(self._extract_metadata(source))             # <<<<<<<<<<<<<<
 * 
 *         logical = self._make_logical(schema, source)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->_extract_metadata(__pyx_v_self, __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyDictProxy_New(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->metadata);
  __Pyx_DECREF(__pyx_v_self->metadata);
  __pyx_v_self->metadata = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "src/type.pxi":93
 *         self.metadata = PyDictProxy_New(self._extract_metadata(source))
 * 
 *         logical = self._make_logical(schema, source)             # <<<<<<<<<<<<<<
 *         if logical is not None:
 *             self.value_adapters = value_adapters + (logical,)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->_make_logical(__pyx_v_self, __pyx_v_schema, __pyx_v_source); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_ValueAdapter))))) __PYX_ERR(2, 93, __pyx_L1_error)
  __pyx_v_logical = ((struct __pyx_obj_5cavro_ValueAdapter *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/type.pxi":94
 * 
 *         logical = self._make_logical(schema, source)
 *         if logical is not None:             # <<<<<<<<<<<<<<
 *             self.value_adapters = value_adapters + (logical,)
 *         else:
 */
  __pyx_t_3 = (((PyObject *)__pyx_v_logical) != Py_None);
  if (__pyx_t_3) {

    /* "src/type.pxi":95
 *         logical = self._make_logical(schema, source)
 *         if logical is not None:
 *             self.value_adapters = value_adapters + (logical,)             # <<<<<<<<<<<<<<
 *         else:
 *             self.value_adapters = value_adapters
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF((PyObject *)__pyx_v_logical);
    __Pyx_GIVEREF((PyObject *)__pyx_v_logical);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_logical))) __PYX_ERR(2, 95, __pyx_L1_error);
    __pyx_t_1 = PyNumber_Add(__pyx_v_value_adapters, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(2, 95, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->value_adapters);
    __Pyx_DECREF(__pyx_v_self->value_adapters);
    __pyx_v_self->value_adapters = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/type.pxi":94
 * 
 *         logical = self._make_logical(schema, source)
 *         if logical is not None:             # <<<<<<<<<<<<<<
 *             self.value_adapters = value_adapters + (logical,)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "src/type.pxi":97
 *             self.value_adapters = value_adapters + (logical,)
 *         else:
 *             self.value_adapters = value_adapters             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType clone_base(self, cls=None):
 */
  /*else*/ {
    if (!(likely(PyTuple_CheckExact(__pyx_v_value_adapters))||((__pyx_v_value_adapters) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v_value_adapters))) __PYX_ERR(2, 97, __pyx_L1_error)
    __pyx_t_1 = __pyx_v_value_adapters;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->value_adapters);
    __Pyx_DECREF(__pyx_v_self->value_adapters);
    __pyx_v_self->value_adapters = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "src/type.pxi":88
 *         return AvroType.for_source(schema, schema.source)
 * 
 *     def __init__(self, schema, source, namespace, value_adapters=tuple()):             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter logical
 *         self.options = schema.options
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.AvroType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_logical);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":99
 *             self.value_adapters = value_adapters
 * 
 *     cdef AvroType clone_base(self, cls=None):             # <<<<<<<<<<<<<<
 *         if cls is None:
 *             cls = self.__class__
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8AvroType_clone_base(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, struct __pyx_opt_args_5cavro_8AvroType_clone_base *__pyx_optional_args) {
  PyObject *__pyx_v_cls = ((PyObject *)Py_None);
  struct __pyx_obj_5cavro_AvroType *__pyx_v_clone = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone_base", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_cls = __pyx_optional_args->cls;
    }
  }
  __Pyx_INCREF(__pyx_v_cls);

  /* "src/type.pxi":100
 * 
 *     cdef AvroType clone_base(self, cls=None):
 *         if cls is None:             # <<<<<<<<<<<<<<
 *             cls = self.__class__
 *         cdef AvroType clone = cls.__new__(cls)
 */
  __pyx_t_1 = (__pyx_v_cls == Py_None);
  if (__pyx_t_1) {

    /* "src/type.pxi":101
 *     cdef AvroType clone_base(self, cls=None):
 *         if cls is None:
 *             cls = self.__class__             # <<<<<<<<<<<<<<
 *         cdef AvroType clone = cls.__new__(cls)
 *         clone.options = self.options
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_cls, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/type.pxi":100
 * 
 *     cdef AvroType clone_base(self, cls=None):
 *         if cls is None:             # <<<<<<<<<<<<<<
 *             cls = self.__class__
 *         cdef AvroType clone = cls.__new__(cls)
 */
  }

  /* "src/type.pxi":102
 *         if cls is None:
 *             cls = self.__class__
 *         cdef AvroType clone = cls.__new__(cls)             # <<<<<<<<<<<<<<
 *         clone.options = self.options
 *         clone.metadata = self.metadata
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_cls};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(2, 102, __pyx_L1_error)
  __pyx_v_clone = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/type.pxi":103
 *             cls = self.__class__
 *         cdef AvroType clone = cls.__new__(cls)
 *         clone.options = self.options             # <<<<<<<<<<<<<<
 *         clone.metadata = self.metadata
 *         clone.value_adapters = self.value_adapters
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_self->options);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_clone->options);
  __Pyx_DECREF((PyObject *)__pyx_v_clone->options);
  __pyx_v_clone->options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/type.pxi":104
 *         cdef AvroType clone = cls.__new__(cls)
 *         clone.options = self.options
 *         clone.metadata = self.metadata             # <<<<<<<<<<<<<<
 *         clone.value_adapters = self.value_adapters
 *         return clone
 */
  __pyx_t_2 = __pyx_v_self->metadata;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_clone->metadata);
  __Pyx_DECREF(__pyx_v_clone->metadata);
  __pyx_v_clone->metadata = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "src/type.pxi":105
 *         clone.options = self.options
 *         clone.metadata = self.metadata
 *         clone.value_adapters = self.value_adapters             # <<<<<<<<<<<<<<
 *         return clone
 * 
 */
  __pyx_t_2 = __pyx_v_self->value_adapters;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_clone->value_adapters);
  __Pyx_DECREF(__pyx_v_clone->value_adapters);
  __pyx_v_clone->value_adapters = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/type.pxi":106
 *         clone.metadata = self.metadata
 *         clone.value_adapters = self.value_adapters
 *         return clone             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_clone);
  __pyx_r = __pyx_v_clone;
  goto __pyx_L0;

  /* "src/type.pxi":99
 *             self.value_adapters = value_adapters
 * 
 *     cdef AvroType clone_base(self, cls=None):             # <<<<<<<<<<<<<<
 *         if cls is None:
 *             cls = self.__class__
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.AvroType.clone_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_clone);
  __Pyx_XDECREF(__pyx_v_cls);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":108
 *         return clone
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement clone")
 */

static PyObject *__pyx_pw_5cavro_8AvroType_7copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8AvroType_copy(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_8AvroType_7copy)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 108, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(2, 108, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/type.pxi":110
 *     cpdef AvroType copy(self):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement clone")             # <<<<<<<<<<<<<<
 * 
 *     cdef _make_logical(self, schema, source):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/type.pxi":109
 * 
 *     cpdef AvroType copy(self):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement clone")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(2, 109, __pyx_L1_error)

  /* "src/type.pxi":108
 *         return clone
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement clone")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.AvroType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_7copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8AvroType_7copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_7copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8AvroType_7copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_8AvroType_6copy(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_6copy(struct __pyx_obj_5cavro_AvroType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_8AvroType_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.AvroType.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":112
 *             f"{type(self).__name__} does not implement clone")
 * 
 *     cdef _make_logical(self, schema, source):             # <<<<<<<<<<<<<<
 *         cdef str logical_type_name = source.get('logicalType')
 *         if logical_type_name is None:
 */

static PyObject *__pyx_f_5cavro_8AvroType__make_logical(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source) {
  PyObject *__pyx_v_logical_type_name = 0;
  PyObject *__pyx_v_logical_type_classes = NULL;
  PyObject *__pyx_v_cls = NULL;
  PyObject *__pyx_v_inst = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_logical", 1);

  /* "src/type.pxi":113
 * 
 *     cdef _make_logical(self, schema, source):
 *         cdef str logical_type_name = source.get('logicalType')             # <<<<<<<<<<<<<<
 *         if logical_type_name is None:
 *             return
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_u_logicalType};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(2, 113, __pyx_L1_error)
  __pyx_v_logical_type_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":114
 *     cdef _make_logical(self, schema, source):
 *         cdef str logical_type_name = source.get('logicalType')
 *         if logical_type_name is None:             # <<<<<<<<<<<<<<
 *             return
 *         logical_type_classes = schema.logical_types.get(logical_type_name, [])
 */
  __pyx_t_5 = (__pyx_v_logical_type_name == ((PyObject*)Py_None));
  if (__pyx_t_5) {

    /* "src/type.pxi":115
 *         cdef str logical_type_name = source.get('logicalType')
 *         if logical_type_name is None:
 *             return             # <<<<<<<<<<<<<<
 *         logical_type_classes = schema.logical_types.get(logical_type_name, [])
 *         for cls in logical_type_classes:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/type.pxi":114
 *     cdef _make_logical(self, schema, source):
 *         cdef str logical_type_name = source.get('logicalType')
 *         if logical_type_name is None:             # <<<<<<<<<<<<<<
 *             return
 *         logical_type_classes = schema.logical_types.get(logical_type_name, [])
 */
  }

  /* "src/type.pxi":116
 *         if logical_type_name is None:
 *             return
 *         logical_type_classes = schema.logical_types.get(logical_type_name, [])             # <<<<<<<<<<<<<<
 *         for cls in logical_type_classes:
 *             inst = cls.for_type(self)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_schema, __pyx_n_s_logical_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_logical_type_name, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_logical_type_classes = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/type.pxi":117
 *             return
 *         logical_type_classes = schema.logical_types.get(logical_type_name, [])
 *         for cls in logical_type_classes:             # <<<<<<<<<<<<<<
 *             inst = cls.for_type(self)
 *             if inst is not None:
 */
  if (likely(PyList_CheckExact(__pyx_v_logical_type_classes)) || PyTuple_CheckExact(__pyx_v_logical_type_classes)) {
    __pyx_t_1 = __pyx_v_logical_type_classes; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_logical_type_classes); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 117, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 117, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(2, 117, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 117, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 117, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(2, 117, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 117, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_1);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 117, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_cls, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/type.pxi":118
 *         logical_type_classes = schema.logical_types.get(logical_type_name, [])
 *         for cls in logical_type_classes:
 *             inst = cls.for_type(self)             # <<<<<<<<<<<<<<
 *             if inst is not None:
 *                 return inst
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_for_type_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, ((PyObject *)__pyx_v_self)};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_inst, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/type.pxi":119
 *         for cls in logical_type_classes:
 *             inst = cls.for_type(self)
 *             if inst is not None:             # <<<<<<<<<<<<<<
 *                 return inst
 * 
 */
    __pyx_t_5 = (__pyx_v_inst != Py_None);
    if (__pyx_t_5) {

      /* "src/type.pxi":120
 *             inst = cls.for_type(self)
 *             if inst is not None:
 *                 return inst             # <<<<<<<<<<<<<<
 * 
 *     property type:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_inst);
      __pyx_r = __pyx_v_inst;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "src/type.pxi":119
 *         for cls in logical_type_classes:
 *             inst = cls.for_type(self)
 *             if inst is not None:             # <<<<<<<<<<<<<<
 *                 return inst
 * 
 */
    }

    /* "src/type.pxi":117
 *             return
 *         logical_type_classes = schema.logical_types.get(logical_type_name, [])
 *         for cls in logical_type_classes:             # <<<<<<<<<<<<<<
 *             inst = cls.for_type(self)
 *             if inst is not None:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/type.pxi":112
 *             f"{type(self).__name__} does not implement clone")
 * 
 *     cdef _make_logical(self, schema, source):             # <<<<<<<<<<<<<<
 *         cdef str logical_type_name = source.get('logicalType')
 *         if logical_type_name is None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.AvroType._make_logical", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_logical_type_name);
  __Pyx_XDECREF(__pyx_v_logical_type_classes);
  __Pyx_XDECREF(__pyx_v_cls);
  __Pyx_XDECREF(__pyx_v_inst);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":123
 * 
 *     property type:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.type_name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_4type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_8AvroType_4type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_8AvroType_4type___get__(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_4type___get__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/type.pxi":124
 *     property type:
 *         def __get__(self):
 *             return self.type_name             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/type.pxi":123
 * 
 *     property type:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.type_name
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.AvroType.type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":126
 *             return self.type_name
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _extract_metadata")
 */

static PyObject *__pyx_f_5cavro_8AvroType__extract_metadata(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/type.pxi":128
 *     cdef dict _extract_metadata(self, source):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _extract_metadata")             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _convert_value(self, object value):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement__extract_met); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/type.pxi":127
 * 
 *     cdef dict _extract_metadata(self, source):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement _extract_metadata")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(2, 127, __pyx_L1_error)

  /* "src/type.pxi":126
 *             return self.type_name
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _extract_metadata")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.AvroType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":130
 *             f"{type(self).__name__} does not implement _extract_metadata")
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _convert_value")
 */

static PyObject *__pyx_pw_5cavro_8AvroType_9_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_8AvroType__convert_value(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_convert_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 130, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_8AvroType_9_convert_value)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_value};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 130, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/type.pxi":132
 *     cpdef object _convert_value(self, object value):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _convert_value")             # <<<<<<<<<<<<<<
 * 
 *     cpdef object convert_value(self, object value, check_value=False):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement__convert_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/type.pxi":131
 * 
 *     cpdef object _convert_value(self, object value):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement _convert_value")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(2, 131, __pyx_L1_error)

  /* "src/type.pxi":130
 *             f"{type(self).__name__} does not implement _extract_metadata")
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _convert_value")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.AvroType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_9_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8AvroType_9_convert_value = {"_convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_9_convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8AvroType_9_convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 130, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_convert_value") < 0)) __PYX_ERR(2, 130, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 130, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.AvroType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8AvroType_8_convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_8_convert_value(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8AvroType__convert_value(__pyx_v_self, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.AvroType._convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":134
 *             f"{type(self).__name__} does not implement _convert_value")
 * 
 *     cpdef object convert_value(self, object value, check_value=False):             # <<<<<<<<<<<<<<
 *         if check_value:
 *             self.assert_value(value)
 */

static PyObject *__pyx_pw_5cavro_8AvroType_11convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_8AvroType_convert_value(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_8AvroType_convert_value *__pyx_optional_args) {
  PyObject *__pyx_v_check_value = ((PyObject *)Py_False);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_value", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_check_value = __pyx_optional_args->check_value;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_convert_value_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 134, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_8AvroType_11convert_value)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_value, __pyx_v_check_value};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 134, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/type.pxi":135
 * 
 *     cpdef object convert_value(self, object value, check_value=False):
 *         if check_value:             # <<<<<<<<<<<<<<
 *             self.assert_value(value)
 *         return self._convert_value(value)
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_check_value); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(2, 135, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "src/type.pxi":136
 *     cpdef object convert_value(self, object value, check_value=False):
 *         if check_value:
 *             self.assert_value(value)             # <<<<<<<<<<<<<<
 *         return self._convert_value(value)
 * 
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->assert_value(__pyx_v_self, __pyx_v_value); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 136, __pyx_L1_error)

    /* "src/type.pxi":135
 * 
 *     cpdef object convert_value(self, object value, check_value=False):
 *         if check_value:             # <<<<<<<<<<<<<<
 *             self.assert_value(value)
 *         return self._convert_value(value)
 */
  }

  /* "src/type.pxi":137
 *         if check_value:
 *             self.assert_value(value)
 *         return self._convert_value(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef int binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->_convert_value(__pyx_v_self, __pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/type.pxi":134
 *             f"{type(self).__name__} does not implement _convert_value")
 * 
 *     cpdef object convert_value(self, object value, check_value=False):             # <<<<<<<<<<<<<<
 *         if check_value:
 *             self.assert_value(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.AvroType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_11convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8AvroType_11convert_value = {"convert_value", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_11convert_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8AvroType_11convert_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_check_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convert_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,&__pyx_n_s_check_value,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_False));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 134, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_check_value);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 134, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "convert_value") < 0)) __PYX_ERR(2, 134, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_value = values[0];
    __pyx_v_check_value = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convert_value", 0, 1, 2, __pyx_nargs); __PYX_ERR(2, 134, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.AvroType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8AvroType_10convert_value(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_value, __pyx_v_check_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_10convert_value(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_check_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_convert_value __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_value", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.check_value = __pyx_v_check_value;
  __pyx_t_1 = __pyx_vtabptr_5cavro_AvroType->convert_value(__pyx_v_self, __pyx_v_value, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.AvroType.convert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":139
 *         return self._convert_value(value)
 * 
 *     cdef int binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter adapter
 * 
 */

static int __pyx_f_5cavro_8AvroType_binary_buffer_encode(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_adapter = 0;
  PyObject *__pyx_v_type_name = NULL;
  PyObject *__pyx_v_inner_value = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_e = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("binary_buffer_encode", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/type.pxi":142
 *         cdef ValueAdapter adapter
 * 
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):             # <<<<<<<<<<<<<<
 *             type_name, inner_value = value
 *             if type_name == self.type:
 */
  if (__pyx_v_self->options->allow_tuple_notation) {
  } else {
    __pyx_t_1 = __pyx_v_self->options->allow_tuple_notation;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 142, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_3 == 2);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_5cavro_ArrayType); 
  __pyx_t_4 = (!__pyx_t_2);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/type.pxi":143
 * 
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):
 *             type_name, inner_value = value             # <<<<<<<<<<<<<<
 *             if type_name == self.type:
 *                 try:
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_value))) || (PyList_CheckExact(__pyx_v_value))) {
      PyObject* sequence = __pyx_v_value;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(2, 143, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(2, 143, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L9_unpacking_done;
      __pyx_L8_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(2, 143, __pyx_L1_error)
      __pyx_L9_unpacking_done:;
    }
    __pyx_v_type_name = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_inner_value = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "src/type.pxi":144
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):
 *             type_name, inner_value = value
 *             if type_name == self.type:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.binary_buffer_encode(buffer, inner_value)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_type_name, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 144, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(2, 144, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_1) {

      /* "src/type.pxi":145
 *             type_name, inner_value = value
 *             if type_name == self.type:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.binary_buffer_encode(buffer, inner_value)
 *                 except Exception as e:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        /*try:*/ {

          /* "src/type.pxi":146
 *             if type_name == self.type:
 *                 try:
 *                     return self.binary_buffer_encode(buffer, inner_value)             # <<<<<<<<<<<<<<
 *                 except Exception as e:
 *                     pass
 */
          __pyx_t_12 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->binary_buffer_encode(__pyx_v_self, __pyx_v_buffer, __pyx_v_inner_value); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(2, 146, __pyx_L11_error)
          __pyx_r = __pyx_t_12;
          goto __pyx_L15_try_return;

          /* "src/type.pxi":145
 *             type_name, inner_value = value
 *             if type_name == self.type:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.binary_buffer_encode(buffer, inner_value)
 *                 except Exception as e:
 */
        }
        __pyx_L11_error:;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "src/type.pxi":147
 *                 try:
 *                     return self.binary_buffer_encode(buffer, inner_value)
 *                 except Exception as e:             # <<<<<<<<<<<<<<
 *                     pass
 * 
 */
        __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
        if (__pyx_t_12) {
          __Pyx_AddTraceback("cavro.AvroType.binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 147, __pyx_L13_except_error)
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          __pyx_v_e = __pyx_t_6;
          /*try:*/ {
          }
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
              goto __pyx_L23;
            }
            __pyx_L23:;
          }
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          goto __pyx_L12_exception_handled;
        }
        goto __pyx_L13_except_error;

        /* "src/type.pxi":145
 *             type_name, inner_value = value
 *             if type_name == self.type:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.binary_buffer_encode(buffer, inner_value)
 *                 except Exception as e:
 */
        __pyx_L13_except_error:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L1_error;
        __pyx_L15_try_return:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L0;
        __pyx_L12_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }

      /* "src/type.pxi":144
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):
 *             type_name, inner_value = value
 *             if type_name == self.type:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.binary_buffer_encode(buffer, inner_value)
 */
    }

    /* "src/type.pxi":142
 *         cdef ValueAdapter adapter
 * 
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):             # <<<<<<<<<<<<<<
 *             type_name, inner_value = value
 *             if type_name == self.type:
 */
  }

  /* "src/type.pxi":150
 *                     pass
 * 
 *         for adapter in self.value_adapters:             # <<<<<<<<<<<<<<
 *             value = adapter.encode_value(value)
 *         return self._binary_buffer_encode(buffer, value)
 */
  if (unlikely(__pyx_v_self->value_adapters == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(2, 150, __pyx_L1_error)
  }
  __pyx_t_7 = __pyx_v_self->value_adapters; __Pyx_INCREF(__pyx_t_7);
  __pyx_t_3 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_7);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 150, __pyx_L1_error)
      #endif
      if (__pyx_t_3 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(2, 150, __pyx_L1_error)
    #else
    __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5cavro_ValueAdapter))))) __PYX_ERR(2, 150, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_adapter, ((struct __pyx_obj_5cavro_ValueAdapter *)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "src/type.pxi":151
 * 
 *         for adapter in self.value_adapters:
 *             value = adapter.encode_value(value)             # <<<<<<<<<<<<<<
 *         return self._binary_buffer_encode(buffer, value)
 * 
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_5cavro_ValueAdapter *)__pyx_v_adapter->__pyx_vtab)->encode_value(__pyx_v_adapter, __pyx_v_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "src/type.pxi":150
 *                     pass
 * 
 *         for adapter in self.value_adapters:             # <<<<<<<<<<<<<<
 *             value = adapter.encode_value(value)
 *         return self._binary_buffer_encode(buffer, value)
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/type.pxi":152
 *         for adapter in self.value_adapters:
 *             value = adapter.encode_value(value)
 *         return self._binary_buffer_encode(buffer, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 */
  __pyx_t_12 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->_binary_buffer_encode(__pyx_v_self, __pyx_v_buffer, __pyx_v_value); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(2, 152, __pyx_L1_error)
  __pyx_r = __pyx_t_12;
  goto __pyx_L0;

  /* "src/type.pxi":139
 *         return self._convert_value(value)
 * 
 *     cdef int binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter adapter
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.AvroType.binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_adapter);
  __Pyx_XDECREF(__pyx_v_type_name);
  __Pyx_XDECREF(__pyx_v_inner_value);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":154
 *         return self._binary_buffer_encode(buffer, value)
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _binary_buffer_encode")
 */

static int __pyx_f_5cavro_8AvroType__binary_buffer_encode(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Writer *__pyx_v_buffer, CYTHON_UNUSED PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_encode", 1);

  /* "src/type.pxi":156
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _binary_buffer_encode")             # <<<<<<<<<<<<<<
 * 
 *     cdef binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement__binary_buff); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/type.pxi":155
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement _binary_buffer_encode")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(2, 155, __pyx_L1_error)

  /* "src/type.pxi":154
 *         return self._binary_buffer_encode(buffer, value)
 * 
 *     cdef int _binary_buffer_encode(self, _Writer buffer, value) except -1:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _binary_buffer_encode")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.AvroType._binary_buffer_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":158
 *             f"{type(self).__name__} does not implement _binary_buffer_encode")
 * 
 *     cdef binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter adapter
 *         value = self._binary_buffer_decode(buffer)
 */

static PyObject *__pyx_f_5cavro_8AvroType_binary_buffer_decode(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_adapter = 0;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("binary_buffer_decode", 1);

  /* "src/type.pxi":160
 *     cdef binary_buffer_decode(self, _Reader buffer):
 *         cdef ValueAdapter adapter
 *         value = self._binary_buffer_decode(buffer)             # <<<<<<<<<<<<<<
 *         for adapter in reversed(self.value_adapters):
 *             value = adapter.decode_value(value)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->_binary_buffer_decode(__pyx_v_self, __pyx_v_buffer); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/type.pxi":161
 *         cdef ValueAdapter adapter
 *         value = self._binary_buffer_decode(buffer)
 *         for adapter in reversed(self.value_adapters):             # <<<<<<<<<<<<<<
 *             value = adapter.decode_value(value)
 *         return value
 */
  if (unlikely(__pyx_v_self->value_adapters == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(2, 161, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->value_adapters; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
  #if !CYTHON_ASSUME_SAFE_MACROS
  if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 161, __pyx_L1_error)
  #endif
  --__pyx_t_2;
  for (;;) {
    if (__pyx_t_2 < 0) break;
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 161, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2--; if (unlikely((0 < 0))) __PYX_ERR(2, 161, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2--; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_ValueAdapter))))) __PYX_ERR(2, 161, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_adapter, ((struct __pyx_obj_5cavro_ValueAdapter *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/type.pxi":162
 *         value = self._binary_buffer_decode(buffer)
 *         for adapter in reversed(self.value_adapters):
 *             value = adapter.decode_value(value)             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_ValueAdapter *)__pyx_v_adapter->__pyx_vtab)->decode_value(__pyx_v_adapter, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/type.pxi":161
 *         cdef ValueAdapter adapter
 *         value = self._binary_buffer_decode(buffer)
 *         for adapter in reversed(self.value_adapters):             # <<<<<<<<<<<<<<
 *             value = adapter.decode_value(value)
 *         return value
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/type.pxi":163
 *         for adapter in reversed(self.value_adapters):
 *             value = adapter.decode_value(value)
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/type.pxi":158
 *             f"{type(self).__name__} does not implement _binary_buffer_encode")
 * 
 *     cdef binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter adapter
 *         value = self._binary_buffer_decode(buffer)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.AvroType.binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_adapter);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":165
 *         return value
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement binary_buffer_decode")
 */

static PyObject *__pyx_f_5cavro_8AvroType__binary_buffer_decode(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Reader *__pyx_v_buffer) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_binary_buffer_decode", 1);

  /* "src/type.pxi":167
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement binary_buffer_decode")             # <<<<<<<<<<<<<<
 * 
 *     cdef int get_value_fitness(self, value) except -1:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement_binary_buffe); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/type.pxi":166
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement binary_buffer_decode")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(2, 166, __pyx_L1_error)

  /* "src/type.pxi":165
 *         return value
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement binary_buffer_decode")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.AvroType._binary_buffer_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":169
 *             f"{type(self).__name__} does not implement binary_buffer_decode")
 * 
 *     cdef int get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter adapter
 * 
 */

static int __pyx_f_5cavro_8AvroType_get_value_fitness(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_adapter = 0;
  PyObject *__pyx_v_type_name = NULL;
  PyObject *__pyx_v_inner_value = NULL;
  int __pyx_v_inner_fitness;
  CYTHON_UNUSED PyObject *__pyx_v_e = NULL;
  int __pyx_v_value_fit;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value_fitness", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/type.pxi":172
 *         cdef ValueAdapter adapter
 * 
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):             # <<<<<<<<<<<<<<
 *             type_name, inner_value = value
 *             if type_name == self.type:
 */
  if (__pyx_v_self->options->allow_tuple_notation) {
  } else {
    __pyx_t_1 = __pyx_v_self->options->allow_tuple_notation;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 172, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_3 == 2);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_5cavro_ArrayType); 
  __pyx_t_4 = (!__pyx_t_2);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/type.pxi":173
 * 
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):
 *             type_name, inner_value = value             # <<<<<<<<<<<<<<
 *             if type_name == self.type:
 *                 inner_fitness = self._get_value_fitness(inner_value)
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_value))) || (PyList_CheckExact(__pyx_v_value))) {
      PyObject* sequence = __pyx_v_value;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(2, 173, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(2, 173, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L9_unpacking_done;
      __pyx_L8_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(2, 173, __pyx_L1_error)
      __pyx_L9_unpacking_done:;
    }
    __pyx_v_type_name = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_inner_value = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "src/type.pxi":174
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):
 *             type_name, inner_value = value
 *             if type_name == self.type:             # <<<<<<<<<<<<<<
 *                 inner_fitness = self._get_value_fitness(inner_value)
 *                 return inner_fitness
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_type_name, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 174, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(2, 174, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_1) {

      /* "src/type.pxi":175
 *             type_name, inner_value = value
 *             if type_name == self.type:
 *                 inner_fitness = self._get_value_fitness(inner_value)             # <<<<<<<<<<<<<<
 *                 return inner_fitness
 * 
 */
      __pyx_t_9 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->_get_value_fitness(__pyx_v_self, __pyx_v_inner_value); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 175, __pyx_L1_error)
      __pyx_v_inner_fitness = __pyx_t_9;

      /* "src/type.pxi":176
 *             if type_name == self.type:
 *                 inner_fitness = self._get_value_fitness(inner_value)
 *                 return inner_fitness             # <<<<<<<<<<<<<<
 * 
 *         for adapter in self.value_adapters:
 */
      __pyx_r = __pyx_v_inner_fitness;
      goto __pyx_L0;

      /* "src/type.pxi":174
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):
 *             type_name, inner_value = value
 *             if type_name == self.type:             # <<<<<<<<<<<<<<
 *                 inner_fitness = self._get_value_fitness(inner_value)
 *                 return inner_fitness
 */
    }

    /* "src/type.pxi":172
 *         cdef ValueAdapter adapter
 * 
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):             # <<<<<<<<<<<<<<
 *             type_name, inner_value = value
 *             if type_name == self.type:
 */
  }

  /* "src/type.pxi":178
 *                 return inner_fitness
 * 
 *         for adapter in self.value_adapters:             # <<<<<<<<<<<<<<
 *             try:
 *                 value = adapter.encode_value(value)
 */
  if (unlikely(__pyx_v_self->value_adapters == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(2, 178, __pyx_L1_error)
  }
  __pyx_t_5 = __pyx_v_self->value_adapters; __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 178, __pyx_L1_error)
      #endif
      if (__pyx_t_3 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(2, 178, __pyx_L1_error)
    #else
    __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5cavro_ValueAdapter))))) __PYX_ERR(2, 178, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_adapter, ((struct __pyx_obj_5cavro_ValueAdapter *)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "src/type.pxi":179
 * 
 *         for adapter in self.value_adapters:
 *             try:             # <<<<<<<<<<<<<<
 *                 value = adapter.encode_value(value)
 *             except Exception as e:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      /*try:*/ {

        /* "src/type.pxi":180
 *         for adapter in self.value_adapters:
 *             try:
 *                 value = adapter.encode_value(value)             # <<<<<<<<<<<<<<
 *             except Exception as e:
 *                 return FIT_NONE
 */
        __pyx_t_6 = ((struct __pyx_vtabstruct_5cavro_ValueAdapter *)__pyx_v_adapter->__pyx_vtab)->encode_value(__pyx_v_adapter, __pyx_v_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 180, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "src/type.pxi":179
 * 
 *         for adapter in self.value_adapters:
 *             try:             # <<<<<<<<<<<<<<
 *                 value = adapter.encode_value(value)
 *             except Exception as e:
 */
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L20_try_end;
      __pyx_L13_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "src/type.pxi":181
 *             try:
 *                 value = adapter.encode_value(value)
 *             except Exception as e:             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *         value_fit = self._get_value_fitness(value)
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("cavro.AvroType.get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_13) < 0) __PYX_ERR(2, 181, __pyx_L15_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_e = __pyx_t_7;
        /*try:*/ {

          /* "src/type.pxi":182
 *                 value = adapter.encode_value(value)
 *             except Exception as e:
 *                 return FIT_NONE             # <<<<<<<<<<<<<<
 *         value_fit = self._get_value_fitness(value)
 *         if value_fit > FIT_NONE:
 */
          __pyx_r = __pyx_v_5cavro_FIT_NONE;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          goto __pyx_L25_return;
        }

        /* "src/type.pxi":181
 *             try:
 *                 value = adapter.encode_value(value)
 *             except Exception as e:             # <<<<<<<<<<<<<<
 *                 return FIT_NONE
 *         value_fit = self._get_value_fitness(value)
 */
        /*finally:*/ {
          __pyx_L25_return: {
            __pyx_t_9 = __pyx_r;
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            __pyx_r = __pyx_t_9;
            goto __pyx_L16_except_return;
          }
        }
      }
      goto __pyx_L15_except_error;

      /* "src/type.pxi":179
 * 
 *         for adapter in self.value_adapters:
 *             try:             # <<<<<<<<<<<<<<
 *                 value = adapter.encode_value(value)
 *             except Exception as e:
 */
      __pyx_L15_except_error:;
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      goto __pyx_L1_error;
      __pyx_L16_except_return:;
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      goto __pyx_L0;
      __pyx_L20_try_end:;
    }

    /* "src/type.pxi":178
 *                 return inner_fitness
 * 
 *         for adapter in self.value_adapters:             # <<<<<<<<<<<<<<
 *             try:
 *                 value = adapter.encode_value(value)
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/type.pxi":183
 *             except Exception as e:
 *                 return FIT_NONE
 *         value_fit = self._get_value_fitness(value)             # <<<<<<<<<<<<<<
 *         if value_fit > FIT_NONE:
 *             return value_fit
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->_get_value_fitness(__pyx_v_self, __pyx_v_value); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 183, __pyx_L1_error)
  __pyx_v_value_fit = __pyx_t_9;

  /* "src/type.pxi":184
 *                 return FIT_NONE
 *         value_fit = self._get_value_fitness(value)
 *         if value_fit > FIT_NONE:             # <<<<<<<<<<<<<<
 *             return value_fit
 * 
 */
  __pyx_t_1 = (__pyx_v_value_fit > __pyx_v_5cavro_FIT_NONE);
  if (__pyx_t_1) {

    /* "src/type.pxi":185
 *         value_fit = self._get_value_fitness(value)
 *         if value_fit > FIT_NONE:
 *             return value_fit             # <<<<<<<<<<<<<<
 * 
 *         return FIT_NONE
 */
    __pyx_r = __pyx_v_value_fit;
    goto __pyx_L0;

    /* "src/type.pxi":184
 *                 return FIT_NONE
 *         value_fit = self._get_value_fitness(value)
 *         if value_fit > FIT_NONE:             # <<<<<<<<<<<<<<
 *             return value_fit
 * 
 */
  }

  /* "src/type.pxi":187
 *             return value_fit
 * 
 *         return FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __pyx_r = __pyx_v_5cavro_FIT_NONE;
  goto __pyx_L0;

  /* "src/type.pxi":169
 *             f"{type(self).__name__} does not implement binary_buffer_decode")
 * 
 *     cdef int get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter adapter
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("cavro.AvroType.get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_adapter);
  __Pyx_XDECREF(__pyx_v_type_name);
  __Pyx_XDECREF(__pyx_v_inner_value);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":189
 *         return FIT_NONE
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _get_value_fitness")
 */

static int __pyx_f_5cavro_8AvroType__get_value_fitness(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_value_fitness", 1);

  /* "src/type.pxi":191
 *     cdef int _get_value_fitness(self, value) except -1:
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _get_value_fitness")             # <<<<<<<<<<<<<<
 * 
 *     cdef int assert_value(self, object value) except -1:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement__get_value_f); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/type.pxi":190
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement _get_value_fitness")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(2, 190, __pyx_L1_error)

  /* "src/type.pxi":189
 *         return FIT_NONE
 * 
 *     cdef int _get_value_fitness(self, value) except -1:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _get_value_fitness")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.AvroType._get_value_fitness", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":193
 *             f"{type(self).__name__} does not implement _get_value_fitness")
 * 
 *     cdef int assert_value(self, object value) except -1:             # <<<<<<<<<<<<<<
 *         cdef int fitness = self.get_value_fitness(value)
 *         if fitness == FIT_NONE:
 */

static int __pyx_f_5cavro_8AvroType_assert_value(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_v_fitness;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assert_value", 1);

  /* "src/type.pxi":194
 * 
 *     cdef int assert_value(self, object value) except -1:
 *         cdef int fitness = self.get_value_fitness(value)             # <<<<<<<<<<<<<<
 *         if fitness == FIT_NONE:
 *             raise ValueError(f"'{value}' not valid for {type(self).__name__}")
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->get_value_fitness(__pyx_v_self, __pyx_v_value); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 194, __pyx_L1_error)
  __pyx_v_fitness = __pyx_t_1;

  /* "src/type.pxi":195
 *     cdef int assert_value(self, object value) except -1:
 *         cdef int fitness = self.get_value_fitness(value)
 *         if fitness == FIT_NONE:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"'{value}' not valid for {type(self).__name__}")
 * 
 */
  __pyx_t_2 = (__pyx_v_fitness == __pyx_v_5cavro_FIT_NONE);
  if (unlikely(__pyx_t_2)) {

    /* "src/type.pxi":196
 *         cdef int fitness = self.get_value_fitness(value)
 *         if fitness == FIT_NONE:
 *             raise ValueError(f"'{value}' not valid for {type(self).__name__}")             # <<<<<<<<<<<<<<
 * 
 *     cdef json_format(self, value):
 */
    __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u__6);
    __pyx_t_4 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__6);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u__6);
    __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_not_valid_for);
    __pyx_t_4 += 16;
    __Pyx_GIVEREF(__pyx_kp_u_not_valid_for);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_not_valid_for);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_6, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_3, 4, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 196, __pyx_L1_error)

    /* "src/type.pxi":195
 *     cdef int assert_value(self, object value) except -1:
 *         cdef int fitness = self.get_value_fitness(value)
 *         if fitness == FIT_NONE:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"'{value}' not valid for {type(self).__name__}")
 * 
 */
  }

  /* "src/type.pxi":193
 *             f"{type(self).__name__} does not implement _get_value_fitness")
 * 
 *     cdef int assert_value(self, object value) except -1:             # <<<<<<<<<<<<<<
 *         cdef int fitness = self.get_value_fitness(value)
 *         if fitness == FIT_NONE:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.AvroType.assert_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":198
 *             raise ValueError(f"'{value}' not valid for {type(self).__name__}")
 * 
 *     cdef json_format(self, value):             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter adapter
 * 
 */

static PyObject *__pyx_f_5cavro_8AvroType_json_format(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_adapter = 0;
  PyObject *__pyx_v_type_name = NULL;
  PyObject *__pyx_v_inner_value = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("json_format", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/type.pxi":201
 *         cdef ValueAdapter adapter
 * 
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):             # <<<<<<<<<<<<<<
 *             type_name, inner_value = value
 *             if type_name == self.type:
 */
  if (__pyx_v_self->options->allow_tuple_notation) {
  } else {
    __pyx_t_1 = __pyx_v_self->options->allow_tuple_notation;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 201, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_3 == 2);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_5cavro_ArrayType); 
  __pyx_t_4 = (!__pyx_t_2);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/type.pxi":202
 * 
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):
 *             type_name, inner_value = value             # <<<<<<<<<<<<<<
 *             if type_name == self.type:
 *                 try:
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_value))) || (PyList_CheckExact(__pyx_v_value))) {
      PyObject* sequence = __pyx_v_value;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(2, 202, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(2, 202, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L9_unpacking_done;
      __pyx_L8_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(2, 202, __pyx_L1_error)
      __pyx_L9_unpacking_done:;
    }
    __pyx_v_type_name = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_inner_value = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "src/type.pxi":203
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):
 *             type_name, inner_value = value
 *             if type_name == self.type:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.json_format(inner_value)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_type_name, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 203, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(2, 203, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_1) {

      /* "src/type.pxi":204
 *             type_name, inner_value = value
 *             if type_name == self.type:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.json_format(inner_value)
 *                 except Exception as e:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        /*try:*/ {

          /* "src/type.pxi":205
 *             if type_name == self.type:
 *                 try:
 *                     return self.json_format(inner_value)             # <<<<<<<<<<<<<<
 *                 except Exception as e:
 *                     pass
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->json_format(__pyx_v_self, __pyx_v_inner_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 205, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_r = __pyx_t_5;
          __pyx_t_5 = 0;
          goto __pyx_L15_try_return;

          /* "src/type.pxi":204
 *             type_name, inner_value = value
 *             if type_name == self.type:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.json_format(inner_value)
 *                 except Exception as e:
 */
        }
        __pyx_L11_error:;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "src/type.pxi":206
 *                 try:
 *                     return self.json_format(inner_value)
 *                 except Exception as e:             # <<<<<<<<<<<<<<
 *                     pass
 * 
 */
        __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
        if (__pyx_t_12) {
          __Pyx_AddTraceback("cavro.AvroType.json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 206, __pyx_L13_except_error)
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          __pyx_v_e = __pyx_t_6;
          /*try:*/ {
          }
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
              goto __pyx_L23;
            }
            __pyx_L23:;
          }
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          goto __pyx_L12_exception_handled;
        }
        goto __pyx_L13_except_error;

        /* "src/type.pxi":204
 *             type_name, inner_value = value
 *             if type_name == self.type:
 *                 try:             # <<<<<<<<<<<<<<
 *                     return self.json_format(inner_value)
 *                 except Exception as e:
 */
        __pyx_L13_except_error:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L1_error;
        __pyx_L15_try_return:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L0;
        __pyx_L12_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }

      /* "src/type.pxi":203
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):
 *             type_name, inner_value = value
 *             if type_name == self.type:             # <<<<<<<<<<<<<<
 *                 try:
 *                     return self.json_format(inner_value)
 */
    }

    /* "src/type.pxi":201
 *         cdef ValueAdapter adapter
 * 
 *         if self.options.allow_tuple_notation and  isinstance(value, tuple) and len(value) == 2 and not isinstance(self, ArrayType):             # <<<<<<<<<<<<<<
 *             type_name, inner_value = value
 *             if type_name == self.type:
 */
  }

  /* "src/type.pxi":209
 *                     pass
 * 
 *         for adapter in self.value_adapters:             # <<<<<<<<<<<<<<
 *             value = adapter.encode_value(value)
 *         return self._json_format(value)
 */
  if (unlikely(__pyx_v_self->value_adapters == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(2, 209, __pyx_L1_error)
  }
  __pyx_t_7 = __pyx_v_self->value_adapters; __Pyx_INCREF(__pyx_t_7);
  __pyx_t_3 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_7);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 209, __pyx_L1_error)
      #endif
      if (__pyx_t_3 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(2, 209, __pyx_L1_error)
    #else
    __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5cavro_ValueAdapter))))) __PYX_ERR(2, 209, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_adapter, ((struct __pyx_obj_5cavro_ValueAdapter *)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "src/type.pxi":210
 * 
 *         for adapter in self.value_adapters:
 *             value = adapter.encode_value(value)             # <<<<<<<<<<<<<<
 *         return self._json_format(value)
 * 
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_5cavro_ValueAdapter *)__pyx_v_adapter->__pyx_vtab)->encode_value(__pyx_v_adapter, __pyx_v_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "src/type.pxi":209
 *                     pass
 * 
 *         for adapter in self.value_adapters:             # <<<<<<<<<<<<<<
 *             value = adapter.encode_value(value)
 *         return self._json_format(value)
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/type.pxi":211
 *         for adapter in self.value_adapters:
 *             value = adapter.encode_value(value)
 *         return self._json_format(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_format(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->_json_format(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "src/type.pxi":198
 *             raise ValueError(f"'{value}' not valid for {type(self).__name__}")
 * 
 *     cdef json_format(self, value):             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter adapter
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.AvroType.json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_adapter);
  __Pyx_XDECREF(__pyx_v_type_name);
  __Pyx_XDECREF(__pyx_v_inner_value);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":213
 *         return self._json_format(value)
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _json_format")
 */

static PyObject *__pyx_f_5cavro_8AvroType__json_format(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_format", 1);

  /* "src/type.pxi":215
 *     cdef _json_format(self, value):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _json_format")             # <<<<<<<<<<<<<<
 * 
 *     cdef json_decode(self, value):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement__json_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/type.pxi":214
 * 
 *     cdef _json_format(self, value):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement _json_format")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(2, 214, __pyx_L1_error)

  /* "src/type.pxi":213
 *         return self._json_format(value)
 * 
 *     cdef _json_format(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _json_format")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.AvroType._json_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":217
 *             f"{type(self).__name__} does not implement _json_format")
 * 
 *     cdef json_decode(self, value):             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter adapter
 *         value = self._json_decode(value)
 */

static PyObject *__pyx_f_5cavro_8AvroType_json_decode(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_adapter = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("json_decode", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/type.pxi":219
 *     cdef json_decode(self, value):
 *         cdef ValueAdapter adapter
 *         value = self._json_decode(value)             # <<<<<<<<<<<<<<
 *         for adapter in reversed(self.value_adapters):
 *             value = adapter.decode_value(value)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->_json_decode(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":220
 *         cdef ValueAdapter adapter
 *         value = self._json_decode(value)
 *         for adapter in reversed(self.value_adapters):             # <<<<<<<<<<<<<<
 *             value = adapter.decode_value(value)
 *         return value
 */
  if (unlikely(__pyx_v_self->value_adapters == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(2, 220, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->value_adapters; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
  #if !CYTHON_ASSUME_SAFE_MACROS
  if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 220, __pyx_L1_error)
  #endif
  --__pyx_t_2;
  for (;;) {
    if (__pyx_t_2 < 0) break;
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 220, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2--; if (unlikely((0 < 0))) __PYX_ERR(2, 220, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2--; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5cavro_ValueAdapter))))) __PYX_ERR(2, 220, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_adapter, ((struct __pyx_obj_5cavro_ValueAdapter *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "src/type.pxi":221
 *         value = self._json_decode(value)
 *         for adapter in reversed(self.value_adapters):
 *             value = adapter.decode_value(value)             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_ValueAdapter *)__pyx_v_adapter->__pyx_vtab)->decode_value(__pyx_v_adapter, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/type.pxi":220
 *         cdef ValueAdapter adapter
 *         value = self._json_decode(value)
 *         for adapter in reversed(self.value_adapters):             # <<<<<<<<<<<<<<
 *             value = adapter.decode_value(value)
 *         return value
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/type.pxi":222
 *         for adapter in reversed(self.value_adapters):
 *             value = adapter.decode_value(value)
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/type.pxi":217
 *             f"{type(self).__name__} does not implement _json_format")
 * 
 *     cdef json_decode(self, value):             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter adapter
 *         value = self._json_decode(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.AvroType.json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_adapter);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":224
 *         return value
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _json_decode")
 */

static PyObject *__pyx_f_5cavro_8AvroType__json_decode(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_json_decode", 1);

  /* "src/type.pxi":226
 *     cdef _json_decode(self, value):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _json_decode")             # <<<<<<<<<<<<<<
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement__json_decode); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/type.pxi":225
 * 
 *     cdef _json_decode(self, value):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement _json_decode")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(2, 225, __pyx_L1_error)

  /* "src/type.pxi":224
 *         return value
 * 
 *     cdef _json_decode(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _json_decode")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.AvroType._json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":228
 *             f"{type(self).__name__} does not implement _json_decode")
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement canonical_form")
 */

static struct __pyx_obj_5cavro__CanonicalForm *__pyx_f_5cavro_8AvroType_canonical_form(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created) {
  struct __pyx_obj_5cavro__CanonicalForm *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canonical_form", 1);

  /* "src/type.pxi":230
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement canonical_form")             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement_canonical_fo); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/type.pxi":229
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement canonical_form")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(2, 229, __pyx_L1_error)

  /* "src/type.pxi":228
 *             f"{type(self).__name__} does not implement _json_decode")
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement canonical_form")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.AvroType.canonical_form", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":232
 *             f"{type(self).__name__} does not implement canonical_form")
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _get_schema_extra")
 */

static PyObject *__pyx_pw_5cavro_8AvroType_13_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_8AvroType__get_schema_extra(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_schema_extra); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_8AvroType_13_get_schema_extra)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_created};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 232, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_2))) __PYX_ERR(2, 232, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/type.pxi":234
 *     cpdef dict _get_schema_extra(self, set created):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _get_schema_extra")             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType for_writer(self, AvroType writer, bint _allow_deferrals=True):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement__get_schema); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/type.pxi":233
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement _get_schema_extra")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(2, 233, __pyx_L1_error)

  /* "src/type.pxi":232
 *             f"{type(self).__name__} does not implement canonical_form")
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _get_schema_extra")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.AvroType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_13_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8AvroType_13_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_13_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8AvroType_13_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 232, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(2, 232, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 232, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.AvroType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(2, 232, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_8AvroType_12_get_schema_extra(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_12_get_schema_extra(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_8AvroType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.AvroType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":236
 *             f"{type(self).__name__} does not implement _get_schema_extra")
 * 
 *     cdef AvroType for_writer(self, AvroType writer, bint _allow_deferrals=True):             # <<<<<<<<<<<<<<
 *         cdef UnionType writer_union
 *         cdef AvroType promoted
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8AvroType_for_writer(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_writer, struct __pyx_opt_args_5cavro_8AvroType_for_writer *__pyx_optional_args) {
  int __pyx_v__allow_deferrals = ((int)1);
  struct __pyx_obj_5cavro_UnionType *__pyx_v_writer_union = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_promoted = 0;
  int __pyx_v_allow_deferrals;
  PyObject *__pyx_v_promote_error = NULL;
  PyObject *__pyx_v_e = NULL;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_cloned = NULL;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("for_writer", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v__allow_deferrals = __pyx_optional_args->_allow_deferrals;
    }
  }

  /* "src/type.pxi":240
 *         cdef AvroType promoted
 *         #cdef AvroType cloned
 *         cdef bint allow_deferrals = _allow_deferrals and self.options.defer_schema_promotion_errors             # <<<<<<<<<<<<<<
 * 
 *         promote_error = None
 */
  if (__pyx_v__allow_deferrals) {
  } else {
    __pyx_t_1 = __pyx_v__allow_deferrals;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_1 = __pyx_v_self->options->defer_schema_promotion_errors;
  __pyx_L3_bool_binop_done:;
  __pyx_v_allow_deferrals = __pyx_t_1;

  /* "src/type.pxi":242
 *         cdef bint allow_deferrals = _allow_deferrals and self.options.defer_schema_promotion_errors
 * 
 *         promote_error = None             # <<<<<<<<<<<<<<
 *         promoted = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_promote_error = Py_None;

  /* "src/type.pxi":243
 * 
 *         promote_error = None
 *         promoted = None             # <<<<<<<<<<<<<<
 * 
 *         if self.canonical_form(set()) == writer.canonical_form(set()):
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_promoted = ((struct __pyx_obj_5cavro_AvroType *)Py_None);

  /* "src/type.pxi":245
 *         promoted = None
 * 
 *         if self.canonical_form(set()) == writer.canonical_form(set()):             # <<<<<<<<<<<<<<
 *             return writer
 *         if isinstance(writer, UnionType) and not isinstance(self, UnionType):
 */
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->canonical_form(__pyx_v_self, ((PyObject*)__pyx_t_2))); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_writer->__pyx_vtab)->canonical_form(__pyx_v_writer, ((PyObject*)__pyx_t_2))); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(2, 245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "src/type.pxi":246
 * 
 *         if self.canonical_form(set()) == writer.canonical_form(set()):
 *             return writer             # <<<<<<<<<<<<<<
 *         if isinstance(writer, UnionType) and not isinstance(self, UnionType):
 *             writer_union = writer
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_writer);
    __pyx_r = __pyx_v_writer;
    goto __pyx_L0;

    /* "src/type.pxi":245
 *         promoted = None
 * 
 *         if self.canonical_form(set()) == writer.canonical_form(set()):             # <<<<<<<<<<<<<<
 *             return writer
 *         if isinstance(writer, UnionType) and not isinstance(self, UnionType):
 */
  }

  /* "src/type.pxi":247
 *         if self.canonical_form(set()) == writer.canonical_form(set()):
 *             return writer
 *         if isinstance(writer, UnionType) and not isinstance(self, UnionType):             # <<<<<<<<<<<<<<
 *             writer_union = writer
 *             try:
 */
  __pyx_t_5 = __Pyx_TypeCheck(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_UnionType); 
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_5cavro_UnionType); 
  __pyx_t_6 = (!__pyx_t_5);
  __pyx_t_1 = __pyx_t_6;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/type.pxi":248
 *             return writer
 *         if isinstance(writer, UnionType) and not isinstance(self, UnionType):
 *             writer_union = writer             # <<<<<<<<<<<<<<
 *             try:
 *                 promoted = writer_union.for_reader(self)
 */
    if (!(likely(((((PyObject *)__pyx_v_writer)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro_UnionType))))) __PYX_ERR(2, 248, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)__pyx_v_writer);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_writer_union = ((struct __pyx_obj_5cavro_UnionType *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/type.pxi":249
 *         if isinstance(writer, UnionType) and not isinstance(self, UnionType):
 *             writer_union = writer
 *             try:             # <<<<<<<<<<<<<<
 *                 promoted = writer_union.for_reader(self)
 *             except CannotPromoteError as e:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "src/type.pxi":250
 *             writer_union = writer
 *             try:
 *                 promoted = writer_union.for_reader(self)             # <<<<<<<<<<<<<<
 *             except CannotPromoteError as e:
 *                 promote_error = e
 */
        __pyx_t_2 = ((PyObject *)__pyx_f_5cavro_9UnionType_for_reader(__pyx_v_writer_union, __pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 250, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_promoted, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "src/type.pxi":249
 *         if isinstance(writer, UnionType) and not isinstance(self, UnionType):
 *             writer_union = writer
 *             try:             # <<<<<<<<<<<<<<
 *                 promoted = writer_union.for_reader(self)
 *             except CannotPromoteError as e:
 */
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L14_try_end;
      __pyx_L9_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "src/type.pxi":251
 *             try:
 *                 promoted = writer_union.for_reader(self)
 *             except CannotPromoteError as e:             # <<<<<<<<<<<<<<
 *                 promote_error = e
 *         else:
 */
      __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_4, &__pyx_t_3);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_CannotPromoteError); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 251, __pyx_L11_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_10);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_ErrRestore(__pyx_t_2, __pyx_t_4, __pyx_t_3);
      __pyx_t_2 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0;
      if (__pyx_t_11) {
        __Pyx_AddTraceback("cavro.AvroType.for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_2) < 0) __PYX_ERR(2, 251, __pyx_L11_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_e = __pyx_t_4;
        /*try:*/ {

          /* "src/type.pxi":252
 *                 promoted = writer_union.for_reader(self)
 *             except CannotPromoteError as e:
 *                 promote_error = e             # <<<<<<<<<<<<<<
 *         else:
 *             try:
 */
          __Pyx_INCREF(__pyx_v_e);
          __Pyx_DECREF_SET(__pyx_v_promote_error, __pyx_v_e);
        }

        /* "src/type.pxi":251
 *             try:
 *                 promoted = writer_union.for_reader(self)
 *             except CannotPromoteError as e:             # <<<<<<<<<<<<<<
 *                 promote_error = e
 *         else:
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            goto __pyx_L21;
          }
          __pyx_L21:;
        }
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L10_exception_handled;
      }
      goto __pyx_L11_except_error;

      /* "src/type.pxi":249
 *         if isinstance(writer, UnionType) and not isinstance(self, UnionType):
 *             writer_union = writer
 *             try:             # <<<<<<<<<<<<<<
 *                 promoted = writer_union.for_reader(self)
 *             except CannotPromoteError as e:
 */
      __pyx_L11_except_error:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L10_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_L14_try_end:;
    }

    /* "src/type.pxi":247
 *         if self.canonical_form(set()) == writer.canonical_form(set()):
 *             return writer
 *         if isinstance(writer, UnionType) and not isinstance(self, UnionType):             # <<<<<<<<<<<<<<
 *             writer_union = writer
 *             try:
 */
    goto __pyx_L6;
  }

  /* "src/type.pxi":254
 *                 promote_error = e
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 promoted = self._for_writer(writer)
 *             except CannotPromoteError as e:
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "src/type.pxi":255
 *         else:
 *             try:
 *                 promoted = self._for_writer(writer)             # <<<<<<<<<<<<<<
 *             except CannotPromoteError as e:
 *                 promote_error = e
 */
        __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->_for_writer(__pyx_v_self, __pyx_v_writer)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 255, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_promoted, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "src/type.pxi":254
 *                 promote_error = e
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 promoted = self._for_writer(writer)
 *             except CannotPromoteError as e:
 */
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L27_try_end;
      __pyx_L22_error:;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "src/type.pxi":256
 *             try:
 *                 promoted = self._for_writer(writer)
 *             except CannotPromoteError as e:             # <<<<<<<<<<<<<<
 *                 promote_error = e
 * 
 */
      __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_4, &__pyx_t_3);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_CannotPromoteError); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 256, __pyx_L24_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_10);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_ErrRestore(__pyx_t_2, __pyx_t_4, __pyx_t_3);
      __pyx_t_2 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0;
      if (__pyx_t_11) {
        __Pyx_AddTraceback("cavro.AvroType.for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_2) < 0) __PYX_ERR(2, 256, __pyx_L24_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_e = __pyx_t_4;
        /*try:*/ {

          /* "src/type.pxi":257
 *                 promoted = self._for_writer(writer)
 *             except CannotPromoteError as e:
 *                 promote_error = e             # <<<<<<<<<<<<<<
 * 
 *         if promoted is None and promote_error is None:
 */
          __Pyx_INCREF(__pyx_v_e);
          __Pyx_DECREF_SET(__pyx_v_promote_error, __pyx_v_e);
        }

        /* "src/type.pxi":256
 *             try:
 *                 promoted = self._for_writer(writer)
 *             except CannotPromoteError as e:             # <<<<<<<<<<<<<<
 *                 promote_error = e
 * 
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            goto __pyx_L34;
          }
          __pyx_L34:;
        }
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L23_exception_handled;
      }
      goto __pyx_L24_except_error;

      /* "src/type.pxi":254
 *                 promote_error = e
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 promoted = self._for_writer(writer)
 *             except CannotPromoteError as e:
 */
      __pyx_L24_except_error:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L23_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      __pyx_L27_try_end:;
    }
  }
  __pyx_L6:;

  /* "src/type.pxi":259
 *                 promote_error = e
 * 
 *         if promoted is None and promote_error is None:             # <<<<<<<<<<<<<<
 *             promote_error = CannotPromoteError(self, writer)
 * 
 */
  __pyx_t_6 = (((PyObject *)__pyx_v_promoted) == Py_None);
  if (__pyx_t_6) {
  } else {
    __pyx_t_1 = __pyx_t_6;
    goto __pyx_L36_bool_binop_done;
  }
  __pyx_t_6 = (__pyx_v_promote_error == Py_None);
  __pyx_t_1 = __pyx_t_6;
  __pyx_L36_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/type.pxi":260
 * 
 *         if promoted is None and promote_error is None:
 *             promote_error = CannotPromoteError(self, writer)             # <<<<<<<<<<<<<<
 * 
 *         if promote_error is not None:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CannotPromoteError); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, ((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_writer)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_11, 2+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_promote_error, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/type.pxi":259
 *                 promote_error = e
 * 
 *         if promoted is None and promote_error is None:             # <<<<<<<<<<<<<<
 *             promote_error = CannotPromoteError(self, writer)
 * 
 */
  }

  /* "src/type.pxi":262
 *             promote_error = CannotPromoteError(self, writer)
 * 
 *         if promote_error is not None:             # <<<<<<<<<<<<<<
 *             if allow_deferrals:
 *                 cloned = writer.copy()
 */
  __pyx_t_1 = (__pyx_v_promote_error != Py_None);
  if (__pyx_t_1) {

    /* "src/type.pxi":263
 * 
 *         if promote_error is not None:
 *             if allow_deferrals:             # <<<<<<<<<<<<<<
 *                 cloned = writer.copy()
 *                 cloned.value_adapters = (CannotPromote(self, writer, promote_error.extra),) + cloned.value_adapters
 */
    if (__pyx_v_allow_deferrals) {

      /* "src/type.pxi":264
 *         if promote_error is not None:
 *             if allow_deferrals:
 *                 cloned = writer.copy()             # <<<<<<<<<<<<<<
 *                 cloned.value_adapters = (CannotPromote(self, writer, promote_error.extra),) + cloned.value_adapters
 *                 return cloned
 */
      __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_writer->__pyx_vtab)->copy(__pyx_v_writer, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_cloned = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "src/type.pxi":265
 *             if allow_deferrals:
 *                 cloned = writer.copy()
 *                 cloned.value_adapters = (CannotPromote(self, writer, promote_error.extra),) + cloned.value_adapters             # <<<<<<<<<<<<<<
 *                 return cloned
 *             raise promote_error
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_promote_error, __pyx_n_s_extra); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF((PyObject *)__pyx_v_self);
      __Pyx_GIVEREF((PyObject *)__pyx_v_self);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_self))) __PYX_ERR(2, 265, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_writer);
      __Pyx_GIVEREF((PyObject *)__pyx_v_writer);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_writer))) __PYX_ERR(2, 265, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_2);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2)) __PYX_ERR(2, 265, __pyx_L1_error);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_CannotPromote), __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_2);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(2, 265, __pyx_L1_error);
      __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Add(__pyx_t_4, __pyx_v_cloned->value_adapters); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_v_cloned->value_adapters);
      __Pyx_DECREF(__pyx_v_cloned->value_adapters);
      __pyx_v_cloned->value_adapters = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "src/type.pxi":266
 *                 cloned = writer.copy()
 *                 cloned.value_adapters = (CannotPromote(self, writer, promote_error.extra),) + cloned.value_adapters
 *                 return cloned             # <<<<<<<<<<<<<<
 *             raise promote_error
 *         return promoted
 */
      __Pyx_XDECREF((PyObject *)__pyx_r);
      __Pyx_INCREF((PyObject *)__pyx_v_cloned);
      __pyx_r = __pyx_v_cloned;
      goto __pyx_L0;

      /* "src/type.pxi":263
 * 
 *         if promote_error is not None:
 *             if allow_deferrals:             # <<<<<<<<<<<<<<
 *                 cloned = writer.copy()
 *                 cloned.value_adapters = (CannotPromote(self, writer, promote_error.extra),) + cloned.value_adapters
 */
    }

    /* "src/type.pxi":267
 *                 cloned.value_adapters = (CannotPromote(self, writer, promote_error.extra),) + cloned.value_adapters
 *                 return cloned
 *             raise promote_error             # <<<<<<<<<<<<<<
 *         return promoted
 * 
 */
    __Pyx_Raise(__pyx_v_promote_error, 0, 0, 0);
    __PYX_ERR(2, 267, __pyx_L1_error)

    /* "src/type.pxi":262
 *             promote_error = CannotPromoteError(self, writer)
 * 
 *         if promote_error is not None:             # <<<<<<<<<<<<<<
 *             if allow_deferrals:
 *                 cloned = writer.copy()
 */
  }

  /* "src/type.pxi":268
 *                 return cloned
 *             raise promote_error
 *         return promoted             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_promoted);
  __pyx_r = __pyx_v_promoted;
  goto __pyx_L0;

  /* "src/type.pxi":236
 *             f"{type(self).__name__} does not implement _get_schema_extra")
 * 
 *     cdef AvroType for_writer(self, AvroType writer, bint _allow_deferrals=True):             # <<<<<<<<<<<<<<
 *         cdef UnionType writer_union
 *         cdef AvroType promoted
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("cavro.AvroType.for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_writer_union);
  __Pyx_XDECREF((PyObject *)__pyx_v_promoted);
  __Pyx_XDECREF(__pyx_v_promote_error);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF((PyObject *)__pyx_v_cloned);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":270
 *         return promoted
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_8AvroType__for_writer(CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_writer) {
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_for_writer", 1);

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":273
 *         pass
 * 
 *     cdef bint accepts_missing_value(self):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */

static int __pyx_f_5cavro_8AvroType_accepts_missing_value(CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_self) {
  int __pyx_r;

  /* "src/type.pxi":274
 * 
 *     cdef bint accepts_missing_value(self):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "src/type.pxi":273
 *         pass
 * 
 *     cdef bint accepts_missing_value(self):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "src/type.pxi":276
 *         return False
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):             # <<<<<<<<<<<<<<
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 */

static PyObject *__pyx_f_5cavro_8AvroType_resolve_default_value(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_schema_default, PyObject *__pyx_v_field) {
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  Py_UCS4 __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resolve_default_value", 1);

  /* "src/type.pxi":277
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if schema_default is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             return NO_DEFAULT
 *         try:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_v_schema_default == __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/type.pxi":278
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT             # <<<<<<<<<<<<<<
 *         try:
 *             return self.json_decode(schema_default)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NO_DEFAULT); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/type.pxi":277
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):
 *         if schema_default is NO_DEFAULT:             # <<<<<<<<<<<<<<
 *             return NO_DEFAULT
 *         try:
 */
  }

  /* "src/type.pxi":279
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 *         try:             # <<<<<<<<<<<<<<
 *             return self.json_decode(schema_default)
 *         except (TypeError, AttributeError, ValueError) as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "src/type.pxi":280
 *             return NO_DEFAULT
 *         try:
 *             return self.json_decode(schema_default)             # <<<<<<<<<<<<<<
 *         except (TypeError, AttributeError, ValueError) as e:
 *             if self.options.allow_invalid_default_values:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->json_decode(__pyx_v_self, __pyx_v_schema_default); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 280, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L8_try_return;

      /* "src/type.pxi":279
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 *         try:             # <<<<<<<<<<<<<<
 *             return self.json_decode(schema_default)
 *         except (TypeError, AttributeError, ValueError) as e:
 */
    }
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/type.pxi":281
 *         try:
 *             return self.json_decode(schema_default)
 *         except (TypeError, AttributeError, ValueError) as e:             # <<<<<<<<<<<<<<
 *             if self.options.allow_invalid_default_values:
 *                 return schema_default
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("cavro.AvroType.resolve_default_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(2, 281, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __pyx_v_e = __pyx_t_7;
      /*try:*/ {

        /* "src/type.pxi":282
 *             return self.json_decode(schema_default)
 *         except (TypeError, AttributeError, ValueError) as e:
 *             if self.options.allow_invalid_default_values:             # <<<<<<<<<<<<<<
 *                 return schema_default
 *             raise TypeError(f"Default value {schema_default!r} is not valid for field: {field}") from e
 */
        if (__pyx_v_self->options->allow_invalid_default_values) {

          /* "src/type.pxi":283
 *         except (TypeError, AttributeError, ValueError) as e:
 *             if self.options.allow_invalid_default_values:
 *                 return schema_default             # <<<<<<<<<<<<<<
 *             raise TypeError(f"Default value {schema_default!r} is not valid for field: {field}") from e
 * 
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(__pyx_v_schema_default);
          __pyx_r = __pyx_v_schema_default;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          goto __pyx_L14_return;

          /* "src/type.pxi":282
 *             return self.json_decode(schema_default)
 *         except (TypeError, AttributeError, ValueError) as e:
 *             if self.options.allow_invalid_default_values:             # <<<<<<<<<<<<<<
 *                 return schema_default
 *             raise TypeError(f"Default value {schema_default!r} is not valid for field: {field}") from e
 */
        }

        /* "src/type.pxi":284
 *             if self.options.allow_invalid_default_values:
 *                 return schema_default
 *             raise TypeError(f"Default value {schema_default!r} is not valid for field: {field}") from e             # <<<<<<<<<<<<<<
 * 
 *     def walk_types(self, visited):
 */
        __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 284, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = 0;
        __pyx_t_11 = 127;
        __Pyx_INCREF(__pyx_kp_u_Default_value_2);
        __pyx_t_10 += 14;
        __Pyx_GIVEREF(__pyx_kp_u_Default_value_2);
        PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_kp_u_Default_value_2);
        __pyx_t_12 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_schema_default), __pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 284, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
        __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_12);
        __pyx_t_12 = 0;
        __Pyx_INCREF(__pyx_kp_u_is_not_valid_for_field);
        __pyx_t_10 += 25;
        __Pyx_GIVEREF(__pyx_kp_u_is_not_valid_for_field);
        PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_kp_u_is_not_valid_for_field);
        __pyx_t_12 = __Pyx_PyUnicode_Unicode(__pyx_v_field); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 284, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
        __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_12);
        __pyx_t_12 = 0;
        __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_9, 4, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 284, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 284, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_Raise(__pyx_t_9, 0, 0, __pyx_v_e);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(2, 284, __pyx_L15_error)
      }

      /* "src/type.pxi":281
 *         try:
 *             return self.json_decode(schema_default)
 *         except (TypeError, AttributeError, ValueError) as e:             # <<<<<<<<<<<<<<
 *             if self.options.allow_invalid_default_values:
 *                 return schema_default
 */
      /*finally:*/ {
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
          }
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
          goto __pyx_L6_except_error;
        }
        __pyx_L14_return: {
          __pyx_t_20 = __pyx_r;
          __pyx_r = 0;
          __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          __pyx_r = __pyx_t_20;
          __pyx_t_20 = 0;
          goto __pyx_L7_except_return;
        }
      }
    }
    goto __pyx_L6_except_error;

    /* "src/type.pxi":279
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 *         try:             # <<<<<<<<<<<<<<
 *             return self.json_decode(schema_default)
 *         except (TypeError, AttributeError, ValueError) as e:
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L8_try_return:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L0;
    __pyx_L7_except_return:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L0;
  }

  /* "src/type.pxi":276
 *         return False
 * 
 *     cdef object resolve_default_value(self, object schema_default, str field):             # <<<<<<<<<<<<<<
 *         if schema_default is NO_DEFAULT:
 *             return NO_DEFAULT
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("cavro.AvroType.resolve_default_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_8AvroType_16generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/type.pxi":286
 *             raise TypeError(f"Default value {schema_default!r} is not valid for field: {field}") from e
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_15walk_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8AvroType_15walk_types = {"walk_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_15walk_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8AvroType_15walk_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_visited = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("walk_types (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_visited,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_visited)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 286, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "walk_types") < 0)) __PYX_ERR(2, 286, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_visited = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("walk_types", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 286, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.AvroType.walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8AvroType_14walk_types(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_visited);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_14walk_types(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_visited) {
  struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("walk_types", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types *)__pyx_tp_new_5cavro___pyx_scope_struct_21_walk_types(__pyx_ptype_5cavro___pyx_scope_struct_21_walk_types, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(2, 286, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_visited = __pyx_v_visited;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_visited);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_visited);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_8AvroType_16generator, __pyx_codeobj__82, (PyObject *) __pyx_cur_scope, __pyx_n_s_walk_types, __pyx_n_s_AvroType_walk_types, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(2, 286, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.AvroType.walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_8AvroType_16generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("walk_types", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L5_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(2, 286, __pyx_L1_error)

  /* "src/type.pxi":287
 * 
 *     def walk_types(self, visited):
 *         if self in visited:             # <<<<<<<<<<<<<<
 *             return
 *         visited.add(self)
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_visited, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(2, 287, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/type.pxi":288
 *     def walk_types(self, visited):
 *         if self in visited:
 *             return             # <<<<<<<<<<<<<<
 *         visited.add(self)
 *         yield self
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "src/type.pxi":287
 * 
 *     def walk_types(self, visited):
 *         if self in visited:             # <<<<<<<<<<<<<<
 *             return
 *         visited.add(self)
 */
  }

  /* "src/type.pxi":289
 *         if self in visited:
 *             return
 *         visited.add(self)             # <<<<<<<<<<<<<<
 *         yield self
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_visited, __pyx_n_s_add); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_cur_scope->__pyx_v_self)};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/type.pxi":290
 *             return
 *         visited.add(self)
 *         yield self             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_schema(self, created=None):
 */
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  /* return from generator, yielding value */
  __pyx_generator->resume_label = 1;
  return __pyx_r;
  __pyx_L5_resume_from_yield:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(2, 290, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "src/type.pxi":286
 *             raise TypeError(f"Default value {schema_default!r} is not valid for field: {field}") from e
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("walk_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":292
 *         yield self
 * 
 *     cpdef get_schema(self, created=None):             # <<<<<<<<<<<<<<
 *         if created is None:
 *             created = set()
 */

static PyObject *__pyx_pw_5cavro_8AvroType_18get_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_8AvroType_get_schema(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_8AvroType_get_schema *__pyx_optional_args) {
  PyObject *__pyx_v_created = ((PyObject *)Py_None);
  PyObject *__pyx_v_type_name = NULL;
  PyObject *__pyx_v_extra = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_schema", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_created = __pyx_optional_args->created;
    }
  }
  __Pyx_INCREF(__pyx_v_created);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_schema); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_8AvroType_18get_schema)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_created};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 292, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/type.pxi":293
 * 
 *     cpdef get_schema(self, created=None):
 *         if created is None:             # <<<<<<<<<<<<<<
 *             created = set()
 *         if self.options.expand_types_in_schema:
 */
  __pyx_t_6 = (__pyx_v_created == Py_None);
  if (__pyx_t_6) {

    /* "src/type.pxi":294
 *     cpdef get_schema(self, created=None):
 *         if created is None:
 *             created = set()             # <<<<<<<<<<<<<<
 *         if self.options.expand_types_in_schema:
 *             created = created.copy()
 */
    __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_created, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/type.pxi":293
 * 
 *     cpdef get_schema(self, created=None):
 *         if created is None:             # <<<<<<<<<<<<<<
 *             created = set()
 *         if self.options.expand_types_in_schema:
 */
  }

  /* "src/type.pxi":295
 *         if created is None:
 *             created = set()
 *         if self.options.expand_types_in_schema:             # <<<<<<<<<<<<<<
 *             created = created.copy()
 *         if isinstance(self, _NamedType):
 */
  if (__pyx_v_self->options->expand_types_in_schema) {

    /* "src/type.pxi":296
 *             created = set()
 *         if self.options.expand_types_in_schema:
 *             created = created.copy()             # <<<<<<<<<<<<<<
 *         if isinstance(self, _NamedType):
 *             type_name = self.type
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_created, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_created, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/type.pxi":295
 *         if created is None:
 *             created = set()
 *         if self.options.expand_types_in_schema:             # <<<<<<<<<<<<<<
 *             created = created.copy()
 *         if isinstance(self, _NamedType):
 */
  }

  /* "src/type.pxi":297
 *         if self.options.expand_types_in_schema:
 *             created = created.copy()
 *         if isinstance(self, _NamedType):             # <<<<<<<<<<<<<<
 *             type_name = self.type
 *             if type_name in created:
 */
  __pyx_t_6 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_5cavro__NamedType); 
  if (__pyx_t_6) {

    /* "src/type.pxi":298
 *             created = created.copy()
 *         if isinstance(self, _NamedType):
 *             type_name = self.type             # <<<<<<<<<<<<<<
 *             if type_name in created:
 *                 return type_name
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_type_name = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "src/type.pxi":299
 *         if isinstance(self, _NamedType):
 *             type_name = self.type
 *             if type_name in created:             # <<<<<<<<<<<<<<
 *                 return type_name
 *             else:
 */
    __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_v_type_name, __pyx_v_created, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(2, 299, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "src/type.pxi":300
 *             type_name = self.type
 *             if type_name in created:
 *                 return type_name             # <<<<<<<<<<<<<<
 *             else:
 *                 created.add(type_name)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_type_name);
      __pyx_r = __pyx_v_type_name;
      goto __pyx_L0;

      /* "src/type.pxi":299
 *         if isinstance(self, _NamedType):
 *             type_name = self.type
 *             if type_name in created:             # <<<<<<<<<<<<<<
 *                 return type_name
 *             else:
 */
    }

    /* "src/type.pxi":302
 *                 return type_name
 *             else:
 *                 created.add(type_name)             # <<<<<<<<<<<<<<
 * 
 *         extra = self._get_schema_extra(created)
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_created, __pyx_n_s_add); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_type_name};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 302, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "src/type.pxi":297
 *         if self.options.expand_types_in_schema:
 *             created = created.copy()
 *         if isinstance(self, _NamedType):             # <<<<<<<<<<<<<<
 *             type_name = self.type
 *             if type_name in created:
 */
  }

  /* "src/type.pxi":304
 *                 created.add(type_name)
 * 
 *         extra = self._get_schema_extra(created)             # <<<<<<<<<<<<<<
 *         if self.metadata:
 *             extra.update(self.metadata)
 */
  if (!(likely(PySet_CheckExact(__pyx_v_created))||((__pyx_v_created) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_v_created))) __PYX_ERR(2, 304, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->_get_schema_extra(__pyx_v_self, ((PyObject*)__pyx_v_created), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_extra = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":305
 * 
 *         extra = self._get_schema_extra(created)
 *         if self.metadata:             # <<<<<<<<<<<<<<
 *             extra.update(self.metadata)
 *         if extra:
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->metadata); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(2, 305, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "src/type.pxi":306
 *         extra = self._get_schema_extra(created)
 *         if self.metadata:
 *             extra.update(self.metadata)             # <<<<<<<<<<<<<<
 *         if extra:
 *             extra['type'] = self.type_name
 */
    __pyx_t_1 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_update, __pyx_v_extra, __pyx_v_self->metadata); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/type.pxi":305
 * 
 *         extra = self._get_schema_extra(created)
 *         if self.metadata:             # <<<<<<<<<<<<<<
 *             extra.update(self.metadata)
 *         if extra:
 */
  }

  /* "src/type.pxi":307
 *         if self.metadata:
 *             extra.update(self.metadata)
 *         if extra:             # <<<<<<<<<<<<<<
 *             extra['type'] = self.type_name
 *             return extra
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_extra); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(2, 307, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "src/type.pxi":308
 *             extra.update(self.metadata)
 *         if extra:
 *             extra['type'] = self.type_name             # <<<<<<<<<<<<<<
 *             return extra
 *         else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_extra == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 308, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__pyx_v_extra, __pyx_n_u_type, __pyx_t_1) < 0))) __PYX_ERR(2, 308, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/type.pxi":309
 *         if extra:
 *             extra['type'] = self.type_name
 *             return extra             # <<<<<<<<<<<<<<
 *         else:
 *             return self.type_name
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_extra);
    __pyx_r = __pyx_v_extra;
    goto __pyx_L0;

    /* "src/type.pxi":307
 *         if self.metadata:
 *             extra.update(self.metadata)
 *         if extra:             # <<<<<<<<<<<<<<
 *             extra['type'] = self.type_name
 *             return extra
 */
  }

  /* "src/type.pxi":311
 *             return extra
 *         else:
 *             return self.type_name             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 311, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "src/type.pxi":292
 *         yield self
 * 
 *     cpdef get_schema(self, created=None):             # <<<<<<<<<<<<<<
 *         if created is None:
 *             created = set()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.AvroType.get_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_type_name);
  __Pyx_XDECREF(__pyx_v_extra);
  __Pyx_XDECREF(__pyx_v_created);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_18get_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8AvroType_18get_schema = {"get_schema", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_18get_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8AvroType_18get_schema(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_schema (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 292, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_schema") < 0)) __PYX_ERR(2, 292, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_created = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_schema", 0, 0, 1, __pyx_nargs); __PYX_ERR(2, 292, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.AvroType.get_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8AvroType_17get_schema(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_17get_schema(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_get_schema __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_schema", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.created = __pyx_v_created;
  __pyx_t_1 = __pyx_vtabptr_5cavro_AvroType->get_schema(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.AvroType.get_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":313
 *             return self.type_name
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.options.types_str_to_schema:
 *             return json.dumps(self.get_schema())
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_20__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_8AvroType_20__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_8AvroType_19__str__(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_19__str__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "src/type.pxi":314
 * 
 *     def __str__(self):
 *         if self.options.types_str_to_schema:             # <<<<<<<<<<<<<<
 *             return json.dumps(self.get_schema())
 *         else:
 */
  if (__pyx_v_self->options->types_str_to_schema) {

    /* "src/type.pxi":315
 *     def __str__(self):
 *         if self.options.types_str_to_schema:
 *             return json.dumps(self.get_schema())             # <<<<<<<<<<<<<<
 *         else:
 *             return super().__str__()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_json); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_dumps); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->__pyx_vtab)->get_schema(__pyx_v_self, 0, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 315, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/type.pxi":314
 * 
 *     def __str__(self):
 *         if self.options.types_str_to_schema:             # <<<<<<<<<<<<<<
 *             return json.dumps(self.get_schema())
 *         else:
 */
  }

  /* "src/type.pxi":317
 *             return json.dumps(self.get_schema())
 *         else:
 *             return super().__str__()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_AvroType);
    __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_AvroType);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_ptype_5cavro_AvroType))) __PYX_ERR(2, 317, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(2, 317, __pyx_L1_error);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 317, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "src/type.pxi":313
 *             return self.type_name
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.options.types_str_to_schema:
 *             return json.dumps(self.get_schema())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.AvroType.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":59
 *     type_name = NotImplemented
 * 
 *     cdef readonly Options options             # <<<<<<<<<<<<<<
 *     cdef readonly object metadata
 *     cdef readonly tuple[ValueAdapter] value_adapters
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_7options_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_8AvroType_7options_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_8AvroType_7options___get__(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_7options___get__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->options);
  __pyx_r = ((PyObject *)__pyx_v_self->options);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":60
 * 
 *     cdef readonly Options options
 *     cdef readonly object metadata             # <<<<<<<<<<<<<<
 *     cdef readonly tuple[ValueAdapter] value_adapters
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_8metadata_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_8AvroType_8metadata_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_8AvroType_8metadata___get__(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_8metadata___get__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->metadata);
  __pyx_r = __pyx_v_self->metadata;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":61
 *     cdef readonly Options options
 *     cdef readonly object metadata
 *     cdef readonly tuple[ValueAdapter] value_adapters             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_14value_adapters_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_8AvroType_14value_adapters_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_8AvroType_14value_adapters___get__(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_14value_adapters___get__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->value_adapters);
  __pyx_r = __pyx_v_self->value_adapters;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_22__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8AvroType_22__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_22__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8AvroType_22__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_8AvroType_21__reduce_cython__(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_21__reduce_cython__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->metadata);
  __Pyx_GIVEREF(__pyx_v_self->metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.metadata, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_AvroType, (type(self), 0x84c5042, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_AvroType, (type(self), 0x84c5042, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle_AvroType, (type(self), 0x84c5042, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_AvroType, (type(self), 0x84c5042, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_AvroType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.metadata is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_AvroType, (type(self), 0x84c5042, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_AvroType, (type(self), 0x84c5042, None), state
 *     else:
 *         return __pyx_unpickle_AvroType, (type(self), 0x84c5042, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_AvroType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_AvroType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_139219010);
    __Pyx_GIVEREF(__pyx_int_139219010);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_139219010)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.AvroType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_AvroType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_AvroType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8AvroType_24__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8AvroType_24__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_24__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8AvroType_24__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.AvroType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8AvroType_23__setstate_cython__(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8AvroType_23__setstate_cython__(struct __pyx_obj_5cavro_AvroType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_AvroType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_AvroType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_AvroType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_AvroType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_AvroType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.AvroType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":328
 *     cdef readonly frozenset aliases
 * 
 *     def __init__(self, schema, source, parse_namespace):             # <<<<<<<<<<<<<<
 *         cdef Schema schema_t = schema
 *         cdef str effective_namespace = None
 */

/* Python wrapper */
static int __pyx_pw_5cavro_10_NamedType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_10_NamedType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_schema = 0;
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_parse_namespace = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_schema,&__pyx_n_s_source,&__pyx_n_s_parse_namespace,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 328, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 328, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(2, 328, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_parse_namespace)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 328, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(2, 328, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(2, 328, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_schema = values[0];
    __pyx_v_source = values[1];
    __pyx_v_parse_namespace = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(2, 328, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._NamedType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10_NamedType___init__(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self), __pyx_v_schema, __pyx_v_source, __pyx_v_parse_namespace);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_10_NamedType___init__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self, PyObject *__pyx_v_schema, PyObject *__pyx_v_source, PyObject *__pyx_v_parse_namespace) {
  struct __pyx_obj_5cavro_Schema *__pyx_v_schema_t = 0;
  PyObject *__pyx_v_effective_namespace = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_name_pattern = NULL;
  PyObject *__pyx_v_namespace_parts = NULL;
  PyObject *__pyx_v_part = NULL;
  PyObject *__pyx_v_alias_val = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/type.pxi":329
 * 
 *     def __init__(self, schema, source, parse_namespace):
 *         cdef Schema schema_t = schema             # <<<<<<<<<<<<<<
 *         cdef str effective_namespace = None
 *         cdef str name = source['name']
 */
  if (!(likely(((__pyx_v_schema) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_schema, __pyx_ptype_5cavro_Schema))))) __PYX_ERR(2, 329, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_schema;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_schema_t = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":330
 *     def __init__(self, schema, source, parse_namespace):
 *         cdef Schema schema_t = schema
 *         cdef str effective_namespace = None             # <<<<<<<<<<<<<<
 *         cdef str name = source['name']
 *         if '.' in name:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_effective_namespace = ((PyObject*)Py_None);

  /* "src/type.pxi":331
 *         cdef Schema schema_t = schema
 *         cdef str effective_namespace = None
 *         cdef str name = source['name']             # <<<<<<<<<<<<<<
 *         if '.' in name:
 *             effective_namespace, name = name.rsplit('.', 1)
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_source, __pyx_n_u_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(2, 331, __pyx_L1_error)
  __pyx_v_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":332
 *         cdef str effective_namespace = None
 *         cdef str name = source['name']
 *         if '.' in name:             # <<<<<<<<<<<<<<
 *             effective_namespace, name = name.rsplit('.', 1)
 * 
 */
  if (unlikely(__pyx_v_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(2, 332, __pyx_L1_error)
  }
  __pyx_t_2 = (__Pyx_PyUnicode_ContainsTF(__pyx_kp_u__2, __pyx_v_name, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 332, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "src/type.pxi":333
 *         cdef str name = source['name']
 *         if '.' in name:
 *             effective_namespace, name = name.rsplit('.', 1)             # <<<<<<<<<<<<<<
 * 
 *         if not schema_t.options.allow_leading_dot_in_names:
 */
    __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyUnicode_Type_rsplit, __pyx_v_name, __pyx_kp_u__2, __pyx_int_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(2, 333, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 333, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 333, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 333, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5);
      index = 0; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) __PYX_ERR(2, 333, __pyx_L1_error)
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L5_unpacking_done;
      __pyx_L4_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(2, 333, __pyx_L1_error)
      __pyx_L5_unpacking_done:;
    }
    if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(2, 333, __pyx_L1_error)
    if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_4))) __PYX_ERR(2, 333, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_effective_namespace, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "src/type.pxi":332
 *         cdef str effective_namespace = None
 *         cdef str name = source['name']
 *         if '.' in name:             # <<<<<<<<<<<<<<
 *             effective_namespace, name = name.rsplit('.', 1)
 * 
 */
  }

  /* "src/type.pxi":335
 *             effective_namespace, name = name.rsplit('.', 1)
 * 
 *         if not schema_t.options.allow_leading_dot_in_names:             # <<<<<<<<<<<<<<
 *             if effective_namespace == '':
 *                 raise InvalidName(f"The null namespace cannot be specified in name fields")
 */
  __pyx_t_2 = (!__pyx_v_schema_t->options->allow_leading_dot_in_names);
  if (__pyx_t_2) {

    /* "src/type.pxi":336
 * 
 *         if not schema_t.options.allow_leading_dot_in_names:
 *             if effective_namespace == '':             # <<<<<<<<<<<<<<
 *                 raise InvalidName(f"The null namespace cannot be specified in name fields")
 * 
 */
    __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_effective_namespace, __pyx_kp_u__13, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 336, __pyx_L1_error)
    if (unlikely(__pyx_t_2)) {

      /* "src/type.pxi":337
 *         if not schema_t.options.allow_leading_dot_in_names:
 *             if effective_namespace == '':
 *                 raise InvalidName(f"The null namespace cannot be specified in name fields")             # <<<<<<<<<<<<<<
 * 
 *         name_pattern = schema_t.options.name_pattern
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidName); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_The_null_namespace_cannot_be_spe};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 337, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(2, 337, __pyx_L1_error)

      /* "src/type.pxi":336
 * 
 *         if not schema_t.options.allow_leading_dot_in_names:
 *             if effective_namespace == '':             # <<<<<<<<<<<<<<
 *                 raise InvalidName(f"The null namespace cannot be specified in name fields")
 * 
 */
    }

    /* "src/type.pxi":335
 *             effective_namespace, name = name.rsplit('.', 1)
 * 
 *         if not schema_t.options.allow_leading_dot_in_names:             # <<<<<<<<<<<<<<
 *             if effective_namespace == '':
 *                 raise InvalidName(f"The null namespace cannot be specified in name fields")
 */
  }

  /* "src/type.pxi":339
 *                 raise InvalidName(f"The null namespace cannot be specified in name fields")
 * 
 *         name_pattern = schema_t.options.name_pattern             # <<<<<<<<<<<<<<
 *         if schema_t.options.enforce_type_name_rules:
 *             if not name_pattern.fullmatch(name):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_schema_t->options), __pyx_n_s_name_pattern); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_name_pattern = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/type.pxi":340
 * 
 *         name_pattern = schema_t.options.name_pattern
 *         if schema_t.options.enforce_type_name_rules:             # <<<<<<<<<<<<<<
 *             if not name_pattern.fullmatch(name):
 *                 raise InvalidName(f"Type name '{name}' is not valid")
 */
  if (__pyx_v_schema_t->options->enforce_type_name_rules) {

    /* "src/type.pxi":341
 *         name_pattern = schema_t.options.name_pattern
 *         if schema_t.options.enforce_type_name_rules:
 *             if not name_pattern.fullmatch(name):             # <<<<<<<<<<<<<<
 *                 raise InvalidName(f"Type name '{name}' is not valid")
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_name_pattern, __pyx_n_s_fullmatch); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_name};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 341, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 341, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = (!__pyx_t_2);
    if (unlikely(__pyx_t_8)) {

      /* "src/type.pxi":342
 *         if schema_t.options.enforce_type_name_rules:
 *             if not name_pattern.fullmatch(name):
 *                 raise InvalidName(f"Type name '{name}' is not valid")             # <<<<<<<<<<<<<<
 * 
 *         self.name = name
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InvalidName); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;
      __Pyx_INCREF(__pyx_kp_u_Type_name);
      __pyx_t_9 += 11;
      __Pyx_GIVEREF(__pyx_kp_u_Type_name);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Type_name);
      __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u_is_not_valid);
      __pyx_t_9 += 14;
      __Pyx_GIVEREF(__pyx_kp_u_is_not_valid);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_is_not_valid);
      __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 342, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(2, 342, __pyx_L1_error)

      /* "src/type.pxi":341
 *         name_pattern = schema_t.options.name_pattern
 *         if schema_t.options.enforce_type_name_rules:
 *             if not name_pattern.fullmatch(name):             # <<<<<<<<<<<<<<
 *                 raise InvalidName(f"Type name '{name}' is not valid")
 * 
 */
    }

    /* "src/type.pxi":340
 * 
 *         name_pattern = schema_t.options.name_pattern
 *         if schema_t.options.enforce_type_name_rules:             # <<<<<<<<<<<<<<
 *             if not name_pattern.fullmatch(name):
 *                 raise InvalidName(f"Type name '{name}' is not valid")
 */
  }

  /* "src/type.pxi":344
 *                 raise InvalidName(f"Type name '{name}' is not valid")
 * 
 *         self.name = name             # <<<<<<<<<<<<<<
 *         self.namespace = source.get('namespace')
 *         if effective_namespace is None:
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "src/type.pxi":345
 * 
 *         self.name = name
 *         self.namespace = source.get('namespace')             # <<<<<<<<<<<<<<
 *         if effective_namespace is None:
 *             effective_namespace = self.namespace
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_n_u_namespace};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(2, 345, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->namespace);
  __Pyx_DECREF(__pyx_v_self->namespace);
  __pyx_v_self->namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":346
 *         self.name = name
 *         self.namespace = source.get('namespace')
 *         if effective_namespace is None:             # <<<<<<<<<<<<<<
 *             effective_namespace = self.namespace
 *             if effective_namespace is None:
 */
  __pyx_t_8 = (__pyx_v_effective_namespace == ((PyObject*)Py_None));
  if (__pyx_t_8) {

    /* "src/type.pxi":347
 *         self.namespace = source.get('namespace')
 *         if effective_namespace is None:
 *             effective_namespace = self.namespace             # <<<<<<<<<<<<<<
 *             if effective_namespace is None:
 *                 effective_namespace = parse_namespace
 */
    __pyx_t_1 = __pyx_v_self->namespace;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_effective_namespace, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "src/type.pxi":348
 *         if effective_namespace is None:
 *             effective_namespace = self.namespace
 *             if effective_namespace is None:             # <<<<<<<<<<<<<<
 *                 effective_namespace = parse_namespace
 * 
 */
    __pyx_t_8 = (__pyx_v_effective_namespace == ((PyObject*)Py_None));
    if (__pyx_t_8) {

      /* "src/type.pxi":349
 *             effective_namespace = self.namespace
 *             if effective_namespace is None:
 *                 effective_namespace = parse_namespace             # <<<<<<<<<<<<<<
 * 
 *         self.effective_namespace = effective_namespace or None
 */
      if (!(likely(PyUnicode_CheckExact(__pyx_v_parse_namespace))||((__pyx_v_parse_namespace) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_parse_namespace))) __PYX_ERR(2, 349, __pyx_L1_error)
      __pyx_t_1 = __pyx_v_parse_namespace;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_effective_namespace, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "src/type.pxi":348
 *         if effective_namespace is None:
 *             effective_namespace = self.namespace
 *             if effective_namespace is None:             # <<<<<<<<<<<<<<
 *                 effective_namespace = parse_namespace
 * 
 */
    }

    /* "src/type.pxi":346
 *         self.name = name
 *         self.namespace = source.get('namespace')
 *         if effective_namespace is None:             # <<<<<<<<<<<<<<
 *             effective_namespace = self.namespace
 *             if effective_namespace is None:
 */
  }

  /* "src/type.pxi":351
 *                 effective_namespace = parse_namespace
 * 
 *         self.effective_namespace = effective_namespace or None             # <<<<<<<<<<<<<<
 * 
 *         if self.effective_namespace is not None and schema_t.options.enforce_namespace_name_rules:
 */
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_effective_namespace); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(2, 351, __pyx_L1_error)
  if (!__pyx_t_8) {
  } else {
    __Pyx_INCREF(__pyx_v_effective_namespace);
    __pyx_t_1 = __pyx_v_effective_namespace;
    goto __pyx_L12_bool_binop_done;
  }
  __Pyx_INCREF(Py_None);
  __pyx_t_1 = Py_None;
  __pyx_L12_bool_binop_done:;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->effective_namespace);
  __Pyx_DECREF(__pyx_v_self->effective_namespace);
  __pyx_v_self->effective_namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":353
 *         self.effective_namespace = effective_namespace or None
 * 
 *         if self.effective_namespace is not None and schema_t.options.enforce_namespace_name_rules:             # <<<<<<<<<<<<<<
 *             namespace_parts = self.effective_namespace.split('.')
 *             for part in namespace_parts:
 */
  __pyx_t_2 = (__pyx_v_self->effective_namespace != ((PyObject*)Py_None));
  if (__pyx_t_2) {
  } else {
    __pyx_t_8 = __pyx_t_2;
    goto __pyx_L15_bool_binop_done;
  }
  __pyx_t_8 = __pyx_v_schema_t->options->enforce_namespace_name_rules;
  __pyx_L15_bool_binop_done:;
  if (__pyx_t_8) {

    /* "src/type.pxi":354
 * 
 *         if self.effective_namespace is not None and schema_t.options.enforce_namespace_name_rules:
 *             namespace_parts = self.effective_namespace.split('.')             # <<<<<<<<<<<<<<
 *             for part in namespace_parts:
 *                 if not part or not name_pattern.fullmatch(part):
 */
    if (unlikely(__pyx_v_self->effective_namespace == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "split");
      __PYX_ERR(2, 354, __pyx_L1_error)
    }
    __pyx_t_1 = PyUnicode_Split(__pyx_v_self->effective_namespace, __Pyx_NoneAsNull(__pyx_kp_u__2), -1L); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_namespace_parts = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "src/type.pxi":355
 *         if self.effective_namespace is not None and schema_t.options.enforce_namespace_name_rules:
 *             namespace_parts = self.effective_namespace.split('.')
 *             for part in namespace_parts:             # <<<<<<<<<<<<<<
 *                 if not part or not name_pattern.fullmatch(part):
 *                     raise InvalidName(f"Namespace '{self.effective_namespace}' is not valid")
 */
    if (likely(PyList_CheckExact(__pyx_v_namespace_parts)) || PyTuple_CheckExact(__pyx_v_namespace_parts)) {
      __pyx_t_1 = __pyx_v_namespace_parts; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_9 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_namespace_parts); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 355, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 355, __pyx_L1_error)
            #endif
            if (__pyx_t_9 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(2, 355, __pyx_L1_error)
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 355, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(2, 355, __pyx_L1_error)
            #endif
            if (__pyx_t_9 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(2, 355, __pyx_L1_error)
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 355, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_11(__pyx_t_1);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(2, 355, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_v_part, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "src/type.pxi":356
 *             namespace_parts = self.effective_namespace.split('.')
 *             for part in namespace_parts:
 *                 if not part or not name_pattern.fullmatch(part):             # <<<<<<<<<<<<<<
 *                     raise InvalidName(f"Namespace '{self.effective_namespace}' is not valid")
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_part); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 356, __pyx_L1_error)
      __pyx_t_12 = (!__pyx_t_2);
      if (!__pyx_t_12) {
      } else {
        __pyx_t_8 = __pyx_t_12;
        goto __pyx_L20_bool_binop_done;
      }
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_name_pattern, __pyx_n_s_fullmatch); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 356, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_part};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 356, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(2, 356, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = (!__pyx_t_12);
      __pyx_t_8 = __pyx_t_2;
      __pyx_L20_bool_binop_done:;
      if (unlikely(__pyx_t_8)) {

        /* "src/type.pxi":357
 *             for part in namespace_parts:
 *                 if not part or not name_pattern.fullmatch(part):
 *                     raise InvalidName(f"Namespace '{self.effective_namespace}' is not valid")             # <<<<<<<<<<<<<<
 * 
 *         if not schema_t.options.allow_primitive_name_collision:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_InvalidName); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 357, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 357, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_13 = 0;
        __pyx_t_10 = 127;
        __Pyx_INCREF(__pyx_kp_u_Namespace);
        __pyx_t_13 += 11;
        __Pyx_GIVEREF(__pyx_kp_u_Namespace);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Namespace);
        __pyx_t_14 = __Pyx_PyUnicode_Unicode(__pyx_v_self->effective_namespace); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 357, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) : __pyx_t_10;
        __pyx_t_13 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14);
        __Pyx_GIVEREF(__pyx_t_14);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_14);
        __pyx_t_14 = 0;
        __Pyx_INCREF(__pyx_kp_u_is_not_valid);
        __pyx_t_13 += 14;
        __Pyx_GIVEREF(__pyx_kp_u_is_not_valid);
        PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_is_not_valid);
        __pyx_t_14 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_13, __pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 357, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_14};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 357, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 357, __pyx_L1_error)

        /* "src/type.pxi":356
 *             namespace_parts = self.effective_namespace.split('.')
 *             for part in namespace_parts:
 *                 if not part or not name_pattern.fullmatch(part):             # <<<<<<<<<<<<<<
 *                     raise InvalidName(f"Namespace '{self.effective_namespace}' is not valid")
 * 
 */
      }

      /* "src/type.pxi":355
 *         if self.effective_namespace is not None and schema_t.options.enforce_namespace_name_rules:
 *             namespace_parts = self.effective_namespace.split('.')
 *             for part in namespace_parts:             # <<<<<<<<<<<<<<
 *                 if not part or not name_pattern.fullmatch(part):
 *                     raise InvalidName(f"Namespace '{self.effective_namespace}' is not valid")
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/type.pxi":353
 *         self.effective_namespace = effective_namespace or None
 * 
 *         if self.effective_namespace is not None and schema_t.options.enforce_namespace_name_rules:             # <<<<<<<<<<<<<<
 *             namespace_parts = self.effective_namespace.split('.')
 *             for part in namespace_parts:
 */
  }

  /* "src/type.pxi":359
 *                     raise InvalidName(f"Namespace '{self.effective_namespace}' is not valid")
 * 
 *         if not schema_t.options.allow_primitive_name_collision:             # <<<<<<<<<<<<<<
 *             if name in PRIMITIVE_TYPES:
 *                 if schema_t.options.allow_primitive_names_in_namespaces and self.effective_namespace is not None:
 */
  __pyx_t_8 = (!__pyx_v_schema_t->options->allow_primitive_name_collision);
  if (__pyx_t_8) {

    /* "src/type.pxi":360
 * 
 *         if not schema_t.options.allow_primitive_name_collision:
 *             if name in PRIMITIVE_TYPES:             # <<<<<<<<<<<<<<
 *                 if schema_t.options.allow_primitive_names_in_namespaces and self.effective_namespace is not None:
 *                     pass
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PRIMITIVE_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 360, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_v_name, __pyx_t_1, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(2, 360, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_8) {

      /* "src/type.pxi":361
 *         if not schema_t.options.allow_primitive_name_collision:
 *             if name in PRIMITIVE_TYPES:
 *                 if schema_t.options.allow_primitive_names_in_namespaces and self.effective_namespace is not None:             # <<<<<<<<<<<<<<
 *                     pass
 *                 else:
 */
      if (__pyx_v_schema_t->options->allow_primitive_names_in_namespaces) {
      } else {
        __pyx_t_8 = __pyx_v_schema_t->options->allow_primitive_names_in_namespaces;
        goto __pyx_L26_bool_binop_done;
      }
      __pyx_t_2 = (__pyx_v_self->effective_namespace != ((PyObject*)Py_None));
      __pyx_t_8 = __pyx_t_2;
      __pyx_L26_bool_binop_done:;
      if (likely(__pyx_t_8)) {
        goto __pyx_L25;
      }

      /* "src/type.pxi":364
 *                     pass
 *                 else:
 *                     raise ValueError(f"'{name}' is not allowed as a name")             # <<<<<<<<<<<<<<
 * 
 *         alias_val = source.get('aliases', [])
 */
      /*else*/ {
        __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = 0;
        __pyx_t_10 = 127;
        __Pyx_INCREF(__pyx_kp_u__6);
        __pyx_t_9 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__6);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u__6);
        __pyx_t_4 = __Pyx_PyUnicode_Unicode(__pyx_v_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_10;
        __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
        __pyx_t_4 = 0;
        __Pyx_INCREF(__pyx_kp_u_is_not_allowed_as_a_name);
        __pyx_t_9 += 26;
        __Pyx_GIVEREF(__pyx_kp_u_is_not_allowed_as_a_name);
        PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_is_not_allowed_as_a_name);
        __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(2, 364, __pyx_L1_error)
      }
      __pyx_L25:;

      /* "src/type.pxi":360
 * 
 *         if not schema_t.options.allow_primitive_name_collision:
 *             if name in PRIMITIVE_TYPES:             # <<<<<<<<<<<<<<
 *                 if schema_t.options.allow_primitive_names_in_namespaces and self.effective_namespace is not None:
 *                     pass
 */
    }

    /* "src/type.pxi":359
 *                     raise InvalidName(f"Namespace '{self.effective_namespace}' is not valid")
 * 
 *         if not schema_t.options.allow_primitive_name_collision:             # <<<<<<<<<<<<<<
 *             if name in PRIMITIVE_TYPES:
 *                 if schema_t.options.allow_primitive_names_in_namespaces and self.effective_namespace is not None:
 */
  }

  /* "src/type.pxi":366
 *                     raise ValueError(f"'{name}' is not allowed as a name")
 * 
 *         alias_val = source.get('aliases', [])             # <<<<<<<<<<<<<<
 *         if not isinstance(alias_val, (list, tuple, set)):
 *             if schema_t.options.allow_aliases_to_be_string:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_source, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_14 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_14)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_14, __pyx_n_u_aliases, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_v_alias_val = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/type.pxi":367
 * 
 *         alias_val = source.get('aliases', [])
 *         if not isinstance(alias_val, (list, tuple, set)):             # <<<<<<<<<<<<<<
 *             if schema_t.options.allow_aliases_to_be_string:
 *                 alias_val = [alias_val]
 */
  __pyx_t_2 = PyList_Check(__pyx_v_alias_val); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_8 = __pyx_t_2;
    goto __pyx_L29_bool_binop_done;
  }
  __pyx_t_2 = PyTuple_Check(__pyx_v_alias_val); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_8 = __pyx_t_2;
    goto __pyx_L29_bool_binop_done;
  }
  __pyx_t_2 = PySet_Check(__pyx_v_alias_val); 
  __pyx_t_8 = __pyx_t_2;
  __pyx_L29_bool_binop_done:;
  __pyx_t_2 = (!__pyx_t_8);
  if (__pyx_t_2) {

    /* "src/type.pxi":368
 *         alias_val = source.get('aliases', [])
 *         if not isinstance(alias_val, (list, tuple, set)):
 *             if schema_t.options.allow_aliases_to_be_string:             # <<<<<<<<<<<<<<
 *                 alias_val = [alias_val]
 *             else:
 */
    if (likely(__pyx_v_schema_t->options->allow_aliases_to_be_string)) {

      /* "src/type.pxi":369
 *         if not isinstance(alias_val, (list, tuple, set)):
 *             if schema_t.options.allow_aliases_to_be_string:
 *                 alias_val = [alias_val]             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(f"Aliases must be a list/tuple/set, got: {repr(alias_val)}")
 */
      __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 369, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_alias_val);
      __Pyx_GIVEREF(__pyx_v_alias_val);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_alias_val)) __PYX_ERR(2, 369, __pyx_L1_error);
      __Pyx_DECREF_SET(__pyx_v_alias_val, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "src/type.pxi":368
 *         alias_val = source.get('aliases', [])
 *         if not isinstance(alias_val, (list, tuple, set)):
 *             if schema_t.options.allow_aliases_to_be_string:             # <<<<<<<<<<<<<<
 *                 alias_val = [alias_val]
 *             else:
 */
      goto __pyx_L32;
    }

    /* "src/type.pxi":371
 *                 alias_val = [alias_val]
 *             else:
 *                 raise ValueError(f"Aliases must be a list/tuple/set, got: {repr(alias_val)}")             # <<<<<<<<<<<<<<
 *         self.aliases = frozenset(alias_val)
 *         super().__init__(schema, source, parse_namespace)
 */
    /*else*/ {
      __pyx_t_1 = PyObject_Repr(__pyx_v_alias_val); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Aliases_must_be_a_list_tuple_set, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(2, 371, __pyx_L1_error)
    }
    __pyx_L32:;

    /* "src/type.pxi":367
 * 
 *         alias_val = source.get('aliases', [])
 *         if not isinstance(alias_val, (list, tuple, set)):             # <<<<<<<<<<<<<<
 *             if schema_t.options.allow_aliases_to_be_string:
 *                 alias_val = [alias_val]
 */
  }

  /* "src/type.pxi":372
 *             else:
 *                 raise ValueError(f"Aliases must be a list/tuple/set, got: {repr(alias_val)}")
 *         self.aliases = frozenset(alias_val)             # <<<<<<<<<<<<<<
 *         super().__init__(schema, source, parse_namespace)
 *         schema_t.register_type(self.effective_namespace, self.name, self)
 */
  __pyx_t_4 = __Pyx_PyFrozenSet_New(__pyx_v_alias_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->aliases);
  __Pyx_DECREF(__pyx_v_self->aliases);
  __pyx_v_self->aliases = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "src/type.pxi":373
 *                 raise ValueError(f"Aliases must be a list/tuple/set, got: {repr(alias_val)}")
 *         self.aliases = frozenset(alias_val)
 *         super().__init__(schema, source, parse_namespace)             # <<<<<<<<<<<<<<
 *         schema_t.register_type(self.effective_namespace, self.name, self)
 * 
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro__NamedType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro__NamedType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_ptype_5cavro__NamedType))) __PYX_ERR(2, 373, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(2, 373, __pyx_L1_error);
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_1, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_v_schema, __pyx_v_source, __pyx_v_parse_namespace};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_7, 3+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/type.pxi":374
 *         self.aliases = frozenset(alias_val)
 *         super().__init__(schema, source, parse_namespace)
 *         schema_t.register_type(self.effective_namespace, self.name, self)             # <<<<<<<<<<<<<<
 * 
 *     property type:
 */
  __pyx_t_4 = __pyx_v_self->effective_namespace;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_1 = __pyx_v_self->name;
  __Pyx_INCREF(__pyx_t_1);
  ((struct __pyx_vtabstruct_5cavro_Schema *)__pyx_v_schema_t->__pyx_vtab)->register_type(__pyx_v_schema_t, ((PyObject*)__pyx_t_4), ((PyObject*)__pyx_t_1), ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/type.pxi":328
 *     cdef readonly frozenset aliases
 * 
 *     def __init__(self, schema, source, parse_namespace):             # <<<<<<<<<<<<<<
 *         cdef Schema schema_t = schema
 *         cdef str effective_namespace = None
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("cavro._NamedType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_schema_t);
  __Pyx_XDECREF(__pyx_v_effective_namespace);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_name_pattern);
  __Pyx_XDECREF(__pyx_v_namespace_parts);
  __Pyx_XDECREF(__pyx_v_part);
  __Pyx_XDECREF(__pyx_v_alias_val);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":377
 * 
 *     property type:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return resolve_namespaced_name(self.effective_namespace, self.name)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10_NamedType_4type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_10_NamedType_4type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_10_NamedType_4type___get__(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10_NamedType_4type___get__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/type.pxi":378
 *     property type:
 *         def __get__(self):
 *             return resolve_namespaced_name(self.effective_namespace, self.name)             # <<<<<<<<<<<<<<
 * 
 *     cdef bint name_matches(self, _NamedType other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->effective_namespace;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_self->name;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_5cavro_resolve_namespaced_name(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/type.pxi":377
 * 
 *     property type:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return resolve_namespaced_name(self.effective_namespace, self.name)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro._NamedType.type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":380
 *             return resolve_namespaced_name(self.effective_namespace, self.name)
 * 
 *     cdef bint name_matches(self, _NamedType other):             # <<<<<<<<<<<<<<
 *         if self.name == other.name:
 *             return True
 */

static int __pyx_f_5cavro_10_NamedType_name_matches(struct __pyx_obj_5cavro__NamedType *__pyx_v_self, struct __pyx_obj_5cavro__NamedType *__pyx_v_other) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("name_matches", 1);

  /* "src/type.pxi":381
 * 
 *     cdef bint name_matches(self, _NamedType other):
 *         if self.name == other.name:             # <<<<<<<<<<<<<<
 *             return True
 *         if other.name in self.aliases:
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->name, __pyx_v_other->name, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(2, 381, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/type.pxi":382
 *     cdef bint name_matches(self, _NamedType other):
 *         if self.name == other.name:
 *             return True             # <<<<<<<<<<<<<<
 *         if other.name in self.aliases:
 *             return True
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "src/type.pxi":381
 * 
 *     cdef bint name_matches(self, _NamedType other):
 *         if self.name == other.name:             # <<<<<<<<<<<<<<
 *             return True
 *         if other.name in self.aliases:
 */
  }

  /* "src/type.pxi":383
 *         if self.name == other.name:
 *             return True
 *         if other.name in self.aliases:             # <<<<<<<<<<<<<<
 *             return True
 *         # Let's get creative
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_other->name, __pyx_v_self->aliases, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(2, 383, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/type.pxi":384
 *             return True
 *         if other.name in self.aliases:
 *             return True             # <<<<<<<<<<<<<<
 *         # Let's get creative
 *         if other.effective_namespace and other.type in self.aliases:
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "src/type.pxi":383
 *         if self.name == other.name:
 *             return True
 *         if other.name in self.aliases:             # <<<<<<<<<<<<<<
 *             return True
 *         # Let's get creative
 */
  }

  /* "src/type.pxi":386
 *             return True
 *         # Let's get creative
 *         if other.effective_namespace and other.type in self.aliases:             # <<<<<<<<<<<<<<
 *             return True
 *         return False
 */
  __pyx_t_2 = (__pyx_v_other->effective_namespace != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_v_other->effective_namespace) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_v_self->aliases, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/type.pxi":387
 *         # Let's get creative
 *         if other.effective_namespace and other.type in self.aliases:
 *             return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "src/type.pxi":386
 *             return True
 *         # Let's get creative
 *         if other.effective_namespace and other.type in self.aliases:             # <<<<<<<<<<<<<<
 *             return True
 *         return False
 */
  }

  /* "src/type.pxi":388
 *         if other.effective_namespace and other.type in self.aliases:
 *             return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cdef frozenset get_namespaced_aliases(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "src/type.pxi":380
 *             return resolve_namespaced_name(self.effective_namespace, self.name)
 * 
 *     cdef bint name_matches(self, _NamedType other):             # <<<<<<<<<<<<<<
 *         if self.name == other.name:
 *             return True
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro._NamedType.name_matches", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5cavro_10_NamedType_22get_namespaced_aliases_2generator19(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "src/type.pxi":394
 *         if not self.effective_namespace:
 *             return all_names
 *         return frozenset(resolve_namespaced_name(self.effective_namespace, alias) for alias in all_names)             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */

static PyObject *__pyx_pf_5cavro_10_NamedType_22get_namespaced_aliases_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr *)__pyx_tp_new_5cavro___pyx_scope_struct_23_genexpr(__pyx_ptype_5cavro___pyx_scope_struct_23_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(2, 394, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5cavro_10_NamedType_22get_namespaced_aliases_2generator19, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_NamedType_get_namespaced_aliase, __pyx_n_s_cavro); if (unlikely(!gen)) __PYX_ERR(2, 394, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro._NamedType.get_namespaced_aliases.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5cavro_10_NamedType_22get_namespaced_aliases_2generator19(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(2, 394, __pyx_L1_error)
  __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(2, 394, __pyx_L1_error) }
  __pyx_t_5 = __Pyx_set_iterator(__pyx_cur_scope->__pyx_genexpr_arg_0, 0, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_set_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(2, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_alias);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_alias, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(2, 394, __pyx_L1_error) }
    __pyx_t_5 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->effective_namespace;
    __Pyx_INCREF(__pyx_t_5);
    if (!(likely(PyUnicode_CheckExact(__pyx_cur_scope->__pyx_v_alias))||((__pyx_cur_scope->__pyx_v_alias) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_cur_scope->__pyx_v_alias))) __PYX_ERR(2, 394, __pyx_L1_error)
    __pyx_t_7 = __pyx_f_5cavro_resolve_namespaced_name(((PyObject*)__pyx_t_5), ((PyObject*)__pyx_cur_scope->__pyx_v_alias)); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(2, 394, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":390
 *         return False
 * 
 *     cdef frozenset get_namespaced_aliases(self):             # <<<<<<<<<<<<<<
 *         cdef frozenset all_names = self.aliases | {self.name}
 *         if not self.effective_namespace:
 */

static PyObject *__pyx_f_5cavro_10_NamedType_get_namespaced_aliases(struct __pyx_obj_5cavro__NamedType *__pyx_v_self) {
  struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *__pyx_cur_scope;
  PyObject *__pyx_v_all_names = 0;
  PyObject *__pyx_gb_5cavro_10_NamedType_22get_namespaced_aliases_2generator19 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_namespaced_aliases", 0);
  __pyx_cur_scope = (struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *)__pyx_tp_new_5cavro___pyx_scope_struct_22_get_namespaced_aliases(__pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(2, 390, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "src/type.pxi":391
 * 
 *     cdef frozenset get_namespaced_aliases(self):
 *         cdef frozenset all_names = self.aliases | {self.name}             # <<<<<<<<<<<<<<
 *         if not self.effective_namespace:
 *             return all_names
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PySet_Add(__pyx_t_1, __pyx_cur_scope->__pyx_v_self->name) < 0) __PYX_ERR(2, 391, __pyx_L1_error)
  __pyx_t_2 = PyNumber_Or(__pyx_cur_scope->__pyx_v_self->aliases, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyFrozenSet_CheckExact(__pyx_t_2)) || __Pyx_RaiseUnexpectedTypeError("frozenset", __pyx_t_2))) __PYX_ERR(2, 391, __pyx_L1_error)
  __pyx_v_all_names = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/type.pxi":392
 *     cdef frozenset get_namespaced_aliases(self):
 *         cdef frozenset all_names = self.aliases | {self.name}
 *         if not self.effective_namespace:             # <<<<<<<<<<<<<<
 *             return all_names
 *         return frozenset(resolve_namespaced_name(self.effective_namespace, alias) for alias in all_names)
 */
  __pyx_t_3 = (__pyx_cur_scope->__pyx_v_self->effective_namespace != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_cur_scope->__pyx_v_self->effective_namespace) != 0);
  __pyx_t_4 = (!__pyx_t_3);
  if (__pyx_t_4) {

    /* "src/type.pxi":393
 *         cdef frozenset all_names = self.aliases | {self.name}
 *         if not self.effective_namespace:
 *             return all_names             # <<<<<<<<<<<<<<
 *         return frozenset(resolve_namespaced_name(self.effective_namespace, alias) for alias in all_names)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_all_names);
    __pyx_r = __pyx_v_all_names;
    goto __pyx_L0;

    /* "src/type.pxi":392
 *     cdef frozenset get_namespaced_aliases(self):
 *         cdef frozenset all_names = self.aliases | {self.name}
 *         if not self.effective_namespace:             # <<<<<<<<<<<<<<
 *             return all_names
 *         return frozenset(resolve_namespaced_name(self.effective_namespace, alias) for alias in all_names)
 */
  }

  /* "src/type.pxi":394
 *         if not self.effective_namespace:
 *             return all_names
 *         return frozenset(resolve_namespaced_name(self.effective_namespace, alias) for alias in all_names)             # <<<<<<<<<<<<<<
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_pf_5cavro_10_NamedType_22get_namespaced_aliases_genexpr(((PyObject*)__pyx_cur_scope), __pyx_v_all_names); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyFrozenSet_New(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/type.pxi":390
 *         return False
 * 
 *     cdef frozenset get_namespaced_aliases(self):             # <<<<<<<<<<<<<<
 *         cdef frozenset all_names = self.aliases | {self.name}
 *         if not self.effective_namespace:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro._NamedType.get_namespaced_aliases", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_all_names);
  __Pyx_XDECREF(__pyx_gb_5cavro_10_NamedType_22get_namespaced_aliases_2generator19);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":396
 *         return frozenset(resolve_namespaced_name(self.effective_namespace, alias) for alias in all_names)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         schema = {}
 *         if self.effective_namespace:
 */

static PyObject *__pyx_pw_5cavro_10_NamedType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_10_NamedType__get_schema_extra(struct __pyx_obj_5cavro__NamedType *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_created, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_schema = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_schema_extra); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_10_NamedType_3_get_schema_extra)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_created};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 396, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_2))) __PYX_ERR(2, 396, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/type.pxi":397
 * 
 *     cpdef dict _get_schema_extra(self, set created):
 *         schema = {}             # <<<<<<<<<<<<<<
 *         if self.effective_namespace:
 *             if self.options.inline_namespaces:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_schema = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":398
 *     cpdef dict _get_schema_extra(self, set created):
 *         schema = {}
 *         if self.effective_namespace:             # <<<<<<<<<<<<<<
 *             if self.options.inline_namespaces:
 *                 schema['name'] = f"{self.effective_namespace}.{self.name}"
 */
  __pyx_t_6 = (__pyx_v_self->effective_namespace != Py_None)&&(__Pyx_PyUnicode_IS_TRUE(__pyx_v_self->effective_namespace) != 0);
  if (__pyx_t_6) {

    /* "src/type.pxi":399
 *         schema = {}
 *         if self.effective_namespace:
 *             if self.options.inline_namespaces:             # <<<<<<<<<<<<<<
 *                 schema['name'] = f"{self.effective_namespace}.{self.name}"
 *             else:
 */
    if (__pyx_v_self->__pyx_base.options->inline_namespaces) {

      /* "src/type.pxi":400
 *         if self.effective_namespace:
 *             if self.options.inline_namespaces:
 *                 schema['name'] = f"{self.effective_namespace}.{self.name}"             # <<<<<<<<<<<<<<
 *             else:
 *                 schema['namespace'] = self.effective_namespace
 */
      __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 400, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = 0;
      __pyx_t_8 = 127;
      __pyx_t_2 = __Pyx_PyUnicode_Unicode(__pyx_v_self->effective_namespace); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 400, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_8;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_INCREF(__pyx_kp_u__2);
      __pyx_t_7 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__2);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__2);
      __pyx_t_2 = __Pyx_PyUnicode_Unicode(__pyx_v_self->name); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 400, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_8;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 400, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely((PyDict_SetItem(__pyx_v_schema, __pyx_n_u_name, __pyx_t_2) < 0))) __PYX_ERR(2, 400, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "src/type.pxi":399
 *         schema = {}
 *         if self.effective_namespace:
 *             if self.options.inline_namespaces:             # <<<<<<<<<<<<<<
 *                 schema['name'] = f"{self.effective_namespace}.{self.name}"
 *             else:
 */
      goto __pyx_L4;
    }

    /* "src/type.pxi":402
 *                 schema['name'] = f"{self.effective_namespace}.{self.name}"
 *             else:
 *                 schema['namespace'] = self.effective_namespace             # <<<<<<<<<<<<<<
 *         if self.aliases:
 *             schema['aliases'] = list(self.aliases)
 */
    /*else*/ {
      __pyx_t_2 = __pyx_v_self->effective_namespace;
      __Pyx_INCREF(__pyx_t_2);
      if (unlikely((PyDict_SetItem(__pyx_v_schema, __pyx_n_u_namespace, __pyx_t_2) < 0))) __PYX_ERR(2, 402, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L4:;

    /* "src/type.pxi":398
 *     cpdef dict _get_schema_extra(self, set created):
 *         schema = {}
 *         if self.effective_namespace:             # <<<<<<<<<<<<<<
 *             if self.options.inline_namespaces:
 *                 schema['name'] = f"{self.effective_namespace}.{self.name}"
 */
  }

  /* "src/type.pxi":403
 *             else:
 *                 schema['namespace'] = self.effective_namespace
 *         if self.aliases:             # <<<<<<<<<<<<<<
 *             schema['aliases'] = list(self.aliases)
 *         if 'name' not in schema:
 */
  __pyx_t_6 = (__pyx_v_self->aliases != Py_None)&&(PySet_GET_SIZE(__pyx_v_self->aliases) != 0);
  if (__pyx_t_6) {

    /* "src/type.pxi":404
 *                 schema['namespace'] = self.effective_namespace
 *         if self.aliases:
 *             schema['aliases'] = list(self.aliases)             # <<<<<<<<<<<<<<
 *         if 'name' not in schema:
 *             schema['name'] = self.name
 */
    __pyx_t_2 = PySequence_List(__pyx_v_self->aliases); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely((PyDict_SetItem(__pyx_v_schema, __pyx_n_u_aliases, __pyx_t_2) < 0))) __PYX_ERR(2, 404, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/type.pxi":403
 *             else:
 *                 schema['namespace'] = self.effective_namespace
 *         if self.aliases:             # <<<<<<<<<<<<<<
 *             schema['aliases'] = list(self.aliases)
 *         if 'name' not in schema:
 */
  }

  /* "src/type.pxi":405
 *         if self.aliases:
 *             schema['aliases'] = list(self.aliases)
 *         if 'name' not in schema:             # <<<<<<<<<<<<<<
 *             schema['name'] = self.name
 *         return schema
 */
  __pyx_t_6 = (__Pyx_PyDict_ContainsTF(__pyx_n_u_name, __pyx_v_schema, Py_NE)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(2, 405, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "src/type.pxi":406
 *             schema['aliases'] = list(self.aliases)
 *         if 'name' not in schema:
 *             schema['name'] = self.name             # <<<<<<<<<<<<<<
 *         return schema
 * 
 */
    __pyx_t_2 = __pyx_v_self->name;
    __Pyx_INCREF(__pyx_t_2);
    if (unlikely((PyDict_SetItem(__pyx_v_schema, __pyx_n_u_name, __pyx_t_2) < 0))) __PYX_ERR(2, 406, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/type.pxi":405
 *         if self.aliases:
 *             schema['aliases'] = list(self.aliases)
 *         if 'name' not in schema:             # <<<<<<<<<<<<<<
 *             schema['name'] = self.name
 *         return schema
 */
  }

  /* "src/type.pxi":407
 *         if 'name' not in schema:
 *             schema['name'] = self.name
 *         return schema             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType clone_base(self, cls=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_schema);
  __pyx_r = __pyx_v_schema;
  goto __pyx_L0;

  /* "src/type.pxi":396
 *         return frozenset(resolve_namespaced_name(self.effective_namespace, alias) for alias in all_names)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         schema = {}
 *         if self.effective_namespace:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._NamedType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_schema);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10_NamedType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10_NamedType_3_get_schema_extra = {"_get_schema_extra", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_NamedType_3_get_schema_extra, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10_NamedType_3_get_schema_extra(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_created = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_schema_extra (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_created,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_created)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 396, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_schema_extra") < 0)) __PYX_ERR(2, 396, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_created = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_schema_extra", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 396, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._NamedType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_created), (&PySet_Type), 1, "created", 1))) __PYX_ERR(2, 396, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_10_NamedType_2_get_schema_extra(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self), __pyx_v_created);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10_NamedType_2_get_schema_extra(struct __pyx_obj_5cavro__NamedType *__pyx_v_self, PyObject *__pyx_v_created) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_schema_extra", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_10_NamedType__get_schema_extra(__pyx_v_self, __pyx_v_created, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._NamedType._get_schema_extra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":409
 *         return schema
 * 
 *     cdef AvroType clone_base(self, cls=None):             # <<<<<<<<<<<<<<
 *         cdef _NamedType inst = AvroType.clone_base(self, cls)
 *         inst.name = self.name
 */

static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_10_NamedType_clone_base(struct __pyx_obj_5cavro__NamedType *__pyx_v_self, struct __pyx_opt_args_5cavro_10_NamedType_clone_base *__pyx_optional_args) {
  PyObject *__pyx_v_cls = ((PyObject *)Py_None);
  struct __pyx_obj_5cavro__NamedType *__pyx_v_inst = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_clone_base __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone_base", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_cls = __pyx_optional_args->cls;
    }
  }

  /* "src/type.pxi":410
 * 
 *     cdef AvroType clone_base(self, cls=None):
 *         cdef _NamedType inst = AvroType.clone_base(self, cls)             # <<<<<<<<<<<<<<
 *         inst.name = self.name
 *         inst.namespace = self.namespace
 */
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.cls = __pyx_v_cls;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_5cavro_AvroType->clone_base(((struct __pyx_obj_5cavro_AvroType *)__pyx_v_self), &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro__NamedType))))) __PYX_ERR(2, 410, __pyx_L1_error)
  __pyx_v_inst = ((struct __pyx_obj_5cavro__NamedType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":411
 *     cdef AvroType clone_base(self, cls=None):
 *         cdef _NamedType inst = AvroType.clone_base(self, cls)
 *         inst.name = self.name             # <<<<<<<<<<<<<<
 *         inst.namespace = self.namespace
 *         inst.effective_namespace = self.effective_namespace
 */
  __pyx_t_1 = __pyx_v_self->name;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_inst->name);
  __Pyx_DECREF(__pyx_v_inst->name);
  __pyx_v_inst->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":412
 *         cdef _NamedType inst = AvroType.clone_base(self, cls)
 *         inst.name = self.name
 *         inst.namespace = self.namespace             # <<<<<<<<<<<<<<
 *         inst.effective_namespace = self.effective_namespace
 *         inst.aliases = self.aliases
 */
  __pyx_t_1 = __pyx_v_self->namespace;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_inst->namespace);
  __Pyx_DECREF(__pyx_v_inst->namespace);
  __pyx_v_inst->namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":413
 *         inst.name = self.name
 *         inst.namespace = self.namespace
 *         inst.effective_namespace = self.effective_namespace             # <<<<<<<<<<<<<<
 *         inst.aliases = self.aliases
 *         return inst
 */
  __pyx_t_1 = __pyx_v_self->effective_namespace;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_inst->effective_namespace);
  __Pyx_DECREF(__pyx_v_inst->effective_namespace);
  __pyx_v_inst->effective_namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":414
 *         inst.namespace = self.namespace
 *         inst.effective_namespace = self.effective_namespace
 *         inst.aliases = self.aliases             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_t_1 = __pyx_v_self->aliases;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_inst->aliases);
  __Pyx_DECREF(__pyx_v_inst->aliases);
  __pyx_v_inst->aliases = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/type.pxi":415
 *         inst.effective_namespace = self.effective_namespace
 *         inst.aliases = self.aliases
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _extract_metadata(self, source):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_inst);
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_v_inst);
  goto __pyx_L0;

  /* "src/type.pxi":409
 *         return schema
 * 
 *     cdef AvroType clone_base(self, cls=None):             # <<<<<<<<<<<<<<
 *         cdef _NamedType inst = AvroType.clone_base(self, cls)
 *         inst.name = self.name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._NamedType.clone_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":417
 *         return inst
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {
 *             'type',
 */

static PyObject *__pyx_f_5cavro_10_NamedType__extract_metadata(CYTHON_UNUSED struct __pyx_obj_5cavro__NamedType *__pyx_v_self, PyObject *__pyx_v_source) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_extract_metadata", 1);

  /* "src/type.pxi":418
 * 
 *     cdef dict _extract_metadata(self, source):
 *         return _strip_keys(dict(source), {             # <<<<<<<<<<<<<<
 *             'type',
 *             'name',
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_source); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_n_u_type) < 0) __PYX_ERR(2, 419, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_name) < 0) __PYX_ERR(2, 419, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_namespace) < 0) __PYX_ERR(2, 419, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_n_u_aliases) < 0) __PYX_ERR(2, 419, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_5cavro__strip_keys(((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/type.pxi":417
 *         return inst
 * 
 *     cdef dict _extract_metadata(self, source):             # <<<<<<<<<<<<<<
 *         return _strip_keys(dict(source), {
 *             'type',
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro._NamedType._extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":323
 *     """Base class for all named types"""
 * 
 *     cdef readonly str name             # <<<<<<<<<<<<<<
 *     cdef readonly str namespace
 *     cdef readonly str effective_namespace
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10_NamedType_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_10_NamedType_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_10_NamedType_4name___get__(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10_NamedType_4name___get__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":324
 * 
 *     cdef readonly str name
 *     cdef readonly str namespace             # <<<<<<<<<<<<<<
 *     cdef readonly str effective_namespace
 *     cdef readonly frozenset aliases
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10_NamedType_9namespace_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_10_NamedType_9namespace_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_10_NamedType_9namespace___get__(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10_NamedType_9namespace___get__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->namespace);
  __pyx_r = __pyx_v_self->namespace;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":325
 *     cdef readonly str name
 *     cdef readonly str namespace
 *     cdef readonly str effective_namespace             # <<<<<<<<<<<<<<
 *     cdef readonly frozenset aliases
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10_NamedType_19effective_namespace_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_10_NamedType_19effective_namespace_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_10_NamedType_19effective_namespace___get__(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10_NamedType_19effective_namespace___get__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->effective_namespace);
  __pyx_r = __pyx_v_self->effective_namespace;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/type.pxi":326
 *     cdef readonly str namespace
 *     cdef readonly str effective_namespace
 *     cdef readonly frozenset aliases             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, schema, source, parse_namespace):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10_NamedType_7aliases_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_10_NamedType_7aliases_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_10_NamedType_7aliases___get__(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10_NamedType_7aliases___get__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->aliases);
  __pyx_r = __pyx_v_self->aliases;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10_NamedType_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10_NamedType_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_NamedType_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10_NamedType_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10_NamedType_4__reduce_cython__(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10_NamedType_4__reduce_cython__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.aliases, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.value_adapters)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->aliases);
  __Pyx_GIVEREF(__pyx_v_self->aliases);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->aliases)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->effective_namespace);
  __Pyx_GIVEREF(__pyx_v_self->effective_namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->effective_namespace)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.metadata);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->__pyx_base.metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->namespace);
  __Pyx_GIVEREF(__pyx_v_self->namespace);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_self->namespace)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.value_adapters);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.value_adapters);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_v_self->__pyx_base.value_adapters)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.aliases, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.aliases, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.aliases is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.value_adapters is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.aliases, self.effective_namespace, self.metadata, self.name, self.namespace, self.options, self.value_adapters)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.aliases is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.value_adapters is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__NamedType, (type(self), 0x24d7e16, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->aliases != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->effective_namespace != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.metadata != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->name != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->namespace != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.value_adapters != ((PyObject*)Py_None));
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__NamedType, (type(self), 0x24d7e16, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.aliases is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:
 *         return __pyx_unpickle__NamedType, (type(self), 0x24d7e16, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__NamedType, (type(self), 0x24d7e16, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle__NamedType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_38632982);
    __Pyx_GIVEREF(__pyx_int_38632982);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_38632982)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aliases is not None or self.effective_namespace is not None or self.metadata is not None or self.name is not None or self.namespace is not None or self.options is not None or self.value_adapters is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__NamedType, (type(self), 0x24d7e16, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__NamedType, (type(self), 0x24d7e16, None), state
 *     else:
 *         return __pyx_unpickle__NamedType, (type(self), 0x24d7e16, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__NamedType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle__NamedType); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_38632982);
    __Pyx_GIVEREF(__pyx_int_38632982);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_38632982)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro._NamedType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__NamedType, (type(self), 0x24d7e16, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__NamedType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10_NamedType_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10_NamedType_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_NamedType_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10_NamedType_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._NamedType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10_NamedType_6__setstate_cython__(((struct __pyx_obj_5cavro__NamedType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10_NamedType_6__setstate_cython__(struct __pyx_obj_5cavro__NamedType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__NamedType, (type(self), 0x24d7e16, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__NamedType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__NamedType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__NamedType, (type(self), 0x24d7e16, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__NamedType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._NamedType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":4
 * import hashlib
 * 
 * cdef str resolve_namespaced_name(str namespace, str name):             # <<<<<<<<<<<<<<
 *     if '.' in name or namespace is None:
 *         return name
 */

static PyObject *__pyx_f_5cavro_resolve_namespaced_name(PyObject *__pyx_v_namespace, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resolve_namespaced_name", 1);

  /* "src/schema.pxi":5
 * 
 * cdef str resolve_namespaced_name(str namespace, str name):
 *     if '.' in name or namespace is None:             # <<<<<<<<<<<<<<
 *         return name
 *     return f'{namespace}.{name}'
 */
  if (unlikely(__pyx_v_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(23, 5, __pyx_L1_error)
  }
  __pyx_t_2 = (__Pyx_PyUnicode_ContainsTF(__pyx_kp_u__2, __pyx_v_name, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(23, 5, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_namespace == ((PyObject*)Py_None));
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/schema.pxi":6
 * cdef str resolve_namespaced_name(str namespace, str name):
 *     if '.' in name or namespace is None:
 *         return name             # <<<<<<<<<<<<<<
 *     return f'{namespace}.{name}'
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_name);
    __pyx_r = __pyx_v_name;
    goto __pyx_L0;

    /* "src/schema.pxi":5
 * 
 * cdef str resolve_namespaced_name(str namespace, str name):
 *     if '.' in name or namespace is None:             # <<<<<<<<<<<<<<
 *         return name
 *     return f'{namespace}.{name}'
 */
  }

  /* "src/schema.pxi":7
 *     if '.' in name or namespace is None:
 *         return name
 *     return f'{namespace}.{name}'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 0;
  __pyx_t_5 = 127;
  __pyx_t_6 = __Pyx_PyUnicode_Unicode(__pyx_v_namespace); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
  __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_4 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_kp_u__2);
  __pyx_t_6 = __Pyx_PyUnicode_Unicode(__pyx_v_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
  __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "src/schema.pxi":4
 * import hashlib
 * 
 * cdef str resolve_namespaced_name(str namespace, str name):             # <<<<<<<<<<<<<<
 *     if '.' in name or namespace is None:
 *         return name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.resolve_namespaced_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":11
 * 
 * class _class_inst_method:
 *     def __init__(self, func):             # <<<<<<<<<<<<<<
 *         self.func = func
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_18_class_inst_method_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_18_class_inst_method_1__init__ = {"__init__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_18_class_inst_method_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_18_class_inst_method_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 11, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 11, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(23, 11, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(23, 11, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_func = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(23, 11, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._class_inst_method.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_18_class_inst_method___init__(__pyx_self, __pyx_v_self, __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_18_class_inst_method___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/schema.pxi":12
 * class _class_inst_method:
 *     def __init__(self, func):
 *         self.func = func             # <<<<<<<<<<<<<<
 * 
 *     def __get__(self, inst, cls):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_func, __pyx_v_func) < 0) __PYX_ERR(23, 12, __pyx_L1_error)

  /* "src/schema.pxi":11
 * 
 * class _class_inst_method:
 *     def __init__(self, func):             # <<<<<<<<<<<<<<
 *         self.func = func
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro._class_inst_method.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":14
 *         self.func = func
 * 
 *     def __get__(self, inst, cls):             # <<<<<<<<<<<<<<
 *         return partial(self.func, inst, cls)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_18_class_inst_method_3__get__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_18_class_inst_method_3__get__ = {"__get__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_18_class_inst_method_3__get__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_18_class_inst_method_3__get__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_inst = 0;
  PyObject *__pyx_v_cls = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_inst,&__pyx_n_s_cls,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 14, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_inst)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 14, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__get__", 1, 3, 3, 1); __PYX_ERR(23, 14, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_cls)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 14, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__get__", 1, 3, 3, 2); __PYX_ERR(23, 14, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__get__") < 0)) __PYX_ERR(23, 14, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_inst = values[1];
    __pyx_v_cls = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__get__", 1, 3, 3, __pyx_nargs); __PYX_ERR(23, 14, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._class_inst_method.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_18_class_inst_method_2__get__(__pyx_self, __pyx_v_self, __pyx_v_inst, __pyx_v_cls);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_18_class_inst_method_2__get__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_inst, PyObject *__pyx_v_cls) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/schema.pxi":15
 * 
 *     def __get__(self, inst, cls):
 *         return partial(self.func, inst, cls)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_partial); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_func); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_t_3, __pyx_v_inst, __pyx_v_cls};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/schema.pxi":14
 *         self.func = func
 * 
 *     def __get__(self, inst, cls):             # <<<<<<<<<<<<<<
 *         return partial(self.func, inst, cls)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._class_inst_method.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":48
 *     cdef readonly dict logical_types
 * 
 *     def __init__(self, source: Union[str, object], Options options: Options=DEFAULT_OPTIONS, *, named_types: dict[str, AvroType]=None, parse_json: bool=True, _type=None, **extra_options):             # <<<<<<<<<<<<<<
 *         if isinstance(source, (str, bytes)) and parse_json:
 *             source = json.loads(source)
 */

/* Python wrapper */
static int __pyx_pw_5cavro_6Schema_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_6Schema_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_source = 0;
  struct __pyx_obj_5cavro_Options *__pyx_v_options = 0;
  PyObject *__pyx_v_named_types = 0;
  __pyx_t_5cavro_bool __pyx_v_parse_json;
  PyObject *__pyx_v__type = 0;
  PyObject *__pyx_v_extra_options = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_v_extra_options = PyDict_New(); if (unlikely(!__pyx_v_extra_options)) return -1;
  __Pyx_GOTREF(__pyx_v_extra_options);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_source,&__pyx_n_s_options,&__pyx_n_s_named_types,&__pyx_n_s_parse_json,&__pyx_n_s_type_3,0};
    values[1] = __Pyx_Arg_NewRef_VARARGS((PyObject *)__pyx_k__83);
    values[2] = __Pyx_Arg_NewRef_VARARGS(((PyObject*)Py_None));
    values[4] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 48, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_options);
          if (value) { values[1] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 48, __pyx_L3_error)
        }
      }
      if (kw_args > 0 && (kw_args <= 3)) {
        Py_ssize_t index;
        for (index = 2; index < 5 && kw_args > 0; index++) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
          if (value) { values[index] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 48, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_extra_options, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(23, 48, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_source = values[0];
    __pyx_v_options = ((struct __pyx_obj_5cavro_Options *)values[1]);
    __pyx_v_named_types = ((PyObject*)values[2]);
    if (values[3]) {
      __pyx_v_parse_json = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_parse_json == (int)-1) && PyErr_Occurred())) __PYX_ERR(23, 48, __pyx_L3_error)
    } else {
      __pyx_v_parse_json = ((__pyx_t_5cavro_bool)1);
    }
    __pyx_v__type = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, __pyx_nargs); __PYX_ERR(23, 48, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_extra_options); __pyx_v_extra_options = 0;
  __Pyx_AddTraceback("cavro.Schema.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_options), __pyx_ptype_5cavro_Options, 0, "options", 0))) __PYX_ERR(23, 48, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_named_types), (&PyDict_Type), 1, "named_types", 1))) __PYX_ERR(23, 48, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_6Schema___init__(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v_source, __pyx_v_options, __pyx_v_named_types, __pyx_v_parse_json, __pyx_v__type, __pyx_v_extra_options);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_DECREF(__pyx_v_extra_options);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_6Schema___init__(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_source, struct __pyx_obj_5cavro_Options *__pyx_v_options, PyObject *__pyx_v_named_types, __pyx_t_5cavro_bool __pyx_v_parse_json, PyObject *__pyx_v__type, PyObject *__pyx_v_extra_options) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_source);
  __Pyx_INCREF((PyObject *)__pyx_v_options);

  /* "src/schema.pxi":49
 * 
 *     def __init__(self, source: Union[str, object], Options options: Options=DEFAULT_OPTIONS, *, named_types: dict[str, AvroType]=None, parse_json: bool=True, _type=None, **extra_options):
 *         if isinstance(source, (str, bytes)) and parse_json:             # <<<<<<<<<<<<<<
 *             source = json.loads(source)
 *         self.source = source
 */
  __pyx_t_3 = PyUnicode_Check(__pyx_v_source); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_3 = PyBytes_Check(__pyx_v_source); 
  __pyx_t_2 = __pyx_t_3;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_parse_json != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/schema.pxi":50
 *     def __init__(self, source: Union[str, object], Options options: Options=DEFAULT_OPTIONS, *, named_types: dict[str, AvroType]=None, parse_json: bool=True, _type=None, **extra_options):
 *         if isinstance(source, (str, bytes)) and parse_json:
 *             source = json.loads(source)             # <<<<<<<<<<<<<<
 *         self.source = source
 *         if extra_options:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_json); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_loads); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_source};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(23, 50, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_source, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/schema.pxi":49
 * 
 *     def __init__(self, source: Union[str, object], Options options: Options=DEFAULT_OPTIONS, *, named_types: dict[str, AvroType]=None, parse_json: bool=True, _type=None, **extra_options):
 *         if isinstance(source, (str, bytes)) and parse_json:             # <<<<<<<<<<<<<<
 *             source = json.loads(source)
 *         self.source = source
 */
  }

  /* "src/schema.pxi":51
 *         if isinstance(source, (str, bytes)) and parse_json:
 *             source = json.loads(source)
 *         self.source = source             # <<<<<<<<<<<<<<
 *         if extra_options:
 *             options = options.replace(**extra_options)
 */
  __Pyx_INCREF(__pyx_v_source);
  __Pyx_GIVEREF(__pyx_v_source);
  __Pyx_GOTREF(__pyx_v_self->source);
  __Pyx_DECREF(__pyx_v_self->source);
  __pyx_v_self->source = __pyx_v_source;

  /* "src/schema.pxi":52
 *             source = json.loads(source)
 *         self.source = source
 *         if extra_options:             # <<<<<<<<<<<<<<
 *             options = options.replace(**extra_options)
 *         self.options = options
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_extra_options); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(23, 52, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/schema.pxi":53
 *         self.source = source
 *         if extra_options:
 *             options = options.replace(**extra_options)             # <<<<<<<<<<<<<<
 *         self.options = options
 *         self.named_types = {} if named_types is None else named_types
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_options), __pyx_n_s_replace); if (unlikely(!__pyx_t_4)) __PYX_ERR(23, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyDict_Copy(__pyx_v_extra_options); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5cavro_Options))))) __PYX_ERR(23, 53, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_options, ((struct __pyx_obj_5cavro_Options *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "src/schema.pxi":52
 *             source = json.loads(source)
 *         self.source = source
 *         if extra_options:             # <<<<<<<<<<<<<<
 *             options = options.replace(**extra_options)
 *         self.options = options
 */
  }

  /* "src/schema.pxi":54
 *         if extra_options:
 *             options = options.replace(**extra_options)
 *         self.options = options             # <<<<<<<<<<<<<<
 *         self.named_types = {} if named_types is None else named_types
 *         self.logical_types = self._make_logical_types(options)
 */
  __Pyx_INCREF((PyObject *)__pyx_v_options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_options);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->options);
  __Pyx_DECREF((PyObject *)__pyx_v_self->options);
  __pyx_v_self->options = __pyx_v_options;

  /* "src/schema.pxi":55
 *             options = options.replace(**extra_options)
 *         self.options = options
 *         self.named_types = {} if named_types is None else named_types             # <<<<<<<<<<<<<<
 *         self.logical_types = self._make_logical_types(options)
 *         self.type = AvroType.for_schema(self) if _type is None else _type
 */
  __pyx_t_1 = (__pyx_v_named_types == ((PyObject*)Py_None));
  if (__pyx_t_1) {
    __pyx_t_6 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __pyx_t_6;
    __pyx_t_6 = 0;
  } else {
    __Pyx_INCREF(__pyx_v_named_types);
    __pyx_t_5 = __pyx_v_named_types;
  }
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->named_types);
  __Pyx_DECREF(__pyx_v_self->named_types);
  __pyx_v_self->named_types = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "src/schema.pxi":56
 *         self.options = options
 *         self.named_types = {} if named_types is None else named_types
 *         self.logical_types = self._make_logical_types(options)             # <<<<<<<<<<<<<<
 *         self.type = AvroType.for_schema(self) if _type is None else _type
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_make_logical_types); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_options)};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  if (!(likely(PyDict_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_5))) __PYX_ERR(23, 56, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->logical_types);
  __Pyx_DECREF(__pyx_v_self->logical_types);
  __pyx_v_self->logical_types = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "src/schema.pxi":57
 *         self.named_types = {} if named_types is None else named_types
 *         self.logical_types = self._make_logical_types(options)
 *         self.type = AvroType.for_schema(self) if _type is None else _type             # <<<<<<<<<<<<<<
 * 
 *     @_class_inst_method
 */
  __pyx_t_1 = (__pyx_v__type == Py_None);
  if (__pyx_t_1) {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_AvroType), __pyx_n_s_for_schema); if (unlikely(!__pyx_t_4)) __PYX_ERR(23, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, ((PyObject *)__pyx_v_self)};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 57, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(23, 57, __pyx_L1_error)
    __pyx_t_5 = __pyx_t_6;
    __pyx_t_6 = 0;
  } else {
    if (!(likely(((__pyx_v__type) == Py_None) || likely(__Pyx_TypeTest(__pyx_v__type, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(23, 57, __pyx_L1_error)
    __Pyx_INCREF(__pyx_v__type);
    __pyx_t_5 = __pyx_v__type;
  }
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->type);
  __Pyx_DECREF((PyObject *)__pyx_v_self->type);
  __pyx_v_self->type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "src/schema.pxi":48
 *     cdef readonly dict logical_types
 * 
 *     def __init__(self, source: Union[str, object], Options options: Options=DEFAULT_OPTIONS, *, named_types: dict[str, AvroType]=None, parse_json: bool=True, _type=None, **extra_options):             # <<<<<<<<<<<<<<
 *         if isinstance(source, (str, bytes)) and parse_json:
 *             source = json.loads(source)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cavro.Schema.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_source);
  __Pyx_XDECREF((PyObject *)__pyx_v_options);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":59
 *         self.type = AvroType.for_schema(self) if _type is None else _type
 * 
 *     @_class_inst_method             # <<<<<<<<<<<<<<
 *     def _wrap_type(inst, cls, AvroType avro_type, Options options=None):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_3_wrap_type(PyObject *__pyx_v_inst, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_6Schema_2_wrap_type, "\n        Used in certain situations to create a `Schema` object from a raw `AvroType` instance.\n        You probably don't need to use this.\n        ");
static PyMethodDef __pyx_mdef_5cavro_6Schema_3_wrap_type = {"_wrap_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_3_wrap_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_2_wrap_type};
static PyObject *__pyx_pw_5cavro_6Schema_3_wrap_type(PyObject *__pyx_v_inst, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_cls = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_avro_type = 0;
  struct __pyx_obj_5cavro_Options *__pyx_v_options = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_wrap_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_avro_type,&__pyx_n_s_options,0};

    /* "src/schema.pxi":60
 * 
 *     @_class_inst_method
 *     def _wrap_type(inst, cls, AvroType avro_type, Options options=None):             # <<<<<<<<<<<<<<
 *         """
 *         Used in certain situations to create a `Schema` object from a raw `AvroType` instance.
 */
    values[2] = __Pyx_Arg_NewRef_FASTCALL((PyObject *)((struct __pyx_obj_5cavro_Options *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_cls)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 59, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_avro_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 59, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_wrap_type", 0, 2, 3, 1); __PYX_ERR(23, 59, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_options);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 59, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_wrap_type") < 0)) __PYX_ERR(23, 59, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cls = values[0];
    __pyx_v_avro_type = ((struct __pyx_obj_5cavro_AvroType *)values[1]);
    __pyx_v_options = ((struct __pyx_obj_5cavro_Options *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_wrap_type", 0, 2, 3, __pyx_nargs); __PYX_ERR(23, 59, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Schema._wrap_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_avro_type), __pyx_ptype_5cavro_AvroType, 1, "avro_type", 0))) __PYX_ERR(23, 60, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_options), __pyx_ptype_5cavro_Options, 1, "options", 0))) __PYX_ERR(23, 60, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_6Schema_2_wrap_type(((struct __pyx_obj_5cavro_Schema *)__pyx_v_inst), __pyx_v_cls, __pyx_v_avro_type, __pyx_v_options);

  /* "src/schema.pxi":59
 *         self.type = AvroType.for_schema(self) if _type is None else _type
 * 
 *     @_class_inst_method             # <<<<<<<<<<<<<<
 *     def _wrap_type(inst, cls, AvroType avro_type, Options options=None):
 *         """
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_2_wrap_type(struct __pyx_obj_5cavro_Schema *__pyx_v_inst, PyObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_avro_type, struct __pyx_obj_5cavro_Options *__pyx_v_options) {
  PyObject *__pyx_v_source = NULL;
  struct __pyx_obj_5cavro_Schema *__pyx_v_new_inst = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_get_schema __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_wrap_type", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_options);

  /* "src/schema.pxi":65
 *         You probably don't need to use this.
 *         """
 *         if inst is not None:             # <<<<<<<<<<<<<<
 *             if options is None:
 *                 options = inst.options
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_inst) != Py_None);
  if (__pyx_t_1) {

    /* "src/schema.pxi":66
 *         """
 *         if inst is not None:
 *             if options is None:             # <<<<<<<<<<<<<<
 *                 options = inst.options
 *         if options is None:
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_options) == Py_None);
    if (__pyx_t_1) {

      /* "src/schema.pxi":67
 *         if inst is not None:
 *             if options is None:
 *                 options = inst.options             # <<<<<<<<<<<<<<
 *         if options is None:
 *             options = DEFAULT_OPTIONS
 */
      __pyx_t_2 = ((PyObject *)__pyx_v_inst->options);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_options, ((struct __pyx_obj_5cavro_Options *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "src/schema.pxi":66
 *         """
 *         if inst is not None:
 *             if options is None:             # <<<<<<<<<<<<<<
 *                 options = inst.options
 *         if options is None:
 */
    }

    /* "src/schema.pxi":65
 *         You probably don't need to use this.
 *         """
 *         if inst is not None:             # <<<<<<<<<<<<<<
 *             if options is None:
 *                 options = inst.options
 */
  }

  /* "src/schema.pxi":68
 *             if options is None:
 *                 options = inst.options
 *         if options is None:             # <<<<<<<<<<<<<<
 *             options = DEFAULT_OPTIONS
 *         source = avro_type.get_schema(set())
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_options) == Py_None);
  if (__pyx_t_1) {

    /* "src/schema.pxi":69
 *                 options = inst.options
 *         if options is None:
 *             options = DEFAULT_OPTIONS             # <<<<<<<<<<<<<<
 *         source = avro_type.get_schema(set())
 *         cdef Schema new_inst = cls(source, options, parse_json=False, _type=avro_type)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_DEFAULT_OPTIONS); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_Options))))) __PYX_ERR(23, 69, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_options, ((struct __pyx_obj_5cavro_Options *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "src/schema.pxi":68
 *             if options is None:
 *                 options = inst.options
 *         if options is None:             # <<<<<<<<<<<<<<
 *             options = DEFAULT_OPTIONS
 *         source = avro_type.get_schema(set())
 */
  }

  /* "src/schema.pxi":70
 *         if options is None:
 *             options = DEFAULT_OPTIONS
 *         source = avro_type.get_schema(set())             # <<<<<<<<<<<<<<
 *         cdef Schema new_inst = cls(source, options, parse_json=False, _type=avro_type)
 *         return new_inst
 */
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4.__pyx_n = 1;
  __pyx_t_4.created = __pyx_t_2;
  __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_avro_type->__pyx_vtab)->get_schema(__pyx_v_avro_type, 0, &__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_source = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "src/schema.pxi":71
 *             options = DEFAULT_OPTIONS
 *         source = avro_type.get_schema(set())
 *         cdef Schema new_inst = cls(source, options, parse_json=False, _type=avro_type)             # <<<<<<<<<<<<<<
 *         return new_inst
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_source);
  __Pyx_GIVEREF(__pyx_v_source);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_source)) __PYX_ERR(23, 71, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_options))) __PYX_ERR(23, 71, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_parse_json, Py_False) < 0) __PYX_ERR(23, 71, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_type_3, ((PyObject *)__pyx_v_avro_type)) < 0) __PYX_ERR(23, 71, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_v_cls, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5cavro_Schema))))) __PYX_ERR(23, 71, __pyx_L1_error)
  __pyx_v_new_inst = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "src/schema.pxi":72
 *         source = avro_type.get_schema(set())
 *         cdef Schema new_inst = cls(source, options, parse_json=False, _type=avro_type)
 *         return new_inst             # <<<<<<<<<<<<<<
 * 
 *     def _make_logical_types(self, options):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_new_inst);
  __pyx_r = ((PyObject *)__pyx_v_new_inst);
  goto __pyx_L0;

  /* "src/schema.pxi":59
 *         self.type = AvroType.for_schema(self) if _type is None else _type
 * 
 *     @_class_inst_method             # <<<<<<<<<<<<<<
 *     def _wrap_type(inst, cls, AvroType avro_type, Options options=None):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.Schema._wrap_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_source);
  __Pyx_XDECREF((PyObject *)__pyx_v_new_inst);
  __Pyx_XDECREF((PyObject *)__pyx_v_options);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":74
 *         return new_inst
 * 
 *     def _make_logical_types(self, options):             # <<<<<<<<<<<<<<
 *         logical_by_name = {}
 *         self.logical_types = logical_by_name
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_5_make_logical_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_6Schema_5_make_logical_types = {"_make_logical_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_5_make_logical_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_6Schema_5_make_logical_types(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_options = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_make_logical_types (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_options,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_options)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 74, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_make_logical_types") < 0)) __PYX_ERR(23, 74, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_options = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_make_logical_types", 1, 1, 1, __pyx_nargs); __PYX_ERR(23, 74, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Schema._make_logical_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6Schema_4_make_logical_types(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v_options);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_4_make_logical_types(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_options) {
  PyObject *__pyx_v_logical_by_name = NULL;
  PyObject *__pyx_v_logical_type = NULL;
  PyObject *__pyx_v_type_name = NULL;
  PyObject *__pyx_v_dest = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_logical_types", 1);

  /* "src/schema.pxi":75
 * 
 *     def _make_logical_types(self, options):
 *         logical_by_name = {}             # <<<<<<<<<<<<<<
 *         self.logical_types = logical_by_name
 *         for logical_type in options.logical_types:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_logical_by_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/schema.pxi":76
 *     def _make_logical_types(self, options):
 *         logical_by_name = {}
 *         self.logical_types = logical_by_name             # <<<<<<<<<<<<<<
 *         for logical_type in options.logical_types:
 *             type_name = logical_type.logical_name
 */
  __Pyx_INCREF(__pyx_v_logical_by_name);
  __Pyx_GIVEREF(__pyx_v_logical_by_name);
  __Pyx_GOTREF(__pyx_v_self->logical_types);
  __Pyx_DECREF(__pyx_v_self->logical_types);
  __pyx_v_self->logical_types = __pyx_v_logical_by_name;

  /* "src/schema.pxi":77
 *         logical_by_name = {}
 *         self.logical_types = logical_by_name
 *         for logical_type in options.logical_types:             # <<<<<<<<<<<<<<
 *             type_name = logical_type.logical_name
 *             dest = logical_by_name.setdefault(type_name, [])
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_options, __pyx_n_s_logical_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(23, 77, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(23, 77, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(23, 77, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 77, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(23, 77, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(23, 77, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 77, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(23, 77, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_logical_type, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/schema.pxi":78
 *         self.logical_types = logical_by_name
 *         for logical_type in options.logical_types:
 *             type_name = logical_type.logical_name             # <<<<<<<<<<<<<<
 *             dest = logical_by_name.setdefault(type_name, [])
 *             dest.append(logical_type)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_logical_type, __pyx_n_s_logical_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_type_name, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/schema.pxi":79
 *         for logical_type in options.logical_types:
 *             type_name = logical_type.logical_name
 *             dest = logical_by_name.setdefault(type_name, [])             # <<<<<<<<<<<<<<
 *             dest.append(logical_type)
 *         return logical_by_name
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyDict_SetDefault(__pyx_v_logical_by_name, __pyx_v_type_name, __pyx_t_1, -1L); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_dest, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "src/schema.pxi":80
 *             type_name = logical_type.logical_name
 *             dest = logical_by_name.setdefault(type_name, [])
 *             dest.append(logical_type)             # <<<<<<<<<<<<<<
 *         return logical_by_name
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_Append(__pyx_v_dest, __pyx_v_logical_type); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(23, 80, __pyx_L1_error)

    /* "src/schema.pxi":77
 *         logical_by_name = {}
 *         self.logical_types = logical_by_name
 *         for logical_type in options.logical_types:             # <<<<<<<<<<<<<<
 *             type_name = logical_type.logical_name
 *             dest = logical_by_name.setdefault(type_name, [])
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/schema.pxi":81
 *             dest = logical_by_name.setdefault(type_name, [])
 *             dest.append(logical_type)
 *         return logical_by_name             # <<<<<<<<<<<<<<
 * 
 *     cdef void register_type(self, str namespace, str name, AvroType avro_type):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_logical_by_name);
  __pyx_r = __pyx_v_logical_by_name;
  goto __pyx_L0;

  /* "src/schema.pxi":74
 *         return new_inst
 * 
 *     def _make_logical_types(self, options):             # <<<<<<<<<<<<<<
 *         logical_by_name = {}
 *         self.logical_types = logical_by_name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.Schema._make_logical_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_logical_by_name);
  __Pyx_XDECREF(__pyx_v_logical_type);
  __Pyx_XDECREF(__pyx_v_type_name);
  __Pyx_XDECREF(__pyx_v_dest);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":83
 *         return logical_by_name
 * 
 *     cdef void register_type(self, str namespace, str name, AvroType avro_type):             # <<<<<<<<<<<<<<
 *         resolved = resolve_namespaced_name(namespace, name)
 *         if self.options.named_type_names_must_be_unique and resolved in self.named_types:
 */

static void __pyx_f_5cavro_6Schema_register_type(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_name, struct __pyx_obj_5cavro_AvroType *__pyx_v_avro_type) {
  PyObject *__pyx_v_resolved = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_existing = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("register_type", 1);

  /* "src/schema.pxi":84
 * 
 *     cdef void register_type(self, str namespace, str name, AvroType avro_type):
 *         resolved = resolve_namespaced_name(namespace, name)             # <<<<<<<<<<<<<<
 *         if self.options.named_type_names_must_be_unique and resolved in self.named_types:
 *             existing = self.named_types[resolved]
 */
  __pyx_t_1 = __pyx_f_5cavro_resolve_namespaced_name(__pyx_v_namespace, __pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_resolved = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/schema.pxi":85
 *     cdef void register_type(self, str namespace, str name, AvroType avro_type):
 *         resolved = resolve_namespaced_name(namespace, name)
 *         if self.options.named_type_names_must_be_unique and resolved in self.named_types:             # <<<<<<<<<<<<<<
 *             existing = self.named_types[resolved]
 *             raise DuplicateName(f'Name {resolved!r} appears multiple times in schema')
 */
  if (__pyx_v_self->options->named_type_names_must_be_unique) {
  } else {
    __pyx_t_2 = __pyx_v_self->options->named_type_names_must_be_unique;
    goto __pyx_L4_bool_binop_done;
  }
  if (unlikely(__pyx_v_self->named_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(23, 85, __pyx_L1_error)
  }
  __pyx_t_3 = (__Pyx_PyDict_ContainsTF(__pyx_v_resolved, __pyx_v_self->named_types, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(23, 85, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "src/schema.pxi":86
 *         resolved = resolve_namespaced_name(namespace, name)
 *         if self.options.named_type_names_must_be_unique and resolved in self.named_types:
 *             existing = self.named_types[resolved]             # <<<<<<<<<<<<<<
 *             raise DuplicateName(f'Name {resolved!r} appears multiple times in schema')
 *         self.named_types[resolved] = avro_type
 */
    if (unlikely(__pyx_v_self->named_types == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(23, 86, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->named_types, __pyx_v_resolved); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_existing = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "src/schema.pxi":87
 *         if self.options.named_type_names_must_be_unique and resolved in self.named_types:
 *             existing = self.named_types[resolved]
 *             raise DuplicateName(f'Name {resolved!r} appears multiple times in schema')             # <<<<<<<<<<<<<<
 *         self.named_types[resolved] = avro_type
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_DuplicateName); if (unlikely(!__pyx_t_4)) __PYX_ERR(23, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 0;
    __pyx_t_7 = 127;
    __Pyx_INCREF(__pyx_kp_u_Name);
    __pyx_t_6 += 5;
    __Pyx_GIVEREF(__pyx_kp_u_Name);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Name);
    __pyx_t_8 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_resolved), __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(23, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_kp_u_appears_multiple_times_in_schem);
    __pyx_t_6 += 33;
    __Pyx_GIVEREF(__pyx_kp_u_appears_multiple_times_in_schem);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u_appears_multiple_times_in_schem);
    __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(23, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 87, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(23, 87, __pyx_L1_error)

    /* "src/schema.pxi":85
 *     cdef void register_type(self, str namespace, str name, AvroType avro_type):
 *         resolved = resolve_namespaced_name(namespace, name)
 *         if self.options.named_type_names_must_be_unique and resolved in self.named_types:             # <<<<<<<<<<<<<<
 *             existing = self.named_types[resolved]
 *             raise DuplicateName(f'Name {resolved!r} appears multiple times in schema')
 */
  }

  /* "src/schema.pxi":88
 *             existing = self.named_types[resolved]
 *             raise DuplicateName(f'Name {resolved!r} appears multiple times in schema')
 *         self.named_types[resolved] = avro_type             # <<<<<<<<<<<<<<
 * 
 *     property canonical_form:
 */
  if (unlikely(__pyx_v_self->named_types == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(23, 88, __pyx_L1_error)
  }
  if (unlikely((PyDict_SetItem(__pyx_v_self->named_types, __pyx_v_resolved, ((PyObject *)__pyx_v_avro_type)) < 0))) __PYX_ERR(23, 88, __pyx_L1_error)

  /* "src/schema.pxi":83
 *         return logical_by_name
 * 
 *     cdef void register_type(self, str namespace, str name, AvroType avro_type):             # <<<<<<<<<<<<<<
 *         resolved = resolve_namespaced_name(namespace, name)
 *         if self.options.named_type_names_must_be_unique and resolved in self.named_types:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cavro.Schema.register_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_resolved);
  __Pyx_XDECREF(__pyx_v_existing);
  __Pyx_RefNannyFinishContext();
}

/* "src/schema.pxi":94
 *         Returns the canonical form of the schema as a string
 *         """
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             # We are effectively immutable, so caching canonical form is safe here.
 *             if self._canonical_form is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_14canonical_form_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_6Schema_14canonical_form_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6Schema_14canonical_form___get__(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_14canonical_form___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/schema.pxi":96
 *         def __get__(self):
 *             # We are effectively immutable, so caching canonical form is safe here.
 *             if self._canonical_form is None:             # <<<<<<<<<<<<<<
 *                 self._canonical_form = self.type.canonical_form(set())
 *             return self._canonical_form
 */
  __pyx_t_1 = (__pyx_v_self->_canonical_form == ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "src/schema.pxi":97
 *             # We are effectively immutable, so caching canonical form is safe here.
 *             if self._canonical_form is None:
 *                 self._canonical_form = self.type.canonical_form(set())             # <<<<<<<<<<<<<<
 *             return self._canonical_form
 * 
 */
    __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->canonical_form(__pyx_v_self->type, ((PyObject*)__pyx_t_2))); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_canonical_form);
    __Pyx_DECREF(__pyx_v_self->_canonical_form);
    __pyx_v_self->_canonical_form = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/schema.pxi":96
 *         def __get__(self):
 *             # We are effectively immutable, so caching canonical form is safe here.
 *             if self._canonical_form is None:             # <<<<<<<<<<<<<<
 *                 self._canonical_form = self.type.canonical_form(set())
 *             return self._canonical_form
 */
  }

  /* "src/schema.pxi":98
 *             if self._canonical_form is None:
 *                 self._canonical_form = self.type.canonical_form(set())
 *             return self._canonical_form             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_canonical_form);
  __pyx_r = __pyx_v_self->_canonical_form;
  goto __pyx_L0;

  /* "src/schema.pxi":94
 *         Returns the canonical form of the schema as a string
 *         """
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             # We are effectively immutable, so caching canonical form is safe here.
 *             if self._canonical_form is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.Schema.canonical_form.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":100
 *             return self._canonical_form
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(self.canonical_form)
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_5cavro_6Schema_7__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_5cavro_6Schema_7__hash__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6Schema_6__hash__(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_5cavro_6Schema_6__hash__(struct __pyx_obj_5cavro_Schema *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 1);

  /* "src/schema.pxi":101
 * 
 *     def __hash__(self):
 *         return hash(self.canonical_form)             # <<<<<<<<<<<<<<
 * 
 *     property schema:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_canonical_form); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(23, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "src/schema.pxi":100
 *             return self._canonical_form
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(self.canonical_form)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Schema.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":108
 *         Note: This will not always be identical to the `source` used to construct this schema object, as it is reconstructed from the types on-demand.
 *         """
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.type.get_schema(set())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_6schema_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_6Schema_6schema_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6Schema_6schema___get__(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_6schema___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_opt_args_5cavro_8AvroType_get_schema __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/schema.pxi":109
 *         """
 *         def __get__(self):
 *             return self.type.get_schema(set())             # <<<<<<<<<<<<<<
 * 
 *     property schema_str:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3.__pyx_n = 1;
  __pyx_t_3.created = __pyx_t_1;
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->get_schema(__pyx_v_self->type, 0, &__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/schema.pxi":108
 *         Note: This will not always be identical to the `source` used to construct this schema object, as it is reconstructed from the types on-demand.
 *         """
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.type.get_schema(set())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.Schema.schema.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":115
 *         `Schema.schema`, but json encoded
 *         """
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return json.dumps(self.schema, indent=2)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_10schema_str_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_6Schema_10schema_str_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6Schema_10schema_str___get__(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_10schema_str___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/schema.pxi":116
 *         """
 *         def __get__(self):
 *             return json.dumps(self.schema, indent=2)             # <<<<<<<<<<<<<<
 * 
 *     def fingerprint(self, method='rabin', **kwargs) -> Union[bytes, hashlib._hashlib.HASH]:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_json); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_dumps); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_schema); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(23, 116, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_indent, __pyx_int_2) < 0) __PYX_ERR(23, 116, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(23, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "src/schema.pxi":115
 *         `Schema.schema`, but json encoded
 *         """
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return json.dumps(self.schema, indent=2)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Schema.schema_str.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":118
 *             return json.dumps(self.schema, indent=2)
 * 
 *     def fingerprint(self, method='rabin', **kwargs) -> Union[bytes, hashlib._hashlib.HASH]:             # <<<<<<<<<<<<<<
 *         """
 *         Return the deterministic fingerprint of the schema, using the given hash method.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_9fingerprint(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_6Schema_8fingerprint, "\n        Return the deterministic fingerprint of the schema, using the given hash method.\n        \n        `**kwargs` are passed to the relevant `hashlib.new()` call.\n        \n        Return type is controlled by the `fingerprint_returns_digest` option.\n        ");
static PyMethodDef __pyx_mdef_5cavro_6Schema_9fingerprint = {"fingerprint", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_9fingerprint, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_8fingerprint};
static PyObject *__pyx_pw_5cavro_6Schema_9fingerprint(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_method = 0;
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fingerprint (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_method,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_n_u_rabin));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_method);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 118, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, kwd_pos_args, "fingerprint") < 0)) __PYX_ERR(23, 118, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_method = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fingerprint", 0, 0, 1, __pyx_nargs); __PYX_ERR(23, 118, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("cavro.Schema.fingerprint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6Schema_8fingerprint(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v_method, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_8fingerprint(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_method, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_hasher = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fingerprint", 1);

  /* "src/schema.pxi":126
 *         Return type is controlled by the `fingerprint_returns_digest` option.
 *         """
 *         if method == 'rabin':             # <<<<<<<<<<<<<<
 *             hasher = Rabin()
 *         else:
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_method, __pyx_n_u_rabin, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(23, 126, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/schema.pxi":127
 *         """
 *         if method == 'rabin':
 *             hasher = Rabin()             # <<<<<<<<<<<<<<
 *         else:
 *             try:
 */
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro_Rabin)); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_hasher = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "src/schema.pxi":126
 *         Return type is controlled by the `fingerprint_returns_digest` option.
 *         """
 *         if method == 'rabin':             # <<<<<<<<<<<<<<
 *             hasher = Rabin()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "src/schema.pxi":129
 *             hasher = Rabin()
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 hasher = hashlib.new(method, **kwargs)
 *             except ValueError:
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "src/schema.pxi":130
 *         else:
 *             try:
 *                 hasher = hashlib.new(method, **kwargs)             # <<<<<<<<<<<<<<
 *             except ValueError:
 *                 raise InvalidHasher(f'Unknown hash method: {method!r}')
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_hashlib); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 130, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_new_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 130, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 130, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_v_method);
        __Pyx_GIVEREF(__pyx_v_method);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_method)) __PYX_ERR(23, 130, __pyx_L4_error);
        __pyx_t_7 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 130, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(23, 130, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_hasher = __pyx_t_8;
        __pyx_t_8 = 0;

        /* "src/schema.pxi":129
 *             hasher = Rabin()
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 hasher = hashlib.new(method, **kwargs)
 *             except ValueError:
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "src/schema.pxi":131
 *             try:
 *                 hasher = hashlib.new(method, **kwargs)
 *             except ValueError:             # <<<<<<<<<<<<<<
 *                 raise InvalidHasher(f'Unknown hash method: {method!r}')
 *         hasher.update(self.canonical_form.encode('utf-8'))
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("cavro.Schema.fingerprint", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(23, 131, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_2);

        /* "src/schema.pxi":132
 *                 hasher = hashlib.new(method, **kwargs)
 *             except ValueError:
 *                 raise InvalidHasher(f'Unknown hash method: {method!r}')             # <<<<<<<<<<<<<<
 *         hasher.update(self.canonical_form.encode('utf-8'))
 *         if self.options.fingerprint_returns_digest:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_InvalidHasher); if (unlikely(!__pyx_t_10)) __PYX_ERR(23, 132, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_method), __pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(23, 132, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Unknown_hash_method, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(23, 132, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = NULL;
        __pyx_t_9 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_9 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_12};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 132, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_Raise(__pyx_t_6, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __PYX_ERR(23, 132, __pyx_L6_except_error)
      }
      goto __pyx_L6_except_error;

      /* "src/schema.pxi":129
 *             hasher = Rabin()
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 hasher = hashlib.new(method, **kwargs)
 *             except ValueError:
 */
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L9_try_end:;
    }
  }
  __pyx_L3:;

  /* "src/schema.pxi":133
 *             except ValueError:
 *                 raise InvalidHasher(f'Unknown hash method: {method!r}')
 *         hasher.update(self.canonical_form.encode('utf-8'))             # <<<<<<<<<<<<<<
 *         if self.options.fingerprint_returns_digest:
 *             return hasher.digest()
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_hasher, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_canonical_form); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_encode); if (unlikely(!__pyx_t_10)) __PYX_ERR(23, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_kp_u_utf_8};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(23, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __pyx_t_10 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/schema.pxi":134
 *                 raise InvalidHasher(f'Unknown hash method: {method!r}')
 *         hasher.update(self.canonical_form.encode('utf-8'))
 *         if self.options.fingerprint_returns_digest:             # <<<<<<<<<<<<<<
 *             return hasher.digest()
 *         return hasher
 */
  if (__pyx_v_self->options->fingerprint_returns_digest) {

    /* "src/schema.pxi":135
 *         hasher.update(self.canonical_form.encode('utf-8'))
 *         if self.options.fingerprint_returns_digest:
 *             return hasher.digest()             # <<<<<<<<<<<<<<
 *         return hasher
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_hasher, __pyx_n_s_digest); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 0+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 135, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "src/schema.pxi":134
 *                 raise InvalidHasher(f'Unknown hash method: {method!r}')
 *         hasher.update(self.canonical_form.encode('utf-8'))
 *         if self.options.fingerprint_returns_digest:             # <<<<<<<<<<<<<<
 *             return hasher.digest()
 *         return hasher
 */
  }

  /* "src/schema.pxi":136
 *         if self.options.fingerprint_returns_digest:
 *             return hasher.digest()
 *         return hasher             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType find_type(self, str namespace, str name, bint _raise=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_hasher);
  __pyx_r = __pyx_v_hasher;
  goto __pyx_L0;

  /* "src/schema.pxi":118
 *             return json.dumps(self.schema, indent=2)
 * 
 *     def fingerprint(self, method='rabin', **kwargs) -> Union[bytes, hashlib._hashlib.HASH]:             # <<<<<<<<<<<<<<
 *         """
 *         Return the deterministic fingerprint of the schema, using the given hash method.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("cavro.Schema.fingerprint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hasher);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":138
 *         return hasher
 * 
 *     cpdef AvroType find_type(self, str namespace, str name, bint _raise=True):             # <<<<<<<<<<<<<<
 *         """
 *         Given a namespace and name (namespace may be None), find and return the `AvroType` instance matching this name.
 */

static PyObject *__pyx_pw_5cavro_6Schema_11find_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_AvroType *__pyx_f_5cavro_6Schema_find_type(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_name, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_6Schema_find_type *__pyx_optional_args) {
  int __pyx_v__raise = ((int)1);
  PyObject *__pyx_v_resolved = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_found = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_type", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v__raise = __pyx_optional_args->_raise;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_find_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_6Schema_11find_type)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v__raise); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 138, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_v_namespace, __pyx_v_name, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 138, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(23, 138, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/schema.pxi":142
 *         Given a namespace and name (namespace may be None), find and return the `AvroType` instance matching this name.
 *         """
 *         cdef str resolved = resolve_namespaced_name(namespace, name)             # <<<<<<<<<<<<<<
 *         cdef AvroType found = self.named_types.get(resolved)
 *         if found is not None:
 */
  __pyx_t_1 = __pyx_f_5cavro_resolve_namespaced_name(__pyx_v_namespace, __pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_resolved = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/schema.pxi":143
 *         """
 *         cdef str resolved = resolve_namespaced_name(namespace, name)
 *         cdef AvroType found = self.named_types.get(resolved)             # <<<<<<<<<<<<<<
 *         if found is not None:
 *             return found
 */
  if (unlikely(__pyx_v_self->named_types == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
    __PYX_ERR(23, 143, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_self->named_types, __pyx_v_resolved, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(23, 143, __pyx_L1_error)
  __pyx_v_found = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/schema.pxi":144
 *         cdef str resolved = resolve_namespaced_name(namespace, name)
 *         cdef AvroType found = self.named_types.get(resolved)
 *         if found is not None:             # <<<<<<<<<<<<<<
 *             return found
 *         found = self.options.externally_defined_types.get(resolved)
 */
  __pyx_t_7 = (((PyObject *)__pyx_v_found) != Py_None);
  if (__pyx_t_7) {

    /* "src/schema.pxi":145
 *         cdef AvroType found = self.named_types.get(resolved)
 *         if found is not None:
 *             return found             # <<<<<<<<<<<<<<
 *         found = self.options.externally_defined_types.get(resolved)
 *         if found is not None:
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_found);
    __pyx_r = __pyx_v_found;
    goto __pyx_L0;

    /* "src/schema.pxi":144
 *         cdef str resolved = resolve_namespaced_name(namespace, name)
 *         cdef AvroType found = self.named_types.get(resolved)
 *         if found is not None:             # <<<<<<<<<<<<<<
 *             return found
 *         found = self.options.externally_defined_types.get(resolved)
 */
  }

  /* "src/schema.pxi":146
 *         if found is not None:
 *             return found
 *         found = self.options.externally_defined_types.get(resolved)             # <<<<<<<<<<<<<<
 *         if found is not None:
 *             self.register_type(namespace, name, found)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->options->externally_defined_types, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_resolved};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(23, 146, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_found, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "src/schema.pxi":147
 *             return found
 *         found = self.options.externally_defined_types.get(resolved)
 *         if found is not None:             # <<<<<<<<<<<<<<
 *             self.register_type(namespace, name, found)
 *             return found
 */
  __pyx_t_7 = (((PyObject *)__pyx_v_found) != Py_None);
  if (__pyx_t_7) {

    /* "src/schema.pxi":148
 *         found = self.options.externally_defined_types.get(resolved)
 *         if found is not None:
 *             self.register_type(namespace, name, found)             # <<<<<<<<<<<<<<
 *             return found
 *         found = self.named_types.get(name)
 */
    ((struct __pyx_vtabstruct_5cavro_Schema *)__pyx_v_self->__pyx_vtab)->register_type(__pyx_v_self, __pyx_v_namespace, __pyx_v_name, __pyx_v_found); if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 148, __pyx_L1_error)

    /* "src/schema.pxi":149
 *         if found is not None:
 *             self.register_type(namespace, name, found)
 *             return found             # <<<<<<<<<<<<<<
 *         found = self.named_types.get(name)
 *         if found is not None:
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_found);
    __pyx_r = __pyx_v_found;
    goto __pyx_L0;

    /* "src/schema.pxi":147
 *             return found
 *         found = self.options.externally_defined_types.get(resolved)
 *         if found is not None:             # <<<<<<<<<<<<<<
 *             self.register_type(namespace, name, found)
 *             return found
 */
  }

  /* "src/schema.pxi":150
 *             self.register_type(namespace, name, found)
 *             return found
 *         found = self.named_types.get(name)             # <<<<<<<<<<<<<<
 *         if found is not None:
 *             return found
 */
  if (unlikely(__pyx_v_self->named_types == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
    __PYX_ERR(23, 150, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_self->named_types, __pyx_v_name, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(23, 150, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_found, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "src/schema.pxi":151
 *             return found
 *         found = self.named_types.get(name)
 *         if found is not None:             # <<<<<<<<<<<<<<
 *             return found
 *         found = self.options.externally_defined_types.get(name)
 */
  __pyx_t_7 = (((PyObject *)__pyx_v_found) != Py_None);
  if (__pyx_t_7) {

    /* "src/schema.pxi":152
 *         found = self.named_types.get(name)
 *         if found is not None:
 *             return found             # <<<<<<<<<<<<<<
 *         found = self.options.externally_defined_types.get(name)
 *         if found is not None:
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_found);
    __pyx_r = __pyx_v_found;
    goto __pyx_L0;

    /* "src/schema.pxi":151
 *             return found
 *         found = self.named_types.get(name)
 *         if found is not None:             # <<<<<<<<<<<<<<
 *             return found
 *         found = self.options.externally_defined_types.get(name)
 */
  }

  /* "src/schema.pxi":153
 *         if found is not None:
 *             return found
 *         found = self.options.externally_defined_types.get(name)             # <<<<<<<<<<<<<<
 *         if found is not None:
 *             return found
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->options->externally_defined_types, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_name};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(23, 153, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_found, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "src/schema.pxi":154
 *             return found
 *         found = self.options.externally_defined_types.get(name)
 *         if found is not None:             # <<<<<<<<<<<<<<
 *             return found
 *         if _raise:
 */
  __pyx_t_7 = (((PyObject *)__pyx_v_found) != Py_None);
  if (__pyx_t_7) {

    /* "src/schema.pxi":155
 *         found = self.options.externally_defined_types.get(name)
 *         if found is not None:
 *             return found             # <<<<<<<<<<<<<<
 *         if _raise:
 *             raise UnknownType(f'Unknown type: {resolved!r}')
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_found);
    __pyx_r = __pyx_v_found;
    goto __pyx_L0;

    /* "src/schema.pxi":154
 *             return found
 *         found = self.options.externally_defined_types.get(name)
 *         if found is not None:             # <<<<<<<<<<<<<<
 *             return found
 *         if _raise:
 */
  }

  /* "src/schema.pxi":156
 *         if found is not None:
 *             return found
 *         if _raise:             # <<<<<<<<<<<<<<
 *             raise UnknownType(f'Unknown type: {resolved!r}')
 * 
 */
  if (unlikely(__pyx_v__raise)) {

    /* "src/schema.pxi":157
 *             return found
 *         if _raise:
 *             raise UnknownType(f'Unknown type: {resolved!r}')             # <<<<<<<<<<<<<<
 * 
 *     def can_encode(self, value: object) -> bool:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_UnknownType); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_resolved), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(23, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Unknown_type, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 157, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(23, 157, __pyx_L1_error)

    /* "src/schema.pxi":156
 *         if found is not None:
 *             return found
 *         if _raise:             # <<<<<<<<<<<<<<
 *             raise UnknownType(f'Unknown type: {resolved!r}')
 * 
 */
  }

  /* "src/schema.pxi":138
 *         return hasher
 * 
 *     cpdef AvroType find_type(self, str namespace, str name, bint _raise=True):             # <<<<<<<<<<<<<<
 *         """
 *         Given a namespace and name (namespace may be None), find and return the `AvroType` instance matching this name.
 */

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_5cavro_AvroType *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.Schema.find_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_resolved);
  __Pyx_XDECREF((PyObject *)__pyx_v_found);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_11find_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_6Schema_10find_type, "\n        Given a namespace and name (namespace may be None), find and return the `AvroType` instance matching this name.\n        ");
static PyMethodDef __pyx_mdef_5cavro_6Schema_11find_type = {"find_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_11find_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_10find_type};
static PyObject *__pyx_pw_5cavro_6Schema_11find_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_namespace = 0;
  PyObject *__pyx_v_name = 0;
  int __pyx_v__raise;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_namespace,&__pyx_n_s_name,&__pyx_n_s_raise,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_namespace)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 138, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 138, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("find_type", 0, 2, 3, 1); __PYX_ERR(23, 138, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_raise);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 138, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "find_type") < 0)) __PYX_ERR(23, 138, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_namespace = ((PyObject*)values[0]);
    __pyx_v_name = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v__raise = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v__raise == (int)-1) && PyErr_Occurred())) __PYX_ERR(23, 138, __pyx_L3_error)
    } else {
      __pyx_v__raise = ((int)1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_type", 0, 2, 3, __pyx_nargs); __PYX_ERR(23, 138, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Schema.find_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_namespace), (&PyUnicode_Type), 1, "namespace", 1))) __PYX_ERR(23, 138, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(23, 138, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_6Schema_10find_type(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v_namespace, __pyx_v_name, __pyx_v__raise);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_10find_type(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_name, int __pyx_v__raise) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5cavro_6Schema_find_type __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_type", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2._raise = __pyx_v__raise;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_5cavro_Schema->find_type(__pyx_v_self, __pyx_v_namespace, __pyx_v_name, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Schema.find_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":159
 *             raise UnknownType(f'Unknown type: {resolved!r}')
 * 
 *     def can_encode(self, value: object) -> bool:             # <<<<<<<<<<<<<<
 *         """
 *         Check if `value` can be encoded using this schema
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_13can_encode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_6Schema_12can_encode, "\n        Check if `value` can be encoded using this schema\n        ");
static PyMethodDef __pyx_mdef_5cavro_6Schema_13can_encode = {"can_encode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_13can_encode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_12can_encode};
static PyObject *__pyx_pw_5cavro_6Schema_13can_encode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("can_encode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 159, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "can_encode") < 0)) __PYX_ERR(23, 159, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("can_encode", 1, 1, 1, __pyx_nargs); __PYX_ERR(23, 159, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Schema.can_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6Schema_12can_encode(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_12can_encode(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_v_fitness;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("can_encode", 1);

  /* "src/schema.pxi":163
 *         Check if `value` can be encoded using this schema
 *         """
 *         fitness = self.type.get_value_fitness(value)             # <<<<<<<<<<<<<<
 *         return fitness > FIT_NONE
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->get_value_fitness(__pyx_v_self->type, __pyx_v_value); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(23, 163, __pyx_L1_error)
  __pyx_v_fitness = __pyx_t_1;

  /* "src/schema.pxi":164
 *         """
 *         fitness = self.type.get_value_fitness(value)
 *         return fitness > FIT_NONE             # <<<<<<<<<<<<<<
 * 
 *     def binary_encode(self, value: object) -> bytes:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong((__pyx_v_fitness > __pyx_v_5cavro_FIT_NONE)); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/schema.pxi":159
 *             raise UnknownType(f'Unknown type: {resolved!r}')
 * 
 *     def can_encode(self, value: object) -> bool:             # <<<<<<<<<<<<<<
 *         """
 *         Check if `value` can be encoded using this schema
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.Schema.can_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":166
 *         return fitness > FIT_NONE
 * 
 *     def binary_encode(self, value: object) -> bytes:             # <<<<<<<<<<<<<<
 *         """
 *         Encode `value` using this schema and return the avro bytes representing it.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_15binary_encode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_6Schema_14binary_encode, "\n        Encode `value` using this schema and return the avro bytes representing it.\n        ");
static PyMethodDef __pyx_mdef_5cavro_6Schema_15binary_encode = {"binary_encode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_15binary_encode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_14binary_encode};
static PyObject *__pyx_pw_5cavro_6Schema_15binary_encode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("binary_encode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 166, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "binary_encode") < 0)) __PYX_ERR(23, 166, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("binary_encode", 1, 1, 1, __pyx_nargs); __PYX_ERR(23, 166, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Schema.binary_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6Schema_14binary_encode(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_14binary_encode(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_MemoryWriter *__pyx_v_buffer = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("binary_encode", 1);

  /* "src/schema.pxi":170
 *         Encode `value` using this schema and return the avro bytes representing it.
 *         """
 *         cdef MemoryWriter buffer = MemoryWriter()             # <<<<<<<<<<<<<<
 *         self.type.binary_buffer_encode(buffer, value)
 *         return buffer.bytes()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro_MemoryWriter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buffer = ((struct __pyx_obj_5cavro_MemoryWriter *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/schema.pxi":171
 *         """
 *         cdef MemoryWriter buffer = MemoryWriter()
 *         self.type.binary_buffer_encode(buffer, value)             # <<<<<<<<<<<<<<
 *         return buffer.bytes()
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->binary_buffer_encode(__pyx_v_self->type, ((struct __pyx_obj_5cavro__Writer *)__pyx_v_buffer), __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(23, 171, __pyx_L1_error)

  /* "src/schema.pxi":172
 *         cdef MemoryWriter buffer = MemoryWriter()
 *         self.type.binary_buffer_encode(buffer, value)
 *         return buffer.bytes()             # <<<<<<<<<<<<<<
 * 
 *     def binary_decode(self, bytes value: bytes) -> object:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_12MemoryWriter_bytes(__pyx_v_buffer); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/schema.pxi":166
 *         return fitness > FIT_NONE
 * 
 *     def binary_encode(self, value: object) -> bytes:             # <<<<<<<<<<<<<<
 *         """
 *         Encode `value` using this schema and return the avro bytes representing it.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Schema.binary_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_buffer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":174
 *         return buffer.bytes()
 * 
 *     def binary_decode(self, bytes value: bytes) -> object:             # <<<<<<<<<<<<<<
 *         """
 *         Decode `value` using this schema and return the decoded value.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_17binary_decode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_6Schema_16binary_decode, "\n        Decode `value` using this schema and return the decoded value.\n        ");
static PyMethodDef __pyx_mdef_5cavro_6Schema_17binary_decode = {"binary_decode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_17binary_decode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_16binary_decode};
static PyObject *__pyx_pw_5cavro_6Schema_17binary_decode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("binary_decode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 174, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "binary_decode") < 0)) __PYX_ERR(23, 174, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("binary_decode", 1, 1, 1, __pyx_nargs); __PYX_ERR(23, 174, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Schema.binary_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyBytes_Type), 0, "value", 1))) __PYX_ERR(23, 174, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_6Schema_16binary_decode(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_16binary_decode(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_5cavro_MemoryReader *__pyx_v_buffer = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("binary_decode", 1);

  /* "src/schema.pxi":178
 *         Decode `value` using this schema and return the decoded value.
 *         """
 *         cdef MemoryReader buffer = MemoryReader(value)             # <<<<<<<<<<<<<<
 *         return self.type.binary_buffer_decode(buffer)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro_MemoryReader), __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buffer = ((struct __pyx_obj_5cavro_MemoryReader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/schema.pxi":179
 *         """
 *         cdef MemoryReader buffer = MemoryReader(value)
 *         return self.type.binary_buffer_decode(buffer)             # <<<<<<<<<<<<<<
 * 
 *     cpdef binary_read(self, _Reader reader: _Reader):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->binary_buffer_decode(__pyx_v_self->type, ((struct __pyx_obj_5cavro__Reader *)__pyx_v_buffer)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/schema.pxi":174
 *         return buffer.bytes()
 * 
 *     def binary_decode(self, bytes value: bytes) -> object:             # <<<<<<<<<<<<<<
 *         """
 *         Decode `value` using this schema and return the decoded value.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Schema.binary_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_buffer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":181
 *         return self.type.binary_buffer_decode(buffer)
 * 
 *     cpdef binary_read(self, _Reader reader: _Reader):             # <<<<<<<<<<<<<<
 *         """
 *         Read a value from `reader` using this schema and return the decoded value.
 */

static PyObject *__pyx_pw_5cavro_6Schema_19binary_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_6Schema_binary_read(struct __pyx_obj_5cavro_Schema *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("binary_read", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_binary_read); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_6Schema_19binary_read)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_reader)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 181, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/schema.pxi":186
 *         `reader` may be a `MemoryReader` or `FileReader` instance.
 *         """
 *         return self.type.binary_buffer_decode(reader)             # <<<<<<<<<<<<<<
 * 
 *     cpdef binary_write(self, _Writer writer: _Writer, value: object):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->binary_buffer_decode(__pyx_v_self->type, __pyx_v_reader); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/schema.pxi":181
 *         return self.type.binary_buffer_decode(buffer)
 * 
 *     cpdef binary_read(self, _Reader reader: _Reader):             # <<<<<<<<<<<<<<
 *         """
 *         Read a value from `reader` using this schema and return the decoded value.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Schema.binary_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_19binary_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_6Schema_18binary_read, "\n        Read a value from `reader` using this schema and return the decoded value.\n        `reader` may be a `MemoryReader` or `FileReader` instance.\n        ");
static PyMethodDef __pyx_mdef_5cavro_6Schema_19binary_read = {"binary_read", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_19binary_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_18binary_read};
static PyObject *__pyx_pw_5cavro_6Schema_19binary_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_5cavro__Reader *__pyx_v_reader = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("binary_read (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_reader,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_reader)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 181, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "binary_read") < 0)) __PYX_ERR(23, 181, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_reader = ((struct __pyx_obj_5cavro__Reader *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("binary_read", 1, 1, 1, __pyx_nargs); __PYX_ERR(23, 181, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Schema.binary_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_reader), __pyx_ptype_5cavro__Reader, 0, "reader", 0))) __PYX_ERR(23, 181, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_6Schema_18binary_read(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v_reader);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_18binary_read(struct __pyx_obj_5cavro_Schema *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("binary_read", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_6Schema_binary_read(__pyx_v_self, __pyx_v_reader, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Schema.binary_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":188
 *         return self.type.binary_buffer_decode(reader)
 * 
 *     cpdef binary_write(self, _Writer writer: _Writer, value: object):             # <<<<<<<<<<<<<<
 *         """
 *         Write `value` to `writer` using this schema.
 */

static PyObject *__pyx_pw_5cavro_6Schema_21binary_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_6Schema_binary_write(struct __pyx_obj_5cavro_Schema *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, PyObject *__pyx_v_value, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("binary_write", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_binary_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_6Schema_21binary_write)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_v_writer), __pyx_v_value};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 188, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/schema.pxi":193
 *         `writer` may be a `MemoryWriter` or `FileWriter` instance.
 *         """
 *         self.type.binary_buffer_encode(writer, value)             # <<<<<<<<<<<<<<
 * 
 *     def json_encode(self, value, serialize=True, **kwargs):
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->binary_buffer_encode(__pyx_v_self->type, __pyx_v_writer, __pyx_v_value); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(23, 193, __pyx_L1_error)

  /* "src/schema.pxi":188
 *         return self.type.binary_buffer_decode(reader)
 * 
 *     cpdef binary_write(self, _Writer writer: _Writer, value: object):             # <<<<<<<<<<<<<<
 *         """
 *         Write `value` to `writer` using this schema.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Schema.binary_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_21binary_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_6Schema_20binary_write, "\n        Write `value` to `writer` using this schema.\n        `writer` may be a `MemoryWriter` or `FileWriter` instance.\n        ");
static PyMethodDef __pyx_mdef_5cavro_6Schema_21binary_write = {"binary_write", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_21binary_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_20binary_write};
static PyObject *__pyx_pw_5cavro_6Schema_21binary_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_5cavro__Writer *__pyx_v_writer = 0;
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("binary_write (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_writer,&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_writer)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 188, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 188, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("binary_write", 1, 2, 2, 1); __PYX_ERR(23, 188, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "binary_write") < 0)) __PYX_ERR(23, 188, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_writer = ((struct __pyx_obj_5cavro__Writer *)values[0]);
    __pyx_v_value = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("binary_write", 1, 2, 2, __pyx_nargs); __PYX_ERR(23, 188, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Schema.binary_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_writer), __pyx_ptype_5cavro__Writer, 0, "writer", 0))) __PYX_ERR(23, 188, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_6Schema_20binary_write(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v_writer, __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_20binary_write(struct __pyx_obj_5cavro_Schema *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("binary_write", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_6Schema_binary_write(__pyx_v_self, __pyx_v_writer, __pyx_v_value, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Schema.binary_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":195
 *         self.type.binary_buffer_encode(writer, value)
 * 
 *     def json_encode(self, value, serialize=True, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Encode `value` using this schema and return the avro json representing it.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_23json_encode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_6Schema_22json_encode, "\n        Encode `value` using this schema and return the avro json representing it.\n        ");
static PyMethodDef __pyx_mdef_5cavro_6Schema_23json_encode = {"json_encode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_23json_encode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_22json_encode};
static PyObject *__pyx_pw_5cavro_6Schema_23json_encode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_serialize = 0;
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("json_encode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,&__pyx_n_s_serialize,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_True));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 195, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_serialize);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 195, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, kwd_pos_args, "json_encode") < 0)) __PYX_ERR(23, 195, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_value = values[0];
    __pyx_v_serialize = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("json_encode", 0, 1, 2, __pyx_nargs); __PYX_ERR(23, 195, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("cavro.Schema.json_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6Schema_22json_encode(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v_value, __pyx_v_serialize, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_22json_encode(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_serialize, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("json_encode", 1);

  /* "src/schema.pxi":199
 *         Encode `value` using this schema and return the avro json representing it.
 *         """
 *         data = self.type.json_format(value)             # <<<<<<<<<<<<<<
 *         if serialize:
 *             return json.dumps(data, **kwargs)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->json_format(__pyx_v_self->type, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/schema.pxi":200
 *         """
 *         data = self.type.json_format(value)
 *         if serialize:             # <<<<<<<<<<<<<<
 *             return json.dumps(data, **kwargs)
 *         return data
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_serialize); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(23, 200, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "src/schema.pxi":201
 *         data = self.type.json_format(value)
 *         if serialize:
 *             return json.dumps(data, **kwargs)             # <<<<<<<<<<<<<<
 *         return data
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_json); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_dumps); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_data);
    __Pyx_GIVEREF(__pyx_v_data);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_data)) __PYX_ERR(23, 201, __pyx_L1_error);
    __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(23, 201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "src/schema.pxi":200
 *         """
 *         data = self.type.json_format(value)
 *         if serialize:             # <<<<<<<<<<<<<<
 *             return json.dumps(data, **kwargs)
 *         return data
 */
  }

  /* "src/schema.pxi":202
 *         if serialize:
 *             return json.dumps(data, **kwargs)
 *         return data             # <<<<<<<<<<<<<<
 * 
 *     def json_decode(self, value: Union[str, object], deserialize: bool=True, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_data);
  __pyx_r = __pyx_v_data;
  goto __pyx_L0;

  /* "src/schema.pxi":195
 *         self.type.binary_buffer_encode(writer, value)
 * 
 *     def json_encode(self, value, serialize=True, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Encode `value` using this schema and return the avro json representing it.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.Schema.json_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":204
 *         return data
 * 
 *     def json_decode(self, value: Union[str, object], deserialize: bool=True, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Decode `value` in JSON form using this schema and return the decoded value.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_25json_decode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_6Schema_24json_decode, "\n        Decode `value` in JSON form using this schema and return the decoded value.\n\n        If `deserialize` is True, then value must be a `str` containing the serialized JSON value.\n        If `deserialize` is False, then value must be a python object representing the JSON value.\n        ");
static PyMethodDef __pyx_mdef_5cavro_6Schema_25json_decode = {"json_decode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_25json_decode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_24json_decode};
static PyObject *__pyx_pw_5cavro_6Schema_25json_decode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  __pyx_t_5cavro_bool __pyx_v_deserialize;
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("json_decode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,&__pyx_n_s_deserialize,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 204, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_deserialize);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 204, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, kwd_pos_args, "json_decode") < 0)) __PYX_ERR(23, 204, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_value = values[0];
    if (values[1]) {
      __pyx_v_deserialize = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_deserialize == (int)-1) && PyErr_Occurred())) __PYX_ERR(23, 204, __pyx_L3_error)
    } else {
      __pyx_v_deserialize = ((__pyx_t_5cavro_bool)1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("json_decode", 0, 1, 2, __pyx_nargs); __PYX_ERR(23, 204, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("cavro.Schema.json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6Schema_24json_decode(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v_value, __pyx_v_deserialize, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_24json_decode(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v_value, __pyx_t_5cavro_bool __pyx_v_deserialize, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("json_decode", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/schema.pxi":211
 *         If `deserialize` is False, then value must be a python object representing the JSON value.
 *         """
 *         if deserialize:             # <<<<<<<<<<<<<<
 *             value = json.loads(value, **kwargs)
 *         return self.type.json_decode(value)
 */
  __pyx_t_1 = (__pyx_v_deserialize != 0);
  if (__pyx_t_1) {

    /* "src/schema.pxi":212
 *         """
 *         if deserialize:
 *             value = json.loads(value, **kwargs)             # <<<<<<<<<<<<<<
 *         return self.type.json_decode(value)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_json); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_loads); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_value)) __PYX_ERR(23, 212, __pyx_L1_error);
    __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(23, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "src/schema.pxi":211
 *         If `deserialize` is False, then value must be a python object representing the JSON value.
 *         """
 *         if deserialize:             # <<<<<<<<<<<<<<
 *             value = json.loads(value, **kwargs)
 *         return self.type.json_decode(value)
 */
  }

  /* "src/schema.pxi":213
 *         if deserialize:
 *             value = json.loads(value, **kwargs)
 *         return self.type.json_decode(value)             # <<<<<<<<<<<<<<
 * 
 *     cpdef Schema reader_for_writer(self, Schema writer_schema: Schema):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->json_decode(__pyx_v_self->type, __pyx_v_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "src/schema.pxi":204
 *         return data
 * 
 *     def json_decode(self, value: Union[str, object], deserialize: bool=True, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Decode `value` in JSON form using this schema and return the decoded value.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.Schema.json_decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":215
 *         return self.type.json_decode(value)
 * 
 *     cpdef Schema reader_for_writer(self, Schema writer_schema: Schema):             # <<<<<<<<<<<<<<
 *         """
 *         Return a schema that is the result of promoting this schema to the writer schema.
 */

static PyObject *__pyx_pw_5cavro_6Schema_27reader_for_writer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static struct __pyx_obj_5cavro_Schema *__pyx_f_5cavro_6Schema_reader_for_writer(struct __pyx_obj_5cavro_Schema *__pyx_v_self, struct __pyx_obj_5cavro_Schema *__pyx_v_writer_schema, int __pyx_skip_dispatch) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_new_type = NULL;
  struct __pyx_obj_5cavro_Schema *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reader_for_writer", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reader_for_writer); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_6Schema_27reader_for_writer)) {
        __Pyx_XDECREF((PyObject *)__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_writer_schema)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 215, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_Schema))))) __PYX_ERR(23, 215, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/schema.pxi":221
 *         The returned schema may only be used for reading, and should return values that match the reader schema.
 *         """
 *         new_type = self.type.for_writer(writer_schema.type)             # <<<<<<<<<<<<<<
 *         return ResolvedSchema(new_type, self.options)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_writer_schema->type);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_AvroType *)__pyx_v_self->type->__pyx_vtab)->for_writer(__pyx_v_self->type, ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1), NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_new_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/schema.pxi":222
 *         """
 *         new_type = self.type.for_writer(writer_schema.type)
 *         return ResolvedSchema(new_type, self.options)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_new_type);
  __Pyx_GIVEREF((PyObject *)__pyx_v_new_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_new_type))) __PYX_ERR(23, 222, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self->options))) __PYX_ERR(23, 222, __pyx_L1_error);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_ResolvedSchema), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/schema.pxi":215
 *         return self.type.json_decode(value)
 * 
 *     cpdef Schema reader_for_writer(self, Schema writer_schema: Schema):             # <<<<<<<<<<<<<<
 *         """
 *         Return a schema that is the result of promoting this schema to the writer schema.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Schema.reader_for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_new_type);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_27reader_for_writer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_5cavro_6Schema_26reader_for_writer, "\n        Return a schema that is the result of promoting this schema to the writer schema.\n        \n        The returned schema may only be used for reading, and should return values that match the reader schema.\n        ");
static PyMethodDef __pyx_mdef_5cavro_6Schema_27reader_for_writer = {"reader_for_writer", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_27reader_for_writer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_26reader_for_writer};
static PyObject *__pyx_pw_5cavro_6Schema_27reader_for_writer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_5cavro_Schema *__pyx_v_writer_schema = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reader_for_writer (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_writer_schema,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_writer_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 215, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "reader_for_writer") < 0)) __PYX_ERR(23, 215, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_writer_schema = ((struct __pyx_obj_5cavro_Schema *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("reader_for_writer", 1, 1, 1, __pyx_nargs); __PYX_ERR(23, 215, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Schema.reader_for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_writer_schema), __pyx_ptype_5cavro_Schema, 0, "writer_schema", 0))) __PYX_ERR(23, 215, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_6Schema_26reader_for_writer(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v_writer_schema);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_26reader_for_writer(struct __pyx_obj_5cavro_Schema *__pyx_v_self, struct __pyx_obj_5cavro_Schema *__pyx_v_writer_schema) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reader_for_writer", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_6Schema_reader_for_writer(__pyx_v_self, __pyx_v_writer_schema, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Schema.reader_for_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":40
 * 
 * 
 *     cdef readonly dict named_types             # <<<<<<<<<<<<<<
 *     cdef readonly object source
 *     cdef readonly Options options
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_11named_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_6Schema_11named_types_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6Schema_11named_types___get__(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_11named_types___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->named_types);
  __pyx_r = __pyx_v_self->named_types;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":41
 * 
 *     cdef readonly dict named_types
 *     cdef readonly object source             # <<<<<<<<<<<<<<
 *     cdef readonly Options options
 *     cdef readonly AvroType type
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_6source_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_6Schema_6source_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6Schema_6source___get__(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_6source___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->source);
  __pyx_r = __pyx_v_self->source;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":42
 *     cdef readonly dict named_types
 *     cdef readonly object source
 *     cdef readonly Options options             # <<<<<<<<<<<<<<
 *     cdef readonly AvroType type
 *     cdef str _canonical_form
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_7options_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_6Schema_7options_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6Schema_7options___get__(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_7options___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->options);
  __pyx_r = ((PyObject *)__pyx_v_self->options);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":43
 *     cdef readonly object source
 *     cdef readonly Options options
 *     cdef readonly AvroType type             # <<<<<<<<<<<<<<
 *     cdef str _canonical_form
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_4type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_6Schema_4type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6Schema_4type___get__(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_4type___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->type);
  __pyx_r = ((PyObject *)__pyx_v_self->type);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":46
 *     cdef str _canonical_form
 * 
 *     cdef readonly dict logical_types             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, source: Union[str, object], Options options: Options=DEFAULT_OPTIONS, *, named_types: dict[str, AvroType]=None, parse_json: bool=True, _type=None, **extra_options):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_13logical_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_6Schema_13logical_types_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_6Schema_13logical_types___get__(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_13logical_types___get__(struct __pyx_obj_5cavro_Schema *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->logical_types);
  __pyx_r = __pyx_v_self->logical_types;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_29__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_6Schema_29__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_29__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_6Schema_29__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_6Schema_28__reduce_cython__(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_28__reduce_cython__(struct __pyx_obj_5cavro_Schema *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._canonical_form, self.logical_types, self.named_types, self.options, self.source, self.type)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_canonical_form);
  __Pyx_GIVEREF(__pyx_v_self->_canonical_form);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->_canonical_form)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->logical_types);
  __Pyx_GIVEREF(__pyx_v_self->logical_types);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->logical_types)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->named_types);
  __Pyx_GIVEREF(__pyx_v_self->named_types);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->named_types)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, ((PyObject *)__pyx_v_self->options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->source);
  __Pyx_GIVEREF(__pyx_v_self->source);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_self->source)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->type);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, ((PyObject *)__pyx_v_self->type))) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._canonical_form, self.logical_types, self.named_types, self.options, self.source, self.type)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._canonical_form, self.logical_types, self.named_types, self.options, self.source, self.type)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._canonical_form is not None or self.logical_types is not None or self.named_types is not None or self.options is not None or self.source is not None or self.type is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._canonical_form, self.logical_types, self.named_types, self.options, self.source, self.type)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._canonical_form is not None or self.logical_types is not None or self.named_types is not None or self.options is not None or self.source is not None or self.type is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Schema, (type(self), 0xc2592e1, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->_canonical_form != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->logical_types != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->named_types != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->source != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->type) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._canonical_form is not None or self.logical_types is not None or self.named_types is not None or self.options is not None or self.source is not None or self.type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Schema, (type(self), 0xc2592e1, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._canonical_form is not None or self.logical_types is not None or self.named_types is not None or self.options is not None or self.source is not None or self.type is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Schema, (type(self), 0xc2592e1, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Schema, (type(self), 0xc2592e1, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Schema); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_203789025);
    __Pyx_GIVEREF(__pyx_int_203789025);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_203789025)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._canonical_form is not None or self.logical_types is not None or self.named_types is not None or self.options is not None or self.source is not None or self.type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Schema, (type(self), 0xc2592e1, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Schema, (type(self), 0xc2592e1, None), state
 *     else:
 *         return __pyx_unpickle_Schema, (type(self), 0xc2592e1, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Schema__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Schema); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_203789025);
    __Pyx_GIVEREF(__pyx_int_203789025);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_203789025)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.Schema.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Schema, (type(self), 0xc2592e1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Schema__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_6Schema_31__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_6Schema_31__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_31__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_6Schema_31__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Schema.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_6Schema_30__setstate_cython__(((struct __pyx_obj_5cavro_Schema *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_6Schema_30__setstate_cython__(struct __pyx_obj_5cavro_Schema *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Schema, (type(self), 0xc2592e1, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Schema__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_Schema__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Schema, (type(self), 0xc2592e1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Schema__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Schema.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/schema.pxi":232
 *     """
 * 
 *     def __init__(self, AvroType resolved_type, Options options=DEFAULT_OPTIONS):             # <<<<<<<<<<<<<<
 *         named_types = {}
 *         for sub_type in resolved_type.walk_types(set()):
 */

/* Python wrapper */
static int __pyx_pw_5cavro_14ResolvedSchema_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_14ResolvedSchema_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_resolved_type = 0;
  struct __pyx_obj_5cavro_Options *__pyx_v_options = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_resolved_type,&__pyx_n_s_options,0};
    values[1] = __Pyx_Arg_NewRef_VARARGS((PyObject *)__pyx_k__84);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_resolved_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 232, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_options);
          if (value) { values[1] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 232, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(23, 232, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_resolved_type = ((struct __pyx_obj_5cavro_AvroType *)values[0]);
    __pyx_v_options = ((struct __pyx_obj_5cavro_Options *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, __pyx_nargs); __PYX_ERR(23, 232, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ResolvedSchema.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_resolved_type), __pyx_ptype_5cavro_AvroType, 1, "resolved_type", 0))) __PYX_ERR(23, 232, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_options), __pyx_ptype_5cavro_Options, 1, "options", 0))) __PYX_ERR(23, 232, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_14ResolvedSchema___init__(((struct __pyx_obj_5cavro_ResolvedSchema *)__pyx_v_self), __pyx_v_resolved_type, __pyx_v_options);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_14ResolvedSchema___init__(struct __pyx_obj_5cavro_ResolvedSchema *__pyx_v_self, struct __pyx_obj_5cavro_AvroType *__pyx_v_resolved_type, struct __pyx_obj_5cavro_Options *__pyx_v_options) {
  PyObject *__pyx_v_named_types = NULL;
  PyObject *__pyx_v_sub_type = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/schema.pxi":233
 * 
 *     def __init__(self, AvroType resolved_type, Options options=DEFAULT_OPTIONS):
 *         named_types = {}             # <<<<<<<<<<<<<<
 *         for sub_type in resolved_type.walk_types(set()):
 *             if isinstance(sub_type, _NamedType):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_named_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/schema.pxi":234
 *     def __init__(self, AvroType resolved_type, Options options=DEFAULT_OPTIONS):
 *         named_types = {}
 *         for sub_type in resolved_type.walk_types(set()):             # <<<<<<<<<<<<<<
 *             if isinstance(sub_type, _NamedType):
 *                 named_types[sub_type.type] = sub_type
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_resolved_type), __pyx_n_s_walk_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 234, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(23, 234, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(23, 234, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 234, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(23, 234, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(23, 234, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 234, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(23, 234, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_sub_type, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/schema.pxi":235
 *         named_types = {}
 *         for sub_type in resolved_type.walk_types(set()):
 *             if isinstance(sub_type, _NamedType):             # <<<<<<<<<<<<<<
 *                 named_types[sub_type.type] = sub_type
 * 
 */
    __pyx_t_8 = __Pyx_TypeCheck(__pyx_v_sub_type, __pyx_ptype_5cavro__NamedType); 
    if (__pyx_t_8) {

      /* "src/schema.pxi":236
 *         for sub_type in resolved_type.walk_types(set()):
 *             if isinstance(sub_type, _NamedType):
 *                 named_types[sub_type.type] = sub_type             # <<<<<<<<<<<<<<
 * 
 *         self.named_types = named_types
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sub_type, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((PyDict_SetItem(__pyx_v_named_types, __pyx_t_1, __pyx_v_sub_type) < 0))) __PYX_ERR(23, 236, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "src/schema.pxi":235
 *         named_types = {}
 *         for sub_type in resolved_type.walk_types(set()):
 *             if isinstance(sub_type, _NamedType):             # <<<<<<<<<<<<<<
 *                 named_types[sub_type.type] = sub_type
 * 
 */
    }

    /* "src/schema.pxi":234
 *     def __init__(self, AvroType resolved_type, Options options=DEFAULT_OPTIONS):
 *         named_types = {}
 *         for sub_type in resolved_type.walk_types(set()):             # <<<<<<<<<<<<<<
 *             if isinstance(sub_type, _NamedType):
 *                 named_types[sub_type.type] = sub_type
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/schema.pxi":238
 *                 named_types[sub_type.type] = sub_type
 * 
 *         self.named_types = named_types             # <<<<<<<<<<<<<<
 *         self.source = None
 *         self.options = options
 */
  __Pyx_INCREF(__pyx_v_named_types);
  __Pyx_GIVEREF(__pyx_v_named_types);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.named_types);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.named_types);
  __pyx_v_self->__pyx_base.named_types = __pyx_v_named_types;

  /* "src/schema.pxi":239
 * 
 *         self.named_types = named_types
 *         self.source = None             # <<<<<<<<<<<<<<
 *         self.options = options
 *         self.type = resolved_type
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.source);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.source);
  __pyx_v_self->__pyx_base.source = Py_None;

  /* "src/schema.pxi":240
 *         self.named_types = named_types
 *         self.source = None
 *         self.options = options             # <<<<<<<<<<<<<<
 *         self.type = resolved_type
 *         self.logical_types = self._make_logical_types(options)
 */
  __Pyx_INCREF((PyObject *)__pyx_v_options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_options);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __pyx_v_self->__pyx_base.options = __pyx_v_options;

  /* "src/schema.pxi":241
 *         self.source = None
 *         self.options = options
 *         self.type = resolved_type             # <<<<<<<<<<<<<<
 *         self.logical_types = self._make_logical_types(options)
 */
  __Pyx_INCREF((PyObject *)__pyx_v_resolved_type);
  __Pyx_GIVEREF((PyObject *)__pyx_v_resolved_type);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->__pyx_base.type);
  __Pyx_DECREF((PyObject *)__pyx_v_self->__pyx_base.type);
  __pyx_v_self->__pyx_base.type = __pyx_v_resolved_type;

  /* "src/schema.pxi":242
 *         self.options = options
 *         self.type = resolved_type
 *         self.logical_types = self._make_logical_types(options)             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_make_logical_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_options)};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_2))) __PYX_ERR(23, 242, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.logical_types);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.logical_types);
  __pyx_v_self->__pyx_base.logical_types = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/schema.pxi":232
 *     """
 * 
 *     def __init__(self, AvroType resolved_type, Options options=DEFAULT_OPTIONS):             # <<<<<<<<<<<<<<
 *         named_types = {}
 *         for sub_type in resolved_type.walk_types(set()):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.ResolvedSchema.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_named_types);
  __Pyx_XDECREF(__pyx_v_sub_type);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14ResolvedSchema_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_14ResolvedSchema_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14ResolvedSchema_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_14ResolvedSchema_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_14ResolvedSchema_2__reduce_cython__(((struct __pyx_obj_5cavro_ResolvedSchema *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14ResolvedSchema_2__reduce_cython__(struct __pyx_obj_5cavro_ResolvedSchema *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._canonical_form, self.logical_types, self.named_types, self.options, self.source, self.type)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._canonical_form);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._canonical_form);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base._canonical_form)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.logical_types);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.logical_types);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->__pyx_base.logical_types)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.named_types);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.named_types);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->__pyx_base.named_types)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.options);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.options);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, ((PyObject *)__pyx_v_self->__pyx_base.options))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.source);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.source);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_self->__pyx_base.source)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.type);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, ((PyObject *)__pyx_v_self->__pyx_base.type))) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._canonical_form, self.logical_types, self.named_types, self.options, self.source, self.type)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._canonical_form, self.logical_types, self.named_types, self.options, self.source, self.type)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._canonical_form is not None or self.logical_types is not None or self.named_types is not None or self.options is not None or self.source is not None or self.type is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._canonical_form, self.logical_types, self.named_types, self.options, self.source, self.type)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._canonical_form is not None or self.logical_types is not None or self.named_types is not None or self.options is not None or self.source is not None or self.type is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ResolvedSchema, (type(self), 0xc2592e1, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->__pyx_base._canonical_form != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.logical_types != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.named_types != ((PyObject*)Py_None));
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.options) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->__pyx_base.source != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base.type) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._canonical_form is not None or self.logical_types is not None or self.named_types is not None or self.options is not None or self.source is not None or self.type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ResolvedSchema, (type(self), 0xc2592e1, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._canonical_form is not None or self.logical_types is not None or self.named_types is not None or self.options is not None or self.source is not None or self.type is not None
 *     if use_setstate:
 *         return __pyx_unpickle_ResolvedSchema, (type(self), 0xc2592e1, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ResolvedSchema, (type(self), 0xc2592e1, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_ResolvedSchema); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_203789025);
    __Pyx_GIVEREF(__pyx_int_203789025);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_203789025)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._canonical_form is not None or self.logical_types is not None or self.named_types is not None or self.options is not None or self.source is not None or self.type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ResolvedSchema, (type(self), 0xc2592e1, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ResolvedSchema, (type(self), 0xc2592e1, None), state
 *     else:
 *         return __pyx_unpickle_ResolvedSchema, (type(self), 0xc2592e1, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ResolvedSchema__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_ResolvedSchema); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_203789025);
    __Pyx_GIVEREF(__pyx_int_203789025);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_203789025)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.ResolvedSchema.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ResolvedSchema, (type(self), 0xc2592e1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ResolvedSchema__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14ResolvedSchema_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_14ResolvedSchema_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14ResolvedSchema_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_14ResolvedSchema_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ResolvedSchema.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_14ResolvedSchema_4__setstate_cython__(((struct __pyx_obj_5cavro_ResolvedSchema *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14ResolvedSchema_4__setstate_cython__(struct __pyx_obj_5cavro_ResolvedSchema *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ResolvedSchema, (type(self), 0xc2592e1, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ResolvedSchema__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_ResolvedSchema__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ResolvedSchema, (type(self), 0xc2592e1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ResolvedSchema__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ResolvedSchema.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":8
 * 
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement encode_value")
 */

static PyObject *__pyx_f_5cavro_12ValueAdapter_encode_value(struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 1);

  /* "src/logical.pxi":10
 *     cdef encode_value(self, value):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement encode_value")             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement_encode_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/logical.pxi":9
 * 
 *     cdef encode_value(self, value):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement encode_value")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(24, 9, __pyx_L1_error)

  /* "src/logical.pxi":8
 * 
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement encode_value")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.ValueAdapter.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":12
 *             f"{type(self).__name__} does not implement encode_value")
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement decode_value")
 */

static PyObject *__pyx_f_5cavro_12ValueAdapter_decode_value(struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/logical.pxi":14
 *     cdef decode_value(self, value):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement decode_value")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement_decode_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/logical.pxi":13
 * 
 *     cdef decode_value(self, value):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement decode_value")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(24, 13, __pyx_L1_error)

  /* "src/logical.pxi":12
 *             f"{type(self).__name__} does not implement encode_value")
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement decode_value")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.ValueAdapter.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12ValueAdapter_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_12ValueAdapter_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12ValueAdapter_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_12ValueAdapter_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_12ValueAdapter___reduce_cython__(((struct __pyx_obj_5cavro_ValueAdapter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12ValueAdapter___reduce_cython__(struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ValueAdapter, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ValueAdapter, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_ValueAdapter, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ValueAdapter, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_ValueAdapter); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ValueAdapter, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ValueAdapter, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_ValueAdapter, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ValueAdapter__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_ValueAdapter); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.ValueAdapter.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ValueAdapter, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ValueAdapter__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12ValueAdapter_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_12ValueAdapter_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12ValueAdapter_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_12ValueAdapter_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ValueAdapter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_12ValueAdapter_2__setstate_cython__(((struct __pyx_obj_5cavro_ValueAdapter *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12ValueAdapter_2__setstate_cython__(struct __pyx_obj_5cavro_ValueAdapter *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ValueAdapter, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ValueAdapter__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_ValueAdapter__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ValueAdapter, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ValueAdapter__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ValueAdapter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":17
 * 
 * 
 * cdef int twos_complement(int value, int bits):             # <<<<<<<<<<<<<<
 *     cdef int mask = (1 << bits) - 1
 *     value = (value ^ mask) + 1
 */

static int __pyx_f_5cavro_twos_complement(int __pyx_v_value, int __pyx_v_bits) {
  int __pyx_v_mask;
  int __pyx_r;

  /* "src/logical.pxi":18
 * 
 * cdef int twos_complement(int value, int bits):
 *     cdef int mask = (1 << bits) - 1             # <<<<<<<<<<<<<<
 *     value = (value ^ mask) + 1
 *     return value & mask
 */
  __pyx_v_mask = ((1 << __pyx_v_bits) - 1);

  /* "src/logical.pxi":19
 * cdef int twos_complement(int value, int bits):
 *     cdef int mask = (1 << bits) - 1
 *     value = (value ^ mask) + 1             # <<<<<<<<<<<<<<
 *     return value & mask
 * 
 */
  __pyx_v_value = ((__pyx_v_value ^ __pyx_v_mask) + 1);

  /* "src/logical.pxi":20
 *     cdef int mask = (1 << bits) - 1
 *     value = (value ^ mask) + 1
 *     return value & mask             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_value & __pyx_v_mask);
  goto __pyx_L0;

  /* "src/logical.pxi":17
 * 
 * 
 * cdef int twos_complement(int value, int bits):             # <<<<<<<<<<<<<<
 *     cdef int mask = (1 << bits) - 1
 *     value = (value ^ mask) + 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "src/logical.pxi":34
 *     underlying_types = NotImplemented
 * 
 *     @_class_inst_method             # <<<<<<<<<<<<<<
 *     def for_type(inst, cls, underlying: AvroType):
 *         cdef object ob = inst
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11LogicalType_1for_type(PyObject *__pyx_v_inst, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_11LogicalType_1for_type = {"for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11LogicalType_1for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_11LogicalType_1for_type(PyObject *__pyx_v_inst, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_cls = 0;
  struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("for_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_underlying,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_cls)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 34, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_underlying)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 34, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("for_type", 1, 2, 2, 1); __PYX_ERR(24, 34, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "for_type") < 0)) __PYX_ERR(24, 34, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_cls = values[0];
    __pyx_v_underlying = ((struct __pyx_obj_5cavro_AvroType *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("for_type", 1, 2, 2, __pyx_nargs); __PYX_ERR(24, 34, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.LogicalType.for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_underlying), __pyx_ptype_5cavro_AvroType, 0, "underlying", 0))) __PYX_ERR(24, 35, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_11LogicalType_for_type(((struct __pyx_obj_5cavro_LogicalType *)__pyx_v_inst), __pyx_v_cls, __pyx_v_underlying);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11LogicalType_for_type(struct __pyx_obj_5cavro_LogicalType *__pyx_v_inst, PyObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying) {
  PyObject *__pyx_v_ob = 0;
  PyObject *__pyx_v_underlying_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("for_type", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_inst);

  /* "src/logical.pxi":36
 *     @_class_inst_method
 *     def for_type(inst, cls, underlying: AvroType):
 *         cdef object ob = inst             # <<<<<<<<<<<<<<
 *         if inst is None:
 *             ob = cls
 */
  __Pyx_INCREF((PyObject *)__pyx_v_inst);
  __pyx_v_ob = ((PyObject *)__pyx_v_inst);

  /* "src/logical.pxi":37
 *     def for_type(inst, cls, underlying: AvroType):
 *         cdef object ob = inst
 *         if inst is None:             # <<<<<<<<<<<<<<
 *             ob = cls
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_inst) == Py_None);
  if (__pyx_t_1) {

    /* "src/logical.pxi":38
 *         cdef object ob = inst
 *         if inst is None:
 *             ob = cls             # <<<<<<<<<<<<<<
 * 
 *         for underlying_type in ob.underlying_types:
 */
    __Pyx_INCREF(__pyx_v_cls);
    __Pyx_DECREF_SET(__pyx_v_ob, __pyx_v_cls);

    /* "src/logical.pxi":37
 *     def for_type(inst, cls, underlying: AvroType):
 *         cdef object ob = inst
 *         if inst is None:             # <<<<<<<<<<<<<<
 *             ob = cls
 * 
 */
  }

  /* "src/logical.pxi":40
 *             ob = cls
 * 
 *         for underlying_type in ob.underlying_types:             # <<<<<<<<<<<<<<
 *             if isinstance(underlying, underlying_type):
 *                 inst = ob._for_type(underlying)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ob, __pyx_n_s_underlying_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 40, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(24, 40, __pyx_L1_error)
          #endif
          if (__pyx_t_4 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(24, 40, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 40, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(24, 40, __pyx_L1_error)
          #endif
          if (__pyx_t_4 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(24, 40, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 40, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(24, 40, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_underlying_type, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/logical.pxi":41
 * 
 *         for underlying_type in ob.underlying_types:
 *             if isinstance(underlying, underlying_type):             # <<<<<<<<<<<<<<
 *                 inst = ob._for_type(underlying)
 *                 if inst is not None:
 */
    __pyx_t_1 = PyObject_IsInstance(((PyObject *)__pyx_v_underlying), __pyx_v_underlying_type); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(24, 41, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "src/logical.pxi":42
 *         for underlying_type in ob.underlying_types:
 *             if isinstance(underlying, underlying_type):
 *                 inst = ob._for_type(underlying)             # <<<<<<<<<<<<<<
 *                 if inst is not None:
 *                     return inst
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_ob, __pyx_n_s_for_type_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(24, 42, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, ((PyObject *)__pyx_v_underlying)};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 42, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_LogicalType))))) __PYX_ERR(24, 42, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_inst, ((struct __pyx_obj_5cavro_LogicalType *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "src/logical.pxi":43
 *             if isinstance(underlying, underlying_type):
 *                 inst = ob._for_type(underlying)
 *                 if inst is not None:             # <<<<<<<<<<<<<<
 *                     return inst
 * 
 */
      __pyx_t_1 = (((PyObject *)__pyx_v_inst) != Py_None);
      if (__pyx_t_1) {

        /* "src/logical.pxi":44
 *                 inst = ob._for_type(underlying)
 *                 if inst is not None:
 *                     return inst             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF((PyObject *)__pyx_v_inst);
        __pyx_r = ((PyObject *)__pyx_v_inst);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L0;

        /* "src/logical.pxi":43
 *             if isinstance(underlying, underlying_type):
 *                 inst = ob._for_type(underlying)
 *                 if inst is not None:             # <<<<<<<<<<<<<<
 *                     return inst
 * 
 */
      }

      /* "src/logical.pxi":41
 * 
 *         for underlying_type in ob.underlying_types:
 *             if isinstance(underlying, underlying_type):             # <<<<<<<<<<<<<<
 *                 inst = ob._for_type(underlying)
 *                 if inst is not None:
 */
    }

    /* "src/logical.pxi":40
 *             ob = cls
 * 
 *         for underlying_type in ob.underlying_types:             # <<<<<<<<<<<<<<
 *             if isinstance(underlying, underlying_type):
 *                 inst = ob._for_type(underlying)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "src/logical.pxi":34
 *     underlying_types = NotImplemented
 * 
 *     @_class_inst_method             # <<<<<<<<<<<<<<
 *     def for_type(inst, cls, underlying: AvroType):
 *         cdef object ob = inst
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.LogicalType.for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ob);
  __Pyx_XDECREF(__pyx_v_underlying_type);
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11LogicalType_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_11LogicalType_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11LogicalType_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_11LogicalType_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_11LogicalType_2__reduce_cython__(((struct __pyx_obj_5cavro_LogicalType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11LogicalType_2__reduce_cython__(struct __pyx_obj_5cavro_LogicalType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_LogicalType, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LogicalType, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_LogicalType, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_LogicalType, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_LogicalType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LogicalType, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_LogicalType, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_LogicalType, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LogicalType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_LogicalType); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.LogicalType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LogicalType, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LogicalType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11LogicalType_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_11LogicalType_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11LogicalType_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_11LogicalType_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.LogicalType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_11LogicalType_4__setstate_cython__(((struct __pyx_obj_5cavro_LogicalType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11LogicalType_4__setstate_cython__(struct __pyx_obj_5cavro_LogicalType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_LogicalType, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LogicalType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_LogicalType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LogicalType, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LogicalType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.LogicalType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":65
 *     """
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         return self.custom_encode_value(value)
 * 
 */

static PyObject *__pyx_f_5cavro_17CustomLogicalType_encode_value(struct __pyx_obj_5cavro_CustomLogicalType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 1);

  /* "src/logical.pxi":66
 * 
 *     cdef encode_value(self, value):
 *         return self.custom_encode_value(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_custom_encode_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":65
 *     """
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         return self.custom_encode_value(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.CustomLogicalType.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":68
 *         return self.custom_encode_value(value)
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return self.custom_decode_value(value)
 * 
 */

static PyObject *__pyx_f_5cavro_17CustomLogicalType_decode_value(struct __pyx_obj_5cavro_CustomLogicalType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/logical.pxi":69
 * 
 *     cdef decode_value(self, value):
 *         return self.custom_decode_value(value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_custom_decode_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":68
 *         return self.custom_encode_value(value)
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return self.custom_decode_value(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.CustomLogicalType.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_17CustomLogicalType_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_17CustomLogicalType_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_17CustomLogicalType_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_17CustomLogicalType_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_17CustomLogicalType___reduce_cython__(((struct __pyx_obj_5cavro_CustomLogicalType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_17CustomLogicalType___reduce_cython__(struct __pyx_obj_5cavro_CustomLogicalType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CustomLogicalType, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CustomLogicalType, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_CustomLogicalType, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CustomLogicalType, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_CustomLogicalType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CustomLogicalType, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CustomLogicalType, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_CustomLogicalType, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CustomLogicalType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_CustomLogicalType); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.CustomLogicalType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CustomLogicalType, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CustomLogicalType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_17CustomLogicalType_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_17CustomLogicalType_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_17CustomLogicalType_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_17CustomLogicalType_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.CustomLogicalType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_17CustomLogicalType_2__setstate_cython__(((struct __pyx_obj_5cavro_CustomLogicalType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_17CustomLogicalType_2__setstate_cython__(struct __pyx_obj_5cavro_CustomLogicalType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CustomLogicalType, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CustomLogicalType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_CustomLogicalType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CustomLogicalType, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CustomLogicalType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.CustomLogicalType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":87
 *     cdef readonly bint check_exp_overflow
 * 
 *     def __init__(self, precision, scale, size, check_exp_overflow, underlying_name):             # <<<<<<<<<<<<<<
 *         self.precision = precision
 *         self.scale = scale
 */

/* Python wrapper */
static int __pyx_pw_5cavro_11DecimalType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_11DecimalType_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_precision = 0;
  PyObject *__pyx_v_scale = 0;
  PyObject *__pyx_v_size = 0;
  PyObject *__pyx_v_check_exp_overflow = 0;
  PyObject *__pyx_v_underlying_name = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_precision,&__pyx_n_s_scale,&__pyx_n_s_size,&__pyx_n_s_check_exp_overflow,&__pyx_n_s_underlying_name,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_precision)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 87, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_scale)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 87, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 1); __PYX_ERR(24, 87, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_size)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 87, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 2); __PYX_ERR(24, 87, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_check_exp_overflow)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 87, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 3); __PYX_ERR(24, 87, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_underlying_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[4]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 87, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 4); __PYX_ERR(24, 87, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(24, 87, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
      values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
    }
    __pyx_v_precision = values[0];
    __pyx_v_scale = values[1];
    __pyx_v_size = values[2];
    __pyx_v_check_exp_overflow = values[3];
    __pyx_v_underlying_name = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, __pyx_nargs); __PYX_ERR(24, 87, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.DecimalType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_11DecimalType___init__(((struct __pyx_obj_5cavro_DecimalType *)__pyx_v_self), __pyx_v_precision, __pyx_v_scale, __pyx_v_size, __pyx_v_check_exp_overflow, __pyx_v_underlying_name);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_11DecimalType___init__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self, PyObject *__pyx_v_precision, PyObject *__pyx_v_scale, PyObject *__pyx_v_size, PyObject *__pyx_v_check_exp_overflow, PyObject *__pyx_v_underlying_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/logical.pxi":88
 * 
 *     def __init__(self, precision, scale, size, check_exp_overflow, underlying_name):
 *         self.precision = precision             # <<<<<<<<<<<<<<
 *         self.scale = scale
 *         self.scale_val = decimal.Decimal('1').scaleb(-scale)
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_precision); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(24, 88, __pyx_L1_error)
  __pyx_v_self->precision = __pyx_t_1;

  /* "src/logical.pxi":89
 *     def __init__(self, precision, scale, size, check_exp_overflow, underlying_name):
 *         self.precision = precision
 *         self.scale = scale             # <<<<<<<<<<<<<<
 *         self.scale_val = decimal.Decimal('1').scaleb(-scale)
 *         self.context = decimal.Context(prec=self.precision, clamp=1)
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_scale); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(24, 89, __pyx_L1_error)
  __pyx_v_self->scale = __pyx_t_1;

  /* "src/logical.pxi":90
 *         self.precision = precision
 *         self.scale = scale
 *         self.scale_val = decimal.Decimal('1').scaleb(-scale)             # <<<<<<<<<<<<<<
 *         self.context = decimal.Context(prec=self.precision, clamp=1)
 *         self.size = size
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_decimal); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_1 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_1 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_1};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_1, 1+__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_scaleb); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Negative(__pyx_v_scale); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_1 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_1 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_1, 1+__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->scale_val);
  __Pyx_DECREF(__pyx_v_self->scale_val);
  __pyx_v_self->scale_val = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "src/logical.pxi":91
 *         self.scale = scale
 *         self.scale_val = decimal.Decimal('1').scaleb(-scale)
 *         self.context = decimal.Context(prec=self.precision, clamp=1)             # <<<<<<<<<<<<<<
 *         self.size = size
 *         self.check_exp_overflow = check_exp_overflow
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Context); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->precision); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_prec, __pyx_t_3) < 0) __PYX_ERR(24, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_clamp, __pyx_int_1) < 0) __PYX_ERR(24, 91, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->context);
  __Pyx_DECREF(__pyx_v_self->context);
  __pyx_v_self->context = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "src/logical.pxi":92
 *         self.scale_val = decimal.Decimal('1').scaleb(-scale)
 *         self.context = decimal.Context(prec=self.precision, clamp=1)
 *         self.size = size             # <<<<<<<<<<<<<<
 *         self.check_exp_overflow = check_exp_overflow
 *         self.underlying_name = underlying_name
 */
  __Pyx_INCREF(__pyx_v_size);
  __Pyx_GIVEREF(__pyx_v_size);
  __Pyx_GOTREF(__pyx_v_self->size);
  __Pyx_DECREF(__pyx_v_self->size);
  __pyx_v_self->size = __pyx_v_size;

  /* "src/logical.pxi":93
 *         self.context = decimal.Context(prec=self.precision, clamp=1)
 *         self.size = size
 *         self.check_exp_overflow = check_exp_overflow             # <<<<<<<<<<<<<<
 *         self.underlying_name = underlying_name
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_check_exp_overflow); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(24, 93, __pyx_L1_error)
  __pyx_v_self->check_exp_overflow = __pyx_t_6;

  /* "src/logical.pxi":94
 *         self.size = size
 *         self.check_exp_overflow = check_exp_overflow
 *         self.underlying_name = underlying_name             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_underlying_name))||((__pyx_v_underlying_name) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_underlying_name))) __PYX_ERR(24, 94, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_underlying_name;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->underlying_name);
  __Pyx_DECREF(__pyx_v_self->underlying_name);
  __pyx_v_self->underlying_name = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/logical.pxi":87
 *     cdef readonly bint check_exp_overflow
 * 
 *     def __init__(self, precision, scale, size, check_exp_overflow, underlying_name):             # <<<<<<<<<<<<<<
 *         self.precision = precision
 *         self.scale = scale
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.DecimalType.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":96
 *         self.underlying_name = underlying_name
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return f'{self.underlying_name}(decimal)'
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11DecimalType_9type_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11DecimalType_9type_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11DecimalType_9type_name___get__(((struct __pyx_obj_5cavro_DecimalType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11DecimalType_9type_name___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/logical.pxi":98
 *     @property
 *     def type_name(self):
 *         return f'{self.underlying_name}(decimal)'             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_v_self->underlying_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_1, __pyx_kp_u_decimal_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":96
 *         self.underlying_name = underlying_name
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return f'{self.underlying_name}(decimal)'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.DecimalType.type_name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":100
 *         return f'{self.underlying_name}(decimal)'
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         meta = underlying.metadata
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11DecimalType_3_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_11DecimalType_3_for_type = {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11DecimalType_3_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_11DecimalType_3_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_for_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_underlying,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_underlying)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 100, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_for_type") < 0)) __PYX_ERR(24, 100, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_underlying = ((struct __pyx_obj_5cavro_AvroType *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_for_type", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 100, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.DecimalType._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_underlying), __pyx_ptype_5cavro_AvroType, 0, "underlying", 0))) __PYX_ERR(24, 101, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_11DecimalType_2_for_type(((PyTypeObject*)__pyx_v_cls), __pyx_v_underlying);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11DecimalType_2_for_type(PyTypeObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying) {
  PyObject *__pyx_v_meta = NULL;
  PyObject *__pyx_v_precision = NULL;
  PyObject *__pyx_v_scale = NULL;
  PyObject *__pyx_v_size = NULL;
  PyObject *__pyx_v_max_precision = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_type", 1);

  /* "src/logical.pxi":102
 *     @classmethod
 *     def _for_type(cls, underlying: AvroType):
 *         meta = underlying.metadata             # <<<<<<<<<<<<<<
 *         precision = meta.get('precision')
 *         if precision is None:
 */
  __pyx_t_1 = __pyx_v_underlying->metadata;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_meta = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/logical.pxi":103
 *     def _for_type(cls, underlying: AvroType):
 *         meta = underlying.metadata
 *         precision = meta.get('precision')             # <<<<<<<<<<<<<<
 *         if precision is None:
 *             return
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_meta, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_u_precision};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_precision = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/logical.pxi":104
 *         meta = underlying.metadata
 *         precision = meta.get('precision')
 *         if precision is None:             # <<<<<<<<<<<<<<
 *             return
 *         if not isinstance(precision, int) or precision < 1 or precision > decimal.MAX_PREC:
 */
  __pyx_t_5 = (__pyx_v_precision == Py_None);
  if (__pyx_t_5) {

    /* "src/logical.pxi":105
 *         precision = meta.get('precision')
 *         if precision is None:
 *             return             # <<<<<<<<<<<<<<
 *         if not isinstance(precision, int) or precision < 1 or precision > decimal.MAX_PREC:
 *             if underlying.options.raise_on_invalid_logical:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/logical.pxi":104
 *         meta = underlying.metadata
 *         precision = meta.get('precision')
 *         if precision is None:             # <<<<<<<<<<<<<<
 *             return
 *         if not isinstance(precision, int) or precision < 1 or precision > decimal.MAX_PREC:
 */
  }

  /* "src/logical.pxi":106
 *         if precision is None:
 *             return
 *         if not isinstance(precision, int) or precision < 1 or precision > decimal.MAX_PREC:             # <<<<<<<<<<<<<<
 *             if underlying.options.raise_on_invalid_logical:
 *                 raise ValueError(f'Precision must be an integer between 1 and {decimal.MAX_PREC}, got: {repr(precision)}')
 */
  __pyx_t_6 = PyInt_Check(__pyx_v_precision); 
  __pyx_t_7 = (!__pyx_t_6);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_5 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_precision, __pyx_int_1, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 106, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(24, 106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_7) {
  } else {
    __pyx_t_5 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_decimal); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_MAX_PREC); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_precision, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(24, 106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_5) {

    /* "src/logical.pxi":107
 *             return
 *         if not isinstance(precision, int) or precision < 1 or precision > decimal.MAX_PREC:
 *             if underlying.options.raise_on_invalid_logical:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f'Precision must be an integer between 1 and {decimal.MAX_PREC}, got: {repr(precision)}')
 *             return
 */
    if (unlikely(__pyx_v_underlying->options->raise_on_invalid_logical)) {

      /* "src/logical.pxi":108
 *         if not isinstance(precision, int) or precision < 1 or precision > decimal.MAX_PREC:
 *             if underlying.options.raise_on_invalid_logical:
 *                 raise ValueError(f'Precision must be an integer between 1 and {decimal.MAX_PREC}, got: {repr(precision)}')             # <<<<<<<<<<<<<<
 *             return
 *         scale = meta.get('scale', 0)
 */
      __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = 0;
      __pyx_t_9 = 127;
      __Pyx_INCREF(__pyx_kp_u_Precision_must_be_an_integer_bet);
      __pyx_t_8 += 43;
      __Pyx_GIVEREF(__pyx_kp_u_Precision_must_be_an_integer_bet);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Precision_must_be_an_integer_bet);
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MAX_PREC); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_9;
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_INCREF(__pyx_kp_u_got_2);
      __pyx_t_8 += 7;
      __Pyx_GIVEREF(__pyx_kp_u_got_2);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_got_2);
      __pyx_t_2 = PyObject_Repr(__pyx_v_precision); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_9;
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(24, 108, __pyx_L1_error)

      /* "src/logical.pxi":107
 *             return
 *         if not isinstance(precision, int) or precision < 1 or precision > decimal.MAX_PREC:
 *             if underlying.options.raise_on_invalid_logical:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f'Precision must be an integer between 1 and {decimal.MAX_PREC}, got: {repr(precision)}')
 *             return
 */
    }

    /* "src/logical.pxi":109
 *             if underlying.options.raise_on_invalid_logical:
 *                 raise ValueError(f'Precision must be an integer between 1 and {decimal.MAX_PREC}, got: {repr(precision)}')
 *             return             # <<<<<<<<<<<<<<
 *         scale = meta.get('scale', 0)
 *         if not isinstance(scale, int) or scale < 0:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/logical.pxi":106
 *         if precision is None:
 *             return
 *         if not isinstance(precision, int) or precision < 1 or precision > decimal.MAX_PREC:             # <<<<<<<<<<<<<<
 *             if underlying.options.raise_on_invalid_logical:
 *                 raise ValueError(f'Precision must be an integer between 1 and {decimal.MAX_PREC}, got: {repr(precision)}')
 */
  }

  /* "src/logical.pxi":110
 *                 raise ValueError(f'Precision must be an integer between 1 and {decimal.MAX_PREC}, got: {repr(precision)}')
 *             return
 *         scale = meta.get('scale', 0)             # <<<<<<<<<<<<<<
 *         if not isinstance(scale, int) or scale < 0:
 *             if underlying.options.raise_on_invalid_logical:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_meta, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__85, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_scale = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "src/logical.pxi":111
 *             return
 *         scale = meta.get('scale', 0)
 *         if not isinstance(scale, int) or scale < 0:             # <<<<<<<<<<<<<<
 *             if underlying.options.raise_on_invalid_logical:
 *                 raise ValueError(f'Scale must be a positive integer, got: {repr(scale)}')
 */
  __pyx_t_7 = PyInt_Check(__pyx_v_scale); 
  __pyx_t_6 = (!__pyx_t_7);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_scale, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 111, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(24, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __pyx_t_6;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_5) {

    /* "src/logical.pxi":112
 *         scale = meta.get('scale', 0)
 *         if not isinstance(scale, int) or scale < 0:
 *             if underlying.options.raise_on_invalid_logical:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f'Scale must be a positive integer, got: {repr(scale)}')
 *             return
 */
    if (unlikely(__pyx_v_underlying->options->raise_on_invalid_logical)) {

      /* "src/logical.pxi":113
 *         if not isinstance(scale, int) or scale < 0:
 *             if underlying.options.raise_on_invalid_logical:
 *                 raise ValueError(f'Scale must be a positive integer, got: {repr(scale)}')             # <<<<<<<<<<<<<<
 *             return
 * 
 */
      __pyx_t_3 = PyObject_Repr(__pyx_v_scale); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Scale_must_be_a_positive_integer, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(24, 113, __pyx_L1_error)

      /* "src/logical.pxi":112
 *         scale = meta.get('scale', 0)
 *         if not isinstance(scale, int) or scale < 0:
 *             if underlying.options.raise_on_invalid_logical:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f'Scale must be a positive integer, got: {repr(scale)}')
 *             return
 */
    }

    /* "src/logical.pxi":114
 *             if underlying.options.raise_on_invalid_logical:
 *                 raise ValueError(f'Scale must be a positive integer, got: {repr(scale)}')
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if scale > precision:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/logical.pxi":111
 *             return
 *         scale = meta.get('scale', 0)
 *         if not isinstance(scale, int) or scale < 0:             # <<<<<<<<<<<<<<
 *             if underlying.options.raise_on_invalid_logical:
 *                 raise ValueError(f'Scale must be a positive integer, got: {repr(scale)}')
 */
  }

  /* "src/logical.pxi":116
 *             return
 * 
 *         if scale > precision:             # <<<<<<<<<<<<<<
 *             if underlying.options.raise_on_invalid_logical:
 *                 raise ValueError(f'Precision must be greater than scale. scale: {scale}, precision: {precision}')
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_scale, __pyx_v_precision, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 116, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(24, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "src/logical.pxi":117
 * 
 *         if scale > precision:
 *             if underlying.options.raise_on_invalid_logical:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f'Precision must be greater than scale. scale: {scale}, precision: {precision}')
 *             return
 */
    if (unlikely(__pyx_v_underlying->options->raise_on_invalid_logical)) {

      /* "src/logical.pxi":118
 *         if scale > precision:
 *             if underlying.options.raise_on_invalid_logical:
 *                 raise ValueError(f'Precision must be greater than scale. scale: {scale}, precision: {precision}')             # <<<<<<<<<<<<<<
 *             return
 * 
 */
      __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = 0;
      __pyx_t_9 = 127;
      __Pyx_INCREF(__pyx_kp_u_Precision_must_be_greater_than_s);
      __pyx_t_8 += 45;
      __Pyx_GIVEREF(__pyx_kp_u_Precision_must_be_greater_than_s);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Precision_must_be_greater_than_s);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_scale, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_9;
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_INCREF(__pyx_kp_u_precision_2);
      __pyx_t_8 += 13;
      __Pyx_GIVEREF(__pyx_kp_u_precision_2);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_precision_2);
      __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_precision, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_9;
      __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(24, 118, __pyx_L1_error)

      /* "src/logical.pxi":117
 * 
 *         if scale > precision:
 *             if underlying.options.raise_on_invalid_logical:             # <<<<<<<<<<<<<<
 *                 raise ValueError(f'Precision must be greater than scale. scale: {scale}, precision: {precision}')
 *             return
 */
    }

    /* "src/logical.pxi":119
 *             if underlying.options.raise_on_invalid_logical:
 *                 raise ValueError(f'Precision must be greater than scale. scale: {scale}, precision: {precision}')
 *             return             # <<<<<<<<<<<<<<
 * 
 *         size = None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/logical.pxi":116
 *             return
 * 
 *         if scale > precision:             # <<<<<<<<<<<<<<
 *             if underlying.options.raise_on_invalid_logical:
 *                 raise ValueError(f'Precision must be greater than scale. scale: {scale}, precision: {precision}')
 */
  }

  /* "src/logical.pxi":121
 *             return
 * 
 *         size = None             # <<<<<<<<<<<<<<
 *         if isinstance(underlying, FixedType):
 *             size = underlying.size
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_size = Py_None;

  /* "src/logical.pxi":122
 * 
 *         size = None
 *         if isinstance(underlying, FixedType):             # <<<<<<<<<<<<<<
 *             size = underlying.size
 *             max_precision = math.floor(math.log10(2) * (8 * size - 1))
 */
  __pyx_t_5 = __Pyx_TypeCheck(((PyObject *)__pyx_v_underlying), __pyx_ptype_5cavro_FixedType); 
  if (__pyx_t_5) {

    /* "src/logical.pxi":123
 *         size = None
 *         if isinstance(underlying, FixedType):
 *             size = underlying.size             # <<<<<<<<<<<<<<
 *             max_precision = math.floor(math.log10(2) * (8 * size - 1))
 *             if precision > max_precision: # Numbers may not fit into the fixed
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_underlying), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/logical.pxi":124
 *         if isinstance(underlying, FixedType):
 *             size = underlying.size
 *             max_precision = math.floor(math.log10(2) * (8 * size - 1))             # <<<<<<<<<<<<<<
 *             if precision > max_precision: # Numbers may not fit into the fixed
 *                 if underlying.options.raise_on_invalid_logical:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_math); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_floor); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_math); if (unlikely(!__pyx_t_10)) __PYX_ERR(24, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_log10); if (unlikely(!__pyx_t_11)) __PYX_ERR(24, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_int_2};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __pyx_t_11 = __Pyx_PyInt_MultiplyCObj(__pyx_int_8, __pyx_v_size, 8, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(24, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_10 = __Pyx_PyInt_SubtractObjC(__pyx_t_11, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(24, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = PyNumber_Multiply(__pyx_t_3, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(24, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_11};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_v_max_precision = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "src/logical.pxi":125
 *             size = underlying.size
 *             max_precision = math.floor(math.log10(2) * (8 * size - 1))
 *             if precision > max_precision: # Numbers may not fit into the fixed             # <<<<<<<<<<<<<<
 *                 if underlying.options.raise_on_invalid_logical:
 *                     raise ValueError(f'Precision is too large for fixed size {size}. Precision: {precision}, Max precision is {max_precision}')
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_precision, __pyx_v_max_precision, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 125, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(24, 125, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "src/logical.pxi":126
 *             max_precision = math.floor(math.log10(2) * (8 * size - 1))
 *             if precision > max_precision: # Numbers may not fit into the fixed
 *                 if underlying.options.raise_on_invalid_logical:             # <<<<<<<<<<<<<<
 *                     raise ValueError(f'Precision is too large for fixed size {size}. Precision: {precision}, Max precision is {max_precision}')
 *                 return None
 */
      if (unlikely(__pyx_v_underlying->options->raise_on_invalid_logical)) {

        /* "src/logical.pxi":127
 *             if precision > max_precision: # Numbers may not fit into the fixed
 *                 if underlying.options.raise_on_invalid_logical:
 *                     raise ValueError(f'Precision is too large for fixed size {size}. Precision: {precision}, Max precision is {max_precision}')             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
        __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = 0;
        __pyx_t_9 = 127;
        __Pyx_INCREF(__pyx_kp_u_Precision_is_too_large_for_fixed);
        __pyx_t_8 += 38;
        __Pyx_GIVEREF(__pyx_kp_u_Precision_is_too_large_for_fixed);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Precision_is_too_large_for_fixed);
        __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_size, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_9;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_INCREF(__pyx_kp_u_Precision);
        __pyx_t_8 += 13;
        __Pyx_GIVEREF(__pyx_kp_u_Precision);
        PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_Precision);
        __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_precision, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_9;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_INCREF(__pyx_kp_u_Max_precision_is);
        __pyx_t_8 += 19;
        __Pyx_GIVEREF(__pyx_kp_u_Max_precision_is);
        PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_Max_precision_is);
        __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_max_precision, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_9;
        __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 6, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(24, 127, __pyx_L1_error)

        /* "src/logical.pxi":126
 *             max_precision = math.floor(math.log10(2) * (8 * size - 1))
 *             if precision > max_precision: # Numbers may not fit into the fixed
 *                 if underlying.options.raise_on_invalid_logical:             # <<<<<<<<<<<<<<
 *                     raise ValueError(f'Precision is too large for fixed size {size}. Precision: {precision}, Max precision is {max_precision}')
 *                 return None
 */
      }

      /* "src/logical.pxi":128
 *                 if underlying.options.raise_on_invalid_logical:
 *                     raise ValueError(f'Precision is too large for fixed size {size}. Precision: {precision}, Max precision is {max_precision}')
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return cls(precision, scale, size, underlying.options.decimal_check_exp_overflow, underlying.type_name)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "src/logical.pxi":125
 *             size = underlying.size
 *             max_precision = math.floor(math.log10(2) * (8 * size - 1))
 *             if precision > max_precision: # Numbers may not fit into the fixed             # <<<<<<<<<<<<<<
 *                 if underlying.options.raise_on_invalid_logical:
 *                     raise ValueError(f'Precision is too large for fixed size {size}. Precision: {precision}, Max precision is {max_precision}')
 */
    }

    /* "src/logical.pxi":122
 * 
 *         size = None
 *         if isinstance(underlying, FixedType):             # <<<<<<<<<<<<<<
 *             size = underlying.size
 *             max_precision = math.floor(math.log10(2) * (8 * size - 1))
 */
  }

  /* "src/logical.pxi":130
 *                 return None
 * 
 *         return cls(precision, scale, size, underlying.options.decimal_check_exp_overflow, underlying.type_name)             # <<<<<<<<<<<<<<
 * 
 *     cdef encode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_underlying->options->decimal_check_exp_overflow); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_underlying), __pyx_n_s_type_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_New(5); if (unlikely(!__pyx_t_11)) __PYX_ERR(24, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_v_precision);
  __Pyx_GIVEREF(__pyx_v_precision);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_precision)) __PYX_ERR(24, 130, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_scale);
  __Pyx_GIVEREF(__pyx_v_scale);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_scale)) __PYX_ERR(24, 130, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_size);
  __Pyx_GIVEREF(__pyx_v_size);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_v_size)) __PYX_ERR(24, 130, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 3, __pyx_t_1)) __PYX_ERR(24, 130, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 4, __pyx_t_2)) __PYX_ERR(24, 130, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_v_cls), __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":100
 *         return f'{self.underlying_name}(decimal)'
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         meta = underlying.metadata
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("cavro.DecimalType._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_meta);
  __Pyx_XDECREF(__pyx_v_precision);
  __Pyx_XDECREF(__pyx_v_scale);
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XDECREF(__pyx_v_max_precision);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":132
 *         return cls(precision, scale, size, underlying.options.decimal_check_exp_overflow, underlying.type_name)
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, bytes):
 *             return value
 */

static PyObject *__pyx_f_5cavro_11DecimalType_encode_value(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_orig_exp = NULL;
  PyObject *__pyx_v_quant = NULL;
  PyObject *__pyx_v_sign = NULL;
  PyObject *__pyx_v_digits = NULL;
  PyObject *__pyx_v_exp = NULL;
  PyObject *__pyx_v_coeff = NULL;
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_v_n_bytes = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  Py_ssize_t __pyx_t_10;
  Py_UCS4 __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/logical.pxi":133
 * 
 *     cdef encode_value(self, value):
 *         if isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *             return value
 *         if not isinstance(value, decimal.Decimal):
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_value); 
  if (__pyx_t_1) {

    /* "src/logical.pxi":134
 *     cdef encode_value(self, value):
 *         if isinstance(value, bytes):
 *             return value             # <<<<<<<<<<<<<<
 *         if not isinstance(value, decimal.Decimal):
 *             value = decimal.Decimal(value)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/logical.pxi":133
 * 
 *     cdef encode_value(self, value):
 *         if isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *             return value
 *         if not isinstance(value, decimal.Decimal):
 */
  }

  /* "src/logical.pxi":135
 *         if isinstance(value, bytes):
 *             return value
 *         if not isinstance(value, decimal.Decimal):             # <<<<<<<<<<<<<<
 *             value = decimal.Decimal(value)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(24, 135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (!__pyx_t_1);
  if (__pyx_t_4) {

    /* "src/logical.pxi":136
 *             return value
 *         if not isinstance(value, decimal.Decimal):
 *             value = decimal.Decimal(value)             # <<<<<<<<<<<<<<
 * 
 *         if self.check_exp_overflow:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_value};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 136, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/logical.pxi":135
 *         if isinstance(value, bytes):
 *             return value
 *         if not isinstance(value, decimal.Decimal):             # <<<<<<<<<<<<<<
 *             value = decimal.Decimal(value)
 * 
 */
  }

  /* "src/logical.pxi":138
 *             value = decimal.Decimal(value)
 * 
 *         if self.check_exp_overflow:             # <<<<<<<<<<<<<<
 *             _, _, orig_exp = value.as_tuple()
 *             if orig_exp < -self.scale:
 */
  if (__pyx_v_self->check_exp_overflow) {

    /* "src/logical.pxi":139
 * 
 *         if self.check_exp_overflow:
 *             _, _, orig_exp = value.as_tuple()             # <<<<<<<<<<<<<<
 *             if orig_exp < -self.scale:
 *                 raise ExponentTooLarge(f"Decimal value {value} has exponent {orig_exp} which is greater than the scale {self.scale}")
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_as_tuple); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(24, 139, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_7 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(24, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8);
      index = 0; __pyx_t_5 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_2 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 2; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 3) < 0) __PYX_ERR(24, 139, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(24, 139, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __pyx_v__ = __pyx_t_5;
    __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v__, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_v_orig_exp = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "src/logical.pxi":140
 *         if self.check_exp_overflow:
 *             _, _, orig_exp = value.as_tuple()
 *             if orig_exp < -self.scale:             # <<<<<<<<<<<<<<
 *                 raise ExponentTooLarge(f"Decimal value {value} has exponent {orig_exp} which is greater than the scale {self.scale}")
 * 
 */
    __pyx_t_3 = __Pyx_PyInt_From_int((-__pyx_v_self->scale)); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = PyObject_RichCompare(__pyx_v_orig_exp, __pyx_t_3, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 140, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(24, 140, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_t_4)) {

      /* "src/logical.pxi":141
 *             _, _, orig_exp = value.as_tuple()
 *             if orig_exp < -self.scale:
 *                 raise ExponentTooLarge(f"Decimal value {value} has exponent {orig_exp} which is greater than the scale {self.scale}")             # <<<<<<<<<<<<<<
 * 
 *         quant = value.quantize(self.scale_val, context=self.context)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ExponentTooLarge); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = PyTuple_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = 0;
      __pyx_t_11 = 127;
      __Pyx_INCREF(__pyx_kp_u_Decimal_value);
      __pyx_t_10 += 14;
      __Pyx_GIVEREF(__pyx_kp_u_Decimal_value);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Decimal_value);
      __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_v_value, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_11;
      __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u_has_exponent);
      __pyx_t_10 += 14;
      __Pyx_GIVEREF(__pyx_kp_u_has_exponent);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_has_exponent);
      __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_v_orig_exp, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_11;
      __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u_which_is_greater_than_the_scale);
      __pyx_t_10 += 33;
      __Pyx_GIVEREF(__pyx_kp_u_which_is_greater_than_the_scale);
      PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_kp_u_which_is_greater_than_the_scale);
      __pyx_t_5 = __Pyx_PyUnicode_From_int(__pyx_v_self->scale, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 6, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_5};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 141, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(24, 141, __pyx_L1_error)

      /* "src/logical.pxi":140
 *         if self.check_exp_overflow:
 *             _, _, orig_exp = value.as_tuple()
 *             if orig_exp < -self.scale:             # <<<<<<<<<<<<<<
 *                 raise ExponentTooLarge(f"Decimal value {value} has exponent {orig_exp} which is greater than the scale {self.scale}")
 * 
 */
    }

    /* "src/logical.pxi":138
 *             value = decimal.Decimal(value)
 * 
 *         if self.check_exp_overflow:             # <<<<<<<<<<<<<<
 *             _, _, orig_exp = value.as_tuple()
 *             if orig_exp < -self.scale:
 */
  }

  /* "src/logical.pxi":143
 *                 raise ExponentTooLarge(f"Decimal value {value} has exponent {orig_exp} which is greater than the scale {self.scale}")
 * 
 *         quant = value.quantize(self.scale_val, context=self.context)             # <<<<<<<<<<<<<<
 * 
 *         sign, digits, exp = quant.as_tuple()
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_quantize); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->scale_val);
  __Pyx_GIVEREF(__pyx_v_self->scale_val);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->scale_val)) __PYX_ERR(24, 143, __pyx_L1_error);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_context, __pyx_v_self->context) < 0) __PYX_ERR(24, 143, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_quant = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "src/logical.pxi":145
 *         quant = value.quantize(self.scale_val, context=self.context)
 * 
 *         sign, digits, exp = quant.as_tuple()             # <<<<<<<<<<<<<<
 *         assert exp == -self.scale
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_quant, __pyx_n_s_as_tuple); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(24, 145, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_7);
    #else
    __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_8 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(24, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8);
    index = 0; __pyx_t_5 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_5)) goto __pyx_L9_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    index = 1; __pyx_t_3 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_3)) goto __pyx_L9_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 2; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L9_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 3) < 0) __PYX_ERR(24, 145, __pyx_L1_error)
    __pyx_t_9 = NULL;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L10_unpacking_done;
    __pyx_L9_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(24, 145, __pyx_L1_error)
    __pyx_L10_unpacking_done:;
  }
  __pyx_v_sign = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_digits = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_exp = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "src/logical.pxi":146
 * 
 *         sign, digits, exp = quant.as_tuple()
 *         assert exp == -self.scale             # <<<<<<<<<<<<<<
 * 
 *         coeff = 0
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_2 = __Pyx_PyInt_From_int((-__pyx_v_self->scale)); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = PyObject_RichCompare(__pyx_v_exp, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 146, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(24, 146, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_4)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(24, 146, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(24, 146, __pyx_L1_error)
  #endif

  /* "src/logical.pxi":148
 *         assert exp == -self.scale
 * 
 *         coeff = 0             # <<<<<<<<<<<<<<
 *         for d in digits:
 *             coeff = coeff * 10 + d
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_coeff = __pyx_int_0;

  /* "src/logical.pxi":149
 * 
 *         coeff = 0
 *         for d in digits:             # <<<<<<<<<<<<<<
 *             coeff = coeff * 10 + d
 *         if sign:
 */
  if (likely(PyList_CheckExact(__pyx_v_digits)) || PyTuple_CheckExact(__pyx_v_digits)) {
    __pyx_t_7 = __pyx_v_digits; __Pyx_INCREF(__pyx_t_7);
    __pyx_t_10 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_v_digits); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_12 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(24, 149, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_7);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(24, 149, __pyx_L1_error)
          #endif
          if (__pyx_t_10 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(24, 149, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 149, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_7);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(24, 149, __pyx_L1_error)
          #endif
          if (__pyx_t_10 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(24, 149, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 149, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_12(__pyx_t_7);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(24, 149, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_d, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/logical.pxi":150
 *         coeff = 0
 *         for d in digits:
 *             coeff = coeff * 10 + d             # <<<<<<<<<<<<<<
 *         if sign:
 *             coeff = -coeff
 */
    __pyx_t_2 = __Pyx_PyInt_MultiplyObjC(__pyx_v_coeff, __pyx_int_10, 10, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_v_d); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_coeff, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/logical.pxi":149
 * 
 *         coeff = 0
 *         for d in digits:             # <<<<<<<<<<<<<<
 *             coeff = coeff * 10 + d
 *         if sign:
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/logical.pxi":151
 *         for d in digits:
 *             coeff = coeff * 10 + d
 *         if sign:             # <<<<<<<<<<<<<<
 *             coeff = -coeff
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_sign); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(24, 151, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "src/logical.pxi":152
 *             coeff = coeff * 10 + d
 *         if sign:
 *             coeff = -coeff             # <<<<<<<<<<<<<<
 * 
 *         if self.size is not None:
 */
    __pyx_t_7 = PyNumber_Negative(__pyx_v_coeff); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF_SET(__pyx_v_coeff, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "src/logical.pxi":151
 *         for d in digits:
 *             coeff = coeff * 10 + d
 *         if sign:             # <<<<<<<<<<<<<<
 *             coeff = -coeff
 * 
 */
  }

  /* "src/logical.pxi":154
 *             coeff = -coeff
 * 
 *         if self.size is not None:             # <<<<<<<<<<<<<<
 *             return coeff.to_bytes(self.size, byteorder='big', signed=True)
 *         else:
 */
  __pyx_t_4 = (__pyx_v_self->size != Py_None);
  if (__pyx_t_4) {

    /* "src/logical.pxi":155
 * 
 *         if self.size is not None:
 *             return coeff.to_bytes(self.size, byteorder='big', signed=True)             # <<<<<<<<<<<<<<
 *         else:
 *             n_bytes = (coeff.bit_length() + 8) // 8
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_coeff, __pyx_n_s_to_bytes); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_self->size);
    __Pyx_GIVEREF(__pyx_v_self->size);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->size)) __PYX_ERR(24, 155, __pyx_L1_error);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_byteorder, __pyx_n_u_big) < 0) __PYX_ERR(24, 155, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_signed, Py_True) < 0) __PYX_ERR(24, 155, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "src/logical.pxi":154
 *             coeff = -coeff
 * 
 *         if self.size is not None:             # <<<<<<<<<<<<<<
 *             return coeff.to_bytes(self.size, byteorder='big', signed=True)
 *         else:
 */
  }

  /* "src/logical.pxi":157
 *             return coeff.to_bytes(self.size, byteorder='big', signed=True)
 *         else:
 *             n_bytes = (coeff.bit_length() + 8) // 8             # <<<<<<<<<<<<<<
 *             return coeff.to_bytes(n_bytes, byteorder='big', signed=True)
 * 
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_coeff, __pyx_n_s_bit_length); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 157, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_5, __pyx_int_8, 8, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyInt_FloorDivideObjC(__pyx_t_2, __pyx_int_8, 8, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_bytes = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "src/logical.pxi":158
 *         else:
 *             n_bytes = (coeff.bit_length() + 8) // 8
 *             return coeff.to_bytes(n_bytes, byteorder='big', signed=True)             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_coeff, __pyx_n_s_to_bytes); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_n_bytes);
    __Pyx_GIVEREF(__pyx_v_n_bytes);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_n_bytes)) __PYX_ERR(24, 158, __pyx_L1_error);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_byteorder, __pyx_n_u_big) < 0) __PYX_ERR(24, 158, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_signed, Py_True) < 0) __PYX_ERR(24, 158, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;
  }

  /* "src/logical.pxi":132
 *         return cls(precision, scale, size, underlying.options.decimal_check_exp_overflow, underlying.type_name)
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, bytes):
 *             return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cavro.DecimalType.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_orig_exp);
  __Pyx_XDECREF(__pyx_v_quant);
  __Pyx_XDECREF(__pyx_v_sign);
  __Pyx_XDECREF(__pyx_v_digits);
  __Pyx_XDECREF(__pyx_v_exp);
  __Pyx_XDECREF(__pyx_v_coeff);
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XDECREF(__pyx_v_n_bytes);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":160
 *             return coeff.to_bytes(n_bytes, byteorder='big', signed=True)
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         coeff = int.from_bytes(value, byteorder='big', signed=True)
 *         decimal_val = self.context.create_decimal(coeff)
 */

static PyObject *__pyx_f_5cavro_11DecimalType_decode_value(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_coeff = NULL;
  PyObject *__pyx_v_decimal_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/logical.pxi":161
 * 
 *     cdef decode_value(self, value):
 *         coeff = int.from_bytes(value, byteorder='big', signed=True)             # <<<<<<<<<<<<<<
 *         decimal_val = self.context.create_decimal(coeff)
 *         return decimal_val.scaleb(-self.scale, context=self.context)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyInt_Type)), __pyx_n_s_from_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_value)) __PYX_ERR(24, 161, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_byteorder, __pyx_n_u_big) < 0) __PYX_ERR(24, 161, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_signed, Py_True) < 0) __PYX_ERR(24, 161, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_coeff = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "src/logical.pxi":162
 *     cdef decode_value(self, value):
 *         coeff = int.from_bytes(value, byteorder='big', signed=True)
 *         decimal_val = self.context.create_decimal(coeff)             # <<<<<<<<<<<<<<
 *         return decimal_val.scaleb(-self.scale, context=self.context)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->context, __pyx_n_s_create_decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_coeff};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_decimal_val = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "src/logical.pxi":163
 *         coeff = int.from_bytes(value, byteorder='big', signed=True)
 *         decimal_val = self.context.create_decimal(coeff)
 *         return decimal_val.scaleb(-self.scale, context=self.context)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_decimal_val, __pyx_n_s_scaleb); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_From_int((-__pyx_v_self->scale)); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3)) __PYX_ERR(24, 163, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_context, __pyx_v_self->context) < 0) __PYX_ERR(24, 163, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":160
 *             return coeff.to_bytes(n_bytes, byteorder='big', signed=True)
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         coeff = int.from_bytes(value, byteorder='big', signed=True)
 *         decimal_val = self.context.create_decimal(coeff)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.DecimalType.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_coeff);
  __Pyx_XDECREF(__pyx_v_decimal_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":78
 *     underlying_types = (BytesType, FixedType)
 * 
 *     cdef readonly int precision             # <<<<<<<<<<<<<<
 *     cdef readonly int scale
 *     cdef readonly object scale_val
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11DecimalType_9precision_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11DecimalType_9precision_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11DecimalType_9precision___get__(((struct __pyx_obj_5cavro_DecimalType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11DecimalType_9precision___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->precision); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.DecimalType.precision.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":79
 * 
 *     cdef readonly int precision
 *     cdef readonly int scale             # <<<<<<<<<<<<<<
 *     cdef readonly object scale_val
 *     cdef readonly object context
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11DecimalType_5scale_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11DecimalType_5scale_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11DecimalType_5scale___get__(((struct __pyx_obj_5cavro_DecimalType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11DecimalType_5scale___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->scale); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.DecimalType.scale.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":80
 *     cdef readonly int precision
 *     cdef readonly int scale
 *     cdef readonly object scale_val             # <<<<<<<<<<<<<<
 *     cdef readonly object context
 *     cdef readonly object size
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11DecimalType_9scale_val_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11DecimalType_9scale_val_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11DecimalType_9scale_val___get__(((struct __pyx_obj_5cavro_DecimalType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11DecimalType_9scale_val___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->scale_val);
  __pyx_r = __pyx_v_self->scale_val;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":81
 *     cdef readonly int scale
 *     cdef readonly object scale_val
 *     cdef readonly object context             # <<<<<<<<<<<<<<
 *     cdef readonly object size
 *     cdef str underlying_name
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11DecimalType_7context_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11DecimalType_7context_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11DecimalType_7context___get__(((struct __pyx_obj_5cavro_DecimalType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11DecimalType_7context___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->context);
  __pyx_r = __pyx_v_self->context;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":82
 *     cdef readonly object scale_val
 *     cdef readonly object context
 *     cdef readonly object size             # <<<<<<<<<<<<<<
 *     cdef str underlying_name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11DecimalType_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11DecimalType_4size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11DecimalType_4size___get__(((struct __pyx_obj_5cavro_DecimalType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11DecimalType_4size___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->size);
  __pyx_r = __pyx_v_self->size;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":85
 *     cdef str underlying_name
 * 
 *     cdef readonly bint check_exp_overflow             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, precision, scale, size, check_exp_overflow, underlying_name):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11DecimalType_18check_exp_overflow_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_11DecimalType_18check_exp_overflow_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_11DecimalType_18check_exp_overflow___get__(((struct __pyx_obj_5cavro_DecimalType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11DecimalType_18check_exp_overflow___get__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->check_exp_overflow); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.DecimalType.check_exp_overflow.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11DecimalType_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_11DecimalType_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11DecimalType_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_11DecimalType_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_11DecimalType_4__reduce_cython__(((struct __pyx_obj_5cavro_DecimalType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11DecimalType_4__reduce_cython__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.check_exp_overflow, self.context, self.precision, self.scale, self.scale_val, self.size, self.underlying_name)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->check_exp_overflow); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->precision); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->scale); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(7); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->context);
  __Pyx_GIVEREF(__pyx_v_self->context);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_self->context)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->scale_val);
  __Pyx_GIVEREF(__pyx_v_self->scale_val);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_self->scale_val)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->size);
  __Pyx_GIVEREF(__pyx_v_self->size);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_v_self->size)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->underlying_name);
  __Pyx_GIVEREF(__pyx_v_self->underlying_name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_v_self->underlying_name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.check_exp_overflow, self.context, self.precision, self.scale, self.scale_val, self.size, self.underlying_name)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.check_exp_overflow, self.context, self.precision, self.scale, self.scale_val, self.size, self.underlying_name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.context is not None or self.scale_val is not None or self.size is not None or self.underlying_name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.check_exp_overflow, self.context, self.precision, self.scale, self.scale_val, self.size, self.underlying_name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.context is not None or self.scale_val is not None or self.size is not None or self.underlying_name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_DecimalType, (type(self), 0xe9db223, None), state
 */
  /*else*/ {
    __pyx_t_6 = (__pyx_v_self->context != Py_None);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->scale_val != Py_None);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->size != Py_None);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->underlying_name != ((PyObject*)Py_None));
    __pyx_t_5 = __pyx_t_6;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_5;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.context is not None or self.scale_val is not None or self.size is not None or self.underlying_name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DecimalType, (type(self), 0xe9db223, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.context is not None or self.scale_val is not None or self.size is not None or self.underlying_name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_DecimalType, (type(self), 0xe9db223, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_DecimalType, (type(self), 0xe9db223, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_DecimalType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_245215779);
    __Pyx_GIVEREF(__pyx_int_245215779);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_245215779)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.context is not None or self.scale_val is not None or self.size is not None or self.underlying_name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DecimalType, (type(self), 0xe9db223, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_DecimalType, (type(self), 0xe9db223, None), state
 *     else:
 *         return __pyx_unpickle_DecimalType, (type(self), 0xe9db223, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DecimalType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_DecimalType); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_245215779);
    __Pyx_GIVEREF(__pyx_int_245215779);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_245215779)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.DecimalType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DecimalType, (type(self), 0xe9db223, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DecimalType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_11DecimalType_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_11DecimalType_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11DecimalType_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_11DecimalType_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.DecimalType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_11DecimalType_6__setstate_cython__(((struct __pyx_obj_5cavro_DecimalType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_11DecimalType_6__setstate_cython__(struct __pyx_obj_5cavro_DecimalType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_DecimalType, (type(self), 0xe9db223, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DecimalType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_DecimalType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DecimalType, (type(self), 0xe9db223, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DecimalType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.DecimalType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":172
 *     cdef bint return_uuid_object
 * 
 *     def __init__(self, return_uuid_object):             # <<<<<<<<<<<<<<
 *         self.return_uuid_object = return_uuid_object
 * 
 */

/* Python wrapper */
static int __pyx_pw_5cavro_8UUIDBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_8UUIDBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_return_uuid_object = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_return_uuid_object,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_return_uuid_object)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 172, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(24, 172, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_return_uuid_object = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 172, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UUIDBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8UUIDBase___init__(((struct __pyx_obj_5cavro_UUIDBase *)__pyx_v_self), __pyx_v_return_uuid_object);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_8UUIDBase___init__(struct __pyx_obj_5cavro_UUIDBase *__pyx_v_self, PyObject *__pyx_v_return_uuid_object) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/logical.pxi":173
 * 
 *     def __init__(self, return_uuid_object):
 *         self.return_uuid_object = return_uuid_object             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_uuid_object); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(24, 173, __pyx_L1_error)
  __pyx_v_self->return_uuid_object = __pyx_t_1;

  /* "src/logical.pxi":172
 *     cdef bint return_uuid_object
 * 
 *     def __init__(self, return_uuid_object):             # <<<<<<<<<<<<<<
 *         self.return_uuid_object = return_uuid_object
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.UUIDBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/logical.pxi":175
 *         self.return_uuid_object = return_uuid_object
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.return_uuid_object)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8UUIDBase_3_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8UUIDBase_3_for_type = {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8UUIDBase_3_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8UUIDBase_3_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_for_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_underlying,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_underlying)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 175, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_for_type") < 0)) __PYX_ERR(24, 175, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_underlying = ((struct __pyx_obj_5cavro_AvroType *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_for_type", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 175, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UUIDBase._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_underlying), __pyx_ptype_5cavro_AvroType, 0, "underlying", 0))) __PYX_ERR(24, 176, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_8UUIDBase_2_for_type(((PyTypeObject*)__pyx_v_cls), __pyx_v_underlying);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8UUIDBase_2_for_type(PyTypeObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_type", 1);

  /* "src/logical.pxi":177
 *     @classmethod
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.return_uuid_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_underlying->options->return_uuid_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_v_cls), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":175
 *         self.return_uuid_object = return_uuid_object
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.return_uuid_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.UUIDBase._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8UUIDBase_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8UUIDBase_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8UUIDBase_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8UUIDBase_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_8UUIDBase_4__reduce_cython__(((struct __pyx_obj_5cavro_UUIDBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8UUIDBase_4__reduce_cython__(struct __pyx_obj_5cavro_UUIDBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.return_uuid_object,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->return_uuid_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.return_uuid_object,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.return_uuid_object,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.return_uuid_object,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_UUIDBase, (type(self), 0x987bfb3, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UUIDBase, (type(self), 0x987bfb3, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_UUIDBase, (type(self), 0x987bfb3, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_UUIDBase, (type(self), 0x987bfb3, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_UUIDBase); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_159891379);
    __Pyx_GIVEREF(__pyx_int_159891379);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_159891379)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UUIDBase, (type(self), 0x987bfb3, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_UUIDBase, (type(self), 0x987bfb3, None), state
 *     else:
 *         return __pyx_unpickle_UUIDBase, (type(self), 0x987bfb3, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UUIDBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_UUIDBase); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_159891379);
    __Pyx_GIVEREF(__pyx_int_159891379);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_159891379)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.UUIDBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UUIDBase, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UUIDBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_8UUIDBase_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_8UUIDBase_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8UUIDBase_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_8UUIDBase_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UUIDBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_8UUIDBase_6__setstate_cython__(((struct __pyx_obj_5cavro_UUIDBase *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_8UUIDBase_6__setstate_cython__(struct __pyx_obj_5cavro_UUIDBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_UUIDBase, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UUIDBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_UUIDBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UUIDBase, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UUIDBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.UUIDBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":183
 *     underlying_types = (StringType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.return_uuid_object)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14UUIDStringType_1_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_14UUIDStringType_1_for_type = {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14UUIDStringType_1_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_14UUIDStringType_1_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_for_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_underlying,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_underlying)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 183, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_for_type") < 0)) __PYX_ERR(24, 183, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_underlying = ((struct __pyx_obj_5cavro_AvroType *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_for_type", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 183, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UUIDStringType._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_underlying), __pyx_ptype_5cavro_AvroType, 0, "underlying", 0))) __PYX_ERR(24, 184, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_14UUIDStringType__for_type(((PyTypeObject*)__pyx_v_cls), __pyx_v_underlying);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14UUIDStringType__for_type(PyTypeObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_type", 1);

  /* "src/logical.pxi":185
 *     @classmethod
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.return_uuid_object)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_underlying->options->return_uuid_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_v_cls), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":183
 *     underlying_types = (StringType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.return_uuid_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.UUIDStringType._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":187
 *         return cls(underlying.options.return_uuid_object)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'string(uuid)'
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14UUIDStringType_9type_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_14UUIDStringType_9type_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_14UUIDStringType_9type_name___get__(((struct __pyx_obj_5cavro_UUIDStringType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14UUIDStringType_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_UUIDStringType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/logical.pxi":189
 *     @property
 *     def type_name(self):
 *         return 'string(uuid)'             # <<<<<<<<<<<<<<
 * 
 *     cdef encode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_kp_u_string_uuid);
  __pyx_r = __pyx_kp_u_string_uuid;
  goto __pyx_L0;

  /* "src/logical.pxi":187
 *         return cls(underlying.options.return_uuid_object)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'string(uuid)'
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":191
 *         return 'string(uuid)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         # apache avro lib behaviour is weird here,
 *         # value should be validated, but unmodified value should
 */

static PyObject *__pyx_f_5cavro_14UUIDStringType_encode_value(struct __pyx_obj_5cavro_UUIDStringType *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 1);

  /* "src/logical.pxi":195
 *         # value should be validated, but unmodified value should
 *         # be encoded (i.e. the presence of '-' or not should be retained)
 *         if isinstance(value, uuid.UUID):             # <<<<<<<<<<<<<<
 *             return str(value)
 *         try:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_uuid); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_UUID); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(24, 195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "src/logical.pxi":196
 *         # be encoded (i.e. the presence of '-' or not should be retained)
 *         if isinstance(value, uuid.UUID):
 *             return str(value)             # <<<<<<<<<<<<<<
 *         try:
 *             _ = uuid.UUID(hex=value)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_Str(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "src/logical.pxi":195
 *         # value should be validated, but unmodified value should
 *         # be encoded (i.e. the presence of '-' or not should be retained)
 *         if isinstance(value, uuid.UUID):             # <<<<<<<<<<<<<<
 *             return str(value)
 *         try:
 */
  }

  /* "src/logical.pxi":197
 *         if isinstance(value, uuid.UUID):
 *             return str(value)
 *         try:             # <<<<<<<<<<<<<<
 *             _ = uuid.UUID(hex=value)
 *         except TypeError as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "src/logical.pxi":198
 *             return str(value)
 *         try:
 *             _ = uuid.UUID(hex=value)             # <<<<<<<<<<<<<<
 *         except TypeError as e:
 *             raise InvalidValue(value, self) from e
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_uuid); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 198, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_UUID); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 198, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 198, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_hex, __pyx_v_value) < 0) __PYX_ERR(24, 198, __pyx_L4_error)
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 198, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v__ = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "src/logical.pxi":197
 *         if isinstance(value, uuid.UUID):
 *             return str(value)
 *         try:             # <<<<<<<<<<<<<<
 *             _ = uuid.UUID(hex=value)
 *         except TypeError as e:
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "src/logical.pxi":199
 *         try:
 *             _ = uuid.UUID(hex=value)
 *         except TypeError as e:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self) from e
 *         return value
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("cavro.UUIDStringType.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(24, 199, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_v_e = __pyx_t_2;
      /*try:*/ {

        /* "src/logical.pxi":200
 *             _ = uuid.UUID(hex=value)
 *         except TypeError as e:
 *             raise InvalidValue(value, self) from e             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_InvalidValue); if (unlikely(!__pyx_t_10)) __PYX_ERR(24, 200, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_11, __pyx_v_value, ((PyObject *)__pyx_v_self)};
          __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(24, 200, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_Raise(__pyx_t_9, 0, 0, __pyx_v_e);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(24, 200, __pyx_L15_error)
      }

      /* "src/logical.pxi":199
 *         try:
 *             _ = uuid.UUID(hex=value)
 *         except TypeError as e:             # <<<<<<<<<<<<<<
 *             raise InvalidValue(value, self) from e
 *         return value
 */
      /*finally:*/ {
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_8 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L6_except_error;
        }
      }
    }
    goto __pyx_L6_except_error;

    /* "src/logical.pxi":197
 *         if isinstance(value, uuid.UUID):
 *             return str(value)
 *         try:             # <<<<<<<<<<<<<<
 *             _ = uuid.UUID(hex=value)
 *         except TypeError as e:
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L9_try_end:;
  }

  /* "src/logical.pxi":201
 *         except TypeError as e:
 *             raise InvalidValue(value, self) from e
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/logical.pxi":191
 *         return 'string(uuid)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         # apache avro lib behaviour is weird here,
 *         # value should be validated, but unmodified value should
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("cavro.UUIDStringType.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":203
 *         return value
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         if self.return_uuid_object:
 *             return uuid.UUID(value)
 */

static PyObject *__pyx_f_5cavro_14UUIDStringType_decode_value(struct __pyx_obj_5cavro_UUIDStringType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/logical.pxi":204
 * 
 *     cdef decode_value(self, value):
 *         if self.return_uuid_object:             # <<<<<<<<<<<<<<
 *             return uuid.UUID(value)
 *         return value
 */
  if (__pyx_v_self->__pyx_base.return_uuid_object) {

    /* "src/logical.pxi":205
 *     cdef decode_value(self, value):
 *         if self.return_uuid_object:
 *             return uuid.UUID(value)             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_uuid); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_UUID); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_value};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "src/logical.pxi":204
 * 
 *     cdef decode_value(self, value):
 *         if self.return_uuid_object:             # <<<<<<<<<<<<<<
 *             return uuid.UUID(value)
 *         return value
 */
  }

  /* "src/logical.pxi":206
 *         if self.return_uuid_object:
 *             return uuid.UUID(value)
 *         return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/logical.pxi":203
 *         return value
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         if self.return_uuid_object:
 *             return uuid.UUID(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.UUIDStringType.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14UUIDStringType_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_14UUIDStringType_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14UUIDStringType_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_14UUIDStringType_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_14UUIDStringType_2__reduce_cython__(((struct __pyx_obj_5cavro_UUIDStringType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14UUIDStringType_2__reduce_cython__(struct __pyx_obj_5cavro_UUIDStringType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.return_uuid_object,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.return_uuid_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.return_uuid_object,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.return_uuid_object,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.return_uuid_object,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_UUIDStringType, (type(self), 0x987bfb3, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UUIDStringType, (type(self), 0x987bfb3, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_UUIDStringType, (type(self), 0x987bfb3, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_UUIDStringType, (type(self), 0x987bfb3, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_UUIDStringType); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_159891379);
    __Pyx_GIVEREF(__pyx_int_159891379);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_159891379)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UUIDStringType, (type(self), 0x987bfb3, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_UUIDStringType, (type(self), 0x987bfb3, None), state
 *     else:
 *         return __pyx_unpickle_UUIDStringType, (type(self), 0x987bfb3, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UUIDStringType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_UUIDStringType); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_159891379);
    __Pyx_GIVEREF(__pyx_int_159891379);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_159891379)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.UUIDStringType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UUIDStringType, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UUIDStringType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14UUIDStringType_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_14UUIDStringType_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14UUIDStringType_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_14UUIDStringType_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UUIDStringType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_14UUIDStringType_4__setstate_cython__(((struct __pyx_obj_5cavro_UUIDStringType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14UUIDStringType_4__setstate_cython__(struct __pyx_obj_5cavro_UUIDStringType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_UUIDStringType, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UUIDStringType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_UUIDStringType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UUIDStringType, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UUIDStringType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.UUIDStringType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":212
 *     underlying_types = (FixedType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         if underlying.size != 16:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_13UUIDFixedType_1_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_13UUIDFixedType_1_for_type = {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13UUIDFixedType_1_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_13UUIDFixedType_1_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_for_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_underlying,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_underlying)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 212, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_for_type") < 0)) __PYX_ERR(24, 212, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_underlying = ((struct __pyx_obj_5cavro_AvroType *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_for_type", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 212, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UUIDFixedType._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_underlying), __pyx_ptype_5cavro_AvroType, 0, "underlying", 0))) __PYX_ERR(24, 213, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_13UUIDFixedType__for_type(((PyTypeObject*)__pyx_v_cls), __pyx_v_underlying);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_13UUIDFixedType__for_type(PyTypeObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_type", 1);

  /* "src/logical.pxi":214
 *     @classmethod
 *     def _for_type(cls, underlying: AvroType):
 *         if underlying.size != 16:             # <<<<<<<<<<<<<<
 *             return None
 *         return cls(underlying.options.return_uuid_object)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_underlying), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PyInt_BoolNeObjC(__pyx_t_1, __pyx_int_16, 16, 0)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(24, 214, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/logical.pxi":215
 *     def _for_type(cls, underlying: AvroType):
 *         if underlying.size != 16:
 *             return None             # <<<<<<<<<<<<<<
 *         return cls(underlying.options.return_uuid_object)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "src/logical.pxi":214
 *     @classmethod
 *     def _for_type(cls, underlying: AvroType):
 *         if underlying.size != 16:             # <<<<<<<<<<<<<<
 *             return None
 *         return cls(underlying.options.return_uuid_object)
 */
  }

  /* "src/logical.pxi":216
 *         if underlying.size != 16:
 *             return None
 *         return cls(underlying.options.return_uuid_object)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_underlying->options->return_uuid_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_v_cls), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":212
 *     underlying_types = (FixedType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         if underlying.size != 16:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.UUIDFixedType._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":218
 *         return cls(underlying.options.return_uuid_object)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'fixed(uuid)'
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_13UUIDFixedType_9type_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_13UUIDFixedType_9type_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_13UUIDFixedType_9type_name___get__(((struct __pyx_obj_5cavro_UUIDFixedType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_13UUIDFixedType_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_UUIDFixedType *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/logical.pxi":220
 *     @property
 *     def type_name(self):
 *         return 'fixed(uuid)'             # <<<<<<<<<<<<<<
 * 
 *     cdef encode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_kp_u_fixed_uuid);
  __pyx_r = __pyx_kp_u_fixed_uuid;
  goto __pyx_L0;

  /* "src/logical.pxi":218
 *         return cls(underlying.options.return_uuid_object)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'fixed(uuid)'
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":222
 *         return 'fixed(uuid)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, uuid.UUID):
 *             value = uuid.UUID(value)
 */

static PyObject *__pyx_f_5cavro_13UUIDFixedType_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_UUIDFixedType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/logical.pxi":223
 * 
 *     cdef encode_value(self, value):
 *         if not isinstance(value, uuid.UUID):             # <<<<<<<<<<<<<<
 *             value = uuid.UUID(value)
 *         return value.bytes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_uuid); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_UUID); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(24, 223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (!__pyx_t_3);
  if (__pyx_t_4) {

    /* "src/logical.pxi":224
 *     cdef encode_value(self, value):
 *         if not isinstance(value, uuid.UUID):
 *             value = uuid.UUID(value)             # <<<<<<<<<<<<<<
 *         return value.bytes
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_uuid); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_UUID); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_value};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "src/logical.pxi":223
 * 
 *     cdef encode_value(self, value):
 *         if not isinstance(value, uuid.UUID):             # <<<<<<<<<<<<<<
 *             value = uuid.UUID(value)
 *         return value.bytes
 */
  }

  /* "src/logical.pxi":225
 *         if not isinstance(value, uuid.UUID):
 *             value = uuid.UUID(value)
 *         return value.bytes             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_bytes_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":222
 *         return 'fixed(uuid)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, uuid.UUID):
 *             value = uuid.UUID(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.UUIDFixedType.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":227
 *         return value.bytes
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         if self.return_uuid_object:
 *             return uuid.UUID(bytes=value)
 */

static PyObject *__pyx_f_5cavro_13UUIDFixedType_decode_value(struct __pyx_obj_5cavro_UUIDFixedType *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/logical.pxi":228
 * 
 *     cdef decode_value(self, value):
 *         if self.return_uuid_object:             # <<<<<<<<<<<<<<
 *             return uuid.UUID(bytes=value)
 *         return value
 */
  if (__pyx_v_self->__pyx_base.return_uuid_object) {

    /* "src/logical.pxi":229
 *     cdef decode_value(self, value):
 *         if self.return_uuid_object:
 *             return uuid.UUID(bytes=value)             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_uuid); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_UUID); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_bytes_3, __pyx_v_value) < 0) __PYX_ERR(24, 229, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "src/logical.pxi":228
 * 
 *     cdef decode_value(self, value):
 *         if self.return_uuid_object:             # <<<<<<<<<<<<<<
 *             return uuid.UUID(bytes=value)
 *         return value
 */
  }

  /* "src/logical.pxi":230
 *         if self.return_uuid_object:
 *             return uuid.UUID(bytes=value)
 *         return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "src/logical.pxi":227
 *         return value.bytes
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         if self.return_uuid_object:
 *             return uuid.UUID(bytes=value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.UUIDFixedType.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_13UUIDFixedType_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_13UUIDFixedType_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13UUIDFixedType_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_13UUIDFixedType_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_13UUIDFixedType_2__reduce_cython__(((struct __pyx_obj_5cavro_UUIDFixedType *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_13UUIDFixedType_2__reduce_cython__(struct __pyx_obj_5cavro_UUIDFixedType *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.return_uuid_object,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.return_uuid_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.return_uuid_object,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.return_uuid_object,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.return_uuid_object,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_UUIDFixedType, (type(self), 0x987bfb3, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UUIDFixedType, (type(self), 0x987bfb3, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_UUIDFixedType, (type(self), 0x987bfb3, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_UUIDFixedType, (type(self), 0x987bfb3, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_UUIDFixedType); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_159891379);
    __Pyx_GIVEREF(__pyx_int_159891379);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_159891379)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_UUIDFixedType, (type(self), 0x987bfb3, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_UUIDFixedType, (type(self), 0x987bfb3, None), state
 *     else:
 *         return __pyx_unpickle_UUIDFixedType, (type(self), 0x987bfb3, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UUIDFixedType__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_UUIDFixedType); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_159891379);
    __Pyx_GIVEREF(__pyx_int_159891379);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_159891379)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.UUIDFixedType.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UUIDFixedType, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UUIDFixedType__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_13UUIDFixedType_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_13UUIDFixedType_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13UUIDFixedType_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_13UUIDFixedType_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.UUIDFixedType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_13UUIDFixedType_4__setstate_cython__(((struct __pyx_obj_5cavro_UUIDFixedType *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_13UUIDFixedType_4__setstate_cython__(struct __pyx_obj_5cavro_UUIDFixedType *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_UUIDFixedType, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_UUIDFixedType__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_UUIDFixedType__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UUIDFixedType, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UUIDFixedType__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.UUIDFixedType.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":244
 *     cdef readonly bint accepts_string
 * 
 *     def __init__(self, accepts_string):             # <<<<<<<<<<<<<<
 *         self.accepts_string = accepts_string
 * 
 */

/* Python wrapper */
static int __pyx_pw_5cavro_4Date_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_4Date_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_accepts_string = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_accepts_string,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_accepts_string)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 244, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(24, 244, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_accepts_string = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 244, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Date.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_4Date___init__(((struct __pyx_obj_5cavro_Date *)__pyx_v_self), __pyx_v_accepts_string);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_4Date___init__(struct __pyx_obj_5cavro_Date *__pyx_v_self, PyObject *__pyx_v_accepts_string) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/logical.pxi":245
 * 
 *     def __init__(self, accepts_string):
 *         self.accepts_string = accepts_string             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_accepts_string); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(24, 245, __pyx_L1_error)
  __pyx_v_self->accepts_string = __pyx_t_1;

  /* "src/logical.pxi":244
 *     cdef readonly bint accepts_string
 * 
 *     def __init__(self, accepts_string):             # <<<<<<<<<<<<<<
 *         self.accepts_string = accepts_string
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.Date.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/logical.pxi":247
 *         self.accepts_string = accepts_string
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.date_type_accepts_string)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_4Date_3_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_4Date_3_for_type = {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_4Date_3_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_4Date_3_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_for_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_underlying,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_underlying)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 247, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_for_type") < 0)) __PYX_ERR(24, 247, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_underlying = ((struct __pyx_obj_5cavro_AvroType *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_for_type", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 247, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Date._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_underlying), __pyx_ptype_5cavro_AvroType, 0, "underlying", 0))) __PYX_ERR(24, 248, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_4Date_2_for_type(((PyTypeObject*)__pyx_v_cls), __pyx_v_underlying);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_4Date_2_for_type(PyTypeObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_type", 1);

  /* "src/logical.pxi":249
 *     @classmethod
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.date_type_accepts_string)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_underlying->options->date_type_accepts_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_v_cls), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":247
 *         self.accepts_string = accepts_string
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.date_type_accepts_string)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.Date._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":251
 *         return cls(underlying.options.date_type_accepts_string)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'int(date)'
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_4Date_9type_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_4Date_9type_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_4Date_9type_name___get__(((struct __pyx_obj_5cavro_Date *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_4Date_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_Date *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/logical.pxi":253
 *     @property
 *     def type_name(self):
 *         return 'int(date)'             # <<<<<<<<<<<<<<
 * 
 *     cdef encode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_kp_u_int_date);
  __pyx_r = __pyx_kp_u_int_date;
  goto __pyx_L0;

  /* "src/logical.pxi":251
 *         return cls(underlying.options.date_type_accepts_string)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'int(date)'
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":255
 *         return 'int(date)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if self.accepts_string and isinstance(value, str):
 *             value = datetime.datetime.strptime(value, '%Y-%m-%d').date()
 */

static PyObject *__pyx_f_5cavro_4Date_encode_value(struct __pyx_obj_5cavro_Date *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/logical.pxi":256
 * 
 *     cdef encode_value(self, value):
 *         if self.accepts_string and isinstance(value, str):             # <<<<<<<<<<<<<<
 *             value = datetime.datetime.strptime(value, '%Y-%m-%d').date()
 *         if isinstance(value, int) and value >= 0 and value <= 2**16:
 */
  if (__pyx_v_self->accepts_string) {
  } else {
    __pyx_t_1 = __pyx_v_self->accepts_string;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyUnicode_Check(__pyx_v_value); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/logical.pxi":257
 *     cdef encode_value(self, value):
 *         if self.accepts_string and isinstance(value, str):
 *             value = datetime.datetime.strptime(value, '%Y-%m-%d').date()             # <<<<<<<<<<<<<<
 *         if isinstance(value, int) and value >= 0 and value <= 2**16:
 *             return value
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(24, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_strptime); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_value, __pyx_kp_u_Y_m_d};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 257, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_date); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 257, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/logical.pxi":256
 * 
 *     cdef encode_value(self, value):
 *         if self.accepts_string and isinstance(value, str):             # <<<<<<<<<<<<<<
 *             value = datetime.datetime.strptime(value, '%Y-%m-%d').date()
 *         if isinstance(value, int) and value >= 0 and value <= 2**16:
 */
  }

  /* "src/logical.pxi":258
 *         if self.accepts_string and isinstance(value, str):
 *             value = datetime.datetime.strptime(value, '%Y-%m-%d').date()
 *         if isinstance(value, int) and value >= 0 and value <= 2**16:             # <<<<<<<<<<<<<<
 *             return value
 *         if isinstance(value, datetime.datetime):
 */
  __pyx_t_2 = PyInt_Check(__pyx_v_value); 
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_value, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 258, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(24, 258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_value, __pyx_int_65536, Py_LE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 258, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(24, 258, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/logical.pxi":259
 *             value = datetime.datetime.strptime(value, '%Y-%m-%d').date()
 *         if isinstance(value, int) and value >= 0 and value <= 2**16:
 *             return value             # <<<<<<<<<<<<<<
 *         if isinstance(value, datetime.datetime):
 *             value = value.date()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/logical.pxi":258
 *         if self.accepts_string and isinstance(value, str):
 *             value = datetime.datetime.strptime(value, '%Y-%m-%d').date()
 *         if isinstance(value, int) and value >= 0 and value <= 2**16:             # <<<<<<<<<<<<<<
 *             return value
 *         if isinstance(value, datetime.datetime):
 */
  }

  /* "src/logical.pxi":260
 *         if isinstance(value, int) and value >= 0 and value <= 2**16:
 *             return value
 *         if isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *             value = value.date()
 *         if isinstance(value, datetime.date):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_value, __pyx_t_5); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(24, 260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_1) {

    /* "src/logical.pxi":261
 *             return value
 *         if isinstance(value, datetime.datetime):
 *             value = value.date()             # <<<<<<<<<<<<<<
 *         if isinstance(value, datetime.date):
 *             return (value - EPOCH_DATE).days
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_date); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "src/logical.pxi":260
 *         if isinstance(value, int) and value >= 0 and value <= 2**16:
 *             return value
 *         if isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *             value = value.date()
 *         if isinstance(value, datetime.date):
 */
  }

  /* "src/logical.pxi":262
 *         if isinstance(value, datetime.datetime):
 *             value = value.date()
 *         if isinstance(value, datetime.date):             # <<<<<<<<<<<<<<
 *             return (value - EPOCH_DATE).days
 *         raise ValueError(f"Expected datetime.date, got {type(value)}")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_date); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(24, 262, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "src/logical.pxi":263
 *             value = value.date()
 *         if isinstance(value, datetime.date):
 *             return (value - EPOCH_DATE).days             # <<<<<<<<<<<<<<
 *         raise ValueError(f"Expected datetime.date, got {type(value)}")
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_EPOCH_DATE); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Subtract(__pyx_v_value, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_days); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "src/logical.pxi":262
 *         if isinstance(value, datetime.datetime):
 *             value = value.date()
 *         if isinstance(value, datetime.date):             # <<<<<<<<<<<<<<
 *             return (value - EPOCH_DATE).days
 *         raise ValueError(f"Expected datetime.date, got {type(value)}")
 */
  }

  /* "src/logical.pxi":264
 *         if isinstance(value, datetime.date):
 *             return (value - EPOCH_DATE).days
 *         raise ValueError(f"Expected datetime.date, got {type(value)}")             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __pyx_t_3 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_value)), __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Expected_datetime_date_got, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_Raise(__pyx_t_3, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __PYX_ERR(24, 264, __pyx_L1_error)

  /* "src/logical.pxi":255
 *         return 'int(date)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if self.accepts_string and isinstance(value, str):
 *             value = datetime.datetime.strptime(value, '%Y-%m-%d').date()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.Date.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":266
 *         raise ValueError(f"Expected datetime.date, got {type(value)}")
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return EPOCH_DATE + datetime.timedelta(days=value)
 * 
 */

static PyObject *__pyx_f_5cavro_4Date_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_Date *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/logical.pxi":267
 * 
 *     cdef decode_value(self, value):
 *         return EPOCH_DATE + datetime.timedelta(days=value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_EPOCH_DATE); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_datetime); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_timedelta); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_days, __pyx_v_value) < 0) __PYX_ERR(24, 267, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":266
 *         raise ValueError(f"Expected datetime.date, got {type(value)}")
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return EPOCH_DATE + datetime.timedelta(days=value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Date.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":242
 *     underlying_types = (IntType, )
 * 
 *     cdef readonly bint accepts_string             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, accepts_string):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_4Date_14accepts_string_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_4Date_14accepts_string_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_4Date_14accepts_string___get__(((struct __pyx_obj_5cavro_Date *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_4Date_14accepts_string___get__(struct __pyx_obj_5cavro_Date *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->accepts_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Date.accepts_string.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_4Date_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_4Date_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_4Date_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_4Date_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_4Date_4__reduce_cython__(((struct __pyx_obj_5cavro_Date *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_4Date_4__reduce_cython__(struct __pyx_obj_5cavro_Date *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.accepts_string,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->accepts_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.accepts_string,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.accepts_string,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.accepts_string,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Date, (type(self), 0xc5331ff, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Date, (type(self), 0xc5331ff, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_Date, (type(self), 0xc5331ff, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Date, (type(self), 0xc5331ff, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Date); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_206778879);
    __Pyx_GIVEREF(__pyx_int_206778879);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_206778879)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Date, (type(self), 0xc5331ff, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Date, (type(self), 0xc5331ff, None), state
 *     else:
 *         return __pyx_unpickle_Date, (type(self), 0xc5331ff, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Date__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Date); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_206778879);
    __Pyx_GIVEREF(__pyx_int_206778879);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_206778879)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Date.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Date, (type(self), 0xc5331ff, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Date__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_4Date_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_4Date_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_4Date_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_4Date_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Date.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_4Date_6__setstate_cython__(((struct __pyx_obj_5cavro_Date *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_4Date_6__setstate_cython__(struct __pyx_obj_5cavro_Date *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Date, (type(self), 0xc5331ff, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Date__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_Date__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Date, (type(self), 0xc5331ff, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Date__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Date.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":270
 * 
 * 
 * def _time_to_micros(value):             # <<<<<<<<<<<<<<
 *     return value.microsecond + 1_000_000 * (value.second + 60 * (value.minute + 60 * value.hour))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15_time_to_micros(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15_time_to_micros = {"_time_to_micros", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15_time_to_micros, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15_time_to_micros(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_time_to_micros (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 270, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_time_to_micros") < 0)) __PYX_ERR(24, 270, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_time_to_micros", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 270, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._time_to_micros", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_14_time_to_micros(__pyx_self, __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14_time_to_micros(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_time_to_micros", 1);

  /* "src/logical.pxi":271
 * 
 * def _time_to_micros(value):
 *     return value.microsecond + 1_000_000 * (value.second + 60 * (value.minute + 60 * value.hour))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_microsecond); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_second); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_minute); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_hour); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_MultiplyCObj(__pyx_int_60, __pyx_t_4, 60, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_MultiplyCObj(__pyx_int_60, __pyx_t_4, 60, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Add(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_MultiplyCObj(__pyx_int_1000000, __pyx_t_4, 0xF4240, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":270
 * 
 * 
 * def _time_to_micros(value):             # <<<<<<<<<<<<<<
 *     return value.microsecond + 1_000_000 * (value.second + 60 * (value.minute + 60 * value.hour))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro._time_to_micros", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":274
 * 
 * 
 * def _micros_to_time(value):             # <<<<<<<<<<<<<<
 *     return datetime.time(
 *         hour=value // 3_600_000_000,
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_17_micros_to_time(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_17_micros_to_time = {"_micros_to_time", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_17_micros_to_time, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_17_micros_to_time(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_micros_to_time (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 274, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_micros_to_time") < 0)) __PYX_ERR(24, 274, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_micros_to_time", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 274, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._micros_to_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_16_micros_to_time(__pyx_self, __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_16_micros_to_time(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_micros_to_time", 1);

  /* "src/logical.pxi":275
 * 
 * def _micros_to_time(value):
 *     return datetime.time(             # <<<<<<<<<<<<<<
 *         hour=value // 3_600_000_000,
 *         minute=(value // 60_000_000) % 60,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_datetime); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/logical.pxi":276
 * def _micros_to_time(value):
 *     return datetime.time(
 *         hour=value // 3_600_000_000,             # <<<<<<<<<<<<<<
 *         minute=(value // 60_000_000) % 60,
 *         second=(value // 1_000_000) % 60,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_FloorDivide(__pyx_v_value, __pyx_int_3600000000); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_hour, __pyx_t_3) < 0) __PYX_ERR(24, 276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "src/logical.pxi":277
 *     return datetime.time(
 *         hour=value // 3_600_000_000,
 *         minute=(value // 60_000_000) % 60,             # <<<<<<<<<<<<<<
 *         second=(value // 1_000_000) % 60,
 *         microsecond=value % 100_0000
 */
  __pyx_t_3 = __Pyx_PyInt_FloorDivideObjC(__pyx_v_value, __pyx_int_60000000, 0x3938700, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_RemainderObjC(__pyx_t_3, __pyx_int_60, 60, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_minute, __pyx_t_4) < 0) __PYX_ERR(24, 276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/logical.pxi":278
 *         hour=value // 3_600_000_000,
 *         minute=(value // 60_000_000) % 60,
 *         second=(value // 1_000_000) % 60,             # <<<<<<<<<<<<<<
 *         microsecond=value % 100_0000
 *     )
 */
  __pyx_t_4 = __Pyx_PyInt_FloorDivideObjC(__pyx_v_value, __pyx_int_1000000, 0xF4240, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_RemainderObjC(__pyx_t_4, __pyx_int_60, 60, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_second, __pyx_t_3) < 0) __PYX_ERR(24, 276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "src/logical.pxi":279
 *         minute=(value // 60_000_000) % 60,
 *         second=(value // 1_000_000) % 60,
 *         microsecond=value % 100_0000             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_RemainderObjC(__pyx_v_value, __pyx_int_1000000, 0xF4240, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_microsecond, __pyx_t_3) < 0) __PYX_ERR(24, 276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "src/logical.pxi":275
 * 
 * def _micros_to_time(value):
 *     return datetime.time(             # <<<<<<<<<<<<<<
 *         hour=value // 3_600_000_000,
 *         minute=(value // 60_000_000) % 60,
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":274
 * 
 * 
 * def _micros_to_time(value):             # <<<<<<<<<<<<<<
 *     return datetime.time(
 *         hour=value // 3_600_000_000,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._micros_to_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":286
 * 
 * 
 * cdef object dt_to_timedelta(value):             # <<<<<<<<<<<<<<
 *     if value.tzinfo is None:
 *         value = value.astimezone(datetime.timezone.utc)
 */

static PyObject *__pyx_f_5cavro_dt_to_timedelta(PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dt_to_timedelta", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/logical.pxi":287
 * 
 * cdef object dt_to_timedelta(value):
 *     if value.tzinfo is None:             # <<<<<<<<<<<<<<
 *         value = value.astimezone(datetime.timezone.utc)
 *     return value - EPOCH_DT
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_tzinfo); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "src/logical.pxi":288
 * cdef object dt_to_timedelta(value):
 *     if value.tzinfo is None:
 *         value = value.astimezone(datetime.timezone.utc)             # <<<<<<<<<<<<<<
 *     return value - EPOCH_DT
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_astimezone); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_timezone); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_utc); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/logical.pxi":287
 * 
 * cdef object dt_to_timedelta(value):
 *     if value.tzinfo is None:             # <<<<<<<<<<<<<<
 *         value = value.astimezone(datetime.timezone.utc)
 *     return value - EPOCH_DT
 */
  }

  /* "src/logical.pxi":289
 *     if value.tzinfo is None:
 *         value = value.astimezone(datetime.timezone.utc)
 *     return value - EPOCH_DT             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_EPOCH_DT); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_Subtract(__pyx_v_value, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":286
 * 
 * 
 * cdef object dt_to_timedelta(value):             # <<<<<<<<<<<<<<
 *     if value.tzinfo is None:
 *         value = value.astimezone(datetime.timezone.utc)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.dt_to_timedelta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":297
 *     underlying_types = (IntType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10TimeMillis_1_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10TimeMillis_1_for_type = {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10TimeMillis_1_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10TimeMillis_1_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_for_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_underlying,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_underlying)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 297, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_for_type") < 0)) __PYX_ERR(24, 297, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_underlying = ((struct __pyx_obj_5cavro_AvroType *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_for_type", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 297, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.TimeMillis._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_underlying), __pyx_ptype_5cavro_AvroType, 0, "underlying", 0))) __PYX_ERR(24, 298, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_10TimeMillis__for_type(((PyTypeObject*)__pyx_v_cls), __pyx_v_underlying);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10TimeMillis__for_type(PyTypeObject *__pyx_v_cls, CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_type", 1);

  /* "src/logical.pxi":299
 *     @classmethod
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_v_cls)); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":297
 *     underlying_types = (IntType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.TimeMillis._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":301
 *         return cls()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'int(time-millis)'
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10TimeMillis_9type_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_10TimeMillis_9type_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_10TimeMillis_9type_name___get__(((struct __pyx_obj_5cavro_TimeMillis *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10TimeMillis_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_TimeMillis *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/logical.pxi":303
 *     @property
 *     def type_name(self):
 *         return 'int(time-millis)'             # <<<<<<<<<<<<<<
 * 
 *     cdef encode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_kp_u_int_time_millis);
  __pyx_r = __pyx_kp_u_int_time_millis;
  goto __pyx_L0;

  /* "src/logical.pxi":301
 *         return cls()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'int(time-millis)'
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":305
 *         return 'int(time-millis)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, int) and value >= 0 and value <= (MAX_TIME // 1_000):
 *             return value
 */

static PyObject *__pyx_f_5cavro_10TimeMillis_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimeMillis *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/logical.pxi":306
 * 
 *     cdef encode_value(self, value):
 *         if isinstance(value, int) and value >= 0 and value <= (MAX_TIME // 1_000):             # <<<<<<<<<<<<<<
 *             return value
 *         if isinstance(value, datetime.datetime):
 */
  __pyx_t_2 = PyInt_Check(__pyx_v_value); 
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_value, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 306, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(24, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_MAX_TIME); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_FloorDivideObjC(__pyx_t_3, __pyx_int_1000, 0x3E8, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_value, __pyx_t_4, Py_LE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(24, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/logical.pxi":307
 *     cdef encode_value(self, value):
 *         if isinstance(value, int) and value >= 0 and value <= (MAX_TIME // 1_000):
 *             return value             # <<<<<<<<<<<<<<
 *         if isinstance(value, datetime.datetime):
 *             value = value.time()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/logical.pxi":306
 * 
 *     cdef encode_value(self, value):
 *         if isinstance(value, int) and value >= 0 and value <= (MAX_TIME // 1_000):             # <<<<<<<<<<<<<<
 *             return value
 *         if isinstance(value, datetime.datetime):
 */
  }

  /* "src/logical.pxi":308
 *         if isinstance(value, int) and value >= 0 and value <= (MAX_TIME // 1_000):
 *             return value
 *         if isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *             value = value.time()
 *         return _time_to_micros(value) // 1_000
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_value, __pyx_t_4); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(24, 308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_1) {

    /* "src/logical.pxi":309
 *             return value
 *         if isinstance(value, datetime.datetime):
 *             value = value.time()             # <<<<<<<<<<<<<<
 *         return _time_to_micros(value) // 1_000
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 309, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/logical.pxi":308
 *         if isinstance(value, int) and value >= 0 and value <= (MAX_TIME // 1_000):
 *             return value
 *         if isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *             value = value.time()
 *         return _time_to_micros(value) // 1_000
 */
  }

  /* "src/logical.pxi":310
 *         if isinstance(value, datetime.datetime):
 *             value = value.time()
 *         return _time_to_micros(value) // 1_000             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_time_to_micros); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_value};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyInt_FloorDivideObjC(__pyx_t_4, __pyx_int_1000, 0x3E8, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":305
 *         return 'int(time-millis)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, int) and value >= 0 and value <= (MAX_TIME // 1_000):
 *             return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.TimeMillis.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":312
 *         return _time_to_micros(value) // 1_000
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return _micros_to_time(value * 1_000)
 * 
 */

static PyObject *__pyx_f_5cavro_10TimeMillis_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimeMillis *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/logical.pxi":313
 * 
 *     cdef decode_value(self, value):
 *         return _micros_to_time(value * 1_000)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_micros_to_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_MultiplyObjC(__pyx_v_value, __pyx_int_1000, 0x3E8, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":312
 *         return _time_to_micros(value) // 1_000
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return _micros_to_time(value * 1_000)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.TimeMillis.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10TimeMillis_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10TimeMillis_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10TimeMillis_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10TimeMillis_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10TimeMillis_2__reduce_cython__(((struct __pyx_obj_5cavro_TimeMillis *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10TimeMillis_2__reduce_cython__(struct __pyx_obj_5cavro_TimeMillis *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_TimeMillis, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TimeMillis, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_TimeMillis, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_TimeMillis, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_TimeMillis); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TimeMillis, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_TimeMillis, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_TimeMillis, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TimeMillis__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_TimeMillis); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.TimeMillis.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimeMillis, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimeMillis__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10TimeMillis_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10TimeMillis_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10TimeMillis_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10TimeMillis_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.TimeMillis.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10TimeMillis_4__setstate_cython__(((struct __pyx_obj_5cavro_TimeMillis *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10TimeMillis_4__setstate_cython__(struct __pyx_obj_5cavro_TimeMillis *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_TimeMillis, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TimeMillis__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_TimeMillis__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimeMillis, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimeMillis__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.TimeMillis.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":321
 *     underlying_types = (LongType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10TimeMicros_1_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10TimeMicros_1_for_type = {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10TimeMicros_1_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10TimeMicros_1_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_for_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_underlying,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_underlying)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 321, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_for_type") < 0)) __PYX_ERR(24, 321, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_underlying = ((struct __pyx_obj_5cavro_AvroType *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_for_type", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 321, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.TimeMicros._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_underlying), __pyx_ptype_5cavro_AvroType, 0, "underlying", 0))) __PYX_ERR(24, 322, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_10TimeMicros__for_type(((PyTypeObject*)__pyx_v_cls), __pyx_v_underlying);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10TimeMicros__for_type(PyTypeObject *__pyx_v_cls, CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_type", 1);

  /* "src/logical.pxi":323
 *     @classmethod
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_v_cls)); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":321
 *     underlying_types = (LongType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.TimeMicros._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":325
 *         return cls()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'long(time-micros)'
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10TimeMicros_9type_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_10TimeMicros_9type_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_10TimeMicros_9type_name___get__(((struct __pyx_obj_5cavro_TimeMicros *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10TimeMicros_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_TimeMicros *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/logical.pxi":327
 *     @property
 *     def type_name(self):
 *         return 'long(time-micros)'             # <<<<<<<<<<<<<<
 * 
 *     cdef encode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_kp_u_long_time_micros);
  __pyx_r = __pyx_kp_u_long_time_micros;
  goto __pyx_L0;

  /* "src/logical.pxi":325
 *         return cls()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'long(time-micros)'
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":329
 *         return 'long(time-micros)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, int) and value >= 0 and value <= MAX_TIME:
 *             return value
 */

static PyObject *__pyx_f_5cavro_10TimeMicros_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimeMicros *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "src/logical.pxi":330
 * 
 *     cdef encode_value(self, value):
 *         if isinstance(value, int) and value >= 0 and value <= MAX_TIME:             # <<<<<<<<<<<<<<
 *             return value
 *         if isinstance(value, datetime.datetime):
 */
  __pyx_t_2 = PyInt_Check(__pyx_v_value); 
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_value, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 330, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(24, 330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_MAX_TIME); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_value, __pyx_t_3, Py_LE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(24, 330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "src/logical.pxi":331
 *     cdef encode_value(self, value):
 *         if isinstance(value, int) and value >= 0 and value <= MAX_TIME:
 *             return value             # <<<<<<<<<<<<<<
 *         if isinstance(value, datetime.datetime):
 *             value = value.time()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/logical.pxi":330
 * 
 *     cdef encode_value(self, value):
 *         if isinstance(value, int) and value >= 0 and value <= MAX_TIME:             # <<<<<<<<<<<<<<
 *             return value
 *         if isinstance(value, datetime.datetime):
 */
  }

  /* "src/logical.pxi":332
 *         if isinstance(value, int) and value >= 0 and value <= MAX_TIME:
 *             return value
 *         if isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *             value = value.time()
 *         return _time_to_micros(value)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(24, 332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "src/logical.pxi":333
 *             return value
 *         if isinstance(value, datetime.datetime):
 *             value = value.time()             # <<<<<<<<<<<<<<
 *         return _time_to_micros(value)
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 333, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/logical.pxi":332
 *         if isinstance(value, int) and value >= 0 and value <= MAX_TIME:
 *             return value
 *         if isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *             value = value.time()
 *         return _time_to_micros(value)
 */
  }

  /* "src/logical.pxi":334
 *         if isinstance(value, datetime.datetime):
 *             value = value.time()
 *         return _time_to_micros(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_time_to_micros); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_value};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":329
 *         return 'long(time-micros)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, int) and value >= 0 and value <= MAX_TIME:
 *             return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.TimeMicros.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":336
 *         return _time_to_micros(value)
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return _micros_to_time(value)
 * 
 */

static PyObject *__pyx_f_5cavro_10TimeMicros_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimeMicros *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/logical.pxi":337
 * 
 *     cdef decode_value(self, value):
 *         return _micros_to_time(value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_micros_to_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":336
 *         return _time_to_micros(value)
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return _micros_to_time(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.TimeMicros.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10TimeMicros_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10TimeMicros_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10TimeMicros_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10TimeMicros_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10TimeMicros_2__reduce_cython__(((struct __pyx_obj_5cavro_TimeMicros *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10TimeMicros_2__reduce_cython__(struct __pyx_obj_5cavro_TimeMicros *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_TimeMicros, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TimeMicros, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_TimeMicros, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_TimeMicros, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_TimeMicros); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TimeMicros, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_TimeMicros, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_TimeMicros, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TimeMicros__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_TimeMicros); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.TimeMicros.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimeMicros, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimeMicros__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10TimeMicros_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10TimeMicros_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10TimeMicros_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10TimeMicros_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.TimeMicros.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10TimeMicros_4__setstate_cython__(((struct __pyx_obj_5cavro_TimeMicros *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10TimeMicros_4__setstate_cython__(struct __pyx_obj_5cavro_TimeMicros *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_TimeMicros, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TimeMicros__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_TimeMicros__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimeMicros, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimeMicros__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.TimeMicros.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":347
 *     cdef readonly bint alternate_timestamp_encoding
 * 
 *     def __init__(self, alternate_timestamp_encoding):             # <<<<<<<<<<<<<<
 *         self.alternate_timestamp_encoding = alternate_timestamp_encoding
 * 
 */

/* Python wrapper */
static int __pyx_pw_5cavro_15TimestampMillis_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_15TimestampMillis_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_alternate_timestamp_encoding = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_alternate_timestamp_encoding,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_alternate_timestamp_encoding)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 347, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(24, 347, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_alternate_timestamp_encoding = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 347, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.TimestampMillis.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_15TimestampMillis___init__(((struct __pyx_obj_5cavro_TimestampMillis *)__pyx_v_self), __pyx_v_alternate_timestamp_encoding);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_15TimestampMillis___init__(struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self, PyObject *__pyx_v_alternate_timestamp_encoding) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/logical.pxi":348
 * 
 *     def __init__(self, alternate_timestamp_encoding):
 *         self.alternate_timestamp_encoding = alternate_timestamp_encoding             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_alternate_timestamp_encoding); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(24, 348, __pyx_L1_error)
  __pyx_v_self->alternate_timestamp_encoding = __pyx_t_1;

  /* "src/logical.pxi":347
 *     cdef readonly bint alternate_timestamp_encoding
 * 
 *     def __init__(self, alternate_timestamp_encoding):             # <<<<<<<<<<<<<<
 *         self.alternate_timestamp_encoding = alternate_timestamp_encoding
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.TimestampMillis.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/logical.pxi":350
 *         self.alternate_timestamp_encoding = alternate_timestamp_encoding
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15TimestampMillis_3_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15TimestampMillis_3_for_type = {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15TimestampMillis_3_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15TimestampMillis_3_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_for_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_underlying,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_underlying)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 350, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_for_type") < 0)) __PYX_ERR(24, 350, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_underlying = ((struct __pyx_obj_5cavro_AvroType *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_for_type", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 350, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.TimestampMillis._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_underlying), __pyx_ptype_5cavro_AvroType, 0, "underlying", 0))) __PYX_ERR(24, 351, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_15TimestampMillis_2_for_type(((PyTypeObject*)__pyx_v_cls), __pyx_v_underlying);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15TimestampMillis_2_for_type(PyTypeObject *__pyx_v_cls, struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_type", 1);

  /* "src/logical.pxi":352
 *     @classmethod
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(             # <<<<<<<<<<<<<<
 *             underlying.options.alternate_timestamp_millis_encoding,
 *         )
 */
  __Pyx_XDECREF(__pyx_r);

  /* "src/logical.pxi":353
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(
 *             underlying.options.alternate_timestamp_millis_encoding,             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_underlying->options->alternate_timestamp_millis_encoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "src/logical.pxi":352
 *     @classmethod
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(             # <<<<<<<<<<<<<<
 *             underlying.options.alternate_timestamp_millis_encoding,
 *         )
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_v_cls), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":350
 *         self.alternate_timestamp_encoding = alternate_timestamp_encoding
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.TimestampMillis._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":356
 *         )
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'long(timestamp-millis)'
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15TimestampMillis_9type_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15TimestampMillis_9type_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15TimestampMillis_9type_name___get__(((struct __pyx_obj_5cavro_TimestampMillis *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15TimestampMillis_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/logical.pxi":358
 *     @property
 *     def type_name(self):
 *         return 'long(timestamp-millis)'             # <<<<<<<<<<<<<<
 * 
 *     cdef encode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_kp_u_long_timestamp_millis);
  __pyx_r = __pyx_kp_u_long_timestamp_millis;
  goto __pyx_L0;

  /* "src/logical.pxi":356
 *         )
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'long(timestamp-millis)'
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":360
 *         return 'long(timestamp-millis)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, int):
 *             return value
 */

static PyObject *__pyx_f_5cavro_15TimestampMillis_encode_value(struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_delta = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 1);

  /* "src/logical.pxi":361
 * 
 *     cdef encode_value(self, value):
 *         if isinstance(value, int):             # <<<<<<<<<<<<<<
 *             return value
 *         if not isinstance(value, datetime.datetime):
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_value); 
  if (__pyx_t_1) {

    /* "src/logical.pxi":362
 *     cdef encode_value(self, value):
 *         if isinstance(value, int):
 *             return value             # <<<<<<<<<<<<<<
 *         if not isinstance(value, datetime.datetime):
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/logical.pxi":361
 * 
 *     cdef encode_value(self, value):
 *         if isinstance(value, int):             # <<<<<<<<<<<<<<
 *             return value
 *         if not isinstance(value, datetime.datetime):
 */
  }

  /* "src/logical.pxi":363
 *         if isinstance(value, int):
 *             return value
 *         if not isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")
 *         if self.alternate_timestamp_encoding:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_datetime); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(24, 363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (!__pyx_t_1);
  if (unlikely(__pyx_t_4)) {

    /* "src/logical.pxi":364
 *             return value
 *         if not isinstance(value, datetime.datetime):
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")             # <<<<<<<<<<<<<<
 *         if self.alternate_timestamp_encoding:
 *             delta = dt_to_timedelta(value)
 */
    __pyx_t_3 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_value)), __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Expected_datetime_datetime_got, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(24, 364, __pyx_L1_error)

    /* "src/logical.pxi":363
 *         if isinstance(value, int):
 *             return value
 *         if not isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")
 *         if self.alternate_timestamp_encoding:
 */
  }

  /* "src/logical.pxi":365
 *         if not isinstance(value, datetime.datetime):
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")
 *         if self.alternate_timestamp_encoding:             # <<<<<<<<<<<<<<
 *             delta = dt_to_timedelta(value)
 *             return (delta.microseconds // 1_000) + (delta.seconds + delta.days * 24 * 3600) * 1_000
 */
  if (__pyx_v_self->alternate_timestamp_encoding) {

    /* "src/logical.pxi":366
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")
 *         if self.alternate_timestamp_encoding:
 *             delta = dt_to_timedelta(value)             # <<<<<<<<<<<<<<
 *             return (delta.microseconds // 1_000) + (delta.seconds + delta.days * 24 * 3600) * 1_000
 *         else:
 */
    __pyx_t_3 = __pyx_f_5cavro_dt_to_timedelta(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_delta = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "src/logical.pxi":367
 *         if self.alternate_timestamp_encoding:
 *             delta = dt_to_timedelta(value)
 *             return (delta.microseconds // 1_000) + (delta.seconds + delta.days * 24 * 3600) * 1_000             # <<<<<<<<<<<<<<
 *         else:
 *             return int(value.timestamp() * 1_000)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_delta, __pyx_n_s_microseconds); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyInt_FloorDivideObjC(__pyx_t_3, __pyx_int_1000, 0x3E8, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_delta, __pyx_n_s_seconds); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_delta, __pyx_n_s_days); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_MultiplyObjC(__pyx_t_5, __pyx_int_24, 24, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(24, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyInt_MultiplyObjC(__pyx_t_6, __pyx_int_3600, 0xE10, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyNumber_Add(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(24, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyInt_MultiplyObjC(__pyx_t_6, __pyx_int_1000, 0x3E8, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyNumber_Add(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(24, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "src/logical.pxi":365
 *         if not isinstance(value, datetime.datetime):
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")
 *         if self.alternate_timestamp_encoding:             # <<<<<<<<<<<<<<
 *             delta = dt_to_timedelta(value)
 *             return (delta.microseconds // 1_000) + (delta.seconds + delta.days * 24 * 3600) * 1_000
 */
  }

  /* "src/logical.pxi":369
 *             return (delta.microseconds // 1_000) + (delta.seconds + delta.days * 24 * 3600) * 1_000
 *         else:
 *             return int(value.timestamp() * 1_000)             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_timestamp); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(24, 369, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_5 = __Pyx_PyInt_MultiplyObjC(__pyx_t_6, __pyx_int_1000, 0x3E8, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(24, 369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* "src/logical.pxi":360
 *         return 'long(timestamp-millis)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, int):
 *             return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.TimestampMillis.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_delta);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":371
 *             return int(value.timestamp() * 1_000)
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return EPOCH_DT + datetime.timedelta(microseconds=value * 1_000)
 * 
 */

static PyObject *__pyx_f_5cavro_15TimestampMillis_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/logical.pxi":372
 * 
 *     cdef decode_value(self, value):
 *         return EPOCH_DT + datetime.timedelta(microseconds=value * 1_000)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_EPOCH_DT); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_datetime); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_timedelta); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_MultiplyObjC(__pyx_v_value, __pyx_int_1000, 0x3E8, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_microseconds, __pyx_t_4) < 0) __PYX_ERR(24, 372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":371
 *             return int(value.timestamp() * 1_000)
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return EPOCH_DT + datetime.timedelta(microseconds=value * 1_000)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.TimestampMillis.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":345
 *     underlying_types = (LongType, )
 * 
 *     cdef readonly bint alternate_timestamp_encoding             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, alternate_timestamp_encoding):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15TimestampMillis_28alternate_timestamp_encoding_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15TimestampMillis_28alternate_timestamp_encoding_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15TimestampMillis_28alternate_timestamp_encoding___get__(((struct __pyx_obj_5cavro_TimestampMillis *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15TimestampMillis_28alternate_timestamp_encoding___get__(struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->alternate_timestamp_encoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.TimestampMillis.alternate_timestamp_encoding.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15TimestampMillis_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15TimestampMillis_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15TimestampMillis_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15TimestampMillis_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_15TimestampMillis_4__reduce_cython__(((struct __pyx_obj_5cavro_TimestampMillis *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15TimestampMillis_4__reduce_cython__(struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.alternate_timestamp_encoding,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->alternate_timestamp_encoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.alternate_timestamp_encoding,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.alternate_timestamp_encoding,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.alternate_timestamp_encoding,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_TimestampMillis, (type(self), 0x8d5b257, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TimestampMillis, (type(self), 0x8d5b257, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_TimestampMillis, (type(self), 0x8d5b257, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_TimestampMillis, (type(self), 0x8d5b257, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_TimestampMillis); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_148222551);
    __Pyx_GIVEREF(__pyx_int_148222551);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_148222551)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TimestampMillis, (type(self), 0x8d5b257, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_TimestampMillis, (type(self), 0x8d5b257, None), state
 *     else:
 *         return __pyx_unpickle_TimestampMillis, (type(self), 0x8d5b257, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TimestampMillis__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_TimestampMillis); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_148222551);
    __Pyx_GIVEREF(__pyx_int_148222551);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_148222551)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.TimestampMillis.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimestampMillis, (type(self), 0x8d5b257, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimestampMillis__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15TimestampMillis_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15TimestampMillis_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15TimestampMillis_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15TimestampMillis_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.TimestampMillis.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_15TimestampMillis_6__setstate_cython__(((struct __pyx_obj_5cavro_TimestampMillis *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15TimestampMillis_6__setstate_cython__(struct __pyx_obj_5cavro_TimestampMillis *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_TimestampMillis, (type(self), 0x8d5b257, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TimestampMillis__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_TimestampMillis__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimestampMillis, (type(self), 0x8d5b257, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimestampMillis__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.TimestampMillis.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":380
 *     underlying_types = (LongType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15TimestampMicros_1_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15TimestampMicros_1_for_type = {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15TimestampMicros_1_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15TimestampMicros_1_for_type(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_for_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_underlying,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_underlying)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 380, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_for_type") < 0)) __PYX_ERR(24, 380, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_underlying = ((struct __pyx_obj_5cavro_AvroType *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_for_type", 1, 1, 1, __pyx_nargs); __PYX_ERR(24, 380, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.TimestampMicros._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_underlying), __pyx_ptype_5cavro_AvroType, 0, "underlying", 0))) __PYX_ERR(24, 381, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_15TimestampMicros__for_type(((PyTypeObject*)__pyx_v_cls), __pyx_v_underlying);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15TimestampMicros__for_type(PyTypeObject *__pyx_v_cls, CYTHON_UNUSED struct __pyx_obj_5cavro_AvroType *__pyx_v_underlying) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_for_type", 1);

  /* "src/logical.pxi":382
 *     @classmethod
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_v_cls)); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":380
 *     underlying_types = (LongType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.TimestampMicros._for_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":384
 *         return cls()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'long(timestamp-micros)'
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15TimestampMicros_9type_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15TimestampMicros_9type_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15TimestampMicros_9type_name___get__(((struct __pyx_obj_5cavro_TimestampMicros *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15TimestampMicros_9type_name___get__(CYTHON_UNUSED struct __pyx_obj_5cavro_TimestampMicros *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/logical.pxi":386
 *     @property
 *     def type_name(self):
 *         return 'long(timestamp-micros)'             # <<<<<<<<<<<<<<
 * 
 *     cdef encode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_kp_u_long_timestamp_micros);
  __pyx_r = __pyx_kp_u_long_timestamp_micros;
  goto __pyx_L0;

  /* "src/logical.pxi":384
 *         return cls()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def type_name(self):
 *         return 'long(timestamp-micros)'
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":388
 *         return 'long(timestamp-micros)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, int):
 *             return value
 */

static PyObject *__pyx_f_5cavro_15TimestampMicros_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimestampMicros *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_delta = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 1);

  /* "src/logical.pxi":389
 * 
 *     cdef encode_value(self, value):
 *         if isinstance(value, int):             # <<<<<<<<<<<<<<
 *             return value
 *         if not isinstance(value, datetime.datetime):
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_value); 
  if (__pyx_t_1) {

    /* "src/logical.pxi":390
 *     cdef encode_value(self, value):
 *         if isinstance(value, int):
 *             return value             # <<<<<<<<<<<<<<
 *         if not isinstance(value, datetime.datetime):
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;

    /* "src/logical.pxi":389
 * 
 *     cdef encode_value(self, value):
 *         if isinstance(value, int):             # <<<<<<<<<<<<<<
 *             return value
 *         if not isinstance(value, datetime.datetime):
 */
  }

  /* "src/logical.pxi":391
 *         if isinstance(value, int):
 *             return value
 *         if not isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")
 *         delta = dt_to_timedelta(value)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_datetime); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(24, 391, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (!__pyx_t_1);
  if (unlikely(__pyx_t_4)) {

    /* "src/logical.pxi":392
 *             return value
 *         if not isinstance(value, datetime.datetime):
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")             # <<<<<<<<<<<<<<
 *         delta = dt_to_timedelta(value)
 *         return delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 1_000_000
 */
    __pyx_t_3 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_value)), __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Expected_datetime_datetime_got, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(24, 392, __pyx_L1_error)

    /* "src/logical.pxi":391
 *         if isinstance(value, int):
 *             return value
 *         if not isinstance(value, datetime.datetime):             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")
 *         delta = dt_to_timedelta(value)
 */
  }

  /* "src/logical.pxi":393
 *         if not isinstance(value, datetime.datetime):
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")
 *         delta = dt_to_timedelta(value)             # <<<<<<<<<<<<<<
 *         return delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 1_000_000
 * 
 */
  __pyx_t_3 = __pyx_f_5cavro_dt_to_timedelta(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_delta = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "src/logical.pxi":394
 *             raise ValueError(f"Expected datetime.datetime, got {type(value)}")
 *         delta = dt_to_timedelta(value)
 *         return delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 1_000_000             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_delta, __pyx_n_s_microseconds); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_delta, __pyx_n_s_seconds); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_delta, __pyx_n_s_days); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_MultiplyObjC(__pyx_t_5, __pyx_int_24, 24, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(24, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_MultiplyObjC(__pyx_t_6, __pyx_int_3600, 0xE10, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyNumber_Add(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(24, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_MultiplyObjC(__pyx_t_6, __pyx_int_1000000, 0xF4240, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyNumber_Add(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(24, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":388
 *         return 'long(timestamp-micros)'
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, int):
 *             return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.TimestampMicros.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_delta);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/logical.pxi":396
 *         return delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 1_000_000
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return EPOCH_DT + datetime.timedelta(microseconds=value)
 * 
 */

static PyObject *__pyx_f_5cavro_15TimestampMicros_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_TimestampMicros *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/logical.pxi":397
 * 
 *     cdef decode_value(self, value):
 *         return EPOCH_DT + datetime.timedelta(microseconds=value)             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_EPOCH_DT); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_datetime); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_timedelta); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_microseconds, __pyx_v_value) < 0) __PYX_ERR(24, 397, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/logical.pxi":396
 *         return delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 1_000_000
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return EPOCH_DT + datetime.timedelta(microseconds=value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.TimestampMicros.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15TimestampMicros_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15TimestampMicros_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15TimestampMicros_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15TimestampMicros_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_15TimestampMicros_2__reduce_cython__(((struct __pyx_obj_5cavro_TimestampMicros *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15TimestampMicros_2__reduce_cython__(struct __pyx_obj_5cavro_TimestampMicros *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_TimestampMicros, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TimestampMicros, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_TimestampMicros, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_TimestampMicros, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_TimestampMicros); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TimestampMicros, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_TimestampMicros, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_TimestampMicros, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TimestampMicros__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_TimestampMicros); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.TimestampMicros.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimestampMicros, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimestampMicros__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15TimestampMicros_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15TimestampMicros_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15TimestampMicros_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15TimestampMicros_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.TimestampMicros.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_15TimestampMicros_4__setstate_cython__(((struct __pyx_obj_5cavro_TimestampMicros *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15TimestampMicros_4__setstate_cython__(struct __pyx_obj_5cavro_TimestampMicros *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_TimestampMicros, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TimestampMicros__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_TimestampMicros__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimestampMicros, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimestampMicros__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.TimestampMicros.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":60
 *     name = NotImplemented
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement read_block")
 */

static __Pyx_memviewslice __pyx_f_5cavro_5Codec_read_block(struct __pyx_obj_5cavro_Codec *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Reader *__pyx_v_reader, CYTHON_UNUSED size_t __pyx_v_length) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_block", 1);

  /* "src/codec.pxi":62
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement read_block")             # <<<<<<<<<<<<<<
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement_read_block); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/codec.pxi":61
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement read_block")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(3, 61, __pyx_L1_error)

  /* "src/codec.pxi":60
 *     name = NotImplemented
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement read_block")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("cavro.Codec.read_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":64
 *             f"{type(self).__name__} does not implement read_block")
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement write_block")
 */

static Py_ssize_t __pyx_f_5cavro_5Codec_write_block(struct __pyx_obj_5cavro_Codec *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_5cavro__Writer *__pyx_v_writer, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_data) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_block", 1);

  /* "src/codec.pxi":66
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement write_block")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u_does_not_implement_write_block); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/codec.pxi":65
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         raise NotImplementedError(             # <<<<<<<<<<<<<<
 *             f"{type(self).__name__} does not implement write_block")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(3, 65, __pyx_L1_error)

  /* "src/codec.pxi":64
 *             f"{type(self).__name__} does not implement read_block")
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement write_block")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.Codec.write_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_5Codec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_5Codec_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Codec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_5Codec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_5Codec___reduce_cython__(((struct __pyx_obj_5cavro_Codec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_5Codec___reduce_cython__(struct __pyx_obj_5cavro_Codec *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Codec, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Codec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_Codec, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Codec, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Codec); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Codec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Codec, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_Codec, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Codec__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Codec); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.Codec.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Codec__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_5Codec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_5Codec_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Codec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_5Codec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.Codec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_5Codec_2__setstate_cython__(((struct __pyx_obj_5cavro_Codec *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_5Codec_2__setstate_cython__(struct __pyx_obj_5cavro_Codec *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Codec__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_Codec__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Codec__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.Codec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":77
 *     name = 'snappy'
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         cdef const uint8_t[:] compressed = reader.read_n(length - 4)
 *         cdef const uint8_t[:] checksum = reader.read_n(4)
 */

static __Pyx_memviewslice __pyx_f_5cavro_12_SnappyCodec_read_block(CYTHON_UNUSED struct __pyx_obj_5cavro__SnappyCodec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length) {
  __Pyx_memviewslice __pyx_v_compressed = { 0, 0, { 0 }, { 0 }, { 0 } };
  CYTHON_UNUSED __Pyx_memviewslice __pyx_v_checksum = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_decompressed = 0;
  __Pyx_memviewslice __pyx_v_view = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_block", 1);

  /* "src/codec.pxi":78
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 *         cdef const uint8_t[:] compressed = reader.read_n(length - 4)             # <<<<<<<<<<<<<<
 *         cdef const uint8_t[:] checksum = reader.read_n(4)
 *         cdef bytes decompressed = snappy.decompress(compressed)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_reader->__pyx_vtab)->read_n(__pyx_v_reader, (__pyx_v_length - 4)); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(3, 78, __pyx_L1_error)
  __pyx_v_compressed = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;

  /* "src/codec.pxi":79
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 *         cdef const uint8_t[:] compressed = reader.read_n(length - 4)
 *         cdef const uint8_t[:] checksum = reader.read_n(4)             # <<<<<<<<<<<<<<
 *         cdef bytes decompressed = snappy.decompress(compressed)
 *         cdef const uint8_t[:] view = decompressed
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_reader->__pyx_vtab)->read_n(__pyx_v_reader, 4); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(3, 79, __pyx_L1_error)
  __pyx_v_checksum = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;

  /* "src/codec.pxi":80
 *         cdef const uint8_t[:] compressed = reader.read_n(length - 4)
 *         cdef const uint8_t[:] checksum = reader.read_n(4)
 *         cdef bytes decompressed = snappy.decompress(compressed)             # <<<<<<<<<<<<<<
 *         cdef const uint8_t[:] view = decompressed
 *         return view
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_snappy); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_decompress); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_compressed, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_2))) __PYX_ERR(3, 80, __pyx_L1_error)
  __pyx_v_decompressed = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/codec.pxi":81
 *         cdef const uint8_t[:] checksum = reader.read_n(4)
 *         cdef bytes decompressed = snappy.decompress(compressed)
 *         cdef const uint8_t[:] view = decompressed             # <<<<<<<<<<<<<<
 *         return view
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_decompressed, 0); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(3, 81, __pyx_L1_error)
  __pyx_v_view = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "src/codec.pxi":82
 *         cdef bytes decompressed = snappy.decompress(compressed)
 *         cdef const uint8_t[:] view = decompressed
 *         return view             # <<<<<<<<<<<<<<
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 */
  __PYX_INC_MEMVIEW(&__pyx_v_view, 1);
  __pyx_r = __pyx_v_view;
  goto __pyx_L0;

  /* "src/codec.pxi":77
 *     name = 'snappy'
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         cdef const uint8_t[:] compressed = reader.read_n(length - 4)
 *         cdef const uint8_t[:] checksum = reader.read_n(4)
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_7, 1);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("cavro._SnappyCodec.read_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __PYX_XCLEAR_MEMVIEW(&__pyx_v_compressed, 1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_v_checksum, 1);
  __Pyx_XDECREF(__pyx_v_decompressed);
  __PYX_XCLEAR_MEMVIEW(&__pyx_v_view, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":84
 *         return view
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef bytes compressed = snappy.compress(data)
 *         cdef uint32_t crc = crc32(compressed)
 */

static Py_ssize_t __pyx_f_5cavro_12_SnappyCodec_write_block(CYTHON_UNUSED struct __pyx_obj_5cavro__SnappyCodec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data) {
  PyObject *__pyx_v_compressed = 0;
  uint32_t __pyx_v_crc;
  uint8_t *__pyx_v_crc_ptr;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  uint32_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_block", 1);

  /* "src/codec.pxi":85
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = snappy.compress(data)             # <<<<<<<<<<<<<<
 *         cdef uint32_t crc = crc32(compressed)
 *         cdef uint8_t *crc_ptr = <uint8_t *>&crc
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_snappy); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compress); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(3, 85, __pyx_L1_error)
  __pyx_v_compressed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/codec.pxi":86
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = snappy.compress(data)
 *         cdef uint32_t crc = crc32(compressed)             # <<<<<<<<<<<<<<
 *         cdef uint8_t *crc_ptr = <uint8_t *>&crc
 *         writer.write_n(compressed)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_crc32); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_compressed};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_6 = __Pyx_PyInt_As_uint32_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(3, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_crc = __pyx_t_6;

  /* "src/codec.pxi":87
 *         cdef bytes compressed = snappy.compress(data)
 *         cdef uint32_t crc = crc32(compressed)
 *         cdef uint8_t *crc_ptr = <uint8_t *>&crc             # <<<<<<<<<<<<<<
 *         writer.write_n(compressed)
 *         writer.write_n(crc_ptr[:4])
 */
  __pyx_v_crc_ptr = ((uint8_t *)(&__pyx_v_crc));

  /* "src/codec.pxi":88
 *         cdef uint32_t crc = crc32(compressed)
 *         cdef uint8_t *crc_ptr = <uint8_t *>&crc
 *         writer.write_n(compressed)             # <<<<<<<<<<<<<<
 *         writer.write_n(crc_ptr[:4])
 *         return len(compressed) + 4
 */
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_compressed, 0); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(3, 88, __pyx_L1_error)
  __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_writer->__pyx_vtab)->write_n(__pyx_v_writer, __pyx_t_7); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(3, 88, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_7, 1);
  __pyx_t_7.memview = NULL; __pyx_t_7.data = NULL;

  /* "src/codec.pxi":89
 *         cdef uint8_t *crc_ptr = <uint8_t *>&crc
 *         writer.write_n(compressed)
 *         writer.write_n(crc_ptr[:4])             # <<<<<<<<<<<<<<
 *         return len(compressed) + 4
 * 
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_crc_ptr) + 0, 4 - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_1, 0); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(3, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_writer->__pyx_vtab)->write_n(__pyx_v_writer, __pyx_t_7); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(3, 89, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_7, 1);
  __pyx_t_7.memview = NULL; __pyx_t_7.data = NULL;

  /* "src/codec.pxi":90
 *         writer.write_n(compressed)
 *         writer.write_n(crc_ptr[:4])
 *         return len(compressed) + 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_compressed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 90, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyBytes_GET_SIZE(__pyx_v_compressed); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(3, 90, __pyx_L1_error)
  __pyx_r = (__pyx_t_8 + 4);
  goto __pyx_L0;

  /* "src/codec.pxi":84
 *         return view
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef bytes compressed = snappy.compress(data)
 *         cdef uint32_t crc = crc32(compressed)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_AddTraceback("cavro._SnappyCodec.write_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_compressed);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12_SnappyCodec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_12_SnappyCodec_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12_SnappyCodec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_12_SnappyCodec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_12_SnappyCodec___reduce_cython__(((struct __pyx_obj_5cavro__SnappyCodec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12_SnappyCodec___reduce_cython__(struct __pyx_obj_5cavro__SnappyCodec *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__SnappyCodec, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__SnappyCodec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle__SnappyCodec, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__SnappyCodec, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle__SnappyCodec); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__SnappyCodec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__SnappyCodec, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle__SnappyCodec, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__SnappyCodec__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle__SnappyCodec); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._SnappyCodec.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__SnappyCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__SnappyCodec__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12_SnappyCodec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_12_SnappyCodec_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12_SnappyCodec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_12_SnappyCodec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._SnappyCodec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_12_SnappyCodec_2__setstate_cython__(((struct __pyx_obj_5cavro__SnappyCodec *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12_SnappyCodec_2__setstate_cython__(struct __pyx_obj_5cavro__SnappyCodec *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__SnappyCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__SnappyCodec__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__SnappyCodec__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__SnappyCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__SnappyCodec__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._SnappyCodec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":99
 *     name = 'null'
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         return reader.read_n(length)
 * 
 */

static __Pyx_memviewslice __pyx_f_5cavro_10_NullCodec_read_block(CYTHON_UNUSED struct __pyx_obj_5cavro__NullCodec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/codec.pxi":100
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 *         return reader.read_n(length)             # <<<<<<<<<<<<<<
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_reader->__pyx_vtab)->read_n(__pyx_v_reader, __pyx_v_length); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(3, 100, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;
  goto __pyx_L0;

  /* "src/codec.pxi":99
 *     name = 'null'
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         return reader.read_n(length)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_1, 1);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("cavro._NullCodec.read_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  return __pyx_r;
}

/* "src/codec.pxi":102
 *         return reader.read_n(length)
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         writer.write_n(data)
 *         return data.shape[0]
 */

static Py_ssize_t __pyx_f_5cavro_10_NullCodec_write_block(CYTHON_UNUSED struct __pyx_obj_5cavro__NullCodec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data) {
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/codec.pxi":103
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         writer.write_n(data)             # <<<<<<<<<<<<<<
 *         return data.shape[0]
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_writer->__pyx_vtab)->write_n(__pyx_v_writer, __pyx_v_data); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 103, __pyx_L1_error)

  /* "src/codec.pxi":104
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         writer.write_n(data)
 *         return data.shape[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_data.shape[0]);
  goto __pyx_L0;

  /* "src/codec.pxi":102
 *         return reader.read_n(length)
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         writer.write_n(data)
 *         return data.shape[0]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro._NullCodec.write_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10_NullCodec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10_NullCodec_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_NullCodec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10_NullCodec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10_NullCodec___reduce_cython__(((struct __pyx_obj_5cavro__NullCodec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10_NullCodec___reduce_cython__(struct __pyx_obj_5cavro__NullCodec *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__NullCodec, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__NullCodec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle__NullCodec, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__NullCodec, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle__NullCodec); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__NullCodec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__NullCodec, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle__NullCodec, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__NullCodec__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle__NullCodec); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._NullCodec.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__NullCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__NullCodec__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10_NullCodec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10_NullCodec_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_NullCodec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10_NullCodec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._NullCodec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10_NullCodec_2__setstate_cython__(((struct __pyx_obj_5cavro__NullCodec *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10_NullCodec_2__setstate_cython__(struct __pyx_obj_5cavro__NullCodec *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__NullCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__NullCodec__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__NullCodec__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__NullCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__NullCodec__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._NullCodec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":113
 *     name = 'deflate'
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         return zlib.decompress(reader.read_n(length), wbits=-15)
 * 
 */

static __Pyx_memviewslice __pyx_f_5cavro_13_DeflateCodec_read_block(CYTHON_UNUSED struct __pyx_obj_5cavro__DeflateCodec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_block", 1);

  /* "src/codec.pxi":114
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 *         return zlib.decompress(reader.read_n(length), wbits=-15)             # <<<<<<<<<<<<<<
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_zlib); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_decompress); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_reader->__pyx_vtab)->read_n(__pyx_v_reader, __pyx_v_length); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(3, 114, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_t_3, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_3, 1);
  __pyx_t_3.memview = NULL; __pyx_t_3.data = NULL;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(3, 114, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_wbits, __pyx_int_neg_15) < 0) __PYX_ERR(3, 114, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_5, 0); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(3, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;
  goto __pyx_L0;

  /* "src/codec.pxi":113
 *     name = 'deflate'
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         return zlib.decompress(reader.read_n(length), wbits=-15)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("cavro._DeflateCodec.read_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":116
 *         return zlib.decompress(reader.read_n(length), wbits=-15)
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef bytes compressed = zlib.compress(data)[2:-1]
 *         writer.write_n(compressed)
 */

static Py_ssize_t __pyx_f_5cavro_13_DeflateCodec_write_block(CYTHON_UNUSED struct __pyx_obj_5cavro__DeflateCodec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data) {
  PyObject *__pyx_v_compressed = 0;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_block", 1);

  /* "src/codec.pxi":117
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = zlib.compress(data)[2:-1]             # <<<<<<<<<<<<<<
 *         writer.write_n(compressed)
 *         return len(compressed)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_zlib); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compress); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, 2, -1L, NULL, NULL, &__pyx_slice__86, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(3, 117, __pyx_L1_error)
  __pyx_v_compressed = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/codec.pxi":118
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = zlib.compress(data)[2:-1]
 *         writer.write_n(compressed)             # <<<<<<<<<<<<<<
 *         return len(compressed)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_compressed, 0); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(3, 118, __pyx_L1_error)
  __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_writer->__pyx_vtab)->write_n(__pyx_v_writer, __pyx_t_6); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(3, 118, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __pyx_t_6.memview = NULL; __pyx_t_6.data = NULL;

  /* "src/codec.pxi":119
 *         cdef bytes compressed = zlib.compress(data)[2:-1]
 *         writer.write_n(compressed)
 *         return len(compressed)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_compressed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 119, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_GET_SIZE(__pyx_v_compressed); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(3, 119, __pyx_L1_error)
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "src/codec.pxi":116
 *         return zlib.decompress(reader.read_n(length), wbits=-15)
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef bytes compressed = zlib.compress(data)[2:-1]
 *         writer.write_n(compressed)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("cavro._DeflateCodec.write_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_compressed);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_13_DeflateCodec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_13_DeflateCodec_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13_DeflateCodec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_13_DeflateCodec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_13_DeflateCodec___reduce_cython__(((struct __pyx_obj_5cavro__DeflateCodec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_13_DeflateCodec___reduce_cython__(struct __pyx_obj_5cavro__DeflateCodec *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__DeflateCodec, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__DeflateCodec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle__DeflateCodec, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__DeflateCodec, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle__DeflateCodec); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__DeflateCodec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__DeflateCodec, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle__DeflateCodec, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__DeflateCodec__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle__DeflateCodec); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._DeflateCodec.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__DeflateCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__DeflateCodec__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_13_DeflateCodec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_13_DeflateCodec_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13_DeflateCodec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_13_DeflateCodec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._DeflateCodec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_13_DeflateCodec_2__setstate_cython__(((struct __pyx_obj_5cavro__DeflateCodec *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_13_DeflateCodec_2__setstate_cython__(struct __pyx_obj_5cavro__DeflateCodec *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__DeflateCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__DeflateCodec__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__DeflateCodec__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__DeflateCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__DeflateCodec__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._DeflateCodec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":128
 *     name = 'bzip2'
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         return bz2.decompress(reader.read_n(length))
 * 
 */

static __Pyx_memviewslice __pyx_f_5cavro_12__Bzip2Codec_read_block(CYTHON_UNUSED struct __pyx_obj_5cavro___Bzip2Codec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_block", 1);

  /* "src/codec.pxi":129
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 *         return bz2.decompress(reader.read_n(length))             # <<<<<<<<<<<<<<
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_bz2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_decompress); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_reader->__pyx_vtab)->read_n(__pyx_v_reader, __pyx_v_length); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(3, 129, __pyx_L1_error)
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_t_4, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_4, 1);
  __pyx_t_4.memview = NULL; __pyx_t_4.data = NULL;
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_1, 0); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(3, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;
  goto __pyx_L0;

  /* "src/codec.pxi":128
 *     name = 'bzip2'
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         return bz2.decompress(reader.read_n(length))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_7, 1);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("cavro.__Bzip2Codec.read_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":131
 *         return bz2.decompress(reader.read_n(length))
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef bytes compressed = bz2.compress(data)
 *         writer.write_n(compressed)
 */

static Py_ssize_t __pyx_f_5cavro_12__Bzip2Codec_write_block(CYTHON_UNUSED struct __pyx_obj_5cavro___Bzip2Codec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data) {
  PyObject *__pyx_v_compressed = 0;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_block", 1);

  /* "src/codec.pxi":132
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = bz2.compress(data)             # <<<<<<<<<<<<<<
 *         writer.write_n(compressed)
 *         return len(compressed)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_bz2); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compress); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(3, 132, __pyx_L1_error)
  __pyx_v_compressed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/codec.pxi":133
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = bz2.compress(data)
 *         writer.write_n(compressed)             # <<<<<<<<<<<<<<
 *         return len(compressed)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_compressed, 0); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(3, 133, __pyx_L1_error)
  __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_writer->__pyx_vtab)->write_n(__pyx_v_writer, __pyx_t_6); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(3, 133, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __pyx_t_6.memview = NULL; __pyx_t_6.data = NULL;

  /* "src/codec.pxi":134
 *         cdef bytes compressed = bz2.compress(data)
 *         writer.write_n(compressed)
 *         return len(compressed)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_compressed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 134, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_GET_SIZE(__pyx_v_compressed); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(3, 134, __pyx_L1_error)
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "src/codec.pxi":131
 *         return bz2.decompress(reader.read_n(length))
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef bytes compressed = bz2.compress(data)
 *         writer.write_n(compressed)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("cavro.__Bzip2Codec.write_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_compressed);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12__Bzip2Codec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_12__Bzip2Codec_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12__Bzip2Codec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_12__Bzip2Codec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_12__Bzip2Codec___reduce_cython__(((struct __pyx_obj_5cavro___Bzip2Codec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12__Bzip2Codec___reduce_cython__(struct __pyx_obj_5cavro___Bzip2Codec *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle___Bzip2Codec, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___Bzip2Codec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle___Bzip2Codec, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle___Bzip2Codec, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle___Bzip2Codec); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___Bzip2Codec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle___Bzip2Codec, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle___Bzip2Codec, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___Bzip2Codec__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle___Bzip2Codec); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__Bzip2Codec.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Bzip2Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Bzip2Codec__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_12__Bzip2Codec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_12__Bzip2Codec_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12__Bzip2Codec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_12__Bzip2Codec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__Bzip2Codec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_12__Bzip2Codec_2__setstate_cython__(((struct __pyx_obj_5cavro___Bzip2Codec *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_12__Bzip2Codec_2__setstate_cython__(struct __pyx_obj_5cavro___Bzip2Codec *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle___Bzip2Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___Bzip2Codec__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle___Bzip2Codec__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Bzip2Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Bzip2Codec__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.__Bzip2Codec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":143
 *     name = 'xz'
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         return lzma.decompress(reader.read_n(length))
 * 
 */

static __Pyx_memviewslice __pyx_f_5cavro_10_LzmaCodec_read_block(CYTHON_UNUSED struct __pyx_obj_5cavro__LzmaCodec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_block", 1);

  /* "src/codec.pxi":144
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 *         return lzma.decompress(reader.read_n(length))             # <<<<<<<<<<<<<<
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_lzma); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_decompress); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_reader->__pyx_vtab)->read_n(__pyx_v_reader, __pyx_v_length); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(3, 144, __pyx_L1_error)
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_t_4, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_4, 1);
  __pyx_t_4.memview = NULL; __pyx_t_4.data = NULL;
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_1, 0); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(3, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;
  goto __pyx_L0;

  /* "src/codec.pxi":143
 *     name = 'xz'
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         return lzma.decompress(reader.read_n(length))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_7, 1);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("cavro._LzmaCodec.read_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":146
 *         return lzma.decompress(reader.read_n(length))
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef bytes compressed = lzma.compress(data)
 *         writer.write_n(compressed)
 */

static Py_ssize_t __pyx_f_5cavro_10_LzmaCodec_write_block(CYTHON_UNUSED struct __pyx_obj_5cavro__LzmaCodec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data) {
  PyObject *__pyx_v_compressed = 0;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_block", 1);

  /* "src/codec.pxi":147
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = lzma.compress(data)             # <<<<<<<<<<<<<<
 *         writer.write_n(compressed)
 *         return len(compressed)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_lzma); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compress); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(3, 147, __pyx_L1_error)
  __pyx_v_compressed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/codec.pxi":148
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = lzma.compress(data)
 *         writer.write_n(compressed)             # <<<<<<<<<<<<<<
 *         return len(compressed)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_compressed, 0); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(3, 148, __pyx_L1_error)
  __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_writer->__pyx_vtab)->write_n(__pyx_v_writer, __pyx_t_6); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(3, 148, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __pyx_t_6.memview = NULL; __pyx_t_6.data = NULL;

  /* "src/codec.pxi":149
 *         cdef bytes compressed = lzma.compress(data)
 *         writer.write_n(compressed)
 *         return len(compressed)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_compressed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 149, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_GET_SIZE(__pyx_v_compressed); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(3, 149, __pyx_L1_error)
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "src/codec.pxi":146
 *         return lzma.decompress(reader.read_n(length))
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef bytes compressed = lzma.compress(data)
 *         writer.write_n(compressed)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("cavro._LzmaCodec.write_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_compressed);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10_LzmaCodec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10_LzmaCodec_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_LzmaCodec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10_LzmaCodec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_10_LzmaCodec___reduce_cython__(((struct __pyx_obj_5cavro__LzmaCodec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10_LzmaCodec___reduce_cython__(struct __pyx_obj_5cavro__LzmaCodec *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__LzmaCodec, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__LzmaCodec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle__LzmaCodec, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__LzmaCodec, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle__LzmaCodec); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__LzmaCodec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__LzmaCodec, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle__LzmaCodec, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__LzmaCodec__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle__LzmaCodec); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._LzmaCodec.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__LzmaCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__LzmaCodec__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_10_LzmaCodec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_10_LzmaCodec_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_LzmaCodec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_10_LzmaCodec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._LzmaCodec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_10_LzmaCodec_2__setstate_cython__(((struct __pyx_obj_5cavro__LzmaCodec *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_10_LzmaCodec_2__setstate_cython__(struct __pyx_obj_5cavro__LzmaCodec *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__LzmaCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__LzmaCodec__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__LzmaCodec__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__LzmaCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__LzmaCodec__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._LzmaCodec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":161
 *     cdef readonly object decompressor
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.compressor = zstandard.ZstdCompressor()
 *         self.decompressor = zstandard.ZstdDecompressor()
 */

/* Python wrapper */
static int __pyx_pw_5cavro_15_ZStandardCodec_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_15_ZStandardCodec_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_5cavro_15_ZStandardCodec___init__(((struct __pyx_obj_5cavro__ZStandardCodec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_15_ZStandardCodec___init__(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/codec.pxi":162
 * 
 *     def __init__(self):
 *         self.compressor = zstandard.ZstdCompressor()             # <<<<<<<<<<<<<<
 *         self.decompressor = zstandard.ZstdDecompressor()
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_zstandard); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ZstdCompressor); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->compressor);
  __Pyx_DECREF(__pyx_v_self->compressor);
  __pyx_v_self->compressor = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/codec.pxi":163
 *     def __init__(self):
 *         self.compressor = zstandard.ZstdCompressor()
 *         self.decompressor = zstandard.ZstdDecompressor()             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_zstandard); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ZstdDecompressor); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->decompressor);
  __Pyx_DECREF(__pyx_v_self->decompressor);
  __pyx_v_self->decompressor = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/codec.pxi":161
 *     cdef readonly object decompressor
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.compressor = zstandard.ZstdCompressor()
 *         self.decompressor = zstandard.ZstdDecompressor()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro._ZStandardCodec.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":165
 *         self.decompressor = zstandard.ZstdDecompressor()
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         decompress_obj = self.decompressor.decompressobj()
 *         return decompress_obj.decompress(reader.read_n(length))
 */

static __Pyx_memviewslice __pyx_f_5cavro_15_ZStandardCodec_read_block(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length) {
  PyObject *__pyx_v_decompress_obj = NULL;
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_block", 1);

  /* "src/codec.pxi":166
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 *         decompress_obj = self.decompressor.decompressobj()             # <<<<<<<<<<<<<<
 *         return decompress_obj.decompress(reader.read_n(length))
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->decompressor, __pyx_n_s_decompressobj); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_decompress_obj = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/codec.pxi":167
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 *         decompress_obj = self.decompressor.decompressobj()
 *         return decompress_obj.decompress(reader.read_n(length))             # <<<<<<<<<<<<<<
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_decompress_obj, __pyx_n_s_decompress); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_reader->__pyx_vtab)->read_n(__pyx_v_reader, __pyx_v_length); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(3, 167, __pyx_L1_error)
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_t_5, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_5, 1);
  __pyx_t_5.memview = NULL; __pyx_t_5.data = NULL;
  __pyx_t_6 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_1, 0); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(3, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;
  goto __pyx_L0;

  /* "src/codec.pxi":165
 *         self.decompressor = zstandard.ZstdDecompressor()
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         decompress_obj = self.decompressor.decompressobj()
 *         return decompress_obj.decompress(reader.read_n(length))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_7, 1);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("cavro._ZStandardCodec.read_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_decompress_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":169
 *         return decompress_obj.decompress(reader.read_n(length))
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef bytes compressed = self.compressor.compress(data)
 *         writer.write_n(compressed)
 */

static Py_ssize_t __pyx_f_5cavro_15_ZStandardCodec_write_block(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data) {
  PyObject *__pyx_v_compressed = 0;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_block", 1);

  /* "src/codec.pxi":170
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = self.compressor.compress(data)             # <<<<<<<<<<<<<<
 *         writer.write_n(compressed)
 *         return len(compressed)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->compressor, __pyx_n_s_compress); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(3, 170, __pyx_L1_error)
  __pyx_v_compressed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/codec.pxi":171
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = self.compressor.compress(data)
 *         writer.write_n(compressed)             # <<<<<<<<<<<<<<
 *         return len(compressed)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_compressed, 0); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(3, 171, __pyx_L1_error)
  __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_writer->__pyx_vtab)->write_n(__pyx_v_writer, __pyx_t_6); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(3, 171, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __pyx_t_6.memview = NULL; __pyx_t_6.data = NULL;

  /* "src/codec.pxi":172
 *         cdef bytes compressed = self.compressor.compress(data)
 *         writer.write_n(compressed)
 *         return len(compressed)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_compressed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 172, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_GET_SIZE(__pyx_v_compressed); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(3, 172, __pyx_L1_error)
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "src/codec.pxi":169
 *         return decompress_obj.decompress(reader.read_n(length))
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef bytes compressed = self.compressor.compress(data)
 *         writer.write_n(compressed)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("cavro._ZStandardCodec.write_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_compressed);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":158
 *     name = 'zstandard'
 * 
 *     cdef readonly object compressor             # <<<<<<<<<<<<<<
 *     cdef readonly object decompressor
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15_ZStandardCodec_10compressor_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15_ZStandardCodec_10compressor_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15_ZStandardCodec_10compressor___get__(((struct __pyx_obj_5cavro__ZStandardCodec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15_ZStandardCodec_10compressor___get__(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->compressor);
  __pyx_r = __pyx_v_self->compressor;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":159
 * 
 *     cdef readonly object compressor
 *     cdef readonly object decompressor             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15_ZStandardCodec_12decompressor_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15_ZStandardCodec_12decompressor_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15_ZStandardCodec_12decompressor___get__(((struct __pyx_obj_5cavro__ZStandardCodec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15_ZStandardCodec_12decompressor___get__(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->decompressor);
  __pyx_r = __pyx_v_self->decompressor;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15_ZStandardCodec_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15_ZStandardCodec_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15_ZStandardCodec_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15_ZStandardCodec_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_15_ZStandardCodec_2__reduce_cython__(((struct __pyx_obj_5cavro__ZStandardCodec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15_ZStandardCodec_2__reduce_cython__(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.compressor, self.decompressor)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->compressor);
  __Pyx_GIVEREF(__pyx_v_self->compressor);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->compressor)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->decompressor);
  __Pyx_GIVEREF(__pyx_v_self->decompressor);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->decompressor)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.compressor, self.decompressor)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.compressor, self.decompressor)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.compressor is not None or self.decompressor is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.compressor, self.decompressor)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.compressor is not None or self.decompressor is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__ZStandardCodec, (type(self), 0x8d57cb2, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->compressor != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->decompressor != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.compressor is not None or self.decompressor is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__ZStandardCodec, (type(self), 0x8d57cb2, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.compressor is not None or self.decompressor is not None
 *     if use_setstate:
 *         return __pyx_unpickle__ZStandardCodec, (type(self), 0x8d57cb2, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__ZStandardCodec, (type(self), 0x8d57cb2, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle__ZStandardCodec); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_148208818);
    __Pyx_GIVEREF(__pyx_int_148208818);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_148208818)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.compressor is not None or self.decompressor is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__ZStandardCodec, (type(self), 0x8d57cb2, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__ZStandardCodec, (type(self), 0x8d57cb2, None), state
 *     else:
 *         return __pyx_unpickle__ZStandardCodec, (type(self), 0x8d57cb2, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__ZStandardCodec__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle__ZStandardCodec); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_148208818);
    __Pyx_GIVEREF(__pyx_int_148208818);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_148208818)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro._ZStandardCodec.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__ZStandardCodec, (type(self), 0x8d57cb2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__ZStandardCodec__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15_ZStandardCodec_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15_ZStandardCodec_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15_ZStandardCodec_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15_ZStandardCodec_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._ZStandardCodec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_15_ZStandardCodec_4__setstate_cython__(((struct __pyx_obj_5cavro__ZStandardCodec *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15_ZStandardCodec_4__setstate_cython__(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__ZStandardCodec, (type(self), 0x8d57cb2, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__ZStandardCodec__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__ZStandardCodec__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__ZStandardCodec, (type(self), 0x8d57cb2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__ZStandardCodec__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._ZStandardCodec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":181
 *     name = 'lz4'
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         return lz4.frame.decompress(reader.read_n(length))
 * 
 */

static __Pyx_memviewslice __pyx_f_5cavro_9_Lz4Codec_read_block(CYTHON_UNUSED struct __pyx_obj_5cavro__Lz4Codec *__pyx_v_self, struct __pyx_obj_5cavro__Reader *__pyx_v_reader, size_t __pyx_v_length) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_block", 1);

  /* "src/codec.pxi":182
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 *         return lz4.frame.decompress(reader.read_n(length))             # <<<<<<<<<<<<<<
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_lz4); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_frame); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_decompress); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_reader->__pyx_vtab)->read_n(__pyx_v_reader, __pyx_v_length); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(3, 182, __pyx_L1_error)
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_t_4, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_4, 1);
  __pyx_t_4.memview = NULL; __pyx_t_4.data = NULL;
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_1, 0); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(3, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;
  goto __pyx_L0;

  /* "src/codec.pxi":181
 *     name = 'lz4'
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):             # <<<<<<<<<<<<<<
 *         return lz4.frame.decompress(reader.read_n(length))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_7, 1);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("cavro._Lz4Codec.read_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/codec.pxi":184
 *         return lz4.frame.decompress(reader.read_n(length))
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef bytes compressed = lz4.frame.compress(data)
 *         writer.write_n(compressed)
 */

static Py_ssize_t __pyx_f_5cavro_9_Lz4Codec_write_block(CYTHON_UNUSED struct __pyx_obj_5cavro__Lz4Codec *__pyx_v_self, struct __pyx_obj_5cavro__Writer *__pyx_v_writer, __Pyx_memviewslice __pyx_v_data) {
  PyObject *__pyx_v_compressed = 0;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_block", 1);

  /* "src/codec.pxi":185
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = lz4.frame.compress(data)             # <<<<<<<<<<<<<<
 *         writer.write_n(compressed)
 *         return len(compressed)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_lz4); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_frame); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_compress); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_data, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(3, 185, __pyx_L1_error)
  __pyx_v_compressed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/codec.pxi":186
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = lz4.frame.compress(data)
 *         writer.write_n(compressed)             # <<<<<<<<<<<<<<
 *         return len(compressed)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_compressed, 0); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(3, 186, __pyx_L1_error)
  __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_writer->__pyx_vtab)->write_n(__pyx_v_writer, __pyx_t_6); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(3, 186, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __pyx_t_6.memview = NULL; __pyx_t_6.data = NULL;

  /* "src/codec.pxi":187
 *         cdef bytes compressed = lz4.frame.compress(data)
 *         writer.write_n(compressed)
 *         return len(compressed)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_compressed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(3, 187, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_GET_SIZE(__pyx_v_compressed); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(3, 187, __pyx_L1_error)
  __pyx_r = __pyx_t_7;
  goto __pyx_L0;

  /* "src/codec.pxi":184
 *         return lz4.frame.decompress(reader.read_n(length))
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:             # <<<<<<<<<<<<<<
 *         cdef bytes compressed = lz4.frame.compress(data)
 *         writer.write_n(compressed)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("cavro._Lz4Codec.write_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_compressed);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9_Lz4Codec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9_Lz4Codec_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9_Lz4Codec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9_Lz4Codec_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_9_Lz4Codec___reduce_cython__(((struct __pyx_obj_5cavro__Lz4Codec *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9_Lz4Codec___reduce_cython__(struct __pyx_obj_5cavro__Lz4Codec *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__Lz4Codec, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__Lz4Codec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle__Lz4Codec, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__Lz4Codec, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle__Lz4Codec); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__Lz4Codec, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__Lz4Codec, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle__Lz4Codec, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__Lz4Codec__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle__Lz4Codec); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro._Lz4Codec.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Lz4Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Lz4Codec__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_9_Lz4Codec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_9_Lz4Codec_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9_Lz4Codec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_9_Lz4Codec_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro._Lz4Codec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_9_Lz4Codec_2__setstate_cython__(((struct __pyx_obj_5cavro__Lz4Codec *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_9_Lz4Codec_2__setstate_cython__(struct __pyx_obj_5cavro__Lz4Codec *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__Lz4Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__Lz4Codec__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__Lz4Codec__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Lz4Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Lz4Codec__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro._Lz4Codec.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":12
 * 
 * 
 * cdef int viewcmp(const uint8_t[:] a, const uint8_t[:] b):             # <<<<<<<<<<<<<<
 *     if len(a) != len(b):
 *         return 1
 */

static int __pyx_f_5cavro_viewcmp(__Pyx_memviewslice __pyx_v_a, __Pyx_memviewslice __pyx_v_b) {
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "src/container.pxi":13
 * 
 * cdef int viewcmp(const uint8_t[:] a, const uint8_t[:] b):
 *     if len(a) != len(b):             # <<<<<<<<<<<<<<
 *         return 1
 *     return memcmp(&a[0], &b[0], len(a))
 */
  __pyx_t_1 = __Pyx_MemoryView_Len(__pyx_v_a); 
  __pyx_t_2 = __Pyx_MemoryView_Len(__pyx_v_b); 
  __pyx_t_3 = (__pyx_t_1 != __pyx_t_2);
  if (__pyx_t_3) {

    /* "src/container.pxi":14
 * cdef int viewcmp(const uint8_t[:] a, const uint8_t[:] b):
 *     if len(a) != len(b):
 *         return 1             # <<<<<<<<<<<<<<
 *     return memcmp(&a[0], &b[0], len(a))
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "src/container.pxi":13
 * 
 * cdef int viewcmp(const uint8_t[:] a, const uint8_t[:] b):
 *     if len(a) != len(b):             # <<<<<<<<<<<<<<
 *         return 1
 *     return memcmp(&a[0], &b[0], len(a))
 */
  }

  /* "src/container.pxi":15
 *     if len(a) != len(b):
 *         return 1
 *     return memcmp(&a[0], &b[0], len(a))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_v_a.shape[0];
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_v_a.shape[0])) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(25, 15, __pyx_L1_error)
  }
  __pyx_t_6 = 0;
  __pyx_t_5 = -1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_v_b.shape[0];
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_5 = 0;
  } else if (unlikely(__pyx_t_6 >= __pyx_v_b.shape[0])) __pyx_t_5 = 0;
  if (unlikely(__pyx_t_5 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_5);
    __PYX_ERR(25, 15, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_MemoryView_Len(__pyx_v_a); 
  __pyx_r = memcmp((&(*((uint8_t const  *) ( /* dim=0 */ (__pyx_v_a.data + __pyx_t_4 * __pyx_v_a.strides[0]) )))), (&(*((uint8_t const  *) ( /* dim=0 */ (__pyx_v_b.data + __pyx_t_6 * __pyx_v_b.strides[0]) )))), __pyx_t_2);
  goto __pyx_L0;

  /* "src/container.pxi":12
 * 
 * 
 * cdef int viewcmp(const uint8_t[:] a, const uint8_t[:] b):             # <<<<<<<<<<<<<<
 *     if len(a) != len(b):
 *         return 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.viewcmp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "src/container.pxi":18
 * 
 * 
 * cdef _Reader make_reader(src):             # <<<<<<<<<<<<<<
 *     if isinstance(src, _Reader):
 *         return src
 */

static struct __pyx_obj_5cavro__Reader *__pyx_f_5cavro_make_reader(PyObject *__pyx_v_src) {
  struct __pyx_obj_5cavro__Reader *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("make_reader", 1);

  /* "src/container.pxi":19
 * 
 * cdef _Reader make_reader(src):
 *     if isinstance(src, _Reader):             # <<<<<<<<<<<<<<
 *         return src
 *     elif isinstance(src, bytes):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_src, __pyx_ptype_5cavro__Reader); 
  if (__pyx_t_1) {

    /* "src/container.pxi":20
 * cdef _Reader make_reader(src):
 *     if isinstance(src, _Reader):
 *         return src             # <<<<<<<<<<<<<<
 *     elif isinstance(src, bytes):
 *         return MemoryReader(src)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_ptype_5cavro__Reader))))) __PYX_ERR(25, 20, __pyx_L1_error)
    __Pyx_INCREF(__pyx_v_src);
    __pyx_r = ((struct __pyx_obj_5cavro__Reader *)__pyx_v_src);
    goto __pyx_L0;

    /* "src/container.pxi":19
 * 
 * cdef _Reader make_reader(src):
 *     if isinstance(src, _Reader):             # <<<<<<<<<<<<<<
 *         return src
 *     elif isinstance(src, bytes):
 */
  }

  /* "src/container.pxi":21
 *     if isinstance(src, _Reader):
 *         return src
 *     elif isinstance(src, bytes):             # <<<<<<<<<<<<<<
 *         return MemoryReader(src)
 *     elif isinstance(src, (str, Path)):
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_src); 
  if (__pyx_t_1) {

    /* "src/container.pxi":22
 *         return src
 *     elif isinstance(src, bytes):
 *         return MemoryReader(src)             # <<<<<<<<<<<<<<
 *     elif isinstance(src, (str, Path)):
 *         return FileReader(Path(src).open('rb'))
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro_MemoryReader), __pyx_v_src); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 22, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = ((struct __pyx_obj_5cavro__Reader *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "src/container.pxi":21
 *     if isinstance(src, _Reader):
 *         return src
 *     elif isinstance(src, bytes):             # <<<<<<<<<<<<<<
 *         return MemoryReader(src)
 *     elif isinstance(src, (str, Path)):
 */
  }

  /* "src/container.pxi":23
 *     elif isinstance(src, bytes):
 *         return MemoryReader(src)
 *     elif isinstance(src, (str, Path)):             # <<<<<<<<<<<<<<
 *         return FileReader(Path(src).open('rb'))
 *     elif hasattr(src, 'read'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Path); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyUnicode_Check(__pyx_v_src); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_src, __pyx_t_2); 
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "src/container.pxi":24
 *         return MemoryReader(src)
 *     elif isinstance(src, (str, Path)):
 *         return FileReader(Path(src).open('rb'))             # <<<<<<<<<<<<<<
 *     elif hasattr(src, 'read'):
 *         return FileReader(src)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Path); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_src};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 24, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_open); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_rb};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 24, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro_FileReader), __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = ((struct __pyx_obj_5cavro__Reader *)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "src/container.pxi":23
 *     elif isinstance(src, bytes):
 *         return MemoryReader(src)
 *     elif isinstance(src, (str, Path)):             # <<<<<<<<<<<<<<
 *         return FileReader(Path(src).open('rb'))
 *     elif hasattr(src, 'read'):
 */
  }

  /* "src/container.pxi":25
 *     elif isinstance(src, (str, Path)):
 *         return FileReader(Path(src).open('rb'))
 *     elif hasattr(src, 'read'):             # <<<<<<<<<<<<<<
 *         return FileReader(src)
 *     else:
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_src, __pyx_n_u_read); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(25, 25, __pyx_L1_error)
  if (likely(__pyx_t_1)) {

    /* "src/container.pxi":26
 *         return FileReader(Path(src).open('rb'))
 *     elif hasattr(src, 'read'):
 *         return FileReader(src)             # <<<<<<<<<<<<<<
 *     else:
 *         raise NotImplementedError(f"Cannot read from '{src}'")
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro_FileReader), __pyx_v_src); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 26, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = ((struct __pyx_obj_5cavro__Reader *)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "src/container.pxi":25
 *     elif isinstance(src, (str, Path)):
 *         return FileReader(Path(src).open('rb'))
 *     elif hasattr(src, 'read'):             # <<<<<<<<<<<<<<
 *         return FileReader(src)
 *     else:
 */
  }

  /* "src/container.pxi":28
 *         return FileReader(src)
 *     else:
 *         raise NotImplementedError(f"Cannot read from '{src}'")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 28, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = 0;
    __pyx_t_9 = 127;
    __Pyx_INCREF(__pyx_kp_u_Cannot_read_from);
    __pyx_t_8 += 18;
    __Pyx_GIVEREF(__pyx_kp_u_Cannot_read_from);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Cannot_read_from);
    __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_src, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 28, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u__6);
    __pyx_t_8 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__6);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u__6);
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 28, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 28, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(25, 28, __pyx_L1_error)
  }

  /* "src/container.pxi":18
 * 
 * 
 * cdef _Reader make_reader(src):             # <<<<<<<<<<<<<<
 *     if isinstance(src, _Reader):
 *         return src
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.make_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":31
 * 
 * 
 * cdef _Writer make_writer(src):             # <<<<<<<<<<<<<<
 *     if isinstance(src, _Writer):
 *         return src
 */

static struct __pyx_obj_5cavro__Writer *__pyx_f_5cavro_make_writer(PyObject *__pyx_v_src) {
  struct __pyx_obj_5cavro__Writer *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("make_writer", 1);

  /* "src/container.pxi":32
 * 
 * cdef _Writer make_writer(src):
 *     if isinstance(src, _Writer):             # <<<<<<<<<<<<<<
 *         return src
 *     elif isinstance(src, (str, Path)):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_src, __pyx_ptype_5cavro__Writer); 
  if (__pyx_t_1) {

    /* "src/container.pxi":33
 * cdef _Writer make_writer(src):
 *     if isinstance(src, _Writer):
 *         return src             # <<<<<<<<<<<<<<
 *     elif isinstance(src, (str, Path)):
 *         return FileWriter(Path(src).open('wb'))
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_ptype_5cavro__Writer))))) __PYX_ERR(25, 33, __pyx_L1_error)
    __Pyx_INCREF(__pyx_v_src);
    __pyx_r = ((struct __pyx_obj_5cavro__Writer *)__pyx_v_src);
    goto __pyx_L0;

    /* "src/container.pxi":32
 * 
 * cdef _Writer make_writer(src):
 *     if isinstance(src, _Writer):             # <<<<<<<<<<<<<<
 *         return src
 *     elif isinstance(src, (str, Path)):
 */
  }

  /* "src/container.pxi":34
 *     if isinstance(src, _Writer):
 *         return src
 *     elif isinstance(src, (str, Path)):             # <<<<<<<<<<<<<<
 *         return FileWriter(Path(src).open('wb'))
 *     elif hasattr(src, 'write'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Path); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyUnicode_Check(__pyx_v_src); 
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_src, __pyx_t_2); 
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "src/container.pxi":35
 *         return src
 *     elif isinstance(src, (str, Path)):
 *         return FileWriter(Path(src).open('wb'))             # <<<<<<<<<<<<<<
 *     elif hasattr(src, 'write'):
 *         return FileWriter(src)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Path); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_src};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 35, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_open); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_wb};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 35, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro_FileWriter), __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = ((struct __pyx_obj_5cavro__Writer *)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "src/container.pxi":34
 *     if isinstance(src, _Writer):
 *         return src
 *     elif isinstance(src, (str, Path)):             # <<<<<<<<<<<<<<
 *         return FileWriter(Path(src).open('wb'))
 *     elif hasattr(src, 'write'):
 */
  }

  /* "src/container.pxi":36
 *     elif isinstance(src, (str, Path)):
 *         return FileWriter(Path(src).open('wb'))
 *     elif hasattr(src, 'write'):             # <<<<<<<<<<<<<<
 *         return FileWriter(src)
 *     raise NotImplementedError(f"Cannot write to '{src}'")
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_src, __pyx_n_u_write); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(25, 36, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "src/container.pxi":37
 *         return FileWriter(Path(src).open('wb'))
 *     elif hasattr(src, 'write'):
 *         return FileWriter(src)             # <<<<<<<<<<<<<<
 *     raise NotImplementedError(f"Cannot write to '{src}'")
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro_FileWriter), __pyx_v_src); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = ((struct __pyx_obj_5cavro__Writer *)__pyx_t_5);
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "src/container.pxi":36
 *     elif isinstance(src, (str, Path)):
 *         return FileWriter(Path(src).open('wb'))
 *     elif hasattr(src, 'write'):             # <<<<<<<<<<<<<<
 *         return FileWriter(src)
 *     raise NotImplementedError(f"Cannot write to '{src}'")
 */
  }

  /* "src/container.pxi":38
 *     elif hasattr(src, 'write'):
 *         return FileWriter(src)
 *     raise NotImplementedError(f"Cannot write to '{src}'")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = 0;
  __pyx_t_9 = 127;
  __Pyx_INCREF(__pyx_kp_u_Cannot_write_to);
  __pyx_t_8 += 17;
  __Pyx_GIVEREF(__pyx_kp_u_Cannot_write_to);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Cannot_write_to);
  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_src, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_9;
  __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_INCREF(__pyx_kp_u__6);
  __pyx_t_8 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__6);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u__6);
  __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_Raise(__pyx_t_5, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __PYX_ERR(25, 38, __pyx_L1_error)

  /* "src/container.pxi":31
 * 
 * 
 * cdef _Writer make_writer(src):             # <<<<<<<<<<<<<<
 *     if isinstance(src, _Writer):
 *         return src
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cavro.make_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":69
 *     cdef _Reader reader
 * 
 *     def __init__(self, src, reader_schema=None, options=DEFAULT_OPTIONS):             # <<<<<<<<<<<<<<
 *         self.reader = make_reader(src)
 *         cdef const uint8_t[:] header = self.reader.read_n(4)
 */

/* Python wrapper */
static int __pyx_pw_5cavro_15ContainerReader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_15ContainerReader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_src = 0;
  PyObject *__pyx_v_reader_schema = 0;
  PyObject *__pyx_v_options = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_src,&__pyx_n_s_reader_schema,&__pyx_n_s_options,0};
    values[1] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    values[2] = __Pyx_Arg_NewRef_VARARGS(__pyx_k__87);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_src)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 69, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_reader_schema);
          if (value) { values[1] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 69, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_options);
          if (value) { values[2] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 69, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(25, 69, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_src = values[0];
    __pyx_v_reader_schema = values[1];
    __pyx_v_options = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 3, __pyx_nargs); __PYX_ERR(25, 69, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ContainerReader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_15ContainerReader___init__(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self), __pyx_v_src, __pyx_v_reader_schema, __pyx_v_options);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_15ContainerReader___init__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self, PyObject *__pyx_v_src, PyObject *__pyx_v_reader_schema, PyObject *__pyx_v_options) {
  __Pyx_memviewslice __pyx_v_header = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_writer_options = NULL;
  PyObject *__pyx_v_codec_name = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  __Pyx_memviewslice __pyx_t_18 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/container.pxi":70
 * 
 *     def __init__(self, src, reader_schema=None, options=DEFAULT_OPTIONS):
 *         self.reader = make_reader(src)             # <<<<<<<<<<<<<<
 *         cdef const uint8_t[:] header = self.reader.read_n(4)
 *         if viewcmp(header, OBJ_MAGIC):
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5cavro_make_reader(__pyx_v_src)); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->reader);
  __Pyx_DECREF((PyObject *)__pyx_v_self->reader);
  __pyx_v_self->reader = ((struct __pyx_obj_5cavro__Reader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/container.pxi":71
 *     def __init__(self, src, reader_schema=None, options=DEFAULT_OPTIONS):
 *         self.reader = make_reader(src)
 *         cdef const uint8_t[:] header = self.reader.read_n(4)             # <<<<<<<<<<<<<<
 *         if viewcmp(header, OBJ_MAGIC):
 *             raise ValueError(f"Invalid file header, expected: {bytes(OBJ_MAGIC)} got {bytes(header)}")
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_self->reader->__pyx_vtab)->read_n(__pyx_v_self->reader, 4); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(25, 71, __pyx_L1_error)
  __pyx_v_header = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "src/container.pxi":72
 *         self.reader = make_reader(src)
 *         cdef const uint8_t[:] header = self.reader.read_n(4)
 *         if viewcmp(header, OBJ_MAGIC):             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Invalid file header, expected: {bytes(OBJ_MAGIC)} got {bytes(header)}")
 *         self.metadata = OBJ_FILE_METADATA.binary_read(self.reader)
 */
  if (unlikely(!__pyx_v_5cavro_OBJ_MAGIC.memview)) { __Pyx_RaiseUnboundLocalError("OBJ_MAGIC"); __PYX_ERR(25, 72, __pyx_L1_error) }
  __pyx_t_3 = __pyx_f_5cavro_viewcmp(__pyx_v_header, __pyx_v_5cavro_OBJ_MAGIC); if (unlikely(__pyx_t_3 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(25, 72, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (unlikely(__pyx_t_4)) {

    /* "src/container.pxi":73
 *         cdef const uint8_t[:] header = self.reader.read_n(4)
 *         if viewcmp(header, OBJ_MAGIC):
 *             raise ValueError(f"Invalid file header, expected: {bytes(OBJ_MAGIC)} got {bytes(header)}")             # <<<<<<<<<<<<<<
 *         self.metadata = OBJ_FILE_METADATA.binary_read(self.reader)
 *         writer_options = options
 */
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_Invalid_file_header_expected);
    __pyx_t_5 += 31;
    __Pyx_GIVEREF(__pyx_kp_u_Invalid_file_header_expected);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Invalid_file_header_expected);
    if (unlikely(!__pyx_v_5cavro_OBJ_MAGIC.memview)) { __Pyx_RaiseUnboundLocalError("OBJ_MAGIC"); __PYX_ERR(25, 73, __pyx_L1_error) }
    __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_5cavro_OBJ_MAGIC, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(25, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_got_3);
    __pyx_t_5 += 5;
    __Pyx_GIVEREF(__pyx_kp_u_got_3);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_got_3);
    __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_header, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(25, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(25, 73, __pyx_L1_error)

    /* "src/container.pxi":72
 *         self.reader = make_reader(src)
 *         cdef const uint8_t[:] header = self.reader.read_n(4)
 *         if viewcmp(header, OBJ_MAGIC):             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Invalid file header, expected: {bytes(OBJ_MAGIC)} got {bytes(header)}")
 *         self.metadata = OBJ_FILE_METADATA.binary_read(self.reader)
 */
  }

  /* "src/container.pxi":74
 *         if viewcmp(header, OBJ_MAGIC):
 *             raise ValueError(f"Invalid file header, expected: {bytes(OBJ_MAGIC)} got {bytes(header)}")
 *         self.metadata = OBJ_FILE_METADATA.binary_read(self.reader)             # <<<<<<<<<<<<<<
 *         writer_options = options
 *         if reader_schema is not None:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_OBJ_FILE_METADATA); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_binary_read); if (unlikely(!__pyx_t_8)) __PYX_ERR(25, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  __pyx_t_3 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
      __pyx_t_3 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, ((PyObject *)__pyx_v_self->reader)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_3, 1+__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->metadata);
  __Pyx_DECREF(__pyx_v_self->metadata);
  __pyx_v_self->metadata = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/container.pxi":75
 *             raise ValueError(f"Invalid file header, expected: {bytes(OBJ_MAGIC)} got {bytes(header)}")
 *         self.metadata = OBJ_FILE_METADATA.binary_read(self.reader)
 *         writer_options = options             # <<<<<<<<<<<<<<
 *         if reader_schema is not None:
 *             writer_options = writer_options.replace(allow_invalid_default_values=True)
 */
  __Pyx_INCREF(__pyx_v_options);
  __pyx_v_writer_options = __pyx_v_options;

  /* "src/container.pxi":76
 *         self.metadata = OBJ_FILE_METADATA.binary_read(self.reader)
 *         writer_options = options
 *         if reader_schema is not None:             # <<<<<<<<<<<<<<
 *             writer_options = writer_options.replace(allow_invalid_default_values=True)
 *         self.writer_schema = Schema(self.metadata['avro.schema'], options=writer_options, allow_union_default_any_member=True)
 */
  __pyx_t_4 = (__pyx_v_reader_schema != Py_None);
  if (__pyx_t_4) {

    /* "src/container.pxi":77
 *         writer_options = options
 *         if reader_schema is not None:
 *             writer_options = writer_options.replace(allow_invalid_default_values=True)             # <<<<<<<<<<<<<<
 *         self.writer_schema = Schema(self.metadata['avro.schema'], options=writer_options, allow_union_default_any_member=True)
 *         if reader_schema is None:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_writer_options, __pyx_n_s_replace); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(25, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_allow_invalid_default_values, Py_True) < 0) __PYX_ERR(25, 77, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF_SET(__pyx_v_writer_options, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "src/container.pxi":76
 *         self.metadata = OBJ_FILE_METADATA.binary_read(self.reader)
 *         writer_options = options
 *         if reader_schema is not None:             # <<<<<<<<<<<<<<
 *             writer_options = writer_options.replace(allow_invalid_default_values=True)
 *         self.writer_schema = Schema(self.metadata['avro.schema'], options=writer_options, allow_union_default_any_member=True)
 */
  }

  /* "src/container.pxi":78
 *         if reader_schema is not None:
 *             writer_options = writer_options.replace(allow_invalid_default_values=True)
 *         self.writer_schema = Schema(self.metadata['avro.schema'], options=writer_options, allow_union_default_any_member=True)             # <<<<<<<<<<<<<<
 *         if reader_schema is None:
 *             self.schema = self.reader_schema = self.writer_schema
 */
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_v_self->metadata, __pyx_kp_u_avro_schema); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(25, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7)) __PYX_ERR(25, 78, __pyx_L1_error);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_options, __pyx_v_writer_options) < 0) __PYX_ERR(25, 78, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_allow_union_default_any_member, Py_True) < 0) __PYX_ERR(25, 78, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_Schema), __pyx_t_8, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->writer_schema);
  __Pyx_DECREF((PyObject *)__pyx_v_self->writer_schema);
  __pyx_v_self->writer_schema = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/container.pxi":79
 *             writer_options = writer_options.replace(allow_invalid_default_values=True)
 *         self.writer_schema = Schema(self.metadata['avro.schema'], options=writer_options, allow_union_default_any_member=True)
 *         if reader_schema is None:             # <<<<<<<<<<<<<<
 *             self.schema = self.reader_schema = self.writer_schema
 *         else:
 */
  __pyx_t_4 = (__pyx_v_reader_schema == Py_None);
  if (__pyx_t_4) {

    /* "src/container.pxi":80
 *         self.writer_schema = Schema(self.metadata['avro.schema'], options=writer_options, allow_union_default_any_member=True)
 *         if reader_schema is None:
 *             self.schema = self.reader_schema = self.writer_schema             # <<<<<<<<<<<<<<
 *         else:
 *             self.reader_schema = reader_schema
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_self->writer_schema);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->schema);
    __Pyx_DECREF((PyObject *)__pyx_v_self->schema);
    __pyx_v_self->schema = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->reader_schema);
    __Pyx_DECREF((PyObject *)__pyx_v_self->reader_schema);
    __pyx_v_self->reader_schema = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "src/container.pxi":79
 *             writer_options = writer_options.replace(allow_invalid_default_values=True)
 *         self.writer_schema = Schema(self.metadata['avro.schema'], options=writer_options, allow_union_default_any_member=True)
 *         if reader_schema is None:             # <<<<<<<<<<<<<<
 *             self.schema = self.reader_schema = self.writer_schema
 *         else:
 */
    goto __pyx_L5;
  }

  /* "src/container.pxi":82
 *             self.schema = self.reader_schema = self.writer_schema
 *         else:
 *             self.reader_schema = reader_schema             # <<<<<<<<<<<<<<
 *             self.schema = self.reader_schema.reader_for_writer(self.writer_schema)
 * 
 */
  /*else*/ {
    if (!(likely(((__pyx_v_reader_schema) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_reader_schema, __pyx_ptype_5cavro_Schema))))) __PYX_ERR(25, 82, __pyx_L1_error)
    __pyx_t_1 = __pyx_v_reader_schema;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->reader_schema);
    __Pyx_DECREF((PyObject *)__pyx_v_self->reader_schema);
    __pyx_v_self->reader_schema = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "src/container.pxi":83
 *         else:
 *             self.reader_schema = reader_schema
 *             self.schema = self.reader_schema.reader_for_writer(self.writer_schema)             # <<<<<<<<<<<<<<
 * 
 *         codec_name = self.metadata.get('avro.codec', b'null')
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_self->writer_schema);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7 = ((PyObject *)((struct __pyx_vtabstruct_5cavro_Schema *)__pyx_v_self->reader_schema->__pyx_vtab)->reader_for_writer(__pyx_v_self->reader_schema, ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1), 0)); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_7);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->schema);
    __Pyx_DECREF((PyObject *)__pyx_v_self->schema);
    __pyx_v_self->schema = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_7);
    __pyx_t_7 = 0;
  }
  __pyx_L5:;

  /* "src/container.pxi":85
 *             self.schema = self.reader_schema.reader_for_writer(self.writer_schema)
 * 
 *         codec_name = self.metadata.get('avro.codec', b'null')             # <<<<<<<<<<<<<<
 *         self.codec_name = codec_name.decode()
 *         try:
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->metadata, __pyx_n_s_get); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__88, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_codec_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "src/container.pxi":86
 * 
 *         codec_name = self.metadata.get('avro.codec', b'null')
 *         self.codec_name = codec_name.decode()             # <<<<<<<<<<<<<<
 *         try:
 *             self.codec = CODECS[codec_name]
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_codec_name, __pyx_n_s_decode); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  __pyx_t_3 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_3 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_3, 0+__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(25, 86, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->codec_name);
  __Pyx_DECREF(__pyx_v_self->codec_name);
  __pyx_v_self->codec_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/container.pxi":87
 *         codec_name = self.metadata.get('avro.codec', b'null')
 *         self.codec_name = codec_name.decode()
 *         try:             # <<<<<<<<<<<<<<
 *             self.codec = CODECS[codec_name]
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "src/container.pxi":88
 *         self.codec_name = codec_name.decode()
 *         try:
 *             self.codec = CODECS[codec_name]             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             raise CodecUnavailable(f"Unsupported codec: '{codec_name.decode('utf-8')}'")
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CODECS); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 88, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_codec_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 88, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5cavro_Codec))))) __PYX_ERR(25, 88, __pyx_L6_error)
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->codec);
      __Pyx_DECREF((PyObject *)__pyx_v_self->codec);
      __pyx_v_self->codec = ((struct __pyx_obj_5cavro_Codec *)__pyx_t_7);
      __pyx_t_7 = 0;

      /* "src/container.pxi":87
 *         codec_name = self.metadata.get('avro.codec', b'null')
 *         self.codec_name = codec_name.decode()
 *         try:             # <<<<<<<<<<<<<<
 *             self.codec = CODECS[codec_name]
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L11_try_end;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_XCLEAR_MEMVIEW(&__pyx_t_2, 1);
    __pyx_t_2.memview = NULL; __pyx_t_2.data = NULL;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "src/container.pxi":89
 *         try:
 *             self.codec = CODECS[codec_name]
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             raise CodecUnavailable(f"Unsupported codec: '{codec_name.decode('utf-8')}'")
 *         self.objects_left_in_block = 0
 */
    __pyx_t_3 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_3) {
      __Pyx_AddTraceback("cavro.ContainerReader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_1, &__pyx_t_8) < 0) __PYX_ERR(25, 89, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "src/container.pxi":90
 *             self.codec = CODECS[codec_name]
 *         except KeyError:
 *             raise CodecUnavailable(f"Unsupported codec: '{codec_name.decode('utf-8')}'")             # <<<<<<<<<<<<<<
 *         self.objects_left_in_block = 0
 *         self.current_block = MemoryReader(empty_buffer)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_CodecUnavailable); if (unlikely(!__pyx_t_13)) __PYX_ERR(25, 90, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_14 = PyTuple_New(3); if (unlikely(!__pyx_t_14)) __PYX_ERR(25, 90, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_5 = 0;
      __pyx_t_6 = 127;
      __Pyx_INCREF(__pyx_kp_u_Unsupported_codec);
      __pyx_t_5 += 20;
      __Pyx_GIVEREF(__pyx_kp_u_Unsupported_codec);
      PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_kp_u_Unsupported_codec);
      __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_codec_name, __pyx_n_s_decode); if (unlikely(!__pyx_t_16)) __PYX_ERR(25, 90, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_17 = NULL;
      __pyx_t_3 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_16))) {
        __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_16);
        if (likely(__pyx_t_17)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
          __Pyx_INCREF(__pyx_t_17);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_16, function);
          __pyx_t_3 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_17, __pyx_kp_u_utf_8};
        __pyx_t_15 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_3, 1+__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        if (unlikely(!__pyx_t_15)) __PYX_ERR(25, 90, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      }
      __pyx_t_16 = __Pyx_PyObject_FormatSimple(__pyx_t_15, __pyx_empty_unicode); if (unlikely(!__pyx_t_16)) __PYX_ERR(25, 90, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_16);
      __Pyx_GIVEREF(__pyx_t_16);
      PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_16);
      __pyx_t_16 = 0;
      __Pyx_INCREF(__pyx_kp_u__6);
      __pyx_t_5 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__6);
      PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_kp_u__6);
      __pyx_t_16 = __Pyx_PyUnicode_Join(__pyx_t_14, 3, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_16)) __PYX_ERR(25, 90, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = NULL;
      __pyx_t_3 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_13))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_13, function);
          __pyx_t_3 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_14, __pyx_t_16};
        __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_3, 1+__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        if (unlikely(!__pyx_t_12)) __PYX_ERR(25, 90, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_Raise(__pyx_t_12, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __PYX_ERR(25, 90, __pyx_L8_except_error)
    }
    goto __pyx_L8_except_error;

    /* "src/container.pxi":87
 *         codec_name = self.metadata.get('avro.codec', b'null')
 *         self.codec_name = codec_name.decode()
 *         try:             # <<<<<<<<<<<<<<
 *             self.codec = CODECS[codec_name]
 *         except KeyError:
 */
    __pyx_L8_except_error:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L1_error;
    __pyx_L11_try_end:;
  }

  /* "src/container.pxi":91
 *         except KeyError:
 *             raise CodecUnavailable(f"Unsupported codec: '{codec_name.decode('utf-8')}'")
 *         self.objects_left_in_block = 0             # <<<<<<<<<<<<<<
 *         self.current_block = MemoryReader(empty_buffer)
 *         self.marker = b''
 */
  __pyx_v_self->objects_left_in_block = 0;

  /* "src/container.pxi":92
 *             raise CodecUnavailable(f"Unsupported codec: '{codec_name.decode('utf-8')}'")
 *         self.objects_left_in_block = 0
 *         self.current_block = MemoryReader(empty_buffer)             # <<<<<<<<<<<<<<
 *         self.marker = b''
 * 
 */
  __pyx_t_8 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro_MemoryReader), __pyx_v_5cavro_empty_buffer); if (unlikely(!__pyx_t_8)) __PYX_ERR(25, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_8);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->current_block);
  __Pyx_DECREF((PyObject *)__pyx_v_self->current_block);
  __pyx_v_self->current_block = ((struct __pyx_obj_5cavro_MemoryReader *)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "src/container.pxi":93
 *         self.objects_left_in_block = 0
 *         self.current_block = MemoryReader(empty_buffer)
 *         self.marker = b''             # <<<<<<<<<<<<<<
 * 
 *     cpdef _read_marker(self):
 */
  __pyx_t_18 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_kp_b__13, 0); if (unlikely(!__pyx_t_18.memview)) __PYX_ERR(25, 93, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_v_self->marker, 0);
  __pyx_v_self->marker = __pyx_t_18;
  __pyx_t_18.memview = NULL;
  __pyx_t_18.data = NULL;

  /* "src/container.pxi":69
 *     cdef _Reader reader
 * 
 *     def __init__(self, src, reader_schema=None, options=DEFAULT_OPTIONS):             # <<<<<<<<<<<<<<
 *         self.reader = make_reader(src)
 *         cdef const uint8_t[:] header = self.reader.read_n(4)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_2, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_18, 1);
  __Pyx_AddTraceback("cavro.ContainerReader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XCLEAR_MEMVIEW(&__pyx_v_header, 1);
  __Pyx_XDECREF(__pyx_v_writer_options);
  __Pyx_XDECREF(__pyx_v_codec_name);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":95
 *         self.marker = b''
 * 
 *     cpdef _read_marker(self):             # <<<<<<<<<<<<<<
 *         value = self.reader.read_n(16)
 *         if not len(self.marker):
 */

static PyObject *__pyx_pw_5cavro_15ContainerReader_3_read_marker(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_15ContainerReader__read_marker(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_value = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_read_marker", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_read_marker); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_15ContainerReader_3_read_marker)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 95, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/container.pxi":96
 * 
 *     cpdef _read_marker(self):
 *         value = self.reader.read_n(16)             # <<<<<<<<<<<<<<
 *         if not len(self.marker):
 *             self.marker = value
 */
  __pyx_t_6 = ((struct __pyx_vtabstruct_5cavro__Reader *)__pyx_v_self->reader->__pyx_vtab)->read_n(__pyx_v_self->reader, 16); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(25, 96, __pyx_L1_error)
  __pyx_v_value = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "src/container.pxi":97
 *     cpdef _read_marker(self):
 *         value = self.reader.read_n(16)
 *         if not len(self.marker):             # <<<<<<<<<<<<<<
 *             self.marker = value
 *         else:
 */
  if (unlikely(!__pyx_v_self->marker.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(25, 97, __pyx_L1_error)}
  __pyx_t_7 = __Pyx_MemoryView_Len(__pyx_v_self->marker); 
  __pyx_t_8 = (!(__pyx_t_7 != 0));
  if (__pyx_t_8) {

    /* "src/container.pxi":98
 *         value = self.reader.read_n(16)
 *         if not len(self.marker):
 *             self.marker = value             # <<<<<<<<<<<<<<
 *         else:
 *             if viewcmp(value, self.marker):
 */
    __PYX_XCLEAR_MEMVIEW(&__pyx_v_self->marker, 0);
    __PYX_INC_MEMVIEW(&__pyx_v_value, 1);
    __pyx_v_self->marker = __pyx_v_value;

    /* "src/container.pxi":97
 *     cpdef _read_marker(self):
 *         value = self.reader.read_n(16)
 *         if not len(self.marker):             # <<<<<<<<<<<<<<
 *             self.marker = value
 *         else:
 */
    goto __pyx_L3;
  }

  /* "src/container.pxi":100
 *             self.marker = value
 *         else:
 *             if viewcmp(value, self.marker):             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Invalid block sync marker, expected {bytes(self.marker)}, got {bytes(value)}")
 * 
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->marker.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(25, 100, __pyx_L1_error)}
    __pyx_t_5 = __pyx_f_5cavro_viewcmp(__pyx_v_value, __pyx_v_self->marker); if (unlikely(__pyx_t_5 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(25, 100, __pyx_L1_error)
    __pyx_t_8 = (__pyx_t_5 != 0);
    if (unlikely(__pyx_t_8)) {

      /* "src/container.pxi":101
 *         else:
 *             if viewcmp(value, self.marker):
 *                 raise ValueError(f"Invalid block sync marker, expected {bytes(self.marker)}, got {bytes(value)}")             # <<<<<<<<<<<<<<
 * 
 *     cdef int next_block(self) except -1:
 */
      __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = 0;
      __pyx_t_9 = 127;
      __Pyx_INCREF(__pyx_kp_u_Invalid_block_sync_marker_expect);
      __pyx_t_7 += 36;
      __Pyx_GIVEREF(__pyx_kp_u_Invalid_block_sync_marker_expect);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Invalid_block_sync_marker_expect);
      if (unlikely(!__pyx_v_self->marker.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(25, 101, __pyx_L1_error)}
      __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->marker, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_9;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_INCREF(__pyx_kp_u_got_4);
      __pyx_t_7 += 6;
      __Pyx_GIVEREF(__pyx_kp_u_got_4);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_got_4);
      __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_value, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_9;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_7, __pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(25, 101, __pyx_L1_error)

      /* "src/container.pxi":100
 *             self.marker = value
 *         else:
 *             if viewcmp(value, self.marker):             # <<<<<<<<<<<<<<
 *                 raise ValueError(f"Invalid block sync marker, expected {bytes(self.marker)}, got {bytes(value)}")
 * 
 */
    }
  }
  __pyx_L3:;

  /* "src/container.pxi":95
 *         self.marker = b''
 * 
 *     cpdef _read_marker(self):             # <<<<<<<<<<<<<<
 *         value = self.reader.read_n(16)
 *         if not len(self.marker):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("cavro.ContainerReader._read_marker", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XCLEAR_MEMVIEW(&__pyx_v_value, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_3_read_marker(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15ContainerReader_3_read_marker = {"_read_marker", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerReader_3_read_marker, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15ContainerReader_3_read_marker(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_read_marker (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("_read_marker", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "_read_marker", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_2_read_marker(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_2_read_marker(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_read_marker", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_15ContainerReader__read_marker(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ContainerReader._read_marker", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":103
 *                 raise ValueError(f"Invalid block sync marker, expected {bytes(self.marker)}, got {bytes(value)}")
 * 
 *     cdef int next_block(self) except -1:             # <<<<<<<<<<<<<<
 *         try:
 *             self._read_marker()
 */

static int __pyx_f_5cavro_15ContainerReader_next_block(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_v_e = NULL;
  size_t __pyx_v_block_size;
  __Pyx_memviewslice __pyx_v_block_bytes = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int64_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_t_18;
  __Pyx_memviewslice __pyx_t_19 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_20 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("next_block", 1);

  /* "src/container.pxi":104
 * 
 *     cdef int next_block(self) except -1:
 *         try:             # <<<<<<<<<<<<<<
 *             self._read_marker()
 *             self.objects_left_in_block = zigzag_decode_long(self.reader)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "src/container.pxi":105
 *     cdef int next_block(self) except -1:
 *         try:
 *             self._read_marker()             # <<<<<<<<<<<<<<
 *             self.objects_left_in_block = zigzag_decode_long(self.reader)
 *         except (EOFError, ValueError) as e:
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro_ContainerReader *)__pyx_v_self->__pyx_vtab)->_read_marker(__pyx_v_self, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 105, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "src/container.pxi":106
 *         try:
 *             self._read_marker()
 *             self.objects_left_in_block = zigzag_decode_long(self.reader)             # <<<<<<<<<<<<<<
 *         except (EOFError, ValueError) as e:
 *             raise StopIteration() from e
 */
      __pyx_t_4 = ((PyObject *)__pyx_v_self->reader);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = __pyx_f_5cavro_zigzag_decode_long(((struct __pyx_obj_5cavro__Reader *)__pyx_t_4)); if (unlikely(__pyx_t_5 == ((int64_t)0x7ffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(25, 106, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_self->objects_left_in_block = __pyx_t_5;

      /* "src/container.pxi":104
 * 
 *     cdef int next_block(self) except -1:
 *         try:             # <<<<<<<<<<<<<<
 *             self._read_marker()
 *             self.objects_left_in_block = zigzag_decode_long(self.reader)
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "src/container.pxi":107
 *             self._read_marker()
 *             self.objects_left_in_block = zigzag_decode_long(self.reader)
 *         except (EOFError, ValueError) as e:             # <<<<<<<<<<<<<<
 *             raise StopIteration() from e
 *         cdef size_t block_size = zigzag_decode_long(self.reader)
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_EOFError, __pyx_builtin_ValueError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("cavro.ContainerReader.next_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(25, 107, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __pyx_v_e = __pyx_t_7;
      /*try:*/ {

        /* "src/container.pxi":108
 *             self.objects_left_in_block = zigzag_decode_long(self.reader)
 *         except (EOFError, ValueError) as e:
 *             raise StopIteration() from e             # <<<<<<<<<<<<<<
 *         cdef size_t block_size = zigzag_decode_long(self.reader)
 *         cdef const uint8_t[:] block_bytes
 */
        __pyx_t_9 = __Pyx_PyObject_CallNoArg(__pyx_builtin_StopIteration); if (unlikely(!__pyx_t_9)) __PYX_ERR(25, 108, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_Raise(__pyx_t_9, 0, 0, __pyx_v_e);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(25, 108, __pyx_L14_error)
      }

      /* "src/container.pxi":107
 *             self._read_marker()
 *             self.objects_left_in_block = zigzag_decode_long(self.reader)
 *         except (EOFError, ValueError) as e:             # <<<<<<<<<<<<<<
 *             raise StopIteration() from e
 *         cdef size_t block_size = zigzag_decode_long(self.reader)
 */
      /*finally:*/ {
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __pyx_t_6 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          }
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;

    /* "src/container.pxi":104
 * 
 *     cdef int next_block(self) except -1:
 *         try:             # <<<<<<<<<<<<<<
 *             self._read_marker()
 *             self.objects_left_in_block = zigzag_decode_long(self.reader)
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "src/container.pxi":109
 *         except (EOFError, ValueError) as e:
 *             raise StopIteration() from e
 *         cdef size_t block_size = zigzag_decode_long(self.reader)             # <<<<<<<<<<<<<<
 *         cdef const uint8_t[:] block_bytes
 *         if block_size > 0:
 */
  __pyx_t_8 = ((PyObject *)__pyx_v_self->reader);
  __Pyx_INCREF(__pyx_t_8);
  __pyx_t_5 = __pyx_f_5cavro_zigzag_decode_long(((struct __pyx_obj_5cavro__Reader *)__pyx_t_8)); if (unlikely(__pyx_t_5 == ((int64_t)0x7ffffffffffffbadULL) && PyErr_Occurred())) __PYX_ERR(25, 109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_block_size = __pyx_t_5;

  /* "src/container.pxi":111
 *         cdef size_t block_size = zigzag_decode_long(self.reader)
 *         cdef const uint8_t[:] block_bytes
 *         if block_size > 0:             # <<<<<<<<<<<<<<
 *             block_bytes = self.codec.read_block(self.reader, block_size)
 *         else:
 */
  __pyx_t_18 = (__pyx_v_block_size > 0);
  if (__pyx_t_18) {

    /* "src/container.pxi":112
 *         cdef const uint8_t[:] block_bytes
 *         if block_size > 0:
 *             block_bytes = self.codec.read_block(self.reader, block_size)             # <<<<<<<<<<<<<<
 *         else:
 *             block_bytes = empty_buffer
 */
    __pyx_t_8 = ((PyObject *)__pyx_v_self->reader);
    __Pyx_INCREF(__pyx_t_8);
    __pyx_t_19 = ((struct __pyx_vtabstruct_5cavro_Codec *)__pyx_v_self->codec->__pyx_vtab)->read_block(__pyx_v_self->codec, ((struct __pyx_obj_5cavro__Reader *)__pyx_t_8), __pyx_v_block_size); if (unlikely(!__pyx_t_19.memview)) __PYX_ERR(25, 112, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_block_bytes = __pyx_t_19;
    __pyx_t_19.memview = NULL;
    __pyx_t_19.data = NULL;

    /* "src/container.pxi":111
 *         cdef size_t block_size = zigzag_decode_long(self.reader)
 *         cdef const uint8_t[:] block_bytes
 *         if block_size > 0:             # <<<<<<<<<<<<<<
 *             block_bytes = self.codec.read_block(self.reader, block_size)
 *         else:
 */
    goto __pyx_L20;
  }

  /* "src/container.pxi":114
 *             block_bytes = self.codec.read_block(self.reader, block_size)
 *         else:
 *             block_bytes = empty_buffer             # <<<<<<<<<<<<<<
 *         self.current_block._reset_to(block_bytes)
 * 
 */
  /*else*/ {
    __pyx_t_20 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_5cavro_empty_buffer, 0); if (unlikely(!__pyx_t_20.memview)) __PYX_ERR(25, 114, __pyx_L1_error)
    __pyx_v_block_bytes = __pyx_t_20;
    __pyx_t_20.memview = NULL;
    __pyx_t_20.data = NULL;
  }
  __pyx_L20:;

  /* "src/container.pxi":115
 *         else:
 *             block_bytes = empty_buffer
 *         self.current_block._reset_to(block_bytes)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object next_object(self):
 */
  ((struct __pyx_vtabstruct_5cavro_MemoryReader *)__pyx_v_self->current_block->__pyx_base.__pyx_vtab)->_reset_to(__pyx_v_self->current_block, __pyx_v_block_bytes); if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 115, __pyx_L1_error)

  /* "src/container.pxi":103
 *                 raise ValueError(f"Invalid block sync marker, expected {bytes(self.marker)}, got {bytes(value)}")
 * 
 *     cdef int next_block(self) except -1:             # <<<<<<<<<<<<<<
 *         try:
 *             self._read_marker()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_19, 1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_20, 1);
  __Pyx_AddTraceback("cavro.ContainerReader.next_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __PYX_XCLEAR_MEMVIEW(&__pyx_v_block_bytes, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":117
 *         self.current_block._reset_to(block_bytes)
 * 
 *     cpdef object next_object(self):             # <<<<<<<<<<<<<<
 *         while self.objects_left_in_block < 1:
 *             self.next_block()
 */

static PyObject *__pyx_pw_5cavro_15ContainerReader_5next_object(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_5cavro_15ContainerReader_next_object(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("next_object", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_next_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_15ContainerReader_5next_object)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 117, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/container.pxi":118
 * 
 *     cpdef object next_object(self):
 *         while self.objects_left_in_block < 1:             # <<<<<<<<<<<<<<
 *             self.next_block()
 *         self.objects_left_in_block -= 1
 */
  while (1) {
    __pyx_t_6 = (__pyx_v_self->objects_left_in_block < 1);
    if (!__pyx_t_6) break;

    /* "src/container.pxi":119
 *     cpdef object next_object(self):
 *         while self.objects_left_in_block < 1:
 *             self.next_block()             # <<<<<<<<<<<<<<
 *         self.objects_left_in_block -= 1
 *         return self.schema.binary_read(self.current_block)
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_ContainerReader *)__pyx_v_self->__pyx_vtab)->next_block(__pyx_v_self); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(25, 119, __pyx_L1_error)
  }

  /* "src/container.pxi":120
 *         while self.objects_left_in_block < 1:
 *             self.next_block()
 *         self.objects_left_in_block -= 1             # <<<<<<<<<<<<<<
 *         return self.schema.binary_read(self.current_block)
 * 
 */
  __pyx_v_self->objects_left_in_block = (__pyx_v_self->objects_left_in_block - 1);

  /* "src/container.pxi":121
 *             self.next_block()
 *         self.objects_left_in_block -= 1
 *         return self.schema.binary_read(self.current_block)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->current_block);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_Schema *)__pyx_v_self->schema->__pyx_vtab)->binary_read(__pyx_v_self->schema, ((struct __pyx_obj_5cavro__Reader *)__pyx_t_1), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/container.pxi":117
 *         self.current_block._reset_to(block_bytes)
 * 
 *     cpdef object next_object(self):             # <<<<<<<<<<<<<<
 *         while self.objects_left_in_block < 1:
 *             self.next_block()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.ContainerReader.next_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_5next_object(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15ContainerReader_5next_object = {"next_object", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerReader_5next_object, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15ContainerReader_5next_object(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("next_object (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("next_object", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "next_object", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_4next_object(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_4next_object(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("next_object", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_15ContainerReader_next_object(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ContainerReader.next_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":123
 *         return self.schema.binary_read(self.current_block)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_7__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerReader_7__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_6__iter__(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_6__iter__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 1);

  /* "src/container.pxi":124
 * 
 *     def __iter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "src/container.pxi":123
 *         return self.schema.binary_read(self.current_block)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":126
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         return self.next_object()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_9__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerReader_9__next__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_8__next__(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_8__next__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__next__", 1);

  /* "src/container.pxi":127
 * 
 *     def __next__(self):
 *         return self.next_object()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro_ContainerReader *)__pyx_v_self->__pyx_vtab)->next_object(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/container.pxi":126
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         return self.next_object()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ContainerReader.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":58
 *     """
 * 
 *     cdef readonly object metadata             # <<<<<<<<<<<<<<
 *     cdef readonly const uint8_t[:] marker
 *     cdef readonly Schema writer_schema
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_8metadata_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerReader_8metadata_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_8metadata___get__(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_8metadata___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->metadata);
  __pyx_r = __pyx_v_self->metadata;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":59
 * 
 *     cdef readonly object metadata
 *     cdef readonly const uint8_t[:] marker             # <<<<<<<<<<<<<<
 *     cdef readonly Schema writer_schema
 *     cdef readonly Schema reader_schema
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_6marker_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerReader_6marker_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_6marker___get__(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_6marker___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_self->marker.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(25, 59, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->marker, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ContainerReader.marker.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":60
 *     cdef readonly object metadata
 *     cdef readonly const uint8_t[:] marker
 *     cdef readonly Schema writer_schema             # <<<<<<<<<<<<<<
 *     cdef readonly Schema reader_schema
 *     cdef readonly Schema schema
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_13writer_schema_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerReader_13writer_schema_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_13writer_schema___get__(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_13writer_schema___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->writer_schema);
  __pyx_r = ((PyObject *)__pyx_v_self->writer_schema);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":61
 *     cdef readonly const uint8_t[:] marker
 *     cdef readonly Schema writer_schema
 *     cdef readonly Schema reader_schema             # <<<<<<<<<<<<<<
 *     cdef readonly Schema schema
 *     cdef readonly str codec_name
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_13reader_schema_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerReader_13reader_schema_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_13reader_schema___get__(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_13reader_schema___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->reader_schema);
  __pyx_r = ((PyObject *)__pyx_v_self->reader_schema);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":62
 *     cdef readonly Schema writer_schema
 *     cdef readonly Schema reader_schema
 *     cdef readonly Schema schema             # <<<<<<<<<<<<<<
 *     cdef readonly str codec_name
 *     cdef readonly size_t objects_left_in_block
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_6schema_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerReader_6schema_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_6schema___get__(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_6schema___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->schema);
  __pyx_r = ((PyObject *)__pyx_v_self->schema);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":63
 *     cdef readonly Schema reader_schema
 *     cdef readonly Schema schema
 *     cdef readonly str codec_name             # <<<<<<<<<<<<<<
 *     cdef readonly size_t objects_left_in_block
 *     cdef MemoryReader current_block
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_10codec_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerReader_10codec_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_10codec_name___get__(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_10codec_name___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->codec_name);
  __pyx_r = __pyx_v_self->codec_name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":64
 *     cdef readonly Schema schema
 *     cdef readonly str codec_name
 *     cdef readonly size_t objects_left_in_block             # <<<<<<<<<<<<<<
 *     cdef MemoryReader current_block
 *     cdef Codec codec
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_21objects_left_in_block_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerReader_21objects_left_in_block_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_21objects_left_in_block___get__(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_21objects_left_in_block___get__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->objects_left_in_block); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ContainerReader.objects_left_in_block.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15ContainerReader_11__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerReader_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15ContainerReader_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_10__reduce_cython__(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_10__reduce_cython__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.codec, self.codec_name, self.current_block, self.marker, self.metadata, self.objects_left_in_block, self.reader, self.reader_schema, self.schema, self.writer_schema)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->marker.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->marker, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->objects_left_in_block); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(10); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_v_self->codec);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->codec);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self->codec))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->codec_name);
  __Pyx_GIVEREF(__pyx_v_self->codec_name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->codec_name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->current_block);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->current_block);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)__pyx_v_self->current_block))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->metadata);
  __Pyx_GIVEREF(__pyx_v_self->metadata);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_self->metadata)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->reader);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->reader);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 6, ((PyObject *)__pyx_v_self->reader))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->reader_schema);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->reader_schema);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 7, ((PyObject *)__pyx_v_self->reader_schema))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->schema);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->schema);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 8, ((PyObject *)__pyx_v_self->schema))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->writer_schema);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->writer_schema);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 9, ((PyObject *)__pyx_v_self->writer_schema))) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.codec, self.codec_name, self.current_block, self.marker, self.metadata, self.objects_left_in_block, self.reader, self.reader_schema, self.schema, self.writer_schema)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.codec, self.codec_name, self.current_block, self.marker, self.metadata, self.objects_left_in_block, self.reader, self.reader_schema, self.schema, self.writer_schema)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.codec is not None or self.codec_name is not None or self.current_block is not None or self.metadata is not None or self.reader is not None or self.reader_schema is not None or self.schema is not None or self.writer_schema is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.codec, self.codec_name, self.current_block, self.marker, self.metadata, self.objects_left_in_block, self.reader, self.reader_schema, self.schema, self.writer_schema)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.codec is not None or self.codec_name is not None or self.current_block is not None or self.metadata is not None or self.reader is not None or self.reader_schema is not None or self.schema is not None or self.writer_schema is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ContainerReader, (type(self), 0x289807d, None), state
 */
  /*else*/ {
    __pyx_t_5 = (((PyObject *)__pyx_v_self->codec) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->codec_name != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->current_block) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->metadata != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->reader) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->reader_schema) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->schema) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->writer_schema) != Py_None);
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.codec is not None or self.codec_name is not None or self.current_block is not None or self.metadata is not None or self.reader is not None or self.reader_schema is not None or self.schema is not None or self.writer_schema is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ContainerReader, (type(self), 0x289807d, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.codec is not None or self.codec_name is not None or self.current_block is not None or self.metadata is not None or self.reader is not None or self.reader_schema is not None or self.schema is not None or self.writer_schema is not None
 *     if use_setstate:
 *         return __pyx_unpickle_ContainerReader, (type(self), 0x289807d, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ContainerReader, (type(self), 0x289807d, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_ContainerReader); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_42565757);
    __Pyx_GIVEREF(__pyx_int_42565757);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_42565757)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.codec is not None or self.codec_name is not None or self.current_block is not None or self.metadata is not None or self.reader is not None or self.reader_schema is not None or self.schema is not None or self.writer_schema is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ContainerReader, (type(self), 0x289807d, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ContainerReader, (type(self), 0x289807d, None), state
 *     else:
 *         return __pyx_unpickle_ContainerReader, (type(self), 0x289807d, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ContainerReader__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_ContainerReader); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_42565757);
    __Pyx_GIVEREF(__pyx_int_42565757);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_42565757)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.ContainerReader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ContainerReader, (type(self), 0x289807d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ContainerReader__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerReader_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15ContainerReader_13__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerReader_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15ContainerReader_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ContainerReader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_15ContainerReader_12__setstate_cython__(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerReader_12__setstate_cython__(struct __pyx_obj_5cavro_ContainerReader *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ContainerReader, (type(self), 0x289807d, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ContainerReader__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_ContainerReader__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ContainerReader, (type(self), 0x289807d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ContainerReader__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ContainerReader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":171
 *     cdef readonly dict metadata
 * 
 *     def __cinit__(self, dest, Schema schema, str codec='null', size_t max_blocksize=16352, write_header=True, metadata=None, marker=None, options=DEFAULT_OPTIONS):             # <<<<<<<<<<<<<<
 *         if schema is None:
 *             raise ValueError('Schema is required')
 */

/* Python wrapper */
static int __pyx_pw_5cavro_15ContainerWriter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_15ContainerWriter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dest = 0;
  struct __pyx_obj_5cavro_Schema *__pyx_v_schema = 0;
  PyObject *__pyx_v_codec = 0;
  size_t __pyx_v_max_blocksize;
  PyObject *__pyx_v_write_header = 0;
  PyObject *__pyx_v_metadata = 0;
  PyObject *__pyx_v_marker = 0;
  PyObject *__pyx_v_options = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[8] = {0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dest,&__pyx_n_s_schema,&__pyx_n_s_codec,&__pyx_n_s_max_blocksize,&__pyx_n_s_write_header,&__pyx_n_s_metadata,&__pyx_n_s_marker,&__pyx_n_s_options,0};
    values[2] = __Pyx_Arg_NewRef_VARARGS(((PyObject*)__pyx_n_u_null_2));
    values[4] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));
    values[5] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    values[6] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    values[7] = __Pyx_Arg_NewRef_VARARGS(__pyx_k__89);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  8: values[7] = __Pyx_Arg_VARARGS(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_VARARGS(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dest)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 171, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_schema)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 171, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 8, 1); __PYX_ERR(25, 171, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_codec);
          if (value) { values[2] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 171, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_max_blocksize);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 171, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_write_header);
          if (value) { values[4] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 171, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_metadata);
          if (value) { values[5] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 171, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_marker);
          if (value) { values[6] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 171, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_options);
          if (value) { values[7] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 171, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(25, 171, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  8: values[7] = __Pyx_Arg_VARARGS(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_VARARGS(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_dest = values[0];
    __pyx_v_schema = ((struct __pyx_obj_5cavro_Schema *)values[1]);
    __pyx_v_codec = ((PyObject*)values[2]);
    if (values[3]) {
      __pyx_v_max_blocksize = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_max_blocksize == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(25, 171, __pyx_L3_error)
    } else {
      __pyx_v_max_blocksize = ((size_t)0x3FE0);
    }
    __pyx_v_write_header = values[4];
    __pyx_v_metadata = values[5];
    __pyx_v_marker = values[6];
    __pyx_v_options = values[7];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 8, __pyx_nargs); __PYX_ERR(25, 171, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ContainerWriter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_schema), __pyx_ptype_5cavro_Schema, 1, "schema", 0))) __PYX_ERR(25, 171, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_codec), (&PyUnicode_Type), 1, "codec", 1))) __PYX_ERR(25, 171, __pyx_L1_error)
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter___cinit__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self), __pyx_v_dest, __pyx_v_schema, __pyx_v_codec, __pyx_v_max_blocksize, __pyx_v_write_header, __pyx_v_metadata, __pyx_v_marker, __pyx_v_options);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_15ContainerWriter___cinit__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, PyObject *__pyx_v_dest, struct __pyx_obj_5cavro_Schema *__pyx_v_schema, PyObject *__pyx_v_codec, size_t __pyx_v_max_blocksize, PyObject *__pyx_v_write_header, PyObject *__pyx_v_metadata, PyObject *__pyx_v_marker, PyObject *__pyx_v_options) {
  PyObject *__pyx_v_codec_b = NULL;
  PyObject *__pyx_v_e = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  __Pyx_memviewslice __pyx_t_25 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_metadata);
  __Pyx_INCREF(__pyx_v_marker);

  /* "src/container.pxi":172
 * 
 *     def __cinit__(self, dest, Schema schema, str codec='null', size_t max_blocksize=16352, write_header=True, metadata=None, marker=None, options=DEFAULT_OPTIONS):
 *         if schema is None:             # <<<<<<<<<<<<<<
 *             raise ValueError('Schema is required')
 *         self.should_write_header = write_header
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_schema) == Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "src/container.pxi":173
 *     def __cinit__(self, dest, Schema schema, str codec='null', size_t max_blocksize=16352, write_header=True, metadata=None, marker=None, options=DEFAULT_OPTIONS):
 *         if schema is None:
 *             raise ValueError('Schema is required')             # <<<<<<<<<<<<<<
 *         self.should_write_header = write_header
 *         self.num_pending = 0
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__90, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(25, 173, __pyx_L1_error)

    /* "src/container.pxi":172
 * 
 *     def __cinit__(self, dest, Schema schema, str codec='null', size_t max_blocksize=16352, write_header=True, metadata=None, marker=None, options=DEFAULT_OPTIONS):
 *         if schema is None:             # <<<<<<<<<<<<<<
 *             raise ValueError('Schema is required')
 *         self.should_write_header = write_header
 */
  }

  /* "src/container.pxi":174
 *         if schema is None:
 *             raise ValueError('Schema is required')
 *         self.should_write_header = write_header             # <<<<<<<<<<<<<<
 *         self.num_pending = 0
 *         self.pending_block = MemoryWriter(max_blocksize)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_write_header); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(25, 174, __pyx_L1_error)
  __pyx_v_self->should_write_header = __pyx_t_1;

  /* "src/container.pxi":175
 *             raise ValueError('Schema is required')
 *         self.should_write_header = write_header
 *         self.num_pending = 0             # <<<<<<<<<<<<<<
 *         self.pending_block = MemoryWriter(max_blocksize)
 *         self.next_item = MemoryWriter()
 */
  __pyx_v_self->num_pending = 0;

  /* "src/container.pxi":176
 *         self.should_write_header = write_header
 *         self.num_pending = 0
 *         self.pending_block = MemoryWriter(max_blocksize)             # <<<<<<<<<<<<<<
 *         self.next_item = MemoryWriter()
 *         self.next_block = MemoryWriter()
 */
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_max_blocksize); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5cavro_MemoryWriter), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->pending_block);
  __Pyx_DECREF((PyObject *)__pyx_v_self->pending_block);
  __pyx_v_self->pending_block = ((struct __pyx_obj_5cavro_MemoryWriter *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/container.pxi":177
 *         self.num_pending = 0
 *         self.pending_block = MemoryWriter(max_blocksize)
 *         self.next_item = MemoryWriter()             # <<<<<<<<<<<<<<
 *         self.next_block = MemoryWriter()
 *         self.options = options
 */
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro_MemoryWriter)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->next_item);
  __Pyx_DECREF((PyObject *)__pyx_v_self->next_item);
  __pyx_v_self->next_item = ((struct __pyx_obj_5cavro_MemoryWriter *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/container.pxi":178
 *         self.pending_block = MemoryWriter(max_blocksize)
 *         self.next_item = MemoryWriter()
 *         self.next_block = MemoryWriter()             # <<<<<<<<<<<<<<
 *         self.options = options
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro_MemoryWriter)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->next_block);
  __Pyx_DECREF((PyObject *)__pyx_v_self->next_block);
  __pyx_v_self->next_block = ((struct __pyx_obj_5cavro_MemoryWriter *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/container.pxi":179
 *         self.next_item = MemoryWriter()
 *         self.next_block = MemoryWriter()
 *         self.options = options             # <<<<<<<<<<<<<<
 * 
 *         if metadata is None:
 */
  if (!(likely(((__pyx_v_options) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_options, __pyx_ptype_5cavro_Options))))) __PYX_ERR(25, 179, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_options;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->options);
  __Pyx_DECREF((PyObject *)__pyx_v_self->options);
  __pyx_v_self->options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/container.pxi":181
 *         self.options = options
 * 
 *         if metadata is None:             # <<<<<<<<<<<<<<
 *             metadata = {}
 *         self.metadata = metadata
 */
  __pyx_t_1 = (__pyx_v_metadata == Py_None);
  if (__pyx_t_1) {

    /* "src/container.pxi":182
 * 
 *         if metadata is None:
 *             metadata = {}             # <<<<<<<<<<<<<<
 *         self.metadata = metadata
 * 
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_metadata, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "src/container.pxi":181
 *         self.options = options
 * 
 *         if metadata is None:             # <<<<<<<<<<<<<<
 *             metadata = {}
 *         self.metadata = metadata
 */
  }

  /* "src/container.pxi":183
 *         if metadata is None:
 *             metadata = {}
 *         self.metadata = metadata             # <<<<<<<<<<<<<<
 * 
 *         self.writer = make_writer(dest)
 */
  if (!(likely(PyDict_CheckExact(__pyx_v_metadata))||((__pyx_v_metadata) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_metadata))) __PYX_ERR(25, 183, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_metadata;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->metadata);
  __Pyx_DECREF(__pyx_v_self->metadata);
  __pyx_v_self->metadata = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/container.pxi":185
 *         self.metadata = metadata
 * 
 *         self.writer = make_writer(dest)             # <<<<<<<<<<<<<<
 *         try:
 *             self.schema = schema
 */
  __pyx_t_3 = ((PyObject *)__pyx_f_5cavro_make_writer(__pyx_v_dest)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->writer);
  __Pyx_DECREF((PyObject *)__pyx_v_self->writer);
  __pyx_v_self->writer = ((struct __pyx_obj_5cavro__Writer *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/container.pxi":186
 * 
 *         self.writer = make_writer(dest)
 *         try:             # <<<<<<<<<<<<<<
 *             self.schema = schema
 *             codec_b = codec.encode('utf8')
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "src/container.pxi":187
 *         self.writer = make_writer(dest)
 *         try:
 *             self.schema = schema             # <<<<<<<<<<<<<<
 *             codec_b = codec.encode('utf8')
 *             try:
 */
      __Pyx_INCREF((PyObject *)__pyx_v_schema);
      __Pyx_GIVEREF((PyObject *)__pyx_v_schema);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->schema);
      __Pyx_DECREF((PyObject *)__pyx_v_self->schema);
      __pyx_v_self->schema = __pyx_v_schema;

      /* "src/container.pxi":188
 *         try:
 *             self.schema = schema
 *             codec_b = codec.encode('utf8')             # <<<<<<<<<<<<<<
 *             try:
 *                 self.codec = CODECS[codec_b]
 */
      if (unlikely(__pyx_v_codec == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
        __PYX_ERR(25, 188, __pyx_L5_error)
      }
      __pyx_t_3 = PyUnicode_AsUTF8String(__pyx_v_codec); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 188, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_codec_b = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "src/container.pxi":189
 *             self.schema = schema
 *             codec_b = codec.encode('utf8')
 *             try:             # <<<<<<<<<<<<<<
 *                 self.codec = CODECS[codec_b]
 *             except KeyError as e:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "src/container.pxi":190
 *             codec_b = codec.encode('utf8')
 *             try:
 *                 self.codec = CODECS[codec_b]             # <<<<<<<<<<<<<<
 *             except KeyError as e:
 *                 raise CodecUnavailable(codec) from e
 */
          __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CODECS); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 190, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_v_codec_b); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 190, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5cavro_Codec))))) __PYX_ERR(25, 190, __pyx_L11_error)
          __Pyx_GIVEREF(__pyx_t_2);
          __Pyx_GOTREF((PyObject *)__pyx_v_self->codec);
          __Pyx_DECREF((PyObject *)__pyx_v_self->codec);
          __pyx_v_self->codec = ((struct __pyx_obj_5cavro_Codec *)__pyx_t_2);
          __pyx_t_2 = 0;

          /* "src/container.pxi":189
 *             self.schema = schema
 *             codec_b = codec.encode('utf8')
 *             try:             # <<<<<<<<<<<<<<
 *                 self.codec = CODECS[codec_b]
 *             except KeyError as e:
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L16_try_end;
        __pyx_L11_error:;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "src/container.pxi":191
 *             try:
 *                 self.codec = CODECS[codec_b]
 *             except KeyError as e:             # <<<<<<<<<<<<<<
 *                 raise CodecUnavailable(codec) from e
 *             if marker is None:
 */
        __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
        if (__pyx_t_10) {
          __Pyx_AddTraceback("cavro.ContainerWriter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_11) < 0) __PYX_ERR(25, 191, __pyx_L13_except_error)
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_3);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_3);
          __pyx_v_e = __pyx_t_3;
          /*try:*/ {

            /* "src/container.pxi":192
 *                 self.codec = CODECS[codec_b]
 *             except KeyError as e:
 *                 raise CodecUnavailable(codec) from e             # <<<<<<<<<<<<<<
 *             if marker is None:
 *                 marker = uuid.uuid4().bytes
 */
            __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_CodecUnavailable); if (unlikely(!__pyx_t_13)) __PYX_ERR(25, 192, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_14 = NULL;
            __pyx_t_10 = 0;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_13))) {
              __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
              if (likely(__pyx_t_14)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                __Pyx_INCREF(__pyx_t_14);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_13, function);
                __pyx_t_10 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_14, __pyx_v_codec};
              __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              if (unlikely(!__pyx_t_12)) __PYX_ERR(25, 192, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            }
            __Pyx_Raise(__pyx_t_12, 0, 0, __pyx_v_e);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __PYX_ERR(25, 192, __pyx_L22_error)
          }

          /* "src/container.pxi":191
 *             try:
 *                 self.codec = CODECS[codec_b]
 *             except KeyError as e:             # <<<<<<<<<<<<<<
 *                 raise CodecUnavailable(codec) from e
 *             if marker is None:
 */
          /*finally:*/ {
            __pyx_L22_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_22);
              __pyx_t_10 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_XGIVEREF(__pyx_t_22);
                __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
              }
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
              __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
              __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
              goto __pyx_L13_except_error;
            }
          }
        }
        goto __pyx_L13_except_error;

        /* "src/container.pxi":189
 *             self.schema = schema
 *             codec_b = codec.encode('utf8')
 *             try:             # <<<<<<<<<<<<<<
 *                 self.codec = CODECS[codec_b]
 *             except KeyError as e:
 */
        __pyx_L13_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L5_error;
        __pyx_L16_try_end:;
      }

      /* "src/container.pxi":193
 *             except KeyError as e:
 *                 raise CodecUnavailable(codec) from e
 *             if marker is None:             # <<<<<<<<<<<<<<
 *                 marker = uuid.uuid4().bytes
 *             else:
 */
      __pyx_t_1 = (__pyx_v_marker == Py_None);
      if (__pyx_t_1) {

        /* "src/container.pxi":194
 *                 raise CodecUnavailable(codec) from e
 *             if marker is None:
 *                 marker = uuid.uuid4().bytes             # <<<<<<<<<<<<<<
 *             else:
 *                 if not isinstance(marker, (bytes, _SLICE_TYPE)):
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_uuid); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 194, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_uuid4); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 194, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        __pyx_t_15 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_15 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_15, 0+__pyx_t_15);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(25, 194, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_bytes_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 194, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF_SET(__pyx_v_marker, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "src/container.pxi":193
 *             except KeyError as e:
 *                 raise CodecUnavailable(codec) from e
 *             if marker is None:             # <<<<<<<<<<<<<<
 *                 marker = uuid.uuid4().bytes
 *             else:
 */
        goto __pyx_L28;
      }

      /* "src/container.pxi":196
 *                 marker = uuid.uuid4().bytes
 *             else:
 *                 if not isinstance(marker, (bytes, _SLICE_TYPE)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(f'Marker must be bytes, got: {type(marker)}')
 *                 if len(marker) != 16:
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SLICE_TYPE); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 196, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_23 = PyBytes_Check(__pyx_v_marker); 
        if (!__pyx_t_23) {
        } else {
          __pyx_t_1 = __pyx_t_23;
          goto __pyx_L30_bool_binop_done;
        }
        __pyx_t_23 = PyObject_IsInstance(__pyx_v_marker, __pyx_t_2); 
        __pyx_t_1 = __pyx_t_23;
        __pyx_L30_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_23 = (!__pyx_t_1);
        if (unlikely(__pyx_t_23)) {

          /* "src/container.pxi":197
 *             else:
 *                 if not isinstance(marker, (bytes, _SLICE_TYPE)):
 *                     raise ValueError(f'Marker must be bytes, got: {type(marker)}')             # <<<<<<<<<<<<<<
 *                 if len(marker) != 16:
 *                     raise ValueError(f'Marker must be exactly 16 bytes, got: {bytes(marker)}')
 */
          __pyx_t_2 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_marker)), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 197, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_11 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Marker_must_be_bytes_got, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(25, 197, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 197, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_Raise(__pyx_t_2, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __PYX_ERR(25, 197, __pyx_L5_error)

          /* "src/container.pxi":196
 *                 marker = uuid.uuid4().bytes
 *             else:
 *                 if not isinstance(marker, (bytes, _SLICE_TYPE)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(f'Marker must be bytes, got: {type(marker)}')
 *                 if len(marker) != 16:
 */
        }

        /* "src/container.pxi":198
 *                 if not isinstance(marker, (bytes, _SLICE_TYPE)):
 *                     raise ValueError(f'Marker must be bytes, got: {type(marker)}')
 *                 if len(marker) != 16:             # <<<<<<<<<<<<<<
 *                     raise ValueError(f'Marker must be exactly 16 bytes, got: {bytes(marker)}')
 *             self.marker = marker
 */
        __pyx_t_24 = PyObject_Length(__pyx_v_marker); if (unlikely(__pyx_t_24 == ((Py_ssize_t)-1))) __PYX_ERR(25, 198, __pyx_L5_error)
        __pyx_t_23 = (__pyx_t_24 != 16);
        if (unlikely(__pyx_t_23)) {

          /* "src/container.pxi":199
 *                     raise ValueError(f'Marker must be bytes, got: {type(marker)}')
 *                 if len(marker) != 16:
 *                     raise ValueError(f'Marker must be exactly 16 bytes, got: {bytes(marker)}')             # <<<<<<<<<<<<<<
 *             self.marker = marker
 *             self.max_blocksize = max_blocksize
 */
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_v_marker); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 199, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(25, 199, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Marker_must_be_exactly_16_bytes, __pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 199, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(25, 199, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_Raise(__pyx_t_11, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __PYX_ERR(25, 199, __pyx_L5_error)

          /* "src/container.pxi":198
 *                 if not isinstance(marker, (bytes, _SLICE_TYPE)):
 *                     raise ValueError(f'Marker must be bytes, got: {type(marker)}')
 *                 if len(marker) != 16:             # <<<<<<<<<<<<<<
 *                     raise ValueError(f'Marker must be exactly 16 bytes, got: {bytes(marker)}')
 *             self.marker = marker
 */
        }
      }
      __pyx_L28:;

      /* "src/container.pxi":200
 *                 if len(marker) != 16:
 *                     raise ValueError(f'Marker must be exactly 16 bytes, got: {bytes(marker)}')
 *             self.marker = marker             # <<<<<<<<<<<<<<
 *             self.max_blocksize = max_blocksize
 *             self.blocks_written = 0
 */
      __pyx_t_25 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_v_marker, 0); if (unlikely(!__pyx_t_25.memview)) __PYX_ERR(25, 200, __pyx_L5_error)
      __PYX_XCLEAR_MEMVIEW(&__pyx_v_self->marker, 0);
      __pyx_v_self->marker = __pyx_t_25;
      __pyx_t_25.memview = NULL;
      __pyx_t_25.data = NULL;

      /* "src/container.pxi":201
 *                     raise ValueError(f'Marker must be exactly 16 bytes, got: {bytes(marker)}')
 *             self.marker = marker
 *             self.max_blocksize = max_blocksize             # <<<<<<<<<<<<<<
 *             self.blocks_written = 0
 *         except:
 */
      __pyx_v_self->max_blocksize = __pyx_v_max_blocksize;

      /* "src/container.pxi":202
 *             self.marker = marker
 *             self.max_blocksize = max_blocksize
 *             self.blocks_written = 0             # <<<<<<<<<<<<<<
 *         except:
 *             self.writer = None
 */
      __pyx_v_self->blocks_written = 0;

      /* "src/container.pxi":186
 * 
 *         self.writer = make_writer(dest)
 *         try:             # <<<<<<<<<<<<<<
 *             self.schema = schema
 *             codec_b = codec.encode('utf8')
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L10_try_end;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_XCLEAR_MEMVIEW(&__pyx_t_25, 1);
    __pyx_t_25.memview = NULL; __pyx_t_25.data = NULL;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "src/container.pxi":203
 *             self.max_blocksize = max_blocksize
 *             self.blocks_written = 0
 *         except:             # <<<<<<<<<<<<<<
 *             self.writer = None
 *             raise
 */
    /*except:*/ {
      __Pyx_AddTraceback("cavro.ContainerWriter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_11, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(25, 203, __pyx_L7_except_error)
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);

      /* "src/container.pxi":204
 *             self.blocks_written = 0
 *         except:
 *             self.writer = None             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->writer);
      __Pyx_DECREF((PyObject *)__pyx_v_self->writer);
      __pyx_v_self->writer = ((struct __pyx_obj_5cavro__Writer *)Py_None);

      /* "src/container.pxi":205
 *         except:
 *             self.writer = None
 *             raise             # <<<<<<<<<<<<<<
 * 
 *     def __enter__(self):
 */
      __Pyx_GIVEREF(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ErrRestoreWithState(__pyx_t_11, __pyx_t_2, __pyx_t_3);
      __pyx_t_11 = 0; __pyx_t_2 = 0; __pyx_t_3 = 0; 
      __PYX_ERR(25, 205, __pyx_L7_except_error)
    }

    /* "src/container.pxi":186
 * 
 *         self.writer = make_writer(dest)
 *         try:             # <<<<<<<<<<<<<<
 *             self.schema = schema
 *             codec_b = codec.encode('utf8')
 */
    __pyx_L7_except_error:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L10_try_end:;
  }

  /* "src/container.pxi":171
 *     cdef readonly dict metadata
 * 
 *     def __cinit__(self, dest, Schema schema, str codec='null', size_t max_blocksize=16352, write_header=True, metadata=None, marker=None, options=DEFAULT_OPTIONS):             # <<<<<<<<<<<<<<
 *         if schema is None:
 *             raise ValueError('Schema is required')
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_25, 1);
  __Pyx_AddTraceback("cavro.ContainerWriter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_codec_b);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_metadata);
  __Pyx_XDECREF(__pyx_v_marker);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":207
 *             raise
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         if self.writer is None:
 *             raise ValueError('Container is closed')
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_3__enter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15ContainerWriter_3__enter__ = {"__enter__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_3__enter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15ContainerWriter_3__enter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__enter__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__enter__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__enter__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_2__enter__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_2__enter__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__enter__", 1);

  /* "src/container.pxi":208
 * 
 *     def __enter__(self):
 *         if self.writer is None:             # <<<<<<<<<<<<<<
 *             raise ValueError('Container is closed')
 *         return self
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->writer) == Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "src/container.pxi":209
 *     def __enter__(self):
 *         if self.writer is None:
 *             raise ValueError('Container is closed')             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__91, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(25, 209, __pyx_L1_error)

    /* "src/container.pxi":208
 * 
 *     def __enter__(self):
 *         if self.writer is None:             # <<<<<<<<<<<<<<
 *             raise ValueError('Container is closed')
 *         return self
 */
  }

  /* "src/container.pxi":210
 *         if self.writer is None:
 *             raise ValueError('Container is closed')
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __exit__(self, exc_type, exc_val, exc_tb):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "src/container.pxi":207
 *             raise
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         if self.writer is None:
 *             raise ValueError('Container is closed')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.ContainerWriter.__enter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":212
 *         return self
 * 
 *     def __exit__(self, exc_type, exc_val, exc_tb):             # <<<<<<<<<<<<<<
 *         if self.writer is not None:
 *             self.close()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_5__exit__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15ContainerWriter_5__exit__ = {"__exit__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_5__exit__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15ContainerWriter_5__exit__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_exc_type = 0;
  CYTHON_UNUSED PyObject *__pyx_v_exc_val = 0;
  CYTHON_UNUSED PyObject *__pyx_v_exc_tb = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__exit__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_exc_type,&__pyx_n_s_exc_val,&__pyx_n_s_exc_tb,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exc_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 212, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exc_val)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 212, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__exit__", 1, 3, 3, 1); __PYX_ERR(25, 212, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exc_tb)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 212, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__exit__", 1, 3, 3, 2); __PYX_ERR(25, 212, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__exit__") < 0)) __PYX_ERR(25, 212, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_exc_type = values[0];
    __pyx_v_exc_val = values[1];
    __pyx_v_exc_tb = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__exit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(25, 212, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ContainerWriter.__exit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_4__exit__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self), __pyx_v_exc_type, __pyx_v_exc_val, __pyx_v_exc_tb);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_4__exit__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_exc_type, CYTHON_UNUSED PyObject *__pyx_v_exc_val, CYTHON_UNUSED PyObject *__pyx_v_exc_tb) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__exit__", 1);

  /* "src/container.pxi":213
 * 
 *     def __exit__(self, exc_type, exc_val, exc_tb):
 *         if self.writer is not None:             # <<<<<<<<<<<<<<
 *             self.close()
 *         return False
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->writer) != Py_None);
  if (__pyx_t_1) {

    /* "src/container.pxi":214
 *     def __exit__(self, exc_type, exc_val, exc_tb):
 *         if self.writer is not None:
 *             self.close()             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_close); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 214, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "src/container.pxi":213
 * 
 *     def __exit__(self, exc_type, exc_val, exc_tb):
 *         if self.writer is not None:             # <<<<<<<<<<<<<<
 *             self.close()
 *         return False
 */
  }

  /* "src/container.pxi":215
 *         if self.writer is not None:
 *             self.close()
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __del__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "src/container.pxi":212
 *         return self
 * 
 *     def __exit__(self, exc_type, exc_val, exc_tb):             # <<<<<<<<<<<<<<
 *         if self.writer is not None:
 *             self.close()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.ContainerWriter.__exit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":217
 *         return False
 * 
 *     def __del__(self):             # <<<<<<<<<<<<<<
 *         if self.writer is not None:
 *             try:
 */

/* Python wrapper */
static void __pyx_pw_5cavro_15ContainerWriter_7__del__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_5cavro_15ContainerWriter_7__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_5cavro_15ContainerWriter_6__del__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5cavro_15ContainerWriter_6__del__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_v_e = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__del__", 1);

  /* "src/container.pxi":218
 * 
 *     def __del__(self):
 *         if self.writer is not None:             # <<<<<<<<<<<<<<
 *             try:
 *                 self.close()
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->writer) != Py_None);
  if (__pyx_t_1) {

    /* "src/container.pxi":219
 *     def __del__(self):
 *         if self.writer is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 self.close()
 *             except ValueError as e:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "src/container.pxi":220
 *         if self.writer is not None:
 *             try:
 *                 self.close()             # <<<<<<<<<<<<<<
 *             except ValueError as e:
 *                 warnings.warn(f'Error closing container file during __del__: {e}', ResourceWarning)
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_close); if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 220, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 220, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "src/container.pxi":219
 *     def __del__(self):
 *         if self.writer is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 self.close()
 *             except ValueError as e:
 */
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "src/container.pxi":221
 *             try:
 *                 self.close()
 *             except ValueError as e:             # <<<<<<<<<<<<<<
 *                 warnings.warn(f'Error closing container file during __del__: {e}', ResourceWarning)
 *                 pass # We might be shutting down, and our writer might have been closed
 */
      __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
      if (__pyx_t_8) {
        __Pyx_AddTraceback("cavro.ContainerWriter.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(25, 221, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __pyx_v_e = __pyx_t_6;
        /*try:*/ {

          /* "src/container.pxi":222
 *                 self.close()
 *             except ValueError as e:
 *                 warnings.warn(f'Error closing container file during __del__: {e}', ResourceWarning)             # <<<<<<<<<<<<<<
 *                 pass # We might be shutting down, and our writer might have been closed
 * 
 */
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_warnings); if (unlikely(!__pyx_t_10)) __PYX_ERR(25, 222, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_warn); if (unlikely(!__pyx_t_11)) __PYX_ERR(25, 222, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = __Pyx_PyObject_FormatSimple(__pyx_v_e, __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(25, 222, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Error_closing_container_file_dur, __pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(25, 222, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_ResourceWarning); if (unlikely(!__pyx_t_10)) __PYX_ERR(25, 222, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_13 = NULL;
          __pyx_t_8 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_8 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_13, __pyx_t_12, __pyx_t_10};
            __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (unlikely(!__pyx_t_9)) __PYX_ERR(25, 222, __pyx_L15_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }

        /* "src/container.pxi":221
 *             try:
 *                 self.close()
 *             except ValueError as e:             # <<<<<<<<<<<<<<
 *                 warnings.warn(f'Error closing container file during __del__: {e}', ResourceWarning)
 *                 pass # We might be shutting down, and our writer might have been closed
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            goto __pyx_L16;
          }
          __pyx_L15_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_8 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L6_except_error;
          }
          __pyx_L16:;
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "src/container.pxi":219
 *     def __del__(self):
 *         if self.writer is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 self.close()
 *             except ValueError as e:
 */
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "src/container.pxi":218
 * 
 *     def __del__(self):
 *         if self.writer is not None:             # <<<<<<<<<<<<<<
 *             try:
 *                 self.close()
 */
  }

  /* "src/container.pxi":217
 *         return False
 * 
 *     def __del__(self):             # <<<<<<<<<<<<<<
 *         if self.writer is not None:
 *             try:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_WriteUnraisable("cavro.ContainerWriter.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
}

/* "src/container.pxi":225
 *                 pass # We might be shutting down, and our writer might have been closed
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def closed(self):
 *         return self.writer is None
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_6closed_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerWriter_6closed_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_6closed___get__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_6closed___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "src/container.pxi":227
 *     @property
 *     def closed(self):
 *         return self.writer is None             # <<<<<<<<<<<<<<
 * 
 *     cdef int _write_header(self) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = (((PyObject *)__pyx_v_self->writer) == Py_None);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/container.pxi":225
 *                 pass # We might be shutting down, and our writer might have been closed
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def closed(self):
 *         return self.writer is None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.ContainerWriter.closed.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":229
 *         return self.writer is None
 * 
 *     cdef int _write_header(self) except -1:             # <<<<<<<<<<<<<<
 *         self.writer.write_n(OBJ_MAGIC)
 *         meta = self.metadata.copy()
 */

static int __pyx_f_5cavro_15ContainerWriter__write_header(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_v_meta = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_write_header", 1);

  /* "src/container.pxi":230
 * 
 *     cdef int _write_header(self) except -1:
 *         self.writer.write_n(OBJ_MAGIC)             # <<<<<<<<<<<<<<
 *         meta = self.metadata.copy()
 *         meta['avro.schema'] = json.dumps(self.schema.schema)
 */
  if (unlikely(!__pyx_v_5cavro_OBJ_MAGIC.memview)) { __Pyx_RaiseUnboundLocalError("OBJ_MAGIC"); __PYX_ERR(25, 230, __pyx_L1_error) }
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_self->writer->__pyx_vtab)->write_n(__pyx_v_self->writer, __pyx_v_5cavro_OBJ_MAGIC); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(25, 230, __pyx_L1_error)

  /* "src/container.pxi":231
 *     cdef int _write_header(self) except -1:
 *         self.writer.write_n(OBJ_MAGIC)
 *         meta = self.metadata.copy()             # <<<<<<<<<<<<<<
 *         meta['avro.schema'] = json.dumps(self.schema.schema)
 *         meta['avro.codec'] = self.codec.name
 */
  if (unlikely(__pyx_v_self->metadata == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "copy");
    __PYX_ERR(25, 231, __pyx_L1_error)
  }
  __pyx_t_2 = PyDict_Copy(__pyx_v_self->metadata); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_meta = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/container.pxi":232
 *         self.writer.write_n(OBJ_MAGIC)
 *         meta = self.metadata.copy()
 *         meta['avro.schema'] = json.dumps(self.schema.schema)             # <<<<<<<<<<<<<<
 *         meta['avro.codec'] = self.codec.name
 *         OBJ_FILE_METADATA.binary_write(self.writer, meta)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_json); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_dumps); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->schema), __pyx_n_s_schema); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_1 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_1 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_1, 1+__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  if (unlikely(__pyx_v_meta == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(25, 232, __pyx_L1_error)
  }
  if (unlikely((PyDict_SetItem(__pyx_v_meta, __pyx_kp_u_avro_schema, __pyx_t_2) < 0))) __PYX_ERR(25, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/container.pxi":233
 *         meta = self.metadata.copy()
 *         meta['avro.schema'] = json.dumps(self.schema.schema)
 *         meta['avro.codec'] = self.codec.name             # <<<<<<<<<<<<<<
 *         OBJ_FILE_METADATA.binary_write(self.writer, meta)
 *         self.writer.write_n(self.marker)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->codec), __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(__pyx_v_meta == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(25, 233, __pyx_L1_error)
  }
  if (unlikely((PyDict_SetItem(__pyx_v_meta, __pyx_kp_u_avro_codec, __pyx_t_2) < 0))) __PYX_ERR(25, 233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/container.pxi":234
 *         meta['avro.schema'] = json.dumps(self.schema.schema)
 *         meta['avro.codec'] = self.codec.name
 *         OBJ_FILE_METADATA.binary_write(self.writer, meta)             # <<<<<<<<<<<<<<
 *         self.writer.write_n(self.marker)
 *         return 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_OBJ_FILE_METADATA); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_binary_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_1 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_1 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_v_self->writer), __pyx_v_meta};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_1, 2+__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/container.pxi":235
 *         meta['avro.codec'] = self.codec.name
 *         OBJ_FILE_METADATA.binary_write(self.writer, meta)
 *         self.writer.write_n(self.marker)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  if (unlikely(!__pyx_v_self->marker.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(25, 235, __pyx_L1_error)}
  __pyx_t_1 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_self->writer->__pyx_vtab)->write_n(__pyx_v_self->writer, __pyx_v_self->marker); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(25, 235, __pyx_L1_error)

  /* "src/container.pxi":236
 *         OBJ_FILE_METADATA.binary_write(self.writer, meta)
 *         self.writer.write_n(self.marker)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "src/container.pxi":229
 *         return self.writer is None
 * 
 *     cdef int _write_header(self) except -1:             # <<<<<<<<<<<<<<
 *         self.writer.write_n(OBJ_MAGIC)
 *         meta = self.metadata.copy()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.ContainerWriter._write_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_meta);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":238
 *         return 0
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         if self.writer is None:
 *             raise ValueError('Trying to close a closed Container')
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_9close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15ContainerWriter_9close = {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_9close, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15ContainerWriter_9close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "close", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_8close(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_8close(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  CYTHON_UNUSED struct __pyx_obj_5cavro__Writer *__pyx_v_writer = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  struct __pyx_opt_args_5cavro_15ContainerWriter__flush_block __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 1);

  /* "src/container.pxi":239
 * 
 *     def close(self):
 *         if self.writer is None:             # <<<<<<<<<<<<<<
 *             raise ValueError('Trying to close a closed Container')
 *         cdef _Writer writer = self.writer
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->writer) == Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "src/container.pxi":240
 *     def close(self):
 *         if self.writer is None:
 *             raise ValueError('Trying to close a closed Container')             # <<<<<<<<<<<<<<
 *         cdef _Writer writer = self.writer
 *         self._flush_block(self.blocks_written == 0)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__92, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 240, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(25, 240, __pyx_L1_error)

    /* "src/container.pxi":239
 * 
 *     def close(self):
 *         if self.writer is None:             # <<<<<<<<<<<<<<
 *             raise ValueError('Trying to close a closed Container')
 *         cdef _Writer writer = self.writer
 */
  }

  /* "src/container.pxi":241
 *         if self.writer is None:
 *             raise ValueError('Trying to close a closed Container')
 *         cdef _Writer writer = self.writer             # <<<<<<<<<<<<<<
 *         self._flush_block(self.blocks_written == 0)
 *         self.writer = None
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_self->writer);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_writer = ((struct __pyx_obj_5cavro__Writer *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "src/container.pxi":242
 *             raise ValueError('Trying to close a closed Container')
 *         cdef _Writer writer = self.writer
 *         self._flush_block(self.blocks_written == 0)             # <<<<<<<<<<<<<<
 *         self.writer = None
 *         self.next_item = None
 */
  __pyx_t_4.__pyx_n = 1;
  __pyx_t_4.force = (__pyx_v_self->blocks_written == 0);
  __pyx_t_3 = ((struct __pyx_vtabstruct_5cavro_ContainerWriter *)__pyx_v_self->__pyx_vtab)->_flush_block(__pyx_v_self, &__pyx_t_4); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(25, 242, __pyx_L1_error)

  /* "src/container.pxi":243
 *         cdef _Writer writer = self.writer
 *         self._flush_block(self.blocks_written == 0)
 *         self.writer = None             # <<<<<<<<<<<<<<
 *         self.next_item = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->writer);
  __Pyx_DECREF((PyObject *)__pyx_v_self->writer);
  __pyx_v_self->writer = ((struct __pyx_obj_5cavro__Writer *)Py_None);

  /* "src/container.pxi":244
 *         self._flush_block(self.blocks_written == 0)
 *         self.writer = None
 *         self.next_item = None             # <<<<<<<<<<<<<<
 * 
 *     cdef int _flush_block(self, int force=False) except -1:
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->next_item);
  __Pyx_DECREF((PyObject *)__pyx_v_self->next_item);
  __pyx_v_self->next_item = ((struct __pyx_obj_5cavro_MemoryWriter *)Py_None);

  /* "src/container.pxi":238
 *         return 0
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         if self.writer is None:
 *             raise ValueError('Trying to close a closed Container')
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.ContainerWriter.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_writer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":246
 *         self.next_item = None
 * 
 *     cdef int _flush_block(self, int force=False) except -1:             # <<<<<<<<<<<<<<
 *         cdef size_t pending_len = self.pending_block.len
 *         if force or self.num_pending > 0:
 */

static int __pyx_f_5cavro_15ContainerWriter__flush_block(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, struct __pyx_opt_args_5cavro_15ContainerWriter__flush_block *__pyx_optional_args) {
  int __pyx_v_force = ((int)0);
  CYTHON_UNUSED size_t __pyx_v_pending_len;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_flush_block", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_force = __pyx_optional_args->force;
    }
  }

  /* "src/container.pxi":247
 * 
 *     cdef int _flush_block(self, int force=False) except -1:
 *         cdef size_t pending_len = self.pending_block.len             # <<<<<<<<<<<<<<
 *         if force or self.num_pending > 0:
 *             if self.blocks_written == 0 and self.should_write_header:
 */
  __pyx_t_1 = __pyx_v_self->pending_block->len;
  __pyx_v_pending_len = __pyx_t_1;

  /* "src/container.pxi":248
 *     cdef int _flush_block(self, int force=False) except -1:
 *         cdef size_t pending_len = self.pending_block.len
 *         if force or self.num_pending > 0:             # <<<<<<<<<<<<<<
 *             if self.blocks_written == 0 and self.should_write_header:
 *                 self._write_header()
 */
  __pyx_t_3 = (__pyx_v_force != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_self->num_pending > 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "src/container.pxi":249
 *         cdef size_t pending_len = self.pending_block.len
 *         if force or self.num_pending > 0:
 *             if self.blocks_written == 0 and self.should_write_header:             # <<<<<<<<<<<<<<
 *                 self._write_header()
 *             zigzag_encode_long(self.writer, self.num_pending)
 */
    __pyx_t_3 = (__pyx_v_self->blocks_written == 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_2 = __pyx_v_self->should_write_header;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_2) {

      /* "src/container.pxi":250
 *         if force or self.num_pending > 0:
 *             if self.blocks_written == 0 and self.should_write_header:
 *                 self._write_header()             # <<<<<<<<<<<<<<
 *             zigzag_encode_long(self.writer, self.num_pending)
 *             self.next_block.reset()
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro_ContainerWriter *)__pyx_v_self->__pyx_vtab)->_write_header(__pyx_v_self); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(25, 250, __pyx_L1_error)

      /* "src/container.pxi":249
 *         cdef size_t pending_len = self.pending_block.len
 *         if force or self.num_pending > 0:
 *             if self.blocks_written == 0 and self.should_write_header:             # <<<<<<<<<<<<<<
 *                 self._write_header()
 *             zigzag_encode_long(self.writer, self.num_pending)
 */
    }

    /* "src/container.pxi":251
 *             if self.blocks_written == 0 and self.should_write_header:
 *                 self._write_header()
 *             zigzag_encode_long(self.writer, self.num_pending)             # <<<<<<<<<<<<<<
 *             self.next_block.reset()
 *             self.codec.write_block(self.next_block, self.pending_block.view())
 */
    __pyx_t_5 = ((PyObject *)__pyx_v_self->writer);
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_4 = __pyx_f_5cavro_zigzag_encode_long(((struct __pyx_obj_5cavro__Writer *)__pyx_t_5), __pyx_v_self->num_pending); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(25, 251, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "src/container.pxi":252
 *                 self._write_header()
 *             zigzag_encode_long(self.writer, self.num_pending)
 *             self.next_block.reset()             # <<<<<<<<<<<<<<
 *             self.codec.write_block(self.next_block, self.pending_block.view())
 *             zigzag_encode_long(self.writer, self.next_block.len)
 */
    __pyx_t_5 = __pyx_f_5cavro_12MemoryWriter_reset(__pyx_v_self->next_block); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "src/container.pxi":253
 *             zigzag_encode_long(self.writer, self.num_pending)
 *             self.next_block.reset()
 *             self.codec.write_block(self.next_block, self.pending_block.view())             # <<<<<<<<<<<<<<
 *             zigzag_encode_long(self.writer, self.next_block.len)
 *             if self.next_block.len > 0:
 */
    __pyx_t_5 = ((PyObject *)__pyx_v_self->next_block);
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_6 = __pyx_f_5cavro_12MemoryWriter_view(__pyx_v_self->pending_block); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(25, 253, __pyx_L1_error)
    __pyx_t_7 = ((struct __pyx_vtabstruct_5cavro_Codec *)__pyx_v_self->codec->__pyx_vtab)->write_block(__pyx_v_self->codec, ((struct __pyx_obj_5cavro__Writer *)__pyx_t_5), __pyx_t_6); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1L))) __PYX_ERR(25, 253, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
    __pyx_t_6.memview = NULL; __pyx_t_6.data = NULL;

    /* "src/container.pxi":254
 *             self.next_block.reset()
 *             self.codec.write_block(self.next_block, self.pending_block.view())
 *             zigzag_encode_long(self.writer, self.next_block.len)             # <<<<<<<<<<<<<<
 *             if self.next_block.len > 0:
 *                 self.writer.write_n(self.next_block.view())
 */
    __pyx_t_5 = ((PyObject *)__pyx_v_self->writer);
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_4 = __pyx_f_5cavro_zigzag_encode_long(((struct __pyx_obj_5cavro__Writer *)__pyx_t_5), __pyx_v_self->next_block->len); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(25, 254, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "src/container.pxi":255
 *             self.codec.write_block(self.next_block, self.pending_block.view())
 *             zigzag_encode_long(self.writer, self.next_block.len)
 *             if self.next_block.len > 0:             # <<<<<<<<<<<<<<
 *                 self.writer.write_n(self.next_block.view())
 *             self.writer.write_n(self.marker)
 */
    __pyx_t_2 = (__pyx_v_self->next_block->len > 0);
    if (__pyx_t_2) {

      /* "src/container.pxi":256
 *             zigzag_encode_long(self.writer, self.next_block.len)
 *             if self.next_block.len > 0:
 *                 self.writer.write_n(self.next_block.view())             # <<<<<<<<<<<<<<
 *             self.writer.write_n(self.marker)
 *             self.blocks_written += 1
 */
      __pyx_t_6 = __pyx_f_5cavro_12MemoryWriter_view(__pyx_v_self->next_block); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(25, 256, __pyx_L1_error)
      __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_self->writer->__pyx_vtab)->write_n(__pyx_v_self->writer, __pyx_t_6); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(25, 256, __pyx_L1_error)
      __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
      __pyx_t_6.memview = NULL; __pyx_t_6.data = NULL;

      /* "src/container.pxi":255
 *             self.codec.write_block(self.next_block, self.pending_block.view())
 *             zigzag_encode_long(self.writer, self.next_block.len)
 *             if self.next_block.len > 0:             # <<<<<<<<<<<<<<
 *                 self.writer.write_n(self.next_block.view())
 *             self.writer.write_n(self.marker)
 */
    }

    /* "src/container.pxi":257
 *             if self.next_block.len > 0:
 *                 self.writer.write_n(self.next_block.view())
 *             self.writer.write_n(self.marker)             # <<<<<<<<<<<<<<
 *             self.blocks_written += 1
 *             self.pending_block.reset()
 */
    if (unlikely(!__pyx_v_self->marker.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(25, 257, __pyx_L1_error)}
    __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_self->writer->__pyx_vtab)->write_n(__pyx_v_self->writer, __pyx_v_self->marker); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(25, 257, __pyx_L1_error)

    /* "src/container.pxi":258
 *                 self.writer.write_n(self.next_block.view())
 *             self.writer.write_n(self.marker)
 *             self.blocks_written += 1             # <<<<<<<<<<<<<<
 *             self.pending_block.reset()
 *             self.num_pending = 0
 */
    __pyx_v_self->blocks_written = (__pyx_v_self->blocks_written + 1);

    /* "src/container.pxi":259
 *             self.writer.write_n(self.marker)
 *             self.blocks_written += 1
 *             self.pending_block.reset()             # <<<<<<<<<<<<<<
 *             self.num_pending = 0
 *         self.writer.flush()
 */
    __pyx_t_5 = __pyx_f_5cavro_12MemoryWriter_reset(__pyx_v_self->pending_block); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "src/container.pxi":260
 *             self.blocks_written += 1
 *             self.pending_block.reset()
 *             self.num_pending = 0             # <<<<<<<<<<<<<<
 *         self.writer.flush()
 *         return 0
 */
    __pyx_v_self->num_pending = 0;

    /* "src/container.pxi":248
 *     cdef int _flush_block(self, int force=False) except -1:
 *         cdef size_t pending_len = self.pending_block.len
 *         if force or self.num_pending > 0:             # <<<<<<<<<<<<<<
 *             if self.blocks_written == 0 and self.should_write_header:
 *                 self._write_header()
 */
  }

  /* "src/container.pxi":261
 *             self.pending_block.reset()
 *             self.num_pending = 0
 *         self.writer.flush()             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_5cavro__Writer *)__pyx_v_self->writer->__pyx_vtab)->flush(__pyx_v_self->writer); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(25, 261, __pyx_L1_error)

  /* "src/container.pxi":262
 *             self.num_pending = 0
 *         self.writer.flush()
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def flush(self, force=False):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "src/container.pxi":246
 *         self.next_item = None
 * 
 *     cdef int _flush_block(self, int force=False) except -1:             # <<<<<<<<<<<<<<
 *         cdef size_t pending_len = self.pending_block.len
 *         if force or self.num_pending > 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("cavro.ContainerWriter._flush_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":264
 *         return 0
 * 
 *     def flush(self, force=False):             # <<<<<<<<<<<<<<
 *         self._flush_block(force)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_11flush(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15ContainerWriter_11flush = {"flush", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_11flush, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15ContainerWriter_11flush(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_force = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("flush (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_force,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_False));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_force);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 264, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "flush") < 0)) __PYX_ERR(25, 264, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_force = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("flush", 0, 0, 1, __pyx_nargs); __PYX_ERR(25, 264, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ContainerWriter.flush", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_10flush(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self), __pyx_v_force);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_10flush(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, PyObject *__pyx_v_force) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  struct __pyx_opt_args_5cavro_15ContainerWriter__flush_block __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("flush", 1);

  /* "src/container.pxi":265
 * 
 *     def flush(self, force=False):
 *         self._flush_block(force)             # <<<<<<<<<<<<<<
 * 
 *     cpdef int write_one(self, obj) except -1:
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_force); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(25, 265, __pyx_L1_error)
  __pyx_t_3.__pyx_n = 1;
  __pyx_t_3.force = __pyx_t_1;
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_ContainerWriter *)__pyx_v_self->__pyx_vtab)->_flush_block(__pyx_v_self, &__pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(25, 265, __pyx_L1_error)

  /* "src/container.pxi":264
 *         return 0
 * 
 *     def flush(self, force=False):             # <<<<<<<<<<<<<<
 *         self._flush_block(force)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.ContainerWriter.flush", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":267
 *         self._flush_block(force)
 * 
 *     cpdef int write_one(self, obj) except -1:             # <<<<<<<<<<<<<<
 *         if self.next_item is None:
 *             raise ValueError('Trying to write to closed Container')
 */

static PyObject *__pyx_pw_5cavro_15ContainerWriter_13write_one(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_5cavro_15ContainerWriter_write_one(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_one", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_write_one); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_5cavro_15ContainerWriter_13write_one)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_obj};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 267, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(25, 267, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "src/container.pxi":268
 * 
 *     cpdef int write_one(self, obj) except -1:
 *         if self.next_item is None:             # <<<<<<<<<<<<<<
 *             raise ValueError('Trying to write to closed Container')
 *         self.next_item.reset()
 */
  __pyx_t_6 = (((PyObject *)__pyx_v_self->next_item) == Py_None);
  if (unlikely(__pyx_t_6)) {

    /* "src/container.pxi":269
 *     cpdef int write_one(self, obj) except -1:
 *         if self.next_item is None:
 *             raise ValueError('Trying to write to closed Container')             # <<<<<<<<<<<<<<
 *         self.next_item.reset()
 *         self.schema.binary_write(self.next_item, obj)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__93, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(25, 269, __pyx_L1_error)

    /* "src/container.pxi":268
 * 
 *     cpdef int write_one(self, obj) except -1:
 *         if self.next_item is None:             # <<<<<<<<<<<<<<
 *             raise ValueError('Trying to write to closed Container')
 *         self.next_item.reset()
 */
  }

  /* "src/container.pxi":270
 *         if self.next_item is None:
 *             raise ValueError('Trying to write to closed Container')
 *         self.next_item.reset()             # <<<<<<<<<<<<<<
 *         self.schema.binary_write(self.next_item, obj)
 * 
 */
  __pyx_t_1 = __pyx_f_5cavro_12MemoryWriter_reset(__pyx_v_self->next_item); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/container.pxi":271
 *             raise ValueError('Trying to write to closed Container')
 *         self.next_item.reset()
 *         self.schema.binary_write(self.next_item, obj)             # <<<<<<<<<<<<<<
 * 
 *         if not self.options.container_fill_blocks:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->next_item);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_5cavro_Schema *)__pyx_v_self->schema->__pyx_vtab)->binary_write(__pyx_v_self->schema, ((struct __pyx_obj_5cavro__Writer *)__pyx_t_1), __pyx_v_obj, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "src/container.pxi":273
 *         self.schema.binary_write(self.next_item, obj)
 * 
 *         if not self.options.container_fill_blocks:             # <<<<<<<<<<<<<<
 *             if self.pending_block.len + self.next_item.len > self.max_blocksize:
 *                 self._flush_block()
 */
  __pyx_t_6 = (!__pyx_v_self->options->container_fill_blocks);
  if (__pyx_t_6) {

    /* "src/container.pxi":274
 * 
 *         if not self.options.container_fill_blocks:
 *             if self.pending_block.len + self.next_item.len > self.max_blocksize:             # <<<<<<<<<<<<<<
 *                 self._flush_block()
 * 
 */
    __pyx_t_6 = ((__pyx_v_self->pending_block->len + __pyx_v_self->next_item->len) > __pyx_v_self->max_blocksize);
    if (__pyx_t_6) {

      /* "src/container.pxi":275
 *         if not self.options.container_fill_blocks:
 *             if self.pending_block.len + self.next_item.len > self.max_blocksize:
 *                 self._flush_block()             # <<<<<<<<<<<<<<
 * 
 *         self.pending_block.write_n(self.next_item.view())
 */
      __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_ContainerWriter *)__pyx_v_self->__pyx_vtab)->_flush_block(__pyx_v_self, NULL); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(25, 275, __pyx_L1_error)

      /* "src/container.pxi":274
 * 
 *         if not self.options.container_fill_blocks:
 *             if self.pending_block.len + self.next_item.len > self.max_blocksize:             # <<<<<<<<<<<<<<
 *                 self._flush_block()
 * 
 */
    }

    /* "src/container.pxi":273
 *         self.schema.binary_write(self.next_item, obj)
 * 
 *         if not self.options.container_fill_blocks:             # <<<<<<<<<<<<<<
 *             if self.pending_block.len + self.next_item.len > self.max_blocksize:
 *                 self._flush_block()
 */
  }

  /* "src/container.pxi":277
 *                 self._flush_block()
 * 
 *         self.pending_block.write_n(self.next_item.view())             # <<<<<<<<<<<<<<
 *         self.num_pending += 1
 * 
 */
  __pyx_t_7 = __pyx_f_5cavro_12MemoryWriter_view(__pyx_v_self->next_item); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(25, 277, __pyx_L1_error)
  __pyx_t_5 = __pyx_f_5cavro_12MemoryWriter_write_n(((struct __pyx_obj_5cavro__Writer *)__pyx_v_self->pending_block), __pyx_t_7); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(25, 277, __pyx_L1_error)
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_7, 1);
  __pyx_t_7.memview = NULL; __pyx_t_7.data = NULL;

  /* "src/container.pxi":278
 * 
 *         self.pending_block.write_n(self.next_item.view())
 *         self.num_pending += 1             # <<<<<<<<<<<<<<
 * 
 *         if self.options.container_fill_blocks:
 */
  __pyx_v_self->num_pending = (__pyx_v_self->num_pending + 1);

  /* "src/container.pxi":280
 *         self.num_pending += 1
 * 
 *         if self.options.container_fill_blocks:             # <<<<<<<<<<<<<<
 *             if self.pending_block.len >= self.max_blocksize:
 *                 self._flush_block()
 */
  if (__pyx_v_self->options->container_fill_blocks) {

    /* "src/container.pxi":281
 * 
 *         if self.options.container_fill_blocks:
 *             if self.pending_block.len >= self.max_blocksize:             # <<<<<<<<<<<<<<
 *                 self._flush_block()
 * 
 */
    __pyx_t_6 = (__pyx_v_self->pending_block->len >= __pyx_v_self->max_blocksize);
    if (__pyx_t_6) {

      /* "src/container.pxi":282
 *         if self.options.container_fill_blocks:
 *             if self.pending_block.len >= self.max_blocksize:
 *                 self._flush_block()             # <<<<<<<<<<<<<<
 * 
 *     def write_many(self, objs):
 */
      __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_ContainerWriter *)__pyx_v_self->__pyx_vtab)->_flush_block(__pyx_v_self, NULL); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(25, 282, __pyx_L1_error)

      /* "src/container.pxi":281
 * 
 *         if self.options.container_fill_blocks:
 *             if self.pending_block.len >= self.max_blocksize:             # <<<<<<<<<<<<<<
 *                 self._flush_block()
 * 
 */
    }

    /* "src/container.pxi":280
 *         self.num_pending += 1
 * 
 *         if self.options.container_fill_blocks:             # <<<<<<<<<<<<<<
 *             if self.pending_block.len >= self.max_blocksize:
 *                 self._flush_block()
 */
  }

  /* "src/container.pxi":267
 *         self._flush_block(force)
 * 
 *     cpdef int write_one(self, obj) except -1:             # <<<<<<<<<<<<<<
 *         if self.next_item is None:
 *             raise ValueError('Trying to write to closed Container')
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_AddTraceback("cavro.ContainerWriter.write_one", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_13write_one(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15ContainerWriter_13write_one = {"write_one", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_13write_one, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15ContainerWriter_13write_one(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_obj = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_one (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_obj)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 267, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "write_one") < 0)) __PYX_ERR(25, 267, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_obj = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("write_one", 1, 1, 1, __pyx_nargs); __PYX_ERR(25, 267, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ContainerWriter.write_one", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_12write_one(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self), __pyx_v_obj);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_12write_one(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_one", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5cavro_15ContainerWriter_write_one(__pyx_v_self, __pyx_v_obj, 1); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(25, 267, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cavro.ContainerWriter.write_one", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":284
 *                 self._flush_block()
 * 
 *     def write_many(self, objs):             # <<<<<<<<<<<<<<
 *         for obj in objs:
 *             self.write_one(obj)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_15write_many(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15ContainerWriter_15write_many = {"write_many", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_15write_many, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15ContainerWriter_15write_many(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_objs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_many (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_objs,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_objs)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 284, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "write_many") < 0)) __PYX_ERR(25, 284, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_objs = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("write_many", 1, 1, 1, __pyx_nargs); __PYX_ERR(25, 284, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ContainerWriter.write_many", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_14write_many(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self), __pyx_v_objs);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_14write_many(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, PyObject *__pyx_v_objs) {
  PyObject *__pyx_v_obj = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_many", 1);

  /* "src/container.pxi":285
 * 
 *     def write_many(self, objs):
 *         for obj in objs:             # <<<<<<<<<<<<<<
 *             self.write_one(obj)
 */
  if (likely(PyList_CheckExact(__pyx_v_objs)) || PyTuple_CheckExact(__pyx_v_objs)) {
    __pyx_t_1 = __pyx_v_objs; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_objs); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 285, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(25, 285, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(25, 285, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 285, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(25, 285, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(25, 285, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 285, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(25, 285, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "src/container.pxi":286
 *     def write_many(self, objs):
 *         for obj in objs:
 *             self.write_one(obj)             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_5cavro_ContainerWriter *)__pyx_v_self->__pyx_vtab)->write_one(__pyx_v_self, __pyx_v_obj, 0); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(25, 286, __pyx_L1_error)

    /* "src/container.pxi":285
 * 
 *     def write_many(self, objs):
 *         for obj in objs:             # <<<<<<<<<<<<<<
 *             self.write_one(obj)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/container.pxi":284
 *                 self._flush_block()
 * 
 *     def write_many(self, objs):             # <<<<<<<<<<<<<<
 *         for obj in objs:
 *             self.write_one(obj)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.ContainerWriter.write_many", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":158
 *     cdef MemoryWriter next_item
 * 
 *     cdef readonly Schema schema             # <<<<<<<<<<<<<<
 *     cdef readonly Codec codec
 *     cdef readonly const uint8_t[:] marker
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_6schema_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerWriter_6schema_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_6schema___get__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_6schema___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->schema);
  __pyx_r = ((PyObject *)__pyx_v_self->schema);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":159
 * 
 *     cdef readonly Schema schema
 *     cdef readonly Codec codec             # <<<<<<<<<<<<<<
 *     cdef readonly const uint8_t[:] marker
 *     cdef readonly size_t max_blocksize
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_5codec_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerWriter_5codec_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_5codec___get__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_5codec___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->codec);
  __pyx_r = ((PyObject *)__pyx_v_self->codec);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":160
 *     cdef readonly Schema schema
 *     cdef readonly Codec codec
 *     cdef readonly const uint8_t[:] marker             # <<<<<<<<<<<<<<
 *     cdef readonly size_t max_blocksize
 *     cdef readonly Options options
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_6marker_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerWriter_6marker_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_6marker___get__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_6marker___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_self->marker.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(25, 160, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->marker, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ContainerWriter.marker.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":161
 *     cdef readonly Codec codec
 *     cdef readonly const uint8_t[:] marker
 *     cdef readonly size_t max_blocksize             # <<<<<<<<<<<<<<
 *     cdef readonly Options options
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_13max_blocksize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerWriter_13max_blocksize_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_13max_blocksize___get__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_13max_blocksize___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->max_blocksize); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ContainerWriter.max_blocksize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":162
 *     cdef readonly const uint8_t[:] marker
 *     cdef readonly size_t max_blocksize
 *     cdef readonly Options options             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly bint should_write_header
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_7options_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerWriter_7options_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_7options___get__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_7options___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->options);
  __pyx_r = ((PyObject *)__pyx_v_self->options);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":164
 *     cdef readonly Options options
 * 
 *     cdef readonly bint should_write_header             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly size_t num_pending
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_19should_write_header_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerWriter_19should_write_header_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_19should_write_header___get__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_19should_write_header___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->should_write_header); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ContainerWriter.should_write_header.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":166
 *     cdef readonly bint should_write_header
 * 
 *     cdef readonly size_t num_pending             # <<<<<<<<<<<<<<
 *     cdef readonly int blocks_written
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_11num_pending_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerWriter_11num_pending_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_11num_pending___get__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_11num_pending___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->num_pending); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ContainerWriter.num_pending.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":167
 * 
 *     cdef readonly size_t num_pending
 *     cdef readonly int blocks_written             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly dict metadata
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_14blocks_written_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerWriter_14blocks_written_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_14blocks_written___get__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_14blocks_written___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->blocks_written); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.ContainerWriter.blocks_written.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/container.pxi":169
 *     cdef readonly int blocks_written
 * 
 *     cdef readonly dict metadata             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, dest, Schema schema, str codec='null', size_t max_blocksize=16352, write_header=True, metadata=None, marker=None, options=DEFAULT_OPTIONS):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_8metadata_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_15ContainerWriter_8metadata_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_8metadata___get__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_8metadata___get__(struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->metadata);
  __pyx_r = __pyx_v_self->metadata;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_17__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15ContainerWriter_17__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_17__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15ContainerWriter_17__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_16__reduce_cython__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.ContainerWriter.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_15ContainerWriter_19__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_15ContainerWriter_19__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_19__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_15ContainerWriter_19__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.ContainerWriter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_15ContainerWriter_18__setstate_cython__(((struct __pyx_obj_5cavro_ContainerWriter *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_15ContainerWriter_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5cavro_ContainerWriter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cavro.ContainerWriter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/promotions.pxi":9
 *     """
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 */

static PyObject *__pyx_f_5cavro_14PromoteToFloat_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_PromoteToFloat *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 1);

  /* "src/promotions.pxi":10
 * 
 *     cdef encode_value(self, value):
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__94, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(26, 10, __pyx_L1_error)

  /* "src/promotions.pxi":9
 *     """
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.PromoteToFloat.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/promotions.pxi":12
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return float(value)
 * 
 */

static PyObject *__pyx_f_5cavro_14PromoteToFloat_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_PromoteToFloat *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/promotions.pxi":13
 * 
 *     cdef decode_value(self, value):
 *         return float(value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/promotions.pxi":12
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return float(value)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.PromoteToFloat.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14PromoteToFloat_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_14PromoteToFloat_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14PromoteToFloat_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_14PromoteToFloat_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_14PromoteToFloat___reduce_cython__(((struct __pyx_obj_5cavro_PromoteToFloat *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14PromoteToFloat___reduce_cython__(struct __pyx_obj_5cavro_PromoteToFloat *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PromoteToFloat, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PromoteToFloat, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_PromoteToFloat, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PromoteToFloat, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_PromoteToFloat); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PromoteToFloat, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PromoteToFloat, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_PromoteToFloat, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PromoteToFloat__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_PromoteToFloat); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.PromoteToFloat.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromoteToFloat, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromoteToFloat__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_14PromoteToFloat_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_14PromoteToFloat_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14PromoteToFloat_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_14PromoteToFloat_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.PromoteToFloat.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_14PromoteToFloat_2__setstate_cython__(((struct __pyx_obj_5cavro_PromoteToFloat *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_14PromoteToFloat_2__setstate_cython__(struct __pyx_obj_5cavro_PromoteToFloat *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PromoteToFloat, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PromoteToFloat__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_PromoteToFloat__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromoteToFloat, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromoteToFloat__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.PromoteToFloat.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/promotions.pxi":22
 *     """
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 */

static PyObject *__pyx_f_5cavro_20PromoteBytesToString_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_PromoteBytesToString *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 1);

  /* "src/promotions.pxi":23
 * 
 *     cdef encode_value(self, value):
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__94, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(26, 23, __pyx_L1_error)

  /* "src/promotions.pxi":22
 *     """
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.PromoteBytesToString.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/promotions.pxi":25
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return value.decode("utf-8")
 * 
 */

static PyObject *__pyx_f_5cavro_20PromoteBytesToString_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_PromoteBytesToString *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/promotions.pxi":26
 * 
 *     cdef decode_value(self, value):
 *         return value.decode("utf-8")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_decode); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_utf_8};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 26, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/promotions.pxi":25
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return value.decode("utf-8")
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.PromoteBytesToString.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_20PromoteBytesToString_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_20PromoteBytesToString_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_20PromoteBytesToString_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_20PromoteBytesToString_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_20PromoteBytesToString___reduce_cython__(((struct __pyx_obj_5cavro_PromoteBytesToString *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_20PromoteBytesToString___reduce_cython__(struct __pyx_obj_5cavro_PromoteBytesToString *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PromoteBytesToString, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PromoteBytesToString, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_PromoteBytesToString, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PromoteBytesToString, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_PromoteBytesToStr); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PromoteBytesToString, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PromoteBytesToString, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_PromoteBytesToString, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PromoteBytesToString__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_PromoteBytesToStr); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.PromoteBytesToString.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromoteBytesToString, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromoteBytesToString__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_20PromoteBytesToString_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_20PromoteBytesToString_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_20PromoteBytesToString_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_20PromoteBytesToString_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.PromoteBytesToString.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_20PromoteBytesToString_2__setstate_cython__(((struct __pyx_obj_5cavro_PromoteBytesToString *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_20PromoteBytesToString_2__setstate_cython__(struct __pyx_obj_5cavro_PromoteBytesToString *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PromoteBytesToString, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PromoteBytesToString__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_PromoteBytesToString__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromoteBytesToString, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromoteBytesToString__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.PromoteBytesToString.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/promotions.pxi":35
 *     """
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 */

static PyObject *__pyx_f_5cavro_20PromoteStringToBytes_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_PromoteStringToBytes *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 1);

  /* "src/promotions.pxi":36
 * 
 *     cdef encode_value(self, value):
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__94, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(26, 36, __pyx_L1_error)

  /* "src/promotions.pxi":35
 *     """
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.PromoteStringToBytes.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/promotions.pxi":38
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return value.encode("utf-8")
 * 
 */

static PyObject *__pyx_f_5cavro_20PromoteStringToBytes_decode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_PromoteStringToBytes *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/promotions.pxi":39
 * 
 *     cdef decode_value(self, value):
 *         return value.encode("utf-8")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_utf_8};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/promotions.pxi":38
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         return value.encode("utf-8")
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.PromoteStringToBytes.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_20PromoteStringToBytes_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_20PromoteStringToBytes_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_20PromoteStringToBytes_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_20PromoteStringToBytes_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_20PromoteStringToBytes___reduce_cython__(((struct __pyx_obj_5cavro_PromoteStringToBytes *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_20PromoteStringToBytes___reduce_cython__(struct __pyx_obj_5cavro_PromoteStringToBytes *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PromoteStringToBytes, (type(self), 0xe3b0c44, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PromoteStringToBytes, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_PromoteStringToBytes, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PromoteStringToBytes, (type(self), 0xe3b0c44, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_PromoteStringToBy); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PromoteStringToBytes, (type(self), 0xe3b0c44, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PromoteStringToBytes, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_PromoteStringToBytes, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PromoteStringToBytes__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_PromoteStringToBy); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_238750788)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.PromoteStringToBytes.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromoteStringToBytes, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromoteStringToBytes__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_20PromoteStringToBytes_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_20PromoteStringToBytes_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_20PromoteStringToBytes_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_20PromoteStringToBytes_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.PromoteStringToBytes.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_20PromoteStringToBytes_2__setstate_cython__(((struct __pyx_obj_5cavro_PromoteStringToBytes *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_20PromoteStringToBytes_2__setstate_cython__(struct __pyx_obj_5cavro_PromoteStringToBytes *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PromoteStringToBytes, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PromoteStringToBytes__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_PromoteStringToBytes__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromoteStringToBytes, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromoteStringToBytes__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.PromoteStringToBytes.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/promotions.pxi":53
 *     cdef readonly object extra
 * 
 *     def __init__(self, reader_type, writer_type, extra=None):             # <<<<<<<<<<<<<<
 *         self.reader_type = reader_type
 *         self.writer_type = writer_type
 */

/* Python wrapper */
static int __pyx_pw_5cavro_13CannotPromote_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5cavro_13CannotPromote_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_reader_type = 0;
  PyObject *__pyx_v_writer_type = 0;
  PyObject *__pyx_v_extra = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_reader_type,&__pyx_n_s_writer_type,&__pyx_n_s_extra,0};
    values[2] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_reader_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 53, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_writer_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 53, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(26, 53, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_extra);
          if (value) { values[2] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 53, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(26, 53, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_reader_type = values[0];
    __pyx_v_writer_type = values[1];
    __pyx_v_extra = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, __pyx_nargs); __PYX_ERR(26, 53, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.CannotPromote.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_13CannotPromote___init__(((struct __pyx_obj_5cavro_CannotPromote *)__pyx_v_self), __pyx_v_reader_type, __pyx_v_writer_type, __pyx_v_extra);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5cavro_13CannotPromote___init__(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self, PyObject *__pyx_v_reader_type, PyObject *__pyx_v_writer_type, PyObject *__pyx_v_extra) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "src/promotions.pxi":54
 * 
 *     def __init__(self, reader_type, writer_type, extra=None):
 *         self.reader_type = reader_type             # <<<<<<<<<<<<<<
 *         self.writer_type = writer_type
 *         self.extra = extra
 */
  if (!(likely(((__pyx_v_reader_type) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_reader_type, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(26, 54, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_reader_type;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->reader_type);
  __Pyx_DECREF((PyObject *)__pyx_v_self->reader_type);
  __pyx_v_self->reader_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/promotions.pxi":55
 *     def __init__(self, reader_type, writer_type, extra=None):
 *         self.reader_type = reader_type
 *         self.writer_type = writer_type             # <<<<<<<<<<<<<<
 *         self.extra = extra
 * 
 */
  if (!(likely(((__pyx_v_writer_type) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_writer_type, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(26, 55, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_writer_type;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->writer_type);
  __Pyx_DECREF((PyObject *)__pyx_v_self->writer_type);
  __pyx_v_self->writer_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "src/promotions.pxi":56
 *         self.reader_type = reader_type
 *         self.writer_type = writer_type
 *         self.extra = extra             # <<<<<<<<<<<<<<
 * 
 *     cdef encode_value(self, value):
 */
  __Pyx_INCREF(__pyx_v_extra);
  __Pyx_GIVEREF(__pyx_v_extra);
  __Pyx_GOTREF(__pyx_v_self->extra);
  __Pyx_DECREF(__pyx_v_self->extra);
  __pyx_v_self->extra = __pyx_v_extra;

  /* "src/promotions.pxi":53
 *     cdef readonly object extra
 * 
 *     def __init__(self, reader_type, writer_type, extra=None):             # <<<<<<<<<<<<<<
 *         self.reader_type = reader_type
 *         self.writer_type = writer_type
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.CannotPromote.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/promotions.pxi":58
 *         self.extra = extra
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 */

static PyObject *__pyx_f_5cavro_13CannotPromote_encode_value(CYTHON_UNUSED struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_value", 1);

  /* "src/promotions.pxi":59
 * 
 *     cdef encode_value(self, value):
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__94, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(26, 59, __pyx_L1_error)

  /* "src/promotions.pxi":58
 *         self.extra = extra
 * 
 *     cdef encode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.CannotPromote.encode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/promotions.pxi":61
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise CannotPromoteError(self.reader_type, self.writer_type, self.extra)
 */

static PyObject *__pyx_f_5cavro_13CannotPromote_decode_value(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_value", 1);

  /* "src/promotions.pxi":62
 * 
 *     cdef decode_value(self, value):
 *         raise CannotPromoteError(self.reader_type, self.writer_type, self.extra)             # <<<<<<<<<<<<<<
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CannotPromoteError); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, ((PyObject *)__pyx_v_self->reader_type), ((PyObject *)__pyx_v_self->writer_type), __pyx_v_self->extra};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(26, 62, __pyx_L1_error)

  /* "src/promotions.pxi":61
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")
 * 
 *     cdef decode_value(self, value):             # <<<<<<<<<<<<<<
 *         raise CannotPromoteError(self.reader_type, self.writer_type, self.extra)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cavro.CannotPromote.decode_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/promotions.pxi":49
 * 
 * 
 *     cdef readonly AvroType reader_type             # <<<<<<<<<<<<<<
 *     cdef readonly AvroType writer_type
 *     cdef readonly object extra
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_13CannotPromote_11reader_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_13CannotPromote_11reader_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_13CannotPromote_11reader_type___get__(((struct __pyx_obj_5cavro_CannotPromote *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_13CannotPromote_11reader_type___get__(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->reader_type);
  __pyx_r = ((PyObject *)__pyx_v_self->reader_type);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/promotions.pxi":50
 * 
 *     cdef readonly AvroType reader_type
 *     cdef readonly AvroType writer_type             # <<<<<<<<<<<<<<
 *     cdef readonly object extra
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_13CannotPromote_11writer_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_13CannotPromote_11writer_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_13CannotPromote_11writer_type___get__(((struct __pyx_obj_5cavro_CannotPromote *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_13CannotPromote_11writer_type___get__(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->writer_type);
  __pyx_r = ((PyObject *)__pyx_v_self->writer_type);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/promotions.pxi":51
 *     cdef readonly AvroType reader_type
 *     cdef readonly AvroType writer_type
 *     cdef readonly object extra             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, reader_type, writer_type, extra=None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_13CannotPromote_5extra_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5cavro_13CannotPromote_5extra_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_5cavro_13CannotPromote_5extra___get__(((struct __pyx_obj_5cavro_CannotPromote *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_13CannotPromote_5extra___get__(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->extra);
  __pyx_r = __pyx_v_self->extra;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_13CannotPromote_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_13CannotPromote_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13CannotPromote_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_13CannotPromote_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_5cavro_13CannotPromote_2__reduce_cython__(((struct __pyx_obj_5cavro_CannotPromote *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_13CannotPromote_2__reduce_cython__(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.extra, self.reader_type, self.writer_type)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->extra);
  __Pyx_GIVEREF(__pyx_v_self->extra);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->extra)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->reader_type);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->reader_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->reader_type))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->writer_type);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->writer_type);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->writer_type))) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.extra, self.reader_type, self.writer_type)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.extra, self.reader_type, self.writer_type)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.extra is not None or self.reader_type is not None or self.writer_type is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.extra, self.reader_type, self.writer_type)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.extra is not None or self.reader_type is not None or self.writer_type is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CannotPromote, (type(self), 0xcbfa7ba, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->extra != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->reader_type) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->writer_type) != Py_None);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.extra is not None or self.reader_type is not None or self.writer_type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CannotPromote, (type(self), 0xcbfa7ba, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.extra is not None or self.reader_type is not None or self.writer_type is not None
 *     if use_setstate:
 *         return __pyx_unpickle_CannotPromote, (type(self), 0xcbfa7ba, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CannotPromote, (type(self), 0xcbfa7ba, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_CannotPromote); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_213886906);
    __Pyx_GIVEREF(__pyx_int_213886906);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_213886906)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.extra is not None or self.reader_type is not None or self.writer_type is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CannotPromote, (type(self), 0xcbfa7ba, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CannotPromote, (type(self), 0xcbfa7ba, None), state
 *     else:
 *         return __pyx_unpickle_CannotPromote, (type(self), 0xcbfa7ba, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CannotPromote__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_CannotPromote); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_213886906);
    __Pyx_GIVEREF(__pyx_int_213886906);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_213886906)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cavro.CannotPromote.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CannotPromote, (type(self), 0xcbfa7ba, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CannotPromote__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_13CannotPromote_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_13CannotPromote_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13CannotPromote_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_13CannotPromote_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.CannotPromote.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_13CannotPromote_4__setstate_cython__(((struct __pyx_obj_5cavro_CannotPromote *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_13CannotPromote_4__setstate_cython__(struct __pyx_obj_5cavro_CannotPromote *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CannotPromote, (type(self), 0xcbfa7ba, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CannotPromote__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_CannotPromote__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CannotPromote, (type(self), 0xcbfa7ba, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CannotPromote__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cavro.CannotPromote.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Options(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_19__pyx_unpickle_Options(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_19__pyx_unpickle_Options = {"__pyx_unpickle_Options", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_19__pyx_unpickle_Options, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_19__pyx_unpickle_Options(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Options (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Options", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Options", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Options") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Options", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Options", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_18__pyx_unpickle_Options(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_18__pyx_unpickle_Options(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Options", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc466972, 0x8dd8fb8, 0xf76cf3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc466972, 0x8dd8fb8, 0xf76cf3c) = (adapt_record_types, allow_aliases_to_be_string, allow_duplicate_union_types, allow_empty_unions, allow_error_type, allow_false_values_for_null, allow_invalid_default_values, allow_leading_dot_in_names, allow_nested_unions, allow_primitive_name_collision, allow_primitive_names_in_namespaces, allow_tuple_notation, allow_union_default_any_member, alternate_timestamp_millis_encoding, ascii_name_rules, bytes_codec, bytes_default_value_utf8, canonical_form_repeat_enum, canonical_form_repeat_fixed, clamp_float_overflow, clamp_int_overflow, coerce_int_to_float, coerce_values_to_boolean, coerce_values_to_float, coerce_values_to_int, coerce_values_to_str, container_fill_blocks, date_type_accepts_string, decimal_check_exp_overflow, defer_schema_promotion_errors, enforce_enum_symbol_name_rules, enforce_namespace_name_rules, enforce_type_name_rules, enum_symbols_must_be_unique, expand_types_in_schema, externally_defined_types, fingerprint_returns_digest, fixed_codec, float_out_of_range_inf, inline_namespaces, invalid_value_include_array_index, invalid_value_include_map_key, invalid_value_includes_record_name, logical_types, missing_values_can_be_empty_container, missing_values_can_be_null, naive_dt_assume_utc, named_type_names_must_be_unique, null_pad_fixed, raise_on_invalid_logical, record_allow_extra_fields, record_can_encode_dict, record_decodes_to_dict, record_encode_use_defaults, record_fields_must_be_unique, record_values_type_hint, return_uuid_object, string_types_default_unchanged, truncate_fixed, types_str_to_schema, unicode_errors, union_decodes_to, union_json_encodes_type_name))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__95, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc466972, 0x8dd8fb8, 0xf76cf3c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc466972, 0x8dd8fb8, 0xf76cf3c) = (adapt_record_types, allow_aliases_to_be_string, allow_duplicate_union_types, allow_empty_unions, allow_error_type, allow_false_values_for_null, allow_invalid_default_values, allow_leading_dot_in_names, allow_nested_unions, allow_primitive_name_collision, allow_primitive_names_in_namespaces, allow_tuple_notation, allow_union_default_any_member, alternate_timestamp_millis_encoding, ascii_name_rules, bytes_codec, bytes_default_value_utf8, canonical_form_repeat_enum, canonical_form_repeat_fixed, clamp_float_overflow, clamp_int_overflow, coerce_int_to_float, coerce_values_to_boolean, coerce_values_to_float, coerce_values_to_int, coerce_values_to_str, container_fill_blocks, date_type_accepts_string, decimal_check_exp_overflow, defer_schema_promotion_errors, enforce_enum_symbol_name_rules, enforce_namespace_name_rules, enforce_type_name_rules, enum_symbols_must_be_unique, expand_types_in_schema, externally_defined_types, fingerprint_returns_digest, fixed_codec, float_out_of_range_inf, inline_namespaces, invalid_value_include_array_index, invalid_value_include_map_key, invalid_value_includes_record_name, logical_types, missing_values_can_be_empty_container, missing_values_can_be_null, naive_dt_assume_utc, named_type_names_must_be_unique, null_pad_fixed, raise_on_invalid_logical, record_allow_extra_fields, record_can_encode_dict, record_decodes_to_dict, record_encode_use_defaults, record_fields_must_be_unique, record_values_type_hint, return_uuid_object, string_types_default_unchanged, truncate_fixed, types_str_to_schema, unicode_errors, union_decodes_to, union_json_encodes_type_name))" % __pyx_checksum
 *     __pyx_result = Options.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc466972, 0x8dd8fb8, 0xf76cf3c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc466972, 0x8dd8fb8, 0xf76cf3c) = (adapt_record_types, allow_aliases_to_be_string, allow_duplicate_union_types, allow_empty_unions, allow_error_type, allow_false_values_for_null, allow_invalid_default_values, allow_leading_dot_in_names, allow_nested_unions, allow_primitive_name_collision, allow_primitive_names_in_namespaces, allow_tuple_notation, allow_union_default_any_member, alternate_timestamp_millis_encoding, ascii_name_rules, bytes_codec, bytes_default_value_utf8, canonical_form_repeat_enum, canonical_form_repeat_fixed, clamp_float_overflow, clamp_int_overflow, coerce_int_to_float, coerce_values_to_boolean, coerce_values_to_float, coerce_values_to_int, coerce_values_to_str, container_fill_blocks, date_type_accepts_string, decimal_check_exp_overflow, defer_schema_promotion_errors, enforce_enum_symbol_name_rules, enforce_namespace_name_rules, enforce_type_name_rules, enum_symbols_must_be_unique, expand_types_in_schema, externally_defined_types, fingerprint_returns_digest, fixed_codec, float_out_of_range_inf, inline_namespaces, invalid_value_include_array_index, invalid_value_include_map_key, invalid_value_includes_record_name, logical_types, missing_values_can_be_empty_container, missing_values_can_be_null, naive_dt_assume_utc, named_type_names_must_be_unique, null_pad_fixed, raise_on_invalid_logical, record_allow_extra_fields, record_can_encode_dict, record_decodes_to_dict, record_encode_use_defaults, record_fields_must_be_unique, record_values_type_hint, return_uuid_object, string_types_default_unchanged, truncate_fixed, types_str_to_schema, unicode_errors, union_decodes_to, union_json_encodes_type_name))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Options.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc466972, 0x8dd8fb8, 0xf76cf3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc466972, 0x8dd8fb8, 0xf76cf3c) = (adapt_record_types, allow_aliases_to_be_string, allow_duplicate_union_types, allow_empty_unions, allow_error_type, allow_false_values_for_null, allow_invalid_default_values, allow_leading_dot_in_names, allow_nested_unions, allow_primitive_name_collision, allow_primitive_names_in_namespaces, allow_tuple_notation, allow_union_default_any_member, alternate_timestamp_millis_encoding, ascii_name_rules, bytes_codec, bytes_default_value_utf8, canonical_form_repeat_enum, canonical_form_repeat_fixed, clamp_float_overflow, clamp_int_overflow, coerce_int_to_float, coerce_values_to_boolean, coerce_values_to_float, coerce_values_to_int, coerce_values_to_str, container_fill_blocks, date_type_accepts_string, decimal_check_exp_overflow, defer_schema_promotion_errors, enforce_enum_symbol_name_rules, enforce_namespace_name_rules, enforce_type_name_rules, enum_symbols_must_be_unique, expand_types_in_schema, externally_defined_types, fingerprint_returns_digest, fixed_codec, float_out_of_range_inf, inline_namespaces, invalid_value_include_array_index, invalid_value_include_map_key, invalid_value_includes_record_name, logical_types, missing_values_can_be_empty_container, missing_values_can_be_null, naive_dt_assume_utc, named_type_names_must_be_unique, null_pad_fixed, raise_on_invalid_logical, record_allow_extra_fields, record_can_encode_dict, record_decodes_to_dict, record_encode_use_defaults, record_fields_must_be_unique, record_values_type_hint, return_uuid_object, string_types_default_unchanged, truncate_fixed, types_str_to_schema, unicode_errors, union_decodes_to, union_json_encodes_type_name))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc466972, 0x8dd8fb8, 0xf76cf3c) = (adapt_record_types, allow_aliases_to_be_string, allow_duplicate_union_types, allow_empty_unions, allow_error_type, allow_false_values_for_null, allow_invalid_default_values, allow_leading_dot_in_names, allow_nested_unions, allow_primitive_name_collision, allow_primitive_names_in_namespaces, allow_tuple_notation, allow_union_default_any_member, alternate_timestamp_millis_encoding, ascii_name_rules, bytes_codec, bytes_default_value_utf8, canonical_form_repeat_enum, canonical_form_repeat_fixed, clamp_float_overflow, clamp_int_overflow, coerce_int_to_float, coerce_values_to_boolean, coerce_values_to_float, coerce_values_to_int, coerce_values_to_str, container_fill_blocks, date_type_accepts_string, decimal_check_exp_overflow, defer_schema_promotion_errors, enforce_enum_symbol_name_rules, enforce_namespace_name_rules, enforce_type_name_rules, enum_symbols_must_be_unique, expand_types_in_schema, externally_defined_types, fingerprint_returns_digest, fixed_codec, float_out_of_range_inf, inline_namespaces, invalid_value_include_array_index, invalid_value_include_map_key, invalid_value_includes_record_name, logical_types, missing_values_can_be_empty_container, missing_values_can_be_null, naive_dt_assume_utc, named_type_names_must_be_unique, null_pad_fixed, raise_on_invalid_logical, record_allow_extra_fields, record_can_encode_dict, record_decodes_to_dict, record_encode_use_defaults, record_fields_must_be_unique, record_values_type_hint, return_uuid_object, string_types_default_unchanged, truncate_fixed, types_str_to_schema, unicode_errors, union_decodes_to, union_json_encodes_type_name))" % __pyx_checksum
 *     __pyx_result = Options.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Options__set_state(<Options> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_Options), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc466972, 0x8dd8fb8, 0xf76cf3c) = (adapt_record_types, allow_aliases_to_be_string, allow_duplicate_union_types, allow_empty_unions, allow_error_type, allow_false_values_for_null, allow_invalid_default_values, allow_leading_dot_in_names, allow_nested_unions, allow_primitive_name_collision, allow_primitive_names_in_namespaces, allow_tuple_notation, allow_union_default_any_member, alternate_timestamp_millis_encoding, ascii_name_rules, bytes_codec, bytes_default_value_utf8, canonical_form_repeat_enum, canonical_form_repeat_fixed, clamp_float_overflow, clamp_int_overflow, coerce_int_to_float, coerce_values_to_boolean, coerce_values_to_float, coerce_values_to_int, coerce_values_to_str, container_fill_blocks, date_type_accepts_string, decimal_check_exp_overflow, defer_schema_promotion_errors, enforce_enum_symbol_name_rules, enforce_namespace_name_rules, enforce_type_name_rules, enum_symbols_must_be_unique, expand_types_in_schema, externally_defined_types, fingerprint_returns_digest, fixed_codec, float_out_of_range_inf, inline_namespaces, invalid_value_include_array_index, invalid_value_include_map_key, invalid_value_includes_record_name, logical_types, missing_values_can_be_empty_container, missing_values_can_be_null, naive_dt_assume_utc, named_type_names_must_be_unique, null_pad_fixed, raise_on_invalid_logical, record_allow_extra_fields, record_can_encode_dict, record_decodes_to_dict, record_encode_use_defaults, record_fields_must_be_unique, record_values_type_hint, return_uuid_object, string_types_default_unchanged, truncate_fixed, types_str_to_schema, unicode_errors, union_decodes_to, union_json_encodes_type_name))" % __pyx_checksum
 *     __pyx_result = Options.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Options__set_state(<Options> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Options.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Options__set_state(<Options> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Options__set_state(Options __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_Options__set_state(((struct __pyx_obj_5cavro_Options *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc466972, 0x8dd8fb8, 0xf76cf3c) = (adapt_record_types, allow_aliases_to_be_string, allow_duplicate_union_types, allow_empty_unions, allow_error_type, allow_false_values_for_null, allow_invalid_default_values, allow_leading_dot_in_names, allow_nested_unions, allow_primitive_name_collision, allow_primitive_names_in_namespaces, allow_tuple_notation, allow_union_default_any_member, alternate_timestamp_millis_encoding, ascii_name_rules, bytes_codec, bytes_default_value_utf8, canonical_form_repeat_enum, canonical_form_repeat_fixed, clamp_float_overflow, clamp_int_overflow, coerce_int_to_float, coerce_values_to_boolean, coerce_values_to_float, coerce_values_to_int, coerce_values_to_str, container_fill_blocks, date_type_accepts_string, decimal_check_exp_overflow, defer_schema_promotion_errors, enforce_enum_symbol_name_rules, enforce_namespace_name_rules, enforce_type_name_rules, enum_symbols_must_be_unique, expand_types_in_schema, externally_defined_types, fingerprint_returns_digest, fixed_codec, float_out_of_range_inf, inline_namespaces, invalid_value_include_array_index, invalid_value_include_map_key, invalid_value_includes_record_name, logical_types, missing_values_can_be_empty_container, missing_values_can_be_null, naive_dt_assume_utc, named_type_names_must_be_unique, null_pad_fixed, raise_on_invalid_logical, record_allow_extra_fields, record_can_encode_dict, record_decodes_to_dict, record_encode_use_defaults, record_fields_must_be_unique, record_values_type_hint, return_uuid_object, string_types_default_unchanged, truncate_fixed, types_str_to_schema, unicode_errors, union_decodes_to, union_json_encodes_type_name))" % __pyx_checksum
 *     __pyx_result = Options.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Options__set_state(<Options> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Options__set_state(<Options> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Options__set_state(Options __pyx_result, tuple __pyx_state):
 *     __pyx_result.adapt_record_types = __pyx_state[0]; __pyx_result.allow_aliases_to_be_string = __pyx_state[1]; __pyx_result.allow_duplicate_union_types = __pyx_state[2]; __pyx_result.allow_empty_unions = __pyx_state[3]; __pyx_result.allow_error_type = __pyx_state[4]; __pyx_result.allow_false_values_for_null = __pyx_state[5]; __pyx_result.allow_invalid_default_values = __pyx_state[6]; __pyx_result.allow_leading_dot_in_names = __pyx_state[7]; __pyx_result.allow_nested_unions = __pyx_state[8]; __pyx_result.allow_primitive_name_collision = __pyx_state[9]; __pyx_result.allow_primitive_names_in_namespaces = __pyx_state[10]; __pyx_result.allow_tuple_notation = __pyx_state[11]; __pyx_result.allow_union_default_any_member = __pyx_state[12]; __pyx_result.alternate_timestamp_millis_encoding = __pyx_state[13]; __pyx_result.ascii_name_rules = __pyx_state[14]; __pyx_result.bytes_codec = __pyx_state[15]; __pyx_result.bytes_default_value_utf8 = __pyx_state[16]; __pyx_result.canonical_form_repeat_enum = __pyx_state[17]; __pyx_result.canonical_form_repeat_fixed = __pyx_state[18]; __pyx_result.clamp_float_overflow = __pyx_state[19]; __pyx_result.clamp_int_overflow = __pyx_state[20]; __pyx_result.coerce_int_to_float = __pyx_state[21]; __pyx_result.coerce_values_to_boolean = __pyx_state[22]; __pyx_result.coerce_values_to_float = __pyx_state[23]; __pyx_result.coerce_values_to_int = __pyx_state[24]; __pyx_result.coerce_values_to_str = __pyx_state[25]; __pyx_result.container_fill_blocks = __pyx_state[26]; __pyx_result.date_type_accepts_string = __pyx_state[27]; __pyx_result.decimal_check_exp_overflow = __pyx_state[28]; __pyx_result.defer_schema_promotion_errors = __pyx_state[29]; __pyx_result.enforce_enum_symbol_name_rules = __pyx_state[30]; __pyx_result.enforce_namespace_name_rules = __pyx_state[31]; __pyx_result.enforce_type_name_rules = __pyx_state[32]; __pyx_result.enum_symbols_must_be_unique = __pyx_state[33]; __pyx_result.expand_types_in_schema = __pyx_state[34]; __pyx_result.externally_defined_types = __pyx_state[35]; __pyx_result.fingerprint_returns_digest = __pyx_state[36]; __pyx_result.fixed_codec = __pyx_state[37]; __pyx_result.float_out_of_range_inf = __pyx_state[38]; __pyx_result.inline_namespaces = __pyx_state[39]; __pyx_result.invalid_value_include_array_index = __pyx_state[40]; __pyx_result.invalid_value_include_map_key = __pyx_state[41]; __pyx_result.invalid_value_includes_record_name = __pyx_state[42]; __pyx_result.logical_types = __pyx_state[43]; __pyx_result.missing_values_can_be_empty_container = __pyx_state[44]; __pyx_result.missing_values_can_be_null = __pyx_state[45]; __pyx_result.naive_dt_assume_utc = __pyx_state[46]; __pyx_result.named_type_names_must_be_unique = __pyx_state[47]; __pyx_result.null_pad_fixed = __pyx_state[48]; __pyx_result.raise_on_invalid_logical = __pyx_state[49]; __pyx_result.record_allow_extra_fields = __pyx_state[50]; __pyx_result.record_can_encode_dict = __pyx_state[51]; __pyx_result.record_decodes_to_dict = __pyx_state[52]; __pyx_result.record_encode_use_defaults = __pyx_state[53]; __pyx_result.record_fields_must_be_unique = __pyx_state[54]; __pyx_result.record_values_type_hint = __pyx_state[55]; __pyx_result.return_uuid_object = __pyx_state[56]; __pyx_result.string_types_default_unchanged = __pyx_state[57]; __pyx_result.truncate_fixed = __pyx_state[58]; __pyx_result.types_str_to_schema = __pyx_state[59]; __pyx_result.unicode_errors = __pyx_state[60]; __pyx_result.union_decodes_to = __pyx_state[61]; __pyx_result.union_json_encodes_type_name = __pyx_state[62]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Options(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Options", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Options__set_state(<Options> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Options__set_state(Options __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.adapt_record_types = __pyx_state[0]; __pyx_result.allow_aliases_to_be_string = __pyx_state[1]; __pyx_result.allow_duplicate_union_types = __pyx_state[2]; __pyx_result.allow_empty_unions = __pyx_state[3]; __pyx_result.allow_error_type = __pyx_state[4]; __pyx_result.allow_false_values_for_null = __pyx_state[5]; __pyx_result.allow_invalid_default_values = __pyx_state[6]; __pyx_result.allow_leading_dot_in_names = __pyx_state[7]; __pyx_result.allow_nested_unions = __pyx_state[8]; __pyx_result.allow_primitive_name_collision = __pyx_state[9]; __pyx_result.allow_primitive_names_in_namespaces = __pyx_state[10]; __pyx_result.allow_tuple_notation = __pyx_state[11]; __pyx_result.allow_union_default_any_member = __pyx_state[12]; __pyx_result.alternate_timestamp_millis_encoding = __pyx_state[13]; __pyx_result.ascii_name_rules = __pyx_state[14]; __pyx_result.bytes_codec = __pyx_state[15]; __pyx_result.bytes_default_value_utf8 = __pyx_state[16]; __pyx_result.canonical_form_repeat_enum = __pyx_state[17]; __pyx_result.canonical_form_repeat_fixed = __pyx_state[18]; __pyx_result.clamp_float_overflow = __pyx_state[19]; __pyx_result.clamp_int_overflow = __pyx_state[20]; __pyx_result.coerce_int_to_float = __pyx_state[21]; __pyx_result.coerce_values_to_boolean = __pyx_state[22]; __pyx_result.coerce_values_to_float = __pyx_state[23]; __pyx_result.coerce_values_to_int = __pyx_state[24]; __pyx_result.coerce_values_to_str = __pyx_state[25]; __pyx_result.container_fill_blocks = __pyx_state[26]; __pyx_result.date_type_accepts_string = __pyx_state[27]; __pyx_result.decimal_check_exp_overflow = __pyx_state[28]; __pyx_result.defer_schema_promotion_errors = __pyx_state[29]; __pyx_result.enforce_enum_symbol_name_rules = __pyx_state[30]; __pyx_result.enforce_namespace_name_rules = __pyx_state[31]; __pyx_result.enforce_type_name_rules = __pyx_state[32]; __pyx_result.enum_symbols_must_be_unique = __pyx_state[33]; __pyx_result.expand_types_in_schema = __pyx_state[34]; __pyx_result.externally_defined_types = __pyx_state[35]; __pyx_result.fingerprint_returns_digest = __pyx_state[36]; __pyx_result.fixed_codec = __pyx_state[37]; __pyx_result.float_out_of_range_inf = __pyx_state[38]; __pyx_result.inline_namespaces = __pyx_state[39]; __pyx_result.invalid_value_include_array_index = __pyx_state[40]; __pyx_result.invalid_value_include_map_key = __pyx_state[41]; __pyx_result.invalid_value_includes_record_name = __pyx_state[42]; __pyx_result.logical_types = __pyx_state[43]; __pyx_result.missing_values_can_be_empty_container = __pyx_state[44]; __pyx_result.missing_values_can_be_null = __pyx_state[45]; __pyx_result.naive_dt_assume_utc = __pyx_state[46]; __pyx_result.named_type_names_must_be_unique = __pyx_state[47]; __pyx_result.null_pad_fixed = __pyx_state[48]; __pyx_result.raise_on_invalid_logical = __pyx_state[49]; __pyx_result.record_allow_extra_fields = __pyx_state[50]; __pyx_result.record_can_encode_dict = __pyx_state[51]; __pyx_result.record_decodes_to_dict = __pyx_state[52]; __pyx_result.record_encode_use_defaults = __pyx_state[53]; __pyx_result.record_fields_must_be_unique = __pyx_state[54]; __pyx_result.record_values_type_hint = __pyx_state[55]; __pyx_result.return_uuid_object = __pyx_state[56]; __pyx_result.string_types_default_unchanged = __pyx_state[57]; __pyx_result.truncate_fixed = __pyx_state[58]; __pyx_result.types_str_to_schema = __pyx_state[59]; __pyx_result.unicode_errors = __pyx_state[60]; __pyx_result.union_decodes_to = __pyx_state[61]; __pyx_result.union_json_encodes_type_name = __pyx_state[62]
 *     if len(__pyx_state) > 63 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_Options__set_state(struct __pyx_obj_5cavro_Options *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Options__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Options__set_state(Options __pyx_result, tuple __pyx_state):
 *     __pyx_result.adapt_record_types = __pyx_state[0]; __pyx_result.allow_aliases_to_be_string = __pyx_state[1]; __pyx_result.allow_duplicate_union_types = __pyx_state[2]; __pyx_result.allow_empty_unions = __pyx_state[3]; __pyx_result.allow_error_type = __pyx_state[4]; __pyx_result.allow_false_values_for_null = __pyx_state[5]; __pyx_result.allow_invalid_default_values = __pyx_state[6]; __pyx_result.allow_leading_dot_in_names = __pyx_state[7]; __pyx_result.allow_nested_unions = __pyx_state[8]; __pyx_result.allow_primitive_name_collision = __pyx_state[9]; __pyx_result.allow_primitive_names_in_namespaces = __pyx_state[10]; __pyx_result.allow_tuple_notation = __pyx_state[11]; __pyx_result.allow_union_default_any_member = __pyx_state[12]; __pyx_result.alternate_timestamp_millis_encoding = __pyx_state[13]; __pyx_result.ascii_name_rules = __pyx_state[14]; __pyx_result.bytes_codec = __pyx_state[15]; __pyx_result.bytes_default_value_utf8 = __pyx_state[16]; __pyx_result.canonical_form_repeat_enum = __pyx_state[17]; __pyx_result.canonical_form_repeat_fixed = __pyx_state[18]; __pyx_result.clamp_float_overflow = __pyx_state[19]; __pyx_result.clamp_int_overflow = __pyx_state[20]; __pyx_result.coerce_int_to_float = __pyx_state[21]; __pyx_result.coerce_values_to_boolean = __pyx_state[22]; __pyx_result.coerce_values_to_float = __pyx_state[23]; __pyx_result.coerce_values_to_int = __pyx_state[24]; __pyx_result.coerce_values_to_str = __pyx_state[25]; __pyx_result.container_fill_blocks = __pyx_state[26]; __pyx_result.date_type_accepts_string = __pyx_state[27]; __pyx_result.decimal_check_exp_overflow = __pyx_state[28]; __pyx_result.defer_schema_promotion_errors = __pyx_state[29]; __pyx_result.enforce_enum_symbol_name_rules = __pyx_state[30]; __pyx_result.enforce_namespace_name_rules = __pyx_state[31]; __pyx_result.enforce_type_name_rules = __pyx_state[32]; __pyx_result.enum_symbols_must_be_unique = __pyx_state[33]; __pyx_result.expand_types_in_schema = __pyx_state[34]; __pyx_result.externally_defined_types = __pyx_state[35]; __pyx_result.fingerprint_returns_digest = __pyx_state[36]; __pyx_result.fixed_codec = __pyx_state[37]; __pyx_result.float_out_of_range_inf = __pyx_state[38]; __pyx_result.inline_namespaces = __pyx_state[39]; __pyx_result.invalid_value_include_array_index = __pyx_state[40]; __pyx_result.invalid_value_include_map_key = __pyx_state[41]; __pyx_result.invalid_value_includes_record_name = __pyx_state[42]; __pyx_result.logical_types = __pyx_state[43]; __pyx_result.missing_values_can_be_empty_container = __pyx_state[44]; __pyx_result.missing_values_can_be_null = __pyx_state[45]; __pyx_result.naive_dt_assume_utc = __pyx_state[46]; __pyx_result.named_type_names_must_be_unique = __pyx_state[47]; __pyx_result.null_pad_fixed = __pyx_state[48]; __pyx_result.raise_on_invalid_logical = __pyx_state[49]; __pyx_result.record_allow_extra_fields = __pyx_state[50]; __pyx_result.record_can_encode_dict = __pyx_state[51]; __pyx_result.record_decodes_to_dict = __pyx_state[52]; __pyx_result.record_encode_use_defaults = __pyx_state[53]; __pyx_result.record_fields_must_be_unique = __pyx_state[54]; __pyx_result.record_values_type_hint = __pyx_state[55]; __pyx_result.return_uuid_object = __pyx_state[56]; __pyx_result.string_types_default_unchanged = __pyx_state[57]; __pyx_result.truncate_fixed = __pyx_state[58]; __pyx_result.types_str_to_schema = __pyx_state[59]; __pyx_result.unicode_errors = __pyx_state[60]; __pyx_result.union_decodes_to = __pyx_state[61]; __pyx_result.union_json_encodes_type_name = __pyx_state[62]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 63 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[63])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->adapt_record_types = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->allow_aliases_to_be_string = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->allow_duplicate_union_types = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->allow_empty_unions = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->allow_error_type = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->allow_false_values_for_null = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->allow_invalid_default_values = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->allow_leading_dot_in_names = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->allow_nested_unions = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->allow_primitive_name_collision = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->allow_primitive_names_in_namespaces = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->allow_tuple_notation = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->allow_union_default_any_member = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->alternate_timestamp_millis_encoding = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->ascii_name_rules = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->bytes_codec);
  __Pyx_DECREF(__pyx_v___pyx_result->bytes_codec);
  __pyx_v___pyx_result->bytes_codec = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->bytes_default_value_utf8 = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 17, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->canonical_form_repeat_enum = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 18, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->canonical_form_repeat_fixed = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 19, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->clamp_float_overflow = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 20, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->clamp_int_overflow = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 21, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->coerce_int_to_float = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 22, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->coerce_values_to_boolean = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 23, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->coerce_values_to_float = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 24, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->coerce_values_to_int = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 25, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->coerce_values_to_str = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 26, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->container_fill_blocks = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 27, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->date_type_accepts_string = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 28, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->decimal_check_exp_overflow = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 29, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->defer_schema_promotion_errors = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 30, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->enforce_enum_symbol_name_rules = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 31, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->enforce_namespace_name_rules = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 32, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->enforce_type_name_rules = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 33, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->enum_symbols_must_be_unique = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 34, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->expand_types_in_schema = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 35, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->externally_defined_types);
  __Pyx_DECREF(__pyx_v___pyx_result->externally_defined_types);
  __pyx_v___pyx_result->externally_defined_types = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 36, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->fingerprint_returns_digest = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 37, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->fixed_codec);
  __Pyx_DECREF(__pyx_v___pyx_result->fixed_codec);
  __pyx_v___pyx_result->fixed_codec = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 38, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->float_out_of_range_inf = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 39, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->inline_namespaces = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 40, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->invalid_value_include_array_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 41, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->invalid_value_include_map_key = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 42, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->invalid_value_includes_record_name = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 43, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->logical_types);
  __Pyx_DECREF(__pyx_v___pyx_result->logical_types);
  __pyx_v___pyx_result->logical_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 44, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->missing_values_can_be_empty_container = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 45, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->missing_values_can_be_null = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 46, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->naive_dt_assume_utc = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 47, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->named_type_names_must_be_unique = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 48, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->null_pad_fixed = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 49, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->raise_on_invalid_logical = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 50, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->record_allow_extra_fields = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 51, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->record_can_encode_dict = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 52, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->record_decodes_to_dict = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 53, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->record_encode_use_defaults = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 54, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->record_fields_must_be_unique = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 55, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->record_values_type_hint = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 56, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->return_uuid_object = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 57, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->string_types_default_unchanged = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 58, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->truncate_fixed = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 59, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->types_str_to_schema = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 60, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->unicode_errors);
  __Pyx_DECREF(__pyx_v___pyx_result->unicode_errors);
  __pyx_v___pyx_result->unicode_errors = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 61, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->union_decodes_to);
  __Pyx_DECREF(__pyx_v___pyx_result->union_decodes_to);
  __pyx_v___pyx_result->union_decodes_to = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 62, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->union_json_encodes_type_name = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Options__set_state(Options __pyx_result, tuple __pyx_state):
 *     __pyx_result.adapt_record_types = __pyx_state[0]; __pyx_result.allow_aliases_to_be_string = __pyx_state[1]; __pyx_result.allow_duplicate_union_types = __pyx_state[2]; __pyx_result.allow_empty_unions = __pyx_state[3]; __pyx_result.allow_error_type = __pyx_state[4]; __pyx_result.allow_false_values_for_null = __pyx_state[5]; __pyx_result.allow_invalid_default_values = __pyx_state[6]; __pyx_result.allow_leading_dot_in_names = __pyx_state[7]; __pyx_result.allow_nested_unions = __pyx_state[8]; __pyx_result.allow_primitive_name_collision = __pyx_state[9]; __pyx_result.allow_primitive_names_in_namespaces = __pyx_state[10]; __pyx_result.allow_tuple_notation = __pyx_state[11]; __pyx_result.allow_union_default_any_member = __pyx_state[12]; __pyx_result.alternate_timestamp_millis_encoding = __pyx_state[13]; __pyx_result.ascii_name_rules = __pyx_state[14]; __pyx_result.bytes_codec = __pyx_state[15]; __pyx_result.bytes_default_value_utf8 = __pyx_state[16]; __pyx_result.canonical_form_repeat_enum = __pyx_state[17]; __pyx_result.canonical_form_repeat_fixed = __pyx_state[18]; __pyx_result.clamp_float_overflow = __pyx_state[19]; __pyx_result.clamp_int_overflow = __pyx_state[20]; __pyx_result.coerce_int_to_float = __pyx_state[21]; __pyx_result.coerce_values_to_boolean = __pyx_state[22]; __pyx_result.coerce_values_to_float = __pyx_state[23]; __pyx_result.coerce_values_to_int = __pyx_state[24]; __pyx_result.coerce_values_to_str = __pyx_state[25]; __pyx_result.container_fill_blocks = __pyx_state[26]; __pyx_result.date_type_accepts_string = __pyx_state[27]; __pyx_result.decimal_check_exp_overflow = __pyx_state[28]; __pyx_result.defer_schema_promotion_errors = __pyx_state[29]; __pyx_result.enforce_enum_symbol_name_rules = __pyx_state[30]; __pyx_result.enforce_namespace_name_rules = __pyx_state[31]; __pyx_result.enforce_type_name_rules = __pyx_state[32]; __pyx_result.enum_symbols_must_be_unique = __pyx_state[33]; __pyx_result.expand_types_in_schema = __pyx_state[34]; __pyx_result.externally_defined_types = __pyx_state[35]; __pyx_result.fingerprint_returns_digest = __pyx_state[36]; __pyx_result.fixed_codec = __pyx_state[37]; __pyx_result.float_out_of_range_inf = __pyx_state[38]; __pyx_result.inline_namespaces = __pyx_state[39]; __pyx_result.invalid_value_include_array_index = __pyx_state[40]; __pyx_result.invalid_value_include_map_key = __pyx_state[41]; __pyx_result.invalid_value_includes_record_name = __pyx_state[42]; __pyx_result.logical_types = __pyx_state[43]; __pyx_result.missing_values_can_be_empty_container = __pyx_state[44]; __pyx_result.missing_values_can_be_null = __pyx_state[45]; __pyx_result.naive_dt_assume_utc = __pyx_state[46]; __pyx_result.named_type_names_must_be_unique = __pyx_state[47]; __pyx_result.null_pad_fixed = __pyx_state[48]; __pyx_result.raise_on_invalid_logical = __pyx_state[49]; __pyx_result.record_allow_extra_fields = __pyx_state[50]; __pyx_result.record_can_encode_dict = __pyx_state[51]; __pyx_result.record_decodes_to_dict = __pyx_state[52]; __pyx_result.record_encode_use_defaults = __pyx_state[53]; __pyx_result.record_fields_must_be_unique = __pyx_state[54]; __pyx_result.record_values_type_hint = __pyx_state[55]; __pyx_result.return_uuid_object = __pyx_state[56]; __pyx_result.string_types_default_unchanged = __pyx_state[57]; __pyx_result.truncate_fixed = __pyx_state[58]; __pyx_result.types_str_to_schema = __pyx_state[59]; __pyx_result.unicode_errors = __pyx_state[60]; __pyx_result.union_decodes_to = __pyx_state[61]; __pyx_result.union_json_encodes_type_name = __pyx_state[62]
 *     if len(__pyx_state) > 63 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[63])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 63);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.adapt_record_types = __pyx_state[0]; __pyx_result.allow_aliases_to_be_string = __pyx_state[1]; __pyx_result.allow_duplicate_union_types = __pyx_state[2]; __pyx_result.allow_empty_unions = __pyx_state[3]; __pyx_result.allow_error_type = __pyx_state[4]; __pyx_result.allow_false_values_for_null = __pyx_state[5]; __pyx_result.allow_invalid_default_values = __pyx_state[6]; __pyx_result.allow_leading_dot_in_names = __pyx_state[7]; __pyx_result.allow_nested_unions = __pyx_state[8]; __pyx_result.allow_primitive_name_collision = __pyx_state[9]; __pyx_result.allow_primitive_names_in_namespaces = __pyx_state[10]; __pyx_result.allow_tuple_notation = __pyx_state[11]; __pyx_result.allow_union_default_any_member = __pyx_state[12]; __pyx_result.alternate_timestamp_millis_encoding = __pyx_state[13]; __pyx_result.ascii_name_rules = __pyx_state[14]; __pyx_result.bytes_codec = __pyx_state[15]; __pyx_result.bytes_default_value_utf8 = __pyx_state[16]; __pyx_result.canonical_form_repeat_enum = __pyx_state[17]; __pyx_result.canonical_form_repeat_fixed = __pyx_state[18]; __pyx_result.clamp_float_overflow = __pyx_state[19]; __pyx_result.clamp_int_overflow = __pyx_state[20]; __pyx_result.coerce_int_to_float = __pyx_state[21]; __pyx_result.coerce_values_to_boolean = __pyx_state[22]; __pyx_result.coerce_values_to_float = __pyx_state[23]; __pyx_result.coerce_values_to_int = __pyx_state[24]; __pyx_result.coerce_values_to_str = __pyx_state[25]; __pyx_result.container_fill_blocks = __pyx_state[26]; __pyx_result.date_type_accepts_string = __pyx_state[27]; __pyx_result.decimal_check_exp_overflow = __pyx_state[28]; __pyx_result.defer_schema_promotion_errors = __pyx_state[29]; __pyx_result.enforce_enum_symbol_name_rules = __pyx_state[30]; __pyx_result.enforce_namespace_name_rules = __pyx_state[31]; __pyx_result.enforce_type_name_rules = __pyx_state[32]; __pyx_result.enum_symbols_must_be_unique = __pyx_state[33]; __pyx_result.expand_types_in_schema = __pyx_state[34]; __pyx_result.externally_defined_types = __pyx_state[35]; __pyx_result.fingerprint_returns_digest = __pyx_state[36]; __pyx_result.fixed_codec = __pyx_state[37]; __pyx_result.float_out_of_range_inf = __pyx_state[38]; __pyx_result.inline_namespaces = __pyx_state[39]; __pyx_result.invalid_value_include_array_index = __pyx_state[40]; __pyx_result.invalid_value_include_map_key = __pyx_state[41]; __pyx_result.invalid_value_includes_record_name = __pyx_state[42]; __pyx_result.logical_types = __pyx_state[43]; __pyx_result.missing_values_can_be_empty_container = __pyx_state[44]; __pyx_result.missing_values_can_be_null = __pyx_state[45]; __pyx_result.naive_dt_assume_utc = __pyx_state[46]; __pyx_result.named_type_names_must_be_unique = __pyx_state[47]; __pyx_result.null_pad_fixed = __pyx_state[48]; __pyx_result.raise_on_invalid_logical = __pyx_state[49]; __pyx_result.record_allow_extra_fields = __pyx_state[50]; __pyx_result.record_can_encode_dict = __pyx_state[51]; __pyx_result.record_decodes_to_dict = __pyx_state[52]; __pyx_result.record_encode_use_defaults = __pyx_state[53]; __pyx_result.record_fields_must_be_unique = __pyx_state[54]; __pyx_result.record_values_type_hint = __pyx_state[55]; __pyx_result.return_uuid_object = __pyx_state[56]; __pyx_result.string_types_default_unchanged = __pyx_state[57]; __pyx_result.truncate_fixed = __pyx_state[58]; __pyx_result.types_str_to_schema = __pyx_state[59]; __pyx_result.unicode_errors = __pyx_state[60]; __pyx_result.union_decodes_to = __pyx_state[61]; __pyx_result.union_json_encodes_type_name = __pyx_state[62]
 *     if len(__pyx_state) > 63 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[63])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 63, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Options__set_state(Options __pyx_result, tuple __pyx_state):
 *     __pyx_result.adapt_record_types = __pyx_state[0]; __pyx_result.allow_aliases_to_be_string = __pyx_state[1]; __pyx_result.allow_duplicate_union_types = __pyx_state[2]; __pyx_result.allow_empty_unions = __pyx_state[3]; __pyx_result.allow_error_type = __pyx_state[4]; __pyx_result.allow_false_values_for_null = __pyx_state[5]; __pyx_result.allow_invalid_default_values = __pyx_state[6]; __pyx_result.allow_leading_dot_in_names = __pyx_state[7]; __pyx_result.allow_nested_unions = __pyx_state[8]; __pyx_result.allow_primitive_name_collision = __pyx_state[9]; __pyx_result.allow_primitive_names_in_namespaces = __pyx_state[10]; __pyx_result.allow_tuple_notation = __pyx_state[11]; __pyx_result.allow_union_default_any_member = __pyx_state[12]; __pyx_result.alternate_timestamp_millis_encoding = __pyx_state[13]; __pyx_result.ascii_name_rules = __pyx_state[14]; __pyx_result.bytes_codec = __pyx_state[15]; __pyx_result.bytes_default_value_utf8 = __pyx_state[16]; __pyx_result.canonical_form_repeat_enum = __pyx_state[17]; __pyx_result.canonical_form_repeat_fixed = __pyx_state[18]; __pyx_result.clamp_float_overflow = __pyx_state[19]; __pyx_result.clamp_int_overflow = __pyx_state[20]; __pyx_result.coerce_int_to_float = __pyx_state[21]; __pyx_result.coerce_values_to_boolean = __pyx_state[22]; __pyx_result.coerce_values_to_float = __pyx_state[23]; __pyx_result.coerce_values_to_int = __pyx_state[24]; __pyx_result.coerce_values_to_str = __pyx_state[25]; __pyx_result.container_fill_blocks = __pyx_state[26]; __pyx_result.date_type_accepts_string = __pyx_state[27]; __pyx_result.decimal_check_exp_overflow = __pyx_state[28]; __pyx_result.defer_schema_promotion_errors = __pyx_state[29]; __pyx_result.enforce_enum_symbol_name_rules = __pyx_state[30]; __pyx_result.enforce_namespace_name_rules = __pyx_state[31]; __pyx_result.enforce_type_name_rules = __pyx_state[32]; __pyx_result.enum_symbols_must_be_unique = __pyx_state[33]; __pyx_result.expand_types_in_schema = __pyx_state[34]; __pyx_result.externally_defined_types = __pyx_state[35]; __pyx_result.fingerprint_returns_digest = __pyx_state[36]; __pyx_result.fixed_codec = __pyx_state[37]; __pyx_result.float_out_of_range_inf = __pyx_state[38]; __pyx_result.inline_namespaces = __pyx_state[39]; __pyx_result.invalid_value_include_array_index = __pyx_state[40]; __pyx_result.invalid_value_include_map_key = __pyx_state[41]; __pyx_result.invalid_value_includes_record_name = __pyx_state[42]; __pyx_result.logical_types = __pyx_state[43]; __pyx_result.missing_values_can_be_empty_container = __pyx_state[44]; __pyx_result.missing_values_can_be_null = __pyx_state[45]; __pyx_result.naive_dt_assume_utc = __pyx_state[46]; __pyx_result.named_type_names_must_be_unique = __pyx_state[47]; __pyx_result.null_pad_fixed = __pyx_state[48]; __pyx_result.raise_on_invalid_logical = __pyx_state[49]; __pyx_result.record_allow_extra_fields = __pyx_state[50]; __pyx_result.record_can_encode_dict = __pyx_state[51]; __pyx_result.record_decodes_to_dict = __pyx_state[52]; __pyx_result.record_encode_use_defaults = __pyx_state[53]; __pyx_result.record_fields_must_be_unique = __pyx_state[54]; __pyx_result.record_values_type_hint = __pyx_state[55]; __pyx_result.return_uuid_object = __pyx_state[56]; __pyx_result.string_types_default_unchanged = __pyx_state[57]; __pyx_result.truncate_fixed = __pyx_state[58]; __pyx_result.types_str_to_schema = __pyx_state[59]; __pyx_result.unicode_errors = __pyx_state[60]; __pyx_result.union_decodes_to = __pyx_state[61]; __pyx_result.union_json_encodes_type_name = __pyx_state[62]
 *     if len(__pyx_state) > 63 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[63])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Options__set_state(<Options> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Options__set_state(Options __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.adapt_record_types = __pyx_state[0]; __pyx_result.allow_aliases_to_be_string = __pyx_state[1]; __pyx_result.allow_duplicate_union_types = __pyx_state[2]; __pyx_result.allow_empty_unions = __pyx_state[3]; __pyx_result.allow_error_type = __pyx_state[4]; __pyx_result.allow_false_values_for_null = __pyx_state[5]; __pyx_result.allow_invalid_default_values = __pyx_state[6]; __pyx_result.allow_leading_dot_in_names = __pyx_state[7]; __pyx_result.allow_nested_unions = __pyx_state[8]; __pyx_result.allow_primitive_name_collision = __pyx_state[9]; __pyx_result.allow_primitive_names_in_namespaces = __pyx_state[10]; __pyx_result.allow_tuple_notation = __pyx_state[11]; __pyx_result.allow_union_default_any_member = __pyx_state[12]; __pyx_result.alternate_timestamp_millis_encoding = __pyx_state[13]; __pyx_result.ascii_name_rules = __pyx_state[14]; __pyx_result.bytes_codec = __pyx_state[15]; __pyx_result.bytes_default_value_utf8 = __pyx_state[16]; __pyx_result.canonical_form_repeat_enum = __pyx_state[17]; __pyx_result.canonical_form_repeat_fixed = __pyx_state[18]; __pyx_result.clamp_float_overflow = __pyx_state[19]; __pyx_result.clamp_int_overflow = __pyx_state[20]; __pyx_result.coerce_int_to_float = __pyx_state[21]; __pyx_result.coerce_values_to_boolean = __pyx_state[22]; __pyx_result.coerce_values_to_float = __pyx_state[23]; __pyx_result.coerce_values_to_int = __pyx_state[24]; __pyx_result.coerce_values_to_str = __pyx_state[25]; __pyx_result.container_fill_blocks = __pyx_state[26]; __pyx_result.date_type_accepts_string = __pyx_state[27]; __pyx_result.decimal_check_exp_overflow = __pyx_state[28]; __pyx_result.defer_schema_promotion_errors = __pyx_state[29]; __pyx_result.enforce_enum_symbol_name_rules = __pyx_state[30]; __pyx_result.enforce_namespace_name_rules = __pyx_state[31]; __pyx_result.enforce_type_name_rules = __pyx_state[32]; __pyx_result.enum_symbols_must_be_unique = __pyx_state[33]; __pyx_result.expand_types_in_schema = __pyx_state[34]; __pyx_result.externally_defined_types = __pyx_state[35]; __pyx_result.fingerprint_returns_digest = __pyx_state[36]; __pyx_result.fixed_codec = __pyx_state[37]; __pyx_result.float_out_of_range_inf = __pyx_state[38]; __pyx_result.inline_namespaces = __pyx_state[39]; __pyx_result.invalid_value_include_array_index = __pyx_state[40]; __pyx_result.invalid_value_include_map_key = __pyx_state[41]; __pyx_result.invalid_value_includes_record_name = __pyx_state[42]; __pyx_result.logical_types = __pyx_state[43]; __pyx_result.missing_values_can_be_empty_container = __pyx_state[44]; __pyx_result.missing_values_can_be_null = __pyx_state[45]; __pyx_result.naive_dt_assume_utc = __pyx_state[46]; __pyx_result.named_type_names_must_be_unique = __pyx_state[47]; __pyx_result.null_pad_fixed = __pyx_state[48]; __pyx_result.raise_on_invalid_logical = __pyx_state[49]; __pyx_result.record_allow_extra_fields = __pyx_state[50]; __pyx_result.record_can_encode_dict = __pyx_state[51]; __pyx_result.record_decodes_to_dict = __pyx_state[52]; __pyx_result.record_encode_use_defaults = __pyx_state[53]; __pyx_result.record_fields_must_be_unique = __pyx_state[54]; __pyx_result.record_values_type_hint = __pyx_state[55]; __pyx_result.return_uuid_object = __pyx_state[56]; __pyx_result.string_types_default_unchanged = __pyx_state[57]; __pyx_result.truncate_fixed = __pyx_state[58]; __pyx_result.types_str_to_schema = __pyx_state[59]; __pyx_result.unicode_errors = __pyx_state[60]; __pyx_result.union_decodes_to = __pyx_state[61]; __pyx_result.union_json_encodes_type_name = __pyx_state[62]
 *     if len(__pyx_state) > 63 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Options__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__Writer(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_21__pyx_unpickle__Writer(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_21__pyx_unpickle__Writer = {"__pyx_unpickle__Writer", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_21__pyx_unpickle__Writer, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_21__pyx_unpickle__Writer(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__Writer (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__Writer", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__Writer", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__Writer") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__Writer", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__Writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_20__pyx_unpickle__Writer(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_20__pyx_unpickle__Writer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__Writer", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _Writer.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _Writer.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _Writer.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__Writer__set_state(<_Writer> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__Writer), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _Writer.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__Writer__set_state(<_Writer> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _Writer.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__Writer__set_state(<_Writer> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__Writer__set_state(_Writer __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__Writer__set_state(((struct __pyx_obj_5cavro__Writer *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _Writer.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__Writer__set_state(<_Writer> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__Writer__set_state(<_Writer> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__Writer__set_state(_Writer __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__Writer(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__Writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__Writer__set_state(<_Writer> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__Writer__set_state(_Writer __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__Writer__set_state(struct __pyx_obj_5cavro__Writer *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__Writer__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__Writer__set_state(_Writer __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__Writer__set_state(_Writer __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__Writer__set_state(_Writer __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__Writer__set_state(<_Writer> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__Writer__set_state(_Writer __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__Writer__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__Reader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_23__pyx_unpickle__Reader(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_23__pyx_unpickle__Reader = {"__pyx_unpickle__Reader", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_23__pyx_unpickle__Reader, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_23__pyx_unpickle__Reader(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__Reader (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__Reader", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__Reader", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__Reader") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__Reader", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__Reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_22__pyx_unpickle__Reader(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_22__pyx_unpickle__Reader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__Reader", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _Reader.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _Reader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _Reader.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__Reader__set_state(<_Reader> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__Reader), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _Reader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__Reader__set_state(<_Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _Reader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__Reader__set_state(<_Reader> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__Reader__set_state(_Reader __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__Reader__set_state(((struct __pyx_obj_5cavro__Reader *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _Reader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__Reader__set_state(<_Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__Reader__set_state(<_Reader> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__Reader__set_state(_Reader __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__Reader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__Reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__Reader__set_state(<_Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__Reader__set_state(_Reader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__Reader__set_state(struct __pyx_obj_5cavro__Reader *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__Reader__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__Reader__set_state(_Reader __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__Reader__set_state(_Reader __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__Reader__set_state(_Reader __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__Reader__set_state(<_Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__Reader__set_state(_Reader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__Reader__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_FileReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_25__pyx_unpickle_FileReader(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_25__pyx_unpickle_FileReader = {"__pyx_unpickle_FileReader", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_25__pyx_unpickle_FileReader, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_25__pyx_unpickle_FileReader(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_FileReader (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FileReader", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FileReader", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_FileReader") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FileReader", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_FileReader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_24__pyx_unpickle_FileReader(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_24__pyx_unpickle_FileReader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FileReader", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x74c6201, 0x3a98199, 0x260679c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__97, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x74c6201, 0x3a98199, 0x260679c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum
 *     __pyx_result = FileReader.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x74c6201, 0x3a98199, 0x260679c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = FileReader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x74c6201, 0x3a98199, 0x260679c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum
 *     __pyx_result = FileReader.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FileReader__set_state(<FileReader> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_FileReader), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum
 *     __pyx_result = FileReader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FileReader__set_state(<FileReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = FileReader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FileReader__set_state(<FileReader> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_FileReader__set_state(FileReader __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_FileReader__set_state(((struct __pyx_obj_5cavro_FileReader *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum
 *     __pyx_result = FileReader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FileReader__set_state(<FileReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FileReader__set_state(<FileReader> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_FileReader__set_state(FileReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.file_obj = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FileReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_FileReader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_FileReader__set_state(<FileReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FileReader__set_state(FileReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.file_obj = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_FileReader__set_state(struct __pyx_obj_5cavro_FileReader *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FileReader__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_FileReader__set_state(FileReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.file_obj = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->file_obj);
  __Pyx_DECREF(__pyx_v___pyx_result->file_obj);
  __pyx_v___pyx_result->file_obj = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_FileReader__set_state(FileReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.file_obj = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.file_obj = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_FileReader__set_state(FileReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.file_obj = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_FileReader__set_state(<FileReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FileReader__set_state(FileReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.file_obj = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_FileReader__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_FileWriter(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_27__pyx_unpickle_FileWriter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_27__pyx_unpickle_FileWriter = {"__pyx_unpickle_FileWriter", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_27__pyx_unpickle_FileWriter, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_27__pyx_unpickle_FileWriter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_FileWriter (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FileWriter", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FileWriter", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_FileWriter") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FileWriter", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_FileWriter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_26__pyx_unpickle_FileWriter(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_26__pyx_unpickle_FileWriter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FileWriter", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x74c6201, 0x3a98199, 0x260679c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__97, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x74c6201, 0x3a98199, 0x260679c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum
 *     __pyx_result = FileWriter.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x74c6201, 0x3a98199, 0x260679c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = FileWriter.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x74c6201, 0x3a98199, 0x260679c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum
 *     __pyx_result = FileWriter.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FileWriter__set_state(<FileWriter> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_FileWriter), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum
 *     __pyx_result = FileWriter.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FileWriter__set_state(<FileWriter> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = FileWriter.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FileWriter__set_state(<FileWriter> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_FileWriter__set_state(FileWriter __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_FileWriter__set_state(((struct __pyx_obj_5cavro_FileWriter *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x74c6201, 0x3a98199, 0x260679c) = (file_obj))" % __pyx_checksum
 *     __pyx_result = FileWriter.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FileWriter__set_state(<FileWriter> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FileWriter__set_state(<FileWriter> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_FileWriter__set_state(FileWriter __pyx_result, tuple __pyx_state):
 *     __pyx_result.file_obj = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FileWriter(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_FileWriter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_FileWriter__set_state(<FileWriter> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FileWriter__set_state(FileWriter __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.file_obj = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_FileWriter__set_state(struct __pyx_obj_5cavro_FileWriter *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FileWriter__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_FileWriter__set_state(FileWriter __pyx_result, tuple __pyx_state):
 *     __pyx_result.file_obj = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->file_obj);
  __Pyx_DECREF(__pyx_v___pyx_result->file_obj);
  __pyx_v___pyx_result->file_obj = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_FileWriter__set_state(FileWriter __pyx_result, tuple __pyx_state):
 *     __pyx_result.file_obj = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.file_obj = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_FileWriter__set_state(FileWriter __pyx_result, tuple __pyx_state):
 *     __pyx_result.file_obj = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_FileWriter__set_state(<FileWriter> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FileWriter__set_state(FileWriter __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.file_obj = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_FileWriter__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_MemoryReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_29__pyx_unpickle_MemoryReader(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_29__pyx_unpickle_MemoryReader = {"__pyx_unpickle_MemoryReader", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_29__pyx_unpickle_MemoryReader, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_29__pyx_unpickle_MemoryReader(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_MemoryReader (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MemoryReader", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MemoryReader", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_MemoryReader") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MemoryReader", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_MemoryReader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_28__pyx_unpickle_MemoryReader(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_28__pyx_unpickle_MemoryReader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MemoryReader", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xec0f6ac, 0xbcf8db1, 0xf4af146):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xec0f6ac, 0xbcf8db1, 0xf4af146) = (data, end_ptr, ptr))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__98, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xec0f6ac, 0xbcf8db1, 0xf4af146):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xec0f6ac, 0xbcf8db1, 0xf4af146) = (data, end_ptr, ptr))" % __pyx_checksum
 *     __pyx_result = MemoryReader.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xec0f6ac, 0xbcf8db1, 0xf4af146):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xec0f6ac, 0xbcf8db1, 0xf4af146) = (data, end_ptr, ptr))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = MemoryReader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xec0f6ac, 0xbcf8db1, 0xf4af146):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xec0f6ac, 0xbcf8db1, 0xf4af146) = (data, end_ptr, ptr))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xec0f6ac, 0xbcf8db1, 0xf4af146) = (data, end_ptr, ptr))" % __pyx_checksum
 *     __pyx_result = MemoryReader.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MemoryReader__set_state(<MemoryReader> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_MemoryReader), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xec0f6ac, 0xbcf8db1, 0xf4af146) = (data, end_ptr, ptr))" % __pyx_checksum
 *     __pyx_result = MemoryReader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MemoryReader__set_state(<MemoryReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = MemoryReader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MemoryReader__set_state(<MemoryReader> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryReader__set_state(MemoryReader __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_MemoryReader__set_state(((struct __pyx_obj_5cavro_MemoryReader *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xec0f6ac, 0xbcf8db1, 0xf4af146) = (data, end_ptr, ptr))" % __pyx_checksum
 *     __pyx_result = MemoryReader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MemoryReader__set_state(<MemoryReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MemoryReader__set_state(<MemoryReader> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_MemoryReader__set_state(MemoryReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end_ptr = __pyx_state[1]; __pyx_result.ptr = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_MemoryReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_MemoryReader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_MemoryReader__set_state(<MemoryReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryReader__set_state(MemoryReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end_ptr = __pyx_state[1]; __pyx_result.ptr = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_MemoryReader__set_state(struct __pyx_obj_5cavro_MemoryReader *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  uint8_t const *__pyx_t_3;
  uint8_t const *__pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MemoryReader__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryReader__set_state(MemoryReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end_ptr = __pyx_state[1]; __pyx_result.ptr = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_1, 0); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XCLEAR_MEMVIEW(&__pyx_v___pyx_result->data, 0);
  __pyx_v___pyx_result->data = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_AsUString(__pyx_t_1); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->end_ptr = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_AsUString(__pyx_t_1); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->ptr = __pyx_t_4;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_MemoryReader__set_state(MemoryReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end_ptr = __pyx_state[1]; __pyx_result.ptr = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 > 3);
  if (__pyx_t_7) {
  } else {
    __pyx_t_5 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_5) {

    /* "(tree fragment)":14
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end_ptr = __pyx_state[1]; __pyx_result.ptr = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_MemoryReader__set_state(MemoryReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end_ptr = __pyx_state[1]; __pyx_result.ptr = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_MemoryReader__set_state(<MemoryReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryReader__set_state(MemoryReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.data = __pyx_state[0]; __pyx_result.end_ptr = __pyx_state[1]; __pyx_result.ptr = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_2, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_MemoryReader__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Rabin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_31__pyx_unpickle_Rabin(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_31__pyx_unpickle_Rabin = {"__pyx_unpickle_Rabin", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_31__pyx_unpickle_Rabin, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_31__pyx_unpickle_Rabin(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Rabin (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Rabin", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Rabin", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Rabin") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Rabin", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Rabin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_30__pyx_unpickle_Rabin(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_30__pyx_unpickle_Rabin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Rabin", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcd42404, 0xf32b67c, 0x2063c16):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__99, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcd42404, 0xf32b67c, 0x2063c16):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum
 *     __pyx_result = Rabin.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xcd42404, 0xf32b67c, 0x2063c16):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Rabin.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcd42404, 0xf32b67c, 0x2063c16):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum
 *     __pyx_result = Rabin.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Rabin__set_state(<Rabin> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_Rabin), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum
 *     __pyx_result = Rabin.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Rabin__set_state(<Rabin> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Rabin.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Rabin__set_state(<Rabin> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Rabin__set_state(Rabin __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_Rabin__set_state(((struct __pyx_obj_5cavro_Rabin *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcd42404, 0xf32b67c, 0x2063c16) = (value))" % __pyx_checksum
 *     __pyx_result = Rabin.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Rabin__set_state(<Rabin> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Rabin__set_state(<Rabin> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Rabin__set_state(Rabin __pyx_result, tuple __pyx_state):
 *     __pyx_result.value = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Rabin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Rabin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Rabin__set_state(<Rabin> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Rabin__set_state(Rabin __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.value = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_Rabin__set_state(struct __pyx_obj_5cavro_Rabin *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Rabin__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Rabin__set_state(Rabin __pyx_result, tuple __pyx_state):
 *     __pyx_result.value = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->value = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Rabin__set_state(Rabin __pyx_result, tuple __pyx_state):
 *     __pyx_result.value = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 1);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.value = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Rabin__set_state(Rabin __pyx_result, tuple __pyx_state):
 *     __pyx_result.value = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Rabin__set_state(<Rabin> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Rabin__set_state(Rabin __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.value = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Rabin__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ArrayType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_33__pyx_unpickle_ArrayType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_33__pyx_unpickle_ArrayType = {"__pyx_unpickle_ArrayType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_33__pyx_unpickle_ArrayType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_33__pyx_unpickle_ArrayType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ArrayType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ArrayType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ArrayType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_ArrayType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ArrayType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_ArrayType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_32__pyx_unpickle_ArrayType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_32__pyx_unpickle_ArrayType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ArrayType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcdf32d8, 0x60f601e, 0xa92a64f):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcdf32d8, 0x60f601e, 0xa92a64f) = (item_type, metadata, options, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__100, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcdf32d8, 0x60f601e, 0xa92a64f):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcdf32d8, 0x60f601e, 0xa92a64f) = (item_type, metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = ArrayType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xcdf32d8, 0x60f601e, 0xa92a64f):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcdf32d8, 0x60f601e, 0xa92a64f) = (item_type, metadata, options, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = ArrayType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcdf32d8, 0x60f601e, 0xa92a64f):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcdf32d8, 0x60f601e, 0xa92a64f) = (item_type, metadata, options, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcdf32d8, 0x60f601e, 0xa92a64f) = (item_type, metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = ArrayType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ArrayType__set_state(<ArrayType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_ArrayType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcdf32d8, 0x60f601e, 0xa92a64f) = (item_type, metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = ArrayType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ArrayType__set_state(<ArrayType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = ArrayType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ArrayType__set_state(<ArrayType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ArrayType__set_state(ArrayType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_ArrayType__set_state(((struct __pyx_obj_5cavro_ArrayType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcdf32d8, 0x60f601e, 0xa92a64f) = (item_type, metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = ArrayType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ArrayType__set_state(<ArrayType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ArrayType__set_state(<ArrayType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ArrayType__set_state(ArrayType __pyx_result, tuple __pyx_state):
 *     __pyx_result.item_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ArrayType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_ArrayType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ArrayType__set_state(<ArrayType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ArrayType__set_state(ArrayType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.item_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_ArrayType__set_state(struct __pyx_obj_5cavro_ArrayType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ArrayType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ArrayType__set_state(ArrayType __pyx_result, tuple __pyx_state):
 *     __pyx_result.item_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->item_type);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->item_type);
  __pyx_v___pyx_result->item_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_ArrayType__set_state(ArrayType __pyx_result, tuple __pyx_state):
 *     __pyx_result.item_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 4);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.item_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ArrayType__set_state(ArrayType __pyx_result, tuple __pyx_state):
 *     __pyx_result.item_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ArrayType__set_state(<ArrayType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ArrayType__set_state(ArrayType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.item_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_ArrayType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_EnumType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_35__pyx_unpickle_EnumType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_35__pyx_unpickle_EnumType = {"__pyx_unpickle_EnumType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_35__pyx_unpickle_EnumType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_35__pyx_unpickle_EnumType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_EnumType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EnumType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EnumType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_EnumType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_EnumType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_EnumType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_34__pyx_unpickle_EnumType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_34__pyx_unpickle_EnumType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_EnumType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8f63b31, 0x117bc6b, 0xb063604):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8f63b31, 0x117bc6b, 0xb063604) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, symbol_indexes, symbols, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__101, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8f63b31, 0x117bc6b, 0xb063604):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8f63b31, 0x117bc6b, 0xb063604) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, symbol_indexes, symbols, value_adapters))" % __pyx_checksum
 *     __pyx_result = EnumType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x8f63b31, 0x117bc6b, 0xb063604):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8f63b31, 0x117bc6b, 0xb063604) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, symbol_indexes, symbols, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = EnumType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8f63b31, 0x117bc6b, 0xb063604):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8f63b31, 0x117bc6b, 0xb063604) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, symbol_indexes, symbols, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8f63b31, 0x117bc6b, 0xb063604) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, symbol_indexes, symbols, value_adapters))" % __pyx_checksum
 *     __pyx_result = EnumType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EnumType__set_state(<EnumType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_EnumType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8f63b31, 0x117bc6b, 0xb063604) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, symbol_indexes, symbols, value_adapters))" % __pyx_checksum
 *     __pyx_result = EnumType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_EnumType__set_state(<EnumType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = EnumType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EnumType__set_state(<EnumType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_EnumType__set_state(EnumType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_EnumType__set_state(((struct __pyx_obj_5cavro_EnumType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8f63b31, 0x117bc6b, 0xb063604) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, symbol_indexes, symbols, value_adapters))" % __pyx_checksum
 *     __pyx_result = EnumType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_EnumType__set_state(<EnumType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_EnumType__set_state(<EnumType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_EnumType__set_state(EnumType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.symbol_indexes = __pyx_state[8]; __pyx_result.symbols = __pyx_state[9]; __pyx_result.value_adapters = __pyx_state[10]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_EnumType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_EnumType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_EnumType__set_state(<EnumType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_EnumType__set_state(EnumType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.symbol_indexes = __pyx_state[8]; __pyx_result.symbols = __pyx_state[9]; __pyx_result.value_adapters = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_EnumType__set_state(struct __pyx_obj_5cavro_EnumType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_EnumType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_EnumType__set_state(EnumType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.symbol_indexes = __pyx_state[8]; __pyx_result.symbols = __pyx_state[9]; __pyx_result.value_adapters = __pyx_state[10]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyFrozenSet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("frozenset", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.aliases);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.aliases);
  __pyx_v___pyx_result->__pyx_base.aliases = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->default_value);
  __Pyx_DECREF(__pyx_v___pyx_result->default_value);
  __pyx_v___pyx_result->default_value = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->doc);
  __Pyx_DECREF(__pyx_v___pyx_result->doc);
  __pyx_v___pyx_result->doc = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.effective_namespace);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.effective_namespace);
  __pyx_v___pyx_result->__pyx_base.effective_namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.name);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.name);
  __pyx_v___pyx_result->__pyx_base.name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.namespace);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.namespace);
  __pyx_v___pyx_result->__pyx_base.namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->symbol_indexes);
  __Pyx_DECREF(__pyx_v___pyx_result->symbol_indexes);
  __pyx_v___pyx_result->symbol_indexes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->symbols);
  __Pyx_DECREF(__pyx_v___pyx_result->symbols);
  __pyx_v___pyx_result->symbols = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_EnumType__set_state(EnumType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.symbol_indexes = __pyx_state[8]; __pyx_result.symbols = __pyx_state[9]; __pyx_result.value_adapters = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 11);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.symbol_indexes = __pyx_state[8]; __pyx_result.symbols = __pyx_state[9]; __pyx_result.value_adapters = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[11])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_EnumType__set_state(EnumType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.symbol_indexes = __pyx_state[8]; __pyx_result.symbols = __pyx_state[9]; __pyx_result.value_adapters = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[11])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_EnumType__set_state(<EnumType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_EnumType__set_state(EnumType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.symbol_indexes = __pyx_state[8]; __pyx_result.symbols = __pyx_state[9]; __pyx_result.value_adapters = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_EnumType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PromotingEnumType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_37__pyx_unpickle_PromotingEnumType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_37__pyx_unpickle_PromotingEnumType = {"__pyx_unpickle_PromotingEnumType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_37__pyx_unpickle_PromotingEnumType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_37__pyx_unpickle_PromotingEnumType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromotingEnumType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromotingEnumType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromotingEnumType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PromotingEnumType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromotingEnumType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromotingEnumType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_36__pyx_unpickle_PromotingEnumType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_36__pyx_unpickle_PromotingEnumType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromotingEnumType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8a61ff8, 0x47d8f26, 0xbae6ea0):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8a61ff8, 0x47d8f26, 0xbae6ea0) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, reader_type, symbol_indexes, symbols, unknown_symbols, value_adapters, writer_type))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__102, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8a61ff8, 0x47d8f26, 0xbae6ea0):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8a61ff8, 0x47d8f26, 0xbae6ea0) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, reader_type, symbol_indexes, symbols, unknown_symbols, value_adapters, writer_type))" % __pyx_checksum
 *     __pyx_result = PromotingEnumType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x8a61ff8, 0x47d8f26, 0xbae6ea0):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8a61ff8, 0x47d8f26, 0xbae6ea0) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, reader_type, symbol_indexes, symbols, unknown_symbols, value_adapters, writer_type))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PromotingEnumType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8a61ff8, 0x47d8f26, 0xbae6ea0):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8a61ff8, 0x47d8f26, 0xbae6ea0) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, reader_type, symbol_indexes, symbols, unknown_symbols, value_adapters, writer_type))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8a61ff8, 0x47d8f26, 0xbae6ea0) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, reader_type, symbol_indexes, symbols, unknown_symbols, value_adapters, writer_type))" % __pyx_checksum
 *     __pyx_result = PromotingEnumType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromotingEnumType__set_state(<PromotingEnumType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_PromotingEnumType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8a61ff8, 0x47d8f26, 0xbae6ea0) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, reader_type, symbol_indexes, symbols, unknown_symbols, value_adapters, writer_type))" % __pyx_checksum
 *     __pyx_result = PromotingEnumType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PromotingEnumType__set_state(<PromotingEnumType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PromotingEnumType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromotingEnumType__set_state(<PromotingEnumType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PromotingEnumType__set_state(PromotingEnumType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_PromotingEnumType__set_state(((struct __pyx_obj_5cavro_PromotingEnumType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8a61ff8, 0x47d8f26, 0xbae6ea0) = (aliases, default_value, doc, effective_namespace, metadata, name, namespace, options, reader_type, symbol_indexes, symbols, unknown_symbols, value_adapters, writer_type))" % __pyx_checksum
 *     __pyx_result = PromotingEnumType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PromotingEnumType__set_state(<PromotingEnumType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromotingEnumType__set_state(<PromotingEnumType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PromotingEnumType__set_state(PromotingEnumType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.reader_type = __pyx_state[8]; __pyx_result.symbol_indexes = __pyx_state[9]; __pyx_result.symbols = __pyx_state[10]; __pyx_result.unknown_symbols = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]; __pyx_result.writer_type = __pyx_state[13]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PromotingEnumType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromotingEnumType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PromotingEnumType__set_state(<PromotingEnumType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromotingEnumType__set_state(PromotingEnumType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.reader_type = __pyx_state[8]; __pyx_result.symbol_indexes = __pyx_state[9]; __pyx_result.symbols = __pyx_state[10]; __pyx_result.unknown_symbols = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]; __pyx_result.writer_type = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_PromotingEnumType__set_state(struct __pyx_obj_5cavro_PromotingEnumType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromotingEnumType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PromotingEnumType__set_state(PromotingEnumType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.reader_type = __pyx_state[8]; __pyx_result.symbol_indexes = __pyx_state[9]; __pyx_result.symbols = __pyx_state[10]; __pyx_result.unknown_symbols = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]; __pyx_result.writer_type = __pyx_state[13]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyFrozenSet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("frozenset", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.aliases);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.aliases);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.aliases = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.default_value);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.default_value);
  __pyx_v___pyx_result->__pyx_base.default_value = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.doc);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.doc);
  __pyx_v___pyx_result->__pyx_base.doc = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.effective_namespace);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.effective_namespace);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.effective_namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.name);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.name);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.namespace);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.namespace);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_EnumType))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->reader_type);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->reader_type);
  __pyx_v___pyx_result->reader_type = ((struct __pyx_obj_5cavro_EnumType *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.symbol_indexes);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.symbol_indexes);
  __pyx_v___pyx_result->__pyx_base.symbol_indexes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.symbols);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.symbols);
  __pyx_v___pyx_result->__pyx_base.symbols = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PySet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->unknown_symbols);
  __Pyx_DECREF(__pyx_v___pyx_result->unknown_symbols);
  __pyx_v___pyx_result->unknown_symbols = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_EnumType))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->writer_type);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->writer_type);
  __pyx_v___pyx_result->writer_type = ((struct __pyx_obj_5cavro_EnumType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PromotingEnumType__set_state(PromotingEnumType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.reader_type = __pyx_state[8]; __pyx_result.symbol_indexes = __pyx_state[9]; __pyx_result.symbols = __pyx_state[10]; __pyx_result.unknown_symbols = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]; __pyx_result.writer_type = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 14);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.reader_type = __pyx_state[8]; __pyx_result.symbol_indexes = __pyx_state[9]; __pyx_result.symbols = __pyx_state[10]; __pyx_result.unknown_symbols = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]; __pyx_result.writer_type = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PromotingEnumType__set_state(PromotingEnumType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.reader_type = __pyx_state[8]; __pyx_result.symbol_indexes = __pyx_state[9]; __pyx_result.symbols = __pyx_state[10]; __pyx_result.unknown_symbols = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]; __pyx_result.writer_type = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PromotingEnumType__set_state(<PromotingEnumType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromotingEnumType__set_state(PromotingEnumType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.reader_type = __pyx_state[8]; __pyx_result.symbol_indexes = __pyx_state[9]; __pyx_result.symbols = __pyx_state[10]; __pyx_result.unknown_symbols = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]; __pyx_result.writer_type = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromotingEnumType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_MapType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_39__pyx_unpickle_MapType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_39__pyx_unpickle_MapType = {"__pyx_unpickle_MapType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_39__pyx_unpickle_MapType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_39__pyx_unpickle_MapType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_MapType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MapType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MapType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_MapType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MapType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_MapType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_38__pyx_unpickle_MapType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_38__pyx_unpickle_MapType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MapType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaa8230c, 0x4c3ce07, 0x4f7d725):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaa8230c, 0x4c3ce07, 0x4f7d725) = (key_type, metadata, options, value_adapters, value_type))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__103, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaa8230c, 0x4c3ce07, 0x4f7d725):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaa8230c, 0x4c3ce07, 0x4f7d725) = (key_type, metadata, options, value_adapters, value_type))" % __pyx_checksum
 *     __pyx_result = MapType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xaa8230c, 0x4c3ce07, 0x4f7d725):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaa8230c, 0x4c3ce07, 0x4f7d725) = (key_type, metadata, options, value_adapters, value_type))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = MapType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xaa8230c, 0x4c3ce07, 0x4f7d725):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaa8230c, 0x4c3ce07, 0x4f7d725) = (key_type, metadata, options, value_adapters, value_type))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaa8230c, 0x4c3ce07, 0x4f7d725) = (key_type, metadata, options, value_adapters, value_type))" % __pyx_checksum
 *     __pyx_result = MapType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MapType__set_state(<MapType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_MapType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaa8230c, 0x4c3ce07, 0x4f7d725) = (key_type, metadata, options, value_adapters, value_type))" % __pyx_checksum
 *     __pyx_result = MapType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MapType__set_state(<MapType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = MapType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MapType__set_state(<MapType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_MapType__set_state(MapType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_MapType__set_state(((struct __pyx_obj_5cavro_MapType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xaa8230c, 0x4c3ce07, 0x4f7d725) = (key_type, metadata, options, value_adapters, value_type))" % __pyx_checksum
 *     __pyx_result = MapType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MapType__set_state(<MapType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MapType__set_state(<MapType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_MapType__set_state(MapType __pyx_result, tuple __pyx_state):
 *     __pyx_result.key_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]; __pyx_result.value_type = __pyx_state[4]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_MapType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_MapType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_MapType__set_state(<MapType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MapType__set_state(MapType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.key_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]; __pyx_result.value_type = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_MapType__set_state(struct __pyx_obj_5cavro_MapType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MapType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_MapType__set_state(MapType __pyx_result, tuple __pyx_state):
 *     __pyx_result.key_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]; __pyx_result.value_type = __pyx_state[4]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_StringType))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->key_type);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->key_type);
  __pyx_v___pyx_result->key_type = ((struct __pyx_obj_5cavro_StringType *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->value_type);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->value_type);
  __pyx_v___pyx_result->value_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_MapType__set_state(MapType __pyx_result, tuple __pyx_state):
 *     __pyx_result.key_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]; __pyx_result.value_type = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 5);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.key_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]; __pyx_result.value_type = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_MapType__set_state(MapType __pyx_result, tuple __pyx_state):
 *     __pyx_result.key_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]; __pyx_result.value_type = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_MapType__set_state(<MapType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MapType__set_state(MapType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.key_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]; __pyx_result.value_type = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_MapType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_NullType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_41__pyx_unpickle_NullType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_41__pyx_unpickle_NullType = {"__pyx_unpickle_NullType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_41__pyx_unpickle_NullType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_41__pyx_unpickle_NullType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_NullType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NullType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NullType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_NullType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NullType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_NullType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_40__pyx_unpickle_NullType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_40__pyx_unpickle_NullType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_NullType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__104, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = NullType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = NullType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = NullType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NullType__set_state(<NullType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_NullType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = NullType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NullType__set_state(<NullType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = NullType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NullType__set_state(<NullType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_NullType__set_state(NullType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_NullType__set_state(((struct __pyx_obj_5cavro_NullType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = NullType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NullType__set_state(<NullType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NullType__set_state(<NullType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_NullType__set_state(NullType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_NullType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_NullType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_NullType__set_state(<NullType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NullType__set_state(NullType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_NullType__set_state(struct __pyx_obj_5cavro_NullType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_NullType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_NullType__set_state(NullType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_NullType__set_state(NullType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_NullType__set_state(NullType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_NullType__set_state(<NullType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NullType__set_state(NullType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_NullType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_BoolType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_43__pyx_unpickle_BoolType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_43__pyx_unpickle_BoolType = {"__pyx_unpickle_BoolType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_43__pyx_unpickle_BoolType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_43__pyx_unpickle_BoolType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_BoolType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BoolType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BoolType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_BoolType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BoolType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_BoolType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_42__pyx_unpickle_BoolType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_42__pyx_unpickle_BoolType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BoolType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__104, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = BoolType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = BoolType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = BoolType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BoolType__set_state(<BoolType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_BoolType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = BoolType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BoolType__set_state(<BoolType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = BoolType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BoolType__set_state(<BoolType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_BoolType__set_state(BoolType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_BoolType__set_state(((struct __pyx_obj_5cavro_BoolType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = BoolType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BoolType__set_state(<BoolType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BoolType__set_state(<BoolType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_BoolType__set_state(BoolType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_BoolType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_BoolType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_BoolType__set_state(<BoolType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BoolType__set_state(BoolType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_BoolType__set_state(struct __pyx_obj_5cavro_BoolType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BoolType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_BoolType__set_state(BoolType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_BoolType__set_state(BoolType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_BoolType__set_state(BoolType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_BoolType__set_state(<BoolType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BoolType__set_state(BoolType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_BoolType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IntType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_45__pyx_unpickle_IntType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_45__pyx_unpickle_IntType = {"__pyx_unpickle_IntType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_45__pyx_unpickle_IntType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_45__pyx_unpickle_IntType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_IntType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_IntType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_44__pyx_unpickle_IntType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_44__pyx_unpickle_IntType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__104, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = IntType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = IntType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = IntType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntType__set_state(<IntType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_IntType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = IntType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IntType__set_state(<IntType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = IntType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntType__set_state(<IntType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IntType__set_state(IntType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_IntType__set_state(((struct __pyx_obj_5cavro_IntType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = IntType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IntType__set_state(<IntType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntType__set_state(<IntType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IntType__set_state(IntType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IntType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_IntType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IntType__set_state(<IntType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntType__set_state(IntType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_IntType__set_state(struct __pyx_obj_5cavro_IntType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IntType__set_state(IntType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_IntType__set_state(IntType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IntType__set_state(IntType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IntType__set_state(<IntType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntType__set_state(IntType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_IntType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_LongType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_47__pyx_unpickle_LongType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_47__pyx_unpickle_LongType = {"__pyx_unpickle_LongType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_47__pyx_unpickle_LongType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_47__pyx_unpickle_LongType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_LongType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LongType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LongType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_LongType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LongType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_LongType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_46__pyx_unpickle_LongType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_46__pyx_unpickle_LongType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LongType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__104, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = LongType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = LongType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = LongType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LongType__set_state(<LongType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_LongType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = LongType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LongType__set_state(<LongType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = LongType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LongType__set_state(<LongType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_LongType__set_state(LongType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_LongType__set_state(((struct __pyx_obj_5cavro_LongType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = LongType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LongType__set_state(<LongType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LongType__set_state(<LongType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_LongType__set_state(LongType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LongType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_LongType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_LongType__set_state(<LongType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LongType__set_state(LongType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_LongType__set_state(struct __pyx_obj_5cavro_LongType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LongType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_LongType__set_state(LongType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_LongType__set_state(LongType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_LongType__set_state(LongType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_LongType__set_state(<LongType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LongType__set_state(LongType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_LongType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_FloatType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_49__pyx_unpickle_FloatType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_49__pyx_unpickle_FloatType = {"__pyx_unpickle_FloatType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_49__pyx_unpickle_FloatType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_49__pyx_unpickle_FloatType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_FloatType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FloatType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FloatType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_FloatType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FloatType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_FloatType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_48__pyx_unpickle_FloatType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_48__pyx_unpickle_FloatType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FloatType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__104, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = FloatType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = FloatType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = FloatType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FloatType__set_state(<FloatType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_FloatType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = FloatType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FloatType__set_state(<FloatType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = FloatType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FloatType__set_state(<FloatType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_FloatType__set_state(FloatType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_FloatType__set_state(((struct __pyx_obj_5cavro_FloatType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = FloatType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FloatType__set_state(<FloatType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FloatType__set_state(<FloatType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_FloatType__set_state(FloatType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FloatType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_FloatType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_FloatType__set_state(<FloatType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FloatType__set_state(FloatType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_FloatType__set_state(struct __pyx_obj_5cavro_FloatType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FloatType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_FloatType__set_state(FloatType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_FloatType__set_state(FloatType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_FloatType__set_state(FloatType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_FloatType__set_state(<FloatType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FloatType__set_state(FloatType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_FloatType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_DoubleType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_51__pyx_unpickle_DoubleType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_51__pyx_unpickle_DoubleType = {"__pyx_unpickle_DoubleType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_51__pyx_unpickle_DoubleType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_51__pyx_unpickle_DoubleType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_DoubleType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DoubleType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DoubleType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_DoubleType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DoubleType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_DoubleType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_50__pyx_unpickle_DoubleType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_50__pyx_unpickle_DoubleType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DoubleType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__104, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = DoubleType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = DoubleType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = DoubleType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DoubleType__set_state(<DoubleType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_DoubleType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = DoubleType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DoubleType__set_state(<DoubleType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = DoubleType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DoubleType__set_state(<DoubleType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_DoubleType__set_state(DoubleType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_DoubleType__set_state(((struct __pyx_obj_5cavro_DoubleType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = DoubleType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DoubleType__set_state(<DoubleType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DoubleType__set_state(<DoubleType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_DoubleType__set_state(DoubleType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_DoubleType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_DoubleType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_DoubleType__set_state(<DoubleType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DoubleType__set_state(DoubleType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_DoubleType__set_state(struct __pyx_obj_5cavro_DoubleType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DoubleType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_DoubleType__set_state(DoubleType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_DoubleType__set_state(DoubleType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_DoubleType__set_state(DoubleType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_DoubleType__set_state(<DoubleType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DoubleType__set_state(DoubleType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_DoubleType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__PlaceholderType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_53__pyx_unpickle__PlaceholderType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_53__pyx_unpickle__PlaceholderType = {"__pyx_unpickle__PlaceholderType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_53__pyx_unpickle__PlaceholderType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_53__pyx_unpickle__PlaceholderType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__PlaceholderType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__PlaceholderType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__PlaceholderType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__PlaceholderType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__PlaceholderType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__PlaceholderType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_52__pyx_unpickle__PlaceholderType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_52__pyx_unpickle__PlaceholderType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__PlaceholderType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x5c2152b, 0x99117dc, 0xe166855):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5c2152b, 0x99117dc, 0xe166855) = (default_value, metadata, options, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__105, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x5c2152b, 0x99117dc, 0xe166855):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5c2152b, 0x99117dc, 0xe166855) = (default_value, metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = _PlaceholderType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x5c2152b, 0x99117dc, 0xe166855):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5c2152b, 0x99117dc, 0xe166855) = (default_value, metadata, options, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _PlaceholderType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x5c2152b, 0x99117dc, 0xe166855):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5c2152b, 0x99117dc, 0xe166855) = (default_value, metadata, options, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5c2152b, 0x99117dc, 0xe166855) = (default_value, metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = _PlaceholderType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__PlaceholderType__set_state(<_PlaceholderType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__PlaceholderType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5c2152b, 0x99117dc, 0xe166855) = (default_value, metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = _PlaceholderType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__PlaceholderType__set_state(<_PlaceholderType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _PlaceholderType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__PlaceholderType__set_state(<_PlaceholderType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__PlaceholderType__set_state(_PlaceholderType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__PlaceholderType__set_state(((struct __pyx_obj_5cavro__PlaceholderType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x5c2152b, 0x99117dc, 0xe166855) = (default_value, metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = _PlaceholderType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__PlaceholderType__set_state(<_PlaceholderType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__PlaceholderType__set_state(<_PlaceholderType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__PlaceholderType__set_state(_PlaceholderType __pyx_result, tuple __pyx_state):
 *     __pyx_result.default_value = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__PlaceholderType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__PlaceholderType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__PlaceholderType__set_state(<_PlaceholderType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__PlaceholderType__set_state(_PlaceholderType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.default_value = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__PlaceholderType__set_state(struct __pyx_obj_5cavro__PlaceholderType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__PlaceholderType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__PlaceholderType__set_state(_PlaceholderType __pyx_result, tuple __pyx_state):
 *     __pyx_result.default_value = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->default_value);
  __Pyx_DECREF(__pyx_v___pyx_result->default_value);
  __pyx_v___pyx_result->default_value = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle__PlaceholderType__set_state(_PlaceholderType __pyx_result, tuple __pyx_state):
 *     __pyx_result.default_value = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 4);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.default_value = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__PlaceholderType__set_state(_PlaceholderType __pyx_result, tuple __pyx_state):
 *     __pyx_result.default_value = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__PlaceholderType__set_state(<_PlaceholderType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__PlaceholderType__set_state(_PlaceholderType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.default_value = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__PlaceholderType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Record(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_55__pyx_unpickle_Record(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_55__pyx_unpickle_Record = {"__pyx_unpickle_Record", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_55__pyx_unpickle_Record, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_55__pyx_unpickle_Record(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Record (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Record", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Record", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Record") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Record", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_54__pyx_unpickle_Record(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_54__pyx_unpickle_Record(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Record", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3a6eb07, 0xa17c9aa, 0x8d777f3):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a6eb07, 0xa17c9aa, 0x8d777f3) = (data))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__106, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3a6eb07, 0xa17c9aa, 0x8d777f3):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a6eb07, 0xa17c9aa, 0x8d777f3) = (data))" % __pyx_checksum
 *     __pyx_result = Record.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x3a6eb07, 0xa17c9aa, 0x8d777f3):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a6eb07, 0xa17c9aa, 0x8d777f3) = (data))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Record.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x3a6eb07, 0xa17c9aa, 0x8d777f3):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a6eb07, 0xa17c9aa, 0x8d777f3) = (data))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a6eb07, 0xa17c9aa, 0x8d777f3) = (data))" % __pyx_checksum
 *     __pyx_result = Record.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Record__set_state(<Record> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_Record), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a6eb07, 0xa17c9aa, 0x8d777f3) = (data))" % __pyx_checksum
 *     __pyx_result = Record.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Record__set_state(<Record> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Record.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Record__set_state(<Record> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Record__set_state(Record __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_Record__set_state(((struct __pyx_obj_5cavro_Record *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x3a6eb07, 0xa17c9aa, 0x8d777f3) = (data))" % __pyx_checksum
 *     __pyx_result = Record.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Record__set_state(<Record> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Record__set_state(<Record> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Record__set_state(Record __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Record(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Record__set_state(<Record> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Record__set_state(Record __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.data = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_Record__set_state(struct __pyx_obj_5cavro_Record *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Record__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Record__set_state(Record __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->data);
  __Pyx_DECREF(__pyx_v___pyx_result->data);
  __pyx_v___pyx_result->data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Record__set_state(Record __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.data = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Record__set_state(Record __pyx_result, tuple __pyx_state):
 *     __pyx_result.data = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Record__set_state(<Record> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Record__set_state(Record __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.data = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Record__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_RecordField(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_57__pyx_unpickle_RecordField(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_57__pyx_unpickle_RecordField = {"__pyx_unpickle_RecordField", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_57__pyx_unpickle_RecordField, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_57__pyx_unpickle_RecordField(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_RecordField (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RecordField", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RecordField", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_RecordField") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RecordField", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_RecordField", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_56__pyx_unpickle_RecordField(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_56__pyx_unpickle_RecordField(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_RecordField", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1d9a0f8, 0x4d78739, 0x24a1a05):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1d9a0f8, 0x4d78739, 0x24a1a05) = (aliases, default_value, doc, name, order, type, writer_name))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__107, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1d9a0f8, 0x4d78739, 0x24a1a05):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1d9a0f8, 0x4d78739, 0x24a1a05) = (aliases, default_value, doc, name, order, type, writer_name))" % __pyx_checksum
 *     __pyx_result = RecordField.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x1d9a0f8, 0x4d78739, 0x24a1a05):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1d9a0f8, 0x4d78739, 0x24a1a05) = (aliases, default_value, doc, name, order, type, writer_name))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = RecordField.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1d9a0f8, 0x4d78739, 0x24a1a05):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1d9a0f8, 0x4d78739, 0x24a1a05) = (aliases, default_value, doc, name, order, type, writer_name))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1d9a0f8, 0x4d78739, 0x24a1a05) = (aliases, default_value, doc, name, order, type, writer_name))" % __pyx_checksum
 *     __pyx_result = RecordField.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RecordField__set_state(<RecordField> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_RecordField), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1d9a0f8, 0x4d78739, 0x24a1a05) = (aliases, default_value, doc, name, order, type, writer_name))" % __pyx_checksum
 *     __pyx_result = RecordField.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_RecordField__set_state(<RecordField> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = RecordField.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RecordField__set_state(<RecordField> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_RecordField__set_state(RecordField __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_RecordField__set_state(((struct __pyx_obj_5cavro_RecordField *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1d9a0f8, 0x4d78739, 0x24a1a05) = (aliases, default_value, doc, name, order, type, writer_name))" % __pyx_checksum
 *     __pyx_result = RecordField.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_RecordField__set_state(<RecordField> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RecordField__set_state(<RecordField> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_RecordField__set_state(RecordField __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.order = __pyx_state[4]; __pyx_result.type = __pyx_state[5]; __pyx_result.writer_name = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_RecordField(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_RecordField", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_RecordField__set_state(<RecordField> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RecordField__set_state(RecordField __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.order = __pyx_state[4]; __pyx_result.type = __pyx_state[5]; __pyx_result.writer_name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_RecordField__set_state(struct __pyx_obj_5cavro_RecordField *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_RecordField__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_RecordField__set_state(RecordField __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.order = __pyx_state[4]; __pyx_result.type = __pyx_state[5]; __pyx_result.writer_name = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyFrozenSet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("frozenset", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->aliases);
  __Pyx_DECREF(__pyx_v___pyx_result->aliases);
  __pyx_v___pyx_result->aliases = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->default_value);
  __Pyx_DECREF(__pyx_v___pyx_result->default_value);
  __pyx_v___pyx_result->default_value = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->doc);
  __Pyx_DECREF(__pyx_v___pyx_result->doc);
  __pyx_v___pyx_result->doc = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->order);
  __Pyx_DECREF(__pyx_v___pyx_result->order);
  __pyx_v___pyx_result->order = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->type);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->type);
  __pyx_v___pyx_result->type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->writer_name);
  __Pyx_DECREF(__pyx_v___pyx_result->writer_name);
  __pyx_v___pyx_result->writer_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_RecordField__set_state(RecordField __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.order = __pyx_state[4]; __pyx_result.type = __pyx_state[5]; __pyx_result.writer_name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 7);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.order = __pyx_state[4]; __pyx_result.type = __pyx_state[5]; __pyx_result.writer_name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_RecordField__set_state(RecordField __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.order = __pyx_state[4]; __pyx_result.type = __pyx_state[5]; __pyx_result.writer_name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_RecordField__set_state(<RecordField> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RecordField__set_state(RecordField __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.order = __pyx_state[4]; __pyx_result.type = __pyx_state[5]; __pyx_result.writer_name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_RecordField__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__FieldAccessor(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_59__pyx_unpickle__FieldAccessor(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_59__pyx_unpickle__FieldAccessor = {"__pyx_unpickle__FieldAccessor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_59__pyx_unpickle__FieldAccessor, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_59__pyx_unpickle__FieldAccessor(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__FieldAccessor (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__FieldAccessor", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__FieldAccessor", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__FieldAccessor") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__FieldAccessor", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__FieldAccessor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_58__pyx_unpickle__FieldAccessor(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_58__pyx_unpickle__FieldAccessor(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__FieldAccessor", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1bc04b5, 0xe540cdd, 0x6a992d5):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bc04b5, 0xe540cdd, 0x6a992d5) = (index))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__108, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1bc04b5, 0xe540cdd, 0x6a992d5):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bc04b5, 0xe540cdd, 0x6a992d5) = (index))" % __pyx_checksum
 *     __pyx_result = _FieldAccessor.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x1bc04b5, 0xe540cdd, 0x6a992d5):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bc04b5, 0xe540cdd, 0x6a992d5) = (index))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _FieldAccessor.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1bc04b5, 0xe540cdd, 0x6a992d5):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bc04b5, 0xe540cdd, 0x6a992d5) = (index))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bc04b5, 0xe540cdd, 0x6a992d5) = (index))" % __pyx_checksum
 *     __pyx_result = _FieldAccessor.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__FieldAccessor__set_state(<_FieldAccessor> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__FieldAccessor), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bc04b5, 0xe540cdd, 0x6a992d5) = (index))" % __pyx_checksum
 *     __pyx_result = _FieldAccessor.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__FieldAccessor__set_state(<_FieldAccessor> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _FieldAccessor.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__FieldAccessor__set_state(<_FieldAccessor> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__FieldAccessor__set_state(_FieldAccessor __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__FieldAccessor__set_state(((struct __pyx_obj_5cavro__FieldAccessor *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bc04b5, 0xe540cdd, 0x6a992d5) = (index))" % __pyx_checksum
 *     __pyx_result = _FieldAccessor.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__FieldAccessor__set_state(<_FieldAccessor> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__FieldAccessor__set_state(<_FieldAccessor> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__FieldAccessor__set_state(_FieldAccessor __pyx_result, tuple __pyx_state):
 *     __pyx_result.index = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__FieldAccessor(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__FieldAccessor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__FieldAccessor__set_state(<_FieldAccessor> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__FieldAccessor__set_state(_FieldAccessor __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.index = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__FieldAccessor__set_state(struct __pyx_obj_5cavro__FieldAccessor *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__FieldAccessor__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__FieldAccessor__set_state(_FieldAccessor __pyx_result, tuple __pyx_state):
 *     __pyx_result.index = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->index = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle__FieldAccessor__set_state(_FieldAccessor __pyx_result, tuple __pyx_state):
 *     __pyx_result.index = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_2 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.index = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__FieldAccessor__set_state(_FieldAccessor __pyx_result, tuple __pyx_state):
 *     __pyx_result.index = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__FieldAccessor__set_state(<_FieldAccessor> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__FieldAccessor__set_state(_FieldAccessor __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.index = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__FieldAccessor__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_RecordType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_61__pyx_unpickle_RecordType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_61__pyx_unpickle_RecordType = {"__pyx_unpickle_RecordType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_61__pyx_unpickle_RecordType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_61__pyx_unpickle_RecordType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_RecordType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RecordType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RecordType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_RecordType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_RecordType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_RecordType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_60__pyx_unpickle_RecordType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_60__pyx_unpickle_RecordType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_RecordType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x182acf2, 0x17ad56f, 0xde7d9e1):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x182acf2, 0x17ad56f, 0xde7d9e1) = (_setting_up, aliases, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__109, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x182acf2, 0x17ad56f, 0xde7d9e1):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x182acf2, 0x17ad56f, 0xde7d9e1) = (_setting_up, aliases, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum
 *     __pyx_result = RecordType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x182acf2, 0x17ad56f, 0xde7d9e1):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x182acf2, 0x17ad56f, 0xde7d9e1) = (_setting_up, aliases, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = RecordType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x182acf2, 0x17ad56f, 0xde7d9e1):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x182acf2, 0x17ad56f, 0xde7d9e1) = (_setting_up, aliases, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x182acf2, 0x17ad56f, 0xde7d9e1) = (_setting_up, aliases, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum
 *     __pyx_result = RecordType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RecordType__set_state(<RecordType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_RecordType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x182acf2, 0x17ad56f, 0xde7d9e1) = (_setting_up, aliases, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum
 *     __pyx_result = RecordType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_RecordType__set_state(<RecordType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = RecordType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RecordType__set_state(<RecordType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_RecordType__set_state(RecordType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_RecordType__set_state(((struct __pyx_obj_5cavro_RecordType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x182acf2, 0x17ad56f, 0xde7d9e1) = (_setting_up, aliases, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum
 *     __pyx_result = RecordType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_RecordType__set_state(<RecordType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_RecordType__set_state(<RecordType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_RecordType__set_state(RecordType __pyx_result, tuple __pyx_state):
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.field_dict = __pyx_state[4]; __pyx_result.fields = __pyx_state[5]; __pyx_result.metadata = __pyx_state[6]; __pyx_result.name = __pyx_state[7]; __pyx_result.namespace = __pyx_state[8]; __pyx_result.options = __pyx_state[9]; __pyx_result.record = __pyx_state[10]; __pyx_result.value_adapters = __pyx_state[11]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_RecordType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_RecordType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_RecordType__set_state(<RecordType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RecordType__set_state(RecordType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.field_dict = __pyx_state[4]; __pyx_result.fields = __pyx_state[5]; __pyx_result.metadata = __pyx_state[6]; __pyx_result.name = __pyx_state[7]; __pyx_result.namespace = __pyx_state[8]; __pyx_result.options = __pyx_state[9]; __pyx_result.record = __pyx_state[10]; __pyx_result.value_adapters = __pyx_state[11]
 *     if len(__pyx_state) > 12 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_RecordType__set_state(struct __pyx_obj_5cavro_RecordType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_RecordType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_RecordType__set_state(RecordType __pyx_result, tuple __pyx_state):
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.field_dict = __pyx_state[4]; __pyx_result.fields = __pyx_state[5]; __pyx_result.metadata = __pyx_state[6]; __pyx_result.name = __pyx_state[7]; __pyx_result.namespace = __pyx_state[8]; __pyx_result.options = __pyx_state[9]; __pyx_result.record = __pyx_state[10]; __pyx_result.value_adapters = __pyx_state[11]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 12 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[12])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_setting_up = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyFrozenSet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("frozenset", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.aliases);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.aliases);
  __pyx_v___pyx_result->__pyx_base.aliases = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->doc);
  __Pyx_DECREF(__pyx_v___pyx_result->doc);
  __pyx_v___pyx_result->doc = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.effective_namespace);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.effective_namespace);
  __pyx_v___pyx_result->__pyx_base.effective_namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->field_dict);
  __Pyx_DECREF(__pyx_v___pyx_result->field_dict);
  __pyx_v___pyx_result->field_dict = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->fields);
  __Pyx_DECREF(__pyx_v___pyx_result->fields);
  __pyx_v___pyx_result->fields = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.name);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.name);
  __pyx_v___pyx_result->__pyx_base.name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.namespace);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.namespace);
  __pyx_v___pyx_result->__pyx_base.namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyType_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("type", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->record);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->record);
  __pyx_v___pyx_result->record = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_RecordType__set_state(RecordType __pyx_result, tuple __pyx_state):
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.field_dict = __pyx_state[4]; __pyx_result.fields = __pyx_state[5]; __pyx_result.metadata = __pyx_state[6]; __pyx_result.name = __pyx_state[7]; __pyx_result.namespace = __pyx_state[8]; __pyx_result.options = __pyx_state[9]; __pyx_result.record = __pyx_state[10]; __pyx_result.value_adapters = __pyx_state[11]
 *     if len(__pyx_state) > 12 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[12])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 12);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.field_dict = __pyx_state[4]; __pyx_result.fields = __pyx_state[5]; __pyx_result.metadata = __pyx_state[6]; __pyx_result.name = __pyx_state[7]; __pyx_result.namespace = __pyx_state[8]; __pyx_result.options = __pyx_state[9]; __pyx_result.record = __pyx_state[10]; __pyx_result.value_adapters = __pyx_state[11]
 *     if len(__pyx_state) > 12 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[12])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_RecordType__set_state(RecordType __pyx_result, tuple __pyx_state):
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.field_dict = __pyx_state[4]; __pyx_result.fields = __pyx_state[5]; __pyx_result.metadata = __pyx_state[6]; __pyx_result.name = __pyx_state[7]; __pyx_result.namespace = __pyx_state[8]; __pyx_result.options = __pyx_state[9]; __pyx_result.record = __pyx_state[10]; __pyx_result.value_adapters = __pyx_state[11]
 *     if len(__pyx_state) > 12 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[12])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_RecordType__set_state(<RecordType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_RecordType__set_state(RecordType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.field_dict = __pyx_state[4]; __pyx_result.fields = __pyx_state[5]; __pyx_result.metadata = __pyx_state[6]; __pyx_result.name = __pyx_state[7]; __pyx_result.namespace = __pyx_state[8]; __pyx_result.options = __pyx_state[9]; __pyx_result.record = __pyx_state[10]; __pyx_result.value_adapters = __pyx_state[11]
 *     if len(__pyx_state) > 12 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_RecordType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PromotingRecordType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_63__pyx_unpickle_PromotingRecordType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_63__pyx_unpickle_PromotingRecordType = {"__pyx_unpickle_PromotingRecordType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_63__pyx_unpickle_PromotingRecordType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_63__pyx_unpickle_PromotingRecordType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromotingRecordType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromotingRecordType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromotingRecordType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PromotingRecordType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromotingRecordType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromotingRecordType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_62__pyx_unpickle_PromotingRecordType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_62__pyx_unpickle_PromotingRecordType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromotingRecordType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x40bba2b, 0x72480de, 0x6ac5437):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x40bba2b, 0x72480de, 0x6ac5437) = (_setting_up, aliases, decode_indexes, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__110, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x40bba2b, 0x72480de, 0x6ac5437):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x40bba2b, 0x72480de, 0x6ac5437) = (_setting_up, aliases, decode_indexes, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum
 *     __pyx_result = PromotingRecordType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x40bba2b, 0x72480de, 0x6ac5437):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x40bba2b, 0x72480de, 0x6ac5437) = (_setting_up, aliases, decode_indexes, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PromotingRecordType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x40bba2b, 0x72480de, 0x6ac5437):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x40bba2b, 0x72480de, 0x6ac5437) = (_setting_up, aliases, decode_indexes, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x40bba2b, 0x72480de, 0x6ac5437) = (_setting_up, aliases, decode_indexes, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum
 *     __pyx_result = PromotingRecordType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromotingRecordType__set_state(<PromotingRecordType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_PromotingRecordType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x40bba2b, 0x72480de, 0x6ac5437) = (_setting_up, aliases, decode_indexes, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum
 *     __pyx_result = PromotingRecordType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PromotingRecordType__set_state(<PromotingRecordType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PromotingRecordType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromotingRecordType__set_state(<PromotingRecordType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PromotingRecordType__set_state(PromotingRecordType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_PromotingRecordType__set_state(((struct __pyx_obj_5cavro_PromotingRecordType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x40bba2b, 0x72480de, 0x6ac5437) = (_setting_up, aliases, decode_indexes, doc, effective_namespace, field_dict, fields, metadata, name, namespace, options, record, value_adapters))" % __pyx_checksum
 *     __pyx_result = PromotingRecordType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PromotingRecordType__set_state(<PromotingRecordType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromotingRecordType__set_state(<PromotingRecordType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PromotingRecordType__set_state(PromotingRecordType __pyx_result, tuple __pyx_state):
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.decode_indexes = __pyx_state[2]; __pyx_result.doc = __pyx_state[3]; __pyx_result.effective_namespace = __pyx_state[4]; __pyx_result.field_dict = __pyx_state[5]; __pyx_result.fields = __pyx_state[6]; __pyx_result.metadata = __pyx_state[7]; __pyx_result.name = __pyx_state[8]; __pyx_result.namespace = __pyx_state[9]; __pyx_result.options = __pyx_state[10]; __pyx_result.record = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PromotingRecordType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromotingRecordType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PromotingRecordType__set_state(<PromotingRecordType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromotingRecordType__set_state(PromotingRecordType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.decode_indexes = __pyx_state[2]; __pyx_result.doc = __pyx_state[3]; __pyx_result.effective_namespace = __pyx_state[4]; __pyx_result.field_dict = __pyx_state[5]; __pyx_result.fields = __pyx_state[6]; __pyx_result.metadata = __pyx_state[7]; __pyx_result.name = __pyx_state[8]; __pyx_result.namespace = __pyx_state[9]; __pyx_result.options = __pyx_state[10]; __pyx_result.record = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_PromotingRecordType__set_state(struct __pyx_obj_5cavro_PromotingRecordType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromotingRecordType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PromotingRecordType__set_state(PromotingRecordType __pyx_result, tuple __pyx_state):
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.decode_indexes = __pyx_state[2]; __pyx_result.doc = __pyx_state[3]; __pyx_result.effective_namespace = __pyx_state[4]; __pyx_result.field_dict = __pyx_state[5]; __pyx_result.fields = __pyx_state[6]; __pyx_result.metadata = __pyx_state[7]; __pyx_result.name = __pyx_state[8]; __pyx_result.namespace = __pyx_state[9]; __pyx_result.options = __pyx_state[10]; __pyx_result.record = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base._setting_up = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyFrozenSet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("frozenset", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.aliases);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.aliases);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.aliases = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_Py_ssize_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XCLEAR_MEMVIEW(&__pyx_v___pyx_result->decode_indexes, 0);
  __pyx_v___pyx_result->decode_indexes = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.doc);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.doc);
  __pyx_v___pyx_result->__pyx_base.doc = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.effective_namespace);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.effective_namespace);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.effective_namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.field_dict);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.field_dict);
  __pyx_v___pyx_result->__pyx_base.field_dict = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.fields);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.fields);
  __pyx_v___pyx_result->__pyx_base.fields = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.name);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.name);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.namespace);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.namespace);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyType_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("type", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.record);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.record);
  __pyx_v___pyx_result->__pyx_base.record = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PromotingRecordType__set_state(PromotingRecordType __pyx_result, tuple __pyx_state):
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.decode_indexes = __pyx_state[2]; __pyx_result.doc = __pyx_state[3]; __pyx_result.effective_namespace = __pyx_state[4]; __pyx_result.field_dict = __pyx_state[5]; __pyx_result.fields = __pyx_state[6]; __pyx_result.metadata = __pyx_state[7]; __pyx_result.name = __pyx_state[8]; __pyx_result.namespace = __pyx_state[9]; __pyx_result.options = __pyx_state[10]; __pyx_result.record = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 13);
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.decode_indexes = __pyx_state[2]; __pyx_result.doc = __pyx_state[3]; __pyx_result.effective_namespace = __pyx_state[4]; __pyx_result.field_dict = __pyx_state[5]; __pyx_result.fields = __pyx_state[6]; __pyx_result.metadata = __pyx_state[7]; __pyx_result.name = __pyx_state[8]; __pyx_result.namespace = __pyx_state[9]; __pyx_result.options = __pyx_state[10]; __pyx_result.record = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PromotingRecordType__set_state(PromotingRecordType __pyx_result, tuple __pyx_state):
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.decode_indexes = __pyx_state[2]; __pyx_result.doc = __pyx_state[3]; __pyx_result.effective_namespace = __pyx_state[4]; __pyx_result.field_dict = __pyx_state[5]; __pyx_result.fields = __pyx_state[6]; __pyx_result.metadata = __pyx_state[7]; __pyx_result.name = __pyx_state[8]; __pyx_result.namespace = __pyx_state[9]; __pyx_result.options = __pyx_state[10]; __pyx_result.record = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PromotingRecordType__set_state(<PromotingRecordType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromotingRecordType__set_state(PromotingRecordType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.decode_indexes = __pyx_state[2]; __pyx_result.doc = __pyx_state[3]; __pyx_result.effective_namespace = __pyx_state[4]; __pyx_result.field_dict = __pyx_state[5]; __pyx_result.fields = __pyx_state[6]; __pyx_result.metadata = __pyx_state[7]; __pyx_result.name = __pyx_state[8]; __pyx_result.namespace = __pyx_state[9]; __pyx_result.options = __pyx_state[10]; __pyx_result.record = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromotingRecordType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_BytesType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_65__pyx_unpickle_BytesType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_65__pyx_unpickle_BytesType = {"__pyx_unpickle_BytesType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_65__pyx_unpickle_BytesType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_65__pyx_unpickle_BytesType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_BytesType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BytesType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BytesType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_BytesType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BytesType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_BytesType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_64__pyx_unpickle_BytesType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_64__pyx_unpickle_BytesType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BytesType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__104, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = BytesType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = BytesType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = BytesType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BytesType__set_state(<BytesType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_BytesType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = BytesType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BytesType__set_state(<BytesType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = BytesType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BytesType__set_state(<BytesType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_BytesType__set_state(BytesType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_BytesType__set_state(((struct __pyx_obj_5cavro_BytesType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = BytesType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BytesType__set_state(<BytesType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BytesType__set_state(<BytesType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_BytesType__set_state(BytesType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_BytesType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_BytesType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_BytesType__set_state(<BytesType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BytesType__set_state(BytesType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_BytesType__set_state(struct __pyx_obj_5cavro_BytesType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BytesType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_BytesType__set_state(BytesType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_BytesType__set_state(BytesType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_BytesType__set_state(BytesType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_BytesType__set_state(<BytesType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BytesType__set_state(BytesType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_BytesType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_StringType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_67__pyx_unpickle_StringType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_67__pyx_unpickle_StringType = {"__pyx_unpickle_StringType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_67__pyx_unpickle_StringType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_67__pyx_unpickle_StringType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_StringType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_StringType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_StringType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_StringType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_StringType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_StringType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_66__pyx_unpickle_StringType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_66__pyx_unpickle_StringType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_StringType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__104, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = StringType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = StringType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = StringType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_StringType__set_state(<StringType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_StringType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = StringType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_StringType__set_state(<StringType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = StringType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_StringType__set_state(<StringType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_StringType__set_state(StringType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_StringType__set_state(((struct __pyx_obj_5cavro_StringType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = StringType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_StringType__set_state(<StringType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_StringType__set_state(<StringType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_StringType__set_state(StringType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_StringType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_StringType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_StringType__set_state(<StringType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_StringType__set_state(StringType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_StringType__set_state(struct __pyx_obj_5cavro_StringType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_StringType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_StringType__set_state(StringType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_StringType__set_state(StringType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_StringType__set_state(StringType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_StringType__set_state(<StringType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_StringType__set_state(StringType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_StringType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_FixedType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_69__pyx_unpickle_FixedType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_69__pyx_unpickle_FixedType = {"__pyx_unpickle_FixedType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_69__pyx_unpickle_FixedType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_69__pyx_unpickle_FixedType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_FixedType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FixedType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FixedType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_FixedType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FixedType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_FixedType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_68__pyx_unpickle_FixedType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_68__pyx_unpickle_FixedType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FixedType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd9f3afe, 0xacb1013, 0x605aa9f):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd9f3afe, 0xacb1013, 0x605aa9f) = (aliases, effective_namespace, metadata, name, namespace, options, size, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__111, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd9f3afe, 0xacb1013, 0x605aa9f):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd9f3afe, 0xacb1013, 0x605aa9f) = (aliases, effective_namespace, metadata, name, namespace, options, size, value_adapters))" % __pyx_checksum
 *     __pyx_result = FixedType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xd9f3afe, 0xacb1013, 0x605aa9f):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd9f3afe, 0xacb1013, 0x605aa9f) = (aliases, effective_namespace, metadata, name, namespace, options, size, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = FixedType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd9f3afe, 0xacb1013, 0x605aa9f):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd9f3afe, 0xacb1013, 0x605aa9f) = (aliases, effective_namespace, metadata, name, namespace, options, size, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd9f3afe, 0xacb1013, 0x605aa9f) = (aliases, effective_namespace, metadata, name, namespace, options, size, value_adapters))" % __pyx_checksum
 *     __pyx_result = FixedType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FixedType__set_state(<FixedType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_FixedType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd9f3afe, 0xacb1013, 0x605aa9f) = (aliases, effective_namespace, metadata, name, namespace, options, size, value_adapters))" % __pyx_checksum
 *     __pyx_result = FixedType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FixedType__set_state(<FixedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = FixedType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FixedType__set_state(<FixedType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_FixedType__set_state(FixedType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_FixedType__set_state(((struct __pyx_obj_5cavro_FixedType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xd9f3afe, 0xacb1013, 0x605aa9f) = (aliases, effective_namespace, metadata, name, namespace, options, size, value_adapters))" % __pyx_checksum
 *     __pyx_result = FixedType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FixedType__set_state(<FixedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FixedType__set_state(<FixedType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_FixedType__set_state(FixedType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.size = __pyx_state[6]; __pyx_result.value_adapters = __pyx_state[7]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FixedType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_FixedType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_FixedType__set_state(<FixedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FixedType__set_state(FixedType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.size = __pyx_state[6]; __pyx_result.value_adapters = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_FixedType__set_state(struct __pyx_obj_5cavro_FixedType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FixedType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_FixedType__set_state(FixedType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.size = __pyx_state[6]; __pyx_result.value_adapters = __pyx_state[7]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyFrozenSet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("frozenset", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.aliases);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.aliases);
  __pyx_v___pyx_result->__pyx_base.aliases = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.effective_namespace);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.effective_namespace);
  __pyx_v___pyx_result->__pyx_base.effective_namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.name);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.name);
  __pyx_v___pyx_result->__pyx_base.name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.namespace);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.namespace);
  __pyx_v___pyx_result->__pyx_base.namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->size = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_FixedType__set_state(FixedType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.size = __pyx_state[6]; __pyx_result.value_adapters = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_2 > 8);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.size = __pyx_state[6]; __pyx_result.value_adapters = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_FixedType__set_state(FixedType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.size = __pyx_state[6]; __pyx_result.value_adapters = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_FixedType__set_state(<FixedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FixedType__set_state(FixedType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.size = __pyx_state[6]; __pyx_result.value_adapters = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_FixedType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_UnionType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_71__pyx_unpickle_UnionType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_71__pyx_unpickle_UnionType = {"__pyx_unpickle_UnionType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_71__pyx_unpickle_UnionType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_71__pyx_unpickle_UnionType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_UnionType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UnionType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UnionType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_UnionType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UnionType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_UnionType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_70__pyx_unpickle_UnionType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_70__pyx_unpickle_UnionType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UnionType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xabef18e, 0x6e50000, 0x97a7fa4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xabef18e, 0x6e50000, 0x97a7fa4) = (by_name_map, metadata, options, return_type_tuple, union_types, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__112, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xabef18e, 0x6e50000, 0x97a7fa4):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xabef18e, 0x6e50000, 0x97a7fa4) = (by_name_map, metadata, options, return_type_tuple, union_types, value_adapters))" % __pyx_checksum
 *     __pyx_result = UnionType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xabef18e, 0x6e50000, 0x97a7fa4):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xabef18e, 0x6e50000, 0x97a7fa4) = (by_name_map, metadata, options, return_type_tuple, union_types, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = UnionType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xabef18e, 0x6e50000, 0x97a7fa4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xabef18e, 0x6e50000, 0x97a7fa4) = (by_name_map, metadata, options, return_type_tuple, union_types, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xabef18e, 0x6e50000, 0x97a7fa4) = (by_name_map, metadata, options, return_type_tuple, union_types, value_adapters))" % __pyx_checksum
 *     __pyx_result = UnionType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UnionType__set_state(<UnionType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_UnionType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xabef18e, 0x6e50000, 0x97a7fa4) = (by_name_map, metadata, options, return_type_tuple, union_types, value_adapters))" % __pyx_checksum
 *     __pyx_result = UnionType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UnionType__set_state(<UnionType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = UnionType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UnionType__set_state(<UnionType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_UnionType__set_state(UnionType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_UnionType__set_state(((struct __pyx_obj_5cavro_UnionType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xabef18e, 0x6e50000, 0x97a7fa4) = (by_name_map, metadata, options, return_type_tuple, union_types, value_adapters))" % __pyx_checksum
 *     __pyx_result = UnionType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UnionType__set_state(<UnionType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UnionType__set_state(<UnionType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_UnionType__set_state(UnionType __pyx_result, tuple __pyx_state):
 *     __pyx_result.by_name_map = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.return_type_tuple = __pyx_state[3]; __pyx_result.union_types = __pyx_state[4]; __pyx_result.value_adapters = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_UnionType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_UnionType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_UnionType__set_state(<UnionType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UnionType__set_state(UnionType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.by_name_map = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.return_type_tuple = __pyx_state[3]; __pyx_result.union_types = __pyx_state[4]; __pyx_result.value_adapters = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_UnionType__set_state(struct __pyx_obj_5cavro_UnionType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UnionType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_UnionType__set_state(UnionType __pyx_result, tuple __pyx_state):
 *     __pyx_result.by_name_map = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.return_type_tuple = __pyx_state[3]; __pyx_result.union_types = __pyx_state[4]; __pyx_result.value_adapters = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->by_name_map);
  __Pyx_DECREF(__pyx_v___pyx_result->by_name_map);
  __pyx_v___pyx_result->by_name_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->return_type_tuple);
  __Pyx_DECREF(__pyx_v___pyx_result->return_type_tuple);
  __pyx_v___pyx_result->return_type_tuple = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->union_types);
  __Pyx_DECREF(__pyx_v___pyx_result->union_types);
  __pyx_v___pyx_result->union_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_UnionType__set_state(UnionType __pyx_result, tuple __pyx_state):
 *     __pyx_result.by_name_map = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.return_type_tuple = __pyx_state[3]; __pyx_result.union_types = __pyx_state[4]; __pyx_result.value_adapters = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 6);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.by_name_map = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.return_type_tuple = __pyx_state[3]; __pyx_result.union_types = __pyx_state[4]; __pyx_result.value_adapters = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_UnionType__set_state(UnionType __pyx_result, tuple __pyx_state):
 *     __pyx_result.by_name_map = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.return_type_tuple = __pyx_state[3]; __pyx_result.union_types = __pyx_state[4]; __pyx_result.value_adapters = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_UnionType__set_state(<UnionType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UnionType__set_state(UnionType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.by_name_map = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.return_type_tuple = __pyx_state[3]; __pyx_result.union_types = __pyx_state[4]; __pyx_result.value_adapters = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_UnionType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__Sentinel(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_73__pyx_unpickle__Sentinel(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_73__pyx_unpickle__Sentinel = {"__pyx_unpickle__Sentinel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_73__pyx_unpickle__Sentinel, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_73__pyx_unpickle__Sentinel(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__Sentinel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__Sentinel", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__Sentinel", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__Sentinel") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__Sentinel", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__Sentinel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_72__pyx_unpickle__Sentinel(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_72__pyx_unpickle__Sentinel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__Sentinel", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__8, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = _Sentinel.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _Sentinel.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = _Sentinel.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__Sentinel__set_state(<_Sentinel> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__Sentinel), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = _Sentinel.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__Sentinel__set_state(<_Sentinel> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _Sentinel.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__Sentinel__set_state(<_Sentinel> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__Sentinel__set_state(_Sentinel __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__Sentinel__set_state(((struct __pyx_obj_5cavro__Sentinel *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = _Sentinel.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__Sentinel__set_state(<_Sentinel> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__Sentinel__set_state(<_Sentinel> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__Sentinel__set_state(_Sentinel __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__Sentinel(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__Sentinel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__Sentinel__set_state(<_Sentinel> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__Sentinel__set_state(_Sentinel __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__Sentinel__set_state(struct __pyx_obj_5cavro__Sentinel *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__Sentinel__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__Sentinel__set_state(_Sentinel __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle__Sentinel__set_state(_Sentinel __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__Sentinel__set_state(_Sentinel __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__Sentinel__set_state(<_Sentinel> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__Sentinel__set_state(_Sentinel __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__Sentinel__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__CanonicalForm(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_75__pyx_unpickle__CanonicalForm(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_75__pyx_unpickle__CanonicalForm = {"__pyx_unpickle__CanonicalForm", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_75__pyx_unpickle__CanonicalForm, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_75__pyx_unpickle__CanonicalForm(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__CanonicalForm (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__CanonicalForm", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__CanonicalForm", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__CanonicalForm") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__CanonicalForm", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__CanonicalForm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_74__pyx_unpickle__CanonicalForm(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_74__pyx_unpickle__CanonicalForm(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__CanonicalForm", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _CanonicalForm.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _CanonicalForm.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _CanonicalForm.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__CanonicalForm__set_state(<_CanonicalForm> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__CanonicalForm), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _CanonicalForm.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__CanonicalForm__set_state(<_CanonicalForm> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _CanonicalForm.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__CanonicalForm__set_state(<_CanonicalForm> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__CanonicalForm__set_state(_CanonicalForm __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__CanonicalForm__set_state(((struct __pyx_obj_5cavro__CanonicalForm *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _CanonicalForm.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__CanonicalForm__set_state(<_CanonicalForm> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__CanonicalForm__set_state(<_CanonicalForm> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__CanonicalForm__set_state(_CanonicalForm __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__CanonicalForm(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__CanonicalForm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__CanonicalForm__set_state(<_CanonicalForm> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__CanonicalForm__set_state(_CanonicalForm __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__CanonicalForm__set_state(struct __pyx_obj_5cavro__CanonicalForm *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__CanonicalForm__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__CanonicalForm__set_state(_CanonicalForm __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__CanonicalForm__set_state(_CanonicalForm __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__CanonicalForm__set_state(_CanonicalForm __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__CanonicalForm__set_state(<_CanonicalForm> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__CanonicalForm__set_state(_CanonicalForm __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__CanonicalForm__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_AvroType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_77__pyx_unpickle_AvroType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_77__pyx_unpickle_AvroType = {"__pyx_unpickle_AvroType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_77__pyx_unpickle_AvroType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_77__pyx_unpickle_AvroType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_AvroType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_AvroType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_AvroType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_AvroType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_AvroType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_AvroType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_76__pyx_unpickle_AvroType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_76__pyx_unpickle_AvroType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_AvroType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__104, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = AvroType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = AvroType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x84c5042, 0xa81a03f, 0x4bb6450):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = AvroType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_AvroType__set_state(<AvroType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_AvroType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = AvroType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_AvroType__set_state(<AvroType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = AvroType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_AvroType__set_state(<AvroType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_AvroType__set_state(AvroType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_AvroType__set_state(((struct __pyx_obj_5cavro_AvroType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x84c5042, 0xa81a03f, 0x4bb6450) = (metadata, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = AvroType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_AvroType__set_state(<AvroType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_AvroType__set_state(<AvroType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_AvroType__set_state(AvroType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_AvroType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_AvroType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_AvroType__set_state(<AvroType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_AvroType__set_state(AvroType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_AvroType__set_state(struct __pyx_obj_5cavro_AvroType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_AvroType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_AvroType__set_state(AvroType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->metadata);
  __pyx_v___pyx_result->metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->options);
  __pyx_v___pyx_result->options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->value_adapters);
  __pyx_v___pyx_result->value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_AvroType__set_state(AvroType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_AvroType__set_state(AvroType __pyx_result, tuple __pyx_state):
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_AvroType__set_state(<AvroType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_AvroType__set_state(AvroType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_AvroType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__NamedType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_79__pyx_unpickle__NamedType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_79__pyx_unpickle__NamedType = {"__pyx_unpickle__NamedType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_79__pyx_unpickle__NamedType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_79__pyx_unpickle__NamedType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__NamedType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__NamedType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__NamedType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__NamedType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__NamedType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__NamedType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_78__pyx_unpickle__NamedType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_78__pyx_unpickle__NamedType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__NamedType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x24d7e16, 0xc138f68, 0xdaff9f5):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x24d7e16, 0xc138f68, 0xdaff9f5) = (aliases, effective_namespace, metadata, name, namespace, options, value_adapters))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__113, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x24d7e16, 0xc138f68, 0xdaff9f5):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x24d7e16, 0xc138f68, 0xdaff9f5) = (aliases, effective_namespace, metadata, name, namespace, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = _NamedType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x24d7e16, 0xc138f68, 0xdaff9f5):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x24d7e16, 0xc138f68, 0xdaff9f5) = (aliases, effective_namespace, metadata, name, namespace, options, value_adapters))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _NamedType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x24d7e16, 0xc138f68, 0xdaff9f5):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x24d7e16, 0xc138f68, 0xdaff9f5) = (aliases, effective_namespace, metadata, name, namespace, options, value_adapters))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x24d7e16, 0xc138f68, 0xdaff9f5) = (aliases, effective_namespace, metadata, name, namespace, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = _NamedType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__NamedType__set_state(<_NamedType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__NamedType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x24d7e16, 0xc138f68, 0xdaff9f5) = (aliases, effective_namespace, metadata, name, namespace, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = _NamedType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__NamedType__set_state(<_NamedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _NamedType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__NamedType__set_state(<_NamedType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__NamedType__set_state(_NamedType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__NamedType__set_state(((struct __pyx_obj_5cavro__NamedType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x24d7e16, 0xc138f68, 0xdaff9f5) = (aliases, effective_namespace, metadata, name, namespace, options, value_adapters))" % __pyx_checksum
 *     __pyx_result = _NamedType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__NamedType__set_state(<_NamedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__NamedType__set_state(<_NamedType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__NamedType__set_state(_NamedType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.value_adapters = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__NamedType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__NamedType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__NamedType__set_state(<_NamedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__NamedType__set_state(_NamedType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.value_adapters = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__NamedType__set_state(struct __pyx_obj_5cavro__NamedType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__NamedType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__NamedType__set_state(_NamedType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.value_adapters = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyFrozenSet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("frozenset", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->aliases);
  __Pyx_DECREF(__pyx_v___pyx_result->aliases);
  __pyx_v___pyx_result->aliases = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->effective_namespace);
  __Pyx_DECREF(__pyx_v___pyx_result->effective_namespace);
  __pyx_v___pyx_result->effective_namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.metadata);
  __pyx_v___pyx_result->__pyx_base.metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->namespace);
  __Pyx_DECREF(__pyx_v___pyx_result->namespace);
  __pyx_v___pyx_result->namespace = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.value_adapters);
  __pyx_v___pyx_result->__pyx_base.value_adapters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle__NamedType__set_state(_NamedType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.value_adapters = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 7);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.value_adapters = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__NamedType__set_state(_NamedType __pyx_result, tuple __pyx_state):
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.value_adapters = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__NamedType__set_state(<_NamedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__NamedType__set_state(_NamedType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.value_adapters = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__NamedType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Schema(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_81__pyx_unpickle_Schema(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_81__pyx_unpickle_Schema = {"__pyx_unpickle_Schema", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_81__pyx_unpickle_Schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_81__pyx_unpickle_Schema(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Schema (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Schema", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Schema", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Schema") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Schema", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_80__pyx_unpickle_Schema(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_80__pyx_unpickle_Schema(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Schema", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc2592e1, 0xe13d498, 0xa7be993):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__114, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc2592e1, 0xe13d498, 0xa7be993):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum
 *     __pyx_result = Schema.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc2592e1, 0xe13d498, 0xa7be993):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Schema.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc2592e1, 0xe13d498, 0xa7be993):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum
 *     __pyx_result = Schema.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Schema__set_state(<Schema> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_Schema), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum
 *     __pyx_result = Schema.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Schema__set_state(<Schema> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Schema.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Schema__set_state(<Schema> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Schema__set_state(Schema __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_Schema__set_state(((struct __pyx_obj_5cavro_Schema *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum
 *     __pyx_result = Schema.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Schema__set_state(<Schema> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Schema__set_state(<Schema> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Schema__set_state(Schema __pyx_result, tuple __pyx_state):
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Schema(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Schema__set_state(<Schema> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Schema__set_state(Schema __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_Schema__set_state(struct __pyx_obj_5cavro_Schema *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Schema__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Schema__set_state(Schema __pyx_result, tuple __pyx_state):
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_canonical_form);
  __Pyx_DECREF(__pyx_v___pyx_result->_canonical_form);
  __pyx_v___pyx_result->_canonical_form = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->logical_types);
  __Pyx_DECREF(__pyx_v___pyx_result->logical_types);
  __pyx_v___pyx_result->logical_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->named_types);
  __Pyx_DECREF(__pyx_v___pyx_result->named_types);
  __pyx_v___pyx_result->named_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->options);
  __pyx_v___pyx_result->options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->source);
  __Pyx_DECREF(__pyx_v___pyx_result->source);
  __pyx_v___pyx_result->source = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->type);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->type);
  __pyx_v___pyx_result->type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Schema__set_state(Schema __pyx_result, tuple __pyx_state):
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 6);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Schema__set_state(Schema __pyx_result, tuple __pyx_state):
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Schema__set_state(<Schema> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Schema__set_state(Schema __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Schema__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ResolvedSchema(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_83__pyx_unpickle_ResolvedSchema(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_83__pyx_unpickle_ResolvedSchema = {"__pyx_unpickle_ResolvedSchema", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_83__pyx_unpickle_ResolvedSchema, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_83__pyx_unpickle_ResolvedSchema(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ResolvedSchema (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ResolvedSchema", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ResolvedSchema", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_ResolvedSchema") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ResolvedSchema", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_ResolvedSchema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_82__pyx_unpickle_ResolvedSchema(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_82__pyx_unpickle_ResolvedSchema(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ResolvedSchema", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc2592e1, 0xe13d498, 0xa7be993):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__114, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc2592e1, 0xe13d498, 0xa7be993):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum
 *     __pyx_result = ResolvedSchema.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc2592e1, 0xe13d498, 0xa7be993):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = ResolvedSchema.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc2592e1, 0xe13d498, 0xa7be993):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum
 *     __pyx_result = ResolvedSchema.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ResolvedSchema__set_state(<ResolvedSchema> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_ResolvedSchema), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum
 *     __pyx_result = ResolvedSchema.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ResolvedSchema__set_state(<ResolvedSchema> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = ResolvedSchema.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ResolvedSchema__set_state(<ResolvedSchema> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ResolvedSchema__set_state(ResolvedSchema __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_ResolvedSchema__set_state(((struct __pyx_obj_5cavro_ResolvedSchema *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc2592e1, 0xe13d498, 0xa7be993) = (_canonical_form, logical_types, named_types, options, source, type))" % __pyx_checksum
 *     __pyx_result = ResolvedSchema.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ResolvedSchema__set_state(<ResolvedSchema> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ResolvedSchema__set_state(<ResolvedSchema> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ResolvedSchema__set_state(ResolvedSchema __pyx_result, tuple __pyx_state):
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ResolvedSchema(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_ResolvedSchema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ResolvedSchema__set_state(<ResolvedSchema> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ResolvedSchema__set_state(ResolvedSchema __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_ResolvedSchema__set_state(struct __pyx_obj_5cavro_ResolvedSchema *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ResolvedSchema__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ResolvedSchema__set_state(ResolvedSchema __pyx_result, tuple __pyx_state):
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._canonical_form);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._canonical_form);
  __pyx_v___pyx_result->__pyx_base._canonical_form = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.logical_types);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.logical_types);
  __pyx_v___pyx_result->__pyx_base.logical_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.named_types);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.named_types);
  __pyx_v___pyx_result->__pyx_base.named_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Options))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.options);
  __pyx_v___pyx_result->__pyx_base.options = ((struct __pyx_obj_5cavro_Options *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.source);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.source);
  __pyx_v___pyx_result->__pyx_base.source = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base.type);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base.type);
  __pyx_v___pyx_result->__pyx_base.type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_ResolvedSchema__set_state(ResolvedSchema __pyx_result, tuple __pyx_state):
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 6);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ResolvedSchema__set_state(ResolvedSchema __pyx_result, tuple __pyx_state):
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ResolvedSchema__set_state(<ResolvedSchema> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ResolvedSchema__set_state(ResolvedSchema __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_ResolvedSchema__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ValueAdapter(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_85__pyx_unpickle_ValueAdapter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_85__pyx_unpickle_ValueAdapter = {"__pyx_unpickle_ValueAdapter", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_85__pyx_unpickle_ValueAdapter, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_85__pyx_unpickle_ValueAdapter(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ValueAdapter (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ValueAdapter", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ValueAdapter", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_ValueAdapter") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ValueAdapter", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_ValueAdapter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_84__pyx_unpickle_ValueAdapter(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_84__pyx_unpickle_ValueAdapter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ValueAdapter", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = ValueAdapter.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = ValueAdapter.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = ValueAdapter.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ValueAdapter__set_state(<ValueAdapter> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_ValueAdapter), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = ValueAdapter.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ValueAdapter__set_state(<ValueAdapter> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = ValueAdapter.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ValueAdapter__set_state(<ValueAdapter> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ValueAdapter__set_state(ValueAdapter __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_ValueAdapter__set_state(((struct __pyx_obj_5cavro_ValueAdapter *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = ValueAdapter.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ValueAdapter__set_state(<ValueAdapter> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ValueAdapter__set_state(<ValueAdapter> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ValueAdapter__set_state(ValueAdapter __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ValueAdapter(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_ValueAdapter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ValueAdapter__set_state(<ValueAdapter> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ValueAdapter__set_state(ValueAdapter __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_ValueAdapter__set_state(struct __pyx_obj_5cavro_ValueAdapter *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ValueAdapter__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ValueAdapter__set_state(ValueAdapter __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ValueAdapter__set_state(ValueAdapter __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ValueAdapter__set_state(ValueAdapter __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ValueAdapter__set_state(<ValueAdapter> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ValueAdapter__set_state(ValueAdapter __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_ValueAdapter__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_LogicalType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_87__pyx_unpickle_LogicalType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_87__pyx_unpickle_LogicalType = {"__pyx_unpickle_LogicalType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_87__pyx_unpickle_LogicalType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_87__pyx_unpickle_LogicalType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_LogicalType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LogicalType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LogicalType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_LogicalType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LogicalType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_LogicalType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_86__pyx_unpickle_LogicalType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_86__pyx_unpickle_LogicalType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LogicalType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = LogicalType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = LogicalType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = LogicalType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LogicalType__set_state(<LogicalType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_LogicalType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = LogicalType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LogicalType__set_state(<LogicalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = LogicalType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LogicalType__set_state(<LogicalType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_LogicalType__set_state(LogicalType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_LogicalType__set_state(((struct __pyx_obj_5cavro_LogicalType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = LogicalType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LogicalType__set_state(<LogicalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LogicalType__set_state(<LogicalType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_LogicalType__set_state(LogicalType __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LogicalType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_LogicalType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_LogicalType__set_state(<LogicalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LogicalType__set_state(LogicalType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_LogicalType__set_state(struct __pyx_obj_5cavro_LogicalType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LogicalType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_LogicalType__set_state(LogicalType __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_LogicalType__set_state(LogicalType __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_LogicalType__set_state(LogicalType __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_LogicalType__set_state(<LogicalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LogicalType__set_state(LogicalType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_LogicalType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_CustomLogicalType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_89__pyx_unpickle_CustomLogicalType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_89__pyx_unpickle_CustomLogicalType = {"__pyx_unpickle_CustomLogicalType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_89__pyx_unpickle_CustomLogicalType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_89__pyx_unpickle_CustomLogicalType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CustomLogicalType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CustomLogicalType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CustomLogicalType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_CustomLogicalType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CustomLogicalType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_CustomLogicalType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_88__pyx_unpickle_CustomLogicalType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_88__pyx_unpickle_CustomLogicalType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CustomLogicalType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = CustomLogicalType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = CustomLogicalType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = CustomLogicalType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CustomLogicalType__set_state(<CustomLogicalType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_CustomLogicalType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = CustomLogicalType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CustomLogicalType__set_state(<CustomLogicalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = CustomLogicalType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CustomLogicalType__set_state(<CustomLogicalType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CustomLogicalType__set_state(CustomLogicalType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_CustomLogicalType__set_state(((struct __pyx_obj_5cavro_CustomLogicalType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = CustomLogicalType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CustomLogicalType__set_state(<CustomLogicalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CustomLogicalType__set_state(<CustomLogicalType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CustomLogicalType__set_state(CustomLogicalType __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CustomLogicalType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_CustomLogicalType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CustomLogicalType__set_state(<CustomLogicalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CustomLogicalType__set_state(CustomLogicalType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_CustomLogicalType__set_state(struct __pyx_obj_5cavro_CustomLogicalType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CustomLogicalType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CustomLogicalType__set_state(CustomLogicalType __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_CustomLogicalType__set_state(CustomLogicalType __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CustomLogicalType__set_state(CustomLogicalType __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_CustomLogicalType__set_state(<CustomLogicalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CustomLogicalType__set_state(CustomLogicalType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_CustomLogicalType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_DecimalType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_91__pyx_unpickle_DecimalType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_91__pyx_unpickle_DecimalType = {"__pyx_unpickle_DecimalType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_91__pyx_unpickle_DecimalType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_91__pyx_unpickle_DecimalType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_DecimalType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DecimalType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DecimalType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_DecimalType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DecimalType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_DecimalType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_90__pyx_unpickle_DecimalType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_90__pyx_unpickle_DecimalType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DecimalType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe9db223, 0xc0cb30d, 0x997aa79):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe9db223, 0xc0cb30d, 0x997aa79) = (check_exp_overflow, context, precision, scale, scale_val, size, underlying_name))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__115, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe9db223, 0xc0cb30d, 0x997aa79):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe9db223, 0xc0cb30d, 0x997aa79) = (check_exp_overflow, context, precision, scale, scale_val, size, underlying_name))" % __pyx_checksum
 *     __pyx_result = DecimalType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe9db223, 0xc0cb30d, 0x997aa79):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe9db223, 0xc0cb30d, 0x997aa79) = (check_exp_overflow, context, precision, scale, scale_val, size, underlying_name))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = DecimalType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe9db223, 0xc0cb30d, 0x997aa79):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe9db223, 0xc0cb30d, 0x997aa79) = (check_exp_overflow, context, precision, scale, scale_val, size, underlying_name))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe9db223, 0xc0cb30d, 0x997aa79) = (check_exp_overflow, context, precision, scale, scale_val, size, underlying_name))" % __pyx_checksum
 *     __pyx_result = DecimalType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DecimalType__set_state(<DecimalType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_DecimalType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe9db223, 0xc0cb30d, 0x997aa79) = (check_exp_overflow, context, precision, scale, scale_val, size, underlying_name))" % __pyx_checksum
 *     __pyx_result = DecimalType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DecimalType__set_state(<DecimalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = DecimalType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DecimalType__set_state(<DecimalType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_DecimalType__set_state(DecimalType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_DecimalType__set_state(((struct __pyx_obj_5cavro_DecimalType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe9db223, 0xc0cb30d, 0x997aa79) = (check_exp_overflow, context, precision, scale, scale_val, size, underlying_name))" % __pyx_checksum
 *     __pyx_result = DecimalType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DecimalType__set_state(<DecimalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DecimalType__set_state(<DecimalType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_DecimalType__set_state(DecimalType __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_exp_overflow = __pyx_state[0]; __pyx_result.context = __pyx_state[1]; __pyx_result.precision = __pyx_state[2]; __pyx_result.scale = __pyx_state[3]; __pyx_result.scale_val = __pyx_state[4]; __pyx_result.size = __pyx_state[5]; __pyx_result.underlying_name = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_DecimalType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_DecimalType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_DecimalType__set_state(<DecimalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DecimalType__set_state(DecimalType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_exp_overflow = __pyx_state[0]; __pyx_result.context = __pyx_state[1]; __pyx_result.precision = __pyx_state[2]; __pyx_result.scale = __pyx_state[3]; __pyx_result.scale_val = __pyx_state[4]; __pyx_result.size = __pyx_state[5]; __pyx_result.underlying_name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_DecimalType__set_state(struct __pyx_obj_5cavro_DecimalType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DecimalType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_DecimalType__set_state(DecimalType __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_exp_overflow = __pyx_state[0]; __pyx_result.context = __pyx_state[1]; __pyx_result.precision = __pyx_state[2]; __pyx_result.scale = __pyx_state[3]; __pyx_result.scale_val = __pyx_state[4]; __pyx_result.size = __pyx_state[5]; __pyx_result.underlying_name = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->check_exp_overflow = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->context);
  __Pyx_DECREF(__pyx_v___pyx_result->context);
  __pyx_v___pyx_result->context = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->precision = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->scale = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->scale_val);
  __Pyx_DECREF(__pyx_v___pyx_result->scale_val);
  __pyx_v___pyx_result->scale_val = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->size);
  __Pyx_DECREF(__pyx_v___pyx_result->size);
  __pyx_v___pyx_result->size = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->underlying_name);
  __Pyx_DECREF(__pyx_v___pyx_result->underlying_name);
  __pyx_v___pyx_result->underlying_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_DecimalType__set_state(DecimalType __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_exp_overflow = __pyx_state[0]; __pyx_result.context = __pyx_state[1]; __pyx_result.precision = __pyx_state[2]; __pyx_result.scale = __pyx_state[3]; __pyx_result.scale_val = __pyx_state[4]; __pyx_result.size = __pyx_state[5]; __pyx_result.underlying_name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 7);
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.check_exp_overflow = __pyx_state[0]; __pyx_result.context = __pyx_state[1]; __pyx_result.precision = __pyx_state[2]; __pyx_result.scale = __pyx_state[3]; __pyx_result.scale_val = __pyx_state[4]; __pyx_result.size = __pyx_state[5]; __pyx_result.underlying_name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_3 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_3 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_3, 1+__pyx_t_3);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_DecimalType__set_state(DecimalType __pyx_result, tuple __pyx_state):
 *     __pyx_result.check_exp_overflow = __pyx_state[0]; __pyx_result.context = __pyx_state[1]; __pyx_result.precision = __pyx_state[2]; __pyx_result.scale = __pyx_state[3]; __pyx_result.scale_val = __pyx_state[4]; __pyx_result.size = __pyx_state[5]; __pyx_result.underlying_name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_DecimalType__set_state(<DecimalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DecimalType__set_state(DecimalType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_exp_overflow = __pyx_state[0]; __pyx_result.context = __pyx_state[1]; __pyx_result.precision = __pyx_state[2]; __pyx_result.scale = __pyx_state[3]; __pyx_result.scale_val = __pyx_state[4]; __pyx_result.size = __pyx_state[5]; __pyx_result.underlying_name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_DecimalType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_UUIDBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_93__pyx_unpickle_UUIDBase(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_93__pyx_unpickle_UUIDBase = {"__pyx_unpickle_UUIDBase", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_93__pyx_unpickle_UUIDBase, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_93__pyx_unpickle_UUIDBase(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_UUIDBase (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UUIDBase", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UUIDBase", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_UUIDBase") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UUIDBase", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_UUIDBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_92__pyx_unpickle_UUIDBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_92__pyx_unpickle_UUIDBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UUIDBase", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x987bfb3, 0xa7eca0c, 0x3ae7748):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__116, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x987bfb3, 0xa7eca0c, 0x3ae7748):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 *     __pyx_result = UUIDBase.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x987bfb3, 0xa7eca0c, 0x3ae7748):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = UUIDBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x987bfb3, 0xa7eca0c, 0x3ae7748):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 *     __pyx_result = UUIDBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UUIDBase__set_state(<UUIDBase> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_UUIDBase), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 *     __pyx_result = UUIDBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UUIDBase__set_state(<UUIDBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = UUIDBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UUIDBase__set_state(<UUIDBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDBase__set_state(UUIDBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_UUIDBase__set_state(((struct __pyx_obj_5cavro_UUIDBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 *     __pyx_result = UUIDBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UUIDBase__set_state(<UUIDBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UUIDBase__set_state(<UUIDBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_UUIDBase__set_state(UUIDBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_UUIDBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_UUIDBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_UUIDBase__set_state(<UUIDBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDBase__set_state(UUIDBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_UUIDBase__set_state(struct __pyx_obj_5cavro_UUIDBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UUIDBase__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDBase__set_state(UUIDBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.return_uuid_object = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->return_uuid_object = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_UUIDBase__set_state(UUIDBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_UUIDBase__set_state(UUIDBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_UUIDBase__set_state(<UUIDBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDBase__set_state(UUIDBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_UUIDBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_UUIDStringType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_95__pyx_unpickle_UUIDStringType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_95__pyx_unpickle_UUIDStringType = {"__pyx_unpickle_UUIDStringType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_95__pyx_unpickle_UUIDStringType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_95__pyx_unpickle_UUIDStringType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_UUIDStringType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UUIDStringType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UUIDStringType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_UUIDStringType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UUIDStringType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_UUIDStringType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_94__pyx_unpickle_UUIDStringType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_94__pyx_unpickle_UUIDStringType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UUIDStringType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x987bfb3, 0xa7eca0c, 0x3ae7748):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__116, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x987bfb3, 0xa7eca0c, 0x3ae7748):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 *     __pyx_result = UUIDStringType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x987bfb3, 0xa7eca0c, 0x3ae7748):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = UUIDStringType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x987bfb3, 0xa7eca0c, 0x3ae7748):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 *     __pyx_result = UUIDStringType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UUIDStringType__set_state(<UUIDStringType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_UUIDStringType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 *     __pyx_result = UUIDStringType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UUIDStringType__set_state(<UUIDStringType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = UUIDStringType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UUIDStringType__set_state(<UUIDStringType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDStringType__set_state(UUIDStringType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_UUIDStringType__set_state(((struct __pyx_obj_5cavro_UUIDStringType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 *     __pyx_result = UUIDStringType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UUIDStringType__set_state(<UUIDStringType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UUIDStringType__set_state(<UUIDStringType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_UUIDStringType__set_state(UUIDStringType __pyx_result, tuple __pyx_state):
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_UUIDStringType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_UUIDStringType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_UUIDStringType__set_state(<UUIDStringType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDStringType__set_state(UUIDStringType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_UUIDStringType__set_state(struct __pyx_obj_5cavro_UUIDStringType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UUIDStringType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDStringType__set_state(UUIDStringType __pyx_result, tuple __pyx_state):
 *     __pyx_result.return_uuid_object = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.return_uuid_object = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_UUIDStringType__set_state(UUIDStringType __pyx_result, tuple __pyx_state):
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_UUIDStringType__set_state(UUIDStringType __pyx_result, tuple __pyx_state):
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_UUIDStringType__set_state(<UUIDStringType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDStringType__set_state(UUIDStringType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_UUIDStringType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_UUIDFixedType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_97__pyx_unpickle_UUIDFixedType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_97__pyx_unpickle_UUIDFixedType = {"__pyx_unpickle_UUIDFixedType", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_97__pyx_unpickle_UUIDFixedType, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_97__pyx_unpickle_UUIDFixedType(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_UUIDFixedType (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UUIDFixedType", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UUIDFixedType", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_UUIDFixedType") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_UUIDFixedType", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_UUIDFixedType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_96__pyx_unpickle_UUIDFixedType(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_96__pyx_unpickle_UUIDFixedType(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UUIDFixedType", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x987bfb3, 0xa7eca0c, 0x3ae7748):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__116, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x987bfb3, 0xa7eca0c, 0x3ae7748):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 *     __pyx_result = UUIDFixedType.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x987bfb3, 0xa7eca0c, 0x3ae7748):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = UUIDFixedType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x987bfb3, 0xa7eca0c, 0x3ae7748):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 *     __pyx_result = UUIDFixedType.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UUIDFixedType__set_state(<UUIDFixedType> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_UUIDFixedType), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 *     __pyx_result = UUIDFixedType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UUIDFixedType__set_state(<UUIDFixedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = UUIDFixedType.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UUIDFixedType__set_state(<UUIDFixedType> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDFixedType__set_state(UUIDFixedType __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_UUIDFixedType__set_state(((struct __pyx_obj_5cavro_UUIDFixedType *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x987bfb3, 0xa7eca0c, 0x3ae7748) = (return_uuid_object))" % __pyx_checksum
 *     __pyx_result = UUIDFixedType.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_UUIDFixedType__set_state(<UUIDFixedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_UUIDFixedType__set_state(<UUIDFixedType> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_UUIDFixedType__set_state(UUIDFixedType __pyx_result, tuple __pyx_state):
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_UUIDFixedType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_UUIDFixedType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_UUIDFixedType__set_state(<UUIDFixedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDFixedType__set_state(UUIDFixedType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_UUIDFixedType__set_state(struct __pyx_obj_5cavro_UUIDFixedType *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_UUIDFixedType__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDFixedType__set_state(UUIDFixedType __pyx_result, tuple __pyx_state):
 *     __pyx_result.return_uuid_object = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.return_uuid_object = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_UUIDFixedType__set_state(UUIDFixedType __pyx_result, tuple __pyx_state):
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_UUIDFixedType__set_state(UUIDFixedType __pyx_result, tuple __pyx_state):
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_UUIDFixedType__set_state(<UUIDFixedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDFixedType__set_state(UUIDFixedType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_UUIDFixedType__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Date(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_99__pyx_unpickle_Date(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_99__pyx_unpickle_Date = {"__pyx_unpickle_Date", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_99__pyx_unpickle_Date, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_99__pyx_unpickle_Date(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Date (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Date", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Date", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Date") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Date", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Date", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_98__pyx_unpickle_Date(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_98__pyx_unpickle_Date(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Date", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc5331ff, 0xabd30c3, 0x97e5bf4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc5331ff, 0xabd30c3, 0x97e5bf4) = (accepts_string))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__117, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc5331ff, 0xabd30c3, 0x97e5bf4):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc5331ff, 0xabd30c3, 0x97e5bf4) = (accepts_string))" % __pyx_checksum
 *     __pyx_result = Date.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc5331ff, 0xabd30c3, 0x97e5bf4):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc5331ff, 0xabd30c3, 0x97e5bf4) = (accepts_string))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Date.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc5331ff, 0xabd30c3, 0x97e5bf4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc5331ff, 0xabd30c3, 0x97e5bf4) = (accepts_string))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc5331ff, 0xabd30c3, 0x97e5bf4) = (accepts_string))" % __pyx_checksum
 *     __pyx_result = Date.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Date__set_state(<Date> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_Date), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc5331ff, 0xabd30c3, 0x97e5bf4) = (accepts_string))" % __pyx_checksum
 *     __pyx_result = Date.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Date__set_state(<Date> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Date.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Date__set_state(<Date> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Date__set_state(Date __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_Date__set_state(((struct __pyx_obj_5cavro_Date *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc5331ff, 0xabd30c3, 0x97e5bf4) = (accepts_string))" % __pyx_checksum
 *     __pyx_result = Date.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Date__set_state(<Date> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Date__set_state(<Date> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Date__set_state(Date __pyx_result, tuple __pyx_state):
 *     __pyx_result.accepts_string = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Date(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Date", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Date__set_state(<Date> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Date__set_state(Date __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.accepts_string = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_Date__set_state(struct __pyx_obj_5cavro_Date *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Date__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Date__set_state(Date __pyx_result, tuple __pyx_state):
 *     __pyx_result.accepts_string = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->accepts_string = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Date__set_state(Date __pyx_result, tuple __pyx_state):
 *     __pyx_result.accepts_string = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.accepts_string = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Date__set_state(Date __pyx_result, tuple __pyx_state):
 *     __pyx_result.accepts_string = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Date__set_state(<Date> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Date__set_state(Date __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.accepts_string = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Date__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_TimeMillis(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_101__pyx_unpickle_TimeMillis(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_101__pyx_unpickle_TimeMillis = {"__pyx_unpickle_TimeMillis", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_101__pyx_unpickle_TimeMillis, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_101__pyx_unpickle_TimeMillis(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_TimeMillis (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TimeMillis", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TimeMillis", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_TimeMillis") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TimeMillis", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_TimeMillis", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_100__pyx_unpickle_TimeMillis(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_100__pyx_unpickle_TimeMillis(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TimeMillis", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = TimeMillis.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = TimeMillis.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = TimeMillis.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TimeMillis__set_state(<TimeMillis> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_TimeMillis), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = TimeMillis.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TimeMillis__set_state(<TimeMillis> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = TimeMillis.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TimeMillis__set_state(<TimeMillis> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_TimeMillis__set_state(TimeMillis __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_TimeMillis__set_state(((struct __pyx_obj_5cavro_TimeMillis *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = TimeMillis.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TimeMillis__set_state(<TimeMillis> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TimeMillis__set_state(<TimeMillis> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_TimeMillis__set_state(TimeMillis __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TimeMillis(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_TimeMillis", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_TimeMillis__set_state(<TimeMillis> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TimeMillis__set_state(TimeMillis __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_TimeMillis__set_state(struct __pyx_obj_5cavro_TimeMillis *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TimeMillis__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TimeMillis__set_state(TimeMillis __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_TimeMillis__set_state(TimeMillis __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TimeMillis__set_state(TimeMillis __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_TimeMillis__set_state(<TimeMillis> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TimeMillis__set_state(TimeMillis __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_TimeMillis__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_TimeMicros(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_103__pyx_unpickle_TimeMicros(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_103__pyx_unpickle_TimeMicros = {"__pyx_unpickle_TimeMicros", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_103__pyx_unpickle_TimeMicros, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_103__pyx_unpickle_TimeMicros(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_TimeMicros (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TimeMicros", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TimeMicros", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_TimeMicros") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TimeMicros", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_TimeMicros", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_102__pyx_unpickle_TimeMicros(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_102__pyx_unpickle_TimeMicros(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TimeMicros", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = TimeMicros.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = TimeMicros.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = TimeMicros.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TimeMicros__set_state(<TimeMicros> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_TimeMicros), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = TimeMicros.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TimeMicros__set_state(<TimeMicros> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = TimeMicros.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TimeMicros__set_state(<TimeMicros> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_TimeMicros__set_state(TimeMicros __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_TimeMicros__set_state(((struct __pyx_obj_5cavro_TimeMicros *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = TimeMicros.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TimeMicros__set_state(<TimeMicros> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TimeMicros__set_state(<TimeMicros> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_TimeMicros__set_state(TimeMicros __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TimeMicros(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_TimeMicros", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_TimeMicros__set_state(<TimeMicros> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TimeMicros__set_state(TimeMicros __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_TimeMicros__set_state(struct __pyx_obj_5cavro_TimeMicros *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TimeMicros__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TimeMicros__set_state(TimeMicros __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_TimeMicros__set_state(TimeMicros __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TimeMicros__set_state(TimeMicros __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_TimeMicros__set_state(<TimeMicros> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TimeMicros__set_state(TimeMicros __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_TimeMicros__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_TimestampMillis(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_105__pyx_unpickle_TimestampMillis(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_105__pyx_unpickle_TimestampMillis = {"__pyx_unpickle_TimestampMillis", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_105__pyx_unpickle_TimestampMillis, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_105__pyx_unpickle_TimestampMillis(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_TimestampMillis (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TimestampMillis", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TimestampMillis", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_TimestampMillis") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TimestampMillis", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_TimestampMillis", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_104__pyx_unpickle_TimestampMillis(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_104__pyx_unpickle_TimestampMillis(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TimestampMillis", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8d5b257, 0x5b53641, 0x2d8f391):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d5b257, 0x5b53641, 0x2d8f391) = (alternate_timestamp_encoding))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__118, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8d5b257, 0x5b53641, 0x2d8f391):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d5b257, 0x5b53641, 0x2d8f391) = (alternate_timestamp_encoding))" % __pyx_checksum
 *     __pyx_result = TimestampMillis.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x8d5b257, 0x5b53641, 0x2d8f391):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d5b257, 0x5b53641, 0x2d8f391) = (alternate_timestamp_encoding))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = TimestampMillis.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_25, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8d5b257, 0x5b53641, 0x2d8f391):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d5b257, 0x5b53641, 0x2d8f391) = (alternate_timestamp_encoding))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d5b257, 0x5b53641, 0x2d8f391) = (alternate_timestamp_encoding))" % __pyx_checksum
 *     __pyx_result = TimestampMillis.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TimestampMillis__set_state(<TimestampMillis> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_TimestampMillis), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d5b257, 0x5b53641, 0x2d8f391) = (alternate_timestamp_encoding))" % __pyx_checksum
 *     __pyx_result = TimestampMillis.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TimestampMillis__set_state(<TimestampMillis> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = TimestampMillis.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TimestampMillis__set_state(<TimestampMillis> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_TimestampMillis__set_state(TimestampMillis __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_TimestampMillis__set_state(((struct __pyx_obj_5cavro_TimestampMillis *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d5b257, 0x5b53641, 0x2d8f391) = (alternate_timestamp_encoding))" % __pyx_checksum
 *     __pyx_result = TimestampMillis.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TimestampMillis__set_state(<TimestampMillis> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TimestampMillis__set_state(<TimestampMillis> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_TimestampMillis__set_state(TimestampMillis __pyx_result, tuple __pyx_state):
 *     __pyx_result.alternate_timestamp_encoding = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TimestampMillis(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_TimestampMillis", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_TimestampMillis__set_state(<TimestampMillis> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TimestampMillis__set_state(TimestampMillis __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.alternate_timestamp_encoding = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_TimestampMillis__set_state(struct __pyx_obj_5cavro_TimestampMillis *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TimestampMillis__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TimestampMillis__set_state(TimestampMillis __pyx_result, tuple __pyx_state):
 *     __pyx_result.alternate_timestamp_encoding = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->alternate_timestamp_encoding = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_TimestampMillis__set_state(TimestampMillis __pyx_result, tuple __pyx_state):
 *     __pyx_result.alternate_timestamp_encoding = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.alternate_timestamp_encoding = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_TimestampMillis__set_state(TimestampMillis __pyx_result, tuple __pyx_state):
 *     __pyx_result.alternate_timestamp_encoding = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_TimestampMillis__set_state(<TimestampMillis> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TimestampMillis__set_state(TimestampMillis __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.alternate_timestamp_encoding = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_TimestampMillis__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_TimestampMicros(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_107__pyx_unpickle_TimestampMicros(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_107__pyx_unpickle_TimestampMicros = {"__pyx_unpickle_TimestampMicros", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_107__pyx_unpickle_TimestampMicros, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_107__pyx_unpickle_TimestampMicros(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_TimestampMicros (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TimestampMicros", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TimestampMicros", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_TimestampMicros") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TimestampMicros", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_TimestampMicros", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_106__pyx_unpickle_TimestampMicros(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_106__pyx_unpickle_TimestampMicros(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TimestampMicros", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = TimestampMicros.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = TimestampMicros.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = TimestampMicros.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TimestampMicros__set_state(<TimestampMicros> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_TimestampMicros), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = TimestampMicros.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TimestampMicros__set_state(<TimestampMicros> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = TimestampMicros.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TimestampMicros__set_state(<TimestampMicros> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_TimestampMicros__set_state(TimestampMicros __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_TimestampMicros__set_state(((struct __pyx_obj_5cavro_TimestampMicros *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = TimestampMicros.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TimestampMicros__set_state(<TimestampMicros> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TimestampMicros__set_state(<TimestampMicros> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_TimestampMicros__set_state(TimestampMicros __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TimestampMicros(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_TimestampMicros", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_TimestampMicros__set_state(<TimestampMicros> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TimestampMicros__set_state(TimestampMicros __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_TimestampMicros__set_state(struct __pyx_obj_5cavro_TimestampMicros *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TimestampMicros__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TimestampMicros__set_state(TimestampMicros __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_TimestampMicros__set_state(TimestampMicros __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TimestampMicros__set_state(TimestampMicros __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_TimestampMicros__set_state(<TimestampMicros> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TimestampMicros__set_state(TimestampMicros __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_TimestampMicros__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Codec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_109__pyx_unpickle_Codec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_109__pyx_unpickle_Codec = {"__pyx_unpickle_Codec", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_109__pyx_unpickle_Codec, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_109__pyx_unpickle_Codec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Codec (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Codec", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Codec", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Codec") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Codec", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Codec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_108__pyx_unpickle_Codec(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_108__pyx_unpickle_Codec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Codec", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = Codec.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Codec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = Codec.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Codec__set_state(<Codec> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_Codec), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = Codec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Codec__set_state(<Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Codec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Codec__set_state(<Codec> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Codec__set_state(Codec __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_Codec__set_state(((struct __pyx_obj_5cavro_Codec *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = Codec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Codec__set_state(<Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Codec__set_state(<Codec> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Codec__set_state(Codec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Codec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Codec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Codec__set_state(<Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Codec__set_state(Codec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_Codec__set_state(struct __pyx_obj_5cavro_Codec *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Codec__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Codec__set_state(Codec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Codec__set_state(Codec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Codec__set_state(Codec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Codec__set_state(<Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Codec__set_state(Codec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_Codec__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__SnappyCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_111__pyx_unpickle__SnappyCodec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_111__pyx_unpickle__SnappyCodec = {"__pyx_unpickle__SnappyCodec", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_111__pyx_unpickle__SnappyCodec, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_111__pyx_unpickle__SnappyCodec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__SnappyCodec (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__SnappyCodec", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__SnappyCodec", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__SnappyCodec") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__SnappyCodec", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__SnappyCodec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_110__pyx_unpickle__SnappyCodec(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_110__pyx_unpickle__SnappyCodec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__SnappyCodec", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _SnappyCodec.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _SnappyCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _SnappyCodec.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__SnappyCodec__set_state(<_SnappyCodec> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__SnappyCodec), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _SnappyCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__SnappyCodec__set_state(<_SnappyCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _SnappyCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__SnappyCodec__set_state(<_SnappyCodec> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__SnappyCodec__set_state(_SnappyCodec __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__SnappyCodec__set_state(((struct __pyx_obj_5cavro__SnappyCodec *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _SnappyCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__SnappyCodec__set_state(<_SnappyCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__SnappyCodec__set_state(<_SnappyCodec> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__SnappyCodec__set_state(_SnappyCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__SnappyCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__SnappyCodec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__SnappyCodec__set_state(<_SnappyCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__SnappyCodec__set_state(_SnappyCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__SnappyCodec__set_state(struct __pyx_obj_5cavro__SnappyCodec *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__SnappyCodec__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__SnappyCodec__set_state(_SnappyCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__SnappyCodec__set_state(_SnappyCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__SnappyCodec__set_state(_SnappyCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__SnappyCodec__set_state(<_SnappyCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__SnappyCodec__set_state(_SnappyCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__SnappyCodec__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__NullCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_113__pyx_unpickle__NullCodec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_113__pyx_unpickle__NullCodec = {"__pyx_unpickle__NullCodec", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_113__pyx_unpickle__NullCodec, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_113__pyx_unpickle__NullCodec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__NullCodec (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__NullCodec", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__NullCodec", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__NullCodec") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__NullCodec", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__NullCodec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_112__pyx_unpickle__NullCodec(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_112__pyx_unpickle__NullCodec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__NullCodec", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _NullCodec.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _NullCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _NullCodec.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__NullCodec__set_state(<_NullCodec> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__NullCodec), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _NullCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__NullCodec__set_state(<_NullCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _NullCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__NullCodec__set_state(<_NullCodec> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__NullCodec__set_state(_NullCodec __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__NullCodec__set_state(((struct __pyx_obj_5cavro__NullCodec *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _NullCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__NullCodec__set_state(<_NullCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__NullCodec__set_state(<_NullCodec> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__NullCodec__set_state(_NullCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__NullCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__NullCodec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__NullCodec__set_state(<_NullCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__NullCodec__set_state(_NullCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__NullCodec__set_state(struct __pyx_obj_5cavro__NullCodec *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__NullCodec__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__NullCodec__set_state(_NullCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__NullCodec__set_state(_NullCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__NullCodec__set_state(_NullCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__NullCodec__set_state(<_NullCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__NullCodec__set_state(_NullCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__NullCodec__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__DeflateCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_115__pyx_unpickle__DeflateCodec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_115__pyx_unpickle__DeflateCodec = {"__pyx_unpickle__DeflateCodec", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_115__pyx_unpickle__DeflateCodec, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_115__pyx_unpickle__DeflateCodec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__DeflateCodec (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__DeflateCodec", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__DeflateCodec", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__DeflateCodec") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__DeflateCodec", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__DeflateCodec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_114__pyx_unpickle__DeflateCodec(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_114__pyx_unpickle__DeflateCodec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__DeflateCodec", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _DeflateCodec.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _DeflateCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _DeflateCodec.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__DeflateCodec__set_state(<_DeflateCodec> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__DeflateCodec), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _DeflateCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__DeflateCodec__set_state(<_DeflateCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _DeflateCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__DeflateCodec__set_state(<_DeflateCodec> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__DeflateCodec__set_state(_DeflateCodec __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__DeflateCodec__set_state(((struct __pyx_obj_5cavro__DeflateCodec *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _DeflateCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__DeflateCodec__set_state(<_DeflateCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__DeflateCodec__set_state(<_DeflateCodec> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__DeflateCodec__set_state(_DeflateCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__DeflateCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__DeflateCodec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__DeflateCodec__set_state(<_DeflateCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__DeflateCodec__set_state(_DeflateCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__DeflateCodec__set_state(struct __pyx_obj_5cavro__DeflateCodec *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__DeflateCodec__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__DeflateCodec__set_state(_DeflateCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__DeflateCodec__set_state(_DeflateCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__DeflateCodec__set_state(_DeflateCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__DeflateCodec__set_state(<_DeflateCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__DeflateCodec__set_state(_DeflateCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__DeflateCodec__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle___Bzip2Codec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_117__pyx_unpickle___Bzip2Codec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_117__pyx_unpickle___Bzip2Codec = {"__pyx_unpickle___Bzip2Codec", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_117__pyx_unpickle___Bzip2Codec, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_117__pyx_unpickle___Bzip2Codec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle___Bzip2Codec (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Bzip2Codec", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Bzip2Codec", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle___Bzip2Codec") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Bzip2Codec", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle___Bzip2Codec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_116__pyx_unpickle___Bzip2Codec(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_116__pyx_unpickle___Bzip2Codec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle___Bzip2Codec", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = __Bzip2Codec.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = __Bzip2Codec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = __Bzip2Codec.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Bzip2Codec__set_state(<__Bzip2Codec> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro___Bzip2Codec), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = __Bzip2Codec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___Bzip2Codec__set_state(<__Bzip2Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = __Bzip2Codec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Bzip2Codec__set_state(<__Bzip2Codec> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle___Bzip2Codec__set_state(__Bzip2Codec __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle___Bzip2Codec__set_state(((struct __pyx_obj_5cavro___Bzip2Codec *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = __Bzip2Codec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___Bzip2Codec__set_state(<__Bzip2Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Bzip2Codec__set_state(<__Bzip2Codec> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle___Bzip2Codec__set_state(__Bzip2Codec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle___Bzip2Codec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle___Bzip2Codec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle___Bzip2Codec__set_state(<__Bzip2Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___Bzip2Codec__set_state(__Bzip2Codec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle___Bzip2Codec__set_state(struct __pyx_obj_5cavro___Bzip2Codec *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle___Bzip2Codec__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___Bzip2Codec__set_state(__Bzip2Codec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle___Bzip2Codec__set_state(__Bzip2Codec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___Bzip2Codec__set_state(__Bzip2Codec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle___Bzip2Codec__set_state(<__Bzip2Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___Bzip2Codec__set_state(__Bzip2Codec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle___Bzip2Codec__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__LzmaCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_119__pyx_unpickle__LzmaCodec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_119__pyx_unpickle__LzmaCodec = {"__pyx_unpickle__LzmaCodec", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_119__pyx_unpickle__LzmaCodec, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_119__pyx_unpickle__LzmaCodec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__LzmaCodec (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__LzmaCodec", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__LzmaCodec", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__LzmaCodec") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__LzmaCodec", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__LzmaCodec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_118__pyx_unpickle__LzmaCodec(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_118__pyx_unpickle__LzmaCodec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__LzmaCodec", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _LzmaCodec.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _LzmaCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _LzmaCodec.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__LzmaCodec__set_state(<_LzmaCodec> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__LzmaCodec), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _LzmaCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__LzmaCodec__set_state(<_LzmaCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _LzmaCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__LzmaCodec__set_state(<_LzmaCodec> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__LzmaCodec__set_state(_LzmaCodec __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__LzmaCodec__set_state(((struct __pyx_obj_5cavro__LzmaCodec *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _LzmaCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__LzmaCodec__set_state(<_LzmaCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__LzmaCodec__set_state(<_LzmaCodec> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__LzmaCodec__set_state(_LzmaCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__LzmaCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__LzmaCodec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__LzmaCodec__set_state(<_LzmaCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__LzmaCodec__set_state(_LzmaCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__LzmaCodec__set_state(struct __pyx_obj_5cavro__LzmaCodec *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__LzmaCodec__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__LzmaCodec__set_state(_LzmaCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__LzmaCodec__set_state(_LzmaCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__LzmaCodec__set_state(_LzmaCodec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__LzmaCodec__set_state(<_LzmaCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__LzmaCodec__set_state(_LzmaCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__LzmaCodec__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__ZStandardCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_121__pyx_unpickle__ZStandardCodec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_121__pyx_unpickle__ZStandardCodec = {"__pyx_unpickle__ZStandardCodec", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_121__pyx_unpickle__ZStandardCodec, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_121__pyx_unpickle__ZStandardCodec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__ZStandardCodec (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__ZStandardCodec", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__ZStandardCodec", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__ZStandardCodec") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__ZStandardCodec", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__ZStandardCodec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_120__pyx_unpickle__ZStandardCodec(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_120__pyx_unpickle__ZStandardCodec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__ZStandardCodec", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8d57cb2, 0xfb98f56, 0xb8cd0c2):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d57cb2, 0xfb98f56, 0xb8cd0c2) = (compressor, decompressor))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__119, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8d57cb2, 0xfb98f56, 0xb8cd0c2):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d57cb2, 0xfb98f56, 0xb8cd0c2) = (compressor, decompressor))" % __pyx_checksum
 *     __pyx_result = _ZStandardCodec.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x8d57cb2, 0xfb98f56, 0xb8cd0c2):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d57cb2, 0xfb98f56, 0xb8cd0c2) = (compressor, decompressor))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _ZStandardCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_26, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x8d57cb2, 0xfb98f56, 0xb8cd0c2):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d57cb2, 0xfb98f56, 0xb8cd0c2) = (compressor, decompressor))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d57cb2, 0xfb98f56, 0xb8cd0c2) = (compressor, decompressor))" % __pyx_checksum
 *     __pyx_result = _ZStandardCodec.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__ZStandardCodec__set_state(<_ZStandardCodec> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__ZStandardCodec), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d57cb2, 0xfb98f56, 0xb8cd0c2) = (compressor, decompressor))" % __pyx_checksum
 *     __pyx_result = _ZStandardCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__ZStandardCodec__set_state(<_ZStandardCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _ZStandardCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__ZStandardCodec__set_state(<_ZStandardCodec> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__ZStandardCodec__set_state(_ZStandardCodec __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__ZStandardCodec__set_state(((struct __pyx_obj_5cavro__ZStandardCodec *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x8d57cb2, 0xfb98f56, 0xb8cd0c2) = (compressor, decompressor))" % __pyx_checksum
 *     __pyx_result = _ZStandardCodec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__ZStandardCodec__set_state(<_ZStandardCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__ZStandardCodec__set_state(<_ZStandardCodec> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__ZStandardCodec__set_state(_ZStandardCodec __pyx_result, tuple __pyx_state):
 *     __pyx_result.compressor = __pyx_state[0]; __pyx_result.decompressor = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__ZStandardCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__ZStandardCodec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__ZStandardCodec__set_state(<_ZStandardCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__ZStandardCodec__set_state(_ZStandardCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.compressor = __pyx_state[0]; __pyx_result.decompressor = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__ZStandardCodec__set_state(struct __pyx_obj_5cavro__ZStandardCodec *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__ZStandardCodec__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__ZStandardCodec__set_state(_ZStandardCodec __pyx_result, tuple __pyx_state):
 *     __pyx_result.compressor = __pyx_state[0]; __pyx_result.decompressor = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->compressor);
  __Pyx_DECREF(__pyx_v___pyx_result->compressor);
  __pyx_v___pyx_result->compressor = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->decompressor);
  __Pyx_DECREF(__pyx_v___pyx_result->decompressor);
  __pyx_v___pyx_result->decompressor = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle__ZStandardCodec__set_state(_ZStandardCodec __pyx_result, tuple __pyx_state):
 *     __pyx_result.compressor = __pyx_state[0]; __pyx_result.decompressor = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 2);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.compressor = __pyx_state[0]; __pyx_result.decompressor = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__ZStandardCodec__set_state(_ZStandardCodec __pyx_result, tuple __pyx_state):
 *     __pyx_result.compressor = __pyx_state[0]; __pyx_result.decompressor = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__ZStandardCodec__set_state(<_ZStandardCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__ZStandardCodec__set_state(_ZStandardCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.compressor = __pyx_state[0]; __pyx_result.decompressor = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__ZStandardCodec__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__Lz4Codec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_123__pyx_unpickle__Lz4Codec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_123__pyx_unpickle__Lz4Codec = {"__pyx_unpickle__Lz4Codec", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_123__pyx_unpickle__Lz4Codec, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_123__pyx_unpickle__Lz4Codec(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__Lz4Codec (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__Lz4Codec", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__Lz4Codec", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle__Lz4Codec") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__Lz4Codec", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle__Lz4Codec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_122__pyx_unpickle__Lz4Codec(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_122__pyx_unpickle__Lz4Codec(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__Lz4Codec", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _Lz4Codec.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = _Lz4Codec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _Lz4Codec.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__Lz4Codec__set_state(<_Lz4Codec> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro__Lz4Codec), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _Lz4Codec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__Lz4Codec__set_state(<_Lz4Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _Lz4Codec.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__Lz4Codec__set_state(<_Lz4Codec> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__Lz4Codec__set_state(_Lz4Codec __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle__Lz4Codec__set_state(((struct __pyx_obj_5cavro__Lz4Codec *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = _Lz4Codec.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__Lz4Codec__set_state(<_Lz4Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__Lz4Codec__set_state(<_Lz4Codec> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__Lz4Codec__set_state(_Lz4Codec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__Lz4Codec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__Lz4Codec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__Lz4Codec__set_state(<_Lz4Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__Lz4Codec__set_state(_Lz4Codec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle__Lz4Codec__set_state(struct __pyx_obj_5cavro__Lz4Codec *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__Lz4Codec__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__Lz4Codec__set_state(_Lz4Codec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__Lz4Codec__set_state(_Lz4Codec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__Lz4Codec__set_state(_Lz4Codec __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__Lz4Codec__set_state(<_Lz4Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__Lz4Codec__set_state(_Lz4Codec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle__Lz4Codec__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ContainerReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_125__pyx_unpickle_ContainerReader(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_125__pyx_unpickle_ContainerReader = {"__pyx_unpickle_ContainerReader", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_125__pyx_unpickle_ContainerReader, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_125__pyx_unpickle_ContainerReader(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ContainerReader (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ContainerReader", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ContainerReader", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_ContainerReader") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ContainerReader", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_ContainerReader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_124__pyx_unpickle_ContainerReader(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_124__pyx_unpickle_ContainerReader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ContainerReader", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x289807d, 0x75481bf, 0x9c7e164):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x289807d, 0x75481bf, 0x9c7e164) = (codec, codec_name, current_block, marker, metadata, objects_left_in_block, reader, reader_schema, schema, writer_schema))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__120, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x289807d, 0x75481bf, 0x9c7e164):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x289807d, 0x75481bf, 0x9c7e164) = (codec, codec_name, current_block, marker, metadata, objects_left_in_block, reader, reader_schema, schema, writer_schema))" % __pyx_checksum
 *     __pyx_result = ContainerReader.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x289807d, 0x75481bf, 0x9c7e164):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x289807d, 0x75481bf, 0x9c7e164) = (codec, codec_name, current_block, marker, metadata, objects_left_in_block, reader, reader_schema, schema, writer_schema))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = ContainerReader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_27, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x289807d, 0x75481bf, 0x9c7e164):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x289807d, 0x75481bf, 0x9c7e164) = (codec, codec_name, current_block, marker, metadata, objects_left_in_block, reader, reader_schema, schema, writer_schema))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x289807d, 0x75481bf, 0x9c7e164) = (codec, codec_name, current_block, marker, metadata, objects_left_in_block, reader, reader_schema, schema, writer_schema))" % __pyx_checksum
 *     __pyx_result = ContainerReader.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ContainerReader__set_state(<ContainerReader> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_ContainerReader), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x289807d, 0x75481bf, 0x9c7e164) = (codec, codec_name, current_block, marker, metadata, objects_left_in_block, reader, reader_schema, schema, writer_schema))" % __pyx_checksum
 *     __pyx_result = ContainerReader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ContainerReader__set_state(<ContainerReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = ContainerReader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ContainerReader__set_state(<ContainerReader> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ContainerReader__set_state(ContainerReader __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_ContainerReader__set_state(((struct __pyx_obj_5cavro_ContainerReader *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x289807d, 0x75481bf, 0x9c7e164) = (codec, codec_name, current_block, marker, metadata, objects_left_in_block, reader, reader_schema, schema, writer_schema))" % __pyx_checksum
 *     __pyx_result = ContainerReader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ContainerReader__set_state(<ContainerReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ContainerReader__set_state(<ContainerReader> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ContainerReader__set_state(ContainerReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.codec = __pyx_state[0]; __pyx_result.codec_name = __pyx_state[1]; __pyx_result.current_block = __pyx_state[2]; __pyx_result.marker = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.objects_left_in_block = __pyx_state[5]; __pyx_result.reader = __pyx_state[6]; __pyx_result.reader_schema = __pyx_state[7]; __pyx_result.schema = __pyx_state[8]; __pyx_result.writer_schema = __pyx_state[9]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ContainerReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_ContainerReader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ContainerReader__set_state(<ContainerReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ContainerReader__set_state(ContainerReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.codec = __pyx_state[0]; __pyx_result.codec_name = __pyx_state[1]; __pyx_result.current_block = __pyx_state[2]; __pyx_result.marker = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.objects_left_in_block = __pyx_state[5]; __pyx_result.reader = __pyx_state[6]; __pyx_result.reader_schema = __pyx_state[7]; __pyx_result.schema = __pyx_state[8]; __pyx_result.writer_schema = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_ContainerReader__set_state(struct __pyx_obj_5cavro_ContainerReader *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  size_t __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ContainerReader__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ContainerReader__set_state(ContainerReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.codec = __pyx_state[0]; __pyx_result.codec_name = __pyx_state[1]; __pyx_result.current_block = __pyx_state[2]; __pyx_result.marker = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.objects_left_in_block = __pyx_state[5]; __pyx_result.reader = __pyx_state[6]; __pyx_result.reader_schema = __pyx_state[7]; __pyx_result.schema = __pyx_state[8]; __pyx_result.writer_schema = __pyx_state[9]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Codec))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->codec);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->codec);
  __pyx_v___pyx_result->codec = ((struct __pyx_obj_5cavro_Codec *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->codec_name);
  __Pyx_DECREF(__pyx_v___pyx_result->codec_name);
  __pyx_v___pyx_result->codec_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_MemoryReader))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->current_block);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->current_block);
  __pyx_v___pyx_result->current_block = ((struct __pyx_obj_5cavro_MemoryReader *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_1, 0); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XCLEAR_MEMVIEW(&__pyx_v___pyx_result->marker, 0);
  __pyx_v___pyx_result->marker = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->metadata);
  __Pyx_DECREF(__pyx_v___pyx_result->metadata);
  __pyx_v___pyx_result->metadata = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->objects_left_in_block = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro__Reader))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->reader);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->reader);
  __pyx_v___pyx_result->reader = ((struct __pyx_obj_5cavro__Reader *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Schema))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->reader_schema);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->reader_schema);
  __pyx_v___pyx_result->reader_schema = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Schema))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->schema);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->schema);
  __pyx_v___pyx_result->schema = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_Schema))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->writer_schema);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->writer_schema);
  __pyx_v___pyx_result->writer_schema = ((struct __pyx_obj_5cavro_Schema *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_ContainerReader__set_state(ContainerReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.codec = __pyx_state[0]; __pyx_result.codec_name = __pyx_state[1]; __pyx_result.current_block = __pyx_state[2]; __pyx_result.marker = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.objects_left_in_block = __pyx_state[5]; __pyx_result.reader = __pyx_state[6]; __pyx_result.reader_schema = __pyx_state[7]; __pyx_result.schema = __pyx_state[8]; __pyx_result.writer_schema = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 > 10);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.codec = __pyx_state[0]; __pyx_result.codec_name = __pyx_state[1]; __pyx_result.current_block = __pyx_state[2]; __pyx_result.marker = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.objects_left_in_block = __pyx_state[5]; __pyx_result.reader = __pyx_state[6]; __pyx_result.reader_schema = __pyx_state[7]; __pyx_result.schema = __pyx_state[8]; __pyx_result.writer_schema = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[10])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ContainerReader__set_state(ContainerReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.codec = __pyx_state[0]; __pyx_result.codec_name = __pyx_state[1]; __pyx_result.current_block = __pyx_state[2]; __pyx_result.marker = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.objects_left_in_block = __pyx_state[5]; __pyx_result.reader = __pyx_state[6]; __pyx_result.reader_schema = __pyx_state[7]; __pyx_result.schema = __pyx_state[8]; __pyx_result.writer_schema = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ContainerReader__set_state(<ContainerReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ContainerReader__set_state(ContainerReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.codec = __pyx_state[0]; __pyx_result.codec_name = __pyx_state[1]; __pyx_result.current_block = __pyx_state[2]; __pyx_result.marker = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.objects_left_in_block = __pyx_state[5]; __pyx_result.reader = __pyx_state[6]; __pyx_result.reader_schema = __pyx_state[7]; __pyx_result.schema = __pyx_state[8]; __pyx_result.writer_schema = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_2, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_ContainerReader__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PromoteToFloat(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_127__pyx_unpickle_PromoteToFloat(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_127__pyx_unpickle_PromoteToFloat = {"__pyx_unpickle_PromoteToFloat", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_127__pyx_unpickle_PromoteToFloat, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_127__pyx_unpickle_PromoteToFloat(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromoteToFloat (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromoteToFloat", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromoteToFloat", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PromoteToFloat") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromoteToFloat", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromoteToFloat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_126__pyx_unpickle_PromoteToFloat(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_126__pyx_unpickle_PromoteToFloat(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromoteToFloat", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = PromoteToFloat.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PromoteToFloat.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = PromoteToFloat.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromoteToFloat__set_state(<PromoteToFloat> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_PromoteToFloat), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = PromoteToFloat.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PromoteToFloat__set_state(<PromoteToFloat> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PromoteToFloat.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromoteToFloat__set_state(<PromoteToFloat> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteToFloat__set_state(PromoteToFloat __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_PromoteToFloat__set_state(((struct __pyx_obj_5cavro_PromoteToFloat *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = PromoteToFloat.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PromoteToFloat__set_state(<PromoteToFloat> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromoteToFloat__set_state(<PromoteToFloat> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PromoteToFloat__set_state(PromoteToFloat __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PromoteToFloat(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromoteToFloat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PromoteToFloat__set_state(<PromoteToFloat> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteToFloat__set_state(PromoteToFloat __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_PromoteToFloat__set_state(struct __pyx_obj_5cavro_PromoteToFloat *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromoteToFloat__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteToFloat__set_state(PromoteToFloat __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PromoteToFloat__set_state(PromoteToFloat __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteToFloat__set_state(PromoteToFloat __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PromoteToFloat__set_state(<PromoteToFloat> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteToFloat__set_state(PromoteToFloat __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromoteToFloat__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PromoteBytesToString(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_129__pyx_unpickle_PromoteBytesToString(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_129__pyx_unpickle_PromoteBytesToString = {"__pyx_unpickle_PromoteBytesToString", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_129__pyx_unpickle_PromoteBytesToString, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_129__pyx_unpickle_PromoteBytesToString(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromoteBytesToString (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromoteBytesToString", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromoteBytesToString", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PromoteBytesToString") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromoteBytesToString", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromoteBytesToString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_128__pyx_unpickle_PromoteBytesToString(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_128__pyx_unpickle_PromoteBytesToString(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromoteBytesToString", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = PromoteBytesToString.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PromoteBytesToString.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = PromoteBytesToString.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromoteBytesToString__set_state(<PromoteBytesToString> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_PromoteBytesToString), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = PromoteBytesToString.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PromoteBytesToString__set_state(<PromoteBytesToString> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PromoteBytesToString.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromoteBytesToString__set_state(<PromoteBytesToString> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteBytesToString__set_state(PromoteBytesToString __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_PromoteBytesToString__set_state(((struct __pyx_obj_5cavro_PromoteBytesToString *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = PromoteBytesToString.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PromoteBytesToString__set_state(<PromoteBytesToString> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromoteBytesToString__set_state(<PromoteBytesToString> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PromoteBytesToString__set_state(PromoteBytesToString __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PromoteBytesToString(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromoteBytesToString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PromoteBytesToString__set_state(<PromoteBytesToString> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteBytesToString__set_state(PromoteBytesToString __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_PromoteBytesToString__set_state(struct __pyx_obj_5cavro_PromoteBytesToString *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromoteBytesToString__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteBytesToString__set_state(PromoteBytesToString __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PromoteBytesToString__set_state(PromoteBytesToString __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteBytesToString__set_state(PromoteBytesToString __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PromoteBytesToString__set_state(<PromoteBytesToString> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteBytesToString__set_state(PromoteBytesToString __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromoteBytesToString__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PromoteStringToBytes(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_131__pyx_unpickle_PromoteStringToBytes(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_131__pyx_unpickle_PromoteStringToBytes = {"__pyx_unpickle_PromoteStringToBytes", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_131__pyx_unpickle_PromoteStringToBytes, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_131__pyx_unpickle_PromoteStringToBytes(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromoteStringToBytes (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromoteStringToBytes", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromoteStringToBytes", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PromoteStringToBytes") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PromoteStringToBytes", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromoteStringToBytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_130__pyx_unpickle_PromoteStringToBytes(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_130__pyx_unpickle_PromoteStringToBytes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromoteStringToBytes", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__96, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = PromoteStringToBytes.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PromoteStringToBytes.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = PromoteStringToBytes.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromoteStringToBytes__set_state(<PromoteStringToBytes> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_PromoteStringToBytes), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = PromoteStringToBytes.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PromoteStringToBytes__set_state(<PromoteStringToBytes> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PromoteStringToBytes.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromoteStringToBytes__set_state(<PromoteStringToBytes> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteStringToBytes__set_state(PromoteStringToBytes __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_PromoteStringToBytes__set_state(((struct __pyx_obj_5cavro_PromoteStringToBytes *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = PromoteStringToBytes.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PromoteStringToBytes__set_state(<PromoteStringToBytes> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PromoteStringToBytes__set_state(<PromoteStringToBytes> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PromoteStringToBytes__set_state(PromoteStringToBytes __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PromoteStringToBytes(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromoteStringToBytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PromoteStringToBytes__set_state(<PromoteStringToBytes> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteStringToBytes__set_state(PromoteStringToBytes __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_PromoteStringToBytes__set_state(struct __pyx_obj_5cavro_PromoteStringToBytes *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PromoteStringToBytes__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteStringToBytes__set_state(PromoteStringToBytes __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PromoteStringToBytes__set_state(PromoteStringToBytes __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteStringToBytes__set_state(PromoteStringToBytes __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PromoteStringToBytes__set_state(<PromoteStringToBytes> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteStringToBytes__set_state(PromoteStringToBytes __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_PromoteStringToBytes__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_CannotPromote(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5cavro_133__pyx_unpickle_CannotPromote(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_5cavro_133__pyx_unpickle_CannotPromote = {"__pyx_unpickle_CannotPromote", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_133__pyx_unpickle_CannotPromote, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5cavro_133__pyx_unpickle_CannotPromote(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CannotPromote (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CannotPromote", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CannotPromote", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_CannotPromote") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CannotPromote", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cavro.__pyx_unpickle_CannotPromote", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5cavro_132__pyx_unpickle_CannotPromote(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5cavro_132__pyx_unpickle_CannotPromote(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CannotPromote", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcbfa7ba, 0xf8f7277, 0xca3fdae):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcbfa7ba, 0xf8f7277, 0xca3fdae) = (extra, reader_type, writer_type))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__121, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcbfa7ba, 0xf8f7277, 0xca3fdae):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcbfa7ba, 0xf8f7277, 0xca3fdae) = (extra, reader_type, writer_type))" % __pyx_checksum
 *     __pyx_result = CannotPromote.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xcbfa7ba, 0xf8f7277, 0xca3fdae):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcbfa7ba, 0xf8f7277, 0xca3fdae) = (extra, reader_type, writer_type))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = CannotPromote.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_28, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xcbfa7ba, 0xf8f7277, 0xca3fdae):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcbfa7ba, 0xf8f7277, 0xca3fdae) = (extra, reader_type, writer_type))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcbfa7ba, 0xf8f7277, 0xca3fdae) = (extra, reader_type, writer_type))" % __pyx_checksum
 *     __pyx_result = CannotPromote.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CannotPromote__set_state(<CannotPromote> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5cavro_CannotPromote), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcbfa7ba, 0xf8f7277, 0xca3fdae) = (extra, reader_type, writer_type))" % __pyx_checksum
 *     __pyx_result = CannotPromote.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CannotPromote__set_state(<CannotPromote> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = CannotPromote.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CannotPromote__set_state(<CannotPromote> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CannotPromote__set_state(CannotPromote __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_5cavro___pyx_unpickle_CannotPromote__set_state(((struct __pyx_obj_5cavro_CannotPromote *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xcbfa7ba, 0xf8f7277, 0xca3fdae) = (extra, reader_type, writer_type))" % __pyx_checksum
 *     __pyx_result = CannotPromote.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CannotPromote__set_state(<CannotPromote> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CannotPromote__set_state(<CannotPromote> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CannotPromote__set_state(CannotPromote __pyx_result, tuple __pyx_state):
 *     __pyx_result.extra = __pyx_state[0]; __pyx_result.reader_type = __pyx_state[1]; __pyx_result.writer_type = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CannotPromote(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_CannotPromote", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CannotPromote__set_state(<CannotPromote> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CannotPromote__set_state(CannotPromote __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.extra = __pyx_state[0]; __pyx_result.reader_type = __pyx_state[1]; __pyx_result.writer_type = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_5cavro___pyx_unpickle_CannotPromote__set_state(struct __pyx_obj_5cavro_CannotPromote *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CannotPromote__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CannotPromote__set_state(CannotPromote __pyx_result, tuple __pyx_state):
 *     __pyx_result.extra = __pyx_state[0]; __pyx_result.reader_type = __pyx_state[1]; __pyx_result.writer_type = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->extra);
  __Pyx_DECREF(__pyx_v___pyx_result->extra);
  __pyx_v___pyx_result->extra = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->reader_type);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->reader_type);
  __pyx_v___pyx_result->reader_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5cavro_AvroType))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->writer_type);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->writer_type);
  __pyx_v___pyx_result->writer_type = ((struct __pyx_obj_5cavro_AvroType *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_CannotPromote__set_state(CannotPromote __pyx_result, tuple __pyx_state):
 *     __pyx_result.extra = __pyx_state[0]; __pyx_result.reader_type = __pyx_state[1]; __pyx_result.writer_type = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 3);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.extra = __pyx_state[0]; __pyx_result.reader_type = __pyx_state[1]; __pyx_result.writer_type = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_CannotPromote__set_state(CannotPromote __pyx_result, tuple __pyx_state):
 *     __pyx_result.extra = __pyx_state[0]; __pyx_result.reader_type = __pyx_state[1]; __pyx_result.writer_type = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_CannotPromote__set_state(<CannotPromote> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CannotPromote__set_state(CannotPromote __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.extra = __pyx_state[0]; __pyx_result.reader_type = __pyx_state[1]; __pyx_result.writer_type = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("cavro.__pyx_unpickle_CannotPromote__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_5cavro_Options __pyx_vtable_5cavro_Options;

static PyObject *__pyx_tp_new_5cavro_Options(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5cavro_Options *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5cavro_Options *)o);
  p->__pyx_vtab = __pyx_vtabptr_5cavro_Options;
  p->union_decodes_to = Py_None; Py_INCREF(Py_None);
  p->bytes_codec = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->fixed_codec = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->logical_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->unicode_errors = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->externally_defined_types = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_Options(PyObject *o) {
  struct __pyx_obj_5cavro_Options *p = (struct __pyx_obj_5cavro_Options *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_Options) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->union_decodes_to);
  Py_CLEAR(p->bytes_codec);
  Py_CLEAR(p->fixed_codec);
  Py_CLEAR(p->logical_types);
  Py_CLEAR(p->unicode_errors);
  Py_CLEAR(p->externally_defined_types);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_5cavro_Options(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_Options *p = (struct __pyx_obj_5cavro_Options *)o;
  if (p->union_decodes_to) {
    e = (*v)(p->union_decodes_to, a); if (e) return e;
  }
  if (p->logical_types) {
    e = (*v)(p->logical_types, a); if (e) return e;
  }
  if (p->externally_defined_types) {
    e = (*v)(p->externally_defined_types, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_Options(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_Options *p = (struct __pyx_obj_5cavro_Options *)o;
  tmp = ((PyObject*)p->union_decodes_to);
  p->union_decodes_to = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->logical_types);
  p->logical_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->externally_defined_types);
  p->externally_defined_types = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_tp_richcompare_5cavro_Options(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_5cavro_7Options_15__eq__(o1, o2);
    }
    case Py_NE: {
      PyObject *ret;
      ret = __pyx_pw_5cavro_7Options_15__eq__(o1, o2);
      if (likely(ret && ret != Py_NotImplemented)) {
        int b = __Pyx_PyObject_IsTrue(ret);
        Py_DECREF(ret);
        if (unlikely(b < 0)) return NULL;
        ret = (b) ? Py_False : Py_True;
        Py_INCREF(ret);
      }
      return ret;
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_5cavro_7Options_name_pattern(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_12name_pattern_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_fingerprint_returns_digest(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_26fingerprint_returns_digest_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_canonical_form_repeat_fixed(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_27canonical_form_repeat_fixed_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_canonical_form_repeat_enum(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_26canonical_form_repeat_enum_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_record_can_encode_dict(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_22record_can_encode_dict_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_record_values_type_hint(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_23record_values_type_hint_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_record_decodes_to_dict(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_22record_decodes_to_dict_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_record_allow_extra_fields(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_25record_allow_extra_fields_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_record_encode_use_defaults(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_26record_encode_use_defaults_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_allow_tuple_notation(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_20allow_tuple_notation_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_union_decodes_to(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_16union_decodes_to_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_union_json_encodes_type_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_28union_json_encodes_type_name_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_allow_primitive_name_collision(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_30allow_primitive_name_collision_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_allow_primitive_names_in_namespaces(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_35allow_primitive_names_in_namespaces_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_named_type_names_must_be_unique(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_31named_type_names_must_be_unique_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_enum_symbols_must_be_unique(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_27enum_symbols_must_be_unique_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_enforce_enum_symbol_name_rules(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_30enforce_enum_symbol_name_rules_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_enforce_type_name_rules(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_23enforce_type_name_rules_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_enforce_namespace_name_rules(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_28enforce_namespace_name_rules_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_record_fields_must_be_unique(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_28record_fields_must_be_unique_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_ascii_name_rules(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_16ascii_name_rules_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_missing_values_can_be_null(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_26missing_values_can_be_null_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_missing_values_can_be_empty_container(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_37missing_values_can_be_empty_container_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_allow_false_values_for_null(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_27allow_false_values_for_null_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_allow_empty_unions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_18allow_empty_unions_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_allow_nested_unions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_19allow_nested_unions_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_allow_duplicate_union_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_27allow_duplicate_union_types_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_allow_union_default_any_member(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_30allow_union_default_any_member_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_allow_aliases_to_be_string(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_26allow_aliases_to_be_string_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_coerce_values_to_int(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_20coerce_values_to_int_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_coerce_values_to_float(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_22coerce_values_to_float_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_coerce_int_to_float(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_19coerce_int_to_float_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_coerce_values_to_boolean(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_24coerce_values_to_boolean_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_coerce_values_to_str(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_20coerce_values_to_str_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_bytes_codec(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_11bytes_codec_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_fixed_codec(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_11fixed_codec_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_null_pad_fixed(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_14null_pad_fixed_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_truncate_fixed(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_14truncate_fixed_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_clamp_int_overflow(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_18clamp_int_overflow_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_clamp_float_overflow(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_20clamp_float_overflow_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_float_out_of_range_inf(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_22float_out_of_range_inf_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_bytes_default_value_utf8(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_24bytes_default_value_utf8_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_string_types_default_unchanged(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_30string_types_default_unchanged_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_decimal_check_exp_overflow(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_26decimal_check_exp_overflow_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_types_str_to_schema(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_19types_str_to_schema_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_logical_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_13logical_types_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_adapt_record_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_18adapt_record_types_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_return_uuid_object(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_18return_uuid_object_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_allow_error_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_16allow_error_type_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_allow_leading_dot_in_names(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_26allow_leading_dot_in_names_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_naive_dt_assume_utc(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_19naive_dt_assume_utc_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_alternate_timestamp_millis_encoding(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_35alternate_timestamp_millis_encoding_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_date_type_accepts_string(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_24date_type_accepts_string_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_raise_on_invalid_logical(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_24raise_on_invalid_logical_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_inline_namespaces(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_17inline_namespaces_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_expand_types_in_schema(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_22expand_types_in_schema_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_unicode_errors(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_14unicode_errors_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_container_fill_blocks(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_21container_fill_blocks_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_defer_schema_promotion_errors(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_29defer_schema_promotion_errors_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_invalid_value_includes_record_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_34invalid_value_includes_record_name_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_invalid_value_include_array_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_33invalid_value_include_array_index_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_invalid_value_include_map_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_29invalid_value_include_map_key_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_allow_invalid_default_values(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_28allow_invalid_default_values_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_7Options_externally_defined_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7Options_24externally_defined_types_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_5cavro_7Options_13__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_5cavro_7Options_13__repr__(self);
}

static PyMethodDef __pyx_methods_5cavro_Options[] = {
  {"replace", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7Options_1replace, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_7Options_replace},
  {"with_logical_types", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5cavro_7Options_3with_logical_types, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5cavro_7Options_2with_logical_types},
  {"with_external_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7Options_5with_external_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_7Options_4with_external_types},
  {"equals", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7Options_7equals, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_7Options_6equals},
  {"diff", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7Options_9diff, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_7Options_8diff},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_5cavro_7Options_13__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7Options_19__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7Options_21__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_Options[] = {
  {(char *)"name_pattern", __pyx_getprop_5cavro_7Options_name_pattern, 0, (char *)PyDoc_STR("\n        Returns the relevant regular expression for validating names based on the current options.\n        "), 0},
  {(char *)"fingerprint_returns_digest", __pyx_getprop_5cavro_7Options_fingerprint_returns_digest, 0, (char *)0, 0},
  {(char *)"canonical_form_repeat_fixed", __pyx_getprop_5cavro_7Options_canonical_form_repeat_fixed, 0, (char *)0, 0},
  {(char *)"canonical_form_repeat_enum", __pyx_getprop_5cavro_7Options_canonical_form_repeat_enum, 0, (char *)0, 0},
  {(char *)"record_can_encode_dict", __pyx_getprop_5cavro_7Options_record_can_encode_dict, 0, (char *)0, 0},
  {(char *)"record_values_type_hint", __pyx_getprop_5cavro_7Options_record_values_type_hint, 0, (char *)0, 0},
  {(char *)"record_decodes_to_dict", __pyx_getprop_5cavro_7Options_record_decodes_to_dict, 0, (char *)0, 0},
  {(char *)"record_allow_extra_fields", __pyx_getprop_5cavro_7Options_record_allow_extra_fields, 0, (char *)0, 0},
  {(char *)"record_encode_use_defaults", __pyx_getprop_5cavro_7Options_record_encode_use_defaults, 0, (char *)0, 0},
  {(char *)"allow_tuple_notation", __pyx_getprop_5cavro_7Options_allow_tuple_notation, 0, (char *)0, 0},
  {(char *)"union_decodes_to", __pyx_getprop_5cavro_7Options_union_decodes_to, 0, (char *)0, 0},
  {(char *)"union_json_encodes_type_name", __pyx_getprop_5cavro_7Options_union_json_encodes_type_name, 0, (char *)0, 0},
  {(char *)"allow_primitive_name_collision", __pyx_getprop_5cavro_7Options_allow_primitive_name_collision, 0, (char *)0, 0},
  {(char *)"allow_primitive_names_in_namespaces", __pyx_getprop_5cavro_7Options_allow_primitive_names_in_namespaces, 0, (char *)0, 0},
  {(char *)"named_type_names_must_be_unique", __pyx_getprop_5cavro_7Options_named_type_names_must_be_unique, 0, (char *)0, 0},
  {(char *)"enum_symbols_must_be_unique", __pyx_getprop_5cavro_7Options_enum_symbols_must_be_unique, 0, (char *)0, 0},
  {(char *)"enforce_enum_symbol_name_rules", __pyx_getprop_5cavro_7Options_enforce_enum_symbol_name_rules, 0, (char *)0, 0},
  {(char *)"enforce_type_name_rules", __pyx_getprop_5cavro_7Options_enforce_type_name_rules, 0, (char *)0, 0},
  {(char *)"enforce_namespace_name_rules", __pyx_getprop_5cavro_7Options_enforce_namespace_name_rules, 0, (char *)0, 0},
  {(char *)"record_fields_must_be_unique", __pyx_getprop_5cavro_7Options_record_fields_must_be_unique, 0, (char *)0, 0},
  {(char *)"ascii_name_rules", __pyx_getprop_5cavro_7Options_ascii_name_rules, 0, (char *)0, 0},
  {(char *)"missing_values_can_be_null", __pyx_getprop_5cavro_7Options_missing_values_can_be_null, 0, (char *)0, 0},
  {(char *)"missing_values_can_be_empty_container", __pyx_getprop_5cavro_7Options_missing_values_can_be_empty_container, 0, (char *)0, 0},
  {(char *)"allow_false_values_for_null", __pyx_getprop_5cavro_7Options_allow_false_values_for_null, 0, (char *)0, 0},
  {(char *)"allow_empty_unions", __pyx_getprop_5cavro_7Options_allow_empty_unions, 0, (char *)0, 0},
  {(char *)"allow_nested_unions", __pyx_getprop_5cavro_7Options_allow_nested_unions, 0, (char *)0, 0},
  {(char *)"allow_duplicate_union_types", __pyx_getprop_5cavro_7Options_allow_duplicate_union_types, 0, (char *)0, 0},
  {(char *)"allow_union_default_any_member", __pyx_getprop_5cavro_7Options_allow_union_default_any_member, 0, (char *)0, 0},
  {(char *)"allow_aliases_to_be_string", __pyx_getprop_5cavro_7Options_allow_aliases_to_be_string, 0, (char *)0, 0},
  {(char *)"coerce_values_to_int", __pyx_getprop_5cavro_7Options_coerce_values_to_int, 0, (char *)0, 0},
  {(char *)"coerce_values_to_float", __pyx_getprop_5cavro_7Options_coerce_values_to_float, 0, (char *)0, 0},
  {(char *)"coerce_int_to_float", __pyx_getprop_5cavro_7Options_coerce_int_to_float, 0, (char *)0, 0},
  {(char *)"coerce_values_to_boolean", __pyx_getprop_5cavro_7Options_coerce_values_to_boolean, 0, (char *)0, 0},
  {(char *)"coerce_values_to_str", __pyx_getprop_5cavro_7Options_coerce_values_to_str, 0, (char *)0, 0},
  {(char *)"bytes_codec", __pyx_getprop_5cavro_7Options_bytes_codec, 0, (char *)0, 0},
  {(char *)"fixed_codec", __pyx_getprop_5cavro_7Options_fixed_codec, 0, (char *)0, 0},
  {(char *)"null_pad_fixed", __pyx_getprop_5cavro_7Options_null_pad_fixed, 0, (char *)0, 0},
  {(char *)"truncate_fixed", __pyx_getprop_5cavro_7Options_truncate_fixed, 0, (char *)0, 0},
  {(char *)"clamp_int_overflow", __pyx_getprop_5cavro_7Options_clamp_int_overflow, 0, (char *)0, 0},
  {(char *)"clamp_float_overflow", __pyx_getprop_5cavro_7Options_clamp_float_overflow, 0, (char *)0, 0},
  {(char *)"float_out_of_range_inf", __pyx_getprop_5cavro_7Options_float_out_of_range_inf, 0, (char *)0, 0},
  {(char *)"bytes_default_value_utf8", __pyx_getprop_5cavro_7Options_bytes_default_value_utf8, 0, (char *)0, 0},
  {(char *)"string_types_default_unchanged", __pyx_getprop_5cavro_7Options_string_types_default_unchanged, 0, (char *)0, 0},
  {(char *)"decimal_check_exp_overflow", __pyx_getprop_5cavro_7Options_decimal_check_exp_overflow, 0, (char *)0, 0},
  {(char *)"types_str_to_schema", __pyx_getprop_5cavro_7Options_types_str_to_schema, 0, (char *)0, 0},
  {(char *)"logical_types", __pyx_getprop_5cavro_7Options_logical_types, 0, (char *)0, 0},
  {(char *)"adapt_record_types", __pyx_getprop_5cavro_7Options_adapt_record_types, 0, (char *)0, 0},
  {(char *)"return_uuid_object", __pyx_getprop_5cavro_7Options_return_uuid_object, 0, (char *)0, 0},
  {(char *)"allow_error_type", __pyx_getprop_5cavro_7Options_allow_error_type, 0, (char *)0, 0},
  {(char *)"allow_leading_dot_in_names", __pyx_getprop_5cavro_7Options_allow_leading_dot_in_names, 0, (char *)0, 0},
  {(char *)"naive_dt_assume_utc", __pyx_getprop_5cavro_7Options_naive_dt_assume_utc, 0, (char *)0, 0},
  {(char *)"alternate_timestamp_millis_encoding", __pyx_getprop_5cavro_7Options_alternate_timestamp_millis_encoding, 0, (char *)0, 0},
  {(char *)"date_type_accepts_string", __pyx_getprop_5cavro_7Options_date_type_accepts_string, 0, (char *)0, 0},
  {(char *)"raise_on_invalid_logical", __pyx_getprop_5cavro_7Options_raise_on_invalid_logical, 0, (char *)0, 0},
  {(char *)"inline_namespaces", __pyx_getprop_5cavro_7Options_inline_namespaces, 0, (char *)0, 0},
  {(char *)"expand_types_in_schema", __pyx_getprop_5cavro_7Options_expand_types_in_schema, 0, (char *)0, 0},
  {(char *)"unicode_errors", __pyx_getprop_5cavro_7Options_unicode_errors, 0, (char *)0, 0},
  {(char *)"container_fill_blocks", __pyx_getprop_5cavro_7Options_container_fill_blocks, 0, (char *)0, 0},
  {(char *)"defer_schema_promotion_errors", __pyx_getprop_5cavro_7Options_defer_schema_promotion_errors, 0, (char *)0, 0},
  {(char *)"invalid_value_includes_record_name", __pyx_getprop_5cavro_7Options_invalid_value_includes_record_name, 0, (char *)0, 0},
  {(char *)"invalid_value_include_array_index", __pyx_getprop_5cavro_7Options_invalid_value_include_array_index, 0, (char *)0, 0},
  {(char *)"invalid_value_include_map_key", __pyx_getprop_5cavro_7Options_invalid_value_include_map_key, 0, (char *)0, 0},
  {(char *)"allow_invalid_default_values", __pyx_getprop_5cavro_7Options_allow_invalid_default_values, 0, (char *)0, 0},
  {(char *)"externally_defined_types", __pyx_getprop_5cavro_7Options_externally_defined_types, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_Options_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_Options},
  {Py_tp_repr, (void *)__pyx_pw_5cavro_7Options_13__repr__},
  {Py_tp_hash, (void *)__pyx_pw_5cavro_7Options_17__hash__},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Runtime configuration options for controlling the behaviour\n    of a schema.\n    Instances of `Options` are immutable, create a modified copy of options, use the `replace` and `with_*` methods.\n\n    Attributes:\n\n    * `fingerprint_returns_digest` If `True`, the `Schema.fingerprint` method returns a hashlib hash object, rather than the digest bytes\n    * `canonical_form_repeat_fixed` Some libraries repeat enum type definitions with the same name/size in canonical form.  Setting this flag to True replicates that\n    * `canonical_form_repeat_enum` Some libraries repeat enum type definitions with the same name/symbols in canonical form.  Setting this flag to True replicates that\n    * `record_can_encode_dict` If `True`, dicts can be encoded as records (provided they have the correct fields).  If false, then an instance of the relevant record type must be used.\n    * `record_values_type_hint`\n        If `True`, then dicts encoded using a record schema can have an optional key `-type` (note the leading '-')\n        with a value that is the name of the record, ensuring the correct record schema is chosen.\n    * `record_decodes_to_dict` If `True`, then records are decoded to a dict, rather than a record class instance\n    * `record_allow_extra_fields` If `True`, then any fields in a dict that are not in the record schema are ignored. If `False`, then an error is raised.\n    * `record_fields_must_be_unique` If `True`, then all fields within a record must have a unique name\n    * `record_encode_use_defaults` \n        If `True`, then when encoding a dict as a record, any fields that are not present in the dict are encoded using their default value from the schema.\n        If `False`, then an error is raised for any missing keys\n    * `missing_values_can_be_null` If `True`, then missing values may be encoded as `null` where this is valid in the schema. (i.e. `['null', 'string']`)\n    * `missing_values_can_be_empty_container` If `True`, then missing val""ues may be encoded as an empty container (list, dict) where this is valid in the schema.\n    * `allow_tuple_notation` If `True`, then values can be encoded by passing a 2-tuple of (<type>, <value>) where <type> is the name of the type to encode as, and <value> is the value to encode.\n    * `union_decodes_to` Controls how union values are decoded.  See `UnionDecodeOption` for details.  The default, `UnionDecodeOption.RAW_VALUES` returns the value from the matching union unmodified.\n    * `union_json_encodes_type_name` \n        If `True` (default), then when JSON encoding a value in a union, the type name is included in the output as per spec.\n        If `False`, then the JSON-encoded value of the matching union type is output directly.\n    * `allow_primitive_name_collision` If `True`, then named types can have the same name as one of the primitive types (e.g. `int`, `float`, `str` etc..)\n    * `allow_primitive_names_in_namespaces` If `True`, then namespace parts can have the same name as one of the primitive types (e.g. `int`, `float`, `str` etc..)\n    * `named_type_names_must_be_unique` If `True`, then all named types must have a unique name within the schema\n    * `enum_symbols_must_be_unique` If `True`, then all symbols within an enum must be unique\n    * `enforce_enum_symbol_name_rules` If `True`, then enum symbol values are checked to ensure they match the rules for valid symbols\n    * `enforce_type_name_rules` If `True`, then type names are checked to ensure they match the rules for valid names\n    * `enforce_namespace_name_rules` If `True`, then namespace names are checked to ensure they match the rules for valid names\n    * `ascii_name_rules` \n        If `True`, then name checking (names/symbols) is done using the strict ASCII rules in the spec.\n        If `False`, then equivalent unicode rules are used.\n    * `allow_false_values_for_null` If `True`, then a null type will accept any value for which `bool(value)` is `False`, otherwise an error ""is raised for any non-`None` value.\n    * `allow_empty_unions` If `True`, then unions can be empty (i.e. `[]`), otherwise an error is raised.\n    * `allow_nested_unions` If `True`, then unions can contain other unions, otherwise an error is raised.\n    * `allow_duplicate_union_types` If `True`, then unions can contain multiple types that are forbidden from being in the same union by the spec.\n    * `allow_union_default_any_member` If `True`, then the default value for a union can match the schema of any member of the union, otherwise the default value must match the first member of the union.\n    * `allow_aliases_to_be_string` \n        If `True`, then aliases can be specified as a string or a list\n        If `False`, then aliases must be specified as a list\n    * `coerce_values_to_int`\n        If `True`, then values encoded in `int` or `long` schemas will be coerced to `int` where possible\n    * `coerce_values_to_float`\n        If `True`, then values encoded in `float` or `double` schemas will be coerced to `float` where possible\n    * `coerce_int_to_float`\n        If `True`, then `float` and `double` types will accept `int` values, and encode them as the closest floating-point value\n    * `coerce_values_to_boolean`\n        If `True`, then values encoded in `boolean` schemas will be coerced to `bool` where possible\n    * `coerce_values_to_str`\n        If `True`, then values encoded in `string` schemas will be coerced to `str` where possible\n    * `bytes_codec` If not `None`, then values passed to bytes schemas will be encoded using the specified codec\n    * `fixed_codec` If not `None`, then values passed to fixed schemas will be encoded using the specified codec\n    * `null_pad_fixed` If `True`, then values encoded in `fixed` schemas will be zero-padded to the specified size\n    * `truncate_fixed` If `True`, then values encoded in `fixed` schemas will be truncated to the specified size (extra bytes will be dropped on encoding)\n    * `clamp_int_""overflow` If `True`, then values encoded in `int` schemas will be clamped to the min/max values for the specified size\n    * `clamp_float_overflow` If `True`, then values encoded in the `float` schema will be clamped to the min/max values 32-bit floats.  (Note, this also replaces INF/NaN values with the closest representable value unless `float_out_of_range_inf` is `True)\n    * `float_out_of_range_inf` If `True`, then values encoded in the `float` schema that are out of range will be encoded as INF/NaN, otherwise an error is raised.\n    * `bytes_default_value_utf8` If `True`, then the default value for a bytes schema is encoded as UTF-8, otherwise it is encoded as latin1, this should probably never be used except for library compatibility reasons\n    * `string_types_default_unchanged` If `True`, then default values for string/bytes/fixed schemas are passed back unmodified (may not be a string), otherwise the relevant JSON decoding is performed. this should probably never be used except for library compatibility reasons\n    * `decimal_check_exp_overflow` If `True`, then values encoded in the `decimal` schema will be checked to ensure they are within the range of the specified precision/scale, otherwise an error is raised.\n    * `types_str_to_schema` If `True`, then calling `str(<schema instance>)` returns a JSON representation of the schema, otherwise the default `str` implementation is used.\n    * `logical_types`\n        A tuple of logical types to use when parsing schemas, each item must be an instance of `LogicalType`.  \n        Typically this means that for custom logical types, you should subclass `cavro.CustomLogicalType`\n        and implement `def custom_encode_value(value)` and `def custom_decode_value(value)`.\n        To add types to this tuple, use the `with_logical_types` method.\n    * `adapt_record_types` \n        If `True`, then when encoding records, if a record type is passed that has come from a different schema,\n        then the record ""type is adapted to the current schema, provided the name and fields match.\n        This situation can easily occur if a source schema is parsed twice.  Each instance of the schema will\n        have its own version of the Record class, and these are not compatible without this option.\n    * `return_uuid_object` If `True`, then UUID values are returned as `uuid.UUID` objects, otherwise they are returned as strings.\n    * `allow_error_type` If `True`, then the `error` type is allowed in schemas (As an alias for 'record').\n    * `allow_leading_dot_in_names` If `True`, then names can start with a dot, indicating the null namespace, otherwise an error is raised.\n    * `naive_dt_assume_utc` If `True`, then naive datetime values are assumed to be in UTC, otherwise they are treated as representing local time (using current locale)\n    * `alternate_timestamp_millis_encoding` If `True`, then an alternate approach to encoding timestamps is used that has slightly different behaviour at extreme values.\n    * `date_type_accepts_string` If `True`, then the `date` logical type will accept string in ISO8601 format as input (YYYY-MM-DD), otherwise an error is raised.\n    * `raise_on_invalid_logical` If `True`, then attempts to parse a schema that contain invalid logical decimal paramters, will raise an error, rather than silently ignoring the logical type (as per spec)\n    * `inline_namespaces` If `True`, then when outputting schema JSON from a parsed schema, namespaces are inlined into the name, otherwise the namespace is output as a separate field.\n    * `expand_types_in_schema` \n        If `True`, then when outputting schema JSON from a parsed schema, repeated types are output in full, rather than being referenced by name.\n        **Note**: This does not apply to nested/recursive types, which are always referenced by name to prevent infinite recursion.\n    * `unicode_errors` The error handling strategy to use when decoding strings/bytes.  See the `errors` parameter to"" `str.decode` for details.\n    * `container_fill_blocks`\n        If `True`, then when writing a container, records are written until the current block is > `max_blocksize` (I.e. blocks will often be larger than `max_blocksize`)\n        By default, container writer will write a new block whenever the next record will take the current block over the `max_blocksize`,\n        meaning that blocks will always be <= `max_blocksize` unless a single value is larger than `max_blocksize`\n    * `defer_schema_promotion_errors`\n        `cavro` performs eager promotion calculation for performance reasons, this means that incompatible reader/writer schemas are typically detected at schema parse time, and errors raised.\n        For compatibility purposes, setting this option to `True` results in schema promotion errors being stored and raised when the first value is read.\n    * `invalid_value_includes_record_name`\n        When raising exceptions based on Invalid Values, the exception path includes the name of the record type, rather than just the field name.\n    * `invalid_value_include_array_index`\n        When raising exceptions based on Invalid Values, the exception path includes the array index of any arrays\n    * `invalid_value_include_map_key`\n        When raising exceptions based on Invalid Values, the exception path includes the key of any maps in the value\n    * `allow_invalid_default_values`\n        Typically, default values must be valid JSON values for the schema, setting this option to `True` allows invalid default values to be used (JSON decoding is still performed, but decode errors result in the raw value being used).\n    * `externally_defined_types`\n        An immutable dict of named types (instances of `AvroType`) that are defined outside of the schema being parsed.  This allows for custom/complex schema loading patterns where type definitions may be spread across multiple locations.\n        For example, if this dict has `{'Foo': <RecordType...>}`"", then a schema: `{\"type\": \"Foo\"}` will be parsed to be the passed-in Foo type.\n        To add types to this dict, use the `with_external_types` method.\n        **Note**: It's possible to end up with some weird situations including infinite recursion when using this option, as it's possible to create reference cycles between schemas resulting in infinite recursion errors.\n        \n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_Options},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_Options},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_5cavro_Options},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_Options},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_Options},
  {Py_tp_init, (void *)__pyx_pw_5cavro_7Options_11__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_Options},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_Options_spec = {
  "cavro.Options",
  sizeof(struct __pyx_obj_5cavro_Options),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_Options_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_Options = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""Options", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_Options), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_Options, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_5cavro_7Options_13__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_5cavro_7Options_17__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Runtime configuration options for controlling the behaviour\n    of a schema.\n    Instances of `Options` are immutable, create a modified copy of options, use the `replace` and `with_*` methods.\n\n    Attributes:\n\n    * `fingerprint_returns_digest` If `True`, the `Schema.fingerprint` method returns a hashlib hash object, rather than the digest bytes\n    * `canonical_form_repeat_fixed` Some libraries repeat enum type definitions with the same name/size in canonical form.  Setting this flag to True replicates that\n    * `canonical_form_repeat_enum` Some libraries repeat enum type definitions with the same name/symbols in canonical form.  Setting this flag to True replicates that\n    * `record_can_encode_dict` If `True`, dicts can be encoded as records (provided they have the correct fields).  If false, then an instance of the relevant record type must be used.\n    * `record_values_type_hint`\n        If `True`, then dicts encoded using a record schema can have an optional key `-type` (note the leading '-')\n        with a value that is the name of the record, ensuring the correct record schema is chosen.\n    * `record_decodes_to_dict` If `True`, then records are decoded to a dict, rather than a record class instance\n    * `record_allow_extra_fields` If `True`, then any fields in a dict that are not in the record schema are ignored. If `False`, then an error is raised.\n    * `record_fields_must_be_unique` If `True`, then all fields within a record must have a unique name\n    * `record_encode_use_defaults` \n        If `True`, then when encoding a dict as a record, any fields that are not present in the dict are encoded using their default value from the schema.\n        If `False`, then an error is raised for any missing keys\n    * `missing_values_can_be_null` If `True`, then missing values may be encoded as `null` where this is valid in the schema. (i.e. `['null', 'string']`)\n    * `missing_values_can_be_empty_container` If `True`, then missing val""ues may be encoded as an empty container (list, dict) where this is valid in the schema.\n    * `allow_tuple_notation` If `True`, then values can be encoded by passing a 2-tuple of (<type>, <value>) where <type> is the name of the type to encode as, and <value> is the value to encode.\n    * `union_decodes_to` Controls how union values are decoded.  See `UnionDecodeOption` for details.  The default, `UnionDecodeOption.RAW_VALUES` returns the value from the matching union unmodified.\n    * `union_json_encodes_type_name` \n        If `True` (default), then when JSON encoding a value in a union, the type name is included in the output as per spec.\n        If `False`, then the JSON-encoded value of the matching union type is output directly.\n    * `allow_primitive_name_collision` If `True`, then named types can have the same name as one of the primitive types (e.g. `int`, `float`, `str` etc..)\n    * `allow_primitive_names_in_namespaces` If `True`, then namespace parts can have the same name as one of the primitive types (e.g. `int`, `float`, `str` etc..)\n    * `named_type_names_must_be_unique` If `True`, then all named types must have a unique name within the schema\n    * `enum_symbols_must_be_unique` If `True`, then all symbols within an enum must be unique\n    * `enforce_enum_symbol_name_rules` If `True`, then enum symbol values are checked to ensure they match the rules for valid symbols\n    * `enforce_type_name_rules` If `True`, then type names are checked to ensure they match the rules for valid names\n    * `enforce_namespace_name_rules` If `True`, then namespace names are checked to ensure they match the rules for valid names\n    * `ascii_name_rules` \n        If `True`, then name checking (names/symbols) is done using the strict ASCII rules in the spec.\n        If `False`, then equivalent unicode rules are used.\n    * `allow_false_values_for_null` If `True`, then a null type will accept any value for which `bool(value)` is `False`, otherwise an error ""is raised for any non-`None` value.\n    * `allow_empty_unions` If `True`, then unions can be empty (i.e. `[]`), otherwise an error is raised.\n    * `allow_nested_unions` If `True`, then unions can contain other unions, otherwise an error is raised.\n    * `allow_duplicate_union_types` If `True`, then unions can contain multiple types that are forbidden from being in the same union by the spec.\n    * `allow_union_default_any_member` If `True`, then the default value for a union can match the schema of any member of the union, otherwise the default value must match the first member of the union.\n    * `allow_aliases_to_be_string` \n        If `True`, then aliases can be specified as a string or a list\n        If `False`, then aliases must be specified as a list\n    * `coerce_values_to_int`\n        If `True`, then values encoded in `int` or `long` schemas will be coerced to `int` where possible\n    * `coerce_values_to_float`\n        If `True`, then values encoded in `float` or `double` schemas will be coerced to `float` where possible\n    * `coerce_int_to_float`\n        If `True`, then `float` and `double` types will accept `int` values, and encode them as the closest floating-point value\n    * `coerce_values_to_boolean`\n        If `True`, then values encoded in `boolean` schemas will be coerced to `bool` where possible\n    * `coerce_values_to_str`\n        If `True`, then values encoded in `string` schemas will be coerced to `str` where possible\n    * `bytes_codec` If not `None`, then values passed to bytes schemas will be encoded using the specified codec\n    * `fixed_codec` If not `None`, then values passed to fixed schemas will be encoded using the specified codec\n    * `null_pad_fixed` If `True`, then values encoded in `fixed` schemas will be zero-padded to the specified size\n    * `truncate_fixed` If `True`, then values encoded in `fixed` schemas will be truncated to the specified size (extra bytes will be dropped on encoding)\n    * `clamp_int_""overflow` If `True`, then values encoded in `int` schemas will be clamped to the min/max values for the specified size\n    * `clamp_float_overflow` If `True`, then values encoded in the `float` schema will be clamped to the min/max values 32-bit floats.  (Note, this also replaces INF/NaN values with the closest representable value unless `float_out_of_range_inf` is `True)\n    * `float_out_of_range_inf` If `True`, then values encoded in the `float` schema that are out of range will be encoded as INF/NaN, otherwise an error is raised.\n    * `bytes_default_value_utf8` If `True`, then the default value for a bytes schema is encoded as UTF-8, otherwise it is encoded as latin1, this should probably never be used except for library compatibility reasons\n    * `string_types_default_unchanged` If `True`, then default values for string/bytes/fixed schemas are passed back unmodified (may not be a string), otherwise the relevant JSON decoding is performed. this should probably never be used except for library compatibility reasons\n    * `decimal_check_exp_overflow` If `True`, then values encoded in the `decimal` schema will be checked to ensure they are within the range of the specified precision/scale, otherwise an error is raised.\n    * `types_str_to_schema` If `True`, then calling `str(<schema instance>)` returns a JSON representation of the schema, otherwise the default `str` implementation is used.\n    * `logical_types`\n        A tuple of logical types to use when parsing schemas, each item must be an instance of `LogicalType`.  \n        Typically this means that for custom logical types, you should subclass `cavro.CustomLogicalType`\n        and implement `def custom_encode_value(value)` and `def custom_decode_value(value)`.\n        To add types to this tuple, use the `with_logical_types` method.\n    * `adapt_record_types` \n        If `True`, then when encoding records, if a record type is passed that has come from a different schema,\n        then the record ""type is adapted to the current schema, provided the name and fields match.\n        This situation can easily occur if a source schema is parsed twice.  Each instance of the schema will\n        have its own version of the Record class, and these are not compatible without this option.\n    * `return_uuid_object` If `True`, then UUID values are returned as `uuid.UUID` objects, otherwise they are returned as strings.\n    * `allow_error_type` If `True`, then the `error` type is allowed in schemas (As an alias for 'record').\n    * `allow_leading_dot_in_names` If `True`, then names can start with a dot, indicating the null namespace, otherwise an error is raised.\n    * `naive_dt_assume_utc` If `True`, then naive datetime values are assumed to be in UTC, otherwise they are treated as representing local time (using current locale)\n    * `alternate_timestamp_millis_encoding` If `True`, then an alternate approach to encoding timestamps is used that has slightly different behaviour at extreme values.\n    * `date_type_accepts_string` If `True`, then the `date` logical type will accept string in ISO8601 format as input (YYYY-MM-DD), otherwise an error is raised.\n    * `raise_on_invalid_logical` If `True`, then attempts to parse a schema that contain invalid logical decimal paramters, will raise an error, rather than silently ignoring the logical type (as per spec)\n    * `inline_namespaces` If `True`, then when outputting schema JSON from a parsed schema, namespaces are inlined into the name, otherwise the namespace is output as a separate field.\n    * `expand_types_in_schema` \n        If `True`, then when outputting schema JSON from a parsed schema, repeated types are output in full, rather than being referenced by name.\n        **Note**: This does not apply to nested/recursive types, which are always referenced by name to prevent infinite recursion.\n    * `unicode_errors` The error handling strategy to use when decoding strings/bytes.  See the `errors` parameter to"" `str.decode` for details.\n    * `container_fill_blocks`\n        If `True`, then when writing a container, records are written until the current block is > `max_blocksize` (I.e. blocks will often be larger than `max_blocksize`)\n        By default, container writer will write a new block whenever the next record will take the current block over the `max_blocksize`,\n        meaning that blocks will always be <= `max_blocksize` unless a single value is larger than `max_blocksize`\n    * `defer_schema_promotion_errors`\n        `cavro` performs eager promotion calculation for performance reasons, this means that incompatible reader/writer schemas are typically detected at schema parse time, and errors raised.\n        For compatibility purposes, setting this option to `True` results in schema promotion errors being stored and raised when the first value is read.\n    * `invalid_value_includes_record_name`\n        When raising exceptions based on Invalid Values, the exception path includes the name of the record type, rather than just the field name.\n    * `invalid_value_include_array_index`\n        When raising exceptions based on Invalid Values, the exception path includes the array index of any arrays\n    * `invalid_value_include_map_key`\n        When raising exceptions based on Invalid Values, the exception path includes the key of any maps in the value\n    * `allow_invalid_default_values`\n        Typically, default values must be valid JSON values for the schema, setting this option to `True` allows invalid default values to be used (JSON decoding is still performed, but decode errors result in the raw value being used).\n    * `externally_defined_types`\n        An immutable dict of named types (instances of `AvroType`) that are defined outside of the schema being parsed.  This allows for custom/complex schema loading patterns where type definitions may be spread across multiple locations.\n        For example, if this dict has `{'Foo': <RecordType...>}`"", then a schema: `{\"type\": \"Foo\"}` will be parsed to be the passed-in Foo type.\n        To add types to this dict, use the `with_external_types` method.\n        **Note**: It's possible to end up with some weird situations including infinite recursion when using this option, as it's possible to create reference cycles between schemas resulting in infinite recursion errors.\n        \n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_Options, /*tp_traverse*/
  __pyx_tp_clear_5cavro_Options, /*tp_clear*/
  __pyx_tp_richcompare_5cavro_Options, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_Options, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_Options, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_7Options_11__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_Options, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro__Writer __pyx_vtable_5cavro__Writer;

static PyObject *__pyx_tp_new_5cavro__Writer(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5cavro__Writer *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5cavro__Writer *)o);
  p->__pyx_vtab = __pyx_vtabptr_5cavro__Writer;
  return o;
}

static void __pyx_tp_dealloc_5cavro__Writer(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro__Writer) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyMethodDef __pyx_methods_5cavro__Writer[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7_Writer_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7_Writer_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__Writer_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro__Writer},
  {Py_tp_methods, (void *)__pyx_methods_5cavro__Writer},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro__Writer},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__Writer_spec = {
  "cavro._Writer",
  sizeof(struct __pyx_obj_5cavro__Writer),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro__Writer_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__Writer = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_Writer", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__Writer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro__Writer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__Writer, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro__Writer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro__Reader __pyx_vtable_5cavro__Reader;

static PyObject *__pyx_tp_new_5cavro__Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5cavro__Reader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5cavro__Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_5cavro__Reader;
  return o;
}

static void __pyx_tp_dealloc_5cavro__Reader(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro__Reader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyMethodDef __pyx_methods_5cavro__Reader[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7_Reader_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7_Reader_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__Reader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro__Reader},
  {Py_tp_methods, (void *)__pyx_methods_5cavro__Reader},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro__Reader},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__Reader_spec = {
  "cavro._Reader",
  sizeof(struct __pyx_obj_5cavro__Reader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro__Reader_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_Reader", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro__Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__Reader, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro__Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_FileReader __pyx_vtable_5cavro_FileReader;

static PyObject *__pyx_tp_new_5cavro_FileReader(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_FileReader *p;
  PyObject *o = __pyx_tp_new_5cavro__Reader(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_FileReader *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro__Reader*)__pyx_vtabptr_5cavro_FileReader;
  p->file_obj = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_FileReader(PyObject *o) {
  struct __pyx_obj_5cavro_FileReader *p = (struct __pyx_obj_5cavro_FileReader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_FileReader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->file_obj);
  __pyx_tp_dealloc_5cavro__Reader(o);
}

static int __pyx_tp_traverse_5cavro_FileReader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_FileReader *p = (struct __pyx_obj_5cavro_FileReader *)o;
  e = ((likely(__pyx_ptype_5cavro__Reader)) ? ((__pyx_ptype_5cavro__Reader->tp_traverse) ? __pyx_ptype_5cavro__Reader->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_5cavro_FileReader)); if (e) return e;
  if (p->file_obj) {
    e = (*v)(p->file_obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_FileReader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_FileReader *p = (struct __pyx_obj_5cavro_FileReader *)o;
  if (likely(__pyx_ptype_5cavro__Reader)) { if (__pyx_ptype_5cavro__Reader->tp_clear) __pyx_ptype_5cavro__Reader->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_5cavro_FileReader);
  tmp = ((PyObject*)p->file_obj);
  p->file_obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_5cavro_FileReader[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10FileReader_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10FileReader_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_FileReader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_FileReader},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A cavro wrapper for reading data from a file-like object (Object that implements `.read(n)`).\n    \n    This class will not close or seek the underlying file object\n\n    Arguments:\n        `file_obj`: The file-like object to read from.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_FileReader},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_FileReader},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_FileReader},
  {Py_tp_init, (void *)__pyx_pw_5cavro_10FileReader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_FileReader},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_FileReader_spec = {
  "cavro.FileReader",
  sizeof(struct __pyx_obj_5cavro_FileReader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_FileReader_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_FileReader = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""FileReader", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_FileReader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_FileReader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    A cavro wrapper for reading data from a file-like object (Object that implements `.read(n)`).\n    \n    This class will not close or seek the underlying file object\n\n    Arguments:\n        `file_obj`: The file-like object to read from.\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_FileReader, /*tp_traverse*/
  __pyx_tp_clear_5cavro_FileReader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_FileReader, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_10FileReader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_FileReader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_FileWriter __pyx_vtable_5cavro_FileWriter;

static PyObject *__pyx_tp_new_5cavro_FileWriter(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_FileWriter *p;
  PyObject *o = __pyx_tp_new_5cavro__Writer(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_FileWriter *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro__Writer*)__pyx_vtabptr_5cavro_FileWriter;
  p->file_obj = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_FileWriter(PyObject *o) {
  struct __pyx_obj_5cavro_FileWriter *p = (struct __pyx_obj_5cavro_FileWriter *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_FileWriter) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->file_obj);
  __pyx_tp_dealloc_5cavro__Writer(o);
}

static int __pyx_tp_traverse_5cavro_FileWriter(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_FileWriter *p = (struct __pyx_obj_5cavro_FileWriter *)o;
  e = ((likely(__pyx_ptype_5cavro__Writer)) ? ((__pyx_ptype_5cavro__Writer->tp_traverse) ? __pyx_ptype_5cavro__Writer->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_5cavro_FileWriter)); if (e) return e;
  if (p->file_obj) {
    e = (*v)(p->file_obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_FileWriter(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_FileWriter *p = (struct __pyx_obj_5cavro_FileWriter *)o;
  if (likely(__pyx_ptype_5cavro__Writer)) { if (__pyx_ptype_5cavro__Writer->tp_clear) __pyx_ptype_5cavro__Writer->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_5cavro_FileWriter);
  tmp = ((PyObject*)p->file_obj);
  p->file_obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_5cavro_FileWriter[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10FileWriter_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10FileWriter_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_FileWriter_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_FileWriter},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A cavro wrapper for writing data to a file-like object (Object that implements `.write(data)` and `.flush()`).\n    \n    This class will not close or seek the underlying file object\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_FileWriter},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_FileWriter},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_FileWriter},
  {Py_tp_init, (void *)__pyx_pw_5cavro_10FileWriter_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_FileWriter},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_FileWriter_spec = {
  "cavro.FileWriter",
  sizeof(struct __pyx_obj_5cavro_FileWriter),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_FileWriter_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_FileWriter = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""FileWriter", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_FileWriter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_FileWriter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    A cavro wrapper for writing data to a file-like object (Object that implements `.write(data)` and `.flush()`).\n    \n    This class will not close or seek the underlying file object\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_FileWriter, /*tp_traverse*/
  __pyx_tp_clear_5cavro_FileWriter, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_FileWriter, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_10FileWriter_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_FileWriter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_MemoryWriter __pyx_vtable_5cavro_MemoryWriter;

static PyObject *__pyx_tp_new_5cavro_MemoryWriter(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_MemoryWriter *p;
  PyObject *o = __pyx_tp_new_5cavro__Writer(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_MemoryWriter *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro__Writer*)__pyx_vtabptr_5cavro_MemoryWriter;
  p->buffer = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_5cavro_12MemoryWriter_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_5cavro_MemoryWriter(PyObject *o) {
  struct __pyx_obj_5cavro_MemoryWriter *p = (struct __pyx_obj_5cavro_MemoryWriter *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->buffer);
  __pyx_tp_dealloc_5cavro__Writer(o);
}

static int __pyx_tp_traverse_5cavro_MemoryWriter(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_MemoryWriter *p = (struct __pyx_obj_5cavro_MemoryWriter *)o;
  e = ((likely(__pyx_ptype_5cavro__Writer)) ? ((__pyx_ptype_5cavro__Writer->tp_traverse) ? __pyx_ptype_5cavro__Writer->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_5cavro_MemoryWriter)); if (e) return e;
  if (p->buffer) {
    e = (*v)(((PyObject *)p->buffer), a); if (e) return e;
  }
  return 0;
}

static PyObject *__pyx_getprop_5cavro_12MemoryWriter_buffer(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_12MemoryWriter_6buffer_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_12MemoryWriter_len(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_12MemoryWriter_3len_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_MemoryWriter[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12MemoryWriter_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12MemoryWriter_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_MemoryWriter[] = {
  {(char *)"buffer", __pyx_getprop_5cavro_12MemoryWriter_buffer, 0, (char *)0, 0},
  {(char *)"len", __pyx_getprop_5cavro_12MemoryWriter_len, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_MemoryWriter_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_MemoryWriter},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A class that writes to a memory buffer. The buffer is automatically resized to fit the data.\n\n    The underlying data is accessible through the `buffer` attribute as an `array.array` of bytes.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_MemoryWriter},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_MemoryWriter},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_MemoryWriter},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_MemoryWriter},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_MemoryWriter_spec = {
  "cavro.MemoryWriter",
  sizeof(struct __pyx_obj_5cavro_MemoryWriter),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_MemoryWriter_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_MemoryWriter = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""MemoryWriter", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_MemoryWriter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_MemoryWriter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    A class that writes to a memory buffer. The buffer is automatically resized to fit the data.\n\n    The underlying data is accessible through the `buffer` attribute as an `array.array` of bytes.\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_MemoryWriter, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_MemoryWriter, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_MemoryWriter, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_MemoryWriter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_MemoryReader __pyx_vtable_5cavro_MemoryReader;

static PyObject *__pyx_tp_new_5cavro_MemoryReader(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_MemoryReader *p;
  PyObject *o = __pyx_tp_new_5cavro__Reader(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_MemoryReader *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro__Reader*)__pyx_vtabptr_5cavro_MemoryReader;
  p->data.data = NULL;
  p->data.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_5cavro_MemoryReader(PyObject *o) {
  struct __pyx_obj_5cavro_MemoryReader *p = (struct __pyx_obj_5cavro_MemoryReader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_MemoryReader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  __PYX_XCLEAR_MEMVIEW(&p->data, 1);
  p->data.memview = NULL; p->data.data = NULL;
  __pyx_tp_dealloc_5cavro__Reader(o);
}

static PyMethodDef __pyx_methods_5cavro_MemoryReader[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12MemoryReader_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12MemoryReader_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_MemoryReader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_MemoryReader},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A class that allows cavro to read binary data from a memory buffer.\n\n    Arguments:\n     * data: The data to read from. Can be a `bytes` object or a `memoryview`.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_MemoryReader},
  {Py_tp_init, (void *)__pyx_pw_5cavro_12MemoryReader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_MemoryReader},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_MemoryReader_spec = {
  "cavro.MemoryReader",
  sizeof(struct __pyx_obj_5cavro_MemoryReader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_MemoryReader_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_MemoryReader = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""MemoryReader", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_MemoryReader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_MemoryReader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    A class that allows cavro to read binary data from a memory buffer.\n\n    Arguments:\n     * data: The data to read from. Can be a `bytes` object or a `memoryview`.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_MemoryReader, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_12MemoryReader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_MemoryReader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_Rabin __pyx_vtable_5cavro_Rabin;

static PyObject *__pyx_tp_new_5cavro_Rabin(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5cavro_Rabin *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5cavro_Rabin *)o);
  p->__pyx_vtab = __pyx_vtabptr_5cavro_Rabin;
  return o;
}

static void __pyx_tp_dealloc_5cavro_Rabin(PyObject *o) {
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_getprop_5cavro_5Rabin_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_5Rabin_5value_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_Rabin[] = {
  {"digest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Rabin_5digest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"hexdigest", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Rabin_7hexdigest, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Rabin_9copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Rabin_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Rabin_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_Rabin[] = {
  {(char *)"value", __pyx_getprop_5cavro_5Rabin_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_Rabin_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_Rabin},
  {Py_tp_doc, (void *)PyDoc_STR("\n    An implementation of the 64-bit Rabin hash function  as described in the avro specification.\n\n    The interface in this class roughly approximates the `hashlib.hash` objects.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_Rabin},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_Rabin},
  {Py_tp_init, (void *)__pyx_pw_5cavro_5Rabin_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_Rabin},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_Rabin_spec = {
  "cavro.Rabin",
  sizeof(struct __pyx_obj_5cavro_Rabin),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_type_5cavro_Rabin_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_Rabin = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""Rabin", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_Rabin), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_Rabin, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  PyDoc_STR("\n    An implementation of the 64-bit Rabin hash function  as described in the avro specification.\n\n    The interface in this class roughly approximates the `hashlib.hash` objects.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_Rabin, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_Rabin, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_5Rabin_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_Rabin, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_Record __pyx_vtable_5cavro_Record;

static struct __pyx_obj_5cavro_Record *__pyx_freelist_5cavro_Record[8];
static int __pyx_freecount_5cavro_Record = 0;

static PyObject *__pyx_tp_new_5cavro_Record(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5cavro_Record *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro_Record > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro_Record)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_5cavro_Record[--__pyx_freecount_5cavro_Record];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro_Record));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_5cavro_Record *)o);
  p->__pyx_vtab = __pyx_vtabptr_5cavro_Record;
  p->data = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_Record(PyObject *o) {
  struct __pyx_obj_5cavro_Record *p = (struct __pyx_obj_5cavro_Record *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_Record) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->data);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro_Record < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro_Record)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_5cavro_Record[__pyx_freecount_5cavro_Record++] = ((struct __pyx_obj_5cavro_Record *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro_Record(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_Record *p = (struct __pyx_obj_5cavro_Record *)o;
  if (p->data) {
    e = (*v)(p->data, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_Record(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_Record *p = (struct __pyx_obj_5cavro_Record *)o;
  tmp = ((PyObject*)p->data);
  p->data = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_5cavro_Record(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_tp_richcompare_5cavro_Record(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_5cavro_6Record_11__eq__(o1, o2);
    }
    case Py_NE: {
      PyObject *ret;
      ret = __pyx_pw_5cavro_6Record_11__eq__(o1, o2);
      if (likely(ret && ret != Py_NotImplemented)) {
        int b = __Pyx_PyObject_IsTrue(ret);
        Py_DECREF(ret);
        if (unlikely(b < 0)) return NULL;
        ret = (b) ? Py_False : Py_True;
        Py_INCREF(ret);
      }
      return ret;
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_5cavro_6Record_7__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_5cavro_6Record_7__repr__(self);
}

static PyMethodDef __pyx_methods_5cavro_Record[] = {
  {"__dir__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Record_3__dir__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_5cavro_6Record_7__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"_asdict", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Record_9_asdict, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Record_8_asdict},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Record_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Record_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_Record_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_Record},
  {Py_tp_repr, (void *)__pyx_pw_5cavro_6Record_7__repr__},
  {Py_sq_item, (void *)__pyx_sq_item_5cavro_Record},
  {Py_mp_subscript, (void *)__pyx_pw_5cavro_6Record_5__getitem__},
  {Py_tp_doc, (void *)PyDoc_STR("\n    An instance of a record value from a schema.\n    This class should never be instantiated directly, instead it forms the base-class for `RecordType.record` classes.\n\n    Field values can be accessed using dot notation, e.g. `record.field_name`, index notation, e.g. `record['field_name']`, or by calling `_asdict()`\n    Subclasses of record have a class attribute: `Type`, which is the `RecordType` schema that the record was created from.\n\n    Internally, records are represented as a list of values, one for each field in the record, with associated field metadata.\n    Subclasses can be instantiated in the following ways:\n     * `Record(data: list|tuple)`: The length of data must match the number of fields in the record, and each value should correspond to the relevant field value\n     * `Record(data: dict)`: The keys of the dict must match the field names, and each value should correspond to the relevant field value\n     * `Record(data: Record)`: The record must be of the same type as the subclass, or must be adaptable to the subclass (Matching name and fields)\n     * `Record(**kwargs)`: Each keyword argument should correspond to a field name, and the value should correspond to the relevant field value\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_Record},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_Record},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_5cavro_Record},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_Record},
  {Py_tp_init, (void *)__pyx_pw_5cavro_6Record_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_Record},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_Record_spec = {
  "cavro.Record",
  sizeof(struct __pyx_obj_5cavro_Record),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_Record_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Record = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_5cavro_Record, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Record = {
  0, /*mp_length*/
  __pyx_pw_5cavro_6Record_5__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_5cavro_Record = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""Record", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_Record), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_Record, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_5cavro_6Record_7__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Record, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Record, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    An instance of a record value from a schema.\n    This class should never be instantiated directly, instead it forms the base-class for `RecordType.record` classes.\n\n    Field values can be accessed using dot notation, e.g. `record.field_name`, index notation, e.g. `record['field_name']`, or by calling `_asdict()`\n    Subclasses of record have a class attribute: `Type`, which is the `RecordType` schema that the record was created from.\n\n    Internally, records are represented as a list of values, one for each field in the record, with associated field metadata.\n    Subclasses can be instantiated in the following ways:\n     * `Record(data: list|tuple)`: The length of data must match the number of fields in the record, and each value should correspond to the relevant field value\n     * `Record(data: dict)`: The keys of the dict must match the field names, and each value should correspond to the relevant field value\n     * `Record(data: Record)`: The record must be of the same type as the subclass, or must be adaptable to the subclass (Matching name and fields)\n     * `Record(**kwargs)`: Each keyword argument should correspond to a field name, and the value should correspond to the relevant field value\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_Record, /*tp_traverse*/
  __pyx_tp_clear_5cavro_Record, /*tp_clear*/
  __pyx_tp_richcompare_5cavro_Record, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_Record, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_6Record_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_Record, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_RecordField __pyx_vtable_5cavro_RecordField;

static PyObject *__pyx_tp_new_5cavro_RecordField(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5cavro_RecordField *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5cavro_RecordField *)o);
  p->__pyx_vtab = __pyx_vtabptr_5cavro_RecordField;
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->writer_name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->doc = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->type = ((struct __pyx_obj_5cavro_AvroType *)Py_None); Py_INCREF(Py_None);
  p->default_value = Py_None; Py_INCREF(Py_None);
  p->order = Py_None; Py_INCREF(Py_None);
  p->aliases = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_RecordField(PyObject *o) {
  struct __pyx_obj_5cavro_RecordField *p = (struct __pyx_obj_5cavro_RecordField *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  Py_CLEAR(p->writer_name);
  Py_CLEAR(p->doc);
  Py_CLEAR(p->type);
  Py_CLEAR(p->default_value);
  Py_CLEAR(p->order);
  Py_CLEAR(p->aliases);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_5cavro_RecordField(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_RecordField *p = (struct __pyx_obj_5cavro_RecordField *)o;
  if (p->type) {
    e = (*v)(((PyObject *)p->type), a); if (e) return e;
  }
  if (p->default_value) {
    e = (*v)(p->default_value, a); if (e) return e;
  }
  if (p->order) {
    e = (*v)(p->order, a); if (e) return e;
  }
  if (p->aliases) {
    e = (*v)(p->aliases, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_RecordField(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_RecordField *p = (struct __pyx_obj_5cavro_RecordField *)o;
  tmp = ((PyObject*)p->type);
  p->type = ((struct __pyx_obj_5cavro_AvroType *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->default_value);
  p->default_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->order);
  p->order = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->aliases);
  p->aliases = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_11RecordField_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11RecordField_4name_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_11RecordField_writer_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11RecordField_11writer_name_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_11RecordField_doc(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11RecordField_3doc_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_11RecordField_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11RecordField_4type_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_11RecordField_default_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11RecordField_13default_value_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_11RecordField_order(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11RecordField_5order_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_11RecordField_aliases(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11RecordField_7aliases_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_RecordField[] = {
  {"get_schema", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11RecordField_3get_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11RecordField_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11RecordField_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_RecordField[] = {
  {(char *)"name", __pyx_getprop_5cavro_11RecordField_name, 0, (char *)0, 0},
  {(char *)"writer_name", __pyx_getprop_5cavro_11RecordField_writer_name, 0, (char *)0, 0},
  {(char *)"doc", __pyx_getprop_5cavro_11RecordField_doc, 0, (char *)0, 0},
  {(char *)"type", __pyx_getprop_5cavro_11RecordField_type, 0, (char *)0, 0},
  {(char *)"default_value", __pyx_getprop_5cavro_11RecordField_default_value, 0, (char *)0, 0},
  {(char *)"order", __pyx_getprop_5cavro_11RecordField_order, 0, (char *)0, 0},
  {(char *)"aliases", __pyx_getprop_5cavro_11RecordField_aliases, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_RecordField_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_RecordField},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Holds the metadata for a record schema field.\n    This class should never be instantiated directly, instead it is created by `RecordType` when parsing a schema.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_RecordField},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_RecordField},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_RecordField},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_RecordField},
  {Py_tp_init, (void *)__pyx_pw_5cavro_11RecordField_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_RecordField},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_RecordField_spec = {
  "cavro.RecordField",
  sizeof(struct __pyx_obj_5cavro_RecordField),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_RecordField_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_RecordField = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""RecordField", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_RecordField), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_RecordField, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Holds the metadata for a record schema field.\n    This class should never be instantiated directly, instead it is created by `RecordType` when parsing a schema.\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_RecordField, /*tp_traverse*/
  __pyx_tp_clear_5cavro_RecordField, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_RecordField, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_RecordField, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_11RecordField_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_RecordField, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_5cavro__FieldAccessor(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro__FieldAccessor(PyObject *o) {
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_tp_descr_get_5cavro__FieldAccessor(PyObject *o, PyObject *i, PyObject *c) {
  PyObject *r = 0;
  if (!i) i = Py_None;
  if (!c) c = Py_None;
  r = __pyx_pw_5cavro_14_FieldAccessor_3__get__(o, i, c);
  return r;
}

static int __pyx_tp_descr_set_5cavro__FieldAccessor(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_5cavro_14_FieldAccessor_5__set__(o, i, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__delete__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_5cavro__FieldAccessor[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14_FieldAccessor_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14_FieldAccessor_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__FieldAccessor_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro__FieldAccessor},
  {Py_tp_methods, (void *)__pyx_methods_5cavro__FieldAccessor},
  {Py_tp_descr_get, (void *)__pyx_tp_descr_get_5cavro__FieldAccessor},
  {Py_tp_descr_set, (void *)__pyx_tp_descr_set_5cavro__FieldAccessor},
  {Py_tp_init, (void *)__pyx_pw_5cavro_14_FieldAccessor_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro__FieldAccessor},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__FieldAccessor_spec = {
  "cavro._FieldAccessor",
  sizeof(struct __pyx_obj_5cavro__FieldAccessor),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_type_5cavro__FieldAccessor_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__FieldAccessor = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_FieldAccessor", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__FieldAccessor), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro__FieldAccessor, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__FieldAccessor, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  __pyx_tp_descr_get_5cavro__FieldAccessor, /*tp_descr_get*/
  __pyx_tp_descr_set_5cavro__FieldAccessor, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_14_FieldAccessor_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro__FieldAccessor, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_5cavro__Sentinel(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5cavro__Sentinel *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5cavro__Sentinel *)o);
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro__Sentinel(PyObject *o) {
  struct __pyx_obj_5cavro__Sentinel *p = (struct __pyx_obj_5cavro__Sentinel *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro__Sentinel) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  Py_CLEAR(p->name);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_getprop_5cavro_9_Sentinel_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_9_Sentinel_4name_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_5cavro_9_Sentinel_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_5cavro_9_Sentinel_3__repr__(self);
}

static PyMethodDef __pyx_methods_5cavro__Sentinel[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_5cavro_9_Sentinel_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9_Sentinel_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9_Sentinel_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro__Sentinel[] = {
  {(char *)"name", __pyx_getprop_5cavro_9_Sentinel_name, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__Sentinel_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro__Sentinel},
  {Py_tp_repr, (void *)__pyx_pw_5cavro_9_Sentinel_3__repr__},
  {Py_tp_methods, (void *)__pyx_methods_5cavro__Sentinel},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro__Sentinel},
  {Py_tp_init, (void *)__pyx_pw_5cavro_9_Sentinel_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro__Sentinel},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__Sentinel_spec = {
  "cavro._Sentinel",
  sizeof(struct __pyx_obj_5cavro__Sentinel),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro__Sentinel_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__Sentinel = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_Sentinel", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__Sentinel), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro__Sentinel, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_5cavro_9_Sentinel_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__Sentinel, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro__Sentinel, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_9_Sentinel_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro__Sentinel, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods_5cavro__CanonicalForm[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14_CanonicalForm_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14_CanonicalForm_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__CanonicalForm_slots[] = {
  {Py_tp_methods, (void *)__pyx_methods_5cavro__CanonicalForm},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__CanonicalForm_spec = {
  "cavro._CanonicalForm",
  sizeof(struct __pyx_obj_5cavro__CanonicalForm),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro__CanonicalForm_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__CanonicalForm = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_CanonicalForm", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__CanonicalForm), /*tp_basicsize*/
  0, /*tp_itemsize*/
  0, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__CanonicalForm, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  0, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_AvroType __pyx_vtable_5cavro_AvroType;

static PyObject *__pyx_tp_new_5cavro_AvroType(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5cavro_AvroType *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5cavro_AvroType *)o);
  p->__pyx_vtab = __pyx_vtabptr_5cavro_AvroType;
  p->options = ((struct __pyx_obj_5cavro_Options *)Py_None); Py_INCREF(Py_None);
  p->metadata = Py_None; Py_INCREF(Py_None);
  p->value_adapters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_AvroType(PyObject *o) {
  struct __pyx_obj_5cavro_AvroType *p = (struct __pyx_obj_5cavro_AvroType *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_AvroType) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->options);
  Py_CLEAR(p->metadata);
  Py_CLEAR(p->value_adapters);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_5cavro_AvroType(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_AvroType *p = (struct __pyx_obj_5cavro_AvroType *)o;
  if (p->options) {
    e = (*v)(((PyObject *)p->options), a); if (e) return e;
  }
  if (p->metadata) {
    e = (*v)(p->metadata, a); if (e) return e;
  }
  if (p->value_adapters) {
    e = (*v)(p->value_adapters, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_AvroType(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_AvroType *p = (struct __pyx_obj_5cavro_AvroType *)o;
  tmp = ((PyObject*)p->options);
  p->options = ((struct __pyx_obj_5cavro_Options *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->metadata);
  p->metadata = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->value_adapters);
  p->value_adapters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_8AvroType_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_8AvroType_4type_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_8AvroType_options(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_8AvroType_7options_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_8AvroType_metadata(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_8AvroType_8metadata_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_8AvroType_value_adapters(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_8AvroType_14value_adapters_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_AvroType[] = {
  {"for_source", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_1for_source, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"for_schema", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_3for_schema, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"walk_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_15walk_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_22__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8AvroType_24__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_AvroType[] = {
  {(char *)"type", __pyx_getprop_5cavro_8AvroType_type, 0, (char *)0, 0},
  {(char *)"options", __pyx_getprop_5cavro_8AvroType_options, 0, (char *)0, 0},
  {(char *)"metadata", __pyx_getprop_5cavro_8AvroType_metadata, 0, (char *)0, 0},
  {(char *)"value_adapters", __pyx_getprop_5cavro_8AvroType_value_adapters, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_AvroType_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_AvroType},
  {Py_tp_str, (void *)__pyx_pw_5cavro_8AvroType_20__str__},
  {Py_tp_doc, (void *)PyDoc_STR("\n    The base class for all Avro types.\n    `cavro` separates the concept of a `Schema` from a `Type`, which is not strictly neccessary, but\n    makes some management of state a bit easier.\n    Here, a `Type` is the specific implementation of the data management, whereas a `Schema` is the\n    public interface for a schema definition, typically (but not always) containing multiple types either in unions or in nested record fields.\n\n    For normal usage, the `AvroType` class and subclasses can largely be ignored.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_AvroType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_AvroType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_AvroType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_AvroType},
  {Py_tp_init, (void *)__pyx_pw_5cavro_8AvroType_5__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_AvroType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_AvroType_spec = {
  "cavro.AvroType",
  sizeof(struct __pyx_obj_5cavro_AvroType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_AvroType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_AvroType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""AvroType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_AvroType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_AvroType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    The base class for all Avro types.\n    `cavro` separates the concept of a `Schema` from a `Type`, which is not strictly neccessary, but\n    makes some management of state a bit easier.\n    Here, a `Type` is the specific implementation of the data management, whereas a `Schema` is the\n    public interface for a schema definition, typically (but not always) containing multiple types either in unions or in nested record fields.\n\n    For normal usage, the `AvroType` class and subclasses can largely be ignored.\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_AvroType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_AvroType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_AvroType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_AvroType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_8AvroType_5__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_AvroType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_UnionType __pyx_vtable_5cavro_UnionType;

static PyObject *__pyx_tp_new_5cavro_UnionType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_UnionType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_UnionType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_UnionType;
  p->union_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->by_name_map = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->return_type_tuple = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_UnionType(PyObject *o) {
  struct __pyx_obj_5cavro_UnionType *p = (struct __pyx_obj_5cavro_UnionType *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->union_types);
  Py_CLEAR(p->by_name_map);
  Py_CLEAR(p->return_type_tuple);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_5cavro_AvroType(o);
}

static int __pyx_tp_traverse_5cavro_UnionType(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_UnionType *p = (struct __pyx_obj_5cavro_UnionType *)o;
  e = __pyx_tp_traverse_5cavro_AvroType(o, v, a); if (e) return e;
  if (p->union_types) {
    e = (*v)(p->union_types, a); if (e) return e;
  }
  if (p->by_name_map) {
    e = (*v)(p->by_name_map, a); if (e) return e;
  }
  if (p->return_type_tuple) {
    e = (*v)(p->return_type_tuple, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_UnionType(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_UnionType *p = (struct __pyx_obj_5cavro_UnionType *)o;
  __pyx_tp_clear_5cavro_AvroType(o);
  tmp = ((PyObject*)p->union_types);
  p->union_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->by_name_map);
  p->by_name_map = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->return_type_tuple);
  p->return_type_tuple = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_9UnionType_union_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_9UnionType_11union_types_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_9UnionType_by_name_map(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_9UnionType_11by_name_map_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_9UnionType_return_type_tuple(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_9UnionType_17return_type_tuple_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_UnionType[] = {
  {"walk_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9UnionType_5walk_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9UnionType_12__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9UnionType_14__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_UnionType[] = {
  {(char *)"union_types", __pyx_getprop_5cavro_9UnionType_union_types, 0, (char *)0, 0},
  {(char *)"by_name_map", __pyx_getprop_5cavro_9UnionType_by_name_map, 0, (char *)0, 0},
  {(char *)"return_type_tuple", __pyx_getprop_5cavro_9UnionType_return_type_tuple, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_UnionType_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_UnionType},
  {Py_tp_doc, (void *)PyDoc_STR("The avro union type")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_UnionType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_UnionType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_UnionType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_UnionType},
  {Py_tp_init, (void *)__pyx_pw_5cavro_9UnionType_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_UnionType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_UnionType_spec = {
  "cavro.UnionType",
  sizeof(struct __pyx_obj_5cavro_UnionType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_UnionType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_UnionType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""UnionType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_UnionType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_UnionType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro union type"), /*tp_doc*/
  __pyx_tp_traverse_5cavro_UnionType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_UnionType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_UnionType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_UnionType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_9UnionType_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_UnionType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_StringType __pyx_vtable_5cavro_StringType;

static PyObject *__pyx_tp_new_5cavro_StringType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_StringType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_StringType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_StringType;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_StringType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10StringType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10StringType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_StringType_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("The avro string type")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_AvroType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_AvroType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_StringType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_StringType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_StringType_spec = {
  "cavro.StringType",
  sizeof(struct __pyx_obj_5cavro_StringType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_StringType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_StringType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""StringType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_StringType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_AvroType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro string type"), /*tp_doc*/
  __pyx_tp_traverse_5cavro_AvroType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_AvroType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_StringType, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_5__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_StringType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_BytesType __pyx_vtable_5cavro_BytesType;

static PyObject *__pyx_tp_new_5cavro_BytesType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_BytesType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_BytesType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_BytesType;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_BytesType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9BytesType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9BytesType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_BytesType_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("The avro bytes type")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_AvroType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_AvroType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_BytesType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_BytesType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_BytesType_spec = {
  "cavro.BytesType",
  sizeof(struct __pyx_obj_5cavro_BytesType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_BytesType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_BytesType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""BytesType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_BytesType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_AvroType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro bytes type"), /*tp_doc*/
  __pyx_tp_traverse_5cavro_AvroType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_AvroType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_BytesType, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_5__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_BytesType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro__PlaceholderType __pyx_vtable_5cavro__PlaceholderType;

static PyObject *__pyx_tp_new_5cavro__PlaceholderType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro__PlaceholderType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro__PlaceholderType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro__PlaceholderType;
  p->default_value = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro__PlaceholderType(PyObject *o) {
  struct __pyx_obj_5cavro__PlaceholderType *p = (struct __pyx_obj_5cavro__PlaceholderType *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro__PlaceholderType) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->default_value);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_5cavro_AvroType(o);
}

static int __pyx_tp_traverse_5cavro__PlaceholderType(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro__PlaceholderType *p = (struct __pyx_obj_5cavro__PlaceholderType *)o;
  e = __pyx_tp_traverse_5cavro_AvroType(o, v, a); if (e) return e;
  if (p->default_value) {
    e = (*v)(p->default_value, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro__PlaceholderType(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro__PlaceholderType *p = (struct __pyx_obj_5cavro__PlaceholderType *)o;
  __pyx_tp_clear_5cavro_AvroType(o);
  tmp = ((PyObject*)p->default_value);
  p->default_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_16_PlaceholderType_default_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_16_PlaceholderType_13default_value_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro__PlaceholderType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_16_PlaceholderType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_16_PlaceholderType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro__PlaceholderType[] = {
  {(char *)"default_value", __pyx_getprop_5cavro_16_PlaceholderType_default_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__PlaceholderType_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro__PlaceholderType},
  {Py_tp_doc, (void *)PyDoc_STR("\n    An avro type that just provied a constant value when read.\n    This is used during schema promotion where the reader has a defaulted value that the writer does not.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro__PlaceholderType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro__PlaceholderType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro__PlaceholderType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro__PlaceholderType},
  {Py_tp_init, (void *)__pyx_pw_5cavro_16_PlaceholderType_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro__PlaceholderType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__PlaceholderType_spec = {
  "cavro._PlaceholderType",
  sizeof(struct __pyx_obj_5cavro__PlaceholderType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro__PlaceholderType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__PlaceholderType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_PlaceholderType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__PlaceholderType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro__PlaceholderType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    An avro type that just provied a constant value when read.\n    This is used during schema promotion where the reader has a defaulted value that the writer does not.\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro__PlaceholderType, /*tp_traverse*/
  __pyx_tp_clear_5cavro__PlaceholderType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__PlaceholderType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro__PlaceholderType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_16_PlaceholderType_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro__PlaceholderType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_DoubleType __pyx_vtable_5cavro_DoubleType;

static PyObject *__pyx_tp_new_5cavro_DoubleType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_DoubleType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_DoubleType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_DoubleType;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_DoubleType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10DoubleType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10DoubleType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_DoubleType_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("The avro double type.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_AvroType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_AvroType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_DoubleType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_DoubleType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_DoubleType_spec = {
  "cavro.DoubleType",
  sizeof(struct __pyx_obj_5cavro_DoubleType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_DoubleType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_DoubleType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""DoubleType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_DoubleType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_AvroType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro double type."), /*tp_doc*/
  __pyx_tp_traverse_5cavro_AvroType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_AvroType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_DoubleType, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_5__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_DoubleType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_FloatType __pyx_vtable_5cavro_FloatType;

static PyObject *__pyx_tp_new_5cavro_FloatType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_FloatType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_FloatType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_FloatType;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_FloatType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FloatType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FloatType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_FloatType_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("The avro float type.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_AvroType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_AvroType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_FloatType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_FloatType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_FloatType_spec = {
  "cavro.FloatType",
  sizeof(struct __pyx_obj_5cavro_FloatType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_FloatType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_FloatType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""FloatType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_FloatType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_AvroType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro float type."), /*tp_doc*/
  __pyx_tp_traverse_5cavro_AvroType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_AvroType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_FloatType, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_5__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_FloatType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_LongType __pyx_vtable_5cavro_LongType;

static PyObject *__pyx_tp_new_5cavro_LongType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_LongType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_LongType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_LongType;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_LongType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8LongType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8LongType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_LongType_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("The avro long type.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_AvroType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_AvroType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_LongType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_LongType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_LongType_spec = {
  "cavro.LongType",
  sizeof(struct __pyx_obj_5cavro_LongType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_LongType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_LongType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""LongType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_LongType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_AvroType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro long type."), /*tp_doc*/
  __pyx_tp_traverse_5cavro_AvroType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_AvroType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_LongType, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_5__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_LongType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_IntType __pyx_vtable_5cavro_IntType;

static PyObject *__pyx_tp_new_5cavro_IntType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_IntType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_IntType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_IntType;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_IntType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7IntType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7IntType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_IntType_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("The avro int type.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_AvroType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_AvroType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_IntType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_IntType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_IntType_spec = {
  "cavro.IntType",
  sizeof(struct __pyx_obj_5cavro_IntType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_IntType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_IntType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""IntType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_IntType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_AvroType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro int type."), /*tp_doc*/
  __pyx_tp_traverse_5cavro_AvroType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_AvroType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_IntType, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_5__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_IntType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_BoolType __pyx_vtable_5cavro_BoolType;

static PyObject *__pyx_tp_new_5cavro_BoolType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_BoolType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_BoolType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_BoolType;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_BoolType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8BoolType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8BoolType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_BoolType_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("The avro boolean type.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_AvroType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_AvroType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_BoolType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_BoolType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_BoolType_spec = {
  "cavro.BoolType",
  sizeof(struct __pyx_obj_5cavro_BoolType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_BoolType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_BoolType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""BoolType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_BoolType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_AvroType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro boolean type."), /*tp_doc*/
  __pyx_tp_traverse_5cavro_AvroType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_AvroType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_BoolType, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_5__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_BoolType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_NullType __pyx_vtable_5cavro_NullType;

static PyObject *__pyx_tp_new_5cavro_NullType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_NullType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_NullType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_NullType;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_NullType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8NullType_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8NullType_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_NullType_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("The avro null type.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_AvroType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_AvroType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_NullType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_NullType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_NullType_spec = {
  "cavro.NullType",
  sizeof(struct __pyx_obj_5cavro_NullType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_NullType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_NullType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""NullType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_NullType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_AvroType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro null type."), /*tp_doc*/
  __pyx_tp_traverse_5cavro_AvroType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_AvroType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_NullType, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_5__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_NullType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_MapType __pyx_vtable_5cavro_MapType;

static PyObject *__pyx_tp_new_5cavro_MapType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_MapType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_MapType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_MapType;
  p->key_type = ((struct __pyx_obj_5cavro_StringType *)Py_None); Py_INCREF(Py_None);
  p->value_type = ((struct __pyx_obj_5cavro_AvroType *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_MapType(PyObject *o) {
  struct __pyx_obj_5cavro_MapType *p = (struct __pyx_obj_5cavro_MapType *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_MapType) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->key_type);
  Py_CLEAR(p->value_type);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_5cavro_AvroType(o);
}

static int __pyx_tp_traverse_5cavro_MapType(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_MapType *p = (struct __pyx_obj_5cavro_MapType *)o;
  e = __pyx_tp_traverse_5cavro_AvroType(o, v, a); if (e) return e;
  if (p->key_type) {
    e = (*v)(((PyObject *)p->key_type), a); if (e) return e;
  }
  if (p->value_type) {
    e = (*v)(((PyObject *)p->value_type), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_MapType(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_MapType *p = (struct __pyx_obj_5cavro_MapType *)o;
  __pyx_tp_clear_5cavro_AvroType(o);
  tmp = ((PyObject*)p->key_type);
  p->key_type = ((struct __pyx_obj_5cavro_StringType *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->value_type);
  p->value_type = ((struct __pyx_obj_5cavro_AvroType *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_7MapType_value_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_7MapType_10value_type_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_MapType[] = {
  {"walk_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7MapType_5walk_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7MapType_12__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_7MapType_14__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_MapType[] = {
  {(char *)"value_type", __pyx_getprop_5cavro_7MapType_value_type, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_MapType_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_MapType},
  {Py_tp_doc, (void *)PyDoc_STR("The avro map type.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_MapType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_MapType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_MapType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_MapType},
  {Py_tp_init, (void *)__pyx_pw_5cavro_7MapType_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_MapType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_MapType_spec = {
  "cavro.MapType",
  sizeof(struct __pyx_obj_5cavro_MapType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_MapType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_MapType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""MapType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_MapType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_MapType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro map type."), /*tp_doc*/
  __pyx_tp_traverse_5cavro_MapType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_MapType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_MapType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_MapType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_7MapType_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_MapType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_ArrayType __pyx_vtable_5cavro_ArrayType;

static PyObject *__pyx_tp_new_5cavro_ArrayType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_ArrayType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_ArrayType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_ArrayType;
  p->item_type = ((struct __pyx_obj_5cavro_AvroType *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_ArrayType(PyObject *o) {
  struct __pyx_obj_5cavro_ArrayType *p = (struct __pyx_obj_5cavro_ArrayType *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->item_type);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_5cavro_AvroType(o);
}

static int __pyx_tp_traverse_5cavro_ArrayType(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_ArrayType *p = (struct __pyx_obj_5cavro_ArrayType *)o;
  e = __pyx_tp_traverse_5cavro_AvroType(o, v, a); if (e) return e;
  if (p->item_type) {
    e = (*v)(((PyObject *)p->item_type), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_ArrayType(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_ArrayType *p = (struct __pyx_obj_5cavro_ArrayType *)o;
  __pyx_tp_clear_5cavro_AvroType(o);
  tmp = ((PyObject*)p->item_type);
  p->item_type = ((struct __pyx_obj_5cavro_AvroType *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_9ArrayType_item_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_9ArrayType_9item_type_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_ArrayType[] = {
  {"walk_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9ArrayType_5walk_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9ArrayType_12__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9ArrayType_14__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_ArrayType[] = {
  {(char *)"item_type", __pyx_getprop_5cavro_9ArrayType_item_type, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_ArrayType_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_ArrayType},
  {Py_tp_doc, (void *)PyDoc_STR("The avro array type.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_ArrayType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_ArrayType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_ArrayType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_ArrayType},
  {Py_tp_init, (void *)__pyx_pw_5cavro_9ArrayType_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_ArrayType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_ArrayType_spec = {
  "cavro.ArrayType",
  sizeof(struct __pyx_obj_5cavro_ArrayType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_ArrayType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_ArrayType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""ArrayType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_ArrayType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ArrayType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro array type."), /*tp_doc*/
  __pyx_tp_traverse_5cavro_ArrayType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_ArrayType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_ArrayType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_ArrayType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_9ArrayType_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_ArrayType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro__NamedType __pyx_vtable_5cavro__NamedType;

static PyObject *__pyx_tp_new_5cavro__NamedType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro__NamedType *p;
  PyObject *o = __pyx_tp_new_5cavro_AvroType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro__NamedType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro__NamedType;
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->namespace = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->effective_namespace = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->aliases = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro__NamedType(PyObject *o) {
  struct __pyx_obj_5cavro__NamedType *p = (struct __pyx_obj_5cavro__NamedType *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro__NamedType) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  Py_CLEAR(p->namespace);
  Py_CLEAR(p->effective_namespace);
  Py_CLEAR(p->aliases);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_5cavro_AvroType(o);
}

static int __pyx_tp_traverse_5cavro__NamedType(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro__NamedType *p = (struct __pyx_obj_5cavro__NamedType *)o;
  e = __pyx_tp_traverse_5cavro_AvroType(o, v, a); if (e) return e;
  if (p->aliases) {
    e = (*v)(p->aliases, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro__NamedType(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro__NamedType *p = (struct __pyx_obj_5cavro__NamedType *)o;
  __pyx_tp_clear_5cavro_AvroType(o);
  tmp = ((PyObject*)p->aliases);
  p->aliases = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_10_NamedType_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_10_NamedType_4type_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_10_NamedType_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_10_NamedType_4name_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_10_NamedType_namespace(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_10_NamedType_9namespace_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_10_NamedType_effective_namespace(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_10_NamedType_19effective_namespace_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_10_NamedType_aliases(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_10_NamedType_7aliases_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro__NamedType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_NamedType_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_NamedType_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro__NamedType[] = {
  {(char *)"type", __pyx_getprop_5cavro_10_NamedType_type, 0, (char *)0, 0},
  {(char *)"name", __pyx_getprop_5cavro_10_NamedType_name, 0, (char *)0, 0},
  {(char *)"namespace", __pyx_getprop_5cavro_10_NamedType_namespace, 0, (char *)0, 0},
  {(char *)"effective_namespace", __pyx_getprop_5cavro_10_NamedType_effective_namespace, 0, (char *)0, 0},
  {(char *)"aliases", __pyx_getprop_5cavro_10_NamedType_aliases, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__NamedType_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro__NamedType},
  {Py_tp_doc, (void *)PyDoc_STR("Base class for all named types")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro__NamedType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro__NamedType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro__NamedType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro__NamedType},
  {Py_tp_init, (void *)__pyx_pw_5cavro_10_NamedType_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro__NamedType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__NamedType_spec = {
  "cavro._NamedType",
  sizeof(struct __pyx_obj_5cavro__NamedType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro__NamedType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__NamedType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_NamedType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__NamedType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro__NamedType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Base class for all named types"), /*tp_doc*/
  __pyx_tp_traverse_5cavro__NamedType, /*tp_traverse*/
  __pyx_tp_clear_5cavro__NamedType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__NamedType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro__NamedType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_10_NamedType_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro__NamedType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_FixedType __pyx_vtable_5cavro_FixedType;

static PyObject *__pyx_tp_new_5cavro_FixedType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_FixedType *p;
  PyObject *o = __pyx_tp_new_5cavro__NamedType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_FixedType *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_FixedType;
  return o;
}

static PyObject *__pyx_getprop_5cavro_9FixedType_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_9FixedType_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_FixedType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FixedType_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9FixedType_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_FixedType[] = {
  {(char *)"size", __pyx_getprop_5cavro_9FixedType_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_FixedType_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("The avro fixed type")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro__NamedType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro__NamedType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_FixedType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_FixedType},
  {Py_tp_init, (void *)__pyx_pw_5cavro_9FixedType_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_FixedType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_FixedType_spec = {
  "cavro.FixedType",
  sizeof(struct __pyx_obj_5cavro_FixedType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_FixedType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_FixedType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""FixedType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_FixedType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro__NamedType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro fixed type"), /*tp_doc*/
  __pyx_tp_traverse_5cavro__NamedType, /*tp_traverse*/
  __pyx_tp_clear_5cavro__NamedType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_FixedType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_FixedType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_9FixedType_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_FixedType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_RecordType __pyx_vtable_5cavro_RecordType;

static PyObject *__pyx_tp_new_5cavro_RecordType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_RecordType *p;
  PyObject *o = __pyx_tp_new_5cavro__NamedType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_RecordType *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_RecordType;
  p->doc = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->fields = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->field_dict = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->record = ((PyTypeObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_RecordType(PyObject *o) {
  struct __pyx_obj_5cavro_RecordType *p = (struct __pyx_obj_5cavro_RecordType *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_RecordType) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->doc);
  Py_CLEAR(p->fields);
  Py_CLEAR(p->field_dict);
  Py_CLEAR(p->record);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_5cavro__NamedType(o);
}

static int __pyx_tp_traverse_5cavro_RecordType(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_RecordType *p = (struct __pyx_obj_5cavro_RecordType *)o;
  e = __pyx_tp_traverse_5cavro__NamedType(o, v, a); if (e) return e;
  if (p->fields) {
    e = (*v)(p->fields, a); if (e) return e;
  }
  if (p->field_dict) {
    e = (*v)(p->field_dict, a); if (e) return e;
  }
  if (p->record) {
    e = (*v)(((PyObject *)p->record), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_RecordType(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_RecordType *p = (struct __pyx_obj_5cavro_RecordType *)o;
  __pyx_tp_clear_5cavro__NamedType(o);
  tmp = ((PyObject*)p->fields);
  p->fields = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->field_dict);
  p->field_dict = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->record);
  p->record = ((PyTypeObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_10RecordType_doc(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_10RecordType_3doc_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_10RecordType_fields(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_10RecordType_6fields_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_10RecordType_record(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_10RecordType_6record_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_RecordType[] = {
  {"walk_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10RecordType_5walk_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10RecordType_12__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10RecordType_14__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_RecordType[] = {
  {(char *)"doc", __pyx_getprop_5cavro_10RecordType_doc, 0, (char *)0, 0},
  {(char *)"fields", __pyx_getprop_5cavro_10RecordType_fields, 0, (char *)0, 0},
  {(char *)"record", __pyx_getprop_5cavro_10RecordType_record, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_RecordType_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_RecordType},
  {Py_tp_doc, (void *)PyDoc_STR("\n    The Type that corresponds to a Record in a Schema.\n\n    Attributes:\n     * `doc` Any \"doc\" metadata defined in the schema\n     * `fields` A tuple of `RecordField` instances, one for each field in the record\n     * `record` A subclass of `Record` that can be used to instantiate records of this type\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_RecordType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_RecordType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_RecordType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_RecordType},
  {Py_tp_init, (void *)__pyx_pw_5cavro_10RecordType_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_RecordType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_RecordType_spec = {
  "cavro.RecordType",
  sizeof(struct __pyx_obj_5cavro_RecordType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_RecordType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_RecordType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""RecordType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_RecordType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_RecordType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    The Type that corresponds to a Record in a Schema.\n\n    Attributes:\n     * `doc` Any \"doc\" metadata defined in the schema\n     * `fields` A tuple of `RecordField` instances, one for each field in the record\n     * `record` A subclass of `Record` that can be used to instantiate records of this type\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_RecordType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_RecordType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_RecordType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_RecordType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_10RecordType_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_RecordType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_PromotingRecordType __pyx_vtable_5cavro_PromotingRecordType;

static PyObject *__pyx_tp_new_5cavro_PromotingRecordType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_PromotingRecordType *p;
  PyObject *o = __pyx_tp_new_5cavro_RecordType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_PromotingRecordType *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_PromotingRecordType;
  p->decode_indexes.data = NULL;
  p->decode_indexes.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_5cavro_PromotingRecordType(PyObject *o) {
  struct __pyx_obj_5cavro_PromotingRecordType *p = (struct __pyx_obj_5cavro_PromotingRecordType *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_PromotingRecordType) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __PYX_XCLEAR_MEMVIEW(&p->decode_indexes, 1);
  p->decode_indexes.memview = NULL; p->decode_indexes.data = NULL;
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_5cavro_RecordType(o);
}

static PyMethodDef __pyx_methods_5cavro_PromotingRecordType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_19PromotingRecordType_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_19PromotingRecordType_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_PromotingRecordType_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_PromotingRecordType},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A variant of a `RecordType`, specialized for reading records from a different schema from a writer.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_RecordType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_RecordType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_PromotingRecordType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_PromotingRecordType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_PromotingRecordType_spec = {
  "cavro.PromotingRecordType",
  sizeof(struct __pyx_obj_5cavro_PromotingRecordType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_PromotingRecordType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_PromotingRecordType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""PromotingRecordType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_PromotingRecordType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_PromotingRecordType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    A variant of a `RecordType`, specialized for reading records from a different schema from a writer.\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_RecordType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_RecordType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_PromotingRecordType, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_10RecordType_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_PromotingRecordType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_EnumType __pyx_vtable_5cavro_EnumType;

static PyObject *__pyx_tp_new_5cavro_EnumType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_EnumType *p;
  PyObject *o = __pyx_tp_new_5cavro__NamedType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_EnumType *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_EnumType;
  p->symbols = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->symbol_indexes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->default_value = Py_None; Py_INCREF(Py_None);
  p->doc = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_EnumType(PyObject *o) {
  struct __pyx_obj_5cavro_EnumType *p = (struct __pyx_obj_5cavro_EnumType *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_EnumType) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->symbols);
  Py_CLEAR(p->symbol_indexes);
  Py_CLEAR(p->default_value);
  Py_CLEAR(p->doc);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_5cavro__NamedType(o);
}

static int __pyx_tp_traverse_5cavro_EnumType(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_EnumType *p = (struct __pyx_obj_5cavro_EnumType *)o;
  e = __pyx_tp_traverse_5cavro__NamedType(o, v, a); if (e) return e;
  if (p->symbols) {
    e = (*v)(p->symbols, a); if (e) return e;
  }
  if (p->symbol_indexes) {
    e = (*v)(p->symbol_indexes, a); if (e) return e;
  }
  if (p->default_value) {
    e = (*v)(p->default_value, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_EnumType(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_EnumType *p = (struct __pyx_obj_5cavro_EnumType *)o;
  __pyx_tp_clear_5cavro__NamedType(o);
  tmp = ((PyObject*)p->symbols);
  p->symbols = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->symbol_indexes);
  p->symbol_indexes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->default_value);
  p->default_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_8EnumType_symbols(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_8EnumType_7symbols_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_8EnumType_default_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_8EnumType_13default_value_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_8EnumType_doc(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_8EnumType_3doc_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_EnumType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8EnumType_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8EnumType_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_EnumType[] = {
  {(char *)"symbols", __pyx_getprop_5cavro_8EnumType_symbols, 0, (char *)0, 0},
  {(char *)"default_value", __pyx_getprop_5cavro_8EnumType_default_value, 0, (char *)0, 0},
  {(char *)"doc", __pyx_getprop_5cavro_8EnumType_doc, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_EnumType_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_EnumType},
  {Py_tp_doc, (void *)PyDoc_STR("The avro enum type")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_EnumType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_EnumType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_EnumType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_EnumType},
  {Py_tp_init, (void *)__pyx_pw_5cavro_8EnumType_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_EnumType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_EnumType_spec = {
  "cavro.EnumType",
  sizeof(struct __pyx_obj_5cavro_EnumType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_EnumType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_EnumType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""EnumType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_EnumType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_EnumType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("The avro enum type"), /*tp_doc*/
  __pyx_tp_traverse_5cavro_EnumType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_EnumType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_EnumType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_EnumType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_8EnumType_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_EnumType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_PromotingEnumType __pyx_vtable_5cavro_PromotingEnumType;

static PyObject *__pyx_tp_new_5cavro_PromotingEnumType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_PromotingEnumType *p;
  PyObject *o = __pyx_tp_new_5cavro_EnumType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_PromotingEnumType *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_AvroType*)__pyx_vtabptr_5cavro_PromotingEnumType;
  p->unknown_symbols = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->reader_type = ((struct __pyx_obj_5cavro_EnumType *)Py_None); Py_INCREF(Py_None);
  p->writer_type = ((struct __pyx_obj_5cavro_EnumType *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_PromotingEnumType(PyObject *o) {
  struct __pyx_obj_5cavro_PromotingEnumType *p = (struct __pyx_obj_5cavro_PromotingEnumType *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_PromotingEnumType) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->unknown_symbols);
  Py_CLEAR(p->reader_type);
  Py_CLEAR(p->writer_type);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_5cavro_EnumType(o);
}

static int __pyx_tp_traverse_5cavro_PromotingEnumType(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_PromotingEnumType *p = (struct __pyx_obj_5cavro_PromotingEnumType *)o;
  e = __pyx_tp_traverse_5cavro_EnumType(o, v, a); if (e) return e;
  if (p->unknown_symbols) {
    e = (*v)(p->unknown_symbols, a); if (e) return e;
  }
  if (p->reader_type) {
    e = (*v)(((PyObject *)p->reader_type), a); if (e) return e;
  }
  if (p->writer_type) {
    e = (*v)(((PyObject *)p->writer_type), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_PromotingEnumType(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_PromotingEnumType *p = (struct __pyx_obj_5cavro_PromotingEnumType *)o;
  __pyx_tp_clear_5cavro_EnumType(o);
  tmp = ((PyObject*)p->unknown_symbols);
  p->unknown_symbols = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->reader_type);
  p->reader_type = ((struct __pyx_obj_5cavro_EnumType *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->writer_type);
  p->writer_type = ((struct __pyx_obj_5cavro_EnumType *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_17PromotingEnumType_unknown_symbols(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_17PromotingEnumType_15unknown_symbols_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_17PromotingEnumType_reader_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_17PromotingEnumType_11reader_type_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_17PromotingEnumType_writer_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_17PromotingEnumType_11writer_type_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_PromotingEnumType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_17PromotingEnumType_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_17PromotingEnumType_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_PromotingEnumType[] = {
  {(char *)"unknown_symbols", __pyx_getprop_5cavro_17PromotingEnumType_unknown_symbols, 0, (char *)0, 0},
  {(char *)"reader_type", __pyx_getprop_5cavro_17PromotingEnumType_reader_type, 0, (char *)0, 0},
  {(char *)"writer_type", __pyx_getprop_5cavro_17PromotingEnumType_writer_type, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_PromotingEnumType_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_PromotingEnumType},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_PromotingEnumType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_PromotingEnumType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_PromotingEnumType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_PromotingEnumType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_PromotingEnumType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_PromotingEnumType_spec = {
  "cavro.PromotingEnumType",
  sizeof(struct __pyx_obj_5cavro_PromotingEnumType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_PromotingEnumType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_PromotingEnumType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""PromotingEnumType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_PromotingEnumType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_PromotingEnumType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8AvroType_20__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro_PromotingEnumType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_PromotingEnumType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_PromotingEnumType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_PromotingEnumType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8EnumType_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_PromotingEnumType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_Schema __pyx_vtable_5cavro_Schema;

static PyObject *__pyx_tp_new_5cavro_Schema(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5cavro_Schema *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5cavro_Schema *)o);
  p->__pyx_vtab = __pyx_vtabptr_5cavro_Schema;
  p->named_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->source = Py_None; Py_INCREF(Py_None);
  p->options = ((struct __pyx_obj_5cavro_Options *)Py_None); Py_INCREF(Py_None);
  p->type = ((struct __pyx_obj_5cavro_AvroType *)Py_None); Py_INCREF(Py_None);
  p->_canonical_form = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->logical_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_Schema(PyObject *o) {
  struct __pyx_obj_5cavro_Schema *p = (struct __pyx_obj_5cavro_Schema *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_Schema) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->named_types);
  Py_CLEAR(p->source);
  Py_CLEAR(p->options);
  Py_CLEAR(p->type);
  Py_CLEAR(p->_canonical_form);
  Py_CLEAR(p->logical_types);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_5cavro_Schema(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_Schema *p = (struct __pyx_obj_5cavro_Schema *)o;
  if (p->named_types) {
    e = (*v)(p->named_types, a); if (e) return e;
  }
  if (p->source) {
    e = (*v)(p->source, a); if (e) return e;
  }
  if (p->options) {
    e = (*v)(((PyObject *)p->options), a); if (e) return e;
  }
  if (p->type) {
    e = (*v)(((PyObject *)p->type), a); if (e) return e;
  }
  if (p->logical_types) {
    e = (*v)(p->logical_types, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_Schema(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_Schema *p = (struct __pyx_obj_5cavro_Schema *)o;
  tmp = ((PyObject*)p->named_types);
  p->named_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->source);
  p->source = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->options);
  p->options = ((struct __pyx_obj_5cavro_Options *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->type);
  p->type = ((struct __pyx_obj_5cavro_AvroType *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->logical_types);
  p->logical_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_6Schema_canonical_form(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_6Schema_14canonical_form_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_6Schema_schema(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_6Schema_6schema_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_6Schema_schema_str(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_6Schema_10schema_str_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_6Schema_named_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_6Schema_11named_types_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_6Schema_source(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_6Schema_6source_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_6Schema_options(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_6Schema_7options_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_6Schema_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_6Schema_4type_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_6Schema_logical_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_6Schema_13logical_types_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_Schema[] = {
  {"_wrap_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_3_wrap_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_2_wrap_type},
  {"_make_logical_types", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_5_make_logical_types, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"fingerprint", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_9fingerprint, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_8fingerprint},
  {"can_encode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_13can_encode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_12can_encode},
  {"binary_encode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_15binary_encode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_14binary_encode},
  {"binary_decode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_17binary_decode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_16binary_decode},
  {"json_encode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_23json_encode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_22json_encode},
  {"json_decode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_25json_decode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_5cavro_6Schema_24json_decode},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_29__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_6Schema_31__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_Schema[] = {
  {(char *)"canonical_form", __pyx_getprop_5cavro_6Schema_canonical_form, 0, (char *)PyDoc_STR("\n        Returns the canonical form of the schema as a string\n        "), 0},
  {(char *)"schema", __pyx_getprop_5cavro_6Schema_schema, 0, (char *)PyDoc_STR("\n        Return an object representing the schema definition.\n        Note: This will not always be identical to the `source` used to construct this schema object, as it is reconstructed from the types on-demand.\n        "), 0},
  {(char *)"schema_str", __pyx_getprop_5cavro_6Schema_schema_str, 0, (char *)PyDoc_STR("\n        `Schema.schema`, but json encoded\n        "), 0},
  {(char *)"named_types", __pyx_getprop_5cavro_6Schema_named_types, 0, (char *)0, 0},
  {(char *)"source", __pyx_getprop_5cavro_6Schema_source, 0, (char *)0, 0},
  {(char *)"options", __pyx_getprop_5cavro_6Schema_options, 0, (char *)0, 0},
  {(char *)"type", __pyx_getprop_5cavro_6Schema_type, 0, (char *)0, 0},
  {(char *)"logical_types", __pyx_getprop_5cavro_6Schema_logical_types, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_Schema_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_Schema},
  {Py_tp_hash, (void *)__pyx_pw_5cavro_6Schema_7__hash__},
  {Py_tp_doc, (void *)PyDoc_STR("\n    The main interface for `cavro`.\n\n    This class represents an avro schema, and is able to encode and decode values appropriately.\n\n    Arguments:\n     * `source`:\n        The source of the schema. This can either be a string that holds the JSON-encoded schema definition, or a python object that represents the schema (e.g. the result of `json.loads`).\n     * `options`:\n        An instance of `Options` that controls how the schema is interpreted. Defaults to `DEFAULT_OPTIONS`.\n     * `named_types`:\n        An optional dictionary that will be updated to contain any named types that are encountered while parsing the schema.\n     * `parse_json`:\n        If `False` then the `source` argument will never be parsed as json, even if it's a string value. Defaults to `True`.\n     * `**extra_options`:\n        Any extra options that should be applied to the schema. These will override any options that are set in the `options` argument.\n        Key-values here must match the attributes of `cavro.Options`.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_Schema},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_Schema},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_Schema},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_Schema},
  {Py_tp_init, (void *)__pyx_pw_5cavro_6Schema_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_Schema},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_Schema_spec = {
  "cavro.Schema",
  sizeof(struct __pyx_obj_5cavro_Schema),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_Schema_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_Schema = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""Schema", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_Schema), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_Schema, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_5cavro_6Schema_7__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    The main interface for `cavro`.\n\n    This class represents an avro schema, and is able to encode and decode values appropriately.\n\n    Arguments:\n     * `source`:\n        The source of the schema. This can either be a string that holds the JSON-encoded schema definition, or a python object that represents the schema (e.g. the result of `json.loads`).\n     * `options`:\n        An instance of `Options` that controls how the schema is interpreted. Defaults to `DEFAULT_OPTIONS`.\n     * `named_types`:\n        An optional dictionary that will be updated to contain any named types that are encountered while parsing the schema.\n     * `parse_json`:\n        If `False` then the `source` argument will never be parsed as json, even if it's a string value. Defaults to `True`.\n     * `**extra_options`:\n        Any extra options that should be applied to the schema. These will override any options that are set in the `options` argument.\n        Key-values here must match the attributes of `cavro.Options`.\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_Schema, /*tp_traverse*/
  __pyx_tp_clear_5cavro_Schema, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_Schema, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_Schema, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_6Schema_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_Schema, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_ResolvedSchema __pyx_vtable_5cavro_ResolvedSchema;

static PyObject *__pyx_tp_new_5cavro_ResolvedSchema(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_ResolvedSchema *p;
  PyObject *o = __pyx_tp_new_5cavro_Schema(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_ResolvedSchema *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_Schema*)__pyx_vtabptr_5cavro_ResolvedSchema;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_ResolvedSchema[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14ResolvedSchema_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14ResolvedSchema_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_ResolvedSchema_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("\n    A variant of a schema that is the result of schema promotion.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_Schema},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_Schema},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_ResolvedSchema},
  {Py_tp_init, (void *)__pyx_pw_5cavro_14ResolvedSchema_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_ResolvedSchema},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_ResolvedSchema_spec = {
  "cavro.ResolvedSchema",
  sizeof(struct __pyx_obj_5cavro_ResolvedSchema),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_ResolvedSchema_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_ResolvedSchema = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""ResolvedSchema", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_ResolvedSchema), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_Schema, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    A variant of a schema that is the result of schema promotion.\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_Schema, /*tp_traverse*/
  __pyx_tp_clear_5cavro_Schema, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_ResolvedSchema, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_14ResolvedSchema_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_ResolvedSchema, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_ValueAdapter __pyx_vtable_5cavro_ValueAdapter;

static PyObject *__pyx_tp_new_5cavro_ValueAdapter(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5cavro_ValueAdapter *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5cavro_ValueAdapter *)o);
  p->__pyx_vtab = __pyx_vtabptr_5cavro_ValueAdapter;
  return o;
}

static void __pyx_tp_dealloc_5cavro_ValueAdapter(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_ValueAdapter) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyMethodDef __pyx_methods_5cavro_ValueAdapter[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12ValueAdapter_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12ValueAdapter_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_ValueAdapter_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_ValueAdapter},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Abstract base class for any helper that affects how values are transformed prior to avro encoding/decoding.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_ValueAdapter},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_ValueAdapter},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_ValueAdapter_spec = {
  "cavro.ValueAdapter",
  sizeof(struct __pyx_obj_5cavro_ValueAdapter),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_ValueAdapter_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_ValueAdapter = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""ValueAdapter", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_ValueAdapter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    Abstract base class for any helper that affects how values are transformed prior to avro encoding/decoding.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_ValueAdapter, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_ValueAdapter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_LogicalType __pyx_vtable_5cavro_LogicalType;

static PyObject *__pyx_tp_new_5cavro_LogicalType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_LogicalType *p;
  PyObject *o = __pyx_tp_new_5cavro_ValueAdapter(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_LogicalType *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_LogicalType;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_LogicalType[] = {
  {"for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11LogicalType_1for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11LogicalType_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11LogicalType_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_LogicalType_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("\n    Semi-abstract class for all logical types.\n    \n    Subclasses must be implemented as cython classes.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_LogicalType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_LogicalType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_LogicalType_spec = {
  "cavro.LogicalType",
  sizeof(struct __pyx_obj_5cavro_LogicalType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_LogicalType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_LogicalType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""LogicalType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_LogicalType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    Semi-abstract class for all logical types.\n    \n    Subclasses must be implemented as cython classes.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_LogicalType, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_LogicalType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_CustomLogicalType __pyx_vtable_5cavro_CustomLogicalType;

static PyObject *__pyx_tp_new_5cavro_CustomLogicalType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_CustomLogicalType *p;
  PyObject *o = __pyx_tp_new_5cavro_LogicalType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_CustomLogicalType *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_CustomLogicalType;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_CustomLogicalType[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_17CustomLogicalType_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_17CustomLogicalType_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_CustomLogicalType_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("\n    Logical type that allows custom encoding/decoding functions to be provided.\n    \n    To use a custom logical type, subclass this class, and implement the:\n     * `encode_value(value: object) -> object` - Transforms a provided value into something that can be encoded by the underlying type.\n     * `decode_value(value: object) -> object` - Transforms a value decoded by the underlying type into a value to return to the user.\n    methods.\n\n    Also implement two class attributes, and a classmethod:\n     * `logical_name` - The name to look for in the schema\n     * `underlying_types` A tuple of `AvtoType` classes that this logical type can be applied to.\n     * `_for_type(cls, underlying: AvroType) -> Cls` \n        A classmethod that returns an instance of the class, optionally customized with information from the underlying type,\n        or None if the logical type is not applicable to the underlying type.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_CustomLogicalType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_CustomLogicalType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_CustomLogicalType_spec = {
  "cavro.CustomLogicalType",
  sizeof(struct __pyx_obj_5cavro_CustomLogicalType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_CustomLogicalType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_CustomLogicalType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""CustomLogicalType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_CustomLogicalType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    Logical type that allows custom encoding/decoding functions to be provided.\n    \n    To use a custom logical type, subclass this class, and implement the:\n     * `encode_value(value: object) -> object` - Transforms a provided value into something that can be encoded by the underlying type.\n     * `decode_value(value: object) -> object` - Transforms a value decoded by the underlying type into a value to return to the user.\n    methods.\n\n    Also implement two class attributes, and a classmethod:\n     * `logical_name` - The name to look for in the schema\n     * `underlying_types` A tuple of `AvtoType` classes that this logical type can be applied to.\n     * `_for_type(cls, underlying: AvroType) -> Cls` \n        A classmethod that returns an instance of the class, optionally customized with information from the underlying type,\n        or None if the logical type is not applicable to the underlying type.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_CustomLogicalType, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_CustomLogicalType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_DecimalType __pyx_vtable_5cavro_DecimalType;

static PyObject *__pyx_tp_new_5cavro_DecimalType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_DecimalType *p;
  PyObject *o = __pyx_tp_new_5cavro_LogicalType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_DecimalType *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_DecimalType;
  p->scale_val = Py_None; Py_INCREF(Py_None);
  p->context = Py_None; Py_INCREF(Py_None);
  p->size = Py_None; Py_INCREF(Py_None);
  p->underlying_name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_DecimalType(PyObject *o) {
  struct __pyx_obj_5cavro_DecimalType *p = (struct __pyx_obj_5cavro_DecimalType *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_DecimalType) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->scale_val);
  Py_CLEAR(p->context);
  Py_CLEAR(p->size);
  Py_CLEAR(p->underlying_name);
  __pyx_tp_dealloc_5cavro_ValueAdapter(o);
}

static int __pyx_tp_traverse_5cavro_DecimalType(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_DecimalType *p = (struct __pyx_obj_5cavro_DecimalType *)o;
  e = ((likely(__pyx_ptype_5cavro_LogicalType)) ? ((__pyx_ptype_5cavro_LogicalType->tp_traverse) ? __pyx_ptype_5cavro_LogicalType->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_5cavro_DecimalType)); if (e) return e;
  if (p->scale_val) {
    e = (*v)(p->scale_val, a); if (e) return e;
  }
  if (p->context) {
    e = (*v)(p->context, a); if (e) return e;
  }
  if (p->size) {
    e = (*v)(p->size, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_DecimalType(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_DecimalType *p = (struct __pyx_obj_5cavro_DecimalType *)o;
  if (likely(__pyx_ptype_5cavro_LogicalType)) { if (__pyx_ptype_5cavro_LogicalType->tp_clear) __pyx_ptype_5cavro_LogicalType->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_5cavro_DecimalType);
  tmp = ((PyObject*)p->scale_val);
  p->scale_val = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->context);
  p->context = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->size);
  p->size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_11DecimalType_type_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11DecimalType_9type_name_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_11DecimalType_precision(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11DecimalType_9precision_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_11DecimalType_scale(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11DecimalType_5scale_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_11DecimalType_scale_val(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11DecimalType_9scale_val_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_11DecimalType_context(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11DecimalType_7context_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_11DecimalType_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11DecimalType_4size_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_11DecimalType_check_exp_overflow(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_11DecimalType_18check_exp_overflow_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_DecimalType[] = {
  {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11DecimalType_3_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11DecimalType_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_11DecimalType_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_DecimalType[] = {
  {(char *)"type_name", __pyx_getprop_5cavro_11DecimalType_type_name, 0, (char *)0, 0},
  {(char *)"precision", __pyx_getprop_5cavro_11DecimalType_precision, 0, (char *)0, 0},
  {(char *)"scale", __pyx_getprop_5cavro_11DecimalType_scale, 0, (char *)0, 0},
  {(char *)"scale_val", __pyx_getprop_5cavro_11DecimalType_scale_val, 0, (char *)0, 0},
  {(char *)"context", __pyx_getprop_5cavro_11DecimalType_context, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop_5cavro_11DecimalType_size, 0, (char *)0, 0},
  {(char *)"check_exp_overflow", __pyx_getprop_5cavro_11DecimalType_check_exp_overflow, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_DecimalType_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_DecimalType},
  {Py_tp_doc, (void *)PyDoc_STR("Logical type for decimal values.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_DecimalType},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_DecimalType},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_DecimalType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_DecimalType},
  {Py_tp_init, (void *)__pyx_pw_5cavro_11DecimalType_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_DecimalType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_DecimalType_spec = {
  "cavro.DecimalType",
  sizeof(struct __pyx_obj_5cavro_DecimalType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_DecimalType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_DecimalType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""DecimalType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_DecimalType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_DecimalType, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Logical type for decimal values."), /*tp_doc*/
  __pyx_tp_traverse_5cavro_DecimalType, /*tp_traverse*/
  __pyx_tp_clear_5cavro_DecimalType, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_DecimalType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_DecimalType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_11DecimalType_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_DecimalType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_UUIDBase __pyx_vtable_5cavro_UUIDBase;

static PyObject *__pyx_tp_new_5cavro_UUIDBase(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_UUIDBase *p;
  PyObject *o = __pyx_tp_new_5cavro_LogicalType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_UUIDBase *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_UUIDBase;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_UUIDBase[] = {
  {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8UUIDBase_3_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8UUIDBase_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_8UUIDBase_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_UUIDBase_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("Logical type for UUID values")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_UUIDBase},
  {Py_tp_init, (void *)__pyx_pw_5cavro_8UUIDBase_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_UUIDBase},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_UUIDBase_spec = {
  "cavro.UUIDBase",
  sizeof(struct __pyx_obj_5cavro_UUIDBase),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_UUIDBase_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_UUIDBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""UUIDBase", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_UUIDBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("Logical type for UUID values"), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_UUIDBase, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_8UUIDBase_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_UUIDBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_UUIDStringType __pyx_vtable_5cavro_UUIDStringType;

static PyObject *__pyx_tp_new_5cavro_UUIDStringType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_UUIDStringType *p;
  PyObject *o = __pyx_tp_new_5cavro_UUIDBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_UUIDStringType *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_UUIDStringType;
  return o;
}

static PyObject *__pyx_getprop_5cavro_14UUIDStringType_type_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_14UUIDStringType_9type_name_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_UUIDStringType[] = {
  {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14UUIDStringType_1_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14UUIDStringType_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14UUIDStringType_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_UUIDStringType[] = {
  {(char *)"type_name", __pyx_getprop_5cavro_14UUIDStringType_type_name, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_UUIDStringType_slots[] = {
  {Py_tp_methods, (void *)__pyx_methods_5cavro_UUIDStringType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_UUIDStringType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_UUIDStringType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_UUIDStringType_spec = {
  "cavro.UUIDStringType",
  sizeof(struct __pyx_obj_5cavro_UUIDStringType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_UUIDStringType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_UUIDStringType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""UUIDStringType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_UUIDStringType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_UUIDStringType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_UUIDStringType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8UUIDBase_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_UUIDStringType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_UUIDFixedType __pyx_vtable_5cavro_UUIDFixedType;

static PyObject *__pyx_tp_new_5cavro_UUIDFixedType(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_UUIDFixedType *p;
  PyObject *o = __pyx_tp_new_5cavro_UUIDBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_UUIDFixedType *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_UUIDFixedType;
  return o;
}

static PyObject *__pyx_getprop_5cavro_13UUIDFixedType_type_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_13UUIDFixedType_9type_name_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_UUIDFixedType[] = {
  {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13UUIDFixedType_1_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13UUIDFixedType_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13UUIDFixedType_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_UUIDFixedType[] = {
  {(char *)"type_name", __pyx_getprop_5cavro_13UUIDFixedType_type_name, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_UUIDFixedType_slots[] = {
  {Py_tp_methods, (void *)__pyx_methods_5cavro_UUIDFixedType},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_UUIDFixedType},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_UUIDFixedType},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_UUIDFixedType_spec = {
  "cavro.UUIDFixedType",
  sizeof(struct __pyx_obj_5cavro_UUIDFixedType),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_UUIDFixedType_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_UUIDFixedType = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""UUIDFixedType", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_UUIDFixedType), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_UUIDFixedType, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_UUIDFixedType, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_5cavro_8UUIDBase_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_UUIDFixedType, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_Date __pyx_vtable_5cavro_Date;

static PyObject *__pyx_tp_new_5cavro_Date(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_Date *p;
  PyObject *o = __pyx_tp_new_5cavro_LogicalType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_Date *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_Date;
  return o;
}

static PyObject *__pyx_getprop_5cavro_4Date_type_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_4Date_9type_name_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_4Date_accepts_string(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_4Date_14accepts_string_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_Date[] = {
  {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_4Date_3_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_4Date_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_4Date_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_Date[] = {
  {(char *)"type_name", __pyx_getprop_5cavro_4Date_type_name, 0, (char *)0, 0},
  {(char *)"accepts_string", __pyx_getprop_5cavro_4Date_accepts_string, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_Date_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("Logical type for Date values")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_Date},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_Date},
  {Py_tp_init, (void *)__pyx_pw_5cavro_4Date_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_Date},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_Date_spec = {
  "cavro.Date",
  sizeof(struct __pyx_obj_5cavro_Date),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_Date_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_Date = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""Date", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_Date), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("Logical type for Date values"), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_Date, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_Date, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_4Date_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_Date, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_TimeMillis __pyx_vtable_5cavro_TimeMillis;

static PyObject *__pyx_tp_new_5cavro_TimeMillis(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_TimeMillis *p;
  PyObject *o = __pyx_tp_new_5cavro_LogicalType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_TimeMillis *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_TimeMillis;
  return o;
}

static PyObject *__pyx_getprop_5cavro_10TimeMillis_type_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_10TimeMillis_9type_name_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_TimeMillis[] = {
  {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10TimeMillis_1_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10TimeMillis_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10TimeMillis_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_TimeMillis[] = {
  {(char *)"type_name", __pyx_getprop_5cavro_10TimeMillis_type_name, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_TimeMillis_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("Logical type for time-millis values")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_TimeMillis},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_TimeMillis},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_TimeMillis},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_TimeMillis_spec = {
  "cavro.TimeMillis",
  sizeof(struct __pyx_obj_5cavro_TimeMillis),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_TimeMillis_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_TimeMillis = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""TimeMillis", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_TimeMillis), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("Logical type for time-millis values"), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_TimeMillis, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_TimeMillis, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_TimeMillis, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_TimeMicros __pyx_vtable_5cavro_TimeMicros;

static PyObject *__pyx_tp_new_5cavro_TimeMicros(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_TimeMicros *p;
  PyObject *o = __pyx_tp_new_5cavro_LogicalType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_TimeMicros *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_TimeMicros;
  return o;
}

static PyObject *__pyx_getprop_5cavro_10TimeMicros_type_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_10TimeMicros_9type_name_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_TimeMicros[] = {
  {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10TimeMicros_1_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10TimeMicros_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10TimeMicros_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_TimeMicros[] = {
  {(char *)"type_name", __pyx_getprop_5cavro_10TimeMicros_type_name, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_TimeMicros_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("Logical type for time-micros values")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_TimeMicros},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_TimeMicros},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_TimeMicros},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_TimeMicros_spec = {
  "cavro.TimeMicros",
  sizeof(struct __pyx_obj_5cavro_TimeMicros),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_TimeMicros_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_TimeMicros = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""TimeMicros", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_TimeMicros), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("Logical type for time-micros values"), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_TimeMicros, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_TimeMicros, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_TimeMicros, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_TimestampMillis __pyx_vtable_5cavro_TimestampMillis;

static PyObject *__pyx_tp_new_5cavro_TimestampMillis(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_TimestampMillis *p;
  PyObject *o = __pyx_tp_new_5cavro_LogicalType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_TimestampMillis *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_TimestampMillis;
  return o;
}

static PyObject *__pyx_getprop_5cavro_15TimestampMillis_type_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15TimestampMillis_9type_name_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15TimestampMillis_alternate_timestamp_encoding(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15TimestampMillis_28alternate_timestamp_encoding_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_TimestampMillis[] = {
  {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15TimestampMillis_3_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15TimestampMillis_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15TimestampMillis_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_TimestampMillis[] = {
  {(char *)"type_name", __pyx_getprop_5cavro_15TimestampMillis_type_name, 0, (char *)0, 0},
  {(char *)"alternate_timestamp_encoding", __pyx_getprop_5cavro_15TimestampMillis_alternate_timestamp_encoding, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_TimestampMillis_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("Logical type for timestamp-micros values")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_TimestampMillis},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_TimestampMillis},
  {Py_tp_init, (void *)__pyx_pw_5cavro_15TimestampMillis_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_TimestampMillis},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_TimestampMillis_spec = {
  "cavro.TimestampMillis",
  sizeof(struct __pyx_obj_5cavro_TimestampMillis),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_TimestampMillis_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_TimestampMillis = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""TimestampMillis", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_TimestampMillis), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("Logical type for timestamp-micros values"), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_TimestampMillis, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_TimestampMillis, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_15TimestampMillis_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_TimestampMillis, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_TimestampMicros __pyx_vtable_5cavro_TimestampMicros;

static PyObject *__pyx_tp_new_5cavro_TimestampMicros(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_TimestampMicros *p;
  PyObject *o = __pyx_tp_new_5cavro_LogicalType(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_TimestampMicros *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_TimestampMicros;
  return o;
}

static PyObject *__pyx_getprop_5cavro_15TimestampMicros_type_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15TimestampMicros_9type_name_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_TimestampMicros[] = {
  {"_for_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15TimestampMicros_1_for_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15TimestampMicros_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15TimestampMicros_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_TimestampMicros[] = {
  {(char *)"type_name", __pyx_getprop_5cavro_15TimestampMicros_type_name, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_TimestampMicros_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("Logical type for timestamp-micros values")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_TimestampMicros},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_TimestampMicros},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_TimestampMicros},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_TimestampMicros_spec = {
  "cavro.TimestampMicros",
  sizeof(struct __pyx_obj_5cavro_TimestampMicros),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_TimestampMicros_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_TimestampMicros = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""TimestampMicros", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_TimestampMicros), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("Logical type for timestamp-micros values"), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_TimestampMicros, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_TimestampMicros, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_TimestampMicros, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_Codec __pyx_vtable_5cavro_Codec;

static PyObject *__pyx_tp_new_5cavro_Codec(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5cavro_Codec *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5cavro_Codec *)o);
  p->__pyx_vtab = __pyx_vtabptr_5cavro_Codec;
  return o;
}

static void __pyx_tp_dealloc_5cavro_Codec(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_Codec) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyMethodDef __pyx_methods_5cavro_Codec[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Codec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_5Codec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_Codec_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_Codec},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Abstract base class for all codecs.  This class is not meant to be used directly.\n\n    Subclasses must be implemented in cython.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_Codec},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_Codec},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_Codec_spec = {
  "cavro.Codec",
  sizeof(struct __pyx_obj_5cavro_Codec),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_Codec_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_Codec = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""Codec", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_Codec), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_Codec, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    Abstract base class for all codecs.  This class is not meant to be used directly.\n\n    Subclasses must be implemented in cython.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_Codec, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_Codec, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro__SnappyCodec __pyx_vtable_5cavro__SnappyCodec;

static PyObject *__pyx_tp_new_5cavro__SnappyCodec(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro__SnappyCodec *p;
  PyObject *o = __pyx_tp_new_5cavro_Codec(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro__SnappyCodec *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_Codec*)__pyx_vtabptr_5cavro__SnappyCodec;
  return o;
}

static PyMethodDef __pyx_methods_5cavro__SnappyCodec[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12_SnappyCodec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12_SnappyCodec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__SnappyCodec_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("\n    Implemented the snappy codec if python-snappy is installed.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro__SnappyCodec},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro__SnappyCodec},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__SnappyCodec_spec = {
  "cavro._SnappyCodec",
  sizeof(struct __pyx_obj_5cavro__SnappyCodec),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_type_5cavro__SnappyCodec_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__SnappyCodec = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_SnappyCodec", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__SnappyCodec), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_Codec, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  PyDoc_STR("\n    Implemented the snappy codec if python-snappy is installed.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__SnappyCodec, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro__SnappyCodec, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro__NullCodec __pyx_vtable_5cavro__NullCodec;

static PyObject *__pyx_tp_new_5cavro__NullCodec(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro__NullCodec *p;
  PyObject *o = __pyx_tp_new_5cavro_Codec(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro__NullCodec *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_Codec*)__pyx_vtabptr_5cavro__NullCodec;
  return o;
}

static PyMethodDef __pyx_methods_5cavro__NullCodec[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_NullCodec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_NullCodec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__NullCodec_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("\n    Implements the null codec (no compression)\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro__NullCodec},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro__NullCodec},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__NullCodec_spec = {
  "cavro._NullCodec",
  sizeof(struct __pyx_obj_5cavro__NullCodec),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro__NullCodec_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__NullCodec = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_NullCodec", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__NullCodec), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_Codec, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    Implements the null codec (no compression)\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__NullCodec, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro__NullCodec, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro__DeflateCodec __pyx_vtable_5cavro__DeflateCodec;

static PyObject *__pyx_tp_new_5cavro__DeflateCodec(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro__DeflateCodec *p;
  PyObject *o = __pyx_tp_new_5cavro_Codec(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro__DeflateCodec *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_Codec*)__pyx_vtabptr_5cavro__DeflateCodec;
  return o;
}

static PyMethodDef __pyx_methods_5cavro__DeflateCodec[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13_DeflateCodec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13_DeflateCodec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__DeflateCodec_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("\n    Implements the deflate codec\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro__DeflateCodec},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro__DeflateCodec},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__DeflateCodec_spec = {
  "cavro._DeflateCodec",
  sizeof(struct __pyx_obj_5cavro__DeflateCodec),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro__DeflateCodec_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__DeflateCodec = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_DeflateCodec", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__DeflateCodec), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_Codec, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    Implements the deflate codec\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__DeflateCodec, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro__DeflateCodec, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro___Bzip2Codec __pyx_vtable_5cavro___Bzip2Codec;

static PyObject *__pyx_tp_new_5cavro___Bzip2Codec(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro___Bzip2Codec *p;
  PyObject *o = __pyx_tp_new_5cavro_Codec(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro___Bzip2Codec *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_Codec*)__pyx_vtabptr_5cavro___Bzip2Codec;
  return o;
}

static PyMethodDef __pyx_methods_5cavro___Bzip2Codec[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12__Bzip2Codec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_12__Bzip2Codec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___Bzip2Codec_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("\n    Implements the bzip2 codec\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro___Bzip2Codec},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___Bzip2Codec},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___Bzip2Codec_spec = {
  "cavro.__Bzip2Codec",
  sizeof(struct __pyx_obj_5cavro___Bzip2Codec),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro___Bzip2Codec_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___Bzip2Codec = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__Bzip2Codec", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___Bzip2Codec), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_Codec, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    Implements the bzip2 codec\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro___Bzip2Codec, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___Bzip2Codec, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro__LzmaCodec __pyx_vtable_5cavro__LzmaCodec;

static PyObject *__pyx_tp_new_5cavro__LzmaCodec(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro__LzmaCodec *p;
  PyObject *o = __pyx_tp_new_5cavro_Codec(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro__LzmaCodec *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_Codec*)__pyx_vtabptr_5cavro__LzmaCodec;
  return o;
}

static PyMethodDef __pyx_methods_5cavro__LzmaCodec[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_LzmaCodec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_10_LzmaCodec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__LzmaCodec_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("\n    Implements the lzma codec\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro__LzmaCodec},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro__LzmaCodec},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__LzmaCodec_spec = {
  "cavro._LzmaCodec",
  sizeof(struct __pyx_obj_5cavro__LzmaCodec),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro__LzmaCodec_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__LzmaCodec = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_LzmaCodec", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__LzmaCodec), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_Codec, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    Implements the lzma codec\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__LzmaCodec, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro__LzmaCodec, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro__ZStandardCodec __pyx_vtable_5cavro__ZStandardCodec;

static PyObject *__pyx_tp_new_5cavro__ZStandardCodec(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro__ZStandardCodec *p;
  PyObject *o = __pyx_tp_new_5cavro_Codec(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro__ZStandardCodec *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_Codec*)__pyx_vtabptr_5cavro__ZStandardCodec;
  p->compressor = Py_None; Py_INCREF(Py_None);
  p->decompressor = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro__ZStandardCodec(PyObject *o) {
  struct __pyx_obj_5cavro__ZStandardCodec *p = (struct __pyx_obj_5cavro__ZStandardCodec *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro__ZStandardCodec) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->compressor);
  Py_CLEAR(p->decompressor);
  __pyx_tp_dealloc_5cavro_Codec(o);
}

static int __pyx_tp_traverse_5cavro__ZStandardCodec(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro__ZStandardCodec *p = (struct __pyx_obj_5cavro__ZStandardCodec *)o;
  e = ((likely(__pyx_ptype_5cavro_Codec)) ? ((__pyx_ptype_5cavro_Codec->tp_traverse) ? __pyx_ptype_5cavro_Codec->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_5cavro__ZStandardCodec)); if (e) return e;
  if (p->compressor) {
    e = (*v)(p->compressor, a); if (e) return e;
  }
  if (p->decompressor) {
    e = (*v)(p->decompressor, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro__ZStandardCodec(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro__ZStandardCodec *p = (struct __pyx_obj_5cavro__ZStandardCodec *)o;
  if (likely(__pyx_ptype_5cavro_Codec)) { if (__pyx_ptype_5cavro_Codec->tp_clear) __pyx_ptype_5cavro_Codec->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_5cavro__ZStandardCodec);
  tmp = ((PyObject*)p->compressor);
  p->compressor = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->decompressor);
  p->decompressor = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_15_ZStandardCodec_compressor(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15_ZStandardCodec_10compressor_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15_ZStandardCodec_decompressor(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15_ZStandardCodec_12decompressor_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro__ZStandardCodec[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15_ZStandardCodec_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15_ZStandardCodec_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro__ZStandardCodec[] = {
  {(char *)"compressor", __pyx_getprop_5cavro_15_ZStandardCodec_compressor, 0, (char *)0, 0},
  {(char *)"decompressor", __pyx_getprop_5cavro_15_ZStandardCodec_decompressor, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__ZStandardCodec_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro__ZStandardCodec},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Implements the zstandard codec if zstandard is installed.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro__ZStandardCodec},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro__ZStandardCodec},
  {Py_tp_methods, (void *)__pyx_methods_5cavro__ZStandardCodec},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro__ZStandardCodec},
  {Py_tp_init, (void *)__pyx_pw_5cavro_15_ZStandardCodec_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro__ZStandardCodec},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__ZStandardCodec_spec = {
  "cavro._ZStandardCodec",
  sizeof(struct __pyx_obj_5cavro__ZStandardCodec),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro__ZStandardCodec_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__ZStandardCodec = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_ZStandardCodec", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__ZStandardCodec), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro__ZStandardCodec, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Implements the zstandard codec if zstandard is installed.\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro__ZStandardCodec, /*tp_traverse*/
  __pyx_tp_clear_5cavro__ZStandardCodec, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__ZStandardCodec, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro__ZStandardCodec, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_15_ZStandardCodec_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro__ZStandardCodec, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro__Lz4Codec __pyx_vtable_5cavro__Lz4Codec;

static PyObject *__pyx_tp_new_5cavro__Lz4Codec(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro__Lz4Codec *p;
  PyObject *o = __pyx_tp_new_5cavro_Codec(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro__Lz4Codec *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_Codec*)__pyx_vtabptr_5cavro__Lz4Codec;
  return o;
}

static PyMethodDef __pyx_methods_5cavro__Lz4Codec[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9_Lz4Codec_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_9_Lz4Codec_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro__Lz4Codec_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("\n    Implements the lz4 codec if lz4 is installed.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro__Lz4Codec},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro__Lz4Codec},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro__Lz4Codec_spec = {
  "cavro._Lz4Codec",
  sizeof(struct __pyx_obj_5cavro__Lz4Codec),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro__Lz4Codec_slots,
};
#else

static PyTypeObject __pyx_type_5cavro__Lz4Codec = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_Lz4Codec", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro__Lz4Codec), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_Codec, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    Implements the lz4 codec if lz4 is installed.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro__Lz4Codec, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro__Lz4Codec, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_ContainerReader __pyx_vtable_5cavro_ContainerReader;

static PyObject *__pyx_tp_new_5cavro_ContainerReader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5cavro_ContainerReader *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5cavro_ContainerReader *)o);
  p->__pyx_vtab = __pyx_vtabptr_5cavro_ContainerReader;
  p->metadata = Py_None; Py_INCREF(Py_None);
  p->writer_schema = ((struct __pyx_obj_5cavro_Schema *)Py_None); Py_INCREF(Py_None);
  p->reader_schema = ((struct __pyx_obj_5cavro_Schema *)Py_None); Py_INCREF(Py_None);
  p->schema = ((struct __pyx_obj_5cavro_Schema *)Py_None); Py_INCREF(Py_None);
  p->codec_name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->current_block = ((struct __pyx_obj_5cavro_MemoryReader *)Py_None); Py_INCREF(Py_None);
  p->codec = ((struct __pyx_obj_5cavro_Codec *)Py_None); Py_INCREF(Py_None);
  p->reader = ((struct __pyx_obj_5cavro__Reader *)Py_None); Py_INCREF(Py_None);
  p->marker.data = NULL;
  p->marker.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_5cavro_ContainerReader(PyObject *o) {
  struct __pyx_obj_5cavro_ContainerReader *p = (struct __pyx_obj_5cavro_ContainerReader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_ContainerReader) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->metadata);
  Py_CLEAR(p->writer_schema);
  Py_CLEAR(p->reader_schema);
  Py_CLEAR(p->schema);
  Py_CLEAR(p->codec_name);
  Py_CLEAR(p->current_block);
  Py_CLEAR(p->codec);
  Py_CLEAR(p->reader);
  __PYX_XCLEAR_MEMVIEW(&p->marker, 1);
  p->marker.memview = NULL; p->marker.data = NULL;
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_5cavro_ContainerReader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_ContainerReader *p = (struct __pyx_obj_5cavro_ContainerReader *)o;
  if (p->metadata) {
    e = (*v)(p->metadata, a); if (e) return e;
  }
  if (p->writer_schema) {
    e = (*v)(((PyObject *)p->writer_schema), a); if (e) return e;
  }
  if (p->reader_schema) {
    e = (*v)(((PyObject *)p->reader_schema), a); if (e) return e;
  }
  if (p->schema) {
    e = (*v)(((PyObject *)p->schema), a); if (e) return e;
  }
  if (p->current_block) {
    e = (*v)(((PyObject *)p->current_block), a); if (e) return e;
  }
  if (p->codec) {
    e = (*v)(((PyObject *)p->codec), a); if (e) return e;
  }
  if (p->reader) {
    e = (*v)(((PyObject *)p->reader), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_ContainerReader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_ContainerReader *p = (struct __pyx_obj_5cavro_ContainerReader *)o;
  tmp = ((PyObject*)p->metadata);
  p->metadata = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->writer_schema);
  p->writer_schema = ((struct __pyx_obj_5cavro_Schema *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->reader_schema);
  p->reader_schema = ((struct __pyx_obj_5cavro_Schema *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->schema);
  p->schema = ((struct __pyx_obj_5cavro_Schema *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_block);
  p->current_block = ((struct __pyx_obj_5cavro_MemoryReader *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->codec);
  p->codec = ((struct __pyx_obj_5cavro_Codec *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->reader);
  p->reader = ((struct __pyx_obj_5cavro__Reader *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_15ContainerReader_metadata(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerReader_8metadata_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerReader_marker(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerReader_6marker_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerReader_writer_schema(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerReader_13writer_schema_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerReader_reader_schema(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerReader_13reader_schema_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerReader_schema(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerReader_6schema_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerReader_codec_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerReader_10codec_name_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerReader_objects_left_in_block(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerReader_21objects_left_in_block_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_5cavro_15ContainerReader_9__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  PyObject *res = __pyx_pw_5cavro_15ContainerReader_9__next__(self);
  if (!res && !PyErr_Occurred()) { PyErr_SetNone(PyExc_StopIteration); }
  return res;
}

static PyMethodDef __pyx_methods_5cavro_ContainerReader[] = {
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_5cavro_15ContainerReader_9__next__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerReader_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerReader_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_ContainerReader[] = {
  {(char *)"metadata", __pyx_getprop_5cavro_15ContainerReader_metadata, 0, (char *)0, 0},
  {(char *)"marker", __pyx_getprop_5cavro_15ContainerReader_marker, 0, (char *)0, 0},
  {(char *)"writer_schema", __pyx_getprop_5cavro_15ContainerReader_writer_schema, 0, (char *)0, 0},
  {(char *)"reader_schema", __pyx_getprop_5cavro_15ContainerReader_reader_schema, 0, (char *)0, 0},
  {(char *)"schema", __pyx_getprop_5cavro_15ContainerReader_schema, 0, (char *)0, 0},
  {(char *)"codec_name", __pyx_getprop_5cavro_15ContainerReader_codec_name, 0, (char *)0, 0},
  {(char *)"objects_left_in_block", __pyx_getprop_5cavro_15ContainerReader_objects_left_in_block, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_ContainerReader_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_ContainerReader},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A class for reading avro object container files.\n\n    The container can ben used as an iterator, in which case it will yield the objects in the file in order:\n    ```\n    for obj in ContainerReader('file.avro'):\n        print(obj)\n    ```\n\n    Arguments:\n     * `src`: The source to read from. Can be a file-like object, instance of `cavro.MemoryReader`, or a path to a file (str|Path)\n     * `reader_schema`: The schema to use when reading objects. If not provided, the writer schema will be used.\n     * `options`: An Options object to use when constructing the writer schema. Defaults to the default options. This does not affect the `reader_schema` options.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_ContainerReader},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_ContainerReader},
  {Py_tp_iter, (void *)__pyx_pw_5cavro_15ContainerReader_7__iter__},
  {Py_tp_iternext, (void *)__pyx_pw_5cavro_15ContainerReader_9__next__},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_ContainerReader},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_ContainerReader},
  {Py_tp_init, (void *)__pyx_pw_5cavro_15ContainerReader_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_ContainerReader},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_ContainerReader_spec = {
  "cavro.ContainerReader",
  sizeof(struct __pyx_obj_5cavro_ContainerReader),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_ContainerReader_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_ContainerReader = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""ContainerReader", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_ContainerReader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ContainerReader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    A class for reading avro object container files.\n\n    The container can ben used as an iterator, in which case it will yield the objects in the file in order:\n    ```\n    for obj in ContainerReader('file.avro'):\n        print(obj)\n    ```\n\n    Arguments:\n     * `src`: The source to read from. Can be a file-like object, instance of `cavro.MemoryReader`, or a path to a file (str|Path)\n     * `reader_schema`: The schema to use when reading objects. If not provided, the writer schema will be used.\n     * `options`: An Options object to use when constructing the writer schema. Defaults to the default options. This does not affect the `reader_schema` options.\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_ContainerReader, /*tp_traverse*/
  __pyx_tp_clear_5cavro_ContainerReader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_5cavro_15ContainerReader_7__iter__, /*tp_iter*/
  __pyx_pw_5cavro_15ContainerReader_9__next__, /*tp_iternext*/
  __pyx_methods_5cavro_ContainerReader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_ContainerReader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_15ContainerReader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_ContainerReader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_ContainerWriter __pyx_vtable_5cavro_ContainerWriter;

static PyObject *__pyx_tp_new_5cavro_ContainerWriter(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_ContainerWriter *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_5cavro_ContainerWriter *)o);
  p->__pyx_vtab = __pyx_vtabptr_5cavro_ContainerWriter;
  p->writer = ((struct __pyx_obj_5cavro__Writer *)Py_None); Py_INCREF(Py_None);
  p->pending_block = ((struct __pyx_obj_5cavro_MemoryWriter *)Py_None); Py_INCREF(Py_None);
  p->next_block = ((struct __pyx_obj_5cavro_MemoryWriter *)Py_None); Py_INCREF(Py_None);
  p->next_item = ((struct __pyx_obj_5cavro_MemoryWriter *)Py_None); Py_INCREF(Py_None);
  p->schema = ((struct __pyx_obj_5cavro_Schema *)Py_None); Py_INCREF(Py_None);
  p->codec = ((struct __pyx_obj_5cavro_Codec *)Py_None); Py_INCREF(Py_None);
  p->options = ((struct __pyx_obj_5cavro_Options *)Py_None); Py_INCREF(Py_None);
  p->metadata = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->marker.data = NULL;
  p->marker.memview = NULL;
  if (unlikely(__pyx_pw_5cavro_15ContainerWriter_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

#if CYTHON_USE_TP_FINALIZE
static void __pyx_tp_finalize_5cavro_ContainerWriter(PyObject *o) {
  PyObject *etype, *eval, *etb;
  PyErr_Fetch(&etype, &eval, &etb);
  __pyx_pw_5cavro_15ContainerWriter_7__del__(o);
  PyErr_Restore(etype, eval, etb);
}
#endif

static void __pyx_tp_dealloc_5cavro_ContainerWriter(PyObject *o) {
  struct __pyx_obj_5cavro_ContainerWriter *p = (struct __pyx_obj_5cavro_ContainerWriter *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_ContainerWriter) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->writer);
  Py_CLEAR(p->pending_block);
  Py_CLEAR(p->next_block);
  Py_CLEAR(p->next_item);
  Py_CLEAR(p->schema);
  Py_CLEAR(p->codec);
  Py_CLEAR(p->options);
  Py_CLEAR(p->metadata);
  __PYX_XCLEAR_MEMVIEW(&p->marker, 1);
  p->marker.memview = NULL; p->marker.data = NULL;
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_5cavro_ContainerWriter(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_ContainerWriter *p = (struct __pyx_obj_5cavro_ContainerWriter *)o;
  if (p->writer) {
    e = (*v)(((PyObject *)p->writer), a); if (e) return e;
  }
  if (p->pending_block) {
    e = (*v)(((PyObject *)p->pending_block), a); if (e) return e;
  }
  if (p->next_block) {
    e = (*v)(((PyObject *)p->next_block), a); if (e) return e;
  }
  if (p->next_item) {
    e = (*v)(((PyObject *)p->next_item), a); if (e) return e;
  }
  if (p->schema) {
    e = (*v)(((PyObject *)p->schema), a); if (e) return e;
  }
  if (p->codec) {
    e = (*v)(((PyObject *)p->codec), a); if (e) return e;
  }
  if (p->options) {
    e = (*v)(((PyObject *)p->options), a); if (e) return e;
  }
  if (p->metadata) {
    e = (*v)(p->metadata, a); if (e) return e;
  }
  return 0;
}

static PyObject *__pyx_getprop_5cavro_15ContainerWriter_closed(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerWriter_6closed_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerWriter_schema(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerWriter_6schema_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerWriter_codec(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerWriter_5codec_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerWriter_marker(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerWriter_6marker_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerWriter_max_blocksize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerWriter_13max_blocksize_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerWriter_options(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerWriter_7options_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerWriter_should_write_header(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerWriter_19should_write_header_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerWriter_num_pending(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerWriter_11num_pending_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerWriter_blocks_written(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerWriter_14blocks_written_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_15ContainerWriter_metadata(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_15ContainerWriter_8metadata_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_ContainerWriter[] = {
  {"__enter__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_3__enter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__exit__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_5__exit__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_9close, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"flush", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_11flush, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"write_many", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_15write_many, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_17__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_15ContainerWriter_19__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_ContainerWriter[] = {
  {(char *)"closed", __pyx_getprop_5cavro_15ContainerWriter_closed, 0, (char *)0, 0},
  {(char *)"schema", __pyx_getprop_5cavro_15ContainerWriter_schema, 0, (char *)0, 0},
  {(char *)"codec", __pyx_getprop_5cavro_15ContainerWriter_codec, 0, (char *)0, 0},
  {(char *)"marker", __pyx_getprop_5cavro_15ContainerWriter_marker, 0, (char *)0, 0},
  {(char *)"max_blocksize", __pyx_getprop_5cavro_15ContainerWriter_max_blocksize, 0, (char *)0, 0},
  {(char *)"options", __pyx_getprop_5cavro_15ContainerWriter_options, 0, (char *)0, 0},
  {(char *)"should_write_header", __pyx_getprop_5cavro_15ContainerWriter_should_write_header, 0, (char *)0, 0},
  {(char *)"num_pending", __pyx_getprop_5cavro_15ContainerWriter_num_pending, 0, (char *)0, 0},
  {(char *)"blocks_written", __pyx_getprop_5cavro_15ContainerWriter_blocks_written, 0, (char *)0, 0},
  {(char *)"metadata", __pyx_getprop_5cavro_15ContainerWriter_metadata, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_ContainerWriter_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_ContainerWriter},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A class for writing avro object container files.\n\n    The writer can be used as a context manager, in which case it will be closed when the context exits:\n    ```\n    with ContainerWriter('file.avro', schema) as writer:\n        writer.write_one(obj)\n    ```\n\n    Arguments:\n     * `dest`: The destination to write to. Can be a file-like object, instance of `cavro.MemoryWriter`, or a path to a file (str|Path)\n     * `schema`: The schema of the objects to be written.\n     * `codec`: The codec to use. Must be one of the supported codecs. Default to `null`\n     * `max_blocksize`: The maximum size of a block. Defaults to `16352`.\n     * `write_header`: Whether to write the avro header to the file before writing blocks. Defaults to `True`.\n     * `metadata`: A dictionary of metadata to write to the file. Defaults to an empty dictionary.\n     * `marker`: A 16-byte marker to use to separate blocks. Defaults to a random UUID.\n     * `options`: An Options object to use when writing. Defaults to the default options.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_ContainerWriter},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_ContainerWriter},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_ContainerWriter},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_ContainerWriter},
  #if PY_VERSION_HEX >= 0x030400a1
  {Py_tp_finalize, (void *)__pyx_tp_finalize_5cavro_ContainerWriter},
  #endif
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_ContainerWriter_spec = {
  "cavro.ContainerWriter",
  sizeof(struct __pyx_obj_5cavro_ContainerWriter),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro_ContainerWriter_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_ContainerWriter = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""ContainerWriter", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_ContainerWriter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ContainerWriter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  PyDoc_STR("\n    A class for writing avro object container files.\n\n    The writer can be used as a context manager, in which case it will be closed when the context exits:\n    ```\n    with ContainerWriter('file.avro', schema) as writer:\n        writer.write_one(obj)\n    ```\n\n    Arguments:\n     * `dest`: The destination to write to. Can be a file-like object, instance of `cavro.MemoryWriter`, or a path to a file (str|Path)\n     * `schema`: The schema of the objects to be written.\n     * `codec`: The codec to use. Must be one of the supported codecs. Default to `null`\n     * `max_blocksize`: The maximum size of a block. Defaults to `16352`.\n     * `write_header`: Whether to write the avro header to the file before writing blocks. Defaults to `True`.\n     * `metadata`: A dictionary of metadata to write to the file. Defaults to an empty dictionary.\n     * `marker`: A 16-byte marker to use to separate blocks. Defaults to a random UUID.\n     * `options`: An Options object to use when writing. Defaults to the default options.\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_ContainerWriter, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_ContainerWriter, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_ContainerWriter, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_ContainerWriter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  __pyx_tp_finalize_5cavro_ContainerWriter, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_PromoteToFloat __pyx_vtable_5cavro_PromoteToFloat;

static PyObject *__pyx_tp_new_5cavro_PromoteToFloat(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_PromoteToFloat *p;
  PyObject *o = __pyx_tp_new_5cavro_ValueAdapter(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_PromoteToFloat *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_PromoteToFloat;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_PromoteToFloat[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14PromoteToFloat_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_14PromoteToFloat_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_PromoteToFloat_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("\n    A value adapter that converts a value to a float on read.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_PromoteToFloat},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_PromoteToFloat},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_PromoteToFloat_spec = {
  "cavro.PromoteToFloat",
  sizeof(struct __pyx_obj_5cavro_PromoteToFloat),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_PromoteToFloat_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_PromoteToFloat = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""PromoteToFloat", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_PromoteToFloat), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    A value adapter that converts a value to a float on read.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_PromoteToFloat, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_PromoteToFloat, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_PromoteBytesToString __pyx_vtable_5cavro_PromoteBytesToString;

static PyObject *__pyx_tp_new_5cavro_PromoteBytesToString(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_PromoteBytesToString *p;
  PyObject *o = __pyx_tp_new_5cavro_ValueAdapter(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_PromoteBytesToString *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_PromoteBytesToString;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_PromoteBytesToString[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_20PromoteBytesToString_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_20PromoteBytesToString_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_PromoteBytesToString_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("\n    A value adapter that decodes bytes to a string (utf8) on read.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_PromoteBytesToString},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_PromoteBytesToString},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_PromoteBytesToString_spec = {
  "cavro.PromoteBytesToString",
  sizeof(struct __pyx_obj_5cavro_PromoteBytesToString),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_PromoteBytesToString_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_PromoteBytesToString = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""PromoteBytesToString", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_PromoteBytesToString), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    A value adapter that decodes bytes to a string (utf8) on read.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_PromoteBytesToString, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_PromoteBytesToString, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_PromoteStringToBytes __pyx_vtable_5cavro_PromoteStringToBytes;

static PyObject *__pyx_tp_new_5cavro_PromoteStringToBytes(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_PromoteStringToBytes *p;
  PyObject *o = __pyx_tp_new_5cavro_ValueAdapter(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_PromoteStringToBytes *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_PromoteStringToBytes;
  return o;
}

static PyMethodDef __pyx_methods_5cavro_PromoteStringToBytes[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_20PromoteStringToBytes_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_20PromoteStringToBytes_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_PromoteStringToBytes_slots[] = {
  {Py_tp_doc, (void *)PyDoc_STR("\n    A value adapter that encodes a string to bytes (utf8) on read.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_PromoteStringToBytes},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_PromoteStringToBytes},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_PromoteStringToBytes_spec = {
  "cavro.PromoteStringToBytes",
  sizeof(struct __pyx_obj_5cavro_PromoteStringToBytes),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_5cavro_PromoteStringToBytes_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_PromoteStringToBytes = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""PromoteStringToBytes", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_PromoteStringToBytes), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_ValueAdapter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    A value adapter that encodes a string to bytes (utf8) on read.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_PromoteStringToBytes, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_PromoteStringToBytes, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_5cavro_CannotPromote __pyx_vtable_5cavro_CannotPromote;

static PyObject *__pyx_tp_new_5cavro_CannotPromote(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5cavro_CannotPromote *p;
  PyObject *o = __pyx_tp_new_5cavro_ValueAdapter(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5cavro_CannotPromote *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5cavro_ValueAdapter*)__pyx_vtabptr_5cavro_CannotPromote;
  p->reader_type = ((struct __pyx_obj_5cavro_AvroType *)Py_None); Py_INCREF(Py_None);
  p->writer_type = ((struct __pyx_obj_5cavro_AvroType *)Py_None); Py_INCREF(Py_None);
  p->extra = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5cavro_CannotPromote(PyObject *o) {
  struct __pyx_obj_5cavro_CannotPromote *p = (struct __pyx_obj_5cavro_CannotPromote *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro_CannotPromote) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->reader_type);
  Py_CLEAR(p->writer_type);
  Py_CLEAR(p->extra);
  __pyx_tp_dealloc_5cavro_ValueAdapter(o);
}

static int __pyx_tp_traverse_5cavro_CannotPromote(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro_CannotPromote *p = (struct __pyx_obj_5cavro_CannotPromote *)o;
  e = ((likely(__pyx_ptype_5cavro_ValueAdapter)) ? ((__pyx_ptype_5cavro_ValueAdapter->tp_traverse) ? __pyx_ptype_5cavro_ValueAdapter->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_5cavro_CannotPromote)); if (e) return e;
  if (p->reader_type) {
    e = (*v)(((PyObject *)p->reader_type), a); if (e) return e;
  }
  if (p->writer_type) {
    e = (*v)(((PyObject *)p->writer_type), a); if (e) return e;
  }
  if (p->extra) {
    e = (*v)(p->extra, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro_CannotPromote(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro_CannotPromote *p = (struct __pyx_obj_5cavro_CannotPromote *)o;
  if (likely(__pyx_ptype_5cavro_ValueAdapter)) { if (__pyx_ptype_5cavro_ValueAdapter->tp_clear) __pyx_ptype_5cavro_ValueAdapter->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_5cavro_CannotPromote);
  tmp = ((PyObject*)p->reader_type);
  p->reader_type = ((struct __pyx_obj_5cavro_AvroType *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->writer_type);
  p->writer_type = ((struct __pyx_obj_5cavro_AvroType *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->extra);
  p->extra = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5cavro_13CannotPromote_reader_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_13CannotPromote_11reader_type_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_13CannotPromote_writer_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_13CannotPromote_11writer_type_1__get__(o);
}

static PyObject *__pyx_getprop_5cavro_13CannotPromote_extra(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5cavro_13CannotPromote_5extra_1__get__(o);
}

static PyMethodDef __pyx_methods_5cavro_CannotPromote[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13CannotPromote_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_5cavro_13CannotPromote_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5cavro_CannotPromote[] = {
  {(char *)"reader_type", __pyx_getprop_5cavro_13CannotPromote_reader_type, 0, (char *)0, 0},
  {(char *)"writer_type", __pyx_getprop_5cavro_13CannotPromote_writer_type, 0, (char *)0, 0},
  {(char *)"extra", __pyx_getprop_5cavro_13CannotPromote_extra, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro_CannotPromote_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro_CannotPromote},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A captured schema promotion error that has been deferred by `Options`, the first time this value is read, the error will be raised.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro_CannotPromote},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro_CannotPromote},
  {Py_tp_methods, (void *)__pyx_methods_5cavro_CannotPromote},
  {Py_tp_getset, (void *)__pyx_getsets_5cavro_CannotPromote},
  {Py_tp_init, (void *)__pyx_pw_5cavro_13CannotPromote_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro_CannotPromote},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro_CannotPromote_spec = {
  "cavro.CannotPromote",
  sizeof(struct __pyx_obj_5cavro_CannotPromote),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_5cavro_CannotPromote_slots,
};
#else

static PyTypeObject __pyx_type_5cavro_CannotPromote = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""CannotPromote", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro_CannotPromote), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro_CannotPromote, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    A captured schema promotion error that has been deferred by `Options`, the first time this value is read, the error will be raised.\n    "), /*tp_doc*/
  __pyx_tp_traverse_5cavro_CannotPromote, /*tp_traverse*/
  __pyx_tp_clear_5cavro_CannotPromote, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5cavro_CannotPromote, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5cavro_CannotPromote, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_5cavro_13CannotPromote_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro_CannotPromote, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct__genexpr *__pyx_freelist_5cavro___pyx_scope_struct__genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct__genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct__genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct__genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct__genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct__genexpr[--__pyx_freecount_5cavro___pyx_scope_struct__genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct__genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct__genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct__genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct__genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct__genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_a);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct__genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct__genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct__genexpr[__pyx_freecount_5cavro___pyx_scope_struct__genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct__genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct__genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct__genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct__genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_a) {
    e = (*v)(p->__pyx_v_a, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct__genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct__genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct__genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct__genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct__genexpr_spec = {
  "cavro.__pyx_scope_struct__genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct__genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct__genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct__genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct__genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct__genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct__genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct__genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct__genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_1_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_1_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_1_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_a);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_1_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_1_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_a) {
    e = (*v)(p->__pyx_v_a, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_1_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_1_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_1_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_1_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_1_genexpr_spec = {
  "cavro.__pyx_scope_struct_1_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_1_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_2_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_2_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_2_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_2_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_2_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_2_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_2_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_2_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_p);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_2_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_2_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_2_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_2_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_p) {
    e = (*v)(p->__pyx_v_p, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_2_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_2_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_2_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_2_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_2_genexpr_spec = {
  "cavro.__pyx_scope_struct_2_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_2_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_2_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_2_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_2_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_2_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_2_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_2_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types *__pyx_freelist_5cavro___pyx_scope_struct_3_walk_types[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_3_walk_types = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_3_walk_types(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_3_walk_types > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_3_walk_types[--__pyx_freecount_5cavro___pyx_scope_struct_3_walk_types];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_3_walk_types(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types *p = (struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_3_walk_types) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_visited);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_3_walk_types < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_3_walk_types[__pyx_freecount_5cavro___pyx_scope_struct_3_walk_types++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_3_walk_types(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types *p = (struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_visited) {
    e = (*v)(p->__pyx_v_visited, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_3_walk_types_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_3_walk_types},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_3_walk_types},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_3_walk_types},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_3_walk_types_spec = {
  "cavro.__pyx_scope_struct_3_walk_types",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_3_walk_types_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_3_walk_types = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_3_walk_types", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_3_walk_types), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_3_walk_types, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_3_walk_types, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_3_walk_types, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *__pyx_freelist_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer[--__pyx_freecount_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *p = (struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_reader_symbols);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer[__pyx_freecount_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *p = (struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *)o;
  if (p->__pyx_v_reader_symbols) {
    e = (*v)(p->__pyx_v_reader_symbols, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *p = (struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer *)o;
  tmp = ((PyObject*)p->__pyx_v_reader_symbols);
  p->__pyx_v_reader_symbols = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_5cavro_EnumType *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer_spec = {
  "cavro.__pyx_scope_struct_4___pyx_base____pyx_base___for_writer",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_4___pyx_base____pyx_base___for_writer", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer, /*tp_traverse*/
  __pyx_tp_clear_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_5_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_5_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_5_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_5_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_5_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_5_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_5_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_5_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_s);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_5_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_5_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_5_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_5_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_5_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_5_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_5_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_5_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_5_genexpr_spec = {
  "cavro.__pyx_scope_struct_5_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_5_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_5_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_5_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_5_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_5_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_5_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_5_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types *__pyx_freelist_5cavro___pyx_scope_struct_6_walk_types[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_6_walk_types = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_6_walk_types(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_6_walk_types > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_6_walk_types[--__pyx_freecount_5cavro___pyx_scope_struct_6_walk_types];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_6_walk_types(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types *p = (struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_6_walk_types) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_visited);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_6_walk_types < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_6_walk_types[__pyx_freecount_5cavro___pyx_scope_struct_6_walk_types++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_6_walk_types(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types *p = (struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_visited) {
    e = (*v)(p->__pyx_v_visited, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_6_walk_types_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_6_walk_types},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_6_walk_types},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_6_walk_types},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_6_walk_types_spec = {
  "cavro.__pyx_scope_struct_6_walk_types",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_6_walk_types_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_6_walk_types = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_6_walk_types", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_6_walk_types), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_6_walk_types, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_6_walk_types, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_6_walk_types, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_7_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_7_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_7_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_7_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_7_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_7_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_7_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_7_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_f);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_7_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_7_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_7_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_7_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_f) {
    e = (*v)(p->__pyx_v_f, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_7_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_7_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_7_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_7_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_7_genexpr_spec = {
  "cavro.__pyx_scope_struct_7_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_7_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_7_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_7_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_7_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_7_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_7_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_7_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *__pyx_freelist_5cavro___pyx_scope_struct_8___init__[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_8___init__ = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_8___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_8___init__ > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_8___init__)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_8___init__[--__pyx_freecount_5cavro___pyx_scope_struct_8___init__];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_8___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_8___init__(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *p = (struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_8___init__) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_schema_);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_8___init__ < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_8___init__)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_8___init__[__pyx_freecount_5cavro___pyx_scope_struct_8___init__++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_8___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *p = (struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *)o;
  if (p->__pyx_v_schema_) {
    e = (*v)(((PyObject *)p->__pyx_v_schema_), a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro___pyx_scope_struct_8___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *p = (struct __pyx_obj_5cavro___pyx_scope_struct_8___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_schema_);
  p->__pyx_v_schema_ = ((struct __pyx_obj_5cavro_Schema *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_5cavro_RecordType *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_8___init___slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_8___init__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_8___init__},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro___pyx_scope_struct_8___init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_8___init__},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_8___init___spec = {
  "cavro.__pyx_scope_struct_8___init__",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_8___init__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_8___init___slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_8___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_8___init__", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_8___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_8___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_8___init__, /*tp_traverse*/
  __pyx_tp_clear_5cavro___pyx_scope_struct_8___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_8___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_9_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_9_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_9_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_9_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_9_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_9_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_9_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_9_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_f);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_9_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_9_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_9_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_9_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_f) {
    e = (*v)(p->__pyx_v_f, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_9_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_9_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_9_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_9_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_9_genexpr_spec = {
  "cavro.__pyx_scope_struct_9_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_9_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_9_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_9_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_9_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_9_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_9_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_9_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types *__pyx_freelist_5cavro___pyx_scope_struct_10_walk_types[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_10_walk_types = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_10_walk_types(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_10_walk_types > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_10_walk_types[--__pyx_freecount_5cavro___pyx_scope_struct_10_walk_types];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_10_walk_types(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types *p = (struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_10_walk_types) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_field);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_visited);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_10_walk_types < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_10_walk_types[__pyx_freecount_5cavro___pyx_scope_struct_10_walk_types++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_10_walk_types(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types *p = (struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types *)o;
  if (p->__pyx_v_field) {
    e = (*v)(p->__pyx_v_field, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_visited) {
    e = (*v)(p->__pyx_v_visited, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_10_walk_types_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_10_walk_types},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_10_walk_types},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_10_walk_types},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_10_walk_types_spec = {
  "cavro.__pyx_scope_struct_10_walk_types",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_10_walk_types_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_10_walk_types = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_10_walk_types", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_10_walk_types), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_10_walk_types, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_10_walk_types, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_10_walk_types, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *__pyx_freelist_5cavro___pyx_scope_struct_11___init__[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_11___init__ = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_11___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_11___init__ > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_11___init__)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_11___init__[--__pyx_freecount_5cavro___pyx_scope_struct_11___init__];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_11___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_11___init__(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *p = (struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_11___init__) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_is_ambiguous);
  Py_CLEAR(p->__pyx_v_namespace);
  Py_CLEAR(p->__pyx_v_schema);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_11___init__ < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_11___init__)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_11___init__[__pyx_freecount_5cavro___pyx_scope_struct_11___init__++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_11___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *p = (struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *)o;
  if (p->__pyx_v_is_ambiguous) {
    e = (*v)(p->__pyx_v_is_ambiguous, a); if (e) return e;
  }
  if (p->__pyx_v_namespace) {
    e = (*v)(p->__pyx_v_namespace, a); if (e) return e;
  }
  if (p->__pyx_v_schema) {
    e = (*v)(p->__pyx_v_schema, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro___pyx_scope_struct_11___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *p = (struct __pyx_obj_5cavro___pyx_scope_struct_11___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_is_ambiguous);
  p->__pyx_v_is_ambiguous = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_namespace);
  p->__pyx_v_namespace = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_schema);
  p->__pyx_v_schema = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_11___init___slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_11___init__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_11___init__},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro___pyx_scope_struct_11___init__},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_11___init__},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_11___init___spec = {
  "cavro.__pyx_scope_struct_11___init__",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_11___init__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_11___init___slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_11___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_11___init__", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_11___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_11___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_11___init__, /*tp_traverse*/
  __pyx_tp_clear_5cavro___pyx_scope_struct_11___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_11___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_12_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_12_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_12_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_12_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_12_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_12_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_12_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_12_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_s);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_12_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_12_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_12_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_12_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_12_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_12_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_12_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_12_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_12_genexpr_spec = {
  "cavro.__pyx_scope_struct_12_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_12_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_12_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_12_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_12_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_12_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_12_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_12_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_13_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_13_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_13_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_13_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_13_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_13_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_13_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_13_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v__);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_13_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_13_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_13_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_13_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v__) {
    e = (*v)(p->__pyx_v__, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_13_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_13_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_13_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_13_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_13_genexpr_spec = {
  "cavro.__pyx_scope_struct_13_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_13_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_13_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_13_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_13_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_13_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_13_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_13_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_14_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_14_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_14_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_14_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_14_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_14_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_14_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_14_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v__);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_14_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_14_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_14_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_14_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v__) {
    e = (*v)(p->__pyx_v__, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_14_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_14_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_14_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_14_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_14_genexpr_spec = {
  "cavro.__pyx_scope_struct_14_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_14_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_14_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_14_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_14_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_14_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_14_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_14_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_15_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_15_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_15_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_15_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_15_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_15_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_15_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_15_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_t);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_15_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_15_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_15_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_15_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_t) {
    e = (*v)(p->__pyx_v_t, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_15_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_15_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_15_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_15_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_15_genexpr_spec = {
  "cavro.__pyx_scope_struct_15_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_15_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_15_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_15_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_15_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_15_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_15_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_15_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_16_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_16_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_16_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_16_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_16_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_16_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_16_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_16_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_ty);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_16_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_16_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_16_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_16_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_ty) {
    e = (*v)(p->__pyx_v_ty, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_16_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_16_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_16_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_16_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_16_genexpr_spec = {
  "cavro.__pyx_scope_struct_16_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_16_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_16_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_16_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_16_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_16_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_16_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_16_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_17_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_17_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_17_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_17_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_17_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_17_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_17_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_17_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_ty);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_17_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_17_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_17_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_17_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_ty) {
    e = (*v)(p->__pyx_v_ty, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_17_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_17_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_17_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_17_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_17_genexpr_spec = {
  "cavro.__pyx_scope_struct_17_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_17_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_17_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_17_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_17_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_17_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_17_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_17_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_18_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_18_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_18_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_18_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_18_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_18_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_18_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_18_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v__);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_18_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_18_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_18_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_18_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v__) {
    e = (*v)(p->__pyx_v__, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_18_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_18_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_18_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_18_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_18_genexpr_spec = {
  "cavro.__pyx_scope_struct_18_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_18_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_18_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_18_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_18_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_18_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_18_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_18_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types *__pyx_freelist_5cavro___pyx_scope_struct_19_walk_types[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_19_walk_types = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_19_walk_types(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_19_walk_types > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_19_walk_types[--__pyx_freecount_5cavro___pyx_scope_struct_19_walk_types];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_19_walk_types(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types *p = (struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_19_walk_types) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_t);
  Py_CLEAR(p->__pyx_v_visited);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_19_walk_types < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_19_walk_types[__pyx_freecount_5cavro___pyx_scope_struct_19_walk_types++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_19_walk_types(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types *p = (struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_t) {
    e = (*v)(p->__pyx_v_t, a); if (e) return e;
  }
  if (p->__pyx_v_visited) {
    e = (*v)(p->__pyx_v_visited, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_19_walk_types_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_19_walk_types},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_19_walk_types},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_19_walk_types},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_19_walk_types_spec = {
  "cavro.__pyx_scope_struct_19_walk_types",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_19_walk_types_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_19_walk_types = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_19_walk_types", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_19_walk_types), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_19_walk_types, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_19_walk_types, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_19_walk_types, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_20_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_20_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_20_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_20_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_20_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_20_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_20_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_20_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_p);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_20_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_20_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_20_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_20_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_p) {
    e = (*v)(p->__pyx_v_p, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_20_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_20_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_20_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_20_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_20_genexpr_spec = {
  "cavro.__pyx_scope_struct_20_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_20_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_20_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_20_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_20_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_20_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_20_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_20_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types *__pyx_freelist_5cavro___pyx_scope_struct_21_walk_types[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_21_walk_types = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_21_walk_types(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_21_walk_types > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_21_walk_types[--__pyx_freecount_5cavro___pyx_scope_struct_21_walk_types];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_21_walk_types(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types *p = (struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_21_walk_types) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_visited);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_21_walk_types < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_21_walk_types[__pyx_freecount_5cavro___pyx_scope_struct_21_walk_types++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_21_walk_types(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types *p = (struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_visited) {
    e = (*v)(p->__pyx_v_visited, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_21_walk_types_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_21_walk_types},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_21_walk_types},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_21_walk_types},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_21_walk_types_spec = {
  "cavro.__pyx_scope_struct_21_walk_types",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_21_walk_types_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_21_walk_types = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_21_walk_types", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_21_walk_types), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_21_walk_types, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_21_walk_types, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_21_walk_types, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *__pyx_freelist_5cavro___pyx_scope_struct_22_get_namespaced_aliases[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_22_get_namespaced_aliases = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_22_get_namespaced_aliases(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_22_get_namespaced_aliases > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_22_get_namespaced_aliases[--__pyx_freecount_5cavro___pyx_scope_struct_22_get_namespaced_aliases];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_22_get_namespaced_aliases(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *p = (struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_22_get_namespaced_aliases) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_22_get_namespaced_aliases < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_22_get_namespaced_aliases[__pyx_freecount_5cavro___pyx_scope_struct_22_get_namespaced_aliases++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_22_get_namespaced_aliases(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *p = (struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5cavro___pyx_scope_struct_22_get_namespaced_aliases(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *p = (struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_5cavro__NamedType *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_22_get_namespaced_aliases_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_22_get_namespaced_aliases},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_22_get_namespaced_aliases},
  {Py_tp_clear, (void *)__pyx_tp_clear_5cavro___pyx_scope_struct_22_get_namespaced_aliases},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_22_get_namespaced_aliases},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_22_get_namespaced_aliases_spec = {
  "cavro.__pyx_scope_struct_22_get_namespaced_aliases",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_22_get_namespaced_aliases_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_22_get_namespaced_aliases = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_22_get_namespaced_aliases", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_22_get_namespaced_aliases), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_22_get_namespaced_aliases, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_22_get_namespaced_aliases, /*tp_traverse*/
  __pyx_tp_clear_5cavro___pyx_scope_struct_22_get_namespaced_aliases, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_22_get_namespaced_aliases, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr *__pyx_freelist_5cavro___pyx_scope_struct_23_genexpr[8];
static int __pyx_freecount_5cavro___pyx_scope_struct_23_genexpr = 0;

static PyObject *__pyx_tp_new_5cavro___pyx_scope_struct_23_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_COMPILING_IN_CPYTHON
  if (likely((int)(__pyx_freecount_5cavro___pyx_scope_struct_23_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5cavro___pyx_scope_struct_23_genexpr[--__pyx_freecount_5cavro___pyx_scope_struct_23_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_5cavro___pyx_scope_struct_23_genexpr(PyObject *o) {
  struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_5cavro___pyx_scope_struct_23_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_alias);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_COMPILING_IN_CPYTHON
  if (((int)(__pyx_freecount_5cavro___pyx_scope_struct_23_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr)))) {
    __pyx_freelist_5cavro___pyx_scope_struct_23_genexpr[__pyx_freecount_5cavro___pyx_scope_struct_23_genexpr++] = ((struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_5cavro___pyx_scope_struct_23_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr *p = (struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_alias) {
    e = (*v)(p->__pyx_v_alias, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_5cavro___pyx_scope_struct_23_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_5cavro___pyx_scope_struct_23_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_5cavro___pyx_scope_struct_23_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_5cavro___pyx_scope_struct_23_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_5cavro___pyx_scope_struct_23_genexpr_spec = {
  "cavro.__pyx_scope_struct_23_genexpr",
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_5cavro___pyx_scope_struct_23_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_5cavro___pyx_scope_struct_23_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""__pyx_scope_struct_23_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5cavro___pyx_scope_struct_23_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5cavro___pyx_scope_struct_23_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5cavro___pyx_scope_struct_23_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5cavro___pyx_scope_struct_23_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_array __pyx_vtable_array;

static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_array_obj *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_array_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_array;
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_array(PyObject *o) {
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_array) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_array___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->mode);
  Py_CLEAR(p->_format);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_array___setitem__(o, i, v);
  }
  else {
    __Pyx_TypeName o_type_name;
    o_type_name = __Pyx_PyType_GetName(Py_TYPE(o));
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by " __Pyx_FMT_TYPENAME, o_type_name);
    __Pyx_DECREF_TypeName(o_type_name);
    return -1;
  }
}

static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_array___getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);
}

static PyMethodDef __pyx_methods_array[] = {
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_array_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_array_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_array[] = {
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
#if !CYTHON_COMPILING_IN_LIMITED_API

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};
#endif
static PyType_Slot __pyx_type___pyx_array_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_array},
  {Py_sq_length, (void *)__pyx_array___len__},
  {Py_sq_item, (void *)__pyx_sq_item_array},
  {Py_mp_length, (void *)__pyx_array___len__},
  {Py_mp_subscript, (void *)__pyx_array___getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_array},
  {Py_tp_getattro, (void *)__pyx_tp_getattro_array},
  #if defined(Py_bf_getbuffer)
  {Py_bf_getbuffer, (void *)__pyx_array_getbuffer},
  #endif
  {Py_tp_methods, (void *)__pyx_methods_array},
  {Py_tp_getset, (void *)__pyx_getsets_array},
  {Py_tp_new, (void *)__pyx_tp_new_array},
  {0, 0},
};
static PyType_Spec __pyx_type___pyx_array_spec = {
  "cavro.array",
  sizeof(struct __pyx_array_obj),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_SEQUENCE,
  __pyx_type___pyx_array_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_array = {
  __pyx_array___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_array, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_array = {
  __pyx_array___len__, /*mp_length*/
  __pyx_array___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_array = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""array", /*tp_name*/
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_array, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_array, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_SEQUENCE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_array, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_array, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_MemviewEnum_obj *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_MemviewEnum_obj *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_Enum(PyObject *o) {
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_Enum) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_Enum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_specialmethod___pyx_MemviewEnum___repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_MemviewEnum___repr__(self);
}

static PyMethodDef __pyx_methods_Enum[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_MemviewEnum___repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type___pyx_MemviewEnum_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_Enum},
  {Py_tp_repr, (void *)__pyx_MemviewEnum___repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_Enum},
  {Py_tp_clear, (void *)__pyx_tp_clear_Enum},
  {Py_tp_methods, (void *)__pyx_methods_Enum},
  {Py_tp_init, (void *)__pyx_MemviewEnum___init__},
  {Py_tp_new, (void *)__pyx_tp_new_Enum},
  {0, 0},
};
static PyType_Spec __pyx_type___pyx_MemviewEnum_spec = {
  "cavro.Enum",
  sizeof(struct __pyx_MemviewEnum_obj),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type___pyx_MemviewEnum_slots,
};
#else

static PyTypeObject __pyx_type___pyx_MemviewEnum = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""Enum", /*tp_name*/
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_MemviewEnum___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_Enum, /*tp_traverse*/
  __pyx_tp_clear_Enum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_Enum, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_MemviewEnum___init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_Enum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryview_obj *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_memoryview_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_memoryview;
  p->obj = Py_None; Py_INCREF(Py_None);
  p->_size = Py_None; Py_INCREF(Py_None);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  p->view.obj = NULL;
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_memoryview(PyObject *o) {
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_memoryview) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryview___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_size);
  Py_CLEAR(p->_array_interface);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  if (p->_size) {
    e = (*v)(p->_size, a); if (e) return e;
  }
  if (p->_array_interface) {
    e = (*v)(p->_array_interface, a); if (e) return e;
  }
  if (p->view.obj) {
    e = (*v)(p->view.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_memoryview(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_size);
  p->_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_array_interface);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->view.obj);
  return 0;
}
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_memoryview___setitem__(o, i, v);
  }
  else {
    __Pyx_TypeName o_type_name;
    o_type_name = __Pyx_PyType_GetName(Py_TYPE(o));
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by " __Pyx_FMT_TYPENAME, o_type_name);
    __Pyx_DECREF_TypeName(o_type_name);
    return -1;
  }
}

static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_memoryview___repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_memoryview___repr__(self);
}

static PyMethodDef __pyx_methods_memoryview[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_memoryview___repr__, METH_NOARGS|METH_COEXIST, 0},
  {"is_c_contig", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_memoryview_is_c_contig, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"is_f_contig", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_memoryview_is_f_contig, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_memoryview_copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"copy_fortran", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_memoryview_copy_fortran, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_memoryview_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_memoryview_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_memoryview[] = {
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
#if !CYTHON_COMPILING_IN_LIMITED_API

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};
#endif
static PyType_Slot __pyx_type___pyx_memoryview_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_memoryview},
  {Py_tp_repr, (void *)__pyx_memoryview___repr__},
  {Py_sq_length, (void *)__pyx_memoryview___len__},
  {Py_sq_item, (void *)__pyx_sq_item_memoryview},
  {Py_mp_length, (void *)__pyx_memoryview___len__},
  {Py_mp_subscript, (void *)__pyx_memoryview___getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_memoryview},
  {Py_tp_str, (void *)__pyx_memoryview___str__},
  #if defined(Py_bf_getbuffer)
  {Py_bf_getbuffer, (void *)__pyx_memoryview_getbuffer},
  #endif
  {Py_tp_traverse, (void *)__pyx_tp_traverse_memoryview},
  {Py_tp_clear, (void *)__pyx_tp_clear_memoryview},
  {Py_tp_methods, (void *)__pyx_methods_memoryview},
  {Py_tp_getset, (void *)__pyx_getsets_memoryview},
  {Py_tp_new, (void *)__pyx_tp_new_memoryview},
  {0, 0},
};
static PyType_Spec __pyx_type___pyx_memoryview_spec = {
  "cavro.memoryview",
  sizeof(struct __pyx_memoryview_obj),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type___pyx_memoryview_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_memoryview = {
  __pyx_memoryview___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_memoryview, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_memoryview = {
  __pyx_memoryview___len__, /*mp_length*/
  __pyx_memoryview___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_memoryview = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""memoryview", /*tp_name*/
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_memoryview___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_memoryview___str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_memoryview, /*tp_traverse*/
  __pyx_tp_clear_memoryview, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_memoryview, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_memoryview, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_memoryview, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryviewslice_obj *p;
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryviewslice_obj *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;
  p->from_object = Py_None; Py_INCREF(Py_None);
  p->from_slice.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc__memoryviewslice) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryviewslice___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->from_object);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_memoryview(o);
}

static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;
  if (p->from_object) {
    e = (*v)(p->from_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear__memoryviewslice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  __pyx_tp_clear_memoryview(o);
  tmp = ((PyObject*)p->from_object);
  p->from_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  __PYX_XCLEAR_MEMVIEW(&p->from_slice, 1);
  return 0;
}

static PyMethodDef __pyx_methods__memoryviewslice[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type___pyx_memoryviewslice_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc__memoryviewslice},
  {Py_tp_doc, (void *)PyDoc_STR("Internal class for passing memoryview slices to Python")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse__memoryviewslice},
  {Py_tp_clear, (void *)__pyx_tp_clear__memoryviewslice},
  {Py_tp_methods, (void *)__pyx_methods__memoryviewslice},
  {Py_tp_new, (void *)__pyx_tp_new__memoryviewslice},
  {0, 0},
};
static PyType_Spec __pyx_type___pyx_memoryviewslice_spec = {
  "cavro._memoryviewslice",
  sizeof(struct __pyx_memoryviewslice_obj),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_SEQUENCE,
  __pyx_type___pyx_memoryviewslice_slots,
};
#else

static PyTypeObject __pyx_type___pyx_memoryviewslice = {
  PyVarObject_HEAD_INIT(0, 0)
  "cavro.""_memoryviewslice", /*tp_name*/
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_memoryview___repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_memoryview___str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_SEQUENCE, /*tp_flags*/
  PyDoc_STR("Internal class for passing memoryview slices to Python"), /*tp_doc*/
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods__memoryviewslice, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new__memoryviewslice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
    {&__pyx_kp_u_0, __pyx_k_0, sizeof(__pyx_k_0), 0, 1, 0, 0},
    {&__pyx_kp_u_1, __pyx_k_1, sizeof(__pyx_k_1), 0, 1, 0, 0},
    {&__pyx_kp_u_1_0_0, __pyx_k_1_0_0, sizeof(__pyx_k_1_0_0), 0, 1, 0, 0},
    {&__pyx_n_s_AMBIGUOUS_TYPES, __pyx_k_AMBIGUOUS_TYPES, sizeof(__pyx_k_AMBIGUOUS_TYPES), 0, 0, 1, 1},
    {&__pyx_n_s_ASC, __pyx_k_ASC, sizeof(__pyx_k_ASC), 0, 0, 1, 1},
    {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
    {&__pyx_kp_u_A_Za_z__A_Za_z0_9, __pyx_k_A_Za_z__A_Za_z0_9, sizeof(__pyx_k_A_Za_z__A_Za_z0_9), 0, 1, 0, 0},
    {&__pyx_kp_s_A_record_contains_multiple_fiel, __pyx_k_A_record_contains_multiple_fiel, sizeof(__pyx_k_A_record_contains_multiple_fiel), 0, 0, 1, 0},
    {&__pyx_kp_s_A_requested_codec_or_codec_in_a, __pyx_k_A_requested_codec_or_codec_in_a, sizeof(__pyx_k_A_requested_codec_or_codec_in_a), 0, 0, 1, 0},
    {&__pyx_kp_s_A_schema_cannot_be_promoted_to, __pyx_k_A_schema_cannot_be_promoted_to, sizeof(__pyx_k_A_schema_cannot_be_promoted_to), 0, 0, 1, 0},
    {&__pyx_kp_s_A_value_is_invalid_for_a_given, __pyx_k_A_value_is_invalid_for_a_given, sizeof(__pyx_k_A_value_is_invalid_for_a_given), 0, 0, 1, 0},
    {&__pyx_kp_u_Aliases_must_be_a_list_tuple_set, __pyx_k_Aliases_must_be_a_list_tuple_set, sizeof(__pyx_k_Aliases_must_be_a_list_tuple_set), 0, 1, 0, 0},
    {&__pyx_kp_s_All_dimensions_preceding_dimensi, __pyx_k_All_dimensions_preceding_dimensi, sizeof(__pyx_k_All_dimensions_preceding_dimensi), 0, 0, 1, 0},
    {&__pyx_kp_s_An_unknown_hash_method_was_requ, __pyx_k_An_unknown_hash_method_was_requ, sizeof(__pyx_k_An_unknown_hash_method_was_requ), 0, 0, 1, 0},
    {&__pyx_n_s_ArrayType, __pyx_k_ArrayType, sizeof(__pyx_k_ArrayType), 0, 0, 1, 1},
    {&__pyx_n_s_ArrayType___reduce_cython, __pyx_k_ArrayType___reduce_cython, sizeof(__pyx_k_ArrayType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ArrayType___setstate_cython, __pyx_k_ArrayType___setstate_cython, sizeof(__pyx_k_ArrayType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ArrayType__get_schema_extra, __pyx_k_ArrayType__get_schema_extra, sizeof(__pyx_k_ArrayType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_ArrayType_convert_value, __pyx_k_ArrayType_convert_value, sizeof(__pyx_k_ArrayType_convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_ArrayType_copy, __pyx_k_ArrayType_copy, sizeof(__pyx_k_ArrayType_copy), 0, 0, 1, 1},
    {&__pyx_n_s_ArrayType_walk_types, __pyx_k_ArrayType_walk_types, sizeof(__pyx_k_ArrayType_walk_types), 0, 0, 1, 1},
    {&__pyx_kp_s_As_of_cython_3_0_2_dataclasses, __pyx_k_As_of_cython_3_0_2_dataclasses, sizeof(__pyx_k_As_of_cython_3_0_2_dataclasses), 0, 0, 1, 0},
    {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
    {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
    {&__pyx_n_s_AvroType, __pyx_k_AvroType, sizeof(__pyx_k_AvroType), 0, 0, 1, 1},
    {&__pyx_n_s_AvroType___reduce_cython, __pyx_k_AvroType___reduce_cython, sizeof(__pyx_k_AvroType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_AvroType___setstate_cython, __pyx_k_AvroType___setstate_cython, sizeof(__pyx_k_AvroType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_AvroType__convert_value, __pyx_k_AvroType__convert_value, sizeof(__pyx_k_AvroType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_AvroType__get_schema_extra, __pyx_k_AvroType__get_schema_extra, sizeof(__pyx_k_AvroType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_AvroType_convert_value, __pyx_k_AvroType_convert_value, sizeof(__pyx_k_AvroType_convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_AvroType_copy, __pyx_k_AvroType_copy, sizeof(__pyx_k_AvroType_copy), 0, 0, 1, 1},
    {&__pyx_n_s_AvroType_for_schema, __pyx_k_AvroType_for_schema, sizeof(__pyx_k_AvroType_for_schema), 0, 0, 1, 1},
    {&__pyx_n_s_AvroType_for_source, __pyx_k_AvroType_for_source, sizeof(__pyx_k_AvroType_for_source), 0, 0, 1, 1},
    {&__pyx_n_s_AvroType_get_schema, __pyx_k_AvroType_get_schema, sizeof(__pyx_k_AvroType_get_schema), 0, 0, 1, 1},
    {&__pyx_n_s_AvroType_walk_types, __pyx_k_AvroType_walk_types, sizeof(__pyx_k_AvroType_walk_types), 0, 0, 1, 1},
    {&__pyx_n_u_B, __pyx_k_B, sizeof(__pyx_k_B), 0, 1, 0, 1},
    {&__pyx_n_s_BATCH_SIZE, __pyx_k_BATCH_SIZE, sizeof(__pyx_k_BATCH_SIZE), 0, 0, 1, 1},
    {&__pyx_kp_s_Base_class_for_exceptions_raise, __pyx_k_Base_class_for_exceptions_raise, sizeof(__pyx_k_Base_class_for_exceptions_raise), 0, 0, 1, 0},
    {&__pyx_n_s_BoolType, __pyx_k_BoolType, sizeof(__pyx_k_BoolType), 0, 0, 1, 1},
    {&__pyx_n_s_BoolType___reduce_cython, __pyx_k_BoolType___reduce_cython, sizeof(__pyx_k_BoolType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BoolType___setstate_cython, __pyx_k_BoolType___setstate_cython, sizeof(__pyx_k_BoolType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BoolType__convert_value, __pyx_k_BoolType__convert_value, sizeof(__pyx_k_BoolType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_BoolType__get_schema_extra, __pyx_k_BoolType__get_schema_extra, sizeof(__pyx_k_BoolType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_BoolType_copy, __pyx_k_BoolType_copy, sizeof(__pyx_k_BoolType_copy), 0, 0, 1, 1},
    {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
    {&__pyx_n_s_BytesType, __pyx_k_BytesType, sizeof(__pyx_k_BytesType), 0, 0, 1, 1},
    {&__pyx_n_s_BytesType___reduce_cython, __pyx_k_BytesType___reduce_cython, sizeof(__pyx_k_BytesType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BytesType___setstate_cython, __pyx_k_BytesType___setstate_cython, sizeof(__pyx_k_BytesType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BytesType__convert_value, __pyx_k_BytesType__convert_value, sizeof(__pyx_k_BytesType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_BytesType__get_schema_extra, __pyx_k_BytesType__get_schema_extra, sizeof(__pyx_k_BytesType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_BytesType_copy, __pyx_k_BytesType_copy, sizeof(__pyx_k_BytesType_copy), 0, 0, 1, 1},
    {&__pyx_n_s_Bzip2Codec, __pyx_k_Bzip2Codec, sizeof(__pyx_k_Bzip2Codec), 0, 0, 1, 1},
    {&__pyx_n_s_Bzip2Codec___reduce_cython, __pyx_k_Bzip2Codec___reduce_cython, sizeof(__pyx_k_Bzip2Codec___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Bzip2Codec___setstate_cython, __pyx_k_Bzip2Codec___setstate_cython, sizeof(__pyx_k_Bzip2Codec___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_CANONICAL_FORM_KEYS, __pyx_k_CANONICAL_FORM_KEYS, sizeof(__pyx_k_CANONICAL_FORM_KEYS), 0, 0, 1, 1},
    {&__pyx_n_s_CODECS, __pyx_k_CODECS, sizeof(__pyx_k_CODECS), 0, 0, 1, 1},
    {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
    {&__pyx_n_s_CannotPromote, __pyx_k_CannotPromote, sizeof(__pyx_k_CannotPromote), 0, 0, 1, 1},
    {&__pyx_n_s_CannotPromoteError, __pyx_k_CannotPromoteError, sizeof(__pyx_k_CannotPromoteError), 0, 0, 1, 1},
    {&__pyx_n_s_CannotPromoteError___init, __pyx_k_CannotPromoteError___init, sizeof(__pyx_k_CannotPromoteError___init), 0, 0, 1, 1},
    {&__pyx_n_s_CannotPromote___reduce_cython, __pyx_k_CannotPromote___reduce_cython, sizeof(__pyx_k_CannotPromote___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_CannotPromote___setstate_cython, __pyx_k_CannotPromote___setstate_cython, sizeof(__pyx_k_CannotPromote___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
    {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
    {&__pyx_kp_u_Cannot_index_with_type, __pyx_k_Cannot_index_with_type, sizeof(__pyx_k_Cannot_index_with_type), 0, 1, 0, 0},
    {&__pyx_kp_u_Cannot_promote, __pyx_k_Cannot_promote, sizeof(__pyx_k_Cannot_promote), 0, 1, 0, 0},
    {&__pyx_kp_u_Cannot_read_from, __pyx_k_Cannot_read_from, sizeof(__pyx_k_Cannot_read_from), 0, 1, 0, 0},
    {&__pyx_kp_s_Cannot_transpose_memoryview_with, __pyx_k_Cannot_transpose_memoryview_with, sizeof(__pyx_k_Cannot_transpose_memoryview_with), 0, 0, 1, 0},
    {&__pyx_kp_u_Cannot_write_to, __pyx_k_Cannot_write_to, sizeof(__pyx_k_Cannot_write_to), 0, 1, 0, 0},
    {&__pyx_n_s_CanonicalForm, __pyx_k_CanonicalForm, sizeof(__pyx_k_CanonicalForm), 0, 0, 1, 1},
    {&__pyx_n_s_CanonicalForm___reduce_cython, __pyx_k_CanonicalForm___reduce_cython, sizeof(__pyx_k_CanonicalForm___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_CanonicalForm___setstate_cython, __pyx_k_CanonicalForm___setstate_cython, sizeof(__pyx_k_CanonicalForm___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_CavroException, __pyx_k_CavroException, sizeof(__pyx_k_CavroException), 0, 0, 1, 1},
    {&__pyx_n_s_Codec, __pyx_k_Codec, sizeof(__pyx_k_Codec), 0, 0, 1, 1},
    {&__pyx_n_s_CodecUnavailable, __pyx_k_CodecUnavailable, sizeof(__pyx_k_CodecUnavailable), 0, 0, 1, 1},
    {&__pyx_n_s_Codec___reduce_cython, __pyx_k_Codec___reduce_cython, sizeof(__pyx_k_Codec___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Codec___setstate_cython, __pyx_k_Codec___setstate_cython, sizeof(__pyx_k_Codec___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerReader, __pyx_k_ContainerReader, sizeof(__pyx_k_ContainerReader), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerReader___reduce_cython, __pyx_k_ContainerReader___reduce_cython, sizeof(__pyx_k_ContainerReader___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerReader___setstate_cytho, __pyx_k_ContainerReader___setstate_cytho, sizeof(__pyx_k_ContainerReader___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerReader__read_marker, __pyx_k_ContainerReader__read_marker, sizeof(__pyx_k_ContainerReader__read_marker), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerReader_next_object, __pyx_k_ContainerReader_next_object, sizeof(__pyx_k_ContainerReader_next_object), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerWriter, __pyx_k_ContainerWriter, sizeof(__pyx_k_ContainerWriter), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerWriter___enter, __pyx_k_ContainerWriter___enter, sizeof(__pyx_k_ContainerWriter___enter), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerWriter___exit, __pyx_k_ContainerWriter___exit, sizeof(__pyx_k_ContainerWriter___exit), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerWriter___reduce_cython, __pyx_k_ContainerWriter___reduce_cython, sizeof(__pyx_k_ContainerWriter___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerWriter___setstate_cytho, __pyx_k_ContainerWriter___setstate_cytho, sizeof(__pyx_k_ContainerWriter___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerWriter_close, __pyx_k_ContainerWriter_close, sizeof(__pyx_k_ContainerWriter_close), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerWriter_flush, __pyx_k_ContainerWriter_flush, sizeof(__pyx_k_ContainerWriter_flush), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerWriter_write_many, __pyx_k_ContainerWriter_write_many, sizeof(__pyx_k_ContainerWriter_write_many), 0, 0, 1, 1},
    {&__pyx_n_s_ContainerWriter_write_one, __pyx_k_ContainerWriter_write_one, sizeof(__pyx_k_ContainerWriter_write_one), 0, 0, 1, 1},
    {&__pyx_kp_u_Container_is_closed, __pyx_k_Container_is_closed, sizeof(__pyx_k_Container_is_closed), 0, 1, 0, 0},
    {&__pyx_n_s_Context, __pyx_k_Context, sizeof(__pyx_k_Context), 0, 0, 1, 1},
    {&__pyx_kp_s_Controls_how_union_values_are_d, __pyx_k_Controls_how_union_values_are_d, sizeof(__pyx_k_Controls_how_union_values_are_d), 0, 0, 1, 0},
    {&__pyx_kp_u_Could_not_find_key_type_in_schem, __pyx_k_Could_not_find_key_type_in_schem, sizeof(__pyx_k_Could_not_find_key_type_in_schem), 0, 1, 0, 0},
    {&__pyx_n_s_CustomLogicalType, __pyx_k_CustomLogicalType, sizeof(__pyx_k_CustomLogicalType), 0, 0, 1, 1},
    {&__pyx_n_s_CustomLogicalType___reduce_cytho, __pyx_k_CustomLogicalType___reduce_cytho, sizeof(__pyx_k_CustomLogicalType___reduce_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_CustomLogicalType___setstate_cyt, __pyx_k_CustomLogicalType___setstate_cyt, sizeof(__pyx_k_CustomLogicalType___setstate_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_DEFAULT_OPTIONS, __pyx_k_DEFAULT_OPTIONS, sizeof(__pyx_k_DEFAULT_OPTIONS), 0, 0, 1, 1},
    {&__pyx_n_s_DESC, __pyx_k_DESC, sizeof(__pyx_k_DESC), 0, 0, 1, 1},
    {&__pyx_n_s_DataclassParams, __pyx_k_DataclassParams, sizeof(__pyx_k_DataclassParams), 0, 0, 1, 1},
    {&__pyx_n_s_Date, __pyx_k_Date, sizeof(__pyx_k_Date), 0, 0, 1, 1},
    {&__pyx_n_s_Date___reduce_cython, __pyx_k_Date___reduce_cython, sizeof(__pyx_k_Date___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Date___setstate_cython, __pyx_k_Date___setstate_cython, sizeof(__pyx_k_Date___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Date__for_type, __pyx_k_Date__for_type, sizeof(__pyx_k_Date__for_type), 0, 0, 1, 1},
    {&__pyx_n_s_Decimal, __pyx_k_Decimal, sizeof(__pyx_k_Decimal), 0, 0, 1, 1},
    {&__pyx_n_s_DecimalType, __pyx_k_DecimalType, sizeof(__pyx_k_DecimalType), 0, 0, 1, 1},
    {&__pyx_n_s_DecimalType___reduce_cython, __pyx_k_DecimalType___reduce_cython, sizeof(__pyx_k_DecimalType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_DecimalType___setstate_cython, __pyx_k_DecimalType___setstate_cython, sizeof(__pyx_k_DecimalType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_DecimalType__for_type, __pyx_k_DecimalType__for_type, sizeof(__pyx_k_DecimalType__for_type), 0, 0, 1, 1},
    {&__pyx_kp_u_Decimal_value, __pyx_k_Decimal_value, sizeof(__pyx_k_Decimal_value), 0, 1, 0, 0},
    {&__pyx_kp_u_Default_value, __pyx_k_Default_value, sizeof(__pyx_k_Default_value), 0, 1, 0, 0},
    {&__pyx_kp_u_Default_value_2, __pyx_k_Default_value_2, sizeof(__pyx_k_Default_value_2), 0, 1, 0, 0},
    {&__pyx_n_s_DeflateCodec, __pyx_k_DeflateCodec, sizeof(__pyx_k_DeflateCodec), 0, 0, 1, 1},
    {&__pyx_n_s_DeflateCodec___reduce_cython, __pyx_k_DeflateCodec___reduce_cython, sizeof(__pyx_k_DeflateCodec___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_DeflateCodec___setstate_cython, __pyx_k_DeflateCodec___setstate_cython, sizeof(__pyx_k_DeflateCodec___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_s_Dimension_d_is_not_direct, __pyx_k_Dimension_d_is_not_direct, sizeof(__pyx_k_Dimension_d_is_not_direct), 0, 0, 1, 0},
    {&__pyx_n_s_DoubleType, __pyx_k_DoubleType, sizeof(__pyx_k_DoubleType), 0, 0, 1, 1},
    {&__pyx_n_s_DoubleType___reduce_cython, __pyx_k_DoubleType___reduce_cython, sizeof(__pyx_k_DoubleType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_DoubleType___setstate_cython, __pyx_k_DoubleType___setstate_cython, sizeof(__pyx_k_DoubleType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_DoubleType__convert_value, __pyx_k_DoubleType__convert_value, sizeof(__pyx_k_DoubleType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_DoubleType__get_schema_extra, __pyx_k_DoubleType__get_schema_extra, sizeof(__pyx_k_DoubleType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_DoubleType_copy, __pyx_k_DoubleType_copy, sizeof(__pyx_k_DoubleType_copy), 0, 0, 1, 1},
    {&__pyx_n_s_DuplicateName, __pyx_k_DuplicateName, sizeof(__pyx_k_DuplicateName), 0, 0, 1, 1},
    {&__pyx_kp_u_Duplicate_field_name, __pyx_k_Duplicate_field_name, sizeof(__pyx_k_Duplicate_field_name), 0, 1, 0, 0},
    {&__pyx_n_s_EMPTY_ARRAY, __pyx_k_EMPTY_ARRAY, sizeof(__pyx_k_EMPTY_ARRAY), 0, 0, 1, 1},
    {&__pyx_n_s_EMPTY_MP, __pyx_k_EMPTY_MP, sizeof(__pyx_k_EMPTY_MP), 0, 0, 1, 1},
    {&__pyx_n_s_EOFError, __pyx_k_EOFError, sizeof(__pyx_k_EOFError), 0, 0, 1, 1},
    {&__pyx_n_s_EPOCH_DATE, __pyx_k_EPOCH_DATE, sizeof(__pyx_k_EPOCH_DATE), 0, 0, 1, 1},
    {&__pyx_n_s_EPOCH_DT, __pyx_k_EPOCH_DT, sizeof(__pyx_k_EPOCH_DT), 0, 0, 1, 1},
    {&__pyx_n_s_EPOCH_DT_NO_TZ, __pyx_k_EPOCH_DT_NO_TZ, sizeof(__pyx_k_EPOCH_DT_NO_TZ), 0, 0, 1, 1},
    {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
    {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
    {&__pyx_kp_u_End_of_file_found_trying_to_read, __pyx_k_End_of_file_found_trying_to_read, sizeof(__pyx_k_End_of_file_found_trying_to_read), 0, 1, 0, 0},
    {&__pyx_kp_u_End_of_file_found_trying_to_read_2, __pyx_k_End_of_file_found_trying_to_read_2, sizeof(__pyx_k_End_of_file_found_trying_to_read_2), 0, 1, 0, 0},
    {&__pyx_n_s_Enum, __pyx_k_Enum, sizeof(__pyx_k_Enum), 0, 0, 1, 1},
    {&__pyx_n_s_EnumType, __pyx_k_EnumType, sizeof(__pyx_k_EnumType), 0, 0, 1, 1},
    {&__pyx_n_s_EnumType___reduce_cython, __pyx_k_EnumType___reduce_cython, sizeof(__pyx_k_EnumType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_EnumType___setstate_cython, __pyx_k_EnumType___setstate_cython, sizeof(__pyx_k_EnumType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_EnumType__convert_value, __pyx_k_EnumType__convert_value, sizeof(__pyx_k_EnumType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_EnumType__for_writer_locals_gene, __pyx_k_EnumType__for_writer_locals_gene, sizeof(__pyx_k_EnumType__for_writer_locals_gene), 0, 0, 1, 1},
    {&__pyx_n_s_EnumType__get_schema_extra, __pyx_k_EnumType__get_schema_extra, sizeof(__pyx_k_EnumType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_EnumType_copy, __pyx_k_EnumType_copy, sizeof(__pyx_k_EnumType_copy), 0, 0, 1, 1},
    {&__pyx_kp_u_Enum_symbol, __pyx_k_Enum_symbol, sizeof(__pyx_k_Enum_symbol), 0, 1, 0, 0},
    {&__pyx_kp_u_Enum_symbols_must_be_a_list_of_s, __pyx_k_Enum_symbols_must_be_a_list_of_s, sizeof(__pyx_k_Enum_symbols_must_be_a_list_of_s), 0, 1, 0, 0},
    {&__pyx_kp_u_Enum_symbols_must_be_unique, __pyx_k_Enum_symbols_must_be_unique, sizeof(__pyx_k_Enum_symbols_must_be_unique), 0, 1, 0, 0},
    {&__pyx_kp_u_Error_closing_container_file_dur, __pyx_k_Error_closing_container_file_dur, sizeof(__pyx_k_Error_closing_container_file_dur), 0, 1, 0, 0},
    {&__pyx_kp_u_Expected_datetime_date_got, __pyx_k_Expected_datetime_date_got, sizeof(__pyx_k_Expected_datetime_date_got), 0, 1, 0, 0},
    {&__pyx_kp_u_Expected_datetime_datetime_got, __pyx_k_Expected_datetime_datetime_got, sizeof(__pyx_k_Expected_datetime_datetime_got), 0, 1, 0, 0},
    {&__pyx_kp_u_Expected_null_got, __pyx_k_Expected_null_got, sizeof(__pyx_k_Expected_null_got), 0, 1, 0, 0},
    {&__pyx_n_s_ExponentTooLarge, __pyx_k_ExponentTooLarge, sizeof(__pyx_k_ExponentTooLarge), 0, 0, 1, 1},
    {&__pyx_n_s_FIELD, __pyx_k_FIELD, sizeof(__pyx_k_FIELD), 0, 0, 1, 1},
    {&__pyx_n_u_False, __pyx_k_False, sizeof(__pyx_k_False), 0, 1, 0, 1},
    {&__pyx_kp_u_Field, __pyx_k_Field, sizeof(__pyx_k_Field), 0, 1, 0, 0},
    {&__pyx_n_s_FieldAccessor, __pyx_k_FieldAccessor, sizeof(__pyx_k_FieldAccessor), 0, 0, 1, 1},
    {&__pyx_n_s_FieldAccessor___reduce_cython, __pyx_k_FieldAccessor___reduce_cython, sizeof(__pyx_k_FieldAccessor___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FieldAccessor___setstate_cython, __pyx_k_FieldAccessor___setstate_cython, sizeof(__pyx_k_FieldAccessor___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FileReader, __pyx_k_FileReader, sizeof(__pyx_k_FileReader), 0, 0, 1, 1},
    {&__pyx_n_s_FileReader___reduce_cython, __pyx_k_FileReader___reduce_cython, sizeof(__pyx_k_FileReader___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FileReader___setstate_cython, __pyx_k_FileReader___setstate_cython, sizeof(__pyx_k_FileReader___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FileWriter, __pyx_k_FileWriter, sizeof(__pyx_k_FileWriter), 0, 0, 1, 1},
    {&__pyx_n_s_FileWriter___reduce_cython, __pyx_k_FileWriter___reduce_cython, sizeof(__pyx_k_FileWriter___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FileWriter___setstate_cython, __pyx_k_FileWriter___setstate_cython, sizeof(__pyx_k_FileWriter___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FixedType, __pyx_k_FixedType, sizeof(__pyx_k_FixedType), 0, 0, 1, 1},
    {&__pyx_n_s_FixedType___reduce_cython, __pyx_k_FixedType___reduce_cython, sizeof(__pyx_k_FixedType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FixedType___setstate_cython, __pyx_k_FixedType___setstate_cython, sizeof(__pyx_k_FixedType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FixedType__convert_value, __pyx_k_FixedType__convert_value, sizeof(__pyx_k_FixedType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_FixedType__get_schema_extra, __pyx_k_FixedType__get_schema_extra, sizeof(__pyx_k_FixedType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_FixedType_copy, __pyx_k_FixedType_copy, sizeof(__pyx_k_FixedType_copy), 0, 0, 1, 1},
    {&__pyx_n_s_FloatType, __pyx_k_FloatType, sizeof(__pyx_k_FloatType), 0, 0, 1, 1},
    {&__pyx_n_s_FloatType___reduce_cython, __pyx_k_FloatType___reduce_cython, sizeof(__pyx_k_FloatType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FloatType___setstate_cython, __pyx_k_FloatType___setstate_cython, sizeof(__pyx_k_FloatType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FloatType__convert_value, __pyx_k_FloatType__convert_value, sizeof(__pyx_k_FloatType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_FloatType__get_schema_extra, __pyx_k_FloatType__get_schema_extra, sizeof(__pyx_k_FloatType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_FloatType_copy, __pyx_k_FloatType_copy, sizeof(__pyx_k_FloatType_copy), 0, 0, 1, 1},
    {&__pyx_n_s_HAVE_BZIP2, __pyx_k_HAVE_BZIP2, sizeof(__pyx_k_HAVE_BZIP2), 0, 0, 1, 1},
    {&__pyx_n_s_HAVE_LZ4, __pyx_k_HAVE_LZ4, sizeof(__pyx_k_HAVE_LZ4), 0, 0, 1, 1},
    {&__pyx_n_s_HAVE_SNAPPY, __pyx_k_HAVE_SNAPPY, sizeof(__pyx_k_HAVE_SNAPPY), 0, 0, 1, 1},
    {&__pyx_n_s_HAVE_XZ, __pyx_k_HAVE_XZ, sizeof(__pyx_k_HAVE_XZ), 0, 0, 1, 1},
    {&__pyx_n_s_HAVE_ZLIB, __pyx_k_HAVE_ZLIB, sizeof(__pyx_k_HAVE_ZLIB), 0, 0, 1, 1},
    {&__pyx_n_s_HAVE_ZSTD, __pyx_k_HAVE_ZSTD, sizeof(__pyx_k_HAVE_ZSTD), 0, 0, 1, 1},
    {&__pyx_n_s_IGNORE, __pyx_k_IGNORE, sizeof(__pyx_k_IGNORE), 0, 0, 1, 1},
    {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_10, __pyx_k_Incompatible_checksums_0x_x_vs_0_10, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_10), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_11, __pyx_k_Incompatible_checksums_0x_x_vs_0_11, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_11), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_12, __pyx_k_Incompatible_checksums_0x_x_vs_0_12, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_12), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_13, __pyx_k_Incompatible_checksums_0x_x_vs_0_13, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_13), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_14, __pyx_k_Incompatible_checksums_0x_x_vs_0_14, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_14), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_15, __pyx_k_Incompatible_checksums_0x_x_vs_0_15, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_15), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_16, __pyx_k_Incompatible_checksums_0x_x_vs_0_16, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_16), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_17, __pyx_k_Incompatible_checksums_0x_x_vs_0_17, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_17), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_18, __pyx_k_Incompatible_checksums_0x_x_vs_0_18, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_18), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_19, __pyx_k_Incompatible_checksums_0x_x_vs_0_19, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_19), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_20, __pyx_k_Incompatible_checksums_0x_x_vs_0_20, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_20), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_21, __pyx_k_Incompatible_checksums_0x_x_vs_0_21, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_21), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_22, __pyx_k_Incompatible_checksums_0x_x_vs_0_22, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_22), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_23, __pyx_k_Incompatible_checksums_0x_x_vs_0_23, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_23), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_24, __pyx_k_Incompatible_checksums_0x_x_vs_0_24, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_24), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_25, __pyx_k_Incompatible_checksums_0x_x_vs_0_25, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_25), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_26, __pyx_k_Incompatible_checksums_0x_x_vs_0_26, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_26), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_27, __pyx_k_Incompatible_checksums_0x_x_vs_0_27, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_27), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_28, __pyx_k_Incompatible_checksums_0x_x_vs_0_28, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_28), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_k_Incompatible_checksums_0x_x_vs_0_4, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_4), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_5, __pyx_k_Incompatible_checksums_0x_x_vs_0_5, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_5), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_6, __pyx_k_Incompatible_checksums_0x_x_vs_0_6, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_6), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_7, __pyx_k_Incompatible_checksums_0x_x_vs_0_7, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_7), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_8, __pyx_k_Incompatible_checksums_0x_x_vs_0_8, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_8), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_9, __pyx_k_Incompatible_checksums_0x_x_vs_0_9, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_9), 0, 0, 1, 0},
    {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
    {&__pyx_kp_s_Index_out_of_bounds_axis_d, __pyx_k_Index_out_of_bounds_axis_d, sizeof(__pyx_k_Index_out_of_bounds_axis_d), 0, 0, 1, 0},
    {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
    {&__pyx_n_s_IntType, __pyx_k_IntType, sizeof(__pyx_k_IntType), 0, 0, 1, 1},
    {&__pyx_n_s_IntType___reduce_cython, __pyx_k_IntType___reduce_cython, sizeof(__pyx_k_IntType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_IntType___setstate_cython, __pyx_k_IntType___setstate_cython, sizeof(__pyx_k_IntType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_IntType__convert_value, __pyx_k_IntType__convert_value, sizeof(__pyx_k_IntType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_IntType__get_schema_extra, __pyx_k_IntType__get_schema_extra, sizeof(__pyx_k_IntType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_IntType_copy, __pyx_k_IntType_copy, sizeof(__pyx_k_IntType_copy), 0, 0, 1, 1},
    {&__pyx_n_s_InvalidHasher, __pyx_k_InvalidHasher, sizeof(__pyx_k_InvalidHasher), 0, 0, 1, 1},
    {&__pyx_n_s_InvalidName, __pyx_k_InvalidName, sizeof(__pyx_k_InvalidName), 0, 0, 1, 1},
    {&__pyx_n_s_InvalidValue, __pyx_k_InvalidValue, sizeof(__pyx_k_InvalidValue), 0, 0, 1, 1},
    {&__pyx_n_s_InvalidValue___init, __pyx_k_InvalidValue___init, sizeof(__pyx_k_InvalidValue___init), 0, 0, 1, 1},
    {&__pyx_n_s_InvalidValue___str, __pyx_k_InvalidValue___str, sizeof(__pyx_k_InvalidValue___str), 0, 0, 1, 1},
    {&__pyx_n_s_InvalidValue___str___locals_gene, __pyx_k_InvalidValue___str___locals_gene, sizeof(__pyx_k_InvalidValue___str___locals_gene), 0, 0, 1, 1},
    {&__pyx_kp_u_Invalid_block_sync_marker_expect, __pyx_k_Invalid_block_sync_marker_expect, sizeof(__pyx_k_Invalid_block_sync_marker_expect), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_file_header_expected, __pyx_k_Invalid_file_header_expected, sizeof(__pyx_k_Invalid_file_header_expected), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_length_for_fixed_field, __pyx_k_Invalid_length_for_fixed_field, sizeof(__pyx_k_Invalid_length_for_fixed_field), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_negative_size_for_fixed, __pyx_k_Invalid_negative_size_for_fixed, sizeof(__pyx_k_Invalid_negative_size_for_fixed), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_non_bytes_value_for_fixe, __pyx_k_Invalid_non_bytes_value_for_fixe, sizeof(__pyx_k_Invalid_non_bytes_value_for_fixe), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_non_str_value_for_fixed, __pyx_k_Invalid_non_str_value_for_fixed, sizeof(__pyx_k_Invalid_non_str_value_for_fixed), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_shape_in_axis, __pyx_k_Invalid_shape_in_axis, sizeof(__pyx_k_Invalid_shape_in_axis), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_union_decodes_to_option, __pyx_k_Invalid_union_decodes_to_option, sizeof(__pyx_k_Invalid_union_decodes_to_option), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_value, __pyx_k_Invalid_value, sizeof(__pyx_k_Invalid_value), 0, 1, 0, 0},
    {&__pyx_n_s_Iterable, __pyx_k_Iterable, sizeof(__pyx_k_Iterable), 0, 0, 1, 1},
    {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
    {&__pyx_n_s_LOGICAL_TYPES, __pyx_k_LOGICAL_TYPES, sizeof(__pyx_k_LOGICAL_TYPES), 0, 0, 1, 1},
    {&__pyx_n_u_Ll, __pyx_k_Ll, sizeof(__pyx_k_Ll), 0, 1, 0, 1},
    {&__pyx_n_s_LogicalType, __pyx_k_LogicalType, sizeof(__pyx_k_LogicalType), 0, 0, 1, 1},
    {&__pyx_n_s_LogicalType___reduce_cython, __pyx_k_LogicalType___reduce_cython, sizeof(__pyx_k_LogicalType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_LogicalType___setstate_cython, __pyx_k_LogicalType___setstate_cython, sizeof(__pyx_k_LogicalType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_LogicalType_for_type, __pyx_k_LogicalType_for_type, sizeof(__pyx_k_LogicalType_for_type), 0, 0, 1, 1},
    {&__pyx_n_s_LongType, __pyx_k_LongType, sizeof(__pyx_k_LongType), 0, 0, 1, 1},
    {&__pyx_n_s_LongType___reduce_cython, __pyx_k_LongType___reduce_cython, sizeof(__pyx_k_LongType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_LongType___setstate_cython, __pyx_k_LongType___setstate_cython, sizeof(__pyx_k_LongType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_LongType__convert_value, __pyx_k_LongType__convert_value, sizeof(__pyx_k_LongType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_LongType__get_schema_extra, __pyx_k_LongType__get_schema_extra, sizeof(__pyx_k_LongType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_LongType_copy, __pyx_k_LongType_copy, sizeof(__pyx_k_LongType_copy), 0, 0, 1, 1},
    {&__pyx_n_u_Lu, __pyx_k_Lu, sizeof(__pyx_k_Lu), 0, 1, 0, 1},
    {&__pyx_n_s_Lz4Codec, __pyx_k_Lz4Codec, sizeof(__pyx_k_Lz4Codec), 0, 0, 1, 1},
    {&__pyx_n_s_Lz4Codec___reduce_cython, __pyx_k_Lz4Codec___reduce_cython, sizeof(__pyx_k_Lz4Codec___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Lz4Codec___setstate_cython, __pyx_k_Lz4Codec___setstate_cython, sizeof(__pyx_k_Lz4Codec___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_LzmaCodec, __pyx_k_LzmaCodec, sizeof(__pyx_k_LzmaCodec), 0, 0, 1, 1},
    {&__pyx_n_s_LzmaCodec___reduce_cython, __pyx_k_LzmaCodec___reduce_cython, sizeof(__pyx_k_LzmaCodec___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_LzmaCodec___setstate_cython, __pyx_k_LzmaCodec___setstate_cython, sizeof(__pyx_k_LzmaCodec___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_MAX_PREC, __pyx_k_MAX_PREC, sizeof(__pyx_k_MAX_PREC), 0, 0, 1, 1},
    {&__pyx_n_s_MAX_TIME, __pyx_k_MAX_TIME, sizeof(__pyx_k_MAX_TIME), 0, 0, 1, 1},
    {&__pyx_n_s_MISSING_VALUE, __pyx_k_MISSING_VALUE, sizeof(__pyx_k_MISSING_VALUE), 0, 0, 1, 1},
    {&__pyx_n_u_MISSING_VALUE, __pyx_k_MISSING_VALUE, sizeof(__pyx_k_MISSING_VALUE), 0, 1, 0, 1},
    {&__pyx_n_s_MapType, __pyx_k_MapType, sizeof(__pyx_k_MapType), 0, 0, 1, 1},
    {&__pyx_n_s_MapType___reduce_cython, __pyx_k_MapType___reduce_cython, sizeof(__pyx_k_MapType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_MapType___setstate_cython, __pyx_k_MapType___setstate_cython, sizeof(__pyx_k_MapType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_MapType__get_schema_extra, __pyx_k_MapType__get_schema_extra, sizeof(__pyx_k_MapType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_MapType_convert_value, __pyx_k_MapType_convert_value, sizeof(__pyx_k_MapType_convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_MapType_copy, __pyx_k_MapType_copy, sizeof(__pyx_k_MapType_copy), 0, 0, 1, 1},
    {&__pyx_n_s_MapType_walk_types, __pyx_k_MapType_walk_types, sizeof(__pyx_k_MapType_walk_types), 0, 0, 1, 1},
    {&__pyx_n_s_MappingProxyType, __pyx_k_MappingProxyType, sizeof(__pyx_k_MappingProxyType), 0, 0, 1, 1},
    {&__pyx_kp_u_Marker_must_be_bytes_got, __pyx_k_Marker_must_be_bytes_got, sizeof(__pyx_k_Marker_must_be_bytes_got), 0, 1, 0, 0},
    {&__pyx_kp_u_Marker_must_be_exactly_16_bytes, __pyx_k_Marker_must_be_exactly_16_bytes, sizeof(__pyx_k_Marker_must_be_exactly_16_bytes), 0, 1, 0, 0},
    {&__pyx_kp_u_Max_precision_is, __pyx_k_Max_precision_is, sizeof(__pyx_k_Max_precision_is), 0, 1, 0, 0},
    {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
    {&__pyx_n_s_MemoryReader, __pyx_k_MemoryReader, sizeof(__pyx_k_MemoryReader), 0, 0, 1, 1},
    {&__pyx_n_s_MemoryReader___reduce_cython, __pyx_k_MemoryReader___reduce_cython, sizeof(__pyx_k_MemoryReader___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_MemoryReader___setstate_cython, __pyx_k_MemoryReader___setstate_cython, sizeof(__pyx_k_MemoryReader___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
    {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
    {&__pyx_n_s_MemoryWriter, __pyx_k_MemoryWriter, sizeof(__pyx_k_MemoryWriter), 0, 0, 1, 1},
    {&__pyx_n_s_MemoryWriter___reduce_cython, __pyx_k_MemoryWriter___reduce_cython, sizeof(__pyx_k_MemoryWriter___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_MemoryWriter___setstate_cython, __pyx_k_MemoryWriter___setstate_cython, sizeof(__pyx_k_MemoryWriter___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_NAME_RE_STRICT, __pyx_k_NAME_RE_STRICT, sizeof(__pyx_k_NAME_RE_STRICT), 0, 0, 1, 1},
    {&__pyx_n_s_NAME_RE_UNICODE, __pyx_k_NAME_RE_UNICODE, sizeof(__pyx_k_NAME_RE_UNICODE), 0, 0, 1, 1},
    {&__pyx_n_s_NO_DEFAULT, __pyx_k_NO_DEFAULT, sizeof(__pyx_k_NO_DEFAULT), 0, 0, 1, 1},
    {&__pyx_n_u_NO_DEFAULT, __pyx_k_NO_DEFAULT, sizeof(__pyx_k_NO_DEFAULT), 0, 1, 0, 1},
    {&__pyx_kp_u_Name, __pyx_k_Name, sizeof(__pyx_k_Name), 0, 1, 0, 0},
    {&__pyx_n_s_NamedType, __pyx_k_NamedType, sizeof(__pyx_k_NamedType), 0, 0, 1, 1},
    {&__pyx_n_s_NamedType___reduce_cython, __pyx_k_NamedType___reduce_cython, sizeof(__pyx_k_NamedType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_NamedType___setstate_cython, __pyx_k_NamedType___setstate_cython, sizeof(__pyx_k_NamedType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_NamedType__get_schema_extra, __pyx_k_NamedType__get_schema_extra, sizeof(__pyx_k_NamedType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_NamedType_get_namespaced_aliase, __pyx_k_NamedType_get_namespaced_aliase, sizeof(__pyx_k_NamedType_get_namespaced_aliase), 0, 0, 1, 1},
    {&__pyx_kp_u_Namespace, __pyx_k_Namespace, sizeof(__pyx_k_Namespace), 0, 1, 0, 0},
    {&__pyx_kp_u_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 1, 0, 0},
    {&__pyx_n_s_NotImplemented, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
    {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
    {&__pyx_kp_u_Not_enough_input_data_to_read_va, __pyx_k_Not_enough_input_data_to_read_va, sizeof(__pyx_k_Not_enough_input_data_to_read_va), 0, 1, 0, 0},
    {&__pyx_n_s_NullCodec, __pyx_k_NullCodec, sizeof(__pyx_k_NullCodec), 0, 0, 1, 1},
    {&__pyx_n_s_NullCodec___reduce_cython, __pyx_k_NullCodec___reduce_cython, sizeof(__pyx_k_NullCodec___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_NullCodec___setstate_cython, __pyx_k_NullCodec___setstate_cython, sizeof(__pyx_k_NullCodec___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_NullType, __pyx_k_NullType, sizeof(__pyx_k_NullType), 0, 0, 1, 1},
    {&__pyx_n_s_NullType___reduce_cython, __pyx_k_NullType___reduce_cython, sizeof(__pyx_k_NullType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_NullType___setstate_cython, __pyx_k_NullType___setstate_cython, sizeof(__pyx_k_NullType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_NullType__convert_value, __pyx_k_NullType__convert_value, sizeof(__pyx_k_NullType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_NullType__get_schema_extra, __pyx_k_NullType__get_schema_extra, sizeof(__pyx_k_NullType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_NullType_copy, __pyx_k_NullType_copy, sizeof(__pyx_k_NullType_copy), 0, 0, 1, 1},
    {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
    {&__pyx_n_s_OBJ_FILE_METADATA, __pyx_k_OBJ_FILE_METADATA, sizeof(__pyx_k_OBJ_FILE_METADATA), 0, 0, 1, 1},
    {&__pyx_n_s_OBJ_MAGIC_BYTES, __pyx_k_OBJ_MAGIC_BYTES, sizeof(__pyx_k_OBJ_MAGIC_BYTES), 0, 0, 1, 1},
    {&__pyx_kp_b_Obj, __pyx_k_Obj, sizeof(__pyx_k_Obj), 0, 0, 0, 0},
    {&__pyx_n_s_Options, __pyx_k_Options, sizeof(__pyx_k_Options), 0, 0, 1, 1},
    {&__pyx_n_s_OptionsWrapper, __pyx_k_OptionsWrapper, sizeof(__pyx_k_OptionsWrapper), 0, 0, 1, 1},
    {&__pyx_n_s_Options___reduce_cython, __pyx_k_Options___reduce_cython, sizeof(__pyx_k_Options___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Options___setstate_cython, __pyx_k_Options___setstate_cython, sizeof(__pyx_k_Options___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Options_diff, __pyx_k_Options_diff, sizeof(__pyx_k_Options_diff), 0, 0, 1, 1},
    {&__pyx_n_s_Options_equals, __pyx_k_Options_equals, sizeof(__pyx_k_Options_equals), 0, 0, 1, 1},
    {&__pyx_n_s_Options_replace, __pyx_k_Options_replace, sizeof(__pyx_k_Options_replace), 0, 0, 1, 1},
    {&__pyx_n_s_Options_with_external_types, __pyx_k_Options_with_external_types, sizeof(__pyx_k_Options_with_external_types), 0, 0, 1, 1},
    {&__pyx_n_s_Options_with_logical_types, __pyx_k_Options_with_logical_types, sizeof(__pyx_k_Options_with_logical_types), 0, 0, 1, 1},
    {&__pyx_n_s_Order, __pyx_k_Order, sizeof(__pyx_k_Order), 0, 0, 1, 1},
    {&__pyx_kp_u_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 1, 0, 0},
    {&__pyx_n_s_OverflowError, __pyx_k_OverflowError, sizeof(__pyx_k_OverflowError), 0, 0, 1, 1},
    {&__pyx_n_s_PERMISSIVE_OPTIONS, __pyx_k_PERMISSIVE_OPTIONS, sizeof(__pyx_k_PERMISSIVE_OPTIONS), 0, 0, 1, 1},
    {&__pyx_n_s_PRIMITIVE_TYPES, __pyx_k_PRIMITIVE_TYPES, sizeof(__pyx_k_PRIMITIVE_TYPES), 0, 0, 1, 1},
    {&__pyx_n_s_Path, __pyx_k_Path, sizeof(__pyx_k_Path), 0, 0, 1, 1},
    {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_PlaceholderType, __pyx_k_PlaceholderType, sizeof(__pyx_k_PlaceholderType), 0, 0, 1, 1},
    {&__pyx_n_s_PlaceholderType___reduce_cython, __pyx_k_PlaceholderType___reduce_cython, sizeof(__pyx_k_PlaceholderType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PlaceholderType___setstate_cyth, __pyx_k_PlaceholderType___setstate_cyth, sizeof(__pyx_k_PlaceholderType___setstate_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_PlaceholderType__convert_value, __pyx_k_PlaceholderType__convert_value, sizeof(__pyx_k_PlaceholderType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_PlaceholderType__get_schema_ext, __pyx_k_PlaceholderType__get_schema_ext, sizeof(__pyx_k_PlaceholderType__get_schema_ext), 0, 0, 1, 1},
    {&__pyx_kp_u_Placeholder_types_cannot_be_enco, __pyx_k_Placeholder_types_cannot_be_enco, sizeof(__pyx_k_Placeholder_types_cannot_be_enco), 0, 1, 0, 0},
    {&__pyx_kp_u_Placeholder_types_have_no_canoni, __pyx_k_Placeholder_types_have_no_canoni, sizeof(__pyx_k_Placeholder_types_have_no_canoni), 0, 1, 0, 0},
    {&__pyx_kp_u_Precision, __pyx_k_Precision, sizeof(__pyx_k_Precision), 0, 1, 0, 0},
    {&__pyx_kp_u_Precision_is_too_large_for_fixed, __pyx_k_Precision_is_too_large_for_fixed, sizeof(__pyx_k_Precision_is_too_large_for_fixed), 0, 1, 0, 0},
    {&__pyx_kp_u_Precision_must_be_an_integer_bet, __pyx_k_Precision_must_be_an_integer_bet, sizeof(__pyx_k_Precision_must_be_an_integer_bet), 0, 1, 0, 0},
    {&__pyx_kp_u_Precision_must_be_greater_than_s, __pyx_k_Precision_must_be_greater_than_s, sizeof(__pyx_k_Precision_must_be_greater_than_s), 0, 1, 0, 0},
    {&__pyx_n_s_PromoteBytesToString, __pyx_k_PromoteBytesToString, sizeof(__pyx_k_PromoteBytesToString), 0, 0, 1, 1},
    {&__pyx_n_s_PromoteBytesToString___reduce_cy, __pyx_k_PromoteBytesToString___reduce_cy, sizeof(__pyx_k_PromoteBytesToString___reduce_cy), 0, 0, 1, 1},
    {&__pyx_n_s_PromoteBytesToString___setstate, __pyx_k_PromoteBytesToString___setstate, sizeof(__pyx_k_PromoteBytesToString___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_PromoteStringToBytes, __pyx_k_PromoteStringToBytes, sizeof(__pyx_k_PromoteStringToBytes), 0, 0, 1, 1},
    {&__pyx_n_s_PromoteStringToBytes___reduce_cy, __pyx_k_PromoteStringToBytes___reduce_cy, sizeof(__pyx_k_PromoteStringToBytes___reduce_cy), 0, 0, 1, 1},
    {&__pyx_n_s_PromoteStringToBytes___setstate, __pyx_k_PromoteStringToBytes___setstate, sizeof(__pyx_k_PromoteStringToBytes___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_PromoteToFloat, __pyx_k_PromoteToFloat, sizeof(__pyx_k_PromoteToFloat), 0, 0, 1, 1},
    {&__pyx_n_s_PromoteToFloat___reduce_cython, __pyx_k_PromoteToFloat___reduce_cython, sizeof(__pyx_k_PromoteToFloat___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PromoteToFloat___setstate_cython, __pyx_k_PromoteToFloat___setstate_cython, sizeof(__pyx_k_PromoteToFloat___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PromotingEnumType, __pyx_k_PromotingEnumType, sizeof(__pyx_k_PromotingEnumType), 0, 0, 1, 1},
    {&__pyx_n_s_PromotingEnumType___reduce_cytho, __pyx_k_PromotingEnumType___reduce_cytho, sizeof(__pyx_k_PromotingEnumType___reduce_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_PromotingEnumType___setstate_cyt, __pyx_k_PromotingEnumType___setstate_cyt, sizeof(__pyx_k_PromotingEnumType___setstate_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_PromotingRecordType, __pyx_k_PromotingRecordType, sizeof(__pyx_k_PromotingRecordType), 0, 0, 1, 1},
    {&__pyx_n_s_PromotingRecordType___reduce_cyt, __pyx_k_PromotingRecordType___reduce_cyt, sizeof(__pyx_k_PromotingRecordType___reduce_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_PromotingRecordType___setstate_c, __pyx_k_PromotingRecordType___setstate_c, sizeof(__pyx_k_PromotingRecordType___setstate_c), 0, 0, 1, 1},
    {&__pyx_kp_u_Promoting_schemas_reader_writer, __pyx_k_Promoting_schemas_reader_writer, sizeof(__pyx_k_Promoting_schemas_reader_writer), 0, 1, 0, 0},
    {&__pyx_n_s_RAW_VALUES, __pyx_k_RAW_VALUES, sizeof(__pyx_k_RAW_VALUES), 0, 0, 1, 1},
    {&__pyx_n_s_Rabin, __pyx_k_Rabin, sizeof(__pyx_k_Rabin), 0, 0, 1, 1},
    {&__pyx_n_s_Rabin___reduce_cython, __pyx_k_Rabin___reduce_cython, sizeof(__pyx_k_Rabin___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Rabin___setstate_cython, __pyx_k_Rabin___setstate_cython, sizeof(__pyx_k_Rabin___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Rabin_copy, __pyx_k_Rabin_copy, sizeof(__pyx_k_Rabin_copy), 0, 0, 1, 1},
    {&__pyx_n_s_Rabin_digest, __pyx_k_Rabin_digest, sizeof(__pyx_k_Rabin_digest), 0, 0, 1, 1},
    {&__pyx_n_s_Rabin_hexdigest, __pyx_k_Rabin_hexdigest, sizeof(__pyx_k_Rabin_hexdigest), 0, 0, 1, 1},
    {&__pyx_n_s_Rabin_update, __pyx_k_Rabin_update, sizeof(__pyx_k_Rabin_update), 0, 0, 1, 1},
    {&__pyx_n_s_Reader, __pyx_k_Reader, sizeof(__pyx_k_Reader), 0, 0, 1, 1},
    {&__pyx_n_s_Reader___reduce_cython, __pyx_k_Reader___reduce_cython, sizeof(__pyx_k_Reader___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Reader___setstate_cython, __pyx_k_Reader___setstate_cython, sizeof(__pyx_k_Reader___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Record, __pyx_k_Record, sizeof(__pyx_k_Record), 0, 1, 0, 0},
    {&__pyx_n_s_RecordField, __pyx_k_RecordField, sizeof(__pyx_k_RecordField), 0, 0, 1, 1},
    {&__pyx_n_s_RecordField___reduce_cython, __pyx_k_RecordField___reduce_cython, sizeof(__pyx_k_RecordField___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_RecordField___setstate_cython, __pyx_k_RecordField___setstate_cython, sizeof(__pyx_k_RecordField___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_RecordField_get_schema, __pyx_k_RecordField_get_schema, sizeof(__pyx_k_RecordField_get_schema), 0, 0, 1, 1},
    {&__pyx_n_s_RecordType, __pyx_k_RecordType, sizeof(__pyx_k_RecordType), 0, 0, 1, 1},
    {&__pyx_n_s_RecordType___reduce_cython, __pyx_k_RecordType___reduce_cython, sizeof(__pyx_k_RecordType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_RecordType___setstate_cython, __pyx_k_RecordType___setstate_cython, sizeof(__pyx_k_RecordType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_RecordType__convert_value, __pyx_k_RecordType__convert_value, sizeof(__pyx_k_RecordType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_RecordType__get_schema_extra, __pyx_k_RecordType__get_schema_extra, sizeof(__pyx_k_RecordType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_RecordType_copy, __pyx_k_RecordType_copy, sizeof(__pyx_k_RecordType_copy), 0, 0, 1, 1},
    {&__pyx_n_s_RecordType_walk_types, __pyx_k_RecordType_walk_types, sizeof(__pyx_k_RecordType_walk_types), 0, 0, 1, 1},
    {&__pyx_kp_u_Record_2, __pyx_k_Record_2, sizeof(__pyx_k_Record_2), 0, 1, 0, 0},
    {&__pyx_n_s_Record_3, __pyx_k_Record_3, sizeof(__pyx_k_Record_3), 0, 0, 1, 1},
    {&__pyx_n_s_Record___dir, __pyx_k_Record___dir, sizeof(__pyx_k_Record___dir), 0, 0, 1, 1},
    {&__pyx_n_s_Record___reduce_cython, __pyx_k_Record___reduce_cython, sizeof(__pyx_k_Record___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Record___setstate_cython, __pyx_k_Record___setstate_cython, sizeof(__pyx_k_Record___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Record__asdict, __pyx_k_Record__asdict, sizeof(__pyx_k_Record__asdict), 0, 0, 1, 1},
    {&__pyx_n_s_Record__repr_children_locals_gen, __pyx_k_Record__repr_children_locals_gen, sizeof(__pyx_k_Record__repr_children_locals_gen), 0, 0, 1, 1},
    {&__pyx_kp_u_Records_may_either_be_instantiat, __pyx_k_Records_may_either_be_instantiat, sizeof(__pyx_k_Records_may_either_be_instantiat), 0, 1, 0, 0},
    {&__pyx_kp_u_Records_may_either_be_instantiat_2, __pyx_k_Records_may_either_be_instantiat_2, sizeof(__pyx_k_Records_may_either_be_instantiat_2), 0, 1, 0, 0},
    {&__pyx_n_s_ResolvedSchema, __pyx_k_ResolvedSchema, sizeof(__pyx_k_ResolvedSchema), 0, 0, 1, 1},
    {&__pyx_n_s_ResolvedSchema___reduce_cython, __pyx_k_ResolvedSchema___reduce_cython, sizeof(__pyx_k_ResolvedSchema___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ResolvedSchema___setstate_cython, __pyx_k_ResolvedSchema___setstate_cython, sizeof(__pyx_k_ResolvedSchema___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ResourceWarning, __pyx_k_ResourceWarning, sizeof(__pyx_k_ResourceWarning), 0, 0, 1, 1},
    {&__pyx_n_s_SLICE_TYPE, __pyx_k_SLICE_TYPE, sizeof(__pyx_k_SLICE_TYPE), 0, 0, 1, 1},
    {&__pyx_n_s_SSIZE_TYPECODE, __pyx_k_SSIZE_TYPECODE, sizeof(__pyx_k_SSIZE_TYPECODE), 0, 0, 1, 1},
    {&__pyx_kp_u_Scale_must_be_a_positive_integer, __pyx_k_Scale_must_be_a_positive_integer, sizeof(__pyx_k_Scale_must_be_a_positive_integer), 0, 1, 0, 0},
    {&__pyx_n_s_Schema, __pyx_k_Schema, sizeof(__pyx_k_Schema), 0, 0, 1, 1},
    {&__pyx_n_s_Schema___reduce_cython, __pyx_k_Schema___reduce_cython, sizeof(__pyx_k_Schema___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Schema___setstate_cython, __pyx_k_Schema___setstate_cython, sizeof(__pyx_k_Schema___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Schema__make_logical_types, __pyx_k_Schema__make_logical_types, sizeof(__pyx_k_Schema__make_logical_types), 0, 0, 1, 1},
    {&__pyx_n_s_Schema__wrap_type, __pyx_k_Schema__wrap_type, sizeof(__pyx_k_Schema__wrap_type), 0, 0, 1, 1},
    {&__pyx_n_s_Schema_binary_decode, __pyx_k_Schema_binary_decode, sizeof(__pyx_k_Schema_binary_decode), 0, 0, 1, 1},
    {&__pyx_n_s_Schema_binary_encode, __pyx_k_Schema_binary_encode, sizeof(__pyx_k_Schema_binary_encode), 0, 0, 1, 1},
    {&__pyx_n_s_Schema_binary_read, __pyx_k_Schema_binary_read, sizeof(__pyx_k_Schema_binary_read), 0, 0, 1, 1},
    {&__pyx_n_s_Schema_binary_write, __pyx_k_Schema_binary_write, sizeof(__pyx_k_Schema_binary_write), 0, 0, 1, 1},
    {&__pyx_n_s_Schema_can_encode, __pyx_k_Schema_can_encode, sizeof(__pyx_k_Schema_can_encode), 0, 0, 1, 1},
    {&__pyx_n_s_Schema_find_type, __pyx_k_Schema_find_type, sizeof(__pyx_k_Schema_find_type), 0, 0, 1, 1},
    {&__pyx_n_s_Schema_fingerprint, __pyx_k_Schema_fingerprint, sizeof(__pyx_k_Schema_fingerprint), 0, 0, 1, 1},
    {&__pyx_kp_u_Schema_is_required, __pyx_k_Schema_is_required, sizeof(__pyx_k_Schema_is_required), 0, 1, 0, 0},
    {&__pyx_n_s_Schema_json_decode, __pyx_k_Schema_json_decode, sizeof(__pyx_k_Schema_json_decode), 0, 0, 1, 1},
    {&__pyx_n_s_Schema_json_encode, __pyx_k_Schema_json_encode, sizeof(__pyx_k_Schema_json_encode), 0, 0, 1, 1},
    {&__pyx_n_s_Schema_reader_for_writer, __pyx_k_Schema_reader_for_writer, sizeof(__pyx_k_Schema_reader_for_writer), 0, 0, 1, 1},
    {&__pyx_n_s_Sentinel, __pyx_k_Sentinel, sizeof(__pyx_k_Sentinel), 0, 0, 1, 1},
    {&__pyx_n_s_Sentinel___reduce_cython, __pyx_k_Sentinel___reduce_cython, sizeof(__pyx_k_Sentinel___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Sentinel___setstate_cython, __pyx_k_Sentinel___setstate_cython, sizeof(__pyx_k_Sentinel___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Sequence, __pyx_k_Sequence, sizeof(__pyx_k_Sequence), 0, 0, 1, 1},
    {&__pyx_kp_s_Sequence_str, __pyx_k_Sequence_str, sizeof(__pyx_k_Sequence_str), 0, 0, 1, 0},
    {&__pyx_n_s_SignatureWrapper, __pyx_k_SignatureWrapper, sizeof(__pyx_k_SignatureWrapper), 0, 0, 1, 1},
    {&__pyx_n_s_SignatureWrapper___init, __pyx_k_SignatureWrapper___init, sizeof(__pyx_k_SignatureWrapper___init), 0, 0, 1, 1},
    {&__pyx_n_s_SignatureWrapper__set_from_data, __pyx_k_SignatureWrapper__set_from_data, sizeof(__pyx_k_SignatureWrapper__set_from_data), 0, 0, 1, 1},
    {&__pyx_n_s_SnappyCodec, __pyx_k_SnappyCodec, sizeof(__pyx_k_SnappyCodec), 0, 0, 1, 1},
    {&__pyx_n_s_SnappyCodec___reduce_cython, __pyx_k_SnappyCodec___reduce_cython, sizeof(__pyx_k_SnappyCodec___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SnappyCodec___setstate_cython, __pyx_k_SnappyCodec___setstate_cython, sizeof(__pyx_k_SnappyCodec___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_s_Step_may_not_be_zero_axis_d, __pyx_k_Step_may_not_be_zero_axis_d, sizeof(__pyx_k_Step_may_not_be_zero_axis_d), 0, 0, 1, 0},
    {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
    {&__pyx_n_s_StringType, __pyx_k_StringType, sizeof(__pyx_k_StringType), 0, 0, 1, 1},
    {&__pyx_n_s_StringType___reduce_cython, __pyx_k_StringType___reduce_cython, sizeof(__pyx_k_StringType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_StringType___setstate_cython, __pyx_k_StringType___setstate_cython, sizeof(__pyx_k_StringType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_StringType__convert_value, __pyx_k_StringType__convert_value, sizeof(__pyx_k_StringType__convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_StringType__get_schema_extra, __pyx_k_StringType__get_schema_extra, sizeof(__pyx_k_StringType__get_schema_extra), 0, 0, 1, 1},
    {&__pyx_n_s_StringType_copy, __pyx_k_StringType_copy, sizeof(__pyx_k_StringType_copy), 0, 0, 1, 1},
    {&__pyx_n_s_TYPES_BY_NAME, __pyx_k_TYPES_BY_NAME, sizeof(__pyx_k_TYPES_BY_NAME), 0, 0, 1, 1},
    {&__pyx_n_s_TYPE_TUPLE_ALWAYS, __pyx_k_TYPE_TUPLE_ALWAYS, sizeof(__pyx_k_TYPE_TUPLE_ALWAYS), 0, 0, 1, 1},
    {&__pyx_n_s_TYPE_TUPLE_IF_AMBIGUOUS, __pyx_k_TYPE_TUPLE_IF_AMBIGUOUS, sizeof(__pyx_k_TYPE_TUPLE_IF_AMBIGUOUS), 0, 0, 1, 1},
    {&__pyx_n_s_TYPE_TUPLE_IF_RECORD, __pyx_k_TYPE_TUPLE_IF_RECORD, sizeof(__pyx_k_TYPE_TUPLE_IF_RECORD), 0, 0, 1, 1},
    {&__pyx_n_s_TYPE_TUPLE_IF_RECORD_AMBIGUOUS, __pyx_k_TYPE_TUPLE_IF_RECORD_AMBIGUOUS, sizeof(__pyx_k_TYPE_TUPLE_IF_RECORD_AMBIGUOUS), 0, 0, 1, 1},
    {&__pyx_kp_s_The_exponent_of_a_decimal_value, __pyx_k_The_exponent_of_a_decimal_value, sizeof(__pyx_k_The_exponent_of_a_decimal_value), 0, 0, 1, 0},
    {&__pyx_kp_u_The_null_namespace_cannot_be_spe, __pyx_k_The_null_namespace_cannot_be_spe, sizeof(__pyx_k_The_null_namespace_cannot_be_spe), 0, 1, 0, 0},
    {&__pyx_kp_s_The_order_of_a_field_in_a_recor, __pyx_k_The_order_of_a_field_in_a_recor, sizeof(__pyx_k_The_order_of_a_field_in_a_recor), 0, 0, 1, 0},
    {&__pyx_kp_s_The_schema_contains_a_type_or_e, __pyx_k_The_schema_contains_a_type_or_e, sizeof(__pyx_k_The_schema_contains_a_type_or_e), 0, 0, 1, 0},
    {&__pyx_kp_s_The_schema_contains_an_unexptec, __pyx_k_The_schema_contains_an_unexptec, sizeof(__pyx_k_The_schema_contains_an_unexptec), 0, 0, 1, 0},
    {&__pyx_n_s_TimeMicros, __pyx_k_TimeMicros, sizeof(__pyx_k_TimeMicros), 0, 0, 1, 1},
    {&__pyx_n_s_TimeMicros___reduce_cython, __pyx_k_TimeMicros___reduce_cython, sizeof(__pyx_k_TimeMicros___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TimeMicros___setstate_cython, __pyx_k_TimeMicros___setstate_cython, sizeof(__pyx_k_TimeMicros___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TimeMicros__for_type, __pyx_k_TimeMicros__for_type, sizeof(__pyx_k_TimeMicros__for_type), 0, 0, 1, 1},
    {&__pyx_n_s_TimeMillis, __pyx_k_TimeMillis, sizeof(__pyx_k_TimeMillis), 0, 0, 1, 1},
    {&__pyx_n_s_TimeMillis___reduce_cython, __pyx_k_TimeMillis___reduce_cython, sizeof(__pyx_k_TimeMillis___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TimeMillis___setstate_cython, __pyx_k_TimeMillis___setstate_cython, sizeof(__pyx_k_TimeMillis___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TimeMillis__for_type, __pyx_k_TimeMillis__for_type, sizeof(__pyx_k_TimeMillis__for_type), 0, 0, 1, 1},
    {&__pyx_n_s_TimestampMicros, __pyx_k_TimestampMicros, sizeof(__pyx_k_TimestampMicros), 0, 0, 1, 1},
    {&__pyx_n_s_TimestampMicros___reduce_cython, __pyx_k_TimestampMicros___reduce_cython, sizeof(__pyx_k_TimestampMicros___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TimestampMicros___setstate_cytho, __pyx_k_TimestampMicros___setstate_cytho, sizeof(__pyx_k_TimestampMicros___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_TimestampMicros__for_type, __pyx_k_TimestampMicros__for_type, sizeof(__pyx_k_TimestampMicros__for_type), 0, 0, 1, 1},
    {&__pyx_n_s_TimestampMillis, __pyx_k_TimestampMillis, sizeof(__pyx_k_TimestampMillis), 0, 0, 1, 1},
    {&__pyx_n_s_TimestampMillis___reduce_cython, __pyx_k_TimestampMillis___reduce_cython, sizeof(__pyx_k_TimestampMillis___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TimestampMillis___setstate_cytho, __pyx_k_TimestampMillis___setstate_cytho, sizeof(__pyx_k_TimestampMillis___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_TimestampMillis__for_type, __pyx_k_TimestampMillis__for_type, sizeof(__pyx_k_TimestampMillis__for_type), 0, 0, 1, 1},
    {&__pyx_n_u_True, __pyx_k_True, sizeof(__pyx_k_True), 0, 1, 0, 1},
    {&__pyx_kp_u_Trying_to_close_a_closed_Contain, __pyx_k_Trying_to_close_a_closed_Contain, sizeof(__pyx_k_Trying_to_close_a_closed_Contain), 0, 1, 0, 0},
    {&__pyx_kp_u_Trying_to_write_to_closed_Contai, __pyx_k_Trying_to_write_to_closed_Contai, sizeof(__pyx_k_Trying_to_write_to_closed_Contai), 0, 1, 0, 0},
    {&__pyx_n_s_Type, __pyx_k_Type, sizeof(__pyx_k_Type), 0, 0, 1, 1},
    {&__pyx_n_u_Type, __pyx_k_Type, sizeof(__pyx_k_Type), 0, 1, 0, 1},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_kp_u_Type_name, __pyx_k_Type_name, sizeof(__pyx_k_Type_name), 0, 1, 0, 0},
    {&__pyx_n_u_UNKNOWN, __pyx_k_UNKNOWN, sizeof(__pyx_k_UNKNOWN), 0, 1, 0, 1},
    {&__pyx_n_s_UUID, __pyx_k_UUID, sizeof(__pyx_k_UUID), 0, 0, 1, 1},
    {&__pyx_n_s_UUIDBase, __pyx_k_UUIDBase, sizeof(__pyx_k_UUIDBase), 0, 0, 1, 1},
    {&__pyx_n_s_UUIDBase___reduce_cython, __pyx_k_UUIDBase___reduce_cython, sizeof(__pyx_k_UUIDBase___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UUIDBase___setstate_cython, __pyx_k_UUIDBase___setstate_cython, sizeof(__pyx_k_UUIDBase___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UUIDBase__for_type, __pyx_k_UUIDBase__for_type, sizeof(__pyx_k_UUIDBase__for_type), 0, 0, 1, 1},
    {&__pyx_n_s_UUIDFixedType, __pyx_k_UUIDFixedType, sizeof(__pyx_k_UUIDFixedType), 0, 0, 1, 1},
    {&__pyx_n_s_UUIDFixedType___reduce_cython, __pyx_k_UUIDFixedType___reduce_cython, sizeof(__pyx_k_UUIDFixedType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UUIDFixedType___setstate_cython, __pyx_k_UUIDFixedType___setstate_cython, sizeof(__pyx_k_UUIDFixedType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UUIDFixedType__for_type, __pyx_k_UUIDFixedType__for_type, sizeof(__pyx_k_UUIDFixedType__for_type), 0, 0, 1, 1},
    {&__pyx_n_s_UUIDStringType, __pyx_k_UUIDStringType, sizeof(__pyx_k_UUIDStringType), 0, 0, 1, 1},
    {&__pyx_n_s_UUIDStringType___reduce_cython, __pyx_k_UUIDStringType___reduce_cython, sizeof(__pyx_k_UUIDStringType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UUIDStringType___setstate_cython, __pyx_k_UUIDStringType___setstate_cython, sizeof(__pyx_k_UUIDStringType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UUIDStringType__for_type, __pyx_k_UUIDStringType__for_type, sizeof(__pyx_k_UUIDStringType__for_type), 0, 0, 1, 1},
    {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
    {&__pyx_n_s_UnicodeEncodeError, __pyx_k_UnicodeEncodeError, sizeof(__pyx_k_UnicodeEncodeError), 0, 0, 1, 1},
    {&__pyx_n_s_Union, __pyx_k_Union, sizeof(__pyx_k_Union), 0, 0, 1, 1},
    {&__pyx_n_s_UnionDecodeOption, __pyx_k_UnionDecodeOption, sizeof(__pyx_k_UnionDecodeOption), 0, 0, 1, 1},
    {&__pyx_n_s_UnionType, __pyx_k_UnionType, sizeof(__pyx_k_UnionType), 0, 0, 1, 1},
    {&__pyx_n_s_UnionType___reduce_cython, __pyx_k_UnionType___reduce_cython, sizeof(__pyx_k_UnionType___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UnionType___setstate_cython, __pyx_k_UnionType___setstate_cython, sizeof(__pyx_k_UnionType___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UnionType_convert_value, __pyx_k_UnionType_convert_value, sizeof(__pyx_k_UnionType_convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_UnionType_copy, __pyx_k_UnionType_copy, sizeof(__pyx_k_UnionType_copy), 0, 0, 1, 1},
    {&__pyx_n_s_UnionType_get_schema, __pyx_k_UnionType_get_schema, sizeof(__pyx_k_UnionType_get_schema), 0, 0, 1, 1},
    {&__pyx_n_s_UnionType_walk_types, __pyx_k_UnionType_walk_types, sizeof(__pyx_k_UnionType_walk_types), 0, 0, 1, 1},
    {&__pyx_kp_s_Union_bytes_hashlib__hashlib_HAS, __pyx_k_Union_bytes_hashlib__hashlib_HAS, sizeof(__pyx_k_Union_bytes_hashlib__hashlib_HAS), 0, 0, 1, 0},
    {&__pyx_kp_s_Union_str_object, __pyx_k_Union_str_object, sizeof(__pyx_k_Union_str_object), 0, 0, 1, 0},
    {&__pyx_kp_u_Unions_may_not_have_more_than_on, __pyx_k_Unions_may_not_have_more_than_on, sizeof(__pyx_k_Unions_may_not_have_more_than_on), 0, 1, 0, 0},
    {&__pyx_kp_u_Unions_may_not_immediately_conta, __pyx_k_Unions_may_not_immediately_conta, sizeof(__pyx_k_Unions_may_not_immediately_conta), 0, 1, 0, 0},
    {&__pyx_kp_u_Unions_must_contain_at_least_one, __pyx_k_Unions_must_contain_at_least_one, sizeof(__pyx_k_Unions_must_contain_at_least_one), 0, 1, 0, 0},
    {&__pyx_n_s_UnknownType, __pyx_k_UnknownType, sizeof(__pyx_k_UnknownType), 0, 0, 1, 1},
    {&__pyx_n_s_UnknownType___init, __pyx_k_UnknownType___init, sizeof(__pyx_k_UnknownType___init), 0, 0, 1, 1},
    {&__pyx_kp_u_Unknown_hash_method, __pyx_k_Unknown_hash_method, sizeof(__pyx_k_Unknown_hash_method), 0, 1, 0, 0},
    {&__pyx_kp_u_Unknown_type, __pyx_k_Unknown_type, sizeof(__pyx_k_Unknown_type), 0, 1, 0, 0},
    {&__pyx_kp_u_Unknown_value_for_enum, __pyx_k_Unknown_value_for_enum, sizeof(__pyx_k_Unknown_value_for_enum), 0, 1, 0, 0},
    {&__pyx_kp_u_Unsupported_codec, __pyx_k_Unsupported_codec, sizeof(__pyx_k_Unsupported_codec), 0, 1, 0, 0},
    {&__pyx_kp_u_Value, __pyx_k_Value, sizeof(__pyx_k_Value), 0, 1, 0, 0},
    {&__pyx_n_s_ValueAdapter, __pyx_k_ValueAdapter, sizeof(__pyx_k_ValueAdapter), 0, 0, 1, 1},
    {&__pyx_n_s_ValueAdapter___reduce_cython, __pyx_k_ValueAdapter___reduce_cython, sizeof(__pyx_k_ValueAdapter___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ValueAdapter___setstate_cython, __pyx_k_ValueAdapter___setstate_cython, sizeof(__pyx_k_ValueAdapter___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
    {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
    {&__pyx_n_s_Writer, __pyx_k_Writer, sizeof(__pyx_k_Writer), 0, 0, 1, 1},
    {&__pyx_n_s_Writer___reduce_cython, __pyx_k_Writer___reduce_cython, sizeof(__pyx_k_Writer___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Writer___setstate_cython, __pyx_k_Writer___setstate_cython, sizeof(__pyx_k_Writer___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Y_m_d, __pyx_k_Y_m_d, sizeof(__pyx_k_Y_m_d), 0, 1, 0, 0},
    {&__pyx_n_s_ZStandardCodec, __pyx_k_ZStandardCodec, sizeof(__pyx_k_ZStandardCodec), 0, 0, 1, 1},
    {&__pyx_n_s_ZStandardCodec___reduce_cython, __pyx_k_ZStandardCodec___reduce_cython, sizeof(__pyx_k_ZStandardCodec___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ZStandardCodec___setstate_cytho, __pyx_k_ZStandardCodec___setstate_cytho, sizeof(__pyx_k_ZStandardCodec___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_ZstdCompressor, __pyx_k_ZstdCompressor, sizeof(__pyx_k_ZstdCompressor), 0, 0, 1, 1},
    {&__pyx_n_s_ZstdDecompressor, __pyx_k_ZstdDecompressor, sizeof(__pyx_k_ZstdDecompressor), 0, 0, 1, 1},
    {&__pyx_n_s__122, __pyx_k__122, sizeof(__pyx_k__122), 0, 0, 1, 1},
    {&__pyx_kp_b__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 0, 0, 0},
    {&__pyx_kp_u__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 1, 0, 0},
    {&__pyx_kp_u__14, __pyx_k__14, sizeof(__pyx_k__14), 0, 1, 0, 0},
    {&__pyx_n_s__152, __pyx_k__152, sizeof(__pyx_k__152), 0, 0, 1, 1},
    {&__pyx_kp_s__18, __pyx_k__18, sizeof(__pyx_k__18), 0, 0, 1, 0},
    {&__pyx_kp_u__18, __pyx_k__18, sizeof(__pyx_k__18), 0, 1, 0, 0},
    {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
    {&__pyx_kp_u__24, __pyx_k__24, sizeof(__pyx_k__24), 0, 1, 0, 0},
    {&__pyx_kp_u__25, __pyx_k__25, sizeof(__pyx_k__25), 0, 1, 0, 0},
    {&__pyx_kp_u__26, __pyx_k__26, sizeof(__pyx_k__26), 0, 1, 0, 0},
    {&__pyx_kp_u__27, __pyx_k__27, sizeof(__pyx_k__27), 0, 1, 0, 0},
    {&__pyx_n_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 1},
    {&__pyx_kp_b__34, __pyx_k__34, sizeof(__pyx_k__34), 0, 0, 0, 0},
    {&__pyx_kp_u__48, __pyx_k__48, sizeof(__pyx_k__48), 0, 1, 0, 0},
    {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
    {&__pyx_kp_u__60, __pyx_k__60, sizeof(__pyx_k__60), 0, 1, 0, 0},
    {&__pyx_kp_u__61, __pyx_k__61, sizeof(__pyx_k__61), 0, 1, 0, 0},
    {&__pyx_kp_u__62, __pyx_k__62, sizeof(__pyx_k__62), 0, 1, 0, 0},
    {&__pyx_kp_u__63, __pyx_k__63, sizeof(__pyx_k__63), 0, 1, 0, 0},
    {&__pyx_kp_u__64, __pyx_k__64, sizeof(__pyx_k__64), 0, 1, 0, 0},
    {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
    {&__pyx_kp_u__70, __pyx_k__70, sizeof(__pyx_k__70), 0, 1, 0, 0},
    {&__pyx_kp_b__72, __pyx_k__72, sizeof(__pyx_k__72), 0, 0, 0, 0},
    {&__pyx_kp_u__76, __pyx_k__76, sizeof(__pyx_k__76), 0, 1, 0, 0},
    {&__pyx_kp_u__77, __pyx_k__77, sizeof(__pyx_k__77), 0, 1, 0, 0},
    {&__pyx_kp_u__78, __pyx_k__78, sizeof(__pyx_k__78), 0, 1, 0, 0},
    {&__pyx_kp_u__79, __pyx_k__79, sizeof(__pyx_k__79), 0, 1, 0, 0},
    {&__pyx_kp_u__80, __pyx_k__80, sizeof(__pyx_k__80), 0, 1, 0, 0},
    {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
    {&__pyx_n_s_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
    {&__pyx_n_b_abacus, __pyx_k_abacus, sizeof(__pyx_k_abacus), 0, 0, 0, 1},
    {&__pyx_n_b_abacuscounter1, __pyx_k_abacuscounter1, sizeof(__pyx_k_abacuscounter1), 0, 0, 0, 1},
    {&__pyx_n_s_abc, __pyx_k_abc, sizeof(__pyx_k_abc), 0, 0, 1, 1},
    {&__pyx_n_s_accepts_string, __pyx_k_accepts_string, sizeof(__pyx_k_accepts_string), 0, 0, 1, 1},
    {&__pyx_n_s_actual, __pyx_k_actual, sizeof(__pyx_k_actual), 0, 0, 1, 1},
    {&__pyx_n_s_adapt_record_types, __pyx_k_adapt_record_types, sizeof(__pyx_k_adapt_record_types), 0, 0, 1, 1},
    {&__pyx_kp_u_adapt_record_types_2, __pyx_k_adapt_record_types_2, sizeof(__pyx_k_adapt_record_types_2), 0, 1, 0, 0},
    {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
    {&__pyx_n_u_aliases, __pyx_k_aliases, sizeof(__pyx_k_aliases), 0, 1, 0, 1},
    {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
    {&__pyx_n_s_allow_aliases_to_be_string, __pyx_k_allow_aliases_to_be_string, sizeof(__pyx_k_allow_aliases_to_be_string), 0, 0, 1, 1},
    {&__pyx_kp_u_allow_aliases_to_be_string_2, __pyx_k_allow_aliases_to_be_string_2, sizeof(__pyx_k_allow_aliases_to_be_string_2), 0, 1, 0, 0},
    {&__pyx_n_s_allow_duplicate_union_types, __pyx_k_allow_duplicate_union_types, sizeof(__pyx_k_allow_duplicate_union_types), 0, 0, 1, 1},
    {&__pyx_kp_u_allow_duplicate_union_types_2, __pyx_k_allow_duplicate_union_types_2, sizeof(__pyx_k_allow_duplicate_union_types_2), 0, 1, 0, 0},
    {&__pyx_n_s_allow_empty_unions, __pyx_k_allow_empty_unions, sizeof(__pyx_k_allow_empty_unions), 0, 0, 1, 1},
    {&__pyx_kp_u_allow_empty_unions_2, __pyx_k_allow_empty_unions_2, sizeof(__pyx_k_allow_empty_unions_2), 0, 1, 0, 0},
    {&__pyx_n_s_allow_error_type, __pyx_k_allow_error_type, sizeof(__pyx_k_allow_error_type), 0, 0, 1, 1},
    {&__pyx_kp_u_allow_error_type_2, __pyx_k_allow_error_type_2, sizeof(__pyx_k_allow_error_type_2), 0, 1, 0, 0},
    {&__pyx_n_s_allow_false_values_for_null, __pyx_k_allow_false_values_for_null, sizeof(__pyx_k_allow_false_values_for_null), 0, 0, 1, 1},
    {&__pyx_kp_u_allow_false_values_for_null_2, __pyx_k_allow_false_values_for_null_2, sizeof(__pyx_k_allow_false_values_for_null_2), 0, 1, 0, 0},
    {&__pyx_n_s_allow_invalid_default_values, __pyx_k_allow_invalid_default_values, sizeof(__pyx_k_allow_invalid_default_values), 0, 0, 1, 1},
    {&__pyx_kp_u_allow_invalid_default_values_2, __pyx_k_allow_invalid_default_values_2, sizeof(__pyx_k_allow_invalid_default_values_2), 0, 1, 0, 0},
    {&__pyx_n_s_allow_leading_dot_in_names, __pyx_k_allow_leading_dot_in_names, sizeof(__pyx_k_allow_leading_dot_in_names), 0, 0, 1, 1},
    {&__pyx_kp_u_allow_leading_dot_in_names_2, __pyx_k_allow_leading_dot_in_names_2, sizeof(__pyx_k_allow_leading_dot_in_names_2), 0, 1, 0, 0},
    {&__pyx_n_s_allow_nested_unions, __pyx_k_allow_nested_unions, sizeof(__pyx_k_allow_nested_unions), 0, 0, 1, 1},
    {&__pyx_kp_u_allow_nested_unions_2, __pyx_k_allow_nested_unions_2, sizeof(__pyx_k_allow_nested_unions_2), 0, 1, 0, 0},
    {&__pyx_n_s_allow_primitive_name_collision, __pyx_k_allow_primitive_name_collision, sizeof(__pyx_k_allow_primitive_name_collision), 0, 0, 1, 1},
    {&__pyx_kp_u_allow_primitive_name_collision_2, __pyx_k_allow_primitive_name_collision_2, sizeof(__pyx_k_allow_primitive_name_collision_2), 0, 1, 0, 0},
    {&__pyx_kp_u_allow_primitive_names_in_namesp, __pyx_k_allow_primitive_names_in_namesp, sizeof(__pyx_k_allow_primitive_names_in_namesp), 0, 1, 0, 0},
    {&__pyx_n_s_allow_primitive_names_in_namespa, __pyx_k_allow_primitive_names_in_namespa, sizeof(__pyx_k_allow_primitive_names_in_namespa), 0, 0, 1, 1},
    {&__pyx_n_s_allow_tuple_notation, __pyx_k_allow_tuple_notation, sizeof(__pyx_k_allow_tuple_notation), 0, 0, 1, 1},
    {&__pyx_kp_u_allow_tuple_notation_2, __pyx_k_allow_tuple_notation_2, sizeof(__pyx_k_allow_tuple_notation_2), 0, 1, 0, 0},
    {&__pyx_n_s_allow_union_default_any_member, __pyx_k_allow_union_default_any_member, sizeof(__pyx_k_allow_union_default_any_member), 0, 0, 1, 1},
    {&__pyx_kp_u_allow_union_default_any_member_2, __pyx_k_allow_union_default_any_member_2, sizeof(__pyx_k_allow_union_default_any_member_2), 0, 1, 0, 0},
    {&__pyx_n_s_alternate_timestamp_encoding, __pyx_k_alternate_timestamp_encoding, sizeof(__pyx_k_alternate_timestamp_encoding), 0, 0, 1, 1},
    {&__pyx_kp_u_alternate_timestamp_millis_enco, __pyx_k_alternate_timestamp_millis_enco, sizeof(__pyx_k_alternate_timestamp_millis_enco), 0, 1, 0, 0},
    {&__pyx_n_s_alternate_timestamp_millis_encod, __pyx_k_alternate_timestamp_millis_encod, sizeof(__pyx_k_alternate_timestamp_millis_encod), 0, 0, 1, 1},
    {&__pyx_kp_u_and, __pyx_k_and, sizeof(__pyx_k_and), 0, 1, 0, 0},
    {&__pyx_n_s_annotations, __pyx_k_annotations, sizeof(__pyx_k_annotations), 0, 0, 1, 1},
    {&__pyx_n_s_annotations_2, __pyx_k_annotations_2, sizeof(__pyx_k_annotations_2), 0, 0, 1, 1},
    {&__pyx_kp_u_appears_multiple_times_in_schem, __pyx_k_appears_multiple_times_in_schem, sizeof(__pyx_k_appears_multiple_times_in_schem), 0, 1, 0, 0},
    {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
    {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
    {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
    {&__pyx_n_u_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 1, 0, 1},
    {&__pyx_n_s_as_tuple, __pyx_k_as_tuple, sizeof(__pyx_k_as_tuple), 0, 0, 1, 1},
    {&__pyx_n_u_ascending, __pyx_k_ascending, sizeof(__pyx_k_ascending), 0, 1, 0, 1},
    {&__pyx_n_s_ascii_name_rules, __pyx_k_ascii_name_rules, sizeof(__pyx_k_ascii_name_rules), 0, 0, 1, 1},
    {&__pyx_kp_u_ascii_name_rules_2, __pyx_k_ascii_name_rules_2, sizeof(__pyx_k_ascii_name_rules_2), 0, 1, 0, 0},
    {&__pyx_n_s_asdict, __pyx_k_asdict, sizeof(__pyx_k_asdict), 0, 0, 1, 1},
    {&__pyx_n_u_asdict, __pyx_k_asdict, sizeof(__pyx_k_asdict), 0, 1, 0, 1},
    {&__pyx_n_s_astimezone, __pyx_k_astimezone, sizeof(__pyx_k_astimezone), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_kp_u_at, __pyx_k_at, sizeof(__pyx_k_at), 0, 1, 0, 0},
    {&__pyx_kp_u_avro_codec, __pyx_k_avro_codec, sizeof(__pyx_k_avro_codec), 0, 1, 0, 0},
    {&__pyx_kp_u_avro_schema, __pyx_k_avro_schema, sizeof(__pyx_k_avro_schema), 0, 1, 0, 0},
    {&__pyx_n_s_avro_type, __pyx_k_avro_type, sizeof(__pyx_k_avro_type), 0, 0, 1, 1},
    {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
    {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
    {&__pyx_n_u_big, __pyx_k_big, sizeof(__pyx_k_big), 0, 1, 0, 1},
    {&__pyx_n_s_bin, __pyx_k_bin, sizeof(__pyx_k_bin), 0, 0, 1, 1},
    {&__pyx_n_s_binary_decode, __pyx_k_binary_decode, sizeof(__pyx_k_binary_decode), 0, 0, 1, 1},
    {&__pyx_n_s_binary_encode, __pyx_k_binary_encode, sizeof(__pyx_k_binary_encode), 0, 0, 1, 1},
    {&__pyx_n_s_binary_read, __pyx_k_binary_read, sizeof(__pyx_k_binary_read), 0, 0, 1, 1},
    {&__pyx_n_s_binary_write, __pyx_k_binary_write, sizeof(__pyx_k_binary_write), 0, 0, 1, 1},
    {&__pyx_n_s_bint, __pyx_k_bint, sizeof(__pyx_k_bint), 0, 0, 1, 1},
    {&__pyx_n_u_bint, __pyx_k_bint, sizeof(__pyx_k_bint), 0, 1, 0, 1},
    {&__pyx_n_s_bit_length, __pyx_k_bit_length, sizeof(__pyx_k_bit_length), 0, 0, 1, 1},
    {&__pyx_n_s_block_size, __pyx_k_block_size, sizeof(__pyx_k_block_size), 0, 0, 1, 1},
    {&__pyx_n_s_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 0, 1, 1},
    {&__pyx_n_u_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 1, 0, 1},
    {&__pyx_n_s_bool_2, __pyx_k_bool_2, sizeof(__pyx_k_bool_2), 0, 0, 1, 1},
    {&__pyx_kp_u_boolean, __pyx_k_boolean, sizeof(__pyx_k_boolean), 0, 1, 0, 0},
    {&__pyx_n_u_boolean_2, __pyx_k_boolean_2, sizeof(__pyx_k_boolean_2), 0, 1, 0, 1},
    {&__pyx_n_s_buffer, __pyx_k_buffer, sizeof(__pyx_k_buffer), 0, 0, 1, 1},
    {&__pyx_n_s_byteorder, __pyx_k_byteorder, sizeof(__pyx_k_byteorder), 0, 0, 1, 1},
    {&__pyx_kp_u_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 1, 0, 0},
    {&__pyx_kp_u_bytes_2, __pyx_k_bytes_2, sizeof(__pyx_k_bytes_2), 0, 1, 0, 0},
    {&__pyx_n_s_bytes_3, __pyx_k_bytes_3, sizeof(__pyx_k_bytes_3), 0, 0, 1, 1},
    {&__pyx_n_u_bytes_3, __pyx_k_bytes_3, sizeof(__pyx_k_bytes_3), 0, 1, 0, 1},
    {&__pyx_n_s_bytes_codec, __pyx_k_bytes_codec, sizeof(__pyx_k_bytes_codec), 0, 0, 1, 1},
    {&__pyx_kp_u_bytes_codec_2, __pyx_k_bytes_codec_2, sizeof(__pyx_k_bytes_codec_2), 0, 1, 0, 0},
    {&__pyx_n_s_bytes_default_value_utf8, __pyx_k_bytes_default_value_utf8, sizeof(__pyx_k_bytes_default_value_utf8), 0, 0, 1, 1},
    {&__pyx_kp_u_bytes_default_value_utf8_2, __pyx_k_bytes_default_value_utf8_2, sizeof(__pyx_k_bytes_default_value_utf8_2), 0, 1, 0, 0},
    {&__pyx_n_s_bz2, __pyx_k_bz2, sizeof(__pyx_k_bz2), 0, 0, 1, 1},
    {&__pyx_n_b_bzip2, __pyx_k_bzip2, sizeof(__pyx_k_bzip2), 0, 0, 0, 1},
    {&__pyx_n_u_bzip2, __pyx_k_bzip2, sizeof(__pyx_k_bzip2), 0, 1, 0, 1},
    {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
    {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
    {&__pyx_n_s_can_encode, __pyx_k_can_encode, sizeof(__pyx_k_can_encode), 0, 0, 1, 1},
    {&__pyx_kp_u_cannot_be_adapted_to, __pyx_k_cannot_be_adapted_to, sizeof(__pyx_k_cannot_be_adapted_to), 0, 1, 0, 0},
    {&__pyx_n_s_canonical_form, __pyx_k_canonical_form, sizeof(__pyx_k_canonical_form), 0, 0, 1, 1},
    {&__pyx_n_s_canonical_form_repeat_enum, __pyx_k_canonical_form_repeat_enum, sizeof(__pyx_k_canonical_form_repeat_enum), 0, 0, 1, 1},
    {&__pyx_kp_u_canonical_form_repeat_enum_2, __pyx_k_canonical_form_repeat_enum_2, sizeof(__pyx_k_canonical_form_repeat_enum_2), 0, 1, 0, 0},
    {&__pyx_n_s_canonical_form_repeat_fixed, __pyx_k_canonical_form_repeat_fixed, sizeof(__pyx_k_canonical_form_repeat_fixed), 0, 0, 1, 1},
    {&__pyx_kp_u_canonical_form_repeat_fixed_2, __pyx_k_canonical_form_repeat_fixed_2, sizeof(__pyx_k_canonical_form_repeat_fixed_2), 0, 1, 0, 0},
    {&__pyx_n_s_category, __pyx_k_category, sizeof(__pyx_k_category), 0, 0, 1, 1},
    {&__pyx_n_s_cavro, __pyx_k_cavro, sizeof(__pyx_k_cavro), 0, 0, 1, 1},
    {&__pyx_kp_s_cavro_pyx, __pyx_k_cavro_pyx, sizeof(__pyx_k_cavro_pyx), 0, 0, 1, 0},
    {&__pyx_n_s_changes, __pyx_k_changes, sizeof(__pyx_k_changes), 0, 0, 1, 1},
    {&__pyx_n_s_check_exp_overflow, __pyx_k_check_exp_overflow, sizeof(__pyx_k_check_exp_overflow), 0, 0, 1, 1},
    {&__pyx_n_s_check_value, __pyx_k_check_value, sizeof(__pyx_k_check_value), 0, 0, 1, 1},
    {&__pyx_n_s_chr, __pyx_k_chr, sizeof(__pyx_k_chr), 0, 0, 1, 1},
    {&__pyx_n_s_clamp, __pyx_k_clamp, sizeof(__pyx_k_clamp), 0, 0, 1, 1},
    {&__pyx_n_s_clamp_float_overflow, __pyx_k_clamp_float_overflow, sizeof(__pyx_k_clamp_float_overflow), 0, 0, 1, 1},
    {&__pyx_kp_u_clamp_float_overflow_2, __pyx_k_clamp_float_overflow_2, sizeof(__pyx_k_clamp_float_overflow_2), 0, 1, 0, 0},
    {&__pyx_n_s_clamp_int_overflow, __pyx_k_clamp_int_overflow, sizeof(__pyx_k_clamp_int_overflow), 0, 0, 1, 1},
    {&__pyx_kp_u_clamp_int_overflow_2, __pyx_k_clamp_int_overflow_2, sizeof(__pyx_k_clamp_int_overflow_2), 0, 1, 0, 0},
    {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_s_class_inst_method, __pyx_k_class_inst_method, sizeof(__pyx_k_class_inst_method), 0, 0, 1, 1},
    {&__pyx_n_s_class_inst_method___get, __pyx_k_class_inst_method___get, sizeof(__pyx_k_class_inst_method___get), 0, 0, 1, 1},
    {&__pyx_n_s_class_inst_method___init, __pyx_k_class_inst_method___init, sizeof(__pyx_k_class_inst_method___init), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
    {&__pyx_n_s_codec, __pyx_k_codec, sizeof(__pyx_k_codec), 0, 0, 1, 1},
    {&__pyx_n_s_coerce_int_to_float, __pyx_k_coerce_int_to_float, sizeof(__pyx_k_coerce_int_to_float), 0, 0, 1, 1},
    {&__pyx_kp_u_coerce_int_to_float_2, __pyx_k_coerce_int_to_float_2, sizeof(__pyx_k_coerce_int_to_float_2), 0, 1, 0, 0},
    {&__pyx_n_s_coerce_values_to_boolean, __pyx_k_coerce_values_to_boolean, sizeof(__pyx_k_coerce_values_to_boolean), 0, 0, 1, 1},
    {&__pyx_kp_u_coerce_values_to_boolean_2, __pyx_k_coerce_values_to_boolean_2, sizeof(__pyx_k_coerce_values_to_boolean_2), 0, 1, 0, 0},
    {&__pyx_n_s_coerce_values_to_float, __pyx_k_coerce_values_to_float, sizeof(__pyx_k_coerce_values_to_float), 0, 0, 1, 1},
    {&__pyx_kp_u_coerce_values_to_float_2, __pyx_k_coerce_values_to_float_2, sizeof(__pyx_k_coerce_values_to_float_2), 0, 1, 0, 0},
    {&__pyx_n_s_coerce_values_to_int, __pyx_k_coerce_values_to_int, sizeof(__pyx_k_coerce_values_to_int), 0, 0, 1, 1},
    {&__pyx_kp_u_coerce_values_to_int_2, __pyx_k_coerce_values_to_int_2, sizeof(__pyx_k_coerce_values_to_int_2), 0, 1, 0, 0},
    {&__pyx_n_s_coerce_values_to_str, __pyx_k_coerce_values_to_str, sizeof(__pyx_k_coerce_values_to_str), 0, 0, 1, 1},
    {&__pyx_kp_u_coerce_values_to_str_2, __pyx_k_coerce_values_to_str_2, sizeof(__pyx_k_coerce_values_to_str_2), 0, 1, 0, 0},
    {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
    {&__pyx_kp_s_collections_abc, __pyx_k_collections_abc, sizeof(__pyx_k_collections_abc), 0, 0, 1, 0},
    {&__pyx_n_s_compare, __pyx_k_compare, sizeof(__pyx_k_compare), 0, 0, 1, 1},
    {&__pyx_n_s_compile, __pyx_k_compile, sizeof(__pyx_k_compile), 0, 0, 1, 1},
    {&__pyx_n_s_compress, __pyx_k_compress, sizeof(__pyx_k_compress), 0, 0, 1, 1},
    {&__pyx_n_s_container_fill_blocks, __pyx_k_container_fill_blocks, sizeof(__pyx_k_container_fill_blocks), 0, 0, 1, 1},
    {&__pyx_kp_u_container_fill_blocks_2, __pyx_k_container_fill_blocks_2, sizeof(__pyx_k_container_fill_blocks_2), 0, 1, 0, 0},
    {&__pyx_n_s_context, __pyx_k_context, sizeof(__pyx_k_context), 0, 0, 1, 1},
    {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
    {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
    {&__pyx_n_s_convert_value, __pyx_k_convert_value, sizeof(__pyx_k_convert_value), 0, 0, 1, 1},
    {&__pyx_n_s_convert_value_2, __pyx_k_convert_value_2, sizeof(__pyx_k_convert_value_2), 0, 0, 1, 1},
    {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
    {&__pyx_n_s_copy_time, __pyx_k_copy_time, sizeof(__pyx_k_copy_time), 0, 0, 1, 1},
    {&__pyx_n_s_count, __pyx_k_count, sizeof(__pyx_k_count), 0, 0, 1, 1},
    {&__pyx_n_b_counter, __pyx_k_counter, sizeof(__pyx_k_counter), 0, 0, 0, 1},
    {&__pyx_n_s_crc32, __pyx_k_crc32, sizeof(__pyx_k_crc32), 0, 0, 1, 1},
    {&__pyx_n_s_create_decimal, __pyx_k_create_decimal, sizeof(__pyx_k_create_decimal), 0, 0, 1, 1},
    {&__pyx_n_s_created, __pyx_k_created, sizeof(__pyx_k_created), 0, 0, 1, 1},
    {&__pyx_n_s_custom_decode_value, __pyx_k_custom_decode_value, sizeof(__pyx_k_custom_decode_value), 0, 0, 1, 1},
    {&__pyx_n_s_custom_encode_value, __pyx_k_custom_encode_value, sizeof(__pyx_k_custom_encode_value), 0, 0, 1, 1},
    {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
    {&__pyx_n_s_dataclass_fields, __pyx_k_dataclass_fields, sizeof(__pyx_k_dataclass_fields), 0, 0, 1, 1},
    {&__pyx_n_s_dataclass_params, __pyx_k_dataclass_params, sizeof(__pyx_k_dataclass_params), 0, 0, 1, 1},
    {&__pyx_n_s_dataclasses, __pyx_k_dataclasses, sizeof(__pyx_k_dataclasses), 0, 0, 1, 1},
    {&__pyx_n_s_date, __pyx_k_date, sizeof(__pyx_k_date), 0, 0, 1, 1},
    {&__pyx_n_u_date, __pyx_k_date, sizeof(__pyx_k_date), 0, 1, 0, 1},
    {&__pyx_n_s_date_type_accepts_string, __pyx_k_date_type_accepts_string, sizeof(__pyx_k_date_type_accepts_string), 0, 0, 1, 1},
    {&__pyx_kp_u_date_type_accepts_string_2, __pyx_k_date_type_accepts_string_2, sizeof(__pyx_k_date_type_accepts_string_2), 0, 1, 0, 0},
    {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
    {&__pyx_n_s_days, __pyx_k_days, sizeof(__pyx_k_days), 0, 0, 1, 1},
    {&__pyx_n_s_decimal, __pyx_k_decimal, sizeof(__pyx_k_decimal), 0, 0, 1, 1},
    {&__pyx_n_u_decimal, __pyx_k_decimal, sizeof(__pyx_k_decimal), 0, 1, 0, 1},
    {&__pyx_kp_u_decimal_2, __pyx_k_decimal_2, sizeof(__pyx_k_decimal_2), 0, 1, 0, 0},
    {&__pyx_n_s_decimal_check_exp_overflow, __pyx_k_decimal_check_exp_overflow, sizeof(__pyx_k_decimal_check_exp_overflow), 0, 0, 1, 1},
    {&__pyx_kp_u_decimal_check_exp_overflow_2, __pyx_k_decimal_check_exp_overflow_2, sizeof(__pyx_k_decimal_check_exp_overflow_2), 0, 1, 0, 0},
    {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
    {&__pyx_n_s_decompress, __pyx_k_decompress, sizeof(__pyx_k_decompress), 0, 0, 1, 1},
    {&__pyx_n_s_decompressobj, __pyx_k_decompressobj, sizeof(__pyx_k_decompressobj), 0, 0, 1, 1},
    {&__pyx_n_s_defaul_value, __pyx_k_defaul_value, sizeof(__pyx_k_defaul_value), 0, 0, 1, 1},
    {&__pyx_n_s_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 0, 1, 1},
    {&__pyx_n_u_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 1, 0, 1},
    {&__pyx_n_s_default_value, __pyx_k_default_value, sizeof(__pyx_k_default_value), 0, 0, 1, 1},
    {&__pyx_n_s_defer_schema_promotion_errors, __pyx_k_defer_schema_promotion_errors, sizeof(__pyx_k_defer_schema_promotion_errors), 0, 0, 1, 1},
    {&__pyx_kp_u_defer_schema_promotion_errors_2, __pyx_k_defer_schema_promotion_errors_2, sizeof(__pyx_k_defer_schema_promotion_errors_2), 0, 1, 0, 0},
    {&__pyx_n_b_deflate, __pyx_k_deflate, sizeof(__pyx_k_deflate), 0, 0, 0, 1},
    {&__pyx_n_u_deflate, __pyx_k_deflate, sizeof(__pyx_k_deflate), 0, 1, 0, 1},
    {&__pyx_n_u_descending, __pyx_k_descending, sizeof(__pyx_k_descending), 0, 1, 0, 1},
    {&__pyx_n_s_deserialize, __pyx_k_deserialize, sizeof(__pyx_k_deserialize), 0, 0, 1, 1},
    {&__pyx_n_s_dest, __pyx_k_dest, sizeof(__pyx_k_dest), 0, 0, 1, 1},
    {&__pyx_n_s_dest_type, __pyx_k_dest_type, sizeof(__pyx_k_dest_type), 0, 0, 1, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
    {&__pyx_kp_s_dict_str_AvroType, __pyx_k_dict_str_AvroType, sizeof(__pyx_k_dict_str_AvroType), 0, 0, 1, 0},
    {&__pyx_kp_s_dict_str_tuple_object_object, __pyx_k_dict_str_tuple_object_object, sizeof(__pyx_k_dict_str_tuple_object_object), 0, 0, 1, 0},
    {&__pyx_n_s_dict_to_canonical_locals_genexpr, __pyx_k_dict_to_canonical_locals_genexpr, sizeof(__pyx_k_dict_to_canonical_locals_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_diff, __pyx_k_diff, sizeof(__pyx_k_diff), 0, 0, 1, 1},
    {&__pyx_n_s_digest, __pyx_k_digest, sizeof(__pyx_k_digest), 0, 0, 1, 1},
    {&__pyx_n_s_digest_size, __pyx_k_digest_size, sizeof(__pyx_k_digest_size), 0, 0, 1, 1},
    {&__pyx_n_s_dir, __pyx_k_dir, sizeof(__pyx_k_dir), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_u_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 1, 0, 1},
    {&__pyx_n_s_doc_2, __pyx_k_doc_2, sizeof(__pyx_k_doc_2), 0, 0, 1, 1},
    {&__pyx_kp_u_does_not_implement__binary_buff, __pyx_k_does_not_implement__binary_buff, sizeof(__pyx_k_does_not_implement__binary_buff), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement__convert_val, __pyx_k_does_not_implement__convert_val, sizeof(__pyx_k_does_not_implement__convert_val), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement__extract_met, __pyx_k_does_not_implement__extract_met, sizeof(__pyx_k_does_not_implement__extract_met), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement__get_schema, __pyx_k_does_not_implement__get_schema, sizeof(__pyx_k_does_not_implement__get_schema), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement__get_value_f, __pyx_k_does_not_implement__get_value_f, sizeof(__pyx_k_does_not_implement__get_value_f), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement__json_decode, __pyx_k_does_not_implement__json_decode, sizeof(__pyx_k_does_not_implement__json_decode), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement__json_format, __pyx_k_does_not_implement__json_format, sizeof(__pyx_k_does_not_implement__json_format), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement_binary_buffe, __pyx_k_does_not_implement_binary_buffe, sizeof(__pyx_k_does_not_implement_binary_buffe), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement_canonical_fo, __pyx_k_does_not_implement_canonical_fo, sizeof(__pyx_k_does_not_implement_canonical_fo), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement_clone, __pyx_k_does_not_implement_clone, sizeof(__pyx_k_does_not_implement_clone), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement_decode_value, __pyx_k_does_not_implement_decode_value, sizeof(__pyx_k_does_not_implement_decode_value), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement_encode_value, __pyx_k_does_not_implement_encode_value, sizeof(__pyx_k_does_not_implement_encode_value), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement_read_block, __pyx_k_does_not_implement_read_block, sizeof(__pyx_k_does_not_implement_read_block), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement_read_bytes, __pyx_k_does_not_implement_read_bytes, sizeof(__pyx_k_does_not_implement_read_bytes), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement_read_n, __pyx_k_does_not_implement_read_n, sizeof(__pyx_k_does_not_implement_read_n), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement_read_u8, __pyx_k_does_not_implement_read_u8, sizeof(__pyx_k_does_not_implement_read_u8), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement_write_block, __pyx_k_does_not_implement_write_block, sizeof(__pyx_k_does_not_implement_write_block), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement_write_n, __pyx_k_does_not_implement_write_n, sizeof(__pyx_k_does_not_implement_write_n), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_implement_write_u8, __pyx_k_does_not_implement_write_u8, sizeof(__pyx_k_does_not_implement_write_u8), 0, 1, 0, 0},
    {&__pyx_kp_u_double, __pyx_k_double, sizeof(__pyx_k_double), 0, 1, 0, 0},
    {&__pyx_n_u_double_2, __pyx_k_double_2, sizeof(__pyx_k_double_2), 0, 1, 0, 1},
    {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
    {&__pyx_n_s_dumps, __pyx_k_dumps, sizeof(__pyx_k_dumps), 0, 0, 1, 1},
    {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
    {&__pyx_n_s_enforce_enum_symbol_name_rules, __pyx_k_enforce_enum_symbol_name_rules, sizeof(__pyx_k_enforce_enum_symbol_name_rules), 0, 0, 1, 1},
    {&__pyx_kp_u_enforce_enum_symbol_name_rules_2, __pyx_k_enforce_enum_symbol_name_rules_2, sizeof(__pyx_k_enforce_enum_symbol_name_rules_2), 0, 1, 0, 0},
    {&__pyx_n_s_enforce_namespace_name_rules, __pyx_k_enforce_namespace_name_rules, sizeof(__pyx_k_enforce_namespace_name_rules), 0, 0, 1, 1},
    {&__pyx_kp_u_enforce_namespace_name_rules_2, __pyx_k_enforce_namespace_name_rules_2, sizeof(__pyx_k_enforce_namespace_name_rules_2), 0, 1, 0, 0},
    {&__pyx_n_s_enforce_type_name_rules, __pyx_k_enforce_type_name_rules, sizeof(__pyx_k_enforce_type_name_rules), 0, 0, 1, 1},
    {&__pyx_kp_u_enforce_type_name_rules_2, __pyx_k_enforce_type_name_rules_2, sizeof(__pyx_k_enforce_type_name_rules_2), 0, 1, 0, 0},
    {&__pyx_n_s_ensure_ascii, __pyx_k_ensure_ascii, sizeof(__pyx_k_ensure_ascii), 0, 0, 1, 1},
    {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
    {&__pyx_n_s_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 0, 1, 1},
    {&__pyx_n_u_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 1, 0, 1},
    {&__pyx_n_s_enum_symbols_must_be_unique, __pyx_k_enum_symbols_must_be_unique, sizeof(__pyx_k_enum_symbols_must_be_unique), 0, 0, 1, 1},
    {&__pyx_kp_u_enum_symbols_must_be_unique_2, __pyx_k_enum_symbols_must_be_unique_2, sizeof(__pyx_k_enum_symbols_must_be_unique_2), 0, 1, 0, 0},
    {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
    {&__pyx_n_s_eq, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
    {&__pyx_n_s_equals, __pyx_k_equals, sizeof(__pyx_k_equals), 0, 0, 1, 1},
    {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
    {&__pyx_n_u_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 1, 0, 1},
    {&__pyx_n_s_errors, __pyx_k_errors, sizeof(__pyx_k_errors), 0, 0, 1, 1},
    {&__pyx_n_s_escape, __pyx_k_escape, sizeof(__pyx_k_escape), 0, 0, 1, 1},
    {&__pyx_n_s_eval, __pyx_k_eval, sizeof(__pyx_k_eval), 0, 0, 1, 1},
    {&__pyx_n_s_eval_2, __pyx_k_eval_2, sizeof(__pyx_k_eval_2), 0, 0, 1, 1},
    {&__pyx_n_s_exc_tb, __pyx_k_exc_tb, sizeof(__pyx_k_exc_tb), 0, 0, 1, 1},
    {&__pyx_n_s_exc_type, __pyx_k_exc_type, sizeof(__pyx_k_exc_type), 0, 0, 1, 1},
    {&__pyx_n_s_exc_val, __pyx_k_exc_val, sizeof(__pyx_k_exc_val), 0, 0, 1, 1},
    {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
    {&__pyx_n_s_expand_types_in_schema, __pyx_k_expand_types_in_schema, sizeof(__pyx_k_expand_types_in_schema), 0, 0, 1, 1},
    {&__pyx_kp_u_expand_types_in_schema_2, __pyx_k_expand_types_in_schema_2, sizeof(__pyx_k_expand_types_in_schema_2), 0, 1, 0, 0},
    {&__pyx_n_s_externally_defined_types, __pyx_k_externally_defined_types, sizeof(__pyx_k_externally_defined_types), 0, 0, 1, 1},
    {&__pyx_kp_u_externally_defined_types_2, __pyx_k_externally_defined_types_2, sizeof(__pyx_k_externally_defined_types_2), 0, 1, 0, 0},
    {&__pyx_n_s_extra, __pyx_k_extra, sizeof(__pyx_k_extra), 0, 0, 1, 1},
    {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
    {&__pyx_n_s_field, __pyx_k_field, sizeof(__pyx_k_field), 0, 0, 1, 1},
    {&__pyx_n_s_field_to_index, __pyx_k_field_to_index, sizeof(__pyx_k_field_to_index), 0, 0, 1, 1},
    {&__pyx_n_u_field_to_index, __pyx_k_field_to_index, sizeof(__pyx_k_field_to_index), 0, 1, 0, 1},
    {&__pyx_n_s_field_type, __pyx_k_field_type, sizeof(__pyx_k_field_type), 0, 0, 1, 1},
    {&__pyx_n_s_fields, __pyx_k_fields, sizeof(__pyx_k_fields), 0, 0, 1, 1},
    {&__pyx_n_u_fields, __pyx_k_fields, sizeof(__pyx_k_fields), 0, 1, 0, 1},
    {&__pyx_n_s_file_obj, __pyx_k_file_obj, sizeof(__pyx_k_file_obj), 0, 0, 1, 1},
    {&__pyx_n_s_find_type, __pyx_k_find_type, sizeof(__pyx_k_find_type), 0, 0, 1, 1},
    {&__pyx_n_s_fingerprint, __pyx_k_fingerprint, sizeof(__pyx_k_fingerprint), 0, 0, 1, 1},
    {&__pyx_n_s_fingerprint_returns_digest, __pyx_k_fingerprint_returns_digest, sizeof(__pyx_k_fingerprint_returns_digest), 0, 0, 1, 1},
    {&__pyx_kp_u_fingerprint_returns_digest_2, __pyx_k_fingerprint_returns_digest_2, sizeof(__pyx_k_fingerprint_returns_digest_2), 0, 1, 0, 0},
    {&__pyx_n_s_fitness, __pyx_k_fitness, sizeof(__pyx_k_fitness), 0, 0, 1, 1},
    {&__pyx_n_s_fixed, __pyx_k_fixed, sizeof(__pyx_k_fixed), 0, 0, 1, 1},
    {&__pyx_n_u_fixed, __pyx_k_fixed, sizeof(__pyx_k_fixed), 0, 1, 0, 1},
    {&__pyx_n_s_fixed_codec, __pyx_k_fixed_codec, sizeof(__pyx_k_fixed_codec), 0, 0, 1, 1},
    {&__pyx_kp_u_fixed_codec_2, __pyx_k_fixed_codec_2, sizeof(__pyx_k_fixed_codec_2), 0, 1, 0, 0},
    {&__pyx_kp_u_fixed_uuid, __pyx_k_fixed_uuid, sizeof(__pyx_k_fixed_uuid), 0, 1, 0, 0},
    {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
    {&__pyx_kp_u_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 1, 0, 0},
    {&__pyx_n_s_float16, __pyx_k_float16, sizeof(__pyx_k_float16), 0, 0, 1, 1},
    {&__pyx_n_s_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 0, 1, 1},
    {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
    {&__pyx_n_u_float_2, __pyx_k_float_2, sizeof(__pyx_k_float_2), 0, 1, 0, 1},
    {&__pyx_n_s_float_out_of_range_inf, __pyx_k_float_out_of_range_inf, sizeof(__pyx_k_float_out_of_range_inf), 0, 0, 1, 1},
    {&__pyx_kp_u_float_out_of_range_inf_2, __pyx_k_float_out_of_range_inf_2, sizeof(__pyx_k_float_out_of_range_inf_2), 0, 1, 0, 0},
    {&__pyx_n_s_floor, __pyx_k_floor, sizeof(__pyx_k_floor), 0, 0, 1, 1},
    {&__pyx_n_s_flush, __pyx_k_flush, sizeof(__pyx_k_flush), 0, 0, 1, 1},
    {&__pyx_n_s_fn, __pyx_k_fn, sizeof(__pyx_k_fn), 0, 0, 1, 1},
    {&__pyx_kp_u_for_field, __pyx_k_for_field, sizeof(__pyx_k_for_field), 0, 1, 0, 0},
    {&__pyx_n_s_for_schema, __pyx_k_for_schema, sizeof(__pyx_k_for_schema), 0, 0, 1, 1},
    {&__pyx_n_s_for_source, __pyx_k_for_source, sizeof(__pyx_k_for_source), 0, 0, 1, 1},
    {&__pyx_kp_u_for_type, __pyx_k_for_type, sizeof(__pyx_k_for_type), 0, 1, 0, 0},
    {&__pyx_n_s_for_type_2, __pyx_k_for_type_2, sizeof(__pyx_k_for_type_2), 0, 0, 1, 1},
    {&__pyx_n_s_for_type_3, __pyx_k_for_type_3, sizeof(__pyx_k_for_type_3), 0, 0, 1, 1},
    {&__pyx_n_s_force, __pyx_k_force, sizeof(__pyx_k_force), 0, 0, 1, 1},
    {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
    {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
    {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
    {&__pyx_kp_u_found_twice, __pyx_k_found_twice, sizeof(__pyx_k_found_twice), 0, 1, 0, 0},
    {&__pyx_n_s_frame, __pyx_k_frame, sizeof(__pyx_k_frame), 0, 0, 1, 1},
    {&__pyx_n_s_from_bytes, __pyx_k_from_bytes, sizeof(__pyx_k_from_bytes), 0, 0, 1, 1},
    {&__pyx_n_s_frozen, __pyx_k_frozen, sizeof(__pyx_k_frozen), 0, 0, 1, 1},
    {&__pyx_n_s_fullmatch, __pyx_k_fullmatch, sizeof(__pyx_k_fullmatch), 0, 0, 1, 1},
    {&__pyx_n_s_func, __pyx_k_func, sizeof(__pyx_k_func), 0, 0, 1, 1},
    {&__pyx_n_s_functools, __pyx_k_functools, sizeof(__pyx_k_functools), 0, 0, 1, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
    {&__pyx_n_s_get_2, __pyx_k_get_2, sizeof(__pyx_k_get_2), 0, 0, 1, 1},
    {&__pyx_n_s_get_schema, __pyx_k_get_schema, sizeof(__pyx_k_get_schema), 0, 0, 1, 1},
    {&__pyx_n_s_get_schema_extra, __pyx_k_get_schema_extra, sizeof(__pyx_k_get_schema_extra), 0, 0, 1, 1},
    {&__pyx_kp_u_getfullargspec, __pyx_k_getfullargspec, sizeof(__pyx_k_getfullargspec), 0, 1, 0, 0},
    {&__pyx_n_u_getitem, __pyx_k_getitem, sizeof(__pyx_k_getitem), 0, 1, 0, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_s_given, __pyx_k_given, sizeof(__pyx_k_given), 0, 0, 1, 1},
    {&__pyx_kp_u_got, __pyx_k_got, sizeof(__pyx_k_got), 0, 1, 0, 0},
    {&__pyx_kp_u_got_2, __pyx_k_got_2, sizeof(__pyx_k_got_2), 0, 1, 0, 0},
    {&__pyx_kp_u_got_3, __pyx_k_got_3, sizeof(__pyx_k_got_3), 0, 1, 0, 0},
    {&__pyx_kp_u_got_4, __pyx_k_got_4, sizeof(__pyx_k_got_4), 0, 1, 0, 0},
    {&__pyx_kp_u_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 1, 0, 0},
    {&__pyx_kp_u_has_exponent, __pyx_k_has_exponent, sizeof(__pyx_k_has_exponent), 0, 1, 0, 0},
    {&__pyx_n_s_hash, __pyx_k_hash, sizeof(__pyx_k_hash), 0, 0, 1, 1},
    {&__pyx_n_s_hasher, __pyx_k_hasher, sizeof(__pyx_k_hasher), 0, 0, 1, 1},
    {&__pyx_n_s_hashlib, __pyx_k_hashlib, sizeof(__pyx_k_hashlib), 0, 0, 1, 1},
    {&__pyx_n_s_hex, __pyx_k_hex, sizeof(__pyx_k_hex), 0, 0, 1, 1},
    {&__pyx_n_s_hexdigest, __pyx_k_hexdigest, sizeof(__pyx_k_hexdigest), 0, 0, 1, 1},
    {&__pyx_n_s_hour, __pyx_k_hour, sizeof(__pyx_k_hour), 0, 0, 1, 1},
    {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
    {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
    {&__pyx_n_s_ignore, __pyx_k_ignore, sizeof(__pyx_k_ignore), 0, 0, 1, 1},
    {&__pyx_n_u_ignore, __pyx_k_ignore, sizeof(__pyx_k_ignore), 0, 1, 0, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_indent, __pyx_k_indent, sizeof(__pyx_k_indent), 0, 0, 1, 1},
    {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
    {&__pyx_kp_u_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 1, 0, 0},
    {&__pyx_n_u_inf_2, __pyx_k_inf_2, sizeof(__pyx_k_inf_2), 0, 1, 0, 1},
    {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
    {&__pyx_n_s_init_2, __pyx_k_init_2, sizeof(__pyx_k_init_2), 0, 0, 1, 1},
    {&__pyx_n_s_init___locals_genexpr, __pyx_k_init___locals_genexpr, sizeof(__pyx_k_init___locals_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_init_subclass, __pyx_k_init_subclass, sizeof(__pyx_k_init_subclass), 0, 0, 1, 1},
    {&__pyx_n_s_initial_size, __pyx_k_initial_size, sizeof(__pyx_k_initial_size), 0, 0, 1, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_n_s_inline_namespaces, __pyx_k_inline_namespaces, sizeof(__pyx_k_inline_namespaces), 0, 0, 1, 1},
    {&__pyx_kp_u_inline_namespaces_2, __pyx_k_inline_namespaces_2, sizeof(__pyx_k_inline_namespaces_2), 0, 1, 0, 0},
    {&__pyx_n_s_inspect, __pyx_k_inspect, sizeof(__pyx_k_inspect), 0, 0, 1, 1},
    {&__pyx_n_s_inst, __pyx_k_inst, sizeof(__pyx_k_inst), 0, 0, 1, 1},
    {&__pyx_kp_u_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 1, 0, 0},
    {&__pyx_n_u_int_2, __pyx_k_int_2, sizeof(__pyx_k_int_2), 0, 1, 0, 1},
    {&__pyx_kp_u_int_date, __pyx_k_int_date, sizeof(__pyx_k_int_date), 0, 1, 0, 0},
    {&__pyx_kp_u_int_time_millis, __pyx_k_int_time_millis, sizeof(__pyx_k_int_time_millis), 0, 1, 0, 0},
    {&__pyx_n_s_integer, __pyx_k_integer, sizeof(__pyx_k_integer), 0, 0, 1, 1},
    {&__pyx_kp_u_invalid_for_enum, __pyx_k_invalid_for_enum, sizeof(__pyx_k_invalid_for_enum), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid_value_include_array_ind, __pyx_k_invalid_value_include_array_ind, sizeof(__pyx_k_invalid_value_include_array_ind), 0, 1, 0, 0},
    {&__pyx_n_s_invalid_value_include_array_inde, __pyx_k_invalid_value_include_array_inde, sizeof(__pyx_k_invalid_value_include_array_inde), 0, 0, 1, 1},
    {&__pyx_n_s_invalid_value_include_map_key, __pyx_k_invalid_value_include_map_key, sizeof(__pyx_k_invalid_value_include_map_key), 0, 0, 1, 1},
    {&__pyx_kp_u_invalid_value_include_map_key_2, __pyx_k_invalid_value_include_map_key_2, sizeof(__pyx_k_invalid_value_include_map_key_2), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid_value_includes_record_n, __pyx_k_invalid_value_includes_record_n, sizeof(__pyx_k_invalid_value_includes_record_n), 0, 1, 0, 0},
    {&__pyx_n_s_invalid_value_includes_record_na, __pyx_k_invalid_value_includes_record_na, sizeof(__pyx_k_invalid_value_includes_record_na), 0, 0, 1, 1},
    {&__pyx_n_s_io, __pyx_k_io, sizeof(__pyx_k_io), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_kp_u_is_not_a_mapping, __pyx_k_is_not_a_mapping, sizeof(__pyx_k_is_not_a_mapping), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_a_valid_key_for_map_type, __pyx_k_is_not_a_valid_key_for_map_type, sizeof(__pyx_k_is_not_a_valid_key_for_map_type), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_a_valid_name, __pyx_k_is_not_a_valid_name, sizeof(__pyx_k_is_not_a_valid_name), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_a_valid_union_value_expe, __pyx_k_is_not_a_valid_union_value_expe, sizeof(__pyx_k_is_not_a_valid_union_value_expe), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_a_valid_union_value_unkn, __pyx_k_is_not_a_valid_union_value_unkn, sizeof(__pyx_k_is_not_a_valid_union_value_unkn), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_a_valid_value_for_array, __pyx_k_is_not_a_valid_value_for_array, sizeof(__pyx_k_is_not_a_valid_value_for_array), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_a_valid_value_for_map_ty, __pyx_k_is_not_a_valid_value_for_map_ty, sizeof(__pyx_k_is_not_a_valid_value_for_map_ty), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_allowed_as_a_name, __pyx_k_is_not_allowed_as_a_name, sizeof(__pyx_k_is_not_allowed_as_a_name), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_valid, __pyx_k_is_not_valid, sizeof(__pyx_k_is_not_valid), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_valid_for_a_union_of, __pyx_k_is_not_valid_for_a_union_of, sizeof(__pyx_k_is_not_valid_for_a_union_of), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_valid_for_bytes_field, __pyx_k_is_not_valid_for_bytes_field, sizeof(__pyx_k_is_not_valid_for_bytes_field), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_valid_for_field, __pyx_k_is_not_valid_for_field, sizeof(__pyx_k_is_not_valid_for_field), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_valid_for_fixed_field, __pyx_k_is_not_valid_for_fixed_field, sizeof(__pyx_k_is_not_valid_for_fixed_field), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_valid_for_union_with_fir, __pyx_k_is_not_valid_for_union_with_fir, sizeof(__pyx_k_is_not_valid_for_union_with_fir), 0, 1, 0, 0},
    {&__pyx_kp_u_is_required_for_record, __pyx_k_is_required_for_record, sizeof(__pyx_k_is_required_for_record), 0, 1, 0, 0},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_isinf, __pyx_k_isinf, sizeof(__pyx_k_isinf), 0, 0, 1, 1},
    {&__pyx_n_s_isnan, __pyx_k_isnan, sizeof(__pyx_k_isnan), 0, 0, 1, 1},
    {&__pyx_n_s_item_type, __pyx_k_item_type, sizeof(__pyx_k_item_type), 0, 0, 1, 1},
    {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
    {&__pyx_n_u_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 1, 0, 1},
    {&__pyx_kp_u_items_2, __pyx_k_items_2, sizeof(__pyx_k_items_2), 0, 1, 0, 0},
    {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
    {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
    {&__pyx_n_s_json, __pyx_k_json, sizeof(__pyx_k_json), 0, 0, 1, 1},
    {&__pyx_n_s_json_decode, __pyx_k_json_decode, sizeof(__pyx_k_json_decode), 0, 0, 1, 1},
    {&__pyx_n_s_json_encode, __pyx_k_json_encode, sizeof(__pyx_k_json_encode), 0, 0, 1, 1},
    {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
    {&__pyx_n_s_kw_only, __pyx_k_kw_only, sizeof(__pyx_k_kw_only), 0, 0, 1, 1},
    {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
    {&__pyx_n_b_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 0, 1},
    {&__pyx_n_u_latin1, __pyx_k_latin1, sizeof(__pyx_k_latin1), 0, 1, 0, 1},
    {&__pyx_kp_u_latin_1, __pyx_k_latin_1, sizeof(__pyx_k_latin_1), 0, 1, 0, 0},
    {&__pyx_n_u_little, __pyx_k_little, sizeof(__pyx_k_little), 0, 1, 0, 1},
    {&__pyx_n_s_loads, __pyx_k_loads, sizeof(__pyx_k_loads), 0, 0, 1, 1},
    {&__pyx_n_s_local, __pyx_k_local, sizeof(__pyx_k_local), 0, 0, 1, 1},
    {&__pyx_n_s_log10, __pyx_k_log10, sizeof(__pyx_k_log10), 0, 0, 1, 1},
    {&__pyx_n_u_logicalType, __pyx_k_logicalType, sizeof(__pyx_k_logicalType), 0, 1, 0, 1},
    {&__pyx_n_s_logical_by_name, __pyx_k_logical_by_name, sizeof(__pyx_k_logical_by_name), 0, 0, 1, 1},
    {&__pyx_n_s_logical_name, __pyx_k_logical_name, sizeof(__pyx_k_logical_name), 0, 0, 1, 1},
    {&__pyx_n_s_logical_type, __pyx_k_logical_type, sizeof(__pyx_k_logical_type), 0, 0, 1, 1},
    {&__pyx_n_s_logical_types, __pyx_k_logical_types, sizeof(__pyx_k_logical_types), 0, 0, 1, 1},
    {&__pyx_kp_u_logical_types_2, __pyx_k_logical_types_2, sizeof(__pyx_k_logical_types_2), 0, 1, 0, 0},
    {&__pyx_kp_u_long, __pyx_k_long, sizeof(__pyx_k_long), 0, 1, 0, 0},
    {&__pyx_n_u_long_2, __pyx_k_long_2, sizeof(__pyx_k_long_2), 0, 1, 0, 1},
    {&__pyx_kp_u_long_time_micros, __pyx_k_long_time_micros, sizeof(__pyx_k_long_time_micros), 0, 1, 0, 0},
    {&__pyx_kp_u_long_timestamp_micros, __pyx_k_long_timestamp_micros, sizeof(__pyx_k_long_timestamp_micros), 0, 1, 0, 0},
    {&__pyx_kp_u_long_timestamp_millis, __pyx_k_long_timestamp_millis, sizeof(__pyx_k_long_timestamp_millis), 0, 1, 0, 0},
    {&__pyx_n_b_lz4, __pyx_k_lz4, sizeof(__pyx_k_lz4), 0, 0, 0, 1},
    {&__pyx_n_s_lz4, __pyx_k_lz4, sizeof(__pyx_k_lz4), 0, 0, 1, 1},
    {&__pyx_n_u_lz4, __pyx_k_lz4, sizeof(__pyx_k_lz4), 0, 1, 0, 1},
    {&__pyx_n_s_lz4_frame, __pyx_k_lz4_frame, sizeof(__pyx_k_lz4_frame), 0, 0, 1, 1},
    {&__pyx_n_s_lzma, __pyx_k_lzma, sizeof(__pyx_k_lzma), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_make_dataclass, __pyx_k_make_dataclass, sizeof(__pyx_k_make_dataclass), 0, 0, 1, 1},
    {&__pyx_n_s_make_logical_types, __pyx_k_make_logical_types, sizeof(__pyx_k_make_logical_types), 0, 0, 1, 1},
    {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
    {&__pyx_n_u_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 1, 0, 1},
    {&__pyx_n_s_marker, __pyx_k_marker, sizeof(__pyx_k_marker), 0, 0, 1, 1},
    {&__pyx_n_s_mask, __pyx_k_mask, sizeof(__pyx_k_mask), 0, 0, 1, 1},
    {&__pyx_n_s_match_args, __pyx_k_match_args, sizeof(__pyx_k_match_args), 0, 0, 1, 1},
    {&__pyx_n_s_math, __pyx_k_math, sizeof(__pyx_k_math), 0, 0, 1, 1},
    {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
    {&__pyx_n_s_max_blocksize, __pyx_k_max_blocksize, sizeof(__pyx_k_max_blocksize), 0, 0, 1, 1},
    {&__pyx_n_s_max_precision, __pyx_k_max_precision, sizeof(__pyx_k_max_precision), 0, 0, 1, 1},
    {&__pyx_n_s_maxunicode, __pyx_k_maxunicode, sizeof(__pyx_k_maxunicode), 0, 0, 1, 1},
    {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
    {&__pyx_n_s_meta, __pyx_k_meta, sizeof(__pyx_k_meta), 0, 0, 1, 1},
    {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
    {&__pyx_n_s_metadata, __pyx_k_metadata, sizeof(__pyx_k_metadata), 0, 0, 1, 1},
    {&__pyx_n_s_method, __pyx_k_method, sizeof(__pyx_k_method), 0, 0, 1, 1},
    {&__pyx_n_s_micros_to_time, __pyx_k_micros_to_time, sizeof(__pyx_k_micros_to_time), 0, 0, 1, 1},
    {&__pyx_n_s_microsecond, __pyx_k_microsecond, sizeof(__pyx_k_microsecond), 0, 0, 1, 1},
    {&__pyx_n_s_microseconds, __pyx_k_microseconds, sizeof(__pyx_k_microseconds), 0, 0, 1, 1},
    {&__pyx_n_s_minute, __pyx_k_minute, sizeof(__pyx_k_minute), 0, 0, 1, 1},
    {&__pyx_kp_u_missing, __pyx_k_missing, sizeof(__pyx_k_missing), 0, 1, 0, 0},
    {&__pyx_kp_u_missing_2, __pyx_k_missing_2, sizeof(__pyx_k_missing_2), 0, 1, 0, 0},
    {&__pyx_kp_u_missing_values_can_be_empty_con, __pyx_k_missing_values_can_be_empty_con, sizeof(__pyx_k_missing_values_can_be_empty_con), 0, 1, 0, 0},
    {&__pyx_n_s_missing_values_can_be_empty_cont, __pyx_k_missing_values_can_be_empty_cont, sizeof(__pyx_k_missing_values_can_be_empty_cont), 0, 0, 1, 1},
    {&__pyx_n_s_missing_values_can_be_null, __pyx_k_missing_values_can_be_null, sizeof(__pyx_k_missing_values_can_be_null), 0, 0, 1, 1},
    {&__pyx_kp_u_missing_values_can_be_null_2, __pyx_k_missing_values_can_be_null_2, sizeof(__pyx_k_missing_values_can_be_null_2), 0, 1, 0, 0},
    {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
    {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
    {&__pyx_n_s_mro_entries, __pyx_k_mro_entries, sizeof(__pyx_k_mro_entries), 0, 0, 1, 1},
    {&__pyx_n_s_msg, __pyx_k_msg, sizeof(__pyx_k_msg), 0, 0, 1, 1},
    {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
    {&__pyx_n_s_n_bytes, __pyx_k_n_bytes, sizeof(__pyx_k_n_bytes), 0, 0, 1, 1},
    {&__pyx_n_s_naive_dt_assume_utc, __pyx_k_naive_dt_assume_utc, sizeof(__pyx_k_naive_dt_assume_utc), 0, 0, 1, 1},
    {&__pyx_kp_u_naive_dt_assume_utc_2, __pyx_k_naive_dt_assume_utc_2, sizeof(__pyx_k_naive_dt_assume_utc_2), 0, 1, 0, 0},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_u_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 1, 0, 1},
    {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
    {&__pyx_n_s_name_pattern, __pyx_k_name_pattern, sizeof(__pyx_k_name_pattern), 0, 0, 1, 1},
    {&__pyx_n_s_named_type_names_must_be_unique, __pyx_k_named_type_names_must_be_unique, sizeof(__pyx_k_named_type_names_must_be_unique), 0, 0, 1, 1},
    {&__pyx_kp_u_named_type_names_must_be_unique_2, __pyx_k_named_type_names_must_be_unique_2, sizeof(__pyx_k_named_type_names_must_be_unique_2), 0, 1, 0, 0},
    {&__pyx_n_s_named_types, __pyx_k_named_types, sizeof(__pyx_k_named_types), 0, 0, 1, 1},
    {&__pyx_n_s_namespace, __pyx_k_namespace, sizeof(__pyx_k_namespace), 0, 0, 1, 1},
    {&__pyx_n_u_namespace, __pyx_k_namespace, sizeof(__pyx_k_namespace), 0, 1, 0, 1},
    {&__pyx_n_s_namespaced, __pyx_k_namespaced, sizeof(__pyx_k_namespaced), 0, 0, 1, 1},
    {&__pyx_n_u_nan, __pyx_k_nan, sizeof(__pyx_k_nan), 0, 1, 0, 1},
    {&__pyx_n_s_ndarray, __pyx_k_ndarray, sizeof(__pyx_k_ndarray), 0, 0, 1, 1},
    {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
    {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
    {&__pyx_n_s_new_2, __pyx_k_new_2, sizeof(__pyx_k_new_2), 0, 0, 1, 1},
    {&__pyx_n_s_new_inst, __pyx_k_new_inst, sizeof(__pyx_k_new_inst), 0, 0, 1, 1},
    {&__pyx_n_s_new_types, __pyx_k_new_types, sizeof(__pyx_k_new_types), 0, 0, 1, 1},
    {&__pyx_n_s_new_ver, __pyx_k_new_ver, sizeof(__pyx_k_new_ver), 0, 0, 1, 1},
    {&__pyx_n_s_next_object, __pyx_k_next_object, sizeof(__pyx_k_next_object), 0, 0, 1, 1},
    {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
    {&__pyx_kp_u_not_compatible_with_NullType, __pyx_k_not_compatible_with_NullType, sizeof(__pyx_k_not_compatible_with_NullType), 0, 1, 0, 0},
    {&__pyx_kp_u_not_in_enum_symbols, __pyx_k_not_in_enum_symbols, sizeof(__pyx_k_not_in_enum_symbols), 0, 1, 0, 0},
    {&__pyx_kp_u_not_valid_for, __pyx_k_not_valid_for, sizeof(__pyx_k_not_valid_for), 0, 1, 0, 0},
    {&__pyx_kp_u_not_valid_for_long, __pyx_k_not_valid_for_long, sizeof(__pyx_k_not_valid_for_long), 0, 1, 0, 0},
    {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
    {&__pyx_n_s_np_f16, __pyx_k_np_f16, sizeof(__pyx_k_np_f16), 0, 0, 1, 1},
    {&__pyx_n_s_np_f32, __pyx_k_np_f32, sizeof(__pyx_k_np_f32), 0, 0, 1, 1},
    {&__pyx_n_s_np_f64, __pyx_k_np_f64, sizeof(__pyx_k_np_f64), 0, 0, 1, 1},
    {&__pyx_kp_u_ns, __pyx_k_ns, sizeof(__pyx_k_ns), 0, 1, 0, 0},
    {&__pyx_kp_u_ns_copy, __pyx_k_ns_copy, sizeof(__pyx_k_ns_copy), 0, 1, 0, 0},
    {&__pyx_kp_u_null, __pyx_k_null, sizeof(__pyx_k_null), 0, 1, 0, 0},
    {&__pyx_n_b_null_2, __pyx_k_null_2, sizeof(__pyx_k_null_2), 0, 0, 0, 1},
    {&__pyx_n_u_null_2, __pyx_k_null_2, sizeof(__pyx_k_null_2), 0, 1, 0, 1},
    {&__pyx_n_s_null_pad_fixed, __pyx_k_null_pad_fixed, sizeof(__pyx_k_null_pad_fixed), 0, 0, 1, 1},
    {&__pyx_kp_u_null_pad_fixed_2, __pyx_k_null_pad_fixed_2, sizeof(__pyx_k_null_pad_fixed_2), 0, 1, 0, 0},
    {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
    {&__pyx_n_s_ob, __pyx_k_ob, sizeof(__pyx_k_ob), 0, 0, 1, 1},
    {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
    {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
    {&__pyx_n_s_objs, __pyx_k_objs, sizeof(__pyx_k_objs), 0, 0, 1, 1},
    {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
    {&__pyx_n_s_options, __pyx_k_options, sizeof(__pyx_k_options), 0, 0, 1, 1},
    {&__pyx_n_s_order, __pyx_k_order, sizeof(__pyx_k_order), 0, 0, 1, 1},
    {&__pyx_n_u_order, __pyx_k_order, sizeof(__pyx_k_order), 0, 1, 0, 1},
    {&__pyx_n_s_orig_value, __pyx_k_orig_value, sizeof(__pyx_k_orig_value), 0, 0, 1, 1},
    {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
    {&__pyx_n_s_other, __pyx_k_other, sizeof(__pyx_k_other), 0, 0, 1, 1},
    {&__pyx_n_s_other_val, __pyx_k_other_val, sizeof(__pyx_k_other_val), 0, 0, 1, 1},
    {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
    {&__pyx_n_s_out_state, __pyx_k_out_state, sizeof(__pyx_k_out_state), 0, 0, 1, 1},
    {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
    {&__pyx_n_s_pad, __pyx_k_pad, sizeof(__pyx_k_pad), 0, 0, 1, 1},
    {&__pyx_n_s_parse_json, __pyx_k_parse_json, sizeof(__pyx_k_parse_json), 0, 0, 1, 1},
    {&__pyx_n_s_parse_namespace, __pyx_k_parse_namespace, sizeof(__pyx_k_parse_namespace), 0, 0, 1, 1},
    {&__pyx_n_s_partial, __pyx_k_partial, sizeof(__pyx_k_partial), 0, 0, 1, 1},
    {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
    {&__pyx_n_s_path_str, __pyx_k_path_str, sizeof(__pyx_k_path_str), 0, 0, 1, 1},
    {&__pyx_n_s_pathlib, __pyx_k_pathlib, sizeof(__pyx_k_pathlib), 0, 0, 1, 1},
    {&__pyx_n_s_perf, __pyx_k_perf, sizeof(__pyx_k_perf), 0, 0, 1, 1},
    {&__pyx_n_s_perf_counter_ns, __pyx_k_perf_counter_ns, sizeof(__pyx_k_perf_counter_ns), 0, 0, 1, 1},
    {&__pyx_n_s_perf_locals__test_zigzag_perf, __pyx_k_perf_locals__test_zigzag_perf, sizeof(__pyx_k_perf_locals__test_zigzag_perf), 0, 0, 1, 1},
    {&__pyx_n_s_perf_locals_add, __pyx_k_perf_locals_add, sizeof(__pyx_k_perf_locals_add), 0, 0, 1, 1},
    {&__pyx_n_s_perf_locals_add_locals_genexpr, __pyx_k_perf_locals_add_locals_genexpr, sizeof(__pyx_k_perf_locals_add_locals_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_perf_tests, __pyx_k_perf_tests, sizeof(__pyx_k_perf_tests), 0, 0, 1, 1},
    {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
    {&__pyx_n_u_placeholder, __pyx_k_placeholder, sizeof(__pyx_k_placeholder), 0, 1, 0, 1},
    {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
    {&__pyx_n_s_prec, __pyx_k_prec, sizeof(__pyx_k_prec), 0, 0, 1, 1},
    {&__pyx_n_s_precision, __pyx_k_precision, sizeof(__pyx_k_precision), 0, 0, 1, 1},
    {&__pyx_n_u_precision, __pyx_k_precision, sizeof(__pyx_k_precision), 0, 1, 0, 1},
    {&__pyx_kp_u_precision_2, __pyx_k_precision_2, sizeof(__pyx_k_precision_2), 0, 1, 0, 0},
    {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
    {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
    {&__pyx_n_s_pytest, __pyx_k_pytest, sizeof(__pyx_k_pytest), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_recursive_repr_guard, __pyx_k_pyx_recursive_repr_guard, sizeof(__pyx_k_pyx_recursive_repr_guard), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_ArrayType, __pyx_k_pyx_unpickle_ArrayType, sizeof(__pyx_k_pyx_unpickle_ArrayType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_AvroType, __pyx_k_pyx_unpickle_AvroType, sizeof(__pyx_k_pyx_unpickle_AvroType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_BoolType, __pyx_k_pyx_unpickle_BoolType, sizeof(__pyx_k_pyx_unpickle_BoolType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_BytesType, __pyx_k_pyx_unpickle_BytesType, sizeof(__pyx_k_pyx_unpickle_BytesType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_CannotPromote, __pyx_k_pyx_unpickle_CannotPromote, sizeof(__pyx_k_pyx_unpickle_CannotPromote), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Codec, __pyx_k_pyx_unpickle_Codec, sizeof(__pyx_k_pyx_unpickle_Codec), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_ContainerReader, __pyx_k_pyx_unpickle_ContainerReader, sizeof(__pyx_k_pyx_unpickle_ContainerReader), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_CustomLogicalType, __pyx_k_pyx_unpickle_CustomLogicalType, sizeof(__pyx_k_pyx_unpickle_CustomLogicalType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Date, __pyx_k_pyx_unpickle_Date, sizeof(__pyx_k_pyx_unpickle_Date), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_DecimalType, __pyx_k_pyx_unpickle_DecimalType, sizeof(__pyx_k_pyx_unpickle_DecimalType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_DoubleType, __pyx_k_pyx_unpickle_DoubleType, sizeof(__pyx_k_pyx_unpickle_DoubleType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_EnumType, __pyx_k_pyx_unpickle_EnumType, sizeof(__pyx_k_pyx_unpickle_EnumType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_FileReader, __pyx_k_pyx_unpickle_FileReader, sizeof(__pyx_k_pyx_unpickle_FileReader), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_FileWriter, __pyx_k_pyx_unpickle_FileWriter, sizeof(__pyx_k_pyx_unpickle_FileWriter), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_FixedType, __pyx_k_pyx_unpickle_FixedType, sizeof(__pyx_k_pyx_unpickle_FixedType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_FloatType, __pyx_k_pyx_unpickle_FloatType, sizeof(__pyx_k_pyx_unpickle_FloatType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_IntType, __pyx_k_pyx_unpickle_IntType, sizeof(__pyx_k_pyx_unpickle_IntType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_LogicalType, __pyx_k_pyx_unpickle_LogicalType, sizeof(__pyx_k_pyx_unpickle_LogicalType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_LongType, __pyx_k_pyx_unpickle_LongType, sizeof(__pyx_k_pyx_unpickle_LongType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_MapType, __pyx_k_pyx_unpickle_MapType, sizeof(__pyx_k_pyx_unpickle_MapType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_MemoryReader, __pyx_k_pyx_unpickle_MemoryReader, sizeof(__pyx_k_pyx_unpickle_MemoryReader), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_NullType, __pyx_k_pyx_unpickle_NullType, sizeof(__pyx_k_pyx_unpickle_NullType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Options, __pyx_k_pyx_unpickle_Options, sizeof(__pyx_k_pyx_unpickle_Options), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_PromoteBytesToStr, __pyx_k_pyx_unpickle_PromoteBytesToStr, sizeof(__pyx_k_pyx_unpickle_PromoteBytesToStr), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_PromoteStringToBy, __pyx_k_pyx_unpickle_PromoteStringToBy, sizeof(__pyx_k_pyx_unpickle_PromoteStringToBy), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_PromoteToFloat, __pyx_k_pyx_unpickle_PromoteToFloat, sizeof(__pyx_k_pyx_unpickle_PromoteToFloat), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_PromotingEnumType, __pyx_k_pyx_unpickle_PromotingEnumType, sizeof(__pyx_k_pyx_unpickle_PromotingEnumType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_PromotingRecordTy, __pyx_k_pyx_unpickle_PromotingRecordTy, sizeof(__pyx_k_pyx_unpickle_PromotingRecordTy), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Rabin, __pyx_k_pyx_unpickle_Rabin, sizeof(__pyx_k_pyx_unpickle_Rabin), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Record, __pyx_k_pyx_unpickle_Record, sizeof(__pyx_k_pyx_unpickle_Record), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_RecordField, __pyx_k_pyx_unpickle_RecordField, sizeof(__pyx_k_pyx_unpickle_RecordField), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_RecordType, __pyx_k_pyx_unpickle_RecordType, sizeof(__pyx_k_pyx_unpickle_RecordType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_ResolvedSchema, __pyx_k_pyx_unpickle_ResolvedSchema, sizeof(__pyx_k_pyx_unpickle_ResolvedSchema), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Schema, __pyx_k_pyx_unpickle_Schema, sizeof(__pyx_k_pyx_unpickle_Schema), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_StringType, __pyx_k_pyx_unpickle_StringType, sizeof(__pyx_k_pyx_unpickle_StringType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_TimeMicros, __pyx_k_pyx_unpickle_TimeMicros, sizeof(__pyx_k_pyx_unpickle_TimeMicros), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_TimeMillis, __pyx_k_pyx_unpickle_TimeMillis, sizeof(__pyx_k_pyx_unpickle_TimeMillis), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_TimestampMicros, __pyx_k_pyx_unpickle_TimestampMicros, sizeof(__pyx_k_pyx_unpickle_TimestampMicros), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_TimestampMillis, __pyx_k_pyx_unpickle_TimestampMillis, sizeof(__pyx_k_pyx_unpickle_TimestampMillis), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_UUIDBase, __pyx_k_pyx_unpickle_UUIDBase, sizeof(__pyx_k_pyx_unpickle_UUIDBase), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_UUIDFixedType, __pyx_k_pyx_unpickle_UUIDFixedType, sizeof(__pyx_k_pyx_unpickle_UUIDFixedType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_UUIDStringType, __pyx_k_pyx_unpickle_UUIDStringType, sizeof(__pyx_k_pyx_unpickle_UUIDStringType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_UnionType, __pyx_k_pyx_unpickle_UnionType, sizeof(__pyx_k_pyx_unpickle_UnionType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_ValueAdapter, __pyx_k_pyx_unpickle_ValueAdapter, sizeof(__pyx_k_pyx_unpickle_ValueAdapter), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__CanonicalForm, __pyx_k_pyx_unpickle__CanonicalForm, sizeof(__pyx_k_pyx_unpickle__CanonicalForm), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__DeflateCodec, __pyx_k_pyx_unpickle__DeflateCodec, sizeof(__pyx_k_pyx_unpickle__DeflateCodec), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__FieldAccessor, __pyx_k_pyx_unpickle__FieldAccessor, sizeof(__pyx_k_pyx_unpickle__FieldAccessor), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__Lz4Codec, __pyx_k_pyx_unpickle__Lz4Codec, sizeof(__pyx_k_pyx_unpickle__Lz4Codec), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__LzmaCodec, __pyx_k_pyx_unpickle__LzmaCodec, sizeof(__pyx_k_pyx_unpickle__LzmaCodec), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__NamedType, __pyx_k_pyx_unpickle__NamedType, sizeof(__pyx_k_pyx_unpickle__NamedType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__NullCodec, __pyx_k_pyx_unpickle__NullCodec, sizeof(__pyx_k_pyx_unpickle__NullCodec), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__PlaceholderType, __pyx_k_pyx_unpickle__PlaceholderType, sizeof(__pyx_k_pyx_unpickle__PlaceholderType), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__Reader, __pyx_k_pyx_unpickle__Reader, sizeof(__pyx_k_pyx_unpickle__Reader), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__Sentinel, __pyx_k_pyx_unpickle__Sentinel, sizeof(__pyx_k_pyx_unpickle__Sentinel), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__SnappyCodec, __pyx_k_pyx_unpickle__SnappyCodec, sizeof(__pyx_k_pyx_unpickle__SnappyCodec), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__Writer, __pyx_k_pyx_unpickle__Writer, sizeof(__pyx_k_pyx_unpickle__Writer), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle__ZStandardCodec, __pyx_k_pyx_unpickle__ZStandardCodec, sizeof(__pyx_k_pyx_unpickle__ZStandardCodec), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle___Bzip2Codec, __pyx_k_pyx_unpickle___Bzip2Codec, sizeof(__pyx_k_pyx_unpickle___Bzip2Codec), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_b_q, __pyx_k_q, sizeof(__pyx_k_q), 0, 0, 0, 1},
    {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
    {&__pyx_n_s_quantize, __pyx_k_quantize, sizeof(__pyx_k_quantize), 0, 0, 1, 1},
    {&__pyx_n_u_rabin, __pyx_k_rabin, sizeof(__pyx_k_rabin), 0, 1, 0, 1},
    {&__pyx_n_s_raise, __pyx_k_raise, sizeof(__pyx_k_raise), 0, 0, 1, 1},
    {&__pyx_n_s_raise_on_invalid_logical, __pyx_k_raise_on_invalid_logical, sizeof(__pyx_k_raise_on_invalid_logical), 0, 0, 1, 1},
    {&__pyx_kp_u_raise_on_invalid_logical_2, __pyx_k_raise_on_invalid_logical_2, sizeof(__pyx_k_raise_on_invalid_logical_2), 0, 1, 0, 0},
    {&__pyx_n_s_raises, __pyx_k_raises, sizeof(__pyx_k_raises), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_u_rb, __pyx_k_rb, sizeof(__pyx_k_rb), 0, 1, 0, 1},
    {&__pyx_n_s_re, __pyx_k_re, sizeof(__pyx_k_re), 0, 0, 1, 1},
    {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
    {&__pyx_n_u_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 1, 0, 1},
    {&__pyx_n_s_read_marker, __pyx_k_read_marker, sizeof(__pyx_k_read_marker), 0, 0, 1, 1},
    {&__pyx_n_s_read_perf, __pyx_k_read_perf, sizeof(__pyx_k_read_perf), 0, 0, 1, 1},
    {&__pyx_n_s_read_time, __pyx_k_read_time, sizeof(__pyx_k_read_time), 0, 0, 1, 1},
    {&__pyx_n_s_reader, __pyx_k_reader, sizeof(__pyx_k_reader), 0, 0, 1, 1},
    {&__pyx_n_s_reader_for_writer, __pyx_k_reader_for_writer, sizeof(__pyx_k_reader_for_writer), 0, 0, 1, 1},
    {&__pyx_n_s_reader_schema, __pyx_k_reader_schema, sizeof(__pyx_k_reader_schema), 0, 0, 1, 1},
    {&__pyx_n_s_reader_type, __pyx_k_reader_type, sizeof(__pyx_k_reader_type), 0, 0, 1, 1},
    {&__pyx_n_s_record, __pyx_k_record, sizeof(__pyx_k_record), 0, 0, 1, 1},
    {&__pyx_n_u_record, __pyx_k_record, sizeof(__pyx_k_record), 0, 1, 0, 1},
    {&__pyx_n_s_record_allow_extra_fields, __pyx_k_record_allow_extra_fields, sizeof(__pyx_k_record_allow_extra_fields), 0, 0, 1, 1},
    {&__pyx_kp_u_record_allow_extra_fields_2, __pyx_k_record_allow_extra_fields_2, sizeof(__pyx_k_record_allow_extra_fields_2), 0, 1, 0, 0},
    {&__pyx_n_s_record_can_encode_dict, __pyx_k_record_can_encode_dict, sizeof(__pyx_k_record_can_encode_dict), 0, 0, 1, 1},
    {&__pyx_kp_u_record_can_encode_dict_2, __pyx_k_record_can_encode_dict_2, sizeof(__pyx_k_record_can_encode_dict_2), 0, 1, 0, 0},
    {&__pyx_n_s_record_decodes_to_dict, __pyx_k_record_decodes_to_dict, sizeof(__pyx_k_record_decodes_to_dict), 0, 0, 1, 1},
    {&__pyx_kp_u_record_decodes_to_dict_2, __pyx_k_record_decodes_to_dict_2, sizeof(__pyx_k_record_decodes_to_dict_2), 0, 1, 0, 0},
    {&__pyx_n_s_record_encode_use_defaults, __pyx_k_record_encode_use_defaults, sizeof(__pyx_k_record_encode_use_defaults), 0, 0, 1, 1},
    {&__pyx_kp_u_record_encode_use_defaults_2, __pyx_k_record_encode_use_defaults_2, sizeof(__pyx_k_record_encode_use_defaults_2), 0, 1, 0, 0},
    {&__pyx_n_s_record_fields_must_be_unique, __pyx_k_record_fields_must_be_unique, sizeof(__pyx_k_record_fields_must_be_unique), 0, 0, 1, 1},
    {&__pyx_kp_u_record_fields_must_be_unique_2, __pyx_k_record_fields_must_be_unique_2, sizeof(__pyx_k_record_fields_must_be_unique_2), 0, 1, 0, 0},
    {&__pyx_n_s_record_values_type_hint, __pyx_k_record_values_type_hint, sizeof(__pyx_k_record_values_type_hint), 0, 0, 1, 1},
    {&__pyx_kp_u_record_values_type_hint_2, __pyx_k_record_values_type_hint_2, sizeof(__pyx_k_record_values_type_hint_2), 0, 1, 0, 0},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_s_register, __pyx_k_register, sizeof(__pyx_k_register), 0, 0, 1, 1},
    {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
    {&__pyx_n_s_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 0, 1, 1},
    {&__pyx_n_u_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 1, 0, 1},
    {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
    {&__pyx_kp_u_required_field, __pyx_k_required_field, sizeof(__pyx_k_required_field), 0, 1, 0, 0},
    {&__pyx_n_s_resolved_type, __pyx_k_resolved_type, sizeof(__pyx_k_resolved_type), 0, 0, 1, 1},
    {&__pyx_n_s_return, __pyx_k_return, sizeof(__pyx_k_return), 0, 0, 1, 1},
    {&__pyx_n_s_return_uuid_object, __pyx_k_return_uuid_object, sizeof(__pyx_k_return_uuid_object), 0, 0, 1, 1},
    {&__pyx_kp_u_return_uuid_object_2, __pyx_k_return_uuid_object_2, sizeof(__pyx_k_return_uuid_object_2), 0, 1, 0, 0},
    {&__pyx_n_s_reversed, __pyx_k_reversed, sizeof(__pyx_k_reversed), 0, 0, 1, 1},
    {&__pyx_n_s_rjust, __pyx_k_rjust, sizeof(__pyx_k_rjust), 0, 0, 1, 1},
    {&__pyx_n_s_rsplit, __pyx_k_rsplit, sizeof(__pyx_k_rsplit), 0, 0, 1, 1},
    {&__pyx_n_s_running, __pyx_k_running, sizeof(__pyx_k_running), 0, 0, 1, 1},
    {&__pyx_kp_u_s_read, __pyx_k_s_read, sizeof(__pyx_k_s_read), 0, 1, 0, 0},
    {&__pyx_n_s_scale, __pyx_k_scale, sizeof(__pyx_k_scale), 0, 0, 1, 1},
    {&__pyx_n_u_scale, __pyx_k_scale, sizeof(__pyx_k_scale), 0, 1, 0, 1},
    {&__pyx_n_s_scaleb, __pyx_k_scaleb, sizeof(__pyx_k_scaleb), 0, 0, 1, 1},
    {&__pyx_n_s_schema, __pyx_k_schema, sizeof(__pyx_k_schema), 0, 0, 1, 1},
    {&__pyx_n_s_schema_ob, __pyx_k_schema_ob, sizeof(__pyx_k_schema_ob), 0, 0, 1, 1},
    {&__pyx_n_s_schema_path, __pyx_k_schema_path, sizeof(__pyx_k_schema_path), 0, 0, 1, 1},
    {&__pyx_n_s_second, __pyx_k_second, sizeof(__pyx_k_second), 0, 0, 1, 1},
    {&__pyx_n_s_seconds, __pyx_k_seconds, sizeof(__pyx_k_seconds), 0, 0, 1, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_n_s_self_val, __pyx_k_self_val, sizeof(__pyx_k_self_val), 0, 0, 1, 1},
    {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
    {&__pyx_n_s_serialize, __pyx_k_serialize, sizeof(__pyx_k_serialize), 0, 0, 1, 1},
    {&__pyx_n_s_set_from_dataclass, __pyx_k_set_from_dataclass, sizeof(__pyx_k_set_from_dataclass), 0, 0, 1, 1},
    {&__pyx_n_s_set_name, __pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 0, 1, 1},
    {&__pyx_n_s_setdefault, __pyx_k_setdefault, sizeof(__pyx_k_setdefault), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
    {&__pyx_n_s_signature, __pyx_k_signature, sizeof(__pyx_k_signature), 0, 0, 1, 1},
    {&__pyx_n_s_signature_2, __pyx_k_signature_2, sizeof(__pyx_k_signature_2), 0, 0, 1, 1},
    {&__pyx_n_s_signed, __pyx_k_signed, sizeof(__pyx_k_signed), 0, 0, 1, 1},
    {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
    {&__pyx_n_u_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 1, 0, 1},
    {&__pyx_n_u_slots, __pyx_k_slots, sizeof(__pyx_k_slots), 0, 1, 0, 1},
    {&__pyx_n_s_slots_2, __pyx_k_slots_2, sizeof(__pyx_k_slots_2), 0, 0, 1, 1},
    {&__pyx_n_b_snappy, __pyx_k_snappy, sizeof(__pyx_k_snappy), 0, 0, 0, 1},
    {&__pyx_n_s_snappy, __pyx_k_snappy, sizeof(__pyx_k_snappy), 0, 0, 1, 1},
    {&__pyx_n_u_snappy, __pyx_k_snappy, sizeof(__pyx_k_snappy), 0, 1, 0, 1},
    {&__pyx_n_s_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 0, 1, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_n_s_src, __pyx_k_src, sizeof(__pyx_k_src), 0, 0, 1, 1},
    {&__pyx_kp_s_src_array_pxi, __pyx_k_src_array_pxi, sizeof(__pyx_k_src_array_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_container_pxi, __pyx_k_src_container_pxi, sizeof(__pyx_k_src_container_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_enum_pxi, __pyx_k_src_enum_pxi, sizeof(__pyx_k_src_enum_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_error_pxi, __pyx_k_src_error_pxi, sizeof(__pyx_k_src_error_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_logical_pxi, __pyx_k_src_logical_pxi, sizeof(__pyx_k_src_logical_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_map_pxi, __pyx_k_src_map_pxi, sizeof(__pyx_k_src_map_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_null_pxi, __pyx_k_src_null_pxi, sizeof(__pyx_k_src_null_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_numeric_types_pxi, __pyx_k_src_numeric_types_pxi, sizeof(__pyx_k_src_numeric_types_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_option_pxi, __pyx_k_src_option_pxi, sizeof(__pyx_k_src_option_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_rabin_pxi, __pyx_k_src_rabin_pxi, sizeof(__pyx_k_src_rabin_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_record_pxi, __pyx_k_src_record_pxi, sizeof(__pyx_k_src_record_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_schema_pxi, __pyx_k_src_schema_pxi, sizeof(__pyx_k_src_schema_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_string_types_pxi, __pyx_k_src_string_types_pxi, sizeof(__pyx_k_src_string_types_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_tests_test_buffer_pxi, __pyx_k_src_tests_test_buffer_pxi, sizeof(__pyx_k_src_tests_test_buffer_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_tests_test_perf_pxi, __pyx_k_src_tests_test_perf_pxi, sizeof(__pyx_k_src_tests_test_perf_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_tests_test_zigzag_pxi, __pyx_k_src_tests_test_zigzag_pxi, sizeof(__pyx_k_src_tests_test_zigzag_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_type_pxi, __pyx_k_src_type_pxi, sizeof(__pyx_k_src_type_pxi), 0, 0, 1, 0},
    {&__pyx_kp_s_src_union_pxi, __pyx_k_src_union_pxi, sizeof(__pyx_k_src_union_pxi), 0, 0, 1, 0},
    {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
    {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
    {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
    {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
    {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
    {&__pyx_n_s_str_2, __pyx_k_str_2, sizeof(__pyx_k_str_2), 0, 0, 1, 1},
    {&__pyx_n_u_strict, __pyx_k_strict, sizeof(__pyx_k_strict), 0, 1, 0, 1},
    {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
    {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
    {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
    {&__pyx_n_u_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 1, 0, 1},
    {&__pyx_kp_u_string_2, __pyx_k_string_2, sizeof(__pyx_k_string_2), 0, 1, 0, 0},
    {&__pyx_n_s_string_types_default_unchanged, __pyx_k_string_types_default_unchanged, sizeof(__pyx_k_string_types_default_unchanged), 0, 0, 1, 1},
    {&__pyx_kp_u_string_types_default_unchanged_2, __pyx_k_string_types_default_unchanged_2, sizeof(__pyx_k_string_types_default_unchanged_2), 0, 1, 0, 0},
    {&__pyx_kp_u_string_uuid, __pyx_k_string_uuid, sizeof(__pyx_k_string_uuid), 0, 1, 0, 0},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_strptime, __pyx_k_strptime, sizeof(__pyx_k_strptime), 0, 0, 1, 1},
    {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
    {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
    {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
    {&__pyx_n_s_symbols, __pyx_k_symbols, sizeof(__pyx_k_symbols), 0, 0, 1, 1},
    {&__pyx_n_u_symbols, __pyx_k_symbols, sizeof(__pyx_k_symbols), 0, 1, 0, 1},
    {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
    {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
    {&__pyx_n_s_t_a, __pyx_k_t_a, sizeof(__pyx_k_t_a), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_test_2, __pyx_k_test_2, sizeof(__pyx_k_test_2), 0, 0, 1, 1},
    {&__pyx_n_s_test_reader_read_n, __pyx_k_test_reader_read_n, sizeof(__pyx_k_test_reader_read_n), 0, 0, 1, 1},
    {&__pyx_n_s_test_reader_read_u8, __pyx_k_test_reader_read_u8, sizeof(__pyx_k_test_reader_read_u8), 0, 0, 1, 1},
    {&__pyx_n_s_test_readvarint, __pyx_k_test_readvarint, sizeof(__pyx_k_test_readvarint), 0, 0, 1, 1},
    {&__pyx_n_s_test_readvarlong, __pyx_k_test_readvarlong, sizeof(__pyx_k_test_readvarlong), 0, 0, 1, 1},
    {&__pyx_n_s_test_zigzag_long, __pyx_k_test_zigzag_long, sizeof(__pyx_k_test_zigzag_long), 0, 0, 1, 1},
    {&__pyx_n_s_test_zigzag_perf, __pyx_k_test_zigzag_perf, sizeof(__pyx_k_test_zigzag_perf), 0, 0, 1, 1},
    {&__pyx_n_s_tests, __pyx_k_tests, sizeof(__pyx_k_tests), 0, 0, 1, 1},
    {&__pyx_n_s_tests_2, __pyx_k_tests_2, sizeof(__pyx_k_tests_2), 0, 0, 1, 1},
    {&__pyx_n_s_tests_locals__test_reader_read, __pyx_k_tests_locals__test_reader_read, sizeof(__pyx_k_tests_locals__test_reader_read), 0, 0, 1, 1},
    {&__pyx_n_s_tests_locals__test_reader_read_2, __pyx_k_tests_locals__test_reader_read_2, sizeof(__pyx_k_tests_locals__test_reader_read_2), 0, 0, 1, 1},
    {&__pyx_n_s_tests_locals_add, __pyx_k_tests_locals_add, sizeof(__pyx_k_tests_locals_add), 0, 0, 1, 1},
    {&__pyx_n_s_tests_locals_add_locals_genexpr, __pyx_k_tests_locals_add_locals_genexpr, sizeof(__pyx_k_tests_locals_add_locals_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_tests_locals_test_readvarint, __pyx_k_tests_locals_test_readvarint, sizeof(__pyx_k_tests_locals_test_readvarint), 0, 0, 1, 1},
    {&__pyx_n_s_tests_locals_test_readvarlong, __pyx_k_tests_locals_test_readvarlong, sizeof(__pyx_k_tests_locals_test_readvarlong), 0, 0, 1, 1},
    {&__pyx_n_s_tests_locals_test_zigzag_long, __pyx_k_tests_locals_test_zigzag_long, sizeof(__pyx_k_tests_locals_test_zigzag_long), 0, 0, 1, 1},
    {&__pyx_n_s_threading, __pyx_k_threading, sizeof(__pyx_k_threading), 0, 0, 1, 1},
    {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
    {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
    {&__pyx_kp_u_time_micros, __pyx_k_time_micros, sizeof(__pyx_k_time_micros), 0, 1, 0, 0},
    {&__pyx_kp_u_time_millis, __pyx_k_time_millis, sizeof(__pyx_k_time_millis), 0, 1, 0, 0},
    {&__pyx_n_s_time_to_micros, __pyx_k_time_to_micros, sizeof(__pyx_k_time_to_micros), 0, 0, 1, 1},
    {&__pyx_n_s_timedelta, __pyx_k_timedelta, sizeof(__pyx_k_timedelta), 0, 0, 1, 1},
    {&__pyx_n_s_timestamp, __pyx_k_timestamp, sizeof(__pyx_k_timestamp), 0, 0, 1, 1},
    {&__pyx_kp_u_timestamp_micros, __pyx_k_timestamp_micros, sizeof(__pyx_k_timestamp_micros), 0, 1, 0, 0},
    {&__pyx_kp_u_timestamp_millis, __pyx_k_timestamp_millis, sizeof(__pyx_k_timestamp_millis), 0, 1, 0, 0},
    {&__pyx_n_s_timezone, __pyx_k_timezone, sizeof(__pyx_k_timezone), 0, 0, 1, 1},
    {&__pyx_kp_u_to, __pyx_k_to, sizeof(__pyx_k_to), 0, 1, 0, 0},
    {&__pyx_n_s_to_bytes, __pyx_k_to_bytes, sizeof(__pyx_k_to_bytes), 0, 0, 1, 1},
    {&__pyx_n_s_total_nums, __pyx_k_total_nums, sizeof(__pyx_k_total_nums), 0, 0, 1, 1},
    {&__pyx_n_s_truncate_fixed, __pyx_k_truncate_fixed, sizeof(__pyx_k_truncate_fixed), 0, 0, 1, 1},
    {&__pyx_kp_u_truncate_fixed_2, __pyx_k_truncate_fixed_2, sizeof(__pyx_k_truncate_fixed_2), 0, 1, 0, 0},
    {&__pyx_kp_s_tuple_LogicalType, __pyx_k_tuple_LogicalType, sizeof(__pyx_k_tuple_LogicalType), 0, 0, 1, 0},
    {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
    {&__pyx_n_u_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 1, 0, 1},
    {&__pyx_kp_u_type_2, __pyx_k_type_2, sizeof(__pyx_k_type_2), 0, 1, 0, 0},
    {&__pyx_n_s_type_3, __pyx_k_type_3, sizeof(__pyx_k_type_3), 0, 0, 1, 1},
    {&__pyx_n_s_type_name, __pyx_k_type_name, sizeof(__pyx_k_type_name), 0, 0, 1, 1},
    {&__pyx_n_s_type_str, __pyx_k_type_str, sizeof(__pyx_k_type_str), 0, 0, 1, 1},
    {&__pyx_n_s_types, __pyx_k_types, sizeof(__pyx_k_types), 0, 0, 1, 1},
    {&__pyx_n_s_types_str_to_schema, __pyx_k_types_str_to_schema, sizeof(__pyx_k_types_str_to_schema), 0, 0, 1, 1},
    {&__pyx_kp_u_types_str_to_schema_2, __pyx_k_types_str_to_schema_2, sizeof(__pyx_k_types_str_to_schema_2), 0, 1, 0, 0},
    {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
    {&__pyx_n_s_tzinfo, __pyx_k_tzinfo, sizeof(__pyx_k_tzinfo), 0, 0, 1, 1},
    {&__pyx_n_s_ubin, __pyx_k_ubin, sizeof(__pyx_k_ubin), 0, 0, 1, 1},
    {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
    {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
    {&__pyx_n_s_underlying, __pyx_k_underlying, sizeof(__pyx_k_underlying), 0, 0, 1, 1},
    {&__pyx_n_s_underlying_name, __pyx_k_underlying_name, sizeof(__pyx_k_underlying_name), 0, 0, 1, 1},
    {&__pyx_n_s_underlying_type, __pyx_k_underlying_type, sizeof(__pyx_k_underlying_type), 0, 0, 1, 1},
    {&__pyx_n_s_underlying_types, __pyx_k_underlying_types, sizeof(__pyx_k_underlying_types), 0, 0, 1, 1},
    {&__pyx_n_s_unicode_errors, __pyx_k_unicode_errors, sizeof(__pyx_k_unicode_errors), 0, 0, 1, 1},
    {&__pyx_kp_u_unicode_errors_2, __pyx_k_unicode_errors_2, sizeof(__pyx_k_unicode_errors_2), 0, 1, 0, 0},
    {&__pyx_n_s_unicodedata, __pyx_k_unicodedata, sizeof(__pyx_k_unicodedata), 0, 0, 1, 1},
    {&__pyx_n_u_union, __pyx_k_union, sizeof(__pyx_k_union), 0, 1, 0, 1},
    {&__pyx_n_s_union_decodes_to, __pyx_k_union_decodes_to, sizeof(__pyx_k_union_decodes_to), 0, 0, 1, 1},
    {&__pyx_kp_u_union_decodes_to_2, __pyx_k_union_decodes_to_2, sizeof(__pyx_k_union_decodes_to_2), 0, 1, 0, 0},
    {&__pyx_n_s_union_json_encodes_type_name, __pyx_k_union_json_encodes_type_name, sizeof(__pyx_k_union_json_encodes_type_name), 0, 0, 1, 1},
    {&__pyx_kp_u_union_json_encodes_type_name_2, __pyx_k_union_json_encodes_type_name_2, sizeof(__pyx_k_union_json_encodes_type_name_2), 0, 1, 0, 0},
    {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
    {&__pyx_n_s_unsafe_hash, __pyx_k_unsafe_hash, sizeof(__pyx_k_unsafe_hash), 0, 0, 1, 1},
    {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
    {&__pyx_n_s_use_setstate, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_utc, __pyx_k_utc, sizeof(__pyx_k_utc), 0, 0, 1, 1},
    {&__pyx_n_u_utf8, __pyx_k_utf8, sizeof(__pyx_k_utf8), 0, 1, 0, 1},
    {&__pyx_kp_u_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 1, 0, 0},
    {&__pyx_n_s_uuid, __pyx_k_uuid, sizeof(__pyx_k_uuid), 0, 0, 1, 1},
    {&__pyx_n_u_uuid, __pyx_k_uuid, sizeof(__pyx_k_uuid), 0, 1, 0, 1},
    {&__pyx_n_s_uuid4, __pyx_k_uuid4, sizeof(__pyx_k_uuid4), 0, 0, 1, 1},
    {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
    {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
    {&__pyx_kp_u_value_2, __pyx_k_value_2, sizeof(__pyx_k_value_2), 0, 1, 0, 0},
    {&__pyx_n_s_value_adapters, __pyx_k_value_adapters, sizeof(__pyx_k_value_adapters), 0, 0, 1, 1},
    {&__pyx_n_s_value_type, __pyx_k_value_type, sizeof(__pyx_k_value_type), 0, 0, 1, 1},
    {&__pyx_n_u_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 1, 0, 1},
    {&__pyx_n_s_version, __pyx_k_version, sizeof(__pyx_k_version), 0, 0, 1, 1},
    {&__pyx_n_s_version_info, __pyx_k_version_info, sizeof(__pyx_k_version_info), 0, 0, 1, 1},
    {&__pyx_n_s_visited, __pyx_k_visited, sizeof(__pyx_k_visited), 0, 0, 1, 1},
    {&__pyx_kp_u_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 1, 0, 0},
    {&__pyx_n_s_walk_types, __pyx_k_walk_types, sizeof(__pyx_k_walk_types), 0, 0, 1, 1},
    {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
    {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
    {&__pyx_n_u_wb, __pyx_k_wb, sizeof(__pyx_k_wb), 0, 1, 0, 1},
    {&__pyx_n_s_wbits, __pyx_k_wbits, sizeof(__pyx_k_wbits), 0, 0, 1, 1},
    {&__pyx_n_s_weakref_slot, __pyx_k_weakref_slot, sizeof(__pyx_k_weakref_slot), 0, 0, 1, 1},
    {&__pyx_kp_u_which_is_greater_than_the_scale, __pyx_k_which_is_greater_than_the_scale, sizeof(__pyx_k_which_is_greater_than_the_scale), 0, 1, 0, 0},
    {&__pyx_n_s_with_external_types, __pyx_k_with_external_types, sizeof(__pyx_k_with_external_types), 0, 0, 1, 1},
    {&__pyx_n_s_with_logical_types, __pyx_k_with_logical_types, sizeof(__pyx_k_with_logical_types), 0, 0, 1, 1},
    {&__pyx_n_s_wrap_type, __pyx_k_wrap_type, sizeof(__pyx_k_wrap_type), 0, 0, 1, 1},
    {&__pyx_n_s_wrapped, __pyx_k_wrapped, sizeof(__pyx_k_wrapped), 0, 0, 1, 1},
    {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
    {&__pyx_n_u_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 1, 0, 1},
    {&__pyx_kp_u_write_2, __pyx_k_write_2, sizeof(__pyx_k_write_2), 0, 1, 0, 0},
    {&__pyx_n_s_write_header, __pyx_k_write_header, sizeof(__pyx_k_write_header), 0, 0, 1, 1},
    {&__pyx_n_s_write_many, __pyx_k_write_many, sizeof(__pyx_k_write_many), 0, 0, 1, 1},
    {&__pyx_n_s_write_one, __pyx_k_write_one, sizeof(__pyx_k_write_one), 0, 0, 1, 1},
    {&__pyx_n_s_write_perf, __pyx_k_write_perf, sizeof(__pyx_k_write_perf), 0, 0, 1, 1},
    {&__pyx_n_s_write_time, __pyx_k_write_time, sizeof(__pyx_k_write_time), 0, 0, 1, 1},
    {&__pyx_n_s_writer, __pyx_k_writer, sizeof(__pyx_k_writer), 0, 0, 1, 1},
    {&__pyx_n_s_writer_schema, __pyx_k_writer_schema, sizeof(__pyx_k_writer_schema), 0, 0, 1, 1},
    {&__pyx_n_s_writer_type, __pyx_k_writer_type, sizeof(__pyx_k_writer_type), 0, 0, 1, 1},
    {&__pyx_n_b_xz, __pyx_k_xz, sizeof(__pyx_k_xz), 0, 0, 0, 1},
    {&__pyx_n_u_xz, __pyx_k_xz, sizeof(__pyx_k_xz), 0, 1, 0, 1},
    {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
    {&__pyx_n_s_zlib, __pyx_k_zlib, sizeof(__pyx_k_zlib), 0, 0, 1, 1},
    {&__pyx_n_b_zstandard, __pyx_k_zstandard, sizeof(__pyx_k_zstandard), 0, 0, 0, 1},
    {&__pyx_n_s_zstandard, __pyx_k_zstandard, sizeof(__pyx_k_zstandard), 0, 0, 1, 1},
    {&__pyx_n_u_zstandard, __pyx_k_zstandard, sizeof(__pyx_k_zstandard), 0, 1, 0, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 71, __pyx_L1_error)
  __pyx_builtin___import__ = __Pyx_GetBuiltinName(__pyx_n_s_import); if (!__pyx_builtin___import__) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_builtin_NotImplemented = __Pyx_GetBuiltinName(__pyx_n_s_NotImplemented); if (!__pyx_builtin_NotImplemented) __PYX_ERR(2, 57, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(3, 6, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 22, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(4, 11, __pyx_L1_error)
  __pyx_builtin_chr = __Pyx_GetBuiltinName(__pyx_n_s_chr); if (!__pyx_builtin_chr) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_builtin_eval = __Pyx_GetBuiltinName(__pyx_n_s_eval); if (!__pyx_builtin_eval) __PYX_ERR(4, 53, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(1, 68, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(5, 7, __pyx_L1_error)
  __pyx_builtin_EOFError = __Pyx_GetBuiltinName(__pyx_n_s_EOFError); if (!__pyx_builtin_EOFError) __PYX_ERR(5, 52, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_bin = __Pyx_GetBuiltinName(__pyx_n_s_bin); if (!__pyx_builtin_bin) __PYX_ERR(6, 3, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(6, 49, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(7, 37, __pyx_L1_error)
  __pyx_builtin_hex = __Pyx_GetBuiltinName(__pyx_n_s_hex); if (!__pyx_builtin_hex) __PYX_ERR(7, 37, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(8, 39, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(8, 92, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(9, 96, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(9, 102, __pyx_L1_error)
  __pyx_builtin_OverflowError = __Pyx_GetBuiltinName(__pyx_n_s_OverflowError); if (!__pyx_builtin_OverflowError) __PYX_ERR(10, 238, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(11, 142, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(11, 410, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(11, 483, __pyx_L1_error)
  __pyx_builtin_UnicodeEncodeError = __Pyx_GetBuiltinName(__pyx_n_s_UnicodeEncodeError); if (!__pyx_builtin_UnicodeEncodeError) __PYX_ERR(12, 45, __pyx_L1_error)
  __pyx_builtin_reversed = __Pyx_GetBuiltinName(__pyx_n_s_reversed); if (!__pyx_builtin_reversed) __PYX_ERR(2, 161, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(1, 156, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(1, 408, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "View.MemoryView":582
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
  __pyx_tuple__4 = PyTuple_New(1); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_tuple__4, 0, __pyx_int_neg_1)) __PYX_ERR(1, 582, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "View.MemoryView":679
 *     tup = <tuple>index if isinstance(index, tuple) else (index,)
 * 
 *     result = [slice(None)] * ndim             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_slice__5 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__5)) __PYX_ERR(1, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__5);
  __Pyx_GIVEREF(__pyx_slice__5);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 */
  __pyx_tuple__8 = PyTuple_Pack(3, __pyx_int_136983863, __pyx_int_112105877, __pyx_int_184977713); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "cavro.pyx":49
 * __tests = {}
 * def _tests(fn):
 *     def add(test, *args):             # <<<<<<<<<<<<<<
 *         name = test.__name__
 *         if args:
 */
  __pyx_tuple__10 = PyTuple_Pack(5, __pyx_n_s_test, __pyx_n_s_args, __pyx_n_s_name, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(13, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cavro_pyx, __pyx_n_s_add, 49, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(13, 49, __pyx_L1_error)

  /* "cavro.pyx":59
 * __perf_tests = {}
 * def _perf(fn):
 *     def add(test, *args):             # <<<<<<<<<<<<<<
 *         name = test.__name__
 *         if args:
 */
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cavro_pyx, __pyx_n_s_add, 59, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(13, 59, __pyx_L1_error)

  /* "src/io.pxi":52
 *         cdef bytes data = self.file_obj.read(1)
 *         if len(data) != 1:
 *             raise EOFError(f"End of file found trying to read 1 byte")             # <<<<<<<<<<<<<<
 *         return ord(data)
 * 
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_u_End_of_file_found_trying_to_read); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(5, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "src/buffer.pxi":79
 *         cdef const uint8_t *ptr = self.ptr
 *         if <size_t>(self.end_ptr - self.ptr) < num:
 *             raise EOFError("Not enough input data to read value")             # <<<<<<<<<<<<<<
 *         self.ptr += num
 *         return ptr
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_u_Not_enough_input_data_to_read_va); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(16, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "src/rabin.pxi":54
 *     def digest(self):
 *         cdef uint64_t value = self.value
 *         return int(value).to_bytes(8, 'little', signed=False)             # <<<<<<<<<<<<<<
 * 
 *     def hexdigest(self):
 */
  __pyx_tuple__22 = PyTuple_Pack(2, __pyx_int_8, __pyx_n_u_little); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(18, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "src/tests/test_zigzag.pxi":3
 * 
 * cpdef str _ubin(val, n_bytes):
 *     cdef str raw = bin(val)[2:]             # <<<<<<<<<<<<<<
 *     padded = raw.rjust(n_bytes * 8, '0')
 *     parts = [padded[i:i+8] for i in range(0, len(padded), 8)]
 */
  __pyx_slice__23 = PySlice_New(__pyx_int_2, Py_None, Py_None); if (unlikely(!__pyx_slice__23)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__23);
  __Pyx_GIVEREF(__pyx_slice__23);

  /* "src/tests/test_zigzag.pxi":33
 * def _tests(add):
 * 
 *     def test_readvarint(uint32_t mask):             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */
  __pyx_tuple__28 = PyTuple_Pack(7, __pyx_n_s_mask, __pyx_n_s_state, __pyx_n_s_actual, __pyx_n_s_given, __pyx_n_s_src, __pyx_n_s_reader, __pyx_n_s_i); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(6, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_tests_test_zigzag_pxi, __pyx_n_s_test_readvarint, 33, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(6, 33, __pyx_L1_error)

  /* "src/tests/test_zigzag.pxi":53
 *         add(test_readvarint, mask)
 * 
 *     def test_zigzag_long(uint64_t mask):             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(8, __pyx_n_s_mask, __pyx_n_s_state, __pyx_n_s_actual, __pyx_n_s_given, __pyx_n_s_writer, __pyx_n_s_reader, __pyx_n_s_i, __pyx_n_s_pad); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(6, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_tests_test_zigzag_pxi, __pyx_n_s_test_zigzag_long, 53, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(6, 53, __pyx_L1_error)

  /* "src/tests/test_zigzag.pxi":74
 *                 raise AssertionError(f"{writer.bytes()}: {given}\n   {_ubin(actual, 8)}\n!= {_ubin(given, 8)}")
 * 
 *     def test_readvarlong(uint64_t mask):             # <<<<<<<<<<<<<<
 *         cdef random_t state
 * 
 */
  __pyx_tuple__32 = PyTuple_Pack(8, __pyx_n_s_mask, __pyx_n_s_state, __pyx_n_s_actual, __pyx_n_s_given, __pyx_n_s_src, __pyx_n_s_reader, __pyx_n_s_i, __pyx_n_s_pad); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(6, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_tests_test_zigzag_pxi, __pyx_n_s_test_readvarlong, 74, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(6, 74, __pyx_L1_error)

  /* "src/tests/test_buffer.pxi":7
 *     @add
 *     def _test_reader_read_u8():
 *         buffer = MemoryReader(b'\x01\x02\x03')             # <<<<<<<<<<<<<<
 *         assert buffer.read_u8() == 1
 *         assert buffer.read_u8() == 2
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_b__34); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(19, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "src/tests/test_buffer.pxi":12
 *         assert buffer.read_u8() == 3
 *         import pytest
 *         with pytest.raises(EOFError):             # <<<<<<<<<<<<<<
 *             buffer.read_u8()
 * 
 */
  __pyx_tuple__36 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(19, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "src/tests/test_buffer.pxi":17
 *     @add
 *     def _test_reader_read_n():
 *         buffer = MemoryReader(b'abacuscounter1')             # <<<<<<<<<<<<<<
 *         a = bytes(buffer.read_n(6)[:6])
 *         assert a == b'abacus', a
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_n_b_abacuscounter1); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(19, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "src/tests/test_buffer.pxi":5
 * def _tests(add):
 * 
 *     @add             # <<<<<<<<<<<<<<
 *     def _test_reader_read_u8():
 *         buffer = MemoryReader(b'\x01\x02\x03')
 */
  __pyx_tuple__38 = PyTuple_Pack(2, __pyx_n_s_buffer, __pyx_n_s_pytest); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(19, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_tests_test_buffer_pxi, __pyx_n_s_test_reader_read_u8, 5, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(19, 5, __pyx_L1_error)

  /* "src/tests/test_buffer.pxi":15
 *             buffer.read_u8()
 * 
 *     @add             # <<<<<<<<<<<<<<
 *     def _test_reader_read_n():
 *         buffer = MemoryReader(b'abacuscounter1')
 */
  __pyx_tuple__40 = PyTuple_Pack(4, __pyx_n_s_buffer, __pyx_n_s_a, __pyx_n_s_b, __pyx_n_s_pytest); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(19, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_tests_test_buffer_pxi, __pyx_n_s_test_reader_read_n, 15, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(19, 15, __pyx_L1_error)

  /* "src/tests/test_perf.pxi":14
 *         cdef uint64_t given
 *         cdef bytes src
 *         cdef MemoryWriter writer = MemoryWriter(3442987655)             # <<<<<<<<<<<<<<
 *         cdef MemoryReader reader
 * 
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_int_3442987655); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(7, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "src/tests/test_perf.pxi":4
 * @_perf
 * def _perf(add):
 *     def _test_zigzag_perf(const uint64_t mask):             # <<<<<<<<<<<<<<
 *         from time import perf_counter_ns
 *         import os
 */
  __pyx_tuple__43 = PyTuple_Pack(20, __pyx_n_s_mask, __pyx_n_s_perf_counter_ns, __pyx_n_s_os, __pyx_n_s_state, __pyx_n_s_BATCH_SIZE, __pyx_n_s_actual, __pyx_n_s_given, __pyx_n_s_src, __pyx_n_s_writer, __pyx_n_s_reader, __pyx_n_s_total_nums, __pyx_n_s_n, __pyx_n_s_i, __pyx_n_s_out_state, __pyx_n_s_t_a, __pyx_n_s_write_time, __pyx_n_s_copy_time, __pyx_n_s_read_time, __pyx_n_s_write_perf, __pyx_n_s_read_perf); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(7, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 20, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_tests_test_perf_pxi, __pyx_n_s_test_zigzag_perf, 4, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(7, 4, __pyx_L1_error)

  /* "src/enum.pxi":16
 *         raw_symbols = source['symbols']
 *         if not isinstance(raw_symbols, (list, tuple)):
 *             raise ValueError("Enum symbols must be a list of strings")             # <<<<<<<<<<<<<<
 * 
 *         if schema_.options.enum_symbols_must_be_unique:
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_kp_u_Enum_symbols_must_be_a_list_of_s); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(8, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "src/enum.pxi":45
 *             if self.default_value not in self.symbol_indexes:
 *                 raise ValueError(f"Default value '{self.default_value}' not in enum symbols")
 *         self.doc = source.get('doc', '')             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  __pyx_tuple__47 = PyTuple_Pack(2, __pyx_n_u_doc, __pyx_kp_u__13); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(8, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "src/null.pxi":44
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"null"')             # <<<<<<<<<<<<<<
 * 
 *     cdef bint accepts_missing_value(self):
 */
  __pyx_tuple__50 = PyTuple_Pack(1, __pyx_kp_u_null); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(21, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "src/numeric_types.pxi":57
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"boolean"')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__51 = PyTuple_Pack(1, __pyx_kp_u_boolean); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(10, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);

  /* "src/numeric_types.pxi":129
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"int"')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__52 = PyTuple_Pack(1, __pyx_kp_u_int); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(10, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);

  /* "src/numeric_types.pxi":201
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"long"')             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 */
  __pyx_tuple__53 = PyTuple_Pack(1, __pyx_kp_u_long); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(10, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);

  /* "src/numeric_types.pxi":314
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"float"')             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_kp_u_float); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(10, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "src/numeric_types.pxi":405
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"double"')             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 */
  __pyx_tuple__55 = PyTuple_Pack(1, __pyx_kp_u_double); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(10, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);

  /* "src/record.pxi":41
 * 
 *     cdef int binary_buffer_encode(self, _Writer buffer, value) except -1:
 *         raise NotImplementedError("Placeholder types cannot be encoded")             # <<<<<<<<<<<<<<
 * 
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 */
  __pyx_tuple__56 = PyTuple_Pack(1, __pyx_kp_u_Placeholder_types_cannot_be_enco); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(11, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "src/record.pxi":59
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         raise NotImplementedError("Placeholder types have no canonical form")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__57 = PyTuple_Pack(1, __pyx_kp_u_Placeholder_types_have_no_canoni); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(11, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);

  /* "src/record.pxi":120
 *                 raise ValueError(f"Record {data} cannot be adapted to {self.Type}")
 *             if kwargs:
 *                 raise ValueError(f"Records may either be instantiated with a record, or **kwargs, not both")             # <<<<<<<<<<<<<<
 *         elif data:
 *             if kwargs:
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_kp_u_Records_may_either_be_instantiat); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(11, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "src/record.pxi":123
 *         elif data:
 *             if kwargs:
 *                 raise ValueError(f"Records may either be instantiated with a single dict, or **kwargs, not both")             # <<<<<<<<<<<<<<
 *             data_dict = data
 *         else:
 */
  __pyx_tuple__59 = PyTuple_Pack(1, __pyx_kp_u_Records_may_either_be_instantiat_2); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(11, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "src/record.pxi":210
 *         self.type = AvroType.for_source(schema, source['type'], namespace)
 *         self.default_value = self.type.resolve_default_value(source.get('default', NO_DEFAULT), self.name)
 *         self.order = Order(source.get('order', 'ascending'))             # <<<<<<<<<<<<<<
 *         alias_val = source.get('aliases', [])
 *         if not isinstance(alias_val, (list, tuple, set)):
 */
  __pyx_tuple__65 = PyTuple_Pack(2, __pyx_n_u_order, __pyx_n_u_ascending); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(11, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "src/string_types.pxi":71
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"bytes"')             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 */
  __pyx_tuple__67 = PyTuple_Pack(1, __pyx_kp_u_bytes_2); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(12, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);

  /* "src/string_types.pxi":115
 *     cdef _binary_buffer_decode(self, _Reader buffer):
 *         cdef uint64_t length = zigzag_decode_long(buffer)
 *         return buffer.read_bytes(length).decode('utf-8', errors=self.options.unicode_errors)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _get_value_fitness(self, value) except -1:
 */
  __pyx_tuple__68 = PyTuple_Pack(1, __pyx_kp_u_utf_8); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(12, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);

  /* "src/string_types.pxi":143
 * 
 *     cdef _CanonicalForm canonical_form(self, set created):
 *         return _CanonicalForm('"string"')             # <<<<<<<<<<<<<<
 * 
 *     cdef AvroType _for_writer(self, AvroType writer):
 */
  __pyx_tuple__69 = PyTuple_Pack(1, __pyx_kp_u_string_2); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(12, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);

  /* "src/string_types.pxi":219
 *     cdef _json_format(self, value):
 *         value = self._convert_value(value)
 *         return value.decode('latin1', errors=self.options.unicode_errors)             # <<<<<<<<<<<<<<
 * 
 *     cdef _json_decode(self, value):
 */
  __pyx_tuple__71 = PyTuple_Pack(1, __pyx_n_u_latin1); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(12, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);

  /* "src/union.pxi":37
 *         self.union_types = tuple(AvroType.for_source(schema, s, namespace) for s in source)
 *         if len(self.union_types) == 0 and not self.options.allow_empty_unions:
 *             raise ValueError("Unions must contain at least one member type")             # <<<<<<<<<<<<<<
 * 
 *         if not (self.options.allow_duplicate_union_types and self.options.allow_nested_unions):
 */
  __pyx_tuple__73 = PyTuple_Pack(1, __pyx_kp_u_Unions_must_contain_at_least_one); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(22, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);

  /* "src/union.pxi":69
 * 
 *             if not self.options.allow_nested_unions and isinstance(member, UnionType):
 *                 raise ValueError("Unions may not immediately contain other unions")             # <<<<<<<<<<<<<<
 *             if not self.options.allow_duplicate_union_types:
 *                 if not isinstance(member, (RecordType, FixedType, EnumType)):
 */
  __pyx_tuple__74 = PyTuple_Pack(1, __pyx_kp_u_Unions_may_not_immediately_conta); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(22, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);

  /* "src/logical.pxi":110
 *                 raise ValueError(f'Precision must be an integer between 1 and {decimal.MAX_PREC}, got: {repr(precision)}')
 *             return
 *         scale = meta.get('scale', 0)             # <<<<<<<<<<<<<<
 *         if not isinstance(scale, int) or scale < 0:
 *             if underlying.options.raise_on_invalid_logical:
 */
  __pyx_tuple__85 = PyTuple_Pack(2, __pyx_n_u_scale, __pyx_int_0); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(24, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);

  /* "src/codec.pxi":117
 * 
 *     cdef ssize_t write_block(self, _Writer writer, const uint8_t[:] data) except -1:
 *         cdef bytes compressed = zlib.compress(data)[2:-1]             # <<<<<<<<<<<<<<
 *         writer.write_n(compressed)
 *         return len(compressed)
 */
  __pyx_slice__86 = PySlice_New(__pyx_int_2, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__86)) __PYX_ERR(3, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__86);
  __Pyx_GIVEREF(__pyx_slice__86);

  /* "src/container.pxi":85
 *             self.schema = self.reader_schema.reader_for_writer(self.writer_schema)
 * 
 *         codec_name = self.metadata.get('avro.codec', b'null')             # <<<<<<<<<<<<<<
 *         self.codec_name = codec_name.decode()
 *         try:
 */
  __pyx_tuple__88 = PyTuple_Pack(2, __pyx_kp_u_avro_codec, __pyx_n_b_null_2); if (unlikely(!__pyx_tuple__88)) __PYX_ERR(25, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);

  /* "src/container.pxi":173
 *     def __cinit__(self, dest, Schema schema, str codec='null', size_t max_blocksize=16352, write_header=True, metadata=None, marker=None, options=DEFAULT_OPTIONS):
 *         if schema is None:
 *             raise ValueError('Schema is required')             # <<<<<<<<<<<<<<
 *         self.should_write_header = write_header
 *         self.num_pending = 0
 */
  __pyx_tuple__90 = PyTuple_Pack(1, __pyx_kp_u_Schema_is_required); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(25, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);

  /* "src/container.pxi":209
 *     def __enter__(self):
 *         if self.writer is None:
 *             raise ValueError('Container is closed')             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_tuple__91 = PyTuple_Pack(1, __pyx_kp_u_Container_is_closed); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(25, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);

  /* "src/container.pxi":240
 *     def close(self):
 *         if self.writer is None:
 *             raise ValueError('Trying to close a closed Container')             # <<<<<<<<<<<<<<
 *         cdef _Writer writer = self.writer
 *         self._flush_block(self.blocks_written == 0)
 */
  __pyx_tuple__92 = PyTuple_Pack(1, __pyx_kp_u_Trying_to_close_a_closed_Contain); if (unlikely(!__pyx_tuple__92)) __PYX_ERR(25, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);

  /* "src/container.pxi":269
 *     cpdef int write_one(self, obj) except -1:
 *         if self.next_item is None:
 *             raise ValueError('Trying to write to closed Container')             # <<<<<<<<<<<<<<
 *         self.next_item.reset()
 *         self.schema.binary_write(self.next_item, obj)
 */
  __pyx_tuple__93 = PyTuple_Pack(1, __pyx_kp_u_Trying_to_write_to_closed_Contai); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(25, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);

  /* "src/promotions.pxi":10
 * 
 *     cdef encode_value(self, value):
 *         raise NotImplementedError(f"Promoting schemas (reader/writer) do no support encoding values, use the writer schema directly.")             # <<<<<<<<<<<<<<
 * 
 *     cdef decode_value(self, value):
 */
  __pyx_tuple__94 = PyTuple_Pack(1, __pyx_kp_u_Promoting_schemas_reader_writer); if (unlikely(!__pyx_tuple__94)) __PYX_ERR(26, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc466972, 0x8dd8fb8, 0xf76cf3c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc466972, 0x8dd8fb8, 0xf76cf3c) = (adapt_record_types, allow_aliases_to_be_string, allow_duplicate_union_types, allow_empty_unions, allow_error_type, allow_false_values_for_null, allow_invalid_default_values, allow_leading_dot_in_names, allow_nested_unions, allow_primitive_name_collision, allow_primitive_names_in_namespaces, allow_tuple_notation, allow_union_default_any_member, alternate_timestamp_millis_encoding, ascii_name_rules, bytes_codec, bytes_default_value_utf8, canonical_form_repeat_enum, canonical_form_repeat_fixed, clamp_float_overflow, clamp_int_overflow, coerce_int_to_float, coerce_values_to_boolean, coerce_values_to_float, coerce_values_to_int, coerce_values_to_str, container_fill_blocks, date_type_accepts_string, decimal_check_exp_overflow, defer_schema_promotion_errors, enforce_enum_symbol_name_rules, enforce_namespace_name_rules, enforce_type_name_rules, enum_symbols_must_be_unique, expand_types_in_schema, externally_defined_types, fingerprint_returns_digest, fixed_codec, float_out_of_range_inf, inline_namespaces, invalid_value_include_array_index, invalid_value_include_map_key, invalid_value_includes_record_name, logical_types, missing_values_can_be_empty_container, missing_values_can_be_null, naive_dt_assume_utc, named_type_names_must_be_unique, null_pad_fixed, raise_on_invalid_logical, record_allow_extra_fields, record_can_encode_dict, record_decodes_to_dict, record_encode_use_defaults, record_fields_must_be_unique, record_values_type_hint, return_uuid_object, string_types_default_unchanged, truncate_fixed, types_str_to_schema, unicode_errors, union_decodes_to, union_json_encodes_type_name))" % __pyx_checksum
 */
  __pyx_tuple__95 = PyTuple_Pack(3, __pyx_int_205941106, __pyx_int_148737976, __pyx_int_259444540); if (unlikely(!__pyx_tuple__95)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);
  __pyx_tuple__96 = PyTuple_Pack(3, __pyx_int_238750788, __pyx_int_228825662, __pyx_int_222419149); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);
  __pyx_tuple__97 = PyTuple_Pack(3, __pyx_int_122446337, __pyx_int_61440409, __pyx_int_39872412); if (unlikely(!__pyx_tuple__97)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__97);
  __Pyx_GIVEREF(__pyx_tuple__97);
  __pyx_tuple__98 = PyTuple_Pack(3, __pyx_int_247527084, __pyx_int_198151601, __pyx_int_256569670); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);
  __pyx_tuple__99 = PyTuple_Pack(3, __pyx_int_215229444, __pyx_int_254981756, __pyx_int_33963030); if (unlikely(!__pyx_tuple__99)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);
  __pyx_tuple__100 = PyTuple_Pack(3, __pyx_int_215954136, __pyx_int_101670942, __pyx_int_177382991); if (unlikely(!__pyx_tuple__100)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);
  __pyx_tuple__101 = PyTuple_Pack(3, __pyx_int_150354737, __pyx_int_18332779, __pyx_int_184956420); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);
  __pyx_tuple__102 = PyTuple_Pack(3, __pyx_int_145104888, __pyx_int_75337510, __pyx_int_195980960); if (unlikely(!__pyx_tuple__102)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);
  __pyx_tuple__103 = PyTuple_Pack(3, __pyx_int_178791180, __pyx_int_79941127, __pyx_int_83351333); if (unlikely(!__pyx_tuple__103)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);
  __pyx_tuple__104 = PyTuple_Pack(3, __pyx_int_139219010, __pyx_int_176267327, __pyx_int_79389776); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);
  __pyx_tuple__105 = PyTuple_Pack(3, __pyx_int_96605483, __pyx_int_160503772, __pyx_int_236349525); if (unlikely(!__pyx_tuple__105)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__105);
  __Pyx_GIVEREF(__pyx_tuple__105);
  __pyx_tuple__106 = PyTuple_Pack(3, __pyx_int_61270791, __pyx_int_169331114, __pyx_int_148338675); if (unlikely(!__pyx_tuple__106)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);
  __pyx_tuple__107 = PyTuple_Pack(3, __pyx_int_31039736, __pyx_int_81233721, __pyx_int_38410757); if (unlikely(!__pyx_tuple__107)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);
  __pyx_tuple__108 = PyTuple_Pack(3, __pyx_int_29099189, __pyx_int_240389341, __pyx_int_111776469); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);
  __pyx_tuple__109 = PyTuple_Pack(3, __pyx_int_25341170, __pyx_int_24827247, __pyx_int_233298401); if (unlikely(!__pyx_tuple__109)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__109);
  __Pyx_GIVEREF(__pyx_tuple__109);
  __pyx_tuple__110 = PyTuple_Pack(3, __pyx_int_67877419, __pyx_int_119832798, __pyx_int_111957047); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);
  __pyx_tuple__111 = PyTuple_Pack(3, __pyx_int_228539134, __pyx_int_181080083, __pyx_int_101034655); if (unlikely(!__pyx_tuple__111)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);
  __pyx_tuple__112 = PyTuple_Pack(3, __pyx_int_180285838, __pyx_int_115671040, __pyx_int_159023012); if (unlikely(!__pyx_tuple__112)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);
  __pyx_tuple__113 = PyTuple_Pack(3, __pyx_int_38632982, __pyx_int_202608488, __pyx_int_229636597); if (unlikely(!__pyx_tuple__113)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__113);
  __Pyx_GIVEREF(__pyx_tuple__113);
  __pyx_tuple__114 = PyTuple_Pack(3, __pyx_int_203789025, __pyx_int_236180632, __pyx_int_175892883); if (unlikely(!__pyx_tuple__114)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);
  __pyx_tuple__115 = PyTuple_Pack(3, __pyx_int_245215779, __pyx_int_202158861, __pyx_int_160934521); if (unlikely(!__pyx_tuple__115)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__115);
  __Pyx_GIVEREF(__pyx_tuple__115);
  __pyx_tuple__116 = PyTuple_Pack(3, __pyx_int_159891379, __pyx_int_176081420, __pyx_int_61765448); if (unlikely(!__pyx_tuple__116)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);
  __pyx_tuple__117 = PyTuple_Pack(3, __pyx_int_206778879, __pyx_int_180170947, __pyx_int_159276020); if (unlikely(!__pyx_tuple__117)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);
  __pyx_tuple__118 = PyTuple_Pack(3, __pyx_int_148222551, __pyx_int_95761985, __pyx_int_47772561); if (unlikely(!__pyx_tuple__118)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);
  __pyx_tuple__119 = PyTuple_Pack(3, __pyx_int_148208818, __pyx_int_263819094, __pyx_int_193777858); if (unlikely(!__pyx_tuple__119)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__119);
  __Pyx_GIVEREF(__pyx_tuple__119);
  __pyx_tuple__120 = PyTuple_Pack(3, __pyx_int_42565757, __pyx_int_122978751, __pyx_int_164094308); if (unlikely(!__pyx_tuple__120)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);
  __pyx_tuple__121 = PyTuple_Pack(3, __pyx_int_213886906, __pyx_int_261059191, __pyx_int_212073902); if (unlikely(!__pyx_tuple__121)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__121);
  __Pyx_GIVEREF(__pyx_tuple__121);

  /* "View.MemoryView":100
 * cdef object __pyx_collections_abc_Sequence "__pyx_collections_abc_Sequence"
 * try:
 *     if __import__("sys").version_info >= (3, 3):             # <<<<<<<<<<<<<<
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 *     else:
 */
  __pyx_tuple__123 = PyTuple_Pack(1, __pyx_n_s_sys); if (unlikely(!__pyx_tuple__123)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);
  __pyx_tuple__124 = PyTuple_Pack(2, __pyx_int_3, __pyx_int_3); if (unlikely(!__pyx_tuple__124)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);

  /* "View.MemoryView":101
 * try:
 *     if __import__("sys").version_info >= (3, 3):
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence             # <<<<<<<<<<<<<<
 *     else:
 *         __pyx_collections_abc_Sequence = __import__("collections").Sequence
 */
  __pyx_tuple__125 = PyTuple_Pack(1, __pyx_kp_s_collections_abc); if (unlikely(!__pyx_tuple__125)) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);

  /* "View.MemoryView":103
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 *     else:
 *         __pyx_collections_abc_Sequence = __import__("collections").Sequence             # <<<<<<<<<<<<<<
 * except:
 * 
 */
  __pyx_tuple__126 = PyTuple_Pack(1, __pyx_n_s_collections); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(1, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);

  /* "View.MemoryView":309
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_tuple__127 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__127)) __PYX_ERR(1, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);

  /* "View.MemoryView":310
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_tuple__128 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__128)) __PYX_ERR(1, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);

  /* "View.MemoryView":311
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__129 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(1, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);

  /* "View.MemoryView":314
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_tuple__130 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__130)) __PYX_ERR(1, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);

  /* "View.MemoryView":315
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__131 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__131)) __PYX_ERR(1, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__132 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);
  __pyx_codeobj__133 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__133)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "cavro.pyx":48
 * 
 * __tests = {}
 * def _tests(fn):             # <<<<<<<<<<<<<<
 *     def add(test, *args):
 *         name = test.__name__
 */
  __pyx_tuple__134 = PyTuple_Pack(3, __pyx_n_s_fn, __pyx_n_s_add, __pyx_n_s_add); if (unlikely(!__pyx_tuple__134)) __PYX_ERR(13, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);
  __pyx_codeobj__135 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__134, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cavro_pyx, __pyx_n_s_tests_2, 48, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__135)) __PYX_ERR(13, 48, __pyx_L1_error)

  /* "cavro.pyx":58
 * 
 * __perf_tests = {}
 * def _perf(fn):             # <<<<<<<<<<<<<<
 *     def add(test, *args):
 *         name = test.__name__
 */
  __pyx_codeobj__136 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__134, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cavro_pyx, __pyx_n_s_perf, 58, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__136)) __PYX_ERR(13, 58, __pyx_L1_error)

  /* "src/error.pxi":20
 *     """
 * 
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *         super().__init__(f"Unknown type: {name}")
 */
  __pyx_tuple__137 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_name); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);
  __pyx_codeobj__138 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_error_pxi, __pyx_n_s_init, 20, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__138)) __PYX_ERR(0, 20, __pyx_L1_error)

  /* "src/error.pxi":61
 *     """
 * 
 *     def __init__(self, reader_type, writer_type, extra=None):             # <<<<<<<<<<<<<<
 *         self.reader_type = reader_type
 *         self.writer_type = writer_type
 */
  __pyx_tuple__139 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_reader_type, __pyx_n_s_writer_type, __pyx_n_s_extra, __pyx_n_s_msg); if (unlikely(!__pyx_tuple__139)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);
  __pyx_codeobj__140 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__139, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_error_pxi, __pyx_n_s_init, 61, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__140)) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_tuple__141 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__141)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__141);
  __Pyx_GIVEREF(__pyx_tuple__141);

  /* "src/error.pxi":82
 *     """
 * 
 *     def __init__(self, value, dest_type, path=()):             # <<<<<<<<<<<<<<
 *         self.value = value
 *         self.dest_type = dest_type
 */
  __pyx_tuple__142 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_value, __pyx_n_s_dest_type, __pyx_n_s_path); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);
  __pyx_codeobj__143 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__142, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_error_pxi, __pyx_n_s_init, 82, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__143)) __PYX_ERR(0, 82, __pyx_L1_error)
  __pyx_tuple__144 = PyTuple_Pack(1, ((PyObject*)__pyx_empty_tuple)); if (unlikely(!__pyx_tuple__144)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);

  /* "src/error.pxi":88
 *         super().__init__()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.schema_path:
 *             path_str = ".".join(str(p) for p in self.schema_path)
 */
  __pyx_tuple__145 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_path_str, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);
  __pyx_codeobj__146 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__145, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_error_pxi, __pyx_n_s_str, 88, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__146)) __PYX_ERR(0, 88, __pyx_L1_error)

  /* "src/option.pxi":3
 * 
 * 
 * NAME_RE_STRICT = re.compile(r'[A-Za-z_][A-Za-z0-9_]*')             # <<<<<<<<<<<<<<
 * NAME_RE_UNICODE = None
 * 
 */
  __pyx_tuple__147 = PyTuple_Pack(1, __pyx_kp_u_A_Za_z__A_Za_z0_9); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(4, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);

  /* "src/option.pxi":49
 * _EMPTY_MP = MappingProxyType({})
 * 
 * def _eval(val):             # <<<<<<<<<<<<<<
 *     if isinstance(val, str):
 *         if val == 'bint':
 */
  __pyx_tuple__148 = PyTuple_Pack(1, __pyx_n_s_val); if (unlikely(!__pyx_tuple__148)) __PYX_ERR(4, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);
  __pyx_codeobj__149 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__148, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_option_pxi, __pyx_n_s_eval_2, 49, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__149)) __PYX_ERR(4, 49, __pyx_L1_error)

  /* "src/option.pxi":70
 *     '''
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.__signature__ = None
 *         self.annotations = {}
 */
  __pyx_tuple__150 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__150)) __PYX_ERR(4, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__150);
  __Pyx_GIVEREF(__pyx_tuple__150);
  __pyx_codeobj__151 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_option_pxi, __pyx_n_s_init, 70, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__151)) __PYX_ERR(4, 70, __pyx_L1_error)

  /* "src/option.pxi":74
 *         self.annotations = {}
 * 
 *     def _set_from_dataclass(self, cls):             # <<<<<<<<<<<<<<
 *         try:
 *             fields = [(f.name, f.type, f) for f in dataclasses.fields(cls)]
 */
  __pyx_tuple__153 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_cls, __pyx_n_s_fields, __pyx_n_s_name, __pyx_n_s_type_str, __pyx_n_s__152, __pyx_n_s_inst, __pyx_n_s_f); if (unlikely(!__pyx_tuple__153)) __PYX_ERR(4, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__153);
  __Pyx_GIVEREF(__pyx_tuple__153);
  __pyx_codeobj__154 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__153, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_option_pxi, __pyx_n_s_set_from_dataclass, 74, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__154)) __PYX_ERR(4, 74, __pyx_L1_error)

  /* "src/option.pxi":285
 *     externally_defined_types: MappingProxyType = _EMPTY_MP
 * 
 *     def replace(self, **changes) -> Options:             # <<<<<<<<<<<<<<
 *         '''
 *         Return a copy of the options with the specified fields replaced.
 */
  __pyx_tuple__155 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_changes); if (unlikely(!__pyx_tuple__155)) __PYX_ERR(4, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__155);
  __Pyx_GIVEREF(__pyx_tuple__155);
  __pyx_codeobj__156 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__155, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_option_pxi, __pyx_n_s_replace, 285, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__156)) __PYX_ERR(4, 285, __pyx_L1_error)

  /* "src/option.pxi":292
 *         return dataclasses.replace(self, **changes)
 * 
 *     def with_logical_types(self, *logical_types):             # <<<<<<<<<<<<<<
 *         '''
 *         Return a copy of the options with additional logical types added.
 */
  __pyx_tuple__157 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_logical_types, __pyx_n_s_new_types); if (unlikely(!__pyx_tuple__157)) __PYX_ERR(4, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__157);
  __Pyx_GIVEREF(__pyx_tuple__157);
  __pyx_codeobj__158 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__157, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_option_pxi, __pyx_n_s_with_logical_types, 292, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__158)) __PYX_ERR(4, 292, __pyx_L1_error)

  /* "src/option.pxi":299
 *         return self.replace(logical_types=new_types)
 * 
 *     def with_external_types(self, named_types: dict[str, AvroType]) -> Options:             # <<<<<<<<<<<<<<
 *         '''
 *         Return a copy of the options with pre-parsed external named types added
 */
  __pyx_tuple__159 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_named_types, __pyx_n_s_new_types); if (unlikely(!__pyx_tuple__159)) __PYX_ERR(4, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__159);
  __Pyx_GIVEREF(__pyx_tuple__159);
  __pyx_codeobj__160 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__159, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_option_pxi, __pyx_n_s_with_external_types, 299, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__160)) __PYX_ERR(4, 299, __pyx_L1_error)

  /* "src/option.pxi":323
 *         return self.missing_values_can_be_empty_container or self.missing_values_can_be_null
 * 
 *     def equals(self, Options other: Options, ignore: Sequence[str]=()) -> bool:             # <<<<<<<<<<<<<<
 *         '''
 *         Compares two Options objects, this is equivalent to
 */
  __pyx_tuple__161 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_ignore, __pyx_n_s_field, __pyx_n_s_self_val, __pyx_n_s_other_val); if (unlikely(!__pyx_tuple__161)) __PYX_ERR(4, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__161);
  __Pyx_GIVEREF(__pyx_tuple__161);
  __pyx_codeobj__162 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__161, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_option_pxi, __pyx_n_s_equals, 323, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__162)) __PYX_ERR(4, 323, __pyx_L1_error)
  __pyx_tuple__163 = PyTuple_Pack(1, __pyx_empty_tuple); if (unlikely(!__pyx_tuple__163)) __PYX_ERR(4, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__163);
  __Pyx_GIVEREF(__pyx_tuple__163);

  /* "src/option.pxi":339
 *         return True
 * 
 *     def diff(self, Options other: Options, ignore: Sequence[str]=()) -> dict[str, tuple[object, object]]:             # <<<<<<<<<<<<<<
 *         '''
 *         Returns a dictionary containing just the fields that are different between two Options objects.
 */
  __pyx_tuple__164 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_ignore, __pyx_n_s_out, __pyx_n_s_field, __pyx_n_s_self_val, __pyx_n_s_other_val); if (unlikely(!__pyx_tuple__164)) __PYX_ERR(4, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__164);
  __Pyx_GIVEREF(__pyx_tuple__164);
  __pyx_codeobj__165 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__164, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_option_pxi, __pyx_n_s_diff, 339, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__165)) __PYX_ERR(4, 339, __pyx_L1_error)

  /* "(tree fragment)":65
 *     self.allow_invalid_default_values = allow_invalid_default_values
 *     self.externally_defined_types = externally_defined_types
 * __pyx_recursive_repr_guard = __import__('threading').local()             # <<<<<<<<<<<<<<
 * __pyx_recursive_repr_guard.running = set()
 * def __repr__(self):
 */
  __pyx_tuple__166 = PyTuple_Pack(1, __pyx_n_s_threading); if (unlikely(!__pyx_tuple__166)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__166);
  __Pyx_GIVEREF(__pyx_tuple__166);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_tuple__167 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_dict_2, __pyx_n_s_use_setstate); if (unlikely(!__pyx_tuple__167)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__167);
  __Pyx_GIVEREF(__pyx_tuple__167);
  __pyx_codeobj__168 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__168)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Options, (type(self), 0xc466972, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Options__set_state(self, __pyx_state)
 */
  __pyx_tuple__169 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__169)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__169);
  __Pyx_GIVEREF(__pyx_tuple__169);
  __pyx_codeobj__170 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__170)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__171 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__171)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Writer, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Writer__set_state(self, __pyx_state)
 */
  __pyx_codeobj__172 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__172)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__173 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__173)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Reader, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Reader__set_state(self, __pyx_state)
 */
  __pyx_codeobj__174 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__174)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__175 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__175)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FileReader, (type(self), 0x74c6201, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FileReader__set_state(self, __pyx_state)
 */
  __pyx_codeobj__176 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__176)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__177 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__177)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FileWriter, (type(self), 0x74c6201, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FileWriter__set_state(self, __pyx_state)
 */
  __pyx_codeobj__178 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__178)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__179 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__179)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__180 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__180)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__181 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__181)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MemoryReader, (type(self), 0xec0f6ac, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MemoryReader__set_state(self, __pyx_state)
 */
  __pyx_codeobj__182 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__182)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/rabin.pxi":44
 *         self.value = value
 * 
 *     cpdef update(self, bytes data):             # <<<<<<<<<<<<<<
 *         cdef int64_t swizzle
 *         cdef int64_t shifted
 */
  __pyx_tuple__183 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_data); if (unlikely(!__pyx_tuple__183)) __PYX_ERR(18, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__183);
  __Pyx_GIVEREF(__pyx_tuple__183);
  __pyx_codeobj__184 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__183, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_rabin_pxi, __pyx_n_s_update, 44, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__184)) __PYX_ERR(18, 44, __pyx_L1_error)

  /* "src/rabin.pxi":52
 *             self.value = shifted ^ swizzle
 * 
 *     def digest(self):             # <<<<<<<<<<<<<<
 *         cdef uint64_t value = self.value
 *         return int(value).to_bytes(8, 'little', signed=False)
 */
  __pyx_tuple__185 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_value); if (unlikely(!__pyx_tuple__185)) __PYX_ERR(18, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__185);
  __Pyx_GIVEREF(__pyx_tuple__185);
  __pyx_codeobj__186 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_rabin_pxi, __pyx_n_s_digest, 52, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__186)) __PYX_ERR(18, 52, __pyx_L1_error)

  /* "src/rabin.pxi":56
 *         return int(value).to_bytes(8, 'little', signed=False)
 * 
 *     def hexdigest(self):             # <<<<<<<<<<<<<<
 *         return self.digest().hex()
 * 
 */
  __pyx_codeobj__187 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_rabin_pxi, __pyx_n_s_hexdigest, 56, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__187)) __PYX_ERR(18, 56, __pyx_L1_error)

  /* "src/rabin.pxi":59
 *         return self.digest().hex()
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         new_ver = Rabin()
 *         new_ver.value = self.value
 */
  __pyx_tuple__188 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_new_ver); if (unlikely(!__pyx_tuple__188)) __PYX_ERR(18, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__188);
  __Pyx_GIVEREF(__pyx_tuple__188);
  __pyx_codeobj__189 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__188, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_rabin_pxi, __pyx_n_s_copy, 59, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__189)) __PYX_ERR(18, 59, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__190 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__190)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Rabin, (type(self), 0xcd42404, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Rabin__set_state(self, __pyx_state)
 */
  __pyx_codeobj__191 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__191)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/tests/test_zigzag.pxi":2
 * 
 * cpdef str _ubin(val, n_bytes):             # <<<<<<<<<<<<<<
 *     cdef str raw = bin(val)[2:]
 *     padded = raw.rjust(n_bytes * 8, '0')
 */
  __pyx_tuple__192 = PyTuple_Pack(2, __pyx_n_s_val, __pyx_n_s_n_bytes); if (unlikely(!__pyx_tuple__192)) __PYX_ERR(6, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__192);
  __Pyx_GIVEREF(__pyx_tuple__192);
  __pyx_codeobj__193 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__192, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_tests_test_zigzag_pxi, __pyx_n_s_ubin, 2, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__193)) __PYX_ERR(6, 2, __pyx_L1_error)

  /* "src/tests/test_zigzag.pxi":30
 *     return bytes(out)
 * 
 * @_tests             # <<<<<<<<<<<<<<
 * def _tests(add):
 * 
 */
  __pyx_tuple__194 = PyTuple_Pack(8, __pyx_n_s_add, __pyx_n_s_test_readvarint, __pyx_n_s_test_readvarint, __pyx_n_s_mask, __pyx_n_s_test_zigzag_long, __pyx_n_s_test_zigzag_long, __pyx_n_s_test_readvarlong, __pyx_n_s_test_readvarlong); if (unlikely(!__pyx_tuple__194)) __PYX_ERR(6, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__194);
  __Pyx_GIVEREF(__pyx_tuple__194);
  __pyx_codeobj__195 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__194, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_tests_test_zigzag_pxi, __pyx_n_s_tests_2, 30, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__195)) __PYX_ERR(6, 30, __pyx_L1_error)

  /* "src/tests/test_buffer.pxi":2
 * 
 * @_tests             # <<<<<<<<<<<<<<
 * def _tests(add):
 * 
 */
  __pyx_tuple__196 = PyTuple_Pack(5, __pyx_n_s_add, __pyx_n_s_test_reader_read_u8, __pyx_n_s_test_reader_read_u8, __pyx_n_s_test_reader_read_n, __pyx_n_s_test_reader_read_n); if (unlikely(!__pyx_tuple__196)) __PYX_ERR(19, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__196);
  __Pyx_GIVEREF(__pyx_tuple__196);
  __pyx_codeobj__197 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__196, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_tests_test_buffer_pxi, __pyx_n_s_tests_2, 2, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__197)) __PYX_ERR(19, 2, __pyx_L1_error)

  /* "src/tests/test_perf.pxi":2
 * 
 * @_perf             # <<<<<<<<<<<<<<
 * def _perf(add):
 *     def _test_zigzag_perf(const uint64_t mask):
 */
  __pyx_tuple__198 = PyTuple_Pack(4, __pyx_n_s_add, __pyx_n_s_test_zigzag_perf, __pyx_n_s_test_zigzag_perf, __pyx_n_s_mask); if (unlikely(!__pyx_tuple__198)) __PYX_ERR(7, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__198);
  __Pyx_GIVEREF(__pyx_tuple__198);
  __pyx_codeobj__199 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__198, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_tests_test_perf_pxi, __pyx_n_s_perf, 2, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__199)) __PYX_ERR(7, 2, __pyx_L1_error)

  /* "src/array.pxi":15
 *         self.item_type = AvroType.for_source(schema, source['items'], namespace)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef ArrayType new_inst = self.clone_base()
 *         new_inst.item_type = self.item_type
 */
  __pyx_codeobj__200 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_array_pxi, __pyx_n_s_copy, 15, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__200)) __PYX_ERR(20, 15, __pyx_L1_error)

  /* "src/array.pxi":23
 *         return _strip_keys(dict(source), {'type', 'items'})
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
  __pyx_tuple__201 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_visited); if (unlikely(!__pyx_tuple__201)) __PYX_ERR(20, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__201);
  __Pyx_GIVEREF(__pyx_tuple__201);
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__201, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_array_pxi, __pyx_n_s_walk_types, 23, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(20, 23, __pyx_L1_error)

  /* "src/array.pxi":29
 *         yield from self.item_type.walk_types(visited)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {'items': self.item_type.get_schema(created)}
 * 
 */
  __pyx_tuple__202 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_created); if (unlikely(!__pyx_tuple__202)) __PYX_ERR(20, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__202);
  __Pyx_GIVEREF(__pyx_tuple__202);
  __pyx_codeobj__203 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_array_pxi, __pyx_n_s_get_schema_extra, 29, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__203)) __PYX_ERR(20, 29, __pyx_L1_error)

  /* "src/array.pxi":92
 *         return out
 * 
 *     cpdef object convert_value(self, object value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef int item_fitness
 *         cdef AvroType item_type = self.item_type
 */
  __pyx_tuple__204 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_value, __pyx_n_s_check_value); if (unlikely(!__pyx_tuple__204)) __PYX_ERR(20, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__204);
  __Pyx_GIVEREF(__pyx_tuple__204);
  __pyx_codeobj__205 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__204, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_array_pxi, __pyx_n_s_convert_value_2, 92, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__205)) __PYX_ERR(20, 92, __pyx_L1_error)
  __pyx_tuple__206 = PyTuple_Pack(1, Py_True); if (unlikely(!__pyx_tuple__206)) __PYX_ERR(20, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__206);
  __Pyx_GIVEREF(__pyx_tuple__206);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__207 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__207)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ArrayType, (type(self), 0xcdf32d8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ArrayType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__208 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__208)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/enum.pxi":47
 *         self.doc = source.get('doc', '')
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef EnumType new_inst = self.clone_base()
 *         new_inst.symbols = self.symbols
 */
  __pyx_codeobj__209 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_enum_pxi, __pyx_n_s_copy, 47, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__209)) __PYX_ERR(8, 47, __pyx_L1_error)

  /* "src/enum.pxi":65
 *         })
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         extra = _NamedType._get_schema_extra(self, created)
 *         if self.default_value is not NO_DEFAULT:
 */
  __pyx_codeobj__210 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_enum_pxi, __pyx_n_s_get_schema_extra, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__210)) __PYX_ERR(8, 65, __pyx_L1_error)

  /* "src/enum.pxi":86
 *             return FIT_NONE
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         # if value is valid for type, then it needs no conversion
 *         return value
 */
  __pyx_codeobj__211 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_enum_pxi, __pyx_n_s_convert_value, 86, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__211)) __PYX_ERR(8, 86, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__212 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__212)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EnumType, (type(self), 0x8f63b31, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EnumType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__213 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__213)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__214 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__214)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromotingEnumType, (type(self), 0x8a61ff8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromotingEnumType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__215 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__215)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/map.pxi":14
 *         self.value_type = AvroType.for_source(schema, source['values'], namespace)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef MapType new_inst = self.clone_base()
 *         new_inst.key_type = self.key_type
 */
  __pyx_codeobj__216 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_map_pxi, __pyx_n_s_copy, 14, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__216)) __PYX_ERR(9, 14, __pyx_L1_error)

  /* "src/map.pxi":23
 *         return _strip_keys(dict(source), {'type', 'values'})
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__201, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_map_pxi, __pyx_n_s_walk_types, 23, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(9, 23, __pyx_L1_error)

  /* "src/map.pxi":29
 *         yield from self.value_type.walk_types(visited)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {'values': self.value_type.get_schema(created)}
 * 
 */
  __pyx_codeobj__217 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_map_pxi, __pyx_n_s_get_schema_extra, 29, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__217)) __PYX_ERR(9, 29, __pyx_L1_error)

  /* "src/map.pxi":122
 *         return out
 * 
 *     cpdef object convert_value(self, object orig_value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef int key_fitness
 *         cdef int value_fitness
 */
  __pyx_tuple__218 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_orig_value, __pyx_n_s_check_value); if (unlikely(!__pyx_tuple__218)) __PYX_ERR(9, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__218);
  __Pyx_GIVEREF(__pyx_tuple__218);
  __pyx_codeobj__219 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__218, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_map_pxi, __pyx_n_s_convert_value_2, 122, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__219)) __PYX_ERR(9, 122, __pyx_L1_error)
  __pyx_tuple__220 = PyTuple_Pack(1, Py_True); if (unlikely(!__pyx_tuple__220)) __PYX_ERR(9, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__220);
  __Pyx_GIVEREF(__pyx_tuple__220);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__221 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__221)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MapType, (type(self), 0xaa8230c, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MapType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__222 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__222)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/null.pxi":6
 *     type_name = "null"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_codeobj__223 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_null_pxi, __pyx_n_s_copy, 6, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__223)) __PYX_ERR(21, 6, __pyx_L1_error)

  /* "src/null.pxi":12
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_codeobj__224 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_null_pxi, __pyx_n_s_get_schema_extra, 12, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__224)) __PYX_ERR(21, 12, __pyx_L1_error)

  /* "src/null.pxi":40
 *         return None
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_codeobj__225 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_null_pxi, __pyx_n_s_convert_value, 40, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__225)) __PYX_ERR(21, 40, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__226 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__226)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NullType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NullType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__227 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__227)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/numeric_types.pxi":15
 *     type_name = "boolean"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_codeobj__228 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_copy, 15, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__228)) __PYX_ERR(10, 15, __pyx_L1_error)

  /* "src/numeric_types.pxi":21
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_codeobj__229 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_get_schema_extra, 21, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__229)) __PYX_ERR(10, 21, __pyx_L1_error)

  /* "src/numeric_types.pxi":42
 *         return FIT_NONE
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (py_bool, bool_)):
 *             return value
 */
  __pyx_codeobj__230 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_convert_value, 42, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__230)) __PYX_ERR(10, 42, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__231 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__231)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BoolType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BoolType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__232 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__232)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/numeric_types.pxi":65
 *     type_name = "int"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_codeobj__233 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_copy, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__233)) __PYX_ERR(10, 65, __pyx_L1_error)

  /* "src/numeric_types.pxi":71
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_codeobj__234 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_get_schema_extra, 71, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__234)) __PYX_ERR(10, 71, __pyx_L1_error)

  /* "src/numeric_types.pxi":100
 *         return max_fit
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')
 */
  __pyx_codeobj__235 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_convert_value, 100, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__235)) __PYX_ERR(10, 100, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__236 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__236)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__237 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__237)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/numeric_types.pxi":137
 *     type_name = "long"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_codeobj__238 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_copy, 137, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__238)) __PYX_ERR(10, 137, __pyx_L1_error)

  /* "src/numeric_types.pxi":143
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_codeobj__239 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_get_schema_extra, 143, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__239)) __PYX_ERR(10, 143, __pyx_L1_error)

  /* "src/numeric_types.pxi":172
 *         return max_fit
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')
 */
  __pyx_codeobj__240 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_convert_value, 172, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__240)) __PYX_ERR(10, 172, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__241 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__241)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LongType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LongType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__242 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__242)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/numeric_types.pxi":213
 *     type_name = "float"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_codeobj__243 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_copy, 213, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__243)) __PYX_ERR(10, 213, __pyx_L1_error)

  /* "src/numeric_types.pxi":219
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_codeobj__244 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_get_schema_extra, 219, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__244)) __PYX_ERR(10, 219, __pyx_L1_error)

  /* "src/numeric_types.pxi":256
 *         return FIT_NONE
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, float):
 *             pass
 */
  __pyx_codeobj__245 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_convert_value, 256, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__245)) __PYX_ERR(10, 256, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__246 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__246)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FloatType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FloatType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__247 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__247)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/numeric_types.pxi":328
 *     type_name = "double"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_codeobj__248 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_copy, 328, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__248)) __PYX_ERR(10, 328, __pyx_L1_error)

  /* "src/numeric_types.pxi":334
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_codeobj__249 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_get_schema_extra, 334, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__249)) __PYX_ERR(10, 334, __pyx_L1_error)

  /* "src/numeric_types.pxi":366
 *         return max_fit
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, float):
 *             pass
 */
  __pyx_codeobj__250 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_numeric_types_pxi, __pyx_n_s_convert_value, 366, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__250)) __PYX_ERR(10, 366, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__251 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__251)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DoubleType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DoubleType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__252 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__252)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/record.pxi":12
 * 
 * 
 * NO_DEFAULT = _Sentinel('NO_DEFAULT')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__253 = PyTuple_Pack(1, __pyx_n_u_NO_DEFAULT); if (unlikely(!__pyx_tuple__253)) __PYX_ERR(11, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__253);
  __Pyx_GIVEREF(__pyx_tuple__253);

  /* "src/record.pxi":34
 *         return {}
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_codeobj__254 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_record_pxi, __pyx_n_s_get_schema_extra, 34, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__254)) __PYX_ERR(11, 34, __pyx_L1_error)

  /* "src/record.pxi":55
 *         return self.default_value
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_codeobj__255 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_record_pxi, __pyx_n_s_convert_value, 55, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__255)) __PYX_ERR(11, 55, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__256 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__256)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__PlaceholderType, (type(self), 0x5c2152b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__PlaceholderType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__257 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__257)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/record.pxi":129
 *         self.data = _record_data_from_dict(self.Type, data_dict)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return ['Type'] + [f.name for f in self.Type.fields] + ['_asdict', '__getitem__']
 * 
 */
  __pyx_tuple__258 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_f); if (unlikely(!__pyx_tuple__258)) __PYX_ERR(11, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__258);
  __Pyx_GIVEREF(__pyx_tuple__258);
  __pyx_codeobj__259 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__258, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_record_pxi, __pyx_n_s_dir, 129, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__259)) __PYX_ERR(11, 129, __pyx_L1_error)

  /* "src/record.pxi":165
 *         return f'<Record:{self.Type.type} {self._repr_children(70)}>'
 * 
 *     def _asdict(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the record as a dict, with field names as keys.
 */
  __pyx_tuple__260 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_items, __pyx_n_s_field, __pyx_n_s_i, __pyx_n_s_data); if (unlikely(!__pyx_tuple__260)) __PYX_ERR(11, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__260);
  __Pyx_GIVEREF(__pyx_tuple__260);
  __pyx_codeobj__261 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__260, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_record_pxi, __pyx_n_s_asdict, 165, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__261)) __PYX_ERR(11, 165, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__262 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__262)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Record, (type(self), 0x3a6eb07, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Record__set_state(self, __pyx_state)
 */
  __pyx_codeobj__263 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__263)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/record.pxi":244
 *         })
 * 
 *     def get_schema(self, created):             # <<<<<<<<<<<<<<
 *         schema = {
 *             'name': self.name,
 */
  __pyx_tuple__264 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_created, __pyx_n_s_schema); if (unlikely(!__pyx_tuple__264)) __PYX_ERR(11, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__264);
  __Pyx_GIVEREF(__pyx_tuple__264);
  __pyx_codeobj__265 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__264, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_record_pxi, __pyx_n_s_get_schema, 244, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__265)) __PYX_ERR(11, 244, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__266 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__266)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RecordField, (type(self), 0x1d9a0f8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RecordField__set_state(self, __pyx_state)
 */
  __pyx_codeobj__267 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__267)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__268 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__268)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__FieldAccessor, (type(self), 0x1bc04b5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__FieldAccessor__set_state(self, __pyx_state)
 */
  __pyx_codeobj__269 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__269)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/record.pxi":348
 *         return AvroType._make_logical(self, schema, source)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef RecordType new_inst = self.clone_base()
 *         new_inst.doc = self.doc
 */
  __pyx_codeobj__270 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_record_pxi, __pyx_n_s_copy, 348, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__270)) __PYX_ERR(11, 348, __pyx_L1_error)

  /* "src/record.pxi":366
 *         })
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
  __pyx_tuple__271 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_visited, __pyx_n_s_field); if (unlikely(!__pyx_tuple__271)) __PYX_ERR(11, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__271);
  __Pyx_GIVEREF(__pyx_tuple__271);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__271, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_record_pxi, __pyx_n_s_walk_types, 366, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(11, 366, __pyx_L1_error)

  /* "src/record.pxi":373
 *             yield from field.type.walk_types(visited)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         extra = super(RecordType, self)._get_schema_extra(created)
 *         return dict(extra, fields=[f.get_schema(created) for f in self.fields])
 */
  __pyx_codeobj__272 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_record_pxi, __pyx_n_s_get_schema_extra, 373, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__272)) __PYX_ERR(11, 373, __pyx_L1_error)

  /* "src/record.pxi":528
 *         return self.json_decode_record(value)
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, self.record):
 *             return value
 */
  __pyx_codeobj__273 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_record_pxi, __pyx_n_s_convert_value, 528, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__273)) __PYX_ERR(11, 528, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__274 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__274)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RecordType, (type(self), 0x182acf2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RecordType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__275 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__275)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__276 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__276)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromotingRecordType, (type(self), 0x40bba2b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromotingRecordType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__277 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__277)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/string_types.pxi":7
 *     type_name = "bytes"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_codeobj__278 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_string_types_pxi, __pyx_n_s_copy, 7, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__278)) __PYX_ERR(12, 7, __pyx_L1_error)

  /* "src/string_types.pxi":13
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_codeobj__279 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_string_types_pxi, __pyx_n_s_get_schema_extra, 13, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__279)) __PYX_ERR(12, 13, __pyx_L1_error)

  /* "src/string_types.pxi":57
 *         return sval.encode('latin-1')
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (bytes, bytearray)):
 *             return value
 */
  __pyx_codeobj__280 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_string_types_pxi, __pyx_n_s_convert_value, 57, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__280)) __PYX_ERR(12, 57, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__281 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__281)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BytesType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BytesType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__282 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__282)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/string_types.pxi":97
 *     type_name = "string"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_codeobj__283 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_string_types_pxi, __pyx_n_s_copy, 97, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__283)) __PYX_ERR(12, 97, __pyx_L1_error)

  /* "src/string_types.pxi":103
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_codeobj__284 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_string_types_pxi, __pyx_n_s_get_schema_extra, 103, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__284)) __PYX_ERR(12, 103, __pyx_L1_error)

  /* "src/string_types.pxi":133
 *         return sval
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, str):
 *             return value
 */
  __pyx_codeobj__285 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_string_types_pxi, __pyx_n_s_convert_value, 133, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__285)) __PYX_ERR(12, 133, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__286 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__286)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_StringType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_StringType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__287 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__287)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/string_types.pxi":163
 *         super().__init__(schema, source, namespace)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef FixedType new_inst =  self.clone_base()
 *         new_inst.size = self.size
 */
  __pyx_codeobj__288 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_string_types_pxi, __pyx_n_s_copy, 163, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__288)) __PYX_ERR(12, 163, __pyx_L1_error)

  /* "src/string_types.pxi":171
 *         return _strip_keys(dict(source), {'type', 'name', 'namespace', 'aliases', 'size'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return dict(_NamedType._get_schema_extra(self, created), size=self.size)
 * 
 */
  __pyx_codeobj__289 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_string_types_pxi, __pyx_n_s_get_schema_extra, 171, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__289)) __PYX_ERR(12, 171, __pyx_L1_error)

  /* "src/string_types.pxi":228
 *         return encoded
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (bytes, bytearray)):
 *             # Let's see if we can recover this
 */
  __pyx_codeobj__290 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_string_types_pxi, __pyx_n_s_convert_value, 228, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__290)) __PYX_ERR(12, 228, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__291 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__291)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FixedType, (type(self), 0xd9f3afe, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FixedType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__292 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__292)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/union.pxi":77
 *                     seen_types.add(member_type)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef UnionType new_inst = self.clone_base()
 *         new_inst.union_types = self.union_types
 */
  __pyx_codeobj__293 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_union_pxi, __pyx_n_s_copy, 77, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__293)) __PYX_ERR(22, 77, __pyx_L1_error)

  /* "src/union.pxi":90
 *         return dict()
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
  __pyx_tuple__294 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_visited, __pyx_n_s_t); if (unlikely(!__pyx_tuple__294)) __PYX_ERR(22, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__294);
  __Pyx_GIVEREF(__pyx_tuple__294);
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__294, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_union_pxi, __pyx_n_s_walk_types, 90, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(22, 90, __pyx_L1_error)

  /* "src/union.pxi":97
 *             yield from t.walk_types(visited)
 * 
 *     cpdef get_schema(self, created=None):             # <<<<<<<<<<<<<<
 *         if created is None:
 *             created = set()
 */
  __pyx_codeobj__295 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_union_pxi, __pyx_n_s_get_schema, 97, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__295)) __PYX_ERR(22, 97, __pyx_L1_error)

  /* "src/union.pxi":170
 *         return item_type.json_decode(item_value)
 * 
 *     cpdef object convert_value(self, object value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t index = self.resolve_from_value(value)
 *         return self.union_types[index].convert_value(value)
 */
  __pyx_codeobj__296 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__204, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_union_pxi, __pyx_n_s_convert_value_2, 170, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__296)) __PYX_ERR(22, 170, __pyx_L1_error)
  __pyx_tuple__297 = PyTuple_Pack(1, Py_True); if (unlikely(!__pyx_tuple__297)) __PYX_ERR(22, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__297);
  __Pyx_GIVEREF(__pyx_tuple__297);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__298 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__298)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UnionType, (type(self), 0xabef18e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UnionType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__299 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__299)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/type.pxi":8
 * 
 * 
 * CANONICAL_FORM_KEYS = ('name', 'type', 'fields', 'symbols', 'items', 'values', 'size')             # <<<<<<<<<<<<<<
 * 
 * cdef class _Sentinel:
 */
  __pyx_tuple__300 = PyTuple_Pack(7, __pyx_n_u_name, __pyx_n_u_type, __pyx_n_u_fields, __pyx_n_u_symbols, __pyx_n_u_items, __pyx_n_u_values, __pyx_n_u_size); if (unlikely(!__pyx_tuple__300)) __PYX_ERR(2, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__300);
  __Pyx_GIVEREF(__pyx_tuple__300);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__301 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__301)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Sentinel, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Sentinel__set_state(self, __pyx_state)
 */
  __pyx_codeobj__302 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__302)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/type.pxi":19
 *         return f'<{self.name}: {id(self)}>'
 * 
 * MISSING_VALUE = _Sentinel('MISSING_VALUE')             # <<<<<<<<<<<<<<
 * 
 * cdef class _CanonicalForm(str):
 */
  __pyx_tuple__303 = PyTuple_Pack(1, __pyx_n_u_MISSING_VALUE); if (unlikely(!__pyx_tuple__303)) __PYX_ERR(2, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__303);
  __Pyx_GIVEREF(__pyx_tuple__303);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__304 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__304)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__CanonicalForm, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__CanonicalForm__set_state(self, __pyx_state)
 */
  __pyx_codeobj__305 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__305)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/type.pxi":63
 *     cdef readonly tuple[ValueAdapter] value_adapters
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def for_source(cls, schema, source, namespace=None):
 *         cdef Schema schema_ob = schema
 */
  __pyx_tuple__306 = PyTuple_Pack(9, __pyx_n_s_cls, __pyx_n_s_schema, __pyx_n_s_source, __pyx_n_s_namespace, __pyx_n_s_schema_ob, __pyx_n_s_type_name, __pyx_n_s_e, __pyx_n_s_inst, __pyx_n_s_namespaced); if (unlikely(!__pyx_tuple__306)) __PYX_ERR(2, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__306);
  __Pyx_GIVEREF(__pyx_tuple__306);
  __pyx_codeobj__307 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__306, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_type_pxi, __pyx_n_s_for_source, 63, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__307)) __PYX_ERR(2, 63, __pyx_L1_error)

  /* "src/type.pxi":84
 *         raise UnknownType(namespaced)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def for_schema(cls, schema):
 *         return AvroType.for_source(schema, schema.source)
 */
  __pyx_tuple__308 = PyTuple_Pack(2, __pyx_n_s_cls, __pyx_n_s_schema); if (unlikely(!__pyx_tuple__308)) __PYX_ERR(2, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__308);
  __Pyx_GIVEREF(__pyx_tuple__308);
  __pyx_codeobj__309 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__308, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_type_pxi, __pyx_n_s_for_schema, 84, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__309)) __PYX_ERR(2, 84, __pyx_L1_error)

  /* "src/type.pxi":108
 *         return clone
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement clone")
 */
  __pyx_codeobj__310 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_type_pxi, __pyx_n_s_copy, 108, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__310)) __PYX_ERR(2, 108, __pyx_L1_error)

  /* "src/type.pxi":130
 *             f"{type(self).__name__} does not implement _extract_metadata")
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _convert_value")
 */
  __pyx_codeobj__311 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_type_pxi, __pyx_n_s_convert_value, 130, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__311)) __PYX_ERR(2, 130, __pyx_L1_error)

  /* "src/type.pxi":134
 *             f"{type(self).__name__} does not implement _convert_value")
 * 
 *     cpdef object convert_value(self, object value, check_value=False):             # <<<<<<<<<<<<<<
 *         if check_value:
 *             self.assert_value(value)
 */
  __pyx_codeobj__312 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__204, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_type_pxi, __pyx_n_s_convert_value_2, 134, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__312)) __PYX_ERR(2, 134, __pyx_L1_error)
  __pyx_tuple__313 = PyTuple_Pack(1, Py_False); if (unlikely(!__pyx_tuple__313)) __PYX_ERR(2, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__313);
  __Pyx_GIVEREF(__pyx_tuple__313);

  /* "src/type.pxi":232
 *             f"{type(self).__name__} does not implement canonical_form")
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _get_schema_extra")
 */
  __pyx_codeobj__314 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_type_pxi, __pyx_n_s_get_schema_extra, 232, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__314)) __PYX_ERR(2, 232, __pyx_L1_error)

  /* "src/type.pxi":286
 *             raise TypeError(f"Default value {schema_default!r} is not valid for field: {field}") from e
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
  __pyx_codeobj__82 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_GENERATOR, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__201, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_type_pxi, __pyx_n_s_walk_types, 286, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__82)) __PYX_ERR(2, 286, __pyx_L1_error)

  /* "src/type.pxi":292
 *         yield self
 * 
 *     cpdef get_schema(self, created=None):             # <<<<<<<<<<<<<<
 *         if created is None:
 *             created = set()
 */
  __pyx_codeobj__315 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_type_pxi, __pyx_n_s_get_schema, 292, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__315)) __PYX_ERR(2, 292, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__316 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__316)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_AvroType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_AvroType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__317 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__317)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/type.pxi":396
 *         return frozenset(resolve_namespaced_name(self.effective_namespace, alias) for alias in all_names)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         schema = {}
 *         if self.effective_namespace:
 */
  __pyx_codeobj__318 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_type_pxi, __pyx_n_s_get_schema_extra, 396, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__318)) __PYX_ERR(2, 396, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__319 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__319)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__NamedType, (type(self), 0x24d7e16, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__NamedType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__320 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__320)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/schema.pxi":11
 * 
 * class _class_inst_method:
 *     def __init__(self, func):             # <<<<<<<<<<<<<<
 *         self.func = func
 * 
 */
  __pyx_tuple__321 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_func); if (unlikely(!__pyx_tuple__321)) __PYX_ERR(23, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__321);
  __Pyx_GIVEREF(__pyx_tuple__321);
  __pyx_codeobj__322 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__321, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_init, 11, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__322)) __PYX_ERR(23, 11, __pyx_L1_error)

  /* "src/schema.pxi":14
 *         self.func = func
 * 
 *     def __get__(self, inst, cls):             # <<<<<<<<<<<<<<
 *         return partial(self.func, inst, cls)
 * 
 */
  __pyx_tuple__323 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_inst, __pyx_n_s_cls); if (unlikely(!__pyx_tuple__323)) __PYX_ERR(23, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__323);
  __Pyx_GIVEREF(__pyx_tuple__323);
  __pyx_codeobj__324 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__323, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_get_2, 14, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__324)) __PYX_ERR(23, 14, __pyx_L1_error)

  /* "src/schema.pxi":59
 *         self.type = AvroType.for_schema(self) if _type is None else _type
 * 
 *     @_class_inst_method             # <<<<<<<<<<<<<<
 *     def _wrap_type(inst, cls, AvroType avro_type, Options options=None):
 *         """
 */
  __pyx_tuple__325 = PyTuple_Pack(6, __pyx_n_s_inst, __pyx_n_s_cls, __pyx_n_s_avro_type, __pyx_n_s_options, __pyx_n_s_source, __pyx_n_s_new_inst); if (unlikely(!__pyx_tuple__325)) __PYX_ERR(23, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__325);
  __Pyx_GIVEREF(__pyx_tuple__325);
  __pyx_codeobj__326 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_wrap_type, 59, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__326)) __PYX_ERR(23, 59, __pyx_L1_error)

  /* "src/schema.pxi":74
 *         return new_inst
 * 
 *     def _make_logical_types(self, options):             # <<<<<<<<<<<<<<
 *         logical_by_name = {}
 *         self.logical_types = logical_by_name
 */
  __pyx_tuple__327 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_options, __pyx_n_s_logical_by_name, __pyx_n_s_logical_type, __pyx_n_s_type_name, __pyx_n_s_dest); if (unlikely(!__pyx_tuple__327)) __PYX_ERR(23, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__327);
  __Pyx_GIVEREF(__pyx_tuple__327);
  __pyx_codeobj__328 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__327, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_make_logical_types, 74, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__328)) __PYX_ERR(23, 74, __pyx_L1_error)

  /* "src/schema.pxi":118
 *             return json.dumps(self.schema, indent=2)
 * 
 *     def fingerprint(self, method='rabin', **kwargs) -> Union[bytes, hashlib._hashlib.HASH]:             # <<<<<<<<<<<<<<
 *         """
 *         Return the deterministic fingerprint of the schema, using the given hash method.
 */
  __pyx_tuple__329 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_method, __pyx_n_s_kwargs, __pyx_n_s_hasher); if (unlikely(!__pyx_tuple__329)) __PYX_ERR(23, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__329);
  __Pyx_GIVEREF(__pyx_tuple__329);
  __pyx_codeobj__330 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__329, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_fingerprint, 118, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__330)) __PYX_ERR(23, 118, __pyx_L1_error)
  __pyx_tuple__331 = PyTuple_Pack(1, __pyx_n_u_rabin); if (unlikely(!__pyx_tuple__331)) __PYX_ERR(23, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__331);
  __Pyx_GIVEREF(__pyx_tuple__331);

  /* "src/schema.pxi":138
 *         return hasher
 * 
 *     cpdef AvroType find_type(self, str namespace, str name, bint _raise=True):             # <<<<<<<<<<<<<<
 *         """
 *         Given a namespace and name (namespace may be None), find and return the `AvroType` instance matching this name.
 */
  __pyx_tuple__332 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_namespace, __pyx_n_s_name, __pyx_n_s_raise); if (unlikely(!__pyx_tuple__332)) __PYX_ERR(23, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__332);
  __Pyx_GIVEREF(__pyx_tuple__332);
  __pyx_codeobj__333 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__332, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_find_type, 138, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__333)) __PYX_ERR(23, 138, __pyx_L1_error)
  __pyx_tuple__334 = PyTuple_Pack(1, Py_True); if (unlikely(!__pyx_tuple__334)) __PYX_ERR(23, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__334);
  __Pyx_GIVEREF(__pyx_tuple__334);

  /* "src/schema.pxi":159
 *             raise UnknownType(f'Unknown type: {resolved!r}')
 * 
 *     def can_encode(self, value: object) -> bool:             # <<<<<<<<<<<<<<
 *         """
 *         Check if `value` can be encoded using this schema
 */
  __pyx_tuple__335 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_value, __pyx_n_s_fitness); if (unlikely(!__pyx_tuple__335)) __PYX_ERR(23, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__335);
  __Pyx_GIVEREF(__pyx_tuple__335);
  __pyx_codeobj__336 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__335, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_can_encode, 159, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__336)) __PYX_ERR(23, 159, __pyx_L1_error)

  /* "src/schema.pxi":166
 *         return fitness > FIT_NONE
 * 
 *     def binary_encode(self, value: object) -> bytes:             # <<<<<<<<<<<<<<
 *         """
 *         Encode `value` using this schema and return the avro bytes representing it.
 */
  __pyx_tuple__337 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_value, __pyx_n_s_buffer); if (unlikely(!__pyx_tuple__337)) __PYX_ERR(23, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__337);
  __Pyx_GIVEREF(__pyx_tuple__337);
  __pyx_codeobj__338 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__337, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_binary_encode, 166, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__338)) __PYX_ERR(23, 166, __pyx_L1_error)

  /* "src/schema.pxi":174
 *         return buffer.bytes()
 * 
 *     def binary_decode(self, bytes value: bytes) -> object:             # <<<<<<<<<<<<<<
 *         """
 *         Decode `value` using this schema and return the decoded value.
 */
  __pyx_codeobj__339 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__337, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_binary_decode, 174, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__339)) __PYX_ERR(23, 174, __pyx_L1_error)

  /* "src/schema.pxi":181
 *         return self.type.binary_buffer_decode(buffer)
 * 
 *     cpdef binary_read(self, _Reader reader: _Reader):             # <<<<<<<<<<<<<<
 *         """
 *         Read a value from `reader` using this schema and return the decoded value.
 */
  __pyx_tuple__340 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_reader); if (unlikely(!__pyx_tuple__340)) __PYX_ERR(23, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__340);
  __Pyx_GIVEREF(__pyx_tuple__340);
  __pyx_codeobj__341 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__340, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_binary_read, 181, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__341)) __PYX_ERR(23, 181, __pyx_L1_error)

  /* "src/schema.pxi":188
 *         return self.type.binary_buffer_decode(reader)
 * 
 *     cpdef binary_write(self, _Writer writer: _Writer, value: object):             # <<<<<<<<<<<<<<
 *         """
 *         Write `value` to `writer` using this schema.
 */
  __pyx_tuple__342 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_writer, __pyx_n_s_value); if (unlikely(!__pyx_tuple__342)) __PYX_ERR(23, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__342);
  __Pyx_GIVEREF(__pyx_tuple__342);
  __pyx_codeobj__343 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__342, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_binary_write, 188, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__343)) __PYX_ERR(23, 188, __pyx_L1_error)

  /* "src/schema.pxi":195
 *         self.type.binary_buffer_encode(writer, value)
 * 
 *     def json_encode(self, value, serialize=True, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Encode `value` using this schema and return the avro json representing it.
 */
  __pyx_tuple__344 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_value, __pyx_n_s_serialize, __pyx_n_s_kwargs, __pyx_n_s_data); if (unlikely(!__pyx_tuple__344)) __PYX_ERR(23, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__344);
  __Pyx_GIVEREF(__pyx_tuple__344);
  __pyx_codeobj__345 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__344, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_json_encode, 195, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__345)) __PYX_ERR(23, 195, __pyx_L1_error)
  __pyx_tuple__346 = PyTuple_Pack(1, Py_True); if (unlikely(!__pyx_tuple__346)) __PYX_ERR(23, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__346);
  __Pyx_GIVEREF(__pyx_tuple__346);

  /* "src/schema.pxi":204
 *         return data
 * 
 *     def json_decode(self, value: Union[str, object], deserialize: bool=True, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Decode `value` in JSON form using this schema and return the decoded value.
 */
  __pyx_tuple__347 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_value, __pyx_n_s_deserialize, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__347)) __PYX_ERR(23, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__347);
  __Pyx_GIVEREF(__pyx_tuple__347);
  __pyx_codeobj__348 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__347, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_json_decode, 204, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__348)) __PYX_ERR(23, 204, __pyx_L1_error)

  /* "src/schema.pxi":215
 *         return self.type.json_decode(value)
 * 
 *     cpdef Schema reader_for_writer(self, Schema writer_schema: Schema):             # <<<<<<<<<<<<<<
 *         """
 *         Return a schema that is the result of promoting this schema to the writer schema.
 */
  __pyx_tuple__349 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_writer_schema); if (unlikely(!__pyx_tuple__349)) __PYX_ERR(23, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__349);
  __Pyx_GIVEREF(__pyx_tuple__349);
  __pyx_codeobj__350 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__349, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_schema_pxi, __pyx_n_s_reader_for_writer, 215, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__350)) __PYX_ERR(23, 215, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__351 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__351)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Schema, (type(self), 0xc2592e1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Schema__set_state(self, __pyx_state)
 */
  __pyx_codeobj__352 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__352)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__353 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__353)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ResolvedSchema, (type(self), 0xc2592e1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ResolvedSchema__set_state(self, __pyx_state)
 */
  __pyx_codeobj__354 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__354)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__355 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__355)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ValueAdapter, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ValueAdapter__set_state(self, __pyx_state)
 */
  __pyx_codeobj__356 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__356)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/logical.pxi":34
 *     underlying_types = NotImplemented
 * 
 *     @_class_inst_method             # <<<<<<<<<<<<<<
 *     def for_type(inst, cls, underlying: AvroType):
 *         cdef object ob = inst
 */
  __pyx_tuple__357 = PyTuple_Pack(5, __pyx_n_s_inst, __pyx_n_s_cls, __pyx_n_s_underlying, __pyx_n_s_ob, __pyx_n_s_underlying_type); if (unlikely(!__pyx_tuple__357)) __PYX_ERR(24, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__357);
  __Pyx_GIVEREF(__pyx_tuple__357);
  __pyx_codeobj__358 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__357, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_logical_pxi, __pyx_n_s_for_type_2, 34, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__358)) __PYX_ERR(24, 34, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__359 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__359)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LogicalType, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LogicalType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__360 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__360)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__361 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__361)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CustomLogicalType, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CustomLogicalType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__362 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__362)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/logical.pxi":100
 *         return f'{self.underlying_name}(decimal)'
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         meta = underlying.metadata
 */
  __pyx_tuple__363 = PyTuple_Pack(7, __pyx_n_s_cls, __pyx_n_s_underlying, __pyx_n_s_meta, __pyx_n_s_precision, __pyx_n_s_scale, __pyx_n_s_size, __pyx_n_s_max_precision); if (unlikely(!__pyx_tuple__363)) __PYX_ERR(24, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__363);
  __Pyx_GIVEREF(__pyx_tuple__363);
  __pyx_codeobj__364 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__363, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_logical_pxi, __pyx_n_s_for_type_3, 100, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__364)) __PYX_ERR(24, 100, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__365 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__365)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DecimalType, (type(self), 0xe9db223, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DecimalType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__366 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__366)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/logical.pxi":175
 *         self.return_uuid_object = return_uuid_object
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.return_uuid_object)
 */
  __pyx_tuple__367 = PyTuple_Pack(2, __pyx_n_s_cls, __pyx_n_s_underlying); if (unlikely(!__pyx_tuple__367)) __PYX_ERR(24, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__367);
  __Pyx_GIVEREF(__pyx_tuple__367);
  __pyx_codeobj__368 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__367, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_logical_pxi, __pyx_n_s_for_type_3, 175, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__368)) __PYX_ERR(24, 175, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__369 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__369)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UUIDBase, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UUIDBase__set_state(self, __pyx_state)
 */
  __pyx_codeobj__370 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__370)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/logical.pxi":183
 *     underlying_types = (StringType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.return_uuid_object)
 */
  __pyx_codeobj__371 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__367, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_logical_pxi, __pyx_n_s_for_type_3, 183, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__371)) __PYX_ERR(24, 183, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__372 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__372)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UUIDStringType, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UUIDStringType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__373 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__373)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/logical.pxi":212
 *     underlying_types = (FixedType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         if underlying.size != 16:
 */
  __pyx_codeobj__374 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__367, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_logical_pxi, __pyx_n_s_for_type_3, 212, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__374)) __PYX_ERR(24, 212, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__375 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__375)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UUIDFixedType, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UUIDFixedType__set_state(self, __pyx_state)
 */
  __pyx_codeobj__376 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__376)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/logical.pxi":233
 * 
 * 
 * EPOCH_DATE = datetime.date(1970, 1, 1)             # <<<<<<<<<<<<<<
 * EPOCH_DT = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
 * EPOCH_DT_NO_TZ = datetime.datetime(1970, 1, 1)
 */
  __pyx_tuple__377 = PyTuple_Pack(3, __pyx_int_1970, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__377)) __PYX_ERR(24, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__377);
  __Pyx_GIVEREF(__pyx_tuple__377);

  /* "src/logical.pxi":247
 *         self.accepts_string = accepts_string
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.date_type_accepts_string)
 */
  __pyx_codeobj__378 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__367, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_logical_pxi, __pyx_n_s_for_type_3, 247, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__378)) __PYX_ERR(24, 247, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__379 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__379)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Date, (type(self), 0xc5331ff, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Date__set_state(self, __pyx_state)
 */
  __pyx_codeobj__380 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__380)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/logical.pxi":270
 * 
 * 
 * def _time_to_micros(value):             # <<<<<<<<<<<<<<
 *     return value.microsecond + 1_000_000 * (value.second + 60 * (value.minute + 60 * value.hour))
 * 
 */
  __pyx_tuple__381 = PyTuple_Pack(1, __pyx_n_s_value); if (unlikely(!__pyx_tuple__381)) __PYX_ERR(24, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__381);
  __Pyx_GIVEREF(__pyx_tuple__381);
  __pyx_codeobj__382 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__381, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_logical_pxi, __pyx_n_s_time_to_micros, 270, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__382)) __PYX_ERR(24, 270, __pyx_L1_error)

  /* "src/logical.pxi":274
 * 
 * 
 * def _micros_to_time(value):             # <<<<<<<<<<<<<<
 *     return datetime.time(
 *         hour=value // 3_600_000_000,
 */
  __pyx_codeobj__383 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__381, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_logical_pxi, __pyx_n_s_micros_to_time, 274, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__383)) __PYX_ERR(24, 274, __pyx_L1_error)

  /* "src/logical.pxi":297
 *     underlying_types = (IntType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()
 */
  __pyx_codeobj__384 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__367, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_logical_pxi, __pyx_n_s_for_type_3, 297, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__384)) __PYX_ERR(24, 297, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__385 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__385)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimeMillis, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimeMillis__set_state(self, __pyx_state)
 */
  __pyx_codeobj__386 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__386)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/logical.pxi":321
 *     underlying_types = (LongType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()
 */
  __pyx_codeobj__387 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__367, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_logical_pxi, __pyx_n_s_for_type_3, 321, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__387)) __PYX_ERR(24, 321, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__388 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__388)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimeMicros, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimeMicros__set_state(self, __pyx_state)
 */
  __pyx_codeobj__389 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__389)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/logical.pxi":350
 *         self.alternate_timestamp_encoding = alternate_timestamp_encoding
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(
 */
  __pyx_codeobj__390 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__367, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_logical_pxi, __pyx_n_s_for_type_3, 350, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__390)) __PYX_ERR(24, 350, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__391 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__391)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimestampMillis, (type(self), 0x8d5b257, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimestampMillis__set_state(self, __pyx_state)
 */
  __pyx_codeobj__392 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__392)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/logical.pxi":380
 *     underlying_types = (LongType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()
 */
  __pyx_codeobj__393 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__367, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_logical_pxi, __pyx_n_s_for_type_3, 380, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__393)) __PYX_ERR(24, 380, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__394 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__394)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimestampMicros, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimestampMicros__set_state(self, __pyx_state)
 */
  __pyx_codeobj__395 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__395)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/codec.pxi":44
 * HAVE_LZ4 = False
 * try:
 *     import lz4.frame             # <<<<<<<<<<<<<<
 *     HAVE_LZ4 = True
 * except ImportError:
 */
  __pyx_tuple__396 = PyTuple_Pack(2, __pyx_n_s_lz4, __pyx_n_s_frame); if (unlikely(!__pyx_tuple__396)) __PYX_ERR(3, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__396);
  __Pyx_GIVEREF(__pyx_tuple__396);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__397 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__397)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Codec__set_state(self, __pyx_state)
 */
  __pyx_codeobj__398 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__398)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__399 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__399)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__SnappyCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__SnappyCodec__set_state(self, __pyx_state)
 */
  __pyx_codeobj__400 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__400)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__401 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__401)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__NullCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__NullCodec__set_state(self, __pyx_state)
 */
  __pyx_codeobj__402 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__402)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__403 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__403)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__DeflateCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__DeflateCodec__set_state(self, __pyx_state)
 */
  __pyx_codeobj__404 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__404)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__405 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__405)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Bzip2Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Bzip2Codec__set_state(self, __pyx_state)
 */
  __pyx_codeobj__406 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__406)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__407 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__407)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__LzmaCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__LzmaCodec__set_state(self, __pyx_state)
 */
  __pyx_codeobj__408 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__408)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__409 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__409)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__ZStandardCodec, (type(self), 0x8d57cb2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__ZStandardCodec__set_state(self, __pyx_state)
 */
  __pyx_codeobj__410 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__410)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__411 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__411)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Lz4Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Lz4Codec__set_state(self, __pyx_state)
 */
  __pyx_codeobj__412 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__412)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/container.pxi":95
 *         self.marker = b''
 * 
 *     cpdef _read_marker(self):             # <<<<<<<<<<<<<<
 *         value = self.reader.read_n(16)
 *         if not len(self.marker):
 */
  __pyx_codeobj__413 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_container_pxi, __pyx_n_s_read_marker, 95, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__413)) __PYX_ERR(25, 95, __pyx_L1_error)

  /* "src/container.pxi":117
 *         self.current_block._reset_to(block_bytes)
 * 
 *     cpdef object next_object(self):             # <<<<<<<<<<<<<<
 *         while self.objects_left_in_block < 1:
 *             self.next_block()
 */
  __pyx_codeobj__414 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_container_pxi, __pyx_n_s_next_object, 117, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__414)) __PYX_ERR(25, 117, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__415 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__415)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ContainerReader, (type(self), 0x289807d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ContainerReader__set_state(self, __pyx_state)
 */
  __pyx_codeobj__416 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__416)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "src/container.pxi":207
 *             raise
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         if self.writer is None:
 *             raise ValueError('Container is closed')
 */
  __pyx_codeobj__417 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_container_pxi, __pyx_n_s_enter, 207, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__417)) __PYX_ERR(25, 207, __pyx_L1_error)

  /* "src/container.pxi":212
 *         return self
 * 
 *     def __exit__(self, exc_type, exc_val, exc_tb):             # <<<<<<<<<<<<<<
 *         if self.writer is not None:
 *             self.close()
 */
  __pyx_tuple__418 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_exc_type, __pyx_n_s_exc_val, __pyx_n_s_exc_tb); if (unlikely(!__pyx_tuple__418)) __PYX_ERR(25, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__418);
  __Pyx_GIVEREF(__pyx_tuple__418);
  __pyx_codeobj__419 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__418, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_container_pxi, __pyx_n_s_exit, 212, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__419)) __PYX_ERR(25, 212, __pyx_L1_error)

  /* "src/container.pxi":238
 *         return 0
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         if self.writer is None:
 *             raise ValueError('Trying to close a closed Container')
 */
  __pyx_tuple__420 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_writer); if (unlikely(!__pyx_tuple__420)) __PYX_ERR(25, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__420);
  __Pyx_GIVEREF(__pyx_tuple__420);
  __pyx_codeobj__421 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__420, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_container_pxi, __pyx_n_s_close, 238, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__421)) __PYX_ERR(25, 238, __pyx_L1_error)

  /* "src/container.pxi":264
 *         return 0
 * 
 *     def flush(self, force=False):             # <<<<<<<<<<<<<<
 *         self._flush_block(force)
 * 
 */
  __pyx_tuple__422 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_force); if (unlikely(!__pyx_tuple__422)) __PYX_ERR(25, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__422);
  __Pyx_GIVEREF(__pyx_tuple__422);
  __pyx_codeobj__423 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__422, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_container_pxi, __pyx_n_s_flush, 264, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__423)) __PYX_ERR(25, 264, __pyx_L1_error)
  __pyx_tuple__424 = PyTuple_Pack(1, Py_False); if (unlikely(!__pyx_tuple__424)) __PYX_ERR(25, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__424);
  __Pyx_GIVEREF(__pyx_tuple__424);

  /* "src/container.pxi":267
 *         self._flush_block(force)
 * 
 *     cpdef int write_one(self, obj) except -1:             # <<<<<<<<<<<<<<
 *         if self.next_item is None:
 *             raise ValueError('Trying to write to closed Container')
 */
  __pyx_tuple__425 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__425)) __PYX_ERR(25, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__425);
  __Pyx_GIVEREF(__pyx_tuple__425);
  __pyx_codeobj__426 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__425, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_container_pxi, __pyx_n_s_write_one, 267, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__426)) __PYX_ERR(25, 267, __pyx_L1_error)

  /* "src/container.pxi":284
 *                 self._flush_block()
 * 
 *     def write_many(self, objs):             # <<<<<<<<<<<<<<
 *         for obj in objs:
 *             self.write_one(obj)
 */
  __pyx_tuple__427 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_objs, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__427)) __PYX_ERR(25, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__427);
  __Pyx_GIVEREF(__pyx_tuple__427);
  __pyx_codeobj__428 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__427, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_container_pxi, __pyx_n_s_write_many, 284, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__428)) __PYX_ERR(25, 284, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__429 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__429)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__430 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__430)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__431 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__431)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromoteToFloat, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromoteToFloat__set_state(self, __pyx_state)
 */
  __pyx_codeobj__432 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__432)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__433 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__433)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromoteBytesToString, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromoteBytesToString__set_state(self, __pyx_state)
 */
  __pyx_codeobj__434 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__434)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__435 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__435)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromoteStringToBytes, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromoteStringToBytes__set_state(self, __pyx_state)
 */
  __pyx_codeobj__436 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__436)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__437 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__437)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CannotPromote, (type(self), 0xcbfa7ba, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CannotPromote__set_state(self, __pyx_state)
 */
  __pyx_codeobj__438 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__438)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_Options(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_codeobj__439 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Options, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__439)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__440 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__Writer, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__440)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__441 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__Reader, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__441)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__442 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_FileReader, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__442)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__443 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_FileWriter, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__443)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__444 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_MemoryReader, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__444)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__445 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Rabin, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__445)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__446 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ArrayType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__446)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__447 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_EnumType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__447)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__448 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PromotingEnumType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__448)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__449 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_MapType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__449)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__450 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_NullType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__450)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__451 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_BoolType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__451)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__452 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IntType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__452)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__453 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_LongType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__453)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__454 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_FloatType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__454)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__455 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_DoubleType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__455)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__456 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__PlaceholderType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__456)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__457 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Record, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__457)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__458 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_RecordField, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__458)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__459 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__FieldAccessor, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__459)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__460 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_RecordType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__460)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__461 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PromotingRecordTy, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__461)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__462 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_BytesType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__462)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__463 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_StringType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__463)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__464 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_FixedType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__464)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__465 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_UnionType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__465)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__466 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__Sentinel, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__466)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__467 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__CanonicalForm, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__467)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__468 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_AvroType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__468)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__469 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__NamedType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__469)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__470 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Schema, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__470)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__471 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ResolvedSchema, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__471)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__472 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ValueAdapter, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__472)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__473 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_LogicalType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__473)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__474 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_CustomLogicalType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__474)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__475 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_DecimalType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__475)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__476 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_UUIDBase, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__476)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__477 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_UUIDStringType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__477)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__478 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_UUIDFixedType, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__478)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__479 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Date, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__479)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__480 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_TimeMillis, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__480)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__481 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_TimeMicros, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__481)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__482 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_TimestampMillis, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__482)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__483 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_TimestampMicros, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__483)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__484 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Codec, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__484)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__485 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__SnappyCodec, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__485)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__486 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__NullCodec, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__486)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__487 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__DeflateCodec, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__487)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__488 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle___Bzip2Codec, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__488)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__489 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__LzmaCodec, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__489)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__490 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__ZStandardCodec, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__490)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__491 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__Lz4Codec, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__491)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__492 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ContainerReader, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__492)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__493 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PromoteToFloat, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__493)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__494 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PromoteBytesToStr, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__494)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__495 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PromoteStringToBy, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__495)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__496 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_CannotPromote, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__496)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_get.method_name = &__pyx_n_s_get;
  __pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_items.method_name = &__pyx_n_s_items;
  __pyx_umethod_PyDict_Type_keys.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_keys.method_name = &__pyx_n_s_keys;
  __pyx_umethod_PyDict_Type_setdefault.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_setdefault.method_name = &__pyx_n_s_setdefault;
  __pyx_umethod_PyDict_Type_update.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_update.method_name = &__pyx_n_s_update;
  __pyx_umethod_PyUnicode_Type_rjust.type = (PyObject*)&PyUnicode_Type;
  __pyx_umethod_PyUnicode_Type_rjust.method_name = &__pyx_n_s_rjust;
  __pyx_umethod_PyUnicode_Type_rsplit.type = (PyObject*)&PyUnicode_Type;
  __pyx_umethod_PyUnicode_Type_rsplit.method_name = &__pyx_n_s_rsplit;
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(13, 2, __pyx_L1_error);
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_16 = PyInt_FromLong(16); if (unlikely(!__pyx_int_16)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_24 = PyInt_FromLong(24); if (unlikely(!__pyx_int_24)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_60 = PyInt_FromLong(60); if (unlikely(!__pyx_int_60)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_70 = PyInt_FromLong(70); if (unlikely(!__pyx_int_70)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_127 = PyInt_FromLong(127); if (unlikely(!__pyx_int_127)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_255 = PyInt_FromLong(255); if (unlikely(!__pyx_int_255)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_1000 = PyInt_FromLong(1000); if (unlikely(!__pyx_int_1000)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_1970 = PyInt_FromLong(1970); if (unlikely(!__pyx_int_1970)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_3600 = PyInt_FromLong(3600); if (unlikely(!__pyx_int_3600)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_4096 = PyInt_FromLong(4096); if (unlikely(!__pyx_int_4096)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_32767 = PyInt_FromLong(32767L); if (unlikely(!__pyx_int_32767)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_65536 = PyInt_FromLong(65536L); if (unlikely(!__pyx_int_65536)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_1000000 = PyInt_FromLong(1000000L); if (unlikely(!__pyx_int_1000000)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_8388607 = PyInt_FromLong(8388607L); if (unlikely(!__pyx_int_8388607)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_18332779 = PyInt_FromLong(18332779L); if (unlikely(!__pyx_int_18332779)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_24827247 = PyInt_FromLong(24827247L); if (unlikely(!__pyx_int_24827247)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_25341170 = PyInt_FromLong(25341170L); if (unlikely(!__pyx_int_25341170)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_29099189 = PyInt_FromLong(29099189L); if (unlikely(!__pyx_int_29099189)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_31039736 = PyInt_FromLong(31039736L); if (unlikely(!__pyx_int_31039736)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_33963030 = PyInt_FromLong(33963030L); if (unlikely(!__pyx_int_33963030)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_38410757 = PyInt_FromLong(38410757L); if (unlikely(!__pyx_int_38410757)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_38632982 = PyInt_FromLong(38632982L); if (unlikely(!__pyx_int_38632982)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_39872412 = PyInt_FromLong(39872412L); if (unlikely(!__pyx_int_39872412)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_42565757 = PyInt_FromLong(42565757L); if (unlikely(!__pyx_int_42565757)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_47772561 = PyInt_FromLong(47772561L); if (unlikely(!__pyx_int_47772561)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_60000000 = PyInt_FromLong(60000000L); if (unlikely(!__pyx_int_60000000)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_61270791 = PyInt_FromLong(61270791L); if (unlikely(!__pyx_int_61270791)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_61440409 = PyInt_FromLong(61440409L); if (unlikely(!__pyx_int_61440409)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_61765448 = PyInt_FromLong(61765448L); if (unlikely(!__pyx_int_61765448)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_67877419 = PyInt_FromLong(67877419L); if (unlikely(!__pyx_int_67877419)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_75337510 = PyInt_FromLong(75337510L); if (unlikely(!__pyx_int_75337510)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_79389776 = PyInt_FromLong(79389776L); if (unlikely(!__pyx_int_79389776)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_79941127 = PyInt_FromLong(79941127L); if (unlikely(!__pyx_int_79941127)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_81233721 = PyInt_FromLong(81233721L); if (unlikely(!__pyx_int_81233721)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_83351333 = PyInt_FromLong(83351333L); if (unlikely(!__pyx_int_83351333)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_95761985 = PyInt_FromLong(95761985L); if (unlikely(!__pyx_int_95761985)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_96605483 = PyInt_FromLong(96605483L); if (unlikely(!__pyx_int_96605483)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_101034655 = PyInt_FromLong(101034655L); if (unlikely(!__pyx_int_101034655)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_101670942 = PyInt_FromLong(101670942L); if (unlikely(!__pyx_int_101670942)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_111776469 = PyInt_FromLong(111776469L); if (unlikely(!__pyx_int_111776469)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_111957047 = PyInt_FromLong(111957047L); if (unlikely(!__pyx_int_111957047)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_112105877 = PyInt_FromLong(112105877L); if (unlikely(!__pyx_int_112105877)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_115671040 = PyInt_FromLong(115671040L); if (unlikely(!__pyx_int_115671040)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_119832798 = PyInt_FromLong(119832798L); if (unlikely(!__pyx_int_119832798)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_122446337 = PyInt_FromLong(122446337L); if (unlikely(!__pyx_int_122446337)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_122978751 = PyInt_FromLong(122978751L); if (unlikely(!__pyx_int_122978751)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_136983863 = PyInt_FromLong(136983863L); if (unlikely(!__pyx_int_136983863)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_139219010 = PyInt_FromLong(139219010L); if (unlikely(!__pyx_int_139219010)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_145104888 = PyInt_FromLong(145104888L); if (unlikely(!__pyx_int_145104888)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_148208818 = PyInt_FromLong(148208818L); if (unlikely(!__pyx_int_148208818)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_148222551 = PyInt_FromLong(148222551L); if (unlikely(!__pyx_int_148222551)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_148338675 = PyInt_FromLong(148338675L); if (unlikely(!__pyx_int_148338675)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_148737976 = PyInt_FromLong(148737976L); if (unlikely(!__pyx_int_148737976)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_150354737 = PyInt_FromLong(150354737L); if (unlikely(!__pyx_int_150354737)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_159023012 = PyInt_FromLong(159023012L); if (unlikely(!__pyx_int_159023012)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_159276020 = PyInt_FromLong(159276020L); if (unlikely(!__pyx_int_159276020)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_159891379 = PyInt_FromLong(159891379L); if (unlikely(!__pyx_int_159891379)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_160503772 = PyInt_FromLong(160503772L); if (unlikely(!__pyx_int_160503772)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_160934521 = PyInt_FromLong(160934521L); if (unlikely(!__pyx_int_160934521)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_164094308 = PyInt_FromLong(164094308L); if (unlikely(!__pyx_int_164094308)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_169331114 = PyInt_FromLong(169331114L); if (unlikely(!__pyx_int_169331114)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_175892883 = PyInt_FromLong(175892883L); if (unlikely(!__pyx_int_175892883)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_176081420 = PyInt_FromLong(176081420L); if (unlikely(!__pyx_int_176081420)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_176267327 = PyInt_FromLong(176267327L); if (unlikely(!__pyx_int_176267327)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_177382991 = PyInt_FromLong(177382991L); if (unlikely(!__pyx_int_177382991)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_178791180 = PyInt_FromLong(178791180L); if (unlikely(!__pyx_int_178791180)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_180170947 = PyInt_FromLong(180170947L); if (unlikely(!__pyx_int_180170947)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_180285838 = PyInt_FromLong(180285838L); if (unlikely(!__pyx_int_180285838)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_181080083 = PyInt_FromLong(181080083L); if (unlikely(!__pyx_int_181080083)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_184956420 = PyInt_FromLong(184956420L); if (unlikely(!__pyx_int_184956420)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_193777858 = PyInt_FromLong(193777858L); if (unlikely(!__pyx_int_193777858)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_195980960 = PyInt_FromLong(195980960L); if (unlikely(!__pyx_int_195980960)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_198151601 = PyInt_FromLong(198151601L); if (unlikely(!__pyx_int_198151601)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_202158861 = PyInt_FromLong(202158861L); if (unlikely(!__pyx_int_202158861)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_202608488 = PyInt_FromLong(202608488L); if (unlikely(!__pyx_int_202608488)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_203789025 = PyInt_FromLong(203789025L); if (unlikely(!__pyx_int_203789025)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_205941106 = PyInt_FromLong(205941106L); if (unlikely(!__pyx_int_205941106)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_206778879 = PyInt_FromLong(206778879L); if (unlikely(!__pyx_int_206778879)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_212073902 = PyInt_FromLong(212073902L); if (unlikely(!__pyx_int_212073902)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_213886906 = PyInt_FromLong(213886906L); if (unlikely(!__pyx_int_213886906)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_215229444 = PyInt_FromLong(215229444L); if (unlikely(!__pyx_int_215229444)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_215954136 = PyInt_FromLong(215954136L); if (unlikely(!__pyx_int_215954136)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_228539134 = PyInt_FromLong(228539134L); if (unlikely(!__pyx_int_228539134)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_228825662 = PyInt_FromLong(228825662L); if (unlikely(!__pyx_int_228825662)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_229636597 = PyInt_FromLong(229636597L); if (unlikely(!__pyx_int_229636597)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_233298401 = PyInt_FromLong(233298401L); if (unlikely(!__pyx_int_233298401)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_236180632 = PyInt_FromLong(236180632L); if (unlikely(!__pyx_int_236180632)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_236349525 = PyInt_FromLong(236349525L); if (unlikely(!__pyx_int_236349525)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_238750788 = PyInt_FromLong(238750788L); if (unlikely(!__pyx_int_238750788)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_240389341 = PyInt_FromLong(240389341L); if (unlikely(!__pyx_int_240389341)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_245215779 = PyInt_FromLong(245215779L); if (unlikely(!__pyx_int_245215779)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_247527084 = PyInt_FromLong(247527084L); if (unlikely(!__pyx_int_247527084)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_254981756 = PyInt_FromLong(254981756L); if (unlikely(!__pyx_int_254981756)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_256569670 = PyInt_FromLong(256569670L); if (unlikely(!__pyx_int_256569670)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_259444540 = PyInt_FromLong(259444540L); if (unlikely(!__pyx_int_259444540)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_261059191 = PyInt_FromLong(261059191L); if (unlikely(!__pyx_int_261059191)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_263819094 = PyInt_FromLong(263819094L); if (unlikely(!__pyx_int_263819094)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_2147483647 = PyInt_FromLong(2147483647L); if (unlikely(!__pyx_int_2147483647)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_3442987655 = PyInt_FromString((char *)"3442987655", 0, 0); if (unlikely(!__pyx_int_3442987655)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_3600000000 = PyInt_FromString((char *)"3600000000", 0, 0); if (unlikely(!__pyx_int_3600000000)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_neg_15 = PyInt_FromLong(-15); if (unlikely(!__pyx_int_neg_15)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_549755813887L = PyLong_FromString((char *)"549755813887", 0, 0); if (unlikely(!__pyx_int_549755813887L)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_0x7fffffffffffL = PyLong_FromString((char *)"0x7fffffffffff", 0, 0); if (unlikely(!__pyx_int_0x7fffffffffffL)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_0x7fffffffffffffL = PyLong_FromString((char *)"0x7fffffffffffff", 0, 0); if (unlikely(!__pyx_int_0x7fffffffffffffL)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_int_0x7fffffffffffffffL = PyLong_FromString((char *)"0x7fffffffffffffff", 0, 0); if (unlikely(!__pyx_int_0x7fffffffffffffffL)) __PYX_ERR(13, 2, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(13, 2, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_5cavro_byte_buffer_template = ((arrayobject *)Py_None); Py_INCREF(Py_None);
  __pyx_v_5cavro_empty_buffer = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_dataclass_dflt5cavro_Options5cavro_union_decodes_to = Py_None; Py_INCREF(Py_None);
  __pyx_collections_abc_Sequence = Py_None; Py_INCREF(Py_None);
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_5cavro_Options = &__pyx_vtable_5cavro_Options;
  __pyx_vtable_5cavro_Options.can_have_missing_values = (int (*)(struct __pyx_obj_5cavro_Options *))__pyx_f_5cavro_7Options_can_have_missing_values;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro_Options = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_Options_spec, NULL); if (unlikely(!__pyx_ptype_5cavro_Options)) __PYX_ERR(4, 89, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_Options_spec, __pyx_ptype_5cavro_Options) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_Options = &__pyx_type_5cavro_Options;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_Options) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_Options->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_Options->tp_dictoffset && __pyx_ptype_5cavro_Options->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_Options->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_Options, __pyx_vtabptr_5cavro_Options) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_Options) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Options, (PyObject *) __pyx_ptype_5cavro_Options) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_Options) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro__Writer = &__pyx_vtable_5cavro__Writer;
  __pyx_vtable_5cavro__Writer.write_u8 = (int (*)(struct __pyx_obj_5cavro__Writer *, uint8_t))__pyx_f_5cavro_7_Writer_write_u8;
  __pyx_vtable_5cavro__Writer.write_n = (int (*)(struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice))__pyx_f_5cavro_7_Writer_write_n;
  __pyx_vtable_5cavro__Writer.flush = (int (*)(struct __pyx_obj_5cavro__Writer *))__pyx_f_5cavro_7_Writer_flush;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro__Writer = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__Writer_spec, NULL); if (unlikely(!__pyx_ptype_5cavro__Writer)) __PYX_ERR(5, 4, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__Writer_spec, __pyx_ptype_5cavro__Writer) < 0) __PYX_ERR(5, 4, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__Writer = &__pyx_type_5cavro__Writer;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__Writer) < 0) __PYX_ERR(5, 4, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__Writer->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__Writer->tp_dictoffset && __pyx_ptype_5cavro__Writer->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__Writer->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro__Writer, __pyx_vtabptr_5cavro__Writer) < 0) __PYX_ERR(5, 4, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro__Writer) < 0) __PYX_ERR(5, 4, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Writer, (PyObject *) __pyx_ptype_5cavro__Writer) < 0) __PYX_ERR(5, 4, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__Writer) < 0) __PYX_ERR(5, 4, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro__Reader = &__pyx_vtable_5cavro__Reader;
  __pyx_vtable_5cavro__Reader.read_u8 = (uint8_t (*)(struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_7_Reader_read_u8;
  __pyx_vtable_5cavro__Reader.read_bytes = (PyObject *(*)(struct __pyx_obj_5cavro__Reader *, Py_ssize_t))__pyx_f_5cavro_7_Reader_read_bytes;
  __pyx_vtable_5cavro__Reader.read_n = (__Pyx_memviewslice (*)(struct __pyx_obj_5cavro__Reader *, Py_ssize_t))__pyx_f_5cavro_7_Reader_read_n;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro__Reader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__Reader_spec, NULL); if (unlikely(!__pyx_ptype_5cavro__Reader)) __PYX_ERR(5, 18, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__Reader_spec, __pyx_ptype_5cavro__Reader) < 0) __PYX_ERR(5, 18, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__Reader = &__pyx_type_5cavro__Reader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__Reader) < 0) __PYX_ERR(5, 18, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__Reader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__Reader->tp_dictoffset && __pyx_ptype_5cavro__Reader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__Reader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro__Reader, __pyx_vtabptr_5cavro__Reader) < 0) __PYX_ERR(5, 18, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro__Reader) < 0) __PYX_ERR(5, 18, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Reader, (PyObject *) __pyx_ptype_5cavro__Reader) < 0) __PYX_ERR(5, 18, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__Reader) < 0) __PYX_ERR(5, 18, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_FileReader = &__pyx_vtable_5cavro_FileReader;
  __pyx_vtable_5cavro_FileReader.__pyx_base = *__pyx_vtabptr_5cavro__Reader;
  __pyx_vtable_5cavro_FileReader.__pyx_base.read_u8 = (uint8_t (*)(struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_10FileReader_read_u8;
  __pyx_vtable_5cavro_FileReader.__pyx_base.read_bytes = (PyObject *(*)(struct __pyx_obj_5cavro__Reader *, Py_ssize_t))__pyx_f_5cavro_10FileReader_read_bytes;
  __pyx_vtable_5cavro_FileReader.__pyx_base.read_n = (__Pyx_memviewslice (*)(struct __pyx_obj_5cavro__Reader *, Py_ssize_t))__pyx_f_5cavro_10FileReader_read_n;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro__Reader); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_FileReader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_FileReader_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_FileReader)) __PYX_ERR(5, 33, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_FileReader_spec, __pyx_ptype_5cavro_FileReader) < 0) __PYX_ERR(5, 33, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_FileReader = &__pyx_type_5cavro_FileReader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_FileReader->tp_base = __pyx_ptype_5cavro__Reader;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_FileReader) < 0) __PYX_ERR(5, 33, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_FileReader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_FileReader->tp_dictoffset && __pyx_ptype_5cavro_FileReader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_FileReader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_FileReader, __pyx_vtabptr_5cavro_FileReader) < 0) __PYX_ERR(5, 33, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_FileReader) < 0) __PYX_ERR(5, 33, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FileReader, (PyObject *) __pyx_ptype_5cavro_FileReader) < 0) __PYX_ERR(5, 33, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_FileReader) < 0) __PYX_ERR(5, 33, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_FileWriter = &__pyx_vtable_5cavro_FileWriter;
  __pyx_vtable_5cavro_FileWriter.__pyx_base = *__pyx_vtabptr_5cavro__Writer;
  __pyx_vtable_5cavro_FileWriter.__pyx_base.write_u8 = (int (*)(struct __pyx_obj_5cavro__Writer *, uint8_t))__pyx_f_5cavro_10FileWriter_write_u8;
  __pyx_vtable_5cavro_FileWriter.__pyx_base.write_n = (int (*)(struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice))__pyx_f_5cavro_10FileWriter_write_n;
  __pyx_vtable_5cavro_FileWriter.__pyx_base.flush = (int (*)(struct __pyx_obj_5cavro__Writer *))__pyx_f_5cavro_10FileWriter_flush;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro__Writer); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_FileWriter = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_FileWriter_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_FileWriter)) __PYX_ERR(5, 68, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_FileWriter_spec, __pyx_ptype_5cavro_FileWriter) < 0) __PYX_ERR(5, 68, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_FileWriter = &__pyx_type_5cavro_FileWriter;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_FileWriter->tp_base = __pyx_ptype_5cavro__Writer;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_FileWriter) < 0) __PYX_ERR(5, 68, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_FileWriter->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_FileWriter->tp_dictoffset && __pyx_ptype_5cavro_FileWriter->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_FileWriter->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_FileWriter, __pyx_vtabptr_5cavro_FileWriter) < 0) __PYX_ERR(5, 68, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_FileWriter) < 0) __PYX_ERR(5, 68, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FileWriter, (PyObject *) __pyx_ptype_5cavro_FileWriter) < 0) __PYX_ERR(5, 68, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_FileWriter) < 0) __PYX_ERR(5, 68, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_MemoryWriter = &__pyx_vtable_5cavro_MemoryWriter;
  __pyx_vtable_5cavro_MemoryWriter.__pyx_base = *__pyx_vtabptr_5cavro__Writer;
  __pyx_vtable_5cavro_MemoryWriter.__pyx_base.write_u8 = (int (*)(struct __pyx_obj_5cavro__Writer *, uint8_t))__pyx_f_5cavro_12MemoryWriter_write_u8;
  __pyx_vtable_5cavro_MemoryWriter.__pyx_base.write_n = (int (*)(struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice))__pyx_f_5cavro_12MemoryWriter_write_n;
  __pyx_vtable_5cavro_MemoryWriter.bytes = (PyObject *(*)(struct __pyx_obj_5cavro_MemoryWriter *))__pyx_f_5cavro_12MemoryWriter_bytes;
  __pyx_vtable_5cavro_MemoryWriter.view = (__Pyx_memviewslice (*)(struct __pyx_obj_5cavro_MemoryWriter *))__pyx_f_5cavro_12MemoryWriter_view;
  __pyx_vtable_5cavro_MemoryWriter.reset = (PyObject *(*)(struct __pyx_obj_5cavro_MemoryWriter *))__pyx_f_5cavro_12MemoryWriter_reset;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro__Writer); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_MemoryWriter = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_MemoryWriter_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_MemoryWriter)) __PYX_ERR(16, 7, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_MemoryWriter_spec, __pyx_ptype_5cavro_MemoryWriter) < 0) __PYX_ERR(16, 7, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_MemoryWriter = &__pyx_type_5cavro_MemoryWriter;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_MemoryWriter->tp_base = __pyx_ptype_5cavro__Writer;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_MemoryWriter) < 0) __PYX_ERR(16, 7, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_MemoryWriter->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_MemoryWriter->tp_dictoffset && __pyx_ptype_5cavro_MemoryWriter->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_MemoryWriter->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_MemoryWriter, __pyx_vtabptr_5cavro_MemoryWriter) < 0) __PYX_ERR(16, 7, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_MemoryWriter) < 0) __PYX_ERR(16, 7, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MemoryWriter, (PyObject *) __pyx_ptype_5cavro_MemoryWriter) < 0) __PYX_ERR(16, 7, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_MemoryWriter) < 0) __PYX_ERR(16, 7, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_MemoryReader = &__pyx_vtable_5cavro_MemoryReader;
  __pyx_vtable_5cavro_MemoryReader.__pyx_base = *__pyx_vtabptr_5cavro__Reader;
  __pyx_vtable_5cavro_MemoryReader.__pyx_base.read_u8 = (uint8_t (*)(struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_12MemoryReader_read_u8;
  __pyx_vtable_5cavro_MemoryReader.__pyx_base.read_bytes = (PyObject *(*)(struct __pyx_obj_5cavro__Reader *, Py_ssize_t))__pyx_f_5cavro_12MemoryReader_read_bytes;
  __pyx_vtable_5cavro_MemoryReader.__pyx_base.read_n = (__Pyx_memviewslice (*)(struct __pyx_obj_5cavro__Reader *, Py_ssize_t))__pyx_f_5cavro_12MemoryReader_read_n;
  __pyx_vtable_5cavro_MemoryReader._reset_to = (void (*)(struct __pyx_obj_5cavro_MemoryReader *, __Pyx_memviewslice))__pyx_f_5cavro_12MemoryReader__reset_to;
  __pyx_vtable_5cavro_MemoryReader.advance = (uint8_t const *(*)(struct __pyx_obj_5cavro_MemoryReader *, size_t))__pyx_f_5cavro_12MemoryReader_advance;
  __pyx_vtable_5cavro_MemoryReader.ensure = (int (*)(struct __pyx_obj_5cavro_MemoryReader *, size_t))__pyx_f_5cavro_12MemoryReader_ensure;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro__Reader); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_MemoryReader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_MemoryReader_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_MemoryReader)) __PYX_ERR(16, 51, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_MemoryReader_spec, __pyx_ptype_5cavro_MemoryReader) < 0) __PYX_ERR(16, 51, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_MemoryReader = &__pyx_type_5cavro_MemoryReader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_MemoryReader->tp_base = __pyx_ptype_5cavro__Reader;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_MemoryReader) < 0) __PYX_ERR(16, 51, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_MemoryReader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_MemoryReader->tp_dictoffset && __pyx_ptype_5cavro_MemoryReader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_MemoryReader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_MemoryReader, __pyx_vtabptr_5cavro_MemoryReader) < 0) __PYX_ERR(16, 51, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_MemoryReader) < 0) __PYX_ERR(16, 51, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MemoryReader, (PyObject *) __pyx_ptype_5cavro_MemoryReader) < 0) __PYX_ERR(16, 51, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_MemoryReader) < 0) __PYX_ERR(16, 51, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_Rabin = &__pyx_vtable_5cavro_Rabin;
  __pyx_vtable_5cavro_Rabin.update = (PyObject *(*)(struct __pyx_obj_5cavro_Rabin *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_5Rabin_update;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro_Rabin = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_Rabin_spec, NULL); if (unlikely(!__pyx_ptype_5cavro_Rabin)) __PYX_ERR(18, 25, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_Rabin_spec, __pyx_ptype_5cavro_Rabin) < 0) __PYX_ERR(18, 25, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_Rabin = &__pyx_type_5cavro_Rabin;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_Rabin) < 0) __PYX_ERR(18, 25, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_Rabin->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_Rabin->tp_dictoffset && __pyx_ptype_5cavro_Rabin->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_Rabin->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_Rabin, __pyx_vtabptr_5cavro_Rabin) < 0) __PYX_ERR(18, 25, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_Rabin) < 0) __PYX_ERR(18, 25, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Rabin, (PyObject *) __pyx_ptype_5cavro_Rabin) < 0) __PYX_ERR(18, 25, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_Rabin) < 0) __PYX_ERR(18, 25, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_Record = &__pyx_vtable_5cavro_Record;
  __pyx_vtable_5cavro_Record._repr_children = (PyObject *(*)(struct __pyx_obj_5cavro_Record *, PyObject *))__pyx_f_5cavro_6Record__repr_children;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro_Record = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_Record_spec, NULL); if (unlikely(!__pyx_ptype_5cavro_Record)) __PYX_ERR(11, 88, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_Record_spec, __pyx_ptype_5cavro_Record) < 0) __PYX_ERR(11, 88, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_Record = &__pyx_type_5cavro_Record;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_Record) < 0) __PYX_ERR(11, 88, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_Record->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_Record->tp_dictoffset && __pyx_ptype_5cavro_Record->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_Record->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_Record, __pyx_vtabptr_5cavro_Record) < 0) __PYX_ERR(11, 88, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_Record) < 0) __PYX_ERR(11, 88, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Record_3, (PyObject *) __pyx_ptype_5cavro_Record) < 0) __PYX_ERR(11, 88, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_Record) < 0) __PYX_ERR(11, 88, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_RecordField = &__pyx_vtable_5cavro_RecordField;
  __pyx_vtable_5cavro_RecordField.for_reader = (PyObject *(*)(struct __pyx_obj_5cavro_RecordField *, PyObject *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_11RecordField_for_reader;
  __pyx_vtable_5cavro_RecordField.placeholder = (PyObject *(*)(struct __pyx_obj_5cavro_RecordField *))__pyx_f_5cavro_11RecordField_placeholder;
  __pyx_vtable_5cavro_RecordField.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_RecordField *, PyObject *))__pyx_f_5cavro_11RecordField_canonical_form;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro_RecordField = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_RecordField_spec, NULL); if (unlikely(!__pyx_ptype_5cavro_RecordField)) __PYX_ERR(11, 188, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_RecordField_spec, __pyx_ptype_5cavro_RecordField) < 0) __PYX_ERR(11, 188, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_RecordField = &__pyx_type_5cavro_RecordField;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_RecordField) < 0) __PYX_ERR(11, 188, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_RecordField->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_RecordField->tp_dictoffset && __pyx_ptype_5cavro_RecordField->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_RecordField->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_RecordField, __pyx_vtabptr_5cavro_RecordField) < 0) __PYX_ERR(11, 188, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_RecordField) < 0) __PYX_ERR(11, 188, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_RecordField, (PyObject *) __pyx_ptype_5cavro_RecordField) < 0) __PYX_ERR(11, 188, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_RecordField) < 0) __PYX_ERR(11, 188, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro__FieldAccessor = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__FieldAccessor_spec, NULL); if (unlikely(!__pyx_ptype_5cavro__FieldAccessor)) __PYX_ERR(11, 267, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__FieldAccessor_spec, __pyx_ptype_5cavro__FieldAccessor) < 0) __PYX_ERR(11, 267, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__FieldAccessor = &__pyx_type_5cavro__FieldAccessor;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__FieldAccessor) < 0) __PYX_ERR(11, 267, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__FieldAccessor->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__FieldAccessor->tp_dictoffset && __pyx_ptype_5cavro__FieldAccessor->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__FieldAccessor->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FieldAccessor, (PyObject *) __pyx_ptype_5cavro__FieldAccessor) < 0) __PYX_ERR(11, 267, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__FieldAccessor) < 0) __PYX_ERR(11, 267, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro__Sentinel = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__Sentinel_spec, NULL); if (unlikely(!__pyx_ptype_5cavro__Sentinel)) __PYX_ERR(2, 10, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__Sentinel_spec, __pyx_ptype_5cavro__Sentinel) < 0) __PYX_ERR(2, 10, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__Sentinel = &__pyx_type_5cavro__Sentinel;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__Sentinel) < 0) __PYX_ERR(2, 10, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__Sentinel->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__Sentinel->tp_dictoffset && __pyx_ptype_5cavro__Sentinel->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__Sentinel->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Sentinel, (PyObject *) __pyx_ptype_5cavro__Sentinel) < 0) __PYX_ERR(2, 10, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__Sentinel) < 0) __PYX_ERR(2, 10, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)(&PyUnicode_Type)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro__CanonicalForm = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__CanonicalForm_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro__CanonicalForm)) __PYX_ERR(2, 21, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__CanonicalForm_spec, __pyx_ptype_5cavro__CanonicalForm) < 0) __PYX_ERR(2, 21, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__CanonicalForm = &__pyx_type_5cavro__CanonicalForm;
  #endif
  if (sizeof(struct __pyx_obj_5cavro__CanonicalForm) != sizeof(PyUnicodeObject)) {
    if (__Pyx_validate_extern_base((&PyUnicode_Type)) < 0) __PYX_ERR(2, 21, __pyx_L1_error)
  }
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro__CanonicalForm->tp_dealloc = (&PyUnicode_Type)->tp_dealloc;
  __pyx_ptype_5cavro__CanonicalForm->tp_base = (&PyUnicode_Type);
  __pyx_ptype_5cavro__CanonicalForm->tp_new = (&PyUnicode_Type)->tp_new;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__CanonicalForm) < 0) __PYX_ERR(2, 21, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__CanonicalForm->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__CanonicalForm->tp_dictoffset && __pyx_ptype_5cavro__CanonicalForm->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__CanonicalForm->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CanonicalForm, (PyObject *) __pyx_ptype_5cavro__CanonicalForm) < 0) __PYX_ERR(2, 21, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__CanonicalForm) < 0) __PYX_ERR(2, 21, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_AvroType = &__pyx_vtable_5cavro_AvroType;
  __pyx_vtable_5cavro_AvroType.clone_base = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_opt_args_5cavro_8AvroType_clone_base *__pyx_optional_args))__pyx_f_5cavro_8AvroType_clone_base;
  __pyx_vtable_5cavro_AvroType.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_8AvroType_copy;
  __pyx_vtable_5cavro_AvroType._make_logical = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, PyObject *))__pyx_f_5cavro_8AvroType__make_logical;
  __pyx_vtable_5cavro_AvroType._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8AvroType__extract_metadata;
  __pyx_vtable_5cavro_AvroType._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_8AvroType__convert_value;
  __pyx_vtable_5cavro_AvroType.convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_8AvroType_convert_value *__pyx_optional_args))__pyx_f_5cavro_8AvroType_convert_value;
  __pyx_vtable_5cavro_AvroType.binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_8AvroType_binary_buffer_encode;
  __pyx_vtable_5cavro_AvroType._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_8AvroType__binary_buffer_encode;
  __pyx_vtable_5cavro_AvroType.binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_8AvroType_binary_buffer_decode;
  __pyx_vtable_5cavro_AvroType._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_8AvroType__binary_buffer_decode;
  __pyx_vtable_5cavro_AvroType.get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8AvroType_get_value_fitness;
  __pyx_vtable_5cavro_AvroType._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8AvroType__get_value_fitness;
  __pyx_vtable_5cavro_AvroType.assert_value = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8AvroType_assert_value;
  __pyx_vtable_5cavro_AvroType.json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8AvroType_json_format;
  __pyx_vtable_5cavro_AvroType._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8AvroType__json_format;
  __pyx_vtable_5cavro_AvroType.json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8AvroType_json_decode;
  __pyx_vtable_5cavro_AvroType._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8AvroType__json_decode;
  __pyx_vtable_5cavro_AvroType.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8AvroType_canonical_form;
  __pyx_vtable_5cavro_AvroType._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_8AvroType__get_schema_extra;
  __pyx_vtable_5cavro_AvroType.for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *, struct __pyx_opt_args_5cavro_8AvroType_for_writer *__pyx_optional_args))__pyx_f_5cavro_8AvroType_for_writer;
  __pyx_vtable_5cavro_AvroType._for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_8AvroType__for_writer;
  __pyx_vtable_5cavro_AvroType.accepts_missing_value = (int (*)(struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_8AvroType_accepts_missing_value;
  __pyx_vtable_5cavro_AvroType.resolve_default_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, PyObject *))__pyx_f_5cavro_8AvroType_resolve_default_value;
  __pyx_vtable_5cavro_AvroType.get_schema = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_8AvroType_get_schema *__pyx_optional_args))__pyx_f_5cavro_8AvroType_get_schema;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro_AvroType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_AvroType_spec, NULL); if (unlikely(!__pyx_ptype_5cavro_AvroType)) __PYX_ERR(2, 45, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_AvroType_spec, __pyx_ptype_5cavro_AvroType) < 0) __PYX_ERR(2, 45, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_AvroType = &__pyx_type_5cavro_AvroType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_AvroType) < 0) __PYX_ERR(2, 45, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_AvroType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_AvroType->tp_dictoffset && __pyx_ptype_5cavro_AvroType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_AvroType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_AvroType, __pyx_vtabptr_5cavro_AvroType) < 0) __PYX_ERR(2, 45, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_AvroType) < 0) __PYX_ERR(2, 45, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AvroType, (PyObject *) __pyx_ptype_5cavro_AvroType) < 0) __PYX_ERR(2, 45, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_AvroType) < 0) __PYX_ERR(2, 45, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_UnionType = &__pyx_vtable_5cavro_UnionType;
  __pyx_vtable_5cavro_UnionType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro_UnionType.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_9UnionType_copy;
  __pyx_vtable_5cavro_UnionType.__pyx_base._make_logical = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, PyObject *))__pyx_f_5cavro_9UnionType__make_logical;
  __pyx_vtable_5cavro_UnionType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9UnionType__extract_metadata;
  __pyx_vtable_5cavro_UnionType.__pyx_base.convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_8AvroType_convert_value *__pyx_optional_args))__pyx_f_5cavro_9UnionType_convert_value;
  __pyx_vtable_5cavro_UnionType.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_9UnionType__binary_buffer_encode;
  __pyx_vtable_5cavro_UnionType.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_9UnionType__binary_buffer_decode;
  __pyx_vtable_5cavro_UnionType.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9UnionType__get_value_fitness;
  __pyx_vtable_5cavro_UnionType.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9UnionType__json_format;
  __pyx_vtable_5cavro_UnionType.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9UnionType__json_decode;
  __pyx_vtable_5cavro_UnionType.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9UnionType_canonical_form;
  __pyx_vtable_5cavro_UnionType.__pyx_base._for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_9UnionType__for_writer;
  __pyx_vtable_5cavro_UnionType.__pyx_base.accepts_missing_value = (int (*)(struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_9UnionType_accepts_missing_value;
  __pyx_vtable_5cavro_UnionType.__pyx_base.resolve_default_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, PyObject *))__pyx_f_5cavro_9UnionType_resolve_default_value;
  __pyx_vtable_5cavro_UnionType.__pyx_base.get_schema = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_8AvroType_get_schema *__pyx_optional_args))__pyx_f_5cavro_9UnionType_get_schema;
  __pyx_vtable_5cavro_UnionType._populate_name_map = (PyObject *(*)(struct __pyx_obj_5cavro_UnionType *))__pyx_f_5cavro_9UnionType__populate_name_map;
  __pyx_vtable_5cavro_UnionType.resolve_from_value = (Py_ssize_t (*)(struct __pyx_obj_5cavro_UnionType *, PyObject *))__pyx_f_5cavro_9UnionType_resolve_from_value;
  __pyx_vtable_5cavro_UnionType.for_reader = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_UnionType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_9UnionType_for_reader;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_UnionType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_UnionType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_UnionType)) __PYX_ERR(22, 25, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_UnionType_spec, __pyx_ptype_5cavro_UnionType) < 0) __PYX_ERR(22, 25, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_UnionType = &__pyx_type_5cavro_UnionType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_UnionType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_UnionType) < 0) __PYX_ERR(22, 25, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_UnionType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_UnionType->tp_dictoffset && __pyx_ptype_5cavro_UnionType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_UnionType->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_UnionType, __pyx_vtabptr_5cavro_UnionType) < 0) __PYX_ERR(22, 25, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_UnionType) < 0) __PYX_ERR(22, 25, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UnionType, (PyObject *) __pyx_ptype_5cavro_UnionType) < 0) __PYX_ERR(22, 25, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_UnionType) < 0) __PYX_ERR(22, 25, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_StringType = &__pyx_vtable_5cavro_StringType;
  __pyx_vtable_5cavro_StringType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro_StringType.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_10StringType_copy;
  __pyx_vtable_5cavro_StringType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10StringType__extract_metadata;
  __pyx_vtable_5cavro_StringType.__pyx_base._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_10StringType__convert_value;
  __pyx_vtable_5cavro_StringType.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_10StringType__binary_buffer_encode;
  __pyx_vtable_5cavro_StringType.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_10StringType__binary_buffer_decode;
  __pyx_vtable_5cavro_StringType.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10StringType__get_value_fitness;
  __pyx_vtable_5cavro_StringType.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10StringType__json_format;
  __pyx_vtable_5cavro_StringType.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10StringType__json_decode;
  __pyx_vtable_5cavro_StringType.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10StringType_canonical_form;
  __pyx_vtable_5cavro_StringType.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_10StringType__get_schema_extra;
  __pyx_vtable_5cavro_StringType.__pyx_base._for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_10StringType__for_writer;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_StringType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_StringType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_StringType)) __PYX_ERR(12, 93, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_StringType_spec, __pyx_ptype_5cavro_StringType) < 0) __PYX_ERR(12, 93, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_StringType = &__pyx_type_5cavro_StringType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_StringType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_StringType) < 0) __PYX_ERR(12, 93, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_StringType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_StringType->tp_dictoffset && __pyx_ptype_5cavro_StringType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_StringType->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_StringType, __pyx_vtabptr_5cavro_StringType) < 0) __PYX_ERR(12, 93, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_StringType) < 0) __PYX_ERR(12, 93, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_StringType, (PyObject *) __pyx_ptype_5cavro_StringType) < 0) __PYX_ERR(12, 93, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_StringType) < 0) __PYX_ERR(12, 93, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_BytesType = &__pyx_vtable_5cavro_BytesType;
  __pyx_vtable_5cavro_BytesType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro_BytesType.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_9BytesType_copy;
  __pyx_vtable_5cavro_BytesType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9BytesType__extract_metadata;
  __pyx_vtable_5cavro_BytesType.__pyx_base._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_9BytesType__convert_value;
  __pyx_vtable_5cavro_BytesType.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_9BytesType__binary_buffer_encode;
  __pyx_vtable_5cavro_BytesType.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_9BytesType__binary_buffer_decode;
  __pyx_vtable_5cavro_BytesType.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9BytesType__get_value_fitness;
  __pyx_vtable_5cavro_BytesType.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9BytesType__json_format;
  __pyx_vtable_5cavro_BytesType.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9BytesType__json_decode;
  __pyx_vtable_5cavro_BytesType.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9BytesType_canonical_form;
  __pyx_vtable_5cavro_BytesType.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_9BytesType__get_schema_extra;
  __pyx_vtable_5cavro_BytesType.__pyx_base._for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_9BytesType__for_writer;
  __pyx_vtable_5cavro_BytesType.__pyx_base.resolve_default_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, PyObject *))__pyx_f_5cavro_9BytesType_resolve_default_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_BytesType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_BytesType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_BytesType)) __PYX_ERR(12, 3, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_BytesType_spec, __pyx_ptype_5cavro_BytesType) < 0) __PYX_ERR(12, 3, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_BytesType = &__pyx_type_5cavro_BytesType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_BytesType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_BytesType) < 0) __PYX_ERR(12, 3, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_BytesType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_BytesType->tp_dictoffset && __pyx_ptype_5cavro_BytesType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_BytesType->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_BytesType, __pyx_vtabptr_5cavro_BytesType) < 0) __PYX_ERR(12, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_BytesType) < 0) __PYX_ERR(12, 3, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BytesType, (PyObject *) __pyx_ptype_5cavro_BytesType) < 0) __PYX_ERR(12, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_BytesType) < 0) __PYX_ERR(12, 3, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro__PlaceholderType = &__pyx_vtable_5cavro__PlaceholderType;
  __pyx_vtable_5cavro__PlaceholderType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro__PlaceholderType.__pyx_base._make_logical = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, PyObject *))__pyx_f_5cavro_16_PlaceholderType__make_logical;
  __pyx_vtable_5cavro__PlaceholderType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_16_PlaceholderType__extract_metadata;
  __pyx_vtable_5cavro__PlaceholderType.__pyx_base._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_16_PlaceholderType__convert_value;
  __pyx_vtable_5cavro__PlaceholderType.__pyx_base.binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_16_PlaceholderType_binary_buffer_encode;
  __pyx_vtable_5cavro__PlaceholderType.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_16_PlaceholderType__binary_buffer_decode;
  __pyx_vtable_5cavro__PlaceholderType.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_16_PlaceholderType__get_value_fitness;
  __pyx_vtable_5cavro__PlaceholderType.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_16_PlaceholderType__json_format;
  __pyx_vtable_5cavro__PlaceholderType.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_16_PlaceholderType__json_decode;
  __pyx_vtable_5cavro__PlaceholderType.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_16_PlaceholderType_canonical_form;
  __pyx_vtable_5cavro__PlaceholderType.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_16_PlaceholderType__get_schema_extra;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro__PlaceholderType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__PlaceholderType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro__PlaceholderType)) __PYX_ERR(11, 15, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__PlaceholderType_spec, __pyx_ptype_5cavro__PlaceholderType) < 0) __PYX_ERR(11, 15, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__PlaceholderType = &__pyx_type_5cavro__PlaceholderType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro__PlaceholderType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__PlaceholderType) < 0) __PYX_ERR(11, 15, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__PlaceholderType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__PlaceholderType->tp_dictoffset && __pyx_ptype_5cavro__PlaceholderType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__PlaceholderType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro__PlaceholderType, __pyx_vtabptr_5cavro__PlaceholderType) < 0) __PYX_ERR(11, 15, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro__PlaceholderType) < 0) __PYX_ERR(11, 15, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PlaceholderType, (PyObject *) __pyx_ptype_5cavro__PlaceholderType) < 0) __PYX_ERR(11, 15, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__PlaceholderType) < 0) __PYX_ERR(11, 15, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_DoubleType = &__pyx_vtable_5cavro_DoubleType;
  __pyx_vtable_5cavro_DoubleType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro_DoubleType.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_10DoubleType_copy;
  __pyx_vtable_5cavro_DoubleType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10DoubleType__extract_metadata;
  __pyx_vtable_5cavro_DoubleType.__pyx_base._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_10DoubleType__convert_value;
  __pyx_vtable_5cavro_DoubleType.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_10DoubleType__binary_buffer_encode;
  __pyx_vtable_5cavro_DoubleType.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_10DoubleType__binary_buffer_decode;
  __pyx_vtable_5cavro_DoubleType.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10DoubleType__get_value_fitness;
  __pyx_vtable_5cavro_DoubleType.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10DoubleType__json_format;
  __pyx_vtable_5cavro_DoubleType.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10DoubleType__json_decode;
  __pyx_vtable_5cavro_DoubleType.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10DoubleType_canonical_form;
  __pyx_vtable_5cavro_DoubleType.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_10DoubleType__get_schema_extra;
  __pyx_vtable_5cavro_DoubleType.__pyx_base._for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_10DoubleType__for_writer;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_DoubleType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_DoubleType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_DoubleType)) __PYX_ERR(10, 324, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_DoubleType_spec, __pyx_ptype_5cavro_DoubleType) < 0) __PYX_ERR(10, 324, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_DoubleType = &__pyx_type_5cavro_DoubleType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_DoubleType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_DoubleType) < 0) __PYX_ERR(10, 324, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_DoubleType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_DoubleType->tp_dictoffset && __pyx_ptype_5cavro_DoubleType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_DoubleType->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_DoubleType, __pyx_vtabptr_5cavro_DoubleType) < 0) __PYX_ERR(10, 324, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_DoubleType) < 0) __PYX_ERR(10, 324, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DoubleType, (PyObject *) __pyx_ptype_5cavro_DoubleType) < 0) __PYX_ERR(10, 324, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_DoubleType) < 0) __PYX_ERR(10, 324, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_FloatType = &__pyx_vtable_5cavro_FloatType;
  __pyx_vtable_5cavro_FloatType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro_FloatType.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_9FloatType_copy;
  __pyx_vtable_5cavro_FloatType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9FloatType__extract_metadata;
  __pyx_vtable_5cavro_FloatType.__pyx_base._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_9FloatType__convert_value;
  __pyx_vtable_5cavro_FloatType.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_9FloatType__binary_buffer_encode;
  __pyx_vtable_5cavro_FloatType.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_9FloatType__binary_buffer_decode;
  __pyx_vtable_5cavro_FloatType.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9FloatType__get_value_fitness;
  __pyx_vtable_5cavro_FloatType.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9FloatType__json_format;
  __pyx_vtable_5cavro_FloatType.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9FloatType__json_decode;
  __pyx_vtable_5cavro_FloatType.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9FloatType_canonical_form;
  __pyx_vtable_5cavro_FloatType.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_9FloatType__get_schema_extra;
  __pyx_vtable_5cavro_FloatType.__pyx_base._for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_9FloatType__for_writer;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_FloatType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_FloatType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_FloatType)) __PYX_ERR(10, 209, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_FloatType_spec, __pyx_ptype_5cavro_FloatType) < 0) __PYX_ERR(10, 209, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_FloatType = &__pyx_type_5cavro_FloatType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_FloatType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_FloatType) < 0) __PYX_ERR(10, 209, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_FloatType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_FloatType->tp_dictoffset && __pyx_ptype_5cavro_FloatType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_FloatType->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_FloatType, __pyx_vtabptr_5cavro_FloatType) < 0) __PYX_ERR(10, 209, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_FloatType) < 0) __PYX_ERR(10, 209, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FloatType, (PyObject *) __pyx_ptype_5cavro_FloatType) < 0) __PYX_ERR(10, 209, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_FloatType) < 0) __PYX_ERR(10, 209, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_LongType = &__pyx_vtable_5cavro_LongType;
  __pyx_vtable_5cavro_LongType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro_LongType.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_8LongType_copy;
  __pyx_vtable_5cavro_LongType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8LongType__extract_metadata;
  __pyx_vtable_5cavro_LongType.__pyx_base._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_8LongType__convert_value;
  __pyx_vtable_5cavro_LongType.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_8LongType__binary_buffer_encode;
  __pyx_vtable_5cavro_LongType.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_8LongType__binary_buffer_decode;
  __pyx_vtable_5cavro_LongType.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8LongType__get_value_fitness;
  __pyx_vtable_5cavro_LongType.__pyx_base.json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8LongType_json_format;
  __pyx_vtable_5cavro_LongType.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8LongType__json_decode;
  __pyx_vtable_5cavro_LongType.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8LongType_canonical_form;
  __pyx_vtable_5cavro_LongType.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_8LongType__get_schema_extra;
  __pyx_vtable_5cavro_LongType.__pyx_base._for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_8LongType__for_writer;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_LongType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_LongType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_LongType)) __PYX_ERR(10, 133, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_LongType_spec, __pyx_ptype_5cavro_LongType) < 0) __PYX_ERR(10, 133, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_LongType = &__pyx_type_5cavro_LongType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_LongType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_LongType) < 0) __PYX_ERR(10, 133, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_LongType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_LongType->tp_dictoffset && __pyx_ptype_5cavro_LongType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_LongType->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_LongType, __pyx_vtabptr_5cavro_LongType) < 0) __PYX_ERR(10, 133, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_LongType) < 0) __PYX_ERR(10, 133, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LongType, (PyObject *) __pyx_ptype_5cavro_LongType) < 0) __PYX_ERR(10, 133, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_LongType) < 0) __PYX_ERR(10, 133, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_IntType = &__pyx_vtable_5cavro_IntType;
  __pyx_vtable_5cavro_IntType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro_IntType.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_7IntType_copy;
  __pyx_vtable_5cavro_IntType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_7IntType__extract_metadata;
  __pyx_vtable_5cavro_IntType.__pyx_base._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_7IntType__convert_value;
  __pyx_vtable_5cavro_IntType.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_7IntType__binary_buffer_encode;
  __pyx_vtable_5cavro_IntType.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_7IntType__binary_buffer_decode;
  __pyx_vtable_5cavro_IntType.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_7IntType__get_value_fitness;
  __pyx_vtable_5cavro_IntType.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_7IntType__json_format;
  __pyx_vtable_5cavro_IntType.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_7IntType__json_decode;
  __pyx_vtable_5cavro_IntType.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_7IntType_canonical_form;
  __pyx_vtable_5cavro_IntType.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_7IntType__get_schema_extra;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_IntType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_IntType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_IntType)) __PYX_ERR(10, 61, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_IntType_spec, __pyx_ptype_5cavro_IntType) < 0) __PYX_ERR(10, 61, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_IntType = &__pyx_type_5cavro_IntType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_IntType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_IntType) < 0) __PYX_ERR(10, 61, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_IntType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_IntType->tp_dictoffset && __pyx_ptype_5cavro_IntType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_IntType->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_IntType, __pyx_vtabptr_5cavro_IntType) < 0) __PYX_ERR(10, 61, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_IntType) < 0) __PYX_ERR(10, 61, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IntType, (PyObject *) __pyx_ptype_5cavro_IntType) < 0) __PYX_ERR(10, 61, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_IntType) < 0) __PYX_ERR(10, 61, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_BoolType = &__pyx_vtable_5cavro_BoolType;
  __pyx_vtable_5cavro_BoolType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro_BoolType.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_8BoolType_copy;
  __pyx_vtable_5cavro_BoolType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8BoolType__extract_metadata;
  __pyx_vtable_5cavro_BoolType.__pyx_base._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_8BoolType__convert_value;
  __pyx_vtable_5cavro_BoolType.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_8BoolType__binary_buffer_encode;
  __pyx_vtable_5cavro_BoolType.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_8BoolType__binary_buffer_decode;
  __pyx_vtable_5cavro_BoolType.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8BoolType__get_value_fitness;
  __pyx_vtable_5cavro_BoolType.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8BoolType__json_format;
  __pyx_vtable_5cavro_BoolType.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8BoolType__json_decode;
  __pyx_vtable_5cavro_BoolType.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8BoolType_canonical_form;
  __pyx_vtable_5cavro_BoolType.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_8BoolType__get_schema_extra;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_BoolType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_BoolType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_BoolType)) __PYX_ERR(10, 11, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_BoolType_spec, __pyx_ptype_5cavro_BoolType) < 0) __PYX_ERR(10, 11, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_BoolType = &__pyx_type_5cavro_BoolType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_BoolType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_BoolType) < 0) __PYX_ERR(10, 11, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_BoolType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_BoolType->tp_dictoffset && __pyx_ptype_5cavro_BoolType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_BoolType->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_BoolType, __pyx_vtabptr_5cavro_BoolType) < 0) __PYX_ERR(10, 11, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_BoolType) < 0) __PYX_ERR(10, 11, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BoolType, (PyObject *) __pyx_ptype_5cavro_BoolType) < 0) __PYX_ERR(10, 11, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_BoolType) < 0) __PYX_ERR(10, 11, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_NullType = &__pyx_vtable_5cavro_NullType;
  __pyx_vtable_5cavro_NullType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro_NullType.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_8NullType_copy;
  __pyx_vtable_5cavro_NullType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8NullType__extract_metadata;
  __pyx_vtable_5cavro_NullType.__pyx_base._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_8NullType__convert_value;
  __pyx_vtable_5cavro_NullType.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_8NullType__binary_buffer_encode;
  __pyx_vtable_5cavro_NullType.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_8NullType__binary_buffer_decode;
  __pyx_vtable_5cavro_NullType.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8NullType__get_value_fitness;
  __pyx_vtable_5cavro_NullType.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8NullType__json_format;
  __pyx_vtable_5cavro_NullType.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8NullType__json_decode;
  __pyx_vtable_5cavro_NullType.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8NullType_canonical_form;
  __pyx_vtable_5cavro_NullType.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_8NullType__get_schema_extra;
  __pyx_vtable_5cavro_NullType.__pyx_base.accepts_missing_value = (int (*)(struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_8NullType_accepts_missing_value;
  __pyx_vtable_5cavro_NullType.__pyx_base.resolve_default_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, PyObject *))__pyx_f_5cavro_8NullType_resolve_default_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_NullType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_NullType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_NullType)) __PYX_ERR(21, 2, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_NullType_spec, __pyx_ptype_5cavro_NullType) < 0) __PYX_ERR(21, 2, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_NullType = &__pyx_type_5cavro_NullType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_NullType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_NullType) < 0) __PYX_ERR(21, 2, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_NullType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_NullType->tp_dictoffset && __pyx_ptype_5cavro_NullType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_NullType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_NullType, __pyx_vtabptr_5cavro_NullType) < 0) __PYX_ERR(21, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_NullType) < 0) __PYX_ERR(21, 2, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_NullType, (PyObject *) __pyx_ptype_5cavro_NullType) < 0) __PYX_ERR(21, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_NullType) < 0) __PYX_ERR(21, 2, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_MapType = &__pyx_vtable_5cavro_MapType;
  __pyx_vtable_5cavro_MapType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro_MapType.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_7MapType_copy;
  __pyx_vtable_5cavro_MapType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_7MapType__extract_metadata;
  __pyx_vtable_5cavro_MapType.__pyx_base.convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_8AvroType_convert_value *__pyx_optional_args))__pyx_f_5cavro_7MapType_convert_value;
  __pyx_vtable_5cavro_MapType.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_7MapType__binary_buffer_encode;
  __pyx_vtable_5cavro_MapType.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_7MapType__binary_buffer_decode;
  __pyx_vtable_5cavro_MapType.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_7MapType__get_value_fitness;
  __pyx_vtable_5cavro_MapType.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_7MapType__json_format;
  __pyx_vtable_5cavro_MapType.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_7MapType__json_decode;
  __pyx_vtable_5cavro_MapType.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_7MapType_canonical_form;
  __pyx_vtable_5cavro_MapType.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_7MapType__get_schema_extra;
  __pyx_vtable_5cavro_MapType.__pyx_base._for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_7MapType__for_writer;
  __pyx_vtable_5cavro_MapType.__pyx_base.accepts_missing_value = (int (*)(struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_7MapType_accepts_missing_value;
  __pyx_vtable_5cavro_MapType._make_converted_map = (PyObject *(*)(struct __pyx_obj_5cavro_MapType *, PyObject *))__pyx_f_5cavro_7MapType__make_converted_map;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_MapType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_MapType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_MapType)) __PYX_ERR(9, 2, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_MapType_spec, __pyx_ptype_5cavro_MapType) < 0) __PYX_ERR(9, 2, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_MapType = &__pyx_type_5cavro_MapType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_MapType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_MapType) < 0) __PYX_ERR(9, 2, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_MapType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_MapType->tp_dictoffset && __pyx_ptype_5cavro_MapType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_MapType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_MapType, __pyx_vtabptr_5cavro_MapType) < 0) __PYX_ERR(9, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_MapType) < 0) __PYX_ERR(9, 2, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MapType, (PyObject *) __pyx_ptype_5cavro_MapType) < 0) __PYX_ERR(9, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_MapType) < 0) __PYX_ERR(9, 2, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_ArrayType = &__pyx_vtable_5cavro_ArrayType;
  __pyx_vtable_5cavro_ArrayType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro_ArrayType.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_9ArrayType_copy;
  __pyx_vtable_5cavro_ArrayType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9ArrayType__extract_metadata;
  __pyx_vtable_5cavro_ArrayType.__pyx_base.convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_8AvroType_convert_value *__pyx_optional_args))__pyx_f_5cavro_9ArrayType_convert_value;
  __pyx_vtable_5cavro_ArrayType.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_9ArrayType__binary_buffer_encode;
  __pyx_vtable_5cavro_ArrayType.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_9ArrayType__binary_buffer_decode;
  __pyx_vtable_5cavro_ArrayType.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9ArrayType__get_value_fitness;
  __pyx_vtable_5cavro_ArrayType.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9ArrayType__json_format;
  __pyx_vtable_5cavro_ArrayType.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9ArrayType__json_decode;
  __pyx_vtable_5cavro_ArrayType.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9ArrayType_canonical_form;
  __pyx_vtable_5cavro_ArrayType.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_9ArrayType__get_schema_extra;
  __pyx_vtable_5cavro_ArrayType.__pyx_base._for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_9ArrayType__for_writer;
  __pyx_vtable_5cavro_ArrayType.__pyx_base.accepts_missing_value = (int (*)(struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_9ArrayType_accepts_missing_value;
  __pyx_vtable_5cavro_ArrayType._make_converted_list = (PyObject *(*)(struct __pyx_obj_5cavro_ArrayType *, PyObject *))__pyx_f_5cavro_9ArrayType__make_converted_list;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_ArrayType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_ArrayType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_ArrayType)) __PYX_ERR(20, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_ArrayType_spec, __pyx_ptype_5cavro_ArrayType) < 0) __PYX_ERR(20, 5, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_ArrayType = &__pyx_type_5cavro_ArrayType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_ArrayType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_ArrayType) < 0) __PYX_ERR(20, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_ArrayType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_ArrayType->tp_dictoffset && __pyx_ptype_5cavro_ArrayType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_ArrayType->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_ArrayType, __pyx_vtabptr_5cavro_ArrayType) < 0) __PYX_ERR(20, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_ArrayType) < 0) __PYX_ERR(20, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ArrayType, (PyObject *) __pyx_ptype_5cavro_ArrayType) < 0) __PYX_ERR(20, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_ArrayType) < 0) __PYX_ERR(20, 5, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro__NamedType = &__pyx_vtable_5cavro__NamedType;
  __pyx_vtable_5cavro__NamedType.__pyx_base = *__pyx_vtabptr_5cavro_AvroType;
  __pyx_vtable_5cavro__NamedType.__pyx_base.clone_base = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_opt_args_5cavro_8AvroType_clone_base *__pyx_optional_args))__pyx_f_5cavro_10_NamedType_clone_base;
  __pyx_vtable_5cavro__NamedType.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10_NamedType__extract_metadata;
  __pyx_vtable_5cavro__NamedType.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_10_NamedType__get_schema_extra;
  __pyx_vtable_5cavro__NamedType.name_matches = (int (*)(struct __pyx_obj_5cavro__NamedType *, struct __pyx_obj_5cavro__NamedType *))__pyx_f_5cavro_10_NamedType_name_matches;
  __pyx_vtable_5cavro__NamedType.get_namespaced_aliases = (PyObject *(*)(struct __pyx_obj_5cavro__NamedType *))__pyx_f_5cavro_10_NamedType_get_namespaced_aliases;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_AvroType); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro__NamedType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__NamedType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro__NamedType)) __PYX_ERR(2, 320, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__NamedType_spec, __pyx_ptype_5cavro__NamedType) < 0) __PYX_ERR(2, 320, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__NamedType = &__pyx_type_5cavro__NamedType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro__NamedType->tp_base = __pyx_ptype_5cavro_AvroType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__NamedType) < 0) __PYX_ERR(2, 320, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__NamedType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__NamedType->tp_dictoffset && __pyx_ptype_5cavro__NamedType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__NamedType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro__NamedType, __pyx_vtabptr_5cavro__NamedType) < 0) __PYX_ERR(2, 320, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro__NamedType) < 0) __PYX_ERR(2, 320, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_NamedType, (PyObject *) __pyx_ptype_5cavro__NamedType) < 0) __PYX_ERR(2, 320, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__NamedType) < 0) __PYX_ERR(2, 320, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_FixedType = &__pyx_vtable_5cavro_FixedType;
  __pyx_vtable_5cavro_FixedType.__pyx_base = *__pyx_vtabptr_5cavro__NamedType;
  __pyx_vtable_5cavro_FixedType.__pyx_base.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_9FixedType_copy;
  __pyx_vtable_5cavro_FixedType.__pyx_base.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9FixedType__extract_metadata;
  __pyx_vtable_5cavro_FixedType.__pyx_base.__pyx_base._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_9FixedType__convert_value;
  __pyx_vtable_5cavro_FixedType.__pyx_base.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_9FixedType__binary_buffer_encode;
  __pyx_vtable_5cavro_FixedType.__pyx_base.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_9FixedType__binary_buffer_decode;
  __pyx_vtable_5cavro_FixedType.__pyx_base.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9FixedType__get_value_fitness;
  __pyx_vtable_5cavro_FixedType.__pyx_base.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9FixedType__json_format;
  __pyx_vtable_5cavro_FixedType.__pyx_base.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9FixedType__json_decode;
  __pyx_vtable_5cavro_FixedType.__pyx_base.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_9FixedType_canonical_form;
  __pyx_vtable_5cavro_FixedType.__pyx_base.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_9FixedType__get_schema_extra;
  __pyx_vtable_5cavro_FixedType.__pyx_base.__pyx_base._for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_9FixedType__for_writer;
  __pyx_vtable_5cavro_FixedType.__pyx_base.__pyx_base.resolve_default_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, PyObject *))__pyx_f_5cavro_9FixedType_resolve_default_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro__NamedType); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_FixedType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_FixedType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_FixedType)) __PYX_ERR(12, 151, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_FixedType_spec, __pyx_ptype_5cavro_FixedType) < 0) __PYX_ERR(12, 151, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_FixedType = &__pyx_type_5cavro_FixedType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_FixedType->tp_base = __pyx_ptype_5cavro__NamedType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_FixedType) < 0) __PYX_ERR(12, 151, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_FixedType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_FixedType->tp_dictoffset && __pyx_ptype_5cavro_FixedType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_FixedType->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_FixedType, __pyx_vtabptr_5cavro_FixedType) < 0) __PYX_ERR(12, 151, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_FixedType) < 0) __PYX_ERR(12, 151, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FixedType, (PyObject *) __pyx_ptype_5cavro_FixedType) < 0) __PYX_ERR(12, 151, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_FixedType) < 0) __PYX_ERR(12, 151, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_RecordType = &__pyx_vtable_5cavro_RecordType;
  __pyx_vtable_5cavro_RecordType.__pyx_base = *__pyx_vtabptr_5cavro__NamedType;
  __pyx_vtable_5cavro_RecordType.__pyx_base.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_10RecordType_copy;
  __pyx_vtable_5cavro_RecordType.__pyx_base.__pyx_base._make_logical = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, PyObject *))__pyx_f_5cavro_10RecordType__make_logical;
  __pyx_vtable_5cavro_RecordType.__pyx_base.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10RecordType__extract_metadata;
  __pyx_vtable_5cavro_RecordType.__pyx_base.__pyx_base._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_10RecordType__convert_value;
  __pyx_vtable_5cavro_RecordType.__pyx_base.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_10RecordType__binary_buffer_encode;
  __pyx_vtable_5cavro_RecordType.__pyx_base.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_10RecordType__binary_buffer_decode;
  __pyx_vtable_5cavro_RecordType.__pyx_base.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10RecordType__get_value_fitness;
  __pyx_vtable_5cavro_RecordType.__pyx_base.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10RecordType__json_format;
  __pyx_vtable_5cavro_RecordType.__pyx_base.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10RecordType__json_decode;
  __pyx_vtable_5cavro_RecordType.__pyx_base.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_10RecordType_canonical_form;
  __pyx_vtable_5cavro_RecordType.__pyx_base.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_10RecordType__get_schema_extra;
  __pyx_vtable_5cavro_RecordType.__pyx_base.__pyx_base._for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_10RecordType__for_writer;
  __pyx_vtable_5cavro_RecordType._binary_buffer_decode_record = (PyObject *(*)(struct __pyx_obj_5cavro_RecordType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_10RecordType__binary_buffer_decode_record;
  __pyx_vtable_5cavro_RecordType._binary_buffer_decode_dict = (PyObject *(*)(struct __pyx_obj_5cavro_RecordType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_10RecordType__binary_buffer_decode_dict;
  __pyx_vtable_5cavro_RecordType.json_decode_record = (PyObject *(*)(struct __pyx_obj_5cavro_RecordType *, PyObject *))__pyx_f_5cavro_10RecordType_json_decode_record;
  __pyx_vtable_5cavro_RecordType.json_decode_dict = (PyObject *(*)(struct __pyx_obj_5cavro_RecordType *, PyObject *))__pyx_f_5cavro_10RecordType_json_decode_dict;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro__NamedType); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_RecordType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_RecordType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_RecordType)) __PYX_ERR(11, 301, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_RecordType_spec, __pyx_ptype_5cavro_RecordType) < 0) __PYX_ERR(11, 301, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_RecordType = &__pyx_type_5cavro_RecordType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_RecordType->tp_base = __pyx_ptype_5cavro__NamedType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_RecordType) < 0) __PYX_ERR(11, 301, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_RecordType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_RecordType->tp_dictoffset && __pyx_ptype_5cavro_RecordType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_RecordType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_RecordType, __pyx_vtabptr_5cavro_RecordType) < 0) __PYX_ERR(11, 301, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_RecordType) < 0) __PYX_ERR(11, 301, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_RecordType, (PyObject *) __pyx_ptype_5cavro_RecordType) < 0) __PYX_ERR(11, 301, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_RecordType) < 0) __PYX_ERR(11, 301, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_PromotingRecordType = &__pyx_vtable_5cavro_PromotingRecordType;
  __pyx_vtable_5cavro_PromotingRecordType.__pyx_base = *__pyx_vtabptr_5cavro_RecordType;
  __pyx_vtable_5cavro_PromotingRecordType.__pyx_base._binary_buffer_decode_record = (PyObject *(*)(struct __pyx_obj_5cavro_RecordType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_19PromotingRecordType__binary_buffer_decode_record;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_RecordType); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_PromotingRecordType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_PromotingRecordType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_PromotingRecordType)) __PYX_ERR(11, 610, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_PromotingRecordType_spec, __pyx_ptype_5cavro_PromotingRecordType) < 0) __PYX_ERR(11, 610, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_PromotingRecordType = &__pyx_type_5cavro_PromotingRecordType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_PromotingRecordType->tp_base = __pyx_ptype_5cavro_RecordType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_PromotingRecordType) < 0) __PYX_ERR(11, 610, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_PromotingRecordType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_PromotingRecordType->tp_dictoffset && __pyx_ptype_5cavro_PromotingRecordType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_PromotingRecordType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_PromotingRecordType, __pyx_vtabptr_5cavro_PromotingRecordType) < 0) __PYX_ERR(11, 610, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_PromotingRecordType) < 0) __PYX_ERR(11, 610, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PromotingRecordType, (PyObject *) __pyx_ptype_5cavro_PromotingRecordType) < 0) __PYX_ERR(11, 610, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_PromotingRecordType) < 0) __PYX_ERR(11, 610, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_EnumType = &__pyx_vtable_5cavro_EnumType;
  __pyx_vtable_5cavro_EnumType.__pyx_base = *__pyx_vtabptr_5cavro__NamedType;
  __pyx_vtable_5cavro_EnumType.__pyx_base.__pyx_base.copy = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, int __pyx_skip_dispatch))__pyx_f_5cavro_8EnumType_copy;
  __pyx_vtable_5cavro_EnumType.__pyx_base.__pyx_base._extract_metadata = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8EnumType__extract_metadata;
  __pyx_vtable_5cavro_EnumType.__pyx_base.__pyx_base._convert_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_8EnumType__convert_value;
  __pyx_vtable_5cavro_EnumType.__pyx_base.__pyx_base._binary_buffer_encode = (int (*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Writer *, PyObject *))__pyx_f_5cavro_8EnumType__binary_buffer_encode;
  __pyx_vtable_5cavro_EnumType.__pyx_base.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_8EnumType__binary_buffer_decode;
  __pyx_vtable_5cavro_EnumType.__pyx_base.__pyx_base._get_value_fitness = (int (*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8EnumType__get_value_fitness;
  __pyx_vtable_5cavro_EnumType.__pyx_base.__pyx_base._json_format = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8EnumType__json_format;
  __pyx_vtable_5cavro_EnumType.__pyx_base.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8EnumType__json_decode;
  __pyx_vtable_5cavro_EnumType.__pyx_base.__pyx_base.canonical_form = (struct __pyx_obj_5cavro__CanonicalForm *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_8EnumType_canonical_form;
  __pyx_vtable_5cavro_EnumType.__pyx_base.__pyx_base._get_schema_extra = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_8EnumType__get_schema_extra;
  __pyx_vtable_5cavro_EnumType.__pyx_base.__pyx_base._for_writer = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_8EnumType__for_writer;
  __pyx_vtable_5cavro_EnumType.__pyx_base.__pyx_base.resolve_default_value = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *, PyObject *))__pyx_f_5cavro_8EnumType_resolve_default_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro__NamedType); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_EnumType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_EnumType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_EnumType)) __PYX_ERR(8, 2, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_EnumType_spec, __pyx_ptype_5cavro_EnumType) < 0) __PYX_ERR(8, 2, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_EnumType = &__pyx_type_5cavro_EnumType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_EnumType->tp_base = __pyx_ptype_5cavro__NamedType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_EnumType) < 0) __PYX_ERR(8, 2, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_EnumType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_EnumType->tp_dictoffset && __pyx_ptype_5cavro_EnumType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_EnumType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_EnumType, __pyx_vtabptr_5cavro_EnumType) < 0) __PYX_ERR(8, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_EnumType) < 0) __PYX_ERR(8, 2, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_EnumType, (PyObject *) __pyx_ptype_5cavro_EnumType) < 0) __PYX_ERR(8, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_EnumType) < 0) __PYX_ERR(8, 2, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_PromotingEnumType = &__pyx_vtable_5cavro_PromotingEnumType;
  __pyx_vtable_5cavro_PromotingEnumType.__pyx_base = *__pyx_vtabptr_5cavro_EnumType;
  __pyx_vtable_5cavro_PromotingEnumType.__pyx_base.__pyx_base.__pyx_base._binary_buffer_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, struct __pyx_obj_5cavro__Reader *))__pyx_f_5cavro_17PromotingEnumType__binary_buffer_decode;
  __pyx_vtable_5cavro_PromotingEnumType.__pyx_base.__pyx_base.__pyx_base._json_decode = (PyObject *(*)(struct __pyx_obj_5cavro_AvroType *, PyObject *))__pyx_f_5cavro_17PromotingEnumType__json_decode;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_EnumType); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_PromotingEnumType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_PromotingEnumType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_PromotingEnumType)) __PYX_ERR(8, 136, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_PromotingEnumType_spec, __pyx_ptype_5cavro_PromotingEnumType) < 0) __PYX_ERR(8, 136, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_PromotingEnumType = &__pyx_type_5cavro_PromotingEnumType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_PromotingEnumType->tp_base = __pyx_ptype_5cavro_EnumType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_PromotingEnumType) < 0) __PYX_ERR(8, 136, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_PromotingEnumType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_PromotingEnumType->tp_dictoffset && __pyx_ptype_5cavro_PromotingEnumType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_PromotingEnumType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_PromotingEnumType, __pyx_vtabptr_5cavro_PromotingEnumType) < 0) __PYX_ERR(8, 136, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_PromotingEnumType) < 0) __PYX_ERR(8, 136, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PromotingEnumType, (PyObject *) __pyx_ptype_5cavro_PromotingEnumType) < 0) __PYX_ERR(8, 136, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_PromotingEnumType) < 0) __PYX_ERR(8, 136, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_Schema = &__pyx_vtable_5cavro_Schema;
  __pyx_vtable_5cavro_Schema.register_type = (void (*)(struct __pyx_obj_5cavro_Schema *, PyObject *, PyObject *, struct __pyx_obj_5cavro_AvroType *))__pyx_f_5cavro_6Schema_register_type;
  __pyx_vtable_5cavro_Schema.find_type = (struct __pyx_obj_5cavro_AvroType *(*)(struct __pyx_obj_5cavro_Schema *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5cavro_6Schema_find_type *__pyx_optional_args))__pyx_f_5cavro_6Schema_find_type;
  __pyx_vtable_5cavro_Schema.binary_read = (PyObject *(*)(struct __pyx_obj_5cavro_Schema *, struct __pyx_obj_5cavro__Reader *, int __pyx_skip_dispatch))__pyx_f_5cavro_6Schema_binary_read;
  __pyx_vtable_5cavro_Schema.binary_write = (PyObject *(*)(struct __pyx_obj_5cavro_Schema *, struct __pyx_obj_5cavro__Writer *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_6Schema_binary_write;
  __pyx_vtable_5cavro_Schema.reader_for_writer = (struct __pyx_obj_5cavro_Schema *(*)(struct __pyx_obj_5cavro_Schema *, struct __pyx_obj_5cavro_Schema *, int __pyx_skip_dispatch))__pyx_f_5cavro_6Schema_reader_for_writer;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro_Schema = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_Schema_spec, NULL); if (unlikely(!__pyx_ptype_5cavro_Schema)) __PYX_ERR(23, 18, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_Schema_spec, __pyx_ptype_5cavro_Schema) < 0) __PYX_ERR(23, 18, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_Schema = &__pyx_type_5cavro_Schema;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_Schema) < 0) __PYX_ERR(23, 18, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_Schema->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_Schema->tp_dictoffset && __pyx_ptype_5cavro_Schema->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_Schema->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_Schema, __pyx_vtabptr_5cavro_Schema) < 0) __PYX_ERR(23, 18, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_Schema) < 0) __PYX_ERR(23, 18, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Schema, (PyObject *) __pyx_ptype_5cavro_Schema) < 0) __PYX_ERR(23, 18, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_Schema) < 0) __PYX_ERR(23, 18, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_ResolvedSchema = &__pyx_vtable_5cavro_ResolvedSchema;
  __pyx_vtable_5cavro_ResolvedSchema.__pyx_base = *__pyx_vtabptr_5cavro_Schema;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_Schema); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_ResolvedSchema = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_ResolvedSchema_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_ResolvedSchema)) __PYX_ERR(23, 226, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_ResolvedSchema_spec, __pyx_ptype_5cavro_ResolvedSchema) < 0) __PYX_ERR(23, 226, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_ResolvedSchema = &__pyx_type_5cavro_ResolvedSchema;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_ResolvedSchema->tp_base = __pyx_ptype_5cavro_Schema;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_ResolvedSchema) < 0) __PYX_ERR(23, 226, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_ResolvedSchema->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_ResolvedSchema->tp_dictoffset && __pyx_ptype_5cavro_ResolvedSchema->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_ResolvedSchema->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_ResolvedSchema, __pyx_vtabptr_5cavro_ResolvedSchema) < 0) __PYX_ERR(23, 226, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_ResolvedSchema) < 0) __PYX_ERR(23, 226, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ResolvedSchema, (PyObject *) __pyx_ptype_5cavro_ResolvedSchema) < 0) __PYX_ERR(23, 226, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_ResolvedSchema) < 0) __PYX_ERR(23, 226, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_ValueAdapter = &__pyx_vtable_5cavro_ValueAdapter;
  __pyx_vtable_5cavro_ValueAdapter.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_12ValueAdapter_encode_value;
  __pyx_vtable_5cavro_ValueAdapter.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_12ValueAdapter_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro_ValueAdapter = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_ValueAdapter_spec, NULL); if (unlikely(!__pyx_ptype_5cavro_ValueAdapter)) __PYX_ERR(24, 1, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_ValueAdapter_spec, __pyx_ptype_5cavro_ValueAdapter) < 0) __PYX_ERR(24, 1, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_ValueAdapter = &__pyx_type_5cavro_ValueAdapter;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_ValueAdapter) < 0) __PYX_ERR(24, 1, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_ValueAdapter->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_ValueAdapter->tp_dictoffset && __pyx_ptype_5cavro_ValueAdapter->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_ValueAdapter->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_ValueAdapter, __pyx_vtabptr_5cavro_ValueAdapter) < 0) __PYX_ERR(24, 1, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_ValueAdapter) < 0) __PYX_ERR(24, 1, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ValueAdapter, (PyObject *) __pyx_ptype_5cavro_ValueAdapter) < 0) __PYX_ERR(24, 1, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_ValueAdapter) < 0) __PYX_ERR(24, 1, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_LogicalType = &__pyx_vtable_5cavro_LogicalType;
  __pyx_vtable_5cavro_LogicalType.__pyx_base = *__pyx_vtabptr_5cavro_ValueAdapter;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_ValueAdapter); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_LogicalType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_LogicalType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_LogicalType)) __PYX_ERR(24, 23, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_LogicalType_spec, __pyx_ptype_5cavro_LogicalType) < 0) __PYX_ERR(24, 23, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_LogicalType = &__pyx_type_5cavro_LogicalType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_LogicalType->tp_base = __pyx_ptype_5cavro_ValueAdapter;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_LogicalType) < 0) __PYX_ERR(24, 23, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_LogicalType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_LogicalType->tp_dictoffset && __pyx_ptype_5cavro_LogicalType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_LogicalType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_LogicalType, __pyx_vtabptr_5cavro_LogicalType) < 0) __PYX_ERR(24, 23, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_LogicalType) < 0) __PYX_ERR(24, 23, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LogicalType, (PyObject *) __pyx_ptype_5cavro_LogicalType) < 0) __PYX_ERR(24, 23, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_LogicalType) < 0) __PYX_ERR(24, 23, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_CustomLogicalType = &__pyx_vtable_5cavro_CustomLogicalType;
  __pyx_vtable_5cavro_CustomLogicalType.__pyx_base = *__pyx_vtabptr_5cavro_LogicalType;
  __pyx_vtable_5cavro_CustomLogicalType.__pyx_base.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_17CustomLogicalType_encode_value;
  __pyx_vtable_5cavro_CustomLogicalType.__pyx_base.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_17CustomLogicalType_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_LogicalType); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_CustomLogicalType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_CustomLogicalType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_CustomLogicalType)) __PYX_ERR(24, 47, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_CustomLogicalType_spec, __pyx_ptype_5cavro_CustomLogicalType) < 0) __PYX_ERR(24, 47, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_CustomLogicalType = &__pyx_type_5cavro_CustomLogicalType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_CustomLogicalType->tp_base = __pyx_ptype_5cavro_LogicalType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_CustomLogicalType) < 0) __PYX_ERR(24, 47, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_CustomLogicalType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_CustomLogicalType->tp_dictoffset && __pyx_ptype_5cavro_CustomLogicalType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_CustomLogicalType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_CustomLogicalType, __pyx_vtabptr_5cavro_CustomLogicalType) < 0) __PYX_ERR(24, 47, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_CustomLogicalType) < 0) __PYX_ERR(24, 47, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CustomLogicalType, (PyObject *) __pyx_ptype_5cavro_CustomLogicalType) < 0) __PYX_ERR(24, 47, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_CustomLogicalType) < 0) __PYX_ERR(24, 47, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_DecimalType = &__pyx_vtable_5cavro_DecimalType;
  __pyx_vtable_5cavro_DecimalType.__pyx_base = *__pyx_vtabptr_5cavro_LogicalType;
  __pyx_vtable_5cavro_DecimalType.__pyx_base.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_11DecimalType_encode_value;
  __pyx_vtable_5cavro_DecimalType.__pyx_base.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_11DecimalType_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_LogicalType); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_DecimalType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_DecimalType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_DecimalType)) __PYX_ERR(24, 72, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_DecimalType_spec, __pyx_ptype_5cavro_DecimalType) < 0) __PYX_ERR(24, 72, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_DecimalType = &__pyx_type_5cavro_DecimalType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_DecimalType->tp_base = __pyx_ptype_5cavro_LogicalType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_DecimalType) < 0) __PYX_ERR(24, 72, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_DecimalType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_DecimalType->tp_dictoffset && __pyx_ptype_5cavro_DecimalType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_DecimalType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_DecimalType, __pyx_vtabptr_5cavro_DecimalType) < 0) __PYX_ERR(24, 72, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_DecimalType) < 0) __PYX_ERR(24, 72, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DecimalType, (PyObject *) __pyx_ptype_5cavro_DecimalType) < 0) __PYX_ERR(24, 72, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_DecimalType) < 0) __PYX_ERR(24, 72, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_UUIDBase = &__pyx_vtable_5cavro_UUIDBase;
  __pyx_vtable_5cavro_UUIDBase.__pyx_base = *__pyx_vtabptr_5cavro_LogicalType;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_LogicalType); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_UUIDBase = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_UUIDBase_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_UUIDBase)) __PYX_ERR(24, 167, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_UUIDBase_spec, __pyx_ptype_5cavro_UUIDBase) < 0) __PYX_ERR(24, 167, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_UUIDBase = &__pyx_type_5cavro_UUIDBase;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_UUIDBase->tp_base = __pyx_ptype_5cavro_LogicalType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_UUIDBase) < 0) __PYX_ERR(24, 167, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_UUIDBase->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_UUIDBase->tp_dictoffset && __pyx_ptype_5cavro_UUIDBase->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_UUIDBase->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_UUIDBase, __pyx_vtabptr_5cavro_UUIDBase) < 0) __PYX_ERR(24, 167, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_UUIDBase) < 0) __PYX_ERR(24, 167, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UUIDBase, (PyObject *) __pyx_ptype_5cavro_UUIDBase) < 0) __PYX_ERR(24, 167, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_UUIDBase) < 0) __PYX_ERR(24, 167, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_UUIDStringType = &__pyx_vtable_5cavro_UUIDStringType;
  __pyx_vtable_5cavro_UUIDStringType.__pyx_base = *__pyx_vtabptr_5cavro_UUIDBase;
  __pyx_vtable_5cavro_UUIDStringType.__pyx_base.__pyx_base.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_14UUIDStringType_encode_value;
  __pyx_vtable_5cavro_UUIDStringType.__pyx_base.__pyx_base.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_14UUIDStringType_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_UUIDBase); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_UUIDStringType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_UUIDStringType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_UUIDStringType)) __PYX_ERR(24, 180, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_UUIDStringType_spec, __pyx_ptype_5cavro_UUIDStringType) < 0) __PYX_ERR(24, 180, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_UUIDStringType = &__pyx_type_5cavro_UUIDStringType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_UUIDStringType->tp_base = __pyx_ptype_5cavro_UUIDBase;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_UUIDStringType) < 0) __PYX_ERR(24, 180, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_UUIDStringType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_UUIDStringType->tp_dictoffset && __pyx_ptype_5cavro_UUIDStringType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_UUIDStringType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_UUIDStringType, __pyx_vtabptr_5cavro_UUIDStringType) < 0) __PYX_ERR(24, 180, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_UUIDStringType) < 0) __PYX_ERR(24, 180, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UUIDStringType, (PyObject *) __pyx_ptype_5cavro_UUIDStringType) < 0) __PYX_ERR(24, 180, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_UUIDStringType) < 0) __PYX_ERR(24, 180, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_UUIDFixedType = &__pyx_vtable_5cavro_UUIDFixedType;
  __pyx_vtable_5cavro_UUIDFixedType.__pyx_base = *__pyx_vtabptr_5cavro_UUIDBase;
  __pyx_vtable_5cavro_UUIDFixedType.__pyx_base.__pyx_base.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_13UUIDFixedType_encode_value;
  __pyx_vtable_5cavro_UUIDFixedType.__pyx_base.__pyx_base.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_13UUIDFixedType_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_UUIDBase); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_UUIDFixedType = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_UUIDFixedType_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_UUIDFixedType)) __PYX_ERR(24, 209, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_UUIDFixedType_spec, __pyx_ptype_5cavro_UUIDFixedType) < 0) __PYX_ERR(24, 209, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_UUIDFixedType = &__pyx_type_5cavro_UUIDFixedType;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_UUIDFixedType->tp_base = __pyx_ptype_5cavro_UUIDBase;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_UUIDFixedType) < 0) __PYX_ERR(24, 209, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_UUIDFixedType->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_UUIDFixedType->tp_dictoffset && __pyx_ptype_5cavro_UUIDFixedType->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_UUIDFixedType->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_UUIDFixedType, __pyx_vtabptr_5cavro_UUIDFixedType) < 0) __PYX_ERR(24, 209, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_UUIDFixedType) < 0) __PYX_ERR(24, 209, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UUIDFixedType, (PyObject *) __pyx_ptype_5cavro_UUIDFixedType) < 0) __PYX_ERR(24, 209, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_UUIDFixedType) < 0) __PYX_ERR(24, 209, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_Date = &__pyx_vtable_5cavro_Date;
  __pyx_vtable_5cavro_Date.__pyx_base = *__pyx_vtabptr_5cavro_LogicalType;
  __pyx_vtable_5cavro_Date.__pyx_base.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_4Date_encode_value;
  __pyx_vtable_5cavro_Date.__pyx_base.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_4Date_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_LogicalType); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_Date = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_Date_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_Date)) __PYX_ERR(24, 237, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_Date_spec, __pyx_ptype_5cavro_Date) < 0) __PYX_ERR(24, 237, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_Date = &__pyx_type_5cavro_Date;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_Date->tp_base = __pyx_ptype_5cavro_LogicalType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_Date) < 0) __PYX_ERR(24, 237, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_Date->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_Date->tp_dictoffset && __pyx_ptype_5cavro_Date->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_Date->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_Date, __pyx_vtabptr_5cavro_Date) < 0) __PYX_ERR(24, 237, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_Date) < 0) __PYX_ERR(24, 237, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Date, (PyObject *) __pyx_ptype_5cavro_Date) < 0) __PYX_ERR(24, 237, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_Date) < 0) __PYX_ERR(24, 237, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_TimeMillis = &__pyx_vtable_5cavro_TimeMillis;
  __pyx_vtable_5cavro_TimeMillis.__pyx_base = *__pyx_vtabptr_5cavro_LogicalType;
  __pyx_vtable_5cavro_TimeMillis.__pyx_base.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_10TimeMillis_encode_value;
  __pyx_vtable_5cavro_TimeMillis.__pyx_base.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_10TimeMillis_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_LogicalType); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_TimeMillis = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_TimeMillis_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_TimeMillis)) __PYX_ERR(24, 292, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_TimeMillis_spec, __pyx_ptype_5cavro_TimeMillis) < 0) __PYX_ERR(24, 292, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_TimeMillis = &__pyx_type_5cavro_TimeMillis;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_TimeMillis->tp_base = __pyx_ptype_5cavro_LogicalType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_TimeMillis) < 0) __PYX_ERR(24, 292, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_TimeMillis->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_TimeMillis->tp_dictoffset && __pyx_ptype_5cavro_TimeMillis->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_TimeMillis->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_TimeMillis, __pyx_vtabptr_5cavro_TimeMillis) < 0) __PYX_ERR(24, 292, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_TimeMillis) < 0) __PYX_ERR(24, 292, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TimeMillis, (PyObject *) __pyx_ptype_5cavro_TimeMillis) < 0) __PYX_ERR(24, 292, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_TimeMillis) < 0) __PYX_ERR(24, 292, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_TimeMicros = &__pyx_vtable_5cavro_TimeMicros;
  __pyx_vtable_5cavro_TimeMicros.__pyx_base = *__pyx_vtabptr_5cavro_LogicalType;
  __pyx_vtable_5cavro_TimeMicros.__pyx_base.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_10TimeMicros_encode_value;
  __pyx_vtable_5cavro_TimeMicros.__pyx_base.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_10TimeMicros_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_LogicalType); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_TimeMicros = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_TimeMicros_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_TimeMicros)) __PYX_ERR(24, 316, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_TimeMicros_spec, __pyx_ptype_5cavro_TimeMicros) < 0) __PYX_ERR(24, 316, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_TimeMicros = &__pyx_type_5cavro_TimeMicros;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_TimeMicros->tp_base = __pyx_ptype_5cavro_LogicalType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_TimeMicros) < 0) __PYX_ERR(24, 316, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_TimeMicros->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_TimeMicros->tp_dictoffset && __pyx_ptype_5cavro_TimeMicros->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_TimeMicros->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_TimeMicros, __pyx_vtabptr_5cavro_TimeMicros) < 0) __PYX_ERR(24, 316, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_TimeMicros) < 0) __PYX_ERR(24, 316, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TimeMicros, (PyObject *) __pyx_ptype_5cavro_TimeMicros) < 0) __PYX_ERR(24, 316, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_TimeMicros) < 0) __PYX_ERR(24, 316, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_TimestampMillis = &__pyx_vtable_5cavro_TimestampMillis;
  __pyx_vtable_5cavro_TimestampMillis.__pyx_base = *__pyx_vtabptr_5cavro_LogicalType;
  __pyx_vtable_5cavro_TimestampMillis.__pyx_base.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_15TimestampMillis_encode_value;
  __pyx_vtable_5cavro_TimestampMillis.__pyx_base.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_15TimestampMillis_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_LogicalType); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_TimestampMillis = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_TimestampMillis_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_TimestampMillis)) __PYX_ERR(24, 340, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_TimestampMillis_spec, __pyx_ptype_5cavro_TimestampMillis) < 0) __PYX_ERR(24, 340, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_TimestampMillis = &__pyx_type_5cavro_TimestampMillis;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_TimestampMillis->tp_base = __pyx_ptype_5cavro_LogicalType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_TimestampMillis) < 0) __PYX_ERR(24, 340, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_TimestampMillis->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_TimestampMillis->tp_dictoffset && __pyx_ptype_5cavro_TimestampMillis->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_TimestampMillis->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_TimestampMillis, __pyx_vtabptr_5cavro_TimestampMillis) < 0) __PYX_ERR(24, 340, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_TimestampMillis) < 0) __PYX_ERR(24, 340, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TimestampMillis, (PyObject *) __pyx_ptype_5cavro_TimestampMillis) < 0) __PYX_ERR(24, 340, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_TimestampMillis) < 0) __PYX_ERR(24, 340, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_TimestampMicros = &__pyx_vtable_5cavro_TimestampMicros;
  __pyx_vtable_5cavro_TimestampMicros.__pyx_base = *__pyx_vtabptr_5cavro_LogicalType;
  __pyx_vtable_5cavro_TimestampMicros.__pyx_base.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_15TimestampMicros_encode_value;
  __pyx_vtable_5cavro_TimestampMicros.__pyx_base.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_15TimestampMicros_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_LogicalType); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_TimestampMicros = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_TimestampMicros_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_TimestampMicros)) __PYX_ERR(24, 375, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_TimestampMicros_spec, __pyx_ptype_5cavro_TimestampMicros) < 0) __PYX_ERR(24, 375, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_TimestampMicros = &__pyx_type_5cavro_TimestampMicros;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_TimestampMicros->tp_base = __pyx_ptype_5cavro_LogicalType;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_TimestampMicros) < 0) __PYX_ERR(24, 375, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_TimestampMicros->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_TimestampMicros->tp_dictoffset && __pyx_ptype_5cavro_TimestampMicros->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_TimestampMicros->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_TimestampMicros, __pyx_vtabptr_5cavro_TimestampMicros) < 0) __PYX_ERR(24, 375, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_TimestampMicros) < 0) __PYX_ERR(24, 375, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TimestampMicros, (PyObject *) __pyx_ptype_5cavro_TimestampMicros) < 0) __PYX_ERR(24, 375, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_TimestampMicros) < 0) __PYX_ERR(24, 375, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_Codec = &__pyx_vtable_5cavro_Codec;
  __pyx_vtable_5cavro_Codec.read_block = (__Pyx_memviewslice (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Reader *, size_t))__pyx_f_5cavro_5Codec_read_block;
  __pyx_vtable_5cavro_Codec.write_block = (Py_ssize_t (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice))__pyx_f_5cavro_5Codec_write_block;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro_Codec = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_Codec_spec, NULL); if (unlikely(!__pyx_ptype_5cavro_Codec)) __PYX_ERR(3, 50, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_Codec_spec, __pyx_ptype_5cavro_Codec) < 0) __PYX_ERR(3, 50, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_Codec = &__pyx_type_5cavro_Codec;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_Codec) < 0) __PYX_ERR(3, 50, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_Codec->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_Codec->tp_dictoffset && __pyx_ptype_5cavro_Codec->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_Codec->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_Codec, __pyx_vtabptr_5cavro_Codec) < 0) __PYX_ERR(3, 50, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_Codec) < 0) __PYX_ERR(3, 50, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Codec, (PyObject *) __pyx_ptype_5cavro_Codec) < 0) __PYX_ERR(3, 50, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_Codec) < 0) __PYX_ERR(3, 50, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro__SnappyCodec = &__pyx_vtable_5cavro__SnappyCodec;
  __pyx_vtable_5cavro__SnappyCodec.__pyx_base = *__pyx_vtabptr_5cavro_Codec;
  __pyx_vtable_5cavro__SnappyCodec.__pyx_base.read_block = (__Pyx_memviewslice (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Reader *, size_t))__pyx_f_5cavro_12_SnappyCodec_read_block;
  __pyx_vtable_5cavro__SnappyCodec.__pyx_base.write_block = (Py_ssize_t (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice))__pyx_f_5cavro_12_SnappyCodec_write_block;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_Codec); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro__SnappyCodec = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__SnappyCodec_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro__SnappyCodec)) __PYX_ERR(3, 70, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__SnappyCodec_spec, __pyx_ptype_5cavro__SnappyCodec) < 0) __PYX_ERR(3, 70, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__SnappyCodec = &__pyx_type_5cavro__SnappyCodec;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro__SnappyCodec->tp_base = __pyx_ptype_5cavro_Codec;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__SnappyCodec) < 0) __PYX_ERR(3, 70, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__SnappyCodec->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__SnappyCodec->tp_dictoffset && __pyx_ptype_5cavro__SnappyCodec->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__SnappyCodec->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro__SnappyCodec, __pyx_vtabptr_5cavro__SnappyCodec) < 0) __PYX_ERR(3, 70, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro__SnappyCodec) < 0) __PYX_ERR(3, 70, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SnappyCodec, (PyObject *) __pyx_ptype_5cavro__SnappyCodec) < 0) __PYX_ERR(3, 70, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__SnappyCodec) < 0) __PYX_ERR(3, 70, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro__NullCodec = &__pyx_vtable_5cavro__NullCodec;
  __pyx_vtable_5cavro__NullCodec.__pyx_base = *__pyx_vtabptr_5cavro_Codec;
  __pyx_vtable_5cavro__NullCodec.__pyx_base.read_block = (__Pyx_memviewslice (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Reader *, size_t))__pyx_f_5cavro_10_NullCodec_read_block;
  __pyx_vtable_5cavro__NullCodec.__pyx_base.write_block = (Py_ssize_t (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice))__pyx_f_5cavro_10_NullCodec_write_block;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_Codec); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro__NullCodec = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__NullCodec_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro__NullCodec)) __PYX_ERR(3, 93, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__NullCodec_spec, __pyx_ptype_5cavro__NullCodec) < 0) __PYX_ERR(3, 93, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__NullCodec = &__pyx_type_5cavro__NullCodec;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro__NullCodec->tp_base = __pyx_ptype_5cavro_Codec;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__NullCodec) < 0) __PYX_ERR(3, 93, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__NullCodec->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__NullCodec->tp_dictoffset && __pyx_ptype_5cavro__NullCodec->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__NullCodec->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro__NullCodec, __pyx_vtabptr_5cavro__NullCodec) < 0) __PYX_ERR(3, 93, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro__NullCodec) < 0) __PYX_ERR(3, 93, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_NullCodec, (PyObject *) __pyx_ptype_5cavro__NullCodec) < 0) __PYX_ERR(3, 93, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__NullCodec) < 0) __PYX_ERR(3, 93, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro__DeflateCodec = &__pyx_vtable_5cavro__DeflateCodec;
  __pyx_vtable_5cavro__DeflateCodec.__pyx_base = *__pyx_vtabptr_5cavro_Codec;
  __pyx_vtable_5cavro__DeflateCodec.__pyx_base.read_block = (__Pyx_memviewslice (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Reader *, size_t))__pyx_f_5cavro_13_DeflateCodec_read_block;
  __pyx_vtable_5cavro__DeflateCodec.__pyx_base.write_block = (Py_ssize_t (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice))__pyx_f_5cavro_13_DeflateCodec_write_block;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_Codec); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro__DeflateCodec = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__DeflateCodec_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro__DeflateCodec)) __PYX_ERR(3, 107, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__DeflateCodec_spec, __pyx_ptype_5cavro__DeflateCodec) < 0) __PYX_ERR(3, 107, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__DeflateCodec = &__pyx_type_5cavro__DeflateCodec;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro__DeflateCodec->tp_base = __pyx_ptype_5cavro_Codec;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__DeflateCodec) < 0) __PYX_ERR(3, 107, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__DeflateCodec->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__DeflateCodec->tp_dictoffset && __pyx_ptype_5cavro__DeflateCodec->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__DeflateCodec->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro__DeflateCodec, __pyx_vtabptr_5cavro__DeflateCodec) < 0) __PYX_ERR(3, 107, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro__DeflateCodec) < 0) __PYX_ERR(3, 107, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DeflateCodec, (PyObject *) __pyx_ptype_5cavro__DeflateCodec) < 0) __PYX_ERR(3, 107, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__DeflateCodec) < 0) __PYX_ERR(3, 107, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro___Bzip2Codec = &__pyx_vtable_5cavro___Bzip2Codec;
  __pyx_vtable_5cavro___Bzip2Codec.__pyx_base = *__pyx_vtabptr_5cavro_Codec;
  __pyx_vtable_5cavro___Bzip2Codec.__pyx_base.read_block = (__Pyx_memviewslice (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Reader *, size_t))__pyx_f_5cavro_12__Bzip2Codec_read_block;
  __pyx_vtable_5cavro___Bzip2Codec.__pyx_base.write_block = (Py_ssize_t (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice))__pyx_f_5cavro_12__Bzip2Codec_write_block;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_Codec); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro___Bzip2Codec = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___Bzip2Codec_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro___Bzip2Codec)) __PYX_ERR(3, 122, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___Bzip2Codec_spec, __pyx_ptype_5cavro___Bzip2Codec) < 0) __PYX_ERR(3, 122, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___Bzip2Codec = &__pyx_type_5cavro___Bzip2Codec;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro___Bzip2Codec->tp_base = __pyx_ptype_5cavro_Codec;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___Bzip2Codec) < 0) __PYX_ERR(3, 122, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___Bzip2Codec->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___Bzip2Codec->tp_dictoffset && __pyx_ptype_5cavro___Bzip2Codec->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___Bzip2Codec->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro___Bzip2Codec, __pyx_vtabptr_5cavro___Bzip2Codec) < 0) __PYX_ERR(3, 122, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro___Bzip2Codec) < 0) __PYX_ERR(3, 122, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Bzip2Codec, (PyObject *) __pyx_ptype_5cavro___Bzip2Codec) < 0) __PYX_ERR(3, 122, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro___Bzip2Codec) < 0) __PYX_ERR(3, 122, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro__LzmaCodec = &__pyx_vtable_5cavro__LzmaCodec;
  __pyx_vtable_5cavro__LzmaCodec.__pyx_base = *__pyx_vtabptr_5cavro_Codec;
  __pyx_vtable_5cavro__LzmaCodec.__pyx_base.read_block = (__Pyx_memviewslice (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Reader *, size_t))__pyx_f_5cavro_10_LzmaCodec_read_block;
  __pyx_vtable_5cavro__LzmaCodec.__pyx_base.write_block = (Py_ssize_t (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice))__pyx_f_5cavro_10_LzmaCodec_write_block;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_Codec); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro__LzmaCodec = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__LzmaCodec_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro__LzmaCodec)) __PYX_ERR(3, 137, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__LzmaCodec_spec, __pyx_ptype_5cavro__LzmaCodec) < 0) __PYX_ERR(3, 137, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__LzmaCodec = &__pyx_type_5cavro__LzmaCodec;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro__LzmaCodec->tp_base = __pyx_ptype_5cavro_Codec;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__LzmaCodec) < 0) __PYX_ERR(3, 137, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__LzmaCodec->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__LzmaCodec->tp_dictoffset && __pyx_ptype_5cavro__LzmaCodec->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__LzmaCodec->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro__LzmaCodec, __pyx_vtabptr_5cavro__LzmaCodec) < 0) __PYX_ERR(3, 137, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro__LzmaCodec) < 0) __PYX_ERR(3, 137, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LzmaCodec, (PyObject *) __pyx_ptype_5cavro__LzmaCodec) < 0) __PYX_ERR(3, 137, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__LzmaCodec) < 0) __PYX_ERR(3, 137, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro__ZStandardCodec = &__pyx_vtable_5cavro__ZStandardCodec;
  __pyx_vtable_5cavro__ZStandardCodec.__pyx_base = *__pyx_vtabptr_5cavro_Codec;
  __pyx_vtable_5cavro__ZStandardCodec.__pyx_base.read_block = (__Pyx_memviewslice (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Reader *, size_t))__pyx_f_5cavro_15_ZStandardCodec_read_block;
  __pyx_vtable_5cavro__ZStandardCodec.__pyx_base.write_block = (Py_ssize_t (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice))__pyx_f_5cavro_15_ZStandardCodec_write_block;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_Codec); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro__ZStandardCodec = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__ZStandardCodec_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro__ZStandardCodec)) __PYX_ERR(3, 152, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__ZStandardCodec_spec, __pyx_ptype_5cavro__ZStandardCodec) < 0) __PYX_ERR(3, 152, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__ZStandardCodec = &__pyx_type_5cavro__ZStandardCodec;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro__ZStandardCodec->tp_base = __pyx_ptype_5cavro_Codec;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__ZStandardCodec) < 0) __PYX_ERR(3, 152, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__ZStandardCodec->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__ZStandardCodec->tp_dictoffset && __pyx_ptype_5cavro__ZStandardCodec->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__ZStandardCodec->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro__ZStandardCodec, __pyx_vtabptr_5cavro__ZStandardCodec) < 0) __PYX_ERR(3, 152, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro__ZStandardCodec) < 0) __PYX_ERR(3, 152, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ZStandardCodec, (PyObject *) __pyx_ptype_5cavro__ZStandardCodec) < 0) __PYX_ERR(3, 152, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__ZStandardCodec) < 0) __PYX_ERR(3, 152, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro__Lz4Codec = &__pyx_vtable_5cavro__Lz4Codec;
  __pyx_vtable_5cavro__Lz4Codec.__pyx_base = *__pyx_vtabptr_5cavro_Codec;
  __pyx_vtable_5cavro__Lz4Codec.__pyx_base.read_block = (__Pyx_memviewslice (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Reader *, size_t))__pyx_f_5cavro_9_Lz4Codec_read_block;
  __pyx_vtable_5cavro__Lz4Codec.__pyx_base.write_block = (Py_ssize_t (*)(struct __pyx_obj_5cavro_Codec *, struct __pyx_obj_5cavro__Writer *, __Pyx_memviewslice))__pyx_f_5cavro_9_Lz4Codec_write_block;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_Codec); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro__Lz4Codec = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro__Lz4Codec_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro__Lz4Codec)) __PYX_ERR(3, 175, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro__Lz4Codec_spec, __pyx_ptype_5cavro__Lz4Codec) < 0) __PYX_ERR(3, 175, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro__Lz4Codec = &__pyx_type_5cavro__Lz4Codec;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro__Lz4Codec->tp_base = __pyx_ptype_5cavro_Codec;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro__Lz4Codec) < 0) __PYX_ERR(3, 175, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro__Lz4Codec->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro__Lz4Codec->tp_dictoffset && __pyx_ptype_5cavro__Lz4Codec->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro__Lz4Codec->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro__Lz4Codec, __pyx_vtabptr_5cavro__Lz4Codec) < 0) __PYX_ERR(3, 175, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro__Lz4Codec) < 0) __PYX_ERR(3, 175, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Lz4Codec, (PyObject *) __pyx_ptype_5cavro__Lz4Codec) < 0) __PYX_ERR(3, 175, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro__Lz4Codec) < 0) __PYX_ERR(3, 175, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_ContainerReader = &__pyx_vtable_5cavro_ContainerReader;
  __pyx_vtable_5cavro_ContainerReader._read_marker = (PyObject *(*)(struct __pyx_obj_5cavro_ContainerReader *, int __pyx_skip_dispatch))__pyx_f_5cavro_15ContainerReader__read_marker;
  __pyx_vtable_5cavro_ContainerReader.next_block = (int (*)(struct __pyx_obj_5cavro_ContainerReader *))__pyx_f_5cavro_15ContainerReader_next_block;
  __pyx_vtable_5cavro_ContainerReader.next_object = (PyObject *(*)(struct __pyx_obj_5cavro_ContainerReader *, int __pyx_skip_dispatch))__pyx_f_5cavro_15ContainerReader_next_object;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro_ContainerReader = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_ContainerReader_spec, NULL); if (unlikely(!__pyx_ptype_5cavro_ContainerReader)) __PYX_ERR(25, 41, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_ContainerReader_spec, __pyx_ptype_5cavro_ContainerReader) < 0) __PYX_ERR(25, 41, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_ContainerReader = &__pyx_type_5cavro_ContainerReader;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_ContainerReader) < 0) __PYX_ERR(25, 41, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_ContainerReader->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_ContainerReader->tp_dictoffset && __pyx_ptype_5cavro_ContainerReader->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_ContainerReader->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_ContainerReader, __pyx_vtabptr_5cavro_ContainerReader) < 0) __PYX_ERR(25, 41, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_ContainerReader) < 0) __PYX_ERR(25, 41, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ContainerReader, (PyObject *) __pyx_ptype_5cavro_ContainerReader) < 0) __PYX_ERR(25, 41, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_ContainerReader) < 0) __PYX_ERR(25, 41, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_ContainerWriter = &__pyx_vtable_5cavro_ContainerWriter;
  __pyx_vtable_5cavro_ContainerWriter._write_header = (int (*)(struct __pyx_obj_5cavro_ContainerWriter *))__pyx_f_5cavro_15ContainerWriter__write_header;
  __pyx_vtable_5cavro_ContainerWriter._flush_block = (int (*)(struct __pyx_obj_5cavro_ContainerWriter *, struct __pyx_opt_args_5cavro_15ContainerWriter__flush_block *__pyx_optional_args))__pyx_f_5cavro_15ContainerWriter__flush_block;
  __pyx_vtable_5cavro_ContainerWriter.write_one = (int (*)(struct __pyx_obj_5cavro_ContainerWriter *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5cavro_15ContainerWriter_write_one;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro_ContainerWriter = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_ContainerWriter_spec, NULL); if (unlikely(!__pyx_ptype_5cavro_ContainerWriter)) __PYX_ERR(25, 131, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_ContainerWriter_spec, __pyx_ptype_5cavro_ContainerWriter) < 0) __PYX_ERR(25, 131, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_ContainerWriter = &__pyx_type_5cavro_ContainerWriter;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_ContainerWriter) < 0) __PYX_ERR(25, 131, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_ContainerWriter->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_ContainerWriter->tp_dictoffset && __pyx_ptype_5cavro_ContainerWriter->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_ContainerWriter->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_ContainerWriter, __pyx_vtabptr_5cavro_ContainerWriter) < 0) __PYX_ERR(25, 131, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_ContainerWriter) < 0) __PYX_ERR(25, 131, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ContainerWriter, (PyObject *) __pyx_ptype_5cavro_ContainerWriter) < 0) __PYX_ERR(25, 131, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_ContainerWriter) < 0) __PYX_ERR(25, 131, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_PromoteToFloat = &__pyx_vtable_5cavro_PromoteToFloat;
  __pyx_vtable_5cavro_PromoteToFloat.__pyx_base = *__pyx_vtabptr_5cavro_ValueAdapter;
  __pyx_vtable_5cavro_PromoteToFloat.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_14PromoteToFloat_encode_value;
  __pyx_vtable_5cavro_PromoteToFloat.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_14PromoteToFloat_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_ValueAdapter); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_PromoteToFloat = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_PromoteToFloat_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_PromoteToFloat)) __PYX_ERR(26, 3, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_PromoteToFloat_spec, __pyx_ptype_5cavro_PromoteToFloat) < 0) __PYX_ERR(26, 3, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_PromoteToFloat = &__pyx_type_5cavro_PromoteToFloat;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_PromoteToFloat->tp_base = __pyx_ptype_5cavro_ValueAdapter;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_PromoteToFloat) < 0) __PYX_ERR(26, 3, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_PromoteToFloat->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_PromoteToFloat->tp_dictoffset && __pyx_ptype_5cavro_PromoteToFloat->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_PromoteToFloat->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_PromoteToFloat, __pyx_vtabptr_5cavro_PromoteToFloat) < 0) __PYX_ERR(26, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_PromoteToFloat) < 0) __PYX_ERR(26, 3, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PromoteToFloat, (PyObject *) __pyx_ptype_5cavro_PromoteToFloat) < 0) __PYX_ERR(26, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_PromoteToFloat) < 0) __PYX_ERR(26, 3, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_PromoteBytesToString = &__pyx_vtable_5cavro_PromoteBytesToString;
  __pyx_vtable_5cavro_PromoteBytesToString.__pyx_base = *__pyx_vtabptr_5cavro_ValueAdapter;
  __pyx_vtable_5cavro_PromoteBytesToString.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_20PromoteBytesToString_encode_value;
  __pyx_vtable_5cavro_PromoteBytesToString.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_20PromoteBytesToString_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_ValueAdapter); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_PromoteBytesToString = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_PromoteBytesToString_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_PromoteBytesToString)) __PYX_ERR(26, 16, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_PromoteBytesToString_spec, __pyx_ptype_5cavro_PromoteBytesToString) < 0) __PYX_ERR(26, 16, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_PromoteBytesToString = &__pyx_type_5cavro_PromoteBytesToString;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_PromoteBytesToString->tp_base = __pyx_ptype_5cavro_ValueAdapter;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_PromoteBytesToString) < 0) __PYX_ERR(26, 16, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_PromoteBytesToString->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_PromoteBytesToString->tp_dictoffset && __pyx_ptype_5cavro_PromoteBytesToString->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_PromoteBytesToString->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_PromoteBytesToString, __pyx_vtabptr_5cavro_PromoteBytesToString) < 0) __PYX_ERR(26, 16, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_PromoteBytesToString) < 0) __PYX_ERR(26, 16, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PromoteBytesToString, (PyObject *) __pyx_ptype_5cavro_PromoteBytesToString) < 0) __PYX_ERR(26, 16, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_PromoteBytesToString) < 0) __PYX_ERR(26, 16, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_PromoteStringToBytes = &__pyx_vtable_5cavro_PromoteStringToBytes;
  __pyx_vtable_5cavro_PromoteStringToBytes.__pyx_base = *__pyx_vtabptr_5cavro_ValueAdapter;
  __pyx_vtable_5cavro_PromoteStringToBytes.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_20PromoteStringToBytes_encode_value;
  __pyx_vtable_5cavro_PromoteStringToBytes.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_20PromoteStringToBytes_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_ValueAdapter); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_PromoteStringToBytes = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_PromoteStringToBytes_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_PromoteStringToBytes)) __PYX_ERR(26, 29, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_PromoteStringToBytes_spec, __pyx_ptype_5cavro_PromoteStringToBytes) < 0) __PYX_ERR(26, 29, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_PromoteStringToBytes = &__pyx_type_5cavro_PromoteStringToBytes;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_PromoteStringToBytes->tp_base = __pyx_ptype_5cavro_ValueAdapter;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_PromoteStringToBytes) < 0) __PYX_ERR(26, 29, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_PromoteStringToBytes->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_PromoteStringToBytes->tp_dictoffset && __pyx_ptype_5cavro_PromoteStringToBytes->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_PromoteStringToBytes->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_PromoteStringToBytes, __pyx_vtabptr_5cavro_PromoteStringToBytes) < 0) __PYX_ERR(26, 29, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_PromoteStringToBytes) < 0) __PYX_ERR(26, 29, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PromoteStringToBytes, (PyObject *) __pyx_ptype_5cavro_PromoteStringToBytes) < 0) __PYX_ERR(26, 29, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_PromoteStringToBytes) < 0) __PYX_ERR(26, 29, __pyx_L1_error)
  #endif
  __pyx_vtabptr_5cavro_CannotPromote = &__pyx_vtable_5cavro_CannotPromote;
  __pyx_vtable_5cavro_CannotPromote.__pyx_base = *__pyx_vtabptr_5cavro_ValueAdapter;
  __pyx_vtable_5cavro_CannotPromote.__pyx_base.encode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_13CannotPromote_encode_value;
  __pyx_vtable_5cavro_CannotPromote.__pyx_base.decode_value = (PyObject *(*)(struct __pyx_obj_5cavro_ValueAdapter *, PyObject *))__pyx_f_5cavro_13CannotPromote_decode_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_5cavro_ValueAdapter); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5cavro_CannotPromote = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro_CannotPromote_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_5cavro_CannotPromote)) __PYX_ERR(26, 42, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro_CannotPromote_spec, __pyx_ptype_5cavro_CannotPromote) < 0) __PYX_ERR(26, 42, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro_CannotPromote = &__pyx_type_5cavro_CannotPromote;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_5cavro_CannotPromote->tp_base = __pyx_ptype_5cavro_ValueAdapter;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro_CannotPromote) < 0) __PYX_ERR(26, 42, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro_CannotPromote->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro_CannotPromote->tp_dictoffset && __pyx_ptype_5cavro_CannotPromote->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro_CannotPromote->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_5cavro_CannotPromote, __pyx_vtabptr_5cavro_CannotPromote) < 0) __PYX_ERR(26, 42, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_5cavro_CannotPromote) < 0) __PYX_ERR(26, 42, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CannotPromote, (PyObject *) __pyx_ptype_5cavro_CannotPromote) < 0) __PYX_ERR(26, 42, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_5cavro_CannotPromote) < 0) __PYX_ERR(26, 42, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct__genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct__genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct__genexpr)) __PYX_ERR(13, 52, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct__genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct__genexpr) < 0) __PYX_ERR(13, 52, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct__genexpr = &__pyx_type_5cavro___pyx_scope_struct__genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct__genexpr) < 0) __PYX_ERR(13, 52, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct__genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct__genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct__genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct__genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_1_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_1_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_1_genexpr)) __PYX_ERR(13, 62, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_1_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(13, 62, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_1_genexpr = &__pyx_type_5cavro___pyx_scope_struct_1_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(13, 62, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_1_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_1_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_1_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_1_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_2_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_2_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_2_genexpr)) __PYX_ERR(0, 90, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_2_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_2_genexpr = &__pyx_type_5cavro___pyx_scope_struct_2_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_2_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_2_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_2_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_2_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_3_walk_types = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_3_walk_types_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_3_walk_types)) __PYX_ERR(20, 23, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_3_walk_types_spec, __pyx_ptype_5cavro___pyx_scope_struct_3_walk_types) < 0) __PYX_ERR(20, 23, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_3_walk_types = &__pyx_type_5cavro___pyx_scope_struct_3_walk_types;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_3_walk_types) < 0) __PYX_ERR(20, 23, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_3_walk_types->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_3_walk_types->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_3_walk_types->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_3_walk_types->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer)) __PYX_ERR(8, 110, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer_spec, __pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer) < 0) __PYX_ERR(8, 110, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer = &__pyx_type_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer) < 0) __PYX_ERR(8, 110, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_4___pyx_base____pyx_base___for_writer->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_5_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_5_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_5_genexpr)) __PYX_ERR(8, 125, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_5_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_5_genexpr) < 0) __PYX_ERR(8, 125, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_5_genexpr = &__pyx_type_5cavro___pyx_scope_struct_5_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_5_genexpr) < 0) __PYX_ERR(8, 125, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_5_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_5_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_5_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_5_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_6_walk_types = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_6_walk_types_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_6_walk_types)) __PYX_ERR(9, 23, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_6_walk_types_spec, __pyx_ptype_5cavro___pyx_scope_struct_6_walk_types) < 0) __PYX_ERR(9, 23, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_6_walk_types = &__pyx_type_5cavro___pyx_scope_struct_6_walk_types;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_6_walk_types) < 0) __PYX_ERR(9, 23, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_6_walk_types->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_6_walk_types->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_6_walk_types->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_6_walk_types->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_7_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_7_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_7_genexpr)) __PYX_ERR(11, 142, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_7_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_7_genexpr) < 0) __PYX_ERR(11, 142, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_7_genexpr = &__pyx_type_5cavro___pyx_scope_struct_7_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_7_genexpr) < 0) __PYX_ERR(11, 142, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_7_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_7_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_7_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_7_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_8___init__ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_8___init___spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_8___init__)) __PYX_ERR(11, 321, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_8___init___spec, __pyx_ptype_5cavro___pyx_scope_struct_8___init__) < 0) __PYX_ERR(11, 321, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_8___init__ = &__pyx_type_5cavro___pyx_scope_struct_8___init__;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_8___init__) < 0) __PYX_ERR(11, 321, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_8___init__->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_8___init__->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_8___init__->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_8___init__->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_9_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_9_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_9_genexpr)) __PYX_ERR(11, 327, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_9_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_9_genexpr) < 0) __PYX_ERR(11, 327, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_9_genexpr = &__pyx_type_5cavro___pyx_scope_struct_9_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_9_genexpr) < 0) __PYX_ERR(11, 327, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_9_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_9_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_9_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_9_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_10_walk_types = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_10_walk_types_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_10_walk_types)) __PYX_ERR(11, 366, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_10_walk_types_spec, __pyx_ptype_5cavro___pyx_scope_struct_10_walk_types) < 0) __PYX_ERR(11, 366, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_10_walk_types = &__pyx_type_5cavro___pyx_scope_struct_10_walk_types;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_10_walk_types) < 0) __PYX_ERR(11, 366, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_10_walk_types->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_10_walk_types->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_10_walk_types->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_10_walk_types->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_11___init__ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_11___init___spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_11___init__)) __PYX_ERR(22, 33, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_11___init___spec, __pyx_ptype_5cavro___pyx_scope_struct_11___init__) < 0) __PYX_ERR(22, 33, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_11___init__ = &__pyx_type_5cavro___pyx_scope_struct_11___init__;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_11___init__) < 0) __PYX_ERR(22, 33, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_11___init__->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_11___init__->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_11___init__->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_11___init__->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_12_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_12_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_12_genexpr)) __PYX_ERR(22, 35, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_12_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_12_genexpr) < 0) __PYX_ERR(22, 35, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_12_genexpr = &__pyx_type_5cavro___pyx_scope_struct_12_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_12_genexpr) < 0) __PYX_ERR(22, 35, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_12_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_12_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_12_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_12_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_13_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_13_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_13_genexpr)) __PYX_ERR(22, 44, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_13_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_13_genexpr) < 0) __PYX_ERR(22, 44, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_13_genexpr = &__pyx_type_5cavro___pyx_scope_struct_13_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_13_genexpr) < 0) __PYX_ERR(22, 44, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_13_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_13_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_13_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_13_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_14_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_14_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_14_genexpr)) __PYX_ERR(22, 47, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_14_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_14_genexpr) < 0) __PYX_ERR(22, 47, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_14_genexpr = &__pyx_type_5cavro___pyx_scope_struct_14_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_14_genexpr) < 0) __PYX_ERR(22, 47, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_14_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_14_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_14_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_14_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_15_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_15_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_15_genexpr)) __PYX_ERR(22, 49, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_15_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_15_genexpr) < 0) __PYX_ERR(22, 49, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_15_genexpr = &__pyx_type_5cavro___pyx_scope_struct_15_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_15_genexpr) < 0) __PYX_ERR(22, 49, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_15_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_15_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_15_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_15_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_16_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_16_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_16_genexpr)) __PYX_ERR(22, 50, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_16_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_16_genexpr) < 0) __PYX_ERR(22, 50, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_16_genexpr = &__pyx_type_5cavro___pyx_scope_struct_16_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_16_genexpr) < 0) __PYX_ERR(22, 50, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_16_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_16_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_16_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_16_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_17_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_17_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_17_genexpr)) __PYX_ERR(22, 52, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_17_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_17_genexpr) < 0) __PYX_ERR(22, 52, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_17_genexpr = &__pyx_type_5cavro___pyx_scope_struct_17_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_17_genexpr) < 0) __PYX_ERR(22, 52, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_17_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_17_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_17_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_17_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_18_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_18_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_18_genexpr)) __PYX_ERR(22, 54, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_18_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_18_genexpr) < 0) __PYX_ERR(22, 54, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_18_genexpr = &__pyx_type_5cavro___pyx_scope_struct_18_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_18_genexpr) < 0) __PYX_ERR(22, 54, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_18_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_18_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_18_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_18_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_19_walk_types = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_19_walk_types_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_19_walk_types)) __PYX_ERR(22, 90, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_19_walk_types_spec, __pyx_ptype_5cavro___pyx_scope_struct_19_walk_types) < 0) __PYX_ERR(22, 90, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_19_walk_types = &__pyx_type_5cavro___pyx_scope_struct_19_walk_types;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_19_walk_types) < 0) __PYX_ERR(22, 90, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_19_walk_types->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_19_walk_types->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_19_walk_types->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_19_walk_types->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_20_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_20_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_20_genexpr)) __PYX_ERR(2, 30, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_20_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_20_genexpr) < 0) __PYX_ERR(2, 30, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_20_genexpr = &__pyx_type_5cavro___pyx_scope_struct_20_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_20_genexpr) < 0) __PYX_ERR(2, 30, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_20_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_20_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_20_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_20_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_21_walk_types = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_21_walk_types_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_21_walk_types)) __PYX_ERR(2, 286, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_21_walk_types_spec, __pyx_ptype_5cavro___pyx_scope_struct_21_walk_types) < 0) __PYX_ERR(2, 286, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_21_walk_types = &__pyx_type_5cavro___pyx_scope_struct_21_walk_types;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_21_walk_types) < 0) __PYX_ERR(2, 286, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_21_walk_types->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_21_walk_types->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_21_walk_types->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_21_walk_types->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_22_get_namespaced_aliases_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases)) __PYX_ERR(2, 390, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_22_get_namespaced_aliases_spec, __pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases) < 0) __PYX_ERR(2, 390, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases = &__pyx_type_5cavro___pyx_scope_struct_22_get_namespaced_aliases;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases) < 0) __PYX_ERR(2, 390, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_22_get_namespaced_aliases->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_5cavro___pyx_scope_struct_23_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_5cavro___pyx_scope_struct_23_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_5cavro___pyx_scope_struct_23_genexpr)) __PYX_ERR(2, 394, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_5cavro___pyx_scope_struct_23_genexpr_spec, __pyx_ptype_5cavro___pyx_scope_struct_23_genexpr) < 0) __PYX_ERR(2, 394, __pyx_L1_error)
  #else
  __pyx_ptype_5cavro___pyx_scope_struct_23_genexpr = &__pyx_type_5cavro___pyx_scope_struct_23_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_5cavro___pyx_scope_struct_23_genexpr) < 0) __PYX_ERR(2, 394, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_5cavro___pyx_scope_struct_23_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_5cavro___pyx_scope_struct_23_genexpr->tp_dictoffset && __pyx_ptype_5cavro___pyx_scope_struct_23_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_5cavro___pyx_scope_struct_23_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_array_type = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type___pyx_array_spec, NULL); if (unlikely(!__pyx_array_type)) __PYX_ERR(1, 114, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_array_type->tp_as_buffer = &__pyx_tp_as_buffer_array;
  if (!__pyx_array_type->tp_as_buffer->bf_releasebuffer && __pyx_array_type->tp_base->tp_as_buffer && __pyx_array_type->tp_base->tp_as_buffer->bf_releasebuffer) {
    __pyx_array_type->tp_as_buffer->bf_releasebuffer = __pyx_array_type->tp_base->tp_as_buffer->bf_releasebuffer;
  }
  #elif defined(Py_bf_getbuffer) && defined(Py_bf_releasebuffer)
  /* PY_VERSION_HEX >= 0x03090000 || Py_LIMITED_API >= 0x030B0000 */
  #elif defined(_MSC_VER)
  #pragma message ("The buffer protocol is not supported in the Limited C-API < 3.11.")
  #else
  #warning "The buffer protocol is not supported in the Limited C-API < 3.11."
  #endif
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type___pyx_array_spec, __pyx_array_type) < 0) __PYX_ERR(1, 114, __pyx_L1_error)
  #else
  __pyx_array_type = &__pyx_type___pyx_array;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_array_type) < 0) __PYX_ERR(1, 114, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_array_type->tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_array_type, __pyx_vtabptr_array) < 0) __PYX_ERR(1, 114, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_array_type) < 0) __PYX_ERR(1, 114, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_array_type) < 0) __PYX_ERR(1, 114, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_MemviewEnum_type = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type___pyx_MemviewEnum_spec, NULL); if (unlikely(!__pyx_MemviewEnum_type)) __PYX_ERR(1, 302, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type___pyx_MemviewEnum_spec, __pyx_MemviewEnum_type) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  #else
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_MemviewEnum_type) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_MemviewEnum_type->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_MemviewEnum_type->tp_dictoffset && __pyx_MemviewEnum_type->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_MemviewEnum_type->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_MemviewEnum_type) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  #endif
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  __pyx_vtable_memoryview._get_base = (PyObject *(*)(struct __pyx_memoryview_obj *))__pyx_memoryview__get_base;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_memoryview_type = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type___pyx_memoryview_spec, NULL); if (unlikely(!__pyx_memoryview_type)) __PYX_ERR(1, 337, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_memoryview_type->tp_as_buffer = &__pyx_tp_as_buffer_memoryview;
  if (!__pyx_memoryview_type->tp_as_buffer->bf_releasebuffer && __pyx_memoryview_type->tp_base->tp_as_buffer && __pyx_memoryview_type->tp_base->tp_as_buffer->bf_releasebuffer) {
    __pyx_memoryview_type->tp_as_buffer->bf_releasebuffer = __pyx_memoryview_type->tp_base->tp_as_buffer->bf_releasebuffer;
  }
  #elif defined(Py_bf_getbuffer) && defined(Py_bf_releasebuffer)
  /* PY_VERSION_HEX >= 0x03090000 || Py_LIMITED_API >= 0x030B0000 */
  #elif defined(_MSC_VER)
  #pragma message ("The buffer protocol is not supported in the Limited C-API < 3.11.")
  #else
  #warning "The buffer protocol is not supported in the Limited C-API < 3.11."
  #endif
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type___pyx_memoryview_spec, __pyx_memoryview_type) < 0) __PYX_ERR(1, 337, __pyx_L1_error)
  #else
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_memoryview_type) < 0) __PYX_ERR(1, 337, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_memoryview_type->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_memoryview_type->tp_dictoffset && __pyx_memoryview_type->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_memoryview_type->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_memoryview_type, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(1, 337, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_memoryview_type) < 0) __PYX_ERR(1, 337, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_memoryview_type) < 0) __PYX_ERR(1, 337, __pyx_L1_error)
  #endif
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_vtable__memoryviewslice.__pyx_base._get_base = (PyObject *(*)(struct __pyx_memoryview_obj *))__pyx_memoryviewslice__get_base;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_memoryview_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_memoryviewslice_type = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type___pyx_memoryviewslice_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_memoryviewslice_type)) __PYX_ERR(1, 952, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type___pyx_memoryviewslice_spec, __pyx_memoryviewslice_type) < 0) __PYX_ERR(1, 952, __pyx_L1_error)
  #else
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_memoryviewslice_type->tp_base = __pyx_memoryview_type;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_memoryviewslice_type) < 0) __PYX_ERR(1, 952, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_memoryviewslice_type->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_memoryviewslice_type->tp_dictoffset && __pyx_memoryviewslice_type->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_memoryviewslice_type->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_memoryviewslice_type, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(1, 952, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_memoryviewslice_type) < 0) __PYX_ERR(1, 952, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_memoryviewslice_type) < 0) __PYX_ERR(1, 952, __pyx_L1_error)
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_4(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_4(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_4(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_4); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(27, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_0_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_4(PyBoolObject),__Pyx_ImportType_CheckSize_Warn_3_0_4); if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(28, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(29, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType_3_0_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_4(PyComplexObject),__Pyx_ImportType_CheckSize_Warn_3_0_4); if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(29, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("array"); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_5array_array = __Pyx_ImportType_3_0_4(__pyx_t_1, "array", "array", sizeof(arrayobject), __PYX_GET_STRUCT_ALIGNMENT_3_0_4(arrayobject),__Pyx_ImportType_CheckSize_Warn_3_0_4); if (!__pyx_ptype_7cpython_5array_array) __PYX_ERR(15, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_cavro(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_cavro},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "cavro",
      __pyx_k_cavro_is_a_library_for_encoding, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcavro(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcavro(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_cavro(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_cavro(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_cavro(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  static PyThread_type_lock __pyx_t_8[8];
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'cavro' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("cavro", __pyx_methods, __pyx_k_cavro_is_a_library_for_encoding, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(13, 2, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 2, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to cavro pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(13, 2, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(13, 2, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(13, 2, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(13, 2, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(13, 2, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_cavro(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(13, 2, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(13, 2, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_cavro) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(13, 2, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "cavro")) {
      if (unlikely((PyDict_SetItemString(modules, "cavro", __pyx_m) < 0))) __PYX_ERR(13, 2, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(13, 2, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(13, 2, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #endif

  /* "View.MemoryView":99
 * 
 * cdef object __pyx_collections_abc_Sequence "__pyx_collections_abc_Sequence"
 * try:             # <<<<<<<<<<<<<<
 *     if __import__("sys").version_info >= (3, 3):
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "View.MemoryView":100
 * cdef object __pyx_collections_abc_Sequence "__pyx_collections_abc_Sequence"
 * try:
 *     if __import__("sys").version_info >= (3, 3):             # <<<<<<<<<<<<<<
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 *     else:
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_tuple__123, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 100, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_version_info); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 100, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_5, __pyx_tuple__124, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 100, __pyx_L2_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(1, 100, __pyx_L2_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {

        /* "View.MemoryView":101
 * try:
 *     if __import__("sys").version_info >= (3, 3):
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence             # <<<<<<<<<<<<<<
 *     else:
 *         __pyx_collections_abc_Sequence = __import__("collections").Sequence
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_tuple__125, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 101, __pyx_L2_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_abc); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 101, __pyx_L2_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Sequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 101, __pyx_L2_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XGOTREF(__pyx_collections_abc_Sequence);
        __Pyx_DECREF_SET(__pyx_collections_abc_Sequence, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        __pyx_t_4 = 0;

        /* "View.MemoryView":100
 * cdef object __pyx_collections_abc_Sequence "__pyx_collections_abc_Sequence"
 * try:
 *     if __import__("sys").version_info >= (3, 3):             # <<<<<<<<<<<<<<
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 *     else:
 */
        goto __pyx_L8;
      }

      /* "View.MemoryView":103
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 *     else:
 *         __pyx_collections_abc_Sequence = __import__("collections").Sequence             # <<<<<<<<<<<<<<
 * except:
 * 
 */
      /*else*/ {
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_tuple__126, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 103, __pyx_L2_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Sequence); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 103, __pyx_L2_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XGOTREF(__pyx_collections_abc_Sequence);
        __Pyx_DECREF_SET(__pyx_collections_abc_Sequence, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_5);
        __pyx_t_5 = 0;
      }
      __pyx_L8:;

      /* "View.MemoryView":99
 * 
 * cdef object __pyx_collections_abc_Sequence "__pyx_collections_abc_Sequence"
 * try:             # <<<<<<<<<<<<<<
 *     if __import__("sys").version_info >= (3, 3):
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L7_try_end;
    __pyx_L2_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "View.MemoryView":104
 *     else:
 *         __pyx_collections_abc_Sequence = __import__("collections").Sequence
 * except:             # <<<<<<<<<<<<<<
 * 
 *     __pyx_collections_abc_Sequence = None
 */
    /*except:*/ {
      __Pyx_AddTraceback("View.MemoryView", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(1, 104, __pyx_L4_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "View.MemoryView":106
 * except:
 * 
 *     __pyx_collections_abc_Sequence = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_INCREF(Py_None);
      __Pyx_XGOTREF(__pyx_collections_abc_Sequence);
      __Pyx_DECREF_SET(__pyx_collections_abc_Sequence, Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L3_exception_handled;
    }

    /* "View.MemoryView":99
 * 
 * cdef object __pyx_collections_abc_Sequence "__pyx_collections_abc_Sequence"
 * try:             # <<<<<<<<<<<<<<
 *     if __import__("sys").version_info >= (3, 3):
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 */
    __pyx_L4_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L3_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L7_try_end:;
  }

  /* "View.MemoryView":241
 * 
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "View.MemoryView":242
 * 
 *     try:
 *         count = __pyx_collections_abc_Sequence.count             # <<<<<<<<<<<<<<
 *         index = __pyx_collections_abc_Sequence.index
 *     except:
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_collections_abc_Sequence, __pyx_n_s_count); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 242, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_SetItemOnTypeDict(__pyx_array_type, __pyx_n_s_count, __pyx_t_7) < 0) __PYX_ERR(1, 242, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      PyType_Modified(__pyx_array_type);

      /* "View.MemoryView":243
 *     try:
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index             # <<<<<<<<<<<<<<
 *     except:
 *         pass
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_collections_abc_Sequence, __pyx_n_s_index); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 243, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_SetItemOnTypeDict(__pyx_array_type, __pyx_n_s_index, __pyx_t_7) < 0) __PYX_ERR(1, 243, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      PyType_Modified(__pyx_array_type);

      /* "View.MemoryView":241
 * 
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L16_try_end;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "View.MemoryView":244
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index
 *     except:             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
    /*except:*/ {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L12_exception_handled;
    }
    __pyx_L12_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    __pyx_L16_try_end:;
  }

  /* "View.MemoryView":309
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__127, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_XGOTREF(generic);
  __Pyx_DECREF_SET(generic, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "View.MemoryView":310
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__128, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_XGOTREF(strided);
  __Pyx_DECREF_SET(strided, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "View.MemoryView":311
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__129, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_XGOTREF(indirect);
  __Pyx_DECREF_SET(indirect, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "View.MemoryView":314
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__130, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_XGOTREF(contiguous);
  __Pyx_DECREF_SET(contiguous, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "View.MemoryView":315
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__131, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_XGOTREF(indirect_contiguous);
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "View.MemoryView":323
 * 
 * 
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[8] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":324
 * 
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[8] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */
  __pyx_t_8[0] = PyThread_allocate_lock();
  __pyx_t_8[1] = PyThread_allocate_lock();
  __pyx_t_8[2] = PyThread_allocate_lock();
  __pyx_t_8[3] = PyThread_allocate_lock();
  __pyx_t_8[4] = PyThread_allocate_lock();
  __pyx_t_8[5] = PyThread_allocate_lock();
  __pyx_t_8[6] = PyThread_allocate_lock();
  __pyx_t_8[7] = PyThread_allocate_lock();
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_8, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":982
 * 
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "View.MemoryView":983
 * 
 *     try:
 *         count = __pyx_collections_abc_Sequence.count             # <<<<<<<<<<<<<<
 *         index = __pyx_collections_abc_Sequence.index
 *     except:
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_collections_abc_Sequence, __pyx_n_s_count); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 983, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_SetItemOnTypeDict(__pyx_memoryviewslice_type, __pyx_n_s_count, __pyx_t_7) < 0) __PYX_ERR(1, 983, __pyx_L17_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      PyType_Modified(__pyx_memoryviewslice_type);

      /* "View.MemoryView":984
 *     try:
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index             # <<<<<<<<<<<<<<
 *     except:
 *         pass
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_collections_abc_Sequence, __pyx_n_s_index); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 984, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_SetItemOnTypeDict(__pyx_memoryviewslice_type, __pyx_n_s_index, __pyx_t_7) < 0) __PYX_ERR(1, 984, __pyx_L17_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      PyType_Modified(__pyx_memoryviewslice_type);

      /* "View.MemoryView":982
 * 
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L22_try_end;
    __pyx_L17_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "View.MemoryView":985
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index
 *     except:             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
    /*except:*/ {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L18_exception_handled;
    }
    __pyx_L18_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L22_try_end:;
  }

  /* "View.MemoryView":988
 *         pass
 * 
 * try:             # <<<<<<<<<<<<<<
 *     if __pyx_collections_abc_Sequence:
 * 
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "View.MemoryView":989
 * 
 * try:
 *     if __pyx_collections_abc_Sequence:             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_collections_abc_Sequence); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(1, 989, __pyx_L23_error)
      if (__pyx_t_6) {

        /* "View.MemoryView":993
 * 
 * 
 *         __pyx_collections_abc_Sequence.register(_memoryviewslice)             # <<<<<<<<<<<<<<
 *         __pyx_collections_abc_Sequence.register(array)
 * except:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_collections_abc_Sequence, __pyx_n_s_register); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 993, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_7, ((PyObject *)__pyx_memoryviewslice_type)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 993, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "View.MemoryView":994
 * 
 *         __pyx_collections_abc_Sequence.register(_memoryviewslice)
 *         __pyx_collections_abc_Sequence.register(array)             # <<<<<<<<<<<<<<
 * except:
 *     pass  # ignore failure, it's a minor issue
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_collections_abc_Sequence, __pyx_n_s_register); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 994, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_array_type)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 994, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":989
 * 
 * try:
 *     if __pyx_collections_abc_Sequence:             # <<<<<<<<<<<<<<
 * 
 * 
 */
      }

      /* "View.MemoryView":988
 *         pass
 * 
 * try:             # <<<<<<<<<<<<<<
 *     if __pyx_collections_abc_Sequence:
 * 
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L28_try_end;
    __pyx_L23_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "View.MemoryView":995
 *         __pyx_collections_abc_Sequence.register(_memoryviewslice)
 *         __pyx_collections_abc_Sequence.register(array)
 * except:             # <<<<<<<<<<<<<<
 *     pass  # ignore failure, it's a minor issue
 * 
 */
    /*except:*/ {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L24_exception_handled;
    }
    __pyx_L24_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    __pyx_L28_try_end:;
  }

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_7 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cavro.pyx":9
 * 
 * cimport cython
 * import warnings             # <<<<<<<<<<<<<<
 * from cpython.object cimport Py_SIZE
 * from libc.stdint cimport *
 */
  __pyx_t_7 = __Pyx_ImportDottedModule(__pyx_n_s_warnings, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_7) < 0) __PYX_ERR(13, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cavro.pyx":13
 * from libc.stdint cimport *
 * from libc.string cimport memcmp
 * import inspect             # <<<<<<<<<<<<<<
 * import dataclasses
 * 
 */
  __pyx_t_7 = __Pyx_patch_inspect(__Pyx_ImportDottedModule(__pyx_n_s_inspect, NULL)); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_inspect, __pyx_t_7) < 0) __PYX_ERR(13, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cavro.pyx":14
 * from libc.string cimport memcmp
 * import inspect
 * import dataclasses             # <<<<<<<<<<<<<<
 * 
 * from collections.abc import Sequence
 */
  __pyx_t_7 = __Pyx_ImportDottedModule(__pyx_n_s_dataclasses, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dataclasses, __pyx_t_7) < 0) __PYX_ERR(13, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cavro.pyx":16
 * import dataclasses
 * 
 * from collections.abc import Sequence             # <<<<<<<<<<<<<<
 * 
 * from cython.dataclasses cimport dataclass
 */
  __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_Sequence);
  __Pyx_GIVEREF(__pyx_n_s_Sequence);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_Sequence)) __PYX_ERR(13, 16, __pyx_L1_error);
  __pyx_t_4 = __Pyx_Import(__pyx_kp_s_collections_abc, __pyx_t_7, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_Sequence); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Sequence, __pyx_t_7) < 0) __PYX_ERR(13, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "cavro.pyx":19
 * 
 * from cython.dataclasses cimport dataclass
 * import datetime             # <<<<<<<<<<<<<<
 * import re
 * import decimal
 */
  __pyx_t_4 = __Pyx_ImportDottedModule(__pyx_n_s_datetime, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_datetime, __pyx_t_4) < 0) __PYX_ERR(13, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "cavro.pyx":20
 * from cython.dataclasses cimport dataclass
 * import datetime
 * import re             # <<<<<<<<<<<<<<
 * import decimal
 * import enum
 */
  __pyx_t_4 = __Pyx_ImportDottedModule(__pyx_n_s_re, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_re, __pyx_t_4) < 0) __PYX_ERR(13, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "cavro.pyx":21
 * import datetime
 * import re
 * import decimal             # <<<<<<<<<<<<<<
 * import enum
 * from cpython cimport array
 */
  __pyx_t_4 = __Pyx_ImportDottedModule(__pyx_n_s_decimal, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_decimal, __pyx_t_4) < 0) __PYX_ERR(13, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "cavro.pyx":22
 * import re
 * import decimal
 * import enum             # <<<<<<<<<<<<<<
 * from cpython cimport array
 * from cython cimport bint
 */
  __pyx_t_4 = __Pyx_ImportDottedModule(__pyx_n_s_enum, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_enum, __pyx_t_4) < 0) __PYX_ERR(13, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "cavro.pyx":25
 * from cpython cimport array
 * from cython cimport bint
 * from functools import partial             # <<<<<<<<<<<<<<
 * import uuid
 * import math
 */
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_partial);
  __Pyx_GIVEREF(__pyx_n_s_partial);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_partial)) __PYX_ERR(13, 25, __pyx_L1_error);
  __pyx_t_7 = __Pyx_Import(__pyx_n_s_functools, __pyx_t_4, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_7, __pyx_n_s_partial); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_partial, __pyx_t_4) < 0) __PYX_ERR(13, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cavro.pyx":26
 * from cython cimport bint
 * from functools import partial
 * import uuid             # <<<<<<<<<<<<<<
 * import math
 * from types import MappingProxyType
 */
  __pyx_t_7 = __Pyx_ImportDottedModule(__pyx_n_s_uuid, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_uuid, __pyx_t_7) < 0) __PYX_ERR(13, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cavro.pyx":27
 * from functools import partial
 * import uuid
 * import math             # <<<<<<<<<<<<<<
 * from types import MappingProxyType
 * from typing import Union
 */
  __pyx_t_7 = __Pyx_ImportDottedModule(__pyx_n_s_math, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_math, __pyx_t_7) < 0) __PYX_ERR(13, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cavro.pyx":28
 * import uuid
 * import math
 * from types import MappingProxyType             # <<<<<<<<<<<<<<
 * from typing import Union
 * from cpython.dict cimport PyDictProxy_New
 */
  __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_MappingProxyType);
  __Pyx_GIVEREF(__pyx_n_s_MappingProxyType);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_MappingProxyType)) __PYX_ERR(13, 28, __pyx_L1_error);
  __pyx_t_4 = __Pyx_Import(__pyx_n_s_types, __pyx_t_7, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_MappingProxyType); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MappingProxyType, __pyx_t_7) < 0) __PYX_ERR(13, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "cavro.pyx":29
 * import math
 * from types import MappingProxyType
 * from typing import Union             # <<<<<<<<<<<<<<
 * from cpython.dict cimport PyDictProxy_New
 * 
 */
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_Union);
  __Pyx_GIVEREF(__pyx_n_s_Union);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_Union)) __PYX_ERR(13, 29, __pyx_L1_error);
  __pyx_t_7 = __Pyx_Import(__pyx_n_s_typing, __pyx_t_4, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_7, __pyx_n_s_Union); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Union, __pyx_t_4) < 0) __PYX_ERR(13, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cavro.pyx":38
 * ctypedef bint bool
 * 
 * if cython.sizeof(Py_ssize_t) == 8:             # <<<<<<<<<<<<<<
 *     SSIZE_TYPECODE = b'q'
 * elif cython.sizeof(Py_ssize_t) == 4:
 */
  __pyx_t_6 = ((sizeof(Py_ssize_t)) == 8);
  if (__pyx_t_6) {

    /* "cavro.pyx":39
 * 
 * if cython.sizeof(Py_ssize_t) == 8:
 *     SSIZE_TYPECODE = b'q'             # <<<<<<<<<<<<<<
 * elif cython.sizeof(Py_ssize_t) == 4:
 *     SSIZE_TYPECODE = b"l"
 */
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_SSIZE_TYPECODE, __pyx_n_b_q) < 0) __PYX_ERR(13, 39, __pyx_L1_error)

    /* "cavro.pyx":38
 * ctypedef bint bool
 * 
 * if cython.sizeof(Py_ssize_t) == 8:             # <<<<<<<<<<<<<<
 *     SSIZE_TYPECODE = b'q'
 * elif cython.sizeof(Py_ssize_t) == 4:
 */
    goto __pyx_L30;
  }

  /* "cavro.pyx":40
 * if cython.sizeof(Py_ssize_t) == 8:
 *     SSIZE_TYPECODE = b'q'
 * elif cython.sizeof(Py_ssize_t) == 4:             # <<<<<<<<<<<<<<
 *     SSIZE_TYPECODE = b"l"
 * else:
 */
  __pyx_t_6 = ((sizeof(Py_ssize_t)) == 4);
  if (__pyx_t_6) {

    /* "cavro.pyx":41
 *     SSIZE_TYPECODE = b'q'
 * elif cython.sizeof(Py_ssize_t) == 4:
 *     SSIZE_TYPECODE = b"l"             # <<<<<<<<<<<<<<
 * else:
 *     SSIZE_TYPECODE = "UNKNOWN"
 */
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_SSIZE_TYPECODE, __pyx_n_b_l) < 0) __PYX_ERR(13, 41, __pyx_L1_error)

    /* "cavro.pyx":40
 * if cython.sizeof(Py_ssize_t) == 8:
 *     SSIZE_TYPECODE = b'q'
 * elif cython.sizeof(Py_ssize_t) == 4:             # <<<<<<<<<<<<<<
 *     SSIZE_TYPECODE = b"l"
 * else:
 */
    goto __pyx_L30;
  }

  /* "cavro.pyx":43
 *     SSIZE_TYPECODE = b"l"
 * else:
 *     SSIZE_TYPECODE = "UNKNOWN"             # <<<<<<<<<<<<<<
 * 
 * __version__ = "1.0.0"
 */
  /*else*/ {
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_SSIZE_TYPECODE, __pyx_n_u_UNKNOWN) < 0) __PYX_ERR(13, 43, __pyx_L1_error)
  }
  __pyx_L30:;

  /* "cavro.pyx":45
 *     SSIZE_TYPECODE = "UNKNOWN"
 * 
 * __version__ = "1.0.0"             # <<<<<<<<<<<<<<
 * 
 * __tests = {}
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_version, __pyx_kp_u_1_0_0) < 0) __PYX_ERR(13, 45, __pyx_L1_error)

  /* "cavro.pyx":47
 * __version__ = "1.0.0"
 * 
 * __tests = {}             # <<<<<<<<<<<<<<
 * def _tests(fn):
 *     def add(test, *args):
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_tests, __pyx_t_7) < 0) __PYX_ERR(13, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cavro.pyx":48
 * 
 * __tests = {}
 * def _tests(fn):             # <<<<<<<<<<<<<<
 *     def add(test, *args):
 *         name = test.__name__
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_1_tests, 0, __pyx_n_s_tests_2, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__135)); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_tests_2, __pyx_t_7) < 0) __PYX_ERR(13, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cavro.pyx":57
 *     return _tests
 * 
 * __perf_tests = {}             # <<<<<<<<<<<<<<
 * def _perf(fn):
 *     def add(test, *args):
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perf_tests, __pyx_t_7) < 0) __PYX_ERR(13, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cavro.pyx":58
 * 
 * __perf_tests = {}
 * def _perf(fn):             # <<<<<<<<<<<<<<
 *     def add(test, *args):
 *         name = test.__name__
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_3_perf, 0, __pyx_n_s_perf, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__136)); if (unlikely(!__pyx_t_7)) __PYX_ERR(13, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perf, __pyx_t_7) < 0) __PYX_ERR(13, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/error.pxi":2
 * 
 * class CavroException(Exception):             # <<<<<<<<<<<<<<
 *     """
 *     Base class for exceptions raised by cavro
 */
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  __Pyx_GIVEREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])))) __PYX_ERR(0, 2, __pyx_L1_error);
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_4, __pyx_n_s_CavroException, __pyx_n_s_CavroException, (PyObject *) NULL, __pyx_n_s_cavro, __pyx_kp_s_Base_class_for_exceptions_raise); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__pyx_t_4 != __pyx_t_7) {
    if (unlikely((PyDict_SetItemString(__pyx_t_9, "__orig_bases__", __pyx_t_7) < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_CavroException, __pyx_t_4, __pyx_t_9, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CavroException, __pyx_t_7) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/error.pxi":8
 * 
 * 
 * class InvalidName(CavroException):             # <<<<<<<<<<<<<<
 *     """
 *     The schema contains a type or enum symbol with an invalid name (as per the avro specification)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CavroException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4)) __PYX_ERR(0, 8, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_9, __pyx_t_4, __pyx_n_s_InvalidName, __pyx_n_s_InvalidName, (PyObject *) NULL, __pyx_n_s_cavro, __pyx_kp_s_The_schema_contains_a_type_or_e); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_4 != __pyx_t_5) {
    if (unlikely((PyDict_SetItemString(__pyx_t_7, "__orig_bases__", __pyx_t_5) < 0))) __PYX_ERR(0, 8, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_9, __pyx_n_s_InvalidName, __pyx_t_4, __pyx_t_7, NULL, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_InvalidName, __pyx_t_5) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/error.pxi":14
 * 
 * 
 * class UnknownType(CavroException):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CavroException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_4)) __PYX_ERR(0, 14, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_4, __pyx_n_s_UnknownType, __pyx_n_s_UnknownType, (PyObject *) NULL, __pyx_n_s_cavro, __pyx_kp_s_The_schema_contains_an_unexptec); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_4 != __pyx_t_9) {
    if (unlikely((PyDict_SetItemString(__pyx_t_5, "__orig_bases__", __pyx_t_9) < 0))) __PYX_ERR(0, 14, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "src/error.pxi":20
 *     """
 * 
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *         super().__init__(f"Unknown type: {name}")
 */
  __pyx_t_10 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_11UnknownType_1__init__, 0, __pyx_n_s_UnknownType___init, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__138)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_t_10);
  PyList_Append(__pyx_t_9, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_init, __pyx_t_10) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "src/error.pxi":14
 * 
 * 
 * class UnknownType(CavroException):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_10 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_UnknownType, __pyx_t_4, __pyx_t_5, NULL, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_9, __pyx_t_10) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UnknownType, __pyx_t_10) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/error.pxi":25
 * 
 * 
 * class DuplicateName(CavroException):             # <<<<<<<<<<<<<<
 *     """
 *     A record contains multiple fields with the same name, a schema contains multiple types of the same name, or an enum has multiple identical symbols.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CavroException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4)) __PYX_ERR(0, 25, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_4, __pyx_n_s_DuplicateName, __pyx_n_s_DuplicateName, (PyObject *) NULL, __pyx_n_s_cavro, __pyx_kp_s_A_record_contains_multiple_fiel); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__pyx_t_4 != __pyx_t_7) {
    if (unlikely((PyDict_SetItemString(__pyx_t_10, "__orig_bases__", __pyx_t_7) < 0))) __PYX_ERR(0, 25, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_DuplicateName, __pyx_t_4, __pyx_t_10, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DuplicateName, __pyx_t_7) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/error.pxi":31
 * 
 * 
 * class InvalidHasher(CavroException):             # <<<<<<<<<<<<<<
 *     """
 *     An unknown hash method was requested
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CavroException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4)) __PYX_ERR(0, 31, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_10 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_10, __pyx_t_4, __pyx_n_s_InvalidHasher, __pyx_n_s_InvalidHasher, (PyObject *) NULL, __pyx_n_s_cavro, __pyx_kp_s_An_unknown_hash_method_was_requ); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_4 != __pyx_t_5) {
    if (unlikely((PyDict_SetItemString(__pyx_t_7, "__orig_bases__", __pyx_t_5) < 0))) __PYX_ERR(0, 31, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_10, __pyx_n_s_InvalidHasher, __pyx_t_4, __pyx_t_7, NULL, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_InvalidHasher, __pyx_t_5) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/error.pxi":37
 * 
 * 
 * class ExponentTooLarge(CavroException):             # <<<<<<<<<<<<<<
 *     """
 *     The exponent of a decimal value is too large to be represented in the given type
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CavroException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4)) __PYX_ERR(0, 37, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_4, __pyx_n_s_ExponentTooLarge, __pyx_n_s_ExponentTooLarge, (PyObject *) NULL, __pyx_n_s_cavro, __pyx_kp_s_The_exponent_of_a_decimal_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_4 != __pyx_t_10) {
    if (unlikely((PyDict_SetItemString(__pyx_t_5, "__orig_bases__", __pyx_t_10) < 0))) __PYX_ERR(0, 37, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_ExponentTooLarge, __pyx_t_4, __pyx_t_5, NULL, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ExponentTooLarge, __pyx_t_10) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/error.pxi":44
 * 
 * 
 * class CodecUnavailable(CavroException):             # <<<<<<<<<<<<<<
 *     """
 *     A requested codec (or codec in a container) is not available or is unknown.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CavroException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4)) __PYX_ERR(0, 44, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_4, __pyx_n_s_CodecUnavailable, __pyx_n_s_CodecUnavailable, (PyObject *) NULL, __pyx_n_s_cavro, __pyx_kp_s_A_requested_codec_or_codec_in_a); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (__pyx_t_4 != __pyx_t_7) {
    if (unlikely((PyDict_SetItemString(__pyx_t_10, "__orig_bases__", __pyx_t_7) < 0))) __PYX_ERR(0, 44, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_CodecUnavailable, __pyx_t_4, __pyx_t_10, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CodecUnavailable, __pyx_t_7) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/error.pxi":50
 * 
 * 
 * class CannotPromoteError(CavroException):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CavroException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4)) __PYX_ERR(0, 50, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_10 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_10, __pyx_t_4, __pyx_n_s_CannotPromoteError, __pyx_n_s_CannotPromoteError, (PyObject *) NULL, __pyx_n_s_cavro, __pyx_kp_s_A_schema_cannot_be_promoted_to); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_4 != __pyx_t_5) {
    if (unlikely((PyDict_SetItemString(__pyx_t_7, "__orig_bases__", __pyx_t_5) < 0))) __PYX_ERR(0, 50, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "src/error.pxi":61
 *     """
 * 
 *     def __init__(self, reader_type, writer_type, extra=None):             # <<<<<<<<<<<<<<
 *         self.reader_type = reader_type
 *         self.writer_type = writer_type
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_18CannotPromoteError_1__init__, 0, __pyx_n_s_CannotPromoteError___init, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__140)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_t_9);
  PyList_Append(__pyx_t_5, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__141);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "src/error.pxi":50
 * 
 * 
 * class CannotPromoteError(CavroException):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_10, __pyx_n_s_CannotPromoteError, __pyx_t_4, __pyx_t_7, NULL, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_5, __pyx_t_9) < 0) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CannotPromoteError, __pyx_t_9) < 0) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/error.pxi":71
 * 
 * 
 * class InvalidValue(CavroException, ValueError):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CavroException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4)) __PYX_ERR(0, 71, __pyx_L1_error);
  __Pyx_INCREF(__pyx_builtin_ValueError);
  __Pyx_GIVEREF(__pyx_builtin_ValueError);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_builtin_ValueError)) __PYX_ERR(0, 71, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_4, __pyx_n_s_InvalidValue, __pyx_n_s_InvalidValue, (PyObject *) NULL, __pyx_n_s_cavro, __pyx_kp_s_A_value_is_invalid_for_a_given); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__pyx_t_4 != __pyx_t_10) {
    if (unlikely((PyDict_SetItemString(__pyx_t_9, "__orig_bases__", __pyx_t_10) < 0))) __PYX_ERR(0, 71, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = PyList_New(0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "src/error.pxi":82
 *     """
 * 
 *     def __init__(self, value, dest_type, path=()):             # <<<<<<<<<<<<<<
 *         self.value = value
 *         self.dest_type = dest_type
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_12InvalidValue_1__init__, 0, __pyx_n_s_InvalidValue___init, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__143)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_t_5);
  PyList_Append(__pyx_t_10, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__144);
  if (__Pyx_SetNameInClass(__pyx_t_9, __pyx_n_s_init, __pyx_t_5) < 0) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/error.pxi":88
 *         super().__init__()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.schema_path:
 *             path_str = ".".join(str(p) for p in self.schema_path)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_12InvalidValue_3__str__, 0, __pyx_n_s_InvalidValue___str, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__146)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_9, __pyx_n_s_str, __pyx_t_5) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/error.pxi":71
 * 
 * 
 * class InvalidValue(CavroException, ValueError):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_InvalidValue, __pyx_t_4, __pyx_t_9, NULL, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_10, __pyx_t_5) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_InvalidValue, __pyx_t_5) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/option.pxi":3
 * 
 * 
 * NAME_RE_STRICT = re.compile(r'[A-Za-z_][A-Za-z0-9_]*')             # <<<<<<<<<<<<<<
 * NAME_RE_UNICODE = None
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_re); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_compile); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__147, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NAME_RE_STRICT, __pyx_t_4) < 0) __PYX_ERR(4, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/option.pxi":4
 * 
 * NAME_RE_STRICT = re.compile(r'[A-Za-z_][A-Za-z0-9_]*')
 * NAME_RE_UNICODE = None             # <<<<<<<<<<<<<<
 * 
 * cdef _get_unicode_name_re():
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NAME_RE_UNICODE, Py_None) < 0) __PYX_ERR(4, 4, __pyx_L1_error)

  /* "src/option.pxi":20
 * 
 * LOGICAL_TYPES = (
 *     DecimalType,             # <<<<<<<<<<<<<<
 *     UUIDStringType,
 *     UUIDFixedType,
 */
  __pyx_t_4 = PyTuple_New(8); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_DecimalType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_DecimalType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_ptype_5cavro_DecimalType))) __PYX_ERR(4, 20, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_UUIDStringType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_UUIDStringType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_ptype_5cavro_UUIDStringType))) __PYX_ERR(4, 20, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_UUIDFixedType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_UUIDFixedType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)__pyx_ptype_5cavro_UUIDFixedType))) __PYX_ERR(4, 20, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_Date);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_Date);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, ((PyObject *)__pyx_ptype_5cavro_Date))) __PYX_ERR(4, 20, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_TimeMillis);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_TimeMillis);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 4, ((PyObject *)__pyx_ptype_5cavro_TimeMillis))) __PYX_ERR(4, 20, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_TimeMicros);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_TimeMicros);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 5, ((PyObject *)__pyx_ptype_5cavro_TimeMicros))) __PYX_ERR(4, 20, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_TimestampMillis);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_TimestampMillis);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 6, ((PyObject *)__pyx_ptype_5cavro_TimestampMillis))) __PYX_ERR(4, 20, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_TimestampMicros);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_TimestampMicros);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 7, ((PyObject *)__pyx_ptype_5cavro_TimestampMicros))) __PYX_ERR(4, 20, __pyx_L1_error);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_LOGICAL_TYPES, __pyx_t_4) < 0) __PYX_ERR(4, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/option.pxi":30
 * )
 * 
 * class UnionDecodeOption(enum.Enum):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Enum); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7)) __PYX_ERR(4, 30, __pyx_L1_error);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PEP560_update_bases(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare(__pyx_t_9, __pyx_t_7, __pyx_n_s_UnionDecodeOption, __pyx_n_s_UnionDecodeOption, (PyObject *) NULL, __pyx_n_s_cavro, __pyx_kp_s_Controls_how_union_values_are_d); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_7 != __pyx_t_4) {
    if (unlikely((PyDict_SetItemString(__pyx_t_5, "__orig_bases__", __pyx_t_4) < 0))) __PYX_ERR(4, 30, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/option.pxi":41
 *     """
 * 
 *     RAW_VALUES = 0             # <<<<<<<<<<<<<<
 *     TYPE_TUPLE_IF_AMBIGUOUS = 1
 *     TYPE_TUPLE_IF_RECORD_AMBIGUOUS = 2
 */
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_RAW_VALUES, __pyx_int_0) < 0) __PYX_ERR(4, 41, __pyx_L1_error)

  /* "src/option.pxi":42
 * 
 *     RAW_VALUES = 0
 *     TYPE_TUPLE_IF_AMBIGUOUS = 1             # <<<<<<<<<<<<<<
 *     TYPE_TUPLE_IF_RECORD_AMBIGUOUS = 2
 *     TYPE_TUPLE_IF_RECORD = 3
 */
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_TYPE_TUPLE_IF_AMBIGUOUS, __pyx_int_1) < 0) __PYX_ERR(4, 42, __pyx_L1_error)

  /* "src/option.pxi":43
 *     RAW_VALUES = 0
 *     TYPE_TUPLE_IF_AMBIGUOUS = 1
 *     TYPE_TUPLE_IF_RECORD_AMBIGUOUS = 2             # <<<<<<<<<<<<<<
 *     TYPE_TUPLE_IF_RECORD = 3
 *     TYPE_TUPLE_ALWAYS = 4
 */
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_TYPE_TUPLE_IF_RECORD_AMBIGUOUS, __pyx_int_2) < 0) __PYX_ERR(4, 43, __pyx_L1_error)

  /* "src/option.pxi":44
 *     TYPE_TUPLE_IF_AMBIGUOUS = 1
 *     TYPE_TUPLE_IF_RECORD_AMBIGUOUS = 2
 *     TYPE_TUPLE_IF_RECORD = 3             # <<<<<<<<<<<<<<
 *     TYPE_TUPLE_ALWAYS = 4
 * 
 */
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_TYPE_TUPLE_IF_RECORD, __pyx_int_3) < 0) __PYX_ERR(4, 44, __pyx_L1_error)

  /* "src/option.pxi":45
 *     TYPE_TUPLE_IF_RECORD_AMBIGUOUS = 2
 *     TYPE_TUPLE_IF_RECORD = 3
 *     TYPE_TUPLE_ALWAYS = 4             # <<<<<<<<<<<<<<
 * 
 * _EMPTY_MP = MappingProxyType({})
 */
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_TYPE_TUPLE_ALWAYS, __pyx_int_4) < 0) __PYX_ERR(4, 45, __pyx_L1_error)

  /* "src/option.pxi":30
 * )
 * 
 * class UnionDecodeOption(enum.Enum):             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_9, __pyx_n_s_UnionDecodeOption, __pyx_t_7, __pyx_t_5, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UnionDecodeOption, __pyx_t_4) < 0) __PYX_ERR(4, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/option.pxi":47
 *     TYPE_TUPLE_ALWAYS = 4
 * 
 * _EMPTY_MP = MappingProxyType({})             # <<<<<<<<<<<<<<
 * 
 * def _eval(val):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_MappingProxyType); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EMPTY_MP, __pyx_t_5) < 0) __PYX_ERR(4, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/option.pxi":49
 * _EMPTY_MP = MappingProxyType({})
 * 
 * def _eval(val):             # <<<<<<<<<<<<<<
 *     if isinstance(val, str):
 *         if val == 'bint':
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_5_eval, 0, __pyx_n_s_eval_2, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__149)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_eval_2, __pyx_t_5) < 0) __PYX_ERR(4, 49, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/option.pxi":57
 * 
 * 
 * class _SignatureWrapper:             # <<<<<<<<<<<<<<
 * 
 *     '''
 */
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_SignatureWrapper, __pyx_n_s_SignatureWrapper, (PyObject *) NULL, __pyx_n_s_cavro, __pyx_kp_s_As_of_cython_3_0_2_dataclasses); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "src/option.pxi":70
 *     '''
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.__signature__ = None
 *         self.annotations = {}
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_17_SignatureWrapper_1__init__, 0, __pyx_n_s_SignatureWrapper___init, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__151)); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(4, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "src/option.pxi":74
 *         self.annotations = {}
 * 
 *     def _set_from_dataclass(self, cls):             # <<<<<<<<<<<<<<
 *         try:
 *             fields = [(f.name, f.type, f) for f in dataclasses.fields(cls)]
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_17_SignatureWrapper_3_set_from_dataclass, 0, __pyx_n_s_SignatureWrapper__set_from_data, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__154)); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_set_from_dataclass, __pyx_t_9) < 0) __PYX_ERR(4, 74, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "src/option.pxi":57
 * 
 * 
 * class _SignatureWrapper:             # <<<<<<<<<<<<<<
 * 
 *     '''
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(((PyObject*)&PyType_Type), __pyx_n_s_SignatureWrapper, __pyx_empty_tuple, __pyx_t_5, NULL, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SignatureWrapper, __pyx_t_9) < 0) __PYX_ERR(4, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/option.pxi":86
 * 
 * 
 * _OptionsWrapper = _SignatureWrapper()             # <<<<<<<<<<<<<<
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_SignatureWrapper); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OptionsWrapper, __pyx_t_9) < 0) __PYX_ERR(4, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "src/option.pxi":202
 *     """
 * 
 *     __wrapped__ = _OptionsWrapper             # <<<<<<<<<<<<<<
 *     __annotations__ = _OptionsWrapper.annotations
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_OptionsWrapper); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_wrapped, __pyx_t_9) < 0) __PYX_ERR(4, 202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Options);

  /* "src/option.pxi":203
 * 
 *     __wrapped__ = _OptionsWrapper
 *     __annotations__ = _OptionsWrapper.annotations             # <<<<<<<<<<<<<<
 * 
 *     fingerprint_returns_digest: bint = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_OptionsWrapper); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_annotations); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_annotations_2, __pyx_t_5) < 0) __PYX_ERR(4, 203, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Options);

  /* "src/option.pxi":285
 *     externally_defined_types: MappingProxyType = _EMPTY_MP
 * 
 *     def replace(self, **changes) -> Options:             # <<<<<<<<<<<<<<
 *         '''
 *         Return a copy of the options with the specified fields replaced.
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_Options) < 0) __PYX_ERR(4, 285, __pyx_L1_error)
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7Options_1replace, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Options_replace, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__156)); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_replace, __pyx_t_9) < 0) __PYX_ERR(4, 285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Options);

  /* "src/option.pxi":292
 *         return dataclasses.replace(self, **changes)
 * 
 *     def with_logical_types(self, *logical_types):             # <<<<<<<<<<<<<<
 *         '''
 *         Return a copy of the options with additional logical types added.
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7Options_3with_logical_types, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Options_with_logical_types, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__158)); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_with_logical_types, __pyx_t_9) < 0) __PYX_ERR(4, 292, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Options);

  /* "src/option.pxi":299
 *         return self.replace(logical_types=new_types)
 * 
 *     def with_external_types(self, named_types: dict[str, AvroType]) -> Options:             # <<<<<<<<<<<<<<
 *         '''
 *         Return a copy of the options with pre-parsed external named types added
 */
  __pyx_t_9 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_named_types, __pyx_kp_s_dict_str_AvroType) < 0) __PYX_ERR(4, 299, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_return, __pyx_n_s_Options) < 0) __PYX_ERR(4, 299, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7Options_5with_external_types, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Options_with_external_types, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__160)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_with_external_types, __pyx_t_5) < 0) __PYX_ERR(4, 299, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Options);

  /* "src/option.pxi":323
 *         return self.missing_values_can_be_empty_container or self.missing_values_can_be_null
 * 
 *     def equals(self, Options other: Options, ignore: Sequence[str]=()) -> bool:             # <<<<<<<<<<<<<<
 *         '''
 *         Compares two Options objects, this is equivalent to
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_other, __pyx_n_s_Options) < 0) __PYX_ERR(4, 323, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_ignore, __pyx_kp_s_Sequence_str) < 0) __PYX_ERR(4, 323, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(4, 323, __pyx_L1_error)
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7Options_7equals, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Options_equals, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__162)); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__163);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_equals, __pyx_t_9) < 0) __PYX_ERR(4, 323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Options);

  /* "src/option.pxi":339
 *         return True
 * 
 *     def diff(self, Options other: Options, ignore: Sequence[str]=()) -> dict[str, tuple[object, object]]:             # <<<<<<<<<<<<<<
 *         '''
 *         Returns a dictionary containing just the fields that are different between two Options objects.
 */
  __pyx_t_9 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_other, __pyx_n_s_Options) < 0) __PYX_ERR(4, 339, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ignore, __pyx_kp_s_Sequence_str) < 0) __PYX_ERR(4, 339, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_return, __pyx_kp_s_dict_str_tuple_object_object) < 0) __PYX_ERR(4, 339, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7Options_9diff, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Options_diff, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__165)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__163);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_diff, __pyx_t_5) < 0) __PYX_ERR(4, 339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Options);

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_DataclassParams); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(11); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_eq, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_order, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_unsafe_hash, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_frozen, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_kw_only, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_kw_only, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_match_args, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_slots_2, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_weakref_slot, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_params, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Options);

  /* "src/option.pxi":215
 *     record_encode_use_defaults: bint = True
 *     allow_tuple_notation: bint = False
 *     union_decodes_to: UnionDecodeOption = UnionDecodeOption.RAW_VALUES             # <<<<<<<<<<<<<<
 *     union_json_encodes_type_name: bint = True
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_UnionDecodeOption); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_RAW_VALUES); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_XGOTREF(__pyx_dataclass_dflt5cavro_Options5cavro_union_decodes_to);
  __Pyx_DECREF_SET(__pyx_dataclass_dflt5cavro_Options5cavro_union_decodes_to, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(63); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":205
 *     __annotations__ = _OptionsWrapper.annotations
 * 
 *     fingerprint_returns_digest: bint = False             # <<<<<<<<<<<<<<
 *     canonical_form_repeat_fixed: bint = False
 *     canonical_form_repeat_enum: bint = False
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_fingerprint_returns_digest, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":206
 * 
 *     fingerprint_returns_digest: bint = False
 *     canonical_form_repeat_fixed: bint = False             # <<<<<<<<<<<<<<
 *     canonical_form_repeat_enum: bint = False
 * 
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_canonical_form_repeat_fixed, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":207
 *     fingerprint_returns_digest: bint = False
 *     canonical_form_repeat_fixed: bint = False
 *     canonical_form_repeat_enum: bint = False             # <<<<<<<<<<<<<<
 * 
 *     record_can_encode_dict: bint = True
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_canonical_form_repeat_enum, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":209
 *     canonical_form_repeat_enum: bint = False
 * 
 *     record_can_encode_dict: bint = True             # <<<<<<<<<<<<<<
 *     record_values_type_hint: bint = False
 *     record_decodes_to_dict: bint = False
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_record_can_encode_dict, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":210
 * 
 *     record_can_encode_dict: bint = True
 *     record_values_type_hint: bint = False             # <<<<<<<<<<<<<<
 *     record_decodes_to_dict: bint = False
 *     record_allow_extra_fields: bint = True
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_record_values_type_hint, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":211
 *     record_can_encode_dict: bint = True
 *     record_values_type_hint: bint = False
 *     record_decodes_to_dict: bint = False             # <<<<<<<<<<<<<<
 *     record_allow_extra_fields: bint = True
 *     record_encode_use_defaults: bint = True
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_record_decodes_to_dict, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":212
 *     record_values_type_hint: bint = False
 *     record_decodes_to_dict: bint = False
 *     record_allow_extra_fields: bint = True             # <<<<<<<<<<<<<<
 *     record_encode_use_defaults: bint = True
 *     allow_tuple_notation: bint = False
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_record_allow_extra_fields, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":213
 *     record_decodes_to_dict: bint = False
 *     record_allow_extra_fields: bint = True
 *     record_encode_use_defaults: bint = True             # <<<<<<<<<<<<<<
 *     allow_tuple_notation: bint = False
 *     union_decodes_to: UnionDecodeOption = UnionDecodeOption.RAW_VALUES
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_record_encode_use_defaults, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":214
 *     record_allow_extra_fields: bint = True
 *     record_encode_use_defaults: bint = True
 *     allow_tuple_notation: bint = False             # <<<<<<<<<<<<<<
 *     union_decodes_to: UnionDecodeOption = UnionDecodeOption.RAW_VALUES
 *     union_json_encodes_type_name: bint = True
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_tuple_notation, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":215
 *     record_encode_use_defaults: bint = True
 *     allow_tuple_notation: bint = False
 *     union_decodes_to: UnionDecodeOption = UnionDecodeOption.RAW_VALUES             # <<<<<<<<<<<<<<
 *     union_json_encodes_type_name: bint = True
 * 
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, __pyx_dataclass_dflt5cavro_Options5cavro_union_decodes_to) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_union_decodes_to, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":216
 *     allow_tuple_notation: bint = False
 *     union_decodes_to: UnionDecodeOption = UnionDecodeOption.RAW_VALUES
 *     union_json_encodes_type_name: bint = True             # <<<<<<<<<<<<<<
 * 
 *     allow_primitive_name_collision: bint = False
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_union_json_encodes_type_name, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":218
 *     union_json_encodes_type_name: bint = True
 * 
 *     allow_primitive_name_collision: bint = False             # <<<<<<<<<<<<<<
 *     allow_primitive_names_in_namespaces: bint = False
 * 
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_primitive_name_collision, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":219
 * 
 *     allow_primitive_name_collision: bint = False
 *     allow_primitive_names_in_namespaces: bint = False             # <<<<<<<<<<<<<<
 * 
 *     named_type_names_must_be_unique: bint = True
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_primitive_names_in_namespa, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":221
 *     allow_primitive_names_in_namespaces: bint = False
 * 
 *     named_type_names_must_be_unique: bint = True             # <<<<<<<<<<<<<<
 *     enum_symbols_must_be_unique: bint = True
 *     enforce_enum_symbol_name_rules: bint = True
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_named_type_names_must_be_unique, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":222
 * 
 *     named_type_names_must_be_unique: bint = True
 *     enum_symbols_must_be_unique: bint = True             # <<<<<<<<<<<<<<
 *     enforce_enum_symbol_name_rules: bint = True
 *     enforce_type_name_rules: bint = True
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_enum_symbols_must_be_unique, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":223
 *     named_type_names_must_be_unique: bint = True
 *     enum_symbols_must_be_unique: bint = True
 *     enforce_enum_symbol_name_rules: bint = True             # <<<<<<<<<<<<<<
 *     enforce_type_name_rules: bint = True
 *     enforce_namespace_name_rules: bint = True
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_enforce_enum_symbol_name_rules, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":224
 *     enum_symbols_must_be_unique: bint = True
 *     enforce_enum_symbol_name_rules: bint = True
 *     enforce_type_name_rules: bint = True             # <<<<<<<<<<<<<<
 *     enforce_namespace_name_rules: bint = True
 *     record_fields_must_be_unique: bint = True
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_enforce_type_name_rules, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":225
 *     enforce_enum_symbol_name_rules: bint = True
 *     enforce_type_name_rules: bint = True
 *     enforce_namespace_name_rules: bint = True             # <<<<<<<<<<<<<<
 *     record_fields_must_be_unique: bint = True
 *     ascii_name_rules: bint = True
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_enforce_namespace_name_rules, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":226
 *     enforce_type_name_rules: bint = True
 *     enforce_namespace_name_rules: bint = True
 *     record_fields_must_be_unique: bint = True             # <<<<<<<<<<<<<<
 *     ascii_name_rules: bint = True
 * 
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_record_fields_must_be_unique, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":227
 *     enforce_namespace_name_rules: bint = True
 *     record_fields_must_be_unique: bint = True
 *     ascii_name_rules: bint = True             # <<<<<<<<<<<<<<
 * 
 *     missing_values_can_be_null: bint = False
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_ascii_name_rules, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":229
 *     ascii_name_rules: bint = True
 * 
 *     missing_values_can_be_null: bint = False             # <<<<<<<<<<<<<<
 *     missing_values_can_be_empty_container: bint = False
 * 
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_missing_values_can_be_null, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":230
 * 
 *     missing_values_can_be_null: bint = False
 *     missing_values_can_be_empty_container: bint = False             # <<<<<<<<<<<<<<
 * 
 *     allow_false_values_for_null: bint = False
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_missing_values_can_be_empty_cont, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":232
 *     missing_values_can_be_empty_container: bint = False
 * 
 *     allow_false_values_for_null: bint = False             # <<<<<<<<<<<<<<
 *     allow_empty_unions: bint = False
 *     allow_nested_unions: bint = False
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_false_values_for_null, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":233
 * 
 *     allow_false_values_for_null: bint = False
 *     allow_empty_unions: bint = False             # <<<<<<<<<<<<<<
 *     allow_nested_unions: bint = False
 *     allow_duplicate_union_types: bint = False
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_empty_unions, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":234
 *     allow_false_values_for_null: bint = False
 *     allow_empty_unions: bint = False
 *     allow_nested_unions: bint = False             # <<<<<<<<<<<<<<
 *     allow_duplicate_union_types: bint = False
 *     allow_union_default_any_member: bint = False
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_nested_unions, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":235
 *     allow_empty_unions: bint = False
 *     allow_nested_unions: bint = False
 *     allow_duplicate_union_types: bint = False             # <<<<<<<<<<<<<<
 *     allow_union_default_any_member: bint = False
 *     allow_aliases_to_be_string: bint = False
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_duplicate_union_types, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":236
 *     allow_nested_unions: bint = False
 *     allow_duplicate_union_types: bint = False
 *     allow_union_default_any_member: bint = False             # <<<<<<<<<<<<<<
 *     allow_aliases_to_be_string: bint = False
 * 
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_union_default_any_member, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":237
 *     allow_duplicate_union_types: bint = False
 *     allow_union_default_any_member: bint = False
 *     allow_aliases_to_be_string: bint = False             # <<<<<<<<<<<<<<
 * 
 *     coerce_values_to_int: bint = False
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_aliases_to_be_string, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":239
 *     allow_aliases_to_be_string: bint = False
 * 
 *     coerce_values_to_int: bint = False             # <<<<<<<<<<<<<<
 *     coerce_values_to_float: bint = False
 *     coerce_int_to_float: bint = True
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_coerce_values_to_int, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":240
 * 
 *     coerce_values_to_int: bint = False
 *     coerce_values_to_float: bint = False             # <<<<<<<<<<<<<<
 *     coerce_int_to_float: bint = True
 *     coerce_values_to_boolean: bint = False
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_coerce_values_to_float, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":241
 *     coerce_values_to_int: bint = False
 *     coerce_values_to_float: bint = False
 *     coerce_int_to_float: bint = True             # <<<<<<<<<<<<<<
 *     coerce_values_to_boolean: bint = False
 *     coerce_values_to_str: bint = False
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_coerce_int_to_float, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":242
 *     coerce_values_to_float: bint = False
 *     coerce_int_to_float: bint = True
 *     coerce_values_to_boolean: bint = False             # <<<<<<<<<<<<<<
 *     coerce_values_to_str: bint = False
 *     bytes_codec: str = None
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_coerce_values_to_boolean, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":243
 *     coerce_int_to_float: bint = True
 *     coerce_values_to_boolean: bint = False
 *     coerce_values_to_str: bint = False             # <<<<<<<<<<<<<<
 *     bytes_codec: str = None
 *     fixed_codec: str = None
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_coerce_values_to_str, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":244
 *     coerce_values_to_boolean: bint = False
 *     coerce_values_to_str: bint = False
 *     bytes_codec: str = None             # <<<<<<<<<<<<<<
 *     fixed_codec: str = None
 * 
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_bytes_codec, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":245
 *     coerce_values_to_str: bint = False
 *     bytes_codec: str = None
 *     fixed_codec: str = None             # <<<<<<<<<<<<<<
 * 
 *     null_pad_fixed: bint = False
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_fixed_codec, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":247
 *     fixed_codec: str = None
 * 
 *     null_pad_fixed: bint = False             # <<<<<<<<<<<<<<
 *     truncate_fixed: bint = False
 *     clamp_int_overflow: bint = False
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_null_pad_fixed, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":248
 * 
 *     null_pad_fixed: bint = False
 *     truncate_fixed: bint = False             # <<<<<<<<<<<<<<
 *     clamp_int_overflow: bint = False
 *     clamp_float_overflow: bint = False
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_truncate_fixed, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":249
 *     null_pad_fixed: bint = False
 *     truncate_fixed: bint = False
 *     clamp_int_overflow: bint = False             # <<<<<<<<<<<<<<
 *     clamp_float_overflow: bint = False
 *     float_out_of_range_inf: bint = False
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_clamp_int_overflow, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":250
 *     truncate_fixed: bint = False
 *     clamp_int_overflow: bint = False
 *     clamp_float_overflow: bint = False             # <<<<<<<<<<<<<<
 *     float_out_of_range_inf: bint = False
 * 
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_clamp_float_overflow, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":251
 *     clamp_int_overflow: bint = False
 *     clamp_float_overflow: bint = False
 *     float_out_of_range_inf: bint = False             # <<<<<<<<<<<<<<
 * 
 *     bytes_default_value_utf8: bint = False  # Avro 1.11/12 utf8 encode the default value for bytes (rather than latin1)
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_float_out_of_range_inf, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":253
 *     float_out_of_range_inf: bint = False
 * 
 *     bytes_default_value_utf8: bint = False  # Avro 1.11/12 utf8 encode the default value for bytes (rather than latin1)             # <<<<<<<<<<<<<<
 *     string_types_default_unchanged: bint = False # Avro < 1.11 pass default value back unmodified for bytes/fixed etc..
 * 
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_bytes_default_value_utf8, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":254
 * 
 *     bytes_default_value_utf8: bint = False  # Avro 1.11/12 utf8 encode the default value for bytes (rather than latin1)
 *     string_types_default_unchanged: bint = False # Avro < 1.11 pass default value back unmodified for bytes/fixed etc..             # <<<<<<<<<<<<<<
 * 
 *     decimal_check_exp_overflow: bint = True
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_string_types_default_unchanged, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":256
 *     string_types_default_unchanged: bint = False # Avro < 1.11 pass default value back unmodified for bytes/fixed etc..
 * 
 *     decimal_check_exp_overflow: bint = True             # <<<<<<<<<<<<<<
 * 
 *     types_str_to_schema: bint = False
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_decimal_check_exp_overflow, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":258
 *     decimal_check_exp_overflow: bint = True
 * 
 *     types_str_to_schema: bint = False             # <<<<<<<<<<<<<<
 *     logical_types: tuple[LogicalType] = LOGICAL_TYPES
 * 
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_types_str_to_schema, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":259
 * 
 *     types_str_to_schema: bint = False
 *     logical_types: tuple[LogicalType] = LOGICAL_TYPES             # <<<<<<<<<<<<<<
 * 
 *     adapt_record_types: bint = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_LOGICAL_TYPES); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_logical_types, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":261
 *     logical_types: tuple[LogicalType] = LOGICAL_TYPES
 * 
 *     adapt_record_types: bint = True             # <<<<<<<<<<<<<<
 *     return_uuid_object: bint = True
 * 
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_adapt_record_types, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":262
 * 
 *     adapt_record_types: bint = True
 *     return_uuid_object: bint = True             # <<<<<<<<<<<<<<
 * 
 *     allow_error_type: bint = False
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return_uuid_object, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":264
 *     return_uuid_object: bint = True
 * 
 *     allow_error_type: bint = False             # <<<<<<<<<<<<<<
 *     allow_leading_dot_in_names: bint = True
 * 
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_error_type, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":265
 * 
 *     allow_error_type: bint = False
 *     allow_leading_dot_in_names: bint = True             # <<<<<<<<<<<<<<
 * 
 *     naive_dt_assume_utc: bint = False
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_leading_dot_in_names, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":267
 *     allow_leading_dot_in_names: bint = True
 * 
 *     naive_dt_assume_utc: bint = False             # <<<<<<<<<<<<<<
 *     alternate_timestamp_millis_encoding: bint = False
 *     date_type_accepts_string: bint = False
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_naive_dt_assume_utc, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":268
 * 
 *     naive_dt_assume_utc: bint = False
 *     alternate_timestamp_millis_encoding: bint = False             # <<<<<<<<<<<<<<
 *     date_type_accepts_string: bint = False
 *     raise_on_invalid_logical: bint = False
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_alternate_timestamp_millis_encod, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":269
 *     naive_dt_assume_utc: bint = False
 *     alternate_timestamp_millis_encoding: bint = False
 *     date_type_accepts_string: bint = False             # <<<<<<<<<<<<<<
 *     raise_on_invalid_logical: bint = False
 *     inline_namespaces: bint = False
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_date_type_accepts_string, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":270
 *     alternate_timestamp_millis_encoding: bint = False
 *     date_type_accepts_string: bint = False
 *     raise_on_invalid_logical: bint = False             # <<<<<<<<<<<<<<
 *     inline_namespaces: bint = False
 *     expand_types_in_schema: bint = False
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_raise_on_invalid_logical, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":271
 *     date_type_accepts_string: bint = False
 *     raise_on_invalid_logical: bint = False
 *     inline_namespaces: bint = False             # <<<<<<<<<<<<<<
 *     expand_types_in_schema: bint = False
 *     unicode_errors: str = 'strict'
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_inline_namespaces, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":272
 *     raise_on_invalid_logical: bint = False
 *     inline_namespaces: bint = False
 *     expand_types_in_schema: bint = False             # <<<<<<<<<<<<<<
 *     unicode_errors: str = 'strict'
 * 
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_expand_types_in_schema, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, __pyx_n_u_strict) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_unicode_errors, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":275
 *     unicode_errors: str = 'strict'
 * 
 *     container_fill_blocks: bint = False             # <<<<<<<<<<<<<<
 *     defer_schema_promotion_errors: bint = False
 * 
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_container_fill_blocks, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":276
 * 
 *     container_fill_blocks: bint = False
 *     defer_schema_promotion_errors: bint = False             # <<<<<<<<<<<<<<
 * 
 *     invalid_value_includes_record_name: bint = False
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_defer_schema_promotion_errors, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":278
 *     defer_schema_promotion_errors: bint = False
 * 
 *     invalid_value_includes_record_name: bint = False             # <<<<<<<<<<<<<<
 *     invalid_value_include_array_index: bint = True
 *     invalid_value_include_map_key: bint = True
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_invalid_value_includes_record_na, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":279
 * 
 *     invalid_value_includes_record_name: bint = False
 *     invalid_value_include_array_index: bint = True             # <<<<<<<<<<<<<<
 *     invalid_value_include_map_key: bint = True
 *     allow_invalid_default_values: bint = False
 */
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_invalid_value_include_array_inde, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_field); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":280
 *     invalid_value_includes_record_name: bint = False
 *     invalid_value_include_array_index: bint = True
 *     invalid_value_include_map_key: bint = True             # <<<<<<<<<<<<<<
 *     allow_invalid_default_values: bint = False
 * 
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_default, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_4 = __Pyx_DataclassesCallHelper(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_invalid_value_include_map_key, __pyx_t_4) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_field); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":281
 *     invalid_value_include_array_index: bint = True
 *     invalid_value_include_map_key: bint = True
 *     allow_invalid_default_values: bint = False             # <<<<<<<<<<<<<<
 * 
 *     externally_defined_types: MappingProxyType = _EMPTY_MP
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_default, Py_False) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_9 = __Pyx_DataclassesCallHelper(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_invalid_default_values, __pyx_t_9) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_field); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_repr, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_hash, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_init_2, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_compare, Py_True) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_metadata, Py_None) < 0) __PYX_ERR(4, 89, __pyx_L1_error)

  /* "src/option.pxi":283
 *     allow_invalid_default_values: bint = False
 * 
 *     externally_defined_types: MappingProxyType = _EMPTY_MP             # <<<<<<<<<<<<<<
 * 
 *     def replace(self, **changes) -> Options:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_EMPTY_MP); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_default, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/option.pxi":89
 * 
 * @dataclasses.dataclass(frozen=True, kw_only=True)
 * cdef class Options:             # <<<<<<<<<<<<<<
 * 
 *     """
 */
  __pyx_t_7 = __Pyx_DataclassesCallHelper(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_externally_defined_types, __pyx_t_7) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields, __pyx_t_5) < 0) __PYX_ERR(4, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Options);

  /* "(tree fragment)":1
 * __dataclass_fields__['fingerprint_returns_digest'].name = 'fingerprint_returns_digest'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['fingerprint_returns_digest'].type = PLACEHOLDER_fingerprint_returns_digest
 * __dataclass_fields__['fingerprint_returns_digest']._field_type = PLACEHOLDER_FIELD_TYPE_fingerprint_returns_digest
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_fingerprint_returns_digest); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_fingerprint_returns_digest) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":2
 * __dataclass_fields__['fingerprint_returns_digest'].name = 'fingerprint_returns_digest'
 * __dataclass_fields__['fingerprint_returns_digest'].type = PLACEHOLDER_fingerprint_returns_digest             # <<<<<<<<<<<<<<
 * __dataclass_fields__['fingerprint_returns_digest']._field_type = PLACEHOLDER_FIELD_TYPE_fingerprint_returns_digest
 * __dataclass_fields__['canonical_form_repeat_fixed'].name = 'canonical_form_repeat_fixed'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_fingerprint_returns_digest); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * __dataclass_fields__['fingerprint_returns_digest'].name = 'fingerprint_returns_digest'
 * __dataclass_fields__['fingerprint_returns_digest'].type = PLACEHOLDER_fingerprint_returns_digest
 * __dataclass_fields__['fingerprint_returns_digest']._field_type = PLACEHOLDER_FIELD_TYPE_fingerprint_returns_digest             # <<<<<<<<<<<<<<
 * __dataclass_fields__['canonical_form_repeat_fixed'].name = 'canonical_form_repeat_fixed'
 * __dataclass_fields__['canonical_form_repeat_fixed'].type = PLACEHOLDER_canonical_form_repeat_fixed
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_fingerprint_returns_digest); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":4
 * __dataclass_fields__['fingerprint_returns_digest'].type = PLACEHOLDER_fingerprint_returns_digest
 * __dataclass_fields__['fingerprint_returns_digest']._field_type = PLACEHOLDER_FIELD_TYPE_fingerprint_returns_digest
 * __dataclass_fields__['canonical_form_repeat_fixed'].name = 'canonical_form_repeat_fixed'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['canonical_form_repeat_fixed'].type = PLACEHOLDER_canonical_form_repeat_fixed
 * __dataclass_fields__['canonical_form_repeat_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_canonical_form_repeat_fixed
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_canonical_form_repeat_fixed); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_canonical_form_repeat_fixed) < 0) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":5
 * __dataclass_fields__['fingerprint_returns_digest']._field_type = PLACEHOLDER_FIELD_TYPE_fingerprint_returns_digest
 * __dataclass_fields__['canonical_form_repeat_fixed'].name = 'canonical_form_repeat_fixed'
 * __dataclass_fields__['canonical_form_repeat_fixed'].type = PLACEHOLDER_canonical_form_repeat_fixed             # <<<<<<<<<<<<<<
 * __dataclass_fields__['canonical_form_repeat_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_canonical_form_repeat_fixed
 * __dataclass_fields__['canonical_form_repeat_enum'].name = 'canonical_form_repeat_enum'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_canonical_form_repeat_fixed); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":6
 * __dataclass_fields__['canonical_form_repeat_fixed'].name = 'canonical_form_repeat_fixed'
 * __dataclass_fields__['canonical_form_repeat_fixed'].type = PLACEHOLDER_canonical_form_repeat_fixed
 * __dataclass_fields__['canonical_form_repeat_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_canonical_form_repeat_fixed             # <<<<<<<<<<<<<<
 * __dataclass_fields__['canonical_form_repeat_enum'].name = 'canonical_form_repeat_enum'
 * __dataclass_fields__['canonical_form_repeat_enum'].type = PLACEHOLDER_canonical_form_repeat_enum
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_canonical_form_repeat_fixed); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":7
 * __dataclass_fields__['canonical_form_repeat_fixed'].type = PLACEHOLDER_canonical_form_repeat_fixed
 * __dataclass_fields__['canonical_form_repeat_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_canonical_form_repeat_fixed
 * __dataclass_fields__['canonical_form_repeat_enum'].name = 'canonical_form_repeat_enum'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['canonical_form_repeat_enum'].type = PLACEHOLDER_canonical_form_repeat_enum
 * __dataclass_fields__['canonical_form_repeat_enum']._field_type = PLACEHOLDER_FIELD_TYPE_canonical_form_repeat_enum
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_canonical_form_repeat_enum); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_canonical_form_repeat_enum) < 0) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":8
 * __dataclass_fields__['canonical_form_repeat_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_canonical_form_repeat_fixed
 * __dataclass_fields__['canonical_form_repeat_enum'].name = 'canonical_form_repeat_enum'
 * __dataclass_fields__['canonical_form_repeat_enum'].type = PLACEHOLDER_canonical_form_repeat_enum             # <<<<<<<<<<<<<<
 * __dataclass_fields__['canonical_form_repeat_enum']._field_type = PLACEHOLDER_FIELD_TYPE_canonical_form_repeat_enum
 * __dataclass_fields__['record_can_encode_dict'].name = 'record_can_encode_dict'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_canonical_form_repeat_enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":9
 * __dataclass_fields__['canonical_form_repeat_enum'].name = 'canonical_form_repeat_enum'
 * __dataclass_fields__['canonical_form_repeat_enum'].type = PLACEHOLDER_canonical_form_repeat_enum
 * __dataclass_fields__['canonical_form_repeat_enum']._field_type = PLACEHOLDER_FIELD_TYPE_canonical_form_repeat_enum             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_can_encode_dict'].name = 'record_can_encode_dict'
 * __dataclass_fields__['record_can_encode_dict'].type = PLACEHOLDER_record_can_encode_dict
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_canonical_form_repeat_enum); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":10
 * __dataclass_fields__['canonical_form_repeat_enum'].type = PLACEHOLDER_canonical_form_repeat_enum
 * __dataclass_fields__['canonical_form_repeat_enum']._field_type = PLACEHOLDER_FIELD_TYPE_canonical_form_repeat_enum
 * __dataclass_fields__['record_can_encode_dict'].name = 'record_can_encode_dict'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_can_encode_dict'].type = PLACEHOLDER_record_can_encode_dict
 * __dataclass_fields__['record_can_encode_dict']._field_type = PLACEHOLDER_FIELD_TYPE_record_can_encode_dict
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_record_can_encode_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_record_can_encode_dict) < 0) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":11
 * __dataclass_fields__['canonical_form_repeat_enum']._field_type = PLACEHOLDER_FIELD_TYPE_canonical_form_repeat_enum
 * __dataclass_fields__['record_can_encode_dict'].name = 'record_can_encode_dict'
 * __dataclass_fields__['record_can_encode_dict'].type = PLACEHOLDER_record_can_encode_dict             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_can_encode_dict']._field_type = PLACEHOLDER_FIELD_TYPE_record_can_encode_dict
 * __dataclass_fields__['record_values_type_hint'].name = 'record_values_type_hint'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_record_can_encode_dict); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":12
 * __dataclass_fields__['record_can_encode_dict'].name = 'record_can_encode_dict'
 * __dataclass_fields__['record_can_encode_dict'].type = PLACEHOLDER_record_can_encode_dict
 * __dataclass_fields__['record_can_encode_dict']._field_type = PLACEHOLDER_FIELD_TYPE_record_can_encode_dict             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_values_type_hint'].name = 'record_values_type_hint'
 * __dataclass_fields__['record_values_type_hint'].type = PLACEHOLDER_record_values_type_hint
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_record_can_encode_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":13
 * __dataclass_fields__['record_can_encode_dict'].type = PLACEHOLDER_record_can_encode_dict
 * __dataclass_fields__['record_can_encode_dict']._field_type = PLACEHOLDER_FIELD_TYPE_record_can_encode_dict
 * __dataclass_fields__['record_values_type_hint'].name = 'record_values_type_hint'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_values_type_hint'].type = PLACEHOLDER_record_values_type_hint
 * __dataclass_fields__['record_values_type_hint']._field_type = PLACEHOLDER_FIELD_TYPE_record_values_type_hint
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_record_values_type_hint); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_record_values_type_hint) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":14
 * __dataclass_fields__['record_can_encode_dict']._field_type = PLACEHOLDER_FIELD_TYPE_record_can_encode_dict
 * __dataclass_fields__['record_values_type_hint'].name = 'record_values_type_hint'
 * __dataclass_fields__['record_values_type_hint'].type = PLACEHOLDER_record_values_type_hint             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_values_type_hint']._field_type = PLACEHOLDER_FIELD_TYPE_record_values_type_hint
 * __dataclass_fields__['record_decodes_to_dict'].name = 'record_decodes_to_dict'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_record_values_type_hint); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":15
 * __dataclass_fields__['record_values_type_hint'].name = 'record_values_type_hint'
 * __dataclass_fields__['record_values_type_hint'].type = PLACEHOLDER_record_values_type_hint
 * __dataclass_fields__['record_values_type_hint']._field_type = PLACEHOLDER_FIELD_TYPE_record_values_type_hint             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_decodes_to_dict'].name = 'record_decodes_to_dict'
 * __dataclass_fields__['record_decodes_to_dict'].type = PLACEHOLDER_record_decodes_to_dict
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_record_values_type_hint); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":16
 * __dataclass_fields__['record_values_type_hint'].type = PLACEHOLDER_record_values_type_hint
 * __dataclass_fields__['record_values_type_hint']._field_type = PLACEHOLDER_FIELD_TYPE_record_values_type_hint
 * __dataclass_fields__['record_decodes_to_dict'].name = 'record_decodes_to_dict'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_decodes_to_dict'].type = PLACEHOLDER_record_decodes_to_dict
 * __dataclass_fields__['record_decodes_to_dict']._field_type = PLACEHOLDER_FIELD_TYPE_record_decodes_to_dict
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_record_decodes_to_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_record_decodes_to_dict) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":17
 * __dataclass_fields__['record_values_type_hint']._field_type = PLACEHOLDER_FIELD_TYPE_record_values_type_hint
 * __dataclass_fields__['record_decodes_to_dict'].name = 'record_decodes_to_dict'
 * __dataclass_fields__['record_decodes_to_dict'].type = PLACEHOLDER_record_decodes_to_dict             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_decodes_to_dict']._field_type = PLACEHOLDER_FIELD_TYPE_record_decodes_to_dict
 * __dataclass_fields__['record_allow_extra_fields'].name = 'record_allow_extra_fields'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_record_decodes_to_dict); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":18
 * __dataclass_fields__['record_decodes_to_dict'].name = 'record_decodes_to_dict'
 * __dataclass_fields__['record_decodes_to_dict'].type = PLACEHOLDER_record_decodes_to_dict
 * __dataclass_fields__['record_decodes_to_dict']._field_type = PLACEHOLDER_FIELD_TYPE_record_decodes_to_dict             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_allow_extra_fields'].name = 'record_allow_extra_fields'
 * __dataclass_fields__['record_allow_extra_fields'].type = PLACEHOLDER_record_allow_extra_fields
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_record_decodes_to_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":19
 * __dataclass_fields__['record_decodes_to_dict'].type = PLACEHOLDER_record_decodes_to_dict
 * __dataclass_fields__['record_decodes_to_dict']._field_type = PLACEHOLDER_FIELD_TYPE_record_decodes_to_dict
 * __dataclass_fields__['record_allow_extra_fields'].name = 'record_allow_extra_fields'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_allow_extra_fields'].type = PLACEHOLDER_record_allow_extra_fields
 * __dataclass_fields__['record_allow_extra_fields']._field_type = PLACEHOLDER_FIELD_TYPE_record_allow_extra_fields
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_record_allow_extra_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_record_allow_extra_fields) < 0) __PYX_ERR(1, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":20
 * __dataclass_fields__['record_decodes_to_dict']._field_type = PLACEHOLDER_FIELD_TYPE_record_decodes_to_dict
 * __dataclass_fields__['record_allow_extra_fields'].name = 'record_allow_extra_fields'
 * __dataclass_fields__['record_allow_extra_fields'].type = PLACEHOLDER_record_allow_extra_fields             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_allow_extra_fields']._field_type = PLACEHOLDER_FIELD_TYPE_record_allow_extra_fields
 * __dataclass_fields__['record_encode_use_defaults'].name = 'record_encode_use_defaults'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_record_allow_extra_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":21
 * __dataclass_fields__['record_allow_extra_fields'].name = 'record_allow_extra_fields'
 * __dataclass_fields__['record_allow_extra_fields'].type = PLACEHOLDER_record_allow_extra_fields
 * __dataclass_fields__['record_allow_extra_fields']._field_type = PLACEHOLDER_FIELD_TYPE_record_allow_extra_fields             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_encode_use_defaults'].name = 'record_encode_use_defaults'
 * __dataclass_fields__['record_encode_use_defaults'].type = PLACEHOLDER_record_encode_use_defaults
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_record_allow_extra_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":22
 * __dataclass_fields__['record_allow_extra_fields'].type = PLACEHOLDER_record_allow_extra_fields
 * __dataclass_fields__['record_allow_extra_fields']._field_type = PLACEHOLDER_FIELD_TYPE_record_allow_extra_fields
 * __dataclass_fields__['record_encode_use_defaults'].name = 'record_encode_use_defaults'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_encode_use_defaults'].type = PLACEHOLDER_record_encode_use_defaults
 * __dataclass_fields__['record_encode_use_defaults']._field_type = PLACEHOLDER_FIELD_TYPE_record_encode_use_defaults
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_record_encode_use_defaults); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_record_encode_use_defaults) < 0) __PYX_ERR(1, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":23
 * __dataclass_fields__['record_allow_extra_fields']._field_type = PLACEHOLDER_FIELD_TYPE_record_allow_extra_fields
 * __dataclass_fields__['record_encode_use_defaults'].name = 'record_encode_use_defaults'
 * __dataclass_fields__['record_encode_use_defaults'].type = PLACEHOLDER_record_encode_use_defaults             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_encode_use_defaults']._field_type = PLACEHOLDER_FIELD_TYPE_record_encode_use_defaults
 * __dataclass_fields__['allow_tuple_notation'].name = 'allow_tuple_notation'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_record_encode_use_defaults); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":24
 * __dataclass_fields__['record_encode_use_defaults'].name = 'record_encode_use_defaults'
 * __dataclass_fields__['record_encode_use_defaults'].type = PLACEHOLDER_record_encode_use_defaults
 * __dataclass_fields__['record_encode_use_defaults']._field_type = PLACEHOLDER_FIELD_TYPE_record_encode_use_defaults             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_tuple_notation'].name = 'allow_tuple_notation'
 * __dataclass_fields__['allow_tuple_notation'].type = PLACEHOLDER_allow_tuple_notation
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_record_encode_use_defaults); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":25
 * __dataclass_fields__['record_encode_use_defaults'].type = PLACEHOLDER_record_encode_use_defaults
 * __dataclass_fields__['record_encode_use_defaults']._field_type = PLACEHOLDER_FIELD_TYPE_record_encode_use_defaults
 * __dataclass_fields__['allow_tuple_notation'].name = 'allow_tuple_notation'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_tuple_notation'].type = PLACEHOLDER_allow_tuple_notation
 * __dataclass_fields__['allow_tuple_notation']._field_type = PLACEHOLDER_FIELD_TYPE_allow_tuple_notation
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_allow_tuple_notation); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_allow_tuple_notation) < 0) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":26
 * __dataclass_fields__['record_encode_use_defaults']._field_type = PLACEHOLDER_FIELD_TYPE_record_encode_use_defaults
 * __dataclass_fields__['allow_tuple_notation'].name = 'allow_tuple_notation'
 * __dataclass_fields__['allow_tuple_notation'].type = PLACEHOLDER_allow_tuple_notation             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_tuple_notation']._field_type = PLACEHOLDER_FIELD_TYPE_allow_tuple_notation
 * __dataclass_fields__['union_decodes_to'].name = 'union_decodes_to'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_allow_tuple_notation); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":27
 * __dataclass_fields__['allow_tuple_notation'].name = 'allow_tuple_notation'
 * __dataclass_fields__['allow_tuple_notation'].type = PLACEHOLDER_allow_tuple_notation
 * __dataclass_fields__['allow_tuple_notation']._field_type = PLACEHOLDER_FIELD_TYPE_allow_tuple_notation             # <<<<<<<<<<<<<<
 * __dataclass_fields__['union_decodes_to'].name = 'union_decodes_to'
 * __dataclass_fields__['union_decodes_to'].type = PLACEHOLDER_union_decodes_to
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_allow_tuple_notation); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":28
 * __dataclass_fields__['allow_tuple_notation'].type = PLACEHOLDER_allow_tuple_notation
 * __dataclass_fields__['allow_tuple_notation']._field_type = PLACEHOLDER_FIELD_TYPE_allow_tuple_notation
 * __dataclass_fields__['union_decodes_to'].name = 'union_decodes_to'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['union_decodes_to'].type = PLACEHOLDER_union_decodes_to
 * __dataclass_fields__['union_decodes_to']._field_type = PLACEHOLDER_FIELD_TYPE_union_decodes_to
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_union_decodes_to); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_union_decodes_to) < 0) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":29
 * __dataclass_fields__['allow_tuple_notation']._field_type = PLACEHOLDER_FIELD_TYPE_allow_tuple_notation
 * __dataclass_fields__['union_decodes_to'].name = 'union_decodes_to'
 * __dataclass_fields__['union_decodes_to'].type = PLACEHOLDER_union_decodes_to             # <<<<<<<<<<<<<<
 * __dataclass_fields__['union_decodes_to']._field_type = PLACEHOLDER_FIELD_TYPE_union_decodes_to
 * __dataclass_fields__['union_json_encodes_type_name'].name = 'union_json_encodes_type_name'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_union_decodes_to); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_UnionDecodeOption) < 0) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":30
 * __dataclass_fields__['union_decodes_to'].name = 'union_decodes_to'
 * __dataclass_fields__['union_decodes_to'].type = PLACEHOLDER_union_decodes_to
 * __dataclass_fields__['union_decodes_to']._field_type = PLACEHOLDER_FIELD_TYPE_union_decodes_to             # <<<<<<<<<<<<<<
 * __dataclass_fields__['union_json_encodes_type_name'].name = 'union_json_encodes_type_name'
 * __dataclass_fields__['union_json_encodes_type_name'].type = PLACEHOLDER_union_json_encodes_type_name
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_union_decodes_to); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":31
 * __dataclass_fields__['union_decodes_to'].type = PLACEHOLDER_union_decodes_to
 * __dataclass_fields__['union_decodes_to']._field_type = PLACEHOLDER_FIELD_TYPE_union_decodes_to
 * __dataclass_fields__['union_json_encodes_type_name'].name = 'union_json_encodes_type_name'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['union_json_encodes_type_name'].type = PLACEHOLDER_union_json_encodes_type_name
 * __dataclass_fields__['union_json_encodes_type_name']._field_type = PLACEHOLDER_FIELD_TYPE_union_json_encodes_type_name
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_union_json_encodes_type_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_union_json_encodes_type_name) < 0) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":32
 * __dataclass_fields__['union_decodes_to']._field_type = PLACEHOLDER_FIELD_TYPE_union_decodes_to
 * __dataclass_fields__['union_json_encodes_type_name'].name = 'union_json_encodes_type_name'
 * __dataclass_fields__['union_json_encodes_type_name'].type = PLACEHOLDER_union_json_encodes_type_name             # <<<<<<<<<<<<<<
 * __dataclass_fields__['union_json_encodes_type_name']._field_type = PLACEHOLDER_FIELD_TYPE_union_json_encodes_type_name
 * __dataclass_fields__['allow_primitive_name_collision'].name = 'allow_primitive_name_collision'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_union_json_encodes_type_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":33
 * __dataclass_fields__['union_json_encodes_type_name'].name = 'union_json_encodes_type_name'
 * __dataclass_fields__['union_json_encodes_type_name'].type = PLACEHOLDER_union_json_encodes_type_name
 * __dataclass_fields__['union_json_encodes_type_name']._field_type = PLACEHOLDER_FIELD_TYPE_union_json_encodes_type_name             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_primitive_name_collision'].name = 'allow_primitive_name_collision'
 * __dataclass_fields__['allow_primitive_name_collision'].type = PLACEHOLDER_allow_primitive_name_collision
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_union_json_encodes_type_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":34
 * __dataclass_fields__['union_json_encodes_type_name'].type = PLACEHOLDER_union_json_encodes_type_name
 * __dataclass_fields__['union_json_encodes_type_name']._field_type = PLACEHOLDER_FIELD_TYPE_union_json_encodes_type_name
 * __dataclass_fields__['allow_primitive_name_collision'].name = 'allow_primitive_name_collision'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_primitive_name_collision'].type = PLACEHOLDER_allow_primitive_name_collision
 * __dataclass_fields__['allow_primitive_name_collision']._field_type = PLACEHOLDER_FIELD_TYPE_allow_primitive_name_collision
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_allow_primitive_name_collision); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_allow_primitive_name_collision) < 0) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":35
 * __dataclass_fields__['union_json_encodes_type_name']._field_type = PLACEHOLDER_FIELD_TYPE_union_json_encodes_type_name
 * __dataclass_fields__['allow_primitive_name_collision'].name = 'allow_primitive_name_collision'
 * __dataclass_fields__['allow_primitive_name_collision'].type = PLACEHOLDER_allow_primitive_name_collision             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_primitive_name_collision']._field_type = PLACEHOLDER_FIELD_TYPE_allow_primitive_name_collision
 * __dataclass_fields__['allow_primitive_names_in_namespaces'].name = 'allow_primitive_names_in_namespaces'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_allow_primitive_name_collision); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":36
 * __dataclass_fields__['allow_primitive_name_collision'].name = 'allow_primitive_name_collision'
 * __dataclass_fields__['allow_primitive_name_collision'].type = PLACEHOLDER_allow_primitive_name_collision
 * __dataclass_fields__['allow_primitive_name_collision']._field_type = PLACEHOLDER_FIELD_TYPE_allow_primitive_name_collision             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_primitive_names_in_namespaces'].name = 'allow_primitive_names_in_namespaces'
 * __dataclass_fields__['allow_primitive_names_in_namespaces'].type = PLACEHOLDER_allow_primitive_names_in_namespaces
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_allow_primitive_name_collision); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":37
 * __dataclass_fields__['allow_primitive_name_collision'].type = PLACEHOLDER_allow_primitive_name_collision
 * __dataclass_fields__['allow_primitive_name_collision']._field_type = PLACEHOLDER_FIELD_TYPE_allow_primitive_name_collision
 * __dataclass_fields__['allow_primitive_names_in_namespaces'].name = 'allow_primitive_names_in_namespaces'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_primitive_names_in_namespaces'].type = PLACEHOLDER_allow_primitive_names_in_namespaces
 * __dataclass_fields__['allow_primitive_names_in_namespaces']._field_type = PLACEHOLDER_FIELD_TYPE_allow_primitive_names_in_namespaces
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_allow_primitive_names_in_namespa); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_allow_primitive_names_in_namespa) < 0) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":38
 * __dataclass_fields__['allow_primitive_name_collision']._field_type = PLACEHOLDER_FIELD_TYPE_allow_primitive_name_collision
 * __dataclass_fields__['allow_primitive_names_in_namespaces'].name = 'allow_primitive_names_in_namespaces'
 * __dataclass_fields__['allow_primitive_names_in_namespaces'].type = PLACEHOLDER_allow_primitive_names_in_namespaces             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_primitive_names_in_namespaces']._field_type = PLACEHOLDER_FIELD_TYPE_allow_primitive_names_in_namespaces
 * __dataclass_fields__['named_type_names_must_be_unique'].name = 'named_type_names_must_be_unique'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_allow_primitive_names_in_namespa); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":39
 * __dataclass_fields__['allow_primitive_names_in_namespaces'].name = 'allow_primitive_names_in_namespaces'
 * __dataclass_fields__['allow_primitive_names_in_namespaces'].type = PLACEHOLDER_allow_primitive_names_in_namespaces
 * __dataclass_fields__['allow_primitive_names_in_namespaces']._field_type = PLACEHOLDER_FIELD_TYPE_allow_primitive_names_in_namespaces             # <<<<<<<<<<<<<<
 * __dataclass_fields__['named_type_names_must_be_unique'].name = 'named_type_names_must_be_unique'
 * __dataclass_fields__['named_type_names_must_be_unique'].type = PLACEHOLDER_named_type_names_must_be_unique
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_allow_primitive_names_in_namespa); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":40
 * __dataclass_fields__['allow_primitive_names_in_namespaces'].type = PLACEHOLDER_allow_primitive_names_in_namespaces
 * __dataclass_fields__['allow_primitive_names_in_namespaces']._field_type = PLACEHOLDER_FIELD_TYPE_allow_primitive_names_in_namespaces
 * __dataclass_fields__['named_type_names_must_be_unique'].name = 'named_type_names_must_be_unique'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['named_type_names_must_be_unique'].type = PLACEHOLDER_named_type_names_must_be_unique
 * __dataclass_fields__['named_type_names_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_named_type_names_must_be_unique
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_named_type_names_must_be_unique); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_named_type_names_must_be_unique) < 0) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":41
 * __dataclass_fields__['allow_primitive_names_in_namespaces']._field_type = PLACEHOLDER_FIELD_TYPE_allow_primitive_names_in_namespaces
 * __dataclass_fields__['named_type_names_must_be_unique'].name = 'named_type_names_must_be_unique'
 * __dataclass_fields__['named_type_names_must_be_unique'].type = PLACEHOLDER_named_type_names_must_be_unique             # <<<<<<<<<<<<<<
 * __dataclass_fields__['named_type_names_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_named_type_names_must_be_unique
 * __dataclass_fields__['enum_symbols_must_be_unique'].name = 'enum_symbols_must_be_unique'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_named_type_names_must_be_unique); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":42
 * __dataclass_fields__['named_type_names_must_be_unique'].name = 'named_type_names_must_be_unique'
 * __dataclass_fields__['named_type_names_must_be_unique'].type = PLACEHOLDER_named_type_names_must_be_unique
 * __dataclass_fields__['named_type_names_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_named_type_names_must_be_unique             # <<<<<<<<<<<<<<
 * __dataclass_fields__['enum_symbols_must_be_unique'].name = 'enum_symbols_must_be_unique'
 * __dataclass_fields__['enum_symbols_must_be_unique'].type = PLACEHOLDER_enum_symbols_must_be_unique
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_named_type_names_must_be_unique); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":43
 * __dataclass_fields__['named_type_names_must_be_unique'].type = PLACEHOLDER_named_type_names_must_be_unique
 * __dataclass_fields__['named_type_names_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_named_type_names_must_be_unique
 * __dataclass_fields__['enum_symbols_must_be_unique'].name = 'enum_symbols_must_be_unique'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['enum_symbols_must_be_unique'].type = PLACEHOLDER_enum_symbols_must_be_unique
 * __dataclass_fields__['enum_symbols_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_enum_symbols_must_be_unique
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_enum_symbols_must_be_unique); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_enum_symbols_must_be_unique) < 0) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":44
 * __dataclass_fields__['named_type_names_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_named_type_names_must_be_unique
 * __dataclass_fields__['enum_symbols_must_be_unique'].name = 'enum_symbols_must_be_unique'
 * __dataclass_fields__['enum_symbols_must_be_unique'].type = PLACEHOLDER_enum_symbols_must_be_unique             # <<<<<<<<<<<<<<
 * __dataclass_fields__['enum_symbols_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_enum_symbols_must_be_unique
 * __dataclass_fields__['enforce_enum_symbol_name_rules'].name = 'enforce_enum_symbol_name_rules'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_enum_symbols_must_be_unique); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":45
 * __dataclass_fields__['enum_symbols_must_be_unique'].name = 'enum_symbols_must_be_unique'
 * __dataclass_fields__['enum_symbols_must_be_unique'].type = PLACEHOLDER_enum_symbols_must_be_unique
 * __dataclass_fields__['enum_symbols_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_enum_symbols_must_be_unique             # <<<<<<<<<<<<<<
 * __dataclass_fields__['enforce_enum_symbol_name_rules'].name = 'enforce_enum_symbol_name_rules'
 * __dataclass_fields__['enforce_enum_symbol_name_rules'].type = PLACEHOLDER_enforce_enum_symbol_name_rules
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_enum_symbols_must_be_unique); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":46
 * __dataclass_fields__['enum_symbols_must_be_unique'].type = PLACEHOLDER_enum_symbols_must_be_unique
 * __dataclass_fields__['enum_symbols_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_enum_symbols_must_be_unique
 * __dataclass_fields__['enforce_enum_symbol_name_rules'].name = 'enforce_enum_symbol_name_rules'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['enforce_enum_symbol_name_rules'].type = PLACEHOLDER_enforce_enum_symbol_name_rules
 * __dataclass_fields__['enforce_enum_symbol_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_enum_symbol_name_rules
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_enforce_enum_symbol_name_rules); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_enforce_enum_symbol_name_rules) < 0) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":47
 * __dataclass_fields__['enum_symbols_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_enum_symbols_must_be_unique
 * __dataclass_fields__['enforce_enum_symbol_name_rules'].name = 'enforce_enum_symbol_name_rules'
 * __dataclass_fields__['enforce_enum_symbol_name_rules'].type = PLACEHOLDER_enforce_enum_symbol_name_rules             # <<<<<<<<<<<<<<
 * __dataclass_fields__['enforce_enum_symbol_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_enum_symbol_name_rules
 * __dataclass_fields__['enforce_type_name_rules'].name = 'enforce_type_name_rules'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_enforce_enum_symbol_name_rules); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":48
 * __dataclass_fields__['enforce_enum_symbol_name_rules'].name = 'enforce_enum_symbol_name_rules'
 * __dataclass_fields__['enforce_enum_symbol_name_rules'].type = PLACEHOLDER_enforce_enum_symbol_name_rules
 * __dataclass_fields__['enforce_enum_symbol_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_enum_symbol_name_rules             # <<<<<<<<<<<<<<
 * __dataclass_fields__['enforce_type_name_rules'].name = 'enforce_type_name_rules'
 * __dataclass_fields__['enforce_type_name_rules'].type = PLACEHOLDER_enforce_type_name_rules
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_enforce_enum_symbol_name_rules); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":49
 * __dataclass_fields__['enforce_enum_symbol_name_rules'].type = PLACEHOLDER_enforce_enum_symbol_name_rules
 * __dataclass_fields__['enforce_enum_symbol_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_enum_symbol_name_rules
 * __dataclass_fields__['enforce_type_name_rules'].name = 'enforce_type_name_rules'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['enforce_type_name_rules'].type = PLACEHOLDER_enforce_type_name_rules
 * __dataclass_fields__['enforce_type_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_type_name_rules
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_enforce_type_name_rules); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_enforce_type_name_rules) < 0) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":50
 * __dataclass_fields__['enforce_enum_symbol_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_enum_symbol_name_rules
 * __dataclass_fields__['enforce_type_name_rules'].name = 'enforce_type_name_rules'
 * __dataclass_fields__['enforce_type_name_rules'].type = PLACEHOLDER_enforce_type_name_rules             # <<<<<<<<<<<<<<
 * __dataclass_fields__['enforce_type_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_type_name_rules
 * __dataclass_fields__['enforce_namespace_name_rules'].name = 'enforce_namespace_name_rules'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_enforce_type_name_rules); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":51
 * __dataclass_fields__['enforce_type_name_rules'].name = 'enforce_type_name_rules'
 * __dataclass_fields__['enforce_type_name_rules'].type = PLACEHOLDER_enforce_type_name_rules
 * __dataclass_fields__['enforce_type_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_type_name_rules             # <<<<<<<<<<<<<<
 * __dataclass_fields__['enforce_namespace_name_rules'].name = 'enforce_namespace_name_rules'
 * __dataclass_fields__['enforce_namespace_name_rules'].type = PLACEHOLDER_enforce_namespace_name_rules
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_enforce_type_name_rules); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":52
 * __dataclass_fields__['enforce_type_name_rules'].type = PLACEHOLDER_enforce_type_name_rules
 * __dataclass_fields__['enforce_type_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_type_name_rules
 * __dataclass_fields__['enforce_namespace_name_rules'].name = 'enforce_namespace_name_rules'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['enforce_namespace_name_rules'].type = PLACEHOLDER_enforce_namespace_name_rules
 * __dataclass_fields__['enforce_namespace_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_namespace_name_rules
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_enforce_namespace_name_rules); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_enforce_namespace_name_rules) < 0) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":53
 * __dataclass_fields__['enforce_type_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_type_name_rules
 * __dataclass_fields__['enforce_namespace_name_rules'].name = 'enforce_namespace_name_rules'
 * __dataclass_fields__['enforce_namespace_name_rules'].type = PLACEHOLDER_enforce_namespace_name_rules             # <<<<<<<<<<<<<<
 * __dataclass_fields__['enforce_namespace_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_namespace_name_rules
 * __dataclass_fields__['record_fields_must_be_unique'].name = 'record_fields_must_be_unique'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_enforce_namespace_name_rules); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":54
 * __dataclass_fields__['enforce_namespace_name_rules'].name = 'enforce_namespace_name_rules'
 * __dataclass_fields__['enforce_namespace_name_rules'].type = PLACEHOLDER_enforce_namespace_name_rules
 * __dataclass_fields__['enforce_namespace_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_namespace_name_rules             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_fields_must_be_unique'].name = 'record_fields_must_be_unique'
 * __dataclass_fields__['record_fields_must_be_unique'].type = PLACEHOLDER_record_fields_must_be_unique
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_enforce_namespace_name_rules); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":55
 * __dataclass_fields__['enforce_namespace_name_rules'].type = PLACEHOLDER_enforce_namespace_name_rules
 * __dataclass_fields__['enforce_namespace_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_namespace_name_rules
 * __dataclass_fields__['record_fields_must_be_unique'].name = 'record_fields_must_be_unique'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_fields_must_be_unique'].type = PLACEHOLDER_record_fields_must_be_unique
 * __dataclass_fields__['record_fields_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_record_fields_must_be_unique
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_record_fields_must_be_unique); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_record_fields_must_be_unique) < 0) __PYX_ERR(1, 55, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":56
 * __dataclass_fields__['enforce_namespace_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_enforce_namespace_name_rules
 * __dataclass_fields__['record_fields_must_be_unique'].name = 'record_fields_must_be_unique'
 * __dataclass_fields__['record_fields_must_be_unique'].type = PLACEHOLDER_record_fields_must_be_unique             # <<<<<<<<<<<<<<
 * __dataclass_fields__['record_fields_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_record_fields_must_be_unique
 * __dataclass_fields__['ascii_name_rules'].name = 'ascii_name_rules'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_record_fields_must_be_unique); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":57
 * __dataclass_fields__['record_fields_must_be_unique'].name = 'record_fields_must_be_unique'
 * __dataclass_fields__['record_fields_must_be_unique'].type = PLACEHOLDER_record_fields_must_be_unique
 * __dataclass_fields__['record_fields_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_record_fields_must_be_unique             # <<<<<<<<<<<<<<
 * __dataclass_fields__['ascii_name_rules'].name = 'ascii_name_rules'
 * __dataclass_fields__['ascii_name_rules'].type = PLACEHOLDER_ascii_name_rules
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_record_fields_must_be_unique); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":58
 * __dataclass_fields__['record_fields_must_be_unique'].type = PLACEHOLDER_record_fields_must_be_unique
 * __dataclass_fields__['record_fields_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_record_fields_must_be_unique
 * __dataclass_fields__['ascii_name_rules'].name = 'ascii_name_rules'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['ascii_name_rules'].type = PLACEHOLDER_ascii_name_rules
 * __dataclass_fields__['ascii_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_ascii_name_rules
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_ascii_name_rules); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_ascii_name_rules) < 0) __PYX_ERR(1, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":59
 * __dataclass_fields__['record_fields_must_be_unique']._field_type = PLACEHOLDER_FIELD_TYPE_record_fields_must_be_unique
 * __dataclass_fields__['ascii_name_rules'].name = 'ascii_name_rules'
 * __dataclass_fields__['ascii_name_rules'].type = PLACEHOLDER_ascii_name_rules             # <<<<<<<<<<<<<<
 * __dataclass_fields__['ascii_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_ascii_name_rules
 * __dataclass_fields__['missing_values_can_be_null'].name = 'missing_values_can_be_null'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_ascii_name_rules); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 59, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":60
 * __dataclass_fields__['ascii_name_rules'].name = 'ascii_name_rules'
 * __dataclass_fields__['ascii_name_rules'].type = PLACEHOLDER_ascii_name_rules
 * __dataclass_fields__['ascii_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_ascii_name_rules             # <<<<<<<<<<<<<<
 * __dataclass_fields__['missing_values_can_be_null'].name = 'missing_values_can_be_null'
 * __dataclass_fields__['missing_values_can_be_null'].type = PLACEHOLDER_missing_values_can_be_null
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_ascii_name_rules); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 60, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":61
 * __dataclass_fields__['ascii_name_rules'].type = PLACEHOLDER_ascii_name_rules
 * __dataclass_fields__['ascii_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_ascii_name_rules
 * __dataclass_fields__['missing_values_can_be_null'].name = 'missing_values_can_be_null'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['missing_values_can_be_null'].type = PLACEHOLDER_missing_values_can_be_null
 * __dataclass_fields__['missing_values_can_be_null']._field_type = PLACEHOLDER_FIELD_TYPE_missing_values_can_be_null
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_missing_values_can_be_null); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_missing_values_can_be_null) < 0) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":62
 * __dataclass_fields__['ascii_name_rules']._field_type = PLACEHOLDER_FIELD_TYPE_ascii_name_rules
 * __dataclass_fields__['missing_values_can_be_null'].name = 'missing_values_can_be_null'
 * __dataclass_fields__['missing_values_can_be_null'].type = PLACEHOLDER_missing_values_can_be_null             # <<<<<<<<<<<<<<
 * __dataclass_fields__['missing_values_can_be_null']._field_type = PLACEHOLDER_FIELD_TYPE_missing_values_can_be_null
 * __dataclass_fields__['missing_values_can_be_empty_container'].name = 'missing_values_can_be_empty_container'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_missing_values_can_be_null); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":63
 * __dataclass_fields__['missing_values_can_be_null'].name = 'missing_values_can_be_null'
 * __dataclass_fields__['missing_values_can_be_null'].type = PLACEHOLDER_missing_values_can_be_null
 * __dataclass_fields__['missing_values_can_be_null']._field_type = PLACEHOLDER_FIELD_TYPE_missing_values_can_be_null             # <<<<<<<<<<<<<<
 * __dataclass_fields__['missing_values_can_be_empty_container'].name = 'missing_values_can_be_empty_container'
 * __dataclass_fields__['missing_values_can_be_empty_container'].type = PLACEHOLDER_missing_values_can_be_empty_container
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_missing_values_can_be_null); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":64
 * __dataclass_fields__['missing_values_can_be_null'].type = PLACEHOLDER_missing_values_can_be_null
 * __dataclass_fields__['missing_values_can_be_null']._field_type = PLACEHOLDER_FIELD_TYPE_missing_values_can_be_null
 * __dataclass_fields__['missing_values_can_be_empty_container'].name = 'missing_values_can_be_empty_container'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['missing_values_can_be_empty_container'].type = PLACEHOLDER_missing_values_can_be_empty_container
 * __dataclass_fields__['missing_values_can_be_empty_container']._field_type = PLACEHOLDER_FIELD_TYPE_missing_values_can_be_empty_container
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_missing_values_can_be_empty_cont); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_missing_values_can_be_empty_cont) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":65
 * __dataclass_fields__['missing_values_can_be_null']._field_type = PLACEHOLDER_FIELD_TYPE_missing_values_can_be_null
 * __dataclass_fields__['missing_values_can_be_empty_container'].name = 'missing_values_can_be_empty_container'
 * __dataclass_fields__['missing_values_can_be_empty_container'].type = PLACEHOLDER_missing_values_can_be_empty_container             # <<<<<<<<<<<<<<
 * __dataclass_fields__['missing_values_can_be_empty_container']._field_type = PLACEHOLDER_FIELD_TYPE_missing_values_can_be_empty_container
 * __dataclass_fields__['allow_false_values_for_null'].name = 'allow_false_values_for_null'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_missing_values_can_be_empty_cont); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":66
 * __dataclass_fields__['missing_values_can_be_empty_container'].name = 'missing_values_can_be_empty_container'
 * __dataclass_fields__['missing_values_can_be_empty_container'].type = PLACEHOLDER_missing_values_can_be_empty_container
 * __dataclass_fields__['missing_values_can_be_empty_container']._field_type = PLACEHOLDER_FIELD_TYPE_missing_values_can_be_empty_container             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_false_values_for_null'].name = 'allow_false_values_for_null'
 * __dataclass_fields__['allow_false_values_for_null'].type = PLACEHOLDER_allow_false_values_for_null
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_missing_values_can_be_empty_cont); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":67
 * __dataclass_fields__['missing_values_can_be_empty_container'].type = PLACEHOLDER_missing_values_can_be_empty_container
 * __dataclass_fields__['missing_values_can_be_empty_container']._field_type = PLACEHOLDER_FIELD_TYPE_missing_values_can_be_empty_container
 * __dataclass_fields__['allow_false_values_for_null'].name = 'allow_false_values_for_null'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_false_values_for_null'].type = PLACEHOLDER_allow_false_values_for_null
 * __dataclass_fields__['allow_false_values_for_null']._field_type = PLACEHOLDER_FIELD_TYPE_allow_false_values_for_null
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_allow_false_values_for_null); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_allow_false_values_for_null) < 0) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":68
 * __dataclass_fields__['missing_values_can_be_empty_container']._field_type = PLACEHOLDER_FIELD_TYPE_missing_values_can_be_empty_container
 * __dataclass_fields__['allow_false_values_for_null'].name = 'allow_false_values_for_null'
 * __dataclass_fields__['allow_false_values_for_null'].type = PLACEHOLDER_allow_false_values_for_null             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_false_values_for_null']._field_type = PLACEHOLDER_FIELD_TYPE_allow_false_values_for_null
 * __dataclass_fields__['allow_empty_unions'].name = 'allow_empty_unions'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_allow_false_values_for_null); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 68, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":69
 * __dataclass_fields__['allow_false_values_for_null'].name = 'allow_false_values_for_null'
 * __dataclass_fields__['allow_false_values_for_null'].type = PLACEHOLDER_allow_false_values_for_null
 * __dataclass_fields__['allow_false_values_for_null']._field_type = PLACEHOLDER_FIELD_TYPE_allow_false_values_for_null             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_empty_unions'].name = 'allow_empty_unions'
 * __dataclass_fields__['allow_empty_unions'].type = PLACEHOLDER_allow_empty_unions
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_allow_false_values_for_null); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":70
 * __dataclass_fields__['allow_false_values_for_null'].type = PLACEHOLDER_allow_false_values_for_null
 * __dataclass_fields__['allow_false_values_for_null']._field_type = PLACEHOLDER_FIELD_TYPE_allow_false_values_for_null
 * __dataclass_fields__['allow_empty_unions'].name = 'allow_empty_unions'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_empty_unions'].type = PLACEHOLDER_allow_empty_unions
 * __dataclass_fields__['allow_empty_unions']._field_type = PLACEHOLDER_FIELD_TYPE_allow_empty_unions
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_allow_empty_unions); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_allow_empty_unions) < 0) __PYX_ERR(1, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":71
 * __dataclass_fields__['allow_false_values_for_null']._field_type = PLACEHOLDER_FIELD_TYPE_allow_false_values_for_null
 * __dataclass_fields__['allow_empty_unions'].name = 'allow_empty_unions'
 * __dataclass_fields__['allow_empty_unions'].type = PLACEHOLDER_allow_empty_unions             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_empty_unions']._field_type = PLACEHOLDER_FIELD_TYPE_allow_empty_unions
 * __dataclass_fields__['allow_nested_unions'].name = 'allow_nested_unions'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_allow_empty_unions); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":72
 * __dataclass_fields__['allow_empty_unions'].name = 'allow_empty_unions'
 * __dataclass_fields__['allow_empty_unions'].type = PLACEHOLDER_allow_empty_unions
 * __dataclass_fields__['allow_empty_unions']._field_type = PLACEHOLDER_FIELD_TYPE_allow_empty_unions             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_nested_unions'].name = 'allow_nested_unions'
 * __dataclass_fields__['allow_nested_unions'].type = PLACEHOLDER_allow_nested_unions
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_allow_empty_unions); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":73
 * __dataclass_fields__['allow_empty_unions'].type = PLACEHOLDER_allow_empty_unions
 * __dataclass_fields__['allow_empty_unions']._field_type = PLACEHOLDER_FIELD_TYPE_allow_empty_unions
 * __dataclass_fields__['allow_nested_unions'].name = 'allow_nested_unions'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_nested_unions'].type = PLACEHOLDER_allow_nested_unions
 * __dataclass_fields__['allow_nested_unions']._field_type = PLACEHOLDER_FIELD_TYPE_allow_nested_unions
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_allow_nested_unions); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_allow_nested_unions) < 0) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":74
 * __dataclass_fields__['allow_empty_unions']._field_type = PLACEHOLDER_FIELD_TYPE_allow_empty_unions
 * __dataclass_fields__['allow_nested_unions'].name = 'allow_nested_unions'
 * __dataclass_fields__['allow_nested_unions'].type = PLACEHOLDER_allow_nested_unions             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_nested_unions']._field_type = PLACEHOLDER_FIELD_TYPE_allow_nested_unions
 * __dataclass_fields__['allow_duplicate_union_types'].name = 'allow_duplicate_union_types'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_allow_nested_unions); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 74, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":75
 * __dataclass_fields__['allow_nested_unions'].name = 'allow_nested_unions'
 * __dataclass_fields__['allow_nested_unions'].type = PLACEHOLDER_allow_nested_unions
 * __dataclass_fields__['allow_nested_unions']._field_type = PLACEHOLDER_FIELD_TYPE_allow_nested_unions             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_duplicate_union_types'].name = 'allow_duplicate_union_types'
 * __dataclass_fields__['allow_duplicate_union_types'].type = PLACEHOLDER_allow_duplicate_union_types
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_allow_nested_unions); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":76
 * __dataclass_fields__['allow_nested_unions'].type = PLACEHOLDER_allow_nested_unions
 * __dataclass_fields__['allow_nested_unions']._field_type = PLACEHOLDER_FIELD_TYPE_allow_nested_unions
 * __dataclass_fields__['allow_duplicate_union_types'].name = 'allow_duplicate_union_types'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_duplicate_union_types'].type = PLACEHOLDER_allow_duplicate_union_types
 * __dataclass_fields__['allow_duplicate_union_types']._field_type = PLACEHOLDER_FIELD_TYPE_allow_duplicate_union_types
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_allow_duplicate_union_types); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_allow_duplicate_union_types) < 0) __PYX_ERR(1, 76, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":77
 * __dataclass_fields__['allow_nested_unions']._field_type = PLACEHOLDER_FIELD_TYPE_allow_nested_unions
 * __dataclass_fields__['allow_duplicate_union_types'].name = 'allow_duplicate_union_types'
 * __dataclass_fields__['allow_duplicate_union_types'].type = PLACEHOLDER_allow_duplicate_union_types             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_duplicate_union_types']._field_type = PLACEHOLDER_FIELD_TYPE_allow_duplicate_union_types
 * __dataclass_fields__['allow_union_default_any_member'].name = 'allow_union_default_any_member'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_allow_duplicate_union_types); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":78
 * __dataclass_fields__['allow_duplicate_union_types'].name = 'allow_duplicate_union_types'
 * __dataclass_fields__['allow_duplicate_union_types'].type = PLACEHOLDER_allow_duplicate_union_types
 * __dataclass_fields__['allow_duplicate_union_types']._field_type = PLACEHOLDER_FIELD_TYPE_allow_duplicate_union_types             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_union_default_any_member'].name = 'allow_union_default_any_member'
 * __dataclass_fields__['allow_union_default_any_member'].type = PLACEHOLDER_allow_union_default_any_member
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_allow_duplicate_union_types); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":79
 * __dataclass_fields__['allow_duplicate_union_types'].type = PLACEHOLDER_allow_duplicate_union_types
 * __dataclass_fields__['allow_duplicate_union_types']._field_type = PLACEHOLDER_FIELD_TYPE_allow_duplicate_union_types
 * __dataclass_fields__['allow_union_default_any_member'].name = 'allow_union_default_any_member'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_union_default_any_member'].type = PLACEHOLDER_allow_union_default_any_member
 * __dataclass_fields__['allow_union_default_any_member']._field_type = PLACEHOLDER_FIELD_TYPE_allow_union_default_any_member
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_allow_union_default_any_member); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_allow_union_default_any_member) < 0) __PYX_ERR(1, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":80
 * __dataclass_fields__['allow_duplicate_union_types']._field_type = PLACEHOLDER_FIELD_TYPE_allow_duplicate_union_types
 * __dataclass_fields__['allow_union_default_any_member'].name = 'allow_union_default_any_member'
 * __dataclass_fields__['allow_union_default_any_member'].type = PLACEHOLDER_allow_union_default_any_member             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_union_default_any_member']._field_type = PLACEHOLDER_FIELD_TYPE_allow_union_default_any_member
 * __dataclass_fields__['allow_aliases_to_be_string'].name = 'allow_aliases_to_be_string'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_allow_union_default_any_member); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 80, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":81
 * __dataclass_fields__['allow_union_default_any_member'].name = 'allow_union_default_any_member'
 * __dataclass_fields__['allow_union_default_any_member'].type = PLACEHOLDER_allow_union_default_any_member
 * __dataclass_fields__['allow_union_default_any_member']._field_type = PLACEHOLDER_FIELD_TYPE_allow_union_default_any_member             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_aliases_to_be_string'].name = 'allow_aliases_to_be_string'
 * __dataclass_fields__['allow_aliases_to_be_string'].type = PLACEHOLDER_allow_aliases_to_be_string
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_allow_union_default_any_member); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":82
 * __dataclass_fields__['allow_union_default_any_member'].type = PLACEHOLDER_allow_union_default_any_member
 * __dataclass_fields__['allow_union_default_any_member']._field_type = PLACEHOLDER_FIELD_TYPE_allow_union_default_any_member
 * __dataclass_fields__['allow_aliases_to_be_string'].name = 'allow_aliases_to_be_string'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_aliases_to_be_string'].type = PLACEHOLDER_allow_aliases_to_be_string
 * __dataclass_fields__['allow_aliases_to_be_string']._field_type = PLACEHOLDER_FIELD_TYPE_allow_aliases_to_be_string
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_allow_aliases_to_be_string); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_allow_aliases_to_be_string) < 0) __PYX_ERR(1, 82, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":83
 * __dataclass_fields__['allow_union_default_any_member']._field_type = PLACEHOLDER_FIELD_TYPE_allow_union_default_any_member
 * __dataclass_fields__['allow_aliases_to_be_string'].name = 'allow_aliases_to_be_string'
 * __dataclass_fields__['allow_aliases_to_be_string'].type = PLACEHOLDER_allow_aliases_to_be_string             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_aliases_to_be_string']._field_type = PLACEHOLDER_FIELD_TYPE_allow_aliases_to_be_string
 * __dataclass_fields__['coerce_values_to_int'].name = 'coerce_values_to_int'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_allow_aliases_to_be_string); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":84
 * __dataclass_fields__['allow_aliases_to_be_string'].name = 'allow_aliases_to_be_string'
 * __dataclass_fields__['allow_aliases_to_be_string'].type = PLACEHOLDER_allow_aliases_to_be_string
 * __dataclass_fields__['allow_aliases_to_be_string']._field_type = PLACEHOLDER_FIELD_TYPE_allow_aliases_to_be_string             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_values_to_int'].name = 'coerce_values_to_int'
 * __dataclass_fields__['coerce_values_to_int'].type = PLACEHOLDER_coerce_values_to_int
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_allow_aliases_to_be_string); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":85
 * __dataclass_fields__['allow_aliases_to_be_string'].type = PLACEHOLDER_allow_aliases_to_be_string
 * __dataclass_fields__['allow_aliases_to_be_string']._field_type = PLACEHOLDER_FIELD_TYPE_allow_aliases_to_be_string
 * __dataclass_fields__['coerce_values_to_int'].name = 'coerce_values_to_int'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_values_to_int'].type = PLACEHOLDER_coerce_values_to_int
 * __dataclass_fields__['coerce_values_to_int']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_int
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_coerce_values_to_int); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_coerce_values_to_int) < 0) __PYX_ERR(1, 85, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":86
 * __dataclass_fields__['allow_aliases_to_be_string']._field_type = PLACEHOLDER_FIELD_TYPE_allow_aliases_to_be_string
 * __dataclass_fields__['coerce_values_to_int'].name = 'coerce_values_to_int'
 * __dataclass_fields__['coerce_values_to_int'].type = PLACEHOLDER_coerce_values_to_int             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_values_to_int']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_int
 * __dataclass_fields__['coerce_values_to_float'].name = 'coerce_values_to_float'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_coerce_values_to_int); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":87
 * __dataclass_fields__['coerce_values_to_int'].name = 'coerce_values_to_int'
 * __dataclass_fields__['coerce_values_to_int'].type = PLACEHOLDER_coerce_values_to_int
 * __dataclass_fields__['coerce_values_to_int']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_int             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_values_to_float'].name = 'coerce_values_to_float'
 * __dataclass_fields__['coerce_values_to_float'].type = PLACEHOLDER_coerce_values_to_float
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_coerce_values_to_int); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 87, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":88
 * __dataclass_fields__['coerce_values_to_int'].type = PLACEHOLDER_coerce_values_to_int
 * __dataclass_fields__['coerce_values_to_int']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_int
 * __dataclass_fields__['coerce_values_to_float'].name = 'coerce_values_to_float'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_values_to_float'].type = PLACEHOLDER_coerce_values_to_float
 * __dataclass_fields__['coerce_values_to_float']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_float
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_coerce_values_to_float); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_coerce_values_to_float) < 0) __PYX_ERR(1, 88, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":89
 * __dataclass_fields__['coerce_values_to_int']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_int
 * __dataclass_fields__['coerce_values_to_float'].name = 'coerce_values_to_float'
 * __dataclass_fields__['coerce_values_to_float'].type = PLACEHOLDER_coerce_values_to_float             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_values_to_float']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_float
 * __dataclass_fields__['coerce_int_to_float'].name = 'coerce_int_to_float'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_coerce_values_to_float); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":90
 * __dataclass_fields__['coerce_values_to_float'].name = 'coerce_values_to_float'
 * __dataclass_fields__['coerce_values_to_float'].type = PLACEHOLDER_coerce_values_to_float
 * __dataclass_fields__['coerce_values_to_float']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_float             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_int_to_float'].name = 'coerce_int_to_float'
 * __dataclass_fields__['coerce_int_to_float'].type = PLACEHOLDER_coerce_int_to_float
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_coerce_values_to_float); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":91
 * __dataclass_fields__['coerce_values_to_float'].type = PLACEHOLDER_coerce_values_to_float
 * __dataclass_fields__['coerce_values_to_float']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_float
 * __dataclass_fields__['coerce_int_to_float'].name = 'coerce_int_to_float'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_int_to_float'].type = PLACEHOLDER_coerce_int_to_float
 * __dataclass_fields__['coerce_int_to_float']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_int_to_float
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_coerce_int_to_float); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_coerce_int_to_float) < 0) __PYX_ERR(1, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":92
 * __dataclass_fields__['coerce_values_to_float']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_float
 * __dataclass_fields__['coerce_int_to_float'].name = 'coerce_int_to_float'
 * __dataclass_fields__['coerce_int_to_float'].type = PLACEHOLDER_coerce_int_to_float             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_int_to_float']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_int_to_float
 * __dataclass_fields__['coerce_values_to_boolean'].name = 'coerce_values_to_boolean'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_coerce_int_to_float); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":93
 * __dataclass_fields__['coerce_int_to_float'].name = 'coerce_int_to_float'
 * __dataclass_fields__['coerce_int_to_float'].type = PLACEHOLDER_coerce_int_to_float
 * __dataclass_fields__['coerce_int_to_float']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_int_to_float             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_values_to_boolean'].name = 'coerce_values_to_boolean'
 * __dataclass_fields__['coerce_values_to_boolean'].type = PLACEHOLDER_coerce_values_to_boolean
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_coerce_int_to_float); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 93, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":94
 * __dataclass_fields__['coerce_int_to_float'].type = PLACEHOLDER_coerce_int_to_float
 * __dataclass_fields__['coerce_int_to_float']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_int_to_float
 * __dataclass_fields__['coerce_values_to_boolean'].name = 'coerce_values_to_boolean'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_values_to_boolean'].type = PLACEHOLDER_coerce_values_to_boolean
 * __dataclass_fields__['coerce_values_to_boolean']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_boolean
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_coerce_values_to_boolean); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_coerce_values_to_boolean) < 0) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":95
 * __dataclass_fields__['coerce_int_to_float']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_int_to_float
 * __dataclass_fields__['coerce_values_to_boolean'].name = 'coerce_values_to_boolean'
 * __dataclass_fields__['coerce_values_to_boolean'].type = PLACEHOLDER_coerce_values_to_boolean             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_values_to_boolean']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_boolean
 * __dataclass_fields__['coerce_values_to_str'].name = 'coerce_values_to_str'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_coerce_values_to_boolean); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":96
 * __dataclass_fields__['coerce_values_to_boolean'].name = 'coerce_values_to_boolean'
 * __dataclass_fields__['coerce_values_to_boolean'].type = PLACEHOLDER_coerce_values_to_boolean
 * __dataclass_fields__['coerce_values_to_boolean']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_boolean             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_values_to_str'].name = 'coerce_values_to_str'
 * __dataclass_fields__['coerce_values_to_str'].type = PLACEHOLDER_coerce_values_to_str
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_coerce_values_to_boolean); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 96, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":97
 * __dataclass_fields__['coerce_values_to_boolean'].type = PLACEHOLDER_coerce_values_to_boolean
 * __dataclass_fields__['coerce_values_to_boolean']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_boolean
 * __dataclass_fields__['coerce_values_to_str'].name = 'coerce_values_to_str'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_values_to_str'].type = PLACEHOLDER_coerce_values_to_str
 * __dataclass_fields__['coerce_values_to_str']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_str
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_coerce_values_to_str); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_coerce_values_to_str) < 0) __PYX_ERR(1, 97, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":98
 * __dataclass_fields__['coerce_values_to_boolean']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_boolean
 * __dataclass_fields__['coerce_values_to_str'].name = 'coerce_values_to_str'
 * __dataclass_fields__['coerce_values_to_str'].type = PLACEHOLDER_coerce_values_to_str             # <<<<<<<<<<<<<<
 * __dataclass_fields__['coerce_values_to_str']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_str
 * __dataclass_fields__['bytes_codec'].name = 'bytes_codec'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_coerce_values_to_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 98, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":99
 * __dataclass_fields__['coerce_values_to_str'].name = 'coerce_values_to_str'
 * __dataclass_fields__['coerce_values_to_str'].type = PLACEHOLDER_coerce_values_to_str
 * __dataclass_fields__['coerce_values_to_str']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_str             # <<<<<<<<<<<<<<
 * __dataclass_fields__['bytes_codec'].name = 'bytes_codec'
 * __dataclass_fields__['bytes_codec'].type = PLACEHOLDER_bytes_codec
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_coerce_values_to_str); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":100
 * __dataclass_fields__['coerce_values_to_str'].type = PLACEHOLDER_coerce_values_to_str
 * __dataclass_fields__['coerce_values_to_str']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_str
 * __dataclass_fields__['bytes_codec'].name = 'bytes_codec'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['bytes_codec'].type = PLACEHOLDER_bytes_codec
 * __dataclass_fields__['bytes_codec']._field_type = PLACEHOLDER_FIELD_TYPE_bytes_codec
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_bytes_codec); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_bytes_codec) < 0) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":101
 * __dataclass_fields__['coerce_values_to_str']._field_type = PLACEHOLDER_FIELD_TYPE_coerce_values_to_str
 * __dataclass_fields__['bytes_codec'].name = 'bytes_codec'
 * __dataclass_fields__['bytes_codec'].type = PLACEHOLDER_bytes_codec             # <<<<<<<<<<<<<<
 * __dataclass_fields__['bytes_codec']._field_type = PLACEHOLDER_FIELD_TYPE_bytes_codec
 * __dataclass_fields__['fixed_codec'].name = 'fixed_codec'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_bytes_codec); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_str_2) < 0) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":102
 * __dataclass_fields__['bytes_codec'].name = 'bytes_codec'
 * __dataclass_fields__['bytes_codec'].type = PLACEHOLDER_bytes_codec
 * __dataclass_fields__['bytes_codec']._field_type = PLACEHOLDER_FIELD_TYPE_bytes_codec             # <<<<<<<<<<<<<<
 * __dataclass_fields__['fixed_codec'].name = 'fixed_codec'
 * __dataclass_fields__['fixed_codec'].type = PLACEHOLDER_fixed_codec
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_bytes_codec); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":103
 * __dataclass_fields__['bytes_codec'].type = PLACEHOLDER_bytes_codec
 * __dataclass_fields__['bytes_codec']._field_type = PLACEHOLDER_FIELD_TYPE_bytes_codec
 * __dataclass_fields__['fixed_codec'].name = 'fixed_codec'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['fixed_codec'].type = PLACEHOLDER_fixed_codec
 * __dataclass_fields__['fixed_codec']._field_type = PLACEHOLDER_FIELD_TYPE_fixed_codec
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_fixed_codec); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_fixed_codec) < 0) __PYX_ERR(1, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":104
 * __dataclass_fields__['bytes_codec']._field_type = PLACEHOLDER_FIELD_TYPE_bytes_codec
 * __dataclass_fields__['fixed_codec'].name = 'fixed_codec'
 * __dataclass_fields__['fixed_codec'].type = PLACEHOLDER_fixed_codec             # <<<<<<<<<<<<<<
 * __dataclass_fields__['fixed_codec']._field_type = PLACEHOLDER_FIELD_TYPE_fixed_codec
 * __dataclass_fields__['null_pad_fixed'].name = 'null_pad_fixed'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_fixed_codec); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_str_2) < 0) __PYX_ERR(1, 104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":105
 * __dataclass_fields__['fixed_codec'].name = 'fixed_codec'
 * __dataclass_fields__['fixed_codec'].type = PLACEHOLDER_fixed_codec
 * __dataclass_fields__['fixed_codec']._field_type = PLACEHOLDER_FIELD_TYPE_fixed_codec             # <<<<<<<<<<<<<<
 * __dataclass_fields__['null_pad_fixed'].name = 'null_pad_fixed'
 * __dataclass_fields__['null_pad_fixed'].type = PLACEHOLDER_null_pad_fixed
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_fixed_codec); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":106
 * __dataclass_fields__['fixed_codec'].type = PLACEHOLDER_fixed_codec
 * __dataclass_fields__['fixed_codec']._field_type = PLACEHOLDER_FIELD_TYPE_fixed_codec
 * __dataclass_fields__['null_pad_fixed'].name = 'null_pad_fixed'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['null_pad_fixed'].type = PLACEHOLDER_null_pad_fixed
 * __dataclass_fields__['null_pad_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_null_pad_fixed
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_null_pad_fixed); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_null_pad_fixed) < 0) __PYX_ERR(1, 106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":107
 * __dataclass_fields__['fixed_codec']._field_type = PLACEHOLDER_FIELD_TYPE_fixed_codec
 * __dataclass_fields__['null_pad_fixed'].name = 'null_pad_fixed'
 * __dataclass_fields__['null_pad_fixed'].type = PLACEHOLDER_null_pad_fixed             # <<<<<<<<<<<<<<
 * __dataclass_fields__['null_pad_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_null_pad_fixed
 * __dataclass_fields__['truncate_fixed'].name = 'truncate_fixed'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_null_pad_fixed); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":108
 * __dataclass_fields__['null_pad_fixed'].name = 'null_pad_fixed'
 * __dataclass_fields__['null_pad_fixed'].type = PLACEHOLDER_null_pad_fixed
 * __dataclass_fields__['null_pad_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_null_pad_fixed             # <<<<<<<<<<<<<<
 * __dataclass_fields__['truncate_fixed'].name = 'truncate_fixed'
 * __dataclass_fields__['truncate_fixed'].type = PLACEHOLDER_truncate_fixed
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_null_pad_fixed); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":109
 * __dataclass_fields__['null_pad_fixed'].type = PLACEHOLDER_null_pad_fixed
 * __dataclass_fields__['null_pad_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_null_pad_fixed
 * __dataclass_fields__['truncate_fixed'].name = 'truncate_fixed'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['truncate_fixed'].type = PLACEHOLDER_truncate_fixed
 * __dataclass_fields__['truncate_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_truncate_fixed
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_truncate_fixed); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_truncate_fixed) < 0) __PYX_ERR(1, 109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":110
 * __dataclass_fields__['null_pad_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_null_pad_fixed
 * __dataclass_fields__['truncate_fixed'].name = 'truncate_fixed'
 * __dataclass_fields__['truncate_fixed'].type = PLACEHOLDER_truncate_fixed             # <<<<<<<<<<<<<<
 * __dataclass_fields__['truncate_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_truncate_fixed
 * __dataclass_fields__['clamp_int_overflow'].name = 'clamp_int_overflow'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_truncate_fixed); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":111
 * __dataclass_fields__['truncate_fixed'].name = 'truncate_fixed'
 * __dataclass_fields__['truncate_fixed'].type = PLACEHOLDER_truncate_fixed
 * __dataclass_fields__['truncate_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_truncate_fixed             # <<<<<<<<<<<<<<
 * __dataclass_fields__['clamp_int_overflow'].name = 'clamp_int_overflow'
 * __dataclass_fields__['clamp_int_overflow'].type = PLACEHOLDER_clamp_int_overflow
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_truncate_fixed); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":112
 * __dataclass_fields__['truncate_fixed'].type = PLACEHOLDER_truncate_fixed
 * __dataclass_fields__['truncate_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_truncate_fixed
 * __dataclass_fields__['clamp_int_overflow'].name = 'clamp_int_overflow'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['clamp_int_overflow'].type = PLACEHOLDER_clamp_int_overflow
 * __dataclass_fields__['clamp_int_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_clamp_int_overflow
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_clamp_int_overflow); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_clamp_int_overflow) < 0) __PYX_ERR(1, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":113
 * __dataclass_fields__['truncate_fixed']._field_type = PLACEHOLDER_FIELD_TYPE_truncate_fixed
 * __dataclass_fields__['clamp_int_overflow'].name = 'clamp_int_overflow'
 * __dataclass_fields__['clamp_int_overflow'].type = PLACEHOLDER_clamp_int_overflow             # <<<<<<<<<<<<<<
 * __dataclass_fields__['clamp_int_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_clamp_int_overflow
 * __dataclass_fields__['clamp_float_overflow'].name = 'clamp_float_overflow'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_clamp_int_overflow); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":114
 * __dataclass_fields__['clamp_int_overflow'].name = 'clamp_int_overflow'
 * __dataclass_fields__['clamp_int_overflow'].type = PLACEHOLDER_clamp_int_overflow
 * __dataclass_fields__['clamp_int_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_clamp_int_overflow             # <<<<<<<<<<<<<<
 * __dataclass_fields__['clamp_float_overflow'].name = 'clamp_float_overflow'
 * __dataclass_fields__['clamp_float_overflow'].type = PLACEHOLDER_clamp_float_overflow
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_clamp_int_overflow); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":115
 * __dataclass_fields__['clamp_int_overflow'].type = PLACEHOLDER_clamp_int_overflow
 * __dataclass_fields__['clamp_int_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_clamp_int_overflow
 * __dataclass_fields__['clamp_float_overflow'].name = 'clamp_float_overflow'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['clamp_float_overflow'].type = PLACEHOLDER_clamp_float_overflow
 * __dataclass_fields__['clamp_float_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_clamp_float_overflow
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_clamp_float_overflow); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_clamp_float_overflow) < 0) __PYX_ERR(1, 115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":116
 * __dataclass_fields__['clamp_int_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_clamp_int_overflow
 * __dataclass_fields__['clamp_float_overflow'].name = 'clamp_float_overflow'
 * __dataclass_fields__['clamp_float_overflow'].type = PLACEHOLDER_clamp_float_overflow             # <<<<<<<<<<<<<<
 * __dataclass_fields__['clamp_float_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_clamp_float_overflow
 * __dataclass_fields__['float_out_of_range_inf'].name = 'float_out_of_range_inf'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_clamp_float_overflow); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":117
 * __dataclass_fields__['clamp_float_overflow'].name = 'clamp_float_overflow'
 * __dataclass_fields__['clamp_float_overflow'].type = PLACEHOLDER_clamp_float_overflow
 * __dataclass_fields__['clamp_float_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_clamp_float_overflow             # <<<<<<<<<<<<<<
 * __dataclass_fields__['float_out_of_range_inf'].name = 'float_out_of_range_inf'
 * __dataclass_fields__['float_out_of_range_inf'].type = PLACEHOLDER_float_out_of_range_inf
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_clamp_float_overflow); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":118
 * __dataclass_fields__['clamp_float_overflow'].type = PLACEHOLDER_clamp_float_overflow
 * __dataclass_fields__['clamp_float_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_clamp_float_overflow
 * __dataclass_fields__['float_out_of_range_inf'].name = 'float_out_of_range_inf'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['float_out_of_range_inf'].type = PLACEHOLDER_float_out_of_range_inf
 * __dataclass_fields__['float_out_of_range_inf']._field_type = PLACEHOLDER_FIELD_TYPE_float_out_of_range_inf
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_float_out_of_range_inf); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_float_out_of_range_inf) < 0) __PYX_ERR(1, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":119
 * __dataclass_fields__['clamp_float_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_clamp_float_overflow
 * __dataclass_fields__['float_out_of_range_inf'].name = 'float_out_of_range_inf'
 * __dataclass_fields__['float_out_of_range_inf'].type = PLACEHOLDER_float_out_of_range_inf             # <<<<<<<<<<<<<<
 * __dataclass_fields__['float_out_of_range_inf']._field_type = PLACEHOLDER_FIELD_TYPE_float_out_of_range_inf
 * __dataclass_fields__['bytes_default_value_utf8'].name = 'bytes_default_value_utf8'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_float_out_of_range_inf); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":120
 * __dataclass_fields__['float_out_of_range_inf'].name = 'float_out_of_range_inf'
 * __dataclass_fields__['float_out_of_range_inf'].type = PLACEHOLDER_float_out_of_range_inf
 * __dataclass_fields__['float_out_of_range_inf']._field_type = PLACEHOLDER_FIELD_TYPE_float_out_of_range_inf             # <<<<<<<<<<<<<<
 * __dataclass_fields__['bytes_default_value_utf8'].name = 'bytes_default_value_utf8'
 * __dataclass_fields__['bytes_default_value_utf8'].type = PLACEHOLDER_bytes_default_value_utf8
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_float_out_of_range_inf); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":121
 * __dataclass_fields__['float_out_of_range_inf'].type = PLACEHOLDER_float_out_of_range_inf
 * __dataclass_fields__['float_out_of_range_inf']._field_type = PLACEHOLDER_FIELD_TYPE_float_out_of_range_inf
 * __dataclass_fields__['bytes_default_value_utf8'].name = 'bytes_default_value_utf8'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['bytes_default_value_utf8'].type = PLACEHOLDER_bytes_default_value_utf8
 * __dataclass_fields__['bytes_default_value_utf8']._field_type = PLACEHOLDER_FIELD_TYPE_bytes_default_value_utf8
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_bytes_default_value_utf8); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_bytes_default_value_utf8) < 0) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":122
 * __dataclass_fields__['float_out_of_range_inf']._field_type = PLACEHOLDER_FIELD_TYPE_float_out_of_range_inf
 * __dataclass_fields__['bytes_default_value_utf8'].name = 'bytes_default_value_utf8'
 * __dataclass_fields__['bytes_default_value_utf8'].type = PLACEHOLDER_bytes_default_value_utf8             # <<<<<<<<<<<<<<
 * __dataclass_fields__['bytes_default_value_utf8']._field_type = PLACEHOLDER_FIELD_TYPE_bytes_default_value_utf8
 * __dataclass_fields__['string_types_default_unchanged'].name = 'string_types_default_unchanged'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_bytes_default_value_utf8); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":123
 * __dataclass_fields__['bytes_default_value_utf8'].name = 'bytes_default_value_utf8'
 * __dataclass_fields__['bytes_default_value_utf8'].type = PLACEHOLDER_bytes_default_value_utf8
 * __dataclass_fields__['bytes_default_value_utf8']._field_type = PLACEHOLDER_FIELD_TYPE_bytes_default_value_utf8             # <<<<<<<<<<<<<<
 * __dataclass_fields__['string_types_default_unchanged'].name = 'string_types_default_unchanged'
 * __dataclass_fields__['string_types_default_unchanged'].type = PLACEHOLDER_string_types_default_unchanged
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_bytes_default_value_utf8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":124
 * __dataclass_fields__['bytes_default_value_utf8'].type = PLACEHOLDER_bytes_default_value_utf8
 * __dataclass_fields__['bytes_default_value_utf8']._field_type = PLACEHOLDER_FIELD_TYPE_bytes_default_value_utf8
 * __dataclass_fields__['string_types_default_unchanged'].name = 'string_types_default_unchanged'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['string_types_default_unchanged'].type = PLACEHOLDER_string_types_default_unchanged
 * __dataclass_fields__['string_types_default_unchanged']._field_type = PLACEHOLDER_FIELD_TYPE_string_types_default_unchanged
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_string_types_default_unchanged); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_string_types_default_unchanged) < 0) __PYX_ERR(1, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":125
 * __dataclass_fields__['bytes_default_value_utf8']._field_type = PLACEHOLDER_FIELD_TYPE_bytes_default_value_utf8
 * __dataclass_fields__['string_types_default_unchanged'].name = 'string_types_default_unchanged'
 * __dataclass_fields__['string_types_default_unchanged'].type = PLACEHOLDER_string_types_default_unchanged             # <<<<<<<<<<<<<<
 * __dataclass_fields__['string_types_default_unchanged']._field_type = PLACEHOLDER_FIELD_TYPE_string_types_default_unchanged
 * __dataclass_fields__['decimal_check_exp_overflow'].name = 'decimal_check_exp_overflow'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_string_types_default_unchanged); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":126
 * __dataclass_fields__['string_types_default_unchanged'].name = 'string_types_default_unchanged'
 * __dataclass_fields__['string_types_default_unchanged'].type = PLACEHOLDER_string_types_default_unchanged
 * __dataclass_fields__['string_types_default_unchanged']._field_type = PLACEHOLDER_FIELD_TYPE_string_types_default_unchanged             # <<<<<<<<<<<<<<
 * __dataclass_fields__['decimal_check_exp_overflow'].name = 'decimal_check_exp_overflow'
 * __dataclass_fields__['decimal_check_exp_overflow'].type = PLACEHOLDER_decimal_check_exp_overflow
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_string_types_default_unchanged); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":127
 * __dataclass_fields__['string_types_default_unchanged'].type = PLACEHOLDER_string_types_default_unchanged
 * __dataclass_fields__['string_types_default_unchanged']._field_type = PLACEHOLDER_FIELD_TYPE_string_types_default_unchanged
 * __dataclass_fields__['decimal_check_exp_overflow'].name = 'decimal_check_exp_overflow'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['decimal_check_exp_overflow'].type = PLACEHOLDER_decimal_check_exp_overflow
 * __dataclass_fields__['decimal_check_exp_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_decimal_check_exp_overflow
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_decimal_check_exp_overflow); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_decimal_check_exp_overflow) < 0) __PYX_ERR(1, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":128
 * __dataclass_fields__['string_types_default_unchanged']._field_type = PLACEHOLDER_FIELD_TYPE_string_types_default_unchanged
 * __dataclass_fields__['decimal_check_exp_overflow'].name = 'decimal_check_exp_overflow'
 * __dataclass_fields__['decimal_check_exp_overflow'].type = PLACEHOLDER_decimal_check_exp_overflow             # <<<<<<<<<<<<<<
 * __dataclass_fields__['decimal_check_exp_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_decimal_check_exp_overflow
 * __dataclass_fields__['types_str_to_schema'].name = 'types_str_to_schema'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_decimal_check_exp_overflow); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":129
 * __dataclass_fields__['decimal_check_exp_overflow'].name = 'decimal_check_exp_overflow'
 * __dataclass_fields__['decimal_check_exp_overflow'].type = PLACEHOLDER_decimal_check_exp_overflow
 * __dataclass_fields__['decimal_check_exp_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_decimal_check_exp_overflow             # <<<<<<<<<<<<<<
 * __dataclass_fields__['types_str_to_schema'].name = 'types_str_to_schema'
 * __dataclass_fields__['types_str_to_schema'].type = PLACEHOLDER_types_str_to_schema
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_decimal_check_exp_overflow); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":130
 * __dataclass_fields__['decimal_check_exp_overflow'].type = PLACEHOLDER_decimal_check_exp_overflow
 * __dataclass_fields__['decimal_check_exp_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_decimal_check_exp_overflow
 * __dataclass_fields__['types_str_to_schema'].name = 'types_str_to_schema'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['types_str_to_schema'].type = PLACEHOLDER_types_str_to_schema
 * __dataclass_fields__['types_str_to_schema']._field_type = PLACEHOLDER_FIELD_TYPE_types_str_to_schema
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_types_str_to_schema); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_types_str_to_schema) < 0) __PYX_ERR(1, 130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":131
 * __dataclass_fields__['decimal_check_exp_overflow']._field_type = PLACEHOLDER_FIELD_TYPE_decimal_check_exp_overflow
 * __dataclass_fields__['types_str_to_schema'].name = 'types_str_to_schema'
 * __dataclass_fields__['types_str_to_schema'].type = PLACEHOLDER_types_str_to_schema             # <<<<<<<<<<<<<<
 * __dataclass_fields__['types_str_to_schema']._field_type = PLACEHOLDER_FIELD_TYPE_types_str_to_schema
 * __dataclass_fields__['logical_types'].name = 'logical_types'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_types_str_to_schema); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":132
 * __dataclass_fields__['types_str_to_schema'].name = 'types_str_to_schema'
 * __dataclass_fields__['types_str_to_schema'].type = PLACEHOLDER_types_str_to_schema
 * __dataclass_fields__['types_str_to_schema']._field_type = PLACEHOLDER_FIELD_TYPE_types_str_to_schema             # <<<<<<<<<<<<<<
 * __dataclass_fields__['logical_types'].name = 'logical_types'
 * __dataclass_fields__['logical_types'].type = PLACEHOLDER_logical_types
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_types_str_to_schema); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":133
 * __dataclass_fields__['types_str_to_schema'].type = PLACEHOLDER_types_str_to_schema
 * __dataclass_fields__['types_str_to_schema']._field_type = PLACEHOLDER_FIELD_TYPE_types_str_to_schema
 * __dataclass_fields__['logical_types'].name = 'logical_types'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['logical_types'].type = PLACEHOLDER_logical_types
 * __dataclass_fields__['logical_types']._field_type = PLACEHOLDER_FIELD_TYPE_logical_types
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_logical_types); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_logical_types) < 0) __PYX_ERR(1, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":134
 * __dataclass_fields__['types_str_to_schema']._field_type = PLACEHOLDER_FIELD_TYPE_types_str_to_schema
 * __dataclass_fields__['logical_types'].name = 'logical_types'
 * __dataclass_fields__['logical_types'].type = PLACEHOLDER_logical_types             # <<<<<<<<<<<<<<
 * __dataclass_fields__['logical_types']._field_type = PLACEHOLDER_FIELD_TYPE_logical_types
 * __dataclass_fields__['adapt_record_types'].name = 'adapt_record_types'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_logical_types); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_kp_s_tuple_LogicalType) < 0) __PYX_ERR(1, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":135
 * __dataclass_fields__['logical_types'].name = 'logical_types'
 * __dataclass_fields__['logical_types'].type = PLACEHOLDER_logical_types
 * __dataclass_fields__['logical_types']._field_type = PLACEHOLDER_FIELD_TYPE_logical_types             # <<<<<<<<<<<<<<
 * __dataclass_fields__['adapt_record_types'].name = 'adapt_record_types'
 * __dataclass_fields__['adapt_record_types'].type = PLACEHOLDER_adapt_record_types
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_logical_types); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":136
 * __dataclass_fields__['logical_types'].type = PLACEHOLDER_logical_types
 * __dataclass_fields__['logical_types']._field_type = PLACEHOLDER_FIELD_TYPE_logical_types
 * __dataclass_fields__['adapt_record_types'].name = 'adapt_record_types'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['adapt_record_types'].type = PLACEHOLDER_adapt_record_types
 * __dataclass_fields__['adapt_record_types']._field_type = PLACEHOLDER_FIELD_TYPE_adapt_record_types
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_adapt_record_types); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_adapt_record_types) < 0) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":137
 * __dataclass_fields__['logical_types']._field_type = PLACEHOLDER_FIELD_TYPE_logical_types
 * __dataclass_fields__['adapt_record_types'].name = 'adapt_record_types'
 * __dataclass_fields__['adapt_record_types'].type = PLACEHOLDER_adapt_record_types             # <<<<<<<<<<<<<<
 * __dataclass_fields__['adapt_record_types']._field_type = PLACEHOLDER_FIELD_TYPE_adapt_record_types
 * __dataclass_fields__['return_uuid_object'].name = 'return_uuid_object'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_adapt_record_types); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":138
 * __dataclass_fields__['adapt_record_types'].name = 'adapt_record_types'
 * __dataclass_fields__['adapt_record_types'].type = PLACEHOLDER_adapt_record_types
 * __dataclass_fields__['adapt_record_types']._field_type = PLACEHOLDER_FIELD_TYPE_adapt_record_types             # <<<<<<<<<<<<<<
 * __dataclass_fields__['return_uuid_object'].name = 'return_uuid_object'
 * __dataclass_fields__['return_uuid_object'].type = PLACEHOLDER_return_uuid_object
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_adapt_record_types); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":139
 * __dataclass_fields__['adapt_record_types'].type = PLACEHOLDER_adapt_record_types
 * __dataclass_fields__['adapt_record_types']._field_type = PLACEHOLDER_FIELD_TYPE_adapt_record_types
 * __dataclass_fields__['return_uuid_object'].name = 'return_uuid_object'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['return_uuid_object'].type = PLACEHOLDER_return_uuid_object
 * __dataclass_fields__['return_uuid_object']._field_type = PLACEHOLDER_FIELD_TYPE_return_uuid_object
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_return_uuid_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_return_uuid_object) < 0) __PYX_ERR(1, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":140
 * __dataclass_fields__['adapt_record_types']._field_type = PLACEHOLDER_FIELD_TYPE_adapt_record_types
 * __dataclass_fields__['return_uuid_object'].name = 'return_uuid_object'
 * __dataclass_fields__['return_uuid_object'].type = PLACEHOLDER_return_uuid_object             # <<<<<<<<<<<<<<
 * __dataclass_fields__['return_uuid_object']._field_type = PLACEHOLDER_FIELD_TYPE_return_uuid_object
 * __dataclass_fields__['allow_error_type'].name = 'allow_error_type'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_return_uuid_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":141
 * __dataclass_fields__['return_uuid_object'].name = 'return_uuid_object'
 * __dataclass_fields__['return_uuid_object'].type = PLACEHOLDER_return_uuid_object
 * __dataclass_fields__['return_uuid_object']._field_type = PLACEHOLDER_FIELD_TYPE_return_uuid_object             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_error_type'].name = 'allow_error_type'
 * __dataclass_fields__['allow_error_type'].type = PLACEHOLDER_allow_error_type
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_return_uuid_object); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":142
 * __dataclass_fields__['return_uuid_object'].type = PLACEHOLDER_return_uuid_object
 * __dataclass_fields__['return_uuid_object']._field_type = PLACEHOLDER_FIELD_TYPE_return_uuid_object
 * __dataclass_fields__['allow_error_type'].name = 'allow_error_type'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_error_type'].type = PLACEHOLDER_allow_error_type
 * __dataclass_fields__['allow_error_type']._field_type = PLACEHOLDER_FIELD_TYPE_allow_error_type
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_allow_error_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_allow_error_type) < 0) __PYX_ERR(1, 142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":143
 * __dataclass_fields__['return_uuid_object']._field_type = PLACEHOLDER_FIELD_TYPE_return_uuid_object
 * __dataclass_fields__['allow_error_type'].name = 'allow_error_type'
 * __dataclass_fields__['allow_error_type'].type = PLACEHOLDER_allow_error_type             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_error_type']._field_type = PLACEHOLDER_FIELD_TYPE_allow_error_type
 * __dataclass_fields__['allow_leading_dot_in_names'].name = 'allow_leading_dot_in_names'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_allow_error_type); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":144
 * __dataclass_fields__['allow_error_type'].name = 'allow_error_type'
 * __dataclass_fields__['allow_error_type'].type = PLACEHOLDER_allow_error_type
 * __dataclass_fields__['allow_error_type']._field_type = PLACEHOLDER_FIELD_TYPE_allow_error_type             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_leading_dot_in_names'].name = 'allow_leading_dot_in_names'
 * __dataclass_fields__['allow_leading_dot_in_names'].type = PLACEHOLDER_allow_leading_dot_in_names
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_allow_error_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":145
 * __dataclass_fields__['allow_error_type'].type = PLACEHOLDER_allow_error_type
 * __dataclass_fields__['allow_error_type']._field_type = PLACEHOLDER_FIELD_TYPE_allow_error_type
 * __dataclass_fields__['allow_leading_dot_in_names'].name = 'allow_leading_dot_in_names'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_leading_dot_in_names'].type = PLACEHOLDER_allow_leading_dot_in_names
 * __dataclass_fields__['allow_leading_dot_in_names']._field_type = PLACEHOLDER_FIELD_TYPE_allow_leading_dot_in_names
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_allow_leading_dot_in_names); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_allow_leading_dot_in_names) < 0) __PYX_ERR(1, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":146
 * __dataclass_fields__['allow_error_type']._field_type = PLACEHOLDER_FIELD_TYPE_allow_error_type
 * __dataclass_fields__['allow_leading_dot_in_names'].name = 'allow_leading_dot_in_names'
 * __dataclass_fields__['allow_leading_dot_in_names'].type = PLACEHOLDER_allow_leading_dot_in_names             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_leading_dot_in_names']._field_type = PLACEHOLDER_FIELD_TYPE_allow_leading_dot_in_names
 * __dataclass_fields__['naive_dt_assume_utc'].name = 'naive_dt_assume_utc'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_allow_leading_dot_in_names); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":147
 * __dataclass_fields__['allow_leading_dot_in_names'].name = 'allow_leading_dot_in_names'
 * __dataclass_fields__['allow_leading_dot_in_names'].type = PLACEHOLDER_allow_leading_dot_in_names
 * __dataclass_fields__['allow_leading_dot_in_names']._field_type = PLACEHOLDER_FIELD_TYPE_allow_leading_dot_in_names             # <<<<<<<<<<<<<<
 * __dataclass_fields__['naive_dt_assume_utc'].name = 'naive_dt_assume_utc'
 * __dataclass_fields__['naive_dt_assume_utc'].type = PLACEHOLDER_naive_dt_assume_utc
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_allow_leading_dot_in_names); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":148
 * __dataclass_fields__['allow_leading_dot_in_names'].type = PLACEHOLDER_allow_leading_dot_in_names
 * __dataclass_fields__['allow_leading_dot_in_names']._field_type = PLACEHOLDER_FIELD_TYPE_allow_leading_dot_in_names
 * __dataclass_fields__['naive_dt_assume_utc'].name = 'naive_dt_assume_utc'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['naive_dt_assume_utc'].type = PLACEHOLDER_naive_dt_assume_utc
 * __dataclass_fields__['naive_dt_assume_utc']._field_type = PLACEHOLDER_FIELD_TYPE_naive_dt_assume_utc
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_naive_dt_assume_utc); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_naive_dt_assume_utc) < 0) __PYX_ERR(1, 148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":149
 * __dataclass_fields__['allow_leading_dot_in_names']._field_type = PLACEHOLDER_FIELD_TYPE_allow_leading_dot_in_names
 * __dataclass_fields__['naive_dt_assume_utc'].name = 'naive_dt_assume_utc'
 * __dataclass_fields__['naive_dt_assume_utc'].type = PLACEHOLDER_naive_dt_assume_utc             # <<<<<<<<<<<<<<
 * __dataclass_fields__['naive_dt_assume_utc']._field_type = PLACEHOLDER_FIELD_TYPE_naive_dt_assume_utc
 * __dataclass_fields__['alternate_timestamp_millis_encoding'].name = 'alternate_timestamp_millis_encoding'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_naive_dt_assume_utc); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":150
 * __dataclass_fields__['naive_dt_assume_utc'].name = 'naive_dt_assume_utc'
 * __dataclass_fields__['naive_dt_assume_utc'].type = PLACEHOLDER_naive_dt_assume_utc
 * __dataclass_fields__['naive_dt_assume_utc']._field_type = PLACEHOLDER_FIELD_TYPE_naive_dt_assume_utc             # <<<<<<<<<<<<<<
 * __dataclass_fields__['alternate_timestamp_millis_encoding'].name = 'alternate_timestamp_millis_encoding'
 * __dataclass_fields__['alternate_timestamp_millis_encoding'].type = PLACEHOLDER_alternate_timestamp_millis_encoding
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_naive_dt_assume_utc); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":151
 * __dataclass_fields__['naive_dt_assume_utc'].type = PLACEHOLDER_naive_dt_assume_utc
 * __dataclass_fields__['naive_dt_assume_utc']._field_type = PLACEHOLDER_FIELD_TYPE_naive_dt_assume_utc
 * __dataclass_fields__['alternate_timestamp_millis_encoding'].name = 'alternate_timestamp_millis_encoding'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['alternate_timestamp_millis_encoding'].type = PLACEHOLDER_alternate_timestamp_millis_encoding
 * __dataclass_fields__['alternate_timestamp_millis_encoding']._field_type = PLACEHOLDER_FIELD_TYPE_alternate_timestamp_millis_encoding
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_alternate_timestamp_millis_encod); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_alternate_timestamp_millis_encod) < 0) __PYX_ERR(1, 151, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":152
 * __dataclass_fields__['naive_dt_assume_utc']._field_type = PLACEHOLDER_FIELD_TYPE_naive_dt_assume_utc
 * __dataclass_fields__['alternate_timestamp_millis_encoding'].name = 'alternate_timestamp_millis_encoding'
 * __dataclass_fields__['alternate_timestamp_millis_encoding'].type = PLACEHOLDER_alternate_timestamp_millis_encoding             # <<<<<<<<<<<<<<
 * __dataclass_fields__['alternate_timestamp_millis_encoding']._field_type = PLACEHOLDER_FIELD_TYPE_alternate_timestamp_millis_encoding
 * __dataclass_fields__['date_type_accepts_string'].name = 'date_type_accepts_string'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_alternate_timestamp_millis_encod); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":153
 * __dataclass_fields__['alternate_timestamp_millis_encoding'].name = 'alternate_timestamp_millis_encoding'
 * __dataclass_fields__['alternate_timestamp_millis_encoding'].type = PLACEHOLDER_alternate_timestamp_millis_encoding
 * __dataclass_fields__['alternate_timestamp_millis_encoding']._field_type = PLACEHOLDER_FIELD_TYPE_alternate_timestamp_millis_encoding             # <<<<<<<<<<<<<<
 * __dataclass_fields__['date_type_accepts_string'].name = 'date_type_accepts_string'
 * __dataclass_fields__['date_type_accepts_string'].type = PLACEHOLDER_date_type_accepts_string
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_alternate_timestamp_millis_encod); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 153, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":154
 * __dataclass_fields__['alternate_timestamp_millis_encoding'].type = PLACEHOLDER_alternate_timestamp_millis_encoding
 * __dataclass_fields__['alternate_timestamp_millis_encoding']._field_type = PLACEHOLDER_FIELD_TYPE_alternate_timestamp_millis_encoding
 * __dataclass_fields__['date_type_accepts_string'].name = 'date_type_accepts_string'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['date_type_accepts_string'].type = PLACEHOLDER_date_type_accepts_string
 * __dataclass_fields__['date_type_accepts_string']._field_type = PLACEHOLDER_FIELD_TYPE_date_type_accepts_string
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_date_type_accepts_string); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_date_type_accepts_string) < 0) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":155
 * __dataclass_fields__['alternate_timestamp_millis_encoding']._field_type = PLACEHOLDER_FIELD_TYPE_alternate_timestamp_millis_encoding
 * __dataclass_fields__['date_type_accepts_string'].name = 'date_type_accepts_string'
 * __dataclass_fields__['date_type_accepts_string'].type = PLACEHOLDER_date_type_accepts_string             # <<<<<<<<<<<<<<
 * __dataclass_fields__['date_type_accepts_string']._field_type = PLACEHOLDER_FIELD_TYPE_date_type_accepts_string
 * __dataclass_fields__['raise_on_invalid_logical'].name = 'raise_on_invalid_logical'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_date_type_accepts_string); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":156
 * __dataclass_fields__['date_type_accepts_string'].name = 'date_type_accepts_string'
 * __dataclass_fields__['date_type_accepts_string'].type = PLACEHOLDER_date_type_accepts_string
 * __dataclass_fields__['date_type_accepts_string']._field_type = PLACEHOLDER_FIELD_TYPE_date_type_accepts_string             # <<<<<<<<<<<<<<
 * __dataclass_fields__['raise_on_invalid_logical'].name = 'raise_on_invalid_logical'
 * __dataclass_fields__['raise_on_invalid_logical'].type = PLACEHOLDER_raise_on_invalid_logical
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_date_type_accepts_string); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":157
 * __dataclass_fields__['date_type_accepts_string'].type = PLACEHOLDER_date_type_accepts_string
 * __dataclass_fields__['date_type_accepts_string']._field_type = PLACEHOLDER_FIELD_TYPE_date_type_accepts_string
 * __dataclass_fields__['raise_on_invalid_logical'].name = 'raise_on_invalid_logical'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['raise_on_invalid_logical'].type = PLACEHOLDER_raise_on_invalid_logical
 * __dataclass_fields__['raise_on_invalid_logical']._field_type = PLACEHOLDER_FIELD_TYPE_raise_on_invalid_logical
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_raise_on_invalid_logical); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_raise_on_invalid_logical) < 0) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":158
 * __dataclass_fields__['date_type_accepts_string']._field_type = PLACEHOLDER_FIELD_TYPE_date_type_accepts_string
 * __dataclass_fields__['raise_on_invalid_logical'].name = 'raise_on_invalid_logical'
 * __dataclass_fields__['raise_on_invalid_logical'].type = PLACEHOLDER_raise_on_invalid_logical             # <<<<<<<<<<<<<<
 * __dataclass_fields__['raise_on_invalid_logical']._field_type = PLACEHOLDER_FIELD_TYPE_raise_on_invalid_logical
 * __dataclass_fields__['inline_namespaces'].name = 'inline_namespaces'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_raise_on_invalid_logical); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":159
 * __dataclass_fields__['raise_on_invalid_logical'].name = 'raise_on_invalid_logical'
 * __dataclass_fields__['raise_on_invalid_logical'].type = PLACEHOLDER_raise_on_invalid_logical
 * __dataclass_fields__['raise_on_invalid_logical']._field_type = PLACEHOLDER_FIELD_TYPE_raise_on_invalid_logical             # <<<<<<<<<<<<<<
 * __dataclass_fields__['inline_namespaces'].name = 'inline_namespaces'
 * __dataclass_fields__['inline_namespaces'].type = PLACEHOLDER_inline_namespaces
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_raise_on_invalid_logical); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":160
 * __dataclass_fields__['raise_on_invalid_logical'].type = PLACEHOLDER_raise_on_invalid_logical
 * __dataclass_fields__['raise_on_invalid_logical']._field_type = PLACEHOLDER_FIELD_TYPE_raise_on_invalid_logical
 * __dataclass_fields__['inline_namespaces'].name = 'inline_namespaces'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['inline_namespaces'].type = PLACEHOLDER_inline_namespaces
 * __dataclass_fields__['inline_namespaces']._field_type = PLACEHOLDER_FIELD_TYPE_inline_namespaces
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_inline_namespaces); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_inline_namespaces) < 0) __PYX_ERR(1, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":161
 * __dataclass_fields__['raise_on_invalid_logical']._field_type = PLACEHOLDER_FIELD_TYPE_raise_on_invalid_logical
 * __dataclass_fields__['inline_namespaces'].name = 'inline_namespaces'
 * __dataclass_fields__['inline_namespaces'].type = PLACEHOLDER_inline_namespaces             # <<<<<<<<<<<<<<
 * __dataclass_fields__['inline_namespaces']._field_type = PLACEHOLDER_FIELD_TYPE_inline_namespaces
 * __dataclass_fields__['expand_types_in_schema'].name = 'expand_types_in_schema'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_inline_namespaces); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":162
 * __dataclass_fields__['inline_namespaces'].name = 'inline_namespaces'
 * __dataclass_fields__['inline_namespaces'].type = PLACEHOLDER_inline_namespaces
 * __dataclass_fields__['inline_namespaces']._field_type = PLACEHOLDER_FIELD_TYPE_inline_namespaces             # <<<<<<<<<<<<<<
 * __dataclass_fields__['expand_types_in_schema'].name = 'expand_types_in_schema'
 * __dataclass_fields__['expand_types_in_schema'].type = PLACEHOLDER_expand_types_in_schema
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_inline_namespaces); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":163
 * __dataclass_fields__['inline_namespaces'].type = PLACEHOLDER_inline_namespaces
 * __dataclass_fields__['inline_namespaces']._field_type = PLACEHOLDER_FIELD_TYPE_inline_namespaces
 * __dataclass_fields__['expand_types_in_schema'].name = 'expand_types_in_schema'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['expand_types_in_schema'].type = PLACEHOLDER_expand_types_in_schema
 * __dataclass_fields__['expand_types_in_schema']._field_type = PLACEHOLDER_FIELD_TYPE_expand_types_in_schema
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_expand_types_in_schema); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_expand_types_in_schema) < 0) __PYX_ERR(1, 163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":164
 * __dataclass_fields__['inline_namespaces']._field_type = PLACEHOLDER_FIELD_TYPE_inline_namespaces
 * __dataclass_fields__['expand_types_in_schema'].name = 'expand_types_in_schema'
 * __dataclass_fields__['expand_types_in_schema'].type = PLACEHOLDER_expand_types_in_schema             # <<<<<<<<<<<<<<
 * __dataclass_fields__['expand_types_in_schema']._field_type = PLACEHOLDER_FIELD_TYPE_expand_types_in_schema
 * __dataclass_fields__['unicode_errors'].name = 'unicode_errors'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_expand_types_in_schema); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 164, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":165
 * __dataclass_fields__['expand_types_in_schema'].name = 'expand_types_in_schema'
 * __dataclass_fields__['expand_types_in_schema'].type = PLACEHOLDER_expand_types_in_schema
 * __dataclass_fields__['expand_types_in_schema']._field_type = PLACEHOLDER_FIELD_TYPE_expand_types_in_schema             # <<<<<<<<<<<<<<
 * __dataclass_fields__['unicode_errors'].name = 'unicode_errors'
 * __dataclass_fields__['unicode_errors'].type = PLACEHOLDER_unicode_errors
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_expand_types_in_schema); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":166
 * __dataclass_fields__['expand_types_in_schema'].type = PLACEHOLDER_expand_types_in_schema
 * __dataclass_fields__['expand_types_in_schema']._field_type = PLACEHOLDER_FIELD_TYPE_expand_types_in_schema
 * __dataclass_fields__['unicode_errors'].name = 'unicode_errors'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['unicode_errors'].type = PLACEHOLDER_unicode_errors
 * __dataclass_fields__['unicode_errors']._field_type = PLACEHOLDER_FIELD_TYPE_unicode_errors
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_unicode_errors); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_unicode_errors) < 0) __PYX_ERR(1, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":167
 * __dataclass_fields__['expand_types_in_schema']._field_type = PLACEHOLDER_FIELD_TYPE_expand_types_in_schema
 * __dataclass_fields__['unicode_errors'].name = 'unicode_errors'
 * __dataclass_fields__['unicode_errors'].type = PLACEHOLDER_unicode_errors             # <<<<<<<<<<<<<<
 * __dataclass_fields__['unicode_errors']._field_type = PLACEHOLDER_FIELD_TYPE_unicode_errors
 * __dataclass_fields__['container_fill_blocks'].name = 'container_fill_blocks'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_unicode_errors); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_str_2) < 0) __PYX_ERR(1, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":168
 * __dataclass_fields__['unicode_errors'].name = 'unicode_errors'
 * __dataclass_fields__['unicode_errors'].type = PLACEHOLDER_unicode_errors
 * __dataclass_fields__['unicode_errors']._field_type = PLACEHOLDER_FIELD_TYPE_unicode_errors             # <<<<<<<<<<<<<<
 * __dataclass_fields__['container_fill_blocks'].name = 'container_fill_blocks'
 * __dataclass_fields__['container_fill_blocks'].type = PLACEHOLDER_container_fill_blocks
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_unicode_errors); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":169
 * __dataclass_fields__['unicode_errors'].type = PLACEHOLDER_unicode_errors
 * __dataclass_fields__['unicode_errors']._field_type = PLACEHOLDER_FIELD_TYPE_unicode_errors
 * __dataclass_fields__['container_fill_blocks'].name = 'container_fill_blocks'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['container_fill_blocks'].type = PLACEHOLDER_container_fill_blocks
 * __dataclass_fields__['container_fill_blocks']._field_type = PLACEHOLDER_FIELD_TYPE_container_fill_blocks
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_container_fill_blocks); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_container_fill_blocks) < 0) __PYX_ERR(1, 169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":170
 * __dataclass_fields__['unicode_errors']._field_type = PLACEHOLDER_FIELD_TYPE_unicode_errors
 * __dataclass_fields__['container_fill_blocks'].name = 'container_fill_blocks'
 * __dataclass_fields__['container_fill_blocks'].type = PLACEHOLDER_container_fill_blocks             # <<<<<<<<<<<<<<
 * __dataclass_fields__['container_fill_blocks']._field_type = PLACEHOLDER_FIELD_TYPE_container_fill_blocks
 * __dataclass_fields__['defer_schema_promotion_errors'].name = 'defer_schema_promotion_errors'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_container_fill_blocks); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":171
 * __dataclass_fields__['container_fill_blocks'].name = 'container_fill_blocks'
 * __dataclass_fields__['container_fill_blocks'].type = PLACEHOLDER_container_fill_blocks
 * __dataclass_fields__['container_fill_blocks']._field_type = PLACEHOLDER_FIELD_TYPE_container_fill_blocks             # <<<<<<<<<<<<<<
 * __dataclass_fields__['defer_schema_promotion_errors'].name = 'defer_schema_promotion_errors'
 * __dataclass_fields__['defer_schema_promotion_errors'].type = PLACEHOLDER_defer_schema_promotion_errors
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_container_fill_blocks); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":172
 * __dataclass_fields__['container_fill_blocks'].type = PLACEHOLDER_container_fill_blocks
 * __dataclass_fields__['container_fill_blocks']._field_type = PLACEHOLDER_FIELD_TYPE_container_fill_blocks
 * __dataclass_fields__['defer_schema_promotion_errors'].name = 'defer_schema_promotion_errors'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['defer_schema_promotion_errors'].type = PLACEHOLDER_defer_schema_promotion_errors
 * __dataclass_fields__['defer_schema_promotion_errors']._field_type = PLACEHOLDER_FIELD_TYPE_defer_schema_promotion_errors
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_defer_schema_promotion_errors); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_defer_schema_promotion_errors) < 0) __PYX_ERR(1, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":173
 * __dataclass_fields__['container_fill_blocks']._field_type = PLACEHOLDER_FIELD_TYPE_container_fill_blocks
 * __dataclass_fields__['defer_schema_promotion_errors'].name = 'defer_schema_promotion_errors'
 * __dataclass_fields__['defer_schema_promotion_errors'].type = PLACEHOLDER_defer_schema_promotion_errors             # <<<<<<<<<<<<<<
 * __dataclass_fields__['defer_schema_promotion_errors']._field_type = PLACEHOLDER_FIELD_TYPE_defer_schema_promotion_errors
 * __dataclass_fields__['invalid_value_includes_record_name'].name = 'invalid_value_includes_record_name'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_defer_schema_promotion_errors); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":174
 * __dataclass_fields__['defer_schema_promotion_errors'].name = 'defer_schema_promotion_errors'
 * __dataclass_fields__['defer_schema_promotion_errors'].type = PLACEHOLDER_defer_schema_promotion_errors
 * __dataclass_fields__['defer_schema_promotion_errors']._field_type = PLACEHOLDER_FIELD_TYPE_defer_schema_promotion_errors             # <<<<<<<<<<<<<<
 * __dataclass_fields__['invalid_value_includes_record_name'].name = 'invalid_value_includes_record_name'
 * __dataclass_fields__['invalid_value_includes_record_name'].type = PLACEHOLDER_invalid_value_includes_record_name
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_defer_schema_promotion_errors); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":175
 * __dataclass_fields__['defer_schema_promotion_errors'].type = PLACEHOLDER_defer_schema_promotion_errors
 * __dataclass_fields__['defer_schema_promotion_errors']._field_type = PLACEHOLDER_FIELD_TYPE_defer_schema_promotion_errors
 * __dataclass_fields__['invalid_value_includes_record_name'].name = 'invalid_value_includes_record_name'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['invalid_value_includes_record_name'].type = PLACEHOLDER_invalid_value_includes_record_name
 * __dataclass_fields__['invalid_value_includes_record_name']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_includes_record_name
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_invalid_value_includes_record_na); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_invalid_value_includes_record_na) < 0) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":176
 * __dataclass_fields__['defer_schema_promotion_errors']._field_type = PLACEHOLDER_FIELD_TYPE_defer_schema_promotion_errors
 * __dataclass_fields__['invalid_value_includes_record_name'].name = 'invalid_value_includes_record_name'
 * __dataclass_fields__['invalid_value_includes_record_name'].type = PLACEHOLDER_invalid_value_includes_record_name             # <<<<<<<<<<<<<<
 * __dataclass_fields__['invalid_value_includes_record_name']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_includes_record_name
 * __dataclass_fields__['invalid_value_include_array_index'].name = 'invalid_value_include_array_index'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_invalid_value_includes_record_na); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":177
 * __dataclass_fields__['invalid_value_includes_record_name'].name = 'invalid_value_includes_record_name'
 * __dataclass_fields__['invalid_value_includes_record_name'].type = PLACEHOLDER_invalid_value_includes_record_name
 * __dataclass_fields__['invalid_value_includes_record_name']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_includes_record_name             # <<<<<<<<<<<<<<
 * __dataclass_fields__['invalid_value_include_array_index'].name = 'invalid_value_include_array_index'
 * __dataclass_fields__['invalid_value_include_array_index'].type = PLACEHOLDER_invalid_value_include_array_index
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_invalid_value_includes_record_na); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":178
 * __dataclass_fields__['invalid_value_includes_record_name'].type = PLACEHOLDER_invalid_value_includes_record_name
 * __dataclass_fields__['invalid_value_includes_record_name']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_includes_record_name
 * __dataclass_fields__['invalid_value_include_array_index'].name = 'invalid_value_include_array_index'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['invalid_value_include_array_index'].type = PLACEHOLDER_invalid_value_include_array_index
 * __dataclass_fields__['invalid_value_include_array_index']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_include_array_index
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_invalid_value_include_array_inde); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_invalid_value_include_array_inde) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":179
 * __dataclass_fields__['invalid_value_includes_record_name']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_includes_record_name
 * __dataclass_fields__['invalid_value_include_array_index'].name = 'invalid_value_include_array_index'
 * __dataclass_fields__['invalid_value_include_array_index'].type = PLACEHOLDER_invalid_value_include_array_index             # <<<<<<<<<<<<<<
 * __dataclass_fields__['invalid_value_include_array_index']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_include_array_index
 * __dataclass_fields__['invalid_value_include_map_key'].name = 'invalid_value_include_map_key'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_invalid_value_include_array_inde); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":180
 * __dataclass_fields__['invalid_value_include_array_index'].name = 'invalid_value_include_array_index'
 * __dataclass_fields__['invalid_value_include_array_index'].type = PLACEHOLDER_invalid_value_include_array_index
 * __dataclass_fields__['invalid_value_include_array_index']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_include_array_index             # <<<<<<<<<<<<<<
 * __dataclass_fields__['invalid_value_include_map_key'].name = 'invalid_value_include_map_key'
 * __dataclass_fields__['invalid_value_include_map_key'].type = PLACEHOLDER_invalid_value_include_map_key
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_invalid_value_include_array_inde); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":181
 * __dataclass_fields__['invalid_value_include_array_index'].type = PLACEHOLDER_invalid_value_include_array_index
 * __dataclass_fields__['invalid_value_include_array_index']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_include_array_index
 * __dataclass_fields__['invalid_value_include_map_key'].name = 'invalid_value_include_map_key'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['invalid_value_include_map_key'].type = PLACEHOLDER_invalid_value_include_map_key
 * __dataclass_fields__['invalid_value_include_map_key']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_include_map_key
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_invalid_value_include_map_key); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_invalid_value_include_map_key) < 0) __PYX_ERR(1, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":182
 * __dataclass_fields__['invalid_value_include_array_index']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_include_array_index
 * __dataclass_fields__['invalid_value_include_map_key'].name = 'invalid_value_include_map_key'
 * __dataclass_fields__['invalid_value_include_map_key'].type = PLACEHOLDER_invalid_value_include_map_key             # <<<<<<<<<<<<<<
 * __dataclass_fields__['invalid_value_include_map_key']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_include_map_key
 * __dataclass_fields__['allow_invalid_default_values'].name = 'allow_invalid_default_values'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_invalid_value_include_map_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":183
 * __dataclass_fields__['invalid_value_include_map_key'].name = 'invalid_value_include_map_key'
 * __dataclass_fields__['invalid_value_include_map_key'].type = PLACEHOLDER_invalid_value_include_map_key
 * __dataclass_fields__['invalid_value_include_map_key']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_include_map_key             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_invalid_default_values'].name = 'allow_invalid_default_values'
 * __dataclass_fields__['allow_invalid_default_values'].type = PLACEHOLDER_allow_invalid_default_values
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_invalid_value_include_map_key); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":184
 * __dataclass_fields__['invalid_value_include_map_key'].type = PLACEHOLDER_invalid_value_include_map_key
 * __dataclass_fields__['invalid_value_include_map_key']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_include_map_key
 * __dataclass_fields__['allow_invalid_default_values'].name = 'allow_invalid_default_values'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_invalid_default_values'].type = PLACEHOLDER_allow_invalid_default_values
 * __dataclass_fields__['allow_invalid_default_values']._field_type = PLACEHOLDER_FIELD_TYPE_allow_invalid_default_values
 */
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_allow_invalid_default_values); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_allow_invalid_default_values) < 0) __PYX_ERR(1, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":185
 * __dataclass_fields__['invalid_value_include_map_key']._field_type = PLACEHOLDER_FIELD_TYPE_invalid_value_include_map_key
 * __dataclass_fields__['allow_invalid_default_values'].name = 'allow_invalid_default_values'
 * __dataclass_fields__['allow_invalid_default_values'].type = PLACEHOLDER_allow_invalid_default_values             # <<<<<<<<<<<<<<
 * __dataclass_fields__['allow_invalid_default_values']._field_type = PLACEHOLDER_FIELD_TYPE_allow_invalid_default_values
 * __dataclass_fields__['externally_defined_types'].name = 'externally_defined_types'
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_allow_invalid_default_values); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_type, __pyx_n_s_bint) < 0) __PYX_ERR(1, 185, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":186
 * __dataclass_fields__['allow_invalid_default_values'].name = 'allow_invalid_default_values'
 * __dataclass_fields__['allow_invalid_default_values'].type = PLACEHOLDER_allow_invalid_default_values
 * __dataclass_fields__['allow_invalid_default_values']._field_type = PLACEHOLDER_FIELD_TYPE_allow_invalid_default_values             # <<<<<<<<<<<<<<
 * __dataclass_fields__['externally_defined_types'].name = 'externally_defined_types'
 * __dataclass_fields__['externally_defined_types'].type = PLACEHOLDER_externally_defined_types
 */
  __pyx_t_9 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GetNameInClass(__pyx_t_9, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_s_allow_invalid_default_values); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":187
 * __dataclass_fields__['allow_invalid_default_values'].type = PLACEHOLDER_allow_invalid_default_values
 * __dataclass_fields__['allow_invalid_default_values']._field_type = PLACEHOLDER_FIELD_TYPE_allow_invalid_default_values
 * __dataclass_fields__['externally_defined_types'].name = 'externally_defined_types'             # <<<<<<<<<<<<<<
 * __dataclass_fields__['externally_defined_types'].type = PLACEHOLDER_externally_defined_types
 * __dataclass_fields__['externally_defined_types']._field_type = PLACEHOLDER_FIELD_TYPE_externally_defined_types
 */
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_externally_defined_types); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_n_s_name, __pyx_n_s_externally_defined_types) < 0) __PYX_ERR(1, 187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":188
 * __dataclass_fields__['allow_invalid_default_values']._field_type = PLACEHOLDER_FIELD_TYPE_allow_invalid_default_values
 * __dataclass_fields__['externally_defined_types'].name = 'externally_defined_types'
 * __dataclass_fields__['externally_defined_types'].type = PLACEHOLDER_externally_defined_types             # <<<<<<<<<<<<<<
 * __dataclass_fields__['externally_defined_types']._field_type = PLACEHOLDER_FIELD_TYPE_externally_defined_types
 */
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_7, __pyx_n_s_externally_defined_types); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_type, __pyx_n_s_MappingProxyType) < 0) __PYX_ERR(1, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":189
 * __dataclass_fields__['externally_defined_types'].name = 'externally_defined_types'
 * __dataclass_fields__['externally_defined_types'].type = PLACEHOLDER_externally_defined_types
 * __dataclass_fields__['externally_defined_types']._field_type = PLACEHOLDER_FIELD_TYPE_externally_defined_types             # <<<<<<<<<<<<<<
 */
  __pyx_t_5 = __Pyx_Load_dataclasses_Module(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_FIELD); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_dataclass_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_s_externally_defined_types); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_field_type, __pyx_t_7) < 0) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":1
 * def __init__(self, *, fingerprint_returns_digest: bint = DATACLASS_PLACEHOLDER_1, canonical_form_repeat_fixed: bint = DATACLASS_PLACEHOLDER_2, canonical_form_repeat_enum: bint = DATACLASS_PLACEHOLDER_3, record_can_encode_dict: bint = DATACLASS_PLACEHOLDER_4, record_values_type_hint: bint = DATACLASS_PLACEHOLDER_5, record_decodes_to_dict: bint = DATACLASS_PLACEHOLDER_6, record_allow_extra_fields: bint = DATACLASS_PLACEHOLDER_7, record_encode_use_defaults: bint = DATACLASS_PLACEHOLDER_8, allow_tuple_notation: bint = DATACLASS_PLACEHOLDER_9, union_decodes_to: UnionDecodeOption = DATACLASS_PLACEHOLDER_10, union_json_encodes_type_name: bint = DATACLASS_PLACEHOLDER_11, allow_primitive_name_collision: bint = DATACLASS_PLACEHOLDER_12, allow_primitive_names_in_namespaces: bint = DATACLASS_PLACEHOLDER_13, named_type_names_must_be_unique: bint = DATACLASS_PLACEHOLDER_14, enum_symbols_must_be_unique: bint = DATACLASS_PLACEHOLDER_15, enforce_enum_symbol_name_rules: bint = DATACLASS_PLACEHOLDER_16, enforce_type_name_rules: bint = DATACLASS_PLACEHOLDER_17, enforce_namespace_name_rules: bint = DATACLASS_PLACEHOLDER_18, record_fields_must_be_unique: bint = DATACLASS_PLACEHOLDER_19, ascii_name_rules: bint = DATACLASS_PLACEHOLDER_20, missing_values_can_be_null: bint = DATACLASS_PLACEHOLDER_21, missing_values_can_be_empty_container: bint = DATACLASS_PLACEHOLDER_22, allow_false_values_for_null: bint = DATACLASS_PLACEHOLDER_23, allow_empty_unions: bint = DATACLASS_PLACEHOLDER_24, allow_nested_unions: bint = DATACLASS_PLACEHOLDER_25, allow_duplicate_union_types: bint = DATACLASS_PLACEHOLDER_26, allow_union_default_any_member: bint = DATACLASS_PLACEHOLDER_27, allow_aliases_to_be_string: bint = DATACLASS_PLACEHOLDER_28, coerce_values_to_int: bint = DATACLASS_PLACEHOLDER_29, coerce_values_to_float: bint = DATACLASS_PLACEHOLDER_30, coerce_int_to_float: bint = DATACLASS_PLACEHOLDER_31, coerce_values_to_boolean: bint = DATACLASS_PLACEHOLDER_32, coerce_values_to_str: bint = DATACLASS_PLACEHOLDER_33, bytes_codec: str = DATACLASS_PLACEHOLDER_34, fixed_codec: str = DATACLASS_PLACEHOLDER_35, null_pad_fixed: bint = DATACLASS_PLACEHOLDER_36, truncate_fixed: bint = DATACLASS_PLACEHOLDER_37, clamp_int_overflow: bint = DATACLASS_PLACEHOLDER_38, clamp_float_overflow: bint = DATACLASS_PLACEHOLDER_39, float_out_of_range_inf: bint = DATACLASS_PLACEHOLDER_40, bytes_default_value_utf8: bint = DATACLASS_PLACEHOLDER_41, string_types_default_unchanged: bint = DATACLASS_PLACEHOLDER_42, decimal_check_exp_overflow: bint = DATACLASS_PLACEHOLDER_43, types_str_to_schema: bint = DATACLASS_PLACEHOLDER_44, logical_types: tuple[LogicalType] = DATACLASS_PLACEHOLDER_45, adapt_record_types: bint = DATACLASS_PLACEHOLDER_46, return_uuid_object: bint = DATACLASS_PLACEHOLDER_47, allow_error_type: bint = DATACLASS_PLACEHOLDER_48, allow_leading_dot_in_names: bint = DATACLASS_PLACEHOLDER_49, naive_dt_assume_utc: bint = DATACLASS_PLACEHOLDER_50, alternate_timestamp_millis_encoding: bint = DATACLASS_PLACEHOLDER_51, date_type_accepts_string: bint = DATACLASS_PLACEHOLDER_52, raise_on_invalid_logical: bint = DATACLASS_PLACEHOLDER_53, inline_namespaces: bint = DATACLASS_PLACEHOLDER_54, expand_types_in_schema: bint = DATACLASS_PLACEHOLDER_55, unicode_errors: str = DATACLASS_PLACEHOLDER_56, container_fill_blocks: bint = DATACLASS_PLACEHOLDER_57, defer_schema_promotion_errors: bint = DATACLASS_PLACEHOLDER_58, invalid_value_includes_record_name: bint = DATACLASS_PLACEHOLDER_59, invalid_value_include_array_index: bint = DATACLASS_PLACEHOLDER_60, invalid_value_include_map_key: bint = DATACLASS_PLACEHOLDER_61, allow_invalid_default_values: bint = DATACLASS_PLACEHOLDER_62, externally_defined_types: MappingProxyType = DATACLASS_PLACEHOLDER_63):             # <<<<<<<<<<<<<<
 *     self.fingerprint_returns_digest = fingerprint_returns_digest
 *     self.canonical_form_repeat_fixed = canonical_form_repeat_fixed
 */
  __Pyx_INCREF(__pyx_dataclass_dflt5cavro_Options5cavro_union_decodes_to);
  __pyx_k__15 = __pyx_dataclass_dflt5cavro_Options5cavro_union_decodes_to;
  __Pyx_GIVEREF(__pyx_dataclass_dflt5cavro_Options5cavro_union_decodes_to);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_LOGICAL_TYPES); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_k__16 = __pyx_t_9;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_EMPTY_MP); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_k__17 = __pyx_t_9;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":65
 *     self.allow_invalid_default_values = allow_invalid_default_values
 *     self.externally_defined_types = externally_defined_types
 * __pyx_recursive_repr_guard = __import__('threading').local()             # <<<<<<<<<<<<<<
 * __pyx_recursive_repr_guard.running = set()
 * def __repr__(self):
 */
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_tuple__166, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_local); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  __pyx_t_11 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_11 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
    __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_11, 0+__pyx_t_11);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_pyx_recursive_repr_guard, __pyx_t_9) < 0) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Options);

  /* "(tree fragment)":66
 *     self.externally_defined_types = externally_defined_types
 * __pyx_recursive_repr_guard = __import__('threading').local()
 * __pyx_recursive_repr_guard.running = set()             # <<<<<<<<<<<<<<
 * def __repr__(self):
 *     key = id(self)
 */
  __pyx_t_9 = PySet_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_pyx_recursive_repr_guard); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_running, __pyx_t_9) < 0) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7Options_19__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Options___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__168)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Options);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Options, (type(self), 0xc466972, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Options__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7Options_21__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Options___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__170)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Options, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Options);

  /* "src/option.pxi":355
 *         return out
 * 
 * _OptionsWrapper._set_from_dataclass(Options)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_OptionsWrapper); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_set_from_dataclass); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_9, ((PyObject *)__pyx_ptype_5cavro_Options)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/option.pxi":358
 * 
 * 
 * DEFAULT_OPTIONS = Options()             # <<<<<<<<<<<<<<
 * PERMISSIVE_OPTIONS = Options(
 *     allow_primitive_name_collision=True,
 */
  __pyx_t_5 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro_Options)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DEFAULT_OPTIONS, __pyx_t_5) < 0) __PYX_ERR(4, 358, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/option.pxi":360
 * DEFAULT_OPTIONS = Options()
 * PERMISSIVE_OPTIONS = Options(
 *     allow_primitive_name_collision=True,             # <<<<<<<<<<<<<<
 *     enum_symbols_must_be_unique=False,
 *     record_fields_must_be_unique=False,
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(30); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_primitive_name_collision, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":361
 * PERMISSIVE_OPTIONS = Options(
 *     allow_primitive_name_collision=True,
 *     enum_symbols_must_be_unique=False,             # <<<<<<<<<<<<<<
 *     record_fields_must_be_unique=False,
 *     enforce_enum_symbol_name_rules=False,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_enum_symbols_must_be_unique, Py_False) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":362
 *     allow_primitive_name_collision=True,
 *     enum_symbols_must_be_unique=False,
 *     record_fields_must_be_unique=False,             # <<<<<<<<<<<<<<
 *     enforce_enum_symbol_name_rules=False,
 *     allow_false_values_for_null=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_record_fields_must_be_unique, Py_False) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":363
 *     enum_symbols_must_be_unique=False,
 *     record_fields_must_be_unique=False,
 *     enforce_enum_symbol_name_rules=False,             # <<<<<<<<<<<<<<
 *     allow_false_values_for_null=True,
 *     allow_empty_unions=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_enforce_enum_symbol_name_rules, Py_False) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":364
 *     record_fields_must_be_unique=False,
 *     enforce_enum_symbol_name_rules=False,
 *     allow_false_values_for_null=True,             # <<<<<<<<<<<<<<
 *     allow_empty_unions=True,
 *     allow_nested_unions=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_false_values_for_null, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":365
 *     enforce_enum_symbol_name_rules=False,
 *     allow_false_values_for_null=True,
 *     allow_empty_unions=True,             # <<<<<<<<<<<<<<
 *     allow_nested_unions=True,
 *     allow_duplicate_union_types=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_empty_unions, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":366
 *     allow_false_values_for_null=True,
 *     allow_empty_unions=True,
 *     allow_nested_unions=True,             # <<<<<<<<<<<<<<
 *     allow_duplicate_union_types=True,
 *     coerce_values_to_boolean=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_nested_unions, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":367
 *     allow_empty_unions=True,
 *     allow_nested_unions=True,
 *     allow_duplicate_union_types=True,             # <<<<<<<<<<<<<<
 *     coerce_values_to_boolean=True,
 *     coerce_values_to_str=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_duplicate_union_types, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":368
 *     allow_nested_unions=True,
 *     allow_duplicate_union_types=True,
 *     coerce_values_to_boolean=True,             # <<<<<<<<<<<<<<
 *     coerce_values_to_str=True,
 *     coerce_values_to_int=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_coerce_values_to_boolean, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":369
 *     allow_duplicate_union_types=True,
 *     coerce_values_to_boolean=True,
 *     coerce_values_to_str=True,             # <<<<<<<<<<<<<<
 *     coerce_values_to_int=True,
 *     bytes_codec='utf8',
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_coerce_values_to_str, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":370
 *     coerce_values_to_boolean=True,
 *     coerce_values_to_str=True,
 *     coerce_values_to_int=True,             # <<<<<<<<<<<<<<
 *     bytes_codec='utf8',
 *     fixed_codec='utf8',
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_coerce_values_to_int, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_bytes_codec, __pyx_n_u_utf8) < 0) __PYX_ERR(4, 360, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_fixed_codec, __pyx_n_u_utf8) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":373
 *     bytes_codec='utf8',
 *     fixed_codec='utf8',
 *     null_pad_fixed=True,             # <<<<<<<<<<<<<<
 *     truncate_fixed=True,
 *     clamp_int_overflow=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_null_pad_fixed, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":374
 *     fixed_codec='utf8',
 *     null_pad_fixed=True,
 *     truncate_fixed=True,             # <<<<<<<<<<<<<<
 *     clamp_int_overflow=True,
 *     clamp_float_overflow=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_truncate_fixed, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":375
 *     null_pad_fixed=True,
 *     truncate_fixed=True,
 *     clamp_int_overflow=True,             # <<<<<<<<<<<<<<
 *     clamp_float_overflow=True,
 *     float_out_of_range_inf=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_clamp_int_overflow, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":376
 *     truncate_fixed=True,
 *     clamp_int_overflow=True,
 *     clamp_float_overflow=True,             # <<<<<<<<<<<<<<
 *     float_out_of_range_inf=True,
 *     coerce_values_to_float=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_clamp_float_overflow, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":377
 *     clamp_int_overflow=True,
 *     clamp_float_overflow=True,
 *     float_out_of_range_inf=True,             # <<<<<<<<<<<<<<
 *     coerce_values_to_float=True,
 *     adapt_record_types=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_float_out_of_range_inf, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":378
 *     clamp_float_overflow=True,
 *     float_out_of_range_inf=True,
 *     coerce_values_to_float=True,             # <<<<<<<<<<<<<<
 *     adapt_record_types=True,
 *     enforce_type_name_rules=False,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_coerce_values_to_float, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":379
 *     float_out_of_range_inf=True,
 *     coerce_values_to_float=True,
 *     adapt_record_types=True,             # <<<<<<<<<<<<<<
 *     enforce_type_name_rules=False,
 *     allow_error_type=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_adapt_record_types, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":380
 *     coerce_values_to_float=True,
 *     adapt_record_types=True,
 *     enforce_type_name_rules=False,             # <<<<<<<<<<<<<<
 *     allow_error_type=True,
 *     allow_leading_dot_in_names=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_enforce_type_name_rules, Py_False) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":381
 *     adapt_record_types=True,
 *     enforce_type_name_rules=False,
 *     allow_error_type=True,             # <<<<<<<<<<<<<<
 *     allow_leading_dot_in_names=True,
 *     enforce_namespace_name_rules=False,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_error_type, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":382
 *     enforce_type_name_rules=False,
 *     allow_error_type=True,
 *     allow_leading_dot_in_names=True,             # <<<<<<<<<<<<<<
 *     enforce_namespace_name_rules=False,
 *     named_type_names_must_be_unique=False,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_leading_dot_in_names, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":383
 *     allow_error_type=True,
 *     allow_leading_dot_in_names=True,
 *     enforce_namespace_name_rules=False,             # <<<<<<<<<<<<<<
 *     named_type_names_must_be_unique=False,
 *     missing_values_can_be_null=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_enforce_namespace_name_rules, Py_False) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":384
 *     allow_leading_dot_in_names=True,
 *     enforce_namespace_name_rules=False,
 *     named_type_names_must_be_unique=False,             # <<<<<<<<<<<<<<
 *     missing_values_can_be_null=True,
 *     missing_values_can_be_empty_container=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_named_type_names_must_be_unique, Py_False) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":385
 *     enforce_namespace_name_rules=False,
 *     named_type_names_must_be_unique=False,
 *     missing_values_can_be_null=True,             # <<<<<<<<<<<<<<
 *     missing_values_can_be_empty_container=True,
 *     allow_tuple_notation=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_missing_values_can_be_null, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":386
 *     named_type_names_must_be_unique=False,
 *     missing_values_can_be_null=True,
 *     missing_values_can_be_empty_container=True,             # <<<<<<<<<<<<<<
 *     allow_tuple_notation=True,
 *     allow_union_default_any_member=True,
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_missing_values_can_be_empty_cont, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":387
 *     missing_values_can_be_null=True,
 *     missing_values_can_be_empty_container=True,
 *     allow_tuple_notation=True,             # <<<<<<<<<<<<<<
 *     allow_union_default_any_member=True,
 *     unicode_errors='replace',
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_tuple_notation, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":388
 *     missing_values_can_be_empty_container=True,
 *     allow_tuple_notation=True,
 *     allow_union_default_any_member=True,             # <<<<<<<<<<<<<<
 *     unicode_errors='replace',
 * )
 */
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_allow_union_default_any_member, Py_True) < 0) __PYX_ERR(4, 360, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_unicode_errors, __pyx_n_u_replace) < 0) __PYX_ERR(4, 360, __pyx_L1_error)

  /* "src/option.pxi":359
 * 
 * DEFAULT_OPTIONS = Options()
 * PERMISSIVE_OPTIONS = Options(             # <<<<<<<<<<<<<<
 *     allow_primitive_name_collision=True,
 *     enum_symbols_must_be_unique=False,
 */
  __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_Options), __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PERMISSIVE_OPTIONS, __pyx_t_9) < 0) __PYX_ERR(4, 359, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "src/io.pxi":1
 * import io             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_9 = __Pyx_ImportDottedModule(__pyx_n_s_io, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_io, __pyx_t_9) < 0) __PYX_ERR(5, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7_Writer_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Writer___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__171)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__Writer, __pyx_n_s_reduce_cython, __pyx_t_9) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro__Writer);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Writer, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Writer__set_state(self, __pyx_state)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7_Writer_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Writer___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__172)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__Writer, __pyx_n_s_setstate_cython, __pyx_t_9) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro__Writer);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7_Reader_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Reader___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__173)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__Reader, __pyx_n_s_reduce_cython, __pyx_t_9) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro__Reader);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Reader, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Reader__set_state(self, __pyx_state)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7_Reader_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Reader___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__174)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__Reader, __pyx_n_s_setstate_cython, __pyx_t_9) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro__Reader);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10FileReader_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FileReader___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__175)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FileReader, __pyx_n_s_reduce_cython, __pyx_t_9) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FileReader);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FileReader, (type(self), 0x74c6201, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FileReader__set_state(self, __pyx_state)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10FileReader_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FileReader___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__176)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FileReader, __pyx_n_s_setstate_cython, __pyx_t_9) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FileReader);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10FileWriter_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FileWriter___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__177)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FileWriter, __pyx_n_s_reduce_cython, __pyx_t_9) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FileWriter);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FileWriter, (type(self), 0x74c6201, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FileWriter__set_state(self, __pyx_state)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10FileWriter_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FileWriter___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__178)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FileWriter, __pyx_n_s_setstate_cython, __pyx_t_9) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FileWriter);

  /* "src/buffer.pxi":3
 * 
 * 
 * cdef array.array byte_buffer_template = array.array('B', [])             # <<<<<<<<<<<<<<
 * 
 * @cython.no_gc_clear
 */
  __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(16, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_B);
  __Pyx_GIVEREF(__pyx_n_u_B);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_B)) __PYX_ERR(16, 3, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_9)) __PYX_ERR(16, 3, __pyx_L1_error);
  __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7cpython_5array_array), __pyx_t_5, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(16, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_XGOTREF((PyObject *)__pyx_v_5cavro_byte_buffer_template);
  __Pyx_DECREF_SET(__pyx_v_5cavro_byte_buffer_template, ((arrayobject *)__pyx_t_9));
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_12MemoryWriter_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MemoryWriter___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__179)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_9) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_12MemoryWriter_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MemoryWriter___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__180)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_9) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "src/buffer.pxi":48
 * 
 * 
 * cdef bytes empty_buffer = b"\x00"             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_kp_b__72);
  __Pyx_XGOTREF(__pyx_v_5cavro_empty_buffer);
  __Pyx_DECREF_SET(__pyx_v_5cavro_empty_buffer, __pyx_kp_b__72);
  __Pyx_GIVEREF(__pyx_kp_b__72);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_12MemoryReader_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MemoryReader___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__181)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_MemoryReader, __pyx_n_s_reduce_cython, __pyx_t_9) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_MemoryReader);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MemoryReader, (type(self), 0xec0f6ac, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MemoryReader__set_state(self, __pyx_state)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_12MemoryReader_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MemoryReader___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__182)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_MemoryReader, __pyx_n_s_setstate_cython, __pyx_t_9) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_MemoryReader);

  /* "src/rabin.pxi":4
 * from libc.stdint cimport *
 * 
 * cdef int64_t RABIN_EMPTY = 0xc15d213aa4d7a795L             # <<<<<<<<<<<<<<
 * 
 * cdef int64_t RABIN_TABLE[256]
 */
  __pyx_v_5cavro_RABIN_EMPTY = 0xc15d213aa4d7a795L;

  /* "src/rabin.pxi":7
 * 
 * cdef int64_t RABIN_TABLE[256]
 * cdef bint RABIN_TABLE_CONFIGURED = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_5cavro_RABIN_TABLE_CONFIGURED = 0;

  /* "src/rabin.pxi":33
 *     """
 * 
 *     name = 'rabin'             # <<<<<<<<<<<<<<
 *     digest_size = 8
 *     block_size = 8
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Rabin, __pyx_n_s_name, __pyx_n_u_rabin) < 0) __PYX_ERR(18, 33, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_Rabin);

  /* "src/rabin.pxi":34
 * 
 *     name = 'rabin'
 *     digest_size = 8             # <<<<<<<<<<<<<<
 *     block_size = 8
 * 
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Rabin, __pyx_n_s_digest_size, __pyx_int_8) < 0) __PYX_ERR(18, 34, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_Rabin);

  /* "src/rabin.pxi":35
 *     name = 'rabin'
 *     digest_size = 8
 *     block_size = 8             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly int64_t value
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Rabin, __pyx_n_s_block_size, __pyx_int_8) < 0) __PYX_ERR(18, 35, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_Rabin);

  /* "src/rabin.pxi":39
 *     cdef readonly int64_t value
 * 
 *     def __init__(self, value=RABIN_EMPTY):             # <<<<<<<<<<<<<<
 *         if not RABIN_TABLE_CONFIGURED:
 *             init_rabin_table()
 */
  __pyx_t_9 = __Pyx_PyInt_From_int64_t(__pyx_v_5cavro_RABIN_EMPTY); if (unlikely(!__pyx_t_9)) __PYX_ERR(18, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_k__21 = __pyx_t_9;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;

  /* "src/rabin.pxi":44
 *         self.value = value
 * 
 *     cpdef update(self, bytes data):             # <<<<<<<<<<<<<<
 *         cdef int64_t swizzle
 *         cdef int64_t shifted
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_5Rabin_3update, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Rabin_update, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__184)); if (unlikely(!__pyx_t_9)) __PYX_ERR(18, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Rabin, __pyx_n_s_update, __pyx_t_9) < 0) __PYX_ERR(18, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Rabin);

  /* "src/rabin.pxi":52
 *             self.value = shifted ^ swizzle
 * 
 *     def digest(self):             # <<<<<<<<<<<<<<
 *         cdef uint64_t value = self.value
 *         return int(value).to_bytes(8, 'little', signed=False)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_5Rabin_5digest, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Rabin_digest, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__186)); if (unlikely(!__pyx_t_9)) __PYX_ERR(18, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Rabin, __pyx_n_s_digest, __pyx_t_9) < 0) __PYX_ERR(18, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Rabin);

  /* "src/rabin.pxi":56
 *         return int(value).to_bytes(8, 'little', signed=False)
 * 
 *     def hexdigest(self):             # <<<<<<<<<<<<<<
 *         return self.digest().hex()
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_5Rabin_7hexdigest, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Rabin_hexdigest, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__187)); if (unlikely(!__pyx_t_9)) __PYX_ERR(18, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Rabin, __pyx_n_s_hexdigest, __pyx_t_9) < 0) __PYX_ERR(18, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Rabin);

  /* "src/rabin.pxi":59
 *         return self.digest().hex()
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         new_ver = Rabin()
 *         new_ver.value = self.value
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_5Rabin_9copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Rabin_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__189)); if (unlikely(!__pyx_t_9)) __PYX_ERR(18, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Rabin, __pyx_n_s_copy, __pyx_t_9) < 0) __PYX_ERR(18, 59, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Rabin);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_5Rabin_11__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Rabin___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__190)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Rabin, __pyx_n_s_reduce_cython, __pyx_t_9) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Rabin);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Rabin, (type(self), 0xcd42404, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Rabin__set_state(self, __pyx_state)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_5Rabin_13__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Rabin___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__191)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Rabin, __pyx_n_s_setstate_cython, __pyx_t_9) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Rabin);

  /* "src/tests/test_zigzag.pxi":2
 * 
 * cpdef str _ubin(val, n_bytes):             # <<<<<<<<<<<<<<
 *     cdef str raw = bin(val)[2:]
 *     padded = raw.rjust(n_bytes * 8, '0')
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7_ubin, 0, __pyx_n_s_ubin, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__193)); if (unlikely(!__pyx_t_9)) __PYX_ERR(6, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ubin, __pyx_t_9) < 0) __PYX_ERR(6, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "src/tests/test_zigzag.pxi":30
 *     return bytes(out)
 * 
 * @_tests             # <<<<<<<<<<<<<<
 * def _tests(add):
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_tests_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(6, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9_tests, 0, __pyx_n_s_tests_2, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__195)); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_tests_2, __pyx_t_7) < 0) __PYX_ERR(6, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/tests/test_buffer.pxi":2
 * 
 * @_tests             # <<<<<<<<<<<<<<
 * def _tests(add):
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_tests_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(19, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_11_tests, 0, __pyx_n_s_tests_2, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__197)); if (unlikely(!__pyx_t_5)) __PYX_ERR(19, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(19, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_tests_2, __pyx_t_9) < 0) __PYX_ERR(19, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "src/tests/test_perf.pxi":2
 * 
 * @_perf             # <<<<<<<<<<<<<<
 * def _perf(add):
 *     def _test_zigzag_perf(const uint64_t mask):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_perf); if (unlikely(!__pyx_t_9)) __PYX_ERR(7, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_13_perf, 0, __pyx_n_s_perf, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__199)); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(7, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perf, __pyx_t_7) < 0) __PYX_ERR(7, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/array.pxi":1
 * import numpy             # <<<<<<<<<<<<<<
 * import collections
 * 
 */
  __pyx_t_7 = __Pyx_ImportDottedModule(__pyx_n_s_numpy, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(20, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numpy, __pyx_t_7) < 0) __PYX_ERR(20, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/array.pxi":2
 * import numpy
 * import collections             # <<<<<<<<<<<<<<
 * 
 * @cython.final
 */
  __pyx_t_7 = __Pyx_ImportDottedModule(__pyx_n_s_collections, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(20, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_collections, __pyx_t_7) < 0) __PYX_ERR(20, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/array.pxi":7
 * cdef class ArrayType(AvroType):
 *     """The avro array type."""
 *     type_name = "array"             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly AvroType item_type
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ArrayType, __pyx_n_s_type_name, __pyx_n_u_array) < 0) __PYX_ERR(20, 7, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_ArrayType);

  /* "src/array.pxi":15
 *         self.item_type = AvroType.for_source(schema, source['items'], namespace)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef ArrayType new_inst = self.clone_base()
 *         new_inst.item_type = self.item_type
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9ArrayType_3copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ArrayType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__200)); if (unlikely(!__pyx_t_7)) __PYX_ERR(20, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ArrayType, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(20, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ArrayType);

  /* "src/array.pxi":23
 *         return _strip_keys(dict(source), {'type', 'items'})
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9ArrayType_5walk_types, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ArrayType_walk_types, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_7)) __PYX_ERR(20, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ArrayType, __pyx_n_s_walk_types, __pyx_t_7) < 0) __PYX_ERR(20, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ArrayType);

  /* "src/array.pxi":29
 *         yield from self.item_type.walk_types(visited)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {'items': self.item_type.get_schema(created)}
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9ArrayType_8_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ArrayType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__203)); if (unlikely(!__pyx_t_7)) __PYX_ERR(20, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ArrayType, __pyx_n_s_get_schema_extra, __pyx_t_7) < 0) __PYX_ERR(20, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ArrayType);

  /* "src/array.pxi":92
 *         return out
 * 
 *     cpdef object convert_value(self, object value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef int item_fitness
 *         cdef AvroType item_type = self.item_type
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9ArrayType_10convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ArrayType_convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__205)); if (unlikely(!__pyx_t_7)) __PYX_ERR(20, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__206);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ArrayType, __pyx_n_s_convert_value_2, __pyx_t_7) < 0) __PYX_ERR(20, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ArrayType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9ArrayType_12__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ArrayType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__207)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ArrayType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ArrayType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ArrayType, (type(self), 0xcdf32d8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ArrayType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9ArrayType_14__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ArrayType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__208)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ArrayType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ArrayType);

  /* "src/enum.pxi":4
 * cdef class EnumType(_NamedType):
 *     """The avro enum type"""
 *     type_name = 'enum'             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly tuple symbols
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_EnumType, __pyx_n_s_type_name, __pyx_n_u_enum) < 0) __PYX_ERR(8, 4, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_EnumType);

  /* "src/enum.pxi":47
 *         self.doc = source.get('doc', '')
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef EnumType new_inst = self.clone_base()
 *         new_inst.symbols = self.symbols
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8EnumType_3copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EnumType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__209)); if (unlikely(!__pyx_t_7)) __PYX_ERR(8, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_EnumType, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(8, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_EnumType);

  /* "src/enum.pxi":65
 *         })
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         extra = _NamedType._get_schema_extra(self, created)
 *         if self.default_value is not NO_DEFAULT:
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8EnumType_5_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EnumType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__210)); if (unlikely(!__pyx_t_7)) __PYX_ERR(8, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_EnumType, __pyx_n_s_get_schema_extra, __pyx_t_7) < 0) __PYX_ERR(8, 65, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_EnumType);

  /* "src/enum.pxi":86
 *             return FIT_NONE
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         # if value is valid for type, then it needs no conversion
 *         return value
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8EnumType_7_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EnumType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__211)); if (unlikely(!__pyx_t_7)) __PYX_ERR(8, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_EnumType, __pyx_n_s_convert_value, __pyx_t_7) < 0) __PYX_ERR(8, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_EnumType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8EnumType_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EnumType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__212)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_EnumType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_EnumType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_EnumType, (type(self), 0x8f63b31, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_EnumType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8EnumType_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_EnumType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__213)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_EnumType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_EnumType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_17PromotingEnumType_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PromotingEnumType___reduce_cytho, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__214)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_PromotingEnumType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_PromotingEnumType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromotingEnumType, (type(self), 0x8a61ff8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromotingEnumType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_17PromotingEnumType_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PromotingEnumType___setstate_cyt, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__215)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_PromotingEnumType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_PromotingEnumType);

  /* "src/map.pxi":4
 * cdef class MapType(AvroType):
 *     """The avro map type."""
 *     type_name = "map"             # <<<<<<<<<<<<<<
 * 
 *     cdef StringType key_type
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_MapType, __pyx_n_s_type_name, __pyx_n_u_map) < 0) __PYX_ERR(9, 4, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_MapType);

  /* "src/map.pxi":14
 *         self.value_type = AvroType.for_source(schema, source['values'], namespace)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef MapType new_inst = self.clone_base()
 *         new_inst.key_type = self.key_type
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7MapType_3copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MapType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__216)); if (unlikely(!__pyx_t_7)) __PYX_ERR(9, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_MapType, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(9, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_MapType);

  /* "src/map.pxi":23
 *         return _strip_keys(dict(source), {'type', 'values'})
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7MapType_5walk_types, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MapType_walk_types, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_7)) __PYX_ERR(9, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_MapType, __pyx_n_s_walk_types, __pyx_t_7) < 0) __PYX_ERR(9, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_MapType);

  /* "src/map.pxi":29
 *         yield from self.value_type.walk_types(visited)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {'values': self.value_type.get_schema(created)}
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7MapType_8_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MapType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__217)); if (unlikely(!__pyx_t_7)) __PYX_ERR(9, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_MapType, __pyx_n_s_get_schema_extra, __pyx_t_7) < 0) __PYX_ERR(9, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_MapType);

  /* "src/map.pxi":122
 *         return out
 * 
 *     cpdef object convert_value(self, object orig_value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef int key_fitness
 *         cdef int value_fitness
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7MapType_10convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MapType_convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__219)); if (unlikely(!__pyx_t_7)) __PYX_ERR(9, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__220);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_MapType, __pyx_n_s_convert_value_2, __pyx_t_7) < 0) __PYX_ERR(9, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_MapType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7MapType_12__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MapType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__221)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_MapType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_MapType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MapType, (type(self), 0xaa8230c, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MapType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7MapType_14__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MapType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__222)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_MapType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_MapType);

  /* "src/null.pxi":4
 * cdef class NullType(AvroType):
 *     """The avro null type."""
 *     type_name = "null"             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_NullType, __pyx_n_s_type_name, __pyx_n_u_null_2) < 0) __PYX_ERR(21, 4, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_NullType);

  /* "src/null.pxi":6
 *     type_name = "null"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8NullType_1copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NullType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__223)); if (unlikely(!__pyx_t_7)) __PYX_ERR(21, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_NullType, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(21, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_NullType);

  /* "src/null.pxi":12
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8NullType_3_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NullType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__224)); if (unlikely(!__pyx_t_7)) __PYX_ERR(21, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_NullType, __pyx_n_s_get_schema_extra, __pyx_t_7) < 0) __PYX_ERR(21, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_NullType);

  /* "src/null.pxi":40
 *         return None
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8NullType_5_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NullType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__225)); if (unlikely(!__pyx_t_7)) __PYX_ERR(21, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_NullType, __pyx_n_s_convert_value, __pyx_t_7) < 0) __PYX_ERR(21, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_NullType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8NullType_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NullType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__226)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_NullType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_NullType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NullType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NullType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8NullType_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NullType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__227)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_NullType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_NullType);

  /* "src/numeric_types.pxi":3
 * from libc.float cimport FLT_MAX, DBL_MAX
 * from cpython cimport bool as py_bool
 * from numpy import bool_, integer, float16 as np_f16, float32 as np_f32, float64 as np_f64             # <<<<<<<<<<<<<<
 * import numpy as np
 * 
 */
  __pyx_t_7 = PyList_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_bool_2);
  __Pyx_GIVEREF(__pyx_n_s_bool_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_bool_2)) __PYX_ERR(10, 3, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_integer);
  __Pyx_GIVEREF(__pyx_n_s_integer);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 1, __pyx_n_s_integer)) __PYX_ERR(10, 3, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_float16);
  __Pyx_GIVEREF(__pyx_n_s_float16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 2, __pyx_n_s_float16)) __PYX_ERR(10, 3, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_float32);
  __Pyx_GIVEREF(__pyx_n_s_float32);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 3, __pyx_n_s_float32)) __PYX_ERR(10, 3, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_float64);
  __Pyx_GIVEREF(__pyx_n_s_float64);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 4, __pyx_n_s_float64)) __PYX_ERR(10, 3, __pyx_L1_error);
  __pyx_t_5 = __Pyx_Import(__pyx_n_s_numpy, __pyx_t_7, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_bool_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bool_2, __pyx_t_7) < 0) __PYX_ERR(10, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_integer); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_integer, __pyx_t_7) < 0) __PYX_ERR(10, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_float16); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np_f16, __pyx_t_7) < 0) __PYX_ERR(10, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_float32); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np_f32, __pyx_t_7) < 0) __PYX_ERR(10, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_float64); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np_f64, __pyx_t_7) < 0) __PYX_ERR(10, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/numeric_types.pxi":4
 * from cpython cimport bool as py_bool
 * from numpy import bool_, integer, float16 as np_f16, float32 as np_f32, float64 as np_f64
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * from math import isnan, isinf
 */
  __pyx_t_5 = __Pyx_ImportDottedModule(__pyx_n_s_numpy, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_5) < 0) __PYX_ERR(10, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/numeric_types.pxi":6
 * import numpy as np
 * 
 * from math import isnan, isinf             # <<<<<<<<<<<<<<
 * 
 * cdef float FLOAT_INT_THRESHOLD = 0.001
 */
  __pyx_t_5 = PyList_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_isnan);
  __Pyx_GIVEREF(__pyx_n_s_isnan);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_isnan)) __PYX_ERR(10, 6, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_isinf);
  __Pyx_GIVEREF(__pyx_n_s_isinf);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 1, __pyx_n_s_isinf)) __PYX_ERR(10, 6, __pyx_L1_error);
  __pyx_t_7 = __Pyx_Import(__pyx_n_s_math, __pyx_t_5, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_7, __pyx_n_s_isnan); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isnan, __pyx_t_5) < 0) __PYX_ERR(10, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_7, __pyx_n_s_isinf); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isinf, __pyx_t_5) < 0) __PYX_ERR(10, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/numeric_types.pxi":8
 * from math import isnan, isinf
 * 
 * cdef float FLOAT_INT_THRESHOLD = 0.001             # <<<<<<<<<<<<<<
 * 
 * @cython.final
 */
  __pyx_v_5cavro_FLOAT_INT_THRESHOLD = 0.001;

  /* "src/numeric_types.pxi":13
 * cdef class BoolType(AvroType):
 *     """The avro boolean type."""
 *     type_name = "boolean"             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_BoolType, __pyx_n_s_type_name, __pyx_n_u_boolean_2) < 0) __PYX_ERR(10, 13, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_BoolType);

  /* "src/numeric_types.pxi":15
 *     type_name = "boolean"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8BoolType_1copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BoolType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__228)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_BoolType, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(10, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_BoolType);

  /* "src/numeric_types.pxi":21
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8BoolType_3_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BoolType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__229)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_BoolType, __pyx_n_s_get_schema_extra, __pyx_t_7) < 0) __PYX_ERR(10, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_BoolType);

  /* "src/numeric_types.pxi":42
 *         return FIT_NONE
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (py_bool, bool_)):
 *             return value
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8BoolType_5_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BoolType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__230)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_BoolType, __pyx_n_s_convert_value, __pyx_t_7) < 0) __PYX_ERR(10, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_BoolType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8BoolType_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BoolType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__231)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_BoolType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_BoolType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BoolType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BoolType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8BoolType_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BoolType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__232)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_BoolType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_BoolType);

  /* "src/numeric_types.pxi":63
 * cdef class IntType(AvroType):
 *     """The avro int type."""
 *     type_name = "int"             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_IntType, __pyx_n_s_type_name, __pyx_n_u_int_2) < 0) __PYX_ERR(10, 63, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_IntType);

  /* "src/numeric_types.pxi":65
 *     type_name = "int"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7IntType_1copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__233)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_IntType, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(10, 65, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_IntType);

  /* "src/numeric_types.pxi":71
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7IntType_3_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__234)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_IntType, __pyx_n_s_get_schema_extra, __pyx_t_7) < 0) __PYX_ERR(10, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_IntType);

  /* "src/numeric_types.pxi":100
 *         return max_fit
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7IntType_5_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__235)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_IntType, __pyx_n_s_convert_value, __pyx_t_7) < 0) __PYX_ERR(10, 100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_IntType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7IntType_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__236)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_IntType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_IntType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_7IntType_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__237)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_IntType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_IntType);

  /* "src/numeric_types.pxi":135
 * cdef class LongType(AvroType):
 *     """The avro long type."""
 *     type_name = "long"             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_LongType, __pyx_n_s_type_name, __pyx_n_u_long_2) < 0) __PYX_ERR(10, 135, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_LongType);

  /* "src/numeric_types.pxi":137
 *     type_name = "long"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8LongType_1copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LongType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__238)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_LongType, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(10, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_LongType);

  /* "src/numeric_types.pxi":143
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8LongType_3_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LongType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__239)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_LongType, __pyx_n_s_get_schema_extra, __pyx_t_7) < 0) __PYX_ERR(10, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_LongType);

  /* "src/numeric_types.pxi":172
 *         return max_fit
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if not self.options.coerce_values_to_int and isinstance(value, (bool_, py_bool)):
 *             raise ValueError(f'{value} not valid for long')
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8LongType_5_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LongType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__240)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_LongType, __pyx_n_s_convert_value, __pyx_t_7) < 0) __PYX_ERR(10, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_LongType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8LongType_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LongType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__241)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_LongType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_LongType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LongType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LongType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8LongType_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LongType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__242)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_LongType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_LongType);

  /* "src/numeric_types.pxi":211
 * cdef class FloatType(AvroType):
 *     """The avro float type."""
 *     type_name = "float"             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FloatType, __pyx_n_s_type_name, __pyx_n_u_float_2) < 0) __PYX_ERR(10, 211, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_FloatType);

  /* "src/numeric_types.pxi":213
 *     type_name = "float"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9FloatType_1copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FloatType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__243)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FloatType, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(10, 213, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FloatType);

  /* "src/numeric_types.pxi":219
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9FloatType_3_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FloatType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__244)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FloatType, __pyx_n_s_get_schema_extra, __pyx_t_7) < 0) __PYX_ERR(10, 219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FloatType);

  /* "src/numeric_types.pxi":256
 *         return FIT_NONE
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, float):
 *             pass
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9FloatType_5_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FloatType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__245)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FloatType, __pyx_n_s_convert_value, __pyx_t_7) < 0) __PYX_ERR(10, 256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FloatType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9FloatType_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FloatType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__246)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FloatType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FloatType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FloatType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FloatType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9FloatType_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FloatType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__247)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FloatType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FloatType);

  /* "src/numeric_types.pxi":326
 * cdef class DoubleType(AvroType):
 *     """The avro double type."""
 *     type_name = "double"             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_DoubleType, __pyx_n_s_type_name, __pyx_n_u_double_2) < 0) __PYX_ERR(10, 326, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_DoubleType);

  /* "src/numeric_types.pxi":328
 *     type_name = "double"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10DoubleType_1copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DoubleType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__248)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_DoubleType, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(10, 328, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_DoubleType);

  /* "src/numeric_types.pxi":334
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10DoubleType_3_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DoubleType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__249)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_DoubleType, __pyx_n_s_get_schema_extra, __pyx_t_7) < 0) __PYX_ERR(10, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_DoubleType);

  /* "src/numeric_types.pxi":366
 *         return max_fit
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, float):
 *             pass
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10DoubleType_5_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DoubleType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__250)); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_DoubleType, __pyx_n_s_convert_value, __pyx_t_7) < 0) __PYX_ERR(10, 366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_DoubleType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10DoubleType_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DoubleType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__251)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_DoubleType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_DoubleType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DoubleType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DoubleType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10DoubleType_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DoubleType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__252)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_DoubleType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_DoubleType);

  /* "src/record.pxi":2
 * 
 * class Order(enum.Enum):             # <<<<<<<<<<<<<<
 *     """
 *     The order of a field in a record.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_enum); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5)) __PYX_ERR(11, 2, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PEP560_update_bases(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = __Pyx_Py3MetaclassPrepare(__pyx_t_9, __pyx_t_5, __pyx_n_s_Order, __pyx_n_s_Order, (PyObject *) NULL, __pyx_n_s_cavro, __pyx_kp_s_The_order_of_a_field_in_a_recor); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_5 != __pyx_t_7) {
    if (unlikely((PyDict_SetItemString(__pyx_t_4, "__orig_bases__", __pyx_t_7) < 0))) __PYX_ERR(11, 2, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/record.pxi":7
 *     """
 * 
 *     ASC = 'ascending'             # <<<<<<<<<<<<<<
 *     DESC = 'descending'
 *     IGNORE = 'ignore'
 */
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_ASC, __pyx_n_u_ascending) < 0) __PYX_ERR(11, 7, __pyx_L1_error)

  /* "src/record.pxi":8
 * 
 *     ASC = 'ascending'
 *     DESC = 'descending'             # <<<<<<<<<<<<<<
 *     IGNORE = 'ignore'
 * 
 */
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_DESC, __pyx_n_u_descending) < 0) __PYX_ERR(11, 8, __pyx_L1_error)

  /* "src/record.pxi":9
 *     ASC = 'ascending'
 *     DESC = 'descending'
 *     IGNORE = 'ignore'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_IGNORE, __pyx_n_u_ignore) < 0) __PYX_ERR(11, 9, __pyx_L1_error)

  /* "src/record.pxi":2
 * 
 * class Order(enum.Enum):             # <<<<<<<<<<<<<<
 *     """
 *     The order of a field in a record.
 */
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_9, __pyx_n_s_Order, __pyx_t_5, __pyx_t_4, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Order, __pyx_t_7) < 0) __PYX_ERR(11, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/record.pxi":12
 * 
 * 
 * NO_DEFAULT = _Sentinel('NO_DEFAULT')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro__Sentinel), __pyx_tuple__253, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NO_DEFAULT, __pyx_t_5) < 0) __PYX_ERR(11, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/record.pxi":22
 *     """
 * 
 *     type_name = "placeholder"             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly object default_value
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__PlaceholderType, __pyx_n_s_type_name, __pyx_n_u_placeholder) < 0) __PYX_ERR(11, 22, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro__PlaceholderType);

  /* "src/record.pxi":34
 *         return {}
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_16_PlaceholderType_3_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PlaceholderType__get_schema_ext, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__254)); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__PlaceholderType, __pyx_n_s_get_schema_extra, __pyx_t_5) < 0) __PYX_ERR(11, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro__PlaceholderType);

  /* "src/record.pxi":55
 *         return self.default_value
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_16_PlaceholderType_5_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PlaceholderType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__255)); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__PlaceholderType, __pyx_n_s_convert_value, __pyx_t_5) < 0) __PYX_ERR(11, 55, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro__PlaceholderType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_16_PlaceholderType_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PlaceholderType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__256)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__PlaceholderType, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro__PlaceholderType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__PlaceholderType, (type(self), 0x5c2152b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__PlaceholderType__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_16_PlaceholderType_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PlaceholderType___setstate_cyth, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__257)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__PlaceholderType, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro__PlaceholderType);

  /* "src/record.pxi":129
 *         self.data = _record_data_from_dict(self.Type, data_dict)
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return ['Type'] + [f.name for f in self.Type.fields] + ['_asdict', '__getitem__']
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Record_3__dir__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Record___dir, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__259)); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Record, __pyx_n_s_dir, __pyx_t_5) < 0) __PYX_ERR(11, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Record);

  /* "src/record.pxi":165
 *         return f'<Record:{self.Type.type} {self._repr_children(70)}>'
 * 
 *     def _asdict(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the record as a dict, with field names as keys.
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Record_9_asdict, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Record__asdict, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__261)); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Record, __pyx_n_s_asdict, __pyx_t_5) < 0) __PYX_ERR(11, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Record);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Record_13__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Record___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__262)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Record, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Record);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Record, (type(self), 0x3a6eb07, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Record__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Record_15__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Record___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__263)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Record, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Record);

  /* "src/record.pxi":244
 *         })
 * 
 *     def get_schema(self, created):             # <<<<<<<<<<<<<<
 *         schema = {
 *             'name': self.name,
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_11RecordField_3get_schema, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RecordField_get_schema, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__265)); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_RecordField, __pyx_n_s_get_schema, __pyx_t_5) < 0) __PYX_ERR(11, 244, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_RecordField);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_11RecordField_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RecordField___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__266)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_RecordField, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_RecordField);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RecordField, (type(self), 0x1d9a0f8, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RecordField__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_11RecordField_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RecordField___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__267)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_RecordField, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_RecordField);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_14_FieldAccessor_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FieldAccessor___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__268)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__FieldAccessor, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro__FieldAccessor);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__FieldAccessor, (type(self), 0x1bc04b5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__FieldAccessor__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_14_FieldAccessor_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FieldAccessor___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__269)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__FieldAccessor, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro__FieldAccessor);

  /* "src/record.pxi":298
 *     )
 * 
 * _EMPTY_ARRAY = array.array(SSIZE_TYPECODE.decode())             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_SSIZE_TYPECODE); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_decode); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7cpython_5array_array), __pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EMPTY_ARRAY, __pyx_t_9) < 0) __PYX_ERR(11, 298, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "src/record.pxi":312
 *     """
 * 
 *     type_name = 'record'             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly str doc
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_RecordType, __pyx_n_s_type_name, __pyx_n_u_record) < 0) __PYX_ERR(11, 312, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_RecordType);

  /* "src/record.pxi":348
 *         return AvroType._make_logical(self, schema, source)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef RecordType new_inst = self.clone_base()
 *         new_inst.doc = self.doc
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10RecordType_3copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RecordType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__270)); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_RecordType, __pyx_n_s_copy, __pyx_t_9) < 0) __PYX_ERR(11, 348, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_RecordType);

  /* "src/record.pxi":366
 *         })
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10RecordType_5walk_types, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RecordType_walk_types, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__66)); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_RecordType, __pyx_n_s_walk_types, __pyx_t_9) < 0) __PYX_ERR(11, 366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_RecordType);

  /* "src/record.pxi":373
 *             yield from field.type.walk_types(visited)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         extra = super(RecordType, self)._get_schema_extra(created)
 *         return dict(extra, fields=[f.get_schema(created) for f in self.fields])
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10RecordType_8_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RecordType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__272)); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_RecordType, __pyx_n_s_get_schema_extra, __pyx_t_9) < 0) __PYX_ERR(11, 373, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_RecordType);

  /* "src/record.pxi":528
 *         return self.json_decode_record(value)
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, self.record):
 *             return value
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10RecordType_10_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RecordType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__273)); if (unlikely(!__pyx_t_9)) __PYX_ERR(11, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_RecordType, __pyx_n_s_convert_value, __pyx_t_9) < 0) __PYX_ERR(11, 528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_RecordType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10RecordType_12__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RecordType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__274)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_RecordType, __pyx_n_s_reduce_cython, __pyx_t_9) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_RecordType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_RecordType, (type(self), 0x182acf2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_RecordType__set_state(self, __pyx_state)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10RecordType_14__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RecordType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__275)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_RecordType, __pyx_n_s_setstate_cython, __pyx_t_9) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_RecordType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_19PromotingRecordType_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PromotingRecordType___reduce_cyt, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__276)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_PromotingRecordType, __pyx_n_s_reduce_cython, __pyx_t_9) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_PromotingRecordType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromotingRecordType, (type(self), 0x40bba2b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromotingRecordType__set_state(self, __pyx_state)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_19PromotingRecordType_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PromotingRecordType___setstate_c, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__277)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_PromotingRecordType, __pyx_n_s_setstate_cython, __pyx_t_9) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_PromotingRecordType);

  /* "src/string_types.pxi":5
 * cdef class BytesType(AvroType):
 *     """The avro bytes type"""
 *     type_name = "bytes"             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_BytesType, __pyx_n_s_type_name, __pyx_n_u_bytes_3) < 0) __PYX_ERR(12, 5, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_BytesType);

  /* "src/string_types.pxi":7
 *     type_name = "bytes"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9BytesType_1copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BytesType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__278)); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_BytesType, __pyx_n_s_copy, __pyx_t_9) < 0) __PYX_ERR(12, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_BytesType);

  /* "src/string_types.pxi":13
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9BytesType_3_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BytesType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__279)); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_BytesType, __pyx_n_s_get_schema_extra, __pyx_t_9) < 0) __PYX_ERR(12, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_BytesType);

  /* "src/string_types.pxi":57
 *         return sval.encode('latin-1')
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, (bytes, bytearray)):
 *             return value
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9BytesType_5_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BytesType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__280)); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_BytesType, __pyx_n_s_convert_value, __pyx_t_9) < 0) __PYX_ERR(12, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_BytesType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9BytesType_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BytesType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__281)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_BytesType, __pyx_n_s_reduce_cython, __pyx_t_9) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_BytesType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BytesType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BytesType__set_state(self, __pyx_state)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9BytesType_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BytesType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__282)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_BytesType, __pyx_n_s_setstate_cython, __pyx_t_9) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_BytesType);

  /* "src/string_types.pxi":95
 * cdef class StringType(AvroType):
 *     """The avro string type"""
 *     type_name = "string"             # <<<<<<<<<<<<<<
 * 
 *     cpdef AvroType copy(self):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_StringType, __pyx_n_s_type_name, __pyx_n_u_string) < 0) __PYX_ERR(12, 95, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_StringType);

  /* "src/string_types.pxi":97
 *     type_name = "string"
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         return self.clone_base()
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10StringType_1copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_StringType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__283)); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_StringType, __pyx_n_s_copy, __pyx_t_9) < 0) __PYX_ERR(12, 97, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_StringType);

  /* "src/string_types.pxi":103
 *         return _strip_keys(dict(source), {'type'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return {}
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10StringType_3_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_StringType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__284)); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_StringType, __pyx_n_s_get_schema_extra, __pyx_t_9) < 0) __PYX_ERR(12, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_StringType);

  /* "src/string_types.pxi":133
 *         return sval
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if isinstance(value, str):
 *             return value
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10StringType_5_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_StringType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__285)); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_StringType, __pyx_n_s_convert_value, __pyx_t_9) < 0) __PYX_ERR(12, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_StringType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10StringType_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_StringType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__286)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_StringType, __pyx_n_s_reduce_cython, __pyx_t_9) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_StringType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_StringType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_StringType__set_state(self, __pyx_state)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10StringType_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_StringType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__287)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_StringType, __pyx_n_s_setstate_cython, __pyx_t_9) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_StringType);

  /* "src/string_types.pxi":153
 * cdef class FixedType(_NamedType):
 *     """The avro fixed type"""
 *     type_name = "fixed"             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly Py_ssize_t size
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FixedType, __pyx_n_s_type_name, __pyx_n_u_fixed) < 0) __PYX_ERR(12, 153, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_FixedType);

  /* "src/string_types.pxi":163
 *         super().__init__(schema, source, namespace)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef FixedType new_inst =  self.clone_base()
 *         new_inst.size = self.size
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9FixedType_3copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FixedType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__288)); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FixedType, __pyx_n_s_copy, __pyx_t_9) < 0) __PYX_ERR(12, 163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FixedType);

  /* "src/string_types.pxi":171
 *         return _strip_keys(dict(source), {'type', 'name', 'namespace', 'aliases', 'size'})
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         return dict(_NamedType._get_schema_extra(self, created), size=self.size)
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9FixedType_5_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FixedType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__289)); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FixedType, __pyx_n_s_get_schema_extra, __pyx_t_9) < 0) __PYX_ERR(12, 171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FixedType);

  /* "src/string_types.pxi":228
 *         return encoded
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         if not isinstance(value, (bytes, bytearray)):
 *             # Let's see if we can recover this
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9FixedType_7_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FixedType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__290)); if (unlikely(!__pyx_t_9)) __PYX_ERR(12, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FixedType, __pyx_n_s_convert_value, __pyx_t_9) < 0) __PYX_ERR(12, 228, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FixedType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9FixedType_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FixedType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__291)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FixedType, __pyx_n_s_reduce_cython, __pyx_t_9) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FixedType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FixedType, (type(self), 0xd9f3afe, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FixedType__set_state(self, __pyx_state)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9FixedType_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FixedType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__292)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_FixedType, __pyx_n_s_setstate_cython, __pyx_t_9) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  PyType_Modified(__pyx_ptype_5cavro_FixedType);

  /* "src/union.pxi":3
 * 
 * AMBIGUOUS_TYPES = [
 *     {RecordType, MapType},             # <<<<<<<<<<<<<<
 *     {EnumType, StringType},
 *     {LongType, IntType},
 */
  __pyx_t_9 = PySet_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(22, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PySet_Add(__pyx_t_9, ((PyObject *)__pyx_ptype_5cavro_RecordType)) < 0) __PYX_ERR(22, 3, __pyx_L1_error)
  if (PySet_Add(__pyx_t_9, ((PyObject *)__pyx_ptype_5cavro_MapType)) < 0) __PYX_ERR(22, 3, __pyx_L1_error)

  /* "src/union.pxi":4
 * AMBIGUOUS_TYPES = [
 *     {RecordType, MapType},
 *     {EnumType, StringType},             # <<<<<<<<<<<<<<
 *     {LongType, IntType},
 * ]
 */
  __pyx_t_5 = PySet_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(22, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PySet_Add(__pyx_t_5, ((PyObject *)__pyx_ptype_5cavro_EnumType)) < 0) __PYX_ERR(22, 4, __pyx_L1_error)
  if (PySet_Add(__pyx_t_5, ((PyObject *)__pyx_ptype_5cavro_StringType)) < 0) __PYX_ERR(22, 4, __pyx_L1_error)

  /* "src/union.pxi":5
 *     {RecordType, MapType},
 *     {EnumType, StringType},
 *     {LongType, IntType},             # <<<<<<<<<<<<<<
 * ]
 * 
 */
  __pyx_t_4 = PySet_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PySet_Add(__pyx_t_4, ((PyObject *)__pyx_ptype_5cavro_LongType)) < 0) __PYX_ERR(22, 5, __pyx_L1_error)
  if (PySet_Add(__pyx_t_4, ((PyObject *)__pyx_ptype_5cavro_IntType)) < 0) __PYX_ERR(22, 5, __pyx_L1_error)

  /* "src/union.pxi":2
 * 
 * AMBIGUOUS_TYPES = [             # <<<<<<<<<<<<<<
 *     {RecordType, MapType},
 *     {EnumType, StringType},
 */
  __pyx_t_7 = PyList_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 0, __pyx_t_9)) __PYX_ERR(22, 2, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 1, __pyx_t_5)) __PYX_ERR(22, 2, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 2, __pyx_t_4)) __PYX_ERR(22, 2, __pyx_L1_error);
  __pyx_t_9 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AMBIGUOUS_TYPES, __pyx_t_7) < 0) __PYX_ERR(22, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/union.pxi":27
 * cdef class UnionType(AvroType):
 *     """The avro union type"""
 *     type_name = "union"             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly tuple union_types
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UnionType, __pyx_n_s_type_name, __pyx_n_u_union) < 0) __PYX_ERR(22, 27, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_UnionType);

  /* "src/union.pxi":77
 *                     seen_types.add(member_type)
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         cdef UnionType new_inst = self.clone_base()
 *         new_inst.union_types = self.union_types
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9UnionType_3copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UnionType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__293)); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UnionType, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(22, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UnionType);

  /* "src/union.pxi":90
 *         return dict()
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9UnionType_5walk_types, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UnionType_walk_types, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__75)); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UnionType, __pyx_n_s_walk_types, __pyx_t_7) < 0) __PYX_ERR(22, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UnionType);

  /* "src/union.pxi":97
 *             yield from t.walk_types(visited)
 * 
 *     cpdef get_schema(self, created=None):             # <<<<<<<<<<<<<<
 *         if created is None:
 *             created = set()
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9UnionType_8get_schema, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UnionType_get_schema, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__295)); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__141);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UnionType, __pyx_n_s_get_schema, __pyx_t_7) < 0) __PYX_ERR(22, 97, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UnionType);

  /* "src/union.pxi":170
 *         return item_type.json_decode(item_value)
 * 
 *     cpdef object convert_value(self, object value, check_value=True):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t index = self.resolve_from_value(value)
 *         return self.union_types[index].convert_value(value)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9UnionType_10convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UnionType_convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__296)); if (unlikely(!__pyx_t_7)) __PYX_ERR(22, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__297);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UnionType, __pyx_n_s_convert_value_2, __pyx_t_7) < 0) __PYX_ERR(22, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UnionType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9UnionType_12__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UnionType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__298)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UnionType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UnionType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UnionType, (type(self), 0xabef18e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UnionType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9UnionType_14__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UnionType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__299)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UnionType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UnionType);

  /* "src/type.pxi":2
 * 
 * cdef int FIT_NONE = 0  # Value is never valid for type             # <<<<<<<<<<<<<<
 * cdef int FIT_POOR = 1  # Value may be valid depending on the schema options
 * cdef int FIT_OK = 2    # Value can be converted to the correct type
 */
  __pyx_v_5cavro_FIT_NONE = 0;

  /* "src/type.pxi":3
 * 
 * cdef int FIT_NONE = 0  # Value is never valid for type
 * cdef int FIT_POOR = 1  # Value may be valid depending on the schema options             # <<<<<<<<<<<<<<
 * cdef int FIT_OK = 2    # Value can be converted to the correct type
 * cdef int FIT_EXACT = 3 # Value is the exact type and needs no further conversion
 */
  __pyx_v_5cavro_FIT_POOR = 1;

  /* "src/type.pxi":4
 * cdef int FIT_NONE = 0  # Value is never valid for type
 * cdef int FIT_POOR = 1  # Value may be valid depending on the schema options
 * cdef int FIT_OK = 2    # Value can be converted to the correct type             # <<<<<<<<<<<<<<
 * cdef int FIT_EXACT = 3 # Value is the exact type and needs no further conversion
 * 
 */
  __pyx_v_5cavro_FIT_OK = 2;

  /* "src/type.pxi":5
 * cdef int FIT_POOR = 1  # Value may be valid depending on the schema options
 * cdef int FIT_OK = 2    # Value can be converted to the correct type
 * cdef int FIT_EXACT = 3 # Value is the exact type and needs no further conversion             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_5cavro_FIT_EXACT = 3;

  /* "src/type.pxi":8
 * 
 * 
 * CANONICAL_FORM_KEYS = ('name', 'type', 'fields', 'symbols', 'items', 'values', 'size')             # <<<<<<<<<<<<<<
 * 
 * cdef class _Sentinel:
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CANONICAL_FORM_KEYS, __pyx_tuple__300) < 0) __PYX_ERR(2, 8, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9_Sentinel_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Sentinel___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__301)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__Sentinel, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__Sentinel);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Sentinel, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Sentinel__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9_Sentinel_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Sentinel___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__302)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__Sentinel, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__Sentinel);

  /* "src/type.pxi":19
 *         return f'<{self.name}: {id(self)}>'
 * 
 * MISSING_VALUE = _Sentinel('MISSING_VALUE')             # <<<<<<<<<<<<<<
 * 
 * cdef class _CanonicalForm(str):
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro__Sentinel), __pyx_tuple__303, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MISSING_VALUE, __pyx_t_7) < 0) __PYX_ERR(2, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_14_CanonicalForm_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CanonicalForm___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__304)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__CanonicalForm, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__CanonicalForm);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__CanonicalForm, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__CanonicalForm__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_14_CanonicalForm_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CanonicalForm___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__305)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__CanonicalForm, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__CanonicalForm);

  /* "src/type.pxi":57
 *     """
 * 
 *     type_name = NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly Options options
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_type_name, __pyx_builtin_NotImplemented) < 0) __PYX_ERR(2, 57, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_AvroType);

  /* "src/type.pxi":63
 *     cdef readonly tuple[ValueAdapter] value_adapters
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def for_source(cls, schema, source, namespace=None):
 *         cdef Schema schema_ob = schema
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8AvroType_1for_source, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvroType_for_source, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__307)); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__141);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_for_source, __pyx_t_7) < 0) __PYX_ERR(2, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_AvroType);
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_for_source); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_Method_ClassMethod(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_for_source, __pyx_t_4) < 0) __PYX_ERR(2, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_AvroType);

  /* "src/type.pxi":84
 *         raise UnknownType(namespaced)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def for_schema(cls, schema):
 *         return AvroType.for_source(schema, schema.source)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8AvroType_3for_schema, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvroType_for_schema, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__309)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_for_schema, __pyx_t_4) < 0) __PYX_ERR(2, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_AvroType);
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_for_schema); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_Method_ClassMethod(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_for_schema, __pyx_t_7) < 0) __PYX_ERR(2, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_AvroType);

  /* "src/type.pxi":88
 *         return AvroType.for_source(schema, schema.source)
 * 
 *     def __init__(self, schema, source, namespace, value_adapters=tuple()):             # <<<<<<<<<<<<<<
 *         cdef ValueAdapter logical
 *         self.options = schema.options
 */
  __pyx_t_7 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyTuple_Type))); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_k__81 = __pyx_t_7;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "src/type.pxi":108
 *         return clone
 * 
 *     cpdef AvroType copy(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement clone")
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8AvroType_7copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvroType_copy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__310)); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(2, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_AvroType);

  /* "src/type.pxi":130
 *             f"{type(self).__name__} does not implement _extract_metadata")
 * 
 *     cpdef object _convert_value(self, object value):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _convert_value")
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8AvroType_9_convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvroType__convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__311)); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_convert_value, __pyx_t_7) < 0) __PYX_ERR(2, 130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_AvroType);

  /* "src/type.pxi":134
 *             f"{type(self).__name__} does not implement _convert_value")
 * 
 *     cpdef object convert_value(self, object value, check_value=False):             # <<<<<<<<<<<<<<
 *         if check_value:
 *             self.assert_value(value)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8AvroType_11convert_value, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvroType_convert_value, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__312)); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__313);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_convert_value_2, __pyx_t_7) < 0) __PYX_ERR(2, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_AvroType);

  /* "src/type.pxi":232
 *             f"{type(self).__name__} does not implement canonical_form")
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError(
 *             f"{type(self).__name__} does not implement _get_schema_extra")
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8AvroType_13_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvroType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__314)); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_get_schema_extra, __pyx_t_7) < 0) __PYX_ERR(2, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_AvroType);

  /* "src/type.pxi":286
 *             raise TypeError(f"Default value {schema_default!r} is not valid for field: {field}") from e
 * 
 *     def walk_types(self, visited):             # <<<<<<<<<<<<<<
 *         if self in visited:
 *             return
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8AvroType_15walk_types, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvroType_walk_types, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_walk_types, __pyx_t_7) < 0) __PYX_ERR(2, 286, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_AvroType);

  /* "src/type.pxi":292
 *         yield self
 * 
 *     cpdef get_schema(self, created=None):             # <<<<<<<<<<<<<<
 *         if created is None:
 *             created = set()
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8AvroType_18get_schema, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvroType_get_schema, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__315)); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__141);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_get_schema, __pyx_t_7) < 0) __PYX_ERR(2, 292, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_AvroType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8AvroType_22__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvroType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__316)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_AvroType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_AvroType, (type(self), 0x84c5042, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_AvroType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8AvroType_24__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvroType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__317)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_AvroType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_AvroType);

  /* "src/type.pxi":396
 *         return frozenset(resolve_namespaced_name(self.effective_namespace, alias) for alias in all_names)
 * 
 *     cpdef dict _get_schema_extra(self, set created):             # <<<<<<<<<<<<<<
 *         schema = {}
 *         if self.effective_namespace:
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10_NamedType_3_get_schema_extra, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NamedType__get_schema_extra, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__318)); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__NamedType, __pyx_n_s_get_schema_extra, __pyx_t_7) < 0) __PYX_ERR(2, 396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__NamedType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10_NamedType_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NamedType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__319)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__NamedType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__NamedType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__NamedType, (type(self), 0x24d7e16, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__NamedType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10_NamedType_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NamedType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__320)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__NamedType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__NamedType);

  /* "src/type.pxi":427
 * 
 * PRIMITIVE_TYPES = {
 *     'null': NullType,             # <<<<<<<<<<<<<<
 *     'boolean': BoolType,
 *     'int': IntType,
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(8); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_null_2, ((PyObject *)__pyx_ptype_5cavro_NullType)) < 0) __PYX_ERR(2, 427, __pyx_L1_error)

  /* "src/type.pxi":428
 * PRIMITIVE_TYPES = {
 *     'null': NullType,
 *     'boolean': BoolType,             # <<<<<<<<<<<<<<
 *     'int': IntType,
 *     'long': LongType,
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_boolean_2, ((PyObject *)__pyx_ptype_5cavro_BoolType)) < 0) __PYX_ERR(2, 427, __pyx_L1_error)

  /* "src/type.pxi":429
 *     'null': NullType,
 *     'boolean': BoolType,
 *     'int': IntType,             # <<<<<<<<<<<<<<
 *     'long': LongType,
 *     'float': FloatType,
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_int_2, ((PyObject *)__pyx_ptype_5cavro_IntType)) < 0) __PYX_ERR(2, 427, __pyx_L1_error)

  /* "src/type.pxi":430
 *     'boolean': BoolType,
 *     'int': IntType,
 *     'long': LongType,             # <<<<<<<<<<<<<<
 *     'float': FloatType,
 *     'double': DoubleType,
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_long_2, ((PyObject *)__pyx_ptype_5cavro_LongType)) < 0) __PYX_ERR(2, 427, __pyx_L1_error)

  /* "src/type.pxi":431
 *     'int': IntType,
 *     'long': LongType,
 *     'float': FloatType,             # <<<<<<<<<<<<<<
 *     'double': DoubleType,
 *     'bytes': BytesType,
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_float_2, ((PyObject *)__pyx_ptype_5cavro_FloatType)) < 0) __PYX_ERR(2, 427, __pyx_L1_error)

  /* "src/type.pxi":432
 *     'long': LongType,
 *     'float': FloatType,
 *     'double': DoubleType,             # <<<<<<<<<<<<<<
 *     'bytes': BytesType,
 *     'string': StringType,
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_double_2, ((PyObject *)__pyx_ptype_5cavro_DoubleType)) < 0) __PYX_ERR(2, 427, __pyx_L1_error)

  /* "src/type.pxi":433
 *     'float': FloatType,
 *     'double': DoubleType,
 *     'bytes': BytesType,             # <<<<<<<<<<<<<<
 *     'string': StringType,
 * }
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_bytes_3, ((PyObject *)__pyx_ptype_5cavro_BytesType)) < 0) __PYX_ERR(2, 427, __pyx_L1_error)

  /* "src/type.pxi":434
 *     'double': DoubleType,
 *     'bytes': BytesType,
 *     'string': StringType,             # <<<<<<<<<<<<<<
 * }
 * 
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_string, ((PyObject *)__pyx_ptype_5cavro_StringType)) < 0) __PYX_ERR(2, 427, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PRIMITIVE_TYPES, __pyx_t_7) < 0) __PYX_ERR(2, 426, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/type.pxi":438
 * 
 * TYPES_BY_NAME = dict(
 *     PRIMITIVE_TYPES,             # <<<<<<<<<<<<<<
 *     map=MapType,
 *     enum=EnumType,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_PRIMITIVE_TYPES); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "src/type.pxi":437
 * }
 * 
 * TYPES_BY_NAME = dict(             # <<<<<<<<<<<<<<
 *     PRIMITIVE_TYPES,
 *     map=MapType,
 */
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7)) __PYX_ERR(2, 437, __pyx_L1_error);
  __pyx_t_7 = 0;

  /* "src/type.pxi":439
 * TYPES_BY_NAME = dict(
 *     PRIMITIVE_TYPES,
 *     map=MapType,             # <<<<<<<<<<<<<<
 *     enum=EnumType,
 *     record=RecordType,
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_map, ((PyObject *)__pyx_ptype_5cavro_MapType)) < 0) __PYX_ERR(2, 439, __pyx_L1_error)

  /* "src/type.pxi":440
 *     PRIMITIVE_TYPES,
 *     map=MapType,
 *     enum=EnumType,             # <<<<<<<<<<<<<<
 *     record=RecordType,
 *     fixed=FixedType,
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_enum, ((PyObject *)__pyx_ptype_5cavro_EnumType)) < 0) __PYX_ERR(2, 439, __pyx_L1_error)

  /* "src/type.pxi":441
 *     map=MapType,
 *     enum=EnumType,
 *     record=RecordType,             # <<<<<<<<<<<<<<
 *     fixed=FixedType,
 *     array=ArrayType,
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_record, ((PyObject *)__pyx_ptype_5cavro_RecordType)) < 0) __PYX_ERR(2, 439, __pyx_L1_error)

  /* "src/type.pxi":442
 *     enum=EnumType,
 *     record=RecordType,
 *     fixed=FixedType,             # <<<<<<<<<<<<<<
 *     array=ArrayType,
 * )
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_fixed, ((PyObject *)__pyx_ptype_5cavro_FixedType)) < 0) __PYX_ERR(2, 439, __pyx_L1_error)

  /* "src/type.pxi":443
 *     record=RecordType,
 *     fixed=FixedType,
 *     array=ArrayType,             # <<<<<<<<<<<<<<
 * )
 */
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_array, ((PyObject *)__pyx_ptype_5cavro_ArrayType)) < 0) __PYX_ERR(2, 439, __pyx_L1_error)

  /* "src/type.pxi":437
 * }
 * 
 * TYPES_BY_NAME = dict(             # <<<<<<<<<<<<<<
 *     PRIMITIVE_TYPES,
 *     map=MapType,
 */
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)(&PyDict_Type)), __pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TYPES_BY_NAME, __pyx_t_5) < 0) __PYX_ERR(2, 437, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/schema.pxi":1
 * import json             # <<<<<<<<<<<<<<
 * import hashlib
 * 
 */
  __pyx_t_5 = __Pyx_ImportDottedModule(__pyx_n_s_json, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_json, __pyx_t_5) < 0) __PYX_ERR(23, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/schema.pxi":2
 * import json
 * import hashlib             # <<<<<<<<<<<<<<
 * 
 * cdef str resolve_namespaced_name(str namespace, str name):
 */
  __pyx_t_5 = __Pyx_ImportDottedModule(__pyx_n_s_hashlib, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_hashlib, __pyx_t_5) < 0) __PYX_ERR(23, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/schema.pxi":10
 * 
 * 
 * class _class_inst_method:             # <<<<<<<<<<<<<<
 *     def __init__(self, func):
 *         self.func = func
 */
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_class_inst_method, __pyx_n_s_class_inst_method, (PyObject *) NULL, __pyx_n_s_cavro, (PyObject *) NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "src/schema.pxi":11
 * 
 * class _class_inst_method:
 *     def __init__(self, func):             # <<<<<<<<<<<<<<
 *         self.func = func
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_18_class_inst_method_1__init__, 0, __pyx_n_s_class_inst_method___init, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__322)); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_init, __pyx_t_7) < 0) __PYX_ERR(23, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/schema.pxi":14
 *         self.func = func
 * 
 *     def __get__(self, inst, cls):             # <<<<<<<<<<<<<<
 *         return partial(self.func, inst, cls)
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_18_class_inst_method_3__get__, 0, __pyx_n_s_class_inst_method___get, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__324)); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_get_2, __pyx_t_7) < 0) __PYX_ERR(23, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/schema.pxi":10
 * 
 * 
 * class _class_inst_method:             # <<<<<<<<<<<<<<
 *     def __init__(self, func):
 *         self.func = func
 */
  __pyx_t_7 = __Pyx_Py3ClassCreate(((PyObject*)&PyType_Type), __pyx_n_s_class_inst_method, __pyx_empty_tuple, __pyx_t_5, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_class_inst_method, __pyx_t_7) < 0) __PYX_ERR(23, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/schema.pxi":48
 *     cdef readonly dict logical_types
 * 
 *     def __init__(self, source: Union[str, object], Options options: Options=DEFAULT_OPTIONS, *, named_types: dict[str, AvroType]=None, parse_json: bool=True, _type=None, **extra_options):             # <<<<<<<<<<<<<<
 *         if isinstance(source, (str, bytes)) and parse_json:
 *             source = json.loads(source)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_DEFAULT_OPTIONS); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5cavro_Options))))) __PYX_ERR(23, 48, __pyx_L1_error)
  __pyx_k__83 = ((struct __pyx_obj_5cavro_Options *)__pyx_t_5);
  __Pyx_GIVEREF((PyObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "src/schema.pxi":59
 *         self.type = AvroType.for_schema(self) if _type is None else _type
 * 
 *     @_class_inst_method             # <<<<<<<<<<<<<<
 *     def _wrap_type(inst, cls, AvroType avro_type, Options options=None):
 *         """
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_3_wrap_type, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema__wrap_type, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__326)); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__141);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_wrap_type, __pyx_t_5) < 0) __PYX_ERR(23, 59, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_class_inst_method); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_wrap_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(23, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = NULL;
  __pyx_t_11 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_11 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_4};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_wrap_type, __pyx_t_5) < 0) __PYX_ERR(23, 59, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "src/schema.pxi":74
 *         return new_inst
 * 
 *     def _make_logical_types(self, options):             # <<<<<<<<<<<<<<
 *         logical_by_name = {}
 *         self.logical_types = logical_by_name
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_5_make_logical_types, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema__make_logical_types, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__328)); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_make_logical_types, __pyx_t_5) < 0) __PYX_ERR(23, 74, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "src/schema.pxi":118
 *             return json.dumps(self.schema, indent=2)
 * 
 *     def fingerprint(self, method='rabin', **kwargs) -> Union[bytes, hashlib._hashlib.HASH]:             # <<<<<<<<<<<<<<
 *         """
 *         Return the deterministic fingerprint of the schema, using the given hash method.
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_kp_s_Union_bytes_hashlib__hashlib_HAS) < 0) __PYX_ERR(23, 118, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_9fingerprint, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema_fingerprint, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__330)); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__331);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_fingerprint, __pyx_t_7) < 0) __PYX_ERR(23, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "src/schema.pxi":138
 *         return hasher
 * 
 *     cpdef AvroType find_type(self, str namespace, str name, bint _raise=True):             # <<<<<<<<<<<<<<
 *         """
 *         Given a namespace and name (namespace may be None), find and return the `AvroType` instance matching this name.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_11find_type, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema_find_type, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__333)); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__334);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_find_type, __pyx_t_7) < 0) __PYX_ERR(23, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "src/schema.pxi":159
 *             raise UnknownType(f'Unknown type: {resolved!r}')
 * 
 *     def can_encode(self, value: object) -> bool:             # <<<<<<<<<<<<<<
 *         """
 *         Check if `value` can be encoded using this schema
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_value, __pyx_n_s_object) < 0) __PYX_ERR(23, 159, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(23, 159, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_13can_encode, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema_can_encode, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__336)); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_can_encode, __pyx_t_5) < 0) __PYX_ERR(23, 159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "src/schema.pxi":166
 *         return fitness > FIT_NONE
 * 
 *     def binary_encode(self, value: object) -> bytes:             # <<<<<<<<<<<<<<
 *         """
 *         Encode `value` using this schema and return the avro bytes representing it.
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_value, __pyx_n_s_object) < 0) __PYX_ERR(23, 166, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_bytes_3) < 0) __PYX_ERR(23, 166, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_15binary_encode, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema_binary_encode, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__338)); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_binary_encode, __pyx_t_7) < 0) __PYX_ERR(23, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "src/schema.pxi":174
 *         return buffer.bytes()
 * 
 *     def binary_decode(self, bytes value: bytes) -> object:             # <<<<<<<<<<<<<<
 *         """
 *         Decode `value` using this schema and return the decoded value.
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_value, __pyx_n_s_bytes_3) < 0) __PYX_ERR(23, 174, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_return, __pyx_n_s_object) < 0) __PYX_ERR(23, 174, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_17binary_decode, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema_binary_decode, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__339)); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_binary_decode, __pyx_t_5) < 0) __PYX_ERR(23, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "src/schema.pxi":181
 *         return self.type.binary_buffer_decode(buffer)
 * 
 *     cpdef binary_read(self, _Reader reader: _Reader):             # <<<<<<<<<<<<<<
 *         """
 *         Read a value from `reader` using this schema and return the decoded value.
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_reader, __pyx_n_s_Reader) < 0) __PYX_ERR(23, 181, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_19binary_read, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema_binary_read, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__341)); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_binary_read, __pyx_t_7) < 0) __PYX_ERR(23, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "src/schema.pxi":188
 *         return self.type.binary_buffer_decode(reader)
 * 
 *     cpdef binary_write(self, _Writer writer: _Writer, value: object):             # <<<<<<<<<<<<<<
 *         """
 *         Write `value` to `writer` using this schema.
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_writer, __pyx_n_s_Writer) < 0) __PYX_ERR(23, 188, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_value, __pyx_n_s_object) < 0) __PYX_ERR(23, 188, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_21binary_write, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema_binary_write, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__343)); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_binary_write, __pyx_t_5) < 0) __PYX_ERR(23, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "src/schema.pxi":195
 *         self.type.binary_buffer_encode(writer, value)
 * 
 *     def json_encode(self, value, serialize=True, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Encode `value` using this schema and return the avro json representing it.
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_23json_encode, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema_json_encode, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__345)); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__346);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_json_encode, __pyx_t_5) < 0) __PYX_ERR(23, 195, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "src/schema.pxi":204
 *         return data
 * 
 *     def json_decode(self, value: Union[str, object], deserialize: bool=True, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Decode `value` in JSON form using this schema and return the decoded value.
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_value, __pyx_kp_s_Union_str_object) < 0) __PYX_ERR(23, 204, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_deserialize, __pyx_n_s_bool) < 0) __PYX_ERR(23, 204, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_25json_decode, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema_json_decode, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__348)); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__346);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_json_decode, __pyx_t_7) < 0) __PYX_ERR(23, 204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "src/schema.pxi":215
 *         return self.type.json_decode(value)
 * 
 *     cpdef Schema reader_for_writer(self, Schema writer_schema: Schema):             # <<<<<<<<<<<<<<
 *         """
 *         Return a schema that is the result of promoting this schema to the writer schema.
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_writer_schema, __pyx_n_s_Schema) < 0) __PYX_ERR(23, 215, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_27reader_for_writer, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema_reader_for_writer, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__350)); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_reader_for_writer, __pyx_t_5) < 0) __PYX_ERR(23, 215, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_29__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__351)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Schema, (type(self), 0xc2592e1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Schema__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_6Schema_31__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Schema___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__352)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Schema, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Schema);

  /* "src/schema.pxi":232
 *     """
 * 
 *     def __init__(self, AvroType resolved_type, Options options=DEFAULT_OPTIONS):             # <<<<<<<<<<<<<<
 *         named_types = {}
 *         for sub_type in resolved_type.walk_types(set()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_DEFAULT_OPTIONS); if (unlikely(!__pyx_t_5)) __PYX_ERR(23, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5cavro_Options))))) __PYX_ERR(23, 232, __pyx_L1_error)
  __pyx_k__84 = ((struct __pyx_obj_5cavro_Options *)__pyx_t_5);
  __Pyx_GIVEREF((PyObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_14ResolvedSchema_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ResolvedSchema___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__353)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ResolvedSchema, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ResolvedSchema);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ResolvedSchema, (type(self), 0xc2592e1, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ResolvedSchema__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_14ResolvedSchema_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ResolvedSchema___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__354)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ResolvedSchema, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ResolvedSchema);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_12ValueAdapter_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ValueAdapter___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__355)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ValueAdapter, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ValueAdapter);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ValueAdapter, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ValueAdapter__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_12ValueAdapter_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ValueAdapter___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__356)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ValueAdapter, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ValueAdapter);

  /* "src/logical.pxi":31
 *     """
 * 
 *     logical_name = NotImplemented             # <<<<<<<<<<<<<<
 *     underlying_types = NotImplemented
 * 
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_LogicalType, __pyx_n_s_logical_name, __pyx_builtin_NotImplemented) < 0) __PYX_ERR(24, 31, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_LogicalType);

  /* "src/logical.pxi":32
 * 
 *     logical_name = NotImplemented
 *     underlying_types = NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     @_class_inst_method
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_LogicalType, __pyx_n_s_underlying_types, __pyx_builtin_NotImplemented) < 0) __PYX_ERR(24, 32, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_LogicalType);

  /* "src/logical.pxi":34
 *     underlying_types = NotImplemented
 * 
 *     @_class_inst_method             # <<<<<<<<<<<<<<
 *     def for_type(inst, cls, underlying: AvroType):
 *         cdef object ob = inst
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_underlying, __pyx_n_s_AvroType) < 0) __PYX_ERR(24, 34, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_11LogicalType_1for_type, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LogicalType_for_type, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__358)); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_LogicalType, __pyx_n_s_for_type_2, __pyx_t_7) < 0) __PYX_ERR(24, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_LogicalType);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_class_inst_method); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetNameInClass(__pyx_t_4, (PyObject *)__pyx_ptype_5cavro_LogicalType, __pyx_n_s_for_type_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = NULL;
  __pyx_t_11 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_11 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_4};
    __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_LogicalType, __pyx_n_s_for_type_2, __pyx_t_7) < 0) __PYX_ERR(24, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_LogicalType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_11LogicalType_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LogicalType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__359)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_LogicalType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_LogicalType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LogicalType, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LogicalType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_11LogicalType_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LogicalType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__360)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_LogicalType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_LogicalType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_17CustomLogicalType_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CustomLogicalType___reduce_cytho, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__361)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_CustomLogicalType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_CustomLogicalType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CustomLogicalType, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CustomLogicalType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_17CustomLogicalType_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CustomLogicalType___setstate_cyt, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__362)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_CustomLogicalType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_CustomLogicalType);

  /* "src/logical.pxi":75
 *     """Logical type for decimal values."""
 * 
 *     logical_name = 'decimal'             # <<<<<<<<<<<<<<
 *     underlying_types = (BytesType, FixedType)
 * 
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_DecimalType, __pyx_n_s_logical_name, __pyx_n_u_decimal) < 0) __PYX_ERR(24, 75, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_DecimalType);

  /* "src/logical.pxi":76
 * 
 *     logical_name = 'decimal'
 *     underlying_types = (BytesType, FixedType)             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly int precision
 */
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_BytesType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_BytesType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_ptype_5cavro_BytesType))) __PYX_ERR(24, 76, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_FixedType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_FixedType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, ((PyObject *)__pyx_ptype_5cavro_FixedType))) __PYX_ERR(24, 76, __pyx_L1_error);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_DecimalType, __pyx_n_s_underlying_types, __pyx_t_7) < 0) __PYX_ERR(24, 76, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_DecimalType);

  /* "src/logical.pxi":100
 *         return f'{self.underlying_name}(decimal)'
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         meta = underlying.metadata
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_underlying, __pyx_n_s_AvroType) < 0) __PYX_ERR(24, 100, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_11DecimalType_3_for_type, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DecimalType__for_type, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__364)); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_DecimalType, __pyx_n_s_for_type_3, __pyx_t_5) < 0) __PYX_ERR(24, 100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_DecimalType);
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_DecimalType, __pyx_n_s_for_type_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_Method_ClassMethod(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_DecimalType, __pyx_n_s_for_type_3, __pyx_t_7) < 0) __PYX_ERR(24, 100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_DecimalType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_11DecimalType_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DecimalType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__365)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_DecimalType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_DecimalType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DecimalType, (type(self), 0xe9db223, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DecimalType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_11DecimalType_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DecimalType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__366)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_DecimalType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_DecimalType);

  /* "src/logical.pxi":169
 * cdef class UUIDBase(LogicalType):
 *     """Logical type for UUID values"""
 *     logical_name = 'uuid'             # <<<<<<<<<<<<<<
 *     cdef bint return_uuid_object
 * 
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDBase, __pyx_n_s_logical_name, __pyx_n_u_uuid) < 0) __PYX_ERR(24, 169, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_UUIDBase);

  /* "src/logical.pxi":175
 *         self.return_uuid_object = return_uuid_object
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.return_uuid_object)
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_underlying, __pyx_n_s_AvroType) < 0) __PYX_ERR(24, 175, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8UUIDBase_3_for_type, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UUIDBase__for_type, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__368)); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDBase, __pyx_n_s_for_type_3, __pyx_t_5) < 0) __PYX_ERR(24, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDBase);
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_UUIDBase, __pyx_n_s_for_type_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_Method_ClassMethod(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDBase, __pyx_n_s_for_type_3, __pyx_t_7) < 0) __PYX_ERR(24, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDBase);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8UUIDBase_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UUIDBase___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__369)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDBase, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDBase);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UUIDBase, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UUIDBase__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_8UUIDBase_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UUIDBase___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__370)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDBase, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDBase);

  /* "src/logical.pxi":181
 * 
 * cdef class UUIDStringType(UUIDBase):
 *     underlying_types = (StringType, )             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_StringType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_StringType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_ptype_5cavro_StringType))) __PYX_ERR(24, 181, __pyx_L1_error);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDStringType, __pyx_n_s_underlying_types, __pyx_t_7) < 0) __PYX_ERR(24, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDStringType);

  /* "src/logical.pxi":183
 *     underlying_types = (StringType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.return_uuid_object)
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_underlying, __pyx_n_s_AvroType) < 0) __PYX_ERR(24, 183, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_14UUIDStringType_1_for_type, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UUIDStringType__for_type, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__371)); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDStringType, __pyx_n_s_for_type_3, __pyx_t_5) < 0) __PYX_ERR(24, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDStringType);
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_UUIDStringType, __pyx_n_s_for_type_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_Method_ClassMethod(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDStringType, __pyx_n_s_for_type_3, __pyx_t_7) < 0) __PYX_ERR(24, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDStringType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_14UUIDStringType_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UUIDStringType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__372)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDStringType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDStringType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UUIDStringType, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UUIDStringType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_14UUIDStringType_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UUIDStringType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__373)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDStringType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDStringType);

  /* "src/logical.pxi":210
 * 
 * cdef class UUIDFixedType(UUIDBase):
 *     underlying_types = (FixedType, )             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_FixedType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_FixedType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_ptype_5cavro_FixedType))) __PYX_ERR(24, 210, __pyx_L1_error);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDFixedType, __pyx_n_s_underlying_types, __pyx_t_7) < 0) __PYX_ERR(24, 210, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDFixedType);

  /* "src/logical.pxi":212
 *     underlying_types = (FixedType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         if underlying.size != 16:
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_underlying, __pyx_n_s_AvroType) < 0) __PYX_ERR(24, 212, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_13UUIDFixedType_1_for_type, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UUIDFixedType__for_type, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__374)); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDFixedType, __pyx_n_s_for_type_3, __pyx_t_5) < 0) __PYX_ERR(24, 212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDFixedType);
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_5cavro_UUIDFixedType, __pyx_n_s_for_type_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_Method_ClassMethod(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDFixedType, __pyx_n_s_for_type_3, __pyx_t_7) < 0) __PYX_ERR(24, 212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDFixedType);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_13UUIDFixedType_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UUIDFixedType___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__375)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDFixedType, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDFixedType);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_UUIDFixedType, (type(self), 0x987bfb3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_UUIDFixedType__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_13UUIDFixedType_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UUIDFixedType___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__376)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_UUIDFixedType, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_UUIDFixedType);

  /* "src/logical.pxi":233
 * 
 * 
 * EPOCH_DATE = datetime.date(1970, 1, 1)             # <<<<<<<<<<<<<<
 * EPOCH_DT = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
 * EPOCH_DT_NO_TZ = datetime.datetime(1970, 1, 1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_datetime); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_date); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__377, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EPOCH_DATE, __pyx_t_7) < 0) __PYX_ERR(24, 233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/logical.pxi":234
 * 
 * EPOCH_DATE = datetime.date(1970, 1, 1)
 * EPOCH_DT = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)             # <<<<<<<<<<<<<<
 * EPOCH_DT_NO_TZ = datetime.datetime(1970, 1, 1)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_datetime); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_timezone); if (unlikely(!__pyx_t_9)) __PYX_ERR(24, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_utc); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_tzinfo, __pyx_t_4) < 0) __PYX_ERR(24, 234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__377, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EPOCH_DT, __pyx_t_4) < 0) __PYX_ERR(24, 234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/logical.pxi":235
 * EPOCH_DATE = datetime.date(1970, 1, 1)
 * EPOCH_DT = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
 * EPOCH_DT_NO_TZ = datetime.datetime(1970, 1, 1)             # <<<<<<<<<<<<<<
 * 
 * cdef class Date(LogicalType):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_datetime); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__377, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EPOCH_DT_NO_TZ, __pyx_t_4) < 0) __PYX_ERR(24, 235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/logical.pxi":239
 * cdef class Date(LogicalType):
 *     """Logical type for Date values"""
 *     logical_name = 'date'             # <<<<<<<<<<<<<<
 *     underlying_types = (IntType, )
 * 
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Date, __pyx_n_s_logical_name, __pyx_n_u_date) < 0) __PYX_ERR(24, 239, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_Date);

  /* "src/logical.pxi":240
 *     """Logical type for Date values"""
 *     logical_name = 'date'
 *     underlying_types = (IntType, )             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly bint accepts_string
 */
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_IntType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_IntType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_ptype_5cavro_IntType))) __PYX_ERR(24, 240, __pyx_L1_error);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Date, __pyx_n_s_underlying_types, __pyx_t_4) < 0) __PYX_ERR(24, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Date);

  /* "src/logical.pxi":247
 *         self.accepts_string = accepts_string
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(underlying.options.date_type_accepts_string)
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_underlying, __pyx_n_s_AvroType) < 0) __PYX_ERR(24, 247, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_4Date_3_for_type, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Date__for_type, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__378)); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Date, __pyx_n_s_for_type_3, __pyx_t_7) < 0) __PYX_ERR(24, 247, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Date);
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_Date, __pyx_n_s_for_type_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_Method_ClassMethod(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Date, __pyx_n_s_for_type_3, __pyx_t_4) < 0) __PYX_ERR(24, 247, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Date);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_4Date_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Date___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__379)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Date, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Date);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Date, (type(self), 0xc5331ff, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Date__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_4Date_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Date___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__380)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Date, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Date);

  /* "src/logical.pxi":270
 * 
 * 
 * def _time_to_micros(value):             # <<<<<<<<<<<<<<
 *     return value.microsecond + 1_000_000 * (value.second + 60 * (value.minute + 60 * value.hour))
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15_time_to_micros, 0, __pyx_n_s_time_to_micros, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__382)); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_time_to_micros, __pyx_t_4) < 0) __PYX_ERR(24, 270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/logical.pxi":274
 * 
 * 
 * def _micros_to_time(value):             # <<<<<<<<<<<<<<
 *     return datetime.time(
 *         hour=value // 3_600_000_000,
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_17_micros_to_time, 0, __pyx_n_s_micros_to_time, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__383)); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_micros_to_time, __pyx_t_4) < 0) __PYX_ERR(24, 274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/logical.pxi":283
 * 
 * 
 * MAX_TIME = _time_to_micros(datetime.time.max)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_time_to_micros); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_datetime); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_max); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MAX_TIME, __pyx_t_5) < 0) __PYX_ERR(24, 283, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/logical.pxi":294
 * cdef class TimeMillis(LogicalType):
 *     """Logical type for time-millis values"""
 *     logical_name = 'time-millis'             # <<<<<<<<<<<<<<
 *     underlying_types = (IntType, )
 * 
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimeMillis, __pyx_n_s_logical_name, __pyx_kp_u_time_millis) < 0) __PYX_ERR(24, 294, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_TimeMillis);

  /* "src/logical.pxi":295
 *     """Logical type for time-millis values"""
 *     logical_name = 'time-millis'
 *     underlying_types = (IntType, )             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_IntType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_IntType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_ptype_5cavro_IntType))) __PYX_ERR(24, 295, __pyx_L1_error);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimeMillis, __pyx_n_s_underlying_types, __pyx_t_5) < 0) __PYX_ERR(24, 295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimeMillis);

  /* "src/logical.pxi":297
 *     underlying_types = (IntType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_underlying, __pyx_n_s_AvroType) < 0) __PYX_ERR(24, 297, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10TimeMillis_1_for_type, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimeMillis__for_type, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__384)); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimeMillis, __pyx_n_s_for_type_3, __pyx_t_7) < 0) __PYX_ERR(24, 297, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimeMillis);
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_TimeMillis, __pyx_n_s_for_type_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_Method_ClassMethod(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimeMillis, __pyx_n_s_for_type_3, __pyx_t_5) < 0) __PYX_ERR(24, 297, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimeMillis);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10TimeMillis_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimeMillis___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__385)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimeMillis, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimeMillis);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimeMillis, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimeMillis__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10TimeMillis_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimeMillis___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__386)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimeMillis, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimeMillis);

  /* "src/logical.pxi":318
 * cdef class TimeMicros(LogicalType):
 *     """Logical type for time-micros values"""
 *     logical_name = 'time-micros'             # <<<<<<<<<<<<<<
 *     underlying_types = (LongType, )
 * 
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimeMicros, __pyx_n_s_logical_name, __pyx_kp_u_time_micros) < 0) __PYX_ERR(24, 318, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_TimeMicros);

  /* "src/logical.pxi":319
 *     """Logical type for time-micros values"""
 *     logical_name = 'time-micros'
 *     underlying_types = (LongType, )             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_LongType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_LongType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_ptype_5cavro_LongType))) __PYX_ERR(24, 319, __pyx_L1_error);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimeMicros, __pyx_n_s_underlying_types, __pyx_t_5) < 0) __PYX_ERR(24, 319, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimeMicros);

  /* "src/logical.pxi":321
 *     underlying_types = (LongType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_underlying, __pyx_n_s_AvroType) < 0) __PYX_ERR(24, 321, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10TimeMicros_1_for_type, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimeMicros__for_type, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__387)); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimeMicros, __pyx_n_s_for_type_3, __pyx_t_7) < 0) __PYX_ERR(24, 321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimeMicros);
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_TimeMicros, __pyx_n_s_for_type_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_Method_ClassMethod(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimeMicros, __pyx_n_s_for_type_3, __pyx_t_5) < 0) __PYX_ERR(24, 321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimeMicros);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10TimeMicros_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimeMicros___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__388)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimeMicros, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimeMicros);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimeMicros, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimeMicros__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10TimeMicros_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimeMicros___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__389)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimeMicros, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimeMicros);

  /* "src/logical.pxi":342
 * cdef class TimestampMillis(LogicalType):
 *     """Logical type for timestamp-micros values"""
 *     logical_name = 'timestamp-millis'             # <<<<<<<<<<<<<<
 *     underlying_types = (LongType, )
 * 
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimestampMillis, __pyx_n_s_logical_name, __pyx_kp_u_timestamp_millis) < 0) __PYX_ERR(24, 342, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_TimestampMillis);

  /* "src/logical.pxi":343
 *     """Logical type for timestamp-micros values"""
 *     logical_name = 'timestamp-millis'
 *     underlying_types = (LongType, )             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly bint alternate_timestamp_encoding
 */
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_LongType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_LongType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_ptype_5cavro_LongType))) __PYX_ERR(24, 343, __pyx_L1_error);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimestampMillis, __pyx_n_s_underlying_types, __pyx_t_5) < 0) __PYX_ERR(24, 343, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimestampMillis);

  /* "src/logical.pxi":350
 *         self.alternate_timestamp_encoding = alternate_timestamp_encoding
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls(
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_underlying, __pyx_n_s_AvroType) < 0) __PYX_ERR(24, 350, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15TimestampMillis_3_for_type, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimestampMillis__for_type, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__390)); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimestampMillis, __pyx_n_s_for_type_3, __pyx_t_7) < 0) __PYX_ERR(24, 350, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimestampMillis);
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_TimestampMillis, __pyx_n_s_for_type_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_Method_ClassMethod(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimestampMillis, __pyx_n_s_for_type_3, __pyx_t_5) < 0) __PYX_ERR(24, 350, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimestampMillis);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15TimestampMillis_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimestampMillis___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__391)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimestampMillis, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimestampMillis);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimestampMillis, (type(self), 0x8d5b257, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimestampMillis__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15TimestampMillis_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimestampMillis___setstate_cytho, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__392)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimestampMillis, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimestampMillis);

  /* "src/logical.pxi":377
 * cdef class TimestampMicros(LogicalType):
 *     """Logical type for timestamp-micros values"""
 *     logical_name = 'timestamp-micros'             # <<<<<<<<<<<<<<
 *     underlying_types = (LongType, )
 * 
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimestampMicros, __pyx_n_s_logical_name, __pyx_kp_u_timestamp_micros) < 0) __PYX_ERR(24, 377, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_TimestampMicros);

  /* "src/logical.pxi":378
 *     """Logical type for timestamp-micros values"""
 *     logical_name = 'timestamp-micros'
 *     underlying_types = (LongType, )             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF((PyObject *)__pyx_ptype_5cavro_LongType);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_5cavro_LongType);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_ptype_5cavro_LongType))) __PYX_ERR(24, 378, __pyx_L1_error);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimestampMicros, __pyx_n_s_underlying_types, __pyx_t_5) < 0) __PYX_ERR(24, 378, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimestampMicros);

  /* "src/logical.pxi":380
 *     underlying_types = (LongType, )
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _for_type(cls, underlying: AvroType):
 *         return cls()
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_underlying, __pyx_n_s_AvroType) < 0) __PYX_ERR(24, 380, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15TimestampMicros_1_for_type, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimestampMicros__for_type, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__393)); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimestampMicros, __pyx_n_s_for_type_3, __pyx_t_7) < 0) __PYX_ERR(24, 380, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimestampMicros);
  __Pyx_GetNameInClass(__pyx_t_7, (PyObject *)__pyx_ptype_5cavro_TimestampMicros, __pyx_n_s_for_type_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(24, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_Method_ClassMethod(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimestampMicros, __pyx_n_s_for_type_3, __pyx_t_5) < 0) __PYX_ERR(24, 380, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimestampMicros);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15TimestampMicros_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimestampMicros___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__394)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimestampMicros, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimestampMicros);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TimestampMicros, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TimestampMicros__set_state(self, __pyx_state)
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15TimestampMicros_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimestampMicros___setstate_cytho, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__395)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_TimestampMicros, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_5cavro_TimestampMicros);

  /* "src/codec.pxi":2
 * 
 * HAVE_BZIP2 = False             # <<<<<<<<<<<<<<
 * try:
 *     import bz2
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_BZIP2, Py_False) < 0) __PYX_ERR(3, 2, __pyx_L1_error)

  /* "src/codec.pxi":3
 * 
 * HAVE_BZIP2 = False
 * try:             # <<<<<<<<<<<<<<
 *     import bz2
 *     HAVE_BZIP2 = True
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "src/codec.pxi":4
 * HAVE_BZIP2 = False
 * try:
 *     import bz2             # <<<<<<<<<<<<<<
 *     HAVE_BZIP2 = True
 * except ImportError:
 */
      __pyx_t_5 = __Pyx_ImportDottedModule(__pyx_n_s_bz2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 4, __pyx_L31_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_bz2, __pyx_t_5) < 0) __PYX_ERR(3, 4, __pyx_L31_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "src/codec.pxi":5
 * try:
 *     import bz2
 *     HAVE_BZIP2 = True             # <<<<<<<<<<<<<<
 * except ImportError:
 *     pass
 */
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_BZIP2, Py_True) < 0) __PYX_ERR(3, 5, __pyx_L31_error)

      /* "src/codec.pxi":3
 * 
 * HAVE_BZIP2 = False
 * try:             # <<<<<<<<<<<<<<
 *     import bz2
 *     HAVE_BZIP2 = True
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L36_try_end;
    __pyx_L31_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "src/codec.pxi":6
 *     import bz2
 *     HAVE_BZIP2 = True
 * except ImportError:             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_11) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L32_exception_handled;
    }
    goto __pyx_L33_except_error;

    /* "src/codec.pxi":3
 * 
 * HAVE_BZIP2 = False
 * try:             # <<<<<<<<<<<<<<
 *     import bz2
 *     HAVE_BZIP2 = True
 */
    __pyx_L33_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L32_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L36_try_end:;
  }

  /* "src/codec.pxi":9
 *     pass
 * 
 * HAVE_XZ = False             # <<<<<<<<<<<<<<
 * try:
 *     import lzma
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_XZ, Py_False) < 0) __PYX_ERR(3, 9, __pyx_L1_error)

  /* "src/codec.pxi":10
 * 
 * HAVE_XZ = False
 * try:             # <<<<<<<<<<<<<<
 *     import lzma
 *     HAVE_XZ = True
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "src/codec.pxi":11
 * HAVE_XZ = False
 * try:
 *     import lzma             # <<<<<<<<<<<<<<
 *     HAVE_XZ = True
 * except ImportError:
 */
      __pyx_t_5 = __Pyx_ImportDottedModule(__pyx_n_s_lzma, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 11, __pyx_L37_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_lzma, __pyx_t_5) < 0) __PYX_ERR(3, 11, __pyx_L37_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "src/codec.pxi":12
 * try:
 *     import lzma
 *     HAVE_XZ = True             # <<<<<<<<<<<<<<
 * except ImportError:
 *     pass
 */
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_XZ, Py_True) < 0) __PYX_ERR(3, 12, __pyx_L37_error)

      /* "src/codec.pxi":10
 * 
 * HAVE_XZ = False
 * try:             # <<<<<<<<<<<<<<
 *     import lzma
 *     HAVE_XZ = True
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L42_try_end;
    __pyx_L37_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "src/codec.pxi":13
 *     import lzma
 *     HAVE_XZ = True
 * except ImportError:             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_11) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L38_exception_handled;
    }
    goto __pyx_L39_except_error;

    /* "src/codec.pxi":10
 * 
 * HAVE_XZ = False
 * try:             # <<<<<<<<<<<<<<
 *     import lzma
 *     HAVE_XZ = True
 */
    __pyx_L39_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L1_error;
    __pyx_L38_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    __pyx_L42_try_end:;
  }

  /* "src/codec.pxi":16
 *     pass
 * 
 * HAVE_ZLIB = False             # <<<<<<<<<<<<<<
 * try:
 *     import zlib
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_ZLIB, Py_False) < 0) __PYX_ERR(3, 16, __pyx_L1_error)

  /* "src/codec.pxi":17
 * 
 * HAVE_ZLIB = False
 * try:             # <<<<<<<<<<<<<<
 *     import zlib
 *     crc32 = zlib.crc32
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "src/codec.pxi":18
 * HAVE_ZLIB = False
 * try:
 *     import zlib             # <<<<<<<<<<<<<<
 *     crc32 = zlib.crc32
 *     HAVE_ZLIB = True
 */
      __pyx_t_5 = __Pyx_ImportDottedModule(__pyx_n_s_zlib, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 18, __pyx_L43_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_zlib, __pyx_t_5) < 0) __PYX_ERR(3, 18, __pyx_L43_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "src/codec.pxi":19
 * try:
 *     import zlib
 *     crc32 = zlib.crc32             # <<<<<<<<<<<<<<
 *     HAVE_ZLIB = True
 * except ImportError:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_zlib); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 19, __pyx_L43_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_crc32); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 19, __pyx_L43_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_crc32, __pyx_t_7) < 0) __PYX_ERR(3, 19, __pyx_L43_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "src/codec.pxi":20
 *     import zlib
 *     crc32 = zlib.crc32
 *     HAVE_ZLIB = True             # <<<<<<<<<<<<<<
 * except ImportError:
 *     pass
 */
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_ZLIB, Py_True) < 0) __PYX_ERR(3, 20, __pyx_L43_error)

      /* "src/codec.pxi":17
 * 
 * HAVE_ZLIB = False
 * try:             # <<<<<<<<<<<<<<
 *     import zlib
 *     crc32 = zlib.crc32
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L48_try_end;
    __pyx_L43_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "src/codec.pxi":21
 *     crc32 = zlib.crc32
 *     HAVE_ZLIB = True
 * except ImportError:             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_11) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L44_exception_handled;
    }
    goto __pyx_L45_except_error;

    /* "src/codec.pxi":17
 * 
 * HAVE_ZLIB = False
 * try:             # <<<<<<<<<<<<<<
 *     import zlib
 *     crc32 = zlib.crc32
 */
    __pyx_L45_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L44_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L48_try_end:;
  }

  /* "src/codec.pxi":25
 * 
 * 
 * HAVE_SNAPPY = False             # <<<<<<<<<<<<<<
 * if HAVE_ZLIB: # Needed for crc32
 *     try:
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_SNAPPY, Py_False) < 0) __PYX_ERR(3, 25, __pyx_L1_error)

  /* "src/codec.pxi":26
 * 
 * HAVE_SNAPPY = False
 * if HAVE_ZLIB: # Needed for crc32             # <<<<<<<<<<<<<<
 *     try:
 *         import snappy
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_HAVE_ZLIB); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_6) {

    /* "src/codec.pxi":27
 * HAVE_SNAPPY = False
 * if HAVE_ZLIB: # Needed for crc32
 *     try:             # <<<<<<<<<<<<<<
 *         import snappy
 *         HAVE_SNAPPY = True
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_1);
      /*try:*/ {

        /* "src/codec.pxi":28
 * if HAVE_ZLIB: # Needed for crc32
 *     try:
 *         import snappy             # <<<<<<<<<<<<<<
 *         HAVE_SNAPPY = True
 *     except ImportError:
 */
        __pyx_t_7 = __Pyx_ImportDottedModule(__pyx_n_s_snappy, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 28, __pyx_L50_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (PyDict_SetItem(__pyx_d, __pyx_n_s_snappy, __pyx_t_7) < 0) __PYX_ERR(3, 28, __pyx_L50_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "src/codec.pxi":29
 *     try:
 *         import snappy
 *         HAVE_SNAPPY = True             # <<<<<<<<<<<<<<
 *     except ImportError:
 *         pass
 */
        if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_SNAPPY, Py_True) < 0) __PYX_ERR(3, 29, __pyx_L50_error)

        /* "src/codec.pxi":27
 * HAVE_SNAPPY = False
 * if HAVE_ZLIB: # Needed for crc32
 *     try:             # <<<<<<<<<<<<<<
 *         import snappy
 *         HAVE_SNAPPY = True
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L55_try_end;
      __pyx_L50_error:;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "src/codec.pxi":30
 *         import snappy
 *         HAVE_SNAPPY = True
 *     except ImportError:             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
      if (__pyx_t_11) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L51_exception_handled;
      }
      goto __pyx_L52_except_error;

      /* "src/codec.pxi":27
 * HAVE_SNAPPY = False
 * if HAVE_ZLIB: # Needed for crc32
 *     try:             # <<<<<<<<<<<<<<
 *         import snappy
 *         HAVE_SNAPPY = True
 */
      __pyx_L52_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
      goto __pyx_L1_error;
      __pyx_L51_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
      __pyx_L55_try_end:;
    }

    /* "src/codec.pxi":26
 * 
 * HAVE_SNAPPY = False
 * if HAVE_ZLIB: # Needed for crc32             # <<<<<<<<<<<<<<
 *     try:
 *         import snappy
 */
  }

  /* "src/codec.pxi":34
 * 
 * 
 * HAVE_ZSTD = False             # <<<<<<<<<<<<<<
 * try:
 *     import zstandard
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_ZSTD, Py_False) < 0) __PYX_ERR(3, 34, __pyx_L1_error)

  /* "src/codec.pxi":35
 * 
 * HAVE_ZSTD = False
 * try:             # <<<<<<<<<<<<<<
 *     import zstandard
 *     HAVE_ZSTD = True
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "src/codec.pxi":36
 * HAVE_ZSTD = False
 * try:
 *     import zstandard             # <<<<<<<<<<<<<<
 *     HAVE_ZSTD = True
 * except ImportError:
 */
      __pyx_t_7 = __Pyx_ImportDottedModule(__pyx_n_s_zstandard, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 36, __pyx_L56_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_zstandard, __pyx_t_7) < 0) __PYX_ERR(3, 36, __pyx_L56_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "src/codec.pxi":37
 * try:
 *     import zstandard
 *     HAVE_ZSTD = True             # <<<<<<<<<<<<<<
 * except ImportError:
 *     pass
 */
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_ZSTD, Py_True) < 0) __PYX_ERR(3, 37, __pyx_L56_error)

      /* "src/codec.pxi":35
 * 
 * HAVE_ZSTD = False
 * try:             # <<<<<<<<<<<<<<
 *     import zstandard
 *     HAVE_ZSTD = True
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L61_try_end;
    __pyx_L56_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "src/codec.pxi":38
 *     import zstandard
 *     HAVE_ZSTD = True
 * except ImportError:             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_11) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L57_exception_handled;
    }
    goto __pyx_L58_except_error;

    /* "src/codec.pxi":35
 * 
 * HAVE_ZSTD = False
 * try:             # <<<<<<<<<<<<<<
 *     import zstandard
 *     HAVE_ZSTD = True
 */
    __pyx_L58_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L57_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L61_try_end:;
  }

  /* "src/codec.pxi":42
 * 
 * 
 * HAVE_LZ4 = False             # <<<<<<<<<<<<<<
 * try:
 *     import lz4.frame
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_LZ4, Py_False) < 0) __PYX_ERR(3, 42, __pyx_L1_error)

  /* "src/codec.pxi":43
 * 
 * HAVE_LZ4 = False
 * try:             # <<<<<<<<<<<<<<
 *     import lz4.frame
 *     HAVE_LZ4 = True
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "src/codec.pxi":44
 * HAVE_LZ4 = False
 * try:
 *     import lz4.frame             # <<<<<<<<<<<<<<
 *     HAVE_LZ4 = True
 * except ImportError:
 */
      __pyx_t_7 = __Pyx_Import(__pyx_n_s_lz4_frame, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 44, __pyx_L62_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_lz4, __pyx_t_7) < 0) __PYX_ERR(3, 44, __pyx_L62_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "src/codec.pxi":45
 * try:
 *     import lz4.frame
 *     HAVE_LZ4 = True             # <<<<<<<<<<<<<<
 * except ImportError:
 *     pass
 */
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_LZ4, Py_True) < 0) __PYX_ERR(3, 45, __pyx_L62_error)

      /* "src/codec.pxi":43
 * 
 * HAVE_LZ4 = False
 * try:             # <<<<<<<<<<<<<<
 *     import lz4.frame
 *     HAVE_LZ4 = True
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L67_try_end;
    __pyx_L62_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "src/codec.pxi":46
 *     import lz4.frame
 *     HAVE_LZ4 = True
 * except ImportError:             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_11) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L63_exception_handled;
    }
    goto __pyx_L64_except_error;

    /* "src/codec.pxi":43
 * 
 * HAVE_LZ4 = False
 * try:             # <<<<<<<<<<<<<<
 *     import lz4.frame
 *     HAVE_LZ4 = True
 */
    __pyx_L64_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L1_error;
    __pyx_L63_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    __pyx_L67_try_end:;
  }

  /* "src/codec.pxi":58
 *     """
 * 
 *     name = NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Codec, __pyx_n_s_name, __pyx_builtin_NotImplemented) < 0) __PYX_ERR(3, 58, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro_Codec);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_5Codec_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Codec___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__397)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Codec, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Codec);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Codec__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_5Codec_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Codec___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__398)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_Codec, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro_Codec);

  /* "src/codec.pxi":75
 *     """
 * 
 *     name = 'snappy'             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__SnappyCodec, __pyx_n_s_name, __pyx_n_u_snappy) < 0) __PYX_ERR(3, 75, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro__SnappyCodec);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_12_SnappyCodec_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SnappyCodec___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__399)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__SnappyCodec, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__SnappyCodec);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__SnappyCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__SnappyCodec__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_12_SnappyCodec_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SnappyCodec___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__400)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__SnappyCodec, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__SnappyCodec);

  /* "src/codec.pxi":97
 *     Implements the null codec (no compression)
 *     """
 *     name = 'null'             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__NullCodec, __pyx_n_s_name, __pyx_n_u_null_2) < 0) __PYX_ERR(3, 97, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro__NullCodec);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10_NullCodec_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NullCodec___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__401)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__NullCodec, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__NullCodec);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__NullCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__NullCodec__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10_NullCodec_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NullCodec___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__402)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__NullCodec, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__NullCodec);

  /* "src/codec.pxi":111
 *     Implements the deflate codec
 *     """
 *     name = 'deflate'             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__DeflateCodec, __pyx_n_s_name, __pyx_n_u_deflate) < 0) __PYX_ERR(3, 111, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro__DeflateCodec);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_13_DeflateCodec_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DeflateCodec___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__403)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__DeflateCodec, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__DeflateCodec);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__DeflateCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__DeflateCodec__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_13_DeflateCodec_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DeflateCodec___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__404)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__DeflateCodec, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__DeflateCodec);

  /* "src/codec.pxi":126
 *     Implements the bzip2 codec
 *     """
 *     name = 'bzip2'             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro___Bzip2Codec, __pyx_n_s_name, __pyx_n_u_bzip2) < 0) __PYX_ERR(3, 126, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro___Bzip2Codec);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_12__Bzip2Codec_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Bzip2Codec___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__405)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro___Bzip2Codec, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro___Bzip2Codec);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Bzip2Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Bzip2Codec__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_12__Bzip2Codec_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Bzip2Codec___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__406)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro___Bzip2Codec, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro___Bzip2Codec);

  /* "src/codec.pxi":141
 *     Implements the lzma codec
 *     """
 *     name = 'xz'             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__LzmaCodec, __pyx_n_s_name, __pyx_n_u_xz) < 0) __PYX_ERR(3, 141, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro__LzmaCodec);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10_LzmaCodec_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LzmaCodec___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__407)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__LzmaCodec, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__LzmaCodec);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__LzmaCodec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__LzmaCodec__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_10_LzmaCodec_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LzmaCodec___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__408)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__LzmaCodec, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__LzmaCodec);

  /* "src/codec.pxi":156
 *     Implements the zstandard codec if zstandard is installed.
 *     """
 *     name = 'zstandard'             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly object compressor
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__ZStandardCodec, __pyx_n_s_name, __pyx_n_u_zstandard) < 0) __PYX_ERR(3, 156, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro__ZStandardCodec);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15_ZStandardCodec_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ZStandardCodec___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__409)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__ZStandardCodec, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__ZStandardCodec);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__ZStandardCodec, (type(self), 0x8d57cb2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__ZStandardCodec__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15_ZStandardCodec_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ZStandardCodec___setstate_cytho, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__410)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__ZStandardCodec, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__ZStandardCodec);

  /* "src/codec.pxi":179
 *     Implements the lz4 codec if lz4 is installed.
 *     """
 *     name = 'lz4'             # <<<<<<<<<<<<<<
 * 
 *     cdef const uint8_t[:] read_block(self, _Reader reader, size_t length):
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__Lz4Codec, __pyx_n_s_name, __pyx_n_u_lz4) < 0) __PYX_ERR(3, 179, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_5cavro__Lz4Codec);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9_Lz4Codec_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Lz4Codec___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__411)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__Lz4Codec, __pyx_n_s_reduce_cython, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__Lz4Codec);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__Lz4Codec, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__Lz4Codec__set_state(self, __pyx_state)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_9_Lz4Codec_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Lz4Codec___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__412)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro__Lz4Codec, __pyx_n_s_setstate_cython, __pyx_t_7) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  PyType_Modified(__pyx_ptype_5cavro__Lz4Codec);

  /* "src/codec.pxi":191
 * 
 * CODECS = {
 *     b'null': _NullCodec()             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro__NullCodec)); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_b_null_2, __pyx_t_5) < 0) __PYX_ERR(3, 191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CODECS, __pyx_t_7) < 0) __PYX_ERR(3, 190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "src/codec.pxi":194
 * }
 * 
 * if HAVE_ZLIB:             # <<<<<<<<<<<<<<
 *     CODECS[b'deflate'] = _DeflateCodec()
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_HAVE_ZLIB); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_6) {

    /* "src/codec.pxi":195
 * 
 * if HAVE_ZLIB:
 *     CODECS[b'deflate'] = _DeflateCodec()             # <<<<<<<<<<<<<<
 * 
 * if HAVE_BZIP2:
 */
    __pyx_t_7 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro__DeflateCodec)); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_CODECS); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely((PyObject_SetItem(__pyx_t_5, __pyx_n_b_deflate, __pyx_t_7) < 0))) __PYX_ERR(3, 195, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "src/codec.pxi":194
 * }
 * 
 * if HAVE_ZLIB:             # <<<<<<<<<<<<<<
 *     CODECS[b'deflate'] = _DeflateCodec()
 * 
 */
  }

  /* "src/codec.pxi":197
 *     CODECS[b'deflate'] = _DeflateCodec()
 * 
 * if HAVE_BZIP2:             # <<<<<<<<<<<<<<
 *     CODECS[b'bzip2'] = __Bzip2Codec()
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_HAVE_BZIP2); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 197, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_6) {

    /* "src/codec.pxi":198
 * 
 * if HAVE_BZIP2:
 *     CODECS[b'bzip2'] = __Bzip2Codec()             # <<<<<<<<<<<<<<
 * 
 * if HAVE_XZ:
 */
    __pyx_t_7 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro___Bzip2Codec)); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_CODECS); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely((PyObject_SetItem(__pyx_t_5, __pyx_n_b_bzip2, __pyx_t_7) < 0))) __PYX_ERR(3, 198, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "src/codec.pxi":197
 *     CODECS[b'deflate'] = _DeflateCodec()
 * 
 * if HAVE_BZIP2:             # <<<<<<<<<<<<<<
 *     CODECS[b'bzip2'] = __Bzip2Codec()
 * 
 */
  }

  /* "src/codec.pxi":200
 *     CODECS[b'bzip2'] = __Bzip2Codec()
 * 
 * if HAVE_XZ:             # <<<<<<<<<<<<<<
 *     CODECS[b'xz'] = _LzmaCodec()
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_HAVE_XZ); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_6) {

    /* "src/codec.pxi":201
 * 
 * if HAVE_XZ:
 *     CODECS[b'xz'] = _LzmaCodec()             # <<<<<<<<<<<<<<
 * 
 * if HAVE_SNAPPY:
 */
    __pyx_t_7 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro__LzmaCodec)); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_CODECS); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely((PyObject_SetItem(__pyx_t_5, __pyx_n_b_xz, __pyx_t_7) < 0))) __PYX_ERR(3, 201, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "src/codec.pxi":200
 *     CODECS[b'bzip2'] = __Bzip2Codec()
 * 
 * if HAVE_XZ:             # <<<<<<<<<<<<<<
 *     CODECS[b'xz'] = _LzmaCodec()
 * 
 */
  }

  /* "src/codec.pxi":203
 *     CODECS[b'xz'] = _LzmaCodec()
 * 
 * if HAVE_SNAPPY:             # <<<<<<<<<<<<<<
 *     CODECS[b'snappy'] = _SnappyCodec()
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_HAVE_SNAPPY); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 203, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_6) {

    /* "src/codec.pxi":204
 * 
 * if HAVE_SNAPPY:
 *     CODECS[b'snappy'] = _SnappyCodec()             # <<<<<<<<<<<<<<
 * 
 * if HAVE_ZSTD:
 */
    __pyx_t_7 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro__SnappyCodec)); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_CODECS); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely((PyObject_SetItem(__pyx_t_5, __pyx_n_b_snappy, __pyx_t_7) < 0))) __PYX_ERR(3, 204, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "src/codec.pxi":203
 *     CODECS[b'xz'] = _LzmaCodec()
 * 
 * if HAVE_SNAPPY:             # <<<<<<<<<<<<<<
 *     CODECS[b'snappy'] = _SnappyCodec()
 * 
 */
  }

  /* "src/codec.pxi":206
 *     CODECS[b'snappy'] = _SnappyCodec()
 * 
 * if HAVE_ZSTD:             # <<<<<<<<<<<<<<
 *     CODECS[b'zstandard'] = _ZStandardCodec()
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_HAVE_ZSTD); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 206, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_6) {

    /* "src/codec.pxi":207
 * 
 * if HAVE_ZSTD:
 *     CODECS[b'zstandard'] = _ZStandardCodec()             # <<<<<<<<<<<<<<
 * 
 * if HAVE_LZ4:
 */
    __pyx_t_7 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro__ZStandardCodec)); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_CODECS); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely((PyObject_SetItem(__pyx_t_5, __pyx_n_b_zstandard, __pyx_t_7) < 0))) __PYX_ERR(3, 207, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "src/codec.pxi":206
 *     CODECS[b'snappy'] = _SnappyCodec()
 * 
 * if HAVE_ZSTD:             # <<<<<<<<<<<<<<
 *     CODECS[b'zstandard'] = _ZStandardCodec()
 * 
 */
  }

  /* "src/codec.pxi":209
 *     CODECS[b'zstandard'] = _ZStandardCodec()
 * 
 * if HAVE_LZ4:             # <<<<<<<<<<<<<<
 *     CODECS[b'lz4'] = _Lz4Codec()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_HAVE_LZ4); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(3, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_6) {

    /* "src/codec.pxi":210
 * 
 * if HAVE_LZ4:
 *     CODECS[b'lz4'] = _Lz4Codec()             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5cavro__Lz4Codec)); if (unlikely(!__pyx_t_7)) __PYX_ERR(3, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_CODECS); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely((PyObject_SetItem(__pyx_t_5, __pyx_n_b_lz4, __pyx_t_7) < 0))) __PYX_ERR(3, 210, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "src/codec.pxi":209
 *     CODECS[b'zstandard'] = _ZStandardCodec()
 * 
 * if HAVE_LZ4:             # <<<<<<<<<<<<<<
 *     CODECS[b'lz4'] = _Lz4Codec()
 */
  }

  /* "src/container.pxi":1
 * from pathlib import Path             # <<<<<<<<<<<<<<
 * import uuid
 * 
 */
  __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_Path);
  __Pyx_GIVEREF(__pyx_n_s_Path);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_Path)) __PYX_ERR(25, 1, __pyx_L1_error);
  __pyx_t_5 = __Pyx_Import(__pyx_n_s_pathlib, __pyx_t_7, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_Path); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Path, __pyx_t_7) < 0) __PYX_ERR(25, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/container.pxi":2
 * from pathlib import Path
 * import uuid             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __Pyx_ImportDottedModule(__pyx_n_s_uuid, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_uuid, __pyx_t_5) < 0) __PYX_ERR(25, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "src/container.pxi":5
 * 
 * 
 * OBJ_MAGIC_BYTES = b'Obj\x01'             # <<<<<<<<<<<<<<
 * cdef const uint8_t[:] OBJ_MAGIC = OBJ_MAGIC_BYTES
 * OBJ_FILE_METADATA = Schema({"type": "map", "values": "bytes"}, bytes_codec='utf8')
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OBJ_MAGIC_BYTES, __pyx_kp_b_Obj) < 0) __PYX_ERR(25, 5, __pyx_L1_error)

  /* "src/container.pxi":6
 * 
 * OBJ_MAGIC_BYTES = b'Obj\x01'
 * cdef const uint8_t[:] OBJ_MAGIC = OBJ_MAGIC_BYTES             # <<<<<<<<<<<<<<
 * OBJ_FILE_METADATA = Schema({"type": "map", "values": "bytes"}, bytes_codec='utf8')
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_OBJ_MAGIC_BYTES); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(__pyx_t_5, 0); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(25, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __PYX_XCLEAR_MEMVIEW(&__pyx_v_5cavro_OBJ_MAGIC, 1);
  __pyx_v_5cavro_OBJ_MAGIC = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "src/container.pxi":7
 * OBJ_MAGIC_BYTES = b'Obj\x01'
 * cdef const uint8_t[:] OBJ_MAGIC = OBJ_MAGIC_BYTES
 * OBJ_FILE_METADATA = Schema({"type": "map", "values": "bytes"}, bytes_codec='utf8')             # <<<<<<<<<<<<<<
 * 
 * _SLICE_TYPE = type(OBJ_MAGIC)
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_u_type, __pyx_n_u_map) < 0) __PYX_ERR(25, 7, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_u_values, __pyx_n_u_bytes_3) < 0) __PYX_ERR(25, 7, __pyx_L1_error)
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5)) __PYX_ERR(25, 7, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_bytes_codec, __pyx_n_u_utf8) < 0) __PYX_ERR(25, 7, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5cavro_Schema), __pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OBJ_FILE_METADATA, __pyx_t_4) < 0) __PYX_ERR(25, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/container.pxi":9
 * OBJ_FILE_METADATA = Schema({"type": "map", "values": "bytes"}, bytes_codec='utf8')
 * 
 * _SLICE_TYPE = type(OBJ_MAGIC)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(!__pyx_v_5cavro_OBJ_MAGIC.memview)) { __Pyx_RaiseUnboundLocalError("OBJ_MAGIC"); __PYX_ERR(25, 9, __pyx_L1_error) }
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_5cavro_OBJ_MAGIC, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_uint8_t__const__, (int (*)(char *, PyObject *)) NULL, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SLICE_TYPE, ((PyObject *)Py_TYPE(__pyx_t_4))) < 0) __PYX_ERR(25, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "src/container.pxi":69
 *     cdef _Reader reader
 * 
 *     def __init__(self, src, reader_schema=None, options=DEFAULT_OPTIONS):             # <<<<<<<<<<<<<<
 *         self.reader = make_reader(src)
 *         cdef const uint8_t[:] header = self.reader.read_n(4)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_DEFAULT_OPTIONS); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_k__87 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "src/container.pxi":95
 *         self.marker = b''
 * 
 *     cpdef _read_marker(self):             # <<<<<<<<<<<<<<
 *         value = self.reader.read_n(16)
 *         if not len(self.marker):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15ContainerReader_3_read_marker, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ContainerReader__read_marker, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__413)); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ContainerReader, __pyx_n_s_read_marker, __pyx_t_4) < 0) __PYX_ERR(25, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ContainerReader);

  /* "src/container.pxi":117
 *         self.current_block._reset_to(block_bytes)
 * 
 *     cpdef object next_object(self):             # <<<<<<<<<<<<<<
 *         while self.objects_left_in_block < 1:
 *             self.next_block()
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15ContainerReader_5next_object, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ContainerReader_next_object, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__414)); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ContainerReader, __pyx_n_s_next_object, __pyx_t_4) < 0) __PYX_ERR(25, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ContainerReader);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15ContainerReader_11__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ContainerReader___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__415)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ContainerReader, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ContainerReader);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ContainerReader, (type(self), 0x289807d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ContainerReader__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15ContainerReader_13__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ContainerReader___setstate_cytho, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__416)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ContainerReader, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ContainerReader);

  /* "src/container.pxi":171
 *     cdef readonly dict metadata
 * 
 *     def __cinit__(self, dest, Schema schema, str codec='null', size_t max_blocksize=16352, write_header=True, metadata=None, marker=None, options=DEFAULT_OPTIONS):             # <<<<<<<<<<<<<<
 *         if schema is None:
 *             raise ValueError('Schema is required')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_DEFAULT_OPTIONS); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_k__89 = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "src/container.pxi":207
 *             raise
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         if self.writer is None:
 *             raise ValueError('Container is closed')
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15ContainerWriter_3__enter__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ContainerWriter___enter, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__417)); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ContainerWriter, __pyx_n_s_enter, __pyx_t_4) < 0) __PYX_ERR(25, 207, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ContainerWriter);

  /* "src/container.pxi":212
 *         return self
 * 
 *     def __exit__(self, exc_type, exc_val, exc_tb):             # <<<<<<<<<<<<<<
 *         if self.writer is not None:
 *             self.close()
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15ContainerWriter_5__exit__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ContainerWriter___exit, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__419)); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ContainerWriter, __pyx_n_s_exit, __pyx_t_4) < 0) __PYX_ERR(25, 212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ContainerWriter);

  /* "src/container.pxi":238
 *         return 0
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         if self.writer is None:
 *             raise ValueError('Trying to close a closed Container')
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15ContainerWriter_9close, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ContainerWriter_close, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__421)); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ContainerWriter, __pyx_n_s_close, __pyx_t_4) < 0) __PYX_ERR(25, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ContainerWriter);

  /* "src/container.pxi":264
 *         return 0
 * 
 *     def flush(self, force=False):             # <<<<<<<<<<<<<<
 *         self._flush_block(force)
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15ContainerWriter_11flush, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ContainerWriter_flush, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__423)); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__424);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ContainerWriter, __pyx_n_s_flush, __pyx_t_4) < 0) __PYX_ERR(25, 264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ContainerWriter);

  /* "src/container.pxi":267
 *         self._flush_block(force)
 * 
 *     cpdef int write_one(self, obj) except -1:             # <<<<<<<<<<<<<<
 *         if self.next_item is None:
 *             raise ValueError('Trying to write to closed Container')
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15ContainerWriter_13write_one, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ContainerWriter_write_one, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__426)); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ContainerWriter, __pyx_n_s_write_one, __pyx_t_4) < 0) __PYX_ERR(25, 267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ContainerWriter);

  /* "src/container.pxi":284
 *                 self._flush_block()
 * 
 *     def write_many(self, objs):             # <<<<<<<<<<<<<<
 *         for obj in objs:
 *             self.write_one(obj)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15ContainerWriter_15write_many, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ContainerWriter_write_many, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__428)); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_ContainerWriter, __pyx_n_s_write_many, __pyx_t_4) < 0) __PYX_ERR(25, 284, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_ContainerWriter);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15ContainerWriter_17__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ContainerWriter___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__429)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_15ContainerWriter_19__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ContainerWriter___setstate_cytho, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__430)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_14PromoteToFloat_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PromoteToFloat___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__431)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_PromoteToFloat, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_PromoteToFloat);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromoteToFloat, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromoteToFloat__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_14PromoteToFloat_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PromoteToFloat___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__432)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_PromoteToFloat, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_PromoteToFloat);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_20PromoteBytesToString_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PromoteBytesToString___reduce_cy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__433)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_PromoteBytesToString, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_PromoteBytesToString);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromoteBytesToString, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromoteBytesToString__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_20PromoteBytesToString_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PromoteBytesToString___setstate, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__434)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_PromoteBytesToString, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_PromoteBytesToString);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_20PromoteStringToBytes_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PromoteStringToBytes___reduce_cy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__435)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_PromoteStringToBytes, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_PromoteStringToBytes);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PromoteStringToBytes, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PromoteStringToBytes__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_20PromoteStringToBytes_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PromoteStringToBytes___setstate, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__436)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_PromoteStringToBytes, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_PromoteStringToBytes);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_13CannotPromote_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CannotPromote___reduce_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__437)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_CannotPromote, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_CannotPromote);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CannotPromote, (type(self), 0xcbfa7ba, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CannotPromote__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_13CannotPromote_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CannotPromote___setstate_cython, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__438)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_5cavro_CannotPromote, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_5cavro_CannotPromote);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Options(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_19__pyx_unpickle_Options, 0, __pyx_n_s_pyx_unpickle_Options, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__439)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Options, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Options__set_state(<Options> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Options__set_state(Options __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.adapt_record_types = __pyx_state[0]; __pyx_result.allow_aliases_to_be_string = __pyx_state[1]; __pyx_result.allow_duplicate_union_types = __pyx_state[2]; __pyx_result.allow_empty_unions = __pyx_state[3]; __pyx_result.allow_error_type = __pyx_state[4]; __pyx_result.allow_false_values_for_null = __pyx_state[5]; __pyx_result.allow_invalid_default_values = __pyx_state[6]; __pyx_result.allow_leading_dot_in_names = __pyx_state[7]; __pyx_result.allow_nested_unions = __pyx_state[8]; __pyx_result.allow_primitive_name_collision = __pyx_state[9]; __pyx_result.allow_primitive_names_in_namespaces = __pyx_state[10]; __pyx_result.allow_tuple_notation = __pyx_state[11]; __pyx_result.allow_union_default_any_member = __pyx_state[12]; __pyx_result.alternate_timestamp_millis_encoding = __pyx_state[13]; __pyx_result.ascii_name_rules = __pyx_state[14]; __pyx_result.bytes_codec = __pyx_state[15]; __pyx_result.bytes_default_value_utf8 = __pyx_state[16]; __pyx_result.canonical_form_repeat_enum = __pyx_state[17]; __pyx_result.canonical_form_repeat_fixed = __pyx_state[18]; __pyx_result.clamp_float_overflow = __pyx_state[19]; __pyx_result.clamp_int_overflow = __pyx_state[20]; __pyx_result.coerce_int_to_float = __pyx_state[21]; __pyx_result.coerce_values_to_boolean = __pyx_state[22]; __pyx_result.coerce_values_to_float = __pyx_state[23]; __pyx_result.coerce_values_to_int = __pyx_state[24]; __pyx_result.coerce_values_to_str = __pyx_state[25]; __pyx_result.container_fill_blocks = __pyx_state[26]; __pyx_result.date_type_accepts_string = __pyx_state[27]; __pyx_result.decimal_check_exp_overflow = __pyx_state[28]; __pyx_result.defer_schema_promotion_errors = __pyx_state[29]; __pyx_result.enforce_enum_symbol_name_rules = __pyx_state[30]; __pyx_result.enforce_namespace_name_rules = __pyx_state[31]; __pyx_result.enforce_type_name_rules = __pyx_state[32]; __pyx_result.enum_symbols_must_be_unique = __pyx_state[33]; __pyx_result.expand_types_in_schema = __pyx_state[34]; __pyx_result.externally_defined_types = __pyx_state[35]; __pyx_result.fingerprint_returns_digest = __pyx_state[36]; __pyx_result.fixed_codec = __pyx_state[37]; __pyx_result.float_out_of_range_inf = __pyx_state[38]; __pyx_result.inline_namespaces = __pyx_state[39]; __pyx_result.invalid_value_include_array_index = __pyx_state[40]; __pyx_result.invalid_value_include_map_key = __pyx_state[41]; __pyx_result.invalid_value_includes_record_name = __pyx_state[42]; __pyx_result.logical_types = __pyx_state[43]; __pyx_result.missing_values_can_be_empty_container = __pyx_state[44]; __pyx_result.missing_values_can_be_null = __pyx_state[45]; __pyx_result.naive_dt_assume_utc = __pyx_state[46]; __pyx_result.named_type_names_must_be_unique = __pyx_state[47]; __pyx_result.null_pad_fixed = __pyx_state[48]; __pyx_result.raise_on_invalid_logical = __pyx_state[49]; __pyx_result.record_allow_extra_fields = __pyx_state[50]; __pyx_result.record_can_encode_dict = __pyx_state[51]; __pyx_result.record_decodes_to_dict = __pyx_state[52]; __pyx_result.record_encode_use_defaults = __pyx_state[53]; __pyx_result.record_fields_must_be_unique = __pyx_state[54]; __pyx_result.record_values_type_hint = __pyx_state[55]; __pyx_result.return_uuid_object = __pyx_state[56]; __pyx_result.string_types_default_unchanged = __pyx_state[57]; __pyx_result.truncate_fixed = __pyx_state[58]; __pyx_result.types_str_to_schema = __pyx_state[59]; __pyx_result.unicode_errors = __pyx_state[60]; __pyx_result.union_decodes_to = __pyx_state[61]; __pyx_result.union_json_encodes_type_name = __pyx_state[62]
 *     if len(__pyx_state) > 63 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_21__pyx_unpickle__Writer, 0, __pyx_n_s_pyx_unpickle__Writer, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__440)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__Writer, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__Reader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_23__pyx_unpickle__Reader, 0, __pyx_n_s_pyx_unpickle__Reader, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__441)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__Reader, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle__Reader__set_state(<_Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__Reader__set_state(_Reader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_25__pyx_unpickle_FileReader, 0, __pyx_n_s_pyx_unpickle_FileReader, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__442)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_FileReader, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FileWriter(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_27__pyx_unpickle_FileWriter, 0, __pyx_n_s_pyx_unpickle_FileWriter, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__443)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_FileWriter, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_FileWriter__set_state(<FileWriter> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FileWriter__set_state(FileWriter __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.file_obj = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_29__pyx_unpickle_MemoryReader, 0, __pyx_n_s_pyx_unpickle_MemoryReader, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__444)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_MemoryReader, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Rabin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_31__pyx_unpickle_Rabin, 0, __pyx_n_s_pyx_unpickle_Rabin, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__445)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Rabin, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Rabin__set_state(<Rabin> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Rabin__set_state(Rabin __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.value = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_33__pyx_unpickle_ArrayType, 0, __pyx_n_s_pyx_unpickle_ArrayType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__446)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ArrayType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_EnumType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_35__pyx_unpickle_EnumType, 0, __pyx_n_s_pyx_unpickle_EnumType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__447)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_EnumType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_EnumType__set_state(<EnumType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_EnumType__set_state(EnumType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.default_value = __pyx_state[1]; __pyx_result.doc = __pyx_state[2]; __pyx_result.effective_namespace = __pyx_state[3]; __pyx_result.metadata = __pyx_state[4]; __pyx_result.name = __pyx_state[5]; __pyx_result.namespace = __pyx_state[6]; __pyx_result.options = __pyx_state[7]; __pyx_result.symbol_indexes = __pyx_state[8]; __pyx_result.symbols = __pyx_state[9]; __pyx_result.value_adapters = __pyx_state[10]
 *     if len(__pyx_state) > 11 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_37__pyx_unpickle_PromotingEnumType, 0, __pyx_n_s_pyx_unpickle_PromotingEnumType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__448)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PromotingEnumType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_MapType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_39__pyx_unpickle_MapType, 0, __pyx_n_s_pyx_unpickle_MapType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__449)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_MapType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_MapType__set_state(<MapType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MapType__set_state(MapType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.key_type = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.value_adapters = __pyx_state[3]; __pyx_result.value_type = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_41__pyx_unpickle_NullType, 0, __pyx_n_s_pyx_unpickle_NullType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__450)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_NullType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_BoolType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_43__pyx_unpickle_BoolType, 0, __pyx_n_s_pyx_unpickle_BoolType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__451)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_BoolType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_BoolType__set_state(<BoolType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BoolType__set_state(BoolType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_45__pyx_unpickle_IntType, 0, __pyx_n_s_pyx_unpickle_IntType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__452)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IntType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LongType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_47__pyx_unpickle_LongType, 0, __pyx_n_s_pyx_unpickle_LongType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__453)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_LongType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_LongType__set_state(<LongType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LongType__set_state(LongType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_49__pyx_unpickle_FloatType, 0, __pyx_n_s_pyx_unpickle_FloatType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__454)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_FloatType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_DoubleType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_51__pyx_unpickle_DoubleType, 0, __pyx_n_s_pyx_unpickle_DoubleType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__455)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_DoubleType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_DoubleType__set_state(<DoubleType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DoubleType__set_state(DoubleType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_53__pyx_unpickle__PlaceholderType, 0, __pyx_n_s_pyx_unpickle__PlaceholderType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__456)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__PlaceholderType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Record(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_55__pyx_unpickle_Record, 0, __pyx_n_s_pyx_unpickle_Record, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__457)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Record, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Record__set_state(<Record> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Record__set_state(Record __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.data = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_57__pyx_unpickle_RecordField, 0, __pyx_n_s_pyx_unpickle_RecordField, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__458)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_RecordField, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__FieldAccessor(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_59__pyx_unpickle__FieldAccessor, 0, __pyx_n_s_pyx_unpickle__FieldAccessor, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__459)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__FieldAccessor, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle__FieldAccessor__set_state(<_FieldAccessor> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__FieldAccessor__set_state(_FieldAccessor __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.index = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_61__pyx_unpickle_RecordType, 0, __pyx_n_s_pyx_unpickle_RecordType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__460)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_RecordType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PromotingRecordType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_63__pyx_unpickle_PromotingRecordType, 0, __pyx_n_s_pyx_unpickle_PromotingRecordTy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__461)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PromotingRecordTy, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_PromotingRecordType__set_state(<PromotingRecordType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromotingRecordType__set_state(PromotingRecordType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._setting_up = __pyx_state[0]; __pyx_result.aliases = __pyx_state[1]; __pyx_result.decode_indexes = __pyx_state[2]; __pyx_result.doc = __pyx_state[3]; __pyx_result.effective_namespace = __pyx_state[4]; __pyx_result.field_dict = __pyx_state[5]; __pyx_result.fields = __pyx_state[6]; __pyx_result.metadata = __pyx_state[7]; __pyx_result.name = __pyx_state[8]; __pyx_result.namespace = __pyx_state[9]; __pyx_result.options = __pyx_state[10]; __pyx_result.record = __pyx_state[11]; __pyx_result.value_adapters = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_65__pyx_unpickle_BytesType, 0, __pyx_n_s_pyx_unpickle_BytesType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__462)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_BytesType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_StringType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_67__pyx_unpickle_StringType, 0, __pyx_n_s_pyx_unpickle_StringType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__463)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_StringType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_StringType__set_state(<StringType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_StringType__set_state(StringType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.metadata = __pyx_state[0]; __pyx_result.options = __pyx_state[1]; __pyx_result.value_adapters = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_69__pyx_unpickle_FixedType, 0, __pyx_n_s_pyx_unpickle_FixedType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__464)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_FixedType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_UnionType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_71__pyx_unpickle_UnionType, 0, __pyx_n_s_pyx_unpickle_UnionType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__465)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_UnionType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_UnionType__set_state(<UnionType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UnionType__set_state(UnionType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.by_name_map = __pyx_state[0]; __pyx_result.metadata = __pyx_state[1]; __pyx_result.options = __pyx_state[2]; __pyx_result.return_type_tuple = __pyx_state[3]; __pyx_result.union_types = __pyx_state[4]; __pyx_result.value_adapters = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_73__pyx_unpickle__Sentinel, 0, __pyx_n_s_pyx_unpickle__Sentinel, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__466)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__Sentinel, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__CanonicalForm(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_75__pyx_unpickle__CanonicalForm, 0, __pyx_n_s_pyx_unpickle__CanonicalForm, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__467)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__CanonicalForm, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle__CanonicalForm__set_state(<_CanonicalForm> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__CanonicalForm__set_state(_CanonicalForm __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_77__pyx_unpickle_AvroType, 0, __pyx_n_s_pyx_unpickle_AvroType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__468)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_AvroType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__NamedType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_79__pyx_unpickle__NamedType, 0, __pyx_n_s_pyx_unpickle__NamedType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__469)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__NamedType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle__NamedType__set_state(<_NamedType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__NamedType__set_state(_NamedType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aliases = __pyx_state[0]; __pyx_result.effective_namespace = __pyx_state[1]; __pyx_result.metadata = __pyx_state[2]; __pyx_result.name = __pyx_state[3]; __pyx_result.namespace = __pyx_state[4]; __pyx_result.options = __pyx_state[5]; __pyx_result.value_adapters = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_81__pyx_unpickle_Schema, 0, __pyx_n_s_pyx_unpickle_Schema, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__470)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Schema, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ResolvedSchema(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_83__pyx_unpickle_ResolvedSchema, 0, __pyx_n_s_pyx_unpickle_ResolvedSchema, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__471)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ResolvedSchema, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_ResolvedSchema__set_state(<ResolvedSchema> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ResolvedSchema__set_state(ResolvedSchema __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._canonical_form = __pyx_state[0]; __pyx_result.logical_types = __pyx_state[1]; __pyx_result.named_types = __pyx_state[2]; __pyx_result.options = __pyx_state[3]; __pyx_result.source = __pyx_state[4]; __pyx_result.type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_85__pyx_unpickle_ValueAdapter, 0, __pyx_n_s_pyx_unpickle_ValueAdapter, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__472)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ValueAdapter, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LogicalType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_87__pyx_unpickle_LogicalType, 0, __pyx_n_s_pyx_unpickle_LogicalType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__473)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_LogicalType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_LogicalType__set_state(<LogicalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LogicalType__set_state(LogicalType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_89__pyx_unpickle_CustomLogicalType, 0, __pyx_n_s_pyx_unpickle_CustomLogicalType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__474)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_CustomLogicalType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_DecimalType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_91__pyx_unpickle_DecimalType, 0, __pyx_n_s_pyx_unpickle_DecimalType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__475)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_DecimalType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_DecimalType__set_state(<DecimalType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DecimalType__set_state(DecimalType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.check_exp_overflow = __pyx_state[0]; __pyx_result.context = __pyx_state[1]; __pyx_result.precision = __pyx_state[2]; __pyx_result.scale = __pyx_state[3]; __pyx_result.scale_val = __pyx_state[4]; __pyx_result.size = __pyx_state[5]; __pyx_result.underlying_name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_93__pyx_unpickle_UUIDBase, 0, __pyx_n_s_pyx_unpickle_UUIDBase, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__476)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_UUIDBase, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_UUIDStringType(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_95__pyx_unpickle_UUIDStringType, 0, __pyx_n_s_pyx_unpickle_UUIDStringType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__477)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_UUIDStringType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_UUIDStringType__set_state(<UUIDStringType> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_UUIDStringType__set_state(UUIDStringType __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.return_uuid_object = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_97__pyx_unpickle_UUIDFixedType, 0, __pyx_n_s_pyx_unpickle_UUIDFixedType, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__478)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_UUIDFixedType, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Date(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_99__pyx_unpickle_Date, 0, __pyx_n_s_pyx_unpickle_Date, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__479)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Date, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Date__set_state(<Date> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Date__set_state(Date __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.accepts_string = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_101__pyx_unpickle_TimeMillis, 0, __pyx_n_s_pyx_unpickle_TimeMillis, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__480)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_TimeMillis, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TimeMicros(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_103__pyx_unpickle_TimeMicros, 0, __pyx_n_s_pyx_unpickle_TimeMicros, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__481)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_TimeMicros, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_TimeMicros__set_state(<TimeMicros> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TimeMicros__set_state(TimeMicros __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_105__pyx_unpickle_TimestampMillis, 0, __pyx_n_s_pyx_unpickle_TimestampMillis, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__482)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_TimestampMillis, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TimestampMicros(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_107__pyx_unpickle_TimestampMicros, 0, __pyx_n_s_pyx_unpickle_TimestampMicros, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__483)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_TimestampMicros, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_TimestampMicros__set_state(<TimestampMicros> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TimestampMicros__set_state(TimestampMicros __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_109__pyx_unpickle_Codec, 0, __pyx_n_s_pyx_unpickle_Codec, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__484)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Codec, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__SnappyCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_111__pyx_unpickle__SnappyCodec, 0, __pyx_n_s_pyx_unpickle__SnappyCodec, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__485)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__SnappyCodec, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle__SnappyCodec__set_state(<_SnappyCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__SnappyCodec__set_state(_SnappyCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_113__pyx_unpickle__NullCodec, 0, __pyx_n_s_pyx_unpickle__NullCodec, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__486)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__NullCodec, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__DeflateCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_115__pyx_unpickle__DeflateCodec, 0, __pyx_n_s_pyx_unpickle__DeflateCodec, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__487)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__DeflateCodec, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle__DeflateCodec__set_state(<_DeflateCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__DeflateCodec__set_state(_DeflateCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_117__pyx_unpickle___Bzip2Codec, 0, __pyx_n_s_pyx_unpickle___Bzip2Codec, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__488)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle___Bzip2Codec, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__LzmaCodec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_119__pyx_unpickle__LzmaCodec, 0, __pyx_n_s_pyx_unpickle__LzmaCodec, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__489)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__LzmaCodec, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle__LzmaCodec__set_state(<_LzmaCodec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__LzmaCodec__set_state(_LzmaCodec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_121__pyx_unpickle__ZStandardCodec, 0, __pyx_n_s_pyx_unpickle__ZStandardCodec, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__490)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__ZStandardCodec, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__Lz4Codec(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_123__pyx_unpickle__Lz4Codec, 0, __pyx_n_s_pyx_unpickle__Lz4Codec, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__491)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__Lz4Codec, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle__Lz4Codec__set_state(<_Lz4Codec> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__Lz4Codec__set_state(_Lz4Codec __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_125__pyx_unpickle_ContainerReader, 0, __pyx_n_s_pyx_unpickle_ContainerReader, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__492)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ContainerReader, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PromoteToFloat(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_127__pyx_unpickle_PromoteToFloat, 0, __pyx_n_s_pyx_unpickle_PromoteToFloat, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__493)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PromoteToFloat, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_PromoteToFloat__set_state(<PromoteToFloat> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteToFloat__set_state(PromoteToFloat __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_129__pyx_unpickle_PromoteBytesToString, 0, __pyx_n_s_pyx_unpickle_PromoteBytesToStr, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__494)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PromoteBytesToStr, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PromoteStringToBytes(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_131__pyx_unpickle_PromoteStringToBytes, 0, __pyx_n_s_pyx_unpickle_PromoteStringToBy, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__495)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PromoteStringToBy, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_PromoteStringToBytes__set_state(<PromoteStringToBytes> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PromoteStringToBytes__set_state(PromoteStringToBytes __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5cavro_133__pyx_unpickle_CannotPromote, 0, __pyx_n_s_pyx_unpickle_CannotPromote, NULL, __pyx_n_s_cavro, __pyx_d, ((PyObject *)__pyx_codeobj__496)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_CannotPromote, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "cavro.pyx":2
 * 
 * """             # <<<<<<<<<<<<<<
 * `cavro` is a library for encoding and decoding data in the Apache Avro format.
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(13, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_2, __pyx_t_4) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XCLEAR_MEMVIEW(&__pyx_t_12, 1);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init cavro", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init cavro");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;  // error
            return kwvalues[i];
        }
    }
    return NULL;  // not found (no exception set)
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value); // transfer ownership of value to values
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;  // ownership transferred to values
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL; // ownership transferred to values
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if Py_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSsize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    CYTHON_MAYBE_UNUSED_VAR(max_char);
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030D0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    CYTHON_UNUSED_VAR(value_count);
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (meth) {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* DivInt[Py_ssize_t] */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t a, Py_ssize_t b) {
    Py_ssize_t q = a / b;
    Py_ssize_t r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
#endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportDottedModule */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__3;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PySequenceMultiply */
static PyObject* __Pyx_PySequence_Multiply_Generic(PyObject *seq, Py_ssize_t mul) {
    PyObject *result, *pymul = PyInt_FromSsize_t(mul);
    if (unlikely(!pymul))
        return NULL;
    result = PyNumber_Multiply(seq, pymul);
    Py_DECREF(pymul);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PySequence_Multiply(PyObject *seq, Py_ssize_t mul) {
#if CYTHON_USE_TYPE_SLOTS
    PyTypeObject *type = Py_TYPE(seq);
    if (likely(type->tp_as_sequence && type->tp_as_sequence->sq_repeat)) {
        return type->tp_as_sequence->sq_repeat(seq, mul);
    } else
#endif
    {
        return __Pyx_PySequence_Multiply_Generic(seq, mul);
    }
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (unlikely(!j)) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_ass_subscript) {
            int r;
            PyObject *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return -1;
            r = mm->mp_ass_subscript(o, key, v);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return sm->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* RaiseUnboundLocalError */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* DivInt[long] */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__2);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    PyObject *abi_module = PyImport_AddModule((char*) __PYX_ABI_MODULE_NAME);
    if (unlikely(!abi_module)) return NULL;
    Py_INCREF(abi_module);
    return abi_module;
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = (long)((unsigned long)a + (unsigned long)b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op2);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* Globals */
static PyObject* __Pyx_Globals(void) {
    return __Pyx_NewRef(__pyx_d);
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* UnpackUnboundCMethod */
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    PyObject *result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
     METH_VARARGS | METH_KEYWORDS,
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #else
    if (likely(!__Pyx_CyOrPyCFunction_Check(method)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            target->method = unbound_method;
        }
    }
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if ((PY_VERSION_HEX >= 0x030600B1) && flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            #else
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            #endif
        } else if ((PY_VERSION_HEX >= 0x030700A0) && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* UnicodeConcatInPlace */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
static int
__Pyx_unicode_modifiable(PyObject *unicode)
{
    if (Py_REFCNT(unicode) != 1)
        return 0;
    if (!PyUnicode_CheckExact(unicode))
        return 0;
    if (PyUnicode_CHECK_INTERNED(unicode))
        return 0;
    return 1;
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    ) {
    PyObject *left = *p_left;
    Py_ssize_t left_len, right_len, new_len;
    if (unlikely(__Pyx_PyUnicode_READY(left) == -1))
        return NULL;
    if (unlikely(__Pyx_PyUnicode_READY(right) == -1))
        return NULL;
    left_len = PyUnicode_GET_LENGTH(left);
    if (left_len == 0) {
        Py_INCREF(right);
        return right;
    }
    right_len = PyUnicode_GET_LENGTH(right);
    if (right_len == 0) {
        Py_INCREF(left);
        return left;
    }
    if (unlikely(left_len > PY_SSIZE_T_MAX - right_len)) {
        PyErr_SetString(PyExc_OverflowError,
                        "strings are too large to concat");
        return NULL;
    }
    new_len = left_len + right_len;
    if (__Pyx_unicode_modifiable(left)
            && PyUnicode_CheckExact(right)
            && PyUnicode_KIND(right) <= PyUnicode_KIND(left)
            && !(PyUnicode_IS_ASCII(left) && !PyUnicode_IS_ASCII(right))) {
        __Pyx_GIVEREF(*p_left);
        if (unlikely(PyUnicode_Resize(p_left, new_len) != 0)) {
            __Pyx_GOTREF(*p_left);
            return NULL;
        }
        __Pyx_INCREF(*p_left);
        #if PY_VERSION_HEX >= 0x030D0000
        if (unlikely(PyUnicode_CopyCharacters(*p_left, left_len, right, 0, right_len) < 0)) return NULL;
        #else
        _PyUnicode_FastCopyCharacters(*p_left, left_len, right, 0, right_len);
        #endif
        return *p_left;
    } else {
        return __Pyx_PyUnicode_Concat(left, right);
    }
  }
#endif

/* UnicodeAsUCS4 */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject* x) {
   Py_ssize_t length;
   #if CYTHON_PEP393_ENABLED
   length = PyUnicode_GET_LENGTH(x);
   if (likely(length == 1)) {
       return PyUnicode_READ_CHAR(x, 0);
   }
   #else
   length = PyUnicode_GET_SIZE(x);
   if (likely(length == 1)) {
       return PyUnicode_AS_UNICODE(x)[0];
   }
   #if Py_UNICODE_SIZE == 2
   else if (PyUnicode_GET_SIZE(x) == 2) {
       Py_UCS4 high_val = PyUnicode_AS_UNICODE(x)[0];
       if (high_val >= 0xD800 && high_val <= 0xDBFF) {
           Py_UCS4 low_val = PyUnicode_AS_UNICODE(x)[1];
           if (low_val >= 0xDC00 && low_val <= 0xDFFF) {
               return 0x10000 + (((high_val & ((1<<10)-1)) << 10) | (low_val & ((1<<10)-1)));
           }
       }
   }
   #endif
   #endif
   PyErr_Format(PyExc_ValueError,
                "only single character unicode strings can be converted to Py_UCS4, "
                "got length %" CYTHON_FORMAT_SSIZE_T "d", length);
   return (Py_UCS4)-1;
}

/* object_ord */
static long __Pyx__PyObject_Ord(PyObject* c) {
    Py_ssize_t size;
    if (PyBytes_Check(c)) {
        size = PyBytes_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyBytes_AS_STRING(c)[0];
        }
#if PY_MAJOR_VERSION < 3
    } else if (PyUnicode_Check(c)) {
        return (long)__Pyx_PyUnicode_AsPy_UCS4(c);
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    } else if (PyByteArray_Check(c)) {
        size = PyByteArray_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyByteArray_AS_STRING(c)[0];
        }
#endif
    } else {
        __Pyx_TypeName c_type_name = __Pyx_PyType_GetName(Py_TYPE(c));
        PyErr_Format(PyExc_TypeError,
            "ord() expected string of length 1, but " __Pyx_FMT_TYPENAME " found",
            c_type_name);
        __Pyx_DECREF_TypeName(c_type_name);
        return (long)(Py_UCS4)-1;
    }
    PyErr_Format(PyExc_TypeError,
        "ord() expected a character, but string of length %zd found", size);
    return (long)(Py_UCS4)-1;
}

/* BufferIndexError */
static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        
            return PyInt_FromLong(a & b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if ((intval & PyLong_MASK) == intval) {
            long last_digit = (long) __Pyx_PyLong_Digits(op1)[0];
            long result = intval & (likely(__Pyx_PyLong_IsPos(op1)) ? last_digit : (PyLong_MASK - last_digit + 1));
            return PyLong_FromLong(result);
        }
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op1);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_and(op1, op2);
            }
        }
                x = a & b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla & llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceAnd : PyNumber_And)(op1, op2);
}
#endif

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is unsliceable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return NULL;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        
#ifdef HAVE_LONG_LONG
            if (sizeof(PY_LONG_LONG) > sizeof(long)) {
                PY_LONG_LONG result = (PY_LONG_LONG)a * (PY_LONG_LONG)b;
                return (result >= LONG_MIN && result <= LONG_MAX) ?
                    PyInt_FromLong((long)result) : PyLong_FromLongLong(result);
            }
#endif
#if CYTHON_USE_TYPE_SLOTS
            return PyInt_Type.tp_as_number->nb_multiply(op1, op2);
#else
            return PyNumber_Multiply(op1, op2);
#endif
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op1);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
            }
        }
                CYTHON_UNUSED_VAR(a);
                CYTHON_UNUSED_VAR(b);
                #ifdef HAVE_LONG_LONG
                lla = a;
                goto long_long;
                #else
                return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
                #endif
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla * llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("multiply", return NULL)
            result = ((double)a) * (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceMultiply : PyNumber_Multiply)(op1, op2);
}
#endif

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* PyObjectLookupSpecial */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return with_error ? __Pyx_PyObject_GetAttrStr(obj, attr_name) : __Pyx_PyObject_GetAttrStrNoError(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_GeneratorType_slots
};
#else
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Generator_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_GeneratorType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
#endif
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* GeneratorYieldFrom */
#if CYTHON_USE_TYPE_SLOTS
static void __Pyx_PyIter_CheckErrorAndDecref(PyObject *source) {
    __Pyx_TypeName source_type_name = __Pyx_PyType_GetName(Py_TYPE(source));
    PyErr_Format(PyExc_TypeError,
        "iter() returned non-iterator of type '" __Pyx_FMT_TYPENAME "'", source_type_name);
    __Pyx_DECREF_TypeName(source_type_name);
    Py_DECREF(source);
}
#endif
static CYTHON_INLINE PyObject* __Pyx_Generator_Yield_From(__pyx_CoroutineObject *gen, PyObject *source) {
    PyObject *source_gen, *retval;
#ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(source)) {
        Py_INCREF(source);
        source_gen = source;
        retval = __Pyx_Generator_Next(source);
    } else
#endif
    {
#if CYTHON_USE_TYPE_SLOTS
        if (likely(Py_TYPE(source)->tp_iter)) {
            source_gen = Py_TYPE(source)->tp_iter(source);
            if (unlikely(!source_gen))
                return NULL;
            if (unlikely(!PyIter_Check(source_gen))) {
                __Pyx_PyIter_CheckErrorAndDecref(source_gen);
                return NULL;
            }
        } else
#endif
        {
            source_gen = PyObject_GetIter(source);
            if (unlikely(!source_gen))
                return NULL;
        }
        retval = __Pyx_PyObject_GetIterNextFunc(source_gen)(source_gen);
    }
    if (likely(retval)) {
        gen->yieldfrom = source_gen;
        return retval;
    }
    Py_DECREF(source_gen);
    return NULL;
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* RaiseClosureNameError */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* IterNext */
static PyObject *__Pyx_PyIter_Next2Default(PyObject* defval) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (!defval || unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return NULL;
        __Pyx_PyErr_Clear();
        Py_INCREF(defval);
        return defval;
    }
    if (defval) {
        Py_INCREF(defval);
        return defval;
    }
    __Pyx_PyErr_SetNone(PyExc_StopIteration);
    return NULL;
}
static void __Pyx_PyIter_Next_ErrorNoIterator(PyObject *iterator) {
    __Pyx_TypeName iterator_type_name = __Pyx_PyType_GetName(Py_TYPE(iterator));
    PyErr_Format(PyExc_TypeError,
        __Pyx_FMT_TYPENAME " object is not an iterator", iterator_type_name);
    __Pyx_DECREF_TypeName(iterator_type_name);
}
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject* iterator, PyObject* defval) {
    PyObject* next;
    iternextfunc iternext = Py_TYPE(iterator)->tp_iternext;
    if (likely(iternext)) {
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
        next = iternext(iterator);
        if (likely(next))
            return next;
#if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(iternext == &_PyObject_NextNotImplemented))
            return NULL;
#endif
#else
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
#endif
    } else if (CYTHON_USE_TYPE_SLOTS || unlikely(!PyIter_Check(iterator))) {
        __Pyx_PyIter_Next_ErrorNoIterator(iterator);
        return NULL;
    }
#if !CYTHON_USE_TYPE_SLOTS
    else {
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
    }
#endif
    return __Pyx_PyIter_Next2Default(defval);
}

/* pybytes_as_double */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj) {
    PyObject *float_value;
#if PY_MAJOR_VERSION >= 3
    float_value = PyFloat_FromString(obj);
#else
    float_value = PyFloat_FromString(obj, 0);
#endif
    if (likely(float_value)) {
        double value = PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
    return (double)-1;
}
static const char* __Pyx__PyBytes_AsDouble_Copy(const char* start, char* buffer, Py_ssize_t length) {
    int last_was_punctuation = 1;
    Py_ssize_t i;
    for (i=0; i < length; i++) {
        char chr = start[i];
        int is_punctuation = (chr == '_') | (chr == '.') | (chr == 'e') | (chr == 'E');
        *buffer = chr;
        buffer += (chr != '_');
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyBytes_AsDouble_inf_nan(const char* start, Py_ssize_t length) {
    int matches = 1;
    char sign = start[0];
    int is_signed = (sign == '+') | (sign == '-');
    start += is_signed;
    length -= is_signed;
    switch (start[0]) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            matches &= (start[1] == 'a' || start[1] == 'A');
            matches &= (start[2] == 'n' || start[2] == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            matches &= (start[1] == 'n' || start[1] == 'N');
            matches &= (start[2] == 'f' || start[2] == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            matches &= (start[3] == 'i' || start[3] == 'I');
            matches &= (start[4] == 'n' || start[4] == 'N');
            matches &= (start[5] == 'i' || start[5] == 'I');
            matches &= (start[6] == 't' || start[6] == 'T');
            matches &= (start[7] == 'y' || start[7] == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static CYTHON_INLINE int __Pyx__PyBytes_AsDouble_IsSpace(char ch) {
    return (ch == 0x20) | !((ch < 0x9) | (ch > 0xd));
}
CYTHON_UNUSED static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length) {
    double value;
    Py_ssize_t i, digits;
    const char *last = start + length;
    char *end;
    while (__Pyx__PyBytes_AsDouble_IsSpace(*start))
        start++;
    while (start < last - 1 && __Pyx__PyBytes_AsDouble_IsSpace(last[-1]))
        last--;
    length = last - start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyBytes_AsDouble_inf_nan(start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    digits = 0;
    for (i=0; i < length; digits += start[i++] != '_');
    if (likely(digits == length)) {
        value = PyOS_string_to_double(start, &end, NULL);
    } else if (digits < 40) {
        char number[40];
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((digits + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}

/* pynumber_float */
static CYTHON_INLINE PyObject* __Pyx__PyNumber_Float(PyObject* obj) {
    double val;
    if (PyLong_CheckExact(obj)) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (likely(__Pyx_PyLong_IsCompact(obj))) {
            val = (double) __Pyx_PyLong_CompactValue(obj);
            goto no_error;
        }
#endif
        val = PyLong_AsDouble(obj);
    } else if (PyUnicode_CheckExact(obj)) {
        val = __Pyx_PyUnicode_AsDouble(obj);
    } else if (PyBytes_CheckExact(obj)) {
        val = __Pyx_PyBytes_AsDouble(obj);
    } else if (PyByteArray_CheckExact(obj)) {
        val = __Pyx_PyByteArray_AsDouble(obj);
    } else {
        return PyNumber_Float(obj);
    }
    if (unlikely(val == -1 && PyErr_Occurred())) {
        return NULL;
    }
#if CYTHON_USE_PYLONG_INTERNALS
no_error:
#endif
    return PyFloat_FromDouble(val);
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_keys */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_keys, d);
    else
        return PyDict_Keys(d);
}

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* py_set_discard_unhashable */
static int __Pyx_PySet_DiscardUnhashable(PyObject *set, PyObject *key) {
    PyObject *tmpkey;
    int rv;
    if (likely(!PySet_Check(key) || !PyErr_ExceptionMatches(PyExc_TypeError)))
        return -1;
    PyErr_Clear();
    tmpkey = __Pyx_PyFrozenSet_New(key);
    if (tmpkey == NULL)
        return -1;
    rv = PySet_Discard(set, tmpkey);
    Py_DECREF(tmpkey);
    return rv;
}

/* py_set_discard */
static CYTHON_INLINE int __Pyx_PySet_Discard(PyObject *set, PyObject *key) {
    int found = PySet_Discard(set, key);
    if (unlikely(found < 0)) {
        found = __Pyx_PySet_DiscardUnhashable(set, key);
    }
    return found;
}

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    CYTHON_UNUSED_VAR(is_set);
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        CYTHON_UNUSED_VAR(orig_length);
        CYTHON_UNUSED_VAR(ppos);
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* py_dict_items */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_items, d);
    else
        return PyDict_Items(d);
}

/* SliceTupleAndList */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_crop_slice(Py_ssize_t* _start, Py_ssize_t* _stop, Py_ssize_t* _length) {
    Py_ssize_t start = *_start, stop = *_stop, length = *_length;
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    *_length = stop - start;
    *_start = start;
    *_stop = stop;
}
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    Py_ssize_t length = PyList_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (length <= 0) {
        return PyList_New(0);
    }
    return __Pyx_PyList_FromArray(((PyListObject*)src)->ob_item + start, length);
}
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    Py_ssize_t length = PyTuple_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    return __Pyx_PyTuple_FromArray(((PyTupleObject*)src)->ob_item + start, length);
}
#endif

/* dict_setdefault */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value,
                                                       int is_safe_type) {
    PyObject* value;
    CYTHON_MAYBE_UNUSED_VAR(is_safe_type);
#if PY_VERSION_HEX >= 0x030400A0
    if ((1)) {
        value = PyDict_SetDefault(d, key, default_value);
        if (unlikely(!value)) return NULL;
        Py_INCREF(value);
#else
    if (is_safe_type == 1 || (is_safe_type == -1 &&
#if PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
            (PyUnicode_CheckExact(key) || PyString_CheckExact(key) || PyLong_CheckExact(key)))) {
        value = PyDict_GetItemWithError(d, key);
        if (unlikely(!value)) {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            if (unlikely(PyDict_SetItem(d, key, default_value) == -1))
                return NULL;
            value = default_value;
        }
        Py_INCREF(value);
#else
            (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key) || PyLong_CheckExact(key)))) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            if (unlikely(PyDict_SetItem(d, key, default_value) == -1))
                return NULL;
            value = default_value;
        }
        Py_INCREF(value);
#endif
#endif
    } else {
        value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_setdefault, d, key, default_value);
    }
    return value;
}

/* append */
    static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = (long)((unsigned long)a - (unsigned long)b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return PyLong_FromLong(-intval);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        const long a = intval;
        long b = PyInt_AS_LONG(op2);
        
#ifdef HAVE_LONG_LONG
            if (sizeof(PY_LONG_LONG) > sizeof(long)) {
                PY_LONG_LONG result = (PY_LONG_LONG)a * (PY_LONG_LONG)b;
                return (result >= LONG_MIN && result <= LONG_MAX) ?
                    PyInt_FromLong((long)result) : PyLong_FromLongLong(result);
            }
#endif
#if CYTHON_USE_TYPE_SLOTS
            return PyInt_Type.tp_as_number->nb_multiply(op1, op2);
#else
            return PyNumber_Multiply(op1, op2);
#endif
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op2))) {
        const long a = intval;
        long b, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG lla = intval;
        PY_LONG_LONG llb, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op2))) {
            return __Pyx_NewRef(op2);
        }
        if (likely(__Pyx_PyLong_IsCompact(op2))) {
            b = __Pyx_PyLong_CompactValue(op2);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op2);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op2);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        b = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        llb = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        b = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        llb = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        b = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        llb = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        b = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        llb = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        b = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        llb = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        b = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        llb = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
            }
        }
                CYTHON_UNUSED_VAR(a);
                CYTHON_UNUSED_VAR(b);
                #ifdef HAVE_LONG_LONG
                llb = b;
                goto long_long;
                #else
                return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
                #endif
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla * llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op2)) {
        const long a = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double b = __pyx_PyFloat_AsDouble(op2);
#else
        double b = PyFloat_AS_DOUBLE(op2);
#endif
            double result;
            
            PyFPE_START_PROTECT("multiply", return NULL)
            result = ((double)a) * (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceMultiply : PyNumber_Multiply)(op1, op2);
}
#endif

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_FloorDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            if (unlikely(b == -1 && ((unsigned long)a) == 0-(unsigned long)a))
                return PyInt_Type.tp_as_number->nb_floor_divide(op1, op2);
            else {
                long q, r;
                q = a / b;
                r = a - q*b;
                q -= ((r != 0) & ((r ^ b) < 0));
                x = q;
            }
            return PyInt_FromLong(x);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op1);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_floor_divide(op1, op2);
            }
        }
                {
                    long q, r;
                    q = a / b;
                    r = a - q*b;
                    q -= ((r != 0) & ((r ^ b) < 0));
                    x = q;
                }
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                {
                    PY_LONG_LONG q, r;
                    q = lla / llb;
                    r = lla - q*llb;
                    q -= ((r != 0) & ((r ^ llb) < 0));
                    llx = q;
                }
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceFloorDivide : PyNumber_FloorDivide)(op1, op2);
}
#endif

/* PyIntCompare */
    static CYTHON_INLINE int __Pyx_PyInt_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 0;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        return (a != b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) != 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 1;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 1;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal != 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        return ((double)a != (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_RemainderObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = a % b;
            x += ((x != 0) & ((x ^ b) < 0)) * b;
            return PyInt_FromLong(x);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op1);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_remainder(op1, op2);
            }
        }
                x = a % b;
                x += ((x != 0) & ((x ^ b) < 0)) * b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla % llb;
                llx += ((llx != 0) & ((llx ^ llb) < 0)) * llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceRemainder : PyNumber_Remainder)(op1, op2);
}
#endif

/* WriteUnraisableException */
    static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* CallNextTpTraverse */
    static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    traverseproc tp_traverse = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc) != current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_traverse = __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc)) == current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_traverse)
        return tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
    static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    inquiry tp_clear = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_clear, inquiry) != current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_clear = __Pyx_PyType_GetSlot(type, tp_clear, inquiry)) == current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_clear)
        tp_clear(obj);
}

/* PyObject_GenericGetAttr */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* ValidateBasesTuple */
    #if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#if !CYTHON_USE_TYPE_SLOTS
        if (dictoffset == 0) {
            PyErr_Format(PyExc_TypeError,
                "extension type '%s.200s': "
                "unable to validate whether bases have a __dict__ "
                "when CYTHON_USE_TYPE_SLOTS is off "
                "(likely because you are building in the limited API). "
                "Therefore, all extension types with multiple bases "
                "must add 'cdef dict __dict__' in this compilation mode",
                type_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#else
        if (dictoffset == 0 && b->tp_dictoffset)
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "extension type '%.200s' has no __dict__ slot, "
                "but base type '" __Pyx_FMT_TYPENAME "' has: "
                "either add 'cdef dict __dict__' to the extension type "
                "or add '__slots__ = [...]' to the base type",
                type_name, b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
    static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetVTable */
    static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
    static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
    #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* SetupReduce */
    #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* FormatTypeName */
    #if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name_2);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__122);
    }
    return name;
}
#endif

/* ValidateExternBase */
    static int __Pyx_validate_extern_base(PyTypeObject *base) {
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_itemsize;
#endif
#if !CYTHON_COMPILING_IN_LIMITED_API
    itemsize = ((PyTypeObject *)base)->tp_itemsize;
#else
    py_itemsize = PyObject_GetAttrString((PyObject*)base, "__itemsize__");
    if (!py_itemsize)
        return -1;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        return -1;
#endif
    if (itemsize) {
        __Pyx_TypeName b_name = __Pyx_PyType_GetName(base);
        PyErr_Format(PyExc_TypeError,
                "inheritance from PyVarObject types like '" __Pyx_FMT_TYPENAME "' not currently supported", b_name);
        __Pyx_DECREF_TypeName(b_name);
        return -1;
    }
    return 0;
}

/* TypeImport */
    #ifndef __PYX_HAVE_RT_ImportType_3_0_4
#define __PYX_HAVE_RT_ImportType_3_0_4
static PyTypeObject *__Pyx_ImportType_3_0_4(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_4 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_4 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_4 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* PatchInspect */
    static PyObject* __Pyx_patch_inspect(PyObject* module) {
#if defined(__Pyx_Generator_USED) && (!defined(CYTHON_PATCH_INSPECT) || CYTHON_PATCH_INSPECT)
    static int inspect_patched = 0;
    if (unlikely((!inspect_patched) && module)) {
        module = __Pyx_Coroutine_patch_module(
            module, ""
"old_types = getattr(_module.isgenerator, '_cython_generator_types', None)\n"
"if old_types is None or not isinstance(old_types, set):\n"
"    old_types = set()\n"
"    def cy_wrap(orig_func, type=type, cython_generator_types=old_types):\n"
"        def cy_isgenerator(obj): return type(obj) in cython_generator_types or orig_func(obj)\n"
"        cy_isgenerator._cython_generator_types = cython_generator_types\n"
"        return cy_isgenerator\n"
"    _module.isgenerator = cy_wrap(_module.isgenerator)\n"
"old_types.add(_cython_generator_type)\n"
        );
        inspect_patched = 1;
    }
#else
    if ((0)) return __Pyx_Coroutine_patch_module(module, NULL);
#endif
    return module;
}

/* Py3UpdateBases */
    static PyObject*
__Pyx_PEP560_update_bases(PyObject *bases)
{
    Py_ssize_t i, j, size_bases;
    PyObject *base, *meth, *new_base, *result, *new_bases = NULL;
    size_bases = PyTuple_GET_SIZE(bases);
    for (i = 0; i < size_bases; i++) {
        base  = PyTuple_GET_ITEM(bases, i);
        if (PyType_Check(base)) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        meth = __Pyx_PyObject_GetAttrStrNoError(base, __pyx_n_s_mro_entries);
        if (!meth && PyErr_Occurred()) {
            goto error;
        }
        if (!meth) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        new_base = __Pyx_PyObject_CallOneArg(meth, bases);
        Py_DECREF(meth);
        if (!new_base) {
            goto error;
        }
        if (!PyTuple_Check(new_base)) {
            PyErr_SetString(PyExc_TypeError,
                            "__mro_entries__ must return a tuple");
            Py_DECREF(new_base);
            goto error;
        }
        if (!new_bases) {
            if (!(new_bases = PyList_New(i))) {
                goto error;
            }
            for (j = 0; j < i; j++) {
                base = PyTuple_GET_ITEM(bases, j);
                PyList_SET_ITEM(new_bases, j, base);
                Py_INCREF(base);
            }
        }
        j = PyList_GET_SIZE(new_bases);
        if (PyList_SetSlice(new_bases, j, j, new_base) < 0) {
            goto error;
        }
        Py_DECREF(new_base);
    }
    if (!new_bases) {
        Py_INCREF(bases);
        return bases;
    }
    result = PyList_AsTuple(new_bases);
    Py_DECREF(new_bases);
    return result;
error:
    Py_XDECREF(new_bases);
    return NULL;
}

/* CalculateMetaclass */
    static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases;
#if CYTHON_ASSUME_SAFE_MACROS
    nbases = PyTuple_GET_SIZE(bases);
#else
    nbases = PyTuple_Size(bases);
    if (nbases < 0) return NULL;
#endif
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
#if CYTHON_ASSUME_SAFE_MACROS
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *tmp = PyTuple_GetItem(bases, i);
        if (!tmp) return NULL;
#endif
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
    static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
#if PY_VERSION_HEX >= 0x03030000
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
#else
    CYTHON_MAYBE_UNUSED_VAR(qualname);
#endif
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc_2, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
static int __Pyx_SetNamesPEP487(PyObject *type_obj) {
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *names_to_set, *key, *value, *set_name, *tmp;
    Py_ssize_t i = 0;
#if CYTHON_USE_TYPE_SLOTS
    names_to_set = PyDict_Copy(type->tp_dict);
#else
    {
        PyObject *d = PyObject_GetAttr(type_obj, __pyx_n_s_dict);
        names_to_set = NULL;
        if (likely(d)) {
            PyObject *names_to_set = PyDict_New();
            int ret = likely(names_to_set) ? PyDict_Update(names_to_set, d) : -1;
            Py_DECREF(d);
            if (unlikely(ret < 0))
                Py_CLEAR(names_to_set);
        }
    }
#endif
    if (unlikely(names_to_set == NULL))
        goto bad;
    while (PyDict_Next(names_to_set, &i, &key, &value)) {
        set_name = __Pyx_PyObject_LookupSpecialNoError(value, __pyx_n_s_set_name);
        if (unlikely(set_name != NULL)) {
            tmp = __Pyx_PyObject_Call2Args(set_name, type_obj, key);
            Py_DECREF(set_name);
            if (unlikely(tmp == NULL)) {
                __Pyx_TypeName value_type_name =
                    __Pyx_PyType_GetName(Py_TYPE(value));
                __Pyx_TypeName type_name = __Pyx_PyType_GetName(type);
                PyErr_Format(PyExc_RuntimeError,
#if PY_MAJOR_VERSION >= 3
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %R " "in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name, key, type_name);
#else
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %.100s in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name,
                    PyString_Check(key) ? PyString_AS_STRING(key) : "?",
                    type_name);
#endif
                goto bad;
            } else {
                Py_DECREF(tmp);
            }
        }
        else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
    Py_DECREF(names_to_set);
    return 0;
bad:
    Py_XDECREF(names_to_set);
    return -1;
}
static PyObject *__Pyx_InitSubclassPEP487(PyObject *type_obj, PyObject *mkw) {
#if CYTHON_USE_TYPE_SLOTS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *mro = type->tp_mro;
    Py_ssize_t i, nbases;
    if (unlikely(!mro)) goto done;
    (void) &__Pyx_GetBuiltinName;
    Py_INCREF(mro);
    nbases = PyTuple_GET_SIZE(mro);
    assert(PyTuple_GET_ITEM(mro, 0) == type_obj);
    for (i = 1; i < nbases-1; i++) {
        PyObject *base, *dict, *meth;
        base = PyTuple_GET_ITEM(mro, i);
        dict = ((PyTypeObject *)base)->tp_dict;
        meth = __Pyx_PyDict_GetItemStrWithError(dict, __pyx_n_s_init_subclass);
        if (unlikely(meth)) {
            descrgetfunc f = Py_TYPE(meth)->tp_descr_get;
            PyObject *res;
            Py_INCREF(meth);
            if (likely(f)) {
                res = f(meth, NULL, type_obj);
                Py_DECREF(meth);
                if (unlikely(!res)) goto bad;
                meth = res;
            }
            res = __Pyx_PyObject_FastCallDict(meth, NULL, 0, mkw);
            Py_DECREF(meth);
            if (unlikely(!res)) goto bad;
            Py_DECREF(res);
            goto done;
        } else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
done:
    Py_XDECREF(mro);
    return type_obj;
bad:
    Py_XDECREF(mro);
    Py_DECREF(type_obj);
    return NULL;
#else
    PyObject *super_type, *super, *func, *res;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    super_type = __Pyx_GetBuiltinName(__pyx_n_s_super);
#else
    super_type = (PyObject*) &PySuper_Type;
    (void) &__Pyx_GetBuiltinName;
#endif
    super = likely(super_type) ? __Pyx_PyObject_Call2Args(super_type, type_obj, type_obj) : NULL;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    Py_XDECREF(super_type);
#endif
    if (unlikely(!super)) {
        Py_CLEAR(type_obj);
        goto done;
    }
    func = __Pyx_PyObject_GetAttrStrNoError(super, __pyx_n_s_init_subclass);
    Py_DECREF(super);
    if (likely(!func)) {
        if (unlikely(PyErr_Occurred()))
            Py_CLEAR(type_obj);
        goto done;
    }
    res = __Pyx_PyObject_FastCallDict(func, NULL, 0, mkw);
    Py_DECREF(func);
    if (unlikely(!res))
        Py_CLEAR(type_obj);
    Py_XDECREF(res);
done:
    return type_obj;
#endif
}
#endif
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET,
#if PY_VERSION_HEX < 0x030600A4
        (metaclass == (PyObject*)&PyType_Type) ? NULL : mkw
#else
        mkw
#endif
    );
    Py_XDECREF(owned_metaclass);
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
    if (likely(result) && likely(PyType_Check(result))) {
        if (unlikely(__Pyx_SetNamesPEP487(result) < 0)) {
            Py_CLEAR(result);
        } else {
            result = __Pyx_InitSubclassPEP487(result, mkw);
        }
    }
#else
    (void) &__Pyx_GetBuiltinName;
#endif
    return result;
}

/* CyFunctionClassCell */
    static int __Pyx_CyFunction_InitClassCell(PyObject *cyfunctions, PyObject *classobj) {
    Py_ssize_t i, count = PyList_GET_SIZE(cyfunctions);
    for (i = 0; i < count; i++) {
        __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *)
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyList_GET_ITEM(cyfunctions, i);
#else
            PySequence_ITEM(cyfunctions, i);
        if (unlikely(!m))
            return -1;
#endif
        __Pyx_CyFunction_SetClassObj(m, classobj);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF((PyObject*)m);
#endif
    }
    return 0;
}

/* DataclassesCallHelper */
    #if PY_MAJOR_VERSION >= 3
static int __Pyx_DataclassesCallHelper_FilterToDict(PyObject *callable, PyObject *kwds, PyObject *new_kwds, PyObject *args_list, int is_kwonly) {
    Py_ssize_t size, i;
    size = PySequence_Size(args_list);
    if (size == -1) return -1;
    for (i=0; i<size; ++i) {
        PyObject *key, *value;
        int setitem_result;
        key = PySequence_GetItem(args_list, i);
        if (!key) return -1;
        if (PyUnicode_Check(key) && (
                PyUnicode_CompareWithASCIIString(key, "self") == 0 ||
                PyUnicode_CompareWithASCIIString(key, "_cls") == 0)) {
            Py_DECREF(key);
            continue;
        }
        value = PyDict_GetItem(kwds, key);
        if (!value) {
            if (is_kwonly) {
                Py_DECREF(key);
                continue;
            } else {
                value = Py_None;
                if (PyErr_WarnFormat(
                        PyExc_RuntimeWarning, 1,
                        "Argument %S not passed to %R. This is likely a bug in Cython so please report it.",
                        key, callable) == -1) {
                    Py_DECREF(key);
                    return -1;
                }
            }
        }
        Py_INCREF(value);
        setitem_result = PyDict_SetItem(new_kwds, key, value);
        Py_DECREF(key);
        Py_DECREF(value);
        if (setitem_result == -1) return -1;
    }
    return 0;
}
#endif
static PyObject* __Pyx_DataclassesCallHelper(PyObject *callable, PyObject *kwds) {
#if PY_MAJOR_VERSION < 3
    return PyObject_Call(callable, __pyx_empty_tuple, kwds);
#else
    PyObject *new_kwds=NULL, *result=NULL;
    PyObject *inspect;
    PyObject *args_list=NULL, *kwonly_args_list=NULL, *getfullargspec_result=NULL;
    inspect = PyImport_ImportModule("inspect");
    if (!inspect) goto bad;
    getfullargspec_result = PyObject_CallMethodObjArgs(inspect, __pyx_kp_u_getfullargspec, callable, NULL);
    Py_DECREF(inspect);
    if (!getfullargspec_result) goto bad;
    args_list = PyObject_GetAttrString(getfullargspec_result, "args");
    if (!args_list) goto bad;
    kwonly_args_list = PyObject_GetAttrString(getfullargspec_result, "kwonlyargs");
    if (!kwonly_args_list) goto bad;
    new_kwds = PyDict_New();
    if (!new_kwds) goto bad;
    if (__Pyx_DataclassesCallHelper_FilterToDict(callable, kwds, new_kwds, args_list, 0) == -1) goto bad;
    if (__Pyx_DataclassesCallHelper_FilterToDict(callable, kwds, new_kwds, kwonly_args_list, 1) == -1) goto bad;
    result = PyObject_Call(callable, __pyx_empty_tuple, new_kwds);
bad:
    Py_XDECREF(getfullargspec_result);
    Py_XDECREF(args_list);
    Py_XDECREF(kwonly_args_list);
    Py_XDECREF(new_kwds);
    return result;
#endif
}

/* ModuleLoader */
    static PyObject* __Pyx_LoadInternalModule(const char* name, const char* fallback_code) {
    PyObject *shared_abi_module = 0, *module = 0;
    shared_abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!shared_abi_module) return NULL;
    if (PyObject_HasAttrString(shared_abi_module, name)) {
        PyObject* result = PyObject_GetAttrString(shared_abi_module, name);
        Py_DECREF(shared_abi_module);
        return result;
    }
    module = PyImport_ImportModule(name);
    if (!module) {
        PyObject *localDict, *runValue, *builtins, *modulename;
        if (!PyErr_ExceptionMatches(PyExc_ImportError)) goto bad;
        PyErr_Clear();  // this is reasonably likely (especially on older versions of Python)
#if PY_MAJOR_VERSION < 3
        modulename = PyBytes_FromFormat("_cython_" CYTHON_ABI ".%s", name);
#else
        modulename = PyUnicode_FromFormat("_cython_" CYTHON_ABI ".%s", name);
#endif
        if (!modulename) goto bad;
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_CPYTHON
        module = PyImport_AddModuleObject(modulename); // borrowed
#else
        module = PyImport_AddModule(PyBytes_AsString(modulename)); // borrowed
#endif
        Py_DECREF(modulename);
        if (!module) goto bad;
        Py_INCREF(module);
        if (PyObject_SetAttrString(shared_abi_module, name, module) < 0) goto bad;
        localDict = PyModule_GetDict(module); // borrowed
        if (!localDict) goto bad;
        builtins = PyEval_GetBuiltins(); // borrowed
        if (!builtins) goto bad;
        if (PyDict_SetItemString(localDict, "__builtins__", builtins) <0) goto bad;
        runValue = PyRun_String(fallback_code, Py_file_input, localDict, localDict);
        if (!runValue) goto bad;
        Py_DECREF(runValue);
    }
    goto shared_cleanup;
    bad:
        Py_CLEAR(module);
    shared_cleanup:
        Py_XDECREF(shared_abi_module);
    return module;
}

/* SpecificModuleLoader */
    static PyObject* __Pyx_Load_dataclasses_Module(void) {
    return __Pyx_LoadInternalModule("dataclasses", "\n\n\n\n\n\n\n\nfrom collections import namedtuple\ntry:\n    from types import MappingProxyType\nexcept ImportError:\n\n    MappingProxyType = lambda x: x\n\nclass _MISSING_TYPE(object):\n    pass\nMISSING = _MISSING_TYPE()\n\n_DataclassParams = namedtuple('_DataclassParams',\n    [\"init\", \"repr\", \"eq\", \"order\", \"unsafe_hash\", \"frozen\",\n     \"match_args\", \"kw_only\", \"slots\", \"weakref_slot\"])\nclass Field(object):\n    __slots__ = ('name',\n                 'type',\n                 'default',\n                 'default_factory',\n                 'repr',\n                 'hash',\n                 'init',\n                 'compare',\n                 'metadata',\n                 'kw_only',\n                 '_field_type',  # Private: not to be used by user code.\n                 )\n\n    def __init__(self, default, default_factory, init, repr, hash, compare,\n                 metadata, kw_only):\n        self.name = None\n        self.type = None\n        self.default = default\n        self.default_factory = default_factory\n        self.init = init\n        self.repr = repr\n        self.hash = hash\n        self.compare = compare\n\n\n        self.metadata = (MappingProxyType({})\n                         if metadata is None else\n                         MappingProxyType(metadata))\n        self.kw_only = kw_only\n        self._field_type = None\n\n    def __repr__(self):\n        return ('Field('\n                'name={0!r},'\n                'type={1!r},'\n                'default={2!r},'\n                'default_factory={3!r},'\n                'init={4!r},'\n                'repr={5!r},'\n                'hash={6!r},'\n                'compare={7!r},'\n                'metadata={8!r},'\n                'kwonly={9!r},'\n                ')'.format(self.name, self.type, self.default,\n                           self.default_factory, self.init,\n                           self.repr, self.hash, self.compare,\n                           ""self.metadata, self.kw_only))\n\n\n\n\nclass _HAS_DEFAULT_FACTORY_CLASS:\n    def __repr__(self):\n        return '<factory>'\n_HAS_DEFAULT_FACTORY = _HAS_DEFAULT_FACTORY_CLASS()\n\ndef dataclass(*args, **kwds):\n    raise NotImplementedError(\"Standard library 'dataclasses' module\"\n        \"is unavailable, likely due to the version of Python you're using.\")\n\n\nclass _FIELD_BASE:\n    def __init__(self, name):\n        self.name = name\n    def __repr__(self):\n        return self.name\n_FIELD = _FIELD_BASE('_FIELD')\n_FIELD_CLASSVAR = _FIELD_BASE('_FIELD_CLASSVAR')\n_FIELD_INITVAR = _FIELD_BASE('_FIELD_INITVAR')\n\ndef field(*ignore, **kwds):\n    default = kwds.pop(\"default\", MISSING)\n    default_factory = kwds.pop(\"default_factory\", MISSING)\n    init = kwds.pop(\"init\", True)\n    repr = kwds.pop(\"repr\", True)\n    hash = kwds.pop(\"hash\", None)\n    compare = kwds.pop(\"compare\", True)\n    metadata = kwds.pop(\"metadata\", None)\n    kw_only = kwds.pop(\"kw_only\", None)\n\n    if kwds:\n        raise ValueError(\"field received unexpected keyword arguments: %s\"\n                         % list(kwds.keys()))\n    if default is not MISSING and default_factory is not MISSING:\n        raise ValueError('cannot specify both default and default_factory')\n    if ignore:\n        raise ValueError(\"'field' does not take any positional arguments\")\n    return Field(default, default_factory, init,\n                 repr, hash, compare, metadata, kw_only)");
}

/* GetNameInClass */
    static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    PyObject *dict;
    assert(PyType_Check(nmspace));
#if CYTHON_USE_TYPE_SLOTS
    dict = ((PyTypeObject*)nmspace)->tp_dict;
    Py_XINCREF(dict);
#else
    dict = PyObject_GetAttr(nmspace, __pyx_n_s_dict);
#endif
    if (likely(dict)) {
        result = PyObject_GetItem(dict, name);
        Py_DECREF(dict);
        if (result) {
            return result;
        }
    }
    PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}

/* ClassMethod */
    static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    #if PY_MAJOR_VERSION == 2
    static PyTypeObject *methoddescr_type = NULL;
    if (unlikely(methoddescr_type == NULL)) {
       PyObject *meth = PyObject_GetAttrString((PyObject*)&PyList_Type, "append");
       if (unlikely(!meth)) return NULL;
       methoddescr_type = Py_TYPE(meth);
       Py_DECREF(meth);
    }
    #else
    PyTypeObject *methoddescr_type = &PyMethodDescr_Type;
    #endif
    if (__Pyx_TypeCheck(method, methoddescr_type))
#endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        #if PY_VERSION_HEX < 0x03020000
        PyTypeObject *d_type = descr->d_type;
        #else
        PyTypeObject *d_type = descr->d_common.d_type;
        #endif
        return PyDescr_NewClassMethod(d_type, descr->d_method);
    }
#endif
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else {
        return PyClassMethod_New(method);
    }
}

/* CLineInTraceback */
    #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
    #if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
    #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    PyErr_Clear();
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    __Pyx_TypeName obj_type_name;
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_ptype_7cpython_5array_array)) return __pyx_pw_7cpython_5array_5array_1__getbuffer__(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
                 "'" __Pyx_FMT_TYPENAME "' does not have the buffer interface",
                 obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
        else if (__Pyx_TypeCheck(obj, __pyx_ptype_7cpython_5array_array)) __pyx_pw_7cpython_5array_5array_3__releasebuffer__(obj, view);
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


    /* MemviewSliceIsContig */
    static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
    static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* MemviewDtypeToObject */
    static CYTHON_INLINE PyObject *__pyx_memview_get_nn_uint8_t__const__(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_uint8_t(*(uint8_t const  *) itemp);
}

/* IsLittleEndian */
    static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
    static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparsable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, int is_complex) {
  CYTHON_UNUSED_VAR(is_complex);
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, int is_complex) {
  CYTHON_UNUSED_VAR(is_complex);
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static int
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return -1;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return -1;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return -1;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i]) {
            PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
            return -1;
        }
        if (*ts != ',' && *ts != ')') {
            PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
            return -1;
        }
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim) {
        PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
        return -1;
    }
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return -1;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return 0;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (__pyx_buffmt_parse_array(ctx, &ts) < 0) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* TypeInfoCompare */
      static int
__pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b)
{
    int i;
    if (!a || !b)
        return 0;
    if (a == b)
        return 1;
    if (a->size != b->size || a->typegroup != b->typegroup ||
            a->is_unsigned != b->is_unsigned || a->ndim != b->ndim) {
        if (a->typegroup == 'H' || b->typegroup == 'H') {
            return a->size == b->size;
        } else {
            return 0;
        }
    }
    if (a->ndim) {
        for (i = 0; i < a->ndim; i++)
            if (a->arraysize[i] != b->arraysize[i])
                return 0;
    }
    if (a->typegroup == 'S') {
        if (a->flags != b->flags)
            return 0;
        if (a->fields || b->fields) {
            if (!(a->fields && b->fields))
                return 0;
            for (i = 0; a->fields[i].type && b->fields[i].type; i++) {
                __Pyx_StructField *field_a = a->fields + i;
                __Pyx_StructField *field_b = b->fields + i;
                if (field_a->offset != field_b->offset ||
                    !__pyx_typeinfo_cmp(field_a->type, field_b->type))
                    return 0;
            }
            return !a->fields[i].type && !b->fields[i].type;
        }
    }
    return 1;
}

/* MemviewSliceValidateAndInit */
      static int
__pyx_check_strides(Py_buffer *buf, int dim, int ndim, int spec)
{
    if (buf->shape[dim] <= 1)
        return 1;
    if (buf->strides) {
        if (spec & __Pyx_MEMVIEW_CONTIG) {
            if (spec & (__Pyx_MEMVIEW_PTR|__Pyx_MEMVIEW_FULL)) {
                if (unlikely(buf->strides[dim] != sizeof(void *))) {
                    PyErr_Format(PyExc_ValueError,
                                 "Buffer is not indirectly contiguous "
                                 "in dimension %d.", dim);
                    goto fail;
                }
            } else if (unlikely(buf->strides[dim] != buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
        if (spec & __Pyx_MEMVIEW_FOLLOW) {
            Py_ssize_t stride = buf->strides[dim];
            if (stride < 0)
                stride = -stride;
            if (unlikely(stride < buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
    } else {
        if (unlikely(spec & __Pyx_MEMVIEW_CONTIG && dim != ndim - 1)) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not contiguous in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(spec & (__Pyx_MEMVIEW_PTR))) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not indirect in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(buf->suboffsets)) {
            PyErr_SetString(PyExc_ValueError,
                            "Buffer exposes suboffsets but no strides");
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_check_suboffsets(Py_buffer *buf, int dim, int ndim, int spec)
{
    CYTHON_UNUSED_VAR(ndim);
    if (spec & __Pyx_MEMVIEW_DIRECT) {
        if (unlikely(buf->suboffsets && buf->suboffsets[dim] >= 0)) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer not compatible with direct access "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    if (spec & __Pyx_MEMVIEW_PTR) {
        if (unlikely(!buf->suboffsets || (buf->suboffsets[dim] < 0))) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer is not indirectly accessible "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_verify_contig(Py_buffer *buf, int ndim, int c_or_f_flag)
{
    int i;
    if (c_or_f_flag & __Pyx_IS_F_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = 0; i < ndim; i++) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not fortran contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    } else if (c_or_f_flag & __Pyx_IS_C_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = ndim - 1; i >- 1; i--) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not C contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    }
    return 1;
fail:
    return 0;
}
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj)
{
    struct __pyx_memoryview_obj *memview, *new_memview;
    __Pyx_RefNannyDeclarations
    Py_buffer *buf;
    int i, spec = 0, retval = -1;
    __Pyx_BufFmt_Context ctx;
    int from_memoryview = __pyx_memoryview_check(original_obj);
    __Pyx_RefNannySetupContext("ValidateAndInit_memviewslice", 0);
    if (from_memoryview && __pyx_typeinfo_cmp(dtype, ((struct __pyx_memoryview_obj *)
                                                            original_obj)->typeinfo)) {
        memview = (struct __pyx_memoryview_obj *) original_obj;
        new_memview = NULL;
    } else {
        memview = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                            original_obj, buf_flags, 0, dtype);
        new_memview = memview;
        if (unlikely(!memview))
            goto fail;
    }
    buf = &memview->view;
    if (unlikely(buf->ndim != ndim)) {
        PyErr_Format(PyExc_ValueError,
                "Buffer has wrong number of dimensions (expected %d, got %d)",
                ndim, buf->ndim);
        goto fail;
    }
    if (new_memview) {
        __Pyx_BufFmt_Init(&ctx, stack, dtype);
        if (unlikely(!__Pyx_BufFmt_CheckString(&ctx, buf->format))) goto fail;
    }
    if (unlikely((unsigned) buf->itemsize != dtype->size)) {
        PyErr_Format(PyExc_ValueError,
                     "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "u byte%s) "
                     "does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "u byte%s)",
                     buf->itemsize,
                     (buf->itemsize > 1) ? "s" : "",
                     dtype->name,
                     dtype->size,
                     (dtype->size > 1) ? "s" : "");
        goto fail;
    }
    if (buf->len > 0) {
        for (i = 0; i < ndim; i++) {
            spec = axes_specs[i];
            if (unlikely(!__pyx_check_strides(buf, i, ndim, spec)))
                goto fail;
            if (unlikely(!__pyx_check_suboffsets(buf, i, ndim, spec)))
                goto fail;
        }
        if (unlikely(buf->strides && !__pyx_verify_contig(buf, ndim, c_or_f_flag)))
            goto fail;
    }
    if (unlikely(__Pyx_init_memviewslice(memview, ndim, memviewslice,
                                         new_memview != NULL) == -1)) {
        goto fail;
    }
    retval = 0;
    goto no_fail;
fail:
    Py_XDECREF(new_memview);
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}

/* ObjectToMemviewSlice */
      static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn_uint8_t__const__(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn_uint8_t__const__, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CIntFromPyVerify */
      #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* MemviewDtypeToObject */
      static CYTHON_INLINE PyObject *__pyx_memview_get_Py_ssize_t(const char *itemp) {
    return (PyObject *) PyInt_FromSsize_t(*(Py_ssize_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_Py_ssize_t(const char *itemp, PyObject *obj) {
    Py_ssize_t value = __Pyx_PyIndex_AsSsize_t(obj);
    if (unlikely((value == (Py_ssize_t)-1) && PyErr_Occurred()))
        return 0;
    *(Py_ssize_t *) itemp = value;
    return 1;
}

/* ObjectToMemviewSlice */
      static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_Py_ssize_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_Py_ssize_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* MemviewSliceCopyTemplate */
      static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (unlikely(from_mvs->suboffsets[i] >= 0)) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* MemviewSliceInit */
      static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (unlikely(memviewslice->memview || memviewslice->data)) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#if PY_VERSION_HEX >= 0x030A0000 || defined(HAVE_STDARG_PROTOTYPES)
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int_type *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int_type *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    __pyx_nonatomic_int_type old_acquisition_count;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        return;
    }
    old_acquisition_count = __pyx_add_acquisition_count(memview);
    if (unlikely(old_acquisition_count <= 0)) {
        if (likely(old_acquisition_count == 0)) {
            if (have_gil) {
                Py_INCREF((PyObject *) memview);
            } else {
                PyGILState_STATE _gilstate = PyGILState_Ensure();
                Py_INCREF((PyObject *) memview);
                PyGILState_Release(_gilstate);
            }
        } else {
            __pyx_fatalerror("Acquisition count is %d (line %d)",
                             old_acquisition_count+1, lineno);
        }
    }
}
static CYTHON_INLINE void __Pyx_XCLEAR_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    __pyx_nonatomic_int_type old_acquisition_count;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        memslice->memview = NULL;
        return;
    }
    old_acquisition_count = __pyx_sub_acquisition_count(memview);
    memslice->data = NULL;
    if (likely(old_acquisition_count > 1)) {
        memslice->memview = NULL;
    } else if (likely(old_acquisition_count == 1)) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         old_acquisition_count-1, lineno);
    }
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint8_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint8_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API
        return _PyLong_FromByteArray(bytes, sizeof(uint8_t),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyInt_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint8_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        kwds = PyDict_New();
        if (!kwds) goto limited_bad;
        if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(!is_unsigned ? Py_True : Py_False))) goto limited_bad;
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(from_bytes);
        Py_XDECREF(py_bytes);
        Py_XDECREF(order_str);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(kwds);
        return result;
#endif
    }
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyInt_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        kwds = PyDict_New();
        if (!kwds) goto limited_bad;
        if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(!is_unsigned ? Py_True : Py_False))) goto limited_bad;
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(from_bytes);
        Py_XDECREF(py_bytes);
        Py_XDECREF(order_str);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(kwds);
        return result;
#endif
    }
}

/* CIntFromPy */
      static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int64_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) >= 2 * PyLong_SHIFT)) {
                                return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) >= 3 * PyLong_SHIFT)) {
                                return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) >= 4 * PyLong_SHIFT)) {
                                return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int64_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(int64_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(int64_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int64_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            int64_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (int64_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (int64_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (int64_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (int64_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (int64_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(int64_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((int64_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(int64_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((int64_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((int64_t) 1) << (sizeof(int64_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (int64_t) -1;
        }
    } else {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyInt_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyInt_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        kwds = PyDict_New();
        if (!kwds) goto limited_bad;
        if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(!is_unsigned ? Py_True : Py_False))) goto limited_bad;
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(from_bytes);
        Py_XDECREF(py_bytes);
        Py_XDECREF(order_str);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(kwds);
        return result;
#endif
    }
}

/* CIntFromPy */
      static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                                return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                                return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                                return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(int) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                                return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                                return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                                return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(int) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (int) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (int) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (int) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (int) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (int) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((int) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((int) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
      static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(size_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                                return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                                return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                                return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(size_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(size_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (size_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (size_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (size_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (size_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (size_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((size_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((size_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
      static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                                return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                                return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                                return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(long) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                                return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                                return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                                return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(long) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (long) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (long) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (long) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (long) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (long) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((long) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((long) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API
        return _PyLong_FromByteArray(bytes, sizeof(uint64_t),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyInt_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        kwds = PyDict_New();
        if (!kwds) goto limited_bad;
        if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(!is_unsigned ? Py_True : Py_False))) goto limited_bad;
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(from_bytes);
        Py_XDECREF(py_bytes);
        Py_XDECREF(order_str);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(kwds);
        return result;
#endif
    }
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyInt_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        kwds = PyDict_New();
        if (!kwds) goto limited_bad;
        if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(!is_unsigned ? Py_True : Py_False))) goto limited_bad;
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(from_bytes);
        Py_XDECREF(py_bytes);
        Py_XDECREF(order_str);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(kwds);
        return result;
#endif
    }
}

/* CIntFromPy */
      static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint32_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT)) {
                                return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT)) {
                                return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT)) {
                                return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(uint32_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(uint32_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint32_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            uint32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (uint32_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (uint32_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (uint32_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (uint32_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (uint32_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(uint32_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((uint32_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(uint32_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((uint32_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((uint32_t) 1) << (sizeof(uint32_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (uint32_t) -1;
        }
    } else {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyInt_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyInt_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        kwds = PyDict_New();
        if (!kwds) goto limited_bad;
        if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(!is_unsigned ? Py_True : Py_False))) goto limited_bad;
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(from_bytes);
        Py_XDECREF(py_bytes);
        Py_XDECREF(order_str);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(kwds);
        return result;
#endif
    }
}

/* CIntFromPy */
      static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint64_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) >= 2 * PyLong_SHIFT)) {
                                return (uint64_t) (((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) >= 3 * PyLong_SHIFT)) {
                                return (uint64_t) (((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) >= 4 * PyLong_SHIFT)) {
                                return (uint64_t) (((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint64_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(uint64_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(uint64_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint64_t) (((uint64_t)-1)*(((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint64_t) ((((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint64_t) (((uint64_t)-1)*(((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint64_t) ((((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint64_t) (((uint64_t)-1)*(((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint64_t) ((((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(uint64_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint64_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            uint64_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (uint64_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (uint64_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (uint64_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (uint64_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (uint64_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(uint64_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((uint64_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(uint64_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((uint64_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((uint64_t) 1) << (sizeof(uint64_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (uint64_t) -1;
        }
    } else {
        uint64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint64_t) -1;
        val = __Pyx_PyInt_As_uint64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint64_t");
    return (uint64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint64_t");
    return (uint64_t) -1;
}

/* CIntFromPy */
      static CYTHON_INLINE int32_t __Pyx_PyInt_As_int32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int32_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(int32_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int32_t) >= 2 * PyLong_SHIFT)) {
                                return (int32_t) (((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int32_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int32_t) >= 3 * PyLong_SHIFT)) {
                                return (int32_t) (((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int32_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int32_t) >= 4 * PyLong_SHIFT)) {
                                return (int32_t) (((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(int32_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(int32_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (int32_t) (((int32_t)-1)*(((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(int32_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (int32_t) ((((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (int32_t) (((int32_t)-1)*(((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int32_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (int32_t) ((((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (int32_t) (((int32_t)-1)*(((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int32_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (int32_t) ((((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(int32_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int32_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            int32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (int32_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (int32_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (int32_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (int32_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (int32_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(int32_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((int32_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(int32_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((int32_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((int32_t) 1) << (sizeof(int32_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (int32_t) -1;
        }
    } else {
        int32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int32_t) -1;
        val = __Pyx_PyInt_As_int32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int32_t");
    return (int32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int32_t");
    return (int32_t) -1;
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API
        return _PyLong_FromByteArray(bytes, sizeof(int32_t),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyInt_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        kwds = PyDict_New();
        if (!kwds) goto limited_bad;
        if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(!is_unsigned ? Py_True : Py_False))) goto limited_bad;
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(from_bytes);
        Py_XDECREF(py_bytes);
        Py_XDECREF(order_str);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(kwds);
        return result;
#endif
    }
}

/* CIntFromPy */
      static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(char) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(char, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(char) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) >= 2 * PyLong_SHIFT)) {
                                return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(char) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) >= 3 * PyLong_SHIFT)) {
                                return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(char) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) >= 4 * PyLong_SHIFT)) {
                                return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(char) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(char) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(char, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(char) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) - 1 > 2 * PyLong_SHIFT)) {
                                return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(char) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) - 1 > 2 * PyLong_SHIFT)) {
                                return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(char) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) - 1 > 3 * PyLong_SHIFT)) {
                                return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(char) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) - 1 > 3 * PyLong_SHIFT)) {
                                return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(char) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) - 1 > 4 * PyLong_SHIFT)) {
                                return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(char) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) - 1 > 4 * PyLong_SHIFT)) {
                                return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(char) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(char) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (char) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (char) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (char) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (char) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (char) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(char) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((char) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(char) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((char) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((char) 1) << (sizeof(char) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CheckBinaryVersion */
      static unsigned long __Pyx_get_runtime_version() {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* InitStrings */
      #if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
